//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "BinaryObjectStreams.h"
//---------------------------------------------------------------------------
#include "X4Exception.h"
#include "MetaClass.h"
#include "StreamClassTable.h"
#include <boost/lexical_cast.hpp>

//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------


//---------------------------------------------------------------------
BinaryUpdateObjectStream::BinaryUpdateObjectStream(const std::string& filename,
   StreamClassTable *classTable, bool doWriteClassTable, BinaryStreamCallback* callback, bool _initStream)
: BinaryOutputObjectStream(classTable, doWriteClassTable)
{
	_callback = callback;

	// if we would like to update the stream, then we have to open it in read/write mode, hence we need first to check if it does exists
	bool exists = true;
	if (fopen_s(&file, filename.c_str(), "r") != 0 || file == NULL)
		exists = false;
	else
		fclose(file);

	fname = filename;

	errno_t result = 0;
	result = fopen_s(&file, filename.c_str(), "r+bR");

	if (result != 0 || file == NULL)
	{
		std::string fileString = filename;
		throw EIOException(std::string("BinaryUpdateObjectStream- could not open file: " + filename + ", error = " + boost::lexical_cast<std::string>(result) ));
	}

	if (_callback)
		_callback->onOpenFile(filename, 0);

	if (_initStream)
		initStream();
}

//---------------------------------------------------------------------
OutputStream* BinaryUpdateObjectStream::createSiblingStream(card64 uid, bool raw)
{
	// use uid to just generate new filenames. this will make sure that siblings are found in the same directory and can be correctly idenitfied
	std::string filename = fname + std::string("__") + boost::lexical_cast<std::string>(uid);
	return new BinaryUpdateObjectStream(filename, classTable, raw ? false : doWriteClassTable, _callback, !raw);
}


//---------------------------------------------------------------------
BinaryOutputObjectStream::BinaryOutputObjectStream(const std::string& filename,
   StreamClassTable *classTable, bool doWriteClassTable, BinaryStreamCallback* callback, bool _initStream)
: OutputObjectStream(classTable, doWriteClassTable),
   _callback(callback)
{
	// if we would like to update the stream, then we have to open it in read/write mode, hence we need first to check if it does exists
	fname = filename;

	errno_t result = fopen_s(&file, filename.c_str(), "w+bR");

	if (result != 0 || file == NULL)
	{
		std::string fileString = filename;
		throw EIOException(std::string("BinaryOutputObjectStream- could not open file: " + filename + ", error = " + boost::lexical_cast<std::string>(result) ));
	}

	if (_callback)
		_callback->onOpenFile(filename, 0);

	if (_initStream)
		initStream();
}

//---------------------------------------------------------------------
BinaryOutputObjectStream::~BinaryOutputObjectStream()
{
	if (doWriteClassTable)
		writeClassTable();

	if (file != NULL)
		fclose(file);

	if (_callback)
		_callback->onCloseFile();
}

//---------------------------------------------------------------------
void BinaryOutputObjectStream::reopen(const std::string& filename)
{
	if (file != NULL)
	{
		writeClassTable();
		fclose(file);

		if (_callback)
			_callback->onCloseFile();
	}

	fname = filename;
	if (fopen_s(&file, filename.c_str(), "wbR") != 0 || file == NULL)
	{
		std::string fileString = filename;
		throw EIOException("BinaryOutputObjectStream::reopen() - could not open file." );
	}
	if (_callback)
		_callback->onOpenFile(filename, 0);

	initStream();
}

//---------------------------------------------------------------------
OutputStream* BinaryOutputObjectStream::createSiblingStream(card64 uid, bool raw)
{
	// use uid to just generate new filenames. this will make sure that siblings are found in the same directory and can be correctly idenitfied
	std::string filename = fname + std::string("__") + boost::lexical_cast<std::string>(uid);
	return new BinaryOutputObjectStream(filename, classTable, raw ? false : doWriteClassTable, _callback, !raw);
}

//---------------------------------------------------------------------
void BinaryOutputObjectStream::writeClassTableHeader()
{
	// Place holder for pointer to class table, will be filled later in writeClassTable()
	__int64 fpos = _ftelli64(file);
	if (fpos < 0)
		throw EIOException("BinaryOutputObjectStream::writeClassTable() - could not get file position");
	classTabPointerPos = fpos;
	write<card64>(0);
}

//---------------------------------------------------------------------
void BinaryOutputObjectStream::writeBuffer(const void* buf, mpcard length)
{
	if (length > 0x3FFFFFFFL)
	{
		mpcard blocks = length / 0x40000000L;
		mpcard rem = length - blocks*0x40000000L;
		for (mpcard i=0; i<blocks; i++)
		{
			if (fwrite(((card8*)buf)+i*0x40000000L, 0x40000000L, 1, file) != 1) 
				throw EIOException("BinaryOutputObjectStream::writeBuffer - could not write all data.");
		}
		if (rem > 0)
		{
			if (fwrite(((card8*)buf)+blocks*0x40000000L, rem, 1, file) != 1) 
				throw EIOException("BinaryOutputObjectStream::writeBuffer - could not write all data.");
		}
	} else {
		if (fwrite(buf, length, 1, file) != 1 && length > 0) 
			throw EIOException("BinaryOutputObjectStream::writeBuffer - could not write all data.");
	}

	if (_callback)
		_callback->onWriteData(buf, length);
}

//---------------------------------------------------------------------
void BinaryOutputObjectStream::writeClassTable()
{
	__int64 fpos = _ftelli64(file);
	if (fpos < 0)
		throw EIOException("BinaryOutputObjectStream::writeClassTable() - could not get file position");
	card64 classTablePos = fpos;

	classTable->write(this);

	if (_fseeki64(file, classTabPointerPos, SEEK_SET) != 0)
		throw EIOException("BinaryOutputObjectStream::writeClassTable() - could not seek to class table pointer");
	write<card64>(classTablePos);
}


//---------------------------------------------------------------------
card64 BinaryOutputObjectStream::seek(int64 rel)
{
	__int64 fpos = _ftelli64(file);
	if (fpos == -1L)
		throw EIOException(std::string("BinaryOutputObjectStream::seek() - can not seek: ") + strerror(errno) );

	if (rel == 0)
		return fpos;

	if (_fseeki64(file, rel, SEEK_CUR))
		throw EIOException(std::string("BinaryOutputObjectStream::seek() - can not seek in the stream") );

	return fpos + rel;
}







//---------------------------------------------------------------------
BinaryInputObjectStream::BinaryInputObjectStream(const std::string& filename, BinaryStreamCallback* callback, bool _initStream) : 
   fileSize(0),
   _callback(callback)
{
	fname = filename;
	if (!fileReader.open(filename.c_str()))
		throw EIOException("BinaryInputObjectStream::BinaryInputObjectStream() - could not open file.");

	fileSize = fileReader.getFileSize();

	// if callback was specified, then call it back
	if (_callback)
		_callback->onOpenFile(filename, fileSize);

	if (_initStream)
		initStream();
}

//---------------------------------------------------------------------
InputStream* BinaryInputObjectStream::createSiblingStream(card64 uid, bool raw)
{
	// use uid to just generate new filenames. this will make sure that siblings are found in the same directory and can be correctly idenitfied
	std::string filename = fname + std::string("__") + boost::lexical_cast<std::string>(uid);
	return new BinaryInputObjectStream(filename, _callback, !raw);
}

//---------------------------------------------------------------------
void BinaryInputObjectStream::readClassTable()
{
	card64 classTablePos;
	read<card64>(classTablePos);

	card64 start = fileReader.seek(0);
	fileReader.seek(classTablePos - start); 
	classTable->read(this);
	card64 stop = fileReader.seek(0);
	fileReader.seek((int64)start - (int64)stop); 

	fileReader.setUseBufferedRead(true);
}

//---------------------------------------------------------------------
void BinaryInputObjectStream::readBuffer(void* buf, mpcard length)
{
	if (length==0) 
		return;

	if( !fileReader.readBuffer(buf, length))
		throw EIOException("BinaryInputObjectStream::readBuffer - could not read all data.");

	if (_callback)
		_callback->onReadData(buf, length);

}

//---------------------------------------------------------------------
BinaryInputObjectStream::~BinaryInputObjectStream()
{
	if (fileReader.file && _callback)
		_callback->onCloseFile();
}

//---------------------------------------------------------------------
BufferedFileReader::BufferedFileReader()
{
	file = NULL;
	filePos = 0;
	fileSize = 0;
	bufferPos = 0;
	useBufferedRead = false;
}

//---------------------------------------------------------------------
BufferedFileReader::~BufferedFileReader()
{
	if (file != NULL)
		fclose(file);
}

//---------------------------------------------------------------------
bool BufferedFileReader::open( const std::string& filename )
{
	// TODO implement support for wide characters and longer file names (current windows limit 260 characters in the filename)
	if( fopen_s(&file, filename.c_str(), "rbS") != 0 || file == NULL)
		return false;

	_fseeki64(file, 0L, SEEK_END);
	fileSize = _ftelli64(file);
	_fseeki64(file, 0L, SEEK_SET);
	filePos = 0;
	return true;
}

//---------------------------------------------------------------------
card64 BufferedFileReader::seek(int64 rel)
{
	if (rel == 0) return filePos;

	if (!useBufferedRead)
	{
		if (_fseeki64(file, rel, SEEK_CUR))
			throw EIOException(std::string("BufferedFileReader::seek() - can not seek in the stream") );
	}else
	{
		int64 bpos = (int64)bufferPos + rel;

		if (bpos < 0 || bpos >= (int64)buffer.size())
		{
			bufferPos = 0;
			buffer.clear();
			_fseeki64(file, rel, SEEK_CUR);
		}else
			bufferPos = bpos;
	}

	filePos += rel;
	return filePos;
}

//---------------------------------------------------------------------
bool BufferedFileReader::readBuffer( void* buf, mpcard length )
{
	// too much data ?
	if( filePos + length > fileSize )
		return false;

	// not buffered, hence read in blocks of 2GB
	if( !useBufferedRead)
	{
		if (length > 0x3FFFFFFFL)
		{
			mpcard blocks = length / 0x40000000L;
			mpcard rem = length - blocks*0x40000000L;
			for (mpcard i=0; i<blocks; i++)
			{
				if (fread(((card8*)buf)+i*0x40000000L, 0x40000000L, 1, file) != 1)
					return false;
				filePos += 0x40000000L;
			}
			if (rem > 0)
			{
				if (fread(((card8*)buf)+blocks*0x40000000L, rem, 1, file) != 1)
					return false;
				filePos += rem;
			}
		} else {
			if (fread(buf, length, 1, file) != 1)
				return false;
			filePos += length;
		}
		return true;
	}

	if( buffer.empty())		
	{
		bufferPos = 0;
		buffer.resize(1024*1024,0);
		size_t size = std::min(fileSize - filePos, (card64)buffer.size());

		if (fread(&buffer[0], size, 1, file) != 1) return false;
	}

	// does fit into buffer, then copy directly from buffer
	if( bufferPos + length <= buffer.size() )
	{
		memcpy(buf, &buffer[bufferPos], length);
		bufferPos += length;
		filePos += length;
		return true;
	}else
	{
		// just read directly from file, since it doesn't fit into the buffer
		_fseeki64(file,filePos,SEEK_SET);
		if( fread(buf, length, 1, file) != 1 )
			return false;
					
		filePos += length;
		bufferPos = 0;
		size_t size = std::min(fileSize-filePos, (card64)buffer.size());
		if (size > 0)
			return fread(&buffer[0],size,1,file) == 1;
		return true;
	}
}

//---------------------------------------------------------------------
const std::string BufferedFileReader::getWorkingPath()
{
	/*char buffer[FILENAME_MAX];

	if (!GetCurrentDir(buffer, sizeof(buffer)))
		return std::string(".");

	buffer[sizeof(buffer) - 1] = '/0';
	return std::string(buffer);
	*/

	char buffer[MAX_PATH];
	GetModuleFileNameA( NULL, buffer, MAX_PATH );
	std::string::size_type pos = std::string( buffer ).find_last_of( "\\/" );
	return std::string( buffer ).substr( 0, pos);
}
