//---------------------------------------------------------------------------
#ifndef BinaryObjectStreamsH
#define BinaryObjectStreamsH
//---------------------------------------------------------------------------
#include "x4core.h"
//---------------------------------------------------------------------------
#include "ObjectStreams.h"
//---------------------------------------------------------------------------
#include <stdio.h>

//---------------------------------------------------------------------------

namespace X4 {

	/**
	* callback for use with binary streams, to be informed on data read/write events
	**/
	class X4CORE_API BinaryStreamCallback
	{
		public:
			BinaryStreamCallback(){};
			virtual ~BinaryStreamCallback(){};

			virtual void onOpenFile(const std::string& filename, size_t fileSize) {forget(filename); forget(fileSize);};
			virtual void onReadData(const void* buf, mpcard length) {forget(buf); forget(length);};
			virtual void onWriteData(const void* buf, mpcard length) {forget(buf); forget(length);};
			virtual void onCloseFile() {};
	};

	/**
	* Object writer class used on binary files.
	**/
	class X4CORE_API BinaryOutputObjectStream : public OutputObjectStream
	{
		protected:
			FILE *file;     
			std::string fname;
			card64 classTabPointerPos;
			BinaryStreamCallback* _callback;

		protected:
			virtual void writeClassTableHeader();
			virtual void writeClassTable();

			BinaryOutputObjectStream(StreamClassTable *classTable = NULL, bool doWriteClassTable = true) : OutputObjectStream(classTable, doWriteClassTable), _callback(NULL), file(NULL) {}

		public:
			BinaryOutputObjectStream(const std::string& filename, StreamClassTable *classTable = NULL, bool doWriteClassTable = true, BinaryStreamCallback* callback = NULL, bool initStream = true);
			virtual ~BinaryOutputObjectStream();

			virtual void reopen(const std::string& filename);
			virtual void writeBuffer(const void* buf, mpcard length);
			inline virtual std::string getFileName() { return fname; }
			FILE* getFilePointer() { return file; }
			virtual card64 seek(int64 rel);
			virtual bool canSeek() const { return true; }

			virtual OutputStream* createSiblingStream(card64 uid, bool raw = false);

			virtual std::string getFileNameIfFileStream() {return fname;}
	};
	
	/**
	* Object writer class used on binary files, the stream is, however, opened in update mode.
	* This means, that the file is overwritten instead of being deleted. This can be used to modify
	* the content of the file without rewritting it completly.
	*
	* Use always this streams in order to update the data! Update can be achieved within the write() method
	* of an object dervied from Persistent.
	**/
	class X4CORE_API BinaryUpdateObjectStream : public BinaryOutputObjectStream
	{
		public:
			BinaryUpdateObjectStream(const std::string& filename, StreamClassTable *classTable = NULL, bool doWriteClassTable = true, BinaryStreamCallback* callback = NULL, bool initStream = true);
			virtual ~BinaryUpdateObjectStream() {}

			virtual OutputStream* createSiblingStream(card64 uid, bool raw = false);
	};

	/**
	* Buffered reader for buffered reads from binary files
	**/
	class X4CORE_API BufferedFileReader
	{
	public:
		BufferedFileReader();
		~BufferedFileReader();

		bool open(const std::string& filename);
		bool readBuffer(void* buf, mpcard length);
		card64 getFileSize() const {return fileSize;}
		void setUseBufferedRead(bool b) {useBufferedRead=b;}
		virtual card64 seek(int64 rel);

		static const std::string getWorkingPath();

		FILE * file;
	private:

		std::vector<unsigned char> buffer;
		card64 filePos;
		card64 bufferPos;
		card64 fileSize;
		bool useBufferedRead;
	};

	/**
	* Read objects from binary files
	**/
	class X4CORE_API BinaryInputObjectStream : public InputObjectStream
	{
		private:
		   //FILE *file;
		   BufferedFileReader fileReader;
		   std::string fname;
		   card64 fileSize;
		   BinaryStreamCallback* _callback;

		protected:
		   virtual void readClassTable();

		public:
			BinaryInputObjectStream(const std::string& filename, BinaryStreamCallback* callback = NULL, bool initStream = true);
			virtual ~BinaryInputObjectStream();

			virtual void readBuffer(void* buf, mpcard length);
			inline virtual std::string getFileName() { return fname; }
			inline virtual card64 seek(int64 rel) { return fileReader.seek(rel); }
			FILE* getFilePointer() { return fileReader.file; }
			virtual InputStream* createSiblingStream(card64 uid, bool raw = false);
			virtual std::string getFileNameIfFileStream() {return fname;}
	};

}

#endif
