//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "BufferObjectStreamComparison.h"
//---------------------------------------------------------------------------
#include "x4core.h"
#include <X4Exception.h>
#include <Persistent.h>
#include <MetaClass.h>
#include <StreamClassTable.h>
#include <BufferObjectStreams.h>
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------


BufferOutputObjectStreamComparison::BufferOutputObjectStreamComparison(bool perObjectClassVersion, StreamClassTable *classTable, bool doWriteClassTable) :
		BufferOutputObjectStream(perObjectClassVersion, classTable, doWriteClassTable), _objPtrCounter(0)
{
}

void 
BufferOutputObjectStreamComparison::removeFromCache(Persistent *obj)
{
	_objPtrMap.erase((mpcard)obj);
}

void 
BufferOutputObjectStreamComparison::clearCache()
{
	_objPtrMap.clear();
}

void
BufferOutputObjectStreamComparison::writeObject(const Persistent * obj)
{
	// Generate new class id or read existing class id from hash table
	card32 classId = 0;
	if (obj != NULL)
	{
		classId = classTable->getOrCreateClassID(obj->getInstanceClass());

		#ifdef _DEBUG
		if (obj != NULL) {
			std::string cName = obj->getInstanceClass()->getClassName();
			if (!MetaClass::isClassRegistered(cName)) {
				//error("OutputObjectStream::writeObject() - writing a class that has not been registered.");
				x4Assert(false);
			}
		}
		#endif

		// Also put all parent classes into the class table
		const MetaClass* mc = (obj->getInstanceClass())->getParentClass();
		while (mc != NULL) {
			classTable->getOrCreateClassID(mc);
			mc = mc->getParentClass();
		}
	}

#ifdef DEBUG_STREAMS
	static card32 DEBUG_IDENT = 0;
	debugOutput.writeIdented(3*DEBUG_IDENT, ((std::string("w objID: ") + intToStr((card64)obj))));
	debugOutput.cr();
	if (obj == NULL) {
		debugOutput.writeIdented(3*DEBUG_IDENT, "w class: NULL");
		debugOutput.cr();
	} else {
		debugOutput.writeIdented(3*DEBUG_IDENT, (std::string("w class: ") + obj->getInstanceClass()->getClassName()));
		debugOutput.cr();
		debugOutput.writeIdented(3*DEBUG_IDENT, (std::string("w classID: ") + intToStr(classId)));
		debugOutput.cr();
	}
#endif

	if (obj == NULL)
	{
		write<card64>(0);
	} else
	{
		card32 nObjPtrCount = _getNumRegisteredObjectPointers();
		card64 objID = 0;

		// Write object if it wasn't written before
		if (!_getObjectID(objID, (const Object*)obj))
		{
			write<card64>(objID);

			write<card32>(classId);

			// get current position of the stream. we use this to write here the size
			// of bytes the object takes in the stream after it has been written
			WriteMarker objSizeMarker = createMarker();
			write<card64>(0); // reserve space to write here later the size of the object

			#ifdef DEBUG_STREAMS
			DEBUG_IDENT++;
			#endif

			obj->write(this);

			#ifdef DEBUG_STREAMS
			DEBUG_IDENT--;
			#endif

			// go back in the stream and write there the object size
			WriteMarker afterObjMarker = createMarker();
			jumpToMarker(objSizeMarker);
			write<card64>(afterObjMarker.difference(objSizeMarker) - sizeof(card64));
			jumpToMarker(afterObjMarker);
		} else
		{
			//x4Assert( ((card64)storedObjPtr > 0 && (card64)storedObjPtr < 100) );
			const card64 knownID = objID;
			x4Assert( nObjPtrCount == _getNumRegisteredObjectPointers());
			write<card64>(knownID);
		}
	}
}

bool operator==(const X4::Persistent & obj1, const X4::Persistent & obj2)
{
    BufferOutputObjectStreamComparison oStream1;
    BufferOutputObjectStreamComparison oStream2;
    //BufferOutputObjectStream oStream1;
    //BufferOutputObjectStream oStream2;

    oStream2.writeObject(&obj2);
    oStream2.finalize();
    oStream1.writeObject(&obj1);
    oStream1.finalize();

    const size_t msgDataLength1 = oStream1.getBufferSize();
    const size_t msgDataLength2 = oStream2.getBufferSize();
    if (msgDataLength1 != msgDataLength2)
        return false;

    const byte *msgData1 = oStream1.getBuffer();
    const byte *msgData2 = oStream2.getBuffer();
    for (size_t i=0; i < msgDataLength2; i++) {
        if (msgData1[i] != msgData2[i])
            return false;
    }
    return true;
}
