//---------------------------------------------------------------------------
#ifndef ClassMethodsH
#define ClassMethodsH
//---------------------------------------------------------------------------
#include "x4core.h"
//---------------------------------------------------------------------------
#include "Object.h"
#include "Variant.h"
#include "VariantList.h"
//---------------------------------------------------------------------------

namespace X4 {

    class ClassProperty;


    /// super class of all objects reflecting a method of a class
    class X4CORE_API ClassMethod : public Object
    {
    private:
        std::string name;
        std::string description;
        std::string icon;
        bool visible;
        /// if set (default = false), the user interface might create a command object for executing the method
        bool packageAsCmdObjHint;

    public:
        ClassMethod(std::string name, std::string description = "", std::string icon = "", bool visible = true);
        std::string getName() const;
        std::string getDescription() const;
        std::string getIcon() const;
        bool getVisible() const;
        void setVisible(bool v);
        bool getPackageAsCmdObjHint() const;
        void setPackageAsCmdObjHint(bool h);
        virtual void call(Object *obj) const {};
        virtual ClassMethod *copyMethod() const = 0;
        virtual bool isValid() const { return false; }
    };

    /// representing a separator, no function class, just for easier separating methods in the list
    class X4CORE_API SeparatorClassMethod : public ClassMethod
    {
    public:
        SeparatorClassMethod(std::string name, std::string icon = "", bool visible = true) : ClassMethod(name,"", icon,visible){}
        virtual ClassMethod *copyMethod() const{ return new SeparatorClassMethod(getName(), getIcon(), getVisible()); }
    };


    /// representing a method with no arguments
    class X4CORE_API NoArgsClassMethod : public ClassMethod
    {
    public:
        typedef void (Object::* MethodPointer)();

    private:
        MethodPointer method;

    public:
        NoArgsClassMethod(MethodPointer method, std::string name, std::string description = "", std::string icon = "", bool visible = true);
        virtual void call(Object *obj) const;
        virtual ClassMethod *copyMethod() const;
        virtual bool isValid() const { return method != NULL; }
    };


    /// representing a function which works like a dispatcher for other class methods
    /// so whenever the method is called, it will invoke another method to run
    class X4CORE_API NoArgsClassMethodDispatcher : public NoArgsClassMethod
    {
    public:
        typedef void (Object::* DispatcherMethodPointer)(ClassMethod*);

    private:
        DispatcherMethodPointer dispatcher;
        ClassMethod* method;

    public:
        NoArgsClassMethodDispatcher(DispatcherMethodPointer dispatcher, ClassMethod* method, std::string name, std::string description = "", std::string icon = "", bool visible = true);
        virtual ~NoArgsClassMethodDispatcher();
        virtual void call(Object *obj) const;
        virtual ClassMethod *copyMethod() const;
        virtual bool isValid() const { return method->isValid() && dispatcher != NULL; }
    };



    /// A general class method, with arguments consisting of basic types defined in X4Types.h (including "Object*").
    /// This thing works as follows:
    ///    Parameters are represented as X4VariantLists, return value is an X4Variant
    ///    When the method is registered, an invoking method has to be specified, which takes the variant as input and calls the actual method
    class X4CORE_API BasicArgsClassMethod : public ClassMethod
    {
    public:
        /// this is a pointer to an invocation method (which has to unmarshal parameters from / to a variant)
        typedef Variant (Object::* MethodPointer)(VariantList &);

    private:
        MethodPointer method;
        VariantTypeList parameterTypes;
        /// enum X4Variant::X4Type (card32 for persistence)
        card32 resultType; 

    public:
        /// constructor - use ADD_BASIC_ARGS_METHOD macro as simplified interface
        /// parameterTypes and resultType specify the types of the formal parameters and the type of the result
        BasicArgsClassMethod(MethodPointer invoker, std::string name, const VariantTypeList parameterTypes, const Variant::Type resultType,
            std::string description = "", std::string icon = "", bool visible = true);

        /// access formal parameter list
        VariantTypeList getParameterTypes() {return parameterTypes;}
        /// access result type
        Variant::Type getResultType() {return (Variant::Type)resultType;}

        virtual void call(Object *obj, VariantList &params, Variant &result) const;
        virtual ClassMethod *copyMethod() const;
        virtual bool isValid() const { return method != NULL; }
    };

    // ------------------------------------------------------------------------------------------------------------------------------------------------
    // -------------------------------------------------------------------- MACROS --------------------------------------------------------------------
    // ------------------------------------------------------------------------------------------------------------------------------------------------


    typedef NoArgsClassMethod::MethodPointer NoArgsMethodPointer;
    typedef NoArgsClassMethodDispatcher::DispatcherMethodPointer NoArgsMethodDispatcherPointer;
    typedef BasicArgsClassMethod::MethodPointer BasicArgsMethodPointer;


#define ADD_NOARGS_METHOD_EX( method, name, desc, icon, visible ) \
    ((MetaClass*)getClass())->addMethod( new NoArgsClassMethod( \
    (NoArgsMethodPointer) & method , name, desc, icon, visible ) );

#define ADD_NOARGS_METHOD( method )							ADD_NOARGS_METHOD_EX ( method, #method, "", "", true )
#define ADD_NOARGS_METHOD_DESC( method, desc )			ADD_NOARGS_METHOD_EX ( method, #method, desc, "", true )
#define ADD_NOARGS_METHOD_ICON( method, desc, icon )	ADD_NOARGS_METHOD_EX ( method, #method, desc, icon, true )
#define ADD_NOARGS_METHOD_INVISIBLE( method )			ADD_NOARGS_METHOD_EX ( method, #method, "", "", false )
#define SET_METHOD_VISIBLE( vis )                     ((X4::MetaClass *)getClass())->getActMethod()->setVisible( vis );
#define SET_METHOD_AS_CMD_OBJ( cmd )                  ((X4::MetaClass *)getClass())->getActMethod()->setPackageAsCmdObjHint( cmd );

#define ADD_BASIC_ARGS_METHOD_EX( method, name, parameterTypes, resultType, desc, icon, visible ) \
    ((MetaClass*)getClass())->addMethod( new BasicArgsClassMethod( \
    (BasicArgsMethodPointer) & method, name, parameterTypes, resultType, desc, icon, visible ) );

#define ADD_BASIC_ARGS_METHOD( method, parameterTypes, resultType )						 ADD_BASIC_ARGS_METHOD_EX ( method, #method, parameterTypes, resultType, "", "", true )
#define ADD_BASIC_ARGS_METHOD_DESC( method, parameterTypes, resultType, desc )		 ADD_BASIC_ARGS_METHOD_EX ( method, #method, parameterTypes, resultType, desc, "", true )
#define ADD_BASIC_ARGS_METHOD_ICON( method, parameterTypes, resultType, desc, icon ) ADD_BASIC_ARGS_METHOD_EX ( method, #method, parameterTypes, resultType, desc, icon, true )
#define ADD_BASIC_ARGS_METHOD_INVISIBLE( method, parameterTypes, resultType )			 ADD_BASIC_ARGS_METHOD_EX ( method, #method, parameterTypes, resultType, "", "", false )


#define ADD_SEPARATOR_METHOD_EX( name, icon, visible ) \
    ((MetaClass*)getClass())->addMethod( new SeparatorClassMethod( name, icon, visible ) );
#define ADD_SEPARATOR_METHOD( name )	ADD_SEPARATOR_METHOD_EX ( name, "", true )

    // dispatcher methods
#define ADD_NOARGS_METHOD_DISPATCHER_EX( dispatcher, classMethod, name, desc, icon, visible ) \
    ((MetaClass*)getClass())->addMethod( new NoArgsClassMethodDispatcher( \
    dispatcher, (ClassMethod*) classMethod , name, desc, icon, visible ) );

#define ADD_NOARGS_METHOD_DISPATCHER( dispatcher, method )\
    ADD_NOARGS_METHOD_DISPATCHER_EX ( (NoArgsMethodDispatcherPointer) & dispatcher, new NoArgsClassMethod((NoArgsMethodPointer) & method, #method, "", "", false ), #method, "", "", true )

#define ADD_NOARGS_METHOD_DISPATCHER_DESC( dispatcher, method, desc)\
    ADD_NOARGS_METHOD_DISPATCHER_EX ( (NoArgsMethodDispatcherPointer) & dispatcher, new NoArgsClassMethod((NoArgsMethodPointer) & method, #method, desc, "", false ), #method, desc, "", true )

}

#endif
