//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "ClassProperty.h"
//---------------------------------------------------------------------------
#include <Tools.h>
#include "ObjectStreams.h"
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

IMPLEMENT_X4_ABSTRACT_CLASS(ClassProperty, 0)
{
	BEGIN_CLASS_INIT(ClassProperty);
}

ClassProperty::ClassProperty() {
   initialize("", 1);
}

ClassProperty::PropClassToEditListMap ClassProperty::propertyTypeDefaultEditors;


ClassProperty::ClassProperty(const std::string cName, const Version introducedInVersion)
{
    initialize(cName, introducedInVersion);
}

void ClassProperty::initialize(const std::string& cName, const Version& introducedInVersion)
{
	this->cName = cName;
	this->introducedInVersion = introducedInVersion;
	this->description = "";
	this->userEdit = true;
	this->editReadonly = false;
	this->streamProperty = true;
	this->compoundAble = true;
	this->ownerClass = NULL;
}

void ClassProperty::setDescription(const std::string description) {
   this->description = description;
}

void ClassProperty::addPropertyEditor(const MetaClass *edClass, bool defaultEditor) {
   if (defaultEditor) {
      editors.push_front(edClass);
   } else {
      editors.push_back(edClass);
   }
}

const std::string ClassProperty::getCName() const {
   return cName;
}

const std::string ClassProperty::getDescription() const {
   return description;
}

const Version ClassProperty::getIntroducedInVersion() const {
   return introducedInVersion;
}

const MetaClass *ClassProperty::getPropertyEditor(card32 i) const {
   // Indices 0 to getNumEditorByType()-1 address type-based editors,
   // indices getNumEditorByType() to getNumProportyEditors()-1 address property-specific editors.

   if (i >= getNumPropertyEditors()) {
      throw ERangeCheck("ClassProperty::getPropertyEditor: index out of range.");
   }

   if (i < getNumEditorByType(getInstanceClass())) {
	  return getEditorByType(getInstanceClass(), i);
   } else {
	  return editors[i - getNumEditorByType(getInstanceClass())];
   }
}

card32 ClassProperty::getNumPropertyEditors() const {
   return getNumEditorByType(getInstanceClass()) + (card32)editors.size();
}

const MetaClass *ClassProperty::getDefaultPropertyEditor() const {
   if (editors.size() == 0) {
      const MetaClass *edClass = getDefaultEditorByType(getInstanceClass());
      return edClass; // NULL if not defined
   } else {
      return getPropertyEditor(0);
   }
}

void ClassProperty::setUserEdit(bool userEdit) {
   this->userEdit = userEdit;
}

void ClassProperty::setEditReadonly(bool editReadonly) {
   this->editReadonly = editReadonly;
}

void ClassProperty::setStreamProperty(bool streamProperty) {
   this->streamProperty = streamProperty;
}

void ClassProperty::assign(const Object* obj, X4CopyContext *context) {
   forget(obj);
   forget(context);
   throw X4Exception("not implemented");
}

void ClassProperty::write(const Object* obj, pointer_diff offset, OutputObjectStream *out) const
{
	// reserve size for the property and write property afterwards
	OutputStream::WriteMarker propSizeMarker = out->createMarker();
	out->write<card64>(0);
	writeProperty(obj, offset, out);

	// seek temporarly back and write the size of the property we have written
	OutputStream::WriteMarker afterPropMarker = out->createMarker();
	out->jumpToMarker(propSizeMarker);
	out->write<card64>(afterPropMarker.difference(propSizeMarker) - sizeof(card64));
	out->jumpToMarker(afterPropMarker);
}

void ClassProperty::read(Object* obj, pointer_diff offset, InputObjectStream *in) const
{
	// read size of the property and parse property
	card64 oldpos = in->seek(0);
	card64 size = 0;
	in->read<card64>(size);
	readProperty(obj, offset, in, size);
	size_t newpos = in->seek(0);

	// compare if we have parsed as much bytes as allowed, otherwise throw an error
	if (newpos - oldpos - sizeof(card64) != size)
		throw EIOException(std::string(getTypeName() + std::string("::readProperty() - size of the property does not match the one in the stream.")));
}

#if 0
void ClassProperty::skip(InputObjectStream *in)
{
	size_t size = 0;
	in->read<card64>(size);
	in->seek(size);
}
#endif 

void ClassProperty::addPropertyEditorByType( const MetaClass *editedType, const MetaClass *editor )
{
   PropClassToEditListMap::iterator pos = propertyTypeDefaultEditors.find(editedType);
   if (pos == propertyTypeDefaultEditors.end()) {
      PropertyTypeDefaultEditors list;
      list.push_back(editor);
      propertyTypeDefaultEditors.insert(std::pair<const MetaClass*, PropertyTypeDefaultEditors>(editedType, list));
   } else {
      pos->second.push_back(editor);
   }
}

const MetaClass *ClassProperty::getDefaultEditorByType( const MetaClass *type )
{
   PropClassToEditListMap::iterator pos = propertyTypeDefaultEditors.find(type);
   if (pos == propertyTypeDefaultEditors.end()) {
      return NULL;
   } else {
      if (pos->second.empty()) return NULL;
      return pos->second.back();
   }
}

const MetaClass * ClassProperty::getEditorByType( const MetaClass *type, mpcard index )
{
   PropClassToEditListMap::iterator pos = propertyTypeDefaultEditors.find(type);
   if (pos == propertyTypeDefaultEditors.end()) {
      return 0;
   } else {
      if (pos->second.size() > index) {
         return pos->second[index];
      } else {
         x4Assert(false);
         return NULL;
      }
   }
}

mpcard ClassProperty::getNumEditorByType( const MetaClass *type )
{
   PropClassToEditListMap::iterator pos = propertyTypeDefaultEditors.find(type);
   if (pos == propertyTypeDefaultEditors.end()) {
      return 0;
   } else {
      return pos->second.size();
   }
}
