//---------------------------------------------------------------------------
#ifndef ClassPropertyH
#define ClassPropertyH
//---------------------------------------------------------------------------
#include "x4core.h"
//---------------------------------------------------------------------------
#include <string>
#include <deque>
#include <map>
#include "X4Exception.h"
#include "X4Types.h"
#include "Object.h"
//---------------------------------------------------------------------------


namespace X4 {
   
	class InputObjectStream;
	class OutputObjectStream;
	class PropertyEditor;
	class MetaClass;

	typedef card32 Version;


	class X4CORE_API EUnparsable : public X4Exception
	{
	public:
		EUnparsable(const std::string msg) : X4Exception(msg) {}
		virtual std::string getTypeString() {return "EUnparsable";}
	};


	class X4CORE_API ClassProperty : public Object
	{
		X4_ABSTRACT_CLASS(ClassProperty);
   protected:
      typedef std::deque<const MetaClass*> PropertyTypeDefaultEditors;
	private:
      // --- default editors according to property type (class related)
      /// maps from type of property to list of default editors
      typedef std::map<const MetaClass*, PropertyTypeDefaultEditors> PropClassToEditListMap;
      static PropClassToEditListMap propertyTypeDefaultEditors;
   
		PropertyTypeDefaultEditors editors;
		std::string cName;
		std::string description;
		Version introducedInVersion;
		bool userEdit;
		bool editReadonly;
		bool streamProperty;
		const MetaClass* ownerClass;   // Owner of this property
		bool compoundAble;

	protected:
		ClassProperty();

		/// write method must be overwritten by derivedclasses to be able to write the property to a stream
		virtual void             writeProperty(const Object* obj, pointer_diff offset, OutputObjectStream *out) const = 0;

		/// read property from a stream
		virtual void             readProperty(Object* obj, pointer_diff offset, InputObjectStream *in, card64 expectedSize) const = 0;

      static mpcard           getNumEditorByType(const MetaClass *type);
      static const MetaClass *getEditorByType(const MetaClass *type, mpcard index);
      static const MetaClass *getDefaultEditorByType(const MetaClass *type);

	public:
		/// constructor: use MetaClass::FIRST_VERSION + i to define introduction of property in class revision i
		ClassProperty(const std::string cName, const Version introducedInVersion);
		virtual void             setDescription(const std::string description);
		virtual void             addPropertyEditor(const MetaClass *edClass, bool defaultEditor = false);
		virtual void             assign(const Object* obj, X4CopyContext *context = NULL);

		virtual const std::string     getTypeName() const = 0;
		virtual const std::string     getCName() const;
		virtual const std::string     getDescription() const;
		virtual const Version    getIntroducedInVersion() const;
		virtual std::string           toString(const Object* obj, pointer_diff offset) const = 0;
		virtual void             fromString(Object* obj, pointer_diff offset, const std::string str) const = 0; // throws EUnparsable
		virtual const MetaClass *getPropertyEditor(card32 i) const;
		virtual const MetaClass *getDefaultPropertyEditor() const;
		virtual card32           getNumPropertyEditors() const;
		virtual void             assignProperty(const Object *from, Object *to, pointer_diff offset) const = 0;


		void write(const Object* obj, pointer_diff offset, OutputObjectStream *out) const;
		void read(Object* obj, pointer_diff offset, InputObjectStream *in) const;
		//static void skip(InputObjectStream *in);

		void setUserEdit(bool userEdit);             // suitable for interactive editing, default = true
		void setEditReadonly(bool editReadonly);     // property cannot be changed interactively, default = false
		void setStreamProperty(bool streamProperty); // property will be written in object stream, default = true
		void setOwner(const MetaClass* ownerClass) {this->ownerClass = ownerClass;}
		void setCompoundAble( bool b ) {compoundAble = b;}
		bool getUserEdit() const {return userEdit;}
		bool getEditReadonly() const {return editReadonly;}
		bool getStreamProperty() const {return streamProperty;}
		const MetaClass* getOwner() const {return ownerClass;}
		bool getCompoundAble() const {return compoundAble;}
		void setCName( std::string name ) {cName = name;}

		virtual ClassProperty* copyClassProperty() = 0;

      /// Adds a property editor that is associated with the type, not the instance of the property.
      /// Type-related editors are used as default in case no instance-based editor is defined.
      /// The editor added last has priority.
      static void addPropertyEditorByType(const MetaClass *editedType, const MetaClass *editor);

    private:
        void initialize(const std::string& cName, const Version& introducedInVersion);
	};

}


#endif
