//---------------------------------------------------------------------------
#ifndef InputStream__H
#define InputStream__H
//---------------------------------------------------------------------------
#include "x4core.h"
#include "StreamContext.h"
//---------------------------------------------------------------------------
#include <string>
//---------------------------------------------------------------------------
#include <X4Types.h>
#include <X4Exception.h>
#include "PropertyTable.h"
#include <Tools.h>
//---------------------------------------------------------------------------

namespace X4
{
	/**
	* Serialization of data is handled by classes derived from this abstract definition
	**/
	class X4CORE_API InputStream
	{
	protected:
		StreamContext* m_context;

	public:
		InputStream() : m_context(NULL) {}
		virtual ~InputStream() { if (m_context) delete m_context; }

		//! Return context associated with this stream
		StreamContext* getContext() { return m_context; }

		//! Set context to be used with this stream. The stream takes ownership of the context
		void setContext(StreamContext* ctx) { if (m_context) delete m_context; m_context = ctx; }

		/**
		* Generate new stream which inherits all the properties of this stream. The new stream
		* can be used to store additional data in parallel with this stream. It depends on the stream's 
		* implementation to provide right behaviour. For example a file stream which
		* generates a sibling stream might make sure that the new stream is stored under the same path.
		* A database stream might for example provide sibling streams to be stored on the same machine.
		* If NULL is given back, then sibling streams are not supported.
		* @param uid Unique identifier used to identify siblings stream among all created. This is useful when opening a sibling stream for input
		*        and required to open the same again. It is in responsibility of the stream to make sure that sibling corresponding to correct UID are opened (when for input).
		* @param raw if stream is raw, then no stream initialization is performed, hence no class information is extracted. Raw streams are pure dump of data.
		**/
		virtual InputStream* createSiblingStream(card64 uid, bool raw = false) { forget(uid); forget(raw); return NULL; }

		/**
		* Indicate if this stream can perform seek operations.
		**/
		virtual bool canSeek() const { return false; }

		/**
		* Seek in the stream. If stream does not support seek operation, then an error is thrown.
		* @param rel Relative seek position, can be negative to seek backward.
		* @return Return the current position within the stream after seek completed
		**/
		virtual card64 seek(int64 rel) { forget(rel); throw EIOException("InputStream::seek() - cannot seek!");  return 0; }

		/**
		* Read a buffer of bytes. to be overridden in descendant classes.
		**/
		virtual void readBuffer(void* buf, mpcard length) = 0;

		/**
		* Read any block of data (not containing pointers or objects that need special construction!)
		**/
		template <class ElementaryType>
		inline void read(ElementaryType &v)
		{
			readBuffer(static_cast<void*>(&v), sizeof(ElementaryType));
		};

		/// read a boolean value
		inline void readBool(bool &v)
		{
			card8 val = 0;
			readBuffer(&val, 1);
			v = val == 1;
		}

		/// read an stl std::string
		inline void readString(std::string &v)
		{
			card64 length = 0;
			readBuffer(&length, sizeof(length));

			if (length > 0)
			{
				std::string::value_type *buf = new std::string::value_type[length+1];
				readBuffer(buf, length);
				buf[length] = '\0';

				v = std::string(buf, length / sizeof(std::string::value_type) );
				delete [] buf;
			}
		}

		/// read a cardinal value with maximum bitlength of the current platform;
		/// in the stream, a 64-bit representation is always used
		inline void readMaxPlatformCard(max_platform_card &v)
		{
			card64 vlarge;
			read<card64>(vlarge);
			#ifdef X4_32BIT
				if (vlarge > card64(0xFFFFFFFF))
					throw EIOException(__FUNCTION__ "() - file contains a number that cannot be represented on a 32bit platform.");
			#endif
			v = static_cast<max_platform_card>(vlarge);		
		}

		/// as readMaxPlatformCard, but for signed values
		inline void readMaxPlatformInt(max_platform_int &v)
		{
			int64 vlarge;
			read<int64>(vlarge);
			#ifdef X4_32BIT
				if (vlarge >= int64(0x7FFFFFFF) || vlarge < -int64(0x7FFFFFFF))
					throw EIOException(__FUNCTION__ "() - file contains a number that cannot be represented on a 32bit platform.");
			#endif
			v = static_cast<max_platform_int>(vlarge);
		}
	};

}


#endif
