//---------------------------------------------------------------------------
#ifndef MemoryManagerH
#define MemoryManagerH
//---------------------------------------------------------------------------
#include "X4GlobalDefinitions.h"
#include "X4Types.h"
#include "x4core.h"
//---------------------------------------------------------------------------
#include <list>
#include <QtCore/QMutex>
//---------------------------------------------------------------------------

namespace X4 {

class AttachedData;
class MemoryManagedObject;


class X4CORE_API MemoryManager
{
private:
   friend MemoryManagedObject;

   virtual void add( MemoryManagedObject * obj );
   virtual void remove( MemoryManagedObject * obj );
   virtual void touched( MemoryManagedObject * obj );
   virtual void resized( MemoryManagedObject * obj, card32 oldSize, card32 newSize );

   std::list<MemoryManagedObject*>  m_Objects;
   card32               m_MaxMemory;
   card32               m_UsedMemory;

public:


   MemoryManager( card32 maxMemory );
   virtual ~MemoryManager();

   void setMaxMemory( card32 maxMemory );
   inline card32 getMaxMemory() const {return m_MaxMemory;}
   inline card32 getUsedMemory() const {return m_UsedMemory;}

   virtual MemoryManagedObject* createNewMMObject(const AttachedData*);

   virtual void free( card32 maxUsedMemory );

	const static size_t PAD_SPACE = 1 << 26;

	/**
	* Reallocate memory in order to fit new data. 
	* The method returns NULL if allocation was not successfull.
	*
	* NOTE: the method is thread-safe. Current implementation locks threads while reallocating the memory
	**/
	static byte* ds_realloc(byte* arr, size_t elemSize, size_t usedElements, size_t& allocatedElements, size_t neededElements);

	//! Free used memory. The given pointer will be cleared and set to NULL
	static void ds_free(byte*& arr);

	//! Allocate memory to be used in the derived classes to manage the data
	static byte* ds_alloc(size_t numBytes);
	static byte* ds_alloc(size_t elemSize, size_t numElements) {return ds_alloc(elemSize * numElements); }

#if 0
	/**
	* Setup memory growing parameters for the reallocation procedure. In general the amount of used elements is smaller then the amount
	* of allocated elements in the memory, if using ds_realloc method. However, as soon as the number of used elements became equal or 
	* larger then the number of allocated elements, then the underlying array has to grow. The growing is achieved by multiplying hte amount
	* of allocated elements by a factor. In a STL std::vector implementation this factor is two, which implies that the std::vector is allocated
	* double as much elements as required when the number of used elements became equal to the number of allocated elements.
	* The benefit of this method is an asymptotic insertion complexity of O(1). However in practice, this strategy might be bad, since
	* if you have already 4GB of RAM in use and you just need to add couple of additional elements, you could end up in an array of 8GB, which
	* might not fit well anymore in the RAM. On the other side, if you store just 1MB of memory doubling this is good, however maybe even multply it by 8 is better :)
	* Hence we propose some kind of adaptive growing factor for the memory allocation here. The growing factor is decided linear between the
	* minimal and maximal amount of memory specified by the user. 
	* @param minMem Minimal amount of memory under which the factor is not changed anymore.
	* @param minFactor Factor for the minimal memory growing. 
	* @param maxMem Maximal amount of memory for which the factor does not change anymore
	* @param maxFactor Maximal factor to be used.
	*
	* @return True if parameters were valid, false otherwise
	*
	* The default parameters emulate the STL std::vector behavior, hence always multiplying by 2.
	*
	* Example: ds_setupFactors(100, 10.0f, 1<<30, 1.1f) results in following behaviour:
	*	when amount of used memory grows somewhere between 0 and 100 bytes, the allocated memory will be multiplied by factor 10. Hence pushing 1 byte in the array
	*	results in 10 bytes allocated. Putting the 10th byte, will end up already in 100 allocated bytes. Putting the 100th byte will increase the size of allocated
	*   memory to 1000 bytes, afterwards the increase factor drops down, e.g. by 1000th byte increase will be somewhere to 9000 bytes.
	**/
	static bool ds_setupFactors(size_t minMem = 0, float minFactor = 2.0f, size_t maxMem = 1, float maxFactor = 2.0f);
#endif

	static QMutex m_mutex;

protected:

#if 0
private:
	static std::pair<size_t,float> m_minMemoryGrowingParam;
	static std::pair<size_t,float> m_maxMemoryGrowingParam;

	static std::pair<double,double> m_MemoryGrowingParam;
#endif

};

class MMOOwner;

class X4CORE_API MemoryManagedObject
{
   friend MemoryManager;

private:
   MemoryManager *   m_MyManager;
   std::list<MemoryManagedObject*>::iterator  m_MyIterator;
   card32            m_MySize;
   MMOOwner       *  m_MyOwner;

public:

   MemoryManagedObject( card32 size, MemoryManager * manager);
   virtual ~MemoryManagedObject();

   void setOwner( MMOOwner * owner ) {m_MyOwner = owner;}
   MMOOwner * getOwner() const {return m_MyOwner;}

   void touched();
   inline card32 getSize() const {return m_MySize;};
   void setSize( card32 size);

};

/// interfaceclass
class X4CORE_API MMOOwner
{
public:
   virtual void onDeleteMMO(MemoryManagedObject*) = 0;
};


/*class X4CORE_API MemoryManagedParameter : public MemoryManagedObject
{
private:
   AttachedData   *    m_AttachedData;
public:

   MemoryManagedParameter( card32 size, MemoryManager * manager, AttachedData * data );
   virtual ~MemoryManagedParameter();

};*/



}
#endif
