//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "MetaClass.h"
//---------------------------------------------------------------------------
#include "ClassProperty.h"
#include "ClassMethods.h"
#include "ClassCreator.h"
#include "PropertyTable.h"
#include <HashTables.inline.h>
#include "OutputStream.h"
#include "InputStream.h"
#include <ConsoleOutput.h>
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

// --- MetaClass

MetaClass::MetaClass_ClassRegHashTable MetaClass::classRegistration;

mpcard MetaClass::MetaClass_ClassRegHashTable::hashFunction(std::string className) const {
	card32 hashValue = 0;
	for (unsigned i = 0; i != className.size(); ++i)
	  hashValue = (hashValue * 37) + className[i];
	return hashValue;
}

MetaClass::MetaClass(std::string className, std::string classDescription, InstanceCreator newInstanceMethod,
					 card32 instanceSize) {
	this->parentClass = NULL;
	this->classVersion = 0;
	this->known = true;
	this->className = className;
	this->classDescription = classDescription;
	this->newInstanceMethod = newInstanceMethod;
	this->instanceSize = instanceSize;
	this->propertyTable = new PropertyTable();
	guiDescr = new GUIDescriptor();
}

/// Obsolete init method, use init(parentClass, classVersion) instead
void MetaClass::init(const MetaClass *parentClass, card32 classVersion, bool known)
{
	this->parentClass = (MetaClass*)parentClass;
	this->classVersion = classVersion;
	this->known = known;
	x4Assert(classVersion >= FIRST_VERSION);

	if (parentClass != NULL)
		((MetaClass*)parentClass)->descendants.push_back(this);

	MetaClass *forgetIt;
	if (!classRegistration.getData(className, forgetIt))
		classRegistration.addElement(className, this);
	else
		throw EInvalidState(std::string("MetaClass::init(): Reregistration of a class with the same class name (Class: ")+className+std::string(")"));
}

void MetaClass::copyParentProperties()
{
	if (parentClass != NULL)
	{
		// add properties from the parent meta class to this one
		std::vector<ClassProperty*>::iterator pos = parentClass->properties.begin();
		std::vector<ClassProperty*>::iterator end = parentClass->properties.end();
		while (pos != end) {
			properties.push_back((*pos)->copyClassProperty());
			++pos;
		}

		// add class methods from the parent class
		std::vector<ClassMethod*>::iterator mpos = parentClass->methods.begin();
		std::vector<ClassMethod*>::iterator mend = parentClass->methods.end();
		while (mpos != mend) {
			addMethod((*mpos)->copyMethod());
			++mpos;
		}

		// merge property table of the parent class with this one
		propertyTable->merge(parentClass->propertyTable);
	}

	GUIDescriptor *parentGuiD = parentClass->getGUIDescriptor();
	GUIDescriptor *thisGuiD = getGUIDescriptor();

	card32 ncc = parentGuiD->getNumClassCreators();
	for (card32 i = 0; i < ncc; ++i) {
		thisGuiD->addClassCreator(parentGuiD->getClassCreator(i));
	}

	card32 nce = parentGuiD->getNumClassEditors();
	for (card32 i = 0; i < nce; ++i) {
		const MetaClass *pce = parentGuiD->getClassEditor(i);
		thisGuiD->addClassEditor(pce, pce == parentGuiD->getDefaultClassEditor(), parentGuiD->getEditorParameters(i));
	}
}

MetaClass::~MetaClass()
{
	delete propertyTable;

	{
	  std::vector<ClassProperty*>::iterator pos = properties.begin();
	  std::vector<ClassProperty*>::iterator end = properties.end();
	  while (pos != end) {
		 delete *pos;
		 ++pos;
	  }
	}
	{
	  std::vector<ClassMethod*>::iterator pos = methods.begin();
	  std::vector<ClassMethod*>::iterator end = methods.end();
	  while (pos != end) {
		 delete *pos;
		 ++pos;
	  }
	}

	delete guiDescr;
/*
	#if defined(_DEBUG)
	MetaClass *forgetIt;
	if (!classRegistration.getData(className, forgetIt)) {
	  std::string msg = std::string("MetaClass::~MetaClass() - class is not registered, className: ")+className;
	  Application->MessageBox(msg.c_str(), "Warning:",
							  MB_OK | MB_ICONEXCLAMATION);
	}
	#endif*/
	classRegistration.deleteElement(className);
}

void MetaClass::checkAllShutdown() {
  /* #if defined(_DEBUG)
	if (classRegistration.count() > 0) {
	  AnsiString badGuys;
	  max_platform_card n = classRegistration.count();
	  for (max_platform_card i=0; i<n; i++) {
		 badGuys += classRegistration.getKey(i).c_str();
		 if (i != n) {
			badGuys += " ";
		 }
	  }
	  AnsiString msg = "Not all MetaClasses have been shutdown properly. Missing: " + badGuys;
	  Application->MessageBox(msg.c_str(), "Warning:", MB_ICONEXCLAMATION | MB_OK);
	}
	#endif
	*/
}

Object* MetaClass::newInstance() const {
	if (newInstanceMethod == NULL) {
	  throw EClassAbstract("MetaClass::newInstance() - cannot create instance, class is abstract.");
	} else {
	  Object* obj = newInstanceMethod();
	  obj->__instanceClass = this;
	  return obj;
	}
}

const std::string MetaClass::getClassName() const {
	return className;
}

const std::string MetaClass::getClassDescription() const {
	return classDescription;
}

card32 MetaClass::getClassVersion() const {
  return this->classVersion;
}

bool MetaClass::isAbstract() const {
	return newInstanceMethod == NULL;
}

card32 MetaClass::getInstanceSize() const {
	return instanceSize;
}

bool MetaClass::inheritsFrom(const MetaClass* mc) const {
	if (mc == NULL) return false;
	const MetaClass *currentClass = this;
	do {
	  if (currentClass == mc) {
		 return true;
	  } else {
		 currentClass = currentClass->getParentClass();
	  }
	} while (currentClass != NULL);
	return false;
}

void MetaClass::addProperty(ClassProperty *prop) {
	x4Assert(prop->getIntroducedInVersion() <= classVersion);
	properties.push_back(prop);
	prop->setOwner(this);

	// check if this property has been registered before
	if (!isClassRegistered(prop->getInstanceClass()->getClassName()))
	{
		consoleError(std::string("MetaClass::addProperty() - property of type '") + prop->getInstanceClass()->getClassName() + std::string("' is not registered. Cannot add this property to ") + getClassName());
	}

	propertyTable->add(prop->getIntroducedInVersion(), prop->getTypeName(), prop->getCName());
}

card32 MetaClass::getNumProperties() const {
	return (card32)properties.size();
}

const ClassProperty *MetaClass::getProperty(card32 i) const {
	return properties[i];
}

const ClassProperty *MetaClass::getProperty(std::string name) const {
	for (size_t i=0; i<properties.size(); i++) {
	  if (properties[i]->getCName() == name) return properties[i];
	}
	return NULL;
}

card32 MetaClass::getNumMethods() const {
	return (card32)methods.size();
}

void MetaClass::addMethod(ClassMethod *m) {
	methods.push_back(m);
}

const ClassMethod *MetaClass::getMethod(card32 i) const {
	return methods[i];
}

const ClassMethod *MetaClass::getMethod(std::string name) const {
	for (size_t i=0; i<methods.size(); i++) {
	  if (methods[i]->getName() == name) return methods[i];
	}
	return NULL;
}

const MetaClass* MetaClass::getClassByName(const std::string className) {
	MetaClass* metaClass;
	bool success = classRegistration.getData(className, metaClass);

	if (success)
	  return metaClass;
	else
	  return NULL;
	  //throw ERangeCheck("MetaClass::getClassByName - invalid class name");
}

bool MetaClass::isClassRegistered(const std::string className) {
	MetaClass* forgetIt;
	return classRegistration.getData(className, forgetIt);
}


bool MetaClass::callNoArgsMethod(Object *obj, std::string methodName) const {
	mpcard n = methods.size();
	for (mpcard i=0; i<n; i++) {
		if (methods[i]->getName() == methodName) {
			NoArgsClassMethod *m = dynamic_cast<NoArgsClassMethod*>(methods[i]);
			if (m != NULL) {
				m->call(obj);
				return true;
			}
		}
	}
	return false;
}


void MetaClass::actProperty_setUserEdit( bool b )
{
	x4Assert( properties.size() > 0 );
	properties.back()->setUserEdit( b );
}

void MetaClass::actProperty_setEditReadonly( bool b )
{
	x4Assert( properties.size() > 0 );
	properties.back()->setEditReadonly( b );
}

void MetaClass::actProperty_setStreamProperty( bool b )
{
	x4Assert( properties.size() > 0 );
	properties.back()->setStreamProperty( b );
}

void MetaClass::actProperty_setPropertyName( std::string name )
{
	x4Assert( properties.size() > 0 );
	properties.back()->setCName( name );
}

void MetaClass::actProperty_addPropertyEditor( const MetaClass *mc, bool defaultEd)
{
	x4Assert( properties.size() > 0 );
	properties.back()->addPropertyEditor(mc, defaultEd);
}

void MetaClass::write(OutputStream* out) const
{
}

void MetaClass::read(InputStream* in)
{
}

void MetaClass::addClassEditorToSubtree( const MetaClass *classEditor, bool defaultEditor, GUIDescriptor::EditorParameters editorParameters )
{
	getGUIDescriptor()->addClassEditor(classEditor, defaultEditor, editorParameters);
	for (size_t i=0; i<descendants.size(); i++) {
		descendants[i]->addClassEditorToSubtree(classEditor, defaultEditor, editorParameters);
	}
}


// --- GUIDescriptor


GUIDescriptor::EditorParameters::EditorParameters()
{
	useAsMainEditor = true;
	useAsSideEditor = false;
}

GUIDescriptor::GUIDescriptor() {
	autoApplyEditChanges = true;
   icon = ":/foundation/images/Classes/Object.png";
}

bool GUIDescriptor::getAutoApplyEditChanges() const {
	return autoApplyEditChanges;
}

void GUIDescriptor::setAutoApplyEditChanges(bool aap) {
	this->autoApplyEditChanges = aap;
}

void GUIDescriptor::addClassEditor(const MetaClass *classEditor, bool defaultEditor, EditorParameters params) {
	if (defaultEditor) {
	  classEditors.push_front(classEditor);
	  editorParameters.push_front(params);
	} else {
	  classEditors.push_back(classEditor);
	  editorParameters.push_back(params);
	}
}

void GUIDescriptor::addEditorTool(const MetaClass *editorTool) {
	editorTools.push_back(editorTool);
}

void GUIDescriptor::addClassCreator(const MetaClass *classCreator) {
	classCreators.push_back(classCreator);
}

void GUIDescriptor::addApplication(ApplicationPtr app) {
	applications.push_back(app);
}


const MetaClass *GUIDescriptor::getDefaultClassCreator() const {
	if (classCreators.empty()) {
	  return NULL;
	} else {
	  return classCreators[0];
	}
}

const MetaClass *GUIDescriptor::getDefaultClassEditor() const {
	if (classEditors.empty()) {
	  return NULL;
	} else {
	  return classEditors[0];
	}
}

card32 GUIDescriptor::getNumClassCreators() const {
	return (card32)classCreators.size();
}

card32 GUIDescriptor::getNumClassEditors() const {
	return (card32)classEditors.size();
}

card32 GUIDescriptor::getNumEditorTools() const {
	return (card32)editorTools.size();
}

card32 GUIDescriptor::getNumApplications() const {
	return (card32)applications.size();
}


const MetaClass *GUIDescriptor::getClassCreator(card32 i) const {
	return classCreators[i];
}

const MetaClass *GUIDescriptor::getClassEditor(card32 i) const {
	return classEditors[i];
}
const MetaClass *GUIDescriptor::getEditorTool(card32 i) const {
	return editorTools[i];
}

GUIDescriptor::EditorParameters GUIDescriptor::getEditorParameters(card32 i) const {
	return editorParameters[i];
}

GUIDescriptor::ApplicationPtr GUIDescriptor::getApplication(card32 i) const {
	return applications[i];
}

const char* GUIDescriptor::getIcon() const {
	return icon.c_str();
}

void GUIDescriptor::setIcon(const char* resourceString) {
	this->icon = resourceString;
}


