//---------------------------------------------------------------------------
#ifndef MetaClassH
#define MetaClassH
//---------------------------------------------------------------------------
#include "x4core.h"
//---------------------------------------------------------------------------
#include <string>
#include <list>
#include <vector>
#include <deque>
#include <X4Exception.h>
#include <X4Types.h>
#include <HashTables.h>
//---------------------------------------------------------------------------


namespace X4 {
   

	class X4CORE_API Object;
	class X4CORE_API ClassProperty;
	class X4CORE_API ClassMethod;
	class X4CORE_API MetaClass;
	class X4CORE_API PropertyTable;

	class X4CORE_API OutputStream;
	class X4CORE_API InputStream;

	class X4CORE_API GUIDescriptor
	{
	public:
		struct X4CORE_API EditorParameters {
			bool useAsMainEditor;
			bool useAsSideEditor;
			EditorParameters();
		};

		/// pointer to a procedure that starts a new application that works with the given object
		class X4CORE_API ApplicationPtr
		{
		public:
			typedef void (*ApplicationStartProc)(Object *object);
			ApplicationPtr(ApplicationStartProc startProc, const char* name) : startProc(startProc), name(name) {}
			void operator()(Object* obj) { startProc(obj); }
		private:
			ApplicationStartProc startProc;
			std::string name;
		};

	private:
		bool autoApplyEditChanges;
		std::vector<const MetaClass*> classCreators; /* class of ClassCreator */
		std::deque<const MetaClass*> classEditors; /* class of ClassEditor */
		std::deque<const MetaClass*> editorTools; /* class of EditorTool */
		std::deque<EditorParameters> editorParameters;
		std::vector<ApplicationPtr> applications;
		/// resource identifier for icon
		std::string icon; 

	public:
		GUIDescriptor();
		bool getAutoApplyEditChanges() const; // default = true
		void setAutoApplyEditChanges(bool aap);
		void addClassEditor(const MetaClass *classEditor, bool defaultEditor = false, EditorParameters params = EditorParameters());
		void addEditorTool(const MetaClass *classEditorTool);
		void addClassCreator(const MetaClass *classCreator);
		void addApplication(ApplicationPtr app);
		const MetaClass *getDefaultClassCreator() const; // may return NULL
		const MetaClass *getDefaultClassEditor() const; // default = property browser
		card32 getNumClassCreators() const;
		card32 getNumClassEditors() const;
		card32 getNumEditorTools() const;
		card32 getNumApplications() const;
		const MetaClass *getClassCreator(card32 i) const;
		const MetaClass *getClassEditor(card32 i) const;
		const MetaClass *getEditorTool(card32 i) const;
		EditorParameters getEditorParameters(card32 i) const; 
		ApplicationPtr getApplication(card32 i) const;
		const char* getIcon() const;
		void setIcon(const char* resourceString);
	};



	class X4CORE_API MetaClass
	{
	private:

		class X4CORE_API MetaClass_ClassRegHashComparator {
			public:
			static bool equals(std::string obj1, std::string obj2) { return obj1 == obj2; }
		};

		class X4CORE_API MetaClass_ClassRegHashTable : public HashTable<std::string, MetaClass*, MetaClass_ClassRegHashComparator> {
		public:
			virtual max_platform_card hashFunction(std::string className) const;
		};

		typedef Object* (*InstanceCreator)();

		static MetaClass_ClassRegHashTable classRegistration;

		card32 instanceSize;
		std::string className;
		std::string classDescription;
		card32 classVersion;
		InstanceCreator newInstanceMethod;
		MetaClass *parentClass;
		GUIDescriptor *guiDescr;
		bool known;

		std::vector<ClassProperty*> properties;
		std::vector<ClassMethod*> methods;
		std::vector<MetaClass*> descendants;
		PropertyTable* propertyTable;  // a property table associated with the Metaclass
		
	public:
		class EClassAbstract : public X4Exception {
		public:
			EClassAbstract(std::string msg) : X4Exception(msg) {}
			virtual std::string getTypeString() {return "EClassAbstract";}
		};

		typedef card32 ClassVersion;

		static const ClassVersion FIRST_VERSION = 0;

		/** create new meta class. newInstanceMethod == NULL for abstract classes. */
		MetaClass(std::string className, std::string classDescription, InstanceCreator newInstanceMethod, card32 instanceSize);
		~MetaClass();

		/// serialize metaclass to the output
		void write(OutputStream* out) const;
		void read(InputStream* in);

		/// initialize meta class.
		/// parentClass - parent class, NULL for root class ("Object")
		/// version -  per class version number, set to FIRST_VERSION + number
		void         init(const MetaClass *parentClass, card32 classVersion = FIRST_VERSION, bool known = true);

		//
		virtual void copyParentProperties();
		card32       getNumProperties() const;
		void         addProperty(ClassProperty *prop);
		const ClassProperty *getProperty(card32 i) const;
		const ClassProperty *getProperty(std::string name) const;
		card32       getNumMethods() const;
		void         addMethod(ClassMethod *m);
		const ClassMethod *getMethod(card32 i) const;
		const ClassMethod *getMethod(std::string name) const;
		Object*      newInstance() const; // throws EClassAbstract
		bool         isAbstract() const; // returns newInstanceMethod == NULL
		bool         isKnown() const { return known; } // check whenever this class was known during hte compilation phase
		card32       getInstanceSize() const;
		bool         inheritsFrom(const MetaClass* mc) const;
		const std::string getClassName() const;
		const std::string getClassDescription() const;
		card32       getClassVersion() const;
		const PropertyTable* getPropertyTable() const { return propertyTable; }

		// interface for the last class property
		void actProperty_setUserEdit( bool b );
		void actProperty_setEditReadonly( bool b );
		void actProperty_setStreamProperty( bool b );
		void actProperty_setPropertyName( std::string name );
		void actProperty_addPropertyEditor( const MetaClass *mc, bool defaultEd = true );
   
		// interface for the last class method
		ClassMethod *getActMethod() {return methods.back();}
   
		card32 getNumDescendants() const {return (card32)descendants.size();}
		const MetaClass *getDescendant(card32 i) const {return descendants[i];}
		const MetaClass *getParentClass() const {return parentClass;}

		const GUIDescriptor *getGUIDescriptor() const {return guiDescr;}
		      GUIDescriptor *getGUIDescriptor() {return guiDescr;}

		inline bool  operator==(MetaClass *c) const {return this == c;}

		/// calls a class method that does not require arguments
		/// returns true if the method was found and called successfully, false otherwise
		bool callNoArgsMethod(Object *obj, std::string methodName) const;

		/// returns NULL if no matching class is found 
		static const MetaClass* getClassByName(const std::string className);
		/// returns true if class is present in global class registration table (useful for debugging)
		static bool isClassRegistered(const std::string className);

		void addClassEditorToSubtree(const MetaClass *classEditor, bool defaultEditor = false, 
											  GUIDescriptor::EditorParameters editorParameters = GUIDescriptor::EditorParameters());


		/// call at the end of the application to check whether all classes have been shutdown (check is ommited if not in debug mode, i.e. _DEBUG defined)
		static void checkAllShutdown();
	};

}


#endif
