//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "Object.h"
//---------------------------------------------------------------------------
#include <Tools.h>
#include <X4Assert.h>
#include "UnknownObject.h"
#include "CopyObjectProperties.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------


MetaClass* Object::objectMetaClass = NULL;

const MetaClass* Object::getInstanceClass() const {
	if (__instanceClass == NULL) return getClass();
	return __instanceClass;
}

const MetaClass* Object::getClass() {
   if (objectMetaClass==NULL) {
      objectMetaClass = new MetaClass("Object", "Base Class for all reflective classes.", NULL, sizeof(Object));
   }
   return objectMetaClass;
}

void Object::init(const MetaClass *parentClass) {
   ((MetaClass*)getClass())->init(parentClass);
}

void Object::shutdown() {
   delete objectMetaClass;
}

Object::Object() : __instanceClass(NULL), __unknownPropertyContainer(NULL)
{
}

Object::~Object()
{
	if (__unknownPropertyContainer) delete __unknownPropertyContainer;
	__unknownPropertyContainer = NULL;
}

UnknownPropertyContainer* Object::getOrCreateUnknownPropertyContainer()
{
	if (__unknownPropertyContainer) return __unknownPropertyContainer;
	__unknownPropertyContainer = new UnknownPropertyContainer;
	return __unknownPropertyContainer;
}

const UnknownPropertyContainer* Object::getOrCreateUnknownPropertyContainer() const
{
	if (__unknownPropertyContainer) return __unknownPropertyContainer;
	__unknownPropertyContainer = new UnknownPropertyContainer;
	return __unknownPropertyContainer;
}

// ----

Object *X4CopyContext::getData(std::string name) {
   const size_t n = tagStack.size();
   if (n == 0) return NULL;
   for (size_t i=n-1; i>0; i--) {
      if (tagStack[i].name == name) return tagStack[i].data;
   }
   if (tagStack[0].name == name) return tagStack[0].data;
   return NULL;
}

void X4CopyContext::pushData(std::string name, Object *obj) {
   Entry e;
   e.name = name;
   e.data = obj;
   tagStack.push_back(e);
}

void X4CopyContext::popData() {
   x4Assert(tagStack.size() > 0);
   tagStack.pop_back();
}



void Object::assign(const Object* obj, X4CopyContext* context)
{
	if (obj->getInstanceClass() != this->getInstanceClass())
	{
		throw ECannotAssign("Object::assign does not implement assignment between different types.");
		return;
	}

	if (context && context->isShallowCopy())
		memcpy(this, obj, getInstanceClass()->getInstanceSize());
	else
		copyObjectProperties(obj, this);
}

Object* Object::copy(X4CopyContext *context) const
{
	if (context == NULL) return copyDeep();
	return copyShallow();
}

Object* Object::copyDeep(X4CopyContext* context) const
{
	Object* result = getInstanceClass()->newInstance();   
	X4CopyContext* ctx = context;
	if (ctx == NULL)
	{
		ctx = new X4CopyContext;
		ctx->setCopyType(X4CopyContext::DEEP_COPY);
	}
	result->assign(this, ctx);

	if (ctx != context) delete ctx;

	return result;
}

Object* Object::copyShallow(X4CopyContext* context) const
{
	Object* result = getInstanceClass()->newInstance();   
	X4CopyContext* ctx = context;
	if (ctx == NULL)
	{
		ctx = new X4CopyContext;
		ctx->setCopyType(X4CopyContext::SHALLOW_COPY);
	}
	result->assign(this, ctx);

	if (ctx != context) delete ctx;

	return result;
}

bool Object::equal(const Object* obj) const {
   if (obj == NULL) return false;

   const MetaClass *thisMetaClass = getInstanceClass();
   const MetaClass *objMetaClass = obj->getInstanceClass();
   if (thisMetaClass != objMetaClass) return false;

   card32 instanceSize = thisMetaClass->getInstanceSize();
   char *thisPos = (char*)this;
   char *objPos = (char*)obj;

   while (instanceSize > 0) {
      if (*thisPos != *objPos) return false;
      ++thisPos;
      ++objPos;
      --instanceSize;
   }
   return true;
}

card32 Object::getFieldsHashValue() const{
   const MetaClass *thisMetaClass = getInstanceClass();

   card32 hashValue = 0;
   card32 DWordInstanceSize = thisMetaClass->getInstanceSize() / 4;
   card32 *thisPos = (card32*)this;

   while (DWordInstanceSize > 0) {
      hashValue = (hashValue << 1) ^ (*thisPos);
      ++thisPos;
      --DWordInstanceSize;
   }
   return hashValue;
}



std::string X4::makeObjectDescription(const Object *obj, bool shortDescr) {
   if (obj == NULL) return "NULL";
   std::string objDescr;
   objDescr = obj->getInstanceClass()->getClassName().c_str();
	char buffer[2048];
   if (!shortDescr && obj != NULL ) 
	{
		sprintf_s( buffer, 2048, "@%p (%d Bytes)", (card32)obj,(card32)obj->getInstanceClass()->getInstanceSize() );
   } else {
		sprintf_s( buffer, 2048, "@%p", (card32)obj );
   }
	objDescr += buffer;
   return objDescr;

}
