//---------------------------------------------------------------------------
#ifndef ObjectH
#define ObjectH
//---------------------------------------------------------------------------
#include "x4core.h"
//---------------------------------------------------------------------------
#include "MetaClass.h"
#include <X4Exception.h>
#include <boost/smart_ptr.hpp>
#include <Tools.h>
//---------------------------------------------------------------------------
namespace X4
{
	class UnknownPropertyContainer;

	/// provides additional context information for copying
	class X4CORE_API X4CopyContext
	{		
		public:
			enum CopyType
			{
				SHALLOW_COPY,  // copy is performed in lightweight way, e.g. memcpy
				DEEP_COPY      // expect to perform deep copy, e.g. copyObjectProperties (default)
			};

			X4CopyContext(const CopyType& type) : copyType(type) {}
			X4CopyContext() : copyType(DEEP_COPY) {}
			Object *getData(std::string name);

			void pushData(std::string name, Object *obj);
			void popData();

			inline void setCopyType(CopyType type) { copyType = type; }

			inline bool isShallowCopy() const {return copyType == SHALLOW_COPY;}
			inline bool isDeepCopy() const {return copyType == DEEP_COPY;}

			operator X4CopyContext*() { return this; }

		private:
			struct Entry
			{
				std::string name;
				Object *data;
			};
			std::vector<Entry> tagStack;
			CopyType copyType;
	};

	class ECannotAssign : public ETypeMismatch
	{
		public:
			ECannotAssign(const std::string msg) : ETypeMismatch(msg) {};
	};

	class X4CORE_API Object
	{
		private:
			static MetaClass* objectMetaClass;
		protected:
			friend class X4::MetaClass;
			const MetaClass* __instanceClass;
			mutable UnknownPropertyContainer* __unknownPropertyContainer;

		public:
			virtual const MetaClass *getInstanceClass() const;
			static const MetaClass *getClass();
			static void init(const MetaClass *parentClass);
			static void shutdown();

			const UnknownPropertyContainer* getOrCreateUnknownPropertyContainer() const;
			UnknownPropertyContainer* getOrCreateUnknownPropertyContainer();

			/**
			* Depending on the context either perform simple copy using memcpy, or reassign each of the properties.
			* If no context is specified, then deep copy is performed.
			**/
			virtual void assign(const Object* obj, X4CopyContext *context = NULL);

			virtual bool equal(const Object* obj) const;
			virtual card32 getFieldsHashValue() const;

			//! Perform copy of the object. if no context is specified, then copyDeep() is executed
			virtual Object* copy(X4CopyContext *context = NULL) const;

			/**
			* A deep copy means that all object properties will be reassigned to the new object.
			* Copying properties is slower but prevents memory corruption for embedded object (non-pointer as well as non-POD member variables).
			* Use copyShallow() to perform lightweight copy, i.e. using memcpy.
			* The method uses assign() to perform object copy.
			*
			* The specified context can be used to pass further context to the assign() function. 
			**/
			Object* copyDeep(X4CopyContext* context = NULL) const;

			//! Copies an object by copying its memory content with memcpy (WARNING: dangerous for embedded member variables). Very fast.
			Object* copyShallow(X4CopyContext* context = NULL) const;

			Object();
			virtual ~Object();
	};


	/**
		Declares an abstract class - example:
		class MyClass : public Persistent
		{
			X4_ABSTRACT_CLASS( MyClass )

		public:
			...
		};
	*/
	#define X4_ABSTRACT_CLASS( ... ) \
	private: \
		static X4::MetaClass* __MetaClass;\
	public:\
		virtual const X4::MetaClass *getInstanceClass() const;\
		static const X4::MetaClass *getClass();\
		static void init(const X4::MetaClass *parentClass);\
		static void initMetaClass();\
		static void shutdown();


	/**
		Declares a class - example:
		class MyClass : public Persistent
		{
			X4_CLASS( MyClass )

		public:
			...
		};
	*/
	#define X4_CLASS( ... ) \
	private:\
		static X4::Object *newInstance();\
		X4_ABSTRACT_CLASS( __VA_ARGS__ )

	#define X_CLASS X4_CLASS();
	#define X_ABSTRACT_CLASS X4_ABSTRACT_CLASS();


	#define BEGIN_CLASS_INIT( ... ) \
        __VA_ARGS__ * obj_null_pointer = (__VA_ARGS__*)nullptr; \
        pointer_diff obj_null_offset = (pointer_diff)nullptr;

	#define SET_USER_EDIT( b ) ((X4::MetaClass *)getClass())->actProperty_setUserEdit(b);
	#define SET_EDIT_READONLY( b ) ((X4::MetaClass *)getClass())->actProperty_setEditReadonly(b);
	#define SET_STREAM_PROPERTY( b ) ((X4::MetaClass *)getClass())->actProperty_setStreamProperty(b);
	#define SET_NAME( name ) ((X4::MetaClass *)getClass())->actProperty_setPropertyName(name);
	#define SET_EDITOR( name ) ((X4::MetaClass *)getClass())->actProperty_addPropertyEditor(name, true);

	#define X4_MEMBER( mType, mName ) \
	private: \
		mType m_##mName ; \
	public: \
		void set##mName##( mType v ) { m_##mName = v; } \
		mType get##mName##() const {return m_##mName ;}


	std::string X4CORE_API makeObjectDescription(const Object *obj, bool shortDescr = false);



/** ------------------------ New Type of class definitions ------------------ **/
#define __X4_CLASS_IMPL__(version, description, tmpl, newInstance, ...) \
	tmpl X4::MetaClass* __VA_ARGS__##::__MetaClass = NULL;\
	tmpl const X4::MetaClass* __VA_ARGS__##::getInstanceClass() const \
	{\
		if (__instanceClass == NULL) return getClass(); \
		return __instanceClass; \
	}\
	tmpl const X4::MetaClass* __VA_ARGS__##::getClass() \
	{\
		if (__VA_ARGS__##::__MetaClass==NULL) {\
			__VA_ARGS__##::__MetaClass = new X4::MetaClass( #__VA_ARGS__, description, newInstance, sizeof( __VA_ARGS__ ));\
		}\
		return __VA_ARGS__##::__MetaClass;\
	}\
	tmpl void __VA_ARGS__##::shutdown() \
	{\
		delete __VA_ARGS__##::__MetaClass;\
	}\
	tmpl void __VA_ARGS__##::init(const X4::MetaClass *parentClass) \
	{\
		((X4::MetaClass*)getClass())->init(parentClass, X4::MetaClass::FIRST_VERSION + version );\
		__VA_ARGS__##::__MetaClass->copyParentProperties();\
		initMetaClass();\
	}\
	tmpl void __VA_ARGS__##::initMetaClass()

#define __X4_INST__(...) \
	X4::Object* __VA_ARGS__##::newInstance() { return new __VA_ARGS__; }
#define __X4_START_INIT__(...) \
	{\
		__VA_ARGS__ * obj_null_pointer = (__VA_ARGS__*)nullptr; \
        __VA_ARGS__ * _the_answer_to_everything = (__VA_ARGS__*)42; \
        pointer_diff obj_null_offset = (size_t)((char*)static_cast<Object*>(_the_answer_to_everything)) - 42;

#define BEGIN_TEMPLATE_CLASS(version, ...)\
	template<> __X4_INST__(__VA_ARGS__)\
	__X4_CLASS_IMPL__(version, "no description", template<>, newInstance, __VA_ARGS__)\
	__X4_START_INIT__(__VA_ARGS__)

#define BEGIN_TEMPLATE_CLASS_DESCR(version, description, ...)\
	template<> __X4_INST__(__VA_ARGS__)\
	__X4_CLASS_IMPL__(version, description, template<>, newInstance, __VA_ARGS__)\
	__X4_START_INIT__(__VA_ARGS__)

#define BEGIN_CLASS(version, ...)\
	__X4_INST__(__VA_ARGS__)\
	__X4_CLASS_IMPL__(version, "no description", , newInstance, __VA_ARGS__)\
	__X4_START_INIT__(__VA_ARGS__)

#define BEGIN_CLASS_DESCR(version, description, ...)\
	__X4_INST__(__VA_ARGS__)\
	__X4_CLASS_IMPL__(version, description, , newInstance, __VA_ARGS__)\
	__X4_START_INIT__(__VA_ARGS__)

#define BEGIN_ABSTRACT_TEMPLATE_CLASS(version, ...)	                   \
	__X4_CLASS_IMPL__(version, "no description", template<>, NULL, __VA_ARGS__)\
	__X4_START_INIT__(__VA_ARGS__)

#define BEGIN_ABSTRACT_TEMPLATE_CLASS_DESCR(version, description, ...) \
	__X4_CLASS_IMPL__(version, description, template<>, NULL, __VA_ARGS__)\
	__X4_START_INIT__(__VA_ARGS__)

#define BEGIN_ABSTRACT_CLASS(version, ...)                    \
	__X4_CLASS_IMPL__(version, "no description", , NULL, __VA_ARGS__)\
	__X4_START_INIT__(__VA_ARGS__)

#define BEGIN_ABSTRACT_CLASS_DESCR(version, description, ...) \
	__X4_CLASS_IMPL__(version, description, , NULL, __VA_ARGS__)\
	__X4_START_INIT__(__VA_ARGS__)

#define END_TEMPLATE_CLASS }
#define END_CLASS }


/**
	Implements a class - example (MyClass.cpp):
	
	IMPLEMENT_X4_CLASS( MyClass, 3 ) // version is MetaClass::FIRST_VERSION + 3 here !
	{
		// init properties and methods here
		BEGIN_CLASS_INIT( MyClass );

		MyClassMetaClass->addProperty...
	}

	MyClass::MyClass() {}
	...

*/

#define IMPLEMENT_X4_CLASS( className, version ) \
	__X4_INST__(className)\
	__X4_CLASS_IMPL__(version, "no description", , newInstance, className)

#define IMPLEMENT_X4_CLASS_DESCR( className, version, description ) \
	__X4_INST__(className)\
	__X4_CLASS_IMPL__(version, description, , newInstance, className)


/**
	Implements an abstract class - example (MyClass.cpp):
	
	IMPLEMENT_X4_ABTRACT_CLASS( MyClass, 3 ) // version is MetaClass::FIRST_VERSION + 3 here !
	{
		// init properties and methods here
		BEGIN_CLASS_INIT( MyClass );

		MyClassMetaClass->addProperty...
	}

	MyClass::MyClass() {}
	...

*/
#define IMPLEMENT_X4_ABSTRACT_CLASS( className, version ) \
	__X4_CLASS_IMPL__(version, "no description", , NULL, className)

#define IMPLEMENT_X4_ABSTRACT_CLASS_DESCR( className, version, description) \
	__X4_CLASS_IMPL__(version, description, , NULL, className)


}


#endif
