//---------------------------------------------------------------------------
#ifndef ObjectListPropertyH
#define ObjectListPropertyH
//---------------------------------------------------------------------------
#include "x4core.h"
//---------------------------------------------------------------------------
#include "ClassProperty.h"
#include "Tools.h"
#include "StringHelper.h"
//---------------------------------------------------------------------------


namespace X4 {

class OutputObjectStream;
class InputObjectStream;
class ObjectList_Prop;
class ObjectListElementEditor;


class X4CORE_API ObjectListProperty : public ClassProperty {
    X4_CLASS(ObjectListProperty);

 public:

   // property-access-methods
   typedef void (Object::* AddObjectMethod)(Object *obj);
   typedef void (Object::* DeleteObjectsMethod)(mpcard begin,mpcard end);
   typedef Object* (Object::* GetObjectMethod)(mpcard index);
   typedef void (Object::* MoveObjectMethod)(mpcard oldPos, mpcard newPos);
   typedef mpcard (Object::* GetNumObjectsMethod)();
   typedef std::string (Object::* GetObjectNameMethod)(mpcard index);

 friend ObjectList_Prop;
 friend ObjectListElementEditor;

 private:

   ObjectListProperty() {}

   AddObjectMethod      m_AddMethod;
   DeleteObjectsMethod   m_DelMethod;
   GetObjectMethod      m_GetMethod;
   MoveObjectMethod     m_MoveMethod;
   GetNumObjectsMethod  m_GetNumMethod;
   GetObjectNameMethod  m_GetNameMethod;

   card32               m_PreferedEditorWidth;
   card32               m_PreferedEditorHeight;
   MetaClass            *m_ObjectBaseClass;

 public:

   ObjectListProperty( const std::string &cName, const Version introducedInVersion,
      const MetaClass *objectBaseClass,
      AddObjectMethod addMethod, DeleteObjectsMethod delMethod,
      GetObjectMethod getMethod, MoveObjectMethod moveMethod, GetNumObjectsMethod getNumMethod,
      GetObjectNameMethod getNameMethod, card32 preferedEditorWidth = 0,
         card32 preferedEditorHeight = 300 );

   card32 getPreferedEditorWidth() const { return m_PreferedEditorWidth;  }
   void setPreferedEditorWidth( const card32 &Width )  {  m_PreferedEditorWidth = Width; }
   card32 getPreferedEditorHeight() const { return m_PreferedEditorHeight;  }
   void setPreferedEditorHeight( const card32 &height )  {  m_PreferedEditorHeight = height; }
   MetaClass * getObjectBaseClass() const {return m_ObjectBaseClass; }
   void setObjectBaseClass( MetaClass * mc ) {m_ObjectBaseClass = mc; }

   virtual const std::string  getTypeName() const
   {
      return "ObjectList";
   }

   virtual void writeProperty(const Object* obj, pointer_diff offset, OutputObjectStream *out) const;
   virtual void readProperty(Object* obj, pointer_diff offset, InputObjectStream *in, card64 expectedSize) const;

   virtual std::string toString(const Object* obj, pointer_diff offset) const {forget(obj ); return "";}
   virtual void   fromString(Object* obj, pointer_diff offset, const std::string str) const;

   virtual void assignProperty(const Object *from, Object *to, pointer_diff offset) const;
   virtual ClassProperty* copyClassProperty();

 private:

   void addObject( Object *obj, Object * toAdd ) const
   {
      (obj->*m_AddMethod)(toAdd);
   }

   void deleteObjects( Object *obj, mpcard begin, mpcard end ) const
   {
      (obj->*m_DelMethod)(begin,end);
   }

   Object * getObject( Object * obj, mpcard index ) const
   {
      return (obj->*m_GetMethod)(index);
   }

   void moveObject(Object * obj,mpcard oldPos, mpcard newPos) const
   {
      (obj->*m_MoveMethod)(oldPos,newPos);
   }
   mpcard getNumObjects(Object * obj) const
   {
      return (obj->*m_GetNumMethod)();
   }

   std::string getObjectName( Object * obj, mpcard index ) const
   {
      return (obj->*m_GetNameMethod)(index);
   }


};

#define IMPLEMENT_OBJECT_LIST_METHODS_STL(propName,cName) \
	void add##propName##( Object * obj ) {m_##propName .push_back( ( cName *) obj );} \
	void delete##propName##s( mpcard begin, mpcard end ) {\
	for( mpcard i=begin;i<end;i++ ) delete m_##propName##[i];  m_##propName##.erase( m_##propName##.begin() + begin, m_##propName##.begin() + end ); \
} \
	Object * get##propName##( mpcard index ) {return m_##propName##[index];} \
	void move##propName##( mpcard oldPos, mpcard newPos ) {\
	cName##* tmp = m_##propName##[oldPos]; \
	m_##propName##.erase(m_##propName##.begin() + oldPos); \
	m_##propName##.insert(m_##propName##.begin() + newPos, tmp); \
} \
	mpcard getNum##propName##s() {return m_##propName##.size(); } \
	std::string get####propName##Name(mpcard index ) {return intToStr(index);}

#define IMPLEMENT_OBJECT_LIST_METHODS_STL_NO_PREFIX_M(propName,cName) \
	void add##propName##( Object * obj ) {propName .push_back( ( cName *) obj );} \
	void delete##propName##s( mpcard begin, mpcard end ) {\
	for( mpcard i=begin;i<end;i++ ) delete propName##[i];  propName##.erase( propName##.begin() + begin, propName##.begin() + end ); \
} \
	Object * get##propName##( mpcard index ) {return static_cast<cName##*>(propName##[index]);} \
	void move##propName##( mpcard oldPos, mpcard newPos ) {\
	cName##* tmp = static_cast<cName##*>(propName##[oldPos]); \
	propName##.erase(propName##.begin() + oldPos); \
	propName##.insert(propName##.begin() + newPos, tmp); \
} \
	mpcard getNum##propName##s() {return propName##.size(); } \
	std::string get####propName##Name(mpcard index ) {return intToStr(index);}


#define IMPLEMENT_OBJECT_LIST_METHODS_STL_WITH_OWNER(propName,cName,isOwner) \
	void add##propName##( Object * obj ) {m_##propName .push_back( ( cName *) obj );} \
	void delete##propName##s( mpcard begin, mpcard end ) {\
	if(isOwner){for( mpcard i=begin;i<end;i++ ) delete m_##propName##[i];}  m_##propName##.erase( m_##propName##.begin() + begin, m_##propName##.begin() + end ); \
} \
	Object * get##propName##( mpcard index ) {return m_##propName##[index];} \
	void move##propName##( mpcard oldPos, mpcard newPos ) {\
	cName##* tmp = m_##propName##[oldPos]; \
	m_##propName##.erase(m_##propName##.begin() + oldPos); \
	m_##propName##.insert(m_##propName##.begin() + newPos, tmp); \
} \
	mpcard getNum##propName##s() {return m_##propName##.size(); } \
	std::string get####propName##Name(mpcard index ) {return intToStr(index);}

#define ADD_OBJECT_LIST_PROP( propName, version, metaClass ) \
	((MetaClass*)getClass())->addProperty( new ObjectListProperty( #propName , MetaClass::FIRST_VERSION + version, \
	metaClass, (ObjectListProperty::AddObjectMethod)&add##propName, \
	(ObjectListProperty::DeleteObjectsMethod)&delete##propName##s , \
	(ObjectListProperty::GetObjectMethod)&get##propName , \
	(ObjectListProperty::MoveObjectMethod)&move##propName , \
	(ObjectListProperty::GetNumObjectsMethod)&getNum##propName##s , \
	(ObjectListProperty::GetObjectNameMethod)&get##propName##Name ) );
}

#endif
