﻿//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "ObjectStreams.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
#include "MetaClass.h"
#include "StreamClassTable.h"
#include "UnknownObject.h"
#include <X4Types.h>
#include <X4Assert.h>
#include <StringHelper.h>
#include <HashTables.inline.h>
#include <ConsoleOutput.h>
//---------------------------------------------------------------------------
#undef DEBUG_STREAMS
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------


static const int FILE_ID_STRING_SIZE = 16;  // this should be always 16 byte in size
static const char FILE_ID_STRING[16] = {
	0x27, 0x18, 0x27, 0x79, // ✘➍
	'O', 'b', 'j',
	'S', 't', 'r', 'e', 'a', 'm',
	64, '\0', '\0'};

static const card32 HASHTABLE_FILLFACTOR = 16;

/// explicit instantiation so that "HashTables.inline.h" does not need to be included in modules using this one
template class HashTable<mpcard, Object*, ObjectStream_ObjectHashComperator>;

max_platform_card ObjectStream_ObjIDHashTable::hashFunction(mpcard objID) const {
	return objID;
}


//---------------------------------------------------------------------------
// Output
//---------------------------------------------------------------------------

OutputObjectStream::OutputObjectStream(StreamClassTable* _classTable, bool _doWriteClassTable) :
		OutputStream(),
		doWriteClassTable(_doWriteClassTable),
		classTable(_classTable),
		externClassTable(true)
{
	objIdTable.setFillFactor(HASHTABLE_FILLFACTOR);

	if (classTable == NULL) {
		classTable = new StreamClassTable();
		externClassTable = false;
	}
}

OutputObjectStream::~OutputObjectStream()
{
	if (!externClassTable) {
		delete classTable;
	}
}

void OutputObjectStream::initStream()
{
	writeBuffer(FILE_ID_STRING, FILE_ID_STRING_SIZE);
	card32 flags = 0;

	if (!doWriteClassTable) flags |= 0x10000000;

	write<card32>(flags);

	if (doWriteClassTable) {
		writeClassTableHeader();
	}
}

void OutputObjectStream::removeFromCache(const Persistent *obj)
{
	objIdTable.deleteElement((mpcard)obj);
}

void OutputObjectStream::clearCache()
{
	objIdTable.clear();
}

void OutputObjectStream::writeObject(const Persistent *obj)
{
	if (obj == NULL)
	{
		write<card64>(0);
	} else
	{
		card32 classId = includeClass(obj->getInstanceClass());

		Object* storedID;

		// Write object if it wasn't written before
		if (!objIdTable.getData((mpcard)obj, storedID))
		{
			card64 objID = (card64)obj;
			write<card64>(objID);

			objIdTable.addElement((mpcard)obj, (Object*)obj);
			write<card32>(classId);

			// get current position of the stream. we use this to write here the size
			// of bytes the object takes in the stream after it has been written
			WriteMarker objSizeMarker = createMarker();
			write<card64>(0); // reserve space to write here later the size of the object

			obj->write(this);

			// go back in the stream and write there the object size
			WriteMarker afterObjMarker = createMarker();
			jumpToMarker(objSizeMarker);
			write<card64>(afterObjMarker.difference(objSizeMarker) - sizeof(card64));
			jumpToMarker(afterObjMarker);
		} else
		{
			x4Assert( (storedID == (Object*)obj || ((card64)storedID > 0 && (card64)storedID < 100)) );
			write<card64>((card64)storedID);
		}
	}
}

void OutputObjectStream::writeMetaClass(const MetaClass *mc)
{
	write<card32>(classTable->getOrCreateClassID(mc));
}

card32 OutputObjectStream::includeClass(const MetaClass *mc)
{
	card32 id = classTable->getOrCreateClassID(mc);

	// Also put all parent classes into the class table
	mc = mc->getParentClass();
	while (mc != NULL)
	{
		classTable->getOrCreateClassID(mc);
		mc = mc->getParentClass();
	}

	return id;
}


//---------------------------------------------------------------------------
// Input
//---------------------------------------------------------------------------
InputObjectStream::InputObjectStream(StreamClassTable *classTable) :
		InputStream(),
		classTable(classTable),
		externClassTable(true)
{
	objIdTable.setFillFactor(HASHTABLE_FILLFACTOR);
	lastClass = NULL;
	if (classTable == NULL) {
	  this->classTable = new StreamClassTable();
	  externClassTable = false;
	}
	majorStreamVersion = SV_UNDEFINED;
}

void InputObjectStream::initStream()
{
	char fileID[FILE_ID_STRING_SIZE];
	readBuffer(fileID, FILE_ID_STRING_SIZE);
	if (memcmp(FILE_ID_STRING, fileID, FILE_ID_STRING_SIZE) != 0) {
		 throw EIOException("BinaryInputObjectStream::BinaryInputObjectStream() - the file is not a X4ObjStream64.");
	} else {
	  majorStreamVersion = SV_X4ObjStream64;
	}
	card32 flags;
	read<card32>(flags);

	if ((flags & 0x10000000) == 0)
		readClassTable();
}

InputObjectStream::~InputObjectStream()
{
	if (!externClassTable) {
	  delete classTable;
	}
}

void InputObjectStream::readMetaClass(const MetaClass *&mc)
{
	card32 classId;
	read<card32>(classId);
	mc = classTable->getClass(classId);
	if (mc == NULL) {
		throw EIOException(std::string("InputObjectStream::readMetaClass() - class not found, id: ") + intToStr(classId).c_str());
	}
}

const PropertyTable* InputObjectStream::getPropertyTable(const MetaClass* mc)
{
	return classTable->getPropertyTable(mc);
}

void InputObjectStream::removeFromCache(const Persistent *obj)
{
	objIdTable.deleteElement((mpcard)obj);
}

void InputObjectStream::clearCache()
{
	objIdTable.clear();
}

void InputObjectStream::readObjectPersistent(Persistent *&obj)
{
	#ifdef DEBUG_STREAMS
	static card32 DEBUG_IDENT = 0;
	#endif

	card64 objID;
	if (majorStreamVersion == SV_X4ObjStream64) {
		read<card64>(objID);
	}else
	{
		throw EIOException(std::string("InputObjectStream::readObject() - unsupported stream version, cannot read objects"));
	}

	#ifdef DEBUG_STREAMS
	debugOutput.cr();
	debugOutput.writeIdented(3*DEBUG_IDENT, ((std::string("r objID: ") + intToStr(objID))));
	debugOutput.cr();
	#endif

	if (objID == 0)
	{
		obj = NULL;
		return;
	} else
	{
		Persistent* storedObj = NULL;
		bool found = objIdTable.getData((mpcard)objID, (Object*&)storedObj);

		// Read object if it wasn't read before
		if (!found)
		{
			card32 uniqueClassID;
			read<card32>(uniqueClassID);

			#ifdef DEBUG_STREAMS
			debugOutput.writeIdented(3*DEBUG_IDENT, ((std::string("r classID: ") + intToStr(uniqueClassID))));
			debugOutput.cr();
			#endif

			const MetaClass* objClass = classTable->getClass(uniqueClassID);
			x4Assert(objClass != NULL); // here we know that even for unknown classes there will be a metaclass
			/*if (objClass == NULL)
			{
				std::string lastClassName = lastClass == NULL ? "no previous class" : lastClass->getClassName();
				warning(std::string("InputObjectStream::readObject() - class not found, id: ")
									+ intToStr(uniqueClassID)
									+ std::string("Last class has been: ")
									+ lastClassName);
			}*/
			lastClass = objClass;

			#ifdef DEBUG_STREAMS
			debugOutput.writeIdented(3*DEBUG_IDENT, std::string("r class name: ") + objClass->getClassName());
			debugOutput.cr();
			#endif

			// if class is known then use it, otherwise use unknown class
			//if (objClass)
				obj = static_cast<Persistent*>(objClass->newInstance());
			//else
			//	obj = static_cast<Persistent*>(UnknownObject::getClass()->newInstance());

			objIdTable.addElement((mpcard)objID, obj);

			// read next the size in bytes an object will take
			card64 objSize = 0;
			read<card64>(objSize);
			size_t oldpos = seek(0);

			#ifdef DEBUG_STREAMS
			DEBUG_IDENT++;
			#endif

			// try to read the object
			// if fails we move the read pointer after the object in the file
			//try
			//{

			// if this class was unknown during the compile time, then we have to parse it just as simple raw stream
			if (!objClass->isKnown())
				dynamic_cast<UnknownObject*>(obj)->readFromRaw(this, uniqueClassID, objSize);
			else
				obj->read(this);

			/*}catch(...)
			{
				warning(std::string("InputObjectStream::readObject() - object of type ")
								  + obj->getInstanceClass()->getClassName()
								+ std::string(" could not be readed"));
			}*/

			size_t curpos = seek(0);
			if (curpos - oldpos < objSize)
			{
				x4Assert(objClass != NULL); // if reading as unknown object, then we should be always safe with amount of readed bytes
				consoleWarning(std::string("InputObjectStream::readObject() - object of type ")
								  + objClass->getClassName()
								+ std::string(" takes less bytes in the stream as in real. Hence the remaining bytes will be ignored."));
				seek(objSize - (curpos - oldpos));
			}else if (curpos - oldpos > objSize)
			{
				throw EIOException(std::string("InputObjectStream::readObject() - object of type ")
								  + objClass->getClassName()
								+ std::string(" gathered more bytes from the stream as it has to. The stream is irreporably corrupted."));
			}


			#ifdef DEBUG_STREAMS
			DEBUG_IDENT--;
			#endif

			return;
		} else
		{
			obj = storedObj;
			return;
		}
	}
}


card32 InputObjectStream::getClassVersion(const MetaClass* metaClass)
{
	card32 id = classTable->getClassID(metaClass);
	if (id == StreamClassTable::BAD_CLASS) {
		return (card32)-1;
	} else {
		return classTable->getCurrentVersion(id);
	}
}

#if 0
	void InputObjectStream::addGlobalObject(Persistent *obj, mpcard index) {
	   x4Assert(index < 100); // EVIL hack - assumes no objects in first 100 bytes and supports only 100 global objects
	   objIdTable.addElement(index+1, (Object*)obj);
	}
#endif 




