//---------------------------------------------------------------------------
#ifndef ObjectStreamsH
#define ObjectStreamsH
//---------------------------------------------------------------------------
#include "x4core.h"
#include "OutputStream.h"
#include "InputStream.h"
//---------------------------------------------------------------------------
#include <X4Types.h>
#include <HashTables.h>
#include <string>
#include <vector>
//---------------------------------------------------------------------------


namespace X4
{
   

	class Object;
	class Persistent;
	class MetaClass;
	class StreamClassTable;

	/// helper class for object streams;
	class X4CORE_API ObjectStream_ObjectHashComperator
	{
	public:
		inline static bool equals(mpcard obj1, mpcard obj2){return obj1 == obj2;}
	};

	/// helper class for object streams;
	class X4CORE_API ObjectStream_ObjIDHashTable : public HashTable<mpcard, Object*, ObjectStream_ObjectHashComperator>
	{
	public:
		virtual max_platform_card hashFunction(mpcard objID) const;
	};



	/// class that allows a serialization of objects of type "Persistent" (output version).
	class X4CORE_API OutputObjectStream : public OutputStream
	{
	private:
		ObjectStream_ObjIDHashTable objIdTable;

	protected:
		StreamClassTable *classTable;
		bool doWriteClassTable;
		bool externClassTable;

		/// some initialization during construction, to be called in subclass after stream has been opened / established
		void initStream();

		/// reserve space for pointer to class table (in non-seeking files, a all-classes table might be submitted at this point)
		/// (only called if constructor parameter doWriteClassTable == true)
		virtual void writeClassTableHeader() = 0;

		/// write the class table after all objects and their classes have been recorded
		/// (only called if constructor parameter doWriteClassTable == true)
		virtual void writeClassTable() = 0;

	public:
		/// Construct stream; usually, the default parameters should be used.
		/// For special applications, the classTable (a dictionary of classes included in the stream, beeing updated during writing)
		/// can be specified manually and its (default) inclusion in the stream can be ommited (e.g. if multiple streams are
		/// packed into one file, only one class table might be diserable).
		OutputObjectStream(StreamClassTable *classTable = NULL, bool doWriteClassTable = true);

		/// close stream again
		virtual ~OutputObjectStream();

		/// write an object (including class id, multiple references are only written once and tagged with an id)
		virtual void writeObject(const Persistent *obj);

		/// remove already written object from cache, so that it can be written again even if it is the same object
		virtual void removeFromCache(const Persistent *obj);

		/// clean the whole cache of already written objects
		virtual void clearCache();

		/// write a meta class (class id)
		virtual void writeMetaClass(const MetaClass *mc);

		/// include a meta class in the class id table (although the class is never encountered by writeObject/writeMetaClasss);
		/// usefull for embedded classes with versioning
		/// returns classID for the given class in the stream
		virtual card32 includeClass(const MetaClass *mc);

		/// this function returns a filename if the stream is being stored to a file
		/// if not, an empty std::string is returned
		virtual std::string getFileNameIfFileStream() {return std::string("");}
	};



	/// class that allows a serialization of objects of type "Persistent" (input version).
	class X4CORE_API InputObjectStream : public InputStream
	{
	private:
		ObjectStream_ObjIDHashTable objIdTable;

		/// type to distinguish between major revisions of the streaming system
		/// will be set according to the first 15 characters of the file, as in the suffix below
		enum MajorStreamVersion
		{
			/// initial value, at streaming-time, it causes an exception      
			SV_UNDEFINED = 0xFFFFFFFF, 

			// TODO add streaming format for X43, in order to be able to read it files

			/// streaming format 64 bit capable
			SV_X4ObjStream64 = 1 << 4
		};
		MajorStreamVersion majorStreamVersion;


	protected:
		const MetaClass *lastClass;
		StreamClassTable *classTable;
		bool externClassTable;

		/// some initialization during construction, to be called in subclass after stream has been opened / established
		void initStream();

		/// retrieve class table from stream (called only if a class table has been included by writing)
		virtual void readClassTable() = 0;

		/// read an object (including class id, multiple references to the same object are reconstructed properly)
		virtual void readObjectPersistent(Persistent *&obj);

	public:
		/// Construct stream object. If the stream to be read does not include a class table, a matching object has to be specified
		/// to the constructor. By default, all streams include a class table, the parameter might be NULL.
		InputObjectStream(StreamClassTable *classTable = NULL);

		/// close stream again
		virtual ~InputObjectStream();


		/// remove readed object from cache, so that it can be loaded again even if it is the same object, which was loaded before
		virtual void removeFromCache(const Persistent *obj);

		/**
		* Get property table associated with the metaclass in the given stream.
		* This property table might be different of that one which is currently compiled in.
		**/
		const PropertyTable* getPropertyTable(const MetaClass* mc);

		/// clean the whole cache of already written objects
		virtual void clearCache();

		/// template to avoid conversion issues with readObject
		template <class T>
		inline void readAnyObject(T *&obj) {Persistent *p; readObjectPersistent(p); x4Assert(p == NULL || p->getInstanceClass()->inheritsFrom(T::getClass())); obj = (T*)p;}

		/// michael: changed read object implementation to default to templated version (avoids bugs) - 07/28/2012
		template <class T>
		inline void readObject(T *&obj) {readAnyObject<T>(obj);}

		/// write a meta class (class id)
		virtual void readMetaClass(const MetaClass *&mc);

		/// return version of the given class in the stream
		card32 getClassVersion(const MetaClass* metaClass);

		/// this function returns a filename if the stream is being stored to a file
		/// if not, an empty std::string is returned
		virtual std::string getFileNameIfFileStream() {return std::string("");}

#if 0
		/// add an object that is not stored in the stream but only as a fixed pointer (obj)
		virtual void addGlobalObject(Persistent *obj, mpcard index);
#endif 

	};


}


#endif
