//---------------------------------------------------------------------------
#ifndef OutputStream__H
#define OutputStream__H
//---------------------------------------------------------------------------
#include "x4core.h"
#include "StreamContext.h" 
//---------------------------------------------------------------------------
#include <string>
//---------------------------------------------------------------------------
#include <X4Types.h>
#include <X4Exception.h>
#include <Tools.h>
//---------------------------------------------------------------------------

namespace X4
{
	/**
	* Serialization of data is handled by classes derived from this abstract definition
	**/
	class X4CORE_API OutputStream
	{
	protected:
		StreamContext* m_context;

	public:
		OutputStream() : m_context(NULL) {}
		virtual ~OutputStream(){ if (m_context) delete m_context; }

		//! Return context assosiated with this stream
		StreamContext* getContext() { return m_context; }

		//! Set context to be used with this stream. The stream takes ownership of the context
		void setContext(StreamContext* ctx) { if (m_context) delete m_context; m_context = ctx; }

		/**
		* Marker which can be used in streams to be able to jump to certain marked positions in the stream
		**/
		class WriteMarker
		{
		private:
			friend class OutputStream;
			size_t pos;
			WriteMarker(size_t pos) : pos(pos) {};
		public:
			WriteMarker(const WriteMarker& mrk) : pos(mrk.pos) {}
			inline size_t difference(const WriteMarker& mrk) { if (mrk.pos <= pos) return pos - mrk.pos; else return mrk.pos - pos; } 
			~WriteMarker() {}
		};

		/**
		* Indicate if this stream can perform seek operations.
		**/
		virtual bool canSeek() const { return false; }

		/**
		* Seek in the stream. If stream does not support seek operation, then an error is thrown.
		* @param rel Relative seek position, can be negative to seek backward.
		* @return Return the current position within the stream after seek completed
		**/
		virtual card64 seek(int64 rel) { forget(rel); throw EIOException("OutputStream::seek() - cannot seek!"); }

		/**
		* Retrieve current position marker. This marker can be used in order to jump to that position later.
		* The marker is not related to the position in the stream in bytes. The marker can only be obtained if canSeek() is true.
		**/
		inline WriteMarker createMarker() { return WriteMarker(seek(0)); }

		/**
		* Jump to a given marker. The marker has to be created before with createMarker(). 
		* If stream does not support seeking then error is thrown
		* NOTE: the jump might be problematic when jumping farther away then max of int64, 
		*       however I suppose we should be ok with that for at least next 50 years, <art tevs> 2012
		**/
		inline void jumpToMarker(const WriteMarker& marker) { seek(static_cast<int64>(marker.pos - seek(0))); }

		/**
		* Generate new stream which inherits all the properties of this stream. The new stream
		* can be used to store additional data in parallel with this stream. It depends on the stream's 
		* implementation to provide right behaviour. For example a file stream which
		* generates a sibling stream might make sure that the new stream is stored under the same path.
		* A database stream might for example provide sibling streams to be stored on the same machine.
		* If NULL is given back, then sibling streams are not supported.
		* @param uid Unique identifier used to identify siblings stream among all created. This is usefull when opening a sibling stream for input
		*        and required to open the same again. It is in responsibility of the stream to make sure that sibling corresponding to correct UID are opened (when for input).
		* @param raw Indicates that this streams is raw. Raw streams does not hold any class tables or other nice stuff, they just used in their raw form to dump data to it.
		**/
		virtual OutputStream* createSiblingStream(card64 uid, bool raw = false) { forget(uid); forget(raw); return NULL; }


		/**
		* Write a buffer of bytes. to be overridden in descendant classes.
		**/
		virtual void writeBuffer(const void* buf, mpcard length) = 0;

		/**
		* Write arbitrary data types as block of bytes
		**/
		template <class ElementaryType>
		inline void write(const ElementaryType &v)
		{
			writeBuffer(static_cast<const void*>(&v), sizeof(ElementaryType));
		};

		/// write boolean value
		inline void writeBool(const bool &v)
		{
			static const card8 t = 1;
			static const card8 f = 0;
			writeBuffer(v ? &t : &f, 1);
		}

		/// write an stl std::string
		inline void writeString(const std::string &v)
		{
			card64 length = v.length() * sizeof(std::string::value_type);
			write(length);
			if (length > 0)
				writeBuffer(v.data(), length);
		}

		/// write a cardinal value with maximum bit length of the current platform;
		/// in the stream, a 64-bit representation is always used
		inline void writeMaxPlatformCard(const max_platform_card &v)
		{
			card64 vlarge = v;
			write<card64>(vlarge);
		}

		/// as writeMaxPlatformCard, but for signed values
		inline void writeMaxPlatformInt(const max_platform_int &v)
		{
			int64 vlarge = v;
			write<int64>(vlarge);
		}
	};

}


#endif
