//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PropertyTableProperty.h"
//---------------------------------------------------------------------------
#include "StringHelper.h"
#include "X4Exception.h"
//---------------------------------------------------------------------------

namespace X4
{

	//---------------------------------------------------------------------------
	PropertyTableProperty::PropertyTableProperty():
		ClassProperty(getPropertyTablePropertyName(), 0) // always with the same name and version
	{
		bIgnoreOnRead = false;
	}

	//---------------------------------------------------------------------------
	void PropertyTableProperty::writeProperty(const Object* obj, pointer_diff offset, OutputObjectStream *out) const
	{
		// first we write our own version number
		out->write<card32>(tableVersion);

		// next write information over all properties found in the MetaClass of the given object
		const MetaClass* meta = obj->getInstanceClass();
		out->write<card32>(meta->getNumProperties());
		for (unsigned i=0; i < meta->getNumProperties(); i++)
		{
			const ClassProperty* prop = meta->getProperty(i);

			// @TODO: write unique property id and/or size of the property
			// if the property for example could not be found in the current class, then we could use
			// the unique property ID to built a dummy property and to read from file so to skip
			// the bytes in the stream for a property which is not registered with an object anymore

			// write properties of the property
			out->write<Version>(prop->getIntroducedInVersion());
			out->writeString(prop->getTypeName());
			out->writeString(prop->getCName());
		}
	}

	//---------------------------------------------------------------------------
	void PropertyTableProperty::readProperty(Object* obj, pointer_diff offset, InputObjectStream *in, card64 expectedSize) const
	{
		if (bIgnoreOnRead) return;

		// first table version of the property must be readed
		card32 tabver = 0;
		in->read<card32>(tabver);

		// for later changes we should be able to make corresponding reads for different table version
		if (tabver != tableVersion)
		{
			throw EIOException("Property table has an unsupported version number. Please contact owner of the file!");
			return;
		}

		// read all properties into an internal property table
		card32 numProperties = 0;
		in->read<card32>(numProperties);
		m_Table.resize(numProperties);
		for (unsigned i=0; i < numProperties; i++)
		{
			in->read<Version>(m_Table[i].introducedInVersion);
			in->readString(m_Table[i].typeName);
			in->readString(m_Table[i].propertyName);
		}
	}

	//---------------------------------------------------------------------------
	std::string PropertyTableProperty::toString(const Object* obj, pointer_diff offset) const
	{
		// copy property table into a std::string
		std::string serialized = intToStr(tableVersion) + '?';
		serialized += intToStr(m_Table.size()) + ':';
		for (unsigned i=0; i < m_Table.size(); i++)
		{
			mpcard ver = m_Table[i].introducedInVersion;
			serialized += intToStr(ver) + ',';
			serialized += m_Table[i].typeName + ',';
			serialized += m_Table[i].propertyName + ';';
		}
		return serialized;
	}

	//---------------------------------------------------------------------------
	void PropertyTableProperty::fromString(Object* obj, pointer_diff offset, const std::string str) const
	{
		#define ParseUntil(a) {\
			temp = ""; \
 			for (; pos < str.length(); pos++) \
			{ \
				if (str[pos] != a) temp += str[pos]; \
				else break; \
			} \
		}

		// built internal structure form given std::string
		std::string temp;
		unsigned pos = 0;

		// read the magic version number
		ParseUntil('?');
		if (temp.length()==0)
			throw EIOException("Serialization std::string for the property table does not contain the magic version number");

		card32 tabver;
		strToInt(temp,tabver);
		if (tabver != tableVersion)
			throw EIOException("Property table has not supported version number. Please contact owner of the file!");

		// read the size of the table
		ParseUntil(':');
		if (temp.length()==0)
			throw EIOException("Serialization std::string for the property table does not contain the number of table entries");
		card32 size;
		strToInt(temp, size);
		m_Table.resize(size);

		// read entries
		for (unsigned i=0; i < m_Table.size(); i++)
		{
			// version
			ParseUntil(',');
			if (temp.length()==0)
				throw EIOException(std::string("Cannot parse table property ") + intToStr(i));
			
			mpcard ver;
			strToInt(temp, ver);
			m_Table[i].introducedInVersion = ver;

			// type name
			ParseUntil(',');
			if (temp.length()==0)
				throw EIOException(std::string("Cannot parse table property ") + intToStr(i));
			m_Table[i].typeName = temp;

			// property name
			ParseUntil(';');
			if (temp.length()==0)
				throw EIOException(std::string("Cannot parse table property ") + intToStr(i));
			m_Table[i].propertyName = temp;
		}

		#undef ParseUntil
	}

	//---------------------------------------------------------------------------
	const std::string PropertyTableProperty::getTypeName() const
	{
		return "property_table";
	}

	//---------------------------------------------------------------------------
	// comment out by CL @ 2014/02/17
	ClassProperty* PropertyTableProperty::copyClassProperty()
	{
		PropertyTableProperty* copy = new PropertyTableProperty();
		copy->m_Table = m_Table;
		return copy;
	}

	//---------------------------------------------------------------------------
	void PropertyTableProperty::assignProperty(const Object *from, Object *to, pointer_diff offset) const
	{
		// do nothing, since this property is some kind of a meta property, so no need to assign this to other object
	}


};



