//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "StreamClassTable.h"
//---------------------------------------------------------------------------
#include "ObjectStreams.h"
//---------------------------------------------------------------------------
#include <X4Assert.h>
#include <X4Exception.h>
#include <HashTables.inline.h>
#include <ConsoleOutput.h>
#include <boost/lexical_cast.hpp>
#include "UnknownObject.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------


MetaClass* StreamClassTable::StreamClassTableMetaClass = NULL;

const MetaClass* StreamClassTable::getInstanceClass() const {
   return getClass();
}

Object* StreamClassTable::newInstance() {
   return new StreamClassTable;
}

const MetaClass* StreamClassTable::getClass() {
   if (StreamClassTableMetaClass==NULL) {
      StreamClassTableMetaClass = new MetaClass("StreamClassTable", "no description.",
                                                newInstance, sizeof(StreamClassTable));
   }
   return StreamClassTableMetaClass;
}

void StreamClassTable::init(const MetaClass *parentClass) {
   ((MetaClass*)getClass())->init(parentClass);
   StreamClassTableMetaClass->copyParentProperties();
}

void StreamClassTable::shutdown() {
   delete StreamClassTableMetaClass;
}


// ----


card32 StreamClassTable::getOrCreateClassID(const MetaClass* mc) {
   card32 id;
   if (idByClass.getData(mc, id)) {
      return id;
   } else {
      id = (card32)metaClasses.size();
      MetaClassEntry me;
      me.mc = mc;
      me.currentStreamIDNumber = id;
      me.mainVersion = mc->getClassVersion();
      metaClasses.push_back(me);
      idByClass.addElement(mc, id);
      return id;
   }
}

card32 StreamClassTable::getClassID(const MetaClass* mc) const {
   card32 id;
   if (idByClass.getData(mc, id)) {
      return id;
   } else {
      return BAD_CLASS;
   }
}

const PropertyTable* StreamClassTable::getPropertyTable(card32 id) const
{
	if (id < metaClasses.size())
		return &metaClasses[id].pt;

	return NULL;
}

const PropertyTable* StreamClassTable::getPropertyTable(const MetaClass* mc) const
{
	card32 id;
	if (idByClass.getData(mc, id))
		return &metaClasses[id].pt;
	else
		return NULL;
}

const MetaClass *StreamClassTable::getClass(card32 id) const
{
	if (id < metaClasses.size()) {
		return metaClasses[id].mc;
	}

	return NULL;

	// throw X4Exception("StreamClassTable::getClass - unknown class.");
}

card32 StreamClassTable::getNumClassIDs() const {
   return (card32)metaClasses.size();
}


card32 StreamClassTable::getCurrentVersion(card32 id) const {
	const MetaClassEntry *me = &(metaClasses[id]);
	return me->mainVersion;
}

card32 StreamClassTable::getCurrentVersion(const MetaClass *mc) const {
	card32 id = getClassID(mc);
	x4Assert(id != BAD_CLASS);
	return getCurrentVersion(id);
}


static void writeClassTableEntry(void* context, const MetaClass* metaClass, card32 id)
{
	OutputObjectStream* s = (OutputObjectStream*)context;

	#ifdef DEBUG_STREAMS
	debugOutput << id << " " << metaClass->getClassName() << " " << metaClass->getClassVersion();
	debugOutput.cr();
	#endif

	// write some info this stream table requires
	s->write<card32>(id);
	s->writeString(metaClass->getClassName());
	s->write<card32>(metaClass->getInstanceSize());
	s->write<card32>(metaClass->getClassVersion());

	// let metaclass serialize itself
	metaClass->getPropertyTable()->write(s);
	metaClass->write(s);
}

static Object* newUnknownObject()
{
	return new UnknownObject();
}

void StreamClassTable::write(OutputObjectStream *s) const {
   s->write<card32>(idByClass.count());
   const_cast<StreamClassTable*>(this)->idByClass.forEachEntry(s, writeClassTableEntry);
}

void StreamClassTable::read(InputObjectStream *s)
{
	card32 classTableCount;
	s->read<card32>(classTableCount);
	metaClasses.resize(classTableCount);

	for (card32 classNum = 0; classNum != classTableCount; ++classNum)
	{
		// Read class ID
		card32 classId;
		s->read<card32>(classId);
		x4Assert(classId < classTableCount);

		// Read class name and instance size, to find out if such class does exists
		std::string className;
		s->readString(className);

		card32 insSize = 0;
		s->read<card32>(insSize);
		x4Assert(insSize > 0);

		// Read class version
		card32 classVersion;
		s->read<card32>(classVersion);

		const MetaClass* metaClass = MetaClass::getClassByName(className);
		if (metaClass == NULL)// || metaClass->getInstanceSize() != insSize)
		{
			// if metaclass is not known, then we just create a new metaclass representing the unknown class
			MetaClass* umc = new MetaClass(className, 
					   std::string("unknown class ") + className + std::string(" found at byte ") + boost::lexical_cast<std::string>(s->seek(0) - sizeof(card32)*2)
					   + std::string(" of version ") + boost::lexical_cast<std::string>(classVersion)
					   + std::string(" (") + boost::lexical_cast<std::string>(insSize) + std::string(" bytes)")					   
					   , newUnknownObject, insSize);

			umc->init(NULL, classVersion, false);
			UnknownObject::unknownClasses.push_back(umc);

			metaClass = umc;

			consoleWarning(std::string("an unknown class '") + className + std::string("' was found in the stream."));
		}


		// add readed class to the class table and read the class from the stream
		MetaClassEntry me;
		me.mc = metaClass;
		me.currentStreamIDNumber = classId;
		me.mainVersion = classVersion;
		//me.pt = new PropertyTable;
		me.pt.read(s);
		metaClasses[classId] = me;

		idByClass.addElement(metaClass, classId);

		MetaClass* cls = const_cast<MetaClass*>(metaClass);
		cls->read(s);

		#ifdef DEBUG_STREAMS
		debugOutput << classId << " " << className << " " << (IntToStr(classVersion)).c_str();
		debugOutput.cr();
		#endif
	}
}


