//---------------------------------------------------------------------------
#ifndef StreamClassTableH
#define StreamClassTableH
//---------------------------------------------------------------------------
#include "x4core.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
#include "HashTables.h"
#include "PropertyTable.h"
//---------------------------------------------------------------------------

namespace X4 {


	class StreamClassTable : public Persistent
	{
	private:
		static MetaClass *StreamClassTableMetaClass;
		static Object* newInstance();

	public:
		struct MetaClassEntry
		{
			const MetaClass *mc;
			card32 currentStreamIDNumber;
			card32 mainVersion;
			PropertyTable pt;

			MetaClassEntry() : mc(NULL), currentStreamIDNumber(0), mainVersion(0) {}
			~MetaClassEntry() { /*if (pt) delete pt;*/ }
		};

	private:
		class X4CORE_API ClassHashComperator {
			public:
			static bool equals(const MetaClass* obj1, const MetaClass* obj2) { return obj1 == obj2; }
		};

		class X4CORE_API ClassCardHashTable : public HashTable<const MetaClass*, card32, ClassHashComperator> {
			public:
			virtual max_platform_card hashFunction(const MetaClass* metaClass) const {
				return (max_platform_card)(metaClass);
			}
		};

		std::vector<MetaClassEntry> metaClasses;
		ClassCardHashTable idByClass;

	public:

		StreamClassTable() {}

		virtual const MetaClass *getInstanceClass() const;
		static const MetaClass *getClass();
		static void init(const MetaClass *parentClass);
		static void shutdown();

		/// Get unique id number for the given class.
		/// The id is unique over the lifetime of this object.
		/// If the class is unknown, a new id is created.
		/// (this also holds if the stream class table has been serialized and reread)
		card32 getOrCreateClassID(const MetaClass* mc);
		/// Get id number for the given class.
		/// If the class is unknown, BAD_CLASS == 0xFFFFFFFF is returned
		/// (this also holds if the stream class table has been serialized and reread)
		card32 getClassID(const MetaClass* mc) const;
		static const card32 BAD_CLASS = 0xFFFFFFFF;

		/// Get class by id. This function must only be called if the
		/// class has been added before by "getClassID()" or "includeClassID()"
		/// (this also holds if the stream class table has been serialized and reread)
		/// If no class with the given id has been found, then NULL is returned
		const MetaClass *getClass(card32 id) const;

		/// Get property table of the metaclass with the given id found in the stream.
		const PropertyTable* getPropertyTable(card32 id) const;
		const PropertyTable* getPropertyTable(const MetaClass* mc) const;

		/// Retrieve number of known class ids. An id is valid if it is smaller than the number of ids.
		card32 getNumClassIDs() const;

		/// Returns current class version (the class must be known, i.e. have an id).
		/// In case a temporary version has been set, the top of the version stack is returned,
		/// otherwise the version at first introduction of the class due to "getClassID()" is returned.
		card32 getCurrentVersion(const MetaClass *mc) const;

		/// Returns current class version.
		/// In case a temporary version has been set, the top of the version stack is returned,
		/// otherwise the version at first introduction of the class due to "getClassID()" is returned.
		card32 getCurrentVersion(card32 id) const;

		/// write to disc (compatible with former "ObjectStreams.h" class tables)
		virtual void write(OutputObjectStream *s) const;
		/// read from disc (compatible with former "ObjectStreams.h" class tables)
		virtual void read(InputObjectStream *s);

		~StreamClassTable() {}
	};


}

#endif
