#ifndef X4VariantH
#define X4VariantH

#include <x4core.h>

#include <X4Types.h>
#include <X4Exception.h>

#include "Persistent.h"


namespace X4
{
	/// this class can take a value of all standard x4 types
	class X4CORE_API Variant : public Persistent
	{
		X_CLASS;

	public:
		enum Type
		{
			X4T_VOID    = 0,
			X4T_CARD8   = 1, X4T_INT8   = 2,
			X4T_CARD16  = 3, X4T_INT16  = 4,
			X4T_CARD32  = 5, X4T_INT32  = 6,
			X4T_CARD64  = 7, X4T_INT64  = 8,
			X4T_MPCARD  = 9, X4T_MPINT  = 10,
			X4T_FLOAT32 = 11, X4T_FLOAT64 = 12,
			X4T_OBJECT  = 13, 
			X4T_STRING  = 14, X4T_MAX_TYPE_INDEX = 14
		};
	private:
		Type type;
		int64 intVal;
		float64 floatVal;
		Object *obj;
		std::string str;

	public:
		Variant();
		virtual ~Variant();

		void assign(const Object* obj, X4CopyContext *context = NULL);

		void write(OutputObjectStream *s) const;
		void read(InputObjectStream *s);

		inline Type getType() const { return type; }
		inline void setVoid() { type = X4T_VOID; }
   
		template<typename T> static Variant* scalar(const T& val)
		{
			throw EInvalidState("Variant invalid scalar type.");
		}
		template<> static Variant* scalar<float32>(const float32& val)
		{
			Variant* var = new Variant; var->setScalar(X4T_FLOAT32, val); return var;
		}
		template<> static Variant* scalar<float64>(const float64& val)
		{
			Variant* var = new Variant; var->setScalar(X4T_FLOAT64, val); return var;
		}
		template<> static Variant* scalar<card8>(const card8& val)
		{
			Variant* var = new Variant; var->setScalar(X4T_CARD8, val); return var;
		}
		template<> static Variant* scalar<card16>(const card16& val)
		{
			Variant* var = new Variant; var->setScalar(X4T_CARD16, val); return var;
		}
		template<> static Variant* scalar<card32>(const card32& val)
		{
			Variant* var = new Variant; var->setScalar(X4T_CARD32, val); return var;
		}
		template<> static Variant* scalar<card64>(const card64& val)
		{
			Variant* var = new Variant; var->setScalar(X4T_CARD64, val); return var;
		}
		template<> static Variant* scalar<int8>(const int8& val)
		{
			Variant* var = new Variant; var->setScalar(X4T_INT8, val); return var;
		}
		template<> static Variant* scalar<int16>(const int16& val)
		{
			Variant* var = new Variant; var->setScalar(X4T_INT16, val); return var;
		}
		template<> static Variant* scalar<int32>(const int32& val)
		{
			Variant* var = new Variant; var->setScalar(X4T_INT32, val); return var;
		}
		template<> static Variant* scalar<int64>(const int64& val)
		{
			Variant* var = new Variant; var->setScalar(X4T_INT64, val); return var;
		}

		/// use this method to read any scalar type
		template<typename ScalarType> 
		ScalarType getScalar() const
		{
			if (type < X4T_CARD8 ||  type > X4T_FLOAT64)
				throw EInvalidState("Variant access via getScalarValue to non scalar type.");

			if (type <= X4T_MPINT) 
				return (ScalarType)(intVal);

			return (ScalarType)(floatVal);
		}

		/// use this method to set a scalar type
		template<typename ScalarType>
		void setScalar(Type newType, const ScalarType &value)
		{
			if (newType < X4T_CARD8 ||  newType > X4T_FLOAT64)
				throw EInvalidState("Variant write via setScalarValue and non scalar type.");

			intVal = (int64)(value);
			floatVal = (float64)(value);

			type = newType;
		}

		/// If the variant contains an object, then retrieve it
		Object *getObject() const
		{
			if (type != X4T_OBJECT)
				throw EInvalidState("Variant access via getObjectValue to non-object type.");
			return obj;
		}

		/// set object type and object pointer
		inline void setObject(Object *obj)
		{
			type = X4T_OBJECT;
			this->obj = obj;
		}

		/// If the variant contains a std::string, then return std::string
		inline const std::string& getString() const 
		{
			if (type != X4T_STRING)
				throw EInvalidState("Variant access via getString to non-object type.");
			return str;
		}
		inline void setString(const std::string& str)
		{
			type = X4T_STRING;
			this->str = str;
		}
	};



}

#endif
