//---------------------------------------------------------------------------
#include "StdAfx.h"
#include "Configuration.h"
//---------------------------------------------------------------------------
#include "ConsoleWidget.h"
//---------------------------------------------------------------------------
#include "Tools.h"
#include "StringHelper.h"
#include "X4Assert.h"
#include "EventViewerWidget.h"
//---------------------------------------------------------------------------
#include <fstream>
#include <QtCore/qtimer.h>
#include <QtCore/qcoreapplication.h>
#include <sstream>
#include <iostream>
#include "ConsoleOutput.h"
#include "ApplicationDirectories.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

bool X4::ALWAYS_PROCESS_EVENTS_ON_FLUSH = false;

ConsoleWidget::ConsoleWidget(QWidget *parent)
:QWidget(parent)
{
    // setup mutex for the text buffer
    bufferMutex = new QMutex(QMutex::Recursive);
	loggedOutput = NULL;

    // connect slots and signals of this object
    connect(this, SIGNAL(sigFlush()), this, SLOT(sltFlush()));
    connect(this, SIGNAL(sigWarningMessage(const std::string&)), this, SLOT(sltWarningMessage(const std::string&)));
    connect(this, SIGNAL(sigErrorMessage(const std::string&)), this, SLOT(sltErrorMessage(const std::string&)));

    stdColor = QColor(0,0,0);				 // not used, initialized only just-in-case
    stdBackground = QColor(192,192,192); // not used, initialized only just-in-case
    //call ui.setupUi(this) to set up the layout from QtDesigner
    ui.setupUi(this);

    //easy access to QtDesigner buttons, etc. by ui object
    //QWidget::connect(ui.BTExec, SIGNAL(clicked()), this, SLOT(sltBTExecClicked()));

    QWidget::connect(ui.CBLogConsole, SIGNAL(stateChanged(int)), this, SLOT(sltCBLogConsoleStateChanged(int)));
    QWidget::connect(ui.CBLogEvents, SIGNAL(stateChanged(int)), this, SLOT(sltCBLogEventsStateChanged(int)));

    timer = new QTimer(this);
    timer->setInterval(500);
    timer->setSingleShot(true);
    timer->stop();
    eventViewer = new EventViewerWidget(ui.tabEvents, 0, false);
    ((QVBoxLayout*)(ui.tabEvents->layout()))->insertWidget(0,eventViewer);
    connect(timer, SIGNAL(timeout()), this, SLOT(textTimer()));

    QDateTime currentTime = QDateTime::currentDateTime();
    std::string now = qString2STLString(currentTime.toString());
    (*this) << "*** X4 started - ";
    (*this) << "System Time: ";
    (*this) << now;
    (*this) << "\n\n";

    //ui.splitter->setStretchFactor(0,1);
    //ui.splitter->setStretchFactor(1,0);

    startTimer(500);

	 (*this) << x4console.unhandledMessages;
}

ConsoleWidget::~ConsoleWidget() {
   try {
		// std::string LogFileName necessary cause of wayward QString
		std::string LogFileName = ui.LEConsoleLogFilename->text().toLatin1();
		std::ofstream out((getConfigurationString("BIN_PATH")+"/"+LogFileName).c_str(), ios_base::app);
      QDateTime currentTime = QDateTime::currentDateTime();
      std::string now = qString2STLString(currentTime.toString());
      out << "\n*** X4 closed -";
      out << "System Time: " << now << "\n\n";
   } catch (...) {
   }
   if (bufferMutex) delete bufferMutex;
}


// ------- user functions -------

void ConsoleWidget::operator<<(bool b) {
   if (b) (*this) << "true"; else (*this) << "false";
}

void ConsoleWidget::operator<<(int i) {
   (*this) << intToStr(i);
}

void ConsoleWidget::operator<<(unsigned int i) {
   (*this) << intToStr((int)i);
}

void ConsoleWidget::operator<<(X4::int64 i) {
   (*this) << intToStr(i);
}

void ConsoleWidget::operator<<(card64 i) {
   (*this) << intToStr((int64)i);
}

void ConsoleWidget::operator<<(long double d)
{
    std::string s;
    std::ostringstream os;
    os << d;
    s = os.str();
    (*this) << s;
}

void ConsoleWidget::operator<<(std::string s)
{
    bufferMutex->lock();

	textBuffer += s;
	if (loggedOutput) *loggedOutput += s;

    bufferMutex->unlock();

	if( textBuffer.size() > 2000 )
		flush();
}


void ConsoleWidget::operator<<(char* s)
{
    if( s != NULL) (*this) << std::string(s);
}

void ConsoleWidget::cr()
{
    bufferMutex->lock();

    textBuffer += "\n";
	if (loggedOutput) *loggedOutput += "\n";

    if (ui.CBLogConsole->isChecked())
    {
        // output to file
        try {
            // std::string LogFileName necessary cause of wayward QString
            std::string LogFileName = ui.LEConsoleLogFilename->text().toLatin1();
            std::ofstream out((getConfigurationString("BIN_PATH")+"/"+LogFileName).c_str(), ios_base::app);
            out << "\n";
        } catch (...) {
        }
    }

    bufferMutex->unlock();
}

void ConsoleWidget::flush()
{
    emit sigFlush();
}

void ConsoleWidget::warningMessage(std::string message)
{
    emit sigWarningMessage(message);
}


void ConsoleWidget::errorMessage(std::string message)
{
    emit sigErrorMessage(message);
}


static int errNum = 1;

static const QColor SIGNAL_COLOR(255,255,255);
static const QColor SIGNAL_BACKGROUND(30,40,128);


void ConsoleWidget::textTimer()
{
	timer->stop();
	QPalette pal = ui.labelErrors->palette();
	pal.setColor(QPalette::Background, stdBackground);
	pal.setColor(QPalette::WindowText, stdColor);
	ui.labelWarnings->setPalette(pal);
	ui.labelErrors->setPalette(pal);
}



void ConsoleWidget::displayEvent(const MetaClass *eventType, std::string name, const EventDisplayerEvent *event) {
   std::string log;
   const char *logPtr = NULL;
	if (ui.CBLogEvents->isChecked()) {
		log = qString2STLString(ui.LEEventsLogFilename->text());
		logPtr = log.c_str();
   }
	eventViewer->displayEvent(eventType, name, event, logPtr);
}


// ------- slots -------
void ConsoleWidget::sltWarningMessage(const std::string& message)
{
    ui.textEditWarnings->setFontWeight(QFont::Bold);
    ui.textEditWarnings->setTextColor(QColor(128,128,0));
    QString str;
    str += QString("(");
    ui.textEditWarnings->insertPlainText(str);
    ui.textEditWarnings->setTextColor(QColor(128,128,0));
    str = QString::number(errNum++);
    str += QString(") ");
    ui.textEditWarnings->insertPlainText(str);
    ui.textEditWarnings->setTextColor(QColor(30,40,128));
    str = QString("warning: ");
    ui.textEditWarnings->insertPlainText(str);
    ui.textEditWarnings->setFontWeight(QFont::Normal);
    ui.textEditWarnings->setTextColor(QColor(0,0,0));
    str = message.c_str();
    str += "\n";
    ui.textEditWarnings->insertPlainText(str);
    ui.tabWidget->setCurrentIndex(3);
    show();
    raise();
    QPalette pal = ui.labelWarnings->palette();
    stdColor = pal.windowText().color();
    stdBackground = pal.background().color();
    pal.setColor(QPalette::Background, SIGNAL_BACKGROUND);
    pal.setColor(QPalette::WindowText, SIGNAL_COLOR);
    ui.labelWarnings->setPalette(pal);
    timer->start();
}

void ConsoleWidget::sltErrorMessage(const std::string& message)
{
    ui.textEditErrors->setFontWeight(QFont::Bold);
    ui.textEditErrors->setTextColor(QColor(128,0,0));
    QString str;
    str += QString("(");
    ui.textEditErrors->insertPlainText(str);
    ui.textEditErrors->setTextColor(QColor(128,0,0));
    str = QString::number(errNum++);
    str += QString(") ");
    ui.textEditErrors->insertPlainText(str);
    ui.textEditErrors->setTextColor(QColor(30,40,128));
    str = QString("error: ");
    ui.textEditErrors->insertPlainText(str);
    ui.textEditErrors->setFontWeight(QFont::Normal);
    ui.textEditErrors->setTextColor(QColor(0,0,0));
    str = message.c_str();
    str += "\n";
    ui.textEditErrors->insertPlainText(str);
    ui.tabWidget->setCurrentIndex(3);
    show();
    raise();
    QPalette pal = ui.labelErrors->palette();
    stdColor = pal.windowText().color();
    stdBackground = pal.background().color();
    pal.setColor(QPalette::Background, SIGNAL_BACKGROUND);
    pal.setColor(QPalette::WindowText, SIGNAL_COLOR);
    ui.labelErrors->setPalette(pal);
    timer->start();
}


void ConsoleWidget::sltFlush()
{
	// michael: contributing to the good old flushing edit war: now you can set "always flush" programmatically by modifying the global
	// variable ALWAYS_PROCESS_EVENTS_ON_FLUSH. Default is false (previous behavior).
	flushText();
	if (ALWAYS_PROCESS_EVENTS_ON_FLUSH) {
		QCoreApplication::processEvents(QEventLoop::ExcludeUserInputEvents);
		if (QApplication::instance()->thread() == QThread::currentThread())
		{
			ui.TEConsoleOutput->repaint();
			ui.TEEventsOutput->repaint();
			QCoreApplication::processEvents(QEventLoop::ExcludeUserInputEvents);
		}
	}
}


// void ConsoleWidget::sltBTExecClicked() {
//    (*this) << "> ";
//    // std::string text necessary cause of wayward QString
// 	std::string text = ui.TEConsoleInput->toPlainText().toLatin1();
// 	(*this) << text;
//    cr();
// 	ui.TEConsoleInput->clear();
// }

void ConsoleWidget::sltCBLogConsoleStateChanged(int state) {
	ui.LEConsoleLogFilename->setEnabled(state==Qt::Checked);
}

void ConsoleWidget::sltCBLogEventsStateChanged(int state) {
	ui.LEEventsLogFilename->setEnabled(state==Qt::Checked);
}

static bool processingFlushEvent = false;

void X4::ConsoleWidget::timerEvent( QTimerEvent * e )
{
	if( !processingFlushEvent )
	{
		processingFlushEvent = true;
		flush();
		processingFlushEvent = false;
	}
}

void X4::ConsoleWidget::flushText()
{
    bufferMutex->lock();

	if( textBuffer.size() )
	{
		ui.TEConsoleOutput->moveCursor( QTextCursor::End );
		QString qs = textBuffer.c_str();
		qs.replace(QRegExp("a[\n]"), "\r\n");
		ui.TEConsoleOutput->insertPlainText(qs);

		if (ui.tabWidget->currentIndex()==1) {
			ui.TEEventsOutput->insertPlainText(qs);
   		ui.TEEventsOutput->moveCursor(QTextCursor::End, QTextCursor::MoveAnchor);
		}


		if (ui.CBLogConsole->isChecked()) {
			// output to file
			try {
				// std::string LogFileName necessary cause of wayward QString
				std::string LogFileName = ui.LEConsoleLogFilename->text().toLatin1();
				std::ofstream out((ApplicationDirectories::getLogDir()+std::string("/")+LogFileName).c_str(), ios_base::app);
				out << textBuffer;
			} catch (...) {
			}
		}

		textBuffer = "";
		ui.TEConsoleOutput->moveCursor(QTextCursor::End, QTextCursor::MoveAnchor);

	}
    
	bufferMutex->unlock();
}

void X4::ConsoleWidget::showPage(Page page)
{
   if (page >= NUM_PAGES)
   {
      page = LAST_PAGE;
      //QList<int> sizes = ui.splitter->sizes();
      //sizes[1] += sizes[0];
      //sizes[0] = 0;
      //ui.splitter->setSizes(sizes);
   }
   ui.tabWidget->setCurrentIndex(page);

}

#include "ConsoleWidget_moc.h"
