#include <QProgressDialog>
#include "InterfaceGlobals.h"
#include "UserSettings.h"
#include "BinaryObjectStreams.h"
#include "Configuration.h"

//---------------------------------------------------------------------------
namespace X4
{

/*
 * Callback class to bring progress dialog when loading files.
 * Class is derived from BinaryStreamCallback, so it 
 * can be used when loading files from binary streams
 **/
class X4INTERFACE_API LoadFileProgressDialog : public BinaryStreamCallback
{
protected:
   size_t fileSize;
   size_t readedBytes;
   QString fileName;
   QProgressDialog* dialog;
   QString fileSizeStr;
   QString fileNameLabel;
   QString unitSizeStr;
   int currentLoadedChunk;
   size_t chunkSize;
   UserSettings* settings;

public:

   LoadFileProgressDialog(UserSettings* userSettings = NULL)
   {
      dialog = NULL;
      readedBytes = 0;
      fileSize = 0;
      currentLoadedChunk = 0;
      chunkSize = size_t(1024) * size_t(1024);
      settings = NULL;
      if (userSettings) {
         settings = userSettings;
      } else {
         AttachedDataContainer *at = Configuration::getConfiguration();
         if (at) {
            settings = dynamic_cast<UserSettings*>(at->getData("USER_SETTINGS"));
         }
      }
   }

   virtual ~LoadFileProgressDialog()
   {
      if (dialog) delete dialog;
   }

   virtual void onOpenFile(const char* filename, size_t fileSize)
   {
      if (settings && !settings->getShowProgressDialogOnLoadSave()) return;

      this->fileName = QString(filename);
      QStringList parts = this->fileName.split(QString("/"), QString::SkipEmptyParts);
      if (!parts.isEmpty())
         this->fileName = parts.last();

      this->fileSize = fileSize;
      this->fileNameLabel = QString("Load file: ") + fileName + QString(" ");
      this->unitSizeStr = QString(" MB");
      this->fileSizeStr = QString::number(fileSize / chunkSize);

      if (!dialog)
      {
         dialog = new QProgressDialog(NULL, Qt::Popup | Qt::WindowStaysOnTopHint);
         dialog->setWindowIcon(QIcon(QString::fromUtf8(":/mainWindowIcons/Icons/images/Icons/MainIcon.png")));
         dialog->setWindowModality(Qt::WindowModal);
         dialog->setMinimum(0);
         dialog->setMaximum(static_cast<int>(fileSize / chunkSize));
         dialog->setLabelText(fileSizeStr + unitSizeStr);
         dialog->setWindowTitle(fileNameLabel);
         dialog->setCancelButton(NULL);
      }
   }

   virtual void onReadData(const void* buf, mpcard length)
   {
      if (settings && !settings->getShowProgressDialogOnLoadSave()) return;

      readedBytes += static_cast<size_t>(length);
      int chunk = static_cast<int>(readedBytes / chunkSize);
      if (dialog && chunk > currentLoadedChunk)
      {
         currentLoadedChunk = chunk;
         dialog->setValue(chunk);
         dialog->setLabelText(fileNameLabel + QString::number(chunk) + QString("/") + fileSizeStr + unitSizeStr);         
      }
   }

   virtual void onCloseFile()
   {
      if (settings && !settings->getShowProgressDialogOnLoadSave()) return;

      if (dialog)
      {
         delete dialog;
         dialog = NULL;
      }
   }

};


};
