//////////////////////////////////////////////////////////////////////////

#include "StdAfx.h"

#include "ModuleManagerWindow.h"

#include "ModuleManagerWindow_moc.h"
#include "ModuleList.h"
#include "Module.h"
#include "TagList.h"
#include "X4Assert.h"
#include "StringHelper.h"
#include "TagHeaderParser.h"
#include "ApplicationDirectories.h"

#include <QListView>
#include <QLabel>
#include <QListWidgetItem>
#include <QMessageBox>
#include <QTreeWidgetItem>
#include <QFileDialog>
#include <QtCore/qfile.h>
#include <QtCore/qtextstream.h>

//////////////////////////////////////////////////////////////////////////

using namespace X4;

//////////////////////////////////////////////////////////////////////////



ModuleManagerWindow::ModuleManagerWindow(QWidget *parent, Qt::WindowFlags f) : QDialog(parent,f) {
	ui.setupUi(this);
	ui.teInfoLoaded->setReadOnly(true);
	ui.teInfoAvailable->setReadOnly(true);
   loadStartupModulesLoadList();
   losTextUnchanged();
}

void ModuleManagerWindow::loadStartupModulesLoadList() {
   ui.pteLoadOnStartup->clear();
   QDir configDir(ApplicationDirectories::getConfigDir().c_str());
   if (configDir.exists(ApplicationDirectories::getStartupModulesListFilename().c_str())) {
      QFile inFile(ApplicationDirectories::getStartupModulesListFilenameAndPath().c_str());
      if( inFile.open(QIODevice::ReadOnly) ) {
         QTextStream inStream(&inFile);
         QString txt = inStream.readAll();
         ui.pteLoadOnStartup->setPlainText(txt);
      } else {
         outputError("Cannot load module startup load list", "File cannot be opened (file might be locked by another process?).");
      }
   }
}


ModuleManagerWindow::~ModuleManagerWindow() {
   for (size_t i=0; i<availableModules.size(); i++) {
      delete availableModules[i].tags;
   }
   availableModules.clear();
}




void ModuleManagerWindow::on_pbClose_clicked() 
{
   setVisible(false);
}

void ModuleManagerWindow::updateLoaded() 
{
   ui.lwModulesLoaded->clear();
   ui.twDependenciesLoaded->clear();
   ui.lbModuleIconLoaded->setText(" ");
   ui.lbModuleNameLoaded->setText("");
   ui.teInfoLoaded->clear();

   unsigned numMods = x4modules.getNumModules();
   for (unsigned i=0; i<numMods; i++) {
      QListWidgetItem *item = new QListWidgetItem();
      const Module *mod = x4modules.getModule(i);
      item->setText(mod->getName().c_str());
      std::string iconRes = mod->getTags().getValue("ICON");
      if (iconRes != "") {
         try {
            QIcon icon(iconRes.c_str());
            item->setIcon(icon);
         } catch (...) {
         }
      }
      ui.lwModulesLoaded->addItem(item);
   }
}

void ModuleManagerWindow::updateAvailable() 
{
   ui.lwModulesAvailable->clear();
   ui.twDependenciesAvailable->clear();
   ui.lbModuleIconAvailable->setText(" ");
   ui.lbModuleNameAvailable->setText("");
   ui.teInfoAvailable->clear();
   for (size_t i=0; i<availableModules.size(); i++) {
      delete availableModules[i].tags;
   }
   availableModules.clear();
   scanForAvailableModules();
   displayAvailableModules();
}

void ModuleManagerWindow::update() 
{
   updateLoaded();
   updateAvailable();
}



void ModuleManagerWindow::on_pbUpdateLoaded_clicked() 
{
   updateLoaded();
}

void ModuleManagerWindow::on_pbUnloadMod_clicked() 
{
   QMessageBox::information(this, "not implemented.", "Unloading modules is not yet implemented. You have to restart to remove modules :-)");
}

void ModuleManagerWindow::on_lwModulesLoaded_clicked(const QModelIndex & index)
{
   // --- commented, weired allocation bug
   //QModelIndexList sel = ui.lwModulesLoaded->selectionModel()->selectedIndexes().size();
   //if (sel.size() != 1) {
   //   updateDependenciesLoaded("");
   //   updateDescrLoaded("");
   //} else {
   //   std::string modName = qString2STLString(ui.lwModulesLoaded->item(sel[0].row())->text());
   //   updateDependenciesLoaded(modName);
   //   updateDescrLoaded(modName);
   //}

   int currentRow = ui.lwModulesLoaded->currentRow();
   if (currentRow >= 0 && currentRow < ui.lwModulesLoaded->count()) {
      std::string modName = qString2STLString(ui.lwModulesLoaded->item(currentRow)->text());
      updateDependenciesLoaded(modName);
      updateDescrLoaded(modName);
      ui.pbUnloadMod->setEnabled(true);
   } else {
      updateDependenciesLoaded("");
      updateDescrLoaded("");
      ui.pbUnloadMod->setEnabled(false);
   }
}

void ModuleManagerWindow::on_lwModulesAvailable_clicked(const QModelIndex & index) 
{

   // --- commented, weired allocation bug
   //QModelIndexList sel = ui.lwModulesAvailable->selectionModel()->selectedIndexes();
   //if (sel.size() != 1) {
   //   updateDependenciesAvailable("");
   //   updateDescrAvailable("");
   //} else {
   //   std::string modName = qString2STLString(ui.lwModulesLoaded->item(sel[0].row())->text());
   //   updateDependenciesAvailable(modName);
   //   updateDescrAvailable(modName);
   //}
   int currentRow = ui.lwModulesAvailable->currentRow();
   if (currentRow >= 0 && currentRow < ui.lwModulesAvailable->count()) {
      QString modString = ui.lwModulesAvailable->item(currentRow)->text();
      int si = modString.indexOf(" "); // check for annotations
      if (si >-1) {
         modString = modString.left(si);
         ui.pbLoadModule->setEnabled(false);
      } else {
         ui.pbLoadModule->setEnabled(true);
      }
      std::string modName = qString2STLString(modString);
      updateDependenciesAvailable(modName);
      updateDescrAvailable(modName);
   } else {
      updateDependenciesAvailable("");
      updateDescrAvailable("");
   }
}


QTreeWidgetItem *ModuleManagerWindow::createTreeWidgetItemLoaded(std::string modName, unsigned depth, bool &cyclic, bool &error) 
{
   if (depth >= 100) {
      cyclic = true;
      error = true;
      return NULL;
   } else {
      QTreeWidgetItem *item = new QTreeWidgetItem();
      const Module *mod = x4modules.getModule(modName);
      if (mod) {
         item->setText(0,modName.c_str());
         std::vector<std::string> dependencies = mod->getDependencies();
         for (size_t i=0; i<dependencies.size(); i++) {
            QTreeWidgetItem *child = createTreeWidgetItemLoaded(dependencies[i],depth+1,cyclic,error);
            if (child) {
               item->addChild(child);
               if (error) {
                  item->setForeground(0, QBrush(QColor(128,30,30)));
                  child->setForeground(0, QBrush(QColor(128,30,30)));
               }
            }
         }
      } else {
         std::string text = std::string("[error] module not found: ")+modName;
         item->setText(0, text.c_str());
         item->setForeground(0, QBrush(QColor(128,30,30)));
      }
      return item;
   }
}

QTreeWidgetItem *ModuleManagerWindow::createTreeWidgetItemAvailable(std::string modName, unsigned depth, bool &cyclic, bool &error) 
{
   if (depth >= 100) {
      cyclic = true;
      error = true;
      return NULL;
   } else {
      QTreeWidgetItem *item = new QTreeWidgetItem();
      int avModIndex = getAvModIndex(modName);
      if (avModIndex > -1) {
         item->setText(0,modName.c_str());
         TagList *tags = availableModules[avModIndex].tags;
         std::vector<std::string> dependencies = tags->getAllValues("DEPENDENCIES");
         for (size_t i=0; i<dependencies.size(); i++) {
            QTreeWidgetItem *child = createTreeWidgetItemLoaded(dependencies[i],depth+1,cyclic,error);
            if (child) {
               item->addChild(child);
               if (error) {
                  item->setForeground(0, QBrush(QColor(128,30,30)));
                  child->setForeground(0, QBrush(QColor(128,30,30)));
               }
            }
         }
      } else {
         std::string text = std::string("[error] module not found: ")+modName;
         item->setText(0, text.c_str());
         item->setForeground(0, QBrush(QColor(128,30,30)));
      }
      return item;
   }
}

void ModuleManagerWindow::updateDependenciesLoaded(std::string modName) 
{
   ui.twDependenciesLoaded->clear();
   const Module *mod = x4modules.getModule(modName);
   if (mod) {
      std::vector<std::string> dependencies = mod->getDependencies();
      for (size_t i=0; i<dependencies.size(); i++) {
         bool cyclic = false;
         bool error = false;
         QTreeWidgetItem *newItem = createTreeWidgetItemLoaded(dependencies[i], 1, cyclic, error);
         if (cyclic) {
            delete newItem;
            newItem = new QTreeWidgetItem();
            std::string text = std::string("[error] cyclic dependency upwards of \"")+dependencies[i]+std::string("\" (direct or indirect).");
            newItem->setText(0, text.c_str());
            newItem->setDisabled(true);
         }
         if (error) {
            newItem->setForeground(0, QBrush(QColor(128,30,30)));
         }
         ui.twDependenciesLoaded->addTopLevelItem(newItem);
      }
   } else {
      QTreeWidgetItem *newItem = new QTreeWidgetItem();
      std::string text = std::string("[error] module not found: ") + modName;
      newItem->setText(0, text.c_str());
      newItem->setForeground(0, QBrush(QColor(128,30,30)));
      ui.twDependenciesLoaded->addTopLevelItem(newItem);
   }
}

int ModuleManagerWindow::getAvModIndex(std::string name) {
   for (int i=0; i<(int)availableModules.size(); i++) {
      if (availableModules[i].name == name) return i;
   }
   return -1;
}

void ModuleManagerWindow::updateDependenciesAvailable(std::string modName) 
{
   ui.twDependenciesAvailable->clear();
   int avModIndex = getAvModIndex(modName);
   if (avModIndex > -1) {
      TagList *tags = availableModules[avModIndex].tags;
      std::vector<std::string> dependencies = tags->getAllValues("DEPENDENCIES");
      for (size_t i=0; i<dependencies.size(); i++) {
         bool cyclic = false;
         bool error = false;
         QTreeWidgetItem *newItem = createTreeWidgetItemAvailable(dependencies[i], 1, cyclic, error);
         if (cyclic) {
            delete newItem;
            newItem = new QTreeWidgetItem();
            std::string text = std::string("[error] cyclic dependency upwards of \"")+dependencies[i]+std::string("\" (direct or indirect).");
            newItem->setText(0, text.c_str());
            newItem->setDisabled(true);
         }
         if (error) {
            newItem->setForeground(0, QBrush(QColor(128,30,30)));
         }
         ui.twDependenciesAvailable->addTopLevelItem(newItem);
      }
   } else {
      QTreeWidgetItem *newItem = new QTreeWidgetItem();
      std::string text = std::string("[error] module not found: ") + modName;
      newItem->setText(0, text.c_str());
      newItem->setForeground(0, QBrush(QColor(128,30,30)));
      ui.twDependenciesAvailable->addTopLevelItem(newItem);
   }
}


void ModuleManagerWindow::updateDescrLoaded(std::string modName) 
{
   ui.teInfoLoaded->clear();
   const Module *mod = x4modules.getModule(modName);
   if (mod) {
      std::string iconResStr = mod->getIconResStr();
      ui.lbModuleIconLoaded->setText("");
      if (iconResStr == "") {
         ui.lbModuleIconLoaded->setPixmap(QPixmap(":/Icons/plugin_32x32.png"));
      } else {
         ui.lbModuleIconLoaded->setPixmap(QPixmap(iconResStr.c_str()));
      }
      ui.lbModuleNameLoaded->setText(mod->getName().c_str());   
      std::string descr = mod->createHTMLDescr(ui.cbTechnicalLoaded->isChecked());
      ui.teInfoLoaded->setHtml(descr.c_str());
   } else {
      ui.lbModuleIconLoaded->setText(" ");
      ui.lbModuleNameLoaded->setText("");   
      ui.teInfoLoaded->setText("");
   }
}

void ModuleManagerWindow::updateDescrAvailable(std::string modName) 
{
   ui.teInfoAvailable->clear();
   int avModIndex = getAvModIndex(modName);
   if (avModIndex > -1) {
      TagList *tags = availableModules[avModIndex].tags;
      ui.lbModuleNameAvailable->setText(availableModules[avModIndex].name.c_str());   
      std::string descr = Module::createHTMLDescrForMod(tags, ui.cbTechnicalAvailable->isChecked());
      ui.teInfoAvailable->setHtml(descr.c_str());
      ui.lbModuleIconAvailable->setPixmap(QPixmap(":/Icons/images/open_docs_32x32.png"));
   } else {
      ui.lbModuleIconAvailable->setText(" ");
      ui.lbModuleNameAvailable->setText("");   
      ui.teInfoAvailable->setText("");
   }
}

void ModuleManagerWindow::on_twDependenciesLoaded_itemDoubleClicked(QTreeWidgetItem *item, int column) 
{
   on_pbViewDependentLoaded_clicked();
}

void ModuleManagerWindow::on_twDependenciesAvailable_itemDoubleClicked(QTreeWidgetItem *item, int column) 
{
   on_pbViewDependentAvailable_clicked();
}

void ModuleManagerWindow::on_pbViewDependentAvailable_clicked() 
{
   QList<QTreeWidgetItem*> sel = ui.twDependenciesAvailable->selectedItems();
   if (sel.size() != 1) return;
   QString modName = sel[0]->text(0);
   if (modName.left(7) == "[error]") return;
   ui.lwModulesAvailable->selectionModel()->clear();
   for (int i=0; i<ui.lwModulesAvailable->count(); i++) {
      QString refText = ui.lwModulesAvailable->item(i)->text();
      int si = refText.indexOf(" "); // check for annotations
      if (si >-1) {
         refText = refText.left(si);
      }
      if (refText == QString(modName)) {
         ui.lwModulesAvailable->item(i)->setSelected(true);
         break;
      }
   }
   std::string modNameStd = qString2STLString(modName);
   updateDependenciesAvailable(modNameStd);
   updateDescrAvailable(modNameStd);
}

void ModuleManagerWindow::on_pbViewDependentLoaded_clicked() 
{
   QList<QTreeWidgetItem*> sel = ui.twDependenciesLoaded->selectedItems();
   if (sel.size() != 1) return;
   QString modName = sel[0]->text(0);
   if (modName.left(7) == "[error]") return;
   ui.lwModulesLoaded->selectionModel()->clear();
   for (int i=0; i<ui.lwModulesLoaded->count(); i++) {
      if (ui.lwModulesLoaded->item(i)->text() == QString(modName)) {
         ui.lwModulesLoaded->item(i)->setSelected(true);
         break;
      }
   }
   std::string modNameStd = qString2STLString(modName);
   updateDependenciesLoaded(modNameStd);
   updateDescrLoaded(modNameStd);
}

//void ModuleManagerWindow::on_pbTestModuleParser_clicked() 
//{
//   QString filename = QFileDialog::getOpenFileName(this, "select module tagging header file...", "", "*.tags.h");
//   TagHeaderParser parser;
//   TagList *test = parser.parseTagFile(qString2STLString(filename));
//   ui.pteOutput->clear();
//   ui.pteOutput->appendPlainText("error level:\n");
//   ui.pteOutput->appendPlainText(intToStr(parser.errorLevel).c_str());
//   ui.pteOutput->appendPlainText("error std::string:\n");
//   ui.pteOutput->appendPlainText(parser.errors.c_str());
//   ui.pteOutput->appendPlainText("\n\n");
//   if (test) {
//      ui.pteOutput->appendPlainText("result:\n");
//      ui.pteOutput->appendPlainText(test->toString().c_str());
//      delete test;
//   }
//}

void ModuleManagerWindow::on_pbUpdateAvailable_clicked() 
{
   updateAvailable();
}

void ModuleManagerWindow::scanForAvailableModules() 
{
   QDir dir(ApplicationDirectories::getBinDir().c_str());
   QStringList filters;
   filters << "*.tags.info";
   dir.setNameFilters(filters);   
   QStringList fileNames = dir.entryList();
   for (size_t i=0; i<availableModules.size(); i++) {
      delete availableModules[i].tags;
   }
   availableModules.clear();

   for (int i=0; i<fileNames.size(); i++) 
   {
      TagList *tags = new TagList();
      tags->readTextFromFile(qString2STLString(QDir::toNativeSeparators(dir.absolutePath()+QString("/"))+fileNames[i]));
      AvailableModule am;
      am.tags = tags;
      static const int extSize = QString(".tags.info").size();
      am.name = qString2STLString(fileNames[i].left(fileNames[i].size()-extSize));
      availableModules.push_back(am);
   }
}

void ModuleManagerWindow::displayAvailableModules() 
{
   for (size_t i=0; i<availableModules.size(); i++) {
      QListWidgetItem *item = new QListWidgetItem();
      QString ext;
      bool isApp = (availableModules[i].tags->getValue("MODULE_TYPE") == "APPLICATION");
      if (isApp) {
         item->setTextColor(QColor(128,128,228));
         ext += " (app)";
      }
      if (x4modules.findModule(availableModules[i].name) != -1) {
         if (!isApp) item->setTextColor(QColor(128,128,128));
         ext += " (loaded)";
      }
      item->setText(QString(availableModules[i].name.c_str())+ext);
      ui.lwModulesAvailable->addItem(item);
   }
}


void ModuleManagerWindow::on_cbTechnicalLoaded_clicked() 
{
   updateDescrLoaded(qString2STLString(ui.lbModuleNameLoaded->text()));
}

void ModuleManagerWindow::on_cbTechnicalAvailable_clicked() 
{
   updateDescrAvailable(qString2STLString(ui.lbModuleNameAvailable->text()));
}

void ModuleManagerWindow::on_pbLoadModule_clicked() 
{
   QModelIndexList *sel = new QModelIndexList;
   *sel = ui.lwModulesAvailable->selectionModel()->selectedIndexes();
   for (int i=0; i<(*sel).size(); i++) {
      QString refText = ui.lwModulesAvailable->item((*sel)[i].row())->text();
      int si = refText.indexOf(" "); // check for annotations
      if (si >-1) {
         refText = refText.left(si);
      }
      if (x4modules.findModule(qString2STLString(refText)) != -1) {
         outputError("module already loaded", qString2STLString(refText));
         return;
      }
      refText += ".x4module.dll";
      std::string err;
      if (!x4modules.tryLoadModule(qString2STLString(refText), err)) {
         outputError("load module failed", err);
      } else {
         outputMsg("module loaded successfully", qString2STLString(refText));
      }
   }  

   // this causes a weired bug (QT bug?)
   // delete sel;

   update();
}

void ModuleManagerWindow::outputError(std::string caption, std::string errMsg) 
{
   static const std::string capStart = "<p><b><font color=\"#903030\"> ERROR [";
   static const std::string capEnd = "] </b></font>";
   static const std::string msgStart = "<font color=\"#B0B0B0\">";
   static const std::string msgEnd = "</font></p>";

   ui.teOutput->append((capStart+caption+capEnd+msgStart+errMsg+msgEnd).c_str());
   ui.tabWidget->setCurrentIndex(3);
}

void ModuleManagerWindow::outputMsg(std::string caption, std::string msg) 
{
   static const std::string capStart = "<p><b><font color=\"#505050\"> MSG [";
   static const std::string capEnd = "] </b></font>";
   static const std::string msgStart = "<font color=\"#B0B0B0\">";
   static const std::string msgEnd = "</font></p>";

   ui.teOutput->append((capStart+caption+capEnd+msgStart+msg+msgEnd).c_str());
}

void ModuleManagerWindow::on_pbLOSUpdate_clicked() 
{
   if (changedLOS) {
      if (QMessageBox::question(this, "Confirm reload...", 
         "You have edited the startup module load list. Do you want to dismiss your changes and reload the file from disk?", 
         QMessageBox::Yes | QMessageBox::No, QMessageBox::Yes )
         == QMessageBox::No) 
      {
         return;
      }
   }
   loadStartupModulesLoadList();
   losTextUnchanged();
}



void ModuleManagerWindow::on_pteLoadOnStartup_textChanged() 
{
   losTextChanged();
}

void ModuleManagerWindow::losTextChanged()
{
   changedLOS = true;
   ui.lbLoadOnStartup->setText("Load on startup: (changed)");
}

void ModuleManagerWindow::losTextUnchanged()
{
   changedLOS = false;
   ui.lbLoadOnStartup->setText("Load on startup:");
}

void ModuleManagerWindow::on_pbSaveLOS_clicked()
{
   QFile outFile(ApplicationDirectories::getStartupModulesListFilenameAndPath().c_str());
   if( outFile.open(QIODevice::WriteOnly) ) {
      QTextStream outStream(&outFile);
      outStream << ui.pteLoadOnStartup->toPlainText();
      losTextUnchanged();
   } else {
      outputError("Cannot save module startup load list", "File cannot be opened for writing (file might be locked by another process?).");
   }
}

void ModuleManagerWindow::on_pbLOSall_clicked() 
{
   if (changedLOS) {
      if (QMessageBox::question(this, "Confirm revert to default...", 
         "You have edited the startup module load list. Do you want to dismiss your changes and set the text to default value?", 
         QMessageBox::Yes | QMessageBox::No, QMessageBox::Yes )
         == QMessageBox::No) 
      {
         return;
      }
   }
   ui.pteLoadOnStartup->setPlainText("*.x4module.dll\n");
   losTextChanged();
}

//void ModuleManagerWindow::hideEvent(QHideEvent *event) {
//   static bool inHereAlready = false;
//   if (!inHereAlready) {
//      if (changedLOS) {
//         show();
//         ui.tabWidget->setCurrentIndex(2);
//         QMessageBox::StandardButton answer = QMessageBox::question(
//            this,
//            "Text changed...", 
//            "You have edited the startup module load list. Do you want to save your changes before closing this dialog?", 
//            QMessageBox::Yes | QMessageBox::No | QMessageBox::Cancel,
//            QMessageBox::Cancel 
//         );
//        
//         if (answer == QMessageBox::No) {
//      	   event->setAccepted(true);
//            inHereAlready = true;
//            hide();
//            inHereAlready = false;
//         } else if (answer == QMessageBox::Yes) {
//            on_pbSaveLOS_clicked();
//      	   event->setAccepted(true);
//            inHereAlready = true;
//            hide();
//            inHereAlready = false;
//         } else {
//      	   event->setAccepted(true);
//         }
//      }
//   }
//}


std::vector<std::string> ModuleManagerWindow::getStartupModulesList() {
   std::vector<std::string> result;
   QDir configDir(ApplicationDirectories::getConfigDir().c_str());
   bool loadAll = true;
   if (configDir.exists(ApplicationDirectories::getStartupModulesListFilename().c_str())) {
      QFile inFile(ApplicationDirectories::getStartupModulesListFilenameAndPath().c_str());
      if( inFile.open(QIODevice::ReadOnly) ) {
         QTextStream inStream(&inFile);
         int lineNo = 0;
         loadAll = false;
         while (!inStream.atEnd()) {
            QString line = inStream.readLine().trimmed();
            lineNo++;
            if (line.size() == 0) continue;

            static const QString ext = ".x4module.dll";

            if (line.toLower() == "*.x4module.dll") {
               result.clear();
               loadAll=true;
               break;
            } else if (line.right(ext.size()).toLower() == ext) {
               result.push_back(qString2STLString(line));
            } else {
               outputMsg("startup modules", std::string("Ignoring line ")+intToStr(lineNo)+ std::string(" due to extension missmatch. Content: ")+qString2STLString(line));
            }
         }
         if (result.empty() && !loadAll) {
            outputMsg("startup modules","module list found but file is empty (or no valid entries found, respectively)");
         }
      } else {
         outputError("Cannot load module startup load list", "File cannot be opened (file might be locked by another process?). Loading all modules instead (*.x4module.dll).");
         loadAll = true;
      }
   } else {
      outputMsg("no startup module load list", std::string("file \"")+ApplicationDirectories::getStartupModulesListFilename()+std::string("\" not found. Loading all modules (*.x4module.dll)."));
      loadAll = true;
   }
   
   if (loadAll) {
      result.clear();
      QDir binDir(ApplicationDirectories::getBinDir().c_str());
      QStringList filters;
      filters << "*.x4module.dll";
      binDir.setNameFilters(filters);   
      QStringList fileNames = binDir.entryList();
      for (int i=0; i<fileNames.size(); i++) {
         result.push_back(qString2STLString(fileNames[i]));
      }
   }
   return result;
}


void ModuleManagerWindow::loadStartupModules() 
{
   std::vector<std::string> modules = getStartupModulesList();
   std::vector<std::string> errors;

   if (!x4modules.tryLoadModules(modules, errors)) {
      outputError("startup modules error", "something went wrong - see below for details.");
      for (size_t i=0; i<errors.size(); i++) {
         if (errors[i] != "") {
            outputError(modules[i],errors[i]);
         }
      }
   } else {
      outputMsg("startup modules", "list of startup modules loaded successfully.");
   }
}
