//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "ProgressWindow.h"
#include "X4Exception.h"
#include "DebugOutput.h"
#include "InterfaceGlobals.h"
#include "Configuration.h"
#include "UserSettings.h"
#include <ctime>
//---------------------------------------------------------------------------
#include "ProgressWindowWidget.h"
#include "ProgressFrameWidget.h"
#include "Object.h"
#include "BooleanClassProperty.h"
//---------------------------------------------------------------------------
using namespace X4;

//---------------------------------------------------------------------------
ProgressWindow* X4::progressWindow = NULL;

//---------------------------------------------------------------------------
IMPLEMENT_X4_CLASS( ProgressWindow, 0 )
{
   BEGIN_CLASS_INIT( ProgressWindow );

   ADD_BOOLEAN_PROP(_enabled, 0);
	ADD_BOOLEAN_PROP(_throwOnCancel, 0);
}


/****************************************************************************
 * ProgressWindow - constructor
 ****************************************************************************/
ProgressWindow::ProgressWindow()
{
    initConstructor(new ProgressWindowWidget(InterfaceGlobals::getMainWindow()));
}

ProgressWindow::ProgressWindow(ProgressWidget* widget)
{
    initConstructor(widget);
}

void ProgressWindow::initConstructor(ProgressWidget* widget)
{
    progressWindowWidget = widget;

    _mutex = new QMutex(QMutex::Recursive);	

    // connect singals with progress widget slots
    QObject::connect(this, SIGNAL(showForm()), progressWindowWidget, SLOT(showForm()));
    QObject::connect(this, SIGNAL(setToolProgress(card32)), progressWindowWidget, SLOT(setToolProgress(card32)));
    QObject::connect(this, SIGNAL(setOverallProgress(card32)), progressWindowWidget, SLOT(setOverallProgress(card32)));
    QObject::connect(this, SIGNAL(setToolProgressAvailable(bool)), progressWindowWidget, SLOT(setToolProgressAvailable(bool)));
    QObject::connect(this, SIGNAL(setOverallProgressAvailable(bool)), progressWindowWidget, SLOT(setOverallProgressAvailable(bool)));
    QObject::connect(this, SIGNAL(setToolName(const std::string& )), progressWindowWidget, SLOT(setToolName(const std::string& )));
    QObject::connect(this, SIGNAL(setToolMessage(const std::string& )), progressWindowWidget, SLOT(setToolMessage(const std::string& )));
    QObject::connect(this, SIGNAL(drawUpdates()), progressWindowWidget, SLOT(drawUpdates()));
    QObject::connect(this, SIGNAL(closeForm()), progressWindowWidget, SLOT(closeForm()));
	QObject::connect(this, SIGNAL(setToolStartTime(time_t)), progressWindowWidget, SLOT(setToolStartTime(time_t)));
	QObject::connect(this, SIGNAL(setToolEstimatedTime(card64,time_t)), progressWindowWidget, SLOT(setToolEstimatedTime(card64,time_t)));

    QObject::connect(progressWindowWidget, SIGNAL(progressCanceled()), this, SLOT(cancelProcess()));
    
    cancelCallback = NULL;
    _enabled = true;
	_throwOnCancel = false;
	overallProgressDone = 0;
	isVisible = false;
	isAborted = false;
	startTime = 0;
	lastProgressTime = 0;
	toolProgressAvailable = true;
	overallProgressAvailable = true;
}

ProgressWindow::~ProgressWindow()
{ 
//  delete ProgressWindowForm;
//	delete progressWindowWidget;
    if (_mutex) delete _mutex;
    //if (progressWindowWidget) delete progressWindowWidget;
}

/****************************************************************************/
// Cancle current running process
/****************************************************************************/
void ProgressWindow::cancelProcess()
{
#if 1
   if (!_enabled) return;

    std::string currentStepName = "";
    //if (!weightStack.empty()) currentStepName = weightStack.top().begin()->stepName;
	if (!weightStack.empty())
		currentStepName = weightStack.front().stepName;

	if (cancelCallback != NULL) 
        (*cancelCallback)(currentStepName, this);
	isAborted = true;
	popSteps();
#endif
}

#if 0
/****************************************************************************/
//  ProgressWindow - pushSteps
//  Pushes a list of steps onto the operations stack
/// If the stack was empty, the progress window will be displayed. If not,
/// the step list will be pushed onto the stack and the total progress for
/// this part of the stack will be the weight of the active step in the level
/// below.
/****************************************************************************/
void ProgressWindow::pushSteps(WeightList steps)
{
   if (!_enabled) return;

  if (steps.empty()) {
    #ifdef _DEBUG
    warning("ProgressWindow: Pushed empty step list");
    #endif
    return;
  }

  _mutex->lock();

  if (weightStack.empty()) {
    // First call to the progress window; initialize it for first display
    startTime = timer.getCurrentValue();
    lastProgressTime = 0;
    emit setOverallProgress(0);
    emit setToolMessage("");
    if (steps.size() == 1 && !steps.begin()->progressAvailable)
      overallProgressAvailable = false;
    else
      overallProgressAvailable = true;
    emit setOverallProgressAvailable(overallProgressAvailable);

    isVisible = false;
    isAborted = false;
    while (!percentageDoneStack.empty()) percentageDoneStack.pop();

    // Clear keyboard buffer for ESC and "q" key
    while ((GetAsyncKeyState(VK_ESCAPE) & 1) == 1);
    while ((GetAsyncKeyState(0x51) & 1) == 1);
  }

  // Normalize weights
  float32 totalWeight = 0.0f;
  for (WeightList::iterator it = steps.begin(); it != steps.end(); ++it)
    totalWeight += it->weight;
  if (!weightStack.empty())
    totalWeight /= weightStack.top().begin()->weight;
  for (WeightList::iterator it = steps.begin(); it != steps.end(); ++it)
    it->weight /= totalWeight;

  weightStack.push(steps);

  if (!updateIntervalStack.empty())
    updateIntervalStack.push(updateIntervalStack.top());
  else
    updateIntervalStack.push(90);

  if (percentageDoneStack.empty()) percentageDoneStack.push(0.0f);
    else percentageDoneStack.push(percentageDoneStack.top());

  initForm();

  _mutex->unlock();
}
#endif 

/****************************************************************************/
//  ProgressWindow - pushStep
//  Pushes a single step onto the operations stack
/// Works just like pushSteps, but instead of pushing a list of steps onto
/// the stack, it just pushes a single step (for convenience).
/****************************************************************************/
void ProgressWindow::pushStep(bool progressAvailable, std::string stepName, float stepWeight)
{
	_mutex->lock();

	if (!_enabled) {
		_mutex->unlock();
		return;
	}

	WeightElem elem;
	elem.stepName = stepName;
	elem.weight = stepWeight;
	elem.progressAvailable = progressAvailable;
	elem.startTime = 0;
	elem.normWeight = stepWeight;

	//WeightList wList;
	//wList.push_back(elem);
	//pushSteps(wList);

	// if nothing on the stack then init whole stuff
	if (weightStack.empty())
	{
		// First call to the progress window; initialize it for first display
		startTime = timer.getCurrentValue();
		lastProgressTime = 0;
		overallProgressAvailable = true; // TODO: do something to decide whenever to show overall progress or not
		overallProgressDone = 0;

		emit setOverallProgress(0);
		emit setOverallProgressAvailable(overallProgressAvailable);

		isVisible = false;
		isAborted = false;

		// Clear keyboard buffer for ESC and "q" key
		while ((GetAsyncKeyState(VK_ESCAPE) & 1) == 1);
		while ((GetAsyncKeyState(0x51) & 1) == 1);
	}

	// add new element on the stack
	weightStack.push_back(elem);

	// Compute new normalized weights
	float32 totalWeight = 0.0f;
	for (WeightList::iterator it = weightStack.begin(); it != weightStack.end(); it++)
		totalWeight += it->weight;
	for (WeightList::iterator it = weightStack.begin(); it != weightStack.end(); it++)
		it->normWeight = it->weight / totalWeight;

	initForm();

	_mutex->unlock();
}

#if 1
/****************************************************************************/
//  ProgressWindow - nextStep
//  Moves to the next step in the list
/// For the overall progress, it is assumed that the step before was
/// completed, no matter what the last progress value was. If the whole
/// process is done, nextStep can not be called. Use popSteps instead.
/// \param aborted Contains true if the progress was aborted at some point.
///                In that case, the tool should not continue with what it's
///                doing.
/****************************************************************************/
bool ProgressWindow::popStep(bool* aborted)
{
	_mutex->lock();

	if (!_enabled) {
		_mutex->unlock();
		return true;
	}

	if (aborted) *aborted = isAborted;
	if (isAborted) {
		_mutex->unlock();
		return isAborted;
	}
	if (weightStack.empty()) {
		_mutex->unlock();
		return false;
	}

	// tools progress is 0 and overall progress increases by the amount of tools progress
	emit setToolProgress(0);
	if (overallProgressAvailable)
		overallProgressDone += weightStack.front().normWeight;

	weightStack.pop_front();

	// if still steps left, then 
	if (!weightStack.empty())
	{
		initForm();

		// do not update last progress time, in order to force appearance of progress bar
		// even if some small steps where computed very fast (TODO: remove hardcoded of 1000)
		if (lastProgressTime != 0 && timer.getDeltaValue(lastProgressTime) < 1000)
		{
			_mutex->unlock();
			return false;
		}

		lastProgressTime = timer.getCurrentValue();
		emit drawUpdates();
		//QCoreApplication::processEvents();
	}else
		emit closeForm();

	_mutex->unlock();

	return false;
}
#endif 

/****************************************************************************/
//  ProgressWindow - popSteps
//  Pops all steps out of the list
/****************************************************************************/
bool ProgressWindow::popSteps(bool* aborted)
{
	_mutex->lock();

	if (!_enabled) {
		_mutex->unlock();
		return false;
	}

	while (!weightStack.empty())
	{
		if (popStep(aborted))
		{
			weightStack.clear();
			isAborted = true;
			break;
		}
	}

	emit setToolMessage("");
	
	emit closeForm();
	_mutex->unlock();

	return isAborted;

#if 0
   if (!_enabled) return;

  aborted = false;

  if (isAborted) {
	  emit closeForm();
    aborted = true;
    return;
  }

  if (weightStack.empty()) {
    #ifdef _DEBUG
    warning("ProgressWindow: Trying to pop empty stack");
    #endif
    return;
  }

  _mutex->lock();

  while (!weightStack.top().empty()) {
    weightStack.top().pop_front();
  }
  weightStack.pop();
  percentageDoneStack.pop();
  updateIntervalStack.pop();

  if (!weightStack.empty()) {
    initForm();
    if (lastProgressTime != 0 && timer.getDeltaValue(lastProgressTime) < updateIntervalStack.top())
	{
		_mutex->unlock();
		return;
	}
    lastProgressTime = timer.getCurrentValue();
    emit drawUpdates();
	// Trigger processing of events
    //QCoreApplication::processEvents();
  } else {
    emit closeForm();
  }

  _mutex->unlock();
#endif 
}

/****************************************************************************/
//  ProgressWindow - initForm [PRIVATE]
//  Updates the form of the progress window
/// The form of the progress window will be updated with the values from the
/// top of the stack, and everything will be initialized for the next step.
/****************************************************************************/
void ProgressWindow::initForm()
{
   _mutex->lock();

   if (!_enabled || weightStack.empty()) {
	   _mutex->unlock();
	   return;
   }

   const WeightElem& firstStep = *weightStack.begin();

   toolProgressAvailable = firstStep.progressAvailable;
   emit setToolName(firstStep.stepName);
   emit setToolProgressAvailable(toolProgressAvailable);
   emit setToolProgress(0);
   emit setOverallProgressAvailable(overallProgressAvailable);
   if( overallProgressAvailable )
	   emit setOverallProgress( overallProgressDone * 100.0f );

   _mutex->unlock();

#if 0
  _mutex->lock();

  WeightElem* firstStep = &(*weightStack.top().begin());

  toolProgressAvailable = firstStep->progressAvailable;
  emit setToolName(firstStep->stepName);
  emit setToolProgressAvailable(toolProgressAvailable);
  emit setToolProgress(0);
  emit setOverallProgressAvailable(overallProgressAvailable);
  if( overallProgressAvailable )
	emit setOverallProgress( percentageDoneStack.top() );

  _mutex->unlock();
#endif 
}

/****************************************************************************/
//  ProgressWindow - showMessage
//  Displays the tool message given
/****************************************************************************/
void ProgressWindow::showMessage(std::string msg)
{
   if (!_enabled) return;

  _mutex->lock();

  emit setToolMessage(msg);
  emit drawUpdates();
  // Trigger processing of events
  //QCoreApplication::processEvents();

  _mutex->unlock();	
}

void ProgressWindow::showWindow()
{
	_mutex->lock();

	if (!isVisible)
	{
		isVisible = true;
		emit showForm();
	}

	_mutex->unlock();
}

void ProgressWindow::hideWindow()
{
	_mutex->lock();

	if (isVisible)
	{
		isVisible = false;
		emit closeForm();
	}

	_mutex->unlock();
}

/****************************************************************************/
//  ProgressWindow - progress
//  Updates the progress window and returns false if ESC was pressed.
/// This method has to be called frequently during the processing in order
/// to update the progress form and check whether the processing was
/// cancelled. If the user requested to cancel the process, false will be
/// returned, otherwise true.
/****************************************************************************/
bool ProgressWindow::progressf(float32 position)
{
	_mutex->lock();

	if (!_enabled) 
	{
		_mutex->unlock();
		return true;
	}

	if (isAborted)
	{
		if (_throwOnCancel)
		{
			std::string currentStepName = "";
			if (!weightStack.empty()) currentStepName = weightStack.begin()->stepName;
			_mutex->unlock();
			throw Canceled(currentStepName);
		}
		_mutex->unlock();
		return false;
	}

	if (weightStack.empty()) {
		_mutex->unlock(); 
		return false;
	}

	if (!isVisible) {
		// Show window if it isn't visible already
		if (timer.getDeltaValue(startTime) > 1000) {
			isVisible = true;
			emit showForm();
		}
	}

	// update internal tool start time if progress is just started
	if (weightStack.begin()->startTime == 0)
	{
		weightStack.begin()->startTime = time(NULL);
		emit setToolStartTime(weightStack.begin()->startTime);
		emit setToolEstimatedTime(0, time(NULL));
	}

	bool returnValue = true;

	if (isVisible)
	{
		// compute how much time was spend on this progress and update the time counter
		//if (position != 0)
		{
			double estimated = (double(timer.getDeltaValue(startTime)) / double(position)) / 1000;
			emit setToolEstimatedTime(card64(estimated), time(NULL));
		}

		// Dont update progress window if last call was less then <updateInterval> ms ago
		if (lastProgressTime != 0 && timer.getDeltaValue(lastProgressTime) < 1000)
		{
			_mutex->unlock();
			return true;
		}
		lastProgressTime = timer.getCurrentValue();

		if (toolProgressAvailable)
		{
			emit setToolProgress(card32(position * 100.0f));
			emit setToolStartTime(weightStack.begin()->startTime);
			if (overallProgressAvailable)
				emit setOverallProgress((overallProgressDone + weightStack.begin()->normWeight * float32(position)) * 100.0f);
		}

		emit drawUpdates();

		// Trigger processing of events
		//if (dynamic_cast<UserSettings*>(Configuration::getConfiguration()->getData( "USER_SETTINGS" ))->getForceIdleProcessing())  // comment out by CL @2014/02/16 
		//{
			InterfaceGlobals::getApplication()->processEvents(QEventLoop::AllEvents, 500);
		//}



		// key pressed, hence abort
		if ((GetAsyncKeyState(VK_ESCAPE) & 1) == 1 && (GetAsyncKeyState(0x51) & 1) == 1)
		{
			std::string currentStepName;
			if (!weightStack.empty()) currentStepName = weightStack.begin()->stepName;

			popSteps();
			isAborted = true;
			returnValue = false;

			if (cancelCallback)
			{
				(*cancelCallback)(currentStepName, this);
				_mutex->unlock();
				return returnValue;
			}
		} else {
			returnValue = true;
		}
	} 

	_mutex->unlock(); 
	return returnValue;


#if 0
   if (!_enabled) return true;

    if (isAborted)
	{
		if (_throwOnCancel)
		{
			std::string currentStepName = "";
			if (!weightStack.empty()) currentStepName = weightStack.top().begin()->stepName;
			throw Canceled(currentStepName);
		}
		return false;
	}
    if (weightStack.empty()) return false;

    _mutex->lock();

    if (!isVisible) {
        // Show window if it isn't visible already
        if (timer.getDeltaValue(startTime) > 1000) {
          isVisible = true;
          emit showForm();
        }
    }


	// update internal tool start time if progress is just started
	if (weightStack.top().begin()->startTime == 0)
	{
		weightStack.top().begin()->startTime = time(NULL);
		emit setToolStartTime(weightStack.top().begin()->startTime);
		emit setToolEstimatedTime(0, time(NULL));
	}

    bool returnValue = true;

    if (isVisible)
    {
		// compute how much time was spend on this progress and update the time counter
		//if (position != 0)
		{
			double estimated = (double(timer.getDeltaValue(startTime)) / double(position)) / 1000;
			emit setToolEstimatedTime(card64(estimated), time(NULL));
		}

        // Dont update progress window if last call was less then <updateInterval> ms ago
        if (lastProgressTime != 0 && timer.getDeltaValue(lastProgressTime) < updateIntervalStack.top())
		{
			_mutex->unlock();
			return true;
		}
        lastProgressTime = timer.getCurrentValue();

        if (toolProgressAvailable)
		{
          emit setToolProgress(card32(position * 100.0f));
		  emit setToolStartTime(weightStack.top().begin()->startTime);
          if (overallProgressAvailable)
            emit setOverallProgress(percentageDoneStack.top() + weightStack.top().begin()->weight * float32(position * 100.0f));
        }

         emit drawUpdates();

		 // Trigger processing of events
		 if (dynamic_cast<UserSettings*>(Configuration::getConfiguration()->getData( "USER_SETTINGS" ))->getForceIdleProcessing())
		 {
			 InterfaceGlobals::getApplication()->processEvents(QEventLoop::AllEvents, 500);
		 }


        // key pressed, hence abort
        if ((GetAsyncKeyState(VK_ESCAPE) & 1) == 1 && (GetAsyncKeyState(0x51) & 1) == 1)
        {
            std::string currentStepName;
            if (!weightStack.empty()) currentStepName = weightStack.top().begin()->stepName;

          bool dummy;
          while (!weightStack.empty()) popSteps(dummy);
          isAborted = true;
          returnValue = false;

          if (cancelCallback)
          {
              (*cancelCallback)(currentStepName, this);
			  _mutex->unlock();
              return returnValue;
          }
        } else {
          returnValue = true;
        }
    } 

    _mutex->unlock(); 
    return returnValue;
#endif 
	return false;
}

/****************************************************************************/
//  ProgressWindow - setUpdateInterval
//  Sets the minimum interval between GUI-updates for this level in the stack
/// This method sets the update interval for the active level in the stack
/// (and all sublevels, if they don't call it again).
/****************************************************************************/
void ProgressWindow::setUpdateInterval(card32 value)
{
   if (!_enabled) return;
#if 0
  if (updateIntervalStack.empty()) {
    #ifdef _DEBUG
    warning("ProgressWindow: setUpdateInterval() called with empty stack");
    #endif
    return;
  }

  _mutex->lock();

  updateIntervalStack.pop();
  updateIntervalStack.push(value);

  _mutex->unlock();
#endif 

}

#include "ProgressWindow_moc.h"
