//---------------------------------------------------------------------------
#ifndef ProgressWindowH
#define ProgressWindowH
//---------------------------------------------------------------------------
#include "x4interface.h"
//---------------------------------------------------------------------------
#include "Object.h"
#include "Timer.h"
#include <list>
#include <stack>
#include <string>
#include <QtCore/QMutex.h>
#include <QtCore/QObject.h>
#include "ProgressWidget.h"

//---------------------------------------------------------------------------


namespace X4 {

/// Class for the global progress window of X4
class X4INTERFACE_API ProgressWindow : public QObject, public Object
{
    Q_OBJECT
    X4_CLASS(ProgressWindow);

	private:
		/// One element represeneted by the progress bar
		struct WeightElem
		{
			WeightElem(bool progressAvailable = true, float32 weight = 1.0f, std::string stepName = "", time_t startTime = 0)
			{
				this->stepName = stepName;
				this->weight = weight;
				this->progressAvailable = progressAvailable;
				this->startTime = startTime;
			}
			std::string stepName;
			float32 weight;
			float32 normWeight;
			bool progressAvailable;
			time_t startTime;
		};
		typedef std::list<WeightElem> WeightList;

		/// Pushes a list of steps onto the operations stack
		//void pushSteps(WeightList steps);

	public:

		/**
		* Subclass which can be used as a callback to be executed when the progressbar receive a cancel operation.
		* If no callback is set, then other mechanism can be used to detect canceling
		**/
        class CancelProgressCallback
        {
        public:
            CancelProgressCallback(){}
            virtual void operator()(const std::string& stepName, ProgressWindow* progressWindow) = 0;
        };

		/**
		* Error class which will be thrown if progress was canceled. If error throwing is disabled, then other mechanism
		* can be used to detect cancel
		**/
		class Canceled
		{
		private:
			const std::string stepName;
		public:
			Canceled(const std::string& stepName) : stepName(stepName){}
			const std::string& getStepName() { return stepName; }
		};

        /**
		* Create default progress window (which use global progress widget)
		**/
        ProgressWindow();

        /** 
		* Create progress window with specified progress widget (can be used also locally)
		**/
        ProgressWindow(ProgressWidget*);

		~ProgressWindow();

        /**
		* Pushes a single step onto the operations stack.
		**/
        void pushStep(bool progressAvailable = true, std::string stepName = "", float stepWeight = 1.0f);
        
		/** 
		* Moves to the next step in the list, so pop out currently computed. Returns true if abort was performed, false otherwise.
		**/
        bool popStep(bool* aborted = NULL);

        /**
		* Pops all used steps out of the list. Same as popStep but executed for all steps. 
		**/
        bool popSteps(bool* aborted = NULL);

		/// check whenever the progress was aborted
        inline bool getIsAborted() {return isAborted;}

        /// Displays the tool message given
        void showMessage(std::string msg);

        /// Sets the minimum interval between GUI-updates for this level in the stack
        void setUpdateInterval(card32 value);

		/// Set if we should throw an error when progress is canceled (this can be used to be catched without using callbacks)
		void setThrowErrorOnCancel(bool en) { _throwOnCancel = en; }

        /// set callback, to be called hwen progress is canceled
        inline void setCancelCallback(CancelProgressCallback* callback) { cancelCallback = callback; }

        /// get currently associated callback
        inline CancelProgressCallback* getCancelCallback() { return cancelCallback; }
        
        /// get used progress windget
        inline ProgressWidget* getWidget() { return progressWindowWidget; }

        /// disable/enable progress window, so that it just do nothing on any call
        inline void setEnableFlag(bool b) { _enabled = b; }
        inline bool getEnableFlag() { return _enabled; }

		/// Lock the ProgressWindow mutex. This allows for atomic sequences of operations.
		inline void lockMutex() { _mutex->lock(); }

		/// Unlock the ProgressWindow mutex. This allows for atomic sequences of operations.
		inline void unlockMutex() { _mutex->unlock(); }

  private:
		void initConstructor(ProgressWidget* widget);

		bool _enabled;
		bool _throwOnCancel;

        bool isVisible, isAborted;
        card32 startTime, lastProgressTime;
        Timer timer;
		float32 overallProgressDone;

        bool toolProgressAvailable, overallProgressAvailable;

        CancelProgressCallback* cancelCallback;

        /// Percentage for the steps that are fully done
        WeightList weightStack;
        //std::stack<float32> percentageDoneStack;
        //std::stack<card32> updateIntervalStack;

        QMutex* _mutex;

        // widget used for showing progress
        ProgressWidget* progressWindowWidget;

		/// Updates the form of the progress window
		void initForm();

    public slots:
        void cancelProcess();
		void showWindow();
		void hideWindow();
		bool progress(card32 position) { return progressf(float32(position) / 100.0f); }
		bool progressf(float32 position) ;

    signals:
        void showForm();
        void setToolProgress(card32 position);
        void setOverallProgress(card32 position);
        void setToolProgressAvailable(bool available);
        void setOverallProgressAvailable(bool available);
        void setToolName(const std::string& name);
        void setToolMessage(const std::string& msg);
		void setToolStartTime(time_t time);
		void setToolEstimatedTime(card64 estTime, time_t currenTime);
        void drawUpdates();
        void closeForm();

};

/// Global instance for the progress window.
/**
 * Use this instead of your own instance to make sure that there aren't several
 * progress windows displayed.
 */
extern X4INTERFACE_API ProgressWindow* progressWindow;

}

#endif