#ifndef CVWrapperH
#define CVWrapperH
//---------------------------------------------------------------------------
#include "x4math.h"
//---------------------------------------------------------------------------
#include "cv_include.h"
//---------------------------------------------------------------------------

namespace X4
{

/**
	provides easier use of open cv functions
*/
class CVWrapper
{
public:

	/// returns open cv type (for example CV_32FC1 for float32)
	template<class FloatType>
	static int getCVType();

	/// convert X4 static matrix to opencv matrix (just points to M.data(), to make sure M is valid when using cvMat!)
	template<unsigned dimension>
	static CvMat convertMatrixf( const StaticMatrix<float,dimension,dimension> &M );

	/// convert X4 dynamic matrix to opencv matrix (make sure to free CvMat!)
	template<class FloatType>
	static CvMat convertDMatrix( const DynamicMatrix<FloatType> &M );

	/// convert opencv matrix to X4 dynamic matrix
	template<class FloatType>
	static DynamicMatrix<FloatType> convertCVMatrix(const CvMat &M);

	/// convert X4 static vector to opencv vector (just points to v.data(), to make sure v is valid when using cvMat!)
	template<unsigned dimension>
	static CvMat convertVectorf( const StaticVector<float,dimension> &v );
	
	/// solve linear system Mx=rhs via SVD, returns true if successful
	template<int dimension>
	static bool solveLinearSystemf( const StaticMatrix<float,dimension,dimension> &M, const StaticVector<float,dimension> &rhs, StaticVector<float,dimension> &x);

	/// solve(dynamic) rectangle linear system Mx=rhs	, returns true if successful
	template<class FloatType>
	static bool solveLinearSystem( const DynamicMatrix<FloatType> &M, const DynamicVector<FloatType> &rhs, DynamicVector<FloatType> &x);

	/// compute inverse or pseudo-inverse
	template<class FloatType>
	static DynamicMatrix<FloatType> invert(const DynamicMatrix<FloatType> &M);


};

template<unsigned dimension>
CvMat CVWrapper::convertMatrixf( const StaticMatrix<float,dimension,dimension> &M )
{
	CvMat ret;
	cvInitMatHeader( &ret, dimension, dimension, CV_32FC1, (float*)M.data());
	return ret;
}

template<class FloatType>
int CVWrapper::getCVType()
{
	return sizeof(FloatType) == sizeof(float32)?CV_32FC1:CV_64FC1;
}

template<class FloatType>
CvMat CVWrapper::convertDMatrix( const DynamicMatrix<FloatType> &M )
{
	CvMat ret;
	FloatType * data = new FloatType[M.getRowsDim()*M.getColsDim()];
	mpcard count = 0;
	for (mpcard i=0; i<M.getRowsDim(); i++)
	{
		for (mpcard j=0; j<M.getColsDim(); j++)
		{
			data[count] = M[j][i];
			count++;
		}
	}
	cvInitMatHeader( &ret, M.getRowsDim(), M.getColsDim(), getCVType<FloatType>(), data);
	return ret;
}

template<class FloatType>
DynamicMatrix<FloatType> CVWrapper::convertCVMatrix(const CvMat &M)
{
	DynamicMatrix<FloatType> ret(M.cols,M.rows);
	FloatType * data = (FloatType*)M.data.ptr;
	mpcard count = 0;
	for( mpcard r=0;r<M.rows;r++ )
	{
		for( mpcard c=0;c<M.cols;c++ )
		{
			ret[c][r] = data[count++];
		}
	}
	return ret;
}

template<unsigned dimension>
CvMat CVWrapper::convertVectorf( const StaticVector<float,dimension> &v )
{
	CvMat ret;
	cvInitMatHeader( &ret, dimension, 1, CV_32FC1, (float*)v.data());
	return ret;
}


template<int dimension>
bool CVWrapper::solveLinearSystemf( const StaticMatrix<float,dimension,dimension> &M, const StaticVector<float,dimension> &rhs, StaticVector<float,dimension> &x)
{
	CvMat cvMat, cvRhs, cvX;
	cvInitMatHeader( &cvMat, dimension, dimension, CV_32FC1, (float*)M.data());
	cvInitMatHeader( &cvRhs, dimension, 1, CV_32FC1, (float*)rhs.data());
	cvInitMatHeader( &cvX, dimension, 1, CV_32FC1, x.data());

	int res = cvSolve( &cvMat,&cvRhs,&cvX, CV_SVD );
	return res != 1;
}

template<class FloatType>
static bool CVWrapper::solveLinearSystem( const DynamicMatrix<FloatType> &M, const DynamicVector<FloatType> &rhs, DynamicVector<FloatType> &x)
{
	CvMat cM = convertDMatrix(M);
	CvMat cvRhs, cvX;
	if( x.getDim() != M.getColsDim())
		x.setDim(M.getColsDim());
	cvInitMatHeader( &cvRhs, rhs.getDim(), 1, getCVType<FloatType>(), (float*)rhs.data());
	cvInitMatHeader( &cvX, x.getDim(), 1, getCVType<FloatType>(), x.data());

	int res = cvSolve( &cM,&cvRhs,&cvX, CV_SVD );
	FloatType * data = (FloatType*)cM.data.ptr;
	delete data;
	return res != 1;
}

template<class FloatType>
static DynamicMatrix<FloatType> CVWrapper::invert(const DynamicMatrix<FloatType> &M)
{
	CvMat cM = convertDMatrix(M);
	CvMat cRet;
	vector<FloatType> retData(M.getRowsDim() * M.getColsDim());
	cvInitMatHeader( &cRet, M.getColsDim(), M.getRowsDim(), getCVType<FloatType>(), &retData[0]);

	cvInvert(&cM,&cRet,CV_SVD);

	FloatType * data = (FloatType*)cM.data.ptr;
	delete data;

	return convertCVMatrix<FloatType>(cRet);	
}



}

#endif