//---------------------------------------------------------------------------
#ifndef DynamicLinearAlgebraH
#define DynamicLinearAlgebraH
//---------------------------------------------------------------------------
#include "x4math.h"
//---------------------------------------------------------------------------
#include "math_include.h"
#include <string>
//---------------------------------------------------------------------------
#include "X4Types.h"
#include "LinearAlgebra.h"
#include "X4Assert.h"
//---------------------------------------------------------------------------
#include "X4Exception.h"
//---------------------------------------------------------------------------

namespace X4 {

    class OutputObjectStream;
    class InputObjectStream;

    // ---------------------------------------------------------------------------
    //
    //                                 DynamicVector
    //
    // ---------------------------------------------------------------------------


    template <typename FloatType>
    class DynamicVector {
    private:
        mpcard dim;
        FloatType *v;

    public:
#if 0
        // -- create instance static function for handling the vectors in metaclasses
        static inline Object* newInstance(){return reinterpret_cast<Object*>(new DynamicVector<FloatType>());}
#endif 
        //  ---- constructors, copy constructor
        inline DynamicVector<FloatType>(mpcard dim = 0);
        inline DynamicVector<FloatType>(const DynamicVector<FloatType> &init);

        inline mpcard getDim() const;
        /// alias to getDim()
        //	inline mpcard size() const;
        inline void setDim(const mpcard dim = 0);
        /// decrease dim, don't change entries < smallerDim
        inline void cutOffDim(const mpcard smallerDim);
        /// set all entries to (FloatType)0
        inline void setZero();
        /// set all entries to (FloatType)value
        inline void setAllToValue(const FloatType &s);


        inline FloatType& operator[](const mpcard &index);
        inline const FloatType& operator[](const mpcard &index) const;

        //  ----  operators +, -, *, /, etc.
        inline DynamicVector<FloatType> operator+(const DynamicVector<FloatType> &op) const;
        inline DynamicVector<FloatType> operator-(const DynamicVector<FloatType> &op) const;
        inline DynamicVector<FloatType> operator-() const;
        inline FloatType operator*(const DynamicVector<FloatType> &op) const;
        inline DynamicVector<FloatType> operator*(const FloatType &s) const;
        inline DynamicVector<FloatType> operator/(const FloatType &s) const;
        inline DynamicVector<FloatType> componentProduct(const DynamicVector<FloatType> &op) const;
        inline DynamicVector<FloatType> componentAddScalar(const FloatType &s) const;
        inline DynamicVector<FloatType> componentSubstractScalar(const FloatType &s) const;
        inline FloatType componentSum() const;
        inline FloatType average() const;

        //  ---- operators +=, -=, *=, /=
        inline DynamicVector<FloatType> operator+=(const DynamicVector<FloatType> &op);
        inline DynamicVector<FloatType> operator-=(const DynamicVector<FloatType> &op);
        inline DynamicVector<FloatType> operator*=(const FloatType &s);
        inline DynamicVector<FloatType> operator/=(const FloatType &s);

        //  ---- operators =, ==, !=
        inline void operator=(const DynamicVector<FloatType> &op);
        inline bool operator==(const DynamicVector<FloatType> &op) const;
        inline bool operator!=(const DynamicVector<FloatType> &op) const;

        inline FloatType maxComponent() const;
        inline FloatType minComponent() const;
        inline DynamicVector<FloatType> absoluteValues() const;

        inline bool allComponentsLargerEqual(const DynamicVector<FloatType> &op) const;
        inline bool allComponentsLarger(const DynamicVector<FloatType> &op) const;
        inline bool allComponentsSmallerEqual(const DynamicVector<FloatType> &op) const;
        inline bool allComponentsSmaller(const DynamicVector<FloatType> &op) const;
        inline FloatType getSqrNorm() const;
        inline FloatType getNorm() const;
        inline DynamicVector<FloatType> append(const DynamicVector<FloatType> &op) const;
        inline DynamicVector<FloatType>& appendInPlace(const DynamicVector<FloatType> &op);
        inline DynamicVector<FloatType> append(const FloatType &op) const;
        inline DynamicVector<FloatType>& appendInPlace(const FloatType &op);

        //  ---- misc.
        void print();
        std::string toString() const;
        std::string toMatString() const;

        inline FloatType* data();
        inline const FloatType* data() const;
        inline FloatType* dataPtr();
        inline const FloatType* dataPtr() const;
        inline DynamicVector<FloatType> copyElements(mpcard howMany) const;
        inline DynamicVector<FloatType> subset(mpcard start, mpcard length) const;

        void write(OutputObjectStream *s) const;
        void read(InputObjectStream *s);	


        //Warning: data at "dataPointer" will be deleted by deleting DynamicVector
        inline void setDataPointer(FloatType* x, mpcard dim);

        // ---- destructor
        inline ~DynamicVector();


        // stl compatible interface
    public:
        typedef FloatType* iterator;
        typedef const FloatType* const_iterator;

        iterator begin() { return data(); }
        const_iterator begin() const { return data(); }
        iterator end() { return data() + getDim(); }
        const_iterator end() const { return data() + getDim(); }
        size_t size() const { return getDim(); }
        bool empty() const { return size() == 0; }
    };

    template<typename FloatType>
    inline DynamicVector<FloatType> operator*(const FloatType &s, const DynamicVector<FloatType>& m) {
        return m*s;
    }




    // ---------------------------------------------------------------------------
    //
    //                                 DynamicMatrix
    //
    // ---------------------------------------------------------------------------



    template <typename FloatType>
    class DynamicMatrix {
    private:
        mpcard rows;
        mpcard columns;
        DynamicVector<FloatType> *theColumns;

    public:

        // ---- constructors, copy constructor
        inline DynamicMatrix( mpcard c = 0, mpcard r = 0, bool initAsIdentity = true);
        inline DynamicMatrix(const DynamicMatrix<FloatType> &init);

        inline void setDimension(mpcard columns = 0, mpcard rows = 0, bool initAsIdentity = true);
        inline void setZero();
        inline mpcard getRowsDim() const;
        inline mpcard getColsDim() const;
        inline mpcard getRows() const;
        inline mpcard getColumns() const;

        // Oliver: not clear whether this returns cols or rows - maybe replace in favor of a col(index) function
        inline DynamicVector<FloatType>& operator[](mpcard index);
        inline const DynamicVector<FloatType>& operator[](mpcard index) const;

        inline DynamicVector<FloatType>& column(mpcard index);
        inline const DynamicVector<FloatType>& column(mpcard index) const;

        //  ----  operators +, -, *, /
        inline DynamicMatrix<FloatType> operator+(const DynamicMatrix<FloatType> &op) const;
        inline DynamicMatrix<FloatType> operator-(const DynamicMatrix<FloatType> &op) const;
        inline DynamicMatrix<FloatType> operator-() const;
        inline DynamicMatrix<FloatType> operator*(const FloatType &s) const;
        inline DynamicMatrix<FloatType> operator*(const DynamicMatrix<FloatType> &op) const;
        inline DynamicMatrix<FloatType> operator/(const FloatType &s) const;

        //  ---- operators +=, -=, *=, /=
        inline DynamicMatrix<FloatType> operator+=(const DynamicMatrix<FloatType> &op);
        inline DynamicMatrix<FloatType> operator-=(const DynamicMatrix<FloatType> &op);
        inline DynamicMatrix<FloatType> operator*=(const FloatType &op);
        inline DynamicMatrix<FloatType> operator*=(const DynamicMatrix<FloatType> &op);
        inline DynamicMatrix<FloatType> operator/=(const FloatType &op);

        //  ---- operators =, ==, !=
        inline void operator=(const DynamicMatrix<FloatType> &op);
        inline bool operator==(const DynamicMatrix<FloatType> &op) const;
        inline bool operator!=(const DynamicMatrix<FloatType> &op) const;

        // ---- multiplying with vectors
        inline DynamicVector<FloatType> operator*(const DynamicVector<FloatType> &v) const;

        /// multiply with another matrix but handle this one as being transposed
        inline DynamicMatrix<FloatType> multTransposed(const DynamicMatrix<FloatType>& m) const;

        // ----  misc.
        inline void print();
        std::string toString() const;
        std::string toMatString() const;

        inline void copyDataToArray(FloatType* target) const;
        inline void copyDataFromArray(FloatType* source);
        ///get top-left 3x3 values (for debug)
        inline Matrix3f getMatrix3f();

        inline void changeRows(mpcard row1, mpcard row2);
        inline void multRow(mpcard row, FloatType value);
        inline void combineRows(mpcard row, mpcard with, FloatType by);
        inline void addRows(mpcard row, mpcard with);
        inline DynamicMatrix<FloatType> transpose() const;

        inline DynamicMatrix<FloatType> append(const DynamicMatrix<FloatType> &op) const;
        inline DynamicMatrix<FloatType>& appendInPlace(const DynamicMatrix<FloatType> &op);
        inline DynamicMatrix<FloatType> append(const DynamicVector<FloatType> &op) const;
        inline DynamicMatrix<FloatType>& appendInPlace(const DynamicVector<FloatType> &op);

        /// copies subset of columns to new matrix
        inline DynamicMatrix<FloatType> copyColumns(mpcard startColumn, mpcard numColumns);

        inline FloatType getDeterminant() const;
        inline FloatType getTrace() const;

        void write(OutputObjectStream *s) const;
        void read(InputObjectStream *s);	


        // ---- destructor
        inline ~DynamicMatrix();
    };


    template<typename FloatType>
    inline DynamicMatrix<FloatType> operator*(const FloatType &s, const DynamicMatrix<FloatType>& m) {
        return m*s;
    }




    // ---------------------------------------------------------------------------
    //
    //                                   instances
    //
    // ---------------------------------------------------------------------------


    typedef DynamicVector<float> DVectorF;
    typedef DynamicVector<double> DVectorD;
    typedef DynamicVector<int> DVectorI;
    typedef DynamicVector<card8> DVectorUB;

    typedef DynamicMatrix<float> DMatrixF;
    typedef DynamicMatrix<double> DMatrixD;
    typedef DynamicMatrix<int> DMatrixI;
	typedef DynamicMatrix<card8> DMatrixUB;

    // ---------------------------------------------------------------------------
    //
    //                                 aux. routines
    //
    // ---------------------------------------------------------------------------


    template <typename FloatType>
    inline DynamicVector<FloatType> scalarDVector(mpcard dim, const FloatType &s);

    template <typename FloatType>
    inline DynamicVector<FloatType> nullDVector(mpcard dim);

    //template <typename FloatType>
    //inline DynamicVector<FloatType> makeDVector(mpcard dim, const FloatType * dataptr);

    template <typename iterator>
    inline DynamicVector<typename iterator_traits<iterator>::value_type> makeDVectorL(mpcard dim, iterator begin);

    template <typename iterator>
    inline DynamicVector<typename iterator_traits<iterator>::value_type> makeDVector(iterator begin, iterator end);

    template <typename Float, int Dim>
    inline DynamicVector<Float> makeDVector(const StaticVector<Float, Dim>& vec) { return makeDVectorL( Dim, vec.data() ); }

    template <typename FloatType1, typename FloatType2>
    inline void convertDVectorType(const DynamicVector<FloatType1> &source, DynamicVector<FloatType2> &dest);


    template <typename FloatType>
    inline DynamicMatrix<FloatType> scalarDMatrix(mpcard c, mpcard r, const FloatType &s);

    template <typename FloatType>
    inline DynamicMatrix<FloatType> nullDMatrix(mpcard c, mpcard r);

    /// v1^t v2
    template <typename FloatType>
    inline DynamicMatrix<FloatType> outerProduct( const DynamicVector<FloatType> &v1,const DynamicVector<FloatType> &v2 );

    template <typename FloatType>
    inline FloatType normQuad(const DynamicVector<FloatType> &v);

    template <typename FloatType>
    inline FloatType norm(const DynamicVector<FloatType> &v);

    template <typename FloatType>
    inline FloatType normInfinity(const DynamicVector<FloatType> &v);

    template <typename FloatType>
    inline mpcard searchPivot(const DynamicMatrix<FloatType> &mat, mpcard step);

    template <typename FloatType>
    inline DynamicMatrix<FloatType> invertMatrix(const DynamicMatrix<FloatType> &mat);


    template <typename FloatType>
    inline DynamicVector<FloatType> maxVector( const DynamicVector<FloatType> &v1,const DynamicVector<FloatType> &v2);


    template <typename FloatType>
    inline DynamicVector<FloatType> minVector( const DynamicVector<FloatType> &v1,const DynamicVector<FloatType> &v2);


#define ADD_DVECTORF_VAR_ARRAY_PROP( propName, version ) \
    ADD_VAR_ARRAY_PROP( \
    propName, \
    version, \
    new EmbeddedObjectClassProperty( \
#propName, MetaClass::FIRST_VERSION + version, (pointer_diff)((pointer_diff)&( obj_null_pointer->##propName ) - obj_null_offset), \
    new MetaClass("DVectorF", "", DVectorF::newInstance, sizeof(DVectorF)) ) \
    )

#define ADD_DVECTORF_PROP( propName, version )\
    ((MetaClass*)getClass())->addProperty( new EmbeddedObjectClassProperty( \
#propName, MetaClass::FIRST_VERSION + version, (pointer_diff)((pointer_diff)&( obj_null_pointer->##propName ) - obj_null_offset), \
    new MetaClass("DVectorF", "", DVectorF::newInstance, sizeof(DVectorF)) )) 

}


#endif
