//---------------------------------------------------------------------------
#ifndef DynamicLinearAlgebraInlineH
#define DynamicLinearAlgebraInlineH
//---------------------------------------------------------------------------
#include "x4math.h"
//---------------------------------------------------------------------------
#include "DynamicLinearAlgebra.h"
//---------------------------------------------------------------------------
#ifdef _DEBUG
#define DVECT_RANGE_CHECK
#endif
//---------------------------------------------------------------------------
#include "LinearAlgebra.inline.h"
#include "ObjectStreams.h"
//---------------------------------------------------------------------------

namespace X4 {

    // ---------------------------------------------------------------------------
    //
    //                                 DynamicVector
    //
    // ---------------------------------------------------------------------------


    template <typename FloatType>
    inline DynamicVector<FloatType>::DynamicVector(mpcard dim) {
        this->dim = dim;
        v = new FloatType[dim];
    }

    template <typename FloatType>
    inline mpcard DynamicVector<FloatType>::getDim() const {
        return dim;
    }

    /*template <typename FloatType>
    inline mpcard DynamicVector<FloatType>::size() const {
    return dim;
    }
    */

    template <typename FloatType>
    inline void DynamicVector<FloatType>::setDim(const mpcard dim) {
        if (this->dim == dim)
            return;

        this->dim = dim;
        delete[] v;
        v = new FloatType[dim];
    }

    template <typename FloatType>
    inline void DynamicVector<FloatType>::cutOffDim(const mpcard smallerDim) {

        x4Assert(smallerDim <= this->dim);
        const DynamicVector<FloatType> backup = *this;
        this->setDim(smallerDim);
        for (mpcard i = 0; i < smallerDim; i++)
        {
            v[i] = backup[i];
        }

    }

    template <typename FloatType>
    inline void DynamicVector<FloatType>::setZero() {
        for (mpcard i=0; i<dim; i++) {
            v[i] = (FloatType)0;
        }
    }


    template <typename FloatType>
    void X4::DynamicVector<FloatType>::setAllToValue( const FloatType &s )
    {
        for (mpcard i=0; i<dim; i++) {
            v[i] = s;
        }
    }

    template <typename FloatType>
    inline DynamicVector<FloatType>::DynamicVector(const DynamicVector<FloatType> &init) {
        this->dim = init.dim;
        v = new FloatType[dim];
        for (mpcard i=0; i<dim; i++) {
            v[i] = init.v[i];
        }
    }

    template <typename FloatType>
    inline FloatType& DynamicVector<FloatType>::operator[](const mpcard &index) {
#ifdef DVECT_RANGE_CHECK
        if (index >= dim) {
            throw ERangeCheck("DynamicVector::operator[] - index exceeds bounds");
        }
#endif
        return v[index];
    }

    template <typename FloatType>
    inline const FloatType& DynamicVector<FloatType>::operator[](const mpcard &index) const{
#ifdef DVECT_RANGE_CHECK
        if (index >= dim) {
            throw ERangeCheck("DynamicVector::operator[] - index exceeds bounds");
        }
#endif
        return v[index];
    }

    template <typename FloatType>
    inline DynamicVector<FloatType> DynamicVector<FloatType>::operator+(const DynamicVector<FloatType> &op) const {
#ifdef DVECT_RANGE_CHECK
        if (dim != op.dim) {
            throw ERangeCheck("DynamicVector::operator+ - incompatible dimensions.");
        }
#endif
        DynamicVector<FloatType> result(dim);
        for (mpcard i=0; i<dim; i++) {
            result[i] = v[i] + op.v[i];
        }
        return result;
    }

    template <typename FloatType>
    inline DynamicVector<FloatType> DynamicVector<FloatType>::operator-(const DynamicVector<FloatType> &op) const {
#ifdef DVECT_RANGE_CHECK
        if (dim != op.dim) {
            throw ERangeCheck("DynamicVector::operator- - incompatible dimensions.");
        }
#endif
        DynamicVector<FloatType> result(dim);
        for (mpcard i=0; i<dim; i++) {
            result[i] = v[i] - op.v[i];
        }
        return result;
    }

    template <typename FloatType>
    inline DynamicVector<FloatType> DynamicVector<FloatType>::operator-() const {
        DynamicVector<FloatType> result(dim);
        for (mpcard i=0; i<dim; i++) {
            result[i] = -v[i];
        }
        return result;
    }

    template <typename FloatType>
    inline FloatType DynamicVector<FloatType>::operator*(const DynamicVector<FloatType>& op) const {
#ifdef DVECT_RANGE_CHECK
        if (dim != op.dim) {
            throw ERangeCheck("DynamicVector::operator* - incompatible dimensions.");
        }
#endif
        FloatType result = 0.0;
        for (mpcard i=0; i<dim; i++) {
            result += v[i] * op.v[i];
        }
        return result;
    }

    template <typename FloatType>
    inline DynamicVector<FloatType> DynamicVector<FloatType>::operator*(const FloatType &s) const {
        DynamicVector<FloatType> result(dim);
        for (mpcard i=0; i<dim; i++) {
            result[i] = s * v[i];
        }
        return result;
    }

    template <typename FloatType>
    inline DynamicVector<FloatType> DynamicVector<FloatType>::operator/(const FloatType &s) const {
        DynamicVector<FloatType> result(dim);
        for (mpcard i=0; i<dim; i++) {
            result[i] = v[i] / s;
        }
        return result;
    }

    template <typename FloatType>
    inline DynamicVector<FloatType> DynamicVector<FloatType>::componentProduct(const DynamicVector<FloatType> &op) const {
#ifdef DVECT_RANGE_CHECK
        if (dim != op.dim) {
            throw ERangeCheck("DynamicVector::componentProduct - incompatible dimensions.");
        }
#endif
        DynamicVector<FloatType> result(dim);
        for (mpcard i=0; i<dim; i++) {
            result[i] = v[i] * op.v[i];
        }
        return result;
    }

    template <typename FloatType>
    inline DynamicVector<FloatType> DynamicVector<FloatType>::componentAddScalar(const FloatType &s) const {
        DynamicVector<FloatType> result(dim);
        for (mpcard i=0; i<dim; i++) {
            result[i] = v[i] + s;
        }
        return result;
    }

    template <typename FloatType>
    inline DynamicVector<FloatType> DynamicVector<FloatType>::componentSubstractScalar(const FloatType &s) const {
        DynamicVector<FloatType> result(dim);
        for (mpcard i=0; i<dim; i++) {
            result[i] = v[i] - s;
        }
        return result;
    }

    template <typename FloatType>
    inline FloatType DynamicVector<FloatType>::componentSum() const {
        FloatType result = 0;
        for (mpcard i=0; i<dim; i++) {
            result += v[i];
        }
        return result;
    }

    template <typename FloatType>
    inline FloatType DynamicVector<FloatType>::average() const {
        x4Assert(dim > 0);
        return componentSum() / dim;
    }

    template <typename FloatType>
    inline DynamicVector<FloatType> DynamicVector<FloatType>::operator+=(const DynamicVector<FloatType> &op) {
#ifdef DVECT_RANGE_CHECK
        if (dim != op.dim) {
            throw ERangeCheck("DynamicVector::operator+= - incompatible dimensions.");
        }
#endif
        for (mpcard i=0; i<dim; i++) {
            v[i] += op.v[i];
        }
        return *this;
    }

    template <typename FloatType>
    inline DynamicVector<FloatType> DynamicVector<FloatType>::operator-=(const DynamicVector<FloatType> &op) {
#ifdef DVECT_RANGE_CHECK
        if (dim != op.dim) {
            throw ERangeCheck("DynamicVector::operator-= - incompatible dimensions.");
        }
#endif
        for (mpcard i=0; i<dim; i++) {
            v[i] -= op.v[i];
        }
        return *this;
    }

    template <typename FloatType>
    inline DynamicVector<FloatType> DynamicVector<FloatType>::operator*=(const FloatType &s) {
        for (mpcard i=0; i<dim; i++) {
            v[i] *= s;
        }
        return *this;
    }

    template <typename FloatType>
    inline DynamicVector<FloatType> DynamicVector<FloatType>::operator/=(const FloatType &s) {
        for (mpcard i=0; i<dim; i++) {
            v[i] /= s;
        }
        return *this;
    }

    template <typename FloatType>
    inline void DynamicVector<FloatType>::operator=(const DynamicVector<FloatType> &op) {
        if (dim != op.dim) {
            delete[] v;
            v = new FloatType[op.dim];
        }
        this->dim = op.dim;
        for (mpcard i=0; i<op.dim; i++) {
            v[i] = op.v[i];
        }
    }

    template <typename FloatType>
    inline bool DynamicVector<FloatType>::operator==(const DynamicVector<FloatType>& op) const {
        if (dim != op.dim) {
            throw ERangeCheck("DynamicVector::operator== - incompatible dimensions.");
        }
        for (mpcard i=0; i<dim; i++) {
            if (v[i] != op.v[i]) return false;
        }
        return true;
    }

    template <typename FloatType>
    inline bool DynamicVector<FloatType>::operator!=(const DynamicVector<FloatType>& op) const {
#ifdef DVECT_RANGE_CHECK
        if (dim != op.dim) {
            throw ERangeCheck("DynamicVector::operator!= - incompatible dimensions.");
        }
#endif
        for (mpcard i=0; i<dim; i++) {
            if (v[i] != op.v[i]) return true;
        }
        return false;
    }

    template <typename FloatType>
    inline FloatType DynamicVector<FloatType>::maxComponent() const {
        x4Assert(dim >= 1);
        FloatType result = v[0];
        for (mpcard i=1; i<dim; i++) {
            if (v[i] > result) {
                result = v[i];
            }
        }
        return result;
    }

    template <typename FloatType>
    inline FloatType DynamicVector<FloatType>::minComponent() const {
        x4Assert(dim >= 1);
        FloatType result = v[0];
        for (mpcard i=1; i<dim; i++) {
            if (v[i] < result) {
                result = v[i];
            }
        }
        return result;
    }

    template <typename FloatType>
    inline DynamicVector<FloatType> DynamicVector<FloatType>::absoluteValues() const {
        DynamicVector<FloatType> result(dim);
        for (mpcard i=0; i<dim; i++) {
            result[i] = fabs(v[i]);
        }
        return result;
    }

    template <typename FloatType>
    inline bool DynamicVector<FloatType>::allComponentsLargerEqual(const DynamicVector<FloatType> &op) const {
        x4Assert(dim == op.dim);
        for (mpcard i=0; i<dim; i++) {
            if (v[i] < op.v[i]) {
                return false;
            }
        }
        return true;
    }

    template <typename FloatType>
    inline bool DynamicVector<FloatType>::allComponentsLarger(const DynamicVector<FloatType> &op) const {
        x4Assert(dim == op.dim);
        for (mpcard i=0; i<dim; i++) {
            if (v[i] <= op.v[i]) {
                return false;
            }
        }
        return true;
    }

    template <typename FloatType>
    inline bool DynamicVector<FloatType>::allComponentsSmallerEqual(const DynamicVector<FloatType> &op) const {
        x4Assert(dim == op.dim);
        for (mpcard i=0; i<dim; i++) {
            if (v[i] > op.v[i]) {
                return false;
            }
        }
        return true;
    }

    template <typename FloatType>
    inline bool DynamicVector<FloatType>::allComponentsSmaller(const DynamicVector<FloatType> &op) const {
        x4Assert(dim == op.dim);
        for (mpcard i=0; i<dim; i++) {
            if (v[i] >= op.v[i]) {
                return false;
            }
        }
        return true;
    }

    template <typename FloatType>
    inline FloatType DynamicVector<FloatType>::getSqrNorm() const {
        FloatType result = (FloatType) 0;
        for (mpcard i=0; i<dim; i++) {
            result +=  v[i] * v[i];
        }
        return result;
    }

    template <typename FloatType>
    inline FloatType DynamicVector<FloatType>::getNorm() const {
        FloatType sqrNorm = getSqrNorm();

        return (FloatType) sqrt( sqrNorm );
    }


    template <typename FloatType>
    DynamicVector<FloatType> X4::DynamicVector<FloatType>::append( const DynamicVector<FloatType> &op ) const
    {
        DynamicVector<FloatType> result(this->getDim() + op.getDim());
        mpcard i = 0;
        for (; i < dim; ++i)
            result[i] = v[i];

        for (; i < result.getDim(); ++i)
            result[i] = op[i-dim];

        return result;
    }

    template <typename FloatType>
    DynamicVector<FloatType>& X4::DynamicVector<FloatType>::appendInPlace(const DynamicVector<FloatType> &op)
    {
        const DynamicVector<FloatType> backup = *this;
        this->setDim(backup.getDim() + op.getDim());
        mpcard i = 0;
        for (; i < backup.getDim(); ++i) {
            v[i] = backup[i];
        }
        for (; i < dim; ++i) {
            v[i] = op[i-backup.getDim()];
        }
        return *this;
    }

    template <typename FloatType>
    DynamicVector<FloatType> X4::DynamicVector<FloatType>::append(const FloatType &op) const
    {
        DynamicVector<FloatType> result(this->getDim()+1);
        mpcard i = 0;
        for (; i < dim; ++i)
            result[i] = v[i];
        result[i] = op;
        return result;
    }

    template <typename FloatType>
    DynamicVector<FloatType>& X4::DynamicVector<FloatType>::appendInPlace(const FloatType &op)
    {
        const DynamicVector<FloatType> backup = *this;
        this->setDim(backup.getDim()+1);
        mpcard i = 0;
        for (; i < backup.getDim(); ++i) {
            v[i] = backup[i];
        }
        v[i] = op;
        return *this;
    }

    template <typename FloatType>
    DynamicMatrix<FloatType> X4::DynamicMatrix<FloatType>::append( const DynamicMatrix<FloatType> &op ) const
    {
        //DynamicMatrix<FloatType> result(this->getColsDim(), this->getRowsDim() + op.getRowsDim());
        //for(mpcard i = 0; i < getColsDim(); i++)
        //    result[i] = theColumns[i].append(op[i]);
        //return result;
        DynamicMatrix<FloatType> result(this->getColsDim() + op.getColsDim(), this->getRowsDim());
        if (0 != columns && rows != op.getRowsDim()) return *this;
        mpcard i = 0;
        for (; i < getColsDim(); ++i)
            result[i] = theColumns[i];
        for (; i < result.getColsDim(); ++i)
            result[i] = op[i-getColsDim()];
        return result;
    }

    template <typename FloatType>
    DynamicMatrix<FloatType>& X4::DynamicMatrix<FloatType>::appendInPlace(const DynamicMatrix<FloatType> &op)
    {
        const DynamicMatrix<FloatType> backup = *this;
        if (0 != columns && rows != op.getRowsDim()) return *this;
        this->setDimension(backup.getColsDim() + op.getColsDim(), backup.getRowsDim(), false);
        mpcard i = 0;
        for (; i < backup.getColsDim(); ++i) {
            theColumns[i] = backup[i];
        }
        for (; i < getColsDim(); ++i) {
            theColumns[i] = op[i-backup.getColsDim()];
        }
        return *this;
    }

    template <typename FloatType>
    DynamicMatrix<FloatType> X4::DynamicMatrix<FloatType>::append(const DynamicVector<FloatType> &op) const
    {
        DynamicMatrix<FloatType> result(this->getColsDim() + 1, this->getRowsDim());
        if (0 != columns && rows != op.getDim()) return *this;
        mpcard i = 0;
        for (; i < getColsDim(); ++i)
            result[i] = theColumns[i];
        result[i] = op;
        return result;
    }

    template <typename FloatType>
    DynamicMatrix<FloatType>& X4::DynamicMatrix<FloatType>::appendInPlace(const DynamicVector<FloatType> &op)
    {
        const DynamicMatrix<FloatType> backup = *this;
        if (0 != columns && rows != op.getDim()) return *this;
        this->setDimension(backup.getColsDim() + 1, backup.getRowsDim(), false);
        mpcard i = 0;
        for (; i < backup.getColsDim(); ++i) {
            theColumns[i] = backup[i];
        }
        theColumns[i] = op;
        return *this;
    }

    //{ // test code for appending
    //    DVectorF a(3), b(2);
    //    a[0] = 0; a[1] = 1; a[2] = 2;
    //    b[0] = 3; b[1] = 4;
    //    DVectorF c = a.append(b);
    //    debugOutput << c << "\n"; // 0 1 2 3 4
    //    a.appendInPlace(b);
    //    debugOutput << a << "\n"; // 0 1 2 3 4
    //    DVectorF d = a.append(5);
    //    debugOutput << d << "\n"; // 0 1 2 3 4 5
    //    a.appendInPlace(6);
    //    debugOutput << a << "\n"; // 0 1 2 3 4 6
    //}
    //{
    //    DMatrixF a(2, 2), b(2, 2);
    //    a[0][0] = 0; a[0][1] = 1; a[1][0] = 2; a[1][1] = 3;
    //    b[0][0] = 4; b[0][1] = 5; b[1][0] = 6; b[1][1] = 7;
    //    DMatrixF c = a.append(b);
    //    debugOutput << c << "\n";
    //    a.appendInPlace(b);
    //    debugOutput << a << "\n";
    //    DVectorF v(2);
    //    v[0] = 3; v[1] = 4;
    //    DMatrixF d = a.append(v);
    //    debugOutput << d << "\n";
    //    a.appendInPlace(v);
    //    debugOutput << a << "\n";
    //}

    template <typename FloatType>
    void DynamicVector<FloatType>::print() {
        debugOutput << '(';
        for (mpcard i=0; i<dim-1; i++) {
            debugOutput << v[i];
            debugOutput << ',';
            debugOutput << ' ';
        }
        debugOutput << v[dim-1];
        debugOutput << ')';
    }

    template <typename FloatType>
    std::string DynamicVector<FloatType>::toString() const {
        std::string result;
        result += '(';
        for (mpcard i=0; i<dim-1; i++) {
            char buffer[1000];
            sprintf(buffer, "%g", v[i]);
            result += buffer;
            result += ',';
            result += ' ';
        }
        char buffer[1000];
        sprintf(buffer, "%g", v[dim-1]);
        result += buffer;
        result += ')';
        return result;
    }

    template <typename FloatType>
    std::string DynamicVector<FloatType>::toMatString() const
    {
        std::ostringstream ss;
        ss << "[";
        for (size_t ii = 0; ii < dim; ++ii) {
            ss << v[ii] << ";";
        }
        ss << "]";
        return ss.str();
    }

    template <typename FloatType>
    inline FloatType* DynamicVector<FloatType>::data() {
        return &(v[0]);
    }

    template <typename FloatType>
    inline const FloatType* DynamicVector<FloatType>::data() const {
        return &(v[0]);
    }

    template <typename FloatType>
    inline FloatType* DynamicVector<FloatType>::dataPtr() {
        return &(v[0]);
    }

    template <typename FloatType>
    inline const FloatType* DynamicVector<FloatType>::dataPtr() const {
        return &(v[0]);
    }


    template <typename FloatType>
    inline DynamicVector<FloatType> DynamicVector<FloatType>::copyElements(mpcard howMany) const {
#ifdef DVECT_RANGE_CHECK
        if (howMany > dim) {
            throw ERangeCheck("DynamicVector::copyElements - incompatible dimensions.");
        }
#endif
        DynamicVector<FloatType> result(howMany);
        for (mpcard i=0; i<howMany; i++) {
            result[i] = v[i];
        }
        return result;
    }

    template <typename FloatType>
    inline DynamicVector<FloatType> DynamicVector<FloatType>::subset(mpcard start, mpcard length) const {
#ifdef DVECT_RANGE_CHECK
        //	if (!( 0<=start && 0<=length && start + length < dim) ) {
        if (!( 0 <= start && 0 <= length && start + length <= dim) ) {
            throw ERangeCheck("DynamicVector::subset - incompatible dimensions.");
        }
#endif
        DynamicVector<FloatType> result(length);
        for ( mpcard i1 = 0; i1 < length; ++ i1 ) {
            result[i1] = v[start+i1];
        }
        return result;
    }


    template <typename FloatType>
    inline void DynamicVector<FloatType>::setDataPointer(FloatType* x, const mpcard dim){
        delete[] v;
        this->dim = dim;
        v=x;
    }

    template <typename FloatType>
    inline DynamicVector<FloatType>::~DynamicVector() {
        delete[] v;
    }


    template <typename FloatType>
    void DynamicVector<FloatType>::write(OutputObjectStream *s) const {
        card64 dd = dim;
        s->write(dd);
        s->writeBuffer(v, dim*sizeof(FloatType));
    }

    template <typename FloatType>
    void DynamicVector<FloatType>::read(InputObjectStream *s) {
        delete[] v;
        card64 dd = dim;
        s->read(dd);
        dim = dd;
        v = new FloatType[dim];
        s->readBuffer(v, dim*sizeof(FloatType));
    }




    // ---------------------------------------------------------------------------
    //
    //                                 DynamicMatrix
    //
    // ---------------------------------------------------------------------------



    template <typename FloatType>
    inline DynamicMatrix<FloatType>::DynamicMatrix( mpcard c, mpcard r, bool initAsIdentity) {
        columns = c;
        rows = r;
        theColumns = new DynamicVector<FloatType>[columns];
        for ( mpcard n = 0; n < columns; n++ )
            theColumns[n] = DynamicVector<FloatType>(rows);
        for (mpcard j=0; j<columns; j++) {
            for (mpcard i=0; i<rows; i++) {
                if (i == j && initAsIdentity) {
                    theColumns[j][i] = 1.0;
                } else {
                    theColumns[j][i] = 0.0;
                }
            }
        }
    }

    template <typename FloatType>
    inline void DynamicMatrix<FloatType>::setDimension(mpcard columns, mpcard rows, bool initAsIdentity) {
        if (this->rows != rows || this->columns != columns)
        {
            delete[] theColumns;
            this->rows = rows;
            this->columns = columns;
            theColumns = new DynamicVector<FloatType>[columns];
            for ( mpcard n = 0; n < columns; n++ )
                theColumns[n] = DynamicVector<FloatType>(rows);
        }


        for (mpcard j=0; j<columns; j++) {
            for (mpcard i=0; i<rows; i++) {
                if (i == j && initAsIdentity) {
                    theColumns[j][i] = 1.0;
                } else {
                    theColumns[j][i] = 0.0;
                }
            }
        }
    }

    template <typename FloatType>
    inline mpcard DynamicMatrix<FloatType>::getRowsDim() const{
        return rows;
    }

    template <typename FloatType>
    inline mpcard DynamicMatrix<FloatType>::getColsDim() const{
        return columns;
    }

    template <typename FloatType>
    inline void DynamicMatrix<FloatType>::setZero() {
        for (mpcard j=0; j<columns; j++)
            for (mpcard i=0; i<rows; i++)
                theColumns[j][i] = 0.0;
    }

    template <typename FloatType>
    inline mpcard DynamicMatrix<FloatType>::getRows() const {
        return rows;
    }

    template <typename FloatType>
    inline mpcard DynamicMatrix<FloatType>::getColumns() const {
        return columns;
    }

    template <typename FloatType>
    inline DynamicMatrix<FloatType>::DynamicMatrix(const DynamicMatrix<FloatType> &init) {
        this->columns = init.columns;
        this->rows = init.rows;
        theColumns = new DynamicVector<FloatType>[columns];
        for (mpcard i=0; i<columns; i++) {
            theColumns[i] = init.theColumns[i];
        }
    }

    template <typename FloatType>
    inline DynamicVector<FloatType>& DynamicMatrix<FloatType>::operator[](mpcard index) {
        return column(index);
    }

    template <typename FloatType>
    inline const DynamicVector<FloatType>& DynamicMatrix<FloatType>::operator[](mpcard index) const {
        return column(index);
    }

    template <typename FloatType>
    inline DynamicVector<FloatType>& DynamicMatrix<FloatType>::column(mpcard index) {
#ifdef DVECT_RANGE_CHECK
        if (index >= columns) {
            throw ERangeCheck("DynamicMatrix::operator[] - index exceeds bounds");
        }
#endif
        return theColumns[index];
    }

    template <typename FloatType>
    inline const DynamicVector<FloatType>& DynamicMatrix<FloatType>::column(mpcard index) const {
#ifdef DVECT_RANGE_CHECK
        if (index >= columns) {
            throw ERangeCheck("DynamicMatrix::operator[] - index exceeds bounds");
        }
#endif
        return theColumns[index];
    }

    template <typename FloatType>
    inline DynamicMatrix<FloatType> DynamicMatrix<FloatType>::operator+(const DynamicMatrix<FloatType> &op) const {
#ifdef DVECT_RANGE_CHECK
        if (columns != op.columns || rows != op.rows) {
            throw ERangeCheck("DynamicMatrix::operator+ - incompatible dimensions.");
        }
#endif
        DynamicMatrix<FloatType> result(columns, rows);
        for (mpcard i=0; i<columns; i++) {
            result[i] = theColumns[i] + op.theColumns[i];
        }
        return result;
    }

    template <typename FloatType>
    inline DynamicMatrix<FloatType> DynamicMatrix<FloatType>::operator-(const DynamicMatrix<FloatType> &op) const {
#ifdef DVECT_RANGE_CHECK
        if (columns != op.columns || rows != op.rows) {
            throw ERangeCheck("DynamicMatrix::operator- - incompatible dimensions.");
        }
#endif
        DynamicMatrix<FloatType> result(columns, rows);
        for (mpcard i=0; i<columns; i++) {
            result[i] = theColumns[i] - op.theColumns[i];
        }
        return result;
    }

    template <typename FloatType>
    inline DynamicMatrix<FloatType> DynamicMatrix<FloatType>::operator-() const {
        DynamicMatrix<FloatType> result(columns, rows);
        for (mpcard i=0; i<columns; i++) {
            result[i] = -theColumns[i];
        }
        return result;
    }

    template <typename FloatType>
    inline DynamicMatrix<FloatType> DynamicMatrix<FloatType>::operator*(const FloatType &s) const {
        DynamicMatrix<FloatType> result(columns, rows);
        for (mpcard i=0; i<columns; i++) {
            result[i] = theColumns[i] * s;
        }
        return result;
    }

    template <typename FloatType>
    inline DynamicMatrix<FloatType> DynamicMatrix<FloatType>::operator*(const DynamicMatrix<FloatType> &op) const {
        if ( columns != op.rows ) throw
            X4Exception("DynamicMatrix::operator* matrix dimensions do not match.");

        DynamicMatrix<FloatType> result( op.columns, rows );
        for ( mpcard r = 0; r < rows; r++ ){
            for ( mpcard c = 0; c < op.columns; c++ ){
                result[c][r] = 0.0;
                for ( mpcard i = 0; i < columns; i++ ) {
                    result[c][r] += (*this)[i][r] * op[c][i];
                }
            }
        }
        return result;
    }

    template <typename FloatType>
    inline DynamicMatrix<FloatType> DynamicMatrix<FloatType>::multTransposed(const DynamicMatrix<FloatType>& op) const
    {
        if ( rows != op.rows ) throw
            X4Exception("DynamicMatrix::multTransposed() matrix dimensions do not match.");

        DynamicMatrix<FloatType> result( op.columns, columns );
        for ( mpcard r = 0; r < columns; r++ ){
            for ( mpcard c = 0; c < op.columns; c++ ){
                result[c][r] = 0.0;
                for ( mpcard i = 0; i < op.rows; i++ ) {
                    result[c][r] += (*this)[r][i] * op[c][i];
                }
            }
        }
        return result;
    }

    template <typename FloatType>
    inline DynamicMatrix<FloatType> DynamicMatrix<FloatType>::operator/(const FloatType &s) const {
        DynamicMatrix<FloatType> result(columns, rows);
        for (mpcard i=0; i<columns; i++) {
            result[i] = theColumns[i] / s;
        }
        return result;
    }

    template <typename FloatType>
    inline DynamicMatrix<FloatType> DynamicMatrix<FloatType>::operator+=(const DynamicMatrix<FloatType> &op) {
#ifdef DVECT_RANGE_CHECK
        if (columns != op.columns || rows != op.rows) {
            throw ERangeCheck("DynamicMatrix::operator+= - incompatible dimensions.");
        }
#endif
        for (mpcard i=0; i<columns; i++) {
            theColumns[i] += op.theColumns[i];
        }
        return *this;
    }

    template <typename FloatType>
    inline DynamicMatrix<FloatType> DynamicMatrix<FloatType>::operator-=(const DynamicMatrix<FloatType> &op) {
#ifdef DVECT_RANGE_CHECK
        if (columns != op.columns || rows != op.rows) {
            throw ERangeCheck("DynamicMatrix::operator-= - incompatible dimensions.");
        }
#endif
        for (mpcard i=0; i<columns; i++) {
            theColumns[i] -= op.theColumns[i];
        }
        return *this;
    }

    template <typename FloatType>
    inline DynamicMatrix<FloatType> DynamicMatrix<FloatType>::operator*=(const FloatType &op) {
        for (mpcard i=0; i<columns; i++) {
            theColumns[i] *= op;
        }
        return *this;
    }

    template <typename FloatType>
    inline DynamicMatrix<FloatType> DynamicMatrix<FloatType>::operator*=(const DynamicMatrix<FloatType> &op) {
#ifdef DVECT_RANGE_CHECK
        if (columns != op.columns || rows != op.rows) {
            throw ERangeCheck("DynamicMatrix::operator*= - incompatible dimensions.");
        }
#endif
        DynamicMatrix<FloatType> result = (*this)*op;
        *this = result;
        return result;
    }

    template <typename FloatType>
    inline DynamicMatrix<FloatType> DynamicMatrix<FloatType>::operator/=(const FloatType &op) {

        for (mpcard i=0; i<columns; i++) {
            theColumns[i] /= op;
        }
        return *this;
    }

    template <typename FloatType>
    inline void DynamicMatrix<FloatType>::operator=(const DynamicMatrix<FloatType> &op) {
        if (columns != op.columns || rows != op.rows) {
            delete[] theColumns;
            theColumns = new DynamicVector<FloatType>[op.columns];
        }
        this->columns = op.columns;
        this->rows = op.rows;
        for (mpcard i=0; i<op.columns; i++) {
            theColumns[i] = op.theColumns[i];
        }
    }

    template <typename FloatType>
    inline bool DynamicMatrix<FloatType>::operator==(const DynamicMatrix<FloatType> &op) const {
#ifdef DVECT_RANGE_CHECK
        if (columns != op.columns || rows != op.rows) {
            throw ERangeCheck("DynamicMatrix::operator== - incompatible dimensions.");
        }
#endif
        for (mpcard i=0; i<columns; i++) {
            if (theColumns[i] != op.theColumns[i]) return false;
        }
        return true;
    }

    template <typename FloatType>
    inline bool DynamicMatrix<FloatType>::operator!=(const DynamicMatrix<FloatType> &op) const {
#ifdef DVECT_RANGE_CHECK
        if (columns != op.columns || rows != op.rows) {
            throw ERangeCheck("DynamicMatrix::operator!= - incompatible dimensions.");
        }
#endif
        for (mpcard i=0; i<columns; i++) {
            if (theColumns[i] != op.theColumns[i]) return true;
        }
        return false;
    }

    template <typename FloatType>
    inline DynamicVector<FloatType> DynamicMatrix<FloatType>::operator*(const DynamicVector<FloatType> &v) const {
#ifdef DVECT_RANGE_CHECK
        if (v.getDim() != columns) {
            throw ERangeCheck("DynamicMatrix::operator* (vect) - incompatible dimensions.");
        }
#endif
        DynamicVector<FloatType> result(rows);
        for (mpcard y=0; y<rows; y++) {
            result[y] = 0.0;
            for (mpcard x=0; x<columns; x++) {
                result[y] += theColumns[x][y]*v[x];
            }
        }
        return result;
    }

    template <typename FloatType>
    inline void DynamicMatrix<FloatType>::print() {
        for (mpcard i=0; i<columns; i++) {
            theColumns[i].print();
            std::cout << endl;
        }
    }

    template <typename FloatType>
    std::string DynamicMatrix<FloatType>::toString() const {
        std::string result;
		  if (columns == 0 || rows == 0) return result;
        for (mpcard r=0; r<rows; r++) {
            result += '(';
            char buffer[1000];
            for (mpcard c=0; c<columns-1; c++) {
                sprintf(buffer, "%.8f", theColumns[c][r]);
                result += buffer;
                result += ',';
                result += ' ';
            }
            sprintf(buffer, "%.8f", theColumns[columns-1][r]);
            result += buffer;
            result += ") \n";
        }
        return result;
    }

    template<typename FloatType>
    std::string DynamicMatrix<FloatType>::toMatString() const
    {
        std::ostringstream ss;
        ss << "[";
        for (size_t rr = 0; rr < rows; ++rr) {
            for (size_t cc = 0; cc < columns-1; ++cc) {
                ss << theColumns[cc][rr] << ",";
            }
            ss << theColumns[columns-1][rr] << ";";
        }
        ss << "]";
        return ss.str();
    }

    template <typename FloatType>
    inline void DynamicMatrix<FloatType>::copyDataToArray(FloatType* target) const{
        mpcard count = 0;
        for (mpcard j=0; j<columns; j++)
            for (mpcard i=0; i<rows; i++){
                target[count] = theColumns[j][i];
                count++;
            }
    }

    template <typename FloatType>
    inline void DynamicMatrix<FloatType>::copyDataFromArray(FloatType* source){
        mpcard count = 0;
        for (mpcard j=0; j<columns; j++)
            for (mpcard i=0; i<rows; i++){
                theColumns[j][i]= source[count];
                count++;
            }
    }

    template <typename FloatType>
    inline Matrix3f DynamicMatrix<FloatType>::getMatrix3f() {
        Matrix3f res;
        for ( unsigned char i = 0; i < 3; i++ )
            for ( unsigned char j = 0; j < 3; j++ )
                res[i][j] = theColumns[i][j];

        return res;
    }

    template <typename FloatType>
    inline void DynamicMatrix<FloatType>::changeRows(mpcard row1, mpcard row2) {
        DynamicVector<FloatType> h(columns);
        mpcard i;
        for (i=0; i<columns; i++) {
            h[i] = theColumns[i][row2];
        }
        for (i=0; i<columns; i++) {
            theColumns[i][row2] = theColumns[i][row1];
        }
        for (i=0; i<columns; i++) {
            theColumns[i][row1] = h[i];
        }
    }

    template <typename FloatType>
    inline void DynamicMatrix<FloatType>::multRow(mpcard row, FloatType value) {
        for (mpcard i=0; i<columns; i++) {
            theColumns[i][row] *= value;
        }
    }

    template <typename FloatType>
    inline void DynamicMatrix<FloatType>::combineRows(mpcard row, mpcard with, FloatType by) {
        for (mpcard i=0; i<columns; i++) {
            theColumns[i][row] += theColumns[i][with] * by;
        }
    }

    template <typename FloatType>
    inline void DynamicMatrix<FloatType>::addRows(mpcard row, mpcard with) {
        for (mpcard i=0; i<columns; i++) {
            theColumns[i][row] += theColumns[i][with];
        }
    }

    template <typename FloatType>
    inline DynamicMatrix<FloatType> DynamicMatrix<FloatType>::transpose() const {
        DynamicMatrix<FloatType> result(rows, columns);
        for (mpcard r=0; r<rows; r++) {
            for (mpcard c=0; c<columns; c++) {
                result[r][c] = theColumns[c][r];
            }
        }
        return result;
    }

    template <typename FloatType>
    inline DynamicMatrix<FloatType> DynamicMatrix<FloatType>::copyColumns(mpcard startColumn, mpcard numColumns) {
        x4Assert(startColumn < columns);
        x4Assert(startColumn + numColumns < columns);
        DynamicMatrix<FloatType> result(numColumns, rows, false);
        for (mpcard i=0; i<numColumns; i++) {
            result[i] = theColumns[i+startColumn];
        }
        return result;
    }

    template <typename FloatType>
    inline FloatType DynamicMatrix<FloatType>::getDeterminant() const {
        if ( ( rows == 2 ) && ( columns == 2 ) ) {
            return (*this)[0][0] * (*this)[1][1] - (*this)[0][1] * (*this)[1][0];
        } else if ( ( rows == 3 ) && ( columns == 3 ) ) {
            return (*this)[0][0] * ( (*this)[1][1] * (*this)[2][2] - (*this)[1][2] * (*this)[2][1] ) -
                (*this)[0][1] * ( (*this)[1][0] * (*this)[2][2] - (*this)[1][2] * (*this)[2][0] ) +
                (*this)[0][2] * ( (*this)[1][0] * (*this)[2][1] - (*this)[1][1] * (*this)[2][0] );
        } else if (rows == 1 && columns == 1) {
            return (*this)[0][0];
        } else {
            // currently does not support higher dimensions
            x4Assert(false);
        }
    }

    template <typename FloatType>
    inline FloatType DynamicMatrix<FloatType>::getTrace() const {
        x4Assert( getColsDim() == getRowsDim() );
        FloatType ret(0);
        for( int i1 = 0; i1 < getColsDim(); ++ i1 ) {
            ret += (*this)[i1][i1];
        }
        return ret;
    }


    // ---- destructor
    template <typename FloatType>
    inline DynamicMatrix<FloatType>::~DynamicMatrix() {
        delete[] theColumns;
    }


    // ---------------------------------------------------------------------------
    //
    //                                 aux. routines
    //
    // ---------------------------------------------------------------------------


    template <typename FloatType>
    inline DynamicVector<FloatType> scalarDVector(mpcard dim, const FloatType &s) {
        DynamicVector<FloatType> result(dim);
        for (mpcard i=0; i<dim; i++) {
            result[i] = s;
        }
        return result;
    }

    template <typename FloatType>
    inline DynamicVector<FloatType> nullDVector(mpcard dim) {
        return scalarDVector(dim, (FloatType)0);
    }

    /*
    // this case is handles by following function with Iterator==const FloatType*
    template <typename FloatType>
    inline DynamicVector<FloatType> makeDVector(mpcard dim, const FloatType * dataptr)
    {
    DynamicVector<FloatType> ret(dim);
    for( mpcard i=0;i<dim;i++)
    ret[i] = dataptr[i];

    return ret;
    }*/

    template <typename iterator>
    inline DynamicVector<typename iterator_traits<iterator>::value_type> makeDVectorL(mpcard dim, iterator it) {
        DynamicVector<typename iterator_traits<iterator>::value_type> ret(dim);
        for( mpcard i=0; i<dim; ++ i ) {
            ret[i] = *it;
            ++ it;
        }

        return ret;
    }

    template <typename iterator>
    inline DynamicVector<typename iterator_traits<iterator>::value_type> makeDVector(iterator begin, iterator end) {
        mpcard size = 0;
        for( iterator it( begin ); it != end; ++ it )
            ++ size;

        return makeDVectorL<iterator>(size, begin);
    }


    template <typename FloatType1, typename FloatType2>
    inline void convertDVectorType(const DynamicVector<FloatType1> &source, DynamicVector<FloatType2> &dest) {
        if (source.getDim() != dest.getDim()) {
            dest.setDim(source.getDim());
        }
        for (mpcard i=0; i<source.getDim(); i++) {
            dest[i] = (FloatType1)source[i];
        }
    }

    template <typename FloatType>
    inline DynamicMatrix<FloatType> scalarDMatrix(mpcard c, mpcard r, const FloatType &s) {
        DynamicMatrix<FloatType> result(c,r);
        for (mpcard i=0; i<c; i++) {
            result[i] = scalarDVector(r, s);
        }
        return result;
    }

    template <typename FloatType>
    inline DynamicMatrix<FloatType> nullDMatrix(mpcard c, mpcard r) {
        return scalarDMatrix(c, r, (FloatType)0);
    }

    template <typename FloatType>
    inline DynamicMatrix<FloatType> outerProduct( const DynamicVector<FloatType> &v1,const DynamicVector<FloatType> &v2 )
    {
        DynamicMatrix<FloatType> ret( v1.getDim(), v2.getDim() );
        for( mpcard i=0;i<v1.getDim();i++ )
        {
            for( mpcard j=0;j<v2.getDim();j++ )
            {
                ret[i][j] = v1[i] * v2[j];
            }
        }
        return ret;
    }


    template <typename FloatType>
    inline FloatType normQuad(const DynamicVector<FloatType> &v) {
        FloatType result = 0.0;
        mpcard dim = v.getDim();
        for (mpcard i=0; i<dim; i++) {
            result+=v[i]*v[i];
        }
        return result;
    };

    template <typename FloatType>
    inline FloatType norm(const DynamicVector<FloatType> &v) {
        return sqrt(normQuad(v));
    };

    template <typename FloatType>
    inline FloatType normInfinity(const DynamicVector<FloatType> &v){
        FloatType max = -1e10;
        for(mpcard index = 0; index < v.getDim(); index++)
        {
            if(fabs(v[index]) > max)
            {
                max = fabs(v[index]);
            }
        }
        return max;
    }


    template <typename FloatType>
    inline mpcard searchPivot(const DynamicMatrix<FloatType> &mat, mpcard step) {
        if (mat.getColumns() != mat.getRows()) {
            throw ERangeCheck("searchPivot - symmetric matrix needed.");
        }
        const mpcard size = mat.getRows();
        const double MAX_PIVOT_SIZE = 0.0;
        mpcard result = step;
        FloatType max = fabs(mat[step][step]);
        for (mpcard p = step+1; p<size; p++) {
            if (fabs(mat[step][p]) > max) {
                max = fabs(mat[step][p]);
                result = p;
            }
        }
        if (max <= MAX_PIVOT_SIZE) {
            throw X4Exception("searchPivot: no pivot element found (matrix singular?).");
        } else {
            return result;
        }
    };

    template <typename FloatType>
    inline DynamicMatrix<FloatType> invertMatrix(const DynamicMatrix<FloatType> &mat) {
        if (mat.getColumns() != mat.getRows()) {
            throw ERangeCheck("searchPivot - symmetric matrix needed.");
        }
        const mpcard size = mat.getRows();
        DynamicMatrix<FloatType> result(size, size);
        DynamicMatrix<FloatType> mCopy = mat;
        for (mpcard step=0; step<size; step++) {
            mpcard pivot = searchPivot(mCopy, step);
            if (step != pivot) {
                mCopy.changeRows(step, pivot);
                result.changeRows(step, pivot);
            }
            if (mCopy[step][step] == 0) throw X4Exception("invertMatrix: Matrix is singular.");
            FloatType m = 1.0f/mCopy[step][step];
            mCopy.multRow(step, m);
            result.multRow(step, m);
            for (mpcard row=step+1; row<size; row++) {
                FloatType m = -mCopy[step][row];
                mCopy.combineRows(row, step, m);
                result.combineRows(row, step, m);
            }
        }
        for (mpint step=size-1; step>=0; step--) {
            for (mpint row=step-1; row>=0; row--) {
                FloatType m = -mCopy[step][row];
                //mCopy.combineRows(row, step, m);
                result.combineRows(row, step, m);
            }
        }
        return result;
    };


    template <typename FloatType>
    void DynamicMatrix<FloatType>::write(OutputObjectStream *s) const {
        card64 rr = rows;
        s->write(rr);

        card64 cc = columns;
        s->write(cc);
        for (mpcard i=0; i<columns; i++) {
            theColumns[i].write(s);
        }
    }

    template <typename FloatType>
    void DynamicMatrix<FloatType>::read(InputObjectStream *s) {
        delete[] theColumns;
        card64 rr;
        s->read(rr); rows = rr;

        card64 cc;
        s->read(cc); columns = cc;
        theColumns = new DynamicVector<FloatType>[columns];
        for (mpcard i=0; i<columns; i++) {
            theColumns[i].read(s);
        }
    }


    template <typename FloatType>
    inline DynamicVector<FloatType> maxVector( const DynamicVector<FloatType> &v1,const DynamicVector<FloatType> &v2)
    {
        x4Assert( v1.size() == v2.size() );
        DynamicVector<FloatType> ret(v1.size());
        for( mpcard i=0;i<v1.size();i++ )
        {
            ret[i] = max( v1[i], v2[i] );
        }
        return ret;
    }


    template <typename FloatType>
    inline DynamicVector<FloatType> minVector( const DynamicVector<FloatType> &v1,const DynamicVector<FloatType> &v2)
    {
        x4Assert( v1.size() == v2.size() );
        DynamicVector<FloatType> ret(v1.size());
        for( mpcard i=0;i<v1.size();i++ )
        {
            ret[i] = min( v1[i], v2[i] );
        }
        return ret;
    }


}

#endif
