//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "DynamicLinearAlgebraTools.h"
#include "DynamicLinearAlgebraTools.inline.h"
//---------------------------------------------------------------------------
//#include "opencv/cv.h"
//#include "opencv/cvaux.h"
#include "opencv/cxcore.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

/// explicit instantiation
template class DynamicPCA<float32>;
template class DynamicPCA<float64>;



// ============================================================================
//         IMPLEMENTATION
// ============================================================================

template <typename FloatType>
DynamicPCA<FloatType>::DynamicPCA(DynamicVector<FloatType> * data, card32 n){
   dataPointer = data;
   numberOfVecs = n;
   x4Assert(n>0);
   dimension=data[0].getDim();
   type=getType();
   useUserDefinedMean = false;
}

template <typename FloatType>
DynamicPCA<FloatType>::DynamicPCA(){
   dataPointer = NULL;
   numberOfVecs = 0;
   dimension=0;
   type=getType();
   useUserDefinedMean = false;
}


/*template <typename FloatType>
const static int DynamicPCA<FloatType>::getType(){
   x4Assert(false); //Tool doesn't support that dataType
   return -1;
}*/

template<>
const int DynamicPCA<float32>::getType(){
    return CV_32FC1;
}

template<>
const int DynamicPCA<float64>::getType(){
	return CV_64FC1;
}

template <typename FloatType>
void DynamicPCA<FloatType>::setUserDefinedMean(const DynamicVector<FloatType> &userMean) {
	userDefinedMean = userMean;
	useUserDefinedMean = true;

}

template <typename FloatType>
DynamicVector<FloatType> DynamicPCA<FloatType>::calcMean(){
	if (useUserDefinedMean) return userDefinedMean;
	
   DynamicVector<FloatType> mean(dimension);

   for (card32 d=0; d<dimension ; d++) //set erg to zero-vec
      mean[d]= 0.0f;

   for (card32 i=0; i<numberOfVecs ; i++){
      mean += dataPointer[i];
   }

   mean/= (float) numberOfVecs;
   return mean;
}

// needed for OpenCV Datasructure

//xxx_open_cv_missing
typedef CvMat* cvMatrixPointer;

template <typename FloatType>
void DynamicPCA<FloatType>::calcCovarMatrix(DynamicMatrix<FloatType> & target){

  // Prepare OpenCV Datasructure
   FloatType* covMatrixArray = new FloatType[dimension*dimension];
   CvMat cv_covMatrix;
   cvInitMatHeader( &cv_covMatrix, dimension, dimension, type, covMatrixArray);

   FloatType* averageArray = new FloatType[dimension];
   CvMat cv_avrVector;
   cvInitMatHeader( &cv_avrVector, dimension, 1, type, averageArray);

   CvMat* cv_trainingVectors= new CvMat[numberOfVecs];
   cvMatrixPointer* vecPointers= new cvMatrixPointer[numberOfVecs];


   for (card32 n=0; n<numberOfVecs; n++){
       cvInitMatHeader( &cv_trainingVectors[n] , dimension, 1, type, dataPointer[n].data());
       vecPointers[n] =  &cv_trainingVectors[n];
   }

   // calc Covariance Matrix
   cvCalcCovarMatrix( (const void **) vecPointers, numberOfVecs, &cv_covMatrix, &cv_avrVector,  CV_COVAR_NORMAL | CV_COVAR_SCALE);


   //clean up
   delete[] averageArray;
   delete[] cv_trainingVectors;
   delete[] vecPointers;

   target.setDimension(dimension, dimension, false);
   target.copyDataFromArray(covMatrixArray);
   delete[] covMatrixArray;
}



template <typename FloatType>
void DynamicPCA<FloatType>::calcCovarMatrixB(DynamicMatrix<FloatType> & cov){

   cov.setDimension(dimension, dimension, false);
   cov.setZero();

   DynamicVector<FloatType> avg = calcMean();

   for (card32 n=0; n<numberOfVecs; n++){

      DynamicVector<FloatType> dataPoint= dataPointer[n];

      for (card32  j=0; j<dimension; j++) {
         for (card32 i=0; i<dimension; i++) {
               cov[j][i] += (dataPoint[j]-avg[j]) *  (dataPoint[i]-avg[i]);
         }
      }
   }

   float oneOverNumberOfVecs= 1.0f/(float)numberOfVecs;
   cov*=oneOverNumberOfVecs;
}



template <typename FloatType>
DynamicVector<FloatType> DynamicPCA<FloatType>::calcEigenValues(bool useU){


  // debugOutput << "CalcCovariance with " << numberOfVecs << " vectors of dim " << dimension << "\n";

   centroid = calcMean();

	DynamicMatrix<FloatType> cov(dimension, dimension);
   calcCovarMatrix(cov);

 //  debugOutput << "Calc SVD, size: " << dimension << " x " << dimension << "\n";

    //================ SVD ====================
   // Convert cov to array
   FloatType* covArray = new FloatType[dimension*dimension];
   cov.copyDataToArray(covArray);

   // SVD Result
   FloatType* resArray = new FloatType[dimension*dimension];
   DynamicVector<FloatType>  eigenValVector(dimension);

   // Prepare OpenCV Datasructure
   CvMat cvCov, cvEigenvalue, cvEigenvector;
   cvInitMatHeader( &cvCov, dimension, dimension, type, covArray);
   cvInitMatHeader( &cvEigenvalue, dimension, 1, type, eigenValVector.data());
   cvInitMatHeader( &cvEigenvector, dimension, dimension, type, resArray);

   // Do SVD
   cvSVD(&cvCov, &cvEigenvalue, &cvEigenvector, 0, CV_SVD_MODIFY_A | (useU?CV_SVD_U_T:CV_SVD_V_T));

   // convert to DynamicMatrix
   eigenVectors.setDimension( dimension, dimension );
   eigenVectors.copyDataFromArray(resArray);

   //clean up
   delete[] covArray;
   delete[] resArray;

	return eigenValVector;

}

template <typename FloatType>
void DynamicPCA<FloatType>::computeSVD(const DynamicMatrix<FloatType>  &input,
   DynamicMatrix<FloatType>  &u,
   DynamicVector<FloatType>  &singularValues,
   DynamicMatrix<FloatType>  &v,
   bool transposeU,
   bool transposeV)
{
   int type =  getType();
//decomposes matrix input into a product of a diagonal matrix and two orthogonal matrices:
        //      A=U*W*Vt   //

   card32 colsDim= input.getRowsDim();
   card32 rowsDim= input.getColsDim();
   //debugOutput << "computeSVD, size: " << colsDim << " x " << rowsDim << "\n";
   x4Assert(colsDim==rowsDim); //only tested for N=M-Matrix


   // Convert input to array (because OpenCv needs an array)
   FloatType* inputArray = new FloatType[colsDim*rowsDim];
   input.copyDataToArray(inputArray);

   // allocate SVD result arrays
   FloatType* resArray_U = new FloatType[colsDim*colsDim];
   FloatType* resArray_V = new FloatType[rowsDim*rowsDim];
   singularValues.setDim(rowsDim);         

   // Prepare OpenCV Datasructure CvMat
   CvMat cvMatInput, cvMat_U, cvMat_W, cvMat_Vt;
   cvInitMatHeader( &cvMatInput, colsDim, rowsDim, type, inputArray);
   cvInitMatHeader( &cvMat_W, rowsDim, 1, type, singularValues.data());
   cvInitMatHeader( &cvMat_U, colsDim, colsDim, type, resArray_U);
   cvInitMatHeader( &cvMat_Vt, rowsDim, rowsDim, type, resArray_V);

   // Do SVD
   cvSVD(&cvMatInput, &cvMat_W, &cvMat_U, &cvMat_Vt, CV_SVD_MODIFY_A | (transposeU?CV_SVD_U_T:0) | (transposeV?CV_SVD_V_T:0) );
   //CV_SVD_MODIFY_A enables modification of matrix A (inputArray) during the operation. It speeds up the processing.

   // convert result to DynamicMatrix
   u.setDimension(colsDim, colsDim);
   u.copyDataFromArray(resArray_U);
   v.setDimension(rowsDim, rowsDim);
   v.copyDataFromArray(resArray_V);

   //clean up                 
   delete[] inputArray;
   delete[] resArray_U;
   delete[] resArray_V;

}





typedef float32 TestDataStructure;


class X4MATH_API TestDynamicPCA {


 public:
   void test(){

      DynamicVector<TestDataStructure> *data;
      card32 dim = 30;
      card32 testNum =2000;

      data = new DynamicVector<TestDataStructure>[testNum];
      for (card32 i=0; i<testNum ; i++){
         data[i].setDim(dim);

         //Vector3f x;
         for (card32 d=0; d<dim ; d++){
           data[i][d]= (TestDataStructure) rnd01()*100.0f;

         }
      }

      DynamicPCA<TestDataStructure>  p(data, testNum);
      DynamicVector<TestDataStructure> erg(dim);

      erg = p.calcMean();
      DynamicVector<TestDataStructure> eigenV= p.calcEigenValues();
   }


   void testSVD(){

      DynamicVector<TestDataStructure> *data;
      card32 dim = 30;
      card32 testNum =2000;

      data = new DynamicVector<TestDataStructure>[testNum];
      for (card32 i=0; i<testNum ; i++){
         data[i].setDim(dim);

         //Vector3f x;
         for (card32 d=0; d<dim ; d++){
           data[i][d]= (TestDataStructure) rnd01()*100.0f;

         }
      }

      DynamicPCA<TestDataStructure>  p(data, testNum);
      DynamicVector<TestDataStructure> erg(dim);
      DynamicMatrix<TestDataStructure>  u;
      DynamicVector<TestDataStructure>  singularValues;
      DynamicMatrix<TestDataStructure>  v;
      DynamicMatrix<TestDataStructure> input;
      p.calcCovarMatrix(input);

      DynamicPCA<TestDataStructure>  pca;
      pca.computeSVD(input, u, singularValues,v);

      DynamicVector<TestDataStructure> eigenV= p.calcEigenValues();
      bool test=(eigenV==singularValues);

   }


};



/*
  //================SVD TEST======================

   // array 1..1000 anlegen
   card32* idx= new card32[dimension];

   // mit zahlen 1,2,3,...
   for (card32 n=0; n<dimension; n++)
      idx[n]=n;


   for (card32 i=0; i<dimension; i++){
      // dann: index wrfeln von i..n
      card32 newIdx= i + (int)(rnd01()*(float)(dimension-1-i));

      // pos i mit pos rnd-index vertauschen
      card32 temp= idx[i];
      idx[i]=idx[newIdx];
      idx[newIdx]=temp;

   }

      for (card32 j=0; j<dimension; j++) {
         for (card32 i=0; i<dimension; i++) {
            card32 index=idx[j];
            if (i == j) {
               cov[index][i] = (float)j+1;
            } else {
               cov[index][i] = 0.0;
            }
         }
      }

   //================SVD TEST ENDE======================

*/

// ============= OLD TESTS ==========================

     /*
   Vector3f mean=makeVector3f(0,0,0);
   for (card32 n=0; n<testNum; n++){
      mean+=testSet3f[n];
   }
   mean/=(float)testNum;

   Matrix3f sigma;


   float covXX ; float covXY ; float covXZ ;
   float covYY ; float covYZ ; float covZZ ;

   for (card32 n=0; n<testNum; n++){

         Vector3f pos = testSet3f[n];

         covXX=(pos[0]-mean[0])*(pos[0]-mean[0]);
         covXY=(pos[0]-mean[0])*(pos[1]-mean[1]);
         covXZ=(pos[0]-mean[0])*(pos[2]-mean[2]);
         covYY=(pos[1]-mean[1])*(pos[1]-mean[1]);
         covYZ=(pos[1]-mean[1])*(pos[2]-mean[2]);
         covZZ=(pos[2]-mean[2])*(pos[2]-mean[2]);

         sigma[0]+=makeVector3f(covXX,covXY,covXZ);
         sigma[1]+=makeVector3f(covXY,covYY,covYZ);
         sigma[2]+=makeVector3f(covXZ,covYZ,covZZ);
      }




   sigma/=testNum;


            Matrix3f cov = sigma;
            Vector3f eigenValV;

            // Result
            Matrix3f eigenVecM;

            // Prepare OpenCV Datasructure
            CvMat covM,eigenvalueM,eigenvectorM;
            cvInitMatHeader( &covM, 3, 3, CV_32FC1, cov.data());
            cvInitMatHeader( &eigenvalueM, 3, 1, CV_32FC1, eigenValV.data());
            cvInitMatHeader( &eigenvectorM, 3, 3, CV_32FC1, eigenVecM.data());

            // Do SVD
            cvSVD(&covM,&eigenvalueM,&eigenvectorM,0, CV_SVD_MODIFY_A | CV_SVD_U_T);





   */












