//---------------------------------------------------------------------------
#ifndef FilterMaskH
#define FilterMaskH
//---------------------------------------------------------------------------
#include "x4math.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
//---------------------------------------------------------------------------
#include "math_include.h"
//---------------------------------------------------------------------------

namespace X4
{

/**
   2d filter mask for 2 dimensional arrays
   no filtering is done in THIS class !
   use this class for big matrices with many zeros
*/
class X4MATH_API FilterMask : public Persistent
{
private:
   static MetaClass *FilterMaskMetaClass;

public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

   struct Element
   {
      int x,y;       // position of sample
      float weight;  // weight of sample
   };

   /// stores the first element of the filter mask in ret
   virtual void getFirstSample( Element &ret ) = 0;

   /// if there is a next sample, then true is returned and
   /// the element is stored in ret
   /// if not, return false + nothing happens to ret
   virtual bool getNextSample( Element &ret ) = 0;

};

/**
	[dim x dim] dynamic filter mask
*/

class X4MATH_API DynamicFilterMask : public FilterMask
{
private:
	std::vector<float> m_Mask;
	int32	m_Dim;
	card32	m_ActSampleIndex;
public:

	DynamicFilterMask() {m_ActSampleIndex=0;}

	void setup( card32 dim ) {m_Mask = std::vector<float>(dim*dim,0); m_Dim = dim; setElement(0,0,1.0f);}

	void setElement( int x, int y, float weight )
	{
		m_Mask[ (x+m_Dim/2) + (y+m_Dim/2)*m_Dim ] = weight;
	}

	virtual void getFirstSample( Element &ret )
	{
		m_ActSampleIndex = 1;
		ret.x = -m_Dim/2;
		ret.y = -m_Dim/2;
		ret.weight = m_Mask[0];
	}

   virtual bool getNextSample( Element &ret )
	{
     x4Assert( m_ActSampleIndex != 0 );   // call getFirstSample() before calling getNextSample() !
      if( (int)m_ActSampleIndex < m_Dim*m_Dim )
      {
         int x = m_ActSampleIndex % m_Dim;
         int y = m_ActSampleIndex / m_Dim;
         ret.weight = m_Mask[m_ActSampleIndex];
         ret.x = x - m_Dim/2;
         ret.y = y - m_Dim/2;
         m_ActSampleIndex++;
         return true;
      }
      else
         return false;

	}

   void makeVerticalGauss( float sigma )
   {
		m_Mask = std::vector<float>(m_Dim*m_Dim,0);
      for( int y = 0; y<m_Dim; y++ )
      {
         float fx2 = (0 - m_Dim/2)*(0 - m_Dim/2);
         float fy2 = (y - m_Dim/2)*(y - m_Dim/2);
         m_Mask[m_Dim/2+y*m_Dim] = 1.0 / sqrt( 2.0 * M_PI * sigma ) *
            exp( - (fx2 + fy2) / (sigma*sigma) );
      }
   }

   void makeHorizontalGauss( float sigma )
   {
		m_Mask = std::vector<float>(m_Dim*m_Dim,0);
      for( int x = 0; x<m_Dim; x++ )
      {
         float fx2 = (x - m_Dim/2)*(x - m_Dim/2);
         float fy2 = (0 - m_Dim/2)*(0 - m_Dim/2);
         m_Mask[x+m_Dim/2*m_Dim] = 1.0 / sqrt( 2.0 * M_PI * sigma ) *
            exp( - (fx2 + fy2) / (sigma*sigma) );
      }
   }

};

/**
   [dim x dim] filter mask
*/
template<int dim> class X4MATH_API TemplatedFixedFilterMask : public FilterMask
{
protected:
   float32     m_FilterMask[dim*dim];
   card32      m_ActSampleIndex;

public:

   TemplatedFixedFilterMask()
   {
      for( int x = 0; x<dim*dim; x++ )
         m_FilterMask[x] = 0.0f;
      m_FilterMask[dim/2 + dim*dim/2] = 1.0f;
      m_ActSampleIndex = 0;
   }

   void getFirstSample( Element &ret )
   {
      ret.x = -dim/2;
      ret.y = -dim/2;
      ret.weight = m_FilterMask[0];
      m_ActSampleIndex = 1;

   }

   bool getNextSample( Element &ret )
   {
      x4Assert( m_ActSampleIndex != 0 );   // call getFirstSample() before calling getNextSample() !
      if( m_ActSampleIndex < dim*dim )
      {
         int x = m_ActSampleIndex % dim;
         int y = m_ActSampleIndex / dim;
         ret.weight = m_FilterMask[m_ActSampleIndex];
         ret.x = x - dim/2;
         ret.y = y - dim/2;
         m_ActSampleIndex++;
         return true;
      }
      else
         return false;

   }

   void setElement( int x, int y, float w );


   /// creates a gauss filter mask
   /// weight = 1.0 / sqrt( 2.0 * PI * sigma ) * exp( - (x^2+ y^2 / (2*sigma^2 )
   void makeGauss( float sigma )
   {
      for( int x = 0; x<dim; x++ )
         for( int y = 0; y<dim; y++ )
         {
            float fx2 = (x - dim/2)*(x - dim/2);
            float fy2 = (y - dim/2)*(y - dim/2);
            m_FilterMask[x+y*dim] = 1.0 / sqrt( 2.0 * M_PI * sigma ) *
               exp( - (fx2 + fy2) / (2.0*sigma*sigma) );
         }
   }

   /// creates a box mask ( every weight = 1 )
   void makeBox()
   {
      for( int x = 0; x<dim*dim; x++ )
      {
         m_FilterMask[x] = 1.0f;
      }
   }

   // TODO: makeTriangle, makeDoG, ...
};


class X4MATH_API FixedFilterMask_9x9 : public TemplatedFixedFilterMask<9>
{
private:
   static MetaClass *FixedFilterMask_9x9MetaClass;
   static Object* newInstance();


public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();
   
};

class X4MATH_API FixedFilterMask_21x21 : public TemplatedFixedFilterMask<21>
{
private:
   static MetaClass *FixedFilterMask_21x21MetaClass;
   static Object* newInstance();


public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

};

}

#endif
 