//---------------------------------------------------------------------------
#ifndef KMeansClusteringH
#define KMeansClusteringH
//---------------------------------------------------------------------------
#include "x4math.h"
//---------------------------------------------------------------------------
#include "DynamicLinearAlgebra.h"
//---------------------------------------------------------------------------
#include <vector>
//---------------------------------------------------------------------------

namespace X4 {

/**
	very simple, straightforward, unoptimized KMeansClustering

	update mb: added sample weights
*/					
template <typename FloatType, typename VectorType>
class KMeansClusteringTempl {
 protected:

	std::vector< VectorType > data;
	std::vector< float > weights;
	bool initWithData;

	void chooseNearest(std::vector<unsigned> &indices, 
							 const std::vector< VectorType > &data, 
							 const std::vector< VectorType > &result);
	FloatType updateResult(const std::vector<unsigned> &indices, 
	                       const std::vector< VectorType > &data, 
	                       std::vector< VectorType > &result);
	FloatType calcQuality(const std::vector<unsigned> &indices, 
	                      const std::vector< VectorType > &data, 
	                      std::vector< VectorType > &result);
	
 public:
	KMeansClusteringTempl() { initWithData = false; }

	/// add data points; needs at least one
	void add(const VectorType &v, float weight=1.0f);

	/// to be called before calc
	/// initWithData: initialize means with random samples from the data
	void setup(bool initWithData = false);
	
	/// get number of data points
	mpcard getDataSize() {return data.size();}
	
	/// retrieve data (in "add" order)
	VectorType getData(unsigned i) {return data[i];}
		
	/// compute clustering, iterate until results do not move by more than relConvThreshold times the bounding box diagonal
	/// also stops after MAX_NUM_ITERATIONS = 100 iterations
	/// numTrials corresponds to the number of random initializations that are tried (the best, i.e. least sqaure distance sum, is returned)
	void calc(card32 numClusters, std::vector< VectorType > &result, FloatType relConvThreshold, card32 numTrials);

	/// same as calc, but provides also indices relating the result to the input
	void calcEx(card32 numClusters, std::vector< VectorType > &result, std::vector<unsigned> &indices, FloatType relConvThreshold, card32 numTrials);

	/// same as calcEx, but only one try
	void calcSingleTry(card32 numClusters, std::vector< VectorType > &result, std::vector<unsigned> &indices, FloatType relConvThreshold);

	// returns (weighted) intra cluster variances
	void calcIntraClusterVariances( const std::vector<unsigned> &indices, const std::vector< VectorType > &result, std::vector< FloatType > &variances);

	~KMeansClusteringTempl() {}
};		


template <typename FloatType>
class KMeansClustering : public KMeansClusteringTempl<FloatType, DynamicVector<FloatType> > {};


class X4MATH_API KMeansClustering3f : public KMeansClusteringTempl<float32, Vector3f> {
 private:
   void prequantize(card32 res = 8);
 public:
   void calcQuantized(card32 numClusters, std::vector<Vector3f> &centroids, float32 relConvThreshold, card32 numTrials, card32 res = 8);
};


}

#endif