//---------------------------------------------------------------------------
#ifndef LinearAlgebraH
#define LinearAlgebraH
//---------------------------------------------------------------------------
#include <x4math.h>
//---------------------------------------------------------------------------
#include <X4Types.h>
#include <x4Types.h>
#include <string>
#include "X4TypeNames.h"
//---------------------------------------------------------------------------

namespace X4 {

typedef float32 Degree;
typedef float32 Radian;

class OutputObjectStream;
class InputObjectStream;


// ---------------------------------------------------------------------------
//
//                                 StaticVector
//
// ---------------------------------------------------------------------------


template <class FloatType, unsigned dim>
class StaticVector {
private:
	FloatType v[dim];
public:
	inline FloatType& operator[](const unsigned &index);
	inline const FloatType& operator[](const unsigned &index) const;

	//  ----  operators +, -, *, /, crossProduct
	inline StaticVector<FloatType, dim> operator+(const StaticVector<FloatType, dim> &op) const;
	inline StaticVector<FloatType, dim> operator-(const StaticVector<FloatType, dim> &op) const;
	inline StaticVector<FloatType, dim> operator-() const;
	inline FloatType operator*(const StaticVector<FloatType, dim> &op) const;
	inline StaticVector<FloatType, dim> operator*(const FloatType &s) const;
	inline StaticVector<FloatType, dim> operator/(const FloatType &s) const;
	StaticVector<FloatType, dim> crossProduct(const StaticVector<FloatType, dim> &op) const;
	inline StaticVector<FloatType, dim> componentProduct(const StaticVector<FloatType, dim> &op) const;
	inline StaticVector<FloatType, dim> componentDivision(const StaticVector<FloatType, dim> &op) const;

	//  ---- operators +=, -=, *=, /=
	inline StaticVector<FloatType, dim> operator+=(const StaticVector<FloatType, dim> &op);
	inline StaticVector<FloatType, dim> operator-=(const StaticVector<FloatType, dim> &op);
	inline StaticVector<FloatType, dim> operator*=(const FloatType &s);
	inline StaticVector<FloatType, dim> operator/=(const FloatType &s);

	//  ---- operators =, ==, !=
	inline bool operator==(const StaticVector<FloatType, dim> &op) const;
	inline bool operator!=(const StaticVector<FloatType, dim> &op) const;

	//  ---- misc.
	void print();
	std::string toString() const;
	std::string toTypeShortString() const;
	inline FloatType* data();
	inline const FloatType* data() const;
	inline void normalize();
	inline FloatType getSqrNorm() const;
	inline FloatType getNorm() const;
	inline void setZero();

	/// compatibility with DynamicVector
	static inline unsigned getDim() {return dim;}
	static inline unsigned size() {return dim;}
	inline const FloatType* begin() const { return data(); }
	inline FloatType* begin() { return data(); }
	inline const FloatType* end() const { return begin() + size(); }
	inline FloatType* end() { return begin() + size(); }

	/// compatibility with persistent methods
	void write(OutputObjectStream *s) const;
	void read(InputObjectStream *s);	
};

template<typename FloatType, int dim>
inline StaticVector<FloatType,dim> operator*(const FloatType &s, const StaticVector<FloatType,dim>& m) {
	return m*s;
}



// ---------------------------------------------------------------------------
//
//                                 StaticMatrix
//
// ---------------------------------------------------------------------------


template <class FloatType, unsigned columns, unsigned rows>
class StaticMatrix {
private:
	StaticVector<FloatType, rows> theColumns[columns];

public:
	// ---- constructors
	inline StaticMatrix<FloatType, columns, rows>();

	// ---- element access
	// Oliver: not clear whether this returns cols or rows - maybe replace in favor of a col(index) function
	inline StaticVector<FloatType, rows>& operator[](mpcard index);
	inline const StaticVector<FloatType, rows>& operator[](mpcard index) const;

	inline StaticVector<FloatType, rows>& column(mpcard index);
	inline const StaticVector<FloatType, rows>& column(mpcard index) const;

	//  ----  operators +, -, *, /
	inline StaticMatrix<FloatType, columns, rows> operator+(const StaticMatrix<FloatType, columns, rows> &op) const;
	inline StaticMatrix<FloatType, columns, rows> operator-(const StaticMatrix<FloatType, columns, rows> &op) const;
	inline StaticMatrix<FloatType, columns, rows> operator-() const;
	inline StaticMatrix<FloatType, columns, rows> operator*(const FloatType &s) const;
	/* This:    colums: columns (b)
	rows: rows (a)
	Operand: colums: colums2 (c)
	rows: columns (b)
	Result:  colums: columns2 (a)
	rows: rows (c)
	*/
	template <unsigned columns2>
	inline StaticMatrix<FloatType, columns2, rows> operator*( const StaticMatrix<FloatType, columns2, columns>& mat ) const;

	inline StaticMatrix<FloatType, columns, rows> operator/(const FloatType &s) const;

	//  ---- operators +=, -=, *=, /=
	inline StaticMatrix<FloatType, columns, rows> operator+=(const StaticMatrix<FloatType, columns, rows> &op);
	inline StaticMatrix<FloatType, columns, rows> operator-=(const StaticMatrix<FloatType, columns, rows> &op);
	inline StaticMatrix<FloatType, columns, rows> operator*=(const FloatType &op);
	inline StaticMatrix<FloatType, columns, rows> operator*=(const StaticMatrix<FloatType, columns, rows> &op);
	inline StaticMatrix<FloatType, columns, rows> operator/=(const FloatType &op);

	//  ---- operators =, ==, !=
	inline StaticMatrix<FloatType, columns, rows> operator=(const StaticMatrix<FloatType, columns, rows> &op);
	inline bool operator==(const StaticMatrix<FloatType, columns, rows> &op) const;
	inline bool operator!=(const StaticMatrix<FloatType, columns, rows> &op) const;

	// ---- multiplying with vectors
	inline StaticVector<FloatType, rows> operator*(const StaticVector<FloatType, columns> &v) const;

	// ----  misc.
	inline void print();
	std::string toString() const;
	inline FloatType* data();
	inline const FloatType* data() const;
	inline void changeRows(unsigned row1, unsigned row2);
	inline void multRow(unsigned row, FloatType value);
	inline void combineRows(unsigned row, unsigned with, FloatType by);
	inline void addRows(unsigned row, unsigned with);
	inline StaticMatrix<FloatType, rows, columns> transpose() const;
	inline FloatType getDeterminant() const;
	inline void setZero();
	inline void setIdentity();

	/// Sort eigenvectors according to eigenvalues in descending order
	void eigenSort( StaticVector<FloatType,rows>& eigenValues, StaticMatrix<FloatType,rows,columns>& eigenVectors ) const;
	/// Compute the eigenstructure (eigenvalues and eigenvectors) of a symmetrical matrix (Numerical Recepies).
	/// The normalized eigenvectors are found as the rows of the eigenVectors-matrix.
	void computeEigenStructure( StaticVector<FloatType,rows>& eigenValues, StaticMatrix<FloatType,rows,columns>& eigenVectors ) const;

	/// compatibility with DynamicVector
	static inline unsigned getRowsDim() {return rows;}
	static inline unsigned getColsDim() {return columns;}
	static inline unsigned getRows()    {return rows;}
	static inline unsigned getColumns() {return columns;}
};

template<typename FloatType, int dim1, int dim2>
inline StaticMatrix<FloatType, dim1, dim2> operator*(const FloatType &s, const StaticMatrix<FloatType, dim1, dim2>& m) {
	return m*s;
}



// ---------------------------------------------------------------------------
//
//                                  instances
//
// ---------------------------------------------------------------------------


typedef StaticVector<float,1> Vector1f;  PROVIDE_MODULE_TYPE(X4MATH_API, Vector1f);
typedef StaticVector<float,2> Vector2f;  PROVIDE_MODULE_TYPE(X4MATH_API, Vector2f);
typedef StaticVector<float,3> Vector3f;  PROVIDE_MODULE_TYPE(X4MATH_API, Vector3f);
typedef StaticVector<float,4> Vector4f;  PROVIDE_MODULE_TYPE(X4MATH_API, Vector4f);
typedef StaticVector<float,6> Vector6f;  PROVIDE_MODULE_TYPE(X4MATH_API, Vector6f);
typedef StaticVector<double,1> Vector1d; PROVIDE_MODULE_TYPE(X4MATH_API, Vector1d);
typedef StaticVector<double,2> Vector2d; PROVIDE_MODULE_TYPE(X4MATH_API, Vector2d);
typedef StaticVector<double,3> Vector3d; PROVIDE_MODULE_TYPE(X4MATH_API, Vector3d);
typedef StaticVector<double,4> Vector4d; PROVIDE_MODULE_TYPE(X4MATH_API, Vector4d);
typedef StaticVector<double,6> Vector6d; PROVIDE_MODULE_TYPE(X4MATH_API, Vector6d);
typedef StaticVector<int32,1> Vector1i;  PROVIDE_MODULE_TYPE(X4MATH_API, Vector1i);
typedef StaticVector<int32,2> Vector2i;  PROVIDE_MODULE_TYPE(X4MATH_API, Vector2i);
typedef StaticVector<int32,3> Vector3i;  PROVIDE_MODULE_TYPE(X4MATH_API, Vector3i);
typedef StaticVector<int32,4> Vector4i;  PROVIDE_MODULE_TYPE(X4MATH_API, Vector4i);
typedef StaticVector<card32,1> Vector1u; PROVIDE_MODULE_TYPE(X4MATH_API, Vector1u);
typedef StaticVector<card32,2> Vector2u; PROVIDE_MODULE_TYPE(X4MATH_API, Vector2u);
typedef StaticVector<card32,3> Vector3u; PROVIDE_MODULE_TYPE(X4MATH_API, Vector3u);
typedef StaticVector<card32,4> Vector4u; PROVIDE_MODULE_TYPE(X4MATH_API, Vector4u);
typedef StaticVector<card16,4> Vector4us;PROVIDE_MODULE_TYPE(X4MATH_API, Vector4us);
typedef StaticVector<card8,1> Vector1ub; PROVIDE_MODULE_TYPE(X4MATH_API, Vector1ub);
typedef StaticVector<card8,2> Vector2ub; PROVIDE_MODULE_TYPE(X4MATH_API, Vector2ub);
typedef StaticVector<card8,3> Vector3ub; PROVIDE_MODULE_TYPE(X4MATH_API, Vector3ub);
typedef StaticVector<card8,4> Vector4ub; PROVIDE_MODULE_TYPE(X4MATH_API, Vector4ub);
typedef StaticVector<int8,1> Vector1b;   PROVIDE_MODULE_TYPE(X4MATH_API, Vector1b);
typedef StaticVector<int8,2> Vector2b;   PROVIDE_MODULE_TYPE(X4MATH_API, Vector2b);
typedef StaticVector<int8,3> Vector3b;   PROVIDE_MODULE_TYPE(X4MATH_API, Vector3b);
typedef StaticVector<int8,4> Vector4b;   PROVIDE_MODULE_TYPE(X4MATH_API, Vector4b);

//typedef Vector1f float1;
//typedef Vector2f float2;
//typedef Vector3f float3;
//typedef Vector4f float4;
//
//typedef Vector1d double1;
//typedef Vector2d double2;
//typedef Vector3d double3;
//typedef Vector4d double4;
//
//typedef Vector1i int1;
//typedef Vector2i int2;
//typedef Vector3i int3;
//typedef Vector4i int4;
//
//typedef Vector1u uint1;
//typedef Vector2u uint2;
//typedef Vector3u uint3;
//typedef Vector4u uint4;
//
//typedef Vector1ub uchar1;
//typedef Vector2ub uchar2;
//typedef Vector3ub uchar3;
//typedef Vector4ub uchar4;
//
//typedef Vector4us ushort4;
//
//typedef Vector1b char1;
//typedef Vector2b char2;
//typedef Vector3b char3;
//typedef Vector4b char4;

typedef StaticMatrix<float,2,2> Matrix2f;
typedef StaticMatrix<float,3,3> Matrix3f;
typedef StaticMatrix<float,4,4> Matrix4f;
typedef StaticMatrix<float,6,6> Matrix6f;
typedef StaticMatrix<double,2,2> Matrix2d;
typedef StaticMatrix<double,3,3> Matrix3d;
typedef StaticMatrix<double,4,4> Matrix4d;
typedef StaticMatrix<double,6,6> Matrix6d;

// ---------------------------------------------------------------------------
//
//                                 misc. functions
//
// ---------------------------------------------------------------------------


template <class FloatTypeDest, class FloatTypeSource, int Dim>
StaticVector<FloatTypeDest, Dim> convertVector(const StaticVector<FloatTypeSource, Dim> &source);



template<typename FloatType>
inline StaticVector<FloatType,1> makeVector1(FloatType x) {
	StaticVector<FloatType,1> v;
	v[0] = x;
	return v;
}

template<typename FloatType>
inline StaticVector<FloatType,2> makeVector2(FloatType x1, FloatType x2) {
	StaticVector<FloatType,2> v;
	v[0] = x1; v[1] = x2;
	return v;
}

template<typename FloatType>
inline StaticVector<FloatType,3> makeVector3(FloatType x1, FloatType x2, FloatType x3) {
	StaticVector<FloatType,3> v;
	v[0] = x1; v[1] = x2; v[2] = x3;
	return v;
}

template<typename FloatType>
inline StaticVector<FloatType,4> makeVector4(FloatType x1, FloatType x2, FloatType x3, FloatType x4) {
	StaticVector<FloatType,4> v;
	v[0] = x1; v[1] = x2; v[2] = x3; v[3] = x4;
	return v;
}

template<typename FloatType>
inline StaticVector<FloatType,6> makeVector6(FloatType x1, FloatType x2, FloatType x3, FloatType x4, FloatType x5, FloatType x6) {
	StaticVector<FloatType,6> v;
	v[0] = x1; v[1] = x2; v[2] = x3; v[3] = x4; v[4] = x5; v[5] = x6;
	return v;
}

template<typename Iterator,int dim>  // implies pointers
inline StaticVector<typename std::iterator_traits<Iterator>::value_type,dim> makeVectorN(Iterator it) { 
	StaticVector<typename std::iterator_traits<Iterator>::value_type,dim> v;
	for( int i1 = 0; i1 < dim; ++ i1 ) {
		v[i1] = *it;
		++ it;
	}
	return v;
}
template<typename Iterator>  // implies pointers
inline StaticVector<typename std::iterator_traits<Iterator>::value_type,1> makeVector1(Iterator it) { return makeVectorN<Iterator,1>(it); }
template<typename Iterator>
inline StaticVector<typename std::iterator_traits<Iterator>::value_type,2> makeVector2(Iterator it) { return makeVectorN<Iterator,2>(it); }
template<typename Iterator>
inline StaticVector<typename std::iterator_traits<Iterator>::value_type,3> makeVector3(Iterator it) { return makeVectorN<Iterator,3>(it); }
template<typename Iterator>
inline StaticVector<typename std::iterator_traits<Iterator>::value_type,4> makeVector4(Iterator it) { return makeVectorN<Iterator,4>(it); }
template<typename Iterator>
inline StaticVector<typename std::iterator_traits<Iterator>::value_type,6> makeVector6(Iterator it) { return makeVectorN<Iterator,6>(it); }




inline Vector1i makeVector1i(int x);
inline Vector1u makeVector1u(unsigned x);
inline Vector1f makeVector1f(float x);
inline Vector1d makeVector1d(double x);
inline Vector1ub makeVector1ub(card8 x);

inline Vector2i makeVector2i(int x, int y);
inline Vector2u makeVector2u(unsigned x, unsigned y);
inline Vector2f makeVector2f(float x, float y);
inline Vector2d makeVector2d(double x, double y);
inline Vector2ub makeVector2ub(card8 x, card8 y);

inline Vector3f makeVector3f(float x, float y, float z);
inline Vector3d makeVector3d(double x, double y, double z);
inline Vector3i makeVector3i(int x, int y, int z);
inline Vector3u makeVector3u(unsigned x, unsigned y, unsigned z);
inline Vector3ub makeVector3ub(card8 x, card8 y, card8 z);

inline Vector4f makeVector4f(float x, float y, float z, float w);
inline Vector4i makeVector4i(int x, int y, int z, int w);
inline Vector4u makeVector4u(unsigned x, unsigned y, unsigned z, unsigned w);
inline Vector4d makeVector4d(double x, double y, double z, double w);
inline Vector4ub makeVector4ub(card8 x, card8 y, card8 z, card8 w);

inline Vector6f makeVector6f(float x, float y, float z, float u, float v, float w);
inline Vector6d makeVector6d(double x, double y, double z, double u, double v, double w);

template<class FloatType,unsigned dim>
inline StaticVector<FloatType,dim> makeVector( const FloatType *data );

inline Vector2f makeVector2fv( const float32 * data );
inline Vector3f makeVector3fv( const float32 * data );
inline Vector4f makeVector4fv( const float32 * data );


inline Matrix2f makeMatrix2f(float m00, float m01,
                             float m10, float m11);

inline Matrix3f makeMatrix3f(float m00, float m10, float m20,
                             float m01, float m11, float m21,
                             float m02, float m12, float m22);

inline Matrix4f makeMatrix4f(float m00, float m10, float m20, float m30,
                             float m01, float m11, float m21, float m31,
                             float m02, float m12, float m22, float m32,
                             float m03, float m13, float m23, float m33);

inline Matrix6f makeMatrix6f(float m00, float m10, float m20, float m30, float m40, float m50,
                             float m01, float m11, float m21, float m31, float m41, float m51,
                             float m02, float m12, float m22, float m32, float m42, float m52,
                             float m03, float m13, float m23, float m33, float m43, float m53,
							 float m04, float m14, float m24, float m34, float m44, float m54,
							 float m05, float m15, float m25, float m35, float m45, float m55);

inline Vector4f expand3To4(Vector3f v);
inline Matrix4f expand3To4(Matrix3f v);
inline Vector3f shrink4To3(Vector4f v);
inline Matrix3f shrink4To3(Matrix4f v);

inline Vector3f expand2To3(const Vector2f &v);
inline Matrix3f expand2To3(const Matrix2f &v);
inline Vector2f shrink3To2(const Vector3f &v);
inline Matrix2f shrink3To2(const Matrix3f &v);
inline Vector3f projectHomogeneous3To2(Vector4f v);
inline Vector3f projectHomogeneous4To3(Vector4f v);

template <class FloatType, unsigned dim>
inline StaticMatrix<FloatType,dim,dim> outerProduct(const StaticVector<FloatType,dim> &v1, const StaticVector<FloatType,dim> &v2);

template <class FloatType, unsigned dim>
inline FloatType normQuad(StaticVector<FloatType, dim> v);

template <class FloatType, unsigned dim>
inline FloatType norm(StaticVector<FloatType, dim> v);

/// squared frobenius norm of a matrix
template <class FloatType, unsigned dim>
inline FloatType frobeniusNormQuad(const StaticMatrix<FloatType,dim, dim> &M);

template <class FloatType, unsigned dim>
inline FloatType frobeniusNorm(const StaticMatrix<FloatType,dim, dim> &M);


template <class FloatType, unsigned dim>
inline StaticVector<FloatType, dim> normalize(const StaticVector<FloatType, dim> &v);


inline Matrix3f makeRotX3f(Radian angle);
inline Matrix4f makeRotX4f(Radian angle);

inline Matrix3f makeRotY3f(Radian angle);
inline Matrix4f makeRotY4f(Radian angle);

inline Matrix3f makeRotZ3f(Radian angle);
inline Matrix4f makeRotZ4f(Radian angle);

inline Matrix3f makeRotVector3f(Vector3f vec, float32 angle);
inline Matrix4f makeRotVector4f(Vector3f vec, float32 angle);

inline Matrix4f makeTranslation4f(const Vector3f &t);
inline Matrix4f makeTranslation4f(float32 x, float32 y, float32 z);
inline Matrix3f makeTranslation3f(const Vector2f &t);
inline Matrix3f makeTranslation3f(float32 x, float32 y);

inline Matrix3f makeScale3f(const Vector3f &s);
inline Matrix4f makeScale4f(const Vector3f &s);
inline Matrix3f makeScale3f(float32 x, float32 y, float32 z);
inline Matrix4f makeScale4f(float32 x, float32 y, float32 z);

inline Vector3f transformVector3f(const Matrix4f &M, const Vector3f &v);
inline Vector2f transformVector2f(const Matrix3f &M, const Vector2f &v);
inline Vector2f projectiveTransformVector2f(const Matrix3f &M, const Vector2f &v);
inline Vector3f projectiveTransformVector3f(const Matrix4f &M, const Vector3f &v);

inline Matrix3f makeNewCoordPrjMatrix3f(Vector3f v1, Vector3f v2, Vector3f v3);


template<class FloatType, unsigned size>
inline unsigned searchPivot(const StaticMatrix<FloatType, size, size> &mat, unsigned step, bool *success = NULL, FloatType pivotEps = (FloatType)0);

/// if success is not NULL, true is returned if inversion was successfull and false if the matrix was found to be singular
/// otherwise, an exception is thrown in case of singularity
template<class FloatType, unsigned size>
inline StaticMatrix<FloatType, size, size> invertMatrix(const StaticMatrix<FloatType, size, size> &mat, bool *success = NULL, FloatType pivotEps = (FloatType)0);


/// fast invert (transpose rotational part and negate translational part)
inline Matrix4f invertFrame( const Matrix4f &m);
inline Matrix3f invertFrame( const Matrix3f &m);

template<class FloatType1, class FloatType2, unsigned dim>
inline void convertMatrix( const StaticMatrix<FloatType1,dim, dim> &source, StaticMatrix<FloatType2,dim, dim> &dest );

// ---------------------------------------------------------------------------
//
//                             predefined constants
//
// ---------------------------------------------------------------------------


extern X4MATH_API const Vector1ub NULL_VECTOR1UB;
extern X4MATH_API const Vector2ub NULL_VECTOR2UB;
extern X4MATH_API const Vector3ub NULL_VECTOR3UB;
extern X4MATH_API const Vector4ub NULL_VECTOR4UB;

extern X4MATH_API const Vector1i NULL_VECTOR1I;
extern X4MATH_API const Vector2i NULL_VECTOR2I;
extern X4MATH_API const Vector3i NULL_VECTOR3I;
extern X4MATH_API const Vector4i NULL_VECTOR4I;

extern X4MATH_API const Vector1f NULL_VECTOR1F;
extern X4MATH_API const Vector2f NULL_VECTOR2F;
extern X4MATH_API const Vector3f NULL_VECTOR3F;
extern X4MATH_API const Vector4f NULL_VECTOR4F;
extern X4MATH_API const Vector6f NULL_VECTOR6F;

extern X4MATH_API const Vector4f HOMOGENEOUS_NULL_VECTOR4F;

extern X4MATH_API const Vector1d NULL_VECTOR1D;
extern X4MATH_API const Vector2d NULL_VECTOR2D;
extern X4MATH_API const Vector3d NULL_VECTOR3D;
extern X4MATH_API const Vector4d NULL_VECTOR4D;
extern X4MATH_API const Vector6d NULL_VECTOR6D;

extern X4MATH_API const Vector3f XAXIS_VECTOR3F;
extern X4MATH_API const Vector3f YAXIS_VECTOR3F;
extern X4MATH_API const Vector3f ZAXIS_VECTOR3F;

extern X4MATH_API const Vector4f XAXIS_VECTOR4F;
extern X4MATH_API const Vector4f YAXIS_VECTOR4F;
extern X4MATH_API const Vector4f ZAXIS_VECTOR4F;

extern X4MATH_API const Matrix2f IDENTITY2F;
extern X4MATH_API const Matrix3f IDENTITY3F;
extern X4MATH_API const Matrix4f IDENTITY4F;
extern X4MATH_API const Matrix6f IDENTITY6F;
extern X4MATH_API const Matrix3d IDENTITY3D;


// -------------------------- console output -------------------------------


// extern template DebugOutput& DebugOutput::operator<<(const StaticVector<float,1> &v);
// extern template DebugOutput& DebugOutput::operator<<(const StaticVector<double,1> &v);
// extern template DebugOutput& DebugOutput::operator<<(const StaticVector<float,2> &v);
// extern template DebugOutput& DebugOutput::operator<<(const StaticVector<double,2> &v);
// extern template DebugOutput& DebugOutput::operator<<(const StaticVector<float,3> &v);
// extern template DebugOutput& DebugOutput::operator<<(const StaticVector<double,3> &v);
// extern template DebugOutput& DebugOutput::operator<<(const StaticVector<float,4> &v);
// extern template DebugOutput& DebugOutput::operator<<(const StaticVector<double,4> &v);
// extern template DebugOutput& DebugOutput::operator<<(const StaticVector<float,6> &v);
// extern template DebugOutput& DebugOutput::operator<<(const StaticVector<double,6> &v);
// 
// 
// extern template DebugOutput& DebugOutput::operator<<(const StaticMatrix<float ,1,1> &v);
// extern template DebugOutput& DebugOutput::operator<<(const StaticMatrix<double,1,1> &v);
// extern template DebugOutput& DebugOutput::operator<<(const StaticMatrix<float ,2,2> &v);
// extern template DebugOutput& DebugOutput::operator<<(const StaticMatrix<double,2,2> &v);
// extern template DebugOutput& DebugOutput::operator<<(const StaticMatrix<float ,3,3> &v);
// extern template DebugOutput& DebugOutput::operator<<(const StaticMatrix<double,3,3> &v);
// extern template DebugOutput& DebugOutput::operator<<(const StaticMatrix<float ,4,4> &v);
// extern template DebugOutput& DebugOutput::operator<<(const StaticMatrix<double,4,4> &v);
// extern template DebugOutput& DebugOutput::operator<<(const StaticMatrix<float ,6,6> &v);
// extern template DebugOutput& DebugOutput::operator<<(const StaticMatrix<double,6,6> &v);
// 
// 
// extern template DebugOutput& DebugOutput::operator<<(const DynamicVector<float> &v);
// extern template DebugOutput& DebugOutput::operator<<(const DynamicVector<double> &v);
// 
// 
// extern template DebugOutput& DebugOutput::operator<<(const DynamicMatrix<float> &m);
// extern template DebugOutput& DebugOutput::operator<<(const DynamicMatrix<double> &m);



// --------------- Properties Macros --------------------------------------
#define ADD_VECTOR1UB_PROP_UPDATE( propName, version, updateMethodPtr ) \
	((MetaClass*)getClass())->addProperty( new FixedArrayClassProperty( #propName , MetaClass::FIRST_VERSION + version, \
	(pointer_diff)((pointer_diff)&( obj_null_pointer->##propName ) - obj_null_offset) , new Card8ClassProperty("", version , NULL, (Card8UpdateMethod)updateMethodPtr) \
	,1, sizeof(card8), 1 ));

#define ADD_VECTOR2F_PROP_UPDATE( propName, version, updateMethodPtr ) \
	((MetaClass*)getClass())->addProperty( new FixedArrayClassProperty( #propName , MetaClass::FIRST_VERSION + version, \
	(pointer_diff)((pointer_diff)&( obj_null_pointer->##propName ) - obj_null_offset) , new Float32ClassProperty("", version , NULL, (Float32UpdateMethod)updateMethodPtr) \
	,2, sizeof(float32), 2 ));

#define ADD_VECTOR2D_PROP_UPDATE( propName, version, updateMethodPtr ) \
   ((MetaClass*)getClass())->addProperty( new FixedArrayClassProperty( #propName , MetaClass::FIRST_VERSION + version, \
   (pointer_diff)((pointer_diff)&( obj_null_pointer->##propName ) - obj_null_offset) , new Float64ClassProperty("", version , NULL, (Float64UpdateMethod)updateMethodPtr) \
   ,2, sizeof(float64), 2 ));

#define ADD_VECTOR2I_PROP_UPDATE( propName, version, updateMethodPtr ) \
	((MetaClass*)getClass())->addProperty( new FixedArrayClassProperty( #propName , MetaClass::FIRST_VERSION + version, \
	(pointer_diff)((pointer_diff)&( obj_null_pointer->##propName ) - obj_null_offset) , new Int32ClassProperty("", version , NULL, (Int32UpdateMethod)updateMethodPtr) \
	,2, sizeof(int32), 2 ));

#define ADD_VECTOR3F_PROP_UPDATE( propName, version, updateMethodPtr ) \
	((MetaClass*)getClass())->addProperty( new FixedArrayClassProperty( #propName , MetaClass::FIRST_VERSION + version, \
	(pointer_diff)((pointer_diff)&( obj_null_pointer->##propName ) - obj_null_offset) , new Float32ClassProperty("",version, NULL, (Float32UpdateMethod)updateMethodPtr) \
	,3, sizeof(float32), 3 ));

#define ADD_VECTOR3D_PROP_UPDATE( propName, version, updateMethodPtr ) \
   ((MetaClass*)getClass())->addProperty( new FixedArrayClassProperty( #propName , MetaClass::FIRST_VERSION + version, \
   (pointer_diff)((pointer_diff)&( obj_null_pointer->##propName ) - obj_null_offset) , new Float64ClassProperty("",version, NULL, (Float64UpdateMethod)updateMethodPtr) \
   ,3, sizeof(float64), 3 ));


#define ADD_VECTOR3I_PROP_UPDATE( propName, version, updateMethodPtr ) \
	((MetaClass*)getClass())->addProperty( new FixedArrayClassProperty( #propName , MetaClass::FIRST_VERSION + version, \
	(pointer_diff)((pointer_diff)&( obj_null_pointer->##propName ) - obj_null_offset) , new Int32ClassProperty("",version, NULL, (Int32UpdateMethod)updateMethodPtr) \
	,3, sizeof(float32), 3 ));

#define ADD_VECTOR4F_PROP_UPDATE( propName, version, updateMethodPtr ) \
	((MetaClass*)getClass())->addProperty( new FixedArrayClassProperty( #propName , MetaClass::FIRST_VERSION + version, \
	(pointer_diff)((pointer_diff)&( obj_null_pointer->##propName ) - obj_null_offset) , new Float32ClassProperty("",version, NULL, (Float32UpdateMethod)updateMethodPtr) \
	,4, sizeof(float32), 4 ));

#define ADD_VECTOR4D_PROP_UPDATE( propName, version, updateMethodPtr ) \
   ((MetaClass*)getClass())->addProperty( new FixedArrayClassProperty( #propName , MetaClass::FIRST_VERSION + version, \
   (pointer_diff)((pointer_diff)&( obj_null_pointer->##propName ) - obj_null_offset) , new Float64ClassProperty("",version, NULL, (Float64UpdateMethod)updateMethodPtr) \
   ,4, sizeof(float64), 4 ));

#define ADD_VECTOR4I_PROP_UPDATE( propName, version, updateMethodPtr ) \
	((MetaClass*)getClass())->addProperty( new FixedArrayClassProperty( #propName , MetaClass::FIRST_VERSION + version, \
	(pointer_diff)((pointer_diff)&( obj_null_pointer->##propName ) - obj_null_offset) , new Int32ClassProperty("",version, NULL, (Int32UpdateMethod)updateMethodPtr) \
	,4, sizeof(int32), 4 ));

#define ADD_VECTOR4U_PROP_UPDATE( propName, version, updateMethodPtr ) \
	((MetaClass*)getClass())->addProperty( new FixedArrayClassProperty( #propName , MetaClass::FIRST_VERSION + version, \
	(pointer_diff)((pointer_diff)&( obj_null_pointer->##propName ) - obj_null_offset) , new Card32ClassProperty("",version, NULL, (Card32UpdateMethod)updateMethodPtr) \
	,4, sizeof(card32), 4 ));

#define ADD_VECTOR6F_PROP_UPDATE( propName, version, updateMethodPtr ) \
	((MetaClass*)getClass())->addProperty( new FixedArrayClassProperty( #propName , MetaClass::FIRST_VERSION + version, \
	(pointer_diff)((pointer_diff)&( obj_null_pointer->##propName ) - obj_null_offset) , new Float32ClassProperty("",version, NULL, (Float32UpdateMethod)updateMethodPtr) \
	,6, sizeof(float32), 6 ));

#define ADD_MATRIX2F_PROP_UPDATE( propName, version, updateMethodPtr ) \
   ((MetaClass*)getClass())->addProperty( new FixedArrayClassProperty( #propName , MetaClass::FIRST_VERSION + version, \
   (pointer_diff)((pointer_diff)&( obj_null_pointer->##propName ) - obj_null_offset) , new Float32ClassProperty("",version, NULL, (Float32UpdateMethod)updateMethodPtr) \
   ,4, sizeof(float32), 2 ));

#define ADD_MATRIX3F_PROP_UPDATE( propName, version, updateMethodPtr ) \
	((MetaClass*)getClass())->addProperty( new FixedArrayClassProperty( #propName , MetaClass::FIRST_VERSION + version, \
	(pointer_diff)((pointer_diff)&( obj_null_pointer->##propName ) - obj_null_offset) , new Float32ClassProperty("",version, NULL, (Float32UpdateMethod)updateMethodPtr) \
	,9, sizeof(float32), 3 ));

#define ADD_MATRIX4F_PROP_UPDATE( propName, version, updateMethodPtr ) \
	((MetaClass*)getClass())->addProperty( new FixedArrayClassProperty( #propName , MetaClass::FIRST_VERSION + version, \
	(pointer_diff)((pointer_diff)&( obj_null_pointer->##propName ) - obj_null_offset) , new Float32ClassProperty("",version, NULL, (Float32UpdateMethod)updateMethodPtr) \
	,16, sizeof(float32), 4 ));

#define ADD_MATRIX2D_PROP_UPDATE( propName, version, updateMethodPtr ) \
   ((MetaClass*)getClass())->addProperty( new FixedArrayClassProperty( #propName , MetaClass::FIRST_VERSION + version, \
   (pointer_diff)((pointer_diff)&( obj_null_pointer->##propName ) - obj_null_offset) , new Float64ClassProperty("",version, NULL, (Float64UpdateMethod)updateMethodPtr) \
   ,4, sizeof(float64), 2 ));

#define ADD_MATRIX3D_PROP_UPDATE( propName, version, updateMethodPtr ) \
   ((MetaClass*)getClass())->addProperty( new FixedArrayClassProperty( #propName , MetaClass::FIRST_VERSION + version, \
   (pointer_diff)((pointer_diff)&( obj_null_pointer->##propName ) - obj_null_offset) , new Float64ClassProperty("",version, NULL, (Float64UpdateMethod)updateMethodPtr) \
   ,9, sizeof(float64), 3 ));

#define ADD_MATRIX4D_PROP_UPDATE( propName, version, updateMethodPtr ) \
   ((MetaClass*)getClass())->addProperty( new FixedArrayClassProperty( #propName , MetaClass::FIRST_VERSION + version, \
   (pointer_diff)((pointer_diff)&( obj_null_pointer->##propName ) - obj_null_offset) , new Float64ClassProperty("",version, NULL, (Float64UpdateMethod)updateMethodPtr) \
   ,16, sizeof(float64), 4 ));

#define ADD_MATRIX6F_PROP_UPDATE( propName, version, updateMethodPtr ) \
	((MetaClass*)getClass())->addProperty( new FixedArrayClassProperty( #propName , MetaClass::FIRST_VERSION + version, \
	(pointer_diff)((pointer_diff)&( obj_null_pointer->##propName ) - obj_null_offset) , new Float32ClassProperty("",version, NULL, (Float32UpdateMethod)updateMethodPtr) \
	,36, sizeof(float32), 6 ));
#define ADD_VECTOR1UB_PROP( propName, version ) ADD_VECTOR1UB_PROP_UPDATE( propName, version, NULL )
#define ADD_VECTOR2F_PROP( propName, version ) ADD_VECTOR2F_PROP_UPDATE( propName, version, NULL )
#define ADD_VECTOR2D_PROP( propName, version ) ADD_VECTOR2D_PROP_UPDATE( propName, version, NULL )
#define ADD_VECTOR2I_PROP( propName, version ) ADD_VECTOR2I_PROP_UPDATE( propName, version, NULL )
#define ADD_VECTOR3F_PROP( propName, version ) ADD_VECTOR3F_PROP_UPDATE( propName, version, NULL )
#define ADD_VECTOR3D_PROP( propName, version ) ADD_VECTOR3D_PROP_UPDATE( propName, version, NULL )
#define ADD_VECTOR6F_PROP( propName, version ) ADD_VECTOR6F_PROP_UPDATE( propName, version, NULL )
#define ADD_VECTOR3I_PROP( propName, version ) ADD_VECTOR3I_PROP_UPDATE( propName, version, NULL )
#define ADD_VECTOR4F_PROP( propName, version ) ADD_VECTOR4F_PROP_UPDATE( propName, version, NULL )
#define ADD_VECTOR4D_PROP( propName, version ) ADD_VECTOR4D_PROP_UPDATE( propName, version, NULL )
#define ADD_VECTOR4I_PROP( propName, version ) ADD_VECTOR4I_PROP_UPDATE( propName, version, NULL )
#define ADD_VECTOR4U_PROP( propName, version ) ADD_VECTOR4U_PROP_UPDATE( propName, version, NULL )
#define ADD_MATRIX2F_PROP( propName, version ) ADD_MATRIX2F_PROP_UPDATE( propName, version, NULL )
#define ADD_MATRIX3F_PROP( propName, version ) ADD_MATRIX3F_PROP_UPDATE( propName, version, NULL )
#define ADD_MATRIX4F_PROP( propName, version ) ADD_MATRIX4F_PROP_UPDATE( propName, version, NULL )
#define ADD_MATRIX2D_PROP( propName, version ) ADD_MATRIX2D_PROP_UPDATE( propName, version, NULL )
#define ADD_MATRIX3D_PROP( propName, version ) ADD_MATRIX3D_PROP_UPDATE( propName, version, NULL )
#define ADD_MATRIX4D_PROP( propName, version ) ADD_MATRIX4D_PROP_UPDATE( propName, version, NULL )
#define ADD_MATRIX6F_PROP( propName, version ) ADD_MATRIX6F_PROP_UPDATE( propName, version, NULL )

#define ADD_VECTOR3F_VAR_ARRAY_PROP( propName, version ) \
	ADD_VAR_ARRAY_PROP( \
		propName, \
		version, \
		new FixedArrayClassProperty( \
			#propName, \
			MetaClass::FIRST_VERSION + version, \
			(pointer_diff)NULL , \
			new Float32ClassProperty( "", version, (pointer_diff)NULL, NULL ), \
			3, \
			sizeof(float32), \
			3 \
		) \
	)


#define ADD_VECTOR3I_VAR_ARRAY_PROP( propName, version ) \
	ADD_VAR_ARRAY_PROP( \
	propName, \
	version, \
	new FixedArrayClassProperty( \
#propName, \
	MetaClass::FIRST_VERSION + version, \
	(pointer_diff)NULL , \
	new Int32ClassProperty( "", version, (pointer_diff)NULL, NULL ), \
	3, \
	sizeof(float32), \
	3 \
	) \
	)

#define ADD_MATRIX3F_VAR_ARRAY_PROP( propName, version ) \
	ADD_VAR_ARRAY_PROP( \
	propName, \
	version, \
	new FixedArrayClassProperty( \
#propName, \
	MetaClass::FIRST_VERSION + version, \
	(pointer_diff)NULL , \
	new Float32ClassProperty( "", version, (pointer_diff)NULL, NULL ), \
	9, \
	sizeof(float32), \
	3 \
	) \
	)

#define ADD_MATRIX4F_VAR_ARRAY_PROP( propName, version ) \
	ADD_VAR_ARRAY_PROP( \
	propName, \
	version, \
	new FixedArrayClassProperty( \
#propName, \
	MetaClass::FIRST_VERSION + version, \
	(pointer_diff)NULL , \
	new Float32ClassProperty( "", version, (pointer_diff)NULL, NULL ), \
	12, \
	sizeof(float32), \
	4 \
	) \
	)

}


#endif
