//---------------------------------------------------------------------------
#ifndef MathToolsH
#define MathToolsH
//---------------------------------------------------------------------------
#include "x4math.h"
//---------------------------------------------------------------------------
#include "X4Types.h"
#include "LinearAlgebra.h"
#include "BoundingBox.h"
#include "BivariatePolynomial.h"
//---------------------------------------------------------------------------


namespace X4 {


// ---- general math


template<class T>
inline T sqr(const T t);

inline float32 power( const float32 b, const card32 e ){
	if ( e == 0 ) return 1;
	else return b * power( b, e-1 );
}

template<class T>
inline T clamp( const T& t, const T& minValue, const T& maxValue )
{
	return max( minValue, min( maxValue, t) );
}

inline card32 log_2( const card32 n ){
	if ( n == 1 ) return 0;
	else return 1 + log_2( n / 2 );
}

// Simple blending function
template <class T>
T blend( T oldVal, T newVal, float32 factor );

inline float hermiteInterpolation(float32 left, float32 leftDiv, float32 right, float32 rightDiv, float32 t);

inline int32 simpleLog2(card32 i);

inline int32 ilog2(card32 i) { return simpleLog2(i); }

// ---- statistics / random numbers


// calculate approximate errfunc [2/sqrt(Pi) * int(exp(-t^2), t=0..x)]
// based on a 65 entry table for arg =-3.2 .. 3.2 and linear interpolation
extern X4MATH_API float32 approxErrFunc(float32 arg);

/// approximate inverse errfunc
extern X4MATH_API float32 approxInvErrFunc(float32 arg);

/// calculate random number with unit gaussian distribution
extern X4MATH_API float32 gaussianRandom();



// ---- low dimensional eigenvalue problems


/// Compute the eigenstaructure of the matrix 'cov'
void X4MATH_API computeEigenStructure( Matrix3f cov, Vector3f& eigenValues, Matrix3f& eigenVectors );

/// computes eigenspace of symmetric matrix
///  m = ( a b )       result: lambda1, lambda2 - eigenvalues
///      ( b c )               v1, v2           - eigenvectors
/// epsilon: minimum value for b before standard basis is returned
/// if normalize == true -> v1, v2 are normalized
extern X4MATH_API void eigenspace2f(const float32 a, const float32 b, const float32 c,
                                 float32 &lambda1, Vector2f &v1, float32 &lambda2, Vector2f &v2,
                                 const float32 epsilon = 0.0f, bool normalize = true);

/// computes eigenvalues of symmetic matrix
///  m = ( a b )
///      ( b c )
/// result: lambda1, lambda2 - eigenvalues
extern void X4MATH_API eigenValuesSym2f(const float32 a, const float32 b, const float32 c,
                                     float32 &lambda1, float32 &lambda2);

extern X4MATH_API void eigenspaceMatQuad2f(const Matrix2f &m, float32 &lambda1, float32 &lambda2, Matrix2f &eSpace,
                                        bool sqrtEigenvalues = true, const float32 epsilon = 0.0f, bool normalize = true);

extern X4MATH_API Matrix2f quadEigenSpaceRecompose(const Matrix2f &eigenspace, const float32 lambda1, const float32 lambda2);

/// v0, v1 define ellipsoid, transform to orthogonal main axises
extern X4MATH_API void mainAxis2f(Vector2f &v0, Vector2f &v1);



// ---- linear algebra  & geometry


/// Solve 2x2 linear system
extern bool X4MATH_API linsolve2f(const Matrix2f &m, const Vector2f &v, Vector2f &solution);

/// Solve 2x2 linear system, more general template version
template <typename FloatType>
bool  linsolve2(
   const StaticMatrix<FloatType, 2, 2> &m,
   const StaticVector<FloatType, 2> &v,
   StaticVector<FloatType, 2> &solution,
   FloatType EPSILON = (FloatType)1E-8
);

/// Solve 3x3 linear system
extern bool X4MATH_API linsolve3f(const Matrix3f &m, const Vector3f &v, Vector3f &solution);

/// Get distance between a point and a section
extern X4MATH_API float32 getDistancePointSection( const Vector3f& p, const Vector3f start, const Vector3f end );

/// Return the world coordinates defined by v
/// alpha is in the x-y-plane from [0;PI];
/// beta is in the x-z-plane from [0;2*PI]
extern X4MATH_API void getAbsoluteAngle( Vector3f v, float& alpha, float& beta );


template<class FloatType>
FloatType determinant(const StaticMatrix<FloatType, 3, 3> &m);

template<class FloatType>
FloatType determinant(const StaticVector<FloatType, 3> &v0,
                      const StaticVector<FloatType, 3> &v1,
                      const StaticVector<FloatType, 3> &v2);

template<class FloatType>
FloatType determinant(const StaticMatrix<FloatType, 2, 2> &m);

template<class FloatType>
FloatType determinant(const StaticVector<FloatType, 2> &v0,
                      const StaticVector<FloatType, 2> &v1);

extern X4MATH_API float32 calcMinProjection(const Vector3f &center, const Vector3f &prjVect, const BoundingBox3f &bb);


/// returns orthogonal componenent of nonOrtho to v. assumes norm(v) == 1. result is not normalized.
extern X4MATH_API Vector3f orthogonalize(const Vector3f &v, const Vector3f &nonOrtho);

/// assumes norm(v) == 1
extern X4MATH_API Vector3f makeOrthonormalVector(const Vector3f &v);

/// projects std::vector x (world coordinates) into local (n,u,v) system,
/// scales it to point into the plane n==1 and returns u and v coordinates
/// u,v,n assumed to be orthonormal
extern X4MATH_API Vector2f projectToN1Plane(const Vector3f &x, const Vector3f &n, const Vector3f &u, const Vector3f &v);

/// projects std::vector v to plane defined by normal n in direction d
/// (adding as much of d to v to reach the plane)
/// minPrjDenominator limits the enlargment caused by projection singularities
extern X4MATH_API Vector3f projectVectByNormalAndDirection(const Vector3f v, const Vector3f n, const Vector3f d,
                                                        const float32 minPrjDenominator = 1.0E-7f);


/// calculate orthonormal coordinate system [tangent_u, tangent_v,normal]
extern X4MATH_API Matrix3f calcTangentSystem( const Vector3f &normal );


// ---- misc


extern X4MATH_API void make16BitNormal(const Vector3f normal, int16 &normalX, int16 &normalY, int16 &normalZ);
inline Vector3f unpack16BitNormal(const int16 normalX, const int16 normalY, const int16 normalZ);

template <class FloatType, unsigned dim>
void limitLengthRatioToSmaller(StaticVector<FloatType, dim> &vect1, StaticVector<FloatType, dim> &vect2,
                               const FloatType maxFact);

template <class FloatType>
void limitRatioToSmaller(FloatType &l1, FloatType &l2, const FloatType maxFact);

template <class FloatType>
void limitRatioToLarger(FloatType &l1, FloatType &l2, const FloatType maxFact);

/*template <class FloatType>
int32 getExponentOfTwo(const FloatType v) {
   notImplemented();
}*/
extern int32 getExponentOfTwo(const float32 v);
extern int32 getExponentOfTwo(const float64 v);


/**			Given a function f, and given a bracketing triplet of abscissas ax, bx, cx (such that bx is
				between ax and cx, and f(bx) is less than both f(ax) and f(cx)), this routine isolates
				the minimum to a fractional precision of about tol using Brents method. The abscissa of
				the minimum is returned as xmin, and the minimum function value is returned as brent, the
				returned function value.*/
extern X4MATH_API float brent(float ax, float bx, float cx, float (*f)(float), float tol,
				float *xmin);


// ---- bisection search



/** search in a monotonicly raising list of entries: search for first entry in a std::vector wich has a value
    larger than searchValue. class EntryType must define operators <= and >= for comparison with NumericalType*/
template<typename EntryType, typename NumericalType>
inline card32 bisectionSearch(const std::vector<EntryType> &entries, card32 lower, card32 upper, NumericalType searchValue) {
   while (upper - lower > 1) {
      card32 mid = (upper + lower) / 2;
      if (entries[mid] <= searchValue) {
         lower = mid;
      } else {
         upper = mid;
      }
   }
   if (entries[lower] >= searchValue) {
      return lower;
   } else {
      return upper;
   }
}

/// returns index of first value >= search value
template<class ElemPtrType, class NumericalType>
inline card32 arrayBisectionSearch(const ElemPtrType entries, card32 lower, card32 upper, NumericalType searchValue) {
   while (upper - lower > 1) {
      card32 mid = (upper + lower) / 2;
      if (entries[mid] <= searchValue) {
         lower = mid;
      } else {
         upper = mid;
      }
   }
   if (entries[lower] >= searchValue) {
      return lower;
   } else {
      return upper;
   }
}

}

#endif
