#ifndef PCA_templCode_H
#define PCA_templCode_H

#include "PCA.h"

// Xrt
#include "LinearAlgebra.h"
#include "LinearAlgebra.inline.h"

// System
#include <vector>

namespace X4 {

/**
 * Create new object
 */
template<class FloatType, unsigned Dim>
PCA<FloatType,Dim>::PCA()
{
   clear();
}

/**
 * Reset PCA
 */
template<class FloatType, unsigned Dim>
void PCA<FloatType,Dim>::clear()
{
	numEntries = 0;
	summedWeight=0;
	if(points.size() > 50000)
	{
		points.clear();
		weights.clear();
	}
	// setup matrix/centroid
	for ( unsigned int i = 0; i < Dim; i++ ){
		centroid[i] = 0;
		for ( unsigned int j = 0; j < Dim; j++ ){
			C[i][j] = 0;
		}
	}
}

/**
 * Add new point
 */
template<class FloatType, unsigned Dim>
void PCA<FloatType,Dim>::addPoint( Point p )
{
   centroid = ( centroid * float32( numEntries ) + p ) / float( numEntries + 1 );
	if( numEntries == points.size() )
		points.push_back( p );
	else
		points[numEntries] =p;
	numEntries++;
}

/**
* Add new point
*/
template<class FloatType, unsigned Dim>
void PCA<FloatType,Dim>::addPoint( Point p, FloatType w )
{
	if( numEntries == points.size() )
	{
		points.push_back(p);
		weights.push_back(w);
	}
	else
	{
		points[numEntries] =p;
		weights[numEntries] = w;
	}
	numEntries++;

	summedWeight += w;
	centroid += p * w;
}

/**
 * Solve eigen system
 */
template<class FloatType, unsigned Dim>
void PCA<FloatType,Dim>::analyze( Point& eigenValues, Matrix& eigenVectors, Point& c )
{
	if( weights.size() > 0)
	{
		if( weights.size() != points.size() )
			throw "Don't mix weighted and unweigted addPoint() methods!";

		// estimate centroid
		/*float summedWeight = 0;
		for( unsigned i=0;i<points.size();i++ )
		{
			centroid += points[i]*weights[i];
			summedWeight += weights[i];
		}*/
		if( summedWeight > 1e-20f )
			centroid /= summedWeight;

		// compute weighted covariance
		float * weightPtr = &weights[0];
		for ( unsigned int n = 0; n < numEntries; n++ ){
			Point tmpP = points[n];
			float weight = *weightPtr;
			weightPtr++;
			for ( unsigned int i = 0; i < Dim; i++ ){
				for ( unsigned int j = 0; j <= i; j++ ){
					C[i][j] += ( tmpP[i]-centroid[i]) * (tmpP[j]-centroid[j])*  weight;
//					C[j][i] = C[i][j];
				}
			}
		}

		for ( unsigned int i = 0; i < Dim; i++ ){
			for ( unsigned int j = 0; j <= i; j++ ){
				C[j][i] = C[i][j];
			}
		}


		if( summedWeight > 1e-20f )
			C /= summedWeight;

	}
	else
	{
		for ( unsigned int n = 0; n < numEntries; n++ ){
			const Point tmp = points[n];
			for ( unsigned int i = 0; i < Dim; i++ ){
				for ( unsigned int j = 0; j <= i; j++ ){
					C[i][j] += ( tmp[i]-centroid[i]) * (tmp[j]-centroid[j]);
					//C[j][i] = C[i][j];
				}
			}
		}

		for ( unsigned int i = 0; i < Dim; i++ ){
			for ( unsigned int j = 0; j <= i; j++ ){
				C[j][i] = C[i][j];
			}
		}

	}

	/// Normalize covariance matrix
	if ( points.size() > 0 ) C /= float32( points.size() );

	// Analyze eigenstructure
	C.computeEigenStructure( eigenValues, eigenVectors );
	c = centroid;
}

}

#endif
