//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PartialDerivative.h"
#include "PartialDerivative.inline.h"
#include <ConsoleOutput.h>
#include <Timer.h>
//#include "GLGraphDisplayer.h"
//#include "DynamicLinearAlgebraTools.h"
//#include "DynamicLinearAlgebraTools.inline.h"
#include "SparseLinearAlgebra.h"
#include "SparseLinearAlgebra.inline.h"
#include "IterativeSolvers.h"
#include "IterativeSolvers.inline.h"
#include "MathTools.h"
#include "MathTools.inline.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------


template <typename FloatType>
static FloatType func(const StaticVector<FloatType, 3> &pos)
{
   StaticMatrix<FloatType, 3, 3> Q;
   Q[0][0] = 1; Q[1][0] = 2; Q[2][0] = 3;
   Q[0][1] = 2; Q[1][1] = 4; Q[2][1] = 5;
   Q[0][2] = 3; Q[1][2] = 5; Q[2][2] = 6;
   StaticVector<FloatType, 3> B;
   B[0] = 1;
   B[1] = 2;
   B[2] = 3;
   FloatType result = (Q*pos)*pos + pos*B + (FloatType)5;
   return result;
}

template <class FloatType>
static FloatType scoreFunc(
      const FloatType &left, 
      const FloatType &middle, 
      const FloatType &right, 
      const FloatType &measuredMiddle,
      float32 exponent) 
{
   FloatType laplace = left + right - middle*(FloatType)2;
   FloatType smoothScore = (exponent==2.0f) ? (laplace*laplace)
                                            : (pow(pow(laplace*laplace+(FloatType)0.0000001, (FloatType)0.5), (FloatType)exponent));
   FloatType dataScore = sqr(middle - measuredMiddle);
   
   return smoothScore + dataScore*(FloatType)0.001;
};

typedef HessianTools<float32, 3> HTools3f;
typedef HTools3f::DiffTwice ADType;

static void computeLocalTaylor(
   const std::vector<float32> &pts, const std::vector<float32> &ptsOriginal, int i, 
   float32 &localScore, Vector3f &localGradient, Matrix3f &localHessian,
   const int NUM_PTS, const float32 exponent) 
{
   int i0 = (i+(int)NUM_PTS-1) % NUM_PTS;
   int i1 = (i+(int)NUM_PTS)   % NUM_PTS;
   int i2 = (i+(int)NUM_PTS+1) % NUM_PTS;
   ADType left = HTools3f::createVariable(pts[i0], 0);
   ADType middle = HTools3f::createVariable(pts[i1], 1);
   ADType right = HTools3f::createVariable(pts[i2], 2);
   ADType measuredMiddle = ptsOriginal[i1];
   ADType score = scoreFunc(left, middle, right, measuredMiddle, exponent);
   localGradient = HTools3f::getGradient(score);
   localHessian = HTools3f::getHessian(score);
}


static void computeTaylor(
   const std::vector<float32> &pts, const std::vector<float32> &ptsOriginal, 
   SparseMatrixF &hessian, DVectorF &gradient, float32 &score, 
   const int NUM_PTS, float32 exponent) 
{
   for (int i=0; i<NUM_PTS; i++) {
      Vector3f localGradient;
      Matrix3f localHessian;
      float32 localScore;
      computeLocalTaylor(pts, ptsOriginal, i, localScore, localGradient, localHessian, NUM_PTS, exponent);
      int i0 = (i+NUM_PTS-1) % NUM_PTS;
      int i1 = (i+NUM_PTS) % NUM_PTS;
      int i2 = (i+NUM_PTS+1) % NUM_PTS;
      gradient[i0] += localGradient[0];
      gradient[i1] += localGradient[1];
      gradient[i2] += localGradient[2];
      hessian[i0].addEntry(i0, localHessian[0][0]); hessian[i0].addEntry(i1, localHessian[0][1]); hessian[i0].addEntry(i2, localHessian[0][2]);
      hessian[i1].addEntry(i0, localHessian[1][0]); hessian[i1].addEntry(i1, localHessian[1][1]); hessian[i1].addEntry(i2, localHessian[1][2]);
      hessian[i2].addEntry(i0, localHessian[2][0]); hessian[i2].addEntry(i1, localHessian[2][1]); hessian[i2].addEntry(i2, localHessian[2][2]);
      score += localScore;
   }
}

void partialDerivativeTest() {
   const unsigned NUM_TESTS = 100;//000;

   Timer T;
   double result = func<double>(makeVector3d(1,2,3));
   for (unsigned i=0; i<NUM_TESTS*100; i++) {
      double result = func<double>(makeVector3d(1,2,3));
   }
   x4console << "time double: " << T.getDeltaValue()/100.0 << "\n";

   const double eps = 1e-7;
   Vector3d grad = 
      makeVector3d(
         (func<double>(makeVector3d(1+eps,2,3)) - func<double>(makeVector3d(1-eps,2,3)))/(2*eps),
         (func<double>(makeVector3d(1,2+eps,3)) - func<double>(makeVector3d(1,2-eps,3)))/(2*eps),
         (func<double>(makeVector3d(1,2,3+eps)) - func<double>(makeVector3d(1,2,3-eps)))/(2*eps)
      );

   StaticVector< PartialDerivative<double,3>, 3> v;
   v[0] = PartialDerivative<double,3>::createVariable(1, 0);
   v[1] = PartialDerivative<double,3>::createVariable(2, 1);
   v[2] = PartialDerivative<double,3>::createVariable(3, 2);
   
   
   PartialDerivative<double,3> resultAD = func< PartialDerivative<double,3> >(v);
   T.getDeltaValue();
   for (unsigned i=0; i<NUM_TESTS; i++) {
      resultAD = func< PartialDerivative<double,3> >(v);
   }
   x4console << "time PartialDerivative<double,3>: " << T.getDeltaValue()<< "\n";;
   
   double resultFB = resultAD.getValue();
   Vector3d gradFB = makeVector3d(
      resultAD.getDerivative(0),
      resultAD.getDerivative(1),
      resultAD.getDerivative(2)
   );

   StaticVector< HessianTools<double, 3>::DiffTwice, 3 > vh;
   vh[0] = HessianTools<double, 3>::createVariable(1,0);
   vh[1] = HessianTools<double, 3>::createVariable(2,1);
   vh[2] = HessianTools<double, 3>::createVariable(3,2);
   HessianTools<double, 3>::DiffTwice resultADH = func< HessianTools<double, 3>::DiffTwice >(vh);

   T.getDeltaValue();
   for (unsigned i=0; i<NUM_TESTS; i++) {
      resultADH = func< HessianTools<double, 3>::DiffTwice >(vh);
   }
   x4console << "time HessianTools<double, 3>::DiffTwice: " << T.getDeltaValue()<< "\n\n";;

   x4console << "\n\nresult: " << result << "  num. grad: " << grad.toString() << "\n";
   x4console << "\nad-result :" << resultFB << "  grad: " << gradFB.toString() << "\n";

   x4console << "Hessian: " << HessianTools<double, 3>::getHessian(resultADH).toString() << "\n";
   x4console << "grad from HessComp: " << HessianTools<double, 3>::getGradient(resultADH).toString() << "\n";
   x4console << "alt-grad from HessComp: " << HessianTools<double, 3>::getGradientAlt(resultADH).toString() << "\n";
   
   /*
   x4console << "\nReconstruction Test\n";
   {
      static const unsigned NUM_PTS = 100;
      std::vector<float32> pts(NUM_PTS);
      for (unsigned i=0; i<NUM_PTS; i++) {
         pts[i] = fabs(i/50.0f-1.0f) + rnd01()*0.1;
      }
      std::vector<float32> ptsOriginal = pts;


      std::vector< std::vector<float32> > graphs;
      std::vector< Vector4ub > colors;
      graphs.push_back(pts);
      colors.push_back(makeVector4ub(220,68,60,255));

      SparseMatrixF hessian(NUM_PTS);
      DVectorF gradient(NUM_PTS);
      std::vector<float32> visGrad(NUM_PTS);

      for (card32 i=0; i<1000; i++) {
         hessian.setZero();
         gradient.setZero();
         float32 score = 0;
         computeTaylor(pts, ptsOriginal, hessian, gradient, score, NUM_PTS, 2.0f);
         if (i==0) {
            convertVectorType(gradient, visGrad, NUM_PTS);
         }
         for (unsigned j=0; j<pts.size(); j++) {
            pts[j] -= gradient[j]*0.05;
         }
      }
      graphs.push_back(pts);
      colors.push_back(makeVector4ub(255,255,0,255));
      
      pts = ptsOriginal;
      for (card32 i=0; i<1; i++) {
         DVectorF x(NUM_PTS);
         convertVectorType(pts, x, NUM_PTS);
         card32 maxNumIterations = 500;
         float32 score = 0;
         hessian.setZero();
         gradient.setZero();
         computeTaylor(pts, ptsOriginal, hessian, gradient, score, NUM_PTS, 2);
         conjugateGradientsSolve<float32>(
            hessian,
            x,
            hessian*x - gradient,
            maxNumIterations,
            1E-7f,
            true
         );
         convertVectorType(x, pts, NUM_PTS);
      }
      
      graphs.push_back(pts);
      colors.push_back(makeVector4ub(255,255,255,255));

      //graphs.push_back(visGrad);
      //colors.push_back(makeVector4ub(220,60,60,255));
      
      x4consoleGraphs("Partial Derivative Reconstruction Test", graphs, colors);
   }
   {
      x4console << "\n\n\n";
      for (card32 i=1; i<10; i++) {
         HessianTools<double,2>::DiffTwice x1 = HessianTools<double,2>::createVariable(i,0);
         HessianTools<double,2>::DiffTwice x2 = HessianTools<double,2>::createVariable(i,1);
         HessianTools<double,2>::DiffTwice y = pow(x1, (HessianTools<double,2>::DiffTwice)2)
                                             + pow(x2, (HessianTools<double,2>::DiffTwice)2);
         x4console << "x1:" << x1.getValue().getValue() 
                     << "\tx2:" << x2.getValue().getValue() 
                     << "\ty: " << y.getValue().getValue()
                     << "\tdy/dx1: " << y.getDerivative(0).getValue() 
                     << "\tdy/dx2: " << y.getDerivative(1).getValue() 
                     << "\tdy^2/dx1^2: " << y.getDerivative(0).getDerivative(0) 
                     << "\tdy^2/dx2^2: " << y.getDerivative(1).getDerivative(1) 
                     << "\tdy^2/dx1dx2: " << y.getDerivative(0).getDerivative(1) 
                     << "\tdy^2/dx2dx1: " << y.getDerivative(1).getDerivative(0) 
                     << "\n";
      }
      
   } */
}


void X4::PartialDerivative_unitTest() {
   partialDerivativeTest();
}



