//---------------------------------------------------------------------------
#ifndef PartialDerivativeH
#define PartialDerivativeH
//---------------------------------------------------------------------------
#include <x4math.h>
#include "LinearAlgebra.h"
//---------------------------------------------------------------------------

namespace X4 {

template<typename FloatType, unsigned dim>
class HessianTools;


template <typename FloatType, unsigned dim>
class PartialDerivative {
 private:
	FloatType value;
	FloatType derivatives[dim];

 public:
	inline PartialDerivative();
	inline PartialDerivative(const FloatType &v, const FloatType *d);
	inline PartialDerivative(const PartialDerivative &op);
	/// constants are created implicitly by conversion; the derivative of constants is zero
   template <typename AnyFloatType>
	inline PartialDerivative(const AnyFloatType &op);
	/// variables have to be created explicitly; derivatives of variables are one; 
	/// the index of the partial derivative has to be specified accordingly
	static inline PartialDerivative<FloatType, dim> createVariable(const FloatType &value, unsigned index);

	inline FloatType &getValue();
	inline const FloatType &getValue() const;
	inline const FloatType *getDerivatives() const;
	inline FloatType &getDerivative(unsigned i);
	inline const FloatType &getDerivative(unsigned i) const;
	inline void setValue(const FloatType &v);
	inline void setDerivative(unsigned index, const FloatType &v);
	inline void setDerivatives(const FloatType *v);

   inline PartialDerivative operator+(const PartialDerivative &op) const;
   inline PartialDerivative operator-(const PartialDerivative &op) const;
   inline PartialDerivative operator*(const PartialDerivative &s) const;
   inline PartialDerivative operator/(const PartialDerivative &s) const;
   inline void operator+=(const PartialDerivative &op);
   inline void operator-=(const PartialDerivative &op);
   inline void operator*=(const PartialDerivative &op);
   inline void operator/=(const PartialDerivative &op);

   inline PartialDerivative operator+(const FloatType &op) const;
   inline PartialDerivative operator-(const FloatType &op) const;
   inline PartialDerivative operator*(const FloatType &s) const;
   inline PartialDerivative operator/(const FloatType &s) const;
   inline void operator+=(const FloatType &s);
   inline void operator-=(const FloatType &s);
   inline void operator*=(const FloatType &s);
   inline void operator/=(const FloatType &s);

   inline PartialDerivative operator-() const;

   inline void operator=(const PartialDerivative op);
  friend HessianTools<FloatType, dim>;
};



template<typename FloatType, unsigned dim>
class HessianTools {
 public:
   typedef typename PartialDerivative< PartialDerivative<FloatType, dim>, dim > DiffTwice;
   
   /// do NOT use DiffTwice::createVariable - this will not work properly!
   static DiffTwice createVariable(FloatType value, unsigned index);
   static StaticMatrix<FloatType, dim, dim> getHessian(const DiffTwice &v);
   static StaticVector<FloatType, dim> getGradient(const DiffTwice &v);
   static StaticVector<FloatType, dim> getGradientAlt(const DiffTwice &v);
};



template <typename FloatType, unsigned dim>
inline PartialDerivative<FloatType, dim> exp(const PartialDerivative<FloatType, dim> &op);

template <typename FloatType>
inline FloatType exp(FloatType &op);


template <typename FloatType, unsigned dim>
inline PartialDerivative<FloatType, dim> log(const PartialDerivative<FloatType, dim> &op);

template <typename FloatType>
inline FloatType log(FloatType &op);


template <typename FloatType, unsigned dim>
inline PartialDerivative<FloatType, dim> pow(
   const PartialDerivative<FloatType, dim> &base, 
   const PartialDerivative<FloatType, dim> &exponent
);


void PartialDerivative_unitTest();

}		


#endif