//---------------------------------------------------------------------------
#ifndef PartialDerivativeInlineH
#define PartialDerivativeInlineH
//---------------------------------------------------------------------------
#include <x4math.h>
//---------------------------------------------------------------------------
#include "PartialDerivative.h"
//---------------------------------------------------------------------------



namespace X4 {


template <typename FloatType, unsigned dim>
inline PartialDerivative<FloatType, dim>::PartialDerivative() {
	value = 0;
	for (unsigned i=0; i<dim; i++) {
		derivatives[i] = 0;
	}
}

template <typename FloatType, unsigned dim>
inline PartialDerivative<FloatType, dim>::PartialDerivative(const FloatType &v, const FloatType *d) {
	value = v;
	for (unsigned i=0; i<dim; i++) {
		derivatives[i] = d[i];
	}
}

template <typename FloatType, unsigned dim>
inline PartialDerivative<FloatType, dim>::PartialDerivative(const PartialDerivative &op) {
	value = op.value;
	for (unsigned i=0; i<dim; i++) {
		derivatives[i] = op.derivatives[i];
	}
}

template <typename FloatType, unsigned dim>
template <typename AnyFloatType>
inline PartialDerivative<FloatType, dim>::PartialDerivative(const AnyFloatType &op) {
	value = (FloatType)op;
	for (unsigned i=0; i<dim; i++) {
		derivatives[i] = (FloatType)0;
	}
}

template <typename FloatType, unsigned dim>
inline PartialDerivative<FloatType, dim> PartialDerivative<FloatType, dim>::createVariable(const FloatType &value, unsigned index) {
	PartialDerivative<FloatType, dim> result;
	result.derivatives[index] = 1;
	result.value = value;
	return result;
}

template <typename FloatType, unsigned dim>
inline const FloatType &PartialDerivative<FloatType, dim>::getValue() const {
	return value;
}

template <typename FloatType, unsigned dim>
inline FloatType &PartialDerivative<FloatType, dim>::getValue() {
	return value;
}

template <typename FloatType, unsigned dim>
inline const FloatType *PartialDerivative<FloatType, dim>::getDerivatives() const {
	return &derivative;
}

template <typename FloatType, unsigned dim>
inline const FloatType &PartialDerivative<FloatType, dim>::getDerivative(unsigned i) const {
   return derivatives[i];
}

template <typename FloatType, unsigned dim>
inline FloatType &PartialDerivative<FloatType, dim>::getDerivative(unsigned i) {
   return derivatives[i];
}

template <typename FloatType, unsigned dim>
inline void PartialDerivative<FloatType, dim>::setValue(const FloatType &v) {
	value = v;
}

template <typename FloatType, unsigned dim>
inline void PartialDerivative<FloatType, dim>::setDerivative(unsigned index, const FloatType &v) {
	derivatives[index] = v;
}

template <typename FloatType, unsigned dim>
inline void PartialDerivative<FloatType, dim>::setDerivatives(const FloatType *v) {
	for (unsigned i=0; i<dim; i++) {
		derivatives[i] = v[i];
	}
}

template <typename FloatType, unsigned dim>
inline PartialDerivative<FloatType, dim> PartialDerivative<FloatType, dim>::operator+(const PartialDerivative &op) const {
	PartialDerivative<FloatType, dim> result;
	result.value      = value      + op.value;
	for (unsigned i=0; i<dim; i++) {
   	result.derivatives[i] = derivatives[i] + op.derivatives[i];
   }
	return result;
}

template <typename FloatType, unsigned dim>
inline PartialDerivative<FloatType, dim> PartialDerivative<FloatType, dim>::operator-(const PartialDerivative &op) const {
	PartialDerivative<FloatType, dim> result;
	result.value      = value      - op.value;
	for (unsigned i=0; i<dim; i++) {
   	result.derivatives[i] = derivatives[i] - op.derivatives[i];
   }
	return result;
}

template <typename FloatType, unsigned dim>
inline PartialDerivative<FloatType, dim> PartialDerivative<FloatType, dim>::operator*(const PartialDerivative &op) const {
	PartialDerivative<FloatType, dim> result;
	result.value      = value * op.value;
	for (unsigned i=0; i<dim; i++) {
	   result.derivatives[i] = derivatives[i]*op.value + op.derivatives[i]*value;
   }
	return result;
}

template <typename FloatType, unsigned dim>
inline PartialDerivative<FloatType, dim> PartialDerivative<FloatType, dim>::operator/(const PartialDerivative &op) const {
	PartialDerivative<FloatType, dim> result;
	result.value      = value / op.value;
	for (unsigned i=0; i<dim; i++) {
	   result.derivatives[i] = (derivatives[i]*op.value - op.derivatives[i]*value)/(op.value*op.value);
   }
	return result;
}

template <typename FloatType, unsigned dim>
inline void PartialDerivative<FloatType, dim>::operator+=(const PartialDerivative &op) {
	value += op.value;
	for (unsigned i=0; i<dim; i++) {
   	derivatives[i] += op.derivatives[i];
   }
}

template <typename FloatType, unsigned dim>
inline void PartialDerivative<FloatType, dim>::operator-=(const PartialDerivative &op) {
	value -= op.value;
	for (unsigned i=0; i<dim; i++) {
   	derivatives[i] -= op.derivatives[i];
   }
}

template <typename FloatType, unsigned dim>
inline void PartialDerivative<FloatType, dim>::operator*=(const PartialDerivative &op) {
	FloatType valueTmp = value * op.value;
	for (unsigned i=0; i<dim; i++) {
   	derivatives[i] = derivatives[i]*op.value + value*op.derivatives[i];
   }
	value = valueTmp;
}

template <typename FloatType, unsigned dim>
inline void PartialDerivative<FloatType, dim>::operator/=(const PartialDerivative &op) {
	FloatType valueTmp = value / op.value;
	for (unsigned i=0; i<dim; i++) {
   	derivatives[i] = (derivatives[i]*op.value - op.derivatives[i]*value)/(op.value*op.value);
   }
	value = valueTmp;
}


template <typename FloatType, unsigned dim>
inline PartialDerivative<FloatType, dim> PartialDerivative<FloatType, dim>::operator+(const FloatType &s) const {
	PartialDerivative<FloatType, dim> result;
	result.value      = value + s;
	for (unsigned i=0; i<dim; i++) {
   	result.derivatives[i] = derivatives[i];
   }
	return result;	
}

template <typename FloatType, unsigned dim>
inline PartialDerivative<FloatType, dim> PartialDerivative<FloatType, dim>::operator-(const FloatType &s) const {
	PartialDerivative<FloatType, dim> result;
	result.value      = value - s;
	for (unsigned i=0; i<dim; i++) {
   	result.derivatives[i] = derivatives[i];
   }
	return result;	
}

template <typename FloatType, unsigned dim>
inline PartialDerivative<FloatType, dim> PartialDerivative<FloatType, dim>::operator*(const FloatType &s) const {
	PartialDerivative<FloatType, dim> result;
	result.value      = value * s;
	for (unsigned i=0; i<dim; i++) {
   	result.derivatives[i] = derivatives[i] * s;
   }
	return result;	
}

template <typename FloatType, unsigned dim>
inline PartialDerivative<FloatType, dim> PartialDerivative<FloatType, dim>::operator/(const FloatType &s) const {
	PartialDerivative<FloatType, dim> result;
	result.value      = value / s;
	for (unsigned i=0; i<dim; i++) {
	   result.derivatives[i] = derivatives[i] / s;
   }
	return result;	
}

template <typename FloatType, unsigned dim>
inline void PartialDerivative<FloatType, dim>::operator+=(const FloatType &s) {
	value += op;
}

template <typename FloatType, unsigned dim>
inline void PartialDerivative<FloatType, dim>::operator-=(const FloatType &s) {
	value -= op;
}

template <typename FloatType, unsigned dim>
inline void PartialDerivative<FloatType, dim>::operator*=(const FloatType &s) {
	value *= op;
	for (unsigned i=0; i<dim; i++) {
	   derivatives[i] *= op;
   }
}

template <typename FloatType, unsigned dim>
inline void PartialDerivative<FloatType, dim>::operator/=(const FloatType &s) {
	value /= op;
	for (unsigned i=0; i<dim; i++) {
   	derivatives[i] /= op;
   }
}


template <typename FloatType, unsigned dim>
inline PartialDerivative<FloatType, dim> PartialDerivative<FloatType, dim>::operator-() const {
	PartialDerivative<FloatType, dim> result;
	result.value = -value;
	for (unsigned i=0; i<dim; i++) {
   	result.derivatives[i] = -derivatives[i];
   }
	return result;
}

template <typename FloatType, unsigned dim>
inline void PartialDerivative<FloatType, dim>::operator=(const PartialDerivative op) {
	value = op.value;
	for (unsigned i=0; i<dim; i++) {
   	derivatives[i] = op.derivatives[i];
   }
}






template<typename FloatType, unsigned dim>
PartialDerivative< PartialDerivative<FloatType, dim>, dim > HessianTools<FloatType, dim>::createVariable(FloatType value, unsigned index) {
   DiffTwice result;
   result.getValue().setValue(value);
   result.getValue().setDerivative(index, 1);
   result.getDerivative(index).setValue(1);
   //result.getDerivative(index).setDerivative(index, 1);
   return result;
}

template<typename FloatType, unsigned dim>
StaticMatrix<FloatType, dim, dim> HessianTools<FloatType, dim>::getHessian(const DiffTwice &v) {
   StaticMatrix<FloatType, dim, dim> result;
   for (unsigned i=0; i<dim; i++) {
      for (unsigned j=0; j<dim; j++) {
         result[i][j] = v.getDerivative(i).getDerivative(j);
      }
   }
   return result;
}

template<typename FloatType, unsigned dim>
StaticVector<FloatType, dim> HessianTools<FloatType, dim>::getGradient(const DiffTwice &v) {
   StaticVector<FloatType, dim> result;
   for (unsigned i=0; i<dim; i++) {
      result[i] = v.getValue().getDerivative(i);
   }
   return result;
}

template<typename FloatType, unsigned dim>
StaticVector<FloatType, dim> HessianTools<FloatType, dim>::getGradientAlt(const DiffTwice &v) {
   StaticVector<FloatType, dim> result;
   for (unsigned i=0; i<dim; i++) {
      result[i] = v.getDerivative(i).getValue();
   }
   return result;
}


template <typename FloatType, unsigned dim>
inline PartialDerivative<FloatType, dim> exp(const PartialDerivative<FloatType, dim> &op) {
	PartialDerivative<FloatType, dim> result;
	result.setValue(exp(op.getValue()));
	for (unsigned i=0; i<dim; i++) {
	   result.setDerivative(i, result.getValue()*op.getDerivative(i));
	}
	return result;
}

template <typename FloatType>
inline FloatType exp(FloatType &op) {
   return ::exp(op);
}

template <typename FloatType, unsigned dim>
inline PartialDerivative<FloatType, dim> log(const PartialDerivative<FloatType, dim> &op) {
	PartialDerivative<FloatType, dim> result;
	result.setValue(log(op.getValue()));
	for (unsigned i=0; i<dim; i++) {
	   result.setDerivative(i, (FloatType(1))/op.getValue()*op.getDerivative(i));
	}
	return result;
}

template <typename FloatType>
inline FloatType log(FloatType &op) {
   return ::log(op);
}

template <typename FloatType, unsigned dim>
inline PartialDerivative<FloatType, dim> pow(
   const PartialDerivative<FloatType, dim> &base, 
   const PartialDerivative<FloatType, dim> &exponent) 
{
	return exp(log(base)*exponent);   
}


}
		
#endif