//---------------------------------------------------------------------------
#include "stdafx.h"
//---------------------------------------------------------------------------
#include "Plane.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

IMPLEMENT_X4_CLASS( Plane, 0 )
{
	BEGIN_CLASS_INIT( Plane );
	ADD_VECTOR3F_PROP( normal, 0 );
	ADD_VECTOR3F_PROP( point, 0 );
}

Plane::Plane()
{
	clear();
}

Plane::Plane( Vector3f n, Vector3f p )
{ 
	clear();
	normal = n; 
	normal.normalize(),
	point = p;
}

Plane::~Plane()
{
}

void Plane::clear()
{
	for ( unsigned int n = 0; n < 3; n++ ){
		point[n] = 0;
		normal[n] = 0;
	}
}

float Plane::getDistanceFromPlane( const Vector3f p ) const
{
	return ( normal * p - normal * point );
}

float Plane::calculateSignedDistance( const Vector3f &p ) const
{
	return p*normal - point*normal;
}

float Plane::calculateLineIntersection( const Vector3f &a, const Vector3f &b ) const
{
	float t = (b-a)*normal;
	if( fabs(t) < 1e-7f ) return 1e20f;
	t = (normal * point - a*normal) / t;
	return t;
}

EPlaneSide Plane::classify( const Line &line, float epsilon ) const
{
	float d1 = calculateSignedDistance( line.getPoint() );
	float d2 = calculateSignedDistance( line.getPoint()+line.getDirection() );

	if( d1 > epsilon && d2 > epsilon )
		return PS_INFRONT;

	if( d1 <= -epsilon && d2 <= -epsilon )
		return PS_BEHIND;

	if( (d1 > epsilon && d2 <= -epsilon) ||
		(d2 > epsilon && d1 <= -epsilon) )
		return PS_SPANNING;

	return PS_COINCIDING;
}

void Plane::split( const Line &line, Line &positive, Line &negative ) const
{
	x4Assert( classify( line ) == PS_SPANNING );

	float d1 = calculateSignedDistance( line.getPoint() );

	Vector3f xm = calculateIntersection( line );

	const float epsilon = 1e-20f;
	positive = line;
	positive.setDirection(xm - positive.getPoint());
	negative.setPoint(xm);
	negative.setDirection(line.getDirection()+line.getPoint()-xm);


	if( d1 <= epsilon )
	{
		swap(positive,negative);
	}

}

Line Plane::calculateIntersection(const Plane& plane) const
{
	float d = plane.getNormal() * getNormal();
	if (fabs(d) > 0.9999f) throw X4Exception("Planes are parallel");

	Vector3f ldir = normalize(getNormal().crossProduct(plane.getNormal()));
	Line tmpline(normalize(ldir.crossProduct(getNormal())), getPoint());

	return Line(ldir, plane.calculateIntersection(tmpline));
}

Vector3f Plane::calculateIntersection( const Line& line ) const
{
	return line.getPoint() + line.getDirection() * ((point*normal - line.getPoint()*normal) / (normal * line.getDirection()) );

}

Vector3f Plane::calculateMirrorPoint( const Vector3f &p)
{
	return p - normal * ((p - point)*normal*2.0f);
}

X4::Vector3f X4::Plane::projectPointOntoPlane( const Vector3f &p )
{
	return p - normal*((p-point)*normal);
}

X4::Plane X4::Plane::transform( const Matrix4f &T ) const
{
	Plane ret;
	ret.point = transformVector3f(T,point);
	ret.normal = shrink4To3(T) * normal;
	return ret;
}

bool X4::Plane::equal( const Plane &p, float angle_threshold, float spatial_threshold ) const
{
	if( normal*p.normal < 1.0f - angle_threshold )
		return false;
	float dist = fabs(calculateSignedDistance(p.point));
	return dist <= spatial_threshold;
}

void X4::Plane::movePoint( const float& dist )
{
    point += dist * normal;
}
