//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "Plane3f.h"
#include "Line3f.h"
#include <FixedArrayClassProperty.h>
#include <ConsoleOutput.h>
#include "MathTools.h"
#include "LinearAlgebra.inline.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

IMPLEMENT_X4_CLASS( Plane3f, 0 )
{
	BEGIN_CLASS_INIT( Plane3f );
	ADD_VECTOR3F_PROP( normal, 0 );
	ADD_VECTOR3F_PROP( point, 0 );
}

Plane3f::Plane3f()
{
	clear();
}

Plane3f::Plane3f( Vector3f n, Vector3f p )
{ 
	clear();
	normal = n; 
	normal.normalize(),
	point = p;
}

Plane3f::~Plane3f()
{
}

void Plane3f::clear()
{
	for ( unsigned int n = 0; n < 3; n++ ){
		point[n] = 0;
		normal[n] = 0;
	}
}

float Plane3f::getDistanceFromPlane( const Vector3f p ) const
{
	return ( normal * p - normal * point );
}

float Plane3f::calculateSignedDistance( const Vector3f &p ) const
{
	return p*normal - point*normal;
}

float Plane3f::calculateLineIntersection( const Vector3f &a, const Vector3f &b ) const
{
	float t = (b-a)*normal;
	if( t == 0.0f )
		return 1e20f;
	t = (normal * point - a*normal) / t;
	return t;
}

EPlaneSide Plane3f::classify( const Line3f &line, float epsilon ) const
{
	float d1 = calculateSignedDistance( line.getPoint() );
	float d2 = calculateSignedDistance( line.getPoint()+line.getDirection() );

	if( d1 > epsilon && d2 > epsilon )
		return PS_INFRONT;

	if( d1 <= -epsilon && d2 <= -epsilon )
		return PS_BEHIND;

	if( (d1 > epsilon && d2 <= -epsilon) ||
		(d2 > epsilon && d1 <= -epsilon) )
		return PS_SPANNING;

	return PS_COINCIDING;
}

void Plane3f::split( const Line3f &line, Line3f &positive, Line3f &negative ) const
{
	x4Assert( classify( line ) == PS_SPANNING );

	float d1 = calculateSignedDistance( line.getPoint() );

	Vector3f xm = calculateIntersection( line );

	const float epsilon = 1e-20f;
	positive = line;
	positive.setDirection(xm - positive.getPoint());
	negative.setPoint(xm);
	negative.setDirection(line.getDirection()+line.getPoint()-xm);


	if( d1 <= epsilon )
	{
		swap(positive,negative);
	}

}

Line3f Plane3f::calculateIntersection(const Plane3f& plane) const
{
	float d = plane.getNormal() * getNormal();
	if (fabs(d) > 0.9999f) throw X4Exception("Planes are parallel");

	Vector3f ldir = normalize(getNormal().crossProduct(plane.getNormal()));
	Line3f tmpline(normalize(ldir.crossProduct(getNormal())), getPoint());

	return Line3f(ldir, plane.calculateIntersection(tmpline));
}

Vector3f Plane3f::calculateIntersection( const Line3f& line ) const
{
	return line.getPoint() + line.getDirection() * ((point*normal - line.getPoint()*normal) / (normal * line.getDirection()) );

}

Vector3f Plane3f::calculateMirrorPoint( const Vector3f &p)
{
	return p - normal * ((p - point)*normal*2.0f);
}

Vector3f Plane3f::projectPointOntoPlane( const Vector3f &p ) const
{
	return p - normal*((p-point)*normal);
}

Plane3f Plane3f::transform( const Matrix4f &T ) const
{
	Plane3f ret;
	ret.point = transformVector3f(T,point);
	ret.normal = shrink4To3(T) * normal;
	return ret;
}

bool Plane3f::equal( const Plane3f &p, float angle_threshold, float spatial_threshold ) const
{
	if( normal*p.normal < 1.0f - angle_threshold )
		return false;
	float dist = fabs(calculateSignedDistance(p.point));
	return dist <= spatial_threshold;
}

void X4::Plane3f::movePoint( const float& dist )
{
    point += dist * normal;
}

/**
 * Projection of a point onto a plane
 */
void X4::projectPointPlane( const Plane3f& plane, const Vector3f point, Vector3f& projectedPoint )
{
   projectedPoint = ( point - plane.getNormal() * plane.getDistanceFromPlane( point ) );
}

/**
 * Intersection between plane and line; return true if possible
 */
bool X4::intersectPlaneLine( const Plane3f& plane, const Line3f& line, Vector3f& intersectionPoint )
{
	float angle = plane.getNormal() * line.getDirection();
	if ( fabs( angle ) < 1e-5 ){
      //warning("intersectPlaneLine: plane and line nearly orthogonal!\n");
		return false;
   }
	
	float a = ( plane.getDistance() - line.getPoint() * plane.getNormal() ) / angle;
	intersectionPoint = line.getPoint() + line.getDirection() * a;
	return true;
}

/**
 * Intersection between tweo planes; return true, if possible
 */
bool X4::intersectPlanePlane( const Plane3f plane1, const Plane3f plane2, Line3f& intersectionLine )
{
	Vector3f n1 = plane1.getNormal(), n2 = plane2.getNormal();
	n1 = normalize( n1 );
	n2 = normalize( n2 ); 

	// Planes are parallel
	if ( ( n1 * n2 ) > 1-1e-6 )
		return false;

	// Compute direction of line
	Vector3f dir = n1.crossProduct( n2 );
	intersectionLine.setDirection( dir );

	// Decide, which variable should be zero
	int zeroDim = ( fabs( dir[0] ) > fabs( dir[1] ) )? 
		              ( ( fabs( dir[0] ) > fabs( dir[2] ) )? 0 : 2 ) : 
	                  ( ( fabs( dir[1] ) > fabs( dir[2] ) )? 1 : 2 );
	Matrix2f M; 
	Vector2f b = makeVector2f( plane1.getDistance(), plane2.getDistance() ), x;
	
	// Setup linear system
	switch ( zeroDim ){
		case 0:
			M[0][0] = n1[1]; M[0][1] = n1[2];
			M[1][0] = n2[1]; M[1][1] = n2[2];		
			break;
		case 1:
			M[0][0] = n1[0]; M[0][1] = n1[2];
			M[1][0] = n2[0]; M[1][1] = n2[2];
			break;
		case 2:
			M[0][0] = n1[0]; M[0][1] = n1[1];
			M[1][0] = n2[0]; M[1][1] = n2[1];
			break;
	}

	// Solve linear system to find point
	linsolve2f( M, b, x );

	// Create point
	switch ( zeroDim ){
		case 0:
			intersectionLine.setPoint( makeVector3f( 0, x[0], x[1] ) );
			break;
		case 1:
			intersectionLine.setPoint( makeVector3f( x[0], 0, x[1] ) );
			break;
		case 2:
			intersectionLine.setPoint( makeVector3f( x[0], x[1], 0 ) );
			break;
	}

	return true;
}
