//---------------------------------------------------------------------------
#ifndef PolygonH
#define PolygonH
//---------------------------------------------------------------------------
#include "x4math.h"
//---------------------------------------------------------------------------
#include "LinearAlgebra.h"
#include "Plane3f.h"
#include "BoundingBox.h"
#include <X4Assert.h>
//---------------------------------------------------------------------------

namespace X4
{
/**
   required methods for type Vertex:
    operator +
    operator -
    operator *
    Vector3f getPos3f() const
*/

const float COINCIDING_EPSILON=0.0000001f;

template<class Vertex> class Polygon
{
private:

   std::vector<Vertex> m_Points;

public:

   typedef Polygon<Vertex> MyType;


   Polygon();
   Polygon( const Polygon &p );
   void operator = ( const Polygon &p );

   void addVertex( const Vertex &vec );
   inline card32 getNumVertices() const;
   void clear() {m_Points.clear();}

   inline Vertex & operator [] ( card32 i );
   inline const Vertex & operator [] ( card32 i ) const;

   inline bool isValid() const;

   EPlaneSide classify( const Plane3f &plane ) const;

   /// calculates normalized normal. if *this is a degenerated polygon, then valid is false
   /// and return std::vector is (0,0,0)
   Vector3f calculateNormal( bool &valid ) const;

   /// calculates polygon-plane, throws exception if polygon is invalid
   Plane3f calculatePlane() const;

   /// cut polygon at plane, polygon-part on the negative plane-side will be removed
   /// when the entire polygon is on the negative plane-side the polygon will have no points
   Polygon cut( const Plane3f &plane ) const;

   /// cut parts outside bounding box, returns empty polygon if polygon (this) is outside of the bounding box bb
   Polygon cut( const BoundingBox3f &bb) const;

   /// split polygon into positive & negative side of the plane, returns empty polygons if polygon
   /// is entirely on the positive/negative side. coinciding polygons are on front side
   /// expects empty front,backPolygons
   void split( const Plane3f &plane, MyType &frontPoly, MyType &backPoly) const;

   /// returns true if poly intersect *this
   /// *** untested ***
   bool intersect( const MyType &poly );

   /// compute center of mass
   Vector3f computeCenter() const;

   /// compute bounding box that encloses polygon
   BoundingBox3f computeBB() const;

   /// convex test (all routines assume convex polygon, so this is usually to reject invalid polygons)
   /// a polygon with insufficient number of points is non-convex
   bool isConvex() const;

   /// computes maximal distance between two polygon points
   float computeMaxInnerEdgeLength() const;

   /// compute points with maximal distance
   void computeMaximalDistancePoints( mpcard &i0,mpcard &i1) const;

 
};

class PolyVector3f// : public Vector3f
{
public:
	Vector3f pos;
   inline Vector3f getPos3f() const {return pos;}
//   inline void assign( const Vector3f & v ) {(*this)[0] = v[0];(*this)[1] = v[1];(*this)[2] = v[2];}

   PolyVector3f operator + ( const PolyVector3f &v ) const {PolyVector3f ret; ret.pos = pos + v.pos; return ret;}
   PolyVector3f operator - ( const PolyVector3f &v ) const {PolyVector3f ret; ret.pos = pos - v.pos; return ret;}
   PolyVector3f operator * ( float f ) const {PolyVector3f ret; ret.pos = pos*f; return ret;}

   static PolyVector3f create( const Vector3f &v ) {PolyVector3f ret; ret.pos = v; return ret;}
};

typedef Polygon< PolyVector3f > SimplePolygon;


}
#include "Polygon.inline.h"



#endif
