//---------------------------------------------------------------------------
#ifndef RegularDataH
#define RegularDataH
//---------------------------------------------------------------------------
#include <x4math.h>
//---------------------------------------------------------------------------
#include "DynamicLinearAlgebra.h"
#include <X4Assert.h>
//---------------------------------------------------------------------------

namespace X4 {

class OutputObjectStream;
class InputObjectStream;


/// this class defines regularly sampled data of potentially higher dimension (such as images, volume data, tensors)
template<typename FloatType>
class RegularData {
 private:
   /// all elements in linear order
   DynamicVector<FloatType> data;
   /// number of spatial directions (1 = std::vector, 2 = matrix, 3 = volume, ...)
   card32 dimension;
   /// number of elements in each direction (width, height, depth, ...)
   DVectorI extends;

   /// helper routine for constructor(s)
   void setup(const DynamicVector<int32> &extends);
   template <typename IndexType, int indexDim>
   void setup(const StaticVector<IndexType, indexDim> &extends);

   /// helper routine for getSubDataSet
   static void RegularData<FloatType>::recursiveGather(
      const RegularData<FloatType> &source, RegularData<FloatType> &dest,
      const DVectorI &lowerSourcePos, DVectorI sourcePos,
      const DVectorI &regionExtends, int32 processingDim);


 public:
   // ---- constructures, element access
   /// create new reg.-data object; specify dimension (extends.size()) and extends
   inline RegularData<FloatType>();
   inline RegularData<FloatType>(const DVectorI &extends);
   inline RegularData<FloatType>(const RegularData<FloatType> &init);

   template <typename IndexType, int indexDim>
   inline RegularData<FloatType>(const StaticVector<IndexType, indexDim> &extends);

   inline card32 getDim() const;
   inline const DVectorI getExtends() const;
   inline int32 getExtends(card32 d) const;
   inline void setExtends(const DVectorI &extends);
   template <typename IndexType, int indexDim>
   inline void setExtends(const StaticVector<IndexType, indexDim> &extends);
   /// computes result[0] = ext[0], result[1] = ext[0] x ext[1], ...
   inline DVectorI computeComulativeSize() const;
   inline mpint getNumElements() const;

   inline FloatType& operator[](const unsigned &index);
   inline const FloatType& operator[](const unsigned &index) const;
   inline int32 calcIndex(const DVectorI &index) const;
   inline const FloatType getElement(const DVectorI &index) const;
   inline void setElement(const DVectorI &index, FloatType value);

   template <typename IndexType, int indexDim>
   inline int32 calcIndex(const StaticVector<IndexType, indexDim> &index) const;
   template <typename IndexType, int indexDim>
   inline const FloatType getElement(const StaticVector<IndexType, indexDim> &index) const;
   template <typename IndexType, int indexDim>
   inline void setElement(const StaticVector<IndexType, indexDim> &index, FloatType value);

   //  ----  operators +, -, *, /
   inline RegularData<FloatType> operator+(const RegularData<FloatType> &op) const;
   inline RegularData<FloatType> operator-(const RegularData<FloatType> &op) const;
   inline RegularData<FloatType> operator-() const;
   inline FloatType operator*(const RegularData<FloatType> &op) const;
   inline RegularData<FloatType> operator*(const FloatType &s) const;
   inline RegularData<FloatType> operator/(const FloatType &s) const;
   inline RegularData<FloatType> componentProduct(const RegularData<FloatType> &op) const;
   inline RegularData<FloatType> componentSubstractScalar(const FloatType &s) const;
   inline RegularData<FloatType> componentAddScalar(const FloatType &s) const;

   //  ---- operators +=, -=, *=, /=
   inline RegularData<FloatType> operator+=(const RegularData<FloatType> &op);
   inline RegularData<FloatType> operator-=(const RegularData<FloatType> &op);
   inline RegularData<FloatType> operator*=(const FloatType &s);
   inline RegularData<FloatType> operator/=(const FloatType &s);

   //  ---- operators =, ==, !=
   inline void operator=(const RegularData<FloatType> &op);
   inline bool operator==(const RegularData<FloatType> &op) const;
   inline bool operator!=(const RegularData<FloatType> &op) const;

   // ---- other computations
   inline FloatType maxComponent() const;
   inline FloatType minComponent() const;
   inline RegularData<FloatType> absoluteValues() const;
   /// copy subblock, target will be resized accordingly
   void getSubDataSet(const DVectorI &lowerPos, const DVectorI &regionExtends, RegularData<FloatType> &target) const;
   inline FloatType componentSum() const;
   inline FloatType average() const;
   /// shift and scale to zero mean and unit L2-norm
   inline void normalizeMeanNorm();
   inline void squareOfComponents();
   inline void squareRootOfComponents();
   inline void logOfComponents();
   inline void expOfComponents();
   inline void clear(FloatType value);

   //  ---- misc.
   void print();
   std::string toString() const;
   inline FloatType* dataPtr();
   inline const FloatType* dataPtr() const;
	inline const DynamicVector<FloatType> getDynamicVector()const;
	
   void write(OutputObjectStream *s) const;
   void read(InputObjectStream *s);	
};

template<typename FloatType>
FloatType normQuad(const RegularData<FloatType> &op);

template<typename FloatType>
FloatType norm(const RegularData<FloatType> &op);


typedef RegularData<float32> RegularDataF;
typedef RegularData<int32> RegularDataI;
typedef RegularData<int8> RegularDataUB;
typedef RegularData<Vector3f> RegularData3F;


}

#endif
