//---------------------------------------------------------------------------
#ifndef RegularDataInlineH
#define RegularDataInlineH
//---------------------------------------------------------------------------
#include <x4math.h>
//---------------------------------------------------------------------------
#include "RegularData.h"
//---------------------------------------------------------------------------
#include "DynamicLinearAlgebra.inline.h"
#include <X4Assert.h>
//---------------------------------------------------------------------------

namespace X4 {


template<typename FloatType>
RegularData<FloatType>::RegularData() {
   dimension = 0;
}

template<typename FloatType>
void RegularData<FloatType>::setup(const DVectorI &extends) {
   x4Assert(extends.size() > 0);
   this->extends = extends;
   x4Assert(extends.minComponent() > 0);
   this->dimension = (card32) extends.size();
   int32 dataSize = 1;
   for (card32 i=0; i<dimension; i++) {
      dataSize *= extends[i];
   }
   data.setDim(dataSize);
}

template<typename FloatType>
template <typename IndexType, int indexDim>
void RegularData<FloatType>::setup(const StaticVector<IndexType, indexDim> &extends) {
   this->dimension = indexDim;
   this->extends.setDim(indexDim);
   for (card32 i=0; i<indexDim; i++) {
      this->extends[i] = extends[i];
   }
   x4Assert(this->extends.minComponent() > 0);
   int32 dataSize = 1;
   for (card32 i=0; i<indexDim; i++) {
      dataSize *= extends[i];
   }
   data.setDim(dataSize);
}

template<typename FloatType>
inline RegularData<FloatType>::RegularData(const DVectorI &extends) {
   setup(extends);
}

template<typename FloatType>
template <typename IndexType, int indexDim>
inline RegularData<FloatType>::RegularData(const StaticVector<IndexType, indexDim> &extends) {
   setup(extends);
}


template<typename FloatType>
inline RegularData<FloatType>::RegularData(const RegularData<FloatType> &init) {
   this->dimension = init.dimension;
   this->data = init.data;
   this->extends = init.extends;
}

template<typename FloatType>
inline card32 RegularData<FloatType>::getDim() const {
   return dimension;
}

template<typename FloatType>
inline const DVectorI RegularData<FloatType>::getExtends() const {
   return extends;
}
template<typename FloatType>
inline int32 RegularData<FloatType>::getExtends(card32 d) const {
   return extends[d];
}

template<typename FloatType>
inline void RegularData<FloatType>::setExtends(const DVectorI &extends) {
   if (extends.getDim() != this->extends.getDim() || extends != this->extends) {
      setup(extends);
   }
}

template<typename FloatType>
inline DVectorI RegularData<FloatType>::computeComulativeSize() const {
   x4Assert(dimension >= 1);
   DVector2i result(dimension);
   result[0] = extends[0];
   for (card32 i=1; i<dimension; i++) {
      result[i] = result[i-1] * extends[i];
   }
   return result;
}

template<typename FloatType>
inline mpint RegularData<FloatType>::getNumElements() const {
   return data.size();
}

template<typename FloatType>
template <typename IndexType, int indexDim>
inline void RegularData<FloatType>::setExtends(const StaticVector<IndexType, indexDim> &extends) {
   bool change = false;
   if (this->extends.getDim() != indexDim) {
      change = true;
   } else {
      for (card32 i=0; i<indexDim; i++) {
         if (extends[i] != this->extends[i]) {
            change = true;
            break;
         }
      }
   }
   if (change) {
      setup<IndexType, indexDim>(extends);
   }
}

template<typename FloatType>
inline FloatType& RegularData<FloatType>::operator[](const unsigned &index) {
   return data[index];
}

template<typename FloatType>
inline const FloatType& RegularData<FloatType>::operator[](const unsigned &index) const{
   return data[index];
}

template<typename FloatType>
inline int32 RegularData<FloatType>::calcIndex(const DVectorI &index) const {
   int32 dataIndex = index[dimension-1];
   for (int32 i=dimension-2; i>=0; i--) {
      dataIndex *= extends[i];
      dataIndex += index[i];
      x4Assert(index[i] >= 0 && index[i] < extends[i]);
   }
   return dataIndex;
}

template<typename FloatType>
inline const FloatType RegularData<FloatType>::getElement(const DVectorI &index) const {
   return data[calcIndex(index)];
}

template<typename FloatType>
inline void RegularData<FloatType>::setElement(const DVectorI &index, FloatType value) {
   data[calcIndex(index)] = value;
}

template<typename FloatType>
template <typename IndexType, int indexDim>
inline int32 RegularData<FloatType>::calcIndex(const StaticVector<IndexType, indexDim> &index) const {
   x4Assert(dimension == indexDim);
   int32 dataIndex = index[dimension-1];
   for (int32 i=dimension-2; i>=0; i--) {
      dataIndex *= extends[i];
      dataIndex += index[i];
      x4Assert(index[i] >= 0 && index[i] < extends[i]);
   }
   return dataIndex;
}

template<typename FloatType>
template <typename IndexType, int indexDim>
inline const FloatType RegularData<FloatType>::getElement(const StaticVector<IndexType, indexDim> &index) const {
   return data[calcIndex(index)];
}

template<typename FloatType>
template <typename IndexType, int indexDim>
inline void RegularData<FloatType>::setElement(const StaticVector<IndexType, indexDim> &index, FloatType value) {
   data[calcIndex<IndexType, indexDim>(index)] = value;
}

template<typename FloatType>
inline RegularData<FloatType> RegularData<FloatType>::operator+(const RegularData<FloatType> &op) const {
   RegularData<FloatType> result(extends);
   result.data = data + op.data;
   return result;
}

template<typename FloatType>
inline RegularData<FloatType> RegularData<FloatType>::operator-(const RegularData<FloatType> &op) const {
   RegularData<FloatType> result(extends);
   result.data = data - op.data;
   return result;
}

template<typename FloatType>
inline RegularData<FloatType> RegularData<FloatType>::operator-() const {
   RegularData<FloatType> result(extends);
   result.data = -data;
   return result;
}

template<typename FloatType>
inline FloatType RegularData<FloatType>::operator*(const RegularData<FloatType> &op) const {
   return data * op.data;
}

template<typename FloatType>
inline RegularData<FloatType> RegularData<FloatType>::operator*(const FloatType &s) const {
   RegularData<FloatType> result(extends);
   result.data = data * s;
   return result;
}

template<typename FloatType>
inline RegularData<FloatType> RegularData<FloatType>::operator/(const FloatType &s) const {
   RegularData<FloatType> result(extends);
   result.data = data / s;
   return result;
}

template<typename FloatType>
inline RegularData<FloatType> RegularData<FloatType>::componentProduct(const RegularData<FloatType> &op) const {
   RegularData<FloatType> result(extends);
   result.data = data.componentProduct(op.data);
   return result;
}

template<typename FloatType>
inline RegularData<FloatType> RegularData<FloatType>::componentSubstractScalar(const FloatType &s) const {
   RegularData<FloatType> result(extends);
   result.data = data.componentSubstractScalar(s);
   return result;
}

template<typename FloatType>
inline RegularData<FloatType> RegularData<FloatType>::componentAddScalar(const FloatType &s) const {
   RegularData<FloatType> result(extends);
   result.data = data.componentAddScalar(s);
   return result;
}

template<typename FloatType>
inline RegularData<FloatType> RegularData<FloatType>::operator+=(const RegularData<FloatType> &op) {
   data += op.data;
   return *this;
}

template<typename FloatType>
inline RegularData<FloatType> RegularData<FloatType>::operator-=(const RegularData<FloatType> &op) {
   data -= op.data;
   return *this;
}

template<typename FloatType>
inline RegularData<FloatType> RegularData<FloatType>::operator*=(const FloatType &s) {
   data *= s;
   return *this;
}

template<typename FloatType>
inline RegularData<FloatType> RegularData<FloatType>::operator/=(const FloatType &s) {
   data /= s;
   return *this;
}

template<typename FloatType>
inline void RegularData<FloatType>::operator=(const RegularData<FloatType> &op) {
   setup(op.extends);
   data = op.data;
}

template<typename FloatType>
inline bool RegularData<FloatType>::operator==(const RegularData<FloatType> &op) const {
   return extends == op.extends && data == op.data;
}

template<typename FloatType>
inline bool RegularData<FloatType>::operator!=(const RegularData<FloatType> &op) const {
   return extends != op.extends || data != op.data;
}

template<typename FloatType>
inline FloatType RegularData<FloatType>::maxComponent() const {
   return data.maxComponent();
}

template<typename FloatType>
inline FloatType RegularData<FloatType>::minComponent() const {
   return data.minComponent();
}

template<typename FloatType>
inline RegularData<FloatType> RegularData<FloatType>::absoluteValues() const {
   RegularData<FloatType> result(extends);
   result.data = data.absoluteValues();
   return result;
}

template<typename FloatType>
void RegularData<FloatType>::recursiveGather(const RegularData<FloatType> &source, RegularData<FloatType> &dest,
                                             const DVectorI &lowerSourcePos, DVectorI sourcePos,
                                             const DVectorI &regionExtends, int32 processingDim) {
   DVectorI destPos = sourcePos - lowerSourcePos;

   if (processingDim <= 0) {
      sourcePos[0] = lowerSourcePos[0];
      destPos[0] = 0;

      int32 sourceIndex = source.calcIndex(sourcePos);
      int32 destIndex = dest.calcIndex(destPos);

      for (int32 i=0; i<regionExtends[0]; i++) {
         dest[destIndex] = source[sourceIndex];
         destIndex++;
         sourceIndex++;
      }
   } else if (processingDim == 1) {
      destPos[0] = 0;
      destPos[1] = 0;

      int32 destIndex = dest.calcIndex(destPos);

      for (int32 y=0; y<regionExtends[1]; y++) {
         sourcePos[0] = lowerSourcePos[0];
         sourcePos[1] = lowerSourcePos[1] + y;
         int32 sourceIndex = source.calcIndex(sourcePos);
         for (int32 x=0; x<regionExtends[0]; x++) {
            dest[destIndex] = source[sourceIndex];
            destIndex++;
            sourceIndex++;
         }
      }
   } else {
      for (int32 i=0; i<regionExtends[processingDim]; i++) {
         sourcePos[processingDim] = i + lowerSourcePos[processingDim];
         recursiveGather(source, dest, lowerSourcePos, sourcePos, regionExtends, processingDim-1);
      }
   }
}

template<typename FloatType>
void RegularData<FloatType>::getSubDataSet(const DVectorI &lowerPos,
                                           const DVectorI &regionExtends,
                                           RegularData<FloatType> &target) const
{
   for (card32 i=0; i<dimension; i++) {
      if (lowerPos[i] < 0 || regionExtends[i] < 0 || lowerPos[i] + regionExtends[i] > extends[i]) {
         x4Assert(false);
      }
   }

   target.setExtends(regionExtends);

   if (dimension > 4) {
      recursiveGather(*this, target, lowerPos, lowerPos, regionExtends, (int32)regionExtends.size()-1);
   } else {
      if (dimension == 0) return;
      const int32 zerothDim = regionExtends[0];
      const int32 firstDim = dimension >= 2 ? regionExtends[1] : 1;
      const int32 secondDim = dimension >= 3 ? regionExtends[2] : 1;
      const int32 thirdDim = dimension >= 4 ? regionExtends[3] : 1;
      int32 sourceFirstIncrement = 0;
      int32 sourceSecondIncrement = 0;
      int32 sourceThirdIncrement = 0;
      if (dimension >= 2) {
         sourceFirstIncrement  = extends[0]
                               - regionExtends[0];
      }
      if (dimension >= 3) {
         sourceSecondIncrement = extends[0]*extends[1]
                               - regionExtends[0]*regionExtends[1]
                               - regionExtends[1]*sourceFirstIncrement;
      }
      if (dimension >= 4) {
         sourceThirdIncrement = extends[0]*extends[1]*extends[2]
                              - regionExtends[0]*regionExtends[1]*regionExtends[2]
                              - regionExtends[1]*regionExtends[2]*sourceFirstIncrement
                              - regionExtends[2]*sourceSecondIncrement;
      }

      FloatType *targetPtr = target.dataPtr();
      const FloatType *sourcePtr = dataPtr() + calcIndex(lowerPos);

      for (int32 p3 = 0; p3 < thirdDim; p3++) {
         for (int32 p2 = 0; p2 < secondDim; p2++) {
            for (int32 p1 = 0; p1 < firstDim; p1++) {
               for (int32 p0 = 0; p0 < zerothDim; p0++) {
                  *targetPtr = *sourcePtr;
                  targetPtr++;
                  sourcePtr++;
               }
               sourcePtr += sourceFirstIncrement;
            }
            sourcePtr += sourceSecondIncrement;
         }
         sourcePtr += sourceThirdIncrement;
      }
   }
}


template<typename FloatType>
inline FloatType RegularData<FloatType>::componentSum() const {
   return data.componentSum();
}


template<typename FloatType>
inline FloatType RegularData<FloatType>::average() const {
   return data.average();
}

template<typename FloatType>
inline void RegularData<FloatType>::normalizeMeanNorm() {
   data = data.componentSubstractScalar(data.average());
   FloatType mynorm = norm(data);
   if (mynorm >= 1E-7) {
      data /= mynorm;
   }
}


template<typename FloatType>
inline void RegularData<FloatType>::squareOfComponents() {
   mpint size = data.getDim();
   for (mpint i=0; i<size; i++) {
      data[i] = data[i]*data[i];
   }
}

template<typename FloatType>
inline void RegularData<FloatType>::squareRootOfComponents() {
   mpcard size = data.getDim();
   for (mpcard i=0; i<size; i++) {
      data[i] = sqrt(data[i]);
   }
}

template<typename FloatType>
inline void RegularData<FloatType>::logOfComponents() {
   mpint size = data.getDim();
   for (mpint i=0; i<size; i++) {
      data[i] = log(data[i]);
   }
}

template<typename FloatType>
inline void RegularData<FloatType>::expOfComponents() {
   mpint size = data.getDim();
   for (mpint i=0; i<size; i++) {
      data[i] = exp(data[i]);
   }
}

template<typename FloatType>
inline void RegularData<FloatType>::clear(FloatType value) {
   const mpcard n = data.size();
   FloatType *dataPtr = data.dataPtr();
   for (mpcard i=0; i<n; i++) {
      *dataPtr++ = value;
   }
}




template<typename FloatType>
void RegularData<FloatType>::print() {
   data.print();
}

template<typename FloatType>

std::string RegularData<FloatType>::toString() const {
   return data.toString();
}

template<typename FloatType>
inline FloatType* RegularData<FloatType>::dataPtr() {
   return data.data();
}

template<typename FloatType>
inline const FloatType* RegularData<FloatType>::dataPtr() const {
   return data.data();
}

template<typename FloatType>
inline const DynamicVector<FloatType> RegularData<FloatType>::getDynamicVector()const{
	return data;
}

template<typename FloatType>
FloatType normQuad(const RegularData<FloatType> &op) {
   return op*op;
}

template<typename FloatType>
FloatType norm(const RegularData<FloatType> &op) {
   return sqrt(normQuad(op));
}

template<typename FloatType>
void RegularData<FloatType>::write(OutputObjectStream *s) const {
   data.write(s);
   s->write(dimension);
   extends.write(s);
}

template<typename FloatType>
void RegularData<FloatType>::read(InputObjectStream *s) {
   data.read(s);
   s->read(dimension);
   extends.read(s);
}


}

#endif
 