//---------------------------------------------------------------------------
#ifndef ScalarDerivativeH
#define ScalarDerivativeH
//---------------------------------------------------------------------------
#include <x4math.h>
//---------------------------------------------------------------------------
#include <X4Types.h>
//---------------------------------------------------------------------------

namespace X4 {

template <typename FloatType>
class ScalarDerivative {
 protected:
	FloatType value;
	FloatType derivative;

 public:
	inline ScalarDerivative();
	inline ScalarDerivative(const FloatType &v, const FloatType &d);
	inline ScalarDerivative(const ScalarDerivative &op);
	/// constants are created implicitly by conversion; the derivative of constants is zero
	inline ScalarDerivative(const FloatType &op);
	/// variables have to be created explicitly; derivatives of variables are one
	static inline ScalarDerivative<FloatType> createVariable(const FloatType &op);

	inline FloatType getValue() const;
	inline FloatType getDerivative() const;
	inline void setValue(const FloatType &v);
	inline void setDerivative(const FloatType &v);

   inline ScalarDerivative operator+(const ScalarDerivative &op) const;
   inline ScalarDerivative operator-(const ScalarDerivative &op) const;
   inline ScalarDerivative operator*(const ScalarDerivative &s) const;
   inline ScalarDerivative operator/(const ScalarDerivative &s) const;
   inline void operator+=(const ScalarDerivative &op);
   inline void operator-=(const ScalarDerivative &op);
   inline void operator*=(const ScalarDerivative &op);
   inline void operator/=(const ScalarDerivative &op);

   inline ScalarDerivative operator+(const FloatType &op) const;
   inline ScalarDerivative operator-(const FloatType &op) const;
   inline ScalarDerivative operator*(const FloatType &s) const;
   inline ScalarDerivative operator/(const FloatType &s) const;
   inline void operator+=(const FloatType &s);
   inline void operator-=(const FloatType &s);
   inline void operator*=(const FloatType &s);
   inline void operator/=(const FloatType &s);

   inline ScalarDerivative operator-() const;

   inline void operator=(const ScalarDerivative op);
};

template<typename FloatType>
inline ScalarDerivative<ScalarDerivative<FloatType> > createVariable2ndOrder(const FloatType &x);

template<typename FloatType>
inline void extract2ndOrderDerivatives(
	const ScalarDerivative<ScalarDerivative<FloatType> > &var,
	FloatType &value,
	FloatType &derivative,
	FloatType &secondOrderDerivative
);

template <typename FloatType>
inline ScalarDerivative<FloatType> sin(const ScalarDerivative<FloatType> &op);
template <typename FloatType>
inline FloatType sin(const FloatType &op);

template <typename FloatType>
inline ScalarDerivative<FloatType> cos(const ScalarDerivative<FloatType> &op);
template <typename FloatType>
inline FloatType cos(const FloatType &op);

template <typename FloatType>
inline ScalarDerivative<FloatType> tan(const ScalarDerivative<FloatType> &op);

template <typename FloatType>
inline ScalarDerivative<FloatType> exp(const ScalarDerivative<FloatType> &op);
template <typename FloatType>
inline FloatType exp(const FloatType &op);

template <typename FloatType>
inline ScalarDerivative<FloatType> log(const ScalarDerivative<FloatType> &op);
template <typename FloatType>
inline FloatType log(const FloatType &op);

template <typename FloatType>
inline ScalarDerivative<FloatType> pow(
   const ScalarDerivative<FloatType> &op, 
   const ScalarDerivative<FloatType> &exponent);

typedef ScalarDerivative<float32> DerivativeF;
typedef ScalarDerivative<float64> DerivativeD;

typedef ScalarDerivative<DerivativeF> Derivative2ndOrderF;
typedef ScalarDerivative<DerivativeD> Derivative2ndOrderD;

inline Derivative2ndOrderF createVariable2ndOrderF(const float32 &x);
inline Derivative2ndOrderD createVariable2ndOrderD(const float64 &x);
inline void extract2ndOrderDerivativesF(
	const Derivative2ndOrderF &var,	
	float32 &value, float32 &derivative, float32 &secondOrderDerivative
);
inline void extract2ndOrderDerivativesD(
	const Derivative2ndOrderD &var,	
	float64 &value, float64 &derivative, float64 &secondOrderDerivative
);


void ScalarDerivative_unitTest();

}


#endif