//---------------------------------------------------------------------------
#ifndef ScalarDerivativeInlineH
#define ScalarDerivativeInlineH
//---------------------------------------------------------------------------
#include <x4math.h>
//---------------------------------------------------------------------------
#include "ScalarDerivative.h"
//---------------------------------------------------------------------------


namespace X4 {

template <typename FloatType>
inline ScalarDerivative<FloatType>::ScalarDerivative() {
	value = 0;
	derivative = 0;
}

template <typename FloatType>
inline ScalarDerivative<FloatType>::ScalarDerivative(const FloatType &v, const FloatType &d) {
	value = v;
	derivative = d;
}

template <typename FloatType>
inline ScalarDerivative<FloatType>::ScalarDerivative(const ScalarDerivative &op) {
	value = op.value;
	derivative = op.derivative;
}

template <typename FloatType>
inline ScalarDerivative<FloatType>::ScalarDerivative(const FloatType &op) {
	value = op;
	derivative = 0;
}

template <typename FloatType>
inline ScalarDerivative<FloatType> ScalarDerivative<FloatType>::createVariable(const FloatType &op) {
	ScalarDerivative<FloatType> result;
	result.value = op;
	result.derivative = 1;
	return result;
}


template <typename FloatType>
inline FloatType ScalarDerivative<FloatType>::getValue() const {
	return value;
}

template <typename FloatType>
inline FloatType ScalarDerivative<FloatType>::getDerivative() const {
	return derivative;
}

template <typename FloatType>
inline void ScalarDerivative<FloatType>::setValue(const FloatType &v) {
	value = v;
}

template <typename FloatType>
inline void ScalarDerivative<FloatType>::setDerivative(const FloatType &v) {
	derivative = v;
}

template <typename FloatType>
inline ScalarDerivative<FloatType> ScalarDerivative<FloatType>::operator+(const ScalarDerivative &op) const {
	ScalarDerivative<FloatType> result;
	result.value      = value      + op.value;
	result.derivative = derivative + op.derivative;
	return result;
}

template <typename FloatType>
inline ScalarDerivative<FloatType> ScalarDerivative<FloatType>::operator-(const ScalarDerivative &op) const {
	ScalarDerivative<FloatType> result;
	result.value      = value      - op.value;
	result.derivative = derivative - op.derivative;
	return result;
}

template <typename FloatType>
inline ScalarDerivative<FloatType> ScalarDerivative<FloatType>::operator*(const ScalarDerivative &op) const {
	ScalarDerivative<FloatType> result;
	result.value      = value * op.value;
	result.derivative = derivative*op.value + op.derivative*value;
	return result;
}

template <typename FloatType>
inline ScalarDerivative<FloatType> ScalarDerivative<FloatType>::operator/(const ScalarDerivative &op) const {
	ScalarDerivative<FloatType> result;
	result.value      = value / op.value;
	result.derivative = (derivative*op.value - op.derivative*value)/(op.value*op.value);
	return result;
}

template <typename FloatType>
inline void ScalarDerivative<FloatType>::operator+=(const ScalarDerivative &op) {
	value += op.value;
	derivative += op.derivative;
}

template <typename FloatType>
inline void ScalarDerivative<FloatType>::operator-=(const ScalarDerivative &op) {
	value -= op.value;
	derivative -= op.derivative;	
}

template <typename FloatType>
inline void ScalarDerivative<FloatType>::operator*=(const ScalarDerivative &op) {
	FloatType valueTmp = value * op.value;
	derivative         = derivative*op.value + value*op.derivative;
	value = valueTmp;
}

template <typename FloatType>
inline void ScalarDerivative<FloatType>::operator/=(const ScalarDerivative &op) {
	FloatType valueTmp = value / op.value;
	derivative         = (derivative*op.value - op.derivative*value)/(op.value*op.value);
	value = valueTmp;
}


template <typename FloatType>
inline ScalarDerivative<FloatType> ScalarDerivative<FloatType>::operator+(const FloatType &s) const {
	ScalarDerivative<FloatType> result;
	result.value      = value + s;
	result.derivative = derivative;
	return result;	
}

template <typename FloatType>
inline ScalarDerivative<FloatType> ScalarDerivative<FloatType>::operator-(const FloatType &s) const {
	ScalarDerivative<FloatType> result;
	result.value      = value - s;
	result.derivative = derivative;
	return result;	
}

template <typename FloatType>
inline ScalarDerivative<FloatType> ScalarDerivative<FloatType>::operator*(const FloatType &s) const {
	ScalarDerivative<FloatType> result;
	result.value      = value * s;
	result.derivative = derivative * s;
	return result;	
}

template <typename FloatType>
inline ScalarDerivative<FloatType> ScalarDerivative<FloatType>::operator/(const FloatType &s) const {
	ScalarDerivative<FloatType> result;
	result.value      = value / s;
	result.derivative = derivative / s;
	return result;	
}

template <typename FloatType>
inline void ScalarDerivative<FloatType>::operator+=(const FloatType &s) {
	value += op;
}

template <typename FloatType>
inline void ScalarDerivative<FloatType>::operator-=(const FloatType &s) {
	value -= op;
}

template <typename FloatType>
inline void ScalarDerivative<FloatType>::operator*=(const FloatType &s) {
	value *= op;
	derivative *= op;
}

template <typename FloatType>
inline void ScalarDerivative<FloatType>::operator/=(const FloatType &s) {
	value /= op;
	derivative /= op;
}


template <typename FloatType>
inline ScalarDerivative<FloatType> ScalarDerivative<FloatType>::operator-() const {
	ScalarDerivative<FloatType> result;
	result.value = -value;
	result.derivative = -derivative;
	return result;
}

template <typename FloatType>
inline void ScalarDerivative<FloatType>::operator=(const ScalarDerivative op) {
	value = op.value;
	derivative = op.derivative;
}



template<typename FloatType>
inline ScalarDerivative<ScalarDerivative<FloatType> > createVariable2ndOrder(const FloatType &x) {
	return ScalarDerivative<ScalarDerivative<FloatType> >(ScalarDerivative<FloatType>(x, 1), ScalarDerivative<FloatType>(1, 0) );
}

template<typename FloatType>
inline void extract2ndOrderDerivatives(
	const ScalarDerivative<ScalarDerivative<FloatType> > &var,
	FloatType &value,
	FloatType &derivative,
	FloatType &secondOrderDerivative
) {
	value = var.getValue().getValue();
	derivative = var.getDerivative().getValue();
	secondOrderDerivative = var.getDerivative().getDerivative();
}




template <typename FloatType>
inline ScalarDerivative<FloatType> sin(const ScalarDerivative<FloatType> &op) {
	ScalarDerivative<FloatType> result;
	result.setValue(sin(op.getValue()) );
	result.setDerivative(cos(op.getValue())*op.getDerivative());
	return result;
}
template <typename FloatType>
inline FloatType sin(const FloatType &op) {
	return ::sin(op);
}

template <typename FloatType>
inline ScalarDerivative<FloatType> cos(const ScalarDerivative<FloatType> &op) {
	ScalarDerivative<FloatType> result;
	result.setValue(cos(op.getValue()));
	result.setDerivative(-sin(op.getValue())*op.getDerivative());
	return result;
}
template <typename FloatType>
inline FloatType cos(const FloatType &op) {
	return ::cos(op);
}

template <typename FloatType>
inline ScalarDerivative<FloatType> tan(const ScalarDerivative<FloatType> &op) {
	return sin(op)/cos(op);
}

template <typename FloatType>
inline ScalarDerivative<FloatType> exp(const ScalarDerivative<FloatType> &op) {
	ScalarDerivative<FloatType> result;
	result.setValue(exp(op.getValue()));
	result.setDerivative(result.getValue()*op.getDerivative());
	return result;
}
template <typename FloatType>
inline FloatType exp(const FloatType &op) {
   return ::exp(op);
}

template <typename FloatType>
inline ScalarDerivative<FloatType> log(const ScalarDerivative<FloatType> &op) {
	ScalarDerivative<FloatType> result;
	result.setValue(log(op.getValue()));
	result.setDerivative((FloatType(1))/op.getValue()*op.getDerivative());
   return result;
}
template <typename FloatType>
inline FloatType log(const FloatType &op) {
   return ::log(op);
}

template <typename FloatType>
inline ScalarDerivative<FloatType> pow(
   const ScalarDerivative<FloatType> &base, 
   const ScalarDerivative<FloatType> &exponent) 
{
	return exp(log(base)*exponent);
}		

inline Derivative2ndOrderF createVariable2ndOrderF(const float32 &x) {
	return createVariable2ndOrder<float32>(x);
}
inline Derivative2ndOrderD createVariable2ndOrderD(const float64 &x) {
	return createVariable2ndOrder<float64>(x);
}
inline void extract2ndOrderDerivativesF(const Derivative2ndOrderF &var,	float32 &value, float32 &derivative, float32 &secondOrderDerivative) {
	extract2ndOrderDerivatives<float32>(var, value, derivative, secondOrderDerivative);
}
inline void extract2ndOrderDerivativesD(const Derivative2ndOrderD &var,	float64 &value, float64 &derivative, float64 &secondOrderDerivative) {
	extract2ndOrderDerivatives<float64>(var, value, derivative, secondOrderDerivative);
}


}

#endif