//---------------------------------------------------------------------------
#ifndef SparseLinearAlgebraH
#define SparseLinearAlgebraH
//---------------------------------------------------------------------------
#include <x4math.h>
//---------------------------------------------------------------------------
#include "DynamicLinearAlgebra.h"
#include <X4Types.h>
//---------------------------------------------------------------------------


namespace X4 {

    /// a sparse std::vector represented by list of (index, value) pairs
    /// efficient if a small number (O(1)) entries are non-zero
    template <typename FloatType>
    class SparseVector {
    private:
        struct Entry {
            mpcard index;
            FloatType value;
        };
        /// std::vector with entries, sorted by index (ascending)
    public:
        std::deque<Entry> entries;
        typedef typename std::deque<Entry> EVec;
        typedef typename std::deque<Entry>::iterator EIterator;
        typedef typename std::deque<Entry>::const_iterator EIteratorConst;
        EIterator begin() { return entries.begin(); }
        EIteratorConst begin() const { return entries.begin(); }
        EIterator end() { return entries.end(); }
        EIteratorConst end() const { return entries.end(); }

    public:
        //  ---- constructors, copy constructor, adm stuff
        inline SparseVector<FloatType>() {}
        inline SparseVector<FloatType>(const SparseVector<FloatType> &init);
        /// construct a sparse std::vector from init std::vector with dimensions shifted by startIndex
        inline SparseVector<FloatType>(const SparseVector<FloatType> &init, mpcard startIndex);
        inline mpcard getNumNonZeroEntries() const;
        /// remove all entries that contain only zeros
        inline void removeZeros();
        /// remove all near zero entries, return number of removed entries
        inline mpcard removeNearZeros(const FloatType &entriesSmallerThanThis);

        /// set all entries to zero (i.e. remove all entries from list)
        inline void clear();
        /// set the values of all entries to zero
        inline void refresh();

        // ---- element access
        inline const FloatType getEntry(const mpcard index) const;
        inline const FloatType operator[](const mpcard index) const;
        inline void setEntry(const mpcard &index, const FloatType value);
		inline void setEntrySorted(const mpcard &index, const FloatType value);
        inline void addEntry(const mpcard &index, const FloatType value);

        /// mb - add entry via binary search
        inline int addEntryBinary(const mpcard &index, const FloatType value);
        inline int setEntryBinary(const mpcard &index, const FloatType value);

        /// tries to use guessed entry position
        inline int addEntryBinaryWithGuessedPos(const mpcard &guessedEntryIndex, const FloatType value, const mpcard &index);
        inline int setEntryBinaryWithGuessedPos(const mpcard &guessedEntryIndex, const FloatType value, const mpcard &index);
        /************************************************************************/
        /* Following is added by peter                                                                     */
        /************************************************************************/
        /// This method is designed for adding group entries
        /************************************************************************/
        /* Combine the two std::vector using binary search                                                                     */
        /************************************************************************/
        inline	void	addEntries_Binary(
            const DVectorI *m_indices, 
            const DynamicVector<FloatType> *m_values, 
            const mpcard numOfEntries);

        inline	void	updateEntries_Binary(
            const	DynamicVector<FloatType> *m_values, 
            const	DVectorI		*m_indices,
            const	DVectorI		*m_mul,
            const	mpcard	numOfElements);

        inline	void	get_Value(mpcard matrixIndex, FloatType &m_value);

        inline	void	binary_Search(mpcard e_index, mpcard start, mpcard end, mpcard &index);

        inline	bool	hasEntry(mpcard	index);

        inline	void	initialize(const	std::vector<int>*m_indices);

        inline	void	initialize(const	std::vector<mpcard>* m_indices, const	std::vector<float32>*	m_values);

        /************************************************************************/
        /* Combine the two vectors using merging sort                                                                     */
        /************************************************************************/
        inline	void	addEntries_Merging(
            const DVectorI *m_indices, 
            const DynamicVector<FloatType> *m_values, 
            const mpcard numOfEntries);

        /************************************************************************/
        /* Above is added by peter                                                                                             */
        /************************************************************************/
        /// number of stored (potentially non-zero) entries
        inline mpcard getNumSparseEntries() const;
        /// get entry from list of stored entries
        inline void getSparseEntry(const mpcard entryIndex, mpcard &matrixIndex, FloatType &value) const;
        /// add entry at the end of the std::vector; entryIndex must be larger than all indices already present
        inline void pushBackEntry(const mpcard entryIndex, const FloatType value);
        /// element list access
        inline FloatType getListEntryValue(mpcard listIndex) const;
        inline mpcard getListEntryIndex(mpcard listIndex) const;
        inline void setListEntryValue(mpcard listIndex, const FloatType &v);
        inline void addListEntryValue(mpcard listIndex, const FloatType &v);
        inline void setListEntryIndex(mpcard listIndex, mpcard ind);
        inline void setNumEntries(mpcard numEntries);
        inline mpcard getLargestIndex() const { return entries.back().index; }
        //  ----  operators
        inline SparseVector<FloatType> operator+(const SparseVector<FloatType> &op) const;
        inline SparseVector<FloatType> operator-(const SparseVector<FloatType> &op) const;
        inline SparseVector<FloatType> operator-() const;
        inline FloatType operator*(const SparseVector<FloatType> &op) const;
        inline FloatType operator*(const DynamicVector<FloatType> &op) const;
        inline SparseVector<FloatType> operator*(const FloatType &s) const;
        inline SparseVector<FloatType> operator/(const FloatType &s) const;
        inline SparseVector<FloatType> componentProduct(const SparseVector<FloatType> &op) const;
        inline void operator+=(const SparseVector<FloatType> &op);
        inline void operator-=(const SparseVector<FloatType> &op);
        inline void operator*=(const FloatType &s);
        inline void operator/=(const FloatType &s);

        inline void operator=(const SparseVector<FloatType> &op);
        inline bool operator==(const SparseVector<FloatType> &op) const;
        inline bool operator!=(const SparseVector<FloatType> &op) const;
        inline bool operator==(const DynamicVector<FloatType> &op) const;
        inline bool operator!=(const DynamicVector<FloatType> &op) const;

        // ----  misc.
        std::string toString(bool printZeros = false, int printZerosLength = 0) const;
        std::string toMatString(size_t length = 0) const;
        void write(OutputObjectStream *s) const;
        void read(InputObjectStream *s);	
        /// assert sorted indices
        void checkConsistency();
    };


    /// Sparse matrix. stores rows as sparse vectors.
    /// Attention: the matrix elements are stored in row/column order
    /// (in contrast to DynamicMatrix, StaticMatrix).
    template<typename FloatType>
    class SparseMatrix {
    private:
        std::deque< SparseVector<FloatType> > theRows;
        typedef typename std::deque< SparseVector<FloatType> > EVec;
        typedef typename std::deque< SparseVector<FloatType> >::iterator EIterator;
        typedef typename std::deque< SparseVector<FloatType> >::const_iterator EIteratorConst;
        EIterator begin() { return theRows.begin(); }
        EIteratorConst& begin() const { return theRows.begin(); }
        EIterator end() { return theRows.end(); }
        EIteratorConst& end() const { return theRows.end(); }

    public:
        // ---- constructors, copy constructor
        inline SparseMatrix(mpcard numRows = 0);
        inline SparseMatrix(const SparseMatrix<FloatType> &init);

        /// set all entries to zero
        void setZero();
        /// alias for set zero
        inline void	clear();
        /// set the values of all entries to zero
        inline void	refresh();

        inline mpcard getRows() const;
        inline mpcard getNumRows() const {return theRows.size();}
        inline void setRows(mpcard numRows = 0);
        inline mpcard getNumCols() const;
        inline mpcard getCols() const { return getNumCols(); }

        /// retrieve ROW std::vector (not column!)
        inline SparseVector<FloatType>& operator[](const mpcard &index);
        /// retrieve ROW std::vector (not column!)
        inline const SparseVector<FloatType>& operator[](const mpcard &index) const;

        /// remove all zero entries in all vectors
        inline void removeZeros();

        /// remove all near zero entries in all vectors, returns number of removed entries
        inline mpcard removeNearZeros(const FloatType &entriesSmallerThanThis);

        //  ---- operators +=, -=, *=, /=
        inline void operator+=(const SparseMatrix<FloatType> &op);
        inline void operator-=(const SparseMatrix<FloatType> &op);
        inline void operator*=(const FloatType &op);
        inline void setToComponentProduct(const SparseMatrix<FloatType> &op);
        inline void operator/=(const FloatType &op);

        //  ---- operators =, ==, !=
        inline void operator=(const SparseMatrix<FloatType> &op);
        inline bool operator==(const SparseMatrix<FloatType> &op) const;
        inline bool operator!=(const SparseMatrix<FloatType> &op) const;

        // ---- multiplying with dynamic vectors. input can be of arbitrary length. output is of length "getRows()"
        inline DynamicVector<FloatType> operator*(const DynamicVector<FloatType> &v) const;
        /// same as operator*(DynamicVector), but faster. result must be of length "getRows()"
        inline void matVecMult(const DynamicVector<FloatType> &v, DynamicVector<FloatType> &result) const;

        // ---- multiplying with sparse vectors
        inline DynamicVector<FloatType> operator*(const SparseVector<FloatType> &v) const;

        // ----  misc.
        void write(OutputObjectStream *s) const;
        void read(InputObjectStream *s);	
        std::string toString(bool printZeros = false, int printZerosLength = 0) const;
        std::string toMatString(size_t length = 0) const;

        SparseMatrix<FloatType> transpose() const;

        /// add row to matrix
        void addRow( const SparseVector<FloatType> &row) {theRows.push_back(row);} 

        /// remove row
        void removeRow(mpcard i) {theRows.erase(theRows.begin()+i);}

        /// append all rows in M to matrix
        void appendRows( const SparseMatrix<FloatType> &M ) {theRows.insert(theRows.end(),M.theRows.begin(),M.theRows.end());}


    };


    // instances
    typedef SparseVector<int32> SparseVectorI;
    typedef SparseVector<float32> SparseVectorF;
    typedef SparseVector<float64> SparseVectorD;

    typedef SparseMatrix<int32> SparseMatrixI;
    typedef SparseMatrix<float32> SparseMatrixF;
    typedef SparseMatrix<float64> SparseMatrixD;


    /// perform various tests of sparse linear algebra operations to test implementation
    void sparse_linear_algebra_unitTest();


    template <typename FloatType1, typename FloatType2>
    inline void convertSparseVectorType(const SparseVector<FloatType1> &source, SparseVector<FloatType2> &dest);

    template <typename FloatType1, typename FloatType2>
    inline void convertSparseMatrixType(const SparseMatrix<FloatType1> &source, SparseMatrix<FloatType2> &dest);

}

#endif
