//---------------------------------------------------------------------------
#ifndef SparseLinearAlgebraInlineH
#define SparseLinearAlgebraInlineH
//---------------------------------------------------------------------------
#include "SparseLinearAlgebra.h"
#include "DynamicLinearAlgebra.inline.h"
//---------------------------------------------------------------------------

namespace X4 {



    // ---- SparseVector


    template <typename FloatType>
    inline SparseVector<FloatType>::SparseVector(const SparseVector<FloatType> &init, mpcard startIndex) {
        const mpcard numEntries = init.entries.size();
        if( numEntries == 0)
        {
            this->entries.clear();
            return;
        }
        entries = EVec(numEntries);
        Entry * entryPtr = &entries[0];
        Entry * srcEntryPtr = (Entry *)&init.entries[0];
        for( mpcard i=0;i<numEntries;i++ )
        {
            entryPtr->index = srcEntryPtr->index+startIndex;
            entryPtr->value = srcEntryPtr->value;
            entryPtr++;
            srcEntryPtr++;
        }
    }

    template <typename FloatType>
    inline SparseVector<FloatType>::SparseVector(const SparseVector<FloatType> &init) {
        this->entries = init.entries;
    }

    template <typename FloatType>
    inline mpcard SparseVector<FloatType>::getNumNonZeroEntries() const{
        return entries.size();
    }

    template <typename FloatType>
    inline void SparseVector<FloatType>::removeZeros() {
        EVec newEntries;
        EIteratorConst pos = entries.begin();
        EIteratorConst end = entries.end();
        while (pos != end) {
            if (pos->value != (FloatType)0) {
                newEntries.push_back(*pos);
            }
            ++pos;
        }
        entries = newEntries;
    }

    template <typename FloatType>
    inline mpcard SparseVector<FloatType>::removeNearZeros(const FloatType &entriesSmallerThanThis) {
        EVec newEntries;
        EIteratorConst pos = entries.begin();
        EIteratorConst end = entries.end();
        while (pos != end) {
            if ( fabs(pos->value) >= entriesSmallerThanThis) {
                newEntries.push_back(*pos);
            }
            ++pos;
        }
        mpcard differenceNum = entries.size() - newEntries.size();
        entries = newEntries;
        return differenceNum;
    }

    template <typename FloatType>
    inline void SparseVector<FloatType>::clear() {
        entries.clear();
    }

    template <typename FloatType>
    inline void SparseVector<FloatType>::refresh() 
    {
        for(EIterator	iter = entries.begin();
            iter != entries.end();
            iter++)
        {
            iter->value -= iter->value;
        }
    }

    template <typename FloatType>
    inline const FloatType SparseVector<FloatType>::getEntry(const mpcard index) const {
        EIteratorConst pos = entries.begin();
        EIteratorConst end = entries.end();
        while (pos != end) {
            if (pos->index >= index) {
                if (pos->index == index) {
                    return pos->value;
                } else {
                    return (FloatType)0;
                }
            }
            ++pos;
        }
        return (FloatType)0;
    }

    template <typename FloatType>
    inline const FloatType SparseVector<FloatType>::operator[](const mpcard index) const {
        return getEntry(index);
    }

    template <typename FloatType>
    inline void SparseVector<FloatType>::setEntry(const mpcard &index, const FloatType value) {
        EIterator pos = entries.begin();
        EIterator end = entries.end();
        while (pos != end) {
            if (pos->index >= index) {
                if (pos->index == index) {
                    pos->value = value;
                    return;
                } else {
                    Entry e; e.index = index; e.value = value;
                    entries.insert(pos, e);
                    return;
                }
            }
            ++pos;
        }
        Entry e; e.index = index; e.value = value;
        entries.push_back(e);
    }


	template <typename FloatType>
	inline void SparseVector<FloatType>::setEntrySorted(const mpcard &index, const FloatType value) {
		Entry e; e.index = index; e.value = value;
		entries.push_back(e);
	}

    template <typename FloatType>
    inline void SparseVector<FloatType>::addEntry(
        const mpcard &index, 
        const FloatType value) 
    {
        EIterator pos = entries.begin();
        EIterator end = entries.end();
        while (pos != end) 
        {
            if (pos->index >= index) 
            {
                if (pos->index == index) 
                {
                    pos->value += value;
                    return;
                } 
                else 
                {
                    Entry e; e.index = index; e.value = value;
                    entries.insert(pos, e);
                    return;
                }
            }
            ++pos;
        }
        Entry e; e.index = index; e.value = value;
        entries.push_back(e);
    }


    template <typename FloatType>
    inline int SparseVector<FloatType>::addEntryBinaryWithGuessedPos( const mpcard &guessedEntryIndex, const FloatType value, const mpcard &index )
    {
        if (entries.size() <= guessedEntryIndex || entries[guessedEntryIndex].index != index ){
            return addEntryBinary(index, value);
        }
        else{
            entries[guessedEntryIndex].value += value;
            return -1;
        }
    }

    template <typename FloatType>
    inline int SparseVector<FloatType>::setEntryBinaryWithGuessedPos( const mpcard &guessedEntryIndex, const FloatType value, const mpcard &index )
    {
        if (entries.size() <= guessedEntryIndex || entries[guessedEntryIndex].index != index ){
            return addEntryBinary(index, value);
        }
        else{
            entries[guessedEntryIndex].value = value;
            return -1;
        }
    }

    template <typename FloatType>
    inline int SparseVector<FloatType>::addEntryBinary(
        const mpcard &index, 
        const FloatType value) 
    {
        if( fabs(value) < 1e-15f )
            return -42;

        if( entries.empty() )
        {
            Entry e; e.index = index; e.value = value;
            entries.push_back(e);
            return (int)entries.size()-1;
        }

        // happens probably most often
        if( entries.back().index <= index )
        {
            if( entries.back().index == index )
            {
                entries.back().value += value;
                return (int)entries.size()-1;
            }
            Entry e; e.index = index; e.value = value;
            entries.push_back(e);
            return (int)entries.size()-1;
        }

        if( entries.front().index >= index)
        {
            if( entries.front().index == index)
            {
                entries.front().value += value;
                return 0;
            }
            Entry e; e.index = index; e.value = value;
            entries.insert(entries.begin(),e);
            return -1;
        }


        mpcard first = 0;
        mpcard last = entries.size()-1;
        while(first+1<last)
        {
            mpcard middle = (first+last)/2;
            Entry &middle_Entry = entries[middle];
            if( middle_Entry.index <= index)
            {
                if( middle_Entry.index == index)
                {
                    middle_Entry.value += value;
                    return middle;
                }
                first = middle;
            }
            else
            {
                last = middle;
            }
        }
        if( entries[last].index == index)
        {
            entries[last].value += value;
            return last;
        }
        else
            if( entries[first].index == index)
            {
                entries[first].value += value;
                return first;
            }
            else
            {
                Entry e; e.index = index; e.value = value;
                entries.insert(entries.begin()+last, e);
                return -1;
            }
    }


    template <typename FloatType>
    inline int SparseVector<FloatType>::setEntryBinary(
        const mpcard &index, 
        const FloatType value) 
    {
        if( fabs(value) < 1e-15f )
            return -42;

        if( entries.empty() )
        {
            Entry e; e.index = index; e.value = value;
            entries.push_back(e);
            return (int)entries.size()-1;
        }

        // happens probably most often
        if( entries.back().index <= index )
        {
            if( entries.back().index == index )
            {
                entries.back().value = value;
                return (int)entries.size()-1;
            }
            Entry e; e.index = index; e.value = value;
            entries.push_back(e);
            return (int)entries.size()-1;
        }

        if( entries.front().index >= index)
        {
            if( entries.front().index == index)
            {
                entries.front().value = value;
                return 0;
            }
            Entry e; e.index = index; e.value = value;
            entries.insert(entries.begin(),e);
            return -1;
        }


        mpcard first = 0;
        mpcard last = entries.size()-1;
        while(first+1<last)
        {
            mpcard middle = (first+last)/2;
            Entry &middle_Entry = entries[middle];
            if( middle_Entry.index <= index)
            {
                if( middle_Entry.index == index)
                {
                    middle_Entry.value = value;
                    return middle;
                }
                first = middle;
            }
            else
            {
                last = middle;
            }
        }
        if( entries[last].index == index)
        {
            entries[last].value = value;
            return last;
        }
        else
            if( entries[first].index == index)
            {
                entries[first].value = value;
                return first;
            }
            else
            {
                Entry e; e.index = index; e.value = value;
                entries.insert(entries.begin()+last, e);
                return -1;
            }
    }

    /************************************************************************/
    /* Following is added by peter                                                                     */
    /************************************************************************/

    template <typename FloatType>
    inline void SparseVector<FloatType>::addEntries_Binary(
        const DVectorI *m_indices, 
        const DynamicVector<FloatType> *m_values, 
        const mpcard numOfEntries)
    {
        EIterator		pos		=	entries.begin();
        size_t length; mpcard id = 0, offset;
        while (pos != entries.end() && id < numOfEntries) 
        {
            mpcard index = (*m_indices)[id];
            if (pos->index >= index) 
            {
                if (pos->index == index) 
                {
                    pos->value += (*m_values)[id];
                    pos++;
                }
                else 
                {
                    Entry e; 
                    e.index = index; 
                    e.value = (*m_values)[id];
                    offset = pos - entries.begin();
                    entries.insert(pos, e);
                    pos = entries.begin() + offset+1;
                }
                id++;
            }
            else
            {
                EIterator begin = entries.begin();
                length = entries.size()-1;
                if(int((begin+length)->index) < (*m_indices)[id])
                    break;
                mpcard	s_id		= pos - begin + 1;
                mpcard	e_id		= (mpcard)length;
                while (s_id < e_id - 1)
                {
                    mpcard mid = (s_id+e_id)/2;
                    pos = begin + mid;
                    if(int(pos->index) < (*m_indices)[id])
                        s_id = mid;
                    else
                        e_id = mid;
                }
                pos = begin + s_id;
                if(int(pos->index) < (*m_indices)[id])
                    pos++;
            }
        }
        while(id < numOfEntries)
        {
            Entry e; 
            e.index = (*m_indices)[id]; 
            e.value = (*m_values)[id];
            entries.push_back(e);
            id++;
        }
    }

    template <typename FloatType>
    inline void SparseVector<FloatType>::updateEntries_Binary(
        const	DynamicVector<FloatType> *m_values, 
        const	DVectorI		*m_indices,
        const	DVectorI		*m_mul,
        const	mpcard	numOfElements)
    {
        EIterator		begin, pos;
        pos = begin = entries.begin();

        size_t length = entries.size()-1; mpcard id = 0, offset;
        for(mpcard i = 0; i < numOfElements; i++)
        {
            mpcard	s_id		= 0;
            if((*m_indices)[i] >= 0)
            {
                if(entries.size() == 0 || (*m_indices)[i]> int32(entries[entries.size()-1].index))
                {
                    for(mpcard j = 0; j < mpcard((*m_mul)[i]); j++)
                    {
                        Entry e;
                        e.value = (*m_values)[id];
                        e.index = (*m_indices)[i]+j;
                        entries.push_back(e);
                        id++;
                    }
                    continue;
                }
                else if((*m_indices)[i] < int32(entries[0].index))
                {
                    for(mpcard j = 0; j < mpcard((*m_mul)[i]); j++)
                    {
                        Entry e;
                        e.value = (*m_values)[id];
                        e.index = (*m_indices)[i]+j;
                        entries.insert(entries.begin()+j, e);
                        id++;
                    }
                    continue;
                }
                else
                {
                    binary_Search((*m_indices)[i], 0, (mpcard)entries.size()-1, offset);
                    pos = entries.begin() + offset;
                    if(pos->index == (*m_indices)[i])
                    {
                        for(mpcard j = 0; j < mpcard((*m_mul)[i]); j++)
                        {
                            pos->value += (*m_values)[id];
                            pos++;
                            id++;
                        }
                    }
                    else
                    {
                        for(mpcard j = 0; j < mpcard((*m_mul)[i]); j++)
                        {
                            Entry e;
                            e.value = (*m_values)[id];
                            e.index = (*m_indices)[i]+j;
                            entries.insert(entries.begin()+offset+j, e);
                            id++;
                        }
                    }
                }
            }
            else
            {
                id += (*m_mul)[i];
            }
        }
    }

    template <typename FloatType>
    inline void SparseVector<FloatType>::get_Value(mpcard matrixIndex, FloatType &m_value)
    {
        mpcard	index;
        binary_Search(matrixIndex, 0, (mpcard)entries.size()-1, index);
        m_value = entries[index].value;
    }

    template <typename FloatType>
    inline void SparseVector<FloatType>::binary_Search(mpcard e_index, mpcard start, mpcard end, mpcard &index)
    {
        if(entries.size() == 0 || e_index <= entries[start].index)
        {
            index = start;
            return;
        }
        else if(e_index >= entries[end].index)
        {
            index = end;
            return;
        }
        else
        {
            mpcard		s_id = start, e_id = end;
            EIterator		begin = entries.begin();
            while (s_id < e_id - 1)
            {
                mpcard	mid = (s_id+e_id)/2;
                EIterator	pos = begin + mid;
                if(int32(pos->index) < e_index)
                    s_id = mid;
                else if(pos->index > e_index)
                    e_id = mid;
                else
                {
                    index = mid;
                    return;
                }
            }
            if(entries[s_id].index == e_index)
            {
                index	=	 s_id;
                return;
            }
            else
            {
                index =  e_id;
                return;
            }
        }
    }

    template<typename FloatType>
    inline bool SparseVector<FloatType>::hasEntry(mpcard e_index)
    {
        if (entries.empty()) return false;
        if(e_index < entries[0].index || e_index > entries[entries.size()-1].index )
            return false;

        mpcard	index;
        binary_Search(e_index, 0, (mpcard)entries.size()-1, index);
        return entries[index].index == e_index;
    }

    template <typename FloatType>
    inline void SparseVector<FloatType>::initialize(const	std::vector<int>*m_indices)
    {
        size_t	m_dimension = m_indices->size();
        entries.resize(m_dimension);
        EIterator		pos = entries.begin(); 
        for(mpcard i = 0; i < m_dimension; i++)
        {
            pos->index = (*m_indices)[i];
            pos->value = 0.0;
            pos++;
        }
    }

    template <typename FloatType>
    inline void SparseVector<FloatType>::initialize(const	std::vector<mpcard>*	m_indices, const	std::vector<float32>* m_values)
    {
        mpcard		m_dimension = (mpcard)m_indices->size();
        entries.resize(m_dimension);
        EIterator		pos = entries.begin(); 
        for(mpcard i = 0; i < m_dimension; i++)
        {
            pos->index = (*m_indices)[i];
            pos->value = (*m_values)[i];
            pos++;
        }
    }

    template <typename FloatType>
    inline void SparseVector<FloatType>::addEntries_Merging(
        const DVectorI *m_indices, 
        const DynamicVector<FloatType> *m_values, 
        const mpcard numOfEntries)
    {
        EIterator		pos	= entries.begin();
        mpcard		id		=	0,offset;
        while (pos != entries.end() && id < numOfEntries) 
        {
            mpcard index = (*m_indices)[id];
            if (pos->index >= index) 
            {
                if (pos->index == index) 
                {
                    pos->value += (*m_values)[id];
                }
                else 
                {
                    Entry e; 
                    e.index = index; 
                    e.value = (*m_values)[id];
                    offset = pos - entries.begin();
                    entries.insert(pos, e);
                    pos = entries.begin() + offset+1;
                }
                id++;
            }
            else
            {
                ++pos;
            }
        }
        while(id < numOfEntries)
        {
            Entry e; 
            e.index = (*m_indices)[id]; 
            e.value = (*m_values)[id];
            entries.push_back(e);
            id++;
        }
    }

    /************************************************************************/
    /* Above is added by peter                                                                   */
    /************************************************************************/

    template <typename FloatType>
    inline mpcard SparseVector<FloatType>::getNumSparseEntries() const {
        return (mpcard)entries.size();
    }

    template <typename FloatType>
    inline void SparseVector<FloatType>::getSparseEntry(const mpcard entryIndex, mpcard &matrixIndex, FloatType &value) const {
        x4Assert(entryIndex < entries.size());
        matrixIndex = entries[entryIndex].index;
        value = entries[entryIndex].value;
    }

    template <typename FloatType>
    inline void SparseVector<FloatType>::pushBackEntry(const mpcard entryIndex, const FloatType value) {
        x4Assert(entries.empty() || entries.back().index < entryIndex);
        Entry e;
        e.index = entryIndex;
        e.value = value;
        entries.push_back(e);
    }

    template <typename FloatType>
    inline FloatType SparseVector<FloatType>::getListEntryValue(mpcard listIndex) const {
        x4Assert(listIndex < entries.size());
        return entries[listIndex].value;
    }

    template <typename FloatType>
    inline mpcard SparseVector<FloatType>::getListEntryIndex(mpcard listIndex) const {
        x4Assert(listIndex < entries.size());
        return entries[listIndex].index;
    }

    template <typename FloatType>
    inline void SparseVector<FloatType>::setListEntryValue(mpcard listIndex, const FloatType &v) {
        x4Assert(listIndex < entries.size());
        entries[listIndex].value = v;
    }

    template <typename FloatType>
    inline void SparseVector<FloatType>::addListEntryValue(mpcard listIndex, const FloatType &v) {
        x4Assert(listIndex < entries.size());
        entries[listIndex].value += v;
    }


    template <typename FloatType>
    inline void SparseVector<FloatType>::setListEntryIndex(mpcard listIndex, mpcard ind) {
        x4Assert(listIndex < entries.size());
        entries[listIndex].index = ind;
    }

    template <typename FloatType>
    inline void SparseVector<FloatType>::setNumEntries(mpcard numEntries) {
        entries.clear();
        entries.resize(numEntries);
    }



    template <typename FloatType>
    inline SparseVector<FloatType> SparseVector<FloatType>::operator+(const SparseVector<FloatType> &op) const {
        SparseVector<FloatType> result;
        EIteratorConst posThis = entries.begin();
        EIteratorConst endThis = entries.end();
        EIteratorConst posOp = op.entries.begin();
        EIteratorConst endOp = op.entries.end();
        while (posThis != endThis && posOp != endOp) {
            if (posThis->index < posOp->index) {
                result.entries.push_back(*posThis);
                ++posThis;
            } else if (posThis->index > posOp->index) {
                result.entries.push_back(*posOp);
                ++posOp;
            } else {
                Entry e;
                x4Assert(posThis->index == posOp->index);
                e.index = posThis->index;
                e.value = posThis->value + posOp->value;
                result.entries.push_back(e);
                ++posOp;
                ++posThis;
            }
        }
        x4Assert(posThis == endThis || posOp == endOp);
        while (posThis != endThis) {
            result.entries.push_back(*posThis);
            ++posThis;
        }
        while (posOp != endOp) {
            result.entries.push_back(*posOp);
            ++posOp;
        }

        return result;
    }

    template <typename FloatType>
    inline SparseVector<FloatType> SparseVector<FloatType>::operator-(const SparseVector<FloatType> &op) const {
        SparseVector<FloatType> result;
        EIteratorConst posThis = entries.begin();
        EIteratorConst endThis = entries.end();
        EIteratorConst posOp = op.entries.begin();
        EIteratorConst endOp = op.entries.end();
        while (posThis != endThis && posOp != endOp) {
            if (posThis->index < posOp->index) {
                result.entries.push_back(*posThis);
                ++posThis;
            } else if (posThis->index > posOp->index) {
                Entry e = *posOp; e.value = -e.value;
                result.entries.push_back(e);
                ++posOp;
            } else {
                Entry e;
                x4Assert(posThis->index == posOp->index);
                e.index = posThis->index;
                e.value = posThis->value - posOp->value;
                result.entries.push_back(e);
                ++posThis;
                ++posOp;
            }
        }
        x4Assert(posThis == endThis || posOp == endOp);
        while (posThis != endThis) {
            result.entries.push_back(*posThis);
            ++posThis;
        }
        while (posOp != endOp) {
            Entry e = *posOp; e.value = -e.value;
            result.entries.push_back(e);
            ++posOp;
        }

        return result;
    }

    template <typename FloatType>
    inline SparseVector<FloatType> SparseVector<FloatType>::operator-() const {
        SparseVector<FloatType> result; /*result.entries.reserve(entries.size())*/;
        EIteratorConst pos = entries.begin();
        EIteratorConst end = entries.end();
        while (pos != end) {
            Entry e = *pos;
            e.value = -e.value;
            result.entries.push_back(e);
            ++pos;
        }
        return result;
    }

    template <typename FloatType>
    inline FloatType SparseVector<FloatType>::operator*(const SparseVector<FloatType> &op) const {
        FloatType result = (FloatType)0;
        EIteratorConst posThis = entries.begin();
        EIteratorConst endThis = entries.end();
        EIteratorConst posOp = op.entries.begin();
        EIteratorConst endOp = op.entries.end();
        while (posThis != endThis && posOp != endOp) {
            if (posThis->index < posOp->index) {
                ++posThis;
            } else if (posThis->index > posOp->index) {
                ++posOp;
            } else {
                result += posThis->value * posOp->value;
                ++posThis;
                ++posOp;
            }
        }
        return result;
    }

    template <typename FloatType>
    inline FloatType SparseVector<FloatType>::operator*(const DynamicVector<FloatType> &op) const {
        FloatType result = (FloatType)0;

        EIteratorConst posThis = entries.begin();
        EIteratorConst endThis = entries.end();
        while (posThis != endThis) {
            x4Assert(posThis->index < op.getDim());
            result += op[posThis->index] * posThis->value;
            ++posThis;
        }

        //const size_t numEntries = entries.size();
        //if( numEntries == 0)
        //    return result;
        //Entry * ptr = (Entry*)&entries[0];
        //FloatType * srcPtr = (FloatType*)op.data();
        //for( int i=0;i<(int)numEntries;i++)
        //{
        //    result += ptr->value * srcPtr[ptr->index];
        //    ptr++;
        //}

        return result; 
    }

    template <typename FloatType>
    inline SparseVector<FloatType> SparseVector<FloatType>::operator*(const FloatType &s) const {
        SparseVector<FloatType> result;
        EIteratorConst posThis = entries.begin();
        EIteratorConst endThis = entries.end();
        while (posThis != endThis) {
            Entry e;
            e = *posThis;
            e.value *= s;
            result.entries.push_back(e);
            ++posThis;
        }
        return result;
    }

    template <typename FloatType>
    inline SparseVector<FloatType> SparseVector<FloatType>::operator/(const FloatType &s) const {
        SparseVector<FloatType> result;
        EIteratorConst posThis = entries.begin();
        EIteratorConst endThis = entries.end();
        while (posThis != endThis) {
            Entry e;
            e = *posThis;
            e.value /= s;
            result.entries.push_back(e);
            ++posThis;
        }
        return result;
    }

    template <typename FloatType>
    inline SparseVector<FloatType> SparseVector<FloatType>::componentProduct(const SparseVector<FloatType> &op) const {
        SparseVector<FloatType> result;
        EIteratorConst posThis = entries.begin();
        EIteratorConst endThis = entries.end();
        EIteratorConst posOp = op.entries.begin();
        EIteratorConst endOp = op.entries.end();
        while (posThis != endThis && posOp != endOp) {
            if (posThis->index < posOp->index) {
                ++posThis;
            } else if (posThis->index > posOp->index) {
                ++posOp;
            } else {
                Entry e;
                x4Assert(posThis->index == posOp->index);
                e.index = posThis->index;
                e.value = posThis->value * posOp->value;
                result.entries.push_back(e);
                ++posOp;
                ++posThis;
            }
        }
        return result;
    }

    template <typename FloatType>
    inline void SparseVector<FloatType>::operator+=(const SparseVector<FloatType> &op) {
        EIterator posThis = entries.begin();
        EIterator endThis = entries.end();
        EIteratorConst posOp = op.entries.begin();
        EIteratorConst endOp = op.entries.end();
        while (posThis != endThis && posOp != endOp) {
            if (posThis->index < posOp->index) {
                ++posThis;
            } else if (posThis->index > posOp->index) {
                Entry e = *posOp;
                posThis = entries.insert(posThis, e);
                ++posThis;
                endThis = entries.end();
                ++posOp;
            } else {
                x4Assert(posThis->index == posOp->index);
                posThis->value += posOp->value;
                ++posOp;
                ++posThis;
            }
        }
        x4Assert(posThis == endThis || posOp == endOp);
        while (posOp != endOp) {
            entries.push_back(*posOp);
            ++posOp;
        }
    }

    template <typename FloatType>
    inline void SparseVector<FloatType>::operator-=(const SparseVector<FloatType> &op) {
        EIterator posThis = entries.begin();
        EIterator endThis = entries.end();
        EIteratorConst posOp = op.entries.begin();
        EIteratorConst endOp = op.entries.end();
        while (posThis != endThis && posOp != endOp) {
            if (posThis->index < posOp->index) {
                ++posThis;
            } else if (posThis->index > posOp->index) {
                Entry e = *posOp;
                e.value = -e.value;
                posThis = entries.insert(posThis, e);
                ++posThis;
                endThis = entries.end();
                ++posOp;
            } else {
                x4Assert(posThis->index == posOp->index);
                posThis->value -= posOp->value;
                ++posOp;
                ++posThis;
            }
        }
        x4Assert(posThis == endThis || posOp == endOp);
        while (posOp != endOp) {
            Entry e = *posOp; e.value = -e.value;
            entries.push_back(e);
            ++posOp;
        }
    }

    template <typename FloatType>
    inline void SparseVector<FloatType>::operator*=(const FloatType &s) {
        EIterator posThis = entries.begin();
        EIterator endThis = entries.end();
        while (posThis != endThis) {
            posThis->value *= s;
            ++posThis;
        }
    }

    template <typename FloatType>
    inline void SparseVector<FloatType>::operator/=(const FloatType &s) {
        EIterator posThis = entries.begin();
        EIterator endThis = entries.end();
        while (posThis != endThis) {
            posThis->value /= s;
            ++posThis;
        }
    }

    template <typename FloatType>
    inline void SparseVector<FloatType>::operator=(const SparseVector<FloatType> &op) {
        this->entries = op.entries;
    }

    template <typename FloatType>
    inline bool SparseVector<FloatType>::operator==(const SparseVector<FloatType> &op) const {
        EIteratorConst posThis = entries.begin();
        EIteratorConst endThis = entries.end();
        EIteratorConst posOp = op.entries.begin();
        EIteratorConst endOp = op.entries.end();
        while (posThis != endThis && posOp != endOp) {
            if (posThis->index < posOp->index) {
                if (posThis->value != (FloatType)0) return false;
                ++posThis;
            } else if (posThis->index > posOp->index) {
                if (posOp->value != (FloatType)0) return false;
                ++posOp;
            } else {
                if (posThis->value != posOp->value) return false;
                ++posOp;
                ++posThis;
            }
        }
        while (posThis != endThis) {
            if (posThis->value != (FloatType)0) return false;
            ++posThis;
        }
        while (posOp != endOp) {
            if (posOp->value != (FloatType)0) return false;
            ++posOp;
        }

        return true;
    }

    template <typename FloatType>
    inline bool SparseVector<FloatType>::operator!=(const SparseVector<FloatType> &op) const {
        return !(*this == op);
    }

    template <typename FloatType>
    inline bool SparseVector<FloatType>::operator==(const DynamicVector<FloatType> &op) const {
        EIteratorConst pos = entries.begin();
        EIteratorConst end = entries.end();
        const mpcard n = op.size();
        for (mpcard i=0; i<n; i++) {
            if (pos != end) {
                mpcard index = pos->index;
                x4Assert(index >= i);
                if (index > i) {
                    if (op[i] != (FloatType)0) return false;
                } else {
                    if (op[i] != pos->value) return false;
                    ++pos;
                }
            } else {
                if (op[i] != (FloatType)0) return false;
            }
        }
        return true;
    }

    template <typename FloatType>
    inline bool SparseVector<FloatType>::operator!=(const DynamicVector<FloatType> &op) const {
        return !(*this == op);
    }


    template <typename FloatType>
    void SparseVector<FloatType>::write(OutputObjectStream *s) const {
        s->write(entries.size());
        for (mpcard i = 0; i < entries.size(); ++i) {
            s->write(entries[i].index);
            s->write(entries[i].value);
        }
    }

    template <typename FloatType>
    void SparseVector<FloatType>::read(InputObjectStream *s) {
        entries.clear();
        mpcard size;
        s->read(size);
        entries.resize(size);
        for (mpcard i = 0; i < size; ++i) {
            s->read(entries[i].index);
            s->read(entries[i].value);
        }
    }

    template <typename FloatType>
    std::string SparseVector<FloatType>::toString(bool printZeros, int printZerosLength) const {
        char buffer[1000];
        std::string result;
        result += '(';
        if (printZeros && printZerosLength > 0) {
            for (int i=0; i<printZerosLength-1; i++) {
                sprintf(buffer, "%g", (*this)[i]);
                result += buffer;
                result += ',';
                result += ' ';
            }
            sprintf(buffer, "%g", (*this)[printZerosLength-1]);
            result += buffer;
        } else {
            if (entries.size() > 0) {
                int n = (int)entries.size();
                for (int i=0; i<n-1; i++) {
                    sprintf(buffer, "%d", entries[i].index);
                    result += buffer;
                    result += ':';
                    sprintf(buffer, "%g", entries[i].value);
                    result += buffer;
                    result += ',';
                    result += ' ';
                }
                sprintf(buffer, "%d", entries[n-1].index);
                result += buffer;
                result += ':';
                sprintf(buffer, "%g", entries[n-1].value);
                result += buffer;
            }
        }
        result += ')';
        return result;
    }

    template <typename FloatType>
    std::string SparseVector<FloatType>::toMatString(size_t length) const
    {
        if (length == 0) length = getLargestIndex();
        std::ostringstream ss;
        ss.precision(16);
        ss << std::scientific;
        ss << "[";
        EIteratorConst it = entries.begin();
        EIteratorConst itEnd = entries.end();
        for (size_t ii = 0; ii < length; ++ii) {
            FloatType value = 0;
            if (it != itEnd) {
                if (ii == it->index) {
                    value = it->value;
                    ++it;
                }
            }
            ss << value << ",";
        }
        ss << "];";
        return ss.str();
    }

    template <typename FloatType>
    void SparseVector<FloatType>::checkConsistency() {
        EIteratorConst posThis = entries.begin();
        EIteratorConst endThis = entries.end();
        int lastIndex = -1;
        while (posThis != endThis) {
            x4Assert(lastIndex < (int)posThis->index);
            lastIndex = posThis->index;
            ++posThis;
        }
    }




    // ---- SparseMatrix





    template<typename FloatType>
    inline SparseMatrix<FloatType>::SparseMatrix(mpcard numRows) {
        theRows.clear();
        theRows.resize(numRows);
    }

    template<typename FloatType>
    inline SparseMatrix<FloatType>::SparseMatrix(const SparseMatrix<FloatType> &init) {
        theRows = init.theRows;
    }

    template<typename FloatType>
    void SparseMatrix<FloatType>::setZero() {
        mpcard n = (mpcard)theRows.size();
        for (mpcard i=0; i<n; i++) {
            theRows[i].clear();
        }
    }

    template<typename FloatType>
    inline void SparseMatrix<FloatType>::clear() {
        setZero();
    }

    template<typename FloatType>
    inline void SparseMatrix<FloatType>::refresh() 
    {
        mpcard n = (mpcard)theRows.size();
        for (mpcard i=0; i<n; i++) 
        {
            theRows[i].refresh();
        }
    }

    template<typename FloatType>
    inline mpcard SparseMatrix<FloatType>::getNumCols() const 
    {
        mpcard largest = 0;
        mpcard n = (mpcard)theRows.size();
        for (mpcard i=0; i<n; i++) 
            largest = std::max<FloatType>(largest, theRows[i].getLargestIndex());
        return largest + 1;
    }

    template<typename FloatType>
    inline mpcard SparseMatrix<FloatType>::getRows() const {
        return (mpcard)theRows.size();
    }

    template<typename FloatType>
    inline void SparseMatrix<FloatType>::setRows(mpcard numRows) {
        theRows.clear();
        theRows.resize(numRows);
    }

    template<typename FloatType>
    inline SparseVector<FloatType>& SparseMatrix<FloatType>::operator[](const mpcard &index) {
        return theRows[index];
    }

    template<typename FloatType>
    inline const SparseVector<FloatType>& SparseMatrix<FloatType>::operator[](const mpcard &index) const {
        return theRows[index];
    }

    template<typename FloatType>
    inline void SparseMatrix<FloatType>::removeZeros() {
        const mpcard n = theRows.size();
        for (mpcard i=0; i<n; i++) {
            theRows[i].removeZeros();
        }
    }

    template<typename FloatType>
    inline mpcard SparseMatrix<FloatType>::removeNearZeros(const FloatType &entriesSmallerThanThis){
        const mpcard n = theRows.size();
        mpcard numRemoved = 0;
        for (mpcard i=0; i<n; i++) {
            numRemoved += theRows[i].removeNearZeros(entriesSmallerThanThis);
        }
        return numRemoved;
    }

    template<typename FloatType>
    inline void SparseMatrix<FloatType>::operator+=(const SparseMatrix<FloatType> &op) {
        const mpcard n = (mpcard)theRows.size();
        x4Assert(n == op.theRows.size());
        for (mpcard i=0; i<n; i++) {
            theRows[i] += op.theRows[i];
        }
    }

    template<typename FloatType>
    inline void SparseMatrix<FloatType>::operator-=(const SparseMatrix<FloatType> &op) {
        const mpcard n = (mpcard)theRows.size();
        x4Assert(n == op.theRows.size());
        for (mpcard i=0; i<n; i++) {
            theRows[i] -= op.theRows[i];
        }
    }

    template<typename FloatType>
    inline void SparseMatrix<FloatType>::operator*=(const FloatType &op) {
        const mpcard n = (mpcard)theRows.size();
        for (mpcard i=0; i<n; i++) {
            theRows[i] *= op;
        }
    }

    template<typename FloatType>
    inline void SparseMatrix<FloatType>::setToComponentProduct(const SparseMatrix<FloatType> &op) {
        const mpcard n = (mpcard)theRows.size();
        x4Assert(n == op.theRows.size());
        for (mpcard i=0; i<n; i++) {
            theRows[i] = theRows[i].componentProduct(op.theRows[i]);
        }
    }

    template<typename FloatType>
    inline void SparseMatrix<FloatType>::operator/=(const FloatType &op) {
        const mpcard n = (mpcard)theRows.size();
        for (mpcard i=0; i<n; i++) {
            theRows[i] /= op;
        }
    }

    template<typename FloatType>
    inline void SparseMatrix<FloatType>::operator=(const SparseMatrix<FloatType> &op) {
        theRows = op.theRows;
    }

    template<typename FloatType>
    inline bool SparseMatrix<FloatType>::operator==(const SparseMatrix<FloatType> &op) const {
        const mpcard n = (mpcard)theRows.size();
        x4Assert(n == op.theRows.size());
        for (mpcard i=0; i<n; i++) {
            if (theRows[i] != op.theRows[i]) return false;
        }
        return true;
    }

    template<typename FloatType>
    inline bool SparseMatrix<FloatType>::operator!=(const SparseMatrix<FloatType> &op) const {
        return !(*this == op);
    }

    template<typename FloatType>
    inline DynamicVector<FloatType> SparseMatrix<FloatType>::operator*(const DynamicVector<FloatType> &v) const {
        const mpcard n = (mpcard)theRows.size();
        DynamicVector<FloatType> result(n);
        for (mpcard i=0; i<n; i++) {
            result[i] = theRows[i]*v;
        }
        return result;
    }

    template<typename FloatType>
    inline void SparseMatrix<FloatType>::matVecMult(const DynamicVector<FloatType> &v, DynamicVector<FloatType> &result) const {
        const int n = (int)theRows.size();
        for (int i=0; i<n; i++) 
        {
            result[i] = theRows[i]*v;
        }
    }

    template<typename FloatType>
    inline DynamicVector<FloatType> SparseMatrix<FloatType>::operator*(const SparseVector<FloatType> &v) const {
        const mpcard n = (mpcard)theRows.size();
        DynamicVector<FloatType> result(n);
        for (mpcard i=0; i<n; i++) {
            result[i] = theRows[i]*v;
        }
        return result;
    }

    template<typename FloatType>
    std::string SparseMatrix<FloatType>::toString(bool printZeros, int printZerosLength) const {
        std::string result = "[ ";
        const mpcard n = (mpcard)theRows.size();
        for (mpcard i=0; i<n; i++) {
            result += theRows[i].toString(printZeros, printZerosLength);
            if (i != n-1) {
                result += ",\n  ";
            }
        }
        result += " ]";
        return result;
    }

    template<typename FloatType>
    std::string SparseMatrix<FloatType>::toMatString(size_t length) const
    {
        if (0 == length) length = theRows.size();
        std::ostringstream ss;
        ss.precision(16);
        ss << std::scientific;
        ss << "[";
        for (size_t rr = 0; rr < theRows.size(); ++rr) {
            ss << theRows[rr].toMatString(length);
        }
        ss << "]";
        return ss.str();
    }

    /// returns transpose matrix
    template<typename FloatType>
    SparseMatrix<FloatType> SparseMatrix<FloatType>::transpose() const {
        const mpcard numRows = theRows.size();
        SparseMatrix<FloatType> result(numRows);
        for( mpcard i=0;i<numRows;i++ )
        {
            const SparseVector<FloatType> &V = theRows[i];
            const mpcard numEntries = V.getNumNonZeroEntries();
            for(mpcard j=0;j<numEntries;j++)
            {
                mpcard col_index;
                FloatType f;
                V.getSparseEntry(j,col_index,f);
                if( col_index < numRows )
                    result[col_index].pushBackEntry(i,f);
                else
                    throw X4Exception("SparseMatrix<FloatType> transpose() - width larger than height, which is not supported.");
            }
        }

        return result;
    }




    template <typename FloatType1, typename FloatType2>
    inline void convertSparseVectorType(const SparseVector<FloatType1> &source, SparseVector<FloatType2> &dest) {
        dest.setNumEntries(source.getNumSparseEntries());
        const mpcard n = dest.getNumSparseEntries();
        for (mpcard i=0; i<n; i++) {
            dest.setListEntryIndex(i, source.getListEntryIndex(i));
            dest.setListEntryValue(i, (FloatType2)source.getListEntryValue(i));
        }
    }

    template <typename FloatType1, typename FloatType2>
    inline void convertSparseMatrixType(const SparseMatrix<FloatType1> &source, SparseMatrix<FloatType2> &dest) {
        dest.setRows(source.getRows());
        const mpcard n = dest.getRows();
        for (mpcard i=0; i<n; i++) {
            convertSparseVectorType(source[i], dest[i]);
        }
    }

    template <typename FloatType>
    void SparseMatrix<FloatType>::write(OutputObjectStream *s) const {
        s->write(theRows.size());
        for (mpcard i = 0; i < theRows.size(); ++i) {
            theRows[i].write(s);
        }
    }

    template <typename FloatType>
    void SparseMatrix<FloatType>::read(InputObjectStream *s) {
        theRows.clear();
        mpcard size;
        s->read(size);
        theRows.resize(size);
        for (mpcard i = 0; i < size; ++i) {
            theRows[i].read(s);
        }
    }



}

#endif