//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "CameraPathRecorder.h"
#include "ExaminerCameraController.h"
#include "FreeCameraController.h"
#include "ObjectViewsTable.h"
#include "ObjectStreams.h"
#include "MathTools.h"
#include "MathTools.inline.h"
#include "Random.h"
#include "RGBAImage.h"
#include "ColorTools.h"
#include "DebugOutput.h"
//---------------------------------------------------------------------------

using namespace X4;


MetaClass* CameraPathRecorder::CameraPathRecorderMetaClass = NULL;

const MetaClass* CameraPathRecorder::getInstanceClass() const {
   return getClass();
}

Object* CameraPathRecorder::newInstance() {
   return new CameraPathRecorder;
}

const MetaClass* CameraPathRecorder::getClass() {
   if (CameraPathRecorderMetaClass==NULL) {
      CameraPathRecorderMetaClass = new MetaClass("CameraPathRecorder", "no description.",
                                  newInstance, sizeof(CameraPathRecorder));
   }
   return CameraPathRecorderMetaClass;
}

void CameraPathRecorder::init(const MetaClass *parentClass) {
   ((MetaClass*)getClass())->init(parentClass);
   CameraPathRecorderMetaClass->copyParentProperties();
}

void CameraPathRecorder::shutdown() {
   delete CameraPathRecorderMetaClass;
}


// ----

CameraPathRecorder::CameraPathRecorder() {
   notifyMe.me = this;
   subController = NULL;
   setSubController( new ExaminerCameraController() );
   mode = rmPlaying;
   currentTime = 0.0f;
   pathTime = 50.0f;
   interpType = interpBSpline;
}

void CameraPathRecorder::ChangesMonitor::changed(Object* obj) {
   ObjectViewsTable::update(obj);
   me->cameraChange();
}

void CameraPathRecorder::setSubController(CameraController *camCtrl) {
   delete subController;
   subController = camCtrl;
   subController->bindCamera(camera);
   subController->setNotifyOnCameraChange(&notifyMe);
}

CameraController *CameraPathRecorder::getSubController() {
   return subController;
}

static Vector3f hermiteInterpolation3f(Vector3f left, Vector3f leftDiv, Vector3f right, Vector3f rightDiv, float32 t) {
   Vector3f result;
   for (card32 i=0; i<3; i++) {
      result[i] = hermiteInterpolation(left[i], leftDiv[i], right[i], rightDiv[i],t);
   }
   return result;
}

double bsplineBasisFunction(double t, mpcard i, mpcard degree)
{
	if( degree == 0)
		return t >= i && t < i+1?1.0:0;
	else
		return (t - (double)i) / (double)degree * bsplineBasisFunction(t,i,degree-1)
		 + (((double)degree + i + 1.0) - t) / (double)degree *  bsplineBasisFunction(t,i+1,degree-1);
}

void CameraPathRecorder::setCameraByTime() {
	if (camera == NULL) return;
	card32 n = (card32)cameraPositions.size();
	float32 t = fmod(currentTime/pathTime, 1.0f);
	CameraPos interp;
	if ( interpType == interpRotX ) {
		FreeCameraController *fcc = dynamic_cast<FreeCameraController*>( subController );
		if ( fcc != NULL ){	
			float32 alpha = t*(float)M_PI*2.0f;
			float32 d = norm( fcc->getLookAt() - fcc->getPosition() );
			const Vector3f& position = makeVector3f( 0, cos( alpha ), sin( alpha ) ) * d;
			const Vector3f& up = makeVector3f( 1,0,0 );
			const Vector3f& lookAt = makeVector3f( 0,0,0 );
			fcc->setUp( up );
			fcc->setPosition( position );
			fcc->setLookAt( lookAt );
			interp.pos = position;
			interp.lookAt = lookAt;
			interp.up = up;
		}
	} else if (interpType == interpRotY) {
		ExaminerCameraController *exc = dynamic_cast<ExaminerCameraController*>( subController );
		FreeCameraController *fcc = dynamic_cast<FreeCameraController*>( subController );
		if ( exc != NULL ){
			ExaminerCameraController lc = *exc;
			lc.setNotifyOnCameraChange(NULL);
			Camera lCam;
			lc.bindCamera(&lCam);
			float32 alpha = t*(float)M_PI*2.0f;
			lc.setAngle(0,alpha);
			lc.apply();
			interp.pos = lCam.getPosition();
			interp.lookAt = lCam.getLookAt();
			interp.up = lCam.getUp();
		} else if ( fcc != NULL ){
			float32 alpha = t*(float)M_PI*2.0f;
			float32 d = norm( fcc->getLookAt() - fcc->getPosition() );
			const Vector3f& position = makeVector3f( cos( alpha ), 0, sin( alpha ) ) * d;
			const Vector3f& up = makeVector3f( 0,1,0 );
			const Vector3f& lookAt = makeVector3f( 0,0,0 );
			fcc->setUp( up );
			fcc->setPosition( position );
			fcc->setLookAt( lookAt );
			interp.pos = position;
			interp.lookAt = lookAt;
			interp.up = up;
		}
   } else  if (interpType == interpRotZ ) {
	   /*FreeCameraController *fcc = dynamic_cast<FreeCameraController*>( subController );
	   if ( fcc != NULL ){	
		   float32 alpha = t*(float)M_PI*2.0f;
		   float32 d = norm( fcc->getLookAt() - fcc->getPosition() );
		   const Vector3f& position = makeVector3f( cos( alpha ), sin( alpha ), 0 ) * d;
		   const Vector3f& up = makeVector3f( 0,0,1 );
		   const Vector3f& lookAt = makeVector3f( 0,0,0 );
		   fcc->setUp( up );
		   fcc->setPosition( position );
		   fcc->setLookAt( lookAt );
		   interp.pos = position;
		   interp.lookAt = lookAt;
		   interp.up = up;
	   }*/
		
		Camera * cam = subController->getCamera();
		CameraPos camPos;
		camPos.pos = cam->getPosition();
		camPos.lookAt = cam->getLookAt();
		camPos.up = cam->getUp();
		if( n > 0 )
			camPos = cameraPositions[0];

		debugOutput << "look at: " << camPos.lookAt << "\n";

		float alpha = t*(float)M_PI*2.0f;
		float32 d = norm( camPos.pos - camPos.lookAt );
		Matrix3f frame;
		frame[0] = camPos.pos - camPos.lookAt;
		frame[0][2] = 0;
		frame[0] = normalize( frame[0] );
		frame[1] = makeVector3f(-frame[0][1],frame[0][0],0);
		frame[2] = frame[1].crossProduct(frame[0]);
		
		interp.pos = frame*makeVector3f( cos( alpha ), sin( alpha ), 0 ) * d + camPos.lookAt;
		interp.pos[2] = camPos.pos[2];
		interp.lookAt = camPos.lookAt;
		interp.up = makeVector3f(0,0,1);


   } else {
      if (n == 0) return;
      card32 l = (card32)floor(t*n);
      card32 u = ((card32)ceil(t*n)) % n;
      CameraPos posl = cameraPositions[l];
      CameraPos posu = cameraPositions[u];
      if (interpType == interpLinear || n < 2) {
		  t -= (float32)l/(float32)n;
		  t *= n;
         interp.pos = posl.pos*(1.0f-t) + posu.pos*t;
         interp.lookAt = posl.lookAt*(1.0f-t) + posu.lookAt*t;
         interp.up = posl.up*(1.0f-t) + posu.up*t;
	  } else if( interpType == interpBSpline)
	  {
			
		  mpint bspline_degree = 3;
// 		  {
// 			  std::vector<Vector2f> points;
// 			  points.push_back(makeVector2f(rnd01(),rnd01()));
// 			  points.push_back(points.back());
// 			  points.push_back(points.back());
// 			  points.push_back(makeVector2f(rnd01(),rnd01()));
// 			  points.push_back(makeVector2f(rnd01(),rnd01()));
// 			  points.push_back(makeVector2f(rnd01(),rnd01()));
// 			  points.push_back(makeVector2f(rnd01(),rnd01()));
// 			  points.push_back(points.back());
// 			  points.push_back(points.back());
// 
// 			  RGBAFloat32Image img(300,300);
// 			  img.clear(NULL_VECTOR4F);
// 			  for( mpcard i=0;i<points.size();i++ )
// 			  {
// 				  float t = (float)i / (points.size()+1);
// 				  img.setPixel(points[i][0] * img.getSizeX(),points[i][1] * img.getSizeY(),rainbowGradient(t));
// 			  }
// 
// 			  double startRange = (double)(bspline_degree) / (points.size()-1);
// 			  double stopRange = 1.0;
// 			  for( double t=startRange;t<stopRange;t += 0.005 )
// 			  {
// 				  Vector2f p = NULL_VECTOR2F;
// 				  double tmp = t * (points.size()-1);
// 				  for( mpcard i=max(tmp - bspline_degree,0);i<points.size() && i<tmp+bspline_degree;i++ )
// 				  {
// 					  p += points[i] * bsplineBasisFunction(tmp,i,bspline_degree);
// 				  }
// 				  img.setPixel(p[0] * img.getSizeX(),p[1] * img.getSizeX(),rainbowGradient(t));
// 			  }
// 			  debugOutput.displayRGBAFloatImage(&img,"muh");
// 
// 		  }


			interp.pos = NULL_VECTOR3F;
			interp.lookAt = NULL_VECTOR3F;
			interp.up = NULL_VECTOR3F;
			double tPath = (double)t * ((double)cameraPositions.size()-2.0 + bspline_degree*2);
			debugOutput << "t: " << t << " tPath: " << tPath << "\n";
			//for( mpint i=tPath - bspline_degree;(double)i<tPath+bspline_degree;i++ )
			for( mpint i=-bspline_degree;i<(mpint)cameraPositions.size()+bspline_degree+1;i++ )
			{
				CameraPos cp;
				if( i < 0 )
					cp = cameraPositions[0];
				else
					if( i >= (mpint)cameraPositions.size())
						cp = cameraPositions.back();
					else
						cp = cameraPositions[i];
				double weight = bsplineBasisFunction(tPath+bspline_degree,i+bspline_degree,bspline_degree);
				interp.pos += cp.pos * weight;
				interp.lookAt += cp.lookAt * weight;
				interp.up += cp.up * weight;
			}
			debugOutput.cr();

			

      } else { // i.e. interpType == interpHermite
         if (n == 1) return;
         int32 lDivL = l - 1; if (lDivL < 0) lDivL = l;
         int32 lDivU = u;
         int32 uDivL = l;
         int32 uDivU = u + 1; if (uDivU >= (int32)n) uDivU = n-1;

         Vector3f divPosL, divPosU, divLookAtL, divLookAtU, divUpL, divUpU;

         divPosL    = cameraPositions[lDivU].pos    - cameraPositions[lDivL].pos;
         divPosU    = cameraPositions[uDivU].pos    - cameraPositions[uDivL].pos;
         divLookAtL = cameraPositions[lDivU].lookAt - cameraPositions[lDivL].lookAt;
         divLookAtU = cameraPositions[uDivU].lookAt - cameraPositions[uDivL].lookAt;
         divUpL     = cameraPositions[lDivU].up     - cameraPositions[lDivL].up;
         divUpU     = cameraPositions[uDivU].up     - cameraPositions[uDivL].up;

         interp.pos    = hermiteInterpolation3f(posl.pos,    divPosL,    posu.pos,    divPosU,    t);
         interp.lookAt = hermiteInterpolation3f(posl.lookAt, divLookAtL, posu.lookAt, divLookAtU, t);
         interp.up     = hermiteInterpolation3f(posl.up,     divUpL,     posu.up,     divUpU,     t);
      }
   }
   camera->setup(interp.pos, interp.lookAt, normalize(interp.up));
}



void CameraPathRecorder::keyDown(GeneralKey key) {
   if (mode == rmRecording) {
      subController->keyDown(key);
   }
}

void CameraPathRecorder::keyUp(GeneralKey key) {
   if (mode == rmRecording) {
      subController->keyUp(key);
   }
}

void CameraPathRecorder::mouseDown(int32 x, int32 y, MouseButtons buttonsState, Modifiers modifiersState) {
   if (mode == rmRecording) {
      subController->mouseDown(x,y,buttonsState, modifiersState);
   }
}

void CameraPathRecorder::mouseMoved(int32 x, int32 y) {
   if (mode == rmRecording) {
      subController->mouseMoved(x,y);
   }
}

void CameraPathRecorder::mouseUp(int32 x, int32 y, MouseButtons buttonsState, Modifiers modifiersState) {
   if (mode == rmRecording) {
      subController->mouseUp(x,y, buttonsState, modifiersState);
   }
}

void CameraPathRecorder::areaResize(card32 width, card32 height) {
   subController->areaResize(width, height);
}

void CameraPathRecorder::apply() {
   if (mode == rmRecording) {
      subController->apply();
   } else {
      setCameraByTime();
   }
}

void CameraPathRecorder::setTime(float32 currentTime) {
   this->currentTime = currentTime;
   if (mode == rmPlaying) {
      setCameraByTime();
   }
}

void CameraPathRecorder::bindCamera(Camera* cam) {
   CameraController::bindCamera( cam );
   subController->bindCamera( cam );
}


float32 CameraPathRecorder::getPathTime() {
   return pathTime;
}

void CameraPathRecorder::setPathTime(float32 time) {
   pathTime = time;
}


void CameraPathRecorder::gotoKeyFrame(card32 frameNum) {
   camera->setup(cameraPositions[frameNum].pos, cameraPositions[frameNum].lookAt, cameraPositions[frameNum].up);
   cameraChange();
}

void CameraPathRecorder::addKeyFrame(card32 frameNum) 
{
   CameraPos camPos;
   camPos.pos = camera->getPosition();
   camPos.lookAt = camera->getLookAt();
   camPos.up = camera->getUp();
	if( cameraPositions.size() == 0 )
		cameraPositions.push_back( camPos );
	else
	{
		cameraPositions.insert(cameraPositions.begin() + frameNum+1, camPos);
	}
}

void CameraPathRecorder::removeKeyFrame(card32 frameNum) {
   cameraPositions.erase(cameraPositions.begin() + frameNum);
}

card32 CameraPathRecorder::getNumKeyFrames() {
   return (card32)cameraPositions.size();
}

void CameraPathRecorder::setRecordFrames(bool rec) {
   if (rec) {
      mode = rmRecording;
   } else {
      mode = rmPlaying;
   }
}

CameraPathRecorder::Interpolation CameraPathRecorder::getInterpolationType() {
   return interpType;

}

void CameraPathRecorder::setInterpolationType(Interpolation interpType) {
   this->interpType = interpType;
}

void CameraPathRecorder::write(OutputObjectStream *s) const {
   s->write<card8>(mode);
   s->write<card32>((card32)cameraPositions.size());
   for (card32 i=0; i<cameraPositions.size(); i++) {
      s->write<float32>(cameraPositions[i].pos[0]);
      s->write<float32>(cameraPositions[i].pos[1]);
      s->write<float32>(cameraPositions[i].pos[2]);

      s->write<float32>(cameraPositions[i].lookAt[0]);
      s->write<float32>(cameraPositions[i].lookAt[1]);
      s->write<float32>(cameraPositions[i].lookAt[2]);

      s->write<float32>(cameraPositions[i].up[0]);
      s->write<float32>(cameraPositions[i].up[1]);
      s->write<float32>(cameraPositions[i].up[2]);
   }

   s->writeObject(subController);
   s->write<float32>(pathTime);
   s->write<card8>(interpType);
}

void CameraPathRecorder::read(InputObjectStream *s) {
   card8 c;
   s->read<card8>(c);
   mode = (CameraPathRecorder::RecorderMode)c;
   card32 n;
   s->read<card32>(n);
   cameraPositions.resize(n);
   for (card32 i=0; i<n; i++) {
      s->read<float32>(cameraPositions[i].pos[0]);
      s->read<float32>(cameraPositions[i].pos[1]);
      s->read<float32>(cameraPositions[i].pos[2]);

      s->read<float32>(cameraPositions[i].lookAt[0]);
      s->read<float32>(cameraPositions[i].lookAt[1]);
      s->read<float32>(cameraPositions[i].lookAt[2]);

      s->read<float32>(cameraPositions[i].up[0]);
      s->read<float32>(cameraPositions[i].up[1]);
      s->read<float32>(cameraPositions[i].up[2]);
   }

   delete subController;
   s->readObject((Persistent*&)subController);
   s->read<float32>(pathTime);
   card8 i8;
   s->read<card8>(i8);
   interpType = (CameraPathRecorder::Interpolation)i8;
}


CameraPathRecorder::~CameraPathRecorder() {
   delete subController;
}

void X4::CameraPathRecorder::mouseWheelRotated( int32 rotatedDelta, Modifiers modifiersState )
{
	if (mode == rmRecording) {
		subController->mouseWheelRotated(rotatedDelta, modifiersState);
	}

}

void X4::CameraPathRecorder::setup( const Vector3f& position, const Vector3f& ref, const Vector3f& up )
{
	if (mode == rmRecording) {
		subController->setup(position,ref,up);
	}
}