//---------------------------------------------------------------------------
#ifndef CubeMap_templCodeH
#define CubeMap_templCodeH
//---------------------------------------------------------------------------
#include "x4miscFoundation.h"
//---------------------------------------------------------------------------
#include "CubeMap.h"
#include "TemplatedImage.h"
//---------------------------------------------------------------------------


namespace X4 {

	template<class PixelType>
	void CubeMap<PixelType>::setup(card32 cube_size, const PixelType &color)
	{
		for( int i=0;i<6;i++ )
		{
			sides[i]->resize(cube_size,cube_size);
			sides[i]->clear(color);
		}
	}

	template<class PixelType>
	CubeMap<PixelType>::CubeMap() {
		for (card32 i=0; i<6; i++) {
			sides[i] = new TemplatedImage<PixelType>();
		}
	}

	template<class PixelType>
	void CubeMap<PixelType>::assign(const Object* obj, X4CopyContext *context) {
		if (this == obj) return;
		const CubeMap<PixelType> *arg = dynamic_cast<const CubeMap<PixelType>*>(obj);
		if (arg == NULL) {
			throw X4Exception("CubeMap<PixelType>::assign - obj == NULL or incompatible type.");
		}
		for (card32 i=0; i<6; i++) {
			delete sides[i];
			sides[i] = (TemplatedImage<PixelType>*)(arg->sides[i]->copy(context));
		}
	}


	template<class PixelType>
	void CubeMap<PixelType>::assignSide(card32 index, const TemplatedImage<PixelType> *img) {
		x4Assert(index <=5);
		sides[index]->assign(img);
	}

	template<class PixelType>
	CubeMap<PixelType>::~CubeMap() {
		for (card32 i=0; i<6; i++) {
			delete sides[i];
		}
	}


	template<class PixelType>
	PixelType CubeMap<PixelType>::lookup(TemplatedImage<PixelType> *img, float32 x, float32 y) {
		const card32 sizeX = img->getSizeX();
		const card32 sizeY = img->getSizeY();

		x *= sizeX/2;
		y *= sizeY/2;
		x += sizeX/2;
		y += sizeY/2;

		int32 xi = (int32)x;
		int32 yi = (int32)y;

		if (xi<0) xi = 0;
		if (xi>=(int32)sizeX) xi = sizeX-1;
		if (yi<0) yi = 0;
		if (yi>=(int32)sizeY) yi = sizeY-1;

		return img->getPixel(xi, yi);
	}

	// "cross format: .. -y ..
	//                -x +z +x
	//                .. +y ..
	//                .. -z ..

	template<class PixelType>
	inline float32 CubeMap<PixelType>::scale(int32 v) {
		const float32 size = (float32)sides[0]->getSizeX();
		const float32 sh = size/2.0f;
		return (((float32)v) - sh) / sh;
	}

	template<class PixelType>
	Vector3f CubeMap<PixelType>::getPixelDirection(CubeMapSide index, int32 x, int32 y) {
		switch (index) {
	case CMS_POSITIVE_X: {
		return makeVector3f(1, scale(y), -scale(x));
						 }
	case CMS_NEGATIVE_X: {
		return makeVector3f(-1, scale(y), scale(x));
						 }
	case CMS_POSITIVE_Y: {
		return makeVector3f(scale(x), 1, -scale(y));
						 }
	case CMS_NEGATIVE_Y: {
		return makeVector3f(scale(x), -1, scale(y));
						 }
	case CMS_POSITIVE_Z: {
		return makeVector3f(scale(x), scale(y), 1);
						 }
	case CMS_NEGATIVE_Z: {
		return makeVector3f(scale(x), -scale(y), -1);
						 }
		}
		return NULL_VECTOR3F; // avoid warning
	}


	template<class PixelType>
	PixelType CubeMap<PixelType>::lookupCubeMap(const Vector3f &dir) {
		float32 maxVal = fabs(dir[0]);
		card32 maxDir = 0;

		float32 comp = fabs(dir[1]);
		if (comp > maxVal) {
			maxVal = comp;
			maxDir = 1;
		}

		comp = fabs(dir[2]);
		if (comp > maxVal) {
			maxVal = comp;
			maxDir = 2;
		}

		bool positive;
		if (dir[maxDir] >= 0) {
			positive = true;
		} else {
			positive = false;
		}

		switch (maxDir) {
	case 0: {
		if (positive) {
			return lookup(sides[CMS_POSITIVE_X], -dir[2]/maxVal,  dir[1]/maxVal);
		} else {
			return lookup(sides[CMS_NEGATIVE_X],  dir[2]/maxVal,  dir[1]/maxVal);
		}
			}
	case 1: {
		if (positive) {
			return lookup(sides[CMS_POSITIVE_Y],  dir[0]/maxVal, -dir[2]/maxVal);
		} else {
			return lookup(sides[CMS_NEGATIVE_Y],  dir[0]/maxVal,  dir[2]/maxVal);
		}
			}
	case 2: {
		if (positive) {
			return lookup(sides[CMS_POSITIVE_Z],  dir[0]/maxVal,  dir[1]/maxVal);
		} else {
			return lookup(sides[CMS_NEGATIVE_Z],  dir[0]/maxVal, -dir[1]/maxVal);
		}
			}
		}
		return lookup(sides[CMS_POSITIVE_X],0,0); // avoid warning
	}


}


#endif
 