//---------------------------------------------------------------------------
#include "StdAfx.h"

//---------------------------------------------------------------------------
#include "FileBlockChainAllocator.h"
//---------------------------------------------------------------------------
#include "X4Assert.h"
#include "Random.h"
#include "DebugOutput.h"
#include "EventCounterDisplay.h"
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------


FileBlockChainAllocator::FileBlockChainAllocator(FileBlockAllocator *baseAlloc) {
   this->baseAlloc = baseAlloc;
   x4Assert(baseAlloc->getDataSizePerBlock() - sizeof(Header) > 0);
   allocatedSize = 0;
}


max_platform_card FileBlockChainAllocator::calcNumSegments(max_platform_card size) {
   max_platform_card numSegments = size / getMaxDataPerBlock();
   if ((size % getMaxDataPerBlock()) != 0) {
      numSegments++;
   }
   return max(numSegments, (mpcard)1);
}


inline void FileBlockChainAllocator::debugDisplay() {
#ifdef _DEBUG
    EventCounterDisplay::NewValueStepOnChange e;
    e.value = (int32)allocatedSize;
    debugOutput.displayEvent(EventCounterDisplay::getClass(), "allocatedSize", &e);
#endif
}

card64 FileBlockChainAllocator::allocBlock(max_platform_card size) {
   allocatedSize += (card64)size;
   debugDisplay();
   const max_platform_card numSegments = calcNumSegments(size);
   card64 *blocks = new card64[numSegments+1];
   for (max_platform_card i=0; i<numSegments; i++) {
      blocks[i] = baseAlloc->allocBlock();
   }
   // end pointer
   blocks[numSegments] = 0;

   for (max_platform_card i=0; i<numSegments; i++) {
      Header newHeader;
      newHeader.numSegments = numSegments;
      newHeader.nextSegment = blocks[i+1];
      newHeader.dataSize = size;
      baseAlloc->writeBlock(blocks[i], newHeader, NULL, 0);
   }

   card64 result = blocks[0];

   delete[] blocks;

   return result;
}

card64 FileBlockChainAllocator::getBlockInChain(card64 startBlock, card64 index, bool updateNumSegments, card64 newNumSegments, card64 newDataSize) { // index is meant as [0..n-1]
   card64 currentBlock = startBlock;
   while (index > 0) {
      x4Assert(currentBlock != 0);
      Header currentHeader;
      baseAlloc->readBlock(currentBlock, currentHeader, NULL, 0);
      if (updateNumSegments) {
         currentHeader.numSegments = newNumSegments;
         currentHeader.dataSize = newDataSize;
         baseAlloc->writeBlock(currentBlock, currentHeader, NULL, 0);
      }
      currentBlock = currentHeader.nextSegment;
      index--;
   }
   return currentBlock;
}

void FileBlockChainAllocator::attachBlocksToChain(card64 lastOldChain, max_platform_card additionalBlocks, max_platform_card overallNumBlocks, card64 newDataSize) {
   card64 currentBlock = lastOldChain;
   while (additionalBlocks > 0) {
      card64 newBlock = baseAlloc->allocBlock();
      Header newHeader;
      newHeader.nextSegment = newBlock;
      newHeader.numSegments = overallNumBlocks;
      newHeader.dataSize = newDataSize;
      baseAlloc->writeBlock(currentBlock, newHeader, NULL, 0);
      currentBlock = newBlock;
      additionalBlocks--;
   }
   Header lastHeader;
   lastHeader.numSegments = overallNumBlocks;
   lastHeader.nextSegment = 0;
   baseAlloc->writeBlock(currentBlock, lastHeader, NULL, 0);
}

max_platform_card FileBlockChainAllocator::deleteBlocksIntern(card64 blockID) {
   max_platform_card delBlocks = 0;
   bool atEnd = false;
   card64 currentBlock = blockID;
   do {
      delBlocks++;
      Header currentHeader;
      baseAlloc->readBlock(currentBlock, currentHeader, NULL, 0);
      baseAlloc->deleteBlock(currentBlock);

      if (delBlocks == 0) {
         allocatedSize -= currentHeader.dataSize;
         debugDisplay();
      }

      if (currentHeader.nextSegment == 0) {
         atEnd = true;
      } else {
         currentBlock = currentHeader.nextSegment;
      }
   } while (!atEnd);
   return delBlocks;
}

void FileBlockChainAllocator::reallocBlock(card64 blockID, max_platform_card newSize) {
   x4Assert(newSize != 0);
   Header firstBlockHeader;
   baseAlloc->readBlock(blockID, firstBlockHeader, NULL, 0);
   max_platform_card newNumSegments = calcNumSegments(newSize);
   max_platform_card oldNumSegments = (max_platform_card)firstBlockHeader.numSegments;

   int64 difference = (int64)newSize - (int64)firstBlockHeader.dataSize;
   allocatedSize += difference;
   debugDisplay();

   if (newNumSegments == oldNumSegments) {
      if (firstBlockHeader.dataSize != newSize) {
         firstBlockHeader.dataSize = newSize;
         baseAlloc->writeBlock(blockID, firstBlockHeader, NULL, 0);
      }
      return;
   } else if (newNumSegments > oldNumSegments) {
      card64 endBlock = getBlockInChain(blockID, oldNumSegments-1, true, newNumSegments, newSize);
      attachBlocksToChain(endBlock, newNumSegments - oldNumSegments, newNumSegments, newSize);
   } else { // newNumSegments < oldNumSegments
      card64 endBlock = getBlockInChain(blockID, newNumSegments-1, true, newNumSegments, newSize);
      x4Assert(endBlock != 0);
      Header endBlockHeader;
      baseAlloc->readBlock(endBlock, endBlockHeader, NULL, 0);
      max_platform_card numDeleted = deleteBlocksIntern(endBlockHeader.nextSegment);
      endBlockHeader.nextSegment = 0;
      endBlockHeader.numSegments = newNumSegments;
      baseAlloc->writeBlock(endBlock, endBlockHeader, NULL, 0);
      x4Assert(numDeleted == oldNumSegments - newNumSegments);
   }
}

max_platform_card FileBlockChainAllocator::getMaxSize(card64 blockID) {
   Header h;
   baseAlloc->readBlock(blockID, h, NULL, 0);
   return getMaxDataPerBlock() * (max_platform_card)h.numSegments;
}


void FileBlockChainAllocator::deleteBlock(card64 blockID) {
   deleteBlocksIntern(blockID);
}

void FileBlockChainAllocator::writeBuffer(card64 blockID, const card8 *buffer, max_platform_card size) {
   max_platform_card numBlocks = calcNumSegments(size);
   max_platform_card currentBlockNr = 0;
   card64 currentBlock = blockID;
   max_platform_card currentPosition = 0;
   max_platform_card remSize = size;
   while (currentBlockNr < numBlocks) {
      Header h;
      baseAlloc->readBlock(currentBlock, h, NULL, 0);
      x4Assert(h.numSegments >= numBlocks);

      max_platform_card writeBytesNum = min(getMaxDataPerBlock(), remSize);

      baseAlloc->writeBlock(currentBlock, h, buffer + currentPosition, writeBytesNum);

      currentPosition += getMaxDataPerBlock();
      remSize -= getMaxDataPerBlock();

      currentBlockNr++;
      currentBlock = h.nextSegment;
   }
}

void FileBlockChainAllocator::readBuffer(card64 blockID, card8 *buffer, max_platform_card size) {
   max_platform_card numBlocks = calcNumSegments(size);
   max_platform_card currentBlockNr = 0;
   card64 currentBlock = blockID;
   max_platform_card currentPosition = 0;
   max_platform_card remSize = size;
   while (currentBlockNr < numBlocks) {
      Header h;
      baseAlloc->readBlock(currentBlock, h, NULL, 0);
      x4Assert(h.numSegments >= numBlocks);

      max_platform_card readBytesNum = min(getMaxDataPerBlock(), remSize);

      baseAlloc->readBlock<Header>(currentBlock, h, buffer + currentPosition, readBytesNum);

      currentPosition += getMaxDataPerBlock();
      remSize -= getMaxDataPerBlock();

      currentBlockNr++;
      currentBlock = h.nextSegment;
   }
}

FileBlockChainAllocator::~FileBlockChainAllocator() {}


bool FileBlockChainAllocator::classTest() {
   struct BlockStruct {
      card64 blockNo;
      card64 size;
   };

   static const card32 TEST_SIZE = 10000;
   static const card32 TESTS_NUMBER = 100000;
   static const card32 MIN_FILL = 5000;
   static const card32 BLOCK_SIZE = 256;

   int actions[TESTS_NUMBER];
   for (card32 i=0; i<TESTS_NUMBER; i++) {
      actions[i] = randomNumber(5);
   }
   int rndNumbers[TESTS_NUMBER];
   for (card32 i=0; i<TESTS_NUMBER; i++) {
      rndNumbers[i] = randomNumber(10000);
   }

   card32 numErrors = 0;

   BlockStruct blocks[TEST_SIZE];
   for (card32 i=0; i<TEST_SIZE; i++) {
      blocks[i].blockNo = 0;
   }
   card32 numBlocks = 0;

   FileBlockAllocator allocTestBase;
   allocTestBase.createNewFile("testFile.tmp", BLOCK_SIZE);
   FileBlockChainAllocator allocTest(&allocTestBase);

   { // a first test
      card64 t = allocTest.allocBlock(13);
      allocTest.deleteBlock(t);
   }

   for (card32 i=0; i<TESTS_NUMBER; i++) {
      card32 action = actions[i];
      if (i < MIN_FILL || i > TESTS_NUMBER - MIN_FILL) action = 0;
      switch (action) {
       case 0: { // create block
         if (numBlocks < TEST_SIZE) {
            card32 size = randomNumber(4096)+10;
            blocks[numBlocks].size = size;
            blocks[numBlocks].blockNo = allocTest.allocBlock(size);
            x4Assert(blocks[numBlocks].blockNo != 0);
            srand(numBlocks);
            card8 *buffer = new card8[size];
            for (card32 i=0; i<size; i++) {
               buffer[i] = (card8)randomNumber(256);
            }
            allocTest.writeBuffer(blocks[numBlocks].blockNo, buffer, size);
            debugOutput << "Write block No. " << numBlocks << " with file start block " << (int)blocks[numBlocks].blockNo << " with size " << size << ".";
            debugOutput.cr();
            numBlocks++;
            delete[] buffer;
         }
         break;
       }
       case 1: case 3: case 4:{ // delete / realloc block
         if (action != 4) {
          if (numBlocks > 0) {
             allocTest.deleteBlock(blocks[numBlocks-1].blockNo);
             debugOutput << "Delete block No. " << (numBlocks - 1) << " (file block " << ((int)blocks[numBlocks - 1].blockNo) << ").";
             debugOutput.cr();
             numBlocks--;
          }
         } else {
            if (numBlocks > 0) {
               card32 testBlock = rndNumbers[i] % numBlocks;
               card32 size = randomNumber(4096)+10;
               blocks[testBlock].size = size;
               allocTest.reallocBlock(blocks[testBlock].blockNo, size);
               srand(testBlock);
               card8 *buffer = new card8[size];
               for (card32 i=0; i<size; i++) {
                  buffer[i] = (card8)randomNumber(256);
               }
               allocTest.writeBuffer(blocks[testBlock].blockNo, buffer, size);
               debugOutput << "Realloc block No. " << testBlock << " with file start block " << (int)blocks[testBlock].blockNo << " with size " << size << ".";
               debugOutput.cr();
               delete[] buffer;
            }
         }
         break;
       }
       case 2: { // read and check block
         if (numBlocks > 0) {
            card32 testBlock = rndNumbers[i] % numBlocks;
            card8 *buffer = new card8[(card32)(blocks[testBlock].size)];
            allocTest.readBuffer(blocks[testBlock].blockNo, buffer, (card32)(blocks[testBlock].size));
            srand(testBlock);
            for (card32 i=0; i<blocks[testBlock].size; i++) {
               if (buffer[i] != (card8)randomNumber(256)) {
                  debugOutput << "test failed in block " << (testBlock) << ", file block " << ((int)blocks[testBlock].blockNo) << " byte number " << i;
                  debugOutput.cr();
                  numErrors++;
               }
            }
            debugOutput << "Test block No. " << (testBlock) << ", file block " << ((int)blocks[testBlock].blockNo) << ", succeded.";
            debugOutput.cr();
            delete[] buffer;
         }
         break;
       }
      }
   }

   allocTestBase.dumpDebugStatistics();

   allocTestBase.close();

   return numErrors == 0;
}
