//---------------------------------------------------------------------------
#include "StdAfx.h"

//---------------------------------------------------------------------------
#include "LargeBinaryFile.h"
//---------------------------------------------------------------------------
#include "X4Exception.h"
#include "X4Assert.h"
#include "Random.h"
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------



#ifdef PLATFORM_WINDOWS

LargeBinaryFile::LargeBinaryFile(const char *filename, bool read, bool write) {
   x4Assert(read || write);
   this->read = read;
   this->write = write;
   DWORD access = (read?GENERIC_READ:0) | (write?GENERIC_WRITE:0);
   file = CreateFileA((LPCSTR)filename, access, 0, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL,	0);
   if (file == 0) throw EIOException("LargeBinaryFile::LargeBinaryFile() - cannot open file.");
}

void LargeBinaryFile::writeBuffer(const void *buffer, const max_platform_card numBytes) {
   x4Assert(write);
   DWORD numWritten = 0;
   bool success = WriteFile(file, buffer, numBytes, &numWritten, NULL);
   if (!success || numBytes != numWritten) {
      throw EIOException("LargeBinaryFile::writeBuffer() - failed to write all data.");
   }
}

void LargeBinaryFile::readBuffer(void *buffer, const max_platform_card numBytes) {
   x4Assert(read);
   DWORD numRead = 0;
   bool success = ReadFile(file, buffer, numBytes, &numRead, NULL);
   if (!success || numBytes != numRead) {
      throw EIOException("LargeBinaryFile::readBuffer() - failed to read all data.");
   }
}


void LargeBinaryFile::seek(card64 position) {
   card32 posHigh = (card32)(position >> 32);
   long posHighCopy = posHigh; // long is 32bit integer on win32
   card32 posLow = (card32)(position & 0xFFFFFFFF);
   x4Assert((card64)posLow + ((card64)posHigh << 32) == position);

   DWORD newLow = SetFilePointer(file, posLow, &posHighCopy, FILE_BEGIN);

   if (newLow != posLow || posHighCopy != (long)posHigh || (posLow == 0xFFFFFFFF && GetLastError() != NO_ERROR)) {
      throw EIOException("LargeBinaryFile::seek() - failed to set new file pointer.");
   }
}

card64 LargeBinaryFile::getCurrentFilePointer() {
   long posHigh = 0;
   DWORD currentLow = SetFilePointer(file, 0, &posHigh, FILE_CURRENT);
   if (currentLow == 0xFFFFFFFF && GetLastError() != NO_ERROR) {
      throw EIOException("LargeBinaryFile::getCurrentFilePointer() - failed to retrieve file pointer.");
   }
   return (card64)currentLow + ((card64)posHigh << 32);
}

card64 LargeBinaryFile::getFileSize() {
   unsigned long high = 0;
   DWORD low = GetFileSize(file, &high);
   return (card64)low + ((card64)high << 32);
}

void LargeBinaryFile::setFileSize(card64 newSize) {
   seek(newSize);
   bool success = SetEndOfFile(file);
   if (!success) {
      throw EIOException("LargeBinaryFile::setFileSize() - failed to resize file.");
   }
}

void LargeBinaryFile::sync() {
   x4Assert(write);
   bool success = FlushFileBuffers(file);
   x4Assert(success);
}


LargeBinaryFile::~LargeBinaryFile() {
   bool success = CloseHandle(file);
   if (!success) {
      throw EIOException("LargeBinaryFile::~LargeBinaryFile() - cannot close file handle.");
   }
}

void LargeBinaryFile::unitTest() {
   LargeBinaryFile f("test.file");
   /*const*/ card32 BYTES = 1024*1024;
   /*const*/ card32 SIZE = BYTES / 8;
   /*const*/ card32 FILE_SIZE = 6*1024;
   card64 *buffer = new card64[SIZE];

   card64 x = 0;
   for (card32 j=0; j<FILE_SIZE; j++) {
      for (card32 i=0; i<SIZE; i++) {
         buffer[i] = x++;
      }
      f.writeBuffer(buffer, BYTES);
   }

   /*const*/ card32 NUM_TESTS = 10000;

   /*const*/ card32 TEST_BUFFER_SIZE = 128;
   card64 *testBuffer = new card64[TEST_BUFFER_SIZE];

   for (card32 j=0; j<NUM_TESTS; j++) {
      card64 pos = ((card64)randomNumber(FILE_SIZE)) * (card64)1024 * (card64)1024 + ((card64)randomNumber(SIZE)*8);
      f.seek(pos);
      f.readBuffer(testBuffer, TEST_BUFFER_SIZE*8);
      for (card32 i=0; i<TEST_BUFFER_SIZE; i++) {
         if (testBuffer[i] != pos/8 + (card64)i) {
            x4Assert(false);
         }
      }
   }
   delete[] buffer;
   delete[] testBuffer;
}

#endif


