//---------------------------------------------------------------------------
#ifndef ManagedMemoryH
#define ManagedMemoryH
//---------------------------------------------------------------------------
#include "x4miscFoundation.h"
#include "X4Exception.h"
#include "X4Assert.h"
#include "X4Types.h"
#include <list>
//---------------------------------------------------------------------------

namespace X4 {

template <class DataType> class ManagedPtr;

class X4MISCFOUNDATION_API EManagedMemory : public X4Exception {
  public:
    EManagedMemory(const std::string msg) : X4Exception(msg) {};
    virtual std::string getTypeString() {return "EManagedMemory";}
};

class X4MISCFOUNDATION_API MMPointerChangeIntf {
  public:
    virtual void handlePointersChanged() = 0;
};

class X4MISCFOUNDATION_API ManagedMemory {
  protected:
    /// Pointer to base address of managed memory
    LPVOID basePtr;
    /// System page size (as power of two, e.g. for 4096, pageSizePow is 12)
    DWORD pageSizePow;
    /// Mask for "% pageSize" (pageSize-1 Bits set to 1)
    DWORD pageModMask;
    /// Address of next free page
    DWORD nextFreePage;
    /// Total number of pages
    SIZE_T totalPages;
    /// Total size in bytes
    SIZE_T totalSize;
    /// Free pages at the moment
    SIZE_T freePages;
    /// Biggest block that can be allocated right now (without defragment)
    SIZE_T maxAllocationSize;

    typedef struct {
      LPVOID ptr;
      SIZE_T numPages;
    } DataListElem;
    std::list<DataListElem> dataPointers;

    std::list<MMPointerChangeIntf*> registeredObjects;

    void defragment();

  public:
    typedef std::list<DataListElem>::iterator DataPtr;

    /// Constructor, only to be called by the global instance; do not instantiiate yourself.
    ManagedMemory(SIZE_T size);

    /// Allocate a new data block with the size given in bytes
    DataPtr allocate(SIZE_T size);
    /// Reallocate a data block
    DataPtr reallocate(DataPtr dataPtr, SIZE_T newSize);
    /// Free a data block
    void free(DataPtr dataPtr);

    /// Returns the free memory
    SIZE_T getFreeMemory() const { return freePages << pageSizePow; }
    /// Returns the total managed memory
    SIZE_T getTotalMemory() const { return totalSize; }

    /// Register object to notify of memory reorganization. Use ManagedPtr instead, if possible.
    void registerObject(MMPointerChangeIntf* obj);
    void unRegisterObject(MMPointerChangeIntf* obj);

    #ifdef _DEBUG
    /// Check for internal errors
    void checkConsistency();
    /// Every call to defragment() increases this counter
    card64 defragCounter;
    #endif

    ~ManagedMemory();
};

extern X4MISCFOUNDATION_API ManagedMemory *managedMemory;

template <class DataType>
class X4MISCFOUNDATION_API ManagedPtr {
  protected:
    /// The actual piece of data
    ManagedMemory::DataPtr dataPtr;
    /// Is this pointer actually allocated?
    bool active;

  public:
    /// Constructor
    ManagedPtr() : active(false) {}
    /// Allocate memory for this pointer
    void allocate() { dataPtr = managedMemory->allocate(sizeof(DataType)); active = true; }
    /// Allocate the number of bytes given for the pointer. Useful for char* buffer for example.
    void allocateSize(SIZE_T size) {
      if (size > 0) { dataPtr = managedMemory->allocate(size); active = true; } }
    /// Reallocate the data for the size given. Only useful in conjunction with allocateSize()
    void reallocateSize(SIZE_T size)
      { if (size > 0) {
          if (active) dataPtr = managedMemory->reallocate(dataPtr, size);
          else allocateSize(size);
        } else if (active) free(); }
    /// Frees the memory allocated by this pointer
    void free()
      { if (active) { managedMemory->free(dataPtr); active = false; }
        else throw EManagedMemory("Memory is not allocated"); }

    /// Returns the C-pointer to the memory area. WARNING! A call to allocate can invalidate this pointer.
    DataType *getCPtr() { if (active) return (DataType*)(dataPtr->ptr); else return NULL; }

    // Dereference
    DataType& operator*()
      {
        #ifdef _DEBUG
        if (active) return *(DataType*)(dataPtr->ptr);
        else throw EManagedMemory("Pointer is NULL");
        #else
        return *(DataType*)(dataPtr->ptr);
        #endif
      };
    DataType* operator->() const
      { if (active) return (DataType*)(dataPtr->ptr); else return NULL; };

    // NULL comparison and comparison with other pointers
    bool operator==(int i) const
      { x4AssertMsg(i == NULL, "Can only be used for NULL comparison");
        return !active; }
    bool operator==(ManagedPtr<DataType>& other)
      { if (active) return other.dataPtr == this->dataPtr;
        else return !other.active; }
    bool operator!=(int i) const
      { x4AssertMsg(i == NULL, "Can only be used for NULL comparison");
        return active; }
    bool operator!=(ManagedPtr<DataType>& other)
      { if (active) return other.dataPtr != this->dataPtr;
        else return other.active; }

};

typedef ManagedPtr<char> ManagedCharPtr;

}

#endif
