//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "RGBAImage.h"
//---------------------------------------------------------------------------
#include "TemplatedImage.inline.h"
#include "ObjectStreams.h"
#include "Tools.h"
#include "X4Assert.h"
#include "ClassMethods.h"
#include "LinearAlgebra.inline.h"
#include "FileDialogs.h"
#include "InterfaceGlobals.h"
#include "DebugOutput.h"
#include "opencv/cv.h"
#include "opencv/highgui.h"
#include "GL/GL.h"
#include "GL/GLU.h"
//---------------------------------------------------------------------------
#include <QtCore/QString>
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

IMPLEMENT_X4_CLASS(RGBAImage, 0)
{
	BEGIN_CLASS_INIT(RGBAImage);
	ADD_NOARGS_METHOD(_interactiveLoadFromFile);
}

IMPLEMENT_X4_CLASS(RGBImage, 0)
{
	BEGIN_CLASS_INIT(RGBImage);
}

IMPLEMENT_X4_ABSTRACT_CLASS_DESCR(OneByteImage, 0, "Container for single byte images, e.g. depth or luminance values.")
{
	BEGIN_CLASS_INIT(OneByteImage);
}

IMPLEMENT_X4_CLASS(RGBAFloat32Image, 0)
{
	BEGIN_CLASS_INIT(RGBAFloat32Image);
}






// ----



inline bool RGBAImage::checkBorder(card32 x, card32 y, Vector4ub &avCol) {
   if (pixels[x+y*sizeX][3] != 0) return false;
   card32 avr = 0;
   card32 avg = 0;
   card32 avb = 0;
   card32 numPix = 0;
   int32 startX = x-1; if (startX<0) startX = 0;
   int32 endX = x+1; if (endX>=(int32)sizeX) endX = sizeX-1;
   int32 startY = y-1; if (startY<0) startY = 0;
   int32 endY = y+1; if (endY>=(int32)sizeY) endY = sizeY-1;
   for (int32 yp = startY; yp<=endY; yp++) {
      for (int32 xp = startX; xp<=endX; xp++) {
         card32 addr = xp+yp*sizeX;
         x4Assert(addr < sizeX*sizeY);
         if (pixels[addr][3] != 0) {
            avr += pixels[addr][0];
            avg += pixels[addr][1];
            avb += pixels[addr][2];
            numPix++;
         }
      }
   }
   if (numPix > 0) {
      avCol[0] = (card8)(avr/numPix);
      avCol[1] = (card8)(avg/numPix);
      avCol[2] = (card8)(avb/numPix);
      avCol[3] = 0;
      return true;
   } else {
      return false;
   }
}

void RGBAImage::maskBorder() {
   for (card32 y=0; y<sizeY; y++) {
      for (card32 x=0; x<sizeX; x++) {
         Vector4ub avColor;
         if (checkBorder(x,y, avColor)) {
            pixels[x+sizeX*y] = avColor;
         }
      }
   }
}


void RGBAImage::loadGLTexture() {
   glTexImage2D(GL_TEXTURE_2D, 0, 4, sizeX, sizeY, 0, GL_RGBA, GL_UNSIGNED_BYTE, pixels);
}

void RGBAImage::loadGLSubTexture(card32 x, card32 y) {
   glTexSubImage2D(GL_TEXTURE_2D, 0, x, y, sizeX, sizeY, GL_RGBA, GL_UNSIGNED_BYTE, pixels);
}

void RGBAImage::loadGLTextureMipmapped() {
   gluBuild2DMipmaps(GL_TEXTURE_2D, 4, sizeX, sizeY, GL_RGBA, GL_UNSIGNED_BYTE, pixels);
}

void RGBAImage::readPixelsGL(card32 x, card32 y) {
   glReadPixels(x, y, sizeX, sizeY, GL_RGBA, GL_UNSIGNED_BYTE, pixels);
}

void RGBAImage::drawPixelsGL(float32 x, float32 y) {
   glPushAttrib(GL_ENABLE_BIT);
   glDisable(GL_DEPTH_TEST);
   glDisable(GL_TEXTURE_2D);
   glDisable(GL_LIGHTING);
   glMatrixMode(GL_MODELVIEW);
   glPushMatrix();
   glLoadIdentity();
   glMatrixMode(GL_PROJECTION);
   glPushMatrix();
   glLoadIdentity();
   glRasterPos2f(x, y);
   glDrawPixels(sizeX, sizeY, GL_RGBA, GL_UNSIGNED_BYTE, pixels);
   glMatrixMode(GL_MODELVIEW);
   glPopMatrix();
   glMatrixMode(GL_PROJECTION);
   glPopMatrix();
   glPopAttrib();
}

void RGBAImage::chromaMask(Vector4ub cMask) {
   Vector4ub *pos = pixels;
   card32 numPixels = sizeX*sizeY;
   for (card32 i=0; i<numPixels; i++) {
      if (((*pos)[0] == cMask[0]) & ((*pos)[1] == cMask[1]) & ((*pos)[2] == cMask[2])) {
         (*pos)[3] = 0;
      } else {
         (*pos)[3] = 255;
      }
      ++pos;
   }
}

void RGBAImage::downsampleWBorder(card32 factor, RGBAImage *dest, Vector4ub maskColor) {
   card32 destWidth = sizeX/factor;
   card32 destHeight = sizeY/factor;
   x4Assert(dest->sizeX >= destWidth && dest->sizeY >= destHeight);
   for (card32 yd = 0; yd<destHeight; yd++) {
      for (card32 xd = 0; xd<destWidth; xd++) {
         card32 r = 0;
         card32 g = 0;
         card32 b = 0;
         card32 a = 0;
         card32 basex = xd*factor;
         card32 basey = yd*factor;
         card32 numPixels = 0;
         for (card32 yso = 0; yso<factor; yso++) {
            for (card32 xso = 0; xso<factor; xso++) {
               card32 addr = xso+basex + (yso+basey)*sizeX;
               x4Assert(addr < sizeX*sizeY);
               if (  (pixels[addr][0] != maskColor[0])
                   | (pixels[addr][1] != maskColor[1])
                   | (pixels[addr][2] != maskColor[2])) {
                  r += pixels[addr][0];
                  g += pixels[addr][1];
                  b += pixels[addr][2];
                  a += pixels[addr][3];
                  numPixels++;
               }
            }
         }
         if (numPixels > 0) {
            r /= numPixels;
            g /= numPixels;
            b /= numPixels;
            a /= numPixels;
         }
         card32 destAddr = xd+yd*dest->sizeX;
         x4Assert(destAddr < dest->sizeX*dest->sizeY);
         dest->pixels[destAddr] = makeVector4ub((card8)r,(card8)g,(card8)b,(card8)a);
      }
   }
}

void RGBAImage::downsampleAA(card32 factor, RGBAImage *dest) {
   card32 destWidth = sizeX/factor;
   card32 destHeight = sizeY/factor;
   x4Assert(dest->sizeX >= destWidth && dest->sizeY >= destHeight);
   for (card32 yd = 0; yd<destHeight; yd++) {
      for (card32 xd = 0; xd<destWidth; xd++) {
         card32 r = 0;
         card32 g = 0;
         card32 b = 0;
         card32 a = 0;
         card32 basex = xd*factor;
         card32 basey = yd*factor;
         card32 numPixels = 0;
         for (card32 yso = 0; yso<factor; yso++) {
            for (card32 xso = 0; xso<factor; xso++) {
               card32 addr = xso+basex + (yso+basey)*sizeX;
               x4Assert(addr < sizeX*sizeY);
               r += pixels[addr][0];
               g += pixels[addr][1];
               b += pixels[addr][2];
               a += pixels[addr][3];
               numPixels++;
            }
         }
         if (numPixels > 0) {
            r /= numPixels;
            g /= numPixels;
            b /= numPixels;
            a /= numPixels;
         }
         card32 destAddr = xd+yd*dest->sizeX;
         x4Assert(destAddr < dest->sizeX*dest->sizeY);
         dest->pixels[destAddr] = makeVector4ub((card8)r,(card8)g,(card8)b,(card8)a);
      }
   }
}

RGBImage *RGBAImage::convertToRGB() {
   RGBImage *dest = new RGBImage(sizeX, sizeY);
   Vector4ub *pos = pixels;
   Vector3ub *destPos = dest->getPixels();
   for (card32 y=0; y<sizeY; y++) {
      for (card32 x=0; x<sizeX; x++) {
         (*destPos)[0] = (*pos)[0];
         (*destPos)[1] = (*pos)[1];
         (*destPos)[2] = (*pos)[2];
         ++pos;
         ++destPos;
      }
   }
   return dest;
}

void RGBAImage::convertToBGRA() {
   Vector4ub *pos = pixels;
   for (card32 y=0; y<sizeY; y++) {
      for (card32 x=0; x<sizeX; x++) {
         Vector4ub c1 = *pos;
         Vector4ub c2;
         c2[0] = c1[2];
         c2[1] = c1[1];
         c2[2] = c1[0];
         c2[3] = c1[3];
         *pos = c2;
         ++pos;
      }
   }
}

void RGBAImage::convertToNormalMap()
{
   std::vector<Vector4f> normalMap( sizeX * sizeY );

   // create height std::map from RGB
   for( card32 i=0;i<normalMap.size();i++ )
   {
      float height = (float)pixels[i][0] +
                     (float)pixels[i][1] +
                     (float)pixels[i][2];
      height /= 3.0f * 128.0f;
      normalMap[i][3] = height;
   }

   // create normal std::map from height
   for( card32 x=0;x<sizeX;x++ )
   {
      for( card32 y=0;y<sizeY;y++ )
      {
         card32 offset = y*sizeX + x;
         card32 offset_x = (x+1==sizeX) ? y*sizeX : y*sizeX + x + 1;
         card32 offset_y = (y+1==sizeY) ? x : (y+1)*sizeX + x;

         Vector3f dx = makeVector3f(1.0f,
            normalMap[offset_x][3] - normalMap[offset][3], 0.0f );
         Vector3f dy = makeVector3f(0.0f,
            normalMap[offset_y][3] - normalMap[offset][3], 1.0f );

         Vector3f normal = normalize( dy.crossProduct( dx ) );
         normalMap[offset][0] = normal[0];
         normalMap[offset][1] = normal[1];
         normalMap[offset][2] = normal[2];
      }
   }

   // store normals & height in image
   for( card32 i=0;i<normalMap.size();i++ )
   {
      pixels[i][0] = (card8)(normalMap[i][0] * 127.0f + 128.0f);
      pixels[i][1] = (card8)(normalMap[i][1] * 127.0f + 128.0f);
      pixels[i][2] = (card8)(normalMap[i][2] * 127.0f + 128.0f);
      pixels[i][0] = (card8)(normalMap[i][3] * 255.0f);
   }

}

void RGBAImage::copyFromBMP(const QImage *img, card32 x, card32 y, card32 width, card32 height) {

   for (card32 yp=y; yp<y+height; yp++) {
      for (card32 xp=x; xp<x+width; xp++) {
			QRgb rgb = img->pixel(xp,yp);
         Vector4ub col = makeVector4ub(qBlue(rgb), qGreen(rgb), qRed(rgb), 0);
         setPixel(xp-x, yp-y, col);
      }
   }
}

void RGBAImage::loadFromFile(const char* bmpFile)
{	
	QString name(bmpFile);
	QImage img(name);
	img=img.convertToFormat(QImage::Format_ARGB32);
	if (img.isNull())
		throw std::string("cannot load image from file ") + std::string(bmpFile);

	int w = img.width();
	int h = img.height();
	
	resize(w,h);

	for( card32 y=0;y<getSizeY();y++ )
	{
		for( card32 x=0;x<getSizeX();x++ )
		{
			QRgb argb = img.pixel(x,y);
			Vector4ub col = makeVector4ub(qBlue(argb), qGreen(argb), qRed(argb), qAlpha(argb));
			setPixel(x, y, col);
		}
	}
	
}

void RGBAImage::loadFromMemory(unsigned w, unsigned h, char* data)
{	
	resize(w,h);
	
	char* srcPtr = (char*)getPixels();

	for( card32 y=0;y<h;y++ )
	{
		for( card32 x=0;x<w;x++ )
		{
			srcPtr[(y*w + x) * 4 + 0] = data[(y*w + x) * 4 + 0];
			srcPtr[(y*w + x) * 4 + 1] = data[(y*w + x) * 4 + 1];
			srcPtr[(y*w + x) * 4 + 2] = data[(y*w + x) * 4 + 2];
			srcPtr[(y*w + x) * 4 + 3] = (char)255;//data[(y*w + x) * 4 + 3];
		}
	}
}


void RGBAImage::_interactiveLoadFromFile() {
	std::string filename = FileDialogs::getOpenFileName(InterfaceGlobals::getMainWindow(), "Open Image...",
	                                               "Bitmap Files (*.bmp; *.jpg; *.png)", NULL, "", FileDialogs::IMPORT_DIR);
   if (filename != "") {
      try {
	      loadFromFile(filename.c_str());
      } catch (...) {
         error("error loading std::map.");
      }
   } 
}

RGBAImage* RGBAImage::interactiveLoadFromFile()
{
	RGBAImage* img = new RGBAImage;

	std::string filename = FileDialogs::getOpenFileName(InterfaceGlobals::getMainWindow(), "Open Image...",
		"Bitmap Files (*.bmp; *.jpg; *.png)", NULL, "", FileDialogs::IMPORT_DIR);
	if (filename != "") {
		try {
			img->loadFromFile(filename.c_str());
		} catch (...) {
			delete img;
			return NULL;
		}
	} 

	return img;
}

RGBAImage* RGBAImage::loadFromFile(const std::string& filename)
{
	RGBAImage* img = new RGBAImage;

	if (filename != "") {
		try {
			img->loadFromFile(filename.c_str());
		} catch (...) {
			delete img;
			return NULL;
		}
	} 

	return img;
}

RGBAImage* RGBAImage::flipY() 
{
   
	RGBAImage* result=new RGBAImage(sizeX,sizeY);
	
	for (card32 y=0; y<sizeY; y++) {
      for (card32 x=0; x<sizeX; x++) {
			result->setPixel(x,y,getPixel(x,sizeY-y-1));
     }
   }
   return result;

   
}


// ----


void RGBImage::loadGLTexture() {
   glTexImage2D(GL_TEXTURE_2D, 0, 3, sizeX, sizeY, 0, GL_RGB, GL_UNSIGNED_BYTE, pixels);
}

void RGBImage::loadGLSubTexture(card32 x, card32 y) {
   glTexSubImage2D(GL_TEXTURE_2D, 0, x, y, sizeX, sizeY, GL_RGB, GL_UNSIGNED_BYTE, pixels);
}

void RGBImage::loadGLTextureMipmapped() {
   gluBuild2DMipmaps(GL_TEXTURE_2D, 3, sizeX, sizeY, GL_RGB, GL_UNSIGNED_BYTE, pixels);
}

void RGBImage::convertToBGR() {
   Vector3ub *pos = pixels;
   for (card32 y=0; y<sizeY; y++) {
      for (card32 x=0; x<sizeX; x++) {
         Vector3ub c1 = *pos;
         Vector3ub c2;
         c2[0] = c1[2];
         c2[1] = c1[1];
         c2[2] = c1[0];
         *pos = c2;
         ++pos;
      }
   }
}

// ----


void OneByteImage::loadGLTexture() {
   glTexImage2D(GL_TEXTURE_2D, 0, 1, sizeX, sizeY, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, pixels);
}

void OneByteImage::loadGLTextureMipmapped() {
   gluBuild2DMipmaps(GL_TEXTURE_2D, 0, sizeX, sizeY, GL_LUMINANCE, GL_UNSIGNED_BYTE, pixels);
}

void OneByteImage::loadGLSubTexture(card32 x, card32 y) {
   glTexSubImage2D(GL_TEXTURE_2D, 0, x, y, sizeX, sizeY, GL_LUMINANCE, GL_UNSIGNED_BYTE, pixels);
}

void OneByteImage::readPixelsFromDepthGL(card32 x, card32 y) {
   glReadPixels(x, y, sizeX, sizeY, GL_DEPTH_COMPONENT, GL_UNSIGNED_BYTE, pixels);
}

void OneByteImage::drawPixelsGL(float32 x, float32 y) {
   glPushAttrib(GL_ENABLE_BIT);
   glDisable(GL_DEPTH_TEST);
   glDisable(GL_TEXTURE_2D);
   glDisable(GL_LIGHTING);
   glMatrixMode(GL_MODELVIEW);
   glPushMatrix();
   glLoadIdentity();
   glMatrixMode(GL_PROJECTION);
   glPushMatrix();
   glLoadIdentity();
   glRasterPos2f(x, y);
   glDrawPixels(sizeX, sizeY, GL_LUMINANCE, GL_UNSIGNED_BYTE, pixels);
   glMatrixMode(GL_MODELVIEW);
   glPopMatrix();
   glMatrixMode(GL_PROJECTION);
   glPopMatrix();
   glPopAttrib();
}


bool X4::RGBAFloat32Image::loadImage( const std::string &filename )
{
	IplImage* image = cvLoadImage( filename.c_str(),1 );
	if( image == NULL || image->imageData == NULL)
		return false;

	resize(image->width, image->height);
	uchar * data = (uchar*)image->imageData;
	for( mpcard x=0;x<(mpcard)image->width;x++ )
	{
		for( mpcard y=0;y<(mpcard)image->height;y++ )
		{
			uchar red = data[y*image->widthStep+x*3+2];
			uchar green = data[y*image->widthStep+x*3+1];
			uchar blue = data[y*image->widthStep+x*3+0];
			setPixel(x,y,makeVector4f(red,green,blue,0)/255.0f);
		}
	}
	return true;
}
QImage *RGBAImage::getQImage() const 
{
	QImage * ret = new QImage(sizeX,sizeY,QImage::Format_ARGB32);
	for( mpcard x=0;x<(mpcard)sizeX;x++ )
	{
		for( mpcard y=0;y<(mpcard)sizeY;y++ )
		{
			Vector4ub p = getPixel(x,y);
			ret->setPixel(x,y,qRgb(p[0],p[1],p[2]));
		}
	}
	return ret;
}
