#ifndef StatisticsH
#define StatisticsH
//---------------------------------------------------------------------------
#include "x4miscFoundation.h"
//---------------------------------------------------------------------------
#include <hash_map>
#include "Persistent.h"
#include "ObjectListProperty.h"
#include "Timer.h"
//---------------------------------------------------------------------------

namespace X4
{


class X4MISCFOUNDATION_API StatisticEntry : public Persistent
{
	X4_CLASS( StatisticEntry )
public:

	StatisticEntry(void);
	~StatisticEntry(void);

	virtual void assign(const Object* obj, X4CopyContext *context = NULL);
	virtual void write(OutputObjectStream *s) const;
	virtual void read(InputObjectStream *s);
	void setup(std::string id);
	void clearEntries();

	// simple timer (every measurement will be stored)
	void clearTimer(const std::string &timer_id );	/// deletes all existing measurements
	void startTimer(const std::string &timer_id );
	void stopTimer(const std::string &timer_id );

	// average timer
	void resetAverageTimer(const std::string &timer_id);
	void startAverageTimer(const std::string &timer_id);
	void stopAverageTimer(const std::string &timer_id);

	// counter
	void resetCounter(const std::string &counter_id);
	void incCounter(const std::string &counter_id, unsigned inc);

	std::string getName() const {return m_Name;}

	std::string report();
	std::string reportCompact();

private:

	struct TAverageTimer
	{
		card32 minEntry;
		card32 maxEntry;
		card32 total;
		card32 numMeasurements;
	};


	stdext::hash_map<std::string, std::vector<unsigned> >	m_SimpleTimer;
	stdext::hash_map<std::string, TAverageTimer >		m_AverageTimer;
	stdext::hash_map<std::string, card64 >					m_Counter;
	stdext::hash_map<std::string, unsigned>				m_TempTimer;
	std::string									m_Name;
	Timer										m_Timer;

};


/**
	gathers statistics (timings, counters, ...)

	Statistics are usually linked to a scene and stored with the scene. 
*/
class X4MISCFOUNDATION_API Statistics : public Persistent
{
	X4_CLASS( Statistics )

private:

	std::vector<StatisticEntry*>	m_Stats;
	int m_CurrentStats;
	//StatisticEntry * m_CurrentStats;

	IMPLEMENT_OBJECT_LIST_METHODS_STL(Stats,StatisticEntry);

public:
	Statistics(void);
	~Statistics(void);

    void assign(const Object* obj, X4CopyContext *context = NULL);

	// outputs summary of all statistics
	std::string outputAll();

	// outputs compact summary of current measurement set
	std::string outputCurrentCompactSummary();

	// allows to group measurements into categories
	// (id=="" refers to the global context)
	void setCurrentMeasurementID(const std::string &id);
	void clearCurrentMeasurements();

	// simple timer (every measurement will be stored)
	void clearTimer(const std::string &timer_id );	/// deletes all existing measurements
	void startTimer(const std::string &timer_id );
	void stopTimer(const std::string &timer_id );

	// average timer
	void resetAverageTimer(const std::string &timer_id);
	void startAverageTimer(const std::string &timer_id);
	void stopAverageTimer(const std::string &timer_id);

	// counter
	void resetCounter(const std::string &counter_id);
	void incCounter(const std::string &counter_id, unsigned inc=1);

	// simple messages
	void addTextLine(const std::string &text);	// adds \n automatically


};


extern X4MISCFOUNDATION_API Statistics * globalStats;
extern X4MISCFOUNDATION_API Statistics * stats;

}

#endif