//---------------------------------------------------------------------------
#ifndef TemplatedImageH
#define TemplatedImageH
//---------------------------------------------------------------------------
#include "x4miscFoundation.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
#include "LinearAlgebra.h"
#include "X4Types.h"
//---------------------------------------------------------------------------

namespace X4 {


template <class PixelType>
class TemplatedImage : public Persistent {
 protected:
   card32 sizeX;
   card32 sizeY;
   PixelType *pixels;

 public:
   /// empty image (size = 0)
   TemplatedImage();
   /// no clear
   TemplatedImage(card32 sizeX, card32 sizeY);

   TemplatedImage(const TemplatedImage &t) { *this = t; }
   TemplatedImage &operator=(const TemplatedImage &t) { assign(&t); return *this; }

   void clear(PixelType col);
   /// destructive resize
   void resize(card32 sizeX, card32 sizeY);

   virtual void assign(const Object* obj, X4CopyContext *context = NULL);

   virtual void write(OutputObjectStream *s) const;
   virtual void read(InputObjectStream *s);

   /// downsampling without anti-aliasing
   void downsample(card32 factor, TemplatedImage<PixelType> *dest);

   PixelType *getPixels() {return pixels;}
   const PixelType *getPixels() const {return pixels;}

   inline PixelType getPixel(card32 x, card32 y) const;
   inline void setPixel(card32 x, card32 y, PixelType pix);

   inline PixelType * operator[](int y) {
      x4Assert(y >= 0 && y < (int)sizeY);
      return pixels + (card32)y*sizeX; 
   }
   inline const PixelType *operator[](int y) const {
      x4Assert(y >= 0 && y < (int)sizeY);
      return pixels + (card32)y*sizeX; 
   }
	/// returns a bilinear interpolated pixel for arbitrary (x,y)
	/// note: this method is designed for float pixeltypes and can result
	///		 in a bad interpolation of 8-bit integers.
	PixelType getBilinearInterpolatedPixel( float x, float y) const;

	enum EImageFilterType
	{
		FT_Gauss = 0,
		FT_Box = 1,
		FT_Triangle = 2
	};
	void filter(float kernelSize, EImageFilterType filterType = FT_Gauss);
	void filterMask(unsigned xSize,unsigned ySize, float * mask, bool normalizeMask=true );
	void smoothGaussian(float sigma);
	void smoothGaussianHorizontal(float sigma);

	void sobelHorizontal();
	void sobelVertical();

	/// draws a rectangle into the image. sets every pixel within [x1,x2) x [y1,y2) with pix
	void drawRectangle( card32 x1,card32 y1, card32 x2, card32 y2, PixelType pix);

	/// draws a solid circle into the image.
	void drawSolidCircle( card32 x,card32 y, card32 radius, PixelType pix);

	/// draws a line 
	void drawLine( int x1, int y1, int x2, int y2, PixelType pix );

	/// bilinear resize
	void resize_bilinear( card32 width, card32 height);

   card32 getSizeX() const {return sizeX;}
   card32 getSizeY() const {return sizeY;}

   ~TemplatedImage();
};


typedef TemplatedImage<int32>	Int32Image;
typedef TemplatedImage<float32> Float32Image;
typedef TemplatedImage<Vector3f> Vector3fImage;

}

#include "TemplatedImage.inline.h"

#endif

