#include "StdAfx.h"

#include "TimerQuery.h"
#include <X4Exception.h>

#ifndef PLATFORM_WINDOWS
#error("only available for windows")
#endif

#ifndef WIN32_LEAN_AND_MEAN
	#define WIN32_LEAN_AND_MEAN
#endif
#include <windows.h>


using namespace X4;

//---------------------------------------------------------------------------
void TimerQuery::startTimerQuery() 
{
	#ifdef __MEASURE_TIME
		_prevMeasuredTime = 0.0;
		ASSERTING(_measureStatus != TimerStarted, "Timer was already started!");
		_measureStatus = TimerStarted;
		QueryPerformanceCounter(&_timerValue);
	#endif
}

//---------------------------------------------------------------------------
void TimerQuery::endTimerQuery() 
{
	#ifdef __MEASURE_TIME
		ASSERTING(_measureStatus == TimerStarted, "Timer has not been started!");
		static double invFreq = 0.0;
		LARGE_INTEGER value;
		QueryPerformanceCounter(&value);
		
		if (invFreq == 0.0) {
			LARGE_INTEGER freq;
			// this needs to be done only once, to measure the count frequency of the current system
			QueryPerformanceFrequency(&freq);
			//TODO: I hope this is thread safe (I mean it should not harm if multiple threads query the performance frequency)
			invFreq = 1.0 / freq.QuadPart;
		}

		_prevMeasuredTime = (value.QuadPart - _timerValue.QuadPart) * invFreq;
		_time.AddSample(_prevMeasuredTime);
		_measureStatus = TimerEnded;
	#endif
}


//---------------------------------------------------------------------------
// static member variables

TimerQueryManager::SharedPtr_t TimerQueryManager::_instance[TimerQueryManager::MAX_NUM_THREADS];
TimerQueryManager::ThreadID_t TimerQueryManager::_threadIDs[TimerQueryManager::MAX_NUM_THREADS];
boost::posix_time::ptime TimerQueryManager::_globalStartDateAndTime;



//---------------------------------------------------------------------------
/*static*/ 
TimerQueryManager::SharedPtr_t
TimerQueryManager::getInstance(ThreadID_t threadID)
{
	static bool initialized = false;
	static boost::shared_mutex mutex;

	if (!initialized) {
		_globalStartDateAndTime = boost::posix_time::second_clock::local_time();
		memset(TimerQueryManager::_threadIDs, 0, sizeof(ThreadID_t));
		initialized = true;
	}
	
	card32 threadIndex = 0;

#ifdef __MULTI_THREADED_TIMER

		// add 1 in order to reserve 0 for 'unset'
		threadID++;
		// perform modulo op. on the thread pointer address
		const card32 trialThreadIndex = (card32)threadID & (card32)(MAX_NUM_THREADS-1);

		if (_threadIDs[trialThreadIndex] != threadID) 
		{
			boost::unique_lock<boost::shared_mutex> lock(mutex);
			int i;
			// we have a hash collision with the new thread (or the thread is a new one)
			for (i=0; i < MAX_NUM_THREADS; i++) {
				threadIndex = (trialThreadIndex + i) & (MAX_NUM_THREADS-1); //modulo
				if (_threadIDs[threadIndex] == 0 || _threadIDs[threadIndex] == threadID)
					break;
			}
			//if i==0 then a new thread called for the first time:
			_threadIDs[threadIndex] = threadID;
			ASSERTING(i < MAX_NUM_THREADS, "Too many threads for measuring the time");
		}
		else // no collision
			threadIndex = trialThreadIndex;
#endif

	if (_instance[threadIndex].get() == NULL) {
		_instance[threadIndex].reset(new TimerQueryManager());
	}
	return _instance[threadIndex];
}

//---------------------------------------------------------------------------
TimerQueryManager::TimerQueryManager()
{
}

//---------------------------------------------------------------------------
TimerQueryManager::~TimerQueryManager()
{
	//PrintToStream<X4Console>(X4Console);

	for (int i=0; i < (int)_timerQueries.size(); i++)
		delete _timerQueries[i];
}


//---------------------------------------------------------------------------
TimerQuery * TimerQueryManager::operator[] (const char * name) 
{
	TTimerEventsMap::const_iterator mapResult = _eventName2IndexMap.find(std::string(name));
	if (mapResult != _eventName2IndexMap.end())
		return _timerQueries[mapResult->second];
	else {
		throw X4Exception("A timer event with name \"" + std::string(name) + "\" does not exist!\n");
		return NULL;
	}
}

//---------------------------------------------------------------------------
TimerQuery & TimerQueryManager::_addTimerEvent(const char * timerEventName, const char * file, int line, const bool gpu) 
{
	(void)gpu;
	TTimerEventsMap::const_iterator mapResult = _eventName2IndexMap.find(std::string(timerEventName));
	if (mapResult != _eventName2IndexMap.end()) {
		throw X4Exception("A timer event with the same name \"" + std::string(timerEventName) + "\" exists already!\n");
		return *_timerQueries[ mapResult->second ];
	}
	else {
		size_t index = _timerQueries.size();
		_timerQueries.push_back(new TimerQuery(timerEventName, file, line));
		_eventName2IndexMap[std::string(timerEventName)] = (int)index;
		return *_timerQueries.back();
	}
}

TimerQueryHandle TimerQueryManager::StartTimer(const char * timerEventName, const char * file, int line, const bool gpu)
{
	(void)gpu;
	//TODO: start and create gpu timer event and check for event type correctness
	TTimerEventsMap::const_iterator mapResult = _eventName2IndexMap.find(std::string(timerEventName));
	if (mapResult != _eventName2IndexMap.end()) {
		_timerQueries[mapResult->second]->startTimerQuery();
		return TimerQueryHandle(*_timerQueries[mapResult->second]);
	}
	else {
		//only done for the first time the timer given this name is called
		TimerQuery & timerEvent = _addTimerEvent(timerEventName, file, line, gpu);
		timerEvent.startTimerQuery();
		return TimerQueryHandle(timerEvent);
	}
}

//---------------------------------------------------------------------------
TimerQueryHandle TimerQueryManager::StartTimer(const char * timerEventName, const bool gpu) 
{
	return StartTimer(timerEventName, "unspecified", 0, gpu);
}

//---------------------------------------------------------------------------
TimerQueryHandle TimerQueryManager::StartTimer_GPU(const char * timerEventName)
{
	ASSERTING(false, "This has not yet been implemented");
	return StartTimer(timerEventName, true);
}

//---------------------------------------------------------------------------
void TimerQueryManager::EndTimer(TimerQueryHandle & timerEvent) 
{
	timerEvent._timerEvent.endTimerQuery();
}
