#ifndef _TIMER_QUERY_H_
#define _TIMER_QUERY_H_

//---------------------------------------------------------------------------
#include "x4tools.h"
//---------------------------------------------------------------------------
#include "X4Types.h"
#include "GuiAssert.h"
//---------------------------------------------------------------------------
#include <boost/thread.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>
#include <limits.h>
#include <string>
//---------------------------------------------------------------------------

#ifdef _MSC_VER
#include <float.h>  // for _isnan() on VC++
#define isnan(x) _isnan(x)  // VC++ uses _isnan() instead of isnan()
#endif
//---------------------------------------------------------------------------


namespace X4
{

// disable this flag if all timers in the application shall be ignored
#define __MEASURE_TIME
// if we have a multi-threaded application which is timed separately for each thread.
#define __MULTI_THREADED_TIMER

	//----------------------------------------------------------------------------
	/// keeps statistics (mean, variance, min, max) for one random variable of a certain type.
	template <typename T>
	class CStatsRecord {
		T _sumValues;
		T _sumValuesSqr;
		T _minValue, _maxValue;
		mpcard _numSamples;
		const static T _max_value_type;
		const static T _min_value_type;

	public:
		CStatsRecord() : _sumValues(0), _sumValuesSqr(0), 
			_minValue(_max_value_type), _maxValue(_min_value_type), _numSamples(0) 
		{}
		void Reset() 
		{ 
			_sumValues = 0.0; _sumValuesSqr = 0.0; 
			_minValue = _max_value_type; _maxValue = _min_value_type; _numSamples = 0; 
		}
		void AddSample(const T value) { 
			{ASSERTING(!isnan(value), "Sample value is not a number!");}
			if (isnan(value)) return;
			_sumValues += value; _sumValuesSqr += value * value;
			_minValue = min(value, _minValue);
			_maxValue = max(value, _maxValue);
			_numSamples++; 
		}
		double GetVariance() const { 
			if (_numSamples <= 0)
				return 0.0;
			const double denom = 1.0 / (double)_numSamples;
			const double mean = _sumValues * denom;
			return (double)(_sumValuesSqr * denom) - mean * mean;
		}
		double GetStandardDeviation() const { 
			if (_numSamples <= 0)
				return 0.0;
			const double denom = 1.0 / (double)_numSamples;
			const double mean = _sumValues * denom;
			return (double)sqrt(max((T)0.0, (_sumValuesSqr * denom) - mean*mean));
		}
		double GetMean() const {
			if (_numSamples <= 0)
				return 0.0;
			return ((double)_sumValues / (double)_numSamples);
		}
		T GetMinimum() const {
			return _minValue;
		}
		T GetMaximum() const {
			return _maxValue;
		}
		mpcard GetSampleSize() const { return _numSamples; }

		T GetSumOfValues() const { return _sumValues; }
	};
	
	//TODO: for some reason std::numeric_limits<T>::max did not work if macro max and min is defined 
	#ifdef max 
		#undef max
	#endif
	#ifdef min
		#undef min
	#endif
	//#define NOMINMAX

	template <typename T> const T CStatsRecord<T>::_max_value_type = std::numeric_limits<T>::max();
	template <typename T> const T CStatsRecord<T>::_min_value_type = std::numeric_limits<T>::min();


	//----------------------------------------------------------------------------
	/**
	* for profiling and measuring the time in specific functions.
	**/
	class X4TOOLS_API TimerQuery {
		friend class TimerQueryManager;
	protected:
		/// the name of the timer query (its identifier in the program)
		const char * _name;
		/// the source file where the timer query is located in
		const char * _file;
		/// the line in the source file where the timer query is issued
		const int _line;
		/// whether the timer has not yet started, is currently running, or stopped.
		int _measureStatus;
		/// the time in milliseconds of the last measurement.
		double _prevMeasuredTime;
		/// the statistics of that timer query.
		CStatsRecord<double> _time;
		/// the start time of the timer query.
		LARGE_INTEGER _timerValue;
	
		enum { TimerUnused = 0, TimerStarted = 1, TimerEnded = 2 };
	public:
	
		TimerQuery(const char * name, const char * file, int line) : _name(name), _file(file), _line(line), 
			_measureStatus(TimerUnused), _prevMeasuredTime(0.0), _time()
		{
			_timerValue.QuadPart = (LONGLONG)0;
		}
		bool hasStopped() const { return _measureStatus != TimerStarted; }
		double GetMean() const { return _time.GetMean(); }
		double GetStandardDeviation() const { return _time.GetStandardDeviation(); }
		/// start the timer.
		virtual void startTimerQuery();
		/// stop the timer and update statistics
		virtual void endTimerQuery();
	};
	
	//----------------------------------------------------------------------------
#if 0
	/// a timer for GPU processing
	class X4TOOLS_API CTimerEventGPU : public CTimerEvent {
		friend class CTimerEventContainer;
	protected:
		GLuint _glTimerQuery[2];
		int _curFrameTimer;

		inline void _recordGPUTime(GLuint64EXT timeElapsed) {
	#ifdef __OUTPUT_STATISTICS
			_time.AddSample(timeElapsed*(1e-9));
	#endif
		}
		inline void _switchTimerBuffer() {
			_curFrameTimer = 1 - _curFrameTimer;
		}
	public:

		CTimerEventGPU(const char * name) : TimerQuery(name), _curFrameTimer(0)
		{
			glGenQueries(2, (GLuint*)_glTimerQuery);
		}
		virtual void startTimerQuery() {
	#ifdef __MEASURE_TIME
			ASSERTING(_measureStatus != TimerStarted, "Timer was already started!");
			_measureStatus = TimerStarted;
			glBeginQuery(GL_TIME_ELAPSED_EXT, _glTimerQuery[_curFrameTimer]);
	#endif
		}
		virtual void endTimerQuery(double time = -1.0) {
	#ifdef __MEASURE_TIME
			ASSERTING(_measureStatus == TimerStarted, "timer query hasn't been started yet!");
			glEndQuery(GL_TIME_ELAPSED_EXT);
			_measureStatus = TimerEnded;
	#endif
		}
		virtual void recordPreviouslyMeasuredTime() {
	#ifdef __MEASURE_TIME
			ASSERTING(_measureStatus != TimerStarted, "Time has not been measured!");
			if (_measureStatus == TimerEnded) {
				// timer queries can contain more than 32 bits of data, so always
				// query them using the 64 bit types to avoid overflow
				GLuint64EXT timeElapsed = 0;
				GLint available = 0;
				// Polling: wait for result to become available
				while (!available) {
					GLint finished = 0;
					available = 1;
					glGetQueryObjectiv(_glTimerQuery[_curFrameTimer], GL_QUERY_RESULT_AVAILABLE, &finished);
					CHECK_GL_ERROR();
					available &= finished;
				}
				// See how much time the rendering of object i took in nanoseconds.
				glGetQueryObjectui64vEXT(_glTimerQuery[_curFrameTimer], GL_QUERY_RESULT, &timeElapsed);
				CHECK_GL_ERROR();
				// Note that care should be
				// taken to use all significant bits of the result, not just the
				// least signficant 32 bits.
				_recordGPUTime(timeElapsed);
				_switchTimerBuffer();
			}
			_measureStatus = TimerUnused; // reset for the next time
	#endif
		}

	};

#endif
	
	// forward declaration
	class TimerQueryManager;

	//----------------------------------------------------------------------------
	/// keeps a reference to the timer event and stops the timer when destructed.
	class X4TOOLS_API TimerQueryHandle 
	{
		friend class TimerQueryManager;
		TimerQuery & _timerEvent;
	public:

		TimerQueryHandle(TimerQuery & te) : _timerEvent(te)
		{}
		/// if the timer handle is destructed then stop the timer
		~TimerQueryHandle() { if (!_timerEvent.hasStopped()) _timerEvent.endTimerQuery(); }
	};

	//----------------------------------------------------------------------------
	/// dynamically stores and logs all the encountered timer queries during its lifetime.
	/// each thread has its own TimerQueryManager.
	class X4TOOLS_API TimerQueryManager 
	{
	public:
		/// keep one instance for each calling thread, up to 64 
		/// the number must be a power of 2.
		static const card32 MAX_NUM_THREADS = 128;
		typedef mpcard ThreadID_t;

	private:

		//typedef std::map<int, TimerQuery*> TTimerEventsMap;
		typedef std::map<std::string, int> TTimerEventsMap;
		TTimerEventsMap _eventName2IndexMap;
		std::vector<TimerQuery*> _timerQueries;
		//boost::posix_time::ptime _startTime;

		typedef boost::shared_ptr<TimerQueryManager> SharedPtr_t;
		static SharedPtr_t _instance[MAX_NUM_THREADS];
		static ThreadID_t _threadIDs[MAX_NUM_THREADS];
		static boost::posix_time::ptime _globalStartDateAndTime;

		/** 
		* add a new timer event (if not existing) to the std::vector and return a reference to it.
		**/
		TimerQuery & _addTimerEvent(const char * timerEventName, const char * file, int line,  const bool gpu);
		
		// private constructor
		TimerQueryManager();

	public:

		~TimerQueryManager();
		
		template<class StreamClassType>
		void PrintToStream(StreamClassType & out) const;

		TimerQuery * operator[] (const char * name);
	
		TimerQueryHandle StartTimer(const char * timerEventName, const bool gpu = false);
		TimerQueryHandle StartTimer(const char * timerEventName, const char * file, int line, const bool gpu = false);
		TimerQueryHandle StartTimer_GPU(const char * timerEventName);

		static void EndTimer(TimerQueryHandle & timerEvent);
		
		///Singleton class (exists only 1 instance for each thread).
		static SharedPtr_t getInstance(ThreadID_t threadID = 0);

		/// print the timer statistics for all timers and all recorded threads
		/// either sorted in order of thread ids or sorted in order by the timer query.
		template<class StreamClassType>
		static void PrintOut(StreamClassType & out, const bool sortByTimerQuery = true);
	};

	
	//-----------------------------------------------------------------------------
	template<int nthDecimal, typename FloatType>
	inline std::string PrintNthDecimalFloat(const FloatType fval)
	{
		stringstream out;
		mpint beforeComma = (mpint)floor(fval);
		mpint afterComma = (mpint)floor((fval - (FloatType)beforeComma) * (FloatType)pow(10.0, (FloatType)nthDecimal));
		out << (beforeComma) << "." << afterComma;
		return out.str();
	}
	
	//-----------------------------------------------------------------------------
	template<class StreamClassType, typename StatsType>
	inline void PrintTimerQueryStatistics(StreamClassType & out, const CStatsRecord<StatsType> & timerStats)
	{
#if 0
		out << "\tavg. "		<< PrintNthDecimalFloat<4>(timerStats.GetMean()*1000.0)
			<< ",\tstdDev.: "	<< PrintNthDecimalFloat<4>(timerStats.GetStandardDeviation()*1000.0) 
			<< ",\tmin: "		<< PrintNthDecimalFloat<4>(timerStats.GetMinimum()*1000.0)
			<< ",\tmax: "		<< PrintNthDecimalFloat<4>(timerStats.GetMaximum()*1000.0) 
			<< ",\t#calls: "	<< timerStats.GetSampleSize();
#else
		out << "\tavg. "		<< (timerStats.GetMean()*1000.0)
			<< ",\tstdDev. "	<< (timerStats.GetStandardDeviation()*1000.0) 
			<< ",\tmin "		<< (timerStats.GetMinimum()*1000.0)
			<< ",\tmax "		<< (timerStats.GetMaximum()*1000.0) 
			<< ",\t#calls "	<< timerStats.GetSampleSize();
#endif
	}

	//-----------------------------------------------------------------------------
	template<class StreamClassType, typename StatsType>
	StreamClassType & operator<< (StreamClassType & out, const CStatsRecord<StatsType> & timerStats)
	{
		PrintTimerQueryStatistics(out, timerStats);
		return out;
	}
	

	//-----------------------------------------------------------------------------
	template<class StreamClassType>
	void TimerQueryManager::PrintToStream(StreamClassType & out) const
	{
#ifdef __MEASURE_TIME
		if (_timerQueries.size() > 0) 
		{
			out << "-----------------------------------------------\n";
			out << "GPU + CPU timings (glTimerQuery) + CPU timings --> (sec)\n";
			for (int i = 0; i < (int)_timerQueries.size(); i++) {
				ASSERTING(_timerQueries[i]->hasStopped(), "A timer has not been stopped");
				out << "File: " << std::string(_timerQueries[i]->_file) << " -- Line " << _timerQueries[i]->_line << ", " 
					<< std::string(_timerQueries[i]->_name) << ": ";
				out << _timerQueries[i]->_time;
				out << "\n";
			}
		}
#endif
	}

	struct TimerQueryThread {
		int threadID;
		TimerQuery * timerQuery;
		TimerQueryThread() 
		{}
		TimerQueryThread(int id, TimerQuery * query) : threadID(id), timerQuery(query)
		{}
	};
	
	//-----------------------------------------------------------------------------
	template<class StreamClassType>
	void TimerQueryManager::PrintOut(StreamClassType & out, const bool sortByTimerQuery)
	{
#ifdef __MEASURE_TIME
		out << "\n\n********************************************************************\n";
		out << "**********************  TIMING STATISTICS    ***********************\n";
		out << " Start: " << to_simple_std::string(_globalStartDateAndTime) << "\n";
		out << " End:   " << to_simple_std::string(second_clock::local_time()) << "\n";
		out << " Machine:\n";
		out << "    #hardware threads: " << boost::thread::hardware_concurrency() << "\n";
		card32 threadsUsed = 0;
		for (int i=0; i < MAX_NUM_THREADS; i++)
			if (_instance[i] && _instance[i]->_timerQueries.size() > 0)
				threadsUsed++;
		out << "    #timed threads:     " << threadsUsed << "\n";
		out << "CPU timings + GPU timings (glTimerQuery) --> [seconds]\n";
		out << "********************************************************************\n";
		if (sortByTimerQuery) {
			std::vector<std::vector<TimerQueryThread> > allTimers;
			TTimerEventsMap timerNameMap;
			for (int i=0; i < MAX_NUM_THREADS; i++)
			{
				if (_instance[i]) {
					for (int j=0; j < _instance[i]->_timerQueries.size(); j++) {
						TimerQuery * curQuery = _instance[i]->_timerQueries[j];
						const std::string & timerQueryKey = curQuery->_name;
						TTimerEventsMap::const_iterator queryIter = timerNameMap.find(timerQueryKey);
						if ( queryIter != timerNameMap.end()) {
							allTimers[queryIter->second].push_back(TimerQueryThread(i, curQuery));
						}
						else {
							timerNameMap[timerQueryKey] = (int)allTimers.size();
							std::vector<TimerQueryThread> newQueryVector;
							newQueryVector.push_back(TimerQueryThread(i, curQuery));
							allTimers.push_back(newQueryVector);
						}
					}
				}
			}

			for (int i=0; i < allTimers.size(); i++) {
				TimerQueryThread & firstQuery = allTimers[i][0];
				out << "\n" << std::string(firstQuery.timerQuery->_name) 
					<< ", File: " << std::string(firstQuery.timerQuery->_file) << " -- Line " << firstQuery.timerQuery->_line << ":\n"; 
				for (int j=0; j < allTimers[i].size(); j++) {
					out << "    thread id = [" << allTimers[i][j].threadID << "]:"
						<< allTimers[i][j].timerQuery->_time
						<< "\n";
				}
			}
		} // else if sorted by thread id
		else {
			for (int i=0; i < MAX_NUM_THREADS; i++)
			{
				if (_instance[i]  && _instance[i]->_timerQueries.size() > 0) {
					out << "\nThread id = [" << i << "] ----->\n";
					_instance[i]->PrintToStream(out);
				}
			}
		}
#endif
	}

	//-----------------------------------------------------------------------------
	// macros for easy using the timer
#ifdef __MEASURE_TIME

	/// get the unsigned id of the current thread calling this function.
	inline mpcard boostGetCurrentThreadID() {
		boost::thread::id threadid = boost::this_thread::get_id();
		//TODO: this is dangerous and might not work in different boost versions
		const boost::detail::thread_data_ptr & threadDataPtr = reinterpret_cast<const boost::detail::thread_data_ptr&>(threadid);
		return (mpcard)threadDataPtr->id;
	}

	/// return an unsigned integer of a pointer address (ignore the least significant bits of the pointer)
	inline X4TOOLS_API mpcard getIndexFromPointer(intptr_t ptr) {
	#ifdef X4_64BIT
			// skip lowest 3 bits since pointers are aligned at 8 byte boundaries
			return (static_cast<mpcard>(ptr) >> 3); 
	#else // X4_32BIT
			// skip lowest 2 bits since pointers are aligned at 4 byte boundaries
			return (static_cast<mpcard>(ptr) >> 2); 
	#endif
	}
	
	#define __START_NAMED_TIMER__(identName, threadID, lineStr, file, line) \
		TimerQueryHandle __timerEvent_ ## identName = \
			TimerQueryManager::getInstance(threadID)->StartTimer(identName ## "_" ## #lineStr, file, line);

	#define __START_UNNAMED_TIMER__(identName, threadID, lineStr, file, line, function) \
		TimerQueryHandle __timerEvent_ ## identName = \
			TimerQueryManager::getInstance(threadID)->StartTimer(function ## "_" ## #lineStr, file, line);
	
	// These macros are just for forcing the preprocessor to expand the macros __LINE__
	#define __STR_EXPAND_UNNAMED_TIMER__(identName, threadID, file, line, function) \
		__START_UNNAMED_TIMER__(identName, threadID, line, file, line, function)
	#define __STR_EXPAND_NAMED_TIMER__(identName, threadID, lineStr, file, line) \
		__START_UNNAMED_TIMER__(identName, threadID, line, file, line)

	#ifdef __MULTI_THREADED_TIMER
		#define START_TIME_MEASURE(identName) \
			__STR_EXPAND_UNNAMED_TIMER__(identName, boostGetCurrentThreadID(), __FILE__, __LINE__, __FUNCTION__)
			//__STR_EXPAND_NAMED_TIMER__(identName, boostGetCurrentThreadID(), __FILE__, __LINE__)
	#else
		#define START_TIME_MEASURE(identName) \
			__STR_EXPAND_UNNAMED_TIMER__(identName, 0, __FILE__, __LINE__, __FUNCTION__)
	#endif
	
	#define END_TIME_MEASURE(identName) \
			/*TimerQueryManager::getInstance()->EndTimer(__timerEvent_ ## identName);*/\
			TimerQueryManager::EndTimer(__timerEvent_ ## identName);
#else
		#define START_TIME_MEASURE(identName)
		#define END_TIME_MEASURE(identName)
#endif

} // namespace X4


#endif 
