//---------------------------------------------------------------------------
#ifndef ConsoleOutputH
#define ConsoleOutputH
//---------------------------------------------------------------------------
#include "x4modules.h"
#include <X4Types.h>
//---------------------------------------------------------------------------
#include <string>
#include <QtCore/QMutex>
//---------------------------------------------------------------------------



namespace X4 {


/// interface for console handlers
class X4MODULES_API ConsoleHandler
{
public:
	virtual ~ConsoleHandler() {}
	/// output text
	virtual void outputText(const std::string& text) = 0;
	/// output formatted text
	virtual void outputHTML(const std::string& html) = 0;
	/// output anything else
	virtual void outputObject(const void*, size_t) = 0;
	/// flush (implement if necessary)
	virtual void flush() {}
   /// show console (if applicable)
   virtual void showConsole() {}
   /// hide console (if applicable)
   virtual void hideConsole() {}
	/// toggle visible
	virtual void toggleVisible() {}
	/// move console window
	virtual void moveConsoleWindow(int x, int y) {}
	/// resize console window
	virtual void resizeConsoleWindow(int w, int h) {}
};



/// x4 console object
class X4MODULES_API X4Console { 
 public:
	std::string unhandledMessages;
 private:
	QMutex *myMutex;
	std::vector<ConsoleHandler*> handlers;

 public:	  

	X4Console();

	inline void lock()   {myMutex->lock();}
	inline void unlock() {myMutex->unlock();}

	/// output strings
	virtual X4Console& operator<<(const std::string &msg);
	virtual void outputText(const std::string& text) {(*this) << text;}
	/// output formatted strings
	virtual void outputHTML(const std::string &text);
	/// output anything else
	virtual void outputObject(const void* data, mpcard size);

	/// output "return"
	virtual void cr();
	/// flush output
	virtual void flush();

	/// add a new handler for console output; ownership is transfered.
	virtual void addHandler(ConsoleHandler *cHandler);

	/// delete all handlers
	virtual void clearHandlers();
   
	/// show console(s) (if applicable)
	virtual void show();
	/// hide console(s) (if applicable)
	virtual void hide();
	/// toggle visible
	virtual void toggleVisible();
	/// move console window
	virtual void moveConsoleWindow(int x, int y);
	/// resize console window
	virtual void resizeConsoleWindow(int w, int h);

	// macros
	X4Console& operator<<(bool b);
	X4Console& operator<<(int i);
	X4Console& operator<<(unsigned int i);
	// this is necessary for 64-bit builds
	X4Console& operator<<(int64 i);
	X4Console& operator<<(card64 i);
	X4Console& operator<<(long double d);
	X4Console& operator<<(float f);
	X4Console& operator<<(double d);
	X4Console& operator<<(char* s);

	virtual ~X4Console();
};


/// printout warning message on debug console, pop-up window if non-visible.
void X4MODULES_API consoleError(std::string msg);

/// printout error message on debug console, pop-up window if non-visible.
void X4MODULES_API consoleWarning(std::string msg);

/// printout message on debug console, pop-up window if non-visible.
void X4MODULES_API consoleMessage(std::string msg);


/// global console object
extern X4MODULES_API X4Console x4console;


}

#endif
