#include "precompiledHeadersBasics.h"
#include "TagList.h"

#include <X4Exception.h>
#include "BasicStringUtils.h"

#include <QtCore/qobject>
#include <fstream>

using namespace X4;



int TagList::findTag(std::string lookForThis) const
{
   lookForThis = strUCase(lookForThis);
   for (size_t i=0; i<tags.size(); i++) {
      if (tags[i] == lookForThis) {
         return (int)i;
      }
   }
   return -1;
}

void TagList::checkArgument(std::string s)
{
   for (size_t i=0; i<s.size(); i++) {
      if (s[i] == 10 || s[i] == 13) {
         throw EInvalidArgument("TagList::addTag() - return characters 10/13 (dec.) are not allowed.");
      }
   }
}

void TagList::addTag(std::string tag, std::string value) 
{
   checkArgument(tag);
   checkArgument(value);
   tag = strUCase(tag);
   int pos = findTag(tag);
   if (pos == -1) {
      tags.push_back(tag);
      values.push_back( std::vector<std::string>() );
      values.back().push_back(value);
   } else {
      values[pos].push_back(value);
   }
}

void TagList::clearTag(std::string tag) 
{
   int pos = findTag(tag);
   if (pos != -1) {
      std::vector< std::string >::iterator posItTags = tags.begin();
      posItTags += pos;
      tags.erase(posItTags);
      std::vector< std::vector<std::string> >::iterator posItValues = values.begin();
      posItValues += pos;
      values.erase(posItValues);
   }   
}

void TagList::clear() 
{
   tags.clear();
   values.clear();   
}


std::string TagList::getValue(std::string tag, unsigned index) const
{
   int pos = findTag(tag);
   if (pos != -1) {
      if (index < values[pos].size()) {
         return values[pos][index];
      } else {
         return "";
      }
   } else {
      return "";
   }
}


std::string TagList::getValue(unsigned tagIndex, unsigned index) const
{
   if (tagIndex < tags.size()) {
      if (index < values[tagIndex].size()) {
         return values[tagIndex][index];
      } else {
         return "";
      }
   } else {
      return "";
   }
}



std::string TagList::toString() const
{
   std::string out;
   out += "X4_Taglist_V1\n\n\n";
   out += intToStr(tags.size()); 
   out += "\n";
   for (size_t i=0; i<tags.size(); i++) 
   {
      out += tags[i] + std::string("\n");
      out += "----------\n";
      out += intToStr(values[i].size()) + std::string("\n");
      for (size_t j=0; j<values[i].size(); j++) {
         out += values[i][j] + std::string("\n");
      }
      out += "\n\n";
   }
   return out;
}

void TagList::writeTextToFile( const std::string& filename ) const
{
	std::ofstream out(filename.c_str());
   out << "X4_Taglist_V1\n\n\n";
   out << tags.size() << "\n";
   for (size_t i=0; i<tags.size(); i++) 
   {
      out << tags[i] << "\n";
      out << "----------\n";
      out << values[i].size() << "\n";
      for (size_t j=0; j<values[i].size(); j++) {
         out << values[i][j] << "\n";
      }
      out << "\n\n";
   }
}

static std::string readline(std::ifstream &in) {
   std::string result;
   bool cont = true;
   while (!in.eof() && cont) {
      char c;
      in.get(c);
      if (c == 10 || c == 13) {
         cont = false;
         in.putback(c);
      } else {
         result += c;
      }
   }
   return result;
}

static void eatWhiteSpace(std::ifstream &in) {
   while (!in.eof()) {
      char c;
      in.get(c);
      if (!(c == ' ' || c == '\t' || c == 10 || c == 13)) {
         in.putback(c);
         break;
      }
   }
}

void TagList::readTextFromFile( const std::string& filename ) 
{
   clear();
   std::ifstream in(filename.c_str());
   std::string s;
   in >> s;
   if (s != "X4_Taglist_V1") 
   {
      throw EIOException("TagList::readTextFromFile - This is not an X4 tag list"); 
   }
   size_t numTags = 0;
   in >> numTags;
   for (size_t i=0; i<numTags; i++) 
   {
      std::string tag, temp;
      eatWhiteSpace(in);
      tag = readline(in);
      in >> temp;
      if (temp != "----------") 
      {
         throw EIOException("TagList::readTextFromFile - parse error, tag marker expected after tag name."); 
      }
      tags.push_back(tag);
      values.push_back(std::vector<std::string>());
      size_t tagSize = 0;
      in >> tagSize;
      for (size_t j=0; j<tagSize; j++) {
         eatWhiteSpace(in);
         std::string c = readline(in);
         values[i].push_back(c);
      }
   }
}

unsigned TagList::getNumTags() const 
{
   return (unsigned)tags.size();
}


std::string TagList::getTag(unsigned tagIndex) const 
{
   return tags[tagIndex];
}

unsigned TagList::getNumValues(unsigned tagIndex) const
{
   return (unsigned)values[tagIndex].size();
}


template <class T>
void forget(T) {}

void TagList::parseModuleDeclaration( const std::string& filename )
{
	forget(filename);
	//   fopen(filename.c_str(), )
	throw X4Exception("TagList::parseModuleDeclaration - not implemented.");
}

bool TagList::isTagPresent(std::string tag) const 
{
   return (findTag(tag) != -1);
}


std::vector<std::string> TagList::getAllValues(std::string tag) const 
{
   std::vector<std::string> result;
   int index = findTag(tag);
   if (index >-1) {
      size_t nv = values[index].size();
      for (size_t i=0; i<nv; i++) {
         result.push_back(values[index][i]);
      }
   }
   return result;
}
