#ifndef TagListH
#define TagListH

#include "x4modules.h"

#include <string>
#include <vector>

class QObject;



namespace X4 {


/// Maintains a list of name tags. Each tag can store a number of additional values, which are also strings. 
/// Each value can be associated with a QObject, for example to store icons etc. The list becomes the owner of these objects
/// (i.e., it will clean up in the end).
/// Tag lists can be written to disk as text files and read back. Binary values (QObjects) are currently ignored.
///
/// Tag lists are the primary tool to manage properties and dependencies in X4 modules.
class X4MODULES_API TagList {
 private:
 
   std::vector< std::string > tags;
   std::vector< std::vector<std::string> > values;
   
   int findTag(std::string lookForThis) const;
   
   void checkArgument(std::string s);
   
 public:
   /// Add a tag to the tag list. Tags are not case-sensitive (converted to upper case automatically).
   /// Tags and values cannot contain return or line feed characters (otherwise: exception)
   void addTag(std::string tag, std::string value);
   /// clear specific tag
   void clearTag(std::string tag);
   /// clear all tags
   void clear();
   /// get tag by name
   std::string getValue(std::string tag, unsigned index = 0) const;
   /// get tag by index
   std::string getValue(unsigned tagIndex, unsigned index = 0) const;
   /// get number of tags
   unsigned getNumTags() const;
   /// get number of values a tag provides
   unsigned getNumValues(unsigned tagIndex) const;
   /// get tagname itself
   std::string getTag(unsigned tagIndex) const;
   /// is a tag present?
   bool isTagPresent(std::string tag) const;
   /// get all values for a tag
   std::vector<std::string> getAllValues(std::string tag) const;
      
   
   /// convert to std::string
   std::string toString() const;

   /// writes as simple ascii format (line-wise storage of tags and values)
   void writeTextToFile( const std::string& filename ) const;
   /// reads from simple ascii format file
   void readTextFromFile( const std::string& filename );

	/// parses an X4 module declaration file by searching for tagging macros
	void parseModuleDeclaration( const std::string& filename );
};

}


#endif
