//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "DeformationTools.h"
//---------------------------------------------------------------------------
#include "VertexDescriptor.h"
#include "GenericSwapAllOutOfCorePointCloud.h"
#include "AttachedData.h"
#include "TopologyAttachment.h"
#include "InCorePCTopologyGraph.h"
#include "PointCloud.h"
#include "ARSequence.h"
#include "ARRecomputeTopology.h"
#include "TopologyRangeSearch.h"
#include "ARSettings.h"
#include "NormalUnifier.h"
#include "FifoStack.h"
#include "IndexGraph.h"
#include "PriorityQueue.h"
#include "PointSet.h"
#include "PointSet.inline.h"
#include "PointCloudTools.h"
#include <opencv/cxcore.h>
//---------------------------------------------------------------------------
#include <queue>
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------



bool X4::compareDescr( const VertexDescriptor *other, const VertexDescriptor *mustHave) 
{
   for( card32 i=0;i<mustHave->getNumAttributes();i++ )
   {
      VAD tmp = mustHave->getAttribute(i);
      if( !other->providesAttribute( tmp.name, tmp.numComponents, tmp.dataFormat ) )
         return false;
   }
   return true;
}


std::string X4::endWithSlash(const std::string &str) {
	std::string result = str;
	if (result.empty()) {
      x4Assert(false);
		return "";
	}
	if (result[result.size()-1] != '/') {
		result.append(std::string("/"));
	}
   return result;
}


std::string X4::endWithBackSlash(const std::string &str) {
	std::string result = str;
	if (result.empty()) {
      x4Assert(false);
		return "";
	}
	if (result[result.size()-1] != '\\') {
		result.append(std::string("\\"));
	}
   return result;
}


AttachedIndexedOctree *X4::fetchAndAttachOctree(PointCloud *pc, card32 createNumPtsPerLeaf) {
	UnstructuredInCorePointCloud *upc = dynamic_cast<UnstructuredInCorePointCloud*>(pc);
	if (!upc) {
	   GenericSwapAllOutOfCorePointCloud *gpc = dynamic_cast<GenericSwapAllOutOfCorePointCloud*>(pc);
	   if (!gpc || !dynamic_cast<UnstructuredInCorePointCloud*>(gpc->fetchPC())) {
	      throw ETypeMismatch("X4::fetchAndAttachOctree() works only for UnstructuredInCorePointCloud and GenericSwapAllOutOfCorePointCloud that contain an UnstructuredInCorePointCloud.");
	   } else {
	      upc = dynamic_cast<UnstructuredInCorePointCloud*>(gpc->fetchPC());
	   }
	}
   AttachedDataContainer *attachments = upc->getAttachments();
   AttachedIndexedOctree *octree = dynamic_cast<AttachedIndexedOctree*>(attachments->getData("SimpleOctree"));

   if (octree == NULL) {
		PointSet *ps = upc->getPointSet();
		octree = new AttachedIndexedOctree();
		octree->setup("SimpleOctree", AttachedData::ADF_DELETEONUPDATE,  NULL);
		octree->octreesetup(ps, createNumPtsPerLeaf);
		attachments->attachData(octree);
   }
   return octree;
}


DVectorF X4::convertPCToVector(PointCloud *pc) {
   DVectorF result;
   result.setDim((card32)pc->getNumPoints()*3);
   AAT positionAAT = pc->getDescr()->getToken("position", 3, VAD::DATA_FORMAT_FLOAT32);
   SceneObjectIterator *it = pc->createIterator(SOIT::CAP_BASIC_PC);
   BasicPointCloudIterator *bit = dynamic_cast<BasicPointCloudIterator*>(it);
   if (!bit) {
      throw X4Exception("X4::convertPCToVector() - pc does not provide a basic iterator.");
   }
   card32 i=0;
   while (!bit->atEnd()) {
      Vector3f p = bit->get3f(positionAAT);
      result[i]   = p[0];
      result[i+1] = p[1];
      result[i+2] = p[2];
      i += 3;
      bit->next();
   }
   delete it;
   return result;
}


InCorePCTopologyGraph *X4::getTopology(PointCloud *pc) {
	AttachedDataContainer *attachments = pc->getAttachments();
	TopologyAttachment *topAtt = dynamic_cast<TopologyAttachment*>(attachments->getData("TopologyAtt"));
	if (topAtt) {
		return topAtt->getTopology();
	} else {
		return NULL;
	}
}   

void X4::attachTopology(PointCloud *pc, InCorePCTopologyGraph *tg, bool visible) {
	AttachedDataContainer *attachments = pc->getAttachments();
	TopologyAttachment *topAtt = dynamic_cast<TopologyAttachment*>(attachments->getData("TopologyAtt"));
	if (topAtt == NULL) {
		topAtt = new TopologyAttachment();
	   topAtt->setup("TopologyAtt", AttachedData::ADF_PERSISTENT);
	   topAtt->setVisible(visible);
		attachments->attachData(topAtt);
	}
	topAtt->setTopology(tg);
}

void X4::assignTopology(PointCloud *source, PointCloud *target) {
	AttachedDataContainer *attachments = source->getAttachments();
	TopologyAttachment *topAtt = dynamic_cast<TopologyAttachment*>(attachments->getData("TopologyAtt"));
	if (topAtt) {
		target->getAttachments()->attachData((TopologyAttachment*)topAtt->copy());
	}  else {
		if (target->getAttachments()->getData("TopologyAtt")) {
			target->getAttachments()->deleteData("TopologyAtt");
		}
	}	
}


// put in LinearAlgebra at some time in the future... (don't want to recompile everything)
static inline Matrix4d makeMatrix4d(
   double m00, double m10, double m20, double m30,
   double m01, double m11, double m21, double m31,
   double m02, double m12, double m22, double m32,
   double m03, double m13, double m23, double m33) 
{
   Matrix4d m;
   m[0][0] = m00;  m[1][0] = m10;  m[2][0] = m20;  m[3][0] = m30;
   m[0][1] = m01;  m[1][1] = m11;  m[2][1] = m21;  m[3][1] = m31;
   m[0][2] = m02;  m[1][2] = m12;  m[2][2] = m22;  m[3][2] = m32;
   m[0][3] = m03;  m[1][3] = m13;  m[2][3] = m23;  m[3][3] = m33;
   return m;
}


X3DEFORMATION_API Matrix4d X4::svdPseudoInvert4d(Matrix4d m, const float64 zeroTolerance) {
	Matrix4d U,V;
	Vector4d evs;
   CvMat cvMatInput, cvMat_U, cvMat_W, cvMat_Vt;
   cvInitMatHeader( &cvMatInput, 4, 4, CV_64FC1, m.data());
	cvInitMatHeader( &cvMat_W, 4, 1, CV_64FC1, evs.data() );
	cvInitMatHeader( &cvMat_U, 4, 4, CV_64FC1, U.data() );
	cvInitMatHeader( &cvMat_Vt, 4, 4, CV_64FC1, V.data() );

   // Do SVD
   cvSVD(&cvMatInput, &cvMat_W, &cvMat_U, &cvMat_Vt, CV_SVD_MODIFY_A | CV_SVD_U_T );

   Matrix4d diagInv = makeMatrix4d(
      ((fabs(evs[0])>zeroTolerance)? 1.0f/evs[0] : 0), 0, 0, 0,
      0, ((fabs(evs[1])>zeroTolerance)? 1.0f/evs[1] : 0), 0, 0,
      0, 0, ((fabs(evs[2])>zeroTolerance)? 1.0f/evs[2] : 0), 0,
      0, 0, 0, ((fabs(evs[3])>zeroTolerance)? 1.0f/evs[3] : 0)
   );

	return (V.transpose())*diagInv*(U.transpose());
}

X3DEFORMATION_API Matrix3d X4::svdPseudoInvert3d(Matrix3d m, const float64 zeroTolerance) {
	Matrix3d U,V;
	Vector3d evs;
   CvMat cvMatInput, cvMat_U, cvMat_W, cvMat_Vt;
   cvInitMatHeader( &cvMatInput, 3, 3, CV_64FC1, m.data());
	cvInitMatHeader( &cvMat_W, 3, 1, CV_64FC1, evs.data() );
	cvInitMatHeader( &cvMat_U, 3, 3, CV_64FC1, U.data() );
	cvInitMatHeader( &cvMat_Vt, 3, 3, CV_64FC1, V.data() );

   // Do SVD
   cvSVD(&cvMatInput, &cvMat_W, &cvMat_U, &cvMat_Vt, CV_SVD_MODIFY_A | CV_SVD_U_T );

   Matrix3d diagInv = makeMatrix3d(
      ((fabs(evs[0])>zeroTolerance)? 1.0f/evs[0] : 0), 0, 0, 
      0, ((fabs(evs[1])>zeroTolerance)? 1.0f/evs[1] : 0), 0,
      0, 0, ((fabs(evs[2])>zeroTolerance)? 1.0f/evs[2] : 0)
   );

	return (V.transpose())*diagInv*(U.transpose());
}

X3DEFORMATION_API void X4::prunePoints(
   UnstructuredInCorePointCloud *inPC, InCorePCTopologyGraph *inTG,
   UnstructuredInCorePointCloud *&outPC, InCorePCTopologyGraph *&outTG,
   const std::vector<bool> &toKeep)
{
   PointSet *ps = inPC->getPointSet();
   x4Assert(ps->getNumEntries() == toKeep.size());
   std::vector<card32> indexRemappingOutToIn;
   std::vector<card32> indexRemappingInToOut(ps->getNumEntries(), -1);
   for (unsigned i=0; i<ps->getNumEntries(); i++) {
      if (toKeep[i]) {
         indexRemappingOutToIn.push_back(i);
         indexRemappingInToOut[i] = (card32)indexRemappingOutToIn.size()-1;
      }
   }
   
   outPC = (UnstructuredInCorePointCloud*)inPC->copy();
   PointSet *copyPS = outPC->getPointSet();
   copyPS->clearAndSetup(1, indexRemappingOutToIn.size(), ps->getDescr());
   outTG = new InCorePCTopologyGraph();
   outTG->clearAndSetup(inTG->getEdges()->getDescr());
   
   for (unsigned i=0; i<indexRemappingOutToIn.size(); i++) {
      copyPS->setVertex(i, ps->getDataPointer(indexRemappingOutToIn[i]));
      card32 currentVertex = indexRemappingOutToIn[i];
      card32 numEdges = inTG->getNumAdjacentVertices(currentVertex);
      for (card32 j=0; j<numEdges; j++) {
         card32 adjacent = inTG->getAdjacentVertex(currentVertex, j);
         if (toKeep[adjacent]) {
            outTG->addEdge(i, indexRemappingInToOut[adjacent]);
         }
      }
   }
}

X3DEFORMATION_API void X4::prunePoints(
   PointCloud *pc, InCorePCTopologyGraph *tg,
   const std::vector<bool> &toKeep)
{
   UnstructuredInCorePointCloud *upc = dynamic_cast<UnstructuredInCorePointCloud*>(pc);
   if (!upc) {
      notImplemented();
   }
   UnstructuredInCorePointCloud *newPC;
   InCorePCTopologyGraph *newTG;
   prunePoints(upc, tg, newPC, newTG, toKeep);
   pc->assign(newPC);
   attachTopology(pc, newTG);
   delete newPC;
}


void X4::prunePointsAndReplace(ARSequence *seq, const std::vector<bool> &toKeep) {
   UnstructuredInCorePointCloud *upc = dynamic_cast<UnstructuredInCorePointCloud*>(seq->getUrshape());
   if (!upc) {
      notImplemented();
   }
   InCorePCTopologyGraph *tg = getTopology(upc);
   if (!tg) {
      throw X4Exception("X4::prunePointsAndReplace() - topology not found.");
   }
   UnstructuredInCorePointCloud *outPC = NULL;
   InCorePCTopologyGraph *outTG = NULL;
   prunePoints(upc, tg, outPC, outTG, toKeep);
   upc->assign(outPC);
   attachTopology(upc, outTG);   
   delete outPC;
}


PointCloud *X4::mergePointClouds(PointCloud *pc1, PointCloud *pc2, ARSettings *settings) {
   UnstructuredInCorePointCloud *upc1 = dynamic_cast<UnstructuredInCorePointCloud*>(pc1);
   UnstructuredInCorePointCloud *upc2 = dynamic_cast<UnstructuredInCorePointCloud*>(pc2);
   if (!upc1 || !upc2) {
      notImplemented();
   }
   UnstructuredInCorePointCloud *result = (UnstructuredInCorePointCloud*)upc1->copy();
   card32 n1 = upc1->getNumPoints();
   card32 n2 = upc2->getNumPoints();
   PointSet *resultPS = result->getPointSet();
   resultPS->changeHeight(resultPS->getHeight()+n2);
   PointSet *p2PS = upc2->getPointSet();
   x4Assert(upc1->getDescr()->equal(upc2->getDescr()));
   for (card32 i=0; i<n2; i++) {
      resultPS->setVertex(i+n1, p2PS->getDataPointer(i));
   }
   ARRecomputeTopology::recomputeTopology(result, settings);
   return result;
}


void X4::removePointsAtEnd(PointCloud *pc, card32 endPtsStartIndex) {
   UnstructuredInCorePointCloud *upc = dynamic_cast<UnstructuredInCorePointCloud*>(pc);
   if (!upc) {
      notImplemented();
   }
   PointSet *ps = upc->getPointSet();
   ps->changeHeight(endPtsStartIndex);
}

struct RNAF_Entry {
	card32 i;
	float32 w;
	
	RNAF_Entry() {}
	inline RNAF_Entry(card32 ni, float32 nw) {i=ni; w=nw;}
	bool operator<(const RNAF_Entry &e) const {return w < e.w;}
};

set<card32> X4::getNeighborIndices(
	const card32 index,
	PointSet* ps,
	InCorePCTopologyGraph* topology,
	const float32 influenceRadius
	)
{
	set<card32> neighborIndices; neighborIndices.clear();

	if ( influenceRadius > 0 ){
		AAT POSITION = ps->getAAT( "position" );
		Vector3f pos = ps->get3f( index, POSITION );
		TopNeighborhoodSphereSearch topoSphereSearch;
		topoSphereSearch.setup( topology, ps, index, pos, influenceRadius, POSITION );
		// Topological neighbors in influenceRadius
		while ( !topoSphereSearch.atEnd() ) {
			neighborIndices.insert( topoSphereSearch.getCurrentIndex() );
			topoSphereSearch.next();
		}
	}
	// All direct topological neighbors
	for ( unsigned int n = 0; n < topology->getNumAdjacentVertices( index ); n++ ){
		neighborIndices.insert( topology->getAdjacentVertex( index, n ) );
	}
	return neighborIndices;
}

X3DEFORMATION_API void X4::recomputeNormalsAndFrames( PointCloud *pc, ARSettings *settings ) 
{
	recomputeFrames( pc, settings );
	recomputeUnifyNormals( pc, settings );
	//recomputeUnifyNormalsHoppe( pc, settings );
}

X3DEFORMATION_API void X4::recomputeFrames( PointCloud *pc, ARSettings *settings ) 
{
	UnstructuredInCorePointCloud *upc = dynamic_cast<UnstructuredInCorePointCloud*>(pc);
	if ( !upc ) {
		notImplemented();
	}
	PointSet *ps = upc->getPointSet();
	InCorePCTopologyGraph *topology = getTopology(pc);
	AAT POSITION = ps->getAAT( "position" );
	AAT NORMAL = ps->getAAT( "normal" );
	AAT TANGENT_U = ps->getAAT( "tangent_u" );
	AAT TANGENT_V = ps->getAAT( "tangent_v" );

	PCA3f pca;
	Matrix3f eigenVectors;
	Vector3f eigenValues, centroid;
	const float32 influenceRadius = settings->topSurfelSpacing * 2.0f;

	// Compute reference frames
	for ( card32 pIndex = 0; pIndex < ps->getNumEntries(); pIndex++ ){
		set<card32> neighborIndices = getNeighborIndices( pIndex, ps, topology, influenceRadius );
		pca.clear();
		for ( set<card32>::const_iterator it = neighborIndices.begin(); it != neighborIndices.end(); it++ ){
			card32 nIndex = *it;
			pca.addPoint( ps->get3f( nIndex, POSITION ) );
		}
		pca.analyze( eigenValues, eigenVectors, centroid );
		ps->set3f( pIndex, NORMAL, eigenVectors[2] );
		ps->set3f( pIndex, TANGENT_U, eigenVectors[0] );
		ps->set3f( pIndex, TANGENT_V, eigenVectors[1] );
	}
}

X3DEFORMATION_API void X4::recomputeUnifyNormals( PointCloud *pc, ARSettings *settings ) 
{
	UnstructuredInCorePointCloud *upc = dynamic_cast<UnstructuredInCorePointCloud*>(pc);
	if ( !upc ) {
		notImplemented();
	}
	PointSet *ps = upc->getPointSet();
	InCorePCTopologyGraph *topology = getTopology(pc);
	AAT NORMAL = ps->getAAT( "normal" );
	AAT FLAGS = ps->getAAT( "flags" );
	for ( unsigned int pIndex = 0; pIndex < ps->getNumEntries(); pIndex++ ){
		ps->set1i( pIndex, FLAGS, ps->get1i( pIndex, FLAGS ) & ( 0xFFFFFFFF ^ NORMAL_ESTIMATOR_FLAG ) );
	}
	for ( unsigned int pIndex = 0; pIndex < ps->getNumEntries(); pIndex++ ){
		if ( !( ps->get1i( pIndex, FLAGS ) & NORMAL_ESTIMATOR_FLAG ) ){
			FifoStack<card32> stack;
			stack.push( pIndex );
			ps->set1i( pIndex, FLAGS, 1 );
			while ( !stack.empty() ){
				card32 currentIndex = stack.pop();
				const Vector3f normal = ps->get3f( currentIndex, NORMAL );
				set<card32> neighborIndices = getNeighborIndices( currentIndex, ps, topology );
				for ( set<card32>::const_iterator it = neighborIndices.begin(); it != neighborIndices.end(); it++ ){
					card32 nIndex = *it;
					if ( !( ps->get1i( nIndex, FLAGS ) & NORMAL_ESTIMATOR_FLAG ) ){
						Vector3f nNormal = ps->get3f( nIndex, NORMAL );
						if ( nNormal * normal < 0 ){
							ps->set3f( nIndex, NORMAL, -nNormal );
						}
						ps->set1i( nIndex, FLAGS, ps->get1i( nIndex, FLAGS ) | NORMAL_ESTIMATOR_FLAG );
						stack.push( nIndex );
					}
				}
			}
		}
	}
}

void X4::recomputeUnifyNormalsHoppe( PointCloud *pc, ARSettings *settings )
{
	if ( pc == NULL ) return;
	UnstructuredInCorePointCloud* upc = dynamic_cast<UnstructuredInCorePointCloud*>( pc );
	if ( upc == NULL ) return;
	PointSet* ps = upc->getPointSet();
	if (ps->getNumEntries() == NULL) return;
	InCorePCTopologyGraph* topology = getTopology( pc );
	AAT POSITION = ps->getAAT( "position" );
	AAT NORMAL = ps->getAAT( "normal" );
	AAT FLAGS = ps->getAAT( "flags" );
	for ( unsigned int pIndex = 0; pIndex < ps->getNumEntries(); pIndex++ ){
		ps->set1i( pIndex, FLAGS, ps->get1i( pIndex, FLAGS ) & ( 0xFFFFFFFF ^ NORMAL_ESTIMATOR_FLAG ) );
	}
	const float32 samplingSpacing = settings->topSurfelSpacing;
	// Compute neighbor graph, assign confidence, remember best
	IGGraph* riemannian = new IGGraph();
	riemannian->createNodes( ps->getNumEntries() );
	for ( unsigned int pIndex = 0; pIndex < ps->getNumEntries(); pIndex++ ){
		if( !( ps->get1i( pIndex,FLAGS ) & NORMAL_ESTIMATOR_FLAG) ){ 
			//create stack
			FifoStack<card32> stack;
			stack.push( 0 );
			// Build riemannian graph
			int pointcounter = 0;
			int knncounter = 0;
			while( !stack.empty()){
				pointcounter++;
				const card32 currentIndex = stack.pop();
				Vector3f current = ps->get3f( currentIndex, POSITION );
				Vector3f n_i, n_j;
				ps->set1i( currentIndex, FLAGS, ps->get1i( currentIndex, FLAGS ) | NORMAL_ESTIMATOR_FLAG );
				n_i = ps->get3f( currentIndex, NORMAL );
				float costs;
				set<card32> neighborIndices = getNeighborIndices( currentIndex, ps, topology, -1 );
				for ( set<card32>::const_iterator it = neighborIndices.begin(); it != neighborIndices.end(); it++ ){
					card32 nIndex = *it;
					Vector3f neighbor = ps->get3f( nIndex, POSITION );
					n_j = ps->get3f( nIndex, NORMAL );
					costs = (1 - abs(n_i * n_j) * exp(-pow((current-neighbor).getSqrNorm(),2)/pow(samplingSpacing,2)));
					//costs = (1 - abs(n_i * n_j));
					riemannian->createEdge(currentIndex, nIndex, costs);
					if(!(ps->get1i(nIndex,FLAGS) & NORMAL_ESTIMATOR_FLAG)){ 
						stack.push(nIndex);
						ps->set1i(nIndex,FLAGS,ps->get1i(nIndex,FLAGS) | NORMAL_ESTIMATOR_FLAG);
					}
				}
			}
			// Build minimal spanning tree
			IGGraph* mst = new IGGraph();
			mst->createNodes(riemannian->getNodeCount());
			mst->setRoot(riemannian->getRoot());
			PriorityQueue<float,IGEdge*>* pq = new PriorityQueue<float,IGEdge*>();
			IGNode* root = riemannian->getNode(riemannian->getRoot());
			root->setFlag(true);
			list<IGEdge*>::iterator it = root->getEdgeIterator();
			for(int i=0;i<(int)root->getEdgeCount();i++){
				IGEdge* current = *it;
				pq->insert(current->getCosts(),current);
				it++;
			}
			IGEdge* currentEdge;
			IGNode* currentNode;
			IGNode* targetNode;
			int connectedNodes = 0;
			while(!pq->isEmpty()){
				currentEdge = pq->pop_first();
				currentNode = riemannian->getNode(currentEdge->getStartNode());
				targetNode = riemannian->getNode(currentEdge->getTargetNode());
				if(targetNode->isFlagSet()) continue;
				mst->createEdge(currentNode->getIndex(),targetNode->getIndex(),currentEdge->getCosts());
				list<IGEdge*>::iterator it = targetNode->getEdgeIterator();
				for(int i=0;i<(int)targetNode->getEdgeCount();i++){
					if(!riemannian->getNode((*it)->getTargetNode())->isFlagSet()) pq->insert((*it)->getCosts(),*it);
					it++;
				}
				targetNode->setFlag(true);
				connectedNodes++;
			}
			// Start growing with best
			root = mst->getNode(mst->getRoot());
			root->setFlag(true);
			card32 currentIndex = root->getIndex();
			Vector3f normal = ps->get3f( currentIndex, NORMAL );
			if ( normal * makeVector3f( 1,0,0 ) < 0 ){
				ps->set3f( currentIndex, NORMAL, -normal );
			}
			currentNode = root;
			list<IGEdge*>* dfs = new list<IGEdge*>();
			it = root->getEdgeIterator();
			for(int i=0;i<(int)root->getEdgeCount();i++){
				dfs->push_front(*it);
				it++;
			}
			//Traversing MST in depth first order
			int nodeCounter = 0;
			while(!dfs->empty()){
				currentEdge = dfs->front();
				dfs->pop_front();
				currentNode = mst->getNode(currentEdge->getStartNode());
				targetNode = mst->getNode(currentEdge->getTargetNode());
				if(!targetNode->isFlagSet()){
					Vector3f currentNormal, targetNormal;
					card32 currentIndex = currentNode->getIndex();
					currentNormal = ps->get3f(currentIndex,NORMAL);
					card32 targetIndex = targetNode->getIndex();
					targetNormal = ps->get3f(targetIndex,NORMAL);
					if(currentNormal * targetNormal < 0) ps->set3f(targetIndex,NORMAL,-targetNormal);
					targetNode->setFlag(true);
					nodeCounter++;
					it = targetNode->getEdgeIterator();
					for(int i=0;i<(int)targetNode->getEdgeCount();i++){
						if(!mst->getNode((*it)->getTargetNode())->isFlagSet()) dfs->push_front(*it);
						it++;
					}
				}
			}

			delete pq;
			delete mst;
			delete dfs;
		}
	}
	delete riemannian;
}

PointSet *X4::copySegment(PointSet *ps, mpcard startIndex, mpcard endIndex) 
{
   x4Assert(ps);
   x4Assert(ps->getNumEntries() > startIndex);
   x4Assert(ps->getNumEntries() > endIndex);
   PointSet *result = new PointSet();
   result->clearAndSetup(1, endIndex - startIndex+1, ps->getDescr());
   for (mpcard i=startIndex; i<=endIndex; i++) {
      result->setVertex(i-startIndex, ps->getDataPointer(i));
   }
   return result;
}
