//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "ICPConstraints.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "PointSetANNQuery.h"
#include "ProgressWindow.h"
#include "ARConstraints.h"
#include "ColorTools.h"
#include "ARSettings.h"
#include "DebugRenderer.h"
#include "Timer.h"
#include "StringHelper.h"
#include "PointSetANNQuery.h"
#include "PrecomputedDistanceField.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------


static std::string getICPModeDescr( const card32 index )
{
	if ( index == ICP_CONSTRAINT_POINT_TO_POINT ) return "Point 2 Point"; 
	else if ( index == ICP_CONSTRAINT_POINT_TO_PLANE ) return "Point 2 Plane";
	else return "Undefined"; 
}

static std::string getOutliersModeDescr( const card32 index )
{
	if ( index == ICP_OUTLIERS_NONE ) return "None"; 
	else if ( index == ICP_OUTLIERS_DISTANCE ) return "Distance";
	else if ( index == ICP_OUTLIERS_PERCENTILE ) return "Percentile";
	else return "Undefined"; 
}

IMPLEMENT_X4_CLASS( ICPConstraints , 1 ) 
{
	BEGIN_CLASS_INIT( ICPConstraints )
	/*
	ADD_CARD32_PROP_UPDATE_DESCR( constraintMode, 0, NULL, getICPModeDescr, 2, NULL )
	ADD_CARD32_PROP_UPDATE_DESCR( distanceOutlierMode, 0, NULL, getOutliersModeDescr, 4, NULL )
	ADD_BOOLEAN_PROP( cullNormalOutliers, 1 )
	ADD_FLOAT32_PROP( outliersNormals, 0 )
	ADD_FLOAT32_PROP( outliersDistanceFactor, 0 )
	ADD_FLOAT32_PROP( outliersPercentile, 0 )
	*/
}

ICPConstraints::ICPConstraints() 
{
	constraintMode = ICP_CONSTRAINT_POINT_TO_PLANE;
	distanceOutlierMode = ICP_OUTLIERS_PERCENTILE;
	cullNormalOutliers = true;
	outliersNormals = 0.7f;
	outliersDistanceFactor = 2.0f;
	outliersPercentile = 0.3f;
	constraintsWeight = 1;
	clearStatistics();
}


void ICPConstraints::assign(const Object* obj, X4CopyContext *context) 
{
	copyObjectProperties( obj, this );
}

ICPConstraints::~ICPConstraints() 
{
}

void ICPConstraints::buildICPConstraints(
   const ARSettings *settings,
   ARConstraintsInterface *receiver,
   PointCloud *referenceShape,
	PointCloud *deformedReference,
   PointCloud *matchingShape,
	const float32 time,
   Delegate callForRepaint,
   bool showCorrVis,
   const float32 skipPercentage,
   bool countCorrespondences,
	bool useDebugRenderer)
{
   clearStatistics();
   AAT numDataPtsAAT;
   if (countCorrespondences && referenceShape->providesAttribute("numDataPts")) {
      numDataPtsAAT = referenceShape->getDescr()->getToken("numDataPts", 1, VAD::DATA_FORMAT_INT32);
   } else {
      countCorrespondences = false;
   }
	if (showCorrVis) {
		corrVis.clear();
	}
	if ( settings == NULL ) {
		debugOutput << "buildICPConstraints: error\n"; debugOutput.flush();
		return;
	}

	if ( settings == NULL ) {
		debugOutput << "buildICPConstraints: error\n"; debugOutput.flush();
		return;
	}

	if ( matchingShape == NULL ){
		debugOutput << "buildICPConstraints: error\n"; debugOutput.flush();
		return;
	}

	UnstructuredInCorePointCloud* referenceUPC = dynamic_cast<UnstructuredInCorePointCloud*>( referenceShape );
	if ( referenceUPC == NULL ){
		debugOutput << "buildICPConstraints: error\n"; debugOutput.flush();
		error("ICPConstraints::buildICPConstraints() - reference surface must be unstructured in core");
		return;
	}
	UnstructuredInCorePointCloud* deformedReferenceUPC = dynamic_cast<UnstructuredInCorePointCloud*>( deformedReference );
	if ( referenceUPC == NULL ){
		debugOutput << "buildICPConstraints: error\n"; debugOutput.flush();
		error("ICPConstraints::buildICPConstraints() - deformed reference surface must be unstructured in core");
		return;
	}
	PointSet* referencePS = referenceUPC->getPointSet();
	PointSet* deformedReferencePS = deformedReferenceUPC->getPointSet();
	
	if (!referencePS->getNumEntries() == deformedReferencePS->getNumEntries()) {
		error("ICPConstraints::buildICPConstraints() - deformed reference and reference should have the same number of points");
		return;
	}
	
	AAT REFERENCE_POSITION			  = referencePS->getAAT( "position" );
	AAT DEFORMED_REFERENCE_POSITION = deformedReferencePS->getAAT( "position" );
	AAT REFERENCE_NORMAL				  = referencePS->getAAT( "normal" );
	AAT DEFORMED_REFERENCE_NORMAL	  = deformedReferencePS->getAAT( "normal" );

	//AttachedIndexedOctree* octree = PointSetKNNQuery::createOctree( deformedReferencePS, 32 );
	//PointSetKNNQuery knn( deformedReferencePS, octree, makeVector3f( 0,0,0 ), DEFORMED_REFERENCE_POSITION );
   PointSetANNQuery knn(deformedReferencePS, 1);
	card32 nIndex;
	if (deformedReferencePS->getNumEntries() <= 0) return;

	//progressWindow->pushStep( true, "ICP Constraints" );
	
	SceneObjectIterator* it = matchingShape->createIterator( SOIT::CAP_BASIC_PC );
	BasicPointCloudIterator* basicIt = dynamic_cast<BasicPointCloudIterator*>( it );
	AAT MATCH_POSITION = matchingShape->getAAT( "position" );
	AAT MATCH_ERROR_QUAD_AAT;
	bool useDataErrorQuad = false;
	if (settings->icpUseDataErrorQuads) {
	   if (matchingShape->getDescr()->providesAttribute("errorQuad", 6, VAD::DATA_FORMAT_FLOAT32)) {
	      MATCH_ERROR_QUAD_AAT = matchingShape->getAAT("errorQuad");
	      useDataErrorQuad = true;
	   }
   }

	float32 sqrPruneDistance = 0;

	// Determine outliers prune distance
	if ( distanceOutlierMode == ICP_OUTLIERS_PERCENTILE ){
		basicIt->reset();
		std::vector<float32> sqrDistances; sqrDistances.resize( matchingShape->getNumPoints() );
		card32 index = 0;
		while( !basicIt->atEnd() ){
			Vector3f matchP = basicIt->get3f( MATCH_POSITION );
			nIndex = knn.getNearestPointIndex(matchP);
			const Vector3f refPos = deformedReferencePS->get3f( nIndex, DEFORMED_REFERENCE_POSITION );
			const float32 sqrD = ( matchP - refPos ) * ( matchP - refPos );
			sqrDistances[index] = sqrD;
			index++;
			basicIt->next();
		}
		sort(sqrDistances.begin(), sqrDistances.end());
		int32 distIndex = (1.0f-settings->icpOutlierPercentile)*(matchingShape->getNumPoints()-1);
		if (distIndex >= (int32)matchingShape->getNumPoints()) {
   		distIndex = matchingShape->getNumPoints()-1;
		}
		if (distIndex < 0) {
		   distIndex = 0;
		}
		sqrPruneDistance = sqrDistances[distIndex];
	}

	basicIt->reset();
	const PointSet *dataPS = basicIt->lockBlockRead();
   card32 currentPointIndex = 0;
	card32 numPrunedPoints = 0;
	srand(42);
	while( !basicIt->atEnd() ) {
	   if (skipPercentage != 0.0f) {
	      if (rnd01()<skipPercentage) {
   	      currentPointIndex++;
	         basicIt->next();
	         continue;
	      }
	   }
		const Vector3f& matchP = basicIt->get3f( MATCH_POSITION );
      nIndex = knn.getNearestPointIndex(matchP);
		
		const Vector3f refPos             = referencePS->get3f( nIndex, REFERENCE_POSITION );
		const Vector3f defRefPos          = deformedReferencePS->get3f( nIndex, DEFORMED_REFERENCE_POSITION );
		const Vector3f refNormal          = referencePS->get3f( nIndex, REFERENCE_NORMAL );
		const Vector3f defRefNormal       = normalize(deformedReferencePS->get3f( nIndex, DEFORMED_REFERENCE_NORMAL ));

		bool acceptMatchPoint = true;
		Vector3f q = matchP - defRefPos;
		const float32 sqrNormQ = q * q;

      /// in the near field, culling is always disabled
      bool inNearField = sqrNormQ < sqr( settings->topSurfelSpacing * outliersDistanceFactor );

		// Check for outliers
		if ( cullNormalOutliers && !inNearField ){
			// Normal-based outlier pruning
			const float32 normQ = sqrt( sqrNormQ ); 
			if ( normQ > 1e-6 ){
				q /= normQ;
				const float32 e = fabs( q * defRefNormal );
				if ( e < outliersNormals ){
					acceptMatchPoint = false;
				}
			}
		}
		if ( distanceOutlierMode == ICP_OUTLIERS_DISTANCE ){
			// Distance-based outlier pruning
			if ( !inNearField ){
				acceptMatchPoint = false;
			}
		} else if ( distanceOutlierMode == ICP_OUTLIERS_PERCENTILE ){
			// Percentile-based outlier pruning
			if ( sqrNormQ > sqrPruneDistance && !inNearField ){
				acceptMatchPoint = false;
			}
		}

		if ( !acceptMatchPoint ){
			numPrunedPoints++; 
		} else {

			PositionARConstraint posConstraint;
			ARConstraintAnchor anchor; 
			anchor.topologicalAnchor = nIndex;
			anchor.pos = refPos;
			if ( constraintMode == ICP_CONSTRAINT_POINT_TO_POINT ){
				posConstraint.setup( anchor, matchP, IDENTITY3F*constraintsWeight, time );
				receiver->addPositionConstraint( &posConstraint );
				if (showCorrVis) {
					corrVis.push_back(VisInfo(defRefPos, posConstraint.toPoint, norm(defRefPos - posConstraint.toPoint)));
				}
            statistics_sumOfSquaredNormalDistances += sqrNormQ;
			} else if ( constraintMode == ICP_CONSTRAINT_POINT_TO_PLANE ){
				Matrix3f quadric;
				for ( unsigned int i = 0; i < 3; i++ ){
					for ( unsigned int j = 0; j < 3; j++ ){
						quadric[i][j] = defRefNormal[i] * defRefNormal[j];
					}
				}
				if (useDataErrorQuad) {
				   Vector6f errVS = dataPS->get<float32, 6>(currentPointIndex, MATCH_ERROR_QUAD_AAT);
				   Matrix3f dataErrorQuad = makeMatrix3f(
				      errVS[0], errVS[3], errVS[5],
				      errVS[3], errVS[1], errVS[4],
				      errVS[5], errVS[4], errVS[2]
				   );
				   Matrix3f projN = quadric;
				   quadric = (projN.transpose()) * dataErrorQuad * projN;
				}
				
				posConstraint.setup( anchor, matchP, quadric*constraintsWeight, time );
				receiver->addPositionConstraint( &posConstraint );
				float32 res = sqrt(fabs((quadric * q)*q));
				if (showCorrVis) {
					corrVis.push_back(VisInfo(defRefPos, posConstraint.toPoint, res));
				}
	         statistics_sumOfSquaredNormalDistances += sqrNormQ;
   		}
			
			if (countCorrespondences) {
			   referencePS->set1i(nIndex, numDataPtsAAT, referencePS->get1i(nIndex, numDataPtsAAT)+1);
			}
		}
	
	   // eats 50% of the computation time	
		//progressWindow->progress( card32( float32( pIndex ) * 100 / float32( matchingShape->getNumPoints() ) ) );
      currentPointIndex++;		
		basicIt->next();
	}
	basicIt->unlock();

	if( showCorrVis && useDebugRenderer)
	{
		debugRenderer->clearDebugData("icp constraints");
		debugRenderer->beginRenderJob("icp constraints");
		for( unsigned i=0;i<corrVis.size();i++ )
		{
			VisInfo corr = corrVis[i];
			debugRenderer->addLine( corr.first,corr.second,makeVector3f(1,0,0),makeVector3f(1,1,0),1 );
		}
		debugRenderer->endRenderJob();
	}

	//progressWindow->popSteps();

	//if (showCorrVis) debugOutput << "      ICPConstraints: " << numPrunedPoints << " outlier points pruned.\n"; debugOutput.flush();
	//if (showCorrVis) debugOutput << "      ICPConstraints: " << statistics_sumOfSquaredNormalDistances << " statistics_sumOfSquaredNormalDistances.\n"; debugOutput.flush();
   
   statistics_numOutliers = numPrunedPoints;

	delete it;
}


void ICPConstraints::buildReverseICPConstraints(
   const ARSettings *settings,
   ARConstraintsInterface *receiver,
   PointCloud *referenceShape,
	PointCloud *deformedReference,
   PointCloud *matchingShape,
	const float32 time,
   const float32 skipPercentage,
	bool useDebugRenderer)
{
   clearStatistics();
	UnstructuredInCorePointCloud* referenceUPC = dynamic_cast<UnstructuredInCorePointCloud*>( referenceShape );
	UnstructuredInCorePointCloud* deformedReferenceUPC = dynamic_cast<UnstructuredInCorePointCloud*>( deformedReference );
	UnstructuredInCorePointCloud* matchingShapeUPC = dynamic_cast<UnstructuredInCorePointCloud*>( matchingShape );

   if (!referenceUPC || !deformedReferenceUPC || ! matchingShapeUPC) {
      throw X4Exception("works only for UPCs");
   }
	
	PointSet* referencePS = referenceUPC->getPointSet();
	PointSet* deformedReferencePS = deformedReferenceUPC->getPointSet();
	PointSet* matchingShapePS = matchingShapeUPC->getPointSet();

	if (!referencePS->getNumEntries() == deformedReferencePS->getNumEntries()) {
		error("ICPConstraints::buildICPConstraints() - deformed reference and reference should have the same number of points");
		return;
	}
	if (matchingShapePS->getNumEntries() <= 0) return;
	
	AAT REFERENCE_POSITION			  = referencePS->getAAT( "position" );
	AAT DEFORMED_REFERENCE_POSITION = deformedReferencePS->getAAT( "position" );
	AAT DEFORMED_REFERENCE_NORMAL	  = deformedReferencePS->getAAT( "normal" );
	AAT MATCH_POSITION              = matchingShape->getAAT( "position" );
	AAT MATCH_NORMAL              = matchingShape->getAAT( "normal" );

	
	HierarchicalKNNIterator hIt(matchingShape,32,NULL);
	SceneObjectIterator* defRefIt = deformedReference->createIterator( SOIT::CAP_BASIC_PC );
	BasicPointCloudIterator* basicDefRefIt = dynamic_cast<BasicPointCloudIterator*>( defRefIt );

	float32 sqrPruneDistance = 0;
	if ( distanceOutlierMode == ICP_OUTLIERS_PERCENTILE ){
		basicDefRefIt->reset();
		std::vector<float32> sqrDistances; sqrDistances.resize( deformedReferencePS->getNumEntries() );
		card32 index = 0;
		while( !basicDefRefIt->atEnd() ){
			Vector3f deformedPos = basicDefRefIt->get3f( DEFORMED_REFERENCE_POSITION );
			hIt.setSeekPointAndReset(deformedPos);
			int32 nIndex = hIt.getCurrentPointIndex();//knn.getNearestPointIndex(deformedPos);
			const Vector3f matchPos = matchingShapePS->get3f( nIndex, MATCH_POSITION );
			const float32 sqrD = ( matchPos - deformedPos ) * ( matchPos - deformedPos );
			sqrDistances[index] = sqrD;
			index++;
			basicDefRefIt->next();
		}
		sort(sqrDistances.begin(), sqrDistances.end());
		int32 distIndex = (1.0f-settings->icpOutlierPercentile)*(matchingShape->getNumPoints()-1);
		if (distIndex >= (int32)matchingShape->getNumPoints()) {
   		distIndex = matchingShape->getNumPoints()-1;
		}
		if (distIndex < 0) {
		   distIndex = 0;
		}
		sqrPruneDistance = sqrDistances[distIndex];
	}

	basicDefRefIt->reset();
   card32 currentPointIndex = 0;
	card32 numPrunedPoints = 0;
	srand(42);
	X4_AVERAGE_TIMER_DEFINE(icpCon);
	X4_AVERAGE_TIMER_DEFINE(addCon);
	X4_AVERAGE_TIMER_START(icpCon);
	while( !basicDefRefIt->atEnd() ) {
	   if (skipPercentage != 0.0f) {
	      if (rnd01()<skipPercentage) {
   	      currentPointIndex++;
	         basicDefRefIt->next();
	         continue;
	      }
	   }
		const Vector3f deformedPos = basicDefRefIt->get3f( DEFORMED_REFERENCE_POSITION );
		const Vector3f deformedNormal = basicDefRefIt->get3f( DEFORMED_REFERENCE_NORMAL );
		hIt.setSeekPointAndReset(deformedPos);
		int32 matchIndex = hIt.getCurrentPointIndex();//knn.getNearestPointIndex(deformedPos);
		const Vector3f matchPos = matchingShapePS->get3f(matchIndex, MATCH_POSITION);
		const Vector3f matchNormal = matchingShapePS->get3f(matchIndex, MATCH_NORMAL);
		
		bool acceptMatchPoint = true;
		Vector3f q = deformedPos - matchPos;
		const float32 sqrNormQ = q * q;
      /// in the near field, culling is always disabled
      bool inNearField = sqrNormQ < sqr( settings->topSurfelSpacing * outliersDistanceFactor );
		// Check for outliers
		if ( cullNormalOutliers  ){
			
			// Normal-based outlier pruning (modified mb)
			const float32 e = deformedNormal * matchNormal;
			if ( e < outliersNormals ){
				acceptMatchPoint = false;
			}
		}
		if ( distanceOutlierMode == ICP_OUTLIERS_DISTANCE ){
			// Distance-based outlier pruning
			if ( !inNearField ){
				acceptMatchPoint = false;
			}
		} else if ( distanceOutlierMode == ICP_OUTLIERS_PERCENTILE ){
			// Percentile-based outlier pruning
			if ( sqrNormQ > sqrPruneDistance && !inNearField ){
				acceptMatchPoint = false;
			}
		}

		if ( !acceptMatchPoint ){
			numPrunedPoints++; 
		} else {
			PositionARConstraint posConstraint;
			ARConstraintAnchor anchor; 
			anchor.topologicalAnchor = currentPointIndex;
			anchor.pos = referencePS->get3f(currentPointIndex, REFERENCE_POSITION);
            Matrix3f quadric =
                (constraintMode == ICP_CONSTRAINT_POINT_TO_POINT) ?
                  (IDENTITY3F) :
                  (outerProduct(matchNormal,matchNormal));
            // + IDENTITY3F * 0.01f;
			/*for (unsigned int i = 0; i < 3; i++){
				for (unsigned int j = 0; j < 3; j++){
					//quadric[i][j] = deformedNormal[i] * deformedNormal[j];
					quadric[i][j] = matchNormal[i] * matchNormal[j];
				}
			}/**/
			posConstraint.setup( anchor, matchPos, quadric*constraintsWeight, time );
			X4_AVERAGE_TIMER_STOP(icpCon);
			X4_AVERAGE_TIMER_START(addCon);
			receiver->addPositionConstraint( &posConstraint );
			X4_AVERAGE_TIMER_STOP(addCon);
			X4_AVERAGE_TIMER_START(icpCon);
			float32 res = sqrt(fabs((quadric * q)*q));
         statistics_sumOfSquaredNormalDistances += sqrNormQ;
	      if(useDebugRenderer) {
				Vector3f matchNormal = matchingShapePS->get3f(matchIndex,MATCH_NORMAL);
				Vector3f projPoint = deformedPos  - matchNormal * (matchNormal * (deformedPos - matchPos));
				float relD = norm(projPoint - deformedPos) / (settings->topSurfelSpacing * outliersDistanceFactor);
				Vector3f col = makeVector3f(1,0,0) * relD + makeVector3f(0,1,0) * (1.0f - relD);
				if( relD > 1.0f )
					col = makeVector3f(1,1,0);
			   debugRenderer->addLine( matchPos,deformedPos,col,col,1 );
				//debugRenderer->addSurfel(projPoint,matchNormal,col);
				float scale = settings->topSurfelSpacing * 0.5f;
				Matrix3f tangentFrame = calcTangentSystem(matchNormal);
				debugRenderer->addQuad(
					matchPos + tangentFrame[0] * scale,
					matchPos + tangentFrame[1] * scale,
					matchPos - tangentFrame[0] * scale,
					matchPos - tangentFrame[1] * scale,col);
	      }
		}
      currentPointIndex++;		
		basicDefRefIt->next();
	}
	X4_AVERAGE_TIMER_STOP(icpCon);
	X4_AVERAGE_TIMER_REPORT_AND_RESET(icpCon);
	X4_AVERAGE_TIMER_REPORT_AND_RESET(addCon);
	basicDefRefIt->unlock();

	debugOutput << "icp constraints: " << referenceShape->getNumPoints() - numPrunedPoints << " pruned: " << numPrunedPoints << "\n";

   statistics_numOutliers = numPrunedPoints;

	delete defRefIt;
}


void ICPConstraints::buildSimpleICPConstraints(
   const ARSettings *settings,
   ARConstraintsInterface *receiver,
   PointCloud *referenceShape,
	PointCloud *deformedReference,
   PointCloud *matchingShape,
	const float32 time,
   Delegate callForRepaint,
   bool showCorrVis)
{
	SceneObjectIterator *refIt = referenceShape->createIterator(SOIT::CAP_BASIC_PC);
	SceneObjectIterator *defRefIt = deformedReference->createIterator(SOIT::CAP_BASIC_PC);
	SceneObjectIterator *matchIt = matchingShape->createIterator(SOIT::CAP_BASIC_PC);
	BasicPointCloudIterator *bRefIt = dynamic_cast<BasicPointCloudIterator*>(refIt);
	BasicPointCloudIterator *bDefRefIt = dynamic_cast<BasicPointCloudIterator*>(defRefIt);
	BasicPointCloudIterator *bMatchIt = dynamic_cast<BasicPointCloudIterator*>(matchIt);
	if (!bRefIt || !bDefRefIt || ! bMatchIt) {
		delete refIt; delete defRefIt; delete matchIt;
		error("something went wrong.");
		return;
	}

	const PointSet *refPS = bRefIt->lockBlockRead();
	const PointSet *defRefPS = bDefRefIt->lockBlockRead();
	const PointSet *matchPS = bMatchIt->lockBlockRead();

	if (!refPS->getNumEntries() == defRefPS->getNumEntries()) {
		error("ICPConstraints::buildICPConstraints() - deformed reference and reference should have the same number of points");
		return;
	}
	
	AAT REFERENCE_POSITION_AAT			   = refPS->getAAT( "position" );
	AAT DEFORMED_REFERENCE_POSITION_AAT = defRefPS->getAAT( "position" );
	AAT MATCH_POSITION_AAT              = matchPS->getAAT( "position" );

	//AttachedIndexedOctree* octree = PointSetKNNQuery::createOctree( defRefPS, 20 );
	//PointSetKNNQuery knn( defRefPS, octree, makeVector3f( 0,0,0 ), DEFORMED_REFERENCE_POSITION_AAT );
	PointSetANNQuery knn(defRefPS, 1);
   if (defRefPS->getNumEntries() > 0) {

	   for (card32 matchIndex = 0; matchIndex<matchPS->getNumEntries(); matchIndex++) {

		   const Vector3f& matchP = matchPS->get3f(matchIndex, MATCH_POSITION_AAT);
		   card32 nIndex = knn.getNearestPointIndex(matchP);
		   const Vector3f refPos = refPS->get3f( nIndex, REFERENCE_POSITION_AAT );
		   PositionARConstraint posConstraint;
		   ARConstraintAnchor anchor; 
		   anchor.topologicalAnchor = nIndex;
		   anchor.pos = refPos;
		   posConstraint.setup( anchor, matchP, IDENTITY3F*constraintsWeight, time );
		   receiver->addPositionConstraint( &posConstraint );
		   //if (showCorrVis) {
   		//   const Vector3f defRefPos = defRefPS->get3f( nIndex, DEFORMED_REFERENCE_POSITION_AAT );
		   //   corrVis.push_back(pair<Vector3f, Vector3f>(defRefPos, matchP));
		   //}
	   }
   }
   
	// delete octree;

	bRefIt->unlock();
	bDefRefIt->unlock();
	bMatchIt->unlock();
	
	delete refIt;
	delete defRefIt;
	delete matchIt;
	
	//if (showCorrVis) callForRepaint.callIfNotNull();
}


void ICPConstraints::buildPerfectConstraints(
   const ARSettings *settings,
   ARConstraintsInterface *receiver,
   PointCloud *referenceShape,
	PointCloud *deformedReference,
   PointCloud *matchingShape,
	const float32 time)
{
	SceneObjectIterator *refIt = referenceShape->createIterator(SOIT::CAP_BASIC_PC);
	SceneObjectIterator *defRefIt = deformedReference->createIterator(SOIT::CAP_BASIC_PC);
	SceneObjectIterator *matchIt = matchingShape->createIterator(SOIT::CAP_BASIC_PC);
	BasicPointCloudIterator *bRefIt = dynamic_cast<BasicPointCloudIterator*>(refIt);
	BasicPointCloudIterator *bDefRefIt = dynamic_cast<BasicPointCloudIterator*>(defRefIt);
	BasicPointCloudIterator *bMatchIt = dynamic_cast<BasicPointCloudIterator*>(matchIt);
	if (!bRefIt || !bDefRefIt || ! bMatchIt) {
		delete refIt; delete defRefIt; delete matchIt;
		error("something went wrong.");
		return;
	}

	const PointSet *refPS = bRefIt->lockBlockRead();
	const PointSet *defRefPS = bDefRefIt->lockBlockRead();
	const PointSet *matchPS = bMatchIt->lockBlockRead();

	if (!refPS->getNumEntries() == defRefPS->getNumEntries()) {
		error("ICPConstraints::buildICPConstraints() - deformed reference and reference should have the same number of points");
		return;
	}
	
	AAT REFERENCE_POSITION_AAT			   = refPS->getAAT( "position" );
	AAT DEFORMED_REFERENCE_POSITION_AAT = defRefPS->getAAT( "position" );
	AAT MATCH_POSITION_AAT              = matchPS->getAAT( "position" );

	for (card32 matchIndex = 0; matchIndex<matchPS->getNumEntries(); matchIndex++) {

		Vector3f matchPos = matchPS->get3f(matchIndex, MATCH_POSITION_AAT);
		Vector3f refPos = refPS->get3f(matchIndex, REFERENCE_POSITION_AAT );
		PositionARConstraint posConstraint;
		ARConstraintAnchor anchor; 
		anchor.topologicalAnchor = matchIndex;
		anchor.pos = refPos;
		posConstraint.setup( anchor, matchPos, IDENTITY3F*constraintsWeight, time );
		receiver->addPositionConstraint( &posConstraint );
	}

	bRefIt->unlock();
	bDefRefIt->unlock();
	bMatchIt->unlock();
	
	delete refIt;
	delete defRefIt;
	delete matchIt;
}

static inline void drawGradientLine(const Vector3f &p, const Vector3f &q) {
   glBegin(GL_LINE_STRIP);
   for (card32 i=0; i<10; i++) {
      float32 x = i/9.0f;
      glColor4fv(blueRedGradient(x, 0.1f).data());
      glVertex3fv((p*(1.0f-x)+q*x).data());
   }
   glEnd();
}


void ICPConstraints::glDrawFeedback(GLContext *glContext, ARVisualizationSettings *visSettings) {
   if (corrVis.empty()) return;
   glPushAttrib(GL_ALL_ATTRIB_BITS);
   glEnable(GL_BLEND);
   glLineWidth(2);
   glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
   glColor4f(1.0f, 1.0f, 1.0f, 0.2f);
   glBegin(GL_LINES);
   for (unsigned i=0; i<corrVis.size(); i++) {
      float32 l = corrVis[i].res;
      Vector4f col = rainbowGradient(l*100);
      glColor4fv(col.data());
      glVertex3fv(corrVis[i].first.data());
      glVertex3fv(corrVis[i].second.data());
      //drawGradientLine(corrVis[i].first, corrVis[i].second);
   }
   glEnd();
   glPopAttrib();
   corrVis.clear();
}


void ICPConstraints::clearStatistics() 
{
   statistics_numOutliers = 0;
   statistics_sumOfSquaredNormalDistances = 0.0;
}

void X4::ICPConstraints::buildConditionalReverseICPConstraints( 
	const ARSettings *settings,
	ARConstraintsInterface *receiver,
	/// parametrization (urshape)
	PointCloud *referenceShape,
	/// urshape as in its current form ("best guess"), used for correspondences
	PointCloud *deformedReference,
	/// data points
	PointCloud *matchingShape,
	/// cached query for matchingShape
	const PointSetANNQuery* annQuery,
	/// the constraints are specified for this time
	const float32 time,
	/// if this is more than 0, the corresponding amount of constraint points is skipped using a fixed random pattern (srand(42))
	const float32 skipPercentage ,
	/// use debug renderer to visualize constraints
	bool useDebugRenderer,
	/// precomputed distance field for nearest neighbor query
	PrecomputedDistanceField * df
	
	)
{

	clearStatistics();
	UnstructuredInCorePointCloud* referenceUPC = dynamic_cast<UnstructuredInCorePointCloud*>( referenceShape );
	UnstructuredInCorePointCloud* deformedReferenceUPC = dynamic_cast<UnstructuredInCorePointCloud*>( deformedReference );
	UnstructuredInCorePointCloud* matchingShapeUPC = dynamic_cast<UnstructuredInCorePointCloud*>( matchingShape );

	if (!referenceUPC || !deformedReferenceUPC || ! matchingShapeUPC) {
		throw X4Exception("works only for UPCs");
	}

	PointSet* referencePS = referenceUPC->getPointSet();
	PointSet* deformedReferencePS = deformedReferenceUPC->getPointSet();
	PointSet* matchingShapePS = matchingShapeUPC->getPointSet();

	if (!referencePS->getNumEntries() == deformedReferencePS->getNumEntries()) {
		error("ICPConstraints::buildICPConstraints() - deformed reference and reference should have the same number of points");
		return;
	}
	if (matchingShapePS->getNumEntries() <= 0) return;

	AAT REFERENCE_POSITION			= referencePS->getAAT( "position" );
	AAT DEFORMED_REFERENCE_POSITION = deformedReferencePS->getAAT( "position" );
	AAT DEFORMED_REFERENCE_NORMAL	= deformedReferencePS->getAAT( "normal" );
	AAT MATCH_POSITION              = matchingShape->getAAT( "position" );
	AAT MATCH_NORMAL				= matchingShape->getAAT( "normal"   );
	AAT MATCH_FLAGS	; bool hasFlags = false;
	if (matchingShape->providesAttribute("flags"))
	{
		MATCH_FLAGS= matchingShape->getAAT( "flags"    );
		hasFlags = true;
	}
	

	//HierarchicalKNNIterator hIt(matchingShape,32,NULL);
	HierarchicalKNNIterator * hIt = NULL;
	if( annQuery == NULL && df == NULL )
	{
		hIt = new HierarchicalKNNIterator(matchingShape,32,NULL);
	}

	SceneObjectIterator* defRefIt = deformedReference->createIterator( SOIT::CAP_BASIC_PC );
	BasicPointCloudIterator* basicDefRefIt = dynamic_cast<BasicPointCloudIterator*>( defRefIt );

	float32 sqrPruneDistance = 0;
	if ( distanceOutlierMode == ICP_OUTLIERS_PERCENTILE ){
		basicDefRefIt->reset();
		std::vector<float32> sqrDistances; sqrDistances.resize( deformedReferencePS->getNumEntries() );
		card32 index = 0;
		while( !basicDefRefIt->atEnd() ){
			Vector3f deformedPos = basicDefRefIt->get3f( DEFORMED_REFERENCE_POSITION );
			
			
			int32 nIndex;
			Vector3f matchPos;
			
			
			if( df == NULL )
			{
				if( annQuery == NULL)
				{
					hIt->setSeekPointAndReset(deformedPos);
					nIndex = hIt->getCurrentPointIndex();
				}
				else
				{
					nIndex =annQuery->getNearestPointIndex(deformedPos);
				}
			}
			else
			{
				nIndex =  df->getClosestPointIndex(deformedPos);
			}
			matchPos = matchingShapePS->get3f( nIndex, MATCH_POSITION );
			
			const float32 sqrD = ( matchPos - deformedPos ) * ( matchPos - deformedPos );
			sqrDistances[index] = sqrD;
			index++;
			basicDefRefIt->next();
		}
		sort(sqrDistances.begin(), sqrDistances.end());
		int32 distIndex = (1.0f-settings->icpOutlierPercentile)*(matchingShape->getNumPoints()-1);
		if (distIndex >= (int32)matchingShape->getNumPoints()) {
			distIndex = matchingShape->getNumPoints()-1;
		}
		if (distIndex < 0) {
			distIndex = 0;
		}
		sqrPruneDistance = sqrDistances[distIndex];
	}

	basicDefRefIt->reset();
	card32 currentPointIndex = 0;
	card32 numPrunedPoints = 0;
	card32 numConstraintsAdded = 0;
	//srand(42);
	//X4_AVERAGE_TIMER_DEFINE(icpCon);
	//X4_AVERAGE_TIMER_DEFINE(addCon);
	//X4_AVERAGE_TIMER_START(icpCon);
	while( !basicDefRefIt->atEnd() ) {
		if (skipPercentage != 0.0f) {
			if (rnd01()<skipPercentage) {
				currentPointIndex++;
				basicDefRefIt->next();
				continue;
			}
		}
		const Vector3f deformedPos = basicDefRefIt->get3f( DEFORMED_REFERENCE_POSITION );
		const Vector3f deformedNormal = basicDefRefIt->get3f( DEFORMED_REFERENCE_NORMAL );
		
		//ann here
		//hIt.setSeekPointAndReset(deformedPos);
		//int32 matchIndex = hIt.getCurrentPointIndex();//knn.getNearestPointIndex(deformedPos);
//		int32 matchIndex =annQuery->getNearestPointIndex(deformedPos);
		Vector3f matchPos, matchNormal;
		bool acceptMatchPoint = true;
		int32 matchIndex;
		if( df == NULL )
		{
			if( annQuery == NULL)
			{
				hIt->setSeekPointAndReset(deformedPos);
				matchIndex = hIt->getCurrentPointIndex();
			}
			else
			{
				matchIndex =annQuery->getNearestPointIndex(deformedPos);
			}

		}
		else
		{
			matchIndex = df->getClosestPointIndex(deformedPos);
		}

		matchPos = matchingShapePS->get3f(matchIndex, MATCH_POSITION);
		matchNormal = matchingShapePS->get3f(matchIndex, MATCH_NORMAL);

		//prune border points
// 		if (hasFlags && (matchingShapePS->get1i(matchIndex, MATCH_FLAGS) & PF_FLAG_SELECTED))
// 		{
// 			acceptMatchPoint = false;
// 		}

		Vector3f q = deformedPos - matchPos;
		const float32 sqrNormQ = q * q;

		if( acceptMatchPoint )
		{

			/// in the near field, culling is always disabled
			bool inNearField = sqrNormQ < sqr( settings->topSurfelSpacing * outliersDistanceFactor );
			// Check for outliers
			if ( cullNormalOutliers  ){

				// Normal-based outlier pruning (modified mb)
				const float32 e = deformedNormal * matchNormal;
				if ( e < outliersNormals ){
					acceptMatchPoint = false;
				}
			}
			if ( distanceOutlierMode == ICP_OUTLIERS_DISTANCE ){
				// Distance-based outlier pruning
				if ( !inNearField ){
					acceptMatchPoint = false;
				}
			} else if ( distanceOutlierMode == ICP_OUTLIERS_PERCENTILE ){
				// Percentile-based outlier pruning
				if ( sqrNormQ > sqrPruneDistance && !inNearField ){
					acceptMatchPoint = false;
				}
			}

		}


		if ( !acceptMatchPoint ){
			numPrunedPoints++; 
		} else {
			PositionARConstraint posConstraint;
			ARConstraintAnchor anchor; 
			anchor.topologicalAnchor = currentPointIndex;
			anchor.pos = referencePS->get3f(currentPointIndex, REFERENCE_POSITION);
			Matrix3f quadric = outerProduct(matchNormal,matchNormal);// + IDENTITY3F * 0.1f;
			posConstraint.setup( anchor, matchPos, quadric*constraintsWeight, time );
			//X4_AVERAGE_TIMER_STOP(icpCon);
			//X4_AVERAGE_TIMER_START(addCon);
			if( receiver ){
				receiver->addPositionConstraint( &posConstraint );
				numConstraintsAdded++;
			}
			//X4_AVERAGE_TIMER_STOP(addCon);
			//X4_AVERAGE_TIMER_START(icpCon);
			float32 res = sqrt(fabs((quadric * q)*q));
			statistics_sumOfSquaredNormalDistances += sqrNormQ;

			if(useDebugRenderer) {
				//Vector3f matchNormal = matchingShapePS->get3f(matchIndex,MATCH_NORMAL);
				Vector3f projPoint = deformedPos  - matchNormal * (matchNormal * (deformedPos - matchPos));
				float relD = norm(projPoint - deformedPos) / (settings->topSurfelSpacing * outliersDistanceFactor);
				Vector3f col = makeVector3f(1,0,0) * relD + makeVector3f(0,1,0) * (1.0f - relD);
				if( relD > 1.0f )
					col = makeVector3f(1,1,0);
				debugRenderer->addLine( matchPos,deformedPos,col,col,1 );
				//debugRenderer->addSurfel(projPoint,matchNormal,col);
				float scale = settings->topSurfelSpacing * 0.01f;
				Matrix3f tangentFrame = calcTangentSystem(matchNormal);
				debugRenderer->addQuad(
					matchPos + tangentFrame[0] * scale,
					matchPos + tangentFrame[1] * scale,
					matchPos - tangentFrame[0] * scale,
					matchPos - tangentFrame[1] * scale,col);
			}
		}
		currentPointIndex++;		
		basicDefRefIt->next();
	}
	//X4_AVERAGE_TIMER_STOP(icpCon);
	//X4_AVERAGE_TIMER_REPORT_AND_RESET(icpCon);
	//X4_AVERAGE_TIMER_REPORT_AND_RESET(addCon);
	basicDefRefIt->unlock();

	statistics_numOutliers = numPrunedPoints;

	debugOutput <<"frame: "<< time<<" - "<< numConstraintsAdded <<" icp constraints added...\n";

	delete defRefIt;

}