//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "FDDeformationSolver.h"
//---------------------------------------------------------------------------
#include "FiniteDifferencesDeformation.h"
#include "FiniteDifferencesDeformation.inline.h"
#include "SGEVolumeGridDeformationSequence.h"
#include "SGEPointCloud.h"
#include "RegularVolume.h"
#include "RegularVolume.inline.h"
#include "RegularInCoreVolume.h"
#include "DynamicLinearAlgebraTools.h"
#include "DynamicLinearAlgebraTools.inline.h"
#include "BoundingBox.h"
#include "BoundingBox.inline.h"
#include "IterativeSolvers.h"
#include "IterativeSolvers.inline.h"
#include "ARSettings.h"
#include "ARConstraints.h"
#include "FDDeformationContext.h"
#include "FDDeformationEvaluator.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------



IMPLEMENT_X4_CLASS( FDDeformationSolver ,0)
{
	BEGIN_CLASS_INIT( FDDeformationSolver );
}

FDDeformationSolver::FDDeformationSolver()
{
	context = NULL;
	settings = NULL;
}

FDDeformationSolver::~FDDeformationSolver() {}



AROptimizationContext *FDDeformationSolver::createOptContext() {
	return new FDDeformationContext();
}

void FDDeformationSolver::setup(AROptimizationContext *context, const ARSettings *settings) {
	if (!dynamic_cast<FDDeformationContext*>(context)) {
		throw ETypeMismatch("FDDeformationSolver::setup() - context does not have the correct dynamic type.");
	}
	this->context = (FDDeformationContext*)context;
	this->settings = settings;
}

void FDDeformationSolver::beginConstraints() 
{
   std::string msg;

   RegularInCoreVolume *vol0 = context->getVolume(0);
   positionAAT = vol0->getAAT("position");
   volWidth = vol0->getVoxels()->getWidth();
   volHeight = vol0->getVoxels()->getHeight();
   volDepth = vol0->getVoxels()->getDepth();

   constraintsQuadric.setRows(volWidth*volHeight*volDepth*3);
   constraintsQuadric.setZero();
   constraintsLinear.setDim(volWidth*volHeight*volDepth*3);
   constraintsLinear.setZero();
}


void FDDeformationSolver::addExampleConstraints() {
   RegularInCoreVolume *vol0 = context->getVolume(0);
   const card32 w = volWidth-2;
   const card32 h = volHeight-2;
   const card32 d = volDepth-2;
   const BoundingBox3f bb = vol0->getVoxels()->getBoundingBox();
   for (card32 z=0; z<2; z++) {
      for (card32 y=0; y<2; y++) {
         for (card32 x=0; x<2; x++) {
         addVoxelConstraint(0+x,0+y,0+z, context->getVolume(0)->getVoxels()->getVoxel<float32, 3>(0+x,0+y,0+z, positionAAT));   
         addVoxelConstraint(w+x,0+y,0+z, context->getVolume(0)->getVoxels()->getVoxel<float32, 3>(w+x,0+y,0+z, positionAAT));   
         addVoxelConstraint(0+x,h+y,0+z, context->getVolume(0)->getVoxels()->getVoxel<float32, 3>(0+x,h+y,0+z, positionAAT) + makeVector3f(0,bb.getSideLength(1) * 0.3f,0));   
         addVoxelConstraint(w+x,h+y,0+z, context->getVolume(0)->getVoxels()->getVoxel<float32, 3>(w+x,h+y,0+z, positionAAT));   
         addVoxelConstraint(0+x,0+y,d+z, context->getVolume(0)->getVoxels()->getVoxel<float32, 3>(0+x,0+y,d+z, positionAAT));
         addVoxelConstraint(w+x,0+y,d+z, context->getVolume(0)->getVoxels()->getVoxel<float32, 3>(w+x,0+y,d+z, positionAAT));   
         addVoxelConstraint(0+x,h+y,d+z, context->getVolume(0)->getVoxels()->getVoxel<float32, 3>(0+x,h+y,d+z, positionAAT) + makeVector3f(0,bb.getSideLength(1) * 0.4f,0));   
         addVoxelConstraint(w+x,h+y,d+z, context->getVolume(0)->getVoxels()->getVoxel<float32, 3>(w+x,h+y,d+z, positionAAT));   
      }
    }
  }
}


void FDDeformationSolver::addPositionConstraint(const PositionARConstraint *c) {
   int32 t = (int32)c->getTime();
   if (t >= (int32)context->getNumVolumes() || t<0) throw ERangeCheck("time out of range");
   Vector3f posAt = c->getAnchor().getPos();
   Vector3f posTo = c->getToPoint();

   Matrix4f transform = bbToVoxelTransform(context->getVolume(t)->getVoxels());
   Vector3f posAtLocal = transformVector3f(transform, posAt);
   int32 w = context->getVolume(t)->getVoxels()->getWidth();
   int32 h = context->getVolume(t)->getVoxels()->getHeight();
   int32 d = context->getVolume(t)->getVoxels()->getDepth();
   Vector3i posAtInt = makeVector3i(
      min(max(0, (int32)posAtLocal[0]), w-1),
      min(max(0, (int32)posAtLocal[1]), h-1),
      min(max(0, (int32)posAtLocal[2]), d-1)
   );
   /// cheap & inaccurate solution - quantize constraint points
   addVoxelConstraint(posAtInt[0], posAtInt[1], posAtInt[2], posTo);//-delta);
}

void FDDeformationSolver::addVelocityConstraint(const VelocityVectorARConstraint *c)
{
   notImplemented();
}

void FDDeformationSolver::addInstantaneousVelocityConstraint(const InstantaneousVelocityARConstraint *c) 
{
   notImplemented();
}

void FDDeformationSolver::addVoxelConstraint(card32 x, card32 y, card32 z, Vector3f pos) {
   card32 baseIndex = 3*(x + volWidth*y + volWidth*volHeight*z);
   constraintsQuadric[baseIndex  ].addEntry(baseIndex  , 2);
   constraintsQuadric[baseIndex+1].addEntry(baseIndex+1, 2);
   constraintsQuadric[baseIndex+2].addEntry(baseIndex+2, 2);
   constraintsLinear[baseIndex  ] -= 2*pos[0];
   constraintsLinear[baseIndex+1] -= 2*pos[1];
   constraintsLinear[baseIndex+2] -= 2*pos[2];
}

void FDDeformationSolver::solve() 
{
   for (card32 seq=0; seq<context->getNumVolumes(); seq++) {
      float32 score = 0;
      DVectorF gradient, x;
      SparseMatrixF hessian;
      for (card32 newtonIter=0; newtonIter<settings->numNewtonIterations; newtonIter++) {
         FDDeformSolverTools::buildRegScoreTaylorApprox(
            context->getVolume(seq)->getVoxels(),
            score,
            gradient,
            hessian,
            settings);
         
         FDDeformSolverTools::convertVolToVec(context->getVolume(seq)->getVoxels(), x);

         const card32 nRows = hessian.getRows();
         for (card32 r=0; r<nRows; r++) {
            hessian[r].addEntry(r, settings->newtonRegularization);
         }

         // convert Taylor approx in quadric form
         gradient = hessian*x - gradient;
         
         // add quadratic constraints term, E(x) = 1/2 * [x]^t [Q] [x] + [x]^t[b] + const
         hessian += constraintsQuadric;
         gradient -= constraintsLinear; // "-=" --> to right hand side
         card32 maxNumIterations = settings->numCGIterations;
         conjugateGradientsSolve<float32>(
            hessian,
            x,
            gradient,
            maxNumIterations,
            settings->convThreshCG,
            true
         );
         FDDeformSolverTools::convertVecToVol(x, context->getVolume(seq)->getVoxels());
         callCallbacks();
      }
   }
   settings = NULL;
}



void FDDeformationSolver::callCallbacks() 
{
   for (unsigned i=0; i<callbacks.size(); i++) {
      callbacks[i]();
   }
}

void FDDeformationSolver::addUserFeedbackCallback(Delegate d) {
   callbacks.push_back(d);
}

void FDDeformationSolver::clearUserFeedbackCallbacks() 
{
   callbacks.clear();
}

DeformationEvaluator *FDDeformationSolver::createEvaluator() 
{
	return new FDDeformationEvaluator();
}

void X4::FDDeformationSolver::addHalfSpaceConstraint( const ARHalfSpaceConstraint *c )
{
// xxx - TODO
   notImplemented();
}

void X4::FDDeformationSolver::addSpatialDerivativeConstraint(const SpatialDerivativeARConstraint *c) {
   notImplemented();
}

void X4::FDDeformationSolver::addSimpleSpatialDerivativeConstraint(const SimpleSpatialDerivativeARConstraint *c) {
   notImplemented();
}
