//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "FiniteDifferencesDeformation.h"
#include "FiniteDifferencesDeformation.inline.h"
//---------------------------------------------------------------------------
#include "SGEVolumeGridDeformationSequence.h"
#include "SGEPointCloud.h"
#include "RegularVolume.h"
#include "RegularVolume.inline.h"
#include "DynamicLinearAlgebra.h"
#include "DynamiclinearAlgebra.inline.h"
#include "SparseLinearAlgebra.h"
#include "SparseLinearAlgebra.inline.h"
#include "PartialDerivative.h"
#include "PartialDerivative.inline.h"
#include "ARSettings.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------



// ------ helper functions

SGEShape* X4::createFDDeformSGEShapeTemplate() 
{
   SGEPointCloud *result = new SGEPointCloud();
   VertexDescriptor *vd = new VertexDescriptor();
   vd->pushAttrib(mVAD("position", 3, VAD::DATA_FORMAT_FLOAT32));
   result->setPointChannels(vd);
   return result;
}

SGEDeformationSequence* X4::createFDDeformSGEDefSeqTemplate() 
{
   SGEVolumeGridDeformationSequence *result = new SGEVolumeGridDeformationSequence();
   VertexDescriptor *vd = new VertexDescriptor();
   vd->pushAttrib(mVAD("position", 3, VAD::DATA_FORMAT_FLOAT32));
   vd->pushAttrib(mVAD("nonRigidWeight", 1, VAD::DATA_FORMAT_FLOAT32));
   result->setDescriptor(vd);
   return result;
}

Matrix4f X4::bbToVoxelTransform(RegularVolume *vol) {
   BoundingBox3f bb = vol->getBoundingBox();
   return makeScale4f(
      (vol->getWidth()-1) / bb.getSideLength(0),
      (vol->getHeight()-1)/ bb.getSideLength(1),
      (vol->getDepth()-1) / bb.getSideLength(2))
   * makeTranslation4f(
      -bb.lowerCorner
   );
}



// ------ deformation engine

static void compVarMap(
   card32 x, card32 y, card32 z, 
   card32 w, card32 h, card32 d, 
   StaticVector<mpcard, 12> &variableIndexMapping)
{
   card32 basePos = x + y*w + z*w*h;
   variableIndexMapping[0] = 3*basePos;
   variableIndexMapping[1] = 3*basePos+1;
   variableIndexMapping[2] = 3*basePos+2;
   card32 basePosX = x+1 + y*w + z*w*h;
   variableIndexMapping[3] = 3*basePosX;
   variableIndexMapping[4] = 3*basePosX+1;
   variableIndexMapping[5] = 3*basePosX+2;
   card32 basePosY = x + (y+1)*w + z*w*h;
   variableIndexMapping[6] = 3*basePosY;
   variableIndexMapping[7] = 3*basePosY+1;
   variableIndexMapping[8] = 3*basePosY+2;
   card32 basePosZ = x + y*w + (z+1)*w*h;
   variableIndexMapping[9] = 3*basePosZ;
   variableIndexMapping[10] = 3*basePosZ+1;
   variableIndexMapping[11] = 3*basePosZ+2;
}

template <typename VarFloatType, typename ConstFloatType>
static VarFloatType computeDeformationScore(
   const StaticVector<VarFloatType, 3> &pos, 
   const StaticVector<VarFloatType, 3> &posX, 
   const StaticVector<VarFloatType, 3> &posY, 
   const StaticVector<VarFloatType, 3> &posZ,
   const StaticVector<ConstFloatType, 3> &gridSpacing)
{
   StaticMatrix<VarFloatType, 3, 3> jac;
   jac[0] = (posX-pos)/gridSpacing[0];
   jac[1] = (posY-pos)/gridSpacing[1];
   jac[2] = (posZ-pos)/gridSpacing[2];
   StaticMatrix<VarFloatType, 3, 3> metricTensor = jac*jac.transpose();
   metricTensor[0][0] -= (VarFloatType)1;
   metricTensor[1][1] -= (VarFloatType)1;
   metricTensor[2][2] -= (VarFloatType)1;
   return sqr(metricTensor[0][0]) + sqr(metricTensor[0][1]) + sqr(metricTensor[0][2])
        + sqr(metricTensor[1][0]) + sqr(metricTensor[1][1]) + sqr(metricTensor[1][2])
        + sqr(metricTensor[2][0]) + sqr(metricTensor[2][1]) + sqr(metricTensor[2][2]);
}

static void computeLocalScoreTApprox(
   const Vector3f &pos, 
   const Vector3f &posX, const Vector3f &posY, const Vector3f &posZ, 
   const Vector3f &gridSpacing,
   float32 &localScore, 
   StaticVector<float32, 12> &localGradient, 
   StaticMatrix<float32, 12, 12> &localHessian,
   const ARSettings *settings)
{
   StaticVector< HessianTools<float32, 12>::DiffTwice, 3 > posAD;
   posAD[0] = HessianTools<float32, 12>::createVariable(pos[0], 0);
   posAD[1] = HessianTools<float32, 12>::createVariable(pos[1], 1);
   posAD[2] = HessianTools<float32, 12>::createVariable(pos[2], 2);
   StaticVector< HessianTools<float32, 12>::DiffTwice, 3 > posXAD;
   posXAD[0] = HessianTools<float32, 12>::createVariable(posX[0], 3);
   posXAD[1] = HessianTools<float32, 12>::createVariable(posX[1], 4);
   posXAD[2] = HessianTools<float32, 12>::createVariable(posX[2], 5);
   StaticVector< HessianTools<float32, 12>::DiffTwice, 3 > posYAD;
   posYAD[0] = HessianTools<float32, 12>::createVariable(posY[0], 6);
   posYAD[1] = HessianTools<float32, 12>::createVariable(posY[1], 7);
   posYAD[2] = HessianTools<float32, 12>::createVariable(posY[2], 8);
   StaticVector< HessianTools<float32, 12>::DiffTwice, 3 > posZAD;
   posZAD[0] = HessianTools<float32, 12>::createVariable(posZ[0], 9);
   posZAD[1] = HessianTools<float32, 12>::createVariable(posZ[1], 10);
   posZAD[2] = HessianTools<float32, 12>::createVariable(posZ[2], 11);
   
   HessianTools<float32, 12>::DiffTwice result = 
      computeDeformationScore(posAD, posXAD, posYAD, posZAD, gridSpacing);
   localScore = result.getValue().getValue();
   localGradient = HessianTools<float32, 12>::getGradient(result) * settings->fdDeformationWeight;
   localHessian = HessianTools<float32, 12>::getHessian(result) * settings->fdDeformationWeight;
}

static inline int mapv(int x, int y, int z, int w, int h) {
   int result = 3*(x+w*y+w*h*z);
   return result;
}

void FDDeformSolverTools::buildRegScoreTaylorApprox(
   RegularVolume *fVol,
   float32 &score,
   DVectorF &gradient,
   SparseMatrixF &hessian,
   const ARSettings *settings)
{
   const AAT posAAT = fVol->getDescr()->getToken("position", 3, VAD::DATA_FORMAT_FLOAT32);
   const int w = fVol->getWidth();
   const int h = fVol->getHeight();
   const int d = fVol->getDepth();

   Vector3f gridSpacing = 
      fVol->getBoundingBox().upperCorner -
      fVol->getBoundingBox().lowerCorner;
   gridSpacing[0] /= w;
   gridSpacing[1] /= h;
   gridSpacing[2] /= d;
   
   x4Assert(w>1 && h>1 && d>1);
   score = 0;
   gradient.setDim(w*h*d*3);
   gradient.setZero();
   hessian.setRows(w*h*d*3);
   hessian.setZero();
   float32 localScore = 0;
   StaticMatrix<float32, 12, 12> localHessian;
   StaticVector<float32, 12> localGradient;
   StaticVector<mpcard, 12> variableIndexMapping;

   for (int z=0; z<d-1; z++) {
      for (int y=0; y<h-1; y++) {
         for (int x=0; x<w-1; x++) {
            Vector3f pos  = fVol->getVoxel<float32, 3>(x, y, z, posAAT);
            Vector3f posX = fVol->getVoxel<float32, 3>(x+1,y,z, posAAT);
            Vector3f posY = fVol->getVoxel<float32, 3>(x,y+1,z, posAAT);
            Vector3f posZ = fVol->getVoxel<float32, 3>(x,y,z+1, posAAT);
            compVarMap(x,y,z, w,h,d, variableIndexMapping);
            computeLocalScoreTApprox(
               pos, posX, posY, posZ, gridSpacing,
               localScore, localGradient, localHessian, settings);
            for(card32 yi=0; yi<12; yi++) {
               for(card32 xi=0; xi<12; xi++) {
                  if (localHessian[xi][yi] != 0.0f) {
                     hessian[variableIndexMapping[yi]].addEntry(variableIndexMapping[xi], localHessian[xi][yi]);
                  }
               }
               gradient[variableIndexMapping[yi]] += localGradient[yi];
            }
            score += localScore;
         }
      }
   }


   //for (int z=0; z<d; z++) {
   //   for (int y=0; y<h; y++) {
   //      for (int x=0; x<w; x++) {
   //         for (int i=0; i<3; i++) {
   //            int s = 0;
   //            if (x<w-1) {hessian[mapv(x,y,z,w,h)+i].addEntry(mapv(x+1,y,z,w,h)+i,-1);s++;}
   //            if (y<h-1) {hessian[mapv(x,y,z,w,h)+i].addEntry(mapv(x,y+1,z,w,h)+i,-1);s++;}
   //            if (z<d-1) {hessian[mapv(x,y,z,w,h)+i].addEntry(mapv(x,y,z+1,w,h)+i,-1);s++;}
   //            if (x>0)   {hessian[mapv(x,y,z,w,h)+i].addEntry(mapv(x-1,y,z,w,h)+i,-1);s++;}
   //            if (y>0)   {hessian[mapv(x,y,z,w,h)+i].addEntry(mapv(x,y-1,z,w,h)+i,-1);s++;}
   //            if (z>0)   {hessian[mapv(x,y,z,w,h)+i].addEntry(mapv(x,y,z-1,w,h)+i,-1);s++;}
   //            hessian[mapv(x,y,z,w,h)+i].addEntry(mapv(x,y,z,  w,h)+i, s);
   //         }
   //         gradient[mapv(x,y,z,w,h)] = 0;
   //      }
   //   }
   //}
}

void FDDeformSolverTools::convertVolToVec(const RegularVolume *vol, DVectorF &vec)
{
   const AAT posAAT = vol->getDescr()->getToken("position", 3, VAD::DATA_FORMAT_FLOAT32);
   const card32 w = vol->getWidth();
   const card32 h = vol->getHeight();
   const card32 d = vol->getDepth();
   vec.setDim(w*h*d*3);
   int i = 0;
   for (card32 z=0; z<d; z++) {
      for (card32 y=0; y<h; y++) {
         for (card32 x=0; x<w; x++) {
            Vector3f p = vol->getVoxel<float32, 3>(x,y,z, posAAT);
            for (card32 j=0; j<3; j++) {
               vec[i++] = p[j];
            }
         }
      }
   }
}

void FDDeformSolverTools::convertVecToVol(const DVectorF &vec, RegularVolume *vol)
{
   const AAT posAAT = vol->getDescr()->getToken("position", 3, VAD::DATA_FORMAT_FLOAT32);
   const card32 w = vol->getWidth();
   const card32 h = vol->getHeight();
   const card32 d = vol->getDepth();
   x4Assert(vec.getDim() >= w*h*d);
   int i = 0;
   for (card32 z=0; z<d; z++) {
      for (card32 y=0; y<h; y++) {
         for (card32 x=0; x<w; x++) {
            Vector3f p = makeVector3f(vec[i], vec[i+1], vec[i+2]);
            vol->setVoxel(x,y,z, posAAT, p);
            i += 3;
         }
      }
   }
}
