#include "StdAfx.h"
#include "MMConstraints.h"
#include "MMDeformations.h"
#include "ARSettings.h"

using namespace X4;

MMConstraints::MMConstraints(const size_t numFrames, const size_t myFrameNum) {
	positionConstraintsConstant = 0;
	numNodes=0;
	std::cout << " created constraint for frame " << myFrameNum << " of " << numFrames <<" frames\n" << flush;
}

MMConstraints::MMConstraints(MMConstraints &copy) {
	for (card32 i=0; i<copy.nbPositionConstraints(); i++) {
		addPositionConstraint(new MMPositionConstraint(*copy.getPositionConstraint(i)));
	}
	for (card32 i=0; i<copy.nbHalfSpaceConstraints(); i++) {
		addHalfSpaceConstraint(new MMHalfSpaceConstraint(*copy.getHalfSpaceConstraint(i)));
	}
	positionConstraintsConstant = copy.positionConstraintsConstant;
	positionConstraintsLinear  = copy.positionConstraintsLinear;
	positionConstraintsQuadric = copy.positionConstraintsQuadric;
	fastGetIt = copy.fastGetIt;
}

MMConstraints::~MMConstraints() {
	clearConstraints();
}

void MMConstraints::clearConstraints() {
	for (card32 i=0; i<positionConstraints.size(); i++) {
		delete positionConstraints[i];
	}
	positionConstraints.clear();
	for (card32 i=0; i<halfSpaceConstraints.size(); i++) {
		delete halfSpaceConstraints[i];
	}
	halfSpaceConstraints.clear();

	positionConstraintsConstant = 0;
	positionConstraintsLinear.clear();
	positionConstraintsQuadric.clear();
	fastGetIt.clear();
	numNodes=0;
}

//MMConstraints* MMConstraints::copy() const {
//	MMConstraints * result = new MMConstraints();
//	for (card32 i=0; i<constraints.size(); i++) {
//		result->addConstraint(constraints[i]->copy());
//	}
//	return result;
//}

void MMConstraints::addPositionConstraint(MMPositionConstraint *c) {
	// TODO: this should disappear if everything works
	//positionConstraints.push_back(c);

	const Vector3f &deformation = c->getDeformation();
	const Matrix3f &covariance = c->getCovariance();

	const Vector3f covariance_deformation = covariance*deformation;

	// constant term
	positionConstraintsConstant += deformation*(covariance_deformation);

	
	MMNeighborContainer * neighbors = c->getNeighbors();
	const card32 nbneighbors = neighbors->nbNeighbors();
	//debugOutput << "   nBneighbours: " << nbneighbors << "\n";

	for (card32 i=0; i<nbneighbors; i++) {
		// linear term
		const card32 neighbor_index1 = neighbors->getNeighborIndex(i);
		hash_map< card32, Vector3f >::iterator iter = positionConstraintsLinear.find(neighbor_index1);
		if (iter==positionConstraintsLinear.end()) {
			positionConstraintsLinear[neighbor_index1] = covariance_deformation*(-2*neighbors->getShapeValue(i));
		}
		else {
			iter->second+=covariance_deformation*(-2*neighbors->getShapeValue(i));
		}

// 		// quadratic term
// 		for (card32 j=0; j<nbneighbors; j++) {
// 			const card32 neighbor_index2 = neighbors->getNeighborIndex(j);
// 			pair<card32, card32> link(neighbor_index1, neighbor_index2);
// 			MyMap::iterator iter = positionConstraintsQuadric.find(link);
// 			if (iter==positionConstraintsQuadric.end()) {
// 				positionConstraintsQuadric[link] = covariance*(neighbors->getShapeValue(i)*neighbors->getShapeValue(j));
// 			}
// 			else {
// 				iter->second+=covariance*(neighbors->getShapeValue(i)*neighbors->getShapeValue(j));
// 			}
// 
// 		}
		// quadratic term
		for (card32 j=0; j<nbneighbors; j++) {
			const card32 neighbor_index2 = neighbors->getNeighborIndex(j);
			fastGetIt.createNewOrAddToExistingMatrix(neighbor_index1, neighbor_index2, covariance*(neighbors->getShapeValue(i)*neighbors->getShapeValue(j)));
		}
	}
	delete c;
}

float32 MMConstraints::fCorrespondences(MMDeformations * deformations, float32 tempCorrespondenceParam) {
	// constant term
	float32 result = positionConstraintsConstant;

	// linear terms
	for (hash_map< card32, Vector3f >::const_iterator iter=positionConstraintsLinear.begin(); iter!=positionConstraintsLinear.end(); ++iter) {
		const card32 index = iter->first;
		const Vector3f &deformation = deformations->getDeformation(index);
		result+=iter->second*deformation;
	}

// 	// quadratic terms
// 	for (MyMap::const_iterator iter=positionConstraintsQuadric.begin(); iter!=positionConstraintsQuadric.end(); ++iter) {
// 		const card32 index1 = iter->first.first;
// 		const card32 index2 = iter->first.second;
// 		const Vector3f &deformation1 = deformations->getDeformation(index1);
// 		const Vector3f &deformation2 = deformations->getDeformation(index2);
// 		result+=deformation1*(iter->second*deformation2);
// 	}

	
		// quadratic terms FastGetABIterator
	    FastGetABIterator iter(fastGetIt);
		for (iter.begin(); !iter.atEnd(); iter.next()) {
			const mpcard index1 = iter.getA();
			const mpcard index2 = iter.getB();
			const Vector3f &deformation1 = deformations->getDeformation(index1);
			const Vector3f &deformation2 = deformations->getDeformation(index2);
			result+=deformation1*(iter.getMatrix()*deformation2);
		}

	// final result
	return tempCorrespondenceParam*result;
}

void MMConstraints::gCorrespondences(std::vector<Vector3f> &gradient, MMDeformations * deformations, float32 tempCorrespondenceParam) {
	// linear terms
	for (hash_map< card32, Vector3f >::const_iterator iter=positionConstraintsLinear.begin(); iter!=positionConstraintsLinear.end(); ++iter) {
		const card32 index = iter->first;
		if (!deformations->isActive(index)) continue;
		const card32 realIndex = deformations->getRealIndex(index);
		gradient[realIndex]+=iter->second*tempCorrespondenceParam;
	}

	// quadratic terms
// 	for (MyMap::const_iterator iter=positionConstraintsQuadric.begin(); iter!=positionConstraintsQuadric.end(); ++iter) {
// 		const card32 index1 = iter->first.first;
// 		const card32 index2 = iter->first.second;
// 		const Vector3f &deformation1 = deformations->getDeformation(index1);
// 		const Vector3f &deformation2 = deformations->getDeformation(index2);
// 		
// 		const card32 realIndex1 = deformations->getRealIndex(index1);
// 		const card32 realIndex2 = deformations->getRealIndex(index2);
// 		
// 		gradient[realIndex1]+=(iter->second*deformation2)*tempCorrespondenceParam;
// 		gradient[realIndex2]+=(iter->second*deformation1)*tempCorrespondenceParam;
// 	}

	// quadratic terms
	FastGetABIterator iter(fastGetIt);
	for (iter.begin(); !iter.atEnd(); iter.next()) {
		const mpcard index1 = iter.getA();
		const mpcard index2 = iter.getB();
		const Vector3f &deformation1 = deformations->getDeformation(index1);
		const Vector3f &deformation2 = deformations->getDeformation(index2);

		const card32 realIndex1 = deformations->getRealIndex(index1);
		const card32 realIndex2 = deformations->getRealIndex(index2);

		gradient[realIndex1]+=(iter.getMatrix()*deformation2)*tempCorrespondenceParam;
		gradient[realIndex2]+=(iter.getMatrix()*deformation1)*tempCorrespondenceParam;
	}
}

void X4::MMConstraints::printMemUsage()
{
	//debugOutput << "vec size: " << positionConstraints.size() << " : " << halfSpaceConstraints.size()<< "\n";
	debugOutput << "positionConstraintsQuadric: " << (positionConstraintsQuadric.size() * (2*sizeof(size_t) + sizeof(Matrix3f))) /1024
		<< " kB  positionConstraintsLinear: " << positionConstraintsLinear.size() *(sizeof(card32) + sizeof(Vector3f))/1024
		<< " kB\n";
//typedef hash_map< pair< size_t, size_t >, X4::Matrix3f , MyHashTraits > MyMap; 
//hash_map< card32, Vector3f > positionConstraintsLinear;
	positionConstraintsQuadric.clear();
	
}

void X4::MMConstraints::setNumNodes( size_t num )
{
	if (numNodes!=num)
	{
		fastGetIt.resize(num);
		numNodes=num;
	}
	
}
