//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "UnstructuredInCorePointCloud.h"
#include "PointSetSphericalQuery.h"
#include "PCCNodesTools.h"
#include "VariableArrayClassProperty.h"
#include "DeformationTools.h"
#include "MMTopology.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------
#include "MMNodes.h"
//---------------------------------------------------------------------------


IMPLEMENT_X4_CLASS( MMNodes , 0 ) {
	BEGIN_CLASS_INIT( MMNodes );
	ADD_VAR_ARRAY_PROP( nodes, 0, new ObjectClassProperty("", 0, NULL, MMNode::getClass(), true, NULL) )
	ADD_FLOAT32_PROP( maxRadius, 0 )
	ADD_FLOAT32_PROP( averageRadius, 0)
	ADD_OBJECT_PROP(topology, 0, MMTopology::getClass(), true);
}

MMNodes::MMNodes() {
   octree = 0;
   pointcloud = 0;
   queryPS = 0;   
   query = 0;
   topology = 0;
}

MMNodes::MMNodes(UnstructuredInCorePointCloud *pc, float32 radius) : pointcloud(pc) {
   octree = 0;
   queryPS = 0;   
   query = 0;
   topology = 0;
   AAT posAAT = pc->getAAT("position");
	PointSet * ps = pc->getPointSet();
	maxRadius = radius;
	averageRadius = radius;
	for (card32 i=0; i<pc->getNumPoints(); i++) {
		Vector3f pos = ps->get3f(i, posAAT);
		addNode(new MMNode(pos, radius));
	}
}

MMNodes::MMNodes(std::vector<Vector3f> &positions, float32 radius) {
   octree = 0;
   pointcloud = 0;
   queryPS = 0;   
   query = 0;
   topology = 0;
   maxRadius = radius;
   averageRadius = radius;
	for (card32 i=0; i<positions.size(); i++) {	
		//std::cout << "Adding node: " << positions[i][0] << " " << positions[i][1] << " " << positions[i][2] << " " << endl;
		addNode(new MMNode(positions[i], radius));
	}
}

MMNodes::MMNodes(std::vector<Vector3f> &positions, std::vector<card32> &baseIndices,  float32 radius) {
   octree = 0;
   pointcloud = 0;
   queryPS = 0;   
   query = 0;
   topology = 0;
   maxRadius = radius;
   averageRadius = radius;
	for (card32 i=0; i<positions.size(); i++) {	
		//std::cout << "Adding node: " << positions[i][0] << " " << positions[i][1] << " " << positions[i][2] << " " << endl;
		addNode(new MMNode(positions[i], baseIndices[i], radius));
	}
}

MMNodes::MMNodes(std::vector<Vector3f> &positions, std::vector<card32> &baseIndices, std::vector<float32> & radii) {
   octree = 0;
   pointcloud = 0;
   queryPS = 0;   
   query = 0;
   topology = 0;
   maxRadius = 0;
   averageRadius = 0;
	for (card32 i=0; i<positions.size(); i++) {	
		//std::cout << "Adding node: " << positions[i][0] << " " << positions[i][1] << " " << positions[i][2] << " " << endl;
		if (radii[i] > maxRadius) maxRadius = radii[i];
		averageRadius+=radii[i];
		addNode(new MMNode(positions[i], baseIndices[i], radii[i]));
	}
	averageRadius/=float32(positions.size());
}

MMNodes::MMNodes(const MMNodes &other) {
   const size_t numNodes = other.nodes.size();
	nodes.resize(numNodes);
	for (unsigned i=0; i<numNodes; i++) {
	   nodes[i] = new MMNode(*(other.nodes[i]));
	}
	if (other.pointcloud) {
	   pointcloud = (UnstructuredInCorePointCloud*)other.pointcloud->copy();
	} else {
	   pointcloud = NULL;
	}
	maxRadius = other.maxRadius;
	averageRadius = other.averageRadius;
	if (other.octree != NULL) {
      octree = new AttachedIndexedOctree();
   	octree->assign(other.octree);
   } else {
      octree = NULL;
   }
   if (other.queryPS != NULL) {
      queryPS = (PointSet*)(other.queryPS->copy());
	   query = new PointSetSphericalQuery(queryPS, octree, Vector3f(), maxRadius);
	} else {
	   queryPS = NULL;
	   query = NULL;
	}

   if (other.topology!=NULL) {
	   topology = new MMTopology(*other.topology);
   }
   else {
	   topology = 0;
   }
}


MMNodes::~MMNodes() {
	delete pointcloud; pointcloud = 0;
	delete octree; octree = 0;
	delete query; query = 0;
	delete queryPS; queryPS = 0;
	for (card32 i=0; i<nodes.size(); i++) {
		delete nodes[i];
	}
	nodes.clear();
	delete topology; topology = 0;
}

void MMNodes::initializeNodes() {
	if (nodes.empty() || nodes[0]->nbNeighbors()>0) return;
	// Compute node neighbors
	std::cout << "Computing neighbors for nodes" << endl;
	computeNodeNeighbors();
	// Now for each node, compute the shape functions (normal and gradient)
	std::cout << "Evaluate nodal shape functions" << endl;
	evaluateNodeShapeFunctions();
	std::cout << "Initializing nodes done" << endl;
}

void MMNodes::createOctree() {
	delete queryPS;
    queryPS = new PointSet();
	VertexDescriptor vd;
	vd.pushAttrib(mVAD("position", 3, VAD::DATA_FORMAT_FLOAT32));
	queryPS->clearAndSetup(1,0,&vd);
	for (card32 i=0; i<nodes.size(); i++) {
		queryPS->addPoint((const char*)&(nodes[i]->getPos()));
	}
	octree = new AttachedIndexedOctree();
	octree->setup("SimpleOctree", AttachedData::ADF_DELETEONUPDATE,  NULL);
	octree->octreesetup(queryPS, 20);


	// TODO: we should keep this, for the correspondences.
	query = new PointSetSphericalQuery(queryPS, octree, Vector3f(), maxRadius);
}

void MMNodes::computeNodeNeighbors() {

	// total number of nodes
	const card32 nbnodes = nbNodes();

	// First compute the maximal possible radius.
	// This radius will be used for range querying (we have to
	// be conservative when nodes have varying radii.
	maxRadius = 0;
	averageRadius = 0;
	for (card32 i=0; i<nbnodes; i++) {
		const float32 radius = nodes[i]->getRadius();
		//std::cout << "Radius of point " << i << " : " << radius << endl;
		if (radius>maxRadius) maxRadius = radius;
		averageRadius+=radius;
	}
	averageRadius/=float32(nbnodes);
	//int bla; cin >> bla;
	
	// octree for neighborhood queries
    if (!query) createOctree();

	// For each node, create the NodeInformation, query for the
	// neighbors and store it in there
	
	if (topology) {		
		for (card32 i=0; i<nbnodes; i++) {
			MMNode * node = nodes[i];
			const Vector3f &pos = node->getPos();
			const card32 baseIndex = node->getBaseIndex();		
			if (baseIndex!=-1) topology->getReachableIndices(baseIndex, pos, 2*maxRadius, node->getReachableBaseIndices());
			else topology->getReachableIndices(pos, 2*maxRadius, node->getReachableBaseIndices());
		}
	}


	for (card32 i=0; i<nbnodes; i++) {		
		computeNeighbors(i);
		
		//std::cout << "Number of neighbors : " << node->nbNeighbors() << endl;
		// TODO: normalize weights so that they sum to 1. Might be more stable for later computations.
	}
	//std::cout << "Using radius: " << maxRadius << endl;
	/*for (card32 i=0; i<nbnodes; i++) {
		NodeInformation * ni = new NodeInformation();
		ani->addNodeInformation(ni);
		const Vector3f posi = ps->get3f(i, positionAAT);
		for (card32 j=0; j<nbnodes; j++) {
			const Vector3f posj = ps->get3f(j, positionAAT);
			const float32 radius = ps->get1f(j, radiusAAT);				
			const float32 neighbor_sqr_dist = (posj-posi).getSqrNorm();
			//std::cout << "Testing: " << radius << " > " << sqrt(neighbor_sqr_dist) << "?" << endl;
			if (radius*radius>=neighbor_sqr_dist) {
				const float32 neighbor_dist = sqrt(neighbor_sqr_dist);				
				float32 weight = 1.0f - neighbor_sqr_dist/(radius*radius); weight = weight*weight*weight;
				ni->addNeighbor(j, neighbor_dist, weight);
			}		
		}
		//std::cout << "Number of neighbors : " << ni->nbNeighbors() << endl;
		// TODO: normalize weights so that they sum to 1. Might be more stable for later computations.
	}*/
	//std::cout << "Using radius: " << maxRadius << endl;
}

// TODO: CHECK TOPOLOGY!
void MMNodes::computeNeighbors(MMNeighborContainer * node) {
	const Vector3f &pos = node->getPos();
	const card32 baseIndex = node->getBaseIndex();

	/*std::set<card32> reachableBaseIndices;
	if (topology) {
		if (baseIndex!=-1) topology->getReachableIndices(baseIndex, 2*maxRadius, reachableBaseIndices);
		else topology->getReachableIndices(pos, 2*maxRadius, reachableBaseIndices);
	}*/

	if (!query) createOctree();
	query->setTargetSphere(pos, maxRadius);
	
	//set<card32> *euclNeighbors = new set<card32>();
	//set<card32> verified;
	
	//PCCNodesTools *nd = new PCCNodesTools();
	
	bool found;
	card32 neighbor_index; // neighbor index
	query->getNextPointIndex(neighbor_index, found);
	while (found) {	
		MMNode * neighbor = nodes[neighbor_index];
		//const card32 neighborBaseIndex = neighbor->getBaseIndex();
		const Vector3f &neighbor_pos = neighbor->getPos();
		//std::cout << neighbor_pos[0] <<"  " << neighbor_pos[1] << " " << neighbor_pos[2] << endl;
		const float32 neighbor_sqr_dist = (pos-neighbor_pos).getSqrNorm(); 			
		const float32 neighbor_radius = neighbor->getRadius(); // radius of neighbor
		if (topology) {
			if (!topology->isNeighbor(baseIndex, neighbor->getBaseIndex(), neighbor->getPos(), 2*neighbor_radius)) continue;
			/*std::set<card32>::const_iterator iter;
			iter = neighbor->getReachableBaseIndices().find(baseIndex);
			if (iter==neighbor->getReachableBaseIndices().end()) {
				query->getNextPointIndex(neighbor_index, found);
				continue;
			}*/
			/*std::set<int> intersection;		
			set_intersection(reachableBaseIndices[index].begin(), reachableBaseIndices[index].end(),
				reachableBaseIndices[neighbor_index].begin(), reachableBaseIndices[neighbor_index].end(),
							inserter(intersection, intersection.begin()));
			if (intersection.empty()) {
				query->getNextPointIndex(neighbor_index, found);
				continue;
			}*/

		}			
		if (neighbor_radius*neighbor_radius>=neighbor_sqr_dist) {

			const float32 neighbor_dist = sqrt(neighbor_sqr_dist);				
			float32 weight = 1.0f - neighbor_sqr_dist/(neighbor_radius*neighbor_radius); weight = weight*weight*weight;
			Vector3f gradientWeight = (pos-neighbor_pos)*3*(1-neighbor_sqr_dist/(neighbor_radius*neighbor_radius))*(1-neighbor_sqr_dist/(neighbor_radius*neighbor_radius))*(-2.0/(neighbor_radius*neighbor_radius));
			node->addNeighbor(neighbor_index, neighbor_dist, weight, gradientWeight);

			//euclNeighbors->insert(neighbor_index);
		}
		query->getNextPointIndex(neighbor_index, found);
	}
	
	/*nd->verify(pointcloud, node->getPos(), 2*maxRadius, euclNeighbors, &verified);
	
	for(set<card32>::iterator it=verified.begin(); it!= verified.end(); it++){
		MMNode * neighbor = nodes[*it];
		card32 neighbor_index = *it;
		const Vector3f &neighbor_pos = neighbor->getPos();
		//std::cout << neighbor_pos[0] <<"  " << neighbor_pos[1] << " " << neighbor_pos[2] << endl;
		const float32 neighbor_sqr_dist = (pos-neighbor_pos).getSqrNorm(); 			
		const float32 neighbor_radius = neighbor->getRadius(); // radius of neighbor

		const float32 neighbor_dist = sqrt(neighbor_sqr_dist);				
		float32 weight = 1.0f - neighbor_sqr_dist/(neighbor_radius*neighbor_radius); weight = weight*weight*weight;
		
		node->addNeighbor(neighbor_index, neighbor_dist, weight);
	}
	

	delete euclNeighbors;
	delete nd;*/
	
	
}




// TODO: CHECK TOPOLOGY!
void MMNodes::computeNeighbors(card32 index) {
	MMNode *node = nodes[index];
	const Vector3f &pos = node->getPos();
	/*const card32 baseIndex = node->getBaseIndex();

	std::set<card32> reachableBaseIndices;
	if (topology) {
		if (baseIndex!=-1) topology->getReachableIndices(baseIndex, 2*maxRadius, reachableBaseIndices);
		else topology->getReachableIndices(pos, 2*maxRadius, reachableBaseIndices);
	}*/

	if (!query) createOctree();
	query->setTargetSphere(pos, maxRadius);
	
	//set<card32> *euclNeighbors = new set<card32>();
	//set<card32> verified;
	
	//PCCNodesTools *nd = new PCCNodesTools();
	
	bool found;
	card32 neighbor_index; // neighbor index
	query->getNextPointIndex(neighbor_index, found);
	while (found) {	
		MMNode * neighbor = nodes[neighbor_index];
		//const card32 neighborBaseIndex = neighbor->getBaseIndex();
		const Vector3f &neighbor_pos = neighbor->getPos();
		//std::cout << neighbor_pos[0] <<"  " << neighbor_pos[1] << " " << neighbor_pos[2] << endl;
		const float32 neighbor_sqr_dist = (pos-neighbor_pos).getSqrNorm(); 			
		const float32 neighbor_radius = neighbor->getRadius(); // radius of neighbor
		if (topology) {
			/*std::set<card32>::const_iterator iter;
			iter = reachableBaseIndices.find(neighborBaseIndex);
			if (iter==reachableBaseIndices.end()) {
				query->getNextPointIndex(neighbor_index, found);
				continue;
			}*/
			std::set<int> intersection;		
			set_intersection(node->getReachableBaseIndices().begin(), node->getReachableBaseIndices().end(),
							 neighbor->getReachableBaseIndices().begin(), neighbor->getReachableBaseIndices().end(),
							 inserter(intersection, intersection.begin()));
			if (intersection.empty()) {
				query->getNextPointIndex(neighbor_index, found);
				continue;
			}

		}			
		if (neighbor_radius*neighbor_radius>=neighbor_sqr_dist) {

			const float32 neighbor_dist = sqrt(neighbor_sqr_dist);				
			float32 weight = 1.0f - neighbor_sqr_dist/(neighbor_radius*neighbor_radius); weight = weight*weight*weight;
			Vector3f gradientWeight = (pos-neighbor_pos)*3*(1-neighbor_sqr_dist/(neighbor_radius*neighbor_radius))*(1-neighbor_sqr_dist/(neighbor_radius*neighbor_radius))*(-2.0/(neighbor_radius*neighbor_radius));
			node->addNeighbor(neighbor_index, neighbor_dist, weight, gradientWeight);

			//euclNeighbors->insert(neighbor_index);
		}
		query->getNextPointIndex(neighbor_index, found);
	}
	
	/*nd->verify(pointcloud, node->getPos(), 2*maxRadius, euclNeighbors, &verified);
	
	for(set<card32>::iterator it=verified.begin(); it!= verified.end(); it++){
		MMNode * neighbor = nodes[*it];
		card32 neighbor_index = *it;
		const Vector3f &neighbor_pos = neighbor->getPos();
		//std::cout << neighbor_pos[0] <<"  " << neighbor_pos[1] << " " << neighbor_pos[2] << endl;
		const float32 neighbor_sqr_dist = (pos-neighbor_pos).getSqrNorm(); 			
		const float32 neighbor_radius = neighbor->getRadius(); // radius of neighbor

		const float32 neighbor_dist = sqrt(neighbor_sqr_dist);				
		float32 weight = 1.0f - neighbor_sqr_dist/(neighbor_radius*neighbor_radius); weight = weight*weight*weight;
		
		node->addNeighbor(neighbor_index, neighbor_dist, weight);
	}
	

	delete euclNeighbors;
	delete nd;*/
	
	
}

void MMNodes::evaluateNodeShapeFunctions() {
	// total number of nodes
	const card32 nbnodes = nbNodes();

	for (card32 i = 0; i<nbnodes; i++) {
		//std::cout << "Processing : " << i << " " << nbnodes << endl;
		bool success1, success2;
		MMNode * node = nodes[i];
		// normal shape functions
		evaluateNodeShapeFunctions(node, success1);	
		// gradient shape functions
		evaluateNodeGradientShapeFunctions(node, success2);	
		if ((!success1) || (!success2)) {
			std::cout << "Shape function evaluation failed: " << success1 << " " << success2 << endl;
		}
	}
}

void MMNodes::evaluateNodeShapeFunctions(MMNeighborContainer *node, bool& success) {
	//if (node->getShapeValuesInitialized()) return;
	const Vector3f &x = node->getPos();	
	const card32 nbneighbors = node->nbNeighbors();
	// First compute the moment matrix (it's a 4x4 matrix)
	Matrix4d M; M.setZero();
	Matrix4d Mx; Mx.setZero();
	Matrix4d My; My.setZero();
	Matrix4d Mz; Mz.setZero();
	//std::cout << nbneighbors << endl;
	for (card32 i=0; i<nbneighbors; i++) {
		const float32 w = node->getWeight(i);
		const Vector3f &gw = node->getGradientWeight(i);
		const float32 wx = gw[0];
		const float32 wy = gw[1];
		const float32 wz = gw[2];

		const card32 index = node->getNeighborIndex(i);
		MMNode * neighbor = nodes[index];
		const Vector3f &pos = neighbor->getPos();
		const float32 radius = neighbor->getRadius();
		const float32 inv_radius = 1.0f/averageRadius;

		// We translate and scale. This improves the condition number
		// of the moment matrix (see for example Fries and Matthies).
		const Vector3f diff = (pos - x)*inv_radius;
		//const Vector3f diff = pos;

		// Rugular M
		M[0][0] += w*diff[2]*diff[2];
		M[0][1] += w*diff[1]*diff[2];
		M[0][2] += w*diff[0]*diff[2];
		M[0][3] += w*diff[2];

		//M[1][0] += w*diff[2]*diff[1];
		M[1][1] += w*diff[1]*diff[1];
		M[1][2] += w*diff[0]*diff[1];
		M[1][3] += w*diff[1];

		//M[2][0] += w*diff[2]*diff[0];
		//M[2][1] += w*diff[1]*diff[0];
		M[2][2] += w*diff[0]*diff[0];
		M[2][3] += w*diff[0];

		//M[3][0] += w*diff[2];
		//M[3][1] += w*diff[1];
		//M[3][2] += w*diff[0];
		M[3][3] += w;

		// dM/dx
		Mx[0][0] += wx*diff[2]*diff[2];
		Mx[0][1] += wx*diff[1]*diff[2];
		Mx[0][2] += wx*diff[0]*diff[2]-w*inv_radius*diff[2];
		Mx[0][3] += wx*diff[2];
		
		Mx[1][1] += wx*diff[1]*diff[1];
		Mx[1][2] += wx*diff[0]*diff[1]-w*inv_radius*diff[1];
		Mx[1][3] += wx*diff[1];

		Mx[2][2] += wx*diff[0]*diff[0]-2*w*inv_radius*diff[0];
		Mx[2][3] += wx*diff[0]-w*inv_radius;

		Mx[3][3] += wx;

		// dM/dy
		My[0][0] += wy*diff[2]*diff[2];
		My[0][1] += wy*diff[1]*diff[2]-w*inv_radius*diff[2];
		My[0][2] += wy*diff[0]*diff[2];
		My[0][3] += wy*diff[2];
		
		My[1][1] += wy*diff[1]*diff[1]-2*w*inv_radius*diff[1];
		My[1][2] += wy*diff[0]*diff[1]-w*inv_radius*diff[0];
		My[1][3] += wy*diff[1]-w*inv_radius;

		My[2][2] += wy*diff[0]*diff[0];
		My[2][3] += wy*diff[0];

		My[3][3] += wy;

		// dM/dz
		Mz[0][0] += wz*diff[2]*diff[2]-2*w*inv_radius*diff[2];
		Mz[0][1] += wz*diff[1]*diff[2]-w*inv_radius*diff[1];
		Mz[0][2] += wz*diff[0]*diff[2]-w*inv_radius*diff[0];
		Mz[0][3] += wz*diff[2]-w*inv_radius;
		
		Mz[1][1] += wz*diff[1]*diff[1];
		Mz[1][2] += wz*diff[0]*diff[1];
		Mz[1][3] += wz*diff[1];

		Mz[2][2] += wz*diff[0]*diff[0];
		Mz[2][3] += wz*diff[0];

		Mz[3][3] += wz;
	}
	for (card32 i=1; i<=3; i++) {
		for (card32 j=0; j<i; j++) {
			M[i][j] = M[j][i];
			Mx[i][j] = Mx[j][i];
			My[i][j] = My[j][i];
			Mz[i][j] = Mz[j][i];
		}
	}
	
	Matrix4d invM;
	try {
		//invM = invertMatrix(M);
		invM = svdPseudoInvert4d(M, 1e-5);
	}
	catch(...) {
		success = false;
		return;
	}

	// TODO: optimize, we only need the last rows of these matrices
	Matrix4d invMx = -invM*Mx*invM;
	Matrix4d invMy = -invM*My*invM;
	Matrix4d invMz = -invM*Mz*invM;


	// Now, finally compute for each neighbor the shape function value.
	for (card32 i=0; i<nbneighbors; i++) {
		const float32 w = node->getWeight(i);
		const Vector3f &gw = node->getGradientWeight(i);
		const float32 wx = gw[0];
		const float32 wy = gw[1];
		const float32 wz = gw[2];
		const card32 index = node->getNeighborIndex(i);
		MMNode * neighbor = nodes[index];
		const Vector3f &pos = neighbor->getPos();
		const float32 radius = neighbor->getRadius();
		const float32 inv_radius = 1.0f/averageRadius;
		const Vector3f diff = (pos - x)*inv_radius;
		//const Vector3f diff = pos;

		// Difficult to follow what I'm doing here.
		// See Fries and Matthies, for the optimized shape function
		// evaluation (basically we translate and scale to improve
		// the conditioning number to compute invM.
		float32 result = invM[3][3]*1;
		for (card32 k=0; k<3; k++) result+= invM[3][k]*diff[2-k];		
		result*=w;
		node->setShapeValue(i, result);

		float32 resultx_1 = invMx[3][3]*1;
		for (card32 k=0; k<3; k++) resultx_1+= invMx[3][k]*diff[2-k];		
		resultx_1*=w;

		float32 resultx_2 = invM[3][3]*1;
		for (card32 k=0; k<3; k++) resultx_2+= invM[3][k]*diff[2-k];		
		resultx_2*=wx;

		float32 resultx_3 = invM[3][2]*(-inv_radius);
		resultx_3*=w;

		float32 resultx = resultx_1+resultx_2+resultx_3;


		float32 resulty_1 = invMy[3][3]*1;
		for (card32 k=0; k<3; k++) resulty_1+= invMy[3][k]*diff[2-k];		
		resulty_1*=w;

		float32 resulty_2 = invM[3][3]*1;
		for (card32 k=0; k<3; k++) resulty_2+= invM[3][k]*diff[2-k];		
		resulty_2*=wy;

		float32 resulty_3 = invM[3][1]*(-inv_radius);
		resulty_3*=w;

		float32 resulty = resulty_1+resulty_2+resulty_3;

		float32 resultz_1 = invMz[3][3]*1;
		for (card32 k=0; k<3; k++) resultz_1+= invMz[3][k]*diff[2-k];		
		resultz_1*=w;

		float32 resultz_2 = invM[3][3]*1;
		for (card32 k=0; k<3; k++) resultz_2+= invM[3][k]*diff[2-k];		
		resultz_2*=wz;

		float32 resultz_3 = invM[3][0]*(-inv_radius);
		resultz_3*=w;

		float32 resultz = resultz_1+resultz_2+resultz_3;

		node->setGradientShapeValue2(i, makeVector3f(resultx, resulty, resultz));
	}
	// Check if shapevalues sum up to 1.
	float32 test = 0;
	Vector3f testg; testg.setZero();
	for (card32 i=0; i<nbneighbors; i++) {
		test+=node->getShapeValue(i);
		//std::cout << node->getShapeValue(i) << " ";
		testg+=node->getGradientShapeValues2(i);
	}
	//std::cout << endl;
	//std::cout << testg[0] << " " << testg[1] << " " << testg[2] << endl;
	if (fabs(test-1.0f)>0.3) {
		std::cout << "Error: Shape functions do not sum to 1, but to: " << test << endl;
		success = false;
		return;
	}
	success = true;
	node->setShapeValuesInitialized();
}

void MMNodes::evaluateNodeGradientShapeFunctions(MMNeighborContainer * node, bool &success) {
	const Vector3f &x = node->getPos();

	const card32 nbneighbors = node->nbNeighbors();

	// First compute the moment matrix (it's a 3x3 matrix for the gradient)
	Matrix3d M; M.setZero();
	for (card32 i=0; i<nbneighbors; i++) {
		const float32 w = node->getWeight(i);
		const card32 index = node->getNeighborIndex(i);
		MMNode * neighbor = nodes[index];
		const Vector3f &pos = neighbor->getPos();		

		// We translate and scale. This improves the condition number
		// of the moment matrix (see for example Fries and Matthies).
		const Vector3f diff = (pos - x);

		M[0][0] += w*diff[0]*diff[0];
		M[0][1] += w*diff[0]*diff[1];
		M[0][2] += w*diff[0]*diff[2];

		M[1][0] += w*diff[1]*diff[0];
		M[1][1] += w*diff[1]*diff[1];
		M[1][2] += w*diff[1]*diff[2];

		M[2][0] += w*diff[2]*diff[0];
		M[2][1] += w*diff[2]*diff[1];
		M[2][2] += w*diff[2]*diff[2];				
	}
	// TODO: use SVD or gauss for safe/faster inversion
	Matrix3d invM;
	try {
		//invM = invertMatrix(M);
		invM = svdPseudoInvert3d(M, 1e-5);
	}
	catch (X4Exception e) {
		success = false;
		return;
	}

	// Now, finally compute for each neighbor the shape function value.
	for (card32 i=0; i<nbneighbors; i++) {
		const float32 w = node->getWeight(i);
		const card32 index = node->getNeighborIndex(i);
		MMNode * neighbor = nodes[index];
		const Vector3f &pos = neighbor->getPos();

		const Vector3f diff = (pos - x);
		Vector3f result; result.setZero();
		// compute invM*diff
		// TODO: is there a helper method for doing this?
		for (card32 j=0; j<3; j++) {
			for (card32 k=0; k<3; k++) {
				result[j] += invM[j][k]*diff[k];
			}
		}
		result*=w;
		node->setGradientShapeValue(i, result);
	}
	success = true;
	node->setGradientShapeValuesInitialized();

	
		/*const Vector3f &pos = node->getPos();

		// Exact
		float32 dfdx = 2*pos[0]+pos[1]-4*pos[2]+1;
		float32 dfdy = pos[0]+1;
		float32 dfdz = -4*pos[0];
		float32 fi = pos[0]*pos[0] + pos[0]*pos[1] - 4*pos[0]*pos[2] + pos[0] + pos[1];
		// Approximationg
		Vector3f df; df.setZero();
		for (card32 k=0; k<node->nbNeighbors(); k++) {
			const Vector3f &posj = getNode(node->getNeighborIndex(k))->getPos();
			float32 fj = posj[0]*posj[0] + posj[0]*posj[1] - 4*posj[0]*posj[2] + posj[0] + posj[1];
			df+=node->getGradientShapeValues(k)*(fj-fi);
		}
		// These should be close together
		std::cout << dfdx << " " << df[0] << endl;
		std::cout << dfdy << " " << df[1] << endl;
		std::cout << dfdz << " " << df[2] << endl;*/
	
}

