//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "MMTopologyGridSampler.h"
#include "UnstructuredInCorePointCloud.h"
#include "ARSettings.h"
#include "MMNodes.h"
#include "SparseGrid.h"
#include "SparseGrid.inline.h"
#include "TopologyRangeSearch.h"
#include "TopologyAttachment.h"
#include "MMTopology.h"
#include <cmath>
#include <cfloat>
#include "DebugOutput.h"
#include "Timer.h"
#include <cmath>
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

IMPLEMENT_X4_CLASS( MMTopologyGridSampler , 0 ) {
	BEGIN_CLASS_INIT( MMTopologyGridSampler );
}



MMNodes* MMTopologyGridSampler::createNodes(const ARSettings *settings, UnstructuredInCorePointCloud * geometry) {



	Timer T;
	if (settings->mmDebugOutput) {
   	debugOutput << "   topology Grid Sampler ... " << "\n";
	   debugOutput.flush();
	}

	bool hierarchical = false;
	AAT rigidityWeightAAT;
	if (geometry->providesAttribute("rigidityWeight")) {
		hierarchical = true;
		rigidityWeightAAT = geometry->getAAT("rigidityWeight");
		debugOutput << "Using rigidity weight channel \n";
	}


	PointSet * ps = geometry->getPointSet();
	AttachedDataContainer *attachments = geometry->getAttachments();
	TopologyAttachment *topAtt = dynamic_cast<TopologyAttachment*>(attachments->getData("TopologyAtt"));
	if (topAtt == NULL) {
		error("Need a topology attachement to the pointCloud");
		return NULL;
	}
	InCorePCTopologyGraph*	topGraph  = topAtt->getTopology();

	BoundingBox3f bb = ps->getBoundingBox();

	Vector3f m = bb.lowerCorner;
	Vector3f M = bb.upperCorner;

	const uint dimension = settings->gridSize;

	float32 maxLength = bb.getMaxSideLength();
	float32 scale = 1.0/maxLength;
	
	// to avoid geometry coinciding with the cell boundaries
	// gives nicer samplings for planes etc.
	//m = m - makeVector3f(cellsize/2, cellsize/2, cellsize/2);

	// Conversion from 3D positions to grid coordinates
	// x <- (x-m)*scale; this is in [0,1]
	// x <- x*(dimension-1)+1; this is in [1, d]
	// x <- round(x); this is to its nearest cell

	SparseGrid<3, GridEntryType, Object> grid;
	std::vector<Vector3f> positions;
	std::vector<float32> radii;
	std::vector<card32> baseIndices;
	

	const card32 numPoints = ps->getNumEntries();

	tempGeometryNeighbors.resize(numPoints);

	AAT posAAT = ps->getAAT("position");
	for (card32 i=0; i<numPoints; i++) {
		//std::cout << i << " " << numPoints << endl;
		const Vector3f &p = ps->get3f(i, posAAT);
		float32 scaleDimension = 1.0f;
		float32 scaleCellSize = 1.0f;
		if (hierarchical) {
			float32 rigidityWeight = ps->get1f(i, rigidityWeightAAT);
			
			/*
			if(fabs(rigidityWeight-1)>0.001)
				rigidityWeight = 0.5f;*/
				
			scaleDimension = pow(2, floor(log(1.0/rigidityWeight)/log(2.0)));
			//scaleCellSize = 1.0f/scaleDimension;
			//if (rigidityWeight<1) std::cout << "rigidityWeight: " << rigidityWeight << " scaleDimension: " << scaleDimension << "    scaleCellSize " << scaleCellSize << endl;
			//std::cout << "rigidityWeight: " << rigidityWeight << " scaleDimension: " << scaleDimension << "    scaleCellSize " << scaleCellSize << endl;
		}
		float32 cellsize = maxLength/float32(scaleDimension*dimension-1);

		//scaleDimension = 2;
		//scaleCellSize = 0.5;
		
		Vector3f index = (p-m)*scale;
		index = index*(scaleDimension*dimension-1)+makeVector3f(1,1,1);
		for (card32 k=0; k<3; k++) index[k] = floor(index[k]+0.5);
		
		//Vector3f indexPos = (index-makeVector3f(1,1,1))*(1.0f/float32(scaleDimension*dimension-1));
		//indexPos = indexPos*(1.0/scale)+m;

		Vector3f indexPos = m + makeVector3f((index[0]-1)*cellsize*scaleCellSize, (index[1]-1)*cellsize*scaleCellSize, (index[2]-1)*cellsize*scaleCellSize);


		const int startx = indexPos[0]<p[0]?0:-1;
		const int starty = indexPos[1]<p[1]?0:-1;
		const int startz = indexPos[2]<p[2]?0:-1;
		const int endx = startx+1;
		const int endy = starty+1;
		const int endz = startz+1;
		for (int dx=startx; dx<=endx; dx++) {
			for (int dy=starty; dy<=endy; dy++) {
				for (int dz=startz; dz<=endz; dz++) {					
					const Vector3i cellIndex = makeVector3i(index[0]+dx,index[1]+dy,index[2]+dz);
					const Vector3f cellPos = indexPos+makeVector3f(dx*cellsize*scaleCellSize, dy*cellsize*scaleCellSize, dz*cellsize*scaleCellSize);
					bool alreadyNeighbor = false;

					std::set<card32>::const_iterator iter;
					for (iter=tempGeometryNeighbors[i].ids.begin(); iter!=tempGeometryNeighbors[i].ids.end(); ++iter) {
					//for (card32 k=0; k<tempGeometryNeighbors[i].ids.size(); k++) {
						const Vector3f &nodePos = positions[*iter];
						float32 dist = (cellPos-nodePos).getSqrNorm();
						if (dist<0.0001) {
							alreadyNeighbor = true;
							break;
						}
					}
					//std::cout << tempGeometryNeighbors[i].ids.size() << "  alreadyneighbor: " << alreadyNeighbor << endl;
					//std::cout << "number of nodes: " << positions.size() << endl;
					/*GridEntryType entry;
					const bool alreadyThere = grid.getObject(cellIndex, entry);
					//std::cout << alreadyThere << endl;
					bool alreadyNeighbor = false;
					if (alreadyThere) {
						// check if one of the nodes is already a neighbor from me
						std::set<int> intersection;
						
						set_intersection(entry.ids.begin(), entry.ids.end(), 
										 tempGeometryNeighbors[i].ids.begin(), tempGeometryNeighbors[i].ids.end(),
										 inserter(intersection, intersection.begin()));

						//std::cout << entry.ids.size() << " " << tempGeometryNeighbors[i].ids.size() << " " << intersection.size() << endl;
						alreadyNeighbor = !(intersection.empty());						
						
					}*/
					
					//if (alreadyThere && !alreadyNeighbor) std::cout << "Found a node in my cell which is not a neighbor yet, suspicious, but can happen in case of nearby features" << endl;
					if (!alreadyNeighbor) {
					//if ((!alreadyThere) || (alreadyThere && !(alreadyNeighbor))) {	
						size_t newNodeIndex = positions.size();
						const float32 newNodeRadius = 1.8*cellsize*scaleCellSize;
						const float32 topReach = settings->topReach;
						//entry.ids.insert(newNodeIndex);
						positions.push_back(cellPos);
						radii.push_back(newNodeRadius);
						// i becomes the base index for this node
						baseIndices.push_back(i);
						tempNodeNeighbors.push_back(TempNeighbors());
						//std::cout << "Adding node: " << cellPos[0] << " " << cellPos[1] << " " << cellPos[2] << endl;
						//grid.setObject(cellIndex, entry);

						// now add this node id as a neighbor to all geometry points
						// that can be reached from the original point p, with radius 1.8*cellsize
						TopNeighborhoodSphereSearch search;
						search.setup(topGraph,ps,i,cellPos,topReach*newNodeRadius,posAAT);
						while (!search.atEnd()) {
							card32 index = search.getCurrentIndex();
																					
							const Vector3f &oldBasePoint = ps->get3f(baseIndices.back(), posAAT);
							const Vector3f &newBasePoint = ps->get3f(index, posAAT);
							const Vector3f &nodePosition = positions.back();
							const float32 oldSqrDist = (oldBasePoint-nodePosition).getSqrNorm();
							const float32 newSqrDist = (newBasePoint-nodePosition).getSqrNorm();
							// check if the point is close enough
							if (newSqrDist < newNodeRadius*newNodeRadius) {
								tempGeometryNeighbors[index].ids.insert((card32)newNodeIndex);
								// check if the point is a better base for the node (closer by)
								if (newSqrDist < oldSqrDist) {
									baseIndices[baseIndices.size()-1] = index;
								}
								// now add the other ids as a neighbor of the new node
								std::set<card32> &otherNodes = tempGeometryNeighbors[index].ids;
								std::set<card32>::iterator iter;
								for (iter=otherNodes.begin(); iter!=otherNodes.end(); ++iter) {
									card32 nodeIndex = *iter;
									// check if close enough
									float32 sqrDist = (cellPos-positions[nodeIndex]).getSqrNorm();
									if (sqrDist < radii[newNodeIndex]*radii[newNodeIndex]) tempNodeNeighbors[nodeIndex].ids.insert((card32)newNodeIndex);
									if (sqrDist < radii[nodeIndex]*radii[nodeIndex]) tempNodeNeighbors[newNodeIndex].ids.insert(nodeIndex);
								}
							}							
							//tempNeighbors[index].neighbors.push_back(TempNeighbor());
							search.next();
						}
					}
				}
			}
		}		
	}
	
	/*for (card32 i=0; i<tempGeometryNeighbors.size(); i++) {
		std::cout << "geometry nb neighbors: " << tempGeometryNeighbors[i].ids.size() << endl;
	}
	for (card32 i=0; i<tempNodeNeighbors.size(); i++) {
		std::cout << "node nb neighbors: " << tempNodeNeighbors[i].ids.size() << endl;
	}*/
	
	if (settings->mmDebugOutput) {
	   debugOutput << "      number of nodes: " << positions.size() << "\n";
	   debugOutput.flush();
   }
	MMNodes * nodes= new MMNodes(positions, baseIndices, radii);
	nodes->setTopology(new MMTopology(ps, topGraph));

	// create range query data structure
	// for each node: collect all nodes with Euclidean distance 2.5*cellsize
	//nodes->computeNodeNeighbors();
	// then check if the basepoint of the node can reach the base point of the neighboring node
	// by walking over the topology graph and staying in a sphere of 2*2.5*cellsize
	// only keep those neighboring nodes that can be reached
	/*const card32 nbnodes = nodes->nbNodes();
	for (card32 i=0; i<nbnodes; i++) {
		MMNode * node = nodes->getNode(i);
		std::vector<int> topologicallyCorrectNeighbors;
		const card32 nbneighbors = node->nbNeighbors();
		const card32 myBaseIndex = baseIndices[i];
		const Vector3f &myBasePoint = ps->get3f(myBaseIndex, posAAT);
		std::set<card32> reachableBaseIndices;
		TopNeighborhoodSphereSearch search;
		search.setup(topGraph,ps,myBaseIndex,myBasePoint,2*2.5*cellsize,posAAT);
		while (!search.atEnd()) {
			const card32 index = search.getCurrentIndex();
			reachableBaseIndices.insert(index);
			search.next();
		}
		for (card32 k=0; k<nbneighbors; k++) {
			const card32 neighbor_index = node->getNeighborIndex(k);
			const card32 neighborBaseIndex = baseIndices[neighbor_index];
			std::set<card32>::const_iterator iter;
			iter = reachableBaseIndices.find(neighborBaseIndex);
			if (iter!=reachableBaseIndices.end()) {
				topologicallyCorrectNeighbors.push_back(neighbor_index);
				//std::cout << "OK: " << neighbor_index << endl;
			}
			else {
				//std::cout << "NOTOK: " << neighbor_index << endl;
			}
		}
		node->clearNeighbors();
		for (card32 k=0; k<topologicallyCorrectNeighbors.size(); k++) {
			const card32 neighbor_index = topologicallyCorrectNeighbors[k];
			MMNode * neighbor = nodes->getNode(neighbor_index);
			const float32 neighbor_sqr_dist = (node->getPos()-neighbor->getPos()).getSqrNorm(); 			
			const float32 neighbor_radius = neighbor->getRadius(); // radius of neighbor					
			const float32 neighbor_dist = sqrt(neighbor_sqr_dist);				
			float32 weight = 1.0f - neighbor_sqr_dist/(neighbor_radius*neighbor_radius); weight = weight*weight*weight;		
			node->addNeighbor(neighbor_index, neighbor_dist, weight);			
		}
	}*/

	// FROM HERE: compute node neighbors and evaluate the shape functions
	
	float32 minnb=FLT_MAX;
	float32 maxnb=-FLT_MAX;
	float32 avnb=0;
	for (card32 i=0; i<positions.size(); i++) {
		MMNode * node = nodes->getNode(i);
		std::set<card32>::iterator iter;
		float32 nb = tempNodeNeighbors[i].ids.size();
		if (nb<minnb) minnb=nb;
		if (nb>maxnb) maxnb=nb;
		avnb+=nb;
		for (iter=tempNodeNeighbors[i].ids.begin(); iter!=tempNodeNeighbors[i].ids.end(); ++iter) {
			const card32 neighbor_index = *iter;
			const Vector3f &neighbor_pos = positions[neighbor_index];
			const float32 neighbor_radius = nodes->getNode(neighbor_index)->getRadius();
			const float32 neighbor_sqr_dist = (node->getPos()-neighbor_pos).getSqrNorm();				
			const float32 neighbor_dist = sqrt(neighbor_sqr_dist);
			float32 weight = 1.0f - neighbor_sqr_dist/(neighbor_radius*neighbor_radius); weight = weight*weight*weight;
			Vector3f gradientWeight = (node->getPos()-neighbor_pos)*3*(1-neighbor_sqr_dist/(neighbor_radius*neighbor_radius))*(1-neighbor_sqr_dist/(neighbor_radius*neighbor_radius))*(-2.0/(neighbor_radius*neighbor_radius));
			node->addNeighbor(neighbor_index, neighbor_dist, weight, gradientWeight);
		}
		
	}
	avnb/=(float32(positions.size()));
	if (settings->mmDebugOutput) {
   	debugOutput << "      min av max nodal neighbors: " << minnb << " " << avnb << " " << maxnb<< "\n";
	}
	nodes->evaluateNodeShapeFunctions();

	// also create the MMNeighborContainers
	minnb=FLT_MAX;
	maxnb=-FLT_MAX;
	avnb=0;
	for (card32 i=0; i<numPoints; i++) {
		const Vector3f &p = ps->get3f(i, posAAT);
		MMNeighborContainer * neighbors = new MMNeighborContainer(p);
		std::set<card32>::iterator iter;
		float32 nb = tempGeometryNeighbors[i].ids.size();
		if (nb<minnb) minnb=nb;
		if (nb>maxnb) maxnb=nb;
		avnb+=nb;
		for (iter=tempGeometryNeighbors[i].ids.begin(); iter!=tempGeometryNeighbors[i].ids.end(); ++iter) {
			const card32 neighbor_index = *iter;
			const Vector3f &neighbor_pos = positions[neighbor_index];
			const float32 neighbor_radius = nodes->getNode(neighbor_index)->getRadius();
			const float32 neighbor_sqr_dist = (p-neighbor_pos).getSqrNorm();				
			const float32 neighbor_dist = sqrt(neighbor_sqr_dist);
			float32 weight = 1.0f - neighbor_sqr_dist/(neighbor_radius*neighbor_radius); weight = weight*weight*weight;
			Vector3f gradientWeight = (p-neighbor_pos)*3*(1-neighbor_sqr_dist/(neighbor_radius*neighbor_radius))*(1-neighbor_sqr_dist/(neighbor_radius*neighbor_radius))*(-2.0/(neighbor_radius*neighbor_radius));
			neighbors->addNeighbor(neighbor_index, neighbor_dist, weight, gradientWeight);
		}
		bool success;
		nodes->evaluateNodeShapeFunctions(neighbors, success);
		//assert(success);
		neighborContainers.push_back(neighbors);
		//delete neighbors;
	}
	avnb/=(float32(numPoints));
	
	if (settings->mmDebugOutput) {
	   debugOutput << "      min av max geometry neighbors: " << minnb << " " << avnb << " " << maxnb<< "\n";	
	   debugOutput.flush();
   }
	// FROM HERE: compute geometry neighbors and evaluate shape functions
	if (settings->mmDebugOutput) {
	   debugOutput << "   ... done: (" << T.getDeltaValue() << " ms) \n";
	   debugOutput.flush();
	}
	
	
	return nodes;
}

MMNeighborContainer* MMTopologyGridSampler::getNeighborContainer(card32 id) {
	return neighborContainers[id];
}


MMTopologyGridSampler::~MMTopologyGridSampler() {
	tempGeometryNeighbors.clear();
	tempNodeNeighbors.clear();
}


void MMTopologyGridSampler::initializeCache(stdext::hash_map<mpint,MMNeighborContainer*> &cached_neighbors, bool &haveTopology) {
	// THIS IS SPECIFIC FOR THE TOPOLOGYGRIDSAMPLER
	//-------------------------------------------------------------------
	haveTopology = true;
	const size_t numPoints = neighborContainers.size();
	for (card32 i=0; i<numPoints; i++) {
		cached_neighbors[i] = getNeighborContainer(i);
	}
	//-------------------------------------------------------------------
}
