//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "RigidDeformationSolver.h"
//---------------------------------------------------------------------------
#include "RigidDeformationEvaluator.h"
#include "RigidDeformationContext.h"
#include "ARConstraints.h"
#include <opencv/cxcore.h>

//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------



IMPLEMENT_X4_CLASS( RigidDeformationSolver , 0 ) {
	BEGIN_CLASS_INIT( RigidDeformationSolver );
}

// ----

RigidDeformationSolver::RigidDeformationSolver() {
	context = NULL;
}


RigidDeformationSolver::~RigidDeformationSolver() {}




AROptimizationContext *RigidDeformationSolver::createOptContext() 
{
	return new RigidDeformationContext();
}

void RigidDeformationSolver::setup(AROptimizationContext *context, const ARSettings *settings) 
{
	RigidDeformationContext *c = dynamic_cast<RigidDeformationContext*>(context);
	if (!c) {
		throw ETypeMismatch("RigidDeformationSolver::setup() - wrong type for context.");
	}
	this->context = c;
	forget(settings);
}

DeformationEvaluator *RigidDeformationSolver::createEvaluator() 
{
	return new RigidDeformationEvaluator();
}

void RigidDeformationSolver::beginConstraints() 
{
	constraints.clear();
	constraints.resize(context->getNumTransformations());
}

void RigidDeformationSolver::addPositionConstraint(const PositionARConstraint *c) 
{
	x4Assert(context);
	pair<Vector3f, Vector3f> p;
	p.first = c->getAnchor().getPos();
	p.second = c->getToPoint();
	int32 time = (int32)c->getTime();
	if (time <0 || time>(int32)constraints.size()) {
		throw ERangeCheck("RigidDeformationSolver::addPositionConstraint() - time out of range.");
	}
	constraints[time].push_back(p);
}

void RigidDeformationSolver::addVelocityConstraint(const VelocityVectorARConstraint *c) {}

void RigidDeformationSolver::addInstantaneousVelocityConstraint(const InstantaneousVelocityARConstraint *c) {}

void RigidDeformationSolver::solve()
{
	x4Assert(context->getNumTransformations() == constraints.size());
	for (unsigned i=0; i<constraints.size(); i++) {
		context->setTransformation(i, calculateTransformation(constraints[i]));
	}
}




Matrix4f RigidDeformationSolver::calculateTransformation(const std::vector<pair<Vector3f, Vector3f> > &pairs) 
{
	// estimate centroids of features
	Vector3f centroidA = NULL_VECTOR3F, centroidB = NULL_VECTOR3F;
	for( card32 i=0;i<pairs.size();i++ )
	{
		centroidA += pairs[i].first / (float)pairs.size();
		centroidB += pairs[i].second / (float)pairs.size();
	}

	// calculate cross-covariance matrix
	Matrix3f C;
	C *= 0;
	for( card32 i=0;i<pairs.size();i++ )
	{
		for( card32 j=0;j<3;j++ )
		{
			for( card32 k=0;k<3;k++ )
			{
				C[j][k] += (pairs[i].first[j] - centroidA[j])
					* (pairs[i].second[k] - centroidB[k]);
			}
		}
	}

   // Prepare OpenCV Datasructure CvMat
	Matrix3f U,V;
	Vector3f eigenValues;
   CvMat cvMatInput, cvMat_U, cvMat_W, cvMat_Vt;
   cvInitMatHeader( &cvMatInput, 3, 3, CV_32FC1, C.data());
	cvInitMatHeader( &cvMat_W, 3, 1, CV_32FC1, eigenValues.data() );
	cvInitMatHeader( &cvMat_U, 3, 3, CV_32FC1, U.data() );
	cvInitMatHeader( &cvMat_Vt, 3, 3, CV_32FC1, V.data() );

   // Do SVD
   cvSVD(&cvMatInput, &cvMat_W, &cvMat_U, &cvMat_Vt, CV_SVD_MODIFY_A | CV_SVD_U_T );

	Matrix3f rotation = U*V;
	Matrix4f ret = makeTranslation4f(centroidB) * expand3To4(rotation.transpose()) * makeTranslation4f(-centroidA);

	return ret;
}

void X4::RigidDeformationSolver::addHalfSpaceConstraint( const ARHalfSpaceConstraint *c )
{
   notImplemented();
}

void X4::RigidDeformationSolver::addSpatialDerivativeConstraint(const SpatialDerivativeARConstraint *c) {
   notImplemented();
}

void X4::RigidDeformationSolver::addSimpleSpatialDerivativeConstraint(const SimpleSpatialDerivativeARConstraint *c) {
   notImplemented();
}
