//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "TSDeformationContext.h"
//---------------------------------------------------------------------------
#include "SparseGrid.inline.h"
#include "DebugRenderer.h"
#include "TSDeformationSolver.h"
#include "TSDeformationEvaluator.h"
#include "PartialDerivative.h"
#include "PartialDerivative.inline.h"
#include "SparseLinearAlgebra.inline.h"
#include "ARSettings.h"
#include "DebugRenderer.h"
#include "TSSmoothDeformationSolver.h"
#include "TSSmoothDeformationEvaluator.h"
#include "ColorTools.h"
#include "PointCloudTools.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------



IMPLEMENT_X4_CLASS( TSDeformationContext , 1 ) {
    BEGIN_CLASS_INIT( TSDeformationContext );
}

// ----


void TSDeformationContext::TimeStep::initializeFromShape(const SceneObject *obj, const ARSettings *settings) {
    this->settings = settings;
    const UnstructuredInCorePointCloud *pc = dynamic_cast<const UnstructuredInCorePointCloud*>(obj);
    if (!pc) {
        throw X4Exception("TSDeformationContext::initializeFromShape() - obj is not of correct type.");
    }

    const PointSet *ps = pc->getPointSet();
    AAT positionAAT = ps->getAAT("position", 3, VAD::DATA_FORMAT_FLOAT32);
    mpcard numPoints = ps->getNumEntries();
    for (mpcard i=0; i<numPoints; i++) {
        checkAddEntryWithNeighbors(ps->get3f(i, positionAAT));
    }
}

inline Vector3i TSDeformationContext::TimeStep::computeGridCell(Vector3f pos) 
{
    return makeVector3i(
        (int32)floor(pos[0]/gridSpacing+0.5f),
        (int32)floor(pos[1]/gridSpacing+0.5f),
        (int32)floor(pos[2]/gridSpacing+0.5f)
        );
}

inline Vector3f TSDeformationContext::TimeStep::computeCenterOfCell(Vector3i cell) 
{
    return makeVector3f(
        (cell[0]+0.5f)*gridSpacing,
        (cell[1]+0.5f)*gridSpacing,
        (cell[2]+0.5f)*gridSpacing
        );
}

inline bool TSDeformationContext::TimeStep::getDeformPos(Vector3i cell, Vector3f &deformPos) {
    int32 index;
    bool result = grid.getObject(cell, index);
    if (result) {
        x4Assert(index < (int32)gridEntries.size());
        deformPos = gridEntries[index].deformPos;
        return true;
    } else {
        return false;
    }
}

inline bool TSDeformationContext::TimeStep::getDeformIndex(Vector3i cell, int32 &index) {
    return grid.getObject(cell, index);
}

void TSDeformationContext::TimeStep::addGridEntry(Vector3i cell) 
{
    int32 forgetme;
    bool haveItAlready = grid.getObject(cell, forgetme);
    if (!haveItAlready) {
        int32 index = (int32)gridEntries.size();
        grid.setObject(cell, index);
        GridEntry e;
        e.gridCell = cell;
        e.deformPos = computeCenterOfCell(cell);
        e.containsUrshapePoints = false;
        gridEntries.push_back(e);
    }
}

void TSDeformationContext::TimeStep::checkAddEntryWithNeighbors(Vector3f pos) 
{
    Vector3i cell = computeGridCell(pos);
    for (int z=-gridExtension; z<=gridExtension; z++) {
        for (int y=-gridExtension; y<=gridExtension; y++) {
            for (int x=-gridExtension; x<=gridExtension; x++) {
                Vector3i cc = cell;
                cc[0] += x;
                cc[1] += y;
                cc[2] += z;
                addGridEntry(cc);
            }
        }
    }

    Vector3i lowerGridPoint = makeVector3i(
        (int32)floor(pos[0]/gridSpacing-0.5f),
        (int32)floor(pos[1]/gridSpacing-0.5f),
        (int32)floor(pos[2]/gridSpacing-0.5f)
        );

    int32 index;
    grid.getObject(lowerGridPoint, index);
    gridEntries[index].containsUrshapePoints = true;

}


void TSDeformationContext::TimeStep::debugOutputVis() 
{
    debugRenderer->beginRenderJob("TSDeformationContext::debugOutputVis()");
    mpcard n = grid.getNumElements();
    for (mpcard i=0; i<n; i++) {
        int32 index = grid.getEntry(i);
        GridEntry e = gridEntries[index];
        Vector3f startPos = computeCenterOfCell(e.gridCell);
        debugRenderer->addFastSphere(startPos, gridSpacing/3, makeVector3f(0.8f,0.7f,0.2f), false);
        debugRenderer->addFineArrow(startPos, e.deformPos, makeVector3f(0.4f,0.8f,0.4f), false);
    }
    debugRenderer->endRenderJob();
}

//---------------------------------------------------------------------------


TSDeformationContext::TSDeformationContext(int32 gridExtension) {
    this->gridExtension = gridExtension;
}

TSDeformationContext::~TSDeformationContext() {
    for( unsigned i=0;i<timeSteps.size();i++ )
    {
        delete timeSteps[i];
    }
}


void TSDeformationContext::debugOutputVis(int32 time) 
{
    if (!timeSteps.empty() && time-startTime <= endTime) {
        timeSteps[time-startTime]->debugOutputVis();
    }
}

DeformationSolver *TSDeformationContext::createSolver() 
{
    if (gridExtension == 2) {
        return new TSDeformationSolver();
    } else {
        return new TSSmoothDeformationSolver();
    }
}

DeformationEvaluator *TSDeformationContext::createEvaluator() 
{
    if (gridExtension == 2) {
        return new TSDeformationEvaluator();
    } else {
        return new TSSmoothDeformationEvaluator();
    }
}

TSDeformationContext::TimeStep::TimeStep(const ARSettings *settings, int32 gridExtension) {
    this->settings = settings;
    this->gridExtension = gridExtension;
    this->gridSpacing = settings->tsGridSpacing;
    lastRegWeight = -0.12345f;
    lastDefModel = 0xFFFF;
    lastIDWeight = -0.12345f;
    lastMix = -0.12345f;
    last_movement = 1e20f;
}

void TSDeformationContext::initializeFromShape(const SceneObject *obj, Vector2f timeInterval, const ARSettings *settings) {
    this->settings = settings;
    startTime = (int32)timeInterval[0];
    endTime = (int32)timeInterval[1];
    if (startTime != timeInterval[0] || endTime != timeInterval[1]) {
        warning("TSDeformationContext::initializeFromShape - supports only integral time steps (fractional time interval given).");
    }
    timeSteps.clear();
    timeSteps.resize(endTime-startTime+1);
    for (int32 i=startTime; i<=endTime; i++) {
        timeSteps[i] = new TimeStep(settings, gridExtension);
        timeSteps[i]->initializeFromShape(obj, settings );
    }
}

float32 TSDeformationContext::getDuration() 
{
    return getNumTimeSteps();
}


void TSDeformationContext::getRegularizer(int32 time, SparseMatrixF &mat, DVectorF &vec) 
{
    time += startTime;
    x4Assert(time >= 0 && time <= endTime);
    timeSteps[time]->getRegularizer(mat, vec);
}

void TSDeformationContext::addPositionConstraint(int32 time, const PositionARConstraint *constr, SparseMatrixF &mat, DVectorF &vec) 
{
    time += startTime;
    x4Assert(time >= 0 && time <= endTime);
    timeSteps[time]->addPositionConstraint(constr, mat, vec);
}

void TSDeformationContext::addPositionConstraintSmoothed(int32 time, const PositionARConstraint *constr, SparseMatrixF &mat, DVectorF &vec) 
{
    time += startTime;
    x4Assert(time >= 0 && time <= endTime);
    timeSteps[time]->addPositionConstraintSmoothed(constr, mat, vec);
}

void TSDeformationContext::addSpatialDerivativeConstraint(int32 time, const SpatialDerivativeARConstraint *constr, SparseMatrixF &mat, DVectorF &vec) 
{
    time += startTime;
    x4Assert(time >= 0 && time <= endTime);
    timeSteps[time]->addSpatialDerivativeConstraint(constr, mat, vec);
}

void TSDeformationContext::addSimpleSpatialDerivativeConstraint(int32 time, const SimpleSpatialDerivativeARConstraint *constr, SparseMatrixF &mat, DVectorF &vec) 
{
    time += startTime;
    x4Assert(time >= 0 && time <= endTime);
    timeSteps[time]->addSimpleSpatialDerivativeConstraint(constr, mat, vec);
}

void TSDeformationContext::setResult(int32 time, const DVectorF &result) 
{
    time += startTime;
    x4Assert(time >= 0 && time <= endTime);
    timeSteps[time]->setResult(result);
}

void TSDeformationContext::getLastResult(int32 time, DVectorF &result)
{
    time += startTime;
    x4Assert(time >= 0 && time <= endTime);
    timeSteps[time]->getLastResult(result);
}


static StaticMatrix<float32, 4,4> computeLPStencilMatrix()
{
    // E = 1/h[ (pcc - ccc)?+ (cpc - ccc)?+ (ccp - ccc)?]
    //
    // MAPLE:
    // diff(E,ccc) = -2 pcc + 6 ccc - 2 cpc - 2 ccp
    // diff(E,pcc) = 2 pcc - 2 ccc
    // diff(E,cpc) = 2 cpc - 2 ccc 
    // diff(E,ccp) = 2 ccp - 2 ccc
    //
    // this means:
    // grad(E(ccc, pcc, cpc, ccp)) =
    //       ccc  pcc  cpc  ccp
    //     (  6,  -2,  -2,  -2 )   (ccc)
    //     ( -2,   2,   0,   0 ) * (pcc)
    //     ( -2,   0,   2,   0 )   (cpc)
    //     ( -2,   0,   0,   2 )   (ccp)

    return makeMatrix4f(
        6,  -2,  -2,  -2,
        -2,   2,   0,   0,
        -2,   0,   2,   0,
        -2,   0,   0,   2
        );
}

static HessianTools<float32, 4>::DiffTwice energyLP(
    HessianTools<float32, 4>::DiffTwice ccc,
    HessianTools<float32, 4>::DiffTwice pcc,
    HessianTools<float32, 4>::DiffTwice cpc,
    HessianTools<float32, 4>::DiffTwice ccp)
{
    return sqr(pcc - ccc) + sqr(cpc - ccc) + sqr(ccp - ccc);
}

static HessianTools<float32, 19>::DiffTwice energyTS(
    HessianTools<float32, 19>::DiffTwice ccc,

    HessianTools<float32, 19>::DiffTwice pcc,
    HessianTools<float32, 19>::DiffTwice cpc,
    HessianTools<float32, 19>::DiffTwice ccp,

    HessianTools<float32, 19>::DiffTwice mcc,
    HessianTools<float32, 19>::DiffTwice cmc,
    HessianTools<float32, 19>::DiffTwice ccm,

    HessianTools<float32, 19>::DiffTwice ppc,
    HessianTools<float32, 19>::DiffTwice cpp,
    HessianTools<float32, 19>::DiffTwice pcp,

    HessianTools<float32, 19>::DiffTwice mmc,
    HessianTools<float32, 19>::DiffTwice mcm,
    HessianTools<float32, 19>::DiffTwice mcp,
    HessianTools<float32, 19>::DiffTwice mpc,
    HessianTools<float32, 19>::DiffTwice cmm,
    HessianTools<float32, 19>::DiffTwice cmp,
    HessianTools<float32, 19>::DiffTwice pmc,
    HessianTools<float32, 19>::DiffTwice pcm,
    HessianTools<float32, 19>::DiffTwice cpm
    )
{
    //return sqr((ccc-mcc) - (pcc-ccc))
    //     + sqr((ccc-cmc) - (cpc-ccc))
    //     + sqr((ccc-ccm) - (ccp-ccc))
    //     + sqr((ppc-pcc) - (cpc-ccc)) + sqr((ppc-pcc) - (cpc-ccc))
    //     + sqr((pcp-pcc) - (ccp-ccc)) + sqr((pcp-pcc) - (ccp-ccc))
    //     + sqr((cpp-cpc) - (ccp-ccc)) + sqr((cpp-cpc) - (ccp-ccc));
    //return sqr((ccc-mcc) - (pcc-ccc))
    //     + sqr((ccc-cmc) - (cpc-ccc))
    //     + sqr((ccc-ccm) - (ccp-ccc))
    //     + sqr((ppc-pcc) - (cpc-ccc)) 
    //     + sqr((pcp-pcc) - (ccp-ccc)) 
    //     + sqr((cpp-cpc) - (ccp-ccc)) 
    //     
    //     + sqr(-(mmc-mcc) + (cmc-ccc)) 
    //     + sqr(-(mcm-mcc) + (ccm-ccc)) 
    //     + sqr(-(cmm-cmc) + (ccm-ccc)) ;
    return sqr((ccc-mcc) - (pcc-ccc))
        + sqr((ccc-cmc) - (cpc-ccc))
        + sqr((ccc-ccm) - (ccp-ccc))

        + sqr((ppc-pmc) - (mpc-mmc)) 
        + sqr((pcp-pcm) - (mcp-mcm)) 
        + sqr((cpp-cpm) - (cmp-cmm));
}

static HessianTools<float32, 19>::DiffTwice energyTS2D(
    HessianTools<float32, 19>::DiffTwice ccc,

    HessianTools<float32, 19>::DiffTwice pcc,
    HessianTools<float32, 19>::DiffTwice cpc,
    HessianTools<float32, 19>::DiffTwice ccp,

    HessianTools<float32, 19>::DiffTwice mcc,
    HessianTools<float32, 19>::DiffTwice cmc,
    HessianTools<float32, 19>::DiffTwice ccm,

    HessianTools<float32, 19>::DiffTwice ppc,
    HessianTools<float32, 19>::DiffTwice cpp,
    HessianTools<float32, 19>::DiffTwice pcp,

    HessianTools<float32, 19>::DiffTwice mmc,
    HessianTools<float32, 19>::DiffTwice mcm,
    HessianTools<float32, 19>::DiffTwice mcp,
    HessianTools<float32, 19>::DiffTwice mpc,
    HessianTools<float32, 19>::DiffTwice cmm,
    HessianTools<float32, 19>::DiffTwice cmp,
    HessianTools<float32, 19>::DiffTwice pmc,
    HessianTools<float32, 19>::DiffTwice pcm,
    HessianTools<float32, 19>::DiffTwice cpm
    )
{
    //return sqr((ccc-mcc) - (pcc-ccc))
    //     + sqr((ccc-cmc) - (cpc-ccc))
    //     + sqr((ccc-ccm) - (ccp-ccc))
    //     + sqr((ppc-pcc) - (cpc-ccc)) + sqr((ppc-pcc) - (cpc-ccc))
    //     + sqr((pcp-pcc) - (ccp-ccc)) + sqr((pcp-pcc) - (ccp-ccc))
    //     + sqr((cpp-cpc) - (ccp-ccc)) + sqr((cpp-cpc) - (ccp-ccc));
    //return sqr((ccc-mcc) - (pcc-ccc))
    //     + sqr((ccc-cmc) - (cpc-ccc))
    //     + sqr((ccc-ccm) - (ccp-ccc))
    //     + sqr((ppc-pcc) - (cpc-ccc)) 
    //     + sqr((pcp-pcc) - (ccp-ccc)) 
    //     + sqr((cpp-cpc) - (ccp-ccc)) 
    //     
    //     + sqr(-(mmc-mcc) + (cmc-ccc)) 
    //     + sqr(-(mcm-mcc) + (ccm-ccc)) 
    //     + sqr(-(cmm-cmc) + (ccm-ccc)) ;
    return sqr((ccc-mcc) - (pcc-ccc))
        + sqr((ccc-cmc) - (cpc-ccc))

        + sqr((ppc-pmc) - (mpc-mmc));
}

void X4::testDiff_for_TS() 
{
    {
        HessianTools<float32, 4>::DiffTwice ccc = HessianTools<float32, 4>::createVariable(0.0f, 0);
        HessianTools<float32, 4>::DiffTwice pcc = HessianTools<float32, 4>::createVariable(0.0f, 1);
        HessianTools<float32, 4>::DiffTwice cpc = HessianTools<float32, 4>::createVariable(0.0f, 2);
        HessianTools<float32, 4>::DiffTwice ccp = HessianTools<float32, 4>::createVariable(0.0f, 3);
        HessianTools<float32, 4>::DiffTwice result = energyLP(ccc, pcc, cpc, ccp);
        Matrix4f m = HessianTools<float32, 4>::getHessian(result);
        debugOutput << m << "\n";
    } 

    {
        HessianTools<float32, 19>::DiffTwice ccc = HessianTools<float32, 19>::createVariable(0.0f, 0);

        HessianTools<float32, 19>::DiffTwice pcc = HessianTools<float32, 19>::createVariable(0.0f, 1);
        HessianTools<float32, 19>::DiffTwice cpc = HessianTools<float32, 19>::createVariable(0.0f, 2);
        HessianTools<float32, 19>::DiffTwice ccp = HessianTools<float32, 19>::createVariable(0.0f, 3);

        HessianTools<float32, 19>::DiffTwice mcc = HessianTools<float32, 19>::createVariable(0.0f, 4);
        HessianTools<float32, 19>::DiffTwice cmc = HessianTools<float32, 19>::createVariable(0.0f, 5);
        HessianTools<float32, 19>::DiffTwice ccm = HessianTools<float32, 19>::createVariable(0.0f, 6);

        HessianTools<float32, 19>::DiffTwice ppc = HessianTools<float32, 19>::createVariable(0.0f, 7);
        HessianTools<float32, 19>::DiffTwice cpp = HessianTools<float32, 19>::createVariable(0.0f, 8);
        HessianTools<float32, 19>::DiffTwice pcp = HessianTools<float32, 19>::createVariable(0.0f, 9);

        HessianTools<float32, 19>::DiffTwice mmc = HessianTools<float32, 19>::createVariable(0.0f,10);
        HessianTools<float32, 19>::DiffTwice mcm = HessianTools<float32, 19>::createVariable(0.0f,11);
        HessianTools<float32, 19>::DiffTwice mcp = HessianTools<float32, 19>::createVariable(0.0f,12);
        HessianTools<float32, 19>::DiffTwice mpc = HessianTools<float32, 19>::createVariable(0.0f,13);
        HessianTools<float32, 19>::DiffTwice cmm = HessianTools<float32, 19>::createVariable(0.0f,14);
        HessianTools<float32, 19>::DiffTwice cmp = HessianTools<float32, 19>::createVariable(0.0f,15);
        HessianTools<float32, 19>::DiffTwice pmc = HessianTools<float32, 19>::createVariable(0.0f,16);
        HessianTools<float32, 19>::DiffTwice pcm = HessianTools<float32, 19>::createVariable(0.0f,17);
        HessianTools<float32, 19>::DiffTwice cpm = HessianTools<float32, 19>::createVariable(0.0f,18);

        HessianTools<float32, 19>::DiffTwice result = energyTS(
            ccc,  
            pcc, cpc, ccp,  
            mcc, cmc, ccm,  
            ppc, cpp, pcp, 
            mmc, mcm, mcp, mpc, cmm, cmp, pmc, pcm, cpm
            );

        StaticMatrix<float32, 19, 19> m2 = HessianTools<float32, 19>::getHessian(result);
        debugOutput << m2.toString() << "\n";
    }
}

StaticMatrix<float32, 19, 19> computeTSStencilMatrix()
{
    StaticMatrix<float32, 19, 19> r;
    HessianTools<float32, 19>::DiffTwice ccc = HessianTools<float32, 19>::createVariable(0.0f, 0);

    HessianTools<float32, 19>::DiffTwice pcc = HessianTools<float32, 19>::createVariable(0.0f, 1);
    HessianTools<float32, 19>::DiffTwice cpc = HessianTools<float32, 19>::createVariable(0.0f, 2);
    HessianTools<float32, 19>::DiffTwice ccp = HessianTools<float32, 19>::createVariable(0.0f, 3);

    HessianTools<float32, 19>::DiffTwice mcc = HessianTools<float32, 19>::createVariable(0.0f, 4);
    HessianTools<float32, 19>::DiffTwice cmc = HessianTools<float32, 19>::createVariable(0.0f, 5);
    HessianTools<float32, 19>::DiffTwice ccm = HessianTools<float32, 19>::createVariable(0.0f, 6);

    HessianTools<float32, 19>::DiffTwice ppc = HessianTools<float32, 19>::createVariable(0.0f, 7);
    HessianTools<float32, 19>::DiffTwice cpp = HessianTools<float32, 19>::createVariable(0.0f, 8);
    HessianTools<float32, 19>::DiffTwice pcp = HessianTools<float32, 19>::createVariable(0.0f, 9);

    HessianTools<float32, 19>::DiffTwice mmc = HessianTools<float32, 19>::createVariable(0.0f,10);
    HessianTools<float32, 19>::DiffTwice mcm = HessianTools<float32, 19>::createVariable(0.0f,11);
    HessianTools<float32, 19>::DiffTwice mcp = HessianTools<float32, 19>::createVariable(0.0f,12);
    HessianTools<float32, 19>::DiffTwice mpc = HessianTools<float32, 19>::createVariable(0.0f,13);
    HessianTools<float32, 19>::DiffTwice cmm = HessianTools<float32, 19>::createVariable(0.0f,14);
    HessianTools<float32, 19>::DiffTwice cmp = HessianTools<float32, 19>::createVariable(0.0f,15);
    HessianTools<float32, 19>::DiffTwice pmc = HessianTools<float32, 19>::createVariable(0.0f,16);
    HessianTools<float32, 19>::DiffTwice pcm = HessianTools<float32, 19>::createVariable(0.0f,17);
    HessianTools<float32, 19>::DiffTwice cpm = HessianTools<float32, 19>::createVariable(0.0f,18);

    HessianTools<float32, 19>::DiffTwice result = energyTS(
        ccc,  
        pcc, cpc, ccp,  
        mcc, cmc, ccm,  
        ppc, cpp, pcp, 
        mmc, mcm, mcp, mpc, cmm, cmp, pmc, pcm, cpm
        );

    return HessianTools<float32, 19>::getHessian(result);
}

StaticMatrix<float32, 19, 19> computeTSStencilMatrix2D()
{
    StaticMatrix<float32, 19, 19> r;
    HessianTools<float32, 19>::DiffTwice ccc = HessianTools<float32, 19>::createVariable(0.0f, 0);

    HessianTools<float32, 19>::DiffTwice pcc = HessianTools<float32, 19>::createVariable(0.0f, 1);
    HessianTools<float32, 19>::DiffTwice cpc = HessianTools<float32, 19>::createVariable(0.0f, 2);
    HessianTools<float32, 19>::DiffTwice ccp = HessianTools<float32, 19>::createVariable(0.0f, 3);

    HessianTools<float32, 19>::DiffTwice mcc = HessianTools<float32, 19>::createVariable(0.0f, 4);
    HessianTools<float32, 19>::DiffTwice cmc = HessianTools<float32, 19>::createVariable(0.0f, 5);
    HessianTools<float32, 19>::DiffTwice ccm = HessianTools<float32, 19>::createVariable(0.0f, 6);

    HessianTools<float32, 19>::DiffTwice ppc = HessianTools<float32, 19>::createVariable(0.0f, 7);
    HessianTools<float32, 19>::DiffTwice cpp = HessianTools<float32, 19>::createVariable(0.0f, 8);
    HessianTools<float32, 19>::DiffTwice pcp = HessianTools<float32, 19>::createVariable(0.0f, 9);

    HessianTools<float32, 19>::DiffTwice mmc = HessianTools<float32, 19>::createVariable(0.0f,10);
    HessianTools<float32, 19>::DiffTwice mcm = HessianTools<float32, 19>::createVariable(0.0f,11);
    HessianTools<float32, 19>::DiffTwice mcp = HessianTools<float32, 19>::createVariable(0.0f,12);
    HessianTools<float32, 19>::DiffTwice mpc = HessianTools<float32, 19>::createVariable(0.0f,13);
    HessianTools<float32, 19>::DiffTwice cmm = HessianTools<float32, 19>::createVariable(0.0f,14);
    HessianTools<float32, 19>::DiffTwice cmp = HessianTools<float32, 19>::createVariable(0.0f,15);
    HessianTools<float32, 19>::DiffTwice pmc = HessianTools<float32, 19>::createVariable(0.0f,16);
    HessianTools<float32, 19>::DiffTwice pcm = HessianTools<float32, 19>::createVariable(0.0f,17);
    HessianTools<float32, 19>::DiffTwice cpm = HessianTools<float32, 19>::createVariable(0.0f,18);

    HessianTools<float32, 19>::DiffTwice result = energyTS2D(
        ccc,  
        pcc, cpc, ccp,  
        mcc, cmc, ccm,  
        ppc, cpp, pcp, 
        mmc, mcm, mcp, mpc, cmm, cmp, pmc, pcm, cpm
        );

    return HessianTools<float32, 19>::getHessian(result);
}


const StaticMatrix<float32, 4, 4> TSDeformationContext::TimeStep::LP_STENCIL_MATRIX = computeLPStencilMatrix();
const StaticMatrix<float32, 19, 19> TSDeformationContext::TimeStep::TS_STENCIL_MATRIX = computeTSStencilMatrix();
const StaticMatrix<float32, 19, 19> TSDeformationContext::TimeStep::TS_STENCIL_MATRIX2D = computeTSStencilMatrix2D();


bool TSDeformationContext::TimeStep::getLPStencil(card32 index, LPStencil &stencil) {
    Vector3i cell = gridEntries[index].gridCell;
    int32 nindex;
    if (grid.getObject(cell+makeVector3i(0,0,0), nindex)) {
        stencil.ccc = nindex;
    } else {
        return false;
    }
    if (grid.getObject(cell+makeVector3i(1,0,0), nindex)) {
        stencil.pcc = nindex;
    } else {
        return false;
    }
    if (grid.getObject(cell+makeVector3i(0,1,0), nindex)) {
        stencil.cpc = nindex;
    } else {
        return false;
    }
    if (grid.getObject(cell+makeVector3i(0,0,1), nindex)) {
        stencil.ccp = nindex;
    } else {
        return false;
    }
    x4Assert(stencil.ccc == index);
    return true;
}

bool TSDeformationContext::TimeStep::getTSStencil(card32 index, TSStencil &stencil) {
    Vector3i cell = gridEntries[index].gridCell;
    int32 nindex;
    if (grid.getObject(cell+makeVector3i(0,0,0), nindex)) {
        stencil.ccc = nindex;
    } else {
        return false;
    }
    if (grid.getObject(cell+makeVector3i(1,0,0), nindex)) {
        stencil.pcc = nindex;
    } else {
        return false;
    }
    if (grid.getObject(cell+makeVector3i(0,1,0), nindex)) {
        stencil.cpc = nindex;
    } else {
        return false;
    }
    if (grid.getObject(cell+makeVector3i(0,0,1), nindex)) {
        stencil.ccp = nindex;
    } else {
        return false;
    }

    if (grid.getObject(cell+makeVector3i(-1,0,0), nindex)) {
        stencil.mcc = nindex;
    } else {
        return false;
    }
    if (grid.getObject(cell+makeVector3i(0,-1,0), nindex)) {
        stencil.cmc = nindex;
    } else {
        return false;
    }
    if (grid.getObject(cell+makeVector3i(0,0,-1), nindex)) {
        stencil.ccm = nindex;
    } else {
        return false;
    }

    if (grid.getObject(cell+makeVector3i(1,1,0), nindex)) {
        stencil.ppc = nindex;
    } else {
        return false;
    }
    if (grid.getObject(cell+makeVector3i(0,1,1), nindex)) {
        stencil.cpp = nindex;
    } else {
        return false;
    }
    if (grid.getObject(cell+makeVector3i(1,0,1), nindex)) {
        stencil.pcp = nindex;
    } else {
        return false;
    }

    if (grid.getObject(cell+makeVector3i(-1,-1,0), nindex)) {         
        stencil.mmc = nindex;                                        
    } else {                                                        
        return false;                                                
    }                                                               
    if (grid.getObject(cell+makeVector3i(-1,0,-1), nindex)) {         
        stencil.mcm = nindex;                                        
    } else {                                                        
        return false;                                                
    }
    if (grid.getObject(cell+makeVector3i(-1,0,1), nindex)) {
        stencil.mcp = nindex;
    } else {
        return false;
    }
    if (grid.getObject(cell+makeVector3i(-1,1,0), nindex)) {
        stencil.mpc = nindex;
    } else {
        return false;
    }
    if (grid.getObject(cell+makeVector3i(0,-1,-1), nindex)) {
        stencil.cmm = nindex;
    } else {
        return false;
    }
    if (grid.getObject(cell+makeVector3i(0,-1,1), nindex)) {
        stencil.cmp = nindex;
    } else {
        return false;
    }
    if (grid.getObject(cell+makeVector3i(1,-1,0), nindex)) {
        stencil.pmc = nindex;
    } else {
        return false;
    }
    if (grid.getObject(cell+makeVector3i(1,0,-1), nindex)) {
        stencil.pcm = nindex;
    } else {
        return false;
    }
    if (grid.getObject(cell+makeVector3i(0,1,-1), nindex)) {
        stencil.cpm = nindex;
    } else {
        return false;
    }

    x4Assert(stencil.ccc == index);
    return true;
}


Vector4i TSDeformationContext::TimeStep::LPStencil::getIndexMapping() 
{
    Vector4i result;
    result[0] = ccc;
    result[1] = pcc;
    result[2] = cpc;
    result[3] = ccp;
    return result;   
}

StaticVector<int32, 19> TSDeformationContext::TimeStep::TSStencil::getIndexMapping() {
    StaticVector<int32, 19> result;
    result[0] = ccc;
    result[1] = pcc;
    result[2] = cpc;
    result[3] = ccp;
    result[4] = mcc;
    result[5] = cmc;
    result[6] = ccm;
    result[7] = ppc;
    result[8] = cpp;
    result[9] = pcp;
    result[10] = mmc;
    result[11] = mcm;
    result[12] = mcp;
    result[13] = mpc;
    result[14] = cmm;
    result[15] = cmp;
    result[16] = pmc;
    result[17] = pcm;
    result[18] = cpm;
    return result;
}

StaticVector<int32, 9> TSDeformationContext::TimeStep::TSStencil2D::getIndexMapping() {
    StaticVector<int32, 9> result;
    result[0] = cc;

    result[1] = pc;
    result[2] = cp;
    result[3] = mc;
    result[4] = cm;

    result[5] = pp;
    result[6] = mm;
    result[7] = mp;
    result[8] = pm;

    return result;
}


template<typename FloatType, int sdim>
static void scatterAddMat3(
    SparseMatrix<FloatType> &dest,
    const StaticMatrix<FloatType, sdim, sdim> &dense, 
    const StaticVector<int32, sdim> &indicesMapping) 
{
    if (!dense.getColumns() == indicesMapping.getDim()) {
        throw ERangeCheck("scatterAndAddMatrix() - indices std::vector does not match matrix dimension.");
    }
    if (indicesMapping.getDim() == 0) {
        throw ERangeCheck("scatterAndAddMatrix() - zero size input is not possible.");
    }
    for (int d=0; d<3; d++) {
        for (int row=0; row<sdim; row++) {
            SparseVector<FloatType> rowVec;
            for (int column=0; column<sdim; column++) {
                if (dense[column][row] != 0) {
                    rowVec.setEntry(indicesMapping[column]*3+d, dense[column][row]);
                }
            }
            dest[indicesMapping[row]*3+d] += rowVec;
        }
    }
}

template<typename FloatType, int sdim>
static void scatterAddMat3SingleD(
    SparseMatrix<FloatType> &dest,
    const StaticMatrix<FloatType, sdim, sdim> &dense, 
    const StaticVector<int32, sdim> &indicesMapping,
    const card32 d) 
{
    if (!dense.getColumns() == indicesMapping.getDim()) {
        throw ERangeCheck("scatterAndAddMatrix() - indices std::vector does not match matrix dimension.");
    }
    if (indicesMapping.getDim() == 0) {
        throw ERangeCheck("scatterAndAddMatrix() - zero size input is not possible.");
    }
    for (int row=0; row<sdim; row++) {
        SparseVector<FloatType> rowVec;
        for (int column=0; column<sdim; column++) {
            if (dense[column][row] != 0) {
                rowVec.setEntry(indicesMapping[column]*3+d, dense[column][row]);
            }
        }
        dest[indicesMapping[row]*3+d] += rowVec;
    }
}

template<typename FloatType, int sdim>
static void scatterAddVec3(
    SparseVector<FloatType> &dest,
    const StaticVector<FloatType, sdim> &dense, 
    const StaticVector<int32, sdim> &indicesMapping) 
{
    for (int i=0; i<sdim; i++) {
        for (int d=0; d<3; d++) {
            dest[indicesMapping[i]*3+d] += dense[i];
        }
    }
}

template<typename FloatType, int sdim>
static void scatterAddVec3SingleD(
    DynamicVector<FloatType> &dest,
    const StaticVector<FloatType, sdim> &dense, 
    const StaticVector<int32, sdim> &indicesMapping,
    card32 d) 
{
    x4Assert(d<3);
    for (int i=0; i<sdim; i++) {
        dest[indicesMapping[i]*3+d] += dense[i];
    }
}


void TSDeformationContext::TimeStep::getRegularizer(SparseMatrixF &mat, DVectorF &vec)
{
    vec.setDim(gridEntries.size()*3);
    vec.setZero();
    if (gridEntries.empty()) {
        throw X4Exception("TSDeformationContext::TimeStep::getRegularizer - empty grid.");
    }
    if (   settings->tsRegularizerWeight != lastRegWeight
        || settings->tsIdentityWeight != lastIDWeight
        || settings->tsDeformationModel != lastDefModel
        || settings->tsMixedModeLPComp != lastMix
        || lastRegMat.getRows() != gridEntries.size()*3) 
    {
        lastRegWeight = settings->tsRegularizerWeight;
        lastIDWeight = settings->tsIdentityWeight;
        lastDefModel = settings->tsDeformationModel;
        lastMix = settings->tsMixedModeLPComp;

        debugOutput << "rebuild reg mat\n";

        size_t numEntries = gridEntries.size();
        mat.setRows((card32)gridEntries.size()*3);
        for (size_t i=0; i<numEntries; i++) {
            if (settings->tsDeformationModel == TS_MODEL_LAPLACE || settings->tsDeformationModel == TS_MODEL_MIXED_MODE) {
                float32 w = settings->tsRegularizerWeight;
                if (settings->tsDeformationModel == TS_MODEL_MIXED_MODE) {
                    w *= settings->tsMixedModeLPComp;
                }
                LPStencil stencil;
                if (getLPStencil((card32)i, stencil)) {
                    Vector4i indexMapping = stencil.getIndexMapping();
                    scatterAddMat3<float32, 4>(mat, LP_STENCIL_MATRIX*(w*gridSpacing), indexMapping);
                }
            }
            if (settings->tsDeformationModel == TS_MODEL_THINPLATESPLINES || settings->tsDeformationModel == TS_MODEL_MIXED_MODE) {
                TSStencil stencil;
                if (getTSStencil((card32)i, stencil)) {
                    StaticVector<int32, 19> indexMapping = stencil.getIndexMapping();
                    if( settings->tsMode2D )
                        scatterAddMat3<float32, 19>(mat, TS_STENCIL_MATRIX2D * (settings->tsRegularizerWeight*gridSpacing), indexMapping);
                    else
                        scatterAddMat3<float32, 19>(mat, TS_STENCIL_MATRIX * (settings->tsRegularizerWeight*gridSpacing), indexMapping);
                }
            }
        }
        const card32 n = (card32)gridEntries.size();
        for (card32 i=0; i<n; i++) {
            for (int d=0; d<3; d++) {
                mat[3*i+d].addEntry(3*i+d, settings->tsIdentityWeight);
                vec[3*i+d] = computeCenterOfCell(gridEntries[i].gridCell)[d] * settings->tsIdentityWeight;
            }
        }
        lastRegMat = mat;
    } else {
        mat = lastRegMat;
    }

    if( settings->tsCachedCellMatrices )
    {
        StaticMatrix<float,24,24> mat_null;
        mat_null.setZero();
        cachedScatterMatrix = std::vector< StaticMatrix<float,24,24> >(gridEntries.size(),mat_null);
    }
}

static HessianTools<float32, 24>::DiffTwice constraintEnergy(
    StaticVector< HessianTools<float32, 24>::DiffTwice, 3 > pos000,
    StaticVector< HessianTools<float32, 24>::DiffTwice, 3 > pos001,
    StaticVector< HessianTools<float32, 24>::DiffTwice, 3 > pos010,
    StaticVector< HessianTools<float32, 24>::DiffTwice, 3 > pos011,
    StaticVector< HessianTools<float32, 24>::DiffTwice, 3 > pos100,
    StaticVector< HessianTools<float32, 24>::DiffTwice, 3 > pos101,
    StaticVector< HessianTools<float32, 24>::DiffTwice, 3 > pos110,
    StaticVector< HessianTools<float32, 24>::DiffTwice, 3 > pos111,
    float32 lamba0, float32 lambda1, float32 lambda2,
    Vector3f constrPoint,
    Matrix3f quad)
{
    StaticVector< HessianTools<float32, 24>::DiffTwice, 3 > interp
        = (   ( pos000*(1-lamba0) + pos100*lamba0 )*(1-lambda1)
        + ( pos010*(1-lamba0) + pos110*lamba0 )*  lambda1  
        ) * (1-lambda2)
        + (   ( pos001*(1-lamba0) + pos101*lamba0 )*(1-lambda1)
        + ( pos011*(1-lamba0) + pos111*lamba0 )*  lambda1 
        ) * lambda2;
    StaticVector< HessianTools<float32, 24>::DiffTwice, 3 > diff = interp;
    diff[0] -= HessianTools<float32, 24>::DiffTwice(constrPoint[0]);
    diff[1] -= HessianTools<float32, 24>::DiffTwice(constrPoint[1]);
    diff[2] -= HessianTools<float32, 24>::DiffTwice(constrPoint[2]);

    StaticMatrix< HessianTools<float32, 24>::DiffTwice, 3, 3> HTDTquad;
    HTDTquad[0][0] = HessianTools<float32, 24>::DiffTwice(quad[0][0]);
    HTDTquad[0][1] = HessianTools<float32, 24>::DiffTwice(quad[0][1]);
    HTDTquad[0][2] = HessianTools<float32, 24>::DiffTwice(quad[0][2]);
    HTDTquad[1][0] = HessianTools<float32, 24>::DiffTwice(quad[1][0]);
    HTDTquad[1][1] = HessianTools<float32, 24>::DiffTwice(quad[1][1]);
    HTDTquad[1][2] = HessianTools<float32, 24>::DiffTwice(quad[1][2]);
    HTDTquad[2][0] = HessianTools<float32, 24>::DiffTwice(quad[2][0]);
    HTDTquad[2][1] = HessianTools<float32, 24>::DiffTwice(quad[2][1]);
    HTDTquad[2][2] = HessianTools<float32, 24>::DiffTwice(quad[2][2]);

    return (HTDTquad*diff)*diff;
}

static void computeConstrTermAD(
    Vector3f lambda,
    Vector3f constrPoint, Matrix3f quad,
    StaticMatrix<float32, 24, 24> &localMat, StaticVector<float32, 24> &localRHS ) 
{
    float32 lamba0 = lambda[0];
    float32 lambda1 = lambda[1];
    float32 lambda2 = lambda[2];

    StaticVector< HessianTools<float32, 24>::DiffTwice, 3 > pos000;
    StaticVector< HessianTools<float32, 24>::DiffTwice, 3 > pos001;
    StaticVector< HessianTools<float32, 24>::DiffTwice, 3 > pos010;
    StaticVector< HessianTools<float32, 24>::DiffTwice, 3 > pos011;
    StaticVector< HessianTools<float32, 24>::DiffTwice, 3 > pos100;
    StaticVector< HessianTools<float32, 24>::DiffTwice, 3 > pos101;
    StaticVector< HessianTools<float32, 24>::DiffTwice, 3 > pos110;
    StaticVector< HessianTools<float32, 24>::DiffTwice, 3 > pos111;

    pos000[0] = HessianTools<float32, 24>::createVariable(0.0f, 0);
    pos000[1] = HessianTools<float32, 24>::createVariable(0.0f, 1);
    pos000[2] = HessianTools<float32, 24>::createVariable(0.0f, 2);

    pos001[0] = HessianTools<float32, 24>::createVariable(0.0f, 3);
    pos001[1] = HessianTools<float32, 24>::createVariable(0.0f, 4);
    pos001[2] = HessianTools<float32, 24>::createVariable(0.0f, 5);

    pos010[0] = HessianTools<float32, 24>::createVariable(0.0f, 6);
    pos010[1] = HessianTools<float32, 24>::createVariable(0.0f, 7);
    pos010[2] = HessianTools<float32, 24>::createVariable(0.0f, 8);

    pos011[0] = HessianTools<float32, 24>::createVariable(0.0f, 9);
    pos011[1] = HessianTools<float32, 24>::createVariable(0.0f,10);
    pos011[2] = HessianTools<float32, 24>::createVariable(0.0f,11);

    pos100[0] = HessianTools<float32, 24>::createVariable(0.0f,12);
    pos100[1] = HessianTools<float32, 24>::createVariable(0.0f,13);
    pos100[2] = HessianTools<float32, 24>::createVariable(0.0f,14);

    pos101[0] = HessianTools<float32, 24>::createVariable(0.0f,15);
    pos101[1] = HessianTools<float32, 24>::createVariable(0.0f,16);
    pos101[2] = HessianTools<float32, 24>::createVariable(0.0f,17);

    pos110[0] = HessianTools<float32, 24>::createVariable(0.0f,18);
    pos110[1] = HessianTools<float32, 24>::createVariable(0.0f,19);
    pos110[2] = HessianTools<float32, 24>::createVariable(0.0f,20);

    pos111[0] = HessianTools<float32, 24>::createVariable(0.0f,21);
    pos111[1] = HessianTools<float32, 24>::createVariable(0.0f,22);
    pos111[2] = HessianTools<float32, 24>::createVariable(0.0f,23);

    HessianTools<float32, 24>::DiffTwice result = constraintEnergy(
        pos000, pos001, pos010, pos011,
        pos100, pos101, pos110, pos111,
        lamba0, lambda1, lambda2,
        constrPoint, quad
        );
    localMat = HessianTools<float32, 24>::getHessian(result);
    localRHS = -HessianTools<float32, 24>::getGradient(result);
}

static inline Vector3f compute_CT_RHS3(
    const Vector3f lambda, const Vector3f constrPoint, const Matrix3f quad, 
    const int index,
    StaticVector<float32, 8> interpFactors) 
{
    Vector3f result = (quad*constrPoint)*(2*interpFactors[index]);
    return result;
}

static inline Matrix3f compute_CT_Mat3(
    const Vector3f lambda, const Vector3f constrPoint, const Matrix3f quad, 
    const int index1, const int index2,
    StaticVector<float32, 8> interpFactors) 
{
    Matrix3f result = quad*(interpFactors[index1]*interpFactors[index2]*2);
    return result;
}

static void computeConstrTerm(
    Vector3f lambda, Vector3f constrPoint, Matrix3f quad,
    StaticMatrix<float32, 24, 24> &localMat, StaticVector<float32, 24> &localRHS ) 
{
    StaticVector<float32, 8> interpFactors;
    for (int variable=0; variable<8; variable++) {
        int index[3];
        index[2] = variable % 2;
        index[1] = (variable/2) % 2;
        index[0] = (variable/4) % 2;

        interpFactors[variable] = 1;
        if (index[0] == 1) {
            interpFactors[variable] *= lambda[0];
        } else {
            interpFactors[variable] *= 1-lambda[0];
        }
        if (index[1] == 1) {
            interpFactors[variable] *= lambda[1];
        } else {
            interpFactors[variable] *= 1-lambda[1];
        }
        if (index[2] == 1) {
            interpFactors[variable] *= lambda[2];
        } else {
            interpFactors[variable] *= 1-lambda[2];
        }
    }

    for (int index1=0; index1<8; index1++) {
        Vector3f rhs3 = compute_CT_RHS3(lambda, constrPoint, quad, index1, interpFactors);
        for (int d=0; d<3; d++) {
            localRHS[3*index1+d] = rhs3[d];
        }
        for (int index2=0; index2<8; index2++) {
            Matrix3f mat3 = compute_CT_Mat3(lambda, constrPoint, quad, index1, index2, interpFactors);
            for (int j=0; j<3; j++) {
                for (int i=0; i<3; i++) {
                    localMat[3*index1+j][3*index2+i] = mat3[j][i];
                }
            }
        }
    }
}


//static void computeConstrTermSmoothed(
//   Vector3f lambda, Vector3f constrPoint, Matrix3f quad,
//   StaticMatrix<float32, 7*7*7*3, 7*7*7*3> &localMat, StaticVector<float32, 7*7*7*3> &localRHS, const PositionARConstraint &at ) 
//{
//   localMat.setZero();
//   localRHS.setZero();
//
//   StaticVector<float32, 7*7*7> interpFactors;
//
//   static const card32 SUPPORT = 7;
//   static const int32 MINUS2 = 0;
//   static const int32 PLUS2 = 6;
//   static const int32 OFFSET = -3;
//
//   int32 index[SUPPORT][SUPPORT][SUPPORT];
//
//   Vector3f queryPos = at.pos;
//   Vector3i lowerGridPoint = makeVector3i(
//      (int32)floor(at.pos[0]/gridSpacing-0.5f),
//      (int32)floor(at.pos[1]/gridSpacing-0.5f),
//      (int32)floor(at.pos[2]/gridSpacing-0.5f)
//   );
//   Vector3f lowerGridCellCenter = computeCenterOfCell(lowerGridPoint);
//   
//   float32 wSum = 0;
//   Vector3f result = NULL_VECTOR3F;
//
//   for (int32 z= MINUS2; z<=PLUS2; z++) {
//      for (int32 y= MINUS2; y<=PLUS2; y++) {
//         for (int32 x= MINUS2; x<=PLUS2; x++) {
//            Vector3i cell = lowerGridPoint+makeVector3i(x+OFFSET, y+OFFSET, z+OFFSET);
//            int32 index = 0;
//            if (!grid.getObject(cell, index)) throw X4Exception("this should never happen");
//            Vector3f funPos = (makeVector3f(cell[0],cell[1], cell[2])+makeVector3f(0.5f, 0.5f, 0.5f))*gridSpacing;
//            Vector3f dest = gridEntries[index].deformPos;
//            float32 w = weightingFunction<float32>(norm(funPos - queryPos)/(gridSpacing*3), settings->tssBasisFunction);
//            result += dest*w;
//            wSum += w;
//         }
//      }
//   }
//   if (wSum >= 1e-7) result /= wSum;
//
//   for (int variable=0; variable<8; variable++) {
//      int index[3];
//      index[2] = variable % 2;
//      index[1] = (variable/2) % 2;
//      index[0] = (variable/4) % 2;
//      
//      interpFactors[variable] = 1;
//      if (index[0] == 1) {
//         interpFactors[variable] *= lambda[0];
//      } else {
//         interpFactors[variable] *= 1-lambda[0];
//      }
//      if (index[1] == 1) {
//         interpFactors[variable] *= lambda[1];
//      } else {
//         interpFactors[variable] *= 1-lambda[1];
//      }
//      if (index[2] == 1) {
//         interpFactors[variable] *= lambda[2];
//      } else {
//         interpFactors[variable] *= 1-lambda[2];
//      }
//   }
//
//   for (int index1=0; index1<8; index1++) {
//      Vector3f rhs3 = compute_CT_RHS3(lambda, constrPoint, quad, index1, interpFactors);
//      for (int d=0; d<3; d++) {
//         localRHS[3*index1+d] = rhs3[d];
//      }
//      for (int index2=0; index2<8; index2++) {
//         Matrix3f mat3 = compute_CT_Mat3(lambda, constrPoint, quad, index1, index2, interpFactors);
//         for (int j=0; j<3; j++) {
//            for (int i=0; i<3; i++) {
//               localMat[3*index1+j][3*index2+i] = mat3[j][i];
//            }
//         }
//      }
//   }
//}


template<typename FloatType, int sdim>
static void scatterAddMat1(
    SparseMatrix<FloatType> &dest,
    const StaticMatrix<FloatType, sdim, sdim> &dense, 
    const StaticVector<int32, sdim> &indicesMapping) 
{
    if (!dense.getColumns() == indicesMapping.getDim()) {
        throw ERangeCheck("scatterAndAddMatrix() - indices std::vector does not match matrix dimension.");
    }
    if (indicesMapping.getDim() == 0) {
        throw ERangeCheck("scatterAndAddMatrix() - zero size input is not possible.");
    }
    for (int row=0; row<sdim; row++) {
        SparseVector<FloatType> rowVec;
        for (int column=0; column<sdim; column++) {
            if (dense[column][row] != 0) {
                rowVec.setEntry(indicesMapping[column], dense[column][row]);
            }
        }
        dest[indicesMapping[row]] += rowVec;
    }
}

template<typename FloatType, int sdim>
static void scatterAddVec1(
    DVectorF &dest,
    const StaticVector<FloatType, sdim> &dense, 
    const StaticVector<int32, sdim> &indicesMapping) 
{
    if (indicesMapping.getDim() == 0) {
        throw ERangeCheck("scatterAndAddMatrix() - zero size input is not possible.");
    }
    for (int i=0; i<sdim; i++) {
        dest[indicesMapping[i]] += dense[i];
    }
}

void TSDeformationContext::TimeStep::addPositionConstraint(const PositionARConstraint *constr, SparseMatrixF &mat, DVectorF &vec)
{
    Vector3f footpoint = constr->anchor.pos;
    footpoint[2] = settings->tsMode2D?0:footpoint[2];
    Vector3i lowerGridPoint = makeVector3i(
        (int32)floor(footpoint[0]/gridSpacing-0.5f),
        (int32)floor(footpoint[1]/gridSpacing-0.5f),
        settings->tsMode2D?0:(int32)floor(footpoint[2]/gridSpacing-0.5f)
        );
    int32 index000, index001, index010, index011,
        index100, index101, index110, index111;

    if (!getDeformIndex(lowerGridPoint, index000)) {
#ifdef XGRT_DEBUG
        throw ERangeCheck("TSDeformationContext::TimeStep::getPosition() - position out of grid range.");
#endif
        return;
    }
    bool canInterpolate = true;
    canInterpolate &= getDeformIndex(lowerGridPoint+makeVector3i(0,1,0), index010);
    canInterpolate &= getDeformIndex(lowerGridPoint+makeVector3i(1,0,0), index100);
    canInterpolate &= getDeformIndex(lowerGridPoint+makeVector3i(1,1,0), index110);
    canInterpolate &= getDeformIndex(lowerGridPoint+makeVector3i(0,0,1), index001);
    canInterpolate &= getDeformIndex(lowerGridPoint+makeVector3i(0,1,1), index011);
    canInterpolate &= getDeformIndex(lowerGridPoint+makeVector3i(1,0,1), index101);
    canInterpolate &= getDeformIndex(lowerGridPoint+makeVector3i(1,1,1), index111);

    if (!canInterpolate) {
#ifdef XGRT_DEBUG
        throw ERangeCheck("TSDeformationContext::TimeStep::getPosition() - position out of grid range.");
#endif
        return;
        //Vector3i constrCell = computeGridCell(constr->anchor.pos);
        //Vector3f target = constr->toPoint;
        //int32 index;
        //if (grid.getObject(constrCell, index)) {
        //   for (card32 d=0; d<3; d++) {
        //      mat[index*3+d].addEntry(3*index+d, 2*settings->tsConstraintsWeight);
        //      vec[index*3+d] += 2*target[d]*settings->tsConstraintsWeight;
        //   }
        //}
    } else {
        Vector3f lowerGridCellCenter = computeCenterOfCell(lowerGridPoint);
        Vector3f frac = (footpoint - lowerGridCellCenter)/gridSpacing;
        x4Assert(frac[0] >= 0 && frac[0] <=1 && frac[1] >= 0 && frac[1] <=1 && frac[2] >= 0 && frac[2] <=1);

        StaticVector<int32, 24> indexMapping;
        indexMapping[ 0] = index000*3;
        indexMapping[ 1] = index000*3 + 1;
        indexMapping[ 2] = index000*3 + 2;

        indexMapping[ 3] = index001*3;
        indexMapping[ 4] = index001*3 + 1;
        indexMapping[ 5] = index001*3 + 2;

        indexMapping[ 6] = index010*3;
        indexMapping[ 7] = index010*3 + 1;
        indexMapping[ 8] = index010*3 + 2;

        indexMapping[ 9] = index011*3;
        indexMapping[10] = index011*3 + 1;
        indexMapping[11] = index011*3 + 2;

        indexMapping[12] = index100*3;
        indexMapping[13] = index100*3 + 1;
        indexMapping[14] = index100*3 + 2;

        indexMapping[15] = index101*3;
        indexMapping[16] = index101*3 + 1;
        indexMapping[17] = index101*3 + 2;

        indexMapping[18] = index110*3;
        indexMapping[19] = index110*3 + 1;
        indexMapping[20] = index110*3 + 2;

        indexMapping[21] = index111*3;
        indexMapping[22] = index111*3 + 1;
        indexMapping[23] = index111*3 + 2;

        StaticMatrix<float32, 24, 24> localMat;
        StaticVector<float32, 24> localRHS;
        Vector3f constrToPoint = constr->toPoint;
        constrToPoint[2] = settings->tsMode2D?0:constrToPoint[2];

        if (settings->tsSlowTest) {
            computeConstrTermAD(frac, constrToPoint, constr->covariance,
                localMat, localRHS);
        } else {
            computeConstrTerm(frac, constrToPoint, constr->covariance,
                localMat, localRHS);
        }
        scatterAddMat1(mat, localMat, indexMapping);
        scatterAddVec1(vec, localRHS, indexMapping);
    }
}

void TSDeformationContext::TimeStep::addPositionConstraintSmoothed(const PositionARConstraint *constr, SparseMatrixF &mat, DVectorF &vec)
{
    addPositionConstraint(constr, mat, vec);
    //Vector3f footpoint = constr->anchor.pos;
    //Vector3i lowerGridPoint = makeVector3i(
    //   (int32)floor(footpoint[0]/gridSpacing-0.5f),
    //   (int32)floor(footpoint[1]/gridSpacing-0.5f),
    //   (int32)floor(footpoint[2]/gridSpacing-0.5f)
    //);

    //static const card32 SUPPORT = 7;
    //static const int32 MINUS2 = 0;
    //static const int32 PLUS2 = 6;
    //static const int32 OFFSET = -3;
    //Vector3f pos[7][7][7];

    //int32 index[SUPPORT][SUPPORT][SUPPORT];
    //bool canInterpolate = true;
    //for (int32 z=MINUS2; z<=PLUS2; z++) {
    //   for (int32 y=MINUS2; y<=PLUS2; y++) {
    //      for (int32 x=MINUS2; x<=PLUS2; x++) {
    //         canInterpolate &= getDeformIndex(lowerGridPoint+makeVector3i(x+OFFSET, y+OFFSET, z+OFFSET), index[x][y][z]);
    //      }
    //   }
    //}


    //if (!canInterpolate) {
    //   #ifdef XGRT_DEBUG
    //   throw ERangeCheck("TSDeformationContext::TimeStep::getPosition() - position out of grid range.");
    //   #endif
    //   return;
    //} else {
    //   Vector3f lowerGridCellCenter = computeCenterOfCell(lowerGridPoint);
    //   Vector3f frac = (footpoint - lowerGridCellCenter)/gridSpacing;
    //   x4Assert(frac[0] >= 0 && frac[0] <=1 && frac[1] >= 0 && frac[1] <=1 && frac[2] >= 0 && frac[2] <=1);

    //   StaticVector<int32, SUPPORT*SUPPORT*SUPPORT*3> indexMapping;
    //   {
    //      card32 i=0;
    //      for (int32 z=MINUS2; z<=PLUS2; z++) {
    //         for (int32 y=MINUS2; y<=PLUS2; y++) {
    //            for (int32 x=MINUS2; x<=PLUS2; x++) {
    //               for (card32 d=0; d<3; d++) {
    //                  indexMapping[i] = index[x][y][z]*3 + (i%3);
    //                  i++;
    //               }
    //            }
    //         }
    //      }
    //   }
    //   
    //   StaticMatrix<float32, SUPPORT*SUPPORT*SUPPORT*3, SUPPORT*SUPPORT*SUPPORT*3> localMat;
    //   StaticVector<float32, SUPPORT*SUPPORT*SUPPORT*3> localRHS;
    //
    //   computeConstrTermSmoothed(frac, constr->toPoint, constr->covariance, localMat, localRHS);

    //   scatterAddMat1<float32, SUPPORT*SUPPORT*SUPPORT*3>(mat, localMat, indexMapping);
    //   scatterAddVec1<float32, SUPPORT*SUPPORT*SUPPORT*3>(vec, localRHS, indexMapping);
    //}
}

void TSDeformationContext::TimeStep::addSpatialDerivativeConstraint(const SpatialDerivativeARConstraint *constr, SparseMatrixF &mat, DVectorF &vec)
{
    notImplemented();
}

static HessianTools<float32, 4>::DiffTwice gradConstrEnergy(
    HessianTools<float32, 4>::DiffTwice p000,
    HessianTools<float32, 4>::DiffTwice p001,
    HessianTools<float32, 4>::DiffTwice p010,
    HessianTools<float32, 4>::DiffTwice p100,
    Vector3f grad,
    float32 weight,
    float32 gridSpacing)
{
    HessianTools<float32, 4>::DiffTwice scale(1.0f/gridSpacing);
    return sqr((p001 - p000)*scale-HessianTools<float32, 4>::DiffTwice(grad[2]))
        + sqr((p010 - p000)*scale-HessianTools<float32, 4>::DiffTwice(grad[1]))
        + sqr((p100 - p000)*scale-HessianTools<float32, 4>::DiffTwice(grad[0]));
}

void TSDeformationContext::TimeStep::addSimpleSpatialDerivativeConstraint(const SimpleSpatialDerivativeARConstraint *constr, SparseMatrixF &mat, DVectorF &vec)
{
    Vector3f footpoint = constr->anchor.pos;
    Vector3i lowerGridPoint = makeVector3i(
        (int32)floor(footpoint[0]/gridSpacing-0.5f),
        (int32)floor(footpoint[1]/gridSpacing-0.5f),
        (int32)floor(footpoint[2]/gridSpacing-0.5f)
        );

    int32 index000, index001, index010, index100;

    bool canInterpolate = getDeformIndex(lowerGridPoint, index000);
    canInterpolate &= getDeformIndex(lowerGridPoint+makeVector3i(0,0,1), index001);
    canInterpolate &= getDeformIndex(lowerGridPoint+makeVector3i(0,1,0), index010);
    canInterpolate &= getDeformIndex(lowerGridPoint+makeVector3i(1,0,0), index100);
    if (canInterpolate) {
        Vector4i indexMapping;
        indexMapping[0] = index000;
        indexMapping[1] = index001;
        indexMapping[2] = index010;
        indexMapping[3] = index100;

        Matrix4f localMat = makeMatrix4f(0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0);
        Vector4f localVec = NULL_VECTOR4F;

        for (card32 d=0; d<3; d++) {
            HessianTools<float32, 4>::DiffTwice p000 = HessianTools<float32, 4>::createVariable(0.0f, 0);
            HessianTools<float32, 4>::DiffTwice p001 = HessianTools<float32, 4>::createVariable(0.0f, 1);
            HessianTools<float32, 4>::DiffTwice p010 = HessianTools<float32, 4>::createVariable(0.0f, 2);
            HessianTools<float32, 4>::DiffTwice p100 = HessianTools<float32, 4>::createVariable(0.0f, 3);
            HessianTools<float32, 4>::DiffTwice result = gradConstrEnergy(p000, p001, p010, p100, constr->targetDerivative[d], constr->columnWeights[d], gridSpacing);
            localMat = HessianTools<float32, 4>::getHessian(result);
            localVec = HessianTools<float32, 4>::getGradient(result);
            scatterAddMat3SingleD<float32, 4>(mat, localMat, indexMapping, d);
            scatterAddVec3SingleD<float32, 4>(vec, localVec, indexMapping, d);
        }      
    } else {
#ifdef XGRT_DEBUG
        warning("TSDeformationContext::TimeStep::addSimpleSpatialDerivativeConstraint() - footpoint out of range.");
#endif
    }   
}

void TSDeformationContext::TimeStep::setResult(const DVectorF &result)
{
    last_movement = 0;
    const card32 numElems = (card32)gridEntries.size();
    x4Assert(numElems == result.getDim()/3);
    for (card32 i=0; i<numElems; i++) {
        Vector3f newPos = makeVector3f(result[3*i], result[3*i+1], result[3*i+2]);
        last_movement = max(last_movement,norm(newPos - gridEntries[i].deformPos));
        x4Assert(i < gridEntries.size());
        gridEntries[i].deformPos = newPos;
    }
}

void TSDeformationContext::TimeStep::getLastResult(DVectorF &result) {
    const card32 numElems = (card32)gridEntries.size();
    result.setDim(numElems*3);
    for (card32 i=0; i<numElems; i++) {
        result[3*i]   = gridEntries[i].deformPos[0];
        result[3*i+1] = gridEntries[i].deformPos[1];
        result[3*i+2] = gridEntries[i].deformPos[2];
    }
}


Vector3f TSDeformationContext::getPosition(int32 time, const ARConstraintAnchor &at) 
{
    x4Assert(time >= 0 && time <= endTime);
    return timeSteps[time]->getPosition(at);
}

Matrix3f TSDeformationContext::getSpatialGradient(int32 time, const ARConstraintAnchor &at) 
{
    x4Assert(time >= 0 && time <= endTime);
    return timeSteps[time]->getSpatialGradient(at);
}

Vector3f TSDeformationContext::getPositionSmoothed(int32 time, const ARConstraintAnchor &at) 
{
    x4Assert(time >= 0 && time <= endTime);
    return timeSteps[time]->getPositionSmoothed(at);
}

Matrix3f TSDeformationContext::getSpatialGradientSmoothed(int32 time, const ARConstraintAnchor &at) 
{
    x4Assert(time >= 0 && time <= endTime);
    return timeSteps[time]->getSpatialGradientSmoothed(at);
}

Vector3f TSDeformationContext::TimeStep::getPosition(const ARConstraintAnchor &at) 
{
    Vector3f queryPos = at.pos;
    Vector3i lowerGridPoint = makeVector3i(
        (int32)floor(at.pos[0]/gridSpacing-0.5f),
        (int32)floor(at.pos[1]/gridSpacing-0.5f),
        (int32)floor(at.pos[2]/gridSpacing-0.5f)
        );
    Vector3f pos000, pos001, pos010, pos011,
        pos100, pos101, pos110, pos111;

    if (!getDeformPos(lowerGridPoint, pos000)) {
#ifdef XGRT_DEBUG
        throw ERangeCheck("TSDeformationContext::TimeStep::getPosition() - position out of grid range.");
#else
        return at.pos;
#endif
    }
    bool canInterpolate = true;
    canInterpolate &= getDeformPos(lowerGridPoint+makeVector3i(0,0,1), pos001);
    canInterpolate &= getDeformPos(lowerGridPoint+makeVector3i(0,1,0), pos010);
    canInterpolate &= getDeformPos(lowerGridPoint+makeVector3i(0,1,1), pos011);
    canInterpolate &= getDeformPos(lowerGridPoint+makeVector3i(1,0,0), pos100);
    canInterpolate &= getDeformPos(lowerGridPoint+makeVector3i(1,0,1), pos101);
    canInterpolate &= getDeformPos(lowerGridPoint+makeVector3i(1,1,0), pos110);
    canInterpolate &= getDeformPos(lowerGridPoint+makeVector3i(1,1,1), pos111);

    if (canInterpolate) {
        Vector3f lowerGridCellCenter = computeCenterOfCell(lowerGridPoint);
        Vector3f frac = (queryPos - lowerGridCellCenter)/gridSpacing;
        x4Assert(frac[0] >= 0 && frac[0] <=1 && frac[1] >= 0 && frac[1] <=1 && frac[2] >= 0 && frac[2] <=1);

        return (   ( pos000*(1-frac[0]) + pos100*frac[0] )*(1-frac[1])  // 000  100   y=0  z=0
            + ( pos010*(1-frac[0]) + pos110*frac[0] )*  frac[1]    // 010  110   y=1  z=0
        ) * (1-frac[2])
            + (   ( pos001*(1-frac[0]) + pos101*frac[0] )*(1-frac[1])  // 001  101   y=0  z=1
            + ( pos011*(1-frac[0]) + pos111*frac[0] )*  frac[1]    // 011  111   y=1  z=1
        ) * frac[2];
    } else {                                                                    
#ifdef XGRT_DEBUG
        warning("interpolation failed.");
#endif
        return pos000;
    }
}

template <typename FloatType>
inline FloatType weightingFunction(FloatType d, card32 tssBasisFunction) {
    static const FloatType gaussianOff = exp(-5.0f);
    if (d<0) d=-d;
    if (d>1) d=1;
    switch (tssBasisFunction) {
    case TSS_COSINE: return cos(d*M_PI)/2.0f+0.5f;
    case TSS_GAUSSIAN: return exp(-d*d*5.0f) - gaussianOff;
    case TSS_POLYSPLINE: return pow((1.0f-d*d),4.0f);
    default: return 1;
    }
}

inline Vector3f weightingFunctionGrad(Vector3f delta, card32 tssBasisFunction) {
    float32 d=normQuad(delta);
    if (d>1) return NULL_VECTOR3F;
    float32 x = delta[0];
    float32 y = delta[1];
    float32 z = delta[2];
    switch (tssBasisFunction) {
    case TSS_COSINE: {
        float32 v = -0.5*sin(sqrt(sqr(x)+sqr(y)+sqr(z))*M_PI)/sqrt(sqr(x)+sqr(y)+sqr(z))*M_PI;
        Vector3f result = -makeVector3f(v*x,v*y,v*z);
        return result;      
                     }
    case TSS_GAUSSIAN: {
        float32 v = -10*exp(-5*(sqr(x)+sqr(y)+sqr(z)));
        Vector3f result = -makeVector3f(v*x,v*y,v*z);
        return result;      
                       }
    case TSS_POLYSPLINE: {
        float32 v = -8*pow(1-(sqr(x)+sqr(y)+sqr(z)),3);
        Vector3f result = -makeVector3f(v*x,v*y,v*z);
        return result;      
                         }
    default: return delta;
    }
}

Vector3f TSDeformationContext::TimeStep::getPositionSmoothed(const ARConstraintAnchor &at) 
{
    Vector3f queryPos = at.pos;
    Vector3i lowerGridPoint = makeVector3i(
        (int32)floor(at.pos[0]/gridSpacing-0.5f),
        (int32)floor(at.pos[1]/gridSpacing-0.5f),
        (int32)floor(at.pos[2]/gridSpacing-0.5f)
        );

    Vector3f pos[7][7][7];
    const int32 MINUS2 = 0;
    const int32 PLUS2 = 6;
    const int32 OFFSET = -3;

    bool canInterpolate = true;
    for (int32 z= MINUS2; z<=PLUS2; z++) {
        for (int32 y= MINUS2; y<=PLUS2; y++) {
            for (int32 x= MINUS2; x<=PLUS2; x++) {
                canInterpolate &= getDeformPos(lowerGridPoint+makeVector3i(x+OFFSET, y+OFFSET, z+OFFSET), pos[x][y][z]);
            }
        }
    }

    if (canInterpolate) {
        Vector3f lowerGridCellCenter = computeCenterOfCell(lowerGridPoint);
        Vector3f frac = (queryPos - lowerGridCellCenter)/gridSpacing;

        float32 wSum = 0;
        Vector3f result = NULL_VECTOR3F;

        for (int32 z= MINUS2; z<=PLUS2; z++) {
            for (int32 y= MINUS2; y<=PLUS2; y++) {
                for (int32 x= MINUS2; x<=PLUS2; x++) {
                    Vector3i cell = lowerGridPoint+makeVector3i(x+OFFSET, y+OFFSET, z+OFFSET);
                    int32 index = 0;
                    if (!grid.getObject(cell, index)) throw X4Exception("this should never happen");
                    Vector3f funPos = (makeVector3f(cell[0],cell[1], cell[2])+makeVector3f(0.5f, 0.5f, 0.5f))*gridSpacing;
                    Vector3f dest = gridEntries[index].deformPos;
                    float32 w = weightingFunction<float32>(norm(funPos - queryPos)/(gridSpacing*3), settings->tssBasisFunction);
                    result += dest*w;
                    wSum += w;
                }
            }
        }
        if (wSum >= 1e-7) result /= wSum;

        return result;      

    } else {                                                                    
#ifdef XGRT_DEBUG
        warning("interpolation failed.");
#endif
        return NULL_VECTOR3F;
    }
}

Matrix3f TSDeformationContext::TimeStep::getSpatialGradient(const ARConstraintAnchor &at) 
{
    Vector3f queryPos = at.pos;
    Vector3i lowerGridPoint = makeVector3i(
        (int32)floor(at.pos[0]/gridSpacing-0.5f),
        (int32)floor(at.pos[1]/gridSpacing-0.5f),
        (int32)floor(at.pos[2]/gridSpacing-0.5f)
        );
    Vector3f pos000, pos001, pos010, pos100;

    bool canInterpolate = getDeformPos(lowerGridPoint, pos000);
    canInterpolate &= getDeformPos(lowerGridPoint+makeVector3i(0,0,1), pos001);
    canInterpolate &= getDeformPos(lowerGridPoint+makeVector3i(0,1,0), pos010);
    canInterpolate &= getDeformPos(lowerGridPoint+makeVector3i(1,0,0), pos100);

    if (canInterpolate) {
        Matrix3f result;
        result[0] = (pos100-pos000)/gridSpacing;
        result[1] = (pos010-pos000)/gridSpacing;
        result[2] = (pos001-pos000)/gridSpacing;
        return result;
    } else {                                                                    
#ifdef XGRT_DEBUG
        warning("interpolation failed.");
#endif
        return IDENTITY3F;
    }
}

Matrix3f TSDeformationContext::TimeStep::getSpatialGradientSmoothed(const ARConstraintAnchor &at) 
{
    Vector3f queryPos = at.pos;
    Vector3i lowerGridPoint = makeVector3i(
        (int32)floor(at.pos[0]/gridSpacing-0.5f),
        (int32)floor(at.pos[1]/gridSpacing-0.5f),
        (int32)floor(at.pos[2]/gridSpacing-0.5f)
        );

    Vector3f pos[7][7][7];
    const int32 MINUS2 = 0;
    const int32 PLUS2 = 6;
    const int32 OFFSET = -3;

    bool canInterpolate = true;
    for (int32 z= MINUS2; z<=PLUS2; z++) {
        for (int32 y= MINUS2; y<=PLUS2; y++) {
            for (int32 x= MINUS2; x<=PLUS2; x++) {
                canInterpolate &= getDeformPos(lowerGridPoint+makeVector3i(x+OFFSET, y+OFFSET, z+OFFSET), pos[x][y][z]);
            }
        }
    }

    if (canInterpolate) {
        Vector3f lowerGridCellCenter = computeCenterOfCell(lowerGridPoint);
        Vector3f frac = (queryPos - lowerGridCellCenter)/gridSpacing;

        float32 wSum = 0;
        Matrix3f result = makeMatrix3f(0,0,0, 0,0,0, 0,0,0);

        for (int32 z= MINUS2; z<=PLUS2; z++) {
            for (int32 y= MINUS2; y<=PLUS2; y++) {
                for (int32 x= MINUS2; x<=PLUS2; x++) {
                    Vector3i cell = lowerGridPoint+makeVector3i(x+OFFSET, y+OFFSET, z+OFFSET);
                    int32 index = 0;
                    if (!grid.getObject(cell, index)) throw X4Exception("this should never happen");
                    Vector3f funPos = (makeVector3f(cell[0],cell[1], cell[2])+makeVector3f(0.5f, 0.5f, 0.5f))*gridSpacing;
                    ARConstraintAnchor atIndex;
                    atIndex.topologicalAnchor = ARConstraintAnchor::NO_TOP_ANCHOR;
                    atIndex.pos = funPos;// lowerGridCellCenter+(makeVector3f(x+OFFSET, y+OFFSET, z+OFFSET)*gridSpacing);
                    Matrix3f dest = getSpatialGradient(atIndex);
                    float32 w = weightingFunction<float32>(norm(funPos - queryPos)/(gridSpacing*3), settings->tssBasisFunction);
                    result += dest*w;
                    wSum += w;
                }
            }
        }
        if (wSum >= 1e-7) result /= wSum;

        return result;      

    } else {                                                                    
#ifdef XGRT_DEBUG
        warning("interpolation failed.");
#endif
        return IDENTITY3F;
    }
}

//Matrix3f TSDeformationContext::TimeStep::getSpatialGradientSmoothed(const ARConstraintAnchor &at) 
//{
//   Vector3f queryPos = at.pos;
//   Vector3i lowerGridPoint = makeVector3i(
//      (int32)floor(at.pos[0]/gridSpacing-0.5f),
//      (int32)floor(at.pos[1]/gridSpacing-0.5f),
//      (int32)floor(at.pos[2]/gridSpacing-0.5f)
//   );
//   
//   Vector3f pos[7][7][7];
//   const int32 MINUS2 = 0;
//   const int32 PLUS2 = 6;
//   const int32 OFFSET = -3;
//   
//   bool canInterpolate = true;
//   for (int32 z= MINUS2; z<=PLUS2; z++) {
//      for (int32 y= MINUS2; y<=PLUS2; y++) {
//         for (int32 x= MINUS2; x<=PLUS2; x++) {
//            canInterpolate &= getDeformPos(lowerGridPoint+makeVector3i(x+OFFSET, y+OFFSET, z+OFFSET), pos[x][y][z]);
//         }
//      }
//   }
//   
//   if (canInterpolate) {
//      Vector3f lowerGridCellCenter = computeCenterOfCell(lowerGridPoint);
//      Vector3f frac = (queryPos - lowerGridCellCenter)/gridSpacing;
//      
//      float32 wSum = 0;
//      Matrix3f result = makeMatrix3f(0,0,0, 0,0,0, 0,0,0);
//
//      for (int32 z= MINUS2; z<=PLUS2; z++) {
//         for (int32 y= MINUS2; y<=PLUS2; y++) {
//            for (int32 x= MINUS2; x<=PLUS2; x++) {
//               Vector3i cell = lowerGridPoint+makeVector3i(x+OFFSET, y+OFFSET, z+OFFSET);
//               int32 index = 0;
//               if (!grid.getObject(cell, index)) throw X4Exception("this should never happen");
//               Vector3f funPos = (makeVector3f(cell[0],cell[1], cell[2])+makeVector3f(0.5f, 0.5f, 0.5f))*gridSpacing;
//               Vector3f grad = weightingFunctionGrad((funPos - queryPos)/(gridSpacing*3), settings->tssBasisFunction);
//               Vector3f dest = gridEntries[index].deformPos;
//               for (card32 d=0; d<3; d++) {
//                  result[d] += grad*dest[d];
//               }
//               float32 w = weightingFunction<float32>(norm(funPos - queryPos)/(gridSpacing*3), settings->tssBasisFunction);
//               wSum += w;
//            }
//         }
//      }
//      if (wSum >= 1e-7) result /= wSum;
//
//      return result;      
//
//   } else {                                                                    
//      #ifdef XGRT_DEBUG
//      warning("interpolation failed.");
//      #endif
//      return makeMatrix3f(0,0,0, 0,0,0, 0,0,0);
//   }
//}

float32 TSDeformationContext::getBaseResolution() 
{
    return settings->tsGridSpacing;
}

void X4::TSDeformationContext::TimeStep::addCachedPositionConstraint( const PositionARConstraint *constr, SparseMatrixF &mat, DVectorF &vec )
{
    Vector3f footpoint = constr->anchor.pos;
    footpoint[2] = settings->tsMode2D?0:footpoint[2];
    Vector3i lowerGridPoint = makeVector3i(
        (int32)floor(footpoint[0]/gridSpacing-0.5f),
        (int32)floor(footpoint[1]/gridSpacing-0.5f),
        settings->tsMode2D?0:(int32)floor(footpoint[2]/gridSpacing-0.5f)
        );

    int32 index000, index001, index010, index011,
        index100, index101, index110, index111;

    if (!getDeformIndex(lowerGridPoint, index000)) {
#ifdef XGRT_DEBUG
        throw ERangeCheck("TSDeformationContext::TimeStep::getPosition() - position out of grid range.");
#endif
        return;
    }
    bool canInterpolate = true;
    canInterpolate &= getDeformIndex(lowerGridPoint+makeVector3i(0,0,1), index001);
    canInterpolate &= getDeformIndex(lowerGridPoint+makeVector3i(0,1,0), index010);
    canInterpolate &= getDeformIndex(lowerGridPoint+makeVector3i(0,1,1), index011);
    canInterpolate &= getDeformIndex(lowerGridPoint+makeVector3i(1,0,0), index100);
    canInterpolate &= getDeformIndex(lowerGridPoint+makeVector3i(1,0,1), index101);
    canInterpolate &= getDeformIndex(lowerGridPoint+makeVector3i(1,1,0), index110);
    canInterpolate &= getDeformIndex(lowerGridPoint+makeVector3i(1,1,1), index111);

    if (!canInterpolate) {
        Vector3i constrCell = computeGridCell(constr->anchor.pos);
        Vector3f target = constr->toPoint;
        int32 index;
        if (grid.getObject(constrCell, index)) {
            for (card32 d=0; d<3; d++) {
                mat[index*3+d].addEntry(3*index+d, 2*settings->tsConstraintsWeight);
                vec[index*3+d] += 2*target[d]*settings->tsConstraintsWeight;
            }
        }
    } else {
        Vector3f lowerGridCellCenter = computeCenterOfCell(lowerGridPoint);
        Vector3f frac = (footpoint - lowerGridCellCenter)/gridSpacing;
        x4Assert(frac[0] >= 0 && frac[0] <=1 && frac[1] >= 0 && frac[1] <=1 && frac[2] >= 0 && frac[2] <=1);

        StaticVector<int32, 24> indexMapping;
        indexMapping[ 0] = index000*3;
        indexMapping[ 1] = index000*3 + 1;
        indexMapping[ 2] = index000*3 + 2;

        indexMapping[ 3] = index001*3;
        indexMapping[ 4] = index001*3 + 1;
        indexMapping[ 5] = index001*3 + 2;

        indexMapping[ 6] = index010*3;
        indexMapping[ 7] = index010*3 + 1;
        indexMapping[ 8] = index010*3 + 2;

        indexMapping[ 9] = index011*3;
        indexMapping[10] = index011*3 + 1;
        indexMapping[11] = index011*3 + 2;

        indexMapping[12] = index100*3;
        indexMapping[13] = index100*3 + 1;
        indexMapping[14] = index100*3 + 2;

        indexMapping[15] = index101*3;
        indexMapping[16] = index101*3 + 1;
        indexMapping[17] = index101*3 + 2;

        indexMapping[18] = index110*3;
        indexMapping[19] = index110*3 + 1;
        indexMapping[20] = index110*3 + 2;

        indexMapping[21] = index111*3;
        indexMapping[22] = index111*3 + 1;
        indexMapping[23] = index111*3 + 2;

        StaticMatrix<float32, 24, 24> localMat;
        StaticVector<float32, 24> localRHS;
        Vector3f constrToPoint = constr->toPoint;
        constrToPoint[2] = settings->tsMode2D?0:constrToPoint[2];
        if (settings->tsSlowTest) {
            computeConstrTermAD(frac, constrToPoint, constr->covariance,
                localMat, localRHS);
        } else {
            computeConstrTerm(frac, constrToPoint, constr->covariance,
                localMat, localRHS);
        }
        //scatterAddMat1(mat, localMat, indexMapping);

        // add to cached matrix
        cachedScatterMatrix[index000] += localMat;

        scatterAddVec1(vec, localRHS, indexMapping);
    }
}

void X4::TSDeformationContext::TimeStep::transferCachedPositionConstraints( SparseMatrixF &mat )
{
    for( unsigned i=0;i<cachedScatterMatrix.size();i++ )
    {
        GridEntry entry = gridEntries[i];
        Vector3i lowerGridPoint = entry.gridCell;

        int32 index000, index001, index010, index011,
            index100, index101, index110, index111;

        if (!getDeformIndex(lowerGridPoint, index000)) {
#ifdef XGRT_DEBUG
            throw ERangeCheck("TSDeformationContext::TimeStep::getPosition() - position out of grid range.");
#endif
            return;
        }
        bool canInterpolate = true;
        canInterpolate &= getDeformIndex(lowerGridPoint+makeVector3i(0,0,1), index001);
        canInterpolate &= getDeformIndex(lowerGridPoint+makeVector3i(0,1,0), index010);
        canInterpolate &= getDeformIndex(lowerGridPoint+makeVector3i(0,1,1), index011);
        canInterpolate &= getDeformIndex(lowerGridPoint+makeVector3i(1,0,0), index100);
        canInterpolate &= getDeformIndex(lowerGridPoint+makeVector3i(1,0,1), index101);
        canInterpolate &= getDeformIndex(lowerGridPoint+makeVector3i(1,1,0), index110);
        canInterpolate &= getDeformIndex(lowerGridPoint+makeVector3i(1,1,1), index111);

        if (!canInterpolate) {
            //throw X4Exception("TSDeformationContext::TimeStep::transferCachedPositionConstraints() - grid cell out of range.");
        } else {

            StaticVector<int32, 24> indexMapping;
            indexMapping[ 0] = index000*3;
            indexMapping[ 1] = index000*3 + 1;
            indexMapping[ 2] = index000*3 + 2;

            indexMapping[ 3] = index001*3;
            indexMapping[ 4] = index001*3 + 1;
            indexMapping[ 5] = index001*3 + 2;

            indexMapping[ 6] = index010*3;
            indexMapping[ 7] = index010*3 + 1;
            indexMapping[ 8] = index010*3 + 2;

            indexMapping[ 9] = index011*3;
            indexMapping[10] = index011*3 + 1;
            indexMapping[11] = index011*3 + 2;

            indexMapping[12] = index100*3;
            indexMapping[13] = index100*3 + 1;
            indexMapping[14] = index100*3 + 2;

            indexMapping[15] = index101*3;
            indexMapping[16] = index101*3 + 1;
            indexMapping[17] = index101*3 + 2;

            indexMapping[18] = index110*3;
            indexMapping[19] = index110*3 + 1;
            indexMapping[20] = index110*3 + 2;

            indexMapping[21] = index111*3;
            indexMapping[22] = index111*3 + 1;
            indexMapping[23] = index111*3 + 2;

            scatterAddMat1(mat, cachedScatterMatrix[i], indexMapping);

        }

    }
    cachedScatterMatrix.clear();
}

void X4::TSDeformationContext::TimeStep::drawGridWithDebugRenderer()
{
    //Vector3f col = shrink4To3(rainbowGradient(rnd01()));
    static int frame = 0;
    for( unsigned i=0;i<gridEntries.size();i++ )
    {
        GridEntry e = gridEntries[i];
        Vector3i lowerGridPoint = e.gridCell;

        int32 index000, index001, index010, index011,
            index100, index101, index110, index111;

        bool canDraw = true;
        canDraw &= getDeformIndex(lowerGridPoint                    , index000);
        canDraw &= getDeformIndex(lowerGridPoint+makeVector3i(0,0,1), index001);
        canDraw &= getDeformIndex(lowerGridPoint+makeVector3i(0,1,0), index010);
        canDraw &= getDeformIndex(lowerGridPoint+makeVector3i(0,1,1), index011);
        canDraw &= getDeformIndex(lowerGridPoint+makeVector3i(1,0,0), index100);
        canDraw &= getDeformIndex(lowerGridPoint+makeVector3i(1,0,1), index101);
        canDraw &= getDeformIndex(lowerGridPoint+makeVector3i(1,1,0), index110);
        canDraw &= getDeformIndex(lowerGridPoint+makeVector3i(1,1,1), index111);

        if( canDraw && e.containsUrshapePoints )
        {

            Vector3f p000 = e.deformPos;
            Vector3f p001 = gridEntries[index001].deformPos;
            Vector3f p010 = gridEntries[index010].deformPos;
            Vector3f p011 = gridEntries[index011].deformPos;
            Vector3f p100 = gridEntries[index100].deformPos;
            Vector3f p101 = gridEntries[index101].deformPos;
            Vector3f p110 = gridEntries[index110].deformPos;
            Vector3f p111 = gridEntries[index111].deformPos;

            //Vector3f color = makeVector3f(0,1,1);
            Matrix3f m;
            m[0] = p001 - p000;
            m[1] = p010 - p000;
            m[2] = p100 - p000;

            float q = determinant(m)*10000;
            //debugOutput << q << " ";
            //Vector3f color = shrink4To3(blueRedGradient(q,1));
            Vector3f color;
            if (q<0) color = makeVector3f(1,0,0); else color = makeVector3f(0,0,1);
            Vector3f color2 = color;
            color2[1] = 0.8f;
            color2[2] = 0.8f;
            float32 l = min(fabs(q),1.f);
            color = color2*l + color*(1-l);


            Vector4f color4 = makeVector4f(0.8f,0.2f,0.2f, 0.6f);
            debugRenderer->setSmoothLine(true);
            debugRenderer->setEnableBlending(true);
            float width = 3.0f;
            debugRenderer->addLine(p000,p001,color4,color4,width);
            debugRenderer->addLine(p001,p011,color4,color4,width);
            debugRenderer->addLine(p011,p010,color4,color4,width);
            debugRenderer->addLine(p010,p000,color4,color4,width);

            debugRenderer->addLine(p100,p101,color4,color4,width);
            debugRenderer->addLine(p101,p111,color4,color4,width);
            debugRenderer->addLine(p111,p110,color4,color4,width);
            debugRenderer->addLine(p110,p100,color4,color4,width);

            debugRenderer->addLine(p000,p100,color4,color4,width);
            debugRenderer->addLine(p001,p101,color4,color4,width);
            debugRenderer->addLine(p010,p110,color4,color4,width);
            debugRenderer->addLine(p011,p111,color4,color4,width);
        }

    }
    debugOutput.cr();
}


void X4::TSDeformationContext::addCachedPositionConstraint( int time, const PositionARConstraint *constr,  SparseMatrixF &mat, DVectorF &vec )
{
    time += startTime;
    x4Assert(time >= 0 && time <= endTime);
    timeSteps[time]->addCachedPositionConstraint(constr, mat, vec);
}

void X4::TSDeformationContext::transferCachedPositionConstraints( int time, SparseMatrixF &mat )
{
    time += startTime;
    x4Assert(time >= 0 && time <= endTime);
    timeSteps[time]->transferCachedPositionConstraints(mat);
}

void X4::TSDeformationContext::drawGridWithDebugRenderer( int time /*= 0*/ )
{
    time += startTime;
    x4Assert(time >= 0 && time <= endTime);
    timeSteps[time]->drawGridWithDebugRenderer();
}

X4::float32 X4::TSDeformationContext::getMovement( int time )
{
    time += startTime;
    x4Assert(time >= 0 && time <= endTime);
    return timeSteps[time]->getLastMovement();

}

bool X4::TSDeformationContext::badDeformation( float threshold ) const
{
    return timeSteps[0]->badDeformation(threshold);
}
void TSDeformationContext::write(OutputObjectStream *s) const 
{
    s->writeObject(settings);
    s->write(startTime);
    s->write(endTime);
    s->write(gridExtension);
    mpcard numTimeSteps = timeSteps.size();
    s->writeMaxPlatformCard(numTimeSteps);
    for (size_t i=0; i<timeSteps.size(); i++) {
        timeSteps[i]->write(s);
    }
}

void TSDeformationContext::read(InputObjectStream *s) 
{
    s->readAnyObject(settings);
    s->read(startTime);
    s->read(endTime);
    //if (s->getClassVersion(TSDeformationContext::getClass()) >= MetaClass::FIRST_VERSION + 1) {
    s->read(gridExtension);
    //}
    for (size_t i=0; i<timeSteps.size(); i++) {
        delete timeSteps[i];
    }   
    timeSteps.clear();
    mpcard numTimeSteps;
    s->readMaxPlatformCard(numTimeSteps);
    timeSteps.resize(numTimeSteps, NULL);
    for (size_t i=0; i<timeSteps.size(); i++) {
        timeSteps[i] = new TimeStep(settings, gridExtension);
    }
    for (size_t i=0; i<timeSteps.size(); i++) {
        timeSteps[i]->read(s);
    }
}

TSDeformationContext::TSDeformationContext( const TSDeformationContext &other)
{
    settings = other.settings;
    startTime = other.startTime;
    endTime = other.endTime;
    gridExtension = other.gridExtension;
    mpcard numTimeSteps = other.timeSteps.size();
    timeSteps.resize(numTimeSteps, NULL);
    for (size_t i=0; i<numTimeSteps; i++)
        timeSteps[i] = new TimeStep(*other.timeSteps[i]);
}

void TSDeformationContext::TimeStep::write(OutputObjectStream *s) const 
{
    s->write(gridSpacing);
    s->writeObject(settings);
    s->write(gridExtension);
    mpcard numEntries = gridEntries.size();
    s->writeMaxPlatformCard(numEntries);
    for (mpcard i=0; i<numEntries; i++) {
        gridEntries[i].write(s);
    }
}

void TSDeformationContext::TimeStep::read(InputObjectStream *s) 
{
    s->read(gridSpacing);
    s->readAnyObject(settings);
    if (s->getClassVersion(TSDeformationContext::getClass()) >= MetaClass::FIRST_VERSION + 1) {
        s->read(gridExtension);
    }
    mpcard numEntries;
    s->readMaxPlatformCard(numEntries);
    gridEntries.resize(numEntries);
    for (mpcard i=0; i<numEntries; i++) {
        gridEntries[i].read(s);
        grid.setObject(gridEntries[i].gridCell, (int32)i);
    }
}

X4::TSDeformationContext::TimeStep::TimeStep( const TimeStep &other )
{
    gridSpacing = other.gridSpacing;
    //settings->assign(other.settings);
    settings = other.settings;	// const ptr...
    gridExtension = other.gridExtension;
    mpcard numEntries = other.gridEntries.size();
    gridEntries.resize(numEntries);
    for( unsigned i=0;i<numEntries;i++ )
    {
        gridEntries[i] = other.gridEntries[i];
        grid.setObject(gridEntries[i].gridCell, (int32)i);
    }
}

bool X4::TSDeformationContext::TimeStep::badDeformation( float threshold )
{
    unsigned numPositiveDeterminants = 0;
    unsigned numNegativeDeterminants = 0;
    unsigned numMeasuredCells = 0;
    float distortion = 0;

    for( unsigned i=0;i<gridEntries.size();i++ )
    {
        GridEntry e = gridEntries[i];
        Vector3i lowerGridPoint = e.gridCell;

        int32 index000, index001, index010, index011,
            index100, index101, index110, index111;

        bool canDraw = true;
        canDraw &= getDeformIndex(lowerGridPoint                    , index000);
        canDraw &= getDeformIndex(lowerGridPoint+makeVector3i(0,0,1), index001);
        canDraw &= getDeformIndex(lowerGridPoint+makeVector3i(0,1,0), index010);
        canDraw &= getDeformIndex(lowerGridPoint+makeVector3i(0,1,1), index011);
        canDraw &= getDeformIndex(lowerGridPoint+makeVector3i(1,0,0), index100);
        canDraw &= getDeformIndex(lowerGridPoint+makeVector3i(1,0,1), index101);
        canDraw &= getDeformIndex(lowerGridPoint+makeVector3i(1,1,0), index110);
        canDraw &= getDeformIndex(lowerGridPoint+makeVector3i(1,1,1), index111);

        if( canDraw && e.containsUrshapePoints )
        {
            numMeasuredCells++;
            Vector3f p000 = e.deformPos;
            Vector3f p001 = gridEntries[index001].deformPos;
            Vector3f p010 = gridEntries[index010].deformPos;
            Vector3f p011 = gridEntries[index011].deformPos;
            Vector3f p100 = gridEntries[index100].deformPos;
            Vector3f p101 = gridEntries[index101].deformPos;
            Vector3f p110 = gridEntries[index110].deformPos;
            Vector3f p111 = gridEntries[index111].deformPos;

            //Vector3f color = makeVector3f(0,1,1);
            Matrix3f m;
            m[0] = p001 - p000;
            m[1] = p010 - p000;
            m[2] = p100 - p000;

            m /= gridSpacing;

            float det = determinant(m);
            if( det > 0 )
                numPositiveDeterminants++;
            else
                numNegativeDeterminants++;

            //debugOutput << "det: " << det << "\n";
            distortion += fabs(det) < 1.0f ? (1.0f - fabs(det)):(1.0f - 1.0f / fabs(det));
        }

    }

    if( (float)numNegativeDeterminants/numMeasuredCells > 0.2f && (float)numPositiveDeterminants/numMeasuredCells > 0.2f )
    {
        debugOutput << "Rejecting deformation (wrong determinants): Positive=" << (float)numNegativeDeterminants/numMeasuredCells
            << " Negative=" << (float)numPositiveDeterminants/numMeasuredCells << "\n";
        return true;
    }

    debugOutput << "distortion: " << distortion / numMeasuredCells << "\n";
    if( distortion / numMeasuredCells > threshold)
    {
        debugOutput << "Rejecting deformation (strong distortion)\n";
        return true;
    }


    return false;
}

void TSDeformationContext::TimeStep::GridEntry::write(OutputObjectStream *s) const
{
    s->write(gridCell);
    s->write(deformPos);
    s->write(containsUrshapePoints);
}

void TSDeformationContext::TimeStep::GridEntry::read(InputObjectStream *s)
{
    s->read(gridCell);
    s->read(deformPos);
    s->read(containsUrshapePoints);

}


