//---------------------------------------------------------------------------
#ifndef TSDeformationContextH
#define TSDeformationContextH
//---------------------------------------------------------------------------
#include "x3deformation.h"
//---------------------------------------------------------------------------
#include "ARDeformationOptimizationContext.h"
#include "SparseGrid.h"
#include "SparseLinearAlgebra.h"
#include "DynamicLinearAlgebra.h"
#include "ARConstraints.h"
#include "TSDeformationHelper.h"
//---------------------------------------------------------------------------

namespace X4 {

class RegularInCoreVolume;
class TSDeformationSolver;
class TSDeformationEvaluator;
class TSSmoothDeformationSolver;
class TSSmoothDeformationEvaluator;

///
///   this class stores all the data to describe a deformation as computed by the TSDeformationSolver/Evaluator
///
class X3DEFORMATION_API TSDeformationContext : public ARDeformationOptimizationContext {
	X4_CLASS( TSDeformationContext )
 protected:
   class TimeStep {
    private:         
      // --- internal data structures
      struct GridEntry {
         Vector3i gridCell;
         Vector3f deformPos;
			bool		containsUrshapePoints;
         void read(InputObjectStream *s);
         void write(OutputObjectStream *s) const;
      };
      struct TSStencil2D {
         card32 cc;

         card32 pc;
         card32 cp;
         card32 mc;
         card32 cm;

         card32 pp;
         card32 mm;
         card32 mp;
         card32 pm;

         StaticVector<int32, 9> getIndexMapping();
      };
      struct TSStencil {
         card32 ccc;

         card32 pcc;
         card32 cpc;
         card32 ccp;

         card32 mcc;
         card32 cmc;
         card32 ccm;

         card32 ppc;
         card32 cpp;
         card32 pcp;

         card32 mmc;
         card32 mcm;
         card32 mcp;
         card32 mpc;
         card32 cmm;
         card32 cmp;
         card32 pmc;
         card32 pcm;
         card32 cpm;    

         StaticVector<int32, 19> getIndexMapping();
      };
      struct LPStencil {
         card32 ccc;
         card32 pcc;
         card32 cpc;
         card32 ccp;
         Vector4i getIndexMapping();
      };

      // --- parameters
      float32 gridSpacing;
		int32 gridExtension;
      const ARSettings *settings;
      
      // --- deformation representation 
      std::vector<GridEntry> gridEntries;
      SparseGrid<3, int32, Object> grid;

      // --- temp
      float32 lastRegWeight;
      card32 lastDefModel;
      float32 lastIDWeight;
      float32 lastMix;
      SparseMatrixF lastRegMat;
		float32 last_movement;	// infinity norm of grid cell movement
		std::vector< StaticMatrix<float,24,24> > cachedScatterMatrix;

      // --- helper functions
      bool getLPStencil(card32 index, LPStencil &stencil);
      bool getTSStencil(card32 index, TSStencil &stencil);
      
      static const StaticMatrix<float32, 4,4> LP_STENCIL_MATRIX;
      static const StaticMatrix<float32, 19,19> TS_STENCIL_MATRIX;
	  static const StaticMatrix<float32, 19,19> TS_STENCIL_MATRIX2D;
    
      inline Vector3i computeGridCell(Vector3f pos);
      inline Vector3f computeCenterOfCell(Vector3i cell);
      inline bool getDeformPos(Vector3i cell, Vector3f &deformPos);
      inline bool getDeformIndex(Vector3i cell, int32 &index);
      
    public:
      TimeStep(const ARSettings *settings, int32 gridExtension);
		TimeStep( const TimeStep &other);

      virtual void write(OutputObjectStream *s) const;
      virtual void read(InputObjectStream *s);

      void addGridEntry(Vector3i cell);
      void checkAddEntryWithNeighbors(Vector3f pos);
   	virtual void initializeFromShape(const SceneObject *obj, const ARSettings *settings);
      virtual void debugOutputVis();

      void getRegularizer(SparseMatrixF &mat, DVectorF &vec);
      void addPositionConstraint(const PositionARConstraint *constr, SparseMatrixF &mat, DVectorF &vec);
      void addPositionConstraintSmoothed(const PositionARConstraint *constr, SparseMatrixF &mat, DVectorF &vec);
      void addSpatialDerivativeConstraint(const SpatialDerivativeARConstraint *constr, SparseMatrixF &mat, DVectorF &vec);
      void addSimpleSpatialDerivativeConstraint(const SimpleSpatialDerivativeARConstraint *constr, SparseMatrixF &mat, DVectorF &vec);
      void setResult(const DVectorF &result);
      void getLastResult(DVectorF &result);
      Vector3f getPosition(const ARConstraintAnchor &at);
      Matrix3f getSpatialGradient(const ARConstraintAnchor &at);
      Vector3f getPositionSmoothed(const ARConstraintAnchor &at);
      Matrix3f getSpatialGradientSmoothed(const ARConstraintAnchor &at);

		void addCachedPositionConstraint(const PositionARConstraint *constr, SparseMatrixF &mat, DVectorF &vec);
		void transferCachedPositionConstraints(SparseMatrixF &mat);
		void drawGridWithDebugRenderer();
		float32 getLastMovement() const {return last_movement;}
		bool badDeformation(float threshold);

	};
   
 private:
   const ARSettings *settings;
   int32 startTime, endTime;
   int32 gridExtension;
   std::vector<TimeStep*> timeSteps;

 protected:
   TimeStep *getTimeStep(int32 time);
    
    // --- solver/evaluator interface
   card32 getNumTimeSteps() {return endTime - startTime +1;}
   int32 getStartTime() {return startTime;}
   int32 getEndTime() {return endTime;}
   void getRegularizer(int32 time, SparseMatrixF &mat, DVectorF &vec);
   void addPositionConstraint(int32 time, const PositionARConstraint *constr, SparseMatrixF &mat, DVectorF &vec);
   void addPositionConstraintSmoothed(int32 time, const PositionARConstraint *constr, SparseMatrixF &mat, DVectorF &vec);
   void addSpatialDerivativeConstraint(int32 time, const SpatialDerivativeARConstraint *constr, SparseMatrixF &mat, DVectorF &vec);
   void addSimpleSpatialDerivativeConstraint(int32 time, const SimpleSpatialDerivativeARConstraint *constr, SparseMatrixF &mat, DVectorF &vec);
   void setResult(int32 time, const DVectorF &result);
   void getLastResult(int32 time, DVectorF &result);
   Vector3f getPosition(int32 time, const ARConstraintAnchor &at);
   Matrix3f getSpatialGradient(int32 time, const ARConstraintAnchor &at);
   Vector3f getPositionSmoothed(int32 time, const ARConstraintAnchor &at);
   Matrix3f getSpatialGradientSmoothed(int32 time, const ARConstraintAnchor &at);
    
	// cached version (each grid cell has a cached matrix)
	void addCachedPositionConstraint(int time, const PositionARConstraint *constr, SparseMatrixF &mat, DVectorF &vec);
	
	// transfer cached partial matrices to the system
	void transferCachedPositionConstraints(int time, SparseMatrixF &mat);

 public:
   // parameter: how many grid cells should surround each point
	TSDeformationContext(int32 gridExtension = 2);
	TSDeformationContext( const TSDeformationContext &other);

   virtual void write(OutputObjectStream *s) const;
   virtual void read(InputObjectStream *s);

	virtual void initializeFromShape(const SceneObject *obj, Vector2f timeInterval, const ARSettings *settings);
	virtual float32 getDuration();

	virtual DeformationSolver *createSolver();
	virtual DeformationEvaluator *createEvaluator();
	virtual float32 getBaseResolution();
	float32 getMovement(int time);

   virtual void debugOutputVis(int32 time = 0);

	// draw deformation grid with debug renderer (call debugRenderer->beginRenderJob() before calling this method)
	void drawGridWithDebugRenderer(int time = 0);

	/// checks deformation for flipped determinants and strong deformation
	bool badDeformation( float threshold ) const;

   ~TSDeformationContext();
 friend TSDeformationSolver;
 friend TSDeformationEvaluator;
 friend TSSmoothDeformationSolver;
 friend TSSmoothDeformationEvaluator;
};

void X3DEFORMATION_API testDiff_for_TS();


}

#endif