//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "TSSmoothDeformationSolver.h"
//---------------------------------------------------------------------------
#include "SGEVolumeGridDeformationSequence.h"
#include "SGEPointCloud.h"
#include "RegularVolume.h"
#include "RegularVolume.inline.h"
#include "RegularInCoreVolume.h"
#include "DynamicLinearAlgebraTools.h"
#include "DynamicLinearAlgebraTools.inline.h"
#include "BoundingBox.h"
#include "BoundingBox.inline.h"
#include "IterativeSolvers.h"
#include "IterativeSolvers.inline.h"
#include "ARSettings.h"
#include "ARConstraints.h"
#include "TSDeformationContext.h"
#include "TSSmoothDeformationEvaluator.h"
#include "Timer.h"
#include "StringHelper.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------



IMPLEMENT_X4_CLASS( TSSmoothDeformationSolver ,0)
{
	BEGIN_CLASS_INIT( TSSmoothDeformationSolver );
}

TSSmoothDeformationSolver::TSSmoothDeformationSolver()
{
	context = NULL;
	settings = NULL;
}

TSSmoothDeformationSolver::~TSSmoothDeformationSolver() {}



AROptimizationContext *TSSmoothDeformationSolver::createOptContext() {
	return new TSDeformationContext(4);
}

void TSSmoothDeformationSolver::setup(AROptimizationContext *context, const ARSettings *settings) {
	if (!dynamic_cast<TSDeformationContext*>(context)) {
		throw ETypeMismatch("TSSmoothDeformationSolver::setup() - context does not have the correct dynamic type.");
	}
	this->context = (TSDeformationContext*)context;
	this->settings = settings;
}

void TSSmoothDeformationSolver::beginConstraints() 
{
   quadTerms = std::vector<SparseMatrixF>();
   linTerms = std::vector<DVectorF>();
   
   int32 numTimeSteps = context->getNumTimeSteps();
   quadTerms.resize(numTimeSteps);
   linTerms.resize(numTimeSteps);
   
   for (int32 i=0; i<numTimeSteps; i++) {
      context->getRegularizer(context->getStartTime()+i, quadTerms[i], linTerms[i]);
   }
}


void TSSmoothDeformationSolver::addPositionConstraint(const PositionARConstraint *c) {
   int32 t = (int32)c->getTime();
   if (t > context->getEndTime() || t<context->getStartTime()) throw ERangeCheck("time out of range");
   card32 index = t - context->getStartTime();

   context->addPositionConstraintSmoothed(t, c, quadTerms[index], linTerms[index]);
}

void TSSmoothDeformationSolver::addVelocityConstraint(const VelocityVectorARConstraint *c)
{
   notImplemented();
}

void TSSmoothDeformationSolver::addInstantaneousVelocityConstraint(const InstantaneousVelocityARConstraint *c) 
{
   notImplemented();
}

void TSSmoothDeformationSolver::addHalfSpaceConstraint( const ARHalfSpaceConstraint *c )
{
   notImplemented();
}

void TSSmoothDeformationSolver::addSpatialDerivativeConstraint(const SpatialDerivativeARConstraint *c)
{
   int32 t = (int32)c->getTime();
   if (t > context->getEndTime() || t<context->getStartTime()) throw ERangeCheck("time out of range");
   card32 index = t - context->getStartTime();

   context->addSpatialDerivativeConstraint(t, c, quadTerms[index], linTerms[index]);
}

void TSSmoothDeformationSolver::addSimpleSpatialDerivativeConstraint(const SimpleSpatialDerivativeARConstraint *c)
{
   int32 t = (int32)c->getTime();
   if (t > context->getEndTime() || t<context->getStartTime()) throw ERangeCheck("time out of range");
   card32 index = t - context->getStartTime();

   context->addSimpleSpatialDerivativeConstraint(t, c, quadTerms[index], linTerms[index]);
}

void TSSmoothDeformationSolver::solve() 
{
   size_t numTimeSteps = quadTerms.size();
   x4Assert(numTimeSteps == linTerms.size());
   for (size_t t=0; t<numTimeSteps; t++) {
      DVectorF result;
      context->getLastResult((int32)t-context->getStartTime(), result);
      card32 numIterations = settings->numCGIterations;
      conjugateGradientsSolve<float32>(
         quadTerms[t],
         result,
         linTerms[t],
         numIterations,
         settings->convThreshCG,
         settings->mmDebugOutput
      );
      //debugOutput << "num CG iterations: " << numIterations << "\n";
      context->setResult((int32)t-context->getStartTime(), result);
   }
   quadTerms = std::vector<SparseMatrixF>();
   linTerms = std::vector<DVectorF>();
}



void TSSmoothDeformationSolver::callCallbacks() 
{
   for (unsigned i=0; i<callbacks.size(); i++) {
      callbacks[i]();
   }
}

void TSSmoothDeformationSolver::addUserFeedbackCallback(Delegate d) {
   callbacks.push_back(d);
}

void TSSmoothDeformationSolver::clearUserFeedbackCallbacks() 
{
   callbacks.clear();
}

DeformationEvaluator *TSSmoothDeformationSolver::createEvaluator() 
{
	return new TSSmoothDeformationEvaluator();
}

