//---------------------------------------------------------------------------
#include "stdafx.h"
//---------------------------------------------------------------------------
#include "PCCLaplacianSurfaceProjection.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "InCorePCTopologyGraph.h"
#include "AttachedData.h"
#include "TopologyAttachment.h"
#include "UnstructuredInCoreTriangleMesh.h"
#include "UnstructuredInCorePointCloud.h"
#include "SparseLinearAlgebra.inline.h"
#include "DynamicLinearAlgebra.inline.h"
#include "PointCloudTools.h"
#include "IterativeSolvers.h"
#include "PointSetKNNQuery.h"
#include "TriangleMeshTools.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

IMPLEMENT_X4_CLASS(PCCLaplacianSurfaceProjection, 1)
{
	BEGIN_CLASS_INIT(PCCLaplacianSurfaceProjection)
   ADD_STRING_PROP(source, 0)
   ADD_STRING_PROP(target, 0)
   ADD_FLOAT32_PROP(regularizerWeight, 0)
   ADD_BOOLEAN_PROP(biLaplacian, 1)
   ADD_FLOAT32_PROP(dataAttrWindow, 1)
}

//---------------------------------------------------------------------------

PCCLaplacianSurfaceProjection::PCCLaplacianSurfaceProjection() {
   source = "root/pc1";
   target = "root/pc1";
   regularizerWeight = 1.0f;
   biLaplacian = false;
   dataAttrWindow = 100000;
}

PCCLaplacianSurfaceProjection::~PCCLaplacianSurfaceProjection() {}

void PCCLaplacianSurfaceProjection::assign(const Object* obj, X4CopyContext *context)
{
	copyObjectProperties(obj, this);
}

//---------------------------------------------------------------------------

static InCorePCTopologyGraph *getTopology(PointCloud *pc) {
   AttachedDataContainer *attachments = pc->getAttachments();
   TopologyAttachment *topAtt = dynamic_cast<TopologyAttachment*>(attachments->getData("TopologyAtt"));
   if (topAtt) {
      return topAtt->getTopology();
   } else {
      return NULL;
   }
}   


void X4::PCCLaplacianSurfaceProjection::execute( Scene *scene )
{
   UnstructuredInCoreTriangleMesh *sourcePC = dynamic_cast<UnstructuredInCoreTriangleMesh*>(getPointCloud(scene, source));
   if (!sourcePC) {
      error("source mesh not found");
      return;
   }
   sourcePC->computeTopologyAttachment(true);
   InCorePCTopologyGraph *tgSource = getTopology(sourcePC);
   x4Assert(tgSource); if (!tgSource) return;
   UnstructuredInCorePointCloud *targetpc = dynamic_cast<UnstructuredInCorePointCloud*>(getPointCloud(scene, target));
   if (!targetpc) {
      error("source upc not found");
      return;
   }
   PointSet *psSource = sourcePC->getPointSet();
   PointSet *psTarget = targetpc->getPointSet();
   AAT sourcePositionAAT = psSource->getAAT("position");
   AAT targetPositionAAT = psTarget->getAAT("position");
   AAT targetNormalAAT = psTarget->getAAT("normal");
   if (targetNormalAAT == NULL_AAT) {
      error("need normals in target");
      return;
   }
   
   AttachedIndexedOctree *octree = PointSetKNNQuery::createOctree(psTarget);
   PointSetKNNQuery knn(psTarget, octree, NULL_VECTOR3F, targetPositionAAT);
   
   mpcard nSrc = psSource->getNumEntries();
   
   DVectorF x(nSrc*3);
   SparseMatrixF A(nSrc*3);
   DVectorF b(nSrc*3);
   x.setZero();
   b.setZero();
   A.setZero();
   
   for (mpcard i=0; i<nSrc; i++) {
      for (mpcard d=0; d<3; d++) {
         mpcard nAdj = tgSource->getNumAdjacentVertices(i);
         for (mpcard j=0; j<nAdj; j++) {
            mpcard vj = tgSource->getAdjacentVertex(i, j);
            A[3*i+d].addEntry(3*i+d, 2*regularizerWeight);
            A[3*i+d].addEntry(3*vj+d, -1*regularizerWeight);
            A[3*vj+d].addEntry(3*i+d, -1*regularizerWeight);
         }
      }
   }
   
   if (biLaplacian) {
      SparseMatrixF BL(nSrc*3);
      BL.setZero();
      for (mpcard r=0; r<nSrc*3; r++) {
         for (mpcard c=0; c<nSrc*3; c++) {
            float32 v = A[r]*A[c];
            if (fabs(v) > 1e-7f) {
               BL[r].setEntry(c, v);
            }
         }         
      }
      A = BL;
   }
   bool selfMatch = (source == target);
   for (mpcard i=0; i<nSrc; i++) {
      Vector3f targetPos;
      float32 weight;
      if (selfMatch) {
         targetPos = psSource->get3f(i, sourcePositionAAT);
         weight = 1;
      } else {
         Vector3f sp = psSource->get3f(i, sourcePositionAAT);
         knn.setTargetPoint(sp);
         card32 index; bool found;
         knn.getNextNearestPointIndex(index, found);
         x4Assert(found);
         Vector3f tp = psTarget->get3f(index, targetPositionAAT);
         Vector3f tn = normalize(psTarget->get3f(index, targetNormalAAT));
         
         Vector3f diff = (tp-sp);
         targetPos = sp + (diff*tn)*tn;
         weight = exp(-(normQuad(diff)/sqr(dataAttrWindow)));
      }
      for (mpcard d=0; d<3; d++) {
         A[3*i+d].addEntry(3*i+d, weight);
         b[3*i+d] = targetPos[d]*weight;
      }
   }
   
   card32 numIt = 10000;
   conjugateGradientsSolve(A,x,b, numIt);
   debugOutput << "solve: " << numIt << " iterations\n";

   UnstructuredInCoreTriangleMesh *output = (UnstructuredInCoreTriangleMesh*)sourcePC->copy();
   PointSet *psOutput = output->getPointSet();
   
   for (mpcard i=0; i<nSrc; i++) {
      psOutput->set3f(i, sourcePositionAAT, makeVector3f(x[3*i], x[3*i+1], x[3*i+2]));
   }
   computeTriangleMeshNormals(output);
   addPointCloud(scene, output, "laplacedPC");   
   delete octree;
}