//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PCCRangeQueryTest.h"
//---------------------------------------------------------------------------
#include "PointSetKNNQuery.h"
#include "GridBasedPointSetKNNQuery.h"
#include "Timer.h"
#include "GridBasedSphericalRangeQuery.h"
#include "SparseGrid.inline.h"
#include "PointCloudTools.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------


static std::string descr(card32 i) {
   if (i==0) return "octree";
   if (i==1) return "grid-knn";
   if (i==2) return "grid-range";
   return "undefined";
}

IMPLEMENT_X4_CLASS( PCCRangeQueryTest , 0 ) {
	BEGIN_CLASS_INIT( PCCRangeQueryTest );
   ADD_STRING_PROP(target, 0)
   ADD_FLOAT32_PROP(radius, 0)
   ADD_CARD32_PROP(numNeighbors, 0)
   ADD_CARD32_PROP(index, 0)
   ADD_CARD32_PROP_UPDATE_DESCR(mode, 0, NULL, descr, 3, NULL)
   ADD_CARD32_PROP(timingIterations, 0)
   ADD_CARD32_PROP(numOctreePoints, 0)
   ADD_FLOAT32_PROP(cellSizeFactor, 0)

   ADD_NOARGS_METHOD(octreeParameterRun)
   ADD_NOARGS_METHOD(rangeParameterRun)
   ADD_NOARGS_METHOD(clean)
   ADD_CARD32_PROP(vectorPreReserve, 0)
}

// ----

PCCRangeQueryTest::PCCRangeQueryTest() {
   target = "root/pc1";
   radius = 0.1f;
   numNeighbors = 1000000;
   index = 0;
   mode = 0;
   timingIterations = 1;
   numOctreePoints = 20;
   cellSizeFactor = 1;
   vectorPreReserve = 0;
}

PCCRangeQueryTest::~PCCRangeQueryTest() {
}




void PCCRangeQueryTest::execute(Scene *scene) 
{
   UnstructuredInCorePointCloud *upc = dynamic_cast<UnstructuredInCorePointCloud*>(getPointCloud(scene, target));
   if (!upc) {warning("no target."); return;}
   
   PointSet *ps = upc->getPointSet();
   
   
   AAT positionAAT = ps->getAAT("position");
   AAT colorAAT = ps->getAAT("color");
   if (mode == 0) {
      debugOutput << "octree:\t";
      Timer timer;
	   AttachedIndexedOctree* octree = PointSetKNNQuery::createOctree( ps, numOctreePoints );
      debugOutput << timer.getDeltaValue() << "\t";
      debugOutput << "query:\t";
      PointSetKNNQuery q(ps, octree, NULL_VECTOR3F, positionAAT);
      for (card32 i=0; i<timingIterations; i++) {
         card32 useIndex = index;
         if (index == -1) useIndex = randomNumber(ps->getNumEntries());
         if (useIndex >= ps->getNumEntries()) useIndex = ps->getNumEntries()-1;
         Vector3f targetPoint = ps->get3f(useIndex, positionAAT);
         q.setTargetPoint(targetPoint);
         card32 numPts = 0;
         float32 distQ;
         do {
            card32 nextIndex;
            bool found;
            q.getNextNearestPointIndex(nextIndex, found);
            if (!found) break;
            Vector3f pos = ps->get3f(nextIndex, positionAAT);
            distQ = normQuad(pos - targetPoint);
            Vector3f col = ps->get3f(nextIndex, colorAAT);
            if (col[0] < 1) col[0] = 1;
            ps->set3f(nextIndex, colorAAT, col);
            numPts++;
         } while (numPts < numNeighbors && distQ < radius*radius);
      }
      debugOutput << timer.getDeltaValue() << "\n";
      delete octree;
   } else if (mode == 1) {
      debugOutput << "gridKNN:\t";
      Timer timer;
      GridBasedPointSetKNNQuery q(ps, radius);
      debugOutput << timer.getDeltaValue() << "\t";
      debugOutput << "query:\t";
      for (card32 i=0; i<timingIterations; i++) {
         card32 useIndex = index;
         if (index == -1) useIndex = randomNumber(ps->getNumEntries());
         if (useIndex >= ps->getNumEntries()) useIndex = ps->getNumEntries()-1;
         Vector3f targetPoint = ps->get3f(useIndex, positionAAT);
         q.setTargetPoint(targetPoint);
         card32 numPts = 0;
         float32 distQ;
         do {
            card32 nextIndex;
            bool found;
            q.getNextNearestPointIndex(nextIndex, found);
            if (!found) break;
            Vector3f pos = ps->get3f(nextIndex, positionAAT);
            distQ = normQuad(pos - targetPoint);
            Vector3f col = ps->get3f(nextIndex, colorAAT);
            if (col[1] < 1) col[1] = 1;
            ps->set3f(nextIndex, colorAAT, col);
            numPts++;
         } while (numPts < numNeighbors && distQ < radius*radius);
      }
      debugOutput << timer.getDeltaValue() << "\n";
   } else if (mode == 2) {
      debugOutput << "gridRange:\t";
      Timer timer;
      GridBasedSphericalRangeQuery q(ps, radius);
      debugOutput << timer.getDeltaValue() << "\t";
      debugOutput << "query:\t";
      for (card32 i=0; i<timingIterations; i++) {
         card32 useIndex = index;
         if (index == -1) useIndex = randomNumber(ps->getNumEntries());
         if (useIndex >= ps->getNumEntries()) useIndex = ps->getNumEntries()-1;
         Vector3f targetPoint = ps->get3f(useIndex, positionAAT);
         std::vector<int32> result;
         if (vectorPreReserve) result.reserve(vectorPreReserve);
         q.getIndices(targetPoint, result);
         for (unsigned j=0; j<result.size(); j++) {
            card32 nextIndex = result[j];
            Vector3f col = ps->get3f(nextIndex, colorAAT);
            if (col[1] < 1) col[1] = 1;
            ps->set3f(nextIndex, colorAAT, col);
         }
      }
      debugOutput << timer.getDeltaValue() << "\n";
   }
}

void PCCRangeQueryTest::octreeParameterRun() 
{
   for (card32 i=500; i>=2; i-=10) {
      numOctreePoints = i;
      debugOutput << i << "\t";
      execute(getScene());
   }
}

void PCCRangeQueryTest::rangeParameterRun() 
{
   numNeighbors = 1000000;
   for (float32 r=2.0f; r>=0.01; r/=sqrt(2.0f)) {
      radius = r;
      debugOutput << "radius:\t" << r << "\t";
      execute(getScene());
   }
}

void PCCRangeQueryTest::clean() 
{
   UnstructuredInCorePointCloud *upc = dynamic_cast<UnstructuredInCorePointCloud*>(getPointCloud(scene, target));
   if (!upc) {warning("no target."); return;}
   
   PointSet *ps = upc->getPointSet();
   AAT colorAAT = ps->getAAT("color");
   card32 n = ps->getNumEntries();
   AAT color = ps->getAAT("color");
   for (card32 i=0; i<n; i++) {
      ps->set3f(i, colorAAT, NULL_VECTOR3F);
   }
   
}
