#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "ShadowVolumeGrid.h"
//---------------------------------------------------------------------------
#include "RegularInCoreVolume.h"
#include "RegularVolume.h"
#include "GenericSwapAllOutOfCorePointCloud.h"
#include "CameraParameterAttachment.h"
#include "ParametricInCorePointCloud.h"
#include "PCCImportDepthAnimation.h"
#include "StringHelper.h"
#include "PointSetKNNQuery.h"
#include "PointSet.inline.h"
#include "PointCloudTools.h"
#include <opencv/cxcore.h>
//---------------------------------------------------------------------------
using namespace X4;

IMPLEMENT_X4_CLASS( ShadowVolumeGrid ,0)
{
	BEGIN_CLASS_INIT( ShadowVolumeGrid );
}

ShadowVolumeGrid::ShadowVolumeGrid(void)
{
	voxelData=NULL;
	resX=0;
	resY=0;
	resZ=0;
	sideLength=0;
	depthTolerance=1;
	knnNearShadow=NULL;
	octree=NULL;
	nearShadowPointSet=NULL;
}

ShadowVolumeGrid::~ShadowVolumeGrid(void)
{
	reset();
}

void X4::ShadowVolumeGrid::reset()
{
	delete voxelData;
	voxelData=NULL;
	resX=0;
	resY=0;
	resZ=0;
	sideLength=0;
	depthTolerance=1;
	delete knnNearShadow;
	knnNearShadow=NULL;
	delete octree;
	octree=NULL;
	delete nearShadowPointSet;
	nearShadowPointSet=NULL;

}

void ShadowVolumeGrid::assign(const Object* obj, X4CopyContext *context)
{
	Persistent::assign(obj,context);
	const ShadowVolumeGrid * o = dynamic_cast<const ShadowVolumeGrid*>(obj);
	x4Assert( o != NULL );

}

void X4::ShadowVolumeGrid::setup( BoundingBox3f bb, card32 minRes, float32 tol )
{

	depthTolerance=tol;
	
	upper=bb.upperCorner;
	lower=bb.lowerCorner;

	float32 maxDist=bb.getMaxSideLength();
	sideLength=maxDist/(float32)minRes;
	resX=(upper[0]-lower[0])/sideLength+1;
	resY=(upper[1]-lower[1])/sideLength+1;
	resZ=(upper[2]-lower[2])/sideLength+1;

	card32 numVoxels=resX*resY*resZ;

	delete voxelData;
	voxelData= new BitVector8[numVoxels];
	//nearShadow= new bool[numVoxels];

	
}

X4::Vector3f X4::ShadowVolumeGrid::getVoxelCenter( const card32 &x, const card32 &y, const card32 &z )
{
	return lower+makeVector3f((float32)x *sideLength, (float32)y *sideLength, (float32)z *sideLength);
}

bool X4::ShadowVolumeGrid::inShadowVolume( Vector3f worldCoordinate )
{
	Vector2f imagePos = myCam->projectWorldtoImage(worldCoordinate);
	float32 x=imagePos[0], y=imagePos[1];
	if ((MIN(x,y)<0)||(x>1023)||(y>767))
		return false;

	float32 depth=depthPicturePointset->get1f( x+0.5f , y+0.5f ,DEPTH);
	float32 depth3D=worldCoordinate[2];
	if (depth3D >= (depth-depthTolerance))
		return true;
	else
		return false;
}

void X4::ShadowVolumeGrid::processDepthImage( card32 camNo, CameraParameterAttachment* cam, const PointSet* picturePointset )
{
	myCam=cam;
	depthPicturePointset=picturePointset;
	DEPTH=picturePointset->getAAT("depth");

	for (card32 z=0; z<resZ; z++) {
		for (card32 y=0; y<resY; y++) {
			for (card32 x=0; x<resX; x++) {

				Vector3f scenePoint = getVoxelCenter(x,y,z);
				if( inShadowVolume(scenePoint) ){
					addCamToVoxel(camNo,x,y,z);
				}
			}
		}
	}
}

X4::card8 X4::ShadowVolumeGrid::getVoxelValue( const card32 &x, const card32 &y, const card32 &z )
{
	return voxelData[x+y*resX+z*resX*resY].countTrueBits();
}

#ifdef USE_VOXELGRID
void X4::ShadowVolumeGrid::displayVoxelGrid( card8 th, Scene* scene )
{
	VoxelGrid* vg = new VoxelGrid(sideLength);
	vg->setMaterialIndex(5);


	for (card32 z=0; z<resZ; z++) {
		for (card32 y=0; y<resY; y++) { 
			for (card32 x=0; x<resX; x++) {

				Vector3f scenePoint = getVoxelCenter(x,y,z);
				if( getVoxelValue(x,y,z)>=th ){
					PCVoxel* vox=vg->getOrCreateEmptyVoxel(vg->getVoxelIndex(scenePoint));
					vox->setPoint();					
				}
			}
		}
	}
	addPointCloud( scene, vg, "volume" );
}
#endif // USE_VOXELGRID

void X4::ShadowVolumeGrid::displayRegularVolume( card8 th, Scene* scene, card32 frameNo )
{
	RegularInCoreVolume *testVol = new RegularInCoreVolume();
	VertexDescriptor vd;
	vd.pushAttrib(VAD("position", 3, VAD::DATA_FORMAT_FLOAT32));
	vd.pushAttrib(VAD("intensity", 1, VAD::DATA_FORMAT_FLOAT32));
	BoundingBox3f sceneBB(lower, upper);
	testVol->clearAndSetup(resX, resY, resZ, &vd, sceneBB  );

	RegularVolume * voxels = testVol->getVoxels();
	PointSet *ps = voxels->getVoxelData();
	AAT POS = ps->getAAT("position");
	AAT INTEN = ps->getAAT("intensity");

	for (card32 z=0; z<resZ; z++) {
		for (card32 y=0; y<resY; y++) {
			for (card32 x=0; x<resX; x++) {
				
				StaticVector<float,1> v;

				/*if(nearShadow[x+y*resX+z*resX*resY])
					v[0]=0.6f;
				else
					v[0]=0;*/
				
				card8 shadowCnt=getVoxelValue(x,y,z);
				if (shadowCnt>th)
					v[0] = (float32)shadowCnt/8.0f;
				else
					v[0] = 0.0f;

				ps->set<float32, 1>(x+y*resX+z*resX*resY , INTEN, v);
				ps->set<float32, 3>(x+y*resX+z*resX*resY , POS, getVoxelCenter(x,y,z));
			}
		}
	}
	testVol->setMaterialIndex(5);
	addPointCloud( scene, testVol, "frame_"+intToStr(frameNo)+"_shadowVolume" );
}

void X4::ShadowVolumeGrid::processFrame( card32 frameNo, Scene* scene )
{

	for (card32 camNo=0; camNo<8; camNo++)
	{
		PointCloud* pc= getPointCloud(scene, PCCImportDepthAnimation::getPointCloudName(frameNo, camNo ,true));
		if (!pc->supportsIteratorType( SOIT::CAP_PARAMETRIC_PC)) {
			x4Assert(false);
			return;
		}
		SceneObjectIterator* iter=pc->createIterator(SOIT::CAP_BASIC_PC|SOIT::CAP_PARAMETRIC_PC );
		BasicPointCloudIterator* basicIt=dynamic_cast<BasicPointCloudIterator*>(iter);
		const PointSet *picturePointset = basicIt->lockBlockRead();
		CameraParameterAttachment* camAtt=dynamic_cast<CameraParameterAttachment*>(pc->getAttachments()->getData(CameraParameterAttachment::getAttachmentName()));

		processDepthImage(camNo, camAtt, picturePointset);

		basicIt->unlock();
		delete iter;
	}

	nearShadowPointSet=new PointSet();
	VertexDescriptor vd;
	vd.pushAttrib(VAD("position", 3, VAD::DATA_FORMAT_FLOAT32));
	vd.pushAttrib(VAD("cameras", 1, VAD::DATA_FORMAT_CARD8));
	nearShadowPointSet->clearAndSetup(0,0,&vd);
	CAM=nearShadowPointSet->getAAT("cameras");
	NSH_POS=nearShadowPointSet->getAAT("position");
	card32 currIdx=0;

	for (card32 z=0; z<resZ; z++) 
		for (card32 y=0; y<resY; y++) 
			for (card32 x=0; x<resX; x++) {
				card32 pos=x+y*resX+z*resX*resY;
				card8 cams=voxelNearShadow(x,y,z);
				if(cams){
//					nearShadow[pos]=true;
					Vector3f center=getVoxelCenter(x,y,z);
					nearShadowPointSet->addPoint((char*)&center);
					nearShadowPointSet->set1ub(currIdx, CAM, cams);
				}
			//	else
			//		nearShadow[pos]=false;
			}

	octree = PointSetKNNQuery::createOctree( nearShadowPointSet, 20 );
	knnNearShadow= new PointSetKNNQuery( nearShadowPointSet, octree, makeVector3f( 0,0,0 ), NSH_POS );


	
	//debug
		//visualizeVolumeInScene(1, scene, frameNo);
	/*	UnstructuredInCorePointCloud* upc=new UnstructuredInCorePointCloud();
		upc->clearAndSetup(&vd,0);
		upc->setPointSet(nearShadowPointSet);
		addPointCloud( scene, upc, "frame_"+intToStr(frameNo)+"_nearShadowPointSet" );
*/
}

card8 ShadowVolumeGrid::voxelNearShadow(const int32 &myX, const int32 &myY, const int32 &myZ){
	
	if (getVoxelValue(myX,myY,myZ))
		return 0; //voxel is IN shadow, not near sh.
	card8 result=0;
	
	for (int32 z=myZ-1; z<=myZ+1; z++) 
		for (int32 y=myY-1; y<=myY+1; y++) 
			for (int32 x=myX-1; x<=myX+1; x++) {
				if (z<0 || y<0 || x<0 || z>=(int32)resZ || y>=(int32)resY || x>=(int32)resX)
					continue;
				result|=getVoxelValue(x,y,z);
			}
	return result;
}

void X4::ShadowVolumeGrid::addCamToVoxel( const card32 &camNo, const card32 &x, const card32 &y, const card32 &z )
{
	voxelData[x+y*resX+z*resX*resY].setBitTrue(camNo);
}

BitVector8* X4::ShadowVolumeGrid::getVoxelPointer( const card32 &x, const card32 &y, const card32 &z )
{
	return &voxelData[x+y*resX+z*resX*resY];
}

X4::Vector3f X4::ShadowVolumeGrid::getNearestPositionAndCams( Vector3f pointPos, BitVector8 &returnedCams )
{
	knnNearShadow->setTargetPoint( pointPos );
	card32 nIndex = 0;
	bool found;
	knnNearShadow->getNextNearestPointIndex( nIndex, found );
	x4Assert(found);
	if (!found) error("ShadowVolumeGrid::getNearestPositionAndCams - no Point found");
	returnedCams.fromCard8(nearShadowPointSet->get1ub(nIndex,CAM));
	return nearShadowPointSet->get3f(nIndex,NSH_POS);
}
