//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "WeightFittingTest.h"
//---------------------------------------------------------------------------
#include "PartialDerivative.h"
#include "PartialDerivative.inline.h"
#include "DebugOutput.h"
#include "Timer.h"
#include "GLGraphDisplayer.h"
#include "DynamicLinearAlgebraTools.h"
#include "DynamicLinearAlgebraTools.inline.h"
#include "SparseLinearAlgebra.h"
#include "SparseLinearAlgebra.inline.h"
#include "ColorTools.h"
#include "IterativeSolvers.h"
#include "IterativeSolvers.inline.h"
#include "PointCloudTools.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

typedef HessianTools<float32, 3> HTools3f;
typedef HTools3f::DiffTwice ADType;


IMPLEMENT_X4_CLASS( WeightFittingTest , 0 ) {
	BEGIN_CLASS_INIT( WeightFittingTest );
   ADD_FLOAT32_PROP(baseWeight, 0)
   ADD_FLOAT32_PROP(denomOffset, 0)
   ADD_INT32_PROP(filterWindowSize, 0)
   ADD_CARD32_PROP(numEigenIterations, 0)
	ADD_NOARGS_METHOD(execute)
}

// ----

WeightFittingTest::WeightFittingTest() {
	baseWeight = 50.0f;
	denomOffset = 0.001f;
	filterWindowSize = 10;
	numEigenIterations = 2;
}

WeightFittingTest::~WeightFittingTest() {
}



template <class FloatType>
static FloatType scoreFunc2(
      const FloatType &left, 
      const FloatType &middle, 
      const FloatType &right, 
      const FloatType &measuredMiddle,
      const FloatType &weight) 
{
   FloatType laplace = left + right - middle*(FloatType)2;
   FloatType smoothScore = (laplace*laplace) * weight;
   FloatType dataScore = sqr(middle - measuredMiddle);
   
   return smoothScore + dataScore*(FloatType)0.001;
};

static void computeLocalTaylor2(
   const std::vector<float32> &pts, const std::vector<float32> &ptsOriginal, int i, const DVectorF &weights, 
   float32 &localScore, Vector3f &localGradient, Matrix3f &localHessian,
   const int NUM_PTS) 
{
   int i0 = (i+(int)NUM_PTS-1) % NUM_PTS;
   int i1 = (i+(int)NUM_PTS)   % NUM_PTS;
   int i2 = (i+(int)NUM_PTS+1) % NUM_PTS;
   ADType left = HTools3f::createVariable(pts[i0], 0);
   ADType middle = HTools3f::createVariable(pts[i1], 1);
   ADType right = HTools3f::createVariable(pts[i2], 2);
   ADType measuredMiddle = ptsOriginal[i1];
   ADType score = scoreFunc2(left, middle, right, measuredMiddle, ADType(weights[i]));
   localGradient = HTools3f::getGradient(score);
   localHessian = HTools3f::getHessian(score);
}


static void computeTaylor2(
   const std::vector<float32> &pts, const std::vector<float32> &ptsOriginal, const DVectorF &weights,
   SparseMatrixF &hessian, DVectorF &gradient, float32 &score, 
   const int NUM_PTS) 
{
   for (int i=0; i<NUM_PTS; i++) {
      Vector3f localGradient;
      Matrix3f localHessian;
      float32 localScore;
      computeLocalTaylor2(pts, ptsOriginal, i, weights, localScore, localGradient, localHessian, NUM_PTS);
      int i0 = (i+NUM_PTS-1) % NUM_PTS;
      int i1 = (i+NUM_PTS) % NUM_PTS;
      int i2 = (i+NUM_PTS+1) % NUM_PTS;
      gradient[i0] += localGradient[0];
      gradient[i1] += localGradient[1];
      gradient[i2] += localGradient[2];
      hessian[i0].addEntry(i0, localHessian[0][0]); hessian[i0].addEntry(i1, localHessian[0][1]); hessian[i0].addEntry(i2, localHessian[0][2]);
      hessian[i1].addEntry(i0, localHessian[1][0]); hessian[i1].addEntry(i1, localHessian[1][1]); hessian[i1].addEntry(i2, localHessian[1][2]);
      hessian[i2].addEntry(i0, localHessian[2][0]); hessian[i2].addEntry(i1, localHessian[2][1]); hessian[i2].addEntry(i2, localHessian[2][2]);
      score += localScore;
   }
}


void WeightFittingTest::execute() {
   static const unsigned NUM_PTS = 100;
   std::vector<float32> pts(NUM_PTS);
   for (unsigned i=0; i<NUM_PTS; i++) {
      pts[i] = fabs(i/50.0f-1.0f) + rnd01()*0.1;
   }
   std::vector<float32> ptsOriginal = pts;

   std::vector< std::vector<float32> > graphs;
   std::vector< Vector4ub > colors;
   graphs.push_back(pts);
   colors.push_back(makeVector4ub(220,68,60,255));

   SparseMatrixF hessian(NUM_PTS);
   DVectorF gradient(NUM_PTS);
   DVectorF weights(NUM_PTS);
   weights = scalarDVector(NUM_PTS, baseWeight);

	debugOutput << norm(weights) << "\n";
    
   for (card32 wi=0; wi<numEigenIterations; wi++) {  
		pts = ptsOriginal;
		DVectorF x(NUM_PTS);
		convertVectorType(pts, x, NUM_PTS);
		card32 maxNumIterations = 1000;
		float32 score = 0;
		hessian.setZero();
		gradient.setZero();
		computeTaylor2(pts, ptsOriginal, weights, hessian, gradient, score, NUM_PTS);
		conjugateGradientsSolve<float32>(
			hessian,
			x,
			hessian*x - gradient,
			maxNumIterations,
			1E-5f,
			false
		);
		convertVectorType(x, pts, NUM_PTS);
		
		DVectorF dataDiff(NUM_PTS);
						
		for (unsigned i=0; i<NUM_PTS; i++) {
			dataDiff[i] = x[i] - ptsOriginal[i];
		}
		
		for (int i=0; i<(int)NUM_PTS; i++) {
			float32 filtered = 0;
			float32 fw = 0;
			if (filterWindowSize == 0) {
				filtered = dataDiff[i];
			} else {
				for (int j=-filterWindowSize; j<filterWindowSize; j++) {
					int p = i+j;
					if (p>=0 && p <(int)NUM_PTS) {
						fw += 1;
						filtered += dataDiff[p];
					}
				}
			}
			filtered /= fw;
			weights[i] = 1.0f/(sqr(filtered)+denomOffset);
		}

		weights /= norm(weights);
		weights *= sqrt(baseWeight*baseWeight*NUM_PTS);
	   
	   debugOutput << norm(weights) << "\n";
	   
		std::vector<float32> weightsV(NUM_PTS);
		DVectorF wv = weights / sqrt(baseWeight*baseWeight*NUM_PTS);
		for (unsigned i=0; i<NUM_PTS; i++) {wv[i] = sqrt(wv[i]);}
		wv += scalarDVector<float32>(NUM_PTS, 1.0f);
		convertVectorType(wv, weightsV, NUM_PTS);
		if (wi == numEigenIterations-1) {
			graphs.push_back(pts);
			colors.push_back(makeVector4ub(255,255,255,255));
			if (numEigenIterations != 2) {
				graphs.push_back(weightsV);
				colors.push_back(makeVector4ub(255,255,255,255));
			}
		} else if (wi == 0) {
			graphs.push_back(pts);
			colors.push_back(makeVector4ub(255,255,0,255));
			graphs.push_back(weightsV);
			colors.push_back(makeVector4ub(255,255,0,255));
		} else {
			graphs.push_back(pts);
			colors.push_back(makeVector4ub(50,50,50,255));
			graphs.push_back(weightsV);
			colors.push_back(makeVector4ub(50,50,50,255));
		}
	}
   
   debugOutputGraphs("Weight Opt Test", graphs, colors);
}
