#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PCCDeformationSurfaceReconstruction.h"
//---------------------------------------------------------------------------
#include "StandardSceneGraphIterator.h"
#include "SGObjectNode.h"
#include "PointCloudTools.h"
#include "DebugOutput.h"
#include "ClassMethods.h"
#include "pointcloud.h"
#include "NumericalClassProperty.h"
#include "CopyObjectProperties.h"
#include "PointSetKNNQuery.h"
#include "SeparatorClassProperty.h"
#include "TopologyAttachment.h"
#include "ProgressWindow.h"
#include "ARConstraints.h"
#include "LocalSurfaceFunction.h"
#include "ShapeEvaluator.h"
#include "DeformationTools.h"
#include "GLShaderMaterial.h"
#include "MarchingCubes3D.h"
#include "Resampling.h"
#include "TopologyRangeSearch.h"
//---------------------------------------------------------------------------
using namespace X4;

IMPLEMENT_X4_CLASS( PCCDeformationSurfaceReconstruction, 0 )
{
	BEGIN_CLASS_INIT( PCCDeformationSurfaceReconstruction )
	ADD_OBJECT_PROP( settings, 0, ARSettings::getClass(), true )
	ADD_OBJECT_PROP( triangulationSettings, 0, BackendSettings::getClass(), true )
	ADD_OBJECT_PROP( initializer, 0, SurfelGraphInit2::getClass(), true )
	ADD_OBJECT_PROP( optimizer, 0, SurfelOptimizer::getClass(), true )
	ADD_OBJECT_PROP( surfels, 0, SGESurfelGraph::getClass(), true )
	ADD_OBJECT_PROP( data, 0, SGEPointCloud::getClass(), true )
	ADD_OBJECT_PROP( icpConstraints, 0, ICPConstraints::getClass(), true )
	ADD_SEPARATOR( "Debugging" )
	ADD_BOOLEAN_PROP( createICPConstraints, 0 )
	ADD_BOOLEAN_PROP( createHSContraint, 0 )
	ADD_VECTOR3F_PROP( hsPoint, 0 )
	ADD_VECTOR3F_PROP( hsNormal, 0 )
	ADD_FLOAT32_PROP( hsWeight, 0 )
	ADD_CARD32_PROP( hsNumSamples, 0 )
	ADD_FLOAT32_PROP( hsRadius, 0 )
	ADD_INT32_PROP( visI, 0 )
	ADD_INT32_PROP( visJ, 0 )
	ADD_INT32_PROP( visK, 0 )
	ADD_SEPARATOR( "Rendering" )
	ADD_INT32_PROP( monitorPoint, 0 )
	ADD_FLOAT32_PROP( renderScale, 0 )
	ADD_BOOLEAN_PROP( showSurfelGraph, 0 )
	ADD_BOOLEAN_PROP( showSurfelFrame, 0 )
	ADD_BOOLEAN_PROP( showSurfelFunction, 0 )
	ADD_BOOLEAN_PROP( showHSConstraints, 0 )
	ADD_BOOLEAN_PROP( showBorderClip, 0 )
	ADD_BOOLEAN_PROP( showMC, 0 )
	ADD_NOARGS_METHOD( initialize )
	ADD_NOARGS_METHOD( optimize )
	ADD_NOARGS_METHOD( triangulate )
	ADD_NOARGS_METHOD( upsample );
}

PCCDeformationSurfaceReconstruction::PCCDeformationSurfaceReconstruction(void)
{
	settings = new ARSettings();
	settings->setPresets( ARDefaultSettings::PCA_TEST );
	triangulationSettings = new BackendSettings();
	initializer = new SurfelGraphInit2();
	optimizer = new SurfelOptimizer();
	surfels = new SGESurfelGraph();
	surfels->setPath( "root/surfelGraph" );
	icpConstraints = new ICPConstraints();
	triangulator = new SurfelGraphTriangulator();
	triangulator->setBackendSettings( triangulationSettings );
	
	// Data
	data = new SGEPointCloud();
	data->setPath( "root/data/frame1" );
	VertexDescriptor* vd = new VertexDescriptor();
	vd->pushAttrib( mVAD( "position", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	data->setPointChannels( vd );

	// Debugging
	createICPConstraints = true;
	createHSContraint = false;
	hsPoint = makeVector3f( 0.4f, 0.0f, 0.0f );
	hsNormal = makeVector3f( 1.0f, 0.0f, 0.0f );
	hsWeight = 1.0f;
	hsNumSamples = 50;
	hsRadius = 0.3f;
	hsConstraints.clear();

	visI = 10;
	visJ = 21;
	visK = 15;

	// Rendering
	monitorPoint = 0;
	renderScale = 1.0f;
	rt = new RenderingTool();
	rt->init();
	showSurfelGraph = false;
	showSurfelFrame = false;
	showSurfelFunction = false;
	showHSConstraints = false;
	showBorderClip = true;
	showMC = false;
}

PCCDeformationSurfaceReconstruction::~PCCDeformationSurfaceReconstruction()
{
	delete settings;
	delete triangulationSettings;
	delete initializer;
	delete rt;
	delete optimizer;
	delete surfels;
	delete data;
	delete icpConstraints;
	delete triangulator;
}

void PCCDeformationSurfaceReconstruction::initialize()
{
	deletePointCloud( scene, surfels->getPath() );
	initializer->initialize( surfels, data, settings );
}

void PCCDeformationSurfaceReconstruction::optimize()
{
	std::string msg;
	if ( data == NULL ) return;
	if ( !data->check( msg ) ) return;
	UnstructuredInCorePointCloud* dataPC = dynamic_cast<UnstructuredInCorePointCloud*>( data->getPointCloud() );
	PointSet* dataPS = dataPC->getPointSet();
	AAT DATA_POSITION = dataPS->getAAT( "position" );
	if ( surfels == NULL ) return;
	if ( !surfels->check( msg ) ) return;
	PointCloud* surfelPC = surfels->getPointCloud();
	PointSet* surfelPS = surfels->getPointSet();
	AAT SURFEL_POSITION = surfelPS->getAAT( "position" );
	AAT SURFEL_NORMAL = surfelPS->getAAT( "normal" );
	AAT SURFEL_TANGENT_U = surfelPS->getAAT( "tangent_u" );
	AAT SURFEL_TANGENT_V = surfelPS->getAAT( "tangent_v" );
	AAT SURFEL_COEFFS = surfelPS->getAAT( "coeffs" );
	SurfelOptimizationContext* context = dynamic_cast<SurfelOptimizationContext*>( optimizer->createOptContext() );
	context->initializeFromShape( getPointCloud( scene, surfels->getPath() ), settings );
	optimizer->setup( context, settings );
	optimizer->beginConstraints();
	if ( createICPConstraints ){

		icpConstraints->setConstraintMode(settings->icpConstraintsMode);
		icpConstraints->setOutliersDistanceFactor(settings->icpNearFieldRelativeCutOff);
		icpConstraints->setDistanceOutliersMode(settings->icpNearFieldCulling?ICP_OUTLIERS_DISTANCE:ICP_OUTLIERS_NONE);
		icpConstraints->setCullNormalOutliers(false);	
		icpConstraints->setOutlierNormals(-1.0f);

		icpConstraints->buildICPConstraints( settings, optimizer, surfels->getPointCloud(), surfels->getPointCloud(), data->getPointCloud(), 0 );
	}
	if ( createHSContraint ){
		if ( hsConstraints.size() != hsNumSamples ){
			hsConstraints.clear();
			Vector3f tu = ( fabs( hsNormal * XAXIS_VECTOR3F ) > 0.8f )? hsNormal.crossProduct( YAXIS_VECTOR3F ) : hsNormal.crossProduct( XAXIS_VECTOR3F );
			Vector3f tv = hsNormal.crossProduct( tu );
			tu.normalize();
			tv.normalize();
			for ( unsigned int n = 0; n < hsNumSamples; n++ ){
				const Vector3f point = hsPoint + tu * ( rand()%200/100.0f-1 ) * hsRadius + tv * ( rand()%200/100.0f-1 ) * hsRadius;
				const Vector3f normal = hsNormal;
				hsConstraints.push_back( Plane3f( normal, point ) );
			}
		}
		for ( unsigned int n = 0; n < hsConstraints.size(); n++ ){
			ARHalfSpaceConstraint* hsConstraint = new ARHalfSpaceConstraint();
			ARConstraintAnchor anchor;
			Plane3f& plane = hsConstraints[n];
			anchor.setup( plane.getPoint(), ARConstraintAnchor::NO_TOP_ANCHOR );
			hsConstraint->setup( anchor, -1, plane.getPoint(), plane.getNormal(), hsWeight );
			optimizer->addHalfSpaceConstraint( hsConstraint );
		}
	}
	optimizer->solve();

	ShapeEvaluator *shapeEval = optimizer->createEvaluator();
	shapeEval->setup( context, settings );
	PointCloud *newSurfels = shapeEval->evaluate();

	InCorePCTopologyGraph *tg = getTopology( surfelPC );
	InCorePCTopologyGraph *copyTG = NULL;
	if (!tg) {
		warning("urshape without topology");
	} else {
		copyTG = tg->copy();
	}
	surfelPC->assign( newSurfels );
	if (copyTG) {
		attachTopology( surfelPC, copyTG );
	}
	delete newSurfels;
	delete shapeEval;
	delete context;

	redrawScene();
}

void PCCDeformationSurfaceReconstruction::triangulate()
{
	// Delete old animation node if existent
	deleteSceneGraphNode( scene, "root/" + triangulator->getMeshNodeName() );

	// Add new meshes
	SGSimpleAnimationNode* animNode = triangulator->triangulate( 0, 0, surfels->getPointCloud(), settings->topSurfelSpacing );
	if ( animNode == NULL ) return;
	scene->getRootNode()->addChildNode( animNode );

	// Rebuild render objects
	for ( unsigned int n = 0; n < scene->getNumGLMaterials(); n++ ){
		GLShaderMaterial* mat = dynamic_cast<GLShaderMaterial*>( scene->getGLMaterial( n ) );
		if ( mat != NULL ){
			mat->increaseVersion();
		}
	}
}

void PCCDeformationSurfaceReconstruction::upsample() 
{
	Resampling::upsample( surfels->getPointSet(), surfels->getTopology(), settings->topSurfelSpacing );
}

void PCCDeformationSurfaceReconstruction::execute( Scene *scene )
{
}

void PCCDeformationSurfaceReconstruction::glDrawTool( GLContext* context )
{
	// Data
	if ( data == NULL ) return;
	std::string msg;
	if ( !data->check( msg ) ) return;
	UnstructuredInCorePointCloud* dataPC = dynamic_cast<UnstructuredInCorePointCloud*>( data->getPointCloud() );
	if ( dataPC == NULL ) return;
	PointSet* dataPS = dataPC->getPointSet();
	if ( dataPS == NULL ) return;
	AAT DATA_POSITION = dataPS->getAAT( "position" );

	glEnable( GL_LIGHTING );
	rt->setRenderColor( makeVector4f( 1.0f, 0.0f, 0.0f, 1.0f ) );
	for ( unsigned int pIndex = 0; pIndex < dataPS->getNumEntries(); pIndex++ ){
		if ( pIndex == monitorPoint ){
			const Vector3f& p = dataPS->get3f( pIndex, DATA_POSITION );
		}
	}
	glDisable( GL_LIGHTING );

	// Surfel-Graph
	if ( surfels == NULL ) return;
	if ( !surfels->check( msg ) ) return;
	PointSet* surfelPS = surfels->getPointSet();
	if ( surfelPS == NULL ) return;
	InCorePCTopologyGraph* surfelTopology = surfels->getTopology();
	AAT SURFEL_POSITION = surfelPS->getAAT( "position" );
	AAT SURFEL_NORMAL = surfelPS->getAAT( "normal" );
	AAT SURFEL_TANGENT_U = surfelPS->getAAT( "tangent_u" );
	AAT SURFEL_TANGENT_V = surfelPS->getAAT( "tangent_v" );
	AAT SURFEL_COEFFS = surfelPS->getAAT( "coeffs" );
	int32 meshResolution = 10;
	std::vector<Vector3f> mesh;
	mesh.resize( (meshResolution + 1)*(meshResolution + 1) );
	glDisable( GL_LIGHTING );
	for ( unsigned int pIndex = 0; pIndex < surfelPS->getNumEntries(); pIndex++ ){
		if ( ( monitorPoint < 0 ) || ( pIndex == monitorPoint ) ){
			const Vector3f& p = surfelPS->get3f( pIndex, SURFEL_POSITION );
			const Vector3f& normal = surfelPS->get3f( pIndex, SURFEL_NORMAL );
			const Vector3f& tangent_u = surfelPS->get3f( pIndex, SURFEL_TANGENT_U );
			const Vector3f& tangent_v = surfelPS->get3f( pIndex, SURFEL_TANGENT_V );
			Matrix3f T;
			T[0] = tangent_u;
			T[1] = tangent_v;
			T[2] = normal;
			const StaticVector<float32,SURFEL_NUM_BASIS_FUNCTIONS>& coeffs = surfelPS->get<float32,SURFEL_NUM_BASIS_FUNCTIONS>( pIndex, SURFEL_COEFFS );
			if ( showSurfelGraph ){
				for ( unsigned int n = 0; n < surfelTopology->getNumAdjacentVertices( pIndex ); n++ ){
					const card32 nIndex = surfelTopology->getAdjacentVertex( pIndex, n );
					const Vector3f& np = surfelPS->get3f( nIndex, SURFEL_POSITION );
					rt->setRenderColor( makeVector4f( 0.5f, 0.75f, 0.0f, 1.0f ) );
					rt->renderLine( p, np );
				}
			}
			if ( showSurfelFrame ){
				rt->setRenderColor( makeVector4f( 0.75f, 0.25f, 0.25f, 1.0f ) );
				rt->renderLine( p, p + normal * 0.02f * renderScale );
				rt->setRenderColor( makeVector4f( 0.25f, 0.75f, 0.25f, 1.0f ) );
				rt->renderLine( p, p + tangent_u * 0.02f * renderScale );
				rt->setRenderColor( makeVector4f( 0.25f, 0.25f, 0.75f, 1.0f ) );
				rt->renderLine( p, p + tangent_v * 0.02f * renderScale );
			}

			if ( showSurfelFunction ){
				rt->setRenderColor( 0.5f, 0.5f, 0.5f );
				const Vector2f domainMin = makeVector2f( -0.5f, -0.5f );
				const Vector2f domainMax = makeVector2f( 0.5f, 0.5f );
				float stepSizeX = fabs( domainMax[0] - domainMin[0] ) / float32( meshResolution );
				float stepSizeY = fabs( domainMax[1] - domainMin[1] ) / float32( meshResolution );
				float x, y;
				for ( int32 i = 0; i <= meshResolution; i++ ){
					for ( int32 j = 0; j <= meshResolution; j++ ){
						x = domainMin[0] + i * stepSizeX;
						y = domainMin[1] + j * stepSizeY;
						mesh[i*(meshResolution+1)+j] = ( T * makeVector3f( x, y, LocalSurfaceFunction<SURFEL_NUM_BASIS_FUNCTIONS>::f( coeffs, x, y ) ) * settings->topSurfelSpacing + p );
					}
				}
				for ( int32 i = 0; i <= meshResolution; i++ ){
					for ( int32 j = 0; j <= meshResolution; j++ ){
						if ( i < meshResolution )
							rt->renderLine( mesh[i*(meshResolution+1)+j], mesh[(i+1)*(meshResolution+1)+j] );
						if ( j < meshResolution )
							rt->renderLine( mesh[i*(meshResolution+1)+j], mesh[i*(meshResolution+1)+j+1] );
					}
				}
			}
		}
	}
	glDisable( GL_LIGHTING );

	glDisable( GL_LIGHTING );
	if ( showHSConstraints ){
		if ( hsConstraints.size() != hsNumSamples ){
			hsConstraints.clear();
			Vector3f tu = ( fabs( hsNormal * XAXIS_VECTOR3F ) > 0.8f )? hsNormal.crossProduct( YAXIS_VECTOR3F ) : hsNormal.crossProduct( XAXIS_VECTOR3F );
			Vector3f tv = hsNormal.crossProduct( tu );
			tu.normalize();
			tv.normalize();
			for ( unsigned int n = 0; n < hsNumSamples; n++ ){
				const Vector3f point = hsPoint + tu * ( rand()%200/100.0f-1 ) * hsRadius + tv * ( rand()%200/100.0f-1 ) * hsRadius;
				const Vector3f normal = hsNormal;
				hsConstraints.push_back( Plane3f( normal, point ) );
			}
		}
		rt->setRenderColor( 0.75f, 0.5f, 0.5f );
		for ( unsigned int n = 0; n < hsConstraints.size(); n++ ){
			rt->renderPlane( hsConstraints[n], 0.05f );
		}
	}
	glDisable( GL_LIGHTING );

	if ( showBorderClip ){
		UnstructuredInCorePointCloud* borderPC = dynamic_cast<UnstructuredInCorePointCloud*>( getPointCloud( scene, "root/" + triangulator->getMeshNodeName() + "/border" ) );
		if ( borderPC != NULL ){
			BoundingBox3f bbox = triangulator->getMCDomain( surfelPS );
			const float32 maxDist = triangulator->getClipInfluence( triangulator->getMCubeSize( bbox, triangulationSettings->mcOctreeLevel ), settings->topSurfelSpacing );
			InCorePCTopologyGraph* borderTopology = getTopology( borderPC );
			PointSet* borderPS = borderPC->getPointSet();
			AAT BORDER_POSITION = borderPS->getAAT( "position" );
			if ( ( borderTopology != NULL ) && ( borderPS != NULL ) ){
				AttachedIndexedOctree* borderOctree = PointSetKNNQuery::createOctree( borderPS, 32 );
				for ( unsigned int bIndex = 0; bIndex < borderPS->getNumEntries(); bIndex++ ){
					if ( ( monitorPoint < 0 ) || ( monitorPoint == int32( bIndex ) ) ){
						const Vector3f p = borderPS->get3f( bIndex, BORDER_POSITION );
						Plane3f plane;

						/*
						if ( triangulator->getClippingPlane( p, maxDist, plane, borderPS, borderTopology, borderOctree, settings ) ){
							glDisable( GL_LIGHTING );
							rt->setRenderColor( makeVector3f( 0.25f, 0.25f, 0.75f ) );
							rt->renderPlane( plane, 0.05f );
							glEnable( GL_LIGHTING );
							rt->setRenderColor( makeVector3f( 0.75f, 0.25f, 0.25f ) );
							rt->renderSphere( p, 0.01f );
						}
						*/
					}
				}
				delete borderOctree;


				/*
				
				for ( unsigned int sIndex = 0; sIndex < surfelPS->getNumEntries(); sIndex++ ){
					if ( ( monitorPoint < 0 ) || ( monitorPoint == int32( sIndex ) ) ){
						const Vector3f p = surfelPS->get3f( sIndex, SURFEL_POSITION );
						Plane3f plane;
						if ( triangulator->getClippingPlane( p, maxDist, plane, borderPS, borderTopology, borderOctree, settings ) ){
							glDisable( GL_LIGHTING );
							rt->setRenderColor( makeVector3f( 0.25f, 0.25f, 0.75f ) );
							rt->renderPlane( plane, 0.05f );
							glEnable( GL_LIGHTING );
							rt->setRenderColor( makeVector3f( 0.75f, 0.25f, 0.25f ) );
							rt->renderSphere( p, 0.01f );
						}
					}
				}
				*/
				glDisable( GL_LIGHTING );
			}
		}
	}

	if ( showMC ){
		BoundingBox3f bbox = triangulator->getMCDomain( surfelPS );
		glDisable( GL_LIGHTING );
		rt->setRenderColor( 1,1,1,0.5f );
		rt->renderBox( bbox );
		const float32 mcCubeSize = triangulator->getMCubeSize( bbox, triangulationSettings->mcOctreeLevel );
		const card32 resolution = power( 2, triangulationSettings->mcOctreeLevel );

		AttachedIndexedOctree* surfelOctree = PointSetKNNQuery::createOctree( surfelPS, 32 );
		PointSetKNNQuery surfelKNN( surfelPS, surfelOctree, makeVector3f( 0,0,0 ), SURFEL_POSITION );
		Vector3f corners[8];
		rt->setRenderColor( 1,0,0 );
		for ( unsigned int k = 0; k < resolution; k++ ){
			for ( unsigned int j = 0; j < resolution; j++ ){
				for ( unsigned int i = 0; i < resolution; i++ ){
					if ( ( ( i == visI ) || ( visI < 0 ) ) && 
						 ( ( j == visJ ) || ( visJ < 0 ) ) && 
						 ( ( k == visK ) || ( visK < 0 ) ) ){
						corners[0] = bbox.lowerCorner + makeVector3f( i*mcCubeSize,     j*mcCubeSize,     k*mcCubeSize );
						corners[1] = bbox.lowerCorner + makeVector3f( (i+1)*mcCubeSize, j*mcCubeSize,     k*mcCubeSize );
						corners[2] = bbox.lowerCorner + makeVector3f( (i+1)*mcCubeSize, (j+1)*mcCubeSize, k*mcCubeSize );
						corners[3] = bbox.lowerCorner + makeVector3f( i*mcCubeSize,     (j+1)*mcCubeSize, k*mcCubeSize );
						corners[4] = bbox.lowerCorner + makeVector3f( i*mcCubeSize,     j*mcCubeSize,     (k+1)*mcCubeSize );
						corners[5] = bbox.lowerCorner + makeVector3f( (i+1)*mcCubeSize, j*mcCubeSize,     (k+1)*mcCubeSize );
						corners[6] = bbox.lowerCorner + makeVector3f( (i+1)*mcCubeSize, (j+1)*mcCubeSize, (k+1)*mcCubeSize );
						corners[7] = bbox.lowerCorner + makeVector3f( i*mcCubeSize,     (j+1)*mcCubeSize, (k+1)*mcCubeSize );
						BoundingBox3f cube;
						cube.lowerCorner = corners[0]; 
						cube.upperCorner = corners[6];
						rt->setRenderColor( 1,1,1 );
						rt->renderBox( cube );

						if ( triangulator->triangulateCube( corners, surfelPS, surfelKNN, settings->topSurfelSpacing ) ){	
							float32 values[8];
							for ( unsigned int n = 0; n < 8; n++ ){
								 values[n] = triangulator->getValue( corners[n], surfelKNN, surfelPS, surfelTopology, settings->topSurfelSpacing );
							}
							std::vector<Vector3f> verts; verts.clear();
							MarchingCubes3D::triangulate( corners, values, 0, verts );
							rt->setRenderColor( 1,1,0 );
							for ( unsigned int n = 0; n < verts.size()/3; n++ ){
								rt->renderLine( verts[n*3], verts[n*3+1] );
								rt->renderLine( verts[n*3+1], verts[n*3+2] );
								rt->renderLine( verts[n*3+2], verts[n*3] );
							}
						}

						card32 nIndex;
						bool found;
						surfelKNN.setTargetPoint( corners[monitorPoint] );
						surfelKNN.getNextNearestPointIndex( nIndex, found );
						const Vector3f closestPoint = surfelPS->get3f( nIndex, SURFEL_POSITION );
						const Vector3f closestNormal = surfelPS->get3f( nIndex, SURFEL_NORMAL );
						Vector3f seedPoint = corners[monitorPoint];
						projectPointPlane( Plane3f( closestNormal, closestPoint ), corners[monitorPoint], seedPoint );
						glEnable( GL_LIGHTING );
						rt->setRenderColor( 1,0,1 );
						rt->renderSphere( seedPoint, cube.getMaxSideLength() / 8.0f );
						TopNeighborhoodSphereSearch topoSphereSearch;
						topoSphereSearch.setup( 
							surfelTopology, 
							surfelPS, 
							nIndex, 
							seedPoint, 
							settings->topSurfelSpacing * triangulationSettings->influence,
							SURFEL_POSITION );
						rt->setRenderColor( 1,1,0 );
						while( !topoSphereSearch.atEnd() ){
							rt->renderSphere( surfelPS->get3f( topoSphereSearch.getCurrentIndex(), SURFEL_POSITION ), cube.getMaxSideLength() / 8.0f );
							topoSphereSearch.next();
						}
						glDisable( GL_LIGHTING );
					}
				}
			}
		}
		delete surfelOctree;
	}
}

void PCCDeformationSurfaceReconstruction::assign( const Object* obj, X4CopyContext *context )
{
	PCCommand::assign( obj, context );
	const PCCDeformationSurfaceReconstruction* o = dynamic_cast<const PCCDeformationSurfaceReconstruction*>(obj);
	x4Assert( o != NULL );
	copyObjectProperties( obj, this );
}

void PCCDeformationSurfaceReconstruction::connectScene( Scene *scene, OnSceneChangeCallback *onChange, QWidget *sceneEditorWidget )
{
	PCCommand::connectScene( scene, onChange, sceneEditorWidget );
	surfels->setScene( scene );
	data->setScene( scene );
}