//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "SmallScaleReconstruction.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "NumericalClassProperty.h"
#include "SeparatorClassProperty.h"
#include "DeformationEvaluator.h"
#include "DynamicLinearAlgebraTools.h"
#include "PointSetKNNQuery.h"
#include "SGSimpleAnimationNode.h"
#include "SGObjectNode.h"
#include "NormalEstimator.h"
#include "NormalUnifier.h"
#include "ProgressWindow.h"
#include "time.h"
#include "ARTopologicalResampling.h"
#include "Resampling.h"
#include "DeformationTools.h"
#include "TopologyRangeSearch.h"
#include <float.h>
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------


IMPLEMENT_X4_CLASS( SmallScaleReconstruction, 0 ) 
{
	BEGIN_CLASS_INIT( SmallScaleReconstruction )
}	

SmallScaleReconstruction::SmallScaleReconstruction()
{
	backendSettings = NULL;
}

SmallScaleReconstruction::~SmallScaleReconstruction()
{
}

void SmallScaleReconstruction::reconstruct(
	UnstructuredInCorePointCloud* originalInUrshapeFrame,
	card32 startFrame,
	card32 endFrame,
	ARSequence* seq, 
	std::vector<PointCloud*>& dataPoints,
	DeformationEvaluator *evaluator, 
	const float32 samplingSpacing,
	SGListNode* listNode )
{
	if ( backendSettings == NULL ) return;
	if ( originalInUrshapeFrame == NULL ) return;
	if ( seq == NULL ) return;
	if ( evaluator == NULL ) return;
	UnstructuredInCorePointCloud* urShapePC = dynamic_cast<UnstructuredInCorePointCloud*>( originalInUrshapeFrame ); //dynamic_cast<UnstructuredInCorePointCloud*>( seq->getUrshape() );
	if ( urShapePC == NULL ) return;
	PointSet* urShapePS = urShapePC->getPointSet();
	AAT URSHAPE_POSITION = urShapePS->getAAT( "position" );
	card32 numFrames = endFrame+1 - startFrame;
	float32 influenceAvgDataPoints = samplingSpacing * 2;
	float32 sqrInfluenceAvgDataPoints = sqr( influenceAvgDataPoints );

	// Create resampled point cloud as patch seed points
	UnstructuredInCorePointCloud* patchPC = dynamic_cast<UnstructuredInCorePointCloud*>( urShapePC->copy() );
	Resampling::downsample( patchPC->getPointSet(), getTopology( patchPC ), backendSettings->patchSamplingSpacing );
	AAT PATCH_POSITION = patchPC->getAAT( "position" );
	PointSet* patchPS = patchPC->getPointSet();

	card32 numParts = patchPC->getNumPoints();
	std::vector<PointSet*> parts; parts.resize( numParts );
	std::vector<DynamicVector<float32>*> partResiduals; partResiduals.resize( numParts );
	
	AAT PART_POSITION;
	for ( unsigned int partIndex = 0; partIndex < numParts; partIndex++ ){
		partResiduals[partIndex] = new DynamicVector<float32>[numFrames];
		parts[partIndex] = new PointSet();
		VertexDescriptor vd;
		vd.pushAttrib( mVAD( "position", 3, VAD::DATA_FORMAT_FLOAT32 ) );
		vd.pushAttrib( mVAD( "color", 3, VAD::DATA_FORMAT_FLOAT32 ) );
		vd.pushAttrib( mVAD( "flags", 1, VAD::DATA_FORMAT_INT32 ) );
		parts[partIndex]->clearAndSetup( 1, 0, &vd );
		PART_POSITION = parts[partIndex]->getAAT( "position" );
	}
	// Assign all urShape points to corresponding part (Voronoi)
	AttachedIndexedOctree* patchPSOctree = PointSetKNNQuery::createOctree( patchPS, 20 );
	PointSetKNNQuery patchPSKNN( patchPS, patchPSOctree, makeVector3f( 0,0,0 ), PATCH_POSITION );
	for ( unsigned int pIndex = 0; pIndex < urShapePS->getNumEntries(); pIndex++ ){
		const Vector3f p = urShapePS->get3f( pIndex, URSHAPE_POSITION );
		patchPSKNN.setTargetPoint( p );
		card32 nIndex;
		bool found;
		patchPSKNN.getNextNearestPointIndex( nIndex, found );
		parts[nIndex]->changeHeight( parts[nIndex]->getNumEntries()+1 );
		parts[nIndex]->set3f( parts[nIndex]->getNumEntries()-1, PART_POSITION, p );
	}
	delete patchPSOctree;
	float32 avgNumPointsPerPart = 0;
	for ( unsigned int partIndex = 0; partIndex < numParts; partIndex++ ){
		avgNumPointsPerPart += parts[partIndex]->getNumEntries();
	}
	avgNumPointsPerPart /= float32( numParts );
	debugOutput << "  PCA-Reconstruction: " << numParts << " parts; avg. #points/part: " << avgNumPointsPerPart << ".\n"; debugOutput.flush();

	// Deform data points into urshape frame
	progressWindow->pushStep( true, "Computing residuals" );
	for ( card32 time = 0; time <= endFrame-startFrame; time++ ){
		//debugOutput << "      deform data points " << card32( time + startFrame ) << ".\n"; debugOutput.flush();
		PointSet* deformedDataPoints = deformDataPoints( 
			dataPoints[time], 
			time-startFrame, 
			dynamic_cast<UnstructuredInCorePointCloud*>( seq->getUrshape() )->getPointSet(), 
			evaluator, 
			samplingSpacing );

		// Compute residuals for all parts
		//debugOutput << "      compute residuals " << card32( time + startFrame ) << ".\n"; debugOutput.flush();
		for ( unsigned int partIndex = 0; partIndex < numParts; partIndex++ ){
			partResiduals[partIndex][time] = computeResidualVector( 
				time, 
				deformedDataPoints, 
				parts[partIndex], 
				samplingSpacing );
		}
		delete deformedDataPoints;
		progressWindow->progress( float32( time ) * 100.0f / float32( numFrames ) );
	}
	progressWindow->popSteps();
	
	// Run PCA analysis
	progressWindow->pushStep( true, "PCA-Analysis" );
	std::vector< std::vector<DynamicVector<float32> > > eigenVectors; eigenVectors.resize( numParts );
	std::vector< DynamicVector<float32> > eigenValues; eigenValues.resize( numParts );
	for ( unsigned int partIndex = 0; partIndex < numParts; partIndex++ ){
		debugOutput << "    Part " << partIndex << "/" << numParts << " with " << parts[partIndex]->getNumEntries() << " points.\n"; debugOutput.flush();
		eigenVectors[partIndex].resize( backendSettings->numEigenVectors );
		DynamicMatrix<float32> evs;
		DynamicVector<float32> mean;
		const card32 dim = parts[partIndex]->getNumEntries() * 3;
		DynamicPCA<float32> pca( partResiduals[partIndex], numFrames );
		mean = pca.calcMean();
		eigenValues[partIndex] = pca.calcEigenValues();
		evs = pca.getEigenVectors();
		for ( unsigned int evIndex = 0; evIndex < backendSettings->numEigenVectors; evIndex++ ){
			eigenVectors[partIndex][evIndex] = evs[evIndex];
		}
		progressWindow->progress( float32( partIndex ) * 100.0f / float32( numParts ) );
	}
	progressWindow->popSteps();

	progressWindow->pushStep( true, "Generate result" );
	SGSimpleAnimationNode* animNode = new SGSimpleAnimationNode();
	animNode->setName( "pca_rec_result" );
	animNode->setVisible( false );
	listNode->addChildNode( animNode );

	UnstructuredInCorePointCloud* usPC = dynamic_cast<UnstructuredInCorePointCloud*>( seq->getUrshape() );
	PointSet* us = usPC->getPointSet();
	AttachedIndexedOctree* usOctree = PointSetKNNQuery::createOctree( us, 32  );
	AAT US_POSITION = us->getAAT( "position" );
	PointSetKNNQuery usKNN( us, usOctree, makeVector3f( 0,0,0 ), US_POSITION );
	for ( unsigned int time = 0; time <= endFrame-startFrame; time++ ){
		evaluator->beginQuery( time );
		// Create deformation of original into frame (create result)
		PointSet* framePS = dynamic_cast<PointSet*>( originalInUrshapeFrame->getPointSet()->copy() );
		AAT FRAME_POSITION = framePS->getAAT( "position" );
		AAT FRAME_NORMAL = framePS->getAAT( "normal" );
		AAT FRAME_FLAGS = framePS->getAAT( "flags" );

		for ( unsigned int pIndex = 0; pIndex < framePS->getNumEntries(); pIndex++ ){
			const Vector3f p = framePS->get3f( pIndex, FRAME_POSITION );
			Vector3f deformedPoint;
			Vector3f deformedNormal;
			deformFromUrshapeToFrame( 
				deformedPoint, 
				deformedNormal, 
				time, 
				p, 
				evaluator, 
				us, 
				getTopology( usPC ),
				usKNN, 
				samplingSpacing );
			framePS->set3f( pIndex, FRAME_POSITION, deformedPoint );
			framePS->set3f( pIndex, FRAME_NORMAL, deformedNormal );
		}	

		char frameID[256];
		itoa( time+startFrame, frameID, 10 );
		SGObjectNode* frameNode = new SGObjectNode();
		frameNode->setName( "frame_" + std::string( frameID ) );
		UnstructuredInCorePointCloud* framePC = new UnstructuredInCorePointCloud();
		framePC->setPointSet( framePS );
		frameNode->clearAndSetup( framePC );
		animNode->addChildNode( frameNode );

		// Compute local deformation from PCA analysis
		card32 globalPIndex = 0;
		for ( unsigned int partIndex = 0; partIndex < numParts; partIndex++ ){
			DynamicVector<float32> delta( eigenValues[partIndex].getDim() );
			delta.setZero();
			// Compute offset for 'time' 
			for ( unsigned int evIndex = 0; evIndex < backendSettings->numEigenVectors; evIndex++ ){
				const float32 c = partResiduals[partIndex][time] * eigenVectors[partIndex][evIndex];
				delta += eigenVectors[partIndex][evIndex] * c;
			}

			AAT PART_POSITION = parts[partIndex]->getAAT( "position" );
			for ( unsigned int pIndex = 0; pIndex < parts[partIndex]->getNumEntries(); pIndex++ ){
				// Point in part
				Vector3f p = parts[partIndex]->get3f( pIndex, PART_POSITION );	
				// Added offset from PCA analysis
				const Vector3f deltaP = makeVector3f( delta[pIndex*3], delta[pIndex*3+1], delta[pIndex*3+2] );
				// Deform point to 'time'
				Vector3f newNormal;
				Vector3f newP;
				deformFromUrshapeToFrame( 
					newP, 
					newNormal, 
					time, 
					p + deltaP, 
					evaluator, 
					us, 
					getTopology( usPC ),
					usKNN, 
					samplingSpacing );
				// Save frame point set
				framePS->set3f( globalPIndex, FRAME_POSITION, newP );
				framePS->set3f( globalPIndex, FRAME_NORMAL, newNormal );
				framePS->set1i( globalPIndex, FRAME_FLAGS, 0 );
				globalPIndex++;
			}
		}
		evaluator->endQuery();
		progressWindow->progress( float32( time ) * 100.0f / float32( numFrames ) );
	}
	delete usOctree;
	progressWindow->popSteps();

	// Debug output
	if ( backendSettings->debugSavePatchPC || backendSettings->debugSaveParts || backendSettings->debugSaveDeformedDataPoints || backendSettings->debugSaveResiduals || backendSettings->debugSaveEigenVectors ){
		if ( backendSettings->debugSavePatchPC ) debugAppendPatchPC( listNode, patchPC );
		if ( backendSettings->debugSaveParts ) debugAppendParts( listNode, parts );
		//if ( backendSettings->debugSaveDeformedDataPoints ) debugAppendDeformedDataPoints( listNode, deformedDataPoints );
		if ( backendSettings->debugSaveResiduals ) debugAppendResiduals( listNode, parts, partResiduals, numFrames );
		if ( backendSettings->debugSaveEigenVectors ) debugAppendEigenVectors( listNode, parts, eigenVectors, eigenValues );
	}

	// Cleanup
	for ( unsigned int n = 0; n < numParts; n++ ){
		delete parts[n];
		delete [] partResiduals[n];
	}
	parts.clear();
	partResiduals.clear();
	/*
	for ( card32 time = 0; time <= endFrame-startFrame; time++ ){
		delete deformedDataPoints[time];
	}
	deformedDataPoints.clear();
	*/
	delete patchPC;
}

// Deform a point into frame 'time' from urshape frame
void SmallScaleReconstruction::deformFromUrshapeToFrame( 
	Vector3f& deformedPoint,
	Vector3f& deformedNormal,
	const card32 time, 
	const Vector3f p,
	DeformationEvaluator *evaluator,
	PointSet* urShapePS,
	InCorePCTopologyGraph* urShapeTopology,
	PointSetKNNQuery& urShapeKNN, 
	const float32 samplingSpacing )
{
	AAT URSHAPE_POSITION = urShapePS->getAAT( "position" );
	AAT URSHAPE_NORMAL = urShapePS->getAAT( "normal" );
	AAT URSHAPE_TANGENT_U = urShapePS->getAAT( "tangent_u" );
	AAT URSHAPE_TANGENT_V = urShapePS->getAAT( "tangent_v" );
	urShapeKNN.setTargetPoint( p );
	//TopNeighborhoodKNNSearch topoKNN( urShapePS, urShapeTopology, 0, URSHAPE_POSITION );
	card32 nIndex;
	bool found;
	float32 summedWeight = 0;
	urShapeKNN.getNextNearestPointIndex( nIndex, found );
	//topoKNN.reset( nIndex );
	deformedPoint = makeVector3f( 0,0,0 );
	deformedNormal = makeVector3f( 0,0,0 );
	ARConstraintAnchor at;
	for ( unsigned int n = 0; n < 5; n++ ){
		const Vector3f urShapeP = urShapePS->get3f( nIndex, URSHAPE_POSITION );
		const float32 sqrD = ( urShapeP - p ) * ( urShapeP - p );
		const float32 weight = exp( - sqrD / sqr( samplingSpacing ) );
		summedWeight += weight;
		at.pos = urShapeP;
		at.topologicalAnchor = nIndex;
		const Vector3f deformedUrShapeP =  evaluator->getPosition( at );
		const Matrix3f jac = evaluator->getSpatialGradient( at );
		const Matrix3f transform = invertMatrix( jac ).transpose();
		Matrix3f T_urshape, T_frame;
		T_urshape[0] = urShapePS->get3f( nIndex, URSHAPE_TANGENT_U );
		T_urshape[1] = urShapePS->get3f( nIndex, URSHAPE_TANGENT_V );
		T_urshape[2] = urShapePS->get3f( nIndex, URSHAPE_NORMAL );
		T_frame[0] = jac * urShapePS->get3f( nIndex, URSHAPE_TANGENT_U );
		T_frame[1] = jac * urShapePS->get3f( nIndex, URSHAPE_TANGENT_V );
		T_frame[2] = transform * urShapePS->get3f( nIndex, URSHAPE_NORMAL );
		Vector3f offset = p - urShapeP;
		offset = T_frame * T_urshape.transpose() * offset;
		deformedPoint += ( deformedUrShapeP + offset ) * weight;
		deformedNormal += T_frame[2] * weight;
		//topoKNN.next();
		//nIndex = topoKNN.getCurrentIndex();
		urShapeKNN.getNextNearestPointIndex( nIndex, found );
	}
	deformedPoint /= summedWeight;
	deformedNormal.normalize();
}

DynamicVector<float32> SmallScaleReconstruction::computeResidualVector( 
	const card32 time,
	PointSet* deformedDataPS,
	PointSet* samplesPS,
	const float32 samplingSpacing )
{
	DynamicVector<float32> res( samplesPS->getNumEntries() * 3 );
	if ( deformedDataPS == NULL ) return res;
	if ( samplesPS == NULL ) return res;
	AAT SAMPLES_POSITION = samplesPS->getAAT( "position" );
	AAT DEFORMED_DATA_POSITION = deformedDataPS->getAAT( "position" );
	card32 nIndex;
	bool found;
	AttachedIndexedOctree* deformedDataPSOctree = PointSetKNNQuery::createOctree( deformedDataPS, 32 );
	PointSetKNNQuery deformedDataPointsKNN( deformedDataPS, deformedDataPSOctree, makeVector3f( 0,0,0 ), DEFORMED_DATA_POSITION );

	const float32 sqrInfluenceAvgDataPoints = sqr( samplingSpacing * backendSettings->dataPointsInfluence );

	// Compute residuii
	for ( unsigned int sIndex = 0; sIndex < samplesPS->getNumEntries(); sIndex++ ){
		const Vector3f p = samplesPS->get3f( sIndex, SAMPLES_POSITION );
		deformedDataPointsKNN.setTargetPoint( p );
		deformedDataPointsKNN.getNextNearestPointIndex( nIndex, found );
		Vector3f np = deformedDataPS->get3f( nIndex, DEFORMED_DATA_POSITION );
		float32 sqrD = ( np - p ) * ( np - p );
		float32 summedWeight = 0;
		Vector3f avgDataPoint = makeVector3f( 0,0,0 );
		card32 numDataPoints = 0;
		while( found && ( sqrD < sqrInfluenceAvgDataPoints ) ){
			float32 weight = exp( - sqrD / sqrInfluenceAvgDataPoints );
			avgDataPoint += np * weight;
			summedWeight += weight;
			deformedDataPointsKNN.getNextNearestPointIndex( nIndex, found );
			if ( found ){
				np = deformedDataPS->get3f( nIndex, DEFORMED_DATA_POSITION );
				sqrD = ( np - p ) * ( np - p );
			}
			numDataPoints++;
		}
		if ( numDataPoints > 0 ){
			avgDataPoint /= summedWeight;
		} else {
			avgDataPoint = p;
		}
		Vector3f residuum = avgDataPoint - p;
		res[sIndex*3] = residuum[0];
		res[sIndex*3+1] = residuum[1];
		res[sIndex*3+2] = residuum[2];
	}
	delete deformedDataPSOctree;
	return res;
}

PointSet* SmallScaleReconstruction::deformDataPoints(
	PointCloud* dataPoints,
	const card32 time,
	PointSet* urShapePS,
	DeformationEvaluator *evaluator,
	const float32 samplingSpacing )
{
	evaluator->beginQuery( time );

	// Data points (live in frame 'time')
	if ( dataPoints == NULL ) return NULL;
	SceneObjectIterator* it = dataPoints->createIterator( SOIT::CAP_BASIC_PC );
	BasicPointCloudIterator* basicIt = dynamic_cast<BasicPointCloudIterator*>( it );
	AAT DATA_POSITION = dataPoints->getAAT( "position" );
	const card32 numDataPoints = dataPoints->getNumPoints();

	// Deformed data points (live in urshape time frame); needs to be deformed
	PointSet* deformedDataPS = new PointSet();
	VertexDescriptor vd;
	vd.pushAttrib( mVAD( "position", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	vd.pushAttrib( mVAD( "flags", 1, VAD::DATA_FORMAT_INT32 ) );
	deformedDataPS->clearAndSetup( 1, numDataPoints, &vd );
	AAT DEFORMED_DATA_POSITION = deformedDataPS->getAAT( "position" );

	// Urshape (lives in urshape time frame)
	if ( urShapePS == NULL ) return NULL;
	AAT URSHAPE_POSITION = urShapePS->getAAT( "position" );
	AAT URSHAPE_TANGENT_U = urShapePS->getAAT( "tangent_u" );
	AAT URSHAPE_TANGENT_V = urShapePS->getAAT( "tangent_v" );
	AAT URSHAPE_NORMAL = urShapePS->getAAT( "normal" );
	card32 numUrShapePoints = urShapePS->getNumEntries();

	// Deformed urshape (lives in frame 'time'); needs to be deformed
	PointSet* deformedUrShapePS = new PointSet();
	VertexDescriptor deformedUrShapeVD;
	deformedUrShapeVD.pushAttrib( mVAD( "position", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	deformedUrShapeVD.pushAttrib( mVAD( "flags", 1, VAD::DATA_FORMAT_INT32 ) );
	deformedUrShapeVD.pushAttrib( mVAD( "normal", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	deformedUrShapeVD.pushAttrib( mVAD( "tangent_u", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	deformedUrShapeVD.pushAttrib( mVAD( "tangent_v", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	deformedUrShapePS->clearAndSetup( 1, numUrShapePoints, &deformedUrShapeVD );
	AAT DEFORMED_URSHAPE_POSITION = deformedUrShapePS->getAAT( "position" );
	AAT DEFORMED_URSHAPE_NORMAL = deformedUrShapePS->getAAT( "normal" );
	AAT DEFORMED_URSHAPE_TANGENT_U = deformedUrShapePS->getAAT( "tangent_u" );
	AAT DEFORMED_URSHAPE_TANGENT_V = deformedUrShapePS->getAAT( "tangent_v" );
	for ( unsigned int sIndex = 0; sIndex < deformedUrShapePS->getNumEntries(); sIndex++ ){
		const Vector3f p = urShapePS->get3f( sIndex, URSHAPE_POSITION );
		ARConstraintAnchor at;
		at.pos = p;
		at.topologicalAnchor = sIndex;
		Matrix3f jac = evaluator->getSpatialGradient( at );
		Matrix3f transform = invertMatrix( jac ).transpose();
		const Vector3f deform_urshape_normal = transform * urShapePS->get3f( sIndex, URSHAPE_NORMAL );
		const Vector3f deform_urshape_tangent_u = jac * urShapePS->get3f( sIndex, URSHAPE_TANGENT_U );
		const Vector3f deform_urshape_tangent_v = jac * urShapePS->get3f( sIndex, URSHAPE_TANGENT_V );
		deformedUrShapePS->set3f( sIndex, DEFORMED_URSHAPE_POSITION, evaluator->getPosition( at ) );
		deformedUrShapePS->set3f( sIndex, DEFORMED_URSHAPE_NORMAL, deform_urshape_normal );
		deformedUrShapePS->set3f( sIndex, DEFORMED_URSHAPE_TANGENT_U, deform_urshape_tangent_u );
		deformedUrShapePS->set3f( sIndex, DEFORMED_URSHAPE_TANGENT_V, deform_urshape_tangent_v );
	}
	AttachedIndexedOctree* deformedUrShapeOctree = PointSetKNNQuery::createOctree( deformedUrShapePS, 32 );
	PointSetKNNQuery deformedUrShapeKNN( deformedUrShapePS, deformedUrShapeOctree, makeVector3f( 0,0,0 ), URSHAPE_POSITION );

	card32 nIndex;
	bool found;
	ARConstraintAnchor at;
	
	basicIt->reset();
	card32 pIndex = 0; 
	while( !basicIt->atEnd() ){
		const Vector3f dataP = basicIt->get3f( DATA_POSITION );
		deformedUrShapeKNN.setTargetPoint( dataP );

		Vector3f deformedDataP = makeVector3f( 0,0,0 );
		float32 summedWeight = 0;
		for ( unsigned int n = 0; n < 5; n++ ){
		
			deformedUrShapeKNN.getNextNearestPointIndex( nIndex, found );
			const Vector3f deformedUrShapeP = deformedUrShapePS->get3f( nIndex, DEFORMED_URSHAPE_POSITION );
			const Vector3f urShapeP = urShapePS->get3f( nIndex, URSHAPE_POSITION );

			// Compute local coordinate system at urshape point
			Matrix3f T_urshape;
			T_urshape[0] = urShapePS->get3f( nIndex, URSHAPE_TANGENT_U );
			T_urshape[1] = urShapePS->get3f( nIndex, URSHAPE_TANGENT_V );
			T_urshape[2] = urShapePS->get3f( nIndex, URSHAPE_NORMAL );

			// Compute local coordinate system at deformed urshape point
			Matrix3f T_deform_urshape;
			T_deform_urshape[0] = deformedUrShapePS->get3f( nIndex, DEFORMED_URSHAPE_TANGENT_U );
			T_deform_urshape[1] = deformedUrShapePS->get3f( nIndex, DEFORMED_URSHAPE_TANGENT_V );
			T_deform_urshape[2] = deformedUrShapePS->get3f( nIndex, DEFORMED_URSHAPE_NORMAL );

			// Transform data point into urshape coordinate system assuming rigid transformation close to urshape point
			Vector3f offset = dataP - deformedUrShapeP;
			offset = T_urshape * T_deform_urshape.transpose() * offset;
			const float32 sqrD = ( deformedUrShapeP - dataP ) * ( deformedUrShapeP - dataP );
			float32 weight = exp( - sqrD / sqr( samplingSpacing ) ) ;
			if ( ( weight < 1e-6 ) && ( n == 1 ) ){
				// Spacial case: no urshape point close to data
				weight = 1.0f;
			}
			summedWeight += weight;
			deformedDataP += ( urShapeP + offset ) * weight;
			
		}
		deformedDataP /= summedWeight;

		if ( _isnan( deformedDataP[0] ) || _isnan( deformedDataP[1] ) || _isnan( deformedDataP[2] ) || !_finite( deformedDataP[0] ) || !_finite( deformedDataP[1] ) || !_finite( deformedDataP[2] ) ){
			debugOutput << "invalid point: " << deformedDataP << ".\n"; debugOutput.flush();
			debugOutput << "invalid summedWeight: " << summedWeight << ".\n"; debugOutput.flush();
		} 

		deformedDataPS->set3f( pIndex, DEFORMED_DATA_POSITION, deformedDataP );

		basicIt->next();
		pIndex++;
	}
	evaluator->endQuery();
	delete it;
	delete deformedUrShapePS;
	delete deformedUrShapeOctree;
	return deformedDataPS;
}

void SmallScaleReconstruction::debugAppendDeformedDataPoints( 
	SGListNode* listNode, 
	std::vector<PointSet*>& deformedDataPoints )
{
	SGAnimationNode* animNode = new SGSimpleAnimationNode();
	animNode->setName( "deformedData" );
	listNode->addChildNode( animNode );
	animNode->setVisible( false );

	// Deformed data points
	for ( unsigned int frameIndex = 0; frameIndex < deformedDataPoints.size(); frameIndex++ ){
		PointSet* deformedDataPS = new PointSet();
		VertexDescriptor vd;
		vd.pushAttrib( mVAD( "position", 3, VAD::DATA_FORMAT_FLOAT32 ) );
		vd.pushAttrib( mVAD( "normal", 3, VAD::DATA_FORMAT_FLOAT32 ) );
		vd.pushAttrib( mVAD( "color", 3, VAD::DATA_FORMAT_FLOAT32 ) );
		vd.pushAttrib( mVAD( "flags", 1, VAD::DATA_FORMAT_INT32 ) );
		deformedDataPS->clearAndSetup( 1, deformedDataPoints[frameIndex]->getNumEntries(), &vd );
		UnstructuredInCorePointCloud* pc = new UnstructuredInCorePointCloud();
		pc->setPointSet( deformedDataPS );
		pc->setMaterialIndex( 0 );
		SGObjectNode* node = new SGObjectNode();
		node->clearAndSetup( pc );
		char frameID[256];
		itoa( frameIndex, frameID, 10 );
		node->setName( "frame" + std::string( frameID ) );
		animNode->addChildNode( node );
		AAT POSITION = deformedDataPS->getAAT( "position" );
		AAT COLOR = deformedDataPS->getAAT( "color" );
		AAT DEFORMED_DATA_POSITION = deformedDataPS->getAAT( "position" );
		for ( unsigned int pIndex = 0; pIndex < deformedDataPoints[frameIndex]->getNumEntries(); pIndex++ ){
			deformedDataPS->set3f( pIndex, POSITION, deformedDataPoints[frameIndex]->get3f( pIndex, DEFORMED_DATA_POSITION ) );
			deformedDataPS->set3f( pIndex, COLOR, makeVector3f( 0,0,1 ) );
		}
	}
}

void SmallScaleReconstruction::debugAppendResiduals( 
	SGListNode* listNode, 
	std::vector<PointSet*>& parts, 
	std::vector< DynamicVector<float32>* >& partResiduals,
	const card32 numFrames )
{
	AAT PART_POSITION = parts[0]->getAAT( "position" );
	for( unsigned int partIndex = 0; partIndex < parts.size(); partIndex++ ){	
		SGAnimationNode* animNode = new SGSimpleAnimationNode();
		char partID[256];
		itoa( partIndex, partID, 10 );
		animNode->setName( "residuals_part_" + std::string( partID ) );
		listNode->addChildNode( animNode );
		animNode->setVisible( false );
		for ( unsigned int frameIndex = 0; frameIndex < numFrames; frameIndex++ ){
			PointSet* ps = new PointSet();
			VertexDescriptor vd;
			vd.pushAttrib( mVAD( "position", 3, VAD::DATA_FORMAT_FLOAT32 ) );
			vd.pushAttrib( mVAD( "normal", 3, VAD::DATA_FORMAT_FLOAT32 ) );
			vd.pushAttrib( mVAD( "color", 3, VAD::DATA_FORMAT_FLOAT32 ) );
			vd.pushAttrib( mVAD( "flags", 1, VAD::DATA_FORMAT_INT32 ) );
			ps->clearAndSetup( 1, parts[partIndex]->getNumEntries(), &vd );
			UnstructuredInCorePointCloud* pc = new UnstructuredInCorePointCloud();
			pc->setPointSet( ps );
			pc->setMaterialIndex( 0 );
			SGObjectNode* node = new SGObjectNode();
			node->clearAndSetup( pc );
			char partID[256];
			char frameID[256];
			itoa( frameIndex, frameID, 10 );
			itoa( partIndex, partID, 10 );
			node->setName( "frame_" + std::string( frameID ) );
			animNode->addChildNode( node );
			AAT EV_POSITION = ps->getAAT( "position" );
			AAT EV_COLOR = ps->getAAT( "color" );
			AAT EV_NORMAL = ps->getAAT( "normal" );
			for ( unsigned int pIndex = 0; pIndex < parts[partIndex]->getNumEntries() ; pIndex++ ){
				Vector3f res = makeVector3f( 
					partResiduals[partIndex][frameIndex][pIndex*3],
					partResiduals[partIndex][frameIndex][pIndex*3+1],
					partResiduals[partIndex][frameIndex][pIndex*3+2]
				);
				ps->set3f( pIndex, EV_POSITION, parts[partIndex]->get3f( pIndex, PART_POSITION ) );
				ps->set3f( pIndex, EV_NORMAL, res );
			}
		}
	}
}

void SmallScaleReconstruction::debugAppendEigenVectors( 
	SGListNode* listNode, 
	std::vector<PointSet*>& parts,
	std::vector< std::vector< DynamicVector<float32> > >& eigenVectors, 
	std::vector< DynamicVector<float32> >& eigenValues )
{
	AAT PART_POSITION = parts[0]->getAAT( "position" );
	for( unsigned int partIndex = 0; partIndex < eigenVectors.size(); partIndex++ ){	
		SGAnimationNode* animNode = new SGSimpleAnimationNode();
		char partID[256];
		itoa( partIndex, partID, 10 );
		animNode->setName( "eigenvectors_part_" + std::string( partID ) );
		listNode->addChildNode( animNode );
		animNode->setVisible( false );
		for ( unsigned int evIndex = 0; evIndex < backendSettings->numEigenVectors; evIndex++ ){
			PointSet* evPS = new PointSet();
			VertexDescriptor vd;
			vd.pushAttrib( mVAD( "position", 3, VAD::DATA_FORMAT_FLOAT32 ) );
			vd.pushAttrib( mVAD( "normal", 3, VAD::DATA_FORMAT_FLOAT32 ) );
			vd.pushAttrib( mVAD( "color", 3, VAD::DATA_FORMAT_FLOAT32 ) );
			vd.pushAttrib( mVAD( "flags", 1, VAD::DATA_FORMAT_INT32 ) );
			evPS->clearAndSetup( 1, parts[partIndex]->getNumEntries(), &vd );
			UnstructuredInCorePointCloud* pc = new UnstructuredInCorePointCloud();
			pc->setPointSet( evPS );
			pc->setMaterialIndex( 0 );
			SGObjectNode* node = new SGObjectNode();
			node->clearAndSetup( pc );
			char evID[256];
			itoa( evIndex, evID, 10 );
			node->setName( "_ev_" + std::string( evID ) );
			animNode->addChildNode( node );
			AAT EV_POSITION = evPS->getAAT( "position" );
			AAT EV_COLOR = evPS->getAAT( "color" );
			AAT EV_NORMAL = evPS->getAAT( "normal" );
			for ( unsigned int pIndex = 0; pIndex < parts[partIndex]->getNumEntries() ; pIndex++ ){
				evPS->set3f( pIndex, EV_POSITION, parts[partIndex]->get3f( pIndex, PART_POSITION ) );
				Vector3f ev_i = makeVector3f( 
					eigenVectors[partIndex][evIndex][pIndex*3], 
					eigenVectors[partIndex][evIndex][pIndex*3+1], 
					eigenVectors[partIndex][evIndex][pIndex*3+2] ) 
					* eigenValues[partIndex][evIndex];
				evPS->set3f( pIndex, EV_NORMAL, ev_i );
			}
		}
	}
}

void SmallScaleReconstruction::debugAppendPatchPC( 
	SGListNode* listNode, 
	UnstructuredInCorePointCloud* resampledPatchPC )
{
	SGObjectNode* node = new SGObjectNode();
	node->setName( "patchPC" );
	node->clearAndSetup( dynamic_cast<UnstructuredInCorePointCloud*>( resampledPatchPC->copy() ) );
	node->setVisible( false );
	listNode->addChildNode( node );
}

void SmallScaleReconstruction::debugAppendParts( 
	SGListNode* listNode, 
	std::vector<PointSet*>& parts )
{
	SGListNode* l2 = new SGListNode();
	l2->setName( "parts" );
	listNode->addChildNode( l2 );
	l2->setVisible( false );

	for ( unsigned int partIndex = 0; partIndex < parts.size(); partIndex++ ){
		SGObjectNode* node = new SGObjectNode();
		char partID[256];
		itoa( partIndex, partID, 10 );
		node->setName( "part_" + std::string( partID ) );
		UnstructuredInCorePointCloud* pc = new UnstructuredInCorePointCloud();
		pc->setMaterialIndex( 2 );
		PointSet* ps = dynamic_cast<PointSet*>( parts[partIndex]->copy() );
		pc->setPointSet( ps );
		node->clearAndSetup( pc );
		l2->addChildNode( node );	

		AAT COLOR = ps->getAAT( "color" );
		const Vector3f color = makeVector3f( rand()%100/100.0f, rand()%100/100.0f, rand()%100/100.0f );
		for ( unsigned int pIndex = 0; pIndex < ps->getNumEntries(); pIndex++ ){
			ps->set3f( pIndex, COLOR, color );
		}
	}	
}

/*
void SmallScaleReconstruction::testDeform( PointSet* ps, const card32 time )
{
	if ( ps == NULL ) return;
	AAT POSITION = ps->getAAT( "position" );
	for ( unsigned int pIndex = 0; pIndex < ps->getNumEntries(); pIndex++ ){
		Vector3f p = ps->get3f( pIndex, POSITION );
		p = makeVector3f( p[0], time * sqr( p[0] ) * 0.2f, p[2] );
		ps->set3f( pIndex, POSITION, p );
	}
}

void SmallScaleReconstruction::testDeformHighFrequency( PointSet* ps, const card32 time )
{
	if ( ps == NULL ) return;
	AAT POSITION = ps->getAAT( "position" );
	AAT NORMAL = ps->getAAT( "normal" );
	for ( unsigned int pIndex = 0; pIndex < ps->getNumEntries(); pIndex++ ){
		Vector3f p = ps->get3f( pIndex, POSITION );
		Vector3f normal = ps->get3f( pIndex, NORMAL );
		float32 offset = 0.1f * sin( 0.5f * float32( time ) ) * exp( -sqr( p[0] ) / sqr( 0.03f ) );
		ps->set3f( pIndex, POSITION, p + makeVector3f( 0,1,0 ) * offset );
	}
}

SGSimpleAnimationNode* SmallScaleReconstruction::createTestExample( const card32 numFrames, const card32 numPoints )
{
	SGSimpleAnimationNode* animNode = new SGSimpleAnimationNode();
	animNode->setName( "data" );
	for ( unsigned int t = 0; t < numFrames; t++ ){
		PointSet* ps = CmdObjGeneratePC::createBox( numPoints * 2 );
		AAT POSITION = ps->getAAT( "position" );
		card32 currentIndex = 0;
		for ( unsigned int pIndex = 0; pIndex < ps->getNumEntries(); pIndex++ ){
			const Vector3f p = ps->get3f( pIndex, POSITION );
			if ( p[1] < -0.25 ){
				ps->set3f( currentIndex, POSITION, p );
				currentIndex++;
			}
		}
		ps->changeHeight( currentIndex );
		testDeformHighFrequency( ps, t );

		UnstructuredInCorePointCloud* pc = new UnstructuredInCorePointCloud();
		pc->setPointSet( ps );
		SGObjectNode* node = new SGObjectNode();
		node->clearAndSetup( pc );
		char id[256];
		itoa( t+1, id, 10 );
		node->setName( "frame" + std::string( id ) );
		animNode->addChildNode( node );
	}
	return animNode;
}
*/