//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "SurfelOptimizer.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "SGEShape.h"
#include "SGESurfelGraph.h"
#include "PointSet.inline.h"
#include "ARConstraints.h"
#include "PointSetKNNQuery.h"
#include "IterativeSolvers.h"
#include "IterativeSolvers.inline.h"
#include <time.h>
#include "InCorePCTopologyGraph.h"
#include "LocalSurfaceFunction.h"
#include "ProgressWindow.h"
#include "TopologyRangeSearch.h"
#include "SurfelOptEvaluator.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

IMPLEMENT_X4_CLASS( SurfelOptimizer, 0 ) 
{
	BEGIN_CLASS_INIT( SurfelOptimizer )
}

SurfelOptimizer::SurfelOptimizer() 
{
	surfelOctree = NULL;
	startBuildMatrixTime = 0;
	settings = NULL;
	context = NULL;
}

SurfelOptimizer::~SurfelOptimizer() 
{
}

void SurfelOptimizer::addGeneralConstraint( const GeneralARConstraint* generalConstraint )
{
	switch (generalConstraint->getConstraintTypesHint()) {
		case GeneralARConstraint::POSITION_CONSTRAINT: {
			addPositionConstraint( (PositionARConstraint*)generalConstraint );
			break;
		}

		case GeneralARConstraint::HALF_SPACE_CONSTRAINT:{
			addHalfSpaceConstraint( (ARHalfSpaceConstraint*)generalConstraint );
		}
		
		default: {
			throw ERangeCheck("SurfelOptimizer::addGeneralConstraint() - unknown constraint type.");
		}
	}
}

void SurfelOptimizer::addSpatialDerivativeConstraint(const SpatialDerivativeARConstraint *c) {
   notImplemented();
}

void SurfelOptimizer::addSimpleSpatialDerivativeConstraint(const SimpleSpatialDerivativeARConstraint *c) {
   notImplemented();
}


void SurfelOptimizer::beginConstraints()
{
	startBuildMatrixTime = clock();
	if ( settings == NULL ) {
		if ( settings->surfaceRecDebugOutput ){
			debugOutput << "beginConstraints: error.\n"; debugOutput.flush();
		}
		return;
	}
	if ( context == NULL ) {
		if ( settings->surfaceRecDebugOutput ){
			debugOutput << "beginConstraints: error.\n"; debugOutput.flush();
		}
		return;
	}
	PointSet* surfelPS = context->getSurfels();
	card32 problemSize = SURFEL_NUM_BASIS_FUNCTIONS * surfelPS->getNumEntries();
	H.setRows( problemSize );
	variables.setDim( problemSize );
	b.setDim( problemSize );
	gradient.setDim( problemSize );
	AAT COEFFS = surfelPS->getAAT( "coeffs" );
	for( unsigned int pIndex = 0; pIndex < surfelPS->getNumEntries(); pIndex++ ){
		StaticVector<float32,SURFEL_NUM_BASIS_FUNCTIONS> coeffs = surfelPS->get<float32,SURFEL_NUM_BASIS_FUNCTIONS>( pIndex, COEFFS );
		card32 vectorIndex = pIndex * SURFEL_NUM_BASIS_FUNCTIONS;
		for ( unsigned int n = 0; n < SURFEL_NUM_BASIS_FUNCTIONS; n++ ){
			variables[vectorIndex+n] = coeffs[n];
			b[vectorIndex+n] = 0;
			gradient[vectorIndex+n] = 0;
			H[vectorIndex+n].clear();
		}
	}
	f = 0;
	if ( surfelOctree != NULL ) delete surfelOctree;
	surfelOctree = PointSetKNNQuery::createOctree( surfelPS, 32 );
}

void SurfelOptimizer::addPositionConstraint( const PositionARConstraint* posConstraint )
{
	/*
	if ( posConstraint == NULL ){
		if ( settings->surfaceRecDebugOutput ){
			debugOutput << "addPositionConstraint: error.\n"; debugOutput.flush();
		}
		return;
	}
	if ( settings == NULL ) {
		if ( settings->surfaceRecDebugOutput ){
			debugOutput << "addPositionConstraint: error.\n"; debugOutput.flush();
		}
		return;
	}
	if ( context == NULL ) {
		if ( settings->surfaceRecDebugOutput ){
			debugOutput << "addPositionConstraint: error.\n"; debugOutput.flush();
		}
		return;
	}
	*/


	PointSet* surfelPS = context->getSurfels();
	const Vector3f& p = posConstraint->getToPoint();
	AAT POSITION = surfelPS->getAAT( "position" );
	AAT NORMAL = surfelPS->getAAT( "normal" );
	AAT TANGENT_U = surfelPS->getAAT( "tangent_u" );
	AAT TANGENT_V = surfelPS->getAAT( "tangent_v" );
	AAT COEFFS = surfelPS->getAAT( "coeffs" );
	
	const ARConstraintAnchor& anchor = posConstraint->getAnchor();
	card32 anchorIndex = anchor.topologicalAnchor;
	Vector3f anchorPos = anchor.pos;
	if ( anchorIndex == ARConstraintAnchor::NO_TOP_ANCHOR ){
		// If anchor index not set, then set it to the closest surfel
		card32 nIndex;
		bool found;
		PointSetKNNQuery knn( surfelPS, surfelOctree, p, POSITION );
		knn.setTargetPoint( anchor.pos );
		knn.getNextNearestPointIndex( nIndex, found );
		anchorIndex = nIndex;
		anchorPos = surfelPS->get3f( nIndex, POSITION );
	}
		
	Matrix3f T;
	const Vector3f np = surfelPS->get3f( anchorIndex, POSITION );
	T[0] = surfelPS->get3f( anchorIndex, TANGENT_U );
	T[1] = surfelPS->get3f( anchorIndex, TANGENT_V );
	const Vector3f nNormal = surfelPS->get3f( anchorIndex, NORMAL );
	T[2] = nNormal;
	T = T.transpose();
	const Vector3f pLocal = T * ( posConstraint->toPoint - np ) / settings->topSurfelSpacing;
	const StaticVector<float32,SURFEL_NUM_BASIS_FUNCTIONS>& coeffs = surfelPS->get<float32,SURFEL_NUM_BASIS_FUNCTIONS>( anchorIndex, COEFFS );

	float32 lambda_lwicp = 1;
	if ( settings->shapeSolveLocallyWeightedICP ){
		const float32 sqrDistFromToPoint = ( np - posConstraint->toPoint ) * ( np - posConstraint->toPoint );
		lambda_lwicp = exp( - sqrDistFromToPoint / sqr( settings->topSurfelSpacing ) );
	}
	const float32 weight_quadric = nNormal* ( posConstraint->covariance * nNormal );
	const float32 Z = lambda_lwicp * weight_quadric;

	StaticMatrix<float32,SURFEL_NUM_BASIS_FUNCTIONS, SURFEL_NUM_BASIS_FUNCTIONS> H_i;
	StaticVector<float32, SURFEL_NUM_BASIS_FUNCTIONS> g_i;
	float32 f_i;
	float32 u_jk = pLocal[0];
	float32 v_jk = pLocal[1];
	float32 n_jk = pLocal[2];
	f_i = Z * sqr( LocalSurfaceFunction<SURFEL_NUM_BASIS_FUNCTIONS>::f( coeffs, u_jk, v_jk ) - n_jk );
	for ( unsigned int bfIndex = 0; bfIndex < SURFEL_NUM_BASIS_FUNCTIONS; bfIndex++ ){
		g_i[bfIndex] = Z * LocalSurfaceFunction<SURFEL_NUM_BASIS_FUNCTIONS>::f_i( u_jk, v_jk, bfIndex ) * ( LocalSurfaceFunction<SURFEL_NUM_BASIS_FUNCTIONS>::f( coeffs, u_jk, v_jk ) - n_jk );
		for ( unsigned int j = 0; j < SURFEL_NUM_BASIS_FUNCTIONS; j++ ){
			H_i[bfIndex][j] = Z * LocalSurfaceFunction<SURFEL_NUM_BASIS_FUNCTIONS>::f_i( u_jk, v_jk, bfIndex ) * LocalSurfaceFunction<SURFEL_NUM_BASIS_FUNCTIONS>::f_i( u_jk, v_jk, j );
		}
		
	}

	f += f_i;
	card32 vectorIndex = anchorIndex * SURFEL_NUM_BASIS_FUNCTIONS;
	for ( unsigned int i = 0; i < SURFEL_NUM_BASIS_FUNCTIONS; i++ ){
		gradient[vectorIndex+i] += g_i[i];
		for ( unsigned int j = 0; j < SURFEL_NUM_BASIS_FUNCTIONS; j++ ){
			H[vectorIndex+i].addEntry( vectorIndex+j, H_i[i][j] );
		}
	}
}

void SurfelOptimizer::addHalfSpaceConstraint( const ARHalfSpaceConstraint* hsConstraint )
{
	if ( hsConstraint == NULL ){
		if ( settings->surfaceRecDebugOutput ){
			debugOutput << "addHalfSpaceConstraint: error.\n"; debugOutput.flush();
		}
		return;
	}
	if ( settings == NULL ) {
		if ( settings->surfaceRecDebugOutput ){
			debugOutput << "addHalfSpaceConstraint: error.\n"; debugOutput.flush();
		}
		return;
	}
	if ( context == NULL ) {
		if ( settings->surfaceRecDebugOutput ){
			debugOutput << "addHalfSpaceConstraint: error.\n"; debugOutput.flush();
		}
		return;
	}
	PointSet* surfelPS = context->getSurfels();
	AAT POSITION = surfelPS->getAAT( "position" );
	AAT NORMAL = surfelPS->getAAT( "normal" );
	AAT TANGENT_U = surfelPS->getAAT( "tangent_u" );
	AAT TANGENT_V = surfelPS->getAAT( "tangent_v" );
	AAT COEFFS = surfelPS->getAAT( "coeffs" );
	PointSetKNNQuery knn( surfelPS, surfelOctree, makeVector3f( -1, -1, -1 ), POSITION );
	card32 nIndex;
	bool found;

	const ARConstraintAnchor& anchor = hsConstraint->getAnchor();
	card32 anchorIndex = anchor.topologicalAnchor;
	Vector3f anchorPos = anchor.pos;
	if ( anchorIndex == ARConstraintAnchor::NO_TOP_ANCHOR ){
		// If anchor index not set, then set it to the closest surfel
		knn.setTargetPoint( anchor.pos );
		knn.getNextNearestPointIndex( nIndex, found );
		anchorIndex = nIndex;
		anchorPos = surfelPS->get3f( nIndex, POSITION );
	}

	const Vector3f hsPoint = hsConstraint->getHalfSpaceFootPoint();
	const Vector3f hsNormal = hsConstraint->getHalfSpaceNormal();
	const float32 sqrSurfelSpacing = sqr( settings->topSurfelSpacing  );
	TopNeighborhoodSphereSearch topoSphereSearch;
	topoSphereSearch.setup( 
		context->getTopologyGraph(), 
		surfelPS, 
		anchorIndex, 
		anchorPos, 
		settings->topSurfelSpacing * 3.0f, 
		POSITION );
	std::vector<card32> influencedSurfels; influencedSurfels.clear();
	while( !topoSphereSearch.atEnd() ){
		const card32 nIndex = topoSphereSearch.getCurrentIndex();
		const Vector3f& np = surfelPS->get3f( nIndex, POSITION );
		const float32 sqrD = ( np - anchorPos ) * ( np - anchorPos );
		if ( sqrD < sqr( settings->topSurfelSpacing * 2.0f ) ){
			influencedSurfels.push_back( nIndex );
		}
		topoSphereSearch.next();
	}

	if ( influencedSurfels.size() > 0 ){
		for ( unsigned int n = 0; n < influencedSurfels.size(); n++ ){
			const card32 nIndex = influencedSurfels[n];
			Matrix3f T;
			const Vector3f np = surfelPS->get3f( nIndex, POSITION );
			T[0] = surfelPS->get3f( nIndex, TANGENT_U );
			T[1] = surfelPS->get3f( nIndex, TANGENT_V );
			T[2] = surfelPS->get3f( nIndex, NORMAL );
			const StaticVector<float32,SURFEL_NUM_BASIS_FUNCTIONS>& coeffs = surfelPS->get<float32,SURFEL_NUM_BASIS_FUNCTIONS>( nIndex, COEFFS );

			const card32 numSamples = 1000;
			for ( unsigned int s = 0; s < numSamples; s++ ){
				const float32 u = rand()%2000/1000.0f - 1;
				const float32 v = rand()%2000/1000.0f - 1;
				Vector3f sample_local = makeVector3f( u,v, LocalSurfaceFunction<SURFEL_NUM_BASIS_FUNCTIONS>::f( coeffs, u, v ) );
				Vector3f sample = T * sample_local * settings->topSurfelSpacing + np;   
				//sample = np;

				const float32 distanceFromPlane = sample * hsNormal - hsPoint * hsNormal;
				if ( distanceFromPlane > 0 ){
					const Vector3f hsPlaneProjection = sample - hsNormal * distanceFromPlane;
					const Vector3f pLocal = T.transpose() * ( hsPlaneProjection - np ) / settings->topSurfelSpacing;
					const float32 Z = hsConstraint->getPenaltyWeight() / float32( numSamples );
					card32 vectorIndex = nIndex * SURFEL_NUM_BASIS_FUNCTIONS;

					// Update function, gradient and hessian
					float32 u_jk = pLocal[0];
					float32 v_jk = pLocal[1];
					float32 n_jk = pLocal[2];
					for ( unsigned int bfIndex = 0; bfIndex < SURFEL_NUM_BASIS_FUNCTIONS; bfIndex++ ){
						for ( unsigned int j = 0; j < SURFEL_NUM_BASIS_FUNCTIONS; j++ ){
							float32 h = Z * LocalSurfaceFunction<SURFEL_NUM_BASIS_FUNCTIONS>::f_i( u_jk, v_jk, bfIndex ) * LocalSurfaceFunction<SURFEL_NUM_BASIS_FUNCTIONS>::f_i( u_jk, v_jk, j );
							H[vectorIndex+bfIndex].addEntry( vectorIndex+j, h );
						}
						gradient[vectorIndex+bfIndex] += Z * LocalSurfaceFunction<SURFEL_NUM_BASIS_FUNCTIONS>::f_i( u_jk, v_jk, bfIndex ) * ( LocalSurfaceFunction<SURFEL_NUM_BASIS_FUNCTIONS>::f( coeffs, u_jk, v_jk ) - n_jk );
					}
					f += Z * sqr( LocalSurfaceFunction<SURFEL_NUM_BASIS_FUNCTIONS>::f( coeffs, u_jk, v_jk ) - n_jk );
				}
			}
		}
	}
}

void SurfelOptimizer::addConsistencyConstraints()
{
	if ( settings->shapeSolveConsistencyMethod == CONSISTENCY_MODE_NONE ){
		return;
	}

	if ( context == NULL ) {
		if ( settings->surfaceRecDebugOutput ){
			debugOutput << "addConsistencyConstraint: error.\n"; debugOutput.flush();
		}
		return;
	}
	PointSet* surfelPS = context->getSurfels();
	if ( surfelPS == NULL ) {
		if ( settings->surfaceRecDebugOutput ){
			debugOutput << "addConsistencyConstraint: error.\n"; debugOutput.flush();
		}
		return;
	}

	if ( settings->shapeSolveConsistencyMethod == CONSISTENCY_MODE_NORMALS_PLANE ){
		if ( SURFEL_NUM_BASIS_FUNCTIONS != 3 ){
			if ( settings->surfaceRecDebugOutput ){
				debugOutput << "CONSISTENCY_MODE_NORMALS only defined for linear basis functions: SURFEL_NUM_BASIS_FUNCIONS = 3.\n"; debugOutput.flush();
			}
			return;
		}
	} else {
		if ( settings->surfaceRecDebugOutput ){
			debugOutput << "Please use normal/plane-based consistency only.\n"; debugOutput.flush();
		}
		return;
	}

	AAT COEFFS = surfelPS->getAAT( "coeffs" );
	AAT POSITION = surfelPS->getAAT( "position" );
	AAT NORMAL = surfelPS->getAAT( "normal" );
	AAT TANGENT_U = surfelPS->getAAT( "tangent_u" );
	AAT TANGENT_V = surfelPS->getAAT( "tangent_v" );
	InCorePCTopologyGraph* topologyGraph = context->getTopologyGraph();
	const float32 sqrSigma = sqr( settings->topSurfelSpacing );
	const card32 numConsistencySamples = 100;

	for ( unsigned int sIndex = 0; sIndex < surfelPS->getNumEntries(); sIndex++ ){
		const StaticVector<float32,SURFEL_NUM_BASIS_FUNCTIONS>& coeffs_i = surfelPS->get<float32,SURFEL_NUM_BASIS_FUNCTIONS>( sIndex, COEFFS );
		const Vector3f& s_i = surfelPS->get3f( sIndex, POSITION );
		Matrix3f T_i;
		T_i[0] = surfelPS->get3f( sIndex, TANGENT_U );
		T_i[1] = surfelPS->get3f( sIndex, TANGENT_V );
		T_i[2] = surfelPS->get3f( sIndex, NORMAL );
		Matrix3f invT_i = T_i.transpose();
		float32 f_i;
		StaticVector<float32,SURFEL_NUM_BASIS_FUNCTIONS> g_i; 
		StaticMatrix<float32,SURFEL_NUM_BASIS_FUNCTIONS, SURFEL_NUM_BASIS_FUNCTIONS> H_i; 
		f_i = 0;
		for ( unsigned int i = 0; i < SURFEL_NUM_BASIS_FUNCTIONS; i++ ){
			g_i[i] = 0;
			for ( unsigned int j = 0; j < SURFEL_NUM_BASIS_FUNCTIONS; j++ ){
				H_i[i][j] = 0;
			}
		}
		
		if ( settings->shapeSolveConsistencyMethod == CONSISTENCY_MODE_NORMALS_PLANE ){
			// Align adjacent normals
			Matrix3f Ta, Tb;
			const Vector3f sa = surfelPS->get3f( sIndex, POSITION );
			Ta[0] = surfelPS->get3f( sIndex, TANGENT_U );
			Ta[1] = surfelPS->get3f( sIndex, TANGENT_V );
			Ta[2] = surfelPS->get3f( sIndex, NORMAL );
			Ta = Ta.transpose();
			const StaticVector<float32,SURFEL_NUM_BASIS_FUNCTIONS>& coeffsa = surfelPS->get<float32,SURFEL_NUM_BASIS_FUNCTIONS>( sIndex, COEFFS );
			const float32 Ta_00 = Ta[0][0];
			const float32 Ta_01 = Ta[0][1];
			const float32 Ta_02 = Ta[0][2];
			const float32 Ta_10 = Ta[1][0];
			const float32 Ta_11 = Ta[1][1];
			const float32 Ta_12 = Ta[1][2];
			const float32 Ta_20 = Ta[2][0];
			const float32 Ta_21 = Ta[2][1];
			const float32 Ta_22 = Ta[2][2];
			const float32 ca_0 = coeffsa[0];
			const float32 ca_1 = coeffsa[1];
			const float32 ca_2 = coeffsa[2];
			const float32 sa_0 = sa[0];
			const float32 sa_1 = sa[1];
			const float32 sa_2 = sa[2];

			const float32 sampling = settings->topSurfelSpacing;
			const float32 sqr_sampling = sqr( sampling );
			float32 summedWeight = 0;
			card32 numNeighbors = topologyGraph->getNumAdjacentVertices( sIndex );

			const float32 weight_plane = settings->weightSmoothPlane;
			const float32 weight_normal = settings->weightSmoothNormal;

			for ( unsigned int n = 0; n < numNeighbors; n++ ){
				const card32 nIndex = topologyGraph->getAdjacentVertex( sIndex, n );
				const Vector3f& sb = surfelPS->get3f( nIndex, POSITION );
				Tb[0] = surfelPS->get3f( nIndex, TANGENT_U );
				Tb[1] = surfelPS->get3f( nIndex, TANGENT_V );
				Tb[2] = surfelPS->get3f( nIndex, NORMAL );
				Tb = Tb.transpose();
				const StaticVector<float32,SURFEL_NUM_BASIS_FUNCTIONS>& coeffsb = surfelPS->get<float32,SURFEL_NUM_BASIS_FUNCTIONS>( nIndex, COEFFS );
				const float32 Tb_00 = Tb[0][0];
				const float32 Tb_01 = Tb[0][1];
				const float32 Tb_02 = Tb[0][2];
				const float32 Tb_10 = Tb[1][0];
				const float32 Tb_11 = Tb[1][1];
				const float32 Tb_12 = Tb[1][2];
				const float32 Tb_20 = Tb[2][0];
				const float32 Tb_21 = Tb[2][1];
				const float32 Tb_22 = Tb[2][2];
				const float32 cb_0 = coeffsb[0];
				const float32 cb_1 = coeffsb[1];
				const float32 cb_2 = coeffsb[2];
				const float32 sb_0 = sb[0];
				const float32 sb_1 = sb[1];
				const float32 sb_2 = sb[2];

				const float32 sqrD = ( sa - sb ) * ( sa - sb );
				float32 weight_dist = exp( - sqrD / sqr_sampling );
				summedWeight += weight_dist;
				
				// Normal alignment
				f_i       += weight_normal*weight_dist * ( sqr(-Ta_00*ca_1-Ta_01*ca_2+Ta_02+Tb_00*cb_1+Tb_01*cb_2-Tb_02)+sqr(-Ta_10*ca_1-Ta_11*ca_2+Ta_12+Tb_10*cb_1+Tb_11*cb_2-Tb_12)+sqr(-Ta_20*ca_1-Ta_21*ca_2+Ta_22+Tb_20*cb_1+Tb_21*cb_2-Tb_22) );
				g_i[0]    += 0;
				g_i[1]    += weight_normal*weight_dist * ( -2*(-Ta_00*ca_1-Ta_01*ca_2+Ta_02+Tb_00*cb_1+Tb_01*cb_2-Tb_02)*Ta_00-2*(-Ta_10*ca_1-Ta_11*ca_2+Ta_12+Tb_10*cb_1+Tb_11*cb_2-Tb_12)*Ta_10-2*(-Ta_20*ca_1-Ta_21*ca_2+Ta_22+Tb_20*cb_1+Tb_21*cb_2-Tb_22)*Ta_20 );
				g_i[2]    += weight_normal*weight_dist * ( -2*(-Ta_00*ca_1-Ta_01*ca_2+Ta_02+Tb_00*cb_1+Tb_01*cb_2-Tb_02)*Ta_01-2*(-Ta_10*ca_1-Ta_11*ca_2+Ta_12+Tb_10*cb_1+Tb_11*cb_2-Tb_12)*Ta_11-2*(-Ta_20*ca_1-Ta_21*ca_2+Ta_22+Tb_20*cb_1+Tb_21*cb_2-Tb_22)*Ta_21 );
				H_i[0][0] += 0;
				H_i[0][1] += 0;
				H_i[0][2] += 0;
				H_i[1][0] += 0;
				H_i[1][1] += weight_normal*weight_dist * ( 2*sqr(Ta_00)+2*sqr(Ta_10)+2*sqr(Ta_20) );
				H_i[1][2] += weight_normal*weight_dist * ( 2*Ta_01*Ta_00+2*Ta_11*Ta_10+2*Ta_21*Ta_20 );
				H_i[2][0] += 0;
				H_i[2][1] += weight_normal*weight_dist * ( 2*Ta_01*Ta_00+2*Ta_11*Ta_10+2*Ta_21*Ta_20 );
				H_i[2][2] += weight_normal*weight_dist * ( 2*sqr(Ta_01)+2*sqr(Ta_11)+2*sqr(Ta_21) );

				// Distance from plane
				f_i       += weight_dist*weight_plane/sqr_sampling*sqr((-Ta_00*ca_1-Ta_01*ca_2+Ta_02)*(sampling*Ta_02*ca_0+sa_0)+(-Ta_10*ca_1-Ta_11*ca_2+Ta_12)*(sampling*Ta_12*ca_0+sa_1)+(-Ta_20*ca_1-Ta_21*ca_2+Ta_22)*(sampling*Ta_22*ca_0+sa_2)-(-Ta_00*ca_1-Ta_01*ca_2+Ta_02)*(sampling*Tb_02*cb_0+sb_0)-(-Ta_10*ca_1-Ta_11*ca_2+Ta_12)*(sampling*Tb_12*cb_0+sb_1)-(-Ta_20*ca_1-Ta_21*ca_2+Ta_22)*(sampling*Tb_22*cb_0+sb_2));
				g_i[0]    += 2*weight_dist*weight_plane/sqr_sampling*((-Ta_00*ca_1-Ta_01*ca_2+Ta_02)*(sampling*Ta_02*ca_0+sa_0)+(-Ta_10*ca_1-Ta_11*ca_2+Ta_12)*(sampling*Ta_12*ca_0+sa_1)+(-Ta_20*ca_1-Ta_21*ca_2+Ta_22)*(sampling*Ta_22*ca_0+sa_2)-(-Ta_00*ca_1-Ta_01*ca_2+Ta_02)*(sampling*Tb_02*cb_0+sb_0)-(-Ta_10*ca_1-Ta_11*ca_2+Ta_12)*(sampling*Tb_12*cb_0+sb_1)-(-Ta_20*ca_1-Ta_21*ca_2+Ta_22)*(sampling*Tb_22*cb_0+sb_2))*((-Ta_00*ca_1-Ta_01*ca_2+Ta_02)*sampling*Ta_02+(-Ta_10*ca_1-Ta_11*ca_2+Ta_12)*sampling*Ta_12+(-Ta_20*ca_1-Ta_21*ca_2+Ta_22)*sampling*Ta_22);
				g_i[1]    += 2*weight_dist*weight_plane/sqr_sampling*((-Ta_00*ca_1-Ta_01*ca_2+Ta_02)*(sampling*Ta_02*ca_0+sa_0)+(-Ta_10*ca_1-Ta_11*ca_2+Ta_12)*(sampling*Ta_12*ca_0+sa_1)+(-Ta_20*ca_1-Ta_21*ca_2+Ta_22)*(sampling*Ta_22*ca_0+sa_2)-(-Ta_00*ca_1-Ta_01*ca_2+Ta_02)*(sampling*Tb_02*cb_0+sb_0)-(-Ta_10*ca_1-Ta_11*ca_2+Ta_12)*(sampling*Tb_12*cb_0+sb_1)-(-Ta_20*ca_1-Ta_21*ca_2+Ta_22)*(sampling*Tb_22*cb_0+sb_2))*(-Ta_00*(sampling*Ta_02*ca_0+sa_0)-Ta_10*(sampling*Ta_12*ca_0+sa_1)-Ta_20*(sampling*Ta_22*ca_0+sa_2)+Ta_00*(sampling*Tb_02*cb_0+sb_0)+Ta_10*(sampling*Tb_12*cb_0+sb_1)+Ta_20*(sampling*Tb_22*cb_0+sb_2));
				g_i[2]    += 2*weight_dist*weight_plane/sqr_sampling*((-Ta_00*ca_1-Ta_01*ca_2+Ta_02)*(sampling*Ta_02*ca_0+sa_0)+(-Ta_10*ca_1-Ta_11*ca_2+Ta_12)*(sampling*Ta_12*ca_0+sa_1)+(-Ta_20*ca_1-Ta_21*ca_2+Ta_22)*(sampling*Ta_22*ca_0+sa_2)-(-Ta_00*ca_1-Ta_01*ca_2+Ta_02)*(sampling*Tb_02*cb_0+sb_0)-(-Ta_10*ca_1-Ta_11*ca_2+Ta_12)*(sampling*Tb_12*cb_0+sb_1)-(-Ta_20*ca_1-Ta_21*ca_2+Ta_22)*(sampling*Tb_22*cb_0+sb_2))*(-Ta_01*(sampling*Ta_02*ca_0+sa_0)-Ta_11*(sampling*Ta_12*ca_0+sa_1)-Ta_21*(sampling*Ta_22*ca_0+sa_2)+Ta_01*(sampling*Tb_02*cb_0+sb_0)+Ta_11*(sampling*Tb_12*cb_0+sb_1)+Ta_21*(sampling*Tb_22*cb_0+sb_2));
				H_i[0][0] += 2*weight_dist*weight_plane/sqr_sampling*sqr((-Ta_00*ca_1-Ta_01*ca_2+Ta_02)*sampling*Ta_02+(-Ta_10*ca_1-Ta_11*ca_2+Ta_12)*sampling*Ta_12+(-Ta_20*ca_1-Ta_21*ca_2+Ta_22)*sampling*Ta_22);
				H_i[0][1] += 2*weight_dist*weight_plane/sqr_sampling*(-Ta_00*(sampling*Ta_02*ca_0+sa_0)-Ta_10*(sampling*Ta_12*ca_0+sa_1)-Ta_20*(sampling*Ta_22*ca_0+sa_2)+Ta_00*(sampling*Tb_02*cb_0+sb_0)+Ta_10*(sampling*Tb_12*cb_0+sb_1)+Ta_20*(sampling*Tb_22*cb_0+sb_2))*((-Ta_00*ca_1-Ta_01*ca_2+Ta_02)*sampling*Ta_02+(-Ta_10*ca_1-Ta_11*ca_2+Ta_12)*sampling*Ta_12+(-Ta_20*ca_1-Ta_21*ca_2+Ta_22)*sampling*Ta_22)+2*weight_dist*weight_plane/sqr_sampling*((-Ta_00*ca_1-Ta_01*ca_2+Ta_02)*(sampling*Ta_02*ca_0+sa_0)+(-Ta_10*ca_1-Ta_11*ca_2+Ta_12)*(sampling*Ta_12*ca_0+sa_1)+(-Ta_20*ca_1-Ta_21*ca_2+Ta_22)*(sampling*Ta_22*ca_0+sa_2)-(-Ta_00*ca_1-Ta_01*ca_2+Ta_02)*(sampling*Tb_02*cb_0+sb_0)-(-Ta_10*ca_1-Ta_11*ca_2+Ta_12)*(sampling*Tb_12*cb_0+sb_1)-(-Ta_20*ca_1-Ta_21*ca_2+Ta_22)*(sampling*Tb_22*cb_0+sb_2))*(-Ta_00*sampling*Ta_02-Ta_10*sampling*Ta_12-Ta_20*sampling*Ta_22);
				H_i[0][2] += 2*weight_dist*weight_plane/sqr_sampling*(-Ta_01*(sampling*Ta_02*ca_0+sa_0)-Ta_11*(sampling*Ta_12*ca_0+sa_1)-Ta_21*(sampling*Ta_22*ca_0+sa_2)+Ta_01*(sampling*Tb_02*cb_0+sb_0)+Ta_11*(sampling*Tb_12*cb_0+sb_1)+Ta_21*(sampling*Tb_22*cb_0+sb_2))*((-Ta_00*ca_1-Ta_01*ca_2+Ta_02)*sampling*Ta_02+(-Ta_10*ca_1-Ta_11*ca_2+Ta_12)*sampling*Ta_12+(-Ta_20*ca_1-Ta_21*ca_2+Ta_22)*sampling*Ta_22)+2*weight_dist*weight_plane/sqr_sampling*((-Ta_00*ca_1-Ta_01*ca_2+Ta_02)*(sampling*Ta_02*ca_0+sa_0)+(-Ta_10*ca_1-Ta_11*ca_2+Ta_12)*(sampling*Ta_12*ca_0+sa_1)+(-Ta_20*ca_1-Ta_21*ca_2+Ta_22)*(sampling*Ta_22*ca_0+sa_2)-(-Ta_00*ca_1-Ta_01*ca_2+Ta_02)*(sampling*Tb_02*cb_0+sb_0)-(-Ta_10*ca_1-Ta_11*ca_2+Ta_12)*(sampling*Tb_12*cb_0+sb_1)-(-Ta_20*ca_1-Ta_21*ca_2+Ta_22)*(sampling*Tb_22*cb_0+sb_2))*(-Ta_01*sampling*Ta_02-Ta_11*sampling*Ta_12-Ta_21*sampling*Ta_22);
				H_i[1][0] += 2*weight_dist*weight_plane/sqr_sampling*(-Ta_00*(sampling*Ta_02*ca_0+sa_0)-Ta_10*(sampling*Ta_12*ca_0+sa_1)-Ta_20*(sampling*Ta_22*ca_0+sa_2)+Ta_00*(sampling*Tb_02*cb_0+sb_0)+Ta_10*(sampling*Tb_12*cb_0+sb_1)+Ta_20*(sampling*Tb_22*cb_0+sb_2))*((-Ta_00*ca_1-Ta_01*ca_2+Ta_02)*sampling*Ta_02+(-Ta_10*ca_1-Ta_11*ca_2+Ta_12)*sampling*Ta_12+(-Ta_20*ca_1-Ta_21*ca_2+Ta_22)*sampling*Ta_22)+2*weight_dist*weight_plane/sqr_sampling*((-Ta_00*ca_1-Ta_01*ca_2+Ta_02)*(sampling*Ta_02*ca_0+sa_0)+(-Ta_10*ca_1-Ta_11*ca_2+Ta_12)*(sampling*Ta_12*ca_0+sa_1)+(-Ta_20*ca_1-Ta_21*ca_2+Ta_22)*(sampling*Ta_22*ca_0+sa_2)-(-Ta_00*ca_1-Ta_01*ca_2+Ta_02)*(sampling*Tb_02*cb_0+sb_0)-(-Ta_10*ca_1-Ta_11*ca_2+Ta_12)*(sampling*Tb_12*cb_0+sb_1)-(-Ta_20*ca_1-Ta_21*ca_2+Ta_22)*(sampling*Tb_22*cb_0+sb_2))*(-Ta_00*sampling*Ta_02-Ta_10*sampling*Ta_12-Ta_20*sampling*Ta_22);
				H_i[1][1] += 2*weight_dist*weight_plane/sqr_sampling*sqr(-Ta_00*(sampling*Ta_02*ca_0+sa_0)-Ta_10*(sampling*Ta_12*ca_0+sa_1)-Ta_20*(sampling*Ta_22*ca_0+sa_2)+Ta_00*(sampling*Tb_02*cb_0+sb_0)+Ta_10*(sampling*Tb_12*cb_0+sb_1)+Ta_20*(sampling*Tb_22*cb_0+sb_2));
				H_i[1][2] += 2*weight_dist*weight_plane/sqr_sampling*(-Ta_01*(sampling*Ta_02*ca_0+sa_0)-Ta_11*(sampling*Ta_12*ca_0+sa_1)-Ta_21*(sampling*Ta_22*ca_0+sa_2)+Ta_01*(sampling*Tb_02*cb_0+sb_0)+Ta_11*(sampling*Tb_12*cb_0+sb_1)+Ta_21*(sampling*Tb_22*cb_0+sb_2))*(-Ta_00*(sampling*Ta_02*ca_0+sa_0)-Ta_10*(sampling*Ta_12*ca_0+sa_1)-Ta_20*(sampling*Ta_22*ca_0+sa_2)+Ta_00*(sampling*Tb_02*cb_0+sb_0)+Ta_10*(sampling*Tb_12*cb_0+sb_1)+Ta_20*(sampling*Tb_22*cb_0+sb_2));
				H_i[2][0] += 2*weight_dist*weight_plane/sqr_sampling*(-Ta_01*(sampling*Ta_02*ca_0+sa_0)-Ta_11*(sampling*Ta_12*ca_0+sa_1)-Ta_21*(sampling*Ta_22*ca_0+sa_2)+Ta_01*(sampling*Tb_02*cb_0+sb_0)+Ta_11*(sampling*Tb_12*cb_0+sb_1)+Ta_21*(sampling*Tb_22*cb_0+sb_2))*((-Ta_00*ca_1-Ta_01*ca_2+Ta_02)*sampling*Ta_02+(-Ta_10*ca_1-Ta_11*ca_2+Ta_12)*sampling*Ta_12+(-Ta_20*ca_1-Ta_21*ca_2+Ta_22)*sampling*Ta_22)+2*weight_dist*weight_plane/sqr_sampling*((-Ta_00*ca_1-Ta_01*ca_2+Ta_02)*(sampling*Ta_02*ca_0+sa_0)+(-Ta_10*ca_1-Ta_11*ca_2+Ta_12)*(sampling*Ta_12*ca_0+sa_1)+(-Ta_20*ca_1-Ta_21*ca_2+Ta_22)*(sampling*Ta_22*ca_0+sa_2)-(-Ta_00*ca_1-Ta_01*ca_2+Ta_02)*(sampling*Tb_02*cb_0+sb_0)-(-Ta_10*ca_1-Ta_11*ca_2+Ta_12)*(sampling*Tb_12*cb_0+sb_1)-(-Ta_20*ca_1-Ta_21*ca_2+Ta_22)*(sampling*Tb_22*cb_0+sb_2))*(-Ta_01*sampling*Ta_02-Ta_11*sampling*Ta_12-Ta_21*sampling*Ta_22);
				H_i[2][1] += 2*weight_dist*weight_plane/sqr_sampling*(-Ta_01*(sampling*Ta_02*ca_0+sa_0)-Ta_11*(sampling*Ta_12*ca_0+sa_1)-Ta_21*(sampling*Ta_22*ca_0+sa_2)+Ta_01*(sampling*Tb_02*cb_0+sb_0)+Ta_11*(sampling*Tb_12*cb_0+sb_1)+Ta_21*(sampling*Tb_22*cb_0+sb_2))*(-Ta_00*(sampling*Ta_02*ca_0+sa_0)-Ta_10*(sampling*Ta_12*ca_0+sa_1)-Ta_20*(sampling*Ta_22*ca_0+sa_2)+Ta_00*(sampling*Tb_02*cb_0+sb_0)+Ta_10*(sampling*Tb_12*cb_0+sb_1)+Ta_20*(sampling*Tb_22*cb_0+sb_2));
				H_i[2][2] += 2*weight_dist*weight_plane/sqr_sampling*sqr(-Ta_01*(sampling*Ta_02*ca_0+sa_0)-Ta_11*(sampling*Ta_12*ca_0+sa_1)-Ta_21*(sampling*Ta_22*ca_0+sa_2)+Ta_01*(sampling*Tb_02*cb_0+sb_0)+Ta_11*(sampling*Tb_12*cb_0+sb_1)+Ta_21*(sampling*Tb_22*cb_0+sb_2));
			}

			if ( numNeighbors > 0 ){
				const float32 Z = 1.0f / summedWeight;
				f_i *= Z;
				g_i *= Z;
				H_i *= Z;
				f += f_i;
				const long vectorIndex = sIndex * SURFEL_NUM_BASIS_FUNCTIONS;
				for ( unsigned int i = 0; i < SURFEL_NUM_BASIS_FUNCTIONS; i++ ){
					gradient[vectorIndex+i] += g_i[i];
					for ( unsigned int j = 0; j < SURFEL_NUM_BASIS_FUNCTIONS; j++ ){
						H[vectorIndex+i].addEntry( vectorIndex+j, H_i[i][j] );
					}
				}
			}

			// Laplacian

		}
	}
}

void SurfelOptimizer::solve()
{

	if ( settings->surfaceRecDebugOutput ){
		float32 timeRequired = float32( ( clock() - startBuildMatrixTime ) / CLK_TCK );
		debugOutput << "  Time required to handle position constraints: " << timeRequired << "\n"; debugOutput.flush();
	}
	startBuildMatrixTime = clock();

	addConsistencyConstraints();

	float32 timeRequired = float32( ( clock() - startBuildMatrixTime ) / CLK_TCK );	
	if ( settings->surfaceRecDebugOutput ){
		debugOutput << "  Time required to handle consistency constraints: " << timeRequired << "\n"; debugOutput.flush();
	}

	const float startSolveTime = clock();

	if ( settings == NULL ) {
		if ( settings->surfaceRecDebugOutput ){
			debugOutput << "solve: error.\n"; debugOutput.flush();
		}
      delete surfelOctree; surfelOctree = NULL;
		return;
	}
	if ( context == NULL ) {
		if ( settings->surfaceRecDebugOutput ){
			debugOutput << "solve: error.\n"; debugOutput.flush();
		}
      delete surfelOctree; surfelOctree = NULL;
		return;
	}
	PointSet* surfelPS = context->getSurfels();
	AAT COEFFS = surfelPS->getAAT( "coeffs" );

	// Init variable std::vector
	for ( unsigned int pIndex = 0; pIndex < surfelPS->getNumEntries(); pIndex++ ){
		const StaticVector<float32,SURFEL_NUM_BASIS_FUNCTIONS> coeffs = surfelPS->get<float32,SURFEL_NUM_BASIS_FUNCTIONS>( pIndex, COEFFS );
		const card32 vectorIndex = pIndex * SURFEL_NUM_BASIS_FUNCTIONS;
		for ( unsigned int n = 0; n < SURFEL_NUM_BASIS_FUNCTIONS; n++ ){
			variables[vectorIndex+n] = coeffs[n];
		}
	}

	// Solve
	card32 numIterations = settings->shapeSolveMaxNumIterations;
	float32 conv = settings->shapeSolveConvergence;
	if ( settings->shapeSolveSolver == SHAPE_SOLVER_STEEPEST_DESCENT ){
		variables = variables - gradient * settings->shapeSolveStepSize;
	} else if ( settings->shapeSolveSolver == SHAPE_SOLVER_NEWTON ){
		b = H * variables - gradient;
		conjugateGradientsSolve( H, variables, b, numIterations, conv );
	} else if ( settings->shapeSolveSolver == SHAPE_SOLVER_DIRECT ){
		b = H * variables - gradient;
		StaticMatrix<float32,SURFEL_NUM_BASIS_FUNCTIONS,SURFEL_NUM_BASIS_FUNCTIONS> A;
		StaticVector<float32,SURFEL_NUM_BASIS_FUNCTIONS> d, c;
		for ( unsigned int sIndex = 0; sIndex < variables.size()/SURFEL_NUM_BASIS_FUNCTIONS; sIndex++ ){
			const long surfelIndex = sIndex * SURFEL_NUM_BASIS_FUNCTIONS;
			for ( unsigned int i = 0; i < SURFEL_NUM_BASIS_FUNCTIONS; i++ ){
				d[i] = b[surfelIndex+i];
				for ( unsigned int j = 0; j < SURFEL_NUM_BASIS_FUNCTIONS; j++ ){
					A[i][j] = H[surfelIndex+i][surfelIndex+j];
				}
			}
			c = invertMatrix( A ) * d;
			for ( unsigned int i = 0; i < SURFEL_NUM_BASIS_FUNCTIONS; i++ ){
				variables[surfelIndex+i] = c[i];
			}
		}
	}
	
	// Write back solution
	for ( unsigned int pIndex = 0; pIndex < surfelPS->getNumEntries(); pIndex++ ){
		const card32 vectorIndex = pIndex * SURFEL_NUM_BASIS_FUNCTIONS;
		StaticVector<float32,SURFEL_NUM_BASIS_FUNCTIONS> coeffs;
		for ( unsigned int n = 0; n < SURFEL_NUM_BASIS_FUNCTIONS; n++ ){
			coeffs[n] = variables[vectorIndex+n];
		}
		surfelPS->set<float32,SURFEL_NUM_BASIS_FUNCTIONS>( pIndex, COEFFS, coeffs );
	}

	if ( settings->surfaceRecDebugOutput ){
		debugOutput << "  Time required to solve system: " << float32( ( clock() - startSolveTime ) / CLK_TCK ) << "\n"; debugOutput.flush();
		debugOutput << "   - #iterations: " << numIterations << "\n"; debugOutput.flush();
		debugOutput << "   - #convergence criterion: " << conv << "\n"; debugOutput.flush();
	}

	settings = NULL;
	context = NULL;
   delete surfelOctree; surfelOctree = NULL;
}

AROptimizationContext* SurfelOptimizer::createOptContext()
{
	SurfelOptimizationContext* context = new SurfelOptimizationContext();
	return context;
}

void SurfelOptimizer::setup( AROptimizationContext* context, const ARSettings *settings )
{
	this->settings = settings;
	this->context = dynamic_cast<SurfelOptimizationContext*>( context );
}

ShapeEvaluator *SurfelOptimizer::createEvaluator() 
{
   return new SurfelOptEvaluator();
}


//// DEPRECARED CODE
/*
if ( settings->shapeSolveConsistencyMethod == CONSISTENCY_MODE_SINGLE_NEIGHBOR ){
/// Idea 1: consistency with each neighbor j, halfway between s_i and s_j
for ( unsigned int n = 0; n < topologyGraph->getNumAdjacentVertices( sIndex ); n++ ){
const card32 nIndex = topologyGraph->getAdjacentVertex( sIndex, n );
const Vector3f& s_j = surfelPS->get3f( nIndex, POSITION );
Matrix3f T_j;
T_j[0] = surfelPS->get3f( nIndex, TANGENT_U );
T_j[1] = surfelPS->get3f( nIndex, TANGENT_V );
T_j[2] = surfelPS->get3f( nIndex, NORMAL );
Matrix3f invT_j = T_j.transpose();
card32 vectorIndex = nIndex * SURFEL_NUM_BASIS_FUNCTIONS;
const Vector3f q = ( s_i - s_j ) / 2.0f;

// Project q on surface of surfel i
Vector3f q_i = invT_i * ( q - s_i ) / settings->topSurfelSpacing;
q_i = makeVector3f( q_i[0], q_i[1], LocalSurfaceFunction<SURFEL_NUM_BASIS_FUNCTIONS>::f( coeffs_i, q_i[0], q_i[1] ) );
q_i = T_i * q_i * settings->topSurfelSpacing + s_i;

// Formulate projection as constraint on surfel j
const StaticVector<float32,SURFEL_NUM_BASIS_FUNCTIONS>& coeffs_j = surfelPS->get<float32,SURFEL_NUM_BASIS_FUNCTIONS>( nIndex, COEFFS );
Vector3f q_j = invT_j * ( q_i - s_j ) / settings->topSurfelSpacing;

const Vector3f& l = ( q_i - q_j ) / 2.0f;

for ( unsigned int i = 0; i < SURFEL_NUM_BASIS_FUNCTIONS; i++ ){
for ( unsigned int j = 0; j < SURFEL_NUM_BASIS_FUNCTIONS; j++ ){
const float32 h = LocalSurfaceFunction<SURFEL_NUM_BASIS_FUNCTIONS>::f_i( l[0], l[1], i ) * LocalSurfaceFunction<SURFEL_NUM_BASIS_FUNCTIONS>::f_i( l[0], l[1], j );
H_i[i][j] += h;
}
const float32 h = LocalSurfaceFunction<SURFEL_NUM_BASIS_FUNCTIONS>::f_i( l[0], l[1], i ) * ( LocalSurfaceFunction<SURFEL_NUM_BASIS_FUNCTIONS>::f( coeffs_j, l[0], l[1] ) - l[2] );
g_i[i] += h;
}
}

const float32 Z = 1.0f / float32( surfelPS->getNumEntries() * topologyGraph->getNumAdjacentVertices( sIndex ) );
f_i *= Z;
g_i *= Z;
H_i *= Z;
f += f_i;

const long vectorIndex = sIndex * SURFEL_NUM_BASIS_FUNCTIONS;
for ( unsigned int i = 0; i < SURFEL_NUM_BASIS_FUNCTIONS; i++ ){
gradient[vectorIndex+i] += g_i[i];
for ( unsigned int j = 0; j < SURFEL_NUM_BASIS_FUNCTIONS; j++ ){
H[vectorIndex+i].addEntry( vectorIndex+j, H_i[i][j] );
}
}
} else if ( settings->shapeSolveConsistencyMethod == CONSISTENCY_MODE_INTEGRATION ){
// For each sample point around surfel s_i
float32 summedSigma = 0;
for ( unsigned int k = 0; k < numConsistencySamples; k++ ){
const float32 u_k = rand()%2000/1000.0f - 1.0f;
const float32 v_k = rand()%2000/1000.0f - 1.0f;
const float32 n_k = LocalSurfaceFunction<SURFEL_NUM_BASIS_FUNCTIONS>::f( coeffs_i, u_k, v_k );
const Vector3f p = makeVector3f( u_k, v_k, n_k );
const Vector3f p_global = ( T_i * p ) * settings->topSurfelSpacing + s_i; 
const float32 omega_i = exp( - ( ( p_global - s_i ) * ( p_global - s_i ) ) / sqrSigma );

// For each candidate (chosen from the neighbors)
for ( unsigned int n = 0; n < topologyGraph->getNumAdjacentVertices( sIndex ); n++ ){
const card32 nIndex = topologyGraph->getAdjacentVertex( sIndex, n );
const Vector3f& s_j = surfelPS->get3f( nIndex, POSITION );

if ( ( s_j - s_i ) * ( p_global - s_i ) > 0 ){
const Vector3f& n_j = surfelPS->get3f( nIndex, NORMAL );
const Vector3f& s_ji = invT_i * ( s_j - s_i ) / settings->topSurfelSpacing;
const Vector3f& n_ji = invT_i * n_j;
const float32 s_ji0 = s_ji[0]; 
const float32 s_ji1 = s_ji[1];
const float32 s_ji2 = s_ji[2];
const float32 n_ji0 = n_ji[0]; 
const float32 n_ji1 = n_ji[1];
const float32 n_ji2 = n_ji[2];
const float32 omega_j = exp( - ( ( p_global - s_j ) * ( p_global - s_j ) ) / sqrSigma );
const float32 c0 = coeffs_i[0];
const float32 c1 = coeffs_i[1];
const float32 c2 = coeffs_i[2];

// Function
summedSigma += omega_i*omega_j;
f_i += omega_i*omega_j*sqr(u_k*n_ji0+v_k*n_ji1+(c0+c1*u_k+c2*v_k)*n_ji2-s_ji0*n_ji0-s_ji1*n_ji1-s_ji2*n_ji2);

// Gradient
g_i[0] += 2*omega_i*omega_j*(u_k*n_ji0+v_k*n_ji1+(c0+c1*u_k+c2*v_k)*n_ji2-s_ji0*n_ji0-s_ji1*n_ji1-s_ji2*n_ji2)*n_ji2;
g_i[1] += 2*omega_i*omega_j*(u_k*n_ji0+v_k*n_ji1+(c0+c1*u_k+c2*v_k)*n_ji2-s_ji0*n_ji0-s_ji1*n_ji1-s_ji2*n_ji2)*u_k*n_ji2;
g_i[2] += 2*omega_i*omega_j*(u_k*n_ji0+v_k*n_ji1+(c0+c1*u_k+c2*v_k)*n_ji2-s_ji0*n_ji0-s_ji1*n_ji1-s_ji2*n_ji2)*v_k*n_ji2;

// Hessian
H_i[0][0] += 2*omega_i*omega_j*sqr(n_ji2);
H_i[0][1] += 2*omega_i*omega_j*u_k*sqr(n_ji2);
H_i[0][2] += 2*omega_i*omega_j*v_k*sqr(n_ji2);
H_i[1][0] += 2*omega_i*omega_j*u_k*sqr(n_ji2);
H_i[1][1] += 2*omega_i*omega_j*sqr(u_k)*sqr(n_ji2);
H_i[1][2] += 2*omega_i*omega_j*v_k*sqr(n_ji2)*u_k;
H_i[2][0] += 2*omega_i*omega_j*v_k*sqr(n_ji2);
H_i[2][1] += 2*omega_i*omega_j*v_k*sqr(n_ji2)*u_k;
H_i[2][2] += 2*omega_i*omega_j*sqr(v_k)*sqr(n_ji2);
}
}
}
const float32 Z = 1.0f / float32( surfelPS->getNumEntries() * summedSigma );
f_i *= Z;
g_i *= Z;
H_i *= Z;
f += f_i;
const long vectorIndex = sIndex * SURFEL_NUM_BASIS_FUNCTIONS;
for ( unsigned int i = 0; i < SURFEL_NUM_BASIS_FUNCTIONS; i++ ){
gradient[vectorIndex+i] += g_i[i];
for ( unsigned int j = 0; j < SURFEL_NUM_BASIS_FUNCTIONS; j++ ){
H[vectorIndex+i].addEntry( vectorIndex+j, H_i[i][j] );
}
}
} else 
*/

/*
/*
// Find all influenced surfels
float32 sqrSurfelSpacing = settings->topSurfelSpacing;
float32 influenceRadius = settings->topSurfelSpacing * settings->recContraintsInfluence;
float32 sqrInfluenceRadius = sqr( influenceRadius );
TopNeighborhoodSphereSearch topoSphereSearch;
topoSphereSearch.setup( 
context->getTopologyGraph(), 
surfelPS, 
anchorIndex, 
anchorPos, 
influenceRadius, 
POSITION );

std::vector<card32> influencedSurfels; influencedSurfels.clear();
float32 weight_sum = 0;
while( !topoSphereSearch.atEnd() ){

const card32 nIndex = topoSphereSearch.getCurrentIndex();
const Vector3f& np = surfelPS->get3f( nIndex, POSITION );
const Vector3f nNormal = surfelPS->get3f( nIndex, NORMAL );
const float32 sqrD = ( np - anchorPos ) * ( np - anchorPos );
if ( sqrD < sqrInfluenceRadius ){

// Weight implied by distance
const float32 weight_distance = exp( - sqrD / sqrInfluenceRadius );

// Weight implied by quadric
const float32 weight_quadric = nNormal * ( posConstraint->covariance * nNormal );

// Combined weight for normalization
influencedSurfels.push_back( nIndex );
weight_sum += weight_distance * weight_quadric;

//if( influencedSurfels.size() > 3 ) break;	/// TODO: remove me; pj
break;
}
topoSphereSearch.next();
}
*/