#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "DebugAnimationRenderer.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "StandardSceneGraphIterator.h"
#include "SGAnimationNode.h"
#include "SGObjectNode.h"
#include "MemoryManager.h"
#include "GLShaderMaterial.h"
#include "DebugAnimationTools.h"
#include "FeatureCorrespondence.h"
#include "GeometricFeature.h"
#include "RenderingTool.h"
#include "PointCloudTools.h"
//---------------------------------------------------------------------------
using namespace X4;

IMPLEMENT_X4_CLASS( DebugAnimationRenderer ,0)
{
	BEGIN_CLASS_INIT( DebugAnimationRenderer );

	ADD_CARD32_PROP( m_NumRenderedFrames,0 );
	ADD_FLOAT32_PROP( m_RenderScale,0 );
	ADD_BOOLEAN_PROP(m_DrawFeatureCorrespondences ,0 );
	ADD_BOOLEAN_PROP(m_DrawFeaturePoints,0);
}

DebugAnimationRenderer::DebugAnimationRenderer(void)
{
	m_NumRenderedFrames = 2;
	m_MemoryManager = NULL;
	m_DrawFeatureCorrespondences = false;
	m_RenderScale = 1.0f;
	m_DrawFeaturePoints = false;
	m_RenderingTool = NULL;
}

DebugAnimationRenderer::~DebugAnimationRenderer(void)
{
}

void DebugAnimationRenderer::assign(const Object* obj, X4CopyContext *context)
{
	const DebugAnimationRenderer * o = dynamic_cast<const DebugAnimationRenderer*>(obj);
	x4Assert( o != NULL );

	copyObjectProperties(obj,this);
}

void X4::DebugAnimationRenderer::render( card32 viewportWidth, card32 viewportHeight )
{
	StandardSceneGraphIterator sIt;
	sIt.clearAndSetup( getScene()->getRootNode(), getScene()->getRootState(), SGDynamicState::tmRendering);
	int currentFrame = getScene()->getRootState()->dynamicState->time;
	int startFrame = currentFrame - (int)(m_NumRenderedFrames-1)/2;
	int lastFrame = currentFrame + (int)m_NumRenderedFrames/2;
	m_ScreenWidth = viewportWidth;
	m_ScreenHeight = viewportHeight;
	while( !sIt.atEnd())
	{
		SGAnimationNode * an = dynamic_cast<SGAnimationNode*>(sIt.getNode());
		if( an && an->getNumChildNodes(NULL) > 0)
		{
			SGObjectNode * firstNode = dynamic_cast<SGObjectNode*>(an->getChildNode(NULL,0));
			BoundingBox3f object_bb = getPCBBox( (PointCloud*)firstNode->getSceneObject() );
			Vector3f translation = makeVector3f( object_bb.getSideLength(0),0,0) * 1.1f;

			for( int i=startFrame;i<=lastFrame;i++ )						
			{
				if( i >= 0 && i<(int)an->getNumChildNodes(NULL) )
				{
					SGObjectNode * objNode = dynamic_cast<SGObjectNode*>(an->getChildNode(NULL,i));
					if( objNode )
					{
						SceneObject * so = objNode->getSceneObject();

						renderSceneObject(objNode, so, makeTranslation4f(translation* (i-currentFrame))  );
					}
				}
			}
			for( int i=startFrame;i<lastFrame;i++ )						
			{
				if( i >= 0 && i+1<(int)an->getNumChildNodes(NULL) )
				{
					SGObjectNode * objNode1 = dynamic_cast<SGObjectNode*>(an->getChildNode(NULL,i));
					SGObjectNode * objNode2 = dynamic_cast<SGObjectNode*>(an->getChildNode(NULL,i+1));
					if( objNode1 && objNode2 )
					{
						SceneObject * so1 = objNode1->getSceneObject();
						SceneObject * so2 = objNode2->getSceneObject();

						renderInBetweenStuff(so1, makeTranslation4f(translation* (i-currentFrame)),
							so2, makeTranslation4f(translation* (i+1-currentFrame)));
					}
				}
			}
		}
		sIt.next();
	}
}

void X4::DebugAnimationRenderer::renderSceneObject( SGObjectNode *node, SceneObject * so, Matrix4f frame)
{
	glPushAttrib(GL_ALL_ATTRIB_BITS);
	glMatrixMode( GL_MODELVIEW );
	glPushMatrix();
	glMultMatrixf( frame.data());
	glEnable(GL_DEPTH_TEST);
	glDepthMask(true);
	SceneGraphState * state = getScene()->getRootState();
	card32 matInd = so->getMaterialIndex();

   HierarchicalIterator *hi = dynamic_cast<HierarchicalIterator*>(node);

	if (getScene()->getNumGLMaterials() > matInd) {
		GLMaterial *material = getScene()->getGLMaterial(matInd);

		GLShaderMaterial *shMat = dynamic_cast<GLShaderMaterial*>(material);
		if (shMat != NULL) {
			float32 minPointSpacing;
			if( hi )
			{
				hi->getNodeInfo(HierarchicalIterator::niPointSpacing, NULL, &minPointSpacing);
				shMat->setParameter1f(getGLContext(), "minPointSpacing", minPointSpacing);
			}
			shMat->setParameter1f(getGLContext(), "vertViewAngle", state->staticState->viewFrustum.getVerticalFieldOfView());
			shMat->setParameter1f(getGLContext(), "tanOfHalfVerticalViewAngle", tan(state->staticState->viewFrustum.getVerticalFieldOfView() / 180.0f * M_PI / 2.0f));
			shMat->setParameter1f(getGLContext(), "vertResolution", m_ScreenHeight);
		}

		if (material) 
		{
			SceneObjectIterator *it = so->createIterator(SOIT(SOIT::CAP_HIERARCHICAL | SOIT::CAP_ATTACH_SECONDARY));
			material->draw(getScene(), state, so, it, getGLContext(), m_MemoryManager );
			delete it;

		}
	}
	glPopMatrix();
	glPopAttrib();

	// other rendering stuff:
	if( m_DrawFeaturePoints )
	{
		FeatureSet * set = dynamic_cast<FeatureSet*>(so->getAttachments()->getData( FeatureSet::getDefaultName()));
		if( set )
			drawFeatureSet(set,frame,m_RenderScale,m_RenderingTool);
	}

}

void X4::DebugAnimationRenderer::disconnect( GLContext* ctxt )
{
	delete m_MemoryManager;
	m_MemoryManager = NULL;
	delete m_RenderingTool;
	m_RenderingTool = NULL;
}

void X4::DebugAnimationRenderer::connect( Scene *scene, GLContext* ctxt )
{
	m_MemoryManager = new MemoryManager( 256 * 1024 * 1024 ); // max. 256 MB for vertex buffers
	m_RenderingTool = new RenderingTool;
	m_RenderingTool->init();
	Renderer::connect(scene,ctxt);
}

void X4::DebugAnimationRenderer::renderInBetweenStuff(SceneObject * so1, Matrix4f frame1, SceneObject * so2, Matrix4f frame2)
{
	PointCloud * pc1 = dynamic_cast<PointCloud*>(so1);
	PointCloud * pc2 = dynamic_cast<PointCloud*>(so2);
	if( m_DrawFeatureCorrespondences && pc1 && pc2)
	{
		FeatureCorrespondenceSet * set = dynamic_cast<FeatureCorrespondenceSet*>(
			pc1->getAttachments()->getData(FeatureCorrespondenceSet::getDefaultName()));
		if( set )
			drawCorrespondenceSet(set, frame1, frame2);
	}
}
