//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "ARMergeSequences.h"
//---------------------------------------------------------------------------
#include "ARMergeSettings.h"
#include "ARSequence.h"
#include "ARDeformationOptimizationContext.h"
#include "ARPairwiseMatching.h"
#include "DeformationSolver.h"
#include "DeformationEvaluator.h"
#include "RigidDeformationContext.h"
#include "ARRecomputeTopology.h"
#include "ARTopologicalResampling.h"
#include "ARSettings.h"
#include "Timer.h"
#include "ARUrshapeOptimization.h"
#include "ARPipelineOptions.h"
#include "ARGlobalDeformationOptimization.h"
#include "AttachedData.h"
#include "TopologyAttachment.h"
#include "DeformationTools.h"
#include "InCorePCTopologyGraph.h"
#include "ARChangeUrshape.h"
#include "ARPruneUnsupportedOutliers.h"
#include "InterfaceGlobals.h"
#include "SurfelGraphInit2.h"
#include "ARUrshapeDeformationOpt.h"
#include "TopologyRangeSearch.h"
#include "StringHelper.h"
#include "AREvaluateDeformation.h"
//---------------------------------------------------------------------------
#include <QtGui/QMessageBox>
#include "DebugRenderer.h"
#include "ARComputeShapeDistance.h"
#include "PointSetANNQuery.h"
#include "Histogram.h"
#include "MMDeformationContext.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

#define IFDBO if (settings->optimizationSettings->mergeDebugOutput)
static Vector3f debugLineColor = makeVector3f(1.0f, 0.25f, 0.25f);
static Vector3f debugLineOffset = NULL_VECTOR3F;

static card32 statistics_align_time = 0;
static card32 statistics_mergeShapes_time = 0;
static card32 statistics_resampling_time = 0;
static card32 statistics_holeFilling_time = 0;
static card32 statistics_topCheck_time = 0;
static card32 statistics_optUrshapeGeometry_time = 0;
static card32 statistics_urshapeDeformation_time = 0;
static card32 statistics_recomputingDeformation_time = 0;
static card32 statistics_globalDeformationOptimization_time = 0;

void X4::ARMergeSequences_DebugOutputTimings() {
   debugOutput << "align_time:\t" << statistics_align_time << "\n";
   debugOutput << "mergeShapes_time:\t" << statistics_mergeShapes_time << "\n";
   debugOutput << "resampling_time:\t" << statistics_resampling_time << "\n";
   debugOutput << "holeFilling_time:\t" << statistics_holeFilling_time << "\n";
   debugOutput << "topCheck_time:\t" << statistics_topCheck_time << "\n";
   debugOutput << "optUrshapeGeometry_time:\t" << statistics_optUrshapeGeometry_time << "\n";
   debugOutput << "urshapeDeformation_time:\t" << statistics_urshapeDeformation_time << "\n";
   debugOutput << "recomputingDeformation_time:\t" << statistics_recomputingDeformation_time << "\n";
   debugOutput << "globalDeformationOptimization_time:\t" << statistics_globalDeformationOptimization_time << "\n";
}



ARMergeSequences::ARMergeSequences() {
   visTool = NULL;
   myVisPC = NULL;
   myVisColor = makeVector4f(1,0,1,1);
}

ARMergeSequences::~ARMergeSequences() {}




ARSequence *ARMergeSequences::mergeSequences(
	Scene* scene,
	ARSequence *seqA, 
	ARSequence *seqB, 
	card32 seqAStartTime,
	card32 seqBStartTime,
   const std::vector<PointCloud*> &dataPoints,
	PointCloud *urshapePropagationInput,
	PointCloud *&urshapePropagationOutput,
	ARMergeSettings *settings, 
	Delegate visualFeedback) 
{
   if (seqA->getDuration() != 0) {
      x4Assert(!urshapePropagationInput);
   }
   
   setupDefRefitChannel(seqA->getUrshape(), seqBStartTime-1);
   setupDefRefitChannel(seqB->getUrshape(), seqBStartTime);
   if (urshapePropagationInput) {
      setupDefRefitChannel(urshapePropagationInput, -1);
   }

   DeformationEvaluator *evaluator = settings->deformationSolver->createEvaluator();
   PointCloud *A_end = seqA->evaluateEnd(evaluator, settings->optimizationSettings);
   PointCloud *B_start = seqB->evaluateStart(evaluator, settings->optimizationSettings);

   InCorePCTopologyGraph *topA = getTopology(seqA->getUrshape());
   if (topA) {
		TopologyAttachment *topAtt = new TopologyAttachment();
		topAtt->setup("TopologyAtt", AttachedData::ADF_PERSISTENT);
		topAtt->setTopology((InCorePCTopologyGraph*)topA->copy());
		A_end->getAttachments()->attachData(topAtt);
   }
         
   InCorePCTopologyGraph *topB = getTopology(seqB->getUrshape());
   if (topB) {
		TopologyAttachment *topAtt = new TopologyAttachment();
		topAtt->setup("TopologyAtt", AttachedData::ADF_PERSISTENT);
		topAtt->setTopology((InCorePCTopologyGraph*)topB->copy());
		B_start->getAttachments()->attachData(topAtt);
   }

   //
   // --- urshape propagation (merge propagated urshape from earlier frames with a_end == a_start
   //
   PointCloud *P_deformedTo_A = NULL;
   Timer timer;
   card32 numIterations_P_to_A = 0;
   if (urshapePropagationInput) {
      x4Assert(settings->pipelineOptions->urshapePropagation);
      if (seqA->getDuration() > 0 || seqB->getDuration() > 0) {
         throw EInvalidState("ARMergeSequences::mergeSequences() - urshape propagation is used only for level-1 merges.");
      }
      IFDBO debugOutput << "   propagating previous urshape...\n";

      ARDeformationOptimizationContext *P_deformedTo_A_ctxt = NULL;
      ARPairwiseMatching matcher;
      visTool = &matcher;
      try {
			matcher.matchGeometry(
					scene,
					seqAStartTime,
            urshapePropagationInput, 
            A_end, 
            settings->deformationSolver, settings,
            P_deformedTo_A,
            P_deformedTo_A_ctxt, 
            numIterations_P_to_A,
            visualFeedback,
            true
         );
      } catch (...) {
         visTool = NULL;
         throw;   	
      }
      visTool = NULL;
      delete P_deformedTo_A_ctxt;


      x4Assert(seqBStartTime > seqAStartTime);
      x4Assert(seqBStartTime == seqAStartTime+seqA->getDuration()+1);
      initPropChannel(P_deformedTo_A, seqBStartTime-1);
      card32 alignTime = timer.getDeltaValue();
      IFDBO debugOutput << "   ...alignment completed (" << alignTime << " ms).\n";
      IFDBO debugOutput.flush();
      statistics_align_time += alignTime;
   }



   //
   // --- pairwise matching
   //
   IFDBO debugOutput << "   aligning successive frames...\n";
   IFDBO debugOutput.flush();
   ARPairwiseMatching matcher;
   PointCloud *A_end_plus_P = NULL;
   int32 P_in_A_deformedTo_B_startIndex = -1;
   if (P_deformedTo_A) {
      A_end_plus_P = mergePointClouds(A_end, P_deformedTo_A, settings->optimizationSettings);
      P_in_A_deformedTo_B_startIndex = (int32)A_end->getNumPoints();
   } else {
      A_end_plus_P = (PointCloud*)A_end->copy();
   }
   delete P_deformedTo_A;
   const card32 numPointsIn_A_plus_P = (card32)A_end_plus_P->getNumPoints();
   
   ARDeformationOptimizationContext *A_plus_P_deformedTo_B_ctxt = NULL;
   ARDeformationOptimizationContext *B_deformedTo_A_ctxt = NULL;
   PointCloud *A_plus_P_deformedTo_B = NULL;
   PointCloud *B_deformedTo_A = NULL;
   visTool = &matcher;
   card32 numIterations_A_to_B = 0;
   card32 numIterations_B_to_A = 0;
   try {
      IFDBO debugOutput << "   matching A --> B\n";
      IFDBO debugOutput.flush();
      matcher.matchGeometry(
         scene,
         seqBStartTime,
         A_end_plus_P, 
         B_start, 
         settings->deformationSolver, settings,
         A_plus_P_deformedTo_B, 
         A_plus_P_deformedTo_B_ctxt,
         numIterations_A_to_B, 
         visualFeedback
      );

      if (settings->optimizationSettings->icpTwoWayMatch) {
         IFDBO debugOutput << "   matching B <-- A\n";
         IFDBO debugOutput.flush();
         matcher.matchGeometry(
			   scene,
			   seqBStartTime-1,
            B_start, 
            A_end_plus_P,
            settings->deformationSolver, settings,
            B_deformedTo_A, 
            B_deformedTo_A_ctxt, 
            numIterations_B_to_A, 
            visualFeedback
         );
      }
   } catch (...) {
      visTool = NULL;
      throw;   	
   }
   visTool = NULL;
   
   bool BA_merge_is_suspicious;
   if (settings->optimizationSettings->icpTwoWayMatch) {
      BA_merge_is_suspicious =
            (numIterations_B_to_A == settings->optimizationSettings->maxNumICPIterations)
         || numIterations_B_to_A >= 2*numIterations_A_to_B;
   } else {
      BA_merge_is_suspicious = true;
   }

   card32 alignTime2 = timer.getDeltaValue();
   IFDBO debugOutput << "   ...alignment completed (" << alignTime2 << " ms).\n";
   statistics_align_time += alignTime2;
   IFDBO debugOutput.flush();

   // //
   // // --- compute mean shape distances
   // //
   // IFDBO debugOutput << "   computing shape distances:\n";
   // IFDBO debugOutput.flush();
   // ARComputeShapeDistance::computeShapeDistance(A_plus_P_deformedTo_B, B_start, settings->optimizationSettings);
   // ARComputeShapeDistance::computeShapeDistance(A_end_plus_P, B_deformedTo_A, settings->optimizationSettings);
   // IFDBO debugOutput << "   ...completed (" << timer.getDeltaValue() << " ms).\n";
   // IFDBO debugOutput.flush();

   if (settings->pipelineOptions->bailOut == ARPipelineOptions::STOP_AFTER_MERGE) return NULL;


   //
   // --- merge aligned shapes (copy in one pc, recompute topology in merge frame)
   //
   IFDBO debugOutput << "   merging shapes...\n";
   IFDBO debugOutput.flush();
   float32 saveTopNeighbors = settings->optimizationSettings->topMaxNumNeighbors;
   settings->optimizationSettings->topMaxNumNeighbors *= 2; // because of double #points
   UnstructuredInCorePointCloud *newUrshapeAB = mergeAlignedShapes(A_plus_P_deformedTo_B, B_start, settings->optimizationSettings);
   UnstructuredInCorePointCloud *newUrshapeBA = NULL;
   if (settings->optimizationSettings->icpTwoWayMatch) {
      newUrshapeBA = mergeAlignedShapes(A_end_plus_P, B_deformedTo_A, settings->optimizationSettings);
   }
   settings->optimizationSettings->topMaxNumNeighbors = saveTopNeighbors;
   card32 mergeTime = timer.getDeltaValue();
   IFDBO debugOutput << "   ...completed (" << mergeTime << " ms).\n";
   IFDBO debugOutput.flush();
   statistics_mergeShapes_time += mergeTime;
   delete A_plus_P_deformedTo_B;
   delete B_deformedTo_A;

   //
   // --- early exit for deformable matching only
   // 

   if (settings->pipelineOptions->matchingOnly) {
      delete A_plus_P_deformedTo_B_ctxt;
      delete B_deformedTo_A_ctxt;
      delete A_end_plus_P;
      delete newUrshapeBA;
      delete evaluator;
      delete A_end;
      delete B_start;
      ARSequence *result = new ARSequence();
      MMDeformationContext *ctxt = new MMDeformationContext();
      ctxt->initializeFromShape(newUrshapeAB, makeVector2f(seqAStartTime,seqBStartTime+seqB->getDuration()), settings->optimizationSettings);
      ctxt->setIdentity();
      result->clearAndSetup(ctxt, newUrshapeAB);
      return result;
   }

   //
   // --- check topology (early version)
   //
   card32 edgesRemovedBA = 0;
   card32 edgesRemovedAB = 0;
   std::vector<bool> keepVertices(newUrshapeAB->getNumPoints(), true);
   if (!settings->pipelineOptions->lateTopologyCheck && !settings->pipelineOptions->noTopologyCheck) {
      IFDBO debugOutput << "   checking topology...\n";
      IFDBO debugOutput.flush();

      debugLineColor = makeVector3f(1.0f, 0.25f, 0.25f);
      debugLineOffset = NULL_VECTOR3F;
      debugRenderer->clearDebugData();
      edgesRemovedAB = checkTopology5(
         newUrshapeAB,
         numPointsIn_A_plus_P,
         true,
         A_end_plus_P,
         settings->deformationSolver,
         settings->optimizationSettings,
         keepVertices,
         seqA, 
         seqB
      );
      if (settings->optimizationSettings->icpTwoWayMatch) {
         debugLineColor = makeVector3f(0.25f, 1.0f, 0.25f);
         debugLineOffset = makeVector3f(2.0f, 0.0f, 0.0f);
         edgesRemovedBA = checkTopology5(
            newUrshapeBA,
            numPointsIn_A_plus_P,
            false,
            B_start,
            settings->deformationSolver,
            settings->optimizationSettings,
            keepVertices,
            seqA, 
            seqB
         );
      }
   }
      
   delete A_plus_P_deformedTo_B_ctxt;
   delete B_deformedTo_A_ctxt;
   delete A_end_plus_P;

   WhereIsTheUrshapeType whereIsTheUrshape;
   UnstructuredInCorePointCloud *newUrshape;
   if (edgesRemovedAB >= edgesRemovedBA || BA_merge_is_suspicious) {
      newUrshape = newUrshapeAB;
      delete newUrshapeBA;
      whereIsTheUrshape = URSHAPE_IN_B;
      IFDBO debugOutput << "   choose first, AB merge (urshape in B)\n";
   } else {
      newUrshape = newUrshapeBA;
      delete newUrshapeAB;
      whereIsTheUrshape = URSHAPE_IN_A;
      IFDBO debugOutput << "   choose second, BA merge (urshape in A)\n";
   }
   
   if (!settings->pipelineOptions->lateTopologyCheck && !settings->pipelineOptions->noTopologyCheck) {
      prunePoints(newUrshape, getTopology(newUrshape), keepVertices);

      IFDBO debugOutput << "   merge AB removes " << edgesRemovedAB << " vertices" << "\n";
      IFDBO debugOutput << "   merge BA removes " << edgesRemovedBA << " vertices" << "\n";
      card32 topTime1 = timer.getDeltaValue();
      IFDBO debugOutput << "   ...completed (" << topTime1 << " ms).\n";
      IFDBO debugOutput.flush();
      statistics_topCheck_time += topTime1;
   }

   show(newUrshape, makeVector4f(1,0.5f,0.5f,1), visualFeedback, settings->optimizationSettings);

   //
   // --- refit deformation
   //
   IFDBO debugOutput << "   refitting deformation ('hole filling')...\n";
   IFDBO debugOutput.flush();
   ARDeformationOptimizationContext *newDeformation = recomputeDeformation3( 
      newUrshape, 
      seqA, 
      seqB,
      numPointsIn_A_plus_P,
      seqBStartTime,
      whereIsTheUrshape,
      settings->deformationSolver,
      settings->optimizationSettings);
   card32 refitTime1 = timer.getDeltaValue();
   IFDBO debugOutput << "   ...completed (" << refitTime1 << " ms).\n";
   IFDBO debugOutput.flush();
   statistics_holeFilling_time += refitTime1;
   
   //
   // --- resample urshape
   //
   PointCloud *resampledUrshape;
   if (settings->optimizationSettings->resampleMergedUrshapes) {
      IFDBO debugOutput << "   resampling urshape...\n";
      IFDBO debugOutput.flush();
      resampledUrshape = ARTopologicalResampling::resample(
         newUrshape, 
         settings->optimizationSettings->topResamplingSurfelSpacing,
         settings->optimizationSettings->topRelRadius,
         settings->optimizationSettings->topMaxNumNeighbors);
      card32 resampleTime = timer.getDeltaValue();
      IFDBO debugOutput << "   ...completed (" << resampleTime << " ms).\n";
      IFDBO debugOutput.flush();
      statistics_resampling_time += resampleTime;
   } else {
      resampledUrshape = (PointCloud*)newUrshape->copy();
   }
   
   delete newUrshape;
   
   ARSequence *result = new ARSequence();
   result->clearAndSetup(newDeformation, resampledUrshape);

   //
   // --- check topology (late version)
   //
   if (settings->pipelineOptions->lateTopologyCheck && !settings->pipelineOptions->noTopologyCheck) {
      IFDBO debugOutput << "   late topology check...\n";

      // change urshape (compute new context)
      IFDBO debugOutput << "      recomputing deformation...\n";
      IFDBO debugOutput.flush();
      ARDeformationOptimizationContext *resampledUrshapeDeformation = ARChangeUrshape::recomputeDeformation(
         result->getUrshape(), 
         NULL,
         result->getDeformation(), 
         settings->deformationSolver, 
         settings->optimizationSettings);
      
      delete result->getDeformation();
      result->setDeformation(resampledUrshapeDeformation);
      card32 recDef1 = timer.getDeltaValue();
      IFDBO debugOutput << "      ...completed (" << recDef1 << " ms).\n";
      IFDBO debugOutput.flush();
      statistics_recomputingDeformation_time += recDef1;
      
      IFDBO debugOutput << "      checking topology...\n";
      IFDBO debugOutput.flush();
      checkTopologyLater(
         result, 
         settings->deformationSolver,
         settings->optimizationSettings
      );
      card32 topologyCheckTimeLater = timer.getDeltaValue();
      IFDBO debugOutput << "      ...completed (" << topologyCheckTimeLater << " ms).\n";
      IFDBO debugOutput.flush();
      statistics_topCheck_time += topologyCheckTimeLater;
   }
      

	if( settings->optimizationSettings->outlier_removeSmallPatches )
		removeSmallComponents( resampledUrshape, settings->optimizationSettings );
   
   show(resampledUrshape, makeVector4f(0.5f,0.7f,1.0f,1), visualFeedback, settings->optimizationSettings);


   //
   // --- urshape geometry optimization
   //
   if (settings->pipelineOptions->mode == ARPipelineOptions::mmExtractSurfacesFirst
       && settings->pipelineOptions->optimizeShapeWhileMerging) {
      IFDBO debugOutput << "   optimizing urshape geometry...\n";
      IFDBO debugOutput.flush();
      recomputeNormalsAndFrames(result->getUrshape(), settings->optimizationSettings);
      ARUrshapeOptimization shapeOpt;
      PointCloud *optUrshape = shapeOpt.optimizeUrshape(result, settings->shapeSolver, evaluator, dataPoints, settings->optimizationSettings);
      delete result->getUrshape();
      result->setUrshape(optUrshape);
      card32 usGeoOptTime = timer.getDeltaValue();
      IFDBO debugOutput << "   ...completed (" << usGeoOptTime << " ms).\n";
      IFDBO debugOutput.flush();
      statistics_optUrshapeGeometry_time += usGeoOptTime;
   }
   show(result->getUrshape(), makeVector4f(0.7f,1.0f,1.0f,1), visualFeedback, settings->optimizationSettings);

   //
   // --- urshape deformation optimization
   //
   if (settings->pipelineOptions->urshapeDeformationOpt) {
      IFDBO debugOutput << "   deforming urshape to best fit the data...\n";
      IFDBO debugOutput.flush();
      ARUrshapeDeformationOpt opt;
      PointCloud *defUrshape = opt.optimizeUrshape(result, settings->deformationSolver, dataPoints, settings->optimizationSettings);
      InCorePCTopologyGraph *tg = (InCorePCTopologyGraph*)getTopology(result->getUrshape())->copy();
      result->getUrshape()->assign(defUrshape);
      delete defUrshape;
      attachTopology(result->getUrshape(), tg);
      card32 urshapeDeformationTime = timer.getDeltaValue();
      IFDBO debugOutput << "   ...completed (" << urshapeDeformationTime << " ms).\n";
      IFDBO debugOutput.flush();
      statistics_urshapeDeformation_time += urshapeDeformationTime;
   }
   show(result->getUrshape(), makeVector4f(0.2f, 0.8f, 0.8f,1), visualFeedback, settings->optimizationSettings);
   
   //
   // --- change urshape (compute new context)
   //
   if (settings->pipelineOptions->recomputeDeformationsInMerge) {
      IFDBO debugOutput << "   recomputing deformation for changed urshape...\n";
      IFDBO debugOutput.flush();
      ARDeformationOptimizationContext *resampledUrshapeDeformation = ARChangeUrshape::recomputeDeformation(
         result->getUrshape(), 
		 NULL,
         result->getDeformation(), 
         settings->deformationSolver, 
         settings->optimizationSettings);
      
      delete result->getDeformation();
      result->setDeformation(resampledUrshapeDeformation);
      card32 recompDeformationTime2 = timer.getDeltaValue();
      IFDBO debugOutput << "   ...completed (" << recompDeformationTime2 << " ms).\n";
      IFDBO debugOutput.flush();
      statistics_recomputingDeformation_time += recompDeformationTime2;
   }
   
   //
   // --- global deformation optimization - xxx
   //
   IFDBO debugOutput << "   global deformation optimization (fit deformation field to data)...\n";
   IFDBO debugOutput.flush();
   bool doUnsupportedCulling = settings->optimizationSettings->cullStartNumFrames <= result->getDuration()+1;
   ARPruneUnsupportedOutliers puo(settings->optimizationSettings, (card32)(result->getDuration()+1));
   ARGlobalDeformationOptimization deformOpt;
   deformOpt.globalOpt(
      result, 
      settings->deformationSolver, 
      dataPoints, settings->optimizationSettings, 
      doUnsupportedCulling?&puo:NULL
   );

   show(result->getUrshape(), makeVector4f(0.5f,0.5f,1.0f,1), visualFeedback, settings->optimizationSettings);

   if (doUnsupportedCulling) {
      IFDBO debugOutput << "   culling urshape points not supported by sufficient data...\n";
      IFDBO debugOutput.flush();
      std::vector<bool> prunePts;
      puo.getResult(prunePts);
      prunePointsAndReplace(result, prunePts);
   }
   card32 globalDeformOptTime = timer.getDeltaValue();
   IFDBO debugOutput << "   ...completed (" << globalDeformOptTime << " ms).\n";
   IFDBO debugOutput.flush();
   statistics_globalDeformationOptimization_time += globalDeformOptTime;

   show(result->getUrshape(), makeVector4f(0.7f, 0.7f, 0.7f,1), visualFeedback, settings->optimizationSettings);
   
   if (settings->pipelineOptions->urshapePropagation) {
      urshapePropagationOutput = AREvaluateDeformation::evaluatePC(
         result->getUrshape(), 
         result->getDeformation(), 
         evaluator, 
         result->getDuration(), 
         settings->optimizationSettings);
      assignTopology(result->getUrshape(), urshapePropagationOutput);
   } else { 
      // otherwise, it is set to NULL and no propagation will start
      urshapePropagationOutput = NULL;
   }
   
   delete evaluator;
   delete A_end;
   delete B_start;

   return result;
}



UnstructuredInCorePointCloud *ARMergeSequences::mergeAlignedShapes(
   PointCloud *shapeA, 
   PointCloud *shapeB, 
   ARSettings *settings)
{
   UnstructuredInCorePointCloud *shapeA_upc = dynamic_cast<UnstructuredInCorePointCloud*>(shapeA);
   UnstructuredInCorePointCloud *shapeB_upc = dynamic_cast<UnstructuredInCorePointCloud*>(shapeB);
   if (!shapeA_upc || !shapeB_upc) {
      notImplemented();
   }
   UnstructuredInCorePointCloud *result = (UnstructuredInCorePointCloud*)shapeA->copy();
   PointSet *resultPS = result->getPointSet();
   PointSet *bPS = shapeB_upc->getPointSet();
   if (!(resultPS->getWidth() == 1 && bPS->getWidth() == 1)) {
      notImplemented();
   }
   if (!resultPS->getDescr()->equal(bPS->getDescr())) {
      notImplemented();
   }
   card32 numPtsA = resultPS->getNumEntries();
   card32 numPtsB = bPS->getNumEntries();
   resultPS->changeHeight(numPtsA + numPtsB);
   for (card32 i=0; i<numPtsB; i++) {
      resultPS->setVertex(numPtsA+i, bPS->getDataPointer(i));
   }

   ARRecomputeTopology::recomputeTopology(result, settings);
   return result;
}


void ARMergeSequences::show(PointCloud *showPC, Vector4f showColor, Delegate repaint, ARSettings *settings) {
   try {
      myVisColor = showColor;
      myVisPC = showPC;
      visTool = NULL;
      repaint.callIfNotNull();
      if (settings->singleStepDebugVis) {
         static bool callOnce = false;
         if (!callOnce) {
            callOnce = true;
            if (QMessageBox::question(
               InterfaceGlobals::getMainWindow(), 
               "Single Step Visualization", "Stay in single step mode?", 
               QMessageBox::Yes | QMessageBox::No, 
               QMessageBox::Yes) == QMessageBox::No) 
            {
               settings->singleStepDebugVis = false;
            }
            callOnce = false;
         }
      }
   } catch (...) {
      myVisPC = NULL;
      myVisColor = makeVector4f(1,0,1,1);
      throw;
   }
   myVisPC = NULL;
   myVisColor = makeVector4f(1,0,1,1);
}

void ARMergeSequences::glDrawFeedback(GLContext *glContext, ARVisualizationSettings *visSettings) 
{
   if (visTool) {
      visTool->glDrawFeedback(glContext, visSettings);
   } else if (myVisPC) {
		SceneObjectIterator *it = myVisPC->createIterator(SOIT::CAP_BASIC_PC);
		BasicPointCloudIterator *bit = dynamic_cast<BasicPointCloudIterator*>(it);
		if (!bit) {delete it; return;}
		AAT positionAAT = myVisPC->getAAT("position");
		AAT normalAAT = myVisPC->getAAT("normal");
		glPushAttrib(GL_ALL_ATTRIB_BITS);
		glDisable(GL_LIGHTING);
		glDisable(GL_BLEND);
		glColor4fv(myVisColor.data());
		glBegin(GL_POINTS);
		while (!bit->atEnd()) {
			Vector3f pos = bit->get3f(positionAAT);
			glVertex3fv(pos.data());
			bit->next();
		}
		glEnd();
      bit->reset();
		glBegin(GL_LINES);
		glColor4f(0.8f,1.0f,0.7f,0.3f);
		while (!bit->atEnd()) {
			Vector3f pos = bit->get3f(positionAAT);
			Vector3f normal = bit->get3f(normalAAT);
			glVertex3fv(pos.data());
			glVertex3fv((pos+normal*0.03f).data());
			bit->next();
		}
		glEnd();
		delete it;
      InCorePCTopologyGraph *tg = getTopology(myVisPC);
      if (tg) {
         UnstructuredInCorePointCloud *upc = dynamic_cast<UnstructuredInCorePointCloud*>(myVisPC);
         if (upc) {
			   glDisable(GL_DEPTH_TEST);
			   glEnable(GL_BLEND);
			   glBlendFunc(GL_SRC_ALPHA, GL_ONE);
      		//glColor4f(1.0f, 1.0f, 1.0f, 0.02f);
      		Vector4f newCol = myVisColor;
      		newCol[3] = 0.05f;
      		glColor4fv(newCol.data());
      		glBegin(GL_LINES);
            PointSet *ps = upc->getPointSet();
            for (card32 i=0; i<ps->getNumEntries(); i++) {
               card32 numAdj = tg->getNumAdjacentVertices(i);
               for (card32 j=0; j<numAdj; j++) {
                  glVertex3fv(ps->get3f(i, positionAAT).data());
                  glVertex3fv(ps->get3f(tg->getAdjacentVertex(i, j), positionAAT).data());
               }
            }
            glEnd();
         }
      }
      glPopAttrib();
   }
}


void ARMergeSequences::initPropChannel(PointCloud *shape, int32 frame) 
{
   SceneObjectIterator *it = shape->createIterator(SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC);
   BasicPointCloudIterator *bit = dynamic_cast<BasicPointCloudIterator*>(it);
   ModifyablePointCloudIterator *mit = dynamic_cast<ModifyablePointCloudIterator*>(it);
   AAT propOriginAAT = shape->getAAT("propOrigin");

   card32 index = 0;
   while (!bit->atEnd()) {
      mit->set2i(propOriginAAT, makeVector2i(frame, index));
      index++;
      bit->next();
   }
   delete it;
}

void ARMergeSequences::removeSmallComponents(PointCloud *pc, ARSettings *settings) 
{
   UnstructuredInCorePointCloud *upc = dynamic_cast<UnstructuredInCorePointCloud*>(pc);
   if (!upc) notImplemented();
   PointSet *ps = upc->getPointSet();
   InCorePCTopologyGraph *tg = getTopology(pc);
   SurfelGraphInit2::removeSmallComponents(ps, tg, settings->outlier_minNumberOfPoints, settings);
   upc->getAttachments()->deleteDelOnWriteData();
}









//static bool checkLength4(float32 length1, float32 length2, ARSettings *settings) {
//   float32 minLength = min(length1, length2);
//   float32 maxLength = max(length1, length2);
//   if (maxLength-minLength > settings->topMaxRelAddLengthChange * settings->topSurfelSpacing) {
//      return false;
//   }
//   if (maxLength > settings->topMinRelLengthRelCheck * settings->topSurfelSpacing) {
//      if (settings->topMaxMultLengthChange > 0) {
//         if (minLength/maxLength < 1.0f/settings->topMaxMultLengthChange) {
//            return false;
//         }
//      }
//   }
//   return true;
//}

static void stressToVertexRemoval(
   const std::vector<Matrix3f> &stressTensors,
   std::vector<bool> &keepVertices,
   ARSettings * settings
);

//card32 ARMergeSequences::checkTopology4(
//   PointCloud *urshape,
//   card32 seqBStart,
//   bool deform_AP_to_B,
//   PointCloud *nonUrshapeFrame,
//   DeformationSolver *deformationSolver,
//   ARSettings *settings,
//   std::vector<bool> &keepVertices)
//{
//   // --- setup
//   UnstructuredInCorePointCloud *upc_urshape = dynamic_cast<UnstructuredInCorePointCloud*>(urshape);
//   UnstructuredInCorePointCloud *upc_nonUrshapeFrame = dynamic_cast<UnstructuredInCorePointCloud*>(nonUrshapeFrame);
//   if (!upc_urshape || !upc_nonUrshapeFrame) {
//      notImplemented();
//   }
//   PointSet *urshapePS         = upc_urshape->getPointSet();
//   PointSet *nonUrshapeFramePS = upc_nonUrshapeFrame->getPointSet();
//
//   AAT positionAAT = urshapePS->getAAT("position");
//   AAT normalAAT = urshapePS->getAAT("normal");
//   
//   x4Assert(seqBStart > 0);   
//   PointSet *shapeAP = copySegment(urshapePS, 0, seqBStart-1);
//   PointSet *shapeB = copySegment(urshapePS, seqBStart, urshapePS->getNumEntries()-1);
//   PointSetANNQuery shapeAP_knn(shapeAP, 1);
//   PointSetANNQuery shapeB_knn(shapeB, 1);
//
//   const card32 num_urshape_points = urshapePS->getNumEntries();
//   const card32 num_nonUrshapeFrame_points = nonUrshapeFramePS->getNumEntries();
//
//   InCorePCTopologyGraph *urshapeTG = getTopology(urshape);
//   if (!urshapeTG) {
//      throw X4Exception("ARMergeSequences::checkTopology4() - topology missing");
//   }
//
//   if (!urshapePS->getDescr()->equal(nonUrshapeFramePS->getDescr()))
//   {
//      notImplemented();
//   }
//
//   Matrix3f nullMatrix = makeMatrix3f(0,0,0, 0,0,0, 0,0,0);
//   std::vector<Matrix3f> stress(num_urshape_points, nullMatrix);
//
//   for (card32 urshape_index_1=0; urshape_index_1<num_urshape_points; urshape_index_1++) {
//      const card32 numAdjacent = urshapeTG->getNumAdjacentVertices(urshape_index_1);
//      Vector3f urshape_normal_1 = normalize(urshapePS->get3f(urshape_index_1, normalAAT));
//      for (card32 ad=0; ad<numAdjacent; ad++) {
//         const card32 urshape_index_2 = urshapeTG->getAdjacentVertex(urshape_index_1, ad);
//        
//         static const int POINT_FROM_AP = 1;
//         static const int POINT_FROM_B = 2;
//
//         int origin1 = (urshape_index_1 >= seqBStart) ? POINT_FROM_B : POINT_FROM_AP;
//         int origin2 = (urshape_index_2 >= seqBStart) ? POINT_FROM_B : POINT_FROM_AP;
//         
//         Vector3f urshape_pos_1 = urshapePS->get3f(urshape_index_1, positionAAT);
//         Vector3f urshape_pos_2 = urshapePS->get3f(urshape_index_2, positionAAT);
//
//         // positions in non-urshape frame
//         Vector3f nonUrshapeFrame_pos_1, nonUrshapeFrame_pos_2;
//
//         // maximum distance at which we look for neighbors to transfer deformations:
//         const float32 maxSearchDistance = urshapePS->getBoundingBox().getMaxSideLength() / settings->gridSize * 0.5f;
//
//         // first case: urshape is obtained by deforming last frame in seqA (+ propagated) to first frame of seqB.
//         // Thus: compare distances in urshape to distances in last frame of seqA.
//         if (deform_AP_to_B) {
//            if (origin1 == POINT_FROM_AP) {
//               nonUrshapeFrame_pos_1 = nonUrshapeFramePS->get3f(urshape_index_1, positionAAT);
//            } else {
//               card32 closestNeighborFromAP = shapeAP_knn.getNearestPointIndex(urshape_pos_1);
//               Vector3f closestNeighborFromAP_in_urshape = urshapePS->get3f(closestNeighborFromAP, positionAAT);
//               if (norm(closestNeighborFromAP_in_urshape - urshape_pos_1) > maxSearchDistance) continue;
//               Vector3f closestNeighborFromAP_in_nonUrshapeFrame = nonUrshapeFramePS->get3f(closestNeighborFromAP, positionAAT);
//               Vector3f delta = closestNeighborFromAP_in_nonUrshapeFrame - closestNeighborFromAP_in_urshape;
//               nonUrshapeFrame_pos_1 = urshape_pos_1 + delta;
//            }
//            if (origin2 == POINT_FROM_AP) {
//               nonUrshapeFrame_pos_2 = nonUrshapeFramePS->get3f(urshape_index_2, positionAAT);
//            } else {
//               card32 closestNeighborFromAP = shapeAP_knn.getNearestPointIndex(urshape_pos_2);
//               Vector3f closestNeighborFromAP_in_urshape = urshapePS->get3f(closestNeighborFromAP, positionAAT);
//               if (norm(closestNeighborFromAP_in_urshape - urshape_pos_2) > maxSearchDistance) continue;
//               Vector3f closestNeighborFromAP_in_nonUrshapeFrame = nonUrshapeFramePS->get3f(closestNeighborFromAP, positionAAT);
//               Vector3f delta = closestNeighborFromAP_in_nonUrshapeFrame - closestNeighborFromAP_in_urshape;
//               nonUrshapeFrame_pos_2 = urshape_pos_2 + delta;
//            }
//            
//         // second case: urshape is obtained by deforming first frame in seqB to the last frame of seqA (+P).
//         // Thus: compare distances in urshape to distances in first frame of seqB.
//         } else {
//            if (origin1 == POINT_FROM_B) {
//               // seqB has larger base index in urshape (+seqBStart)
//               nonUrshapeFrame_pos_1 = nonUrshapeFramePS->get3f(urshape_index_1-seqBStart, positionAAT);
//            } else {
//               card32 closestNeighborFromB = shapeB_knn.getNearestPointIndex(urshape_pos_1);
//               Vector3f closestNeighborFromB_in_urshape = urshapePS->get3f(closestNeighborFromB+seqBStart, positionAAT);
//               if (norm(closestNeighborFromB_in_urshape - urshape_pos_1) > maxSearchDistance) continue;
//               Vector3f closestNeighborFromB_in_nonUrshapeFrame = nonUrshapeFramePS->get3f(closestNeighborFromB, positionAAT);
//               Vector3f delta = closestNeighborFromB_in_nonUrshapeFrame - closestNeighborFromB_in_urshape;
//               nonUrshapeFrame_pos_1 = urshape_pos_1 + delta;
//            }
//            if (origin2 == POINT_FROM_B) {
//               // seqB has larger base index in urshape (+seqBStart)
//               nonUrshapeFrame_pos_2 = nonUrshapeFramePS->get3f(urshape_index_2-seqBStart, positionAAT);
//            } else {
//               card32 closestNeighborFromB = shapeB_knn.getNearestPointIndex(urshape_pos_2);
//               Vector3f closestNeighborFromB_in_urshape = urshapePS->get3f(closestNeighborFromB+seqBStart, positionAAT);
//               if (norm(closestNeighborFromB_in_urshape - urshape_pos_2) > maxSearchDistance) continue;
//               Vector3f closestNeighborFromB_in_nonUrshapeFrame = nonUrshapeFramePS->get3f(closestNeighborFromB, positionAAT);
//               Vector3f delta = closestNeighborFromB_in_nonUrshapeFrame - closestNeighborFromB_in_urshape;
//               nonUrshapeFrame_pos_2 = urshape_pos_2 + delta;
//            }
//
//         }
//
//         const float32 length1 = norm(urshape_pos_1 - urshape_pos_2);
//         const float32 length2 = norm(nonUrshapeFrame_pos_1 - nonUrshapeFrame_pos_2);
//         
//         // if points (almost) coincide, forget it
//         if (length1 <= 1e-10) continue;
//
//         Vector3f stressDir = urshape_pos_1 - urshape_pos_2;
//         Vector3f stressDirUV = stressDir - urshape_normal_1*(urshape_normal_1*stressDir);
//         float32 stressDirUVLength = norm(stressDirUV);
//         if (stressDirUVLength < 1e-10) continue;
//         stressDirUV /= stressDirUVLength;
//         
//         Matrix3f stressTensor = outerProduct(stressDirUV, stressDirUV)*(length2/length1);
//         stress[urshape_index_1] += stressTensor*(1.0f/numAdjacent);
//         stress[urshape_index_2] += stressTensor*(1.0f/numAdjacent);
//      }
//   }
//
//   card32 numDeletedVertices = 0;
//   for (size_t i=0; i<num_urshape_points; i++) {
//      Matrix3f m = stress[i];
//      Matrix3f eigenVectors;
//      Vector3f eigenValues;
//      m.computeEigenStructure(eigenValues, eigenVectors);
//      m.eigenSort(eigenValues, eigenVectors); // descending --> largest first
//
//      bool removeIt = false;
//      float32 c = (sqr(eigenValues[0]-1)+sqr(eigenValues[1]-1))+1;
//      if (c > settings->topCutMaxStressFactor) {
//         removeIt = true;
//      }
//      if (removeIt) {
//         keepVertices[i] = false;
//         numDeletedVertices++;
//      }
//   }
//
//   for (card32 k=0; k<settings->topCutBorder; k++) {
//      std::vector<bool> removeNext(num_urshape_points, false);
//      for (card32 i=0; i<num_urshape_points; i++) {
//         const card32 numAdjacent = urshapeTG->getNumAdjacentVertices(i);
//         for (card32 ad=0; ad<numAdjacent; ad++) {
//            const card32 j = urshapeTG->getAdjacentVertex(i, ad);
//            if (!keepVertices[j]) {
//               removeNext[i] = false;
//               break;
//            }
//         }
//      }
//      for (card32 i=0; i<num_urshape_points; i++) {
//         if (removeNext[i]) keepVertices[i] = false;
//      }
//   }
//
//   if (settings->debugOutputStressTensors) {
//      static int num = 0;
//      debugRenderer->beginRenderJob(std::string("topCheck")+intToStr(num++));
//      Histogram minStress, maxStress;
//      for (card32 i=0; i<num_urshape_points; i++) {
//         Matrix3f m = stress[i];
//         Matrix3f eigenVectors;
//         Vector3f eigenValues;
//         m.computeEigenStructure(eigenValues, eigenVectors);
//         m.eigenSort(eigenValues, eigenVectors);
//         float32 c = 1.0f/(sqr(eigenValues[0]-1)+sqr(eigenValues[1]-1));
//         debugRenderer->addFastEllipse(urshapePS->get3f(i, positionAAT)+debugLineOffset, eigenVectors, eigenValues*0.002f, debugLineColor*c, false);
//
//         float32 xmax = max(eigenValues[0], eigenValues[1]);
//         float32 xmin = min(eigenValues[0], eigenValues[1]);
//      
//         minStress.addEntry((int)(xmin*10), xmin);
//         maxStress.addEntry((int)(xmax*10), xmax);
//      
//      }
//      debugOutput << "Min Stress Histogram\n" << minStress.toString() << "\n\n";
//      debugOutput << "Max Stress Histogram\n" << maxStress.toString() << "\n\n";
//
//      debugRenderer->endRenderJob();
//   }
//   
//   delete shapeAP;
//   delete shapeB;
//
//   return numDeletedVertices; 
//}



card32 ARMergeSequences::checkTopology5(
   PointCloud *urshape,
   card32 seqBStart,
   bool deform_AP_to_B,
   PointCloud *nonUrshapeFrame,
   DeformationSolver *deformationSolver,
   ARSettings *settings,
   std::vector<bool> &keepVertices,
	ARSequence *seqA, 
	ARSequence *seqB)
{
   // --- setup
   UnstructuredInCorePointCloud *upc_urshape = dynamic_cast<UnstructuredInCorePointCloud*>(urshape);
   UnstructuredInCorePointCloud *upc_nonUrshapeFrame = dynamic_cast<UnstructuredInCorePointCloud*>(nonUrshapeFrame);
   if (!upc_urshape || !upc_nonUrshapeFrame) {
      notImplemented();
   }
   PointSet *urshapePS         = upc_urshape->getPointSet();
   PointSet *nonUrshapeFramePS = upc_nonUrshapeFrame->getPointSet();

   AAT positionAAT = urshapePS->getAAT("position");
   AAT normalAAT = urshapePS->getAAT("normal");
   
   x4Assert(seqBStart > 0);   
   PointSet *shapeAP = copySegment(urshapePS, 0, seqBStart-1);
   PointSet *shapeB = copySegment(urshapePS, seqBStart, urshapePS->getNumEntries()-1);
   PointSetANNQuery shapeAP_knn(shapeAP, 1);
   PointSetANNQuery shapeB_knn(shapeB, 1);

   const card32 num_urshape_points = urshapePS->getNumEntries();
   const card32 num_nonUrshapeFrame_points = nonUrshapeFramePS->getNumEntries();
   const card32 numPointsInSeqA = (card32)seqA->getUrshape()->getNumPoints();
   const card32 numPointsInSeqB = (card32)seqB->getUrshape()->getNumPoints();

   InCorePCTopologyGraph *urshapeTG = getTopology(urshape);
   if (!urshapeTG) {
      throw X4Exception("ARMergeSequences::checkTopology5() - topology missing");
   }

   if (!urshapePS->getDescr()->equal(nonUrshapeFramePS->getDescr()))
   {
      notImplemented();
   }

   Matrix3f nullMatrix = makeMatrix3f(0,0,0, 0,0,0, 0,0,0);
   std::vector<Matrix3f> stress(num_urshape_points, nullMatrix);
   std::vector<Matrix3f> stress2(num_urshape_points, nullMatrix);
   std::vector<Matrix3f> stress3(num_urshape_points, nullMatrix);

   DeformationEvaluator *evalA = deformationSolver->createEvaluator();
   evalA->setup(seqA->getDeformation(), settings);
   DeformationEvaluator *evalB = deformationSolver->createEvaluator();
   evalB->setup(seqB->getDeformation(), settings);
   
   std::vector<card32> dbgFlags(num_urshape_points, 0);

   for (card32 urshape_index_1=0; urshape_index_1<num_urshape_points; urshape_index_1++) {
      const card32 numAdjacent = urshapeTG->getNumAdjacentVertices(urshape_index_1);
      Vector3f urshape_normal_1 = normalize(urshapePS->get3f(urshape_index_1, normalAAT));
      for (card32 ad=0; ad<numAdjacent; ad++) {
         const card32 urshape_index_2 = urshapeTG->getAdjacentVertex(urshape_index_1, ad);
        
         static const int POINT_FROM_AP = 1;
         static const int POINT_FROM_B = 2;

         int origin1 = (urshape_index_1 >= seqBStart) ? POINT_FROM_B : POINT_FROM_AP;
         int origin2 = (urshape_index_2 >= seqBStart) ? POINT_FROM_B : POINT_FROM_AP;
         
         Vector3f urshape_pos_1 = urshapePS->get3f(urshape_index_1, positionAAT);
         Vector3f urshape_pos_2 = urshapePS->get3f(urshape_index_2, positionAAT);
         
         // positions in non-urshape frame
         Vector3f nonUrshapeFrame_pos_1, nonUrshapeFrame_pos_2;

         // maximum distance at which we look for neighbors to transfer deformations:
         const float32 maxSearchDistance = urshapePS->getBoundingBox().getMaxSideLength() / settings->gridSize * 0.5f;

         { // ---------------------------------------------------------------------------------------------------
           // -------------- first part: check urshape only (necessary to handle propagated "P"-points correctly)
           // -------------- this is copied from checkTopology4
           // ---------------------------------------------------------------------------------------------------
           
            // first case: urshape is obtained by deforming last frame in seqA (+ propagated) to first frame of seqB.
            // Thus: compare distances in urshape to distances in last frame of seqA.
            if (deform_AP_to_B) {
               if (origin1 == POINT_FROM_AP) {
                  nonUrshapeFrame_pos_1 = nonUrshapeFramePS->get3f(urshape_index_1, positionAAT);
               } else {
                  card32 closestNeighborFromAP = shapeAP_knn.getNearestPointIndex(urshape_pos_1);
                  Vector3f closestNeighborFromAP_in_urshape = urshapePS->get3f(closestNeighborFromAP, positionAAT);
                  float32 d = norm(closestNeighborFromAP_in_urshape - urshape_pos_1);
                  if (d > maxSearchDistance) continue;
                  Vector3f closestNeighborFromAP_in_nonUrshapeFrame = nonUrshapeFramePS->get3f(closestNeighborFromAP, positionAAT);
                  Vector3f delta = closestNeighborFromAP_in_nonUrshapeFrame - closestNeighborFromAP_in_urshape;
                  nonUrshapeFrame_pos_1 = urshape_pos_1 + delta;
               }
               if (origin2 == POINT_FROM_AP) {
                  nonUrshapeFrame_pos_2 = nonUrshapeFramePS->get3f(urshape_index_2, positionAAT);
               } else {
                  card32 closestNeighborFromAP = shapeAP_knn.getNearestPointIndex(urshape_pos_2);
                  Vector3f closestNeighborFromAP_in_urshape = urshapePS->get3f(closestNeighborFromAP, positionAAT);
                  float32 d = norm(closestNeighborFromAP_in_urshape - urshape_pos_2);
                  if (d > maxSearchDistance) continue;
                  Vector3f closestNeighborFromAP_in_nonUrshapeFrame = nonUrshapeFramePS->get3f(closestNeighborFromAP, positionAAT);
                  Vector3f delta = closestNeighborFromAP_in_nonUrshapeFrame - closestNeighborFromAP_in_urshape;
                  nonUrshapeFrame_pos_2 = urshape_pos_2 + delta;
               }
               
            // second case: urshape is obtained by deforming first frame in seqB to the last frame of seqA (+P).
            // Thus: compare distances in urshape to distances in first frame of seqB.
            } else {
               if (origin1 == POINT_FROM_B) {
                  // seqB has larger base index in urshape (+seqBStart)
                  nonUrshapeFrame_pos_1 = nonUrshapeFramePS->get3f(urshape_index_1-seqBStart, positionAAT);
               } else {
                  card32 closestNeighborFromB = shapeB_knn.getNearestPointIndex(urshape_pos_1);
                  Vector3f closestNeighborFromB_in_urshape = urshapePS->get3f(closestNeighborFromB+seqBStart, positionAAT);
                  float32 d = norm(closestNeighborFromB_in_urshape - urshape_pos_1);
                  if (d > maxSearchDistance) continue;
                  Vector3f closestNeighborFromB_in_nonUrshapeFrame = nonUrshapeFramePS->get3f(closestNeighborFromB, positionAAT);
                  Vector3f delta = closestNeighborFromB_in_nonUrshapeFrame - closestNeighborFromB_in_urshape;
                  nonUrshapeFrame_pos_1 = urshape_pos_1 + delta;
               }
               if (origin2 == POINT_FROM_B) {
                  // seqB has larger base index in urshape (+seqBStart)
                  nonUrshapeFrame_pos_2 = nonUrshapeFramePS->get3f(urshape_index_2-seqBStart, positionAAT);
               } else {
                  card32 closestNeighborFromB = shapeB_knn.getNearestPointIndex(urshape_pos_2);
                  Vector3f closestNeighborFromB_in_urshape = urshapePS->get3f(closestNeighborFromB+seqBStart, positionAAT);
                  float32 d = norm(closestNeighborFromB_in_urshape - urshape_pos_2);
                  if (d > maxSearchDistance) continue;
                  Vector3f closestNeighborFromB_in_nonUrshapeFrame = nonUrshapeFramePS->get3f(closestNeighborFromB, positionAAT);
                  Vector3f delta = closestNeighborFromB_in_nonUrshapeFrame - closestNeighborFromB_in_urshape;
                  nonUrshapeFrame_pos_2 = urshape_pos_2 + delta;
               }

            }

            const float32 length1 = norm(urshape_pos_1 - urshape_pos_2);
            const float32 length2 = norm(nonUrshapeFrame_pos_1 - nonUrshapeFrame_pos_2);
            
            // if points (almost) coincide, forget it
            if (length1 <= 1e-10) continue;

            Vector3f stressDir = urshape_pos_1 - urshape_pos_2;
            Vector3f stressDirUV = stressDir - urshape_normal_1*(urshape_normal_1*stressDir);
            float32 stressDirUVLength = norm(stressDirUV);
            if (stressDirUVLength < 1e-10) continue;
            stressDirUV /= stressDirUVLength;
            
            Matrix3f stressTensor = outerProduct(stressDirUV, stressDirUV)*(length2/length1 - 1);   // 1
            stress[urshape_index_1] += stressTensor*(1.0f/numAdjacent);
            stress[urshape_index_2] += stressTensor*(1.0f/numAdjacent);
         
         } // ---------------------------------------------------------------------------------------------------


         const card32 numFramesSeqA = seqA->getDeformation()->getDuration() + 1;
         const card32 numFramesSeqB = seqB->getDeformation()->getDuration() + 1;
         if (numFramesSeqA > 1 || numFramesSeqB > 1 && settings->topCheckWholeAnimation) {
           // ---------------------------------------------------------------------------------------------------
           // -------------- second part: check the whole sequence
           // -------------- this is necessary to avoid false connections in some situations
           // -------------- (points meet in merging frames without stress, but move apart at other times)
           // ---------------------------------------------------------------------------------------------------


            static const int32 UNDEFINED_POINT = -1;
            // these variables store the reference point indices to be used when looking for length changes in different sequences
            int32 index_of_1_in_A = UNDEFINED_POINT; 
            int32 index_of_2_in_A = UNDEFINED_POINT;
            int32 index_of_1_in_B = UNDEFINED_POINT;
            int32 index_of_2_in_B = UNDEFINED_POINT;
            
            // these variables store the difference vectors (in the urshape) between the points that are actually used and the original urshape points
            Vector3f delta_of_1_in_A = NULL_VECTOR3F;
            Vector3f delta_of_2_in_A = NULL_VECTOR3F;
            Vector3f delta_of_1_in_B = NULL_VECTOR3F;
            Vector3f delta_of_2_in_B = NULL_VECTOR3F;

            // ---- point 1
            if (origin1 == POINT_FROM_AP) {
               if (urshape_index_1 < numPointsInSeqA) { // check whether the the point is from SeqA, not from propagated points (P)
                  index_of_1_in_A = urshape_index_1;
               } // otherwise: undefined
               card32 closestNeighborFromB = shapeB_knn.getNearestPointIndex(urshape_pos_1);
               Vector3f closestNeighborFromB_in_urshape = shapeB->get3f(closestNeighborFromB, positionAAT);
               delta_of_1_in_B = closestNeighborFromB_in_urshape - urshape_pos_1;
               if (norm(delta_of_1_in_B) > maxSearchDistance) {
                  index_of_1_in_B = UNDEFINED_POINT;
               } else {
                  index_of_1_in_B = closestNeighborFromB;
               }
            } else { // origin1 == POINT_FROM_B
               index_of_1_in_B = urshape_index_1 - seqBStart;
               card32 closestNeighborFromAP = shapeAP_knn.getNearestPointIndex(urshape_pos_1);
               if (closestNeighborFromAP < numPointsInSeqA) {
                  Vector3f closestNeighborFromA_in_urshape = shapeAP->get3f(closestNeighborFromAP, positionAAT);
                  delta_of_1_in_A = closestNeighborFromA_in_urshape - urshape_pos_1;
                  if (norm(delta_of_1_in_A) > maxSearchDistance) {
                     index_of_1_in_A = UNDEFINED_POINT;
                  } else {
                     index_of_1_in_A = closestNeighborFromAP;
                  }
               }
            }
            
            // ---- point 2
            if (origin2 == POINT_FROM_AP) {
               if (urshape_index_2 < numPointsInSeqA) { // check whether the the point is from SeqA, not from propagated points (P)
                  index_of_2_in_A = urshape_index_2;
               } // otherwise: undefined
               card32 closestNeighborFromB = shapeB_knn.getNearestPointIndex(urshape_pos_2);
               Vector3f closestNeighborFromB_in_urshape = shapeB->get3f(closestNeighborFromB, positionAAT);
               delta_of_2_in_B = closestNeighborFromB_in_urshape - urshape_pos_2;
               if (norm(delta_of_2_in_B) > maxSearchDistance) {
                  index_of_2_in_B = UNDEFINED_POINT;
               } else {
                  index_of_2_in_B = closestNeighborFromB;
               }
            } else { // origin2 == POINT_FROM_B
               index_of_2_in_B = urshape_index_2 - seqBStart;
               card32 closestNeighborFromAP = shapeAP_knn.getNearestPointIndex(urshape_pos_2);
               if (closestNeighborFromAP < numPointsInSeqA) {
                  Vector3f closestNeighborFromA_in_urshape = shapeAP->get3f(closestNeighborFromAP, positionAAT);
                  delta_of_2_in_A = closestNeighborFromA_in_urshape - urshape_pos_2;
                  if (norm(delta_of_2_in_A) > maxSearchDistance) {
                     index_of_2_in_A = UNDEFINED_POINT;
                  } else {
                     index_of_2_in_A = closestNeighborFromAP;
                  }
               }
            }

            if (index_of_1_in_A == UNDEFINED_POINT) dbgFlags[urshape_index_1] |= 1;
            if (index_of_1_in_B == UNDEFINED_POINT) dbgFlags[urshape_index_1] |= 2;
            if (index_of_2_in_A == UNDEFINED_POINT) dbgFlags[urshape_index_2] |= 4;
            if (index_of_2_in_B == UNDEFINED_POINT) dbgFlags[urshape_index_2] |= 8;

            // base lengths to compare against
            float32 urshapeLength = norm(urshape_pos_1 - urshape_pos_2);
            if (urshapeLength <= 1e-10) continue;

            if (numFramesSeqA > 1) {
               if (index_of_1_in_A != UNDEFINED_POINT && index_of_2_in_A != UNDEFINED_POINT) {
                  for (card32 t_a = 0; t_a <numFramesSeqA; t_a++) {
                     evalA->beginQuery(t_a);
                     Vector3f pos_1_inSeqA = evalA->getPosition(ARConstraintAnchor(shapeAP->get3f(index_of_1_in_A, positionAAT), index_of_1_in_A));
                     pos_1_inSeqA -= delta_of_1_in_A;
                     Vector3f pos_2_inSeqA = evalA->getPosition(ARConstraintAnchor(shapeAP->get3f(index_of_2_in_A, positionAAT), index_of_2_in_A));
                     pos_2_inSeqA -= delta_of_2_in_A;
                     evalA->endQuery();

                     float32 newLength = norm(pos_1_inSeqA - pos_2_inSeqA);
                     Vector3f stressDir = urshape_pos_1 - urshape_pos_2;
                     Vector3f stressDirUV = stressDir - urshape_normal_1*(urshape_normal_1*stressDir);
                     float32 stressDirUVLength = norm(stressDirUV);
                     if (stressDirUVLength < 1e-10) continue;
                     stressDirUV /= stressDirUVLength;
                     
                     Matrix3f stressTensor = outerProduct(stressDirUV, stressDirUV)*(newLength/urshapeLength - 1);  // 2
                     stress2[urshape_index_1] += stressTensor*(1.0f/numAdjacent);
                     stress2[urshape_index_2] += stressTensor*(1.0f/numAdjacent);
                  }
               }
            }

            if (numFramesSeqB > 1) {
               if (index_of_1_in_B != UNDEFINED_POINT && index_of_2_in_B != UNDEFINED_POINT) {
                  for (card32 t_b = 0; t_b <numFramesSeqB; t_b++) {
                     evalB->beginQuery(t_b);
                     Vector3f pos_1_inSeqB = evalB->getPosition(ARConstraintAnchor(shapeB->get3f(index_of_1_in_B, positionAAT), index_of_1_in_B));
                     pos_1_inSeqB -= delta_of_1_in_B;
                     Vector3f pos_2_inSeqB = evalB->getPosition(ARConstraintAnchor(shapeB->get3f(index_of_2_in_B, positionAAT), index_of_2_in_B));
                     pos_2_inSeqB -= delta_of_2_in_B;
                     evalB->endQuery();

                     float32 newLength = norm(pos_1_inSeqB - pos_2_inSeqB);
                     Vector3f stressDir = urshape_pos_1 - urshape_pos_2;
                     Vector3f stressDirUV = stressDir - urshape_normal_1*(urshape_normal_1*stressDir);
                     float32 stressDirUVLength = norm(stressDirUV);
                     if (stressDirUVLength < 1e-10) continue;
                     stressDirUV /= stressDirUVLength;
                     
                     Matrix3f stressTensor = outerProduct(stressDirUV, stressDirUV)*(newLength/urshapeLength - 1); // 3
                     stress3[urshape_index_1] += stressTensor*(1.0f/numAdjacent);
                     stress3[urshape_index_2] += stressTensor*(1.0f/numAdjacent);
                  }
               }
            }
         }
      }
   }

   delete evalA;
   delete evalB;

   card32 numDeletedVertices = 0;
   for (size_t i=0; i<num_urshape_points; i++) {
      Matrix3f m = stress[i]+stress2[i]+stress3[i];
      Matrix3f eigenVectors;
      Vector3f eigenValues;
      m.computeEigenStructure(eigenValues, eigenVectors);
      m.eigenSort(eigenValues, eigenVectors); // descending --> largest first

      bool removeIt = false;
      float32 c = (sqr(eigenValues[0])+sqr(eigenValues[1]))+1;
      if (c > settings->topCutMaxStressFactor) {
         removeIt = true;
      }
      if (removeIt) {
         keepVertices[i] = false;
         numDeletedVertices++;
      }
   }

   for (card32 k=0; k<settings->topCutBorder; k++) {
      std::vector<bool> removeNext(num_urshape_points, false);
      for (card32 i=0; i<num_urshape_points; i++) {
         const card32 numAdjacent = urshapeTG->getNumAdjacentVertices(i);
         for (card32 ad=0; ad<numAdjacent; ad++) {
            const card32 j = urshapeTG->getAdjacentVertex(i, ad);
            if (!keepVertices[j]) {
               removeNext[i] = true;
               break;
            }
         }
      }
      for (card32 i=0; i<num_urshape_points; i++) {
         if (removeNext[i]) keepVertices[i] = false;
      }
   }

   if (settings->debugOutputStressTensors) {
      static int num = 0;
      debugRenderer->beginRenderJob(std::string("topCheck")+intToStr(num++));
      Histogram minStress, maxStress;
      for (card32 i=0; i<num_urshape_points; i++) {
         Matrix3f m = stress[i];
         Matrix3f eigenVectors;
         Vector3f eigenValues;
         m.computeEigenStructure(eigenValues, eigenVectors);
         m.eigenSort(eigenValues, eigenVectors);
         float32 c = (sqr(eigenValues[0])+sqr(eigenValues[1]));
         Vector3f evoff = eigenValues + makeVector3f(0.2f, 0.2f, 0.2f);
         float32 col = c / (settings->topCutMaxStressFactor-1);
         debugRenderer->addFastEllipse(urshapePS->get3f(i, positionAAT)+debugLineOffset, eigenVectors, evoff*0.02f, 
                                       (c+1 > settings->topCutMaxStressFactor)?makeVector3f(0.7f,0.4f,0.7f):(makeVector3f(1,1,1)*(1-col)+debugLineColor*col), false);

         float32 xmax = max(eigenValues[0], eigenValues[1]);
         float32 xmin = min(eigenValues[0], eigenValues[1]);
      
         minStress.addEntry((int)(xmin*10), xmin);
         maxStress.addEntry((int)(xmax*10), xmax);
      
      }
      debugOutput << "Min Stress Histogram\n" << minStress.toString() << "\n\n";
      debugOutput << "Max Stress Histogram\n" << maxStress.toString() << "\n\n";

      debugRenderer->endRenderJob();

      debugRenderer->beginRenderJob(std::string("topCheck")+intToStr(num++));
      for (card32 i=0; i<num_urshape_points; i++) {
         Matrix3f m = stress2[i];
         Matrix3f eigenVectors;
         Vector3f eigenValues;
         m.computeEigenStructure(eigenValues, eigenVectors);
         m.eigenSort(eigenValues, eigenVectors);
         float32 c = (sqr(eigenValues[0])+sqr(eigenValues[1]));
         Vector3f evoff = eigenValues + makeVector3f(0.2f, 0.2f, 0.2f);
         float32 col = c / (settings->topCutMaxStressFactor-1);
         debugRenderer->addFastEllipse(urshapePS->get3f(i, positionAAT)+debugLineOffset+makeVector3f(0,2,0), eigenVectors, evoff*0.02f, 
                                       (c+1 > settings->topCutMaxStressFactor)?makeVector3f(0.7f,0.4f,0.7f):(makeVector3f(1,1,1)*(1-col)+debugLineColor*col), false);
      }
      debugRenderer->endRenderJob();

      debugRenderer->beginRenderJob(std::string("topCheck")+intToStr(num++));
      for (card32 i=0; i<num_urshape_points; i++) {
         Matrix3f m = stress3[i];
         Matrix3f eigenVectors;
         Vector3f eigenValues;
         m.computeEigenStructure(eigenValues, eigenVectors);
         m.eigenSort(eigenValues, eigenVectors);
         float32 c = (sqr(eigenValues[0])+sqr(eigenValues[1]));
         Vector3f evoff = eigenValues + makeVector3f(0.2f, 0.2f, 0.2f);
         float32 col = c / (settings->topCutMaxStressFactor-1);
         debugRenderer->addFastEllipse(urshapePS->get3f(i, positionAAT)+debugLineOffset+makeVector3f(0,4,0), eigenVectors, evoff*0.02f, 
                                       (c+1 > settings->topCutMaxStressFactor)?makeVector3f(0.7f,0.4f,0.7f):(makeVector3f(1,1,1)*(1-col)+debugLineColor*col), false);
      }
      debugRenderer->endRenderJob();
      
      debugRenderer->beginRenderJob(std::string("topCheck")+intToStr(num++));
      for (card32 i=0; i<num_urshape_points; i++) {
         Vector3f originCol = (i >= seqBStart) ? makeVector3f(0.7f, 0.5f, 0.2f) : makeVector3f(1, 1, 0.2f);
         debugRenderer->addFastEllipse(urshapePS->get3f(i, positionAAT)+debugLineOffset+makeVector3f(0,-2,0), IDENTITY3F, makeVector3f(1,1,1)*0.02f, 
                                       originCol, false);
      }
      debugRenderer->endRenderJob();

      debugRenderer->beginRenderJob(std::string("topCheck")+intToStr(num++));
      for (card32 i=0; i<num_urshape_points; i++) {
         Vector3f col = NULL_VECTOR3F;
         if (dbgFlags[i] & 1) col += makeVector3f(0.5f,0,0);
         if (dbgFlags[i] & 2) col += makeVector3f(0,0.5f,0);
         if (dbgFlags[i] & 4) col += makeVector3f(0.5f,0,0);
         if (dbgFlags[i] & 8) col += makeVector3f(0,0.5f,0);
         debugRenderer->addFastEllipse(urshapePS->get3f(i, positionAAT)+debugLineOffset+makeVector3f(0,-4,0), IDENTITY3F, makeVector3f(1,1,1)*0.02f, 
                                       col, false);
      }
      debugRenderer->endRenderJob();

   }
   
   delete shapeAP;
   delete shapeB;

   return numDeletedVertices; 
}






















ARDeformationOptimizationContext *ARMergeSequences::recomputeDeformation3( 
      PointCloud *newUrshape, 
      ARSequence *seqA, 
      ARSequence *seqB,
      card32 seqBStart,
      card32 seqBStartTime,
      WhereIsTheUrshapeType whereIsTheUrshape,
      DeformationSolver *solver,
      ARSettings *settings)
{
	ARDeformationOptimizationContext *newDeformation = (ARDeformationOptimizationContext*)solver->createOptContext();
	Vector2f newTimeInterval = makeVector2f(0, seqA->getDuration()+seqB->getDuration()+1);
	newDeformation->initializeFromShape(newUrshape, newTimeInterval, settings);
	solver->setup(newDeformation, settings);

   UnstructuredInCorePointCloud *upc_newUrshape = dynamic_cast<UnstructuredInCorePointCloud*>(newUrshape);
   if (!upc_newUrshape) notImplemented();
   PointSet *ps_newUrshape = upc_newUrshape->getPointSet();

   UnstructuredInCorePointCloud *upc_seqA = dynamic_cast<UnstructuredInCorePointCloud*>(seqA->getUrshape());
   if (!upc_seqA) notImplemented();
   PointSet *ps_seqA = upc_seqA->getPointSet();

   UnstructuredInCorePointCloud *upc_seqB = dynamic_cast<UnstructuredInCorePointCloud*>(seqB->getUrshape());
   if (!upc_seqB) notImplemented();
   PointSet *ps_seqB = upc_seqB->getPointSet();
   
   const card32 numNewUrshapePoints = (card32)newUrshape->getNumPoints();
   const card32 numSeqAPoints = (card32)seqA->getUrshape()->getNumPoints();
   const card32 numSeqBPoints = (card32)seqB->getUrshape()->getNumPoints();

   const card32 numSeqAFrames = (card32)seqA->getDuration()+1;
   const card32 numSeqBFrames = (card32)seqB->getDuration()+1;
   
   AAT positionAAT = ps_newUrshape->getAAT("position");
   AAT defRefitChannelAAT = ps_newUrshape->getAAT("defRefitChannel", 2, VAD::DATA_FORMAT_INT32);
   
	solver->beginConstraints();		

   // ---- seqA
   DeformationEvaluator *evalA = solver->createEvaluator();
   evalA->setup(seqA->getDeformation(), settings);
   for (card32 frame=0; frame<numSeqAFrames-(whereIsTheUrshape==URSHAPE_IN_A?1:0); frame++)
   {
      evalA->beginQuery(frame);
      for (card32 point=0; point<numNewUrshapePoints; point++) {

         Vector2i defRefitInfo = ps_newUrshape->get2i(point, defRefitChannelAAT);
         // defRefitInfo[0] == original frame from merge from {seqBStartTime-1, seqBStartTime, -1}
         // defRefitInfo[1] == point index in sequences (a or b)

         if (defRefitInfo[0] == -1) continue;
         if (defRefitInfo[0] == seqBStartTime) continue;
         x4Assert(defRefitInfo[0] == seqBStartTime-1);

         PositionARConstraint cPos;
         cPos.anchor = ARConstraintAnchor(
            ps_newUrshape->get3f(point, positionAAT),
            point
         );
         cPos.covariance = IDENTITY3F;
         cPos.time = frame;
         cPos.toPoint = evalA->getPosition(
            ARConstraintAnchor(
               ps_seqA->get3f(defRefitInfo[1], positionAAT),
               defRefitInfo[1]               
            )
         );
         solver->addPositionConstraint(&cPos);
         
      }
      evalA->endQuery();
   }
   delete evalA; evalA = NULL;

   for (card32 point=0; point<numNewUrshapePoints; point++) {

      PositionARConstraint cPos;
      cPos.anchor = ARConstraintAnchor(
         ps_newUrshape->get3f(point, positionAAT),
         point
      );
      cPos.covariance = IDENTITY3F;
      cPos.time = numSeqAFrames-(whereIsTheUrshape==URSHAPE_IN_A?1:0);
      cPos.toPoint = ps_newUrshape->get3f(point, positionAAT);
   }


   // ---- seqB
   DeformationEvaluator *evalB = solver->createEvaluator();
   evalB->setup(seqB->getDeformation(), settings);
   for (card32 frame=1; frame<numSeqBFrames; frame++)
   {
      evalB->beginQuery(frame);
      for (card32 point=(whereIsTheUrshape==URSHAPE_IN_B?1:0); point<numNewUrshapePoints; point++) {

         Vector2i defRefitInfo = ps_newUrshape->get2i(point, defRefitChannelAAT);
         // defRefitInfo[0] == original frame from merge from {seqBStartTime-1, seqBStartTime, -1}
         // defRefitInfo[1] == point index in sequences (a or b)

         if (defRefitInfo[0] == -1) continue;
         if (defRefitInfo[0] == seqBStartTime-1) continue;
         x4Assert(defRefitInfo[0] == seqBStartTime);

         PositionARConstraint cPos;
         cPos.anchor = ARConstraintAnchor(
            ps_newUrshape->get3f(point, positionAAT),
            point
         );
         cPos.covariance = IDENTITY3F;
         cPos.time = frame+numSeqAFrames;
         cPos.toPoint = evalB->getPosition(
            ARConstraintAnchor(
               ps_seqB->get3f(defRefitInfo[1], positionAAT),
               defRefitInfo[1]               
            )
         );
         solver->addPositionConstraint(&cPos);
      }
      evalB->endQuery();
   }
   delete evalB; evalB = NULL;
	
	solver->solve();
	return newDeformation;
}



void ARMergeSequences::setupDefRefitChannel(PointCloud *p, int32 frameIndex) 
{
   UnstructuredInCorePointCloud *upc = dynamic_cast<UnstructuredInCorePointCloud*>(p);
   if (!upc) {
      notImplemented();
   }
   PointSet *ps = upc->getPointSet();
   AAT defRefitChannelAAT = ps->getAAT("defRefitChannel", 2, VAD::DATA_FORMAT_INT32);
   Vector2i v = makeVector2i(-1,-1);
   for (card32 i=0; i<ps->getNumEntries(); i++) {
      if (frameIndex != -1) {
         v = makeVector2i(frameIndex, i);
      }
      ps->set2i(i, defRefitChannelAAT, v);
   }
}


static void recomputeUrshapeGraph(
   PointCloud *urshape, 
   const std::vector<bool> &keepVertices, 
   const std::vector< std::vector<bool> > &keepEdges)
{
   UnstructuredInCorePointCloud *upc = dynamic_cast<UnstructuredInCorePointCloud*>(urshape);
   if (!upc) {
      notImplemented();
   }
   InCorePCTopologyGraph *tg = getTopology(urshape);
   
   InCorePCTopologyGraph *newTG = new InCorePCTopologyGraph();
   newTG->clearAndSetup(tg->getEdges()->getDescr());
   
   for (unsigned v=0; v<keepEdges.size(); v++) {
      for (unsigned e=0; e<keepEdges[v].size(); e++) {
         if (keepEdges[v][e]) {
            card32 otherV = tg->getAdjacentVertex(v,e);
            newTG->addEdge(v, otherV);
         }
      }
   }
   
   UnstructuredInCorePointCloud *finalPC;
   InCorePCTopologyGraph *finalTG;
   prunePoints(upc, newTG, finalPC, finalTG, keepVertices);
   delete newTG;
   upc->assign(finalPC);
   attachTopology(upc, finalTG);
   delete finalPC;
}


void ARMergeSequences::checkTopologyLater(
   ARSequence *seq,
   DeformationSolver *deformationSolver,
   ARSettings *settings)
{
   // --- setup
   UnstructuredInCorePointCloud *upc_urshape = dynamic_cast<UnstructuredInCorePointCloud*>(seq->getUrshape());
   if (!upc_urshape) {
      notImplemented();
   }
   PointSet *urshapePS         = upc_urshape->getPointSet();
   AAT positionAAT = urshapePS->getAAT("position");
   const card32 num_urshape_points = urshapePS->getNumEntries();
   InCorePCTopologyGraph *urshapeTG = getTopology(upc_urshape);
   if (!urshapeTG) {
      throw X4Exception("ARMergeSequences::checkTopology5() - topology missing");
   }

   DeformationEvaluator *eval = deformationSolver->createEvaluator();
   eval->setup(seq->getDeformation(), settings);

   std::vector< std::vector<bool> > keepEdge;
   std::vector< std::vector<float32> > urshapeEdgeLength;
   
   for (card32 urshape_index_1=0; urshape_index_1<num_urshape_points; urshape_index_1++) {
      const card32 numAdjacent = urshapeTG->getNumAdjacentVertices(urshape_index_1);
      keepEdge.push_back(std::vector<bool>(numAdjacent, true));
      urshapeEdgeLength.push_back(std::vector<float32>(numAdjacent, 0));
      Vector3f pos1 = urshapePS->get3f(urshape_index_1, positionAAT);
      for (card32 ad=0; ad<numAdjacent; ad++) {
         const card32 urshape_index_2 = urshapeTG->getAdjacentVertex(urshape_index_1, ad);
         Vector3f pos2 = urshapePS->get3f(urshape_index_2, positionAAT);
         urshapeEdgeLength[urshape_index_1][ad] = norm(pos2-pos1);
      }
   }
   
   card32 seqTime = (card32)(seq->getDuration()+1);
   
   card32 numEDeleted = 0;
   
   for (card32 t=0; t<seqTime; t++) {
      eval->beginQuery(t);
      for (card32 urshape_index_1=0; urshape_index_1<num_urshape_points; urshape_index_1++) {
         const card32 numAdjacent = urshapeTG->getNumAdjacentVertices(urshape_index_1);
         Vector3f pos1 = eval->getPosition(ARConstraintAnchor(urshapePS->get3f(urshape_index_1, positionAAT),urshape_index_1));
         for (card32 ad=0; ad<numAdjacent; ad++) {
            const card32 urshape_index_2 = urshapeTG->getAdjacentVertex(urshape_index_1, ad);
            Vector3f pos2 = eval->getPosition(ARConstraintAnchor(urshapePS->get3f(urshape_index_2, positionAAT),urshape_index_2));
            float32 newLength = norm(pos2-pos1);
            if (newLength > settings->topSurfelSpacing * settings->topRelRadius) {
               keepEdge[urshape_index_1][ad] = false;
               numEDeleted++;
            }
            float32 urshapeLength = urshapeEdgeLength[urshape_index_1][ad];
            if (urshapeLength > settings->topCutMinRelativeLength * settings->topSurfelSpacing) {
               float32 minLength = min(newLength, urshapeLength);
               float32 maxLength = max(newLength, urshapeLength);
               if (minLength > 1e-15) {
                  float32 ratio = maxLength / minLength;
                  if (ratio > settings->topCutMaxStressFactor) {
                     keepEdge[urshape_index_1][ad] = false;
                     numEDeleted++;
                  }
               }
            }
         }

      }
      eval->endQuery();
   }   
   delete eval;  

   card32 numVDeleted = 0;
   std::vector<bool> keepVertex(num_urshape_points, true);   
   for (card32 urshape_index_1=0; urshape_index_1<num_urshape_points; urshape_index_1++) {
      const card32 numAdjacent = urshapeTG->getNumAdjacentVertices(urshape_index_1);
      bool allDeleted = true;
      bool oneDeleted = false;
      for (card32 ad=0; ad<numAdjacent; ad++) {
         if (keepEdge[urshape_index_1][ad]) {
            allDeleted = false;
         } else {
            oneDeleted = true;
         }
      }

      if (allDeleted) {
         keepVertex[urshape_index_1] = false;
      }
      if (oneDeleted && settings->topCutBorder > 0) {
         keepVertex[urshape_index_1] = false;
      }
      if (!keepVertex[urshape_index_1]) numVDeleted++;
   }
   debugOutput << "      late top check: delete edges " << numEDeleted << ", delete vertices " << numVDeleted << "\n";
   recomputeUrshapeGraph(seq->getUrshape(), keepVertex, keepEdge);
}
