//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "AREvaluateDeformation.h"
//---------------------------------------------------------------------------
#include "PointCloud.h"
#include "ARDeformationOptimizationContext.h"
#include "DeformationEvaluator.h"
#include "UnstructuredInCorePointCloud.h"
#include "VertexDescriptor.h"
#include "ARConstraints.h"
#include "UnstructuredInCoreTriangleMesh.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------




UnstructuredInCorePointCloud *AREvaluateDeformation::evaluatePC(
   PointCloud *referenceShape, 
   ARDeformationOptimizationContext *deformation, 
   DeformationEvaluator *evaluator,
   float32 time,
   const ARSettings *settings,
	bool writeGradients)
{
   if (!referenceShape->getDescr()->providesAttribute("position", 3, VAD::DATA_FORMAT_FLOAT32)) {
      throw X4Exception("AREvaluateDeformation::evaluatePC() - referenceShape must provide a 3f position channel.");
   }
   SceneObjectIterator *it = referenceShape->createIterator(SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC);
   BasicPointCloudIterator *bit = dynamic_cast<BasicPointCloudIterator*>(it);
   if (!bit) {
      delete it;
      throw ETypeMismatch("AREvaluateDeformation::evaluatePC() - referenceShape does not provide basic access.");
   }

   UnstructuredInCoreTriangleMesh* mesh = dynamic_cast<UnstructuredInCoreTriangleMesh*>(referenceShape);

   UnstructuredInCorePointCloud *result;
   if (mesh)
   {
	   result = new UnstructuredInCoreTriangleMesh;
   }else
	   result = new UnstructuredInCorePointCloud;

	VertexDescriptor * vdResult = (VertexDescriptor*)referenceShape->getDescr()->copy();
	if( writeGradients )
	{
		vdResult->pushAttrib(VAD("deform_gradient",9,VAD::DATA_FORMAT_FLOAT32));
		vdResult->pushAttrib(VAD("deformation",2,VAD::DATA_FORMAT_FLOAT32));
	}

	result->clearAndSetup(vdResult, (card32)referenceShape->getNumPoints());
	delete vdResult;



   const PointSet *psSource = bit->lockBlockRead();
   PointSet *psTarget = result->getPointSet();
   const card32 numPoints = psSource->getNumEntries();
   AAT positionAAT = referenceShape->getAAT("position");
   AAT normalAAT = NULL_AAT;
   if (referenceShape->getDescr()->providesAttribute("normal", 3, VAD::DATA_FORMAT_FLOAT32)) {
      normalAAT = referenceShape->getAAT("normal");
   }
   AAT tangentUAAT = NULL_AAT;
   if (referenceShape->getDescr()->providesAttribute("tangent_u", 3, VAD::DATA_FORMAT_FLOAT32)) {
      tangentUAAT = referenceShape->getAAT("tangent_u");
   }
   AAT tangentVAAT = NULL_AAT;
   if (referenceShape->getDescr()->providesAttribute("tangent_v", 3, VAD::DATA_FORMAT_FLOAT32)) {
      tangentVAAT = referenceShape->getAAT("tangent_v");
   }
   bool transformCoords = (normalAAT != NULL_AAT) || (tangentUAAT != NULL_AAT) || (tangentVAAT != NULL_AAT);
	AAT deformGradientAAT = NULL_AAT;
	AAT deformGradientAAT_source = NULL_AAT;
	AAT deformGradientVisAAT = NULL_AAT;
	bool useSourceGradients = false;
	if( writeGradients )
	{
		deformGradientAAT = result->getAAT("deform_gradient");
		deformGradientVisAAT = result->getAAT("deformation");
		if( referenceShape->providesAttribute("deform_gradient") )
		{
			deformGradientAAT_source = referenceShape->getAAT("deform_gradient");
			useSourceGradients = true;
		}
	}
   
   evaluator->setup(deformation, settings);
   evaluator->beginQuery(time);
   for (card32 i=0; i<numPoints; i++) {
      ARConstraintAnchor at;
      Vector3f pos = psSource->get3f(i, positionAAT);
      at.setup(pos, i);
      Vector3f transformedPos = evaluator->getPosition(at);  // todo: normals
      psTarget->setVertex(i, psSource->getDataPointer(i));
      psTarget->set3f(i, positionAAT, transformedPos);
			

      if (transformCoords || writeGradients) {
         Matrix3f jac = evaluator->getSpatialGradient(at);
         Matrix3f transform = invertMatrix(jac).transpose();
			Vector3f newNormal;
         if (normalAAT != NULL_AAT) {
            newNormal = /*transform*/jac*psSource->get3f(i, normalAAT);
            float32 nn = norm(newNormal);
            if (nn > 1E-7) newNormal *= (1.0f/nn);
            psTarget->set3f(i, normalAAT, newNormal);
         }
         if (tangentUAAT != NULL_AAT) {
            psTarget->set3f(i, tangentUAAT, jac*psSource->get3f(i, tangentUAAT));
         }
         if (tangentVAAT != NULL_AAT) {
            psTarget->set3f(i, tangentVAAT, jac*psSource->get3f(i, tangentVAAT));
         }
			if( writeGradients )
			{
				Matrix3f grad = jac;
				if( useSourceGradients )
				{
					Matrix3f sourceGradient = psSource->get9f(i,deformGradientAAT_source);
					grad = grad * sourceGradient;
				}

				Matrix3f tmp = grad;//*(IDENTITY3F - outerProduct(newNormal,newNormal));
				tmp = tmp * tmp;
				Vector3f eigenValues;
				Matrix3f eigenVectors;
				tmp.computeEigenStructure( eigenValues, eigenVectors);
				psTarget->set2f(i,deformGradientVisAAT, makeVector2f(eigenValues[0],eigenValues[1]));
				psTarget->set9f(i,deformGradientAAT, grad);
			}
      }
   }
   evaluator->endQuery();
   
   bit->unlock();
   delete it;

	if (mesh)
	{
		UnstructuredInCoreTriangleMesh* resMesh = dynamic_cast<UnstructuredInCoreTriangleMesh*>(result);
		resMesh->setTriangles((PointSet*)mesh->getTriangles()->copy());
	}
	
   
   return result;
}

UnstructuredInCorePointCloud *AREvaluateDeformation::evaluatePC(
	PointCloud *referenceShape, 
	PointCloud *shape, 
	ARDeformationOptimizationContext *deformation, 
	DeformationEvaluator *evaluator,
	float32 time,
	const ARSettings *settings,
	bool writeGradients)
{
	if (!shape->getDescr()->providesAttribute("position", 3, VAD::DATA_FORMAT_FLOAT32)) {
		throw X4Exception("AREvaluateDeformation::evaluatePC() - shape must provide a 3f position channel.");
	}
	SceneObjectIterator *it = shape->createIterator(SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC);
	BasicPointCloudIterator *bit = dynamic_cast<BasicPointCloudIterator*>(it);
	if (!bit) {
		delete it;
		throw ETypeMismatch("AREvaluateDeformation::evaluatePC() - shape does not provide basic access.");
	}

	UnstructuredInCoreTriangleMesh* mesh = dynamic_cast<UnstructuredInCoreTriangleMesh*>(shape);
	HierarchicalKNNIterator hIt_ref(referenceShape,32,NULL);

	UnstructuredInCorePointCloud *result;
	if (mesh)
	{
		result = new UnstructuredInCoreTriangleMesh;
	}else
		result = new UnstructuredInCorePointCloud;

	VertexDescriptor * vdResult = (VertexDescriptor*)shape->getDescr()->copy();
	if( writeGradients )
	{
		vdResult->pushAttrib(VAD("deform_gradient",9,VAD::DATA_FORMAT_FLOAT32));
		vdResult->pushAttrib(VAD("deformation",2,VAD::DATA_FORMAT_FLOAT32));
	}

	result->clearAndSetup(vdResult, (card32)shape->getNumPoints());
	delete vdResult;



	const PointSet *psSource = bit->lockBlockRead();
	PointSet *psTarget = result->getPointSet();
	const card32 numPoints = psSource->getNumEntries();
	AAT positionAAT = shape->getAAT("position");
	AAT normalAAT = NULL_AAT;
	if (shape->getDescr()->providesAttribute("normal", 3, VAD::DATA_FORMAT_FLOAT32)) {
		normalAAT = shape->getAAT("normal");
	}
	AAT tangentUAAT = NULL_AAT;
	if (shape->getDescr()->providesAttribute("tangent_u", 3, VAD::DATA_FORMAT_FLOAT32)) {
		tangentUAAT = shape->getAAT("tangent_u");
	}
	AAT tangentVAAT = NULL_AAT;
	if (shape->getDescr()->providesAttribute("tangent_v", 3, VAD::DATA_FORMAT_FLOAT32)) {
		tangentVAAT = shape->getAAT("tangent_v");
	}
	bool transformCoords = (normalAAT != NULL_AAT) || (tangentUAAT != NULL_AAT) || (tangentVAAT != NULL_AAT);
	AAT deformGradientAAT = NULL_AAT;
	AAT deformGradientAAT_source = NULL_AAT;
	AAT deformGradientVisAAT = NULL_AAT;
	bool useSourceGradients = false;
	if( writeGradients )
	{
		deformGradientAAT = result->getAAT("deform_gradient");
		deformGradientVisAAT = result->getAAT("deformation");
		if( shape->providesAttribute("deform_gradient") )
		{
			deformGradientAAT_source = shape->getAAT("deform_gradient");
			useSourceGradients = true;
		}
	}

	evaluator->setup(deformation, settings);
	evaluator->beginQuery(time);
	for (card32 i=0; i<numPoints; i++) {
		ARConstraintAnchor at;
		Vector3f pos = psSource->get3f(i, positionAAT);
		hIt_ref.setSeekPointAndReset(pos);
		at.setup(pos, hIt_ref.getCurrentPointIndex());
		Vector3f transformedPos = evaluator->getPosition(at);  // todo: normals
		psTarget->setVertex(i, psSource->getDataPointer(i));
		psTarget->set3f(i, positionAAT, transformedPos);


		if (transformCoords || writeGradients) {
			Matrix3f jac = evaluator->getSpatialGradient(at);
			Matrix3f transform = invertMatrix(jac).transpose();
			Vector3f newNormal;
			if (normalAAT != NULL_AAT) {
				newNormal = transform*psSource->get3f(i, normalAAT);
				float32 nn = norm(newNormal);
				if (nn > 1E-7) newNormal *= (1.0f/nn);
				psTarget->set3f(i, normalAAT, newNormal);
			}
			if (tangentUAAT != NULL_AAT) {
				psTarget->set3f(i, tangentUAAT, jac*psSource->get3f(i, tangentUAAT));
			}
			if (tangentVAAT != NULL_AAT) {
				psTarget->set3f(i, tangentVAAT, jac*psSource->get3f(i, tangentVAAT));
			}
			if( writeGradients )
			{
				Matrix3f grad = jac;
				if( useSourceGradients )
				{
					Matrix3f sourceGradient = psSource->get9f(i,deformGradientAAT_source);
					grad = grad * sourceGradient;
				}

				Matrix3f tmp = grad;//*(IDENTITY3F - outerProduct(newNormal,newNormal));
				tmp = tmp * tmp;
				Vector3f eigenValues;
				Matrix3f eigenVectors;
				tmp.computeEigenStructure( eigenValues, eigenVectors);
				psTarget->set2f(i,deformGradientVisAAT, makeVector2f(eigenValues[0],eigenValues[1]));
				psTarget->set9f(i,deformGradientAAT, grad);
			}
		}
	}
	evaluator->endQuery();

	bit->unlock();
	delete it;

	if (mesh)
	{
		UnstructuredInCoreTriangleMesh* resMesh = dynamic_cast<UnstructuredInCoreTriangleMesh*>(result);
		resMesh->setTriangles((PointSet*)mesh->getTriangles()->copy());
	}


	return result;
}