//---------------------------------------------------------------------------
#include "stdafx.h"
//---------------------------------------------------------------------------
#include "NearIsometricEmbedding.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "InCorePCTopologyGraph.h"
#include "PointSet.inline.h"
#include "PointSetANNQuery.h"
#include "TopologyRangeSearch.h"
#include "PCA.h"
#include "IterativeSolvers.inline.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

IMPLEMENT_X4_ABSTRACT_CLASS(NIEManifold, 0)
{
	BEGIN_CLASS_INIT(NIEManifold)
}

IMPLEMENT_X4_CLASS(NIEPointSetManifold, 0)
{
	BEGIN_CLASS_INIT(NIEPointSetManifold)
}

IMPLEMENT_X4_CLASS(NearIsometricEmbedding, 0)
{
	BEGIN_CLASS_INIT(NearIsometricEmbedding)
}

//---------------------------------------------------------------------------

void NIEManifold::assign(const Object* obj, X4CopyContext *context)
{
	copyObjectProperties(obj, this);
}

//---------------------------------------------------------------------------


NearIsometricEmbedding::NearIsometricEmbedding() {}

NearIsometricEmbedding::~NearIsometricEmbedding() {}

void NearIsometricEmbedding::assign(const Object* obj, X4CopyContext *context)
{
	copyObjectProperties(obj, this);
}

//---------------------------------------------------------------------------

NIEPointSetManifold::NIEPointSetManifold() 
{
   ps = 0;
   tg = 0;
   knn = 0;
   topSearch = 0;
}

void NIEPointSetManifold::assign(const Object* obj, X4CopyContext *context)
{
	copyObjectProperties(obj, this);
}

void NIEPointSetManifold::clearAndSetup(const PointSet *ps, const InCorePCTopologyGraph *tg, float32 sampleSpacing, float32 mlsRadiusFact) 
{
   // point set
   {
      delete this->ps;
      this->ps = new PointSet();
      VertexDescriptor vd;
      vd.pushAttrib(mVAD("position", 3, VAD::DATA_FORMAT_FLOAT32));
      this->ps->clearAndSetup(1,0,&vd);
      positionAAT = vd.getToken("position");
      AAT otherPosAAT = ps->getAAT("position");
      mpcard n = ps->getNumEntries();
      for (mpcard i=0; i<n; i++) {
         Vector3f pos = ps->get3f(i, otherPosAAT);
         this->ps->addPoint((char*)&pos);
      }
   }
   // topology
   {
      delete this->tg;
      this->tg = new InCorePCTopologyGraph();
	   VertexDescriptor vd;
	   vd.pushAttrib(VAD("vindex", 2, VAD::DATA_FORMAT_INT32));
	   this->tg->clearAndSetup(&vd);
      const card32 numVertices = tg->getNumVertices();
      for (card32 i=0; i<numVertices; i++) {
         const card32 nVert = tg->getNumAdjacentVertices(i);
         for (card32 j=0; j<nVert; j++) {
            this->tg->addEdge(i, tg->getAdjacentVertex(i, j));
         }
      }
   }
   // parameters
   {
      this->sampleSpacing = sampleSpacing;
      this->mlsRadiusFact = mlsRadiusFact;
   }
   // knn
   {
      knn = new PointSetANNQuery(this->ps, 1);
      topSearch = new TopNeighborhoodSphereSearch();
   }
}

//void NIEPointSetManifold::sampleSurface(float32 epsilon, std::vector<Vector3f> &result) 
//{
//}

void NIEPointSetManifold::computeMLS(const Vector3f &point, const int32 index, Matrix3f &uvn, Vector3f &proj) 
{
   topSearch->setup(tg, ps, index, ps->get3f(index, positionAAT), sampleSpacing*mlsRadiusFact, positionAAT);
   //std::vector<Vector3f> pts;
   PCA3f pca;
   while (!topSearch->atEnd()) {
      card32 cInd = topSearch->getCurrentIndex();
      pca.addPoint(ps->get3f(cInd, positionAAT));
      topSearch->next();
   }
   Vector3f lambda;
   x4Assert(fabs(lambda[0]) <= fabs(lambda[1]) && fabs(lambda[1]) <= fabs(lambda[2]));
   pca.analyze(lambda, uvn, proj);
   // normal last
   swap(uvn[0],uvn[2]);
   uvn[0] = normalize(uvn[0]);
   uvn[1] = normalize(uvn[1]);
   uvn[2] = normalize(uvn[2]);
}

Vector3f NIEPointSetManifold::project(const Vector3f &point) 
{
   int32 index = knn->getNearestPointIndex(point);
   Matrix3f uvn;
   Vector3f proj;
   computeMLS(point, index, uvn, proj);
   return proj;
}

void NIEPointSetManifold::getTangentSystem(const Vector3f &point, Vector3f &projection, Matrix3f &uvn) 
{
   int32 index = knn->getNearestPointIndex(point);
   computeMLS(point, index, uvn, projection);
}

mpcard NIEPointSetManifold::getOverallNumEdges() {
   mpcard n = tg->getNumVertices();
   mpcard result = 0;
   for (mpcard i=0; i<n; i++) {
      result += tg->getNumAdjacentVertices(i);
   }
   return result;
}


NIEPointSetManifold::~NIEPointSetManifold() {
   delete ps;
   delete tg;
   delete knn;
   delete topSearch;
}

//---------------------------------------------------------------------------

void NearIsometricEmbedding::clearAndSetup(NIEPointSetManifold *sourceShape, NIEManifold *targetShape, NIEPointSetManifold *embedding) 
{
   this->sourceShape = sourceShape;
   this->targetShape = targetShape;
   this->embedding = embedding;
   constraints.resize(embedding->getNumPts());
}

void NearIsometricEmbedding::addConstraint(const card32 embIndex, const Vector3f &pos, const float32 weight) 
{
   if (constraints[embIndex].weight > 0) {
      warning("deleting previous constraint");
   }
   constraints[embIndex].target = pos;
   constraints[embIndex].weight = weight;
}

void NearIsometricEmbedding::solveStep()
{
   const mpcard numPts = embedding->getNumPts();

   // initial projection on manifold;   
   std::vector< Matrix3f > tangentSystems(numPts);
   for (mpcard i=0; i<numPts; i++) {
      Vector3f prj;
      targetShape->getTangentSystem(embedding->getPoint(i), prj, tangentSystems[i]);
      embedding->setPoint(i, prj);
   }

   // gradient computation   
   std::vector< std::vector<Vector2f> > grads(numPts);
   std::vector< std::vector<float32> >  lengthsSource(numPts);
   std::vector< std::vector<float32> >  lengthsDest(numPts);
   for (mpcard i=0; i<numPts; i++) {
      const mpcard numAdj = embedding->getNumAdjacent(i);
      lengthsSource[i].resize(numAdj, 0.0f);
      lengthsDest[i].resize(numAdj, 0.0f);
      grads[i].resize(numAdj, NULL_VECTOR2F);
      for (mpcard j=0; j<numAdj; j++) {
         mpcard adjIndex = embedding->getAdjacentIndex(i,j);

         lengthsSource[i][j] = norm(sourceShape->getPoint(i) - sourceShape->getPoint(adjIndex));
         lengthsDest[i][j] = norm(embedding->getPoint(i) - embedding->getPoint(adjIndex));
         Vector3f gradD = (embedding->getPoint(i) - embedding->getPoint(adjIndex));
         gradD /= norm(gradD);
         Vector3f inUVN = tangentSystems[i].transpose()*gradD;
         grads[i][j] = makeVector2f(inUVN[0], inUVN[1]);
      }
   }

   SparseMatrixF M(2*numPts);
   DVectorF b(2*numPts);
   b.setZero();
   
   for (card32 k=0; k<numPts; k++) {
      card32 numAdj = embedding->getNumAdjacent(k);
      for (card32 i_ind=0; i_ind<numAdj; i_ind++) {
         const card32 i = embedding->getAdjacentIndex(k, i_ind);
         if (k == i) continue;
         M[2*k + 0].addEntry(2*i+0, grads[i][k][0] * grads[k][i][0]);
         M[2*k + 0].addEntry(2*i+1, grads[i][k][1] * grads[k][i][0]);
         M[2*k + 1].addEntry(2*i+0, grads[i][k][0] * grads[k][i][1]);
         M[2*k + 1].addEntry(2*i+1, grads[i][k][1] * grads[k][i][1]);
         
         M[2*k + 0].addEntry(2*k+0, grads[k][i][0] * grads[k][i][0]);
         M[2*k + 0].addEntry(2*k+1, grads[k][i][1] * grads[k][i][0]);
         M[2*k + 1].addEntry(2*k+0, grads[k][i][0] * grads[k][i][1]);
         M[2*k + 1].addEntry(2*k+1, grads[k][i][1] * grads[k][i][1]);

         b[2*k + 0] += (lengthsSource[i][k] - lengthsDest[i][k]) * grads[k][i][0];
         b[2*k + 1] += (lengthsSource[i][k] - lengthsDest[i][k]) * grads[k][i][1];
      }
   }

   DVectorF x(2*numPts);
   card32 numIt = 1000;
   conjugateGradientsSolve<float32>(M, x, b, numIt, 1E-7f, true);
   
   for (mpcard i=0; i<numPts; i++) {
      embedding->setPoint(i, embedding->getPoint(i) + tangentSystems[i]*makeVector3f(x[2*i+0], x[2*i+1], 0));
   }
}