//---------------------------------------------------------------------------
#ifndef NearIsometricEmbeddingH
#define NearIsometricEmbeddingH
//---------------------------------------------------------------------------
#include "x3deformation.h"
//---------------------------------------------------------------------------
#include "X4Types.h"
#include "InCorePCTopologyGraph.h"
#include "PointSet.inline.h"
//---------------------------------------------------------------------------


namespace X4 {

class PointSetANNQuery;
class TopNeighborhoodSphereSearch;


class NIESettings : public Persistent {
   X4_CLASS( NIESettings )
 public:
   float32 epsilon;
   
};

/// abstract class that describes a manifold embedded in 3-space
class NIEManifold : public Persistent {
	X4_ABSTRACT_CLASS( NIEManifold )
 public:
   NIEManifold() {}
	virtual void assign(const Object* obj, X4CopyContext *context = NULL);
   
   //virtual void sampleSurface(float32 epsilon, std::vector<Vector3f> &result) = 0;
   virtual Vector3f project(const Vector3f &point) = 0;
   virtual void getTangentSystem(const Vector3f &point, Vector3f &projection, Matrix3f &uvn) = 0;
   
   ~NIEManifold() {}
};


/// abstract class that describes a manifold embedded in 3-space
class NIEPointSetManifold : public NIEManifold {
	X4_CLASS( NIEPointSetManifold )
 private:
   PointSet *ps;
   AAT positionAAT;
   InCorePCTopologyGraph *tg;
   float32 sampleSpacing;
   float32 mlsRadiusFact;
   PointSetANNQuery *knn;
   TopNeighborhoodSphereSearch *topSearch;
   
 protected:
   void computeMLS(const Vector3f &point, const int32 index, Matrix3f &uvn, Vector3f &proj);
   
 public:
   NIEPointSetManifold();
	virtual void assign(const Object* obj, X4CopyContext *context = NULL);
   
   virtual void clearAndSetup(const PointSet *ps, const InCorePCTopologyGraph *tg, 
                              float32 sampleSpacing, float32 mlsRadiusFact = 3);
   
   //virtual void sampleSurface(float32 epsilon, std::vector<Vector3f> &result);
   virtual Vector3f project(const Vector3f &point);
   virtual void getTangentSystem(const Vector3f &point, Vector3f &projection, Matrix3f &uvn);
   
   mpcard getNumPts() {return ps->getNumEntries();}
   Vector3f getPoint(mpcard i) {return ps->get3f(i, positionAAT);}
   void setPoint(mpcard i, const Vector3f &pos) {ps->set3f(i, positionAAT, pos);}
   
   mpcard getNumAdjacent(mpcard i) {return tg->getNumAdjacentVertices(i);}
   mpcard getOverallNumEdges();
   mpcard getAdjacentIndex(mpcard v, mpcard i) {return tg->getAdjacentVertex(v,i);}
   
   ~NIEPointSetManifold();
};


/// Near-Isometric Surface Embedding
class X3DEFORMATION_API NearIsometricEmbedding : public Persistent {
	X4_CLASS( NearIsometricEmbedding )
 private:
   NIEPointSetManifold *sourceShape;
   NIEManifold *targetShape;
   NIEPointSetManifold *embedding;
   
   struct Constr {
      Vector3f target;
      float32 weight;

      Constr(Vector3f v, float32 w) {target = v; weight = w;}
      Constr() {target = NULL_VECTOR3F; weight = 0;}
   };
   
   std::vector<Constr> constraints;
 
 public:
	NearIsometricEmbedding();
	virtual void assign(const Object* obj, X4CopyContext *context = NULL);

   /// stored as references (not deleted with this object)
	virtual void clearAndSetup(NIEPointSetManifold *sourceShape, NIEManifold *targetShape, NIEPointSetManifold *embedding);
	NIEPointSetManifold *getCurrentEmbedding() {return embedding;}
	
	void addConstraint(const card32 embIndex, const Vector3f &pos, const float32 weight);
	
	/// returns max-norm of differences of embedding before/after
	void solveStep();
	
	~NearIsometricEmbedding();
};


}

#endif
