//---------------------------------------------------------------------------
#include "stdafx.h"
//---------------------------------------------------------------------------
#include "PCCBoundaryDetector.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "PointSetANNQuery.h"
#include "PCA.h"
#include "PointCloudTools.h"
#include "SeparatorClassProperty.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

IMPLEMENT_X4_CLASS(PCCBoundaryDetector, 3)
{
	BEGIN_CLASS_INIT(PCCBoundaryDetector)
	ADD_SEPARATOR("find point cloud boundaries")
   ADD_CARD32_PROP(numNeighbors,0)
   ADD_FLOAT32_PROP(sigmaRel, 0)
   ADD_FLOAT32_PROP(quantizeThreshold, 1)
   ADD_BOOLEAN_PROP(quantize, 1)
   ADD_BOOLEAN_PROP(boundaryTangents, 1)
	ADD_SEPARATOR("find attribute boundaries")
	ADD_BOOLEAN_PROP(attributeBoundaries, 2)
	ADD_STRING_PROP(boundaryChannel, 2)
   ADD_BOOLEAN_PROP(boundaryTangents, 1)
	ADD_SEPARATOR("additional output")
	ADD_BOOLEAN_PROP(createBoundaryPC, 3)
	ADD_STRING_PROP(boundaryPCName, 3)
	ADD_FLOAT32_PROP(normalOffset, 3)
}

//---------------------------------------------------------------------------

PCCBoundaryDetector::PCCBoundaryDetector() {
   numNeighbors = 20;
   sigmaRel = 1;
   quantizeThreshold = 0.75f;
   quantize = true;
   boundaryTangents = true;
	attributeBoundaries = false;
	boundaryChannel = "color";
	createBoundaryPC = false;
	boundaryPCName = "boundaries";
	normalOffset = 0.01f;
}

PCCBoundaryDetector::~PCCBoundaryDetector() {}

void PCCBoundaryDetector::assign(const Object* obj, X4CopyContext *context)
{
	copyObjectProperties(obj, this);
}

//---------------------------------------------------------------------------

struct BoundaryPCVertex {
	Vector3f position;
	Vector3f normal;
	Vector3f tangent;
	int32 flags;
	int buffer[32]; // unused, you never know...
};

void PCCBoundaryDetector::detectBoundaries(PointCloud *pc, float32 sr, float32 qt, card32 nn) {
   sigmaRel = sr;
   quantizeThreshold = qt;
   numNeighbors = nn;
   if (!pc->providesAttribute("position")) {
       warning("no position channel - selected pc ignored.");
       return;
   }
   if (!pc->providesAttribute("normal")) {
       warning("no normal channel - selected pc ignored.");
       return;
   }
   checkAttribute(pc, "boundary", 1, VAD::DATA_FORMAT_FLOAT32);
   if (boundaryTangents) {
       checkAttribute(pc, "boundaryTangent", 3, VAD::DATA_FORMAT_FLOAT32);
   }


   SceneObjectIterator * it = pc->createIterator(SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC);
   if (!it) return;
   BasicPointCloudIterator *bit = dynamic_cast<BasicPointCloudIterator*>(it);
   ModifyablePointCloudIterator *mit = dynamic_cast<ModifyablePointCloudIterator*>(it);
   PointSet *ps = mit->lockBlockReadWrite();
   AAT positionAAT = ps->getAAT("position", 3, VAD::DATA_FORMAT_FLOAT32);
   AAT normalAAT = ps->getAAT("normal", 3, VAD::DATA_FORMAT_FLOAT32);
   AAT boundaryAAT = ps->getAAT("boundary", 1, VAD::DATA_FORMAT_FLOAT32);
	
	AAT attributeBoundaryAAT = NULL_AAT;
	VAD attributBoundariesVAD;
	if (attributeBoundaries) {
		 attributeBoundaryAAT = ps->getAAT(boundaryChannel);
		 if (attributeBoundaryAAT == NULL_AAT) throw X4Exception("PCCBoundaryDetector::detectBoundaries() - boundaryChannel not found (attributeBoundaries mode is set)");
		 attributBoundariesVAD = ps->getDescr()->getAttribute(boundaryChannel);
		 if (attributBoundariesVAD.numComponents >4) throw X4Exception("PCCBoundaryDetector::detectBoundaries() - boundaryChannel dimension bigger than 4 - this is currently not supported (attributeBoundaries mode is set)");
		 card32 dataFormat = attributBoundariesVAD.dataFormat ;
		 if (dataFormat != VAD::DATA_FORMAT_CARD8 && dataFormat != VAD::DATA_FORMAT_INT32 && dataFormat != VAD::DATA_FORMAT_FLOAT32)
			 throw X4Exception("PCCBoundaryDetector::detectBoundaries() - boundaryChannel data format is neither card8,int32, nor float32 - this is currently not supported (attributeBoundaries mode is set)");
	}

	PointSetANNQuery query(ps, numNeighbors);
   mpcard numPts = ps->getNumEntries();
   for (mpcard i=0; i<numPts; i++) {
      Vector3f pos = ps->get3f(i, positionAAT);
      //Vector3f tangent_u = ps->get3f(i, tangent_uAAT);
      //Vector3f tangent_v = ps->get3f(i, tangent_vAAT);
      Vector3f normal = ps->get3f(i, normalAAT);
      Matrix3f tangentFrame = calcTangentSystem(normal);
      const Vector3f tangent_u = tangentFrame[0];
      const Vector3f tangent_v = tangentFrame[1];
		Vector4d attributeValue = NULL_VECTOR4D;
		if (attributeBoundaries) {
			  attributeValue = attributBoundariesVAD.getAnyAttrib((card8*)ps->getDataPointer(i), attributeBoundaryAAT);
		}

      query.setTargetPoint(pos);
      card32 k;
      int32 *buf = query.getAllIndices(k);
      std::vector<Vector2f> buf2D(k);
		card32 actualK = 0;
		bool homogeneousAttributesOverrideBoundaries = attributeBoundaries; // this flag is used to avoid detecting just boundaries of homogeneous area (when no other attributes are present)
      for (card32 j=0; j<k; j++) {
			if (attributeBoundaries) {
				if (attributeValue != attributBoundariesVAD.getAnyAttrib((card8*)ps->getDataPointer(buf[j]), attributeBoundaryAAT)) {
					homogeneousAttributesOverrideBoundaries = false;
					continue;
				}
			}
         Vector3f d = pos - ps->get3f(buf[j], positionAAT);
			buf2D[j] = makeVector2f(d*tangent_u, d*tangent_v);
				actualK++;
       }
		  if (attributeBoundaries) {
				buf2D.resize(actualK);
				k = actualK;
		  }

       float32 avDist = 0;
       for (card32 j=0; j<k; j++) {
           avDist += norm(buf2D[j]);
       }
       avDist /= k;
       static const card32 NUM_DISCR = 24;
       float32 minMaxR = 1e20f;
       for (card32 r=0; r<NUM_DISCR; r++) {
           float32 alpha = (float32)r/(float32)NUM_DISCR*M_PI*2;
           Vector2f rv = makeVector2f(sin(alpha), cos(alpha));
           float32 maxR = -1e20f;
           for (card32 j=0; j<k; j++) {
               float32 rd = rv*buf2D[j];
               if (maxR < rd) maxR = rd;
           }
           if (maxR < minMaxR) minMaxR = maxR;
       }
       float32 value = exp(-minMaxR*minMaxR/(avDist*avDist*sigmaRel*sigmaRel));
		 if (homogeneousAttributesOverrideBoundaries) value = 0;
       ps->set1f(i, boundaryAAT, value);
   }
   if (boundaryTangents) {
        AAT boundaryTangentsAAT = ps->getAAT("boundaryTangent", 3, VAD::DATA_FORMAT_FLOAT32);
        for (mpcard i=0; i<numPts; i++) {
            Vector3f pos = ps->get3f(i, positionAAT);
            Vector3f normal = ps->get3f(i, normalAAT);
            Matrix3f tangentFrame = calcTangentSystem(normal);
            const Vector3f tangent_u = tangentFrame[0];
            const Vector3f tangent_v = tangentFrame[1];
            query.setTargetPoint(pos);
            card32 k;
            int32 *buf = query.getAllIndices(k);
            PCA<float32, 2> pca;
            card32 ptsInPCA = 0;
            for (card32 j=0; j<k; j++) {
                Vector3f d = pos - ps->get3f(buf[j], positionAAT);
                Vector2f d2 = makeVector2f(d*tangent_u, d*tangent_v);
                float32 b = ps->get1f(buf[j], boundaryAAT);
                pca.addPoint(d2, b); 
                if (b > 0.01) ptsInPCA++;
            }
            Vector2f center, values;
            Matrix2f vects;
            if (ptsInPCA >= 2) {
                pca.analyze(values, vects, center);
                ps->set3f(i, boundaryTangentsAAT, tangent_u*vects[0][0] + tangent_v*vects[0][1]);
            } else {
                ps->set3f(i, boundaryTangentsAAT, NULL_VECTOR3F);
            }
        }
   }
   if (quantize) {
       for (mpcard i=0; i<numPts; i++) {
          float32 value = ps->get1f(i, boundaryAAT);
          if (value > quantizeThreshold) {
             value = 1;
          } else {
             value = 0;
          }
          ps->set1f(i, boundaryAAT, value);
       }
   }

	if (createBoundaryPC) {
		AAT boundaryTangentsAAT = NULL_AAT;
		if (boundaryTangents) boundaryTangentsAAT = ps->getAAT("boundaryTangent", 3, VAD::DATA_FORMAT_FLOAT32);
		UnstructuredInCorePointCloud *extraBoundariesPC = NULL;
		extraBoundariesPC = new UnstructuredInCorePointCloud;
		VertexDescriptor vd;
		vd.pushAttrib(mVAD("position",3, VAD::DATA_FORMAT_FLOAT32));
		vd.pushAttrib(mVAD("normal",3, VAD::DATA_FORMAT_FLOAT32));
		vd.pushAttrib(mVAD("tangent",3, VAD::DATA_FORMAT_FLOAT32));
		vd.pushAttrib(mVAD("flags",1, VAD::DATA_FORMAT_INT32));
		extraBoundariesPC->clearAndSetup(&vd, 0);
		AAT extraPositionAAT =vd.getToken("position");
		AAT extraNormalAAT = vd.getToken("normal");
		AAT extraTangentAAT = vd.getToken("tangent");

		mpcard numPts = ps->getNumEntries();
		 for (mpcard i=0; i<numPts; i++) {
			 Vector3f pos = ps->get3f(i, positionAAT);
			 Vector3f normal = ps->get3f(i, normalAAT);
			 Vector3f boundaryTangent = NULL_VECTOR3F;
			 float32 b = ps->get1f(i, boundaryAAT);
			 if (b < 0.01) continue;
			 if (boundaryTangents) boundaryTangent = ps->get3f(i, boundaryTangentsAAT);
			 Vector3f newPos = pos+normal*normalOffset;
			 BoundaryPCVertex newPoint;
			 newPoint.position = newPos;
			 newPoint.normal = normal;
			 newPoint.tangent = boundaryTangent;
			 newPoint.flags = 0;
			 extraBoundariesPC->getPointSet()->addPoint((char*)&newPoint);
		 }
		 addPointCloud(getScene(), extraBoundariesPC, boundaryPCName);
	 }

    bit->unlock();
    delete it;
}

void PCCBoundaryDetector::execute(Scene *scene) 
{
   std::vector<std::string> selection = scene->getRootState()->staticState->getSelectedNodes();
   for (size_t m=0; m<selection.size(); m++) {
      PointCloud *pc = getPointCloud(scene, selection[m]);
      detectBoundaries(pc);
   }
}
