//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "Handle3DHandler.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "Handle3D.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

static const float32 CLICK_RADIUS = 12.0f;


IMPLEMENT_X4_CLASS( Handle3DHandler , 0 ) {
	BEGIN_CLASS_INIT( Handle3DHandler );
}

// ----

Handle3DHandler::Handle3DHandler() {
	handles = NULL;
	width = 1;
	height = 1;
	selection = false;
	selected_h = 0;
	selected_p = 0;
	last_x = 0;
	last_y = 0;
	tanOfHalfViewingAngle = 1;
}

void Handle3DHandler::assign(const Object* obj, X4CopyContext *context) {
	copyObjectProperties(obj, this);
}

Handle3DHandler::~Handle3DHandler() {
}




void Handle3DHandler::setParams(const Camera &cam, const ViewFrustum &vf, const Matrix4f &baseTransform) 
{
	Matrix4f viewMatrix = cam.getViewMatrix();
	Matrix4f projMatrix = vf.getProjectionMatrix();
	mvMatrix = projMatrix * viewMatrix * baseTransform;
	tanOfHalfViewingAngle = tan(M_PI*vf.getVerticalFieldOfView()/180.0f/2.0f);
	camViewDir = cam.getOrthoNormViewDirection();
	camRightDir = cam.getOrthoNormRightDirection();
	camUpDir = cam.getOrthoNormUpDirection();
	camPos = cam.getPosition();
}

void Handle3DHandler::setHandles(std::vector<Handle3D*> *handles) 
{
	this->handles = handles;
}

void Handle3DHandler::mouseDown(int32 x, int32 y, MouseButtons buttonsState, Modifiers modifiersState) 
{
	if (!handles) return;
	if (buttonsState.getLeft()) {
		unsigned h;
		card32 p;
		if (pick(x,y, h, p)) {
			selection = true;
			selected_h = h;
			selected_p = p;
			last_x = x;
			last_y = y;
			last_modifiers = modifiersState;
		}	
	}
}

bool Handle3DHandler::selected() {
   return selection;
}

void Handle3DHandler::getSelection(bool &selected, card32 &handle, card32 &point) {
   selected = selection;
   handle = selected_h;
   point = selected_p;
}

void Handle3DHandler::deselect() {
	selection = false;
	selected_h = 0;
	selected_p = 0;
}



void Handle3DHandler::mouseMoved(int32 x, int32 y) 
{
	for (unsigned h=0; h<handles->size(); h++) {
		(*handles)[h]->clearHighlights();
	}
	if (!handles) return;
	if (!selection) {
		unsigned h;
		card32 p;
		if (pick(x,y, h, p)) {
			(*handles)[h]->hoverPoint(p);
		}
	} else {
	   if (selected_h < handles->size()) {
		   (*handles)[selected_h]->hoverPoint(selected_p);
		   Vector3f delta = -reproject(x-last_x, y-last_y, last_prj_depth);
		   Vector3f newPos = (*handles)[selected_h]->getPoint(selected_p) + delta;
		   (*handles)[selected_h]->movePoint(selected_p, newPos, last_modifiers);
		   for (unsigned l=0; l<listeners.size(); l++) {
			   listeners[l]->handleChanged(selected_h, selected_p);
		   }
		   last_x = x;
		   last_y = y;
      }
	}
}

void Handle3DHandler::mouseUp(int32 x, int32 y, MouseButtons buttonsState, Modifiers modifiersState) 
{
	if (!handles) return;
	if (buttonsState.getLeft()) {
		selection = false;
	}
}

void Handle3DHandler::areaResize(card32 width, card32 height) 
{
	this->width = width;
	this->height = height;
}

void Handle3DHandler::keyDown(GeneralKey key) 
{
	if (!handles) return;
}

void Handle3DHandler::keyUp(GeneralKey key) 
{
	if (!handles) return;
}


float32 Handle3DHandler::rayQuery(Vector3f pos, int32 x, int32 y, float32 &depth) 
{
	Vector3f proj = projectiveTransformVector3f(mvMatrix, pos);
	Vector2f click;
	click[0] = ((float32)x)/(width*0.5f)-1.0f;
	click[1] = ((float32)y)/(height*0.5f)-1.0f;
	Vector2f proj2 = makeVector2f(proj[0], proj[1]);
	depth = (pos-camPos)*camViewDir;
	return norm(click-proj2)*height*0.5f;
}

bool Handle3DHandler::pick(int32 x, int32 y, unsigned &handle, card32 &point) {
	if (!handles) return false;
	bool result = false;
	float32 lastDepth = 1E30f;
	for (unsigned h=0; h<handles->size(); h++) {
		Handle3D *currentHandle = (*handles)[h];
		card32 n = currentHandle->getNumPoints();
		for (card32 i=0; i<n; i++) {
		   if (!currentHandle->getPointMovable(i)) continue;
           if (!currentHandle->getPointVisible(i)) continue;
			Vector3f p = currentHandle->getPoint(i);
			float32 depth;
			float32 dist = rayQuery(p,x,y, depth);
			if (dist <= CLICK_RADIUS*0.5f) {
				if (!result) {
					handle = h;
					point = i;
					lastDepth = depth;
					result = true;
				} else {
					if (depth < lastDepth) {
						handle = h;
						point = i;
						lastDepth = depth;
					}
				}
			}
		}
	}
	if (result) last_prj_depth = lastDepth;
	return result;
}


Vector3f Handle3DHandler::reproject(float32 x, float32 y, float32 depth) 
{
	float32 scale = depth*tanOfHalfViewingAngle/(((float32)height)/2.0f);
	x *= scale;
	y *= scale;
	return camRightDir * x + camUpDir * y;
}

void Handle3DHandler::addChangesListener(Handle3DUpdatesListener *x) 
{
	listeners.push_back(x);
}

void Handle3DHandler::manuallySelectLatest(int x, int y) 
{
   if (!handles->empty()) {
      if ((*handles)[handles->size()-1]->getNumPoints() > 0) {
         selection = true;
         selected_h = (card32)handles->size()-1;
         selected_p = (*handles)[handles->size()-1]->getNumPoints()-1;
         last_x = x;
         last_y = y;
         last_modifiers = Modifiers();
      }
   }
}

unsigned Handle3DHandler::getLastSelectedHandle() 
{
   return selected_h;
}
