//---------------------------------------------------------------------------
#include "stdafx.h"
//---------------------------------------------------------------------------
#include "CmdObjGeneratePC.h"
#include "FixedArrayClassProperty.h"
#include "StringClassProperty.h"
#include "ObjectClassProperty.h"
#include "PCCommand.h"
#include "EmbeddedObjectClassProperty.h"
#include "BooleanClassProperty.h"
#include "DebugOutput.h"
#include "Random.h"
#include "PolarCoordinates.h"
#include "NormalDistribution.h"
#include "PointCloud.h"
#include "PointCloudTools.h"
#include "PointSet.h"
#include "PointCloudIterators.h"
#include "UnstructuredInCorePointCloud.h"
#include "SGObjectNode.h"
#include "StandardSceneGraphIterator.h"
#include "SGListNode.h"
#include "PointSet.inline.h"
#include "MathTools.h"
#include "MathTools.inline.h"
//---------------------------------------------------------------------------
#include <algorithm>
 
using namespace X4;

IMPLEMENT_X4_CLASS( CmdObjGeneratePC, 0 )
{
	 BEGIN_CLASS_INIT( CmdObjGeneratePC );
	 ADD_STRING_PROP( pcName, 0 );
	 ADD_CARD32_PROP( numPoints, 0 )
	 ADD_CARD32_PROP( type, 0 )
}

/**
 * Constructor
 */
CmdObjGeneratePC::CmdObjGeneratePC()
{
	type = GENERATE_MERLONS;
	numPoints = 3;
	pcName = "pc1";
}

void CmdObjGeneratePC::doAction(Object *context)
{
   Scene *scene = dynamic_cast<Scene*>(context); x4Assert(scene);
   PointSet* ps = NULL;

   if ( type == GENERATE_DOWNWARD_SPIRAL ){
		ps = createDownwardSpiral( numPoints );  
   } else if ( type == GENERATE_SPHERE ){
		ps = createSphere( numPoints );
   } else if ( type == GENERATE_BOX ){
	   ps = createBox( numPoints );
   } else if ( type == GENERATE_PLANE ){
	   ps = createPlane( numPoints );
	} else if ( type == GENERATE_PARABEL ){
		ps = createParabel( numPoints );
   } else if ( type == GENERATE_SINUS ){
	   ps = createSinusPlane( numPoints );
   } else if ( type == GENERATE_CYLINDER ){
	   ps = createCylinder( numPoints );
   } else if ( type == GENERATE_MERLONS ){
	   ps = createMerlons( numPoints );
   }   else{
	   debugOutput << "CmdObjGeneratePC::doAction: invalid object selected!\n";
	   debugOutput.flush();
   }

   if ( ps != NULL ){
	   UnstructuredInCorePointCloud* pc = new UnstructuredInCorePointCloud();
	   pc->setPointSet( ps );
	   addPointCloud( scene, pc, pcName );
   }
}

void CmdObjGeneratePC::undoAction(Object *context)
{
   forget(context);
   throw ECannotUndo( "CmdObjGeneratePC does not provide any undo." );
}

bool CmdObjGeneratePC::canUndo() {
   return false;
}


std::string CmdObjGeneratePC::getDescription()
{
   return "Create a new point cloud!";
}

/// Setup object
void CmdObjGeneratePC::setup( unsigned int type, unsigned int numPoints, std::string pcName )
{
	this->type = type;
	this->numPoints = numPoints;
	this->pcName = pcName;
}

/**
* Create downward spiral
*/
PointSet* CmdObjGeneratePC::createDownwardSpiral( const card32& numPoints )
{
	std::vector<Vector3f> points;
	for ( card32 i = 0; i < numPoints; i++ ){
		const float32 T_MAX = 10.0f*(2.0f*(float32)M_PI);
		float32 t =  T_MAX * (sqrt((float32)(i))/sqrt((float32)numPoints));
		Vector3f pos = makeVector3f ( (t/T_MAX) * sin(t),
			(t/T_MAX) * cos(t),
			sqr(t/T_MAX)*T_MAX/30.0f );
		Vector3f tangent = normalize(
			makeVector3f (
			(1/T_MAX) * sin(t) + (t/T_MAX) * cos(t),
			(1/T_MAX) * cos(t) - (t/T_MAX) * sin(t),
			2.0f/T_MAX*t/30.0f ) );
		///1.0f/(sqrt(t/T_MAX)+0.0001f)*1.0/30.0f ) );
		Vector3f n1 = normalize(makeOrthonormalVector(tangent));
		Vector3f n2 = tangent.crossProduct(n1);
		float32 w = rnd01() * 2.0f * M_PI;
		float32 l = 0.1 * (t/T_MAX);
		pos +=   n1 * (l * sin(w))
			+ n2 * (l * cos(w));
		points.push_back( pos );
	}
	VertexDescriptor vd;
	vd.pushAttrib( mVAD( "position", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	vd.pushAttrib( mVAD( "flags", 1, VAD::DATA_FORMAT_INT32 ) );
	vd.pushAttrib( mVAD( "normal", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	vd.pushAttrib( mVAD( "color", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	PointSet* ps = new PointSet();
	ps->clearAndSetup( 1, points.size(), &vd );
	AAT POSITION = ps->getAAT("position");
	AAT NORMAL = ps->getAAT("normal");
	AAT COLOR = ps->getAAT( "color" );
	for ( card32 pointCounter = 0; pointCounter < points.size(); pointCounter++ ){
		Vector3f p = points[pointCounter];
		ps->set3f( pointCounter, POSITION, p );
		ps->set3f( pointCounter, NORMAL, NULL_VECTOR3F );
		ps->set3f( pointCounter, COLOR, makeVector3f( 0.5f,0.5f,0.5f ) );
	}
	return ps;
}

/**
* Create plane
*/
PointSet* CmdObjGeneratePC::createPlane( const card32& numPoints )
{
	VertexDescriptor vd;
	vd.pushAttrib( mVAD( "position", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	vd.pushAttrib( mVAD( "flags", 1, VAD::DATA_FORMAT_INT32 ) );
	vd.pushAttrib( mVAD( "normal", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	vd.pushAttrib( mVAD( "color", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	std::vector<Vector3f> points, normals;
	card32 sqrtNumPoints = sqrt( float32( numPoints ) );
	points.clear();
	normals.clear();
	for ( unsigned int x = 0; x <= sqrtNumPoints; x++ ){
		for ( unsigned int z = 0; z <= sqrtNumPoints; z++ ){
			points.push_back( makeVector3f( x / float32( sqrtNumPoints ) - 0.5f, 0, z / float32( sqrtNumPoints ) - 0.5f ) );
			normals.push_back( makeVector3f( 0,1,0 ) ); 
		}
	}
	PointSet* ps = new PointSet();
	ps->clearAndSetup( 1, points.size(), &vd );
	AAT POSITION = ps->getAAT( "position" );
	AAT NORMAL = ps->getAAT( "normal" );
	AAT COLOR = ps->getAAT( "color" );
	for ( unsigned int pointIndex = 0; pointIndex < points.size(); pointIndex++ ){
		ps->set3f( pointIndex, POSITION, points[pointIndex] );
		ps->set3f( pointIndex, NORMAL, normals[pointIndex] );		
		ps->set3f( pointIndex, COLOR, makeVector3f( 0.5f,0.5f,0.5f ) );
	}
	return ps;
}

/**
* Create sphere
*/
PointSet* CmdObjGeneratePC::createSphere( const card32& numPoints )
{
	VertexDescriptor vd;
	vd.pushAttrib( mVAD( "position", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	vd.pushAttrib( mVAD( "flags", 1, VAD::DATA_FORMAT_INT32 ) );
	vd.pushAttrib( mVAD( "normal", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	vd.pushAttrib( mVAD( "color", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	std::vector<Vector3f> points, normals;
	float32 alpha, beta, radius, height;
	points.clear();
	normals.clear();
	unsigned int i, j, pointsOnCircle, NUM_HORIZONTAL_CIRLES;
	Vector3f p;
	// Approximate number of points
	NUM_HORIZONTAL_CIRLES = card32( sqrt( (float32)numPoints ) * 1.25f );
	for ( i = 0; i < NUM_HORIZONTAL_CIRLES; i++ ){
		alpha = (float)i/(float)NUM_HORIZONTAL_CIRLES * float32( M_PI );
		radius = sin( alpha );
		height = cos( alpha );
		pointsOnCircle = std::max( 1u, card32( radius * NUM_HORIZONTAL_CIRLES ) );
		if ( radius < 1e-6 ) pointsOnCircle = 1;
		for ( j = 0; j < pointsOnCircle; j++ ){
			beta = (float)j/(float)pointsOnCircle * 2.0f * float32( M_PI );
			p = makeVector3f( radius * sin( beta ), height, radius * cos( beta ) ) * 0.5f;
			points.push_back( p );
			normals.push_back( normalize( p ) ); 
		}
	}		
	PointSet* ps = new PointSet();
	ps->clearAndSetup( 1, points.size(), &vd );
	AAT POSITION = ps->getAAT( "position" );
	AAT NORMAL = ps->getAAT( "normal" );
	AAT COLOR = ps->getAAT( "color" );
	for ( unsigned int pointIndex = 0; pointIndex < points.size(); pointIndex++ ){
		ps->set3f( pointIndex, POSITION, points[pointIndex] );
		ps->set3f( pointIndex, NORMAL, normals[pointIndex] );		
		ps->set3f( pointIndex, COLOR, makeVector3f( 0.5f,0.5f,0.5f ) );
	}
	return ps;
}

/**
* Create box
*/
PointSet* CmdObjGeneratePC::createBox( const card32& numPoints )
{
	VertexDescriptor vd;
	vd.pushAttrib( mVAD( "position", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	vd.pushAttrib( mVAD( "flags", 1, VAD::DATA_FORMAT_INT32 ) );
	vd.pushAttrib( mVAD( "normal", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	vd.pushAttrib( mVAD( "color", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	unsigned int RESOLUTION, i, j, n;
	Vector3f origin[6], dx[6], dy[6], p, normal[6];
	std::vector<Vector3f> points;
	std::vector<Vector3f> normals;
	RESOLUTION = card32( sqrt( (float32)numPoints / 6.0f ) );
	// Create box
	points.clear();
	normals.clear();
	// Front
	origin[0] = makeVector3f( -0.5,-0.5,-0.5 );
	dx[0] = makeVector3f( 1,0,0 );
	dy[0] = makeVector3f( 0,1,0 );
	normal[0] = makeVector3f( 0,0,-1 );
	// Back
	origin[1] = makeVector3f( 0.5,-0.5,0.5 );
	dx[1] = makeVector3f( -1,0,0 );
	dy[1] = makeVector3f( 0,1,0 );
	normal[1] = makeVector3f( 0,0,1 );
	// Left
	origin[2] = makeVector3f( 0.5,-0.5,-0.5 );
	dx[2] = makeVector3f( 0,0,1 );
	dy[2] = makeVector3f( 0,1,0 );
	normal[2] = makeVector3f( 1,0,0 );
	// Left
	origin[3] = makeVector3f( -0.5,-0.5,0.5 );
	dx[3] = makeVector3f( 0,0,-1 );
	dy[3] = makeVector3f( 0,1,0 );
	normal[3] = makeVector3f( -1,0,0 );
	// Bottom
	origin[4] = makeVector3f( -0.5,-0.5,-0.5 );
	dx[4] = makeVector3f( 1,0,0 );
	dy[4] = makeVector3f( 0,0,1 );
	normal[4] = makeVector3f( 0,-1,0 );
	// Bottom
	origin[5] = makeVector3f( -0.5,0.5,-0.5 );
	dx[5] = makeVector3f( 1,0,0 );
	dy[5] = makeVector3f( 0,0,1 );
	normal[5] = makeVector3f( 0,1,0 );
	for ( n = 0; n < 6; n++ ){
		for ( i = 0; i < RESOLUTION; i++ ){
			for ( j = 0; j < RESOLUTION; j++ ){
				p = origin[n] +
					dx[n] * ((float32)i+0.5f)/(float32)RESOLUTION +
					dy[n] * ((float32)j+0.5f)/(float32)RESOLUTION;
				points.push_back( p );
				normals.push_back( normalize( normal[n] ) );
			}
		}
	}
	// Create iterators
	PointSet* ps = new PointSet();
	ps->clearAndSetup( 1, points.size(), &vd );
	AAT POSITION = ps->getAAT( "position" );
	AAT NORMAL = ps->getAAT( "normal" );
	AAT COLOR = ps->getAAT( "color" );
	for ( unsigned int pointIndex = 0; pointIndex < points.size(); pointIndex++ ){
		ps->set3f( pointIndex, POSITION, points[pointIndex] );
		ps->set3f( pointIndex, NORMAL, normals[pointIndex] );
		ps->set3f( pointIndex, COLOR, makeVector3f( 0.5f,0.5f,0.5f ) );
	}
	return ps;
}

PointSet* X4::CmdObjGeneratePC::createSinusPlane( const card32& numPoints )
{
	VertexDescriptor vd;
	vd.pushAttrib( mVAD( "position", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	vd.pushAttrib( mVAD( "flags", 1, VAD::DATA_FORMAT_INT32 ) );
	vd.pushAttrib( mVAD( "normal", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	vd.pushAttrib( mVAD( "color", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	std::vector<Vector3f> points, normals;
	card32 sqrtNumPoints = sqrt( float32( numPoints ) );
	points.clear();
	normals.clear();
	
	float32 xForSin=0;
	float32 multiplier= (10.0*M_PI_4)/(float32)sqrtNumPoints;
	
	for ( unsigned int x = 0; x <= sqrtNumPoints; x++ ){
		
		for ( unsigned int z = 0; z <= sqrtNumPoints; z++ ){
			points.push_back( makeVector3f( xForSin, sin(xForSin), float32(z) * multiplier ) * (1/(10.0f*(float32)M_PI_4)));
			float32 derivated= cos(xForSin);
			Vector3f normal= makeVector3f( 1,derivated,0 ).crossProduct(makeVector3f(0,0,1));
			if (normal*makeVector3f(0,1,0)<0)
				normal*= -1.0f;
			
			normals.push_back( normalize( normal )); 
		}
		xForSin += multiplier * 1/(sqrt(1+cos(xForSin)*cos(xForSin)));
	}

	PointSet* ps = new PointSet();
	ps->clearAndSetup( 1, points.size(), &vd );
	AAT POSITION = ps->getAAT( "position" );
	AAT NORMAL = ps->getAAT( "normal" );
	AAT COLOR = ps->getAAT( "color" );
	for ( unsigned int pointIndex = 0; pointIndex < points.size(); pointIndex++ ){
		ps->set3f( pointIndex, POSITION, points[pointIndex] );
		ps->set3f( pointIndex, NORMAL, normals[pointIndex] );		
		ps->set3f( pointIndex, COLOR, makeVector3f( 0.5f,0.5f,0.5f ) );
	}
	return ps;
}

PointSet* X4::CmdObjGeneratePC::createParabel( const card32& numPoints )
{
	VertexDescriptor vd;
	vd.pushAttrib( mVAD( "position", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	vd.pushAttrib( mVAD( "flags", 1, VAD::DATA_FORMAT_INT32 ) );
	vd.pushAttrib( mVAD( "normal", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	vd.pushAttrib( mVAD( "color", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	std::vector<Vector3f> points, normals;
	card32 sqrtNumPoints = sqrt( float32( numPoints ) );
	points.clear();
	normals.clear();
	for ( unsigned int x = 0; x <= sqrtNumPoints; x++ ){
		for ( unsigned int z = 0; z <= sqrtNumPoints; z++ ){
			float fx = x / float32( sqrtNumPoints ) - 0.5f;
			points.push_back( makeVector3f( fx, pow(2.0f * fx,2.0f), z / float32( sqrtNumPoints ) - 0.5f ) );
			Vector3f tangentX = makeVector3f(1,4.0f * x,0);
			normals.push_back( normalize( tangentX.crossProduct(makeVector3f(0,1,0))) ); 
		}
	}
	PointSet* ps = new PointSet();
	ps->clearAndSetup( 1, points.size(), &vd );
	AAT POSITION = ps->getAAT( "position" );
	AAT NORMAL = ps->getAAT( "normal" );
	AAT COLOR = ps->getAAT( "color" );
	for ( unsigned int pointIndex = 0; pointIndex < points.size(); pointIndex++ ){
		ps->set3f( pointIndex, POSITION, points[pointIndex] );
		ps->set3f( pointIndex, NORMAL, normals[pointIndex] );		
		ps->set3f( pointIndex, COLOR, makeVector3f( 0.5f,0.5f,0.5f ) );
	}
	return ps;
}

PointSet* X4::CmdObjGeneratePC::createCylinder( const card32& numPoints )
{
	VertexDescriptor vd;
	vd.pushAttrib( mVAD( "position", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	vd.pushAttrib( mVAD( "flags", 1, VAD::DATA_FORMAT_INT32 ) );
	vd.pushAttrib( mVAD( "normal", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	vd.pushAttrib( mVAD( "color", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	std::vector<Vector3f> points, normals;
	float32 alpha, beta, radius, height=0;
	points.clear();
	normals.clear();
	unsigned int i, j, pointsOnCircle, NUM_HORIZONTAL_CIRLES;
	Vector3f p;
	// Approximate number of points
	NUM_HORIZONTAL_CIRLES = card32( sqrt( (float32)numPoints ) * 1.25f );
	float32 hightAdd= 1/(float32)NUM_HORIZONTAL_CIRLES;
	radius = 1; 
	for ( i = 0; i < NUM_HORIZONTAL_CIRLES; i++ ){
		alpha = (float)i/(float)NUM_HORIZONTAL_CIRLES * float32( M_PI );
		height += hightAdd;
		pointsOnCircle = 100 ;
		for ( j = 0; j < pointsOnCircle; j++ ){
			beta = (float)j/(float)pointsOnCircle * 2.0f * float32( M_PI );
			p = makeVector3f( radius * sin( beta ), height, radius * cos( beta ) ) * 0.5f;
			points.push_back( p );
			normals.push_back( normalize( makeVector3f( radius * sin( beta ), 0, radius * cos( beta ) ) ) ); 
		}
	}		
	PointSet* ps = new PointSet();
	ps->clearAndSetup( 1, points.size(), &vd );
	AAT POSITION = ps->getAAT( "position" );
	AAT NORMAL = ps->getAAT( "normal" );
	AAT COLOR = ps->getAAT( "color" );
	for ( unsigned int pointIndex = 0; pointIndex < points.size(); pointIndex++ ){
		ps->set3f( pointIndex, POSITION, points[pointIndex] );
		ps->set3f( pointIndex, NORMAL, normals[pointIndex] );		
		ps->set3f( pointIndex, COLOR, makeVector3f( 0.5f,0.5f,0.5f ) );
	}
	return ps;
}

PointSet* X4::CmdObjGeneratePC::createMerlons( const card32& numMerlons )
{
	if (numMerlons<1)
		return new PointSet();
	

	
	VertexDescriptor vd;
	vd.pushAttrib( mVAD( "position", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	vd.pushAttrib( mVAD( "flags", 1, VAD::DATA_FORMAT_INT32 ) );
	vd.pushAttrib( mVAD( "normal", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	vd.pushAttrib( mVAD( "color", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	unsigned int RESOLUTION, i, j, n;
	Vector3f origin[6], dx[6], dy[6], p, normal[6];
	std::vector<Vector3f> points;
	std::vector<Vector3f> normals;
	RESOLUTION = 20;

	points.clear();
	normals.clear();


	// Create one merlon

	// Front
	origin[0] = makeVector3f( -0.5,-0.5,-0.5 );
	dx[0] = makeVector3f( 1,0,0 );
	dy[0] = makeVector3f( 0,1,0 );
	normal[0] = makeVector3f( 0,0,-1 );
	// Front2
	origin[1] = makeVector3f( -0.5,-1.5,-0.5 );
	dx[1] = makeVector3f( 2,0,0 );
	dy[1] = makeVector3f( 0,1,0 );
	normal[1] = makeVector3f( 0,0,-1 );
	// Right
	origin[2] = makeVector3f( 0.5,-0.5,-0.5 );
	dx[2] = makeVector3f( 0,0,1 );
	dy[2] = makeVector3f( 0,1,0 );
	normal[2] = makeVector3f( 1,0,0 );
	// Left
	origin[3] = makeVector3f( -0.5,-0.5,0.5 );
	dx[3] = makeVector3f( 0,0,-1 );
	dy[3] = makeVector3f( 0,1,0 );
	normal[3] = makeVector3f( -1,0,0 );
	// Bottom Left
	origin[4] = makeVector3f( 0.5,-0.5,-0.5 );
	dx[4] = makeVector3f( 1,0,0 );
	dy[4] = makeVector3f( 0,0,1 );
	normal[4] = makeVector3f( 0,1,0 );
	// Top
	origin[5] = makeVector3f( -0.5,0.5,-0.5 );
	dx[5] = makeVector3f( 1,0,0 );
	dy[5] = makeVector3f( 0,0,1 );
	normal[5] = makeVector3f( 0,1,0 );
	for ( n = 0; n < 6; n++ ){
		for ( i = 0; i < RESOLUTION; i++ ){
			for ( j = 0; j < RESOLUTION; j++ ){
				p = origin[n] +
					dx[n] * ((float32)i+0.5f)/(float32)RESOLUTION +
					dy[n] * ((float32)j+0.5f)/(float32)RESOLUTION;
				points.push_back( p );
				normals.push_back( normalize( normal[n] ) );
			}
		}
	}


	// add other merlons
	const mpcard numPoints = points.size();
	for (mpcard i=1; i<numMerlons; ++i)
	{
		Vector3f moveVec = makeVector3f(2.0f*(float)(i),0,0);
		for (mpcard p=0; p<numPoints; p++)
		{
			points.push_back(points[p]+moveVec);
			normals.push_back(normals[p]);
		}
	}


	// Create iterators
	PointSet* ps = new PointSet();
	ps->clearAndSetup( 1, points.size(), &vd );
	AAT POSITION = ps->getAAT( "position" );
	AAT NORMAL = ps->getAAT( "normal" );
	AAT COLOR = ps->getAAT( "color" );
	for ( unsigned int pointIndex = 0; pointIndex < points.size(); pointIndex++ ){
		ps->set3f( pointIndex, POSITION, points[pointIndex] );
		ps->set3f( pointIndex, NORMAL, normals[pointIndex] );
		ps->set3f( pointIndex, COLOR, makeVector3f( 0.7f,0.7f,0.7f ) );
	}
	return ps;
}