//---------------------------------------------------------------------------
#ifndef CmdObjSelectionH
#define CmdObjSelectionH
//---------------------------------------------------------------------------
#include "x3editor.h"
//---------------------------------------------------------------------------
#include "CommandObject.h"
#include "PointCloud.h"
#include "ViewFrustum.h"
#include "Camera.h"
#include "Scene.h"
//---------------------------------------------------------------------------


namespace X4 {


// in this file you find:
// CmdObjChangeSelectionOnClick
// CmdObjSelectionByButton
// CmdObjSelectionOnArea
// CmdObjSelectionByAttribute



/// command object for all selections performed by click
class X3EDITOR_API CmdObjChangeSelectionOnClick : public CommandObject {

 private:
   static MetaClass *CmdObjChangeSelectionOnClickMetaClass;
   static Object* newInstance();

   struct clickPoint {
      int32 x;
      int32 y;
   };

 protected:

   // properties
   /// list of consecutive mouse clicks
   std::vector<clickPoint> clickList;
   /// actual mouse click x
   int32 clickX;
   /// actual mouse click y
   int32 clickY;
   /// in brushmode: size of brush in pixel
   card32 brushSize;
   /// selection mode which is used. 0=Cloud, 1=Point, 2=CircleBrush, 4=QuadBrush
   card32 selectMode;
   /// width and height of window, needed for ray calculation
   card32 areaWidth, areaHeight;
   /// if true then in cloud mode multiple clouds may be selected by mouse clicks, in point or brush modes this enables deselection
   bool multiselect;
   /// saves the camera used at selection time, needed for ray calculation
   Camera cam;
   /// saves the viewing frustum used at selection time, needed for ray calculation
   ViewFrustum viewFrust;
   /// if true then surface mode is enabled. See corresponding X4 help file for details of surface mode.
   bool surfaceMode;

   // user functions
   /// function which is called to select single clouds
   void selectCloud(Scene *scene);
   /// function which is called to select single points
   void selectPoint(Scene *scene);
   /// function which is called to select by brush
   void selectPointsBrush(Scene *scene);


 public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

   virtual void assign(const Object* obj, X4CopyContext *context = NULL);
   virtual void write(OutputObjectStream *s) const;
   virtual void read(InputObjectStream *s);

   /// setup parameters: see properties and implementation
   virtual void setup(int32 x, int32 y, bool multsel, card32 tol, card32 mode, Camera &ca, ViewFrustum &vfr, card32 w, card32 h, bool sphsel);

   virtual void doAction(Object *context);
   virtual void undoAction(Object *context);
   virtual bool canUndo();

   virtual std::string getDescription();

   virtual bool mergeCommand(const CommandObject *nextCmd);

};



/// command object for all selections performed by pressing a button
class X3EDITOR_API CmdObjSelectionByButton : public CommandObject {
 private:
   static MetaClass *CmdObjSelectionByButtonMetaClass;
   static Object* newInstance();

   // properties
   /// defines the action that is done by this command object: 0=invert points, 1=select all clouds, 2=select all points, 3=deselect all clouds, 4=deselect all points;
   card32 button;

 public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

   /// setup parameters: see properties and implementation
   virtual void setup(card32 btn);

   virtual void doAction(Object *context);
   virtual void undoAction(Object *context);
   virtual bool canUndo();

   virtual std::string getDescription();

   virtual bool mergeCommand(const CommandObject *nextCmd);
};



/// command object for all selections by opening an area with the mouse
class X3EDITOR_API CmdObjSelectionOnArea : public CommandObject {

 private:
   static MetaClass *CmdObjSelectionOnAreaMetaClass;
   static Object* newInstance();

   struct rectangle {
      int32 x1;
      int32 y1;
      int32 x2;
      int32 y2;
   };

 protected:

   // properties
   /// list of consecutive rectangles gathered from mouse interaction
   std::vector<rectangle> rectList;
   /// actual mouse x when pressed mouse button, x1 of last rectangle in list
   int32 clickX1;
   /// actual mouse y when pressed mouse button, y1 of last rectangle in list
   int32 clickY1;
   /// actual mouse x when released mouse button, x2 of last rectangle in list
   int32 clickX2;
   /// actual mouse y when released mouse button, y2 of last rectangle in list
   int32 clickY2;
   /// obsolete. Not used at the moment.
   card32 tolerance;
   /// selection mode which is used. 0=Cloud, 1=Point, 2=CircleBrush, 4=QuadBrush
   card32 selectMode;
   /// width and height of window, needed for ray calculation
   card32 areaWidth, areaHeight;
   /// if true this enables deselection
   bool multiselect;
   /// saves the camera used at selection time, needed for ray calculation
   Camera cam;
   /// saves the viewing frustum used at selection time, needed for ray calculation
   ViewFrustum viewFrust;

 public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

   virtual void assign(const Object* obj, X4CopyContext *context = NULL);
   virtual void write(OutputObjectStream *s) const;
   virtual void read(InputObjectStream *s);

   /// setup parameters: see properties and implementation
   virtual void setup(int32 x1, int32 y1, int32 x2, int32 y2, bool multsel, card32 tol, card32 mode, Camera &ca, ViewFrustum &vfr, card32 w, card32 h);

   virtual void doAction(Object *context);
   virtual void undoAction(Object *context);
   virtual bool canUndo();

   virtual std::string getDescription();

   virtual bool mergeCommand(const CommandObject *nextCmd);

};



/// command object for all selections referring to attributes
class X3EDITOR_API CmdObjSelectionByAttribute : public CommandObject {
 private:
   static MetaClass *CmdObjSelectionByAttributeMetaClass;
   static Object* newInstance();

   // properties
   /// name of attribute to select
   std::string AttribName;
   /// values to search in chosen attribute channel
   Vector4f AttribValues;
   /// tolerance for more inexact search
   card32 Tolerance;
   /// true if deselection is enabled
   bool deselect;

 public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

   /// setup parameters: see properties and implementation
   virtual void setup(std::string attrname, Vector4f attrval, card32 tol, bool desel);

   virtual void doAction(Object *context);
   virtual void undoAction(Object *context);
   virtual bool canUndo();

   virtual std::string getDescription();

   virtual bool mergeCommand(const CommandObject *nextCmd);
};

/// command object for growing selection based on normal differences
class X3EDITOR_API CmdObjGrowSelectionByNormal : public CommandObject {
private:
	static MetaClass *CmdObjGrowSelectionByNormalMetaClass;
	static Object* newInstance();

	// properties
	/// tolerated difference of neighbours 
	float angleDiff;
	/// use topology instead of knn
	bool useTopology;
	/// true if deselection is enabled
	bool deselect;

public:
	virtual const MetaClass *getInstanceClass() const;
	static const MetaClass *getClass();
	static void init(const MetaClass *parentClass);
	static void shutdown();

	/// setup parameters: see properties and implementation
	virtual void setup(float angleDiff, bool useTopology, bool desel);

	virtual void doAction(Object *context);
	virtual void undoAction(Object *context);
	virtual bool canUndo();

	virtual std::string getDescription();
};

}

#endif