#include "stdafx.h"
//---------------------------------------------------------------------------
#include "NormalEstimator.h"
#include "DebugOutput.h"
#include "HierarchicalKNNIterator.h"
#include "ProgressWindow.h"
#include "PointSetKNNQuery.h"
#include "PointSet.inline.h"
#include <time.h>
#include "NumericalClassProperty.h"
#include "PCA.h"
#include "MathTools.h"
#include "StringHelper.h"
//#include "PointSetANNQuery.h"
#include <float.h>
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

void NormalEstimator::estimateNormals( 
	PointSet* ps, 
	const card32 numKNN, 
	float32 epsilon, 
	float32 timesSamplingSpacing, 
	card32 mode,
	bool saveTangents )
{
	if ( ps == NULL ){
		debugOutput << "estimateNormals: invalid point set.\n"; debugOutput.flush();
		return;
	}

	if ( !ps->providesAttribute( "position" ) || !ps->providesAttribute( "normal" ) ){
		debugOutput << "estimateNormals: channels missing.\n"; debugOutput.flush();
		return;
	}

	float32 startTime = clock();

	AAT POSITION = ps->getAAT( "position" );
	AAT NORMAL = ps->getAAT( "normal" );
	AAT SAMPLING_SPACING;
	AAT TANGENT_U;
	AAT TANGENT_V;

	// Sampling spacing
	if ( mode == ENVIRONMENT_TIMES_SAMPLING_SPACING ){
		if ( !ps->providesAttribute( "samplingSpacing" ) ){
			debugOutput << "estimateNormals: channels missing.\n"; debugOutput.flush();
			return;
		} else {
			SAMPLING_SPACING = ps->getAAT( "samplingSpacing" );
		}
	} 

	// Tangents
	if ( saveTangents ){
		if ( !ps->providesAttribute( "tangent_u" ) || !ps->providesAttribute( "tangent_v" ) ){
			debugOutput << "estimateNormals: channel missing.\n"; debugOutput.flush();
			return;
		} else {
			TANGENT_U = ps->getAAT( "tangent_u" );
			TANGENT_V = ps->getAAT( "tangent_v" );
		}
	}

	Timer T; card32 vt;

	T.getDeltaValue();
	AttachedIndexedOctree* octree = PointSetKNNQuery::createOctree( ps, 32 );
	PointSetKNNQuery knn( ps, octree, makeVector3f( 0,0,0 ), POSITION );
	//PointSetANNQuery annQuery( ps, numKNN + 1 );
	vt = T.getDeltaValue();
	//debugOutput << "Build search structure: " <<  vt << " ms\n";debugOutput.flush();


	card32 nIndex;
	bool found;
	PCA3f pca;
	Matrix3f eigenVectors;
	Vector3f eigenValues, centroid;
	float32 sqrEpsilon = sqr( epsilon );

	card32 knnSetSeekTime=0;
	card32 knnTime=0;

	float32 avgNumNeighbors = 0;
	for ( unsigned int pIndex = 0; pIndex < ps->getNumEntries(); pIndex++ ){
		const Vector3f& p = ps->get3f( pIndex, POSITION );
		
		T.getDeltaValue();
		knn.setTargetPoint( p );
		knn.getNextNearestPointIndex( nIndex, found );	// Skip self
		//annQuery.setTargetPoint(p);
		//annQuery.getNextNearestPointIndex(nIndex, found);
		knnSetSeekTime+= T.getDeltaValue();



		float32 sqrD = 0;
		if ( mode == ENVIRONMENT_TIMES_SAMPLING_SPACING ) {
			float32 samplingSpacing = ps->get1f( pIndex, SAMPLING_SPACING );
			sqrEpsilon = sqr( timesSamplingSpacing * samplingSpacing );
		}

		pca.clear();
		if ( ( mode == ENVIRONMENT_EPSILON ) || ( mode == ENVIRONMENT_TIMES_SAMPLING_SPACING ) ){
			pca.addPoint( p, exp( - sqrt( sqrD ) / sqrEpsilon ) );
		} else { 
			pca.addPoint( p );
		}
		card32 numNeighbors = 0;
		bool converged = false;
		while ( !converged ){
			
			T.getDeltaValue();
			knn.getNextNearestPointIndex( nIndex, found );
			//annQuery.getNextNearestPointIndex(nIndex, found);
			knnTime+= T.getDeltaValue();
			
			
			const Vector3f& np = ps->get3f( nIndex, POSITION );
			sqrD = ( np - p ) * ( np - p );
			if ( ( mode == ENVIRONMENT_EPSILON ) || ( mode == ENVIRONMENT_TIMES_SAMPLING_SPACING ) ){
				pca.addPoint( np, exp( - sqrt( sqrD ) / sqrEpsilon ) );
			} else { 
				pca.addPoint( np );
			}
			numNeighbors++;

			if ( mode == ENVIRONMENT_KNN ){ 
				if ( numNeighbors >= numKNN ) 
					converged = true;
			} else if ( mode == ENVIRONMENT_EPSILON ) {
				if ( sqrD > sqrEpsilon ) 
					converged = true;
			} else if ( mode == ENVIRONMENT_TIMES_SAMPLING_SPACING ) {
				if ( sqrD > sqrEpsilon ) 
					converged = true;
			} else {
				converged = true;	// Invalid
			}
		}

		avgNumNeighbors = ( avgNumNeighbors * pIndex + numNeighbors ) / float32 ( pIndex + 1 );

		pca.analyze( eigenValues, eigenVectors, centroid );
		eigenVectors[2].normalize();

		Vector3f calculatedNormal = eigenVectors[2];
		if ( _isnan( calculatedNormal[0] ) || _isnan( calculatedNormal[1] ) || _isnan( calculatedNormal[2] ) ){
			debugOutput << "ERROR! - estimateNormals: calculated normal is NAN point "<< pIndex <<"\n"; debugOutput.flush();
			calculatedNormal = makeVector3f(1,0,0);
			if (saveTangents)
				throw X4Exception("ERROR! - estimateNormals: calculated normal is NAN point ");

		}
		if ( !_finite( calculatedNormal[0] ) || !_finite( calculatedNormal[1] ) || !_finite( calculatedNormal[2] ) ){
			debugOutput << "ERROR! - estimateNormals: calculated normal is INF point "<<pIndex << "\n"; debugOutput.flush();
			calculatedNormal = makeVector3f(1,0,0);
			if (saveTangents)
				throw X4Exception("ERROR! - estimateNormals: calculated normal is INF point ");
		}



		Vector3f oldNormal = ps->get3f(pIndex, NORMAL);
		if (calculatedNormal*oldNormal>=0)
		{
			ps->set3f(pIndex,  NORMAL, calculatedNormal );
		}else
		 ps->set3f(pIndex,  NORMAL, -calculatedNormal );

		if ( saveTangents ){
			ps->set3f( pIndex, TANGENT_U, eigenVectors[0] );
			ps->set3f( pIndex, TANGENT_V, eigenVectors[1] );
		}
	}
/*
	debugOutput << "seek time: " <<  knnSetSeekTime << " ms\n";
	debugOutput << "knn time: " <<   knnTime << " ms\n";
	debugOutput << "together: " <<  knnSetSeekTime+knnTime << " ms\n";



	debugOutput << "Average #neighbors: (estimate normals): " << avgNumNeighbors << ".\n"; debugOutput.flush();
	debugOutput << "Time required (estimate normals): " << float32( ( clock() - startTime ) / CLK_TCK ) << "\n\n"; debugOutput.flush();
*/

	delete octree;
}

void NormalEstimator::estimateNormals( 
	PointCloud* pc, 
	const card32 numKNN, 
	float32 epsilon, 
	float32 timesSamplingSpacing, 
	card32 mode,
	bool saveTangents )
{
	if ( pc == NULL ){
		debugOutput << "estimateNormals: invalid point set.\n"; debugOutput.flush();
		return;
	}

	if ( !pc->providesAttribute( "position" ) || !pc->providesAttribute( "normal" ) ){
		debugOutput << "estimateNormals: channels missing.\n"; debugOutput.flush();
		return;
	}

	float32 startTime = clock();

	AAT POSITION = pc->getAAT( "position" );
	AAT NORMAL = pc->getAAT( "normal" );
	AAT SAMPLING_SPACING;
	AAT TANGENT_U;
	AAT TANGENT_V;

	// Sampling spacing
	if ( mode == ENVIRONMENT_TIMES_SAMPLING_SPACING ){
		if ( !pc->providesAttribute( "samplingSpacing" ) ){
			debugOutput << "estimateNormals: channels missing.\n"; debugOutput.flush();
			return;
		} else {
			SAMPLING_SPACING = pc->getAAT( "samplingSpacing" );
		}
	} 

	// Tangents
	if ( saveTangents ){
		if ( !pc->providesAttribute( "tangent_u" ) || !pc->providesAttribute( "tangent_v" ) ){
			debugOutput << "estimateNormals: channel missing.\n"; debugOutput.flush();
			return;
		} else {
			TANGENT_U = pc->getAAT( "tangent_u" );
			TANGENT_V = pc->getAAT( "tangent_v" );
		}
	}

	SceneObjectIterator* it = pc->createIterator( SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC );
	BasicPointCloudIterator* basicIt = dynamic_cast<BasicPointCloudIterator*>( it );
	ModifyablePointCloudIterator* modIt = dynamic_cast<ModifyablePointCloudIterator*>( it );
	HierarchicalKNNIterator* knnIt = new HierarchicalKNNIterator( pc, 32, NULL );
	PCA3f pca;
	Matrix3f eigenVectors;
	Vector3f eigenValues, centroid;
	float32 sqrEpsilon = sqr( epsilon );
	float32 radius_2 = sqr( 2.0f * epsilon );

	basicIt->reset();
	long pIndex = 0;
	float32 avgNumNeighbors = 0;
	Timer reportTimer;
	card32 lastTimeReported = reportTimer.getCurrentValue();
	card32 reportStartTime = lastTimeReported;

	progressWindow->pushStep(true, "normal estimator");

	while( !basicIt->atEnd() ){
		const Vector3f& p = basicIt->get3f( POSITION );
		knnIt->setSeekPointAndReset( p );
		knnIt->next();	// Skip self
		float32 sqrD = 0;
		float32 samplingSpacing = 0;
		if ( mode == ENVIRONMENT_TIMES_SAMPLING_SPACING ) {
			samplingSpacing = basicIt->get1f( SAMPLING_SPACING );
			sqrEpsilon = sqr( timesSamplingSpacing * samplingSpacing );
		}

		pca.clear();
		if ( ( mode == ENVIRONMENT_EPSILON ) || ( mode == ENVIRONMENT_TIMES_SAMPLING_SPACING ) ){
			pca.addPoint( p, exp( - ( sqrD ) / sqrEpsilon ) );
		} else { 
			pca.addPoint( p );
		}
		card32 numNeighbors = 0;
		bool converged = false;
		while ( !converged ){
			knnIt->next();
			const Vector3f& np = knnIt->get3f( POSITION );
			sqrD = ( np - p ) * ( np - p );
			if ( mode == ENVIRONMENT_EPSILON ){
				pca.addPoint( np, exp( - ( sqrD ) / sqrEpsilon ) );
			} else { 
				pca.addPoint( np );
			}

			numNeighbors++;

			if ( mode == ENVIRONMENT_KNN ){ 
				if ( numNeighbors >= numKNN ) 
					converged = true;
			} else if ( mode == ENVIRONMENT_EPSILON ) {
				if ( sqrD > radius_2 ) 
					converged = true;
			} else if ( mode == ENVIRONMENT_TIMES_SAMPLING_SPACING ) {
				if ( sqrD > radius_2 ) 
					converged = true;
			} else {
				converged = true;	// Invalid
			}
		}

		avgNumNeighbors = ( avgNumNeighbors * pIndex + numNeighbors ) / float32 ( pIndex + 1 );

		pca.analyze( eigenValues, eigenVectors, centroid );
		eigenVectors[2].normalize();

		Vector3f calculatedNormal = eigenVectors[2];
		if ( _isnan( calculatedNormal[0] ) || _isnan( calculatedNormal[1] ) || _isnan( calculatedNormal[2] ) ){
			debugOutput << "ERROR! - estimateNormals: calculated normal is NAN point "<< pIndex <<"\n"; debugOutput.flush();
			calculatedNormal = makeVector3f(1,0,0);
			if (saveTangents)
				throw X4Exception("ERROR! - estimateNormals: calculated normal is NAN point ");

		}
		if ( !_finite( calculatedNormal[0] ) || !_finite( calculatedNormal[1] ) || !_finite( calculatedNormal[2] ) ){
			debugOutput << "ERROR! - estimateNormals: calculated normal is INF point "<<pIndex << "\n"; debugOutput.flush();
			calculatedNormal = makeVector3f(1,0,0);
			if (saveTangents)
				throw X4Exception("ERROR! - estimateNormals: calculated normal is INF point ");
		}


		Vector3f oldNormal = basicIt->get3f(NORMAL);
		if (calculatedNormal*oldNormal>=0)
		{
			modIt->set3f( NORMAL, calculatedNormal );
		}else
		 modIt->set3f( NORMAL, -calculatedNormal );

		if ( saveTangents ){
			modIt->set3f( TANGENT_U, eigenVectors[0] );
			modIt->set3f( TANGENT_V, eigenVectors[1] );
		}

		if( reportTimer.getCurrentValue() - lastTimeReported > 1000 * 10 )
		{
			float perc = (float)pIndex / pc->getNumPoints();
			unsigned remainingTime = (1.0f - perc) * (reportTimer.getCurrentValue() - reportStartTime); 
			//debugOutput <<  perc*100.0f << "% (completed in ~" << convertTimeToString(remainingTime) << "\n";
			lastTimeReported = reportTimer.getCurrentValue();
		}



		if (pIndex%11111==0)
		{	
			progressWindow->progress((float32)pIndex/(float32) pc->getNumPoints() *100.0f  );
		}

		pIndex++;
		basicIt->next();


	}
	progressWindow->popSteps();

/*
	debugOutput << "Average #neighbors: (estimate normals): " << avgNumNeighbors << ".\n"; debugOutput.flush();
	debugOutput << "Time required (estimate normals): " << float32( ( clock() - startTime ) / CLK_TCK ) << "\n"; debugOutput.flush();
*/

	delete it;
	delete knnIt;
};