#include "stdafx.h"
//---------------------------------------------------------------------------
#include "NormalUnifier.h"
#include "DebugOutput.h"
#include "HierarchicalKNNIterator.h"
#include "PointSetKNNQuery.h"
#include "PointSet.inline.h"
#include <time.h>
#include "NumericalClassProperty.h"
#include "MathTools.h"
#include "FifoStack.h"
#include "PointSetKNNQuery.h"
#include "IndexGraph.h"
#include "PriorityQueue.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

void NormalUnifier::unifyNormals( PointSet* ps, const card32 numKNN )
{
	if ( ps == NULL ){
		debugOutput << "unifyNormals: invalid point set.\n"; debugOutput.flush();
		return;
	}

	if ( !ps->providesAttribute( "position" ) || !ps->providesAttribute( "normal" ) || !ps->providesAttribute( "flags" ) ){
		debugOutput << "unify normals: channel missing.\n"; debugOutput.flush();
		return;
	}

	AAT POSITION = ps->getAAT( "position" );
	AAT NORMAL = ps->getAAT( "normal" );
	AAT FLAGS = ps->getAAT( "flags" );

	AttachedIndexedOctree* octree = PointSetKNNQuery::createOctree( ps, 32 );
	PointSetKNNQuery knn( ps, octree, makeVector3f( 100000, 0, 0 ), POSITION );

	Vector3f pos, normal, nNormal, seedPoint;

	// Reset all flags
	for ( unsigned int pIndex = 0; pIndex < ps->getNumEntries(); pIndex++ ){
		ps->set1i( pIndex, FLAGS, ps->get1i( pIndex, FLAGS ) & ( 0xFFFFFFFF ^ NORMAL_ESTIMATOR_FLAG ) );
	}

	/// Set std::vector with largest x coordinate
	card32 nIndex;
	bool found;
	knn.getNextNearestPointIndex( nIndex, found );
	seedPoint = ps->get3f( nIndex, POSITION );
	ps->set1i( nIndex, FLAGS, ps->get1i( nIndex, FLAGS ) | NORMAL_ESTIMATOR_FLAG );
	normal = ps->get3f( nIndex, NORMAL );
	if ( normal * makeVector3f( 1,0,0 ) < 0 ){
		ps->set3f( nIndex, NORMAL, -normal );
	}

	/// Start region growing
	card32 unifiedNormals = 0;
	FifoStack<card32> stack;
	stack.push( nIndex );
	while ( !stack.empty() ){
		card32 currentIndex = stack.pop();
		pos = ps->get3f( currentIndex, POSITION );
		knn.setTargetPoint( pos );	
		normal = ps->get3f( currentIndex, NORMAL );
		/// Push those nearest neighbors that are inconsistent
		card32 numNeighbors = 0;
		float32 sqrD = 0; 
		bool converged = false;
		while( !converged ){
			knn.getNextNearestPointIndex( nIndex, found );
			const Vector3f& nPos = ps->get3f( nIndex, POSITION );
			if ( !( ps->get1i( nIndex, FLAGS ) & NORMAL_ESTIMATOR_FLAG ) ){
				// Set normal according to
				nNormal = ps->get3f( nIndex, NORMAL );
				if ( nNormal * normal < 0 ){
					ps->set3f( nIndex, NORMAL, -nNormal );
				}
				ps->set1i( nIndex, FLAGS, ps->get1i( nIndex, FLAGS ) | NORMAL_ESTIMATOR_FLAG );
				stack.push( nIndex );
			}
			numNeighbors++;
			sqrD = ( pos - nPos ) * ( pos - nPos );
			if ( numNeighbors >= numKNN ) converged = true;
		}
		unifiedNormals++;
	}
	// Test if all normals are consistent
	int inconsistentNormals = 0;
	for ( unsigned int pIndex = 0; pIndex < ps->getNumEntries(); pIndex++ ){
		if ( !( ps->get1i( pIndex, FLAGS ) & NORMAL_ESTIMATOR_FLAG ) ){
			inconsistentNormals++;
		}
	}
	if ( inconsistentNormals > 0 )
		debugOutput << "Inconsistent normals: " << inconsistentNormals <<"!\n";

	delete octree;
}

void NormalUnifier::unifyNormals( PointCloud* pc, const card32 numKNN )
{
	if ( pc == NULL ){
		debugOutput << "unifyNormals: invalid point cloud.\n"; debugOutput.flush();
		return;
	}

	if ( !pc->providesAttribute( "position" ) || !pc->providesAttribute( "normal" ) || !pc->providesAttribute( "flags" ) ){
		debugOutput << "unify normals: channel missing.\n"; debugOutput.flush();
		return;
	}

	SceneObjectIterator *newIt = pc->createIterator( SOIT( SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC ) );
	BasicPointCloudIterator *basicIt = dynamic_cast<BasicPointCloudIterator*>(newIt);
	ModifyablePointCloudIterator *modIt = dynamic_cast<ModifyablePointCloudIterator*>(newIt);
	ModifyableKNNIterator *knnIt = new ModifyableKNNIterator( pc, 32, NULL );

	AAT POSITION = pc->getAAT( "position" );
	AAT NORMAL = pc->getAAT( "normal" );
	AAT FLAGS = pc->getAAT( "flags" );

	Vector3f pos, normal, nNormal, seedPoint;

	// Reset all flags
	basicIt->reset();
	while( !basicIt->atEnd() ){
		modIt->set1i( FLAGS, basicIt->get1i( FLAGS ) & ( 0xFFFFFFFF ^ NORMAL_ESTIMATOR_FLAG ) );
		basicIt->next();
	}

	/// Set std::vector with largest x coordinate
	knnIt->setSeekPointAndReset( makeVector3f( 100000, 0, 0 ) );
	seedPoint = knnIt->get3f( POSITION );
	knnIt->set1i( FLAGS, knnIt->get1i( FLAGS ) | NORMAL_ESTIMATOR_FLAG );
	normal = knnIt->get3f( NORMAL );
	if ( normal * makeVector3f( 1,0,0 ) < 0 ){
		knnIt->set3f( NORMAL, -normal );
	}

	/// Start region growing
	card32 unifiedNormals = 0;
	FifoStack<Vector3f>* stack = new FifoStack<Vector3f>();
	stack->push( seedPoint );
	while ( !stack->empty() ){
		pos = stack->pop();
		knnIt->setSeekPointAndReset( pos );
		normal = knnIt->get3f( NORMAL );
		/// Push those nearest neighbors that are inconsistent
		card32 numNeighbors = 0;
		float32 sqrD = 0; 
		bool converged = false;
		while( !converged ){
			const Vector3f& nPos = knnIt->get3f( POSITION );
			if ( !( knnIt->get1i( FLAGS ) & NORMAL_ESTIMATOR_FLAG ) ){
				// Set normal according to
				nNormal = knnIt->get3f( NORMAL );
				if ( nNormal * normal < 0 ){
					knnIt->set3f( NORMAL, -nNormal );
				}
				knnIt->set1i( FLAGS, knnIt->get1i( FLAGS ) | NORMAL_ESTIMATOR_FLAG );
				stack->push( nPos );
			}
			knnIt->next();

			numNeighbors++;
			sqrD = ( pos - nPos ) * ( pos - nPos );
			if ( numNeighbors >= numKNN ) converged = true;
		}
		unifiedNormals++;
	}
	// Test if all normals are consistent
	basicIt->reset();
	int inconsistentNormals = 0;
	while( !basicIt->atEnd() ){
		if ( !( knnIt->get1i( FLAGS ) & NORMAL_ESTIMATOR_FLAG ) ){
			inconsistentNormals++;
		}
		basicIt->next();
	}
	if ( inconsistentNormals > 0 ){
		debugOutput << "Inconsistent normals: " << inconsistentNormals <<"!\n";
	}

	// Free allocated memory
	delete stack;
	delete knnIt;
	delete newIt;
}

void NormalUnifier::unifyNormalsHoppe( PointSet* ps, const card32 numKNN )
{

	if ( ps == NULL ){
		debugOutput << "unifyNormalsImproved: invalid point set.\n"; debugOutput.flush();
		return;
	}

	if (ps->getNumEntries()==0)
	{
		debugOutput << "!!! unifyNormalsImproved: empty point set.!!!\n"; debugOutput.flush();
		return;
	}

	AAT POSITION = ps->getAAT( "position" );
	AAT NORMAL = ps->getAAT( "normal" );
	AAT FLAGS = ps->getAAT( "flags" );

	// Reset all flags
	for ( unsigned int pIndex = 0; pIndex < ps->getNumEntries(); pIndex++ ){
		ps->set1i( pIndex, FLAGS, ps->get1i( pIndex, FLAGS ) & ( 0xFFFFFFFF ^ NORMAL_ESTIMATOR_FLAG ) );
	}

	// Compute neighbor graph, assign confidence, remember best

	IGGraph* riemannian = new IGGraph();
	riemannian->createNodes( ps->getNumEntries() );
	AttachedIndexedOctree* octree = PointSetKNNQuery::createOctree( ps, 32 );
	PointSetKNNQuery knn( ps, octree, makeVector3f( 100000, 0, 0 ), POSITION );

	// Set std::vector with largest z coordinate
	knn.setTargetPoint( makeVector3f( 0, 0, 100000 ) );
	card32 nIndex;
	bool found;
	knn.getNextNearestPointIndex( nIndex, found );
	Vector3f rootPoint = ps->get3f( nIndex, POSITION );
	riemannian->setRoot( nIndex );

	//create stack
	FifoStack<card32> stack;
	stack.push( nIndex );

	//Build riemannian graph

	int pointcounter = 0;
	int knncounter = 0;
	while( !stack.empty()){
		pointcounter++;
		const card32 currentIndex = stack.pop();
		Vector3f current = ps->get3f( currentIndex, POSITION );
		//calculate average point distance
		knn.setTargetPoint( current );
		float avgDist = 0;
		knncounter = 0;
		while(knncounter++ < (int)numKNN){
			knn.getNextNearestPointIndex( nIndex, found );
			Vector3f neighbor = ps->get3f( nIndex, POSITION );
			avgDist += (current - neighbor).getSqrNorm();
		}
		avgDist /= float32( numKNN );
		Vector3f n_i, n_j;
		knn.setTargetPoint( current );
		ps->set1i( currentIndex, FLAGS, ps->get1i( currentIndex, FLAGS ) | NORMAL_ESTIMATOR_FLAG );
		n_i = ps->get3f( currentIndex, NORMAL );
		knncounter = 0;
		float costs;
		while(knncounter < (int)numKNN){
			knn.getNextNearestPointIndex( nIndex, found );
			Vector3f neighbor = ps->get3f( nIndex, POSITION );
			n_j = ps->get3f( nIndex, NORMAL );
			costs = (1 - abs(n_i * n_j) * exp(-pow((current-neighbor).getSqrNorm(),2)/pow(avgDist,2)));
			//costs = (1 - abs(n_i * n_j));
			riemannian->createEdge(currentIndex, nIndex, costs);
			if(!(ps->get1i(nIndex,FLAGS) & NORMAL_ESTIMATOR_FLAG)){ 
				stack.push(nIndex);
				ps->set1i(nIndex,FLAGS,ps->get1i(nIndex,FLAGS) | NORMAL_ESTIMATOR_FLAG);
			}
			knncounter++;
		}
	}

	// Build minimal spanning tree
	IGGraph* mst = new IGGraph();
	mst->createNodes(riemannian->getNodeCount());
	mst->setRoot(riemannian->getRoot());
	PriorityQueue<float,IGEdge*>* pq = new PriorityQueue<float,IGEdge*>();

	IGNode* root = riemannian->getNode(riemannian->getRoot());
	root->setFlag(true);
	list<IGEdge*>::iterator it = root->getEdgeIterator();
	for(unsigned i=0;i<root->getEdgeCount();i++){
		IGEdge* current = *it;
		pq->insert(current->getCosts(),current);
		it++;
	}

	IGEdge* currentEdge;
	IGNode* currentNode;
	IGNode* targetNode;
	int connectedNodes = 0;

	while(!pq->isEmpty()){
		currentEdge = pq->pop_first();
		currentNode = riemannian->getNode(currentEdge->getStartNode());
		targetNode = riemannian->getNode(currentEdge->getTargetNode());
		if(targetNode->isFlagSet()) continue;
		mst->createEdge(currentNode->getIndex(),targetNode->getIndex(),currentEdge->getCosts());
		list<IGEdge*>::iterator it = targetNode->getEdgeIterator();
		for(unsigned i=0;i<targetNode->getEdgeCount();i++){
			if(!riemannian->getNode((*it)->getTargetNode())->isFlagSet()) pq->insert((*it)->getCosts(),*it);
			it++;
		}
		targetNode->setFlag(true);
		connectedNodes++;
	}

	// Start growing with best
	root = mst->getNode(mst->getRoot());
	root->setFlag(true);
	card32 currentIndex = root->getIndex();
	knn.setTargetPoint( ps->get3f( currentIndex, POSITION ) );
	Vector3f normal = ps->get3f( currentIndex, NORMAL );
	if ( normal * makeVector3f( 1,0,0 ) < 0 ){
		ps->set3f( currentIndex, NORMAL, -normal );
	}

	currentNode = root;
	list<IGEdge*>* dfs = new list<IGEdge*>();
	it = root->getEdgeIterator();
	for(unsigned i=0;i<root->getEdgeCount();i++){
		dfs->push_front(*it);
		it++;
	}

	/***********************************
	* FOR TESTING ONLY! SHUFFLES NORMALS
	************************************/
	/*
	for (int i=0;i<mst->getNodeCount();i++){
	knnIt->setSeekPointAndReset(ps->get3f(i,POSITION));
	if(i%2==0) knnIt->set3f(NORMAL,-(knnIt->get3f(NORMAL)));
	}
	*/

	//Traversing MST in depth first order
	int nodeCounter = 0;
	while(!dfs->empty()){
		currentEdge = dfs->front();
		dfs->pop_front();
		currentNode = mst->getNode(currentEdge->getStartNode());
		targetNode = mst->getNode(currentEdge->getTargetNode());
		if(!targetNode->isFlagSet()){
			Vector3f currentNormal, targetNormal;
			card32 currentIndex = currentNode->getIndex();
			knn.setTargetPoint(ps->get3f(currentIndex,POSITION));
			currentNormal = ps->get3f(currentIndex,NORMAL);
			card32 targetIndex = targetNode->getIndex();
			knn.setTargetPoint(ps->get3f(targetIndex,POSITION));
			targetNormal = ps->get3f(targetIndex,NORMAL);
			if(currentNormal * targetNormal < 0) ps->set3f(targetIndex,NORMAL,-targetNormal);
			targetNode->setFlag(true);
			nodeCounter++;
			it = targetNode->getEdgeIterator();
			for(unsigned i=0;i<targetNode->getEdgeCount();i++){
				if(!mst->getNode((*it)->getTargetNode())->isFlagSet()) dfs->push_front(*it);
				it++;
			}
		}
	}

	delete pq;
	delete riemannian;
	delete mst;
	delete dfs;
	delete octree;
}