#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PCCImplicitMesher.h"
//---------------------------------------------------------------------------
#include "MarchingCubes.h"
#include "HierarchicalKNNIterator.h"
#include "Scene.h"
#include "PointCloud.h"
#include "NumericalClassProperty.h"
#include "PointCloud.h"
#include "PointCloudTools.h"
#include "PointCloudIterators.h"
#include "StandardSceneGraphIterator.h"
#include "SGObjectNode.h"
#include "UnstructuredInCoreTriangleMesh.h"
#include "VertexArray.h"
#include "RegularData.h"
#include "RegularData.inline.h"
#include "AttachedIndexedOctree.h"
#include "PointCloudTools.h"
#include "SeparatorClassProperty.h"
#include "ClassMethods.h"
#include "RegularVolume.inline.h"
#include "PointSet.inline.h"
#include "InCoreVolume.h"
#include "ExceptionHandling.h"
//---------------------------------------------------------------------------
using namespace X4;

IMPLEMENT_X4_CLASS(PCCImplicitMesher, 5)
{
	BEGIN_CLASS_INIT( PCCImplicitMesher );
	ADD_CARD32_PROP( m_Resolution, 0);
	ADD_FLOAT32_PROP( m_Sigma, 0 );
	ADD_FLOAT32_PROP( m_GridSnapTolerance, 0);
	ADD_BOOLEAN_PROP (smoothNormals,1)
	ADD_FLOAT32_PROP(smoothNormals_Sigma,2)
	ADD_FLOAT32_PROP(borderFactor, 5)
	ADD_SEPARATOR("line feature sigma");
	ADD_STRING_PROP(m_LineChannel,3);
	ADD_FLOAT32_PROP(m_IsoValue,3);
	ADD_BOOLEAN_PROP(simpleLineDistance, 4)
	ADD_NOARGS_METHOD(meshLineFeatures);
	ADD_NOARGS_METHOD(transferCurvatureFromVolume);
}

PCCImplicitMesher::PCCImplicitMesher(void)
{
	//m_Sigma = 1.0f;
	m_Sigma = 2.0f;
	m_Resolution = 30;
	//m_GridSnapTolerance = 0.1f;
	m_GridSnapTolerance = 0.0001f;
	smoothNormals=true;
	smoothNormals_Sigma=1.0f;
	//m_IsoValue = 0.1f;
	m_IsoValue = 0.55f;
	//m_LineChannel = "kMin";
	m_LineChannel = "tangent";
	simpleLineDistance = false;
	borderFactor = 1.5f;
}

PCCImplicitMesher::~PCCImplicitMesher(void)
{
}

void PCCImplicitMesher::assign(const Object* obj, X4CopyContext *context)
{
	PCCommand::assign(obj,context);
	const PCCImplicitMesher * o = dynamic_cast<const PCCImplicitMesher*>(obj);
	x4Assert( o != NULL );

	m_Sigma = o->m_Sigma;
	m_Resolution = o->m_Resolution;
	borderFactor = o->borderFactor;
}

struct TMesherContext
{
	AttachedIndexedOctree * octree;
	HierarchicalKNNIterator *	hIt;
	VertexArray * vertexArray;
	float		sigma;
	float    borderFactor;
	float		averagePointDistance;
	float		cubeSize;
	AAT		posAAT;
	AAT		normalAAT;
	
	AAT		lineAAT;
	float	isoValue;
	bool simpleLineDistance;
};

TMesherContext * sdContext = NULL;

float signedDistanceFunction( void *context, float x, float y, float z )
{
	Vector3f pos = makeVector3f(x,y,z);
	sdContext->hIt->setSeekPointAndReset(pos);
	float summedWeight = 0.0f;
	card32 numSamples = 0;
	float signedDistance = 0.0f;
	float scale = 1.0f / sdContext->averagePointDistance;
	while(!sdContext->hIt->atEnd())
	{
		Vector3f samplePos = sdContext->hIt->get3f( sdContext->posAAT );
		Vector3f normal = sdContext->hIt->get3f( sdContext->normalAAT );
		float distance = norm( samplePos - pos );

		if( distance > sdContext->averagePointDistance * 3.0f && numSamples == 0)
		{
			Vector3f dir = -normalize(samplePos - pos);
			Vector3f testPos = samplePos + dir*0.5f*sdContext->averagePointDistance;
			float ret = signedDistanceFunction( context, testPos[0],testPos[1],testPos[2] );
			if( ret < 0 )
				return -distance;
			else
				return distance;
		}

		distance *=  scale;
		sdContext->hIt->next();

		float weight = //0.5f / M_PI / sdContext->sigma * 
			exp( -0.5f * pow(distance/sdContext->sigma,2.0f) );

		signedDistance += (samplePos*normal - pos*normal) * weight;

		summedWeight += weight;
		

		if( numSamples++ > 100 || weight * 100<summedWeight || weight < 0.0001f)
			if( numSamples > 1 )
				break;
			else
				return (samplePos*normal - pos*normal);
	}

	if( summedWeight > 0.1e-25f )
		return signedDistance / summedWeight;
	else
		return 1.0f;
}

float lineDistanceFunction(void *context, float x, float y, float z )
{
	if (sdContext->lineAAT == NULL_AAT) return 0;
	Vector3f pos = makeVector3f(x,y,z);
	sdContext->hIt->setMaxDistanceToSeekPoint(max(sdContext->averagePointDistance * 3.0f,sdContext->isoValue*1.5f));
	sdContext->hIt->setSeekPointAndReset(pos);
	if( sdContext->hIt->atEnd())
		return 1e20f;

	float summedWeight = 0.0f;
	card32 numSamples = 0;
	float retDistance = 0.0f;
	float scale = 1.0f / sdContext->averagePointDistance;
	while(!sdContext->hIt->atEnd())
	{
		Vector3f samplePos = sdContext->hIt->get3f( sdContext->posAAT );
		Vector3f normal = sdContext->hIt->get3f( sdContext->normalAAT );
		Vector3f lineDir = sdContext->hIt->get3f(sdContext->lineAAT);
		float distance = norm( samplePos - pos );

		distance *=  scale;
		sdContext->hIt->next();

		float weight = exp( -0.5f * pow(distance/sdContext->sigma,2.0f) );

		float dot = 0;
		float lineDistance = 0;
		
		if (sdContext->simpleLineDistance) {
			dot = norm(samplePos - pos);

		} else {
			dot = (samplePos - pos)*lineDir;
			lineDistance = norm( (samplePos - pos) - lineDir*dot );
		}

		summedWeight += weight;
		retDistance += (lineDistance*.8f + distance*.2f) * weight;


// 		if( numSamples++ > 100 || weight * 100<summedWeight || weight < 0.0001f)
// 			if( numSamples > 1 )
// 				break;
// 			else
// 				return (retDistance);
	}

	if( summedWeight > 0.1e-25f )
		return retDistance / summedWeight;
	else
		return 1e20f;
}

Vector3f computeFunctionGradient_LineFeatures( const Vector3f &pos )
{
	sdContext->hIt->setMaxDistanceToSeekPoint(max(sdContext->averagePointDistance * 3.0f,sdContext->isoValue*1.5f));
	sdContext->hIt->setSeekPointAndReset(pos);
	if( sdContext->hIt->atEnd())
		return makeVector3f(1,1,1);

	float summedWeight = 0.0f;
	card32 numSamples = 0;
	Vector3f gradient = NULL_VECTOR3F;
	float scale = 1.0f / sdContext->averagePointDistance;
	while(!sdContext->hIt->atEnd())
	{
		Vector3f samplePos = sdContext->hIt->get3f( sdContext->posAAT );
		Vector3f normal = sdContext->hIt->get3f( sdContext->normalAAT );
		Vector3f lineDir = sdContext->hIt->get3f(sdContext->lineAAT);
		float distance = norm( samplePos - pos );

		distance *=  scale;
		sdContext->hIt->next();

		float weight = exp( -0.5f * pow(distance/sdContext->sigma,2.0f) );
		float dot = (pos - samplePos)*lineDir;
		summedWeight += weight;
		gradient += ((pos - samplePos) - lineDir*dot ) * weight;

	}

	if( summedWeight > 0.1e-25f )
		return gradient / summedWeight;
	else
		return makeVector3f(1,1,1);
}

bool isBoundingBoxEmpty( const BoundingBox3f &bb, AIOctreeNode * node )
{
   if( node->isLeaf() )
   {
      for( card32 i=0;i<node->getContent().m_index.size();i++ )
      {
         card32 index = node->getContent().m_index[i];
			if( bb.inBoundingBox( sdContext->vertexArray->getPosition3f( index ) ) )
            return false;
      }
      return true;
   }
   else
   {
      if( node->getContent().m_bb.intersectBoundingBox( bb ) )
      {
          for( card32 i=0;i<8;i++ )
          {
            if( !isBoundingBoxEmpty( bb, node->getChild(i) ) )
               return false;
          }
          return true;
      }
   }
   return true;
}

bool trianglulate_isBoundingBoxEmpty( void *context, const BoundingBox3f &bb )
{
	BoundingBox3f checkBB = bb;
	checkBB.addBorder( max(sdContext->cubeSize * 0.55f,sdContext->averagePointDistance*sdContext->borderFactor) );
	return isBoundingBoxEmpty( checkBB, sdContext->octree->getOctree() );
}

bool trianglulate_isBoundingBoxEmptyLines( void *context, const BoundingBox3f &bb )
{
	BoundingBox3f checkBB = bb;
	checkBB.addBorder( max(sdContext->cubeSize * 0.55f,sdContext->averagePointDistance*1.5f+sdContext->sigma*4.0f + sdContext->isoValue) );
	return isBoundingBoxEmpty( checkBB, sdContext->octree->getOctree() );
}


void PCCImplicitMesher::execute(Scene *scene) 
{
   std::vector<std::string> selection = scene->getRootState()->staticState->getSelectedNodes();

	if( selection.size() == 0 )
	{
		// collect all point clouds
		StandardSceneGraphIterator* it = new StandardSceneGraphIterator();
		it->clearAndSetup( scene->getRootNode(), scene->getRootState() );
		while ( !it->atEnd() )
		{
			SGObjectNode* sgon = dynamic_cast<SGObjectNode*>( it->getNode() );
			if( sgon )
			{
				PointCloud * pc = dynamic_cast<PointCloud*>( sgon->getSceneObject() );
				UnstructuredInCoreTriangleMesh * mesh = dynamic_cast<UnstructuredInCoreTriangleMesh*>( sgon->getSceneObject() );
				if( pc && !mesh)
				{
					selection.push_back( it->getNameWithPath() );
				}
			}
			it->next();
		}
	}


	for ( unsigned m = 0; m < selection.size(); m++ )
	{
		triangulatePointCloud( scene, selection[m],false );
	}
	
}

void PCCImplicitMesher::triangulatePointCloud( Scene * scene, std::string pcName, bool lineMode ) 
{
	STD_GLOBAL_EXCEPTION_HANDLER_START
	PointCloud * pc =  getPointCloud( scene, pcName );
	UnstructuredInCorePointCloud * upc = dynamic_cast<UnstructuredInCorePointCloud*>(pc);
	x4Assert( pc->providesAttribute( "position" ) );
	x4Assert( pc->providesAttribute( "normal" ) );
	PointCloud * cornerPC = getPointCloud(scene, "root/corners");
	
	// prepare context
	TMesherContext context;
	context.hIt = new HierarchicalKNNIterator(pc, 32, NULL );
	context.posAAT = pc->getAAT( "position" );
	context.normalAAT = pc->getAAT( "normal" );
	context.sigma = m_Sigma;
	context.borderFactor = borderFactor;
	context.octree = new AttachedIndexedOctree;
	context.octree->octreesetup( upc->getPointSet(), 32 );
	context.vertexArray = new VertexArray( upc->getPointSet() );
	if( lineMode )
	{
		context.lineAAT = upc->getAAT(m_LineChannel);
		context.isoValue = m_IsoValue;
		context.simpleLineDistance = simpleLineDistance;
	}

	sdContext = &context;

	SceneObjectIterator *newIt = pc->createIterator( SOIT( SOIT::CAP_HIERARCHICAL | SOIT::CAP_BASIC_PC ) );
   HierarchicalIterator * hIt = dynamic_cast<HierarchicalIterator*>(newIt);
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(newIt);

	// calculate average point distance
	context.averagePointDistance = getMedianPointDistance( pc );

	// calc boundingbox and enlarge it a little bit ( 10% each dim + average point distance*2)
   BoundingBox3f bb = hIt->getBoundingBox();
   for( card32 i=0;i<3;i++ )
   {
      float delta = max(bb.upperCorner[i] - bb.lowerCorner[i],0.0001f);
      bb.lowerCorner[i] -= delta * 0.05f + context.averagePointDistance*2.0f;
      bb.upperCorner[i] += delta * 0.05f + context.averagePointDistance*2.0f;
   }
	bb.makeBoundingCube();
	context.cubeSize = bb.getSideLength(0) / (float)m_Resolution;


	// calculate triangulation
	MarchingCubes mc;
	Vector3i dim = makeVector3i(1,1,1) * m_Resolution;
	mc.setGridSnappingTolerance( m_GridSnapTolerance );
	if( !lineMode )
		mc.calculateHierarchical( &signedDistanceFunction, NULL, &trianglulate_isBoundingBoxEmpty, bb, dim, 0.0f );
	else
		mc.calculateHierarchical( &lineDistanceFunction, NULL, &trianglulate_isBoundingBoxEmptyLines, bb, dim, m_IsoValue );


	// create output point cloud
	UnstructuredInCoreTriangleMesh * mesh = new UnstructuredInCoreTriangleMesh;
   mesh->setMaterialIndex( 3 );

	VertexDescriptor vdesc;
   vdesc.pushAttrib( VAD( "position", 3, VAD::DATA_FORMAT_FLOAT32 ) );
   vdesc.pushAttrib( VAD( "normal", 3, VAD::DATA_FORMAT_FLOAT32 ) );

   if( lineMode )
   {
	   vdesc.pushAttrib( VAD("color", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	   if( cornerPC )
		   vdesc.pushAttrib( VAD("distToCorner",1,VAD::DATA_FORMAT_FLOAT32));
   }

	VertexDescriptor idesc;
   idesc.pushAttrib( VAD( "index", 3, VAD::DATA_FORMAT_INT32 ) );

	mesh->clearAndSetup( &vdesc, mc.getNumVertices(), &idesc, mc.getNumTriangles() );
	VertexArray vertices( mesh->getPointSet() );
	VertexArray indices( mesh->getTriangles() );
	for( card32 i=0;i<mc.getNumVertices();i++ )
	{
		vertices.setPosition3f( i, mc.getVertex(i) );
	}
	for( card32 i=0;i<mc.getNumTriangles();i++ )
	{
		indices.setIndex3i( i, mc.getTriangle(i) );
	}

	mesh->snapMesh( sdContext->cubeSize * m_GridSnapTolerance );
// 	mesh->snapMesh( sdContext->cubeSize * m_GridSnapTolerance );
// 	mesh->snapMesh( sdContext->cubeSize * m_GridSnapTolerance );
	mesh->closeMesh();

	if( !lineMode )
		mesh->calculateVertexNormals();
	else
	{
		// compute normals via implicit function
		VertexArray vertices( mesh->getPointSet() );
		for( mpcard i=0;i<vertices.getNumElements();i++ )
		{
			Vector3f pos = vertices.getPosition3f(i);
			Vector3f gradient = computeFunctionGradient_LineFeatures(pos);
			float gradLendgth = norm(gradient);
			if( gradLendgth )
				vertices.setNormal3f(i,gradient / gradLendgth);
		}

		// compute distance to corners
		if( cornerPC )
		{
			HierarchicalKNNIterator hItCorners(cornerPC,32,NULL);
			AAT distToCornerAAT = mesh->getAAT("distToCorner");
			AAT corners_PosAAT = cornerPC->getAAT("position");
			for( mpcard i=0;i<vertices.getNumElements();i++ )
			{
				Vector3f pos = vertices.getPosition3f(i);
				hItCorners.setSeekPointAndReset(pos);
				Vector3f corner = hItCorners.get3f(corners_PosAAT);
				Vector1f dist; dist[0] = norm(corner - pos);
				vertices.set<float,1>(i,distToCornerAAT,dist);
			}
		}

	}


	// output mesh
	std::string outputName = pcName + "_mesh";
	if( getPointCloud(scene, outputName ) )
		replacePointCloud( scene, outputName, mesh );
	else
		addPointCloud( scene, mesh, outputName );

	delete newIt;
	delete context.hIt;
	delete context.octree;
	delete context.vertexArray;


	if(smoothNormals){
		checkAttribute( mesh, "smoothNormal", 3, VAD::DATA_FORMAT_FLOAT32 ) ;
		smoothAttribute3f( mesh, "normal", "smoothNormal", smoothNormals_Sigma, true);	
	}
	STD_GLOBAL_EXCEPTION_HANDLER_END
}

void X4::PCCImplicitMesher::meshLineFeatures()
{
	Scene * scene = getScene();
	std::vector<std::string> selection = scene->getRootState()->staticState->getSelectedNodes();

	for ( unsigned m = 0; m < selection.size(); m++ )
	{
		triangulatePointCloud( scene, selection[m],true );
	}
}

void X4::PCCImplicitMesher::transferCurvatureFromVolume()
{
	Scene * scene = getScene();
	std::vector<std::string> selection = scene->getRootState()->staticState->getSelectedNodes();

	// search for volume, UPC and mesh, assumption is that those three are selected, no checks whatsoever are performed
	UnstructuredInCoreTriangleMesh * mesh = NULL;
	InCoreVolume * volume = NULL;
	UnstructuredInCorePointCloud * upc = NULL;
	for ( unsigned m = 0; m < selection.size(); m++ )
	{
		PointCloud * pc = getPointCloud(getScene(),selection[m]);
		if(!mesh)
			mesh = dynamic_cast<UnstructuredInCoreTriangleMesh*>(pc);
		if(!upc && dynamic_cast<UnstructuredInCoreTriangleMesh*>(pc) == NULL)
			upc = dynamic_cast<UnstructuredInCorePointCloud*>(pc);
		if( !volume)
			volume = dynamic_cast<InCoreVolume*>(pc);
	}

	if( !volume || !mesh || !upc )
	{
		message("Select exaclty one mesh, one volume and one upc");
		return;
	}

	checkAttribute(mesh,"curv",2,VAD::DATA_FORMAT_FLOAT32);
	AAT curvAAT = mesh->getAAT("curv");
	
	AAT posAAT = upc->getAAT("position");
	HierarchicalKNNIterator hIt_upc(upc,32,NULL);
	float medPointDistance = getMedianPointDistance(upc);
	float sigma = medPointDistance;
	float inv_sigma_2 = 1.0f / sigma / sigma;
	hIt_upc.setMaxDistanceToSeekPoint(sigma*2.25f);
	VertexArray va_points(mesh);
	RegularVolume * regvolume = volume->getVoxels();
	AAT kMaxAAT = volume->getAAT("kMax");
	AAT kMinAAT = volume->getAAT("kMin");
	for( mpcard i=0;i<mesh->getNumPoints();i++ )
	{
		Vector3f mesh_pos = va_points.getPosition3f(i);
		

		hIt_upc.setSeekPointAndReset(mesh_pos);
		float summedWeight = 0;
		Vector2f curv = NULL_VECTOR2F;
		while(!hIt_upc.atEnd())
		{
			Vector3f upc_pos = hIt_upc.get3f(posAAT);
			
			// transform position into volume coordinates
			Vector3f vol_pos = regvolume->worldPosToVoxel(upc_pos);
			Vector4f kMin = regvolume->getVoxel<float,4>(vol_pos[0],vol_pos[1],vol_pos[2],kMinAAT);
			Vector4f kMax = regvolume->getVoxel<float,4>(vol_pos[0],vol_pos[1],vol_pos[2],kMaxAAT);
			float weight = exp( -normQuad(upc_pos - mesh_pos)*inv_sigma_2 );
			curv += makeVector2f(kMin[3],kMax[3])*weight;
			summedWeight += weight;
			hIt_upc.next();
		}
		va_points.set<float,2>(i,curvAAT,curv/summedWeight);
	}
}