//---------------------------------------------------------------------------
#include "StdAfx.h"

//---------------------------------------------------------------------------
#include "PCCNormalEstimatorPCA.h"
//---------------------------------------------------------------------------
#include "NumericalClassProperty.h"
#include "FixedArrayClassProperty.h"
#include "DebugOutput.h"

#include "PointCloud.h"
#include "PointCloudTools.h"

#include "Scene.h"
#include "SceneGraphState.h"
#include "StandardSceneGraphIterator.h"
#include "SGObjectNode.h"
#include "LinearAlgebra.inline.h"
#include "NormalUnifier.h"
#include "NormalEstimator.h"
//---------------------------------------------------------------------------
 
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

MetaClass* PCCNormalEstimatorPCA::PCCNormalEstimatorPCAMetaClass = NULL;

const MetaClass* PCCNormalEstimatorPCA::getInstanceClass() const {
   return getClass();
}

Object* PCCNormalEstimatorPCA::newInstance() {
   return new PCCNormalEstimatorPCA;
}

const MetaClass* PCCNormalEstimatorPCA::getClass() {
   if (PCCNormalEstimatorPCAMetaClass==NULL) {
      PCCNormalEstimatorPCAMetaClass = new MetaClass("PCCNormalEstimatorPCA", "no description.",
                                  newInstance, sizeof(PCCNormalEstimatorPCA));
   }
   return PCCNormalEstimatorPCAMetaClass;
}

void PCCNormalEstimatorPCA::init(const MetaClass *parentClass) {
   ((MetaClass*)getClass())->init(parentClass, MetaClass::FIRST_VERSION + 7);
   PCCNormalEstimatorPCAMetaClass->copyParentProperties();
   BEGIN_CLASS_INIT(PCCNormalEstimatorPCA)
   ADD_BOOLEAN_PROP( computeTangents , 0)
   ADD_CARD32_PROP(nearestNeighbors ,0)
   ADD_BOOLEAN_PROP(unifyNormals, 4)
   ADD_BOOLEAN_PROP(useImprovedAlgorithm, 7)
   ADD_BOOLEAN_PROP(usePreviousOrientation, 4)
   ADD_BOOLEAN_PROP(camInOrigin,5)
	ADD_BOOLEAN_PROP(scaleTangentVectors,6);
}



void PCCNormalEstimatorPCA::shutdown() {
   delete PCCNormalEstimatorPCAMetaClass;
}

//----------------------------------------------------------------------

/**
 * Create new object
 */
PCCNormalEstimatorPCA::PCCNormalEstimatorPCA()
{
	computeTangents = false;
	nearestNeighbors = 20;
	camInOrigin = false;
	unifyNormals = false;
	usePreviousOrientation = true;
	scaleTangentVectors = false;
	useImprovedAlgorithm = false;
}

/// compatibility - mw, added Oct/02/05
void PCCNormalEstimatorPCA::read(InputObjectStream *s) {
   if (s->getClassVersion(getClass()) <= MetaClass::FIRST_VERSION + 1) {
      float32 subTreeSize;
      s->read(subTreeSize);
      card32 NumPoints;
      s->read(NumPoints);
      float32 m_scale;
      s->read(m_scale);
      if (s->getClassVersion(getClass()) == MetaClass::FIRST_VERSION + 1) {
         bool consistent_chapel_normals;
         s->readBool(consistent_chapel_normals);
      }
   } else {
      PCCommand::read(s);
   }
}


/**
 * Execute command
 */
void PCCNormalEstimatorPCA::execute( Scene *scene )
{
   /// Find selected point cloud
   unsigned int m;
   std::vector<std::string> selection = scene->getRootState()->staticState->getSelectedNodes();
   PointCloud* pc;
   /// Use all point clouds in selection
   for ( m = 0; m < selection.size(); m++ ){
      /// Check for required attributes
      checkAttribute( scene, selection[m], "position", 3, VAD::DATA_FORMAT_FLOAT32 );
      checkAttribute( scene, selection[m], "normal", 3, VAD::DATA_FORMAT_FLOAT32 );
      if ( computeTangents ){
         checkAttribute( scene, selection[m], "tangent_u", 3, VAD::DATA_FORMAT_FLOAT32 );
         checkAttribute( scene, selection[m], "tangent_v", 3, VAD::DATA_FORMAT_FLOAT32 );
      }
		if( unifyNormals )
			checkAttribute( scene, selection[m], "intAtt", 1, VAD::DATA_FORMAT_INT32 );
      /// Estimate normals
      pc = getPointCloud( scene, selection[m] );
	  NormalEstimator::estimateNormals( pc, nearestNeighbors, -1, -1, ENVIRONMENT_KNN, computeTangents );
      if (!usePreviousOrientation) {
         if (unifyNormals) {
			 if( !useImprovedAlgorithm ) {
				 NormalUnifier::unifyNormals( pc, nearestNeighbors );
			 } else{
				 UnstructuredInCorePointCloud* upc = dynamic_cast<UnstructuredInCorePointCloud*>( pc );
				 if ( upc != NULL )
					 NormalUnifier::unifyNormalsHoppe( upc->getPointSet(), nearestNeighbors );
			 }
         }
      }
		if (camInOrigin)
		{
			unifyNormalsByCameraPosition( pc );
		}
   }
}

void PCCNormalEstimatorPCA::unifyNormalsByCameraPosition( PointCloud *pc )
{
	SceneObjectIterator *newIt = pc->createIterator( SOIT( SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC ) );
	BasicPointCloudIterator *basicIt = dynamic_cast<BasicPointCloudIterator*>(newIt);
	ModifyablePointCloudIterator *modIt = dynamic_cast<ModifyablePointCloudIterator*>(newIt);

	AAT NORMAL = pc->getAAT( "normal" );
	AAT POSITION = pc->getAAT( "position" );
	
	// CL 2014/11/08: to handle pc that has a origin channel
	AAT ORIGIN;
	bool flag_origin = pc->providesAttribute("origin");
	if (flag_origin){
		ORIGIN = pc->getAAT("origin");
	}

	while ( !basicIt->atEnd() ){
		Vector3f normal = basicIt->get3f( NORMAL );
		Vector3f position = basicIt->get3f(POSITION);
		//if((getScene()->getRootState()->staticState->camera.getPosition()*normal)<0.0f){
		if (flag_origin){
			Vector3f ori = basicIt->get3f(ORIGIN);
			position = position - ori;
		}
		if((position*normal)>0.0f){
				modIt->set3f( NORMAL, -normal );
		}
		basicIt->next();
	}
	delete newIt;
}