//---------------------------------------------------------------------------
#include "StdAfx.h"

//---------------------------------------------------------------------------
#include "PCCTriangleMeshSampler.h"
//---------------------------------------------------------------------------
#include "PointCloud.h"
#include "PointCloudTools.h"
#include "TriangleMeshPointCloud.h"
#include "UnstructuredInCoreTriangleMesh.h"
#include "VertexDescriptor.h"
#include "NumericalClassProperty.h"
#include "StringClassProperty.h"
#include "PointSet.h"
#include "Random.h"
#include "SGObjectNode.h"
#include "MathTools.h"
#include "ProgressWindow.h"
#include "ClassMethods.h"
#ifndef _DEBUG
#include "PointSet.inline.h"
#endif
#include "VertexArray.h"
//---------------------------------------------------------------------------
 
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------


MetaClass* PCCTriangleMeshSampler::PCCTriangleMeshSamplerMetaClass = NULL;

const MetaClass* PCCTriangleMeshSampler::getInstanceClass() const {
   return getClass();
}

Object* PCCTriangleMeshSampler::newInstance() {
   return new PCCTriangleMeshSampler;
}

const MetaClass* PCCTriangleMeshSampler::getClass() {
   if (PCCTriangleMeshSamplerMetaClass==NULL) {
      PCCTriangleMeshSamplerMetaClass = new MetaClass("PCCTriangleMeshSampler", "no description.",
                                  newInstance, sizeof(PCCTriangleMeshSampler));
   }
   return PCCTriangleMeshSamplerMetaClass;
}

void PCCTriangleMeshSampler::init(const MetaClass *parentClass) {
   ((MetaClass*)getClass())->init(parentClass, MetaClass::FIRST_VERSION + 4);
   PCCTriangleMeshSamplerMetaClass->copyParentProperties();

	BEGIN_CLASS_INIT( PCCTriangleMeshSampler )
	ADD_STRING_PROP(sourceNode,0)
	ADD_STRING_PROP(destNode,0)
	ADD_CARD64_PROP(expNumPoints,0)
	ADD_BOOLEAN_PROP(createTangentDiscs,0)
	ADD_FLOAT32_PROP(tangentDiscOversampling,0)
	ADD_BOOLEAN_PROP(surfaceNormalSmoothing,2);
	ADD_FLOAT32_PROP(sampleSpacing,3);
	ADD_INT32_PROP(copyDataFromVertex, 4);
	ADD_NOARGS_METHOD(sampleMeshPoission);
}

void PCCTriangleMeshSampler::shutdown() {
   delete PCCTriangleMeshSamplerMetaClass;
}


// ----


void X4::sampleMeshRandomly(Scene * scene, std::string destID, SceneObject *sourceMesh, SceneObject *destPointCloud, card64 numPoints, bool createTangentDiscs, 
										float32 tangentDiscOversampling, bool surfaceNormalSmoothing) {

   // scene object types...
   TriangleMeshPointCloud *sourceAsMesh = dynamic_cast<TriangleMeshPointCloud*>(sourceMesh);
   if (!sourceAsMesh) {
      throw X4Exception("X4::sampleMeshRandomly() - source is no TriangleMeshPointCloud.");
   }
   PointCloud *sourceAsPC = dynamic_cast<PointCloud*>(sourceMesh);

	if(!sourceAsMesh) {
      throw X4Exception("X4::sampleMeshRandomly() - source is no PointCloud.");
   }
	
	x4Assert(sourceAsPC);
   PointCloud *destAsPC = dynamic_cast<PointCloud*>(destPointCloud);
   if (!destAsPC) {
      throw X4Exception("X4::sampleMeshRandomly() - dest is no PointCloud.");
   }

	// append necessary fields to destPointCloud
	if( !destAsPC->getDescr()->equal( sourceAsPC->getDescr() ) )
	{
		const VertexDescriptor * vd = sourceAsPC->getDescr();
		for( card32 i=0;i<vd->getNumAttributes();i++ )
		{
			checkAttribute( scene, destID, vd->getAttribute(i).name, vd->getAttribute(i).numComponents, vd->getAttribute(i).dataFormat );
		}

	   SceneGraphNode *destN = getSceneGraphNodeByName(scene, destID);
	   SGObjectNode *destNO = dynamic_cast<SGObjectNode*>(destN);
		destPointCloud = destNO->getSceneObject();
		destAsPC = dynamic_cast<PointCloud*>(destNO->getSceneObject());
	}


   // iterators...
   SceneObjectIterator *sourceIt = sourceMesh->createIterator(SOIT::CAP_BASIC_PC);
   BasicPointCloudIterator *sourceIt_basic = dynamic_cast<BasicPointCloudIterator*>(sourceIt);
   if (sourceIt == NULL || sourceIt_basic == NULL) {
      delete sourceIt;
      throw X4Exception("X4::sampleMeshRandomly() - source is no basic point cloud.");
   }

   SceneObjectIterator *sourceMeshIt = sourceMesh->createIterator(SOIT::CAP_BASIC_PC | SOIT::CAP_TRIANGLE_MESH);
   BasicPointCloudIterator *sourceMeshIt_basic = dynamic_cast<BasicPointCloudIterator*>(sourceMeshIt);
   if (sourceMeshIt == NULL) {
      delete sourceIt; delete sourceMeshIt;
      throw X4Exception("X4::sampleMeshRandomly() - source is no basic mesh.");
   }

   SceneObjectIterator *destIt = destPointCloud->createIterator(SOIT::CAP_DYNAMIC_PC);
   DynamicPointCloudIterator *destIt_dyn = dynamic_cast<DynamicPointCloudIterator*>(destIt);
   if (destIt == NULL) {
      delete destIt; delete sourceMeshIt; delete sourceIt;
      throw X4Exception("X4::sampleMeshRandomly() - dest is not a dynamic point cloud.");
   }



   // vertex descriptors, AATS...
   const VertexDescriptor *sourceVertexDescr = sourceAsPC->getDescr();
   const VertexDescriptor *destVertexDescr = destAsPC->getDescr();

   if (!sourceVertexDescr->equal(destVertexDescr)) {
      delete sourceIt; delete sourceMeshIt; delete destIt;
      throw X4Exception("X4::sampleMeshRandomly() - source and dest VertexDescriptors do not match.");
   }
   if (!sourceAsPC->providesAttribute("position")) {
      delete sourceIt; delete sourceMeshIt; delete destIt;
      throw X4Exception("X4::sampleMeshRandomly() - source does not have position channel.");
   }
   AAT SOURCE_POS = sourceAsPC->getAAT("position");
   AAT TANGENT_U;
   if (destAsPC->providesAttribute("tangent_u")) {
      TANGENT_U = destAsPC->getAAT("tangent_u");
   } else {
      if (createTangentDiscs) throw X4Exception("X4::sampleMeshRandomly() - cannot create tangent discs, attribute channel is missing.");
   }
   AAT TANGENT_V;
   if (destAsPC->providesAttribute("tangent_v")) {
      TANGENT_V = destAsPC->getAAT("tangent_v");
   } else {
      if (createTangentDiscs) throw X4Exception("X4::sampleMeshRandomly() - cannot create tangent discs, attribute channel is missing.");
   }
	AAT normalAAT;
	if( surfaceNormalSmoothing )
	{
		normalAAT = sourceAsPC->getAAT("normal");
	}

   if (!sourceAsMesh->getTriangleDescr()->providesAttribute("index")) {
      delete sourceIt; delete sourceMeshIt; delete destIt;
      throw X4Exception("X4::sampleMeshRandomly() - source mesh pc does not provide index channel.");
   }
   AAT SOURCE_INDEX = sourceAsMesh->getTriangleDescr()->getToken("index");

   // lock vertices...
   const PointSet *vertices = sourceIt_basic->lockBlockRead();

   
   // setup progress window
	progressWindow->pushStep(true, "area calc", 0.5f);
	progressWindow->pushStep(true, "sampling", 0.5f );
	
	// step 1: calculate overall area...
   float64 area = 0;
	card32 numTriangles = sourceAsMesh->getNumTriangles();
	card32 actTriangle = 0;
	bool aborted = false;
   while (!sourceMeshIt_basic->atEnd() && !aborted) {
      Vector3i indices = sourceMeshIt_basic->get3i(SOURCE_INDEX);
      Vector3f pos[3];
      pos[0] = vertices->get3f(indices[0], SOURCE_POS);
      pos[1] = vertices->get3f(indices[1], SOURCE_POS);
      pos[2] = vertices->get3f(indices[2], SOURCE_POS);

      float32 thisarea = 0.5f*norm((pos[1]-pos[0]).crossProduct(pos[2]-pos[0]));

      area += (float64)thisarea;

      sourceMeshIt_basic->next();

		// progress
		aborted = !progressWindow->progress( (float)actTriangle++ / (float)numTriangles * 100.0f );

   }
   sourceMeshIt_basic->reset();
	progressWindow->popStep( &aborted );

   float64 areaPerPoint = area / numPoints;
   float32 tangentSize = sqrt(areaPerPoint * tangentDiscOversampling);


	// step 2: sampling...
   char *destBuffer = new char[destVertexDescr->getSize()];
   //typedef const char *constCharPtr;
   const char **vbuff = new const char*[3];
	card32 actPoint = 0;
   while (!sourceMeshIt_basic->atEnd() && !aborted) {
      Vector3i indices = sourceMeshIt_basic->get3i(SOURCE_INDEX);
      Vector3f pos[3];
      pos[0] = vertices->get3f(indices[0], SOURCE_POS);
      pos[1] = vertices->get3f(indices[1], SOURCE_POS);
      pos[2] = vertices->get3f(indices[2], SOURCE_POS);

      Vector3f tNormal = (pos[1]-pos[0]).crossProduct(pos[2]-pos[0]);
      float64 thisarea = 0.5f*norm(tNormal);

      float64 numPoints = thisarea / areaPerPoint;

      card64 numPointsInt = floor(numPoints);
      float32 remPts = (float32)(numPoints - numPointsInt);

      if (remPts > rnd01()) numPointsInt++;

      for (card64 pnum = 0; pnum < numPointsInt; pnum++) {

         vbuff[0] = vertices->getDataPointer(indices[0]);
         vbuff[1] = vertices->getDataPointer(indices[1]);
         vbuff[2] = vertices->getDataPointer(indices[2]);


         float32 u = rnd01();
         float32 v = rnd01();
         if (u+v > 1) {
            float32 nu = 1.0f - v;
            v = 1.0f - u;
            u = nu;
         }

         float32 factors[3];
         factors[0] = u;
         factors[1] = v;
         factors[2] = 1.0f - u - v;

         sourceVertexDescr->interpolate(vbuff, factors, 3, destBuffer);

			if( surfaceNormalSmoothing )
			{
				Vector3f * posPtr = (Vector3f*)destBuffer;
				Vector3f * normalPtr = (Vector3f*)(&destBuffer[normalAAT.getOffset()]);
				Vector3f p[3];
				p[0] = *((Vector3f*)vbuff[0]);
				p[1] = *((Vector3f*)vbuff[1]);
				p[2] = *((Vector3f*)vbuff[2]);

				float dist1 = norm(p[0] - *posPtr);
				float dist2 = norm(p[1] - *posPtr);
				float dist3 = norm(p[2] - *posPtr);

				Vector3f normal[3];
				normal[0] = *((Vector3f*)(vbuff[0]+normalAAT.getOffset()));
				normal[1] = *((Vector3f*)(vbuff[1]+normalAAT.getOffset()));
				normal[2] = *((Vector3f*)(vbuff[2]+normalAAT.getOffset()));

				// implementation of PN curved triangles [Vlachos et al.]
				Vector3f b300 = p[0];
				Vector3f b030 = p[1];
				Vector3f b003 = p[2];
				float weight[3][3];
				for( unsigned i=0;i<3;i++ )
				{
					for( unsigned j=0;j<3;j++ )
					{
						weight[i][j] = (p[j]-p[i])*normal[i];
					}
				}
				Vector3f b210 = (p[0]*2.0f + p[1] - normal[0]*weight[0][1])/3.0f;
				Vector3f b120 = (p[1]*2.0f + p[0] - normal[1]*weight[1][0])/3.0f;
				Vector3f b021 = (p[1]*2.0f + p[2] - normal[1]*weight[1][2])/3.0f;
				Vector3f b012 = (p[2]*2.0f + p[1] - normal[2]*weight[2][1])/3.0f;

				Vector3f b102 = (p[2]*2.0f + p[0] - normal[2]*weight[2][0])/3.0f;
				Vector3f b201 = (p[0]*2.0f + p[2] - normal[0]*weight[0][2])/3.0f;

				Vector3f E = (b210 + b120 + b021 + b012 + b102 + b201) / 6.0f;
				Vector3f V = (p[0]+p[1]+p[2])/3.0f;

				Vector3f b111 = E + (E-V)/2.0f;

				float w = factors[2];
				float u_2 = u*u; float u_3 = u_2 * u;
				float v_2 = v*v; float v_3 = v_2 * v;
				float w_2 = w*w; float w_3 = w_2 * w;
				*posPtr = b300 * w_3 + b030 * u_3 + b003 * v_3
					+ b210*3.0f*w_2*u + b120*3.0f*w*u_2 + b201*3.0f*w_2*v
					+ b021*3.0f*u_2*v + b102*3.0f*w*v_2 + b012*3.0f*u*v_2
					+ b111*6.0f*w*u*v;

				Vector3f middleNormals[3];
				for( unsigned i=0;i<3;i++ )
				{
					unsigned j=(i+1)%3;
					float nweight = (p[j]-p[i])*(normal[i]+normal[j])*2.0f
						/ normQuad(p[j]-p[i]);
					middleNormals[i] = normal[i] + normal[j] - (p[j]-p[i])*nweight;
					normalize( middleNormals[i]);
				}

				*normalPtr = normal[0]*w_2 + normal[1]*u_2 + normal[0]*v_2
					+middleNormals[0]*w*u + middleNormals[1]*u*v + middleNormals[2]*w*v;
			}

         if (createTangentDiscs) {
            Vector3f normTNormal = tNormal / (thisarea * 2.0f);
            Vector3f tangent_u = makeOrthonormalVector(normTNormal);
            Vector3f tangent_v = tangent_u.crossProduct(normTNormal);
            tangent_u *= tangentSize;
            tangent_v *= tangentSize;
            TANGENT_U.setAttrib<float32,3>(destBuffer, tangent_u);
            TANGENT_V.setAttrib<float32,3>(destBuffer, tangent_v);
         }

         destIt_dyn->add(destBuffer);

      }

      sourceMeshIt_basic->next();

	}
   delete[] vbuff;
   delete[] destBuffer;

   sourceIt_basic->unlock();
	
	progressWindow->popSteps(  );
}


PCCTriangleMeshSampler::PCCTriangleMeshSampler() {
   expNumPoints = 100;
   createTangentDiscs = true;
   tangentDiscOversampling = 10;
	surfaceNormalSmoothing = false;
	sourceNode="root/pc1";
	destNode="root/pc2";
	sampleSpacing = 1.0f;
	copyDataFromVertex = -1;
}


void PCCTriangleMeshSampler::execute(Scene *scene) {
   SceneGraphNode *sourceN = getSceneGraphNodeByName(scene, sourceNode);
   SceneGraphNode *destN = getSceneGraphNodeByName(scene, destNode);
   SGObjectNode *sourceNO = dynamic_cast<SGObjectNode*>(sourceN);
   SGObjectNode *destNO = dynamic_cast<SGObjectNode*>(destN);
   if (!sourceNO) throw X4Exception("PCCTriangleMeshSampler::execute() - sourceNode not found.");
   if (!destNO) throw X4Exception("PCCTriangleMeshSampler::execute() - destNode not found.");;

   SceneObject *sourceSO = sourceNO->getSceneObject();
   SceneObject *destSO = destNO->getSceneObject();
   if (sourceSO == NULL || destSO == NULL) return;
   sampleMeshRandomly(scene, destNode, sourceSO, destSO, expNumPoints, createTangentDiscs, tangentDiscOversampling,surfaceNormalSmoothing);
}


UnstructuredInCorePointCloud * X4::PCCTriangleMeshSampler::sampleMeshPoisson( UnstructuredInCoreTriangleMesh * mesh, float sampleDist, int copyDataFromVertex )
{
	UnstructuredInCorePointCloud * ret= new UnstructuredInCorePointCloud;
	ret->clearAndSetup(mesh->getDescr(),0);

	const VertexDescriptor *sourceVertexDescr = mesh->getDescr();
	const VertexDescriptor *destVertexDescr = ret->getDescr();

	// oversample
	VertexArray va_points(mesh);
	VertexArray va_triangles(mesh->getTriangles());
	AAT posAAT = ret->getAAT("position");
	AAT normalAAT = ret->getAAT("normal");
	const char **vbuff = new const char*[3];
	char *destBuffer = new char[destVertexDescr->getSize()];
	for( mpcard i=0;i<va_triangles.getNumElements();i++)
	{
		Vector3i tri = va_triangles.getIndex3i(i);

		Vector3f p[3];
		for( mpcard j=0;j<3;j++ )
			p[j] = va_points.getPosition3f(tri[j]);

		vbuff[0] = mesh->getPointSet()->getDataPointer(tri[0]);
		vbuff[1] = mesh->getPointSet()->getDataPointer(tri[1]);
		vbuff[2] = mesh->getPointSet()->getDataPointer(tri[2]);

		// compute coordinate frame (longest edge is always dim 0)
		Vector3f longestEdge = p[1]-p[0];
		for( mpcard j=1;j<3;j++ )
		{
			if( norm(p[(j+1)%3]-p[j]) > norm(longestEdge) )
				longestEdge = p[(j+1)%3]-p[j];
		}

		Vector3f normal = normalize((p[1]-p[0]).crossProduct(p[2]-p[0]));
		Matrix3f frame;
		frame[0] = normalize(longestEdge);
		frame[1] = normalize(normal.crossProduct(longestEdge));
		frame[2] = normal;
		Matrix3f invFrame = frame.transpose();
		Vector3f frameCenter = p[0];

		// compute bounding box in frame
		BoundingBox2f bb;
		for( mpcard j=0;j<3;j++ )
		{
			Vector2f pf = shrink3To2(invFrame*(p[j]-frameCenter));
			if( j == 0)
				bb.lowerCorner = bb.upperCorner = pf;
			else
				bb.addPoint(pf);
		}
		bb.lowerCorner/=sampleDist;
		bb.upperCorner/=sampleDist;
						

		// setup pineda rasterization planes
		Plane3f edgePlanes[3];
		for( mpcard j=0;j<3;j++ )
		{
			edgePlanes[j].setNormal(normal.crossProduct(p[(j+1)%3]-p[j]));
			edgePlanes[j].setPoint(p[j]);
			if( edgePlanes[j].calculateSignedDistance(p[(j+2)%3]) < 0)
				edgePlanes[j].normal *= -1;
		}

		float area = 0.5f*norm((p[1]-p[0]).crossProduct(p[2]-p[0]));

		// special case: super flat triangle or very small triangle
		if( bb.getSideLength(1)/bb.getSideLength(0) < .1f || bb.getSideLength(0)*bb.getSideLength(1) < 4.0f )
		{
			// apply uniform sampling
			mpcard numSamples = ceil(area/sqr(sampleDist)*4.0f);
			for( mpcard j=0;j<numSamples;j++ )
			{
				float32 u = rnd01();
				float32 v = rnd01();
				if (u+v > 1) {
					float32 nu = 1.0f - v;
					v = 1.0f - u;
					u = nu;
				}
				float32 factors[3];
				factors[0] = u;
				factors[1] = v;
				factors[2] = 1.0f - u - v;

				mpcard pi = ret->getNumPoints();

				Vector3f sample = NULL_VECTOR3F;
				for( mpcard k=0;k<3;k++ )
					sample += p[k]*factors[k];

				if (copyDataFromVertex < 0) // interpolate
					sourceVertexDescr->interpolate(vbuff, factors, 3, destBuffer);
				else if (copyDataFromVertex >=0 && copyDataFromVertex <= 2) // just copy
					memcpy(destBuffer, vbuff[copyDataFromVertex], destVertexDescr->getSize());

				ret->getPointSet()->addPoint(destBuffer);
				ret->getPointSet()->set3f(pi,normalAAT,normal);				
				
				if (copyDataFromVertex >=0 && copyDataFromVertex <= 2)
					ret->getPointSet()->set3f(pi,posAAT,sample);
			}

			continue;
		}


		// perform pineda rasterization inside bounding box
		for( int x=floor(bb.lowerCorner[0]); x<ceil(bb.upperCorner[0]); x++ )
		{
			for( int y=floor(bb.lowerCorner[1]); y<ceil(bb.upperCorner[1]); y++ )
			{
				// transform std::vector to world space
				Vector3f v = makeVector3f(rnd01()+x,rnd01()+y,0)*sampleDist;
				//Vector3f v = makeVector3f(x,y,0)*sampleDist;
				v = frame*v+frameCenter;
				
				// test if std::vector is inside triangle
				bool skipPoint = false;
				float dist[3];
				for( mpcard j=0;j<3;j++ )
				{
					dist[j] = edgePlanes[j].calculateSignedDistance(v);
					if( dist[j] <0 )
					{
						skipPoint = true;
						break;
					}
				}

				if( skipPoint)
					continue;

				// compute factors (barycentric coordinates taking into account that surfaces of subtriangles are proportional to them)
				float factors[3] = {
					0.5f*norm((p[2]-p[1]).crossProduct(v-p[1])) / area,
					0.5f*norm((p[2]-p[0]).crossProduct(v-p[0])) / area,
					1
					/*0.5f*norm((p[1]-p[0]).crossProduct(v-p[0])) / area*/};
				factors[2] = 1.0f - factors[0] - factors[1];

				if (copyDataFromVertex < 0) // interpolate
					sourceVertexDescr->interpolate(vbuff, factors, 3, destBuffer);
				else if (copyDataFromVertex >=0 && copyDataFromVertex <= 2) // just copy
					memcpy(destBuffer, vbuff[copyDataFromVertex], destVertexDescr->getSize());

				mpcard pi = ret->getNumPoints();
				ret->getPointSet()->addPoint(destBuffer);
				ret->getPointSet()->set3f(pi,normalAAT,normal);

				if (copyDataFromVertex >=0 && copyDataFromVertex <= 2)
					ret->getPointSet()->set3f(pi,posAAT,v);
			}
		}
		
	}
	delete [] vbuff;
	delete [] destBuffer;

	// bilateral poisson disk sampling (normal direction)
	HierarchicalKNNIterator hIt(ret,32,NULL);
	hIt.setMaxDistanceToSeekPoint(sampleDist);
	LotteryDraw lottery;
	lottery.setup(ret->getNumPoints());
	std::vector<card32> visited(ret->getNumPoints(),0);
	VertexArray va_retpoints(ret);
	for( mpcard i=0;i<ret->getNumPoints();i++)
	{
		mpcard index = lottery.draw();
		if( visited[index])
			continue;
		visited[index] = 1;

		Vector3f normal = va_retpoints.getNormal3f(index);
		hIt.setSeekPointAndReset(va_retpoints.getPosition3f(index));
		hIt.next();
		while(!hIt.atEnd())
		{
			mpcard index2 = hIt.getCurrentPointIndex();
			if( va_retpoints.getNormal3f(index2)*normal > 0.9f )
			{
				visited[index2] = 2;	// delete
			}
			hIt.next();
		}
	}

	// delete all points marked with 2
	for( mpcard i=ret->getNumPoints();i>0;i-- )
	{
		if( visited[i-1] == 2)
			ret->getPointSet()->deletePoint(i-1);
	}

	ret->clearAttachments();

	return ret;
}

void X4::PCCTriangleMeshSampler::sampleMeshPoission()
{
	SGObjectNode *sourceNO = dynamic_cast<SGObjectNode*>(getSceneGraphNodeByName(scene, sourceNode));
	if (!sourceNO) throw X4Exception("PCCTriangleMeshSampler::sampleMeshPoission() - sourceNode mesh not found.");

	UnstructuredInCoreTriangleMesh * mesh = dynamic_cast<UnstructuredInCoreTriangleMesh*>(sourceNO->getSceneObject());
	if( !mesh )
		throw X4Exception("PCCTriangleMeshSampler::sampleMeshPoission() - source node is not a mesh!");

	UnstructuredInCorePointCloud * ret = PCCTriangleMeshSampler::sampleMeshPoisson(mesh,sampleSpacing,copyDataFromVertex);
	addPointCloud(getScene(),ret,"sampledMesh");
}