//---------------------------------------------------------------------------
#ifndef PCInteractionToolH
#define PCInteractionToolH
//---------------------------------------------------------------------------
#include "x3editor.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
#include "KeyboardController.h"
#include "MouseController.h"
#include "Tools.h"
#include "OnSceneChangeCallback.h"
#include <QWidget>
#include "LinearAlgebra.h"
#include "PointCloud.h"
#include "OperatorNoArgs.h"
//---------------------------------------------------------------------------


namespace X4 {

class GLContext;
class Scene;
class ProgressWindow;

class X3EDITOR_API PCInteractionToolSceneSettings : public AttachedData
{
	X4_CLASS( PCInteractionToolSceneSettings )	
public:
	PCInteractionToolSceneSettings(){}
	virtual ~PCInteractionToolSceneSettings() {}
};

class X3EDITOR_API PCInteractionTool : public Persistent, public KeyboardController, public MouseController 
{
	X4_ABSTRACT_CLASS( PCInteractionTool )	

	 protected:
		 /// for compatibiltiy with threaded tools
		 virtual void execOperatorNoArgs(OperatorNoArgs *op);

    public:
        PCInteractionTool();
        virtual ~PCInteractionTool();

        // --- tool definition methods

        /// called once to connect tool with scene and editor
        /// scene: point cloud scene
        /// onChange: callback to notify editor of changes
        /// doRedraw: callback to cause redraw of current editor view of scene
        /// sceneEditorWidget: top level widget that contains the editor (used to attach child windows to the right parents, do not cast to SceneEditorWidget to access it directly)
		/// toolWidget: Widget used to represent the tool (usually the widget created by tool's class editor)
        void connectToScene( Scene *scene, OnSceneChangeCallback *onChange, QWidget *sceneEditorWidget, QWidget* toolWidget );

        /// tool has been choosen (again) by user, re-initialize tool
        virtual void activateTool() {}
        /// tool has been deactivated by user (other tool has been choosen)
        virtual void deactivateTool();
        /// override this method to provide user feedback, e.g. handles / bounding boxes etc... (called after main rendering pass)
        virtual void glDrawTool(GLContext *glContext) {forget(glContext);}
        /// override this method to peform operations _before_ the main rendering pass (mw - added: Nov 4th 2011)
        virtual void beforeGlRendering(GLContext *glContext) {forget(glContext);}
        /// determines whether the main application should draw a 3d view of the scene (before glDrawTool)
        /// default: true - if false, all visualization must be performed in glDrawTool
        virtual bool drawGLSceneRendering() {return true;}
        /// override this method and return true if the tool should operate in camera mode by default
        virtual bool forceCamByDefault();

        // ---

        inline Scene *getScene() {return scene;}

        /// resize pci-intern screensize
        void areaResize_base(card32 w, card32 h) {screenSizeX=w;screenSizeY=h;}

        virtual void mouseWheelRotated(int32 rotatedDelta, Modifiers modifiersState ){};

		/// get widget associated with this tool (widget might be created by a ClassEditor)
		inline QWidget* getToolWidget() { return toolWidget; }

		/// Get tool's progress window. 
		ProgressWindow* getProgressWindow() { return toolProgressWindow; }

		bool autoSaveAndRestoreToolParameters;
		virtual void loadPCISettings();
		virtual void savePCISettings();
		

		bool autoTriggerFullLayout;  /// if set to true, then auto layout (debug window + full screen) will be triggered automatically when scene connected

		/// Load(or create and attach) new settings object from the scene with the given name. If settingsClass is not given, meta class is retrieved from the given name
		virtual PCInteractionToolSceneSettings* getOrCreateSceneSettings(Scene* scene, const std::string& name, const MetaClass* settingsClass = NULL);
		inline  PCInteractionToolSceneSettings* getOrCreateSceneSettings(Scene* scene, const MetaClass* settingsClass) { return getOrCreateSceneSettings(scene, settingsClass->getClassName(), settingsClass); }

		template<class T>
		T* getOrCreateSceneSettings(Scene* scene, const std::string& name = std::string()) { return dynamic_cast<T*>(getOrCreateSceneSettings(scene, name.length() ? name : T::getClass()->getClassName(), T::getClass())); }

    protected:
        Scene *scene;
        OnSceneChangeCallback *onChange;
        QWidget *sceneEditorWidget;
		QWidget *toolWidget;
        card32 screenSizeX,screenSizeY;
		ProgressWindow* toolProgressWindow;

		/// method for derived classes called when connected to a scene
		virtual void connectToSceneImpl( Scene *scene, OnSceneChangeCallback *onChange, QWidget *sceneEditorWidget, QWidget* toolWidget = NULL ){}
        /// call to notify editor concerning changes
        virtual void sceneChanged(PCInteractionToolChanges whatChange) {if (onChange != NULL) {onChange->changed(whatChange);}}
        /// call to cause immediate redraw
        virtual void redrawScene() {if (onChange != NULL) {onChange->changed(PCI_CHANGE_VIEW);}}
		/// call to cause redraw on the next event loop
		virtual void updateSceneView() {if (onChange != NULL) {onChange->changed(PCI_UPDATE_VIEW_SIGNAL);}}
        /// call to write a frame as bitmap to disc (if activated in main editor, otherwise nothing happens)
        virtual void writeFrameToDisk() {if (onChange != NULL) {onChange->changed(PCI_WRITE_FRAME);}}

        /// returns true and the corresponding point, point cloud if cursor points on point cloud
        bool getIntersectionFromCursor( int x, int y, Vector3f &retPoint, std::string &pcName, PointCloud ** pc );

		/// guess click point (zbuffer readback around x,y with radius rad -> median of depth...)
		bool guessClickPoint( int x, int y, unsigned rad, Vector3f &retPoint);

		/// estimates oobb in world space from a rectangular area (zbuffer read within area, median of depth for outlier removal,...)
		/// returns true if succeeded
		bool estimateOOBBFromScreenArea( const Vector2i &x1, const Vector2i &x2, BoundingBox3f &bb, Matrix4f &worldToBB, float depthOutlierTolerance=0.1f, Vector3f * firstAxis=NULL);
		 

        /// returns a std::vector of point cloud from the scene
        std::vector<PointCloud*> getAllPointCloudsFromScene();

        /// returns all selected point clouds from the scene
        std::vector<PointCloud*> getSelectedPointCloudsFromScene();
public:
		template<typename Base> 
		std::vector<Base*> getSelectedPointCloudFromSceneAs()
		{ 
			std::vector<PointCloud*> pcs = getSelectedPointCloudsFromScene(); 
			std::vector<Base*> ret;
			ret.reserve(pcs.size());
			for( int i1 = 0; i1 < pcs.size(); ++ i1 ) {
				Base* b = dynamic_cast<Base*>(pcs[i1]);
				if( b ) ret.push_back(b);
			}
			return ret;
		}

		/// returns selected point cloud (if no/more than one is selected, method returns NULL)
		PointCloud* getSingleSelectedPointCloudFromScene();

		template<typename Base> 
		Base* getSingleSelectedPointCloudFromSceneAs() { return dynamic_cast<Base*>( getSingleSelectedPointCloudFromScene() ); }

        /// pass the mouse event on to the camera controller
        virtual void mouseDownPassToCam(int32 x, int32 y, MouseButtons buttonsState, Modifiers modifiersState);
        /// pass the mouse event on to the camera controller
        virtual void mouseMovedPassToCam(int32 x, int32 y);
        /// pass the mouse event on to the camera controller
        virtual void mouseUpPassToCam(int32 x, int32 y, MouseButtons buttonsState, Modifiers modifiersState);
        /// pass the mouse event on to the camera controller
        virtual void areaResizePassToCam(card32 width, card32 height);
        /// pass the mouse event on to the camera controller
        virtual void mouseWheelRotatedPassToCam(int32 rotatedDelta, Modifiers modifiersState);

        /// pass the keyboard event on to the camera controller
        virtual void keyDownPassToCam(GeneralKey key);
        /// pass the keyboard event on to the camera controller
        virtual void keyUpPassToCam(GeneralKey key);


};

}

#endif
