//---------------------------------------------------------------------------
#ifndef PCInteractionTool_ThreadedH
#define PCInteractionTool_ThreadedH
//---------------------------------------------------------------------------
#include "PCInteractionTool.h"
#include "ClassMethods.h"
#include "OperatorNoArgs.h"
//---------------------------------------------------------------------------
#include <qtcore/QThread>
#include <qtcore/QMetaType>

namespace X4
{

	template<class ResultType>
	struct OperatorNoArgsLockedResult : OperatorNoArgs
	{
		QMutex __mutex;
		QWaitCondition __cond;

		ResultType __result;

		//! Get result from the operator. This will block until other thread set the result
		inline const ResultType& get()
		{
			__mutex.lock(); 
			__cond.wait(&__mutex);
			__mutex.unlock();
			return __result;
		}

		//! Set result. This will unblock the thread waiting for the get result
		inline void set(const ResultType& res) { __result = res; __cond.wakeAll();}
		virtual void operator()() {};		
	};

    /**
    * Same as PCInterationTool, however allows to specify certain methods to be executed in a thread.
    * In any class derived from this class, use macro
    * ADD_NOARGS_METHOD_THREADED (...)
    * to specify methods which you would like to be executed in parallel.
    * NOTE: Currently only one method can be executed concurrent to the main program!!!
    * TODO: Multiple methods to be executed in parallel
    **/
    class X3EDITOR_API PCInteractionToolThreaded :  public QThread, public PCInteractionTool
    {
        Q_OBJECT

        X4_ABSTRACT_CLASS( PCInteractionToolThreaded )	

	public:

	signals:
		/// MW (24.05.2015) - evil hack to make threaded and non-threaded version compatible (so that we can switch easily)
		void execOperatorNoArgsSignal(OperatorNoArgs*);

	protected:
		/// MW (24.05.2015) - evil hack to make threaded and non-threaded version compatible (so that we can switch easily)
		virtual void execOperatorNoArgs(OperatorNoArgs *op) {execOperatorNoArgsSignal(op);}


    private slots:

        // -- QThread interface
        void threadStarted();
        void threadTerminated();
        void threadFinished();
		
		//! Execute given lambda in the main thread, except of in the running thread
		void slotExecCallback(OperatorNoArgs*);

    private:
        bool _isActive;
		card32 _randomSeed;

        // current method which is executed in thread
        ClassMethod* _method;

        // make sure we don't run thread if it is already running
        QMutex _mutex;

        void run();

    protected:
        /// run a dispatcher to run the thread in a special context
        void dispatchMethod(ClassMethod*);

        /// implement any user reaction to the activation of the tool here
        virtual void activateToolImplementation() {}

        /// user reaction if tool get deactivated
        virtual void deactivateToolImplementation() {}

        /// user implemented opengl rendering routines
        virtual void glDrawToolImplementation ( GLContext *glContext ) {};

        /// check if the tool is currently active or not
        inline bool isActive() { return _isActive; }

        /// callback which is executed whenever the execution of the according method starts (just before the main execution)
        virtual void onExecutionStart() {}

        /// called back from base class when execution is terminated (use this for gracefully cleaning up of resources)
        virtual void onExecutionStop() {}

    public:
        PCInteractionToolThreaded();
        virtual ~PCInteractionToolThreaded();

        // --- tool definition methods

        /// tool has been chosen (again) by user, re-initialize tool
        void activateTool();

        /// tool has been deactivated by user (other tool has been chosen)
        /// TODO: implement a differentiation if thread has to be terminated or not during tool changes
        void deactivateTool();

        /// tool requires to redraw, make context current and call implementation method to draw
        void glDrawTool( GLContext *glContext );
    };


    /// use this macro to create methods which will be executed in a separate thread
    /// this will be done by creating a dispatcher method to call the specified method (works only inside of PCInteractionToolThreaded)
    #define ADD_NOARGS_METHOD_THREADED( method ) ADD_NOARGS_METHOD_DISPATCHER (dispatchMethod, method);
	#define ADD_NOARGS_METHOD_THREADED_DESC( method, desc ) ADD_NOARGS_METHOD_DISPATCHER_DESC (dispatchMethod, method, desc);

	Q_DECLARE_METATYPE(OperatorNoArgs*);

}

#endif
