//---------------------------------------------------------------------------
#include "stdafx.h"
//---------------------------------------------------------------------------
#include "PointCloudImporters.h"
#include "LinearAlgebra.h"
#include "LinearAlgebra.inline.h"
#include "PointSet.h"
#include "PointSet.inline.h"
#include "Tools.h"
#include "DebugOutput.h"
#include <fstream>
//---------------------------------------------------------------------------



using namespace X4;

static void skipToEOL(std::ifstream &in) {
   bool cont = true;
   while (!in.eof() && cont) {
      char c;
      in.get(c);
      if (c == 10 || c == 13) {
         cont = false;
      }
   }
}

struct ZFPoint {
   Vector3f pos;
   StaticVector<float32, 1> intensity;
   Vector2f param;
   StaticVector<int32, 1> flags;
};


PointSet *X4::doZFImport(const char *zfFile) {

   card32 maxParamU = 0;
   card32 maxParamV = 0;

   bool mod2OddFoundU = false;
   bool mod2OddFoundV = false;

   std::vector<ZFPoint> points;

   card32 numPoints = 0;

   //ShowProgress::setProgress(0.0f, "ZF import: checking file size...");
   { // first: count points
      std::ifstream in(zfFile);

      skipToEOL(in); // skip comments
      while (!in.eof()) {
         ZFPoint p;
         p.flags[0] = 0;
         card32 paramU;
         card32 paramV;
         in >> p.pos[0];
         in >> p.pos[1];
         in >> p.pos[2];
         in >> p.intensity[0];
         in >> paramU;
         in >> paramV;
         p.param[0] = paramU;
         p.param[1] = paramV;

         if (maxParamU < paramU) {
            maxParamU = paramU;
         }
         if (maxParamV < paramV) {
            maxParamV = paramV;
         }
         if (paramU % 2 == 1) {
            mod2OddFoundU = true;
         }
         if (paramV % 2 == 1) {
            mod2OddFoundV = true;
         }

         numPoints++;
      }
   }

   points.resize(numPoints);

   //ShowProgress::setProgress(0.0f, "ZF import: reading...");
   { // second: store points
      card32 pos = 0;
      std::ifstream in(zfFile);

      skipToEOL(in); // skip comments
      while (!in.eof()) {
         ZFPoint p;
         p.flags[0] = 0;
         card32 paramU;
         card32 paramV;
         in >> p.pos[0];
         in >> p.pos[1];
         in >> p.pos[2];
         in >> p.intensity[0];
         in >> paramU;
         in >> paramV;
         p.param[0] = paramU;
         p.param[1] = paramV;

         if (maxParamU < paramU) {
            maxParamU = paramU;
         }
         if (maxParamV < paramV) {
            maxParamV = paramV;
         }
         if (paramU % 2 == 1) {
            mod2OddFoundU = true;
         }
         if (paramV % 2 == 1) {
            mod2OddFoundV = true;
         }

         points[pos] = p;
         pos++;
         // ShowProgress::setProgress((float32)pos/(float32)numPoints, "ZF import: reading...");
     }
   }

   PointSet *pc = new PointSet();
   VertexDescriptor componentDescription;
   componentDescription.pushAttrib(mVAD("position", 3, VAD::DATA_FORMAT_FLOAT32));
   componentDescription.pushAttrib(mVAD("intensity", 1, VAD::DATA_FORMAT_FLOAT32));
   componentDescription.pushAttrib(mVAD("param", 2, VAD::DATA_FORMAT_FLOAT32));
   componentDescription.pushAttrib(mVAD("flags", 1, VAD::DATA_FORMAT_INT32));

   const AAT POSITION = componentDescription.getToken("position");
   const AAT INTENSITY = componentDescription.getToken("intensity");
   const AAT PARAM = componentDescription.getToken("param");
   const AAT FLAGS = componentDescription.getToken("flags");

   card32 width = maxParamU;
   card32 height = maxParamV;
   if (!mod2OddFoundU) width /=2;
   if (!mod2OddFoundV) height /=2;
   width++;
   height++; // count from 0
   pc->clearAndSetup(width, height, &componentDescription);

   ZFPoint NULL_POINT;
   NULL_POINT.pos = NULL_VECTOR3F;
   NULL_POINT.intensity[0] = 0.0f;
   NULL_POINT.param = NULL_VECTOR2F;
   NULL_POINT.flags[0] = PF_FLAG_INVALID;

   for (card32 i=0; i<width*height; i++) {
      pc->set<float32, 3>(i, POSITION, NULL_POINT.pos);
      pc->set<float32, 1>(i, INTENSITY, NULL_POINT.intensity);
      pc->set<int32, 1>(i, FLAGS, NULL_POINT.flags);
      pc->set<float32, 2>(i, PARAM, NULL_POINT.param);
   }
   for (card32 i=0; i<numPoints; i++) {
      card32 u = points[i].param[0];
      card32 v = points[i].param[1];
      if (!mod2OddFoundU) {
         u /= 2;
      }
      if (!mod2OddFoundV) {
         v /= 2;
      }
      pc->set<float32, 3>(u, v, POSITION, points[i].pos);
      pc->set<float32, 1>(u, v, INTENSITY, points[i].intensity);
      pc->set<int32, 1>(u, v, FLAGS, points[i].flags);
      pc->set<float32, 2>(u, v, PARAM, points[i].param);
   }

   return pc;
}


