//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "HoughTransformProcessing.h"
//---------------------------------------------------------------------------
#include "NumericalClassProperty.h"
#include "BooleanClassProperty.h"
#include "RegularData.inline.h"
#include "MathTools.h"
#include "MathTools.inline.h"
//---------------------------------------------------------------------------
#include <set>
#include <deque>
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------




IMPLEMENT_X4_ABSTRACT_CLASS( HoughTransformCornerDetector ,0 )
{
	BEGIN_CLASS_INIT( HoughTransformCornerDetector );
}

IMPLEMENT_X4_CLASS( HTCornerDetectorFiniteDiff, 0 )
{
	BEGIN_CLASS_INIT( HTCornerDetectorFiniteDiff );
	ADD_FLOAT32_PROP( epsGradient, 0);
	ADD_FLOAT32_PROP( epsHessian, 0);
}

IMPLEMENT_X4_CLASS( HTPickMaximum, 0 )
{
	BEGIN_CLASS_INIT( HTPickMaximum );
	ADD_CARD32_PROP( numCorners, 0 );
}

IMPLEMENT_X4_CLASS( HTConvexSegmentation, 0 )
{
	BEGIN_CLASS_INIT( HTConvexSegmentation );
	ADD_CARD32_PROP( numCorners, 0);
	ADD_CARD32_PROP( maxSegmentSize, 0 );
}



// -----


HTCornerDetectorFiniteDiff::HTCornerDetectorFiniteDiff() {
	epsGradient = 0.01f;
	epsHessian = 0.01f;
}

static Vector2f compute2dHessianEigenvalues(HTRegularDataAdaptor &data, int rot, int dist, float32 *normGrad = NULL) {
	Vector2f grad;
	grad[0] = data.get(rot+1, dist) - data.get(rot-1, dist);
	grad[1] = data.get(rot, dist+1) - data.get(rot, dist-1);
	float32 hessian00 = data.get(rot-1, dist) -2*data.get(rot, dist) + data.get(rot+1, dist);
	float32 hessian11 = data.get(rot, dist-1) -2*data.get(rot, dist) + data.get(rot, dist+1);
	float32 hessian01 = data.get(rot-1, dist-1) + data.get(rot+1, dist+1)
					  - data.get(rot+1, dist-1) - data.get(rot-1, dist+1);
	hessian01 /= 4; // different discretization

	Vector2f lambda;
	eigenValuesSym2f(hessian00, hessian01, hessian11, lambda[0], lambda[1]);
	if (normGrad) {
		*normGrad = norm(grad);
	}
	return lambda;
}

void HTCornerDetectorFiniteDiff::detectCorners(X4::RegularDataF &ht, std::vector<HTCorner> &result) {
	HTRegularDataAdaptor data(&ht);
	
	if (ht.getExtends().getDim() == 2) {
		x4Assert(data.getHeight() > 2);
		// 2d ---
		for (int32 dist=1; dist<data.getHeight()-1; dist++) {
			for (int32 rot=0; rot<data.getWidth(); rot++) {
				float32 normGrad;
				Vector2f lambda = compute2dHessianEigenvalues(data, rot, dist, &normGrad);
				if (lambda[0] < -epsHessian && lambda[1] < -epsHessian && normGrad < epsGradient) {
					result.push_back(HTCorner(makeVector3f(rot, dist, 0), data.get(rot, dist)));
				}
			}
		}
	} else {
		x4Assert(false);/*
		// 3d ---
		for (int32 dist=1; dist<data.getDepth()-1; dist++) {
			for (int32 rlong=0; rlong<data.getHeight(); rlong++) {
				for (int32 rlat=0; rlat<data.getWidth(); rlat++) {
				}
			}
		}*/
	}
}


// --

HTPickMaximum::HTPickMaximum() {
	numCorners = 10;
}
	
void HTPickMaximum::detectCorners(RegularDataF &ht, std::vector<HTCorner> &result) {
	HTRegularDataAdaptor data(&ht);
	std::vector<HTCorner> temp;
	const float32 *pos = ht.dataPtr();
 	for (int32 z=0; z<data.getDepth(); z++) {
		for (int32 y=0; y<data.getHeight(); y++) {
			for (int32 x=0; x<data.getWidth(); x++) {
				temp.push_back(HTCorner(makeVector3f(x,y,z), *pos));
				++pos;
			}
		}
	}
	sort(temp.begin(), temp.end());
	size_t pickNum = (size_t)numCorners;
	if (pickNum > temp.size()) pickNum = temp.size();
	for (size_t i=temp.size()-1; i>= temp.size()-pickNum; i--) {
		result.push_back(temp[i]);
	}
}

// --

HTConvexSegmentation::HTConvexSegmentation() {
	numCorners = 10;
	maxSegmentSize = 10;
}

void HTConvexSegmentation::detectCorners(X4::RegularDataF &ht, std::vector<HTCorner> &result) {
	set<Entry> points;
	HTRegularDataAdaptor data(&ht);
	bool in2d = (ht.getDim() == 2);
	std::vector<HTCorner> temp;
	const float32 *pos = ht.dataPtr();
 	for (int32 z=0; z<data.getDepth(); z++) {
		for (int32 y=0; y<data.getHeight(); y++) {
			for (int32 x=0; x<data.getWidth(); x++) {
				Vector3f lambda;
				if (in2d) {
					Vector2f lambda2 = compute2dHessianEigenvalues(data, x, y);
					lambda = makeVector3f(lambda2[0], lambda2[1], 0);
				} else {
					x4Assert(false);
					return;
				}
				if (lambda[0] <=0 && lambda[1] <=0 && lambda[2] <=0) {
					points.insert(Entry(x,y,z, *pos));
				}
				++pos;
			}
		}
	}
	
	card32 numSelected = 0;
	int middle_x;
	int middle_y;
	do {
	
		set<Entry>::iterator lastPos = points.end(); lastPos--;
		/// initialization
		std::deque<Entry> regionGrowingQueue;
		{
			Entry e = *lastPos;
			result.push_back(HTCorner(makeVector3f(e.x, e.y, e.z), e.magnitude));
		
			numSelected++;
			
			regionGrowingQueue.push_back(e);
			middle_x = e.x;
			middle_y = e.y;
		}
		while (!regionGrowingQueue.empty()) {
			Entry e = regionGrowingQueue.front();
			regionGrowingQueue.pop_front();
			set<Entry>::iterator newPos = points.find(e);
			if (newPos != points.end()) {
				points.erase(newPos);
				if (in2d) {
					if (abs(middle_x-e.x) <= (int)maxSegmentSize && abs(middle_y-e.y) <= (int)maxSegmentSize) {
						if (e.x > 0) {
							set<Entry>::iterator nextPos = points.find(Entry(e.x-1, e.y, e.z, data.get(e.x-1, e.y)));
							if (nextPos != points.end()) {
								regionGrowingQueue.push_back(*nextPos);
							}
						}			
						if (e.x < data.getWidth()-1) {
							set<Entry>::iterator nextPos = points.find(Entry(e.x+1, e.y, e.z, data.get(e.x+1, e.y)));
							if (nextPos != points.end()) {
								regionGrowingQueue.push_back(*nextPos);
							}
						}			
						if (e.y > 0) {
							set<Entry>::iterator nextPos = points.find(Entry(e.x, e.y-1, e.z, data.get(e.x, e.y-1)));
							if (nextPos != points.end()) {
								regionGrowingQueue.push_back(*nextPos);
							}
						}			
						if (e.y < data.getHeight()-1) {
							set<Entry>::iterator nextPos = points.find(Entry(e.x, e.y+1, e.z, data.get(e.x, e.y+1)));
							if (nextPos != points.end()) {
								regionGrowingQueue.push_back(*nextPos);
							}
						}			
					}
				} else {
					x4Assert(false);
					return;
				}
			}
		}
		
	} while (numSelected < numCorners && !points.empty());  
}



// --



HTRegularDataAdaptor::HTRegularDataAdaptor(RegularDataF *data) {
	this->data = data;
	if (data == NULL) {
		throw X4Exception("HTRegularDataAdaptor::HTRegularDataAdaptor() - data must not be NULL.");
	}
	card32 dim = data->getExtends().getDim();
	if (dim == 2) {
		width = data->getExtends(0);
		height = data->getExtends(1);
		depth = 1;
	} else if (dim == 3) {
		width = data->getExtends(0);
		height = data->getExtends(1);
		depth = data->getExtends(2);
	} else {
		throw X4Exception("HTRegularDataAdaptor::HTRegularDataAdaptor() - data must have dimension 2 or 3.");
	}
	dataPtr = data->dataPtr();
}


