//---------------------------------------------------------------------------
#ifndef HoughTransformProcessingH
#define HoughTransformProcessingH
//---------------------------------------------------------------------------
#include "x3editor2.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
#include "LinearAlgebra.h"
#include "DynamicLinearAlgebra.h"
#include "RegularData.h"
//---------------------------------------------------------------------------
#include <vector>
//---------------------------------------------------------------------------


namespace X4 {

class X3EDITOR2_API HTCorner {
 public:
	Vector3f pos;
	float32 magnitude;
	inline HTCorner() {}
	inline HTCorner(Vector3f p, float32 mag) {
		pos = p;
		magnitude = mag;
	}
	inline bool operator<(const HTCorner &other) const {
		return magnitude < other.magnitude;
	}
};

/// detects local maxima in hogh transform diagrams produces by "LocalModelStatistics".
/// abstract base class.
class X3EDITOR2_API HoughTransformCornerDetector : public Persistent {
	X4_ABSTRACT_CLASS(HoughTransformCornerDetector);
 public:
 	virtual void detectCorners(RegularDataF &ht, std::vector<HTCorner> &result) = 0;
};


/// simplest strategy ever: pick brightest points
class X3EDITOR2_API HTPickMaximum : public HoughTransformCornerDetector {
	X4_CLASS(HTPickMaximum);
 private:
	card32 numCorners;

 public:
	HTPickMaximum();
	
 	virtual void detectCorners(RegularDataF &ht, std::vector<HTCorner> &result);
};


/// implementation of ht-corner detector using simple finite differencing
class X3EDITOR2_API HTCornerDetectorFiniteDiff : public HoughTransformCornerDetector {
	X4_CLASS(HTCornerDetectorFiniteDiff);
 private:
	//card32 numCorners;
	float32 epsGradient;
	float32 epsHessian;
		
 public:
	HTCornerDetectorFiniteDiff();
	
 	virtual void detectCorners(RegularDataF &ht, std::vector<HTCorner> &result);
};


/// segments the diagram in convex areas, picks maximum in each area
class X3EDITOR2_API HTConvexSegmentation : public HoughTransformCornerDetector {
	X4_CLASS(HTConvexSegmentation);
 private:
	/// number of corners to be extracted
	card32 numCorners;
	/// maximum number of pixels belonging to one convex segment
	card32 maxSegmentSize;
		
	struct Entry {
		int32 x,y,z;
		float32 magnitude;

		inline Entry(int32 x=0, int32 y=0, int32 z=0, float32 magnitude=0) {
			this->x = x; 
			this->y = y; 
			this->z = z; 
			this->magnitude = magnitude;
		}

		inline bool operator<(const Entry &other) const {
			if (magnitude < other.magnitude) return true;
			if (magnitude > other.magnitude) return false;
			if (x < other.x) return true;
			if (x > other.x) return false;
			if (y < other.y) return true;
			if (y > other.y) return false;
			if (z < other.z) return true;
			return false;
		}
	};

 public:
	HTConvexSegmentation();
	
 	virtual void detectCorners(RegularDataF &ht, std::vector<HTCorner> &result);
};


/// helper class for easy access of ht's stored in RegularData objects,
/// provides the correct boundary conditions
class HTRegularDataAdaptor {
 private:
	RegularDataF *data;
	int32 width, height, depth;
	float32 *dataPtr;
 public:
	/// construct adaptor; data must be 2 or 3 dim
	HTRegularDataAdaptor(RegularDataF *data);
	
	/// order: [rot, dist] ~~ [width, height]
	inline float32 &get(int32 rot, int32 dist);
	/// order: [lat, long, dist] ~~ [width, height, depth]
	inline float32 &get(int32 rotLat, int32 rotLong, int32 dist);
	
	inline int32 getWidth() {return width;}
	inline int32 getHeight() {return height;}
	inline int32 getDepth() {return depth;}
};


inline float32 &HTRegularDataAdaptor::get(int32 rot, int32 dist) {
	if (rot < 0) rot += width;
	rot = rot % width;
	if (dist < 0) dist = 0;
	if (dist >= height) dist = height-1;
	return dataPtr[rot + dist*width];
}

inline float32 &HTRegularDataAdaptor::get(int32 rotLat, int32 rotLong, int32 dist) {
	x4Assert(rotLong > -height);
	x4Assert(rotLong < 2*height);
	
	if (rotLong < 0) {
		rotLong += height;
		rotLat = width/2 - rotLat;
	}
	if (rotLong > 0) {
		rotLong -= height;
		rotLat = width/2 - rotLat;
	}
	if (rotLat < 0) rotLat  += width;
	x4Assert(rotLat > 0);
	rotLat = rotLat % width;
	if (dist >= depth) dist = depth-1;
	if (dist < 0) dist = 0;
	
	return dataPtr[rotLat  +  rotLong * width  +  dist * width*height];
}

}


#endif