//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "LocalGeometryModelStatistics.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "NumericalClassProperty.h"
#include "BooleanClassProperty.h"
#include "RegularData.inline.h"
#include "ImageDisplayer.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------




IMPLEMENT_X4_ABSTRACT_CLASS( LocalGeometryModelStatistics ,0 )
{
	BEGIN_CLASS_INIT( LocalGeometryModelStatistics );
	ADD_CARD32_PROP( quantization, 0 );
	ADD_BOOLEAN_PROP( in2d, 0 );
}
									 
IMPLEMENT_X4_ABSTRACT_CLASS( LGMSPlaneModel ,0 )
{
	BEGIN_CLASS_INIT( LGMSPlaneModel )
	ADD_FLOAT32_PROP( maxDistance, 0 )
}

IMPLEMENT_X4_CLASS( LGMSLeastSquaresPlaneModel , 1 )
{
	BEGIN_CLASS_INIT( LGMSLeastSquaresPlaneModel )
	ADD_BOOLEAN_PROP( outputLogDensity, 0)
	ADD_FLOAT32_PROP( sigma, 0 )
	ADD_FLOAT32_PROP( cutOffValue, 1 )
	ADD_FLOAT32_PROP( exponent, 1 )
}

IMPLEMENT_X4_CLASS( LGMSKernelPlaneModel ,0 )
{
	BEGIN_CLASS_INIT( LGMSKernelPlaneModel )
	ADD_BOOLEAN_PROP( outputLogDensity, 0)
	ADD_FLOAT32_PROP( sigma, 0 )
	ADD_BOOLEAN_PROP( refit2ndOrder, 0 )
}

IMPLEMENT_X4_CLASS( LGMSKernelLineEntropyPlaneModel ,0 )
{
	BEGIN_CLASS_INIT( LGMSKernelLineEntropyPlaneModel )
	ADD_BOOLEAN_PROP( outputLogDensity, 0)
	ADD_FLOAT32_PROP( sigma, 0 )
	ADD_FLOAT32_PROP( scale, 0 )
	ADD_BOOLEAN_PROP( entropyWeighting, 0 )
	ADD_CARD32_PROP( resolution, 0 )
}

// ----


LocalGeometryModelStatistics::LocalGeometryModelStatistics() {
	quantization = 32;
	in2d = true;
	updateResultExtends();
}


void LocalGeometryModelStatistics::updateResultExtends() {
	if (in2d) {	
		resultExtends = DVectorI(2);
		resultExtends[0] = quantization;
		resultExtends[1] = quantization;
	} else {
		resultExtends = DVectorI(3);
		resultExtends[0] = quantization;
		resultExtends[1] = quantization;
		resultExtends[2] = quantization;
	}
}


void LocalGeometryModelStatistics::setQuantization(card32 q) {
	quantization = q;
	updateResultExtends();
}

void LocalGeometryModelStatistics::setIn2d(bool d) {
	in2d = d;
	updateResultExtends();
}


const DVectorI &LocalGeometryModelStatistics::getResultExtends() {
	if (in2d) {
		resultExtends.setDim(2);
		resultExtends[0] = quantization;		
		resultExtends[1] = quantization;		
	} else {
		resultExtends.setDim(3);
		resultExtends[0] = quantization;		
		resultExtends[1] = quantization;		
		resultExtends[2] = quantization;		
	}
	return resultExtends;
}


void LocalGeometryModelStatistics::assign(const Object* obj, X4CopyContext *context) {
	Persistent::assign(obj,context);
	const LocalGeometryModelStatistics *other = dynamic_cast<const LocalGeometryModelStatistics*>(obj);
	if (!other) throw X4Exception("LocalGeometryModelStatistics::assign() - unrecognized operand class.");
	copyObjectProperties(obj, this);
	this->resultExtends = other->resultExtends;
}


// --

LGMSPlaneModel::LGMSPlaneModel() {
	maxDistance = 1.0f;
}

void LGMSPlaneModel::compute(const PointList &points, const Vector3f &origin, RegularDataF &probDens) {
	card32 quantization = getQuantization();
	if (quantization < 2) quantization = 2; 
	probDens.setExtends(getResultExtends());
	
	float32 *destPos = probDens.dataPtr();
	
	if (in2d) {	
		for (card32 dist=0; dist<quantization; dist++) {
			for (card32 rot=0; rot<quantization; rot++) {
				float32 rotf = (float32)rot/(float32)quantization * M_PI;
				float32 distf = (2.0f*(float32)dist/(float32)(quantization-1) - 1.0f) * maxDistance;
				const Vector3f normal = get2dNormal(rotf);
				*destPos = computeLocalFit(points, origin, normal,	distf);
				++destPos;		
			}			
		}
	} else {
		for (card32 dist=0; dist<quantization; dist++) {
			for (card32 rotLong=0; rotLong<quantization; rotLong++) {
				for (card32 rotLat=0; rotLat<quantization; rotLat++) {
					float32 rotLongf = (float32)rotLong/(float32)quantization * M_PI;
					float32 rotLatf = (float32)rotLat/(float32)quantization * M_PI;
					float32 distf = (2.0f*(float32)dist/(float32)(quantization-1) - 1.0f) * maxDistance;
					const Vector3f normal = polarCoordsToNormal(makeVector2f(rotLatf, rotLongf));
					*destPos = computeLocalFit(points, origin, normal,	distf);
					++destPos;		
				}			
			}
		}
	}
}


// --


LGMSLeastSquaresPlaneModel::LGMSLeastSquaresPlaneModel() {
	outputLogDensity = true;
	sigma = 1.0f;
	cutOffValue = 1E10;
	exponent = 1;
}

float32 LGMSLeastSquaresPlaneModel::computeLocalFit(
	const PointList &points, 
	const Vector3f &origin, 
	const Vector3f &normalizedPlaneNormal,
	const float32 planeDist)
{
	PointList::const_iterator pos = points.begin();
	PointList::const_iterator end = points.end();
	float32 result = 0.0f;
	while (pos != end) {
		Vector3f posv = *pos;
		float32 posDist = (posv-origin) * normalizedPlaneNormal - planeDist;
		result += min(pow(posDist*posDist, exponent), cutOffValue);
		++pos;
	}
	if (outputLogDensity) {
		return -result;
	} else {
		return exp(-result*result/sigma*sigma);
	}
}


// --


LGMSKernelPlaneModel::LGMSKernelPlaneModel() {
	outputLogDensity = true;
	sigma = 0.05f;
	refit2ndOrder = false;
}

/*
static void computeFixedNormalQuadricCoeffs(
   const std::vector<Vector3f> &points,
   StaticVector<float32, 4> &coeffs)
{
   const card32 NUM_NEIGHBORS = points.size();
   coeffs[0] = 0.0f; coeffs[1] = 0.0f; coeffs[2] = 0.0f; coeffs[3] = 0.0f; 
   StaticVector<float32, 4> cd;
   StaticMatrix<float32, 4,4> dm;
   for (card32 i=0; i<4;i++) {
      for (card32 j=0; j<4;j++) {
         dm[i][j] = 0.0f;
      }
   }

   for (card32 pIndex=0; pIndex<NUM_NEIGHBORS; pIndex++){
      const float32 u = points[pIndex][0];
      const float32 v = points[pIndex][1];
      const float32 n = points[pIndex][2];

      coeffs[0] += 1   * n;
      coeffs[1] += u*v * n;
      coeffs[2] += u*u * n;
      coeffs[3] += v*v * n;

      // 1 uv uu vv

      dm[0][0] += w;      dm[0][1] += u*v*w;     dm[0][2] += u*u*w;     dm[0][3] += v*v*w;
      dm[1][0] += w*u*v;  dm[1][1] += u*v*w*u*v; dm[1][2] += u*u*w*u*v; dm[1][3] += v*v*w*u*v;
      dm[2][0] += w*u*u;  dm[2][1] += u*v*w*u*u; dm[2][2] += u*u*w*u*u; dm[2][3] += v*v*w*u*u;
      dm[3][0] += w*v*v;  dm[3][1] += u*v*w*v*v; dm[3][2] += u*u*w*v*v; dm[3][3] += v*v*w*v*v;
   }
   StaticMatrix<float32, 4,4> invDm;
   try {
      invDm = invertMatrix(dm);
   } catch (...) {
      for (card32 i=0; i<4; i++) {
         for (card32 j=0; j<4; j++) {
            invDm[i][j] = 0.0f;
         }
      }
   }
   coeffs = invDm * coeffs;
}

float32 evaluateFixedNormalQuadric(
   const StaticVector<float32, 4> &coeffs,
   const Vector2f &position)
{
   const float32 u = position[0];
   const float32 v = position[1];
   return   coeffs[0]
          + coeffs[1] * u*v
          + coeffs[2] * u*u
          + coeffs[3] * v*v;
}*/


///// rotates 2d points so that normal is pointing upward
//static std::vector<Vector2f> transformPoints(const std::vector<Vector3f> &points, Vector2f normal) {
//	for (card32 i=0; i<points.size(); i++) {
//		Vector2f p = makeVector2f(points[i][0], points[i][1]);
//	}	
//}

float32 LGMSKernelPlaneModel::computeLocalFit(
	const PointList &points, 
	const Vector3f &origin, 
	const Vector3f &normalizedPlaneNormal,
	const float32 planeDist)
{
	PointList::const_iterator pos = points.begin();
	PointList::const_iterator end = points.end();
	float32 result = 0.0f;
	while (pos != end) {
		Vector3f posv = *pos;
		float32 posDist = (posv-origin) * normalizedPlaneNormal - planeDist;
		result += exp(-posDist*posDist/(2*sigma*sigma));
		++pos;
	}
//	if (refit2ndOrder) {
//		if (in2d) {
//			PointList transformedPoints = transformPoints(points, normalizedPlaneNormal);
//			Vector2f coeff = fitCoeff(points);
//			PointList::const_iterator pos = points.begin();
//			PointList::const_iterator end = points.end();
//			float32 result = 0.0f;
//			while (pos != end) {
//				Vector3f posv = *pos;
//				float32 posDist = xxx; //(posv-origin) * normalizedPlaneNormal - planeDist;
//				result += exp(-posDist*posDist/(2*sigma*sigma));
//				++pos;
//			}
//		} else {
//			warning("LGMSKernelPlaneModel::computeLocalFit - refit 2nd order in 3d is not implemented.");
//		}
//	}
	if (outputLogDensity) {
		return log(max(static_cast<float>(result), static_cast<float>(1E-7)));
	} else {
		return result;
	}
}


// --


LGMSKernelLineEntropyPlaneModel::LGMSKernelLineEntropyPlaneModel() {
	outputLogDensity = true;
	sigma = 0.05f;
	scale = 1.0f;
	entropyWeighting = true;
	resolution = 10;
}

float32 LGMSKernelLineEntropyPlaneModel::computeLocalFit(
	const PointList &points, 
	const Vector3f &origin, 
	const Vector3f &normalizedPlaneNormal,
	const float32 planeDist)
{
	if (!getIn2d()) {x4Assert(false);}
	std::vector<float32> bins( resolution, 0.0f );
	std::vector<Vector3f> p(resolution);
	Vector3f mid = origin + normalizedPlaneNormal*planeDist;
	Vector3f normal = makeVector3f(normalizedPlaneNormal[1], -normalizedPlaneNormal[0], 0);
	for (card32 i=0; i<resolution; i++) {
		float32 ifl = 2.0f*((float32)i + 0.5f)/(float32)resolution - 1.0f;
		ifl *= scale;
		p[i] = mid + normal*ifl;
	}
	PointList::const_iterator pos = points.begin();
	PointList::const_iterator end = points.end();
	while (pos != end) {
		Vector3f posv = *pos;
		float32 posDist = (posv-origin) * normalizedPlaneNormal - planeDist;
		for (card32 i=0;i<resolution;i++) {
			Vector3f dist = p[i] - *pos;
			bins[i] += exp(-dist*dist/(2*sigma*sigma));
		}
		++pos;
	}
	float32 result = 0.0f;
	if (entropyWeighting) {
		float32 normalize = 0.0f;
		for (card32 i=0; i<resolution; i++) {
			normalize += bins[i];
		}
		float32 fact;
		if (normalize > 1e-20) {	
			fact = 1.0f / normalize;
		} else {
			fact = 0.0f;
		}
		for (card32 i=0; i<resolution; i++) {
			if (bins[i] > 0) {
				result += bins[i]*normalize * log(bins[i]*normalize);
			}
		}
		result *= normalize;
	} else {
		for (card32 i=0; i<resolution; i++) {
			result += bins[i];
		}
	}
	result /= resolution;
	
	if (outputLogDensity) {
		return log(max(static_cast<float>(result), static_cast<float>(1E-7)));
	} else {
		return result;
	}
}
