//---------------------------------------------------------------------------
#ifndef LocalGeometryModelStatisticsH
#define LocalGeometryModelStatisticsH
//---------------------------------------------------------------------------
#include "x3editor2.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
#include "LinearAlgebra.h"
#include "DynamicLinearAlgebra.h"
#include "RegularData.h"
//---------------------------------------------------------------------------
#include <vector>
//---------------------------------------------------------------------------


namespace X4 {

/// represents a small, local piece of geometry
/// computes probability densities for parameters of a model of such pieces
class X3EDITOR2_API LocalGeometryModelStatistics : public Persistent {
	X4_ABSTRACT_CLASS( LocalGeometryModelStatistics )

public:
	typedef std::vector<Vector3f> PointList;

	/// quantization of the local model (Hough Transform)
	card32 quantization;
	/// compute in 2d or 3d?
	bool in2d;
	/// cached std::vector	of result extends
	DVectorI resultExtends;
	
	void updateResultExtends();

public:
	LocalGeometryModelStatistics();

	void setQuantization(card32 q);
	card32 getQuantization()		 {return quantization;}
	
	void setIn2d(bool d);
	bool getIn2d()                 {return in2d;}
	
	/// extends of resulting hough transform - depends on "in2d" and "quantization"
	const DVectorI &getResultExtends();

   virtual void assign(const Object* obj, X4CopyContext *context = NULL);
	virtual void compute(const PointList &points, const Vector3f &origin, RegularDataF &probDens) = 0;
};


// --

/// models that fit planes
class X3EDITOR2_API LGMSPlaneModel : public	LocalGeometryModelStatistics  {
	X4_ABSTRACT_CLASS( LGMSPlaneModel )

 private:
	/// maximum distance from origin
	float32 maxDistance;
	
 public:
	float32 getMaxDistance()          {return maxDistance;}
	void    setMaxDistance(float32 d) {maxDistance = d;}
 
 protected:
	virtual float32 computeLocalFit(
		const PointList &points, 
		const Vector3f &origin, 
		const Vector3f &normalizedPlaneNormal,
		const float32 planeDist) = 0;

	Vector3f get2dNormal(float32 angle) {
		return makeVector3f(cos(angle), sin(angle), 0);
	}

 public:
	LGMSPlaneModel();
	virtual void compute(const PointList &points, const Vector3f &origin, RegularDataF &probDens); 	
};


// --


/// simplistic least squares model
class X3EDITOR2_API LGMSLeastSquaresPlaneModel : public LGMSPlaneModel {
	X4_CLASS( LGMSLeastSquaresPlaneModel )

 private:
	bool outputLogDensity;
	float32 sigma;
	float32 cutOffValue;
	float32 exponent;

 protected:	
	virtual float32 computeLocalFit(
		const PointList &points, 
		const Vector3f &origin, 
		const Vector3f &normalizedPlaneNormal,
		const float32 planeDist);

 public:
	LGMSLeastSquaresPlaneModel();	
};


/// kernel model - add to score if line runs through neighborhood of points
class X3EDITOR2_API LGMSKernelPlaneModel : public LGMSPlaneModel {
	X4_CLASS( LGMSKernelPlaneModel )

 private:
	bool outputLogDensity;
	float32 sigma;
	bool refit2ndOrder;

 protected:	
	virtual float32 computeLocalFit(
		const PointList &points, 
		const Vector3f &origin, 
		const Vector3f &normalizedPlaneNormal,
		const float32 planeDist);

 public:
	LGMSKernelPlaneModel();	
};



/// augmented kernel model - favours equally distributed points along fitted line
class X3EDITOR2_API LGMSKernelLineEntropyPlaneModel : public LGMSPlaneModel {
	X4_CLASS( LGMSKernelLineEntropyPlaneModel )

 private:
	bool outputLogDensity;
	float32 sigma;
	float32 scale;
	bool entropyWeighting;
	card32 resolution;

 protected:	
	virtual float32 computeLocalFit(
		const PointList &points, 
		const Vector3f &origin, 
		const Vector3f &normalizedPlaneNormal,
		const float32 planeDist);

 public:
	LGMSKernelLineEntropyPlaneModel();	
};


}

#endif