#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "MajorKDTree.h"
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
using namespace X4;


MajorKDTree::MajorKDTree(void)
{
	m_RootNode = NULL;
	m_MaxNumElementsPerNode = 32;
}

MajorKDTree::~MajorKDTree(void)
{
	delete m_RootNode;
}

mpcard X4::MajorKDTree::addElement( const BoundingBox3f &bb )
{
	TElement el;
	el.bb = bb;
	mpcard elementIndex = m_Elements.size();
	m_Elements.push_back(el);
	return elementIndex;
}


void X4::MajorKDTree::setupTree(mpcard maxNumElementsPerNode)
{
	m_MaxNumElementsPerNode = maxNumElementsPerNode;
	delete m_RootNode;
	m_RootNode = new TNode;
	for( unsigned i=0;i<m_Elements.size();i++ )
		m_RootNode->indices.push_back(i);
	
	if( m_RootNode->indices.size() > m_MaxNumElementsPerNode)
		setupNode(m_RootNode);
}

void X4::MajorKDTree::setupNode( TNode * node )
{
	// compute bb
	node->bb = m_Elements[node->indices[0]].bb;
	for( unsigned i=1;i<node->indices.size();i++ )
	{
		node->bb.addBB(m_Elements[node->indices[i]].bb);
	}

	float maxSideLength = node->bb.getMaxSideLength();
	for( unsigned i=0;i<3;i++ )
	{
		if( node->bb.getSideLength(i) < 1e-7f )
		{
			node->bb.lowerCorner[i] -= maxSideLength * 0.01f;
			node->bb.upperCorner[i] += maxSideLength * 0.01f;
		}
	}

	// shall we split?
	if( node->indices.size() > m_MaxNumElementsPerNode )
	{
		// choose longest bb side
		std::vector<mpcard> bestSplit_child0,bestSplit_child1,bestSplit_parent;
		unsigned bestDim = 0;
		for( unsigned dim=1;dim<3;dim++ )
		{
			if(node->bb.getSideLength(dim) > node->bb.getSideLength(bestDim))
				bestDim = dim;
		}
		splitNode(node,bestDim,bestSplit_child0,bestSplit_child1,bestSplit_parent);

		// split 
		if( min(bestSplit_child0.size(),bestSplit_child1.size()) > 0 )
		{
			node->child[0] = new TNode;
			node->child[1] = new TNode;
			node->child[0]->indices = bestSplit_child0;
			node->child[1]->indices = bestSplit_child1;
			node->indices = bestSplit_parent;
			node->splitPlane.normal = NULL_VECTOR3F;
			node->splitPlane.normal[bestDim] = 1;
			node->splitPlane.point = node->bb.getCenter();
			setupNode(node->child[0]);
			setupNode(node->child[1]);
		}

	}


}

void X4::MajorKDTree::splitNode( TNode * node, unsigned dim, std::vector<mpcard> &child0_indices, std::vector<mpcard> &child1_indices, std::vector<mpcard> &parent_indices )
{
	Vector3f center = node->bb.getCenter();

	float tolerance = 0.2f * node->bb.getSideLength(dim);
	float splitPlane = center[dim];
	float splitPlane_min = center[dim] - tolerance;
	float splitPlane_max = center[dim] + tolerance;

	for( unsigned i=0;i<node->indices.size();i++ )
	{
		mpcard index = node->indices[i];
		TElement * el = &m_Elements[index];
		float el_center = (el->bb.lowerCorner[dim] + el->bb.upperCorner[dim]) / 2.0f;
		if( el_center < splitPlane )				
		{
			if( el->bb.upperCorner[dim] < splitPlane_max )
			{
				child0_indices.push_back(index);
				continue;
			}
		}
		else
		{
			if( el->bb.lowerCorner[dim] > splitPlane_min )
			{
				child1_indices.push_back(index);
				continue;
			}
		}
		parent_indices.push_back(index);
	}
}


std::vector<mpcard> X4::MajorKDTree::computePossibleCollision( const BoundingBox3f &bb, const Matrix4f &T )
{
	std::vector<mpcard> ret;
	if( m_RootNode == NULL )
		return ret;

	BoundingBox3f transformedBB;
	transformedBB.lowerCorner = transformedBB.upperCorner = transformVector3f(T,bb.getCenter());
	for( unsigned i=0;i<8;i++ )
	{
		Vector3f corner = makeVector3f(i & 1 ? bb.lowerCorner[0]:bb.upperCorner[0],
			i & 2 ? bb.lowerCorner[1]:bb.upperCorner[1],
			i & 4 ? bb.lowerCorner[2]:bb.upperCorner[2]);
		transformedBB.addPoint(transformVector3f(T,corner));
	}

	std::vector<TNode*> nodes;
	nodes.push_back(m_RootNode);
	while (!nodes.empty())
	{
		TNode * node = nodes.back();
		nodes.pop_back();

		if( transformedBB.intersectBoundingBox(node->bb) != BVQueryDisjoint )
		{
			for( unsigned i=0;i<node->indices.size();i++ )
			{
				if( transformedBB.intersectBoundingBox(m_Elements[node->indices[i]].bb)!= BVQueryDisjoint)
					ret.push_back(node->indices[i]);
			}

			if( !node->isLeaf() )
			{
				nodes.push_back(node->child[0]);
				nodes.push_back(node->child[1]);
			}
		}
	}
	return ret;		
}

void X4::MajorKDTree::getSubtreeIndices( TNode * node, std::vector<mpcard> &result )
{
	for( unsigned i=0;i<node->indices.size();i++ )
		result.push_back(node->indices[i]);
	if( !node->isLeaf())
	{
		getSubtreeIndices(node->child[0],result);
		getSubtreeIndices(node->child[1],result);
	}
}