#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PCCBilateralSmoothing.h"
//---------------------------------------------------------------------------
#include "copyObjectProperties.h"
#include <time.h>
#include "DebugOutput.h"
#include "NormalEstimator.h"
#include "PointCloudTools.h"
//---------------------------------------------------------------------------
using namespace X4;

IMPLEMENT_X4_CLASS( PCCBilateralSmoothing ,0)
{
	BEGIN_CLASS_INIT( PCCBilateralSmoothing );
	ADD_FLOAT32_PROP(m_SigmaEuklid,0);
	ADD_FLOAT32_PROP(m_SigmaPlane,0);
	SET_USER_EDIT(false);
	ADD_FLOAT32_PROP(m_SigmaPointToPlane,0);

	ADD_CARD64_PROP( m_OOC_MaxNumPointsPerPart, 0);

	ADD_STRING_PROP(m_DestHierarchicalPointCloud,0);
	ADD_FLOAT32_PROP(m_OutlierThreshold,0);
	ADD_BOOLEAN_PROP(m_CalculateNormals,0);
	ADD_CARD32_PROP(m_CalcNormals_NumPoints,0);
}

PCCBilateralSmoothing::PCCBilateralSmoothing(void)
{
	m_SigmaEuklid = 0.1f;
	m_SigmaPlane = 0.03f;
	m_SigmaPointToPlane = 0.025f;
	m_OOC_MaxNumPointsPerPart = 400*1000;
	m_DestHierarchicalPointCloud = "";
	m_OutlierThreshold = 3.0f;
	m_CalculateNormals  = false;
	m_CalcNormals_NumPoints = 20;
}

PCCBilateralSmoothing::~PCCBilateralSmoothing(void)
{
}

void PCCBilateralSmoothing::assign(const Object* obj, X4CopyContext *context)
{
	PCCommand::assign(obj,context);
	const PCCBilateralSmoothing * o = dynamic_cast<const PCCBilateralSmoothing*>(obj);
	x4Assert( o != NULL );

	copyObjectProperties(obj,this);
}

void X4::PCCBilateralSmoothing::execute( Scene * scene )
{
	float32 startTime = clock();
	const std::vector<std::string> selection = getScene()->getRootState()->staticState->getSelectedNodes();
	for( card32 i=0;i<selection.size();i++ )
	{
		PointCloud * pc = getPointCloud( getScene(), selection[i] );
		if( pc )
		{
			if( dynamic_cast<UnstructuredInCorePointCloud*>(pc) != NULL && 
				getPointCloud(getScene(), m_DestHierarchicalPointCloud ) == NULL )
				smooth(pc);
			else
			{
				PointCloud * destPC = getPointCloud(getScene(), m_DestHierarchicalPointCloud );
				if( destPC == NULL )
				{
					warning( "PCCBilateralSmoothing - Destination-pointcloud not found!\n");
					break;
				}
				hierarchicalSmoothing(pc,destPC);
				break;
			}
		}
	}
	float32 timeRequired = float32( ( clock() - startTime ) / CLK_TCK );
	debugOutput << "Bilateral reconstruction filter, time required: " << timeRequired << ".\n"; debugOutput.flush();
}

void X4::PCCBilateralSmoothing::smooth( PointCloud * pc )
{
	if( m_CalculateNormals )
	{
		NormalEstimator::estimateNormals( pc, m_CalcNormals_NumPoints );
	}

	AAT posAAT = pc->getAAT( "position" );
	AAT normalAAT;
	normalAAT = pc->getAAT( "normal" );

	SceneObjectIterator * iter = pc->createIterator( SceneObjectIteratorType::CAP_BASIC_PC );
	x4Assert( iter != NULL );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(iter);

	HierarchicalKNNIterator hIt(pc,32,NULL);

	float inv_sigmaEuklid_2 = 1.0f / pow(m_SigmaEuklid,2.0f);
	//float inv_sigmaPlane_2 = 1.0f / pow(m_SigmaPlane,2.0f);
	float rad_2 = m_SigmaEuklid * m_SigmaEuklid*4.0f;
	float inv_sigmaPointToPlane_2 = 1.0f / pow(m_SigmaPointToPlane,2.0f);

	while( !bIt->atEnd() )
	{
		Vector3f pos = bIt->get3f(posAAT);
		Vector3f normal = bIt->get3f(normalAAT);

		hIt.setSeekPointAndReset(pos);
		float summed_weight=0;
		Vector3f averagedNormal = NULL_VECTOR3F;
		card32 flipCount = 0;
		card32 numNeighbourSamples = 0;
		while( !hIt.atEnd())
		{
			Vector3f spos = hIt.get3f(posAAT);
			Vector3f snormal = hIt.get3f(normalAAT);
			float dist_2 = normQuad(spos - pos);
			if( dist_2 > rad_2 )
				break;

			numNeighbourSamples++;
			if( normal * snormal < 0)
			{
				snormal *= -1;
				flipCount++;
			}

			float planeDist_2 = max(0.0f, 1.0f - snormal * normal);
			float pointToPlaneDist_2 = pow(snormal * (spos - pos),2.0f);
			
			float euklidWeight = exp( -dist_2 * inv_sigmaEuklid_2 );
			//float planeWeight = exp(-planeDist_2 * inv_sigmaPlane_2);
			float planeDistWeight = exp(-pointToPlaneDist_2 * inv_sigmaPointToPlane_2 );
			float weight = euklidWeight * planeDistWeight;
			
			summed_weight += weight;
			averagedNormal += snormal * weight;

			hIt.next();
		}
		averagedNormal = normalize(averagedNormal);
		if( flipCount > numNeighbourSamples*4/3 )
			averagedNormal *= -1;
		mIt->set3f(normalAAT, averagedNormal);

		bIt->next();
	}

	bIt->reset();
	while( !bIt->atEnd() )
	{
		Vector3f pos = bIt->get3f(posAAT);
		Vector3f normal = bIt->get3f(normalAAT);

		hIt.setSeekPointAndReset(pos);
		float summed_weight = 0;
		Vector3f averagedPoint = NULL_VECTOR3F;
		while( !hIt.atEnd())
		{
			Vector3f spos = hIt.get3f(posAAT);
			Vector3f snormal = hIt.get3f(normalAAT);
			float dist_2 = normQuad(spos - pos);
			if( dist_2 > rad_2 )
				break;

			float planeDist_2 = max(0.0f, 1.0f - snormal * normal);
			float pointToPlaneDist_2 = pow(snormal * (spos - pos),2.0f);
			
			float euklidWeight = exp( -dist_2 * inv_sigmaEuklid_2 );
//			float planeWeight = exp(-planeDist_2 * inv_sigmaPlane_2);
			float planeDistWeight = exp(-pointToPlaneDist_2 * inv_sigmaPointToPlane_2 );
			float weight = euklidWeight * planeDistWeight;

			Vector3f projectedPoint = pos + snormal * (snormal * (spos - pos));
			
			summed_weight += weight;
			averagedPoint += projectedPoint * weight;

			hIt.next();
		}
		averagedPoint /= summed_weight;
		mIt->set3f(posAAT, averagedPoint);

		bIt->next();
	}
	delete iter;
}

void X4::PCCBilateralSmoothing::hierarchicalSmoothing( PointCloud * pc, PointCloud * destPC )
{
	SceneObjectIterator * iter = pc->createIterator( SceneObjectIteratorType::CAP_BASIC_PC 
		| SceneObjectIteratorType::CAP_HIERARCHICAL | SceneObjectIteratorType::CAP_ATTACH_SECONDARY
		| SceneObjectIteratorType::CAP_MODIFYABLE_PC);
	x4Assert( iter != NULL );
//	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
//	HierarchicalIterator * hIt = dynamic_cast<HierarchicalIterator*>(iter);

	SceneObjectIterator * dest_iter = destPC->createIterator( SceneObjectIteratorType::CAP_DYNAMIC_PC );


	traversePointCloud(pc, iter, dest_iter);

	delete iter;
	delete dest_iter;
	
}

void X4::PCCBilateralSmoothing::traversePointCloud( PointCloud * pc, SceneObjectIterator * iter, SceneObjectIterator *destIter )
{
	HierarchicalIterator * hIt = dynamic_cast<HierarchicalIterator *>(iter);
	
	card64 numPoints;
	hIt->getNodeInfo(HierarchicalIterator::niNumPointsInSubtree,NULL,&numPoints);
	if( hIt->isLeaf() || numPoints < m_OOC_MaxNumPointsPerPart )
	{
		processPointInNode(pc, iter,destIter);
	}
	else
	{
		for( unsigned i=0;i<hIt->getNumChildren();i++ )
		{
			if( hIt->goToChild(i) )
			{
				traversePointCloud(pc,iter,destIter);
				hIt->goToParent();
			}
		}
	}
	
	
}

void X4::PCCBilateralSmoothing::processPointInNode( PointCloud * pc, SceneObjectIterator * iter, SceneObjectIterator *destOutputIter )
{
	// --------------------------------------------------
	// gather points & copy them to a temporal point cloud (unstructured in-core)
	HierarchicalIterator * hIt = dynamic_cast<HierarchicalIterator *>(iter);
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);

	// get current bb and add a border
	BoundingBox3f bb = hIt->getBoundingBox();
	BoundingBox3f bb_bordered = bb;
	bb_bordered.addBorder(m_SigmaEuklid*4.0f);

	UnstructuredInCorePointCloud * subpart_pc = new UnstructuredInCorePointCloud;
	subpart_pc->clearAndSetup( pc->getDescr(),0);

	SceneObjectIterator * destIter = subpart_pc->createIterator( SceneObjectIteratorType::CAP_BASIC_PC |
		SceneObjectIteratorType::CAP_MODIFYABLE_PC | SceneObjectIteratorType::CAP_DYNAMIC_PC);

	gatherPoints( iter, bb_bordered, destIter);
	delete destIter;

	// ---------------------------------------------------
	// smooth normals/points and store all points inside bb into the result pc
	checkAttribute(subpart_pc,"normal",3,VAD::DATA_FORMAT_FLOAT32);
	smooth(subpart_pc);

	DynamicPointCloudIterator * destDynIt = dynamic_cast<DynamicPointCloudIterator*>(destOutputIter);
	PointSet * ps = subpart_pc->getPointSet();
	AAT posAAT = ps->getAAT("position");
	std::vector<char> buffer( ps->getDescr()->getSize());
	for( unsigned i=0;i<ps->getNumEntries();i++ )
	{
		Vector3f pos = ps->get3f(i,posAAT);
		if( bb.inBoundingBox(pos))
		{
			ps->getVertex(i,&buffer[0]);
			destDynIt->add(&buffer[0]);
		}
	}

	message("Block proceeded - NumPoints: ");
	debugOutput << (card32)subpart_pc->getNumPoints() << "\n";
	debugOutput << "Percentage:" << (float)subpart_pc->getNumPoints() / pc->getNumPoints()*100.0f << "\n;";
	debugOutput.flush();
	delete subpart_pc;
}

void X4::PCCBilateralSmoothing::gatherPoints( SceneObjectIterator * sourceIter, const BoundingBox3f &bb, SceneObjectIterator *destIter )
{
	HierarchicalIterator * hIt = dynamic_cast<HierarchicalIterator *>(sourceIter);
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(sourceIter);

	DynamicPointCloudIterator * destDynIt = dynamic_cast<DynamicPointCloudIterator*>(destIter);

	
	if( hIt->isLeaf() )
	{
		const PointSet * ps = bIt->lockBlockRead();
		AAT posAAT = ps->getAAT("position");
		std::vector<char> buffer(ps->getDescr()->getSize());

		bool hasSurfaceness = ps->providesAttribute("surfaceness");
		AAT surfacenessAAT;
		if( hasSurfaceness )
			surfacenessAAT = ps->getAAT("surfaceness");

		for( unsigned i=0;i<ps->getNumEntries();i++ )
		{
			if( hasSurfaceness )
			{
				if( ps->get1f(i, surfacenessAAT) < m_OutlierThreshold ) 
					continue;
			}


			Vector3f pos = ps->get3f(i,posAAT);
			if( bb.inBoundingBox(pos) )
			{
				ps->getVertex(i, &buffer[0]);
				destDynIt->add(&buffer[0]);
			}
		}
		bIt->unlock();
	}
	else
	{
		for( unsigned i=0;i<hIt->getNumChildren();i++ )
		{
			SceneObjectIterator * iter_copy = (SceneObjectIterator*)sourceIter->copy();
			HierarchicalIterator * hIt_copy = dynamic_cast<HierarchicalIterator *>(iter_copy);
			if( hIt_copy->goToChild(i) )
			{
				if( bb.intersectBoundingBox(hIt_copy->getBoundingBox()) != BVQueryDisjoint )
				{
					gatherPoints(iter_copy, bb, destIter);
				}
			}
			delete iter_copy;
		}
	}
}
