//---------------------------------------------------------------------------
#include "stdafx.h"
//---------------------------------------------------------------------------
#include "PCCComputeTopology.h"
//---------------------------------------------------------------------------
#include "BooleanClassProperty.h"
#include "NumericalClassProperty.h"
#include "PointCloud.h"
#include "UnstructuredInCorePointCloud.h"
#include "ParametricInCorePointCloud.h"
#include "AttachedData.h"
#include "TopologyAttachment.h"
#include "Random.h"
#include "PointSet.h"
#include "InCorePCTopologyGraph.h"
#include "ProgressWindow.h"
#include "GenericSwapAllOutOfCorePointCloud.h"
#include "UnstructuredInCoreTriangleMesh.h"
#include "PointCloudTools.h"
#include "FastSphereQuerry.h"
//---------------------------------------------------------------------------
#include <set>
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------


std::string descrTopologyType(card32 i) {
	switch (i) {
		case PCCComputeTopology::TOPTYPE_RANDOM: return "RANDOM";
		case PCCComputeTopology::TOPTYPE_KNN: return "KNN";
		case PCCComputeTopology::TOPTYPE_EPS: return "EPS";
		case PCCComputeTopology::TOPTYPE_KNN_AND_EPS: return "EPS_AND_KNN";
		case PCCComputeTopology::TOPTYPE_TRIMESH_CONNECTIVITY: return "MESH_TOPOLOGY";
		default: return "unknown";
	}
}

IMPLEMENT_X4_CLASS( PCCComputeTopology, 1 ) 
{
	BEGIN_CLASS_INIT( PCCComputeTopology );
	ADD_CARD32_PROP_UPDATE_DESCR( topologyType, 0, NULL, descrTopologyType, 5, NULL)
	ADD_FLOAT32_PROP( epsilon , 0)
	ADD_CARD32_PROP( k , 0)
	ADD_BOOLEAN_PROP( attachToOOCData, 1);
}



PCCComputeTopology::PCCComputeTopology() {
	topologyType = TOPTYPE_KNN;
	epsilon = 0.1f;
	k = 20;
	attachToOOCData = false;
}

InCorePCTopologyGraph *PCCComputeTopology::computeRandomTopology(PointSet *ps) {
	card32 n = ps->getNumEntries();
	set< pair <card32, card32> > pairs;
	InCorePCTopologyGraph *top = new InCorePCTopologyGraph();
	VertexDescriptor vd;
	vd.pushAttrib(VAD("vindex", 2, VAD::DATA_FORMAT_INT32));
	vd.pushAttrib(VAD("color", 4, VAD::DATA_FORMAT_FLOAT32));
	top->clearAndSetup(&vd);
	for (card32 i=0; i<k; i++) {
		card32 v1 = randomNumber(n);
		card32 v2 = randomNumber(n);
		pair<card32, card32> p(v1, v2);
		if (pairs.find(p) == pairs.end()) {
			pairs.insert(p);
			top->addEdge(v1, v2);
		}
	}
	return top;
}

InCorePCTopologyGraph *PCCComputeTopology::computeKNNTopology(PointSet *ps, PointCloud *pc){
    if (!ps->getDescr()->providesAttribute("position", 3, VAD::DATA_FORMAT_FLOAT32)) return NULL;
    InCorePCTopologyGraph *top = new InCorePCTopologyGraph();
    VertexDescriptor vd;
    vd.pushAttrib(VAD("vindex", 2, VAD::DATA_FORMAT_INT32));
    //vd.pushAttrib(VAD("color", 4, VAD::DATA_FORMAT_FLOAT32));
    top->clearAndSetup(&vd);
    AAT POS = ps->getAAT("position");
    card32 n = ps->getNumEntries();
    HierarchicalKNNIterator knn(pc, 20, NULL);
    progressWindow->pushStep(true, "knn topology");
    for (card32 i=0; i<n; i++) {
        Vector3f pos = ps->get3f(i, POS);
        knn.setSeekPointAndReset(pos);
        card32 j = 0;
        while (!knn.atEnd() && j < k) {
            int32 i2 = knn.getPointSetPointNumber();
            if (i != i2) {
                top->addEdge(i, i2);
            }
            knn.next();
            j++;
        }
        if (i%5555==0) {
            progressWindow->progress((card32)((double)i/(double)n*100.0));
        }
    }
    bool aborted;
    progressWindow->popStep(&aborted);
    //if (aborted) {
    //	delete top;
    //	return NULL;
    //} else {
    return top;
    //}
}

InCorePCTopologyGraph *PCCComputeTopology::computeTopologyFromMesh(PointSet *ps, PointCloud *pc)
{
    UnstructuredInCoreTriangleMesh* mesh = dynamic_cast<UnstructuredInCoreTriangleMesh*>(pc);
    if (mesh == NULL)
    {

        warning("This type of topology works only for triangle meshes!");
        return NULL;
    }

    mesh->computeTopologyAttachment();
    TopologyAttachment* tatt = dynamic_cast<TopologyAttachment*>(mesh->getAttachments()->getData(TopologyAttachment::getDefaultName()));
    return (InCorePCTopologyGraph*)tatt->getTopology()->copy();
}

InCorePCTopologyGraph *PCCComputeTopology::computeEpsTopology(const PointSet *ps, PointCloud *pc)
{
    if (!ps->getDescr()->providesAttribute("position", 3, VAD::DATA_FORMAT_FLOAT32)) return NULL;
    InCorePCTopologyGraph *top = new InCorePCTopologyGraph();
    VertexDescriptor vd;
    vd.pushAttrib(VAD("vindex", 2, VAD::DATA_FORMAT_INT32));
    top->clearAndSetup(&vd);
    AAT POS = ps->getAAT("position");
    card32 n = ps->getNumEntries();
    progressWindow->pushStep(true, "eps topology");
    FastSphereQuerry query(pc);
    mpcard numPoints;
    mpcard *pts;
    for (card32 i=0; i<n; i++) {
        Vector3f pos = ps->get3f(i, POS);
        query.querry(pos, epsilon, &pts, numPoints);
        for (card32 j=0; j<numPoints; j++) {
            if (i == pts[j]) continue;
            top->addEdge(i, pts[j]);
        }
        if ((i+1)%100==0)
        {
            progressWindow->progress((card32)((float32)(i+1)/(float32)n*100.0));
        }
    }
    bool aborted = false;
    progressWindow->popSteps(&aborted);
    if (aborted) {
        delete top;
        return NULL;
    } else {
        return top;
    }



}

InCorePCTopologyGraph *PCCComputeTopology::computeEpsKNNTopology(const PointSet *ps, PointCloud *pc, card32 numKNN){
    if (!ps->getDescr()->providesAttribute("position", 3, VAD::DATA_FORMAT_FLOAT32)) return NULL;
    InCorePCTopologyGraph *top = new InCorePCTopologyGraph();
    VertexDescriptor vd;
    vd.pushAttrib(VAD("vindex", 2, VAD::DATA_FORMAT_INT32));
    //vd.pushAttrib(VAD("color", 4, VAD::DATA_FORMAT_FLOAT32));
    top->clearAndSetup(&vd);
    AAT POS = ps->getAAT("position");
    card32 n = ps->getNumEntries();
    progressWindow->pushStep(true, "eps topology");
    HierarchicalKNNIterator knn(pc, 20, NULL);
    for (card32 i=0; i<n; i++) {
        Vector3f pos = ps->get3f(i, POS);
        knn.setSeekPointAndReset(pos);
        Vector3f currentPos = knn.get3f(POS);
        float32 dist = norm(pos - currentPos);
        card32 numNeighbors=0;
        while (!knn.atEnd() && dist <= epsilon && numNeighbors < numKNN) {
            int32 i2 = knn.getPointSetPointNumber();
            if (i != i2) {
                top->addEdge(i, i2);
                numNeighbors++;
            }

            knn.next();
            currentPos = knn.get3f(POS);
            dist = norm(pos - currentPos);
        }
        if ((i+1)%100==0)
        {
            progressWindow->progress((card32)((float32)(i+1)/(float32)n*100.0));
        }
    }

    bool aborted = false;
    progressWindow->popSteps(&aborted);
    if (aborted) {
        delete top;
        return NULL;
    } else {
        return top;
    }

}


void PCCComputeTopology::execute(Scene *scene) {
    unsigned int m;
    std::vector<std::string> selection = scene->getRootState()->staticState->getSelectedNodes();
    for ( m = 0; m < selection.size(); m++ ){
        PointCloud *pc = getPointCloud( scene, selection[m] );
        PointSet *ps = NULL;
        SceneObjectIterator * iter = NULL;

        // handle cases
        if (dynamic_cast<UnstructuredInCorePointCloud*>(pc) != NULL) {
            ps = dynamic_cast<UnstructuredInCorePointCloud*>(pc)->getPointSet();
        }
        if (dynamic_cast<ParametricInCorePointCloud*>(pc) != NULL) {
            ps = dynamic_cast<ParametricInCorePointCloud*>(pc)->getPointSet();
        }
        if( dynamic_cast<GenericSwapAllOutOfCorePointCloud*>(pc) != NULL )
        {
            iter = pc->createIterator(SOIT::CAP_BASIC_PC);
            BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
            ps = (PointSet*)bIt->lockBlockRead();
        }

        // create topology graph
        if (ps == NULL) continue;
        InCorePCTopologyGraph *top = NULL;
        switch (topologyType) {
        case TOPTYPE_RANDOM: top = computeRandomTopology(ps); break;
        case TOPTYPE_KNN: top = computeKNNTopology(ps, pc); break;
        case TOPTYPE_EPS: top = computeEpsKNNTopology(ps, pc); break;
        case TOPTYPE_KNN_AND_EPS: top = computeEpsKNNTopology(ps, pc, k); break;
        case TOPTYPE_TRIMESH_CONNECTIVITY: top = computeTopologyFromMesh(ps, pc); break;
        default: break;
        }      
        if (top == NULL) 
            throw X4Exception("Could not compute topology");
        // attach to point cloud
        AttachedDataContainer *attachments;
        if( attachToOOCData && iter)
        {
            GenericSwapAllOutOfCorePointCloud * gpc = dynamic_cast<GenericSwapAllOutOfCorePointCloud*>(pc);
            PointCloud * innerPointCloud = gpc->fetchPC();
            attachments = innerPointCloud->getAttachments();
        }
        else
        {
            attachments = pc->getAttachments();
        }
        TopologyAttachment *topAtt = dynamic_cast<TopologyAttachment*>(attachments->getData(TopologyAttachment::getDefaultName()));
        if (topAtt == NULL) {
            topAtt = new TopologyAttachment();
            topAtt->setup(TopologyAttachment::getDefaultName(), AttachedData::ADF_PERSISTENT);
            attachments->attachData(topAtt);
        }
        topAtt->setTopology(top);

        // cleanup
        if( iter )
        {
            BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
            bIt->unlock();
            delete iter;
        }
    }
}


void PCCComputeTopology::execute_ar(Scene *scene, UnstructuredInCorePointCloud* upc) {
    PointSet *ps = upc->getPointSet();

    // create topology graph
    InCorePCTopologyGraph *top = NULL;
    top = computeKNNTopology_ar(ps, upc);
    if (top == NULL) 
        throw X4Exception("Could not compute topology");

    // attach to point cloud
    AttachedDataContainer *attachments;
    attachments = upc->getAttachments();

    TopologyAttachment *topAtt = dynamic_cast<TopologyAttachment*>(attachments->getData(TopologyAttachment::getDefaultName()));
    if (topAtt == NULL) {
        topAtt = new TopologyAttachment();
        topAtt->setup(TopologyAttachment::getDefaultName(), AttachedData::ADF_PERSISTENT);
        attachments->attachData(topAtt);
    }

    topAtt->setTopology(top);
}

InCorePCTopologyGraph *PCCComputeTopology::computeKNNTopology_ar(PointSet *ps, UnstructuredInCorePointCloud *upc){
    InCorePCTopologyGraph *top = new InCorePCTopologyGraph();
    VertexDescriptor vd;
    vd.pushAttrib(VAD("vindex", 2, VAD::DATA_FORMAT_INT32));
    top->clearAndSetup(&vd);
    AAT POS = ps->getAAT("position");
    card32 n = ps->getNumEntries();
    HierarchicalKNNIterator knn(upc, 20, NULL);
    for (card32 i=0; i<n; i++) {
        Vector3f pos = ps->get3f(i, POS);
        knn.setSeekPointAndReset(pos);
        card32 j = 0;
        while (!knn.atEnd() && j < k) {
            int32 i2 = knn.getPointSetPointNumber();
            if (i != i2) {
                top->addEdge(i, i2);
            }
            knn.next();
            j++;
        }
    }
    return top;
}

InCorePCTopologyGraph * PCCComputeTopology::getOrCreateTopology( PointCloud* pc )
{
    TopologyAttachment *topAtt = dynamic_cast<TopologyAttachment*>(pc->getAttachments()->getData(TopologyAttachment::getDefaultName()));
    if( topAtt )
    {

        return topAtt->getTopology();
    }


    // create new topology attachment
    topAtt = new TopologyAttachment();
    topAtt->setup(TopologyAttachment::getDefaultName(), AttachedData::ADF_PERSISTENT);
    pc->getAttachments()->attachData(topAtt);

    // compute topology
    SceneObjectIterator * iter_pc = pc->createIterator(SOIT::CAP_BASIC_PC|SOIT::CAP_MODIFYABLE_PC);
    BasicPointCloudIterator * bIt_pc = dynamic_cast<BasicPointCloudIterator*>(iter_pc);
    const PointSet * ps = bIt_pc->lockBlockRead();
    InCorePCTopologyGraph * top = computeTopology((PointSet*)ps,pc);
    topAtt->setTopology(top);
    bIt_pc->unlock();

    return top;	
}