#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PCCGridResampler.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "HashTables.inline.h"
#include "PointCloudTools.h"
//---------------------------------------------------------------------------
using namespace X4;

IMPLEMENT_X4_CLASS( PCCGridResampler ,0)
{
	BEGIN_CLASS_INIT( PCCGridResampler );

	ADD_FLOAT32_PROP( m_AbsoluteSamplingSpace, 0 );
	ADD_FLOAT32_PROP( m_RelativeSamplingSpace, 0 );
	ADD_BOOLEAN_PROP( m_UseRelativeSamplingSpace, 0 );
}

PCCGridResampler::PCCGridResampler(void)
{
	m_AbsoluteSamplingSpace = 0.01f;
	m_RelativeSamplingSpace = 10.0f;
	m_UseRelativeSamplingSpace = false;
}

PCCGridResampler::~PCCGridResampler(void)
{
}

void PCCGridResampler::assign(const Object* obj, X4CopyContext *context)
{
	PCCommand::assign(obj,context);
	const PCCGridResampler * o = dynamic_cast<const PCCGridResampler*>(obj);
	x4Assert( o != NULL );

	copyObjectProperties( obj, this );
}

void PCCGridResampler::execute(Scene *scene) 
{
	unsigned int m;
   std::vector<std::string> selection = getScene()->getRootState()->staticState->getSelectedNodes();
   PointCloud* pc;
   for ( m = 0; m < selection.size(); m++ )
	{
      pc = getPointCloud( getScene(), selection[m] );
		if( pc )
		{
			float samplingSpace;
			if( m_UseRelativeSamplingSpace )
			{
				float medianPointDistance = getMedianPointDistance(pc);
				samplingSpace = medianPointDistance * m_RelativeSamplingSpace;
			}
			else
			{
				samplingSpace = m_AbsoluteSamplingSpace;
			}
			
			PointCloud * newPC = resamplePointCloud( pc, samplingSpace );
			replacePointCloud( getScene(), selection[m], newPC );
		}

   }
}

void deleteGridEntries( PCCGridResampler::TGridKey key, PCCGridResampler::TGridEntry * entry )
{
	delete entry;
}

max_platform_card PCCGridResampler::MyHashTable::hashFunction(TGridKey key) const
{
	max_platform_card ret;
	ret = (max_platform_card)key.x + (max_platform_card)key.y * 1000 +  (max_platform_card)key.z * 100000;
	return ret;
}

PointCloud *PCCGridResampler::resamplePointCloud( PointCloud * pc, float samplingSpace ) 
{
	MyHashTable table;
	samplingSpace /= sqrt(3.0f);
	
	SceneObjectIterator * iter = pc->createIterator( SceneObjectIteratorType::CAP_BASIC_PC );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	const VertexDescriptor * vd = pc->getDescr();
	std::vector<char> tmp_vertex( vd->getSize());
	UnstructuredInCorePointCloud * upc = new UnstructuredInCorePointCloud;
	upc->clearAndSetup( vd, 0);
	PointSet * ps = upc->getPointSet(); 
	
	AAT posAAT = pc->getAAT( "position" );
	bool hasNormal = pc->providesAttribute( "normal" );
	AAT normalAAT;
	if( hasNormal )
		normalAAT = pc->getAAT( "normal" );

	bool hasColor = pc->providesAttribute( "color" );
	AAT colorAAT;
	if( hasColor )
		colorAAT = pc->getAAT( "color" );

	TGridEntry *currentEntry;
	TGridKey currentKey;
	currentKey.x = 0x7fffffff;
	currentKey.y = 0x7fffffff;
	currentKey.z = 0x7fffffff;
	card32 numGridElements = 8;
	float gridEntrySamplingSpace = samplingSpace * (float)numGridElements;
	std::vector<float> weight;
	
	while( !bIt->atEnd() )
	{
		Vector3f pos = bIt->get3f( posAAT );

		// transfom position into grid entry "space"
		TGridKey key;
		Vector3f keyf = pos / gridEntrySamplingSpace;
		key.x = keyf[0] >= 0.0f ? keyf[0] : keyf[0]-1.0f;
		key.y = keyf[1] >= 0.0f ? keyf[1] : keyf[1]-1.0f;
		key.z = keyf[2] >= 0.0f ? keyf[2] : keyf[2]-1.0f;

		if( !gridComp::equals( key, currentKey ) )
		{
			currentKey = key;
			if( !table.getElement( key, currentEntry ) )
			{
				currentEntry = new TGridEntry;
				currentEntry->vertexIndex = std::vector<card32>( numGridElements*numGridElements*numGridElements,0x7fffffff);
				table.addElement( key, currentEntry );
			}
						
		}

		// transform position into local gridspace
		int x = (keyf[0] - key.x) * numGridElements;
		int y = (keyf[1] - key.y) * numGridElements;
		int z = (keyf[2] - key.z) * numGridElements;

		card32 cellIndex = (z*numGridElements+y)*numGridElements+x;
		if( cellIndex > numGridElements * numGridElements * numGridElements )
			debugOutput << "cellIndex out of range! (x,y,z) = (" << x << "," << y << "," << ")\n";

		if( currentEntry->vertexIndex[cellIndex] == 0x7fffffff )
		{
			bIt->getVertex( &tmp_vertex[0] );
			ps->addPoint( &tmp_vertex[0] );
			currentEntry->vertexIndex[cellIndex] = ps->getNumEntries()-1;
			weight.push_back( 1.0f );
		}
		else
		{
			card32 vi = currentEntry->vertexIndex[cellIndex];
			weight[vi] += 1.0f;
			ps->set3f( vi, posAAT, ps->get3f(vi,posAAT) + pos);
		}

		bIt->next();
	}

	table.forEachEntry( deleteGridEntries );

	for( card32 i=0;i<ps->getNumEntries();i++ )
		ps->set3f(i, posAAT, ps->get3f(i,posAAT) / weight[i]);

	return upc;
}
