//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PCCImportMPIStructuredLight.h"
//---------------------------------------------------------------------------
#include "NumericalClassProperty.h"

#include "DebugOutput.h"
#include "PointCloud.h"
#include "PointCloudTools.h"
#include "Scene.h"
#include "SceneGraphState.h"
#include "LinearAlgebra.inline.h"
#include <queue> 
#include "NormalUnifier.h"
#include "QtCore/QFileInfo.h"
#include "SGSimpleAnimationNode.h"
#include "SGObjectNode.h"
#include "ProgressWindow.h"
#include "NormalEstimator.h"
#include "GenericSwapAllOutOfCorePointCloud.h"
#include "StandardSceneGraphIterator.h"
#include "SGIndexedAndByNameSceneGraphIteratorAdaptor.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

IMPLEMENT_X4_CLASS( PCCImportMPIStructuredLight, 0 )
{
	BEGIN_CLASS_INIT( PCCImportMPIStructuredLight )
	ADD_INT32_PROP( startFrame, 0 )
	ADD_INT32_PROP( endFrame, 0 )
	ADD_STRING_PROP( inputFilename, 0 )
	ADD_BOOLEAN_PROP( computeNormals, 0 )
}

PCCImportMPIStructuredLight::PCCImportMPIStructuredLight() 
{
	startFrame = 1;
	endFrame = 10;
	inputFilename = "D:/LocalProject/Data/StructuredLight/dynamic_hand.bin";
	computeNormals = false;
}

PCCImportMPIStructuredLight::~PCCImportMPIStructuredLight() 
{
}

void PCCImportMPIStructuredLight::execute( Scene *scene )
{

	QString filename = QString::fromStdString( inputFilename );

	QFileInfo fileInfo( filename );
	if ( !fileInfo.exists() ){
		std::cout << "Can't find file!" << endl;
		return;
	}

	// Open file
	FILE* scannerFile;
	fopen_s( &scannerFile, QString( filename ).toLatin1(), "rb" );
	StructuredLightHeader scannerFileheader;
	fread( &scannerFileheader, 1, sizeof(StructuredLightHeader), scannerFile );
	int32 scannerFileCurrentFrame = 0;
	int32 scannerFileNumFrames = scannerFileheader.timesteps;

	debugOutput << "#frames in file: " << scannerFileNumFrames << ".\n"; debugOutput.flush();

	// Skip frames
	while( ( scannerFileCurrentFrame < startFrame ) && ( scannerFileCurrentFrame < scannerFileNumFrames ) ){
		/// Calculate size of skipping offset;
		unsigned long offset = 0;
		if ( scannerFileheader.content & DATA_Z )
			offset += scannerFileheader.width * scannerFileheader.height * sizeof( float );
		if ( scannerFileheader.content & DATA_XY )
			offset += scannerFileheader.width * scannerFileheader.height * sizeof( float ) * 2;
		if ( scannerFileheader.content & DATA_MASK )
			offset += scannerFileheader.width * scannerFileheader.height * sizeof( unsigned char  );
		if ( scannerFileheader.content & DATA_GREY )
			offset += scannerFileheader.width * scannerFileheader.height * sizeof( unsigned char );	

		/// Skip data in file
		_fseeki64( scannerFile, offset, SEEK_CUR );
		scannerFileCurrentFrame++;
	}

	SGSimpleAnimationNode* animNode = new SGSimpleAnimationNode();
	animNode->setName( "data" );
	scene->getRootNode()->addChildNode( animNode );

	progressWindow->pushStep( true, "MPI Importer" );
	progressWindow->progress( 1 );

	card32 nodeIndex = 0;
	while( ( scannerFileCurrentFrame <= endFrame ) && ( scannerFileCurrentFrame < scannerFileNumFrames ) ){
		// Read frame
		unsigned long numValues = scannerFileheader.width * scannerFileheader.height;
		StructuredLightFrame frameData;
		frameData.greyValues = new unsigned char[numValues];
		frameData.zValues = new float[numValues];
		frameData.xValues = new float[numValues];
		frameData.yValues = new float[numValues];
		frameData.masks = new unsigned char[numValues];
		frameData.numValues = numValues;
		frameData.width = scannerFileheader.width;
		frameData.height = scannerFileheader.height;
		if ( scannerFileheader.content & DATA_Z ){
			for ( unsigned long i = 0; i < numValues; i++ ){
				fread( &(frameData.zValues[i]), sizeof(float), 1, scannerFile );
			}
		}
		if ( scannerFileheader.content & DATA_MASK ){
			for ( unsigned long i = 0; i < numValues; i++){
				fread( &(frameData.masks[i]), sizeof(unsigned char), 1, scannerFile );
			}
		}
		if ( scannerFileheader.content & DATA_XY ){
			for ( unsigned long i = 0; i < numValues; i++ ){
				fread( &(frameData.xValues[i]), sizeof( float ), 1, scannerFile );
			}
			for ( unsigned long i = 0; i < numValues; i++ ){
				fread( &(frameData.yValues[i]), sizeof( float ), 1, scannerFile );
			}
		}
		if ( scannerFileheader.content & DATA_GREY ){
			for ( unsigned long i = 0; i < numValues; i++ ){
				fread( &(frameData.greyValues[i]), sizeof( unsigned char ), 1, scannerFile );
			}
		}

		// Create point set
		PointSet* ps = new PointSet();
		VertexDescriptor vd;
		vd.pushAttrib( VAD( "position", 3, VAD::DATA_FORMAT_FLOAT32 ) );
		vd.pushAttrib( VAD( "color", 3, VAD::DATA_FORMAT_FLOAT32 ) );
		if ( computeNormals ) vd.pushAttrib( VAD( "normal", 3, VAD::DATA_FORMAT_FLOAT32 ) );
		vd.pushAttrib( VAD( "flags", 1, VAD::DATA_FORMAT_INT32 ) );
		ps->clearAndSetup( 1, numValues, &vd );
		AAT POSITION = ps->getAAT( "position" );
		AAT COLOR = ps->getAAT( "color" );
		card32 pIndex = 0;
		for ( unsigned int n = 0; n < numValues; n++ ){
			if ( !frameData.masks[n] ){
				Vector3f p = makeVector3f( frameData.xValues[n], frameData.yValues[n], frameData.zValues[n] );
				ps->set3f( pIndex, POSITION, p );
				Vector3f color = makeVector3f( frameData.greyValues[n], frameData.greyValues[n], frameData.greyValues[n] );
				ps->set3f( pIndex, COLOR, color );
				pIndex++;
			}
		}
		ps->changeHeight( pIndex );
		if ( computeNormals ) NormalEstimator::estimateNormals( ps, 20 );
		UnstructuredInCorePointCloud* upc = new UnstructuredInCorePointCloud();
		upc->setMaterialIndex( 5 );
		upc->setPointSet( ps );

		// Save in animation sequence
		std::string nodeName = animNode->getName() + "_Frame" + QString::number( nodeIndex ).toStdString();
		std::string nodeNameWithPath = "root/" + animNode->getName() + "/" + nodeName;
		SGObjectNode* node = new SGObjectNode();
		animNode->addNode( node );
		if ( node != NULL ){
			GenericSwapAllOutOfCorePointCloud *newGSA = new GenericSwapAllOutOfCorePointCloud();
			std::string m_oocSwapFileNamePath = "C:/Temp/";
			node->setName( nodeName );
			newGSA->clearAndSetup( upc, m_oocSwapFileNamePath + node->getName()+"_anim.swap.object", true );
			node->clearAndSetup( newGSA );
			delete upc;	
		}

		progressWindow->progress( float32 ( scannerFileCurrentFrame - startFrame ) * 100 / float32( endFrame - startFrame ) );

		scannerFileCurrentFrame++;
		nodeIndex++;
	}
	progressWindow->popSteps();
}

