//---------------------------------------------------------------------------
#include "stdafx.h"
//---------------------------------------------------------------------------
#include "PCCMLSHoleFiller.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "PointSetANNQuery.h"
#include "SparseGrid.h"
#include "SparseGrid.inline.h"
#include "Monomials2ndOrder.h"
#include "SeparatorClassProperty.h"
#include "ClassMethods.h"
#include "ProgressWindow.h"
#include "ObjectViewsTable.h"
#include "PointCloudTools.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

IMPLEMENT_X4_CLASS(PCCMLSHoleFiller, 3)
{
	BEGIN_CLASS_INIT(PCCMLSHoleFiller)
	ADD_SEPARATOR("Parameters")
   ADD_FLOAT32_PROP(fillRadius, 0)
   ADD_FLOAT32_PROP(fitRadius, 1)
   ADD_FLOAT32_PROP(sampleGrid, 0)
   ADD_BOOLEAN_PROP(selectedOnly, 0)
   ADD_BOOLEAN_PROP(quadraticPrec, 3)
	ADD_SEPARATOR("Mark Low Density Points")
	ADD_FLOAT32_PROP(markSampleSpacing, 0)
   ADD_CARD32_PROP(markSamSpNumNeighbors, 0)
	ADD_SEPARATOR("Mark Outliers")
   ADD_FLOAT32_PROP(maxRelResiduum, 2)
	
	ADD_NOARGS_METHOD(markLowDensityPoints)
	SET_METHOD_AS_CMD_OBJ(true)
	ADD_NOARGS_METHOD(markOutliers)
	SET_METHOD_AS_CMD_OBJ(true)
	ADD_NOARGS_METHOD(clearSelection)
	SET_METHOD_AS_CMD_OBJ(true)
}

//---------------------------------------------------------------------------

PCCMLSHoleFiller::PCCMLSHoleFiller() {
   fillRadius = 0.005f;
   fitRadius = 0.01f;
   sampleGrid = 0.0025f;
   maxRelResiduum = 1;
   selectedOnly = false;
   quadraticPrec = false;
   markSampleSpacing = 0.005f;
   markSamSpNumNeighbors = 20;
}

PCCMLSHoleFiller::~PCCMLSHoleFiller() {}

void PCCMLSHoleFiller::assign(const Object* obj, X4CopyContext *context)
{
	copyObjectProperties(obj, this);
}

//---------------------------------------------------------------------------

void PCCMLSHoleFiller::execute(Scene *scene) 
{
   int32 subdivisions = ceil(fillRadius / sampleGrid);
   if (subdivisions <= 1) {
      warning("sample grid too sparse - will not have any effect. aborted.");
      return;
   }
   std::vector<std::string> selection = scene->getRootState()->staticState->getSelectedNodes();
   for (size_t m = 0; m < selection.size(); m++ ){
   	progressWindow->pushStep(true, "upsamling");
      mpcard numAdded = 0;
      mpcard numUndersampled = 0;
      mpcard outliers = 0;
      PointCloud *pc = getPointCloud( scene, selection[m] );
      UnstructuredInCorePointCloud *upc = dynamic_cast<UnstructuredInCorePointCloud*>(pc);
      if (!upc) {
         warning("currently only supports UnstructuredInCorePointClouds.");
         continue;
      }
		PointSet *ps = upc->getPointSet();
		PointSet *result = (PointSet*)ps->copy();
		if (   !ps->providesAttribute("position", 3, VAD::DATA_FORMAT_FLOAT32)
		    || !ps->providesAttribute("normal", 3, VAD::DATA_FORMAT_FLOAT32)
		    || !ps->providesAttribute("flags", 1, VAD::DATA_FORMAT_INT32) ) 
      {
		    warning("needs position, flags and normal channels");
		    continue;
      }
		AAT normalAAT = ps->getAAT("normal");
		AAT positionAAT = ps->getAAT("position");
		AAT flagsAAT = ps->getAAT("flags");
		const mpcard n = ps->getNumEntries();
		PointSetRangeQuery query(ps, 4*fitRadius*fitRadius);
		const float32 subdivDist = fillRadius/(float32)subdivisions;
   	std::vector<Vector3f> posBuffer;
   	SparseGrid<3, bool, Object> occGrid;
		for (mpcard ptIndex=0; ptIndex<n; ptIndex++) {
		   Vector3f pos = ps->get3f(ptIndex, positionAAT);
		   Vector3i cell = makeVector3i((int32)(pos[0]/sampleGrid), (int32)(pos[1]/sampleGrid), (int32)(pos[2]/sampleGrid));
		   occGrid.setObject(cell, true);
      }
		for (mpcard ptIndex=0; ptIndex<n; ptIndex++) {
   	   if ((ptIndex % 1000) == 0) {
   	      if( !progressWindow->progress((float)ptIndex/(float)n*100.0f) ) break;
   	   }
		   Vector3f currentPosition = ps->get3f(ptIndex, positionAAT);
		   Vector3f currentNormal = normalize(ps->get3f(ptIndex, normalAAT));
		   Vector3f current_tangent_u = makeOrthonormalVector(currentNormal);
		   Vector3f current_tangent_v = current_tangent_u.crossProduct(currentNormal);
		   if (selectedOnly && !(ps->get1i(ptIndex, flagsAAT) & PF_FLAG_SELECTED)) {
		      continue;
		   }

	      for (int32 v=-subdivisions; v<subdivisions; v++) {
   		   for (int32 u=-subdivisions; u<subdivisions; u++) {
   		      float32 fu = (float32)u*subdivDist;
   		      float32 fv = (float32)v*subdivDist;
   		      Vector3f startPoint = currentPosition + current_tangent_u*fu + current_tangent_v *fv;
   		      
		         query.setTargetPoint(startPoint);
		         card32 localNumPoints;
		         int32 *pts = query.getAllIndices(localNumPoints);

               if (localNumPoints < 10) {
                  numUndersampled++;
                  continue;
               }

		         PCA<float,3> pca;
		         for (card32 i=0; i<localNumPoints; i++) {
			         Vector3f sample_pos = ps->get3f(pts[i], positionAAT);
			         float dist_2 = normQuad(sample_pos - startPoint);
			         float w = exp(-dist_2/(fillRadius*fillRadius));
			         pca.addPoint(sample_pos, w);
		         }
               float32 normalDisplacement;
		         Vector3f pcaNormal;
		         {
			         Vector3f eigenValues,centroid;
			         Matrix3f eigenVectors;
			         pca.analyze(eigenValues,eigenVectors,centroid);
			         if( eigenVectors[2] * currentNormal < 0.0f )
				         eigenVectors *= -1.0f;
			         pcaNormal = normalize(eigenVectors[2]);
			         normalDisplacement = (centroid - startPoint)*pcaNormal;
		         }

               if (quadraticPrec) {
		            Matrix3f tangentSystem = calcTangentSystem( pcaNormal );
		            Matrix3f invTangentSystem = tangentSystem.transpose();

		            if( posBuffer.size() < localNumPoints )
		            {
			            posBuffer.resize(localNumPoints);
		            }

		            for(card32 i=0; i<localNumPoints; i++)
		            {
			            Vector3f tmp = ps->get3f(pts[i], positionAAT) - startPoint;
			            posBuffer[i] = invTangentSystem * tmp / fitRadius;
		            }

		            StaticVector<float32, 6> coeff;
		            Monomials2ndOrder::computeCoeffs( &posBuffer[0], (card32)localNumPoints, coeff, 1 );

                  coeff[0] *= fitRadius;
                  normalDisplacement = coeff[0];
               }
               
               Vector3f newPos = startPoint + pcaNormal*normalDisplacement;
               
      		   Vector3i cell = makeVector3i((int32)(newPos[0]/sampleGrid), (int32)(newPos[1]/sampleGrid), (int32)(newPos[2]/sampleGrid));
      		   bool forgetit;
               if (!occGrid.getObject(cell, forgetit)) {
                  occGrid.setObject(cell, true);               
                  result->addPoint(ps->getDataPointer(ptIndex));
                  result->set3f(result->getNumEntries()-1, positionAAT, newPos);
                  result->set3f(result->getNumEntries()-1, normalAAT, pcaNormal);
                  numAdded++;
               }
		      }
   	   }
   	}
   	progressWindow->popSteps();
		
		upc->setPointSet(result);
		debugOutput << "num added: " << numAdded << "\n";
		debugOutput << "numUndersampled (less than 10 neighbors, skipped): " << numUndersampled << "\n";
		debugOutput << "outliers (rejected): " << outliers << "\n";
      upc->deleteDelOnWriteAttachments();
   }
}

void PCCMLSHoleFiller::markLowDensityPoints() 
{
   std::vector<std::string> selection = scene->getRootState()->staticState->getSelectedNodes();
   for (size_t m = 0; m < selection.size(); m++ ){
      PointCloud *pc = getPointCloud( scene, selection[m] );
      UnstructuredInCorePointCloud *upc = dynamic_cast<UnstructuredInCorePointCloud*>(pc);
      if (!upc) {
         warning("currently only supports UnstructuredInCorePointClouds.");
         continue;
      }
		PointSet *ps = upc->getPointSet();
		PointSet *result = (PointSet*)ps->copy();
		if (   !ps->providesAttribute("position", 3, VAD::DATA_FORMAT_FLOAT32)
		    || !ps->providesAttribute("flags", 1, VAD::DATA_FORMAT_INT32) ) 
      {
		    warning("needs position, flags channels");
		    continue;
      }
		AAT positionAAT = ps->getAAT("position");
		AAT flagsAAT = ps->getAAT("flags");
		const mpcard n = ps->getNumEntries();
		PointSetANNQuery queryKNN(ps, markSamSpNumNeighbors);
		for (mpcard ptIndex=0; ptIndex<n; ptIndex++) {
		   Vector3f currentPosition = ps->get3f(ptIndex, positionAAT);

         queryKNN.setTargetPoint(currentPosition);
         card32 localNumPoints;
         int32 *pts = queryKNN.getAllIndices(localNumPoints);

         if (localNumPoints < 2) {
            continue;
         }
         float32 avDist = 0;
         for (card32 i=0; i<localNumPoints; i++) {
            Vector3f otherPos = ps->get3f(pts[i], positionAAT);
            avDist += norm(currentPosition - otherPos);
         }
         avDist /= localNumPoints;
         if (avDist >= markSampleSpacing) {
            ps->set1i(ptIndex, flagsAAT, ps->get1i(ptIndex, flagsAAT) | PF_FLAG_SELECTED);
         }
      }
      upc->deleteDelOnWriteAttachments();
   }
   selectedOnly = true;
   ObjectViewsTable::update(this);
}

void PCCMLSHoleFiller::clearSelection() 
{
   std::vector<std::string> selection = scene->getRootState()->staticState->getSelectedNodes();
   for (size_t m = 0; m < selection.size(); m++ ){
      PointCloud *pc = getPointCloud( scene, selection[m] );
      UnstructuredInCorePointCloud *upc = dynamic_cast<UnstructuredInCorePointCloud*>(pc);
      if (!upc) {
         warning("currently only supports UnstructuredInCorePointClouds.");
         continue;
      }
		PointSet *ps = upc->getPointSet();
		PointSet *result = (PointSet*)ps->copy();
		if ( !ps->providesAttribute("flags", 1, VAD::DATA_FORMAT_INT32) ) 
      {
		    warning("needs flags channel");
		    continue;
      }
		AAT positionAAT = ps->getAAT("position");
		AAT flagsAAT = ps->getAAT("flags");
		const mpcard n = ps->getNumEntries();
		for (mpcard ptIndex=0; ptIndex<n; ptIndex++) {
         ps->set1i(ptIndex, flagsAAT, ps->get1i(ptIndex, flagsAAT) & (0xFFFFFFFF ^ PF_FLAG_SELECTED));
      }
      upc->deleteDelOnWriteAttachments();
   }
   selectedOnly = false;
   ObjectViewsTable::update(this);
}

void PCCMLSHoleFiller::markOutliers() 
{
   std::vector<std::string> selection = scene->getRootState()->staticState->getSelectedNodes();
   for (size_t m = 0; m < selection.size(); m++ ){
   	progressWindow->pushStep(true, "marking outliers");
      mpcard outliers = 0;
      PointCloud *pc = getPointCloud( scene, selection[m] );
      UnstructuredInCorePointCloud *upc = dynamic_cast<UnstructuredInCorePointCloud*>(pc);
      if (!upc) {
         warning("currently only supports UnstructuredInCorePointClouds.");
         continue;
      }
		PointSet *ps = upc->getPointSet();
		if (   !ps->providesAttribute("position", 3, VAD::DATA_FORMAT_FLOAT32)
		    || !ps->providesAttribute("normal", 3, VAD::DATA_FORMAT_FLOAT32)
		    || !ps->providesAttribute("flags", 1, VAD::DATA_FORMAT_INT32) ) 
      {
		    warning("needs position, flags and normal channels");
		    continue;
      }
		AAT normalAAT = ps->getAAT("normal");
		AAT positionAAT = ps->getAAT("position");
		AAT flagsAAT = ps->getAAT("flags");
		const mpcard n = ps->getNumEntries();
		PointSetRangeQuery query(ps, 4*fitRadius*fitRadius);
   	std::vector<Vector3f> posBuffer;
   	SparseGrid<3, bool, Object> occGrid;
		for (mpcard ptIndex=0; ptIndex<n; ptIndex++) {
		   Vector3f pos = ps->get3f(ptIndex, positionAAT);
		   Vector3i cell = makeVector3i((int32)(pos[0]/sampleGrid), (int32)(pos[1]/sampleGrid), (int32)(pos[2]/sampleGrid));
		   occGrid.setObject(cell, true);
      }
		for (mpcard ptIndex=0; ptIndex<n; ptIndex++) {
   	   if ((ptIndex % 1000) == 0) {
   	      if( !progressWindow->progress((float)ptIndex/(float)n*100.0f) ) break;
   	   }
		   Vector3f currentPosition = ps->get3f(ptIndex, positionAAT);
		   Vector3f currentNormal = normalize(ps->get3f(ptIndex, normalAAT));
		   Vector3f current_tangent_u = makeOrthonormalVector(currentNormal);
		   Vector3f current_tangent_v = current_tangent_u.crossProduct(currentNormal);

         query.setTargetPoint(currentPosition);
         card32 localNumPoints;
         int32 *pts = query.getAllIndices(localNumPoints);

         if (localNumPoints < 10) {
            continue;
         }

         PCA<float,3> pca;
         for (card32 i=0; i<localNumPoints; i++) {
	         Vector3f sample_pos = ps->get3f(pts[i], positionAAT);
	         float dist_2 = normQuad(sample_pos - currentPosition);
	         float w = exp(-dist_2/(fillRadius*fillRadius));
	         pca.addPoint(sample_pos, w);
         }
         Vector3f pcaNormal;
         float32 normalDisplacement;
         {
	         Vector3f eigenValues,centroid;
	         Matrix3f eigenVectors;
	         pca.analyze(eigenValues,eigenVectors,centroid);
	         if( eigenVectors[2] * currentNormal < 0.0f )
		         eigenVectors *= -1.0f;
	         pcaNormal = normalize(eigenVectors[2]);
            normalDisplacement = (centroid - currentPosition)*pcaNormal;
         }

         if (quadraticPrec) {
            Matrix3f tangentSystem = calcTangentSystem( pcaNormal );
            Matrix3f invTangentSystem = tangentSystem.transpose();

            if( posBuffer.size() < localNumPoints )
            {
	            posBuffer.resize(localNumPoints);
            }
            
            //debugOutput << localNumPoints << "\n";

            for(card32 i=0; i<localNumPoints; i++)
            {
	            Vector3f tmp = ps->get3f(pts[i], positionAAT) - currentPosition;
	            posBuffer[i] = invTangentSystem * tmp / fitRadius;
            }

            StaticVector<float32, 6> coeff;
            Monomials2ndOrder::computeCoeffs( &posBuffer[0], (card32)localNumPoints, coeff, 1 );

            coeff[0] *= fitRadius;
            normalDisplacement = coeff[0];
         }
         

         if (fabs(normalDisplacement) > maxRelResiduum*sampleGrid) {
            outliers++;
            ps->set1i(ptIndex, flagsAAT, ps->get1i(ptIndex, flagsAAT) | PF_FLAG_SELECTED);
         }
   	}
   	progressWindow->popSteps();
		
		debugOutput << "outliers (rejected): " << outliers << "\n";
      upc->deleteDelOnWriteAttachments();
   }
}




