#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PCCNormalTools.h"
//---------------------------------------------------------------------------
#include "Scene.h"
#include "HierarchicalKNNIterator.h"
#include "SceneGraphState.h"
#include "PointCloud.h"
#include "PointCloudTools.h"
#include "PointCloudIterators.h"
#include "HierarchicalKNNIterator.h"
#include "StandardSceneGraphIterator.h"
#include "SGObjectNode.h"
#include "ProgressWindow.h"
#include "ClassMethods.h"
#include "SelectionTools.h"
#include "TriangleMeshTools.h"
//---------------------------------------------------------------------------
using namespace X4;

IMPLEMENT_X4_CLASS( PCCNormalTools ,1)
{
	BEGIN_CLASS_INIT( PCCNormalTools );
	ADD_FLOAT32_PROP( m_Sigma, 0 );
	ADD_MAX_PLATFORM_CARD_PROP(m_NumNeighbors, 1);
	ADD_NOARGS_METHOD_DESC( computeMeshNormalsFromTopNeighborsByPCA, "calculate normals from topological neighbors keeping their orientation");
	SET_METHOD_AS_CMD_OBJ( true )
	ADD_NOARGS_METHOD_DESC( calculateRangeImageNormals, "calculate normals facing to the current camera position");
	SET_METHOD_AS_CMD_OBJ( true )
	ADD_NOARGS_METHOD_DESC( recalculateBadCalculatedNormals, "recalculates normals from other scans with better view angle");
	SET_METHOD_AS_CMD_OBJ( true )
	ADD_NOARGS_METHOD(smoothNormals);
	SET_METHOD_AS_CMD_OBJ( true );
	ADD_NOARGS_METHOD(flipNormalsToCamera);
	SET_METHOD_AS_CMD_OBJ( true );
	ADD_NOARGS_METHOD(flipSelectedNormals);
	SET_METHOD_AS_CMD_OBJ( true );
	ADD_NOARGS_METHOD(flipSelectedNormalsToCamera);
	SET_METHOD_AS_CMD_OBJ( true );
	ADD_NOARGS_METHOD(recomputeAndOrientNormalsKNN);
	SET_METHOD_AS_CMD_OBJ( true );

}

PCCNormalTools::PCCNormalTools(void)
{
	m_Sigma = 2.5f;
	m_NumNeighbors = 40;
}

PCCNormalTools::~PCCNormalTools(void)
{
}

void PCCNormalTools::assign(const Object* obj, X4CopyContext *context)
{
	PCCommand::assign(obj,context);
	const PCCNormalTools * o = dynamic_cast<const PCCNormalTools*>(obj);
	x4Assert( o != NULL );

}
void PCCNormalTools::execute(Scene *scene) 
{
}

void PCCNormalTools::computeMeshNormalsFromTopNeighborsByPCA() 
{
	std::vector<UnstructuredInCoreTriangleMesh*> sel = getSelectedPointCloudsAs<UnstructuredInCoreTriangleMesh>(scene);
	for( int i1 = 0; i1 < (int)sel.size(); ++ i1 )
		computeTriangleMeshNormalsByPCA(sel[i1]);
}

void PCCNormalTools::calculateRangeImageNormals() 
{
	// get camera position from scene
	Vector3f camPos = getScene()->getRootState()->staticState->camera.getPosition();

	unsigned int m;
   std::vector<std::string> selection = getScene()->getRootState()->staticState->getSelectedNodes();
   PointCloud* pc;
   for ( m = 0; m < selection.size(); m++ )
	{
      pc = getPointCloud( getScene(), selection[m] );
		if( pc )
		{
			float averagePointDistance = getMedianPointDistance(pc);
			float sigma = averagePointDistance * m_Sigma;
			calculateRangeImageNormals( pc, sigma, camPos );
		}

   }

}

void PCCNormalTools::calculateRangeImageNormals( PointCloud * pc, float sigma, Vector3f cameraPosition ) 
{
	BoundingBox3f bb = getPCBBox(pc);
	debugOutput << "BoundingBox_sidelength: " << bb.getSideLength(0) << "," << bb.getSideLength(1) << "," << bb.getSideLength(2) << "\n";
	checkAttribute( pc, "normal", 3, VAD::DATA_FORMAT_FLOAT32 );
	checkAttribute( pc, "camAngle", 1, VAD::DATA_FORMAT_FLOAT32 );

	// setup iterators and AATs
	SceneObjectIterator * iter = pc->createIterator( SceneObjectIteratorType::CAP_BASIC_PC | 
																	SceneObjectIteratorType::CAP_MODIFYABLE_PC );
	x4Assert( iter );

	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(iter);

	HierarchicalKNNIterator * hIt = new HierarchicalKNNIterator( pc, 32, NULL );

	AAT posAAT = pc->getAAT( "position" );
	AAT normalAAT = pc->getAAT( "normal" );
	AAT camAngleAAT = pc->getAAT( "camAngle" );

	// kernel support radius x = ln 0.01 * s^2
	float sigma_2 = sigma * sigma;
	float kernelRadius = 2.5f * sigma;
	float sigma_2_inv = 1.0f / sigma_2;

	
	progressWindow->pushStep( true, "calc normals" );
	card32 currentPointIndex = 0;
	std::vector<Vector3f> points( 100 );
	while( !bIt->atEnd() )
	{
		Vector3f pos = bIt->get3f(posAAT);
		hIt->setSeekPointAndReset(pos);

		float summed_weight = 0;
		card32 numPoints = 0;
		Vector3f centroid = NULL_VECTOR3F;

		// collect points and calulate centroid
		while(!hIt->atEnd() )
		{
			Vector3f tmpPos = hIt->get3f( posAAT );
			float dist = norm( tmpPos - pos );
			if( dist > kernelRadius && numPoints > 10)
				break;


			while( numPoints >= points.size() )
				points.push_back( NULL_VECTOR3F ); // grow point buffer

			float weight = exp( - dist * dist * sigma_2_inv );
			centroid += tmpPos * weight;
			summed_weight += weight;
			
			points[numPoints] = tmpPos;
			numPoints++;

			hIt->next();
		}

		centroid /= summed_weight;

		if( numPoints > 1 )
		{
			// calculate covariance matrix:
			Matrix3f C;
			C *= 0.0f;
			for( card32 k=0;k<numPoints;k++ )
			{
				Vector3f p = points[k];
				float dist_2 = normQuad( p - pos );
				float weight = exp( - dist_2 * sigma_2_inv );
				for ( unsigned int i = 0; i < 3; i++ )
				{
					for ( unsigned int j = 0; j <= i; j++ )
					{
						C[i][j] += ( p[i]-centroid[i]) * (p[j]-centroid[j]) * weight;
						C[j][i] = C[i][j];
					}
				}
			}

			// compute eigenvectors
			Vector3f eigenValues;
			Matrix3f eigenVectors;
			C.computeEigenStructure( eigenValues, eigenVectors );

			// calculate normal
			Vector3f normal = eigenVectors[0].crossProduct( eigenVectors[1] );
			normal.normalize();

			// flip normals which aren't facing to the camera
			Vector3f camDir = normalize( pos - cameraPosition);
			if( normal * camDir > 0 )
				normal = -normal;
			mIt->set1f( camAngleAAT, -normal * camDir );

			// store normal in point cloud
			mIt->set3f( normalAAT, normal );
		}
		else
			debugOutput << "Point " << currentPointIndex << " has no neighbours within kernelraidius!\n";

		currentPointIndex++;
		if( currentPointIndex % 300 == 1 )
			if( !progressWindow->progress((float)currentPointIndex/(float)pc->getNumPoints() * 100.0f ) )
				break;

		bIt->next();
	}

	progressWindow->popSteps();

	delete hIt;
	delete iter;
}

struct TPC_RecalcBadNormals
{
	PointCloud * pc;
	Matrix3f rot;
	Matrix3f invrot;
	Vector3f transl;
};

void PCCNormalTools::recalculateBadCalculatedNormals() 
{
	unsigned int m;
   std::vector<std::string> selection = getScene()->getRootState()->staticState->getSelectedNodes();
   std::vector<TPC_RecalcBadNormals> pcs;
	UnstructuredInCorePointCloud * tmpPC = new UnstructuredInCorePointCloud;
   for ( m = 0; m < selection.size(); m++ )
	{
      PointCloud * pc = getPointCloud( getScene(), selection[m] );
		if( !pc )
			continue;

		TPC_RecalcBadNormals tmp;
		tmp.pc = pc;
		SGObjectNode * node = dynamic_cast<SGObjectNode*>( getSceneGraphNodeByName( getScene(), selection[m]) );
		if( node )
		{
			Matrix4f transf = node->getSceneObject()->getTransformation();
			tmp.rot = shrink4To3( transf );
			tmp.invrot = tmp.rot.transpose();
			tmp.transl = shrink4To3( transf[3] );
		}
		else
		{
			throw "muh";	// shouldn't happen
		}

		if( pcs.size() == 0 )
			tmpPC->assign( pc );
		else
		{
			UnstructuredInCorePointCloud * upc = dynamic_cast<UnstructuredInCorePointCloud *>(pc);
			PointSet * ps = upc->getPointSet();
			PointSet * destps = tmpPC->getPointSet();
			std::vector<char> buffer(destps->getDescr()->getSize() );
			AAT posAAT = destps->getAAT( "position" );
			AAT normalAAT = destps->getAAT( "normal" );
			for( card32 i=0;i<ps->getNumEntries();i++ )
			{
				ps->getVertex( i, &buffer[0] );
				destps->addPoint( &buffer[0] );
				card32 index = destps->getNumEntries()-1;
				Vector3f pos = destps->get3f( index, posAAT );
				Vector3f normal = destps->get3f(index,normalAAT );
				destps->set3f( index, posAAT, tmp.rot * pos + tmp.transl);
				destps->set3f( index, normalAAT, tmp.rot * normal );
			}
		}

		pcs.push_back( tmp );


	}

	float averagePointDistance = getAveragePointDistance(tmpPC);
	float sigma = averagePointDistance * m_Sigma;
	float sigma_2 = sigma * sigma;
	float kernelRadius = sqrt( -log(0.01f) * sigma_2 );
	float sigma_2_inv = 1.0f / sigma_2;
	HierarchicalKNNIterator hIt( tmpPC,32,NULL);
	std::vector<Vector3f> points( 100 );

	for( card32 i=0;i<pcs.size();i++ )
	{
		PointSet * ps = ((UnstructuredInCorePointCloud*)pcs[i].pc)->getPointSet();
		AAT posAAT = ps->getAAT( "position" );
		AAT normalAAT = ps->getAAT("normal");
		AAT angleAAT = ps->getAAT("camAngle");
		AAT colorAAT = ps->getAAT("color");

		for( card32 j=0;j<ps->getNumEntries();j++ )
		{
			float camAngle = ps->get<float,1>(j, angleAAT)[0];
			//if( camAngle > 0.4f )
			//	continue;

			
			Vector3f pos = ps->get3f(j,posAAT);
			pos = pcs[i].rot * pos + pcs[i].transl;

			hIt.setSeekPointAndReset(pos);

			float summed_weight = 0;
			card32 numPoints = 0;
			Vector3f centroid = NULL_VECTOR3F;
			float bestCamAngle = 0;
			Vector3f bestCamAngleNormal;

			// collect points and calulate centroid
			while(!hIt.atEnd() )
			{
				Vector3f tmpPos = hIt.get3f( posAAT );
				float dist = norm( tmpPos - pos );
				if( dist > kernelRadius && numPoints > 10)
					break;

				float tmpAngle = hIt.get1f(angleAAT);
				if( tmpAngle < 0.4f )
				{
					hIt.next();
					continue;
				}
				if( tmpAngle > bestCamAngle )
				{
					bestCamAngle = tmpAngle;
					bestCamAngleNormal = hIt.get3f( normalAAT);
				}

				while( numPoints >= points.size() )
					points.push_back( NULL_VECTOR3F ); // grow point buffer

				float weight = exp( - dist * dist * sigma_2_inv );
				centroid += tmpPos * weight;
				summed_weight += weight;
				
				points[numPoints] = tmpPos;
				numPoints++;

				hIt.next();
			}

			centroid /= summed_weight;

			if( numPoints > 1 )
			{
				// calculate covariance matrix:
				Matrix3f C;
				C *= 0.0f;
				for( card32 k=0;k<numPoints;k++ )
				{
					Vector3f p = points[k];
					float dist_2 = normQuad( p - pos );
					float weight = exp( - dist_2 * sigma_2_inv );
					for ( unsigned int i = 0; i < 3; i++ )
					{
						for ( unsigned int j = 0; j <= i; j++ )
						{
							C[i][j] += ( p[i]-centroid[i]) * (p[j]-centroid[j]) * weight;
							C[j][i] = C[i][j];
						}
					}
				}

				// compute eigenvectors
				Vector3f eigenValues;
				Matrix3f eigenVectors;
				C.computeEigenStructure( eigenValues, eigenVectors );

				// calculate normal
				Vector3f normal = eigenVectors[0].crossProduct( eigenVectors[1] );
				normal.normalize();

				if( normal * bestCamAngleNormal < 0 )
					normal = -normal;

				// store normal in point cloud
				ps->set3f( j, normalAAT, pcs[i].invrot * normal );
			}

		}

	}

	delete tmpPC;
}

void X4::PCCNormalTools::smoothNormals()
{
	unsigned int m;
	std::vector<std::string> selection = getScene()->getRootState()->staticState->getSelectedNodes();
	PointCloud* pc;
	for ( m = 0; m < selection.size(); m++ )
	{
		pc = getPointCloud( getScene(), selection[m] );
		if( pc )
		{
			float medPointDistance = getMedianPointDistance(pc);
			float sigma = medPointDistance * m_Sigma;
			smoothNormals( pc, sigma );
		}

	}

}

void X4::PCCNormalTools::smoothNormals( PointCloud * pc, float sigma )
{
	if( !pc->providesAttribute("normal"))
		return;

	// setup iterators and AATs
	SceneObjectIterator * iter = pc->createIterator( SceneObjectIteratorType::CAP_BASIC_PC | 
		SceneObjectIteratorType::CAP_MODIFYABLE_PC );
	x4Assert( iter );

	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(iter);

	HierarchicalKNNIterator * hIt = new HierarchicalKNNIterator( pc, 32, NULL );

	AAT posAAT = pc->getAAT( "position" );
	AAT normalAAT = pc->getAAT( "normal" );

	float sigma_2 = sigma * sigma;
	float kernelRadius_2 = 4.0f * sigma_2;
	float sigma_2_inv = 1.0f / sigma_2;


	progressWindow->pushStep( true, "smooth normals" );
	card32 currentPointIndex = 0;
	while( !bIt->atEnd() )
	{
		Vector3f pos = bIt->get3f(posAAT);
		hIt->setSeekPointAndReset(pos);

		card32 numPoints = 0;
		Vector3f averageNormal = NULL_VECTOR3F;

		// collect points and calulate centroid
		while(!hIt->atEnd() )
		{
			Vector3f tmpPos = hIt->get3f( posAAT );
			float dist_2 = normQuad( tmpPos - pos );
			if( dist_2 > kernelRadius_2 && numPoints >=1)
				break;

			float weight = exp( - dist_2 * sigma_2_inv );
			averageNormal += hIt->get3f(normalAAT) * weight;
			numPoints++;

			hIt->next();
		}

		averageNormal = normalize(averageNormal);
		mIt->set3f(normalAAT,averageNormal);

		currentPointIndex++;
		if( currentPointIndex % 300 == 1 )
			if( !progressWindow->progress((float)currentPointIndex/(float)pc->getNumPoints() * 100.0f ) )
				break;

		bIt->next();
	}

	progressWindow->popSteps();

	delete hIt;
	delete iter;

}

void X4::PCCNormalTools::flipNormalsToCamera()
{
	// get camera position from scene
	Vector3f camPos = getScene()->getRootState()->staticState->camera.getPosition();

	unsigned int m;
	std::vector<std::string> selection = getScene()->getRootState()->staticState->getSelectedNodes();
	PointCloud* pc;
	for ( m = 0; m < selection.size(); m++ )
	{
		pc = getPointCloud( getScene(), selection[m] );
		if( pc )
		{
			AAT posAAT = pc->getAAT("position");
			AAT normalAAT = pc->getAAT("normal");

			SceneObjectIterator * iter = pc->createIterator(SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC);
			BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
			ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(iter);
			RandomAccessPointCloudIterator * rIt = dynamic_cast<RandomAccessPointCloudIterator*>(iter);

			while(!bIt->atEnd())
			{
				Vector3f pos = bIt->get3f(posAAT);
				Vector3f normal = bIt->get3f(normalAAT);

				if( normal * (camPos - pos) < 0.0f )
					normal *= -1;

				mIt->set3f(normalAAT,normal);


				bIt->next();
			}

			delete iter;
		}
	}
}

void X4::PCCNormalTools::flipSelectedNormals()
{
	std::vector<std::string> selection = getScene()->getRootState()->staticState->getSelectedNodes();
	PointCloud* pc;
	for ( unsigned m = 0; m < selection.size(); m++ )
	{
		pc = getPointCloud( getScene(), selection[m] );
		if( pc )
		{
			AAT posAAT = pc->getAAT("position");
			AAT normalAAT = pc->getAAT("normal");
			AAT flagsAAT = pc->getAAT("flags");

			SceneObjectIterator * iter = pc->createIterator(SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC);
			BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
			ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(iter);
			RandomAccessPointCloudIterator * rIt = dynamic_cast<RandomAccessPointCloudIterator*>(iter);

			while(!bIt->atEnd())
			{
				Vector3f pos = bIt->get3f(posAAT);
				Vector3f normal = bIt->get3f(normalAAT);

				if( bIt->get1i(flagsAAT) & PF_FLAG_SELECTED )
				{
					normal *= -1;
					mIt->set3f(normalAAT,normal);
				}

				bIt->next();
			}

			delete iter;
		}
	}
}

void X4::PCCNormalTools::flipSelectedNormalsToCamera()
{
	// get camera position from scene
	Vector3f camPos = getScene()->getRootState()->staticState->camera.getPosition();

	std::vector<std::string> selection = getScene()->getRootState()->staticState->getSelectedNodes();
	PointCloud* pc;
	for ( unsigned m = 0; m < selection.size(); m++ )
	{
		pc = getPointCloud( getScene(), selection[m] );
		if( pc )
		{
			AAT posAAT = pc->getAAT("position");
			AAT normalAAT = pc->getAAT("normal");
			AAT flagsAAT = pc->getAAT("flags");

			SceneObjectIterator * iter = pc->createIterator(SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC);
			BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
			ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(iter);
			RandomAccessPointCloudIterator * rIt = dynamic_cast<RandomAccessPointCloudIterator*>(iter);

			while(!bIt->atEnd())
			{
				Vector3f pos = bIt->get3f(posAAT);
				Vector3f normal = bIt->get3f(normalAAT);

				if( bIt->get1i(flagsAAT) & PF_FLAG_SELECTED )
				{
					if( normal * (camPos - pos) < 0.0f )
						normal *= -1;
					mIt->set3f(normalAAT,normal);
				}

				bIt->next();
			}

			delete iter;
		}
	}
}

void X4::PCCNormalTools::recomputeAndOrientNormalsKNN( PointCloud * pc, mpcard numNeighbors )
{
	if( !pc->providesAttribute("normal") )
		throw X4Exception("PCCNormalTools::recomputeAndOrientNormalsKNN() - no normal channel found in point cloud!");

	PCA<float,3> pca;
	SceneObjectIterator * iter = pc->createIterator(SOIT::CAP_BASIC_PC|SOIT::CAP_MODIFYABLE_PC);
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(iter);
	AAT posAAT = pc->getAAT("position");
	AAT normalAAT = pc->getAAT("normal");

	HierarchicalKNNIterator hIt(pc,32,NULL);

	// first pass: compute pca normals, use previous orientation
	while (!bIt->atEnd())
	{
		Vector3f p = bIt->get3f(posAAT);
		Vector3f oldNormal = bIt->get3f(normalAAT);
		hIt.setSeekPointAndReset(p);
		pca.clear();
		mpcard numPoints = 0;
		while(!hIt.atEnd() && numPoints++ < numNeighbors)
		{
			pca.addPoint(hIt.get3f(posAAT));
			hIt.next();
		}
		Vector3f centroid,eigenValues;
		Matrix3f eigenVectors;
		pca.analyze(eigenValues,eigenVectors,centroid);

		if( oldNormal*eigenVectors[2] < 0)
			eigenVectors[2] *= -1;
		mIt->set3f(normalAAT,eigenVectors[2]);

		bIt->next();
	}

	// second pass: orient normals via bilateral voting
	bIt->reset();
	while (!bIt->atEnd())
	{
		mpint dontflip = 0;
		Vector3f p = bIt->get3f(posAAT);
		Vector3f normal = bIt->get3f(normalAAT);
		hIt.setSeekPointAndReset(p);
		mpcard numPoints = 0;
		while(!hIt.atEnd() && numPoints++ < numNeighbors)
		{
			Vector3f tmpNormal = hIt.get3f(normalAAT);
			dontflip += (tmpNormal*normal < 0)?-1:1;
			hIt.next();
		}
		
		if( dontflip < 0)
			mIt->set3f(normalAAT,-normal);

		bIt->next();
	}

	delete iter;
}

void X4::PCCNormalTools::recomputeAndOrientNormalsKNN()
{
	unsigned int m;
	std::vector<std::string> selection = getScene()->getRootState()->staticState->getSelectedNodes();
	PointCloud* pc;
	for ( m = 0; m < selection.size(); m++ )
	{
		pc = getPointCloud( getScene(), selection[m] );
		if( pc )
		{
			recomputeAndOrientNormalsKNN( pc, m_NumNeighbors );
		}

	}

}