//---------------------------------------------------------------------------
#include "StdAfx.h"

//---------------------------------------------------------------------------
#include "PCCParametricDerivativeEstimator.h"
//---------------------------------------------------------------------------
#include "PointCloud.h"
#include "Scene.h"
#include "PointCloudIterators.h"
#include "BooleanClassProperty.h"
#include "NumericalClassProperty.h"
#include "PointFlags.h"
#include "SGLegacyPCSceneAdaptor.h"
#include "LinearAlgebra.inline.h"
#include "GlobalMacros.h"
//---------------------------------------------------------------------------
 
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

MetaClass* PCCParametricDerivativeEstimator::PCCParametricDerivativeEstimatorMetaClass = NULL;

const MetaClass* PCCParametricDerivativeEstimator::getInstanceClass() const {
   return getClass();
}

Object* PCCParametricDerivativeEstimator::newInstance() {
   return new PCCParametricDerivativeEstimator;
}

const MetaClass* PCCParametricDerivativeEstimator::getClass() {
   if (PCCParametricDerivativeEstimatorMetaClass==NULL) {
      PCCParametricDerivativeEstimatorMetaClass = new MetaClass("PCCParametricDerivativeEstimator", "no description.",
                                  newInstance, sizeof(PCCParametricDerivativeEstimator));
   }
   return PCCParametricDerivativeEstimatorMetaClass;
}

void PCCParametricDerivativeEstimator::init(const MetaClass *parentClass) {
   ((MetaClass*)getClass())->init(parentClass);
   PCCParametricDerivativeEstimatorMetaClass->copyParentProperties();

   BEGIN_CLASS_INIT( PCCParametricDerivativeEstimator )
   ADD_BOOLEAN_PROP(estimateNormals,0)
   ADD_BOOLEAN_PROP(estimateTangents,0)
   ADD_BOOLEAN_PROP(normalizeNormals,0)
   ADD_BOOLEAN_PROP(normalizeTangents,0)
   ADD_BOOLEAN_PROP(invalidateOutliers,0)
   ADD_FLOAT32_PROP(outlierFactor,0)

}

void PCCParametricDerivativeEstimator::shutdown() {
   delete PCCParametricDerivativeEstimatorMetaClass;
}


// ----


PCCParametricDerivativeEstimator::PCCParametricDerivativeEstimator() {
   estimateNormals = true;
   estimateTangents = true;
   normalizeNormals = true;
   normalizeTangents = false;
   invalidateOutliers = true;
   outlierFactor = 3.0f;
}

bool PCCParametricDerivativeEstimator::checkNormalTangentsChannels(Scene *scene, card32 pointCloudIndex, PointCloud* &pc)
{
   SceneObjectIterator *it = pc->createIterator(SOIT(SOIT::CAP_BASIC_PC | SOIT::CAP_PARAMETRIC_PC));
   ModifyablePointCloudIterator *modIt = dynamic_cast<ModifyablePointCloudIterator*>(it);
   ParametricPointCloudIterator *paramIt = dynamic_cast<ParametricPointCloudIterator*>(it);
   if (modIt == NULL || paramIt == NULL) {
      return false;
   }
   delete it;

   bool needChange = false;
   if (estimateNormals) {
      if (!pc->providesAttribute("normal")) needChange = true;
   }
   if (estimateTangents) {
      if (!pc->providesAttribute("tangent_u")) needChange = true;
   }
   if (estimateTangents) {
      if (!pc->providesAttribute("tangent_v")) needChange = true;
   }
   SGLegacyPCSceneAdaptor sceneLA(scene);
   if (needChange) {
      VertexDescriptor newFields;
      if (estimateNormals) {
         newFields.pushAttrib(mVAD("normal", 3, VAD::DATA_FORMAT_FLOAT32));
      }
      if (estimateTangents) {
         newFields.pushAttrib(mVAD("tangent_u", 3, VAD::DATA_FORMAT_FLOAT32));
      }
      if (estimateTangents) {
         newFields.pushAttrib(mVAD("tangent_v", 3, VAD::DATA_FORMAT_FLOAT32));
      }
      PointCloud *newpc = pc->appendFields(&newFields);
      sceneLA.replacePointCloud(pointCloudIndex, newpc);
      pc = newpc;
   }
   return true;
}

void PCCParametricDerivativeEstimator::setAttribs(ModifyablePointCloudIterator *modIt, ParametricPointCloudIterator *paramIt,
   const card32 x, const card32 y, const Vector3f normal, const Vector3f tangentU, const Vector3f tangentV,
   const AAT NORMAL, const AAT TANGENT_U, const AAT TANGENT_V)
{
   if (estimateNormals) {
      paramIt->goTo(x,y);
      modIt->set3f(NORMAL, normal);
   }
   if (estimateTangents) {
      modIt->set3f(TANGENT_U, tangentU);
      modIt->set3f(TANGENT_V, tangentV);
   }
}

static bool isOutlier(float32 *tangentLengths, bool *invalidArray, int32 width, int32 height, int32 xpos, int32 ypos, float32 nFact) {
   x4Assert(xpos < width && ypos < height);
   float32 nl = 1E30F;
   if (xpos <= 1 || xpos >= width - 2 || ypos <= 1 || ypos >= height - 2) return true;
   int32 xl = xpos - 2; // if (xl < 0) xl = 0;
   int32 xr = xpos + 2; // if (xr >= width) xr = width-1;
   int32 yl = ypos - 2; // if (yl < 0) yl = 0;
   int32 yu = ypos + 2; // if (yu >= height) yu = height-1;
   for (int32 y = yl; y < yu; y++) {
      for (int32 x = xl; x < xr; x++) {
         if (x != xpos && y != ypos) {
            float32 l = tangentLengths[x + width*y];
            if (l < nl) nl = l;
            if (invalidArray[x +width*y]) return true;
         }
      }
   }
   float32 cl = tangentLengths[xpos + width*ypos];
   if (cl > nl*nFact) {
      return true;
   } else {
      return false;
   }
}

void PCCParametricDerivativeEstimator::execute(Scene *scene) {
   SGLegacyPCSceneAdaptor sceneLA(scene);
   std::vector<card32> selection = sceneLA.getSelectedPCs();

   if (selection.size() == 0) {
      //Application->MessageBox("Nothing selected.", "Cannot Execute Command...", MB_OK | MB_ICONERROR);
	   WARNINGMESSAGE("Nothing selected.");
      return;
   }
   const size_t n = selection.size();

   if (!estimateNormals && !estimateTangents) return;

   for (card32 sel=0; sel<n; sel++) {
      PointCloud *pc = sceneLA.getPointCloud(selection[sel]);
      if (!checkNormalTangentsChannels(scene, selection[sel], pc)) continue;
      SceneObjectIterator *it = pc->createIterator(SOIT(SOIT::CAP_BASIC_PC | SOIT::CAP_PARAMETRIC_PC));
      BasicPointCloudIterator *bIt = dynamic_cast<BasicPointCloudIterator*>(it);
      ModifyablePointCloudIterator *modIt = dynamic_cast<ModifyablePointCloudIterator*>(it);
      ParametricPointCloudIterator *paramIt = dynamic_cast<ParametricPointCloudIterator*>(it);
      if (modIt != NULL && paramIt != NULL) {
         const card32 width = paramIt->getWidth();
         const card32 height = paramIt->getHeight();

         if (!pc->providesAttribute("position")) continue;

         AAT POSITION = pc->getAAT("position");
         AAT FLAGS;
         AAT NORMAL;
         AAT TANGENT_U;
         AAT TANGENT_V;

         bool check_flags = false;
         if (    pc->providesAttribute("flags")
              && (*(pc->getDescr()))["flags"].dataFormat == VAD::DATA_FORMAT_INT32
              && (*(pc->getDescr()))["flags"].numComponents == 1)
         {
            FLAGS = pc->getAAT("flags");
            check_flags = true;
         }
         if (estimateNormals) NORMAL = pc->getAAT("normal");
         if (estimateTangents) {
            TANGENT_U = pc->getAAT("tangent_u");
            TANGENT_V = pc->getAAT("tangent_v");
         }

         float32 *tangentLengths = new float32[width*height];
         bool *invalid_array = new bool[width*height];

         for (card32 y=0; y<height-1; y++) {
            for (card32 x=0; x<width-1; x++) {
               bool invalid = false;

               paramIt->goTo(x,y);
               Vector3f pos = bIt->get3f(POSITION);
               if (check_flags) invalid = invalid && ((bIt->get1i(FLAGS) & PF_FLAG_INVALID) != 0);
               paramIt->goTo(x+1,y);
               Vector3f posDX = bIt->get3f(POSITION);
               if (check_flags) invalid = invalid && ((bIt->get1i(FLAGS) & PF_FLAG_INVALID) != 0);
               paramIt->goTo(x,y+1);
               Vector3f posDY = bIt->get3f(POSITION);
               if (check_flags) invalid = invalid && ((bIt->get1i(FLAGS) & PF_FLAG_INVALID) != 0);

               invalid_array[x + width*y] = (bIt->get1i(FLAGS) & PF_FLAG_INVALID) != 0 || invalid;

               if (invalid) {
                  paramIt->goTo(x,y);
                  modIt->set1i(FLAGS, bIt->get1i(FLAGS) | PF_FLAG_INVALID);
                  continue;
               }

               Vector3f tangentU = posDX - pos;
               Vector3f tangentV = posDY - pos;
               tangentLengths[x + width*y] = max(norm(tangentU), norm(tangentV));
               if (normalizeTangents) {
                  float32 n = norm(tangentU);
                  if (n >= 1E-7) tangentU /= n;
                  n = norm(tangentV);
                  if (n >= 1E-7) tangentV /= n;
               }
               Vector3f normal = tangentU.crossProduct(tangentV);
               if (normalizeNormals) {
                  float32 n = norm(normal);
                  if (n >= 1E-7) normal /= n;
               }

               setAttribs(modIt, paramIt, x, y, normal, tangentU, tangentV, NORMAL, TANGENT_U, TANGENT_V);

               if (x == width-2) {
                  setAttribs(modIt, paramIt, x+1, y, normal, tangentU, tangentV, NORMAL, TANGENT_U, TANGENT_V);
               }
               if (y == height-2) {
                  setAttribs(modIt, paramIt, x, y+1, normal, tangentU, tangentV, NORMAL, TANGENT_U, TANGENT_V);
                  if (x == width-2) {
                     setAttribs(modIt, paramIt, x+1, y+1, normal, tangentU, tangentV, NORMAL, TANGENT_U, TANGENT_V);
                  }
               }
            }
         }

         if (check_flags && invalidateOutliers) {
            for (card32 y=0; y<height-1; y++) {
               for (card32 x=0; x<width-1; x++) {
                  if (isOutlier(tangentLengths, invalid_array, width, height, x, y, outlierFactor)) {
                     paramIt->goTo(x,y);
                     modIt->set1i(FLAGS, (bIt->get1i(FLAGS) | PF_FLAG_INVALID));
                  }
               }
            }
         }

         for (card32 x=0; x<width; x++) {
            paramIt->goTo(x, height-1);
            modIt->set1i(FLAGS, (bIt->get1i(FLAGS) | PF_FLAG_INVALID));
         }
         for (card32 y=0; y<height; y++) {
            paramIt->goTo(width-1,y);
            modIt->set1i(FLAGS, (bIt->get1i(FLAGS) | PF_FLAG_INVALID));
         }


         // for (int32 y=0; y<(int32)height-1; y++) {
         //    for (int32 x=0; x<(int32)width-1; x++) {
         //       bool neighbourInv = false;
         // 
         //       int32 xxStart = x-1; if (xxStart <0) xxStart = 0;
         //       int32 yyStart = y-1; if (yyStart <0) yyStart = 0;
         //       int32 xxEnd = x+1; if (xxEnd >= (int32)width) xxEnd = width-1;
         //       int32 yyEnd = y+1; if (yyEnd >= (int32)height) yyEnd = height-1;
         // 
         //       for (int32 yy = yyStart; yy<=yyEnd; yy++) {
         //          for (int32 xx = xxStart; xx<=xxEnd; xx++) {
         //             neighbourInv |= invalid_array[xx + width*yy];
         //          }
         //       }
         // 
         //       if (neighbourInv) {
         //          paramIt->goTo(x,y);
         //          modIt->set1i(FLAGS, (bIt->get1i(FLAGS) | PF_FLAG_INVALID));
         //       }
         // 
         //    }
         // }


         delete[] tangentLengths;
         delete[] invalid_array;
      }
      delete it;
   }
}
