//---------------------------------------------------------------------------
#include "StdAfx.h"

//---------------------------------------------------------------------------
#include "PCCParametricDerivator.h"
//---------------------------------------------------------------------------
#include <vector>
#include <algorithm>
#include "Scene.h"
#include "PointCloud.h"
#include "PointCloudIterators.h"
#include "BooleanClassProperty.h"
#include "NumericalClassProperty.h"
#include "VertexDescriptor.h"
#include "SGLegacyPCSceneAdaptor.h"
#include "LinearAlgebra.inline.h"
#include "GlobalMacros.h"
//---------------------------------------------------------------------------
 
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

MetaClass* PCCParametricDerivator::PCCParametricDerivatorMetaClass = NULL;

const MetaClass* PCCParametricDerivator::getInstanceClass() const {
   return getClass();
}

Object* PCCParametricDerivator::newInstance() {
   return new PCCParametricDerivator;
}

const MetaClass* PCCParametricDerivator::getClass() {
   if (PCCParametricDerivatorMetaClass==NULL) {
      PCCParametricDerivatorMetaClass = new MetaClass("PCCParametricDerivator", "no description.",
                                  newInstance, sizeof(PCCParametricDerivator));
   }
   return PCCParametricDerivatorMetaClass;
}

void PCCParametricDerivator::init(const MetaClass *parentClass) {
   ((MetaClass*)getClass())->init(parentClass, MetaClass::FIRST_VERSION + 4);
   PCCParametricDerivatorMetaClass->copyParentProperties();

	BEGIN_CLASS_INIT( PCCParametricDerivator )
	ADD_BOOLEAN_PROP(m_EstimateNormals,0)
	ADD_BOOLEAN_PROP(m_EstimateTangents,0)
	ADD_BOOLEAN_PROP(m_NormalizeNormals,0)
	ADD_BOOLEAN_PROP(m_NormalizeTangents,0)
	ADD_FLOAT32_PROP(m_ScaleTangents,0)
	ADD_BOOLEAN_PROP(m_InvalidateOutliers,0)
	ADD_CARD32_PROP(m_InvalidateKernelSize,0)
	ADD_FLOAT32_PROP(m_MaxRelativeQuadSize,0)
	ADD_BOOLEAN_PROP(m_CheckAbsolute,0)
	ADD_FLOAT32_PROP(m_MaxAbsoluteQuadSize,0)
	ADD_BOOLEAN_PROP(m_CheckDiag,0)
	ADD_FLOAT32_PROP(m_MaxQuadDiag,0)
	ADD_FLOAT32_PROP(m_MaxQuadRatio,0)
	ADD_FLOAT32_PROP(m_MinTangentAngle,0)

}

void PCCParametricDerivator::shutdown() {
   delete PCCParametricDerivatorMetaClass;
}

// ----

PCCParametricDerivator::PCCParametricDerivator()
{
   m_MaxAbsoluteQuadSize = 0.01f;
   m_MaxRelativeQuadSize = 2.0f;
   m_ScaleTangents = 1.0f;
   m_MinTangentAngle = 50.0f;

   m_InvalidateKernelSize = 1;

   m_MaxQuadRatio = 7.0f;
   m_MaxQuadDiag = 0.1f;

   m_EstimateNormals = true;
   m_EstimateTangents = true;
   m_NormalizeNormals = true;
   m_NormalizeTangents = false;

   m_InvalidateOutliers = true;
   m_CheckAbsolute = false;
   m_CheckDiag = false;

}

PCCParametricDerivator::~PCCParametricDerivator()
{
}


bool PCCParametricDerivator::checkNormalTangentsChannels(Scene *scene, card32 pointCloudIndex, PointCloud* &pc)
{
   SGLegacyPCSceneAdaptor sceneLA(scene);
   SceneObjectIterator *it = pc->createIterator(SOIT(SOIT::CAP_BASIC_PC | SOIT::CAP_PARAMETRIC_PC | SOIT::CAP_MODIFYABLE_PC ));
   ModifyablePointCloudIterator *modIt = dynamic_cast<ModifyablePointCloudIterator*>(it);
   ParametricPointCloudIterator *paramIt = dynamic_cast<ParametricPointCloudIterator*>(it);
   if (modIt == NULL || paramIt == NULL) {
      return false;
   }
   delete it;

   bool needChange = false;
   if (m_EstimateNormals) {
      if (!pc->providesAttribute(getAttName_Normal())) needChange = true;
   }
   if (m_EstimateTangents) {
      if (!pc->providesAttribute(getAttName_TangentU())) needChange = true;
   }
   if (m_EstimateTangents) {
      if (!pc->providesAttribute(getAttName_TangentV())) needChange = true;
   }
   if (needChange) {
      VertexDescriptor newFields;
      if (m_EstimateNormals) {
         newFields.pushAttrib(mVAD( getAttName_Normal(), 3, VAD::DATA_FORMAT_FLOAT32));
      }
      if (m_EstimateTangents) {
         newFields.pushAttrib(mVAD(getAttName_TangentU(), 3, VAD::DATA_FORMAT_FLOAT32));
      }
      if (m_EstimateTangents) {
         newFields.pushAttrib(mVAD(getAttName_TangentV(), 3, VAD::DATA_FORMAT_FLOAT32));
      }
      PointCloud *newpc = pc->appendFields(&newFields);
      sceneLA.replacePointCloud(pointCloudIndex, newpc);
      pc = newpc;
   }
   return true;
}

void PCCParametricDerivator::execute(Scene *scene)
{
   if( !m_EstimateNormals && !m_EstimateTangents )
   {
      //Application->MessageBox("Nothing to do - chose estimate normals oder estimate tangents", "Cannot Execute Command...", MB_OK | MB_ICONERROR);
      WARNINGMESSAGE("Nothing to do - chose estimate normals oder estimate tangents");
		return;
   }

   SGLegacyPCSceneAdaptor sceneLA(scene);
   std::vector<card32> selection = sceneLA.getSelectedPCs();

   if (selection.size() == 0) {
     // Application->MessageBox("Nothing selected.", "Cannot Execute Command...", MB_OK | MB_ICONERROR);
		WARNINGMESSAGE("Nothing selected.");
      return;
   }

   for( card32 i=0; i<selection.size(); i++ )
   {
      estimate( scene, selection[i] );
   }

}

bool compareQuads( PCCParametricDerivator::TQuad *a, PCCParametricDerivator::TQuad  * b )
{
   return a->size < b->size;
}

#define getMyQuad( x, y ) quads[ (y)*(paramIt->getWidth()-1)+(x) ]

void PCCParametricDerivator::estimate(Scene *scene, card32 index)
{
   SGLegacyPCSceneAdaptor sceneLA(scene);
   PointCloud *pc = sceneLA.getPointCloud( index );
   if( !checkNormalTangentsChannels( scene, index, pc ) )
   {
      return;
   }

   SceneObjectIterator *it = pc->createIterator(SOIT(SOIT::CAP_BASIC_PC | SOIT::CAP_PARAMETRIC_PC | SOIT::CAP_MODIFYABLE_PC ));
   ModifyablePointCloudIterator *modIt = dynamic_cast<ModifyablePointCloudIterator*>(it);
   ParametricPointCloudIterator *paramIt = dynamic_cast<ParametricPointCloudIterator*>(it);
   BasicPointCloudIterator *bIt = dynamic_cast<BasicPointCloudIterator*>(it);

   AAT token_pos = pc->getAAT( getAttName_Position() );
   AAT token_normal = pc->getAAT( getAttName_Normal() );
   AAT token_tangentU = pc->getAAT( getAttName_TangentU() );
   AAT token_tangentV = pc->getAAT( getAttName_TangentV() );
   AAT token_flags = pc->getAAT( "flags" );

   // -----------------------------------------------
   // calculate quads
   TQuad q;
   q.normal = makeVector3f(0,0,0);
   q.tangent_u = makeVector3f(0,0,0);
   q.tangent_u = makeVector3f(0,0,0);
   std::vector<TQuad>  quads((paramIt->getWidth()-1) * (paramIt->getHeight()-1), q);
   for( card32 x=0;x<paramIt->getWidth()-1;x++ )
      for( card32 y=0;y<paramIt->getHeight()-1;y++ )
      {
         paramIt->goTo(x,y);
         Vector3f p1 = bIt->get3f( token_pos );
         int32 flag1 = bIt->get1i( token_flags );
         paramIt->goTo(x+1,y);
         Vector3f p2 = bIt->get3f( token_pos );
         int32 flag2 = bIt->get1i( token_flags );
         paramIt->goTo(x,y+1);
         Vector3f p3 = bIt->get3f( token_pos );
         int32 flag3 = bIt->get1i( token_flags );
         paramIt->goTo(x+1,y+1);
         Vector3f p4 = bIt->get3f( token_pos );
         int32 flag4 = bIt->get1i( token_flags );

         TQuad quad;
         quad.tangent_u = ((p1-p2) * 0.25f + (p3-p4) * 0.25f)*0.6f;
         quad.tangent_v = ((p1-p3) * 0.25f + (p2-p4) * 0.25f)*0.6f;
         quad.normal = quad.tangent_u.crossProduct( quad.tangent_v );
         quad.size = norm( quad.normal );
         //quad.valid = true;
         quad.valid = ((flag1 | flag2 | flag3 | flag4) & PF_FLAG_INVALID) ? false : true;
         if( norm( quad.tangent_v ) < 1e-20f )
            quad.valid = false;
         if( norm( quad.tangent_u ) < 1e-20f )
            quad.valid = false;

         if( fabs(normalize(quad.tangent_u) * normalize(quad.tangent_v) ) > cos(m_MinTangentAngle/180.0f*M_PI) )
            quad.valid = false;

         if( m_CheckDiag )
         {
            if( norm( quad.tangent_u + quad.tangent_v ) > m_MaxQuadDiag )
               quad.valid = false;
            if( norm( quad.tangent_u ) > m_MaxQuadDiag / 2.0f )
               quad.valid = false;
            if( norm( quad.tangent_v ) > m_MaxQuadDiag / 2.0f )
               quad.valid = false;
         }

         quads[y * (paramIt->getWidth()-1) + x] = quad;
      }

   std::vector<TQuad*> medianCache( (m_InvalidateKernelSize*2+1)*(m_InvalidateKernelSize*2+1) );

   // -----------------------------------------------
   // validate quads
   // inner quads ( no edge test )
   if( m_InvalidateKernelSize > 0 )
   {
      for( card32 x=0;x<paramIt->getWidth()-1;x++ )
         for( card32 y=0;y<paramIt->getHeight()-1;y++ )
         {
            TQuad &quad = getMyQuad(x,y);
            if( !quad.valid )
               continue;

            if( m_CheckAbsolute )
               if( quad.size > m_MaxAbsoluteQuadSize )
               {
                  quad.valid = false;
                  continue;
               }

            float32 l1 = norm( quad.tangent_u );

            Vector3f v2 = quad.tangent_v - quad.tangent_u * ((quad.tangent_u*quad.tangent_v) / normQuad(quad.tangent_u));
            float32 l2 = norm( v2 );
            if( l2 < 1e-20f )
               quad.valid = false;
            else
               if( (max(l1,l2) / min( l1,l2)) > m_MaxQuadRatio )
                  quad.valid = false;
         }

      for( card32 x=m_InvalidateKernelSize;x<paramIt->getWidth()-1-m_InvalidateKernelSize;x++ )
         for( card32 y=m_InvalidateKernelSize;y<paramIt->getHeight()-1-m_InvalidateKernelSize;y++ )
         {
            TQuad &quad = getMyQuad(x,y);
            if( !quad.valid )
               continue;

            int numUsed = 0;
            for( int xk = -(int)m_InvalidateKernelSize; xk<=(int)m_InvalidateKernelSize; xk++ )
               for( int yk = -(int)m_InvalidateKernelSize; yk<=(int)m_InvalidateKernelSize; yk++ )
               {
                  TQuad  * tmp = &getMyQuad(x+xk,y+yk);
                  if( tmp->valid )
                     medianCache[numUsed++] = tmp;
               }
                                                                                       
            sort( medianCache.begin(), medianCache.begin() + numUsed/2, compareQuads );

            if( medianCache[numUsed/2]->size / quad.size > m_MaxRelativeQuadSize )
            {
               quad.valid = false;
               continue;
            }

         }


   }
   // -----------------------------------------------
   // validate & write results

   for(  card32 x=0;x<paramIt->getWidth()-1;x++ )
      for( card32 y=0;y<paramIt->getHeight()-1;y++ )
      {
         paramIt->goTo(x,y);
         TQuad quad = quads[ y*(paramIt->getWidth()-1)+x];

         if( !quad.valid )
         {
            quad.normal = makeVector3f(0,0,0);
            quad.tangent_u = makeVector3f(0,0,0);
            quad.tangent_v = makeVector3f(0,0,0);
         }


         if( m_InvalidateOutliers )
         {
            //if( !quad.valid )
            //   modIt->set1i(token_flags, (bIt->get1i(token_flags) | PF_FLAG_INVALID));
         }


         if( m_EstimateNormals )
         {
            Vector3f normal = makeVector3f(0,0,0);

            if( m_NormalizeNormals )
               modIt->set3f( token_normal , normalize( quad.normal ) );
            else
               modIt->set3f( token_normal , quad.normal );
         }
         if( m_EstimateTangents )
         {
            if( m_NormalizeTangents )
            {
               modIt->set3f( token_tangentU, normalize( quad.tangent_u ) );
               modIt->set3f( token_tangentV, normalize( quad.tangent_v ) );
            }
            else
            {
               modIt->set3f( token_tangentU, quad.tangent_u * m_ScaleTangents );
               modIt->set3f( token_tangentV, quad.tangent_v * m_ScaleTangents );
            }
         }
      }

   for( card32 x=0;x<paramIt->getWidth();x++ )
   {
      paramIt->goTo(x, paramIt->getHeight()-1 );
      modIt->set1i(token_flags, (bIt->get1i(token_flags) | PF_FLAG_INVALID));
   }

   for( card32 y=0;y<paramIt->getHeight()-1;y++ )
   {
      paramIt->goTo(paramIt->getWidth()-1,y );
      modIt->set1i(token_flags, (bIt->get1i(token_flags) | PF_FLAG_INVALID));
   }



   delete it;
   pc->clearAttachments();
}
