//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PCCRemoveOutliers.h"
#include "StringClassProperty.h"
#include "ObjectClassProperty.h"
#include "SeparatorClassProperty.h"
#include "ProgressWindow.h"
#include <time.h>
#include "scene.h"
#include "GenericSwapAllOutOfCorePointCloud.h"
#include "qt/qimage.h"
#include "PointCloudTools.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

static std::string getTechniqueDescr( const card32 index )
{
	if ( index == OUTLIER_REMOVER_TENSOR_VOTING ) return "Tensor Voting";
	else return "Undefined";
}

static std::string getActionDescr( const card32 index )
{
	if ( index == OUTLIER_ACTION_COMPUTE ) return "Compute";
	else if ( index == OUTLIER_ACTION_COLORIZE ) return "Colorize";
	else if ( index == OUTLIER_ACTION_DELETE ) return "Delete";
	else return "Undefined";
}

IMPLEMENT_X4_CLASS( PCCRemoveOutliers, 6 )
{
	BEGIN_CLASS_INIT( PCCRemoveOutliers )
	ADD_STRING_PROP( pcName, 1 )
	ADD_CARD32_PROP_UPDATE_DESCR( technique, 0, NULL, getTechniqueDescr, 1, NULL )
	ADD_SEPARATOR( "Tensor Voting" )
	ADD_OBJECT_PROP( tv, 6, TensorVoting::getClass(), true );
	ADD_FLOAT32_PROP( tvSigma, 0 )
	ADD_FLOAT32_PROP( tvThreshold, 0 )
	ADD_CARD32_PROP_UPDATE_DESCR( action, 0, NULL, getActionDescr, 3, NULL )
	ADD_BOOLEAN_PROP( batch, 3 );
	ADD_BOOLEAN_PROP( ignoreOutliers, 4 );
	ADD_BOOLEAN_PROP( selectionMode, 5);
	ADD_SEPARATOR( "Visualization" )
	ADD_BOOLEAN_PROP( showTVSigma, 0 )
	ADD_FLOAT32_PROP( renderScale, 0 )
	ADD_CARD32_PROP( showPointIndex, 2 )
}

PCCRemoveOutliers::PCCRemoveOutliers() 
{
	pcName = "root/data";
	technique = OUTLIER_REMOVER_TENSOR_VOTING;
	action = OUTLIER_ACTION_COMPUTE;
	tv = new TensorVoting();
	renderScale = 1.0f;
	tvThreshold = 3.0f;
	showTVSigma = false;
	showPointIndex = 0;
	rt = new RenderingTool();
	rt->init();
	selectionMode = false;

	// Ephesus
	tvSigma = 0.06f;
	renderScale = 1.0f;
	showPointIndex = 0;

	/*
	// Buddha
	tvSigma = 1.0f;
	showPointIndex = 0;
	*/

	// Sphere
	//tvSigma = 0.05f;	
	//renderScale = 0.05f;

	/*
	// Happy Buddha
	tvSigma = 0.002f;	
	renderScale = 0.005f;
	*/

	/*
	// Floor
	tvSigma = 4.0f;	
	renderScale = 5.0f;
	*/

	/*
	// Bunny
	tvSigma = 0.05f;
	*/


	// DEPRECATED
	batch = false;
	ignoreOutliers = false;
	
}

PCCRemoveOutliers::~PCCRemoveOutliers() 
{
	delete rt;
	delete tv;
}

void PCCRemoveOutliers::execute( Scene *scene )
{
	if( !selectionMode )
	{
		processPointCloud(pcName, scene);
	}
	else
	{
		const std::vector<std::string> selection = scene->getRootState()->staticState->getSelectedNodes();
		debugOutput << "  PCCRemoveOutliers:\n";
		for( card32 i=0;i<selection.size();i++ )
		{
			PointCloud * pc = getPointCloud( getScene(), selection[i] );
			if( pc ){
				debugOutput <<"  pc "<<i<< " "; debugOutput.flush();
				processPointCloud(selection[i],scene);
			}
		}
		debugOutput << "  PCCRemoveOutliers done.\n";

	}

}

void PCCRemoveOutliers::glDrawTool( GLContext *glContext )
{
	if ( !showTVSigma ) return;
	float32 sqrSigma = sqr( tvSigma );
	PointCloud* pc = getPointCloud( scene, pcName );
	if ( pc == NULL ) return;
	AAT POSITION = pc->getAAT( "position" );
	SceneObjectIterator* it = pc->createIterator( SOIT::CAP_BASIC_PC );
	BasicPointCloudIterator* basicIt = dynamic_cast<BasicPointCloudIterator*>( it );
	HierarchicalKNNIterator* knnIt = new HierarchicalKNNIterator( pc, 32, NULL );
	basicIt->reset();
	long pointIndex = 0;
	rt->enableLighting();
	rt->setRenderColor( makeVector3f( 0.25f, 0.25f, 0.75f ) );
	while( !basicIt->atEnd() ){
		if ( pointIndex == showPointIndex ){
			const Vector3f& p = basicIt->get3f( POSITION );
			knnIt->setSeekPointAndReset( p );
			Vector3f np = knnIt->get3f( POSITION );
			float32 sqrD = ( np - p ) * ( np - p );
			while( sqrD < sqrSigma ){
				rt->renderSphere( np, 0.01f * renderScale );
				np = knnIt->get3f( POSITION );
				sqrD = ( np - p ) * ( np - p );
				knnIt->next();
			}
			break;
		}
		basicIt->next();
		pointIndex++;
	}
	delete it;
	delete knnIt;
}

void PCCRemoveOutliers::processPointCloud( std::string currentPCName, Scene * scene )
{
	if (action == OUTLIER_ACTION_COLORIZE) checkAttribute( scene, currentPCName, "color", 3, VAD::DATA_FORMAT_FLOAT32 );
	checkAttribute( scene, currentPCName, "surfaceness", 1, VAD::DATA_FORMAT_FLOAT32 );
	checkAttribute( scene, currentPCName, "curveness", 1, VAD::DATA_FORMAT_FLOAT32 );
	//checkAttribute( scene, currentPCName, "tensor", 9, VAD::DATA_FORMAT_FLOAT32 );
	PointCloud* pc = getPointCloud( scene, currentPCName );
	if ( pc == NULL ){
		debugOutput << "Invalid point cloud.\n"; debugOutput.flush();
		return;
	}


	AAT POSITION = pc->getAAT( "position" );
	AAT SURFACENESS = pc->getAAT( "surfaceness" );
	AAT CURVENESS = pc->getAAT( "curveness" );
	AAT COLOR;
	if (action == OUTLIER_ACTION_COLORIZE) COLOR = pc->getAAT( "color" );

	if ( action == OUTLIER_ACTION_COMPUTE ){
		GenericSwapAllOutOfCorePointCloud * gpc =  dynamic_cast<GenericSwapAllOutOfCorePointCloud*>(pc);
		UnstructuredInCorePointCloud * upc = dynamic_cast<UnstructuredInCorePointCloud*>(pc);
		if( upc || (gpc && dynamic_cast<UnstructuredInCorePointCloud*>(gpc->fetchPC())))
		{
			SceneObjectIterator* it = pc->createIterator( SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC | SOIT::CAP_DYNAMIC_PC );
			BasicPointCloudIterator* basicIt = dynamic_cast<BasicPointCloudIterator*>( it );
			ModifyablePointCloudIterator* modIt = dynamic_cast<ModifyablePointCloudIterator*>( it );
			PointSet * ps = modIt->lockBlockReadWrite();
			tv->tensorVote(true, ps,tvSigma);
			basicIt->unlock();
			delete it;
		}
		else
			tv->tensorVote( true, pc, tvSigma );
		return;
	} else if ( action == OUTLIER_ACTION_DELETE ){
		SceneObjectIterator* it = pc->createIterator( SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC | SOIT::CAP_DYNAMIC_PC );
		BasicPointCloudIterator* basicIt = dynamic_cast<BasicPointCloudIterator*>( it );
		ModifyablePointCloudIterator* modIt = dynamic_cast<ModifyablePointCloudIterator*>( it );
		if ( modIt == NULL ){
			debugOutput << "Cannot create modifyable iterator.\n"; debugOutput.flush();
			return;
		}
		DynamicPointCloudIterator* dynIt = dynamic_cast<DynamicPointCloudIterator*>( it );
		if ( dynIt == NULL  ){
			debugOutput << "Cannot create dynamic iterator; required for deletion of points.\n"; debugOutput.flush();
			return;
		}
		progressWindow->pushStep( true, "Tensor Voting" );
		basicIt->reset();
		long pointIndex = 0;
		long numDeleted = 0;
		long numOutliers = 0;
		while( !basicIt->atEnd() ){
			const Vector3f& p = basicIt->get3f( POSITION );
			const float32& surfaceness = basicIt->get1f( SURFACENESS );
			bool del = false;
			bool isOutlier = basicIt->get1f( SURFACENESS ) < tvThreshold;
			if ( isOutlier ) {
				//if ( rand()%10 == 0 ){
				dynIt->deleteCurrent();
				del = true;
				numDeleted++;
				pointIndex--;
			}
			if ( !del ) basicIt->next();
			pointIndex++;
			progressWindow->progress( card32( float32( pointIndex ) / float32( pc->getNumPoints() ) * 100 ) );
		}
		pc->clearAttachments();
		debugOutput << "Outlier remover: deleted " << numDeleted << " points.\n"; debugOutput.flush();
		progressWindow->popSteps( );
		delete it;
	} else if ( action == OUTLIER_ACTION_COLORIZE ){
		SceneObjectIterator* it = pc->createIterator( SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC | SOIT::CAP_DYNAMIC_PC );
		BasicPointCloudIterator* basicIt = dynamic_cast<BasicPointCloudIterator*>( it );
		ModifyablePointCloudIterator* modIt = dynamic_cast<ModifyablePointCloudIterator*>( it );
		if ( modIt == NULL ){
			debugOutput << "Cannot create modifyable iterator.\n"; debugOutput.flush();
			return;
		}
		progressWindow->pushStep( true, "Tensor Voting" );
		basicIt->reset();
		long pointIndex = 0;
		long numDeleted = 0;
		long numOutliers = 0;

		QImage * outImg = new QImage(160,120,QImage::Format_RGB32);

		if (tvThreshold>0.0f) // standard behaviour
		{
			while( !basicIt->atEnd() ){
				const Vector3f& p = basicIt->get3f( POSITION );
				const float32& surfaceness = basicIt->get1f( SURFACENESS );
				bool del = false;
				bool isOutlier = basicIt->get1f( SURFACENESS ) < tvThreshold;
				if ( isOutlier ) {
					modIt->set3f( COLOR, makeVector3f( 1.0f, 0.0f, 0.0f ) );
				} else {
					modIt->set3f( COLOR, makeVector3f( 0.0f, 1.0f, 0.0f ) );
				}
				basicIt->next();
				pointIndex++;
				if( pointIndex % 1000 == 0)
					progressWindow->progress( card32( float32( pointIndex ) / float32( pc->getNumPoints() ) * 100 ) );
			}
		}
		else // continuous colors
		{
			/* compute surfaceness maximum
			float32 maxSurfaceness = -1.0f;
			while( !basicIt->atEnd() ){
				const float32& surfaceness = basicIt->get1f( SURFACENESS );
				if (surfaceness>maxSurfaceness) maxSurfaceness = surfaceness;
				basicIt->next();
			}*/

			basicIt->reset();
			while( !basicIt->atEnd() ){
				float32 prob = basicIt->get1f( CURVENESS )/(-1.0f*tvThreshold);
				modIt->set3f( COLOR, makeVector3f( 1.0f-prob, prob, 0.0f ) );
				basicIt->next();
				card32 j= pointIndex % 160;
				card32 i= pointIndex/160;
				QColor c = (prob>0.5) ? QColor(255,0,0) : QColor(0,0,255);
				outImg->setPixel(j,i,c.rgb());
				pointIndex++;
				if( pointIndex % 1000 == 0)
					progressWindow->progress( card32( float32( pointIndex ) / float32( pc->getNumPoints() ) * 100 ) );

			}
			outImg->save("c:/daten/outlierImg.bmp");

		}
		progressWindow->popSteps( );
		delete it;
	}
	
}


void PCCRemoveOutliers::execute_ar( Scene *scene, const std::vector<std::string> selection )
{
	for( card32 i=0;i<selection.size();i++ )
	{
		PointCloud * pc = getPointCloud( getScene(), selection[i] );
		processPointCloud(selection[i],scene);
	}
}