//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PCCRochusKapelle.h"
//---------------------------------------------------------------------------
#include "ImportTools.h"
#include "PointCloudTools.h"
#include "VertexDescriptor.h"
#include "ProgressWindow.h"
#include "StringClassProperty.h"
#include "BooleanClassProperty.h"
#include "BucketSplitPointsAddingAdaptor.h"
#include "NumericalClassProperty.h"
#include "ClassMethods.h"
#include "NormalEstimator.h"
#include "PCCNormalEstimatorPCA.h"
//---------------------------------------------------------------------------
#include <string>
#include <fstream>
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------



IMPLEMENT_X4_CLASS(PCCRochusKapelle , 1)
{
	BEGIN_CLASS_INIT(PCCRochusKapelle);
	ADD_STRING_PROP(targetNode, 0)
		ADD_STRING_PROP(dataSetDir, 0)
		ADD_BOOLEAN_PROP(normalizeIntensity, 0)
		ADD_BOOLEAN_PROP(storeCameraPos, 0)
		ADD_STRING_PROP(bucketTempFileName, 1)
		ADD_FLOAT32_PROP(bucketsSideLength, 1)
		ADD_BOOLEAN_PROP(useBuckets, 1)
		ADD_NOARGS_METHOD(importToSeparatePCs);
}


// --



PCCRochusKapelle::PCCRochusKapelle() {
	targetNode = "root/pc1";
	dataSetDir = "I:\\ZF\\full_res\\";
	//dataSetDir = "I:\\ZF\\tenth_res\\";

	normalizeIntensity = false;
	storeCameraPos = true;
	bucketTempFileName = "C:\\temp\\buckets";
	bucketsSideLength = 2.0f;
	useBuckets = false;
}


void PCCRochusKapelle::execute(Scene *scene) {
	Matrix4f reg[12];
	std::string fname[12];
	setNamesAndInitialTransformations(reg, fname);

	PointCloud *pc = getPointCloud(scene, targetNode);
	if (!pc) throw ECommandFailed("PCCRochusKapelle::execute() - targetNode not found");

	const VertexDescriptor *vd = pc->getDescr();
	if (!(vd->providesAttribute("position", 3, VAD::DATA_FORMAT_FLOAT32) 
		&& vd->providesAttribute("intensity", 1, VAD::DATA_FORMAT_FLOAT32))) {
			throw ECommandFailed("PCCRochusKapelle::execute() - need channels position3f and intensity1f");   
	}
	if (storeCameraPos) {
		if (!vd->providesAttribute("camPos", 3, VAD::DATA_FORMAT_FLOAT32)) {
			throw ECommandFailed("PCCRochusKapelle::execute() - current settings demand an additional channel camPos3f");
		}
	}

	SceneObjectIterator *it = pc->createIterator(SOIT::CAP_DYNAMIC_PC);
	DynamicPointCloudIterator *dynIt = dynamic_cast<DynamicPointCloudIterator*>(it);
	if (dynIt == NULL) {
		delete it;
		throw ECommandFailed("PCCRochusKapelle::execute() - targetNode is not dynamic");
	}


	card8 *buffer = new card8[vd->getSize()];

	vd->zero((char*)buffer);
	VAD posVad = vd->getAttribute("position");
	AAT posAAT = vd->getToken("position");
	VAD intensityVad = vd->getAttribute("intensity");
	AAT intensityAAT = vd->getToken("intensity");
	VAD camPosVad;
	AAT camPosAAT;
	if (storeCameraPos) {
		camPosVad = vd->getAttribute("camPos");
		camPosAAT = vd->getToken("camPos");
	}

	BucketSplitPointsAddingAdaptor *bucketIt 
		= new BucketSplitPointsAddingAdaptor(dynIt, bucketsSideLength, bucketTempFileName, posAAT, vd->getSize());

	progressWindow->pushStep(true, "import");
	for (card32 i=0; i<12; i++) {
		std::string filename = dataSetDir+fname[i];
		std::ifstream in(filename.c_str());
		if (!in.good()) throw ECommandFailed("PCCRochusKapelle::execute() - file not found");
		skipToEOL(in);
		while (!in.eof()) {
			Vector3f pos;
			in >> pos[0];
			in >> pos[1];
			in >> pos[2];
			StaticVector<float32, 1> intensity;
			in >> intensity[0]; 
			intensity[0] /= 16384.0f;
			if (normalizeIntensity) {
				intensity[0] /= max(1.0E-5f, norm(pos));
			}
			int32 forgetthis;
			in >> forgetthis;
			in >> forgetthis;

			pos = transformVector3f(reg[i], pos);

			posVad.setAttrib<float32, 3>(buffer, pos, posAAT);
			intensityVad.setAttrib<float32, 1>(buffer, intensity, intensityAAT);
			if (storeCameraPos) {
				Vector3f camPos = transformVector3f(reg[i], NULL_VECTOR3F);
				camPosVad.setAttrib<float32, 3>(buffer, camPos, camPosAAT);				
			}
			if (useBuckets) {
				bucketIt->add((const char*)buffer);
			} else {
				dynIt->add((const char*)buffer);
			}
			progressWindow->progress((card32)(float(i)/12.0f*100.0f));
		}		
	}	
	progressWindow->popSteps();

	delete bucketIt;
	delete it;
}

void X4::PCCRochusKapelle::importToSeparatePCs()
{
	Matrix4f reg[12];
	std::string fname[12];
	setNamesAndInitialTransformations(reg, fname);

	VertexDescriptor vd;
	vd.pushAttrib( mVAD( "position", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	vd.pushAttrib( mVAD( "intensity", 1, VAD::DATA_FORMAT_FLOAT32 ) );
	vd.pushAttrib( mVAD( "normal", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	vd.pushAttrib( mVAD( "flags", 1, VAD::DATA_FORMAT_INT32 ) );
	//vd.pushAttrib( mVAD( "color", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	card8 *buffer = new card8[vd.getSize()];
	vd.zero((char*)buffer);
	Vector4f* pos = (Vector4f*) buffer;




	progressWindow->pushStep(true, "import");
	for (card32 i=0; i<12; i++) {
		std::string filename = dataSetDir+fname[i];
		std::ifstream in(filename.c_str());
		if (!in.good()) throw ECommandFailed("PCCRochusKapelle::execute() - file not found: " + filename);

		UnstructuredInCorePointCloud* pc = new UnstructuredInCorePointCloud;

		pc->clearAndSetup(&vd, 0);
		PointSet* ps = pc->getPointSet();
		AAT POS = ps->getAAT("position");
		AAT INT = ps->getAAT("intensity");

		skipToEOL(in);
		mpcard pointIndex = 0;
		while (!in.eof()/* && pointIndex < 100000*/) {
			
			in >> (*pos)[0];
			in >> (*pos)[1];
			in >> (*pos)[2];

			float32 intensity = 0;
			in >> intensity; 
			//debugOutput << intensity << "\n";

			intensity /= 16384.0f;
// 			if (normalizeIntensity) {
// 				intensity /= max(1.0E-5f, norm(*pos));
// 			}
			(*pos)[3] = intensity;
			//debugOutput << (*pos)[3] << "\n";

			int32 forgetthis;
			in >> forgetthis;
			in >> forgetthis;

			ps->addPoint((char*)buffer);
			pointIndex++;
			//ps->set1f(pointIndex++, INT, intensity );
			//pos = transformVector3f(reg[i], pos);

		}		
		progressWindow->progress((card32)(float(i)/12.0f*100.0f));
		//NormalEstimator::estimateNormals(ps, 20);
		//PCCNormalEstimatorPCA::unifyNormalsByCameraPosition(pc);

		pc->setTransformation(reg[i]);
		addPointCloud(scene, pc, "part_" + intToStr(i));
		progressWindow->progress((card32)(float(i)/12.0f*100.0f + 100.0f/24.0f));
		ObjectViewsTable::update(scene);

	}	
	progressWindow->popSteps();

	delete buffer;


}





void PCCRochusKapelle::setNamesAndInitialTransformations( Matrix4f * reg, std::string * fname ) 
{

	reg[0] = IDENTITY4F;
	reg[1] = makeMatrix4f(
		-0.084685601f,	-0.9950884f,	-0.05125149f,	0.0f,
		0.99384385f,	-0.088043094f,	0.067245722f,	0.0f,
		-0.071427733f,	-0.045241214f,	0.99641937f,	0.0f,
		3.2169688f,		-1.5138476f,	0.10139322f,	1.0f
		).transpose();

	reg[2] = makeMatrix4f(
		-0.99667066f,		-0.0053207139f,	0.081369542f,	0.0f,
		-0.0014077654f,	-0.99659902f,		-0.082406245f,	0.0f,
		0.081531085f,		-0.082246177f,		0.99327195f,	0.0f,
		6.5255594f,			-3.2952032f,		0.2273435f,		1.0f
		).transpose();

	reg[3] = makeMatrix4f(
		-0.98343372f,	0.057755727f,	-0.17182194f,	0.0f,
		-0.074275933f, -0.99304771f,	0.091322586f,	0.0f,
		-0.16535257f,	0.1025713f,		0.98088688f,	0.0f,
		6.3888216f,		-3.1991196f,	0.21578683f,	1.0f
		).transpose();

	reg[4] = makeMatrix4f(
		-0.98496139f,	-0.063827805f, -0.16054492f,	0.0f,
		0.050512265f,	-0.99503869f,	0.085698329f,	0.0f,
		-0.16521829f,	0.076300174f,	0.98330134f,	0.0f,
		4.5179977f,		-2.029129f,		0.1449634f,		1.0f
		).transpose();

	reg[5] = makeMatrix4f(
		-0.99748409f,	0.063801445f,	-0.030888975f, 0.0f,
		-0.064677328f,	-0.99750656f,	0.028237417f,	0.0f,
		-0.029010361f,	0.030164205f,	0.99912393f,	0.0f,
		-3.1145897f,	-2.7324498f,	-0.099038243f, 1.0f
		).transpose();

	reg[6] = makeMatrix4f(
		0.38166234f,	-0.91471595f,	0.13277116f,	0.0f,
		0.92348683f,	0.38340163f,	-0.013228061f, 0.0f,
		-0.038804762f, 0.12766103f,	0.99105853f,	0.0f,
		-0.53601211f,	0.078203753f,	-0.057978343f, 1.0f
		).transpose();

	reg[7] = makeMatrix4f(
		-0.97514713f,	-0.16160582f,	0.15155947f,	0.0f,
		0.14471899f,	-0.98258096f,	-0.11657734f,	0.0f,
		0.16775902f,	-0.091746621f,	0.98154986f,	0.0f,
		13.094265f,		-4.3865814f,	0.56535244f,	1.0f
		).transpose();

	reg[8] = makeMatrix4f(
		-0.99382389f,	0.018279484f,	0.10945091f,	0.0f,
		-0.025332654f, -0.99766701f,	-0.063402638f, 0.0f,
		0.10803647f,	-0.065783828f, 0.99196929f,	0.0f,
		11.978042f,		-7.4336138f,	0.79717082f,	1.0f
		).transpose();

	reg[9] = makeMatrix4f(
		0.17098458f,	0.98475993f, -0.031812817f, 0.0f,
		-0.98382127f,	0.16889147f, -0.059749201f, 0.0f,
		-0.053465717f, 0.04151433f, 0.99770588f,	 0.0f,
		7.9571834f,		-3.6674364f, 0.5792675f,	 1.0f
		).transpose();

	reg[10] = makeMatrix4f(
		0.84243715f,	-0.53853148f,		0.016823107f,	0.0f,
		0.53824425f,	0.84257722f,		0.018867796f,	0.0f,
		-0.024335621f, -0.0068400423f,	0.99968106f,	0.0f,
		-1.9287333f,	2.1387477f,			2.8555129f,		1.0f
		).transpose();

	reg[11] = makeMatrix4f(
		0.79058784f,	0.60965407f, -0.057393607f, 0.0f,
		-0.61015308f,	0.79221523f, 0.01041652f,	 0.0f,
		0.051818509f,	0.02678366f, 0.99829793f,	 0.0f,
		-3.2408426f,	-1.8796377f, 2.7659883f,	 1.0f
		).transpose();


	fname[0] = "r01.asc";
	fname[1] = "r02.asc";
	fname[2] = "r03.asc";
	fname[3] = "r04.asc";
	fname[4] = "r05.asc";
	fname[5] = "r06.asc";
	fname[6] = "r07.asc";
	fname[7] = "r08.asc";
	fname[8] = "r09.asc";
	fname[9] = "r10.asc";
	fname[10] = "r11.asc";
	fname[11] = "r12.asc";

}
