//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PCCSmoothing.h"
#include "LinearAlgebra.h"
#include "LinearAlgebra.inline.h"
#include "StandardSceneGraphIterator.h"
#include "SGObjectNode.h"
#include "ClassMethods.h"
#include "StringClassProperty.h"
#include "ObjectClassProperty.h"
#include "PointSetKNNQuery.h"
#include "Monomials2ndOrder.h"
#include "TopologicalKNNIterator.h"
#include "TopologyAttachment.h"
#include "InCorePCTopologyGraph.h"
#include "PointCloudTools.h"

#include <time.h>
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------
IMPLEMENT_X4_CLASS( PCCSmoothing, 1 )
{
	BEGIN_CLASS_INIT( PCCSmoothing )
	ADD_STRING_PROP( pcName, 0 )
	ADD_CARD32_PROP( numNeighbors, 0 )
	ADD_BOOLEAN_PROP( useTopology, 1);
	ADD_BOOLEAN_PROP(useLaplacian, 1);
	ADD_CARD32_PROP(numIterations, 1);
}

/***
 * Constructor
 */
PCCSmoothing::PCCSmoothing() 
{
	pcName = "root/pc";
	numNeighbors = 20;
	useTopology = false;
	useLaplacian = false;
	numIterations = 1;
}

/***
* Destructor
*/
PCCSmoothing::~PCCSmoothing() 
{
}

/**
* Run systems
**/
void PCCSmoothing::execute_at( Scene *scene, PointCloud* pc )
{
	UnstructuredInCorePointCloud* upc = dynamic_cast<UnstructuredInCorePointCloud*>( pc );
	if ( upc == NULL ){
		debugOutput << "execute: invalid point point cloud.\n"; debugOutput.flush();
		return;
	}

	PointSet* ps = upc->getPointSet();
	AAT POSITION = ps->getAAT( "position" );
	//AttachedIndexedOctree* octree = PointSetKNNQuery::createOctree( ps, 32 );
	//PointSetKNNQuery knn( ps, octree, NULL_VECTOR3F, ps->getAAT("position") );
	HierarchicalKNNIterator* knn = NULL;
	TopologicalKNNIterator* tnn = NULL;
	InCorePCTopologyGraph* tpg = NULL;

	if (useTopology)
	{
		TopologyAttachment* tatt = dynamic_cast<TopologyAttachment*>(pc->getAttachments()->getData(TopologyAttachment::getDefaultName()));
		PointCloudIterator* tIt = NULL;
		if (tatt) tpg = tatt->getTopology();
		if (tpg) tIt = tpg->createIterator(ps);
		if (tIt) tnn = new TopologicalKNNIterator(pc, tIt);
	}else
		knn = new HierarchicalKNNIterator(pc, 32, NULL);

	card32 nIndex;
	bool found;
	Matrix3f T, invT;
	Vector3f eigenValues, centroid;
	PCA<float32,3> pca;
	std::vector<card32> neighbors;
	std::vector<Vector3f> points;
	StaticVector<float32,6> coeff;

	progressWindow->pushStep( true, "Smoothing" );
	progressWindow->progress( 0 );

	float32 start = clock();

	for (unsigned iter = 0; iter < numIterations; iter++)
	{
		std::vector<Vector3f> positions;
		if (useLaplacian)
		{
			positions.resize(ps->getNumEntries());
			for (unsigned i=0; i < ps->getNumEntries(); i++)
				positions[i] = ps->get3f(i, POSITION);
		}

		for ( unsigned int pointIndex = 0; pointIndex < ps->getNumEntries(); pointIndex++ )
		{
			Vector3f pos = ps->get3f( pointIndex, POSITION );
			//knn.setTargetPoint( pos );
			if (knn) knn->setSeekPointAndReset(pos);
			if (tnn) tnn->setStartPointAndReset(pointIndex);

			unsigned int nNeighbours = numNeighbors;
			if (tpg) nNeighbours = tpg->getNumAdjacentVertices(pointIndex);

			neighbors.clear();
			pca.clear();
			for ( unsigned int n = 0; n < nNeighbours; n++ )
			{
				found = false;

				//knn.getNextNearestPointIndex( nIndex, found );
				if (knn)
				{
					knn->next();
					if (knn->atEnd() == false)
					{
						found = true;
						nIndex = knn->getCurrentPointIndex();
					}
				}
				if (tnn)
				{
					if (useLaplacian && tpg)
					{
						nIndex = tpg->getAdjacentVertex(pointIndex, n);
						found = true;
					}else
					{
						tnn->next();
						if (tnn->atEnd() == false)
						{
							found = true;
							nIndex = tnn->getPointSetPointNumber();
						}
					}
				}

				if (found == false) break;

				neighbors.push_back( nIndex );
				pca.addPoint( ps->get3f( nIndex, POSITION ) );
			}

			if (useLaplacian)
			{
				Vector3f avg = makeVector3f(0,0,0);
				for (unsigned n=0; n < neighbors.size(); n++)
					avg += positions[neighbors[n]];
				avg /= float(neighbors.size());
				ps->set3f(pointIndex, POSITION, avg);

			}else
			{
				// Compute regression plane
				pca.analyze( eigenValues, T, centroid );
				invT = invertMatrix( T );
				/// Compute regression polynomial
				points.clear();
				for ( unsigned int n = 0; n < neighbors.size(); n++ ){
					Vector3f nPos = ps->get3f( neighbors[n], POSITION );
					nPos = invT * ( nPos - pos );
					points.push_back( nPos );
				}
				Monomials2ndOrder::computeCoeffs( points, coeff );
				float32 h = Monomials2ndOrder::evaluate( coeff, makeVector2f( 0,0 ) ); 
				Vector3f newPos = T * makeVector3f( 0,0,h ) + pos;
				ps->set3f( pointIndex, POSITION, newPos );
			}

			progressWindow->progress( 100.0f * (pointIndex + ps->getNumEntries() * iter) / float32( ps->getNumEntries() * numIterations));
		}
	}

	float32 time = ( clock() - start ) / CLK_TCK;
	debugOutput << "Time required: " << time << " seconds.\n"; debugOutput.flush();

	if (tnn) delete tnn;
	if (knn) delete knn;

	//delete octree;
	//bool dummy;
	progressWindow->popSteps();// dummy );
}

/** 
* Run system
*/
void PCCSmoothing::execute( Scene *scene )
{
	PointCloud* pc = getPointCloud( scene, pcName );
	execute_at(scene, pc);
}

/** 
 * Connect to scene
 */
void PCCSmoothing::connectScene( Scene *scene, OnSceneChangeCallback *onChange, QWidget *sceneEditorWidget )
{
	PCCommand::connectScene( scene, onChange, sceneEditorWidget );
}