#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PCI2DPointPainter.h"
//---------------------------------------------------------------------------
#include "SceneGraphState.h"
#include "PointCloudTools.h"
#include "UnstructuredInCorePointCloud.h"
#include "VertexDescriptor.h"
#include "PointCloudIterators.h"
#include "MathTools.h"
#include "SelectionTools.h"
//---------------------------------------------------------------------------
using namespace X4;

IMPLEMENT_X4_CLASS(PointPainterCmdObj, 1)
{
	BEGIN_CLASS_INIT(PointPainterCmdObj);
	ADD_STRING_PROP(targetNode, 0);
	ADD_BOOLEAN_PROP(createIfNotFound, 0);
	ADD_BOOLEAN_PROP(storeSampleSpacing, 1)
	ADD_BOOLEAN_PROP(storeNoiseCovariance, 1)
	ADD_VAR_ARRAY_PROP( points, 0, 
		new FixedArrayClassProperty( "element" , MetaClass::FIRST_VERSION, NULL, 
		                             new Float32ClassProperty("", 0, (pointer_diff)NULL, NULL), 3, sizeof(float32), 3 )
	)
	ADD_VAR_ARRAY_PROP( sampleSpacings, 1, 
		new Float32ClassProperty( "element" , MetaClass::FIRST_VERSION + 1, (pointer_diff)NULL, NULL )
	)
	ADD_VAR_ARRAY_PROP( noiseCovariances, 1, 
		new FixedArrayClassProperty( "element" , MetaClass::FIRST_VERSION, NULL, 
		                             new Float32ClassProperty("", 0, (pointer_diff)NULL, NULL), 9, sizeof(float32), 3 )
	)
}


IMPLEMENT_X4_CLASS(PCI2DPointPainter ,0)
{
	BEGIN_CLASS_INIT(PCI2DPointPainter);
	ADD_STRING_PROP(targetNode, 0)
	ADD_BOOLEAN_PROP(createIfNotFound, 0)
	ADD_BOOLEAN_PROP(storeSampleSpacing, 0)
	ADD_BOOLEAN_PROP(storeNoiseCovariance, 0)
	ADD_FLOAT32_PROP(zDist, 0)
	ADD_FLOAT32_PROP(randomSigma, 0)
}


// ----

PointPainterCmdObj::PointPainterCmdObj() {
	createIfNotFound = true;
	storeSampleSpacing = true;
	storeNoiseCovariance = true;
}


void PointPainterCmdObj::setup(std::string targetNode, bool createIfNotFound, bool storeSampleSpacing, bool storeNoiseCovariance, 
                               Vector3f point, float32 sampleSpacing, Matrix3f noiseCovariance) {
	this->targetNode = targetNode;
	this->createIfNotFound = createIfNotFound;
	this->storeSampleSpacing = storeSampleSpacing;
	this->storeNoiseCovariance = storeNoiseCovariance;
	points.push_back(point);
	sampleSpacings.push_back(sampleSpacing);
	noiseCovariances.push_back(noiseCovariance);
}

void PointPainterCmdObj::assign(const Object* obj, X4CopyContext *context) {
	const PointPainterCmdObj *other = dynamic_cast<const PointPainterCmdObj*>(obj);
	if (other == NULL) throw ETypeMismatch("PointPainterCmdObj::assign() - wrong type.");
	this->targetNode = other->targetNode;
	this->createIfNotFound = other->createIfNotFound;
	this->storeSampleSpacing = other->storeSampleSpacing;
	this->storeNoiseCovariance = other->storeNoiseCovariance;
	this->points = other->points;
	this->sampleSpacings = other->sampleSpacings;
	this->noiseCovariances = other->noiseCovariances;
}

void PointPainterCmdObj::doAction(Object *context) {
	Scene *scene = dynamic_cast<Scene*>(context);
	if (scene == NULL) {
		throw ECommandFailed("PointPainterCmdObj::doAction() - context must be a scene object.");
	}
	PointCloud *pc = getPointCloud(scene, targetNode);
	if (pc == NULL) {
		if (createIfNotFound) {
			UnstructuredInCorePointCloud *upc = new UnstructuredInCorePointCloud();
			pc = upc;
			VertexDescriptor vd;
			vd.pushAttrib(VAD("position", 3, VAD::DATA_FORMAT_FLOAT32));
			if (storeSampleSpacing) {
				vd.pushAttrib(VAD("sampleSpacing", 1, VAD::DATA_FORMAT_FLOAT32));			
			}
			if (storeNoiseCovariance) {
				vd.pushAttrib(VAD("noiseCovariance", 6, VAD::DATA_FORMAT_FLOAT32));			
			}
			vd.pushAttrib(VAD("flags", 1, VAD::DATA_FORMAT_INT32));
			upc->clearAndSetup(&vd, 0);
			addPointCloud(scene, pc, targetNode);
		} else {
			throw ECommandFailed("PointPainterCmdObj::doAction() - target not found.");
		}
	}
	SceneObjectIterator *it = pc->createIterator(SOIT::CAP_DYNAMIC_PC);
	DynamicPointCloudIterator *dynIt = dynamic_cast<DynamicPointCloudIterator*>(it);
	if (dynIt == NULL) {
		delete it;
		throw ECommandFailed("PointPainterCmdObj::doAction() - target is not dynamic.");		
	}
	const VertexDescriptor *vd = pc->getDescr();
	if (!vd->providesAttribute("position", 3, VAD::DATA_FORMAT_FLOAT32)) {
		throw ECommandFailed("PointPainterCmdObj::doAction() - no position3f channel in target.");
	}
	if (storeSampleSpacing && !vd->providesAttribute("sampleSpacing", 1, VAD::DATA_FORMAT_FLOAT32)) {
		throw ECommandFailed("PointPainterCmdObj::doAction() - no sampleSpacing1f channel in target.");
	}
	if (storeNoiseCovariance && !vd->providesAttribute("noiseCovariance", 6, VAD::DATA_FORMAT_FLOAT32)) {
		throw ECommandFailed("PointPainterCmdObj::doAction() - no noiseCovariance6f channel in target.");
	}
	AAT posAAT = vd->getToken("position");
	VAD posVAD = vd->getAttribute("position");
	card8 *buf = new card8[vd->getSize()];
	
	AAT sampleSpacingAAT;
	VAD sampleSpacingVAD;
	if (storeSampleSpacing) {
		sampleSpacingAAT = vd->getToken("sampleSpacing");
		sampleSpacingVAD = vd->getAttribute("sampleSpacing");		
	}
	AAT noiseCovarianceAAT;
	VAD noiseCovarianceVAD;
	if (storeNoiseCovariance) {
		noiseCovarianceAAT = vd->getToken("noiseCovariance");
		noiseCovarianceVAD = vd->getAttribute("noiseCovariance");		
	}
	
	vd->zero((char*)buf);
	
	for (card32 i=0; i<points.size(); i++) {
		posVAD.setAttrib<float, 3>(buf, points[i], posAAT);
		if (storeSampleSpacing) {
			StaticVector<float32, 1> v;
			v[0] = sampleSpacings[i];
			sampleSpacingVAD.setAttrib<float, 1>(buf, v, sampleSpacingAAT);
		}
		if (storeNoiseCovariance) {
			StaticVector<float32, 6> cv;
			cv[0] = noiseCovariances[i][0][0]; cv[1] = noiseCovariances[i][1][0]; cv[2] = noiseCovariances[i][2][0];
													 	  cv[3] = noiseCovariances[i][1][1]; cv[4] = noiseCovariances[i][2][1];
																										 cv[5] = noiseCovariances[i][2][2];
			noiseCovarianceVAD.setAttrib<float, 6>(buf, cv, noiseCovarianceAAT);
		}
		dynIt->add((char*)buf);
	}
	
	delete[] buf;
	delete it;
}

void PointPainterCmdObj::undoAction(Object *context) 
{
	throw ECannotUndo("PointPainterCmdObj::undoAction() - cannot undo.");
}

bool PointPainterCmdObj::canUndo() 
{
	return false;
}


std::string PointPainterCmdObj::getDescription() {
	return "painting 2d point curve";
}

bool PointPainterCmdObj::mergeCommand(const CommandObject *nextCmd) {
	const PointPainterCmdObj *other = dynamic_cast<const PointPainterCmdObj*>(nextCmd);
	if (other != NULL) {
		for (card32 i=0; i<other->points.size(); i++) {
			points.push_back(other->points[i]);
		}
		return true;
	} else {
		return false;
	}
}


// --


PCI2DPointPainter::PCI2DPointPainter()
{
	targetNode = "root/pc1";
	createIfNotFound = true;
	storeSampleSpacing = true;
	zDist = 10.0f;
	randomSigma = 0.0f;
	lastPointSet = false;
	leftDown = false;
   firstPoint = false;
	// try to guess this... ;-)
	lastPoint = makeVector3f(0.12345678f, 98765431.0f, 0.00000001234678234f);
}

PCI2DPointPainter::~PCI2DPointPainter() {}



void PCI2DPointPainter::keyDown(GeneralKey key) {forget(key);}

void PCI2DPointPainter::keyUp(GeneralKey key) {forget(key);}

void PCI2DPointPainter::mouseDown(int32 x, int32 y, MouseButtons buttonsState, Modifiers modifiersState) {
	if (buttonsState.getLeft()) {
		leftDown = true;
		lastPointSet = false;
      firstPoint = true;
		createPoint(x,y);
      firstPoint = false;
	}						  	
}

void PCI2DPointPainter::mouseMoved(int32 x, int32 y) 
{
	if (leftDown) {
		createPoint(x,y);
	}
}

void PCI2DPointPainter::mouseUp(int32 x, int32 y, MouseButtons buttonsState, Modifiers modifiersState) 
{
	if (buttonsState.getLeft()) {
		leftDown = false;
		lastPointSet = false;
	}
}

void PCI2DPointPainter::areaResize(card32 width, card32 height) 
{
	this->width = width;
	this->height = height;
}

void PCI2DPointPainter::createPoint(int32 x, int32 y) {
   Camera cam = scene->getRootState()->staticState->camera;
   RayConeRange* ray = getRayConeRangeFromMouseClick(x, y, width, height, 0, 
		&cam, &scene->getRootState()->staticState->viewFrustum);		
	Vector3f origin = ray->getRayOrigin();
	Vector3f dir = ray->getRayDir();		
	delete ray;
	
	Vector3f viewDir = cam.getOrthoNormViewDirection();
	try {
		/// transform with root ctm
		viewDir = transformVector3f(invertMatrix(scene->getRootState()->dynamicState->ctm).transpose(), viewDir);
	} catch (...) {}

	Matrix3f viewCoords;
	viewCoords[0] = cam.getOrthoNormViewDirection();
	viewCoords[1] = cam.getOrthoNormRightDirection();
	viewCoords[2] = cam.getOrthoNormUpDirection();
	
	Vector3f dirInCamCoords = viewCoords.transpose() * dir;
	if (dirInCamCoords[0] != 0) {
		dirInCamCoords[1] *= zDist / dirInCamCoords[0];
		dirInCamCoords[2] *= zDist / dirInCamCoords[0];
		dirInCamCoords[0] = zDist;
		if (randomSigma != 0) {
			dirInCamCoords[1] += gaussianRandom() * randomSigma;
			dirInCamCoords[2] += gaussianRandom() * randomSigma;
		}
	}
	Vector3f dirTransf = viewCoords * dirInCamCoords;

	Vector3f pos = scene->getRootState()->staticState->camera.getPosition() + dirTransf;

	if ((lastPointSet && pos != lastPoint) || firstPoint) {
		PointPainterCmdObj *cmdObj = new PointPainterCmdObj();
		Matrix3f noiseCov = makeMatrix3f(
			randomSigma, 0, 0,
			0, randomSigma, 0,
			0, 0, randomSigma
		);
		cmdObj->setup(targetNode, createIfNotFound, storeSampleSpacing, storeNoiseCovariance, pos, norm(pos-lastPoint), noiseCov);
		scene->getCommandHistory()->executeCommand(scene, cmdObj);	
	}
	lastPoint = pos;	
	lastPointSet = true;
}
