//---------------------------------------------------------------------------
#include "StdAfx.h"

//---------------------------------------------------------------------------
#include "PCIImageEditing.h"
//---------------------------------------------------------------------------
#include "PCImageTools.h"
#include "ObjectClassProperty.h"
#include "ParametricInCorePointCloud.h"
#include "PointSet.h"
#include "VertexDescriptor.h"
#include "PointCloudTools.h"
#include "LinearAlgebra.inline.h"
#include "DynamicLinearAlgebra.inline.h"
#ifndef _DEBUG
#include "PointSet.inline.h"
#endif
//---------------------------------------------------------------------------
 
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------



MetaClass* PCIImageEditing::PCIImageEditingMetaClass = NULL;

const MetaClass* PCIImageEditing::getInstanceClass() const {
   return getClass();
}

Object* PCIImageEditing::newInstance() {
   return new PCIImageEditing;
}

const MetaClass* PCIImageEditing::getClass() {
   if (PCIImageEditingMetaClass==NULL) {
      PCIImageEditingMetaClass = new MetaClass("PCIImageEditing", "no description.",
                                               newInstance, sizeof(PCIImageEditing));
   }
   return PCIImageEditingMetaClass;
}

void PCIImageEditing::init(const MetaClass *parentClass) {
   ((MetaClass*)getClass())->init(parentClass);
   PCIImageEditingMetaClass->copyParentProperties();
	BEGIN_CLASS_INIT(PCIImageEditing)
	ADD_OBJECT_PROP(currentTool,0,PCImageTool::getClass(),true)
}

void PCIImageEditing::shutdown() {
   delete PCIImageEditingMetaClass;
}


// ----


PCIImageEditing::PCIImageEditing() {
   channel = "color";
   targetNode = "";
   currentTool = NULL;
   targetPointSet = NULL;
   channelAAT = NULL_AAT;
   zoom = 1.0f;
   pan = NULL_VECTOR2F;
   scale = 255.0f;
   offset = 0.0f;
   flipY = true;
   width = 0;
   height = 0;
}


//static inline float clamp255(float v) {
//   if (v > 255) v = 255;
//   if (v < 0) v = 0;
//   return v;
//}

void PCIImageEditing::glDrawTool(GLContext *glContext) {
   glPushAttrib(GL_ALL_ATTRIB_BITS);
   glClearColor(0,0,0,0);
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   if (targetPointSet != NULL && channelAAT != NULL_AAT) {

      int w = targetPointSet->getWidth();
      int w2 = 2 << simpleLog2(w);
      int h = targetPointSet->getHeight();
      int h2 = 2 << simpleLog2(h);

      if (w == 0 || w2 == 0 || h == 0 || h2 == 0) return;
      float tw = (float)w/(float)w2;
      float th = (float)h/(float)h2;

      glMatrixMode(GL_MODELVIEW);
      glLoadIdentity();
      glViewport(0, 0, width, height);
      if (width == 0) return;
      glScalef((float)w*(float)height/((float)h*(float)width), 1, 1);
      if (flipY) {
         glScalef(zoom, -zoom, zoom);
      } else {
         glScalef(zoom, zoom, zoom);
      }
      glTranslatef(pan[0], pan[1], 0);
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();

      Vector4ub *copyData = new Vector4ub[w2*h2];

      for (int i=0; i<h2*w2; i++) {
         copyData[i] = NULL_VECTOR4UB;
      }

      const char *data = targetPointSet->getDataPointer();
      data += channelAAT.getOffset();

      int comp = targetPointSet->getDescr()->getAttribute(channel).numComponents;
      card32 type = targetPointSet->getDescr()->getAttribute(channel).dataFormat;
      int addOffset = targetPointSet->getDescr()->getSize() - comp * (type==VAD::DATA_FORMAT_CARD8?1:4);

      for (int y=0; y<h; y++) {
         for (int x=0; x<w; x++) {
            Vector4ub vec = NULL_VECTOR4UB;
            for (int c=0; c<comp; c++) {
               float v = 0;
               switch(type) {
                case VAD::DATA_FORMAT_CARD8:   v = *data; data += 1; break;
                case VAD::DATA_FORMAT_FLOAT32: v = *((float*)data); data += 4; break;
                case VAD::DATA_FORMAT_INT32:   v = *((int*)data); data += 4; break;
               }
               if (c<4) {
                  v = (v+offset)*scale;
                  if (v > 255) v = 255;
                  if (v < 0) v = 0;
                  vec[c] = v;
               }
            }
            copyData[x + y*w2] = vec;
            data += addOffset;
         }
      }

      glTexImage2D(GL_TEXTURE_2D, 0, 4, w2, h2, 0, GL_RGBA, GL_UNSIGNED_BYTE, copyData);
      delete[] copyData;
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
      glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
      glColor4f(1,1,1,1);
      glEnable(GL_TEXTURE_2D);
      glBegin(GL_QUADS);
      glTexCoord2f(0,0);
      glVertex3f(-1,-1,0);
      glTexCoord2f(tw,0);
      glVertex3f( 1,-1,0);
      glTexCoord2f(tw,th);
      glVertex3f( 1, 1,0);
      glTexCoord2f(0,th);
      glVertex3f(-1, 1,0);
      glEnd();
   }
   glPopAttrib();
   if (currentTool != NULL) {
      currentTool->glDrawTool(glContext);
   }
   forget(glContext);
}

void PCIImageEditing::activateTool() {
   setTargetNode(targetNode);
   setChannel(channel);      
}



void PCIImageEditing::setTargetNode(std::string targetNode) {
   this->targetNode = targetNode;
   if (scene == NULL) {
      targetPointSet = NULL;
   } else {
      ParametricInCorePointCloud *pc = dynamic_cast<ParametricInCorePointCloud*>(getPointCloud(scene, targetNode));
      if (pc == NULL) {
         targetPointSet = NULL;
      } else {
         targetPointSet = pc->getPointSet();

         if (targetPointSet != NULL) {
            int w = targetPointSet->getWidth();
            int h = targetPointSet->getHeight();
            if (currentTool != NULL) currentTool->areaResize(w, h);
         }

      }
   }
   if (currentTool != NULL) {
      currentTool->setup(targetPointSet, channel);
   }
}

void PCIImageEditing::setChannel(std::string channel) {
   this->channel = channel;
   channelAAT = NULL_AAT;
   if (targetPointSet != NULL) {
      if (targetPointSet->providesAttribute(channel)) {
         channelAAT = targetPointSet->getAAT(channel);
      }
   }
   if (currentTool != NULL) {
      currentTool->setup(targetPointSet, channel);
   }
}

void PCIImageEditing::setTool(PCImageTool *tool) {
   if (currentTool != tool) {
      delete currentTool;
      currentTool = NULL;
   }
   currentTool = tool;
   if (currentTool != NULL) {
      currentTool->setup(targetPointSet, channel);
   }
}



void PCIImageEditing::keyDown(GeneralKey key) {
   if (currentTool != NULL) currentTool->keyDown(key);
}

void PCIImageEditing::keyUp(GeneralKey key) {
   if (currentTool != NULL) currentTool->keyUp(key);
}


Vector2f PCIImageEditing::unproject(float32 x, float32 y) {
   if (targetPointSet == NULL) return NULL_VECTOR2F;
   int imgw = targetPointSet->getWidth();
   int imgh = targetPointSet->getHeight();

   if (imgw <= 0 || imgh <=0) return NULL_VECTOR2F;

   x -= width / 2.0f;
   y -= height / 2.0f;
   x /= 0.5f*height;
   y /= 0.5f*height;

   x *= (float32)imgh/(float32)imgw;

   x /= zoom;
   y /= zoom;

   x -= pan[0];
   y -= pan[1];

   x /= 2.0f;
   y /= 2.0f;
   
   x += 0.5f;
   y += 0.5f;

   return makeVector2f(x,y);
}


void PCIImageEditing::mouseDown(int32 x, int32 y, MouseButtons buttonsState, Modifiers modifiersState) {
   Vector2f pos = unproject(x,y);
   if (currentTool != NULL) currentTool->mouseDown(pos[0], pos[1], buttonsState, modifiersState);
   }

void PCIImageEditing::mouseMoved(int32 x, int32 y) {
   Vector2f pos = unproject(x,y);
   if (currentTool != NULL) currentTool->mouseMoved(pos[0], pos[1]);
}

void PCIImageEditing::mouseUp(int32 x, int32 y, MouseButtons buttonsState, Modifiers modifiersState) {
   Vector2f pos = unproject(x,y);
   if (currentTool != NULL) currentTool->mouseUp(pos[0], pos[1], buttonsState, modifiersState);
}

void PCIImageEditing::areaResize(card32 width, card32 height) {
   this->width = width;
   this->height = height;
}

PCIImageEditing::~PCIImageEditing() {
   delete currentTool;
}
