#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PCIManualAlignment.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "ClassMethods.h"
#include "FSIcp.h"
#include "GeometricTools.h"
#include "PointCloudTools.h"
#include <opencv\cxcore.h>

//---------------------------------------------------------------------------
using namespace X4;

IMPLEMENT_X4_CLASS( PCIManualAlignment ,0)
{
	BEGIN_CLASS_INIT( PCIManualAlignment );
	ADD_NOARGS_METHOD_DESC( clear, "clear list of corresponding points and restart" );
	ADD_NOARGS_METHOD_DESC( align, "try to align corresponding points (you need 4 points at least)" );
	ADD_NOARGS_METHOD_DESC( icp, "fine alignment with icp" );
	ADD_NOARGS_METHOD_DESC( basePlaneAlignment, "pure magic");
	ADD_OBJECT_PROP(m_ICP,0,FSIcp::getClass(),true);
	ADD_CARD32_PROP(m_NumICPPoints,0);
	ADD_BOOLEAN_PROP(m_WriteResiduals,0);
	
}

PCIManualAlignment::PCIManualAlignment(void)
{
	m_ICP = new FSIcp;
	clear();
	m_NumICPPoints = 500;
	m_WriteResiduals = false;
}

PCIManualAlignment::~PCIManualAlignment(void)
{
	delete m_ICP;
}

void PCIManualAlignment::assign(const Object* obj, X4CopyContext *context)
{
	PCInteractionTool::assign(obj,context);
	const PCIManualAlignment * o = dynamic_cast<const PCIManualAlignment*>(obj);
	x4Assert( o != NULL );

	copyObjectProperties(obj,this);
}

void X4::PCIManualAlignment::mouseDown( int32 x, int32 y, MouseButtons buttonsState, Modifiers modifiersState )
{
	if( buttonsState.getLeft() || buttonsState.getRight() )
	{
		bool cloud1 = buttonsState.getLeft();
		if( modifiersState.getCtrl() )
		{
			if( cloud1 && m_PointList1.size())
				m_PointList1.pop_back();
			if( !cloud1 && m_PointList2.size())
				m_PointList2.pop_back();
			if( m_PointList1.size() == 0 )
				m_PC1 = NULL;
			if( m_PointList2.size() == 0 )
				m_PC2 = NULL;
			return;
		}

		std::string pcName;
		PointCloud * pc;
		Vector3f point;
		if( getIntersectionFromCursor(x,y,point,pcName,&pc) )
		{
			if( cloud1 )
			{
				if( m_PC1 == NULL )
					m_PC1 = pc;
				if( m_PC1 == pc )
					m_PointList1.push_back(point);
			}
			else
			{
				if( m_PC2 == NULL && m_PC1 != pc)
					m_PC2 = pc;
				if( m_PC2 == pc )
					m_PointList2.push_back(point);
			}
				
		}


	}
}

void X4::PCIManualAlignment::glDrawTool( GLContext *glContext )
{
	// transform points based on point cloud transformation
	std::vector<Vector3f> points1,points2;
	for( unsigned i=0;i<m_PointList1.size();i++ )
	{
		points1.push_back( transformVector3f(m_PC1->getTransformation(),m_PointList1[i]) );
	}

	for( unsigned i=0;i<m_PointList2.size();i++ )
	{
		points2.push_back( transformVector3f(m_PC2->getTransformation(),m_PointList2[i]) );
	}


	glLineWidth(1.0f);
	glBegin(GL_LINES);
	glColor3f(0.65f,0.65f,0.65f);
	for( unsigned i=0;i<points1.size() && i<points2.size();i++ )
	{
		glVertex3fv( points1[i].data());
		glVertex3fv( points2[i].data());
	}
	glEnd();


	glPointSize(5.0f);
	glBegin(GL_POINTS);
	glColor3f(1,0.5,0);
	for( unsigned i=0;i<points1.size();i++ )
	{
		glVertex3fv( points1[i].data());
	}
	glColor3f(0,0.75,1);
	for( unsigned i=0;i<points2.size();i++ )
	{
		glVertex3fv( points2[i].data());
	}
	glEnd();
}

void X4::PCIManualAlignment::clear()
{
	m_PC1 = NULL;
	m_PC2 = NULL;
	m_PointList1.clear();
	m_PointList2.clear();
}

void X4::PCIManualAlignment::align()
{
	if( m_PointList1.size() != m_PointList2.size() )
	{
		message("you have to select exactly the same number of points in each point cloud");
		return;
	}

	if( m_PointList1.size() < 4 )
	{
		message("not enough points");
		return;		
	}

	// transform points based on point cloud transformation
	std::vector<Vector3f> points1,points2;
	for( unsigned i=0;i<m_PointList1.size();i++ )
	{
		points1.push_back( transformVector3f(m_PC1->getTransformation(),m_PointList1[i]) );
		points2.push_back( transformVector3f(m_PC2->getTransformation(),m_PointList2[i]) );
	}
	
	// compute centroids
	Vector3f centroid1 = NULL_VECTOR3F;
	Vector3f centroid2 = NULL_VECTOR3F;
	for( unsigned i=0;i<points1.size();i++ )
	{
		centroid1 += points1[i];
		centroid2 += points2[i];
	}
	centroid1 /= m_PointList1.size();
	centroid2 /= m_PointList2.size();

	// calculate cross-covariance matrix
	Matrix3f C = IDENTITY3F * 0;
	std::vector<Vector3f> points1_s(points1.size());
	std::vector<Vector3f> points2_s(points2.size());
	for( card32 i=0;i<points1.size();i++ )
	{
		points1_s[i] = points1[i]-centroid1;
		points2_s[i] = points2[i]-centroid2;
		for( card32 j=0;j<3;j++ )
		{
			for( card32 k=0;k<3;k++ )
			{
				C[j][k] += (points1[i][j] - centroid1[j])
					* (points2[i][k] - centroid2[k]);
			}
		}
	}
	C /= points1.size();

	// Prepare OpenCV Datastructure CvMat
	Matrix3f U,V;
	Vector3f eigenValues;
	CvMat cvMatInput, cvMat_U, cvMat_W, cvMat_Vt;
	cvInitMatHeader( &cvMatInput, 3, 3, CV_32FC1, C.data());
	cvInitMatHeader( &cvMat_W, 3, 1, CV_32FC1, eigenValues.data() );
	cvInitMatHeader( &cvMat_U, 3, 3, CV_32FC1, U.data() );
	cvInitMatHeader( &cvMat_Vt, 3, 3, CV_32FC1, V.data() );

	// Do SVD
	cvSVD(&cvMatInput, &cvMat_W, &cvMat_U, &cvMat_Vt, CV_SVD_MODIFY_A | CV_SVD_U_T );

	Matrix3f rotation = U*V;

	//calc scale
	for( card32 i=0;i<points2.size();i++ )
	{
		points2_s[i] = rotation*points2_s[i];
	}
	float scale = 0.0f;
	float nominator = 0.0f;
	float denominator = 0.0f;
	for( card32 i=0;i<points2.size();i++ )
	{
		for( card32 k=0;k<3;k++ )
		{
			nominator += points1_s[k]*points2_s[k];
			denominator += points2_s[k]*points2_s[k];
		}
	}
	scale = nominator/denominator;
	debugOutput<<"estimated scale "<<scale<<"\n";
	rotation[0][0] *= scale;
	rotation[1][1] *= scale;
	rotation[2][2] *= scale;
	Matrix4f transformation = makeTranslation4f(centroid1) * expand3To4(rotation) * makeTranslation4f(-centroid2);


	m_PC2->setTransformation( transformation * m_PC2->getTransformation() );

}

void X4::PCIManualAlignment::basePlaneAlignment()
{
	if( m_PointList1.size() < 3 )
	{
		message("not enough points");
		return;		
	}

	Vector3f normal = normalize((m_PointList1[2] - m_PointList1[0]).crossProduct(m_PointList1[1] - m_PointList1[0]));
	Matrix3f tangentSystem = calcTangentSystem(normal);

	m_PC1->setTransformation(expand3To4(tangentSystem.transpose()));

}

void X4::PCIManualAlignment::icp()
{
	if( m_PC1 && m_PC2)
	{
		PointSet * resampledSet = resamplePointCloudForICP(m_PC2, m_NumICPPoints);
		Matrix4f transf = invertFrame(m_PC1->getTransformation()) * m_PC2->getTransformation();

		m_ICP->setupData(m_PC1, getScene());
		m_ICP->calculateResidualAfterICP( resampledSet, transf);
		m_PC2->setTransformation(m_PC1->getTransformation() * transf);

		if( m_WriteResiduals && m_PC2->providesAttribute("color") )
		{
			AAT colorAAT = m_PC2->getAAT("color");
			AAT posAAT = m_PC2->getAAT("position");
			AAT normalAAT  = m_PC1->getAAT("normal");
			HierarchicalKNNIterator hIt(m_PC1,32,NULL);
			SceneObjectIterator * iter = m_PC2->createIterator(SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC);
			BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
			ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(iter);
			float medPointDistance = getMedianPointDistance(m_PC2);

			while (!bIt->atEnd())
			{
				Vector3f p = bIt->get3f(posAAT);
				p = transformVector3f(transf,p);
				hIt.setSeekPointAndReset( p );
				Vector3f np = hIt.get3f(posAAT);
				Vector3f nn = hIt.get3f(normalAAT);
				float dist = (p - np) * nn / medPointDistance;
				mIt->set3f(colorAAT,makeVector3f(dist,fabs(dist),0));
				bIt->next();
			}

			delete iter;
		}


	}
}