//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PCITopologyAnalyzer.h"
//---------------------------------------------------------------------------
#include "Scene.h"
#include "SceneGraphState.h"
#include "ObjectViewsTable.h"
#include "PointCloudIterators.h"
#include "PointCloudTools.h"
#include "UnstructuredInCorePointCloud.h"
#include "InCorePCTopologyGraph.h"
#include "PointCloud.h"
#include "PointSet.h"
#include "AttachedData.h"
#include "TopologyAttachment.h"
#include "DebugOutput.h"
#include "TopologyTools.h"
#include "PCTopologyNeighborhoodIterator.h"
#include "ExceptionHandling.h"
#include "ProgressWindow.h"
#include "VertexDescriptor.h"
#include "PointSet.inline.h"
#include "SelectionTools.h"

//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

IMPLEMENT_X4_CLASS(PCITopologyAnalyzer, 0)
{
	BEGIN_CLASS_INIT(PCITopologyAnalyzer)
}

PCITopologyAnalyzer::PCITopologyAnalyzer() {
	leftMouseDown = false;
	rightMouseDown = false;
	width = 0;
	height = 0;
	localModel = lmKernel;

	selectedCloud = "";
	selectedPoint = 0;
	comparePoint = 0;
	quantization = 32;
	epsilonFact = 0.05f;
	imageGain = 10.0f;
	imageOffset = 0.0f;
	sigma = 0.005f;
	expNumPts = 10;
}

void PCITopologyAnalyzer::mouseDown(int32 x, int32 y, MouseButtons buttonsState, Modifiers modifiersState) {
	STD_GLOBAL_EXCEPTION_HANDLER_START
	if (buttonsState.getLeft()) {
		leftMouseDown = true;
		processSelection(x, y, true);
	}
	if (buttonsState.getRight()) {
		rightMouseDown = true;
		processSelection(x, y, false);
	}
	forget(modifiersState);
	STD_GLOBAL_EXCEPTION_HANDLER_END
}

void PCITopologyAnalyzer::mouseMoved(int32 x, int32 y) {
	STD_GLOBAL_EXCEPTION_HANDLER_START
	if (leftMouseDown) {
		processSelection(x, y, true);
	}
	if (rightMouseDown) {
		processSelection(x, y, false);
	}
	STD_GLOBAL_EXCEPTION_HANDLER_END
}

void PCITopologyAnalyzer::mouseUp(int32 x, int32 y, MouseButtons buttonsState, Modifiers modifiersState) {
	STD_GLOBAL_EXCEPTION_HANDLER_START
	leftMouseDown = false;
	rightMouseDown = false;
	forget(x+y);
	forget(buttonsState);
	forget(modifiersState);
	STD_GLOBAL_EXCEPTION_HANDLER_END
}

void PCITopologyAnalyzer::areaResize(card32 width, card32 height) {
	STD_GLOBAL_EXCEPTION_HANDLER_START
	this->width = width;
	this->height = height;
	STD_GLOBAL_EXCEPTION_HANDLER_END
}


void PCITopologyAnalyzer::keyDown(GeneralKey key) {
	STD_GLOBAL_EXCEPTION_HANDLER_START
	forget(key);
	STD_GLOBAL_EXCEPTION_HANDLER_END
}

void PCITopologyAnalyzer::keyUp(GeneralKey key) {
	STD_GLOBAL_EXCEPTION_HANDLER_START
	forget(key);
	STD_GLOBAL_EXCEPTION_HANDLER_END
}

bool PCITopologyAnalyzer::drawGLSceneRendering() {
	PointCloud *pc;
	PointSet *ps;
	InCorePCTopologyGraph *tg;
	getEnv(pc, ps, tg);
	if (pc != NULL && ps != NULL && tg != NULL) {
		card32 v = selectedPoint;
		if (v >= ps->getNumEntries()) return true;
		Matrix4f transform = getTransformation(getScene(), selectedCloud);
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glMultMatrixf(transform.data());
		glPushAttrib(GL_ALL_ATTRIB_BITS);
		AAT POS = ps->getAAT("position");
		if (POS == NULL_AAT) return true;
		try {
			glDisable(GL_LIGHTING);
			glColor3f(1.0f, 1.0f, 0.3f);
			glBegin(GL_LINES);
				InCorePCTopologyGraph::VertexIndex ne = tg->getNumAdjacentVertices(v);
				for (InCorePCTopologyGraph::VertexIndex i=0; i<ne; i++) {
					int32 v2 = tg->getAdjacentVertex(v, i);
					Vector3f pos1 = ps->get3f(v, POS);
					Vector3f pos2 = ps->get3f(v2, POS);
					glVertex3fv(pos1.data());
					glVertex3fv(pos2.data());
				}
			glEnd();
			glEnable(GL_BLEND);
			glBlendFunc(GL_SRC_ALPHA, GL_ONE);
			glDisable(GL_DEPTH_TEST);
			glColor4f(1.0f, 1.0f, 1.0f, 0.1f);
			GLUquadricObj* qo = gluNewQuadric();
			const float32 bbSize = ps->getBoundingBox().getMaxSideLength();
			glMatrixMode(GL_MODELVIEW);
			Vector3f pos = ps->get3f(selectedPoint, POS);
			glPushMatrix();
			glTranslatef(pos[0], pos[1], pos[2]);
			gluSphere(qo, epsilonFact*bbSize, 16, 8);
			glPopMatrix();
			gluDeleteQuadric(qo);
		} catch (...) {
			glPopAttrib();
			glPopMatrix();
			throw;
		}
		glPopMatrix();
		glPopAttrib();
	}
	return true;
}

void PCITopologyAnalyzer::processSelection(int x, int y, bool left) {
   PointCloudIterator* it = getClosestPointIteratorFromAllPoints(x, y, width, height, 5, 
			&getScene()->getRootState()->staticState->camera, 
			&getScene()->getRootState()->staticState->viewFrustum,
			getScene(),
			selectedCloud
	);
	IndexedPointCloudIterator *indIt = dynamic_cast<IndexedPointCloudIterator*>(it);
	card32 sel = 0;
	if (indIt != NULL && selectedCloud != "") {
		sel = (card32)indIt->getCurrentPointIndex();
	} else {
		sel = 0;
	}
	if (left) {
		selectedPoint = sel;
	} else {
		comparePoint = sel;
	}
	ObjectViewsTable::update(this);
	delete it;
}

void PCITopologyAnalyzer::computeViz(QImage &img, bool compCompare) {
	card32 selection;
	if (compCompare) {
		selection = comparePoint;
	} else {
		selection = selectedPoint;
	}
	if (selection < 0) selection = 0;
	if (quantization < 4) quantization = 4;
	img = QImage(quantization, quantization, QImage::Format_RGB32);
	img.fill(QColor(255,255,255,255).rgba());

   PointCloud *pc;
   PointSet *ps;
   InCorePCTopologyGraph *tg;
	getEnv(pc, ps, tg);
   if (!(pc && ps && tg)) return;

	const float32 bbSize = getPCBBox(pc).getMaxSideLength();
	const float32 epsilon = epsilonFact * bbSize;
	debugOutput << "\nPCITopologyAnalyzer::computeViz()\n";
	debugOutput << "absolute epsilon = " << epsilon << "\n";
	AAT posAAT = pc->getAAT("position");
	if (posAAT == NULL_AAT) return;

	if (pc->getNumPoints() <= selection) return;

	PCTopologyNeighborhoodIterator it(tg, ps);
	it.goToVertex(selection);
	if (!it.atEnd()) {
		
		const Vector3f currentPointPos = ps->get<float32, 3>(selection, posAAT);
		
		for (card32 qangle=0; qangle<quantization; qangle++) {
			Degree angle = (float32)qangle / (float32)quantization * M_PI;
			for (card32 qdist=0; qdist<quantization; qdist++) {
				float32 dist = (float32)qdist / ((float32)(quantization)-1.0f) * 2.0f - 1.0f;
				dist *= epsilon;
				it.reset();
				const Vector3f normal = get2dNormal(angle);
				const float32 pointDist = normal * currentPointPos;
				float32 value;
				int vi;
				if (localModel == lmMeanSquare) {
					value = computeSSDPlaneFitError(normal, dist + pointDist, &it, posAAT);
					vi = (int)(255.0f*exp(-value*imageGain));
				} else if (localModel == lmRobustStat) {
					value = computeRobustStatisticsFitError(normal, dist + pointDist, &it, posAAT, expNumPts, sigma);
					vi = (int)(255.0f*exp(-value*imageGain));
				} else {
					value = computeKernelPlaneFitError(normal, dist + pointDist, &it, posAAT, sigma);
					vi = (int)(value*imageGain + imageOffset);
					if (vi<0) vi = 0;
					if (vi>255) vi = 255;
				}
				
				QRgb c = qRgb ( vi, vi, vi );
				img.setPixel(qangle, qdist, c);
			}
		}

	} else {
		debugOutput << "no edges in neighborhood.\n";
	}
	debugOutput << "\n";
}

void PCITopologyAnalyzer::getEnv(PointCloud *&pc, PointSet *&ps, InCorePCTopologyGraph *&tg) {
	tg = NULL;
	ps = NULL;
   pc = getPointCloud(scene, selectedCloud);
   if (!pc) return;
	UnstructuredInCorePointCloud *upc = dynamic_cast<UnstructuredInCorePointCloud*>(pc);
	if (!upc) return;
	ps = upc->getPointSet();
   AttachedDataContainer *att = upc->getAttachments();
   TopologyAttachment *topAtt = dynamic_cast<TopologyAttachment*>(att->getData("TopologyAtt"));
   if (topAtt) {
	   tg = topAtt->getTopology();
	}
}

static double computeSum(QImage &img) {
	const int w = img.width();
	const int h = img.height();
	double result = 0;
	for (int y=0; y<h; y++) {
		for (int x=0; x<w; x++) {
			QRgb c = img.pixel(x,y);
			result += double(qRed(c))/255.0f/(double(w)*double(h));
		}
	}
	return result;
}

void PCITopologyAnalyzer::computeLocalVariance() {
	PointCloud *pc;
	PointSet *ps;
	InCorePCTopologyGraph *tg;
	getEnv(pc, ps, tg);	
	if (!ps || !pc || !tg) {
		warning("PCITopologyAnalyzer::computeLocalVariance() - pc, ps or tg is NULL.");
		return;
	}
	
   checkAttribute(getScene(), selectedCloud, "color", 4, VAD::DATA_FORMAT_FLOAT32 ) ;
	getEnv(pc, ps, tg);	
	if (!ps || !pc || !tg) {
		warning("PCITopologyAnalyzer::computeLocalVariance() - pc, ps or tg is NULL after checking attributes (internal error).");
		return;
	}

	const AAT COL = ps->getAAT("color");
	if (COL == NULL_AAT) {
		warning("PCITopologyAnalyzer::computeLocalVariance() - no color channel found in point cloud to write result to (also it was tried to create it), aborted.");
		return;
	}
	VAD VCOL = ps->getDescr()->getAttribute("color");
	
	progressWindow->pushStep(true, "computing local variance");
	card32 n = ps->getNumEntries();
	card32 selSave = selectedPoint;
	QImage img(quantization, quantization, QImage::Format_ARGB32);
	for (card32 i=0; i<n; i++) {
		selectedPoint = i;
		computeViz(img, false);
		float64 v = computeSum(img);
		Vector4d val = makeVector4d(v,v,v,1);
		VCOL.setAnyAttrib((card8*)ps->getDataPointer(i), val, COL);
		if (!progressWindow->progress((card32)(float(i)/float(n)*100.0f))) break;
	}
	bool x;
	progressWindow->popSteps(&x);
	selectedPoint = selSave;
}

void PCITopologyAnalyzer::computeLinkVariance() {
	PointCloud *pc;
	PointSet *ps;
	InCorePCTopologyGraph *tg;
	getEnv(pc, ps, tg);	
	if (!ps || !pc || !tg) {
		warning("PCITopologyAnalyzer::computeLocalVariance() - pc, ps or tg is NULL.");
		return;
	}
	PointSet *edges = tg->getEdges();
	if (edges == NULL) {
		warning("PCITopologyAnalyzer::computeLocalVariance() - no edges.");
		return;		
	}
	
	const AAT COL = edges->getAAT("color");
	if (COL == NULL_AAT || !edges->getDescr()->providesAttribute("color", 4, VAD::DATA_FORMAT_FLOAT32)) {
		warning("PCITopologyAnalyzer::computeLocalVariance() - no matching color channel found in edges to write result to, aborted (try recomputing topology with latest PCCComputeTopology).");
		return;
	}

	const AAT IDX = edges->getAAT("vindex");
	if (IDX == NULL_AAT) {
		warning("PCITopologyAnalyzer::computeLocalVariance() - no index channel found in edge table.");
		return;
	}
	
	progressWindow->pushStep(true, "computing edge variance");
	card32 n = edges->getNumEntries();
	card32 selSave = selectedPoint;
	card32 compSave = comparePoint;
	QImage img(quantization, quantization, QImage::Format_ARGB32);
	for (card32 i=0; i<n; i++) {
		Vector2i currentIndices = edges->get<int32, 2>(i, IDX);
		selectedPoint = currentIndices[0];
		comparePoint = currentIndices[1];
		computeViz(img, false);
		QImage img2(quantization, quantization, QImage::Format_ARGB32);
		computeViz(img2, true);

		const int w = img.width();
		const int h = img.height();
		double diff = 0;
		for (int y=0; y<h; y++) {
			for (int x=0; x<w; x++) {
				QRgb c1 = img.pixel(x,y);
				QRgb c2 = img2.pixel(x,y);
				diff += double(abs(qRed(c1)-qRed(c2)))/255.0f/(double(w)*double(h));
			}
		}
		
		Vector4f val = makeVector4f(diff,diff,diff,1);
		edges->set<float32, 4>(i, COL, val);
		if (i % 100 == 0) {
			if (!progressWindow->progress((card32)(float(i)/float(n)*100.0f))) break;
		}
	}
	selectedPoint = selSave;
	comparePoint = compSave;
	bool x;
	progressWindow->popSteps(&x);
}
