//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PCITopologyComputation.h"
//---------------------------------------------------------------------------
#include "ClassMethods.h"
#include "SeparatorClassProperty.h"
#include "PointCloudTools.h"
#include "ExceptionHandling.h"
#include "BooleanClassProperty.h"
#include "NumericalClassProperty.h"
#include "DebugOutput.h"
#include "InCorePCTopologyGraph.h"
#include "TopologyAttachment.h"
#include "TopologyTools.h"
#include "PCTopologyNeighborhoodIterator.h"
#include "PointSet.h"
#include "PointSet.inline.h"
#include "VertexDescriptor.h"
#include "ColorTools.h"
#include "ImageDisplayer.h"
#include "LocalGeometryModelStatistics.h"
#include "InterfaceGlobals.h"
#include "EventViewerWidget.h"
#include "SceneCommandObjects.h"
#include "ObjectClassProperty.h"
#include "PCCComputeTopology.h"
#include "GLVolumeDisplayer.h"
#include "GLEventDisplayer.h"
#include "RGBAImage.h"
#include "HoughTransformProcessing.h"
#include "CmdObjMemberMethod.h"
#include "LocalModelAttachment.h"
#include "ProgressWindow.h"
#include "RegularData.inline.h"
#include "LocalGeometryUncertainty.h"
#include "TopologyEstimator.h"
#include "SelectionTools.h"
//---------------------------------------------------------------------------
#include <queue>
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------



typedef InCorePCTopologyGraph::VertexIndex VIT;



static std::string descrFeedbackMode(card32 i) {
	if (i == PCITopologyComputation::FM_NONE) return "FM_NONE";
	if (i == PCITopologyComputation::FM_SHOW_LOCAL_MODEL) return "FM_SHOW_LOCAL_MODEL";
	return "invalid";
}

IMPLEMENT_X4_CLASS( PCITopologyComputation , 4 )
{
	BEGIN_CLASS_INIT( PCITopologyComputation );
	ADD_SEPARATOR( "target" )
	ADD_STRING_PROP( targetNode, 0 )
	ADD_SEPARATOR( "settings" )

	ADD_OBJECT_PROP(baseTopologyEstimator, 0, PCCComputeTopology::getClass(), true)
	ADD_OBJECT_PROP(localModel, 0, LocalGeometryModelStatistics::getClass(), true)
	ADD_OBJECT_PROP(htCornerDetector, 0, HoughTransformCornerDetector::getClass(), true)
	ADD_OBJECT_PROP(uncertModel, 2, LocalGeometryUncertainty::getClass(), true);
	ADD_OBJECT_PROP(topologyEstimator, 4, TopologyEstimator::getClass(), true);

	ADD_BOOLEAN_PROP( readStdDevFromPC, 0 )
	ADD_FLOAT32_PROP( constStdDev, 0 )
	ADD_BOOLEAN_PROP( readSamplingDensityFromPC, 0 )
	ADD_FLOAT32_PROP( constSamplingDensity, 0 )
	ADD_BOOLEAN_PROP( colorPoints, 0 )

	ADD_SEPARATOR( "rendering" )
	ADD_BOOLEAN_PROP( showCompleteTopology, 3 )
	ADD_BOOLEAN_PROP( showUncertainty, 0 )
	ADD_BOOLEAN_PROP( showLocalModelsVis, 1 )
	ADD_BOOLEAN_PROP( showCurrentLocalModelVis, 1 )
	ADD_BOOLEAN_PROP( showAllLocalModelVis, 1 )
	ADD_FLOAT32_PROP( localModelVisRadius, 1 )
	ADD_FLOAT32_PROP( localModelVisGamma, 1 )
	ADD_CARD32_PROP( maxNumMaximaVis, 1)
	ADD_CARD32_PROP_UPDATE_DESCR( feedbackMode, 0, NULL, descrFeedbackMode, 2, NULL )

	ADD_SEPARATOR( "computation" )
	ADD_BOOLEAN_PROP( addExtraPosChannel, 0 )
	ADD_CARD32_PROP( numDiffusionSteps, 3 )
	ADD_BOOLEAN_PROP( useMaxDiffusion, 3 );
	 
	ADD_NOARGS_METHOD_DESC( showFeedbackWindow, "open feedback window" )
	ADD_NOARGS_METHOD_DESC( computeBaseTopology, "execute baseTopologyEstimator to create base topology estimate (e.g. kNN graph)" )

	ADD_NOARGS_METHOD_DESC( computeLocalModelMaxima, "computes local model Hough-transforms and estimates maxima; stores them in additional channel \"localModel\"." )
	ADD_NOARGS_METHOD_DESC( computeUncertainty, "compute local PCA eigenvalue ratio per point." )
	ADD_NOARGS_METHOD_DESC( diffuseUncertainty, "Laplacian diffusion w. renormalization for uncertainty values" );
	ADD_NOARGS_METHOD_DESC( computeTopologyGreedy, "topology estimation using the \"greedy\" algorithm: compute statistical point cloud topology estimate - computes all other properties as well as intermediate results." )

	ADD_NOARGS_METHOD_INVISIBLE( computeLocalModelMaxima_internCmdObjCall)
	ADD_NOARGS_METHOD_INVISIBLE( computeUncertainty_internCmdObjCall )
	ADD_NOARGS_METHOD_INVISIBLE( diffuseUncertainty_internCmdObjCall )
	ADD_NOARGS_METHOD_INVISIBLE( computeTopologyGreedy_internCmdObjCall )
}


// ---


PCITopologyComputation::PCITopologyComputation(void) {
	width = 0;
	height = 0;
	readStdDevFromPC = false;
	constStdDev = 0.1f;
	readSamplingDensityFromPC = false;
	constSamplingDensity = 10.0f;
	colorPoints = true;
	showCompleteTopology = false;
	showUncertainty = false;
	showLocalModelsVis = true;
	showCurrentLocalModelVis = false;
	showAllLocalModelVis = true;
	localModelVisRadius = 100;
	localModelVisGamma = 1;
	maxNumMaximaVis = 10000;
	feedbackMode = FM_NONE;
	leftMouseDown = false;
	rightMouseDown = false;	
	baseTopologyEstimator = new PCCComputeTopology();
	localModel = new LGMSKernelPlaneModel();
	htCornerDetector = new HTConvexSegmentation();
	uncertModel = new LGULeastSquarePlanePCA();
	topologyEstimator = new GreedyTopologyEstimator();

	feedbackWindow = NULL;
	addExtraPosChannel = false;
	numDiffusionSteps = 1;
	useMaxDiffusion = true;
}

PCITopologyComputation::~PCITopologyComputation() {
	delete localModel;
	delete baseTopologyEstimator;
	delete htCornerDetector;
}

void PCITopologyComputation::connectToSceneImpl(Scene *scene, OnSceneChangeCallback *onChange, QWidget *sceneEditorWidget, QWidget* toolWidget) {
	feedbackWindow = new EventViewerWidget(sceneEditorWidget, Qt::Window);
	feedbackWindow->setWindowTitle("PCITopologyComputation Feedback/Visualizations Window");
}


void PCITopologyComputation::showFeedbackWindow() {
	if (feedbackWindow == NULL) return;
	feedbackWindow->show();
//	feedbackWindow->resize(300, 400);
}


void PCITopologyComputation::areaResize(card32 width, card32 height) {
	STD_GLOBAL_EXCEPTION_HANDLER_START
	this->width = width;
	this->height = height;
	STD_GLOBAL_EXCEPTION_HANDLER_END
}

void PCITopologyComputation::processSelection(int x, int y, bool left) {
   PointCloudIterator* it = getClosestPointIteratorFromAllPoints(x, y, width, height, 5, 
			&getScene()->getRootState()->staticState->camera, 
			&getScene()->getRootState()->staticState->viewFrustum,
			getScene(),
			targetNode
	);
	IndexedPointCloudIterator *indIt = dynamic_cast<IndexedPointCloudIterator*>(it);
	card32 sel = 0;
	if (indIt != NULL && targetNode != "") {
		sel = (card32)indIt->getCurrentPointIndex();
		if (feedbackMode == FM_SHOW_LOCAL_MODEL) {
			showLocalModel(!left);
		}
	} else {
		sel = 0;
	}
	if (left) {
		selectedPoint = sel;
	} else {
		comparePoint = sel;
	}
	ObjectViewsTable::update(this);
	delete it;
}

void PCITopologyComputation::mouseDown(int32 x, int32 y, MouseButtons buttonsState, Modifiers modifiersState) {
	STD_GLOBAL_EXCEPTION_HANDLER_START
	if (buttonsState.getLeft()) {
		leftMouseDown = true;
		processSelection(x, y, true);
	}
	if (buttonsState.getRight()) {
		rightMouseDown = true;
		processSelection(x, y, false);
	}
	forget(modifiersState);
	STD_GLOBAL_EXCEPTION_HANDLER_END
}

void PCITopologyComputation::mouseMoved(int32 x, int32 y) {
	STD_GLOBAL_EXCEPTION_HANDLER_START
	if (leftMouseDown) {
		processSelection(x, y, true);
	}
	if (rightMouseDown) {
		processSelection(x, y, false);
	}
	STD_GLOBAL_EXCEPTION_HANDLER_END
}

void PCITopologyComputation::mouseUp(int32 x, int32 y, MouseButtons buttonsState, Modifiers modifiersState) {
	STD_GLOBAL_EXCEPTION_HANDLER_START
	leftMouseDown = false;
	rightMouseDown = false;
	forget(x+y);
	forget(buttonsState);
	forget(modifiersState);
	STD_GLOBAL_EXCEPTION_HANDLER_END
}

bool PCITopologyComputation::getEnv(PointCloud *&pc, PointSet *&ps, InCorePCTopologyGraph *&tg) {
	tg = NULL;
	ps = NULL;
   pc = getPointCloud(scene, targetNode);
   if (!pc) return false;
	UnstructuredInCorePointCloud *upc = dynamic_cast<UnstructuredInCorePointCloud*>(pc);
	if (!upc) return false;
	ps = upc->getPointSet();
   AttachedDataContainer *att = upc->getAttachments();
   TopologyAttachment *topAtt = dynamic_cast<TopologyAttachment*>(att->getData("TopologyAtt"));
   if (topAtt) {
	   tg = topAtt->getTopology();
	}
	return true;
}

void PCITopologyComputation::invalidateAttachments() {
	PointCloud *pc;
	PointSet *ps;
	InCorePCTopologyGraph *tg;
	getEnv(pc, ps, tg);
	if (pc != NULL) {
		SceneObjectIterator *it = pc->createIterator(SOIT::CAP_ATTACH_SECONDARY);
		SecondaryDataStructureIterator *ait = dynamic_cast<SecondaryDataStructureIterator*>(it);
		if (ait) {
			AttachedDataContainer *c = ait->lockAttachments();
			c->deleteDelOnWriteData();
			ait->unlockAttachments();
		}
		delete it;
	}	
}


bool PCITopologyComputation::drawGLSceneRendering() {
	PointCloud *pc;
	PointSet *ps;
	InCorePCTopologyGraph *tg;
	getEnv(pc, ps, tg);
	if (pc != NULL && ps != NULL && tg != NULL) {
		LocalModelAttachment *lmAtt = dynamic_cast<LocalModelAttachment*>(pc->getAttachments()->getData("localModelMaxima"));
		card32 v = selectedPoint;
		if (v >= ps->getNumEntries()) return true;
		Matrix4f transform = getTransformation(getScene(), targetNode);
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glMultMatrixf(transform.data());
		glPushAttrib(GL_ALL_ATTRIB_BITS);
		AAT POS = ps->getAAT("position");
		if (POS == NULL_AAT) return true;
		try {
			glDisable(GL_LIGHTING);
			glColor3f(1.0f, 1.0f, 0.3f);
			glBegin(GL_LINES);
				InCorePCTopologyGraph::VertexIndex ne = tg->getNumAdjacentVertices(v);
				for (InCorePCTopologyGraph::VertexIndex i=0; i<ne; i++) {
					int32 v2 = tg->getAdjacentVertex(v, i);
					Vector3f pos1 = ps->get3f(v, POS);
					Vector3f pos2 = ps->get3f(v2, POS);
					glVertex3fv(pos1.data());
					glVertex3fv(pos2.data());
				}
			glEnd();
			glEnable(GL_BLEND);
			glBlendFunc(GL_SRC_ALPHA, GL_ONE);
			glDisable(GL_DEPTH_TEST);

			if (showCompleteTopology) {			
				card32 numPoints = ps->getNumEntries();
				glBegin(GL_LINES);
					glColor4f(0.3f, 0.3f, 1.0f, 0.08f);
					for (card32 j=0; j<numPoints; j++) {
						InCorePCTopologyGraph::VertexIndex ne = tg->getNumAdjacentVertices(j);
						for (InCorePCTopologyGraph::VertexIndex i=0; i<ne; i++) {
							int32 v2 = tg->getAdjacentVertex(j, i);
							Vector3f pos1 = ps->get3f(j, POS);
							Vector3f pos2 = ps->get3f(v2, POS);
							glVertex3fv(pos1.data());
							glVertex3fv(pos2.data());
						}
					}
				glEnd();
			}																
			
			glColor4f(1.0f, 1.0f, 1.0f, 0.1f);
			GLUquadricObj* qo = gluNewQuadric();
			const float32 bbSize = ps->getBoundingBox().getMaxSideLength();
			glMatrixMode(GL_MODELVIEW);
			Vector3f pos = ps->get3f(selectedPoint, POS);
			glPushMatrix();
			glTranslatef(pos[0], pos[1], pos[2]);
			gluSphere(qo, constStdDev, 16, 8);
			glPopMatrix();
			gluDeleteQuadric(qo);

			if (ps != NULL) {
				AAT POS = ps->getAAT("position");
				if (POS == NULL_AAT) return true;
				Matrix4f transform = getTransformation(getScene(), targetNode);
				glMatrixMode(GL_MODELVIEW);
				glPushMatrix();
				glMultMatrixf(transform.data());
				glPushAttrib(GL_ALL_ATTRIB_BITS);
				try {
					glDisable(GL_LIGHTING);
					glEnable(GL_BLEND);
					glBlendFunc(GL_SRC_ALPHA, GL_ONE);
					glDisable(GL_DEPTH_TEST);
					glColor4f(0.01f, 0.05f, 0.1f, 1.0f);
					GLUquadricObj* qo = gluNewQuadric();
					glMatrixMode(GL_MODELVIEW);
					card32 numPoints = ps->getNumEntries();
					
					if (showUncertainty) {
						/// uncertainty visualization
						for (card32 i=0; i<numPoints; i++) {
							Vector3f pos = ps->get3f(i, POS);
							glPushMatrix();
							glTranslatef(pos[0], pos[1], pos[2]);
							gluSphere(qo, constStdDev, 10, 5);
							glPopMatrix();
						}
						gluDeleteQuadric(qo);
					}

					/// local model visualization
					if (lmAtt != NULL && showLocalModelsVis) {
						glBegin(GL_LINES);
						for (card32 i=0; i<numPoints; i++) {
							if (i == selectedPoint) {
								glColor4f(0.4f, 0.2f, 0.2f, 1.0f);
							} else {
								if (!showAllLocalModelVis) continue;
								glColor4f(0.02f, 0.01f, 0.01f, 1.0f);
							}
							
							Vector3f pos = ps->get3f(i, POS);
							if (lmAtt->getNumLists() > i) {
								// draw local model maxima in image
								for (unsigned j=0; j<min(maxNumMaximaVis,lmAtt->getNumMaxima(i)); j++) {
									Vector3f params = lmAtt->getMaxium(i, j).pos;
									card32 quantization = lmAtt->getQuantization();
									float32 rotf = (float32)params[0]/(float32)quantization * M_PI;
									float32 distf = (2.0f*(float32)params[1]/(float32)(quantization-1) - 1.0f) /** xxx_missing mult: * maxDistance */;
									Vector3f normal = makeVector3f(cos(rotf), sin(rotf), 0);
									Vector3f dir = makeVector3f(normal[1], -normal[0], 0)*localModelVisRadius;
									pos += normal*distf;
									glVertex3fv((pos-dir).data());
									glVertex3fv((pos+dir).data());
								}
							}
						}							 
						glEnd();
					}
					
					int width = 0;
					int height = 0;
					if (lastResult.getDim() == 2) {
						width = lastResult.getExtends(0);
						height = lastResult.getExtends(1);
					}

					/// current local model vis					
					if (showCurrentLocalModelVis && lastResult.getDim() == 2 && width > 0 && height > 0) {

						glBegin(GL_LINES);
						try {
							AAT POS = ps->getAAT("position");
							const float32 *datapos = lastResult.dataPtr();
							float32 minv = lastResult.minComponent();
							float32 maxv = lastResult.maxComponent();
							float32 localModelVisScale = 1.0f/max(static_cast<float>(maxv-minv),static_cast<float>(1e-10));
							float32 localModelVisOffset = -minv;
							for (int y=0; y<height; y++) {
								for (int x=0; x<width; x++) {
									glColor4f(0.4f, 0.2f, 0.2f, pow((*datapos+localModelVisOffset) * localModelVisScale, 1.0f/localModelVisGamma));
									++datapos;
									
									Vector3f pos = ps->get3f(selectedPoint, POS);
									// draw local model maxima in image

									card32 quantization = localModel->getQuantization();
									float32 rotf = (float32)x/(float32)quantization * M_PI;
									float32 distf = (2.0f*(float32)y/(float32)(quantization-1) - 1.0f) /** xxx_missing mult: * maxDistance */;
									Vector3f normal = makeVector3f(cos(rotf), sin(rotf), 0);
									Vector3f dir = makeVector3f(normal[1], -normal[0], 0)*localModelVisRadius;
									pos += normal*distf;
									glVertex3fv((pos-dir).data());
									glVertex3fv((pos+dir).data());
								}
							}							 
						} catch(...) {
							glEnd();
							throw;
						}
						glEnd();
					}
					
				} catch (...) {
					glPopMatrix();
					glPopAttrib();
					throw;
				}
				glPopMatrix();
				glPopAttrib();
			}

		} catch (...) {
			glMatrixMode(GL_MODELVIEW);
			glPopMatrix();
			glPopAttrib();
			throw;
		}
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();
		glPopAttrib();
	}
	return true;
}

void PCITopologyComputation::showLocalModel(bool compare) {
	card32 selection;
	if (compare) {
		selection = comparePoint;
	} else {
		selection = selectedPoint;
	}
	if (selection < 0) selection = 0;

   PointCloud *pc;
   PointSet *ps;
   InCorePCTopologyGraph *tg;
	getEnv(pc, ps, tg);
   if (!(pc && ps && tg)) return;

	if (!pc->getDescr()->providesAttribute("position", 3, VAD::DATA_FORMAT_FLOAT32)) return;  
	AAT posAAT = pc->getAAT("position");

	if (pc->getNumPoints() <= selection) return;

	VIT numVert = tg->getNumAdjacentVertices(selection);
	std::vector<Vector3f> points;
	points.reserve(numVert);
	for (VIT i=0; i<numVert; i++) {
		VIT j = tg->getAdjacentVertex(selection, i);
		const Vector3f p = ps->get<float32, 3>(j, posAAT);
		points.push_back(p);
	}

	RegularDataF result;
	Vector3f origin = ps->get3f(selection, posAAT);
	localModel->compute(points, origin, result);		
	lastResult = result;
	if (result.getDim() == 2) {
		std::vector<HTCorner> corners;
		htCornerDetector->detectCorners(result, corners);
		const int width = result.getExtends(0);
		const int height = result.getExtends(1);
		RGBAFloat32Image img(width, height);
		const float32 *pos = result.dataPtr();
		for (int y=0; y<height; y++) {
			for (int x=0; x<width; x++) {
				float32 c = *pos;
				img.setPixel(x,y, makeVector4f(c,c,c,1));
				++pos;
			}
		}
		for (unsigned i=0; i<corners.size(); i++) {
			int x = (int)corners[i].pos[0];
			int y = (int)corners[i].pos[1];
			Vector4f pix = img.getPixel(x,y);
			pix[1] = 0;
			pix[2] = corners[i].magnitude;
			img.setPixel(x,y, pix);
		}
		ImageDisplayer::DisplayRGBAFloat32Image ev(&img);
		std::string descrString;
		if (compare) {
			descrString = "Comparison";
		} else {
			descrString = "Local Model";
		}
		showFeedbackWindow();
		if (feedbackWindow != NULL) {
			feedbackWindow->displayEvent(ImageDisplayer::getClass(), descrString, &ev);
			double zoomFact = 200.0 / result.getExtends(1);
			ImageDisplayer::ChangeZoom zev(zoomFact);
			feedbackWindow->displayEvent(ImageDisplayer::getClass(), descrString, &zev);
		}

	} else if (result.getDim() == 3) {
		GLVolumeDisplayer *voldisp = new GLVolumeDisplayer(&result, "lattitude", "longitude", "distance");
		GLEventDisplayer::SetNewGLEventObject ev(voldisp);

		std::string descrString;
		if (compare) {
			descrString = "Comparison (Vol)";
		} else {
			descrString = "Local Model (Vol)";
		}
		showFeedbackWindow();
		if (feedbackWindow != NULL) {
			feedbackWindow->displayEvent(GLEventDisplayer::getClass(), descrString, &ev);
		}
	}  else {
		x4Assert(false);
	}
}



void PCITopologyComputation::computeUncertainty() {
	CmdObjMemberMethod *cmdObj = new CmdObjMemberMethod();
	cmdObj->setup("computeUncertainty_internCmdObjCall", this);
	getScene()->getCommandHistory()->executeCommand(getScene(), cmdObj);
	redrawScene();
}

void PCITopologyComputation::computeUncertainty_internCmdObjCall() {
	PointCloud *pc;
	PointSet *ps;
	InCorePCTopologyGraph *tg;
	if (!getEnv(pc, ps, tg)) {
		throw X4Exception("PCITopologyComputation::computeUncertainty() - cannot retrieve point cloud, point set or topology graph. Please check selection.");
	}
   checkAttribute(getScene(), targetNode, "localModelUncertainty", 1, VAD::DATA_FORMAT_FLOAT32) ;
	if (!getEnv(pc, ps, tg)) {
		throw X4Exception("PCITopologyComputation::computeUncertainty() - adding attributes failed (point cloud not found after adding localModelUncertainty channel).");
	}	
	const AAT UNCERTAIN = ps->getAAT("localModelUncertainty");
	if (UNCERTAIN == NULL_AAT) {
		throw X4Exception("PCITopologyComputation::computeUncertainty() - adding attributes failed (channel not found).");
	}
	const AAT POS = ps->getAAT("position");
	if (POS == NULL_AAT) {
		throw X4Exception("PCITopologyComputation::computeUncertainty() - cannot find position channel.");
	}
	AAT COL;
	bool doColorPoints = false;
	if (colorPoints) {
	   checkAttribute(getScene(), targetNode, "color", 4, VAD::DATA_FORMAT_FLOAT32) ;
		if (!getEnv(pc, ps, tg)) {
			throw X4Exception("PCITopologyComputation::computeUncertainty() - adding attributes failed (point cloud not found after adding color channel).");
		}
		COL = ps->getAAT("color");
		if (COL != NULL_AAT) {
			doColorPoints = true;
		}
	}
	PCTopologyNeighborhoodIterator it(tg, ps);
	const card32 n = ps->getNumEntries();
	std::vector<Vector3f> points;
	float32 maxUncert = 0;
	progressWindow->pushStep(true, "est. uncertainty");
	for (card32 i=0; i<n; i++) {
		it.goToVertex(i);
		StaticVector<float32, 1> uncertain;
		collectPoints(&it, POS, points);
		uncertain[0] = uncertModel->compute(points);
		ps->set<float32, 1>(i, UNCERTAIN, uncertain);
		if (doColorPoints) {
			if (fabs(uncertain[0]) > maxUncert) maxUncert = fabs(uncertain[0]);
		}
		progressWindow->progress((card32)((double)i/(double)n*100.0));
	}
	if (maxUncert < 1e-15f) maxUncert = 1e-15f;	
	if (doColorPoints) {
		for (card32 i=0; i<n; i++) {
				ps->set<float32, 4>(i, COL, rainbowGradient(ps->get<float32, 1>(i, UNCERTAIN)[0]/maxUncert));
		}
	}
	progressWindow->popSteps();
	invalidateAttachments();
	redrawScene();
}

void PCITopologyComputation::computeBaseTopology() {
	if (targetNode == "") {
		warning("PCITopologyComputation::computeBaseTopology() - TargetNode not specified, done nothing.");
		return;
	}
	CmdObjChangeSelection *select = new CmdObjChangeSelection;
	std::vector<std::string> sel; 
	sel.push_back(targetNode);
	select->setup(sel);
	getScene()->getCommandHistory()->executeCommand(getScene(), select);

	CmdObjPCCommand *compute = new CmdObjPCCommand;
	compute->setup(baseTopologyEstimator);
	getScene()->getCommandHistory()->executeCommand(getScene(), compute);	
	redrawScene();
}

void PCITopologyComputation::diffuseUncertainty() {
	CmdObjMemberMethod *cmdObj = new CmdObjMemberMethod();
	cmdObj->setup("diffuseUncertainty_internCmdObjCall", this);
	getScene()->getCommandHistory()->executeCommand(getScene(), cmdObj);
	redrawScene();
}

void PCITopologyComputation::diffuseUncertainty_internCmdObjCall() {
	PointCloud *pc;
	PointSet *ps;
	InCorePCTopologyGraph *tg;
	if (!getEnv(pc, ps, tg)) {
		throw X4Exception("PCITopologyComputation::diffuseUncertainty_internCmdObjCall() - cannot retrieve point cloud, point set or topology graph. Please check selection.");
	}
   if (!ps->getDescr()->providesAttribute("localModelUncertainty", 1, VAD::DATA_FORMAT_FLOAT32)) {
		throw X4Exception("PCITopologyComputation::diffuseUncertainty_internCmdObjCall() - localModelUncertainty1f channel not found.");
	}	
	const AAT UNCERTAIN = ps->getAAT("localModelUncertainty");
	x4Assert(UNCERTAIN != NULL_AAT);
	const AAT POS = ps->getAAT("position");
	if (POS == NULL_AAT) {
		throw X4Exception("PCITopologyComputation::diffuseUncertainty_internCmdObjCall() - cannot find position channel.");
	}
	AAT COL;
	bool doColorPoints = false;
	if (colorPoints) {
	   checkAttribute(getScene(), targetNode, "color", 4, VAD::DATA_FORMAT_FLOAT32) ;
		if (!getEnv(pc, ps, tg)) {
			throw X4Exception("PCITopologyComputation::diffuseUncertainty_internCmdObjCall() - adding attributes failed (point cloud not found after adding color channel).");
		}
		COL = ps->getAAT("color");
		if (COL != NULL_AAT) {
			doColorPoints = true;
		}
	}
	
	PCTopologyNeighborhoodIterator it(tg, ps);
	const card32 n = ps->getNumEntries();

	progressWindow->pushStep(true, "diffusing. uncertainty");
	for (card32 step=0; step<numDiffusionSteps; step++) {
		std::vector<float32> newVals(n,0);
		for (card32 i=0; i<n; i++) {
			card32 numOthers = 0;
			float32 othersValuesSum = 0.0f;
			float32 maxVal = 1e-30f;
			it.goToVertex(i);
			while (!it.atEnd()) {
				float32 v = it.get1f(UNCERTAIN);
				if (v>maxVal) maxVal = v;
				othersValuesSum += v;
				numOthers++;
				it.next();
			}
			
			if (numOthers != 0) {
				if (useMaxDiffusion) {
					newVals[i] = maxVal;
				} else {
					newVals[i] = ( othersValuesSum + ps->get<float32, 1>(i, UNCERTAIN)[0] )  /  (numOthers+1);
				}
			}
			
			progressWindow->progress((card32)((double)(i+n*step)/(double)(n*numDiffusionSteps)*100.0));
		}
		//float32 maxv = -1e30f;
		//float32 minv = 1e30f;
		//for (card32 i=0; i<n; i++) {
		//	float32 v = newVals[i];
		//	if (minv > v) minv = v;
		//	if (maxv < v) maxv = v;
		//}
		//for (card32 i=0; i<n; i++) {
		//	StaticVector<float32, 1> uncertain;
		//	uncertain[0] = (newVals[i] - minv)/(maxv-minv);
		//	ps->set<float32, 1>(i, UNCERTAIN, uncertain);
		//}
		for (card32 i=0; i<n; i++) {
			StaticVector<float32, 1> uncertain;
			uncertain[0] = newVals[i];
			ps->set<float32, 1>(i, UNCERTAIN, uncertain);
		}
	}
	
	if (doColorPoints) {
		float32 maxv = -1e30f;
		float32 minv = 1e30f;
		for (card32 i=0; i<n; i++) {
			float32 v = ps->get<float32, 1>(i, UNCERTAIN)[0];
			if (minv > v) minv = v;
			if (maxv < v) maxv = v;
		}
		for (card32 i=0; i<n; i++) {
			ps->set<float32, 4>(i, COL, rainbowGradient((ps->get<float32, 1>(i, UNCERTAIN)[0] - minv)/(maxv-minv)));
		}
	}
	progressWindow->popSteps();
	invalidateAttachments();
	redrawScene();
}


void PCITopologyComputation::computeLocalModelMaxima() {
	CmdObjMemberMethod *cmdObj = new CmdObjMemberMethod();
	cmdObj->setup("computeLocalModelMaxima_internCmdObjCall", this);
	getScene()->getCommandHistory()->executeCommand(getScene(), cmdObj);
	redrawScene();
}

void PCITopologyComputation::computeLocalModelMaxima_internCmdObjCall() {
	if (addExtraPosChannel) {
		checkAttribute(getScene(), targetNode, "pos2", 3, VAD::DATA_FORMAT_FLOAT32 ) ;
	}
   PointCloud *pc;
   PointSet *ps;
   InCorePCTopologyGraph *tg;
	getEnv(pc, ps, tg);
   if (!(pc && ps && tg)) return;
	if (!pc->getDescr()->providesAttribute("position", 3, VAD::DATA_FORMAT_FLOAT32)) return;  
	AAT posAAT = pc->getAAT("position");
   AAT pos2AAT;
   if (addExtraPosChannel) {
		pos2AAT = pc->getAAT("pos2");
   }
   card32 numPoints = ps->getNumEntries();
   
	LocalModelAttachment *resultAtt = new LocalModelAttachment();
   resultAtt->setup("localModelMaxima", AttachedData::ADF_PERSISTENT);
   resultAtt->setupLocalModel(localModel->getQuantization());
	
	progressWindow->pushStep();
	
   for (card32 pos=0; pos<numPoints; pos++) {
		VIT numVert = tg->getNumAdjacentVertices(pos);
		std::vector<Vector3f> points;
		points.reserve(numVert);
		for (VIT i=0; i<numVert; i++) {
			VIT j = tg->getAdjacentVertex(pos, i);
			const Vector3f p = ps->get<float32, 3>(j, posAAT);
			points.push_back(p);
		}

		RegularDataF result;
		Vector3f origin = ps->get3f(pos, posAAT);
		localModel->compute(points, origin, result);		

		std::vector<HTCorner> corners;
		htCornerDetector->detectCorners(result, corners);
		
		resultAtt->addMaximaList(corners);
		if (addExtraPosChannel && !corners.empty()) {
			ps->set3f(pos, pos2AAT, corners[0].pos);
		}
		if (!progressWindow->progress((card32)((float)pos/(float)numPoints*100.0f))) break;
	}  
   progressWindow->popSteps();
   pc->getAttachments()->attachData(resultAtt);
}


void PCITopologyComputation::computeTopologyGreedy()
{	
	progressWindow->pushStep(false, "comp. base topology", 5);
	progressWindow->pushStep(true, "est. local uncertainty", 1);
	progressWindow->pushStep(true, "diffuse uncertainty", 2);
	progressWindow->pushStep(true, "greedy topology", 2);

	/// est base topology
	computeBaseTopology();
	if (progressWindow->popStep()) return;
	
	/// est uncertainty
	computeUncertainty();
	if (progressWindow->popStep()) return;

	/// diffuse uncertainty
	diffuseUncertainty();
	if (progressWindow->popStep()) return;

	/// greedy topology computation
	CmdObjMemberMethod *cmdObj = new CmdObjMemberMethod();
	cmdObj->setup("computeTopologyGreedy_internCmdObjCall", this);
	getScene()->getCommandHistory()->executeCommand(getScene(), cmdObj);
	redrawScene();
	progressWindow->popStep();
}


struct PrioQueueEntry {
	card32 index;
	float32 weight;

	PrioQueueEntry(card32 i, float32 w) {
		index = i; weight = w;
	}
	inline bool operator<(const PrioQueueEntry &other) const {
		return weight < other.weight;
	}
};

void PCITopologyComputation::computeTopologyGreedy_internCmdObjCall() {
	PointCloud *pc;
	PointSet *ps;
	InCorePCTopologyGraph *tg;
	if (!getEnv(pc, ps, tg)) {
		throw X4Exception("PCITopologyComputation::computeTopologyGreedy_internCmdObjCall() - cannot retrieve point cloud, point set or topology graph. Please check selection.");
	}

	/// create new topology graph, exchange for current topology graph
	InCorePCTopologyGraph *newTg = topologyEstimator->estimateTopology(pc, ps, tg, true);
	
	AttachedDataContainer *att = pc->getAttachments();
	TopologyAttachment *topAtt = dynamic_cast<TopologyAttachment*>(att->getData("TopologyAtt"));
	if (topAtt) {
		topAtt->swapTopology(newTg);
	} 
}
