//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PCITriangleMesh.h"
#include "TriangleMeshTools.h" 
#include "ClassMethods.h"
#include "SeparatorClassProperty.h"
#include "GLShaderMaterial.h"
#include "VertexArray.h"
#include "CopyObjectProperties.h"
#include <QtGui/qinputdialog>
#include "DebugRenderer.h"
#include "Line.h"
#include "InCorePCTopologyGraph.h"
#include "TopologyAttachment.h"
#include "PCCResampler.h"
#include "PointCloudTools.h"
#include "Polygon.inline.h"
#include "SelectionTools.h"
#include "MajorKDTree.h"
#include "RaytracingMath.h"
#include "Plane.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------


std::string getParamModeDescr( const card32 index )
{
	if ( index == MESH_PARAMETERIZATION_MODE_SPHERE ){
		return "Sphere";
	} else if ( index == MESH_PARAMETERIZATION_MODE_PLANE ){
		return "Plane";
	} else if ( index == MESH_PARAMETERIZATION_MODE_CYLINDER ){
		return "Cylinder";
	} else {
		return "Undefined";
	}
}


std::string getUpDirectionDescr(const card32 index )
{
	switch(index)
	{
	case 0:
		return "x-direction";
	case 1:
		return "y-direction";
	case 2:
		return "z-direction";
	}
	return "unknown";
}

IMPLEMENT_X4_CLASS( PCITriangleMesh, 8 )
{
	BEGIN_CLASS_INIT( PCITriangleMesh )
	ADD_SEPARATOR( "General" )
	ADD_BOOLEAN_PROP(m_DeleteTriangleMode,6)
	ADD_STRING_PROP( meshNode, 0 )
	//ADD_SEPARATOR( "Snapping" )
	ADD_FLOAT32_PROP( maxSnapDistance, 2 )
	ADD_FLOAT32_PROP( m_SampleRadius, 5 )
	
	//ADD_SEPARATOR( "Parameterization" )
	//ADD_CARD32_PROP_UPDATE_DESCR( parameterizationMode, 0, NULL, getParamModeDescr, 3 );
	//ADD_FLOAT32_PROP( textureRepeat, 0 )
	
	//ADD_BOOLEAN_PROP(pcaROT, 1)
	//ADD_CARD32_PROP(coordIndex0, 1)
	//ADD_CARD32_PROP(coordIndex1, 1)
	ADD_CARD32_PROP(m_GroundPlaneResolution,1);

	ADD_SEPARATOR( "Mesh closing" )
	ADD_BOOLEAN_PROP(checkNormals, 3);
	ADD_SEPARATOR( "Groundplane" );
	ADD_CARD32_PROP_UPDATE_DESCR( upDirection, 4, NULL, getUpDirectionDescr, 3, NULL);
	ADD_BOOLEAN_PROP(m_DeleteAllTrianglesWithinPlane,7);
	ADD_BOOLEAN_PROP(m_SnapSelectedPointsOnly,8);

	ADD_NOARGS_METHOD( computeNormals );
   SET_METHOD_AS_CMD_OBJ( true )
   ADD_NOARGS_METHOD( adjustTriangleIndicesToNormals );
	ADD_NOARGS_METHOD( smoothNormals )
   SET_METHOD_AS_CMD_OBJ( true )
//	ADD_NOARGS_METHOD( parameterizeMesh );
   SET_METHOD_AS_CMD_OBJ( true )
	ADD_NOARGS_METHOD( createGroundPlane );
   SET_METHOD_AS_CMD_OBJ( true )
	ADD_NOARGS_METHOD( subdivide );
   SET_METHOD_AS_CMD_OBJ( true )
	ADD_NOARGS_METHOD( upsampleMesh );
	SET_METHOD_AS_CMD_OBJ( true )
	ADD_NOARGS_METHOD( closeMesh );
	SET_METHOD_AS_CMD_OBJ( true )
	ADD_NOARGS_METHOD( openMesh );
	SET_METHOD_AS_CMD_OBJ( true )
	ADD_NOARGS_METHOD( snapAndClose );
	SET_METHOD_AS_CMD_OBJ( true );
	ADD_NOARGS_METHOD( snap );
	SET_METHOD_AS_CMD_OBJ( true )
	ADD_NOARGS_METHOD( cleanMesh );
	SET_METHOD_AS_CMD_OBJ( true )
	ADD_NOARGS_METHOD( pointSampleAndCreateTop );
	SET_METHOD_AS_CMD_OBJ( true )
	ADD_NOARGS_METHOD( computeCoTanWeights );
	ADD_NOARGS_METHOD( tryToRepairTopology );
	ADD_NOARGS_METHOD( mergeSelected );
	SET_METHOD_AS_CMD_OBJ( true )
	ADD_NOARGS_METHOD(pcIntersection);
	SET_METHOD_AS_CMD_OBJ( false )
}

/***
 * Constructor
 */
PCITriangleMesh::PCITriangleMesh() 
{	
	meshNode = "root/pc1";
	parameterizationMode = MESH_PARAMETERIZATION_MODE_SPHERE;
	textureRepeat = 1.0f;
	pcaROT = true;
	coordIndex0 = 0;
	coordIndex1 = 1;
	m_GroundPlaneResolution = 100;
	maxSnapDistance = 1e-7f;
	checkNormals = false;
	upDirection = 1;
	m_SampleRadius = 0.01f;
	m_Temp_KDtree = NULL;
	m_DeleteAllTrianglesWithinPlane = false;
	m_SnapSelectedPointsOnly = false;
}

/***
 * Destructor
 */
PCITriangleMesh::~PCITriangleMesh()
{
}

void PCITriangleMesh::assign(const Object* obj, X4CopyContext *context) {
	copyObjectProperties(obj, this);
}

/** 
 * Find mesh in scene graph
 */
UnstructuredInCoreTriangleMesh* PCITriangleMesh::getMesh()
{
	return dynamic_cast<UnstructuredInCoreTriangleMesh*>( getPointCloud( scene, meshNode ) );
}

/**
* Merge all selected triangle meshes into a big one
**/
void PCITriangleMesh::mergeSelected()
{
	std::vector<std::string> selection = scene->getRootState()->staticState->getSelectedNodes();
	if (selection.size() < 2)
	{
		error("At least two meshes must be selected");
		return;
	}

	UnstructuredInCoreTriangleMesh* sel0 = dynamic_cast<UnstructuredInCoreTriangleMesh*>(getPointCloud(scene,selection[0]));

	const VertexDescriptor* vd = sel0->getDescr();
	const VertexDescriptor* td = sel0->getTriangleDescr();

	UnstructuredInCoreTriangleMesh* mesh = (UnstructuredInCoreTriangleMesh*)sel0->copy();;

	for (unsigned m = 1; m < selection.size(); m++ )
	{
		UnstructuredInCoreTriangleMesh* pc = dynamic_cast<UnstructuredInCoreTriangleMesh*>(getPointCloud(scene, selection[m]));
		if (pc)
		{
			if (!pc->getDescr()->equal(vd) || !pc->getTriangleDescr()->equal(td))
				warning(selection[m] + " descriptor mismatch. will be ignored");
			else
				mesh->joinOtherMesh(pc);
		}else
			warning(selection[m] + " is not a triangle mesh. will be ignored");
	}

	addSceneObject(scene, mesh, "merged");

	scene->rebuildRenderObjects();
	ObjectViewsTable::update(scene);
}

void PCITriangleMesh::pcIntersection()
{
	std::vector<std::string> selection = scene->getRootState()->staticState->getSelectedNodes();
	if (selection.size() < 1)
	{
		error("At least one point cloud must be selected");
		return;
	}

	UnstructuredInCoreTriangleMesh* mesh = getMesh();
	if ( mesh == NULL ){
		debugOutput << "invalid mesh.\n"; debugOutput.flush();
		return;
	}

	AAT flagsAAT = mesh->getAAT("flags");
	AAT posAAT = mesh->getAAT("position");

	std::vector<char> marked(mesh->getNumPoints(),0);

	for (unsigned s=0; s < selection.size(); s++)
	{
		UnstructuredInCorePointCloud* sel = dynamic_cast<UnstructuredInCorePointCloud*>(getPointCloud(scene,selection[s]));
		AAT posSelAAT = sel->getAAT("position");

		HierarchicalKNNIterator hIt(sel, 32, NULL);
		hIt.setMaxDistanceToSeekPoint(m_SampleRadius);
		for (unsigned i=0; i < mesh->getNumPoints(); i++)
		{
			if (marked[i]) continue;
			hIt.setSeekPointAndReset(mesh->getPointSet()->get3f(i,posAAT));
			if (hIt.atEnd() || (hIt.get3f(posSelAAT) - mesh->getPointSet()->get3f(i,posAAT)).getSqrNorm() > m_SampleRadius * m_SampleRadius) continue; 
			marked[i] = 1;
		}
	}
	std::vector<mpcard> todelete;
	for (unsigned i=0; i < mesh->getNumPoints(); i++)
		mesh->getPointSet()->set1i(i, flagsAAT, mesh->getPointSet()->get1i(i, flagsAAT) | (!marked[i] ? PF_FLAG_SELECTED : 0));

	scene->rebuildRenderObjects();
	ObjectViewsTable::update(scene);
}

/*** 
 * Compute vertex normals based on triangles
 */
void PCITriangleMesh::computeNormals()
{
	unsigned int m;
	std::vector<std::string> selection = scene->getRootState()->staticState->getSelectedNodes();
	PointCloud* pc;
	/// Use all point clouds in selection
	for ( m = 0; m < selection.size(); m++ ){
		pc = getPointCloud( scene, selection[m] );
		UnstructuredInCoreTriangleMesh* mesh = dynamic_cast<UnstructuredInCoreTriangleMesh*>(pc);
		if (mesh)
		{
			computeTriangleMeshNormals( mesh );
		}
	}

	scene->rebuildRenderObjects();
	ObjectViewsTable::update(scene);
}

void PCITriangleMesh::adjustTriangleIndicesToNormals() {
	std::vector<std::string> selection = scene->getRootState()->staticState->getSelectedNodes();

	for ( int i1 = 0; i1 < (int)selection.size(); ++ i1 ) {
		UnstructuredInCoreTriangleMesh* mesh = dynamic_cast<UnstructuredInCoreTriangleMesh*>( getPointCloud( scene, selection[i1] ) );
		if (!mesh)
			continue;

		X4::adjustTriangleIndicesToNormals( mesh );
	}

	scene->rebuildRenderObjects();
	ObjectViewsTable::update(scene);
}


/*** 
* Smooth out normals
*/
void PCITriangleMesh::smoothNormals()
{
	//smoothTriangleMeshNormals( getMesh(), gridSigma );
	debugOutput << "smoothNormals: not implemented yet!\n"; debugOutput.flush();
}

/** 
 * Parameterize mesh
 */
void PCITriangleMesh::parameterizeMesh()
{
	UnstructuredInCoreTriangleMesh* mesh = getMesh();
	if ( !mesh != NULL ){
		debugOutput << "parameterizeMesh: invalid mesh.\n"; debugOutput.flush();
		return;
	}

	// Check for texture coordinates attribute
	PointSet* vertices = mesh->getPointSet();
	if ( !vertices->providesAttribute( "texcoord" ) ){
		VertexDescriptor newAttribs;
		newAttribs.pushAttrib( VAD( "texcoord", 2, VAD::DATA_FORMAT_FLOAT32 ) );
		PointSet* newVertices = vertices->appendFields( &newAttribs );
		mesh->setPointSet( newVertices );
	}
	parameterizeTriangleMesh( mesh, parameterizationMode, textureRepeat, pcaROT, coordIndex0, coordIndex1 );

	// Rebuild render objects
	for ( unsigned int n = 0; n < scene->getNumGLMaterials(); n++ ){
		GLShaderMaterial* mat = dynamic_cast<GLShaderMaterial*>( scene->getGLMaterial( n ) );
		if ( mat != NULL ){
			mat->increaseVersion();
		}
	}
}

void X4::PCITriangleMesh::createGroundPlane()
{
	BoundingBox3f bb = getScenePCBBox( getScene());
	float border = bb.getDiagonalLength();
	bb.addBorder(border);
	bb.lowerCorner[upDirection] += border;
	bb.upperCorner[upDirection] -= border;
	UnstructuredInCoreTriangleMesh * mesh = new UnstructuredInCoreTriangleMesh;
	mesh->clearAndSetup(createDefaultVertexDescriptor(true,true,true,true), (m_GroundPlaneResolution+1)*(m_GroundPlaneResolution+1));
	VertexDescriptor * vdTri = new VertexDescriptor;
	vdTri->pushAttrib(VAD("index",3, VAD::DATA_FORMAT_INT32));
	mesh->getTriangles()->clearAndSetup(1,2*m_GroundPlaneResolution*m_GroundPlaneResolution, vdTri);
	VertexArray va_vert(mesh);
	VertexArray va_tri(mesh->getTriangles());

	Vector3f upVector = NULL_VECTOR3F;
	upVector[upDirection%3] = 1.0f;
	Vector3f xVec = NULL_VECTOR3F;
	xVec[(upDirection+1) % 3] = bb.getSideLength((upDirection+1) % 3) / m_GroundPlaneResolution;
	Vector3f yVec = NULL_VECTOR3F;
	yVec[(upDirection+2) % 3] = bb.getSideLength((upDirection+2) % 3) / m_GroundPlaneResolution;
	Vector3f groundPos = bb.lowerCorner;

	for( unsigned x=0;x<m_GroundPlaneResolution+1;x++)
	{
		for( unsigned y=0;y<m_GroundPlaneResolution+1;y++)
		{
			unsigned index = y*(m_GroundPlaneResolution+1)+x;
			Vector3f pos = groundPos + xVec*x + yVec*y;
			va_vert.setPosition3f(index,pos);
			va_vert.setNormal3f(index,upVector);
			va_vert.setColor3f(index,makeVector3f(1,1,1));
		}
	}

	card32 t_index = 0;
	for( unsigned x=0;x<m_GroundPlaneResolution;x++)
	{
		for( unsigned y=0;y<m_GroundPlaneResolution;y++)
		{
			unsigned index1 = y*(m_GroundPlaneResolution+1)+x;
			unsigned index2 = y*(m_GroundPlaneResolution+1)+x+1;
			unsigned index3 = (y+1)*(m_GroundPlaneResolution+1)+x;
			unsigned index4 = (y+1)*(m_GroundPlaneResolution+1)+x+1;
			Vector3i t1 = makeVector3i(index1,index2,index3);			
			Vector3i t2 = makeVector3i(index3,index2,index4);			
			va_tri.setIndex3i(t_index++,t1);
			va_tri.setIndex3i(t_index++,t2);
		}
	}

	addPointCloud(getScene(),mesh,"groundplane");
}

void X4::PCITriangleMesh::subdivide()
{
/*	UnstructuredInCoreTriangleMesh* mesh = getMesh();
	if( !mesh )
	{
		unsigned int m;
		std::vector<std::string> selection = scene->getRootState()->staticState->getSelectedNodes();
		PointCloud* pc;
		/// Use all point clouds in selection
		for ( m = 0; m < selection.size(); m++ ){
			pc = getPointCloud( scene, selection[m] );
			UnstructuredInCoreTriangleMesh* mesh = dynamic_cast<UnstructuredInCoreTriangleMesh*>(pc);
			if (mesh)
				subdivideMesh( mesh );
		}

		scene->rebuildRenderObjects();
		ObjectViewsTable::update(scene);

	}
	else
	  subdivideMesh(mesh);

*/

	std::vector<UnstructuredInCoreTriangleMesh*> meshes = getSelectedPointCloudsAs<UnstructuredInCoreTriangleMesh>(scene);
	for( int i1 = 0; i1 < (int)meshes.size(); ++ i1 )
		subdivideMesh(meshes[i1]);

	scene->rebuildRenderObjects();
	ObjectViewsTable::update(scene);
}



void X4::PCITriangleMesh::constraintSubdivision()
{
	UnstructuredInCoreTriangleMesh* mesh = getMesh();
	bool ok;
	float maxRelativeEdgeLength = QInputDialog::getDouble(NULL,"max edge length", "Specify maximal edge length relative to mesh bounding box",0.02f,0.001, 0.5f,10,&ok);
	if( !ok )
		return;

	BoundingBox3f bb = getPCBBox(mesh);
	subdivideMesh(mesh, bb.getMaxSideLength()*maxRelativeEdgeLength);
}

float PCITriangleMesh_computeTriangleScore( const Vector3f &p0, const Vector3f &p1, const Vector3f &p2)
{
	float l0 = norm(p2-p0);
	float l1 = norm(p2-p1);
	float l2 = norm(p1-p0);
	float lmin = min( l0, min(l1,l2) );
	float lmax = max( l0, max(l1,l2) );

	return lmin / lmax;
}

void X4::PCITriangleMesh::upsampleMesh()
{
	UnstructuredInCoreTriangleMesh* mesh = getMesh();
	if( !mesh )
		return;

	PointSet * ps_tri = mesh->getTriangles();
	PointSet * ps = mesh->getPointSet();

	float medPointDistance = getMedianPointDistance(mesh);
	
	bool ok;
	float sampleSpacing = QInputDialog::getDouble(NULL,"sample spacing", "Target sample spacing of new mesh",medPointDistance,0 , 1e20f,10,&ok);
	if( !ok )
		return;

	AAT posAAT = mesh->getAAT("position");
	AAT indexAAT = ps_tri->getAAT("index");

	// compute total area of mesh
	float total_area = 0;
	for( unsigned i=0;i<ps_tri->getNumEntries();i++ )
	{
		Vector3i triangle = ps_tri->get3i(i,indexAAT);
		Vector3f p0 = ps->get3f(triangle[0],posAAT);
		Vector3f p1 = ps->get3f(triangle[1],posAAT);
		Vector3f p2 = ps->get3f(triangle[2],posAAT);
		float area = norm((p1-p0).crossProduct(p2-p0)) * .5f;
		total_area += area;
	}

	unsigned approxNumVertices = (total_area) / sampleSpacing / sampleSpacing / sqrt(0.75f);
	QString str = "This will produce approximately ";
	str += QString::number( approxNumVertices );
	str += " vertices. OK?";
	QInputDialog::getText(NULL,"sample spacing",str,QLineEdit::Normal,"",&ok);
	if( !ok)
		return;

	//bool useDebugRenderer = ps_tri->getNumEntries() <= 3;

	mesh->upsample2(sampleSpacing);//, useDebugRenderer);
}


void X4::PCITriangleMesh::closeMesh()
{
	UnstructuredInCoreTriangleMesh * mesh = getMesh();
	if( mesh )
	{
		mesh->closeMesh(checkNormals);
	}

}

void X4::PCITriangleMesh::openMesh()
{
	UnstructuredInCoreTriangleMesh * mesh = getMesh();
	if( mesh )
	{
		mesh->openMesh();
	}

}

void X4::PCITriangleMesh::snapAndClose()
{
	UnstructuredInCoreTriangleMesh * mesh = getMesh();
	if( mesh )
	{
		mesh->snapMesh(maxSnapDistance, m_SnapSelectedPointsOnly);
		closeMesh();
	}

}


void X4::PCITriangleMesh::snap()
{
	UnstructuredInCoreTriangleMesh * mesh = getMesh();
	if( mesh )
	{
		mesh->snapMesh(maxSnapDistance, m_SnapSelectedPointsOnly);
	}
}


void X4::PCITriangleMesh::cleanMesh()
{
	UnstructuredInCoreTriangleMesh * mesh = getMesh();
	if( mesh )
	{
		mesh->cleanTopology();
	}	
}

void X4::PCITriangleMesh::read( InputObjectStream *s )
{
	if( s->getClassVersion(PCITriangleMesh::getClass()) < MetaClass::FIRST_VERSION + 2 )
	{
		s->readString(meshNode);
		s->read(maxSnapDistance);
		s->read(parameterizationMode);
		s->read(textureRepeat);
		if( s->getClassVersion(PCITriangleMesh::getClass()) == MetaClass::FIRST_VERSION + 1)
		{
			s->read(pcaROT);
			s->read(coordIndex0);
			s->read(coordIndex1);
			s->read(m_GroundPlaneResolution);
		}
	}
	else
		PCInteractionTool::read(s);
}

void X4::PCITriangleMesh::pointSampleAndCreateTop()
{
	UnstructuredInCoreTriangleMesh * mesh = getMesh();
	if( !mesh )
		return;

	checkAttribute(mesh,"tri_id",1,VAD::DATA_FORMAT_INT32);

	UnstructuredInCorePointCloud * upc = new UnstructuredInCorePointCloud;
	upc->clearAndSetup(mesh->getDescr(),0);
	const VertexDescriptor * sourceVertexDescr = mesh->getDescr();


	// create (oversampled) point cloud
	VertexArray va_vertex(mesh);
	VertexArray va_tri(mesh->getTriangles());
	PointSet * vertices = mesh->getPointSet();
	float oversample_radius = m_SampleRadius*0.15f;
	std::vector<char*> vbuff(3);
	std::vector<char> destPointBuffer(sourceVertexDescr->getSize());
	AAT tri_idAAT = upc->getAAT("tri_id");
	for( unsigned i=0;i<va_tri.getNumElements();i++ )
	{
		Vector3i tri = va_tri.getIndex3i(i);
		Vector3f p[3];
		for( unsigned j=0;j<3;j++ )
		{
			p[j] = va_vertex.getPosition3f(tri[j]);
		}

		Vector3f normal = (p[1]-p[0]).crossProduct(p[2] - p[0]);
		float area = norm(normal)*0.5f;
		normal = normalize(normal);

		unsigned numSamples = ceil(area / oversample_radius / oversample_radius);
		vbuff[0] = vertices->getDataPointer(tri[0]);
		vbuff[1] = vertices->getDataPointer(tri[1]);
		vbuff[2] = vertices->getDataPointer(tri[2]);
		for( unsigned j=0;j<numSamples;j++ )
		{

			float32 u = rnd01();
			float32 v = rnd01();
			if (u+v > 1) {
				float32 nu = 1.0f - v;
				v = 1.0f - u;
				u = nu;
			}

			float32 factors[3];
			factors[0] = u;
			factors[1] = v;
			factors[2] = 1.0f - u - v;

			sourceVertexDescr->interpolate((const char**)&vbuff[0], factors, 3, &destPointBuffer[0]);
			upc->getPointSet()->addPoint(&destPointBuffer[0]);
			
			// store face index
			upc->getPointSet()->set1i(upc->getNumPoints()-1,tri_idAAT,i);
		}
	}

	// resample point cloud
	PCCResampler::resamplePC(upc,m_SampleRadius);

	// setup lookup table vertex->neighbor vertex & vertex->face
	std::vector< std::vector<unsigned> > vertex_neighbors(mesh->getNumPoints());
	std::vector< std::vector<unsigned> > vertex_to_face(mesh->getNumPoints());
	for( unsigned i=0;i<va_tri.getNumElements();i++ )
	{
		Vector3i tri = va_tri.getIndex3i(i);
		for( unsigned j=0;j<3;j++ )
		{
			vertex_neighbors[tri[j]].push_back(tri[(j+1)%3]);
			vertex_neighbors[tri[j]].push_back(tri[(j+2)%3]);
			vertex_to_face[tri[j]].push_back(i);
		}
	}

	// unify vertex topology
	for( unsigned i=0;i<vertex_neighbors.size();i++ )
	{
		std::vector<unsigned> vn = vertex_neighbors[i];
		sort(vn.begin(),vn.end());
		for( unsigned j=0;j+1<vn.size(); )
		{
			if( vn[j] == vn[j+1] )
			{
				vn.erase(vn.begin()+j);
			}
			else
				j++;
		}
		vertex_neighbors[i] = vn;
	}

	// setup face topology
	/*std::vector<mpint> faceNeighbors(mesh->getTriangles()->getNumEntries()*3,-1);
	std::map<mpcard,mpcard> faceNMap;
	for( unsigned i=0;i<va_tri.getNumElements();i++ )
	{
		Vector3i tri = va_tri.getIndex3i(i);
		for( unsigned j=0;j<3;j++ )
		{
			mpcard v0 = tri[j];
			mpcard v1 = tri[(j+1)%3];
			mpcard edgeID = min(v0,v1)*mesh->getNumPoints()+max(v0,v1);
			if( faceNMap.find(edgeID) != faceNMap.end() )
			{
				mpcard neighborTri = faceNMap[edgeID];
				mpcard startIndex0 = i*3;
				mpcard startIndex1 = neighborTri*3;
				for( unsigned k=0;k<3;k++ )
				{
					if( faceNeighbors[startIndex0+k] == -1 )
					{
						faceNeighbors[startIndex0+k] = neighborTri;
						break;
					}
				}
				for( unsigned k=0;k<3;k++ )
				{
					if( faceNeighbors[startIndex1+k] == -1 )
					{
						faceNeighbors[startIndex1+k] = i;
						break;
					}
				}
			}
			else
			{
				faceNMap[edgeID] = i;
			}
		}
	}*/

	// create topology
	InCorePCTopologyGraph * graph = new InCorePCTopologyGraph;
	VertexDescriptor vd;
	vd.pushAttrib(VAD("vindex", 2, VAD::DATA_FORMAT_INT32));
	graph->clearAndSetup(&vd);
	VertexArray va_result(upc);
	HierarchicalKNNIterator hIt(upc,32,NULL);
	hIt.setMaxDistanceToSeekPoint(m_SampleRadius*2.5f);
	HierarchicalKNNIterator hIt_mesh(mesh,32,NULL);
	float maxSampleRadMesh = m_SampleRadius*4.0f;
	std::vector<unsigned> meshVertexTag(mesh->getNumPoints(),0);
	unsigned currentVertexTag = 0;
	std::vector<unsigned> faceTag(mesh->getTriangles()->getNumEntries(),0);
	unsigned currentFaceTag = 0;
	for( unsigned i=0;i<va_result.getNumElements();i++ )
	{
		// query vertices in original mesh
		unsigned myFaceID = upc->getPointSet()->get1i(i,tri_idAAT);
		Vector3i myTri = va_tri.getIndex3i(myFaceID);
		unsigned meshAnchor = myTri[0];
		std::vector<unsigned> stack; stack.reserve(10); stack.push_back(meshAnchor);
		currentVertexTag++;
		currentFaceTag++;
		meshVertexTag[meshAnchor] = currentVertexTag;

		Vector3f pos = va_result.getPosition3f(i);
		while (!stack.empty())
		{
			unsigned vindex = stack.back();
			stack.pop_back();
			
			// tag active face index
			std::vector<unsigned> &faces = vertex_to_face[vindex];
			const unsigned numFaces = (unsigned)faces.size();
			for( unsigned j=0;j<numFaces;j++ )
			{
				faceTag[faces[j]] = currentFaceTag;
			}

			// iterate
			for( unsigned j=0;j<vertex_neighbors[vindex].size();j++ )
			{
				unsigned nindex = vertex_neighbors[vindex][j];
				if( meshVertexTag[nindex] != currentVertexTag && norm(pos - va_vertex.getPosition3f(nindex)) < maxSampleRadMesh)
				{
					meshVertexTag[nindex] = currentVertexTag;
					stack.push_back(nindex);
				}
			}
		}
		
		// query all points in sphere
		hIt.setSeekPointAndReset(va_result.getPosition3f(i));
		hIt.next();
		static int test = 0;
		while(!hIt.atEnd())
		{
			unsigned fIndex	= hIt.get1i(tri_idAAT);
			if( faceTag[fIndex] != currentFaceTag )
			{
				hIt.next();
				continue;
			}

			unsigned nIndex = hIt.getCurrentPointIndex();
			graph->addEdge(i,nIndex);

			hIt.next();
		}
	}

	TopologyAttachment * topAtt = new TopologyAttachment;
	topAtt->setup(TopologyAttachment::getDefaultName(),AttachedData::ADF_PERSISTENT);
	topAtt->setTopology(graph);
	upc->getAttachments()->attachData(topAtt);

	addPointCloud(getScene(),upc,"sampled");
}

void X4::PCITriangleMesh::subdivideMesh( UnstructuredInCoreTriangleMesh* mesh )
{
	UnstructuredInCoreTriangleMesh * newMesh = (UnstructuredInCoreTriangleMesh*)mesh->copy();

	PointSet * ps = mesh->getPointSet();
	PointSet * ps_tri = mesh->getTriangles();
	PointSet * ps_new = newMesh->getPointSet();
	PointSet * ps_newtri = newMesh->getTriangles();

	AAT posAAT = mesh->getAAT("position");
	AAT normalAAT = mesh->getAAT("normal");
	AAT colorAAT = mesh->getAAT("color");
	AAT indexAAT = ps_tri->getAAT("index");

	std::vector<char> vbuffer(mesh->getDescr()->getSize());
	char * vptr = &vbuffer[0];
	std::map<card64,unsigned> edgeMap;
	for( card64 i=0;i<ps_tri->getNumEntries();i++ )
	{
		Vector3i triangle = ps_tri->get3i(i,indexAAT);
		unsigned newVIndex[3];
		for( card64 j=0;j<3;j++ )
		{
			unsigned i1 = triangle[j];
			unsigned i2 = triangle[(j+1)%3];
			card64 newVertexID = min(i1,i2)*ps->getNumEntries()+ max(i1,i2);
			if( edgeMap.find(newVertexID) != edgeMap.end() )
			{
				newVIndex[j] = edgeMap[newVertexID];
			}
			else
			{
				newVIndex[j] = newMesh->getNumPoints();
				edgeMap[newVertexID] = newVIndex[j];
				const VertexDescriptor * vd = mesh->getDescr();
				char * ptr[2];
				ptr[0] = ps->getDataPointer(i1);
				ptr[1] = ps->getDataPointer(i2);
				float32 interp[] = {0.5f,0.5f};
				vd->interpolate( (const char**)ptr, interp,2,vptr);
				ps_new->addPoint(vptr);
				ps_new->set3f(newVIndex[j],normalAAT, normalize(ps_new->get3f(newVIndex[j],normalAAT)));
			}
		}
		Vector3i newTri1 = makeVector3i(triangle[0], newVIndex[0],newVIndex[2]);
		Vector3i newTri2 = makeVector3i(triangle[1], newVIndex[1],newVIndex[0]);
		Vector3i newTri3 = makeVector3i(triangle[2], newVIndex[2],newVIndex[1]);
		Vector3i newTri4 = makeVector3i(newVIndex[0], newVIndex[1],newVIndex[2]);
		ps_newtri->set3i(i,indexAAT,newTri1);
		ps_newtri->addPoint((char*)newTri2.data());
		ps_newtri->addPoint((char*)newTri3.data());
		ps_newtri->addPoint((char*)newTri4.data());
	}

	mesh->assign(newMesh);
	delete newMesh;
}


struct PCITriangleMesh_TPolygonalMesh
{
	std::vector< SimplePolygon > polygons;
};

/**
convert triangle mesh to polygonal mesh, tries to convert triangle pairs into quads but nothing more than that...
*/
void PCITriangleMesh_convertToPolygonalMesh(UnstructuredInCoreTriangleMesh * mesh, PCITriangleMesh_TPolygonalMesh &ret)
{
	VertexArray va_tri(mesh->getTriangles());
	VertexArray va_points(mesh);

	std::map<card64,mpcard> edgeToTriangle;
	std::vector<Vector3f> faceNormals(mesh->getNumTriangles());
	std::vector<card32> faceMarkedAsQuad(mesh->getNumTriangles(),0);
	for( mpcard i=0;i<va_tri.getNumElements();i++ )
	{
		Vector3i tri = va_tri.getIndex3i(i);
		Vector3f p[3];
		for( mpcard j=0;j<3;j++ )
			p[j] = va_points.getPosition3f(tri[j]);
		faceNormals[i] = normalize((p[1]-p[0]).crossProduct(p[2]-p[0]));

		for( mpcard j=0;j<3;j++)
		{
			card64 i0 = tri[(j+1)%3];
			card64 i1 = tri[j];
			card64 edgeID = min(i0,i1)*mesh->getNumPoints()+max(i0,i1);
			if( edgeToTriangle.find(edgeID)!=edgeToTriangle.end() )
			{
				mpcard triID1 = edgeToTriangle[edgeID];
				if( faceMarkedAsQuad[triID1] == 0)
					if( faceNormals[i]*faceNormals[triID1] >= 0.99f )
					{
						// merge triangles
						faceMarkedAsQuad[triID1] = 1;
						faceMarkedAsQuad[i] = 1;
						SimplePolygon poly;
						poly.addVertex(PolyVector3f::create(va_points.getPosition3f(tri[(j+1)%3])));
						poly.addVertex(PolyVector3f::create(va_points.getPosition3f(tri[(j+2)%3])));
						poly.addVertex(PolyVector3f::create(va_points.getPosition3f(tri[j])));

						Vector3i tri1 = va_tri.getIndex3i(triID1);
						for( mpcard k=0;k<3;k++ )
							if( tri1[k] != tri[j] && tri1[k] !=tri[(j+1)%3])
								poly.addVertex(PolyVector3f::create(va_points.getPosition3f(tri1[k])));

						//debugRenderPolyLine(poly,makeVector3f(1,0.5f,0));
						ret.polygons.push_back(poly);
						edgeToTriangle.erase(edgeID);

						break;
					}
			}
			else
			{
				edgeToTriangle[edgeID] = i;
			}
		}
	}

	// add all remaining unmerged triangles
	for( mpcard i=0;i<va_tri.getNumElements();i++ )
	{
		if( faceMarkedAsQuad[i])
			continue;
		Vector3i tri = va_tri.getIndex3i(i);
		SimplePolygon poly;
		poly.addVertex(PolyVector3f::create(va_points.getPosition3f(tri[0])));
		poly.addVertex(PolyVector3f::create(va_points.getPosition3f(tri[1])));
		poly.addVertex(PolyVector3f::create(va_points.getPosition3f(tri[2])));
		ret.polygons.push_back(poly);
	}
}

void PCITriangleMesh_addpolyToMesh( UnstructuredInCoreTriangleMesh* mesh, const SimplePolygon &poly)
{

}

template<class PolyType>
float PCITriangleMesh_computeMaximalPolyLength(const PolyType &poly)
{
	const mpcard numPoints = poly.getNumVertices();
	float ret = 0;
	for( mpcard i=0;i<numPoints;i++ )
	{
		for( mpcard j=i+1;j<numPoints;j++ )
		{
			ret = max(ret, normQuad(poly[i].getPos3f() - poly[j].getPos3f()));
		}
	}
	return sqrt(ret);
}

void X4::PCITriangleMesh::subdivideMesh( UnstructuredInCoreTriangleMesh* mesh, float maxEdgeLength )
{
	PCITriangleMesh_TPolygonalMesh polyMesh;
	PCITriangleMesh_convertToPolygonalMesh(mesh, polyMesh);

	mesh->getPointSet()->clear();
	mesh->getTriangles()->clear();

	for( mpcard i=0;i<polyMesh.polygons.size();i++ )
	{
		SimplePolygon poly = polyMesh.polygons[i];

		// check if we have to subdivide this polygon
		float maxL = PCITriangleMesh_computeMaximalPolyLength(poly);
		if( maxL <= maxEdgeLength)
		{
			PCITriangleMesh_addpolyToMesh(mesh, poly);
		}

		// subdivide
		std::vector<SimplePolygon> subdivided;
		if( poly.getNumVertices() == 3)
		{

		}
	}
}

void X4::PCITriangleMesh::computeCoTanWeights()
{
	unsigned int m;
	std::vector<std::string> selection = scene->getRootState()->staticState->getSelectedNodes();
	PointCloud* pc;
	/// Use all point clouds in selection
	for ( m = 0; m < selection.size(); m++ ){
		pc = getPointCloud( scene, selection[m] );
		UnstructuredInCoreTriangleMesh* mesh = dynamic_cast<UnstructuredInCoreTriangleMesh*>(pc);
		if (mesh)
		{
			 mesh->computeTopologyAttachment(true);
		}
	}
}

void X4::PCITriangleMesh::tryToRepairTopology()
{
	UnstructuredInCoreTriangleMesh* mesh = getMesh();
	if ( !mesh != NULL ){
		debugOutput << "no mesh.\n"; debugOutput.flush();
		return;
	}


	// search for zero-area triangles
	std::vector<mpcard> deleteTri;
	VertexArray vaTri(mesh->getTriangles());
	VertexArray points(mesh);
	for( mpcard i=0;i<vaTri.getNumElements();i++ )
	{
		Vector3i tri = vaTri.getIndex3i(i);
		Vector3f p[3];
		for( mpcard d=0;d<3;d++)
			p[d] = points.getPosition3f(tri[d]);

		// pick longest dir
		mpcard md=0;
		float l=0;
		for( mpcard j=0;j<3;j++ )
		{
			float tmp = norm(p[j] - p[(j+1)%3]);
			if( tmp > l)
			{
				md = j;
				l = tmp;
			}
		}

	
		if( l < 1e-10f)
		{
			deleteTri.push_back(i);
			continue;
		}

		Vector3f dir = normalize(p[md] - p[(md+1)%3]);
		Vector3f p3 = p[(md+2)%3];
		float projDist = norm(p[md] + dir*(dir*(p3-p[md])) - p3);
		if( projDist < maxSnapDistance )
			deleteTri.push_back(i);


// 		
// 		Vector3f dir = p[1]-p[0];
// 		float l = norm(dir);
// 		if( l < 1e-10f)
// 		{
// 			deleteTri.push_back(i);
// 			continue;
// 		}
// 
// 		Vector3f dir2 = p[2]-p[0];
// 		float l2 = norm(dir2);
// 		if( l2 < 1e-10f)
// 		{
// 			deleteTri.push_back(i);
// 			continue;
// 		}

// 		if( fabs(dir*dir2/l/l2) > 0.9999f )
// 			deleteTri.push_back(i);

	}

	while (!deleteTri.empty())
	{
		mpcard i = deleteTri.back();
		deleteTri.pop_back();

		mesh->getTriangles()->deletePoint(i);
	}

	mesh->clearAttachments();

}

void X4::PCITriangleMesh::mouseDown( int32 x, int32 y, MouseButtons buttonsState, Modifiers modifiersState )
{
	if( !m_DeleteTriangleMode)
		return;

	if( buttonsState.getLeft())
	{
		Vector3f clickPoint;
		if( guessClickPoint(x,y,5,clickPoint))
		{
			UnstructuredInCoreTriangleMesh * mesh = getMesh();
			if( mesh != m_Temp_Mesh )
			{
				m_Temp_Mesh = mesh;
				delete m_Temp_KDtree;
				m_Temp_KDtree = new MajorKDTree;
				m_Temp_DeleteCache = std::vector<mpint>(mesh->getNumTriangles());
				m_Temp_OrgIndexCache = std::vector<mpcard>(mesh->getNumTriangles());
				VertexArray vaTri(mesh->getTriangles());
				VertexArray vaPoints(mesh);
				for( mpcard i=0;i<vaTri.getNumElements();i++ )
				{
					m_Temp_DeleteCache[i] = i;
					m_Temp_OrgIndexCache[i] = i;
					Vector3i tri = vaTri.getIndex3i(i);
					BoundingBox3f bb(vaPoints.getPosition3f(tri[0]));
					for( mpcard j=1;j<3;j++ )
						bb.addPoint(vaPoints.getPosition3f(tri[j]));
					bb.addBorder(1e-7f);
					m_Temp_KDtree->addElement(bb);
				}
				m_Temp_KDtree->setupTree(16);
			}

			VertexArray vaTri(mesh->getTriangles());
			VertexArray vaPoints(mesh);
			bool hasPolyInfo = mesh->getTriangles()->providesAttribute("polyIndex");
			AAT polyInfoAAT;
			if( hasPolyInfo)
				polyInfoAAT = mesh->getTriangles()->getAAT("polyIndex");
			BoundingBox3f meshBB = getPCBBox(mesh);
			BoundingBox3f queryBB(clickPoint);
			queryBB.addBorder(meshBB.getDiagonalLength()*1e-4f);
			std::vector<mpcard> possibleCol = m_Temp_KDtree->computePossibleCollision(queryBB,IDENTITY4F);
			Ray3f ray;
			ray.origin = getScene()->getRootState()->staticState->camera.getPosition();
			ray.direction = clickPoint-ray.origin;
			mpint bestHit=-1;
			float bestDist = 1e20f;
			for( mpcard i=0;i<possibleCol.size();i++ )
			{
				mpcard index = possibleCol[i];
				if( m_Temp_DeleteCache[index] < 0 )
					continue;

				Vector3i tri = vaTri.getIndex3i(m_Temp_DeleteCache[index]);
				Vector3f p[3];
				for( mpcard j=0;j<3;j++)
					p[j] = vaPoints.getPosition3f(tri[j]);

				float dist;
				if( intersectTriangle3f(ray,p[0],p[1],p[2],dist))
				{
					if( dist < bestDist)
					{
						bestDist = dist;
						bestHit = index;
					}
				}
			}

			if( bestHit >= 0)
			{
				std::vector<mpcard> toDelete;
				if( m_DeleteAllTrianglesWithinPlane )
				{
					
					mpcard actualIndex = m_Temp_DeleteCache[bestHit];
					Vector3i tri = vaTri.getIndex3i(actualIndex);
					Vector3f normal0 = normalize( (vaPoints.getPosition3f(tri[1])-vaPoints.getPosition3f(tri[0]))
						.crossProduct(vaPoints.getPosition3f(tri[2])-vaPoints.getPosition3f(tri[0])));
					Vector3f p0 = vaPoints.getPosition3f(tri[0]);
					Plane plane0(normal0,p0);

					for( mpcard j=0;j<vaTri.getNumElements();j++ )
					{
						Vector3i tri = vaTri.getIndex3i(m_Temp_OrgIndexCache[j]);
						Vector3f normal1 = normalize( (vaPoints.getPosition3f(tri[1])-vaPoints.getPosition3f(tri[0]))
							.crossProduct(vaPoints.getPosition3f(tri[2])-vaPoints.getPosition3f(tri[0])));
						Vector3f p1 = vaPoints.getPosition3f(tri[0]);

						if( checkNormals && normal0*normal1 < 0.95f )
							continue;

						if( fabs(normal0*normal1) > 0.99f && fabs(plane0.calculateSignedDistance(p1))<0.0001 * meshBB.getDiagonalLength())
							toDelete.push_back(m_Temp_OrgIndexCache[j]);
					}
				}
				else
				if( hasPolyInfo )
				{
					mpcard actualIndex = m_Temp_DeleteCache[bestHit];
					int32 polyID = vaTri.get<int32,1>(actualIndex,polyInfoAAT)[0];
					for( mpcard j=0;j<vaTri.getNumElements();j++ )
					{
						if( vaTri.get<int32,1>(j,polyInfoAAT)[0] == polyID )
						{
							toDelete.push_back(m_Temp_OrgIndexCache[j]);
						}
					}

				}
				else
					toDelete.push_back(bestHit);


				for( mpcard i=0;i<toDelete.size();i++ )
				{
					mpcard toDeleteIndex = toDelete[i];
					mpcard actualIndex = m_Temp_DeleteCache[toDeleteIndex];
					mesh->getTriangles()->deletePoint(actualIndex);
					m_Temp_DeleteCache[m_Temp_OrgIndexCache.back()] = actualIndex;
					m_Temp_OrgIndexCache[actualIndex] = m_Temp_OrgIndexCache.back();
					m_Temp_OrgIndexCache.pop_back();
					m_Temp_DeleteCache[toDeleteIndex] = -1;
				}
			}
		}
	}
}

