//---------------------------------------------------------------------------
#ifndef RegularDataProcessingH
#define RegularDataProcessingH
//---------------------------------------------------------------------------
#include "x3editor2.h"
//---------------------------------------------------------------------------
#include "DynamicLinearAlgebra.h"
#include "RegularData.h"
//---------------------------------------------------------------------------

namespace X4 {

class PointSet;


//  --  basic stuff



/// concatenate data sets along last (dimension-1) axis. all other dimensions must match.
template <typename FloatType>
RegularData<FloatType> concatenate(
   const RegularData<FloatType> &one,
   const RegularData<FloatType> &two
);



//  --  conversion from point sets



/// creates a RegularData object from a point set.
/// multiple channels are represented as first dimension, width and height are second and third.
/// some extends might be one (not optimized to avoid complicated conditionals in processing code)
template<typename FloatType>
void createRegDataFromPointSet(
   const PointSet *ps,
   std::string channel,
   RegularData<FloatType> &target
);


/// the same as the first variant of createRegDataFromPointSet, but read only one component of the given channel
template<typename FloatType>
void createRegDataFromPointSet(
   const PointSet *ps,
   std::string channel,
   int32 componentNumber,
   RegularData<FloatType> &target
);



//  --  conversion to point sets



/// write data (back) to a point set channel.
/// the dimension of the channel must be no smaller than the innermost dimension of the data object (ext[0] <= channel.numAttributes)
/// width and height of the point set must be larger than the second and third extend of thd data object
/// the data object must have dimension 3
template<typename FloatType>
void writeRegDataToPointSet(
   PointSet *ps,
   std::string channel,
   const RegularData<FloatType> &data
);


/// the same as the first variant of writeRegDataToPointSet, but read only one component of the given channel
template<typename FloatType>
void writeRegDataToPointSet(
   PointSet *ps,
   std::string channel,
   int32 componentNumber,
   const RegularData<FloatType> &data,
   card32 xOffset = 0,
   card32 yOffset = 0
);



//  --  fourier transform



/// computes fourier transform of an image. If inputImg == NULL, a constant zero imaginary part is assumed
template<typename FloatType>
void fourierTransform(
   const RegularData<FloatType> *inputReal,
   const RegularData<FloatType> *inputImg,
   RegularData<FloatType> &outputReal,
   RegularData<FloatType> &outputImg,
   bool invTransform = false
);


//  --  geometric transformation


template <typename FloatType>
struct AffineTransform {
   /// center - translate to the center before transforming, then translate back
   DynamicVector<FloatType> center;
   /// linear transform
   DynamicMatrix<FloatType> transform;

   const DynamicVector<FloatType> doTransform(const DynamicVector<FloatType> &input) const {
      return transform*(input - center) + center;
   }
   void invertTransform() {
      transform = invertMatrix(transform);
   }
};


template <typename FloatType>
void affineTransform(
   const RegularData<FloatType> &source,
   RegularData<FloatType> &dest,
   const AffineTransform<FloatType> &transf
);


//  --  Gaussian parameter estimation, PCA-Alignment


template <typename FloatType>
void computeCenterOfMass(
   const RegularData<FloatType> &source,
   DynamicVector<FloatType> &result
);


template <typename FloatType>
void computeMassCovariance(
   const RegularData<FloatType> &source,
   const DynamicVector<FloatType> &mean,
   DynamicMatrix<FloatType> &result
);


/// computes coordinate system that aligns patch to PCA main axes.
/// the new coordinate system is guaranteed to have a positive determinant (if not, the 0th axis [x-axis] is swapped).
template <typename FloatType>
void computePCACoordinateSystem(
   const RegularData<FloatType> &source,
   DynamicVector<FloatType> &mean,
   DynamicMatrix<FloatType> &rot
);


/// set all values outside a sphere around center to zero
/// uses minimum extend as diameter of clipping sphere
template <typename FloatType>
void sphereClip(RegularData<FloatType> &data);


/// performs a PCA transform for the vectors of the innermost dimension (extends[0])
/// all other dimensions are treated as list of independent examples
/// the result is returned in place
/// optionally (pass non-null pointer) the eigenvalues, eigenvectors and the mean std::vector are returned
template <typename FloatType>
void computePCATransform(
	RegularData<FloatType> &data, 
	DynamicVector<FloatType> *eigenvalues = NULL, 
	DynamicVector<FloatType> *mean = NULL,
	DynamicMatrix<FloatType> *eigenvects = NULL,
	bool normalize = false
);


}

#endif
