//---------------------------------------------------------------------------
#ifndef RegularDataProcessing_templCodeH
#define RegularDataProcessing_templCodeH
//---------------------------------------------------------------------------
#include "x3editor2.h"
//---------------------------------------------------------------------------
#include "RegularDataProcessing.h"
//---------------------------------------------------------------------------
#include "DynamicLinearAlgebra.inline.h"
#include "DynamicLinearAlgebraTools.h"
#include "RegularData.inline.h"
#include "PointSet.h"
#include "PointSet.inline.h"
#include "MathTools.h"
#include "MathTools.inline.h"
#include "DynamicLinearAlgebraTools.h"
#include "fftw3.h"
//---------------------------------------------------------------------------

namespace X4 {


// ------------- basic stuff


template <typename FloatType>
void concatenate(const RegularData<FloatType> &one, const RegularData<FloatType> &two, RegularData<FloatType> &result) {
   x4Assert(one.getDim() == two.getDim());
   x4Assert(one.getDim() >= 1);
   card32 dim = one.getDim();
   for (card32 i=0; i<dim-1; i++) {
      x4Assert(one.getExtends(i) == two.getExtends(i));
   }

   DVectorI ext = one.getExtends();
   ext[dim-1] = one.getExtends(dim-1) + two.getExtends(dim-1);

   result.setExtends(ext);
   card32 num1 = one.getNumElements();
   card32 num2 = two.getNumElements();
   const FloatType *inOne = one.dataPtr();
   const FloatType *inTwo = two.dataPtr();
   FloatType *out = result.dataPtr();
   for (card32 i=0; i<num1; i++) {
      *out++ = *inOne++;
   }
   for (card32 i=0; i<num2; i++) {
      *out++ = *inTwo++;
   }
}


// ------------- point set stuff


template<typename FloatType>
void createRegDataFromPointSet(const PointSet *ps, std::string channel, RegularData<FloatType> &target) {
   x4Assert(ps->providesAttribute(channel));
   VAD channelVAD = ps->getDescr()->getAttribute(channel);
   AAT channelAAT = ps->getAAT(channel);
   card32 offset = channelAAT.getOffset();
   DynamicVector<int32> ext(3);
   ext[0] = channelVAD.numComponents;
   ext[1] = ps->getWidth();
   ext[2] = ps->getHeight();

   target.setExtends(ext);

   card32 numElements = ps->getNumEntries();
   card32 numAttribs = channelVAD.numComponents;
   card32 index = 0;
   for (card32 i=0; i<numElements; i++) {
      for (card32 a=0; a<numAttribs; a++) {
         FloatType f = 0;
         const char *dataPointer = ps->getDataPointer(i);
         switch (channelVAD.dataFormat) {
            case VAD::DATA_FORMAT_CARD8: {
               f = (FloatType)*( ((card8*)(dataPointer+offset)) + a );
               break;
            }
            case VAD::DATA_FORMAT_FLOAT32: {
               f = (FloatType)*( ((float32*)(dataPointer+offset)) + a );
               break;
            }
            case VAD::DATA_FORMAT_INT32: {
               f = (FloatType)*( ((int32*)(dataPointer+offset)) + a );
               break;
            }
         }
         target[index] = f;
         index++;
      }
   }
}

template<typename FloatType>
void createRegDataFromPointSet(const PointSet *ps, std::string channel, int32 componentNumber, RegularData<FloatType> &target) {
   x4Assert(ps->providesAttribute(channel));
   VAD channelVAD = ps->getDescr()->getAttribute(channel);
   AAT channelAAT = ps->getAAT(channel);
   card32 offset = channelAAT.getOffset();
   DynamicVector<int32> ext(2);
   ext[0] = ps->getWidth();
   ext[1] = ps->getHeight();

   target.setExtends(ext);

   card32 numElements = ps->getNumEntries();
   card32 index = 0;
   for (card32 i=0; i<numElements; i++) {
      FloatType f = 0;
      const char *dataPointer = ps->getDataPointer(i);
      switch (channelVAD.dataFormat) {
         case VAD::DATA_FORMAT_CARD8: {
            f = (FloatType)*( ((card8*)(dataPointer+offset)) + componentNumber );
            break;
         }
         case VAD::DATA_FORMAT_FLOAT32: {
            f = (FloatType)*( ((float32*)(dataPointer+offset)) + componentNumber );
            break;
         }
         case VAD::DATA_FORMAT_INT32: {
            f = (FloatType)*( ((int32*)(dataPointer+offset)) + componentNumber );
            break;
         }
      }
      target[index] = f;
      index++;
   }
}


template<typename FloatType>
void writeRegDataToPointSet(PointSet *ps, std::string channel, const RegularData<FloatType> &data) {
   x4Assert(ps->providesAttribute(channel));
   VAD channelVAD = ps->getDescr()->getAttribute(channel);
   AAT channelAAT = ps->getAAT(channel);
   card32 offset = channelAAT.getOffset();

   DynamicVector<int32> extends = data.getExtends();

   x4Assert((int32)channelVAD.numComponents >= extends[0]);
   x4Assert((int32)ps->getWidth() >= extends[1]);
   x4Assert((int32)ps->getHeight() >= extends[2]);

   card32 index = 0;
   for (int32 y=0; y<extends[2]; y++) {
      for (int32 x=0; x<extends[1]; x++) {
         for (int32 a=0; a<extends[0]; a++) {
            char *dataPointer = ps->getDataPointer(x, y);
            switch (channelVAD.dataFormat) {
               case VAD::DATA_FORMAT_CARD8: {
                  *( ((card8*)(dataPointer+offset)) + a ) = (card8)data[index];
                  break;
               }
               case VAD::DATA_FORMAT_FLOAT32: {
                  *( ((float32*)(dataPointer+offset)) + a ) = (float32)data[index];
                  break;
               }
               case VAD::DATA_FORMAT_INT32: {
                  *( ((int32*)(dataPointer+offset)) + a ) = (int32)data[index];
                  break;
               }
            }
            index++;

         }
      }
   }
}

template<typename FloatType>
void writeRegDataToPointSet(PointSet *ps, std::string channel, int32 componentNumber,
                            const RegularData<FloatType> &data, card32 xOffset, card32 yOffset)
{
   x4Assert(ps->providesAttribute(channel));
   VAD channelVAD = ps->getDescr()->getAttribute(channel);
   AAT channelAAT = ps->getAAT(channel);
   card32 offset = channelAAT.getOffset();

   DynamicVector<int32> extends = data.getExtends();

   x4Assert((int32)channelVAD.numComponents >= componentNumber);
   x4Assert((int32)ps->getWidth()-(int32)xOffset >= extends[0]);
   // no problem to allow for the height to exceed target...
   // x4Assert((int32)ps->getHeight() >= extends[1]);
   int32 targetHeight = min((int32)ps->getHeight()-(int32)yOffset, extends[1]);

   card32 index = 0;
   for (int32 y=0; y<targetHeight; y++) {
      for (int32 x=0; x<extends[0]; x++) {
         char *dataPointer = ps->getDataPointer(x+xOffset, y+yOffset);
         switch (channelVAD.dataFormat) {
            case VAD::DATA_FORMAT_CARD8: {
               *( ((card8*)(dataPointer+offset)) + componentNumber ) = (card8)data[index];
               break;
            }
            case VAD::DATA_FORMAT_FLOAT32: {
               *( ((float32*)(dataPointer+offset)) + componentNumber ) = (float32)data[index];
               break;
            }
            case VAD::DATA_FORMAT_INT32: {
               *( ((int32*)(dataPointer+offset)) + componentNumber ) = (int32)data[index];
               break;
            }
         }
         index++;
      }
   }
}


// ------------- fourier stuff


template<typename FloatType>
void fourierTransform(const RegularData<FloatType> *inputReal,
                      const RegularData<FloatType> *inputImg,
                      RegularData<FloatType> &outputReal,
                      RegularData<FloatType> &outputImg,
                      bool invTransform)
{
   x4Assert(inputReal != NULL);
   const DVectorI ext = inputReal->getExtends();
   DVectorI extReverse(ext.size());
   for (card32 i=0; i<ext.size(); i++) {
      extReverse[ext.size() - 1 - i] = ext[i];
   }
   const int32 n = inputReal->getNumElements();

   fftwf_complex *in = (fftwf_complex*)fftwf_malloc(sizeof(fftwf_complex) * n);;
   fftwf_complex *out = (fftwf_complex*)fftwf_malloc(sizeof(fftwf_complex) * n);

   const FloatType *dataPtr = inputReal->dataPtr();
   fftwf_complex *inPos = in;
   if (inputImg == NULL) {
      for (int i=0; i<n; i++) {
         (*inPos)[0] = (float)*dataPtr++;
         (*inPos)[1] = 0;
         inPos++;
      }
   } else {
      x4Assert(inputImg->getExtends() == inputReal->getExtends());
      const FloatType *dataPtrImg = inputImg->dataPtr();
      for (int i=0; i<n; i++) {
         (*inPos)[0] = (float)*dataPtr++;
         (*inPos)[1] = (float)*dataPtrImg++;
         inPos++;
      }
   }                                      

   fftwf_plan plan = fftwf_plan_dft((int) extReverse.size(), extReverse.data(), in, out,
                                     invTransform?FFTW_BACKWARD:FFTW_FORWARD, FFTW_ESTIMATE );
   fftwf_execute(plan);

   outputReal.setExtends(ext);
   outputImg.setExtends(ext);

   FloatType *dataPtrReal = outputReal.dataPtr();
   FloatType *dataPtrImg  = outputImg.dataPtr();
   fftwf_complex *outPos = out;
   for (int i=0; i<n; i++) {
      *dataPtrReal++ = (FloatType)((*outPos)[0]);
      *dataPtrImg++  = (FloatType)((*outPos)[1]);
      outPos++;
   }
   fftwf_destroy_plan(plan);
   fftwf_free(in);
   fftwf_free(out);
}


// ------------- geometric transformations


template<typename FloatType>
void affineTransformInternRek(
   const RegularData<FloatType> &source, RegularData<FloatType> &dest,
   const AffineTransform<FloatType> &invTransf,
   DVectorI &pos, int32 procDim
) {
   if (procDim < 0) return;
   if (procDim <= 2) {
      int width  =                source.getExtends(0);
      int height = procDim >= 1 ? source.getExtends(1) : 1;
      int depth  = procDim >= 2 ? source.getExtends(2) : 1;

      DVectorF floatPos(pos.getDim()); convertDVectorType(pos, floatPos);
      DVectorF sourcePos = invTransf.doTransform(floatPos);
      DVectorF saveSourcePosY;
      DVectorF saveSourcePosX;
      DVectorI convertSourcePos(source.getDim());
      DVectorI destExtends = dest.getExtends();

      FloatType *destDataPtr = dest.dataPtr();
      destDataPtr += dest.calcIndex(pos);

      for (int z=0; z<depth; z++) {
         saveSourcePosY = sourcePos;
         for (int y=0; y<height; y++) {
            saveSourcePosX = sourcePos;
            for (int x=0; x<width; x++) {
               convertDVectorType(sourcePos, convertSourcePos);
               FloatType v = 0;
               if (convertSourcePos.minComponent() >= 0 && convertSourcePos.allComponentsSmaller(destExtends)) {
                  v = source.getElement(convertSourcePos);
               }
               *destDataPtr++ = v;
               sourcePos += invTransf.transform[0];
            }
            sourcePos = saveSourcePosX;
            if (procDim >= 1) sourcePos += invTransf.transform[1];
         }
         sourcePos = saveSourcePosY;
         if (procDim >= 2) sourcePos += invTransf.transform[2];
      }
   } else {
      card32 extDim = source.getExtends(procDim);
      for (card32 i=0; i<extDim; i++) {
         pos[i] = i;
         affineTransformInternRek(source, dest, invTransf, pos, procDim-1);
      }
   }
}

template<typename FloatType>
void affineTransform(const RegularData<FloatType> &source, RegularData<FloatType> &dest, const AffineTransform<FloatType> &invTransf)
{
   card32 dim = source.getDim();
   x4Assert(dim >= 1);
   dest.setExtends(source.getExtends());
   DVectorI pos = nullDVector<int32>(dim);
   affineTransformInternRek(source, dest, invTransf, pos, dim-1);
}


template<typename FloatType>
void computeCenterOfMassInternRek(
   const RegularData<FloatType> &source,
   DynamicVector<FloatType> &result,
   DVectorI &pos,
   FloatType &normalize,
   int32 procDim)
{
   DynamicVector<FloatType> posConv(result.getDim());

   x4Assert(procDim >= 0);
   if (procDim <= 2) {
      int width  =                source.getExtends(0);
      int height = procDim >= 1 ? source.getExtends(1) : 1;
      int depth  = procDim >= 2 ? source.getExtends(2) : 1;

      for (int z=0; z<depth; z++) {
         if (procDim >=2) pos[2] = z;
         for (int y=0; y<height; y++) {
            if (procDim >=1) pos[1] = y;
            for (int x=0; x<width; x++) {
               pos[0] = x;
               convertDVectorType(pos, posConv);
               FloatType v = source.getElement(pos);
               posConv *= v;
               normalize += v;
               result += posConv;
            }
         }
      }
   } else {
      card32 extDim = source.getExtends(procDim);
      for (card32 i=0; i<extDim; i++) {
         pos[i] = i;
         computeCenterOfMassInternRek(source, result, pos, normalize, procDim-1);
      }
   }
}

template<typename FloatType>
void computeCenterOfMass(const RegularData<FloatType> &source, DynamicVector<FloatType> &result)
{
   card32 dim = source.getDim();
   x4Assert(dim >= 1);
   result = nullDVector<FloatType>(dim);
   DVectorI pos = nullDVector<int32>(dim);
   FloatType normalize = (FloatType)0;
   computeCenterOfMassInternRek(source, result, pos, normalize, dim-1);
   if (normalize >= 1E-7) {
      result *= ((FloatType)1)/normalize;
   } else {
      DVectorI ext = source.getExtends();
      ext /= 2;
      convertDVectorType(ext, result);
   }
}


template<typename FloatType>
void computeMassCovarianceInternRek(
   const RegularData<FloatType> &source,
   const DVectorF &mean,
   DynamicMatrix<FloatType> &result,
   DVectorI &pos,
   FloatType &normalize,
   int32 procDim)
{
   DynamicVector<FloatType> posConv(pos.getDim());

   x4Assert(procDim >= 0);
   if (procDim <= 2) {
      int width  =                source.getExtends(0);
      int height = procDim >= 1 ? source.getExtends(1) : 1;
      int depth  = procDim >= 2 ? source.getExtends(2) : 1;
      int dim = source.getDim();


      for (int z=0; z<depth; z++) {
         if (procDim >=2) pos[2] = z;
         for (int y=0; y<height; y++) {
            if (procDim >=1) pos[1] = y;
            for (int x=0; x<width; x++) {
               pos[0] = x;
               FloatType v = source.getElement(pos);

               for (int my = 0; my<dim; my++) {
                  for (int mx = 0; mx<dim; mx++) {
                     result[mx][my] += v * ((FloatType)pos[mx] - mean[mx]) * ((FloatType)pos[my] - mean[my]);
                  }
               }

               normalize += v;
            }
         }
      }
   } else {
      card32 extDim = source.getExtends(procDim);
      for (card32 i=0; i<extDim; i++) {
         pos[i] = i;
         computeMassCovarianceInternRek(source, mean, result, pos, normalize, procDim-1);
      }
   }
}


template<typename FloatType>
void computeMassCovariance(const RegularData<FloatType> &source, const DynamicVector<FloatType> &mean,
                           DynamicMatrix<FloatType> &result)
{
   card32 dim = source.getDim();
   x4Assert(dim >= 1 && mean.getDim() == source.getDim());
   result = nullDMatrix<FloatType>(dim, dim);
   DVectorI pos = nullDVector<int32>(dim);
   FloatType normalize = (FloatType)0;
   computeMassCovarianceInternRek(source, mean, result, pos, normalize, dim-1);
   if (normalize >= 1E-7) {
      result *= ((FloatType)1)/normalize;
   }
}


template <typename FloatType>
void computePCACoordinateSystem(const RegularData<FloatType> &source, DynamicVector<FloatType> &mean, DynamicMatrix<FloatType> &rot) {
   computeCenterOfMass(source, mean);
   DynamicMatrix<FloatType> cov;
   computeMassCovariance(source, mean, cov);
   int32 dim = source.getDim();
   if (dim <= 2 && sizeof(FloatType) == sizeof(float32)) {
      rot.setDimension(dim, dim);
      float32 l1, l2;
      Vector2f v0, v1;
      eigenspace2f(cov[0][0], cov[0][1], cov[1][1], l1, v0, l2, v1);
      rot[0][0] = v0[0]; rot[1][0] = v1[0];
      rot[0][1] = v0[1]; rot[1][1] = v1[1];
   } else {
      DynamicMatrix<FloatType> v;
      DynamicVector<FloatType> d;
      DynamicPCA<FloatType>::computeSVD(cov, rot, d, v);
   }
}



template <typename FloatType>
void sphereClip(RegularData<FloatType> &data) {
   int dim = data.getDim();
   x4Assert(dim == 2 || dim == 3);
   int width  =           data.getExtends(0);
   int height =           data.getExtends(1);
   int depth  = dim > 2 ? data.getExtends(2) : 1;

   float32 halfWidth  = width  / 2.0f;
   float32 halfHeight = height / 2.0f;
   float32 halfDepth  = depth  / 2.0f;

   int diameter = data.getExtends().minComponent();
   float32 halfDiameter = diameter / 2.0f;

   FloatType *dataPtr = data.dataPtr();
   for (int z=0; z<depth; z++) {
      float32 zr = (z - halfDepth) / halfDiameter;
      for (int y=0; y<height; y++) {
         float32 yr = (y - halfHeight) / halfDiameter;
         for (int x=0; x<width; x++) {
            float32 xr = (x - halfWidth) / halfDiameter;
            if (xr*xr+yr*yr+zr*zr > 1.0f) {
               *dataPtr = (FloatType)0;
            }
            dataPtr++;
         }
      }
   }
}

template <typename FloatType>
void computePCATransform(
	RegularData<FloatType> &data, 
	DynamicVector<FloatType> *eigenvalues, 
	DynamicVector<FloatType> *mean,
	DynamicMatrix<FloatType> *eigenvects,
	bool normalize)
{
	x4Assert(data.getDim() >= 1);
	int32 numElements = data.getNumElements();
	int32 innerDim = data.getExtends(0);
	int32 numEx = numElements/innerDim;
	DynamicVector<FloatType> v(innerDim);
	FloatType *dataPtr = data.dataPtr();
	std::vector< DynamicVector<FloatType> > ex;
	for (int32 i=0; i<numEx; i++) {
		for (int32 j=0; j<innerDim; j++) {
			v[j] = *dataPtr;
			++dataPtr;
		}
		ex.push_back(v);
	}
	DynamicPCA<FloatType> dpca(&ex[0],	numEx);
	DynamicVector<FloatType> internMean = dpca.calcMean();
	DVectorF eValues = dpca.calcEigenValues(false);
	if (eigenvalues != NULL) {
		*eigenvalues = eValues;
	}
	DynamicMatrix<FloatType> internEVCs = dpca.getEigenVectors();	  
	card32 numEVCs = internEVCs.getColumns();
	for (card32 i=0; i<numEVCs; i++) {
		internEVCs[i] /= norm(internEVCs[i]);
	}

	dataPtr = data.dataPtr();
	for (int32 i=0; i<numEx; i++) {
		float32 *vecBegin = dataPtr;
		for (int32 j=0; j<innerDim; j++) {
			v[j] = *dataPtr;
			++dataPtr;
		}
		v -= internMean;
		dataPtr = vecBegin;
		for (int32 j=0; j<innerDim; j++) {
			*dataPtr = internEVCs[j] * v;
			if (normalize) *dataPtr /= eValues[j];
			dataPtr++;
		}
	}

	if (eigenvects != NULL) {
		*eigenvects = internEVCs;
	}	
	if (mean != NULL) {
		*mean = internMean;
	}
}



}

#endif
 