//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "RegularDataProcessingLD.h"
//---------------------------------------------------------------------------
#include "MathTools.h"
#include "MathTools.inline.h"
#include "LinearAlgebra.inline.h"
#include "RegularData.inline.h"
#include "fftw3.h"
#include "RGBAImage.h"
#include "FFTWrapper.h"
//---------------------------------------------------------------------------
#include <Qt/QImage.h> 
#include <qtcore/qstring>
#include <complex>
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------



void X4::affineTransform2DF(const RegularDataF &source, RegularDataF &dest, const Matrix2f &invTransf, const Vector2f &center) {
   dest.setExtends(source.getExtends());
   x4Assert(source.getDim() == 2);
   int32 width = source.getExtends(0);  float32 floatWidthMinOne = width-1;
   int32 height = source.getExtends(1); float32 floatHeightMinOne = height-1;

   Vector2f firstLookup = invTransf*(-center)+center;
   float32 lookupX = firstLookup[0];
   float32 lookupY = firstLookup[1];
   float32 lookupX_incrX = invTransf[0][0];
   float32 lookupY_incrX = invTransf[0][1];
   float32 lookupX_incrY = invTransf[1][0];
   float32 lookupY_incrY = invTransf[1][1];

   float32 *destDataPtr = dest.dataPtr();
   const float32 *sourceDataPtr = source.dataPtr();

   for (int y=0; y<height; y++) {
      float32 saveX = lookupX;
      float32 saveY = lookupY;
      for (int x=0; x<width; x++) {
         if (   lookupX >= 0.0f && lookupX <= floatWidthMinOne
             && lookupY >= 0.0f && lookupY <= floatHeightMinOne)
         {
            int lx = lookupX;
            int ly = lookupY;
            *destDataPtr = sourceDataPtr[lx + width*ly];
         }
         destDataPtr++;
         lookupX += lookupX_incrX;
         lookupY += lookupY_incrX;
      }
      lookupX = saveX;
      lookupX += lookupX_incrY;
      lookupY = saveY;
      lookupY += lookupY_incrY;
   }
}

void X4::affineTransformLinearInterp2DF(const RegularDataF &source, RegularDataF &dest, const Matrix2f &invTransf, const Vector2f &center) {
   dest.setExtends(source.getExtends());
   x4Assert(source.getDim() == 2);
   int32 width = source.getExtends(0);  float32 floatWidthMinOne = width-1;   float32 floatWidthMinTwo = width-2;
   int32 height = source.getExtends(1); float32 floatHeightMinOne = height-1; float32 floatHeightMinTwo = height-2;

   Vector2f firstLookup = invTransf*(-center)+center;
   float32 lookupX = firstLookup[0];
   float32 lookupY = firstLookup[1];
   float32 lookupX_incrX = invTransf[0][0];
   float32 lookupY_incrX = invTransf[0][1];
   float32 lookupX_incrY = invTransf[1][0];
   float32 lookupY_incrY = invTransf[1][1];

   float32 *destDataPtr = dest.dataPtr();
   const float32 *sourceDataPtr = source.dataPtr();

   for (int y=0; y<height; y++) {
      float32 saveX = lookupX;
      float32 saveY = lookupY;
      for (int x=0; x<width; x++) {
         if (   lookupX >= 0.0f && lookupX <= floatWidthMinOne
             && lookupY >= 0.0f && lookupY <= floatHeightMinOne)
         {
            if (lookupX <= floatWidthMinTwo && lookupY <= floatHeightMinTwo) {
               int lx = lookupX;
               float fx = lookupX - lx;
               int ly = lookupY;
               float fy = lookupY - ly;
               const float *sp = sourceDataPtr + lx + width*ly;
               *destDataPtr = (1-fy)*((1-fx)*(*sp) + fx*(*(sp+1)))
                            +     fy*((1-fx)*(*(sp+width)) + fx*(*(sp+1+width)));

            } else {
               int lx = lookupX;
               int ly = lookupY;
               *destDataPtr = sourceDataPtr[lx + width*ly];
            }
         }
         destDataPtr++;
         lookupX += lookupX_incrX;
         lookupY += lookupY_incrX;
      }
      lookupX = saveX;
      lookupX += lookupX_incrY;
      lookupY = saveY;
      lookupY += lookupY_incrY;
   }
}



void X4::computeCenterOfMass2DF(const RegularDataF &source, Vector2f &result) {
   x4Assert(source.getDim() == 2);
   int32 width = source.getExtends(0);
   int32 height = source.getExtends(1);
   float32 w = 0;
   float32 resultX = 0;
   float32 resultY = 0;
   const float32 *sourceDataPtr = source.dataPtr();
   for (int y=0; y<height; y++) {
      for (int x=0; x<width; x++) {
         float32 v = *sourceDataPtr;
         resultX += v*x;
         resultY += v*y;
         w += v;
      }
   }
   if (w >= 1E-7) {
      w = 1.0f/w;
      resultX *= w;
      resultY *= w;
   }
   result[0] = resultX;
   result[1] = resultY;
}

void X4::computeMassCovariance2DF(const RegularDataF &source, const Vector2f &mean, float32 &covXX, float32 &covXY, float32 &covYY) {
   x4Assert(source.getDim() == 2);
   int32 width = source.getExtends(0);
   int32 height = source.getExtends(1);

   float32 w = 0;
   covXX = 0;
   covXY = 0;
   covYY = 0;

   const float32 meanX = mean[0];
   const float32 meanY = mean[1];

   const float32 *sourceDataPtr = source.dataPtr();
   for (int y=0; y<height; y++) {
      float32 yFloat = y;
      for (int x=0; x<width; x++) {
         float32 v = *sourceDataPtr;
         float32 xFloat = x;
         float32 tx = (xFloat-meanX);
         covXX += v * tx * tx;
         float32 ty = (yFloat-meanY);
         covXY += v * tx * ty;
         covYY += v * ty * ty;
         w += v;
      }
   }
   if (w >= 1E-7) {
      w = 1.0f/w;
      covXX *= w;
      covXY *= w;
      covYY *= w;
   }
}

void X4::computePCACoordinateSystem2DF(const RegularDataF &source, Vector2f &mean, Matrix2f &rot) {
   computeCenterOfMass2DF(source, mean);
   float32 covXX, covXY, covYY;
   computeMassCovariance2DF(source, mean, covXX, covXY, covYY);
   float32 l1, l2;
   Vector2f v0, v1;
   eigenspace2f(covXX, covXY, covYY, l1, v0, l2, v1);
   rot[0][0] = v0[0]; rot[1][0] = v1[0];
   rot[0][1] = v0[1]; rot[1][1] = v1[1];
}

void X4::cubeSphereClip2DF(RegularDataF &data) {
   x4Assert(data.getDim() == 2);
   x4Assert(data.getExtends(0) == data.getExtends(1));
   int32 size = data.getExtends(0);
   int32 hs = size/2;
   int32 RQ = hs*hs;
   float32 *dataPtr = data.dataPtr();
   for (int32 y=0; y<size; y++) {
      for (int32 x=0; x<size; x++) {
         if ((x-hs)*(x-hs)+(y-hs)*(y-hs) > RQ) {
            *dataPtr = 0.0f;
         }
         dataPtr++;
      }
   }
}

void X4::getSubDataSet2DF(const RegularDataF &source, const Vector2i &lowerPos, const Vector2i &regionExtends, RegularDataF &target) {
   x4Assert(source.getDim() == 2);
   int32 width = source.getExtends(0);

   target.setExtends<int32, 2>(regionExtends);

   int startX = lowerPos[0];
   int startY = lowerPos[1];
   int endX = lowerPos[0] + regionExtends[0];
   int endY = lowerPos[1] + regionExtends[1];

   float32 *dataPtr = target.dataPtr();
   const float32 *sourcePtr = source.dataPtr();

   for (int y=startY; y<endY; y++) {
      for (int x=startX; x<endX; x++) {
         *dataPtr++ = sourcePtr[x+width*y];
      }
   }
}


inline float32 bilinearSample2DF(const float32 *data, const card32 width, const card32 height, const Vector2f &pos) {
	int x = (int)pos[0];
	float32 fracx = pos[0]-x;
	int y = (int)pos[1];
	float32 fracy = pos[1]-y;
	if (x<0) 
		x=0;
	if (y<0)
		y = 0;

	const float32 *dp = data + (x + width*y);
	if( x<(int)width-1 && y<(int)height-1)
		return 	(1-fracy) * ((1-fracx)*(*dp) + fracx*(*(dp+1)))
			+    fracy  * ((1-fracx)*(*(dp+width)) + fracx*(*(dp+1+width)));

	if( x<(int)width-1 )
		return (1-fracx)*(*dp) + fracx*(*(dp+1));
	if( y<(int)height-1 )
		return (1-fracy)*(*dp) + fracy*(*(dp+width));
	return (*dp);
}

void X4::createRDFromImage(RegularDataF &dest, const QImage *source) {
	const card32 width = source->width();
	const card32 height = source->height();
	
	DVectorI extends(2);
	extends[0] = width;
	extends[1] = height;
	
	dest.setExtends(extends);
	
	float32 *destPtr = dest.dataPtr();
	
	for (card32 y=0; y<height; y++) {
		for (card32 x=0; x<width; x++) {
			QRgb c = source->pixel(x,y);
			float32 v = (qRed(c)+qGreen(c)+qBlue(c)) / (3.0f * 255.0f);
			*destPtr = v;
			destPtr++;
		}
	}
}


void X4::createImageFromRD(QImage &dest, const RegularDataF &source, float32 intensityScaleFactor) {
	x4Assert(source.getDim() == 2);
	if (!(source.getDim() == 2)) return;

	DVectorI extends = source.getExtends();
	const int width = extends[0];
	const int height = extends[1];
	
	dest = QImage(width, height, QImage::Format_ARGB32);
	const float32 *sourcePtr = source.dataPtr();
	
	
	for (int y=0; y<height; y++) {
		for (int x=0; x<width; x++) {
			int c = (*sourcePtr) * intensityScaleFactor * 255.0f;
			if (c<0) c = 0;
			if (c>255) c = 255;
			dest.setPixel(x,y, qRgb(c,c,c));
			sourcePtr++;
		}
	}
}

static inline int clamp0_255(float32 v) {
	if (v>255) v = 255;
	if (v<0) v = 0;
	return (int)v;
}

void X4::createRGBImageFromRD(QImage &dest, const RegularDataF &source, float32 intensityScaleFactor, float32 offset) {
	x4Assert(source.getDim() == 3);
	if (!(source.getDim() == 3)) return;

	DVectorI extends = source.getExtends();
	const int rgbExt = extends[0];
	const int extraData = rgbExt - 3;
	const int width = extends[1];
	const int height = extends[2];
	
	dest = QImage(width, height, QImage::Format_ARGB32);
	const float32 *sourcePtr = source.dataPtr();
		
	for (int y=0; y<height; y++) {
		for (int x=0; x<width; x++) {
			QRgb c;
			if (rgbExt >= 3) {
				//debugOutput << *sourcePtr << "\t";
				int r = clamp0_255((*sourcePtr++) * intensityScaleFactor * 255.0f + offset);
				//debugOutput << *sourcePtr << "\t";
				int g = clamp0_255((*sourcePtr++) * intensityScaleFactor * 255.0f + offset);
				//debugOutput << *sourcePtr << "\n";
				int b = clamp0_255((*sourcePtr++) * intensityScaleFactor * 255.0f + offset);
				c = qRgb(r,g,b);
				sourcePtr += extraData;
				
			} else if (rgbExt == 2) {
				int r = clamp0_255((*sourcePtr++) * intensityScaleFactor * 255.0f + offset);
				int g = clamp0_255((*sourcePtr++) * intensityScaleFactor * 255.0f + offset);
				c = qRgb(r,g,0);
			} else if (rgbExt == 1) {
				int r = clamp0_255((*sourcePtr++) * intensityScaleFactor * 255.0f + offset);
				c = qRgb(r,0,0);
			}
			dest.setPixel(x,y, c);
		}
	}
}

void X4::createRGBAFloatImageFromRD(RGBAFloat32Image &dest, const RegularDataF &source) {
	x4Assert(source.getDim() == 3);
	if (!(source.getDim() == 3)) return;

	DVectorI extends = source.getExtends();
	const int rgbExt = extends[0];
	const int extraData = rgbExt - 4;
	const int width = extends[1];
	const int height = extends[2];
	
	dest.resize(width, height);
	const float32 *sourcePtr = source.dataPtr();
		
	for (int y=0; y<height; y++) {
		for (int x=0; x<width; x++) {
			Vector4f c;
			if (rgbExt >= 4) {
				float r = *sourcePtr++;
				float g = *sourcePtr++;
				float b = *sourcePtr++;
				float a = *sourcePtr++;
				c = makeVector4f(r,g,b,a);
				sourcePtr += extraData;
			} else if (rgbExt >= 3) {
				float r = *sourcePtr++;
				float g = *sourcePtr++;
				float b = *sourcePtr++;
				c = makeVector4f(r,g,b,0);
			} else if (rgbExt == 2) {
				float r = *sourcePtr++;
				float g = *sourcePtr++;
				c = makeVector4f(r,g,0,0);
			} else if (rgbExt == 1) {
				float r = *sourcePtr++;
				c = makeVector4f(r,0,0,0);
			}
			dest.setPixel(x,y, c);
		}
	}
}


void X4::polarTransform2DF(const RegularDataF &source, RegularDataF &dest, card32 numCircularSamples, card32 numRadialSamples) {
	x4Assert(source.getDim() == 2);
	if (!(source.getDim() == 2)) return;
	DVectorI destExtends(2);
	destExtends[0] = numCircularSamples;
	destExtends[1] = numRadialSamples;
	dest.setExtends(destExtends);
	
	const card32 sourceWidth = source.getExtends(0);
	const card32 sourceHeight = source.getExtends(1);
	const card32 minSourceExtend = min(sourceWidth, sourceHeight);
	
	const Vector2f sourceCenter = makeVector2f((float32)sourceWidth / 2.0f, (float32)sourceHeight / 2.0f);
	
	
	std::vector<Vector2f> angleVects(numCircularSamples);
	for (card32 c=0; c<numCircularSamples; c++) {
		float32 cf = (float32)c/(float32)numCircularSamples * 2.0f * (float32)(M_PI);
		angleVects[c] = makeVector2f(cos(cf), sin(cf));
	}
	
	float32 radialScale = (float32)minSourceExtend / 2.0f / (float32)numRadialSamples;
	float32 *destData = dest.dataPtr();
	const float32 *sourceData = source.dataPtr();
	
	for (card32 r=0; r<numRadialSamples; r++) {
		for (card32 c=0; c<numCircularSamples; c++) {
			float32 rf = r * radialScale;
			Vector2f samplePos = sourceCenter + angleVects[c] * rf;
			*destData = bilinearSample2DF(sourceData, sourceWidth, sourceHeight, samplePos);
			destData++;
		}
	}
}

void X4::fastLineConvolve2DF(const RegularDataF &sourceA, const RegularDataF &sourceB, RegularDataF &dest, bool reverseB) {
	DVectorI extends = sourceA.getExtends();
	x4Assert(sourceB.getExtends() == extends);
	if (sourceB.getExtends() != extends) return;
	dest.setExtends(extends);
	
	const card32 width = extends[0];
	const card32 height = extends[1];
	
   fftwf_complex *in = (fftwf_complex*)fftwf_malloc(sizeof(fftwf_complex) * width);;
   fftwf_complex *outA = (fftwf_complex*)fftwf_malloc(sizeof(fftwf_complex) * width);
   fftwf_complex *outB = (fftwf_complex*)fftwf_malloc(sizeof(fftwf_complex) * width);

   fftwf_plan planThereA = fftwf_plan_dft_1d( width, in, outA, FFTW_FORWARD, FFTW_ESTIMATE );
   fftwf_plan planThereB = fftwf_plan_dft_1d( width, in, outB, FFTW_FORWARD, FFTW_ESTIMATE );
   fftwf_plan planBack = fftwf_plan_dft_1d( width, in, outA, FFTW_BACKWARD, FFTW_ESTIMATE );
   
   const float32 *sourcePtrA = sourceA.dataPtr();
   const float32 *sourcePtrB = sourceB.dataPtr();
   float32 *destPtr = dest.dataPtr();
   
   for (card32 y=0; y<height; y++) {
		for (card32 x=0; x<width; x++) {
			in[x][0] = *sourcePtrA;
			in[x][1] = 0;
			sourcePtrA++;
		}
	   fftwf_execute(planThereA);

		if (reverseB) {
			const float32 *sourceSafeB = sourcePtrB;
			sourcePtrB += y*width + (width-1);
			for (card32 x=0; x<width; x++) {
				in[x][0] = *sourcePtrB;
				in[x][1] = 0;
				sourcePtrB--;
			}
			sourcePtrB = sourceSafeB;
		} else {
			for (card32 x=0; x<width; x++) {
				in[x][0] = *sourcePtrB;
				in[x][1] = 0;
				sourcePtrB++;
			}
		}
	   fftwf_execute(planThereB);
	   
	   for (card32 x=0; x<width; x++) {
			complex<float> valA(outA[x][0], outA[x][1]);
			complex<float> valB(outB[x][0], outB[x][1]);
			complex<float>	valO = valA * valB;
			in[x][0] = valO.real();
			in[x][1] = valO.imag();
	   }
	   fftwf_execute(planBack);
	   
	   for (card32 x=0; x<width; x++) {
			*destPtr = outA[x][0] / (float)width; // fftw computes unnormalized transform
			destPtr++;
	   }
	}
	
   fftwf_destroy_plan(planThereA);
   fftwf_destroy_plan(planThereB);
   fftwf_destroy_plan(planBack);

   fftwf_free(in);
   fftwf_free(outA);
   fftwf_free(outB);
	
}

float32 X4::computeMaximumCorrelationRotation(
	const RegularDataF &sourceA, const RegularDataF &sourceB,
	card32 numRadialSamples, card32 numCircularSamples) 
{
	RegularDataF polarA;
	polarTransform2DF(sourceA, polarA, numCircularSamples, numRadialSamples);

	RegularDataF polarB;
	polarTransform2DF(sourceB, polarB, numCircularSamples, numRadialSamples);

	RegularDataF result;
	fastLineConvolve2DF(polarA, polarB, result, true);
	
	const int width = result.getExtends(0);
	const int height = result.getExtends(1);

	DVectorI coord(2);
	float32 maxCor = 0;
	int maxX = 0;
	for (int x=0; x<width; x++) {
		float32 vsum = 0;
		for (int y=1; y<height; y++) {
			coord[0] = x;
			coord[1] = y;
			vsum += result.getElement(coord) * y;
		}
		vsum /= height;
		if (vsum > maxCor) {
			maxX = x;
			maxCor = vsum;
		}
	}
	return 2*M_PI * (float32)maxX / (float32)width;
}

void X4::rotateToMaximumCorrelationRotation(
	const RegularDataF &reference, const RegularDataF &source, RegularDataF &dest,
	card32 numRadialSamples, card32 numCircularSamples)
{
	float32 angle = -computeMaximumCorrelationRotation(reference, source, numRadialSamples, numCircularSamples);
	affineTransformLinearInterp2DF(source, dest, 
		makeMatrix2f(cos(angle), sin(angle), 
		            -sin(angle), cos(angle)),
		makeVector2f(source.getExtends(0)/2, source.getExtends(1)/2)
	);
	
}

float32 X4::computeMaxCorrelationSSDOverAllAngles2DF(RegularDataF sourceA, const RegularDataF &sourceB) {
	x4Assert(sourceA.getExtends(0) == sourceA.getExtends(1));
	x4Assert(sourceB.getExtends(0) == sourceA.getExtends(1));
	x4Assert(sourceB.getExtends(0) == sourceB.getExtends(1));
	card32 numRadialSamples = sourceA.getExtends(0) / 2;
	card32 numCircularSamples = (card32)( sourceA.getExtends(0) * M_PI );
	RegularDataF sourceBRot;
	rotateToMaximumCorrelationRotation(sourceA, sourceB, sourceBRot, numRadialSamples, numCircularSamples);
	cubeSphereClip2DF(sourceA);
	cubeSphereClip2DF(sourceBRot);
	RegularDataF diff = sourceA - sourceBRot;
	return diff * diff; 
}

void X4::polarCirc1DFourierTransform2DF(
	const RegularDataF &source, 
	RegularDataF &destReal, RegularDataF &destImg,
	card32 numRadialSamples, card32 numCircularSamples) 
{
	RegularDataF polar;
	polarTransform2DF(source, polar, numCircularSamples, numRadialSamples);
	DVectorI extends = polar.getExtends();
	destReal.setExtends(extends);
	destImg.setExtends(extends);
	const card32 width = extends[0];
	const card32 height = extends[1];
	
   fftwf_complex *in = (fftwf_complex*)fftwf_malloc(sizeof(fftwf_complex) * width);
   fftwf_complex *out = (fftwf_complex*)fftwf_malloc(sizeof(fftwf_complex) * width);

   fftwf_plan plan = fftwf_plan_dft_1d( width, in, out, FFTW_FORWARD, FFTW_ESTIMATE );
   
   const float32 *sourcePtr = source.dataPtr();
   float32 *destPtrReal = destReal.dataPtr();
   float32 *destPtrImg  = destImg.dataPtr();
   
   const float32 fact = 1.0f/sqrt((float32)width);	// fftw computes unnormalized transform
   
   for (card32 y=0; y<height; y++) {
		for (card32 x=0; x<width; x++) {
			in[x][0] = *sourcePtr;
			in[x][1] = 0;
			sourcePtr++;
		}
	   fftwf_execute(plan);

	   for (card32 x=0; x<width; x++) {
			*destPtrReal = out[x][0] * fact; 
			*destPtrImg  = out[x][1] * fact; 
			destPtrReal++;
			destPtrImg++;
	   }
	}
	
   fftwf_destroy_plan(plan);

   fftwf_free(in);
   fftwf_free(out);			
}

float32 X4::computeSSD2DF(const RegularDataF &sourceA, const RegularDataF &sourceB) {
	x4Assert(sourceA.getExtends() == sourceB.getExtends());
	x4Assert(sourceA.getDim() == sourceB.getDim());
	
	const card32 width = sourceA.getExtends(0);
	const card32 height = sourceA.getExtends(1);

	const float32 *dataPtrA = sourceA.dataPtr();
	const float32 *dataPtrB = sourceB.dataPtr();

	float32 ssd = 0;

	for (card32 i=0;i<width*height;i++) {
		float32 v = (*dataPtrA) - (*dataPtrB);
		ssd += v*v;
		dataPtrA++;
		dataPtrB++;
	}
	return ssd;
}


float32 X4::computeMaxSSDOverAllAnglesBruteForce(RegularDataF sourceA, RegularDataF sourceB) {
	x4Assert(sourceA.getExtends() == sourceB.getExtends());
	x4Assert(sourceA.getExtends(0) == sourceB.getExtends(0));

	card32 size = sourceA.getExtends(0);

	cubeSphereClip2DF(sourceA);
	cubeSphereClip2DF(sourceB);

	card32 numCircularSamples = (int)( 2*size * M_PI );

	RegularDataF rotatedA;
	float32 diff = -1;
	for (card32 i=0; i<numCircularSamples; i++) {
		float32 angle = i/(float32)numCircularSamples * M_PI * 2.0f;
		affineTransformLinearInterp2DF(sourceA, rotatedA, 
			makeMatrix2f(cos(angle), sin(angle), 
			            -sin(angle), cos(angle)),
			makeVector2f(size/2, size/2)
		);
		float32 newDiff = computeSSD2DF(rotatedA, sourceB);
																							 
		if (diff == -1) {
			diff = newDiff;
		} else {
			if (newDiff < diff) {
				diff = newDiff;
			}
		}
	}	
	
	return diff;	
}



void X4::polarCirc1DPowerSpectrum2DF(
	const RegularDataF &source, 
	RegularDataF &dest,
	card32 numRadialSamples, card32 numCircularSamples) 
{
	RegularDataF polar;
	polarTransform2DF(source, polar, numCircularSamples, numRadialSamples);
	DVectorI extends = polar.getExtends();
	dest.setExtends(extends);
	const card32 width = extends[0];
	const card32 height = extends[1];
	
   fftwf_complex *in = (fftwf_complex*)fftwf_malloc(sizeof(fftwf_complex) * width);
   fftwf_complex *out = (fftwf_complex*)fftwf_malloc(sizeof(fftwf_complex) * width);

   fftwf_plan plan = fftwf_plan_dft_1d( width, in, out, FFTW_FORWARD, FFTW_ESTIMATE );
   
   const float32 *sourcePtr = polar.dataPtr();
   float32 *destPtr = dest.dataPtr();
   
   const float32 fact = 1.0f/(float32)width;	// fftw computes unnormalized transform
   
   for (card32 y=0; y<height; y++) {
		for (card32 x=0; x<width; x++) {
			in[x][0] = *sourcePtr;
			in[x][1] = 0;
			sourcePtr++;
		}
	   fftwf_execute(plan);

	   for (card32 x=0; x<width; x++) {
			*destPtr = (sqr(out[x][0]) + sqr(out[x][1])) * fact; 
			destPtr++;
	   }
	}
	
   fftwf_destroy_plan(plan);

   fftwf_free(in);
   fftwf_free(out);			
}

struct TPolarSampleRing
{
	std::vector<Vector2f> samples;
};


std::vector<float> X4::adaptivePolarCirc1DPowerSpectrum2DF(const RegularDataF &source, card32 numRadialSamples)
{
	std::vector<float> result;

	int width = source.getExtends(0);
	int height = source.getExtends(1);
	if( width != height )
		throw "Don't use adaptivePolarCirc1DPowerSpectrum2DF with non-square images!\n";

	float sampleSpacing = ((float)width/2.0f - 1.0f) / (float)numRadialSamples;
	Vector2f center = makeVector2f( width-1,height-1) / 2.0f;

	// create sampling buffer
	static std::map<int,std::vector<TPolarSampleRing> > sampleBuffer;
	int index = numRadialSamples + width * 100000;
	if( sampleBuffer.find( index ) == sampleBuffer.end() )
	{
//		QImage image(width,height,QImage::Format_RGB32);
//		image.fill( 0xffffff);

		std::vector<TPolarSampleRing> rings;
		for( card32 i=1;i<numRadialSamples;i++ )
		{
			TPolarSampleRing ring;
			float radius = sampleSpacing * i;
			
			int numCircularSamples = ceil( 2.0f * M_PI * radius / sampleSpacing );
			float angleSteps = 2.0f * M_PI / (float)numCircularSamples;
			float startAngle = (i % 2 == 1) ? angleSteps / 2.0f : 0;
			for( card32 j=0;j<(card32)numCircularSamples;j++ )
			{
				float angle = startAngle + angleSteps*j;
				Vector2f samplePoint = makeVector2f( sin(angle), cos(angle) ) * radius + center;
				ring.samples.push_back( samplePoint );								
//				image.setPixel( samplePoint[0],samplePoint[1], 0);
			}
			rings.push_back(ring);
		}
		sampleBuffer[ index ] = rings;
//		image.save( "funkhousersampling.png", "PNG" );
		
	}
	std::map<int,std::vector<TPolarSampleRing> >::iterator iter = sampleBuffer.find( index );
	
	/*QImage image( width,height, QImage::Format_RGB32);
	
	for( card32 x=0;x<width;x++ )
		for( card32 y=0;y<height;y++ )
	{
		int gray = max(0,min(255,source.getElement( makeVector2f(x,y)) * 255.f));
		card32 col = gray + (gray << 8) + (gray << 16);
		image.setPixel(x,y, col );
	}
	static int numImg = 0;
	image.save( QString("funkCurv") + QString::number(numImg++) + ".png", "PNG" );*/
	
	// first sample from midpoint
	//result.push_back( bilinearSample2DF( source.dataPtr(), width, height, center ) );

	for( card32 i=1;i<numRadialSamples;i++ )
	{
		size_t numCircularSamples = iter->second[i-1].samples.size();
		FFTWPlan1D * plan = MainFFTWrapper->getPlan( (card32)numCircularSamples );

		// sample
		for( card32 j=0;j<numCircularSamples;j++ )
		{
			float sample = bilinearSample2DF( source.dataPtr(), width, height, iter->second[i-1].samples[j] );
			plan->in[j][0] = sample;
			plan->in[j][1] = 0;
		}

		// fft tranform
		fftwf_execute(plan->plan);

		// save power spectrum
		float32 fact = 1.0f;//(float32)numCircularSamples;
		for( card32 j=0;j<numCircularSamples;j++ )
		{
			result.push_back(sqrt(sqr(plan->out[j][0]) + sqr(plan->out[j][1])));
		}
	}

	// everything fine and done!
	return result;
}


card32 X4::calcPowerSpectrumDescriptorSize2DF(card32 numRadialSamples, card32 numCircularSamples) {
	card32 dim =0;
	for (card32 y=0; y<numRadialSamples; y++) {
		card32 numFromThisRow = (card32)ceil((double)numCircularSamples/(double)numRadialSamples*(double)(y+1));
		dim += numFromThisRow;
	}
	return dim;
}

void X4::calcPowerSpectrumDescriptor2DF(const RegularDataF &source, DVectorF &descr,
                                         card32 numRadialSamples, card32 numCircularSamples)
{
	RegularDataF pspec;
	DVectorI test(2); test[0] = numCircularSamples; test[1] = numRadialSamples; 
	pspec.setExtends(test);
	polarCirc1DPowerSpectrum2DF(source, pspec, numRadialSamples, numCircularSamples);
	DVectorI extends = pspec.getExtends();
	const card32 width = extends[0];
	const card32 height = extends[1];
	card32 dim = calcPowerSpectrumDescriptorSize2DF(numRadialSamples, numCircularSamples);
	descr.setDim(dim);
	card32 i=0;
	const float32 *dataPtr = pspec.dataPtr();
	for (card32 y=0; y<height; y++) {
		card32 numFromThisRow = (card32)ceil((double)width/(double)height*(double)(y+1));
		for (card32 x=0; x<numFromThisRow; x++) {
			descr[i] = dataPtr[x+width*y];
			i++;
		}
	}
}


void X4::calcPowerSpectrumTransform2DF(const RegularDataF &source, RegularDataF &dest, card32 diameter){
	const DVectorI extends = source.getExtends();
	x4Assert(extends.getDim() == 2);
	x4Assert(extends[0] > (int32)diameter);
	x4Assert(extends[1] > (int32)diameter);
	const card32 width = extends[0]-diameter;
	const card32 height = extends[1]-diameter;

	DVectorI destExtends(3);
	card32 numRadialSamples = diameter / 2;
	card32 numCircularSamples = (int)( diameter * M_PI );
	card32 dim = calcPowerSpectrumDescriptorSize2DF(numRadialSamples, numCircularSamples);
	destExtends[0] = dim;
	destExtends[1] = width;
	destExtends[2] = height;

	dest.setExtends(destExtends);
	float32 *destPtr = dest.dataPtr();
	DVectorF descr(dim);
	RegularDataF subBlock;
	DVectorI lowerPos(2);
	DVectorI regionExtends(2);
	regionExtends[0] = diameter;
	regionExtends[1] = diameter;
	
	for (card32 y=0; y<height; y++) {
		for (card32 x=0; x<width; x++) {
			lowerPos[0] = x;
			lowerPos[1] = y;
			source.getSubDataSet(lowerPos, regionExtends, subBlock);
			calcPowerSpectrumDescriptor2DF(subBlock, descr, numRadialSamples, numCircularSamples);
			for (card32 i=0; i<dim; i++) {
				*destPtr = descr[i];
				destPtr++;
			}
		}
	}
}


void X4::computerInnerDimMinMax(const RegularDataF &source, DVectorF &minResult, DVectorF &maxResult) {
	const DVectorI extends = source.getExtends();
	x4Assert(extends.getDim() == 3);
	const card32 dim = extends[0];
	const card32 width = extends[1];
	const card32 height = extends[2];

	minResult.setDim(dim);
	maxResult.setDim(dim);

	const float32 *sourcePtr = source.dataPtr();

	bool first = true;
	for (card32 y=0; y<height; y++) {
		for (card32 x=0; x<width; x++) {
			for (card32 i=0; i<dim; i++) {
				float32 v = *sourcePtr;
				if (!first) {
					if (v < minResult[i]) minResult = v;
					if (v > maxResult[i]) maxResult = v;
				} else {
					minResult[i] = v;
					maxResult[i] = v;					
				}
				sourcePtr++;
			}
			first = false;
		}
	}
}


/*void X4::computerInnerDimMovingHistogram(const RegularDataF &source, RegularDataF &dest, card32 diameter, card32 numBucketsPerDim) {
	const DVectorI extends = source.getExtends();
	x4Assert(extends.getDim() == 3);
	x4Assert(extends[1] > (int32)diameter);
	x4Assert(extends[2] > (int32)diameter);
	const card32 dim = extends[0];
	const card32 width = extends[1]-diameter;
	const card32 height = extends[2]-diameter;

	const DVectorI destExtends(4);
	destExtends[0] = numBucketsPerDim;
	destExtends[1] = 

	for (card32 y=0; y<height; y++) {
		for (card32 x=0; x<width; x++) {
			lowerPos[0] = x;
			lowerPos[1] = y;
			source.getSubDataSet(lowerPos, regionExtends, subBlock);
			calcPowerSpectrumDescriptor2DF(subBlock, descr, numRadialSamples, numCircularSamples);
			for (card32 i=0; i<dim; i++) {
				*destPtr = descr[i];
				destPtr++;
			}
		}
	}
}*/
