//---------------------------------------------------------------------------
#ifndef RegularDataProcessingLDH
#define RegularDataProcessingLDH
//---------------------------------------------------------------------------
#include "x3editor2.h"
//---------------------------------------------------------------------------
#include "LinearAlgebra.h"
#include "RegularData.h"
//---------------------------------------------------------------------------

namespace X4 {

class RGBAFloat32Image;

// these are optimized special cases of methods from RegularDataProcessing.h
// restriction: fixed float type, lower dimension ("LD") only

X3EDITOR2_API void affineTransform2DF(const RegularDataF &source, RegularDataF &dest, const Matrix2f &invTransf, const Vector2f &center);
X3EDITOR2_API void affineTransformLinearInterp2DF(const RegularDataF &source, RegularDataF &dest, const Matrix2f &invTransf, const Vector2f &center);

X3EDITOR2_API void computeCenterOfMass2DF(const RegularDataF &source, Vector2f &result);

X3EDITOR2_API void computeMassCovariance2DF(const RegularDataF &source, const Vector2f &mean, float32 &covXX, float32 &covXY, float32 &covYY);

X3EDITOR2_API void computePCACoordinateSystem2DF(const RegularDataF &source, Vector2f &mean, Matrix2f &rot);

X3EDITOR2_API void cubeSphereClip2DF(RegularDataF &data);

X3EDITOR2_API void getSubDataSet2DF(const RegularDataF &source, const Vector2i &lowerPos, const Vector2i &regionExtends, RegularDataF &target);


X3EDITOR2_API void createRDFromImage(RegularDataF &dest, const QImage *source);
X3EDITOR2_API void createImageFromRD(QImage &dest, const RegularDataF &source, float32 intensityScaleFactor = 1.0f);
/// maps 0,1,2 or the innermost dimension to RGB
X3EDITOR2_API void createRGBImageFromRD(QImage &dest, const RegularDataF &source, float32 intensityScaleFactor = 1.0f, float32 offset = 0.0f);
X3EDITOR2_API void createRGBAFloatImageFromRD(RGBAFloat32Image &dest, const RegularDataF &source);
                 

/// transform 2d image in polar coordinates
X3EDITOR2_API void polarTransform2DF(const RegularDataF &source, RegularDataF &dest, card32 numCircularSamples, card32 numRadialSamples);

/// computes convolution of dim. zero (x) direction lines of pair of 2d images
/// sourceA and sourceB must have the same extends. An extends[0] size beeing a power of two is advantageous in terms of efficiency.
/// if reverseB is true, a corrolation rather than a convolution is computed (i.e. B is flipped in x direction before Fourier transform)
X3EDITOR2_API void fastLineConvolve2DF(const RegularDataF &sourceA, const RegularDataF &sourceB, RegularDataF &dest, bool reverseB = false);

/// compares two 2d-images and computes the angle at which the maximum correlation is found when rotating sourceB over sourceA
/// returns the angle in radian
/// the sampling parameters affect the size / accuracy of the interal polar transform
X3EDITOR2_API float32 computeMaximumCorrelationRotation(const RegularDataF &sourceA, const RegularDataF &sourceB,
                                                      card32 numRadialSamples, card32 numCircularSamples);

/// similar to computeMaximumCorrelationRotation, but rotates "source" to "reference" match best and stores the result in "dest"
X3EDITOR2_API void rotateToMaximumCorrelationRotation(const RegularDataF &reference, const RegularDataF &source, RegularDataF &dest,
                                                    card32 numRadialSamples, card32 numCircularSamples);

/// computes the SSD of two 2d image patches (circular) at the angle of maximum correlation
X3EDITOR2_API float32 computeMaxCorrelationSSDOverAllAngles2DF(RegularDataF sourceA, const RegularDataF &sourceB);

X3EDITOR2_API float32 computeSSD2DF(const RegularDataF &sourceA, const RegularDataF &sourceB);

/// compute the SSD of two 2d patches by repeated comparison
X3EDITOR2_API float32 computeMaxSSDOverAllAnglesBruteForce(RegularDataF sourceA, RegularDataF sourceB);

/// computes circle-wise 1D Fourier transform of polar transform around "source" center
/// source must be 2DF, dest is 3DF, with the innermost extend referring to real/img of the transform [0] = real, [1] = img
X3EDITOR2_API void polarCirc1DFourierTransform2DF(const RegularDataF &source, RegularDataF &destReal, RegularDataF &destImg,
                                             card32 numRadialSamples, card32 numCircularSamples);

/// like polarCirc1DFourierTransform, but computes power spectrum
X3EDITOR2_API void polarCirc1DPowerSpectrum2DF(const RegularDataF &source, RegularDataF &dest,
                                          card32 numRadialSamples, card32 numCircularSamples);

/// computes a rotationinvariant power spectrum using an adaptive sampling rate 
/// adapts circular sampling rate to radial sampling rate
X3EDITOR2_API std::vector<float> adaptivePolarCirc1DPowerSpectrum2DF(const RegularDataF &source, card32 numRadialSamples);
/// computes the dimension of a power spectrum descriptor (see	calcPowerSpectrumDescriptor2DF)
X3EDITOR2_API card32 calcPowerSpectrumDescriptorSize2DF(card32 numRadialSamples, card32 numCircularSamples);

/// compute local, rotation invariant power spectrum descriptor
X3EDITOR2_API void calcPowerSpectrumDescriptor2DF(const RegularDataF &source, DVectorF &descr,
                                                card32 numRadialSamples, card32 numCircularSamples);

/// computes a power spectrum descriptor for each "radius"-neighborhood (that fits in) and stores the result in a 3DF output dest
/// the inner dimension (dest.extends[0]) refers to the descriptor index
X3EDITOR2_API void calcPowerSpectrumTransform2DF(const RegularDataF &source, RegularDataF &dest, card32 diameter);

/// computes minimum and maximum of inner dimension vectors entries
X3EDITOR2_API void computerInnerDimMinMax(const RegularDataF &source, DVectorF &minResult, DVectorF &maxResult);

// /// computes a histogramm over the vectors of the inner dimension within a diameter^2 neighborhood
// X3EDITOR2_API void computerInnerDimMovingHistogram(const RegularDataF &source, RegularDataF &dest, card32 diameter, card32 numBucketsPerDim);

// /// filters inner dimension std::vector entries
// X3EDITOR2_API void filterInnerDim(const RegularDataF &source, RegularDataF &dest, card32 diameter);


}

#endif
 