//---------------------------------------------------------------------------
#include "StdAfx.h"

//---------------------------------------------------------------------------
#include "S2kit.h"
//---------------------------------------------------------------------------
#include "LinearAlgebra.inline.h"
#include "UnstructuredInCorePointCloud.h"
#include <fstream>
//---------------------------------------------------------------------------
#include "fftw3.h"
//---------------------------------------------------------------------------
#include "pmls.h"
#include "makeweights.h"
#include "naive_synthesis.h"
#include "cospmls.h"
#include "FST_semi_fly.h"
#include "FST_semi_memo.h"
#include "primitive.h"
#include "seminaive.h"
//---------------------------------------------------------------------------
#ifndef PI
#define PI 3.14159265358979
#endif
//---------------------------------------------------------------------------
 
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

int S2kit::seanindex(int m, int l, int bw) {
    return ::seanindex(m, l, bw);
}

void S2kit::computeSpectrum(const DVectorD &dataReal, const DVectorD &dataImg, DVectorD *spectrum, card32 bw) {
    x4Assert(dataReal.getDim() == bw*bw*4);
    x4Assert(dataImg.getDim() == bw*bw*4);

    double *coeffsReal = new double[bw*bw];
    double *coeffsImg = new double[bw*bw];

    doSHT(const_cast<double *>(dataReal.data()), const_cast<double *>(dataImg.data()), coeffsReal, coeffsImg, bw, false);

    spectrum->setDim(bw*bw);
    spectrum->setZero();
    for (card32 l = 0; l < bw; ++l) {
        for (int m = -(int)l; m <= (int)l; ++m) {
            int index = seanindex(m, l, bw);
            (*spectrum)[l] += coeffsReal[index] * coeffsReal[index] + coeffsImg[index] * coeffsImg[index];
        }
    }

    delete[] coeffsReal;
    delete[] coeffsImg;
}

PointCloud* S2kit::createPointCloudFromS2File(const char* file){
  double *rdata;
  double x;
  card32 bw=64;
  card32 size = 2*bw;
  card32 pointCount = size * size;

  /* allocate memory */
  rdata = new double[pointCount];

  /* now read in samples */
  std::ifstream in(file);

  for(card32 i = 0 ; i < pointCount ; i++ )
    {
      x4Assert(!in.eof());
      in >> x; rdata[i]=x;
    }

	PointCloud* result= createPointCloudFromS2Data(rdata,bw);

 
  delete(rdata);
  return result;
}

PointCloud* S2kit::createPointCloudFromS2DataFile(const char* file){
  double *rcoeffs, *icoeffs, *rdata, *idata=NULL;
  double x;
  card32 bw=17;
  card32 size = 2*bw;

  /* allocate memory */
  rdata = new double[size * size];
  idata = new double[size * size];
  rcoeffs = new double[bw * bw];
  icoeffs = new double[bw * bw];
  x4Assert(rcoeffs&&icoeffs && rdata&&idata);

  card32 pointCount = size * size;

  /* now read in samples */
  std::ifstream in(file);

  for(card32 i = 0 ; i < pointCount ; i++ )
    {
      x4Assert(!in.eof());
      /* first the real part of the sample */
      in >> x; rdata[i]=x;
      /* now the imaginary part */
      in >> x; idata[i]=x;
    }
//	doSHT(rdata,idata,rcoeffs ,icoeffs ,bw, true);

	PointCloud* result= createPointCloudFromS2Data(rdata,bw);
	//PointCloud* result= createPointCloudFromS2Data(idata,bw);
 
  delete(rdata);
  delete(idata);
    delete (icoeffs);
    delete (rcoeffs);

  return result;
}



PointCloud* S2kit::createPointCloudFromS2Data(double *rdata, card32 bw){
  card32 size = 2*bw;
  card32 pointCount = size * size;


//Generate Pointcloudsphere
//========================

  //Generate Points
  std::vector<Vector3f> position;
  position.resize(pointCount);


  //size = 2*bw

  float32 phi_k, theta_j;
  card32 at=0;
  for (card32 j=0; j<size; j++)
	  for(card32 k=0; k<size; k++) {
        theta_j = PI*(float)(2*j+1)/(float)(2*size);
        phi_k = 2*PI*(float)k/(float) size;

        position[at]=makeVector3f(sin(theta_j)*cos(phi_k),
               sin(theta_j)*sin(phi_k),
               cos(theta_j));
        at++;
  }




  //Create Pointset, color from "rdata"
  PointSet *ps = new PointSet();
  VertexDescriptor vd;
   vd.pushAttrib(mVAD("position", 3, VAD::DATA_FORMAT_FLOAT32));
   vd.pushAttrib(mVAD("color", 3, VAD::DATA_FORMAT_FLOAT32));
   vd.pushAttrib(mVAD("flags", 1, VAD::DATA_FORMAT_INT32));
   vd.pushAttrib(mVAD("normal", 3, VAD::DATA_FORMAT_FLOAT32));
   const AAT POSITION = vd.getToken("position");
   const AAT FLAGS = vd.getToken("flags");
   const AAT COLOR = vd.getToken("color");
   const AAT NORMAL = vd.getToken("normal");
   ps->clearAndSetup(1, pointCount, &vd);
   StaticVector<int32, 1> NULL_FLAG; NULL_FLAG[0] = 0;

   for (card32 n=0; n<pointCount; ++n)
   { // store points
         ps->set3f(n , POSITION, position[n]);

         float32 graylevel=rdata[n];
			if (graylevel>=0.0f)
				ps->set3f(n , COLOR, makeVector3f(0,graylevel,0) );
			else
				ps->set3f(n , COLOR, makeVector3f(-graylevel,0,0) );
         ps->set1i(n , FLAGS, NULL_FLAG);
         ps->set3f(n , NORMAL, normalize( position[n]) );
   }



  UnstructuredInCorePointCloud* result= new UnstructuredInCorePointCloud();
  result->setPointSet(ps);

	return result;
}


/*
PointCloud* S2kit::createPointCloudFromS2File(const char* file){
  double *rdata, *idata ;
  double x;
  card32 bw=64;
  card32 size = 2*bw;


  rdata = new double[size * size];
  idata = new double[size * size];

  card32 pointCount = size * size;


  std::ifstream in(file);

  for(card32 i = 0 ; i < pointCount ; i++ )
    {
      x4Assert(!in.eof());

      in >> x; rdata[i]=x;

    //  in >> x; idata[i]=x;
    }

 // x4Assert(in.eof());                                        
  //generate position
  std::vector<Vector3f> position;
  position.resize(pointCount);

  float32 phi_k, theta_j;

  card32 at=0;
  for(card32 k=0; k<size; k++)
    for (card32 j=0; j<size; j++){
        theta_j = PI*(float)(2*j+1)/(float)(4*bw);
        phi_k = 2*PI*(float)k/(float)(2*bw);

        position[at]=makeVector3f(sin(phi_k)*sin(theta_j),
               sin(phi_k)*cos(theta_j),
               cos(phi_k));
        at++;
  }


  PointSet *ps = new PointSet();
  VertexDescriptor vd;
   vd.pushAttrib(mVAD("position", 3, VAD::DATA_FORMAT_FLOAT32));
   vd.pushAttrib(mVAD("color", 3, VAD::DATA_FORMAT_FLOAT32));
   vd.pushAttrib(mVAD("flags", 1, VAD::DATA_FORMAT_INT32));
   vd.pushAttrib(mVAD("normal", 3, VAD::DATA_FORMAT_FLOAT32));
   const AAT POSITION = vd.getToken("position");
   const AAT FLAGS = vd.getToken("flags");
   const AAT COLOR = vd.getToken("color");
   const AAT NORMAL = vd.getToken("normal");
   ps->clearAndSetup(1, pointCount, &vd);
   StaticVector<int32, 1> NULL_FLAG; NULL_FLAG[0] = 0;

   for (card32 n=0; n<pointCount; ++n)
   { // store points
         ps->set3f(n , POSITION, position[n]);

         float32 graylevel=3.0*fabs(rdata[n]);
			if (rdata[n]!=0.0f)
				ps->set3f(n , COLOR, makeVector3f(graylevel,graylevel,graylevel) );
			else
				ps->set3f(n , COLOR, makeVector3f(1,0,0) );
         ps->set1i(n , FLAGS, NULL_FLAG);
         ps->set3f(n , NORMAL, normalize( position[n]) );
   }




  //get rdata into UnstructuredInCorePointCloud ....
  UnstructuredInCorePointCloud* result= new UnstructuredInCorePointCloud();
  result->setPointSet(ps);

  delete(rdata);
  delete(idata);
  return result;
}
*/


void S2kit::test(){
    double *rcoeffs, *icoeffs, *rdata, *idata=NULL;
    card32 bw=101;
    card32 size = 2*bw;
    rcoeffs = new double[bw * bw];
    icoeffs = new double[bw * bw];
    rdata = new double[size * size];
    idata =  new double[size * size];
    x4Assert(rcoeffs&&icoeffs && rdata&&idata);

    for (card32 i=0; i<(size * size); i++){
          rdata[i]=0.8;
          idata[i]=0.0;
          rcoeffs[i/4]=0.0;
          icoeffs[i/4]=0.0;
    }

    const int dim=80;

        // StaticVector is best to view 
          StaticVector<double,dim>  rc=makeVector<double,dim>(rcoeffs);
          StaticVector<double,dim>  ic=makeVector<double,dim>(icoeffs);
          StaticVector<double,dim>  rd=makeVector<double,dim>(rdata);
          StaticVector<double,dim>  id=makeVector<double,dim>(idata);

    doSHT(rdata,idata,rcoeffs ,icoeffs ,bw, false);

          rc=makeVector<double,dim>(rcoeffs);
          ic=makeVector<double,dim>(icoeffs);
          rd=makeVector<double,dim>(rdata);
          id=makeVector<double,dim>(idata);

    doSHT(rdata,idata,rcoeffs ,icoeffs ,bw, true);

          rc=makeVector<double,dim>(rcoeffs);
          ic=makeVector<double,dim>(icoeffs);
          rd=makeVector<double,dim>(rdata);
          id=makeVector<double,dim>(idata);

    for (card32 i=0; i<(size * size); i++){
          rdata[i]=0.8;
          idata[i]=0.0;
          rcoeffs[i/4]=0.0;
          icoeffs[i/4]=0.0;
    }

    doForwardMemSHT (rdata,idata,rcoeffs ,icoeffs ,bw);

          rc=makeVector<double,dim>(rcoeffs);
          ic=makeVector<double,dim>(icoeffs);
          rd=makeVector<double,dim>(rdata);
          id=makeVector<double,dim>(idata);

    doInverseMemSHT (rdata,idata,rcoeffs ,icoeffs ,bw);

          rc=makeVector<double,dim>(rcoeffs);
          ic=makeVector<double,dim>(icoeffs);
          rd=makeVector<double,dim>(rdata);
          id=makeVector<double,dim>(idata);


    delete (idata);
    delete (rdata);
    delete (icoeffs);
    delete (rcoeffs);
}



void S2kit::doForwardMemSHT(double *dataReal, double *dataImg, double *coeffsReal, double *coeffsImg, card32 bw){
  /*** ASSUMING WILL SEMINAIVE ALL ORDERS ***/
  int cutoff = bw ;
 // int size = 2*bw;
  /* allocate memory */
  double* weights = (double *) malloc(sizeof(double) * 4 * bw);
  double* seminaive_naive_tablespace =
    (double *) malloc(sizeof(double) *
		      (Reduced_Naive_TableSize(bw,cutoff) +
		       Reduced_SpharmonicTableSize(bw,cutoff)));
  double* workspace = (double *) malloc(sizeof(double) *
				((8 * (bw*bw)) +
				 (7 * bw)));

  /* now precompute the Legendres */
  fprintf(stdout,"Generating seminaive_naive tables...\n");
  double **seminaive_naive_table = SemiNaive_Naive_Pml_Table(bw, cutoff,
						    seminaive_naive_tablespace,
                                                    workspace);

  /* construct fftw plans */

  /* make DCT plan -> note that I will be using the GURU
     interface to execute these plans within the routines*/

  /* forward DCT */
  fftw_plan dctPlan = fftw_plan_r2r_1d( 2*bw, weights, dataReal,
			      FFTW_REDFT10, FFTW_ESTIMATE ) ;

  /*
    fftw "preamble" ;
    note that this plan places the output in a transposed array
  */
  int rank, howmany_rank ;
  fftw_iodim dims[1], howmany_dims[1];
  rank = 1 ;
  dims[0].n = 2*bw ;
  dims[0].is = 1 ;
  dims[0].os = 2*bw ;
  howmany_rank = 1 ;
  howmany_dims[0].n = 2*bw ;
  howmany_dims[0].is = 2*bw ;
  howmany_dims[0].os = 1 ;

  /* forward fft */
  fftw_plan fftPlan = fftw_plan_guru_split_dft( rank, dims,
				      howmany_rank, howmany_dims,
				      dataReal, dataImg,
				      workspace, workspace+(4*bw*bw),
				      FFTW_ESTIMATE );


  /* now make the weights */
  makeweights( bw, weights );						    

  /* now do the forward spherical transform */
  FST_semi_memo(dataReal, dataImg,
		coeffsReal, coeffsImg,
		bw,
		seminaive_naive_table,
		workspace,
		0,
		cutoff,
		&dctPlan,
		&fftPlan,
		weights );





  /* clean up */
  fftw_destroy_plan( fftPlan );
  fftw_destroy_plan( dctPlan );

  free(workspace);
  free(seminaive_naive_table);
  free(seminaive_naive_tablespace);
  free(weights);

}

void S2kit::doInverseMemSHT(double *dataReal, double *dataImg, double *coeffsReal, double *coeffsImg, card32 bw){
  /*** ASSUMING WILL SEMINAIVE ALL ORDERS ***/
  int cutoff = bw ;
 // int size = 2*bw;

  /* allocate lots of memory */
  double *weights = (double *) malloc(sizeof(double) * 4 * bw);

  double *workspace = (double *) malloc(sizeof(double) *
				((8 * (bw*bw)) +
				 (10 * bw)));

  double *seminaive_naive_tablespace =
    (double *) malloc(sizeof(double) *
		      (Reduced_Naive_TableSize(bw,cutoff) +
		       Reduced_SpharmonicTableSize(bw,cutoff)));

  double *trans_seminaive_naive_tablespace =
    (double *) malloc(sizeof(double) *
		      (Reduced_Naive_TableSize(bw,cutoff) +
		       Reduced_SpharmonicTableSize(bw,cutoff)));

  /* now precompute the Legendres */
  double **seminaive_naive_table = SemiNaive_Naive_Pml_Table(bw, cutoff,
						    seminaive_naive_tablespace,
						    workspace);

  double **trans_seminaive_naive_table =
    Transpose_SemiNaive_Naive_Pml_Table(seminaive_naive_table,
					bw, cutoff,
					trans_seminaive_naive_tablespace,
					workspace);

  /* construct fftw plans */

  /* make iDCT plan -> note that I will be using the GURU
     interface to execute this plan within the routine*/
      
  /* inverse DCT */
  fftw_plan idctPlan = fftw_plan_r2r_1d( 2*bw, weights, dataReal,
			       FFTW_REDFT01, FFTW_ESTIMATE );

  /*
    now plan for inverse fft - note that this plans assumes
    that I'm working with a transposed array, e.g. the inputs
    for a length 2*bw transform are placed every 2*bw apart,
    the output will be consecutive entries in the array
  */
  int rank, howmany_rank ;
  fftw_iodim dims[1], howmany_dims[1];
  rank = 1 ;
  dims[0].n = 2*bw ;
  dims[0].is = 2*bw ;
  dims[0].os = 1 ;
  howmany_rank = 1 ;
  howmany_dims[0].n = 2*bw ;
  howmany_dims[0].is = 1 ;
  howmany_dims[0].os = 2*bw ;

  /* inverse fft */
  fftw_plan ifftPlan = fftw_plan_guru_split_dft( rank, dims,
				       howmany_rank, howmany_dims,
				       dataReal, dataImg,
				       workspace, workspace+(4*bw*bw),
				       FFTW_ESTIMATE );

  /* now make the weights */
  makeweights( bw, weights );

  
  /* do the inverse spherical transform */
  InvFST_semi_memo(coeffsReal, coeffsImg,
		   dataReal, dataImg,
		   bw,
		   trans_seminaive_naive_table,
		   workspace,
		   0,
		   cutoff,
		   &idctPlan,
		   &ifftPlan );

  /* now clean up */

  fftw_destroy_plan( ifftPlan );
  fftw_destroy_plan( idctPlan );

  free(trans_seminaive_naive_table);
  free(seminaive_naive_table);
  free(trans_seminaive_naive_tablespace);
  free(seminaive_naive_tablespace);
  free(workspace);
  free(weights);
}


void S2kit::doSHT(double *dataReal, double *dataImg, double *coeffsReal, double *coeffsImg, card32 bw, bool inverse){
  /*** ASSUMING WILL SEMINAIVE ALL ORDERS ***/
  int cutoff = bw ;

  /* make array for weights */
  double *weights = (double *) malloc(sizeof(double) * 4 * bw);

    /* forward DCT */
  fftw_plan dctPlan = fftw_plan_r2r_1d( 2*bw, weights, dataReal,
		          FFTW_REDFT10, FFTW_ESTIMATE ) ;

  /* inverse DCT */
  fftw_plan idctPlan = fftw_plan_r2r_1d( 2*bw, weights, dataReal,
		           FFTW_REDFT01, FFTW_ESTIMATE );

  /*
    fftw "preamble" ;
    note that this plan places the output in a transposed array
  */
  fftw_iodim dims[1], howmany_dims[1];
  int rank = 1 ;
  dims[0].n = 2*bw ;
  dims[0].is = 1 ;
  dims[0].os = 2*bw ;
  int howmany_rank = 1 ;
  howmany_dims[0].n = 2*bw ;
  howmany_dims[0].is = 2*bw ;
  howmany_dims[0].os = 1 ;
  double *workspace= (double *) malloc(sizeof(double) * 
			    ((10 * (bw*bw)) + 
			     (24 * bw)));
  
  /* forward fft */
  fftw_plan fftPlan = fftw_plan_guru_split_dft( rank, dims,
			          howmany_rank, howmany_dims,
			          dataReal, dataImg,
			          workspace, workspace+(4*bw*bw),
			          FFTW_ESTIMATE );
  
  /*
    now plan for inverse fft - note that this plans assumes
    that I'm working with a transposed array, e.g. the inputs
    for a length 2*bw transform are placed every 2*bw apart,
    the output will be consecutive entries in the array
  */
  rank = 1 ;
  dims[0].n = 2*bw ;
  dims[0].is = 2*bw ;
  dims[0].os = 1 ;
  howmany_rank = 1 ;
  howmany_dims[0].n = 2*bw ;
  howmany_dims[0].is = 1 ;
  howmany_dims[0].os = 2*bw ;

  /* inverse fft */
  fftw_plan ifftPlan = fftw_plan_guru_split_dft( rank, dims,
			           howmany_rank, howmany_dims,
			           dataReal, dataImg,
			           workspace, workspace+(4*bw*bw),
			           FFTW_ESTIMATE );

  /* now make the weights */
  makeweights( bw, weights );

  if(inverse)
      InvFST_semi_fly(coeffsReal, coeffsImg,
	        dataReal, dataImg,
	        bw,
	        workspace,
	        0,
	        cutoff,
	        &idctPlan,
	        &ifftPlan );

  else
      FST_semi_fly(dataReal, dataImg,
	     coeffsReal, coeffsImg,
	     bw,
	     workspace,
	     0,
	     cutoff,
	     &dctPlan,
	     &fftPlan,
	     weights ) ;


   /* destroy fftw plans */
  fftw_destroy_plan( ifftPlan );
  fftw_destroy_plan( fftPlan );
  fftw_destroy_plan( idctPlan );
  fftw_destroy_plan( dctPlan );

  /* free memory */
  free(weights );
  free(workspace);
}
