//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "SurfaceElements.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "SurfaceElementsInteractionContext.h"
#include "surfaceelementsviewers.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

IMPLEMENT_X4_ABSTRACT_CLASS( SurfaceElement ,0)
{
	BEGIN_CLASS_INIT( SurfaceElement );
}

SurfaceElement::SurfaceElement() {}

SurfaceElement::~SurfaceElement() {}

void SurfaceElement::assign(const Object* obj, X4CopyContext *context) {
   copyObjectProperties(obj, this);
}


// ----


IMPLEMENT_X4_ABSTRACT_CLASS( SurfaceElementPoint ,0)
{
   BEGIN_CLASS_INIT( SurfaceElementPoint );
}

SurfaceElementPoint::SurfaceElementPoint() {}

SurfaceElementPoint::~SurfaceElementPoint() {}


Vector3f SurfaceElementPoint::getCenter()
{
   return getPoint();
}

BoundingBox3f SurfaceElementPoint::getBoundingBox()
{
   return BoundingBox3f(getPoint(), getPoint());
}

SurfaceElementViewer *SurfaceElementPoint::createDefaultViewer(bool viewerOwnsModel)
{
   SurfaceElementViewerPointAsSphere *result = new SurfaceElementViewerPointAsSphere();
   result->setup(this, NULL, viewerOwnsModel);
   return result;
}




// ----


IMPLEMENT_X4_CLASS( SurfaceElementVertex ,0)
{
   BEGIN_CLASS_INIT( SurfaceElementVertex );
}

SurfaceElementVertex::SurfaceElementVertex() {}

SurfaceElementVertex::~SurfaceElementVertex() {}



void SurfaceElementVertex::setup( UnstructuredInCorePointCloud *referencePC, mpcard index )
{
   this->referencePC = referencePC;
   this->index = index;
   positionAAT = referencePC->getPointSet()->getAAT("position", 3, VAD::DATA_FORMAT_FLOAT32);
   if (positionAAT == NULL_AAT) {
      throw ETypeMismatch("SurfaceElementVertex::setup() - no position channel found.");
   }
   ps = referencePC->getPointSet();
}

Vector3f SurfaceElementVertex::getPoint()
{
   if (index > ps->getNumEntries()) throw EInvalidState("SurfaceElementVertex::getPoint() - index out of range");
   return ps->get3f(index, positionAAT);
}


// ----


IMPLEMENT_X4_CLASS( SurfaceElementFreePoint ,0)
{
   BEGIN_CLASS_INIT( SurfaceElementFreePoint );
   ADD_VECTOR3F_PROP(point, 0)
}

SurfaceElementFreePoint::SurfaceElementFreePoint() {}

SurfaceElementFreePoint::~SurfaceElementFreePoint() {}



void SurfaceElementFreePoint::setup(Vector3f point)
{
   this->point = point;
}

Vector3f SurfaceElementFreePoint::getPoint()
{
   return point;
}

bool SurfaceElementPoint::mouseInRange(int32 x, int32 y, SurfaceElementsInteractionContext *context, float32 *dist, float32 *depth, bool importantElement)
{
   float32 qdepth;
   float32 qscreenDist;
   context->rayQuery(getPoint(), x,y, qdepth, qscreenDist);
   if (importantElement) {
      qscreenDist /= 4.0f;
      qdepth -= 4*context->worldSpaceSnap;
   }
   if (dist) *dist = qscreenDist;
   if (depth) *depth = qdepth;
   return context->pickDistance >= qscreenDist;
}



// ----


IMPLEMENT_X4_CLASS( SurfaceElementPolygonPoint ,0)
{
   BEGIN_CLASS_INIT( SurfaceElementPolygonPoint );
}

SurfaceElementPolygonPoint::SurfaceElementPolygonPoint() {viewerHint = NULL; parent = NULL;}

SurfaceElementPolygonPoint::~SurfaceElementPolygonPoint() {}



void SurfaceElementPolygonPoint::setup(Vector3f point, SurfaceElementPolygon *parent)
{
   SurfaceElementFreePoint::setup(point);
   this->parent = parent;
}




// ----


IMPLEMENT_X4_ABSTRACT_CLASS( SurfaceElementAbstractPolygon ,0)
{
   BEGIN_CLASS_INIT( SurfaceElementAbstractPolygon );
}


// ----


IMPLEMENT_X4_CLASS( SurfaceElementPolygon ,0)
{
   BEGIN_CLASS_INIT( SurfaceElementPolygon );
}

SurfaceElementPolygon::SurfaceElementPolygon() {}

SurfaceElementPolygon::~SurfaceElementPolygon() {}

X4::Vector3f X4::SurfaceElementPolygon::getCenter()
{
   return getBoundingBox().getCenter();
}

void X4::SurfaceElementPolygon::write( OutputObjectStream *s ) const
{
   SurfaceElement::write(s);
   s->writeMaxPlatformCard(points.size());
   for (mpcard i=0; i<points.size(); i++) {
      s->write(points[i][0]);
      s->write(points[i][1]);
      s->write(points[i][2]);
   }
}

void X4::SurfaceElementPolygon::read( InputObjectStream *s )
{
   SurfaceElement::read(s);
   mpcard size;
   s->readMaxPlatformCard(size);
   points.resize(size);
   for (mpcard i=0; i<points.size(); i++) {
      s->read(points[i][0]);
      s->read(points[i][1]);
      s->read(points[i][2]);
   }
}

X4::mpint X4::SurfaceElementPolygon::getIndex( Vector3f point )
{
   for (size_t i=0; i<points.size(); i++) {
      if (points[i] == point) return i;
   }
   return -1;
}

void X4::SurfaceElementPolygon::subdivideEdge( mpint index )
{
   if (index >= (mpint)points.size() || index < 0) return;
   if (points.size() == 1) {
      points.push_back(points.front());
      return;
   }
   if (index == points.size()-1) index = points.size()-2;
   std::vector<Vector3f>::iterator pos = points.begin();
   pos += index+1;
   points.insert(pos, (points[index]+points[index+1])*0.5f);
}

void X4::SurfaceElementPolygon::deleteVertex( mpint index )
{
   if (index >= (mpint)points.size() || index < 0) return;
   std::vector<Vector3f>::iterator pos = points.begin();
   pos += index;
   points.erase(pos);
}

SurfaceElementPolygon * X4::SurfaceElementPolygon::createTransformedCopy( Matrix4f matrix )
{
   SurfaceElementPolygon *copy = (SurfaceElementPolygon*)getInstanceClass()->newInstance();
   for (size_t i=0; i<points.size(); i++) {
      copy->points.push_back(transformVector3f(matrix, points[i]));
   }
   return copy;
}

SurfaceElementViewer *SurfaceElementPolygon::createDefaultViewer(bool viewerOwnsModel)
{
   SurfaceElementViewerPolyLine *result = new SurfaceElementViewerPolyLine();
   result->setup(this, NULL, viewerOwnsModel);
   return result;
}

BoundingBox3f SurfaceElementPolygon::getBoundingBox()
{
   if (points.empty()) return BoundingBox3f(NULL_VECTOR3F,NULL_VECTOR3F);
   BoundingBox3f bb(points[0]);
   for (size_t i=1; i<points.size(); i++)
   {
      bb.addPoint(points[i]);
   }
   return bb;
}

bool SurfaceElementPolygon::mouseInRange(int32 x, int32 y, SurfaceElementsInteractionContext *context, 
   float32 *dist, float32 *depth, bool importantElement)
{
   return false;
}


// ----


IMPLEMENT_X4_CLASS( SurfaceElementPolygonAnnotated ,0)
{
   BEGIN_CLASS_INIT( SurfaceElementPolygonAnnotated );
   ADD_INT32_PROP(id, 0)
   ADD_VECTOR4F_PROP(color, 0)
}

SurfaceElementViewer *SurfaceElementPolygonAnnotated::createDefaultViewer(bool viewerOwnsModel)
{
   SurfaceElementViewerPolyLineAnnotated *result = new SurfaceElementViewerPolyLineAnnotated();
   result->setup(this, NULL, viewerOwnsModel);
   return result;
}


void X4::SurfaceElementPolygonAnnotated::write( OutputObjectStream *s ) const
{
   SurfaceElementPolygon::write(s);
   s->writeMaxPlatformCard(annotations.size());
   for (mpcard i=0; i<annotations.size(); i++) {
      s->write(annotations[i][0]);
      s->write(annotations[i][1]);
      s->write(annotations[i][2]);
   }

}

void X4::SurfaceElementPolygonAnnotated::read( InputObjectStream *s )
{
   SurfaceElementPolygon::read(s);
   mpcard size;
   s->readMaxPlatformCard(size);
   annotations.resize(size);
   for (mpcard i=0; i<annotations.size(); i++) {
      s->read(annotations[i][0]);
      s->read(annotations[i][1]);
      s->read(annotations[i][2]);
   }

}

SurfaceElementPolygon * X4::SurfaceElementPolygonAnnotated::createTransformedCopy( Matrix4f matrix )
{
   SurfaceElementPolygonAnnotated *copy = dynamic_cast<SurfaceElementPolygonAnnotated*>(SurfaceElementPolygon::createTransformedCopy(matrix));
   copy->id = id;
   copy->color = color;
   for (size_t i=0; i<annotations.size(); i++) {
      copy->annotations.push_back(transformVector3f(matrix, annotations[i]));
   }
   return copy;

}


// ----


IMPLEMENT_X4_CLASS( SurfaceElementMeshPolygon ,0)
{
   BEGIN_CLASS_INIT( SurfaceElementMeshPolygon );
}




X4::SurfaceElementMeshPolygon::SurfaceElementMeshPolygon()
{
   basePC = NULL;
   positionAAT = NULL_AAT;
}

void X4::SurfaceElementMeshPolygon::write( OutputObjectStream *s ) const
{
   SurfaceElementAbstractPolygon::write(s);
   s->writeMaxPlatformCard(indices.size());
   for (size_t i=0; i<indices.size(); i++) {
      s->writeMaxPlatformCard(indices[i]);
   }
   s->writeObject(basePC);
}

void X4::SurfaceElementMeshPolygon::read( InputObjectStream *s )
{
   SurfaceElementAbstractPolygon::read(s);
   mpcard size;
   s->readMaxPlatformCard(size);
   indices.resize(size);
   for (mpcard i=0; i<size; i++) {
      s->readMaxPlatformCard(indices[i]);
   }
   s->readAnyObject(basePC);
   positionAAT = basePC->getAAT("position");
}

Vector3f X4::SurfaceElementMeshPolygon::getPoint( mpcard i )
{
   return basePC->getPointSet()->get3f(indices[i], positionAAT);
}

X4::Vector3f X4::SurfaceElementMeshPolygon::getCenter()
{
   return getBoundingBox().getCenter();
}

X4::BoundingBox3f X4::SurfaceElementMeshPolygon::getBoundingBox()
{
   return basePC->getPointSet()->getBoundingBox();
}

SurfaceElementViewer * X4::SurfaceElementMeshPolygon::createDefaultViewer( bool viewerOwnsModel )
{
   SurfaceElementViewerPolyLine *result = new SurfaceElementViewerPolyLine();
   result->setShowVertices(false);
   result->setup(this, NULL, viewerOwnsModel);
   return result;   
}

bool X4::SurfaceElementMeshPolygon::mouseInRange( int32 x, int32 y, SurfaceElementsInteractionContext *context, float32 *dist, float32 *depth, bool importantElement )
{
   return false;
}

X4::SurfaceElementMeshPolygon::~SurfaceElementMeshPolygon() {}

void X4::SurfaceElementMeshPolygon::addIndices( const std::vector<int32> &indices )
{
   mpint start = 0;
   if (!indices.empty() && !this->indices.empty()) {
      if (indices.front() == this->indices.back()) start = 1;
   }
   for (mpint i=start; i<((mpint)indices.size()); i++) {
      this->indices.push_back(indices[i]);
   }
}

void X4::SurfaceElementMeshPolygon::setup( UnstructuredInCorePointCloud *basePC )
{
   this->basePC = basePC;
   this->positionAAT = basePC->getAAT("position");
   if (positionAAT == NULL_AAT) {
      x4Assert(false);
   }
}

bool X4::SurfaceElementPoint::pointInRange( Vector3f p, float32 radius )
{
   return norm(p-getPoint()) <= radius;
}

bool X4::SurfaceElementAbstractPolygon::pointInRange( Vector3f p, float32 radius )
{
   mpcard n = getNumPoints();
   for (mpcard i=0; i<n; i++) {
      if (norm(p-getPoint(i)) <= radius) return true;
   }
   return false;
}
