//---------------------------------------------------------------------------
#ifndef SurfaceElementH
#define SurfaceElementH
//---------------------------------------------------------------------------
#include "x3editor2.h"
//---------------------------------------------------------------------------
#include "AttachedData.h"
//---------------------------------------------------------------------------

namespace X4 {

class SurfaceElementViewer;
class SurfaceElementsInteractionContext;


/**
   An abstract surface structure element.
*/
class X3EDITOR2_API SurfaceElement : public Persistent {
	X4_ABSTRACT_CLASS( SurfaceElement )
 public:
	SurfaceElement();
		
   virtual void assign(const Object* obj, X4CopyContext *context = NULL);
   virtual SurfaceElementViewer *createDefaultViewer(bool viewerOwnsModel = false) = 0;


   virtual Vector3f getCenter() = 0;
   virtual BoundingBox3f getBoundingBox() = 0;
   virtual bool pointInRange(Vector3f p, float32 radius) = 0;
   virtual bool mouseInRange(int32 x, int32 y, SurfaceElementsInteractionContext *context, float32 *dist = NULL, float32 *depth = NULL, bool importantElement = false)  = 0;

	~SurfaceElement();
};



/**
   A point feature
*/
class X3EDITOR2_API SurfaceElementPoint : public SurfaceElement {
   X4_ABSTRACT_CLASS( SurfaceElementPoint )
 protected:
   virtual Vector3f getPoint() = 0;

 public:
   SurfaceElementPoint();

   virtual Vector3f getCenter();
   virtual BoundingBox3f getBoundingBox();
   virtual bool pointInRange(Vector3f p, float32 radius);
   virtual SurfaceElementViewer *createDefaultViewer(bool viewerOwnsModel = false);

   virtual bool mouseInRange(int32 x, int32 y, SurfaceElementsInteractionContext *context, float32 *dist = NULL, float32 *depth = NULL, bool importantElement = false);

   ~SurfaceElementPoint();
};


/**
   A mesh vertex
*/
class X3EDITOR2_API SurfaceElementVertex : public SurfaceElementPoint {
   X4_CLASS( SurfaceElementVertex )
 private:
    UnstructuredInCorePointCloud *referencePC;
    mpcard index;
    AAT positionAAT;
    PointSet *ps;
    
protected:
   virtual Vector3f getPoint();
   
 public:
   SurfaceElementVertex();
   void setup(UnstructuredInCorePointCloud *referencePC, mpcard index);

   UnstructuredInCorePointCloud *getReferencePC() const {return referencePC;}
   mpcard getIndex() const {return index;}

   ~SurfaceElementVertex();
};


/**
   A free point in space
*/
class X3EDITOR2_API SurfaceElementFreePoint : public SurfaceElementPoint {
   X4_CLASS( SurfaceElementFreePoint )
 private:
   Vector3f point;
 public:
   SurfaceElementFreePoint();
   void setup(Vector3f point);

   virtual Vector3f getPoint();

   ~SurfaceElementFreePoint();
};


class SurfaceElementPolygon;

/**
   A polygon vertex
*/
class X3EDITOR2_API SurfaceElementPolygonPoint : public SurfaceElementFreePoint {
   X4_CLASS( SurfaceElementPolygonPoint )
 private:
   SurfaceElementPolygon *parent;
   SurfaceElementViewer *viewerHint;

 public:
   SurfaceElementPolygonPoint();
   void setup(Vector3f point, SurfaceElementPolygon *parent);

   SurfaceElementPolygon *getParent() {return parent;}
   
   SurfaceElementViewer *getViewerHint() const { return viewerHint; }
   void setViewerHint(SurfaceElementViewer * val) { viewerHint = val; }

   ~SurfaceElementPolygonPoint();
};

///**
//   an edge
//*/
//class X3EDITOR2_API SurfaceElementEdge : public SurfaceElement {
//   X4_CLASS( SurfaceElementEdge )
//public:
//   SurfaceElementEdge();
//
//   ~SurfaceElementEdge();
//};
//
//
///**
//   a path
//*/
//class X3EDITOR2_API SurfaceElementPath : public SurfaceElement {
//   X4_CLASS( SurfaceElementPath )
//public:
//   SurfaceElementPath();
//
//   ~SurfaceElementPath();
//};

///
/// super class for polygons
///
class X3EDITOR2_API SurfaceElementAbstractPolygon : public SurfaceElement {
   X4_ABSTRACT_CLASS( SurfaceElementAbstractPolygon )
 public:
   virtual mpcard getNumPoints() = 0;
   virtual Vector3f getPoint(mpcard i) = 0;
   virtual bool pointInRange(Vector3f p, float32 radius);
};


/**
  a polygonal region with explicit vertex coordinates
*/
class X3EDITOR2_API SurfaceElementPolygon : public SurfaceElementAbstractPolygon {
  X4_CLASS( SurfaceElementPolygon )
 private:
   std::vector<Vector3f> points;
   
 public:
   SurfaceElementPolygon();

   virtual void write(OutputObjectStream *s) const;
   virtual void read(InputObjectStream *s);

   std::vector<Vector3f> &getPoints() {return points;}
   virtual mpcard getNumPoints() {return points.size();}
   virtual Vector3f getPoint(mpcard i) {return points[i];}

   virtual Vector3f getCenter();
   virtual BoundingBox3f getBoundingBox();
   virtual SurfaceElementViewer *createDefaultViewer(bool viewerOwnsModel = false);
   virtual bool mouseInRange(int32 x, int32 y, SurfaceElementsInteractionContext *context, float32 *dist = NULL, float32 *depth = NULL, bool importantElement = false);

   mpint getIndex(Vector3f point);
   void subdivideEdge(mpint index);
   void deleteVertex(mpint index);

   virtual SurfaceElementPolygon *createTransformedCopy(Matrix4f matrix);

   ~SurfaceElementPolygon();
};


/// polygon with annoations
class X3EDITOR2_API SurfaceElementPolygonAnnotated : public SurfaceElementPolygon {
   X4_CLASS( SurfaceElementPolygonAnnotated )
private:
   std::vector<Vector3f> annotations;
   int32 id;
   Vector4f color;

public:
   SurfaceElementPolygonAnnotated() {id = -1; color = NULL_VECTOR4F;}

   virtual void write(OutputObjectStream *s) const;
   virtual void read(InputObjectStream *s);

   X4::int32 getId() const { return id; }
   void setId(X4::int32 val) { id = val; }
   X4::Vector4f getColor() const { return color; }
   void setColor(X4::Vector4f val) { color = val; }

   std::vector<Vector3f> &getAnnotations() {return annotations;}

   virtual SurfaceElementViewer *createDefaultViewer(bool viewerOwnsModel = false);

   virtual SurfaceElementPolygon *createTransformedCopy(Matrix4f matrix);

   ~SurfaceElementPolygonAnnotated() {}
};


/**
  a polygonal region with vertex coordinates from a point cloud
*/
class X3EDITOR2_API SurfaceElementMeshPolygon : public SurfaceElementAbstractPolygon {
  X4_CLASS( SurfaceElementMeshPolygon )
 private:
   std::vector<mpcard> indices;
   UnstructuredInCorePointCloud *basePC;
   AAT positionAAT;
   
 public:
   SurfaceElementMeshPolygon();
   
   void setup(UnstructuredInCorePointCloud *basePC);

   virtual void write(OutputObjectStream *s) const;
   virtual void read(InputObjectStream *s);

   std::vector<mpcard> &getIndices() {return indices;}
   virtual mpcard getNumPoints() {return indices.size();}
   virtual Vector3f getPoint(mpcard i);
   
   void addIndices(const std::vector<int32> &indices);

   virtual Vector3f getCenter();
   virtual BoundingBox3f getBoundingBox();
   virtual SurfaceElementViewer *createDefaultViewer(bool viewerOwnsModel = false);
   virtual bool mouseInRange(int32 x, int32 y, SurfaceElementsInteractionContext *context, float32 *dist = NULL, float32 *depth = NULL, bool importantElement = false);

   ~SurfaceElementMeshPolygon();
};


}

#endif