//---------------------------------------------------------------------------
#ifndef SurfaceElementsInteractionToolsH
#define SurfaceElementsInteractionToolsH
//---------------------------------------------------------------------------
#include "x3editor2.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
#include "PCInteractionTool.h"
#include "SurfaceElementsViewers.h"
#include "SurfaceElements.h"
//---------------------------------------------------------------------------

namespace X4 {

class SEVContext;
class SurfaceElement;
class MeshIndexTablesHack;
class SurfaceElementPolygon;
class SurfaceElementController;
class SurfaceElementPolygonPoint;
class SurfaceElementPolygonAnnotated;
class UnstructuredInCoreTriangleMesh;
class SurfaceElementStructureAttachment;
class SurfaceElementsInteractionContext;
class SurfaceElementsPolyMeshSegmentation;
class SurfaceElementEditingEventDispatcher;



/**
  Interaction controller for a surface element
*/
class X3EDITOR2_API PCISurfaceElementInteractionTool : public PCInteractionTool {
	X4_ABSTRACT_CLASS( PCISurfaceElementInteractionTool )
 public:
   struct ClickInfo {
      float32 depth;
      float32 dist;      
   };
 protected:

   class MyRootViewer : public SurfaceElementViewer {
    private:
      SurfaceElementViewerList *meshViewer;
      SurfaceElementViewerList *tempViewer;
      SurfaceElementViewerList *resultsViewer;
    public:
       MyRootViewer();

       SurfaceElementViewerList *getMeshViewer() const { return meshViewer; }
       SurfaceElementViewerList *getTempViewer() const { return tempViewer; }
       SurfaceElementViewerList *getResultsViewer() const { return resultsViewer; }

       void addMeshViewer(SurfaceElementViewer *v);
       void addResultsViewer(SurfaceElementViewer *v);
       void addTempViewer(SurfaceElementViewer *v);
       void clearTempViewers();
       
       virtual mpcard getNumChildViewers();
       virtual SurfaceElementViewer *getChildViewer(mpcard index);

       virtual void drawGL(SEVContext *context);

       ~MyRootViewer();
   };

   struct Click {
      ClickInfo info;
      SurfaceElementController *issuer;
      SurfaceElement *element;
      inline bool operator<(const Click &other) const {
         return (info.depth < other.info.depth) || (info.depth == other.info.depth && info.dist > other.info.dist);
      }
   };
   
   void updateViewerHierarchy();
   
 private:

   // --- persistent state
   std::string meshPC;
   std::string continuousPC;
   SEVContext *visContext;
   SurfaceElementsInteractionContext *interactionContext;

   // --- volatile state (invisible)
   SurfaceElementEditingEventDispatcher *dispatcher;
   // - editing state
   SurfaceElementStructureAttachment *resultsAttachment;
   std::vector<SurfaceElement*> meshSurfaceElements;
   std::vector<SurfaceElement*> helperSurfaceElements;
   std::vector<SurfaceElement*> tempSurfaceElements;
   MyRootViewer *viewerHierarchy;
   // - instances vars
   card32 width;
   card32 height;
   std::vector<Click> clickBuffer;
   bool leftDown;
   bool camMode;

   // --- helper functions
   void doGuessClickPoint(int32 x, int32 y);
   void tryInit();
   void setMeshPC(std::string val) { meshPC = val; tryInit();}
   void setContinuousPC(std::string val) { continuousPC = val; tryInit();}
   void updateIntContext();

 protected:
   // --- member access
   SurfaceElementStructureAttachment *getResultsAttachment() {return resultsAttachment;}
   SurfaceElementEditingEventDispatcher *getDispatcher() {return dispatcher;}
   SurfaceElementsInteractionContext *getInteractionContext() {return interactionContext;}
   //MyRootViewer * getViewerHierarchy() const { return viewerHierarchy; }
   SEVContext *getVisContext() const {return visContext;}
   std::string getMeshPC() const { return meshPC; }
   std::string getContinuousPC() const { return continuousPC; }
   bool getLeftDown() const { return leftDown; }
   void setLeftDown(bool val) { leftDown = val; }
   const std::vector<SurfaceElement*> &getMeshSurfaceElements() const {return meshSurfaceElements;}
   const std::vector<SurfaceElement*> &getHelperSurfaceElements() const {return meshSurfaceElements;}

   /// setup controllers & viewers
   /// override in subclass (or define custom createViewersControllers method)
   virtual SurfaceElementController *createLocalControler(SurfaceElement *element) const = 0;
   /// create a viewer hierarchy with attached controllers
   virtual SurfaceElementViewer *createViewerControllers(std::vector<SurfaceElement*> &elementList, bool ownTheElements, bool lightWeightVis = false) const;
   /// add a viewer for a a new mesh element (element owned by this object) (no update of viewers/controllers)
   virtual void addHelperElementWithViewer(SurfaceElement *element);
   /// explicit update of controller list in dispatcher by viewers 
   virtual void updateAllControllers();
   /// add a viewer for a result just created (model not owned) and update controller list (implicit update of viewers/ctrls)
   virtual void addResultsViewer(SurfaceElement *element);
   /// add helper viewer (model not owned) and update controller list (implicit update of viewers/ctrls)
   virtual void addTempViewer(SurfaceElement *element);
   /// add only viewer, nothing else
   virtual void addTempViewerOnly(SurfaceElementViewer *viewer);
   /// remove temporary viewers and update controller list
   virtual void clearTempViewers();

  
   // ---- click management
   void newClicks();
   void addClick(const Click &c);
   Click *getTopClick();

 public:
	PCISurfaceElementInteractionTool();
			
   virtual void assign(const Object* obj, X4CopyContext *context = NULL);

   /// add a (set of) surface elements for editing
   virtual void addMeshSurfaceElement(SurfaceElement *editMe) {meshSurfaceElements.push_back(editMe);}
   virtual void addTempSurfaceElement(SurfaceElement *editMe) {tempSurfaceElements.push_back(editMe);}
   virtual void clearSurfaceElements() {meshSurfaceElements.clear();}

   // --- external (PCI) interface
   virtual void keyDown(GeneralKey key);
   virtual void keyUp(GeneralKey key);

   virtual void mouseDown(int32 x, int32 y, MouseButtons buttonsState, Modifiers modifiersState);
   virtual void mouseMoved(int32 x, int32 y);
   virtual void mouseUp(int32 x, int32 y, MouseButtons buttonsState, Modifiers modifiersState);
   virtual void mouseWheelRotated(int32 rotatedDelta, Modifiers modifiersState);

   virtual void areaResize(card32 width, card32 height);

   // --- internal (controller) interface
   static const card32 TYPE_DOWN = 0;
   static const card32 TYPE_MOVE = 1;
   static const card32 TYPE_UP = 2;
   virtual void startInternalEvent(Modifiers modifiers, card32 type) {}
   virtual bool tryFreeClickPoint(Vector3f surfPoint) {return false;}
   virtual bool tryFreeHoverPoint(Vector3f surfPoint, bool hit) {return false;}
   virtual void tryHoverElement(SurfaceElementController *issuer, SurfaceElement *element, ClickInfo &info) {}
   virtual void tryClickElement(SurfaceElementController *issuer, SurfaceElement *element, ClickInfo &info) {}
   virtual void dragElement(SurfaceElementController *issuer, SurfaceElement *element, Vector3f oldPos, Vector3f newPos) {}
   virtual void rightClick() {}
   virtual void endInternalEvent(card32 type) {}
   
   virtual void glDrawTool(GLContext *glContext);

	~PCISurfaceElementInteractionTool();
};


/// abstract selection tool
class X3EDITOR2_API PCISESelection : public PCISurfaceElementInteractionTool {
   X4_ABSTRACT_CLASS(PCISESelection)
 private:
   Vector3f lastSurfacePoint;
   bool surfaceHit;
   bool hoverHit;
   bool clicked;
   bool freePointClicked;
   Modifiers modifiers;
   
 protected:
   /// override in subclass (or define custom createViewersControllers method)
   virtual SurfaceElementController *createLocalControler(SurfaceElement *element) const;
   
   // -- Selection interface
   virtual void processHoverPoint(Click *click) = 0;
   virtual void processDragPoint(Vector3f where) = 0;
   virtual void processClickPoint(Click *click) = 0;
   virtual void processFreeClickPoint(Vector3f point) = 0;
   virtual void canceled() {}
   Modifiers getModifiers() {return modifiers;}

 public:
   PCISESelection();
   
   virtual void startInternalEvent(Modifiers modifiers, card32 type);
   virtual bool tryFreeHoverPoint(Vector3f surfPoint, bool hit);
   virtual bool tryFreeClickPoint(Vector3f surfPoint);
   virtual void tryHoverElement(SurfaceElementController *issuer, SurfaceElement *element, ClickInfo &info);
   virtual void tryClickElement(SurfaceElementController *issuer, SurfaceElement *element, ClickInfo &info);
   virtual void rightClick();
   virtual void endInternalEvent(card32 type);

   virtual void glDrawTool(GLContext *glContext);
};


/// vertex selection tool
class X3EDITOR2_API PCISEPointSelection : public PCISESelection {
   X4_CLASS(PCISEPointSelection)
 private:
 protected:
   // -- Selection interface
   virtual void processHoverPoint(Click *click);
   virtual void processDragPoint(Vector3f where) {}
   virtual void processClickPoint(Click *click);
   virtual void processFreeClickPoint(Vector3f point);

public:
   PCISEPointSelection();
};


/// polygon spanning tool
class X3EDITOR2_API PCISEPolygonRegionTool : public PCISESelection {
   X4_CLASS(PCISEPolygonRegionTool)
 private:
   // -- persistent state
   card32 mode;
   int32 regionID;
   int32 autoColorNumIDs;
   Vector4f regionColor;
   SurfaceElementsPolyMeshSegmentation *segmentation;
   Vector3f upDirection;
   float32 symSearchRadius;
   float32 symAngleThreshold;
   float32 lineFeatureCosAngle;
   std::string cutOutMesh;
   float32 lineFeatureSamplingFact;
   float32 lfDistFactor;
   float32 noOverlapTest;
   float32 minCoverageFact;
   float32 lineTestDownSample;
   float32 diagTemplateBorderFact;
   
   // -- volatile state
   SurfaceElementPolygonAnnotated *currentPoly;
   SurfaceElementPolygon *selectedPoly;
   SurfaceElementPolygonPoint *selectedPoint;
   SurfaceElementViewer *selectedViewer;

   void changeMode();

   void addIntersections();
   void changeRegionID();
   void deletePoly();
   void subdivideEdge();
   void subdivideEdgeAllRegions();
   void removeVertex();
   void removeVertexAllRegions();   
   void findAllSimilar();
   void matchLineFeatures();
   void cutRegions();

   void regionIDChanged();
   
 protected:
   // -- Selection interface
   virtual void processHoverPoint(Click *click);
   virtual void processDragPoint(Vector3f where);   
   virtual void processClickPoint(Click *click);
   virtual void processFreeClickPoint(Vector3f point);
   virtual void canceled();
   
   struct DescriptorContext {
      Vector3f upDirection;
      float32 angleThreshold;
      float32 lineFeatureCosAngle;
      DescriptorContext(Vector3f upDirection, float32 angleThreshold, float32 lineFeatureCosAngle) {
         this->upDirection = upDirection;
         this->angleThreshold = angleThreshold;
         this->lineFeatureCosAngle = lineFeatureCosAngle;
      }
   };
   
   class Descriptor {
    private:
       //void growEdge(DescriptorContext &ctxt, MeshIndexTablesHack *table, mpcard edgeIndex, float32 &angle, float32 &length);    
       bool eq(const DescriptorContext &ctxt, const Descriptor &other) const;
    public:
      struct DEntry {
         float32 angleUX;
         float32 angleUN;
         // bool operator<(const DEntry &other) const {return angle < other.angle;}
      };
      std::vector<DEntry> entries;
      bool equal(const DescriptorContext &ctxt, const Descriptor &other) const;
      Descriptor() {}
      Descriptor(const DescriptorContext &ctxt, MeshIndexTablesHack *table, mpcard vIndex);
   };
   
   class DescriptorSet {
    private:
      std::vector<Descriptor> descriptors;
    public:
      DescriptorSet(DescriptorContext &ctxt, UnstructuredInCoreTriangleMesh *mesh, MeshIndexTablesHack *table);
      mpcard getNumDescriptors() {return descriptors.size();}
      Descriptor getDescriptor(mpcard i) {return descriptors[i];}
   };
   
 public:
   PCISEPolygonRegionTool();

   virtual void keyDown(GeneralKey key);

   virtual void glDrawTool(GLContext *glContext);
   
   ~PCISEPolygonRegionTool();
};



}

#endif