//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "SurfaceElementsViewers.h"
#include "SurfaceElements.h"
#include "SurfaceElementsControllers.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "ColorTools.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------


const float32 X4::SurfaceElementViewer::SMALL  = 0.6f;
const float32 X4::SurfaceElementViewer::NORMAL = 1.0f;
const float32 X4::SurfaceElementViewer::BIG    = 1.3f;

const Vector4f X4::SurfaceElementViewer::RED    = makeVector4f(0.7f, 0.1f, 0.1f, 1.0f);
const Vector4f X4::SurfaceElementViewer::YELLOW = makeVector4f(0.9f, 0.9f, 0.3f, 1.0f);
const Vector4f X4::SurfaceElementViewer::GREEN  = makeVector4f(0.1f, 0.7f, 0.1f, 1.0f);
const Vector4f X4::SurfaceElementViewer::BLUE   = makeVector4f(0.1f, 0.1f, 0.7f, 1.0f);

const Vector4f X4::SurfaceElementViewer::HOVER_COLOR   = makeVector4f(0.8f, 0.3f, 0.4f, 1.0f);
const Vector4f X4::SurfaceElementViewer::SURFACE_HOVER_COLOR   = makeVector4f(0.9f, 0.8f, 0.3f, 1.0f);
const Vector4f X4::SurfaceElementViewer::PLAIN_COLOR   = makeVector4f(0.7f, 0.7f, 0.7f, 1.0f);
const Vector4f X4::SurfaceElementViewer::SELECTION_COLOR   = makeVector4f(0.8f, 0.6f, 0.7f, 1.0f);

const float32 X4::SurfaceElementViewer::SELECTION_SIZE = 1.5f;
const float32 X4::SurfaceElementViewer::HOVER_SIZE = 2.0f;


IMPLEMENT_X4_CLASS( SEVContext ,0)
{
   BEGIN_CLASS_INIT( SEVContext );
   ADD_FLOAT32_PROP(visScale, 0)
   ADD_FLOAT32_PROP(worldSpaceScale, 0)
   ADD_VECTOR4F_PROP(lightPos, 0)
   ADD_VECTOR4F_PROP(ambient, 0)
   ADD_FLOAT32_PROP(specular, 0)
   ADD_FLOAT32_PROP(shininess, 0)
   ADD_BOOLEAN_PROP(quick, 0)
}

SEVContext::SEVContext()
{
   width = 0;
   height = 0;
   visScale = 5.0f;
   worldSpaceScale = 0.002f;
   lightPos = makeVector4f(0,0,1,0);
   ambient = makeVector4f(0.1f,0.1f,0.1f,1.0f);
   specular = 0.5f;
   shininess = 10;
   sphereQuadric = NULL;
   quick = false;
}

SEVContext::~SEVContext() {
   gluDeleteQuadric(sphereQuadric);
}

void X4::SEVContext::pushRenderState()
{
   glPushAttrib(GL_ALL_ATTRIB_BITS);
   glEnable(GL_DEPTH_TEST);
   glDisable(GL_BLEND);
   //if (quick) {
   //   glDisable(GL_LIGHTING);
   //} else {
   glEnable(GL_LIGHTING);
   glEnable(GL_COLOR_MATERIAL);
   glEnable(GL_LIGHT0);
//   }
   
   const Vector4f spec =  makeVector4f(specular,specular,specular,1.0f);

   glMaterialfv( GL_FRONT, GL_AMBIENT, ambient.data() );
   glMaterialfv( GL_FRONT, GL_SPECULAR, spec.data() );
   glMaterialf( GL_FRONT, GL_SHININESS, shininess );
   glLightfv( GL_LIGHT0, GL_POSITION, lightPos.data() );
   
   if (!sphereQuadric) {
      sphereQuadric = gluNewQuadric();
   }
}

void X4::SEVContext::popRenderState()
{
   glPopAttrib();
}

void SEVContext::setupUpdate( int32 height, int32 widht, GLContext *context )
{
   this->height = height;
   this->width = width;
   this->context = context;
}

// ----



IMPLEMENT_X4_ABSTRACT_CLASS( SurfaceElementViewer ,0)
{
	BEGIN_CLASS_INIT( SurfaceElementViewer );
}

SurfaceElementViewer::SurfaceElementViewer() 
{
   model = NULL;
   controller = NULL;
   mouseHover = false;
   selected = false;
   ownerOfModel = false;
   lightWeightVis = false;
}

SurfaceElementViewer::~SurfaceElementViewer() 
{
   if (ownerOfModel) delete model;
   delete controller;
}

void SurfaceElementViewer::assign(const Object* obj, X4CopyContext *context) {
   copyObjectProperties(obj, this);
}


void SurfaceElementViewer::setup( SurfaceElement *model, SurfaceElementController *controller, bool ownerOfModel)
{
   this->model = model;
   this->controller = controller;
   this->ownerOfModel = ownerOfModel;
}


// ----


IMPLEMENT_X4_CLASS( SurfaceElementViewerPointAsSphere ,0)
{
   BEGIN_CLASS_INIT( SurfaceElementViewerPointAsSphere );
}

SurfaceElementViewerPointAsSphere::SurfaceElementViewerPointAsSphere()
{
   size = SEV::NORMAL;
   color = SEV::PLAIN_COLOR;
   plainStyleScale = 1.0f;
}

SurfaceElementViewerPointAsSphere::~SurfaceElementViewerPointAsSphere() {}



void X4::SurfaceElementViewerPointAsSphere::setPlainStyleScale( float32 plainStyleScale )
{
   this->plainStyleScale = plainStyleScale;
}

void SurfaceElementViewerPointAsSphere::drawGL(SEVContext *context)
{
   Vector3f position = getModel()->getCenter();
   Vector4f col = color;
   float32 exScale = 1.0f;
   bool forceVis = false;
   if (getMouseHover()) {
      col = HOVER_COLOR;
      exScale = HOVER_SIZE;
      forceVis = true;
   } else if (getSelected()) {
      col = SELECTION_COLOR;
      exScale = SELECTION_SIZE;
      if (dynamic_cast<SurfaceElementPolygonPoint*>(getModel())) {
         exScale *= 2;
      }
      forceVis = true;
   } else {
      exScale *= plainStyleScale;
   }
   
   glColor4fv(col.data());
   if (context->quick && getLightWeightVis()) {
      if (!forceVis) return;
   }
//          glPointSize(size*context->visScale*exScale);
//          glBegin(GL_POINTS);
//          glVertex3fv(position.data());
//          glEnd();
   glMatrixMode(GL_MODELVIEW);
   glPushMatrix();
   glTranslatef(position[0], position[1], position[2]);
   gluSphere(context->sphereQuadric, size*context->visScale*context->worldSpaceScale*exScale, 5,5);
   glPopMatrix();
   
}




// ----


IMPLEMENT_X4_ABSTRACT_CLASS( SurfaceElementNullViewer, 0)
{
   BEGIN_CLASS_INIT( SurfaceElementNullViewer );
}

// ----


IMPLEMENT_X4_ABSTRACT_CLASS( SurfaceElementViewerList ,0)
{
   BEGIN_CLASS_INIT( SurfaceElementViewerList );
}

SurfaceElementViewerList::SurfaceElementViewerList()
{
}

SurfaceElementViewerList::~SurfaceElementViewerList() {
}




void SurfaceElementViewerList::addViewer(SurfaceElementViewer *childViewer)
{
   childViewers.push_back(childViewer);

}

mpcard SurfaceElementViewerList::getNumChildViewers()
{
   return childViewers.size();
}

SurfaceElementViewer *SurfaceElementViewerList::getChildViewer(mpcard index)
{
   if (index < childViewers.size()) return childViewers[index];
   warning("index out of range");
   return 0;
}

void SurfaceElementViewerList::drawGL(SEVContext *context)
{
   for (size_t i=0; i<childViewers.size(); i++)
      childViewers[i]->drawGL(context);
}


// ----


IMPLEMENT_X4_ABSTRACT_CLASS( SurfaceElementViewerPolyLine ,0)
{
   BEGIN_CLASS_INIT( SurfaceElementViewerPolyLine );
}

SurfaceElementViewerPolyLine::SurfaceElementViewerPolyLine()
{
   size = SEV::BIG;
   thickness = 5;
   color = SEV::PLAIN_COLOR;
   rainbowColor = false;
   showVertices = true;
}

SurfaceElementViewerPolyLine::~SurfaceElementViewerPolyLine() {}



void SurfaceElementViewerPolyLine::setup(
   SurfaceElement *model, SurfaceElementController *controller, bool ownerOfModel)
{
   SurfaceElementViewer::setup(model, controller, ownerOfModel);
}

void SurfaceElementViewerPolyLine::drawGL(SEVContext *context)
{
   SurfaceElementAbstractPolygon *poly = dynamic_cast<SurfaceElementAbstractPolygon*>(getModel());
   if (!poly) return;
   Vector4f col = color;
   SurfaceElementPolygonAnnotated *polyA = dynamic_cast<SurfaceElementPolygonAnnotated*>(getModel());
   if (polyA) {
      if (polyA->getId() >= 0) color = polyA->getColor();
   }
   float32 exScale = 1.0f;
   if (getMouseHover()) {
      col = HOVER_COLOR;
      exScale = HOVER_SIZE;
   } else if (getSelected()) {
      col = SELECTION_COLOR;
      exScale = SELECTION_SIZE*1.3;
   }

   mpcard N = poly->getNumPoints();
   mpcard N2 = N;
   if (N>1 && poly->getPoint(0) == poly->getPoint(N-1)) N2--;

   glColor4fv(col.data());
   glLineWidth(thickness*exScale);
   glPushAttrib(GL_ENABLE_BIT);
   glDisable(GL_LIGHTING);
   glBegin(GL_LINE_STRIP);
   for (mpcard i=0; i<N; i++) {
      glVertex3fv(poly->getPoint(i).data());
   }
   glEnd();
   glPopAttrib();
   glLineWidth(1);
//    if (context->quick) {
//       glPointSize(size*context->visScale*exScale);
//       glBegin(GL_POINTS);
//       for (size_t i=0; i<N;i++) {
//          if (rainbowColor) {
//             float32 val = (float32)i/((float32)N-1);
//             glColor4fv(rainbowGradient(val).data());
//          }
//          glVertex3fv(poly->getPoints()[i].data());
//       }
//       glEnd();
//    } else {
   glMatrixMode(GL_MODELVIEW);
   if (showVertices) for (size_t i=0; i<N;i++) {
      if (rainbowColor) {
         float32 val = (float32)i/((float32)N2-1);
         glColor4fv(rainbowGradient(val).data());
      }
      Vector3f position = poly->getPoint(i);
      glPushMatrix();
      glTranslatef(position[0], position[1], position[2]);
      gluSphere(context->sphereQuadric, size*context->visScale*context->worldSpaceScale*exScale, 5,5);
      glPopMatrix();
   }
//   }
}


// ----


IMPLEMENT_X4_ABSTRACT_CLASS( SurfaceElementViewerPolyLineAnnotated ,0)
{
   BEGIN_CLASS_INIT( SurfaceElementViewerPolyLineAnnotated );
}

void SurfaceElementViewerPolyLineAnnotated::drawGL(SEVContext *context)
{
   SurfaceElementViewerPolyLine::drawGL(context);
   SurfaceElementPolygonAnnotated *poly = dynamic_cast<SurfaceElementPolygonAnnotated*>(getModel());
   if (!poly) return;

   int closed = 1;
   if (!poly->getPoints().empty() && poly->getPoints().front() == poly->getPoints().back()) closed = 0;

   glColor4f(0.2f,0.7f,0.3f,1);
   //if (context->quick) {
   //   glPointSize(getSize()*context->visScale*1.1);
   //   glBegin(GL_POINTS);
   //   for (size_t i=0; i<poly->getAnnotations().size();i++) {
   //      if (getRainbowColor()) {
   //         float32 val = (float32)i/((float32)poly->getAnnotations().size()-closed);
   //         glColor4fv((rainbowGradient(val)*0.5f+makeVector4f(0.7f,0.7f,0.7f,1.0f)*0.5f).data());
   //      }
   //      glVertex3fv(poly->getAnnotations()[i].data());
   //   }
   //   glEnd();
   //} else {
   glMatrixMode(GL_MODELVIEW);
   for (size_t i=0; i<poly->getAnnotations().size();i++) {
      if (getRainbowColor()) {
         float32 val = (float32)i/((float32)poly->getAnnotations().size()-closed);
         glColor4fv((rainbowGradient(val)*0.5f+makeVector4f(0.7f,0.7f,0.7f,1.0f)*0.5f).data());
      }
      Vector3f position = poly->getAnnotations()[i];
      glPushMatrix();
      glTranslatef(position[0], position[1], position[2]);
      gluSphere(context->sphereQuadric, getSize()*context->visScale*context->worldSpaceScale*1.1, 5,5);
      glPopMatrix();
   }
//   }
}

X4::SurfaceElementViewerPolyLineAnnotated::SurfaceElementViewerPolyLineAnnotated()
{
   setRainbowColor(true);
}

