//---------------------------------------------------------------------------
#ifndef SurfaceElementViewersH
#define SurfaceElementViewersH
//---------------------------------------------------------------------------
#include "x3editor2.h"
//---------------------------------------------------------------------------
#include "AttachedData.h"
#include <string>
//---------------------------------------------------------------------------

namespace X4 {

class SurfaceElement;
class SurfaceElementController;


class X3EDITOR2_API SEVContext : public Persistent {
   X4_CLASS(SEVContext)
 public:
   // --- persistent
   float32 visScale;
   float32 worldSpaceScale;
   Vector4f lightPos;
   Vector4f ambient;
   
   float32 specular;
   float32 shininess;
   
   bool quick;
   
   // --- volatile
   int32 width;
   int32 height;   
   GLUquadricObj *sphereQuadric;
   GLContext *context;
   
   SEVContext();
   
   void setupUpdate(int32 height, int32 widht, GLContext *context);

   void pushRenderState();
   void popRenderState();
   
   virtual ~SEVContext();
};



/**
  Visualization for a surface element
*/
class X3EDITOR2_API SurfaceElementViewer : public Object {
	X4_ABSTRACT_CLASS( SurfaceElementViewer )
public:
   // --- global constants
   static const float32 SMALL;
   static const float32 NORMAL;
   static const float32 BIG;

   static const Vector4f RED;
   static const Vector4f YELLOW;
   static const Vector4f GREEN;
   static const Vector4f BLUE;
   
   static const Vector4f HOVER_COLOR;
   static const Vector4f PLAIN_COLOR;
   static const Vector4f SELECTION_COLOR;
   static const Vector4f SURFACE_HOVER_COLOR;

   static const float32 SELECTION_SIZE;
   static const float32 HOVER_SIZE;

 private:
   SurfaceElement *model;
   SurfaceElementController *controller;

   bool mouseHover;
   bool selected;
   bool ownerOfModel;
   bool lightWeightVis;
   
 public:
	SurfaceElementViewer();
	
   /// the model and the controller are owned by the viewer
   virtual void setup(SurfaceElement *model, SurfaceElementController *controller = NULL, bool ownerOfModel = true);

   virtual void assign(const Object* obj, X4CopyContext *context = NULL);

   bool getMouseHover() {return mouseHover;}
   bool getSelected() {return selected;}
   SurfaceElement *getModel() {return model;}
   bool getOwnerOfModel() const { return ownerOfModel; }
   void setOwnerOfModel(bool val) { ownerOfModel = val; }
   bool getLightWeightVis() const { return lightWeightVis; }
   void setLightWeightVis(bool val) { lightWeightVis = val; }
   SurfaceElementController *getController() {return controller;}

   virtual mpcard getNumChildViewers() {return 0;}
   virtual SurfaceElementViewer *getChildViewer(mpcard index) {return NULL;}

   virtual void setHoverState(bool mouseHover) {this->mouseHover = mouseHover;}
   virtual void setSelected(bool selected) {this->selected = selected;}
		
   virtual void drawGL(SEVContext *context) = 0;

	~SurfaceElementViewer();
};

typedef SurfaceElementViewer SEV;


/**
   Point-type visualization
*/
class X3EDITOR2_API SurfaceElementViewerPointAsSphere : public SurfaceElementViewer {
   X4_CLASS( SurfaceElementViewerPointAsSphere )
 private:
    float32 size;
    float32 plainStyleScale;
    Vector4f color;

 protected:
    
 public:
   SurfaceElementViewerPointAsSphere();
   
   float32 getSize() const { return size; }
   void setSize(float32 val) { size = val; }
   Vector4f getColor() const { return color; }
   void setColor(Vector4f val) { color = val; }

   void setPlainStyleScale(float32 plainStyleScale);

   virtual void drawGL(SEVContext *context);

   ~SurfaceElementViewerPointAsSphere();
};



/**
   no visualization
*/
class X3EDITOR2_API SurfaceElementNullViewer : public SurfaceElementViewer {
   X4_CLASS( SurfaceElementNullViewer )
 public:
   SurfaceElementNullViewer() {}
   
   virtual void drawGL(SEVContext *context) {}

   ~SurfaceElementNullViewer() {}
};



/**
   Aggregate visualization
*/
class X3EDITOR2_API SurfaceElementViewerList : public SurfaceElementViewer {
   X4_CLASS( SurfaceElementViewerList )
 private:
   std::vector<SurfaceElementViewer*> childViewers;   

 public:
   SurfaceElementViewerList();

   virtual void addViewer(SurfaceElementViewer *childViewer);
   virtual mpcard getNumChildViewers();
   virtual SurfaceElementViewer *getChildViewer(mpcard index);

   virtual void drawGL(SEVContext *context);

   ~SurfaceElementViewerList();
};


/**
 polygon visualization
*/
class X3EDITOR2_API SurfaceElementViewerPolyLine : public SurfaceElementViewer {
   X4_CLASS( SurfaceElementViewerPolyLine )
 private:
   float32 size;
   float32 thickness;
   Vector4f color;
   bool rainbowColor;
   bool showVertices;

 public:
   SurfaceElementViewerPolyLine();

   virtual void setup(SurfaceElement *model, SurfaceElementController *controller = NULL, bool ownerOfModel = true);

   float32 getSize() const { return size; }
   void setSize(float32 val) { size = val; }
   float32 getThickness() const { return thickness; }
   void setThickness(float32 val) { thickness = val; }
   Vector4f getColor() const { return color; }
   void setColor(Vector4f val) { color = val; }
   bool getRainbowColor() const { return rainbowColor; }
   void setRainbowColor(bool val) { rainbowColor = val; }
   bool getShowVertices() const { return showVertices; }
   void setShowVertices(bool val) { showVertices = val; }

   virtual void drawGL(SEVContext *context);

   ~SurfaceElementViewerPolyLine();
};


/**
Point-type visualization
*/
class X3EDITOR2_API SurfaceElementViewerPolyLineAnnotated : public SurfaceElementViewerPolyLine {
   X4_CLASS( SurfaceElementViewerPolyLineAnnotated )
 public:
   SurfaceElementViewerPolyLineAnnotated();

   virtual void drawGL(SEVContext *context);
};



}

#endif