//---------------------------------------------------------------------------
#include "stdafx.h"
//---------------------------------------------------------------------------
#include "ThreadCommandQueue.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------



ThreadCommandQueue::ThreadCommandQueue(card32 maxNumMsgs) {
	this->maxNumMsgs = maxNumMsgs;
	queueSlotsAvailable.release(maxNumMsgs);
}


// preconditions:
// queueMutex    messagesAvailable    answerAvailable
//   locked              0                   0           -	  server in getMessage(), will unlock soon
//   unlocked            0                   0           -    works
//   l/u                >=1						0           -    works
//   l/u                any					  >=1				-    impossible, result of wrong usage of answerMessage
// queueSlotsAvailable
//    0  - server works on message queue
//         server can be blocked by messagesAvailable, but only if queue empty (in that case queueSlotsAvailable is not locked)
//         server cannot be blocked by mutex
//   >=1 - works
void ThreadCommandQueue::sendAsyncMessage(Message *msg, bool immediately) {
	if (maxNumMsgs != 0) queueSlotsAvailable.acquire();
	queueMutex.lock();
		if (immediately) {
			messages.push_front(msg);
		} else {
			messages.push_back(msg);
		}
	queueMutex.unlock();
	// "add" a message to the semaphore
	messagesAvailable.release();	
}


// preconditions:
// same as sendAsyncMessage, additionally:
// answerAvailable
//        0           -	  will block until server answers
//       >=1          -   impossible on entry, possible if server answers before messagesAvailable.acquire(), will work due to semaphore counting mechanism
// queueSlotsAvailable
//    0 - server works on message queue, will hopefully release (i.e. receive) message slots
//   >=1 - works
// queueSlotsAvailable
//    0  - server works on message queue
//         server can be blocked by messagesAvailable, but only if queue empty (in that case queueSlotsAvailable is not locked)
//         server cannot be blocked by mutex
//   >=1 - works
void ThreadCommandQueue::sendMessageAndReceiveReply(Message *msg, bool immediately) {
	sendAsyncMessage(msg, immediately);
	// wait for reply (there can be only one)
	answersAvailable.acquire();
}
	


// preconditions:
// queueMutex    
//   locked      -	 client in getMessage(), will unlock soon
//   unlocked    -    works
// messagesAvailable    
//   0           -    will block until request becomes available
//                    does not hold any other resources, unless an answerable message has not been answered immediately (then it deadlocks)
//   1           -    works
// answerAvailable
//   any         -    no effect
// queueSlotsAvailable
//   any         -    does not wait on this semaphore
ThreadCommandQueue::Message *ThreadCommandQueue::getMessage() {
	if (maxNumMsgs != 0) queueSlotsAvailable.release();
	Message *result;
	messagesAvailable.acquire();
	queueMutex.lock();
		result = messages.front();
		messages.pop_front();
	queueMutex.unlock();
	return result;	
}


// preconditions:
// queueMutex    
//   any      -	 no effect
// messagesAvailable    
//   any      -    no effect
// answerAvailable
//   >=1      -    impossible
// queueSlotsAvailable
//   any      -    no effect
void ThreadCommandQueue::answerMessage() {
	answersAvailable.release();
}

int32 ThreadCommandQueue::getNumAvailableMessageSlots() {
	return queueSlotsAvailable.available();		
}
