//---------------------------------------------------------------------------
#ifndef TopologyEstimatorH
#define TopologyEstimatorH
//---------------------------------------------------------------------------
#include "x3editor2.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
#include "InCorePCTopologyGraph.h"
#include "LinearAlgebra.h"
//---------------------------------------------------------------------------

namespace X4 {

class PointSet;
class PointCloud;


/// Estimates the topology of a point cloud given the point positions and a superset of the topology.
/// "Topology" refers to a graph connecting points that are nearby on the original surface (not points from different surfaces).
class X3EDITOR2_API TopologyEstimator : public Persistent {
	X4_ABSTRACT_CLASS(TopologyEstimator)

 public:
	TopologyEstimator();

	virtual InCorePCTopologyGraph *estimateTopology(PointCloud *pc, PointSet *ps, InCorePCTopologyGraph *supersetTopology, bool in2d) = 0;
};


/// computes the topology using a greedy region growing approach;
/// starting from "certain" regions, edges are added that fit best to the current local geometry model
class X3EDITOR2_API GreedyTopologyEstimator : public TopologyEstimator {
	X4_CLASS(GreedyTopologyEstimator)

 private:
 
	// --- member types
 
	struct Edge {
		card32 v1;
		card32 v2;
		float32 weight;
		
		Edge(card32 v1=0, card32 v2=0, float32 w=0) {
			this->v1 = v1;
			this->v2 = v2;
			weight = w;
		}
		Edge(const Edge &other) {
			v1 = other.v1; 
			v2 = other.v2;
		}
		
		inline bool operator==(const Edge &other) const {
			return (v1 == other.v1 && v2 == other.v2) || (v1 == other.v2 && v2 == other.v1) || (weight == other.weight);
		}
		inline bool operator<(const Edge &other) const {
			if (weight < other.weight) return true;
			if (weight > other.weight) return false;
			if (v1<other.v1) return true;
			if (v1>other.v1) return false;
			if (v2<other.v2) return true;
			return false;
		}
		inline void operator=(const Edge &other) {
			v1 = other.v1;
			v2 = other.v2;
		}
	};

	struct LocalModel {
		Vector3f normal;
		Vector3f center;
		
		inline LocalModel(Vector3f normal = NULL_VECTOR3F, Vector3f center = NULL_VECTOR3F) {
			this->normal = normal;
			this->center = center;
		}
		inline bool isInitialized() {return normal != NULL_VECTOR3F;}
	};
	
	static const LocalModel LM_NULL;

	struct Vertex {
		card32 index;
		Vector3f position;
		float32 uncertainty;
		LocalModel localModel;
		bool selected;
		
		inline Vertex(card32 index=0, Vector3f position = NULL_VECTOR3F, float32 uncertainty = 0) {
			this->index = index;
			this->position = position;
			this->uncertainty = uncertainty;
			this->localModel = LM_NULL;
			selected = false;
		}
		inline bool operator<(const Vertex &v) const {
			if (uncertainty < v.uncertainty) return true;
			if (uncertainty > v.uncertainty) return false;
			return index < v.index;
		}
	}; 
	
	// --- data members
	// - persistent
	float32 initVerticesFraction;
	
 public:		
	GreedyTopologyEstimator();

	virtual InCorePCTopologyGraph *estimateTopology(PointCloud *pc, PointSet *ps, InCorePCTopologyGraph *supersetTopology, bool in2d);
};

}

#endif