//---------------------------------------------------------------------------
#ifndef TopologyToolsH
#define TopologyToolsH
//---------------------------------------------------------------------------
#include "x3editor2.h"
//---------------------------------------------------------------------------
#include "LinearAlgebra.h"
#include "VertexDescriptor.h"
//---------------------------------------------------------------------------

namespace X4 {

class BasicPointCloudIterator;


/// computes how well a given plane fits through a set of points
/// the measure is the average of squared distances
/// the points iterator must provide attribute position, float, 3-dim; the AAT is expected as parameter posAAT
/// points are read from the current position up to the end or until maxNumPts are processed
X3EDITOR2_API float32 computeSSDPlaneFitError(
	Vector3f normalizedPlaneNormal, float32 dist, 
	BasicPointCloudIterator *points, AAT posAAT,
	card32 maxNumPts = 0xFFFFFFFF
);

/// computes how well a given plane fits through a set of points
/// the measure is the (approx.) intergral over the line over radial exponential kernel functions around each point
/// the points iterator must provide attribute position, float, 3-dim; the AAT is expected as parameter posAAT
/// points are read from the current position up to the end or until maxNumPts are processed
X3EDITOR2_API float32 computeKernelPlaneFitError(
	Vector3f normalizedPlaneNormal, float32 dist, 
	BasicPointCloudIterator *points, AAT posAAT,
	float32 sigma,
	card32 maxNumPts = 0xFFFFFFFF
);


/// computes log poission distribution
/// uses table based k! computation - k must be <= 200
X3EDITOR2_API float32 calcNegLogPoissonProb(float32 lambda, card32 k);

/// computes how well a given plane fits through a set of points
/// a statistical measure is used that computes the set of points which maximizes the probability
/// of both having enough points and the likelihood of distance variations
/// the points iterator must provide attribute position, float, 3-dim; the AAT is expected as parameter posAAT
/// points are read from the current position up to the end or until maxNumPts are processed
X3EDITOR2_API float32 computeRobustStatisticsFitError(
	Vector3f normalizedPlaneNormal, float32 dist, 
	BasicPointCloudIterator *points, AAT posAAT,
	card32 expectedNumberOfPoints, // expected number of points in the sample from the plane
	float32 stdDev,                // standard deviation of of distance to the plane
	card32 maxNumPts = 0xFFFFFFFF
);

/// helper method - collect all points from iterator (up to maxNumPoints) and write result to std::vector)
X3EDITOR2_API void collectPoints(BasicPointCloudIterator *points, AAT posAAT, std::vector<Vector3f> &pointsPos, card32 maxNumPts = 0xFFFFFFFF);


/// computes PCA of 3d point cloud, sorts eigenvalues in ascending order ev[0] <= ev[1] <= ev[2]
X3EDITOR2_API void computePCA(const std::vector<Vector3f> &points, Vector3f &centroid, Matrix3f &eigenVectors, Vector3f &eigenValues);

/// computes leastSquare plane fit by PCA analysis
X3EDITOR2_API void computePCAPlane(const std::vector<Vector3f> &points, Vector3f &center, Vector3f &normal, bool in2d = false);

/// computes leastSquare plane fit by PCA analysis
X3EDITOR2_API void computePCAPlane(BasicPointCloudIterator *points, AAT posAAT, Vector3f &center, Vector3f &normal, card32 maxNumPts = 0xFFFFFFFF);

/// compute local uncertainty assuming locally flat surfaces
/// the result is the ratio of smallest to largest eigenvalue for a local PCA of the points
/// if "in2D" is true, the smallest eigenvalue is neglected
X3EDITOR2_API float32 computePlaneUncertainty(const std::vector<Vector3f> &points, bool in2d, bool useRatio = true);

/// same as computePlaneUncertainty(const std::vector &lt; Vector3f &gt; &amp;, bool), but operating on an iterator
X3EDITOR2_API float32 computePlaneUncertainty(BasicPointCloudIterator *points, AAT posAAT, bool in2d, card32 maxNumPts = 0xFFFFFFFF);


/// creates normal std::vector in x,y plane according to given angle
X3EDITOR2_API Vector3f get2dNormal(Degree angle);




}

#endif