//---------------------------------------------------------------------------
#ifndef ABS_FrameworkH
#define ABS_FrameworkH
//---------------------------------------------------------------------------
#include "x3foundation_private.h"
//---------------------------------------------------------------------------
#include <string>
#include "Persistent.h"
//---------------------------------------------------------------------------

namespace X4
{

class Scene;
class GLRenderTextureCache;
class GLImage;

/**
    meta info for a parameter
*/
class X3FOUNDATION_PRIVATE_API ABS_ParameterDescriptor : public Persistent
{
private:
   static MetaClass *ABS_ParameterDescriptorMetaClass;
   static Object* newInstance();

   float    m_MinValue;
   float    m_MaxValue;
   float    m_MaxStep;
   float    m_DefaultValue;
   std::string   m_Name;

public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

   void setup( float minValue, float maxValue, float maxStep, float defaultValue,
      std::string name ) {m_MinValue = minValue; m_MaxValue = maxValue; m_MaxStep = maxStep;
      m_DefaultValue = defaultValue; m_Name = name; }

   float getMinValue() const {return m_MinValue;}
   float getMaxValue() const {return m_MaxValue;}
   float getMaxStep() const {return m_MaxStep;}
   float getDefaultValue() const {return m_DefaultValue;}
   std::string getName() const {return m_Name; }

};

/**
   describes a parameter std::vector
*/
class X3FOUNDATION_PRIVATE_API ABS_ParameterDescriptorVector : public Persistent
{
private:
   static MetaClass *ABS_ParameterDescriptorVectorMetaClass;
   static Object* newInstance();

   std::vector<ABS_ParameterDescriptor*>   m_Parameters;

public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

   bool parameterExists( const std::string &name );

   // ---- objectlist property interface
   void addParameter( Object * obj );
   void deleteParameters( card32 begin, card32 end );
   Object * getParameter( card32 index );
   void moveParameter( card32 oldPos, card32 newPos );
   mpcard getNumParameters() {return m_Parameters.size();}
   std::string getParameterName(card32 index );

   void addAndSetupParameter( std::string name, float minV, float maxV, float maxStep, float defaultValue );

   card32 getParameterIndex( const std::string &name );
   float getMaxValue( card32 index );
   float getMinValue( card32 index );
   float getMaxStep( card32 index );
   float getDefaultValue( card32 index );

};

class X3FOUNDATION_PRIVATE_API ABS_ParameterVector
{
private:
   std::vector<float>  m_Data;
   bool           m_ValidResult;
   float          m_ResultValue;

public:

   ABS_ParameterVector() {m_ValidResult = false;}

   /// creates paramter std::vector with default values
   void setup( ABS_ParameterDescriptorVector * pdv );

   void setParameter( card32 i, float value )
      {x4Assert( i < m_Data.size()); m_Data[i] = value;}
   float getParameter( card32 i )
      {x4Assert( i < m_Data.size()); return m_Data[i]; }

   void setResultValue( float value ) {m_ResultValue = value; m_ValidResult = true;}
   float getResultValue() {x4Assert(m_ValidResult); return m_ResultValue;}

   bool isValidResult() const {return m_ValidResult;}

   /// returns true if parameter[name] exists and stores parameter into param
   bool readParameter( ABS_ParameterDescriptorVector * pdv, std::string name, float &param );

   /// returns true if parameter[name] exists and writes param into parameter
   bool writeParameter(  ABS_ParameterDescriptorVector * pdv, std::string name, float param );


   bool operator < ( const ABS_ParameterVector &other ) const
   {
      x4Assert( m_ValidResult );
      x4Assert( other.m_ValidResult );

      return m_ResultValue < other.m_ResultValue;
   }

};

class X3FOUNDATION_PRIVATE_API ABS_ImageSynthesizer : public Persistent
{
private:
   static MetaClass *ABS_ImageSynthesizerMetaClass;

protected:
   ABS_ParameterDescriptorVector    *  m_PDV;

public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

   /// synthesize image from scene into destImage with parameters params
   virtual void execute( GLRenderTextureCache * cache, GLImage * destImage, Scene * scene, ABS_ParameterVector * params  ) = 0;

   void setPDV( ABS_ParameterDescriptorVector * pdv )
   {
      m_PDV = pdv;
   }

};

class X3FOUNDATION_PRIVATE_API ABS_Comparator : public Persistent
{
private:
   static MetaClass *ABS_ComparatorMetaClass;

protected:

   GLImage  *  m_ReferenceImage;
   std::vector<float>  m_Results;

public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

   void setReferenceImage( GLImage * image ) {m_ReferenceImage = image;}
   GLImage * getReferenceImage() { return m_ReferenceImage; }

   /// adds image to compare-queue and returns the result-index
   /// Don't save a pointer to img !!!!
   virtual card32 addCompareImage( GLImage * img, GLRenderTextureCache * cache ) = 0;

   /// compares all remaining images and publishes result
   virtual void execute( GLRenderTextureCache * cache )=0;

   mpcard getNumResults() {return m_Results.size();}
   float getResult( card32 i) {x4Assert(i<m_Results.size()); return m_Results[i];}

};

class X3FOUNDATION_PRIVATE_API ABS_Analysator : public Persistent
{
private:
   static MetaClass *ABS_AnalysatorMetaClass;

protected:

   /// sort inputvectors with respect to resultvalue
   void sortInputVectors();

   std::vector<ABS_ParameterVector>       m_InputVectors;
   std::vector<ABS_ParameterVector>       m_OutputVectors;
   ABS_ParameterDescriptorVector *   m_PDV;

public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

   /// generates new set of #num vectors
   virtual void generateNewVectors(card32 num) = 0;

   /// returns the best parameter found
   virtual ABS_ParameterVector getBestVector() = 0;

   void setPDV( ABS_ParameterDescriptorVector * pdv )
   {
      m_PDV = pdv;
   }


   mpcard getNumOutputVectors() const {return m_OutputVectors.size();}
   ABS_ParameterVector getOutputVector( card32 i ) {x4Assert( i < m_OutputVectors.size() );
      return m_OutputVectors[i];}

   void clearInputVectors() {m_InputVectors.clear();}
   void addInputVector( const ABS_ParameterVector &vec ) {m_InputVectors.push_back(vec);}

};

class X3FOUNDATION_PRIVATE_API ABS_Framework : public Persistent
{
private:
   static MetaClass *ABS_FrameworkMetaClass;
   static Object* newInstance();

   ABS_ParameterDescriptorVector    *  m_PDV;
   ABS_Analysator                   *  m_Analysator;
   ABS_ImageSynthesizer             *  m_Synthesizer;
   ABS_Comparator                   *  m_Comparator;

   GLImage                          *  m_ReferenceImage;

   card32                              m_NumSamples;

public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

   ABS_Framework();
   ~ABS_Framework();

   virtual void calculateIteration(GLRenderTextureCache *cache, Scene * scene);

   ABS_ParameterVector getBestVector() {return m_Analysator->getBestVector();}

   void setPDV( ABS_ParameterDescriptorVector * pdv );
   ABS_ParameterDescriptorVector * getPDV() {return m_PDV;}

   void setAnalysator( ABS_Analysator * a);
   ABS_Analysator * getAnalysator() {return m_Analysator;}

   void setImageSynthesizer( ABS_ImageSynthesizer * s );
   ABS_ImageSynthesizer * getImageSynthesizer() {return m_Synthesizer;}

   void setComparator( ABS_Comparator * c );
   ABS_Comparator * getComparator() {return m_Comparator;}

   void setNumSamples( card32 num ) {m_NumSamples = num;}
   card32 getNumSamples() {return m_NumSamples;}

   void setReferenceImage( GLImage * image ) {m_ReferenceImage = image;}
   GLImage * getReferenceImage() {return m_ReferenceImage;}



};

}

#endif
 