#ifndef BRDFAttachmentH
#define BRDFAttachmentH
//---------------------------------------------------------------------------
#include "x3foundation_private.h"
//---------------------------------------------------------------------------
#include "MaterialAttachment.h"
#include "GLContext.h"
#include "ObjectListProperty.h"
#include "GLTexture.h"
//---------------------------------------------------------------------------

namespace X4
{
	class GLTexture3D;

	/**
	* BRDF attachment for storing tabelized BRDF function as an attachment.
	* Provides a couple of useful tools for BRDF handling
	* NOTE: Theta = azimuth xy, Phi = zenith z
	**/
	class X3FOUNDATION_PRIVATE_API BRDFAttachment : public TextureAttachment
	{
		X4_CLASS(BRDFAttachment);
	protected:

		float* m_BRDF;

		Vector3f m_Normal;
		
		card16 m_outPhiDomain;
		card16 m_outThetaDomain;
		card16 m_inPhiDomain;
		card16 m_inThetaDomain;
		card8 m_numChannels;

		std::vector<float> m_pcaBasisScalar;
		bool m_usePCAprojection;

		bool m_use16bit;
		bool m_isotrop;
		card16 m_isotropOutPhi;

	public:
		BRDFAttachment();
		virtual ~BRDFAttachment();

		typedef pair<card16, card16> DiscreteUniformPolarCoordinates;

		virtual card32 getTextureType() const { return TextureAttachment::TYPE_3D; }

		virtual void assign(const Object* obj, X4CopyContext *context = NULL);
		virtual void write(OutputObjectStream *s) const;
		virtual void read(InputObjectStream *s);

		//! BRDF summation
		BRDFAttachment& operator+=(const BRDFAttachment& b);
		inline BRDFAttachment& operator+(const BRDFAttachment& b) { return *this+=b; }
		BRDFAttachment& operator-=(const BRDFAttachment& b);
		inline BRDFAttachment& operator-(const BRDFAttachment& b) { return *this-=b; }
		BRDFAttachment& operator*=(float v);
		inline BRDFAttachment& operator*(float v) { return *this *= v;}
		float operator*=(const BRDFAttachment& b)const;
		inline float operator*(const BRDFAttachment& b)const { return *this *=b; }

		/** 
		* Specify dimensions of the discretization for each of the parameters.
		* The discretization will be chosen mostly uniform around the hemisphere.
		* NOTE: When the data might get invalid when you set new discretization.
		**/
		void setDiscretization(card16 outPhi, card16 outTheta, card16 inPhi, card16 inTheta, card8 numChannels);

		//! get discretization
		unsigned getInPhiDomainSize() const { return m_inPhiDomain; }
		unsigned getInThetaDomainSize() const { return m_inThetaDomain; }
		unsigned getOutPhiDomainSize() const { return m_outPhiDomain; }
		unsigned getOutThetaDomainSize() const { return m_outThetaDomain; }
		card8 getNumChannels() const { return m_numChannels; }
		size_t getNumEntries() const { return m_outPhiDomain*m_outThetaDomain*m_inPhiDomain*m_inThetaDomain*m_numChannels; }
		size_t getMemorySize() const { return getNumEntries() * sizeof(float); }

		bool isProjectedToPCA() const { return m_usePCAprojection; }

		//! Directly access to the data of this BRDF
		float* data() { return m_BRDF; }
		float& at(card16 outphi, card16 outtheta, card16 inphi, card16 intheta, card8 channel);

		/**
		* Set BRDF normal, this is a polar axis for the coordinates.
		**/
		void setNormal(Vector3f normal) { m_Normal = normal; }

		/** 
		* Convert given cartesian coordinates into polar coordinates which can be directly used to access values.
		* The coordinates are distributed uniformly along the surface, i.e.:
		*   theta = 2 * PI * u
		*   phi   = cos^-1(2v-1)
		* uv are the coordinates returned back from this method. Dir is first converted into usual polar coordinate phi, theta
		**/
		DiscreteUniformPolarCoordinates getCoordinatesForOutputDirection(Vector3f dir);


		/**
		* Create 3D OpenGL texture which stored the BRDF. The data is stored slicewise for inPhi/inTheta.
		* Hence the texture represents in x/y dimension the slice at outPhi/outTheta. The z direction
		* is packed by slices of outer directions. So the indexing is following:
		*  inPhi = x;
		*  inTheta = y;
		*  outPhi = z % outPhiDimension
		*  outTheta = z / outPhiDiemsion
		* @param makeIsotrop if true, the data BRDF is handled as isotrop BRDf. Hereby the outPhi is ignored.
		* This means the resulting 3D texture, contains directly accessable 3D BRDF with [x,y,z] = [inPhi, inTheta, outTheta] parametrization.
		* @param isotropOutPhi Which entry for the outPhi should be used as assumed as constant. The 3D BRDF is constructed as a 3D slice
		* of the 4D BRDF for the given outPhi angle.
		* @param use16bit if false texture is created as 32bit float, otherwise 16bit
		* NOTE: the texture can only be created, if the BRDF is not encoded in the PCA domain.
		**/
		GLTexture3D* createBRDFTexture(GLContext* glContext, bool makeIsotrop = false, card16 isotropOutPhi = 0, bool use16bit = false);
		virtual GLTexture * createGLTexture(GLContext * context)
		{
			return createBRDFTexture(context, m_isotrop, m_isotropOutPhi, m_use16bit);
		}
		void setTextureParameters(bool use16bit, bool makeIsotrop, card16 isotropOutPhi)
		{
			m_use16bit = use16bit;
			m_isotrop = makeIsotrop;
			m_isotropOutPhi = isotropOutPhi;
		}


		/**
		* Project BRDf into PCA space. This return linear components of each PCA basis BRDF.
		* This fully reduce BRDF to this factors only. In order to access
		* BRDF values afterwards PCA backprojection will be required. This will drastically save memory.
		**/
		void reduceToPCAbasis(const std::vector<BRDFAttachment*>& pcaBasis);

		/**
		* Reconstruct full BRDF as a linear combination of the pcaBasis vectors. 
		* NOTE: the brdf must be previoulsy encoded into pca domain with reduceToPCAbasis()
		* @return false if cannot reconstruct (this is the case if BRDf was not previously encoded in the PCA domain
		**/
		bool reconstructFromPCAbasis(const std::vector<BRDFAttachment*>& pcaBasis);

		/**
		* Return PCA parameters required to reconstruct the BRDF from pca basis vectors.
		**/
		const std::vector<float>& getPCAParameters() const { return m_pcaBasisScalar; }

		/**
		* Perform PCA analysis of given BRDFs and return number of "main" BRDFs [result].
		**/
		static card32 pcaAnalysis(const std::vector<BRDFAttachment*>& brdfs, std::vector<BRDFAttachment*>& result, card32 maxNumBRDFS, float eigenValueThreshold = 1e-4f);

	};

	/**
	* List of all BRDFs used in the scene. Every point of the point cloud which using BRDFs must correspond
	* within an index to BRDF used.
	**/
	class X3FOUNDATION_PRIVATE_API BRDFContainer : public AttachedData
	{
		X4_CLASS(BRDFContainer);
		std::vector<BRDFAttachment*> m_BRDF;

	public:
		IMPLEMENT_OBJECT_LIST_METHODS_STL(BRDF, BRDFAttachment);

		static std::string getDefaultName() {return "BRDFContainer";}
	};

}; // end namespace

#endif 

