#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "Circle3D.h"
//---------------------------------------------------------------------------
#include "DebugRenderer.h"
//---------------------------------------------------------------------------
using namespace X4;



Circle3D::Circle3D()
{
	m_Radius = -1;
}

Vector3f expandV( const Vector2f &v, float x2=0)
{
	return makeVector3f(v[0],v[1],x2);
}

bool Circle3D::constructFromPoints(const Vector3f &p0,const Vector3f &p1,const Vector3f &p2)
{
	m_Radius = -1;

	// try to construct base frame
	m_Frame[0] = p1-p0;
	float l0 = norm(m_Frame[0]);
	if( l0 < 1e-10f )
		return false;
	m_Frame[1] = p2-p0;
	float l1 = norm(m_Frame[1]);
	if( l1 < 1e-10f )
		return false;
	if( norm(p2-p1) < 1e-10f )
		return false;
	m_Frame[0] /= l0;
	m_Frame[2] = normalize(m_Frame[0].crossProduct(m_Frame[1]));
	m_Frame[1] = normalize(m_Frame[0].crossProduct(m_Frame[2]));
	m_FrameT = m_Frame.transpose();

	const Vector2f p1_2d = shrink3To2(m_FrameT * (p1-p0));
	const Vector2f p2_2d = shrink3To2(m_FrameT * (p2-p0));

	// construct lines through center
	const Vector2f dir0 = makeVector2f(-p1_2d[1],p1_2d[0]);
	const Vector2f mp0 = p1_2d * 0.5f;
	const Vector2f dir1 = makeVector2f(-p2_2d[1],p2_2d[0]);
	const Vector2f mp1 = p2_2d * 0.5f;

	// compute intersection as center
	Matrix2f M; M[0] = dir0; M[1] = -dir1;
	Vector2f t;
	if( !linsolve2f(M.transpose(),mp1-mp0,t) )
		return false;
	Vector2f center = mp0 + dir0*t[0];
	if( norm(dir0) < norm(dir1) )
		center = mp1 + dir1*t[1];

	m_Center = m_Frame*expandV(center)+p0;
	m_Radius = (norm(m_Center - p0)+norm(m_Center - p1)+norm(m_Center - p2))/3.0f;

	return true;
}

float Circle3D::distance(const Vector3f &p) const
{
	Vector3f p_2d = m_FrameT * (p - m_Center);
	return sqrt(  sqr(norm(shrink3To2(p_2d)) - m_Radius) + sqr(p_2d[2]));
}

float Circle3D::computeProjection(const Vector3f &p) const
{
	static bool test = true;
	if( test)
	{
		test = false;
		for( mpcard i=0;i<200;i++ )
		{
			float angle = ((float)i-100.0f)/100.0f * M_PI;
			float proj = computeProjection(computeCirclePoint(angle));
			debugOutput << "angle: " << angle << " proj: " << proj << "\n";
		}
	}


	Vector3f p_2d = m_FrameT * (p - m_Center);
	float angle = atan2(p_2d[1],p_2d[0]);
	if( angle < 0 )
		angle += 2.0f * (float)M_PI;
	return angle;
}

Vector3f Circle3D::computeCirclePoint(float angle) const
{
	return m_Frame * makeVector3f(cos(angle),sin(angle),0)*m_Radius + m_Center;
}

void Circle3D::drawDR(const Vector3f &color, mpcard numSegments)
{
	Vector3f lastPos;
	for( mpcard i=0;i<=numSegments;i++ )
	{
		float angle = (float)i / (float)numSegments * 2.0f * (float)M_PI;
		Vector3f pos = computeCirclePoint(angle);
		if( i > 0 )
			debugRenderer->addLine(lastPos,pos,color,color,1);
		lastPos = pos;
	}
}

Matrix4f Circle3D::computeT( float angle) const 
{
	Matrix4f frame = makeTranslation4f(m_Center)*expand3To4(m_Frame);
	return frame * makeRotZ4f(-angle) * invertFrame(frame);
}

void Circle3D::setup( const Matrix3f &frame, const Vector3f &center, float radius) 
{
	m_Frame = frame;
	m_FrameT = frame.transpose();
	m_Center = center;
	m_Radius = radius;
}
