//------------------------------------------------------------------------------
#include "stdafx.h"

//------------------------------------------------------------------------------
#include "ConvexHull2D.h"
//------------------------------------------------------------------------------
#include "LinearAlgebra.inline.h"
#include "DebugOutput.h"
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
using namespace X4;
//------------------------------------------------------------------------------

/// Create new object
ConvexHull2D::ConvexHull2D( unsigned int bound )
{
   init( bound );
}

/// Delete object
ConvexHull2D::~ConvexHull2D() 
{

}

/// Create new object
void ConvexHull2D::init( unsigned int bound )
{
   points = new Vector2f[bound];
	numPoints = 0;
	pointIndices = new int[bound];
	numConvexHullIndices = 0;
   convexHullIndices = new int[bound];
}

/// Clear point container
void ConvexHull2D::clear()
{
	numPoints = 0;
}

/// Add point
void ConvexHull2D::addPoint( Vector2f p )
{ 
	points[numPoints] = p; 
	pointIndices[numPoints] = numPoints; 
	numPoints++;  
}

/// Delete point from array. Attention: pretty expensive operation
void ConvexHull2D::deletePoint( unsigned int index )
{
   assert( numPoints > 0 );
	unsigned int n;
	for ( n = index; n < numPoints - 1; n++ ){
		points[n] = points[n+1];
		pointIndices[n] = pointIndices[n+1];
	}
	numPoints--;
}

/// Swap two points in array
void ConvexHull2D::swapPoints( unsigned int i, unsigned int j )
{
	Vector2f help = points[i];
	int helpIndex;
	points[i] = points[j];
	points[j] = help;
	helpIndex = pointIndices[i];
	pointIndices[i] = pointIndices[j];
	pointIndices[j] = helpIndex;
}

/// Test if point p is left of the line through p0 and p1
float ConvexHull2D::isLeft( Vector2f p0, Vector2f p1, Vector2f p ) const
{
	return ( p1[0] - p0[0] )*( p[1] - p0[1] ) - ( p[0] - p0[0] ) * ( p1[1] - p0[1] );
}

/// Compare points p0 and p1.; p1 left (pivot <-> p0) -> return < 0; p1 on (pivot <-> p0) -> return 0; p1 right of (pivot <-> p0) -> return > 0
float ConvexHull2D::compare( int p0, int p1 ) const
{
	return isLeft( pivot, points[p0], points[p1] );
}

/// Perform quicksort on point array
void ConvexHull2D::quickSortPoints()
{
	/// Start with first point, point 0 is pivot point
	unsigned int currentIndex = 0;
	while ( currentIndex < numPoints ){
		for ( unsigned int n = currentIndex + 1; n < numPoints; n++ ){
			float comparison = compare( currentIndex, n );
			if ( comparison > 0 ){
				swapPoints( currentIndex, n );
			} else if ( comparison == 0 ){
				/// Delete closer point
				float dist0 = ( pivot - points[currentIndex] ).getSqrNorm();
				float dist1 = ( pivot - points[n] ).getSqrNorm();
				if ( dist0 < dist1 ){
					deletePoint( currentIndex );
					currentIndex--;
				} else {
					deletePoint( n );
					n--;
				}
			}
		}
		currentIndex++;
	}
}


/// Compute convex hull
void ConvexHull2D::computeConvexHull()
{
	unsigned int n;

    // Select the rightmost lowest point as pivot
	int pivotIndex = 0;
	for ( n = 1; n < numPoints; n++ ){
		if ( ( points[n][1] < points[pivotIndex][1] ) ||
			 ( ( points[n][1] == points[pivotIndex][1] ) &&
			   ( points[n][0] > points[pivotIndex][0] ) ) )
			pivotIndex = n;
	}
	pivot = points[pivotIndex];
	// Save index of pivot point
	deletePoint( pivotIndex );

	//// Sort points angularly around pivot as a center.
	quickSortPoints();

	//// Push pivot and points[0] onto stack.
	stack.clear();
	stack.push( pivot );
    stack.push( points[0] );
    convexHullIndices[0] = 0; numConvexHullIndices = 1;
    n = 1;

	while (n < numPoints){
		if (isLeft(stack.getSecond(), stack.getTop(), points[n]) < 0){
			stack.push(points[n]);
			convexHullIndices[numConvexHullIndices] = n;
			numConvexHullIndices++;
			n++;
		}
		else {
			stack.pop();
			numConvexHullIndices--;
		}
	}

   // Hack: save pivot point
   points[numPoints+1] = pivot;
   convexHullIndices[numConvexHullIndices] = numPoints+1;
   numConvexHullIndices++;
}

/// Test if point is withing convex hull (only valid after computeConvexHull)
bool ConvexHull2D::isInside( const Vector2f& testPoint )
{
   for ( int n = 0; n < numConvexHullIndices; n++ ){
      Vector2f p0 = points[convexHullIndices[n]];
      Vector2f p1 = points[convexHullIndices[(n+1)%numConvexHullIndices]];
      if ( isLeft( p0, p1, testPoint ) >= 0 ) return false;
   }
   return true;
}


Vector2f ConvexHull2D::getHullCenter() const
{
	Vector2f center = makeVector2f( 0,0 );
	for ( int32 n = 0; n < numConvexHullIndices; n++ ){
		center += points[convexHullIndices[n]];
	}
	if ( numConvexHullIndices > 0 ){
		center /= float32( numConvexHullIndices );
	}
	return center;
}
