#ifndef ConvexHull2DH
#define ConvexHull2DH    

// Xrt
#include "LinearAlgebra.h"
#include "FiloStack.h"

#include "x3foundation_private.h"

// System
#include <vector>

namespace X4 {

/// Compute the convex hull of a point set in 2D using Graham's algorith.
class X3FOUNDATION_PRIVATE_API ConvexHull2D
{
	private:
		/// Container for the points defining the object
		Vector2f* points;
		/// Number of points in the array
		unsigned int numPoints;
		/// Current pivot point of the algorithm
		Vector2f pivot;
		/// Indices of original points in 'points' array
		int* pointIndices;
		/// Number of points in the convex hull
		int numConvexHullIndices;
		/// Indices of the points in the convex hull
		int* convexHullIndices;

      /// Stack
      FiloStack<Vector2f> stack;
		/// Test if point 'p' is left of line p0, p1
		float isLeft( Vector2f p0, Vector2f p1, Vector2f p ) const;
		/// Quicksort point array
		void quickSortPoints();
		/// Swap two points in array
		void swapPoints( unsigned int i, unsigned int j );
		/// Compare points p0 and p1 for quicksort
		float compare( int p0, int p1 ) const; 
		// Delete point (Expensive!)
		void deletePoint( unsigned int index );

	public:
		ConvexHull2D( unsigned int bound );
		~ConvexHull2D();
		
		void init( unsigned int bound );									/// Init members
		void clear();														/// Clear object
		void addPoint( Vector2f p );										/// Add point
		void computeConvexHull();											/// Compute convex hull (Graham Scan)
		inline int* getHullPointsIndices( unsigned int& numHullPoints ){	/// Return the sorted list of indices of the points in the convex hull (only valid after computeConvexHull)
			numHullPoints = numConvexHullIndices; return convexHullIndices; 
		}		
		bool isInside( const Vector2f& testPoint );							/// Test if point is withing convex hull (only valid after computeConvexHull)
		inline card32 getNumHullPoints() const { return numConvexHullIndices; }
		inline Vector2f getHullPoint( const card32 index ) const{ return points[convexHullIndices[index]]; }
		Vector2f getHullCenter() const;
};

}

#endif