#include "stdafx.h"
//---------------------------------------------------------------------------
#include "CurvatureEstimator.h"
#include "NumericalClassProperty.h"
#include "DebugOutput.h"
#include "HierarchicalKNNIterator.h"
#include "ProgressWindow.h"
#include "PointSetKNNQuery.h"
#include "PointSet.inline.h"
#include <time.h>
#include "HierarchicalRangeQueryIterator.h"
#include "MathTools.h"
#include "Monomials2ndOrder.h"
#include "LocalSurfaceFunction.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------


void CurvatureEstimator::estimateCurvature( PointSet* ps, const float32 epsilon )
{
	if ( ps == NULL ){
		debugOutput << "estimateCurvature: invalid point set\n"; debugOutput.flush();
		return;
	}

	if ( !ps->providesAttribute( "coeffs" ) || !ps->providesAttribute( "curv0" ) || 
		 !ps->providesAttribute( "curv1" ) || !ps->providesAttribute( "polyScale" ) ){
		debugOutput << "estimateCurvature: channel missing\n"; debugOutput.flush();
		return;
	}

	const float32 startTime = clock();

	AAT COEFFS = ps->getAAT( "coeffs" );
	AAT CURV_0 = ps->getAAT( "curv0" );
	AAT CURV_1 = ps->getAAT( "curv1" );
	AAT POLY_SCALE = ps->getAAT( "polyScale" );

	progressWindow->pushStep( true, "Curvatures" );
	for ( unsigned int pointIndex = 0; pointIndex < ps->getNumEntries(); pointIndex++ ){
		StaticVector<float32,6> coeffs = ps->get<float32,6>( pointIndex, COEFFS );
		const float32 polyScale = ps->get1f( pointIndex, POLY_SCALE );

		// Estimate curvatures
		// http://mathworld.wolfram.com/GaussianCurvature.html and 
		// http://mathworld.wolfram.com/MeanCurvature.html
		float32 u = 0;
		float32 v = 0;
		float32 f_u = coeffs[1] + coeffs[3] * v + 2.0f * coeffs[4] * u;
		float32 f_v = coeffs[2] + coeffs[3] * u + 2.0f * coeffs[5] * v;
		float32 f_uu = 2 * coeffs[4];
		float32 f_vv = 2 * coeffs[5];
		float32 f_uv = coeffs[3];
		float32 K = ( f_uu * f_vv - sqr( f_uv ) ) / sqr( 1 + sqr( f_uu ) + sqr( f_vv ) );
		float32 H = ( ( 1 + sqr( f_v ) ) * f_uu - 2.0f * f_u * f_v * f_uv + ( 1 + sqr( f_u ) ) * f_vv ) / ( 2.0f * pow( 1 + sqr( f_u ) + sqr( f_v ), 1.5f ) );
		float32 kappa1 = fabs( H + sqrt( sqr( H ) - K ) ) / polyScale;
		float32 kappa2 = fabs( H - sqrt( sqr( H ) - K ) ) / polyScale;
		ps->set1f( pointIndex, CURV_0, kappa1 );
		ps->set1f( pointIndex, CURV_1, kappa2 );

		progressWindow->progress( card32( float( pointIndex ) * 100 / float32( ps->getNumEntries() ) ) );
	}
	progressWindow->popSteps();
	debugOutput << "Time required (estimate curvature): " << float32( ( clock() - startTime ) / CLK_TCK ) << "\n"; debugOutput.flush(); 
}

void CurvatureEstimator::estimateCurvature( PointCloud* pc, const float32 epsilon )
{
	debugOutput << "estimateCurvature: not implemented for point cloud\n"; debugOutput.flush();
}