#include "StdAfx.h"
//---------------------------------------------------------------------------
#include <windows.h>
#include <mmsystem.h>
#include "FSIcp.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "PointSetANNQuery.h"
#include "DebugRenderer.h"
#include "ObjectViewsTable.h"
#include "Scene.h"
#include "Random.h"
#include "UnstructuredInCorePointCloud.h"
#include "HierarchicalKNNIterator.h"
#include "PointCloudTools.h"
#include "ColorTools.h"
#include <opencv\cxcore.h>

//---------------------------------------------------------------------------
using namespace X4;

IMPLEMENT_X4_CLASS( FSIcp ,0)
{
	BEGIN_CLASS_INIT( FSIcp );
	ADD_CARD32_PROP (maxICPSteps, 0);
	ADD_FLOAT32_PROP(allowedNormalDiff, 0);
	ADD_FLOAT32_PROP(m_MedPointDistance, 0);
	ADD_FLOAT32_PROP(outlierDistanceFactor, 0);
	ADD_FLOAT32_PROP(minPercentInliers, 0);
	ADD_FLOAT32_PROP(icpAllowedMoveDist, 0);
	ADD_FLOAT32_PROP(icpMaxResidual, 0);
	ADD_FLOAT32_PROP(icpConvergenceDiff, 0);
	ADD_BOOLEAN_PROP(visualization, 0);
}

FSIcp::FSIcp(void)
{
	this->setupParameters();
	targetPS=NULL;
	scene=NULL;
	annQuery=NULL;
	useANN = false;
	hIt = NULL;
}

FSIcp::~FSIcp(void)
{
	delete annQuery;
	delete hIt;
}

void FSIcp::assign(const Object* obj, X4CopyContext *context)
{
	// Persistent::assign(obj,context);
	const FSIcp * o = dynamic_cast<const FSIcp*>(obj);
	x4Assert( o != NULL );

	copyObjectProperties(obj,this);
}

void X4::FSIcp::setupParameters( card32 maxICPSteps /*= 32*/, float32 allowedNormalDiff /*= 2*/, float32 m_MedPointDistance /*= 0.1f*/, float32 outlierDistanceFactor /*= 1000*/, float32 minPercentInliers, float32 icpAllowedMoveDist/*= 1e20f */, float32 icpMaxResidual/*= 0.01f*/, float32 icpConvergenceDiff /*= 0.01f*/, bool visualization/*= false*/ )
{

	this->maxICPSteps=maxICPSteps;
	this->allowedNormalDiff=allowedNormalDiff;
	this->m_MedPointDistance=m_MedPointDistance;
	this->outlierDistanceFactor=outlierDistanceFactor;
	this->minPercentInliers=minPercentInliers;
	this->icpAllowedMoveDist=icpAllowedMoveDist;
	this->icpMaxResidual=icpMaxResidual;
	this->icpConvergenceDiff=icpConvergenceDiff;
	this->visualization=visualization;
}

void X4::FSIcp::setupData( PointSet* ps, Scene* scene )
{
	this->targetPS=ps;
	this->scene=scene;
	useANN = true;
	delete annQuery;
	annQuery= new PointSetANNQuery(targetPS, 5);
	NORMAL= ps->getAAT("normal");
	POSITION= ps->getAAT("position");

}

void X4::FSIcp::setupData( PointCloud* pc, Scene* scene )
{
	this->targetPS=NULL;
	this->scene=scene;
	useANN = false;
	if (hIt) delete hIt;
	hIt = new HierarchicalKNNIterator(pc,32,NULL);
	NORMAL= pc->getAAT("normal");
	POSITION= pc->getAAT("position");
	m_MedPointDistance = X4::getMedianPointDistance(pc);
}
X4::float32 X4::FSIcp::calculateResidualAfterICP( PointSet* startPS, Matrix4f& originalTransformation )
{
	if (!startPS) return 1e20f;

	Matrix4f trans=originalTransformation;
	Matrix4f initialTransformation = trans;

	float residual = 0;
	float lastResidual = 1e20f;

	if (m_MedPointDistance==0.0f)
	{
		if( targetPS )
			m_MedPointDistance= getMedianPointDistance(targetPS);
		else
			throw X4Exception("FSIcp::calculateResidualAfterICP() - Median Point Distance is not defined!");
	}

	// compute bounding box of sartPS
	card32 numPoints = startPS->getNumEntries();
	BoundingBox3f bb(startPS->get3f(0,POSITION));
	for( unsigned i=1;i<numPoints;i++ )
		bb.addPoint(startPS->get3f(i,POSITION));

	startPScenter=bb.getCenter();
	startPScenterOrgTransformed=transformVector3f(originalTransformation, startPScenter);

	bb.makeBoundingCube();


	// compute transformation from world space to normed local space [-1,1]?
	//Matrix3f relativeRotation = shrink4To3(initialTransformation);
	Matrix4f sourceFrame = makeTranslation4f(bb.getCenter());
	Matrix4f initialSourceFrame = sourceFrame;
	Matrix4f destFrame = initialTransformation * sourceFrame;

	float scaleFactor = bb.getSideLength(0) / 2.0f;
	Matrix4f destToLocal = makeScale4f(makeVector3f(1,1,1) / scaleFactor)* invertFrame(destFrame);
	Matrix4f localToDest = destFrame * makeScale4f(makeVector3f(1,1,1) * scaleFactor);
	Matrix4f refToLocal = makeScale4f(makeVector3f(1,1,1) / scaleFactor)* invertFrame(sourceFrame);
	Matrix4f localToRef = sourceFrame * makeScale4f(makeVector3f(1,1,1) * scaleFactor);
	float summed_weight = 0;
	std::string icpPoints = "debug icp";



	for( unsigned pass=0;pass<maxICPSteps;pass++)
	{

		if( visualization )
			debugRenderer->beginRenderJob(icpPoints);

		residual = 0;
		summed_weight = 0;
		Matrix3f zeroMat = IDENTITY3F * 0;
		Matrix3f P = zeroMat;
		Matrix3f S = zeroMat;
		Matrix3f R = zeroMat;
		Vector6f resB = NULL_VECTOR6F;

		// calculate corresponding points
		for( unsigned i=0;i<numPoints;i++)
		{
			Vector3f point = startPS->get3f(i,POSITION);
			Vector3f transformedPos = transformVector3f(trans, point);
			//			Vector3f transformedNormal = relativeRotation * lf->m_Normal;

			// seek closest point
			Vector3f closestPoint,closestPointNormal;
			if( useANN )
			{
				card32 transformedPointIndex = annQuery->getNearestPointIndex(transformedPos);
				closestPoint = targetPS->get3f(transformedPointIndex, POSITION);
				closestPointNormal =  targetPS->get3f(transformedPointIndex, NORMAL);
			}
			else
			{
				hIt->setSeekPointAndReset(transformedPos);
				closestPoint = hIt->get3f(POSITION);
				closestPointNormal = hIt->get3f(NORMAL);
			}

			//float surfelWeight=1.0f;
			float surfelWeight = norm(closestPoint-transformedPos) > m_MedPointDistance * outlierDistanceFactor ? 0.0f : 1.0f;
			if(surfelWeight>0.0f){
				summed_weight+=surfelWeight;
				float weight = surfelWeight;


				// compute local quadratic approximation
				// M =  n n^t
				// F(x) = (x-p)^tM(x-p) = x^tMx - 2x^tMp + ...
				Vector3f relP = transformVector3f(refToLocal, point);
				Matrix3f matM = outerProduct(closestPointNormal, closestPointNormal );
				Vector3f destPosLocal = transformVector3f(destToLocal, closestPoint);
				Vector3f b = matM * destPosLocal * (-2.0f);

				// F(x) = Ax?+ Bxy + Cy?+ Dxz + Eyz + Fz?+ Gx + Hy + Iz (Mitra et al 2004)
				float A = matM[0][0];
				float B = matM[1][0] + matM[0][1];
				float C = matM[1][1];
				float D = matM[2][0] + matM[0][2];
				float E = matM[2][1] + matM[1][2];
				float F = matM[2][2];
				float G = b[0];
				float H = b[1];
				float I = b[2];

				// compute derivates
				float x = relP[0];
				float y = relP[1];
				float z = relP[2];
				float J = 2.0f * (C*x*x - B*x*y + A*y*y);
				float K = 2.0f * (F*x*x - D*x*z + A*z*z);
				float L = 2.0f * (F*y*y - E*y*z + C*z*z);
				float M = -E*x*x + D*x*y + B*x*z - 2.0f*A*y*z;
				float N = -D*y*y + E*x*y -2.0f*C*x*z + B*y*z;
				float T = -B*z*z -2.0f*F*x*y + E*x*z + D*y*z;
				float U = B*(x*x-y*y) + 2.0f*(C-A)*x*y + E*x*z - D*y*z + H*x - G*y;
				float V = D*(z*z-x*x) - E*x*y + 2.0f*(A-F)*x*z + B*y*z -I*x + G*z;
				float W = E*(y*y-z*z) + D*x*y - B*x*z + 2.0f*(F-C)*y*z + I*y - H*z;

				// setup linear system
				P += makeMatrix3f(	J, M, N, 
									M, K, T, 
									N, T, L) * weight;
				S += makeMatrix3f(	B*x - A*y*2.0f	, C*x*2.0f - B*y	, E*x - D*y,
									-D*x + A*z*2.0f	, -E*x + B*z		, -2.0f*F*x + D*z,
									D*y - B*z		, E*y - 2.0f*C*z	, 2.0f*F*y - E*z) * weight;
				R += makeMatrix3f(	2.0f*A, B, D, 
									B, 2.0f*C, E, 
									D, E, 2.0f*F) * weight;
				resB -= makeVector6f(U,V,W,
					2.0f*A*x + B*y + D*z + G,
					B*x + 2.0f*C*y + E*z + H,
					D*x + E*y + 2.0f*F*z + I) * weight;


				// current residual
				summed_weight +=  weight;
			}

			// project onto surfel
			Vector3f projectedPoint = transformedPos - closestPointNormal * (closestPointNormal * (transformedPos- closestPoint));
			float projectedDistance = (transformedPos - projectedPoint)*(transformedPos - projectedPoint);
			if (visualization) debugRenderer->addSphere(transformedPos, m_MedPointDistance*0.6f ,makeVector3f(0.3f,0.7f*surfelWeight,0.86f));

			// current residual
			residual += projectedDistance; 

		}



		if (summed_weight<minPercentInliers*(float32)numPoints)
		{
			residual = 1e20f;
		}

		if (visualization) {
			debugRenderer->endRenderJob();
			ObjectViewsTable::update(scene);
		}

		if (!checkAllowedMovement(trans)){
			if (visualization) debugRenderer->clearDebugData( icpPoints );	// clear data from last call
			return 1e20f;
		}

		if(pass>0 && fabs(lastResidual-residual) < icpConvergenceDiff )
		{

			if (visualization) debugOutput << "converged after " <<pass<< " steps\n";
			if (visualization)debugRenderer->clearDebugData( icpPoints );	// clear data from last call

			originalTransformation=trans;
			return residual/(float32)numPoints;

		}


		// setup linear system
		Matrix3f S_T = S.transpose();
		Matrix6f matM;
		for( unsigned a=0;a<3;a++ )
		{
			for( unsigned b=0;b<3;b++ )
			{
				matM[a][b] = P[a][b];
				matM[a+3][b] = S[a][b];
				matM[a][b+3] = S_T[a][b];
				matM[a+3][b+3] = R[a][b];
			}
		}

		CvMat cvMatInput, cvMatOutput;
		Matrix6f matM_inv;
		cvInitMatHeader( &cvMatInput, 6, 6, CV_32FC1, matM.data());
		cvInitMatHeader( &cvMatOutput, 6, 6, CV_32FC1, matM_inv.data());
		cvInvert(&cvMatInput, &cvMatOutput,CV_SVD_SYM);

		Vector6f coeffs = matM_inv * resB * 0.5;

		float cos_a = cos(coeffs[0]);
		float sin_a = sin(coeffs[0]);
		float cos_b = cos(coeffs[1]);
		float sin_b = sin(coeffs[1]);
		float cos_c = cos(coeffs[2]);
		float sin_c = sin(coeffs[2]);
		Matrix3f delta_rotation = 
			makeMatrix3f(cos_a, -sin_a,0,
			sin_a, cos_a, 0,
			0,0,1) * 
			makeMatrix3f(cos_b, 0, sin_b,
			0,1,0,
			-sin_b, 0, cos_b) *
			makeMatrix3f(1, 0, 0,
			0, cos_c, -sin_c,
			0, sin_c, cos_c );

		


		
		Matrix3f rot=   makeMatrix3f(   1     , -coeffs[0],  coeffs[1],
									 coeffs[0],    1      , -coeffs[2],
									 -coeffs[1],  coeffs[2],      1     );
// 		debugOutput << "\nvorher:\n";
// 		debugOutput << rot;

		Matrix3f U,V;
		Vector3f eigenValues;
		CvMat cvMatInputR, cvMat_U, cvMat_W, cvMat_Vt;
		cvInitMatHeader( &cvMatInputR, 3, 3, CV_32FC1, rot.data());
		cvInitMatHeader( &cvMat_W, 3, 1, CV_32FC1, eigenValues.data() );
		cvInitMatHeader( &cvMat_U, 3, 3, CV_32FC1, U.data() );
		cvInitMatHeader( &cvMat_Vt, 3, 3, CV_32FC1, V.data() );



		// Do SVD
		cvSVD(&cvMatInputR, &cvMat_W, &cvMat_U, &cvMat_Vt, CV_SVD_MODIFY_A | CV_SVD_U_T );

		delta_rotation = U*V;
		delta_rotation = delta_rotation.transpose();

// 		debugOutput << "\nnachher:\n";
// 		debugOutput << delta_rotation << "\n";

		Matrix4f deltaTransformation = makeTranslation4f(makeVector3f(coeffs[3],coeffs[4],coeffs[5]))
			* expand3To4(delta_rotation);

		refToLocal = deltaTransformation * refToLocal;
		sourceFrame = invertFrame(localToRef*refToLocal)*initialSourceFrame;
		trans = destFrame * invertFrame(sourceFrame);	
		lastResidual=residual;
		if (visualization) debugRenderer->clearDebugData( icpPoints );

	}
	
	originalTransformation=trans;  // xxx remove


	if (visualization)debugRenderer->clearDebugData( icpPoints );
	return 1e20f;
}

X4::float32 X4::FSIcp::calculateResidualAfterICP( const std::vector<Vector3f> &positions, Matrix4f& originalTransformation )
{
	Matrix4f trans=originalTransformation;
	Matrix4f initialTransformation = trans;

	float residual = 0;
	float lastResidual = 1e20f;

	if (m_MedPointDistance==0.0f)
	{
		if( targetPS )
			m_MedPointDistance= getMedianPointDistance(targetPS);
		else
			throw X4Exception("FSIcp::calculateResidualAfterICP() - Median Point Distance is not defined!");
	}

	// compute bounding box of sartPS
	mpcard numPoints = positions.size();
	BoundingBox3f bb(positions[0]);
	for( unsigned i=1;i<numPoints;i++ )
		bb.addPoint(positions[i]);

	startPScenter=bb.getCenter();
	startPScenterOrgTransformed=transformVector3f(originalTransformation, startPScenter);


	bb.makeBoundingCube();


	// compute transformation from world space to normed local space [-1,1]?
	//Matrix3f relativeRotation = shrink4To3(initialTransformation);
	Matrix4f sourceFrame = makeTranslation4f(bb.getCenter());
	Matrix4f initialSourceFrame = sourceFrame;
	Matrix4f destFrame = initialTransformation * sourceFrame;

	float scaleFactor = bb.getSideLength(0) / 2.0f;
	Matrix4f destToLocal = makeScale4f(makeVector3f(1,1,1) / scaleFactor)* invertFrame(destFrame);
	Matrix4f localToDest = destFrame * makeScale4f(makeVector3f(1,1,1) * scaleFactor);
	Matrix4f refToLocal = makeScale4f(makeVector3f(1,1,1) / scaleFactor)* invertFrame(sourceFrame);
	Matrix4f localToRef = sourceFrame * makeScale4f(makeVector3f(1,1,1) * scaleFactor);
	float summed_weight = 0;
	std::string icpPoints = "debug icp";



	for( unsigned pass=0;pass<maxICPSteps;pass++)
	{

		if( visualization )
			debugRenderer->beginRenderJob(icpPoints);
		if( visualization )
			debugRenderer->addBoundingBox(bb, makeVector3f(1,1,1));

		residual = 0;
		summed_weight = 0;
		Matrix3f zeroMat = IDENTITY3F * 0;
		Matrix3f P = zeroMat;
		Matrix3f S = zeroMat;
		Matrix3f R = zeroMat;
		Vector6f resB = NULL_VECTOR6F;

		// calculate corresponding points
		for( unsigned i=0;i<numPoints;i++)
		{
			Vector3f point = positions[i];
			Vector3f transformedPos = transformVector3f(trans, point);
			//			Vector3f transformedNormal = relativeRotation * lf->m_Normal;

			// seek closest point
			Vector3f closestPoint,closestPointNormal;
			if( useANN )
			{
				card32 transformedPointIndex = annQuery->getNearestPointIndex(transformedPos);
				closestPoint = targetPS->get3f(transformedPointIndex, POSITION);
				closestPointNormal =  targetPS->get3f(transformedPointIndex, NORMAL);
			}
			else
			{
				hIt->setSeekPointAndReset(transformedPos);
				closestPoint = hIt->get3f(POSITION);
				closestPointNormal = hIt->get3f(NORMAL);
			}

			//float surfelWeight=1.0f;
			float surfelWeight =norm(closestPoint-transformedPos) > m_MedPointDistance * outlierDistanceFactor ? 0.0f : 1.0f;
			if(surfelWeight>0.0f){
				summed_weight+=surfelWeight;
				float weight = surfelWeight;



				// compute local quadratic approximation
				// M =  n n^t
				// F(x) = (x-p)^tM(x-p) = x^tMx - 2x^tMp + ...
				Vector3f relP = transformVector3f(refToLocal, point);
				Matrix3f matM = outerProduct(closestPointNormal, closestPointNormal );
				Vector3f destPosLocal = transformVector3f(destToLocal, closestPoint);
				Vector3f b = matM * destPosLocal * (-2.0f);

				// F(x) = Ax?+ Bxy + Cy?+ Dxz + Eyz + Fz?+ Gx + Hy + Iz (Mitra et al 2004)
				float A = matM[0][0];
				float B = matM[1][0] + matM[0][1];
				float C = matM[1][1];
				float D = matM[2][0] + matM[0][2];
				float E = matM[2][1] + matM[1][2];
				float F = matM[2][2];
				float G = b[0];
				float H = b[1];
				float I = b[2];

				// compute derivates
				float x = relP[0];
				float y = relP[1];
				float z = relP[2];
				float J = 2.0f * (C*x*x - B*x*y + A*y*y);
				float K = 2.0f * (F*x*x - D*x*z + A*z*z);
				float L = 2.0f * (F*y*y - E*y*z + C*z*z);
				float M = -E*x*x + D*x*y + B*x*z - 2.0f*A*y*z;
				float N = -D*y*y + E*x*y -2.0f*C*x*z + B*y*z;
				float T = -B*z*z -2.0f*F*x*y + E*x*z + D*y*z;
				float U = B*(x*x-y*y) + 2.0f*(C-A)*x*y + E*x*z - D*y*z + H*x - G*y;
				float V = D*(z*z-x*x) - E*x*y + 2.0f*(A-F)*x*z + B*y*z -I*x + G*z;
				float W = E*(y*y-z*z) + D*x*y - B*x*z + 2.0f*(F-C)*y*z + I*y - H*z;

				// setup linear system
				P += makeMatrix3f(J, M, N, M, K, T, N, T, L) * weight;
				S += makeMatrix3f(B*x - A*y*2.0f, C*x*2.0f - B*y, E*x - D*y,
					-D*x + A*z*2.0f, -E*x + B*z, -2.0f*F*x + D*z,
					D*y - B*z, E*y - 2.0f*C*z, 2.0f*F*y - E*z) * weight;
				R += makeMatrix3f(2.0f*A, B, D, 
					B, 2.0f*C, E, 
					D, E, 2.0f*F) * weight;
				resB -= makeVector6f(U,V,W,
					2.0f*A*x + B*y + D*z + G,
					B*x + 2.0f*C*y + E*z + H,
					D*x + E*y + 2.0f*F*z + I) * weight;


				// current residual
				summed_weight +=  weight;
			}

			// project onto surfel
			Vector3f projectedPoint = transformedPos - closestPointNormal * (closestPointNormal * (transformedPos- closestPoint));
			float projectedDistance = norm(transformedPos - projectedPoint);//*(transformedPos - projectedPoint);
			if (visualization) {//
				debugRenderer->addSphere(transformedPos, m_MedPointDistance*2.6f ,makeVector3f(0.3f,0.7f*surfelWeight,0.86f));
				Vector3f col = makeVector3f(projectedDistance*3000, 0 ,0);
				debugRenderer->addLine(transformedPos, projectedPoint, col, col,1);

			}



			// current residual
			residual += projectedDistance; 

		}



		if (summed_weight<minPercentInliers*(float32)numPoints)
		{
			residual = 1e20f;
		}

		if (visualization) {
			debugRenderer->endRenderJob();
			ObjectViewsTable::update(scene);
		}

		if (!checkAllowedMovement(trans)){
			if (visualization) debugRenderer->clearDebugData( icpPoints );	// clear data from last call
			return 1e20f;
		}

		if(pass>0 && fabs(lastResidual-residual) < icpConvergenceDiff )
		{

			if (visualization) debugOutput << "converged after " <<pass<< " steps\n";
			if (visualization)debugRenderer->clearDebugData( icpPoints );	// clear data from last call

			if(false){
				AAT COL = targetPS->getAAT("color");
				for( unsigned i=0;i<numPoints;i++)
				{
					
					Vector3f point = positions[i];
					Vector3f transformedPos = transformVector3f(trans, point);
					Vector3f closestPoint,closestPointNormal;

					card32 transformedPointIndex = annQuery->getNearestPointIndex(transformedPos);
					closestPoint = targetPS->get3f(transformedPointIndex, POSITION);
					closestPointNormal =  targetPS->get3f(transformedPointIndex, NORMAL);


					Vector3f projectedPoint = transformedPos - closestPointNormal * (closestPointNormal * (transformedPos- closestPoint));
					float projectedDistance = norm(transformedPos - projectedPoint);
					float colVal=projectedDistance/ (2*m_MedPointDistance);
					Vector3f col = shrink4To3(rainbowGradient(colVal));
					targetPS->set3f(transformedPointIndex, COL, col );
				}


			}

			originalTransformation=trans;
			return residual;

		}


		// setup linear system
		Matrix3f S_T = S.transpose();
		Matrix6f matM;
		for( unsigned a=0;a<3;a++ )
		{
			for( unsigned b=0;b<3;b++ )
			{
				matM[a][b] = P[a][b];
				matM[a+3][b] = S[a][b];
				matM[a][b+3] = S_T[a][b];
				matM[a+3][b+3] = R[a][b];
			}
		}

		CvMat cvMatInput, cvMatOutput;
		Matrix6f matM_inv;
		cvInitMatHeader( &cvMatInput, 6, 6, CV_32FC1, matM.data());
		cvInitMatHeader( &cvMatOutput, 6, 6, CV_32FC1, matM_inv.data());
		cvInvert(&cvMatInput, &cvMatOutput,CV_SVD_SYM);

		Vector6f coeffs = matM_inv * resB * 0.5;

		float cos_a = cos(coeffs[0]);
		float sin_a = sin(coeffs[0]);
		float cos_b = cos(coeffs[1]);
		float sin_b = sin(coeffs[1]);
		float cos_c = cos(coeffs[2]);
		float sin_c = sin(coeffs[2]);

		Matrix3f delta_rotation = 
			makeMatrix3f(cos_a, -sin_a,0,
			sin_a, cos_a, 0,
			0,0,1) * 
			makeMatrix3f(cos_b, 0, sin_b,
			0,1,0,
			-sin_b, 0, cos_b) *
			makeMatrix3f(1, 0, 0,
			0, cos_c, -sin_c,
			0, sin_c, cos_c );

		Matrix4f deltaTransformation = makeTranslation4f(makeVector3f(coeffs[3],coeffs[4],coeffs[5]))
			* expand3To4(delta_rotation);


		refToLocal = deltaTransformation * refToLocal;
		sourceFrame = invertFrame(localToRef*refToLocal)*initialSourceFrame;
		trans = destFrame * invertFrame(sourceFrame);	
		lastResidual=residual;
		if (visualization) debugRenderer->clearDebugData( icpPoints );

	}

	originalTransformation=trans;  // xxx remove


	if (visualization)debugRenderer->clearDebugData( icpPoints );
	return 1e20f;
}



bool X4::FSIcp::checkAllowedMovement(const Matrix4f &trans)
{
	Vector3f transformedCenter =transformVector3f(trans, startPScenter);

	Vector3f moveVecA=transformedCenter-startPScenterOrgTransformed;
	float32 moveDist=norm(moveVecA);

	if (false && visualization)
	{
		std::string myid = "move";
		debugRenderer->clearDebugData( myid );	// clear data from last call
		debugRenderer->beginRenderJob( myid );	// specify job name 

		debugRenderer->addLine(transformedCenter, startPScenter, makeVector3f(1,0,0),makeVector3f(0,1,0),5);
		
		debugRenderer->endRenderJob();
		ObjectViewsTable::update(scene);
	}


	if (moveDist>icpAllowedMoveDist )
	{
		return false;
	}else
		return true;
}

void X4::FSIcp::setMedianPointDistance( float32 d )
{
	m_MedPointDistance=d;
}

X4::float32 X4::FSIcp::getMedianPointDistance( PointSet* ps )
{
	std::vector<float> distances;
	distances.resize( min( (card32)ps->getNumEntries(), (card32)1000) * 5 );
	mpcard targetIndex=0;

	mpcard numPts=ps->getNumEntries();
	for( card32 i=0;i<1000;i++ )
	{

		Vector3f samplePos = ps->get3f(random64bit(numPts), POSITION);
		annQuery->setTargetPoint(samplePos);
		card32 index; bool dummy;
		annQuery->getNextNearestPointIndex(index, dummy);

		for( card32 j=1;j<5;j++ )
		{		
			annQuery->getNextNearestPointIndex(index, dummy);
			distances[targetIndex++] = norm( samplePos - ps->get3f(index, POSITION) );
		}
	}

	
	sort( distances.begin(), distances.end() );

	return distances[ distances.size() / 2];
}

float32 X4::FSIcp::getMedianPointDistance() const
{
	return m_MedPointDistance;
}
X4::float32 X4::FSIcp::calculateResidualAfterStandardICP( const std::vector<Vector3f> &positions, Matrix4f& originalTransformation )
{


	//+===================

	Matrix4f trans=originalTransformation;


	float residual = 0;
	float lastResidual = 1e20f;

	mpcard numPoints = positions.size();


	std::vector<Vector3f> targetPositions, transformedA;
	targetPositions.resize(numPoints);
	transformedA.resize(numPoints);

	for( unsigned pass=0;pass<maxICPSteps;pass++)
	{

		float residual = 0;
		if( visualization )
			debugRenderer->beginRenderJob("icpPoints");

		Vector3f centerA= NULL_VECTOR3F;
		Vector3f centerB= NULL_VECTOR3F;

		for( mpcard i=0;i<numPoints;i++)
		{
			Vector3f point = positions[i];
			Vector3f transformedPos = transformVector3f(trans, point);
			//			Vector3f transformedNormal = relativeRotation * lf->m_Normal;

			// seek closest point
			Vector3f closestPoint,closestPointNormal;
			getClosestPointAndNormal(transformedPos, closestPoint, closestPointNormal);


			Vector3f projectedPoint = transformedPos - closestPointNormal * (closestPointNormal * (transformedPos- closestPoint));
			float projectedDistance = norm(transformedPos - projectedPoint);//*(transformedPos - projectedPoint);
			residual =+ projectedDistance;
			if (visualization) {//debugRenderer->addSphere(transformedPos, m_MedPointDistance*0.6f ,makeVector3f(0.3f,0.7f*surfelWeight,0.86f));
				Vector3f col = makeVector3f(projectedDistance*3000, 0 ,0);
				debugRenderer->addLine(transformedPos, projectedPoint, col, col,1);
			}

			targetPositions[i]=projectedPoint;
			transformedA[i]=transformedPos;

			centerA += transformedPos;
			centerB += projectedPoint;
		}
		centerA/= (float)numPoints;
		centerB/= (float)numPoints;


		// calculate cross-covariance matrix
		Matrix3f C ;
		C.setZero();
		for( card32 i=0;i<numPoints;i++ )
		{
			for( card32 j=0;j<3;j++ )
			{
				for( card32 k=0;k<3;k++ )
				{
					C[j][k] += (transformedA[i][j] - centerA[j])
						* (targetPositions[i][k] - centerB[k]);
				}
			}
		}

		// Prepare OpenCV data structure CvMat
		Matrix3f U,V;
		Vector3f eigenValues;
		CvMat cvMatInput, cvMat_U, cvMat_W, cvMat_Vt;
		cvInitMatHeader( &cvMatInput, 3, 3, CV_32FC1, C.data());
		cvInitMatHeader( &cvMat_W, 3, 1, CV_32FC1, eigenValues.data() );
		cvInitMatHeader( &cvMat_U, 3, 3, CV_32FC1, U.data() );
		cvInitMatHeader( &cvMat_Vt, 3, 3, CV_32FC1, V.data() );

		// Do SVD
		cvSVD(&cvMatInput, &cvMat_W, &cvMat_U, &cvMat_Vt, CV_SVD_MODIFY_A | CV_SVD_U_T );

		Matrix3f rotation = U*V;
		trans = makeTranslation4f(centerB) * expand3To4(rotation.transpose()) * makeTranslation4f(-centerA) *trans;



		if (visualization) {
			debugRenderer->endRenderJob();
			ObjectViewsTable::update(scene);
		}

		

		if(pass>0 && fabs(lastResidual-residual) < icpConvergenceDiff )
		{

			if (visualization) debugOutput << "converged after " <<pass<< " steps\n";
			if (visualization)debugRenderer->clearDebugData( "icpPoints" );	// clear data from last call

			originalTransformation=trans;
			return residual;

		}

		lastResidual=residual;
		if (visualization) debugRenderer->clearDebugData( "icpPoints" );

	}

	originalTransformation=trans;  // xxx remove


	if (visualization)debugRenderer->clearDebugData( "icpPoints" );
	return 1e20f;

}



X4::float32 X4::FSIcp::calculateResidualAfterStandardICP( PointSet* startPS, Matrix4f& transformation )
{
	std::vector<Vector3f> positions;
	AAT POS = startPS->getAAT("position");
	for (mpcard i = 0; i < startPS->getNumEntries(); i++)
	{
		positions.push_back(startPS->get3f(i, POS));
	}
	return calculateResidualAfterStandardICP(positions, transformation);
}
void X4::FSIcp::getClosestPointAndNormal( Vector3f transformedPos, Vector3f &closestPoint, Vector3f &closestPointNormal )const
{
	if( useANN )
	{
		card32 transformedPointIndex = annQuery->getNearestPointIndex(transformedPos);
		closestPoint = targetPS->get3f(transformedPointIndex, POSITION);
		closestPointNormal =  targetPS->get3f(transformedPointIndex, NORMAL);
	}
	else
	{
		hIt->setSeekPointAndReset(transformedPos);
		closestPoint = hIt->get3f(POSITION);
		closestPointNormal = hIt->get3f(NORMAL);
	}
}