#ifndef FastSpatialQueryH
#define FastSpatialQueryH

#include "StdAfx.h"

//---------------------------------------------------------------------------
#include "x3foundation_private.h"
#include "Circle3D.h"
#include "AttachedIndexedOctree.h"
#include "OOBBox3f.h"
#include "Plane.h"

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

namespace X4
{

class AttachedIndexedOctree;
class PointCloud;
class SecondaryDataStructureIterator;
class SceneObjectIterator;
class BasicPointCloudIterator;


/**
	stack that keeps internal memory reserved.
	faster but much more memory consuming
*/
template<class T> 
class CustomMemoryStack
{
	std::vector<T> data;
	mpcard mySize;

public:
	CustomMemoryStack() {mySize = 0;}

	void clear()
	{
		mySize = 0;
	}

	inline void push_back(const T& v)
	{
		if( data.size() == mySize )
			data.push_back(v);

		data[mySize] = v;
		mySize++;
	}

	inline T pop_back()
	{
		mySize--;
		return data[mySize];
	}

	inline bool empty() const 
	{
		return mySize == 0;
	}

};


/**
	fast spatial point queries based on octree structure (AIOctree)
*/
class X3FOUNDATION_PRIVATE_API FastSpatialQuery 
{

public:
	FastSpatialQuery(PointCloud * pc);
	~FastSpatialQuery(void);

	AttachedIndexedOctree * getOctree() {return m_Octree;}


	/// queries all points near line p+dir*t
	/// indices are valid until next call or destruction of this object or another query call
	void lineQuery( const Vector3f &p, const Vector3f &dir, float maxDistance, mpcard ** indices, mpcard &numPoints);

	/// queries all points near plane
	/// indices are valid until next call or destruction of this object or another query call
	void planeQuery( const Plane &plane, float maxDistance,mpcard ** indices, mpcard &numPoints);
	/// queries all points near plane within sphere
	void planeAndSphereQuery( const Plane &plane, float maxDistance, const Vector3f &sphere_center, float sphere_radius, mpcard ** indices, mpcard &numPoints);

	/// queries all points in torus (all p with circle.distance(p) < tubeRadius)
	void torusQuery( const Circle3D & circle, float tubeRadius, mpcard ** indices, mpcard &numPoints );

	/// queries all points near triangle
	void triangleQuery( const Vector3f p0, const Vector3f p1, const Vector3f p2, float maxDistance,mpcard ** indices, mpcard &numPoints);

	/// oobb query
	void oobbQuery( const OOBBox3f &oobb, mpcard ** indices, mpcard &numPoints);

private:

	inline void addIndexToCache(const mpcard i)
	{
		if( m_NumPoints == m_IndexCache.size() )
		{
			m_IndexCache.resize(m_NumPoints*2);

		}
		m_IndexCache[m_NumPoints++] = i;
	}

	std::vector<mpcard>	m_IndexCache;
	mpcard			m_NumPoints;

	AttachedIndexedOctree	*	m_Octree;
	SecondaryDataStructureIterator	* m_SDS;
	SceneObjectIterator		*	m_SceneIter;
	BasicPointCloudIterator *	m_BasicIt;
	PointSet		*			m_PS;

	CustomMemoryStack<AIOctreeNode*>	m_NodeStack;

};

}

#endif