#include "StdAfx.h"
#include <windows.h>
#include <mmsystem.h>
//---------------------------------------------------------------------------
#include "FeatureMatching.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "SpectralCorrespondenceValidation.h"
#include "GeometricFeature.h"
#include "FeatureCorrespondence.h"
#include "MultiScaleFeatureDetector.h"
#include "ObjectClassProperty.h"
#include "PointCloud.h"
#include "SlippageFeatureDetector.h"
#include "debugRenderer.h"
//---------------------------------------------------------------------------
using namespace X4;

IMPLEMENT_X4_CLASS( FeatureMatching ,2)
{
	BEGIN_CLASS_INIT( FeatureMatching );

	ADD_OBJECT_PROP( m_FeatureDetector, 0, FeaturePointDetector::getClass(), true);
	ADD_OBJECT_PROP( m_SpectralCV, 0, SpectralCorrespondenceValidation::getClass(), true);
	SET_USER_EDIT(false);
	ADD_CARD32_PROP(m_MaxNumCorrespondingFeatures,1);
	SET_USER_EDIT(false);
	ADD_CARD32_PROP(m_MaxNumInitialCorrespondences,2);
	SET_USER_EDIT(false);
}

FeatureMatching::FeatureMatching(void)
{
	m_FeatureDetector = new SlippageFeatureDetector;
	m_SpectralCV = new SpectralCorrespondenceValidation;
	m_MaxNumCorrespondingFeatures = 10;
	m_MaxNumInitialCorrespondences = 500;
}

FeatureMatching::~FeatureMatching(void)
{
	delete m_FeatureDetector;
	delete m_SpectralCV;
}

void FeatureMatching::assign(const Object* obj, X4CopyContext *context)
{
	const FeatureMatching * o = dynamic_cast<const FeatureMatching*>(obj);
	x4Assert( o != NULL );

	copyObjectProperties(obj,this);
}

X4::FeatureCorrespondenceSet * X4::FeatureMatching::computeInitialCorrespondences( FeatureSet * f1, FeatureSet * f2, FeaturePointDetector* fd )
{
	FeatureCorrespondenceSet * ret = new FeatureCorrespondenceSet;
   FeaturePointDetector* detector = fd ? fd : m_FeatureDetector;
   
   //float maxScore = 0;
   
	for( unsigned i=0;i<f1->m_Features.size();i++ )
	{
		GeometricFeature * feature1 = f1->m_Features[i];
		for( unsigned j=0;j<f2->m_Features.size();j++ )
		{
			GeometricFeature * feature2 = f2->m_Features[j];
			float score = feature1->compare(feature2, detector);
			if( score )
			{
				FeatureCorrespondence * fcorr = new FeatureCorrespondence;
				fcorr->m_FeaturePoint[0] = feature1->m_Position;
				fcorr->m_FeaturePoint[1] = feature2->m_Position;
				fcorr->m_FeatureIndex[0] = i;
				fcorr->m_FeatureIndex[1] = j;
            
				fcorr->m_Distance = score;
				ret->m_Correspondences.push_back( fcorr );
				
				//if (score > maxScore) maxScore = score;
			}
		}
	}

	return ret;
}

FeatureCorrespondenceSet * X4::FeatureMatching::computeInitialCorrespondences( PointCloud * pc1,PointCloud * pc2, FeaturePointDetector* fd )
{
	FeatureSet * f1 = dynamic_cast<FeatureSet*>(pc1->getAttachments()->getData(FeatureSet::getDefaultName()));
	FeatureSet * f2 = dynamic_cast<FeatureSet*>(pc2->getAttachments()->getData(FeatureSet::getDefaultName()));
	
	if( f1 == NULL || f2 == NULL )
		throw "No feature attachments found!";

	return computeInitialCorrespondences(f1,f2,fd);
}

X4::FeatureCorrespondenceSet * X4::FeatureMatching::computeValidatedCorrespondences( FeatureSet * f1, FeatureSet * f2, FeaturePointDetector* fd )
{
	FeatureCorrespondenceSet * initialSet = computeInitialCorrespondences(f1,f2,fd);
	FeatureCorrespondenceSet * ret = computeValidatedCorrespondences(f1,f2,initialSet,fd);
	delete initialSet;
	return ret;
}

FeatureCorrespondenceSet * X4::FeatureMatching::computeValidatedCorrespondences( PointCloud * pc1,PointCloud * pc2, FeaturePointDetector* fd )
{
	FeatureSet * f1 = dynamic_cast<FeatureSet*>(pc1->getAttachments()->getData(FeatureSet::getDefaultName()));
	FeatureSet * f2 = dynamic_cast<FeatureSet*>(pc2->getAttachments()->getData(FeatureSet::getDefaultName()));

	if( f1 == NULL || f2 == NULL )
		throw "No feature attachments found!";

	return computeValidatedCorrespondences(f1,f2,fd);

}

FeatureCorrespondenceSet * X4::FeatureMatching::computeValidatedCorrespondences( FeatureSet * f1, FeatureSet * f2,FeatureCorrespondenceSet * initialSet, FeaturePointDetector* fd  )
{
	// setup pairs
	m_SpectralCV->setup( (card32)initialSet->m_Correspondences.size(), (card32)f1->m_Features.size(),(card32)f2->m_Features.size());
	for( unsigned i=0;i<initialSet->m_Correspondences.size();i++ )
	{
		FeatureCorrespondence * fcorr = initialSet->m_Correspondences[i];
		m_SpectralCV->setCorrespondencePair(i, fcorr->m_FeatureIndex[0],fcorr->m_FeatureIndex[1],
			fcorr->m_FeatureIndex[0],fcorr->m_FeatureIndex[1],0.0f, fcorr->m_FeaturePoint[0], fcorr->m_FeaturePoint[1]);

	}

	bool hasGeoDist = f1->m_GeodeticDistances.size() > 0;

	// setup distance matrix1,2
	for( unsigned i=0;i<f1->m_Features.size();i++ )
	{
		for( unsigned j=0;j<f1->m_Features.size();j++ )
		{
			float dist = norm( f1->m_Features[i]->m_Position - f1->m_Features[j]->m_Position);
			float geo_dist = dist;
			if( hasGeoDist )
				geo_dist = f1->getGeodeticDistance(i,j);
			m_SpectralCV->setDistanceEntry1(i,j, geo_dist, dist );
			m_SpectralCV->setNormalAngle1(i,j,f1->m_Features[i]->m_Normal * f1->m_Features[j]->m_Normal );
		}

	}
	for( unsigned i=0;i<f2->m_Features.size();i++ )
	{
		for( unsigned j=0;j<f2->m_Features.size();j++ )
		{
			float dist = norm( f2->m_Features[i]->m_Position - f2->m_Features[j]->m_Position);
			float geo_dist = dist;
			if( hasGeoDist )
				geo_dist = f2->getGeodeticDistance(i,j);
			m_SpectralCV->setDistanceEntry2(i,j, geo_dist, dist );
			m_SpectralCV->setNormalAngle2(i,j,f2->m_Features[i]->m_Normal * f2->m_Features[j]->m_Normal );
		}

	}


	// validate correspondences
	m_SpectralCV->calculate();

	// store valid correspondences
	FeatureCorrespondenceSet * ret_validated = new FeatureCorrespondenceSet;
	std::vector<SpectralCorrespondenceValidation::TFeaturePair> validPairs = 
		m_SpectralCV->getValidPairs();

	for( unsigned i=0;i<validPairs.size();i++ )
	{
		FeatureCorrespondence * fcorr = new FeatureCorrespondence;
		fcorr->m_FeaturePoint[0] = validPairs[i].p1;
		fcorr->m_FeatureIndex[0] = validPairs[i].i1;
		fcorr->m_FeaturePoint[1] = validPairs[i].p2;
		fcorr->m_FeatureIndex[1] = validPairs[i].i2;
		ret_validated->m_Correspondences.push_back(fcorr);
	}

	return ret_validated;

}