#ifndef FeaturePointDetectorH
#define FeaturePointDetectorH
//---------------------------------------------------------------------------
#include "x3foundation_private.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
//---------------------------------------------------------------------------

namespace X4
{

class ProgressWindow;
class FeatureSet;
class LocalDescriptor;
class UnstructuredInCorePointCloud;

/**
	base class for feature detection in point clouds - 
	computes a set of features with signatures, first signature is always
	created by the current detector.
*/
class X3FOUNDATION_PRIVATE_API FeaturePointDetector : public Persistent
{
	X4_ABSTRACT_CLASS( FeaturePointDetector )

protected:

	// properties
	std::vector<LocalDescriptor*>	m_Descriptors;
	float32			m_DescriptorRadius;	// relative to median point distance

	// members
	float32			m_MedianPointDistance;

	/// virtual version of "begin"
	virtual void beginDetection(const std::vector<UnstructuredInCorePointCloud*>& clouds) {}

	// progress widget which can be used to indicate job progress
	ProgressWindow* progress;

public:
	FeaturePointDetector(void);
	~FeaturePointDetector(void);

   virtual void assign(const Object* obj, X4CopyContext *context = NULL);

	/// setup feature detection
	/// some detectors need a set of representative clouds
	/// one may derive beginDetection()
	void begin( const std::vector<UnstructuredInCorePointCloud*>& clouds );


	/// cleanup temporal data
	virtual void end() {}

	/// detect features in point cloud
	/// FeatureSet is also attached to the point cloud
	/// when attachmentName is not defined, then the default name is used
	virtual FeatureSet * detect( UnstructuredInCorePointCloud * upc, std::string attachmentName="" ) = 0;

	/// recomputes feature signatures for every feature 
	/// first signature (from the detector) is not modified
	virtual void recomputeFeatureSignatures( UnstructuredInCorePointCloud * upc, std::string attachmentName="" );
	virtual void recomputeFeatureSignatures( UnstructuredInCorePointCloud * upc, FeatureSet * fs);

	mpcard getNumLocalDescriptors() const {return m_Descriptors.size();}
	LocalDescriptor * getLocalDescriptor( unsigned index) {return m_Descriptors[index];}

	float32 getDescriptorRadius() const {return m_DescriptorRadius;}
	void setDescriptorRadius(float32 v) {m_DescriptorRadius=v;}

	void addDescriptor( Object * obj ) {m_Descriptors.push_back((LocalDescriptor*)obj);}
	void deleteDescriptors( card32 begin, card32 end );
	Object * getDescriptor( card32 index );
	void moveDescriptor( card32 oldPos, card32 newPos ) {}
	mpcard getNumDescriptors() {return m_Descriptors.size();}
	std::string getDescriptorName(card32 index ) {return "-";}

	/// If progress window is set, it can then be used by derived classes to indicate progress
	void setProgress(ProgressWindow* prg) { progress = prg; }
	ProgressWindow* getProgress() { return progress; }

};

}

#endif