#ifndef FiloStackH
#define FiloStackH

// System
#include <assert.h>

#include "LinearAlgebra.h"

namespace X4 {

template <class ItemType>
struct FiloStackElement
{
	/// Item
	ItemType item;
	/// Stack management: lower and upper item in the stack
	FiloStackElement<ItemType> *upper, *lower;
};

/**
 * Provides a stack of arbitrary items
 */
template <class ItemType>
class  FiloStack
{
	private:
		/// Pointer towards start and end of list
		FiloStackElement<ItemType> *top, *bottom;
		/// Free all allocated memory
		void freeMemory();
		/// Number of items on the stack
		int numItems;

	public:
		/// Constructor
		FiloStack();
		/// Destructor
		~FiloStack();
		/// Empty stack
		void clear();
		/// Put new item in top of stack
		void push( ItemType item );
		/// Pop lowest item from stack
		ItemType pop();
		/// Test if stack is empty
		inline bool empty() const { return bottom == NULL; }
		/// Debugging purpose
		void printDebugInfo() const;
		/// Return number of items on stack
		inline int getNumItems() const { return numItems; }
		/// Return the second item in the stack
		inline ItemType getTop(){ return top->item; }
		/// Return the second item in the stack
		inline ItemType getSecond(){ return top->lower->item; }
		/// Create a copy of the stack
		FiloStack copy();
		/// Return lowest item from stack; no pop'ing
		ItemType next() const;
};

/**
 * Create new SortedLinkedList object
 */
template<class ItemType>
FiloStack<ItemType>::FiloStack()
{
	// Init empty stack
	top = bottom = NULL;
   numItems = 0;
}

/**
 * Delete object
 */
template<class ItemType>
FiloStack<ItemType>::~FiloStack()
{
	freeMemory();
}

/**
 * Free allocated memory
 */
template<class ItemType>
void FiloStack<ItemType>::freeMemory()
{
	/*
	// Free mem for each item in stack
	FiloStackElement<ItemType> *e = top, *next;
	while ( e != NULL ){
		next = e->lower;
		delete e;
		e = next;
	}
	*/
	top = bottom = NULL;
   numItems = 0;
}

/**
 * Empty stack
 */
template<class ItemType>
void FiloStack<ItemType>::clear()
{
   freeMemory();
}

/**
 * Put new item on top of stack
 */
template<class ItemType>
void FiloStack<ItemType>::push( ItemType item )
{
	// Create new item
	FiloStackElement<ItemType>* e = new FiloStackElement<ItemType>;
	e->item = item;
	// Update data structure
	if ( top == NULL ){
		// First and only item in the stack
		bottom = top = e;
		e->upper = NULL;
		e->lower = NULL;
	} else {
		// At least one element already in stack
		top->upper = e;
		e->lower = top;
		e->upper = NULL;
		top = e;
	}
   numItems++;
}

/**
 * Remove the lowest element in the stack and return item
 */
template<class ItemType>
ItemType FiloStack<ItemType>::pop()
{
	// Never pop an empty stack!
	assert( top != NULL );
	// Temp save bottom element
	FiloStackElement<ItemType>* e = top;
	ItemType item = e->item;
	// Move bottom pointer up
	top = top->lower;
	if ( top == NULL ){
		// No more elements in the stack
		bottom = NULL;
	} else {
		// At least one element remains, update bottom of stack
		top->upper = NULL;
	}
   numItems--;
	// Free allocated memory
	delete e;
	return item;
}

template<class ItemType>
ItemType FiloStack<ItemType>::next() const
{
	assert( top != NULL );
	return top->item;
}

/**
 * Print information about the state of the stack
 */
template<class ItemType>
void FiloStack<ItemType>::printDebugInfo() const
{	
	std::cout << "* FiloStack state * " << endl;
	//std::cout << "  Top element:    " << top << endl;
	//std::cout << "  Bottom element: " << bottom << endl;

   FiloStackElement<ItemType>* c = bottom;
   while( c != NULL ){
      debugOutput << c->item << "\n";
      c = c->upper;
   }
}

}

#endif