//---------------------------------------------------------------------------
#include "stdafx.h"
 
//---------------------------------------------------------------------------
#include "GLImageFilter.h"
//---------------------------------------------------------------------------
#include "BooleanClassProperty.h"
#include "ObjectClassProperty.h"
#include "StringClassProperty.h"
#include "FilterMask.h"
#include "GLShaderMaterial.h"
#include "Configuration.h"
#include "StringHelper.h"
#include "GLRenderTextureCache.h"
#include "GLImage.h"
#include "LinearAlgebra.h"
#include "LinearAlgebra.inline.h"
#include "GLHelper.h"
#include "ApplicationDirectories.h"

//---------------------------------------------------------------------------
 
//---------------------------------------------------------------------------
using namespace X4;

MetaClass* GLImageFilter::GLImageFilterMetaClass = NULL;

const MetaClass* GLImageFilter::getInstanceClass() const {
   return getClass();
}

Object* GLImageFilter::newInstance() {
   return new GLImageFilter;
}

const MetaClass* GLImageFilter::getClass() {
   if (GLImageFilterMetaClass==NULL) {
      GLImageFilterMetaClass = new MetaClass("GLImageFilter", "no description.",
                                  newInstance, sizeof(GLImageFilter));
   }
   return GLImageFilterMetaClass;
}

void GLImageFilter::init(const MetaClass *parentClass) {
   ((MetaClass*)getClass())->init(parentClass, MetaClass::FIRST_VERSION + 0);
   GLImageFilterMetaClass->copyParentProperties();
   BEGIN_CLASS_INIT(GLImageFilter)
   ADD_BOOLEAN_PROP_METHOD(m_NormalizeMask,0, &getNormalizeMask, &setNormalizeMask)
   ADD_STRING_PROP_METHOD(m_PostProcess, 0, &getPostProcess, &setPostProcess)
	ADD_OBJECT_PROP(m_FilterMask, 0,FilterMask::getClass(), true)
   ADD_BOOLEAN_PROP_METHOD(m_Modified, 0, &getModified, &setModified)
}

void GLImageFilter::shutdown() {
   delete GLImageFilterMetaClass;
}


// ----

GLImageFilter::GLImageFilter()
{
   m_NormalizeMask = false;
   m_PostProcess = "O = O;";
   m_FilterMask = new FixedFilterMask_9x9;

   loadTemplates();
}

GLImageFilter::~GLImageFilter()
{
   delete m_FilterMask;
}

void GLImageFilter::assign(const Object* obj, X4CopyContext *context)
{
   const GLImageFilter * filter = dynamic_cast<const GLImageFilter*>(obj);
   x4Assert( filter != NULL );

   GLImageMaterialOperator::assign( obj, context );

   delete m_FilterMask;
   m_FilterMask = (FilterMask *)filter->m_FilterMask->copy( context );
   m_PostProcess = filter->m_PostProcess;
   m_NormalizeMask = filter->m_NormalizeMask;
}

GLImage * GLImageFilter::execute(GLRenderTextureCache * cache)
{
   x4Assert( m_InputImages.size() != 0 );

   // get output rendertexture from cache
   GLImage * ret = new GLImage;
   bool isFloat = false;
   bool isFloat32 = false;
   switch( getOutputFormat() )
   {
   case OF_Undefined:
      // take inputformat
      isFloat = m_InputImages[0]->isFloat();
      isFloat32 = m_InputImages[0]->isFloat32();
   break;
   case OF_Float32:
      isFloat32 = true;
   case OF_Float16:
      isFloat = true;
   break;
   }

   card32 width = m_InputImages[0]->getWidth();
   card32 height = m_InputImages[0]->getHeight();
   ret->setup( cache, width , height,
      m_InputImages[0]->getNumComponents(), m_InputImages[0]->isRectangle(),
         isFloat, isFloat32 );

   // #####################################
   // --------- render to texture --------------
   ret->activate();
   glPushAttrib( GL_ALL_ATTRIB_BITS );

   GLHelper::begin2D( width, height );
   bindMaterial( cache->getContext() );

   for( card32 i=0;i<m_InputImages.size();i++ )
   {
       m_InputImages[i]->bind(i);
   }

   glBegin( GL_QUADS );

   std::vector<Vector2f> coords(m_InputImages.size()*2);
   for( card32 i=0;i<m_InputImages.size();i++ )
   {
      coords[i*2+0] = makeVector2f(0,0);
      if( m_InputImages[i]->isRectangle() )
         coords[i*2+1] = makeVector2f(m_InputImages[i]->getWidth(),
                     m_InputImages[i]->getHeight());
      else
         coords[i*2+1] = makeVector2f(1,1);
   }

   for( card32 i=0;i<m_InputImages.size();i++ )
   {
      glVertexAttrib2fARB( i+1,     coords[i*2+0][0],coords[i*2+0][1] );
   }
   glVertexAttrib2fARB(0, 0,0);

   for( card32 i=0;i<m_InputImages.size();i++ )
   {
      glVertexAttrib2fARB( i+1,     coords[i*2+1][0],coords[i*2+0][1] );
   }
   glVertexAttrib2fARB(0, width, 0);

   for( card32 i=0;i<m_InputImages.size();i++ )
   {
      glVertexAttrib2fARB( i+1,     coords[i*2+1][0],coords[i*2+1][1] );
   }
   glVertexAttrib2fARB(0, width, height);

   for( card32 i=0;i<m_InputImages.size();i++ )
   {
      glVertexAttrib2fARB( i+1,     coords[i*2+0][0],coords[i*2+1][1] );
   }
   glVertexAttrib2fARB(0, 0, height);

   glEnd();


   // #####################################
   // ----------- clean up ----------------

   for( card32 i=0;i<m_InputImages.size();i++ )
   {
       m_InputImages[i]->unbind();
   }

   GLHelper::end2D();
   glPopAttrib();

   ret->deactivate();
   unbindMaterial( cache->getContext() );

   return ret;
}


void GLImageFilter::loadTemplates()
{
   m_VSTemplate = readTextFile( ApplicationDirectories::getResourcesDir() + "shaders/hqRenderer/filter_vs.txt" );
   m_PSTemplate = readTextFile( ApplicationDirectories::getResourcesDir() + "shaders/hqRenderer/filter_ps.txt" );
}


void GLImageFilter::modify( std::string &vsTemplate, std::string &psTemplate )
{
   x4Assert( getNumInputImages() > 0 );

   forget(vsTemplate);
   psTemplate = replaceSubStrings( psTemplate, "$PS_POST_PROCESS", m_PostProcess );

   float weight_sum = 1.0f;
   if( m_NormalizeMask )
   {
      FilterMask::Element e;
      m_FilterMask->getFirstSample(e);
      weight_sum = e.weight;
      while( m_FilterMask->getNextSample(e) )
      {
         weight_sum += e.weight;
      }

   }

   std::string filtering;
   FilterMask::Element e;
   m_FilterMask->getFirstSample(e);
   bool rect = m_InputImages[0]->isRectangle();

   do
   {
      float weight = e.weight / weight_sum;

      if( fabs(weight) < 0.01f )
         continue;

      if( filtering.length() == 0 )
         filtering = "O = ";
      else
         filtering += "O += ";

      filtering += numToStr( weight ) + " * tex";
      if( rect )
         filtering += "RECT";
      else
         filtering += "2D";

      std::string x = numToStr(e.x);
      std::string y = numToStr(e.y);
      filtering += "( texture0, IN.tex0.xy + float2("+x+","+y+") );\n";

   }
   while( m_FilterMask->getNextSample(e) );

   if( filtering.length() == 0 )
      filtering = "O = 0.0;\n";

   psTemplate = replaceSubStrings( psTemplate, "$PS_FILTER", filtering );

}

void GLImageFilter::setFilterMask( FilterMask * fm )
{
   delete m_FilterMask;
   m_FilterMask = fm;
}



