//---------------------------------------------------------------------------
#include "stdafx.h"
 
//---------------------------------------------------------------------------
#include "GLImageSummarizer.h"
//---------------------------------------------------------------------------
#include "GLImage.h"
#include "GLRenderTextureFBO.h"
#include "GLRenderTextureCache.h"
#include "Configuration.h"
#include "GLImageReducer.h"
#include "RGBAImage.h"
#include "LinearAlgebra.h"
#include "LinearAlgebra.inline.h"
//---------------------------------------------------------------------------
 
//---------------------------------------------------------------------------
using namespace X4;

GLImageSummarizer::GLImageSummarizer( EMode mode )
{
   m_Mode = mode;
   GLImageReducer * reducer = new GLImageReducer;
   reducer->setNumInputImages(1);
   reducer->setOutputFormat( GLImageOperator::OF_Float16 );
   switch( mode )
   {
      case Mode_Sum:
         reducer->setReduceOp( "O = I0 + I1 + I2 + I3;" );
      break;
      case Mode_Average:
         reducer->setReduceOp( "O = (I0 + I1 + I2 + I3)*0.25;" );
      break;
      case Mode_Minimum:
         reducer->setReduceOp(
            "O.x = min( min(I0.x,I1.x), min(I2.x,I3.x));\n"
            "O.y = min( min(I0.y,I1.y), min(I2.y,I3.y));\n"
            "O.z = min( min(I0.z,I1.z), min(I2.z,I3.z));\n"
            "O.w = min( min(I0.w,I1.w), min(I2.w,I3.w));\n"
            );
      break;
      case Mode_Maximum:
         reducer->setReduceOp(
            "O.x = max( max(I0.x,I1.x), max(I2.x,I3.x));\n"
            "O.y = max( max(I0.y,I1.y), max(I2.y,I3.y));\n"
            "O.z = max( max(I0.z,I1.z), max(I2.z,I3.z));\n"
            "O.w = max( max(I0.w,I1.w), max(I2.w,I3.w));\n"
            );
      break;
   }
   m_Reduce2x2 = reducer;
   m_NumImages = 0;
	m_MinBlockSize = 4;

}

GLImageSummarizer::~GLImageSummarizer()
{
   delete m_Reduce2x2;
   for( card32 i=0;i<m_Blocks.size();i++ )
      delete m_Blocks[i].blockImage;
}

card32 GLImageSummarizer::addImage( GLImage * img, GLRenderTextureCache * cache )
{
   initBlocks(cache);

   card32 firstBlockElementSize = m_Blocks.back().elementSize;

   GLImage * actImage = img;
   GLImage * newImage = NULL;

   // reduce image until it's next reduction fits in an block ( = elementSize*2 )
   while( actImage->getWidth() > firstBlockElementSize*2 )
   {
      m_Reduce2x2->setInputImage(0, actImage );
      newImage = m_Reduce2x2->execute( cache );
      if( actImage != img )   // don't delete input image !
         delete actImage;
      actImage = newImage;
   }

   TResult tmp;
   tmp.index = m_NumImages;
   tmp.x = m_Blocks.back().nextX;
   tmp.y = m_Blocks.back().nextY;
   m_Blocks.back().images.push_back(tmp);

   renderToBlock( cache, (card32)m_Blocks.size()-1, actImage );

   if( actImage != img )
      delete actImage;

   m_NumImages++;
   return tmp.index;
}

void GLImageSummarizer::execute( GLRenderTextureCache * cache )
{
   for( card32 i=0;i<m_Blocks.size();i++ )
   {
      size_t b = m_Blocks.size()-1-i;
      if( m_Blocks[b].images.size() )
      {
          processBlock( cache, (card32)b );
      }
   }

   m_NumImages = 0;
   m_Results = m_TemporalResults;
   m_TemporalResults.clear();
}

void GLImageSummarizer::initBlocks(GLRenderTextureCache * cache)
{
   if( m_Blocks.size() > 0 )
      return;

   card32 blockSize = 512;
   card32 minElementSize = m_MinBlockSize;
   card32 numBlocks = 1;
   card32 tmpElementSize = minElementSize*2;
   while( blockSize > tmpElementSize )
   {
      tmpElementSize*=2;
      numBlocks++;
   }

   for( card32 i=0;i<numBlocks;i++ )
   {
      TBlock block;
      block.size = blockSize;
      block.elementSize = minElementSize << i;
      block.blockImage = new GLImage;
      block.blockImage->setup( cache, block.size, block.size, 4, true, true );
      block.nextX = block.nextY = 0;
      block.blockInc = blockSize/2;
      m_Blocks.push_back( block );
   }
}

void GLImageSummarizer::renderToBlock( GLRenderTextureCache * cache, card32 blockIndex, GLImage * img )
{
   x4Assert( blockIndex < m_Blocks.size() );
   TBlock * block = &m_Blocks[blockIndex];

   m_Reduce2x2->setInputImage(0,img);
   m_Reduce2x2->execute( cache, block->blockImage, block->nextX, block->nextY );

   block->nextX += block->blockInc;
   if( block->nextX >= block->size )
   {
      block->nextX = 0;
      block->nextY += block->blockInc;
      if( block->nextY >= block->size )
      {
         block->nextY = 0;
         processBlock( cache, blockIndex );
      }
   }
}

void GLImageSummarizer::processBlock( GLRenderTextureCache * cache, card32 blockIndex )
{
   TBlock * block = &m_Blocks[ blockIndex ];

   if( blockIndex > 0 )
   {
      card32 blockXPos = m_Blocks[ blockIndex-1 ].nextX;
      card32 blockYPos = m_Blocks[ blockIndex-1 ].nextY;
      for( card32 i=0;i<block->images.size();i++ )
      {
         TResult tmpRes = block->images[i];
         tmpRes.x = blockXPos + tmpRes.x/2;
         tmpRes.y = blockYPos + tmpRes.y/2;
         m_Blocks[ blockIndex-1 ].images.push_back( tmpRes );
      }

      block->images.clear();
      renderToBlock( cache, blockIndex-1, block->blockImage );


   }
   else
   {
      perfomReadback();
      block->images.clear();
   }
}


void GLImageSummarizer::perfomReadback()
{
   x4Assert( m_Blocks.size() > 0 );

   TBlock * block = &m_Blocks[0];
   if( block->images.size() == 0 )
      return;

   std::vector<Vector4f> result( block->blockImage->getWidth() * block->blockImage->getHeight() );
   block->blockImage->bind(0);
   glGetTexImage( block->blockImage->getRenderTexture()->getTextureTarget(),0, GL_RGBA, GL_FLOAT,
      &result[0] );
   block->blockImage->unbind();

   for( card32 i=0;i<block->images.size();i++ )
   {
      TResult tmp = block->images[i];
      tmp.result = makeVector4f(0,0,0,0);
		tmp.resultBlocks.resize(block->elementSize*block->elementSize);
      for( card32 x=0;x<block->elementSize;x++ )
      {
         for( card32 y=0;y<block->elementSize;y++ )
         {
				tmp.resultBlocks[y*block->elementSize +x] = result[(tmp.y+y) * block->size + tmp.x+x];
            tmp.result += result[(tmp.y+y) * block->size + tmp.x+x];
         }
      }

      m_TemporalResults.push_back( tmp );
   }
}
