//---------------------------------------------------------------------------
#ifndef GLImageSummarizerH
#define GLImageSummarizerH
//---------------------------------------------------------------------------
#include "x3foundation_private.h"
//---------------------------------------------------------------------------
#include <vector>
#include <map>
#include "LinearAlgebra.h"
#include "X4Assert.h"
//---------------------------------------------------------------------------

namespace X4
{

class GLImage;
class GLRenderTextureCache;
class GLImageOperator;

/**
   summarizes one ore more images into 1x1 pixel
   sum modes are
    - Minimum
    - Maximum
    - Average
    - Sum

   Usage:
    GLImageSummarizer img_sum( mode );

    for( some images )
    {
       img_sum.addImage( image[i] )
       delete image[i]; // you don't have to delete the image, but you can
    }

    img_sum.execute( renderTextureCache );

    for( ... )
      Vector4f result = img_sum.getResult(i);
      ...


*/
class X3FOUNDATION_PRIVATE_API GLImageSummarizer
{
private:

   void initBlocks(GLRenderTextureCache * cache);
   void renderToBlock( GLRenderTextureCache * cache, card32 blockIndex, GLImage * img );
   void processBlock( GLRenderTextureCache * cache, card32 blockIndex );
   void perfomReadback();

   struct TResult
   {
      card32 index;  // index of input image
      card32 x,y;    // block coordinate
      Vector4f result;
		std::vector<Vector4f> resultBlocks;
   };

   std::vector<TResult>      m_Results;
   std::vector<TResult>      m_TemporalResults;   // will be copied to Results in execute()
   card32               m_NumImages;
	card32					m_MinBlockSize;


   struct TBlock
   {
      std::vector<TResult>   images;
      card32            size;          // blocksize
      card32            elementSize;
      card32            nextX,nextY;
      card32            blockInc;
      GLImage     *     blockImage;
   };

   std::vector<TBlock>       m_Blocks;

   GLImageOperator  *   m_Reduce2x2;


public:

   enum EMode
   {
      Mode_Minimum,
      Mode_Maximum,
      Mode_Average,
      Mode_Sum
   };

   GLImageSummarizer( EMode mode );
   ~GLImageSummarizer();

   /// enqueue image, you may delete img after calling addImage
   /// returns queue-index
   card32 addImage( GLImage * img, GLRenderTextureCache * cache );

   /// summarize all images in queue and clears existing result
   void execute( GLRenderTextureCache * cache );

   mpcard getNumResults() {return m_Results.size();}
   Vector4f getResult( card32 i ) {x4Assert( i < m_Results.size() );return m_Results[i].result;}
	std::vector<Vector4f> & getResultBlock( card32 i ){x4Assert( i < m_Results.size() );return m_Results[i].resultBlocks;}

   EMode getMode() const {return m_Mode;}

	card32 getMinBlockSize() const {return m_MinBlockSize;}
	/// has to be power of two!
	void setMinBlockSize(card32 v) {m_MinBlockSize=v;}

private:

   EMode    m_Mode;

};

}


#endif
