//---------------------------------------------------------------------------
#include "stdafx.h"
 
//---------------------------------------------------------------------------
#include "GLImageWarper.h"
//---------------------------------------------------------------------------
#include "StringClassProperty.h"
#include "ClassMethods.h"
#include "GLShaderMaterial.h"
#include "Configuration.h"
#include "StringHelper.h"
#include "GLRenderTextureCache.h"
#include "GLImage.h"
#include "LinearAlgebra.h"
#include "LinearAlgebra.inline.h"
#include "GLHelper.h"
#include "GLTexture.h"
#include "ApplicationDirectories.h"

//---------------------------------------------------------------------------
 
//---------------------------------------------------------------------------

using namespace X4;

MetaClass* GLImageWarper::GLImageWarperMetaClass = NULL;

const MetaClass* GLImageWarper::getInstanceClass() const {
   return getClass();
}

const MetaClass* GLImageWarper::getClass() {
   if (GLImageWarperMetaClass==NULL) {
      GLImageWarperMetaClass = new MetaClass("GLImageWarper", "no description.",
                                  NULL, sizeof(GLImageWarper));
   }
   return GLImageWarperMetaClass;
}

void GLImageWarper::init(const MetaClass *parentClass) {
   ((MetaClass*)getClass())->init(parentClass, MetaClass::FIRST_VERSION + 0);
   GLImageWarperMetaClass->copyParentProperties();
	BEGIN_CLASS_INIT(GLImageWarper)
	ADD_STRING_PROP(m_WarpOp,0)
	ADD_NOARGS_METHOD( updateWarpTexture );
 
}

void GLImageWarper::shutdown() {
   delete GLImageWarperMetaClass;
}


// ----


GLImageWarper::GLImageWarper()
{
   m_WarpOp = "float2 warp = texRECT( texture0, IN.tex0 );\n O = texRECT(texture1, warp );";

   setNumInputImages(1);
   setOutputFormat( OF_Undefined );

   loadTemplates();
}

GLImageWarper::~GLImageWarper()
{
   clearWarpTextures();
}

void GLImageWarper::assign(const Object* obj, X4CopyContext *context)
{
   GLImageMaterialOperator::assign( obj, context );
   const GLImageWarper * iw = dynamic_cast<const GLImageWarper*>(obj);
   x4Assert( iw );

   m_WarpOp = iw->m_WarpOp;
   clearWarpTextures();
}

GLImage * GLImageWarper::execute( GLRenderTextureCache * cache )
{
   x4Assert( m_InputImages.size() >= 1 );

   // get output rendertexture from cache
   GLImage * ret = new GLImage;
   bool isFloat = false;
   bool isFloat32 = false;
   switch( getOutputFormat() )
   {
   case OF_Undefined:
      // take inputformat
      isFloat = m_InputImages[0]->isFloat();
      isFloat32 = m_InputImages[0]->isFloat32();
   break;
   case OF_Float32:
      isFloat32 = true;
   case OF_Float16:
      isFloat = true;
   break;
   }

   card32 width = m_InputImages[0]->getWidth();
   card32 height = m_InputImages[0]->getHeight();
   ret->setup( cache, width , height,
      m_InputImages[0]->getNumComponents(), m_InputImages[0]->isRectangle(),
         isFloat, isFloat32 );


   // #####################################
   // --------- render to texture --------------
   ret->activate();
   glPushAttrib( GL_ALL_ATTRIB_BITS );

   GLHelper::begin2D( width, height );
   bindMaterial( cache->getContext() );


   x4Assert( width < (1 << 16) );
   x4Assert( height < (1 << 16) );
   card32 textureID = width * (1 << 16) + height;

   if( m_WarpTextures.find(textureID) == m_WarpTextures.end() )
   {
      m_WarpTextures[textureID] = buildWarpTexture( cache->getContext() );
   }
   m_WarpTextures[textureID]->bind(0);

   for( card32 i=0;i<m_InputImages.size();i++ )
   {
      m_InputImages[i]->bind(i+1);
   }

   glBegin( GL_QUADS );

   std::vector<Vector2f> coords(m_InputImages.size()*2);
   for( card32 i=0;i<m_InputImages.size();i++ )
   {
      coords[i*2+0] = makeVector2f(0,0);
      if( m_InputImages[i]->isRectangle() )
         coords[i*2+1] = makeVector2f(m_InputImages[i]->getWidth(),
                     m_InputImages[i]->getHeight());
      else
         coords[i*2+1] = makeVector2f(1,1);
   }

   for( card32 i=0;i<m_InputImages.size();i++ )
   {
      glVertexAttrib2fARB( i+1,     coords[i*2+0][0],coords[i*2+0][1] );
   }
   glVertexAttrib2fARB(0, 0,0);

   for( card32 i=0;i<m_InputImages.size();i++ )
   {
      glVertexAttrib2fARB( i+1,     coords[i*2+1][0],coords[i*2+0][1] );
   }
   glVertexAttrib2fARB(0, width, 0);

   for( card32 i=0;i<m_InputImages.size();i++ )
   {
      glVertexAttrib2fARB( i+1,     coords[i*2+1][0],coords[i*2+1][1] );
   }
   glVertexAttrib2fARB(0, width, height);

   for( card32 i=0;i<m_InputImages.size();i++ )
   {
      glVertexAttrib2fARB( i+1,     coords[i*2+0][0],coords[i*2+1][1] );
   }
   glVertexAttrib2fARB(0, 0, height);

   glEnd();


   // #####################################
   // ----------- clean up ----------------

   m_WarpTextures[textureID]->unbind();
   for( card32 i=0;i<m_InputImages.size();i++ )
   {
       m_InputImages[i]->unbind();
   }

   GLHelper::end2D();
   glPopAttrib();

   ret->deactivate();
   unbindMaterial( cache->getContext() );

   return ret;

}

void GLImageWarper::loadTemplates()
{
   m_VSTemplate = readTextFile( ApplicationDirectories::getResourcesDir() + "shaders/hqRenderer/warp_vs.txt" );
   m_PSTemplate = readTextFile( ApplicationDirectories::getResourcesDir() + "shaders/hqRenderer/warp_ps.txt" );
}

void GLImageWarper::modify(std::string &vsTemplate, std::string &psTemplate)
{
   forget(vsTemplate);
   psTemplate = replaceSubStrings( psTemplate, "$WARP_OP", m_WarpOp );
   psTemplate = replaceSubStrings( psTemplate, "$WARP_PS_PARAM", ", uniform samplerRECT texture0\n, uniform samplerRECT texture1\n"  );

}

void GLImageWarper::updateWarpTexture()
{
   clearWarpTextures();
}

GLTextureRect * GLImageWarper::buildWarpTexture( GLContext * context )
{
   x4Assert( m_InputImages.size() >= 1 );

   Vector2i dim = makeVector2i( m_InputImages[0]->getWidth(), m_InputImages[0]->getHeight() );
   static std::vector<Vector3f> buffer;
   buffer.resize( dim[0] * dim[1] );

   for( int x=0;x<dim[0];x++ )
   {
      for( int y=0;y<dim[1];y++ )
      {
         Vector2f pos = makeVector2f( x,y );
         Vector2f newPos = calculateWarpCoords( pos, dim ); 
         buffer[y * dim[0] + x] = makeVector3f(newPos[0],newPos[1],0 );
      }
   }

   GLTextureRect * ret = new GLTextureRect( context );
   ret->createFloatTextureFromMemory( dim[0], dim[1], 3, &buffer[0][0], false );
   return ret;
}


void GLImageWarper::clearWarpTextures()
{
   for( std::map<card32,GLTextureRect *>::iterator iter = m_WarpTextures.begin();
      iter != m_WarpTextures.end(); ++iter )
   {
      delete iter->second;
   }

   m_WarpTextures.clear();
}

