#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "GLOpticalFlow.h"
//---------------------------------------------------------------------------
#include "GLBuffer.h"
#include "GLRenderTextureCache.h"
#include "GLImage.h"
#include "GLShaderMaterial.h"
#include "Configuration.h"
#include "RGBAImage.h"
#include "GLImageReducer.h"
#include "StringHelper.h"
#include "GLImageSummarizer.h"
#include "ApplicationDirectories.h"

//---------------------------------------------------------------------------
using namespace X4;

GLOpticalFlow::GLOpticalFlow(GLContext * context)
{
	m_GLContext = context;
	m_ComputationBufferSize = 2048;
	m_Mat_ComputeDifference = GLShaderMaterial::createFromShaderFile(ApplicationDirectories::getResourcesDir() + "shaders/SC_opticalFlow_CompDifference.txt");
	m_ImageOp_Reduce2x2 = new GLImageReducer;
	m_ImageOp_Reduce2x2->setOutputFormat(GLImageOperator::OF_Float16);
	m_ImageOp_Reduce2x2->setNumInputImages(1);

	m_ImageOp_Summarize = new GLImageSummarizer(GLImageSummarizer::Mode_Sum);
}

GLOpticalFlow::~GLOpticalFlow(void)
{
	cleanupVertexBuffers();
	delete m_Mat_ComputeDifference;
	delete m_ImageOp_Reduce2x2;
	delete m_ImageOp_Summarize;
}

void X4::GLOpticalFlow::beginInitSamples( mpcard windowSize, mpcard searchWindow )
{
	m_WindowSize = windowSize;
	m_SearchWindow = searchWindow;
	m_Samples.clear();
}

void X4::GLOpticalFlow::addSample( const Vector2f &p )
{
	m_Samples.push_back(p);
}

void X4::GLOpticalFlow::endInitSamples()
{
	cleanupVertexBuffers();


	// setup vertex buffers
	mpcard numComparisonsPerSample = sqr(m_SearchWindow*2+1);
	mpcard numComparisons = m_Samples.size() * numComparisonsPerSample;	// total number of comparisons
	m_Result = std::vector<float>(numComparisons);	// setup result array
	mpcard windowSize_pot = 1;
	while (windowSize_pot < m_WindowSize*2+1)
	{
		windowSize_pot *= 2;
	}
	m_WindowsSize_pot = windowSize_pot;


	

	VertexDescriptor vd;
	vd.pushAttrib( VAD("position",2,VAD::DATA_FORMAT_FLOAT32));
	vd.pushAttrib( VAD("texcoord0",2,VAD::DATA_FORMAT_FLOAT32));
	vd.pushAttrib( VAD("texcoord1",2,VAD::DATA_FORMAT_FLOAT32));

	
	mpcard currentX = 0; mpcard currentY = 0;
	mpcard numCompPerRow = m_ComputationBufferSize / windowSize_pot;
	std::vector<float> vbuffer;
	for( unsigned i=0;i<numComparisons;i++ )
	{
		if( currentX == numCompPerRow )
		{
			currentX = 0;
			currentY++;
			if( currentY == numCompPerRow )
			{
				// image complete, create & store vb, start new one
				GLVertexBuffer * vb = new GLVertexBuffer( m_GLContext, vd);
				vb->init((GLuint)vbuffer.size() / vd.getSize()*4);
				vb->writeToBuffer( &vbuffer[0], 0, (GLuint)vbuffer.size() / vd.getSize()*4);
				m_SampleVertexBuffers.push_back(vb);
				
				vbuffer.clear();
				currentY = 0;
			}
		}

		// setup quad
		Vector2f s = m_Samples[i / numComparisonsPerSample];
		Vector2f p = makeVector2f(currentX,currentY)*windowSize_pot;
		mpcard compID = i % numComparisonsPerSample;
		Vector2f translation = makeVector2f(compID % (m_SearchWindow*2+1),compID / (m_SearchWindow*2+1)) - makeVector2f(1,1) * m_SearchWindow;
// 		if( i < numComparisonsPerSample )
//  		{
//  			debugOutput << "---------------\n";
//  			debugOutput << "p: " << p << "\n";
//  			debugOutput << "s: " << s << "\n";
//  			debugOutput << "compID: " << compID << "\n";
//  			debugOutput << "transl: " << translation << "\n";
//  		}
		float tmpWindowSize = m_WindowSize;
		for( unsigned j=0;j<4;j++ )
		{
			bool multX = (j == 1 || j == 2);
			bool multY = (j >= 2);
			Vector2f tex0 = s;
			tex0[0] += multX?-tmpWindowSize:tmpWindowSize;
			tex0[1] += multY?-tmpWindowSize:tmpWindowSize;
			Vector2f tex1 = tex0 + translation;
			Vector2f pos = p;
			pos[0] += multX?m_WindowSize*2+1:0;
			pos[1] += multY?m_WindowSize*2+1:0;
//			debugOutput << "tex0: " << tex0 << "   pos: " << pos << "\n";
			vbuffer.push_back(pos[0] / (float)m_ComputationBufferSize * 2.0f - 1.0f);
			vbuffer.push_back((pos[1] + 0.0f) / (float)m_ComputationBufferSize * 2.0f - 1.0f);
			vbuffer.push_back(tex0[0]);
			vbuffer.push_back(tex0[1]);
			vbuffer.push_back(tex1[0]);
			vbuffer.push_back(tex1[1]);
		}

		currentX++;
		
	}
	
	// add remaining vertices to buffer
	if( vbuffer.size() > 0)
	{
		GLVertexBuffer * vb = new GLVertexBuffer( m_GLContext, vd);
		vb->init((GLuint)vbuffer.size() / vd.getSize()*4);
		vb->writeToBuffer( &vbuffer[0], 0, (GLuint)vbuffer.size() / vd.getSize()*4);
		m_SampleVertexBuffers.push_back(vb);

		vbuffer.clear();
	}
	



/*	mpcard numCompPerImage = sqr(m_ComputationBufferSize / windowSize_pot);
	mpcard numCompImages = (numComparisons+numCompPerImage-1) / numCompPerImage;
	for( unsigned imgID=0;imgID<numCompImages;imgID++ )
	{

	}*/

}

void X4::GLOpticalFlow::cleanupVertexBuffers()
{
	for( mpcard i=0;i<m_SampleVertexBuffers.size();i++)
		delete m_SampleVertexBuffers[i];
	m_SampleVertexBuffers.clear();
}

void X4::GLOpticalFlow::compute( GLRenderTextureCache * cache, GLImage * img1, GLImage * img2 )
{
	mpcard currentResultIndex = 0;
	mpcard numComparisonsPerSample = sqr(m_SearchWindow*2+1);
	m_ImageOp_Summarize->setMinBlockSize(m_ComputationBufferSize / m_WindowsSize_pot);
	for( unsigned i=0;i<m_SampleVertexBuffers.size();i++ )
	{
		// setup rt
		GLImage * renderTarget = new GLImage;
		renderTarget->setup(cache, m_ComputationBufferSize, m_ComputationBufferSize, 4, true, true, false, false, false);
		renderTarget->activate();

		// clear & setup
		glClearColor(0,0,0,0);
		glClear( GL_COLOR_BUFFER_BIT );
		glPushAttrib( GL_ALL_ATTRIB_BITS );
		glDisable( GL_DEPTH_TEST);
		glDisable(GL_CULL_FACE);

		// setup shader
		m_Mat_ComputeDifference->bindShaders(m_GLContext);

		// setup images
		img1->bind(0);
		img2->bind(1);

		// render
		m_SampleVertexBuffers[i]->bind();
		m_SampleVertexBuffers[i]->drawArrays(GL_QUADS);
		m_SampleVertexBuffers[i]->unbind();

		// cleanup
		img1->unbind();
		img2->unbind();
		m_Mat_ComputeDifference->unbindShaders(m_GLContext);
		glPopAttrib();
		renderTarget->deactivate();

		m_ImageOp_Summarize->addImage(renderTarget,cache);

		// perform hierarchical reduction
		/*std::vector<float> result = computeSums(cache,renderTarget);
		mpcard numTotalSamples = m_Samples.size()*numComparisonsPerSample;
		mpcard numCurrentSamples = min(numTotalSamples, currentResultIndex+result.size()) - currentResultIndex;
		for( unsigned j=0;j<numCurrentSamples;j++ )
			m_Result[currentResultIndex++] = result[j];*/

		// debug output
// 		RGBAFloat32Image * rgbaimg = renderTarget->getRGBAFloat32Image();
// 		debugOutput.displayRGBAFloatImage( rgbaimg, "test");
// 		delete rgbaimg;

		delete renderTarget;
	}

	m_ImageOp_Summarize->execute(cache);
	
	for( unsigned i=0;i<m_SampleVertexBuffers.size();i++ )
	{
		std::vector<Vector4f> res = m_ImageOp_Summarize->getResultBlock(i);
		mpcard numTotalSamples = m_Samples.size()*numComparisonsPerSample;
		mpcard numCurrentSamples = min(numTotalSamples, currentResultIndex+res.size()) - currentResultIndex;
		for( unsigned j=0;j<numCurrentSamples;j++ )
			m_Result[currentResultIndex++] = res[j][0];
	}
	
	
}


std::vector<float> X4::GLOpticalFlow::computeSums( GLRenderTextureCache * cache, GLImage * img )
{
	mpcard winSize = m_WindowsSize_pot;
	GLImage * currentImage = img; 
	RGBAFloat32Image * rgba_org = currentImage->getRGBAFloat32Image();

	while (winSize>1)
	{
		winSize /= 2;
		m_ImageOp_Reduce2x2->setInputImage(0,currentImage);
		GLImage * nextImage = m_ImageOp_Reduce2x2->execute(cache);
		if( currentImage != img )
			delete currentImage;
		currentImage = nextImage;

//		RGBAFloat32Image * rgba = currentImage->getRGBAFloat32Image();
//		debugOutput.displayRGBAFloatImage(rgba,std::string("step") + intToStr(winSize));
//		delete rgba;
	}

	// read back result
	RGBAFloat32Image * rgba = currentImage->getRGBAFloat32Image();
	std::vector<float> ret(rgba->getSizeX()*rgba->getSizeY());
	for( unsigned i=0;i<ret.size();i++ )
		ret[i] = rgba->getPixels()[i][0];

	delete currentImage;
	return ret;
}

float * X4::GLOpticalFlow::getFlowImage( mpcard sampleID )
{
	mpcard numComparisonsPerSample = sqr(m_SearchWindow*2+1);
	return &m_Result[sampleID*numComparisonsPerSample];
}