#ifndef GLOpticalFlowH
#define GLOpticalFlowH
//---------------------------------------------------------------------------
#include "x3foundation_private.h"
//---------------------------------------------------------------------------

namespace X4
{

class GLContext;
class GLVertexBuffer;
class GLRenderTextureCache;
class GLImage;
class GLImageReducer;
class GLImageSummarizer;

/**
	performs optical flow for a set of sample points
	expects two rectified images as input and a set of sample
	points where the optical flow should be computed

	Usage:
		// setup flow samples
		opticalFlow->beginInitSamples(windowSize,searchWindow);
		for all samples
			opticalFlow->addSample( sample);
		opticalFlow->endInitSamples();

		// loop over several different images (or transformed images or ...)
		for something
			opticalFlow->compute(cache, img1, img2)
			

*/
class X3FOUNDATION_PRIVATE_API GLOpticalFlow
{

private:

	GLRenderTextureCache *	m_Cache;
	mpcard		m_WindowSize;		// comparison window 
	mpcard		m_SearchWindow;		// transformation space window
	mpcard		m_ComputationBufferSize;	// size of render target
	std::vector<Vector2f>		m_Samples;
	GLContext	*		m_GLContext;
	std::vector<GLVertexBuffer *>	m_SampleVertexBuffers;
	std::vector<float>		m_Result;	// image this as a stack of 2d images with size (searchWindow*2+1)^2


	GLShaderMaterial	*	m_Mat_ComputeDifference;
	GLImageReducer		*	m_ImageOp_Reduce2x2;
	GLImageSummarizer	*	m_ImageOp_Summarize;



	// temp members
	mpcard		m_WindowsSize_pot;

public:
	GLOpticalFlow(GLContext * context);
	~GLOpticalFlow(void);

	/// begin setup sample points
	/// for each sample there will be (searchWindow*2+1) ssd comparisons of
	/// image patches (windowSize*2+1)
	void beginInitSamples(mpcard windowSize, mpcard searchWindow);

	/// add sample point (pixel coordinates)
	void addSample(const Vector2f &p);

	/// init vertex buffer for samples
	void endInitSamples();

	/// compute optical flow for sample points
	void compute(GLRenderTextureCache * cache, GLImage * img1, GLImage * img2);

	/// get ssd image for a sample, size is (searchWindow*2+1)^2
	float * getFlowImage(mpcard sampleID);

	/// get number of samples
	mpcard getNumSamples() const {return m_Samples.size();}

	/// get sample
	Vector2f getSample(mpcard i) const {return m_Samples[i];}

	/// get dimension of flow image for a sample
	mpcard getFlowImageDim() const {return m_SearchWindow*2+1;}

	/// get search window size
	mpcard getSearchWindow() const {return m_SearchWindow;}

private:

	void cleanupVertexBuffers();
	std::vector<float> computeSums( GLRenderTextureCache * cache, GLImage * img );
};

}

#endif