#include "StdAfx.h"
//---------------------------------------------------------------------------
#include <windows.h>
#include <mmsystem.h>
#include "GeometricFeature.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "ObjectListProperty.h"
#include "FeaturePointDetector.h"
#include "VariableArrayClassProperty.h"
#include "GeometricTools.h"
#include "PointSetIterator.h"
#include "UnstructuredInCorePointCloud.h"
#include "ProgressWindow.h"
#include "HierarchicalKNNIterator.h"
#include "PCA.h"

#include <SeparatorClassProperty.h>
#include <PointCloudTools.h>
#include <InCorePCTopologyGraph.h>
#include <TopologyAttachment.h>
#include <PointCloudIterators.h>
#include <StringHelper.h>
#include <ObjectClassProperty.h>

#include <float.h>

//---------------------------------------------------------------------------
using namespace X4;

IMPLEMENT_X4_CLASS( GeometricFeature , 3)
{
	BEGIN_CLASS_INIT( GeometricFeature );

	ADD_VECTOR3F_PROP( m_Position,0 );
	ADD_VECTOR3F_PROP( m_Normal,0 );
	ADD_OBJECT_LIST_PROP( Signature, 0, LocalSignature::getClass());
	ADD_BOOLEAN_PROP(mBadFeature, 1);
	ADD_BOOLEAN_PROP(mDirtyBadFeatureFlag, 1);
	ADD_BOOLEAN_PROP(mIsPseudoFeature, 1);

    ADD_OBJECT_PROP(mUserData, 2, Object::getClass(), false );

	ADD_INT32_PROP(nearestPointIndex, 3);

}

GeometricFeature::GeometricFeature(void)
{
   mBadFeature = false;
   mDirtyBadFeatureFlag = true;
   mIsPseudoFeature = false;
   mUserData = NULL;
   nearestPointIndex = -1; // -1 == nearest point not set/used
}

GeometricFeature::~GeometricFeature(void)
{
	for( unsigned i=0;i<m_Signatures.size();i++ )
		delete m_Signatures[i];
}

void GeometricFeature::assign(const Object* obj, X4CopyContext *context)
{
	const GeometricFeature * o = dynamic_cast<const GeometricFeature*>(obj);
	x4Assert( o != NULL );

	copyObjectProperties(obj,this);
	mUserData = o->mUserData;
}

void X4::GeometricFeature::deleteSignatures( card32 begin, card32 end )
{
	for( unsigned i=begin;i<end;i++)
		delete m_Signatures[i];
	m_Signatures.erase(m_Signatures.begin() + begin, m_Signatures.begin() + end);
}

float X4::GeometricFeature::compare( GeometricFeature * other, FeaturePointDetector * detector, float minSignatureScoreToCount )
{
    if (badFeature(detector) || other->badFeature(detector)) return FLT_MAX;

	float score = 0;

	for( unsigned i=0;i<m_Signatures.size();i++ )
	{
		float lscore = 0;
		if (detector)
			lscore = m_Signatures[i]->compare( other->m_Signatures[i], detector->getLocalDescriptor(i) );
		else
			lscore = m_Signatures[i]->compare( other->m_Signatures[i] );

		if (lscore > minSignatureScoreToCount) score += lscore;
	}
	return score;
}

#if 0
float X4::GeometricFeature::compare( card32 signatureIndex, GeometricFeature * other, FeaturePointDetector * detector )
{
   if (badFeature(detector) || other->badFeature(detector)) return FLT_MAX;

   if (detector)
	   return m_Signatures[signatureIndex]->compare( other->m_Signatures[signatureIndex], detector->getLocalDescriptor(signatureIndex) );
   else
	   return m_Signatures[signatureIndex]->compare( other->m_Signatures[signatureIndex] );
}
#endif

bool X4::GeometricFeature::badFeature( FeaturePointDetector * detector )
{
	mBadFeature = false;
	mDirtyBadFeatureFlag = false;
	mpcard numSig = min(mpcard(m_Signatures.size()), detector ? detector->getNumLocalDescriptors() : mpcard(m_Signatures.size()));
	for( unsigned i=0;i<numSig;i++ )
	{
		LocalSignature * sig = m_Signatures[i];
		if( sig->badFeature( detector ? detector->getLocalDescriptor(i) : NULL) )
		{
		    mBadFeature = true;
			break;
		}
	}

	return mBadFeature;
}

void X4::GeometricFeature::transformFeature( const Matrix4f &mat )
{
	m_Position = transformVector3f(mat,m_Position);
	m_Normal = shrink4To3(mat)*m_Normal;

	// todo: rotate signatures
}



//---------------------------------------------------------------------------
IMPLEMENT_X4_CLASS( FeatureSet ,3)
{
	BEGIN_CLASS_INIT( FeatureSet );

	ADD_OBJECT_LIST_PROP( Feature, 0, GeometricFeature::getClass());
	ADD_FLOAT32_VAR_ARRAY_PROP( GeodeticDistance, 1);
	
	ADD_SEPARATOR("Rendering");
	ADD_BOOLEAN_PROP(m_Visible, 2);
	ADD_FLOAT32_PROP(m_FeatureSize, 2);
   ADD_OBJECT_PROP (m_RenderingTool, 3, RenderingTool::getClass(), true);

}


FeatureSet::FeatureSet() : AttachedData(), RenderableObject()
{
   m_Visible = true;
   m_FeatureSize = 0.25f;
	m_RenderingTool = NULL;
}

FeatureSet::~FeatureSet()
{
	for( unsigned i=0;i<m_Features.size();i++ )
		delete m_Features[i];	
   if (m_RenderingTool) 
      delete m_RenderingTool;
}

void X4::FeatureSet::deleteFeatures( mpcard begin, mpcard end )
{
	for( unsigned i=begin;i<end;i++)
		delete m_Features[i];
	m_Features.erase(m_Features.begin() + begin, m_Features.begin() + end);
}

void X4::FeatureSet::assign( const Object* obj, X4CopyContext *context /*= NULL*/ )
{
	AttachedData::assign(obj,context);
	const FeatureSet * o = dynamic_cast<const FeatureSet*>(obj);
	x4Assert( o != NULL );

	copyObjectProperties(obj,this);
}

void FeatureSet::copyData(FeatureSet* to)
{
   if (to == NULL) return;
   to->m_GeodeticDistances = m_GeodeticDistances;
   
   // copy each feature explicitely
   for (std::vector<GeometricFeature*>::iterator it = m_Features.begin(); it != m_Features.end(); it++)
   {
      GeometricFeature* gp = dynamic_cast<GeometricFeature*>((*it)->copy());
      to->m_Features.push_back(gp);
   }
}



void X4::FeatureSet::computeGeodeticDistances( UnstructuredInCorePointCloud * pc )
{
   InCorePCTopologyGraph *tg = NULL;
   AttachedDataContainer *attachments = pc->getAttachments();
   TopologyAttachment *topAtt = dynamic_cast<TopologyAttachment*>(attachments->getData("TopologyAtt"));
   if (topAtt) {
      tg = topAtt->getTopology();
   } else {
      throw EInvalidState("X4::FeatureSet::computeGeodeticDistances - the point cloud does not have a topology graph (it is currently not created on the fly)");
   }
   
   PointCloudIterator *top_it = tg->createIterator(pc->getPointSet());   

	std::vector<Vector3f> fpoints;
	for( unsigned i=0;i<m_Features.size();i++ )
	{
		fpoints.push_back(m_Features[i]->m_Position);
	}

	m_GeodeticDistances.resize(fpoints.size()*fpoints.size());

	progressWindow->pushStep();
	for( unsigned i=0;i<fpoints.size();i++ )
	{
		std::vector<float> distances = ::computeGeodeticDistances(pc, top_it, fpoints[i], fpoints);
		progressWindow->progress((float)i/(float)fpoints.size()*100.0f);
		for( unsigned j=0;j<fpoints.size();j++ )
		{
			m_GeodeticDistances[i*fpoints.size()+j] = distances[j];
		}
	}
	progressWindow->popSteps();
	
	delete top_it;	
}

GeometricFeature * X4::FeatureSet::getNearestFeature( const Vector3f &pos )
{
	float dist = 1e20f;
	GeometricFeature * ret = NULL;
	for( unsigned i=0;i<m_Features.size();i++ )
	{
		float tmpDist = norm(pos - m_Features[i]->m_Position);
		if( tmpDist < dist )
		{
			dist = tmpDist;
			ret = m_Features[i];
		}
	}
	return ret;
}

void X4::FeatureSet::orientateNormals( PointCloud * pc, float radius )
{
	HierarchicalKNNIterator	 hIt(pc,32,NULL);
	AAT posAAT = pc->getAAT("position");
	AAT normalAAT = pc->getAAT("normal");


	for( unsigned i=0;i<m_Features.size();i++ )
	{
		PCA<float,3> pca;
		GeometricFeature * f = m_Features[i];
		Vector3f averageNormal = NULL_VECTOR3F;
		
		hIt.setSeekPointAndReset(f->m_Position);
		while(!hIt.atEnd())
		{
			Vector3f pos = hIt.get3f(posAAT);
			if( norm( pos - f->m_Position) > radius )
				break;

			averageNormal += hIt.get3f(normalAAT);
			pca.addPoint(pos);

			hIt.next();
		}

		Vector3f centroid,eigenValues;
		Matrix3f eigenVectors;
		pca.analyze(eigenValues,eigenVectors,centroid);
		
		f->m_Normal = eigenVectors[2];
		if( f->m_Normal * averageNormal < 0)
			f->m_Normal *= -1.0f;
		
	}
}

//---------------------------------------------------------------------------
void FeatureSet::removeBadFeatures(FeaturePointDetector* detector)
{
   // if geodetic distances are valid, then do also change them
   bool haveDistancies = false;
   if (m_GeodeticDistances.size() == m_Features.size())
      haveDistancies = true;
      
   // iterate over all features and remove bad features
   std::vector<GeometricFeature*>::iterator it = m_Features.begin();
   std::vector<float>::iterator jt = m_GeodeticDistances.begin();
   for (; it != m_Features.end(); )
   {
      if ((*it)->badFeature(detector))
      {
         delete *it;
         it = m_Features.erase(it);
         if (haveDistancies) jt = m_GeodeticDistances.erase(jt);
      }else
      {
         it ++;
         if (haveDistancies) jt ++;
      }
   }
}

//---------------------------------------------------------------------------
void FeatureSet::computeBadFeatures(FeaturePointDetector* detector)
{
   // iterate over all features and check whenever thay are bad or not
	for( unsigned i=0; i < m_Features.size(); i++ )
	   m_Features[i]->badFeature(detector);      
}

//---------------------------------------------------------------------------
void FeatureSet::renderGL(X4::SceneObject *obj)
{
   if (!m_Visible) return;
   if (!m_RenderingTool)
   {
   	m_RenderingTool = new RenderingTool();
   	m_RenderingTool->init();
   }
   
	glPushAttrib(GL_ALL_ATTRIB_BITS);

   // color   
	m_RenderingTool->setRenderColor(0.9f,0.1f,0.1f, 1.0f);
	
	// get point cloud and check for the cached median point distance
	UnstructuredInCorePointCloud* upc = dynamic_cast<UnstructuredInCorePointCloud*>(obj);
	float medPointDistance = 0;
	if( upc ) 
		medPointDistance = getCachedMedianPointDistance(upc);
	else
		medPointDistance = 0.03f;
   
   // draw each feature
	for( unsigned i=0; i < m_Features.size(); i++ )
	{
      // only render if feature point is valid
		if(!m_Features[i]->getLastBadFeatureState())
			m_RenderingTool->renderSphere(m_Features[i]->m_Position, medPointDistance * m_FeatureSize);
		
	}
   
   glPopAttrib();
}
