#ifndef GeometricFeatureH
#define GeometricFeatureH
//---------------------------------------------------------------------------
#include "x3foundation_private.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
#include "AttachedData.h"
#include "LocalDescriptor.h"

#include "RenderableObject.h"
#include "RenderingTool.h"
//---------------------------------------------------------------------------

namespace X4
{
class FeaturePointDetector;
/**
	defines a geometric feature
*/
class X3FOUNDATION_PRIVATE_API GeometricFeature : public Persistent
{
	X4_CLASS( GeometricFeature )

      
public:
   bool mBadFeature;
   bool mDirtyBadFeatureFlag;
   bool mIsPseudoFeature;
   Object* mUserData;

	GeometricFeature(void);
	~GeometricFeature(void);

	Vector3f		m_Position;
	Vector3f		m_Normal;
	std::vector<LocalSignature*>	m_Signatures;
	int32 nearestPointIndex;

	float compare( GeometricFeature * other, FeaturePointDetector * detector, float minSignatureScoreToCount = 0 );
	
	/// Compare to given feature by using the specified local signature. If no comparison is possible, then a max_float value is returned.
   //float compare( card32 signatureIndex, GeometricFeature * other, FeaturePointDetector * detector );

   /// Check whenever the feature is not valid in the context of the given detector
	bool badFeature(FeaturePointDetector * detector);

   inline bool isBadFeature() const { return mBadFeature; }
   
   inline void setIsPseudoFeature(bool b) { mIsPseudoFeature = b;}
   inline bool getIsPseudoFeature() const { return mIsPseudoFeature; }
   
   /**
    * Get last valid state of the feature in the context of previously used detector.
    * The returned value might be not valid. The validty state of the returned value 
    * can be checked with the validState paramter. To ensure a valid result you have
    * to precompute the bad feature state with the badFeature() method.
    **/
   inline bool getLastBadFeatureState(bool* validState = NULL) const { if (validState) *validState = mDirtyBadFeatureFlag; return mBadFeature; }

	virtual void assign(const Object* obj, X4CopyContext *context = NULL);


	void addSignature( Object * obj ) {m_Signatures.push_back((LocalSignature*)obj);}
	void deleteSignatures( card32 begin, card32 end );
	Object * getSignature( card32 index ) {return m_Signatures[index];}
	void moveSignature( card32 oldPos, card32 newPos ) {throw "ToDo";} // not necessary
	mpcard getNumSignatures() {return m_Signatures.size();}
	std::string getSignatureName(card32 index ) {return "-";}

	void transformFeature( const Matrix4f &mat);

    Object* getUserData() { return mUserData; }
    void setUserData(Object* data) { mUserData = data; }

};

class PointCloud;

/**
	holds a set of features
*/
class X3FOUNDATION_PRIVATE_API FeatureSet : public AttachedData, public RenderableObject 
{
	X4_CLASS( FeatureSet )

private:
   bool  m_Visible;
   float32 m_FeatureSize;
   RenderingTool* m_RenderingTool;
   
public:
   FeatureSet();
	virtual ~FeatureSet();
	std::vector<GeometricFeature*>	m_Features;
	std::vector<float>					m_GeodeticDistances;
   
	virtual void assign(const Object* obj, X4CopyContext *context = NULL);

   /**
    * Copy all the data (distancies and features) to another feature set.
    * This method is required to prevent copying of parent classes data which will
    * be initiated by usign assign() method. The given feature set must be preconfigured 
    * properly (name, etc.) before calling this method.
    * NOTE: Using this method the m_name of the AttachData parent class is not copied ;)
    **/
   void copyData(FeatureSet* to);
   
   /**
    * Let the feature detector run over the computed features to recheck which
    * features are bad and which not. Call this method everytime you want to 
    * update the validity information of the feature set.
    **/
   void computeBadFeatures(FeaturePointDetector* detector);
   
   /**
    * Cleanup the feature set of bad features. This method will go over all the features containing
    * in this set and remove all the bad features. computeBadFeatures() will be called implicitely
    * within this method.
    * @param detector Detector which will be used to detect bad features.
    * NOTE: You have to recompute the features if detector or its parameters are changed
    **/
   void removeBadFeatures(FeaturePointDetector* detector);
   
	/// default for attachment name
	static std::string getDefaultName() {return "FeatureSet";}

	/// get feature next to pos (slow)
	GeometricFeature * getNearestFeature( const Vector3f &pos);

	/// computes geodetic distances (creates a fixed-knn-topology if no topology found)
	void computeGeodeticDistances(UnstructuredInCorePointCloud * pc);
	
	/// geodetic distance between two features
	float getGeodeticDistance( card32 feature1, card32 feature2 ) 
		{return m_GeodeticDistances[feature1*m_Features.size()+feature2];}

	/// flip wrong orientated normals
	void orientateNormals( PointCloud * pc, float radius );

	void addFeature( Object * obj ) {m_Features.push_back((GeometricFeature*)obj);}
	void deleteFeatures( mpcard begin, mpcard end );
	Object * getFeature( mpcard index ) {return m_Features[index];}
	void moveFeature( mpcard oldPos, mpcard newPos ) {throw "ToDo";} // not necessary}
	mpcard getNumFeatures() {return (mpcard)m_Features.size();}
	std::string getFeatureName(mpcard index ) {return "-";}

	void setNumGeodeticDistances(mpcard size) {m_GeodeticDistances.resize(size);}
	void deleteGeodeticDistance( mpcard index ) {m_GeodeticDistances.erase( m_GeodeticDistances.begin() + index);}
	char * getGeodeticDistance( mpcard index ) {return (char*)&m_GeodeticDistances[index];}
	mpcard getNumGeodeticDistances() {return m_GeodeticDistances.size();}
	std::string getGeodeticDistanceName(mpcard index ) {return "-";}
   
   // Render object interface
	virtual void renderGL(SceneObject *obj);
   
   inline bool getVisible() const { return m_Visible; }
   inline void setVisible(bool v) { m_Visible = v; }

   inline float getFeatureSize() const { return m_FeatureSize; }
   inline void setFeatureSize(float r) { m_FeatureSize = r; }
   
};

}

#endif