#include "StdAfx.h"
//---------------------------------------------------------------------------
#include <windows.h>
#include <mmsystem.h>
#include "GeometricTools.h"
//---------------------------------------------------------------------------
#include "PCA.inline.h"
#include "PointCloud.h"
#include "HierarchicalKNNIterator.h"
#include "NormalDistribution.h"
#include "Monomials2ndOrder.h"
#include "SuperIterator.inline.h"
#include "PointCloudTools.h"
#include "UnstructuredInCorePointCloud.h"
#include "DebugRenderer.h"
#include "InCorePCTopologyGraph.h"
#include "Random.h"
#include "PCCResampler.h"
#include "UnstructuredInCoreTriangleMesh.h"
#include "TopologyAttachment.h"
#include "VertexArray.h"
#include <opencv\cxcore.h>
#include <TopologicalKNNIterator.h>

//---------------------------------------------------------------------------
using namespace X4;


/// uses weighted pca to estimate a local coordinate frame. Uses at least 20 points
void X4::computeLocalFrame( const Vector3f &point, float sigma, PointCloud * pc,  Matrix3f &frame, Vector3f &center, bool orientWithNormals )
{
	HierarchicalKNNIterator hIt(pc,32,NULL);
	computeLocalFrame(point,sigma,pc,&hIt,frame,center,orientWithNormals);
}

X3FOUNDATION_PRIVATE_API void X4::computeLocalFrame( const Vector3f &point, float sigma, PointCloud * pc, HierarchicalKNNIterator * hIt, Matrix3f &frame, Vector3f &center, bool orientWithNormals )
{
	float inv_sigma_2 = 1.0f / sigma / sigma;
	float rad_2 = sigma * sigma * 4.0f;
	PCA<float,3> pca;
	AAT posAAT = pc->getAAT("position");
	AAT normalAAT;
	if( orientWithNormals )
		normalAAT = pc->getAAT("normal");

	hIt->setSeekPointAndReset(point);
	Vector3f firstNormal;
	if( orientWithNormals )
		firstNormal = hIt->get3f(normalAAT);
	card32 numPoints = 0;
	while( !hIt->atEnd())
	{
		Vector3f sample_pos = hIt->get3f(posAAT);
		float dist_2 = normQuad(sample_pos - point);
		if( dist_2 > rad_2 && numPoints >= 40 )
			break;

		numPoints++;
		pca.addPoint(sample_pos, exp(-dist_2*inv_sigma_2));

		hIt->next();
	}

	Vector3f eigenValues;
	pca.analyze(eigenValues,frame,center);
	if( orientWithNormals )
	{
		if( frame[2]*firstNormal < 0.0f )
		{
			frame[2] = -frame[2];
			frame[1] = -frame[1];
		}
	}
}

PointCloud * X4::createNoisyCopy( PointCloud * pc, float stdDeviation, float maxAmplitude, bool noiseInNormalDirection )
{
	PointCloud * pcCopy = (PointCloud*)pc->copy();
	SceneObjectIterator * iter = pcCopy->createIterator( SceneObjectIteratorType::CAP_BASIC_PC 
		| SceneObjectIteratorType::CAP_MODIFYABLE_PC);

	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(iter);

	AAT posAAT = pc->getAAT("position");
	AAT normalAAT;
	if( noiseInNormalDirection )
		normalAAT = pc->getAAT("normal");

	while( !bIt->atEnd())
	{
		Vector3f noisyPos = bIt->get3f(posAAT);
		if( noiseInNormalDirection )
			noisyPos += bIt->get3f(normalAAT) * 
				max( min( NormalDistribution::randomNumber( 0, stdDeviation ), maxAmplitude), - maxAmplitude);
		else
		{
			for( unsigned i=0;i<3;i++ )
				noisyPos[i] += max( min( NormalDistribution::randomNumber( 0, stdDeviation ), maxAmplitude), - maxAmplitude);
		}
			
		mIt->set3f(posAAT,noisyPos);
		bIt->next();
	}

	delete iter;
	pcCopy->clearAttachments();
	return pcCopy;
}

bool X4::projectPointOntoPatch( const Vector3f &point, float sigma, PointCloud * pc,  Vector3f &projPoint, Vector3f &projNormal )
{
	HierarchicalKNNIterator hIt(pc,32,NULL);
	return projectPointOntoPatch(point,sigma,pc,&hIt, projPoint, projNormal);
}

bool X4::projectPointOntoPatch( const Vector3f &point, float sigma, PointCloud * pc, HierarchicalKNNIterator * hIt,  Vector3f &projPoint, Vector3f &projNormal )
{
	//HierarchicalKNNIterator hIt(pc,32,NULL);
	float inv_sigma_2 = 1.0f / sigma / sigma;
	float rad_2 = sigma * sigma * 4.0f;
	float rad = 2.0f * sigma;
	PCA<float,3> pca;
	AAT posAAT = pc->getAAT("position");
	AAT normalAAT;
	bool useNormals = pc->providesAttribute("normal");
	if( useNormals)
		normalAAT = pc->getAAT("normal");

	// estimate coordinate frame
	hIt->setSeekPointAndReset(point);
	card32 numPoints = 0;
	std::vector<float> weights;
	std::vector<Vector3f> positions;
	Vector3f firstNormal;
	if( useNormals )
		firstNormal = hIt->get3f(normalAAT);
	mpcard numNormalsFlipped = 0;
	mpcard numNormalsCorrect = 0;
	while( !hIt->atEnd())
	{
		Vector3f sample_pos = hIt->get3f(posAAT);
		float dist_2 = normQuad(sample_pos - point);
		if( dist_2 > rad_2 && numPoints >= 20 )
			break;

		numPoints++;
		float w = exp(-dist_2*inv_sigma_2);
		pca.addPoint(sample_pos, w);
		weights.push_back(w);
		positions.push_back(sample_pos);
		if( useNormals )
		{
			bool flipped = hIt->get3f(normalAAT)*firstNormal < 0;
			if( flipped)
				numNormalsFlipped++;
			else
				numNormalsCorrect++;

		}

		hIt->next();
	}
   
	Vector3f eigenValues,centroid;
	Matrix3f eigenVectors;
	pca.analyze(eigenValues,eigenVectors,centroid);

	centroid = point - eigenVectors[2] * (eigenVectors[2] * (point - centroid));
	Matrix3f frame = eigenVectors.transpose();
	
	// transform into coordinate frame
	for( unsigned i=0;i<positions.size();i++ )
	{
		positions[i] = frame * (positions[i]-centroid) / rad;
	}

	// fit monomials
	Vector6f coeffs;
	Monomials2ndOrder::computeCoeffs( &positions[0], &weights[0], (card32)positions.size(), coeffs);

	// project point and retransform into reference frame
	projPoint = eigenVectors * makeVector3f(0,0,coeffs[0])*rad + centroid;
	projNormal = eigenVectors * Monomials2ndOrder::evaluateNormal( coeffs, makeVector2f(0,0));

	if( useNormals )
	{
		if( projNormal * firstNormal < 0)
			projNormal *= -1;
		if( numNormalsCorrect < numNormalsFlipped)
			projNormal *= -1;
	}

   return true;	
}

Vector6f X4::getPatchCoefficients( const Vector3f &point, float sigma, PointCloud * pc,  Matrix3f &frame, HierarchicalKNNIterator * hIt)
{
	float inv_sigma_2 = 1.0f / sigma / sigma;
	float rad_2 = sigma * sigma * 4.0f;
	float rad = 2.0f * sigma;
	PCA<float,3> pca;
	AAT posAAT = pc->getAAT("position");
	AAT normalAAT = pc->getAAT("normal");
	

	// estimate coordinate frame
	hIt->setSeekPointAndReset(point);
	card32 numPoints = 0;
	std::vector<float> weights;
	std::vector<Vector3f> positions;
	Vector3f firstNormal;
	firstNormal = hIt->get3f(normalAAT);
	while( !hIt->atEnd())
	{
		Vector3f sample_pos = hIt->get3f(posAAT);
		float dist_2 = normQuad(sample_pos - point);
		if( dist_2 > rad_2 && numPoints >= 20 )
			break;

		numPoints++;
		float w = exp(-dist_2*inv_sigma_2);
		pca.addPoint(sample_pos, w);
		weights.push_back(w);
		positions.push_back(sample_pos);

		hIt->next();
	}

	Vector3f eigenValues,centroid;
	Matrix3f eigenVectors;
	pca.analyze(eigenValues,eigenVectors,centroid);

	if( eigenVectors[2]*firstNormal < 0.0f )
	{
		eigenVectors[2] = -eigenVectors[2];
		eigenVectors[1] = -eigenVectors[1];
		//frame[0] = -frame[0];
	}
	frame = eigenVectors.transpose();

	// transform into coordinate frame
	//frame /= rad;

	for( unsigned i=0;i<positions.size();i++ )
	{
		positions[i] = frame * (positions[i]-point)/rad;
	}

	// fit monomials
	Vector6f coeffs;
	Monomials2ndOrder::computeCoeffs( &positions[0], &weights[0], (card32)positions.size(), coeffs);
	coeffs[0] *= rad;
	coeffs[3] /= rad;
	coeffs[4] /= rad;
	coeffs[5] /= rad;

	return coeffs;	
}

struct TDistanceIndexPair
{
	float dist;
	PointIndex index;
	bool operator < (const TDistanceIndexPair &o) const {return dist > o.dist;}
	TDistanceIndexPair(float d, PointIndex i) {dist = d; index = i;}
};


std::vector<float> X4::computeGeodeticDistances( PointCloud * pc, PointCloudIterator * topologyIter,
											 const Vector3f startPoint, const std::vector<Vector3f> points,
											 std::vector< std::vector<Vector3f> > * paths)
{
	TopologyIterator * tIter = dynamic_cast<TopologyIterator*>(topologyIter);
	x4Assert( tIter );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(topologyIter);
	RandomAccessPointCloudIterator * rIt = dynamic_cast<RandomAccessPointCloudIterator*>(topologyIter);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(topologyIter);
	AAT posAAT = pc->getAAT( "position" );

	SceneObjectIterator * iter = pc->createIterator( SceneObjectIteratorType::CAP_BASIC_PC |
		SceneObjectIteratorType::CAP_RANDOM_ACCESS_PC );
	x4Assert( iter != NULL );
	BasicPointCloudIterator * neighbour_bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	RandomAccessPointCloudIterator * neighbour_rIt = dynamic_cast<RandomAccessPointCloudIterator*>(iter);

	// -----------------------------------------
	// distance transformation for the complete point cloud
	std::vector<float> distanceField( pc->getNumPoints(), 1e30f );
	HierarchicalKNNIterator hIt(pc,32,NULL);
	hIt.setSeekPointAndReset( startPoint );
	//queue<PointIndex>        indexQueue;
	priority_queue<TDistanceIndexPair> pq;
	PointIndex startIndex;
	for( card32 i=0;i<6 && !hIt.atEnd();i++ )
	{
		startIndex = hIt.getPointSetPointNumber();
		float dist = norm( hIt.get3f( posAAT ) - startPoint);
		//distanceField[startIndex] = dist;
		//indexQueue.push( startIndex );
		pq.push(TDistanceIndexPair(dist,startIndex));
		hIt.next();
	}

	//while( !indexQueue.empty() )
	while(!pq.empty())
	{
		//PointIndex index = indexQueue.front();
		//indexQueue.pop();        // remove first element
		PointIndex index = pq.top().index;
		float currentDistance = pq.top().dist;
		pq.pop();

		if( distanceField[index] < 1e30f )
			continue;

		distanceField[index] = currentDistance;

		rIt->goToIndex( index );
		Vector3f pos = bIt->get3f(posAAT);
		card32 numNeighbors = tIter->getNumNeighbours();
		for( card32 i=0;i<numNeighbors;i++ )
		{
			PointIndex neighbour = tIter->getNeighbour(i);
			if( distanceField[neighbour] < 1e30f )
				continue;

			neighbour_rIt->goToIndex(neighbour);
			Vector3f neighbour_pos = neighbour_bIt->get3f(posAAT);

			float distance = max(1e-20f, norm( neighbour_pos - pos )) + currentDistance;
			if( distanceField[neighbour] > distance )
			{
				//distanceField[neighbour] = distance;
				//indexQueue.push( neighbour );
				pq.push(TDistanceIndexPair(distance,neighbour));
			}

		}
	}

	if( paths != NULL )
	{
		paths->clear();
	}

	// get distances from distance field
	std::vector<float> distances( points.size() );
	for( card32 i=0;i<points.size();i++ )
	{
		hIt.setSeekPointAndReset( points[i] );
		PointIndex index = hIt.getPointSetPointNumber();
		for( card32 j=0;j<6 && !hIt.atEnd();j++ )
		{
			float tmpDist = distanceField[index] + norm( points[i] - hIt.get3f(posAAT));
			if( j == 0 || tmpDist < distances[i] )
			{
				distances[i] = tmpDist;
				index = hIt.getPointSetPointNumber();
			}
			hIt.next();
		}

		if( points[i] == startPoint )
			distances[i] = 0.0f;

		if( paths != NULL )
		{
			std::vector<Vector3f> path;
			float lastDistance = 1e30f;
			while( index != startIndex )
			{
				rIt->goToIndex( index );
				path.push_back( bIt->get3f(posAAT) );
				if( path.size() > 50000 )
					throw X4Exception("computeGeodeticDistances path too long");

				float minDist = 1e30f;
				PointIndex bestNeighbour = tIter->getNeighbour(0);
				for( card32 j=0;j<tIter->getNumNeighbours();j++ )
				{
					PointIndex neighbour = tIter->getNeighbour(j);
					if( distanceField[neighbour] < minDist )
					{
						minDist = distanceField[neighbour];
						bestNeighbour = neighbour;
					}
				}
				if( minDist >= lastDistance )
				{
					break;
				}
				lastDistance = minDist;

				index = bestNeighbour;
			}
			paths->push_back( path );

		}
	}


	delete iter;
	return distances;
}

X3FOUNDATION_PRIVATE_API std::vector<float> X4::computeGeodeticDistances( PointCloud * pc, PointCloudIterator * topologyIter, float sigma,
																						  const Vector3f startPoint, const std::vector<Vector3f> points, 
																						  std::vector<Vector3f> &derivates, std::vector< std::vector<Vector3f> > * paths /*= NULL*/)
{
	// find corresponding indices to input points
	HierarchicalKNNIterator hIt(pc,32,NULL);
	hIt.setSeekPointAndReset(startPoint);
	unsigned startPointIndex = hIt.getCurrentPointIndex();
	std::vector<unsigned> indices(points.size());
	for( unsigned i=0;i<points.size();i++ )
	{
		hIt.setSeekPointAndReset(points[i]);
		indices[i] = hIt.getCurrentPointIndex();
	}
	return computeGeodeticDistances(pc,topologyIter,sigma, startPointIndex, indices, derivates, paths);
}

X3FOUNDATION_PRIVATE_API std::vector<float> X4::computeGeodeticDistances( PointCloud * pc, PointCloudIterator * topologyIter, float sigma, 
																						  const unsigned startIndex, const std::vector<unsigned> &indices, 
																						  std::vector<Vector3f> &derivates, std::vector< std::vector<Vector3f> > * paths /*= NULL*/)
{
	TopologyIterator * tIter = dynamic_cast<TopologyIterator*>(topologyIter);
	x4Assert( tIter );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(topologyIter);
	RandomAccessPointCloudIterator * rIt = dynamic_cast<RandomAccessPointCloudIterator*>(topologyIter);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(topologyIter);
	AAT posAAT = pc->getAAT( "position" );

	SceneObjectIterator * iter = pc->createIterator( SceneObjectIteratorType::CAP_BASIC_PC | 
		SceneObjectIteratorType::CAP_RANDOM_ACCESS_PC );
	x4Assert( iter != NULL );
	BasicPointCloudIterator * neighbour_bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	RandomAccessPointCloudIterator * neighbour_rIt = dynamic_cast<RandomAccessPointCloudIterator*>(iter);

	// -----------------------------------------
	// distance transformation for the complete point cloud
	std::vector<float> distanceField( pc->getNumPoints(), 1e30f );
	derivates.resize(indices.size());
	HierarchicalKNNIterator hIt(pc,32,NULL);
	queue<PointIndex>	indexQueue;
	indexQueue.push(startIndex);
	distanceField[startIndex] = 0.0f;

// 	float medPointDistance = getMedianPointDistance(pc);
// 	rIt->goToIndex(startIndex);
// 	Vector3f pos = bIt->get3f(posAAT);
// 	debugRenderer->addFastSphere(pos, medPointDistance*2.0f, makeVector3f(1,0,0));

	while( !indexQueue.empty() )
	{
		PointIndex index = indexQueue.front();
		indexQueue.pop();	// remove first element

		rIt->goToIndex( index );
		Vector3f pos = bIt->get3f(posAAT);
		float currentDistance = distanceField[index];
		card32 numNeighbors = tIter->getNumNeighbours();
		for( card32 i=0;i<numNeighbors;i++ )
		{
			PointIndex neighbour = tIter->getNeighbour(i);
			neighbour_rIt->goToIndex(neighbour);
			Vector3f neighbour_pos = neighbour_bIt->get3f(posAAT);

			float distance = max(1e-20f, norm( neighbour_pos - pos )) + currentDistance;
			if( distanceField[neighbour] > distance )
			{
				distanceField[neighbour] = distance;
				indexQueue.push( neighbour );
			}

		}
	}

	if( paths != NULL )
	{
		paths->clear();
	}

	// get distances from distance field
	std::vector<float> distances( indices.size() );
	float inv_sigma_2 = 1.0f / sigma / sigma;
	float rad_2 = pow(sigma * 2.0f,2.0f);
	for( card32 i=0;i<indices.size();i++ )
	{
		PointIndex index = indices[i];
		rIt->goToIndex(index);
		Vector3f currentPoint = bIt->get3f(posAAT);

		// no path leads to this point
		if( distanceField[index] >= 1e20f )
		{
			distances[i] = 1e20f;
			derivates[i] = NULL_VECTOR3F;
			continue;
		}
      
		// compute a tangent plane
		Vector3f center;
		Matrix3f frame;
		computeLocalFrame(currentPoint, sigma, pc, &hIt, frame, center);
		const Matrix3f worldToFrame = frame.transpose();

		// gather neighbor points and transform into tangent space
		// height is set to distance, ignores outliers with infinity distance
		hIt.setSeekPointAndReset(currentPoint);
		std::vector<Vector3f> tangentSpacePoints;
		while( !hIt.atEnd())
		{
			if( distanceField[hIt.getCurrentPointIndex()] >= 1e20f)
			{
			   hIt.next();
				continue;
		   }
			Vector3f tmpPos = hIt.get3f(posAAT);
			float dist_2 = normQuad(tmpPos-currentPoint);
			if( dist_2 > rad_2)
				break;

			tmpPos = worldToFrame * (tmpPos-currentPoint);
			tmpPos[2] = distanceField[hIt.getCurrentPointIndex()];
			tangentSpacePoints.push_back(tmpPos);

			hIt.next();
		}

		// compute first order fit
		Vector3f coeffs;
		Monomials1stOrder::computeCoeffs(tangentSpacePoints, coeffs, sigma*sigma);
		
		// set slightly more precise distance
		distances[i] = distanceField[index];//coeffs[0];

		// compute derivates
		Vector3f tangentSpaceDerivate = makeVector3f(coeffs[1],coeffs[2],0);
		derivates[i] = frame * tangentSpaceDerivate;
// 		debugRenderer->addLine(currentPoint, currentPoint + derivates[i]*medPointDistance*2,
// 			makeVector3f(0,1,0),makeVector3f(1,1,0.5f),2);

      
		if( paths != NULL )
		{
			std::vector<Vector3f> path;
			float lastDistance = 1e30f;
			while( index != startIndex )
			{
				rIt->goToIndex( index );
				path.push_back( bIt->get3f(posAAT) );
				if( path.size() > 50000 )
					break;

				if( path.size()>1 )
				{
					size_t s = path.size();
					//debugRenderer->addLine(path[s-1],path[s-2], makeVector3f(0.5f,0.5f,0.5f),makeVector3f(0.5f,0.5f,0.5f),1);
				}

				float minDist = 1e30f;
				PointIndex bestNeighbour = tIter->getNeighbour(0);
				for( card32 j=0;j<tIter->getNumNeighbours();j++ )
				{
					PointIndex neighbour = tIter->getNeighbour(j);
					if( distanceField[neighbour] < minDist )
					{
						minDist = distanceField[neighbour];
						bestNeighbour = neighbour;
					}
				}
				if( minDist > lastDistance )
				{
					break;
				}
				lastDistance = minDist;

				index = bestNeighbour;
			}
			paths->push_back( path );

		}
	}


	delete iter;
	return distances;	
}

X3FOUNDATION_PRIVATE_API std::vector<float> X4::computeGeodeticDistances( PointCloud * pc, PointCloudIterator * topologyIter, 
                                                              Vector3f startPoint, const std::vector<Vector3f> &points, TGeneralDistanceFunction distFunc, TGDFContext * context,
                                                              std::vector< std::vector<Vector3f> > * paths, float visualizeIsoLines)
{
   // find corresponding indices to input points
   HierarchicalKNNIterator hIt(pc,32,NULL);
   hIt.setSeekPointAndReset(startPoint);
   unsigned startPointIndex = hIt.getCurrentPointIndex();
   std::vector<mpcard> indices(points.size());
   for( unsigned i=0;i<points.size();i++ )
   {
      hIt.setSeekPointAndReset(points[i]);
      indices[i] = hIt.getCurrentPointIndex();
   }
   return computeGeodeticDistances(pc, topologyIter, startPointIndex, indices, distFunc, context, paths, visualizeIsoLines);
}

struct TDistanceIndexPair_Pre
{
   float dist;
   mpcard index;
   mpcard predecessor;
   bool operator < (const TDistanceIndexPair_Pre &o) const {return dist > o.dist;}
   TDistanceIndexPair_Pre(float d, mpcard i,mpcard pre) {dist = d; index = i; predecessor = pre;}
};


X3FOUNDATION_PRIVATE_API std::vector<float> X4::computeGeodeticDistances( PointCloud * pc, PointCloudIterator * topologyIter, mpcard startIndex, const std::vector<mpcard> &points,
																	TGeneralDistanceFunction distFunc, TGDFContext * context, std::vector< std::vector<Vector3f> > * paths /*= NULL*/,
																	float visualizeIsoLines)
{
	TopologyIterator * tIter = dynamic_cast<TopologyIterator*>(topologyIter);
	x4Assert( tIter );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(topologyIter);
	RandomAccessPointCloudIterator * rIt = dynamic_cast<RandomAccessPointCloudIterator*>(topologyIter);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(topologyIter);
	AAT posAAT = pc->getAAT( "position" );
	AAT normalAAT = NULL_AAT;
	if( pc->providesAttribute("normal"))
		normalAAT = pc->getAAT("normal");

	SceneObjectIterator * iter = pc->createIterator( SceneObjectIteratorType::CAP_BASIC_PC |
		SceneObjectIteratorType::CAP_RANDOM_ACCESS_PC );
	x4Assert( iter != NULL );
	BasicPointCloudIterator * neighbour_bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	RandomAccessPointCloudIterator * neighbour_rIt = dynamic_cast<RandomAccessPointCloudIterator*>(iter);

	
	// -----------------------------------------
	// distance transformation for the complete point cloud
	std::vector<float> distanceField( pc->getNumPoints(), 1e30f );
	const int PRED_Unassigned = pc->getNumPoints();
	const int PRED_UnassignedAndTarget = PRED_Unassigned+1;
	std::vector<int> predecessor( pc->getNumPoints(), PRED_Unassigned );
	priority_queue<TDistanceIndexPair_Pre> pq;

	// flag target points
	for( unsigned i=0;i<points.size();i++ )
	{
		predecessor[points[i]] = PRED_UnassignedAndTarget;
	}
	mpcard numTargetPoints = points.size();
	pq.push(TDistanceIndexPair_Pre(0,startIndex,startIndex));

	while(!pq.empty())
	{
		mpcard index = pq.top().index;
		float currentDistance = pq.top().dist;
		mpcard pre = pq.top().predecessor;
		pq.pop();

		if( predecessor[index] < PRED_Unassigned)
			continue;

		distanceField[index] = currentDistance;
		if( predecessor[index] == PRED_UnassignedAndTarget )
		{
			numTargetPoints--;
			predecessor[index] = pre;
			if( numTargetPoints == 0)
				break;
		}
		else
			predecessor[index] = pre;

		if(0)
		{
			rIt->goToIndex( pre );
			Vector3f pre_pos = bIt->get3f(posAAT);
			rIt->goToIndex( index );
			Vector3f pos = bIt->get3f(posAAT);
			debugRenderer->addLine(pre_pos,pos);
		}


		if( visualizeIsoLines > 0.0f )
		{
			rIt->goToIndex( pre );
			Vector3f pre_pos = bIt->get3f(posAAT);
			rIt->goToIndex( index );
			Vector3f pos = bIt->get3f(posAAT);

			Vector3f dir = pos - pre_pos;
			rIt->goToIndex( pre );
			Vector3f pre_n = bIt->get3f(normalAAT);
			rIt->goToIndex( index );
			Vector3f n = bIt->get3f(normalAAT);
			float l = norm(dir);
			Matrix3f mat = calcTangentSystem(normalize(dir));
			float tmp_dist0 = distanceField[pre] / visualizeIsoLines;
			float tmp_dist1 = currentDistance / visualizeIsoLines;
			//debugRenderer->addLine(pre_pos,pos,makeVector3f(1,1,1)*0.5f,makeVector3f(1,1,1)*0.5f,1);
			for( int d=ceil(tmp_dist0);d<tmp_dist1;d++ )
			{
				float tmp = ((float)d  - tmp_dist0) / (tmp_dist1-tmp_dist0);
				Vector3f normal = normalize(pre_n * (1.0f - tmp) + n * tmp);
				Vector3f tmpPos = pre_pos + dir * tmp;
				Vector3f ldir = normalize(normal.crossProduct(dir))*l;				
				debugRenderer->addLine(tmpPos + ldir*0.35f,tmpPos - ldir*0.35f);
			}

		}

		rIt->goToIndex( index );
		card32 numNeighbors = tIter->getNumNeighbours();
		for( card32 i=0;i<numNeighbors;i++ )
		{
			mpcard neighbour = tIter->getNeighbour(i);
			if( distanceField[neighbour] < 1e30f )
				continue;

			neighbour_rIt->goToIndex(neighbour);

			float distance = (*distFunc)(pre,index,neighbour,context) + currentDistance;
			if( distanceField[neighbour] > distance )
			{
				pq.push(TDistanceIndexPair_Pre(distance,neighbour,index));
			
			}

		}
	}

	if( paths != NULL )
	{
		paths->clear();
	}

	// get distances from distance field
	std::vector<float> distances( points.size() );
	for( card32 i=0;i<points.size();i++ )
	{
		PointIndex index = points[i];
		if( startIndex == index )
			distances[i] = 0.0f;
		else
			distances[i] = distanceField[index];

		if( paths != NULL )
		{
			std::vector<Vector3f> path;
			float lastDistance = 1e30f;
			while( true )
			{
				rIt->goToIndex( index );
				path.push_back( bIt->get3f(posAAT) );
				if( path.size() > 50000 )
					throw X4Exception("computeGeodeticDistances path too long");
				if( index == startIndex )
					break;
				index = predecessor[index];
				if ( index >= (PointIndex)PRED_Unassigned )
					break;
			}
			paths->push_back( path );
		}

	}


	delete iter;
	return distances;
}

struct TComputeGeodeticDistance_PointIndex
{
	PointIndex m_Index;
	unsigned m_NumHops;
	TComputeGeodeticDistance_PointIndex(unsigned index, unsigned numHops) {m_Index =index; m_NumHops = numHops;}
};

X3FOUNDATION_PRIVATE_API std::vector<float> X4::computeGeodeticDistancesAndPathes( PointCloud * pc, PointCloudIterator * topologyIter, const Vector3f &startPoint, const std::vector<Vector3f> points, std::vector< std::vector<int32> > * paths /*= NULL*/ )
{
	
	//setup iterators
	TopologyIterator * tIter = dynamic_cast<TopologyIterator*>(topologyIter);
	x4Assert( tIter );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(topologyIter);
	RandomAccessPointCloudIterator * rIt = dynamic_cast<RandomAccessPointCloudIterator*>(topologyIter);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(topologyIter);
	AAT posAAT = pc->getAAT( "position" );

	SceneObjectIterator * iter = pc->createIterator( SceneObjectIteratorType::CAP_BASIC_PC | 
		SceneObjectIteratorType::CAP_RANDOM_ACCESS_PC );
	x4Assert( iter != NULL );
	BasicPointCloudIterator * neighbour_bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	RandomAccessPointCloudIterator * neighbour_rIt = dynamic_cast<RandomAccessPointCloudIterator*>(iter);




	// -----------------------------------------
	// do distance transformation for the complete point cloud  (TODO: only for target point range)
	
	//initialize distance field
	std::vector<float> distanceField( pc->getNumPoints(), 1e30f );
	
	//put the 5 nearest neighbors of the start point on queue
	HierarchicalKNNIterator hIt(pc,32,NULL);
	hIt.setSeekPointAndReset( startPoint );
	queue<PointIndex>	indexQueue;
	PointIndex startIndex;
	for( card32 i=0;i<5 && !hIt.atEnd();i++ )
	{
		startIndex = hIt.getPointSetPointNumber();
		distanceField[startIndex] = norm( hIt.get3f( posAAT ) - startPoint);
		indexQueue.push( startIndex );
		hIt.next();
	}

	//grow over the pc
	while( !indexQueue.empty() )
	{
		PointIndex index = indexQueue.front();
		indexQueue.pop();	

		rIt->goToIndex( index );
		Vector3f pos = bIt->get3f(posAAT);
		float currentDistance = distanceField[index];
		card32 numNeighbors = tIter->getNumNeighbours();
		for( card32 i=0;i<numNeighbors;i++ )
		{
			PointIndex neighbour = tIter->getNeighbour(i);
			neighbour_rIt->goToIndex(neighbour);
			Vector3f neighbour_pos = neighbour_bIt->get3f(posAAT);

			float distance = max(1e-20f, norm( neighbour_pos - pos )) + currentDistance;
			if( distanceField[neighbour] > distance )
			{
				distanceField[neighbour] = distance;
				indexQueue.push( neighbour );
			}
		}
	}




//======================================================================================================
	//use distance field to read distances and calculate paths

	if (paths != NULL)
		paths->clear();
	std::vector<float> distances( points.size() );

	// for all query points
	for( card32 i=0;i<points.size();i++ )
	{
		//find best start point from the 5 nearest neighbors
		hIt.setSeekPointAndReset( points[i] );
		PointIndex index = hIt.getPointSetPointNumber();
		for( card32 j=0;j<5 && !hIt.atEnd();j++ )
		{
			float tmpDist = distanceField[index] + norm( points[i] - hIt.get3f(posAAT));
			if( j == 0 || tmpDist < distances[i] )
			{
				distances[i] = tmpDist;
				index = hIt.getPointSetPointNumber();
			}
			hIt.next();
		}

		if( points[i] == startPoint )
			distances[i] = 0.0f;


		//calculate paths

		std::vector<int32> path;
		float lastDistance = 1e30f;
		while( index != startIndex )
		{
			rIt->goToIndex( index );
			path.push_back( index );
			if( path.size() > 100000 )
				throw X4Exception("computeGeodeticDistances path too long");

			float minDist = 1e30f;
			PointIndex bestNeighbour = tIter->getNeighbour(0);
			for( card32 j=0;j<tIter->getNumNeighbours();j++ )
			{
				PointIndex neighbour = tIter->getNeighbour(j);
				if( distanceField[neighbour] < minDist )
				{
					minDist = distanceField[neighbour];
					bestNeighbour = neighbour;
				}
			}
			if( minDist >= lastDistance )
			{
				break;
			}
			lastDistance = minDist;

			index = bestNeighbour;
		}
		//path.push_back( startIndex ); //xxx schould be added

		if (paths != NULL)
			paths->push_back( path );
	}

	delete iter;
	return distances;
}


X3FOUNDATION_PRIVATE_API void X4::smoothNormals( PointCloud * sourcePC, PointCloud * destPC, std::string destChannel, float sigma )
{
	checkAttribute(destPC, "normal",3,VAD::DATA_FORMAT_FLOAT32);
	AAT sourcePosAAT = sourcePC->getAAT("position");
	AAT destPosAAT = destPC->getAAT("position");
	AAT sourceNormalAAT = sourcePC->getAAT("normal");
	AAT destNormalAAT = destPC->getAAT("normal");

	SceneObjectIterator * iter = destPC->createIterator( SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC);
	SuperIterator sIt(iter,destPC->getDescr());
	HierarchicalKNNIterator hIt(sourcePC,32,NULL);

	float invSigma_2 = 1.0f / pow(sigma,2.0f);
	float maxRad_2 = 4.0f * sigma * sigma;
	
	while( !sIt.atEnd())
	{
		Vector3f pos = sIt.get<Vector3f>(destPosAAT);
		hIt.setSeekPointAndReset(pos);
		Vector3f firstNormal = hIt.get3f(sourceNormalAAT);
		Vector3f averageNormal = NULL_VECTOR3F;
		float summedWeight = 0;
		while( !hIt.atEnd())
		{
			// check bounds
			Vector3f sourcePos = hIt.get3f(sourcePosAAT);
			float dist_2 = normQuad(sourcePos - pos);
			if( dist_2 > maxRad_2 )
				break;

			// compute weight and average normal
			float weight = exp(-dist_2 * invSigma_2);
			summedWeight += weight;
			Vector3f normal = hIt.get3f(sourceNormalAAT);
			if( normal * firstNormal < 0)
				normal = -normal;
			averageNormal += normal * weight;
			
			hIt.next();
		}

		sIt.set(destNormalAAT, normalize(averageNormal));

		sIt.flush();
		sIt.next();
	}
}

X3FOUNDATION_PRIVATE_API void X4::computeCurvature_PatchFitting( PointCloud * pc, const Vector3f &point, float sigma, float &meanCurvature, float &gaussianCurvature )
{
	HierarchicalKNNIterator hIt(pc,32,NULL);
	computeCurvature_PatchFitting(&hIt, pc, point, sigma, meanCurvature, gaussianCurvature);
}

X3FOUNDATION_PRIVATE_API void X4::computeCurvature_PatchFitting( PointCloud * pc, const Vector3f &point, float sigma, 
																					 float &meanCurvature, float &gaussianCurvature,
																					 Vector3f &principalDir1, Vector3f &principalDir2)
{
	HierarchicalKNNIterator hIt(pc,32,NULL);
	float inv_sigma_2 = 1.0f / sigma / sigma;
	float rad_2 = sigma * sigma * 4.0f;
	float rad = 2.0f * sigma;
	PCA<float,3> pca;
	AAT posAAT = pc->getAAT("position");
	AAT normalAAT;
	bool useNormals = pc->providesAttribute("normal");
	if( useNormals)
		normalAAT = pc->getAAT("normal");

	// estimate coordinate frame
	hIt.setSeekPointAndReset(point);
	card32 numPoints = 0;
	std::vector<float> weights;
	std::vector<Vector3f> positions;
	Vector3f firstNormal;
	if( useNormals )
		firstNormal = hIt.get3f(normalAAT);
	while( !hIt.atEnd())
	{
		Vector3f sample_pos = hIt.get3f(posAAT);
		float dist_2 = normQuad(sample_pos - point);
		if( dist_2 > rad_2 && numPoints >= 20 )
			break;

		numPoints++;
		float w = exp(-dist_2*inv_sigma_2);
		pca.addPoint(sample_pos, w);
		weights.push_back(w);
		positions.push_back(sample_pos);

		hIt.next();
	}

	Vector3f eigenValues,centroid;
	Matrix3f eigenVectors;
	pca.analyze(eigenValues,eigenVectors,centroid);
	if( useNormals )
	{
		if( eigenVectors[2] * firstNormal < 0.0f )
			eigenVectors *= -1.0f;
	}

	Matrix3f worldToParam = eigenVectors.transpose();
	const Matrix3f paramToWorld = eigenVectors;


	// transform into coordinate frame
	for( unsigned i=0;i<positions.size();i++ )
	{
		positions[i] = worldToParam * (positions[i]-centroid) / rad;
	}

	// fit monomials
	Vector6f coeff;
	Monomials2ndOrder::computeCoeffs( &positions[0], &weights[0], (card32)positions.size(), coeff);

//	debugRenderer->addPatch(point, paramToWorld, coeff, 8,8, makeVector3f(1,0,0), 1.0f, rad);

	float f_x = coeff[1];
	float f_y = coeff[2];
	float f_xy = coeff[4];
	float f_xx = 2.0f * coeff[3];
	float f_yy = 2.0f * coeff[5];

	gaussianCurvature = (f_xx*f_yy - pow( f_xy, 2.0f ))
		/ pow(1.0f + pow(f_x,2.0f) + pow(f_y,2.0f),2.0f );

	meanCurvature = ((1.0f + pow(f_x,2.0f))*f_yy - 2.0f * f_x * f_y * f_xy + (1.0f + pow(f_y,2.0f))*f_xx)
		/ pow(1.0f + pow(f_x,2.0f) + pow(f_y,2.0f),2.0f/3.0f ) / 2.0f;

	// compute principal frame from patch
	Matrix2f M = makeMatrix2f(f_xx, f_xy, f_xy, f_yy);
	Matrix2f eigenVectors2f;
	Vector2f eigenValues2f;
	M.computeEigenStructure(eigenValues2f,eigenVectors2f);
	principalDir1 = paramToWorld[0] * eigenVectors2f[1][0]
	+ paramToWorld[1] * eigenVectors2f[1][1];
	principalDir2 = principalDir1.crossProduct(paramToWorld[2]);

}

X3FOUNDATION_PRIVATE_API void X4::computeCurvature_PatchFittingBoxed( HierarchicalKNNIterator * hIt, PointCloud * pc, const Vector3f &point, float sigma, 
																					 float &meanCurvature, float &gaussianCurvature,
																					 Vector3f &principalDir1, Vector3f &principalDir2)
{
	float inv_sigma_2 = 1.0f / sigma / sigma;
	float rad_2 = sigma * sigma * 4.0f * 4.0f;
	float rad = 2.0f * sigma * sqrt(3.0f);
	PCA<float,3> pca;
	AAT posAAT = pc->getAAT("position");
	AAT normalAAT;
	bool useNormals = pc->providesAttribute("normal");
	if( useNormals)
		normalAAT = pc->getAAT("normal");

	// estimate coordinate frame
	hIt->setSeekPointAndReset(point);
	card32 numPoints = 0;
	std::vector<float> weights;
	std::vector<Vector3f> positions;
	Vector3f firstNormal;
	if( useNormals )
		firstNormal = hIt->get3f(normalAAT);
	while( !hIt->atEnd())
	{
		Vector3f sample_pos = hIt->get3f(posAAT);
		float dist_2 = normQuad(sample_pos - point);
		if( dist_2 > rad_2 && numPoints >= 20 )
			break;

		numPoints++;
		float w = exp(-dist_2*inv_sigma_2);
		pca.addPoint(sample_pos, w);
		weights.push_back(w);
		positions.push_back(sample_pos);

		hIt->next();
	}

	Vector3f eigenValues,centroid;
	Matrix3f eigenVectors;
	pca.analyze(eigenValues,eigenVectors,centroid);
	if( useNormals )
	{
		if( eigenVectors[2] * firstNormal < 0.0f )
			eigenVectors *= -1.0f;
	}

	Matrix3f worldToParam = eigenVectors.transpose();
	const Matrix3f paramToWorld = eigenVectors;

	//centroid = point + eigenVectors[2] * (eigenVectors[2]*(centroid - point));


	// transform into coordinate frame
	for( unsigned i=0;i<positions.size();i++ )
	{
		positions[i] = worldToParam * (positions[i]-point) / rad;
	}

	// fit monomials
	Vector6f coeff;
	Monomials2ndOrder::computeCoeffs( &positions[0], &weights[0], (card32)positions.size(), coeff);

	//	debugRenderer->addPatch(point, paramToWorld, coeff, 8,8, makeVector3f(1,0,0), 1.0f, rad);

	float f_x = coeff[1];
	float f_y = coeff[2];
	float f_xy = coeff[4];
	float f_xx = 2.0f * coeff[3];
	float f_yy = 2.0f * coeff[5];

	gaussianCurvature = (f_xx*f_yy - pow( f_xy, 2.0f ))
		/ pow(1.0f + pow(f_x,2.0f) + pow(f_y,2.0f),2.0f );

	meanCurvature = ((1.0f + pow(f_x,2.0f))*f_yy - 2.0f * f_x * f_y * f_xy + (1.0f + pow(f_y,2.0f))*f_xx)
		/ pow(1.0f + pow(f_x,2.0f) + pow(f_y,2.0f),2.0f/3.0f ) / 2.0f;

	// compute principal frame from patch
	Matrix2f M = makeMatrix2f(f_xx, f_xy, f_xy, f_yy);
	Matrix2f eigenVectors2f;
	Vector2f eigenValues2f;
	M.computeEigenStructure(eigenValues2f,eigenVectors2f);
	principalDir1 = paramToWorld[0] * eigenVectors2f[1][0]
	+ paramToWorld[1] * eigenVectors2f[1][1];
	principalDir2 = principalDir1.crossProduct(paramToWorld[2]);

	// recompute weights (boxed window function)
	float boxSize = 1.0f / sqrt(3.0f);
	float b1 = boxSize * 0.85f;
	float b2 = boxSize - b1;
	Matrix2f localToPFrame = eigenVectors2f.transpose();
	for( unsigned i=0;i<positions.size();i++ )
	{
		Vector2f pframe_point = localToPFrame * makeVector2f(positions[i][0],positions[i][1]);
		float w1 = 1.0f - clamp<float>((fabs(pframe_point[0]) - b1) / b2,0.0f,1.0f);
		float w2 = 1.0f - clamp<float>((fabs(pframe_point[1]) - b1) / b2,0.0f,1.0f);
		weights[i] = min(w1,w2);
		
		//if( weights[i] > 0 )
		//	debugRenderer->addPoint(paramToWorld*positions[i]*rad + point, makeVector3f(1,weights[i],0));

		//positions[i][0] = pframe_point[0];
		//positions[i][1] = pframe_point[1];
	}

	//Matrix3f tmp;
	//tmp[0] = expand2To3(eige)

	Monomials2ndOrder::computeCoeffs( &positions[0], &weights[0], (card32)positions.size(), coeff);
	f_x = coeff[1];
	f_y = coeff[2];
	f_xy = coeff[4];
	f_xx = 2.0f * coeff[3];
	f_yy = 2.0f * coeff[5];

	gaussianCurvature = (f_xx*f_yy - pow( f_xy, 2.0f ))
		/ pow(1.0f + pow(f_x,2.0f) + pow(f_y,2.0f),2.0f );

	meanCurvature = ((1.0f + pow(f_x,2.0f))*f_yy - 2.0f * f_x * f_y * f_xy + (1.0f + pow(f_y,2.0f))*f_xx)
		/ pow(1.0f + pow(f_x,2.0f) + pow(f_y,2.0f),2.0f/3.0f ) / 2.0f;


	//debugRenderer->addPatch( point, paramToWorld, coeff, 10, 10, makeVector3f(0.8f,0.5f,0.1f), 1.0f, rad);
}


X3FOUNDATION_PRIVATE_API void X4::computeCurvature_PatchFitting( HierarchicalKNNIterator * hIt, PointCloud * pc, const Vector3f &point, float sigma, float &meanCurvature, float &gaussianCurvature )
{
	float inv_sigma_2 = 1.0f / sigma / sigma;
	float rad_2 = sigma * sigma * 4.0f;
	float rad = 2.0f * sigma;
	PCA<float,3> pca;
	AAT posAAT = pc->getAAT("position");
	AAT normalAAT;
	bool useNormals = pc->providesAttribute("normal");
	if( useNormals)
		normalAAT = pc->getAAT("normal");

	// estimate coordinate frame
	hIt->setSeekPointAndReset(point);
	card32 numPoints = 0;
	std::vector<float> weights;
	std::vector<Vector3f> positions;
	Vector3f firstNormal;
	if( useNormals )
		firstNormal = hIt->get3f(normalAAT);
	while( !hIt->atEnd())
	{
		Vector3f sample_pos = hIt->get3f(posAAT);
		float dist_2 = normQuad(sample_pos - point);
		if( dist_2 > rad_2 && numPoints >= 20 )
			break;

		numPoints++;
		float w = exp(-dist_2*inv_sigma_2);
		pca.addPoint(sample_pos, w);
		weights.push_back(w);
		positions.push_back(sample_pos);

		hIt->next();
	}

	Vector3f eigenValues,centroid;
	Matrix3f eigenVectors;
	pca.analyze(eigenValues,eigenVectors,centroid);

	Matrix3f frame = eigenVectors.transpose();

	// transform into coordinate frame
	for( unsigned i=0;i<positions.size();i++ )
	{
		positions[i] = frame * (positions[i]-centroid) / rad;
	}

	// fit monomials
	Vector6f coeff;
	Monomials2ndOrder::computeCoeffs( &positions[0], &weights[0], (card32)positions.size(), coeff);


	float f_x = coeff[1];
	float f_y = coeff[2];
	float f_xy = coeff[4];
	float f_xx = 2.0f * coeff[3];
	float f_yy = 2.0f * coeff[5];

	gaussianCurvature = (f_xx*f_yy - pow( f_xy, 2.0f ))
		/ pow(1.0f + pow(f_x,2.0f) + pow(f_y,2.0f),2.0f );

	meanCurvature = ((1.0f + pow(f_x,2.0f))*f_yy - 2.0f * f_x * f_y * f_xy + (1.0f + pow(f_y,2.0f))*f_xx)
		/ pow(1.0f + pow(f_x,2.0f) + pow(f_y,2.0f),2.0f/3.0f ) / 2.0f;
}

X3FOUNDATION_PRIVATE_API void X4::computeCurvature_PatchFitting( 
		HierarchicalKNNIterator * hIt, PointCloud * pc, const Vector3f &point, float sigma, 
		Vector3f &principalDir1, Vector3f &principalDir2, 
		float &kappa1, float &kappa2 )
{
	float inv_sigma_2 = 1.0f / sigma / sigma;
	float rad_2 = sigma * sigma * 4.0f;
	float rad = 2.0f * sigma;
	PCA<float,3> pca;
	AAT posAAT = pc->getAAT("position");
	AAT normalAAT;
	bool useNormals = pc->providesAttribute("normal");
	if( useNormals)
		normalAAT = pc->getAAT("normal");

	// estimate coordinate frame
	hIt->setSeekPointAndReset(point);
	card32 numPoints = 0;
	std::vector<float> weights;
	std::vector<Vector3f> positions;
	Vector3f firstNormal;
	if( useNormals )
		firstNormal = hIt->get3f(normalAAT);
	while( !hIt->atEnd())
	{
		Vector3f sample_pos = hIt->get3f(posAAT);
		float dist_2 = normQuad(sample_pos - point);
		if( dist_2 > rad_2 && numPoints >= 20 )
			break;

		numPoints++;
		float w = exp(-dist_2*inv_sigma_2);
		pca.addPoint(sample_pos, w);
		weights.push_back(w);
		positions.push_back(sample_pos);

		hIt->next();
	}

	Vector3f eigenValues,centroid;
	Matrix3f eigenVectors;
	pca.analyze(eigenValues,eigenVectors,centroid);

	Matrix3f worldToParam = eigenVectors.transpose();
	const Matrix3f paramToWorld = eigenVectors;

	// transform into coordinate frame
	for( unsigned i=0;i<positions.size();i++ )
	{
		positions[i] = worldToParam * (positions[i]-centroid) / rad;
	}

	// fit monomials
	Vector6f coeff;
	Monomials2ndOrder::computeCoeffs( &positions[0], &weights[0], (card32)positions.size(), coeff);


	float f_x = coeff[1];
	float f_y = coeff[2];
	float f_xy = coeff[4];
	float f_xx = coeff[3];
	float f_yy = coeff[5];

	// compute principal frame from patch
	Matrix2f M = makeMatrix2f(f_xx, f_xy, f_xy, f_yy);
	Matrix2f eigenVectors2f;
	Vector2f eigenValues2f;
	M.computeEigenStructure(eigenValues2f,eigenVectors2f);
	principalDir1 = paramToWorld[0] * eigenVectors2f[1][0]
	+ paramToWorld[1] * eigenVectors2f[1][1];
	principalDir2 = principalDir1.crossProduct(paramToWorld[2]);
	kappa1 = eigenValues[0];
	kappa2 = eigenValues[1];
}


UnstructuredInCorePointCloud* X4::copyToUnstructuredInCorePointCloud(PointCloud* input)
{
   PointCloudIterator *it = dynamic_cast<PointCloudIterator*>(input->createIterator(SOIT::CAP_BASIC_PC));
   BasicPointCloudIterator *bit = dynamic_cast<BasicPointCloudIterator*>(it);
   const PointSet *ps = bit->lockBlockRead();

   X4::UnstructuredInCorePointCloud *output = new X4::UnstructuredInCorePointCloud();
   output->setPointSet((PointSet*)ps->copy());

   bit->unlock();
   delete it;

   return output;
}

X3FOUNDATION_PRIVATE_API void X4::computeLocalSlippageAnalysis( PointCloud * pc, const Vector3f &point, float sigma, Matrix6f &eigenVectors, Vector6f &eigenValues )
{
	HierarchicalKNNIterator hIt(pc,32,NULL);
	computeLocalSlippageAnalysis(&hIt,pc,point,sigma,eigenVectors,eigenValues);
}

X3FOUNDATION_PRIVATE_API void X4::computeLocalSlippageAnalysis( HierarchicalKNNIterator * hIt, PointCloud * pc, const Vector3f &point, float sigma, Matrix6f &eigenVectors, Vector6f &eigenValues )
{
	AAT posAAT = pc->getAAT("position");
	AAT normalAAT = pc->getAAT("normal");

	// setup matrices
	static Matrix3f zeroMat3f = IDENTITY3F * 0;
	Matrix3f mat_upperleft = zeroMat3f;
	Matrix3f mat_lowerright = zeroMat3f;
	Matrix3f mat_upperright = zeroMat3f;

	// compute matrices
	float sigma_2 = sigma * sigma;
	float inv_sigma_2 = 1.0f / sigma_2;
	float rad_2 = 4.0f * sigma_2;
	float summedWeight = 0;
	hIt->setSeekPointAndReset(point);
	while(!hIt->atEnd())
	{
		Vector3f p = hIt->get3f(posAAT) - point;				
		float dist_2 = normQuad(p);
		if( dist_2 > rad_2)
		{
			break;			
		}

		Vector3f n = hIt->get3f(normalAAT);

		float weight = exp(-dist_2 * inv_sigma_2);
		summedWeight += weight;

		Vector3f c = p.crossProduct(n) / sigma;
		mat_upperleft += outerProduct(c,c) * weight;
		mat_lowerright += outerProduct(n,n) * weight;
		mat_upperright += outerProduct(n,c) * weight;

		hIt->next();
	}

	// compute slippage analysis
	Matrix6f covariance;
	for( unsigned m=0;m<3;m++ )
	{
		for( unsigned j=0;j<3;j++ )
		{
			covariance[m][j] = mat_upperleft[m][j];
			covariance[m+3][j+3] = mat_lowerright[m][j];
			covariance[m+3][j] = mat_upperright[m][j];
			covariance[m][j+3] = mat_upperright[j][m];
		}
	}

	if( summedWeight > 0.0f)
		covariance /= summedWeight;
	covariance.computeEigenStructure(eigenValues,eigenVectors);

	
}

X4::PointSet * X4::resamplePointCloudForICP( PointCloud * pc, card32 numPoints)
{
	// setup result point set 
	PointSet * retPS = new PointSet;
	VertexDescriptor * vd = createDefaultVertexDescriptor(true,false,true,false);
	retPS->clearAndSetup(1,0,vd);
	delete vd;

	// TODO: geometric stable sampling (slippage...)

	// prepare
	SceneObjectIterator * iter = pc->createIterator(SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC);
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	AAT posAAT_source = pc->getAAT("position");
	AAT normalAAT_source = pc->getAAT("normal");
	//AAT colorAAT_source = pc->getAAT("color");
	AAT posAAT_dest = retPS->getAAT("position");
	AAT normalAAT_dest = retPS->getAAT("normal");

	
	// create a spherical parameterization & lookup table
	static std::vector<unsigned> lookupTable;
	const card32 tableSizeX = 512;
	const card32 tableSizeY = 256;
	static bool parameterizationInitialized = false;
	static card32 numBins = 0;
	if( !parameterizationInitialized )
	{
		parameterizationInitialized = true;

		// create a point cloud (sphere) with 10 times the number of desired points
		UnstructuredInCorePointCloud * upc = new UnstructuredInCorePointCloud;
		VertexDescriptor * vd_tmp = createDefaultVertexDescriptor(true,false,false,true);
		upc->clearAndSetup(vd_tmp,3000);
		AAT posAAT_tmp = upc->getAAT("position");
		delete vd_tmp;
		for( unsigned i=0;i<upc->getNumPoints();i++ )
		{
			float theta = M_PI/2.0f - acos(rnd01());
			if( rand() % 2 == 0)
				theta *= -1.0f;
			float phi = rnd01() * M_PI * 2.0f;
			Vector3f pos = makeVector3f(cos(theta)*sin(phi),cos(theta)*cos(phi),sin(theta));
			upc->getPointSet()->set3f(i,posAAT_tmp,pos);

			
		}
	
		// resample
		float medPointDistance = getMedianPointDistance(upc);
		PCCResampler::resamplePC(upc,medPointDistance * 5.0f);
		numBins = upc->getNumPoints();

		// create lookup table
		lookupTable = std::vector<unsigned>(tableSizeX * tableSizeY);
		HierarchicalKNNIterator hIt(upc,32,NULL);
		for( unsigned x=0;x<tableSizeX;x++ )
		{
			for( unsigned y=0;y<tableSizeY;y++ )
			{
				unsigned tableEntry = y * tableSizeX + x;
				float theta = ((float)y / tableSizeY*2.0f - 1.0f) * M_PI / 2.0f;
				float phi = (float)x / tableSizeX * 2.0f * M_PI;
				Vector3f pos = makeVector3f(cos(theta)*sin(phi),cos(theta)*cos(phi),sin(theta));
				hIt.setSeekPointAndReset(pos);
				lookupTable[tableEntry] = hIt.getCurrentPointIndex();
			}
		}

		delete upc;
	}

	// setup density function
	std::vector<float> bins(numBins,0 );
	std::vector<std::vector<unsigned> > binIndices(numBins);
	unsigned currentIndex = 0;
	std::vector<unsigned> flags(pc->getNumPoints(),0);
	while( !bIt->atEnd())
	{
		Vector3f normal = bIt->get3f(normalAAT_source);
		float theta = asin(normal[2]);
		float phi = atan2(normal[0],normal[1]);
		int entryX = phi / 2.0f / M_PI * (float)tableSizeX;
		int entryY = (theta / M_PI * 2.0f + 1.0f) * (float)tableSizeY / 2.0f;
		if( entryX < 0 )
			entryX += tableSizeX;
		if( entryY < 0 )
			entryY += tableSizeY;
		unsigned entry = entryY * tableSizeX + entryX;
		if( entry >= lookupTable.size() )
		{
			debugOutput << "lookup error: normal = " << normal 
				<< "\ntheta = " << theta
				<< "\nphi = " << phi << "\n";
		}
		else
		{
			bins[ lookupTable[entry] ] ++;
			binIndices[lookupTable[entry]].push_back(currentIndex);
		}

		currentIndex++;
		bIt->next();
	}

	// invert density function 
	float summedDensity = 0;
	float minDensity = 10;
	for( unsigned  i=0;i<bins.size();i++ )
	{
		if( bins[i] >= minDensity )
		{
			bins[i] = 1.0f / bins[i];
		}
		else
			bins[i] = 0;
		summedDensity += bins[i];
	}


	// sampling
	float remainingSummedDensity = 0;
	std::vector<unsigned> sampledIndices;
	unsigned numPoints_normalSpaceSampled = numPoints * 0.5f;
	for( unsigned i=0;i<bins.size();i++ )
	{
		if( bins[i] > 0 )
		{
			float numPointsThisBin = bins[i] / summedDensity * numPoints_normalSpaceSampled;
			float fracPart = numPointsThisBin - (int)numPointsThisBin;
			numPointsThisBin += (rnd01() > fracPart) ? 1:0;
			for( unsigned j=0;j<numPointsThisBin && binIndices[i].size() > 0;j++ )
			{
				unsigned e = rand() % binIndices[i].size();
				sampledIndices.push_back( binIndices[i][e] );
				flags[ binIndices[i][e]  ] = 1;
				if( binIndices[i].size() == 1)
				{
					binIndices[i].clear();
					break;
				}
				binIndices[i][e] = binIndices[i].back();
				binIndices[i].pop_back();
			}

			remainingSummedDensity += bins[i];
		}
	}
	
	//debugOutput << "num normal-space samples: "  << sampledIndices.size() << "\n";
	
	// uniform sampling to match the number of required samples
	std::vector<unsigned> indices;
	for( unsigned i=0;i<pc->getNumPoints();i++ )
	{
		if( flags[i] == 0)
			indices.push_back(i);
	}

	while( sampledIndices.size() < numPoints && indices.size())
	{
		unsigned i = random32bit((card32)indices.size());
		sampledIndices.push_back( indices[i] );
		indices[i] = indices.back();
		indices.pop_back();
	}

	// copy data
	retPS->changeHeight(sampledIndices.size());
	const PointSet * sourcePS = bIt->lockBlockRead();
	for( unsigned i=0;i<sampledIndices.size();i++ )
	{
		retPS->set3f(i,posAAT_dest,sourcePS->get3f(sampledIndices[i],posAAT_source));
		retPS->set3f(i,normalAAT_dest,sourcePS->get3f(sampledIndices[i],normalAAT_source));
	}

	bIt->unlock();
	delete iter;

	return retPS;
}

void X4::computeLocalFrameAndPatchCoeffs( const Vector3f &point, float sigma, PointCloud * pc, HierarchicalKNNIterator * hIt, Matrix3f &frame, Vector3f &center, Vector6f & coeffs, bool orientWithNormals/*=false */ )
{
	float inv_sigma_2 = 1.0f / sigma / sigma;
	float rad_2 = sigma * sigma * 4.0f;
	float rad = 2.0f * sigma;
	PCA<float,3> pca;
	AAT posAAT = pc->getAAT("position");
	AAT normalAAT;
	bool useNormals = pc->providesAttribute("normal");
	if( useNormals)
		normalAAT = pc->getAAT("normal");

	// estimate coordinate frame
	hIt->setSeekPointAndReset(point);
	card32 numPoints = 0;
	std::vector<float> weights;
	std::vector<Vector3f> positions;
	Vector3f firstNormal;
	if( useNormals )
		firstNormal = hIt->get3f(normalAAT);
	while( !hIt->atEnd())
	{
		Vector3f sample_pos = hIt->get3f(posAAT);
		float dist_2 = normQuad(sample_pos - point);
		if( dist_2 > rad_2 && numPoints >= 20 )
			break;

		numPoints++;
		float w = 1;//exp(-dist_2*inv_sigma_2);
		pca.addPoint(sample_pos, w);
		weights.push_back(w);
		positions.push_back(sample_pos);

		hIt->next();
	}

	Vector3f eigenValues;
	Matrix3f eigenVectors;
	pca.analyze(eigenValues,eigenVectors,center);
	
	if( eigenVectors[2] * firstNormal < 0 )
	{
		eigenVectors[2] *= -1;
		eigenVectors[1] *= -1;
	}

	center = point - eigenVectors[2] * (eigenVectors[2] * (point - center));
	frame = eigenVectors;
	Matrix3f frameT = frame.transpose();

	// transform into coordinate frame
	for( unsigned i=0;i<positions.size();i++ )
	{
		positions[i] = frameT * (positions[i]-center) / rad;
	}

	// fit monomials
	Monomials2ndOrder::computeCoeffs( &positions[0], &weights[0], (card32)positions.size(), coeffs);

}

bool X4::projectPointRIMLS( const Vector3f &point, float sigmaR, float sigmaN, float h, PointCloud * pc, Vector3f &projPoint, Vector3f &projNormal )
{
	HierarchicalKNNIterator hIt(pc,32,NULL);
	return projectPointRIMLS(point,sigmaR,sigmaN,h, pc, &hIt, projPoint, projNormal);
}

float RIMLS_phi(float x, float h)
{
	return pow( 1.0f - x*x / (h*h), 4.0f);
}

float RIMLS_dphi(float x, float h)
{
	return (x)*(-2.0f / h * 4.0f * pow( 1.0f - x*x / (h*h), 3.0f));
}

bool X4::projectPointRIMLS( const Vector3f &point, float sigmaR, float sigmaN, float h, PointCloud * pc, HierarchicalKNNIterator * hIt, Vector3f &projPoint, Vector3f &projNormal )
{
	AAT posAAT = pc->getAAT("position");
	AAT normalAAT = pc->getAAT("normal");

	float rad_2 = h * h;
	float inv_sigmaR_2 = 1.0f / sigmaR / sigmaR;
	float inv_sigmaN_2 = 1.0f / sigmaN / sigmaN;

	projPoint = point;
	
	// estimate initial normal
	Vector3f grad_f = NULL_VECTOR3F;
	hIt->setSeekPointAndReset(point);
	Vector3f n0 = hIt->get3f(normalAAT);
	hIt->next();
	unsigned numPoints = 0;
	float sumW = 0;
	while(!hIt->atEnd())
	{
		float dist_2 = normQuad(point - hIt->get3f(posAAT));
		if( dist_2 > rad_2 && numPoints > 10)
			break;

		numPoints++;
		float weight = exp(-dist_2 * inv_sigmaR_2) * RIMLS_phi(sqrt(dist_2),h);
		sumW += weight;
		grad_f += hIt->get3f(normalAAT) * weight;

		hIt->next();
	}

	grad_f /= sumW;
	
	const unsigned maxNumIterations = 20;
	for( unsigned i=0;i<maxNumIterations;i++ )
	{
		float f = 0;
		for( unsigned j=0;j<maxNumIterations;j++ )
		{
			float sumW = 0;
			Vector3f sumGw = NULL_VECTOR3F;
			float sumF = 0;
			Vector3f sumGF = NULL_VECTOR3F;
			Vector3f sumN = NULL_VECTOR3F;

			hIt->setSeekPointAndReset(projPoint);
			unsigned numPoints = 0;
			while( !hIt->atEnd() )
			{
				Vector3f px = projPoint - hIt->get3f(posAAT);
				float px_normQ = normQuad(px);
				if( px_normQ > rad_2 && numPoints > 10 )
					break;
				numPoints++;
				float px_norm = sqrt(px_normQ);

				Vector3f normal = hIt->get3f(normalAAT);
				float fx = px * normal;
				float alpha = 1.0f;
				if( i != 0)
					alpha = exp( - pow(fx - f,2.0f) * inv_sigmaR_2 ) 
								* 	exp( - normQuad(normal - grad_f) * inv_sigmaN_2);

				float w = alpha * RIMLS_phi(px_norm,h);
				Vector3f grad_w = px * alpha * 2.0f * RIMLS_dphi(px_norm,h);

				sumW += w;
				sumGw += grad_w;
				sumF += w * fx;	sumGF += grad_w * fx;
				sumN += normal * w;

				hIt->next();
			}

			f = sumF / sumW;
			Vector3f newGrad = (sumGF - sumGw * f + sumN) / sumW;
			bool conv = norm(newGrad - grad_f) < 0.001f;
			grad_f = newGrad;
			if( conv)
				break;
			
		}

		projPoint -= grad_f * f;
		if( norm(grad_f * f) < h * 0.005f )		
			break;
	}
	projNormal = normalize(grad_f);
	return true;
}

float X4::computeSignedAngle( const Vector3f &dir1, const Vector3f &dir2, const Vector3f &surfaceNormal )
{
	Matrix3f frame;
	frame[0] = normalize(dir1);
	frame[1] = normalize(frame[0].crossProduct(surfaceNormal));
	frame[2] = normalize(frame[1].crossProduct(frame[0]));

	const Vector3f tdir2 = frame.transpose() * dir2;
	float ret=atan2(tdir2[0], tdir2[1]);
	if( ret < -(float32)M_PI_2 )
		ret += ((float32)M_PI_2 + (float32)M_PI);
	else
		ret -= (float32)M_PI_2;
	return ret;
}

//-------------------------------------------------------------------------------------------------
X4::InCorePCTopologyGraph* X4::createTopologyFromTriangleMesh(UnstructuredInCoreTriangleMesh * mesh)
{
	InCorePCTopologyGraph* tpg = new InCorePCTopologyGraph;
	VertexDescriptor vd;
	vd.pushAttrib( VAD("vindex", 2, VAD::DATA_FORMAT_INT32));
	tpg->clearAndSetup(&vd);

	AAT indexAAT = mesh->getTriangles()->getAAT("index");
	for (card64 i=0; i < mesh->getTriangles()->getNumEntries(); i++)
	{
		Vector3i tri = mesh->getTriangles()->get3i(i, indexAAT);

		if (!tpg->areVerticesAdjacent(tri[0], tri[1]))
			tpg->addEdge(tri[0], tri[1]);

		if (!tpg->areVerticesAdjacent(tri[0], tri[2]))
			tpg->addEdge(tri[0], tri[2]);

		if (!tpg->areVerticesAdjacent(tri[1], tri[2]))
			tpg->addEdge(tri[1], tri[2]);
	}

	return tpg;
}

//-------------------------------------------------------------------------------------------------
UnstructuredInCorePointCloud* X4::createTopologyUPCFromTriangleMesh(UnstructuredInCoreTriangleMesh * mesh, float sampleRadius)
{
	checkAttribute(mesh,"tri_id",1,VAD::DATA_FORMAT_INT32);

	/*float epsilon = getMedianPointDistance(mesh);
	float sampleRadius = subsampleEpsilonFactor * epsilon;
	if (subsampleEpsilonFactor < 1e-10f) sampleRadius = epsilon;
	*/

	UnstructuredInCorePointCloud * upc = new UnstructuredInCorePointCloud;
	upc->clearAndSetup(mesh->getDescr(),0);
	const VertexDescriptor * sourceVertexDescr = mesh->getDescr();

	// create (oversampled) point cloud
	VertexArray va_vertex(mesh);
	VertexArray va_tri(mesh->getTriangles());
	PointSet * vertices = mesh->getPointSet();
	float oversample_radius = sampleRadius*0.15f;
	std::vector<char*> vbuff(3);
	std::vector<char> destPointBuffer(sourceVertexDescr->getSize());
	AAT tri_idAAT = upc->getAAT("tri_id");
	for( unsigned i=0;i<va_tri.getNumElements();i++ )
	{
		Vector3i tri = va_tri.getIndex3i(i);
		Vector3f p[3];
		for( unsigned j=0;j<3;j++ )
		{
			p[j] = va_vertex.getPosition3f(tri[j]);
		}

		Vector3f normal = (p[1]-p[0]).crossProduct(p[2] - p[0]);
		float area = norm(normal)*0.5f;
		normal = normalize(normal);

		unsigned numSamples = ceil(area / oversample_radius / oversample_radius);
		vbuff[0] = vertices->getDataPointer(tri[0]);
		vbuff[1] = vertices->getDataPointer(tri[1]);
		vbuff[2] = vertices->getDataPointer(tri[2]);
		for( unsigned j=0;j<numSamples;j++ )
		{

			float32 u = rnd01();
			float32 v = rnd01();
			if (u+v > 1) {
				float32 nu = 1.0f - v;
				v = 1.0f - u;
				u = nu;
			}

			float32 factors[3];
			factors[0] = u;
			factors[1] = v;
			factors[2] = 1.0f - u - v;

			sourceVertexDescr->interpolate((const char**)&vbuff[0], factors, 3, &destPointBuffer[0]);
			upc->getPointSet()->addPoint(&destPointBuffer[0]);

			// store face index
			upc->getPointSet()->set1i(upc->getNumPoints()-1,tri_idAAT,i);
		}
	}

	// resample point cloud
	PCCResampler::resamplePC(upc, sampleRadius);

	// setup lookup table vertex->neighbor vertex & vertex->face
	std::vector< std::vector<unsigned> > vertex_neighbors(mesh->getNumPoints());
	std::vector< std::vector<unsigned> > vertex_to_face(mesh->getNumPoints());
	for( unsigned i=0;i<va_tri.getNumElements();i++ )
	{
		Vector3i tri = va_tri.getIndex3i(i);
		for( unsigned j=0;j<3;j++ )
		{
			vertex_neighbors[tri[j]].push_back(tri[(j+1)%3]);
			vertex_neighbors[tri[j]].push_back(tri[(j+2)%3]);
			vertex_to_face[tri[j]].push_back(i);
		}
	}

	// unify vertex topology
	for( unsigned i=0;i<vertex_neighbors.size();i++ )
	{
		std::vector<unsigned> vn = vertex_neighbors[i];
		sort(vn.begin(),vn.end());
		for( unsigned j=0;j+1<vn.size(); )
		{
			if( vn[j] == vn[j+1] )
			{
				vn.erase(vn.begin()+j);
			}
			else
				j++;
		}
		vertex_neighbors[i] = vn;
	}


	// create topology
	InCorePCTopologyGraph * graph = new InCorePCTopologyGraph;
	VertexDescriptor vd;
	vd.pushAttrib(VAD("vindex", 2, VAD::DATA_FORMAT_INT32));
	graph->clearAndSetup(&vd);
	VertexArray va_result(upc);
	HierarchicalKNNIterator hIt(upc,32,NULL);
	hIt.setMaxDistanceToSeekPoint(sampleRadius*2.5f);
	HierarchicalKNNIterator hIt_mesh(mesh,32,NULL);
	float maxSampleRadMesh = sampleRadius*4.0f;
	std::vector<unsigned> meshVertexTag(mesh->getNumPoints(),0);
	unsigned currentVertexTag = 0;
	std::vector<unsigned> faceTag(mesh->getTriangles()->getNumEntries(),0);
	unsigned currentFaceTag = 0;
	for( unsigned i=0;i<va_result.getNumElements();i++ )
	{
		// query vertices in original mesh
		unsigned myFaceID = upc->getPointSet()->get1i(i,tri_idAAT);
		Vector3i myTri = va_tri.getIndex3i(myFaceID);
		unsigned meshAnchor = myTri[0];
		std::vector<unsigned> stack; stack.reserve(10); stack.push_back(meshAnchor);
		currentVertexTag++;
		currentFaceTag++;
		meshVertexTag[meshAnchor] = currentVertexTag;

		Vector3f pos = va_result.getPosition3f(i);
		while (!stack.empty())
		{
			unsigned vindex = stack.back();
			stack.pop_back();

			// tag active face index
			std::vector<unsigned> &faces = vertex_to_face[vindex];
			const unsigned numFaces = (unsigned)faces.size();
			for( unsigned j=0;j<numFaces;j++ )
			{
				faceTag[faces[j]] = currentFaceTag;
			}

			// iterate
			for( unsigned j=0;j<vertex_neighbors[vindex].size();j++ )
			{
				unsigned nindex = vertex_neighbors[vindex][j];
				if( meshVertexTag[nindex] != currentVertexTag && norm(pos - va_vertex.getPosition3f(nindex)) < maxSampleRadMesh)
				{
					meshVertexTag[nindex] = currentVertexTag;
					stack.push_back(nindex);
				}
			}
		}

		// query all points in sphere
		hIt.setSeekPointAndReset(va_result.getPosition3f(i));
		hIt.next();
		static int test = 0;
		while(!hIt.atEnd())
		{
			unsigned fIndex	= hIt.get1i(tri_idAAT);
			if( faceTag[fIndex] != currentFaceTag )
			{
				hIt.next();
				continue;
			}

			unsigned nIndex = hIt.getCurrentPointIndex();
			graph->addEdge(i,nIndex);

			hIt.next();
		}
	}

	TopologyAttachment * topAtt = new TopologyAttachment;
	topAtt->setup(TopologyAttachment::getDefaultName(),AttachedData::ADF_PERSISTENT);
	topAtt->setTopology(graph);
	upc->getAttachments()->attachData(topAtt);

	return upc;
}

X3FOUNDATION_PRIVATE_API Matrix4f X4::computeTransformationFromCorrespondences( const std::vector<Vector3f> &points1, const std::vector<Vector3f> &points2 )
{
	if( points1.size() != points1.size() )
	{
		message("you have to select exactly the same number of points in each point cloud - returning identity");
		return IDENTITY4F;
	}

	if( points1.size() < 4 )
	{
		message("not enough points- returning identity");
		return IDENTITY4F;		
	}

	// compute centroids
	Vector3f centroid1 = NULL_VECTOR3F;
	Vector3f centroid2 = NULL_VECTOR3F;
	for( unsigned i=0;i<points1.size();i++ )
	{
		centroid1 += points1[i];
		centroid2 += points2[i];
	}
	centroid1 /= points1.size();
	centroid2 /= points2.size();

	// calculate cross-covariance matrix
	Matrix3f C = IDENTITY3F * 0;
	for( card32 i=0;i<points1.size();i++ )
	{
		for( card32 j=0;j<3;j++ )
		{
			for( card32 k=0;k<3;k++ )
			{
				C[j][k] += (points1[i][j] - centroid1[j])
					* (points2[i][k] - centroid2[k]);
			}
		}
	}
	C /= points1.size();

	// Prepare OpenCV Datastructure CvMat
	Matrix3f U,V;
	Vector3f eigenValues;
	CvMat cvMatInput, cvMat_U, cvMat_W, cvMat_Vt;
	cvInitMatHeader( &cvMatInput, 3, 3, CV_32FC1, C.data());
	cvInitMatHeader( &cvMat_W, 3, 1, CV_32FC1, eigenValues.data() );
	cvInitMatHeader( &cvMat_U, 3, 3, CV_32FC1, U.data() );
	cvInitMatHeader( &cvMat_Vt, 3, 3, CV_32FC1, V.data() );

	// Do SVD
	cvSVD(&cvMatInput, &cvMat_W, &cvMat_U, &cvMat_Vt, CV_SVD_MODIFY_A | CV_SVD_U_T );

	Matrix3f rotation = U*V;
	Matrix4f transformation = makeTranslation4f(centroid1) * expand3To4(rotation) * makeTranslation4f(-centroid2);

	return transformation;
}

X3FOUNDATION_PRIVATE_API Matrix4f X4::computeRigidTransformationFromCorrespondences( const std::vector<Vector3f> &points1, const std::vector<Vector3f> &points2 )
{
    const size_t num_points = points1.size();

	if ( num_points != points2.size() ) {
		error("you have to select exactly the same number of points in each point cloud - returning identity");
		return IDENTITY4F;
	}

	if ( num_points < 4 ) {
		error("not enough points- returning identity");
		return IDENTITY4F;		
	}

	// compute centroids
	Vector3f centroid1 = NULL_VECTOR3F;
	Vector3f centroid2 = NULL_VECTOR3F;
	for ( size_t i=0;i<num_points;i++ )
    {
		centroid1 += points1[i];
		centroid2 += points2[i];
	}
	centroid1 /= num_points;
	centroid2 /= num_points;

	// calculate cross-covariance matrix
	Matrix3f C = IDENTITY3F * 0;
	for( size_t i=0;i<num_points;i++ )
	{
		for( card32 j=0;j<3;j++ )
		{
			for( card32 k=0;k<3;k++ )
			{
				C[j][k] += (points1[i][j] - centroid1[j])
					* (points2[i][k] - centroid2[k]);
			}
		}
	}
	C /= num_points;

	// Prepare OpenCV Datastructure CvMat
	Matrix3f U,V;
	Vector3f eigenValues;
	CvMat cvMatInput, cvMat_U, cvMat_W, cvMat_Vt;
	cvInitMatHeader( &cvMatInput, 3, 3, CV_32FC1, C.data());
	cvInitMatHeader( &cvMat_W, 3, 1, CV_32FC1, eigenValues.data() );
	cvInitMatHeader( &cvMat_U, 3, 3, CV_32FC1, U.data() );
	cvInitMatHeader( &cvMat_Vt, 3, 3, CV_32FC1, V.data() );

	// Do SVD
	cvSVD(&cvMatInput, &cvMat_W, &cvMat_U, &cvMat_Vt, CV_SVD_MODIFY_A | CV_SVD_U_T );

	Matrix3f rotation = U*V;
	U[2] *= rotation.getDeterminant();
	rotation = U*V;
	Matrix4f transformation = makeTranslation4f(centroid1) * expand3To4(rotation) * makeTranslation4f(-centroid2);

	return transformation;
}

void X4::resamplePointCloud_Grid( UnstructuredInCorePointCloud * upc, float cellSize )
{
	Simple3DHashedGrid grid(cellSize);

	AAT posAAT = upc->getAAT("position");
	PointSet * ps = upc->getPointSet();
	for( mpcard i=upc->getNumPoints();i>0;i-- )
	{
		Vector3f p = ps->get3f(i-1,posAAT);
		if( grid.insert(p) )
		{
			ps->deletePoint(i-1);
		}
	}
}

PointCloud * X4::upsamplePointCloud( PointCloud * pc, float32 targetSampleSpacing )
{
	// setup output point cloud
	UnstructuredInCorePointCloud * outputPC = (UnstructuredInCorePointCloud*)pc->copy();
	PointSet * outputPS = outputPC->getPointSet();

	// setup stuff
	HierarchicalKNNIterator hItSource(pc,32,NULL);
	SceneObjectIterator *iter = pc->createIterator( SOIT( SOIT::CAP_BASIC_PC ) );
	BasicPointCloudIterator* bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	AAT posAAT = pc->getAAT("position");
	AAT normalAAT = pc->getAAT("normal");
	AAT sampleSpacingAAT = pc->getAAT("sampleSpacing");
	
	// insert new points (creates sample points within radius sampleSpacing around a point,
	// project samples on the surface with sigma=sampleSpacing*eps)
	bIt->reset();
	while (!bIt->atEnd())
	{
		Vector3f pos = bIt->get3f(posAAT);
		Vector3f normal = bIt->get3f(normalAAT);
		float sampleSpacing = bIt->get1f(sampleSpacingAAT);
			
		if( sampleSpacing < targetSampleSpacing )
		{
			float smoothingSigma = max(sampleSpacing*1.5f,targetSampleSpacing*1.5f);
			float sampleSpacingOrg = sampleSpacing;
			sampleSpacing *= 1.75f;
			Matrix3f tangentSystem = calcTangentSystem(normal);

			// compute number of samples necessary (overestimated)
			float numSamplesf = sqr<float>(sampleSpacing/targetSampleSpacing*2.0f)*4;
			mpcard numSamples = numSamplesf;
			numSamplesf += rnd01() < (numSamplesf-(float)numSamples)?1:0;
			for( mpcard i=0;i<numSamples;i++ )
			{
				// compute random sample point on disk
				float r = sqrt(rnd01())*sampleSpacing;
				float angle = rnd01()*M_PI*2.0f;
				Vector3f p = pos + tangentSystem*makeVector3f(sin(angle),cos(angle),0)*r;
				
				// project point on smooth surface
				Vector3f projPoint,projNormal;
				projectPointOntoPatch(p,smoothingSigma, pc, &hItSource, projPoint, projNormal);

				mpcard index = outputPS->getNumEntries();
				outputPS->addPoint(NULL);
				outputPS->set3f(index,posAAT,projPoint);
				outputPS->set3f(index,normalAAT,projNormal);
				outputPS->set1f(index,sampleSpacingAAT,sampleSpacingOrg);
			}
		}

		bIt->next();
	}

	// resample
	PCCResampler::resamplePC(outputPC,targetSampleSpacing*.75f,true);
	AAT flagsAAT = outputPC->getAAT("flags");
	for( mpint i=(mpint)outputPC->getNumPoints()-1;i>=(mpint)pc->getNumPoints();i-- )
	{
		if( outputPS->get1i(i,flagsAAT) != 0 )
		{
			outputPS->deletePoint(i);
		}
	}
	outputPC->clearAttachments();

	delete iter;
	return outputPC;
}

X3FOUNDATION_PRIVATE_API float X4::compareSlippageMotion( const Matrix6f &eigenVectors0, mpcard numSlipDim, const Vector6f &eigenVectorTest )
{
	if( numSlipDim == 1)
	{
		float x1 = eigenVectors0[5]*eigenVectorTest;
		return normQuad(eigenVectors0[5]*x1 - eigenVectorTest);
	}
	else
	if( numSlipDim == 2)
	{
		Matrix2f M;
		Vector2f d;
		for( mpcard i=0;i<2;i++ )
		{
			d[i] = eigenVectorTest*eigenVectors0[5-i];
			for( mpcard j=i;j<2;j++ )
			{
				M[i][j] = M[j][i] = eigenVectors0[5-i]*eigenVectors0[5-j];
			}
		}
		Vector2f x = invertMatrix(M)*d;
		Vector6f tmp = NULL_VECTOR6F;
		for( mpcard i=0;i<2;i++)
			tmp += eigenVectors0[5-i]*x[i];
		return normQuad(tmp - eigenVectorTest);
	}
	else
		if( numSlipDim == 3)
		{
			Matrix3f M;
			Vector3f d;
			for( mpcard i=0;i<3;i++ )
			{
				d[i] = eigenVectorTest*eigenVectors0[5-i];
				for( mpcard j=i;j<3;j++ )
				{
					M[i][j] = M[j][i] = eigenVectors0[5-i]*eigenVectors0[5-j];
				}
			}
			Vector3f x = invertMatrix(M)*d;
			Vector6f tmp = NULL_VECTOR6F;
			for( mpcard i=0;i<3;i++)
				tmp += eigenVectors0[5-i]*x[i];
			return normQuad(tmp - eigenVectorTest);
		}

	return 1e20f;
}

X3FOUNDATION_PRIVATE_API float X4::compareSlippageMotionRotated( const Matrix6f &eigenVectors0, const Vector3f &pos0, mpcard numSlipDim, const Vector6f &eigenVectorTest, const Vector3f &testPos, float sigma )
{
	float residual = compareSlippageMotion(eigenVectors0,numSlipDim,eigenVectorTest);

	Vector3f rot = makeVector3f(eigenVectorTest[0],eigenVectorTest[1],eigenVectorTest[2]);

	if( norm(rot) < 0.1f)
	{
		return residual;
	}

	// rotate translation part by pure, numerically unstable magic
	Matrix3f rot_frame;
	Vector3f trans = makeVector3f(eigenVectorTest[3],eigenVectorTest[4],eigenVectorTest[5]);
	rot_frame[0] = normalize(rot);
	rot_frame[1] = normalize(trans);
	rot_frame[2] = normalize(rot_frame[0].crossProduct(rot_frame[1]));

	float beta = M_PI/2.0f - norm(rot);
	float m = norm(trans) * sigma;
	float d = m * tan(beta);
	Vector3f center = rot_frame * makeVector3f(0,0,d) + testPos;
	Vector3f p1 = rot_frame.transpose() * (pos0 - center);

	float angle = -atan2(p1[1],-p1[2]);
	Vector3f newTrans = rot_frame*makeRotX3f(angle)*rot_frame.transpose()*trans;
	Vector6f rotatedTestVector = eigenVectorTest;
	rotatedTestVector[3] = newTrans[0];
	rotatedTestVector[4] = newTrans[1];
	rotatedTestVector[5] = newTrans[2];
	return min(residual,compareSlippageMotion(eigenVectors0, numSlipDim,rotatedTestVector));
}


X3FOUNDATION_PRIVATE_API void X4::doTopologicalMarchingApplyCallback(mpcard startIndex, UnstructuredInCorePointCloud* pc, InCorePCTopologyGraph* tpg, TopologicalMarchingCallback* callback)
{
	std::vector<char> visited(pc->getNumPoints(), 0);
	queue<card32> visit; visit.push(startIndex);
	while(!visit.empty())
	{
		card32 pt = visit.front(); visit.pop();
		if (visited[pt]) continue;
		visited[pt] = 1;
		if (!(*callback)(pc, tpg, pt)) continue;

		for (unsigned t=0; t < tpg->getNumAdjacentVertices(pt); t++)
		{
			card32 v = tpg->getAdjacentVertex(pt,t);
			if (visited[v]) continue;
			visit.push(v);
		}
	}
}

X3FOUNDATION_PRIVATE_API void X4::doMarchingFromGivenPointCompareDistancesSetCallback(mpcard startIndex, UnstructuredInCorePointCloud* pc, InCorePCTopologyGraph* tpg, 
		AAT posAAT, AAT distAAT, MarchingCallback* callback)
{
	PointSet* ps = pc->getPointSet();

	struct TDistanceIndexPair_Pre
	{
		float dist;
		mpcard index;
		bool operator < (const TDistanceIndexPair_Pre &o) const {return dist > o.dist;}
		TDistanceIndexPair_Pre(float d, mpcard i) {dist = d; index = i;}
	};

	// we store in the queue elements by their distance 
	std::vector<float> inqueue(pc->getNumPoints(), 1e20f);
	priority_queue<TDistanceIndexPair_Pre> pq;
	pq.push(TDistanceIndexPair_Pre(0,startIndex));

	while(!pq.empty())
	{
		// get next element from the queue
		TDistanceIndexPair_Pre top = pq.top(); pq.pop();
		float dist = ps->get1f(top.index, distAAT);

		if (dist < top.dist) continue;

		// set current distance from the start point to the reached point
		ps->set1f(top.index, distAAT, top.dist);
		if (callback && !(*callback)(pc, tpg, top.index, top.dist)) break;

		// for each of the neighbors check if their distance is higher then currently reached
		// then include them into the next queue
		// if their distance is lower, then we can't propagate their further
		for( unsigned i=0; i < tpg->getNumAdjacentVertices(top.index); i++ )
		{
			mpcard neighbour = tpg->getAdjacentVertex(top.index, i);

			float distance = top.dist + (ps->get3f(top.index, posAAT) - ps->get3f(neighbour, posAAT)).getNorm();

			// in case if the distance on the neighbor is larger then it could be, then we update it on the next step
			if (ps->get1f(neighbour, distAAT) > distance && inqueue[neighbour] > distance)
			{
				inqueue[neighbour] = distance;
				pq.push(TDistanceIndexPair_Pre(distance,neighbour));
			}
		}
	}
}

//--------------------------------------------------------------------------------------
X3FOUNDATION_PRIVATE_API float X4::computeHausdorffDistance(UnstructuredInCorePointCloud* pc1, UnstructuredInCorePointCloud* pc2)
{
	return computeHausdorffDistance(pc1, pc2, IDENTITY4F, IDENTITY4F);
}

X3FOUNDATION_PRIVATE_API float X4::computeHausdorffDistance(UnstructuredInCorePointCloud* pc1, UnstructuredInCorePointCloud* pc2, const Matrix4f& oneTotwo, const Matrix4f& twoToone)
{
	AAT pos1AAT = pc1->getAAT("position");
	AAT pos2AAT = pc2->getAAT("position");

	float maxdist = 0;
	unsigned _p = 0;
	unsigned _q = 0;

	if (pc1->getNumPoints() == 0 || pc2->getNumPoints() == 0) return 1e20f;

	// check the nereast distance of each point of shape B on shape A
	{
		HierarchicalKNNIterator hIt(pc1, 32, NULL);
		for (unsigned p=0; p < pc2->getNumPoints(); p++)
		{
			Vector3f pos2 = shrink4To3(twoToone * expand3To4(pc2->getPointSet()->get3f(p, pos2AAT)));

			hIt.setSeekPointAndReset(pos2);

			float dist = (hIt.get3f(pos1AAT) - pos2).getNorm();

			if (dist > maxdist)
			{
				_q = p;
				_p = hIt.getCurrentPointIndex();
				maxdist = dist;
			}
		}
	}

	// perform same check but now of shape A on shape B
	{
		HierarchicalKNNIterator hIt(pc2, 32, NULL);
		for (unsigned p=0; p < pc1->getNumPoints(); p++)
		{
			Vector3f pos1 = shrink4To3(oneTotwo * expand3To4(pc1->getPointSet()->get3f(p, pos1AAT)));

			hIt.setSeekPointAndReset(pos1);

			float dist = (hIt.get3f(pos2AAT) - pos1).getNorm();

			if (dist > maxdist)
			{
				_q = hIt.getCurrentPointIndex();
				_p = p;
				maxdist = dist;
			}
		}
	}

	/*if (maxdist > 0.01)
	{
		debugOutput << "p = " << _p << ", q = " << _q << "\n";
		debugRenderer->addLine(pc1->getPointSet()->get3f(_p, pos1AAT), pc2->getPointSet()->get3f(_q, pos2AAT), makeVector4f(1,1,0,1), makeVector4f(1,1,0,1), 3.0f);
	}*/
	return maxdist;
}

//--------------------------------------------------------------------------------------
X3FOUNDATION_PRIVATE_API float X4::computeAvgPointToPlaneDistance(UnstructuredInCorePointCloud* pc1, UnstructuredInCorePointCloud* pc2, float cutoffThreshold)
{
	AAT pos1AAT = pc1->getAAT("position");
	AAT pos2AAT = pc2->getAAT("position");

	AAT nrm1AAT = pc1->getAAT("normal");
	AAT nrm2AAT = pc2->getAAT("normal");

	if (pc1->getNumPoints() == 0 || pc2->getNumPoints() == 0) return 1e20f;

	double sumdist = 0;

	// check the nereast distance of each point of shape B on shape A
	HierarchicalKNNIterator hIt(pc1, 32, NULL);
	for (unsigned i=0; i < pc2->getNumPoints(); i++)
	{
		Vector3f pos2 = pc2->getPointSet()->get3f(i, pos2AAT);
		Vector3f nrm2 = pc2->getPointSet()->get3f(i, nrm2AAT);
		hIt.setSeekPointAndReset(pos2);

		Vector3f pos1 = pc1->getPointSet()->get3f(hIt.getCurrentPointIndex(), pos1AAT);
		Vector3f nrm1 = pc1->getPointSet()->get3f(hIt.getCurrentPointIndex(), nrm1AAT);

		Plane3f plane(nrm1, pos1);
		float dist = fabs(plane.calculateSignedDistance(pos2));
		if (dist > cutoffThreshold) dist = cutoffThreshold;

		sumdist += double(dist);
	}
	return sumdist / double(pc2->getNumPoints());
}

X3FOUNDATION_PRIVATE_API float X4::computeAvgPointToPlaneDistance(UnstructuredInCorePointCloud* pc1, UnstructuredInCorePointCloud* pc2, float cutoffThreshold, int32 subTreeSize)
{
	AAT pos1AAT = pc1->getAAT("position");
	AAT pos2AAT = pc2->getAAT("position");

	AAT nrm1AAT = pc1->getAAT("normal");
	AAT nrm2AAT = pc2->getAAT("normal");

	if (pc1->getNumPoints() == 0 || pc2->getNumPoints() == 0) return 1e20f;

	double sumdist = 0;

	// check the nereast distance of each point of shape B on shape A
	HierarchicalKNNIterator hIt(pc1, subTreeSize, NULL);
	for (unsigned i = 0; i < pc2->getNumPoints(); i++)
	{
		Vector3f pos2 = pc2->getPointSet()->get3f(i, pos2AAT);
		Vector3f nrm2 = pc2->getPointSet()->get3f(i, nrm2AAT);
		hIt.setSeekPointAndReset(pos2);

		Vector3f pos1 = pc1->getPointSet()->get3f(hIt.getCurrentPointIndex(), pos1AAT);
		Vector3f nrm1 = pc1->getPointSet()->get3f(hIt.getCurrentPointIndex(), nrm1AAT);

		Plane3f plane(nrm1, pos1);
		float dist = fabs(plane.calculateSignedDistance(pos2));
		if (dist > cutoffThreshold) dist = cutoffThreshold;

		sumdist += double(dist);
	}
	return sumdist / double(pc2->getNumPoints());
}


X3FOUNDATION_PRIVATE_API float X4::computeAvgPointToPointDistance(UnstructuredInCorePointCloud* pc1, UnstructuredInCorePointCloud* pc2, float cutoffThreshold)
{
	AAT pos1AAT = pc1->getAAT("position");
	AAT pos2AAT = pc2->getAAT("position");

	AAT nrm1AAT = pc1->getAAT("normal");
	AAT nrm2AAT = pc2->getAAT("normal");

	if (pc1->getNumPoints() == 0 || pc2->getNumPoints() == 0) return 1e20f;

	double sumdist = 0;

	// check the nereast distance of each point of shape B on shape A
	HierarchicalKNNIterator hIt(pc1, 32, NULL);
	for (unsigned i = 0; i < pc2->getNumPoints(); i++)
	{
		Vector3f pos2 = pc2->getPointSet()->get3f(i, pos2AAT);
		hIt.setSeekPointAndReset(pos2);

		Vector3f pos1 = pc1->getPointSet()->get3f(hIt.getCurrentPointIndex(), pos1AAT);
		Vector3f diff = pos2 - pos1;
		float dist = diff.getNorm();
		if (dist > cutoffThreshold) dist = cutoffThreshold;

		sumdist += double(dist);
	}
	return sumdist / double(pc2->getNumPoints());
}


//--------------------------------------------------------------------------------------
bool SortRemovePointPredicate (mpcard i, mpcard j) { return (i>j); }

X3FOUNDATION_PRIVATE_API void X4::keepPointsInPointCloud(UnstructuredInCorePointCloud* upc, const std::vector<mpcard>& _points, InCorePCTopologyGraph* tpg, InCorePCTopologyGraph** newtpg)
{
	if (_points.size() == upc->getNumPoints()) return;

	std::vector<mpcard> all(upc->getNumPoints());
	for (unsigned i=0; i < upc->getNumPoints(); i++)
		all[i] = i;

	std::vector<mpcard> points = _points;
	sort(points.begin(), points.end());

	std::vector<mpcard> tmp(all.size());
	std::vector<mpcard>::iterator it = set_difference(all.begin(), all.end(), points.begin(), points.end(), tmp.begin());

	removePointsFromPointCloud(upc, std::vector<mpcard>(tmp.begin(), it), tpg, newtpg);
}

//--------------------------------------------------------------------------------------
X3FOUNDATION_PRIVATE_API UnstructuredInCorePointCloud* X4::sampleFromPointCloud(UnstructuredInCorePointCloud* upc, const std::vector<mpcard>& points, InCorePCTopologyGraph* tpg, InCorePCTopologyGraph** newtpg)
{
	UnstructuredInCorePointCloud* newpc = new UnstructuredInCorePointCloud;
	newpc->clearAndSetup(upc->getDescr(), points.size());

	PointSet* newps = newpc->getPointSet();
	PointSet* ps = upc->getPointSet();

	std::vector<mpcard> mapped(ps->getNumEntries(), (mpcard)-1);

	// TODO: incorporate width as well
	for (size_t i=0; i < points.size(); i++)
	{
		newps->setVertex(i, ps->getDataPointer(points[i]));
		mapped[points[i]] = i;
	}

	if (newtpg && tpg)
	{
		if (*newtpg) delete (*newtpg);
		(*newtpg) = new InCorePCTopologyGraph;
		(*newtpg)->clearAndSetup(tpg->getDescr());

		AAT indexAAT = tpg->getEdges()->getAAT("vindex");
		char* tmpBuffer = new char[tpg->getDescr()->getSize()];

		for (size_t j=0; j < points.size(); j++)
		{
			PointCloudTopologyGraph::VertexIndex v = points[j];
			x4Assert(mapped[v] != (mpcard)-1);

			for (unsigned i=0; i < tpg->getNumAdjacentVertices(v); i++)
			{
				InCorePCTopologyGraph::VertexIndex vi = tpg->getAdjacentVertex(v,i);
				if (mapped[vi] == (mpcard)-1) continue;

				memset(tmpBuffer, 0, tpg->getDescr()->getSize() * sizeof(char));

				indexAAT.setAttrib<int32,2>(tmpBuffer, makeVector2i(mapped[v] , mapped[vi]));

				(*newtpg)->addEdge(tmpBuffer);
			}
		}
		delete [] tmpBuffer;
	}

	return newpc;
}

X3FOUNDATION_PRIVATE_API std::vector<mpcard> X4::removeSmallPatches(UnstructuredInCorePointCloud* geometry, InCorePCTopologyGraph* tpg, InCorePCTopologyGraph** newTpg, int minNumberOfPoints)
{
	if (tpg == NULL) return std::vector<mpcard>();

	std::vector<int> regionId(geometry->getNumPoints(), -1);
	int currentRegion = -1;
	std::vector<int> numPointsInRegion;
	std::vector<card32> startPointIndex;

	// mark point with regions
	for (unsigned v=0; v < regionId.size(); v++)
	{
		if (regionId[v] < 0)
		{
			currentRegion++;
			numPointsInRegion.push_back(0);
			startPointIndex.push_back(v);

			TopologicalKNNIterator_Fast tIt(geometry, tpg->createIterator(geometry->getPointSet()));
			tIt.setStartPointAndReset(v);
			while(!tIt.atEnd())
			{
				numPointsInRegion[currentRegion]++;
				regionId[tIt.getPointSetPointNumber()] = currentRegion;
				tIt.next();
			}
		}
	}


	// collect all points which are inside of a small region
	std::vector<mpcard> pointsToDelete;
	for (unsigned r=0; r < numPointsInRegion.size(); r++)
	{
		if (numPointsInRegion[r] < minNumberOfPoints)
		{
			TopologicalKNNIterator_Fast tIt(geometry, tpg->createIterator(geometry->getPointSet()));
			tIt.setStartPointAndReset(startPointIndex[r]);
			while(!tIt.atEnd())
			{
				pointsToDelete.push_back(tIt.getPointSetPointNumber());
				tIt.next();
			}
		}
	}

	if (dynamic_cast<UnstructuredInCoreTriangleMesh*>(geometry))
		dynamic_cast<UnstructuredInCoreTriangleMesh*>(geometry)->deletePoints(pointsToDelete);
	else
		removePointsFromPointCloud(geometry, pointsToDelete, tpg, newTpg);

	return pointsToDelete;
}

X3FOUNDATION_PRIVATE_API  std::vector<mpcard>  X4::removeUnconnectedPoints(UnstructuredInCorePointCloud* geometry, InCorePCTopologyGraph* tpg, InCorePCTopologyGraph** newTpg)
{
	std::vector<mpcard> toDelete;
	for (unsigned i=0; i < geometry->getNumPoints(); i++)
		if (tpg->getNumAdjacentVertices(i) <= 1)
			toDelete.push_back(i);

	if (dynamic_cast<UnstructuredInCoreTriangleMesh*>(geometry))
		dynamic_cast<UnstructuredInCoreTriangleMesh*>(geometry)->deletePoints(toDelete);
	else
		removePointsFromPointCloud(geometry, toDelete, tpg, newTpg);
	
	return toDelete;
}

X3FOUNDATION_PRIVATE_API void X4::removePointsFromPointCloud(UnstructuredInCorePointCloud* upc, const std::vector<mpcard>& _points, InCorePCTopologyGraph* tpg, InCorePCTopologyGraph** newtpg)
{
	if (_points.size() == 0) return;

	std::vector<mpcard> points = _points;
	std::sort(points.begin(), points.end(), SortRemovePointPredicate);

	unsigned oldSize = upc->getPointSet()->getHeight();

	// std::map of new indices
	std::vector<int> indices(upc->getPointSet()->getHeight());
	std::vector<int> lastVertices(upc->getPointSet()->getHeight(), -1);
	for (size_t v=0; v < indices.size(); v++)
		indices[v] = (int)v;

	int lastIndex = upc->getPointSet()->getHeight() - 1;
	size_t newSize = upc->getPointSet()->getHeight() - points.size();
	for (std::vector<mpcard>::iterator it = points.begin(); it != points.end(); it++)
	{
		if (*it < newSize)
		{
			for (; indices[lastIndex] == -1; lastIndex--)
				x4Assert((mpint)lastIndex >= (mpint)newSize);

			indices[*it] = lastIndex;
			lastVertices[lastIndex] = *it;
			indices[lastIndex] = -1;
		}else
		{
			x4Assert(indices[*it] >= 0);
			indices[*it] = -1;
		}
	}

	for (size_t v=0; v < indices.size(); v++)
		if (indices[v] == v) lastVertices[v] = (int)v;

	// do swapping
	for (unsigned i=0; i < newSize; i++)
	{
		// do nothing if index not changed
		x4Assert(indices[i] >= 0);
		if (i == indices[i]) continue;

		// copy from the stored index to correct position
		for (mpcard j=0; j < upc->getPointSet()->getWidth(); j++)
			upc->getPointSet()->setVertex(j, i, upc->getPointSet()->getDataPointer(j, indices[i]));
	}

	upc->getPointSet()->changeHeight(newSize);

	// change topology
	if (newtpg && tpg)
	{
		if (*newtpg) delete (*newtpg);
		(*newtpg) = new InCorePCTopologyGraph;
		(*newtpg)->clearAndSetup(tpg->getDescr());

		AAT indexAAT = tpg->getEdges()->getAAT("vindex");

		for (unsigned v=0; v < oldSize; v++)
		{
			for (unsigned i=0; i < tpg->getNumAdjacentVertices(v); i++)
			{
				InCorePCTopologyGraph::VertexIndex vi = tpg->getAdjacentVertex(v,i);
				if (lastVertices[v] < 0 || lastVertices[vi] < 0) continue;

				char* tmpBuffer = new char[tpg->getDescr()->getSize()];
				memset(tmpBuffer, 0, tpg->getDescr()->getSize() * sizeof(char));

				indexAAT.setAttrib<int32,2>(tmpBuffer, makeVector2i(lastVertices[v] , lastVertices[vi]));

				(*newtpg)->addEdge(tmpBuffer);

				delete [] tmpBuffer;
			}
		}
	}

}

X3FOUNDATION_PRIVATE_API unsigned X4::findConnectedComponents(UnstructuredInCorePointCloud* upc, InCorePCTopologyGraph* tpg, std::vector<std::vector<mpcard>>& components)
{
	std::vector<int> ptComponent(upc->getNumPoints(), -1);

	components.clear();

	// compute connected components
	unsigned numsegments = 0;
	PointSet* ps = upc->getPointSet();
	{
		for (unsigned c=0; c < upc->getNumPoints(); c++)
		{
			if (ptComponent[c] > -1) continue;

			int numpoints = 0;

			std::vector<char> marked(upc->getNumPoints(), 0);
			queue<card32> next;
			next.push(c);

			std::vector<mpcard> component;

			while(!next.empty())
			{
				card32 idx = next.front(); next.pop();
				if (ptComponent[idx] > -1) continue;

				ptComponent[idx] = (int)components.size();
				component.push_back(idx);
				numpoints++;

				for (unsigned t=0; t < tpg->getNumAdjacentVertices(idx); t++)
				{
					if (ptComponent[tpg->getAdjacentVertex(idx,t)] < 0 && marked[tpg->getAdjacentVertex(idx,t)] == 0)
					{
						next.push(tpg->getAdjacentVertex(idx,t));
						marked[tpg->getAdjacentVertex(idx,t)] = 1;
					}
				}
			}

			numsegments++;
			components.push_back(component);
		}
	}

	return numsegments;
}

X4::ComputeGeodeticDistanceShortRange::ComputeGeodeticDistanceShortRange( PointCloud * pc, InCorePCTopologyGraph * topGraph, card32 maxNumNearestNeighbors, float maxNeighborDistance )
{
	m_DistanceField = std::vector<float>(pc->getNumPoints(),CGD_MAX_DISTANCE());
	SceneObjectIterator * iter = pc->createIterator(SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(iter);
	m_Iter = iter;
	m_PointSet = mIt->lockBlockReadWrite();
	m_TopGraph = topGraph;
	m_hIt = new HierarchicalKNNIterator(pc,32,NULL);
	m_PC = pc;
	m_MaxNeighborDistance = maxNeighborDistance;
	m_MaxNumNearestNeighbors = maxNumNearestNeighbors;
}


std::vector<float> X4::ComputeGeodeticDistanceShortRange::compute(
    const Vector3f &startPoint, const std::vector<Vector3f> points,std::vector< std::vector<int32> > * paths,
    DistanceFunction customDistance, void *distanceContext)
{
	std::vector<float> distances(points.size(),CGD_MAX_DISTANCE());
	AAT posAAT = m_PC->getAAT("position");

	unsigned maxNearPoints = m_MaxNumNearestNeighbors;
	// compute nearest points to destination points and mark entries in distance field with -1
	// also stores a std::map for every marked point
	std::map<unsigned,std::vector<unsigned> > indexToPoint;
	for( unsigned i=0;i<points.size();i++ )
	{
		// don't add start point
		if( norm(startPoint - points[i]) < 1e-20f )
		{
			distances[i] = 0.0f;
			continue;
		}

		m_hIt->setSeekPointAndReset(points[i]);
		card32 numNeighbors = 0;
		while( !m_hIt->atEnd() && numNeighbors < maxNearPoints)
		{
			Vector3f p = m_hIt->get3f(posAAT);
			float dist = norm(p - points[i]);
			if( dist > m_MaxNeighborDistance)
				break;

			unsigned index = m_hIt->getCurrentPointIndex();
			indexToPoint[index].push_back(i);
			m_DistanceField[index] = CGD_NEG_MAX_DISTANCE();

			numNeighbors++;
			if( dist < 0.000001f )
				break;	// target point is on the graph
			m_hIt->next();
		}
	}

	list<TComputeGeodeticDistance_PointIndex> index_queue;

	// add neighbors of start point to queue
	m_hIt->setSeekPointAndReset(startPoint);
	std::map<unsigned,unsigned> entryPoints;
	for( unsigned i=0;i<maxNearPoints;i++ )
	{
		Vector3f pos = m_hIt->get3f(posAAT);
		float dist = norm(pos - startPoint);
		if( dist > m_MaxNeighborDistance)
			break;

		TComputeGeodeticDistance_PointIndex entry(m_hIt->getCurrentPointIndex(),1);
		index_queue.push_back(entry);
		m_DistanceField[entry.m_Index] = norm(startPoint - pos);
		entryPoints[entry.m_Index] = 1;

		if( dist< 0.000001f )
			break;	// source point is on the graph
		m_hIt->next();
	}

	unsigned currentHop = 1;
	float maxDestintaionPathDistance = 1e31f;
	float minPathDistance_ThisHop = 1e30f;
	
	// grow 
	while( !index_queue.empty() )
	{
		TComputeGeodeticDistance_PointIndex entry = index_queue.front();
		index_queue.pop_front();

		// next hop? 
		if( currentHop < entry.m_NumHops )
		{
			currentHop = entry.m_NumHops;
			if( minPathDistance_ThisHop > maxDestintaionPathDistance )
			{
				break;
			}
			minPathDistance_ThisHop = 1e30f;
		}

		const Vector3f pos = m_PointSet->get3f(entry.m_Index,posAAT);
		const float currentPathDistance = fabs(m_DistanceField[entry.m_Index]);

		//debugRenderer->addPoint(pos,makeVector3f(1,0.5,0));

		const unsigned numNeighbors = m_TopGraph->getNumAdjacentVertices(entry.m_Index);
		for( unsigned i=0;i<numNeighbors;i++ )
		{
			unsigned nIndex = m_TopGraph->getAdjacentVertex(entry.m_Index,i);
			const Vector3f nPos = m_PointSet->get3f(nIndex,posAAT);
			float dist = (customDistance==NULL) ?
			   norm(nPos - pos) :
   			customDistance(entry.m_Index, nIndex, distanceContext);
			
			float pathDistance = dist + currentPathDistance;
			minPathDistance_ThisHop = min(minPathDistance_ThisHop,pathDistance);

			const float distanceFieldEntry = m_DistanceField[nIndex];
			if( distanceFieldEntry < 0.0f)
				if( distanceFieldEntry == CGD_NEG_MAX_DISTANCE() || fabs(distanceFieldEntry) > pathDistance )
			{
				// we found an "exit point"
				for( unsigned j=0;j<indexToPoint[nIndex].size();j++ )
				{
					unsigned dest_point_index = indexToPoint[nIndex][j];
					float exitPointDistance = norm(nPos - points[dest_point_index]);
					if( distances[dest_point_index] > pathDistance + exitPointDistance )
					{
						distances[dest_point_index] = pathDistance + exitPointDistance;

						// check other distance entries
						maxDestintaionPathDistance = 0;
						for( unsigned k=0;k<distances.size();k++ )
								maxDestintaionPathDistance = max(maxDestintaionPathDistance,distances[k]);

					}
				}

				// expand node
				m_DistanceField[nIndex] = -pathDistance;
				TComputeGeodeticDistance_PointIndex nEntry(nIndex,entry.m_NumHops+1);
				index_queue.push_back(nEntry);

			}

			if( distanceFieldEntry > pathDistance )
			{
				m_DistanceField[nIndex] = pathDistance;
				TComputeGeodeticDistance_PointIndex nEntry(nIndex,entry.m_NumHops+1);
				index_queue.push_back(nEntry);
			}

		}

	}

	// path backtracking
	paths->clear();

	// for all query points
	for( card32 i=0;i<points.size();i++ )
	{
		//find best start point from the 5 nearest neighbors
		m_hIt->setSeekPointAndReset( points[i] );
		PointIndex index = m_hIt->getPointSetPointNumber();
		for( card32 j=0;j<maxNearPoints && !m_hIt->atEnd();j++ )
		{
			unsigned currentIndex = m_hIt->getPointSetPointNumber();
			float distToNeighbor = norm( points[i] - m_hIt->get3f(posAAT));
			float tmpDist = fabs(m_DistanceField[currentIndex]) + distToNeighbor;
			if( j == 0 || tmpDist < distances[i] && distToNeighbor < m_MaxNeighborDistance)
			{
				distances[i] = tmpDist;
				index = currentIndex;
			}
			if( distToNeighbor< 0.000001f )
				break;	// source point is on the graph
			m_hIt->next();
		}

		if( points[i] == startPoint )
			distances[i] = 0.0f;


		//calculate paths

		std::vector<int32> path;
		float lastDistance = 1e30f;
		while( true )
		{
			path.push_back( index );
			if( path.size() > 100000 )
				throw X4Exception("computeGeodeticDistances path too long");

			float minDist = 1e30f;
			PointIndex bestNeighbour;
			for( card32 j=0;j<m_TopGraph->getNumAdjacentVertices(index);j++ )
			{
				PointIndex neighbour = m_TopGraph->getAdjacentVertex(index,j);
				if( fabs(m_DistanceField[neighbour]) < minDist )
				{
					minDist = fabs(m_DistanceField[neighbour]);
					bestNeighbour = neighbour;
				}
			}
			if( minDist >= lastDistance )
			{
				break;
			}
			lastDistance = minDist;

			index = bestNeighbour;
		}

		// remove double entry points
		while( path.size() > 3)
		{
			if( entryPoints.find(path[path.size() - 2]) != entryPoints.end() )
				path.pop_back();
			else
				break;
		}

		paths->push_back( path );
	}

	// clear distances
	m_hIt->setSeekPointAndReset(startPoint);
	for( unsigned i=0;i<maxNearPoints;i++ )
	{
		TComputeGeodeticDistance_PointIndex entry(m_hIt->getCurrentPointIndex(),1);
		index_queue.push_back(entry);
		m_DistanceField[entry.m_Index] = CGD_MAX_DISTANCE();

		m_hIt->next();
	}

	while( !index_queue.empty() )
	{
		TComputeGeodeticDistance_PointIndex entry = index_queue.front();
		index_queue.pop_front();

		for( unsigned i=0;i<m_TopGraph->getNumAdjacentVertices(entry.m_Index);i++ )
		{
			unsigned nIndex = m_TopGraph->getAdjacentVertex(entry.m_Index,i);
			if( m_DistanceField[nIndex] != CGD_MAX_DISTANCE() )
			{
				m_DistanceField[nIndex] = CGD_MAX_DISTANCE();
				TComputeGeodeticDistance_PointIndex e(nIndex,entry.m_NumHops + 1);
				index_queue.push_back(e);
			}
		}
	}

	// reset near points
	for( std::map<unsigned,std::vector<unsigned> >::iterator iter = indexToPoint.begin();
		iter != indexToPoint.end(); ++iter)
		m_DistanceField[iter->first] = CGD_MAX_DISTANCE();

	// check
	/*for( unsigned i=0;i<m_DistanceField.size();i++ )
	{
		if( m_DistanceField[i] != CGD_MAX_DISTANCE() )
		{
			throw X4Exception("ComputeGeodeticDistanceShortRange::compte() - found invalid distance entry after cleanup!");
		}
	}/**/

	return distances;
	
}


std::vector<float> X4::ComputeGeodeticDistanceShortRange::compute( const Vector3f &startPoint, const std::vector<Vector3f> points,std::vector< std::vector<mpcard> > * paths /*= NULL*/ )
{
	std::vector< std::vector< int32> > dummy;
	std::vector<float> floatVals = compute(startPoint, points, &dummy);

	for (mpcard i = 0; i < dummy.size(); i++)
	{
		std::vector<mpcard> tmp;
		for (mpcard j = 0; j < dummy[i].size();  j++)
		{
			int32 val = dummy[i][j];
			tmp.push_back(val);


		}
		paths->push_back(tmp);
	}

	return floatVals;
}


X4::ComputeGeodeticDistanceShortRange::~ComputeGeodeticDistanceShortRange()
{
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(m_Iter);
	bIt->unlock();
	delete m_Iter;
	delete m_hIt;
}


bool X4::Simple3DHashedGrid::entryExists( const Vector3f &p ) const
{
	card32 data;
	return m_Grid.getElement( computeCellIndex(p), data);
}

bool X4::Simple3DHashedGrid::insert( const Vector3f &p )
{
	card32 data = 0;
	Vector3i key = computeCellIndex(p);
	bool element_exists = m_Grid.getElement( key, data);
	if( !element_exists )
	{
		data = 0;
		m_Grid.addElement(key,data);
	}
	return element_exists;
}

X4::Vector3i X4::Simple3DHashedGrid::computeCellIndex( const Vector3f &p ) const
{
	return makeVector3i(	(p[0]<0.0f)?p[0]*m_InvGridSize-1:p[0]*m_InvGridSize,
							(p[1]<0.0f)?p[1]*m_InvGridSize-1:p[1]*m_InvGridSize,
							(p[2]<0.0f)?p[2]*m_InvGridSize-1:p[2]*m_InvGridSize);
}