#ifndef GeometricToolsH
#define GeometricToolsH
//---------------------------------------------------------------------------
#include "x3foundation_private.h"
//---------------------------------------------------------------------------
#include "HashTables.inline.h"
//---------------------------------------------------------------------------

namespace X4
{
	class PointCloud;
	class HierarchicalKNNIterator;
   class UnstructuredInCorePointCloud;
	class SceneObjectIterator;
	class InCorePCTopologyGraph;
	class HierarchicalKNNIterator;
	class Scene;
	class UnstructuredInCoreTriangleMesh;

	/// computes a local frame via weighted pca
	X3FOUNDATION_PRIVATE_API void computeLocalFrame( const Vector3f &point, float sigma, PointCloud * pc, 
		Matrix3f &frame, Vector3f &center, bool orientWithNormals=false );

	/// computes a local frame via weighted pca
	X3FOUNDATION_PRIVATE_API void computeLocalFrame( const Vector3f &point, float sigma, PointCloud * pc,
		HierarchicalKNNIterator * hIt, Matrix3f &frame, Vector3f &center, bool orientWithNormals=false );

	/// compute local frame and patch coefficients
	X3FOUNDATION_PRIVATE_API void computeLocalFrameAndPatchCoeffs( const Vector3f &point, float sigma, PointCloud * pc,
		HierarchicalKNNIterator * hIt, Matrix3f &frame, Vector3f &center, Vector6f & coeffs, bool orientWithNormals=false );


	/// creates a noisy copy of a point cloud
	X3FOUNDATION_PRIVATE_API PointCloud * createNoisyCopy( PointCloud * pc, float stdDeviation, float maxAmplitude,
		bool noiseInNormalDirection);

// ---------------------------------------------------------------------------------------------------------------
//				surface projection 
// ---------------------------------------------------------------------------------------------------------------

	/// computes a projection onto a quadratic patch (which is computed by least squares fitting)
	X3FOUNDATION_PRIVATE_API bool projectPointOntoPatch( const Vector3f &point, float sigma, PointCloud * pc, 
		Vector3f &projPoint, Vector3f &projNormal);
	X3FOUNDATION_PRIVATE_API bool projectPointOntoPatch( const Vector3f &point, float sigma, PointCloud * pc, HierarchicalKNNIterator * hIt, 
		Vector3f &projPoint, Vector3f &projNormal);
	
	/// computes a RIMLS projection
	X3FOUNDATION_PRIVATE_API bool projectPointRIMLS( const Vector3f &point, float sigmaR, float sigmaN, float h, PointCloud * pc, 
		Vector3f &projPoint, Vector3f &projNormal);
	X3FOUNDATION_PRIVATE_API bool projectPointRIMLS( const Vector3f &point, float sigmaR, float sigmaN, float h, PointCloud * pc, HierarchicalKNNIterator * hIt, 
		Vector3f &projPoint, Vector3f &projNormal);

	/// returns the coefficients of the quadratic function that describes a quadratic patch
	X3FOUNDATION_PRIVATE_API Vector6f getPatchCoefficients( const Vector3f &point, float sigma,
		PointCloud * pc,  Matrix3f &frame, HierarchicalKNNIterator * hIt);


	/// computes a high resolution point cloud with desired sample spacing, requires channel sample spacing
	/// uses MLS projection to insert new points
	X3FOUNDATION_PRIVATE_API PointCloud * upsamplePointCloud(PointCloud * pc, float32 targetSampleSpacing);

// ---------------------------------------------------------------------------------------------------------------
//				dijkstra / geodetic distance
// ---------------------------------------------------------------------------------------------------------------

	/// computes geodetic distances between a startPoint and a set of points
	/// for debug purposes: returns geodetic paths from startPoint to all points
	X3FOUNDATION_PRIVATE_API std::vector<float> computeGeodeticDistances( PointCloud * pc, PointCloudIterator * topologyIter, 
							const Vector3f startPoint, const std::vector<Vector3f> points,
							std::vector< std::vector<Vector3f> > * paths = NULL);


	struct TGDFContext
	{
		PointCloud * pc;
		PointSet*  ps;
		AAT posAAT;
		AAT normalAAT;
	};
	typedef float (*TGeneralDistanceFunction)(int,int,int,TGDFContext*);
	typedef float (*TDistanceFunction)(card32 edgeIndex, TGDFContext*);

	/// computes dijkstra paths for general distance functions
	X3FOUNDATION_PRIVATE_API std::vector<float> computeGeodeticDistances( PointCloud * pc, PointCloudIterator * topologyIter, 
		mpcard startIndex, const std::vector<mpcard> &points, TGeneralDistanceFunction distFunc, TGDFContext * context,
		std::vector< std::vector<Vector3f> > * paths = NULL, float visualizeIsoLines = -1.0f);
   X3FOUNDATION_PRIVATE_API std::vector<float> computeGeodeticDistances( PointCloud * pc, PointCloudIterator * topologyIter, 
      Vector3f startPoint, const std::vector<Vector3f> &points, TGeneralDistanceFunction distFunc, TGDFContext * context,
      std::vector< std::vector<Vector3f> > * paths = NULL, float visualizeIsoLines = -1.0f);

	/// computes geodetic distances between a startPoint  and a set of points
	/// returns geodetic paths of indexes from startPoint to all points
	X3FOUNDATION_PRIVATE_API std::vector<float> computeGeodeticDistancesAndPathes( PointCloud * pc, PointCloudIterator * topologyIter, 
		const Vector3f &startPoint, const std::vector<Vector3f> points,
		std::vector< std::vector<int32> > * paths = NULL);

	/// computes geodetic distances&derivates between a start point and a set of points
	/// uses a weighted least squares fit to estimate the intrinsic gradient
	/// for debug purposes: returns geodetic paths from startPoint to all points
	X3FOUNDATION_PRIVATE_API std::vector<float> computeGeodeticDistances( PointCloud * pc, PointCloudIterator * topologyIter, 
		float sigma, 
		const Vector3f startPoint, const std::vector<Vector3f> points, std::vector<Vector3f> &derivates,
		std::vector< std::vector<Vector3f> > * paths = NULL);

	/// computes geodetic distances&derivates between a start point(index) and a set of points (indices)
	/// for debug purposes: returns geodetic paths from startPoint to all points
	X3FOUNDATION_PRIVATE_API std::vector<float> computeGeodeticDistances( PointCloud * pc, PointCloudIterator * topologyIter, 
		float sigma,
		const unsigned startIndex, const std::vector<unsigned> &indices, std::vector<Vector3f> &derivates,
		std::vector< std::vector<Vector3f> > * paths = NULL);


	class X3FOUNDATION_PRIVATE_API ComputeGeodeticDistanceShortRange
	{
	private:
		std::vector<float>	m_DistanceField;
		PointCloud *	m_PC;
		PointSet *		m_PointSet;
		SceneObjectIterator	*	m_Iter;
		InCorePCTopologyGraph	*	m_TopGraph;
		HierarchicalKNNIterator	* m_hIt;
		card32 m_MaxNumNearestNeighbors;
		float m_MaxNeighborDistance;

		inline float CGD_MAX_DISTANCE() {return 1e30f; }
		inline float CGD_NEG_MAX_DISTANCE() {return -1e30f;}


	public:
	   /// arguments: first vertex index, second vertex index, user defined context ptr
      typedef float (*DistanceFunction)(int,int,void*);

		ComputeGeodeticDistanceShortRange(PointCloud * pc, InCorePCTopologyGraph * topGraph, card32 maxNumNearestNeighbors = 6, float maxNeighborDistance=1e20f);
		~ComputeGeodeticDistanceShortRange();

		std::vector<float>	compute(const Vector3f &startPoint, const std::vector<Vector3f> points,std::vector< std::vector<int32> > * paths = NULL,
		                       DistanceFunction customDistance = NULL, void *distanceContext = NULL);
		std::vector<float>	compute(const Vector3f &startPoint, const std::vector<Vector3f> points,std::vector< std::vector<mpcard> > * paths = NULL);

	};


	struct X3FOUNDATION_PRIVATE_API MarchingCallback
	{
		virtual bool operator()(UnstructuredInCorePointCloud* pc, InCorePCTopologyGraph* tpg, mpcard index, float dist) = 0;
	};
	struct X3FOUNDATION_PRIVATE_API TopologicalMarchingCallback
	{
		virtual bool operator()(UnstructuredInCorePointCloud* pc, InCorePCTopologyGraph* tpg, mpcard index) = 0;
	};

	/// Compute Hausdorff Distance between two point clouds. Given matrices will be applied on each of the point cloud before upfront
	X3FOUNDATION_PRIVATE_API float computeHausdorffDistance(UnstructuredInCorePointCloud* pc1, UnstructuredInCorePointCloud* pc2, const Matrix4f& oneTotwo, const Matrix4f& twoToone);

	/// Compute Hausdorff Distance between two point clouds. Shapes must be already prealigned
	X3FOUNDATION_PRIVATE_API float computeHausdorffDistance(UnstructuredInCorePointCloud* pc1, UnstructuredInCorePointCloud* pc2);
 
	/// Compute avg ICP (point to plane) distance. Given thresholds can be used to cut the distance (if larger then given threshold, then set to threshold)
	X3FOUNDATION_PRIVATE_API float computeAvgPointToPlaneDistance(UnstructuredInCorePointCloud* pc1, UnstructuredInCorePointCloud* pc2, float cutoffThreshold = 1e20f);
	X3FOUNDATION_PRIVATE_API float computeAvgPointToPlaneDistance(UnstructuredInCorePointCloud* pc1, UnstructuredInCorePointCloud* pc2, float cutoffThreshold, int32 subTreeSize);

	X3FOUNDATION_PRIVATE_API float computeAvgPointToPointDistance(UnstructuredInCorePointCloud* pc1, UnstructuredInCorePointCloud* pc2, float cutoffThreshold = 1e20f);

	X3FOUNDATION_PRIVATE_API void doMarchingFromGivenPointCompareDistancesSetCallback(mpcard startIndex, UnstructuredInCorePointCloud* pc, InCorePCTopologyGraph* tpg, 
		AAT posAAT, AAT distAAT, MarchingCallback* callback = NULL);

	/// run starting from given point over all topologically reachable points (in any order) and apply callback. If callback returns false, then neighbors of current point are not further processed
	X3FOUNDATION_PRIVATE_API void doTopologicalMarchingApplyCallback(mpcard startIndex, UnstructuredInCorePointCloud* pc, InCorePCTopologyGraph* tpg, TopologicalMarchingCallback* callback = NULL);
// ---------------------------------------------------------------------------------------------------------------
//				normals, curvature, & slippage
// ---------------------------------------------------------------------------------------------------------------

	/// smooth normals
	X3FOUNDATION_PRIVATE_API void smoothNormals( PointCloud * sourcePC, PointCloud * destPC, std::string destChannel, float sigma );

	/// compute local curvature
	X3FOUNDATION_PRIVATE_API void computeCurvature_PatchFitting( PointCloud * pc, const Vector3f &point, float sigma, float &meanCurvature, float &gaussianCurvature );
	X3FOUNDATION_PRIVATE_API void computeCurvature_PatchFitting( HierarchicalKNNIterator * hIt, PointCloud * pc, const Vector3f &point, float sigma, float &meanCurvature, float &gaussianCurvature );

	/// compute local curvature
	X3FOUNDATION_PRIVATE_API void computeCurvature_PatchFitting( PointCloud * pc, const Vector3f &point, float sigma, float &meanCurvature, float &gaussianCurvature, 
		Vector3f &principalDir1, Vector3f &principalDir2 );
	
	X3FOUNDATION_PRIVATE_API void computeCurvature_PatchFitting( HierarchicalKNNIterator * hIt, PointCloud * pc, const Vector3f &point, float sigma, 
		Vector3f &principalDir1, Vector3f &principalDir2, float &kappa1, float &kappa2 );


	/// compute local curvature using a boxed window function
	X3FOUNDATION_PRIVATE_API void computeCurvature_PatchFittingBoxed( HierarchicalKNNIterator * hIt, PointCloud * pc, const Vector3f &point, float sigma, float &meanCurvature, float &gaussianCurvature, 
		Vector3f &principalDir1, Vector3f &principalDir2 );



   /// get unstructured incor epoint cloud from any point cloud
   X3FOUNDATION_PRIVATE_API UnstructuredInCorePointCloud* copyToUnstructuredInCorePointCloud(PointCloud* pc);
   

	/// compute slippage analysis for a given surface patch (p,sigma)
	X3FOUNDATION_PRIVATE_API void computeLocalSlippageAnalysis( PointCloud * pc, const Vector3f &point, float sigma, Matrix6f &eigenVectors, Vector6f &eigenValues );
	X3FOUNDATION_PRIVATE_API void computeLocalSlippageAnalysis( HierarchicalKNNIterator * hIt, PointCloud * pc, const Vector3f &point, float sigma, Matrix6f &eigenVectors, Vector6f &eigenValues );

	/// computes residual of (eigenVectors[1,2,3]*(x1,x2,x3)- eigenVectorTest)^2 (numSlipDim specifies the number of used eigenvectors)
	/// assumes orthogonal vectors
	X3FOUNDATION_PRIVATE_API float compareSlippageMotion(const Matrix6f &eigenVectors0, mpcard numSlipDim, const Vector6f &eigenVectorTest);

	/// rotates motion of eigenVectorTest into frame of eigenVectors (if rotation is involved) before comparing
	X3FOUNDATION_PRIVATE_API float compareSlippageMotionRotated(const Matrix6f &eigenVectors0, const Vector3f &pos0, mpcard numSlipDim, const Vector6f &eigenVectorTest, const Vector3f &testPos, float sigma);

// ---------------------------------------------------------------------------------------------------------------
//				misc
// ---------------------------------------------------------------------------------------------------------------

	
	/// subsample point cloud for icp registration
	X3FOUNDATION_PRIVATE_API PointSet *  resamplePointCloudForICP( PointCloud * pc, card32 numPoints);

	/// creates a coordinate frame with frame[0] = dir1 and frame[2] ~ surfaceNormal
	/// projects dir2 into that frame and returns atan2 from the projected dir2
	X3FOUNDATION_PRIVATE_API float computeSignedAngle(const Vector3f &dir1, const Vector3f &dir2, const Vector3f &surfaceNormal);

	/// computes best transformation (in a least squares sense) based on corresponding points
	/// number of source/dest points must be equal and at least 4
	X3FOUNDATION_PRIVATE_API Matrix4f computeTransformationFromCorrespondences( const std::vector<Vector3f> &sourcePoints, const std::vector<Vector3f> &destPoints);

	/// computes best transformation (in a least squares sense) based on corresponding points and avoids mirroring
	/// number of source/dest points must be equal and at least 4
	X3FOUNDATION_PRIVATE_API Matrix4f computeRigidTransformationFromCorrespondences( const std::vector<Vector3f> &sourcePoints, const std::vector<Vector3f> &destPoints);

	/// create point cloud with topology graph from given triangle mesh
	X3FOUNDATION_PRIVATE_API UnstructuredInCorePointCloud* createTopologyUPCFromTriangleMesh(UnstructuredInCoreTriangleMesh * mesh, float sampleRadius);

	/// create topology graph from the given triangle mesh (topology graph corresponds to the point cloud of the mesh)
	X3FOUNDATION_PRIVATE_API InCorePCTopologyGraph* createTopologyFromTriangleMesh(UnstructuredInCoreTriangleMesh * mesh);

	/// remove given points from the point cloud. If tpg is given, then it will be replaced by an updated version. the old one will be deleted.
	X3FOUNDATION_PRIVATE_API void removePointsFromPointCloud(UnstructuredInCorePointCloud* upc, const std::vector<mpcard>& points, InCorePCTopologyGraph* tpg = NULL, InCorePCTopologyGraph** newTpg = NULL);
	X3FOUNDATION_PRIVATE_API void keepPointsInPointCloud(UnstructuredInCorePointCloud* upc, const std::vector<mpcard>& points, InCorePCTopologyGraph* tpg = NULL, InCorePCTopologyGraph** newTpg = NULL);
	X3FOUNDATION_PRIVATE_API std::vector<mpcard> removeSmallPatches(UnstructuredInCorePointCloud* geometry, InCorePCTopologyGraph* tpg, InCorePCTopologyGraph** newTpg, int minNumberOfPoints);
	X3FOUNDATION_PRIVATE_API std::vector<mpcard> removeUnconnectedPoints(UnstructuredInCorePointCloud* geometry, InCorePCTopologyGraph* tpg, InCorePCTopologyGraph** newTpg);
	X3FOUNDATION_PRIVATE_API UnstructuredInCorePointCloud* sampleFromPointCloud(UnstructuredInCorePointCloud* upc, const std::vector<mpcard>& points, InCorePCTopologyGraph* tpg = NULL, InCorePCTopologyGraph** newTpg = NULL);

	// return connected components
	X3FOUNDATION_PRIVATE_API unsigned findConnectedComponents(UnstructuredInCorePointCloud* upc, InCorePCTopologyGraph* tpg, std::vector<std::vector<mpcard>>& components);

	class X3FOUNDATION_PRIVATE_API Simple3DHashedGrid
	{
	public:
		Simple3DHashedGrid(float gridSize) {m_GridSize = gridSize;m_InvGridSize = 1.0f / m_GridSize;}

		bool entryExists(const Vector3f &p) const;

		/// returns true if cell entry already exists (or inserts cell and returns false)
		bool insert(const Vector3f &p);

		inline Vector3i computeCellIndex(const Vector3f &p) const;

	private:
		
		typedef StaticVector<int32, 3> KeyType;

		struct KeyComp {
			static bool equals(KeyType pos1, KeyType pos2) { return pos1 == pos2; }
		};

		class SparseGridHashTable : public HashTable<KeyType, card32, KeyComp> {
		protected:
			virtual max_platform_card hashFunction(KeyType key) const {
				return (key[0]+521) + 1031*((key[1]+521) + 1031*(key[2]+521));
			};
		public:
		};

		SparseGridHashTable		m_Grid;
		float32					m_GridSize;
		float32					m_InvGridSize;
	};

	X3FOUNDATION_PRIVATE_API void resamplePointCloud_Grid(UnstructuredInCorePointCloud * upc, float cellSize);
}

#endif
