//---------------------------------------------------------------------------
#include "stdafx.h"
//---------------------------------------------------------------------------
#include "GridBuilder.h"
#include "ClassMethods.h"
#include "ObjectClassProperty.h"
#include "SeparatorClassProperty.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

IMPLEMENT_X4_CLASS( GridBuilder, 1 )
{
	BEGIN_CLASS_INIT( GridBuilder )
	ADD_FLOAT32_PROP( maxEigenValueRatio, 0 )
	ADD_FLOAT32_PROP( minPointsPerCellScale, 0 )
	ADD_VECTOR3I_PROP( maxEigenValueFactors, 1 )
}

GridBuilder::GridBuilder()
{
	maxEigenValueRatio = 0.3f;
	minPointsPerCell = 20;
	minPointsPerCellScale = 0.5f;
	maxEigenValueFactors = makeVector3f( 0.5f, 0.3f, 0.15f );
}

GridBuilder::~GridBuilder()
{
}

void GridBuilder::fillGrid( AbstractGrid* grid, PointSet* points, const Matrix3f& R, const Vector3f& t )
{
	AAT POSITION = points->getAAT( "position" );
	long numCells = grid->getNumCells();
	std::vector< std::vector<card32> > indices; indices.resize( numCells );
	std::vector<Vector3f> means; means.resize( numCells );
	
	// Reset
	for ( long cellIndex = 0; cellIndex < numCells; cellIndex++ ) {
		indices[cellIndex].clear();
		means[cellIndex] = NULL_VECTOR3F;
	}
	// Sort points into cells
	for( unsigned int pointIndex = 0; pointIndex < points->getNumEntries(); pointIndex++ ){
		Vector3f pos = R * points->get3f( pointIndex, POSITION ) + t;
		pos = ( pos - grid->getNormalizationT() ) * grid->getNormalizationS();
		Vector3i index3D = grid->getIndex3D( pos );
		long index1D = grid->getIndex1D( index3D );
		if ( index1D >= 0 ){
			means[index1D] = ( means[index1D] * indices[index1D].size() + pos ) / ( indices[index1D].size() + 1 );
			indices[index1D].push_back( pointIndex );
		} else {
			//debugOutput << "Should nor happen!\n"; debugOutput.flush();
		}
	}

	float32 avgNumPointsPerCell = 0;
	card32 numContributers = 0;
	for ( long index1D = 0; index1D < numCells; index1D++ ) {
		if ( indices[index1D].size() > 5 ){
			numContributers++;
			avgNumPointsPerCell += indices[index1D].size();
		}
	}
	minPointsPerCell = card32( avgNumPointsPerCell / float32( numContributers ) );
	//debugOutput << "minPointsPerCell set to: " << minPointsPerCell << ".\n"; debugOutput.flush();

	// Create cell if #points in cell is sufficient
	float32 maxEigenValue = 0;
	for ( long index1D = 0; index1D < numCells; index1D++ ) {
		if ( indices[index1D].size() > minPointsPerCell ){
			Matrix3f sigma;
			for ( unsigned int i = 0; i < 3; i++ ){
				for ( unsigned int j = 0; j < 3; j++ ){
					sigma[i][j] = 0;
				}
			}
			for ( unsigned int m = 0; m < indices[index1D].size(); m++ ){
				Vector3f pos = R * points->get3f( indices[index1D][m], POSITION ) + t;
				pos = ( pos - grid->getNormalizationT() ) * grid->getNormalizationS();
				Vector3f d = pos - means[index1D];
				sigma[0][0] += d[0] * d[0];
				sigma[0][1] += d[0] * d[1];
				sigma[0][2] += d[0] * d[2];
				sigma[1][0] += d[0] * d[1];
				sigma[1][1] += d[1] * d[1];
				sigma[1][2] += d[1] * d[2];
				sigma[2][0] += d[0] * d[2];
				sigma[2][1] += d[1] * d[2];
				sigma[2][2] += d[2] * d[2];
			}

			// Position
			Vector3f eigenValues;
			Matrix3f eigenVectors;
			sigma.computeEigenStructure( eigenValues, eigenVectors );

			// Discard tube-like distributions
			if ( eigenValues[1] / eigenValues[0] > maxEigenValueRatio ){
				if ( eigenValues[0] > maxEigenValue ) maxEigenValue = eigenValues[0];

				Matrix3f inv_singular = IDENTITY3F;
				inv_singular[0][0] = 1.0f / eigenValues[0];
				inv_singular[1][1] = 1.0f / eigenValues[1];
				inv_singular[2][2] = 1.0f / eigenValues[2];

				Vector3i index3D = grid->getIndex3D( index1D );
				NDTCell* cell = new NDTCell();
				cell->setMean( means[index1D] );
				cell->setInvSigma( eigenVectors.transpose() * inv_singular * eigenVectors );
				cell->setEigenValues( eigenValues );
				cell->setEigenVectors( eigenVectors );
				cell->setBBox( grid->computeCellBBox( index3D ) );
				cell->setIndex3D( index3D );
				cell->setValid( true );
				grid->setCell( index1D, cell );
			}
		}
	}

	for( long filledCellIndex = 0; filledCellIndex < grid->getNumFilledCells(); filledCellIndex++ ){
		NDTCell* cell = (NDTCell*) grid->getFilledCell( filledCellIndex );
		Vector3f eigenValues = cell->getEigenValues();
		Matrix3f eigenVectors = cell->getEigenVectors();
		if ( ( eigenValues[0] < maxEigenValue * maxEigenValueFactors[0] ) ||
			 ( eigenValues[1] < maxEigenValue * maxEigenValueFactors[1] ) || 
			 ( eigenValues[2] < maxEigenValue * maxEigenValueFactors[2] ) ){
			eigenValues = makeVector3f( 
				max( eigenValues[0], maxEigenValue * maxEigenValueFactors[0] ),
				max( eigenValues[1], maxEigenValue * maxEigenValueFactors[1] ),
				max( eigenValues[2], maxEigenValue * maxEigenValueFactors[2] )
				);
			cell->setEigenValues( eigenValues );
			Matrix3f inv_singular = IDENTITY3F;
			inv_singular[0][0] = 1.0f / eigenValues[0];
			inv_singular[1][1] = 1.0f / eigenValues[1];
			inv_singular[2][2] = 1.0f / eigenValues[2];
			cell->setInvSigma( eigenVectors.transpose() * inv_singular * eigenVectors );
		}
	}
}