#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "HDRLightprobeSampling.h"
//---------------------------------------------------------------------------
#include <queue>
#include "CopyObjectProperties.h"
#include "BooleanClassProperty.h"
#include "ObjectClassProperty.h"
#include "Random.h"
#include <fstream>
//---------------------------------------------------------------------------
using namespace X4;

IMPLEMENT_X4_CLASS( HDRLightprobeSampling ,1)
{
	BEGIN_CLASS_INIT( HDRLightprobeSampling );
	ADD_BOOLEAN_PROP( m_HalfSphere, 0);
	ADD_CARD32_PROP( m_NumSamples, 0 );
	ADD_OBJECT_PROP( m_LightImage, 0, RGBAFloat32Image::getClass(), true );
	ADD_FLOAT32_PROP( m_InterpToGray , 1);
}

HDRLightprobeSampling::HDRLightprobeSampling(void)
{
	m_LightImage = new RGBAFloat32Image(512,96*2);
	m_LightImage->clear( makeVector4f(1,1,1,1));
	m_SummedAreaTable = NULL;
	m_HalfSphere = true;
	m_NumSamples = 100;
	m_Version = 0;
	m_InterpToGray = 0.5f;
}

HDRLightprobeSampling::~HDRLightprobeSampling(void)
{
	delete m_LightImage;
	delete m_SummedAreaTable;
}

void HDRLightprobeSampling::assign(const Object* obj, X4CopyContext *context)
{
	const HDRLightprobeSampling * o = dynamic_cast<const HDRLightprobeSampling*>(obj);
	x4Assert( o != NULL );

	copyObjectProperties( obj, this );
	delete m_SummedAreaTable;
	m_SummedAreaTable = NULL;
	
	m_Version = (card32)rand();
	setup();
}

void HDRLightprobeSampling::setupSamples( card32 numSamples ) 
{
	m_NumSamples = numSamples;
	RGBAFloat32Image * weightedImage = new RGBAFloat32Image( m_LightImage->getSizeX(),
		m_LightImage->getSizeY() / (m_HalfSphere ? 2.0f:1.0f) );

	for(card32 y=0;y<weightedImage->getSizeY();y++ )
	{
		float weight;
		if( m_HalfSphere )
			weight = cos((1.0f - (float)y / (float)weightedImage->getSizeY()) * M_PI / 2.0f );
		else
			weight = cos((1.0f - 2.0f * (float)y / (float)weightedImage->getSizeY()) * M_PI / 2.0f );
		for( card32 x=0;x<weightedImage->getSizeX();x++ )
		{
			weightedImage->setPixel(x,y, m_LightImage->getPixel(x,y) * weight );
		}
	}

	delete m_SummedAreaTable;
	m_SummedAreaTable = new HDRSummedAreaTable( weightedImage,0,0, weightedImage->getSizeX(),
		weightedImage->getSizeY() );

	MedianCutLightNode rootNode;
	rootNode.m_SummedEnergy = 1.0f;
	rootNode.m_X1 = 0;
	rootNode.m_Y1 = 0;
	rootNode.m_X2 = weightedImage->getSizeX();
	rootNode.m_Y2 = weightedImage->getSizeY();
	list<MedianCutLightNode> queue;
	std::vector<MedianCutLightNode> finishedNodes;
	queue.push_back(rootNode);


	while( queue.size() + finishedNodes.size() != numSamples )
	{
		MedianCutLightNode node = queue.front();
		queue.pop_front();

		if( node.m_X1 + 1 == node.m_X2 && node.m_Y1 + 1 == node.m_Y2 )
		{
			finishedNodes.push_back( node );
			continue;
		}
		
		int w = node.m_X2 - node.m_X1;
		int h = node.m_Y2 - node.m_Y1;

		MedianCutLightNode node1, node2;
		node1 = node;
		node2 = node;
		int stepSize;
		if( w > h )
		{
			node1.m_X2 = node.m_X1 + w/2;
			node2.m_X1 = node1.m_X2;
			stepSize = node1.m_X2 - node1.m_X1;
		}
		else
		{
			node1.m_Y2 = node.m_Y1 + h/2;
			node2.m_Y1 = node1.m_Y2;
			stepSize = node1.m_Y2 - node1.m_Y1;
		}

		stepSize /= 2;

		node1.m_SummedEnergy = m_SummedAreaTable->getEnergy(node1.m_X1, node1.m_Y1, node1.m_X2, node1.m_Y2);
		node2.m_SummedEnergy = m_SummedAreaTable->getEnergy(node2.m_X1, node2.m_Y1, node2.m_X2, node2.m_Y2);
		while( stepSize > 0 )
		{
			if( node1.m_SummedEnergy < node2.m_SummedEnergy )
			{
				if( w > h )
				{
					node1.m_X2 = node1.m_X2 + stepSize;
					node2.m_X1 = node1.m_X2;
				}
				else
				{
					node1.m_Y2 = node1.m_Y2 + stepSize;
					node2.m_Y1 = node1.m_Y2;
				}
			}
			else
			{
				if( w > h )
				{
					node1.m_X2 = node1.m_X2 - stepSize;
					node2.m_X1 = node1.m_X2;
				}
				else
				{
					node1.m_Y2 = node1.m_Y2 - stepSize;
					node2.m_Y1 = node1.m_Y2;
				}
			}

			node1.m_SummedEnergy = m_SummedAreaTable->getEnergy(node1.m_X1, node1.m_Y1, node1.m_X2, node1.m_Y2);
			node2.m_SummedEnergy = m_SummedAreaTable->getEnergy(node2.m_X1, node2.m_Y1, node2.m_X2, node2.m_Y2);

			stepSize /= 2;
		}

		queue.push_back( node1 );
		queue.push_back( node2 );

	}

	while( !queue.empty () )
	{
		finishedNodes.push_back( queue.front() );
		queue.pop_front();
	}

	m_Samples.clear();
	for( card32 i=0;i<finishedNodes.size();i++ )
	{
		MedianCutLightNode node = finishedNodes[i];
		// calculate rgb energy and centroid
		TLightSample sample;
		sample.m_PixelPosition = NULL_VECTOR2F;
		sample.m_Color = NULL_VECTOR3F;

		float summedEnergy = 0;

		for( int x=node.m_X1;x<node.m_X2;x++ )
		{
			for( int y=node.m_Y1;y<node.m_Y2;y++ )
			{
				Vector4f col = weightedImage->getPixel(x,y);
				float energy = 0.2125f * col[0] + 0.7154f * col[1]+0.0721f * col[2];
				sample.m_Color += shrink4To3(col);
				sample.m_PixelPosition += makeVector2f(x,y) * energy;
				summedEnergy += energy;
			}
		}
		//sample.m_PixelPosition /= max(1e-10f,summedEnergy);
		int xsize = node.m_X2 - node.m_X1;
		int ysize = node.m_Y2 - node.m_Y1;
		sample.m_PixelPosition[0] = (node.m_X1 + node.m_X2) / 2.0f + rnd01()*xsize - (float)xsize*0.5f;
		sample.m_PixelPosition[1] = (node.m_Y1 + node.m_Y2) / 2.0f + rnd01()*ysize - (float)ysize*0.5f;
		//sample.m_PixelPosition = makeVector2f(node.m_X1+node.m_X2,node.m_Y1+node.m_Y2) * 0.5f;
		sample.m_Bounds.lowerCorner =  makeVector2f(node.m_X1,node.m_Y1);
		sample.m_Bounds.upperCorner =  makeVector2f(node.m_X2,node.m_Y2);
		float phi, theta = sample.m_PixelPosition[0] / (float)weightedImage->getSizeX() * 2.0f * M_PI;
		if( m_HalfSphere )
			phi = sample.m_PixelPosition[1] / (float)weightedImage->getSizeY() * M_PI / 2.0f;
		else
			phi = sample.m_PixelPosition[1] / (float)weightedImage->getSizeY() * M_PI;

		sample.m_Position = makeVector3f( sin(theta)*sin(phi), cos(theta)*sin(phi), cos(phi));
		sample.m_Color *= 1.0f / m_SummedAreaTable->m_SummedEnergy * 4.0f / 3.0f * M_PI * M_PI * 0.15f;
		float grayValue = makeVector3f(0.3f,0.59f, 0.11f) * sample.m_Color;
		sample.m_Color = sample.m_Color * (1.0f - m_InterpToGray) + makeVector3f(1,1,1)*grayValue*m_InterpToGray;
		m_Samples.push_back( sample );
	}

	delete weightedImage;
	m_Version++;
}

HDRSummedAreaTable::HDRSummedAreaTable(RGBAFloat32Image * image, int x1, int y1, int x2, int y2)
{
	m_ChildA = NULL;
	m_ChildB = NULL;
	m_X1 = x1;
	m_Y1 = y1;
	m_X2 = x2;
	m_Y2 = y2;
	if( x1+1 == x2 && y1+1 == y2 )
	{
		Vector4f col = image->getPixel(x1,y1);
		m_SummedEnergy = 0.2125f * col[0] + 0.7154f * col[1]+0.0721f * col[2];
	}
	else
	{
		int w = x2-x1;
		int h = y2-y1;
		if( w > h )
		{
			m_ChildA = new	HDRSummedAreaTable( image, x1, y1, x1+w/2, y2 );
			m_ChildB = new	HDRSummedAreaTable( image, x1+w/2, y1, x2, y2 );
		}
		else
		{
			m_ChildA = new	HDRSummedAreaTable( image, x1, y1, x2, y1+h/2 );
			m_ChildB = new	HDRSummedAreaTable( image, x1, y1+h/2, x2, y2 );
		}
		m_SummedEnergy = m_ChildA->m_SummedEnergy + m_ChildB->m_SummedEnergy;
	}
}

HDRSummedAreaTable::~HDRSummedAreaTable()
{
	delete m_ChildA;
	delete m_ChildB;
}

float HDRSummedAreaTable::getEnergy( int x1, int y1, int x2, int y2 )
{
	if( x1 <= m_X1 && y1 <= m_Y1 && x2 >= m_X2 && y2 >= m_Y2 )
		return m_SummedEnergy;
	else
		if( x1 >= m_X2 ||y1 >= m_Y2 || x2 <= m_X1 || y2 <= m_Y1 )
			return 0;
		else
		{
			return m_ChildA->getEnergy(x1,y1,x2,y2) + m_ChildB->getEnergy(x1,y1,x2,y2);
		}
}

MedianCutLightNode::MedianCutLightNode()
{
}

MedianCutLightNode::~MedianCutLightNode()
{
}

void HDRLightprobeSampling::setImage( RGBAFloat32Image * image ) 
{
	delete m_LightImage;
	m_LightImage = image;
}

void X4::HDRLightprobeSampling::resetWhiteImage()
{
	delete m_LightImage;
	m_LightImage = new RGBAFloat32Image(256,96);
	m_LightImage->clear( makeVector4f(1,1,1,1));
}

void X4::HDRLightprobeSampling::read( InputObjectStream *s )
{
	Persistent::read(s);
	setup();
}

bool X4::HDRLightprobeSampling::loadFromPFM( const char * str )
{
	fstream input( str );
	if( !input.is_open() )
		return false;

	std::string line;
	getline( input, line );
	int width,height;
	float aspect;
	input >> width;
	input >> height;
	input >> aspect;
	input.close();

	std::ifstream bin_input( str, ios::binary );
	bin_input.seekg( 0, ios::end );
	int fSize = bin_input.tellg();
	int newPos = fSize - width*height*4*3;
	bin_input.seekg( newPos, ios::beg );

	RGBAFloat32Image * ret = new RGBAFloat32Image;
	ret->resize(width,height);
	std::vector<char> data(width * height *4 * 3);
	char * ptr = &data[0];
	bin_input.read( (char*)ptr, width * height *4 * 3);
	bin_input.close();

	// big/little endian?
	if( aspect > 0 )
	{
		for( int i=0;i<width * height;i++ )
		{
			char tmp = ptr[0];
			ptr[0] = ptr[3];
			ptr[3] = tmp;
			tmp = ptr[1];
			ptr[1] = ptr[2];
			ptr[2] = tmp;

			ptr += 4;
		}
	}

	Vector4f * retPtr = ret->getPixels();
	Vector3f * srcPtr = (Vector3f*)&data[0];
	for( card32 x=0;x<ret->getSizeX();x++ )
		for( card32 y=0;y<ret->getSizeY();y++ )
		{
			Vector3f col = srcPtr[(ret->getSizeY()-1-y)*ret->getSizeX() + x ];
			retPtr[y*ret->getSizeX()+x] = makeVector4f( col[0],col[1],col[2],1.0f);
		}

	setImage(ret);
	return true;
}