#include "stdafx.h"
//---------------------------------------------------------------------------
#include "HoppeMesher.h"
#include "DebugOutput.h"
#include "HierarchicalKNNIterator.h"
#include "ProgressWindow.h"
#include "PointSet.inline.h"
#include <time.h>
#include "NumericalClassProperty.h"
#include "PointSetKNNQuery.h"
#include "MarchingCubes3D.h"
#include "MathTools.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

float32 HoppeMesher::eval( 
	PointSet* ps, 
	const Vector3f& x, 
	PointSetKNNQuery& knn, 
	const float32 tau,
	const float32 epsilon,
	const card32 distanceMode )
{
	AAT POSITION = ps->getAAT( "position" );
	AAT NORMAL = ps->getAAT( "normal" );
	
	knn.setTargetPoint( x );
	bool found;
	card32 nIndex;
	knn.getNextNearestPointIndex( nIndex, found );
	const Vector3f& p = ps->get3f( nIndex, POSITION );
	if ( norm ( p - x ) > tau * 3 ) return -1;

	if ( distanceMode == DISTANCE_CLOSEST_POINT ){
		const Vector3f& normal = ps->get3f( nIndex, NORMAL );
		return p * normal - x * normal;
	} else if ( distanceMode == DISTANCE_EPSILON ){
		float32 summedWeight = 0;
		const float32 sqrEpsilon = sqr( epsilon );
		bool converged = false;
		float32 value = 0;
		while( !converged && found ){
			const Vector3f& np = ps->get3f( nIndex, POSITION );
			const Vector3f nNormal = ps->get3f( nIndex, NORMAL );
			const float32 sqrD = ( np - p ) * ( np - p ); 
			const float32 weight = exp( -sqrD / sqrEpsilon );
			summedWeight += weight;
			value += ( np * nNormal - x * nNormal ) * weight;
			if ( sqrD > sqrEpsilon ) converged = true;
			knn.getNextNearestPointIndex( nIndex, found );
		}
		return value / summedWeight;
	}

	return 0;
}


UnstructuredInCoreTriangleMesh* HoppeMesher::triangulate( 
	PointSet* ps, 
	const card32 res, 
	const float32 epsilon,
	const card32 distanceMode )
{
	if ( ps == NULL ){
		debugOutput << "triangulate: invalid point set.\n"; debugOutput.flush();
		return NULL;
	}

	if ( !ps->providesAttribute( "position" ) || !ps->providesAttribute( "normal" )){
		debugOutput << "triangulate: channel missing.\n"; debugOutput.flush();
		return NULL;
	}

	const float32 startTime = clock();

	AAT POSITION = ps->getAAT( "position" );

	AttachedIndexedOctree* octree = PointSetKNNQuery::createOctree( ps, 32 );
	PointSetKNNQuery knn( ps, octree, makeVector3f( 0, 0, 0 ), POSITION );

	// Grid subdivision
	BoundingBox3f bbox = ps->getBoundingBox();
	bbox.enlarge( 1.05f );
	const float32 d = bbox.getMaxSideLength()/2.0f;
	const Vector3f center = bbox.getCenter();
	bbox.lowerCorner = center - makeVector3f( d, d, d );
	bbox.upperCorner = center + makeVector3f( d, d, d );
	const float32 tau = ( d * 2.0f ) / float32( res ); 
	Vector3f corners[8];
	float32 values[8];
	std::vector<Vector3f> verts;

	std::vector<Vector3f> vertices; vertices.clear();

	progressWindow->pushStep( true, "Hoppe Mesher" );
	for ( unsigned int k = 0; k < res; k++ ){
		for ( unsigned int j = 0; j < res; j++ ){
			for ( unsigned int i = 0; i < res; i++ ){
				corners[0] = bbox.lowerCorner + makeVector3f( i*tau,     j*tau,     k*tau );
				corners[1] = bbox.lowerCorner + makeVector3f( (i+1)*tau, j*tau,     k*tau );
				corners[2] = bbox.lowerCorner + makeVector3f( (i+1)*tau, (j+1)*tau, k*tau );
				corners[3] = bbox.lowerCorner + makeVector3f( i*tau,     (j+1)*tau, k*tau );
				corners[4] = bbox.lowerCorner + makeVector3f( i*tau,     j*tau,     (k+1)*tau );
				corners[5] = bbox.lowerCorner + makeVector3f( (i+1)*tau, j*tau,     (k+1)*tau );
				corners[6] = bbox.lowerCorner + makeVector3f( (i+1)*tau, (j+1)*tau, (k+1)*tau );
				corners[7] = bbox.lowerCorner + makeVector3f( i*tau,     (j+1)*tau, (k+1)*tau );
				for ( unsigned int n = 0; n < 8; n++ ){
					values[n] = eval( ps, corners[n], knn, tau, epsilon, distanceMode );
				}
				MarchingCubes3D::triangulate( corners, values, 0, verts );
				for ( unsigned int n = 0; n < verts.size(); n++ )
					vertices.push_back( verts[n] );
			}
		}	
		progressWindow->progress( float32( k * 100 ) / float( res ) );	
	}

	// Create mesh
	UnstructuredInCoreTriangleMesh* mesh = new UnstructuredInCoreTriangleMesh();
	mesh->setMaterialIndex( 3 );
	VertexDescriptor vdesc;
	vdesc.pushAttrib( VAD( "position", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	vdesc.pushAttrib( VAD( "normal", 3, VAD::DATA_FORMAT_FLOAT32 ) );
	VertexDescriptor idesc;
	idesc.pushAttrib( VAD( "index", 3, VAD::DATA_FORMAT_INT32 ) );
	mesh->clearAndSetup( &vdesc, vertices.size(), &idesc, vertices.size()/3 );
	AAT VERTEX_POSITION = mesh->getPointSet()->getAAT( "position" );
	AAT TRIANGLE_INDICES = mesh->getTriangles()->getAAT( "index" );
	// Set mesh content
	for( card32 i=0; i < vertices.size(); i++ ){
		mesh->getPointSet()->set3f( i, VERTEX_POSITION, vertices[i] );
	}
	for( card32 i=0; i < vertices.size()/3; i++ ){
		mesh->getTriangles()->set3i( i, TRIANGLE_INDICES, makeVector3i( i*3, i*3+1, i*3+2 ) );
	}
	// Finalize mesh
	mesh->snapMesh( tau * 0.1f );
	mesh->closeMesh();
	mesh->calculateVertexNormals();

	debugOutput << "Time required (Hoppe Mesher): " << float32( ( clock() - startTime ) / CLK_TCK ) << "\n"; debugOutput.flush(); 
	progressWindow->popSteps();
	delete octree;

	return mesh;
}

UnstructuredInCoreTriangleMesh* HoppeMesher::triangulate( 
	PointCloud* pc, 
	const card32 res, 
	const float32 epsilon,	
	const card32 distanceMode )
{
	debugOutput << "not implemented yet\n"; debugOutput.flush();
	return NULL;
}