//---------------------------------------------------------------------------
#include "stdafx.h"
//---------------------------------------------------------------------------
#include "BoundingBox.h"
#include "ClassMethods.h"
#include "ObjectClassProperty.h"
#include "SeparatorClassProperty.h"
#include "ICP.h"
#include <time.h>
#include "FixedArrayClassProperty.h"
#include "NumericalClassProperty.h"
#include "DebugOutput.h"
#include "linearalgebra.inline.h"
#include "PointSet.inline.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

IMPLEMENT_X4_CLASS( ICP, 3 )
{
	BEGIN_CLASS_INIT( ICP )
	ADD_SEPARATOR( "ICP" )
	ADD_VECTOR3F_PROP( normalizationT, 3 )
	ADD_FLOAT32_PROP( normalizationS, 3 )
}

ICP::ICP()
{
	subsampling = 5;
	maxDistance = -1.0f;
	candidatesRatio = 0.5f;
	updateCandidatesInterval = 9;
	stepSize = 0.25f;
	numSolves = 0;
	T->setMode( TRANSFORMATION_EULER_ANGLES );
	solver = SOLVER_GRADIENT_DESCENT;
	candSelectionMode = CANDIDATE_SELECTION_SUBSAMPLING;
	colorCode = COLOR_CODE_NONE;
	knnBase = NULL;
	octree = NULL;
	ownOctree = false;
	normalizationS = 1.0f;
	normalizationT = NULL_VECTOR3F;
	registerTranslation = true;
	registerRotation = true;
}

ICP::~ICP()
{
	if ( ownOctree && ( octree != NULL ) ) {
		delete octree;
		ownOctree = false; 
	}
	if ( knnBase != NULL ) delete knnBase;
}

void ICP::setup( PointSet* basePS, const Matrix3f& baseR, const Vector3f& baseT )
{
	Registerizer::setup( basePS, baseR, baseT );
	// Compute normalization to [-1;1]^3
	Vector3f rotCenter = makeVector3f( 0,0,0 );
	BoundingBox3f bbox = BoundingBox3f( basePS->get3f( 0, BASE_POSITION ) ); 
	for ( unsigned int pointIndex = 0; pointIndex < basePS->getNumEntries(); pointIndex++ ){
		Vector3f pos = baseR * basePS->get3f( pointIndex, BASE_POSITION ) + baseT;
		rotCenter = ( rotCenter * pointIndex + pos ) / float32( pointIndex + 1 );
		bbox.addPoint( pos );
	}
	normalizationT = rotCenter;
	normalizationS = 2.0f / bbox.getMaxSideLength();
	if ( octree != NULL ) delete octree;
	octree = new ICPOctree();
	ownOctree = true;
	octree->octreesetup( basePS, 32, baseR, baseT );
	if ( knnBase != NULL ) delete knnBase;
	knnBase = new ICPKNNQuery( basePS, octree, NULL_VECTOR3F, BASE_POSITION );
}


void ICP::setupWithOctree( PointSet* basePS, ICPOctree* octree )
{
	Registerizer::setup( basePS, octree->getRotation(), octree->getTranslation() );
	// Compute normalization to [-1;1]^3
	Vector3f rotCenter = makeVector3f( 0,0,0 );
	BoundingBox3f bbox = BoundingBox3f( basePS->get3f( 0, BASE_POSITION ) ); 
	for ( unsigned int pointIndex = 0; pointIndex < basePS->getNumEntries(); pointIndex++ ){
		Vector3f pos = baseR * basePS->get3f( pointIndex, BASE_POSITION ) + baseT;
		rotCenter = ( rotCenter * pointIndex + pos ) / float32( pointIndex + 1 );
		bbox.addPoint( pos );
	}
	normalizationT = rotCenter;
	normalizationS = 2.0f / bbox.getMaxSideLength();
	if ( ownOctree && ( this->octree != NULL ) ) {
		delete this->octree;
		ownOctree = false;
	}
	this->octree = octree;
	if ( knnBase != NULL ) delete knnBase;
	knnBase = new ICPKNNQuery( basePS, octree, NULL_VECTOR3F, BASE_POSITION );
}

float32 ICP::computeGradient( StaticVector<float32,7>& gradient )
{
	for ( unsigned int n = 0; n < 7; n++ ) { 
		gradient[n] = 0; 
	}

	if ( !ready ){
		debugOutput << "computeGradient: call setup first.\n"; debugOutput.flush();
		return -1;
	}

	StaticMatrix<float32,7,3> jacobian;
	card32 nIndex; 
	bool found;
	
	float32 error = 0;
	Vector3f c = normalizationT;
	float32 s = normalizationS;

	card32 numPointsUsed = 0;
	for ( unsigned int pointIndex = 0; pointIndex < registerPS->getNumEntries(); pointIndex++ ){
		if ( colorCode == COLOR_CODE_USED_POINTS ) registerPS->set3f( pointIndex, REGISTER_COLOR, makeVector3f( 0.85f, 0.15f, 0.15f ) );
		if ( registerPS->get1i( pointIndex, REGISTER_FLAG ) & FLAG_REGISTER_CANDIDATE ){
			Vector3f r = registerR * registerPS->get3f( pointIndex, REGISTER_POSITION ) + registerT;
			knnBase->setTargetPoint( r );
			knnBase->getNextNearestPointIndex( nIndex, found );
			if ( found ){
				Vector3f b = baseR * basePS->get3f( nIndex, BASE_POSITION ) + baseT;
				// Transform points to local system [-1;1]^3
				r = ( r - c ) * s;
				b = ( b - c ) * s;
				float32 d = norm( r - b );
				error += d;
				T->jacobian( jacobian, r );
				for ( unsigned int i = 0; i < 7; i++ ){
					gradient[i] -= ( r - b ) * jacobian[i] * 2.0f;
				}
				numPointsUsed++;
				if ( colorCode == COLOR_CODE_USED_POINTS ) registerPS->set3f( pointIndex, REGISTER_COLOR, makeVector3f( 0.15f, 0.85f, 0.15f ) );
			}
		}
	}
	gradient = gradient / float( numPointsUsed );
	return error;
}
		
float32 ICP::computeHessian( StaticMatrix<float32,7,7>& hessian, StaticVector<float32,7>& gradient )
{
	debugOutput << "computeHessian: not implemented yet.\n"; debugOutput.flush();
	return -1;
}

void ICP::selectRegisterCandidates()
{
	if ( ( basePS == NULL ) || ( registerPS == NULL ) ){
		debugOutput << "selectRegisterCandidates: invalid point sets.\n"; debugOutput.flush();
		return;
	}
	if ( candSelectionMode == CANDIDATE_SELECTION_CLOSEST ){
		if ( knnBase == NULL ){
			debugOutput << "selectRegisterCandidates: knn not initialized.\n"; debugOutput.flush();
			return;
		}
		// Select closest candidates
		std::vector<Candidate> candidates;
		bool found;
		card32 nIndex;
		// Get candidates
		for ( unsigned int pointIndex = 0; pointIndex < registerPS->getNumEntries(); pointIndex++ ){
			Vector3f r = registerR * registerPS->get3f( pointIndex, REGISTER_POSITION ) + registerT;
			knnBase->setTargetPoint( r );
			knnBase->getNextNearestPointIndex( nIndex, found );
			Vector3f b = baseR * basePS->get3f( nIndex, BASE_POSITION ) + baseT;
			Candidate candidate;
			candidate.dist = ( r - b ).getSqrNorm();
			candidate.index = pointIndex;
			candidates.push_back( candidate );
			// Init all point with not candidate flag
			registerPS->set1i( pointIndex, REGISTER_FLAG, registerPS->get1i( pointIndex, REGISTER_FLAG ) & (0xFFFFFFFF ^FLAG_REGISTER_CANDIDATE) );
		}
		std::sort( candidates.begin(), candidates.end(), CandidateSort() );
		// Mark candidates
		for ( unsigned int n = 0; n < candidatesRatio * candidates.size(); n++ ){
			card32 pointIndex = candidates[n].index;
			registerPS->set1i( pointIndex, REGISTER_FLAG, registerPS->get1i( pointIndex, REGISTER_FLAG ) | FLAG_REGISTER_CANDIDATE );
		}
	} else if ( candSelectionMode == CANDIDATE_SELECTION_MAXDIST ){
		// Select closest candidates
		bool found;
		card32 nIndex;
		// Get candidates
		for ( unsigned int pointIndex = 0; pointIndex < registerPS->getNumEntries(); pointIndex++ ){
			Vector3f r = registerR * registerPS->get3f( pointIndex, REGISTER_POSITION ) + registerT;
			knnBase->setTargetPoint( r );
			knnBase->getNextNearestPointIndex( nIndex, found );
			Vector3f b = baseR * basePS->get3f( nIndex, BASE_POSITION ) + baseT;
			float32 d = norm( r - b );
			if ( ( d < maxDistance ) || ( maxDistance < 0 ) ){
				registerPS->set1i( pointIndex, REGISTER_FLAG, registerPS->get1i( pointIndex, REGISTER_FLAG ) | FLAG_REGISTER_CANDIDATE );
			} else {
				registerPS->set1i( pointIndex, REGISTER_FLAG, registerPS->get1i( pointIndex, REGISTER_FLAG ) & (0xFFFFFFFF ^FLAG_REGISTER_CANDIDATE) );
			}
		} 
	} else if ( candSelectionMode == CANDIDATE_SELECTION_CLOSEST_SUBSAMPLING ){
		if ( knnBase == NULL ){
			debugOutput << "selectRegisterCandidates: knn not initialized.\n"; debugOutput.flush();
			return;
		}
		// Select closest candidates
		std::vector<Candidate> candidates;
		bool found;
		card32 nIndex;
		// Get candidates
		for ( unsigned int pointIndex = 0; pointIndex < registerPS->getNumEntries(); pointIndex++ ){
			Vector3f r = registerR * registerPS->get3f( pointIndex, REGISTER_POSITION ) + registerT;
			knnBase->setTargetPoint( r );
			knnBase->getNextNearestPointIndex( nIndex, found );
			Vector3f b = baseR * basePS->get3f( nIndex, BASE_POSITION ) + baseT;
			Candidate candidate;
			candidate.dist = ( r - b ).getSqrNorm();
			candidate.index = pointIndex;
			candidates.push_back( candidate );
			// Init all point with not candidate flag
			registerPS->set1i( pointIndex, REGISTER_FLAG, registerPS->get1i( pointIndex, REGISTER_FLAG ) & (0xFFFFFFFF ^FLAG_REGISTER_CANDIDATE) );
		}
		std::sort( candidates.begin(), candidates.end(), CandidateSort() );
		// Mark candidates
		for ( unsigned int n = 0; n < candidatesRatio * candidates.size() / float32( subsampling ); n++ ){
			if ( rand()%subsampling == 0 ){
				card32 pointIndex = candidates[n].index;
				registerPS->set1i( pointIndex, REGISTER_FLAG, registerPS->get1i( pointIndex, REGISTER_FLAG ) | FLAG_REGISTER_CANDIDATE );
			}
		} 
	} else {
		Registerizer::selectRegisterCandidates();
		return;
	}

	// Mark candidates
	if ( colorCode == COLOR_CODE_CANDIDATES ){
		for ( unsigned int pointIndex = 0; pointIndex < registerPS->getNumEntries(); pointIndex++ ){
			int32 flag = registerPS->get1i( pointIndex, REGISTER_FLAG );
			if ( flag & FLAG_REGISTER_CANDIDATE ){
				registerPS->set3f( pointIndex, REGISTER_COLOR, makeVector3f( 0.15f, 0.85f, 0.15f ) );
			} else {
				registerPS->set3f( pointIndex, REGISTER_COLOR, makeVector3f( 0.85f, 0.15f, 0.15f ) );
			}
		}
	}
}

Matrix4f ICP::getTransformation() const
{
	return getInvNormalization() * T->getTransformation() * getNormalization();
}

Matrix4f ICP::getNormalization() const
{
	Matrix4f scale;
	Matrix4f translate;
	for ( unsigned int i = 0; i < 3; i++ ){
		scale[i][i] = normalizationS;
	}
	for ( unsigned int i = 0; i < 3; i++ ){
		translate[3][i] = -normalizationT[i];
	}
	return scale * translate;
}

Matrix4f ICP::getInvNormalization() const
{
	Matrix4f scale;
	Matrix4f translate;
	for ( unsigned int i = 0; i < 3; i++ ){
		scale[i][i] = 1.0f / normalizationS;
	}
	for ( unsigned int i = 0; i < 3; i++ ){
		translate[3][i] = normalizationT[i];
	}
	return translate * scale;
}