//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "ICPKNNQuery.h"
//---------------------------------------------------------------------------
#include "BoundingBox.inline.h"
#include "Persistent.h"
#include "Tools.h"
#include "DebugOutput.h"
#include "PointSet.inline.h"
#include "X4Exception.h"

//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

ICPKNNQuery::ICPKNNQuery( const PointSet *ps, ICPOctree *octree, const Vector3f &targetPoint, const AAT &positionAAT ) 
{
	if (ps == NULL) throw X4Exception("ICPKNNQuery - ps must not be NULL.");
	if (octree == NULL) throw X4Exception("ICPKNNQuery - octree must not be NULL.");
	this->ps = ps;
	this->targetPoint = targetPoint;
	this->octree = octree;
	this->POSITION = positionAAT;
	rootNode = octree->getOctree();
	putBoxOnQueue( rootNode );
}

void ICPKNNQuery::setTargetPoint(const Vector3f& targetPoint) {
   this->targetPoint = targetPoint;

   static const std::priority_queue<ICPPointEntry> ICP_EMPTY_POINTS;
   static const std::priority_queue<ICPBoxEntry> ICP_EMPTY_BOXES;

   pqBoxes = ICP_EMPTY_BOXES;
   pqPoints = ICP_EMPTY_POINTS;
   putBoxOnQueue(rootNode);
}

void ICPKNNQuery::putBoxOnQueue(ICPOctreeNode* box) {
   // calc box-point distance
   float32 qDistance;

   Sphere3f sp=box->getBBox().getBoundingSphere();
   qDistance=(sp.center-targetPoint)*(sp.center-targetPoint);
   if (qDistance<sp.radius*sp.radius) {
      qDistance=-1.0f;
   } else {
      qDistance=sqrt(qDistance);
      qDistance-=sp.radius;
      qDistance*=qDistance;
   }

   ICPBoxEntry bEntry(box, qDistance);
   pqBoxes.push( bEntry );
}

void ICPKNNQuery::putPointOnQueue(const card32 &indexOfPoint) {
   // calc point-point distance
   Vector3f pos = octree->getRotation() * ps->get<float32,3>( indexOfPoint, POSITION ) + octree->getTranslation();
   float32 qDistance = (pos-targetPoint)*(pos-targetPoint);
   ICPPointEntry pnt(indexOfPoint, qDistance);
   pqPoints.push(pnt);
}

void ICPKNNQuery::getNextNearestPointIndex(card32 &nextIndex, bool &found) {
   // check for end of points
   if (pqBoxes.empty() && pqPoints.empty()) {   // no more points
      nextIndex = 0xFFFFFFFF;
      found = false;
      return;
   }
   
	// start search
   while (true) {
      if (!pqPoints.empty()) {
         if (pqBoxes.empty() || pqPoints.top().quDist <= pqBoxes.top().quDist) { // point nearer than any box
            card32 i = pqPoints.top().index;
            pqPoints.pop();
				nextIndex = i;
				found = true;
            return;
         }
      }

      // step into node
      if (pqBoxes.empty()) {found = false; return;}
      ICPOctreeNode *thisBox = pqBoxes.top().node;
      pqBoxes.pop();

      if(thisBox->isLeaf()){  //Put Boxes Points on Queue
         size_t size = thisBox->getPointIndices().size();
         for(card32 i=0; i<size; i++){
            putPointOnQueue( thisBox->getPointIndex( i ) );
         }

      } else { // put child boxes on queue
         for (card32 i=0; i<8; i++){
            putBoxOnQueue(thisBox->getChild(i));
         }
      }
   } //while(true)
}



