//---------------------------------------------------------------------------
#ifndef ImprotanceSamplingH
#define ImprotanceSamplingH
//---------------------------------------------------------------------------
#include "x3foundation_private.h"
#include "Persistent.h"
#include <ObjectListProperty.h>
//#include <random>

//---------------------------------------------------------------------------

namespace X4
{

	/**
	* Class representing a data element which is sampled. Derive your own
	* classes from this one to use importance sampler. Probability value
	* of the sample defines how it is likely that this sample is sampled.
	* The probability value has not to be normalized.
	**/
	class X3FOUNDATION_PRIVATE_API ImportanceSample
	{
	private:
		float m_SampleProbability;
		card32 m_SampleID;

		void* m_UserData;

		friend class ImportanceSampler;

	public:
		ImportanceSample(float p) : m_SampleProbability(p), m_SampleID(0) {}
		ImportanceSample(const ImportanceSample& sample) : 
			m_SampleProbability(sample.m_SampleProbability), m_SampleID(sample.m_SampleID), m_UserData(sample.m_UserData) {}
		virtual ~ImportanceSample() {}

		inline float getProbability() const {return m_SampleProbability;}
		inline void    setProbability(float prob) { m_SampleProbability = prob; }

		inline void* getUserData() const {return m_UserData;}
		inline void setUserData(void* v) {m_UserData=v;}

		inline card32 getSampleID() const {return m_SampleID;}
		inline void setSampleID(card32 v) {m_SampleID=v;}

		//void assign(const Object* obj, X4CopyContext *context = NULL);
	};

	/**
	* Perform weighted sampling. The sampling time complexity is O(logn), where n number of samples
	*/
	class X3FOUNDATION_PRIVATE_API ImportanceSampler
	{
      public:

         ImportanceSampler();
         virtual ~ImportanceSampler();

         /// Sample the database and return the best matched sampler
         ImportanceSample* sample(bool getJustTheBest = false);
         
         /// Add a new sample (sample will not be owned by the sampler !). @return id of the sample
         card32 addSample(ImportanceSample* sample);
         
         /// Remove sample
         void removeSample(card32 id);
         
         /// remove all samples
         inline void clearSamples() { m_Samples.clear(); m_Weights.clear(); }
         
         /// get num of samples
         inline size_t getNumSamples() const { return m_Samples.size(); }
         
         /// get sample based on the index and not on id
         inline ImportanceSample* getSample(card32 index) { return m_Samples[index]; }
         
         /// set sample based on index and not on id
         inline void setSample(ImportanceSample* sample, card32 index) { m_Samples[index] = sample; }
         
		 /// force sampler to recompute weights, because probabilities might have changed
		 void recomputeWeights();

      protected:
		  typedef std::vector<ImportanceSample*> SampleList;
		  std::vector<ImportanceSample*> m_Samples;
		  std::vector<pair<unsigned, float> > m_Weights;
		  ImportanceSample* binarySearch(double val, mpint low, mpint high, double totalWeight);

		  //IMPLEMENT_OBJECT_LIST_METHODS_STL(Samples, ImportanceSample);
   };

};

#endif
