#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "InteractiveHandleSet.h"
//---------------------------------------------------------------------------
#include "GLHelper.h"
#include "RaytracingMath.inline.h"
#include "TriangleBBIntersection.h"
#include "Scene.h"
#include "Camera.h"
#include "SceneGraphState.h"
#include "DebugRenderer.h"
#include "InteractiveHandleLength.h"
//---------------------------------------------------------------------------
using namespace X4;

X4::InteractiveHandlePoint2D::InteractiveHandlePoint2D( const Vector2f &point, const std::string &name )
{
	setName(name);
	m_Point = point;
	m_Hover = false;
}

void X4::InteractiveHandlePoint2D::draw( HandleSetInformation * info, bool selected )
{
	// color changes when mouse hovers over point
	Vector3f color = m_Hover?getColor():getColor()*0.9f;
	
	// determine radius
	float radiusX = getSize()*info->m_HandleSizeScale / info->m_ScreenWidth*(selected?1.2f:1.0f);
	float radiusY = getSize()*info->m_HandleSizeScale / info->m_ScreenHeight*(selected?1.2f:1.0f);

	// draw
	GLHelper::begin2D(1,1);
  	glBegin(GL_TRIANGLE_FAN);
 	glColor3fv(color.data());
 	glVertex2fv(m_Point.data());
	int numSegments = 10;
	for( int i=0;i<=numSegments;i++ )
	{
		float angle = (float)i / numSegments * M_PI*2.0f;
		glVertex2f(m_Point[0] + sin(angle)*radiusX,m_Point[1] + cos(angle)*radiusY);
	}
 	glEnd();
	GLHelper::end2D();
}

bool X4::InteractiveHandlePoint2D::selectHandle( float x, float y, const Vector3f &worldSpacePoint,HandleSetInformation * info, bool isSelected )
{
	float radiusX = (float)HandleSize / info->m_ScreenWidth*1.75f;
	float radiusY = (float)HandleSize / info->m_ScreenHeight*1.75f;
	float dist = norm((makeVector2f(x,y)-m_Point).componentProduct(makeVector2f(1.0f/radiusX,1.0f/radiusY)));
	return ( dist <= 1.0f);
}

void X4::InteractiveHandlePoint2D::moveHandle( float dx, float dy, const Vector3f &worldSpacePointInc,HandleSetInformation * info )
{
	m_Point += makeVector2f(dx,dy);
}

void X4::InteractiveHandlePoint2D::hover( float x, float y, const Vector3f &worldSpacePoint,HandleSetInformation * info )
{
	float radiusX = (float)HandleSize / info->m_ScreenWidth*1.75f;
	float radiusY = (float)HandleSize / info->m_ScreenHeight*1.75f;
	float dist = norm((makeVector2f(x,y)-m_Point).componentProduct(makeVector2f(1.0f/radiusX,1.0f/radiusY)));
	m_Hover = ( dist <= 1.0f);
}

X4::InteractiveHandleSet::InteractiveHandleSet( Scene * scene )
{
	m_Scene = scene;
	m_Info.m_Scene = scene;
	m_Info.m_HandleSizeScale = 1.0f;
	m_UseProjectionMatrix = false;
}

bool X4::InteractiveHandleSet::mousePressEvent( int x, int y, const Vector3f &worldSpacePoint, MouseButtons buttons, Modifiers modifiers )
{
	if( !buttons.getLeft())
		return false;

	// compute x,y in normalized coordinates
	float fx = (float)x/m_Info.m_ScreenWidth;
	float fy = 1.0f - (float)y/m_Info.m_ScreenHeight;
	m_Info.m_LastPoint = makeVector2f(fx,fy);
	m_Info.m_InitialPoint = makeVector2f(x,y);

	// setup 3d info
	m_Info.m_InitialWorldSpacePoint = m_Info.m_LastWorldSpacePoint = worldSpacePoint;
	m_Info.m_ScreenTangentSystem = m_Scene->getRootState()->staticState->camera.getLocalCoordsMatrix();
	Vector3f camPos = m_Scene->getRootState()->staticState->camera.getPosition();
	float distance = (m_Info.m_ScreenTangentSystem.transpose()*(worldSpacePoint-camPos))[2];
	float fovY = m_Scene->getRootState()->staticState->viewFrustum.getVerticalFieldOfView();
	float h = tan(fovY/180.0f*M_PI)*distance / m_Info.m_ScreenHeight;
	m_Info.m_ScreenTangentSystem[0] *= -h;
	m_Info.m_ScreenTangentSystem[1] *= -h;


	// ask single active handles first if they want to keep focus
	if( !modifiers.getCtrl())
	if( m_ActiveHandles.size())
	{
		for( mpcard j=0;j<m_ActiveHandles.size();j++ )
			if( m_ActiveHandles[j]->selectHandle(fx,fy,worldSpacePoint,&m_Info,true) )
			{
				return true;
			}
			m_ActiveHandles.clear();
	}

	for( mpcard i=0;i<m_Handles.size();i++ )
	{
		if( m_Handles[i]->selectHandle(fx,fy,worldSpacePoint,&m_Info,true) )
		{
			// try to find handle in active handles
			for( mpcard j=0;j<m_ActiveHandles.size();j++ )
			{
				if( m_ActiveHandles[j] == m_Handles[i])
				{
					m_ActiveHandles.erase(m_ActiveHandles.begin()+j);
					return false;
				}
			}

			m_ActiveHandles.push_back(m_Handles[i]);
			return true;
		}
	}
	return false;
}

void X4::InteractiveHandleSet::mouseMoveEvent( int x, int y )
{
	if( m_ActiveHandles.size() )
	{
		// 2d movement
		float fx = (float)x/m_Info.m_ScreenWidth;
		float fy = 1.0f - (float)y/m_Info.m_ScreenHeight;
		Vector2f newPos = makeVector2f(fx,fy);
		Vector2f m = newPos-m_Info.m_LastPoint;
		
		// 3d movement
		Vector3f tmp = makeVector3f(x,y,0) - makeVector3f(m_Info.m_InitialPoint[0],m_Info.m_InitialPoint[1],0);
		Vector3f newPos3f = m_Info.m_ScreenTangentSystem*tmp + m_Info.m_InitialWorldSpacePoint;
 		Vector3f pos3finc = newPos3f - m_Info.m_LastWorldSpacePoint;

		// project 3d movement onto subspace
		if( m_UseProjectionMatrix )
		{
			pos3finc = m_ProjectionMatrix * m_ProjectionMatrix.transpose() * pos3finc;
		}


		for( mpcard i=0;i<m_ActiveHandles.size();i++ )
			m_ActiveHandles[i]->moveHandle(m[0],m[1],pos3finc,&m_Info);
		m_Info.m_LastPoint = newPos;
		m_Info.m_LastWorldSpacePoint = newPos3f;
	}

}

void X4::InteractiveHandleSet::hover( int x, int y, const Vector3f &worldSpacePoint )
{
	// compute x,y in normalized coordinates
	float fx = (float)x/m_Info.m_ScreenWidth;
	float fy = 1.0f - (float)y/m_Info.m_ScreenHeight;

	for( mpcard i=0;i<m_Handles.size();i++)
		m_Handles[i]->hover(fx,fy,worldSpacePoint,&m_Info);
}

void X4::InteractiveHandleSet::keyPressEvent( GeneralKey key )
{
	if( m_ActiveHandles.size())
	{
		for( mpcard i=0;i<m_ActiveHandles.size();i++ )
			m_ActiveHandles[i]->keyPressEvent(key,&m_Info);
	}
}

void X4::InteractiveHandleSet::draw( GLContext * context )
{
	debugRenderer->clearDebugData("handles");
	debugRenderer->beginRenderJob("handles");
	set<InteractiveHandle*> activeHandles;
	for( mpcard i=0;i<m_ActiveHandles.size();i++ )
		activeHandles.insert(m_ActiveHandles[i]);
	HandleSetInformation info = m_Info;
	info.m_GLContext = context;
	info.m_Scene = m_Scene;
	for( mpcard i=0;i<m_Handles.size();i++ )
	{
		m_Handles[i]->draw(&info,activeHandles.find(m_Handles[i]) != activeHandles.end());
	}
	debugRenderer->endRenderJob();
	//debugRenderer->renderJob(context,getScene(),"handles");
}

InteractiveHandlePoint2D * X4::InteractiveHandleSet::addPoint2DHandle( const Vector2f &pos,const Vector3f &color,const std::string &name )
{
	InteractiveHandlePoint2D * h = new InteractiveHandlePoint2D(pos,name);
	h->setColor(color);
	m_Handles.push_back(h);
	return h;
}

InteractiveHandlePoint3D * X4::InteractiveHandleSet::addPoint3DHandle( const Vector3f &pos,const Vector3f &color,const std::string &name )
{
	InteractiveHandlePoint3D * h = new InteractiveHandlePoint3D(pos,name);
	h->setColor(color);
	m_Handles.push_back(h);
	return h;
}


InteractiveHandleConstraint2D * X4::InteractiveHandleSet::addConstraint2DHandle( const Vector2f &pos,const Vector2f &source, const Vector3f &color,const std::string &name/*=""*/ )
{
	InteractiveHandleConstraint2D * h = new InteractiveHandleConstraint2D(pos,source,name);
	h->setColor(color);
	m_Handles.push_back(h);
	return h;
}


InteractiveHandleBox3D * X4::InteractiveHandleSet::addBox3DHandle( const BoundingBox3f &bb, const Vector3f &color, const Matrix4f &frame, const std::string &name/*=""*/ )
{
	InteractiveHandleBox3D * h = new InteractiveHandleBox3D(bb,frame);
	h->setColor(color);
	h->setName(name);
	m_Handles.push_back(h);
	return h;
}


InteractiveHandleLength * X4::InteractiveHandleSet::addLengthHandle( const Vector3f &p0,const Vector3f &p1, const Vector3f &color, const std::string &name/*=""*/ )
{
	InteractiveHandleLength * h = new InteractiveHandleLength(p0,p1,name);
	h->setColor(color);
	h->setName(name);
	m_Handles.push_back(h);
	return h;
}


InteractiveHandle * X4::InteractiveHandleSet::getHandleByName( const std::string &name )
{
	for( mpcard i=0;i<m_Handles.size();i++ )
	{
		if( m_Handles[i]->getName() == name)
			return m_Handles[i];
	}
	return NULL;
}

void X4::InteractiveHandleSet::resizeEvent( card32 w, card32 h )
{
	m_Info.m_ScreenWidth = w;
	m_Info.m_ScreenHeight = h;
}

InteractiveHandle * X4::InteractiveHandleSet::deleteClosestHandle( const Vector3f &p )
{
	InteractiveHandle * closestHandle = NULL;
	float bestdist = 1e20f;
	mpcard index;
	for( mpcard i=0;i<m_Handles.size();i++ )
	{
		InteractiveHandlePoint3D * pointHandle = dynamic_cast<InteractiveHandlePoint3D*>(m_Handles[i]);
		if( pointHandle )
		{
			float dist = norm(pointHandle->getPoint() - p);
			if( dist < bestdist )
			{
				bestdist = dist;
				index = i;
				closestHandle = pointHandle;
			}
		}
	}
	if( closestHandle )
	{
		delete closestHandle;
		m_Handles.erase(m_Handles.begin() + index);
		for( mpcard i=0;i<m_ActiveHandles.size();i++ )
		{
			if( m_ActiveHandles[i] == closestHandle)
			{
				m_ActiveHandles.erase(m_ActiveHandles.begin() + i);
			}
		}
	}
	return closestHandle;
}

void X4::InteractiveHandleSet::setProjectionMatrix( const Matrix3f &M )
{
	m_UseProjectionMatrix = true;
	m_ProjectionMatrix = M;
}

void X4::InteractiveHandleSet::deactivateMovementProjectionAxis()
{
	m_UseProjectionMatrix = false;
}


X4::InteractiveHandleConstraint2D::InteractiveHandleConstraint2D( const Vector2f &point, const Vector2f &sourcePoint, const std::string &name )
:InteractiveHandlePoint2D(point,name)
{
	m_SourcePoint = sourcePoint;
}

void X4::InteractiveHandleConstraint2D::draw( HandleSetInformation * info, bool selected )
{
	InteractiveHandlePoint2D::draw(info,selected);
	GLHelper::begin2D(1,1);
	glBegin(GL_LINES);
		glColor3f(0.25f,0.25f,0.25f);
		glVertex2fv(m_SourcePoint.data());
		glVertex2fv(getPoint().data());
	glEnd();
	GLHelper::end2D();
}

X4::InteractiveHandlePoint3D::InteractiveHandlePoint3D( const Vector3f &point, const std::string &name )
{
	m_Point = point;
	setName(name);
}

void X4::InteractiveHandlePoint3D::draw( HandleSetInformation * info, bool selected )
{
	if( norm(getColor() - makeVector3f(0.1f,1,1))  < 0.01 )
		return;
	float r = computeHandleSize_worldSpace(m_Point,info);
	debugRenderer->addFastSphere(m_Point,r*(selected?1.2f:1.0f),m_Hover?getColor():getColor()*0.7f,false);
}

bool X4::InteractiveHandlePoint3D::selectHandle( float x, float y, const Vector3f &worldSpacePoint,HandleSetInformation * info, bool isSelected )
{
	float r = computeHandleSize_worldSpace(worldSpacePoint,info)*1.5f;
	return norm(m_Point-worldSpacePoint) <= r;
}

void X4::InteractiveHandlePoint3D::moveHandle( float dx, float dy, const Vector3f &worldSpacePointInc,HandleSetInformation * info )
{
	m_Point += worldSpacePointInc;
}

void X4::InteractiveHandlePoint3D::hover( float x, float y, const Vector3f &worldSpacePoint,HandleSetInformation * info )
{
	if( worldSpacePoint[0] > 1e19f )
	{
		m_Hover = false;
		return;
	}
	float r = computeHandleSize_worldSpace(worldSpacePoint,info)*1.5f;
	m_Hover = norm(m_Point-worldSpacePoint) <= r;
}

float X4::InteractiveHandle::computeHandleSize_worldSpace( const Vector3f &point_worldSpace, HandleSetInformation * info )
{
	Vector3f camPos = info->m_Scene->getRootState()->staticState->camera.getPosition();
	float distance = norm(point_worldSpace-camPos);
	float fovY = info->m_Scene->getRootState()->staticState->viewFrustum.getVerticalFieldOfView();
	float h = tan(fovY/180.0f*M_PI) * info->m_ScreenHeight;
	return m_Size*info->m_HandleSizeScale / h*distance;
}

X4::InteractiveHandle::InteractiveHandle()
{
	m_Size = 5.0f;
}



////////////////////////////////////////////////////////////////////////////////////////////
//		Box Handle
////////////////////////////////////////////////////////////////////////////////////////////

X4::InteractiveHandleBox3D::InteractiveHandleBox3D( const BoundingBox3f &bb, const Matrix4f &initialframe )
:m_Box(bb),m_InitialFrame(initialframe),m_DestinationFrame(IDENTITY4F)
{
	setColor(makeVector3f(0.2f,0.7f,1.0f));
	m_Status = HS_NotSelected;

	for( mpcard i=0;i<6;i++ )
	{
		Vector3f center = bb.getCenter();
		float size = bb.getMinSideLength()*0.6f;
		center[i/2] += (bb.getSideLength(i/2)*0.5f+size*1.5f) * ((i % 2 == 0)?-1:1);
		BoundingBox3f tmpBB(center);
		tmpBB.addBorder(size);
		m_DirectionBB.push_back(tmpBB);		
	}
}

void X4::InteractiveHandleBox3D::draw( HandleSetInformation * info, bool selected )
{
	Vector3f col = selected?getColor():(getColor() + makeVector3f(1,1,1)*0.5f)*0.5f;
	Matrix4f frame = m_DestinationFrame*m_InitialFrame;
	BoundingBox3f bb = m_Box;
	bb.addBorder(bb.getMaxSideLength()*0.025f);
	debugRenderer->addBox(transformVector3f(frame,bb.lowerCorner),shrink4To3(frame), bb.upperCorner-bb.lowerCorner,col);

	if( m_Status == HS_DirectionSelectMode || 
		m_Status == HS_MoveDirectionX || 
		m_Status == HS_MoveDirectionZ || 
		m_Status == HS_MoveDirectionY )
	{
		Vector3f dirBoxColor = getColor() * 0.5f + (makeVector3f(1,1,1)*0.6f)*0.5f;
		for( mpcard i=0;i<6;i++ )
		{
			bool dirSelected = ( i < 2 && m_Status == HS_MoveDirectionX );
			dirSelected |= ( i >=2 && i<4 && m_Status == HS_MoveDirectionY );
			dirSelected |= ( i >=4 && m_Status == HS_MoveDirectionZ );

			debugRenderer->addBox(transformVector3f(frame,m_DirectionBB[i].lowerCorner),shrink4To3(frame), 
				m_DirectionBB[i].upperCorner-m_DirectionBB[i].lowerCorner,dirSelected?dirBoxColor*1.25f:dirBoxColor);
		}
		
		
	}

	//debugRenderer->addBox(transformVector3f(m_DestinationFrame,m_Box.lowerCorner),shrink4To3(m_DestinationFrame), m_Box.upperCorner-m_Box.lowerCorner,col);
	if( selected )
	{
/*		Vector3f center = transformVector3f(m_DestinationFrame,m_Box.getCenter());
		Matrix3f orientation = shrink4To3(m_DestinationFrame);

		Vector3f camPos = info->m_Scene->getRootState()->staticState->camera.getPosition();
		float distance = norm(center-camPos);
		float fovY = info->m_Scene->getRootState()->staticState->viewFrustum.getVerticalFieldOfView();
		float h = tan(fovY/180.0f*M_PI) * info->m_ScreenHeight;
		float tmp =  5.0f*info->m_HandleSizeScale / h*distance;
		
		debugRenderer->addFineArrow(center,center+orientation[0]*(m_Box.getSideLength(0)*0.50f+tmp*10.0f),makeVector3f(1,0,0),tmp, false);
		debugRenderer->addFineArrow(center,center+orientation[1]*(m_Box.getSideLength(1)*0.50f+tmp*10.0f),makeVector3f(0,1,0),tmp, false);
		debugRenderer->addFineArrow(center,center+orientation[2]*(m_Box.getSideLength(2)*0.50f+tmp*10.0f),makeVector3f(0,0,1),tmp, false);*/
	}
}

bool X4::InteractiveHandleBox3D::selectHandle( float x, float y, const Vector3f &worldSpacePoint,HandleSetInformation * info, bool isSelected )
{
	const Vector3f camPos = info->m_Scene->getRootState()->staticState->camera.getPosition();
	// transform ray into bounding box frame
	Ray3f r(camPos,worldSpacePoint-camPos);
	r = transformRay(r,invertFrame(m_DestinationFrame));
	bool intersectsMainBox = rayIntersectsBoundingBox(r.origin,r.direction,m_Box);

	switch(m_Status)
	{
	case HS_NotSelected:
		if( intersectsMainBox )
		{
			m_Status = HS_ScreenSpaceMode;
			return true;
		}
	case HS_ScreenSpaceMode:
		if( intersectsMainBox )
		{
			m_Status = HS_DirectionSelectMode;
			return true;
		}
	case HS_DirectionSelectMode:
	case HS_MoveDirectionX:
	case HS_MoveDirectionZ:
	case HS_MoveDirectionY:
		{
			int bestBox = -1;
			float bestT = 1e20f;
			float t0,t1;
			for( mpcard i=0;i<6;i++ )
			{
				if( !rayIntersectsBoundingBox(r.origin,r.direction,m_DirectionBB[i],t0,t1) )
					continue;
				t0 = min(t0,t1);
				if( t0 < bestT )
				{
					bestBox = i;
					bestT = t0;
				}
			}
			
			if( rayIntersectsBoundingBox(r.origin,r.direction,m_Box,t0,t1) )
			{
				t0 = min(t0,t1);
				if( t0 < bestT )
				{
					m_Status = HS_ScreenSpaceMode;
					return true;
				}
			}
			if( bestBox >= 0 )
			{
				if( bestBox < 2 )
					m_Status = HS_MoveDirectionX;
				if( bestBox >= 2 && bestBox<4 )
					m_Status = HS_MoveDirectionY;
				if( bestBox >= 4 )
					m_Status = HS_MoveDirectionZ;
				return true;
			}
		}
	}

	m_Status = HS_NotSelected;
	return false;
}

void X4::InteractiveHandleBox3D::moveHandle( float dx, float dy, const Vector3f &worldSpacePointInc,HandleSetInformation * info )
{
	Vector3f inc = shrink4To3(m_DestinationFrame).transpose()*worldSpacePointInc;
	switch(m_Status)
	{
	case HS_MoveDirectionX:
	{
		inc = inc.componentProduct(makeVector3f(1,0,0));
		break;
	}
	case HS_MoveDirectionY:
		inc = inc.componentProduct(makeVector3f(0,1,0));
		break;
	case HS_MoveDirectionZ:
		inc = inc.componentProduct(makeVector3f(0,0,1));
		break;
	}
	m_DestinationFrame = m_DestinationFrame*makeTranslation4f(inc);
}

void X4::InteractiveHandleBox3D::keyPressEvent( GeneralKey key,HandleSetInformation * info )
{
	Vector3f rotCenter = transformVector3f(m_InitialFrame,m_Box.getCenter());
	switch(key.getKey())
	{
	case keyNumPad4:
		m_DestinationFrame = m_DestinationFrame*makeTranslation4f(rotCenter)*makeRotY4f((float)M_PI/50.0f)*makeTranslation4f(-rotCenter);
		break;
	case keyNumPad6:
		m_DestinationFrame = m_DestinationFrame*makeTranslation4f(rotCenter)*makeRotY4f(-(float)M_PI/50.0f)*makeTranslation4f(-rotCenter);
		break;
	case keyNumPad5:
		m_DestinationFrame = m_DestinationFrame*makeTranslation4f(rotCenter)*makeRotX4f((float)M_PI/50.0f)*makeTranslation4f(-rotCenter);
		break;
	case keyNumPad8:
		m_DestinationFrame = m_DestinationFrame*makeTranslation4f(rotCenter)*makeRotX4f(-(float)M_PI/50.0f)*makeTranslation4f(-rotCenter);
		break;
	case keyNumPad7:
		m_DestinationFrame = m_DestinationFrame*makeTranslation4f(rotCenter)*makeRotZ4f((float)M_PI/50.0f)*makeTranslation4f(-rotCenter);
		break;
	case keyNumPad9:
		m_DestinationFrame = m_DestinationFrame*makeTranslation4f(rotCenter)*makeRotZ4f(-(float)M_PI/50.0f)*makeTranslation4f(-rotCenter);
		break;
	}
}

X4::Vector3f X4::InteractiveHandleBox3D::transformPoint( const Vector3f &p ) const
{
	return transformVector3f(m_DestinationFrame,p);
}

X4::Matrix3f X4::InteractiveHandleBox3D::transformMatrix( const Matrix3f &M ) const
{
	return shrink4To3(m_DestinationFrame)*M;
}

bool X4::InteractiveHandleBox3D::insideBox( const Vector3f &p ) const
{
	return m_Box.inBoundingBox(transformVector3f(invertFrame(m_InitialFrame),p));
}

bool X4::InteractiveHandleBox3D::intersect( const Vector3f &p,float radius ) const
{
	return m_Box.intersectBoundingSphere(transformVector3f(invertFrame(m_InitialFrame),p),radius);
}

float X4::InteractiveHandleBox3D::distToBox( const Vector3f &p ) const
{
	return sqrt(m_Box.computeSquaredDistance(transformVector3f(invertFrame(m_InitialFrame),p)));
}

