#ifndef InteractiveHandleSetH
#define InteractiveHandleSetH
//---------------------------------------------------------------------------
#include "x3foundation_private.h"
//---------------------------------------------------------------------------
#include "PCInteractionTool.h"
//---------------------------------------------------------------------------

namespace X4
{

class X3FOUNDATION_PRIVATE_API HandleSetInformation
{
public:
	card32 m_ScreenWidth;
	card32 m_ScreenHeight;
	GLContext * m_GLContext;
	Scene*	m_Scene;
	float m_HandleSizeScale;

	Vector2f	m_LastPoint;
	Vector2f	m_InitialPoint;
	Vector3f	m_InitialWorldSpacePoint;
	Vector3f	m_LastWorldSpacePoint;
	Matrix3f	m_ScreenTangentSystem;
};

class X3FOUNDATION_PRIVATE_API InteractiveHandle
{
public:
	InteractiveHandle();
	virtual ~InteractiveHandle() {}
	virtual void draw(HandleSetInformation * info, bool selected) {}

	virtual bool selectHandle(float x, float y, const Vector3f &worldSpacePoint,HandleSetInformation * info, bool isSelected) = 0;
	virtual void moveHandle(float dx, float dy, const Vector3f &worldSpacePointInc,HandleSetInformation * info) = 0;
	virtual void hover(float x, float y, const Vector3f &worldSpacePoint,HandleSetInformation * info) {}
	virtual void keyPressEvent(GeneralKey key,HandleSetInformation * info) {}

	std::string getName() const {return m_Name;}
	void setName(const std::string &name) {m_Name = name;}
	Vector3f getColor() const {return m_Color;}
	void setColor( const Vector3f &color) {m_Color = color;}
	float getSize() const {return m_Size;}
	void setSize(float s) {m_Size=s;}
	mpcard getIndex() const {return m_Index;}
	void setIndex(mpcard v) {m_Index=v;}

protected:

	float computeHandleSize_worldSpace(const Vector3f &point_worldSpace, HandleSetInformation * info);

private:

	std::string		m_Name;
	Vector3f	m_Color;
	float		m_Size;
	mpcard		m_Index;

};

class X3FOUNDATION_PRIVATE_API InteractiveHandlePoint2D : public InteractiveHandle
{
private:
	Vector2f	m_Point;
	bool		m_Hover;

public:

	InteractiveHandlePoint2D(const Vector2f &point, const std::string &name);
	virtual void draw(HandleSetInformation * info, bool selected);
	static const int HandleSize = 5;

	virtual bool selectHandle(float x, float y, const Vector3f &worldSpacePoint,HandleSetInformation * info, bool isSelected);
	virtual void moveHandle(float dx, float dy, const Vector3f &worldSpacePointInc,HandleSetInformation * info);
	virtual void hover(float x, float y, const Vector3f &worldSpacePoint,HandleSetInformation * info);

	Vector2f getPoint() const {return m_Point;}
};

class X3FOUNDATION_PRIVATE_API InteractiveHandlePoint3D : public InteractiveHandle
{
private:
	Vector3f	m_Point;
	bool		m_Hover;

public:

	InteractiveHandlePoint3D(const Vector3f &point, const std::string &name);
	virtual void draw(HandleSetInformation * info, bool selected);
	static const int HandleSize = 5;

	virtual bool selectHandle(float x, float y, const Vector3f &worldSpacePoint,HandleSetInformation * info, bool isSelected);
	virtual void moveHandle(float dx, float dy, const Vector3f &worldSpacePointInc,HandleSetInformation * info);
	virtual void hover(float x, float y, const Vector3f &worldSpacePoint,HandleSetInformation * info);

	Vector3f getPoint() const {return m_Point;}
	void resetPoint(const Vector3f &p) {m_Point = p;}
};

class X3FOUNDATION_PRIVATE_API InteractiveHandleConstraint2D : public InteractiveHandlePoint2D
{
private:
	Vector2f	m_SourcePoint;
public:
	InteractiveHandleConstraint2D(const Vector2f &point, const Vector2f &sourcePoint, const std::string &name);
	virtual void draw(HandleSetInformation * info, bool selected);

	Vector2f	getSourcePoint() const {return m_SourcePoint;}
};
class InteractiveHandleLength;

/**
	Interactive box handle - IMPORTANT! initialframe does not work properly yet, use IDENTITY4F
*/
class X3FOUNDATION_PRIVATE_API InteractiveHandleBox3D : public InteractiveHandle
{

public:
	InteractiveHandleBox3D(const BoundingBox3f &bb, const Matrix4f &initialframe);
	
	virtual void draw(HandleSetInformation * info, bool selected);

	virtual bool selectHandle(float x, float y, const Vector3f &worldSpacePoint,HandleSetInformation * info, bool isSelected);
	virtual void moveHandle(float dx, float dy, const Vector3f &worldSpacePointInc,HandleSetInformation * info);
	virtual void hover(float x, float y, const Vector3f &worldSpacePoint,HandleSetInformation * info) {}
	virtual void keyPressEvent(GeneralKey key,HandleSetInformation * info);

	BoundingBox3f getBox() const {return m_Box;}
	void setBox(BoundingBox3f v) {m_Box=v;}

	Matrix4f getInitialFrame() const {return m_InitialFrame;}
	void setInitialFrame(Matrix4f v) {m_InitialFrame=v;}

	Matrix4f getDestinationFrame() const {return m_DestinationFrame;}
	void setDestinationFrame(Matrix4f v) {m_DestinationFrame=v;}

	Vector3f transformPoint(const Vector3f &p) const;	// compute destination point
	Matrix3f transformMatrix(const Matrix3f &M) const;	// compute destination matrix (for covariance)
	bool insideBox(const Vector3f &p) const;			// returns true if point is inside initial box
	bool intersect(const Vector3f &p,float radius) const;			// returns true if sphere intersects with initial box
	float distToBox(const Vector3f &p) const;	// compute distance to box

private:
	BoundingBox3f	m_Box;
	Matrix4f		m_InitialFrame;
	Matrix4f		m_DestinationFrame;

	enum EHandleStatus
	{
		HS_NotSelected = 0,
		HS_ScreenSpaceMode = 1,		// move handle parallel to screen plane
		HS_DirectionSelectMode = 2,	// shows direction handles
		HS_MoveDirectionX = 3,		// move in x direction only...
		HS_MoveDirectionY = 4,
		HS_MoveDirectionZ = 5
	};

	EHandleStatus	m_Status;
	std::vector<BoundingBox3f>	m_DirectionBB;

};

class X3FOUNDATION_PRIVATE_API InteractiveHandleSet
{
public:

	InteractiveHandleSet(Scene * scene);

	/// project handle movement onto subspace (movement d -> M M^t d)
	void setProjectionMatrix(const Matrix3f &M);

	/// reset to normal movement mode
	void deactivateMovementProjectionAxis();
	
	/// activates handle, returns true if a nearby handle was found
	bool mousePressEvent(int x, int y, const Vector3f &worldSpacePoint, MouseButtons buttons, Modifiers modifiers);

	/// pass movement to active handle or handle other events
	void mouseMoveEvent(int x, int y);

	/// pass mouse movement to handles (for visualization)
	void hover(int x, int y, const Vector3f &worldSpacePoint);

	/// pass movement/keys to active handle
	void keyPressEvent(GeneralKey key);

	/// updates resize event
	void resizeEvent(card32 w, card32 h);

	/// guess what
	void draw(GLContext * context);

	/// returns handle with name or NULL if no such handle exists
	InteractiveHandle * getHandleByName(const std::string &name);

	/// set global handle scale
	void setHandleScale(float s) {m_Info.m_HandleSizeScale = s;}

	/// deletes all handles (you have to take care of all existing pointers to handles)
	void clear();

	/// delete closest handle (returns handle pointer for identification reasons)
	InteractiveHandle * deleteClosestHandle(const Vector3f &p);

	// add different handle types (HandleSet is owner!!)
	// ------------------------------

	/// create handles
	InteractiveHandlePoint2D * addPoint2DHandle(const Vector2f &pos, const Vector3f &color, const std::string &name="");
	InteractiveHandlePoint3D * addPoint3DHandle(const Vector3f &pos, const Vector3f &color, const std::string &name="");
	InteractiveHandleConstraint2D * addConstraint2DHandle(const Vector2f &pos,const Vector2f &source, const Vector3f &color, const std::string &name="");
	InteractiveHandleBox3D * addBox3DHandle(const BoundingBox3f &bb, const Vector3f &color, const Matrix4f &frame, const std::string &name="");
	InteractiveHandleLength * addLengthHandle(const Vector3f &p0,const Vector3f &p1, const Vector3f &color, const std::string &name="");

private:

	std::vector<InteractiveHandle*>	m_Handles;
	std::vector<InteractiveHandle*>	m_ActiveHandles;
	Scene	*	m_Scene;

	HandleSetInformation	m_Info;
	
	Matrix3f				m_ProjectionMatrix;		// optional: project movements into subspace
	bool					m_UseProjectionMatrix;

};

}

#endif