//---------------------------------------------------------------------------
#ifndef Dynamic_KMeansClusteringH
#define Dynamic_KMeansClusteringH
//---------------------------------------------------------------------------
#include "x3foundation_private.h"
//---------------------------------------------------------------------------
#include <vector>
#include <queue>
//---------------------------------------------------------------------------

namespace X4
{

	#define HAS_MEM_FUNC(func, name)                                        \
    template<typename T, typename Sign>                                 \
    struct name {                                                       \
        typedef char yes[1];                                            \
        typedef char no [2];                                            \
        template <typename U, U> struct type_check;                     \
        template <typename _1> static yes &chk(type_check<Sign, &_1::func> *); \
        template <typename   > static no  &chk(...);                    \
        static bool const value = sizeof(chk<T>(0)) == sizeof(yes);     \
    }

	template<bool C, typename T = void>
	struct enable_if{ typedef T type; };

	template<typename T>
	struct enable_if<false, T> { };


	/**
	* Simple K-meadian clustering algorithm with a possibility to recluster the given elements.
	* The algorithm is implemented in a templated fashion. Each element to be clustered must provide a distance
	* metric, to be able to be clustered. The distance metric must either provide a static or a non-static method distance()
	* which returns the distance between two elements.
	*
	* The advantage of this algorithm is, that the number of clusters not need to be known in advance.
	* The disadvantage is, that there are no theoretical guarantees about the quality and running time of the clustering given.
	*
	* First all elements are added into one cluster. Next we look for elements which are above the given distance threshold.
	* This elements cannot belong to a cluster because they are to far away from the median of the cluster. Points, which are far from the median
	* are reculstered until every point is assigned to a cluster.
	*
	* @param Element A class which does support simple arithmetic operations +,-,*,/ 
	* @param ElementDistance An class which has a (static) float distance(const Element&, const Element&) method
	**/					
	template <class Element, class ElementDistance>
	class InverseDynamicKMedianClustering
	{
	protected:
		ElementDistance* distanceFunction;

		template<class Dist>
		struct Cluster
		{
			HAS_MEM_FUNC(distance, has_distance);

			template<typename T, typename Element> 
			typename enable_if<has_distance<T, float32(T::*)(const Element&, const Element&)>::value, float32>::type
			computeElementDistance(T* t, const Element& a, const Element& b)
			{
				return t->distance(a,b);
			}

			template<typename T, typename Element> 
			typename enable_if<!has_distance<T, float32(T::*)(const Element&, const Element&)>::value, float32>::type
			computeElementDistance(T* t, const Element& a, const Element& b)
			{
				return T::distance(a,b);
			}

			Dist* dist;

			std::vector<Element> elements;
			size_t median;
			Cluster(const Element& el, Dist* distanceFunction) : dist(distanceFunction)
			{
				elements.push_back(el);
				median = 0;
			}
			inline float distance(const Element& a, const Element& b)
			{
				return computeElementDistance<Dist>(dist, a,b);
			}
			inline void recomputeMedian()
			{
				// compute the median elements
				// here we use a brute force approach to compute geometric mean
				float minsum = 1e30f;
				for (unsigned i=0; i < elements.size(); i++)
				{
					float distsum = 0;
					for (unsigned j=0; j < elements.size(); j++)
						distsum += distance(elements[i], elements[j]);
					if (distsum < minsum)
					{
						minsum = distsum;
						median = i;
					}
				}
			}
			inline void add(const Element& elem)
			{
				elements.push_back(elem);
				recomputeMedian();
			}
			inline Element pop(unsigned i)
			{
				Element el = elements[i];
				elements.erase(elements.begin() + i);
				recomputeMedian();
				return el;
			}
		};

		std::vector<Cluster<ElementDistance>> clusters;

	public:

		InverseDynamicKMedianClustering() : distanceFunction(NULL) {}
		InverseDynamicKMedianClustering(ElementDistance& distance) : distanceFunction(&distance) {}

		//! To be executed everytime before new computation
		inline void reset()
		{
			elements.clear();
			clusters.clear();
		}

		//! Create specific number of clusters
		inline void createClusters(unsigned num) { clusters.resize(num); }

		//! Add new element into the set of elements to be clustered
		inline void add(const Element& elem, unsigned clusterid=0)
		{
			if (clusters.size() == 0)
				clusters.push_back(Cluster<ElementDistance>(elem, distanceFunction));
			else
				clusters[clusterid].add(elem);
		}

		//! Get number of clusters after the last clustering
		inline size_t getNumClusters() const { return clusters.size(); }

		//! Retrieve a cluster containing alist of elements
		inline const std::vector<Element>& getCluster(size_t index) { return clusters[index].elements; }

		//! Getr index of the median element of a cluster
		inline size_t getClusterMedian(size_t index) const { return clusters[index].median; }

		/**
		* Perform clustering process. The returned number is the number of clusters.
		* @param distance a threshold for the largest distance for an alement allowed to be distant to a mean of a cluster
		**/
		inline size_t cluster(float distanceThreshold)
		{
			// at this point we should have all elements organized in one cluster
			// we now need to rearrange the clusters, so that similar elements are better clustered
			// this is achieved by recomputing the medians and reculster the elements
			while(true)
			{
				queue<Element> reschedule;

				// recompute means
				for (unsigned j=0; j < clusters.size(); j++)
				{
					// check what elements from this cluster might need to be rescheduled
					for (unsigned k=0; k < clusters[j].elements.size(); k++)
						if (clusters[j].distance(clusters[j].elements[k], clusters[j].elements[clusters[j].median]) > distanceThreshold)
							reschedule.push(clusters[j].pop(k));
				}

				// if we converged
				if (reschedule.empty()) break;

				// recluster elements
				// go through all elements which need to be rescheduled and either put them into other clusters or build new clusters
				while(!reschedule.empty())
				{
					Element el = reschedule.front(); reschedule.pop();

					// look for new cluster where this one can be fit into
					// if nothing found, then create new cluster with this element as the mean				
					int addtocluster = -1;
					float mind = 1e30f;
					for (unsigned j=0; j < clusters.size(); j++)
					{
						float dvw = clusters[j].elements.size() != 0 ? clusters[j].distance(el, clusters[j].elements[clusters[j].median]) : 0;			
						if (dvw < distanceThreshold && dvw < mind)
						{
							addtocluster = j;
							mind = dvw;
						}
					}

					if (addtocluster == -1)
						clusters.push_back(Cluster<ElementDistance>(el, distanceFunction));
					else if (addtocluster > -1)
						clusters[addtocluster].add(el);
				} // queue of a cluster to reschedule


			} // end of reschedule iterations

			return clusters.size();
		}

	};

} // end namespace
#endif