#include "StdAfx.h"
//------------------------------------------------------------------------------------------------
#include "JointHistogram.h"
#include <PointSet.h>
#include <CopyObjectProperties.h>
#include <ObjectClassproperty.h>
#include <EmbeddedObjectClassProperty.h>

//------------------------------------------------------------------------------------------------
using namespace X4;

IMPLEMENT_X4_CLASS( JointHistogram , 0)
{
    BEGIN_CLASS_INIT( JointHistogram );

    ADD_CARD32_PROP (m_count, 0);
    ADD_FLOAT32_FIXED_ARRAY_PROP(m_min, 0, 2, 1);
    ADD_FLOAT32_FIXED_ARRAY_PROP(m_max, 0, 2, 1);
    ADD_CARD32_FIXED_ARRAY_PROP(m_binSize, 0, 2, 1);
    ADD_FLOAT32_VAR_ARRAY_PROP(m_jointHist, 0);
    ADD_FLOAT32_VAR_ARRAY_PROP(m_hist, 0);
}

//------------------------------------------------------------------------------------------------
JointHistogram::JointHistogram()
{
    setup(0,0);
}

//------------------------------------------------------------------------------------------------
void JointHistogram::setup(card32 binSize1, card32 binSize2)
{
    m_count = 0;
    m_binSize[0] = binSize1;
    m_binSize[1] = binSize2;
    for (int i=0; i < 2; i++)
    {
        m_min[i] = 0.0f;
        m_max[i] = 0.0f;
    }
    m_jointHist.clear();
    m_hist.clear();

    m_hist.resize(binSize1 + binSize2, 0);
    m_jointHist.resize(binSize2 * binSize1, 0);
}

//------------------------------------------------------------------------------------------------
JointHistogram::~JointHistogram()
{

}

//------------------------------------------------------------------------------------------------
void JointHistogram::setMinMax(float min, float max, card32 binIndex)
{
    x4Assert(binIndex < 2);
    m_min[binIndex] = min;
    m_max[binIndex] = max;
}

//------------------------------------------------------------------------------------------------
void JointHistogram::add(float32 val1, float32 val2)
{
    // scale to [0, no_bins-1]
    int bin1 = static_cast<int>((val1 - m_min[0])/(m_max[0]-m_min[0]) * (m_binSize[0] - 1) + 0.5f);
    int bin2 = static_cast<int>((val2 - m_min[1])/(m_max[1]-m_min[1]) * (m_binSize[1] - 1) + 0.5f);

	x4Assert(bin1 < (int)m_binSize[0]);
	x4Assert(bin2 < (int)m_binSize[1]);

    m_jointHist[(bin2 * m_binSize[0] + bin1) % m_jointHist.size()]++;
    m_hist[bin1 % m_hist.size()]++;
    m_hist[(m_binSize[0] + bin2) % m_hist.size()]++;
    m_count++;
}

//------------------------------------------------------------------------------------------------
float32 JointHistogram::computeJointEntropy()
{
    x4Assert(m_count > 0);
    float val =0.0f; 
    float entropy = 0.0f;

    //compute joint histogram entropy
    for (unsigned j = 0; j < m_binSize[1]; ++j)
        for (unsigned i = 0; i < m_binSize[0]; ++i)
        {
            val = float32(m_jointHist[j * m_binSize[0] + i])/m_count;
            if (val > 0.0f) entropy -= val * log(val);
        }

    return entropy;
}

//------------------------------------------------------------------------------------------------
float32 JointHistogram::computeEntropy(card32 binIndex)
{
    x4Assert(m_count > 0);
    x4Assert(binIndex < 2);

    float32 val = 0.0f;
    float32 entropy = 0.0f;
    for(unsigned i = 0; i < m_binSize[binIndex]; ++i)
    {
        val = m_hist[binIndex * m_binSize[0] + i]/m_count;
        if (val > 0) entropy -= val*log(val);
    }
    return entropy;
}

//------------------------------------------------------------------------------------------------
float32 JointHistogram::computeNMI()
{
    float32 jE = computeJointEntropy();
    float32 j1 = computeEntropy(0);
    float32 j2 = computeEntropy(1);

    float32 NMI = 2 - 2*jE/(j1 + j2);
    return NMI;
}
