#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "LBase.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "LineFeature.h"
#include "DebugRenderer.h"
#include "AnnSearch.h"
#include "ObjectClassProperty.h"
#include "X4Exception.h"

//---------------------------------------------------------------------------
using namespace X4;

IMPLEMENT_X4_CLASS( LBase ,0)
{
	BEGIN_CLASS_INIT( LBase );
	ADD_VECTOR3F_PROP(m_BasePoint, 0);
	ADD_CARD32_PROP(m_Index1, 0);
	ADD_CARD32_PROP(m_Index2, 0);
	ADD_CARD32_PROP(m_BaseType,0);
	ADD_OBJECT_PROP(m_F1,0, LineFeature::getClass(),true);
	ADD_OBJECT_PROP(m_F2,0, LineFeature::getClass(),true);
}

LBase::LBase(void)
{
	m_F1 = NULL;
	m_F2 = NULL;
}

LBase::~LBase(void)
{
	delete m_F1;
	delete m_F2;
}

void LBase::assign(const Object* obj, X4CopyContext *context)
{
	// Persistent::assign(obj,context);
	const LBase * o = dynamic_cast<const LBase*>(obj);
	x4Assert( o != NULL );

	//copyObjectProperties(obj,this);
	delete m_F1; m_F1 = (LineFeature *)o->m_F1->copy();
	delete m_F2; m_F2 = (LineFeature *)o->m_F2->copy();
	m_BaseType = o->m_BaseType;
	m_BasePoint = o->m_BasePoint;
	m_Index1 = o->m_Index1;
	m_Index2 = o->m_Index2;

}

bool LBase::isSimilar( LBase * b, float maxSurfaceCurvDiff, float maxSlippageCurvatureDiff, float maxAngleDiff )
{
	if( fabs(m_F1->m_Curvature - b->m_F1->m_Curvature) > maxSurfaceCurvDiff )
		return false;
	if( fabs(m_F2->m_Curvature - b->m_F2->m_Curvature) > maxSurfaceCurvDiff )
		return false;

	if( fabs(m_F1->m_Curvature2 - b->m_F1->m_Curvature2) > maxSurfaceCurvDiff )
		return false;
	if( fabs(m_F2->m_Curvature2 - b->m_F2->m_Curvature2) > maxSurfaceCurvDiff )
		return false;

	if( fabs(norm(m_F1->m_RotationDirection)-norm(b->m_F1->m_RotationDirection) ) > maxSlippageCurvatureDiff )
		return false;
	if( fabs(norm(m_F2->m_RotationDirection)-norm(b->m_F2->m_RotationDirection) ) > maxSlippageCurvatureDiff )
		return false;

	if( norm(m_F1->m_RotationDirection) < 0.05f && norm(m_F2->m_RotationDirection) < 0.05f )
	{
		float cos_angle1 = fabs(normalize(m_F1->m_LineDirection)*normalize(m_F2->m_LineDirection));
		float cos_angle2 = fabs(normalize(b->m_F1->m_LineDirection)*normalize(b->m_F2->m_LineDirection));
		if( fabs(cos_angle1 - cos_angle2) > maxAngleDiff )
			return false;
	}

	return true;
}

void LBase::setup( LineFeature * f1, LineFeature * f2, card32 index1, card32 index2, float rotThres )
{
	delete m_F1; m_F1 = (LineFeature*)f1->copy();
	delete m_F2; m_F2 = (LineFeature*)f2->copy();
	m_Index1 = index1;
	m_Index2 = index2;

	if( norm(f1->m_RotationDirection) < rotThres )
	{
		if( norm(f2->m_RotationDirection) < rotThres )
		{
			m_BasePoint = f1->computePseudoIntersectionPoint( f2 );
			m_BaseType = BT_NoRotation;
		}
		else
		{
			m_BaseType = BT_F2Rotates;
			Vector3f center = m_F2->computeRotationCenter();
			Vector3f d = normalize(m_F1->m_LineDirection);
			Vector3f p = m_F1->m_Position;
			m_BasePoint = p + d * (d*(center-p));
		}
	}
	else
	{
		if( norm(f2->m_RotationDirection) < rotThres )
		{
			m_BaseType = BT_F1Rotates;
			Vector3f center = m_F1->computeRotationCenter();
			Vector3f d = normalize(m_F2->m_LineDirection);
			Vector3f p = m_F2->m_Position;
			m_BasePoint = p + d * (d*(center-p));
		}
		else
			throw X4Exception("LBase::setup() - two lines with rotational slippage are not implemented yet!\n"
			"(Sry, but we thought that would never happen. We didn't listen. We didn't listen!)");
	}
	
}

Matrix4f LBase::computeTransformation( LBase * b )
{
	LineFeature* olb_f1 = b->m_F1;
    LineFeature* olb_f2 = b->m_F2;
    {
        Vector3f nmli = m_F1->m_LineDirection.crossProduct(m_F2->m_LineDirection);
        Vector3f nmlj = olb_f1->m_LineDirection.crossProduct(olb_f2->m_LineDirection);
        if (nmli * nmlj < 0) {
            LineFeature* tmp = olb_f1;
            olb_f1 = olb_f2;
            olb_f2 = tmp;
        }
    }

	if( m_BaseType == BT_NoRotation )
		return ::computeTransformation(m_F1,m_F2,olb_f1,olb_f2,false);
	else
	{
		if( m_BaseType == BT_F1Rotates )
			return computeTransformationForRotationalBases(m_F2, m_F1,olb_f2,olb_f1,b->getBasePoint());
		else
			return computeTransformationForRotationalBases(m_F1, m_F2,olb_f1,olb_f2,b->getBasePoint());

	}
}

void LBase::drawWithDebugRenderer(const Vector3f &color, float radius, float length)
{
//	debugRenderer->addFastCylinder(m_F1->m_Position, m_F1->m_LineDirection, 1.0f, 0.05f, makeVector3f(0,1,0),false,false );
//	debugRenderer->addFastCylinder(m_F2->m_Position, m_F2->m_LineDirection, 1.0f, 0.05f, makeVector3f(0,1,0),false,false );

    if (length < 0) {
        debugRenderer->addFineArrow(m_BasePoint, m_F1->m_Position, color, radius,false );
        debugRenderer->addFineArrow(m_BasePoint, m_F2->m_Position, color, radius,false );
    } else {
        Vector3f d1 = m_F1->m_Position - m_BasePoint;
        Vector3f d2 = m_F2->m_Position - m_BasePoint;
        debugRenderer->addFineArrow(m_BasePoint, length*d1/norm(d1)+m_BasePoint, color, radius,false );
        debugRenderer->addFineArrow(m_BasePoint, length*d2/norm(d2)+m_BasePoint, color, radius,false );
    }
}

void LBase::drawFrameWithDebugRenderer()
{
	
}

DVectorF LBase::getSignature()
{
	DVectorF ret;
	ret.setDim(6);
	ret[0] = m_F1->m_Curvature;
	ret[1] = m_F1->m_Curvature2;
	ret[2] = m_F2->m_Curvature;
	ret[3] = m_F2->m_Curvature2;
	ret[4] = norm(m_F1->m_RotationDirection);
	ret[5] = norm(m_F2->m_RotationDirection);

	return ret;
}

Matrix4f LBase::computeTransformationForRotationalBases( LineFeature * fLine, LineFeature * fRot,
																						  LineFeature * fLine2, LineFeature * fRot2, Vector3f basePoint2)
{
	// setup frame of base 1 
	Matrix3f frame;
	frame[0] = normalize(fLine->m_LineDirection);
	if( frame[0] * (fLine->m_Position - m_BasePoint) < 0 )
		frame[0] *= -1;
	Vector3f rot_center = fRot->computeRotationCenter();
	frame[1] = normalize(rot_center - m_BasePoint);
	frame[2] = normalize(frame[0].crossProduct(frame[1]));
	frame[1] = normalize(frame[2].crossProduct(frame[0]));
	if( fLine->m_Normal * frame[2] < 0.0f )
		frame[2] *= -1.0f;

	// setup frame of base 2 
	Matrix3f frame2;
	frame2[0] = normalize(fLine2->m_LineDirection);
	if( frame2[0] * (fLine2->m_Position - basePoint2) < 0 )
		frame2[0] *= -1;
	Vector3f rot_center2 = fRot2->computeRotationCenter();
	frame2[1] = normalize(rot_center2 - basePoint2);
	frame2[2] = normalize(frame2[0].crossProduct(frame2[1]));
	frame2[1] = normalize(frame2[2].crossProduct(frame2[0]));
	if( fLine2->m_Normal * frame2[2] < 0.0f )
		frame2[2] *= -1.0f;

	Matrix4f hframe1 = makeTranslation4f(m_BasePoint) * expand3To4(frame);
	Matrix4f hframe2 = makeTranslation4f(basePoint2) * expand3To4(frame2);
	return hframe2 * invertFrame(hframe1);
	
}

bool LBase::searchTestedTransformation( const Matrix4f transf, std::vector<Matrix4f> &transformations, float translationTolerance )
{
	Vector3f basePoint = getBasePoint();
	Vector3f transformedBasePoint = transformVector3f(transf,basePoint);
	const float rotTolerance = 0.1f;
	for( unsigned i=0;i<m_TransformationList.size();i++ )
	{
		Matrix4f compTransf = transformations[m_TransformationList[i]];
		Matrix3f diffMat = shrink4To3(compTransf - transf);
		if( norm(diffMat[0])>rotTolerance || 
			norm(diffMat[1])>rotTolerance ||
			norm(diffMat[2])>rotTolerance )
			continue;

		Vector3f compTransformedBasePoint = transformVector3f(compTransf,basePoint);
		float dist = norm( transformedBasePoint - compTransformedBasePoint );
		if( dist < translationTolerance )
			return true;

	}
	return false;
}

void LBase::drawTransformationsWithDR(std::vector<Matrix4f> &transformations)
{
	Vector3f bp = getBasePoint();
	for( unsigned i=0;i<m_TransformationList.size();i++ )
	{
		Matrix4f T = transformations[ m_TransformationList[i] ];
		debugRenderer->addLine(bp, transformVector3f(T,bp));
	}
}

Vector3f computeNearestPointTmp( const Vector3f &dir0, const Vector3f &p0, const Vector3f &dir1, const Vector3f &p1 )
{
	// transform line 0 into local frame of line 1
	Matrix3f frame1 = calcTangentSystem(dir1);
	Vector3f g = frame1.transpose() * dir0;
	Vector3f g0 = frame1.transpose() * (p0 - p1);
	g[2] = 0;
	g0[2] = 0;

	float t = (g * g0) / (g*g);

	return p0 - dir0*t;
}

Vector3f computePseudoIntersectionPointTmp(const Vector3f &dir0, const Vector3f &p0, const Vector3f &dir1, const Vector3f &p1 )
{
	return computeNearestPointTmp(dir0, p0, dir1, p1)*0.5f + computeNearestPointTmp(dir1, p1, dir0, p0) * 0.5f;
}


Matrix4f LBase::getCoordinateFrame() const
{
	if( m_BaseType == BT_F1Rotates || m_BaseType == BT_F2Rotates)
	{
		Matrix3f frame2;
		LineFeature * fLine2 = m_BaseType == BT_F1Rotates?m_F2:m_F1;
		LineFeature * fRot2 = m_BaseType == BT_F1Rotates?m_F1:m_F2;
		Vector3f basePoint2 = m_BasePoint;
		frame2[0] = normalize(fLine2->m_LineDirection);
		if( frame2[0] * (fLine2->m_Position - basePoint2) < 0 )
			frame2[0] *= -1;
		Vector3f rot_center2 = fRot2->computeRotationCenter();
		frame2[1] = normalize(rot_center2 - basePoint2);
		frame2[2] = normalize(frame2[0].crossProduct(frame2[1]));
		frame2[1] = normalize(frame2[2].crossProduct(frame2[0]));
		if( fLine2->m_Normal * frame2[2] < 0.0f )
			frame2[2] *= -1.0f;
		return makeTranslation4f(basePoint2) * expand3To4(frame2);
	}


	Vector3f pseudoIntersectionPoint1 = computePseudoIntersectionPointTmp(m_F1->m_LineDirection,m_F1->m_Position,
		m_F2->m_LineDirection,m_F2->m_Position);

	Vector3f l1 = normalize(m_F1->m_LineDirection);
	Vector3f l2 = normalize(m_F2->m_LineDirection);
	if( l1 * (m_F1->m_Position - pseudoIntersectionPoint1) < 0.0f )
		l1 *= -1;
	if( l2 * (m_F2->m_Position - pseudoIntersectionPoint1) < 0.0f )
		l2 *= -1;

	Matrix3f rot_frame;
	rot_frame[0] = l1;
	rot_frame[2] = normalize(l1.crossProduct(l2));
	rot_frame[1] = normalize(rot_frame[2].crossProduct(l1));
	if( m_F1->m_Normal * rot_frame[2] < 0.0f )
		rot_frame[2] *= -1.0f;
	return makeTranslation4f(pseudoIntersectionPoint1) * expand3To4(rot_frame);

}

LBaseQuerry::LBaseQuerry( LBase ** base_array, const card32 numBases, DVectorF scaleSignature) 
{
	m_ScaleSignature = scaleSignature;
	card32 dim = base_array[0]->getSignature().getDim();
	m_Signatures.resize(numBases * dim);
	m_SignaturePointers.resize(numBases);
	for( unsigned i=0;i<numBases;i++ )
	{
		DVectorF sig = base_array[i]->getSignature().componentProduct(scaleSignature);
		for( unsigned j=0;j<dim;j++ )
		{
			m_Signatures[i*dim+j] = sig[j];
		}
		m_SignaturePointers[i] = &m_Signatures[i*dim];
	}

	m_ANN = new AnnSearch(0.01f);
	m_ANN->buildSearchStructure(numBases, dim, &m_SignaturePointers[0]);
	m_Indices.resize(numBases);
}

int LBaseQuerry::neighbourQuerry(DVectorF signature, float radius, int *& indices) 
{
	DVectorF transformedSig = signature.componentProduct(m_ScaleSignature);
	m_NumValidIndices = m_ANN->getkNearestNeighborsInRadius(&transformedSig[0],(card32)m_SignaturePointers.size(), radius*radius, NULL,&m_Indices[0]);
	indices = &m_Indices[0];
	return m_NumValidIndices;
}

LBaseQuerry::~LBaseQuerry()
{
	delete m_ANN;
}