//---------------------------------------------------------------------------
#include "StdAfx.h"
#include "LBaseSet.h"
#include "CopyObjectProperties.h"
#include "LineFeature.h"
#include "ObjectListProperty.h"
#include <AnnSearch.h>
#include "RGBAImage.h"
#include "FFTWrapper.h"
#include <complex>
#include "DebugRenderer.h"
#include "PCCResampler.h"
#include "FastSphereQuerry.h"
#include "VertexArray.h"
#include "Random.h"
#include "ClassMethods.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
IMPLEMENT_X4_CLASS(SHDDescriptorCluster,0)
{
	BEGIN_CLASS_INIT(SHDDescriptorCluster);
	ADD_MAX_PLATFORM_CARD_PROP(m_Representative,0);
	ADD_VAR_ARRAY_PROP( m_Elements,0, new MaxPlatformCardClassProperty("element", MetaClass::FIRST_VERSION + 0,NULL));
	//ADD_MATRIX4F_VAR_ARRAY_PROP( m_Transforms,0);
}

IMPLEMENT_X4_CLASS(LBaseSetTranSet,0)
{
    BEGIN_CLASS_INIT(LBaseSetTranSet);
}

void SHDDescriptorCluster::write(OutputObjectStream *s) const
{
	Persistent::write(s);

	mpcard numTransformations = m_Transforms.size();
	s->writeMaxPlatformCard(numTransformations);
	for( mpcard i=0;i<numTransformations;i++ )
	{
		s->write(m_Transforms[i]);
	}
}

void SHDDescriptorCluster::read(InputObjectStream *s)
{
	Persistent::read(s);

	mpcard numTransformations;
	s->readMaxPlatformCard(numTransformations);
	m_Transforms = std::vector<Matrix4f>(numTransformations);
	for( mpcard i=0;i<numTransformations;i++ )
	{
		s->read(m_Transforms[i]);
	}
}



IMPLEMENT_X4_CLASS(LBaseSet, 5)
{
	BEGIN_CLASS_INIT(LBaseSet);
	ADD_OBJECT_LIST_PROP( Base, 0, LBase::getClass() );
	ADD_MATRIX4F_VAR_ARRAY_PROP(transformationVector, 1);
	ADD_OBJECT_LIST_PROP(DescriptorClusters,4,SHDDescriptorCluster::getClass());
    ADD_OBJECT_LIST_PROP(symmtranset, 0, LBaseSetTranSet::getClass());

	ADD_NOARGS_METHOD(clearSymAppResults);
}

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
LBaseSet::LBaseSet()
{
	m_ANNHistogram = NULL;
	m_BaseUPC = NULL;
	m_BaseFQ = NULL;
}

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
LBaseSet::~LBaseSet()
{
	delete m_ANNHistogram;
	delete m_BaseFQ;
	delete m_BaseUPC;
}

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
void LBaseSet::computeAngles()
{
	angleVector.clear();
	const float32 radToDeg = 180.0f / (float)M_PI;
	for(int i = 0; i < (int)baseVector.size(); i++)
	{
		// compute dot product
		// swap sign if angle larger than 90 degree
		float32 dotPDirection	= min(fabs(baseVector[i]->getF1()->m_LineDirection	* baseVector[i]->getF2()->m_LineDirection),1.0f);

		// compute angels in degree
		const float32 angleDirection	= acos(dotPDirection) * radToDeg;

		// store them at the appropriate position
		angleVector.push_back(angleDirection);
	}
}


//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
void LBaseSet::deleteBases( card32 begin, card32 end )
{
	for( unsigned i=begin;i<end;i++ )
	{
		delete baseVector[i];
	}
	baseVector.erase( baseVector.begin() + begin, baseVector.begin() + end);
}

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
void LBaseSet::setDescriptorVector(std::vector<RGBAFloat32Image*> dV)
{
	for (int i = 0; i < (int)descriptorVector.size(); i++)
	{
		delete descriptorVector[i];
	}
	descriptorVector.clear();
	descriptorVector = dV;
}

void LBaseSet::setupHistogramQuery( float precision ) 
{
	delete m_ANNHistogram;
	m_ANNHistogram = NULL;
	if( histogramVector.size() == 0 )
		throw X4Exception("LBaseSet::setupHistogramQuery() - no histograms in baseset!");

	// setup pointers
	m_ANNHistogram_Pointers = std::vector<float*>(histogramVector.size());
	for( mpcard i=0;i<histogramVector.size();i++ )
	{
		m_ANNHistogram_Pointers[i] = histogramVector[i].data();
	}

	m_ANNHistogram = new AnnSearch(precision);
	m_ANNHistogram->buildSearchStructure((card32)histogramVector.size(), (card32)histogramVector[0].size(), &m_ANNHistogram_Pointers[0]);

}

void LBaseSet::getKNearestNeighbors_Histogram( const DVectorF &query, mpcard k, std::vector<int32> &neighbors, std::vector<float> &distances) 
{
	neighbors.resize(k);
	distances.reserve(k);
	ANNpoint q = (float *)query.data();
	m_ANNHistogram->getkNearestNeighbors(q,k,&distances[0],&neighbors[0]);
}


void LBaseSet::clearHistogramInfo() 
{
	for( mpcard i=0;i<m_HistogramInfo.size();i++ )
		delete m_HistogramInfo[i].alignmentHistograms;
	m_HistogramInfo.clear();
}

void LBaseSet::addHistogramInfo( const Matrix4f &frame, RGBAFloat32Image *	alignmentHistograms, float alignmentScale ) 
{
	THistogramInfo s;
	s.alignmentHistograms = alignmentHistograms;
	s.alignmentScale = alignmentScale;
	s.frame = frame;
	m_HistogramInfo.push_back(s);
}

void showHistogramInOutput( FFTWPlan1D * plan, std::string name)
{
}

int computeMaximumCorrelation( Vector4f * image0, Vector4f * image1, mpcard size )
{
	FFTWPlan1D * plan = MainFFTWrapper->getPlan(size,false);
	// convolve by multiply in fourier domain
	for( mpcard i=0;i<size;i++ )
	{
		complex<double> v1(image0[i][0],image0[i][1]);
		complex<double> v2(image1[i][0],-image1[i][1]);	// conjugate
		complex<double> res = v1 * v2;

		plan->in[i][0] = res.real();
		plan->in[i][1] = res.imag();
	}
	// transform back
	fftwf_execute(plan->plan);
	// search maximum correlation
	int ret = 0;
	float maxV = plan->out[0][0];
	for( mpcard i=1;i<size;i++ )
	{
		if( maxV < plan->out[i][0] )
		{
			maxV = plan->out[i][0];
			ret = i;
		}
	}

	if( ret > (int)size/2 )
		return ret-(int)size;
	else
		return ret;
}

Matrix4f LBaseSet::computeAlignment(mpcard hisID1, mpcard hisID2, mpint visualizeID) 
{
	const THistogramInfo s1 = m_HistogramInfo[hisID1];
	const THistogramInfo s2 = m_HistogramInfo[hisID2];

	Vector3f ret;
	ret[2] = 0;	// dont care about depth yet
	mpcard size = s1.alignmentHistograms->getSizeX();
	ret[0] = -s1.alignmentScale/(float)size * (float)computeMaximumCorrelation(s1.alignmentHistograms->getPixels(),s2.alignmentHistograms->getPixels(),size);
	ret[1] = -s1.alignmentScale/(float)size * (float)computeMaximumCorrelation(&(s1.alignmentHistograms->getPixels()[size]),
		&(s2.alignmentHistograms->getPixels()[size]),size);

	if( visualizeID>=0 )
	{
		
	}

//	return s2.frame * makeTranslation4f(ret)*invertFrame(s1.frame);
	return s2.frame * invertFrame(s1.frame);
}


void LBaseSet::write(OutputObjectStream *s) const 
{
	Persistent::write(s);
	mpcard numHistograms = histogramVector.size();
	s->writeMaxPlatformCard(numHistograms);
	if( numHistograms == 0 )
		return;

	mpcard histogramDim = histogramVector.front().getDim();
	s->writeMaxPlatformCard(histogramDim);
	for( mpcard i=0;i<numHistograms;i++ )
	{
		for( mpcard j=0;j<histogramDim;j++ )
			s->write( histogramVector[i][j] );

		const THistogramInfo &info = m_HistogramInfo[i];
		s->write( info.frame );
		s->write( info.alignmentScale );
		s->writeObject( info.alignmentHistograms );
	}

	mpcard numLocalFrames = m_LocalFrames.size();
	s->writeMaxPlatformCard(numLocalFrames);
	for( mpcard i=0;i<numLocalFrames;i++ )
	{
		s->write<Matrix3f>(m_LocalFrames[i]);
	}

	mpcard numResultSymmetries = m_SymApp_Results.size();
	s->writeMaxPlatformCard( numResultSymmetries );
	for( mpcard i=0;i<numResultSymmetries;i++ )
	{
		const TResultSymmetry & e = m_SymApp_Results[i];
		s->write<OOBBox3f>(e.oobb);	// let's pray  that oobb never changes
		card32 numT = (card32)e.transformation.size();
		s->write(numT);
		for( mpcard j=0;j<numT;j++ )
			s->write<Matrix4f>(e.transformation[j]);
	}

	mpcard num_symmtranset = symmtranset.size();
    s->writeMaxPlatformCard( num_symmtranset );
    for (unsigned ii = 0; ii < num_symmtranset; ++ii) {
        symmtranset[ii]->write(s);
    }
}


void LBaseSet::writeResults(OutputObjectStream *s) const 
{
	mpcard numResultSymmetries = m_SymApp_Results.size();
	s->writeMaxPlatformCard( numResultSymmetries );
	for( mpcard i=0;i<numResultSymmetries;i++ )
	{
		const TResultSymmetry & e = m_SymApp_Results[i];
		s->write<OOBBox3f>(e.oobb);	// let's pray  that oobb never changes
		card32 numT = (card32)e.transformation.size();
		s->write(numT);
		for( mpcard j=0;j<numT;j++ )
			s->write<Matrix4f>(e.transformation[j]);
	}
}

void LBaseSet::readResults(InputObjectStream *s) 
{
		// read stored results from symmetry browser app
	mpcard numResultSymmetries;
	s->readMaxPlatformCard( numResultSymmetries );
	m_SymApp_Results = std::vector<TResultSymmetry>(numResultSymmetries);
	for( mpcard i=0;i<numResultSymmetries;i++ )
	{
		TResultSymmetry & e = m_SymApp_Results[i];
		s->read<OOBBox3f>(e.oobb);	// let's pray  that oobb never changes
		card32 numT;
		s->read(numT);
		e.transformation = std::vector<Matrix4f>(numT);
		for( mpcard j=0;j<numT;j++ )
			s->read<Matrix4f>(e.transformation[j]);
	}

}


void LBaseSet::read(InputObjectStream *s) 
{
	Persistent::read(s);
	//if( s->getClassVersion(LBaseSet::getClass()) < MetaClass::FIRST_VERSION + 2 )
	//	return;

	clearHistogramInfo();
	for (int i = 0; i < (int)descriptorVector.size(); i++)
	{
		delete descriptorVector[i];
	}
	descriptorVector.clear();

	mpcard numHistograms;
	s->readMaxPlatformCard(numHistograms);
	if( numHistograms == 0 )
		return;

	mpcard histogramDim;
	s->readMaxPlatformCard(histogramDim);
	DVectorF tmp(histogramDim);
	histogramVector = std::vector< DVectorF > (numHistograms,tmp);
	for( mpcard i=0;i<numHistograms;i++ )
	{
		for( mpcard j=0;j<histogramDim;j++ )
			s->read( histogramVector[i][j] );

		THistogramInfo info;
		s->read( info.frame );
		s->read( info.alignmentScale );
		s->readObject( (Persistent *&)info.alignmentHistograms );
		m_HistogramInfo.push_back(info);
	}

	//if( s->getClassVersion(LBaseSet::getClass()) < MetaClass::FIRST_VERSION + 3 )
	//	return;
	mpcard numLocalFrames;
	s->readMaxPlatformCard(numLocalFrames);
	m_LocalFrames = std::vector<Matrix3f>(numLocalFrames);
	for( mpcard i=0;i<numLocalFrames;i++ )
	{
		s->read<Matrix3f>(m_LocalFrames[i]);
	}

	//// debug render symmetric clusters with lines
	//debugRenderer->beginRenderJob_OneFrame("sym test",1);
	//for( mpcard i=0;i<m_DescriptorClusters.size();i++ )
	//{
	//	SHDDescriptorCluster * c = m_DescriptorClusters[i];
	//	LBase * reprBase = baseVector[ c->m_Representative];
	//	Vector3f col = makeVector3f(rnd01(),rnd01(),rnd01());
	//	for( mpcard j=0;j<c->m_Transforms.size();j++ )
	//	{
	//		debugRenderer->addLine(reprBase->getBasePoint(), transformVector3f(c->m_Transforms[j],reprBase->getBasePoint()),col,col,1);			
	//	}
	//}
	//debugRenderer->endRenderJob();

	//if( s->getClassVersion(LBaseSet::getClass()) < MetaClass::FIRST_VERSION + 5 )
	//	return;

	// read stored results from symmetry browser app
	mpcard numResultSymmetries;
	s->readMaxPlatformCard( numResultSymmetries );
	m_SymApp_Results = std::vector<TResultSymmetry>(numResultSymmetries);
	for( mpcard i=0;i<numResultSymmetries;i++ )
	{
		TResultSymmetry & e = m_SymApp_Results[i];
		s->read<OOBBox3f>(e.oobb);	// let's pray  that oobb never changes
		card32 numT;
		s->read(numT);
		e.transformation = std::vector<Matrix4f>(numT);
		for( mpcard j=0;j<numT;j++ )
			s->read<Matrix4f>(e.transformation[j]);
	}

    mpcard num_symmtranset;
    s->read( num_symmtranset );
    symmtranset.resize(num_symmtranset);
    for (unsigned ii = 0; ii < num_symmtranset; ++ii) {
        symmtranset[ii]->read(s);
    }

}

void LBaseSet::clearDescriptorClusters() 
{
	for( mpcard i=0;i<m_DescriptorClusters.size();i++ )
		delete m_DescriptorClusters[i];
	m_DescriptorClusters.clear();
}

void LBaseSet::resampleBases(const float32 spacing)
{
	std::vector<LBase*> resampled;
	VertexDescriptor vd;
	vd.pushAttrib( mVAD( "position",	3, VAD::DATA_FORMAT_FLOAT32 ) );
	vd.pushAttrib( mVAD( "flags",		1, VAD::DATA_FORMAT_INT32	) );
	vd.pushAttrib( mVAD( "index",		1, VAD::DATA_FORMAT_INT32 ) );
	PointSet* ps = new PointSet();
	ps->clearAndSetup( 1, baseVector.size(), &vd );
	AAT POSITION = ps->getAAT( "position" );
	AAT FLAGS = ps->getAAT( "flags" );
	AAT INDEX = ps->getAAT( "index" );

	for(int i = 0; i < (int)baseVector.size(); i++)
	{
		ps->set3f(i,POSITION,baseVector[i]->getBasePoint());
		ps->set1i(i,INDEX,i);
	}
	UnstructuredInCorePointCloud* pc = new UnstructuredInCorePointCloud();
	pc->setPointSet(ps);

	PCCResampler::resamplePC(pc,spacing);

	for(int i = 0; i < pc->getNumPoints(); i++)
	{
		resampled.push_back(baseVector[ps->get1i(i,INDEX)]);
	}
	debugOutput << "#bases before resampling = " << baseVector.size() << "\n";
	baseVector.clear();
	baseVector = resampled;
	debugOutput << "#bases after resampling = " << baseVector.size() << "\n";
	delete pc;
}

void LBaseSet::initBaseSearchStructures()
{
	delete m_BaseFQ;
	delete m_BaseUPC;

	m_BaseUPC = new UnstructuredInCorePointCloud;
	m_BaseUPC->clearAndSetup(baseVector.size(),true,false,false,true);
	VertexArray baseArray(m_BaseUPC);
	for( mpcard i=0;i<baseArray.getNumElements();i++ )
	{
		baseArray.setPosition3f(i,baseVector[i]->getBasePoint());
	}
	m_BaseFQ = new FastSphereQuerry(m_BaseUPC);

	// init base cluster info (for fast query of individual symmetry transformations)
	m_BaseClusters = std::vector<std::vector<TBaseClusterInfo> >(baseVector.size());
	m_BaseToClusterMap.clear();
	for( mpcard i=0;i<m_DescriptorClusters.size();i++ )
	{
		SHDDescriptorCluster * c = m_DescriptorClusters[i];
		m_BaseToClusterMap[c->m_Representative] = i;
		for( mpcard j=0;j<c->m_Elements.size();j++ )
		{
			TBaseClusterInfo info;
			info.cluster = i;
			info.T = invertFrame(c->m_Transforms[j]);
			m_BaseClusters[c->m_Elements[j]].push_back(info);
		}
	}
}


std::vector<Matrix4f> LBaseSet::gatherSymmetryTransformations( const OOBBox3f &oobb )
{
	Sphere3f bs = oobb.computeAABB().getBoundingSphere();
	if( m_BaseClusters.size() != baseVector.size() )
		initBaseSearchStructures();

	// get all bases in sphere, test if base is in oobb and if true, add symmetry transformations to result
	std::vector<Matrix4f> result;
	mpcard numPoints;
	mpcard * indices;
	m_BaseFQ->querry(bs.center,bs.radius,&indices,numPoints);
	for( mpcard i=0;i<numPoints;i++ )
	{
		mpcard index = indices[i];
		LBase * b = baseVector[index];
		if( oobb.inBoundingBox(b->getBasePoint()))
		{
			for( mpcard j=0;j<m_BaseClusters[index].size();j++ )
			{
				const TBaseClusterInfo &info = m_BaseClusters[index][j];
				result.push_back(info.T);
			}
			
			if(m_BaseToClusterMap.find(index) != m_BaseToClusterMap.end() )
			{
				SHDDescriptorCluster * mycluster = m_DescriptorClusters[ m_BaseToClusterMap[index]];
				for( mpcard j=0;j<mycluster->m_Transforms.size();j++ )
					result.push_back(mycluster->m_Transforms[j]);
			}

		}
	}

	return result;
}

void SHDDescriptorCluster::assign(const Object* obj, X4CopyContext *context)
{
	copyObjectProperties(obj,this);
}

void LBaseSet::clearSymAppResults() 
{
	m_SymApp_Results.clear();
}
