//---------------------------------------------------------------------------
#ifndef LBaseSetH
#define LBaseSetH
//---------------------------------------------------------------------------
#include "x3foundation_private.h"
//---------------------------------------------------------------------------
#include "AttachedData.h"
#include "LBase.h"
#include "OOBBox3f.h"
#include "ObjectListProperty.h"
//---------------------------------------------------------------------------


namespace X4 {

    class AnnSearch;
    class FastSphereQuerry;
    class UnstructuredInCorePointCloud;
    /**
    represents a set of (aligned) bases with similar descriptor
    */
    class X3FOUNDATION_PRIVATE_API SHDDescriptorCluster : public Persistent
    {
        X4_CLASS(SHDDescriptorCluster);

    public:

        virtual void assign(const Object* obj, X4CopyContext *context = NULL);

        mpcard m_Representative;		// descriptor index of representative
        std::vector<mpcard> m_Elements;		// descriptor indices of assigned elements
        std::vector<Matrix4f> m_Transforms;		// transformation from representative to elements
        static bool sortCluster(const SHDDescriptorCluster* a, const SHDDescriptorCluster* b)
        {
            return (a->m_Elements.size() > b->m_Elements.size()); // ATTENTION!!!! > instead of <
        }

        IMPLEMENT_VAR_ARRAY_METHODS_STL(m_Elements);

        virtual void write(OutputObjectStream *s) const;
        virtual void read(InputObjectStream *s);

        //IMPLEMENT_VAR_ARRAY_METHODS_STL(m_Transforms);
    };

    class X3FOUNDATION_PRIVATE_API LBaseSetTranSet : public Persistent
    {
        X4_CLASS(LBaseSetTranSet);
        Matrix4f T;
        pair<card32, card32> bseed;
        std::vector<card32> subset; // feature indices
        std::vector<card32> supp_list; // point indices
        float32 score;

        //bool operator<(LBaseSetTranSet const& v) {
        //    return (score > v.score);
        //}
        //bool operator<(LBaseSetTranSet const* v) {
        //    return (score > v->score);
        //}

        bool operator<(LBaseSetTranSet const& v) {
            return (subset.size() > v.subset.size());
        }
        bool operator<(LBaseSetTranSet const* v) {
            return (subset.size() > v->subset.size());
        }

        bool static com_func(const LBaseSetTranSet* lhs, const LBaseSetTranSet* rhs) {
            //debugOutput << lhs->subset.size() << " - " << rhs->subset.size() << "; " << lhs->supp_list.size() << " - " << rhs->supp_list.size() << "\n";
            return (lhs->supp_list.empty() || rhs->supp_list.empty()) ?
                (lhs->subset.size() > rhs->subset.size()) :
            (lhs->supp_list.size() > rhs->supp_list.size());
        }

        virtual void write(OutputObjectStream *s) const
        {
            s->write<Matrix4f>(T);
            s->write(bseed.first);
            s->write(bseed.second);
            {
                mpcard nss = subset.size();
                s->writeMaxPlatformCard(nss);
                for (unsigned ii = 0; ii < nss; ++ii) {
                    s->write(subset[ii]);
                }
            }
            {
                mpcard nss = supp_list.size();
                s->writeMaxPlatformCard(nss);
                for (unsigned ii = 0; ii < nss; ++ii) {
                    s->write(supp_list[ii]);
                }
            }
            s->write(score);
        }
        virtual void read(InputObjectStream *s)
        {
            s->read<Matrix4f>(T);
            s->read(bseed.first);
            s->read(bseed.second);
            {
                mpcard nss;
                s->read(nss);
                subset.resize(nss);
                for (unsigned ii = 0; ii < nss; ++ii) {
                    s->read(subset[ii]);
                }
            }
            {
                mpcard nss;
                s->read(nss);
                supp_list.resize(nss);
                for (unsigned ii = 0; ii < nss; ++ii) {
                    s->read(supp_list[ii]);
                }
            }
            s->read(score);
        }
    };

    /**
    plz comment your class - solves NP-hard problems in O(1), 
    calculates events before big bang, or someone did not comment this class
    */
    class X3FOUNDATION_PRIVATE_API LBaseSet : public AttachedData
    {
        X4_CLASS( LBaseSet )
    private:
        std::vector<LBase*> baseVector;
        std::vector<float32> angleVector;
        std::vector<Matrix4f> transformationVector;
        IMPLEMENT_VAR_ARRAY_METHODS_STL( transformationVector );
        std::vector<RGBAFloat32Image*> descriptorVector;
        std::vector<DynamicVector<float32>> histogramVector;
        std::vector<Matrix3f> m_LocalFrames;

        struct THistogramInfo
        {
            Matrix4f		frame;
            RGBAFloat32Image *	alignmentHistograms;	// contains 1d histograms for translational alignment
            float			alignmentScale;	// size of local neighborhood (a bit redudant because the same in all entries)
        };

        std::vector<THistogramInfo>	m_HistogramInfo;

        std::vector<SHDDescriptorCluster*>	m_DescriptorClusters;	// results from clustering

        // temporary search structures
        AnnSearch	*		m_ANNHistogram;				// search structure for histograms
        std::vector<float*>		m_ANNHistogram_Pointers;	// helper for ANN structure
        UnstructuredInCorePointCloud *	m_BaseUPC;		// point cloud with base points
        struct TBaseClusterInfo
        {
            mpcard cluster;
            Matrix4f T;	// transforms from base to representative of cluster
        };
        std::vector< std::vector<TBaseClusterInfo> >		m_BaseClusters;	// each base is assigned to a set of clusters
        FastSphereQuerry			*	m_BaseFQ;		// sphere query for baseUPC
        std::map<mpcard,mpcard>				m_BaseToClusterMap;	// representative base to cluster id (sparse set of entries, one for each cluster)

        void initBaseSearchStructures();

    public:
        std::vector<LBaseSetTranSet*> symmtranset;
        IMPLEMENT_OBJECT_LIST_METHODS_STL_NO_PREFIX_M( symmtranset, LBaseSetTranSet );

    public:
        static std::string getDefaultName() {return "LBaseSet";}

        LBaseSet();
        virtual ~LBaseSet();

        virtual void write(OutputObjectStream *s) const;
        virtual void read(InputObjectStream *s);

        void writeResults(OutputObjectStream *s) const;
        void readResults(InputObjectStream *s);

        std::vector<LBase*> getBaseVector() {return baseVector;}
        void setBaseVector(std::vector<LBase*> bv) {baseVector = bv;}
        void computeAngles();
        std::vector<float32> getAngleVector() {return angleVector;};
        void setTransformationVector(std::vector<Matrix4f> v) {transformationVector = v;};
        std::vector<Matrix4f> getTransformationVector() {
            if (transformationVector.size() != symmtranset.size()) {
                transformationVector.clear();
                transformationVector.resize(symmtranset.size());
                for (unsigned ii = 0; ii < symmtranset.size(); ++ii) {
                    transformationVector[ii] = symmtranset[ii]->T;
                }
            }
            return transformationVector;
        };
        void setDescriptorVector(std::vector<RGBAFloat32Image*> dV);
        std::vector<RGBAFloat32Image*> getDescriptorVector() {return descriptorVector;};
        void setHistogramVector(std::vector<DynamicVector<float32>> dV) {histogramVector.clear();histogramVector = dV;};
        std::vector<DynamicVector<float32>> getHistogramVector() {return histogramVector;};

        void clearHistogramInfo();


        Matrix4f getHistogramInfoFrame(int index){return m_HistogramInfo[index].frame;}


        /// add extra information assigned to a histogram entry
        void addHistogramInfo( const Matrix4f &frame, RGBAFloat32Image *	alignmentHistograms, float alignmentScale );

        /// align bases by correlation analysis of 1d histograms (see view finder alignment)
        /// expects two indices of histogram entries (not bases!)
        /// returns transformation from frame[his1] to frame[his2]
        /// if visualizeID >= 0 the alignment is visualized in debugOutput with images
        Matrix4f computeAlignment(mpcard hisID1, mpcard hisID2, mpint visualizeID=-1);


        // histogram interface
        mpcard getNumHistograms() const {return histogramVector.size();} 
        DVectorF getHistogram( mpcard i ) const {return histogramVector[i];}

        /// initialize ANN structure for histogram entries
        void setupHistogramQuery( float precision );

        /// return approximate k-nearest-neighbors to query histogram (call setupHistogramQuery() at least once before calling this method!)
        /// assumes k < #histograms
        void getKNearestNeighbors_Histogram( const DVectorF &query, mpcard k, std::vector<int32> &neighbors, std::vector<float> &distances);

        /// get ANN structure for histogram queries
        AnnSearch *	getANNHistogram() {return m_ANNHistogram;}

        std::vector<Matrix3f> getLocalFrames(){return m_LocalFrames;}
        void setLocalFrames(std::vector<Matrix3f> lF){m_LocalFrames = lF;}

        // gather symmetry transformations within OOBB
        std::vector<Matrix4f> gatherSymmetryTransformations( const OOBBox3f &oobb );



        // descriptorcluster interface 
        void addDescriptorCluster( SHDDescriptorCluster * cluster ) {m_DescriptorClusters.push_back(cluster);m_BaseClusters.clear();}
        mpcard getNumDescriptorClusters() const {return m_DescriptorClusters.size();}
        SHDDescriptorCluster * getDescriptorCluster( mpcard i ) {return m_DescriptorClusters[i];}
        void replaceDescriptorCluster(mpcard i, SHDDescriptorCluster* c){delete m_DescriptorClusters[i];m_DescriptorClusters[i] = c;}
        void clearDescriptorClusters();
        void sortDescriptorClusters(){sort(m_DescriptorClusters.begin(),m_DescriptorClusters.end(),SHDDescriptorCluster::sortCluster);}
        IMPLEMENT_OBJECT_LIST_METHODS_STL(DescriptorClusters, SHDDescriptorCluster);

        // clear descriptorvector1 set und get descriptorvector 2 dynamicvector

        void addBase( Object * obj ) {baseVector.push_back((LBase*)obj);}
        void deleteBases( card32 begin, card32 end );
        Object * getBase( card32 index ) {return baseVector[index];}
        void moveBase( card32 oldPos, card32 newPos ) {}
        card32 getNumBases() {return (card32)baseVector.size();}
        std::string getBaseName(card32 index ) {return "ToDo :-)";}
        void resampleBases(const float32 spacing);


        // results from symmetry browser app
        // sorry for deadlinestyle
        struct TResultSymmetry
        {
            OOBBox3f oobb;
            std::vector<Matrix4f> transformation;
        };
        std::vector<TResultSymmetry>	m_SymApp_Results;
        void clearSymAppResults();


    };


}
#endif