#include "StdAfx.h"
#include <windows.h>
#include <mmsystem.h>
//---------------------------------------------------------------------------
#include "LDCentroidDescriptor.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "HierarchicalKNNIterator.h"
#include "UnstructuredInCorePointCloud.h"
#include "Random.h"
#include "DebugRenderer.h"
#include "PCA.h"
#include <opencv\cxcore.h>
//---------------------------------------------------------------------------
using namespace X4;

IMPLEMENT_X4_CLASS( LDCentroidDescriptor ,0)
{
	BEGIN_CLASS_INIT( LDCentroidDescriptor );

	ADD_FLOAT32_PROP( m_SpatialSigma,0);
	ADD_CARD32_PROP( m_NumRings,0);
	ADD_STRING_PROP( m_Attribute,0);
	ADD_FLOAT32_PROP( m_AttributeSigma,0);
	ADD_CARD32_PROP( m_NumAttributeSamples,0);
	ADD_FLOAT32_PROP(m_MaxHistogramDistance,0);
}

IMPLEMENT_X4_CLASS( LDCentroidSignature ,1)
{
	BEGIN_CLASS_INIT( LDCentroidSignature );
	ADD_VECTOR3F_VAR_ARRAY_PROP(m_Centroids,0);
	ADD_FLOAT32_VAR_ARRAY_PROP(m_CentroidWeights,0);
	ADD_VECTOR3F_PROP(m_PCANormal,1);
}

LDCentroidDescriptor::LDCentroidDescriptor(void)
{
	m_hIter = NULL;
	m_SpatialSigma = 1.0f;
	m_NumRings = 2;
	m_Attribute = "curvature";
	m_AttributeSigma = 0.2f;
	m_NumAttributeSamples = 5;
	m_MaxHistogramDistance = 100.0f;
}

LDCentroidDescriptor::~LDCentroidDescriptor(void)
{
	delete m_hIter;
}

void LDCentroidDescriptor::assign(const Object* obj, X4CopyContext *context)
{
	const LDCentroidDescriptor * o = dynamic_cast<const LDCentroidDescriptor*>(obj);
	x4Assert( o != NULL );

	copyObjectProperties(obj,this);
}

void X4::LDCentroidDescriptor::beginSignatureComputations( UnstructuredInCorePointCloud * upc )
{
	m_hIter = new HierarchicalKNNIterator( upc, 32, NULL);
	m_PosAAT = upc->getAAT("position");
	m_NormalAAT = upc->getAAT("normal");
	m_AttributeAAT = upc->getAAT(m_Attribute);

}

void X4::LDCentroidDescriptor::endSignatureComputations()
{
	delete m_hIter;
	m_hIter = NULL;
}

LocalSignature * X4::LDCentroidDescriptor::computeSignature( const Vector3f &point, const Vector3f &normal, float size, UnstructuredInCorePointCloud * upc )
{
	// prepare signature
	LDCentroidSignature * sig = new LDCentroidSignature;
	sig->m_Centroids.resize(m_NumRings * m_NumAttributeSamples, NULL_VECTOR3F);
	sig->m_CentroidWeights.resize(m_NumRings * m_NumAttributeSamples, 0.0f);

	// init settings
	float32 invSpatialSigma_2 = 1.0f / pow(m_SpatialSigma,2.0f);
	float32 invAttributeSigma_2 = 1.0f / pow(m_AttributeSigma,2.0f);
	float32 maxRad = m_SpatialSigma * (1.0f+ m_NumRings);

	static bool drawSupport = true;
	if( drawSupport )
	{
		drawSupport = false;
		debugRenderer->beginRenderJob("asdfasd");
		debugRenderer->addPlanarEllipse( point, makeVector3f(1,0,0)*maxRad, makeVector3f(0,1,0)*maxRad, makeVector3f(1,0,0),2);
		debugRenderer->addPlanarEllipse( point, makeVector3f(0,0,1)*maxRad, makeVector3f(0,1,0)*maxRad, makeVector3f(1,0,0),2);
		debugRenderer->addPlanarEllipse( point, makeVector3f(1,0,0)*maxRad, makeVector3f(0,0,1)*maxRad, makeVector3f(1,0,0),2);
		debugRenderer->endRenderJob();
	}
	PCA<float,3> pca;

	m_hIter->setSeekPointAndReset(point);
	while( !m_hIter->atEnd())
	{
		// get point and attribute, break if out of support radius
		Vector3f pos = m_hIter->get3f(m_PosAAT);
		float dist = norm(pos - point);
		if( dist > maxRad)
			break;

		float32 attribute = m_hIter->get1f(m_AttributeAAT);
		pca.addPoint(pos);

		// add weighted position to centroids
		for( unsigned r=0;r<m_NumRings;r++)
		{
			float targetDistance = m_SpatialSigma * (float)(r+1);
			float spatialWeight = exp( -pow(targetDistance - dist,2.0f) * invSpatialSigma_2 );
			if( spatialWeight < 0.001f )
				continue;

			for( unsigned i=0;i<m_NumAttributeSamples;i++ )
			{
				float targetAttribute = ((float)i + 0.5f) / (float)m_NumAttributeSamples;
				float attributeWeight = exp( -pow(targetAttribute - attribute,2.0f) * invAttributeSigma_2 );
				sig->m_CentroidWeights[r*m_NumAttributeSamples + i] += attributeWeight * spatialWeight;
				sig->m_Centroids[r*m_NumAttributeSamples + i] += pos * attributeWeight * spatialWeight;
			}
		}


		m_hIter->next();
	}

	// normalize centroids
	for( unsigned i=0;i<sig->m_Centroids.size();i++ )
	{
		float weight = sig->m_CentroidWeights[i];
		if( weight > 1e-7f)
			sig->m_Centroids[i] /= weight;
	}

	// compute normal via pca
	Vector3f eigenValues, centroid;
	Matrix3f eigenVectors;
	pca.analyze(eigenValues, eigenVectors, centroid);
	sig->m_PCANormal = eigenVectors[2];
	if( sig->m_PCANormal*normal < 0)
		sig->m_PCANormal *= -1;
	sig->m_PCANormal *= maxRad;

	return sig;
}


float X4::LDCentroidSignature::compare( LocalSignature * other, LocalDescriptor * myDescriptor )
{
	LDCentroidSignature * sig = dynamic_cast<LDCentroidSignature*>(other);
	LDCentroidDescriptor * cd = dynamic_cast<LDCentroidDescriptor*>(myDescriptor);

	if (cd)
	{
		if( computeHistogramDistance(sig) > cd->m_MaxHistogramDistance )
			return 1e20f;
	}else
	{
		if( computeHistogramDistance(sig) > 100.0f )
			return 1e20f;
	}

	return compareCentroidDistances(sig);
}

float X4::LDCentroidSignature::compareCentroidDistances( LDCentroidSignature * sig ) const
{
	float dist = 0;
	float summedWeight = 0;
	for( unsigned i=0;i<m_Centroids.size();i++ )
	{
		float maxWeight = max( m_CentroidWeights[i], sig->m_CentroidWeights[i]);
		for( unsigned j=1;j<m_Centroids.size();j++ )
		{
			float weight = max(maxWeight, max(m_CentroidWeights[j], sig->m_CentroidWeights[j]));
			float cDist = norm( m_Centroids[i] - m_Centroids[j] ) - norm(sig->m_Centroids[i] - sig->m_Centroids[j]);
			dist +=  pow( weight * cDist, 2.0f);
			summedWeight += weight;
		}
	}
	return sqrt(dist) / summedWeight;
}

void X4::LDCentroidSignature::debug_draw()
{
	for( unsigned i=0;i<m_Centroids.size()-1;i++)
	{
		float t = (float)i/(float)(m_Centroids.size()-1);
		Vector3f col = makeVector3f(1,0,0) * t + makeVector3f(0.5f,1.0f,0.3f)*(1-t);
		debugRenderer->addLine(m_Centroids[i], m_Centroids[i+1], col, col, 2);
	}
}

float LDCentroidSignature::computeHistogramDistance( LDCentroidSignature * sig ) const
{
	float dist = 0;
	unsigned j=0;
	for( unsigned i=0;i<m_CentroidWeights.size();i++ )
	{
		dist += pow((m_CentroidWeights[i]-sig->m_CentroidWeights[i]),2.0f);
	}
	dist = sqrt(dist);
	return dist;
}

Matrix4f LDCentroidSignature::calculateTransformation(LDCentroidSignature * sourceSignature, bool usePCANormal)
{
	debugRenderer->beginRenderJob("test");
	for( unsigned i=0;i<m_Centroids.size()-1;i++)
	{
		Vector3f col = makeVector3f(rnd01(),rnd01(),rnd01());
		debugRenderer->addLine(m_Centroids[i],m_Centroids[i+1], col, col, 2);
		debugRenderer->addLine(sourceSignature->m_Centroids[i],sourceSignature->m_Centroids[i+1], col, col, 2);
	}
	debugRenderer->endRenderJob();
	std::vector<Vector4f> centroidsA, centroidsB;
	for( card32 i=0;i<m_Centroids.size();i++ )
	{
		Vector4f c1 = expand3To4(m_Centroids[i] );
		c1[3] = m_CentroidWeights[i];
		centroidsA.push_back( c1 );
		Vector4f c2 = expand3To4(sourceSignature->m_Centroids[i] );
		c2[3] = sourceSignature->m_CentroidWeights[i];
		centroidsB.push_back( c2 );
	}

	if( usePCANormal )
	{
		for( card32 i=0;i<m_Centroids.size();i++ )
		{
			Vector4f c1 = expand3To4(m_Centroids[i]+m_PCANormal );
			c1[3] = m_CentroidWeights[i];
			centroidsA.push_back( c1 );
			Vector4f c2 = expand3To4(sourceSignature->m_Centroids[i]+sourceSignature->m_PCANormal );
			c2[3] = sourceSignature->m_CentroidWeights[i];
			centroidsB.push_back( c2 );
		}

	}

	// estimate weighted centroids of centroids
	Vector3f centroidA = NULL_VECTOR3F, centroidB = NULL_VECTOR3F;
	std::vector<float> weights(centroidsA.size());
	float summedWeight = 0;
	for( card32 i=0;i<centroidsA.size();i++ )
	{
		weights[i] = min(centroidsA[i][3],centroidsB[i][3]);
		summedWeight += weights[i];
		centroidA += shrink4To3(centroidsA[i]) * weights[i];
		centroidB += shrink4To3(centroidsB[i]) * weights[i];
	}
	centroidA /= summedWeight;
	centroidB /= summedWeight;

	// calculate cross-covariance matrix
	Matrix3f C;
	C *= 0;
	for( card32 i=0;i<centroidsA.size();i++ )
	{
		for( card32 j=0;j<3;j++ )
		{
			for( card32 k=0;k<3;k++ )
			{
				C[j][k] += (centroidsA[i][j] - centroidA[j])
					* (centroidsB[i][k] - centroidB[k]) * weights[i];
			}
		}
	}
	C /=summedWeight;

	// Prepare OpenCV Datastructure CvMat
	Matrix3f U,V;
	Vector3f eigenValues;
	CvMat cvMatInput, cvMat_U, cvMat_W, cvMat_Vt;
	cvInitMatHeader( &cvMatInput, 3, 3, CV_32FC1, C.data());
	cvInitMatHeader( &cvMat_W, 3, 1, CV_32FC1, eigenValues.data() );
	cvInitMatHeader( &cvMat_U, 3, 3, CV_32FC1, U.data() );
	cvInitMatHeader( &cvMat_Vt, 3, 3, CV_32FC1, V.data() );

	// Do SVD
	cvSVD(&cvMatInput, &cvMat_W, &cvMat_U, &cvMat_Vt, CV_SVD_MODIFY_A | CV_SVD_U_T );

	Matrix3f rotation = U*V;
	Matrix4f ret = makeTranslation4f(centroidA) * expand3To4(rotation) * makeTranslation4f(-centroidB);

	return ret;
}