#include "StdAfx.h"
#include <windows.h>
#include <mmsystem.h>
//---------------------------------------------------------------------------
#include "LDCorrelation.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "HierarchicalKNNIterator.h"
#include "UnstructuredInCorePointCloud.h"
#include "SplatPullPushInterpolation.h"
#include "ImageDisplayer.h"
#include "GeometricTools.h"
#include "FFTWrapper.h"
#include "fftw3.h"
#include "RGBAImage.h"
#include "DebugRenderer.h"
#include "MathTools.inline.h"
#include "Timer.h"
#include "StringHelper.h"
#include <complex>
#include "SeparatorClassProperty.h"
#include "LocalICP.h"
#include "CachedRangeQuerry.h"
#include "TemplatedImage.inline.h"
//---------------------------------------------------------------------------
using namespace X4;

IMPLEMENT_X4_CLASS( LDCorrelation ,1)
{
	BEGIN_CLASS_INIT( LDCorrelation );
	ADD_CARD32_PROP( m_NumSamplesPerRing,0);
	ADD_CARD32_PROP( m_NumRings,0);
	ADD_STRING_PROP( m_Attribute,0);
	ADD_SEPARATOR("pruning");
	ADD_CARD32_PROP(m_MaxNumberOfOrientations,1);

}

IMPLEMENT_X4_CLASS( LDCorrelationSignature ,1)
{
	BEGIN_CLASS_INIT( LDCorrelationSignature );
	ADD_MATRIX3F_PROP( m_LocalFrame,0 );
	ADD_FLOAT32_VAR_ARRAY_PROP( m_Data, 0);
	ADD_FLOAT32_VAR_ARRAY_PROP( m_DataInv, 0);
	ADD_FLOAT32_VAR_ARRAY_PROP( m_Spectrum, 0);
	ADD_CARD32_PROP(m_NumPossibleOrientations,1);
}

LDCorrelation::LDCorrelation(void)
{
	m_hIt = NULL;
	m_Attribute = "height";
	m_NumSamplesPerRing = 16;
	m_NumRings = 6;
	m_MaxNumberOfOrientations = 6;
	
}

LDCorrelation::~LDCorrelation(void)
{
	delete m_hIt;
}

void LDCorrelation::assign(const Object* obj, X4CopyContext *context)
{
	LocalDescriptor::assign(obj,context);
	const LDCorrelation * o = dynamic_cast<const LDCorrelation*>(obj);
	x4Assert( o != NULL );

	copyObjectProperties(obj,this);
}

LocalSignature * X4::LDCorrelation::computeSignature( const Vector3f &point, const Vector3f &normal, float size, UnstructuredInCorePointCloud * upc )
{
	LDCorrelationSignature * ret = new LDCorrelationSignature;
	bool debugThis = false;
	
	// compute pca frame
	float sigma = size / 2.0f;
	float inv_sigma_2 = 1.0f / pow(sigma,2.0f);
	float rad_2 = pow(2.0f * sigma,2.0f);
	Matrix3f frame;
	Vector3f center;
	computeLocalFrame(point,sigma,upc,m_hIt,frame,center);
	if( normal * frame[2] < 0)
		frame[2] = -frame[2];
	frame[1] = normalize(frame[2].crossProduct(frame[0]));	// make non-mirrored frame
	Matrix3f frame_t = frame.transpose();
	Vector3f projPoint = point + frame[2] * ((point - center) * frame[2]);
	ret->m_LocalFrame = frame;
	ret->parametricImage = NULL;
	ret->polarImage = NULL;

	// transform points into parametric domain
	std::vector<SplatPullPushInterpolation::TWorkingVertex> vertices;
	AAT posAAT = upc->getAAT("position");
	AAT attrAAT;
	bool dataAttribute = m_Attribute != "height";
	if( dataAttribute )
		attrAAT = upc->getAAT(m_Attribute);
	m_hIt->setSeekPointAndReset(point);
	while( !m_hIt->atEnd())
	{
		Vector3f pos = m_hIt->get3f(posAAT);
		float dist_2 = normQuad(pos - point);
		if( dist_2 > rad_2 )
			break;

		Vector3f ppos3f = frame_t * (pos - projPoint) / size / 2.0f;

		SplatPullPushInterpolation::TWorkingVertex v;
		v.paramPos = makeVector2f(ppos3f[0]+0.5f,ppos3f[1]+0.5f);
		if( dataAttribute )
			v.attribute[0] = m_hIt->get1f(attrAAT);
		else
			v.attribute[0] = ppos3f[2] / size;

		vertices.push_back(v);
		
		m_hIt->next();
	}

	// create parametric image
	card32 imageSize = 32;
	SplatPullPushInterpolation interp;
	float sampleSpacing = ((float)imageSize/2.0f - 1.0) / (float)m_NumSamplesPerRing;
	interp.setSigma(max(2.0f,sampleSpacing * 2.0f));
	TemplatedImage< StaticVector<float,1> > * param_image = interp.createImage1f(imageSize,vertices,false);
	
	// create ring samples
	std::vector<Vector2f> samples(m_NumSamplesPerRing);
	for( unsigned i=0;i<m_NumSamplesPerRing;i++ )
	{
		float angle = (float)i / (float)m_NumSamplesPerRing * 2.0f * M_PI;
		samples[i] = makeVector2f(sin(angle),cos(angle));
	}

	// transform image into polar image
	const Vector2f image_center = makeVector2f(1,1) * ((float)imageSize / 2);
	std::vector<Vector2f> samplePoints;
	const float maxRadius = (float)(imageSize-2) / 2.0f;
	FFTWPlan2D* plan = MainFFTWrapper->getPlan(m_NumRings,m_NumSamplesPerRing,true);
	for( unsigned r=0;r<m_NumRings;r++)
	{
		float rad = (float)(r+1)/(float)m_NumRings * maxRadius;
		float ring_weight = sqr(rad) - sqr((float)r/(float)m_NumRings * maxRadius);
		for( unsigned i=0;i<m_NumSamplesPerRing;i++ )
		{
			Vector2f sample = samples[i] * rad + image_center;
			StaticVector<float,1> attr = param_image->getBilinearInterpolatedPixel(sample[0],sample[1]);
			card32 imageIndex = r * m_NumSamplesPerRing + i;
			plan->in[imageIndex][0] = attr[0] * ring_weight;
			plan->in[imageIndex][1] = 0;
			if( debugThis )
				samplePoints.push_back(sample);
		}
	}

	// debug: show parametric image with sample points
	// and show polar transformed image
	if( debugThis )
	{
		static unsigned num = 0;
		num++;
		char c = num % 2 == 0 ? 'a':'b';
		RGBAFloat32Image image1;
		image1.resize(imageSize,imageSize);
		for( unsigned x=0;x<imageSize;x++)
			for( unsigned y=0;y<imageSize;y++ )
				image1.setPixel(x,y, makeVector4f(1,1,1,0)*param_image->getPixel(x,y)[0]);

		/*for( unsigned i=0;i<samplePoints.size();i++ )
		{
			image1.setPixel(samplePoints[i][0],samplePoints[i][1],makeVector4f(0,0.25f,0,0));
		}*/
		//ImageDisplayer::DisplayRGBAFloat32Image ev(&image1);
		//debugOutput.displayEvent(ImageDisplayer::getClass(), std::string("param_image") + c, &ev);

		RGBAFloat32Image image2;
		image2.resize(m_NumSamplesPerRing,m_NumRings);
		for( unsigned x=0;x<m_NumSamplesPerRing;x++)
		{
			for( unsigned y=0;y<m_NumRings;y++ )
			{
				card32 index = y*m_NumSamplesPerRing+x;
				image2.setPixel(x,y,makeVector4f(1,1,1,0)*plan->in[index][0]);
			}
		}
		//ImageDisplayer::DisplayRGBAFloat32Image ev2(&image2);
		//debugOutput.displayEvent(ImageDisplayer::getClass(), std::string("polar_image")+c, &ev2);
		ret->parametricImage = (RGBAFloat32Image*)image1.copy();
		ret->polarImage = (RGBAFloat32Image*)image2.copy();

		debugRenderer->clearDebugData("debug");
		debugRenderer->beginRenderJob("debug");
		debugRenderer->addPlanarEllipse(projPoint,frame[0]*size,frame[1]*size, makeVector3f(1,0,0),1.0f);
		debugRenderer->endRenderJob();
	}

	// fourier transform polar-image and store
	Timer time;
	fftwf_execute(plan->plan);
	ret->m_Data.resize(m_NumSamplesPerRing*m_NumRings*2);
	for( unsigned x=0;x<m_NumSamplesPerRing*m_NumRings;x++ )
	{
		ret->m_Data[x*2+0] = plan->out[x][0];
		ret->m_Data[x*2+1] = plan->out[x][1];
	}

	// compute spectrum
	ret->m_Spectrum.resize(m_NumSamplesPerRing*m_NumRings);
	for( unsigned x=0;x<m_NumSamplesPerRing*m_NumRings;x++ )
	{
		ret->m_Spectrum[x] = sqrt( pow(plan->out[x][0],2.0f) + pow(plan->out[x][1],2.0f));
	}

	// flip image, fft transform and store
	for( unsigned x=0;x<m_NumSamplesPerRing/2;x++)
	{
		for( unsigned y=0;y<m_NumRings;y++ )
		{
			unsigned index1 = y*m_NumSamplesPerRing+x;
			unsigned index2 = (y+1)*m_NumSamplesPerRing-1-x;
			float tmp = plan->in[index1][0];
			plan->in[index1][0] = plan->in[index2][0];
			plan->in[index2][0] = tmp;
			tmp = plan->in[index1][1];
			plan->in[index1][1] = plan->in[index2][1];
			plan->in[index2][1] = tmp;
		}
	}
	fftwf_execute(plan->plan);
	ret->m_DataInv.resize(m_NumSamplesPerRing*m_NumRings*2);
	for( unsigned x=0;x<m_NumSamplesPerRing*m_NumRings;x++ )
	{
		ret->m_DataInv[x*2+0] = plan->out[x][0];
		ret->m_DataInv[x*2+1] = plan->out[x][1];
	}

	std::vector<Matrix3f> frames;
	ret->computeAlignment(ret, this, frames);
	ret->m_NumPossibleOrientations = (card32)frames.size();


	return ret;
}

void X4::LDCorrelation::beginSignatureComputations( UnstructuredInCorePointCloud * upc )
{
	m_hIt = new HierarchicalKNNIterator(upc,32,NULL);
}

void X4::LDCorrelation::endSignatureComputations()
{
	delete m_hIt;
	m_hIt = NULL;
}

bool compareCorrelationAngles(const Vector3f &v1, const Vector3f &v2)
{
	return v1[1] > v2[1];	// bigger correlation first
}

float X4::LDCorrelationSignature::compare( LocalSignature * other, LocalDescriptor * myDescriptor )
{
	LDCorrelationSignature * sig = (LDCorrelationSignature*)other;
	float dist = 0;
	for( unsigned i=1;i<m_Spectrum.size();i++ )	// 0 value holds average height/attribute
	{
		dist += sqr(m_Spectrum[i]-sig->m_Spectrum[i]);
	}
	return sqrt(dist) * (myDescriptor ? myDescriptor->getDifferenceScaleFactor() : 1.0f);
	return 0;
}

void X4::LDCorrelationSignature::computeAlignment( LDCorrelationSignature * dest, LDCorrelation * myDescriptor, std::vector<Matrix3f> &dest_frames, bool mirrored )
{
	std::vector<Vector3f> angles;
	if( 0 && mirrored )
	{
		computeAngles(dest,myDescriptor,dest->m_DataInv, angles);
		for( unsigned i=0;i<angles.size();i++ )
			angles[i][2] = -1.0f;
	}
	computeAngles(dest,myDescriptor,dest->m_Data, angles);
	sort( angles.begin(), angles.end(), compareCorrelationAngles);


	for( unsigned i=0;i<angles.size();i++ )
	{
		if( i > 0)
		if( angles[i][1] < angles[0][1]*0.8f )
			return;

		float angle = angles[i][0];
		if( angles[i][2] < 0 )
		{
			Matrix3f frame = dest->m_LocalFrame;
			frame[0] = -frame[0];
			frame = frame * makeRotZ3f(-angle);

			dest_frames.push_back(frame);
		}
		else
		{
			Matrix3f frame = dest->m_LocalFrame * makeRotZ3f(angle);
			dest_frames.push_back(frame);
		}
	}
}

void X4::LDCorrelationSignature::computeAngles( LDCorrelationSignature * dest, LDCorrelation * myDescriptor, const std::vector<float32> &destData, std::vector<Vector3f> &result )
{
	FFTWPlan2D * plan = MainFFTWrapper->getPlan(myDescriptor->m_NumRings,myDescriptor->m_NumSamplesPerRing,false);

	// multiply signatures (frequency domain)
	const unsigned numValues = plan->width * plan->height;
	double length = 0;
	std::vector< complex<double> > tmp;
	for( unsigned i=0;i<numValues;i++ )
	{
		complex<double> v1(m_Data[i*2+0],m_Data[i*2+1]);
		complex<double> v2(destData[i*2+0],-destData[i*2+1]);	// conjugate
		complex<double> res = v1 * v2;

		length += sqrt(sqr(res.real()) + sqr(res.imag()));

		//plan->in[i][0] = res.real();
		//plan->in[i][1] = res.imag();
		tmp.push_back( res );
	}

	for( unsigned i=0;i<numValues;i++ )
	{
		plan->in[i][0] = tmp[i].real() / length;
		plan->in[i][1] = tmp[i].imag() / length;
	}

	// retransform to time domain
	fftwf_execute(plan->plan);

	// debug: output correlation image
	bool debugThis = false;
	if( debugThis )
	{

		// slow convolution
		std::vector<float> conv_ref;
		float minCorrRef=1e20f;
		float maxCorrRef=-1e20f;
		unsigned maxCorrIndex = 0;

		RGBAFloat32Image image1;
		image1.resize(myDescriptor->m_NumSamplesPerRing,20);
		image1.clear( makeVector4f(0,0,0,0));

		float minCorr = 1e20f;
		float maxCorr = -1e20f;
		for( unsigned i=0;i<myDescriptor->m_NumSamplesPerRing;i++ )
		{
			minCorr = min( minCorr, plan->out[i][0]);
			maxCorr = max( maxCorr, plan->out[i][0]);
		}

		debugRenderer->clearDebugData("corr");
		debugRenderer->beginRenderJob("corr");
		for( unsigned i=0;i<myDescriptor->m_NumSamplesPerRing;i++ )
		{
			Vector2f v = makeVector2f((float)i/(float)myDescriptor->m_NumSamplesPerRing, (plan->out[i][0]-minCorr)/(maxCorr - minCorr)*0.3f) * 0.1f;
			debugRenderer->add2DPoint(v, makeVector3f(0,1,0));
		}
		debugRenderer->endRenderJob();

		debugOutput << "minCorr/maxCorr: " << minCorr << "," << maxCorr << "\n";

		unsigned numSamplesPerRing_2 = myDescriptor->m_NumSamplesPerRing/2;
		unsigned numRings_2 = myDescriptor->m_NumRings/2;
		for( unsigned i=0;i<image1.getSizeX();i++ )
		{
			float y = (plan->out[i][0]-minCorr) / (maxCorr-minCorr) * ((float)image1.getSizeY()-1.0f);
			if( y >= 0.0f)
				image1.setPixel(i,image1.getSizeY()-1-y,makeVector4f(1,1,0,0));

		}
		ImageDisplayer::DisplayRGBAFloat32Image ev(&image1);
		debugOutput.displayEvent(ImageDisplayer::getClass(), "correlation", &ev);

		if( polarImage )
		{
			ImageDisplayer::DisplayRGBAFloat32Image ev(polarImage);
			debugOutput.displayEvent(ImageDisplayer::getClass(), "polar_1", &ev);
			ImageDisplayer::DisplayRGBAFloat32Image ev2(dest->polarImage);
			debugOutput.displayEvent(ImageDisplayer::getClass(), "polar_2", &ev2);
			ImageDisplayer::DisplayRGBAFloat32Image ev3(parametricImage);
			debugOutput.displayEvent(ImageDisplayer::getClass(), "parametric1", &ev3);
			ImageDisplayer::DisplayRGBAFloat32Image ev4(dest->parametricImage);
			debugOutput.displayEvent(ImageDisplayer::getClass(), "parametric2", &ev4);
		}

	}

	std::vector<bool> validEntries(myDescriptor->m_NumSamplesPerRing,true);
	float overallPeak = -1e20f;
	for( unsigned peak=0;peak<4;peak++ ) // limit number of peaks
	{
		// find maximum
		bool foundPeak = false;
		float maxCorr = -1e20f;
		float minCorr = 0;
		unsigned maxCorrIndex = 0;
		unsigned numSamplesPerRing_2 = myDescriptor->m_NumSamplesPerRing/2;
		unsigned numRings_2 = myDescriptor->m_NumRings/2;
		for( unsigned i=0;i<myDescriptor->m_NumSamplesPerRing;i++ )
		{
			if( !validEntries[i] )
				continue;
			unsigned index = (i+numSamplesPerRing_2) % myDescriptor->m_NumSamplesPerRing;
			index += numRings_2 * myDescriptor->m_NumSamplesPerRing;
			index = i;
			minCorr = min(minCorr,plan->out[index][0]);
			if( maxCorr < plan->out[index][0])
			{
				maxCorr = plan->out[index][0];
				maxCorrIndex = i;
				foundPeak = true;
				overallPeak = max( maxCorr, overallPeak);
			}
		}

		// drop small local maxima
		if( maxCorr/overallPeak < 0.8f )
			return;

		// flag peak and neighbors as visited
		validEntries[maxCorrIndex] = false;
		validEntries[(maxCorrIndex+1) % myDescriptor->m_NumSamplesPerRing] = false;
		validEntries[(maxCorrIndex+myDescriptor->m_NumSamplesPerRing-1) % myDescriptor->m_NumSamplesPerRing] = false;


		// compute maximum
		float y1 = plan->out[(maxCorrIndex+numSamplesPerRing_2-1)%numSamplesPerRing_2][0];
		float y2 = plan->out[maxCorrIndex][0];
		float y3 = plan->out[(maxCorrIndex+1)%numSamplesPerRing_2][0];
		float c = y2;
		float b = -(y1+y3) / 2.0f;
		float a = y1+b-c;
		float fineAngle = -b/2/a + maxCorrIndex;

		// drop invalid peaks
		if( y2 < y1 || y2 < y3 )
			continue;


		// compute frame
		float angle = -fineAngle / (float)myDescriptor->m_NumSamplesPerRing * 2.0f * M_PI;
		result.push_back( makeVector3f(angle,maxCorr,1));
	}



}

bool X4::LDCorrelationSignature::badFeature( LocalDescriptor * myDescriptor ) const
{
	LDCorrelation * corr = dynamic_cast<LDCorrelation*>(myDescriptor);
	if( corr == NULL )
		throw X4Exception("LDCorrelationSignature::badFeature() - invalid descriptor!");
	return m_NumPossibleOrientations > corr->m_MaxNumberOfOrientations;
}

float X4::LDCorrelationSignature::computeICPDistance(const Vector3f &sourceFeature, const Vector3f &destFeature, LDCorrelationSignature * destSig, float32 icpradius, bool mirrored/*=true*/ )
{

	LDCorrelationSignature * corrSigRef = this;
	Matrix4f refFrame = makeTranslation4f(sourceFeature) * expand3To4(corrSigRef->m_LocalFrame);
	refFrame = invertFrame(refFrame);

	// gather points around selected feature for icp
	CachedRangeQuerry querry(m_SourcePC);
	PointSet * ps = querry.getSpherePoints(sourceFeature,icpradius,0);

	LDCorrelationSignature * corrSigDest = destSig;

	std::vector<Matrix3f> dest_frames;
	corrSigRef->computeAlignment( corrSigDest, m_correlationDescriptor, dest_frames, false);

	//debugRenderer->addFineCoordinateFrame(makeTranslation4f(f2->m_Position) * expand3To4(dest_frames[0]), medPointDistance * det->getDescriptorRadius()
	//	, makeVector3f(1.0f,0.0f,0.0f),false);

	float bestResidual = 1e20f;
	Matrix4f bestFrame = IDENTITY4F;

	for( unsigned i=0;i<dest_frames.size();i++ )
	{
		Matrix4f tmpRefFrame = refFrame;
		Matrix4f destFrame = invertFrame(makeTranslation4f(destFeature) * expand3To4(dest_frames[i]));
		//debugRenderer->clearDebugData("icptest");
		//debugRenderer->beginRenderJob("icptest");
		float residual = m_icp->computeFastICP(ps,tmpRefFrame, destFrame, icpradius *0.5f );

		if( residual < bestResidual )
		{
			Matrix4f dest0 = invertFrame(destFrame);
			Matrix4f r0 = invertFrame(refFrame);
			Matrix4f r1 = invertFrame(tmpRefFrame);
			Matrix4f T = invertFrame(r1)*r0;
			bestFrame = dest0 * T;
			bestResidual = residual;
		}
		//debugRenderer->endRenderJob();
	}
	//debugOutput << "best residual: " << bestResidual << "\n";

	return bestResidual;
}

void X4::LDCorrelationSignature::beginICPCompare( LocalICP * icp, PointCloud * sourcePC, PointCloud * destPC, LDCorrelation * correlationDescriptor )
{
	m_icp = icp;
	m_SourcePC = sourcePC;
	m_DestPC = destPC;
	m_correlationDescriptor = correlationDescriptor;
}

void X4::LDCorrelationSignature::endICPCompare()
{

}