#ifndef LDCorrelationH
#define LDCorrelationH
//---------------------------------------------------------------------------
#include "x3foundation_private.h"
//---------------------------------------------------------------------------
#include "LocalDescriptor.h"
//---------------------------------------------------------------------------

namespace X4
{

class HierarchicalKNNIterator;
class LDCorrelationSignature;
class RGBAFloat32Image;
class LocalICP;
class PointCloud;

/**
	geometry descriptor - takes an attribute/height as input
	 * transform local environment via pca into a parametric image
	 * sample image in circles
	 * transform resulting image into frequency domain
   can be used for comparison and approximate alignment (phase correlation)
*/
class X3FOUNDATION_PRIVATE_API LDCorrelation : public LocalDescriptor
{
	X4_CLASS( LDCorrelation )

private:

	std::string	m_Attribute;
	card32	m_NumSamplesPerRing;
	card32	m_NumRings;
	card32	m_MaxNumberOfOrientations;

	// members
	HierarchicalKNNIterator	*	m_hIt;


public:
	LDCorrelation(void);
	~LDCorrelation(void);

   virtual void assign(const Object* obj, X4CopyContext *context = NULL);

	/// setup descriptor
	virtual void beginSignatureComputations(UnstructuredInCorePointCloud * upc);

	/// compute local signature for point
	virtual LocalSignature * computeSignature(const Vector3f &point, const Vector3f &normal, float size, UnstructuredInCorePointCloud * upc);

	/// cleanup
	virtual void endSignatureComputations();

	friend LDCorrelationSignature;

};

class X3FOUNDATION_PRIVATE_API LDCorrelationSignature : public LocalSignature
{
	X4_CLASS(LDCorrelationSignature)
public:
	LDCorrelationSignature() {m_NumPossibleOrientations = 0;}

	/// compares two signatures (descriptor is required for context information)
	virtual float compare( LocalSignature * other, LocalDescriptor * myDescriptor);
	virtual float compare( LocalSignature * other ) { return compare(other, NULL); }

	/// compute best alignments
	void computeAlignment( LDCorrelationSignature * dest, LDCorrelation * myDescriptor, std::vector<Matrix3f> &dest_frames, bool mirrored=true);

	virtual bool badFeature(LocalDescriptor * myDescriptor) const;

	void beginICPCompare(LocalICP * icp, PointCloud * sourcePC, PointCloud * destPC, LDCorrelation * correlationDescriptor);
	float computeICPDistance(const Vector3f &sourceFeature, const Vector3f &destFeature, LDCorrelationSignature * destSig, float32 icpradius, bool mirrored=true);
	void endICPCompare();

	Matrix3f				m_LocalFrame;
	std::vector<float32>	m_Data,m_DataInv;	// contains complex numbers
	std::vector<float32>	m_Spectrum;
	card32				m_NumPossibleOrientations;

	// temporal members
	LocalICP *			m_icp;
	PointCloud *		m_SourcePC, *m_DestPC;
	LDCorrelation	*	m_correlationDescriptor;


	// debug
	RGBAFloat32Image * parametricImage;
	RGBAFloat32Image * polarImage;

	IMPLEMENT_VAR_ARRAY_METHODS_STL(m_Data);
	IMPLEMENT_VAR_ARRAY_METHODS_STL(m_DataInv);
	IMPLEMENT_VAR_ARRAY_METHODS_STL(m_Spectrum);

private:

	// compute optimal angles between two features
	// returns angles + correlation in result
	void computeAngles( LDCorrelationSignature * dest, LDCorrelation * myDescriptor, const std::vector<float32> &destData, std::vector<Vector3f> &result );

};

}

#endif