#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "LDHistogram.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "UnstructuredInCorePointCloud.h"
#include "HierarchicalKNNIterator.h"
#include "MathTools.h"
#include "dynamiclinearalgebra.inline.h"
#include "TopologicalKNNIterator.h"
//---------------------------------------------------------------------------
using namespace X4;

IMPLEMENT_X4_CLASS( LDHistogram ,1)
{
	BEGIN_CLASS_INIT( LDHistogram );

	ADD_CARD32_PROP( m_NumBins ,0 );
	ADD_CARD32_PROP( m_NumHistograms, 0);
	ADD_STRING_PROP( m_Attribute, 0);
	ADD_CARD32_PROP( m_AttributeCompNumber , 0);
	ADD_FLOAT32_PROP( m_MaxValue, 0);
	ADD_FLOAT32_PROP( m_MinValue, 0);
	ADD_BOOLEAN_PROP( m_UseTopology, 1);

}

LDHistogram::LDHistogram(void)
{
	m_NumBins = 8;
	m_NumHistograms = 3;
	m_Attribute = "curvature";
	m_AttributeCompNumber = 2;
	m_MaxValue = 1.0f;
	m_MinValue = -1;

	m_hIt = NULL;
	m_TopQuery = NULL;
	m_UseTopology = false;
}

LDHistogram::~LDHistogram(void)
{
}

void LDHistogram::assign(const Object* obj, X4CopyContext *context)
{
	LocalDescriptor::assign(obj,context);
	const LDHistogram * o = dynamic_cast<const LDHistogram*>(obj);
	x4Assert( o != NULL );

	copyObjectProperties(obj,this);
}

void X4::LDHistogram::beginSignatureComputations( UnstructuredInCorePointCloud * upc )
{
	m_hIt = new HierarchicalKNNIterator( upc, 32, NULL);
	if( m_UseTopology )
	{
		m_TopQuery = new TopologicalRangeQuerry(upc);
	}
}

LocalSignature * X4::LDHistogram::computeSignature( const Vector3f &point, const Vector3f &normal, float size, UnstructuredInCorePointCloud * upc )
{
	float h_2 = - log(0.01f) / pow(size,2.0f);
	PointCloud * pc = upc;

   if (!pc->providesAttribute(m_Attribute) || !pc->getDescr()->providesAttribute(m_Attribute))
   {
      error(std::string(__FUNCTION__ "Given point cloud do not provide the attribute ") + m_Attribute + std::string(". Hence no signature computation is possible."));
      return NULL;
   }

	AAT attrAAT = pc->getAAT( m_Attribute );
	AAT posAAT = pc->getAAT( "position" );
	AAT normalAAT = pc->getAAT("normal");
	VAD attrVAD = pc->getDescr()->getAttribute( m_Attribute );
	x4Assert( attrVAD.dataFormat == VAD::DATA_FORMAT_FLOAT32 );

	DVectorF sigdv;
	sigdv.setDim(m_NumBins * m_NumHistograms);
	sigdv.setZero();
	m_hIt->setSeekPointAndReset( point );
	if( m_UseTopology)
	{
		m_TopQuery->rangeQuerry(point,size);
	}
	std::vector<float> ringWeight;
	for( card32 i=0;i<m_NumHistograms;i++ )
	{
		float radius = (float)(i+1) / (float)m_NumHistograms * size;
		float radius0 = (float)(i) / (float)m_NumHistograms * size;
		ringWeight.push_back( sqr(radius) - sqr(radius0) );
	}

	//bool hasCurv = pc->providesAttribute("curvature");
	//AAT curvAAT;
	//if( hasCurv )
	//	curvAAT = pc->getAAT("curvature");

	if (attrVAD.numComponents < m_AttributeCompNumber)
		m_AttributeCompNumber = attrVAD.numComponents - 1;

	float scale = 1.0f;
	float summed_magnitude = 0;
	unsigned currentPoint = 0;
	while( true )
	{
		
		if( !m_UseTopology && m_hIt->atEnd())
			break;

		Vector3f pos;
		float dist;
		if( !m_UseTopology )
		{
			pos = m_hIt->get3f( posAAT );
			dist = norm( point - pos );
		}
		else
		{
			pos = m_TopQuery->getPosition(currentPoint);
			dist = m_TopQuery->getDistance(currentPoint);
		}
		
		if( dist > size )
			break;

		float attr = 0;
		switch( attrVAD.numComponents )
		{
		case 1:
			attr = m_hIt->get1f( attrAAT );
			break;
		case 2:
			attr = m_hIt->get2f( attrAAT )[m_AttributeCompNumber];
			break;
		case 3:
			attr = m_hIt->get3f( attrAAT )[m_AttributeCompNumber];
			break;
		case 4:
			attr = m_hIt->get4f( attrAAT )[m_AttributeCompNumber];
			break;
		}

		float magnitude = 1.0f;//exp( -pow(dist,2.0f) / h_2 ) * scale;
		summed_magnitude += magnitude;
		float binID = (attr - m_MinValue) / (m_MaxValue - m_MinValue) * (float)m_NumBins;
		int bin1 = floor(binID);
		int bin2 = ceil(binID);
		float binInterp = binID - floor(binID);

		// use linear interpolation in radial direction 
		float hCoord = dist / (float)size * (m_NumHistograms) - 0.5f;
		if( hCoord <= 0.0f )
		{
			if( bin1 >= 0 && bin1 < (int)m_NumBins )
				sigdv[ bin1 ] += magnitude * (1.0f - binInterp);
			if( bin2 >= 0 && bin2 < (int)m_NumBins )
				sigdv[ bin2 ] += magnitude * binInterp;
			//ringWeights[0] += magnitude;
		}
		else
		{
			int h1 = floor(hCoord);
			int h2 = ceil(hCoord);
			float hInterp = hCoord - floor(hCoord);
			if( h1 < (int)m_NumHistograms )
			{
				if( bin1 >= 0 && bin1 < (int)m_NumBins )
					sigdv[ h1 * m_NumBins + bin1] += magnitude * (1.0f - binInterp) * (1.0f - hInterp);
				if( bin2 >= 0 && bin2 < (int)m_NumBins )
					sigdv[ h1 * m_NumBins + bin2] += magnitude * binInterp * (1.0f - hInterp);
			}
			if( h2 < (int)m_NumHistograms )
			{
				if( bin1 >= 0 && bin1 < (int)m_NumBins )
					sigdv[ h2 * m_NumBins + bin1] += magnitude * (1.0f - binInterp) * hInterp;
				if( bin2 >= 0 && bin2 < (int)m_NumBins )
					sigdv[ h2 * m_NumBins + bin2] += magnitude * binInterp * hInterp;
			}
		}
		if(!m_UseTopology)
			m_hIt->next();

		if( m_UseTopology && currentPoint++ >= m_TopQuery->getNumPoints())
			break;
	}

	float inv_summed_magnitude = 1.0f;
	if( summed_magnitude > 1e-10f )
		inv_summed_magnitude = 1.0f / summed_magnitude;
	for( card32 i=0;i<sigdv.getDim();i++ )
		sigdv[i] *= inv_summed_magnitude;


	FloatVectorSignature * sig = new FloatVectorSignature;
	sig->setVector(sigdv);

	return sig;
}

void X4::LDHistogram::endSignatureComputations()
{
	delete m_hIt;
	m_hIt = NULL;
	delete m_TopQuery;
	m_TopQuery = NULL;
}