#include "StdAfx.h"
#include <windows.h>
#include <mmsystem.h>
//---------------------------------------------------------------------------
#include "LineFeature.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "UnstructuredInCorePointCloud.h"
#include "FastSphereQuerry.h"
#include "HierarchicalKNNIterator.h"
#include "PointCloudTools.h"
#include "Random.h"
#include "Timer.h"
#include "StringHelper.h"
#include "CachedRangeQuerry.h"
#include "ObjectListProperty.h"
#include "DebugRenderer.h"
#include "GeometricTools.h"
#include "Monomials2ndOrder.h"
#include "ProgressWindow.h"
#include "EventCounterDisplay.h"
#include "SeparatorClassProperty.h"
#include "Monomials3rdOrder.h"
#include "Statistics.h"
#include "NormalDistribution.h"
#include <set>
//---------------------------------------------------------------------------
using namespace X4;


IMPLEMENT_X4_CLASS( LineFeature ,3)
{
	BEGIN_CLASS_INIT( LineFeature );
	ADD_VECTOR3F_PROP( m_Position ,0 );
	ADD_VECTOR3F_PROP( m_Normal ,0 );
	ADD_VECTOR3F_PROP( m_LineDirection ,0 );
	ADD_VECTOR3F_PROP( m_RotationDirection, 0 );
	ADD_FLOAT32_PROP( m_Curvature, 0);
	ADD_FLOAT32_PROP( m_Curvature2, 2);

	ADD_FLOAT32_PROP( m_Sigma, 1);
	ADD_FLOAT32_PROP( m_Length, 3);
}
IMPLEMENT_X4_CLASS( LineFeatureSet ,1)
{
	BEGIN_CLASS_INIT( LineFeatureSet );
	ADD_OBJECT_LIST_PROP( Feature, 0, LineFeature::getClass() );
	ADD_CARD32_VAR_ARRAY_PROP( m_ClusterID, 1);
}

IMPLEMENT_X4_CLASS( LineFeatureDetector ,6)
{

	
	BEGIN_CLASS_INIT( LineFeatureDetector );
	ADD_FLOAT32_PROP( m_Sigma, 0);
	ADD_FLOAT32_PROP( m_MinEigenRatio, 0);
	ADD_CARD32_PROP( m_MinNumberOfPoints, 0);
	ADD_FLOAT32_PROP( m_SamplingFactor, 1);
	ADD_FLOAT32_PROP( m_MaxContrainedFactor, 2);
	ADD_SEPARATOR("ooc threading");
	ADD_CARD32_PROP(m_Threading_NumThreads,6);
	ADD_SEPARATOR("point features");
	ADD_FLOAT32_PROP(m_PF_MinGaussianCurv,3);
	SET_USER_EDIT(false);
	ADD_FLOAT32_PROP(m_PF_MinSmallestEigen,4);
	SET_USER_EDIT(false);
	ADD_CARD32_PROP(m_PF_MinClustersize,5);
	SET_USER_EDIT(false);
}
LineFeature::LineFeature(void)
{
	m_Curvature2 = 0;
	m_Length = 0;
}

LineFeature::~LineFeature(void)
{
}

void LineFeature::assign(const Object* obj, X4CopyContext *context)
{
    const LineFeature * o = dynamic_cast<const LineFeature*>(obj);
    x4Assert( o != NULL );

    copyObjectProperties(obj,this);
}

void LineFeatureDetector::assign( const Object* obj, X4CopyContext *context /*= NULL */ )
{
    const LineFeatureDetector * o = dynamic_cast<const LineFeatureDetector*>(obj);
    x4Assert( o != NULL );

    copyObjectProperties(obj,this);
}

LineFeatureDetector::LineFeatureDetector()
{
    m_Sigma = 0.25f;
    m_MinEigenRatio = 2.0f;
    m_MinNumberOfPoints = 40;
    m_SamplingFactor = 1.0f;
    m_MaxContrainedFactor = 0.01f;
    m_PF_MinGaussianCurv = 0.01f;
    m_PF_MinSmallestEigen = 0.2f;
    m_PF_MinClustersize = 4;

    m_Querry = NULL;
    m_hIt = NULL;
    //m_VisitedID = 1;
    m_CachedQuerry = NULL;

    m_ThreadPool = NULL;
    m_Mutex = new QMutex;
    m_Threading_NumThreads = 0;
#if defined(WIN32) || defined(WIN64)
    SYSTEM_INFO    si;
    GetSystemInfo(&si);
    m_Threading_NumThreads = si.dwNumberOfProcessors;
#endif
}

void LineFeatureDetector::begin( UnstructuredInCorePointCloud * pc )
{
    m_Querry = new FastSphereQuerry(pc);
    //m_Visited = std::vector<int>(pc->getNumPoints(),0);
    m_PC = pc;
    m_hIt = new HierarchicalKNNIterator(pc,32,NULL);
    m_MedianPointDistance = getMedianPointDistance(pc);
    m_PosAAT = pc->getAAT("position");
    m_NormalAAT = pc->getAAT("normal");
    m_FlagsAAT = pc->getAAT("flags");
    m_CachedQuerry = new CachedRangeQuerry(pc);
    m_CacheOverlap = m_MedianPointDistance * 2.7f;
}

void LineFeatureDetector::end()
{
    PointSet * ps = m_PC->getPointSet();
    for( unsigned i=0;i<ps->getNumEntries();i++ )
        ps->set1i(i,m_FlagsAAT,0);

    delete m_Querry;
    m_Querry = NULL;
    delete m_hIt;
    m_hIt = NULL;
    delete m_CachedQuerry;
    m_CachedQuerry = NULL;
}

LineFeature * LineFeatureDetector::extractLineFeature( const Vector3f &pos )
{

    // compute slippage analysis
    Matrix6f eigenVectors;
    Vector6f eigenValues;
    if( !computeSlippageAnalysis(pos, m_Sigma, eigenVectors, eigenValues, true) )
        return NULL;

    // check for 1d slippage
    float div_eps = 0.001f;
    float eigenRatio = eigenValues[4] / (eigenValues[5] + div_eps);
    if( eigenRatio < m_MinEigenRatio || eigenValues[5] / eigenValues[0] > m_MaxContrainedFactor)
        return NULL;

    // fix coordinate frame
    AAT normalAAT = m_PC->getAAT("normal");
    PointSet * ps = m_PC->getPointSet();
    m_hIt->setSeekPointAndReset(pos);
    /*Matrix3f pcaFrame;
    Vector3f center;
    computeLocalFrame( pos, m_Sigma, m_PC, m_hIt, pcaFrame, center, true );*/
    Vector3f normal = ps->get3f(m_hIt->getCurrentPointIndex(), normalAAT);
    Vector3f slippageDirection = normalize(makeVector3f(eigenVectors[5][3],eigenVectors[5][4],eigenVectors[5][5]));






    /*	1. fix frame (pca)
    2. adapt curv
    3. brent*/










    //Vector3f r = pcaFrame[2].crossProduct(slippageDirection);
    Vector3f r = normal.crossProduct(slippageDirection);




    /*Matrix3f coordinateFrame;
    coordinateFrame[0] = normalize(r);
    coordinateFrame[1] = slippageDirection;
    coordinateFrame[2] = normal;
    Matrix3f worldToFrame = coordinateFrame.transpose();









    mpcard * indices;
    mpcard numPoints;
    float scaleFactor = m_Sigma * 2.5f;
    m_Querry->querry(pos, scaleFactor, &indices, numPoints);
    if( numPoints < m_MinNumberOfPoints)
    {
    return NULL;
    }

    std::vector<Vector3f> points(numPoints);
    std::vector<float> weights(numPoints);
    AAT posAAT = ps->getAAT("position");
    Matrix3f scaleElipsMat = makeScale3f(1.0f,4.0f,1.0f) / 2;
    for( unsigned i=0;i<numPoints;i++ )
    {
    Vector3f p = worldToFrame * (ps->get3f(indices[i],posAAT) - pos) / scaleFactor;
    points[i] = p;
    weights[i] = exp( - normQuad(scaleElipsMat*p));
    }




















    Vector4f coeffs = Monomials3rdOrder1D::computeCoeffs( &points[0], &weights[0], numPoints);
    float optX;
    if( !Monomials3rdOrder1D::computeCurvatureMinMax(coeffs, optX))
    {

    return NULL;
    }


    Vector3f optPos = coordinateFrame * makeVector3f(optX,0,Monomials3rdOrder1D::evalate(coeffs,optX)) * scaleFactor + pos;
    if( norm(optPos- pos) > m_Sigma )
    return NULL;


    Matrix6f tmpEigenVectors;
    Vector6f tmpEigenValues;
    computeSlippageAnalysis(optPos, m_Sigma, tmpEigenVectors, tmpEigenValues,false);


    LineFeature * ret = new LineFeature;
    ret->m_Position = optPos;
    ret->m_InitialPosition = pos;
    ret->m_LineDirection = makeVector3f(tmpEigenVectors[5][3],tmpEigenVectors[5][4],tmpEigenVectors[5][5]);
    ret->m_RotationDirection = makeVector3f(tmpEigenVectors[5][0],tmpEigenVectors[5][1],tmpEigenVectors[5][2]);
    ret->m_Sigma = m_Sigma;

    // compute normal
    Matrix3f nframe;
    Vector3f ncenter;
    computeLocalFrame( ret->m_Position, m_Sigma, m_PC, m_hIt, nframe, ncenter,true);
    ret->m_Normal = nframe[2];

    float H,K;
    computeCurvatures(optPos,ret->m_Normal,m_Sigma, H, K);
    ret->m_Curvature = H;
    computeCurvatures(optPos,ret->m_Normal,m_Sigma/sqrt(2.0f), H, K);
    ret->m_Curvature2 = H;

    return ret;	

    /**/
    // optimize
    float projSigma = m_MedianPointDistance;
    float tl = -m_Sigma;
    float tr = m_Sigma;
    float tm = 0;
    Matrix6f tmpEigenVectors;
    Vector6f tmpEigenValues;




    Vector3f bestPos = pos;
    float summedWeight = 0;
    Vector3f pos_left = projectPoint(bestPos - r * m_MedianPointDistance, projSigma, slippageDirection, summedWeight  );
    if( summedWeight < 0.001f )
        return NULL;
    Vector3f pos_right = projectPoint(bestPos + r * m_MedianPointDistance, projSigma, slippageDirection, summedWeight  );
    if( summedWeight < 0.001f )
        return NULL;
    float score_l = computeScore( pos_left, m_Sigma);
    float score_r = computeScore( pos_right, m_Sigma);
    eigenRatio = computeScore( bestPos, m_Sigma);

    for( unsigned i=0;i<15;i++ )
    {

        if( eigenRatio >= score_l && eigenRatio >= score_r )
        {

            bestPos = (bestPos * eigenRatio + pos_left * score_l + pos_right * score_r) / (eigenRatio + score_l + score_r);
            bestPos = projectPoint(bestPos, projSigma, slippageDirection, summedWeight);
            if( summedWeight < 0.001f )
                return NULL;
            break;
        }
        else
        {




            if( score_l > score_r )
            {

                Vector3f newPos = projectPoint(pos_left - r * m_MedianPointDistance, projSigma, slippageDirection, summedWeight );
                if( summedWeight < 0.001f )
                    return NULL;
                score_r = eigenRatio;
                eigenRatio = score_l;
                pos_right = bestPos;
                bestPos = pos_left;
                pos_left = newPos;
                score_l = computeScore(pos_left, m_Sigma);
            }
            else
            {



                Vector3f newPos = projectPoint(pos_right + r * m_MedianPointDistance, projSigma, slippageDirection, summedWeight  );
                if( summedWeight < 0.001f )
                    return NULL;
                score_l = eigenRatio;
                eigenRatio = score_r;
                pos_left = bestPos;
                bestPos = pos_right;
                pos_right = newPos;
                score_r = computeScore(pos_right, m_Sigma);
            }
        }


        if( norm(bestPos - pos) > m_Sigma )
            return NULL;

    }


    computeSlippageAnalysis(bestPos, m_Sigma, tmpEigenVectors, tmpEigenValues,false);


    LineFeature * ret = new LineFeature;
    ret->m_Position = bestPos;
    ret->m_InitialPosition = pos;
    ret->m_LineDirection = makeVector3f(tmpEigenVectors[5][3],tmpEigenVectors[5][4],tmpEigenVectors[5][5]);
    ret->m_RotationDirection = makeVector3f(tmpEigenVectors[5][0],tmpEigenVectors[5][1],tmpEigenVectors[5][2]);
    ret->m_Sigma = m_Sigma;

    // compute normal
    Matrix3f nframe;
    Vector3f ncenter;
    computeLocalFrame( ret->m_Position, m_Sigma, m_PC, m_hIt, nframe, ncenter,true);
    ret->m_Normal = nframe[2];

    float H,K;
    computeCurvatures(bestPos,ret->m_Normal,m_Sigma, H, K);
    ret->m_Curvature = H;
    computeCurvatures(bestPos,ret->m_Normal,m_Sigma/sqrt(2.0f), H, K);
    ret->m_Curvature2 = H;

    return ret;
}

LineFeature * LineFeatureDetector::extractLineFeature_old( const Vector3f &pos )
{
    // compute slippage analysis
    Matrix6f eigenVectors;
    Vector6f eigenValues;
    if( !computeSlippageAnalysis(pos, m_Sigma, eigenVectors, eigenValues, true) )
        return NULL;

    // check for 1d slippage
    float div_eps = 0.001f;
    float eigenRatio = eigenValues[4] / (eigenValues[5] + div_eps);
    if( eigenRatio < m_MinEigenRatio )
        return NULL;

    // fix coordinate frame
    AAT normalAAT = m_PC->getAAT("normal");
    PointSet * ps = m_PC->getPointSet();
    m_hIt->setSeekPointAndReset(pos);
    Vector3f normal = ps->get3f(m_hIt->getCurrentPointIndex(), normalAAT);
    Vector3f slippageDirection = normalize(makeVector3f(eigenVectors[5][3],eigenVectors[5][4],eigenVectors[5][5]));
    Vector3f r = normal.crossProduct(slippageDirection);

    // optimize
    float projSigma = m_MedianPointDistance;
    float tl = -m_Sigma;
    float tr = m_Sigma;
    float tm = 0;
    Matrix6f tmpEigenVectors;
    Vector6f tmpEigenValues;



    Vector3f bestPos = pos;
    float summedWeight = 0;
    Vector3f pos_left = projectPoint(bestPos - r * m_MedianPointDistance, projSigma, slippageDirection, summedWeight  );
    if( summedWeight < 0.001f )
        return NULL;
    Vector3f pos_right = projectPoint(bestPos + r * m_MedianPointDistance, projSigma, slippageDirection, summedWeight  );
    if( summedWeight < 0.001f )
        return NULL;
    float score_l = computeScore( pos_left, m_Sigma);
    float score_r = computeScore( pos_right, m_Sigma);
    eigenRatio = computeScore( bestPos, m_Sigma);

    for( unsigned i=0;i<15;i++ )
    {

        if( eigenRatio >= score_l && eigenRatio >= score_r )
        {

            bestPos = (bestPos * eigenRatio + pos_left * score_l + pos_right * score_r) / (eigenRatio + score_l + score_r);
            bestPos = projectPoint(bestPos, projSigma, slippageDirection, summedWeight);
            if( summedWeight < 0.001f )
                return NULL;
            break;
        }
        else
        {




            if( score_l > score_r )
            {

                Vector3f newPos = projectPoint(pos_left - r * m_MedianPointDistance, projSigma, slippageDirection, summedWeight );
                if( summedWeight < 0.001f )
                    return NULL;
                score_r = eigenRatio;
                eigenRatio = score_l;
                pos_right = bestPos;
                bestPos = pos_left;
                pos_left = newPos;
                score_l = computeScore(pos_left, m_Sigma);
            }
            else
            {



                Vector3f newPos = projectPoint(pos_right + r * m_MedianPointDistance, projSigma, slippageDirection, summedWeight  );
                if( summedWeight < 0.001f )
                    return NULL;
                score_l = eigenRatio;
                eigenRatio = score_r;
                pos_left = bestPos;
                bestPos = pos_right;
                pos_right = newPos;
                score_r = computeScore(pos_right, m_Sigma);
            }
        }


        if( norm(bestPos - pos) > m_Sigma )
            return NULL;

    }


    computeSlippageAnalysis(bestPos, m_Sigma, tmpEigenVectors, tmpEigenValues,false);


    LineFeature * ret = new LineFeature;
    ret->m_Position = bestPos;
    ret->m_InitialPosition = pos;
    ret->m_LineDirection = makeVector3f(tmpEigenVectors[5][3],tmpEigenVectors[5][4],tmpEigenVectors[5][5]);
    ret->m_RotationDirection = makeVector3f(tmpEigenVectors[5][0],tmpEigenVectors[5][1],tmpEigenVectors[5][2]);
    ret->m_Sigma = m_Sigma;

    // compute normal
    Matrix3f nframe;
    Vector3f ncenter;
    computeLocalFrame( ret->m_Position, m_Sigma, m_PC, m_hIt, nframe, ncenter,true);
    ret->m_Normal = nframe[2];

    float H,K;
    computeCurvatures(bestPos,ret->m_Normal,m_Sigma, H, K);
    ret->m_Curvature = H;
    computeCurvatures(bestPos,ret->m_Normal,m_Sigma/sqrt(2.0f), H, K);
    ret->m_Curvature2 = H;

    return ret;
}

bool LineFeatureDetector::computeSlippageAnalysis( const Vector3f &pos, float sigma, Matrix6f &eigenVectors, Vector6f &eigenValues, bool markVisited )
{
    // querry points in sphere
    mpcard * indices;
    mpcard numPoints;
    //std::vector<Vector3f> * positions = m_CachedQuerry->getSpherePositions(pos, m_Sigma * 2.0f, 0,10,&indices);
    //numPoints = positions->size();
    m_Querry->querry( pos, sigma * 2.0f, &indices, numPoints);	


    if( numPoints < m_MinNumberOfPoints )
        return false;	// not enough points

    // setup access variables
    PointSet * ps = m_PC->getPointSet();

    // setup matrices
    static Matrix3f zeroMat3f = IDENTITY3F * 0;
    Matrix3f mat_upperleft = zeroMat3f;
    Matrix3f mat_lowerright = zeroMat3f;
    Matrix3f mat_upperright = zeroMat3f;



    // compute matrices
    float sigma_2 = sigma * sigma;
    float markAsVisitedRadius_2 = pow(sigma * m_SamplingFactor, 2.0f);
    float inv_sigma_2 = 1.0f / sigma_2;
    float summedWeight = 0;
    for( unsigned i=0;i<numPoints;i++ )
    {
        mpcard index = *indices++;
        Vector3f p = ps->get3f(index,m_PosAAT) - pos;				
        Vector3f n = ps->get3f(index,m_NormalAAT);





        float dist_2 = normQuad(p);
        float weight = //exp(-dist_2 * inv_sigma_2);
            NormalDistribution::expApproximation(dist_2 * inv_sigma_2);




        summedWeight += weight;











        // mark as visited
        if( markVisited && dist_2 < markAsVisitedRadius_2 )
            //m_Visited[index] = m_VisitedID;
                ps->set1i(index,m_FlagsAAT, PF_MARKED_FLAG);

        Vector3f c = p.crossProduct(n) / m_Sigma;
        mat_upperleft += outerProduct(c,c) * weight;
        mat_lowerright += outerProduct(n,n) * weight;
        mat_upperright += outerProduct(n,c) * weight;

    }


    // compute slippage analysis
    Matrix6f covariance;
    for( unsigned m=0;m<3;m++ )
    {

        for( unsigned j=0;j<3;j++ )
        {

            covariance[m][j] = mat_upperleft[m][j];
            covariance[m+3][j+3] = mat_lowerright[m][j];
            covariance[m+3][j] = mat_upperright[m][j];
            covariance[m][j+3] = mat_upperright[j][m];
        }
    }



    if( summedWeight > 0.0f)
        covariance /= summedWeight;
    covariance.computeEigenStructure(eigenValues,eigenVectors);


    return true;
}

LineFeatureSet * LineFeatureDetector::extractFeatures(bool attachToPointCloud)
{
    if( m_Threading_NumThreads == 0 )
    {

        debugRenderer->clearDebugData("testn");
        debugRenderer->beginRenderJob("testn");
    }

    //X4_TIMER_START( FeatureExtraction )
    LineFeatureSet * ret = NULL;

    if( attachToPointCloud)
    {
        ret = dynamic_cast<LineFeatureSet*>(m_PC->getAttachments()->getData(LineFeatureSet::getDefaultName()));
        if( ret )
        {
            delete ret;
        }
    }
    ret = new LineFeatureSet;
    ret->setup( LineFeatureSet::getDefaultName(), AttachedData::ADF_PERSISTENT, NULL);
    if( attachToPointCloud)
        m_PC->getAttachments()->attachData(ret);


















    // setup access variables
    AAT posAAT = m_PC->getAAT("position");
    PointSet * ps = m_PC->getPointSet();

    UnstructuredInCoreTriangleMesh * mesh = dynamic_cast<UnstructuredInCoreTriangleMesh*>(m_PC);
    /*if( mesh )
    {
    extractFeatures_TriangleMesh(mesh,ret);
    }
    else*/
    {
        if( m_Threading_NumThreads == 0 )
            progressWindow->pushStep();
        for( unsigned i=0;i<ps->getNumEntries();i++ )
        {
            if( m_Threading_NumThreads == 0 )
            {
                if( i % 100 )
                {
                    progressWindow->progress(i*100 / ps->getNumEntries());
                }
            }
            if( ps->get1i(i,m_FlagsAAT) & PF_MARKED_FLAG)
                continue;




            Vector3f pos = ps->get3f(i,posAAT);
            //stats->startAverageTimer("extractLineFeature()");
            LineFeature * fl = extractLineFeature(pos);
            //stats->stopAverageTimer("extractLineFeature()");
            if( fl )
                ret->m_LineFeatures.push_back(fl);
            else
                ret->m_RejectedSamples.push_back(pos);
        }











        if( m_Threading_NumThreads == 0 )



            progressWindow->popSteps();
    }














    if( m_Threading_NumThreads == 0 )




        debugRenderer->endRenderJob();
    if( m_Threading_NumThreads == 0 )
        debugOutput << "Line Feature Detected: " << ret->m_LineFeatures.size() << "\n";
    //XGRT_TIMER_END( FeatureExtraction )
    return ret;
}

Vector3f LineFeatureDetector::projectPoint( const Vector3f &pos, float sigma, const Vector3f slippageDirection, float &summedWeight )
{
    float sigma_2 = sigma * sigma;
    float inv_sigma_2 = 1.0f / sigma_2;
    Vector3f projectedPos = NULL_VECTOR3F;
    summedWeight = 0;
    mpcard numPoints = 0;
    mpcard * indices;

    m_Querry->querry(pos,sigma,&indices,numPoints);
    if( numPoints > 5 )
    {
        PointSet * ps = ((UnstructuredInCorePointCloud*)m_PC)->getPointSet();
        for( unsigned i=0;i<numPoints;i++ )
        {
            Vector3f tmpPos = ps->get3f(indices[i],m_PosAAT);
            //Vector3f diff = pos - tmpPos;
            float dist_2 = m_Querry->getSquaredDistance(i);

            float weight = //exp(-dist_2 * inv_sigma_2);
                NormalDistribution::expApproximation(dist_2 * inv_sigma_2);
            Vector3f tmpNormal = ps->get3f(indices[i],m_NormalAAT);

            Vector3f lineDir = tmpNormal.crossProduct(slippageDirection);	// intersection of surface and optimization plane
            float t = (pos - tmpPos) * lineDir / (lineDir * lineDir);
            Vector3f intersectionPoint = tmpPos + lineDir * t;


            projectedPos += intersectionPoint * weight;
            summedWeight += weight;
        }
        return projectedPos / summedWeight;










    }




    numPoints = 0;
    m_hIt->setSeekPointAndReset(pos);
    while( !m_hIt->atEnd())
    {
        Vector3f tmpPos = m_hIt->get3f(m_PosAAT);
        Vector3f diff = pos - tmpPos;
        float dist_2 = normQuad(diff);
        if( dist_2 > sigma_2 && numPoints > 5)
            break;









        numPoints++;
        float weight = exp(-dist_2 * inv_sigma_2);
        Vector3f tmpNormal = m_hIt->get3f(m_NormalAAT);


        //projectedPos += (tmpPos - tmpNormal * (tmpNormal * diff)) * weight;










        Vector3f lineDir = tmpNormal.crossProduct(slippageDirection);	// intersection of surface and optimization plane
        float t = (pos - tmpPos) * lineDir / (lineDir * lineDir);
        Vector3f intersectionPoint = tmpPos + lineDir * t;










        projectedPos += intersectionPoint * weight;
        summedWeight += weight;

        m_hIt->next();
    }


    return projectedPos / summedWeight;
}

float LineFeatureDetector::computeScore( const Vector3f &pos, float sigma )
{
    /*		

    Matrix6f tmpEigenVectors;
    Vector6f tmpEigenValues;
    computeSlippageAnalysis(pos, sigma, tmpEigenVectors, tmpEigenValues,false);
    float ratio = tmpEigenValues[4] / (tmpEigenValues[5] + 0.001f);

    return ratio;*/

    float H,K;
    //computeCurvature_PatchFitting(m_PC, pos, sigma, H, K);
    computeCurvatures(pos, makeVector3f(1,0,0),sigma, H, K);

    //	debugOutput << fabs(H) << "\n";

    return fabs(H);//fabs(H + sqrt( H*H - K));
}

void LineFeatureDetector::computeCurvatures( const Vector3f &pos, const Vector3f &normal, float sigma, float &meanCurvature, float &gaussianCurvature )
{
    //std::vector<Vector3f> * positionCache = m_CachedQuerry->getSpherePositions(pos, 2.0f * sigma, 2.7f * m_MedianPointDistance);
    //const unsigned numPositions = positionCache->size();
    mpcard numPositions;
    mpcard * indices;
    m_Querry->querry(pos,2.0f*sigma,&indices,numPositions);

    while( m_CompCurv_PositionCache.size() < numPositions )
    {
        m_CompCurv_PositionCache.push_back( NULL_VECTOR3F );
        m_CompCurv_TransformedPositions.push_back(NULL_VECTOR3F);
        m_CompCurv_Weights.push_back(0);
    }
    Vector3f * points = &(m_CompCurv_PositionCache[0]);

    PointSet * ps = m_PC->getPointSet();










    float inv_sigma_2 = 1.0f / sigma / sigma;
    float rad = 2.0f * sigma;
    float rad_2 = sigma * sigma * 4.0f;

    // compute local frame via pca
    PCA<float,3> pca;
    //pca.clear();
    Vector3f * points_ptr = points;
    for( unsigned i=0;i<numPositions;i++ )
    {
        *points_ptr = ps->get3f( indices[i], m_PosAAT );
        float dist_2 = m_Querry->getSquaredDistance(i);//normQuad(*points_ptr - pos);
        /*if( dist_2 > rad_2 )
        {
        points_ptr++;
        continue;
        }*/




        float weight = //exp( -dist_2 * inv_sigma_2);
            NormalDistribution::expApproximation(dist_2*inv_sigma_2);
        m_CompCurv_Weights[i] = weight;







        pca.addPoint(*points_ptr,weight);


        points_ptr++;
    }



    Matrix3f eigenVectors;
    Vector3f eigenValues, centroid;
    pca.analyze(eigenValues, eigenVectors, centroid);







    if( eigenVectors[2] * normal < 0.0f )


        eigenVectors[2] *= -1;

    Matrix3f frame = eigenVectors.transpose() / rad;
    points_ptr = points;
    for( unsigned i=0;i<numPositions;i++ )
    {
        m_CompCurv_TransformedPositions[i] = frame * (*points_ptr++ - pos);
    }



    Vector6f coeff;
    Monomials2ndOrder::computeCoeffs( &m_CompCurv_TransformedPositions[0], &m_CompCurv_Weights[0], numPositions, coeff);










    float f_x = coeff[1];
    float f_y = coeff[2];
    float f_xy = coeff[4];
    float f_xx = 2.0f * coeff[3];
    float f_yy = 2.0f * coeff[5];

    gaussianCurvature = (f_xx*f_yy - pow( f_xy, 2.0f ))
        / pow(1.0f + pow(f_x,2.0f) + pow(f_y,2.0f),2.0f );

    meanCurvature = ((1.0f + pow(f_x,2.0f))*f_yy - 2.0f * f_x * f_y * f_xy + (1.0f + pow(f_y,2.0f))*f_xx)
        / pow(1.0f + pow(f_x,2.0f) + pow(f_y,2.0f),2.0f/3.0f ) / 2.0f;



}

class LineFeatureDetector_Comp {
public:
    static bool equals(card32 obj1, card32 obj2) { return obj1 == obj2; }
};

class LineFeatureDetector_MyHashTable : public HashTable<card32, Vector3f, LineFeatureDetector_Comp> {
protected:
    virtual max_platform_card hashFunction(card32 key) const {
        return key;
    };

public:
    LineFeatureDetector_MyHashTable(card32 fillFactor = 1, card32 startSizeIndex = 8) : HashTable<card32, Vector3f, LineFeatureDetector_Comp>(fillFactor, startSizeIndex) {}

};


void LineFeatureDetector::extractFeatures_TriangleMesh( UnstructuredInCoreTriangleMesh * mesh, LineFeatureSet * lf_set )
{
    UnstructuredInCoreTriangleMesh * meshCopy = (UnstructuredInCoreTriangleMesh*)mesh->copy();
    meshCopy->closeMesh();
    AAT posAAT_mesh = mesh->getAAT("position");
    AAT indexAAT = mesh->getTriangles()->getAAT("index");

    LineFeatureDetector_MyHashTable hashTable;

    for( unsigned i=0;i<meshCopy->getNumTriangles();i++ )
    {
        Vector3i triangle	= meshCopy->getTriangles()->get3i(i,indexAAT);
        Vector3f p1 = meshCopy->getPointSet()->get3f(triangle[0],posAAT_mesh);
        Vector3f p2 = meshCopy->getPointSet()->get3f(triangle[1],posAAT_mesh);
        Vector3f p3 = meshCopy->getPointSet()->get3f(triangle[2],posAAT_mesh);
        Vector3f normal = normalize( (p2-p1).crossProduct(p3-p1));
        for( unsigned j=0;j<3;j++ )
        {
            unsigned index1 = triangle[j];
            unsigned index2 = triangle[(j+1)%3];
            if( index1 > index2 )
            {
                unsigned tmp = index1;
                index1 = index2;
                index2 = tmp;
            }

            card32 hashKey = index1 * meshCopy->getNumPoints() + index2;
            Vector3f normal1;
            if( hashTable.getElement(hashKey,normal1) )
            {
                hashTable.deleteElement(hashKey);
                if( normal * normal1 < cos(30.0f/180.0f*M_PI) )
                {
                    Vector3f pos1 = meshCopy->getPointSet()->get3f(index1,posAAT_mesh);
                    Vector3f pos2 = meshCopy->getPointSet()->get3f(index2,posAAT_mesh);
                    float l = norm(pos2-pos1);
                    card32 numSamples = (l / m_Sigma)*(rnd01()+1.0f);


                    for( unsigned k=0;k<numSamples;k++ )
                    {
                        LineFeature * lf = new LineFeature;
                        float r = rnd01();
                        lf->m_Position = pos1 * (1-r) + pos2*r;
                        lf->m_Normal = normalize(normal + normal1);
                        lf->m_LineDirection = normalize(pos2-pos1);
                        lf->m_Sigma = m_Sigma;
                        lf->m_RotationDirection = NULL_VECTOR3F;
                        lf->m_InitialPosition = lf->m_Position;
                        lf->m_Curvature = normal * normal1;
                        lf->m_Curvature2 = lf->m_Curvature;
                        lf_set->m_LineFeatures.push_back(lf);
                    }





























                }
            }
            else
            {
                hashTable.addElement( hashKey, normal);
            }






















        }
    }






    // create a pointcloud of barycenters
    /*UnstructuredInCorePointCloud upc;
    upc.clearAndSetup(createDefaultVertexDescriptor(true,false,false,false),mesh->getNumTriangles());
    AAT posAAT_upc = upc.getAAT("position");
    for( unsigned i=0;i<mesh->getNumTriangles();i++ )
    {

    Vector3i triangle = mesh->getTriangles()->get3i(i,indexAAT);
    Vector3f p1 = mesh->getPointSet()->get3f(triangle[0],posAAT_mesh);
    Vector3f p2 = mesh->getPointSet()->get3f(triangle[1],posAAT_mesh);
    Vector3f p3 = mesh->getPointSet()->get3f(triangle[2],posAAT_mesh);
    Vector3f baryCenter = (p1+p2+p3) / 3.0f;
    upc.getPointSet()->set3f(i,posAAT_upc,baryCenter);
    }

    HierarchicalKNNIterator hIt_upc(&upc,32,NULL);

    // find edge with high curvature
    for( unsigned i=0;i<upc.getNumPoints();i++ )
    {

    // gather neighboring triangles
    Vector3f currentTriangleBarycenter = upc.getPointSet()->get3f(i,posAAT_upc);
    Vector3i triangle = mesh->getTriangles()->get3i(i,indexAAT);
    Vector3f p1 = mesh->getPointSet()->get3f(triangle[0],posAAT_mesh);
    Vector3f p2 = mesh->getPointSet()->get3f(triangle[1],posAAT_mesh);
    Vector3f p3 = mesh->getPointSet()->get3f(triangle[2],posAAT_mesh);

    hIt_upc.setSeekPointAndReset(currentTriangleBarycenter);
    hIt_upc.next();

    while(!hIt_upc.atEnd())
    {

    Vector3f n_bc = hIt_upc.get3f(posAAT_upc);



    hIt_upc.next();
    }

    }*/

}

void LineFeatureDetector::extractPointFeatures( LineFeatureSet * lf_set )
{
    if( lf_set == NULL || lf_set->m_ClusterID.size() == 0)
        return;


    clearMarkedFlag(m_PC);

    // count cluster elements
    std::map<unsigned,unsigned> clusterSize;
    for( unsigned i=0;i<lf_set->m_ClusterID.size();i++ )
    {
        clusterSize[ lf_set->m_ClusterID[i] ]++;
    }

    SceneObjectIterator * iter = m_PC->createIterator(SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC);
    BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
    ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(iter);

    SceneObjectIterator * iter2 = m_PC->createIterator(SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC | SOIT::CAP_RANDOM_ACCESS_PC);
    BasicPointCloudIterator * bIt2 = dynamic_cast<BasicPointCloudIterator*>(iter2);
    ModifyablePointCloudIterator * mIt2 = dynamic_cast<ModifyablePointCloudIterator*>(iter2);
    RandomAccessPointCloudIterator * rIt2 = dynamic_cast<RandomAccessPointCloudIterator*>(iter2);


    UnstructuredInCorePointCloud * upc_lines = lf_set->getFeaturesAsPointCloud();
    HierarchicalKNNIterator hItLines(upc_lines,32,NULL);







    float mark_rad_2 = pow(m_Sigma * m_SamplingFactor,2.0f);
    float inv_line_sigma_2 = 1.0f / pow(m_Sigma,2.0f);
    float line_rad_2 = pow(2.0f * m_Sigma,2.0f);
    debugRenderer->clearDebugData("testPF");
    debugRenderer->beginRenderJob("testPF");
    while( !bIt->atEnd())
    {




        if( (bIt->get1i(m_FlagsAAT) & PF_MARKED_FLAG) == 0 )
        {
            Vector3f pos = bIt->get3f(m_PosAAT);
            Vector3f normal = bIt->get3f(m_NormalAAT);





            // mark region & store points
            std::vector<Vector3f> points;
            m_hIt->setSeekPointAndReset(pos);
            while( !m_hIt->atEnd())
            {
                Vector3f p = m_hIt->get3f(m_PosAAT);
                if( normQuad(p-pos) > mark_rad_2 )
                    break;

                points.push_back(p);
                mpcard pIndex = m_hIt->getCurrentPointIndex();
                rIt2->goToIndex(pIndex);
                mIt2->set1i(m_FlagsAAT, PF_MARKED_FLAG);
                m_hIt->next();
            }



            int numConvSteps = 0;
            Vector3f currentFeaturesPos = pos;
            bool validFeature = true;
            while( numConvSteps++ < 5 )
            {













                // get all neighboring lines & compute quadratic form
                Matrix3f M = IDENTITY3F * 0;
                Vector3f b = NULL_VECTOR3F;
                hItLines.setSeekPointAndReset(currentFeaturesPos);
                std::vector<LineFeature *> localSet;
















                //static int abc = 0;
                //bool visConstraints =(abc++ % 10) == 0;






                while( !hItLines.atEnd())
                {
                    unsigned lIndex = hItLines.getCurrentPointIndex();
                    LineFeature * lf = lf_set->m_LineFeatures[lIndex];
                    if( clusterSize[lf_set->m_ClusterID[lIndex]] < m_PF_MinClustersize )
                    {
                        hItLines.next();
                        continue;
                    }
                    const Vector3f p = lf->m_Position;
                    float dist_2 = normQuad(p-currentFeaturesPos);
                    if( dist_2 > line_rad_2 )
                    {
                        break;
                    }






                    Vector3f nearestPos,nearestPosLineDir;
                    lf->computeNearestCurvePoint(currentFeaturesPos, nearestPos, nearestPosLineDir);



                    //Vector3f lineDir = normalize(lf->m_LineDirection.crossProduct(lf->m_Normal));
                    //Matrix3f Q = outerProduct(lineDir,lineDir) + outerProduct(lf->m_Normal,lf->m_Normal);
                    Matrix3f Q = IDENTITY3F - outerProduct(nearestPosLineDir,nearestPosLineDir);
                    float weight = exp(-dist_2 * inv_line_sigma_2);
















                    //if( numConvSteps > 1 )
                    {
                        float tmp_dist_2 = (Q * (currentFeaturesPos - nearestPos))*(currentFeaturesPos - nearestPos);






                        weight *= exp( -tmp_dist_2 * inv_line_sigma_2 );
                    }

                    M += Q*weight;
                    //b -= Q*p*weight * 2.0f;
                    b -= Q*nearestPos*weight * 2.0f;






                    localSet.push_back(lf_set->m_LineFeatures[lIndex]);
                    hItLines.next();
                }




                if( localSet.size() < 5 )
                {
                    validFeature = false;
                    break;
                }




                // eigen analysis of M - reject bad constrained points
                Matrix3f eigenVectors;
                Vector3f eigenValues;
                M.computeEigenStructure( eigenValues, eigenVectors);
                //debugOutput << "eigenValues: " << eigenValues << "\n";
                if( sqrt(eigenValues[2]) / m_Sigma < m_PF_MinSmallestEigen )
                {


                    validFeature = false;
                    break;
                }



                bool inv_success;
                Matrix3f invM = invertMatrix(M,&inv_success);
                if( !inv_success )
                {







                    validFeature = false;
                    break;
                }



                Vector3f f = - (invM * b) / 2.0f;








                if( normQuad(f - pos) > mark_rad_2)
                {
                    validFeature = false;
                    break;
                }
                debugRenderer->addLine(currentFeaturesPos,f);
                currentFeaturesPos = f;				
            }


            if( validFeature )
            {





                //debugRenderer->addLine(pos,currentFeaturesPos,makeVector3f(1,1,1)*0.25f,makeVector3f(1,1,1)*0.25f,1);
                debugRenderer->addFastSphere(currentFeaturesPos,m_MedianPointDistance*2.5f,makeVector3f(1,0.5f,0),false);
            }











        }


        bIt->next();
    }


    debugRenderer->endRenderJob();
    delete iter;
    delete iter2;
}

void LineFeatureDetector::clearMarkedFlag( PointCloud * pc )
{
    AAT flagsAAT = pc->getAAT("flags");
    SceneObjectIterator * iter = pc->createIterator(SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC);
    BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
    ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(iter);

    while(!bIt->atEnd())
    {

        mIt->set1i(flagsAAT,0);
        bIt->next();
    }

    delete iter;
}

void LineFeatureDetector::debug_visualizeQuadricLines( const Vector3f &point, LineFeatureSet * lf_set )
{
    if( lf_set == NULL )// || lf_set->m_ClusterID.size() == 0)
        return;

    clearMarkedFlag(m_PC);


    UnstructuredInCorePointCloud * upc_lines = lf_set->getFeaturesAsPointCloud();
    HierarchicalKNNIterator hItLines(upc_lines,32,NULL);

    float line_rad_2 = pow(2.0f * m_Sigma,2.0f);
    debugRenderer->clearDebugData("visualizeQ");
    debugRenderer->beginRenderJob("visualizeQ");

    // get all neighboring lines & compute quadratic form
    Matrix3f M = IDENTITY3F * 0;
    Vector3f b = NULL_VECTOR3F;
    hItLines.setSeekPointAndReset(point);

    while( !hItLines.atEnd())
    {

        unsigned lIndex = hItLines.getCurrentPointIndex();
        LineFeature * lf = lf_set->m_LineFeatures[lIndex];
        const Vector3f p = lf->m_Position;
        float dist_2 = normQuad(p-point);
        if( dist_2 > line_rad_2 )
        {
            break;
        }




        Vector3f nearestPos,nearestPosLineDir;
        lf->computeNearestCurvePoint(point, nearestPos, nearestPosLineDir);
        nearestPosLineDir = normalize(nearestPosLineDir);

        //Vector3f lineDir = normalize(lf->m_LineDirection.crossProduct(lf->m_Normal));
        //Matrix3f Q = outerProduct(lineDir,lineDir) + outerProduct(lf->m_Normal,lf->m_Normal);
        Matrix3f Q = IDENTITY3F - outerProduct(nearestPosLineDir,nearestPosLineDir);
        float weight = 1;//exp(-dist_2 * inv_line_sigma_2);

        debugRenderer->addLine(nearestPos + nearestPosLineDir*lf->m_Sigma*1.2f,nearestPos - nearestPosLineDir*1.2f*lf->m_Sigma,
            makeVector3f(1,0,0),makeVector3f(1,0,0),2);

        M += Q*weight;
        //b -= Q*p*weight * 2.0f;
        b -= Q*nearestPos*weight * 2.0f;

        hItLines.next();
    }


    /*		if( localSet.size() < 5 )
    {

    validFeature = false;
    break;
    }


    // eigen analysis of M - reject bad constrained points
    Matrix3f eigenVectors;
    Vector3f eigenValues;
    M.computeEigenStructure( eigenValues, eigenVectors);
    //debugOutput << "eigenValues: " << eigenValues << "\n";
    if( sqrt(eigenValues[2]) / m_Sigma < m_PF_MinSmallestEigen )
    {

    validFeature = false;
    break;
    }


    bool inv_success;
    Matrix3f invM = invertMatrix(M,&inv_success);
    if( !inv_success )
    {

    validFeature = false;
    break;
    }


    Vector3f f = - (invM * b) / 2.0f;

    if( normQuad(f - pos) > line_rad_2)
    {

    validFeature = false;
    break;
    }

    debugRenderer->addLine(currentFeaturesPos,f);
    currentFeaturesPos = f;				*/
    debugRenderer->endRenderJob();
}

void LineFeatureDetector::smoothLineFeatures( LineFeatureSet * lf_set )
{
    LineFeatureSet * lf_copy = (LineFeatureSet*)lf_set->copy();
    UnstructuredInCorePointCloud * upc_lines = lf_copy->getFeaturesAsPointCloud();
    HierarchicalKNNIterator hItLines(upc_lines,32,NULL);
    float inv_line_sigma_2 = 1.0f / pow(m_Sigma,2.0f);
    float line_rad_2 = pow(2.0f * m_Sigma,2.0f);

    for( unsigned i=0;i<lf_set->m_LineFeatures.size();i++ )
    {
        LineFeature * lf = lf_set->m_LineFeatures[i];
        Vector3f lf_lineDir = normalize(lf->m_LineDirection);

        hItLines.setSeekPointAndReset(lf->m_Position);
        Vector6f filtered = NULL_VECTOR6F;
        while (!hItLines.atEnd())
        {
            unsigned lIndex = hItLines.getCurrentPointIndex();			
            LineFeature * lf2 = lf_copy->m_LineFeatures[lIndex];
            float dist_2 = normQuad(lf->m_Position - lf2->m_Position);
            if( dist_2 > line_rad_2 ) 
            {
                break;
            }

            // compute nearest point
            Vector3f nearestPos,nearestPosLineDir;
            lf2->computeNearestCurvePoint(lf->m_Position, nearestPos, nearestPosLineDir);
            nearestPosLineDir *= -1.0f;























            // compute weight
            float settings_minDirAngle = cos(30.0f);
            float settings_maxLineNormDiff = 0.5f;
            float weight_euclid = exp(-dist_2 * inv_line_sigma_2);
            float weight_dir = max(0.0f, (fabs(normalize(nearestPosLineDir) * lf_lineDir) - settings_minDirAngle) / (1.0f - settings_minDirAngle));
            float weight_normLine = max(0.0f,1.0f - fabs(norm(lf->m_LineDirection) - norm(lf2->m_LineDirection)) / settings_maxLineNormDiff);
            float weight = weight_euclid * weight_dir * weight_normLine;


            Vector6f tmp = makeVector6f( nearestPosLineDir[0],nearestPosLineDir[1],nearestPosLineDir[2],
                lf2->m_RotationDirection[0],lf2->m_RotationDirection[1],lf2->m_RotationDirection[2]);
            if( nearestPosLineDir * lf_lineDir<0)
                tmp *= -1.0f;
            filtered += tmp * weight;

            hItLines.next();
        }


        filtered = normalize(filtered);
        lf->m_LineDirection = makeVector3f( filtered[0],filtered[1],filtered[2]);
        lf->m_RotationDirection = makeVector3f( filtered[3],filtered[4],filtered[5]);

    }

    delete lf_copy;
}

struct TTmpSFeature
{
    Vector3f pos;
    float curv;
};

void LineFeatureDetector::testSlippageClustering()
{
    PointSet * ps = m_PC->getPointSet();
    AAT posAAT = ps->getAAT("position");
    AAT flagsAAT = ps->getAAT("flags");
    AAT colorAAT = ps->getAAT("color");

    // reset flags
    for( unsigned i=0;i<ps->getNumEntries();i++ )
    {

        ps->set1i(i,flagsAAT,0);
    }


    debugRenderer->clearDebugData("slipClustering");
    debugRenderer->beginRenderJob("slipClustering");

    std::vector<TTmpSFeature> features;
    float mark_rad_2 = pow(m_Sigma*m_SamplingFactor,2.0f);

    for( unsigned i=0;i<ps->getNumEntries();i++ )
    {

        int f = ps->get1i(i,flagsAAT);
        if( (f&PF_MARKED_FLAG) != 0)
        {

            continue;
        }

        Vector3f pos = ps->get3f(i,posAAT);
        Matrix6f eigenVectors;
        Vector6f eigenValues;
        computeSlippageAnalysis(pos, m_Sigma, eigenVectors, eigenValues, true );

        Vector3f color = makeVector3f(eigenValues[5] / eigenValues[0],eigenValues[4] / eigenValues[0],eigenValues[3] / eigenValues[0]);
        if( color[2] == 0.0f )
        {

            TTmpSFeature f;
            f.pos = pos;
            //features.push_back(f);

        }

        m_hIt->setSeekPointAndReset(pos);
        while(!m_hIt->atEnd())
        {

            if( normQuad(m_hIt->get3f(posAAT) - pos) > mark_rad_2 ) 
                break;


            ps->set3f(m_hIt->getCurrentPointIndex(),colorAAT,color);
            m_hIt->next();
        }
    }





}

X4_AVERAGE_TIMER_DEFINE(Detect_GatherData);
X4_AVERAGE_TIMER_DEFINE(Detect_ComputeLF);

LineFeatureSet * LineFeatureDetector::extractFeaturesOOC( PointCloud * pc )
{
    delete m_ThreadPool;
    if( m_Threading_NumThreads > 0 )
    {

        m_ThreadPool = new QThreadPool;
        m_ThreadPool->setExpiryTimeout(100);
        m_ThreadPool->setMaxThreadCount(m_Threading_NumThreads);

    }

    else
        m_ThreadPool = NULL;

    m_NumProcessedPoints = 0;
    X4_TIMER_START(ExtractLinesOOC);
    LineFeatureSet * ret = new LineFeatureSet;
    SceneObjectIterator * iter = pc->createIterator(SOIT::CAP_BASIC_PC | SOIT::CAP_HIERARCHICAL);
    traversePointCloudAndExtractLines(pc,iter,ret);
    debugOutput << "finished traversing, waiting for remaing threads\n";
    debugOutput.flush();
    if( m_ThreadPool )
    {

        m_ThreadPool->waitForDone();
        for( mpcard i=0;i<m_ThreadResults.size();i++ )
        {

            LineFeatureSet * tmp_lf = m_ThreadResults[i];
            for( mpcard j=0;j<tmp_lf->m_LineFeatures.size();j++ ) 
            {

                ret->m_LineFeatures.push_back((LineFeature*)tmp_lf->m_LineFeatures[j]->copy());
            }


            delete m_ThreadResults[i];
        }

        m_ThreadResults.clear();
    }

    debugOutput << "#Line features" << ret->m_LineFeatures.size() << "\n";
    ret->setup(LineFeatureSet::getDefaultName(),AttachedData::ADF_PERSISTENT,NULL);
    pc->getAttachments()->attachData(ret);
    delete iter;
    X4_TIMER_STOP(ExtractLinesOOC);
    X4_AVERAGE_TIMER_REPORT_AND_RESET(Detect_GatherData);
    X4_AVERAGE_TIMER_REPORT_AND_RESET(Detect_ComputeLF);


    return ret;
}

class LFD_DetectorThread : public QRunnable
{
public:

    LFD_DetectorThread( QMutex * mutex, LineFeatureDetector * rootDetector, UnstructuredInCorePointCloud * upc, std::vector<LineFeatureSet*> * results )
        :m_Mutex(mutex),m_Results(results),m_UPC(upc)
    {

        setAutoDelete(true);
        m_MyDetector = (LineFeatureDetector*)rootDetector->copy();
        m_MyDetector->begin(upc);
    }


    ~LFD_DetectorThread()
    {
    }



private:
    QMutex * m_Mutex;
    LineFeatureDetector * m_MyDetector;
    std::vector<LineFeatureSet*> * m_Results;
    UnstructuredInCorePointCloud * m_UPC;

    void recomputeAndOrientNormalsKNN( PointCloud * pc, mpcard numNeighbors )
    {

        if( !pc->providesAttribute("normal") )
            return;

        PCA<float,3> pca;
        SceneObjectIterator * iter = pc->createIterator(SOIT::CAP_BASIC_PC|SOIT::CAP_MODIFYABLE_PC);
        BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
        ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(iter);
        AAT posAAT = pc->getAAT("position");
        AAT normalAAT = pc->getAAT("normal");

        HierarchicalKNNIterator hIt(pc,32,NULL);

        // first pass: compute pca normals, use previous orientation
        while (!bIt->atEnd())
        {

            Vector3f p = bIt->get3f(posAAT);
            Vector3f oldNormal = bIt->get3f(normalAAT);
            hIt.setSeekPointAndReset(p);
            pca.clear();
            mpcard numPoints = 0;
            while(!hIt.atEnd() && numPoints++ < numNeighbors)
            {

                pca.addPoint(hIt.get3f(posAAT));
                hIt.next();
            }

            Vector3f centroid,eigenValues;
            Matrix3f eigenVectors;
            pca.analyze(eigenValues,eigenVectors,centroid);

            if( oldNormal*eigenVectors[2] < 0)
                eigenVectors[2] *= -1;
            mIt->set3f(normalAAT,eigenVectors[2]);

            bIt->next();
        }


        // second pass: orient normals via bilateral voting
        bIt->reset();
        while (!bIt->atEnd())
        {

            mpint dontflip = 0;
            Vector3f p = bIt->get3f(posAAT);
            Vector3f normal = bIt->get3f(normalAAT);
            hIt.setSeekPointAndReset(p);
            mpcard numPoints = 0;
            while(!hIt.atEnd() && numPoints++ < numNeighbors)
            {

                Vector3f tmpNormal = hIt.get3f(normalAAT);
                dontflip += (tmpNormal*normal < 0)?-1:1;
                hIt.next();
            }


            if( dontflip < 0)
                mIt->set3f(normalAAT,-normal);

            bIt->next();
        }


        delete iter;
    }


    void run()
    {

        recomputeAndOrientNormalsKNN(m_UPC,40);
        LineFeatureSet * lfs = m_MyDetector->extractFeatures(false);
        m_MyDetector->end();
        delete m_MyDetector;
        delete m_UPC;
        m_Mutex->lock();
        m_Results->push_back(lfs);
        m_Mutex->unlock();
    }

};

void LineFeatureDetector::traversePointCloudAndExtractLines( PointCloud * pc, SceneObjectIterator * iter, LineFeatureSet * lf_set )
{
    BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
    HierarchicalIterator * hIt = dynamic_cast<HierarchicalIterator*>(iter);

    card64 numPoints;
    hIt->getNodeInfo(HierarchicalIterator::niNumPointsInSubtree,NULL,&numPoints);

    const card64 maxNumPointsPerBlock = 5*1000 * 1000;

    if( numPoints > maxNumPointsPerBlock )
    {

        unsigned numChilds = hIt->getNumChildren();
        for( unsigned i=0;i<numChilds;i++ )
        {

            if( hIt->goToChild(i) )
            {

                traversePointCloudAndExtractLines(pc,iter,lf_set);
                hIt->goToParent();
            }
        }
    }



    else
    {

        UnstructuredInCorePointCloud * upc = new UnstructuredInCorePointCloud;
        upc->clearAndSetup( pc->getDescr(),0);
        BoundingBox3f bb = hIt->getBoundingBox();
        float border = m_Sigma * 2.0f;
        for( unsigned i=0;i<3;i++ )
        {

            bb.lowerCorner[i] -= border;
            bb.upperCorner[i] += border;
        }


        X4_AVERAGE_TIMER_START(Detect_GatherData);
        SceneObjectIterator * iter_gather = pc->createIterator(SOIT::CAP_BASIC_PC | SOIT::CAP_HIERARCHICAL);
        gatherPoints(pc,iter_gather,bb,upc);
        delete iter_gather;
        X4_AVERAGE_TIMER_STOP(Detect_GatherData);
        static mpcard lastReport = 0;
        m_NumProcessedPoints += upc->getNumPoints();
        if( m_NumProcessedPoints - lastReport > 1000*1000 )
        {

            debugOutput << " loaded: " << 100.0f*(float)m_NumProcessedPoints / (float)pc->getNumPoints() << "\n";
            lastReport = m_NumProcessedPoints;
            mpcard numLines = 0;
            for( mpcard i=0;i<m_ThreadResults.size();i++ )
            {

                numLines += m_ThreadResults[i]->m_LineFeatures.size();
            }

            debugOutput << "MB used for lines[" << numLines << "]: " << numLines * sizeof(LineFeature) / 1024 / 1024 << "\n";
            debugOutput.flush();
        }


        if( upc->getNumPoints() > 100 )
        {

            if( m_ThreadPool )
            {

                LFD_DetectorThread * t = new LFD_DetectorThread(m_Mutex, this, upc,&m_ThreadResults);
                m_ThreadPool->start(t);

            }
            else
            {



                X4_AVERAGE_TIMER_START(Detect_ComputeLF);
                begin(upc);
                LineFeatureSet * tmp_lf = extractFeatures(false);
                end();

                debugOutput << "num lines: " << tmp_lf->m_LineFeatures.size() + lf_set->m_LineFeatures.size();
                debugOutput << " loaded: " << 100.0f*(float)m_NumProcessedPoints / (float)pc->getNumPoints() << "\n";
                debugOutput.flush();

                for( unsigned i=0;i<tmp_lf->m_LineFeatures.size();i++ )
                {

                    lf_set->m_LineFeatures.push_back( (LineFeature*)tmp_lf->m_LineFeatures[i]->copy() );
                }

                delete tmp_lf;
                X4_AVERAGE_TIMER_STOP(Detect_ComputeLF);
                delete upc;
            }
        }
        else



            delete upc;
    }


}

struct LFD_Vector3f {
    bool operator() (const Vector3f& p1, const Vector3f& p2) const
    {

        if( p1[0] != p2[0] )
            return p1[0] < p2[0];
        else
        {


            if( p1[1] != p2[1] )
                return p1[1] < p2[1];
            else

                return p1[2] < p2[2];
        }
    }


};



void LineFeatureDetector::gatherPoints( PointCloud * pc, SceneObjectIterator * iter, const BoundingBox3f & bb, UnstructuredInCorePointCloud * retPC )
{
    HierarchicalIterator * hIt = dynamic_cast<HierarchicalIterator*>(iter);
    if( !hIt->getBoundingBox().intersectBoundingBox(bb))
        return;


    if( hIt->isLeaf() )
    {

        BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
        const PointSet * ps = bIt->lockBlockRead();
        std::vector<char> buffer(pc->getDescr()->getSize(),0);
        set<Vector3f,LFD_Vector3f> removeDoublePoints;
        AAT posAAT = pc->getAAT("position");

        if( hIt->getBoundingBox().containedIn(bb) )
        {

            // copy all points
            for( unsigned i=0;i<ps->getNumEntries();i++ )
            {

                Vector3f p = ps->get3f(i,posAAT);
                if( removeDoublePoints.find(p) != removeDoublePoints.end() )
                    continue;

                ps->getVertex(i,&buffer[0]);
                retPC->getPointSet()->addPoint( &buffer[0]);
                removeDoublePoints.insert(p);
            }
        }
        else
        {




            // test every point individually
            for( unsigned i=0;i<ps->getNumEntries();i++ )
            {

                if( !bb.inBoundingBox(ps->get3f(i,posAAT)) )
                    continue;

                Vector3f p = ps->get3f(i,posAAT);
                if( removeDoublePoints.find(p) != removeDoublePoints.end() )
                    continue;


                ps->getVertex(i,&buffer[0]);
                retPC->getPointSet()->addPoint( &buffer[0]);
                removeDoublePoints.insert(p);
            }
        }



        bIt->unlock();
    }

    else
    {

        unsigned numChilds = hIt->getNumChildren();
        for( unsigned i=0;i<numChilds;i++ )
        {

            if( hIt->goToChild(i) )
            {

                gatherPoints(pc,iter,bb,retPC);
                hIt->goToParent();
            }
        }



    }

}

Vector3f LineFeature::computeRotationCenter()
{
    Matrix3f rot_frame;
    rot_frame[0] = normalize(m_LineDirection);
    rot_frame[1] = normalize(m_RotationDirection);
    rot_frame[2] = normalize(rot_frame[0].crossProduct(rot_frame[1]));

    float beta = M_PI/2.0f - norm(m_RotationDirection);
    float m = norm(m_LineDirection) * m_Sigma;
    float d = m * tan(beta);
    Vector3f center = rot_frame * makeVector3f(0,0,-d) + m_Position;

    return center;
}

LineFeatureDetector::~LineFeatureDetector()
{
    delete m_Querry;
    delete m_hIt;
    delete m_Mutex;
    delete m_ThreadPool;
}

void LineFeatureSet::deleteFeatures( card32 begin, card32 end )
{
    for( unsigned i=begin;i<end;i++ )
    {

        delete m_LineFeatures[i];
    }

    m_LineFeatures.erase( m_LineFeatures.begin() + begin,m_LineFeatures.begin() + end);
}

LineFeatureSet::~LineFeatureSet()
{
    deleteFeatures(0, (card32)m_LineFeatures.size());
    delete m_PC;
}

LineFeature * LineFeatureSet::getNearestFeature( const Vector3f &point )
{
    LineFeature * ret = NULL;
    float closestDistance = 1e20f;
    for(unsigned i=0;i<m_LineFeatures.size();i++ )
    {

        float dist = norm(m_LineFeatures[i]->m_Position - point);
        if( dist < closestDistance )
        {

            ret = m_LineFeatures[i];
            closestDistance = dist;
        }
    }


    return ret;
}

UnstructuredInCorePointCloud * LineFeatureSet::getFeaturesAsPointCloud()
{
    if( m_PC != NULL && m_PC->getNumPoints() != m_LineFeatures.size() )
    {

        delete m_PC;
        m_PC = NULL;
    }

    if( m_PC == NULL )
    {

        m_PC = new UnstructuredInCorePointCloud;
        VertexDescriptor * vd = createDefaultVertexDescriptor(true,false,false,true);
        m_PC->clearAndSetup(vd, m_LineFeatures.size());
        AAT posAAT = m_PC->getAAT("position");
        PointSet * ps = m_PC->getPointSet();
        for( unsigned i=0;i<m_LineFeatures.size();i++ )
        {

            ps->set3f(i, posAAT, m_LineFeatures[i]->m_Position);
        }
    }


    return m_PC;
}

void LineFeatureSet::setupGraph( float maxEdgeLength )
{
    m_Graph = std::vector< std::vector<unsigned> >(m_LineFeatures.size());
    UnstructuredInCorePointCloud * upc = getFeaturesAsPointCloud();
    HierarchicalKNNIterator hIt(upc,32,NULL);
    AAT posAAT = upc->getAAT("position");
    card32 numLinks = 0;

    for( unsigned i=0;i<m_LineFeatures.size();i++ )
    {

        LineFeature * lf = m_LineFeatures[i];
        Vector3f firstFeatureFromCurrentCluster;
        hIt.setSeekPointAndReset(lf->m_Position);
        hIt.next();
        std::map<unsigned,unsigned> clusterMap;
        while(!hIt.atEnd())
        {

            Vector3f tmpPos = hIt.get3f(posAAT);
            if( norm(tmpPos - lf->m_Position) > maxEdgeLength || 
                m_Graph[i].size() > 100 )
                break;


            unsigned tmpClusterID = m_ClusterID[hIt.getCurrentPointIndex()];
            bool connect = false;
            if( clusterMap.find(tmpClusterID) == clusterMap.end() )
            {

                connect = true;
            }
            else


                if( tmpClusterID == m_ClusterID[i] && clusterMap[tmpClusterID] < 2)
                {

                    // provide graph link to own cluster into both directions
                    if( (tmpPos - lf->m_Position) * (firstFeatureFromCurrentCluster - lf->m_Position) < 0.0f )
                    {
                        connect = true;
                    }
                }





                if( !connect )	// add random links to 
                {

                    //connect = (rand() % 4) == 0;
                }


                if( connect )
                {

                    if( tmpClusterID != m_ClusterID[i] )
                    {

                        m_Graph[i].push_back(hIt.getCurrentPointIndex());
                    }
                    else
                    {



                        m_Graph[i].insert(m_Graph[i].begin(),(unsigned)hIt.getCurrentPointIndex());
                    }

                    clusterMap[tmpClusterID]++;
                    if( tmpClusterID == m_ClusterID[i] )
                        firstFeatureFromCurrentCluster = tmpPos;
                    numLinks++;
                }


                hIt.next();
        }
    }


    debugOutput << "num average links per lf: " << (float)numLinks / (float)getNumFeatures() << "\n";
}

void LineFeatureSet::clearFeaturePointCloud()
{
    delete m_PC;
    m_PC = NULL;
}

void LineFeatureSet::assign( const Object* obj, X4CopyContext *context /*= NULL */ )
{
    //	copyObjectProperties(obj,this);
    const LineFeatureSet * lfs = dynamic_cast<const LineFeatureSet*>(obj);
    if(lfs == NULL ) 
        throw X4Exception("LineFeatureSet::assign() - Trying to assign from wrong object type\n");

    m_Graph.clear();
    m_ClusterID = lfs->m_ClusterID;
    m_LineFeatures.resize(0);
    for( unsigned i=0;i<lfs->m_LineFeatures.size();i++ )
    {

        m_LineFeatures.push_back( (LineFeature*)lfs->m_LineFeatures[i]->copy() );		
    }

}


std::vector<unsigned> LineFeatureSet::removeSmallClusters( const card32& minClusterSize, PointSet** ps_ref )
{ // note: must keep order!

    if ( m_ClusterID.size() != m_LineFeatures.size() )
        throw X4Exception("LineFeatureSet::removeSmallClusters() - clusterIDs invalid!");
    const unsigned num_feature = (unsigned) m_LineFeatures.size();

    std::vector<unsigned> deleteMe(num_feature, 0);
    std::vector<unsigned> clusterSize(num_feature, 0);
    std::vector<unsigned> updated_indices(num_feature);
    std::vector<LineFeature*> updated_features(num_feature);
    std::vector<card32> updated_cluster_id(num_feature);

    // compute cluster sizes
    for (unsigned fi = 0; fi < num_feature; ++fi) {
        clusterSize[m_ClusterID[fi]]++;
    }



    // mark lines that we want to delete
    for (unsigned fi = 0; fi < num_feature; ++fi) {



        if (clusterSize[m_ClusterID[fi]] < minClusterSize)
            deleteMe[fi] = 1;
    }


    unsigned num_indices = 0;
    for (unsigned fi = 0; fi < num_feature; ++fi) {
        if ( deleteMe[fi] ) {
            delete m_LineFeatures[fi];
        } else {
            updated_features[num_indices] = m_LineFeatures[fi];
            updated_cluster_id[num_indices] = m_ClusterID[fi];
            updated_indices[num_indices] = fi;
            ++num_indices;
        }
    }
    updated_features.resize(num_indices);
    updated_cluster_id.resize(num_indices);
    updated_indices.resize(num_indices);







    m_LineFeatures.clear();
    m_LineFeatures.resize(num_indices);
    m_ClusterID.clear();
    m_ClusterID.resize(num_indices);
    for (unsigned fi = 0; fi < num_indices; ++fi) {
        m_LineFeatures[fi] = updated_features[fi];
        m_ClusterID[fi] = updated_cluster_id[fi];
    }


	 if (ps_ref) {
		 PointSet* ps = *ps_ref;
		 PointSet* newps = ps->subset(updated_indices);
		 delete ps;
		 *ps_ref = newps;
	 }
    clearFeaturePointCloud();

    return updated_indices;
}

void LineFeatureSet::resample( float gridSize )
{
    m_ClusterID.clear();
    m_Graph.clear();

    Simple3DHashedGrid grid(gridSize);
    for( mpcard i=m_LineFeatures.size();i>0;i-- )
    {

        Vector3f p = m_LineFeatures[i-1]->m_Position;
        if( grid.insert(p) )
        {

            delete m_LineFeatures[i-1];
            m_LineFeatures.erase(m_LineFeatures.begin()+i-1);
        }
    }


}

Vector3f computeNearestPoint( const Vector3f &dir0, const Vector3f &p0, const Vector3f &dir1, const Vector3f &p1 )
{
    // transform line 0 into local frame of line 1
    Matrix3f frame1 = calcTangentSystem(dir1);
    Vector3f g = frame1.transpose() * dir0;
    Vector3f g0 = frame1.transpose() * (p0 - p1);
    g[2] = 0;
    g0[2] = 0;

    float t = (g * g0) / (g*g);

    return p0 - dir0*t;
}

Vector3f computePseudoIntersectionPoint(const Vector3f &dir0, const Vector3f &p0, const Vector3f &dir1, const Vector3f &p1 )
{
    return computeNearestPoint(dir0, p0, dir1, p1)*0.5f + computeNearestPoint(dir1, p1, dir0, p0) * 0.5f;
}

Matrix4f base_computeCoordinateFrame(LineFeature * base_f1, LineFeature * base_f2)
{
    Vector3f pseudoIntersectionPoint1 = computePseudoIntersectionPoint(base_f1->m_LineDirection,base_f1->m_Position,
        base_f2->m_LineDirection,base_f2->m_Position);

    Vector3f l1 = normalize(base_f1->m_LineDirection);
    Vector3f l2 = normalize(base_f2->m_LineDirection);
    if( l1 * (base_f1->m_Position - pseudoIntersectionPoint1) < 0.0f )
        l1 *= -1;
    if( l2 * (base_f2->m_Position - pseudoIntersectionPoint1) < 0.0f )
        l2 *= -1;

    Matrix3f rot_frame;
    rot_frame[0] = l1;
    rot_frame[2] = normalize(l1.crossProduct(l2));
    rot_frame[1] = normalize(rot_frame[2].crossProduct(l1));
    if( base_f1->m_Normal * rot_frame[2] < 0.0f )
        rot_frame[2] *= -1.0f;
    return makeTranslation4f(pseudoIntersectionPoint1) * expand3To4(rot_frame);
}

Matrix4f X4::computeTransformation( LineFeature * base1_f1, LineFeature * base1_f2, LineFeature * base2_f1, LineFeature * base2_f2,bool mirrored )
{
    Matrix4f frame1 = base_computeCoordinateFrame(base1_f1,base1_f2);
    Matrix4f frame2 = base_computeCoordinateFrame(base2_f1,base2_f2);
    return frame2 * invertFrame(frame1);


    /*	// easy method: compute translation with T(f1+f2) / 2 = (f1+f2)/2
    Vector3f middle1 = (base1_f1->m_Position + base1_f2->m_Position) / 2;
    Vector3f middle2 = (base2_f1->m_Position + base2_f2->m_Position) / 2;

    // compute pseudo intersection points
    Vector3f pseudoIntersectionPoint1 = computePseudoIntersectionPoint(base1_f1->m_LineDirection,base1_f1->m_Position,base1_f2->m_LineDirection,base1_f2->m_Position);
    Vector3f pseudoIntersectionPoint2 = computePseudoIntersectionPoint(base2_f1->m_LineDirection,base2_f1->m_Position,base2_f2->m_LineDirection,base2_f2->m_Position);

    float invert_ld1 = base1_f1->m_LineDirection * base1_f2->m_LineDirection < 0 ? -1.0f : 0;
    float invert_ld2 = base2_f1->m_LineDirection * base2_f2->m_LineDirection < 0 ? -1.0f : 0;
    Vector3f l1 = normalize(base1_f1->m_LineDirection);//normalize(normalize(base1_f1->m_LineDirection)*invert_ld1 + normalize(base1_f2->m_LineDirection));
    Vector3f l2 = normalize(base2_f1->m_LineDirection);//normalize(normalize(base2_f1->m_LineDirection)*invert_ld2 + normalize(base2_f2->m_LineDirection));

    if( l1 * (base1_f1->m_Position - pseudoIntersectionPoint1) < 0 )
    l1 *= -1;
    if( l2 * (base2_f1->m_Position - pseudoIntersectionPoint2) < 0 )
    l2 *= -1;

    Vector3f n1 = normalize( (base1_f1->m_Position - base1_f2->m_Position).crossProduct(l1) );
    Vector3f n2 = normalize( (base2_f1->m_Position - base2_f2->m_Position).crossProduct(l2) );

    Matrix3f frame1, frame2;
    frame1[0] = l1;
    frame2[0] = l2;
    frame1[1] = l1.crossProduct(n1);
    frame2[1] = l2.crossProduct(n2);
    frame1[2] = n1;
    frame2[2] = n2;

    if( mirrored )
    frame2[2] = -frame2[2];

    Matrix4f ret = makeTranslation4f(middle2) * expand3To4(frame2) * expand3To4(frame1.transpose()) * makeTranslation4f(-middle1);

    /*	debugRenderer->beginRenderJob("test ");
    Matrix4f c1 = makeTranslation4f(middle1) * expand3To4(frame1);
    Matrix4f c2 = makeTranslation4f(middle2) * expand3To4(frame2);

    debugRenderer->addFineCoordinateFrame(c1,2.0f, makeVector3f(1,0.8f,0));
    debugRenderer->addFineCoordinateFrame(c2,2.0f, makeVector3f(1,0.8f,0));

    debugRenderer->addSphere(pseudoIntersectionPoint1, 0.5f, makeVector3f(0,0,1));
    debugRenderer->addSphere(pseudoIntersectionPoint2, 0.5f, makeVector3f(0,0,1));

    debugRenderer->addFastCylinder(base1_f1->m_Position, base1_f1->m_LineDirection, 4.0f, 0.025f,makeVector3f(1,1,0),true,false);
    debugRenderer->addFastCylinder(base1_f2->m_Position, base1_f2->m_LineDirection, 4.0f, 0.025f,makeVector3f(1,1,0),true,false);
    debugRenderer->addFastCylinder(base2_f1->m_Position, base2_f1->m_LineDirection, 4.0f, 0.025f,makeVector3f(1,1,0),true,false);
    debugRenderer->addFastCylinder(base2_f2->m_Position, base2_f2->m_LineDirection, 4.0f, 0.025f,makeVector3f(1,1,0),true,false);

    debugRenderer->endRenderJob();
    */


    //return ret;
}

void LineFeature::drawWithDebugRenderer()
{
    Vector3f dir = normalize(m_LineDirection) * m_Sigma;
    debugRenderer->addFastCylinder(m_Position - dir, normalize(dir), norm(dir)*2.0f, m_Sigma * 0.25f, makeVector3f(1,0,0),false,false);
}

void LineFeature::computeNearestCurvePoint( const Vector3f &querry_pos, Vector3f &nearest_pos, Vector3f &nearest_pos_derivate )
{
    if( norm(m_RotationDirection) < 0.05f )
    {

        nearest_pos = m_Position + m_LineDirection * (m_LineDirection * (querry_pos - m_Position));
        nearest_pos_derivate = normalize(m_LineDirection);
        return;
    }



    Matrix3f rot_frame;
    rot_frame[0] = normalize(m_LineDirection);
    rot_frame[1] = normalize(m_RotationDirection);
    rot_frame[2] = normalize(rot_frame[0].crossProduct(rot_frame[1]));

    float beta = M_PI/2.0f - norm(m_RotationDirection);
    float m = norm(m_LineDirection) * m_Sigma;
    float d = m * tan(beta);
    Vector3f center = rot_frame * makeVector3f(0,0,-d) + m_Position;

    float radius = norm(center - m_Position);

    nearest_pos = rot_frame.transpose() * (querry_pos - center);
    nearest_pos[1] = 0;
    nearest_pos = normalize(nearest_pos);

    nearest_pos_derivate = rot_frame * makeVector3f(-nearest_pos[2],0,nearest_pos[0]);
    nearest_pos = rot_frame * nearest_pos*radius + center;
}

Vector3f LineFeature::computePseudoIntersectionPoint( LineFeature * f )
{
    return ::computePseudoIntersectionPoint( m_LineDirection, m_Position, f->m_LineDirection, f->m_Position);
}

void LineFeature::drawSlippableMotionWithDebugRenderer( const Vector3f &origin, const Vector3f &color )
{
    //if( norm(m_RotationDirection) <  )

    Vector3f lastVec;
    int samples = 10;
    Matrix3f rot_frame;
    rot_frame[0] = normalize(m_LineDirection);
    rot_frame[1] = normalize(m_RotationDirection);
    rot_frame[2] = normalize(rot_frame[0].crossProduct(rot_frame[1]));

    float beta = M_PI/2.0f - norm(m_RotationDirection);
    float m = norm(m_LineDirection) * m_Sigma;
    float d = m * tan(beta);
    Vector3f center = rot_frame * makeVector3f(0,0,-d) + m_Position;

    float radius = norm(center - m_Position);

    for( int i=-samples;i<=samples;i++ )
    {

        float angle = (float)i / (float)samples*norm(m_RotationDirection);
        Vector3f vec = rot_frame * makeVector3f(sin(angle),0,cos(angle)) * radius + center;


        if( i > -samples )
        {

            debugRenderer->addLine(lastVec-m_Position + origin,vec-m_Position + origin,color,color,2);
        }


        lastVec = vec;
    }

}

void LineFeature::transform( const Matrix4f &M )
{
    m_Position = transformVector3f(M,m_Position);
    Matrix3f rotMat = shrink4To3(M);
    m_LineDirection = rotMat * m_LineDirection;
    m_RotationDirection = rotMat * m_RotationDirection;
    m_Normal = rotMat * m_Normal;
}






UnstructuredInCorePointCloud * LineFeatureSet::getFeaturesAsPointCloudJK()
{
    if( m_PC != NULL && m_PC->getNumPoints() != m_LineFeatures.size() )
    {

        delete m_PC;
        m_PC = NULL;
    }

    if( m_PC == NULL )
    {

        m_PC = new UnstructuredInCorePointCloud;
        VertexDescriptor * vd = createDefaultVertexDescriptor(true,false,true,false,false);
        vd->pushAttrib(VAD("linedirection",3,VAD::DATA_FORMAT_FLOAT32));
        vd->pushAttrib(VAD("rotationdirection",3,VAD::DATA_FORMAT_FLOAT32));
        vd->pushAttrib(VAD("sigma",1,VAD::DATA_FORMAT_FLOAT32));
        vd->pushAttrib(VAD("curvature",1,VAD::DATA_FORMAT_FLOAT32));
        m_PC->clearAndSetup(vd, m_LineFeatures.size());
        AAT POSITION = m_PC->getAAT("position");
        AAT LINEDIR = m_PC->getAAT("linedirection");
        AAT ROTDIR = m_PC->getAAT("rotationdirection");
        AAT NORMAL = m_PC->getAAT("normal");
        AAT SIGMA = m_PC->getAAT("sigma");
        AAT CURVATURE = m_PC->getAAT("curvature");
        PointSet * ps = m_PC->getPointSet();
        for( unsigned i=0;i<m_LineFeatures.size();i++ )
        {

            ps->set3f(i, POSITION, m_LineFeatures[i]->m_Position);
            ps->set3f(i, LINEDIR, m_LineFeatures[i]->m_LineDirection);
            ps->set3f(i, ROTDIR, m_LineFeatures[i]->m_RotationDirection);
            ps->set3f(i, NORMAL, m_LineFeatures[i]->m_Normal);
            ps->set1f(i, SIGMA, m_LineFeatures[i]->m_Sigma);
            ps->set1f(i, CURVATURE, m_LineFeatures[i]->m_Curvature);
        }
    }


    return m_PC;
}