#ifndef LineFeatureH
#define LineFeatureH
//---------------------------------------------------------------------------
#include "x3foundation_private.h"
//---------------------------------------------------------------------------
#include "AttachedData.h"
#include "VariableArrayClassProperty.h"
#include <QtCore/QtCore>
//---------------------------------------------------------------------------

namespace X4
{

class UnstructuredInCorePointCloud;
class FastSphereQuerry;
class HierarchicalKNNIterator;
class CachedRangeQuerry;
class InCorePCTopologyGraph;
class UnstructuredInCoreTriangleMesh;
class PointCloud;
class SceneObjectIterator;
/**
	line features
*/
class X3FOUNDATION_PRIVATE_API LineFeature : public Persistent
{
	X4_CLASS( LineFeature )

private:

public:
	LineFeature(void);
	~LineFeature(void);

   virtual void assign(const Object* obj, X4CopyContext *context = NULL);

	void drawWithDebugRenderer();
	void drawSlippableMotionWithDebugRenderer(const Vector3f &origin, const Vector3f &color);

	/// project querry point onto curve and return pos+derivate
	void computeNearestCurvePoint( const Vector3f &querry_pos, Vector3f &nearest_pos, Vector3f &nearest_pos_derivate );

	/// compute pseudo intersection point
	Vector3f computePseudoIntersectionPoint( LineFeature * f);

	/// returns center of rotation, doesn't really work with low rotation
	Vector3f computeRotationCenter();

	/// transform feature
	void transform(const Matrix4f &m);

	Vector3f		m_InitialPosition;
	Vector3f		m_Position;
	Vector3f		m_Normal;				// averaged normal from support
	Vector3f		m_LineDirection;		// translational part of the eigenvector pointing in slippable direction
	Vector3f		m_RotationDirection;	// rotational part
	
	float32		m_Curvature;
	float32		m_Curvature2;			// curvature on a bigger scale (2.0)
	float32		m_Sigma;
	float32     m_Length;				// in case when line is bound
};

class X3FOUNDATION_PRIVATE_API LineFeatureSet : public AttachedData
{
		X4_CLASS( LineFeatureSet )

private:

	UnstructuredInCorePointCloud * m_PC;

public:

	static std::string getDefaultName() {return "LineFeatureSet";}

	LineFeatureSet() {m_PC = NULL;}
	~LineFeatureSet();
	std::vector<LineFeature*> m_LineFeatures;
	std::vector<card32> m_ClusterID;
	std::vector< std::vector< unsigned > > m_Graph;
	std::vector<Vector3f> m_RejectedSamples;

	void assign(const Object* obj, X4CopyContext *context = NULL );

	void addFeature( Object * obj ) {m_LineFeatures.push_back((LineFeature*)obj);}
	void deleteFeatures( card32 begin, card32 end );
	Object * getFeature( card32 index ) {return m_LineFeatures[index];}
	void moveFeature( card32 oldPos, card32 newPos ) {}
	card32 getNumFeatures() {return (card32)m_LineFeatures.size();}
	std::string getFeatureName(card32 index ) {return "ToDo :-)";}

	LineFeature * getNearestFeature( const Vector3f &point);

	// use a knn iterator on this point cloud for efficient feature queries
	UnstructuredInCorePointCloud * getFeaturesAsPointCloud();
	UnstructuredInCorePointCloud * getFeaturesAsPointCloudJK();

	// resample (grid based)
	void resample(float gridSize);

	// delete point cloud
	void clearFeaturePointCloud();

	// removes all lines that belong to a small clusters
	std::vector<unsigned> removeSmallClusters(const card32& minClusterSize, PointSet** ps_ref);
	
	/// create graph
	void setupGraph(float maxEdgeLength);

	IMPLEMENT_VAR_ARRAY_METHODS_STL( m_ClusterID );
};

/**
	computation of line features
*/

class X3FOUNDATION_PRIVATE_API LineFeatureDetector : public Persistent
{
		X4_CLASS( LineFeatureDetector )
private:

	// persistent properties
	float32	m_Sigma;
	float32	m_MinEigenRatio;		// reject when lambda5/(lambda6 + eps) < minEigenRatio
	card32	m_MinNumberOfPoints;	// minimum number of points for slippage analysis
	float32	m_SamplingFactor;
	float32	m_MaxContrainedFactor;	// reject when lambda6/lambda1 > m_MaxContrainedFactor

	float32	m_PF_MinGaussianCurv;
	float32	m_PF_MinSmallestEigen;
	card32	m_PF_MinClustersize;
	card32	m_Threading_NumThreads;

	// members
	std::vector<int> m_Visited;
	float			m_MedianPointDistance;
	int			m_VisitedID;
	AAT			m_PosAAT, m_NormalAAT,m_FlagsAAT;
	static const unsigned PF_MARKED_FLAG = 8;
	UnstructuredInCorePointCloud * m_PC;
	FastSphereQuerry * m_Querry;
	HierarchicalKNNIterator * m_hIt;
	CachedRangeQuerry * m_CachedQuerry;
	float	m_CacheOverlap;
	mpcard	m_NumProcessedPoints;

	std::vector<float> m_CompCurv_Weights;
	std::vector<Vector3f> m_CompCurv_TransformedPositions;
	std::vector<Vector3f> m_CompCurv_PositionCache;


	// threading
	QThreadPool	*	m_ThreadPool;
	QMutex		*	m_Mutex;
	std::vector<LineFeatureSet*>	m_ThreadResults;

	/// compute slippage analysis, return false if support containts not enough points
	bool computeSlippageAnalysis( const Vector3f &pos, float sigma, Matrix6f &eigenVectors, Vector6f &eigenValues, bool markVisited );

	/// compute curvatures for a local patch (uses patch fitting)
	void computeCurvatures( const Vector3f &pos, const Vector3f &normal, float sigma, float &meanCurvature, float &gaussianCurvature );

	/// project point onto the intersection of the plane orthogonal to the slippable motion and the surface
	Vector3f projectPoint( const Vector3f &pos, float sigma, const Vector3f slippageDirection, float &summedWeight);

	/// compute score for line optimization
	float computeScore(const Vector3f &pos, float sigma);

	void clearMarkedFlag( PointCloud * pc);

	/// traverse hierarchical point cloud, fetch parts of the point cloud and combine resulting features in lf_set
	void traversePointCloudAndExtractLines(PointCloud * pc, SceneObjectIterator * iter, LineFeatureSet * lf_set);

	/// fetch all points in bb and store them in retPC
	void gatherPoints(PointCloud * pc, SceneObjectIterator * iter, const BoundingBox3f & bb, UnstructuredInCorePointCloud * retPC  );

public:

	LineFeatureDetector();
	~LineFeatureDetector();

	void assign(const Object* obj, X4CopyContext *context = NULL );

	/// prepare data structures
	void begin( UnstructuredInCorePointCloud * pc);

	/// clean up
	void end();

	/// detect a single line, returns NULL if no line was found
	LineFeature * extractLineFeature( const Vector3f &pos );
	LineFeature * extractLineFeature_old( const Vector3f &pos );

	/// detect line features for compute point cloud
	LineFeatureSet * extractFeatures(bool attachToPointCloud=true);

	/// detect line features in ooc point clouds
	LineFeatureSet * extractFeaturesOOC(PointCloud * pc);

	/// smooth line features
	void smoothLineFeatures(LineFeatureSet * lf_set );

	/// detect point features with respect to a set of line features
	void extractPointFeatures( LineFeatureSet * lf_set );

	/// debug: visualize computed line segments for a given point
	void debug_visualizeQuadricLines(const Vector3f &p, LineFeatureSet * lf_set);

	/// test something
	void testSlippageClustering();

	void extractFeatures_TriangleMesh( UnstructuredInCoreTriangleMesh * mesh, LineFeatureSet * lf_set );


	// access methods
	float32 getSigma() const {return m_Sigma;}
	void setSigma(float32 s ) {m_Sigma = s;}

};

/// compute transformation from base1 to base2
X3FOUNDATION_PRIVATE_API Matrix4f computeTransformation( LineFeature * base1_f1, LineFeature * base1_f2, LineFeature * base2_f1, LineFeature * base2_f2, bool mirrored);

}

#endif