#ifndef LocalDescriptorH
#define LocalDescriptorH
//---------------------------------------------------------------------------
#include "x3foundation_private.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
#include "LinearAlgebra.h"
//---------------------------------------------------------------------------

namespace X4
{

class LocalSignature;
class UnstructuredInCorePointCloud;

/**
	base class for local surface/environment descriptors

*/
class X3FOUNDATION_PRIVATE_API LocalDescriptor : public Persistent
{
	X4_ABSTRACT_CLASS( LocalDescriptor )

private:

	float32	m_DifferenceScaleFactor;

public:

	LocalDescriptor() {m_DifferenceScaleFactor = 1.0f;}

   virtual void assign(const Object* obj, X4CopyContext *context = NULL);

	/// setup descriptor
	virtual void beginSignatureComputations(UnstructuredInCorePointCloud * upc) {}
	
	/// compute local signature for point
	virtual LocalSignature * computeSignature(const Vector3f &point, const Vector3f &normal, float size, UnstructuredInCorePointCloud * upc) = 0;

	/// cleanup
	virtual void endSignatureComputations() {}


	inline float32 getDifferenceScaleFactor() {return m_DifferenceScaleFactor;}
   inline void setDifferenceScaleFactor(float32 d) { m_DifferenceScaleFactor = d; }
   
	/// returns true if signatures can be recomputed (e.g. false for all detector-based signatures)
	virtual bool isRecomputeable() const {return true;}

};

/************************************************************************/
/*		base class for local signatures                                   */
/************************************************************************/
class X3FOUNDATION_PRIVATE_API LocalSignature : public Persistent
{
	X4_ABSTRACT_CLASS( LocalSignature )

private:

public:
	virtual void assign(const Object* obj, X4CopyContext *context = NULL);

	/// compares two signatures (descriptor is required for context information)
	virtual float compare( LocalSignature * other, LocalDescriptor * myDescriptor ) = 0;

	/// compares two signatures (no context information required)
	virtual float compare( LocalSignature * other) = 0;

	/// used for signature based feature filtering
	virtual bool badFeature(LocalDescriptor * myDescriptor) const {return false;}

};

/************************************************************************/
/*		simple float std::vector signature with l2 norm                        */
/************************************************************************/
class X3FOUNDATION_PRIVATE_API FloatVectorSignature : public LocalSignature
{
	X4_CLASS(FloatVectorSignature)
private:

	void setNumDatas(card32 size) {setSize(size);}
	void deleteData( card32 index );
	char * getData( card32 index ) {return (char*)&m_Data[index];}
	card32 getNumDatas() {return m_Data.getDim();}
	std::string getDataName(card32 index ) {return "-";}

	DVectorF		m_Data;
public:

	virtual void assign(const Object* obj, X4CopyContext *context = NULL);

	// access methods
	void setVector( const DVectorF &v) {m_Data = v;}
	DVectorF getVector() const {return m_Data;}
	void setSize( unsigned numEntries ) {m_Data.setDim(numEntries);}

	virtual float compare( LocalSignature * other, LocalDescriptor * myDescriptor );
	virtual float compare( LocalSignature * other ) { return compare(other, NULL); }
	float compare( FloatVectorSignature * other );

	std::string printToString() const;


};

}

#endif