#include "StdAfx.h"
#include <windows.h>
#include <mmsystem.h>
//---------------------------------------------------------------------------
#include "MLSProjector.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "CachedRangeQuerry.h"
#include "PCA.h"
#include "Monomials2ndOrder.h"
#include "PointCloudTools.h"
#include "SeparatorClassProperty.h"
//---------------------------------------------------------------------------
using namespace X4;

IMPLEMENT_X4_CLASS( MLSProjector ,1)
{
	BEGIN_CLASS_INIT( MLSProjector );
	ADD_FLOAT32_PROP(m_Sigma,0);

	ADD_SEPARATOR("RILMS");
	ADD_FLOAT32_PROP(m_SigmaR,0);
	ADD_FLOAT32_PROP(m_SigmaN,0);
	ADD_CARD32_PROP(m_MinRefittingIterations,0);
	ADD_CARD32_PROP(m_MaxRefittingIterations,0);

	ADD_SEPARATOR("Settings");
	ADD_CARD32_PROP(m_minNumberOfSamples, 1);
}

MLSProjector::MLSProjector(void)
{
	m_RangeQuery = NULL;
	m_PC = NULL;
	m_Sigma = 2.0f;

	m_SigmaN = 0.8f;
	m_SigmaR = 0;
	m_MinRefittingIterations = 1;
	m_MaxRefittingIterations = 3;

	m_minNumberOfSamples = 15;
}

MLSProjector::~MLSProjector(void)
{
	delete m_RangeQuery;
}

void MLSProjector::assign(const Object* obj, X4CopyContext *context)
{
	// Persistent::assign(obj,context);
	const MLSProjector * o = dynamic_cast<const MLSProjector*>(obj);
	x4Assert( o != NULL );

	copyObjectProperties(obj,this);
}

void X4::MLSProjector::begin( PointCloud * pc, float medPointDistance )
{
	if (m_RangeQuery) delete m_RangeQuery;
	m_RangeQuery = new CachedRangeQuerry(pc);
	m_PC = pc;
	if( medPointDistance == 0)
		m_MedPointDistance = getMedianPointDistance(pc);
	else
		m_MedPointDistance = medPointDistance;
	m_PositionAAT = pc->getAAT("position");
	m_NormalAAT = pc->getAAT("normal");
	m_ProvideWeightChannel = pc->providesAttribute("weight");
	if( m_ProvideWeightChannel )
		m_WeightAAT = pc->getAAT("weight");

}

X4::Vector3f X4::MLSProjector::project( const Vector3f &pos, Vector3f *projDerivate )
{
	m_LastPointOK = true;
	return project_RIMLS(pos,projDerivate);
}

X4::Vector3f X4::MLSProjector::project_SimpleMLS( const Vector3f &pos, Vector3f *projDerivate, bool* valid, float* summedWeight )
{
	if (valid == NULL)
		valid = &m_LastPointOK;

	if (summedWeight == NULL)
		summedWeight = &m_LastSummedWeight;

	Vector3f projPos = pos;
	float sigma = m_MedPointDistance * m_Sigma;
	PointSet * ps = m_RangeQuery->getSpherePoints(projPos, sigma*2.0f, m_MedPointDistance * 0.3f);
	if( ps->getNumEntries() < m_minNumberOfSamples )
	{
		*valid = false;
		return pos;
	}

	//if( rand() % 100 == 0 )
	//	debugOutput << "num pc: " << ps->getNumEntries() << "\n";

	// 01.2011 (c) art - removed static optimization here. Please consider to use threads instead! Static doesn't bring so much nowadays
	std::vector<float> weights(ps->getNumEntries(), 0);
	std::vector<Vector3f> positions(ps->getNumEntries(), NULL_VECTOR3F);
	float inv_sigma_2 = 1.0f / pow(sigma,2.0f);
	card32 numIter = 0;
	Vector3f lastPos = pos;

	while(numIter++ < 2)
	{
		*summedWeight = 0;
		Vector3f firstNormal = ps->get3f(0,m_NormalAAT);

		while( weights.size() < ps->getNumEntries() )
		{
			weights.push_back(0);
			positions.push_back(NULL_VECTOR3F);
		}

		// compute pca
		PCA<float,3> pca;
		float * currentWeightPtr = &weights[0];
		for( unsigned i=0; i<ps->getNumEntries();i++ )
		{
			Vector3f p = ps->get3f(i,m_PositionAAT);
			float weight = exp( - normQuad(p - projPos)  * inv_sigma_2);
			if( m_ProvideWeightChannel )
				weight *= ps->get1f(i,m_WeightAAT);
			*currentWeightPtr++ = weight;
			*summedWeight  += weight;
			pca.addPoint(p,weight);
		}

		Vector3f centroid, eigenValues;
		Matrix3f eigenVectors;
		pca.analyze(eigenValues, eigenVectors, centroid);
		if(  eigenVectors[2] * firstNormal < 0)
		{
			eigenVectors[1] *= -1;
			eigenVectors[2] *= -1;
		}

		const Matrix3f frameToWorld = eigenVectors;
		const Matrix3f worldToFrame = eigenVectors.transpose();

		Vector3f planePos = projPos + frameToWorld[2] * (frameToWorld[2] * (centroid - projPos));
		
		// transform positions into local frame
		for( unsigned i=0; i<ps->getNumEntries();i++ )
		{
			Vector3f p = ps->get3f(i,m_PositionAAT);
			positions[i] = worldToFrame * (p - planePos) / sigma;
		}


		// fit monomial
		Vector6f coeffs;
		Monomials2ndOrder::computeCoeffs( &positions[0], &weights[0], ps->getNumEntries(), coeffs);

		projPos = makeVector3f(0,0, Monomials2ndOrder::evaluate(coeffs,makeVector2f(0,0)));
		projPos = frameToWorld * projPos * sigma + planePos;
		
		if( projDerivate )
		{
			*projDerivate = frameToWorld * Monomials2ndOrder::evaluateNormal(coeffs,makeVector2f(0,0));
		}

		if( norm(lastPos - projPos) < m_MedPointDistance * 0.01f  )
			break;
		lastPos = projPos;

		ps = m_RangeQuery->getSpherePoints(projPos, sigma*2.0f, m_MedPointDistance * 1.7f);
		if( ps->getNumEntries() < m_minNumberOfSamples)
		{
			*valid = false;
			return pos;
		}
	}

	*valid = true;
	return projPos;
}

X4::Vector3f X4::MLSProjector::project_RIMLS( const Vector3f &pos, Vector3f *projDerivate, bool* valid )
{
	if (valid == NULL)
		valid = &m_LastPointOK;

	int iterationCount = 0;
	Vector3f position = pos;
	Vector3f normal;
	float potential;
	float epsilon = m_MedPointDistance * 0.01f;
	*valid = true;
	do {
		Vector3f gradient;
		*valid = RIMLS_computePotentialAndGradient(position, gradient, potential);

		if (gradient.getSqrNorm() > 1e-6f)
		{
			normal = normalize(gradient);
			position = position - normal*potential;
		}

		if( iterationCount++ > 10 )
			*valid = false;
	} while ( fabs(potential)>epsilon && *valid);

	if (projDerivate)
		*projDerivate = normal;

	return position;
}

bool X4::MLSProjector::RIMLS_computePotentialAndGradient( const Vector3f &x, Vector3f &gradient_out, float &potential_out)
{
	PointSet * ps = m_RangeQuery->getSpherePoints(x, m_MedPointDistance * m_Sigma*2.5f, m_MedPointDistance * 1.7f);
	unsigned int nofSamples = ps->getNumEntries();

	if (nofSamples<m_minNumberOfSamples)
	{
		gradient_out = NULL_VECTOR3F;
		potential_out = 1e20f;
		//m_LastPointOK = false;
		return false;
	}

	// 01.2011 (c)art - this optimization is almost useless, it does not bring a lot of performance benefit (it is better to optimize for thread safety, hence multicore)
	/*static std::vector<float> cachedWeights;
	static std::vector<float> cachedWeightDerivatives;
	static std::vector<Vector3f> cachedWeightGradients;

	while( cachedWeights.size() < nofSamples )
	{
		cachedWeights.push_back(0);
		cachedWeightDerivatives.push_back(0);
		cachedWeightGradients.push_back(NULL_VECTOR3F);
	}*/

	std::vector<float> cachedWeights(nofSamples);
	std::vector<float> cachedWeightDerivatives(nofSamples);
	std::vector<Vector3f> cachedWeightGradients(nofSamples);
	
	// compute weights
	for (size_t i=0; i<nofSamples; i++)
	{
		//Scalar s = 1./(mPoints[id].cR()*mFilterScale);
		float s = 1.0f / m_MedPointDistance / m_Sigma;
		s = s*s;
		Vector3f p = ps->get3f(i,m_PositionAAT);
		float w = 1.0f - normQuad(x-p) * s;

		if( m_ProvideWeightChannel )
			w *= ps->get1f(i,m_WeightAAT);

		w = max(0.0f,w);
		float aux = w;
		w = w * w;
		w = w * w;
		cachedWeights[i] = w;
		cachedWeightDerivatives[i] = (-2. * s) * (4. * aux * aux * aux);
		cachedWeightGradients[i]  = (x - p) * cachedWeightDerivatives[i];
	}


	/*static std::vector<float> cachedRefittingWeights;
	while( cachedRefittingWeights.size() < nofSamples )
		cachedRefittingWeights.push_back(0);
	*/
	std::vector<float> cachedRefittingWeights(nofSamples);


	Vector3f source     = x;
	Vector3f grad = NULL_VECTOR3F;
	Vector3f previousGrad;
	float potentialPrev;
	Vector3f sumN = NULL_VECTOR3F;
	float potential      = 0.;
	float invSigma2      = 1.0f / (m_SigmaN*m_SigmaN);
	float invSigmaR2 = 0;
	if( m_SigmaR > 0 )
		invSigmaR2     = 1.0f / pow(m_SigmaR,2.0f);
	Vector3f sumGradWeight;
	Vector3f sumGradWeightPotential;
	float sumW;

	int iterationCount = 0;
	do
	{
		previousGrad = grad;
		sumGradWeight = NULL_VECTOR3F;
		sumGradWeightPotential = NULL_VECTOR3F;
		sumN = NULL_VECTOR3F;
		potentialPrev = potential;
		potential = 0.0f;
		sumW = 0.0f;

		for (unsigned int i=0; i<nofSamples; i++)
		{
			Vector3f diff = source - ps->get3f(i,m_PositionAAT);
			Vector3f normal = ps->get3f(i,m_NormalAAT);
			float f = diff * normal;

			float refittingWeight = 1;
			if (iterationCount > 0)
			{
				refittingWeight = exp(-normQuad(normal - previousGrad) * invSigma2);
				if (m_SigmaR>0)
				{
					float residual = (f - potentialPrev) / m_MedPointDistance;
					refittingWeight *= exp(-residual*residual * invSigmaR2);
				}
			}
			cachedRefittingWeights[i] = refittingWeight;
			float w = cachedWeights[i] * refittingWeight;
			Vector3f gw = cachedWeightGradients[i] * refittingWeight;

			sumGradWeight += gw;
			sumGradWeightPotential += gw * f;
			sumN += normal * w;
			potential += w * f;
			sumW += w;
		}

		if(sumW==0.)
		{
			//m_LastPointOK = false;
			return false;
		}

		potential /= sumW;
		grad = (-sumGradWeight*potential + sumGradWeightPotential + sumN) * (1./sumW);

		iterationCount++;

	} while ( (iterationCount < (int)m_MinRefittingIterations)
		|| ( normQuad(grad - previousGrad) > 1e-3f && iterationCount < (int)m_MaxRefittingIterations) );

	gradient_out   = grad;
	potential_out  = potential;

	return true;

/*	mCachedSumGradWeight = sumGradWeight;
	mCachedSumN = sumN;
	mCachedSumW = sumW;
	mCachedSumGradPotential = sumGradWeightPotential;*/
}