//---------------------------------------------------------------------------
#ifndef MiniBall_templCodeH
#define MiniBall_templCodeH
//---------------------------------------------------------------------------
#include "x3foundation_private.h"
//---------------------------------------------------------------------------
#include "MiniBall.h"
//---------------------------------------------------------------------------

namespace X4 {

	template<class FloatType>  MiniBall<FloatType>::MiniBall()
	{
		center = makeVector3f(-1e10f, -1e10f, -1e10f);
		radius = -1;
	}

	template<class FloatType>	MiniBall<FloatType>::MiniBall(
		const MiniBall<FloatType> &S)
	{
		radius = S.radius;
		center = S.center;
	}

	template<class FloatType>	MiniBall<FloatType>::MiniBall(
		const StaticVector<FloatType, 3> &O, 
		FloatType R)
	{
		radius = R;
		center = O;
	}

	template<class FloatType>	MiniBall<FloatType>::MiniBall(
		const StaticVector<FloatType, 3> &O, 
		const StaticVector<FloatType, 3> &A,
		const FloatType radiusEpsilon)
	{
		StaticVector<FloatType, 3> a = A - O;

		StaticVector<FloatType, 3> o = a*(FloatType)0.5;

		radius = sqrt(o.getSqrNorm()) + radiusEpsilon;
		center = O + o;
	}

	template<class FloatType>	MiniBall<FloatType>::MiniBall(
		const StaticVector<FloatType, 3> &O, 
		const StaticVector<FloatType, 3> &A, 
		const StaticVector<FloatType, 3> &B,
		const FloatType radiusEpsilon)
	{
		StaticVector<FloatType, 3> a = A - O;
		StaticVector<FloatType, 3> b = B - O;

		FloatType AA = a*a, BB = b*b, AB = a*b;

		FloatType det = AA*BB - AB*AB;

		StaticVector<FloatType, 3> o = (a*(AA*AA-AB*BB)+b*(-AB*AA+BB*BB))/det/2;

		radius = sqrt(o.getSqrNorm()) + radiusEpsilon;
		center = O + o;
	}

	template<class FloatType>	MiniBall<FloatType>::MiniBall(
		const StaticVector<FloatType, 3> &O, 
		const StaticVector<FloatType, 3> &A,
		const StaticVector<FloatType, 3> &B, 
		const StaticVector<FloatType, 3> &C,
		const FloatType radiusEpsilon)
	{
		StaticVector<FloatType, 3> a = A - O;
		StaticVector<FloatType, 3> b = B - O;
		StaticVector<FloatType, 3> c = C - O;

		FloatType AA = a*a, BB = b*b, CC = c*c,
							AB = a*b, AC = a*c, BC = b*c;

		FloatType det = AA*(BB*CC - BC*BC) + AB*(BC*AC-AB*CC) + AC*(AB*BC - BB*AC);

		FloatType x = (BB*CC - BC*BC)*AA +(AC*BC-AB*CC)*BB +(AB*BC-AC*BB)*CC,
			           y = (BC*AC - AB*CC)*AA +(AA*CC-AC*AC)*BB+(AB*AC-AA*BC)*CC, 
						  z = (AB*BC - BB*AC)*AA +(AB*AC-AA*BC)*BB +(AA*BB-AB*AB)*CC;


		StaticVector<FloatType, 3> o = (a*x+b*y+c*z) / det/2;

		radius = sqrt(o.getSqrNorm())+ radiusEpsilon;
		center = O + o;
	}

	template<class FloatType>	MiniBall<FloatType> &MiniBall<FloatType>::operator=(
		const MiniBall<FloatType> &ball)
	{
		radius = ball.radius;
		center = ball.center;

		return *this;
	}

	template<class FloatType>	FloatType MiniBall<FloatType>::d(
		const StaticVector<FloatType, 3> &position) const
	{
		return sqrt((position - center).getSqrNorm())- radius;
	}

	template<class FloatType>	FloatType MiniBall<FloatType>::d2(
		const StaticVector<FloatType, 3> &position) const
	{
		return (position - center).getSqrNorm() - radius*radius;
	}

	template<class FloatType>	FloatType MiniBall<FloatType>::d(
		const MiniBall<FloatType> &ball, const StaticVector<FloatType, 3> &position)
	{
		return sqrt((position - ball.center).getSqrNorm()) - ball.radius;
	}

	template<class FloatType>	FloatType MiniBall<FloatType>::d(
		const StaticVector<FloatType, 3> &position, const MiniBall<FloatType> &ball)
	{
		return sqrt((position - ball.center).getSqrNorm()) - ball.radius;
	}

	template<class FloatType>	FloatType MiniBall<FloatType>::d2(
		const MiniBall<FloatType> &ball, const StaticVector<FloatType, 3> &position)
	{
		return (position - ball.center).getSqrNorm()  - ball.radius*ball.radius;
	}

	template<class FloatType>	FloatType MiniBall<FloatType>::d2(
		const StaticVector<FloatType, 3> &position, const MiniBall<FloatType> &ball)
	{
		return (position - ball.center).getSqrNorm()  - ball.radius*ball.radius;
	}

	template<class FloatType>	MiniBall<FloatType> MiniBall<FloatType>::recurseMini(
		StaticVector<FloatType, 3> *Positions[], 
		card32 numOfPointsInside, 
		card32 numOfPointsOnBoundary,
		const FloatType radiusEpsilon)
	{
		MiniBall<FloatType> MB;

		switch(numOfPointsOnBoundary)
		{
		case 0:
			MB = MiniBall<FloatType>();
			break;
		case 1:
			MB = MiniBall<FloatType>(*Positions[-1], radiusEpsilon);
			break;
		case 2:
			MB = MiniBall<FloatType>(*Positions[-1], *Positions[-2], radiusEpsilon);
			break;
		case 3:
			MB = MiniBall<FloatType>(*Positions[-1], *Positions[-2], *Positions[-3], radiusEpsilon);
			break;
		case 4:
			MB = MiniBall<FloatType>(*Positions[-1], *Positions[-2], *Positions[-3], *Positions[-4], radiusEpsilon);
			return MB;
		}

		for(card32 i = 0; i < numOfPointsInside; i++)
		{
			if(MB.d2(*Positions[i]) > 0)   // Signed square distance to sphere
			{
				for(card32 j = i; j > 0; j--)
				{
					StaticVector<FloatType, 3> *T = Positions[j];
					Positions[j] = Positions[j - 1];
					Positions[j - 1] = T;
				}

				MB = recurseMini(Positions + 1, i, numOfPointsOnBoundary + 1, radiusEpsilon);
			}
		}
		return MB;
	}

	template<class FloatType>	MiniBall<FloatType> MiniBall<FloatType>::miniBall(
		StaticVector<FloatType, 3>  Positions[], 
		card32 numOfPointsInside, 
		FloatType radiusEpsilon)
	{
		StaticVector<FloatType, 3>  **L = new StaticVector<FloatType, 3> *[numOfPointsInside];

		for(card32 i = 0; i < numOfPointsInside; i++)
			L[i] = &Positions[i];

		MiniBall<FloatType> MB = recurseMini(L, numOfPointsInside, 0, radiusEpsilon);

		delete[] L;

		return MB;
	}

	template<class FloatType>	MiniBall<FloatType> MiniBall<FloatType>::smallBall(
		StaticVector<FloatType, 3>  Positions[],
		card32 numOfPointsInside,
		FloatType radiusEpsilon)
	{
		StaticVector<FloatType, 3> center;
		FloatType radius = -1;

		if(p > 0)
		{
			center = 0;

			for(card32 i = 0; i < p; i++)
				center += Positions[i];

			center /= (FloatType)numOfPointsInside;

			for(card32 i = 0; i < numOfPointsInside; i++)
			{
				FloatType d2 = (Positions[i] - center).getSqrNorm();
				
				if(d2 > radius)
					radius = d2;
			}

			radius = sqrtf(radius) + radiusEpsilon;
		}

		return MiniBall<FloatType>(center, radius);
	}

	template<class FloatType>	MiniBall<FloatType> MiniBall<FloatType>::miniBall(
		std::vector<StaticVector<FloatType, 3> >  *positions, 
		card32 numOfPointsInside, 
		FloatType radiusEpsilon)
	{
		x4Assert(positions != NULL && positions->size() >= numOfPointsInside);
		StaticVector<FloatType, 3>  **L = new StaticVector<FloatType, 3> *[numOfPointsInside];

		for(card32 i = 0; i < numOfPointsInside; i++)
			L[i] = &(*positions)[i];

		MiniBall<FloatType> MB = recurseMini(L, numOfPointsInside, 0, radiusEpsilon);

		delete[] L;

		return MB;
	}

	template<class FloatType>	MiniBall<FloatType> MiniBall<FloatType>::smallBall(
		std::vector<StaticVector<FloatType, 3> >  *positions,
		card32 numOfPointsInside,
		FloatType radiusEpsilon)
	{
		x4Assert(positions != NULL && positions->size() >= numOfPointsInside);
		StaticVector<FloatType, 3> center;
		FloatType radius = -1;

		if(p > 0)
		{
			center = 0;

			for(card32 i = 0; i < p; i++)
				center += positions[i];

			center /= (FloatType)numOfPointsInside;

			for(card32 i = 0; i < numOfPointsInside; i++)
			{
				FloatType d2 = (positions[i] - center).getSqrNorm();

				if(d2 > radius)
					radius = d2;
			}

			radius = sqrtf(radius) + radiusEpsilon;
		}

		return MiniBall<FloatType>(center, radius);
	}
}

#endif
 