//---------------------------------------------------------------------------
#ifndef Monomials2ndOrderH
#define Monomials2ndOrderH
//---------------------------------------------------------------------------
#include "Persistent.h"
#include "x3foundation_private.h"
//---------------------------------------------------------------------------

namespace X4 {


// --- 1st order


/// this class is compatible with HFBFBivariateMonomials of degree 2
/// (i.e. the ordering of basis functions is the same)
class X3FOUNDATION_PRIVATE_API Monomials1stOrder : public Persistent {
	X4_CLASS( Monomials1stOrder )
 private:
	
 public:
	static const card32 NUM_BASIS_FUNS = 3;

	/// computes a least squares fit and its derivatives for moving the orginal points in normal direction
	/// a Gaussian window can be activated for weighted least squares by setting "gaussianWindowVariance" to a positive value
	static void computeCoeffs( const std::vector<Vector3f> &points,
                              StaticVector<float32, 3> &coeffs,
                              std::vector< StaticVector<float32,3> > &coeffsDerivs,
                              float32 gaussianWindowVariance = -1 );
	/// same as above, but no derivatives (slightly faster)
	static void computeCoeffs( const std::vector<Vector3f> &points,
                              StaticVector<float32, 3> &coeffs,
                              float32 gaussianWindowVariance = -1 );
	static float32 evaluate( const StaticVector<float32,3> &coeffs,
                            const Vector2f &position );
};


/// this class fits a affine 3x3 matrix + contstant std::vector mapping to sparse 3d -> 3d point mappings
class X3FOUNDATION_PRIVATE_API AffineMapping3dFit {
 public:
   struct MapPoint {
      Vector3f from;
      Vector3f to;
   };
   struct AffineMap {
      Matrix3f M;
      Vector3f b;
   };

	/// computes a least squares fit and its derivatives for moving the orginal points in normal direction
	/// a Gaussian window can be activated for weighted least squares by setting "gaussianWindowVariance" to a positive value
	static void computeCoeffs( const std::vector<MapPoint> &points,
                              AffineMap &leastSquareFit,
                              std::vector< AffineMap > &coeffsDerivs,
									   float32 inversionSingularValThreshold = 1E-5,
                              float32 gaussianWindowVariance = -1 );
	/// same as above, but no derivatives (slightly faster)
	static void computeCoeffs( const std::vector<MapPoint> &points,
                              AffineMap &leastSquareFit,
									   float32 inversionSingularValThreshold = 1E-5,
                              float32 gaussianWindowVariance = -1 );
	/// project affine mapping to least-squares closest orthogonal std::map
	static void orthoghonalize(AffineMap &map);
};


// --- 2nd order


/// this class is compatible with HFBFBivariateMonomials of degree 2
/// (i.e. the ordering of basis functions is the same)
class X3FOUNDATION_PRIVATE_API Monomials2ndOrder : public Persistent {
	X4_CLASS( Monomials2ndOrder )
 private:
	
 public:
	static const card32 NUM_BASIS_FUNS = 6;

	/// computes a least squares fit and its derivatives for moving the original points in normal direction
	/// a Gaussian window can be activated for weighted least squares by setting "gaussianWindowVariance" to a positive value
	static void computeCoeffs( const std::vector<Vector3f> &points,
                              StaticVector<float32, 6> &coeffs,
                              std::vector< StaticVector<float32,6> > &coeffsDerivs,
                              float32 gaussianWindowVariance = -1 );
	/// same as above, but no derivatives (slightly faster)
	static void computeCoeffs( const std::vector<Vector3f> &points,
                              StaticVector<float32, 6> &coeffs,
                              float32 gaussianWindowVariance = -1 );

	static void computeCoeffs( const Vector3f * dataPtr, card32 numPoints, StaticVector<float32, 6> &coeffs,
		float32 gaussianWindowVariance = -1);

	static void computeCoeffs( const Vector3f * dataPtr, const float * weightPtr, card32 numPoints, StaticVector<float32, 6> &coeffs );

	static float32 evaluate( const StaticVector<float32,6> &coeffs,
                            const Vector2f &position );

	static Vector3f evaluateNormal( const StaticVector<float32,6> &coeffs,
                            const Vector2f &position );

	/// computes critical point of a patch ( both derivates 0), returns true if patch has a critical point
	static bool evaluateCriticalPoint( const StaticVector<float32,6> &coeffs, Vector3f &criticalPoint );

	/// computes principal curvatures
	static void computeCurvatures(const StaticVector<float32,6> &coeffs, float &k1, float &k2, Vector2f &k1_dir, Vector2f &k2_dir );

};



/**
computes least squares fit of 3rd order 1d function to a set of data points
polynomial: f(x) = c0 + c1 x + c2 x^2 + c3 x^3
*/
class X3FOUNDATION_PRIVATE_API Monomials2ndOrder1D
{

public:

	/// computes least squares fit of 3rd order 1d function to a set of data points
	/// y components are ignored
	static Vector3f computeCoeffs( Vector2f * dataPoints , float * weights, const unsigned numPoints);
};



//class ANIMATIONRECONSTRUCTION_API Monomials2ndOrder3D {
// public:
//	static const card32 NUM_BASIS_FUNS = 10;
//
//   // 1  u  u^2  v  v^2  uv  w  wu  wv  w^2
//
//   static void computeInvDesignMatrix( const std::vector<Vector4f> &points,
//                                       StaticMatrix<float32, 10, 10> &invDM);
//	static void computeCoeffs( const std::vector<Vector4f> &points,
//	                           const StaticMatrix<float32, 10, 10> &invDM,
//                              StaticVector<float32, 10> &coeffs);
//	static float32 evaluate( const StaticVector<float32,10> &coeffs,
//                            const Vector3f &position );
//};


extern void Monomials2ndOrder_unitTest();

}

#endif