#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "Monomials3rdOrder.h"
//---------------------------------------------------------------------------
using namespace X4;


X4::Vector4f X4::Monomials3rdOrder1D::computeCoeffs( Vector3f * dataPoints , float * weights, const unsigned numPoints )
{
	// compute gradient of squared difference function & solve grad = 0
	Matrix4f mat = IDENTITY4F * 0;
	Vector4f v = NULL_VECTOR4F;

	for( unsigned i=0;i<numPoints;i++ )
	{
		const float x = dataPoints[i][0];
		const float y = dataPoints[i][2];
		const float weight = weights[i];

		const float x2 = x*x;
		const float x3 = x2*x;
		const float x4 = x3*x;
		const float x5 = x4*x;
		const float x6 = x5*x;

		v[0] += y * weight;
		v[1] += y * weight * x;
		v[2] += y * weight * x2;
		v[3] += y * weight * x3;

		mat[0][0] += weight;		mat[0][1] += weight*x;		mat[0][2] += weight*x2;		mat[0][3] += weight*x3;
		mat[1][0] += weight*x;	mat[1][1] += weight*x2;		mat[1][2] += weight*x3;		mat[1][3] += weight*x4;
		mat[2][0] += weight*x2;	mat[2][1] += weight*x3;		mat[2][2] += weight*x4;		mat[2][3] += weight*x5;
		mat[3][0] += weight*x3;	mat[3][1] += weight*x4;		mat[3][2] += weight*x5;		mat[3][3] += weight*x6;
	}

	Matrix4f invDm;
	bool success;
	invDm = invertMatrix(mat, &success, 1e-7f );
	if (!success) 
	{
		return makeVector4f(0,0,0,0);
	}
	return invDm * v;
}

float X4::Monomials3rdOrder1D::evalate( const Vector4f & coeffs, float x )
{
	return  coeffs[0] 
			+ coeffs[1] * x
			+ coeffs[2] * x*x
			+ coeffs[3] * x*x*x;
}

bool X4::Monomials3rdOrder1D::computeCurvatureMinMax( const Vector4f &coeffs, float &result )
{
	float bestValue = 0;
	for( int i=-100;i<=100;i++ )
	{
		float x = (float)i / 100.0f;
		float curv = fabs(evaluateCurvature(coeffs,x));
		if( curv > bestValue )
		{
			result = x;
			bestValue = curv;
		}
	}
	return 0.995f - fabs(result) > 0.0f;
}

float X4::Monomials3rdOrder1D::evaluateD1( const Vector4f & coeffs, float x )
{
	return coeffs[1] +
		2.0f * coeffs[2] * x +
		3.0f * coeffs[3] * x * x;
}

float X4::Monomials3rdOrder1D::evaluateD2( const Vector4f & coeffs, float x )
{
	return 2.0f * coeffs[2] +
		6.0f * coeffs[3] * x;
}

float X4::Monomials3rdOrder1D::evaluateCurvature( const Vector4f & coeffs, float x )
{
	return evaluateD2(coeffs,x) / (1.0f + pow(evaluateD1(coeffs,x),2.0f));
}

typedef double myOptType;

StaticVector<float,5> X4::Monomials4thOrder1D::computeCoeffs( Vector3f * dataPoints , float * weights, const unsigned numPoints )
{
	// compute gradient of squared difference function & solve grad = 0
	StaticMatrix<myOptType,5,5> mat;
	mat *= 0;
	static StaticVector<myOptType,5> nullVector5f = StaticVector<myOptType,5>() * 0;
	StaticVector<myOptType,5> v = nullVector5f;

	for( unsigned i=0;i<numPoints;i++ )
	{
		const myOptType x = dataPoints[i][0];
		const myOptType y = dataPoints[i][2];
		const myOptType weight = weights[i];

		const myOptType x2 = x*x;
		const myOptType x3 = x2*x;
		const myOptType x4 = x3*x;
		const myOptType x5 = x4*x;
		const myOptType x6 = x5*x;
		const myOptType x7 = x6*x;
		const myOptType x8 = x7*x;

		v[0] += y * weight;
		v[1] += y * weight * x;
		v[2] += y * weight * x2;
		v[3] += y * weight * x3;
		v[4] += y * weight * x4;

		mat[0][0] += weight;		mat[0][1] += weight*x;		mat[0][2] += weight*x2;		mat[0][3] += weight*x3;		mat[0][4] += weight*x4;
		mat[1][0] += weight*x;	mat[1][1] += weight*x2;		mat[1][2] += weight*x3;		mat[1][3] += weight*x4;		mat[1][4] += weight*x5;
		mat[2][0] += weight*x2;	mat[2][1] += weight*x3;		mat[2][2] += weight*x4;		mat[2][3] += weight*x5;		mat[2][4] += weight*x6;
		mat[3][0] += weight*x3;	mat[3][1] += weight*x4;		mat[3][2] += weight*x5;		mat[3][3] += weight*x6;		mat[3][4] += weight*x7;
		mat[4][0] += weight*x4;	mat[4][1] += weight*x5;		mat[4][2] += weight*x6;		mat[4][3] += weight*x7;		mat[4][4] += weight*x8;
	}

	StaticMatrix<myOptType,5,5> invDm;
	bool success;
	invDm = invertMatrix<myOptType,5>(mat, &success, 1e-7f );
	if (!success) 
	{
		return StaticVector<float,5>() * 0;
	}
	v = invDm * v;
	StaticVector<float,5> coeffs;
	for( unsigned i=0;i<5;i++ )
		coeffs[i] = v[i];

	return coeffs;
}

bool X4::Monomials4thOrder1D::computeCurvatureMinMax( const StaticVector<float,5> &coeffs, float &result )
{
	if( fabs(coeffs[4]) > 1e-7f )
	{
		result = coeffs[3] / coeffs[4] / 4.0f;
		return true;
	}
	return false;
}

float X4::Monomials4thOrder1D::evalate( const StaticVector<float,5> & coeffs, float x )
{
	return  coeffs[0] 
	+ coeffs[1] * x
		+ coeffs[2] * x*x
		+ coeffs[3] * x*x*x
		+ coeffs[4] * x*x*x*x;

}