#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "MultiPassRenderer.h"
//---------------------------------------------------------------------------
#include "Scene.h"
#include "RenderTextureAttachment.h"
#include "RGBATextureAttachment.h"
#include "MaterialAttachment.h"
#include "GLRenderTextureFBO.h"
#include "DebugOutput.h"
#include "StandardSceneGraphIterator.h"
#include "LCFilters.h"
#include "MemoryManager.h"
#include "SceneGraphNode.h"
#include "SGObjectNode.h"
#include "SceneObject.h"
#include "GLMaterial.h"
#include "SGCameraNode.h"
#include "ViewFrustum.h"
#include "SGIndexedAndByNameSceneGraphIteratorAdaptor.h"
#include "CameraController.h"
#include "GLShaderMaterial.h"
#include "Configuration.h"
#include "Shader.h"
#include "GLHelper.h"
#include "SGTransformationNode.h"
#include "PointCloudTools.h"
#include "ApplicationDirectories.h"
#include "GLDebuggingTools.h"
#include <QtCore\QRegExp>
//---------------------------------------------------------------------------
using namespace X4;

IMPLEMENT_X4_ABSTRACT_CLASS( MultiPassRenderer ,0)
{
	BEGIN_CLASS_INIT( MultiPassRenderer );
}

IMPLEMENT_X4_CLASS( LCMultiPassRender_DrawNode,0)
{
	BEGIN_CLASS_INIT( LCMultiPassRender_DrawNode );
}


MultiPassRenderer::MultiPassRenderer(void)
{
	m_DrawNodeState.material = NULL;
	m_MemoryManager = NULL;
	m_ScreenWidth = m_CurrentWidth = 640;
	m_ScreenHeight = m_CurrentHeight = 480;
	m_ScreenSizeChanged = true;

	m_Mat_DrawRTToScreen = NULL;
	// init draw rt to screen material
	m_Mat_DrawRTToScreen = new GLShaderMaterial;
	if( !m_Mat_DrawRTToScreen->importFromTextFile( ApplicationDirectories::getResourcesDir() 
		+ "shaders/hqRenderer/MR_DrawRTToScreen.txt" ) )
	{
		debugOutput << "Couldn't find " + ApplicationDirectories::getResourcesDir() 
		+ "shaders/hqRenderer/MR_DrawRTToScreen.txt" + "\n";
	}

	m_Mat_DrawRTToScreen->getVertexShader()->setShaderType( Shader::VERTEX );
   m_Mat_DrawRTToScreen->getVertexShader()->setShaderClass( Shader::CLASS_CG );
   m_Mat_DrawRTToScreen->getFragmentShader()->setShaderClass( Shader::CLASS_CG );

	m_Traversal = NULL;

}

MultiPassRenderer::~MultiPassRenderer(void)
{
	delete m_Traversal;
	delete m_Mat_DrawRTToScreen;
}

void MultiPassRenderer::assign(const Object* obj, X4CopyContext *context)
{
	Renderer::assign(obj,context);
	const MultiPassRenderer * o = dynamic_cast<const MultiPassRenderer*>(obj);
	x4Assert( o != NULL );

}


void MultiPassRenderer::connect(Scene* scene, GLContext* ctxt) 
{
	Renderer::connect(scene,ctxt);

   x4Assert( m_MemoryManager == NULL );
   m_MemoryManager = new MemoryManager( 1024 * 1024 * 1024 ); // max. 256 MB for vertex buffers

	if( m_Traversal == NULL )
	{
		m_Traversal = new LCTraversal( new LCState );

		// inner nodes processing
		m_Traversal->addSubdividedNodeProcessors(new LCDrawInnerBoxesNodeProcessor());

		// processing for inner nodes at which the traversal stops
		m_Traversal->addNodeProcessors(new LCDrawBoxesNodeProcessor());
		m_Traversal->addNodeProcessors(new LCMultiPassRender_DrawNode());
	//	m_Traversal->addNodeProcessors(new LCDrawPointsMaterial());

		// leaf node processing
		m_Traversal->addUnSubdividableNodesProcessors(new LCDrawBoxesNodeProcessor());
		m_Traversal->addUnSubdividableNodesProcessors( new LCMultiPassRender_DrawNode() );
	//   m_Traversal->addUnSubdividableNodesProcessors( new LCDrawPointsMaterial() );
		
		// traversal
		m_Traversal->addNodeFilters(new LCViewFrustumCulling());
		m_Traversal->addNodeFilters(new LCProjectiveClassification());
	}

}

void MultiPassRenderer::render(card32 viewportWidth, card32 viewportHeight) 
{
	// store screensize
	m_ScreenSizeChanged = !( (m_ScreenWidth == viewportWidth) && (m_ScreenHeight == viewportHeight) );
	m_ScreenWidth = m_CurrentWidth = viewportWidth;
	m_ScreenHeight = m_CurrentHeight = viewportHeight;

	// store camera & viewfrustum
	m_DefaultCamera = getScene()->getRootState()->staticState->camera;
	m_DefaultViewFrustum = getScene()->getRootState()->staticState->viewFrustum;
	glGetFloatv( GL_PROJECTION_MATRIX, m_DefaultProjectionMatrix.data());
	glGetFloatv( GL_MODELVIEW_MATRIX, m_DefaultModelviewMatrix.data());

	// call descendants rendering method
	renderMultiPass();

	// cleanup
	if( m_CurrentRenderTarget.length()>0 )
	{
		// deactivate current render texture
		x4Assert( renderTargetExists( m_CurrentRenderTarget ));
		m_RenderTargets[m_CurrentRenderTarget].rt->deactivate();
		X4_GL_FB_DBG

		m_CurrentRenderTarget = "";
	}

	setDefaultCamera();
}

void MultiPassRenderer::disconnect(GLContext* ctxt) 
{
	eraseRenderTargets();
	X4_GL_FB_DBG

	x4Assert( m_MemoryManager != NULL );
   delete m_MemoryManager;
   m_MemoryManager = NULL;
}

void MultiPassRenderer::setRenderTarget( std::string rt_name, card32 flags, card32 width, card32 height ) 
{
	x4Assert( (flags & (RT_RGBA16 & RT_RGBA32)) == 0 ); // allow only 16 OR 32 bit float

	// --------------------------------------------------------------
	// delete existing render texture if specifications are different
	if( renderTargetExists( rt_name ) )
	{
		bool validRenderTexture = true;
		validRenderTexture &= m_RenderTargets[rt_name].rt->getWidth() == width;
		validRenderTexture &= m_RenderTargets[rt_name].rt->getHeight() == height;
		if( flags != RT_Depth )
		{
			// color texture
			validRenderTexture &= m_RenderTargets[rt_name].rt->getNumComponents() == 4;
			validRenderTexture &= m_RenderTargets[rt_name].rt->hasDepth() == ((flags & RT_Depth)!=0);
			validRenderTexture &= m_RenderTargets[rt_name].rt->isFloat16() == ((flags & RT_RGBA16)!=0);
			validRenderTexture &= m_RenderTargets[rt_name].rt->isFloat32() == ((flags & RT_RGBA32)!=0);
		}
		else
		{
			// pure depth texture
			validRenderTexture &= m_RenderTargets[rt_name].rt->isPureDepthTexture();
		}

		if( !validRenderTexture )
		{
			delete m_RenderTargets[rt_name].rt;
			m_RenderTargets.erase(rt_name);
		}
	}


	// --------------------------------------------------------------
	// create render texture (if not already exists)
	if( !renderTargetExists( rt_name ) )
	{
		// create render texture
		TRenderTarget tmp;
		tmp.target_name = rt_name;
		tmp.rt = new GLRenderTextureFBO( getGLContext() );
		X4_GL_FB_DBG
		if( flags != RT_Depth )
		{
			// create color-rt
			card32 rt_flags = flags & RT_2D ? 0 : GLRenderTextureFBO::FLAG_RECTANGLE;
			rt_flags |= (flags & RT_Depth ) ? GLRenderTextureFBO::FLAG_DEPTH : 0;
			rt_flags |= (flags & RT_RGBA16 ) ? GLRenderTextureFBO::FLAG_FLOAT16 : 0;
			rt_flags |= (flags & RT_RGBA32 ) ? GLRenderTextureFBO::FLAG_FLOAT32 : 0;
			rt_flags |= (flags & RT_MIPMAP) ? GLRenderTextureFBO::FLAG_MIPMAP : 0;
			bool create = tmp.rt->create(width, height, 4, rt_flags );
			x4Assert( create );
			X4_GL_FB_DBG
		}
		else
		{
			// create depth-rt
			bool create = tmp.rt->createDepthTexture(width, height);
			x4Assert( create );
			X4_GL_FB_DBG
		}
		m_RenderTargets[ rt_name ] = tmp;

	}
	X4_GL_FB_DBG

	
	// unbind all render textures
	unbindAllRenderTargets();
	X4_GL_FB_DBG
	
	// deactivate previous target
	if( m_CurrentRenderTarget != rt_name )
		if( m_CurrentRenderTarget.length() > 0 )
			m_RenderTargets[ m_CurrentRenderTarget ].rt->deactivate();
	X4_GL_FB_DBG
	
	// activate render target
	m_CurrentRenderTarget = rt_name;
	if( !m_RenderTargets[ rt_name ].rt->isActive() )
	{
		m_RenderTargets[ rt_name ].rt->activate();
		X4_GL_FB_DBG
	}
	
	m_CurrentWidth = width;
	m_CurrentHeight = height;

}

void MultiPassRenderer::setDefaultRenderTarget() 
{
	// unbind all render textures
	unbindAllRenderTargets();
	X4_GL_FB_DBG
	
	// deactivate previous target
	if( m_CurrentRenderTarget.length() > 0 )
		m_RenderTargets[ m_CurrentRenderTarget ].rt->deactivate();
	X4_GL_FB_DBG

	m_CurrentRenderTarget = "";
	m_CurrentWidth = m_ScreenWidth;
	m_CurrentHeight = m_ScreenHeight;
}

void MultiPassRenderer::setMaterial( GLMaterial * mat ) 
{
	m_DrawNodeState.material = mat;
	getScene()->setHQRMaterial( mat );
}

void MultiPassRenderer::setMaterialFromScene( card32 index ) 
{
	if( index < getScene()->getNumGLMaterials() )
		setMaterial( getScene()->getGLMaterial(index) );
	//m_DrawNodeState.matIndex = index;
}

void MultiPassRenderer::resetMaterial() 
{
	setMaterial(NULL);
}

void MultiPassRenderer::renderScene(bool ignoreVisibilityFlag, bool useObjMaterial, string pathWildcards) 
{
   Scene *scene = getScene();
   glPushAttrib(GL_ALL_ATTRIB_BITS);

   try {

	   StandardSceneGraphIterator sceneIt;
	   sceneIt.clearAndSetup(scene->getRootNode(), scene->getRootState(), SGDynamicState::tmRendering, 0, 0);
		renderSceneRecursive(&sceneIt, IDENTITY4F, ignoreVisibilityFlag, useObjMaterial, pathWildcards);

   } catch(...) {
      glPopAttrib();
		throw;
   }

   glPopAttrib();


}

void MultiPassRenderer::bindTarget( std::string rt_name, std::string texture_name ) 
{
	if( texture_name.length() == 0 )
	{
		debugOutput << "MultiPassRenderer::bindTarget() - You have to spefify texture_name when using bindTarget!\n";
		return;
	}

	if( !renderTargetExists( rt_name ) )
	{
		debugOutput << "MultiPassRenderer::bindTarget() - Trying to bind unkown render target " << rt_name << "!\n";
		return;
	}

	// check if attachment-name is already in use
	AttachedData * data = getScene()->getAttachments()->getData( texture_name );
	RenderTextureAttachment * rta = dynamic_cast<RenderTextureAttachment*>(data);
	if( data != NULL && rta == NULL )
	{
		debugOutput << "MultiPassRenderer::bindTarget() - texture_name[" << texture_name << "] already in use by another tool!\n";
		return;
	}

	TRenderTarget t = m_RenderTargets[rt_name];

	// check if already bound with other name
	if( t.texture_name.length() > 0 )
	{
		if( t.texture_name == texture_name )
			return;	// already bound
		else
		{
			// remove attachment
			RenderTextureAttachment * rta = dynamic_cast<RenderTextureAttachment*>(
				getScene()->getAttachments()->getData( t.texture_name ) );
			delete rta;
			m_RenderTargets[rt_name].texture_name = "";
			
			// unbind texture
			if( t.rt->isBound() )
				t.rt->unbind();
			X4_GL_FB_DBG
		}
	}

	// try to find other rendertarget which is already bound to texture_name
	for( std::map<std::string,TRenderTarget>::iterator iter = m_RenderTargets.begin();
		iter != m_RenderTargets.end(); ++iter )
	{
		if( iter->second.target_name != rt_name )
			if( iter->second.texture_name == texture_name )
			{
				// remove attachment
				RenderTextureAttachment * rta = dynamic_cast<RenderTextureAttachment*>(
					getScene()->getAttachments()->getData( texture_name ) );
				delete rta;
				iter->second.texture_name = "";
				
				// unbind texture
				if( iter->second.rt->isBound() )
					iter->second.rt->unbind();
			}
	}

	// create attachment (indirect texture binding)
	rta = new RenderTextureAttachment;
	rta->setup( texture_name, 0 );
	rta->setRenderTexture( m_RenderTargets[rt_name].rt );
	X4_GL_FB_DBG
	getScene()->getAttachments()->attachData( rta );
}


void MultiPassRenderer::bindTarget( std::string target_name, card32 texture_stage ) 
{
	if( !renderTargetExists( target_name ) )
	{
		debugOutput << "MultiPassRenderer::bindTarget() - Trying to bind unkown render target " << target_name << "!\n";
		return;
	}

	TRenderTarget t = m_RenderTargets[target_name];
	if( t.rt->isBound() )
		t.rt->unbind();
	X4_GL_FB_DBG

	t.rt->bind(texture_stage);
	X4_GL_FB_DBG

}

void MultiPassRenderer::bindTexture( std::string texture_name, RGBAImage * image  ) 
{
	// check if attachment already exists or in use
	AttachedData * data = getScene()->getAttachments()->getData( texture_name );
	RGBATextureAttachment * texAtt = dynamic_cast<RGBATextureAttachment*>(data);
	if( data != NULL && texAtt == NULL )
	{
		debugOutput << "MultiPassRenderer::bindTexture() - texture_name[" << texture_name << "] already in use by another tool!\n";
		return;
	}
	delete texAtt;

	// uploads every frame ...
	texAtt = new RGBATextureAttachment;
	texAtt->setup( texture_name,0 );
	texAtt->createFromImage( image );
	getScene()->getAttachments()->attachData( texAtt );
}

void MultiPassRenderer::setMaterialParameter1f( std::string param_name, float f ) 
{
	float tmp[4] = {0,0,0,0};
	tmp[0] = f;
	setMaterialParameterxfv( param_name, 1, tmp );
}

void MultiPassRenderer::setMaterialParameter2f( std::string param_name, float f1, float f2 ) 
{
	float tmp[4] = {0,0,0,0};
	tmp[0] = f1;
	tmp[1] = f2;
	setMaterialParameterxfv( param_name, 2, tmp );
}

void MultiPassRenderer::setMaterialParameter3f( std::string param_name, float f1, float f2, float f3 ) 
{
	float tmp[4] = {0,0,0,0};
	tmp[0] = f1;
	tmp[1] = f2;
	tmp[2] = f3;
	setMaterialParameterxfv( param_name, 3, tmp );
}

void MultiPassRenderer::setMaterialParameter4f( std::string param_name, float f1, float f2, float f3, float f4 ) 
{
	float tmp[4] = {0,0,0,0};
	tmp[0] = f1;
	tmp[1] = f2;
	tmp[2] = f3;
	tmp[3] = f4;
	setMaterialParameterxfv( param_name, 4, tmp );
}


void MultiPassRenderer::setMaterialParameter4fv( std::string param_name, const Vector4f &vec ) 
{
	setMaterialParameterxfv( param_name, 4, (float*)&vec[0] );
}

void MultiPassRenderer::setMaterialParameter4fv( std::string param_name, float * f ) 
{
	setMaterialParameterxfv( param_name, 4, f );
}

void MultiPassRenderer::setMatrix4f( std::string param_name, const Matrix4f &mat ) 
{
	AttachedData * data = getScene()->getAttachments()->getData( param_name );
	Matrix4fAttachment * att = dynamic_cast<Matrix4fAttachment*>(data);

	if( data != NULL && att == NULL )
	{
		debugOutput << "MultiPassRenderer::setMatrix4f() - parameter name \"" << param_name << "\" already in use by another tool!\n";
		return;
	}

	if( att == NULL )
	{
		att = new Matrix4fAttachment;
		att->setup( param_name,0 );
		getScene()->getAttachments()->attachData( att );
	}

	att->setMatrix( mat );
}

void MultiPassRenderer::setMatrix4f( std::string param_name, float * mat ) 
{
	Matrix4f m;
	float * fPtr = &m[0][0];
	for( card32 i=0;i<4*4;i++ )
		fPtr[i] = mat[i];
	setMatrix4f( param_name, m );
}

void MultiPassRenderer::setMaterialParameter1i( std::string param_name, int32 f ) 
{
	int32 tmp[4] = {0,0,0,0};
	tmp[0] = f;
	setMaterialParameterxiv( param_name, 1, tmp );
}


void MultiPassRenderer::setMaterialParameterxfv( std::string param_name, card32 numComponents, float * f ) 
{
	AttachedData * data = getScene()->getAttachments()->getData( param_name );
	Vector4fAttachment * att = dynamic_cast<Vector4fAttachment*>(data);

	if( data != NULL && att == NULL )
	{
		debugOutput << "MultiPassRenderer::setMaterialParameterxfv() - parameter name \"" << param_name << "\" already in use by another tool!\n";
		return;
	}

	if( att == NULL )
	{
		att = new Vector4fAttachment;
		att->setup( param_name,0 );
		getScene()->getAttachments()->attachData( att );
	}
	
	Vector4f vec = makeVector4f( f[0],f[1],f[2],f[3] );
	att->setVector( vec );
	att->setNumComponentsUsed( numComponents );
}

void MultiPassRenderer::setMaterialParameterxiv( std::string param_name, card32 numComponents, int32* f ) 
{
	AttachedData * data = getScene()->getAttachments()->getData( param_name );
	Vector4iAttachment * att = dynamic_cast<Vector4iAttachment*>(data);

	if( data != NULL && att == NULL )
	{
		debugOutput << "MultiPassRenderer::setMaterialParameterxiv() - parameter name \"" << param_name << "\" already in use by another tool!\n";
		return;
	}

	if( att == NULL )
	{
		att = new Vector4iAttachment;
		att->setup( param_name,0 );
		getScene()->getAttachments()->attachData( att );
	}

	Vector4i vec = makeVector4i( f[0],f[1],f[2],f[3] );
	att->setVector( vec );
	att->setNumComponentsUsed( numComponents );
}

void MultiPassRenderer::eraseRenderTargets() 
{
	for( std::map<std::string, TRenderTarget>::iterator iter = m_RenderTargets.begin();
		iter != m_RenderTargets.end(); ++iter )
	{
		delete iter->second.rt;
	}
	m_RenderTargets.clear();
}

bool MultiPassRenderer::renderTargetExists( std::string target_name ) 
{
	return m_RenderTargets.find( target_name ) != m_RenderTargets.end();
}

void MultiPassRenderer::unbindAllRenderTargets() 
{
	// unbind all bound textures
	for( std::map<std::string,TRenderTarget>::iterator iter = m_RenderTargets.begin();
		iter != m_RenderTargets.end(); ++iter )
	{

		// remove attachment
		if( iter->second.texture_name.length() > 0 )
		{
			RenderTextureAttachment * rta = dynamic_cast<RenderTextureAttachment*>(
				getScene()->getAttachments()->getData( iter->second.texture_name ) );
			if( rta )
				getScene()->getAttachments()->deleteData( iter->second.texture_name );

			iter->second.texture_name = "";
		}

		// unbind texture
		if( iter->second.rt->isBound() )
			iter->second.rt->unbind();
	}
}

#include "RenderableObject.h"
void LCMultiPassRender_DrawNode::processNode(X4::LCState *state, X4::SceneObjectIterator *node)
{
   glPushAttrib(GL_ALL_ATTRIB_BITS);

	card32 matInd = state->sceneObject->getMaterialIndex();

	GLMaterial *material = NULL;

	MultiPassRenderer * renderer = dynamic_cast<MultiPassRenderer*>(state->renderer);
	if( renderer != NULL )
		material = state->scene->getHQRMaterial();//renderer->getDrawNodeState().material;
	if( material == NULL )
		material = state->scene->getGLMaterial(matInd);
		

	//GLMaterial * test_material = state->scene->getGLMaterial(matInd);


	if (material == NULL) return;

	// render attachments
	{
		AttachedDataContainer *attachments = state->sceneObject->getAttachments();
		std::vector<AttachedData*> allAtt;
		attachments->getAll(allAtt);
		for (size_t i=0; i<allAtt.size(); i++)
		{
			RenderableObject* renderableObj = dynamic_cast<RenderableObject*>(allAtt[i]);
			if (renderableObj != NULL)
				renderableObj->renderGL(state->sceneObject);
		}
	}

	Vector4f globalColor;
	if( state->sceneGraphState->dynamicState->getStateVariable("SC_globalColor", globalColor) )
	{
		GLShaderMaterial * shaderMat = dynamic_cast<GLShaderMaterial*>(material);
		if( shaderMat)
			shaderMat->setParameter4fv(state->glContext,"globalColor",globalColor.data());
	}

   material->draw(state->scene, state->sceneGraphState, state->sceneObject, node, state->glContext, state->memManager );
   glPopAttrib();

}

void MultiPassRenderer::setCamera( std::string camera_name ) 
{
   StandardSceneGraphIterator sceneIt;
   sceneIt.clearAndSetup(getScene()->getRootNode(), getScene()->getRootState(), SGDynamicState::tmRendering, SceneGraphNode::FLAG_VISIBLE, 0);
	SGIndexedAndByNameSceneGraphIteratorAdaptor nameIter;
	nameIter.clearAndSetup( &sceneIt );
	if( !nameIter.goToName( camera_name ) )
		return;

	SGCameraNode * node = dynamic_cast<SGCameraNode*>(nameIter.getNode());
	if( node != NULL )
	{
		node->getCameraController()->setTime( getScene()->getRootState()->dynamicState->time );
		if( node->getCamera() )
		{
			getScene()->getRootState()->staticState->camera = *(node->getCamera());
		}

		ViewFrustum * v = node->getViewFrustum();
		if( v )
		{
			getScene()->getRootState()->staticState->viewFrustum.setAll(
				v->getZNearClip(), v->getZFarClip(), v->getVerticalFieldOfView(),
				(float)m_CurrentWidth/(float)m_CurrentHeight );
			v->setXToYAspectRatio((float)m_CurrentWidth/(float)m_CurrentHeight);
		}
		setGLCameraFromState();
	}
	else
		debugOutput << "Camera " << camera_name << " wurde nicht gefunden\n!";
}

void MultiPassRenderer::setCamera( Camera cam, ViewFrustum vf) 
{
	getScene()->getRootState()->staticState->camera = cam;
	getScene()->getRootState()->staticState->viewFrustum = vf;
	setGLCameraFromState();
}

void MultiPassRenderer::setOrthoCamera( Camera camera, ViewFrustum viewFrustum) 
{
	getScene()->getRootState()->staticState->camera = camera;
	getScene()->getRootState()->staticState->viewFrustum = viewFrustum;
	setGLOrthoCameraFromState();
}

void MultiPassRenderer::setOrthoCameraAsParameter( Camera camera, ViewFrustum viewFrustum, std::string parameter_name) 
{
	Scene * scene = getScene();
	BoundingBox3f sceneBB = getScenePCBBox(scene);
	Sphere3f bsphere = sceneBB.getBoundingSphere();

	Vector3f camPos = camera.getPosition();
	Vector3f lookAt = camera.getLookAt();
	Vector3f upVec = camera.getUp();

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	bsphere.radius *= 1.2f;
	glOrtho(-bsphere.radius,bsphere.radius,-bsphere.radius, bsphere.radius, 
		max(1e-10f,norm(camPos-bsphere.center)-bsphere.radius*1.1f),
		max(1e-10f,norm(camPos-bsphere.center)+bsphere.radius*1.1f));
	gluLookAt(	camPos[0], camPos[1], camPos[2],
				lookAt[0], lookAt[1], lookAt[2],

				upVec[0], upVec[1], upVec[2]);
	
	Matrix4f mvp;
	glGetFloatv( GL_PROJECTION_MATRIX, &mvp[0][0] );

	glPopMatrix();
	glMatrixMode( GL_MODELVIEW );

	setMatrix4f( parameter_name, mvp );
	Matrix4f invMVP = invertMatrix(mvp);
	setMatrix4f( parameter_name+"inv",invMVP);

	setMaterialParameter3f( parameter_name+"CamPos", camera.getPosition()[0],
				 camera.getPosition()[1],
				 camera.getPosition()[2] );

	float offset = (viewFrustum.getZFarClip() - viewFrustum.getZNearClip()) * 0.0001f;
	setMaterialParameter1f( parameter_name+"Offset", offset );
}


void MultiPassRenderer::setCameraAsParameter( Camera camera, ViewFrustum viewFrustum, std::string parameter_name)
{
		Matrix4f mvp;
		glMatrixMode( GL_PROJECTION );
		glPushMatrix();

		glLoadIdentity();
		gluPerspective(viewFrustum.getVerticalFieldOfView(),
							viewFrustum.getXToYAspectRatio(),
							viewFrustum.getZNearClip(),
							viewFrustum.getZFarClip());
		gluLookAt(camera.getPosition()[0],
					 camera.getPosition()[1],
					 camera.getPosition()[2],

					 camera.getLookAt()[0],
					 camera.getLookAt()[1],
					 camera.getLookAt()[2],

					 camera.getUp()[0],
					 camera.getUp()[1],
					 camera.getUp()[2]);
		glGetFloatv( GL_PROJECTION_MATRIX, &mvp[0][0] );

		glPopMatrix();
		glMatrixMode( GL_MODELVIEW );

		setMatrix4f( parameter_name, mvp );

		setMaterialParameter3f( parameter_name+"CamPos", camera.getPosition()[0],
					 camera.getPosition()[1],
					 camera.getPosition()[2] );

		float offset = (viewFrustum.getZFarClip() - viewFrustum.getZNearClip()) * 0.0001f;
		setMaterialParameter1f( parameter_name+"Offset", offset );
}

void MultiPassRenderer::setCameraAsParameter( std::string camera_name, std::string parameter_name) 
{
   StandardSceneGraphIterator sceneIt;
   sceneIt.clearAndSetup(getScene()->getRootNode(), getScene()->getRootState(), SGDynamicState::tmRendering, SceneGraphNode::FLAG_VISIBLE, 0);
	SGIndexedAndByNameSceneGraphIteratorAdaptor nameIter;
	nameIter.clearAndSetup( &sceneIt );
	if( !nameIter.goToName( camera_name ) )
		return;

	SGCameraNode * node = dynamic_cast<SGCameraNode*>(nameIter.getNode());
	if( node != NULL )
	{
		//Matrix4f mvp = node->getViewFrustum()->getProjectionMatrix() * node->getCamera()->getViewMatrix();
		ViewFrustum viewFrustum = *node->getViewFrustum();
		Camera camera = *node->getCamera();

		setCameraAsParameter( camera, viewFrustum, parameter_name);
	}
}

void MultiPassRenderer::setDefaultCamera() 
{
	getScene()->getRootState()->staticState->camera = m_DefaultCamera;

	ViewFrustum * v = &m_DefaultViewFrustum;
	getScene()->getRootState()->staticState->viewFrustum.setAll(
			v->getZNearClip(), v->getZFarClip(), v->getVerticalFieldOfView(),
			(float)m_CurrentWidth/(float)m_CurrentHeight );

//	setGLCameraFromState();

	Scene * scene = getScene();
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMultMatrixf(m_DefaultProjectionMatrix.data());
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glMultMatrixf(m_DefaultModelviewMatrix.data());
	/*gluLookAt(scene->getRootState()->staticState->camera.getPosition()[0],
		scene->getRootState()->staticState->camera.getPosition()[1],
		scene->getRootState()->staticState->camera.getPosition()[2],

		scene->getRootState()->staticState->camera.getLookAt()[0],
		scene->getRootState()->staticState->camera.getLookAt()[1],
		scene->getRootState()->staticState->camera.getLookAt()[2],

		scene->getRootState()->staticState->camera.getUp()[0],
		scene->getRootState()->staticState->camera.getUp()[1],
		scene->getRootState()->staticState->camera.getUp()[2]);*/

	

}

void MultiPassRenderer::setDefaultCameraAsParameter( std::string parameter_name) 
{
	Matrix4f mvp = m_DefaultViewFrustum.getProjectionMatrix() * 
		m_DefaultCamera.getViewMatrix();
	setMatrix4f( parameter_name, mvp );
}

/*void MultiPassRenderer::saveRenderTarget( std::string target_name, std::string filename, EFileFormat, int quality ) 
{
	if( !renderTargetExists( target_name ) )
	{
		debugOutput << "MultiPassRenderer::saveRenderTarget() - rendertarget '" << target_name << "' doesn't exist!\n";
		return;
	}

	m_RenderTargets[target_name].rt->
}*/

void MultiPassRenderer::setGLCameraFromState() 
{
	Scene * scene = getScene();

   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   gluPerspective(scene->getRootState()->staticState->viewFrustum.getVerticalFieldOfView(),
                  (float)m_CurrentWidth/ (float)m_CurrentHeight,
                  scene->getRootState()->staticState->viewFrustum.getZNearClip(),
                  scene->getRootState()->staticState->viewFrustum.getZFarClip());
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
   gluLookAt(scene->getRootState()->staticState->camera.getPosition()[0],
             scene->getRootState()->staticState->camera.getPosition()[1],
             scene->getRootState()->staticState->camera.getPosition()[2],

             scene->getRootState()->staticState->camera.getLookAt()[0],
             scene->getRootState()->staticState->camera.getLookAt()[1],
             scene->getRootState()->staticState->camera.getLookAt()[2],

             scene->getRootState()->staticState->camera.getUp()[0],
             scene->getRootState()->staticState->camera.getUp()[1],
             scene->getRootState()->staticState->camera.getUp()[2]);
}

void MultiPassRenderer::setGLOrthoCameraFromState() 
{
	Scene * scene = getScene();
	BoundingBox3f sceneBB = getScenePCBBox(scene);
	Sphere3f bsphere = sceneBB.getBoundingSphere();

	Vector3f camPos = scene->getRootState()->staticState->camera.getPosition();
	Vector3f lookAt = scene->getRootState()->staticState->camera.getLookAt();
	Vector3f upVec = scene->getRootState()->staticState->camera.getUp();

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	bsphere.radius *= 1.2f;
	glOrtho(-bsphere.radius,bsphere.radius,-bsphere.radius, bsphere.radius, 
		max(1e-10f,norm(camPos-bsphere.center)-bsphere.radius*1.1f),
		max(1e-10f,norm(camPos-bsphere.center)+bsphere.radius*1.1f));
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	gluLookAt(	camPos[0], camPos[1], camPos[2],
				lookAt[0], lookAt[1], lookAt[2],
				upVec[0], upVec[1], upVec[2]);
}

void MultiPassRenderer::drawTargetToScreen( std::string target_name ) 
{
	if( m_RenderTargets[target_name].rt->isActive() )
		m_RenderTargets[target_name].rt->deactivate();
	X4_GL_FB_DBG
	if( m_RenderTargets[target_name].rt->isBound() )
		m_RenderTargets[target_name].rt->unbind();
	X4_GL_FB_DBG

	m_RenderTargets[target_name].rt->bind();
	X4_GL_FB_DBG
   int vp[4];
   glGetIntegerv( GL_VIEWPORT, vp );
   card32 width = vp[2]-vp[0];
   card32 height = vp[3]-vp[1];

   glPushAttrib( GL_ALL_ATTRIB_BITS );
   GLHelper::begin2D(width,height);

   glDisable( GL_DEPTH_TEST );
	m_Mat_DrawRTToScreen->bindShaders( getGLContext() );
	int tex_width = m_RenderTargets[target_name].rt->getWidth();
	int tex_height = m_RenderTargets[target_name].rt->getHeight();

   glBegin( GL_QUADS );
   glColor3f(1,1,1);
   glTexCoord2f( 0,0 );
   glVertex2f(0,0);
   glTexCoord2f( tex_width,0 );
   glVertex2f(width,0);
   glTexCoord2f( tex_width,tex_height );
   glVertex2f(width,height);
   glTexCoord2f( 0,tex_height );
   glVertex2f(0,height);
   glEnd();

	m_RenderTargets[target_name].rt->unbind();
	X4_GL_FB_DBG
   GLHelper::end2D();
	m_Mat_DrawRTToScreen->unbindShaders(getGLContext());
   glPopAttrib();

}

GLRenderTextureFBO *MultiPassRenderer::getRenderTargetFBO( std::string target_name ) 
{
	if( m_RenderTargets.find( target_name ) != m_RenderTargets.end() )
	{
		return m_RenderTargets[target_name].rt;
	}
	else
		return NULL;
}

void X4::MultiPassRenderer::renderSceneRecursive( StandardSceneGraphIterator * sceneIter, const Matrix4f &modelMatrix, bool ignoreVisibilityFlag, bool useObjMaterial, string pathWildcards )
{
	if (!sceneIter->getNode()->visible() && !ignoreVisibilityFlag) return;
	bool performRendering = true;
	if (pathWildcards != "") {
		string path = sceneIter->getNameWithPath();
		QString pathq = path.c_str();
		QRegExp rx(pathWildcards.c_str());
		rx.setPatternSyntax(QRegExp::Wildcard);
		performRendering  = rx.exactMatch(pathq);
	}

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glMultMatrixf(sceneIter->getState()->dynamicState->ctm.data());
	Matrix4f currentModelMatrix = modelMatrix;
	SGTransformationNode * tnode = dynamic_cast<SGTransformationNode*>(sceneIter->getNode());
	if( tnode )
	{
		currentModelMatrix = currentModelMatrix * tnode->getTransformation();
	}

	setMatrix4f( "ModelMatrix", currentModelMatrix.data() );

	SGObjectNode *objNode = dynamic_cast<SGObjectNode*>(sceneIter->getNode());
	if (objNode != NULL) {
		SceneObject* obj = objNode->getSceneObject();

		if (obj != NULL && (objNode->visible() || ignoreVisibilityFlag)  && performRendering) {

			if (useObjMaterial) {
				mpcard matIndex = obj->getMaterialIndex();
				if (matIndex < getScene()->getNumGLMaterials()) {
					setMaterial( getScene()->getGLMaterial(matIndex) );
				}
			}

			SceneGraphState *state = sceneIter->getState();
			SceneObjectIterator *it = obj->createIterator(SOIT(SOIT::CAP_HIERARCHICAL | SOIT::CAP_ATTACH_SECONDARY));
			HierarchicalIterator *hit = dynamic_cast<HierarchicalIterator*>(it);
			Matrix4f tmpM = currentModelMatrix * obj->getTransformation();
			setMatrix4f( "ModelMatrix", tmpM.data() );

			if (hit != NULL) {
				try {
					m_Traversal->performTraversal(it, sceneIter->getState(), this, m_CurrentWidth, m_CurrentHeight,
						getScene(), obj, getGLContext(), m_MemoryManager);
				} catch(...) {
					glMatrixMode(GL_MODELVIEW);
					glPopMatrix();
					delete it;
					throw;
				}
			}
			delete it;
		}
	}

	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	unsigned numChildren = sceneIter->getNumChildren();
	for( unsigned i=0;i<numChildren;i++ )
	{
		sceneIter->goToChild(i);
		renderSceneRecursive(sceneIter,currentModelMatrix, ignoreVisibilityFlag, useObjMaterial, pathWildcards);
		sceneIter->goToParent();
	}

}
