#ifndef MultiPassRendererH
#define MultiPassRendererH
//---------------------------------------------------------------------------
#include "x3foundation_private.h"
//---------------------------------------------------------------------------
#include "Renderer.h"
#include "LCTraversal.h"
//---------------------------------------------------------------------------

namespace X4
{

class GLRenderTextureFBO;
class GLMaterial;
class LCMultiPassRender_DrawNode;
class RGBAImage;
class StandardSceneGraphIterator;

/**
	provides methods for multi-pass-rendering including managed rendertargets.
	offers some extra functions for cameras,materials,...
*/
class X3FOUNDATION_PRIVATE_API MultiPassRenderer : public Renderer
{
	X4_ABSTRACT_CLASS( MultiPassRenderer )

public:
	MultiPassRenderer(void);
	~MultiPassRenderer(void);

   virtual void assign(const Object* obj, X4CopyContext *context = NULL);

public:

   virtual void connect(Scene* scene, GLContext* ctxt);

	/// do not overwrite this method (unless you are very sure of what you doing)
	virtual void render(card32 viewportWidth, card32 viewportHeight);

   /// frees rendertargets etc.
	virtual void disconnect(GLContext* ctxt);

	struct TDrawNodeState
	{
		GLMaterial	*	material;	// when != NULL all objects will be drawn with this material
		card32			matIndex;
	};

	TDrawNodeState getDrawNodeState() {return m_DrawNodeState;}

protected:

	/// render
	virtual void renderMultiPass() = 0;

	enum ERenderTargetFlags
	{
		RT_RGBA		=   0x01,
		RT_Depth    =   0x02,
		RT_RGBA16	=   0x04,
		RT_RGBA32	=   0x08,
		RT_MIPMAP   =   0x10,
		RT_2D       =   0x20
	};

	/// creates and activates a rendertarget(always render texture) for rendering
	/// possible flags are: RT_RGBA, RT_DEPTH, RT_RGBA16, RT_RGBA32
	/// to create a shadowmap use RT_DEPTH only and make sure that width&height are power of two
	void setRenderTarget( std::string target_name, card32 flags, card32 width, card32 height );

	/// switch to main rendertarget (screen,gl-window,...)
	void setDefaultRenderTarget();

	/// returns existing render target or null
	GLRenderTextureFBO * getRenderTargetFBO( std::string target_name );

	/// sets a global material, every sceneobject will be drawn with this material
	void setMaterial( GLMaterial * mat );
	void setMaterialFromScene( card32 index );

	/// sets default material (default per-pointcloud material)
	void resetMaterial();
	
	/// renders scene with default renderer
	/// mw: extended method to permit a targeted traversal with wildcards, ignoring visibility annotations
	void renderScene(bool ignoreVisibilityFlag=false, bool useObjMaterial=false, string pathWildcards="");

	/// binds a rendertarget as texture (creates an attachment)
	void bindTarget( std::string target_name, std::string texture_name );

	/// binds a rendertarget to a specific texturestage (direct binding)
	void bindTarget( std::string target_name, card32 texture_stage );

	/// binds an image as texture (uploads every frame)
	void bindTexture( std::string texture_name, RGBAImage * image  );

	/// draws a rendertarget over the complete screen
	void drawTargetToScreen( std::string target_name );

	/// sets material parameter
	void setMaterialParameter1f( std::string param_name, float f );
	void setMaterialParameter2f( std::string param_name, float f1, float f2 );
	void setMaterialParameter3f( std::string param_name, float f1, float f2, float f3 );
	void setMaterialParameter4f( std::string param_name, float f1, float f2, float f3, float f4 );

	/// sets material parameter integer
	void setMaterialParameter1i( std::string param_name, int32 f );

	/// sets material parameter (4 floats)
	void setMaterialParameter4fv( std::string param_name, const Vector4f &vec );
	void setMaterialParameter4fv( std::string param_name, float * f );

	/// sets material matrix parameter (4x4 matrix)
	void setMatrix4f( std::string param_name, const Matrix4f &mat );
	void setMatrix4f( std::string param_name, float * mat );

	/// sets camera
	void setCamera( std::string camera_name );
	void setCamera( Camera camera, ViewFrustum viewFrustum);

	/// sets camera-matrix (mvp) as material parameter
	/// sets camera position as 'parameter_name+"CamPos"'
	void setCameraAsParameter( std::string camera_name, std::string parameter_name);
	void setCameraAsParameter( Camera cam, ViewFrustum vf, std::string parameter_name);

	/// sets std::map of gl state as ortho camera
	void setOrthoCamera( Camera camera, ViewFrustum viewFrustum);

	/// sets ortho camera as material parameter
	void setOrthoCameraAsParameter( Camera camera, ViewFrustum viewFrustum, std::string parameter_name);


	/// sets current default camera
	void setDefaultCamera();
	Camera& getDefaultCamera() { return m_DefaultCamera; }

	/// sets default-camera-matrix(mvp) as material parameter
	void setDefaultCameraAsParameter( std::string parameter_name);

	enum EFileFormat
	{
		BMP = 0,
		PNG = 1,
		JPEG = 2
	};
	
	/// saves rendertarget to image, quality is only used in jpeg-compression
	//void saveRenderTarget( std::string target_name, std::string filename, EFileFormat, int quality=80 );

	card32 getScreenWidth() const {return m_ScreenWidth;}
	card32 getScreenHeight() const {return m_ScreenHeight;}

private:

	void renderSceneRecursive( StandardSceneGraphIterator * sceneIter, const Matrix4f &modelMatrix, bool ignoreVisibilityFlag, bool useObjMaterial, string pathWildcards );
	void setMaterialParameterxfv( std::string param_name, card32 numComponents, float * f );
	void setMaterialParameterxiv( std::string param_name, card32 numComponents, int32 * f );
	void unbindAllRenderTargets();
	void eraseRenderTargets();
	bool renderTargetExists( std::string target_name );
	void setGLCameraFromState();
	void setGLOrthoCameraFromState();

	struct TRenderTarget
	{
		GLRenderTextureFBO	* rt;
		std::string	target_name;
		std::string	texture_name;
	};

	std::map<std::string, TRenderTarget>	m_RenderTargets;
	std::string							m_CurrentRenderTarget;
	float								m_CurrentTime;

	card32							m_ScreenWidth, m_ScreenHeight;
	card32							m_CurrentWidth, m_CurrentHeight;
	bool								m_ScreenSizeChanged;

	Camera							m_DefaultCamera;
	ViewFrustum						m_DefaultViewFrustum;
	Matrix4f						m_DefaultProjectionMatrix;
	Matrix4f						m_DefaultModelviewMatrix;

	LCTraversal			*			m_Traversal;

	MemoryManager		*			m_MemoryManager;

	TDrawNodeState					m_DrawNodeState;

	GLShaderMaterial	*			m_Mat_DrawRTToScreen;
};

class X3FOUNDATION_PRIVATE_API LCMultiPassRender_DrawNode : public LCNodeProcessor
{
	X4_CLASS( LCMultiPassRender_DrawNode )
private:
	
public:

	virtual void processNode(LCState *state, SceneObjectIterator *node);

};

}

#endif