#include "StdAfx.h"
#include <windows.h>
#include <mmsystem.h>
//---------------------------------------------------------------------------
#include "MultiScaleFeatureDetector.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "CachedRangeQuerry.h"
#include "VertexArray.h"
#include "PCA.inline.h"
#include "UnstructuredInCorePointCloud.h"
#include "PointCloudTools.h"
#include "GeometricTools.h"
#include "GeometricFeature.h"
#include "DebugRenderer.h"
#include "Timer.h"
#include "StringHelper.h"
#include "LDHistogram.h"
#include "Monomials2ndOrder.h"
#include "TopologyRangeSearch.h"
#include "TopologyAttachment.h"
#include "SeparatorClassProperty.h"
#include "random.h"
#include "PCCResampler.h"
#include "InCorePCTopologyGraph.h"
#include "PCCSurfaceTools.h"
#include "LDHistogram.h"
#include "SpinImageDescriptor.h"
#include "IntrinsicWaveDescriptor.h"
#include "ProgressWindow.h"

//---------------------------------------------------------------------------
using namespace X4;


std::string descSmoothingMethod (card32 i)
{
	switch (i)
	{
	case 0:
		return "Plane";
	case 1:
		return "QuadPatch";
	case 2:
		return "MLS";
	}
	return "unkown";
}

IMPLEMENT_X4_CLASS( MultiScaleFeatureDetector ,1)
{
	BEGIN_CLASS_INIT( MultiScaleFeatureDetector );

	ADD_CARD32_PROP(m_NumScales,0);
	ADD_FLOAT32_PROP(m_H0,0);
	ADD_FLOAT32_PROP(m_F,0);
	ADD_BOOLEAN_PROP(m_UseNormalOrientation,0);
	ADD_CARD32_PROP_UPDATE_DESCR(m_SmoothingMethod,0,NULL,descSmoothingMethod,3, NULL);
	ADD_SEPARATOR("Topological version");
	ADD_BOOLEAN_PROP(m_UseTopology,1);
	ADD_FLOAT32_PROP(m_TargetMedPointDistance,1);
	ADD_SEPARATOR("Debug");
	ADD_BOOLEAN_PROP(m_DebugOutputScales,0);
}

IMPLEMENT_X4_CLASS( ScaleSpaceDescriptor ,0)
{
	BEGIN_CLASS_INIT( ScaleSpaceDescriptor );

	ADD_FLOAT32_PROP( m_MinScaleSpaceDifference, 0);

}

IMPLEMENT_X4_CLASS( ScaleSpaceSignature ,0)
{
	BEGIN_CLASS_INIT( ScaleSpaceSignature );

	ADD_FLOAT32_PROP( m_ScaleSpaceDifference,0);
	ADD_CARD32_PROP( m_ScaleIndex,0);
	ADD_BOOLEAN_PROP( m_IsMaxima, 0);


}

MultiScaleFeatureDetector::MultiScaleFeatureDetector(void)
{
	m_NumScales = 4;
	m_H0 = 2.0f;
	m_F = pow(2.0f,0.25f);
	m_UseNormalOrientation = true;
	m_SmoothingMethod = 0;
	m_DebugOutputScales = false;
	m_UseTopology = false;
	m_TargetMedPointDistance = 0.01f;

	m_Descriptors.push_back( new ScaleSpaceDescriptor);
}

MultiScaleFeatureDetector::~MultiScaleFeatureDetector(void)
{
}

void MultiScaleFeatureDetector::assign(const Object* obj, X4CopyContext *context)
{
	FeaturePointDetector::assign(obj,context);
	const MultiScaleFeatureDetector * o = dynamic_cast<const MultiScaleFeatureDetector*>(obj);
	x4Assert( o != NULL );

	copyObjectProperties(obj,this);
}

FeatureSet * X4::MultiScaleFeatureDetector::detect( UnstructuredInCorePointCloud * upc, std::string attachmentName)//=""  )
{
	Timer timer;
	m_ScaleDifferences.resize(upc->getNumPoints()*(m_NumScales-1));

	// create/get feature set attachment
	if( attachmentName.empty() )
		attachmentName = FeatureSet::getDefaultName();

	FeatureSet * fs = dynamic_cast<FeatureSet*>(upc->getAttachments()->getData(attachmentName));
	delete fs;
	fs = new FeatureSet;
	fs->setup( attachmentName, AttachedData::ADF_PERSISTENT );
	upc->getAttachments()->attachData(fs);

	UnstructuredInCoreTriangleMesh * mesh = dynamic_cast<UnstructuredInCoreTriangleMesh*>(upc);
	if( m_UseTopology && mesh)
	{
		UnstructuredInCorePointCloud * resampledMesh = NULL;
		if (m_TargetMedPointDistance > 0) resampledMesh = createTopologyUPCFromTriangleMesh(mesh, m_TargetMedPointDistance);
		else
		{
			UnstructuredInCoreTriangleMesh* pcCopy = new UnstructuredInCoreTriangleMesh;
			pcCopy->setPointSet((PointSet*)mesh->getPointSet()->copy());
			pcCopy->setTriangles((PointSet*)mesh->getTriangles()->copy());
			resampledMesh = pcCopy;

			if (mesh->getAttachments()->getData(TopologyAttachment::getDefaultName()) == NULL)
			{
				pcCopy->computeTopologyAttachment();
			}else
			{
				TopologyAttachment* copyt = new TopologyAttachment;
				copyt->setup(TopologyAttachment::getDefaultName(), AttachedData::ADF_PERSISTENT);
				copyt->assign(mesh->getAttachments()->getData(TopologyAttachment::getDefaultName()));
				resampledMesh->getAttachments()->attachData(copyt);
			}
		}

		computeScaleSpaceDifferences(resampledMesh);
		extractFeatures(resampledMesh,fs);

		/*PCCSurfaceTools st;
		st.setKernelSigma((m_TargetMedPointDistance < 0 ? m_MedianPointDistance : m_TargetMedPointDistance)*2.5f);
		st.setUseTopogicalRangeQuerry(true);
		st.estimateCurvature(resampledMesh);
		*/

		// compute signatures
		recomputeFeatureSignatures(resampledMesh,fs );

		delete resampledMesh;
	}
	else
	{
		computeScaleSpaceDifferences(upc);

		// search for extreme points in scale space differences
		extractFeatures(upc,fs);

		// compute signatures
		recomputeFeatureSignatures(upc,attachmentName );
	}


	// clean up
	m_ScaleDifferences.clear();


	debugOutput << "Feature Detection Time: " << convertTimeToString(timer.getDeltaValue()) << "\n";

//	fs->computeGeodeticDistances(upc);

	return fs;
}

void X4::MultiScaleFeatureDetector::beginDetection(const std::vector<UnstructuredInCorePointCloud*>& clouds )
{
	if( m_UseTopology && m_TargetMedPointDistance > 0)
	{
		m_MedianPointDistance = m_TargetMedPointDistance;
		return;
	}

	m_MedianPointDistance = 0;
	for( unsigned i=0;i<clouds.size();i++ )
		m_MedianPointDistance += getMedianPointDistance(clouds[i]) / (float)clouds.size();
}

void X4::MultiScaleFeatureDetector::computeScaleSpaceDifferences( UnstructuredInCorePointCloud * upc )
{
	std::vector<float> h(m_NumScales),radius(m_NumScales),radius_2(m_NumScales),weightMultipliers(m_NumScales);
	for( unsigned i=0;i<h.size();i++ )
	{
		h[i] = m_H0 * m_MedianPointDistance * pow(m_F,(float)i);
		weightMultipliers[i] = 1.0f / h[i]/ h[i];
		radius[i] = h[i] * 2.0f;
		radius_2[i] = pow(radius[i],2.0f);
	}
	float maxRad = radius.back();
	CachedRangeQuerry querry(upc);
	TopNeighborhoodSphereSearch * topQuery = NULL;
	InCorePCTopologyGraph * tg = NULL;
	if( m_UseTopology )
	{
		TopologyAttachment * topAtt = dynamic_cast<TopologyAttachment*>(upc->getAttachments()->getData(TopologyAttachment::getDefaultName()));
		if( !topAtt)
			throw X4Exception("MultiScaleFeatureDetector::computeScaleSpaceDifferences() - Could not find topology attachments!");
		topQuery = new TopNeighborhoodSphereSearch;
		tg = topAtt->getTopology();
	}

	VertexArray points(upc->getPointSet());
	const card32 numPoints = upc->getNumPoints();
	const card32 numScales = m_NumScales;
	std::vector<float> distances_2;
	std::vector<Vector3f> currentProjectedPoints(m_NumScales);
	std::vector<Vector3f> currentNormals(m_NumScales);
	bool hasNormals = upc->providesAttribute("normal");
	BoundingBox3f bb = getPCBBox(upc);
	debugRenderer->clearDebugData("computeScales");
	debugRenderer->beginRenderJob("computeScales");
	AAT posAAT = upc->getAAT("position");
	std::vector<Vector3f> topPositions; topPositions.reserve(1000);
	for( unsigned i=0;i<numPoints;i++ )	
	{
		Vector3f pos = points.getPosition3f(i);
		Vector3f normal;
		if( hasNormals )
			normal = points.getNormal3f(i);
		std::vector<Vector3f> * positions = NULL;
		if(!m_UseTopology)
			positions = querry.getSpherePositions(pos, maxRad, 2.7f * m_MedianPointDistance);
		else
		{
			topQuery->setup(tg,upc->getPointSet(),i,pos,maxRad, posAAT);
			topPositions.clear();
			while(!topQuery->atEnd())
			{
				unsigned index = topQuery->getCurrentIndex();
				topPositions.push_back(upc->getPointSet()->get3f(index,posAAT));
				topQuery->next();
			}
			positions = &topPositions;
		}
		distances_2.resize(positions->size());
		for( unsigned j=0;j<positions->size();j++ )
			distances_2[j] = normQuad(pos - (*positions)[j] );

		// estimate local frame via pca and project point onto plane
		for(unsigned scale=0;scale<numScales;scale++)
		{
			PCA<float,3> pca;
			std::vector<Vector3f> localPositions;
			std::vector<float> weights;
			for( unsigned j=0;j<positions->size();j++ )
			{
				if( distances_2[j] > radius_2[scale] )
					continue;

				float w = exp(-distances_2[j] * weightMultipliers[scale]);
				pca.addPoint( (*positions)[j], w );
				
				// store positions and weights for later use
				if(m_SmoothingMethod == SM_QuadPatch )
				{
					localPositions.push_back((*positions)[j]);
					weights.push_back(w);
				}
			}

			Matrix3f eigenVectors;
			Vector3f eigenValues, centroid;
			pca.analyze(eigenValues, eigenVectors, centroid);

			if( m_UseNormalOrientation )
			{
				if( eigenVectors[2]*normal < 0 )
					eigenVectors[2] *= -1;
			}

			// project point & store normal
			Vector3f plane_normal;
			if( m_SmoothingMethod == SM_Plane )
			{
				plane_normal= eigenVectors[2];
				currentProjectedPoints[scale] = pos - plane_normal * (plane_normal * (pos - centroid));
				currentNormals[scale] = plane_normal;
			}
			else
			{
				// transform into local frame
				Matrix3f frame = eigenVectors.transpose();
				for( unsigned j=0;j<localPositions.size();j++ )
				{
					localPositions[j] = frame * (localPositions[j] - centroid) / radius[scale];
				}

				// compute patch
				Vector6f coeffs;
				Monomials2ndOrder::computeCoeffs( &localPositions[0], &weights[0], (card32)localPositions.size(), coeffs );

				// project and transform back into reference frame
				Vector3f transf_pos = makeVector3f(0,0,coeffs[0]);
				transf_pos = eigenVectors * (transf_pos*radius[scale]) + centroid;
				currentProjectedPoints[scale] = transf_pos;
				currentNormals[scale] = Monomials2ndOrder::evaluateNormal( coeffs, makeVector2f(0,0));
				currentNormals[scale] = eigenVectors * currentNormals[scale];

			}
			if( m_DebugOutputScales )
			{
				Vector3f scale_pos = currentProjectedPoints[scale] + makeVector3f(1,0,0)*bb.getSideLength(0)*(float)(scale+1);
				//debugRenderer->addSurfel(scale_pos , currentNormals[scale], makeVector3f(0,0.3f,0.8f));
				debugRenderer->addPoint(scale_pos , (currentNormals[scale] + makeVector3f(1,1,1))*0.5);
			}

		}

		// compute scale difference
		for( unsigned scale=0;scale<numScales-1;scale++)
		{
			float diff = currentNormals[scale+1] * (currentProjectedPoints[scale]-currentProjectedPoints[scale+1]);
			if( !m_UseNormalOrientation )
				diff = fabs(diff);
			m_ScaleDifferences[i*(numScales-1)+scale] = diff;
		}
	}
	debugRenderer->endRenderJob();
	delete topQuery;
}

void X4::MultiScaleFeatureDetector::extractFeatures( UnstructuredInCorePointCloud * upc, FeatureSet * fs )
{
	HierarchicalKNNIterator hIt(upc,32,NULL);
	VertexArray points(upc->getPointSet());
	AAT posAAT = upc->getAAT("position");

	std::vector<float> h(m_NumScales),radius(m_NumScales),radius_2(m_NumScales),weightMultipliers(m_NumScales);
	for( unsigned i=0;i<h.size();i++ )
	{
		h[i] = m_H0 * m_MedianPointDistance * pow(m_F,(float)i);
		weightMultipliers[i] = 1.0f / h[i]/ h[i];
		radius[i] = h[i] * 2.0f;
		radius_2[i] = pow(radius[i],2.0f);
	}

	progressWindow->pushStep(true, "Multiscale Features");
	for( unsigned i=0;i<points.getNumElements();i++ )
	{
		progressWindow->progress(100.0f * float(i) / float(points.getNumElements()));
		Vector3f pos = points.getPosition3f(i);
		Vector3f normal;
		if( upc->providesAttribute("normal"))
			normal = points.getNormal3f(i);
		for( unsigned scale=1;scale<m_NumScales-1;scale++ )
		{
			hIt.setSeekPointAndReset(pos);

			float value = m_ScaleDifferences[i*(m_NumScales-1)+scale];
			bool minPoint = value < 0;
			bool maxPoint = value > 0;

			while( !hIt.atEnd() && (minPoint || maxPoint))
			{
				Vector3f tmpPos = hIt.get3f(posAAT);
				
				if( normQuad(tmpPos - pos) > radius_2[scale+1]*0.5f)
					break;

				card32 pindex = hIt.getCurrentPointIndex();
				minPoint = minPoint && (value <= m_ScaleDifferences[pindex*(m_NumScales-1)+scale]);
				maxPoint = maxPoint && (value >= m_ScaleDifferences[pindex*(m_NumScales-1)+scale]);
				
				if( scale > 0 )
				{
					minPoint = minPoint && (value < m_ScaleDifferences[pindex*(m_NumScales-1)+scale-1]);
					maxPoint = maxPoint && (value > m_ScaleDifferences[pindex*(m_NumScales-1)+scale-1]);
				}

				if( scale < m_NumScales-2 )
				{
					minPoint = minPoint && (value < m_ScaleDifferences[pindex*(m_NumScales-1)+scale+1]);
					maxPoint = maxPoint && (value > m_ScaleDifferences[pindex*(m_NumScales-1)+scale+1]);
				}

				hIt.next();
			}

			if( minPoint || maxPoint )
			{
				Vector3f projPos;
				Vector3f projNormal;
				if( m_SmoothingMethod == SM_Plane)
				{
					Matrix3f frame;
					Vector3f center;
					computeLocalFrame( pos, h[scale+1], upc, frame,center);
					if( m_UseNormalOrientation )
						if( normal * frame[2] < 0)
						frame[2] *= -1;

					projPos = pos - frame[2]*(frame[2]*(pos - center));

					// refine feature point
					Vector3f fpoint = projPos;
					Vector3f fnormal = frame[2];
					float bestvalue = m_ScaleDifferences[i*(m_NumScales-1)+scale];
					int numSamples = 5;
					for( int x=-numSamples;x<=numSamples;x++ )
					{
						for( int y=-numSamples;y<=numSamples;y++ )
						{
							Vector3f testpos = pos + frame[0]*(m_MedianPointDistance*(float)x/(float)numSamples)
								+ frame[1]*(m_MedianPointDistance*(float)y/(float)numSamples);
													
							Matrix3f frame_current;
							Vector3f center_current;
							computeLocalFrame( testpos, h[scale], upc, frame_current,center_current);
							Vector3f projPos1 = testpos - frame_current[2]*(frame_current[2]*(testpos - center_current));
							computeLocalFrame( testpos, h[scale+1], upc, frame_current,center_current);
							Vector3f projPos2 = testpos - frame_current[2]*(frame_current[2]*(testpos - center_current));
							if( m_UseNormalOrientation )
							{
								if( frame_current[2]*normal < 0)
									frame_current[2] *= -1;
							}

							float diff = frame_current[2] * (projPos1 - projPos2);
							if( !m_UseNormalOrientation )
								diff = fabs(diff);

							if( (minPoint && diff < bestvalue) || (maxPoint && diff > bestvalue))
							{
								bestvalue = diff;
								fpoint = projPos2;
								fnormal = frame_current[2];
							}

						}

					}
					projNormal = frame[2];

				}
				else
				{
					// project onto smoothed surface 
					projectPointOntoPatch(pos,h[scale+1], upc, projPos, projNormal);
					if( normal * projNormal < 0 )
						projNormal *= -1;
				}


				GeometricFeature * f = new GeometricFeature;
				f->m_Position = m_UseTopology?pos:projPos;
				f->m_Normal = projNormal;

				// check if scale space descriptor is attached, then add such signature
				int found = -1;
				for (unsigned de=0; de < getNumDescriptors(); de++)
				{
					if (dynamic_cast<ScaleSpaceDescriptor*>(getDescriptor(de)))
					{
						found = de;
						break;
					}
				}

				if (found >= 0)
				{
					ScaleSpaceSignature * sig = new ScaleSpaceSignature;
					sig->m_IsMaxima = maxPoint || !m_UseNormalOrientation;
					sig->m_ScaleIndex = scale;
					sig->m_ScaleSpaceDifference = m_ScaleDifferences[i*(m_NumScales-1)+scale];
					f->m_Signatures.push_back(sig);
				}

				fs->m_Features.push_back(f);
			}
		}
	}
	progressWindow->popSteps();
}


void X4::MultiScaleFeatureDetector::setupSHREC()
{
	deleteDescriptors(0, getNumDescriptors());

	ScaleSpaceDescriptor* sd = new ScaleSpaceDescriptor;
	sd->m_MinScaleSpaceDifference = 0.0005f;
	addDescriptor(sd);

	m_TargetMedPointDistance = -1;
	m_UseTopology = true;
	m_H0 = 1.5f;
	m_NumScales = 5;
	//IntrinsicWaveDescriptor* iwd = new IntrinsicWaveDescriptor;
	//addDescriptor(iwd);
	//SpinImageDescriptor* spd = new SpinImageDescriptor;
	//addDescriptor(spd);
	//LDHistogram * ldh = new LDHistogram;
	//ldh->setUseTopology(true);
	//addDescriptor( ldh);
}

X4::ScaleSpaceDescriptor::ScaleSpaceDescriptor()
{
	m_MinScaleSpaceDifference = 0;
}

X4::ScaleSpaceSignature::ScaleSpaceSignature()
{
}

bool X4::ScaleSpaceSignature::badFeature( LocalDescriptor * myDescriptor ) const
{
	ScaleSpaceDescriptor * desc = dynamic_cast<ScaleSpaceDescriptor*>(myDescriptor);
	return fabs(m_ScaleSpaceDifference) < (desc ? desc->m_MinScaleSpaceDifference : 0);
}

float X4::ScaleSpaceSignature::compare( LocalSignature * other, LocalDescriptor *  )
{
	ScaleSpaceSignature * o = dynamic_cast<ScaleSpaceSignature*>(other);
	return (m_ScaleIndex == o->m_ScaleIndex) && (m_IsMaxima == o->m_IsMaxima);
}
