//---------------------------------------------------------------------------
#include "stdafx.h"
//---------------------------------------------------------------------------
#include "ClassMethods.h"
#include "ObjectClassProperty.h"
#include "SeparatorClassProperty.h"
#include "NDT.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------


std::string getGridSetupModeDescr( const card32 index )
{
	if ( index == CREATE_GRID_FROM_TAU ){
		return "From tau";
	} else if ( index == CREATE_GRID_FROM_RES ){
		return "From res";
	} else {
		return "undefined";
	}
}


std::string getGridTypeDescr( const card32 index )
{
	if ( index == NDT_GRID_REGULAR ){
		return "Regular";
	} else if ( index == NDT_GRID_REGULAR_OVERLAPPING ){
		return "Regular/Overlapping";
	} else {
		return "undefined";
	}
}

IMPLEMENT_X4_CLASS( NDT, 5 )
{
	BEGIN_CLASS_INIT( NDT )
	ADD_SEPARATOR( "NDT" )
	ADD_SEPARATOR( "Grid" )
	ADD_CARD32_PROP( numGrids, 2 )
	ADD_VAR_ARRAY_PROP( grids, 3, new ObjectClassProperty( "grid" , 0, NULL, AbstractGrid::getClass(), true, NULL ) )
	ADD_CARD32_PROP_UPDATE_DESCR( setupGridMode, 4, NULL, getGridSetupModeDescr, 2, NULL )
	ADD_CARD32_PROP_UPDATE_DESCR( gridType, 5, NULL, getGridTypeDescr, 3, NULL )
	ADD_CARD32_PROP( setupGridRes, 4 )
	ADD_FLOAT32_PROP( setupGridTau, 4 )
	ADD_OBJECT_PROP( gridBuilder, 5, GridBuilder::getClass(), true );
	ADD_NOARGS_METHOD( printScore )
}

NDT::NDT()
{
	subsampling = 5;
	stepSize = 0.5f;
	numGrids = 0;
	setupGridTau = 0.3f;
	setupGridRes = 10;
	grids.clear();
	T->setMode( TRANSFORMATION_EULER_ANGLES );
	solver = SOLVER_GRADIENT_DESCENT;
	colorCode = COLOR_CODE_NONE;
	candSelectionMode = CANDIDATE_SELECTION_ALL;
	setupGridMode = CREATE_GRID_FROM_RES;
	gridType = NDT_GRID_REGULAR;
	gridBuilder = new GridBuilder();
	ownGrids = false;
	registerTranslation = true;
	registerRotation = true;
}

NDT::~NDT()
{
	clear();
	delete gridBuilder;
}

void NDT::clear()
{
	if ( ownGrids ){
		for ( unsigned int n = 0; n < numGrids; n++ ){
			grids[n]->clear();
			delete grids[n];
		}
	}
	numGrids = 0;
	grids.clear();
	ownGrids = false;
}

void NDT::setup( PointSet* basePS, const Matrix3f& baseR, const Vector3f& baseT )
{
	Registerizer::setup( basePS, baseR, baseT );
	clear();
	createGrids( gridType );
	fillGrids();
	for ( unsigned int gridIndex = 0; gridIndex < numGrids; gridIndex++ ){
		grids[gridIndex]->printStats();
	}
}

void NDT::setupWithGrid( 
	PointSet* basePS, 
	AbstractGrid* grid, 
	const Matrix3f& baseR, 
	const Vector3f& baseT )
{
	Registerizer::setup( basePS, baseR, baseT );
	clear();
	numGrids = 1;
	grids.resize( 1 );
	grids[0] = grid;
}

void NDT::createGrids( const card32 type )
{
	// Compute bounding box
	AAT POSITION = basePS->getAAT( "position" );
	BoundingBox3f bbox;
	for ( unsigned int pointIndex = 0; pointIndex < basePS->getNumEntries(); pointIndex++ ){
		Vector3f pos = basePS->get3f( pointIndex, POSITION );
		if ( pointIndex == 0 ) bbox = BoundingBox3f( pos );
		else bbox.addPoint( pos );
	}
	Vector3f center = bbox.getCenter();
	float32 extend = bbox.getMaxSideLength() * 0.525;
	bbox.lowerCorner = center - makeVector3f( extend, extend, extend );
	bbox.upperCorner = center + makeVector3f( extend, extend, extend );
	Vector3f normalizationT = bbox.getCenter();
	float32 normalizationS = 2.0f / bbox.getMaxSideLength();
	bbox.lowerCorner = ( bbox.lowerCorner - normalizationT ) * normalizationS;
	bbox.upperCorner = ( bbox.upperCorner - normalizationT ) * normalizationS;

	clear();
	if ( type == NDT_GRID_REGULAR ){
		numGrids = 1;
		grids.resize( 1 );
		RegularGrid* grid = new RegularGrid();
		grid->setup( bbox, normalizationS, normalizationT, setupGridMode, setupGridTau, setupGridRes );
		grids[0] = grid;
	} else if ( type == NDT_GRID_REGULAR_OVERLAPPING ){
		numGrids = 8;
		grids.resize( numGrids );
		for ( unsigned int gridIndex = 0; gridIndex < numGrids; gridIndex++ ){
			RegularGrid* grid = new RegularGrid();
			grid->setup( bbox, normalizationS, normalizationT, setupGridMode, setupGridTau, setupGridRes );
			// Adjust bounding box
			float32 offset = grid->getTau()/4.0f;
			float32 dx = ( gridIndex%2 == 0 )? -offset : offset;
			float32 dy = ( (gridIndex/2)%2==0 )? -offset : offset;
			float32 dz = ( gridIndex<4 )? -offset : offset;	
			BoundingBox3f bbox = grid->getBBox();
			bbox.lowerCorner += makeVector3f( dx, dy, dz );
			bbox.upperCorner += makeVector3f( dx, dy, dz );
			grid->setBBox( bbox );
			grids[gridIndex] = grid;
		}
	} else {
		debugOutput << "setGrid: type " << type << " not implemented yet.\n"; debugOutput.flush();
		return;
	}
	ownGrids = true;
}

void NDT::fillGrids()
{
	// Compute covariance matrices
	for ( unsigned int gridIndex = 0; gridIndex < numGrids; gridIndex++ ){
		AbstractGrid* grid = grids[gridIndex];
		gridBuilder->fillGrid( grid, basePS );
	}
}

float32 NDT::computeGradient( StaticVector<float32,7>& gradient )
{
	for ( unsigned int n = 0; n < 7; n++ ) gradient[n] = 0; 
	
	if ( !ready ){
		debugOutput << "computeGradient: call setup first.\n"; debugOutput.flush();
		return -1;
	}
	
	NDTCell cell;
	StaticMatrix<float32,7,3> jacobian;
	card32 numContributers = 0;
	float32 score = 0;
	for ( unsigned int gridIndex = 0; gridIndex < numGrids; gridIndex++ ){
		AbstractGrid* grid = grids[gridIndex];
		Vector3f normalizationT = grid->getNormalizationT();
		float32 normalizationS = grid->getNormalizationS();
		for ( unsigned int pointIndex = 0; pointIndex < registerPS->getNumEntries(); pointIndex++ ){
			if ( colorCode == COLOR_CODE_USED_POINTS ) registerPS->set3f( pointIndex, REGISTER_COLOR, makeVector3f( 1,0,0 ) );
			if ( registerPS->get1i( pointIndex, REGISTER_FLAG ) & FLAG_REGISTER_CANDIDATE ){
				Vector3f r = registerR * registerPS->get3f( pointIndex, REGISTER_POSITION ) + registerT;
				r = ( r - normalizationT ) * normalizationS;
				if ( grid->contains( r ) ){
					NDTCell* cell = (NDTCell*) grid->getCell( r );
					if ( ( cell != NULL ) && ( cell->getValid() ) ){
						// Position
						Vector3f q = cell->getMean() - r;
						T->jacobian( jacobian, r );
						float32 localScore = exp( -q * ( cell->getInvSigma() * q ) / 2.0f );
						for ( unsigned int i = 0; i < 7; i++ ){
							gradient[i] += q * ( cell->getInvSigma() * jacobian[i] ) * localScore;
						}
						numContributers++;
						score += localScore;
						if ( colorCode == COLOR_CODE_USED_POINTS )
						{
							registerPS->set3f( pointIndex, REGISTER_COLOR, makeVector3f( 0.15f, 0.85f, 0.15f ) );
						}
					}
				}
			}
		}
	}
	if ( numContributers > 0 ){
		gradient /= float32( numContributers );
		score /= float32( numContributers );
	}

	debugOutput << "computeGradient: ";
	debugOutput << gradient;
	debugOutput << "\n"; 
	debugOutput.flush();

	return score;
}

float32 NDT::computeHessian( StaticMatrix<float32,7,7>& hessian, StaticVector<float32,7>& gradient )
{
	for ( unsigned int i = 0; i < 7; i++ ){
		gradient[i] = 0;
		for ( unsigned int j = 0; j < 7; j++ ){
			hessian[i][j] = 0;
		}
	}

	if ( !ready ){
		debugOutput << "computeGradient: call setup first.\n"; debugOutput.flush();
		return -1;
	}

	StaticMatrix<float32,7,3> jacobian;
	StaticMatrix<float32,7,7> H;
	StaticVector<float32,7> g;
	Vector3f r, q;
	card32 numContributers;
	numContributers = 0;
	NDTCell cell;
	float32 score = 0;
	lastMaxGrad = maxGrad;
	maxGrad = 0;

	for ( unsigned int gridIndex = 0; gridIndex < numGrids; gridIndex++ ){
		AbstractGrid* grid = grids[gridIndex];
		Vector3f normalizationT = grid->getNormalizationT();
		float32 normalizationS = grid->getNormalizationS();
		for ( unsigned int pointIndex = 0; pointIndex < registerPS->getNumEntries(); pointIndex++ ){
			if ( colorCode == COLOR_CODE_USED_POINTS ) registerPS->set3f( pointIndex, REGISTER_COLOR, makeVector3f( 1,0,0 ) );
			if ( registerPS->get1i( pointIndex, REGISTER_FLAG ) & FLAG_REGISTER_CANDIDATE ){
				r = registerR * registerPS->get3f( pointIndex, REGISTER_POSITION ) + registerT;
				r = ( r - normalizationT ) * normalizationS;
				if ( grid->contains( r ) ){
					NDTCell* cell = (NDTCell*) grid->getCell( r );
					if ( ( cell != NULL ) && ( cell->getValid() ) ){
						q = cell->getMean() - r;
						T->jacobian( jacobian, r );
						float32 localScore = exp( -q * ( cell->getInvSigma() * q ) / 2.0f );
						for ( unsigned int i = 0; i < 7; i++ ){
							g[i] = q * ( cell->getInvSigma() * jacobian[i] ) * localScore;
							for ( unsigned int j = 0; j < 7; j++ ){
								H[i][j] = ( -q * ( cell->getInvSigma() * jacobian[i] ) ) 
										* ( -q * ( cell->getInvSigma() * jacobian[j] ) )
										+ ( -jacobian[j] * ( cell->getInvSigma() * jacobian[i] ) );
							}
						}
						if ( T->getMode() != TRANSFORMATION_QUATERNIONS ) {
							H[6][6] = 1;
							g[6] = 0;
						}
						hessian += H * localScore;
						gradient += g;
						numContributers++;
						score += localScore;
						if ( colorCode == COLOR_CODE_GRADIENT_STRENGTH ) {
							float32 ng = norm( g );
							if ( ng > maxGrad ) maxGrad = ng;
							registerPS->set3f( pointIndex, REGISTER_COLOR, makeVector3f( 1, 1, 0 ) * ng / lastMaxGrad );
						} else if ( colorCode == COLOR_CODE_USED_POINTS ) {
							registerPS->set3f( pointIndex, REGISTER_COLOR, makeVector3f( 0.15f, 0.85f, 0.15f ) );
						}
					}	
				}
			}
		}
	}
	if ( numContributers > 5 ) {
		hessian /= float32( numContributers );
		gradient /= float32( numContributers );
		score /= float32( numContributers );
	} else {
		for ( unsigned int i = 0; i < 7; i++ ){
			for ( unsigned int j = 0; j < 7; j++ ){
				hessian[i][j] = ( i == j )? 1 : 0;
			}
			gradient[i] = 0;
		}
		debugOutput << "computeHessian: too few candidates found: " << numContributers << ".\n"; debugOutput.flush();
		return -1;
	}

	return score;
}

void NDT::printScore()
{
	debugOutput << "NDT score: " << computeScore() << ".\n"; debugOutput.flush();
}

void NDT::getDeriv( Vector3f& df_dp, const card32 component, const StaticMatrix<float32,7,3>& jacobian )
{
	for ( unsigned int n = 0; n < 3; n++ ){ 
		df_dp[n] = jacobian[component][n]; 
	}
}

void NDT::get2ndDeriv( Vector3f& d2f_dpdp, const card32 comp1, const card32 comp2, const StaticMatrix<float32,7,3>& jacobian )
{
	for ( unsigned int n = 0; n < 3; n++ ){ 
		d2f_dpdp[n] = 0; 
	}
}

float32 NDT::computeScore(Matrix3f rot, Vector3f t)
{
	Vector3f q;
	float32 score = 0;
	NDTCell cell;
	card32 numContributers = 0;
	for ( unsigned int gridIndex = 0; gridIndex < numGrids; gridIndex++ ){
		AbstractGrid* grid = grids[gridIndex];
		Vector3f normalizationT = grid->getNormalizationT();
		float32 normalizationS = grid->getNormalizationS();
		for ( unsigned int pointIndex = 0; pointIndex < registerPS->getNumEntries(); pointIndex++ ){
			Vector3f pos = registerPS->get3f( pointIndex, REGISTER_POSITION );
			pos = ( pos - normalizationT ) * normalizationS;
			NDTCell* cell = (NDTCell*) grids[gridIndex]->getCell( grids[gridIndex]->getIndex3D( pos ) );
			if ( cell != NULL ){
				q = cell->getMean() - pos;
				score += exp( -q * ( cell->getInvSigma() * q ) / 2.0f );
				numContributers++;
			}
		}	
	}
	if (numContributers>0) 
		score /= (float32) numContributers;
	return score;
}

bool NDT::contains( BoundingBox3f bbox, Vector3f p )
{
	for ( unsigned int d = 0; d < 3; d++ ){
		if ( ( p[d] <= bbox.lowerCorner[d] ) || ( p[d] >= bbox.upperCorner[d] ) ) return false;
	}
	return true;
}

void NDT::printStats() const
{
	if ( !ready || ( registerPS == NULL ) ){
		debugOutput << "printStats: ndt not ready.\n"; debugOutput.flush();
		return;
	}

	NDTCell cell;

	// Regular grid
	card32 numCoveredPoints = 0;
	for ( unsigned int gridIndex = 0; gridIndex < numGrids; gridIndex++ ){
		AbstractGrid* grid = grids[gridIndex];
		Vector3f normalizationT = grid->getNormalizationT();
		float32 normalizationS = grid->getNormalizationS();
		for ( unsigned int pointIndex = 0; pointIndex < registerPS->getNumEntries(); pointIndex++ ){
			Vector3f pos = registerPS->get3f( pointIndex, REGISTER_POSITION );
			pos = ( pos - normalizationT ) * normalizationS;
			NDTCell* cell = (NDTCell*) grids[gridIndex]->getCell( grids[gridIndex]->getIndex3D( pos ) );
			if ( cell != NULL ){
				numCoveredPoints++;
			}
		}
		debugOutput << "Covered register points: " << float32( 100.0f * numCoveredPoints ) / float32( registerPS->getNumEntries() * numGrids ) << ".\n"; debugOutput.flush();
	}
}

Matrix4f NDT::getTransformation() const
{
	if ( numGrids != 1 ){
		debugOutput << "Currently only implemented for single grid.\n"; debugOutput.flush();
		return IDENTITY4F;
	}
	return grids[0]->getInvNormalization() * T->getTransformation() * grids[0]->getNormalization();
}