#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "OOBBox3f.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "VertexArray.h"
#include "PCA.h"
#include "BoundingBox.h"
#include "DebugRenderer.h"
#include "PCA.inline.h"
#include "X4Exception.h"

//---------------------------------------------------------------------------
using namespace X4;

IMPLEMENT_X4_CLASS( OOBBox3f ,1)
{
	BEGIN_CLASS_INIT( OOBBox3f );
	ADD_BOUNDINGBOX3F_PROP(internalBB,1);
	ADD_MATRIX4F_PROP(toInternal,1);
}

OOBBox3f::OOBBox3f(void)
{
}

OOBBox3f::~OOBBox3f(void)
{
}

void OOBBox3f::assign(const Object* obj, X4CopyContext *context)
{
	const OOBBox3f * o = dynamic_cast<const OOBBox3f*>(obj);
	x4Assert( o != NULL );

	copyObjectProperties(obj,this);

	internalBB = o->internalBB;
	toInternal = o->toInternal;
}



Matrix4f fastInvertFrameMatrix( const Matrix4f &mat)
{
	Matrix3f rot = shrink4To3(mat);
	return expand3To4(rot.transpose()) * makeTranslation4f(shrink4To3(-mat[3]));
}


X4::OOBBox3f::OOBBox3f( PointCloud* pc )
{
	VertexArray vert(pc);
	PCA3f pca;
	for (mpcard i = 0; i < pc->getNumPoints(); i++)
	{
		pca.addPoint(vert.getPosition3f(i));
	}
	Matrix3f rotPCA;
	Vector3f eigenValues, center;
	pca.analyze( eigenValues, rotPCA, center );

	rotPCA = rotPCA.transpose();

	BoundingBox3f bb(rotPCA* (vert.getPosition3f(0)-center));
	for (mpcard i = 1; i < pc->getNumPoints(); i++)
	{
		Vector3f target = rotPCA* (vert.getPosition3f(i)-center);
		bb.addPoint(target);
	}

	internalBB = bb;
	toInternal = fastInvertFrameMatrix(makeTranslation4f(center)* expand3To4(rotPCA.transpose()));

/*	debugRenderer->beginRenderJob("bb");
	debugRenderer->addFineBoundingBox(makeTranslation4f(center)* expand3To4(rotPCA.transpose())  , bb, makeVector3f(1,1,0));
	debugRenderer->endRenderJob();
*/
}

X4::OOBBox3f::OOBBox3f( const std::vector<Vector3f> &points )
{
//	debugRenderer->beginRenderJob("oobb");	
	
	mpcard numPoints = points.size();
	PCA3f pca;
	for (mpcard i = 0; i < numPoints; i++)
	{
		pca.addPoint(points[i]);
//		debugRenderer->addFastSphere(points[i], 0.002f, makeVector3f(1,1,1));
	}
	Matrix3f rotPCA;
	Vector3f eigenValues, center;
	pca.analyze( eigenValues, rotPCA, center );



	rotPCA = rotPCA.transpose();

	BoundingBox3f bb(rotPCA* (points[0]-center));
	for (mpcard i = 1; i < numPoints; i++)
	{
		Vector3f target = rotPCA* (points[i]-center);
		bb.addPoint(target);
	}

	internalBB = bb;
	toInternal = fastInvertFrameMatrix(makeTranslation4f(center)* expand3To4(rotPCA.transpose()));

//	debugRenderer->addFineBoundingBox(makeTranslation4f(center)* expand3To4(rotPCA.transpose())  , internalBB, makeVector3f(1,1,0));
//	debugRenderer->endRenderJob();
}

X4::OOBBox3f::OOBBox3f( const std::vector<Vector3f> &points, const Vector3f &firstAxis )
{
	if( points.size() == 0)
		throw X4Exception("OOBBox3f::OOBBox3f(points,firstaxis) - no points provided!");
	// compute 3d centroid
	Vector3f centroid = NULL_VECTOR3F;
	for( mpcard i=0;i<points.size();i++ )
	{
		centroid += points[i];
	}
	centroid /= points.size();

	// transform points into tangent system & compute 2d pca in local frame
	Matrix3f tangentSystem = calcTangentSystem(firstAxis);
	Matrix3f invTangentSystem = tangentSystem.transpose();
	PCA<float,2> pca;
	for( mpcard i=0;i<points.size();i++ )
	{
		Vector2f p2d = shrink3To2( invTangentSystem * (points[i]-centroid));
		pca.addPoint(p2d);
	}
	Matrix2f eigenVectors;
	Vector2f eigenValues,centroid2d;
	pca.analyze(eigenValues,eigenVectors,centroid2d);

	// construct bb frame
	Matrix3f frame;
	frame[0] = firstAxis;
	frame[1] = tangentSystem[0]*eigenVectors[0][0] + tangentSystem[1]*eigenVectors[0][1];
	frame[2] = tangentSystem[0]*eigenVectors[1][0] + tangentSystem[1]*eigenVectors[1][1];
	Matrix4f frameToWorld = makeTranslation4f(centroid)*expand3To4(frame);
	toInternal = invertFrame(frameToWorld);

	for(mpcard i=0;i<points.size();i++ )
	{
		Vector3f tpoint = transformVector3f(toInternal,points[i]);
		if( i == 0)
		{
			internalBB.lowerCorner = internalBB.upperCorner = tpoint;
		}
		else
			internalBB.addPoint(tpoint);
	}
}

OOBBox3f::OOBBox3f(const std::vector<Vector3f> &points, const Matrix3f &frame) 
{
	if( points.size() == 0)
		throw X4Exception("OOBBox3f::OOBBox3f(points,frame) - no points provided!");
	// compute 3d centroid
	Vector3f centroid = NULL_VECTOR3F;
	for( mpcard i=0;i<points.size();i++ )
	{
		centroid += points[i];
	}
	centroid /= points.size();
	Matrix4f frameToWorld = makeTranslation4f(centroid) * expand3To4(frame);
	toInternal = invertFrame(frameToWorld);
	for(mpcard i=0;i<points.size();i++ )
	{
		Vector3f tpoint = transformVector3f(toInternal,points[i]);
		if( i == 0)
		{
			internalBB.lowerCorner = internalBB.upperCorner = tpoint;
		}
		else
			internalBB.addPoint(tpoint);
	}

}


void X4::OOBBox3f::visualize( const Vector3f &color, float linewidth ) const
{
	Vector3f corner[8];
	Matrix4f localToWorld = invertFrame(toInternal);
	for( unsigned i=0;i<8;i++ )
	{
		for( unsigned j=0;j<3;j++ )
		{
			corner[i][j] = (i & (1 << j)) != 0 ? internalBB.upperCorner[j] : internalBB.lowerCorner[j];
		}
		corner[i] = transformVector3f(localToWorld,corner[i]);
	}
	debugRenderer->addLine(corner[0],corner[1],color,color,linewidth*3.0f);
	debugRenderer->addLine(corner[0],corner[2],color,color,linewidth);
	debugRenderer->addLine(corner[0],corner[4],color,color,linewidth);

	debugRenderer->addLine(corner[7],corner[3],color,color,linewidth);
	debugRenderer->addLine(corner[7],corner[5],color,color,linewidth);
	debugRenderer->addLine(corner[7],corner[6],color,color,linewidth);

	debugRenderer->addLine(corner[3],corner[1],color,color,linewidth);
	debugRenderer->addLine(corner[4],corner[5],color,color,linewidth);
	debugRenderer->addLine(corner[5],corner[1],color,color,linewidth);

	debugRenderer->addLine(corner[2],corner[3],color,color,linewidth);
	debugRenderer->addLine(corner[4],corner[6],color,color,linewidth);
	debugRenderer->addLine(corner[2],corner[6],color,color,linewidth);

}

X4::Vector3f X4::OOBBox3f::getCenter() const
{
	Matrix4f invF = invertFrame(toInternal);
	return transformVector3f(invF,internalBB.getCenter());
}

void X4::OOBBox3f::normalize() {
	Vector3f center = (internalBB.upperCorner - internalBB.lowerCorner)/2.0f;

	toInternal = makeTranslation4f(-center) * toInternal;
	internalBB.upperCorner -= center;
	internalBB.lowerCorner -= center;
}


bool X4::OOBBox3f::intersectSphere( const Vector3f &center, float radius ) const
{
	Vector3f tcenter = transformVector3f(toInternal,center);
	return internalBB.computeSquaredDistance(tcenter) <= sqr(radius);
}

bool X4::OOBBox3f::intersect( const OOBBox3f &oobb ) const
{
	return intersect(*this,oobb);
}

bool X4::OOBBox3f::intersect( const BoundingBox3f &aabb ) const
{
	OOBBox3f tmpBB;
	tmpBB.setup(aabb,IDENTITY4F);
	return intersect(tmpBB);
}

bool X4::OOBBox3f::intersect(const OOBBox3f &oobb0,const OOBBox3f &oobb1)
{
	// transform oobb1 into coordinate frame of oobb0 and test for disjoint bb
	Matrix4f t1To0 = oobb0.toInternal * invertFrame(oobb1.toInternal);
	BoundingBox3f bb1_frame0(transformVector3f(t1To0,oobb1.internalBB.getCorner(0)));
	for( mpcard i=1;i<8;i++ )
	{
		Vector3f p = transformVector3f(t1To0,oobb1.internalBB.getCorner(i));
		bb1_frame0.addPoint(p);
	}

	if( bb1_frame0.intersectBoundingBox(oobb0.internalBB) == BVQueryDisjoint)
		return false;

	// test vice versa: transform oobb0 into coordinate frame of oobb1 and test for disjoint bb
	Matrix4f t0To1 = oobb1.toInternal * invertFrame(oobb0.toInternal);
	BoundingBox3f bb0_frame1(transformVector3f(t0To1,oobb0.internalBB.getCorner(0)));
	for( mpcard i=1;i<8;i++ )
	{
		Vector3f p = transformVector3f(t0To1,oobb0.internalBB.getCorner(i));
		bb0_frame1.addPoint(p);
	}

	return( bb0_frame1.intersectBoundingBox(oobb1.internalBB) != BVQueryDisjoint);
}

bool X4::OOBBox3f::isInside( const OOBBox3f &oobb0,const OOBBox3f &oobb1 )
{
	// transform oobb0 into coordinate frame of oobb1 and test for inclusion
	Matrix4f t0To1 = oobb1.toInternal * invertFrame(oobb0.toInternal);
	BoundingBox3f bb0_frame1(transformVector3f(t0To1,oobb0.internalBB.getCorner(0)));
	for( mpcard i=1;i<8;i++ )
	{
		Vector3f p = transformVector3f(t0To1,oobb0.internalBB.getCorner(i));
		bb0_frame1.addPoint(p);
	}

	return( oobb1.internalBB.intersectBoundingBox(bb0_frame1) == BVInsideQuery);
}

void X4::OOBBox3f::enlarge( const Vector3f &factors )
{
	for( mpcard d=0;d<3;d++ )
	{
		float c = (internalBB.upperCorner[d]+internalBB.lowerCorner[d])*0.5f;
		float l = (internalBB.upperCorner[d]-internalBB.lowerCorner[d])*0.5f*factors[d];
		internalBB.lowerCorner[d] = c-l;
		internalBB.upperCorner[d] = c+l;
	}
}

X4::BoundingBox3f X4::OOBBox3f::computeAABB() const
{
	Matrix4f bbToWorld = invertFrame(toInternal);
	BoundingBox3f retBB( transformVector3f(bbToWorld,internalBB.getCorner(0)));
	for( int i=1;i<8;i++ )
	{
		retBB.addPoint(transformVector3f(bbToWorld,internalBB.getCorner(i)));
	}
	return retBB;
}


std::vector<Vector3f> X4::OOBBox3f::getCorners() const
{
	Matrix4f localToWorld = invertFrame(toInternal);
	std::vector<Vector3f> ret(8);
	for( mpcard i=0;i<8;i++ )
	{
		ret[i] = transformVector3f(localToWorld,internalBB.getCorner(i));
	}
	return ret;
}

X4::BoundingBox3f X4::OOBBox3f::computeApproximateOverlap( const OOBBox3f &other ) const
{
	BoundingBox3f tmpBB;
	std::vector<Vector3f> corners = other.getCorners();
	for( mpcard i=0;i<8;i++)
	{
		Vector3f corner = corners[i];
		corner = transformVector3f(toInternal,corner);
		if( i == 0)
			tmpBB.lowerCorner = tmpBB.upperCorner = corner;
		else
			tmpBB.addPoint(corner);
	}

	return internalBB.computeIntersection(tmpBB);
}

float X4::OOBBox3f::computeBoundingSphereRadius( const Vector3f &v ) const
{
	float r=0;

	const std::vector<Vector3f> corners = getCorners();
	for( mpcard i=0;i<8;i++ )
		r = max(r,normQuad(v-corners[i]));
	return sqrt(r);
}

void OOBBox3f::addOOBB( const OOBBox3f &o) 
{
	Matrix4f frameToWorld = invertFrame(o.getTransformation());
	for( mpcard i=0;i<8;i++ )
	{
		Vector3f corner = o.getInternalBoundingBox().getCorner(i);
		addPoint(transformVector3f(frameToWorld,corner));
	}
}

void OOBBox3f::addPoint( const Vector3f &v) 
{
	internalBB.addPoint( transformVector3f(toInternal,v));
}
