#ifndef OOBBox3fH
#define OOBBox3fH
//---------------------------------------------------------------------------
#include "x3foundation_private.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
//---------------------------------------------------------------------------

namespace X4
{
	class PointCloud;

/**
	oriented bounding box
*/
class X3FOUNDATION_PRIVATE_API OOBBox3f : public Persistent
{
	X4_CLASS( OOBBox3f )

private:
	BoundingBox3f internalBB;
	Matrix4f toInternal;


public:
	OOBBox3f(void);
	OOBBox3f(PointCloud* pc);
	~OOBBox3f(void);

	OOBBox3f(const std::vector<Vector3f> &points);

	/// compute oriented bounding box with a specified axis, axis 2 and 3 are computed via pca in 2d subspace
	OOBBox3f(const std::vector<Vector3f> &points, const Vector3f &firstAxis);

	/// compute oriented bounding box with predefined coordinate axes
	OOBBox3f(const std::vector<Vector3f> &points, const Matrix3f &frame);

	/// direct setup
	void setup(const BoundingBox3f &bb, const Matrix4f &worldToLocal) {internalBB = bb; toInternal = worldToLocal;}

	/// add other oriented bounding box (conservative, just adds corner points)
	void addOOBB( const OOBBox3f &o);

	/// add single point (in world space) to bounding box, coordinate frame is not modified
	void addPoint( const Vector3f &v);

   virtual void assign(const Object* obj, X4CopyContext *context = NULL);

   void enlarge(float32 factor){
		internalBB.enlarge(factor);
   }

   /// enlarge internal bb (anisotropic scaling, (1,1,1) means unmodified)
   void enlarge( const Vector3f &factors);

	/// add border to inner bb
	void addBorder(float b) {internalBB.addBorder(b);}
	
	inline bool inBoundingBox(const Vector3f &point) const{
	   Vector3f transformedPoint = transformVector3f(toInternal, point);
	   return internalBB.inBoundingBox(transformedPoint);
   }

   /// does sphere intersect oobb?
   bool intersectSphere(const Vector3f &center, float radius) const;

   /// intersection with other oobb
   bool intersect( const OOBBox3f &oobb) const;

   /// intersection with aabb
   bool intersect( const BoundingBox3f &aabb) const;

   /// test two oobbs for intersection
   static bool intersect(const OOBBox3f &oobb0,const OOBBox3f &oobb1);

   /// test if oobb0 is included in oobb1
   static bool isInside(const OOBBox3f &oobb0,const OOBBox3f &oobb1);


   /// get bounding box in local coordinates
   BoundingBox3f getInternalBoundingBox() const {return internalBB;}

   /// get transformation to local coordinate frame
   Matrix4f getTransformation() const {return toInternal;}

   /// transform oobb
   void transform(const Matrix4f &T) {toInternal = toInternal * invertFrame(T);}

   /// get oobb center
   Vector3f getCenter() const;

   /// compute radius of bounding sphere centered at p
   float computeBoundingSphereRadius( const Vector3f &v) const;

   /// visualize via dr
   void visualize(const Vector3f &color, float linewidth) const;

   /// compute minimal enclosing aabb
   BoundingBox3f computeAABB() const;

   /// boundingbox lowercorner = - uppercorner;
   void normalize();

   /// get oobb corners
   std::vector<Vector3f> getCorners() const;

   /// compute approximate overlap by setting up a new bounding box in this domain
   /// (transform corners of other into this frame). returns intersection box
   BoundingBox3f computeApproximateOverlap( const OOBBox3f &other) const;

   

};

}

#endif