#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PCCCacheOptimizer.h"
//---------------------------------------------------------------------------
#include "UnstructuredInCorePointCloud.h"
#include "SuperIterator.inline.h"
#include "HierarchicalKNNIterator.h"
#include "Scene.h"
#include "SceneGraphState.h"
#include "PointCloudTools.h"
#include "PointSet.h"
#include "UniformPointCloudResampling.h"
#include "CopyObjectProperties.h"
#include "timer.h"
#include "StringHelper.h"
#include "ClassMethods.h"
#include "StringHelper.h"
#include <QtCore/QDir>
#include <QtGUI/QFileDialog>
#include <qtGUI/QInputDialog>
#include <fstream>
#include "FileDialogs.h"
//---------------------------------------------------------------------------
using namespace X4;

IMPLEMENT_X4_CLASS( PCCCacheOptimizer ,0)
{
	BEGIN_CLASS_INIT( PCCCacheOptimizer );

	ADD_FLOAT32_PROP( m_Radius,0 );
	ADD_NOARGS_METHOD(optimizeForOctree);
	ADD_NOARGS_METHOD(countLinesInFiles);
	ADD_NOARGS_METHOD(specialSearchInFile);
//	ADD_BOOLEAN_PROP( m_UseUniformResampler, 0 );
}

PCCCacheOptimizer::PCCCacheOptimizer(void)
{
	m_Radius = 2.7f;
	m_UseUniformResampler = false;	// slow and didn't do ANY improvement!
}

PCCCacheOptimizer::~PCCCacheOptimizer(void)
{
}

void PCCCacheOptimizer::assign(const Object* obj, X4CopyContext *context)
{
	PCCommand::assign(obj,context);
	const PCCCacheOptimizer * o = dynamic_cast<const PCCCacheOptimizer*>(obj);
	x4Assert( o != NULL );

	copyObjectProperties(obj,this);
}

void PCCCacheOptimizer::execute(Scene *scene)
{
	const std::vector<std::string> selection = getScene()->getRootState()->staticState->getSelectedNodes();
	for( card32 i=0;i<selection.size();i++ )
	{
		UnstructuredInCorePointCloud * upc = dynamic_cast<UnstructuredInCorePointCloud*>(getPointCloud( getScene(), selection[i] ));
		if( upc )
		{
			UnstructuredInCorePointCloud * res = optimize( upc, m_Radius,m_UseUniformResampler );
			upc->getPointSet()->assign(res->getPointSet());
			upc->clearAttachments();
			delete res;
		}
	}
}

UnstructuredInCorePointCloud * PCCCacheOptimizer::optimize( UnstructuredInCorePointCloud * upc, float radius_cache,
																			  bool useUniformResampler)
{
	Timer timer;
	float medianPointDistance = getMedianPointDistance(upc);
	radius_cache *= medianPointDistance;

	UnstructuredInCorePointCloud * ret = new UnstructuredInCorePointCloud;
	ret->clearAndSetup( upc->getDescr(), upc->getNumPoints() );
	PointSet * ps = upc->getPointSet();
	PointSet * retPS = ret->getPointSet();
	AAT posAAT = upc->getAAT("position");
	//AAT colorAAT = upc->getAAT("color");

	std::vector<card32> indices( upc->getNumPoints());
	std::vector<bool> mask( upc->getNumPoints(), false);
	for( card32 i=0;i<indices.size();i++ )
		indices[i] = i;

	// permutate indices
	for( card32 i=0;i<indices.size()-1;i++ )
	{
		card32 newIndex = i + rand() % (indices.size()-i);
		card32 tmp = indices[newIndex];
		indices[newIndex] = indices[i];
		indices[i] = tmp;
	}

	UnstructuredInCorePointCloud * resampledPC = NULL;
	SuperIterator * resampledIter = NULL;

	if( useUniformResampler )
	{
		UniformPointCloudResampling resampler;
		resampler.setup( radius_cache*1.8f);
		resampledPC = new UnstructuredInCorePointCloud;
		resampledPC->getPointSet()->assign( resampler.resamplePointSet( upc->getPointSet() ));
		resampledIter = new SuperIterator( resampledPC->createIterator( 
			SceneObjectIteratorType::CAP_BASIC_PC ), upc->getDescr() );
		debugOutput << "resampled #numpoints: " << (card32)resampledPC->getNumPoints() << "\n";
	}
	
	std::vector<char> vbuffer( upc->getDescr()->getSize() );

	HierarchicalKNNIterator hIt( upc,32,NULL);
	card32 resPointIndex = 0;
	Vector3f col = makeVector3f(0,0,0);
	for( card32 i=0;i<indices.size();i++)
	{
		Vector3f samplePos;
		if( resampledIter != NULL && !resampledIter->atEnd() )
		{
			i=-1;
			samplePos = resampledIter->get<Vector3f>(posAAT);
			resampledIter->next();
		}
		else
		{
			card32 index = indices[i];
			if( mask[index] )
				continue;

			samplePos = ps->get3f(index,posAAT);
		}

		hIt.setSeekPointAndReset(samplePos);
		if( mask[hIt.getCurrentPointIndex()] )
		{
			continue;
		}

		while( !hIt.atEnd())
		{
			Vector3f pos = hIt.get3f(posAAT);
			if( norm( pos - samplePos) >= radius_cache )
			{
				break;
			}

			card64 tmpIndex = hIt.getCurrentPointIndex();
			if( !mask[tmpIndex] )
			{
				//ps->set3f(tmpIndex, colorAAT, col );
				ps->getVertex( tmpIndex, &vbuffer[0] );
				retPS->setVertex(resPointIndex, &vbuffer[0]);
				resPointIndex++;
				mask[tmpIndex] = true;
			}

			hIt.next();
		}

		col += makeVector3f(1,1,1) * 0.1f;

	}

	delete resampledIter;
	delete resampledPC;

	debugOutput << "cache optimizer: " << convertTimeToString(timer.getDeltaValue()) << "\n";
	return ret;
}

void X4::PCCCacheOptimizer::optimizeForOctree()
{
	const std::vector<std::string> selection = getScene()->getRootState()->staticState->getSelectedNodes();
	for( card32 i=0;i<selection.size();i++ )
	{
		UnstructuredInCorePointCloud * upc = dynamic_cast<UnstructuredInCorePointCloud*>(getPointCloud( getScene(), selection[i] ));
		if( upc )
		{
			optimizeForOctree( upc);
		}
	}

}

void X4::PCCCacheOptimizer::optimizeForOctree( UnstructuredInCorePointCloud * upc )
{
	AttachedIndexedOctree * octree = NULL;


	octree = dynamic_cast<AttachedIndexedOctree*>(upc->getAttachments()->getData("SimpleOctree"));
	if(octree==NULL){
		octree=new AttachedIndexedOctree();
		octree->setup("SimpleOctree", AttachedData::ADF_QUERY_DELETE_ON_UPDATE,  NULL); // attachedOctreeMemManager !!
		octree->octreesetup(upc->getPointSet(), 32);
		upc->getAttachments()->attachData(octree);
	}

	PointSet * psNew = (PointSet*)upc->getPointSet()->copy();
	unsigned currentIndex = 0;

	list<AIOctreeNode *> nodeQueue;
	nodeQueue.push_back(octree->getOctree());

	std::vector<char> vbuffer(upc->getDescr()->getSize(),0);
	while( !nodeQueue.empty() )
	{
		AIOctreeNode * node = nodeQueue.front();
		nodeQueue.pop_front();

		if( node->isLeaf() )
		{
			std::vector<unsigned> indices = node->getContent().m_index;
			for( unsigned i=0;i<indices.size();i++ )
			{
				upc->getPointSet()->getVertex(indices[i], &vbuffer[0]);
				psNew->setVertex(currentIndex,&vbuffer[0]);
				node->getContent().m_index[i] = currentIndex;
				currentIndex++;
			}

		}
		else
		{
			for( unsigned i=0;i<8;i++ )
			{
				AIOctreeNode * c = node->getChild(i);
				if( c )
					nodeQueue.push_back(c);
			}
		}

	}

	//delete octree;
	upc->getPointSet()->assign(psNew);
	delete psNew;

}

void recursiveCollectFileNames(std::string file, std::vector<std::string> &filenames) 
{
	QFileInfo info(file.c_str());
	if (info.isDir()) 
	{
		QDir dir(file.c_str());
		for (unsigned i=0; i<dir.count(); i++) 
		{
			if (!(dir[i] == QString(".") || dir[i] == QString("..") || dir[i].isEmpty())) {
				recursiveCollectFileNames(qString2STLString(dir.absoluteFilePath(dir[i])),filenames);
			}
		}
	} else 
	{
		if (info.isFile() && info.exists() && info.isReadable()) 
		{
			filenames.push_back(file);
		}
	}
}


void X4::PCCCacheOptimizer::countLinesInFiles()
{
	QString dir = QFileDialog::getExistingDirectory(NULL,"project directory");

	std::vector<std::string> filenames;
	recursiveCollectFileNames(qString2STLString(dir),filenames);

	mpcard numLines = 0;
	mpcard numNonEmptyLines = 0;
	mpcard numFiles = 0;

	for( mpcard i=0;i<filenames.size();i++ )
	{
		std::string ext = extractFileExt(filenames[i]);
		if( ext != ".cpp" && ext != ".h" && ext != ".c" && ext != ".hpp" && ext != ".cc")
			continue;

		std::ifstream f(filenames[i].c_str());
		if( !f.is_open())
			continue;

		numFiles++;
		while (!f.eof())
		{
			std::string line;
			getline(f,line);

			numLines++;
			if( line.empty() )
				continue;

			std::vector<std::string> words = splitString(line," \t");
			if( words.size() == 0)
				continue;
			numNonEmptyLines++;
		}
	}

	message("Summary:\n");
	debugOutput << "num files: " << numFiles
		<< "\nnum lines: " << numLines <<
		"\nnum non-empty lines: " << numNonEmptyLines << "\n";
}

void X4::PCCCacheOptimizer::specialSearchInFile()
{
	std::string fname = FileDialogs::getOpenFileName(NULL,"file","*.*");
	if( fname.empty())
		return;
	QString searchStrQ = QInputDialog::getText(NULL,"query text","enter a query text");
	std::string qstr = qString2STLString(searchStrQ);
	if( qstr.empty())
		return;

	message("\n");
	
	std::ifstream f(fname.c_str());
	if( !f.is_open())
		return;

	while (!f.eof())
	{
		std::string line;
		getline(f,line);

		if( line.length()<qstr.length() )
			continue;

		std::string ostr;
		mpcard pos = 0;
		for( mpcard i=0;i<line.length();i++ )
		{
			if( qstr[pos] == line[i] )
			{
				ostr += '#';
				ostr += line[i];

				pos++;
				if( pos == qstr.length() )
				{
					debugOutput << ostr << line.substr(i,line.length()-i) << "\n";
					break;
				}
			}
			else
				ostr += line[i];
		}

	}
}
