#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PCCPoissionSampling.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "GeometricTools.h"
#include "UnstructuredInCorePointCloud.h"
#include "PointCloudTools.h"
#include "SceneGraphState.h"
#include "SeparatorClassProperty.h"
#include "SGObjectNode.h"
#include "SGRelativeTimeAnimationNode.h"
#include "SceneGraphState.h"
#include "GenericSwapAllOutOfCorePointCloud.h"
#include "StringHelper.h"
#include "ProgressWindow.h"
//---------------------------------------------------------------------------
using namespace X4;

IMPLEMENT_X4_CLASS( PCCPoissionSampling ,1)
{
	BEGIN_CLASS_INIT( PCCPoissionSampling );
	ADD_FLOAT32_PROP(m_MaxSamplingSpacing,0 );
	ADD_FLOAT32_PROP(m_RelativeNormalSigma,0 );
	ADD_SEPARATOR("OOC");
	ADD_BOOLEAN_PROP(m_CreateOCCPointCloud,1);
	ADD_STRING_PROP(m_OCCFileName,1);

}

PCCPoissionSampling::PCCPoissionSampling(void)
{
	m_MaxSamplingSpacing = 0.01f;
	m_RelativeNormalSigma = 1.0f;
	m_OCCFileName = "resampled";
}

PCCPoissionSampling::~PCCPoissionSampling(void)
{
}

void PCCPoissionSampling::assign(const Object* obj, X4CopyContext *context)
{
	const PCCPoissionSampling * o = dynamic_cast<const PCCPoissionSampling*>(obj);
	x4Assert( o != NULL );

	copyObjectProperties(obj,this);
}

PointCloud * X4::PCCPoissionSampling::resampleAndComputeNormals( PointCloud * pc, const Vector3f &cameraPos )
{
	UnstructuredInCorePointCloud * ret = new UnstructuredInCorePointCloud;
	ret->clearAndSetup(pc->getDescr(),0);
	checkAttribute(ret,"normal",3,VAD::DATA_FORMAT_FLOAT32);

	std::vector<unsigned> marked(pc->getNumPoints(),0);
	SceneObjectIterator * iter = pc->createIterator(SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC);
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	HierarchicalKNNIterator hIt(pc,32,NULL);

	AAT posAAT = ret->getAAT("position");
	AAT normalAAT = ret->getAAT("normal");
	float rad_2 = pow(m_MaxSamplingSpacing,2.0f);

	unsigned index = 0;
	std::vector<char> vbuffer(ret->getDescr()->getSize());
	while(!bIt->atEnd())
	{
		if( marked[index] == 0 )
		{
			marked[index] = 1;
			Vector3f pos = bIt->get3f(posAAT);
			hIt.setSeekPointAndReset(pos);
			while(!hIt.atEnd())
			{
				Vector3f tmpPos = hIt.get3f(posAAT);
				hIt.next();
				if( normQuad(tmpPos - pos) > rad_2 )
				{
					break;
				}

				marked[hIt.getCurrentPointIndex()] = 1;

			}

			bIt->getVertex( &vbuffer[0]);
			ret->getPointSet()->addPoint(&vbuffer[0]);

			Vector3f projPos,projNormal;
			projectPointOntoPatch( pos, m_MaxSamplingSpacing*m_RelativeNormalSigma,pc,&hIt,projPos,projNormal);
			if( (projPos - cameraPos) * projNormal > 0 )
				projNormal *= -1;

			ret->getPointSet()->set3f(ret->getNumPoints()-1,posAAT,projPos);
			ret->getPointSet()->set3f(ret->getNumPoints()-1,normalAAT,projNormal);
		
		}


		bIt->next();
		index++;
	}


	delete iter;
	return ret;
}

void X4::PCCPoissionSampling::execute( Scene * scene )
{
	std::vector<std::string> selection = scene->getRootState()->staticState->getSelectedNodes();
	if( m_CreateOCCPointCloud && m_OCCFileName.length() == 0 )
		throw X4Exception("PCCPoissionSampling::execute() - No ooc filename was specified!");

	SGListNode * rnode = dynamic_cast<SGListNode*>(getSceneGraphNodeByName(scene,"root/resampled"));
	if( rnode == NULL )
	{
		rnode = new SGRelativeTimeAnimationNode;
		rnode->setName("resampled");
		scene->getRootNode()->addChildNode(rnode);
	}


	for( unsigned i=0;i<selection.size();i++ )
	{
		PointCloud * pc = getPointCloud(scene,selection[i]);
		if( pc )
		{
			PointCloud * newPC = resampleAndComputeNormals(pc,scene->getRootState()->staticState->camera.getPosition());
			if( !m_CreateOCCPointCloud )
				addPointCloud(scene,newPC,selection[i] + "_resampled");
			else
			{
				GenericSwapAllOutOfCorePointCloud * gpc = new GenericSwapAllOutOfCorePointCloud;
				std::string num = intToStr(i);
				while( num.length() < 5)
					num = std::string("0") + num;
				gpc->clearAndSetup(newPC, m_OCCFileName + num + ".object");
				addPointCloud(scene,gpc,std::string("root/resampled/resampled") + num);
			}
		}

		SGListNode * ld = dynamic_cast<SGListNode*>(getSceneGraphNodeByName(scene,selection[i]));
		if( ld  && m_CreateOCCPointCloud)
		{
			SceneGraphState * state = scene->getRootState();
			progressWindow->pushStep();
			for( unsigned j=0;j<ld->getNumChildNodes(state); j++)
			{
				progressWindow->progress(j * 100 / ld->getNumChildNodes(state));
				SGObjectNode * objNode = dynamic_cast<SGObjectNode*>(ld->getChildNode(state,j));
				if( objNode )
				{
					PointCloud * pc = (PointCloud*)objNode->getSceneObject();
					PointCloud * newPC = resampleAndComputeNormals(pc,scene->getRootState()->staticState->camera.getPosition());
					GenericSwapAllOutOfCorePointCloud * gpc = new GenericSwapAllOutOfCorePointCloud;
					std::string num = intToStr(j);
					while( num.length() < 5)
						num = std::string("0") + num;
					gpc->clearAndSetup(newPC, m_OCCFileName + num + ".object");
					addPointCloud(scene,gpc,std::string("root/resampled/resampled") + num);
				}
			}
			progressWindow->popSteps();
		}
	}
}