//---------------------------------------------------------------------------
#include "StdAfx.h"
#include <windows.h>
#include <mmsystem.h>
//---------------------------------------------------------------------------
#include "PCCRegistration.h"
#include "ClassMethods.h"
#include "StringClassProperty.h"
#include "ObjectClassProperty.h"
#include "BooleanClassProperty.h"
#include "SeparatorClassProperty.h"
#include <time.h>
#include <qtcore/QTextStream>
#include "GLShaderMaterial.h"
#include "NumericalClassProperty.h"
#include "PointCloudTools.h"
#include "UnstructuredInCorePointCloud.h"
#include "DebugOutput.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

IMPLEMENT_X4_CLASS( PCCRegistration, 6 )
{
	BEGIN_CLASS_INIT( PCCRegistration )
	ADD_STRING_PROP( basePCName, 0 )
	ADD_STRING_PROP( registerPCName, 0 )
	ADD_INT32_PROP( numIterations, 0 )
	ADD_INT32_PROP( maxNumIterations, 5 )
	ADD_OBJECT_PROP( registerizer, 0, Registerizer::getClass(), NULL )
	ADD_OBJECT_PROP( regDebug, 0, RegistrationDebug::getClass(), NULL )
	ADD_FLOAT32_PROP( convergenceChange, 4 )
	ADD_BOOLEAN_PROP( writeFrames, 6 )

	ADD_NOARGS_METHOD( setNDT )
   SET_METHOD_AS_CMD_OBJ( true )
	ADD_NOARGS_METHOD( setICP )
   SET_METHOD_AS_CMD_OBJ( true )
	ADD_NOARGS_METHOD( setup )
   SET_METHOD_AS_CMD_OBJ( true )
	ADD_NOARGS_METHOD( step )
   SET_METHOD_AS_CMD_OBJ( true )
	ADD_NOARGS_METHOD( nSteps )
   SET_METHOD_AS_CMD_OBJ( true )
	ADD_NOARGS_METHOD( exportData )
	ADD_NOARGS_METHOD( printNDTStats )
}

PCCRegistration::PCCRegistration() 
{
	registerizer = new NDT();
	basePCName = "root/base";
	registerPCName = "root/register";
	numIterations = 30;
	maxNumIterations = 200;
	regDebug = new RegistrationDebug();
	regDebug->setRegisterizer( registerizer );
	convergenceChange = 0.9999f;
	writeFrames = true;
}

/***
* Destructor
*/
PCCRegistration::~PCCRegistration() 
{
	delete regDebug;
	delete registerizer;
}

void PCCRegistration::setup()
{
	PointCloud* basePC = getPointCloud( scene, basePCName );
	UnstructuredInCorePointCloud* baseUPC = dynamic_cast<UnstructuredInCorePointCloud*>( basePC );
	if ( baseUPC == NULL ){
		debugOutput << "setup: invalid base point cloud.\n"; debugOutput.flush();
		return;
	}
	PointCloud* registerPC = getPointCloud( scene, registerPCName );
	UnstructuredInCorePointCloud* registerUPC = dynamic_cast<UnstructuredInCorePointCloud*>( registerPC );
	if ( registerUPC == NULL ){
		debugOutput << "setup: invalid register point cloud.\n"; debugOutput.flush();
		return;
	}
	PointSet* basePS = baseUPC->getPointSet();
	PointSet* registerPS = registerUPC->getPointSet();
	float32 startTime = clock();
	registerizer->setup( basePS );	
	registerizer->setRegisterPS( registerPS );
	registerizer->selectRegisterCandidates();
	#if (VERBOSE_TIME)
	debugOutput << "Time required (setup): " << float32( ( clock() - startTime ) / CLK_TCK ) << ".\n"; debugOutput.flush();
	#endif

	redrawScene();
	for ( unsigned int n = 0; n < scene->getNumGLMaterials(); n++ ){
		GLShaderMaterial* mat = dynamic_cast<GLShaderMaterial*>( scene->getGLMaterial( n ) );
		if ( mat != NULL ){
			mat->increaseVersion();
		}
	}
	redrawScene();
}

void PCCRegistration::glDrawTool( GLContext *glContext )
{
	regDebug->renderGL();
}

void PCCRegistration::connectScene( Scene *scene, OnSceneChangeCallback *onChange, QWidget *sceneEditorWidget )
{
	PCCommand::connectScene( scene, onChange, sceneEditorWidget );
}

void PCCRegistration::exportData()
{
	PointCloud* basePC = getPointCloud( scene, basePCName );
	UnstructuredInCorePointCloud* baseUPC = dynamic_cast<UnstructuredInCorePointCloud*>( basePC );
	if ( basePC == NULL ){
		debugOutput << "exportData: invalid base point cloud.\n"; debugOutput.flush();
		return;
	}
	PointCloud* registerPC = getPointCloud( scene, registerPCName );
	UnstructuredInCorePointCloud* registerUPC = dynamic_cast<UnstructuredInCorePointCloud*>( registerPC );
	if ( registerPC == NULL ){
		debugOutput << "exportData: invalid base register cloud.\n"; debugOutput.flush();
		return;
	}
	PointSet* basePS = baseUPC->getPointSet();
	PointSet* registerPS = registerUPC->getPointSet();

	// Base
	QFile baseFile( "basePS.txt" );
	if ( !baseFile.open( QFile::WriteOnly | QFile::Truncate ) ) {
		debugOutput << "exportData: cannot open file for writing.\n"; debugOutput.flush();
	} else {
		QTextStream out( &baseFile );
		AAT BASE_POSITION = basePS->getAAT( "position" ); 
		AAT BASE_COLOR = basePS->getAAT( "color" );
		for ( unsigned int pointIndex = 0; pointIndex < basePS->getNumEntries(); pointIndex++ ){
			Vector3f pos = basePS->get3f( pointIndex, BASE_POSITION );
			Vector3f color = basePS->get3f( pointIndex, BASE_COLOR );
			out << pos[0] << " " << pos[1] << " " << pos[2] << " ";
			out << color[0] << " " << color[1] << " " << color[2] << "\n";
		}
	}

	// Register
	QFile registerFile( "registerPS.txt" );
	if ( !registerFile.open( QFile::WriteOnly | QFile::Truncate ) ) {
		debugOutput << "exportData: cannot open file for writing.\n"; debugOutput.flush();
	} else {
		QTextStream out( &registerFile );
		AAT REGISTER_POSITION = registerPS->getAAT( "position" ); 
		AAT REGISTER_COLOR = registerPS->getAAT( "color" );
		for ( unsigned int pointIndex = 0; pointIndex < registerPS->getNumEntries(); pointIndex++ ){
			Vector3f pos = registerPS->get3f( pointIndex, REGISTER_POSITION );
			Vector3f color = registerPS->get3f( pointIndex, REGISTER_COLOR );
			out << pos[0] << " " << pos[1] << " " << pos[2] << " ";
			out << color[0] << " " << color[1] << " " << color[2] << "\n";
		}
	}
}

void PCCRegistration::printNDTStats()
{
	NDT* ndt = dynamic_cast<NDT*>( registerizer );
	if ( ndt != NULL ){
		ndt->printStats();
	}
}

void PCCRegistration::step()
{
	if ( !registerizer->isReady() ){
		debugOutput << "nSteps: call setup first!\n"; debugOutput.flush();
		return;
	}

	float32 startTime = clock();

	registerizer->reset();
	registerizer->solve( 1 );

	// Apply transformation
	Matrix4f T = registerizer->getTransformation();
	PointSet* registerPS = registerizer->getRegisterPS();
	AAT POSITION = registerPS->getAAT( "position" );
	for ( unsigned int pointIndex = 0; pointIndex < registerPS->getNumEntries(); pointIndex++ ){
		Vector3f p = registerPS->get3f( pointIndex, POSITION );
		Vector4f p4 = makeVector4f( p[0], p[1], p[2], 1 );
		p4 = T * p4;
		registerPS->set3f( pointIndex, POSITION, makeVector3f( p4[0], p4[1], p4[2] ) );
	}
	#if (VERBOSE_TIME)
	debugOutput << "Time required (step): " << float32( ( clock() - startTime ) / CLK_TCK ) << ".\n"; debugOutput.flush();
	#endif

	// Update view
	for ( unsigned int n = 0; n < scene->getNumGLMaterials(); n++ ){
		GLShaderMaterial* mat = dynamic_cast<GLShaderMaterial*>( scene->getGLMaterial( n ) );
		if ( mat != NULL ){
			mat->increaseVersion();
		}
	}
	redrawScene();
}

void PCCRegistration::nSteps()
{
	if ( !registerizer->isReady() ){
		debugOutput << "nSteps: call setup first!\n"; debugOutput.flush();
		return;
	}

	PointCloud* basePC = getPointCloud( scene, basePCName );
	UnstructuredInCorePointCloud* baseUPC = dynamic_cast<UnstructuredInCorePointCloud*>( basePC );
	if ( basePC == NULL ){
		debugOutput << "nSteps: invalid base point cloud.\n"; debugOutput.flush();
		return;
	}
	PointCloud* registerPC = getPointCloud( scene, registerPCName );
	UnstructuredInCorePointCloud* registerUPC = dynamic_cast<UnstructuredInCorePointCloud*>( registerPC );
	if ( registerPC == NULL ){
		debugOutput << "nSteps: invalid base register cloud.\n"; debugOutput.flush();
		return;
	}
	PointSet* basePS = baseUPC->getPointSet();
	PointSet* registerPS = registerUPC->getPointSet();
	AAT REGISTER_POSITION = registerPS->getAAT( "position" );
	float32 startTime = clock();
	for ( int32 n = 0; n < numIterations; n++ ){
		registerizer->reset();
		float32 error = registerizer->solve( 1 );

		// Apply transformation
		Matrix4f T = registerizer->getTransformation();
		for ( unsigned int pointIndex = 0; pointIndex < registerPS->getNumEntries(); pointIndex++ ){
			Vector3f p = registerPS->get3f( pointIndex, REGISTER_POSITION );
			Vector4f p4 = makeVector4f( p[0], p[1], p[2], 1 );
			p4 = T * p4;
			registerPS->set3f( pointIndex, REGISTER_POSITION, makeVector3f( p4[0], p4[1], p4[2] ) );
		}

		// Update view
		for ( unsigned int n = 0; n < scene->getNumGLMaterials(); n++ ){
			GLShaderMaterial* mat = dynamic_cast<GLShaderMaterial*>( scene->getGLMaterial( n ) );
			if ( mat != NULL ){
				mat->increaseVersion();
			}
		}
		redrawScene();
		writeFrameToDisk();
	}
	#if (VERBOSE_TIME)
	debugOutput << "Time required (nSteps): " << float32( ( clock() - startTime ) / CLK_TCK ) << ".\n"; debugOutput.flush();
	#endif
}

void PCCRegistration::execute( Scene *scene )
{
	if ( !registerizer->isReady() ){
		debugOutput << "execute: call setup first!\n"; debugOutput.flush();
		return;
	}

	float32 startTime = clock();

	ICP* icp = dynamic_cast<ICP*>( registerizer );
	NDT* ndt = dynamic_cast<NDT*>( registerizer );
	if ( ( icp == NULL ) && ( ndt == NULL ) ){
		debugOutput << "execute: only implemented for ICP and NDT.\n"; debugOutput.flush();
		return;
	}

	float32 score = -1, lastScore = -1;
	int32 numIterations = 0;
	float32 convergenceVar = 0;
	PointSet* registerPS = registerizer->getRegisterPS();
	AAT REGISTER_POSITION = registerPS->getAAT( "position" );
	while( ( ( numIterations < 2 ) || ( convergenceVar < convergenceChange ) ) && ( numIterations < maxNumIterations ) ){
		lastScore = score;
		registerizer->reset();
		score = registerizer->solve( 1 );

		
		// Apply transformation
		Matrix4f T = registerizer->getTransformation();
		for ( unsigned int pointIndex = 0; pointIndex < registerPS->getNumEntries(); pointIndex++ ){
			Vector3f p = registerPS->get3f( pointIndex, REGISTER_POSITION );
			Vector4f p4 = makeVector4f( p[0], p[1], p[2], 1 );
			p4 = T * p4;
			registerPS->set3f( pointIndex, REGISTER_POSITION, makeVector3f( p4[0], p4[1], p4[2] ) );
		}

		convergenceVar = min( score, lastScore ) / max( score, lastScore );

		//debugOutput << "Score: " << score << ", convergence rate: " << convergenceVar << "/" << convergenceChange << ".\n"; debugOutput.flush();

		for ( unsigned int n = 0; n < scene->getNumGLMaterials(); n++ ){
			GLShaderMaterial* mat = dynamic_cast<GLShaderMaterial*>( scene->getGLMaterial( n ) );
			if ( mat != NULL ){
				mat->increaseVersion();
			}
		}
		redrawScene();
		writeFrameToDisk();

		numIterations++;
	}

	#if (VERBOSE_TIME)
	debugOutput << "Time required: " << float32( ( clock() - startTime ) / CLK_TCK ) << ".\n";
	debugOutput << "Iterations required: " << numIterations << ".\n"; 
	debugOutput.flush();
	#endif

	// Update view
	for ( unsigned int n = 0; n < scene->getNumGLMaterials(); n++ ){
		GLShaderMaterial* mat = dynamic_cast<GLShaderMaterial*>( scene->getGLMaterial( n ) );
		if ( mat != NULL ){
			mat->increaseVersion();
		}
	}
	redrawScene();
}

void PCCRegistration::setNDT()
{
	if ( registerizer != NULL ) delete registerizer;
	registerizer = new NDT();
	gridCellType = NDT_CELL_TYPE;
	regDebug->setRegisterizer( registerizer );
}

void PCCRegistration::setICP()
{
	if ( registerizer != NULL ) delete registerizer;
	registerizer = new ICP();
	regDebug->setRegisterizer( registerizer );
}