//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PCCResampleByCurvature.h"
//---------------------------------------------------------------------------
#include "BooleanClassProperty.h"
#include "NumericalClassProperty.h"
#include "DebugOutput.h"
#include "PointCloudIterators.h"
#include "PointSet.h"
#include "PointSet.inline.h"
#include "UnstructuredInCorePointCloud.h"
#include "PointCloudTools.h"
#include "SceneGraphState.h"
#include "StandardSceneGraphIterator.h"
#include "SGObjectNode.h"
#include "SGAnimationNode.h"
#include "ProgressWindow.h"
#include "SeparatorClassProperty.h"
#include "InCorePCTopologyGraph.h"
#include "TopologyAttachment.h"
#include "TopologicalKNNIterator.h"
#include "GeometricTools.h"
#include <opencv\cxcore.h>
#include <float.h>

//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

IMPLEMENT_X4_CLASS( PCCResampleByCurvature, 3)
{
   BEGIN_CLASS_INIT( PCCResampleByCurvature )
   ADD_FLOAT32_PROP( minSpacing, 1 )
   ADD_FLOAT32_PROP( maxSpacing, 1 )
   //ADD_INT32_PROP (curvatureChannel, 1)
   ADD_FLOAT32_PROP (mTopologyReconnectMaxRadius, 2);
   ADD_BOOLEAN_PROP (useTopologicalRangeQuery, 3);
}

PCCResampleByCurvature::PCCResampleByCurvature()
{
   minSpacing = 0.01f;
   maxSpacing = 0.1f;
   //curvatureChannel = 2;
   mTopologyReconnectMaxRadius = 0;
   useTopologicalRangeQuery = false;
}

PCCResampleByCurvature::~PCCResampleByCurvature()
{
}


void PCCResampleByCurvature::resamplePC(UnstructuredInCorePointCloud* pc, float minSpacing, float maxSpacing, float mTopologyReconnectMaxRadius, bool useTopologicalRangeQuery, bool writeToFlagsOldIndex, bool onlyWriteFlagsToDelete)
{
   if (pc == NULL) return;

   // TODO: check if point cloud is hierarchical

   if ( !pc->providesAttribute( "position" ) ){
      debugOutput << "Resampler: point cloud does not provide attribute 'position'\n"; debugOutput.flush();
      return;
   }

   if ( !pc->providesAttribute( "flags" ) ){
      debugOutput << "Resampler: point cloud does not provide attribute 'flags'\n"; debugOutput.flush();
      return;
   }

   if ( !pc->providesAttribute( "curvature" ) ){
      debugOutput << "Resampler: point cloud does not provide attribute 'curvature'\n"; debugOutput.flush();
      return;
   }

   if( pc->getDescr()->getAttribute( "curvature" ).numComponents != 4)
   {
      debugOutput << "Resampler: point cloud does provide 'curvature' attribute with wrong element count, must be 4\n"; debugOutput.flush();
      return;   
   }
   
   // copy point cloud
   UnstructuredInCorePointCloud* pcCopy = copyToUnstructuredInCorePointCloud(pc);
   AAT POSITION = pcCopy->getAAT( "position" );
   AAT CURVATURE = pcCopy->getAAT( "curvature" );
   AAT FLAGS = pcCopy->getAAT( "flags" );

   // get topology
   InCorePCTopologyGraph* tpg = NULL;
   InCorePCTopologyGraph* tpgCopy = NULL;
   if (pc->getAttachments()->getData(TopologyAttachment::getDefaultName()))
   {
      TopologyAttachment* tatt = dynamic_cast<TopologyAttachment*>(pc->getAttachments()->getData(TopologyAttachment::getDefaultName()));
      tpg = tatt->getTopology();

      // copy the given point cloud for later use
      if (mTopologyReconnectMaxRadius < 0.000001)
         mTopologyReconnectMaxRadius = maxSpacing * 2.0;
         
      tpgCopy = tpg->copy();
      tatt = new TopologyAttachment;
      tatt->setup(TopologyAttachment::getDefaultName(), AttachedData::ADF_PERSISTENT);
      tatt->setTopology(tpgCopy);            
      pcCopy->getAttachments()->attachData(tatt);         
   }
   

   // Create iterators
   SceneObjectIterator* it = pcCopy->createIterator( SOIT( SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC | SOIT::CAP_DYNAMIC_PC | SOIT::CAP_INDEXED_PC) );
   BasicPointCloudIterator* basicIt = dynamic_cast<BasicPointCloudIterator*>( it );
   ModifyablePointCloudIterator* modIt = dynamic_cast<ModifyablePointCloudIterator*>( it );
   DynamicPointCloudIterator* dynIt = dynamic_cast<DynamicPointCloudIterator*>( it );
   IndexedPointCloudIterator* indIt = dynamic_cast<IndexedPointCloudIterator*>(it);
   
   if ( basicIt == NULL ){ 
      debugOutput << "Cannot create basic iterator.\n"; debugOutput.flush();
      return;
   }

   if ( modIt == NULL ){ 
      debugOutput << "Cannot create modifyable iterator.\n"; debugOutput.flush();
      return;
   }

   if ( dynIt == NULL ){ 
      debugOutput << "Cannot create dynamic iterator.\n"; debugOutput.flush();
      return;
   }

   // Reset all flags and get max/min curvature
   float minCurvature = FLT_MAX;
   float maxCurvature = 0.0f;
   debugOutput << "Resampler: resetting flags ...\n"; debugOutput.flush();
   progressWindow->pushStep( true, "Reset" );
   basicIt->reset();
   unsigned long pointIndex = 0;
   while( !basicIt->atEnd() )
   {
      const Vector3f& p = basicIt->get3f( POSITION );
      if ( _isnan( p[0] ) || _isnan( p[1] ) || _isnan( p[2] ) ){
         debugOutput << "Resampler: point cloud contains NAN point ... deleting point\n"; debugOutput.flush();
         dynIt->deleteCurrent();
      }
      if ( !_finite( p[0] ) || !_finite( p[1] ) || !_finite( p[2] ) ){
         debugOutput << "Resampler: point cloud contains INF point ... deleting point\n"; debugOutput.flush();
         dynIt->deleteCurrent();
      }

      float attr = abs(basicIt->get4f( CURVATURE )[0]) + abs(basicIt->get4f( CURVATURE )[1]);

      if (attr < minCurvature) minCurvature = attr;
      if (attr > maxCurvature) maxCurvature = attr;

      modIt->set1i( FLAGS, 0 );
      basicIt->next();
      pointIndex++;
      progressWindow->progress( float32( pointIndex ) / float32( pcCopy->getNumPoints() ) * 100 );
   }
   bool aborted;
   progressWindow->popSteps( &aborted );

   card32 numPointStart = pc->getNumPoints();
   card32 numNewPoints = numPointStart;

   ModifyableTopologicalKNNIterator* topIt = NULL;
   PointCloudIterator* tpgIter = NULL;
   if (useTopologicalRangeQuery && tpgCopy)
   {
	   tpgIter = tpgCopy->createIterator(pcCopy->getPointSet());
		topIt = new ModifyableTopologicalKNNIterator(pcCopy, tpgIter);		
   }

	ModifyableKNNIterator knnIt( pcCopy, 32, NULL );
	knnIt.reset();

   if (topIt) topIt->reset();

   // Mark all delete points
   debugOutput << "Resampler: tagging points (minCurv=" << minCurvature <<", maxCurv=" << maxCurvature << ") ...\n"; debugOutput.flush();
   progressWindow->pushStep( true, "Tagging" );
   //float32 sqrSpacing = sqr( spacing );
   basicIt->reset();
   pointIndex = 0;
   while( !basicIt->atEnd() )
   {
      const Vector3f& p = basicIt->get3f( POSITION );
      const int32 flag = basicIt->get1i( FLAGS );
      
      // point was not visited before
      if (flag == 0)
      {      
         // mark point as visited
         modIt->set1i(FLAGS, 2);

         //get curvature
         float curv =  abs(basicIt->get4f( CURVATURE )[0]) + abs(basicIt->get4f( CURVATURE )[1]);

         // compute spacing based on curvature
         //float32 factor = pow((curv - minCurvature) / (maxCurvature - minCurvature), 0.5f)
         float32 factor = (curv - minCurvature) / (maxCurvature - minCurvature);
         float32 sqrSpacing = sqr( minSpacing * factor + (1.0 - factor) * maxSpacing );


         // find the point and get to the nearest neighbour
		 knnIt.setSeekPointAndReset( p );
		 bool atEnd = false;
		 if (topIt)
		 {
			 topIt->setStartPointAndReset(knnIt.getCurrentPointIndex());
			 topIt->next();
			 atEnd = topIt->atEnd();
		 }else
		 {
			 knnIt.next();			// Skip self 
			 atEnd = knnIt.atEnd();
		 }

         if (!atEnd)
         {
             float32 sqrD = 0;         
             do 
             {
				 atEnd = topIt ? topIt->atEnd() : knnIt.atEnd();
				 int flag = 0;
				 if (topIt) flag = topIt->get1i(FLAGS);
				 else flag = knnIt.get1i(FLAGS);

                if (flag == 0) 
                {            
					const Vector3f& np = topIt ? topIt->get3f(POSITION) : knnIt.get3f( POSITION );
                   sqrD = ( p - np ) * ( p - np );
                   if (sqrD < sqrSpacing)
                   {
					   if (topIt) topIt->set1i(FLAGS, 1);
					   else knnIt.set1i( FLAGS, 1 );
                      numNewPoints--;
                   }
                }
				if (topIt) topIt->next();
				else knnIt.next();
             }while( sqrD < sqrSpacing && !atEnd);
         }
      }
      basicIt->next();
      pointIndex++;
      progressWindow->progress( float32( pointIndex ) / float32( pcCopy->getNumPoints() ) * 100 );
   }
   progressWindow->popSteps( &aborted );
   if (topIt)
   {
	   delete topIt;
	   //delete tpgIter;
   }

   x4Assert(numPointStart == pcCopy->getNumPoints());

   debugOutput << "Resampled  from " << numPointStart << " points to " << card32( numNewPoints ) << " points (minSpacing=" << minSpacing << ", maxSpacing=" << maxSpacing << ")\n"; debugOutput.flush();

   if (onlyWriteFlagsToDelete)
   {
	   for (unsigned i=0; i < pcCopy->getNumPoints(); i++)
	   {
		   if (pcCopy->getPointSet()->get1i(i, FLAGS) != 1)
			   pc->getPointSet()->set1i(i, FLAGS, 0);
		   else
			   pc->getPointSet()->set1i(i, FLAGS, 1);
	   }
	   delete pcCopy;
		delete it;
		return;
   }

   // -1 means point has been removed
   std::vector<int32> newToOldIndex(numNewPoints, -1);
   std::vector<int32> oldToNewIndex(numPointStart, -1);

   // setup indices correctly and copy points
   progressWindow->pushStep( true, "Deleting" );
   //debugOutput << "Resampler: deleting points upto" << numNewPoints << " ...\n"; debugOutput.flush();
   card32 numDeleted = 0;
   pc->clearAndSetup(pcCopy->getDescr(), numNewPoints);
   for (unsigned i=0, newindex=0; i < pcCopy->getNumPoints(); i++)
   {
      if (pcCopy->getPointSet()->get1i(i, FLAGS) != 1)
      {
         newToOldIndex[newindex] = i;
         oldToNewIndex[i] = newindex;
         pc->getPointSet()->setVertex(newindex, pcCopy->getPointSet()->getDataPointer(i));
         if (writeToFlagsOldIndex) pc->getPointSet()->set1i(newindex, FLAGS, i);
         newindex++;
      }else
         numDeleted++;
      progressWindow->progress( 100.0 * float32( i ) / float32( numPointStart ));
   }
   progressWindow->popSteps( &aborted );
   
   
   x4Assert(numNewPoints == pc->getNumPoints());

   //debugOutput << "deleted: " << numDeleted << " points\n";
   //debugOutput << "Resampled  from " << numPointStart << " points to " << card32( numNewPoints ) << " points (minSpacing=" << minSpacing << ", maxSpacing=" << maxSpacing << ")\n"; debugOutput.flush();

   delete it;
   it = NULL;

   // create new topology graph
   if (tpg)
   {
      progressWindow->pushStep( true, "Recreate topology" );
      debugOutput << "Resampler: recreate topology graph...\n"; debugOutput.flush();
      
      // this would be our new topology graph
      InCorePCTopologyGraph *resultTg = new InCorePCTopologyGraph();
      resultTg->clearAndSetup(tpg->getEdges()->getDescr());

      // use this iterator to go through the topology graph
      PointCloudIterator* titer = tpg->createIterator(pcCopy->getPointSet());
      TopologicalKNNIterator knn(pcCopy, titer);

      // for each new we have to establishe new topological connection based on previous one
      for (unsigned i=0; i < pc->getNumPoints(); i++)
      {
         int32 searchCenter = newToOldIndex[i];
         x4Assert(searchCenter >= 0);
         Vector3f searchCenterPos = pcCopy->getPointSet()->get3f(searchCenter, POSITION);
         knn.setStartPointAndReset(searchCenter);
         while(!knn.atEnd())
         {
            card32 currentOldPoint = knn.getPointSetPointNumber();
            if (currentOldPoint != searchCenter)
            {
               Vector3f currentOldPointPos = pcCopy->getPointSet()->get3f(currentOldPoint, POSITION);
               float32 distance = norm(currentOldPointPos-searchCenterPos);
               if (distance > mTopologyReconnectMaxRadius)
                  break;
                  
               if (oldToNewIndex[currentOldPoint] != -1)
                  resultTg->addEdge(i, oldToNewIndex[currentOldPoint]);
            }
            knn.next();
         }
         progressWindow->progress( 100.0 * float32( i ) / float32( pc->getNumPoints()));
      }
      progressWindow->popSteps( &aborted );

      // attach new topology graph
      TopologyAttachment* tatt = NULL;
      //if (pc->getAttachments()->getData(TopologyAttachment::getDefaultName()))
      //{
      //   tatt = dynamic_cast<TopologyAttachment*>(pc->getAttachments()->getData(TopologyAttachment::getDefaultName()));
      //}else
      //{
         tatt = new TopologyAttachment;
         tatt->setup(TopologyAttachment::getDefaultName(), AttachedData::ADF_PERSISTENT);
         tatt->setTopology(resultTg);            
         pc->getAttachments()->attachData(tatt);	
      //}
   }

   delete pcCopy;
}

void PCCResampleByCurvature::execute( Scene *scene )
{
   std::vector<std::string> selection = scene->getRootState()->staticState->getSelectedNodes();
   for ( unsigned int m = 0; m < selection.size(); m++ ){

      // Get point cloud
      UnstructuredInCorePointCloud* pc = dynamic_cast<UnstructuredInCorePointCloud*>(getPointCloud( scene, selection[m] ));
      resamplePC(pc);
   }

}
