//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PCCSimplification.h"
//---------------------------------------------------------------------------
#include "BooleanClassProperty.h"
#include "NumericalClassProperty.h"
#include "DebugOutput.h"
#include "PointCloudIterators.h"
#include "PointSet.h"
#include "PointSet.inline.h"
#include "UnstructuredInCorePointCloud.h"
#include "SeparatorClassProperty.h"
#include "BooleanClassProperty.h"
#include "HierarchicalKNNIterator.h"
#include "InCorePCTopologyGraph.h"
#include "TopologicalKNNIterator.h"
#include "TopologyAttachment.h"
#include "PointCloudTools.h"
#include "ClassMethods.h"
#include "StringClassProperty.h"
#include "SGObjectNode.h"
#include "SceneGraphState.h"
#include "ColorTools.h"
#include <set>
#include "UnstructuredInCoreTriangleMesh.h"
#include "ProgressWindow.h"
#include <opencv\cxcore.h>

//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

IMPLEMENT_X4_CLASS( PCCSimplification, 0)
{
	BEGIN_CLASS_INIT( PCCSimplification );
	
	ADD_SEPARATOR("Settings");
	ADD_CARD32_PROP(knn, 0);
	ADD_BOOLEAN_PROP(useTopology, 0);
	ADD_STRING_PROP(variationChannel, 0);
	ADD_BOOLEAN_PROP(normalizeVariation, 0);
	//ADD_BOOLEAN_PROP(estimateNormals, 0);

	ADD_SEPARATOR("Iterative Simplification");
	ADD_FLOAT32_PROP(contractionPairEpsilonFactor, 0);
	ADD_CARD32_PROP(targetNumPoints, 0);
	ADD_FLOAT64_PROP(maxError, 0);

	ADD_NOARGS_METHOD_DESC(covarianceAnalysis, "Perform covariance analysis and store in variation channel [0;1]. The value can be used to check which part are more important for surface description.");
	ADD_NOARGS_METHOD_DESC(iterativeSimplification, "Run iterative simplification (similar to edge-contraction) until desired amount of points is reached.");
	ADD_NOARGS_METHOD_DESC(showVariationInBlueRed, "Copy variation into color channel");
}

//---------------------------------------------------------------------------
PCCSimplification::PCCSimplification()
{
	knn = 20;
	useTopology = true;
	variationChannel = "variation";
	normalizeVariation = true;
	contractionPairEpsilonFactor = 2.0;
	targetNumPoints = 1000;
	maxError = -1;
	estimateNormals = false;
}

//---------------------------------------------------------------------------
PCCSimplification::~PCCSimplification()
{

}

//---------------------------------------------------------------------------

std::vector<PointCloud*> PCCSimplification::getSelectedPointClouds()
{	
	std::vector<PointCloud*> ret;
	const std::vector<std::string> selection = getScene()->getRootState()->staticState->getSelectedNodes();
	for( card32 i=0;i<selection.size();i++ )
	{
		PointCloud * pc = getPointCloud( getScene(), selection[i] );
		if( pc )
			ret.push_back(pc);
	}

	return ret;
}

//---------------------------------------------------------------------------
void PCCSimplification::covarianceAnalysis()
{
	std::vector<PointCloud*> pcs = getSelectedPointClouds();

	for (unsigned i=0; i < pcs.size(); i++)
		covarianceAnalysisImpl(pcs[i], variationChannel);
}

//---------------------------------------------------------------------------
void PCCSimplification::showVariationInBlueRed()
{
	std::vector<PointCloud*> pcs = getSelectedPointClouds();

	for (unsigned i=0; i < pcs.size(); i++)
	{
		// Create iterators
		SceneObjectIterator* it = pcs[i]->createIterator( SOIT( SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC | SOIT::CAP_DYNAMIC_PC | SOIT::CAP_INDEXED_PC) );
		BasicPointCloudIterator* basicIt = dynamic_cast<BasicPointCloudIterator*>( it );
		ModifyablePointCloudIterator* modIt = dynamic_cast<ModifyablePointCloudIterator*>( it );

		AAT colorAAT = pcs[i]->getAAT("color");
		AAT varAAT = pcs[i]->getAAT(variationChannel);

		while(!basicIt->atEnd())
		{
			float var = basicIt->get1f(varAAT);
			Vector3f col = rainbowGradient3f(var);
			modIt->set3f(colorAAT, col);

			basicIt->next();
		}

		delete it;
	}
}

//---------------------------------------------------------------------------
void PCCSimplification::covarianceAnalysisImpl(PointCloud* pc, const std::string& channelName /*= "variation"*/)
{
	if (pc == NULL) return;

	UnstructuredInCorePointCloud* upc = dynamic_cast<UnstructuredInCorePointCloud*>(pc);
	if (upc == NULL)
	{
		error(std::string(__FUNCTION__) + " - currently only implemented for UnstructoredInCorePointClouds");
		return;
	}

	// check that all required fileds are provided
	if ( !pc->providesAttribute( "position" ) ){
		debugOutput << __FUNCTION__ << " - point cloud does not provide attribute 'position'\n"; debugOutput.flush();
		return;
	}

	checkAttribute(pc, channelName, 1, VAD::DATA_FORMAT_FLOAT32);

	
	// check if we have topology information available
	TopologicalKNNIterator* tIt = NULL;
	PointCloudIterator* pIt = NULL;
	if (useTopology && pc->getAttachments()->getData(TopologyAttachment::getDefaultName()))
	{
		TopologyAttachment* tatt = dynamic_cast<TopologyAttachment*>(pc->getAttachments()->getData(TopologyAttachment::getDefaultName()));
		InCorePCTopologyGraph* tpg = tatt->getTopology();
		if (tpg)
		{
			pIt = tpg->createIterator(upc->getPointSet());
			tIt = new TopologicalKNNIterator(pc, pIt);
		}
	}

	// create hierarchical point cloud iterator if topological fails
	HierarchicalKNNIterator* hIt = NULL;
	if (tIt == NULL) hIt = new HierarchicalKNNIterator(pc, 32, NULL);

	AAT variationAAT = pc->getAAT("variation");
	AAT posAAT = pc->getAAT("position");
	PointSet* ps = upc->getPointSet();

	float maxVar = 0;
	float minVar = FLT_MAX;

	// for every point reset values and compute variation value
	for (unsigned p=0; p < upc->getNumPoints(); p++)
	{
		ps->set1f(p, variationAAT, 0);

		Vector3f pos = ps->get3f(p, posAAT);

		// get neighbourhood
		std::vector<Vector3f> N;
		N.reserve(knn);
		if (tIt) { tIt->setStartPointAndReset(p); tIt->next(); }
		if (hIt) { hIt->setSeekPointAndReset(pos); hIt->next(); }
		while ( ((tIt && !tIt->atEnd()) || (hIt && !hIt->atEnd())) && N.size() < knn )
		{
			if (tIt) { N.push_back(ps->get3f(tIt->getPointSetPointNumber(), posAAT)); tIt->next(); }
			if (hIt) { N.push_back(ps->get3f(hIt->getPointSetPointNumber(), posAAT)); hIt->next(); }
		}

		// compute covariance matrix and variation
		/*DynamicMatrix<float> M;
		M.setDimension(3, N.size());
		for (unsigned j=0; j < N.size(); j++)
		{
			M[0][j] = N[j][0] - pos[0];
			M[1][j] = N[j][1] - pos[1];
			M[2][j] = N[j][2] - pos[2];
		}
		DynamicMatrix<float> C = M.multTransposed(M);*/

		// compute covariance matrix
		CvMat* cov = cvCreateMat(3, 3,CV_32FC1);
		for (unsigned i=0; i < 3; i++)
		{
			for (unsigned j=0; j < 3; j++)
			{
				float sum = 0;
				for (unsigned k=0; k < N.size(); k++)
					sum += (N[k][i] - pos[i]) * (N[k][j] - pos[j]);
				cvmSet(cov, i, j, sum);
			}
		}

		CvMat* U  = cvCreateMat(3, 3, CV_32FC1);
		CvMat* D  = cvCreateMat(3, 1, CV_32FC1);
		CvMat* V  = cvCreateMat(3, 3, CV_32FC1);

		cvSVD(cov, D, U, V, CV_SVD_U_T|CV_SVD_V_T|CV_SVD_MODIFY_A); // A = U D V^T

		// sort lambda values
		float lambda[3] = {cvmGet(D, 0, 0), cvmGet(D, 1, 0), cvmGet(D, 2, 0)};
		if (lambda[0] > lambda[1]) { float t=lambda[0]; lambda[0] = lambda[1]; lambda[1] = t; }
		if (lambda[1] > lambda[2]) { float t=lambda[1]; lambda[1] = lambda[2]; lambda[2] = t; }
		if (lambda[0] > lambda[1]) { float t=lambda[0]; lambda[0] = lambda[1]; lambda[1] = t; }
		if (lambda[1] > lambda[2]) { float t=lambda[1]; lambda[1] = lambda[2]; lambda[2] = t; }

		cvReleaseMat(&U);
		cvReleaseMat(&D);
		cvReleaseMat(&V);
		cvReleaseMat(&cov);
		
		// compute variation value
		float var = lambda[0] / (lambda[0] + lambda[1] + lambda[2]);
		ps->set1f(p, variationAAT, var);

		maxVar = std::max<float>(var, maxVar);
		minVar = std::min<float>(var, minVar);
	}

	if (normalizeVariation)
	{
		for (unsigned p=0; p < upc->getNumPoints(); p++)
		{
			float var = ps->get1f(p, variationAAT);
			ps->set1f(p, variationAAT, (var - minVar) / (maxVar - minVar));
		}
	}

	if (tIt) {delete tIt; delete pIt;}
	if (hIt) delete hIt;
}

//---------------------------------------------------------------------------
void PCCSimplification::iterativeSimplification()
{
	std::vector<PointCloud*> pcs = getSelectedPointClouds();

	for (unsigned i=0; i < pcs.size(); i++)
		iterativeSimplificationImpl(dynamic_cast<UnstructuredInCorePointCloud*>(pcs[i]));
}

//---------------------------------------------------------------------------
void PCCSimplification::computeFundamentalQuadrics(UnstructuredInCorePointCloud* pc, HierarchicalKNNIterator* hIt, float medDist, std::vector<FundamentalQuadric*>* quadrics)
{
	if (quadrics == NULL) return;
	if (useTopology && hIt == NULL)
	{
		if (pc->getAttachments()->getData(TopologyAttachment::getDefaultName()) == NULL)
		{
			error(std::string(__FUNCTION__) + " - given mesh must have a topology graph attached or you have to provide hIt");
			return;
		}
	}
	InCorePCTopologyGraph* tpg = NULL;
	if (useTopology && pc->getAttachments()->getData(TopologyAttachment::getDefaultName()) != NULL)
		tpg = dynamic_cast<TopologyAttachment*>(pc->getAttachments()->getData(TopologyAttachment::getDefaultName()))->getTopology();

	PointSet* ps = pc->getPointSet();

	AAT posAAT = ps->getAAT("position");
	AAT flagsAAT = ps->getAAT("flags");
	AAT normalAAT = ps->getAAT("normal");

	quadrics->clear();
	quadrics->resize(ps->getNumEntries(), NULL);

	for (unsigned p=0; p < ps->getNumEntries(); p++)
	{
		if (ps->get1i(p, flagsAAT) == 1) continue;
		if (ps->get1i(p, flagsAAT) != 2) continue;
		ps->set1i(p, flagsAAT, 0);//ps->get1i(p, flagsAAT) & (~2));

		// we work on this points
		Vector3f pos = ps->get3f(p, posAAT);
		quadrics->at(p) = new FundamentalQuadric();

		// get full neighborhood (incl. knn neighbors, if hIt specified)
		std::set<PointIndex> N;
		if (tpg)
		{
			for (unsigned n=0; n < tpg->getNumAdjacentVertices(p) && N.size() < knn; n++)
			{
				PointIndex pi = tpg->getAdjacentVertex(p, n);
				if (ps->get1i(pi, flagsAAT) != 1)
					N.insert(pi);
			}
		}

		if (hIt)
		{
			hIt->setSeekPointAndReset(pos); 
			hIt->setMaxDistanceToSeekPoint(medDist);
			hIt->next();
			while ((!hIt->atEnd()) && (N.size() < knn) )
			{
				if (ps->get1i(hIt->getPointSetPointNumber(), flagsAAT) != 1)
					N.insert(hIt->getPointSetPointNumber());
				hIt->next();
			}
		}

		// compute error quadric as a sum of all quadrics describing for local planes
		for (std::set<PointIndex>::const_iterator it=N.begin(); it != N.end(); it++)
		{
			Vector3f norm = ps->get3f(*it, normalAAT); norm.normalize();
			Vector3f e = pos - ps->get3f(*it, posAAT);
			Vector3f b = e.crossProduct(norm);
			Vector3f n = e.crossProduct(b); n.normalize();
			Vector4f plane = makeVector4f(n[0], n[1], n[2], - (n * pos));

			(*quadrics->at(p)) += FundamentalQuadric(plane);
		}
	}

}

//---------------------------------------------------------------------------
void PCCSimplification::iterativeSimplificationImpl(UnstructuredInCorePointCloud* tri, bool onlyChangeFlag)
{
	if (tri == NULL) return;

	// check that all required fileds are provided
	if ( !tri->providesAttribute( "position" ) ){
		debugOutput << __FUNCTION__ << " - point cloud does not provide attribute 'position'\n"; debugOutput.flush();
		return;
	}

	if ( !tri->providesAttribute( "flags" ) ){
		debugOutput << __FUNCTION__ << " - point cloud does not provide attribute 'flags'\n"; debugOutput.flush();
		return;
	}

	if ( !tri->providesAttribute( "normal" ) ){
		debugOutput << __FUNCTION__ << " - point cloud does not provide attribute 'normal'\n"; debugOutput.flush();
		return;
	}

	InCorePCTopologyGraph* tpg = NULL;
	if (useTopology && tri->getAttachments()->getData(TopologyAttachment::getDefaultName()) != NULL)
		tpg = dynamic_cast<TopologyAttachment*>(tri->getAttachments()->getData(TopologyAttachment::getDefaultName()))->getTopology();

	// get epsilon and compute distance for valid pairs
	float medDist = getCachedMedianPointDistance(tri);
	float epsDist = medDist * contractionPairEpsilonFactor;

	// create hierarchical point cloud iterator if topological fails
	HierarchicalKNNIterator* hIt = NULL;
	if (contractionPairEpsilonFactor > 1e-5) hIt = new HierarchicalKNNIterator(tri, 32, NULL);

	AAT posAAT = tri->getAAT("position");
	AAT flagsAAT = tri->getAAT("flags");
	AAT normalAAT = tri->getAAT("normal");

	PointSet* ps = tri->getPointSet();

	// compute all error quadrics
	std::vector<FundamentalQuadric*> quadrics;
	for (unsigned p=0; p < tri->getNumPoints(); p++)
		ps->set1i(p, flagsAAT, 2);
	computeFundamentalQuadrics(tri, hIt, epsDist, &quadrics);

	// create a sorted list of all contraction pairs
	// two points are in a pair if they are either adjacent or their distance is below given threshold
	std::list<DecimationPair> pairList;
	for (unsigned i=0; i < ps->getNumEntries(); i++)
	{
		Vector3f pos1 = ps->get3f(i, posAAT);
		for (unsigned j=i+1; j < ps->getNumEntries(); j++)
		{
			Vector3f pos2 = ps->get3f(j, posAAT);

			float sqrDist = (pos1 - pos2).getSqrNorm();
			bool validPair = sqrDist < (epsDist * epsDist);
			if (tpg && tpg->areVerticesAdjacent(i,j)) validPair = true;

			if (validPair)
				pairList.push_back(DecimationPair(pos1,pos2,i,j,*quadrics[i],*quadrics[j]));
		}
	}
	pairList.sort();
	for (unsigned i=0; i < quadrics.size(); i++) if (quadrics[i]) delete quadrics[i];

	// decimate points
	debugOutput << "Simplificate: " << ps->getNumEntries() << " -> " << targetNumPoints << " points\n";
	int targetIt = ps->getNumEntries()-targetNumPoints;
	double lastError = 0;
	progressWindow->pushStep(true, "Iterative Simplification");
	for (int k=0; k < targetIt && pairList.size(); k++)
	{
		progressWindow->progressf(float(k) / float(ps->getNumEntries()-targetNumPoints));

		// get a pair which is most far away to the next best
		DecimationPair pair;
		{
			std::list<DecimationPair>::iterator it = pairList.begin();

			if (maxError > 0 && it->error > maxError) break;

			pair = *it;
			pairList.erase(it);
		}
		lastError = pair.error;

		// replace the first element of the pair by the computed best element
		ps->set3f(pair.v.first, posAAT, pair.vhat);

		// remove the v2 out of the point list and topology graph
		//tri->getPointSet()->deletePoint(pair.v.second);
		ps->set1i(pair.v.second, flagsAAT, 1);
		/*if (tpg)
		{
			for (unsigned n=0; n < tpg->getNumAdjacentVertices(pair.v.second); n++)
				tpg->addEdge(pair.v.first, tpg->getAdjacentVertex(pair.v.second, n));

			tpg->deleteVertex(pair.v.second);
		}*/
		if (hIt) // TODO: check if HierarchicalKNNIterator can be reused when position changed
		{
			delete hIt;
			hIt = new HierarchicalKNNIterator(tri, 32, NULL);
		}

		// now we remove all pairs of the pair list, which contains v2, this are copied in a temporary list (replace v2 with v1)
		std::list<DecimationPair> templist;
		for (std::list<DecimationPair>::iterator jt=pairList.begin(); jt != pairList.end(); )
		{
			if ((jt->v.second == pair.v.second) || (jt->v.first == pair.v.second))
			{
				DecimationPair listPair = *jt;

				if (jt->v.second == pair.v.second) listPair.v.second = pair.v.first;
				else listPair.v.first = pair.v.first;

				if (std::find(pairList.begin(), pairList.end(), listPair) == pairList.end())
				{
					if (listPair.v.first == listPair.v.second)
					{
						int k =0;
					}

					// if there is no duplicate, then add this one
					if (std::find(templist.begin(), templist.end(), listPair) == templist.end())
						templist.push_back(listPair);
				}

				jt = pairList.erase(jt);
			}else 
				jt++;
		}
		
		// any of the remaining pair which contains one of the extracted vertices, has to got its error updated too
		/*for (std::list<DecimationPair>::iterator jt=pairList.begin(); jt != pairList.end(); )
		{
			// check if this pair contains a vertex which belong to this temporary list
			bool found = false;
			for (std::list<DecimationPair>::iterator kt=templist.begin(); kt != templist.end(); kt++)
			{
				if ((jt->v.first == kt->v.first) || (jt->v.first == kt->v.second)
				||  (jt->v.second == kt->v.first) || (jt->v.second == kt->v.second))
				{
					found = true;
					break;
				}
			}
			if (found)
			{
				templist.push_back(*jt);
				jt = pairList.erase(jt);
			}else
				jt++;
		}*/

		// update quadrics for all vertices in the temporary list
		for (std::list<DecimationPair>::iterator jt=templist.begin(); jt != templist.end(); jt++)
		{
			x4Assert(ps->get1i(jt->v.first, flagsAAT) != 1 && ps->get1i(jt->v.second, flagsAAT) != 1);
			ps->set1i(jt->v.first, flagsAAT, 2);
			ps->set1i(jt->v.second, flagsAAT, 2);
		}
		std::vector<FundamentalQuadric*> _quadrics;
		computeFundamentalQuadrics(tri, hIt, epsDist, &_quadrics);

		// update all pairs in the temporary buffer
		for (std::list<DecimationPair>::iterator jt=templist.begin(); jt != templist.end(); jt++)
		{
			x4Assert(_quadrics[jt->v.first] && _quadrics[jt->v.second]);
			jt->update(ps->get3f(jt->v.first, posAAT), ps->get3f(jt->v.second, posAAT), (*_quadrics[jt->v.first]), (*_quadrics[jt->v.second]));
		}
		templist.sort();
		for (unsigned i=0; i < _quadrics.size(); i++) if (_quadrics[i]) delete _quadrics[i];

		// now we add back the decimation pairs from the temporary list
		// hereby we add them in a sorted way and check for duplicates
		std::list<DecimationPair>::iterator jt = templist.begin();
		for (std::list<DecimationPair>::iterator it = pairList.begin(); it != pairList.end() && jt != templist.end(); it++)
		{
			for (std::list<DecimationPair>::iterator jt=templist.begin(); jt != templist.end();)
			{
				// sorted add
				if ((*jt) < (*it))
				{
					pairList.insert(it, *jt);
					jt = templist.erase(jt);
					continue;
				}else break;
			}
		}
		for (std::list<DecimationPair>::iterator jt=templist.begin(); jt != templist.end();)
		{
			pairList.push_back(*jt);
			jt = templist.erase(jt);
		}		
	}
	progressWindow->popSteps();

	if (onlyChangeFlag == false)
	{
		std::vector<PointIndex> toDelete;
		for (unsigned i=0; i < ps->getNumEntries(); i++)
			if (ps->get1i(i, flagsAAT) == 1)
				toDelete.push_back(i);
		std::sort(toDelete.begin(), toDelete.end());
		for (std::vector<PointIndex>::reverse_iterator it = toDelete.rbegin(); it != toDelete.rend(); it++)
			ps->deletePoint(*it);
		for (unsigned p=0; p < ps->getNumEntries(); p++)
			ps->set1i(p, flagsAAT, 0);
	}

	debugOutput << "Simplification done: lastError = " << lastError << "\n";

	// free data
	if (hIt) delete hIt;
}

//---------------------------------------------------------------------------
void PCCSimplification::execute( Scene *scene )
{

}

