//---------------------------------------------------------------------------
#ifndef PCCSimplificationH
#define PCCSimplificationH
//---------------------------------------------------------------------------
#include "x3foundation_private.h"
//---------------------------------------------------------------------------
#include "PCCommand.h"
#include "PointCloud.h"
#include "UnstructuredInCorePointCloud.h"
#include "Scene.h"
#include <float.h>

//---------------------------------------------------------------------------

namespace X4 {

	class TopologicalKNNIterator;
	class HierarchicalKNNIterator;
	class UnstructuredInCoreTriangleMesh;

	/**
	* Point cloud simplification algorithms implemented by the Mark Paulys Paper: "Efficient Simplification of Point-Sampled Surfaces"
	*/
	class X3FOUNDATION_PRIVATE_API PCCSimplification : public PCCommand 
	{
		X4_CLASS( PCCSimplification )

	protected:
		card32 knn;
		bool useTopology;
		std::string variationChannel;
		bool normalizeVariation;
		float contractionPairEpsilonFactor;
		card32 targetNumPoints;

		bool estimateNormals;
		std::vector<PointCloud*> getSelectedPointClouds();
		double maxError;

	public:
		PCCSimplification();
		virtual ~PCCSimplification();

		/// Compute curvature distribution based on covariance analysis
		void covarianceAnalysisImpl(PointCloud* pc, const std::string& channelName = "variation");
		void iterativeSimplificationImpl(UnstructuredInCorePointCloud* tri, bool onlyChangeFlag = false);

		/// Apply to scene
		void execute( Scene *scene );
		void covarianceAnalysis();
		void iterativeSimplification();
		void showVariationInBlueRed();

	protected:
		// first we compute quadrics describing local planes in a neighbourhood of each point
		struct FundamentalQuadric
		{
			double v[10];
			FundamentalQuadric(const FundamentalQuadric& f)
			{
				memcpy(&v[0], &f.v[0], sizeof(double) * 10);
			}
			FundamentalQuadric(Vector4f plane)
			{
				v[0] = plane[0] * plane[0];
				v[1] = plane[0] * plane[1];
				v[2] = plane[0] * plane[2];
				v[3] = plane[0] * plane[3];

				v[4] = plane[1] * plane[1];
				v[5] = plane[1] * plane[2];
				v[6] = plane[1] * plane[3];

				v[7] = plane[2] * plane[2];
				v[8] = plane[2] * plane[3];

				v[9] = plane[3] * plane[3];
			}
			FundamentalQuadric()
			{
				memset(&v[0], 0, sizeof(double) * 10);
			}
			inline FundamentalQuadric& operator+=(const FundamentalQuadric& f)
			{
				for (int i=0; i < 10; i++) v[i]+=f.v[i];
				return *this;
			}

			// compute v^TQv
			inline double operator*(const Vector4f& p)
			{
				return p[0] * (v[0]*p[0] + v[1]*p[1] + v[2]*p[2] + v[3]) + 
					   p[1] * (v[1]*p[0] + v[4]*p[1] + v[5]*p[2] + v[6]) +
					   p[2] * (v[2]*p[0] + v[5]*p[1] + v[7]*p[2] + v[8]) +
					    1   * (v[3]*p[0] + v[6]*p[1] + v[8]*p[2] + v[9]);
			}

			// convert to 4x4 matrix
			inline Matrix4d toMatrix() const 
			{
				Matrix4d mat;
				mat[0][0] = v[0];
				mat[1][0] = v[1];
				mat[2][0] = v[2];
				mat[3][0] = v[3];

				mat[0][1] = v[1];
				mat[1][1] = v[4];
				mat[2][1] = v[5];
				mat[3][1] = v[6];

				mat[0][2] = v[2];
				mat[1][2] = v[5];
				mat[2][2] = v[7];
				mat[3][2] = v[8];

				mat[0][3] = v[3];
				mat[1][3] = v[6];
				mat[2][3] = v[8];
				mat[3][3] = v[9];

				return mat;
			}
		};

		struct DecimationPair
		{
			std::pair<PointIndex,PointIndex> v;
			double error;
			Vector3f vhat;
			DecimationPair()
			{
				error = FLT_MAX;
				v = std::pair<PointIndex,PointIndex>(0,0);
			}
			DecimationPair(const DecimationPair& dc)
			{
				v = dc.v;
				error = dc.error;
				vhat = dc.vhat;
			}
			void update(const Vector3f& p1, const Vector3f& p2, const FundamentalQuadric& Q1, const FundamentalQuadric& Q2)
			{
				FundamentalQuadric Q = Q1; Q += Q2;
				Matrix4d invQ;
				bool sucess = true;

				// try to invert the matrix
				try
				{
					Matrix4d qm = Q.toMatrix();
					qm[0][3] = 0;
					qm[1][3] = 0;
					qm[2][3] = 0;
					qm[3][3] = 1;
					invQ = invertMatrix(qm, &sucess);
				}
				catch (...)
				{
					sucess = false;
				}

				// if inversion happened well, then we can compute the solution directly
				if (sucess)
				{
					Vector4d vv = invQ * makeVector4d(0,0,0,1);
					vhat = makeVector3f(vv[0]/vv[3], vv[1]/vv[3], vv[2]/vv[3]);
					error = abs(Q * expand3To4(vhat));
				}else
				{
					// check which of the points has smallest error and use this as vhat
					vhat = (p1 + p2) * 0.5;
					error = abs(Q * expand3To4(vhat));
					double e1 = abs(Q * expand3To4(p1));
					double e2 = abs(Q * expand3To4(p2));
					if (e1 < error){ error = e1; vhat = p1; }
					if (e2 < error){ error = e1; vhat = p2; }
				}
			}
			DecimationPair(const Vector3f& p1, const Vector3f& p2, PointIndex pi1, PointIndex pi2, const FundamentalQuadric& Q1, const FundamentalQuadric& Q2)
			{
				v.first = pi1;
				v.second = pi2;
				update(p1,p2,Q1,Q2);
			}
			bool operator<(const DecimationPair& p) const
			{
				return error < p.error;
			}
			bool operator==(const DecimationPair& p) const 
			{
				return (p.v.first == v.first && p.v.second == v.second) || (p.v.first == v.second && p.v.second == v.first);
			}
		};

		/// compute quadrics for all points from pc.getPointSet which (flagAAT & 2)==2. Can use also knn neighbourhood if hIt is given. Note flag will be removed
		void computeFundamentalQuadrics(UnstructuredInCorePointCloud* pc, HierarchicalKNNIterator* hIt, float medDist , std::vector<FundamentalQuadric*>* quadrics);

	};

}

#endif 
