#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PCCSurfaceReconTools.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "MLSProjector.h"
#include "UnstructuredInCorePointCloud.h"
#include "PointCloudTools.h"
#include "HierarchicalKNNIterator.h"
#include "Scene.h"
#include "SceneGraphState.h"
#include "ObjectClassProperty.h"
#include "ClassMethods.h"
#include "Random.h"
//---------------------------------------------------------------------------
using namespace X4;

IMPLEMENT_X4_CLASS( PCCSurfaceReconTools ,0)
{
	BEGIN_CLASS_INIT( PCCSurfaceReconTools );
	ADD_OBJECT_PROP(m_Projector,0,MLSProjector::getClass(),true);
	ADD_FLOAT32_PROP(m_TargetSampleSpacing,0);
	ADD_NOARGS_METHOD(upsampleAndProject);
}

PCCSurfaceReconTools::PCCSurfaceReconTools(void)
{
	m_Projector = new MLSProjector;
	m_TargetSampleSpacing = 1.0f;
}

PCCSurfaceReconTools::~PCCSurfaceReconTools(void)
{
	delete m_Projector;
}

void PCCSurfaceReconTools::assign(const Object* obj, X4CopyContext *context)
{
	PCCommand::assign(obj,context);
	const PCCSurfaceReconTools * o = dynamic_cast<const PCCSurfaceReconTools*>(obj);
	x4Assert( o != NULL );

	copyObjectProperties(obj,this);
}

void X4::PCCSurfaceReconTools::upsampleAndProject()
{
	if( m_TargetSampleSpacing > 1.0f )
	{
		message("m_TargetSampleSpacing must not be > 1.0 !\n");
		return;
	}

	std::vector<PointCloud*> pcs = getSelectedPointClouds();
	for( unsigned i=0;i<pcs.size();i++ )
	{
		PointCloud * pc = pcs[0];
		UnstructuredInCorePointCloud * ret = new UnstructuredInCorePointCloud;
		VertexDescriptor * vd = createDefaultVertexDescriptor(true,true,true,true);
		ret->clearAndSetup(vd,0);

		float medPointDistance = getMedianPointDistance(pc);
		HierarchicalKNNIterator hIt(pc,32,NULL);

		SceneObjectIterator * iter = pc->createIterator(SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC);
		BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
		AAT posAAT = pc->getAAT("position");
		AAT normalAAT = pc->getAAT("normal");
		AAT dest_normalAAT = ret->getAAT("normal");

		m_Projector->begin(pc);
		unsigned numNewPoints = pow(1.0f / m_TargetSampleSpacing,2.0f);

		while( !bIt->atEnd())
		{
			Vector3f p = bIt->get3f(posAAT);

			if( m_TargetSampleSpacing == 1.0f )
			{
				Vector3f projNormal;
				Vector3f projPos = m_Projector->project(p, &projNormal);
				if( m_Projector->lastProjectionValid())
				{
					unsigned ni = ret->getNumPoints();
					ret->getPointSet()->changeHeight(ni+1);
					ret->getPointSet()->set3f(ni,posAAT,projPos);
					ret->getPointSet()->set3f(ni,dest_normalAAT,projNormal);
				}
			}
			else
			{
				Vector3f n = bIt->get3f(normalAAT);
				Matrix3f tangSys = calcTangentSystem(n);
				for( unsigned i=0;i<numNewPoints;i++ )
				{
					Vector3f tmpP = p + tangSys[0] * (rnd01() - 0.5f) * 2.0f * medPointDistance + tangSys[1] * (rnd01() - 0.5f) * 2.0f * medPointDistance;
					Vector3f projNormal;
					Vector3f projPos = m_Projector->project(tmpP, &projNormal);
					if( !m_Projector->lastProjectionValid())
						continue;

					unsigned ni = ret->getNumPoints();
					ret->getPointSet()->changeHeight(ni+1);
					ret->getPointSet()->set3f(ni,posAAT,projPos);
					ret->getPointSet()->set3f(ni,dest_normalAAT,projNormal);
				}


			}

			bIt->next();

		}

		ret->setMaterialIndex(pc->getMaterialIndex());
		addPointCloud(getScene(),ret,"upsampled");

		delete iter;

	}
}


std::vector<PointCloud *> X4::PCCSurfaceReconTools::getSelectedPointClouds()
{
	std::vector<PointCloud*> ret;
	std::vector<std::string> sel = getScene()->getRootState()->staticState->getSelectedNodes();
	for( unsigned i=0;i<sel.size();i++ )
	{
		PointCloud * pc = dynamic_cast<PointCloud*>(
			getPointCloud(getScene(), sel[i]));
		if( pc )
			ret.push_back( pc );
	}

	return ret;
}