#include "StdAfx.h"
//---------------------------------------------------------------------------
#include <windows.h>
#include <mmsystem.h>
#include "PCCSurfaceTools.h"
//---------------------------------------------------------------------------
#include "StandardSceneGraphIterator.h"
#include "SGObjectNode.h"
#include "PointCloudTools.h"
#include "DebugOutput.h"
#include "ClassMethods.h"
#include "CopyObjectProperties.h"
#include "pointcloud.h"
#include "ProgressWindow.h"
#include "Monomials2ndOrder.h"
#include "Random.h"
#include "ImageDisplayer.h"
#include "RGBAImage.h"
#include "Timer.h"
#include "StringHelper.h"
#include "CachedRangeQuerry.h"
#include "TopologicalKNNIterator.h"
#include "GeometricTools.h"
#include <QtGui/QFileDialog>
#include <QtGUI/QImage>
#include "UnstructuredInCoreTriangleMesh.h"
#include "VertexArray.h"
#include "AttachedIndexedOctree.h"
#include "FastSphereQuerry.h"
#include "DebugRenderer.h"
#include "PointSetANNQuery.h"
#include "ObjectViewsTable.h"
#include "NormalDistribution.h"
#include <xmmintrin.h>
#include "SGListNode.h"
#include "SceneGraphTools.h"
#include "InCorePCTopologyGraph.h"
#include "TopologyAttachment.h"

//---------------------------------------------------------------------------
using namespace X4;

IMPLEMENT_X4_CLASS( PCCSurfaceTools ,2)
{
	BEGIN_CLASS_INIT( PCCSurfaceTools );

	ADD_FLOAT32_PROP( m_KernelSigma, 0);
	ADD_BOOLEAN_PROP( m_UseNormalOrientation,0);
	ADD_FLOAT32_PROP(m_MlsMaxOffset, 2);
	ADD_STRING_PROP( m_DestChannelName, 0);
	ADD_BOOLEAN_PROP( m_ScaleCurvatures,0);
	ADD_FLOAT32_PROP( m_ScaleCurvatureBorder, 0);
	ADD_FLOAT32_PROP( m_CacheRadius, 0 );
	ADD_BOOLEAN_PROP( m_UseMedianFilter,1);
	SET_USER_EDIT(false);
	ADD_BOOLEAN_PROP(m_UseTopogicalRangeQuerry,0);
	

	ADD_NOARGS_METHOD( estimateCurvature );
   SET_METHOD_AS_CMD_OBJ( true )
	ADD_NOARGS_METHOD( calculateAveragePointDistance );
   SET_METHOD_AS_CMD_OBJ( true )
	ADD_NOARGS_METHOD( calculateMedianPointDistance );
   SET_METHOD_AS_CMD_OBJ( true )
	ADD_NOARGS_METHOD( computeCurvatureHistogram );
   SET_METHOD_AS_CMD_OBJ( true )
	ADD_NOARGS_METHOD( colorizePointClouds );
   SET_METHOD_AS_CMD_OBJ( true )
	ADD_NOARGS_METHOD( resample );
   SET_METHOD_AS_CMD_OBJ( true )
	ADD_NOARGS_METHOD( importHeightfield );
   SET_METHOD_AS_CMD_OBJ( true )
	ADD_NOARGS_METHOD( selectBorderPoints );
   SET_METHOD_AS_CMD_OBJ( true )
	ADD_NOARGS_METHOD( colorizeFromOtherPC );
   SET_METHOD_AS_CMD_OBJ( true )
	ADD_NOARGS_METHOD( mlsProjectPC );
   SET_METHOD_AS_CMD_OBJ( true )
	ADD_NOARGS_METHOD( deleteInvalidPoints );
   SET_METHOD_AS_CMD_OBJ( true )
	ADD_NOARGS_METHOD(alignBoundingBoxes);
   SET_METHOD_AS_CMD_OBJ( true )
   ADD_NOARGS_METHOD(eraseBadPoints)
}

PCCSurfaceTools::PCCSurfaceTools(void)
{
	m_KernelSigma = 0.01f;
	m_CacheRadius = 2.7f;
	m_ScaleCurvatureBorder = 0.01f;
	m_UseNormalOrientation = true;
	m_DestChannelName = "curvature";
	m_ScaleCurvatures = false;
	m_UseMedianFilter = false;
	m_UseTopogicalRangeQuerry = false;
	m_MlsMaxOffset = 0.0f;
}

PCCSurfaceTools::~PCCSurfaceTools(void)
{
}

void PCCSurfaceTools::assign(const Object* obj, X4CopyContext *context)
{
	PCCommand::assign(obj,context);
	const PCCSurfaceTools * o = dynamic_cast<const PCCSurfaceTools*>(obj);
	x4Assert( o != NULL );

	copyObjectProperties(obj, this);
}

void PCCSurfaceTools::estimateCurvature()
{

	std::vector<PointCloud*> pcs = getSelectedPointClouds();
	m_Curv1.clear();
	m_Curv2.clear();
	m_Curv3.clear();
	for( card32 i=0;i<pcs.size();i++ )
	{
		estimateCurvature(pcs[i]);
	}

	if( m_ScaleCurvatures && pcs.size() > 0 )
	{
		sort( m_Curv1.begin(), m_Curv1.end());
		sort( m_Curv2.begin(), m_Curv2.end());
		sort( m_Curv3.begin(), m_Curv3.end());

		// consider upper/lower 5% as noise
		Vector4f minCurv = makeVector4f( m_Curv1[ m_ScaleCurvatureBorder * (float)m_Curv1.size() ], m_Curv2[ m_ScaleCurvatureBorder * (float)m_Curv2.size() ],
			m_Curv3[ m_ScaleCurvatureBorder * (float)m_Curv3.size() ],-1.0f );
		Vector4f maxCurv = makeVector4f( m_Curv1[ (1.0f - m_ScaleCurvatureBorder) * (float)m_Curv1.size() ], m_Curv2[ (1.0f - m_ScaleCurvatureBorder) * (float)m_Curv2.size() ], 
			m_Curv3[ (1.0f - m_ScaleCurvatureBorder) * (float)m_Curv3.size() ], 1.0f );
		
		// avoid division by zero
		maxCurv[0] = max( minCurv[0] + 1e-5f,maxCurv[0] );
		maxCurv[1] = max( minCurv[1] + 1e-5f,maxCurv[1] );
		maxCurv[2] = max( minCurv[2] + 1e-5f,maxCurv[2] );

		for( card32 i=0;i<pcs.size();i++ )
		{
			scaleCurvature( pcs[i], minCurv, maxCurv);
		}

	}
}

void PCCSurfaceTools::calculateAveragePointDistance()
{
	std::vector<PointCloud*> pcs = getSelectedPointClouds();
	for( card32 i=0;i<pcs.size();i++ )
	{
		float averagePointDistance = getAveragePointDistance(pcs[i]);
		debugOutput << "PointCloud[" << i << "]: averagePointDistance = " << averagePointDistance << "\n";
	}
}

void PCCSurfaceTools::calculateMedianPointDistance()
{
	std::vector<PointCloud*> pcs = getSelectedPointClouds();
	for( card32 i=0;i<pcs.size();i++ )
	{
		float medianPointDistance = getMedianPointDistance(pcs[i]);
		debugOutput << "PointCloud[" << i << "]: medianPointDistance = " << medianPointDistance << "\n";
	}
}


std::vector<PointCloud*> PCCSurfaceTools::getSelectedPointClouds()
{	
	std::vector<PointCloud*> ret;
	const std::vector<std::string> selection = getScene()->getRootState()->staticState->getSelectedNodes();
	for( card32 i=0;i<selection.size();i++ )
	{
		PointCloud * pc = getPointCloud( getScene(), selection[i] );
		if( pc )
			ret.push_back(pc);
	}

	return ret;
}

struct TDebugKNNPoint
{
	float dist;
	unsigned index;
	bool operator < (const TDebugKNNPoint &p) const {return dist < p.dist;}
};

void PCCSurfaceTools::estimateCurvature( PointCloud * pc)
{

	float cacheRadius = getMedianPointDistance(pc)*m_CacheRadius;

	Timer timer;
	checkAttribute( pc, m_DestChannelName, 4, VAD::DATA_FORMAT_FLOAT32 );
	AAT posAAT = pc->getAAT( "position" );
	AAT normalAAT;
	normalAAT = pc->getAAT( "normal" );
	AAT destAAT = pc->getAAT( m_DestChannelName );

	SceneObjectIterator * iter = pc->createIterator( SceneObjectIteratorType::CAP_BASIC_PC );
	x4Assert( iter != NULL );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(iter);

	//HierarchicalKNNIterator * hIt = new HierarchicalKNNIterator(pc, 32, NULL );
	//CachedRangeQuerry querry(pc);
	FastSphereQuerry fq(pc);
	TopologicalRangeQuerry * topQuerry = NULL;
	if( m_UseTopogicalRangeQuerry )
	{
		UnstructuredInCorePointCloud* upc = dynamic_cast<UnstructuredInCorePointCloud*>(pc);
		PointCloudIterator* tIt = NULL;
		if (upc && dynamic_cast<TopologyAttachment*>(pc->getAttachments()->getData(TopologyAttachment::getDefaultName())))
		{
			InCorePCTopologyGraph* tpg = dynamic_cast<TopologyAttachment*>(pc->getAttachments()->getData(TopologyAttachment::getDefaultName()))->getTopology();
			if (tpg) tIt = tpg->createIterator(upc->getPointSet());
		}
		topQuerry = new TopologicalRangeQuerry(pc,tIt);
	}

	card32 index = 0;
	progressWindow->pushStep(true, "curv");
	std::vector<Vector3f> posBuffer;
	std::vector<float> weights;
	float kernelRadSqr = sqr(m_KernelSigma * 2.0f);
	PointSet * ps = ((UnstructuredInCorePointCloud*)pc)->getPointSet();

	while( !bIt->atEnd() )
	{
		Vector3f normal = bIt->get3f( normalAAT );
		Matrix3f invTangentSystem = calcTangentSystem( normal ).transpose();

		Vector3f pos = bIt->get3f( posAAT );

		mpcard numPoints;
		if( !m_UseTopogicalRangeQuerry )
		{
			//std::vector<Vector3f> * cache = querry.getSpherePositions( pos, m_KernelSigma * 2.0f, cacheRadius );
			mpcard * indices;
			fq.querry(pos, m_KernelSigma * 2.0f, &indices, numPoints);
			//numPoints = cache->size();
			if( posBuffer.size() < numPoints )
			{
				posBuffer.resize(numPoints);
				weights.resize(numPoints);
			}

			card32 current = 0;
			float invKernelSigma = 1.0f / m_KernelSigma;
			float invKernelSigma_2 = 1.0f / m_KernelSigma / m_KernelSigma;

			for( card32 i=0;i<numPoints;i++ )
			{
				Vector3f tmp = //(*cache)[i] - pos;
					ps->get3f(indices[i],posAAT) - pos;
				float dist_2 = fq.getSquaredDistance(i);//normQuad(tmp);
				//if( dist_2 < kernelRadSqr )
				{
					weights[current] = //sqr(exp(-dist_2));
						NormalDistribution::expApproximation(-dist_2 * invKernelSigma_2);
					posBuffer[current++] = invTangentSystem * tmp * invKernelSigma;
				}
			}
			//numPoints = current;
		}
		else
		{
			topQuerry->rangeQuerry(pos, m_KernelSigma * 2.0f );
			numPoints = topQuerry->getNumPoints();
			if( posBuffer.size() < numPoints )
			{
				posBuffer.resize(numPoints);
				weights.resize(numPoints);
			}

			for( card32 i=0;i<numPoints;i++ )
			{
				Vector3f tmp = topQuerry->getPosition(i) - pos;
				float dist_2 = pow(topQuerry->getDistance(i),2.0f);
				weights[i] = sqr(exp(-dist_2));
				posBuffer[i] = invTangentSystem * tmp / m_KernelSigma;
			}

		}

		StaticVector<float32, 6> coeff;
		//Monomials2ndOrder::computeCoeffs( &posBuffer[0], (card32)numPoints, coeff, 1.0f );
		Monomials2ndOrder::computeCoeffs( &posBuffer[0], &weights[0], (card32)numPoints, coeff);

		float f_x = coeff[1];
		float f_y = coeff[2];
		float f_xy = coeff[4];
		float f_xx = 2.0f * coeff[3];
		float f_yy = 2.0f * coeff[5];

		float K = (f_xx*f_yy - pow( f_xy, 2.0f ))
			/ pow(1.0f + pow(f_x,2.0f) + pow(f_y,2.0f),2.0f );

		float H = ((1.0f + pow(f_x,2.0f))*f_yy - 2.0f * f_x * f_y * f_xy + (1.0f + pow(f_y,2.0f))*f_xx)
			/ pow(1.0f + pow(f_x,2.0f) + pow(f_y,2.0f),2.0f/3.0f ) / 2.0f;


		Vector4f res_vec;
		float sign = H<0 ? -1:1;
		res_vec[ 2 ] = H;
		res_vec[ 3 ] = K;
		res_vec[ 0 ] = H + sqrt( H*H - K);
		res_vec[ 1 ] = H - sqrt( H*H - K);

		if( !m_UseNormalOrientation )
		{
			res_vec[ 0 ] = fabs(res_vec[ 0 ]);
			res_vec[ 1 ] = fabs(res_vec[ 1 ]);
			res_vec[ 2 ] = fabs(res_vec[ 2 ]);
			res_vec[ 3 ] = fabs(res_vec[ 3 ]);
		}
		mIt->set4f( destAAT, res_vec );
		index++;

		if( rnd01() <= 0.1f && !m_UseMedianFilter)
		{
			m_Curv1.push_back(res_vec[0]);
			m_Curv2.push_back(res_vec[1]);
			m_Curv3.push_back(res_vec[2]);
		}

		bIt->next();

		if( index % 100 == 0)
			if( !progressWindow->progress((float)index/(float)pc->getNumPoints()*100.0f) )
				break;
	}
	bool aborted  = false;
	progressWindow->popSteps(&aborted);

	if( m_UseMedianFilter )
	{
		card32 index = 0;
		progressWindow->pushStep(true, "curv(med)");
		std::vector<float> curv1,curv2;
		float kernelRadSqr = sqr(m_KernelSigma);
		CachedRangeQuerry querry(pc);
		bIt->reset();
		while( !bIt->atEnd() )
		{
			Vector3f pos = bIt->get3f( posAAT );

			PointSet * cache = querry.getSpherePoints( pos, m_KernelSigma, cacheRadius );
			const card32 numPoints = cache->getNumEntries();
			if( curv1.size() < numPoints )
			{
				curv1.resize(numPoints);
				curv2.resize(numPoints);
			}

			card32 current = 0;
			for( card32 i=0;i<numPoints;i++ )
			{
				Vector3f tmp = cache->get3f(i,posAAT) - pos;
				float dist_2 = normQuad(tmp);
				if( dist_2 < kernelRadSqr )
				{
					Vector4f tmpCurv = cache->get4f(i,destAAT);
					curv1[current] = tmpCurv[0];
					curv2[current] = tmpCurv[1];
					current++;
				}
			}

			sort( curv1.begin(),curv1.begin()+current);
			sort( curv2.begin(),curv2.begin()+current);

			Vector4f curv = bIt->get4f(destAAT);
			curv[0] = curv1[ current / 2 ];
			curv[1] = curv2[ current / 2 ];
			mIt->set4f( destAAT, curv );
			index++;

			if( rnd01() <= 0.1f )
			{
				m_Curv1.push_back(curv[0]);
				m_Curv2.push_back(curv[1]);
				m_Curv3.push_back(curv[2]);
			}

			bIt->next();

			if( index % 100 == 0)
				if( !progressWindow->progress((float)index/(float)pc->getNumPoints()*100.0f) )
					break;
		}
		bool aborted  = false;
		progressWindow->popSteps(&aborted);	}

//	delete hIt;
	delete iter;
	delete topQuerry;
	debugOutput << "Curvature time: " << convertTimeToString(timer.getDeltaValue()) << "\n";
}

void PCCSurfaceTools::scaleCurvature( PointCloud * pc, const Vector4f &minCurv, const Vector4f &maxCurv)
{
	AAT destAAT = pc->getAAT( m_DestChannelName );

	SceneObjectIterator * iter = pc->createIterator( SceneObjectIteratorType::CAP_BASIC_PC );
	x4Assert( iter != NULL );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(iter);

	Vector4f scaleCurv = maxCurv - minCurv;
	for( card32 i=0;i<4;i++)
		scaleCurv[i] = 1.0f / scaleCurv[i];


	int imgSize = 16;
	RGBAFloat32Image his2d(imgSize,imgSize);
	his2d.clear( makeVector4f(0,0,0,0) );

	while( !bIt->atEnd() )
	{
		Vector4f curv = bIt->get4f( destAAT );
		curv = (curv - minCurv).componentProduct(scaleCurv);
		for( card32 i=0;i<4;i++)
			curv[i] = max( 0.0f, min( 1.0f,curv[i]));

		Vector4f tmp = curv * imgSize;
		int x = min( tmp[0], (float)imgSize-1 );
		int y = min( tmp[1], (float)imgSize-1 );
		his2d.setPixel(x,y,his2d.getPixel(x,y) + makeVector4f(1,1,1,1));

		mIt->set4f( destAAT, curv);

		bIt->next();
	}
	
	ImageDisplayer::DisplayRGBAFloat32Image ev(&his2d);
	debugOutput.displayEvent(ImageDisplayer::getClass(), "his2d", &ev);


}

void X4::PCCSurfaceTools::computeCurvatureHistogram()
{
	int numBins = 256;
	int numBinsY = 256;
	float minMeanCurv = -1.0f;
	float maxMeanCurv =  1.0f;
	float minGaussCurv = -0.6f;
	float maxGaussCurv =  0.6f;
	unsigned numOutlier=0;
	unsigned numPoints = 0;
	std::vector<float> histogram(numBins,0.0f);
	RGBAFloat32Image histogram2D(numBins,numBinsY);
	histogram2D.clear( makeVector4f(0,0,0,0));
	const std::vector<std::string> selection = getScene()->getRootState()->staticState->getSelectedNodes();
	for( card32 i=0;i<selection.size();i++ )
	{
		PointCloud * pc = getPointCloud( getScene(), selection[i] );
		if( !pc )
			continue;
		AAT posAAT = pc->getAAT( "position" );
		AAT destAAT = pc->getAAT( m_DestChannelName );
		SceneObjectIterator * iter = pc->createIterator( SceneObjectIteratorType::CAP_BASIC_PC );
		x4Assert( iter != NULL );
		BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);

		while( !bIt->atEnd() )
		{
			Vector4f curv = bIt->get4f(destAAT);
			float meanCurv = curv[2];
			float gaussCurv = curv[3];
			numPoints++;
			
			int index = (meanCurv - minMeanCurv)/(maxMeanCurv-minMeanCurv)*(float)(numBins-1);
			if( index >= 0 && index < numBins)
				histogram[index] += 1.0f;
			else
				numOutlier++;

			int indexY = (gaussCurv - minGaussCurv)/(maxGaussCurv-minGaussCurv)*(float)(numBinsY-1);

			if( index >= 0 && index < numBins && indexY >= 0 && indexY < numBinsY )
				histogram2D.setPixel(index,indexY, histogram2D.getPixel(index,indexY) + makeVector4f(1,1,1,1));

			bIt->next();
		}

	}

	// filter histogram
	int kernelSize = 2;
	std::vector<float> kernel(kernelSize*2+1);
	float summedWeight = 0;
	for( int j=-kernelSize;j<=kernelSize;j++ )
	{
		float d = (float)j/(float)kernelSize;
		kernel[j+kernelSize] = exp(-pow(d,2.0f));
		summedWeight += kernel[j+kernelSize];
	}
	for( unsigned i=0;i<kernel.size();i++ )
		kernel[i] /= summedWeight;

	std::vector<float> filteredHistogram = histogram;
	float maxEntry = 0;
	for( int i=0;i<numBins;i++)
	{
		float tmp = 0;
		for( int j=-kernelSize;j<=kernelSize;j++ )
		{
			if( i+j >= 0 && i+j < numBins )
				tmp += histogram[i+j] * kernel[j+kernelSize];
		}
		filteredHistogram[i] = tmp;
		maxEntry = max(maxEntry,tmp);
	}

	// draw image
	RGBAFloat32Image image;
	image.resize(numBins, 40);
	image.clear(makeVector4f(0,0,0,0));
	for( unsigned i=0;i<(unsigned)numBins;i++ )
	{
		float height = filteredHistogram[i] / maxEntry * (float)image.getSizeY();

		for( unsigned j=0;j<height && j<image.getSizeY();j++ )
		{
			image.setPixel(i,image.getSizeY()-1-j,makeVector4f(1,1,1,1));
		}
	}

	ImageDisplayer::DisplayRGBAFloat32Image ev(&image);
	debugOutput.displayEvent(ImageDisplayer::getClass(), "curv-histogram", &ev);


	debugOutput << "num outliers: " << (float)numOutlier*100.0f / (float)numPoints << "%\n";


	// 2d histogram
	histogram2D.filter( 3.0f);
	ImageDisplayer::DisplayRGBAFloat32Image ev2(&histogram2D);
	debugOutput.displayEvent(ImageDisplayer::getClass(), "2dcurv-histogram", &ev2);
	

}

void X4::PCCSurfaceTools::colorizePointClouds()
{
	std::vector<PointCloud*> pcs = getSelectedPointClouds();
	
	std::vector<Vector3f> colors = getRandomColorTable((card32)pcs.size());




	for( unsigned i=0;i<pcs.size();i++ )
	{
		PointCloud * pc = pcs[i];
		checkAttribute(pc,"color",3,VAD::DATA_FORMAT_FLOAT32);
		SceneObjectIterator * iter = pc->createIterator(SOIT::CAP_BASIC_PC|SOIT::CAP_MODIFYABLE_PC);
		BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
		ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(iter);
		AAT colAAT = pc->getAAT("color");

		while(!bIt->atEnd())
		{
			mIt->set3f(colAAT,colors[i]);

			bIt->next();
		}

		delete iter;
	}
}

void X4::PCCSurfaceTools::resample()
{
	std::vector<PointCloud*> pcs = getSelectedPointClouds();

	for( unsigned i=0;i<pcs.size();i++ )	
	{
		PointCloud * pc = pcs[i];
		PointCloud * resPC = resamplePC(pc,m_KernelSigma);
		addPointCloud(getScene(), resPC, "resampled");
	}
}

void X4::PCCSurfaceTools::importHeightfield()
{
	QString file = QFileDialog::getOpenFileName(NULL,"filename", ".bmp");
	if( !file.isEmpty() )
	{
		QImage image;
		if( !image.load( file ) )
			return;

		UnstructuredInCoreTriangleMesh * mesh = new UnstructuredInCoreTriangleMesh;
		VertexDescriptor * vd_mesh = new VertexDescriptor;
		VertexDescriptor * vd = createDefaultVertexDescriptor(true,true,true,true,false);
		vd_mesh->pushAttrib(VAD("index",3,VAD::DATA_FORMAT_INT32));
		mesh->clearAndSetup(vd,image.width()*image.height(),vd_mesh,(image.width()-1)*(image.height()-1)*2);
		delete vd;
		delete vd_mesh;

		VertexArray va_points(mesh->getPointSet());
		VertexArray va_indices(mesh->getTriangles());

		unsigned triIndex = 0;
		for( unsigned x=0;x<(unsigned)image.width();x++ )		
		{
			for( unsigned y=0;y<(unsigned)image.height();y++ )
			{
				unsigned index = y*image.width()+x;
				va_points.setPosition3f(index, makeVector3f(x,y,0) / max(image.width(),image.height()) + 
					makeVector3f(0,0,(float)qGray(image.pixel(x,y))/255.0f*0.1f));

				if( x > 0 && y > 0)
				{
					Vector3i t1 = makeVector3i(index-1,index,index-1-image.width());
					Vector3i t2 = makeVector3i(index-1-image.width(),index,index-image.width());
					va_indices.setIndex3i(triIndex++,t1);
					va_indices.setIndex3i(triIndex++,t2);
				}
			}
		}

		mesh->calculateVertexNormals();
		mesh->setMaterialIndex(3);
		addPointCloud(getScene(),mesh,"heightfield");
	}
}

void drawBBFunc(BoundingBox3f bb)
{
	debugRenderer->addBoundingBox(bb,makeVector3f(1,1,1),1);
}

void X4::PCCSurfaceTools::selectBorderPoints()
{
	std::vector<PointCloud*> pcs = getSelectedPointClouds();
	if( pcs.size() == 0 )
		return;

	PointCloud * pc = pcs[0];
	UnstructuredInCorePointCloud * upc = dynamic_cast<UnstructuredInCorePointCloud*>(pc);
	AttachedIndexedOctree * m_octree = dynamic_cast<AttachedIndexedOctree*>(pc->getAttachments()->getData("SimpleOctree"));
	if(m_octree==NULL){
		m_octree=new AttachedIndexedOctree();
		m_octree->setup("SimpleOctree", AttachedData::ADF_QUERY_DELETE_ON_UPDATE,  NULL);
		m_octree->octreesetup(upc->getPointSet(), 32);
		pc->getAttachments()->attachData(m_octree);
	}
	
	/* debugRenderer->clearDebugData("testFastQuerry");
	debugRenderer->beginRenderJob("testFastQuerry");

	AAT posAAT = pc->getAAT("position");
	float medPointDistance = getMedianPointDistance(pc);
	Vector3f pos = upc->getPointSet()->get3f(rand() % pc->getNumPoints(), posAAT );
	FastSphereQuerry fq(m_octree);
	fq.m_func = drawBBFunc;
	mpcard * indices;
	mpcard numPoints;
	float rad = medPointDistance * 4.0f;
	fq.querry(pos, rad, &indices, numPoints);
	
	debugRenderer->addSphereWiremesh(pos, rad, 1, makeVector3f(1,0,0));
	debugRenderer->addSphere(pos,medPointDistance*0.5f, makeVector3f(1,0,0));

	for( unsigned i=0;i<numPoints;i++ )
		debugRenderer->addSphere(upc->getPointSet()->get3f(indices[i],posAAT), medPointDistance * 0.1f, makeVector3f(0,1,1));

	debugRenderer->endRenderJob();
/*

	AAT posAAT = pc->getAAT("position");
	float medPointDistance = getMedianPointDistance(pc);
	PointSet * ps = upc->getPointSet();
	FastSphereQuerry fq(m_octree);
	mpcard * indices;
	mpcard numPoints;
	float rad = m_KernelSigma * medPointDistance;
	HierarchicalKNNIterator hIt(pc,32,NULL);

	Timer t;
	for( unsigned i=0;i<ps->getNumEntries();i++ )
	{
		fq.querry(ps->get3f(i,posAAT), rad, &indices, numPoints);
	}


	debugOutput << "Timings new: " << convertTimeToString(t.getDeltaValue()) << "\n";

	float rad_2 = rad * rad;
	for( unsigned i=0;i<ps->getNumEntries();i++ )
	{
		Vector3f pos = ps->get3f(i,posAAT);
		hIt.setSeekPointAndReset(pos);
		while(!hIt.atEnd())
		{
			Vector3f tmpPos = hIt.get3f(posAAT);
			float dist_2 = normQuad(tmpPos-pos);
			if( dist_2 > rad_2 )
				break;
			hIt.next();
		}
	}

	debugOutput << "Timings knn: " << convertTimeToString(t.getDeltaValue()) << "\n";

/**/



	for( unsigned i=0;i<pcs.size();i++ )	
	{
		float relRadius = 5.0f;
		float minAngle = 2.0f;
		PointCloud * pc = pcs[i];

		selectBorderPoints(pc, relRadius, minAngle);

	}
}

void X4::PCCSurfaceTools::selectBorderPoints( PointCloud * pc, float relRadius, float minAngle, unsigned flag )
{
	float medPointDist = getMedianPointDistance(pc);
	float radius = relRadius * medPointDist;

	SceneObjectIterator * iter = pc->createIterator( SceneObjectIteratorType::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC
		| SOIT::CAP_RANDOM_ACCESS_PC);
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(iter);
	RandomAccessPointCloudIterator * rIt = dynamic_cast<RandomAccessPointCloudIterator*>(iter);
	IndexedPointCloudIterator * iIt = dynamic_cast<IndexedPointCloudIterator*>(iter);
	AAT posAAT = pc->getAAT("position");
	AAT flagsAAT = pc->getAAT("flags");
	HierarchicalKNNIterator hIt(pc,32,NULL);
	std::vector<unsigned> borderIndices;

	while( !bIt->atEnd() )
	{
		std::vector<Vector3f> positions;
		PCA<float,3> pca;
		Vector3f pos = bIt->get3f(posAAT);
		hIt.setSeekPointAndReset(pos);
		hIt.next();

		// get local environment and compute pca frame
		pca.addPoint(pos);
		while( !hIt.atEnd() )
		{
			Vector3f tmp = hIt.get3f(posAAT);
			if( norm(tmp - pos) > radius )
				break;
			pca.addPoint(tmp);
			positions.push_back(tmp);

			hIt.next();
		}

		Vector3f centroid,eigenvalues;
		Matrix3f eigenVectors;
		pca.analyze(eigenvalues,eigenVectors,centroid);
		Matrix3f worldToFrame = eigenVectors.transpose();

		std::vector<float> angles;
		for( unsigned p=0;p<positions.size();p++ )
		{
			Vector3f tmp = worldToFrame * (positions[p]-pos);
			float angle = atan2(tmp[0],tmp[1]);
			angles.push_back(angle);
		}

		sort(angles.begin(),angles.end());

		// check for big angles
		bool border = false;
		for( mpint i=0;i<(mpint)angles.size()-1;i++ )
		{
			float deltaAngle = angles[i+1] - angles[i];
			if( deltaAngle > minAngle )
			{
				border = true;
				break;
			}
		}

		border |= (fabs(angles[0] - (angles.back() - 2.0f * M_PI)) > minAngle);

		if( border )
		{
			borderIndices.push_back( iIt->getCurrentPointIndex());
			int f = bIt->get1i(flagsAAT);
			f |= flag;
			mIt->set1i(flagsAAT,f);
			//AAT colAAT = pc->getAAT("color");
			//mIt->set3f(colAAT,makeVector3f(1,1,0));
			//debugRenderer->addPoint(pos,makeVector3f(1,0,0));
		}
		else
		{
			int f = bIt->get1i(flagsAAT);
			f = f & (~flag);
			mIt->set1i(flagsAAT,f);
			//AAT colAAT = pc->getAAT("color");
			//mIt->set3f(colAAT,makeVector3f(0.5f,0.5f,0.5f));
		}

		bIt->next();
	}

	std::vector<unsigned> invalidBorderPoints;
	for( unsigned i=0;i<borderIndices.size();i++ )
	{
		rIt->goToIndex(borderIndices[i]);
		Vector3f pos = bIt->get3f(posAAT);
		hIt.setSeekPointAndReset(pos);
		hIt.next();

		unsigned numBorderPoints=0;
		while( !hIt.atEnd() )
		{
			Vector3f tmp = hIt.get3f(posAAT);
			if( norm(tmp - pos) > radius )
				break;

			if( hIt.get1i(flagsAAT) & flag )
			{
				numBorderPoints++;
			}

			hIt.next();
		}

		if( numBorderPoints < 5)
			invalidBorderPoints.push_back(borderIndices[i]);
	}

	for( unsigned i=0;i<invalidBorderPoints.size();i++ )
	{
		rIt->goToIndex(borderIndices[i]);
		int f = bIt->get1i(flagsAAT);
		f = f & (~flag);
		mIt->set1i(flagsAAT,f);
	}

	delete iter;

}

std::vector<Vector3f> X4::PCCSurfaceTools::getRandomColorTable( card32 numColors )
{
	// create rnd colors
	std::vector<Vector3f> colors;
	for( unsigned i=0;i<numColors;i++ )
	{
		float bestDist = 0;
		Vector3f bestColor = makeVector3f(rnd01(),rnd01(),rnd01());
		for( unsigned j=0;j<100;j++ )
		{
			Vector3f rndColor;

			do
			{
				rndColor = makeVector3f(rnd01(),rnd01(),rnd01());
			} while(norm(rndColor) < 0.15f || norm(rndColor) > 0.85);

			float minDist = 1;
			for( unsigned c=0;c<colors.size();c++ )			
			{
				float dist=norm(rndColor - colors[c]);
				minDist = min(minDist,dist);
			}

			if( minDist > bestDist )
			{
				bestDist = minDist;
				bestColor = rndColor;
			}
		}

		colors.push_back(bestColor);
	}

	return colors;
}

PointCloud * X4::PCCSurfaceTools::resamplePC( PointCloud * pc, float sigma )
{
	UnstructuredInCorePointCloud * resPC = new UnstructuredInCorePointCloud;

	resPC->clearAndSetup(pc->getDescr(),0);
	resPC->setMaterialIndex(pc->getMaterialIndex());
	std::vector<char> vbuffer(pc->getDescr()->getSize());
	std::map<std::string,Vector3f> grid;
	SceneObjectIterator * iter = pc->createIterator( SceneObjectIteratorType::CAP_BASIC_PC );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	AAT posAAT = pc->getAAT("position");
	AAT normalAAT = pc->getAAT("normal");

	float gridSamplingSpace = sigma;
	while( !bIt->atEnd())
	{		
		Vector3f pos = bIt->get3f(posAAT);
		pos /= gridSamplingSpace;
		std::string entry;
		Vector3f gridcellCenter;
		for( unsigned i=0;i<3;i++ )
		{
			int index = pos[i] >= 0.0f ? pos[i] : pos[i] - 1.0f;
			char * p = (char*)&index;
			entry += p[0];
			entry += p[1];
			entry += p[2];
			entry += p[3];
			gridcellCenter[i] = (float)index * gridSamplingSpace;
		}

		if( grid.find(entry) == grid.end() )			
		{
			grid[entry] = gridcellCenter;
			bIt->getVertex(&vbuffer[0]);
			resPC->getPointSet()->addPoint(&vbuffer[0]);

			Vector3f projPoint, projNormal;
			projectPointOntoPatch( gridcellCenter, sigma, pc, projPoint, projNormal );
			Vector3f normal = bIt->get3f(normalAAT);
			if( projNormal * normal < 0 )
				projNormal = -projNormal;

			resPC->getPointSet()->set3f(resPC->getNumPoints()-1, posAAT, projPoint );
			resPC->getPointSet()->set3f(resPC->getNumPoints()-1, normalAAT, projNormal );
		}

		bIt->next();
	}

	return resPC;
}

void X4::PCCSurfaceTools::deleteInvalidPoints()
{
	std::vector<PointCloud*> pcs = getSelectedPointClouds();
	if( pcs.size() == 0 )
		return;

	for( unsigned i=0;i<pcs.size();i++ )
	{
		PointCloud * pc = pcs[i];
		SceneObjectIterator * iter = pc->createIterator(SOIT::CAP_BASIC_PC | SOIT::CAP_DYNAMIC_PC);
		BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
		DynamicPointCloudIterator * dIt = dynamic_cast<DynamicPointCloudIterator*>(iter);

		AAT flagsAAT = pc->getAAT("flags");
		while( !bIt->atEnd())
		{
			int flag = bIt->get1i(flagsAAT);
			if( flag & PF_FLAG_INVALID )
			{
				dIt->deleteCurrent();
			}
			else
				bIt->next();
		}
		delete iter;
	}
}

void X4::PCCSurfaceTools::colorizeFromOtherPC()
{
	std::vector<PointCloud*> pcs = getSelectedPointClouds();
	if(pcs.size()==0)
		return;
	UnstructuredInCorePointCloud* destPC= dynamic_cast<UnstructuredInCorePointCloud*>(pcs[0]);
	UnstructuredInCorePointCloud* sourcePC= dynamic_cast<UnstructuredInCorePointCloud*>(getPointCloud(scene, "root/colorPC"));
	if (!destPC||!sourcePC)
		return;

	AAT DESTPOS= destPC->getAAT("position");
	AAT DESTCOL= destPC->getAAT("color");
	AAT SOURCECOL= sourcePC->getAAT("color");

	PointSet* sourcePS=sourcePC->getPointSet();
	PointSet* destPS=destPC->getPointSet();
	PointSetANNQuery ann(sourcePS,1);
	
	mpcard destNumPts=destPC->getNumPoints();
	for (mpcard i=0; i<destNumPts; i++)
	{
		int32 nearestSourceIndex = ann.getNearestPointIndex(destPS->get3f(i,DESTPOS));
		destPS->set3f(i, DESTCOL, sourcePS->get3f(nearestSourceIndex, SOURCECOL));
	}
	

	//repaint
	SceneObjectIterator* it=destPC->createIterator(SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC);
	BasicPointCloudIterator* basicIt = dynamic_cast<BasicPointCloudIterator*>(it);
	ModifyablePointCloudIterator* modIt = dynamic_cast<ModifyablePointCloudIterator*>(it);
	modIt->set3f(DESTCOL, basicIt->get3f(DESTCOL));
	delete it;
	ObjectViewsTable::update(scene);
}

void X4::PCCSurfaceTools::mlsProjectPC()
{
	std::vector<PointCloud*> pcs = getSelectedPointClouds();

	for( unsigned i=0;i<pcs.size();i++ )	
	{
		PointCloud * pc = pcs[i];
		PointCloud * resPC = mlsProjectPC(pc,m_KernelSigma);
		addPointCloud(getScene(), resPC, "mlsProjected");
	}

// 	PointCloud * resPC = mlsProjectPC(getPointCloud(scene, "root/pc1"), m_KernelSigma, getPointCloud(scene, "root/pc2"));
// 	addPointCloud(getScene(), resPC, "mlsProjected");
}

PointCloud *  X4::PCCSurfaceTools::mlsProjectPC( PointCloud * dataPointset, float sigma, PointCloud * dataPC )
{	
	X4_TIMER_START(mlsProjectPC);

	if( dataPC == NULL )
		dataPC = dataPointset;

	PointCloud * resPC = dynamic_cast<UnstructuredInCorePointCloud *>(dataPointset->copy());
	SceneObjectIterator* it=resPC->createIterator(SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC);
	BasicPointCloudIterator* basicIt = dynamic_cast<BasicPointCloudIterator*>(it);
	ModifyablePointCloudIterator* modIt = dynamic_cast<ModifyablePointCloudIterator*>(it);

	PointSet * ps= modIt->lockBlockReadWrite();
	
	//float32	myMedianPointDistance=getMedianPointDistance(dataPointset);

	//sigma*=myMedianPointDistance;
	FastSphereQuerry* sphereQuery= new FastSphereQuerry(dataPC);


	mpcard numPoints= dataPointset->getNumPoints();
	mpcard * indices;
	mpcard numPointsInRadius=0;
	AAT POSITION=ps->getAAT("position");
	AAT NORMAL_MOD = ps->getAAT("normal");
	AAT NORMAL_DATA = dataPC->getAAT("normal");
	PointSet* dataPS=dynamic_cast<UnstructuredInCorePointCloud *>(dataPC)->getPointSet();

	float inv_sigma_2 = 1.0f / sigma / sigma;
	float rad_2 = sigma * sigma * 4.0f;
	float rad = 2.0f * sigma;
	PCA<float,3> pca;

	progressWindow->pushStep(true, "mls");
	card32 index=0;

	for (mpcard i=0; i< numPoints; i++)
	{
		if( index++ % 100 == 0)
			if( !progressWindow->progress((float)index/(float)numPoints*100.0f) )
				break;
		
		pca.clear();
		Vector3f pointToOptPos=ps->get3f(i, POSITION);
		sphereQuery->querry(pointToOptPos, 2*sigma, &indices, numPointsInRadius);
		if (numPointsInRadius<10)
			continue;
		
		std::vector<float> weights;
		weights.resize(numPointsInRadius);
		std::vector<Vector3f> positions;
		positions.resize(numPointsInRadius);
		Vector3f formerNormal = ps->get3f(i, NORMAL_MOD);
		card32 dataPointsCount=0;

		for (mpcard n=0; n < numPointsInRadius; n++)
		{
			mpcard index = *indices++;
			Vector3f sample_pos = dataPS->get3f(index, POSITION);
			float dist_2 = sphereQuery->getSquaredDistance(n);
				//normQuad(sample_pos - pointToOptPos);
			
			

			float debugDist = sqrtf(dist_2);
			float debugInv  = 1.0/sqrtf(inv_sigma_2);


			float w = exp(-dist_2*inv_sigma_2);

			Vector3f dataNormal = dataPS->get3f(index, NORMAL_DATA);
			float32 angle = formerNormal * dataNormal;
			float32 normalWeight = std::max<float>(0, angle);
			w *= normalWeight;


			//float w2 = NormalDistribution::expApproximation(dist_2*inv_sigma_2);
			//debugOutput << w << " <-w  wND-> " << w2 << "\n";
			pca.addPoint(sample_pos, w);

			weights[n]= w ;
			positions[n]= sample_pos;

			if (w > 0.0f)
			{
				dataPointsCount++;
			}
		}

		if (dataPointsCount<5)
			continue;
		
		Vector3f eigenValues,centroid;
		Matrix3f eigenVectors;
		pca.analyze(eigenValues,eigenVectors,centroid);

		Matrix3f frame = eigenVectors.transpose();

		centroid = pointToOptPos - eigenVectors[2] * (eigenVectors[2] * (pointToOptPos - centroid));

		// transform into coordinate frame
		Matrix3f worldToLocal = frame / rad;
		for( unsigned q=0;q<positions.size();q++ )
		{
			positions[q] = worldToLocal * (positions[q]-centroid);
		}
		// fit monomials
		Vector6f coeffs;
		Monomials2ndOrder::computeCoeffs( &positions[0], &weights[0], (card32)positions.size(), coeffs);

		// project point and re transform into reference frame
		Vector3f projPoint = eigenVectors * makeVector3f(0,0,coeffs[0])*rad + centroid;
		float32 y = projPoint[1];
		if (y==0.0f)
		{
			debugOutput << 0;
		}
		Vector3f projNormal = eigenVectors * Monomials2ndOrder::evaluateNormal( coeffs, makeVector2f(0,0));

		if( projNormal * formerNormal < 0)
			projNormal *= -1;
		const float32 offsetLength = norm(projPoint - ps->get3f(i, POSITION));
		//if (offsetLength <= m_MlsMaxOffset && m_MlsMaxOffset != 0.0f)
		//{
		// TODO
			ps->set3f(i, POSITION, projPoint);
			ps->set3f(i, NORMAL_MOD, projNormal);
		//}
	}

	progressWindow->popSteps();

	delete sphereQuery;


	//Repaint:
	basicIt->unlock();
	delete it;

	X4_TIMER_STOP(mlsProjectPC);
	return resPC;
}

void X4::PCCSurfaceTools::alignBoundingBoxes()
{
	std::vector<PointCloud*> pcs = getSelectedPointClouds();

	if( pcs.size() < 2)
		return;

	BoundingBox3f bb1 = getPCBBox(pcs[0]);
	

	for( unsigned i=1;i<pcs.size();i++ )
	{
		BoundingBox3f bb2 = getPCBBox(pcs[i]);
		Vector3f d1 = bb1.upperCorner - bb1.lowerCorner;
		Vector3f d2 = bb2.upperCorner - bb2.lowerCorner;
		Vector3f scale = d1.componentProduct(makeVector3f(1.0f / d2[0],1.0f / d2[1],1.0f / d2[2]));
		Matrix4f T = makeTranslation4f(bb1.lowerCorner)*makeScale4f(scale)*makeTranslation4f(-bb2.lowerCorner);
		pcs[i]->setTransformation(T);
	}


}

void X4::PCCSurfaceTools::eraseBadPoints()
{
	SGListNode* anim = (SGListNode*)getSceneGraphNode(scene, "root/data");
	if (anim == NULL) return;

	bool goOn = true;
	mpcard childNo = 0;
	while(goOn)
	{

		SGObjectNode* sgobj = dynamic_cast<SGObjectNode*>(anim->getChildNode(getScene()->getRootState(), childNo));
		
		if(sgobj)
		{
			UnstructuredInCorePointCloud* pc = dynamic_cast<UnstructuredInCorePointCloud*>(sgobj->getSceneObject());
			AAT CONF = pc->getAAT("confidence");
			PointSet* ps = pc->getPointSet();
			for (mpint i = pc->getNumPoints()-1; i >= 0 ; i--)
			{
				if (ps->get1f(i, CONF)<8)
				{
					ps->deletePoint(i);
				}
			}
			childNo++;
		}
		else
		{
			goOn = false;
		}
	}
}