//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PCCTopologyPreservingResampling.h"
//---------------------------------------------------------------------------
#include "NumericalClassProperty.h"
#include "DebugOutput.h"
#include "PointCloudIterators.h"
#include "PointSet.h"
#include "PointSet.inline.h"
#include "UnstructuredInCorePointCloud.h"
#include "PointCloudTools.h"
#include "SceneGraphState.h"
#include "StandardSceneGraphIterator.h"
#include "SGObjectNode.h"
#include "SGAnimationNode.h"
#include "ProgressWindow.h"
#include "SeparatorClassProperty.h"
#include "InCorePCTopologyGraph.h"
#include "TopologyAttachment.h"
#include "TopologicalKNNIterator.h"
#include "GeometricTools.h"

//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

IMPLEMENT_X4_CLASS( PCCTopologyPreservingResampling, 2)
{
   BEGIN_CLASS_INIT( PCCTopologyPreservingResampling )
   ADD_FLOAT32_PROP( m_Spacing, 1 )
   ADD_FLOAT32_PROP ( m_ReconnectFactor, 2 );
}

PCCTopologyPreservingResampling::PCCTopologyPreservingResampling()
{
   m_Spacing = 0.01f;
   m_ReconnectFactor = 3.1f;
}

PCCTopologyPreservingResampling::~PCCTopologyPreservingResampling()
{
}

void PCCTopologyPreservingResampling::resamplePC(UnstructuredInCorePointCloud* pc)
{
	// copy point cloud
	UnstructuredInCorePointCloud* pcCopy = copyToUnstructuredInCorePointCloud(pc);
	AAT POSITION = pcCopy->getAAT( "position" );
	AAT FLAGS = pcCopy->getAAT( "flags" );

	// get topology
	if (!pc->getAttachments()->getData(TopologyAttachment::getDefaultName()))
	{
		warning("Selected pointcloud does not contain topology information");
		return;
	}

	InCorePCTopologyGraph* tpg = NULL;
	TopologyAttachment* tatt = dynamic_cast<TopologyAttachment*>(pc->getAttachments()->getData(TopologyAttachment::getDefaultName()));
	tpg = tatt->getTopology();
	InCorePCTopologyGraph* tpgCopy = tpg->copy();
	tatt = new TopologyAttachment;
	tatt->setup(TopologyAttachment::getDefaultName(), AttachedData::ADF_PERSISTENT);
	tatt->setTopology(tpgCopy);            
	pcCopy->getAttachments()->attachData(tatt);         

	// Create iterators
	SceneObjectIterator* it = pcCopy->createIterator( SOIT( SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC | SOIT::CAP_DYNAMIC_PC | SOIT::CAP_INDEXED_PC) );
	BasicPointCloudIterator* basicIt = dynamic_cast<BasicPointCloudIterator*>( it );
	ModifyablePointCloudIterator* modIt = dynamic_cast<ModifyablePointCloudIterator*>( it );
	DynamicPointCloudIterator* dynIt = dynamic_cast<DynamicPointCloudIterator*>( it );
	IndexedPointCloudIterator* indIt = dynamic_cast<IndexedPointCloudIterator*>(it);

	ModifyableKNNIterator* knnIt = new ModifyableKNNIterator( pcCopy, 32, NULL );
	knnIt->reset();

	// Reset all flags
	basicIt->reset();
	while( !basicIt->atEnd() )
	{
		modIt->set1i( FLAGS, 0 );
		basicIt->next();
	}

	card32 numPointStart = pc->getNumPoints();
	card32 numNewPoints = numPointStart;

	// Mark all delete points

	float32 sqrSpacing = sqr( m_Spacing );
	basicIt->reset();
	while( !basicIt->atEnd() )
	{
		const Vector3f& p = basicIt->get3f( POSITION );
		const int32 flag = basicIt->get1i( FLAGS );

		// point was not visited before
		if (flag == 0)
		{      
			// mark point as visited
			modIt->set1i(FLAGS, 2);

			// find the point and get to the nearest neighbour
			knnIt->setSeekPointAndReset( p );
			knnIt->next();			// Skip self 
			if (!knnIt->atEnd())
			{
				float32 sqrD = 0;         
				do 
				{
					if (knnIt->get1i(FLAGS) == 0) 
					{            
						const Vector3f& np = knnIt->get3f( POSITION );
						sqrD = ( p - np ) * ( p - np );
						if (sqrD < sqrSpacing)
						{
							knnIt->set1i( FLAGS, 1 );
							numNewPoints--;
						}
					}
					knnIt->next();
				}while( sqrD < sqrSpacing && !knnIt->atEnd());
			}
		}
		basicIt->next();
	}
	delete knnIt;

	x4Assert(numPointStart == pcCopy->getNumPoints());

	// -1 means point has been removed
	std::vector<int32> newToOldIndex(numNewPoints, -1);
	std::vector<int32> oldToNewIndex(numPointStart, -1);

	// setup indices correctly and copy points
	card32 numDeleted = 0;
	pc->clearAndSetup(pcCopy->getDescr(), numNewPoints);
	for (unsigned i=0, newindex=0; i < pcCopy->getNumPoints(); i++)
	{
		if (pcCopy->getPointSet()->get1i(i, FLAGS) != 1)
		{
			newToOldIndex[newindex] = i;
			oldToNewIndex[i] = newindex;
			pc->getPointSet()->setVertex(newindex, pcCopy->getPointSet()->getDataPointer(i));
			newindex++;
		}
		else
			numDeleted++;
	}

	// Reset all flags
	basicIt->reset();
	while( !basicIt->atEnd() )
	{
		modIt->set1i( FLAGS, 0 );
		basicIt->next();
	}

	delete it;
	it = NULL;

	x4Assert(numNewPoints == pc->getNumPoints());

	const float32 maxDistance = m_Spacing * m_ReconnectFactor;
	// create new topology graph
	if (tpg)
	{
		// this would be our new topology graph
		InCorePCTopologyGraph *resultTg = new InCorePCTopologyGraph();
		resultTg->clearAndSetup(tpg->getEdges()->getDescr());

		// use this iterator to go through the topology graph
		PointCloudIterator* titer = tpg->createIterator(pcCopy->getPointSet());
		TopologicalKNNIterator knn(pcCopy, titer);

		// for each new we have to establish a new topological connection based on previous one
		for (unsigned i=0; i < pc->getNumPoints(); i++)
		{
			int32 searchCenter = newToOldIndex[i];
			Vector3f searchCenterPos = pcCopy->getPointSet()->get3f(searchCenter, POSITION);
			knn.setStartPointAndReset(searchCenter);
			while(!knn.atEnd())
			{
				card32 currentOldPoint = knn.getPointSetPointNumber();
				if (currentOldPoint != searchCenter)
				{
					Vector3f currentOldPointPos = pcCopy->getPointSet()->get3f(currentOldPoint, POSITION);
					float32 distance = norm(currentOldPointPos-searchCenterPos);
					if (distance > maxDistance)
						break;

					if (oldToNewIndex[currentOldPoint] != -1)
						resultTg->addEdge(i, oldToNewIndex[currentOldPoint]);
				}
				knn.next();
			}
		}

		TopologyAttachment* tatt = NULL;
		tatt = new TopologyAttachment;
		tatt->setup(TopologyAttachment::getDefaultName(), AttachedData::ADF_PERSISTENT);
		tatt->setTopology(resultTg);            
		pc->getAttachments()->attachData(tatt);	
	}

	delete pcCopy;
}

void PCCTopologyPreservingResampling::execute( Scene *scene )
{
   std::vector<std::string> selection = scene->getRootState()->staticState->getSelectedNodes();
   for ( unsigned int m = 0; m < selection.size(); m++ )
   {
      UnstructuredInCorePointCloud* pc = dynamic_cast<UnstructuredInCorePointCloud*>(getPointCloud( scene, selection[m] ));
      PCCTopologyPreservingResampling::resamplePC(pc);
   }
}