//---------------------------------------------------------------------------
#include "StdAfx.h"
#include <windows.h>
#include <mmsystem.h>
//---------------------------------------------------------------------------
#include "PCHOCDescriptor.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "FastSphereQuerry.h"
#include "RegressionPlane3f.h"
#include "UnstructuredInCorePointCloud.h"
#include "FastSphereQuerry.h"
#include "DebugRenderer.h"
#include <float.h>
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------


#define UP makeVector3f(0,0,1)


IMPLEMENT_X4_CLASS(PCHOCDescriptor ,0)
{
	BEGIN_CLASS_INIT(PCHOCDescriptor);
   ADD_CARD32_PROP(numOrientationBins, 0)
   ADD_CARD32_PROP(numSpatialBins, 0)
   ADD_FLOAT32_PROP(smoothness, 0)
   ADD_FLOAT32_PROP(orientationUniquenessThreshold, 0)
   ADD_BOOLEAN_PROP(normalizeDescriptor, 0)
}

PCHOCDescriptor::PCHOCDescriptor()
{
   numOrientationBins = 8;
   numSpatialBins = 4;
   smoothness = 1.0f;
   orientationUniquenessThreshold = 0.9f;
   normalizeDescriptor = true;

	m_up = UP;
	useUp = false;

   query = NULL;
   upc = NULL;
   ps = NULL;
   positionAAT = NULL_AAT;
   curv2AAT = NULL_AAT;
   singleWarning = true;
}


PCHOCDescriptor::PCHOCDescriptor(float32 rad, mpcard orientationBins, mpcard spatialBins, 
								 float32 OrientationUniqueness, float32 smoothnessThreshold, bool normalize) 
{
	radius = rad;
	numOrientationBins = orientationBins;
	numSpatialBins = spatialBins;
	smoothness = smoothnessThreshold;
	orientationUniquenessThreshold = OrientationUniqueness;
	normalizeDescriptor = normalize;

	m_up = UP;
	useUp = false;

   query = NULL;
   upc = NULL;
   ps = NULL;
   positionAAT = NULL_AAT;
   curv2AAT = NULL_AAT;
   singleWarning = true;
}

PCHOCDescriptor::~PCHOCDescriptor()
{

	if (query != NULL) delete query;
	query = NULL;
   
}

void PCHOCDescriptor::setUpDir(Vector3f& up) {
	m_up = normalize(up);
}

void PCHOCDescriptor::useUpDir(bool use) {
	useUp = use;
}


void PCHOCDescriptor::assign(const Object* obj, X4CopyContext *context)
{
	copyObjectProperties(obj,this);
}


void X4::PCHOCDescriptor::beginComputation(UnstructuredInCorePointCloud *upc)
{
   if (this->upc || query) {
      throw EInvalidState("PCHOCDescriptor::beginComputation() - beginComputation called twice without end in between.");
   }
   if (upc->getPointSet()->providesAttribute("position", 3, VAD::DATA_FORMAT_FLOAT32)
    && upc->getPointSet()->providesAttribute("curv2", 2, VAD::DATA_FORMAT_FLOAT32)
    && upc->getPointSet()->providesAttribute("tangent_v", 3, VAD::DATA_FORMAT_FLOAT32)
    && upc->getPointSet()->providesAttribute("normal", 3, VAD::DATA_FORMAT_FLOAT32)) 
   {
       this->upc = upc;
       ps = upc->getPointSet();
       tangentVAAT = ps->getAAT("tangent_v");
       positionAAT = ps->getAAT("position");
       curv2AAT    = ps->getAAT("curv2");
       normalAAT   = ps->getAAT("normal");
       query = new FastSphereQuerry(upc);
   } else {
      throw ETypeMismatch("PCHOCDescriptor::beginComputation() - position channel missing or curv2 channel missing.");
   }
   singleWarning = true;
}

bool X4::PCHOCDescriptor::getDescriptor(mpcard index, float32 radius, DVectorF &result, bool debugVis)
{
	Vector3f& pos = ps->get3f(index, positionAAT);
	Vector3f& n = ps->get3f(index, normalAAT);
	return getDescriptor(pos, n, radius, result, debugVis);
}

bool X4::PCHOCDescriptor::getDescriptor(const Vector3f& pos, const Vector3f& n, float32 radius, DVectorF &result, bool debugVis)
{

   bool valid = true;
   mpcard *indices = NULL;
   mpcard numPoints = 0;
   
   // Vector3f pos = ps->get3f(index, positionAAT);
   query->querry(pos, radius, &indices, numPoints);
   
   //debugOutput << "#neighbours " << numPoints << "\n";
   if (numPoints < 10) {
	   // throw ERangeCheck("no points found");
	   // debugOutput << "NumNeighbours < 10 !\n";
	   return false; 
   }
   
   /// tangent space
   PCA3f pca;
   Vector3f curvWeightedMean = NULL_VECTOR3F;
   float32 weights = 0;
   for (mpcard i=0; i<numPoints; i++) {
      Vector3f cPos = ps->get3f(indices[i], positionAAT);
      Vector2f curv = ps->get2f(indices[i], curv2AAT);
      pca.addPoint(cPos);
      float32 w = fabs(curv[0]);
      curvWeightedMean += cPos * w;
      weights += w;
   }
   curvWeightedMean /= max(weights, 1E-10f);
   Vector3f eigenvalues;
   Matrix3f eigenvectors;
   Vector3f centroid;
   pca.analyze(eigenvalues, eigenvectors, centroid);


   Vector3f normal   = normalize(eigenvectors[2]);
   // if (normal*ps->get3f(index, normalAAT) <0) normal *= -1;
   if (normal*n <0) normal *= -1;

   Vector3f tangentU;
   Vector3f tangentV;
   if (useUp) {

	   if (sqr(normal*m_up) > 0.98f) {
		   // debugOutput << "Invalid normal \n";
		   valid = false;
		   return false;
	   }

	   tangentU = normalize( m_up.crossProduct(normal) );
	   tangentV = normalize( normal.crossProduct(tangentU) );

   } else {
	   // if (eigenvalues[2] < eigenvalues[1]*0.90f) {
	   if (eigenvalues[0] < eigenvalues[1]*0.75f) {
		   // debugOutput << "No reliable geometric orientation! \n";
		  valid = false;
		  return false;
	   }
	   tangentU = normalize(eigenvectors[0]);
	   tangentV = normalize(eigenvectors[1]);
   }



   Vector3f curvMeanDiff = curvWeightedMean - pos;
   if (!useUp && (norm(curvMeanDiff) < radius*orientationUniquenessThreshold)) {

		// debugOutput << "Curvature not uniquely oriented! \n";

      valid = false;
	  return false;
   }
   if (!useUp) {
      curvMeanDiff = normalize(curvMeanDiff);
      Vector2f newCoords = makeVector2f(tangentU*curvMeanDiff, tangentV*curvMeanDiff);
      tangentU = tangentU*newCoords[0] + tangentV*newCoords[1];
      tangentV = tangentU.crossProduct(normal);
   }
   
   /// building descriptor
   const float32  spatialBinSpacing = 2*radius / numSpatialBins;
   const Vector3f upperLeft  = pos - tangentU*radius - tangentV*radius;
   const Vector3f lowerRight = pos + tangentU*radius + tangentV*radius;
   const float32  binRadius = spatialBinSpacing * 2.0f * smoothness;
   const float32  oBinSpacing = 1.0f / numOrientationBins;
   
   result.setDim(numSpatialBins*numSpatialBins*numOrientationBins);
   result.setZero();
   
   Vector3f binPosV = upperLeft;
   card32 baseIndex = 0;
   for (card32 v=0; v<numSpatialBins; v++) {
      Vector3f binPosU = binPosV;
      for (card32 u=0; u<numSpatialBins; u++) {
      
         mpcard numLocalPoints;
         mpcard *localIndices;
         query->querry(binPosU, binRadius, &localIndices, numLocalPoints);
         
         for (mpcard j=0; j<numLocalPoints; j++) {
            Vector3f localPosVec  = ps->get3f(localIndices[j], positionAAT);
            float32  magnitude    = ps->get2f(localIndices[j], curv2AAT)[0];
            Vector3f localCurvVec = ps->get3f(localIndices[j], tangentVAAT);
            Vector2f localCoords  = makeVector2f(tangentU*localCurvVec, tangentV*localCurvVec);
            if (_isnan(magnitude)) {
               magnitude = 0;
               if (singleWarning) {
                  warning("NaN principal curvature encountered - this is not in Indian restaurant!");
                  singleWarning = false;
               }
            }
            localCoords = makeVector2f(1,0);
            localCurvVec = NULL_VECTOR3F;
            float32 sBinWeight = exp(-((localPosVec-binPosU)*(localPosVec-binPosU))/sqr(spatialBinSpacing*smoothness));

            float32 angle = atan2(localCoords[0],localCoords[1]) / (float32)M_PI;
            if (angle > 1.0f) angle -= 1.0f;
            for (card32 oBin=0; oBin<numOrientationBins; oBin++) {
               float32 oBinAngle = (float32)oBin / (float32)numOrientationBins;
               float32 oBinWeight = exp(-sqr(angle-oBinAngle)/sqr(oBinSpacing*smoothness));
               result[baseIndex+oBin] += oBinWeight*sBinWeight*fabs(magnitude);
            }
         }
         
         baseIndex += numOrientationBins;
         binPosU += spatialBinSpacing*tangentU;
      }      
      binPosV += spatialBinSpacing*tangentV;
   }
   if (normalizeDescriptor) {
      float32 m = 0;
      for (mpcard i=0; i<result.getDim(); i++) {
         m += result[i];
      }
      if (m>= 1e-7) result /= m;
   }
   
   if (debugVis) {
      debugRenderer->beginRenderJob("HOC descriptors");
      Matrix3f frameRot;
      frameRot[0] = tangentU;
      frameRot[1] = tangentV;
      frameRot[2] = normal;

      debugRenderer->addLine(
         pos-tangentU*radius-tangentV*radius, 
         pos+tangentU*radius-tangentV*radius, 
         makeVector3f(0,0.5f,0), makeVector3f(0,0.5f,0), 2);
      debugRenderer->addLine(
         pos+tangentU*radius-tangentV*radius, 
         pos+tangentU*radius+tangentV*radius, 
         makeVector3f(0,0.5f,0), makeVector3f(0,0.5f,0), 2);
      debugRenderer->addLine(
         pos+tangentU*radius+tangentV*radius, 
         pos-tangentU*radius+tangentV*radius, 
         makeVector3f(0,0.5f,0), makeVector3f(0,0.5f,0), 2);
      debugRenderer->addLine(
         pos-tangentU*radius+tangentV*radius, 
         pos-tangentU*radius-tangentV*radius, 
         makeVector3f(0,0.5f,0), makeVector3f(0,0.5f,0), 2);

      Vector3f binPosV = upperLeft;
      card32 baseIndex = 0;
      
      for (card32 v=0; v<numSpatialBins; v++) {
         Vector3f binPosU = binPosV;
         for (card32 u=0; u<numSpatialBins; u++) {

            debugRenderer->addLine(
               binPosU, 
               binPosU+spatialBinSpacing*tangentU, 
               makeVector3f(0.5f,0,0), makeVector3f(0.5f,0,0), 1);
            debugRenderer->addLine(
               binPosU+spatialBinSpacing*tangentU, 
               binPosU+spatialBinSpacing*tangentU+spatialBinSpacing*tangentV, 
               makeVector3f(0.5f,0,0), makeVector3f(0.5f,0,0), 1);
            debugRenderer->addLine(
               binPosU+spatialBinSpacing*tangentU+spatialBinSpacing*tangentV, 
               binPosU+spatialBinSpacing*tangentV, 
               makeVector3f(0.5f,0,0), makeVector3f(0.5f,0,0), 1);
            debugRenderer->addLine(
               binPosU+spatialBinSpacing*tangentV, 
               binPosU, 
               makeVector3f(0.5f,0,0), makeVector3f(0.5f,0,0), 1);
            for (card32 oBin=0; oBin<numOrientationBins; oBin++) {
               float32 oBinAngle = (float32)oBin / (float32)numOrientationBins * (float32)M_PI;
               Vector3f startPoint = binPosU + frameRot*((XAXIS_VECTOR3F+YAXIS_VECTOR3F)*spatialBinSpacing*0.5f);
               float32 magnitude = result[baseIndex+oBin]*numOrientationBins*numSpatialBins*0.5f;
               Vector3f endPoint = startPoint + frameRot*((XAXIS_VECTOR3F*cos(oBinAngle) + YAXIS_VECTOR3F*sin(oBinAngle))*(magnitude*spatialBinSpacing));
               startPoint = startPoint - (endPoint-startPoint);
               debugRenderer->addLine(startPoint, endPoint, makeVector4f(1,1,0.3f,1),makeVector4f(1,1,0.7f,1), 1);
            }

            baseIndex += numOrientationBins;
            binPosU += spatialBinSpacing*tangentU;
         }      
         binPosV += spatialBinSpacing*tangentV;
      }
      
      debugRenderer->endRenderJob();
   }

   return valid;
}

void X4::PCHOCDescriptor::endComputation()
{
   if (query != NULL) delete query;
   query = NULL;
   upc = NULL;
   ps = NULL;
   positionAAT = NULL_AAT;
   curv2AAT = NULL_AAT;
}

