#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PCIAnimFeatures.h"
#include <windows.h>
#include <mmsystem.h>
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "ObjectClassProperty.h"
#include "SeparatorClassProperty.h"
#include "ClassMethods.h"
#include "SlippageFeatureDetector.h"
#include "GenericSwapAllOutOfCorePointCloud.h"
#include "PCCSurfaceTools.h"
#include "StandardSceneGraphIterator.h"
#include "SGObjectNode.h"
#include "SGRelativeTimeAnimationNode.h"
#include "Scene.h"
#include "StringHelper.h"
#include "UnstructuredInCorePointCloud.h"
#include "GeometricFeature.h"
#include "PointCloudTools.h"
#include "NormalEstimator.h"
#include "VertexArray.h"
#include "RenderingTool.h"
//---------------------------------------------------------------------------
using namespace X4;

IMPLEMENT_X4_CLASS( PCIAnimFeatures ,0)
{
	BEGIN_CLASS_INIT( PCIAnimFeatures );

	ADD_OBJECT_PROP( m_FeaturePointDetector, 0, FeaturePointDetector::getClass(), true);
	ADD_SEPARATOR("Rendering");
	ADD_BOOLEAN_PROP(m_ShowFeaturePoints,0);
	ADD_FLOAT32_PROP(m_FeatureSize,0);

	ADD_NOARGS_METHOD(preprocessAnimation);
	ADD_NOARGS_METHOD(precomputeFeatures);
	ADD_NOARGS_METHOD(createGroundTruthChannel);
}

PCIAnimFeatures::PCIAnimFeatures(void)
{
	m_FeaturePointDetector = new SlippageFeatureDetector;
	m_ShowFeaturePoints = true;
	m_FeatureSize = 0.25f;
	
	m_RenderingTool = new RenderingTool;
	m_RenderingTool->init();

}

PCIAnimFeatures::~PCIAnimFeatures(void)
{
	delete m_FeaturePointDetector;
	delete m_RenderingTool;
}

void PCIAnimFeatures::assign(const Object* obj, X4CopyContext *context)
{
	const PCIAnimFeatures * o = dynamic_cast<const PCIAnimFeatures*>(obj);
	x4Assert( o != NULL );

	copyObjectProperties(obj,this);
}

void X4::PCIAnimFeatures::precomputeFeatures()
{
	FeaturePointDetector * detector = m_FeaturePointDetector;
	bool firstFrame = true;

	// iterate over each scene object & process every point cloud of type GenericSwapAllOutOfCorePointCloud
	StandardSceneGraphIterator * sceneIt = new StandardSceneGraphIterator();
	sceneIt->clearAndSetup( getScene()->getRootNode(), getScene()->getRootState() );
	while( !sceneIt->atEnd() )
	{
		SGObjectNode * objNode = dynamic_cast<SGObjectNode*>(sceneIt->getNode());
		if( objNode )
		{
			GenericSwapAllOutOfCorePointCloud * gpc = dynamic_cast<GenericSwapAllOutOfCorePointCloud*>(objNode->getSceneObject());
			if( gpc )
			{
				UnstructuredInCorePointCloud * upc = dynamic_cast<UnstructuredInCorePointCloud*>(gpc->fetchPC());
				if( firstFrame )
				{
					// feature point detectors need to initialize some variables (e.g. median point distance)
					// it's sufficient here to use the first frame for initialization
					firstFrame = false;
					std::vector<UnstructuredInCorePointCloud*> clouds;
					clouds.push_back(upc);
					detector->begin(clouds);
				}

				// create in-core copy and compute features
				UnstructuredInCorePointCloud * upc_copy = (UnstructuredInCorePointCloud*)upc->copy();
				FeatureSet * fs = detector->detect(upc_copy);

				// copy attachment and attach to generic...pc
				if( fs )
				{
					FeatureSet * fs_copy = (FeatureSet*)fs->copy();
					gpc->getAttachments()->attachData(fs_copy);
				}
				delete upc_copy;

			}
		}
		sceneIt->next();
	}			
	if( !firstFrame )
		detector->end();
}

void X4::PCIAnimFeatures::preprocessAnimation()
{
	bool firstFrame = true;

	StandardSceneGraphIterator * sceneIt = new StandardSceneGraphIterator();
	sceneIt->clearAndSetup( getScene()->getRootNode(), getScene()->getRootState() );

	// setup destination node
	SGRelativeTimeAnimationNode * animNode = dynamic_cast<SGRelativeTimeAnimationNode*>(getSceneGraphNodeByName(getScene(), "root/prepro"));
	if( !animNode )
	{
		animNode = new SGRelativeTimeAnimationNode;
		animNode->setName("prepro");
		getScene()->getRootNode()->addChildNode(animNode);
	}


	// get a list of point cloud names 
	float medPointDistance = 0.000001f;
	std::vector<GenericSwapAllOutOfCorePointCloud*> inputPointClouds;
	std::vector<std::string> nodeNames;
	while( !sceneIt->atEnd() )
	{
		SGObjectNode * objNode = dynamic_cast<SGObjectNode*>(sceneIt->getNode());
		if( objNode )
		{
			GenericSwapAllOutOfCorePointCloud * gpc = dynamic_cast<GenericSwapAllOutOfCorePointCloud*>(objNode->getSceneObject());
			if( gpc )
			{
				inputPointClouds.push_back(gpc);
				nodeNames.push_back(objNode->getName());
			}
		}
		sceneIt->next();
	}			


	// preprocess
	for( unsigned pc_index=0;pc_index<inputPointClouds.size();pc_index++ )
	{
		GenericSwapAllOutOfCorePointCloud * gpc = inputPointClouds[pc_index];
		UnstructuredInCorePointCloud * upc = dynamic_cast<UnstructuredInCorePointCloud*>(gpc->fetchPC());
		UnstructuredInCorePointCloud * upc_copy = (UnstructuredInCorePointCloud*)upc->copy();
		if( firstFrame )
		{
			firstFrame = false;
			medPointDistance = getMedianPointDistance(upc_copy);
		}


		// estimate normals
		NormalEstimator ne;
		checkAttribute(upc_copy, "normal",3,VAD::DATA_FORMAT_FLOAT32);
		ne.estimateNormals(upc_copy, 40,medPointDistance * 3.5f,-1, ENVIRONMENT_EPSILON);

		// orientate normals
		Vector3f camPos = getScene()->getRootState()->staticState->camera.getPosition();
		VertexArray varr(upc_copy);
		for( unsigned i=0;i<varr.getNumElements();i++ )
		{
			Vector3f pos = varr.getPosition3f(i);
			Vector3f normal = varr.getNormal3f(i);
			if( normal * (camPos - pos) < 0 )
				normal *= -1;
			varr.setNormal3f(i,normal);
		}

		// resample
		PCCSurfaceTools st;
		PointCloud * resampledPC = st.resamplePC(upc_copy, medPointDistance * 4.0f);

		// store
		GenericSwapAllOutOfCorePointCloud * gpc_res = new GenericSwapAllOutOfCorePointCloud;
		std::string path = gpc->getFileName();
		std::string dir = removeFilenameFromPath(path);
		std::string filename = removePathFromFilename(path);
		gpc_res->clearAndSetup(resampledPC, dir + std::string("res_") + filename);
		SGObjectNode * resObjNode = new SGObjectNode;
		resObjNode->setName(std::string("res_") + nodeNames[pc_index]);
		resObjNode->clearAndSetup(gpc_res);
		animNode->addChildNode(resObjNode);

		delete upc_copy;
		delete resampledPC;
	}

}

void X4::PCIAnimFeatures::glDrawTool( GLContext *glContext )
{
	StandardSceneGraphIterator sIt;
	sIt.clearAndSetup( getScene()->getRootNode(), getScene()->getRootState(), SGDynamicState::tmRendering);
	while( !sIt.atEnd())
	{
		SGObjectNode * objNode = dynamic_cast<SGObjectNode*>(sIt.getNode());
		if( objNode )
		{
			SceneObject * so = objNode->getSceneObject();
			Matrix4f T = so->getTransformation();
			if( m_ShowFeaturePoints )
			{
				FeatureSet * fs = dynamic_cast<FeatureSet*>(so->getAttachments()->getData(FeatureSet::getDefaultName()) );
				if( fs )
				{
					m_RenderingTool->setRenderColor(0.9f,0.1f,0.1f);
					UnstructuredInCorePointCloud * upc = dynamic_cast<UnstructuredInCorePointCloud*>(so);
					float medPointDistance = 0;
					if( upc ) 
						medPointDistance = getCachedMedianPointDistance(upc);
					else
						medPointDistance = 0.03f;
					for( unsigned i=0;i<fs->m_Features.size();i++ )
					{
						GeometricFeature * f = fs->m_Features[i];
						bool badFeature = false;
						for( unsigned j=0;j<f->m_Signatures.size() && j<m_FeaturePointDetector->getNumLocalDescriptors();j++ )
							badFeature |= f->m_Signatures[j]->badFeature( m_FeaturePointDetector->getLocalDescriptor(j) );
						if( !badFeature)
						{
							Vector3f transformedPos = transformVector3f(T, f->m_Position);
							m_RenderingTool->renderSphere(transformedPos, medPointDistance * m_FeatureSize);
						}
					}
				}
			}
		}
		sIt.next();
	}


}

void X4::PCIAnimFeatures::createGroundTruthChannel()
{
	StandardSceneGraphIterator sIt;
	sIt.clearAndSetup( getScene()->getRootNode(), getScene()->getRootState(), SGDynamicState::tmAllNodes);
	PointSet *firstFrame = NULL;
	while( !sIt.atEnd())
	{
		SGObjectNode * objNode = dynamic_cast<SGObjectNode*>(sIt.getNode());
		if( objNode )
		{
			PointCloud * pc = dynamic_cast<PointCloud*>(objNode->getSceneObject());
			if( pc == NULL )
			{
				sIt.next();
				continue;
			}
			checkAttribute(pc,"groundTruth",3,VAD::DATA_FORMAT_FLOAT32);

			if( firstFrame == NULL )
			{
			   SceneObjectIterator *it = pc->createIterator(SOIT::CAP_BASIC_PC);
			   if (!it) continue;
			   const PointSet *ps = dynamic_cast<BasicPointCloudIterator*>(it)->lockBlockRead();
				firstFrame = (PointSet*)ps->copy();
				dynamic_cast<BasicPointCloudIterator*>(it)->unlock();
			}

			SceneObjectIterator * iter = pc->createIterator(SOIT::CAP_MODIFYABLE_PC|SOIT::CAP_BASIC_PC);
			BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
			ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(iter);

			unsigned index = 0;
			AAT posAAT = firstFrame->getAAT("position");
			AAT gtAAT = pc->getAAT("groundTruth");
			while( !bIt->atEnd())
			{
				mIt->set3f(gtAAT, firstFrame->get3f(index,posAAT));
				bIt->next();
				index++;
			}
			
		}
		sIt.next();
	}

	delete firstFrame;
	
}