//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include <windows.h>
#include <mmsystem.h>
#include "PCIPointCloudProcessing.h"
#include "DebugOutput.h"
#include "NumericalClassProperty.h"
#include "ObjectClassProperty.h"
#include "ClassMethods.h"
#include "ProgressWindow.h"
#include "StandardSceneGraphIterator.h"
#include "SGObjectNode.h"
#include "SGListNode.h"
#include "StringHelper.h"
#include "NormalEstimator.h"
#include "SamplingSpacingEstimator.h"
#include "SeparatorClassProperty.h"
#include "PointCloudTools.h"
#include "NormalUnifier.h"
#include "TensorVoting.h"
#include "MLSSmooth.h"
#include "HoppeMesher.h"
#include "CurvatureEstimator.h"
#include "TangentsEstimator.h"
#include "MonomialFitter.h"
//------------------------------------------------------------------------------
using namespace X4;
//------------------------------------------------------------------------------

static std::string getModeDescr( const card32 index )
{
	if ( index == ENVIRONMENT_KNN ) return "KNN";
	else if ( index == ENVIRONMENT_EPSILON ) return "Epsilon";
	else if ( index == ENVIRONMENT_TIMES_SAMPLING_SPACING ) return "x SamplingSpacing";
	else return "Undefined";
}

static std::string getMeshDistanceMode( const card32 index )
{
	if ( index == DISTANCE_CLOSEST_POINT ) return "Closest Point";
	else if ( index == DISTANCE_EPSILON ) return "Epsilon";
	else return "Undefined";
}

IMPLEMENT_X4_CLASS( PCIPointCloudProcessing, 0 )
{
	BEGIN_CLASS_INIT( PCIPointCloudProcessing )
	ADD_OBJECT_PROP( renderer, 0, PointCloudProcessingRenderer::getClass(), true )
	ADD_OBJECT_PROP( checkAtt, 0, PointCloudAttributes::getClass(), true )
	ADD_STRING_PROP( pcName, 0 )
	ADD_BOOLEAN_PROP( usePointSet, 0 )
	ADD_CARD32_PROP( knn, 0 )
	ADD_FLOAT32_PROP( epsilon, 0 )
	ADD_FLOAT32_PROP( timesSamplingSpacing, 0 )
	ADD_CARD32_PROP_UPDATE_DESCR( mode, 0, NULL, getModeDescr, 3, NULL)
	ADD_SEPARATOR( "Meshing" )
	ADD_CARD32_PROP( mcRes, 0 )
	ADD_CARD32_PROP_UPDATE_DESCR( meshDistanceMode, 0, NULL, getMeshDistanceMode, 2, NULL )
	ADD_SEPARATOR( "Tensor Voting" )
	ADD_BOOLEAN_PROP( saveTensors, 0 )
	ADD_NOARGS_METHOD( samplingSpacing )
	ADD_NOARGS_METHOD( normalsPCA )
	ADD_NOARGS_METHOD( unifyNormals )
	ADD_NOARGS_METHOD( unifyNormalsHoppe )
	ADD_NOARGS_METHOD( estimateTangents )
	ADD_NOARGS_METHOD( fitMonomial )
	ADD_NOARGS_METHOD( estimateCurvature )
	ADD_NOARGS_METHOD( mlsSmooth )
	ADD_NOARGS_METHOD( hoppeTriangulate )
	ADD_NOARGS_METHOD( tensorVoteInit )
	ADD_NOARGS_METHOD( tensorVote )
}

PCIPointCloudProcessing::PCIPointCloudProcessing()
{
	pcName = "root/data";
	usePointSet = true;
	renderer = new PointCloudProcessingRenderer();
	checkAtt = new PointCloudAttributes();

	// Normal estimator
	knn = 20;
	epsilon = 5.0f;	// Godzilla
	//epsilon = 0.1f;	// Sphere
	timesSamplingSpacing = 3.0f;
	mode = ENVIRONMENT_EPSILON;

	// Mesher
	mcRes = 20;
	meshDistanceMode = DISTANCE_CLOSEST_POINT;

	// Tensor Voting
	saveTensors = true;
}

PCIPointCloudProcessing::~PCIPointCloudProcessing()
{
	delete checkAtt;
	delete renderer;
}

void PCIPointCloudProcessing::samplingSpacing()
{
	PointCloud* pc = getPointCloud( scene, pcName );
	if ( usePointSet ){
		UnstructuredInCorePointCloud* upc = dynamic_cast<UnstructuredInCorePointCloud*>( pc );
		if ( upc != NULL ) SamplingSpacingEstimator::estimateSamplingSpacing( upc->getPointSet() );
	} else { 
		SamplingSpacingEstimator::estimateSamplingSpacing( pc );
	}
}

void PCIPointCloudProcessing::normalsPCA()
{
	PointCloud* pc = getPointCloud( scene, pcName );
	if ( usePointSet ){
		UnstructuredInCorePointCloud* upc = dynamic_cast<UnstructuredInCorePointCloud*>( pc );
		if ( upc != NULL ) NormalEstimator::estimateNormals( upc->getPointSet(), knn, epsilon, timesSamplingSpacing, mode );
	} else { 
		NormalEstimator::estimateNormals( pc, knn, epsilon, timesSamplingSpacing, mode );
	}
}

void PCIPointCloudProcessing::unifyNormals()
{
	PointCloud* pc = getPointCloud( scene, pcName );
	if ( usePointSet ){
		UnstructuredInCorePointCloud* upc = dynamic_cast<UnstructuredInCorePointCloud*>( pc );
		if ( upc != NULL ) NormalUnifier::unifyNormals( upc->getPointSet(), knn );
	} else { 
		NormalUnifier::unifyNormals( pc, knn );
	}
}

void PCIPointCloudProcessing::unifyNormalsHoppe()
{
	PointCloud* pc = getPointCloud( scene, pcName );
	UnstructuredInCorePointCloud* upc = dynamic_cast<UnstructuredInCorePointCloud*>( pc );
	if ( upc != NULL ) NormalUnifier::unifyNormalsHoppe( upc->getPointSet(), knn );
}

void PCIPointCloudProcessing::mlsSmooth()
{
	PointCloud* pc = getPointCloud( scene, pcName );
	if ( usePointSet ){
		UnstructuredInCorePointCloud* upc = dynamic_cast<UnstructuredInCorePointCloud*>( pc );
		if ( upc != NULL ) MLSSmooth::smooth( upc->getPointSet(), epsilon );
	} else { 
		MLSSmooth::smooth( pc, epsilon );
	}
}

void PCIPointCloudProcessing::estimateCurvature()
{
	PointCloud* pc = getPointCloud( scene, pcName );
	if ( usePointSet ){
		UnstructuredInCorePointCloud* upc = dynamic_cast<UnstructuredInCorePointCloud*>( pc );
		if ( upc != NULL ) CurvatureEstimator::estimateCurvature( upc->getPointSet(), epsilon );
	} else { 
		CurvatureEstimator::estimateCurvature( pc, epsilon );
	}
}

void PCIPointCloudProcessing::tensorVoteInit()
{
	PointCloud* pc = getPointCloud( scene, pcName );
	if ( usePointSet ){
		UnstructuredInCorePointCloud* upc = dynamic_cast<UnstructuredInCorePointCloud*>( pc );
		if ( upc != NULL ) TensorVoting::tensorVote( true, upc->getPointSet(), epsilon, saveTensors, knn, timesSamplingSpacing, mode,true );
	} else { 
		TensorVoting::tensorVote( true, pc, epsilon, saveTensors, knn, timesSamplingSpacing, mode );
	}
}

void PCIPointCloudProcessing::tensorVote()
{
	PointCloud* pc = getPointCloud( scene, pcName );
	if ( usePointSet ){
		UnstructuredInCorePointCloud* upc = dynamic_cast<UnstructuredInCorePointCloud*>( pc );
		if ( upc != NULL ) TensorVoting::tensorVote( false, upc->getPointSet(), epsilon, saveTensors, knn, timesSamplingSpacing, mode,true );
	} else { 
		TensorVoting::tensorVote( false, pc, epsilon, saveTensors, knn, timesSamplingSpacing, mode );
	}
}

void PCIPointCloudProcessing::estimateTangents()
{
	PointCloud* pc = getPointCloud( scene, pcName );
	if ( usePointSet ){
		UnstructuredInCorePointCloud* upc = dynamic_cast<UnstructuredInCorePointCloud*>( pc );
		if ( upc != NULL ) TangentsEstimator::estimateTangents( upc->getPointSet() );
	} else { 
		TangentsEstimator::estimateTangents( pc );
	}
}

void PCIPointCloudProcessing::fitMonomial()
{
	PointCloud* pc = getPointCloud( scene, pcName );
	if ( usePointSet ){
		UnstructuredInCorePointCloud* upc = dynamic_cast<UnstructuredInCorePointCloud*>( pc );
		if ( upc != NULL ) MonomialFitter::fitMonomial( upc->getPointSet(), epsilon );
	} else { 
		MonomialFitter::fitMonomial( pc, epsilon );
	}
}



void PCIPointCloudProcessing::hoppeTriangulate()
{
	PointCloud* pc = getPointCloud( scene, pcName );
	UnstructuredInCoreTriangleMesh* mesh = NULL;
	if ( usePointSet ){
		UnstructuredInCorePointCloud* upc = dynamic_cast<UnstructuredInCorePointCloud*>( pc );
		if ( upc != NULL ) mesh = HoppeMesher::triangulate( upc->getPointSet(), mcRes, epsilon, meshDistanceMode );
	} else { 
		mesh = HoppeMesher::triangulate( pc, mcRes, epsilon, meshDistanceMode );
	}

	if ( mesh != NULL ){
		std::string meshPCName = "root/hoppeMesh";
		if ( getPointCloud( scene, meshPCName ) == NULL ){
			addPointCloud( scene, mesh, meshPCName );
		} else {
			replacePointCloud( scene, meshPCName, mesh );
		}
	}
}

void PCIPointCloudProcessing::glDrawTool( GLContext *glContext )
{
	PointCloud* pc = getPointCloud( scene, pcName );
	renderer->render( epsilon, knn, pc );
}

void PCIPointCloudProcessing::keyUp( GeneralKey key )
{
	if ( key.getkeyAlphaNumKey() == 'A' ){
		renderer->incMonitorPoint();
	} else if ( key.getkeyAlphaNumKey() == 'B' ){
		renderer->decMonitorPoint();
	}
	redrawScene();
}

void PCIPointCloudProcessing::connectToSceneImpl(Scene *scene, OnSceneChangeCallback *onChange, QWidget *sceneEditorWidget, QWidget* toolWidget)
{
	checkAtt->setup( scene, &pcName );
	renderer->setup( scene, &pcName );
}