#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PCISHREC2010.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "UnstructuredInCoreTriangleMesh.h"
#include "PointCloudTools.h"
#include "StringHelper.h"
#include "ClassMethods.h"
#include "ManifoldCorrespondences.h"
#include "DiscreteManifold.h"
#include "UnstructuredInCoreTriangleMeshManifold.h"

#include "DebugRenderer.h"
//---------------------------------------------------------------------------
using namespace X4;

IMPLEMENT_X4_CLASS( PCISHREC2010 ,0)
{
	BEGIN_CLASS_INIT( PCISHREC2010 );
	ADD_STRING_PROP(nullPose,0);
	ADD_STRING_PROP(corrPose,0);
	ADD_STRING_PROP(outputFile,0);

	ADD_NOARGS_METHOD(writeSHRECOutput);
	ADD_NOARGS_METHOD(readCorrFile);
}

PCISHREC2010::PCISHREC2010(void)
{
	// 	nullPose = "root/data/0001.null.0.off";
	// 	corrPose = "root/data/0001.isometry.1.off";

	nullPose = "root/Manifolds/mf1";
	corrPose = "root/Manifolds/mf2";
	outputFile = "E:/aberner/tosca/output.corr";
	//outputFile = "E:/aberner/geodesics/example_correspondence/shape1.corr";
}

PCISHREC2010::~PCISHREC2010(void)
{
}

void PCISHREC2010::assign(const Object* obj, X4CopyContext *context)
{
	PCInteractionTool::assign(obj,context);
	const PCISHREC2010 * o = dynamic_cast<const PCISHREC2010*>(obj);
	x4Assert( o != NULL );

	copyObjectProperties(obj,this);
}

void X4::PCISHREC2010::writeSHRECOutput()
{

	std::vector<int32> correspondingVertices;

	//loading Manifolds
	UnstructuredInCoreTriangleMeshManifold* nullManifold = dynamic_cast<UnstructuredInCoreTriangleMeshManifold*>(getSceneObject(getScene(), nullPose));
	UnstructuredInCoreTriangleMeshManifold* corrManifold = dynamic_cast<UnstructuredInCoreTriangleMeshManifold*>(getSceneObject(getScene(), corrPose));

	if (!nullManifold || !corrManifold)
	{
		debugOutput << "writeSHRECOutput: manifold not found\n";
		return;
	}

	// loading Null pose
	UnstructuredInCoreTriangleMesh* nullMesh = nullManifold->getMesh();

	// loading corr pose
	UnstructuredInCoreTriangleMesh* corrMesh = corrManifold->getMesh();
	ManifoldCorrespondenceSet* mCorSet = dynamic_cast<ManifoldCorrespondenceSet*>(nullManifold->getAttachments()->getData(ManifoldCorrespondenceSet::getDefaultName()));

	if (!nullMesh || !corrMesh || !mCorSet)
	{
		debugOutput << "writeSHRECOutput: mesh or correspondences not found\n";
		return;
	}


	//read ManifoldCorrespondenceSet
	correspondingVertices.resize(nullMesh->getNumPoints(), -1);
	for (mpcard i = 0; i < mCorSet->getNumCorrespondences(); i++)
	{
		ManifoldCorrespondence* corr = mCorSet->getManifoldCorrespondence(i);
		pair<ManifoldPoint*,ManifoldPoint*> corrPair = corr->getCorrespondence();
		DiscreteManifoldPoint* pointA = dynamic_cast<DiscreteManifoldPoint*>(corrPair.first);
		DiscreteManifoldPoint* pointB = dynamic_cast<DiscreteManifoldPoint*>(corrPair.second);
		if (pointA && pointB)
		{
			int32 indexA = pointA->getIndex();
			int32 indexB = pointB->getIndex();
			correspondingVertices[indexA] = indexB;
		}
		else
		{
			// todo (no DiscreteManifoldPoint, this should not occur, because result is vertex to vertex)
			notImplemented();
		}
	}


	// write shrec format
	//Each file will contain correspondence to the null shape (000n.null.0), represented as pairs of matching points in 
	//barycentric coordinates, as text lines of the format: tyi uyi vyi wyi txi uxi vxi wxi, where: txi is a triangle in the 
	//transformed shape, uxi+vxi+wxi=1 are the non-negative barycentric weights in triangle txi; tyi is the 
	//corresponding triangle in the null shape, uyi+vyi+wyi=1 are the non-negative barycentric weights in 
	//triangle tyi; i=1,..., number of corresponding points. Example:
	//	68116 0.23998567 0.24572606 0.51428826 50375 0.33333333 0.33333333 0.33333333
	//The baricentric coordinates are the most generic representation for correspondences. 
	//If the participating algorithm is designed to find vertex-to-vertex correspondences, use the baricentric representation
	//to encode vertices (e.g., 68116 1 0 0).

	// => we use the 2nd "encode vertices" way

	BaricentricMeshCoordinates nullMeshCoords(nullMesh);
	BaricentricMeshCoordinates corrMeshCoords(corrMesh);

	debugOutput << corrPose << "\n--------------\n";

	QFile outTxt(outputFile.c_str());
	if (outTxt.open(QFile::WriteOnly | QFile::Truncate)) {
		QTextStream outTextStream(&outTxt);

		for (mpcard i = 0; i < correspondingVertices.size(); i++)
		{
			if (correspondingVertices[i] != -1)
			{
				// "writecorr.m" says: fprintf(fout, '%d %.8f %.8f %.8f %d %.8f %.8f %.8f\n', [tx(:) ux(:,1) ux(:,2) ux(:,3) ty(:) uy(:,1) uy(:,2) uy(:,3)]');
				std::string oneLine = corrMeshCoords.getShrecStringForVertex(correspondingVertices[i]) + " " + nullMeshCoords.getShrecStringForVertex(i) + "\n";
				debugOutput << oneLine;

				QString qstr(oneLine.c_str());
				outTextStream << qstr;

			}
		}
	}
}

void X4::PCISHREC2010::readCorrFile()
{
	QFile file(outputFile.c_str());
	if (!file.open(QIODevice::ReadOnly | QIODevice::Text))
		return;

#if 1

	//loading Manifolds
	UnstructuredInCoreTriangleMeshManifold* nullManifold = dynamic_cast<UnstructuredInCoreTriangleMeshManifold*>(getSceneObject(getScene(), nullPose));
	UnstructuredInCoreTriangleMeshManifold* corrManifold = dynamic_cast<UnstructuredInCoreTriangleMeshManifold*>(getSceneObject(getScene(), corrPose));

	if (!nullManifold || !corrManifold)
	{
		debugOutput << "readCorrFile: manifolds not found\n";
		return;
	}

	// loading Null pose
	UnstructuredInCoreTriangleMesh* nullMesh = nullManifold->getMesh();

	// loading corr pose
	UnstructuredInCoreTriangleMesh* corrMesh = corrManifold->getMesh();
#endif
#if 0

	// loading Null pose
	UnstructuredInCoreTriangleMesh* nullMesh = (UnstructuredInCoreTriangleMesh*)getPointCloud(getScene(), "root/pc1");

	// loading corr pose
	UnstructuredInCoreTriangleMesh* corrMesh = (UnstructuredInCoreTriangleMesh*)getPointCloud(getScene(), "root/pc2");

#endif
	debugRenderer->clearDebugData("loaded");
	debugRenderer->beginRenderJob("loaded");

	QTextStream in(&file);
	QString line = in.readLine();
	while (!line.isNull()) {

		QStringList lineList = line.split(" ", QString::SkipEmptyParts);
		//	68116 0.23998567 0.24572606 0.51428826 50375 0.33333333 0.33333333 0.33333333
		x4Assert(lineList.size() == 8);
		mpint indexA = lineList[0].toInt();
		Vector3f barA = makeVector3f(lineList[1].toFloat(), lineList[2].toFloat(), lineList[3].toFloat());
		mpint indexB = lineList[4].toInt();
		Vector3f barB = makeVector3f(lineList[5].toFloat(), lineList[6].toFloat(), lineList[7].toFloat());	

		Vector3f realA = BaricentricMeshCoordinates::getAbsoluteCoordinateOfBaricentricMeshCoord(indexA, barA, nullMesh);
		Vector3f realB = BaricentricMeshCoordinates::getAbsoluteCoordinateOfBaricentricMeshCoord(indexB, barB, corrMesh);

		debugRenderer->addLine(realA, realB);


		line = in.readLine();
	}

	debugRenderer->endRenderJob();

	

}


X4::BaricentricMeshCoordinates::BaricentricMeshCoordinates( UnstructuredInCoreTriangleMesh * nullMesh)
{
	//find a face for each vertex
	partOfFace.resize(nullMesh->getNumPoints(), -1);
	vertexABC.resize(nullMesh->getNumPoints(), -1);

	PointSet* tris = nullMesh->getTriangles();
	AAT IND = tris->getAAT("index");
	for (mpcard i = 0; i < tris->getNumEntries(); i++)
	{
		Vector3i ti = tris->get3i(i, IND);
		for (mpcard j = 0; j < 3; j++)
		{
			if (partOfFace[ti[j]]<0)
			{
				partOfFace[ti[j]] = i;
				vertexABC[ti[j]] = j;
			}
		}
	}

	for (mpcard i = 0;  i < partOfFace.size(); i++)
	{
		x4Assert(partOfFace[i]!= -1);
		x4Assert(vertexABC[i]!= -1);
	}
}

std::string X4::BaricentricMeshCoordinates::getShrecStringForVertex( mpcard index )
{
	x4Assert(index<partOfFace.size());
	int32 faceIndex = partOfFace[index];
	int32 vertexIndexInFace = vertexABC[index];
	std::string result = intToStr(faceIndex) ;
	for (mpcard i = 0; i < 3; i++)
	{
		if (vertexIndexInFace == i)
			result += " 1";
		else
			result += " 0";
	}

	return result;

}

X4::Vector3f X4::BaricentricMeshCoordinates::getAbsoluteCoordinateOfBaricentricMeshCoord( mpcard triangleIndex, Vector3f bCoords, UnstructuredInCoreTriangleMesh* mesh )
{
	PointSet* tris = mesh->getTriangles();
	AAT IND = tris->getAAT("index");
	Vector3i verticeIndex = tris->get3i(triangleIndex, IND);

	PointSet* ps = mesh->getPointSet();
	AAT POS = ps->getAAT("position");

	Vector3f resultPos = ps->get3f(verticeIndex[0], POS) *  bCoords[0] +  ps->get3f(verticeIndex[1], POS) * bCoords[1] +  ps->get3f(verticeIndex[2], POS) * bCoords[2];

	return resultPos;
}