//---------------------------------------------------------------------------
#include "StdAfx.h"
#include <windows.h>
#include <mmsystem.h>
//---------------------------------------------------------------------------
#include "PointCloudProcessingRenderer.h"
#include "NumericalClassProperty.h"
#include "ObjectClassProperty.h"
#include "ClassMethods.h"
#include "SeparatorClassProperty.h"
#include "HierarchicalKNNIterator.h"
#include "MathTools.h"
#include "PointCloudTools.h"
#include "Monomials2ndOrder.h"
#include "PointSet.inline.h"
//------------------------------------------------------------------------------
using namespace X4;
//------------------------------------------------------------------------------

IMPLEMENT_X4_CLASS( PointCloudProcessingRenderer, 0 )
{
	BEGIN_CLASS_INIT( PointCloudProcessingRenderer )
	ADD_OBJECT_PROP( rt, 0, RenderingTool::getClass(), true )
	ADD_BOOLEAN_PROP( showEpsilon, 0 )
	ADD_CARD32_PROP( monitorPoint, 0 )
	ADD_FLOAT32_PROP( renderScale, 0 )
	ADD_SEPARATOR( "Properties" )
	ADD_BOOLEAN_PROP( showNormals, 0 )
	ADD_BOOLEAN_PROP( showTangents, 0 )
	ADD_BOOLEAN_PROP( showMonomials, 0 )
	ADD_BOOLEAN_PROP( showTensor, 0 )
	ADD_BOOLEAN_PROP( showMeanCurvature, 0 )
	ADD_BOOLEAN_PROP( showGaussianCurvature, 0 )
	ADD_SEPARATOR( "Color Code" )
	ADD_FLOAT32_PROP( curvMin, 0 )
	ADD_FLOAT32_PROP( curvMax, 0 )
	ADD_NOARGS_METHOD( colorCode )
}

PointCloudProcessingRenderer::PointCloudProcessingRenderer()
{
	rt = new RenderingTool();
	rt->init();

	scene = NULL;
	pcName = NULL;
	
	monitorPoint = 278;
	renderScale = 1.0f;
	
	showEpsilon = false;
	showNormals = true;
	showTangents = true;
	showTensor = false;
	showMonomials = true;
	showMeanCurvature = false;
	showGaussianCurvature = false;

	// Color code
	curvMin = 0;
	curvMax = 1;
}

PointCloudProcessingRenderer::~PointCloudProcessingRenderer()
{
	delete rt;
}

void PointCloudProcessingRenderer::render(
	const float32 epsilon, 
	const card32 knn,
	PointCloud* pc )
{

	if ( pc == NULL ) return;

	AAT POSITION = pc->getAAT( "position" );
	AAT NORMAL;
	AAT TANGENT_U, TANGENT_V;
	AAT TENSOR_0, TENSOR_1;
	AAT COEFFS;
	AAT POLY_SCALE;
	AAT CURV_0, CURV_1;

	UnstructuredInCorePointCloud* upc = dynamic_cast<UnstructuredInCorePointCloud*>( pc );
	PointSet* ps = NULL;
	if ( upc != NULL ) ps = upc->getPointSet();

	if ( showNormals ){
		if ( pc->providesAttribute( "normal" ) ){
			NORMAL = pc->getAAT( "normal" ); 
		} else {
			return;
		}
	}

	if ( showTangents ){
		if ( pc->providesAttribute( "tangent_u" ) && pc->providesAttribute( "tangent_v" ) ){
			TANGENT_U = pc->getAAT( "tangent_u" ); 
			TANGENT_V = pc->getAAT( "tangent_v" ); 
		} else {
			return;
		}
	}

	if ( showMonomials ){
		if ( pc->providesAttribute( "normal" ) && pc->providesAttribute( "tangent_u" ) && 
			 pc->providesAttribute( "tangent_v" ) && pc->providesAttribute( "coeffs" ) && 
			 pc->providesAttribute( "polyScale" ) ){
			TANGENT_U = pc->getAAT( "tangent_u" ); 
			TANGENT_V = pc->getAAT( "tangent_v" ); 
			NORMAL = pc->getAAT( "normal" ); 
			COEFFS = pc->getAAT( "coeffs" ); 
			POLY_SCALE = pc->getAAT( "polyScale" ); 
		} else {
			return;
		}
	}

	if ( showTensor ){
		if ( pc->providesAttribute( "tensor_0" ) && pc->providesAttribute( "tensor_1" ) ){
			TENSOR_0 = pc->getAAT( "tensor_0" );
			TENSOR_1 = pc->getAAT( "tensor_1" );
		} else {
			return;
		}
	}

	if ( showMeanCurvature || showGaussianCurvature ){
		if ( pc->providesAttribute( "curv0" ) && pc->providesAttribute( "curv1" ) && 
			pc->providesAttribute( "normal" ) ){
			NORMAL = pc->getAAT( "normal" );
			CURV_0 = pc->getAAT( "curv0" );
			CURV_1 = pc->getAAT( "curv1" );
		} else {
			return;
		}
	}

	SceneObjectIterator *newIt = pc->createIterator( SOIT( SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC ) );
	BasicPointCloudIterator *basicIt = dynamic_cast<BasicPointCloudIterator*>(newIt);
	ModifyablePointCloudIterator *modIt = dynamic_cast<ModifyablePointCloudIterator*>(newIt);
	HierarchicalKNNIterator *knnIt = new HierarchicalKNNIterator( pc, 32, NULL );

	basicIt->reset();
	long pIndex = 0;
	while( !basicIt->atEnd() ){
		if ( ( monitorPoint == pIndex ) || ( monitorPoint < 0 )  ){
			const Vector3f& p = basicIt->get3f( POSITION );

			if ( showEpsilon ){
				rt->enableLighting();
				rt->setRenderColor( 1,0,0 );
				knnIt->setSeekPointAndReset( p );
				float32 sqrD = 0;
				Vector3f np = p;
				const float sqrEpsilon = sqr( epsilon );
				while( !knnIt->atEnd() && ( sqrD < sqrEpsilon ) ){
					rt->renderSphere( np, 0.1f * epsilon * renderScale );
					np = knnIt->get3f( POSITION );
					knnIt->next();
					sqrD = ( np - p ) * ( np - p );
				}
			}

			if ( showTensor ){
				rt->enableLighting();
				rt->setRenderColor( 0,0,1 );
				const Vector3f tensor_0 = basicIt->get3f( TENSOR_0 );
				const Vector3f tensor_1 = basicIt->get3f( TENSOR_1 );
				Matrix3f tensor;
				tensor[0][0] = tensor_0[0];
				tensor[0][1] = tensor_0[1];
				tensor[0][2] = tensor_0[2];
				tensor[1][1] = tensor_1[0];
				tensor[1][2] = tensor_1[1];
				tensor[2][2] = tensor_1[2];
				tensor[1][0] = tensor[0][1];
				tensor[2][0] = tensor[0][2];
				tensor[2][1] = tensor[1][2];

				Vector3f eigenValues;
				Matrix3f eigenVectors;
				tensor.computeEigenStructure( eigenValues, eigenVectors );

				rt->renderEllipse( p, eigenVectors, eigenValues * renderScale * epsilon );
			}

			if ( showNormals ){
				rt->disableLighting();
				rt->setRenderColor( 1,0,0 );
				const Vector3f normal = basicIt->get3f( NORMAL );
				rt->renderLine( p, p + normal * 0.5f * epsilon * renderScale );
			}

			if ( showTangents ){
				rt->disableLighting();
				const Vector3f tangent_u = basicIt->get3f( TANGENT_U );
				const Vector3f tangent_v = basicIt->get3f( TANGENT_V );
				rt->setRenderColor( 0,1,0 );
				rt->renderLine( p, p + tangent_u * 0.5f * epsilon * renderScale );
				rt->setRenderColor( 0,0,1 );
				rt->renderLine( p, p + tangent_v * 0.5f * epsilon * renderScale );
			}

			if ( showMonomials ){
				const Vector3f normal = basicIt->get3f( NORMAL );
				const Vector3f tangent_u = basicIt->get3f( TANGENT_U );
				const Vector3f tangent_v = basicIt->get3f( TANGENT_V );
				Matrix3f T;
				T[0] = tangent_u;
				T[1] = tangent_v;
				T[2] = normal;
				StaticVector<float32,6> coeffs;
				if ( ps != NULL ) coeffs = ps->get<float32,6>( pIndex, COEFFS );
				const float32 polyScale = basicIt->get1f( POLY_SCALE );
				const Vector2f domainMin = makeVector2f( -0.5f, -0.5f );
				const Vector2f domainMax = makeVector2f( 0.5f, 0.5f );
				const card32 resolution = 10;
				Vector3f* mesh = new Vector3f[(resolution + 1)*(resolution + 1)];
				float stepSizeX = fabs( domainMax[0] - domainMin[0] ) / resolution;
				float stepSizeY = fabs( domainMax[1] - domainMin[1] ) / resolution;
				float x, y;
				for ( int i = 0; i <= resolution; i++ ){
					for ( int j = 0; j <= resolution; j++ ){
						x = domainMin[0] + i * stepSizeX;
						y = domainMin[1] + j * stepSizeY;
						mesh[i*(resolution+1)+j] = ( T * makeVector3f( x, y, Monomials2ndOrder::evaluate( coeffs, makeVector2f( x, y ) ) ) * polyScale + p );
					}
				}
				// Compute vertices
				for ( int i = 0; i <= resolution; i++ ){
					for ( int j = 0; j <= resolution; j++ ){
						//renderSphere( mesh[i*(resolution+1)+j], 0.01f * s );
					}
				}
				// Compute vertices
				rt->disableLighting();
				rt->setRenderColor( makeVector3f( 0.5f, 0.5f, 0.5f ) );
				for ( int i = 0; i <= resolution; i++ ){
					for ( int j = 0; j <= resolution; j++ ){
						if ( i < resolution )
							rt->renderLine( mesh[i*(resolution+1)+j], mesh[(i+1)*(resolution+1)+j] );
						if ( j < resolution )
							rt->renderLine( mesh[i*(resolution+1)+j], mesh[i*(resolution+1)+j+1] );
					}
				}
				delete [] mesh;
			}

			if ( showMeanCurvature ){
				Vector3f normal = basicIt->get3f( NORMAL );
				const float32 curv0 = basicIt->get1f( CURV_0 );
				const float32 curv1 = basicIt->get1f( CURV_1 );
				const float32 meanCurvature = 0.5f * ( curv0 + curv1 ); 
				float32 r = 1;
				if ( meanCurvature > 1e-7 ) r = 1.0f / meanCurvature;
				rt->enableLighting();
				rt->renderSphere( p - normal * r, r );
			}

			if ( showGaussianCurvature ){
				Vector3f normal = basicIt->get3f( NORMAL );
				const float32 curv0 = basicIt->get1f( CURV_0 );
				const float32 curv1 = basicIt->get1f( CURV_1 );
				const float32 gaussianCurvature = curv0 * curv1; 
				float32 r = 1;
				if ( gaussianCurvature > 1e-7 ) r = 1.0f / gaussianCurvature;
				rt->enableLighting();
				rt->renderSphere( p - normal * r, r );
			}

		}
		basicIt->next();
		pIndex++;
	}

	delete newIt;
	delete knnIt;
}

void PointCloudProcessingRenderer::colorCode()
{
	if ( ( scene == NULL ) || ( pcName == NULL ) ) return;
	PointCloud* pc = getPointCloud( scene, *pcName );

	if ( pc == NULL ) return;
	if ( !pc->providesAttribute( "curv0" ) ) return;
	if ( !pc->providesAttribute( "curv1" ) ) return;
	if ( !pc->providesAttribute( "color" ) ) return;

	SceneObjectIterator* it = pc->createIterator( SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC );
	BasicPointCloudIterator* basicIt = dynamic_cast<BasicPointCloudIterator*>( it );
	ModifyablePointCloudIterator* modIt = dynamic_cast<ModifyablePointCloudIterator*>( it );

	AAT CURV_0 = pc->getAAT( "curv0" );
	AAT CURV_1 = pc->getAAT( "curv1" );
	AAT COLOR = pc->getAAT( "color" );

	basicIt->reset();
	while( !basicIt->atEnd() ){
		const float32 curv0 = basicIt->get1f( CURV_0 );
		const float32 curv1 = basicIt->get1f( CURV_1 );
		const float32 meanCurvature = 0.5f * ( curv0 + curv1 );
		const float32 gaussianCurvature = curv0 * curv1;

		float32 tau = ( meanCurvature - curvMin ) / ( curvMax - curvMin );
		if ( tau < 0 ) tau = 0;
		if ( tau > 1 ) tau = 1;

		Vector3f color = makeVector3f( 1,0,0 ) * tau + makeVector3f( 0,1,0 ) * ( 1.0f - tau );
		modIt->set3f( COLOR, color );
		basicIt->next();
	}

	delete it;
}