//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "Quaternion.h"
#include "DebugOutput.h"
#include "MathTools.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

Quaternion::Quaternion()
{
	v[0] = 1;
	v[1] = 0;
	v[2] = 0;
	v[3] = 0;
}

Quaternion::Quaternion( float32 q0, float32 q1, float32 q2, float32 q3 )
{
	v[0] = q0;
	v[1] = q1;
	v[2] = q2;
	v[3] = q3;
}

Quaternion Quaternion::operator*( const Quaternion& q ) const
{
	float tmp_00 = (v[3] - v[2]) * (q[2] - q[3]);
	float tmp_01 = (v[0] + v[1]) * (q[0] + q[1]);
	float tmp_02 = (v[0] - v[1]) * (q[2] + q[3]);
	float tmp_03 = (v[2] + v[3]) * (q[0] - q[1]);
	float tmp_04 = (v[3] - v[1]) * (q[1] - q[2]);
	float tmp_05 = (v[3] + v[1]) * (q[1] + q[2]);
	float tmp_06 = (v[0] + v[2]) * (q[0] - q[3]);
	float tmp_07 = (v[0] - v[2]) * (q[0] + q[3]);
	float tmp_08 = tmp_05 + tmp_06 + tmp_07;
	float tmp_09 = 0.5f * (tmp_04 + tmp_08);

	Quaternion ret;
	ret[0] = tmp_00 + tmp_09 - tmp_05;
	ret[1] = tmp_01 + tmp_09 - tmp_08;
	ret[2] = tmp_02 + tmp_09 - tmp_07;
	ret[3] = tmp_03 + tmp_09 - tmp_06;
	return ret;
}

Quaternion Quaternion::operator+( const Quaternion& q ) const
{
	Quaternion ret;
	ret[0] = v[0] + q[0];
	ret[1] = v[1] + q[1];
	ret[2] = v[2] + q[2];
	ret[3] = v[3] + q[3];
	return ret;
}

void Quaternion::operator+=( const Quaternion& q ) 
{
	v[0] += q[0];
	v[1] += q[1];
	v[2] += q[2];
	v[3] += q[3];
}

void Quaternion::operator/=( const float32 s ) 
{
	Quaternion ret;
	v[0] /= s;
	v[1] /= s;
	v[2] /= s;
	v[3] /= s;
}

Quaternion Quaternion::operator*( const float32& s ) const
{
	return Quaternion( v[0] * s, v[1] * s, v[2] * s, v[3] * s );
}

Vector3f Quaternion::rotate( const Vector3f& p ) const
{
	//return q_mult( q_mult ( q_inv(q), p), q);
	Quaternion qp;
	qp[0] = 0;
	qp[1] = p[0];
	qp[2] = p[1];
	qp[3] = p[2];
	Quaternion res = ( inverse() * qp ) * (*this);
	return makeVector3f( res[1], res[2], res[3] );
}

Quaternion Quaternion::inverse() const
{
	Quaternion inv;
	inv[0] = v[0];
	inv[1] = -v[1];
	inv[2] = -v[2];
	inv[3] = -v[3];
	return inv;
}

Quaternion Quaternion::copy() const
{
	Quaternion c;
	for ( unsigned int n = 0; n < 4; n++ ) c[n] = v[n];
	return c;
}

Quaternion Quaternion::fromRotMat(Matrix3f R)
{	

	/* Diebel 2006, Representing Attitude, p.15 */

	Quaternion q;

	R = R.transpose();

	if (R[1][1]>(-R[2][2]) && R[0][0]>(-R[1][1]) && R[0][0]>(-R[2][2]))
	{
		float32 den = sqrt(1+R[0][0]+R[1][1]+R[2][2]);
		q = Quaternion( 0.5f * den,
						0.5f * (R[1][2]-R[2][1]) / den,
						0.5f * (R[2][0]-R[0][2]) / den,
						0.5f * (R[0][1]-R[1][0]) / den );
	}
	else if (R[1][1]<(-R[2][2]) && R[0][0]>R[1][1] && R[0][0]>R[2][2])
	{
		float32 den = sqrt(1+R[0][0]-R[1][1]-R[2][2]);
		q = Quaternion( 0.5f * (R[1][2]-R[2][1]) / den,
						0.5f * den,
						0.5f * (R[0][1]+R[1][0]) / den,
						0.5f * (R[2][0]+R[0][2]) / den );
	}
	else if (R[1][1]>R[2][2] && R[0][0]<R[1][1] && R[0][0]<(-R[2][2]))
	{
		float32 den = sqrt(1-R[0][0]+R[1][1]-R[2][2]);
		q = Quaternion( 0.5f * (R[2][0]-R[0][2]) / den,
						0.5f * (R[0][1]+R[1][0]) / den,
						0.5f * den,
						0.5f * (R[1][2]+R[2][1]) / den );
	}
	else if (R[1][1]<R[2][2] && R[0][0]<(-R[1][1]) && R[0][0]<R[2][2])
	{
		float32 den = sqrt(1-R[0][0]-R[1][1]+R[2][2]);
		q = Quaternion( 0.5f * (R[0][1]-R[1][0]) / den,
						0.5f * (R[2][0]+R[0][2]) / den,
						0.5f * (R[1][2]+R[2][1]) / den,
						0.5f * den );
	}
	
	return q;
}

Matrix3f Quaternion::getRotMat() const
{
	// Diebel
	Matrix3f rot;
	rot[0][0] = v[0]*v[0] + v[1]*v[1] - v[2]*v[2] - v[3]*v[3];
	rot[1][0] = 2*(v[1]*v[2] + v[0]*v[3]);
	rot[2][0] = 2*(v[1]*v[3] - v[0]*v[2]);

	rot[0][1] = 2*(v[1]*v[2] - v[0]*v[3]);
	rot[1][1] = v[0]*v[0] - v[1]*v[1] + v[2]*v[2] - v[3]*v[3];
	rot[2][1] = 2*(v[2]*v[3] + v[0]*v[1]);

	rot[0][2] = 2*(v[1]*v[3] + v[0]*v[2]);
	rot[1][2] = 2*(v[2]*v[3] - v[0]*v[1]);
	rot[2][2] = v[0]*v[0] - v[1]*v[1] - v[2]*v[2] + v[3]*v[3];

	return rot;
}

/// Print to console
void Quaternion::println() const
{
	debugOutput << "(" << v[0] << " " << v[1] << " " << v[2] << " " << v[3] << ")\n"; debugOutput.flush();
}

void Quaternion::normalize()
{
	float32 n = sqrt( sqr(v[0]) + sqr(v[1]) + sqr(v[2]) + sqr(v[3]) );
	v[0] /= n;
	v[1] /= n;
	v[2] /= n;
	v[3] /= n;
}