/***************************************************************************
*   Copyright (c) 2010   Art Tevs                                         *
*                                                                         *
*   This library is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU Lesser General Public License as        *
*   published by the Free Software Foundation; either version 3 of        *
*   the License, or (at your option) any later version.                   *
*                                                                         *
*   This library is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU Lesse General Public License for more details.                    *
*                                                                         *
*   The full license is in LICENSE file included with this distribution.  *
***************************************************************************/

//---------------------------------------------------------------------------
#ifndef RANSAC_REGISTRATION__H
#define RANSAC_REGISTRATION__H

//---------------------------------------------------------------------------
#include "x3foundation_private.h"
#include "Persistent.h"
#include "DiscreteManifold.h"
#include "ImportanceSampling.h"

//---------------------------------------------------------------------------
namespace X4
{
	class FeatureSet;
	class ImportanceSampler;
	class FeatureCorrespondence;
	class FeatureCorrespondenceSet;

	/**
	* Solution returned acquired with ransac matching.
	**/
	class X3FOUNDATION_PRIVATE_API RansacSolution : public X4::Persistent
	{
		X4_CLASS(RansacSolution);

	public:
		float error;
		float geodeticsLength;
		unsigned numGeodetics;
		FeatureCorrespondenceSet* fcs;

		RansacSolution() : error(0), geodeticsLength(0), numGeodetics(0), fcs(NULL) {}
		RansacSolution(float _error, unsigned _numGeodetics, float _geodeticsLength, FeatureCorrespondenceSet* _fcs) : 
			error(_error), geodeticsLength(_geodeticsLength), numGeodetics(_numGeodetics), fcs(_fcs) {}

		inline bool operator<(const RansacSolution& b)
		{
			return error < b.error;
		}
		inline bool operator>(const RansacSolution& b)
		{
			return error > b.error;
		}
		inline bool operator==(const RansacSolution& b)
		{
			return fabs(error - b.error) < 1e-6;
		}

		static bool SortSolutionPredicate(const RansacSolution& a, const RansacSolution& b);
	};

	/**
	* Reimplementation of the CVPR'09 Algorithm based on ransac registration of two manifolds.
	**/
	class X3FOUNDATION_PRIVATE_API RansacRegistration : public Persistent
	{
		X4_CLASS(RansacRegistration)

	public:
		//! Return manifold point corresponding to a feature point
		virtual DiscreteManifoldPoint* getFeatureManifoldPoint(DiscreteManifold* manifold, FeatureSet* featureSet, mpcard featureIndex);

	public:
		FeatureSet* m_FeatureSet[2];
		DiscreteManifold*   m_Manifold[2];
		//ManifoldPointList* m_ManifoldPointList[2];

		//ImportanceSampler* m_InitialImportanceSampler;
		//ImportanceSampler* m_CandidateImportanceSampler;

		float m_Sigma;
		card32 m_maxIterations;
		card32 m_minNumberOfMatches;
		float32 m_minMatchError;
		card32 m_initialMatchCount;
		float32 m_maxSupportDifference;
		float32 m_supportByAtLeastPercent;
		float32 m_descriptorScoreFactor;
		int32 m_useSupportedOfNumGeodesics;
		float32 m_UncertaintyEpsilonFactor;

		bool m_ignoreSameInitialCorrespondence;
		bool m_putInitialMatchIntoSolution;
		bool m_bUseBestMatchWhenRansac;
		bool m_bSortByNumberOfCorrespondencies;
		bool m_bCheckMirroring;
		bool m_bLookForGoodFeatureWhenRansac;

		/**
		* Compute score for a potential correspondence. The score is computed by checking how well given
		* correspondence is suited to the already contained list of correspondences
		**/
		virtual float computeCorrespondenceScore(const std::vector<X4::FeatureCorrespondence*>& previousSet, FeatureCorrespondence* corr,
			float* geodeticLength = NULL, unsigned* supportedBy = NULL);

		/**
		* Compute error for given correspondence set. The error is computed on the manifold.
		**/
		float computeCorrespondenceError(FeatureCorrespondenceSet* fcs);

		/**
		* Compute e-score for a difference and a sigma. This score is used when ransacing
		**/
		inline float32 computeScore(float32 distance)
		{
			return exp(-(distance*distance) / (m_Sigma*m_Sigma));
		}

		/**
		* Find best feature point on the given manifold, which supports by the already computed correspondences well
		* @param side Look on the source side (0) or target side (1)
		* @return index of the feature point or -1 if nothing found
		**/
		int findBestFeaturePoint(const std::vector<X4::FeatureCorrespondence*>& previousSet, int side = 0);


	public:
		RansacRegistration();
		virtual ~RansacRegistration();

		typedef std::vector<RansacSolution> SolutionList;

		/** 
		* Set both manifolds and corresponding feature sets used for registration.
		**/
		void setManifold(DiscreteManifold* ms1, DiscreteManifold* ms2, FeatureSet* fs1, FeatureSet* fs2)
		{
			m_Manifold[0] = ms1;
			m_Manifold[1] = ms2;
			m_FeatureSet[0] = fs1;
			m_FeatureSet[1] = fs2;
		}
		std::pair<DiscreteManifold*,DiscreteManifold*> getManifold() { return std::pair<DiscreteManifold*,DiscreteManifold*>(m_Manifold[0], m_Manifold[1]); }
		std::pair<FeatureSet*,FeatureSet*> getFeatureSet() { return std::pair<FeatureSet*,FeatureSet*>(m_FeatureSet[0], m_FeatureSet[1]); }

		//! Set sigma used when computing scores
		void setSigma(float32 sigma) { m_Sigma = sigma; }
		float getSigma() const { return m_Sigma; }


		//! Callback implementation for initial Ransac guess.
		class InitialCallback
		{
			protected:
				float _sigma;
				ImportanceSampler sampler;
			public:
				InitialCallback(float sigma) : _sigma(sigma) {};
				virtual std::vector<FeatureCorrespondence*> operator()(DiscreteManifold* mfSide, FeatureSet* fsSide, FeatureCorrespondenceSet* fcs, unsigned number);
		};

		//! Callback when a solution has been chosen (user app can react on callback, i.e. rendering)
		class SolutionFoundCallback
		{
			public:
				virtual void foundSolution(RansacSolution* solution) = 0;
		};

		/**
		* Run ransac and find correspondences between both feature sets on the manifold.
		* If initial match callback is given, then use it, otherwise use a default one.
		**/
		SolutionList computeMatching(FeatureCorrespondenceSet* fcs, InitialCallback* callback = NULL, SolutionFoundCallback* foundCallback = NULL);

		/** 
		* Sort solution list.
		**/
		void sortSolutions(SolutionList& solution, bool sortByNumberOfCorrespondenciesToo);

		/** 
		* Setup internal structures. Call this to add some attachments to manifold required for proper work of the algorthm.
		* Call this also directly before useing ransac.
		**/
		void setupManifolds();

		/**
		* Compute stability for a point in respect to the other points
		**/
		float computeStability(DiscreteManifoldPoint* point, const std::vector<const std::vector<DVectorF>*> paths);

	};
};

#endif 
