//---------------------------------------------------------------------------
#include "stdafx.h"
//---------------------------------------------------------------------------
#include "ClassMethods.h"
#include "ObjectClassProperty.h"
#include "Registerizer.h"
#include "SeparatorClassProperty.h"
#include "NumericalClassProperty.h"
#include "BooleanClassProperty.h"
#include "DebugOutput.h"
#include <time.h>
#include "PointSet.inline.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------


std::string candSelectionModeDescr( const card32 index )
{
	if ( index == CANDIDATE_SELECTION_ALL ){
		return "All";
	} else if ( index == CANDIDATE_SELECTION_CLOSEST ){
		return "Closest";
	} else if ( index == CANDIDATE_SELECTION_MAXDIST ){
		return "Max distance";
	} else if ( index == CANDIDATE_SELECTION_SUBSAMPLING ){
		return "Subsampling";
	} else if ( index == CANDIDATE_SELECTION_CLOSEST_SUBSAMPLING ){
		return "Closest & Subsampling";
	} else {
		return "Undefined";
	}
}


std::string getSolverDesc( const card32 i )
{
	if ( i == SOLVER_GRADIENT_DESCENT ){
		return "Gradient Descent";
	} else if ( i == SOLVER_NEWTON ){
		return "Newton";
	} else if ( i == SOLVER_LINE_SEARCH ){
		return "Line Search";
	} else if ( i == SOLVER_NEWTON_LINE_SEARCH ){
		return "Newton with Line Search";
	} else return "Undefined";
}


std::string colorCodeDescr( const card32 index )
{
	if ( index == COLOR_CODE_NONE ){
		return "None";
	} else if ( index == COLOR_CODE_CANDIDATES ){
		return "Candidates";
	} else if ( index == COLOR_CODE_GRADIENT_STRENGTH ){
		return "Gradient strength";
	} else if ( index == COLOR_CODE_USED_POINTS ){
		return "Used points";
	} else {
		return "Undefined";
	}
}

IMPLEMENT_X4_ABSTRACT_CLASS( Registerizer, 4 )
{
	BEGIN_CLASS_INIT( Registerizer )
	ADD_SEPARATOR( "Registerizer" )
	ADD_CARD32_PROP_UPDATE_DESCR( solver, 4, NULL, getSolverDesc, 4, NULL )
	ADD_CARD32_PROP_UPDATE_DESCR( colorCode, 1, NULL, colorCodeDescr, 4, NULL )
	ADD_BOOLEAN_PROP( registerTranslation, 3 )
	ADD_BOOLEAN_PROP( registerRotation, 3 )
	ADD_OBJECT_PROP( T, 0, Transformation::getClass(), true );
	ADD_FLOAT32_PROP( stepSize, 0 )
	ADD_SEPARATOR( "Candidates" )
	ADD_CARD32_PROP_UPDATE_DESCR( candSelectionMode, 1, NULL, candSelectionModeDescr, 5, NULL )
	ADD_FLOAT32_PROP( candidatesRatio, 1 )
	ADD_CARD32_PROP( updateCandidatesInterval, 1 )
	ADD_FLOAT32_PROP( maxDistance, 2 )
	ADD_CARD32_PROP( subsampling, 2 )
	ADD_NOARGS_METHOD( selectRegisterCandidates )
}

Registerizer::Registerizer()
{
	stepSize = 0.1f;
	maxDistance = -1.0f;
	candidatesRatio = 0.33f;
	updateCandidatesInterval = 9;
	registerTranslation = true;
	registerRotation = true;
	ready = false;
	T = new Transformation();
	T->setMode( TRANSFORMATION_QUATERNIONS );
	ownT = true;
	solver = SOLVER_GRADIENT_DESCENT;
	colorCode = COLOR_CODE_NONE;
	candSelectionMode = CANDIDATE_SELECTION_ALL;
	reset();
	maxGrad = lastMaxGrad = 1;
	registerR = baseR = IDENTITY3F;
	registerT = baseT = NULL_VECTOR3F;
	basePS = NULL;
	registerPS = NULL;
}

Registerizer::~Registerizer()
{
	if ( ownT && ( T != NULL ) ){
		delete T;
		T = NULL;
	}
}

void Registerizer::setup( PointSet* basePS, const Matrix3f& baseR, const Vector3f& baseT )
{
	if ( basePS == NULL ){
		debugOutput << "setup: failed; invalid point sets.\n"; debugOutput.flush();
		ready = false;
		return;
	}
	this->basePS = basePS;
	this->baseR = baseR;
	this->baseT = baseT;
	BASE_POSITION = basePS->getAAT( "position" );
	numSolves = 1;
	ready = true;
}

float32 Registerizer::solve( const int32 numIterations )
{
	if ( ( registerPS == NULL ) || ( basePS == NULL ) ){
		debugOutput << "Invalid point sets.\n"; debugOutput.flush();
		return -1;
	}

	if ( !ready ){
		debugOutput << "solve: call setup() first!\n"; debugOutput.flush();
		return -1;
	}

	int32 iteration = 0;
	StaticMatrix<float32,7,7> hessian;
	StaticVector<float32,7> gradient;
	float32 score = -1;

	if( numSolves%updateCandidatesInterval == 0 ) selectRegisterCandidates();
	
	while( iteration < numIterations ){
		if (solver == SOLVER_NEWTON )
		{
			if ( supportsNewton() )
			{					
				score = computeHessian( hessian, gradient );

				// matrix inversion with opencv
				StaticMatrix<float32,7,7> invHessian;
				if ( !registerTranslation ) for ( unsigned int i = 0; i < 3; i++ ) gradient[i] = 0;
				if ( !registerRotation ) for ( unsigned int i = 3; i < 7; i++ ) gradient[i] = 0;				

				invHessian = invertMatrix( hessian );
				
				T->add( -invHessian * gradient );								
			} 
			else 
			{
				debugOutput << "registerizer does not support newton method.\n"; 
				debugOutput.flush();
				return -1;
			}
		} 
		else if (solver == SOLVER_GRADIENT_DESCENT )
		{
			if ( supportsGradientDescent() )
			{
				score = computeGradient( gradient );
				if ( !registerTranslation ) for ( unsigned int i = 0; i < 3; i++ ) gradient[i] = 0;
				if ( !registerRotation ) for ( unsigned int i = 3; i < 7; i++ ) gradient[i] = 0;
				T->add( gradient * stepSize );
			} 
			else 
			{
				debugOutput << "registerizer does not support gradient descent method.\n"; 
				debugOutput.flush();
				return -1;
			}
		} 
		else if (solver == SOLVER_LINE_SEARCH )
		{
			if ( supportsGradientDescent() )
			{
				// computeLineSerach calls computeGradient
				score = computeLineSearch( gradient, stepSize );
				if ( !registerTranslation || !registerRotation ) 
				{
					debugOutput << "registerizer does not support restriction to rotation/translation only in lineSearch mode!\n";
					debugOutput.flush();
					// T->add is done in computeLineSearch!
				}
			}
			else 
			{
				debugOutput << "registerizer does not support gradient descent method needed for line search.\n"; 
				debugOutput.flush();
				return -1;
			}
		}
		else if (solver == SOLVER_NEWTON_LINE_SEARCH )
		{
			if ( supportsNewton() )
			{
				// computeLineSerach calls computeGradient
				score = computeNewtonLineSearch( gradient, stepSize );
				if ( !registerTranslation || !registerRotation ) 
				{
					debugOutput << "registerizer does not support restriction to rotation/translation only in lineSearch mode!\n";
					debugOutput.flush();
					// T->add is done in computeNewtonLineSearch!
				}
			}
			else 
			{
				debugOutput << "registerizer does not support gradient descent method needed for line search.\n"; 
				debugOutput.flush();
				return -1;
			}
		}
		else 
		{
			debugOutput << "step: invalid solver.\n"; debugOutput.flush();
			return -1;
		}	
		iteration++;
		numSolves++;
		T->cleanup();
	}

	return score;
}

float32 Registerizer::computeLineSearch( StaticVector<float32,7>& gradient, float32 &stepSize )
{	
	float32 s = computeGradient( gradient ); // call to child class		
	/*
	debugOutput << "line search starting at: " << s << "\n";
	debugOutput.flush();
	*/			
	float32 increment = this->stepSize;
	const card32 numTrials = 100;

	bool increasing;
	float32 last_s;
	card32 i = 1;
	float32 inc_fac = 2.0f;
		
	do 
	{
		last_s = s;
		increment *= inc_fac;
		T->setState( gradient * increment );
		T->cleanup();
		s = computeScore( T->getRotation(), T->getTranslation() );
/*		
		debugOutput << "score: " << s << "\n";
		debugOutput.flush();
*/
		increasing = (s > last_s);
		i++;
	} while (increasing && i<numTrials);

	if (!increasing)
	{
		// one step back
		increment /= inc_fac;
		T->setState( gradient * increment );
		T->cleanup();
		s = last_s;
	}
/*
	debugOutput << "line search: " << i << " iterations (score = " << s << ").\n";
	debugOutput.flush();
*/
	return s;
}
/* LM-like approach
float32 Registerizer::computeNewtonLineSearch( StaticVector<float32,7>& gradient, float32 &stepSize )
{	
	StaticMatrix<float32,7,7> hessian;
	StaticMatrix<float32,7,7> inv_hessian;
	StaticVector<float32,7> step;
	
	float32 last_s = computeHessian( hessian, gradient );
	
	StaticMatrix<float32,7,7> diag_hessian;
	for (card32 i=0; i<7; i++)
	{
		for (card32 j=0; j<7; j++)
		{
			diag_hessian[i][j] = (i==j) ? hessian[i][j] : 0.0f;
		}
	}
	
	float32 increment = this->stepSize;
	float32 inc_fac = 10.0f;

	card32 iter = 1;
	const card32 numTrials = 100;

	bool increasing;		
	float32 s;
			
	do 
	{				
		inv_hessian = hessian + (diag_hessian*increment);

		inv_hessian = invertMatrix( inv_hessian );
		step = -inv_hessian * gradient;

		T->setState( step );
		T->cleanup();

		s = computeScore( T->getRotation(), T->getTranslation() );
		
		debugOutput << "increment: " << increment << " -> score: " << s << "\n";		
		debugOutput.flush();

		increasing = (s > last_s);		
	
		if (!increasing)
		{
			iter++;
			// one step back
			if (increment < 1.0f) increment *= inc_fac;
		}
		else
		{
			if (increment > 1e-6f) increment /= inc_fac;
		}

	} while (!increasing && i<numTrials);

	this->stepSize = increment;

	debugOutput << "line search: " << iter << " iterations (score = " << s << ").\n";
	debugOutput.flush();

	return s;
}
*/

float32 Registerizer::computeNewtonLineSearch( StaticVector<float32,7>& gradient, float32 &stepSize )
{	
	StaticMatrix<float32,7,7> hessian;
	float32 s = computeHessian( hessian, gradient );
	if (s < 0.0f)
	{
		return -1.0f;
	}
	hessian = invertMatrix( hessian );
	gradient = -hessian * gradient;

			
	float32 increment = this->stepSize;
	const card32 numTrials = 100;

	bool increasing;
	float32 last_s;
	card32 i = 1;
	float32 inc_fac = 2.0f;
		
	do 
	{
		last_s = s;
		increment *= inc_fac;
		T->setState( gradient * increment );
		T->cleanup();
		s = computeScore( T->getRotation(), T->getTranslation() );

		increasing = (s > last_s);
		i++;
	} while (increasing && i<numTrials);

	if (!increasing)
	{
		// one step back
		increment /= inc_fac;
		T->setState( gradient * increment );
		T->cleanup();
		s = last_s;
	}

	return s;
}

void Registerizer::printVector( StaticVector<float32,7> v )
{
	for ( unsigned int i = 0; i < 7; i++ ){
		debugOutput << v[i] << " ";
	}
	debugOutput << "\n";
	debugOutput.flush();
}

void Registerizer::printMatrix( StaticMatrix<float32,7,7> M )
{
	for ( unsigned int i = 0; i < 7; i++ ){
		for ( unsigned int j = 0; j < 7; j++ ){
			debugOutput << M[i][j] << " ";
		}
		debugOutput << "\n";
	}
	debugOutput << "\n";
	debugOutput.flush();
}

void Registerizer::selectRegisterCandidates()
{
	if ( registerPS == NULL ){
		debugOutput << "selectRegisterCandidates: invalid point sets.\n"; debugOutput.flush();
		return;
	}
	if ( candSelectionMode == CANDIDATE_SELECTION_ALL ){
		// Select all candidates
		for ( unsigned int pointIndex = 0; pointIndex < registerPS->getNumEntries(); pointIndex++ ){
			registerPS->set1i( pointIndex, REGISTER_FLAG, registerPS->get1i( pointIndex, REGISTER_FLAG ) | FLAG_REGISTER_CANDIDATE );
		}
		
	} else if ( candSelectionMode == CANDIDATE_SELECTION_SUBSAMPLING ){
		for ( unsigned int pointIndex = 0; pointIndex < registerPS->getNumEntries(); pointIndex++ ){
			if ( rand()%subsampling == 0 ){
				registerPS->set1i( pointIndex, REGISTER_FLAG, registerPS->get1i( pointIndex, REGISTER_FLAG ) | FLAG_REGISTER_CANDIDATE );
			} else {
				registerPS->set1i( pointIndex, REGISTER_FLAG, registerPS->get1i( pointIndex, REGISTER_FLAG ) & (0xFFFFFFFF ^FLAG_REGISTER_CANDIDATE) );
			}	
		}
	} else {
		debugOutput << "selectRegisterCandidates: invalid mode.\n"; debugOutput.flush();
	}

	// Mark candidates
	if ( colorCode == COLOR_CODE_CANDIDATES ){
		for ( unsigned int pointIndex = 0; pointIndex < registerPS->getNumEntries(); pointIndex++ ){
			int32 flag = registerPS->get1i( pointIndex, REGISTER_FLAG );
			if ( flag & FLAG_REGISTER_CANDIDATE ){
				registerPS->set3f( pointIndex, REGISTER_COLOR, makeVector3f( 0.15f, 0.85f, 0.15f ) );
			} else {
				registerPS->set3f( pointIndex, REGISTER_COLOR, makeVector3f( 0.85f, 0.15f, 0.15f ) );
			}
		}
	}
}

void Registerizer::reset()
{
	T->reset();
}

void Registerizer::setRegisterPS( PointSet* registerPS, const Matrix3f& registerR, const Vector3f& registerT )
{
	if ( registerPS == NULL ){
		debugOutput << "setRegisterPS: invalid point set.\n"; debugOutput.flush();
	}
	this->registerR = registerR;
	this->registerT = registerT;
	this->registerPS = registerPS;
	REGISTER_POSITION = registerPS->getAAT( "position" );
	REGISTER_COLOR = registerPS->getAAT( "color" );
	REGISTER_FLAG = registerPS->getAAT( "flags" );
}

void Registerizer::replaceTransformation( Transformation* T )
{
	if ( this->T != NULL ) delete this->T;
	this->T = T;
	ownT = false;
}