#ifndef RegisterizerH
#define RegisterizerH

#include "Persistent.h"
#include "Transformation.h"
#include "PointSetKNNQuery.h"

namespace X4{

#define SOLVER_GRADIENT_DESCENT 0					// Gradient descent
#define SOLVER_NEWTON 1								// Newton iteration
#define SOLVER_LINE_SEARCH 2						// Gradient descent with line search
#define SOLVER_NEWTON_LINE_SEARCH 3					// Newton iteration with line search

#define COLOR_CODE_NONE 0							// Leave point color
#define COLOR_CODE_CANDIDATES 1						// Paint candidate points
#define COLOR_CODE_GRADIENT_STRENGTH 2				// Strength of gradient
#define COLOR_CODE_USED_POINTS 3					// Strength of gradient

#define CANDIDATE_SELECTION_ALL 0					// Use all points
#define CANDIDATE_SELECTION_CLOSEST 1				// Use only the 'candidatesRatio' % closest points
#define CANDIDATE_SELECTION_MAXDIST 2				// Use all points within 'maxDistance'
#define CANDIDATE_SELECTION_SUBSAMPLING 3			// Use all points within 'maxDistance'
#define CANDIDATE_SELECTION_CLOSEST_SUBSAMPLING 4	// Subsample clostest points

#define FLAG_REGISTER_CANDIDATE 16	// Bit set: point used in registration process

struct Candidate{
	float32 dist;
	card32 index;
};

struct CandidateSort
{
	bool operator()( Candidate& rpStart, Candidate& rpEnd )
	{
		return rpStart.dist < rpEnd.dist;
	}
};

/// Debugging and visualization functionality for the registration tools
class X3FOUNDATION_PRIVATE_API Registerizer : public Persistent
{
	X4_ABSTRACT_CLASS( Registerizer ) 
	protected:
		card32 solver;						// Solver: Gradient descent vs. Newton iterations
		float32 stepSize;					// Gradient descent step size
		PointSet *registerPS, *basePS;		// Data
		bool ready;							// Point set was assigned properly
		Transformation* T;					// Transformation state
		card32 colorCode;					// Color coding
		float32 candidatesRatio;			/// Percentage of all register candidates with smallest error to be used
		card32 updateCandidatesInterval;	/// Update interval for candidate selection;
		card32 candSelectionMode;
		card32 subsampling;					// Subsampling of the registration point cloud; 1 means no subsampling
		float32 maxDistance;				// Above this distance towards closest point, registration points are discarded
		card32 numSolves;					// Internal: number of solves
		float32 maxGrad, lastMaxGrad;		// Used for gradient visualization
		Matrix3f registerR;					// Initial transformation of register point set
		Vector3f registerT;					// Initial transformation of register point set
		Matrix3f baseR;						// Initial transformation of base point set
		Vector3f baseT;						// Initial transformation of base point set
		bool registerTranslation;			// Include translation part
		bool registerRotation;				// Include rotation part
		bool ownT;							// Flag for own T object

		AAT BASE_POSITION;
		AAT REGISTER_POSITION;
		AAT REGISTER_COLOR;
		AAT REGISTER_FLAG;

		void printVector( StaticVector<float32,7> v );
		void printMatrix( StaticMatrix<float32,7,7> M );
		
	public:
		Registerizer();
		~Registerizer();

		float32 solve( const int32 numIterations );

		/// Return the overall transformation of the registration
		virtual Matrix4f getTransformation() const = 0;

		virtual float32 computeGradient( StaticVector<float32,7>& gradient ) = 0;
		float32 computeLineSearch( StaticVector<float32,7>& gradient, float32 &stepSize );
		float32 computeNewtonLineSearch( StaticVector<float32,7>& gradient, float32 &stepSize );
		virtual float32 computeHessian( StaticMatrix<float32,7,7>& hessian, StaticVector<float32,7>& gradient ) = 0;
		virtual float32 computeScore(Matrix3f rot, Vector3f t) = 0;
		/// R, t are initial transformations of basePS (data structure (octree/grid) is required correspondingly)
		virtual void setup( PointSet* basePS, const Matrix3f& baseR = IDENTITY3F, const Vector3f& baseT = NULL_VECTOR3F );
		/// R, t are initial transformations of registerPS, applied during optimization
		virtual void setRegisterPS( PointSet* registerPS, const Matrix3f& registerR = IDENTITY3F, const Vector3f& registerT = NULL_VECTOR3F );
		virtual inline bool supportsGradientDescent() = 0;
		virtual inline bool supportsNewton() = 0;		

		// if registerPS transformation changes between multiple calls of solve
		void updateRegisterTransformation(const Matrix3f &registerR, const Vector3f& registerT) { this->registerR = registerR; this->registerT = registerT; }		

		virtual void selectRegisterCandidates();

		inline bool isReady() const { return ready; }
		inline void setColorCode( const card32 code ){ this->colorCode = code; }
		inline void setCandidateSelectionMode( const card32 mode ){ candSelectionMode = mode; }
		inline void setCandidatesRatio( const float32 r ){ candidatesRatio = r; }
		inline void setSolver( const card32 solver ){ this->solver = solver; }
		inline void reset();
		inline PointSet* getRegisterPS() { return registerPS; }
		inline PointSet* getBasePS() { return basePS; }
		inline void setSubsampling( const card32 subsampling ) { this->subsampling = subsampling; }
		inline void setMaxDistance( const float32 maxDistance ){ this->maxDistance = maxDistance; }

		/*** MEMBER ACCESS ***/
		inline void setStepSize( const float32 stepSize ){ this->stepSize = stepSize; }
		inline void setRegisterTranslation( const bool b ) { registerTranslation = b; }
		inline void setRegisterRotation( const bool b ) { registerRotation = b; }
		inline void setUpdateCandidatesInterval( const card32 i ){ updateCandidatesInterval = i; }
		void replaceTransformation( Transformation* T );
		Transformation* getT(){ return T; };
};

}

#endif
