//---------------------------------------------------------------------------
#include "stdafx.h"
//---------------------------------------------------------------------------
#include "RegularGrid.h"
#include "ClassMethods.h"
#include "ObjectClassProperty.h"
#include "SeparatorClassProperty.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

IMPLEMENT_X4_CLASS( RegularGrid, 1 )
{
	BEGIN_CLASS_INIT( RegularGrid )
	ADD_SEPARATOR( "Regular Grid" )
	ADD_FLOAT32_PROP( tau, 1 )
	ADD_CARD32_PROP( res, 1 )
	ADD_VAR_ARRAY_PROP( cells, 0, new ObjectClassProperty( "cell" , 0, NULL, AbstractNDTCell::getClass(), true, NULL ) )
}

RegularGrid::RegularGrid()
{
	res = -1;
	tau = -1;
	cells.clear();
	cellMapping.clear();
	numCells = 0;
}

RegularGrid::~RegularGrid()
{
	clear();
}

void RegularGrid::clear()
{
	for ( unsigned int cellIndex = 0; cellIndex < cells.size(); cellIndex++ ){
		if ( cells[cellIndex] != NULL ) delete cells[cellIndex];
	}
	cells.clear();
	cellMapping.clear();
	numCells = 0;
}

void RegularGrid::setup(
	const BoundingBox3f& bbox, 
	const float32 normalizationS,
	const Vector3f& normalizationT,
	const card32 setupMode, 
	const float32 setupTau, 
	const card32 setupRes )
{
	clear();

	if ( res < 1 ){
		debugOutput << "setup: invalid resolution for grid.\n"; debugOutput.flush();
		return;
	}
	this->bbox = bbox;
	this->normalizationT = normalizationT;
	this->normalizationS = normalizationS;
	if ( setupMode == CREATE_GRID_FROM_RES ){
		res = setupRes;
		tau = ( bbox.upperCorner[0] - bbox.lowerCorner[0] ) / setupRes;
		numCells = res * res * res;
	} else if ( setupMode == CREATE_GRID_FROM_TAU ){
		res = bbox.getMaxSideLength() / setupTau;
		tau = bbox.getMaxSideLength() / float32( res );
		numCells = res * res * res;
	} else {
		debugOutput << "setup: invalid setup mode.\n"; debugOutput.flush();
	}
}

long RegularGrid::getIndex1D( const Vector3i& index3D ) const 
{
	long index1D = ( index3D[2] * res + index3D[1] ) * res + index3D[0]; 
	if ( ( index1D < 0 ) || ( index1D >= long( numCells ) ) ) { 
		//debugOutput << "Invalid index!\n"; debugOutput.flush();
		//debugOutput << "3D index:   " << index3D[0] << " " << index3D[1] << " " << index3D[2] << "\n";
		//debugOutput << "Resolution: " << res << "\n";
		//debugOutput.flush();
		return -1;
	}
	return index1D; 
}

Vector3i RegularGrid::getIndex3D( const Vector3f& p ) const
{
	Vector3i index3D;
	index3D[0] = ( p[0] - bbox.lowerCorner[0] ) / tau;
	index3D[1] = ( p[1] - bbox.lowerCorner[1] ) / tau;
	index3D[2] = ( p[2] - bbox.lowerCorner[2] ) / tau;
	if ( ( index3D[0] < 0 ) || ( index3D[0] >= long( res ) ) ) index3D[0] = -1;
	if ( ( index3D[1] < 0 ) || ( index3D[1] >= long( res ) ) ) index3D[1] = -1;
	if ( ( index3D[2] < 0 ) || ( index3D[2] >= long( res ) ) ) index3D[2] = -1;
	return index3D;
}

AbstractNDTCell* RegularGrid::getCell( const long index1D )
{ 
	if ( ( index1D < 0 ) || ( index1D > long( numCells ) ) ) return NULL;
	std::map<long,long>::const_iterator it = cellMapping.find( index1D );
	if ( it == cellMapping.end() ) return NULL;
	else return cells[it->second];
}

AbstractNDTCell* RegularGrid::getCell( const Vector3i& index3D )
{ 
	return getCell( getIndex1D( index3D ) );
}

AbstractNDTCell* RegularGrid::getCell( const Vector3f& p )
{ 
	return getCell( getIndex3D( p ) );
}

void RegularGrid::setCell( const long index1D, AbstractNDTCell* cell )
{ 
	std::map<long,long>::const_iterator it = cellMapping.find( index1D );
	if ( it != cellMapping.end() ){
		// Cell existed before
		cells[index1D] = cell; 
	} else {
		// Add new cell
		cellMapping.insert( LongPair( index1D, (card32)cells.size() ) );
		cells.push_back( cell );
	}
}

void RegularGrid::printStats()
{
	debugOutput << "Regular grid: #cells: " << mpcard( cells.size() ) << " cells; cellMapping entries: " << (mpcard)cellMapping.size() << ".\n"; 
	debugOutput.flush();
}

BoundingBox3f RegularGrid::computeCellBBox( const long index1D )
{
	Vector3i index3D = getIndex3D( index1D );
	return computeCellBBox( index3D );
}

BoundingBox3f RegularGrid::computeCellBBox( const Vector3i& index3D )
{
	return BoundingBox3f( 
		bbox.lowerCorner + makeVector3f( index3D[0] * tau, index3D[1] * tau, index3D[2] * tau ),
		bbox.lowerCorner + makeVector3f( (index3D[0]+1) * tau, (index3D[1]+1) * tau, (index3D[2]+1) * tau )
	);
}

Vector3i RegularGrid::getIndex3D( const long index1D )
{
	Vector3i index3D;
	index3D[2] = index1D / ( res * res );
	index3D[1] = ( index1D - res * res * index3D[2] ) / res;
	index3D[0] = index1D - res * res * index3D[2] - res * index3D[1];
	return index3D;
}