#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "RigidMotion.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include <opencv\cxcore.h>
#include "DebugRenderer.h"
#include "StringHelper.h"

#include "X4Exception.h"
//---------------------------------------------------------------------------
using namespace X4;


std::string describeMotionClass(card32 d)
{
	switch(d)
	{
	case 1:return "Translation";break;
	case 2:return "Rotation";break;
	case 3:return "Helix";break;
	}
	return "drunken monkey";
}

IMPLEMENT_X4_CLASS( RigidMotion ,1)
{
	BEGIN_CLASS_INIT( RigidMotion );
	ADD_MATRIX4F_PROP(m_Frame,0);
	ADD_MATRIX4F_PROP(m_InvertedFrame,0);
	ADD_FLOAT32_PROP(m_AngleVelocity,0);
	ADD_FLOAT32_PROP(m_SpatialVelocity,0);
	ADD_CARD32_PROP_UPDATE_DESCR(m_MotionClass,0,NULL,describeMotionClass,3, NULL);

	ADD_VECTOR3F_PROP(m_rotAxisCenter, 1);
}

RigidMotion::RigidMotion(void)
{
}

RigidMotion::~RigidMotion(void)
{
}

void RigidMotion::assign(const Object* obj, X4CopyContext *context)
{
	// Persistent::assign(obj,context);
	const RigidMotion * o = dynamic_cast<const RigidMotion*>(obj);
	x4Assert( o != NULL );

	copyObjectProperties(obj,this);
}

void X4::RigidMotion::setup( const Matrix4f &frame, float angleVel, float spatialVel )
{
	m_Frame = frame;
	m_AngleVelocity = angleVel;
	m_SpatialVelocity = spatialVel;
	if( m_AngleVelocity == 0 )
		m_MotionClass = TranslationMotion;
	else
		m_MotionClass = spatialVel == 0?RotationMotion:HelicalMotion;
	m_InvertedFrame = invertFrame(m_Frame);
}

bool X4::RigidMotion::setupFromTransformation( const Matrix4f &T, const Vector3f &referencePoint )
{
	// special test for translation patterns
	if( frobeniusNorm(shrink4To3(T)-IDENTITY3F) < 0.01f )
	{
		Matrix3f frame = calcTangentSystem(normalize(shrink4To3(T[3])));
		m_Frame = makeTranslation4f(referencePoint) * expand3To4(frame);
		m_AngleVelocity = 0;
		m_SpatialVelocity = norm(shrink4To3(T[3]));
		m_MotionClass = TranslationMotion;
	}
	else
	{
		// rigid motion with rotational part
		Matrix3f TR = shrink4To3(T);
		Vector3f Tt = shrink4To3(T[3]);

		// compute rigid motion
		// we search for a rotation axis A(t)=v*t+a0 
		// properties of x \in A(t): x = T*x = TR*x+Tt
		// leads to: Tt=A(t)-TR*A(t)=(I-TR)*x
		// --> x = inverse(I-TR) * Tt
		// we use a svd decomposition to find the rotation axis
		Matrix3f M = IDENTITY3F-TR;
		Matrix3f U,V;
		Vector3f evs;
		CvMat cvMatInput, cvMat_U, cvMat_W, cvMat_Vt;
		cvInitMatHeader( &cvMatInput, 3, 3, CV_32FC1, (void *)M.data());
		cvInitMatHeader( &cvMat_W, 3, 1, CV_32FC1, evs.data() );
		cvInitMatHeader( &cvMat_U, 3, 3, CV_32FC1, U.data() );
		cvInitMatHeader( &cvMat_Vt, 3, 3, CV_32FC1, V.data() );

		// Do SVD
		cvSVD(&cvMatInput, &cvMat_W, &cvMat_U, &cvMat_Vt, CV_SVD_U_T );

		const float zeroTolerance = 0.001f;
		Matrix3f diagInv = makeMatrix3f(
			((fabs(evs[0])>zeroTolerance)? 1.0f/evs[0] : 0), 0, 0, 
			0, ((fabs(evs[1])>zeroTolerance)? 1.0f/evs[1] : 0), 0,
			0, 0, ((fabs(evs[2])>zeroTolerance)? 1.0f/evs[2] : 0)
			);


		Matrix3f pseudo_inv = (V.transpose())*diagInv*(U.transpose());
		Vector3f rotAxisCenter = pseudo_inv.transpose()*(Tt);
		Vector3f rotAxis = U[2];

		Matrix3f tangentFrame = calcTangentSystem(rotAxis);
		Vector3f point0_tf = makeVector3f(1,0,0);
		Vector3f point1_w = transformVector3f(T,rotAxisCenter+tangentFrame*point0_tf);
		Vector3f point1_tf = tangentFrame.transpose() * (point1_w-rotAxisCenter);
		////////////////////////////
		// TODO: atan 2!!! or sign ..., ambiguous...
		///////////////////////////////
		m_AngleVelocity = acos(clamp(point0_tf*point1_tf,-0.999f, 0.999f));
		if( _isnan(m_AngleVelocity) || !_finite(m_AngleVelocity))
			debugOutput << "NAN velocity: acos(" << point0_tf*point1_tf << ")\nT=\n" << T << "\n";
		m_SpatialVelocity = point1_tf[2]-point0_tf[2];
		m_MotionClass = ( fabs(m_SpatialVelocity) < norm(point1_tf-point0_tf)*0.01f)?RotationMotion:HelicalMotion;
		if( isRotationMotion())
			m_SpatialVelocity = 0;


		// project reference point to rotation axis
		float projT = (referencePoint-rotAxisCenter)*rotAxis;
		Vector3f projRefPoint = rotAxisCenter + rotAxis*projT;

		if( m_MotionClass == RotationMotion )
		{
			tangentFrame[0] = normalize(referencePoint-projRefPoint);
			tangentFrame[1] = normalize(tangentFrame[2].crossProduct(tangentFrame[0]));
		}

		m_rotAxisCenter = projRefPoint;//rotAxisCenter;
		m_Frame = makeTranslation4f(projRefPoint) * expand3To4(tangentFrame);
	}

	// make sure we have a right hand oriented coordinate frame
	Matrix3f f = shrink4To3(m_Frame);
	if( f[0].crossProduct(f[1])*f[2] < 0 )
		m_Frame[1] *= -1.0f;
	m_InvertedFrame = invertFrame(m_Frame);
	return true;
}



void X4::RigidMotion::setupRotationFromTransformation( const Matrix4f &T, const Vector3f &referencePoint )
{
	// rigid motion with rotational part
	Matrix3f TR = shrink4To3(T);
	Vector3f Tt = shrink4To3(T[3]);

	// compute rigid motion
	// we search for a rotation axis A(t)=v*t+a0 
	// properties of x \in A(t): x = T*x = TR*x+Tt
	// leads to: Tt=A(t)-TR*A(t)=(I-TR)*x
	// --> x = inverse(I-TR) * Tt
	// we use a svd decomposition to find the rotation axis
	Matrix3f M = IDENTITY3F-TR;
	Matrix3f U,V;
	Vector3f evs;
	CvMat cvMatInput, cvMat_U, cvMat_W, cvMat_Vt;
	cvInitMatHeader( &cvMatInput, 3, 3, CV_32FC1, (void *)M.data());
	cvInitMatHeader( &cvMat_W, 3, 1, CV_32FC1, evs.data() );
	cvInitMatHeader( &cvMat_U, 3, 3, CV_32FC1, U.data() );
	cvInitMatHeader( &cvMat_Vt, 3, 3, CV_32FC1, V.data() );

	// Do SVD
	cvSVD(&cvMatInput, &cvMat_W, &cvMat_U, &cvMat_Vt, CV_SVD_U_T );

	const float zeroTolerance = 0.001f;
	Matrix3f diagInv = makeMatrix3f(
		((fabs(evs[0])>zeroTolerance)? 1.0f/evs[0] : 0), 0, 0, 
		0, ((fabs(evs[1])>zeroTolerance)? 1.0f/evs[1] : 0), 0,
		0, 0, ((fabs(evs[2])>zeroTolerance)? 1.0f/evs[2] : 0)
		);


	Matrix3f pseudo_inv = (V.transpose())*diagInv*(U.transpose());
	Vector3f rotAxisCenter = pseudo_inv.transpose()*(Tt);
	Vector3f rotAxis = U[2];

	Matrix3f tangentFrame = calcTangentSystem(rotAxis);
	Vector3f point0_tf = makeVector3f(1,0,0);
	Vector3f point1_w = transformVector3f(T,rotAxisCenter+tangentFrame*point0_tf);
	Vector3f point1_tf = tangentFrame.transpose() * (point1_w-rotAxisCenter);
	////////////////////////////
	// TODO: atan 2!!! or sign ..., ambiguous...
	///////////////////////////////
	m_AngleVelocity = acos(clamp(point0_tf*point1_tf,-0.999f, 0.999f));
	if( _isnan(m_AngleVelocity) || !_finite(m_AngleVelocity))
		debugOutput << "NAN velocity: acos(" << point0_tf*point1_tf << ")\nT=\n" << T << "\n";
	m_MotionClass = RotationMotion;
	m_SpatialVelocity = 0;


	// project reference point to rotation axis
	float projT = (referencePoint-rotAxisCenter)*rotAxis;
	Vector3f projRefPoint = rotAxisCenter + rotAxis*projT;

	tangentFrame[0] = normalize(referencePoint-projRefPoint);
	tangentFrame[1] = normalize(tangentFrame[2].crossProduct(tangentFrame[0]));

	m_Frame = makeTranslation4f(projRefPoint) * expand3To4(tangentFrame);

	// make sure we have a right hand oriented coordinate frame
	Matrix3f f = shrink4To3(m_Frame);
	if( f[0].crossProduct(f[1])*f[2] < 0 )
	m_Frame[1] *= -1.0f;
	m_InvertedFrame = invertFrame(m_Frame);
}

void X4::RigidMotion::draw( const Vector3f &p, const Vector3f &color, mpcard numSegments ) const
{
	// draw curve
	mpcard numSamplesPerSegment = 5;
	mpcard numTotalSamples = numSamplesPerSegment*numSegments+1;
	Vector3f p_param = parameterCoordinates(p);
	Vector3f lastSample;
	for( mpcard j=0;j<numTotalSamples;j++ )
	{
		float t = p_param[2] + (float)j / (float)numSamplesPerSegment;
		Vector3f sample = worldSpaceCoordinates(makeVector3f(p_param[0],p_param[1],t));
		if( j%numSamplesPerSegment == 0 )
			debugRenderer->addPoint(sample,color);
		if( j > 0)
			debugRenderer->addLine(sample,lastSample,color,color,1);
		lastSample = sample;
	}
}

X4::Vector3f X4::RigidMotion::parameterCoordinates( const Vector3f &p_worldSpace ) const
{
	Vector3f p_frame = transformVector3f(m_InvertedFrame,p_worldSpace);
	if( m_MotionClass != RotationMotion )
	{
		float t = p_frame[2] / m_SpatialVelocity;
		if( m_MotionClass == TranslationMotion )
		{
			// translation pattern case
			return makeVector3f(p_frame[0],p_frame[1],t);
		}
		
		// helical motion
		float angle = t * m_AngleVelocity;
		float sin_angle = sin(-angle);
		float cos_angle = cos(-angle);
		return makeVector3f(p_frame[0]*cos_angle-p_frame[1]*sin_angle,p_frame[0]*sin_angle+p_frame[1]*cos_angle,t);
	}
	else
	{
		// pure rotation
		float angle = atan2(p_frame[1],p_frame[0]);
		if( angle < 0)
			angle += 2.0f * (float)M_PI;
		angle/=m_AngleVelocity;
		return makeVector3f(sqrt(sqr(p_frame[0])+sqr(p_frame[1])),p_frame[2],angle);
	}

}

X4::Vector3f X4::RigidMotion::worldSpaceCoordinates( const Vector3f &p_parameterSpace ) const
{
	float angle = p_parameterSpace[2] * m_AngleVelocity;
	float cos_angle = cos(angle);
	float sin_angle = sin(angle);
	Vector3f p;
	if( m_MotionClass != RotationMotion )
		p = makeVector3f( p_parameterSpace[0]*cos_angle-p_parameterSpace[1]*sin_angle,
								p_parameterSpace[0]*sin_angle+p_parameterSpace[1]*cos_angle,
								p_parameterSpace[2] * m_SpatialVelocity);
	else
		p = makeVector3f( p_parameterSpace[0]*cos_angle,	p_parameterSpace[0]*sin_angle,
			p_parameterSpace[1]);
	return transformVector3f(m_Frame,p);
}

bool X4::RigidMotion::equivalentCoordinateFrames( const Matrix4f &frame0, const Matrix4f &frame1, float spatialTolerance, float angleTolerance ) const
{
	Vector3f origin0 = parameterCoordinates(shrink4To3(frame0[3]));
	Vector3f origin1 = parameterCoordinates(shrink4To3(frame1[3]));

	Vector2f diff = shrink3To2(origin1-origin0);
	if( normQuad(diff) > sqr(spatialTolerance))
		return false;

	if( m_MotionClass == TranslationMotion )
	{
		Matrix3f f0 = shrink4To3(frame0);
		Matrix3f f1 = shrink4To3(frame1);
		return (f0[0]*f1[0] >= 1.0f - angleTolerance) && (f0[1]*f1[1] >= 1.0f -  angleTolerance) && (f0[2]*f1[2] >= 1.0f - angleTolerance);
	}

	Matrix3f f0 = shrink4To3(frame0);
	Matrix3f f1 = makeRotVector3f( shrink4To3(m_Frame[2]), (origin0[2]-origin1[2])*m_AngleVelocity) * shrink4To3(frame1);

	float d1 = f0[0]*f1[0];
	float d2 = f0[1]*f1[1];
	float d3 = f0[2]*f1[2];

	return (d1 >= 1.0f - angleTolerance) && (d2 >= 1.0f -  angleTolerance) && (d3 >= 1.0f - angleTolerance);
}



bool X4::RigidMotion::equivalentPoints( const Vector3f &p0,const Vector3f &p1, float spatialTolerance )
{
	Vector3f origin0 = parameterCoordinates(p0);
	Vector3f origin1 = parameterCoordinates(p1);

	Vector2f diff = shrink3To2(origin1-origin0);
	return( normQuad(diff) < sqr(spatialTolerance));
}

X4::Matrix4f X4::RigidMotion::computeTransformation( float t0, float t1 ) const
{
	float deltaT = t1-t0;
	return m_Frame * makeTranslation4f(makeVector3f(0,0,deltaT*m_SpatialVelocity))*makeRotZ4f(-deltaT*m_AngleVelocity) * m_InvertedFrame;
}

bool X4::RigidMotion::isEqualMotion( const Matrix4f &T0,const Matrix4f &T1, const Vector3f &refPoint, float spatialTol, float angleTol )
{
	if( normQuad(transformVector3f(T0,refPoint)-transformVector3f(T1,refPoint)) > sqr(spatialTol))
		return false;

	Matrix3f R0 = shrink4To3(T0);
	Matrix3f R1 = shrink4To3(T1);
	return (R0[0]*R1[0] >= 1.0f - angleTol) && (R0[1]*R1[1] >= 1.0f - angleTol) && (R0[2]*R1[2] >= 1.0f - angleTol);
}

bool X4::RigidMotion::isEqualAxis( const RigidMotion * rm0,const RigidMotion * rm1,float spatialTol, float angleTol )
{
	if( rm0->m_MotionClass != rm1->m_MotionClass)
		return false;

	Vector3f axis0 = shrink4To3(rm0->m_Frame[2]);
	Vector3f axis1 = shrink4To3(rm1->m_Frame[2]);

	if( fabs(axis0*axis1) < 1.0f - angleTol )
		return false;

	if( rm0->isTranslationMotion() )
		return true;

	// for rotation/helical motions we need to check if both axis are collinear
	Vector3f refp0 = shrink4To3(rm0->m_Frame[3]);
	Vector3f refp1 = shrink4To3(rm1->m_Frame[3]);
	Vector3f x = refp0 + axis0* ( (refp1-refp0)*axis0 );

	return ( norm(x-refp1) < spatialTol );

}

bool X4::RigidMotion::isEqualAxisOrientationAware( const RigidMotion * rm0,const RigidMotion * rm1,float spatialTol, float angleTol )
{
	if( rm0->m_MotionClass != rm1->m_MotionClass)
		return false;

	Vector3f axis0 = shrink4To3(rm0->m_Frame[2]);
	Vector3f axis1 = shrink4To3(rm1->m_Frame[2]);

	if( (axis0*axis1) < 1.0f - angleTol )
		return false;

	if( rm0->isTranslationMotion() )
		return true;

	// for rotation/helical motions we need to check if both axis are collinear
	Vector3f refp0 = shrink4To3(rm0->m_Frame[3]);
	Vector3f refp1 = shrink4To3(rm1->m_Frame[3]);
	Vector3f x = refp0 + axis0* ( (refp1-refp0)*axis0 );

	return ( norm(x-refp1) < spatialTol );

}

void X4::RigidMotion::setupFromInstantaneousMotion( const Vector6f &inst_motion, float motionScale, const Vector3f &point )
{
	Vector3f trans = makeVector3f(inst_motion[3],inst_motion[4],inst_motion[5]);
	// easy case: pure translation
	if( norm(trans) > 0.99f )
	{
		setupFromTransformation( makeTranslation4f(trans*motionScale),point);
		return;
	}

	// motion includes rotation
	Vector3f rot = makeVector3f(inst_motion[0],inst_motion[1],inst_motion[2]);
	Matrix4f rotM = makeRotVector4f( normalize(rot), norm(rot));
	setupFromTransformation(makeTranslation4f(point)*rotM*makeTranslation4f(trans*motionScale)*makeTranslation4f(-point),point);
}

X4::Vector3f X4::RigidMotion::rotateVector( const Vector3f &p0, const Vector3f &dir, const Vector3f &targetPoint ) const
{
	Vector3f p0_pspace = parameterCoordinates(p0);
	Vector3f target_pspace = parameterCoordinates(targetPoint);

	Matrix4f T = computeTransformation(p0_pspace[2],target_pspace[2]);
	return shrink4To3(T)*dir;
}

X4::Matrix4f X4::RigidMotion::rotateFrame( const Vector3f &p0, const Matrix4f &frame, const Vector3f &targetPoint ) const
{
	Vector3f p0_pspace = parameterCoordinates(p0);
	Vector3f target_pspace = parameterCoordinates(targetPoint);

	Matrix4f T = computeTransformation(p0_pspace[2],target_pspace[2]);
	Matrix3f R = shrink4To3(T);
	if( !isTranslationMotion())
		target_pspace[0] = p0_pspace[0];

	return makeTranslation4f(worldSpaceCoordinates(target_pspace))*expand3To4(R*shrink4To3(frame));
}

float X4::RigidMotion::findMotion( const Vector3f &p, float dist ) const
{
	if( isTranslationMotion())
		return dist/m_SpatialVelocity;

	if( isRotationMotion() )
	{
		Vector3f p_pspace = parameterCoordinates(p);
		if( fabs(p_pspace[0]) < dist)
			return 1e20f;	// not precisely true , but does not matter for the application

		float arclength = atan(dist/p_pspace[0]);//dist*2.0f*M_PI/fabs(p_pspace[0]);
		return fabs(arclength/m_AngleVelocity);
	}

	// ToDo
	return dist/m_SpatialVelocity;
}

float X4::RigidMotion::computePathLength( const Vector3f &worldSpacePoint, float motionLength )
{
	Vector3f paramCoordsPoint = parameterCoordinates(worldSpacePoint);
	if( m_MotionClass==TranslationMotion )
	{
		paramCoordsPoint[2] += motionLength;
		return norm(worldSpacePoint - worldSpaceCoordinates(paramCoordsPoint));		
	}
	if( m_MotionClass==RotationMotion )
	{
		return motionLength * paramCoordsPoint[0] * m_AngleVelocity;		
	}
	throw X4Exception("Not implemented yet! RigidMotion::computePathLength() for helical motions");
}

X4::Vector3f X4::RigidMotion::computeMotionTangent( const Vector3f &p ) const
{
	if( m_MotionClass == TranslationMotion)
	{
		return shrink4To3(m_Frame[2]);
	}
	else
	{
		Vector3f p_param = parameterCoordinates(p);
		if( m_MotionClass == RotationMotion)
		{
			// numerical approx.
			p_param[2] += 0.01f;
			return normalize(worldSpaceCoordinates(p_param)-p);
		}
		else
		{
			// numerical approx.
			p_param[2] += 0.01f;
			return normalize(worldSpaceCoordinates(p_param)-p);
		}
	}
}

bool X4::RigidMotion::equivalentMotion( const RigidMotion * rm, float spatialTolerance, float angleTolerance/*=0.01f*/ ) const 
{
	if( !isEqualAxisOrientationAware(this,rm,spatialTolerance,angleTolerance))
		return false;
	if( fabs(m_SpatialVelocity-rm->m_SpatialVelocity) > spatialTolerance )
		return false;
	return( fabs(m_AngleVelocity-rm->m_AngleVelocity) <= angleTolerance );
}

void X4::RigidMotion::invertMotion()
{
	if( isTranslationMotion())
	{
		m_Frame[2]  *= -1;
		m_Frame[1]  *= -1;
	}
	else
	{
		m_AngleVelocity *= -1.0f;
	}
}

float X4::RigidMotion::getFullRotationLength() const
{
	if( !isRotationMotion())
		throw X4Exception("RigidMotion::getFullRotationLength() was called for non-rotational motion!");
	
	return 2.0f * M_PI / m_AngleVelocity;
}

void X4::RigidMotion::setNewReferencePoint( const Vector3f &p_worldSpace )
{
	if( isRotationMotion() )
	{
		Vector3f refPoint = shrink4To3(m_Frame[3]);
		Vector3f rotAxis = shrink4To3(m_Frame[2]);

		Vector3f dir0 = p_worldSpace-refPoint;
		dir0 = normalize(dir0 - rotAxis*(rotAxis*dir0));
		Vector3f dir1 = normalize(rotAxis.crossProduct(dir0));
		for( mpcard i=0;i<3;i++ )
		{
			m_Frame[0][i] = dir0[i];
			m_Frame[1][i] = dir1[i];
		}
	}
	else
	{
		Vector3f refPoint = shrink4To3(m_Frame[3]);
		Vector3f motionDir = shrink4To3(m_Frame[2]);
		refPoint += motionDir * (motionDir*(p_worldSpace-refPoint));
		m_Frame[3] = expand3To4(refPoint);
	}
	m_InvertedFrame = invertFrame(m_Frame);
}

std::string X4::RigidMotion::describe() const
{
	std::string ret;
	ret += describeMotionClass(m_MotionClass);
	ret += "Spatial: ";
	ret += floatToStr(m_SpatialVelocity);
	ret += "Angle: ";
	ret += floatToStr(m_AngleVelocity);
	ret += "frame: ";
	ret += m_Frame.toString();
	return ret;
}

X4::Vector3f X4::RigidMotion::parameterCoordinates_nullCentered( const Vector3f &p_worldSpace ) const
{
	Vector3f p_frame = transformVector3f(m_InvertedFrame,p_worldSpace);
	if( m_MotionClass != RotationMotion )
	{
		float t = p_frame[2] / m_SpatialVelocity;
		if( m_MotionClass == TranslationMotion )
		{
			// translation pattern case
			return makeVector3f(p_frame[0],p_frame[1],t);
		}

		// helical motion
		float angle = t * m_AngleVelocity;
		float sin_angle = sin(-angle);
		float cos_angle = cos(-angle);
		return makeVector3f(p_frame[0]*cos_angle-p_frame[1]*sin_angle,p_frame[0]*sin_angle+p_frame[1]*cos_angle,t);
	}
	else
	{
		// pure rotation
		float angle = atan2(p_frame[1],p_frame[0]);
/*		if( angle < 0)
			angle += 2.0f * M_PI;*/
		angle/=m_AngleVelocity;
		return makeVector3f(sqrt(sqr(p_frame[0])+sqr(p_frame[1])),p_frame[2],angle);
	}
}