#ifndef RigidMotionH
#define RigidMotionH
//---------------------------------------------------------------------------
#include "x3foundation_private.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
//---------------------------------------------------------------------------

namespace X4
{

/**
	represents a rigid motion, provides methods for construction, evaluation and inversion

	explanation of motion coordinates:
	a motion moves a 2d slice along the motion. therefore, we parameterize a motion with (r,s,t)
	where r,s are 2d slice coordinates and t is the motion parameter
*/
class X3FOUNDATION_PRIVATE_API RigidMotion : public Persistent
{
	X4_CLASS( RigidMotion )

private:

	Matrix4f	m_Frame;				//	fixed frame in which rigid motion lives
										//	for translation-only motions we use frame[2] as motion direction
	Matrix4f	m_InvertedFrame;
	float32		m_AngleVelocity;		//	motion rotates around frame[2] axis
	float32		m_SpatialVelocity;		//	motion moves in frame[2] direction

	Vector3f    m_rotAxisCenter;

	enum MotionClass
	{
		TranslationMotion=1,
		RotationMotion=2,
		HelicalMotion=3
	};

	card32		m_MotionClass;

public:
	RigidMotion(void);
	~RigidMotion(void);

   virtual void assign(const Object* obj, X4CopyContext *context = NULL);

   /// direct setup
   void setup(const Matrix4f &frame, float angleVel, float spatialVel);

   /// setup from rigid transformation (to avoid ambiguity we always take the lowest angle velocity)
   /// reference point defines origin of motion parameterization (t=0)
   /// returns true if transformation is a rigid motion
   bool setupFromTransformation(const Matrix4f &T, const Vector3f &referencePoint);
   void setupRotationFromTransformation(const Matrix4f &T, const Vector3f &referencePoint);

   /// setup from instantaneous motion (resulting e.g. from slippage analysis, motionScale = sigma)
   void setupFromInstantaneousMotion(const Vector6f &inst_motion, float motionScale, const Vector3f &point);

   /// set new reference point (only motion dimension will be modified)
   void setNewReferencePoint(const Vector3f &p_worldSpace);

   /// scale parameter speed
   void scaleSpeed(float s) {m_AngleVelocity *= s;m_SpatialVelocity *= s;}

   /// draw motion curve of point p with debug renderer
   void draw(const Vector3f &p, const Vector3f &color, mpcard numSegments) const;

   /// compute parameter coordinates for a point in world space
   inline Vector3f parameterCoordinates(const Vector3f &p_worldSpace) const;

   /// compute world space coordinates of a point in parameter space
   inline Vector3f worldSpaceCoordinates(const Vector3f &p_parameterSpace) const;

   /// compute parameter coordinates for a point in world space (for rotation: 
   inline Vector3f parameterCoordinates_nullCentered(const Vector3f &p_worldSpace) const;

   /// invert motion
   void invertMotion();

   /// compare with other rigid motion
   bool equivalentMotion( const RigidMotion * rm, float spatialTolerance, float angleTolerance=0.01f) const;

   /// test if two coordinate frames are equivalent with respect to rigid motion
   /// assumes that coordinate frames are right hand orthonormal coordinate frames without distortion
   bool equivalentCoordinateFrames(const Matrix4f &frame0, const Matrix4f &frame1, float spatialTolerance, float angleTolerance=0.01f) const;

   /// test if two points are equivalent w.r.t rigid motion (on the same curve)
   bool equivalentPoints(const Vector3f &p0,const Vector3f &p1, float spatialTolerance);

   /// computes world space transformation between two points on a motion path (only motion parameter required)
   Matrix4f computeTransformation(float t0, float t1) const;

   /// rotate std::vector located at p0 by rigid motion required to reach target point
   Vector3f rotateVector(const Vector3f &p0, const Vector3f &dir, const Vector3f &targetPoint) const;

   /// rotate frame by rigid motion  required to reach target point
   Matrix4f rotateFrame(const Vector3f &p0, const Matrix4f &frame, const Vector3f &targetPoint) const;

   /// find motion distance that moves p approximately dist in world space (returns delta t in parameter coordinates, t positive)
   /// return 1e20f if impossible (rotation motion near p)
   float findMotion(const Vector3f &p, float dist) const;

   bool isHelicalMotion() const {return m_MotionClass == HelicalMotion;}
   bool isTranslationMotion() const {return m_MotionClass == TranslationMotion;}
   bool isRotationMotion() const {return m_MotionClass == RotationMotion;}

   /// compare two rigid motions (spatial difference is measured w.r.t a reference point)
   static bool isEqualMotion(const Matrix4f &T0,const Matrix4f &T1, const Vector3f &refPoint, float spatialTol, float angleTol);

   /// compare rot/transl-axis of two rigid motions, does not care about axis orientation
   static bool isEqualAxis(const RigidMotion * rm0,const RigidMotion * rm1,float spatialTol, float angleTol);

   /// compare rot/transl-axis of two rigid motions including axis orientation
   static bool isEqualAxisOrientationAware(const RigidMotion * rm0,const RigidMotion * rm1,float spatialTol, float angleTol);

   /// returns local coordinate frame
   const Matrix4f& getFrame() const {return m_Frame;}
   const Matrix4f& getInvertedFrame() const {return m_InvertedFrame;}

   inline const Vector3f& getRotationAxisCenter() const { return m_rotAxisCenter; }
   inline void setRotationAxisCenter(const Vector3f& center) { m_rotAxisCenter = center; }

   card32 getMotionClass() const {return m_MotionClass;}

   /// compute world space length of motion path
   float computePathLength(const Vector3f &worldSpacePoint, float motionLength);

   /// get tangential motion std::vector at point p
   Vector3f computeMotionTangent(const Vector3f &p) const;

   /// for rotations only: compute path length for full rotation (2pi)
   float getFullRotationLength() const;

   // basic interface
   float getAngleVelocity() const {return m_AngleVelocity;}
   float getSpatialVelocity() const {return m_SpatialVelocity;}

   // describe motion
   std::string describe() const;


};

}

#endif