#include "stdafx.h"
//---------------------------------------------------------------------------
#include "SamplingSpacingEstimator.h"
#include "DebugOutput.h"
#include "HierarchicalKNNIterator.h"
#include "PointSetKNNQuery.h"
#include "PointSet.inline.h"
#include <time.h>
#include "NumericalClassProperty.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

void SamplingSpacingEstimator::estimateSamplingSpacing( PointSet* ps )
{
	if ( ps == NULL ){
		debugOutput << "estimateSamplingSpacing: invalid point set.\n"; debugOutput.flush();
		return;
	}

	if ( !ps->providesAttribute( "position" ) || !ps->providesAttribute( "samplingSpacing" ) ){
		debugOutput << "estimateSamplingSpacing: channels missing.\n"; debugOutput.flush();
		return;
	}

	float32 startTime = clock();

	AAT POSITION = ps->getAAT( "position" );
	AAT SAMPLING_SPACING = ps->getAAT( "samplingSpacing" );

	AttachedIndexedOctree* octree = PointSetKNNQuery::createOctree( ps, 32 );
	PointSetKNNQuery knn( ps, octree, makeVector3f( 0,0,0 ), POSITION );
	card32 nIndex;
	bool found;

	for ( unsigned int pIndex = 0; pIndex < ps->getNumEntries(); pIndex++ ){
		const Vector3f& p = ps->get3f( pIndex, POSITION );
		knn.setTargetPoint( p );
		knn.getNextNearestPointIndex( nIndex, found );	// Skip self
		float32 avgDistance = 0;
		for ( unsigned int n = 0; n < 8; n++ ){
			knn.getNextNearestPointIndex( nIndex, found );
			const Vector3f& np = ps->get3f( nIndex, POSITION );
			avgDistance = ( avgDistance * n + norm( np - p ) ) / ( float32( n+1 ) );
		}
		ps->set1f( pIndex, SAMPLING_SPACING, avgDistance );
	}

	debugOutput << "Time required (sampling spacing): " << float32( ( clock() - startTime ) / CLK_TCK ) << "\n"; debugOutput.flush();

	delete octree;
}

void SamplingSpacingEstimator::estimateSamplingSpacing( PointCloud* pc )
{
	if ( pc == NULL ){
		debugOutput << "estimateSamplingSpacing: invalid point cloud.\n"; debugOutput.flush();
		return;
	}

	if ( !pc->providesAttribute( "position" ) ||
		 !pc->providesAttribute( "samplingSpacing" ) ){
			debugOutput << "estimateSamplingSpacing: channels missing.\n"; debugOutput.flush();
			return;
	}

	float32 startTime = clock();

	AAT POSITION = pc->getAAT( "position" );
	AAT SAMPLING_SPACING = pc->getAAT( "samplingSpacing" );

	SceneObjectIterator* it = pc->createIterator( SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC );
	BasicPointCloudIterator* basicIt = dynamic_cast<BasicPointCloudIterator*>( it );
	ModifyablePointCloudIterator* modIt = dynamic_cast<ModifyablePointCloudIterator*>( it );
	HierarchicalKNNIterator* knnIt = new HierarchicalKNNIterator( pc, 32, NULL );

	basicIt->reset();
	long pIndex = 0;
	while( !basicIt->atEnd() ) {
		const Vector3f& p = basicIt->get3f( POSITION );
		knnIt->setSeekPointAndReset( p );
		knnIt->next();	// Skip self
		float32 avgDistance = 0;
		for ( unsigned int n = 0; n < 8; n++ ){
			knnIt->next();
			const Vector3f& np = knnIt->get3f( POSITION );
			avgDistance = ( avgDistance * n + norm( np - p ) ) / ( float32( n + 1 ) );
		}
		modIt->set1f( SAMPLING_SPACING, avgDistance );

		basicIt->next();
		pIndex++;
	}

	debugOutput << "Time required (sampling spacing): " << float32( ( clock() - startTime ) / CLK_TCK ) << "\n"; debugOutput.flush();

	delete it;
	delete knnIt;
}

float32 SamplingSpacingEstimator::estimateAvgSamplingSpacing( PointSet* ps )
{
	if ( ps == NULL ){
		debugOutput << "estimateAvgSamplingSpacing: invalid point set.\n"; debugOutput.flush();
		return -1;
	}

	if ( !ps->providesAttribute( "position" ) ){
		debugOutput << "estimateAvgSamplingSpacing: channels missing.\n"; debugOutput.flush();
		return -1;
	}

	float32 startTime = clock();

	AAT POSITION = ps->getAAT( "position" );

	AttachedIndexedOctree* octree = PointSetKNNQuery::createOctree( ps, 32 );
	PointSetKNNQuery knn( ps, octree, makeVector3f( 0,0,0 ), POSITION );
	card32 nIndex;
	bool found;
	float32 avgSamplingSpacing = 0;

	for ( unsigned int pIndex = 0; pIndex < ps->getNumEntries(); pIndex++ ){
		const Vector3f& p = ps->get3f( pIndex, POSITION );
		knn.setTargetPoint( p );
		knn.getNextNearestPointIndex( nIndex, found );	// Skip self
		float32 avgDistance = 0;
		for ( unsigned int n = 0; n < 8; n++ ){
			knn.getNextNearestPointIndex( nIndex, found );
			const Vector3f& np = ps->get3f( nIndex, POSITION );
			avgDistance = ( avgDistance * n + norm( np - p ) ) / ( float32( n+1 ) );
		}
		avgSamplingSpacing = ( avgSamplingSpacing * pIndex + avgDistance ) / float32( pIndex + 1 );
	}

	debugOutput << "Time required (sampling spacing): " << float32( ( clock() - startTime ) / CLK_TCK ) << "\n"; debugOutput.flush();

	delete octree;
	return avgSamplingSpacing;
}