#include "StdAfx.h"
//---------------------------------------------------------------------------
#include <windows.h>
#include <mmsystem.h>
#include "SlippageFeatureDetector.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "LocalICP.h"
#include "GeometricFeature.h"
#include "ObjectClassProperty.h"
#include "UnstructuredInCorePointCloud.h"
#include "HierarchicalKNNIterator.h"
#include "GeometricTools.h"
#include "DebugRenderer.h"
#include "Timer.h"
#include "StringHelper.h"
#include "Random.h"
#include "LogOutput.h"
#include "PointCloudTools.h"
#include "CachedRangeQuerry.h"
//---------------------------------------------------------------------------
using namespace X4;

IMPLEMENT_X4_CLASS( SlippageFeatureDetector ,3)
{
	BEGIN_CLASS_INIT( SlippageFeatureDetector );
	ADD_FLOAT32_PROP( m_Sigma0, 0);
	ADD_FLOAT32_PROP( m_SigmaSmoothFactor, 1);
	
	ADD_FLOAT32_PROP( m_ScaleBase, 1);
	ADD_CARD32_PROP( m_NumScales, 1);

	ADD_OBJECT_PROP(m_icp,0,LocalICP::getClass(),true);
	ADD_CARD32_PROP(m_MaxMeanshiftSteps,0);
	SET_USER_EDIT(false);
	ADD_FLOAT32_PROP(m_MeanShiftSigma,1);
	ADD_FLOAT32_PROP(m_Select_RelativeRadius,3);
	ADD_BOOLEAN_PROP(m_Debug_WriteScales,2);
}

IMPLEMENT_X4_CLASS( SlippageDescriptor, 3)
{
	BEGIN_CLASS_INIT( SlippageDescriptor );
	ADD_FLOAT32_PROP( m_MinUnslippability,0 );
	ADD_FLOAT32_PROP( m_MaxConvSteps, 1);
	ADD_FLOAT32_PROP( m_MinAverageUnslippability,2);
	ADD_FLOAT32_PROP( m_SlippageThreshold,3);
}

IMPLEMENT_X4_CLASS( SlippageSignature, 5)
{
	BEGIN_CLASS_INIT( SlippageSignature );
	ADD_FLOAT32_PROP( m_Unslippability,0 );
	ADD_FLOAT32_PROP(m_FeatureWeight,1);
	ADD_FLOAT32_PROP(m_ConvergenceSteps,2);
	ADD_FLOAT32_PROP(m_Scale,1);
	ADD_FLOAT32_PROP(m_AverageUnslippability,3);
	ADD_FLOAT32_FIXED_ARRAY_PROP(m_Slippage,4,6,3);
	ADD_FLOAT32_PROP(m_Sigma,5);
}


SlippageFeatureDetector::SlippageFeatureDetector(void)
{
	m_Sigma0 = 1.5f;
	m_SigmaSmoothFactor = 1.0f;
	m_ScaleBase = pow( 2.0f, 0.25f);
	m_NumScales = 6;
	m_MeanShiftSigma = 2.0f;
	m_Debug_WriteScales = false;
	m_Select_RelativeRadius = 1.0f;

	m_icp = new LocalICP;
	m_hIt = NULL;
	m_MaxMeanshiftSteps = 200;
	m_Descriptors.push_back( new SlippageDescriptor);
	m_PC = NULL;

}

SlippageFeatureDetector::~SlippageFeatureDetector(void)
{
	delete m_icp;
}

void SlippageFeatureDetector::assign(const Object* obj, X4CopyContext *context)
{
	const SlippageFeatureDetector * o = dynamic_cast<const SlippageFeatureDetector*>(obj);
	x4Assert( o != NULL );

	copyObjectProperties(obj,this);
}

FeatureSet * X4::SlippageFeatureDetector::detect( UnstructuredInCorePointCloud * upc, std::string attachmentName/*="" */ )
{
	m_NormalAAT = upc->getAAT("normal");
	bool outputDebugInfo = true;
	Timer overallTimer;

   // detect on multiscale and compute all signatures
	FeatureSet* fs = detectMultiScaleMS(upc,attachmentName);
   recomputeFeatureSignatures(upc,attachmentName );


   return fs;

#if 0   
	// create and attach new feature set, delete old one
	if(attachmentName == "")
		attachmentName = FeatureSet::getDefaultName();
	FeatureSet * ret = dynamic_cast<FeatureSet*>(upc->getAttachments()->getData(attachmentName));
	delete ret;
	ret = new FeatureSet;
	ret->setup( attachmentName, AttachedData::ADF_PERSISTENT );
	upc->getAttachments()->attachData(ret);

	UnstructuredInCorePointCloud * smoothedPointCloud = (UnstructuredInCorePointCloud*)upc->copy();
	std::vector<FeatureSet*> featursets;
	float meanShiftSigma = m_MeanShiftSigma * m_MedianPointDistance;

	for( unsigned scale=0;scale<m_NumScales;scale++ )
	{
		FeatureSet * fs = new FeatureSet;
		float slippage_sigma = m_Sigma0 * m_MedianPointDistance * pow( m_ScaleBase, (float)scale);
		float normal_sigma = slippage_sigma * m_SigmaSmoothFactor;

		// smooth point cloud
		Timer timer;
		smoothNormals( upc, smoothedPointCloud, "normal", normal_sigma);
		if( outputDebugInfo )
		{
			logOutput << "normal smoothing: "<< convertTimeToString(timer.getDeltaValue()) << "\n";
			debugOutput.flush();
		}
		// init
		beginPC(smoothedPointCloud);

		// compute slippage factor
		std::vector<float> slippage = computeSlippageFactor(smoothedPointCloud, slippage_sigma);
		if( outputDebugInfo )
		{
			logOutput << "slippage analysis: "<< convertTimeToString(timer.getDeltaValue()) << "\n";
			debugOutput.flush();
		}

		// cluster points
		std::vector<Vector3f> featurePoints;
		std::vector<float> unslippabilitys;
		//std::vector<unsigned> region = 
		gridBasedMeanShiftClustering( smoothedPointCloud, &slippage[0], meanShiftSigma, featurePoints,unslippabilitys );
		if( outputDebugInfo )
		{
			logOutput << "mean shift clustering: "<< convertTimeToString(timer.getDeltaValue()) << "\n";
			debugOutput.flush();
		}

		// get maximum slippage
		float maxUnslippability = 0;
		for( unsigned i=0;i<unslippabilitys.size();i++ )
			maxUnslippability = max( maxUnslippability, unslippabilitys[i]);

		// debug: write slippability in the color channel
		/*float maxS = 0;
		for( unsigned i=0;i<slippage.size();i++ )
			maxS = max( maxS, slippage[i]);
		PointSet * ps = upc->getPointSet();
		AAT colAAT = ps->getAAT("color");
		for( unsigned i=0;i<ps->getNumEntries();i++ )
		{
			ps->set3f(i,colAAT, makeVector3f(0,0.5f,1) * slippage[i] / maxS * 2.0f);
		}*/
		
		// store features
		m_icp->beginICP(upc,upc);
		m_icp->setSigma(slippage_sigma);
		for( unsigned i=0;i<featurePoints.size();i++ )
		{
			float minDist = 1e20f;
			unsigned bestIndex = 0;
			for( unsigned j=0;j<fs->m_Features.size();j++ )
			{
				minDist = min(minDist,norm(fs->m_Features[j]->m_Position - featurePoints[i] ) );
				if( minDist < 2.0f * m_MedianPointDistance )
				{
					bestIndex = j;
					break;
				}
			}
			if( minDist < 2.0f * m_MedianPointDistance )
			{
				SlippageSignature * sig = (SlippageSignature*)fs->m_Features[bestIndex]->m_Signatures[0];
				if( sig->m_Unslippability < unslippabilitys[i] )
				{
					delete fs->m_Features[bestIndex];
					fs->m_Features.erase( fs->m_Features.begin() + bestIndex);
				}
				else
					continue;
			}

			float unslippability = 0;//unslippabilitys[i];
			if( unslippabilitys[i] < 0.001f * maxUnslippability )
				continue;

			float32 numConvSteps = 1e20f;
			if( !testFeaturePointConvergence(featurePoints[i], &slippage[0], meanShiftSigma, unslippability, numConvSteps))
			{
				continue;
			}

			GeometricFeature * f = new GeometricFeature;
			f->m_Position = featurePoints[i];

			SlippageSignature * sig = new SlippageSignature;
			sig->m_Unslippability = unslippability;
			sig->m_AverageUnslippability = computeAverageUnslippabilityAndNormal(f->m_Position, &slippage[0],meanShiftSigma,f->m_Normal);
			sig->m_Scale = scale;
			sig->m_FeatureWeight = 0.0f;
			sig->m_ConvergenceSteps = numConvSteps;
			Matrix6f cov;
			m_icp->computeSlippage6d(f->m_Position, cov, sig->m_Slippage);

			f->m_Signatures.push_back(sig);
			fs->m_Features.push_back(f);
		}
		m_icp->endICP();

		// debug: store features in point seperate cloud
		/*UnstructuredInCorePointCloud * featurePC = new UnstructuredInCorePointCloud;
		VertexDescriptor * vd = new VertexDescriptor;
		vd->pushAttrib(VAD("position",3,VAD::DATA_FORMAT_FLOAT32));
		vd->pushAttrib(VAD("color",3,VAD::DATA_FORMAT_FLOAT32));
		vd->pushAttrib(VAD("flags",1,VAD::DATA_FORMAT_INT32));
		PointSet * features_ps = featurePC->getPointSet();
		features_ps->clearAndSetup(1,fs->getNumFeatures(),vd);
		AAT posAAT_ = features_ps->getAAT("position");
		AAT colorAAT_ = features_ps->getAAT("color");
		Vector3f color[8];
		color[0] = makeVector3f(1,0,0);
		color[1] = makeVector3f(1,1,0);
		color[2] = makeVector3f(0,1,0);
		color[3] = makeVector3f(0,1,1);
		color[4] = makeVector3f(0,0,1);
		color[5] = makeVector3f(1,0,1);
		color[6] = makeVector3f(0.5f,1,0);
		color[7] = makeVector3f(1,0.5f,0);
		for( unsigned i=0;i<features_ps->getNumEntries();i++ )
		{
			features_ps->set3f(i,posAAT_, fs->m_Features[i]->m_Position);
			features_ps->set3f(i,colorAAT_, color[scale] );
		}
		addPointCloud(sceneHack, featurePC, "features");*/

		if( outputDebugInfo )
		{
			logOutput << "feature pruning: "<< convertTimeToString(timer.getDeltaValue()) << "\n####################\n";
			debugOutput.flush();
		}

		featursets.push_back(fs);

		// cleanup
		endPC();
	}

	// combine multiple feature sets
	for( unsigned i=0;i<featursets.size();i++ )
	{
		FeatureSet * fs = featursets[i];
		for( unsigned j=0;j<fs->m_Features.size();j++ )
		{
			GeometricFeature * f = (GeometricFeature*)fs->m_Features[j]->copy();
			ret->m_Features.push_back(f);
		}
	}

	// compute overlap
	std::vector<bool> drop(ret->m_Features.size(),false);
	for( unsigned i=0;i<ret->m_Features.size();i++ )
	{
		for( unsigned j=i+1;j<ret->m_Features.size();j++ )
		{
			GeometricFeature * f1 = ret->m_Features[i];
			GeometricFeature * f2 = ret->m_Features[j];

			if( norm(f1->m_Position - f2->m_Position) < meanShiftSigma )
			{
				SlippageSignature * sig1 = (SlippageSignature*)f1->m_Signatures[0];
				SlippageSignature * sig2 = (SlippageSignature*)f2->m_Signatures[0];
				if( sig1->m_Scale != sig2->m_Scale )
				{
					sig1->m_FeatureWeight += 1.0f;
					sig2->m_FeatureWeight += 1.0f;
				}
				
				// ToDo: take feature with best convergence
				if( sig1->m_Unslippability/sig1->m_AverageUnslippability > sig2->m_Unslippability/sig2->m_AverageUnslippability )
				{
					drop[i] = true;
				}
				else
				{
					drop[j] = true;
				}
			}
		}
	}

	// drop features
	for( int i= int(ret->m_Features.size())-1; i>=0; i-- )
	{
		if( drop[i] )
		{
			ret->deleteFeatures(i,i+1);
		}
	}

	debugOutput << "slippage-features: " << convertTimeToString(overallTimer.getDeltaValue()) << "\n";
	return ret;
#endif
}


std::vector<float> X4::SlippageFeatureDetector::computeSlippageFactor( UnstructuredInCorePointCloud * upc, float sigma )
{
	PointSet * ps = upc->getPointSet();
	std::vector<float> ret(ps->getNumEntries());

	m_icp->setSigma(sigma);
	m_icp->beginICP(upc,upc);
	AAT normalAAT = ps->getAAT("normal");
	for( unsigned i=0;i<ps->getNumEntries();i++ )
	{
		Vector3f pos = ps->get3f(i,m_PosAAT);
		Vector3f normal = ps->get3f(i,normalAAT);

		Matrix6f covariance;
		Vector6f eigenValues;
		m_icp->computeSlippage6d(pos,covariance, eigenValues);
		ret[i] = eigenValues[5];
	}
	m_icp->endICP();

	return ret;
}

X4::Vector3f X4::SlippageFeatureDetector::computeMeanShiftPoint( const Vector3f &point, float * slippage, float sigma, float &summedWeight, bool smallCache )
{
	float radius_2 = pow(sigma*2.0f,2.0f);
	float inv_sigma_2 = 1.0f / pow( sigma, 2.0f);
	float overlap = m_MedianPointDistance * (smallCache?0.5f:2.0f);
	float rad_overlap_2 = pow(sigma*2.0f + overlap, 2.0f);
	float overlap_2 = pow(overlap,2.0f);

	if( m_ShiftCache.size() == 0 || normQuad(m_ShiftCacheLastPos - point) > overlap_2 )
	{
		m_ShiftCache.clear();
		m_ShiftCacheLastPos = point;
		m_hIt->setSeekPointAndReset(point);
		while( !m_hIt->atEnd())
		{
			Vector3f x = m_hIt->get3f(m_PosAAT);
			unsigned index = m_hIt->getCurrentPointIndex();
			float dist_2 = normQuad(point - x);
			if( dist_2 > rad_overlap_2)
				break;

			TMeanShiftVertex v;
			v.pos = x;
			v.slippage = slippage[index];
			v.index = index;
			m_ShiftCache.push_back(v);

			m_hIt->next();
		}

	}

	summedWeight = 0;
	Vector3f nextPos = NULL_VECTOR3F;
	const size_t numPoints = m_ShiftCache.size();
	TMeanShiftVertex * vertices = &m_ShiftCache[0];
	for( size_t i=0;i<numPoints;i++ )
	{
		const TMeanShiftVertex v = vertices[i];
		float dist_2 = normQuad(v.pos - point);
		float weight = exp( -dist_2 * inv_sigma_2) * v.slippage;
		summedWeight += weight;
		nextPos += v.pos * weight;
	}

	nextPos /= summedWeight;
	summedWeight /= pow(sigma,2.0f);
	return nextPos;

}

void X4::SlippageFeatureDetector::beginPC( UnstructuredInCorePointCloud * upc )
{
	m_hIt = new HierarchicalKNNIterator(upc,32,NULL);
	m_PosAAT = upc->getAAT("position");
	m_PC = upc;
}

void X4::SlippageFeatureDetector::endPC()
{
	delete m_hIt;
	m_hIt = NULL;
	m_PC = NULL;
}


std::vector<unsigned> X4::SlippageFeatureDetector::meanShiftClustering( UnstructuredInCorePointCloud * upc, float * slippage, float sigma, std::vector<Vector3f> &points,
																						  std::vector<float> &unslippablity)
{
	debugRenderer->beginRenderJob("testms");

	PointSet * ps = upc->getPointSet();
	std::vector<unsigned> region(ps->getNumEntries(),ps->getNumEntries());
	unsigned regionID = 0;
	std::map<unsigned,unsigned> mapRegion;
	AAT colAAT = ps->getAAT("color");
	for( unsigned i=0;i<ps->getNumEntries();i++ )
	{
		if( region[i] != ps->getNumEntries() )
			continue;

		Vector3f pos = ps->get3f(i,m_PosAAT);
		Vector3f initialPos = pos;
		regionID++;
		region[i] = regionID;
		mapRegion[regionID] = regionID;


		// iterative mean shifting until mode found or already marked closest point found
		bool modeFound = false;
		for( unsigned j=0;j<m_MaxMeanshiftSteps;j++ )
		{
			float summed_weight;
			Vector3f nextPos = computeMeanShiftPoint( pos, slippage, sigma, summed_weight);
			float stepLength = norm(nextPos - pos);
			if( stepLength < 0.05f * m_MedianPointDistance)
			{
				modeFound = true;
				points.push_back( nextPos );
				unslippablity.push_back( summed_weight );
				break;
			}

			const int numSteps = stepLength / m_MedianPointDistance / 2.0f + 1;
			Vector3f stepDir = (nextPos - pos);
			Vector3f stepDirNormed = normalize(stepDir);
			for( int x=0;x<numSteps;x++ )
			{
				Vector3f tmpPos = pos + stepDir * (float)(x+1) / (float)numSteps;
				//debugRenderer->addPoint(tmpPos, makeVector3f(0,1,1));
				m_hIt->setSeekPointAndReset(tmpPos);
				bool stop = false;
				while(!stop)
				{
					unsigned cp_index = m_hIt->getCurrentPointIndex();
					Vector3f closestPoint = ps->get3f(cp_index, m_PosAAT);
					float dist = norm( closestPoint - tmpPos);
					if( dist < 1.0f * m_MedianPointDistance &&
						 norm(tmpPos + stepDirNormed * (stepDirNormed * ( closestPoint - tmpPos)) - closestPoint) < 0.45f * m_MedianPointDistance
						)
					{
						if( region[cp_index] != ps->getNumEntries() && region[cp_index] != regionID )
						{
							mapRegion[regionID] = mapRegion[region[cp_index]];
							//ps->set3f(cp_index, colAAT, makeVector3f(0,1,0));
							stop = true;
							break;	// we already visitied this point
						}

						region[cp_index] = regionID;
						//ps->set3f(cp_index, colAAT, makeVector3f(1,0,0));
					}
					if( dist > m_MedianPointDistance )
						break;
					m_hIt->next();
				}
				if( stop )
					break;
			}

			//debugRenderer->addLine(pos,nextPos,makeVector3f(0.25f,0.25f,0),makeVector3f(0.25f,0.25f,0),1);
			pos = nextPos;
		}
	}
	debugRenderer->endRenderJob();

	std::map<unsigned,Vector3f> regionColors;
	for( unsigned i=0;i<region.size();i++ )
	{
		region[i] = mapRegion[region[i]];
		if( regionColors.find(region[i]) == regionColors.end())
			regionColors[region[i]] = makeVector3f(rnd01(), rnd01(), rnd01());

		ps->set3f(i,colAAT,regionColors[region[i]]);
	}

	return region;

	// start by assigning a "mean shift topology"
	/*HierarchicalKNNIterator hIt(upc,32,NULL);
	FixedKNNTopologyIterator topIt(upc->getPointSet(), upc->getAttachments(), upc, 10);
	std::map<unsigned,std::vector<unsigned> > edgeMap;
	PointSet * ps = upc->getPointSet();
	for( unsigned i=0;i<ps->getNumEntries();i++ )
	{
		Vector3f pos = ps->get3f(i,m_PosAAT);
		Vector3f initialPos = pos;

		// get topological neighbours
		std::vector<Vector3f> topNeighbours;
		std::vector<unsigned> topNeighbourIndices;
		topIt.goToIndex(i);
		for( unsigned j=0;j<topIt.getNumNeighbours();j++)
		{
			unsigned n = topIt.getNeighbour(j);
			topNeighbourIndices.push_back(n);
			topNeighbours.push_back(ps->get3f(n,m_PosAAT));
		}

		// keep best topological neighbour in mind
		float bestNeighbourDistance = 1e20f;
		unsigned bestNeighbourIndex = topNeighbourIndices[0];

		// iterative mean shifting until clusterpoint becomes close to top. neighbour
		for( unsigned j=0;j<m_MaxMeanshiftSteps;j++ )
		{
			float summed_weight;
			Vector3f nextPos = computeMeanShiftPoint( pos, slippage, sigma, summed_weight);

			// get closest neighbour
			for( unsigned n=0;n<topNeighbours.size();n++ )
			{
				float dist = norm(nextPos - topNeighbours[n]);
				if( dist < bestNeighbourDistance )
				{
					bestNeighbourDistance = dist;
					bestNeighbourIndex = topNeighbourIndices[n];
				}
			}

			pos = nextPos;
			if( bestNeighbourDistance < m_MedianPointDistance * 0.2f || norm(initialPos - pos) > m_MedianPointDistance * 1.5f || j == m_MaxMeanshiftSteps-1)
			{
				edgeMap[bestNeighbourIndex].push_back(i);
				edgeMap[i].push_back(bestNeighbourIndex);
				break;
			}
		}
	}

	// assign a regionID to every topological cluster ("mean shift topology")
	unsigned regionID = 0;
	std::vector<unsigned> region( ps->getNumEntries(), 0);
	std::vector<Vector3f> regionCentroids;
	for( unsigned i=0;i<ps->getNumEntries();i++ )
	{
		if( region[i] != 0)
			continue;

		regionID++;
		Vector3f averagePoint = NULL_VECTOR3F;
		float summedWeight = 0;

		list<unsigned> queue;
		queue.push_back(i);

		while( !queue.empty())
		{
			unsigned index = queue.front();
			queue.pop_front();

			if( region[index] != 0)
				continue;

			region[index] = regionID;

			float weight = slippage[index];
			summedWeight+=weight;
			averagePoint += ps->get3f(index,m_PosAAT)*weight;

			for( unsigned j=0;j<edgeMap[index].size();j++ )
			{
				queue.push_back(edgeMap[index][j]);
			}

		}

		regionCentroids.push_back( averagePoint / summedWeight);

	}

	// meanshift every region centroid into region mode
	for( unsigned i=0;i<regionCentroids.size();i++ )
	{
		Vector3f pos = regionCentroids[i];
		for( unsigned j=0;j<100;j++ )
		{
			float summed_weight;
			Vector3f nextPos = computeMeanShiftPoint( pos, slippage, sigma, summed_weight);
			if( norm(pos - nextPos) < m_MedianPointDistance * 0.05f )
			{
				points.push_back(nextPos);
				unslippablity.push_back(summed_weight);
			}

			pos = nextPos;
		}
	}

	return region;*/
}

bool X4::SlippageFeatureDetector::testFeaturePointConvergence( const Vector3f &point, float * slippage, float sigma, float &bestSummedWeight, float32 &numConvSteps )
{
	Matrix3f frame;
	Vector3f center;
	computeLocalFrame( point, sigma, m_PC, frame, center );

	std::vector<Vector3f> testpoints;
	float startDist=sigma * 0.75f;
	testpoints.push_back( point + frame[0]*startDist);
	testpoints.push_back( point + frame[1]*startDist);
	testpoints.push_back( point - frame[0]*startDist);
	testpoints.push_back( point - frame[1]*startDist);
	std::vector<Vector3f> modes;
	numConvSteps = 0;

	for( unsigned i=0;i<testpoints.size();i++ )
	{
		Vector3f testPoint = testpoints[i];
		bool modeFound = false;
		for( unsigned j=0;j<200;j++ )
		{
			float summed_weight;
			Vector3f nextPos = computeMeanShiftPoint( testPoint, slippage, sigma, summed_weight, true);
			if( norm(testPoint - nextPos) < m_MedianPointDistance * 0.05f )
			{
				modes.push_back(nextPos);
				bestSummedWeight = summed_weight;
				modeFound = true;

				if( norm( nextPos - point) > sigma * 0.25f)
					return false;

				numConvSteps += (float)j / (float)testpoints.size();
				break;
			}

			testPoint = nextPos;
		}

		if( !modeFound )
			return false;

	}


	return true;
}

Vector3i meanshift_computeGridPoint( const Vector3f &v, float cellsize, std::string &cellID)
{
	int32 x = (int32)(v[0] / cellsize - 0.5f);
	int32 y = (int32)(v[1] / cellsize - 0.5f);
	int32 z = (int32)(v[2] / cellsize - 0.5f);
	cellID = intToStr(x) + "|" + intToStr(y) + "|" + intToStr(z);
	/*BoundingBox3f bb;
	Vector3f center = (makeVector3f(x,y,z) + makeVector3f(1,1,1)*0.5f)*cellsize;
	bb.lowerCorner = center - makeVector3f(1,1,1) * cellsize / 2.0f;
	bb.upperCorner = center + makeVector3f(1,1,1) * cellsize / 2.0f;
	debugRenderer->addBoundingBox(bb, makeVector3f(1,0,0));*/
	return makeVector3i(x,y,z);
}

void X4::SlippageFeatureDetector::gridBasedMeanShiftClustering( UnstructuredInCorePointCloud * upc, float * slippage, float sigma,  std::vector<Vector3f> &points,std::vector<float> &unslippablity )
{
	debugRenderer->clearDebugData("testms");
	debugRenderer->beginRenderJob("testms");
	std::map<std::string,Vector3i> visited;	// grid
	float cellsize = sigma * 0.5f;

	std::vector<float> tmpSlippage;
	for( unsigned i=0;i<upc->getNumPoints();i++ )
		tmpSlippage.push_back(slippage[i]);
	sort( tmpSlippage.begin(), tmpSlippage.end());
	float minSlippage = /*max( tmpSlippage.back() * 0.01f,*/tmpSlippage[tmpSlippage.size() * 0.95f] * 0.01f;
	
	PointSet * ps = upc->getPointSet();
	float averageNumShifts = 0;
	float maxNumShifts = 0;
	for( unsigned i=0;i<ps->getNumEntries();i++ )
	{
		// check minimum slippage
		if( slippage[i] < minSlippage )
			continue;

		// quantize point
		Vector3f pos = ps->get3f(i,m_PosAAT);
		std::string cellID;
		Vector3i gridPos = meanshift_computeGridPoint(pos, cellsize,cellID);
		std::map<std::string,Vector3i>::iterator iter = visited.find(cellID);
		if( iter != visited.end() )
			continue;	// we've already been here

		visited[cellID] = gridPos;
		Vector3i lastGridPos = gridPos;

		// mean shift to local mode or until a visited gridpoint is found
		unsigned numShifts = 0;
		for( unsigned j=0;j<500;j++ )
		{
			// mean shift
			float summedWeight;
			Vector3f nextPos = computeMeanShiftPoint( pos, slippage, sigma, summedWeight);
			numShifts = j;

			/*if( j < 10 )
				debugRenderer->addLine(pos,nextPos, makeVector3f(0.2f,0.8f,1),makeVector3f(0.5f,0.2f,1),1);
			else
				if( j<60 )
					debugRenderer->addLine(pos,nextPos, makeVector3f(1.0f,0.8f,0),makeVector3f(1.0f,0.0,0),1);
				else
				{
					debugRenderer->addLine(pos,nextPos, makeVector3f(1.0f,0.8f,0),makeVector3f(0.0f,0.0,1),1);
					if( j == 200)
						debugRenderer->addSphere(nextPos, m_MedianPointDistance * 0.5f, makeVector3f(0,1,1));
				}*/



			// test if we already visited this grid point
			Vector3i gridPos = meanshift_computeGridPoint(nextPos, cellsize, cellID);
			if( lastGridPos != gridPos )	// did we leave the cell?
			{
				std::map<std::string,Vector3i>::iterator iter2 = visited.find(cellID);
				if( iter2 != visited.end() )
					break;	// we've already been here

				visited[cellID] = gridPos;
				lastGridPos = gridPos;
			}

			// test for local mode
			if( norm( nextPos - pos ) < 0.05f * m_MedianPointDistance )
			{
				points.push_back(nextPos);
				unslippablity.push_back(summedWeight);
				break;
			}

			pos = nextPos;

		}

		//debugOutput << "numshifts: " << numShifts << "\n";
		averageNumShifts += numShifts;
		maxNumShifts = max( maxNumShifts, (float)numShifts);

	}	

	debugRenderer->endRenderJob();
//	debugOutput << "num (initial) modes found: " << points.size() << "\n";
//	debugOutput << "averageNumShifts: " << averageNumShifts / (float)ps->getNumEntries() << "\n"
//		<< "maxNumShifts: " << maxNumShifts << "\n";
}


float X4::SlippageFeatureDetector::computeAverageUnslippabilityAndNormal( const Vector3f &point, float * slippage, float sigma, Vector3f &normal )
{
	// gather sample points
	float ring_sigma = sigma * 0.25f;
	float inv_ring_sigma_2 = 1.0f/ pow(ring_sigma,2.0f);
	float rad = 2.0f * sigma + ring_sigma*2.0f;
	float inv_sigma_2 = 1.0f / pow(sigma,2.0f);
	PCA<float,3> pca;
	std::vector<Vector3f> points;
	m_hIt->setSeekPointAndReset(point);
	Vector3f firstNormal = m_hIt->get3f(m_NormalAAT);
	while( !m_hIt->atEnd())
	{
		Vector3f pos = m_hIt->get3f(m_PosAAT);
		if( norm(pos - point) > rad )
			break;

		pca.addPoint(pos, exp(-normQuad(pos-point) * inv_sigma_2));
		points.push_back(pos);
		m_hIt->next();
	}

	Matrix3f eigenVectors;
	Vector3f eigenValues,centroid;
	pca.analyze(eigenValues,eigenVectors,centroid);
	normal = eigenVectors[2];
	if( normal * firstNormal < 0)
		normal = -normal;

	// compute average
	float ret = 0;
	float summedWeight = 0;
	for( unsigned i=0;i<points.size();i++ )
	{
		float s;
		float weight = exp( - max(0.0f,normQuad(points[i]-point) - 2.0f * sigma)*inv_ring_sigma_2 ) ;
		if( weight < 0.01f )
			continue;
		computeMeanShiftPoint(points[i], slippage, sigma, s);
		ret += s * weight;
		summedWeight += weight;
	}
	if( summedWeight > 0.0f)	// should always be true
		ret /= summedWeight;

	return ret;
}

FeatureSet * X4::SlippageFeatureDetector::detectMultiScaleMS( UnstructuredInCorePointCloud * upc, std::string attachmentName/*="" */ )
{
	m_PosAAT = upc->getAAT("position");
	m_NormalAAT = upc->getAAT("normal");
	m_UseSamplingSpacing = upc->providesAttribute("samplingSpacing");
	if( m_UseSamplingSpacing)
		m_SamplingSpacingAAT = upc->getAAT("samplingSpacing");
	bool outputDebugInfo = true;
	Timer overallTimer;

	// create and attach new feature set, delete old one
	if(attachmentName == "")
		attachmentName = FeatureSet::getDefaultName();
	FeatureSet * ret = dynamic_cast<FeatureSet*>(upc->getAttachments()->getData(attachmentName));
	delete ret;
	ret = new FeatureSet;
	ret->setup( attachmentName, AttachedData::ADF_PERSISTENT );
	upc->getAttachments()->attachData(ret);

	float meanShiftSigma = m_MeanShiftSigma * m_MedianPointDistance;
	CachedRangeQuerry querry(upc);
	
	// ----------------------------------------------------
	// ----------------------------------------------------
	// create smoothed normals
	// ----------------------------------------------------
	std::vector<Vector3f> smoothedNormals(m_NumScales*upc->getNumPoints());
	const unsigned numPoints = upc->getNumPoints();
	PointSet * currentPS = upc->getPointSet();
	float overlap = 2.7f * m_MedianPointDistance;
	std::vector<float> slippage_sigma,inv_slippage_sigma_2, slippage_rad_2, slippage_rad;
	std::vector<float> normal_sigma,inv_normal_sigma_2, normal_rad_2,normal_rad;
	for( unsigned i=0;i<m_NumScales;i++ )
	{
		slippage_sigma.push_back(m_Sigma0 * m_MedianPointDistance * pow( m_ScaleBase, (float)i));
		inv_slippage_sigma_2.push_back(1.0f/pow(slippage_sigma[i],2.0f));
		slippage_rad_2.push_back(pow(slippage_sigma[i]*2.0f,2.0f));
		slippage_rad.push_back(slippage_sigma[i]*2.0f);
		normal_sigma.push_back(slippage_sigma[i]*m_SigmaSmoothFactor);
		inv_normal_sigma_2.push_back(1.0f / pow(normal_sigma[i],2.0f));
		normal_rad_2.push_back(pow(normal_sigma[i]*2.0f,2.0f));
		normal_rad.push_back(normal_sigma[i]*2.0f);
	}
	float maxRadius = normal_sigma.back() * 2.0f;

	Timer timer;
	for( unsigned i=0;i<numPoints;i++ )
	{
		Vector3f currentPoint = currentPS->get3f(i,m_PosAAT);
		PointSet * cache = querry.getSpherePoints(currentPoint, maxRadius, overlap);
		Vector3f cacheCenter = querry.getCacheCenter();
		float dist_point_center = norm(currentPoint - cacheCenter);

		std::vector<Vector3f> normals(m_NumScales,makeVector3f(0,0,0));
		unsigned lastScale = 0;
		for( unsigned p=0;p<cache->getNumEntries();p++ )
		{
			Vector3f samplePoint = cache->get3f(p,m_PosAAT);
			Vector3f currentNormal = cache->get3f(p,m_NormalAAT);
			float dist_2 = normQuad(samplePoint-currentPoint);
			float cdist = norm(samplePoint-cacheCenter);
			float sampleWeight = 1.0f;
			if( m_UseSamplingSpacing )
				sampleWeight = cache->get1f(p,m_SamplingSpacingAAT);
			for( unsigned j=m_NumScales;j>lastScale;j-- )
			{
				unsigned scale = j-1;
				/*if( cdist - dist_point_center > normal_rad[scale])
				{
					lastScale = j;
					break;
				}*/

				if( dist_2 > normal_rad_2[scale] )
					continue;

				float weight = exp(-dist_2 * inv_normal_sigma_2[scale]) * sampleWeight;
				normals[scale] += currentNormal * weight;
			}

		}

		for( unsigned j=0;j<m_NumScales;j++ )
		{
			smoothedNormals[i*m_NumScales+j] = normalize(normals[j]);
		}

		if( i % 5000 == 4999)
		{
			debugOutput << "percentage: " << i * 100 / upc->getNumPoints() << "%\n";
			debugOutput.flush();
		}
	}

	debugOutput << "###############\nSmoothing: " << convertTimeToString(timer.getDeltaValue()) << "\n";

	// ----------------------------------------------------
	// ----------------------------------------------------
	// compute slippage
	// ----------------------------------------------------
	std::vector<float> slippage(m_NumScales*numPoints);
	CachedRangeQuerry slippageQuerry(upc);
	std::vector<float> maxSlippage(m_NumScales,0);
	for( unsigned i=0;i<numPoints;i++ )
	{
		unsigned lastScale = 0;
		Vector3f currentPoint = currentPS->get3f(i,m_PosAAT);

		unsigned * indices;
		std::vector<Vector3f> *positions = slippageQuerry.getSpherePositions(currentPoint,slippage_sigma.back()*2.0f, overlap,10,&indices);
		Vector3f cacheCenter = querry.getCacheCenter();
		float dist_point_center = norm(currentPoint - cacheCenter);

		std::vector<Matrix3f> posMatrix(m_NumScales,IDENTITY3F*0.0f);
		std::vector<Matrix3f> normalMatrix(m_NumScales,IDENTITY3F*0.0f);
		std::vector<Matrix3f> correlationMatrix(m_NumScales,IDENTITY3F*0.0f);
		std::vector<float> summedWeight(m_NumScales,0);
		const size_t numCachePoints = positions->size();
		Vector3f *posPtr = &((*positions)[0]);
		char * dataPtr = upc->getPointSet()->getDataPointer();
		if( m_UseSamplingSpacing )
			dataPtr += m_SamplingSpacingAAT.getOffset();
		float * sampleSpacingPtr =  (float*)dataPtr;
		unsigned vertexSize_4 = upc->getDescr()->getSize() / sizeof(float);

		for( unsigned p=0;p<numCachePoints;p++ )
		{
			Vector3f samplePoint = posPtr[p];
			float sampleWeight = 1.0f;
			if( m_UseSamplingSpacing )
				sampleWeight = sampleSpacingPtr[indices[p]*vertexSize_4];

			float dist_2 = normQuad(samplePoint-currentPoint);
			float cdist_2 = normQuad(samplePoint-cacheCenter);
			float cdist = norm(samplePoint-cacheCenter);
			for( unsigned j=m_NumScales;j>lastScale;j-- )
			{
				unsigned scale = j-1;
				/*if( cdist - dist_point_center > slippage_rad[scale])
				{
					lastScale = j;
					break;
				}*/

				if( dist_2 > slippage_rad_2[scale] )
					continue;

				Vector3f currentNormal = smoothedNormals[ indices[p]*m_NumScales + scale];
				Vector3f p = samplePoint - currentPoint;
				p /= slippage_sigma[scale];
				Vector3f c = p.crossProduct(currentNormal);

				float weight = exp(-dist_2 * inv_normal_sigma_2[scale]) * sampleWeight;
				summedWeight[scale] += weight;
				posMatrix[scale] += outerProduct(c,c) * weight;
				normalMatrix[scale] += outerProduct(currentNormal,currentNormal) * weight;
				correlationMatrix[scale] += outerProduct(currentNormal,c) * weight;
				
			}

		}

		for( unsigned scale=0;scale<m_NumScales;scale++ )
		{
			Matrix6f covariance;
			for( unsigned m=0;m<3;m++ )
			{
				for( unsigned j=0;j<3;j++ )
				{
					covariance[m][j] = posMatrix[scale][m][j];
					covariance[m+3][j+3] = normalMatrix[scale][m][j];
					covariance[m+3][j] = correlationMatrix[scale][m][j];
					covariance[m][j+3] = correlationMatrix[scale][j][m];
				}
			}

			Matrix6f eigenVectors;
			if( summedWeight[scale] > 0.0f)
				covariance /= summedWeight[scale];
			Vector6f eigenValues;
			covariance.computeEigenStructure(eigenValues, eigenVectors);

			slippage[i*m_NumScales+scale] = eigenValues[5];// / eigenValues[0];
			maxSlippage[scale] = max(maxSlippage[scale],slippage[i*m_NumScales+scale]);
		}

		if( i % 5000 == 4999)
		{
			debugOutput << "percentage: " << i * 100 / upc->getNumPoints() << "%\n";
			debugOutput.flush();
		}
		
	}

	debugOutput << "Slippage computation: " << convertTimeToString(timer.getDeltaValue()) << "\n";
	smoothedNormals.clear();

	//debug
	if( m_Debug_WriteScales )
	{
		AAT colorAAT = upc->getAAT("color");
		for( unsigned i=0;i<numPoints;i++ )
		{
			Vector3f col = makeVector3f(slippage[i*m_NumScales+0],slippage[i*m_NumScales+1],slippage[i*m_NumScales+2]);
			currentPS->set3f(i,colorAAT,col);
			currentPS->set3f(i,m_NormalAAT, makeVector3f(slippage[i*m_NumScales+3],slippage[i*m_NumScales+4],slippage[i*m_NumScales+5]));
		}
	}


	beginPC(upc);
	debugRenderer->clearDebugData("testASDF");
	debugRenderer->beginRenderJob("testASDF");
	std::vector<float> averagedSlippage(numPoints*m_NumScales);

	for( unsigned i=0;i<numPoints;i++ )
	{
		Vector3f point = currentPS->get3f(i,m_PosAAT);
		std::vector<float> weights = computeMeanShiftWeights_Scale(point, &slippage[0], meanShiftSigma);
		for( unsigned scale=0;scale<m_NumScales;scale++)
		{
			averagedSlippage[i*m_NumScales+scale] = weights[scale];
		}
	}

	debugOutput << "Pre-MeanShift: " << convertTimeToString(timer.getDeltaValue()) << "\n";
	debugOutput.flush();

	std::vector<Vector3f> color( max((card32)6,m_NumScales), makeVector3f(1,1,1));
	color[0] = makeVector3f(0,1,1);
	color[1] = makeVector3f(1,0,1);
	color[2] = makeVector3f(0.5f,.5f,0.5f);
	color[3] = makeVector3f(0,1,0);
	color[4] = makeVector3f(1,1,0);
	color[5] = makeVector3f(1,0,0);
	std::vector<bool> hidden(numPoints*m_NumScales,false);
	std::vector<Vector4f> startPoints;
	for( unsigned i=0;i<numPoints;i++ )
	{
		Vector3f point = currentPS->get3f(i,m_PosAAT);
		for( unsigned scale=1;scale<m_NumScales;scale++ )
		{
			if( hidden[i*m_NumScales+scale] )
				continue;

			float currentSlippage = averagedSlippage[i*m_NumScales+scale];
			if( currentSlippage < maxSlippage[scale]*0.001f )
			{
				continue;
			}

			unsigned startScale = (scale==0)?0:scale-1;
			unsigned endScale = (scale==m_NumScales-1)?m_NumScales-1:scale+1;

			m_hIt->setSeekPointAndReset(point);
			bool maxima = true;
			while( !m_hIt->atEnd() && maxima)
			{
				Vector3f samplePos = m_hIt->get3f(m_PosAAT);
				float dist_2 = normQuad(point - samplePos);
				float dist = sqrt(dist_2);
				if( dist > max(slippage_sigma[endScale]*0.5f,2.0f*m_MedianPointDistance) )
					break;

				unsigned index = m_hIt->getCurrentPointIndex();
				for( unsigned s=startScale;s<=endScale;s++ )
				{
					if( dist < slippage_sigma[s]*m_Select_RelativeRadius )
					{
						if( averagedSlippage[index*m_NumScales+s] >= currentSlippage &&
							(index != i || scale != s))
						{
							maxima = false;
							break;
						}
						else
							hidden[index*m_NumScales+s] = true;
					}
				}

				m_hIt->next();
			}
			if( maxima )
				startPoints.push_back(makeVector4f(point[0],point[1],point[2],scale));
				
		}

	}

	end();
	debugOutput << "Select Maxima: " << convertTimeToString(timer.getDeltaValue()) << "\n";
	debugOutput.flush();

	float convMeanShift_minMovement = meanShiftSigma * 0.001f;
	card32 numConv = 0;
	card32 numDiv = 0;
	card32 numMeanShifts = 100;
	for( unsigned i=0;i<startPoints.size();i++ )
	{
		Vector4f point = startPoints[i];
		Vector3f point3d = shrink4To3(point);
		float summedSlippage = 0;
		float scale1 = point[3];
		for( unsigned x=0;x<numMeanShifts;x++ )
		{
			float currentSummedWeight;
			Vector4f nextPoint = computeMeanShiftPoint_4D(point, &slippage[0], meanShiftSigma, currentSummedWeight);
			Vector3f nextPoint3d = shrink4To3(nextPoint);
			if( norm(nextPoint3d - point3d) < convMeanShift_minMovement	&& fabs(point[3] - nextPoint[3]) < 0.01f)
			{
				float radius = m_MedianPointDistance * m_Sigma0 * pow(m_ScaleBase,nextPoint[3]);
				//debugRenderer->addSphereWiremesh( nextPoint3d, radius, 1, makeVector3f(1,0,0));

				GeometricFeature * f = new GeometricFeature;
				f->m_Position = nextPoint3d;
				
				SlippageSignature * sig = new SlippageSignature;
				sig->m_Scale = nextPoint[3];
				sig->m_Unslippability = currentSummedWeight;
				sig->m_AverageUnslippability = 2.0f;
				sig->m_FeatureWeight = 1;
				sig->m_ConvergenceSteps = 1;
				sig->m_Sigma = radius/2.0f;
				sig->m_Slippage = makeVector6f(1,1,1,1,1,1);
				f->m_Signatures.push_back(sig);

				Matrix3f frame;
				Vector3f center;
				computeLocalFrame( f->m_Position, meanShiftSigma, upc, frame, center, true);
				f->m_Normal = frame[2];

				ret->m_Features.push_back(f);

				numConv++;
				break;
			}

			//debugRenderer->addLine(point3d, nextPoint3d, makeVector3f(0,0.2f,1),makeVector3f(1,1,1),2);
			point = nextPoint;
			point3d = nextPoint3d;
			if( x == numMeanShifts-1 )
				numDiv++;
		}
	}

	debugOutput << "num conv: " << numConv << "\nnumdiv: " << numDiv << "\n";
	debugOutput << "mean shift: " << convertTimeToString(timer.getDeltaValue()) << "\n";

	// compute overlap
	std::vector<bool> drop(ret->m_Features.size(),false);
	UnstructuredInCorePointCloud fpc;
	fpc.clearAndSetup(createDefaultVertexDescriptor(true,false,false,false), ret->m_Features.size());
	AAT f_posAAT = fpc.getAAT("position");
	for( unsigned i=0;i<fpc.getNumPoints();i++ )
	{
		fpc.getPointSet()->set3f(i,f_posAAT, ret->m_Features[i]->m_Position);
	}
	HierarchicalKNNIterator hIt(&fpc,32,NULL);

	for( unsigned i=0;i<ret->m_Features.size();i++ )
	{
		GeometricFeature * f1 = ret->m_Features[i];
		//for( unsigned j=i+1;j<ret->m_Features.size();j++ )
		hIt.setSeekPointAndReset(f1->m_Position);
		while( !hIt.atEnd())
		{
			unsigned j=hIt.getCurrentPointIndex();
			if( i == j )
			{
				hIt.next();
				continue;
			}
			GeometricFeature * f2 = ret->m_Features[j];
			
			if( norm(f1->m_Position - f2->m_Position) < meanShiftSigma )
			{
				SlippageSignature * sig1 = (SlippageSignature*)f1->m_Signatures[0];
				SlippageSignature * sig2 = (SlippageSignature*)f2->m_Signatures[0];
				if( sig1->m_Scale != sig2->m_Scale )
				{
					sig1->m_FeatureWeight += 1.0f;
					sig2->m_FeatureWeight += 1.0f;
				}

				// ToDo: take feature with best convergence
				if( fabs(sig1->m_Scale - sig2->m_Scale) < 1.5f )
				{
					if( sig1->m_Unslippability < sig2->m_Unslippability )
					{
						drop[i] = true;
					}
					else
					{
						drop[j] = true;
					}
				}
			}
			else
				break;

			hIt.next();
		}
	}

	// drop features
	for( int i=(int)ret->m_Features.size()-1;i>=0;i-- )
	{
		if( drop[i] )
		{
			ret->deleteFeatures(i,i+1);
		}
	}
	
	debugRenderer->endRenderJob();


	return ret;
}


std::vector<float> X4::SlippageFeatureDetector::computeMeanShiftWeights_Scale( const Vector3f &point, float * slippage, float sigma )
{
	m_hIt->setSeekPointAndReset(point);
	float inv_sigma_2 = 1.0f / pow( sigma,2.0f);
	float rad_2 = pow(sigma*2.0f,2.0f);
	std::vector<float> weights(m_NumScales,0);
	while( !m_hIt->atEnd())
	{
		Vector3f pos = m_hIt->get3f(m_PosAAT);
		float dist_2 = normQuad(pos - point);
		if( dist_2 > rad_2)
			break;

		unsigned index = m_hIt->getCurrentPointIndex() * m_NumScales;
		for( unsigned i=0;i<m_NumScales;i++ )
		{
			weights[i] += exp(-dist_2 * inv_sigma_2) * slippage[index+i];
		}
		

		m_hIt->next();
	}

	return weights;
}

X4::Vector4f X4::SlippageFeatureDetector::computeMeanShiftPoint_4D( const Vector4f &point, float * slippage, float sigma, float &summedWeight )
{
	float radius_2 = pow(sigma*2.0f,2.0f);
	float inv_sigma_2 = 1.0f / pow( sigma, 2.0f);
	float overlap = m_MedianPointDistance * 0.5f;
	float rad_overlap_2 = pow(sigma*2.0f + overlap, 2.0f);
	float overlap_2 = pow(overlap,2.0f);

	Vector3f point3d = shrink4To3(point);
	if( m_ShiftCache.size() == 0 || normQuad(m_ShiftCacheLastPos - point3d) > overlap_2 )
	{
		m_ShiftCache.clear();
		m_ShiftCacheLastPos = point3d;
		m_hIt->setSeekPointAndReset(point3d);
		while( !m_hIt->atEnd())
		{
			Vector3f x = m_hIt->get3f(m_PosAAT);
			unsigned index = m_hIt->getCurrentPointIndex();
			float dist_2 = normQuad(point3d - x);
			if( dist_2 > rad_overlap_2)
				break;

			TMeanShiftVertex v;
			v.pos = x;
			v.index = index;
			m_ShiftCache.push_back(v);

			m_hIt->next();
		}

	}


	summedWeight = 0;
	Vector4f nextPos = NULL_VECTOR4F;
	const unsigned numPoints = (unsigned)m_ShiftCache.size();
	TMeanShiftVertex * vertices = &m_ShiftCache[0];
	unsigned startScale = (point[3] >= 1) ? point[3] - 1:0;
	unsigned stopScale = (point[3] < m_NumScales - 2) ? max(point[3]+1,0.0f):m_NumScales-1;

	std::vector<float> scaleweights(m_NumScales,0);
	float scaleSigma = 1.0f;
	float inv_scaleSigma_2 = 1.0f / pow(scaleSigma,2.0f);
	for( unsigned i=startScale;i<=stopScale;i++ )
	{
		scaleweights[i] = exp( -pow((float)i - point[3],2.0f)*inv_scaleSigma_2);
	}

	for( unsigned i=0;i<numPoints;i++ )
	{
		const TMeanShiftVertex v = vertices[i];
		float dist_2 = normQuad(v.pos - point3d);
		float * slippage_ptr = &slippage[v.index*m_NumScales];
		Vector4f pos4d = expand3To4(v.pos);
		for( unsigned s=startScale;s<=stopScale;s++ )
		{
			float weight = exp( -dist_2 * inv_sigma_2) * slippage_ptr[s] * scaleweights[s];
			//float dog_factor = 2.0f;
			//float weight = (exp(-dist_2 * inv_sigma_2 *dog_factor*dog_factor) * dog_factor*1.2f - exp( -dist_2 * inv_sigma_2)) * slippage_ptr[s] * scaleweights[s];
			summedWeight += weight;
			pos4d[3] = s;
			nextPos += pos4d * weight;
		}
		
	}

	//if( fabs(summedWeight) > 1e-10f )
		nextPos /= summedWeight;
	summedWeight /= pow(sigma,2.0f);

	return nextPos;	

}

X4::SlippageDescriptor::SlippageDescriptor()
{
	m_MinUnslippability = 0.0001f;
	m_MinAverageUnslippability = 1.3f;
	m_MaxConvSteps = 30;
	m_SlippageThreshold = 300;
}

bool X4::SlippageSignature::badFeature( LocalDescriptor * myDescriptor ) const
{
	SlippageDescriptor * d = (SlippageDescriptor*)myDescriptor;
	x4Assert( dynamic_cast<SlippageDescriptor*>(myDescriptor) );
//	if( m_ConvergenceSteps > d->m_MaxConvSteps )
//		return true
	if( d->m_MinUnslippability > m_Unslippability )
		return true;

	return false;

	if( d->m_SlippageThreshold < m_Slippage[0]/m_Slippage[5] )
		return true;

	return d->m_MinAverageUnslippability > m_Unslippability/m_AverageUnslippability;
}

float X4::SlippageSignature::compare( LocalSignature * other, LocalDescriptor * myDescriptor )
{
	return 0.0f;
}

X4::SlippageSignature::SlippageSignature()
{
	m_Unslippability = 0;
	m_ConvergenceSteps = 0;
	m_AverageUnslippability = 1.0f;
}