#ifndef SlippageFeatureDetectorH
#define SlippageFeatureDetectorH
//---------------------------------------------------------------------------
#include "x3foundation_private.h"
//---------------------------------------------------------------------------
#include "FeaturePointDetector.h"
#include "LocalDescriptor.h"
//---------------------------------------------------------------------------

namespace X4
{
class LocalICP;
class HierarchicalKNNIterator;
class PointCloud;
class Scene;
/**
	uses slippage analysis for feature detection
*/
class X3FOUNDATION_PRIVATE_API SlippageFeatureDetector : public FeaturePointDetector
{
	X4_CLASS( SlippageFeatureDetector )

private:

	// properties
	float32	m_Sigma0;
	float32	m_SigmaSmoothFactor;		// sigmaSmooth = sigma * sigmaSmoothFactor
	float32	m_ScaleBase;				// scale_i = scale_0 * pow(scaleBase,i)

   card32	m_NumScales;
	card32	m_MaxMeanshiftSteps;
	float32	m_MeanShiftSigma;
	float32	m_Select_RelativeRadius;

	bool		m_Debug_WriteScales;
   LocalICP * m_icp;


	// members
	PointCloud * m_PC;
	HierarchicalKNNIterator * m_hIt;
	AAT		m_PosAAT;
	AAT		m_NormalAAT;
	AAT		m_SamplingSpacingAAT;
	bool		m_UseSamplingSpacing;

	struct TMeanShiftVertex
	{
		Vector3f pos;
		float slippage;
		unsigned index;
	};
	std::vector<TMeanShiftVertex> m_ShiftCache;
	Vector3f m_ShiftCacheLastPos;


	void beginPC(UnstructuredInCorePointCloud * upc);
	void endPC();
	std::vector<float> computeSlippageFactor( UnstructuredInCorePointCloud * upc, float sigma );
	Vector3f computeMeanShiftPoint( const Vector3f &point, float * slippage, float sigma, float &summedWeight, bool cached = false );
	std::vector<unsigned> meanShiftClustering( UnstructuredInCorePointCloud * upc, float * slippage, float sigma, 
		std::vector<Vector3f> &points,std::vector<float> &unslippablity );
	
	void gridBasedMeanShiftClustering( UnstructuredInCorePointCloud * upc, float * slippage, float sigma, 
		std::vector<Vector3f> &points,std::vector<float> &unslippablity );

	/// tests convergence from a few start positions next to point
	/// returns true, if all converged. sets summedWeight to the best unslippablility
	bool testFeaturePointConvergence( const Vector3f &point, float * slippage, float sigma, float &summedWeight, float32 &numConvSteps );

	float computeAverageUnslippabilityAndNormal( const Vector3f &point, float * slippage, float sigma, Vector3f &normal);

	FeatureSet * detectMultiScaleMS( UnstructuredInCorePointCloud * upc, std::string attachmentName="" );

	std::vector<float> computeMeanShiftWeights_Scale( const Vector3f &point, float * slippage, float sigma);

	Vector4f computeMeanShiftPoint_4D( const Vector4f &point, float * slippage, float sigma, float &summedWeight);


public:
	SlippageFeatureDetector(void);
	~SlippageFeatureDetector(void);

   virtual void assign(const Object* obj, X4CopyContext *context = NULL);

	virtual FeatureSet * detect( UnstructuredInCorePointCloud * upc, std::string attachmentName="" );

	float32 getSigma() const {return m_Sigma0;}
	void setSigma( float s ){m_Sigma0 = s;}

   float32 getSigmaSmoothFactor() const {return m_SigmaSmoothFactor;}
   void setSigmaSmoothFactor(float32 v) {m_SigmaSmoothFactor=v;}

   float32 getScaleBase() const {return m_ScaleBase;}
   void setScaleBase(float32 v) {m_ScaleBase=v;}


   card32 getNumScales() const {return m_NumScales;}
   void setNumScales(card32 v) {m_NumScales=v;}

   card32 getMaxMeanshiftSteps() const {return m_MaxMeanshiftSteps;}
   void setMaxMeanshiftSteps(card32 v) {m_MaxMeanshiftSteps=v;}

   float32 getMeanShiftSigma() const {return m_MeanShiftSigma;}
   void setMeanShiftSigma(float32 v) {m_MeanShiftSigma=v;}

   float32 getSelect_RelativeRadius() const {return m_Select_RelativeRadius;}
   void setSelect_RelativeRadius(float32 v) {m_Select_RelativeRadius=v;}

	Scene * sceneHack;

};


class SlippageSignature;

/**
descriptor for scale space signatures
*/
class SlippageDescriptor : public LocalDescriptor
{
	X4_CLASS( SlippageDescriptor )
public:

	SlippageDescriptor();

	virtual LocalSignature * computeSignature(const Vector3f &point, const Vector3f &normal, float size, UnstructuredInCorePointCloud * upc)
	{return NULL;}

	virtual bool isRecomputeable() const {return false;}

   float32 getMinUnslippability() const {return m_MinUnslippability;}
   void setMinUnslippability(float32 v) {m_MinUnslippability=v;}

   float32 getMaxConvSteps() const {return m_MaxConvSteps;}
   void setMaxConvSteps(float32 v) {m_MaxConvSteps=v;}

   float32 getMinAverageUnslippability() const {return m_MinAverageUnslippability;}
   void setMinAverageUnslippability(float32 v) {m_MinAverageUnslippability=v;}

   float32 getSlippageThreshold() const {return m_SlippageThreshold;}
   void setSlippageThreshold(float32 v) {m_SlippageThreshold=v;}


private:

	float32	m_MinUnslippability;
	float32	m_MaxConvSteps;
	float32	m_MinAverageUnslippability;
	float32	m_SlippageThreshold;

	friend SlippageSignature;
};

/**
signature for scale space features
*/
class SlippageSignature : public LocalSignature
{
	X4_CLASS( SlippageSignature )

public:

	SlippageSignature();

	virtual bool badFeature(LocalDescriptor * myDescriptor) const;
	virtual float compare( LocalSignature * other, LocalDescriptor * myDescriptor );
	virtual float compare( LocalSignature * other ) { return compare(other, NULL); }

	float32	m_Unslippability;
	float32	m_FeatureWeight;	// encodes how often a feature was detected at different scales
	float32	m_Scale;
	float32	m_ConvergenceSteps;	// average convergence steps
	float32	m_AverageUnslippability;
	Vector6f m_Slippage;
	float32	m_Sigma;				// redundant to scale, but easier to use

};


}

#endif