#ifndef SORTEDLINKEDLIST_H
#define SORTEDLINKEDLIST_H

/// Node object for the SortedLinkedList, container for the content and pointer towards neighbors
template <class KeyType, class ItemType>
struct SortedLinkedListItem
{
	/// Key value
	KeyType key;
	/// Item value 
	ItemType value;
	/// List management: last and next item in list
	SortedLinkedListItem<KeyType,ItemType> *next, *last;
};

/**
 * Iterator object for SortedLinkedList access
 */
template <class KeyType, class ItemType>
class SortedLinkedListIterator
{
	private:
		/// Pointer to start item in the asociated SortedLinkedList
		SortedLinkedListItem<KeyType,ItemType> *start;
		/// Pointer to current item in list
		SortedLinkedListItem<KeyType,ItemType> *current;

	public:
		/**
       * Constructor, requires start item of the list and the current pointer
       * item of the iterator.
       */
		SortedLinkedListIterator( SortedLinkedListItem<KeyType,ItemType> *s,
			SortedLinkedListItem<KeyType,ItemType> *c ){
			start = s;
			current = c;
		}
		/// Reset iterator to beginning
		inline void reset(){ current = start; }
		/// Test if list has more items after current position
		inline bool hasMoreItems(){ return current->next != NULL; }
		/// Inc pointer in list
		inline void nextItem(){ current = current->next; }
		/// Return current item
		inline ItemType getItem(){ return current->value; };
		/// Return key of current item
		inline KeyType getKey(){ return current->key; };
};

/**
 * Provides a sorted list of arbitrary items
 */
template <class KeyType, class ItemType>
class SortedLinkedList 
{
	private:
		/// Pointer towards start and end of list
		SortedLinkedListItem<KeyType,ItemType> *start, *end;
		/// Free all allocated memory
		void freeMemory();
      
	public:
		/// Constructor
		SortedLinkedList();
		/// Destructor
		~SortedLinkedList();
		/// Insert item into list
		void insertItem( KeyType, ItemType value );
		/// Print list to console
		void printToConsole();
		/// Create a iterator for the list
		SortedLinkedListIterator<KeyType,ItemType>* createIterator();
};

/**
 * Create new SortedLinkedList object
 */
template<class KeyType, class ItemType>
SortedLinkedList<KeyType,ItemType>::SortedLinkedList()
{
	start = NULL;
	end = NULL;
}

/**
 * Delete object
 */
template<class KeyType, class ItemType>
SortedLinkedList<KeyType,ItemType>::~SortedLinkedList()
{
	freeMemory();
}

/**
 * Free all allocated memory, delete all nodes and reset list
 */
template<class KeyType, class ItemType>
void SortedLinkedList<KeyType,ItemType>::freeMemory()
{
	SortedLinkedListItem<KeyType,ItemType> *currentItem = start, *deleteItem;
	while ( currentItem != NULL ){
		deleteItem = currentItem;
		currentItem = currentItem->next;
		delete deleteItem;
	}
	start = end = NULL;
}

/**
 * Find appropriate position within list and add item in a sorted manner.
 */
template<class KeyType, class ItemType>
void SortedLinkedList<KeyType,ItemType>::insertItem( KeyType key, ItemType value )
{
	SortedLinkedListItem<KeyType,ItemType>* newItem = new SortedLinkedListItem<KeyType,ItemType>;
	newItem->key = key;
	newItem->value = value;

	if ( start == NULL ){
		// Insert into empty list
		start = end = newItem;
		start->last = start->next = NULL;
	} else {
		// Find appropriate position in list for insertion
		SortedLinkedListItem<KeyType,ItemType>* currentItem = start;
		while ( currentItem != NULL && currentItem->key < newItem->key ){
			currentItem = currentItem->next;
		}
		if ( currentItem != NULL ){
			// Insert before existing
			newItem->last = currentItem->last;
			newItem->next = currentItem;
			currentItem->last = newItem;
			if ( newItem->last != NULL ){
				newItem->last->next = newItem;
			} else {
				// Insertion at the beginning of the list
				start = newItem;
			}
		} else {
			// Insert at the end of the list;
			newItem->last = end;
			newItem->next = NULL;
			end->next = newItem;
			end = newItem;
		}
	}
}

/**
 * Print the content of the current list to console.
 */
template<class KeyType, class ItemType>
void SortedLinkedList<KeyType,ItemType>::printToConsole()
{
	SortedLinkedListIterator<KeyType,ItemType>* it = createIterator();
	it->reset();
	while ( it->hasMoreItems() ){
		std::cout << /*it->getItem() << */"(" << it->getKey() << ") - ";
		it->nextItem();
	}
	std::cout << endl;
	delete it;
}

/** 
 * Create new iterator for the list, has to be deleted manually.
 */
template<class KeyType, class ItemType>
SortedLinkedListIterator<KeyType,ItemType>* SortedLinkedList<KeyType,ItemType>::createIterator()
{
	SortedLinkedListIterator<KeyType,ItemType> *it = new SortedLinkedListIterator<KeyType,ItemType>( start, start );
	return it;
}

#endif