#ifndef SpectralCorrespondenceValidationH
#define SpectralCorrespondenceValidationH
//---------------------------------------------------------------------------
#include "x3foundation_private.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
//---------------------------------------------------------------------------
#include "SparseLinearAlgebra.h"
//---------------------------------------------------------------------------

namespace X4
{

class RegistrationContext;

/**
	implements a spectral technique for correspondence matching
	Marius Leordeanu, Martial Hebert - "A Spectral Technique for Correspondence 
	Problems Using Pairwise Constraints"
*/
class X3FOUNDATION_PRIVATE_API SpectralCorrespondenceValidation : public Persistent
{
	X4_CLASS( SpectralCorrespondenceValidation )

public:
	struct TFeaturePair
	{
		card32	i1,i2;
		card32	orgIndex1, orgIndex2;
		float32	confidence;
		Vector3f p1,p2;
		bool		valid;	// forward search
		bool		conflicting; 
	};

private:

	SparseMatrixF				m_Matrix;
	DynamicMatrix<float>		m_DistanceMatrix1,m_DistanceMatrix2;
	DynamicMatrix<float>		m_EuclidDistanceMatrix1,m_EuclidDistanceMatrix2;
	DynamicMatrix<float>		m_NormalAngle1,m_NormalAngle2;
	std::vector<TFeaturePair>		m_Solution;
	std::vector<TFeaturePair>		m_Pairs;

	float32						m_ScaleDiff;		// |d1-d2| < min(d1,d2)*scaleDiff + addDiff
	float32						m_AddDiff;
	float32						m_Sigma;
	float32						m_MaxDistance;
	float32						m_MinConfidence;
	card32						m_NumPowerIterations;
	float32						m_MaxNormalAngleDiff;
	float32						m_NormalAngleSigma;


	//RegistrationContext *	m_RegistrationContext;

	// uses euklidian distances if dist_e < m_UseEuclidRadius
	// uses geodetic distances if dist_e > m_UseGeodeticRadius
	// blends between both distances if m_UseEuclidRadius <= dist_e <= m_UseGeodeticRadius
	// uses geodetic distances if dist_g / dist_e > m_DistanceSimilarityFactor
	float32						m_UseEuclidRadius;
	float32						m_UseGeodeticRadius;
	float32						m_DistanceSimilarityFactor;
	bool							m_Deformable;
	float32						m_DistanceTolerance;


	/// set (x,y) and (y,x) matrix entry to value
	void setMatrixEntry( card32 x, card32 y, float value );

	/// returns true if both pairs are in conflict with each other
	bool conflict( card32 pair1, card32 pair2 );

	/// returns distances between two feature pairs
	void getDistances( card32 pair1, card32 pair2, float &distance1, float &distance2);


public:
	SpectralCorrespondenceValidation(void);
	~SpectralCorrespondenceValidation(void);

   virtual void assign(const Object* obj, X4CopyContext *context = NULL);

	/// setup matrices etc.
	void setup( card32 numberOfPairs, card32 numberOfFeatures1, card32 numberOfFeatures2, card32 numPowerIterations=1 );
	
	/// set entry for distance matrix
	void setDistanceEntry1( card32 index1, card32 index2, float geo_distance, float euclid_distance );
	void setDistanceEntry2( card32 index1, card32 index2, float geo_distance, float euclid_distance );

	void setNormalAngle1( card32 index1, card32 index2, float diff_angle);
	void setNormalAngle2( card32 index1, card32 index2, float diff_angle);

	void setCorrespondencePair( card32 index, card32 i1, card32 i2, card32 orgI1, card32 orgI2, float confidence,
		const Vector3f &pos1, const Vector3f &pos2, bool valid = false);

	// estimate valid correspondences (returns first eigenValue)
	float32 calculate();

	void forwardSearch();

	std::vector<TFeaturePair>	getValidPairs() {return m_Solution;}

	float32 getScaleDiff() const {return m_ScaleDiff;}
	void setScaleDiff(float32 v) {m_ScaleDiff=v;}

	float32 getAddDiff() const {return m_AddDiff;}
	void setAddDiff(float32 v) {m_AddDiff=v;}

	float32 getSigma() const {return m_Sigma;}
	void setSigma(float32 v) {m_Sigma=v;}

	float32 getMaxDistance() const {return m_MaxDistance;}
	void setMaxDistance(float32 v) {m_MaxDistance=v;}

	float32 getMinConfidence() const {return m_MinConfidence;}
	void setMinConfidence(float32 v) {m_MinConfidence=v;}

	//RegistrationContext * getRegistrationContext() const {return m_RegistrationContext;}
	//void setRegistrationContext(RegistrationContext * v) {m_RegistrationContext=v;}

};

}

#endif