#include "StdAfx.h"
//------------------------------------------------------------------------------------------------
#include "SpinImageDescriptor.h"
#include <PointSet.h>
#include <CopyObjectProperties.h>
#include <ObjectClassproperty.h>
#include <TopologicalKNNIterator.h>
#include <TopologyRangeSearch.h>
#include <FastSphereQuerry.h>
#include <UnstructuredInCorePointCloud.h>
#include <HierarchicalKNNIterator.h>
#include <TopologyAttachment.h>
#include <JointHistogram.h>

//------------------------------------------------------------------------------------------------
using namespace X4;

IMPLEMENT_X4_CLASS( SpinImageDescriptor ,1)
{
    BEGIN_CLASS_INIT( SpinImageDescriptor );

    ADD_CARD32_PROP (m_spinSize, 0);
    ADD_CARD32_PROP (m_histSize, 0);
    ADD_BOOLEAN_PROP (m_useNormals, 1);
}

IMPLEMENT_X4_CLASS( SpinImageSignature, 0)
{
    BEGIN_CLASS_INIT( SpinImageSignature );
    
    ADD_FLOAT32_VAR_ARRAY_PROP(m_spinImage, 0);

    ADD_VECTOR3F_PROP(m_center, 0);
    ADD_VECTOR3F_PROP(m_normal, 0);

    ADD_CARD32_PROP(m_binsize, 0);
    ADD_CARD32_PROP(m_bins, 0);

    ADD_FLOAT32_PROP(m_radius, 0);
}

//------------------------------------------------------------------------------------------------

//------------------------------------------------------------------------------------------------
SpinImageDescriptor::SpinImageDescriptor(bool useNormals) : LocalDescriptor()
{
    m_spinSize = 16;
    m_histSize = 256;
    m_useNormals = useNormals;
}

//------------------------------------------------------------------------------------------------
SpinImageDescriptor::~SpinImageDescriptor()
{

}

//------------------------------------------------------------------------------------------------
void SpinImageDescriptor::assign(const Object* obj, X4CopyContext *context)
{
    const SpinImageDescriptor * o = dynamic_cast<const SpinImageDescriptor*>(obj);
    x4Assert( o != NULL );

    copyObjectProperties(obj,this);
}

//------------------------------------------------------------------------------------------------
LocalSignature* SpinImageDescriptor::computeSignature(const Vector3f &point, const Vector3f &normal, float size, UnstructuredInCorePointCloud * upc)
{
    if (upc == NULL) return NULL;

    // query point cloud atts
    PointSet* ps = upc->getPointSet();		
    AAT positionAAT = ps->getAAT("position");
    AAT normalAAT = ps->getAAT("normal");

    // get index of the center point
    card64 index = 0;
    std::vector<Vector3f> points;
    std::vector<Vector3f> normals;
    HierarchicalKNNIterator hIt(upc,32,NULL);
    hIt.setSeekPointAndReset(point);
    index = hIt.getCurrentPointIndex();

    // check if the point cloud contains a topology graph
    // use the graph to query the points
    TopologyAttachment* tatt = dynamic_cast<TopologyAttachment*>(upc->getAttachments()->getData(TopologyAttachment::getDefaultName()));
    InCorePCTopologyGraph* tg = NULL;
    if (tatt != NULL)
        tg = tatt->getTopology();

    // graph is there, hence 
    if(tg)
    {
        // do a range query on topology
        TopNeighborhoodSphereSearch querry;
        querry.setup(tg, ps, index, point, size, positionAAT);
        while(!querry.atEnd())
        {
            points.push_back(ps->get3f(querry.getCurrentIndex(), positionAAT));
            normals.push_back(ps->get3f(querry.getCurrentIndex(), normalAAT));
            querry.next();
        }

    // ok no topology, then do just simple sphere query
    }else
    {
        FastSphereQuerry querry(upc);
        mpcard numPoints;
        mpcard* indices;
        querry.querry( point, size, &indices, numPoints);	
        points.resize(numPoints);
		normals.resize(numPoints);
        for( unsigned i=0;i < numPoints;i++,indices++)
        {
            points[i] = ps->get3f(*indices,positionAAT);
            normals[i] = ps->get3f(*indices,normalAAT);
        }
    }

    // create and hence compute spin image signature
    SpinImageSignature* spin = new SpinImageSignature();
    if (m_useNormals)
        spin->create(point, normal, size, points, normals, m_spinSize);
    else
        spin->create(point, normal, size, points, m_spinSize); // ignore normals!

    return spin;
}

//------------------------------------------------------------------------------------------------
float SpinImageDescriptor::compareLinear(SpinImageSignature* one, SpinImageSignature* two)
{
    if (one == NULL || two == NULL) return 0;
    if (one->getNumElements() != two->getNumElements()) return 0;

	float N = one->getNumElements();
	float sumpq = 0;
	float sump = 0;
	float sumq = 0;
	float sumps = 0;
	float sumqs = 0;
	for (unsigned i=0; i < one->getNumElements(); i++)
	{
		sumpq += one->getElement(i) * two->getElement(i);
		sump  += one->getElement(i);
		sumq  += two->getElement(i);

		sumps += (one->getElement(i) * one->getElement(i));
		sumqs += (two->getElement(i) * two->getElement(i));
	}

	float R = (N * sumpq - sump*sumq) / sqrt((N * sumps - sump*sump) * (N * sumqs - sumq*sumq));

	return (R+1.0) * 0.5;
}

//------------------------------------------------------------------------------------------------
float SpinImageDescriptor::compareByNMI(SpinImageSignature* one, SpinImageSignature* two)
{
    if (one == NULL || two == NULL) return 0;
    if (one->getNumElements() != two->getNumElements()) return 0;

    // compute NMI using joint histograms
    JointHistogram* jH = new JointHistogram();
    jH->setup(m_histSize, m_histSize);
    jH->setMinMax(0,1, 0);
    jH->setMinMax(0,1, 1);

    double nmi = 0.0;
    for(card32 k = 0; k < one->getNumElements(); k++)
    {
        jH->add(one->getElement(k), two->getElement(k));
    }
    nmi = jH->computeNMI();		

    delete jH;
    return 1.0 - nmi;
}



//------------------------------------------------------------------------------------------------
// SpinImageSignature
//------------------------------------------------------------------------------------------------
SpinImageSignature::SpinImageSignature()
{
    m_bins = 0;
    m_binsize = 0;
}

//------------------------------------------------------------------------------------------------
void SpinImageSignature::assign(const Object* obj, X4CopyContext *context)
{
    const SpinImageSignature * o = dynamic_cast<const SpinImageSignature*>(obj);
    x4Assert( o != NULL );

    //copyObjectProperties(obj,this);
    m_spinImage = o->m_spinImage;
    m_center = o->m_center;
    m_normal = o->m_normal;
    m_binsize = o->m_binsize;
    m_bins = o->m_bins;
    m_radius = o->m_radius;
}

//------------------------------------------------------------------------------------------------
void SpinImageSignature::deletem_spinImage( card32 index )
{
    DVectorF tmp = m_spinImage;
    m_spinImage.setDim(tmp.getDim()-1);
    for( unsigned i=0;i<index;i++ )
        m_spinImage[i] = tmp[i];
    for( unsigned i=index+1;i<tmp.getDim();i++ )
        m_spinImage[i-1] = tmp[i];
}

//------------------------------------------------------------------------------------------------
/// construct from scratch
//------------------------------------------------------------------------------------------------
void SpinImageSignature::create(Vector3f cnt, Vector3f nml, float rad, const std::vector<Vector3f> &points, const std::vector<Vector3f> &normals, card32 _n)
{
    int n = _n;

    //m_spinImage.resize(n*n, 0.0f);
    m_spinImage.setDim(n*n);
    m_spinImage.setZero();

    m_binsize = n;
    m_bins = n*n;

    m_center = cnt;
    m_normal = nml;
    m_radius = rad;

    int fn = n-1;

    // clear the spin image
    float init = 0.0;
    //fill(&init);

    bool useNormals = normals.size() == points.size();

    // init some values
    float32 r2 = m_radius*m_radius;
    float32 ir = 1.0 / m_radius;

    std::vector< std::vector<float32> > spinImgs(n);
    std::vector<float32> cols(n,0.0f);
    for (int i = 0; i < n; ++i) spinImgs[i] = cols;
    float32 m = 10e10f;   //initialize minimum bin count (impossible number)
    float32 M = 0.0f;     //initialize maximum bin count (impossible number)
    // go over all points and determine distances and bin position
    for (card32 i = 0; i < points.size(); ++i)
    {
        Vector3f p = points[i];
        Vector3f pd = p - m_center;
        float32 beta = m_normal * pd;
        float32 d = pd.getSqrNorm();

        // if point is outside the radius, ignore it
        if (d >= r2) continue;

        // ignore points with very distinct normal
        if (useNormals && fabs(normals[i] * m_normal) < M_PI/6.0) //What is this PI doing here?
            continue;

        float32 diff1 = d - beta*beta;
        if (diff1 < 0) continue;
        float32 alpha = sqrt(diff1);

        // normalize values to 0..n
        alpha = (alpha * ir) * (float32)fn;
        beta = ((beta + m_radius) * 0.5) * ir * (float32)fn;

        // find integer bin positions and fractions
        int a = int(floor(alpha));
        if (a >= n) a = n -1;
        int b = int(floor(beta));
        if (b >= n) b = n -1;
        float32 af = alpha - (float32)a;
        float32 bf = beta - (float32)b;
        if (a < 0 || b < 0) std::cout<<"a or b smaller zero"<<std::endl; 
        spinImgs[a][b] += (1.0-af) * (1.0-bf);
        if ( (a + 1) < (int)m_binsize) spinImgs[a+1][b] += (af    ) * (1.0-bf);
        if ( (b + 1) < (int)m_binsize) spinImgs[a][b+1] += (1.0-af) * (bf    );
        if ( (a + 1) < (int)m_binsize && (b + 1) < (int)m_binsize) spinImgs[a+1][b+1] += (af    ) * (bf    );

    }
    for (int j = 0; j < n; ++j)
    {
        for ( int i = 0; i < n; ++i)
        {
            M = max( M, spinImgs[i][j]);
            m = min( m, spinImgs[i][j]);
        }
    }
    const float32 rescaleRange = max(0.000001f, M - m);
    for ( int j = 0; j < n; ++j)
    {
        for ( int i = 0; i < n; ++i)
        {
            m_spinImage[j*n+i] = (spinImgs[i][j] - m)/rescaleRange;
        }
    }

/*
	// go over all points and determine distances and bin position
	for (card32 i = 0; i < points.size(); ++i){
		Vector3f p = points[i];
		Vector3f pd = p - m_center;

		// determine alpha (distance to normal) and beta (height difference) components
		float32 d = pd.getSqrNorm();

		// if point is outside the radius, ignore it
		if (d >= r2) continue;

		float32 beta = m_normal * pd;
		float32 diff1 = d - beta*beta;
		if (diff1 < 0) continue;
		float32 alpha = sqrt(diff1);

		// normalize values to 0..n
		alpha = (alpha * ir) * (float32)fn;
		beta = ((beta + m_radius) * 0.5) * ir * (float32)fn;

		// find integer bin positions and fractions
		int a = int(floor(alpha));
		if (a >= n) a = n -1;
		int b = int(floor(beta));
		if (b >= n) b = n -1;
		float32 af = alpha - (float32)a;
		float32 bf = beta - (float32)b;
		if (a < 0 || b < 0) std::cout<<"a or b smaller zero"<<std::endl; 
		spinImgs[a][b] += (1.0-af) * (1.0-bf);
		if ( (a + 1) < (int)m_binsize) spinImgs[a+1][b] += (af    ) * (1.0-bf);
		if ( (b + 1) < (int)m_binsize)spinImgs[a][b+1] += (1.0-af) * (bf    );
		if ( (a + 1) < (int)m_binsize && (b + 1) < (int)m_binsize) spinImgs[a+1][b+1] += (af    ) * (bf    );
				
	}
	for (int j = 0; j < n; ++j)
    {
		for ( int i = 0; i < n; ++i)
        {
			if (spinImgs[i][j] < m) m = spinImgs[i][j];
			if (spinImgs[i][j] > M) M = spinImgs[i][j];
		}
	}

	for ( int j = 0; j < n; ++j)
    {
		for ( int i = 0; i < n; ++i)
        {
			m_spinImage[j*n+i] = (spinImgs[i][j] - m)/(M - m);
		}
	}
*/
}

//------------------------------------------------------------------------------------------------
/// destruct from scratch
//------------------------------------------------------------------------------------------------
SpinImageSignature::~SpinImageSignature()
{

}

//------------------------------------------------------------------------------------------------
float SpinImageSignature::compare(const SpinImageSignature& src)
{
    return compare(const_cast<SpinImageSignature*>(&src), NULL);
}

//------------------------------------------------------------------------------------------------
/// comparison operator, based solely on spin image content values
//------------------------------------------------------------------------------------------------
float SpinImageSignature::compare( LocalSignature * other, LocalDescriptor *  descriptor)
{
	return compareByNMI(other, descriptor);
}

//------------------------------------------------------------------------------------------------
float SpinImageSignature::compareByNMI(LocalSignature* other, LocalDescriptor* myDescriptor) 
{
    SpinImageSignature* src = dynamic_cast<SpinImageSignature*>(other);
    SpinImageDescriptor* descr = dynamic_cast<SpinImageDescriptor*>(myDescriptor);

	if (other == NULL) return 0;
	if (getNumElements() != src->getNumElements()) return 0;

	// compute NMI using joint histograms
	JointHistogram* jH = new JointHistogram();
	jH->setup(descr ? descr->m_histSize : 16, descr ? descr->m_histSize : 16);
	jH->setMinMax(0,1, 0);
	jH->setMinMax(0,1, 1);

	double nmi = 0.0;
	for(card32 k = 0; k < getNumElements(); k++)
	{
		jH->add(getElement(k), src->getElement(k));
	}
	nmi = jH->computeNMI();		

	delete jH;
	return 1.0 - nmi;
}

//------------------------------------------------------------------------------------------------
/// distance operator, based on center distance
//------------------------------------------------------------------------------------------------
float32 SpinImageSignature::distance(const SpinImageSignature& src, const float32& rad)
{
	Vector3f p = m_center - src.m_center;
	float32 d = sqrt(p.getSqrNorm()) / rad;
	float32 dw = pow( (1.0 - d ) , 4.0 ) * ( 4.0 * d + 1);
	if (d > 1.0)
		dw = 0.0;

	return dw;
}

//------------------------------------------------------------------------------------------------
/// distance operator, based on normal deviation
//------------------------------------------------------------------------------------------------
float32 SpinImageSignature::deviation(const SpinImageSignature& src)
{
	float32 d = 1.0 - ((m_normal[0]*src.m_normal[0]+m_normal[1]*src.m_normal[1]+m_normal[2]*src.m_normal[2]) * 0.5 + 0.5);
	return pow( (1.0 - d ) , 4.0 ) * ( 4.0 * d + 1);
}

#if 0
//------------------------------------------------------------------------------------------------
void SpinImageSignature::write(OutputObjectStream *s) const
{
	s->writeMaxPlatformCard(m_binsize);
	s->writeMaxPlatformInt(m_bins);
	s->writeBuffer(&m_radius, sizeof(float32));
	s->writeBuffer(m_center.data(), sizeof(Vector3f));
	s->writeBuffer(m_normal.data(), sizeof(Vector3f));
	m_spinImage.write(s);
}


//------------------------------------------------------------------------------------------------
void SpinImageSignature::read(InputObjectStream *s)
{
	mpcard size;
	s->readMaxPlatformCard(size);
	m_binsize = size;
	s->readMaxPlatformCard(size);
	m_bins = size;
	
	float radius;
	s->readBuffer(&radius, sizeof(float32));

	Vector3f vec;
	s->readBuffer(vec.data(), sizeof(Vector3f));
	m_center = vec;

	s->readBuffer(vec.data(), sizeof(Vector3f));
	m_normal = vec;

	m_spinImage.read(s);
}
#endif 

