#ifndef SpinImageDescriptor_H
#define SpinImageDescriptor_H
//--------------------------------------------------------------------------
#include "x3foundation_private.h"
//--------------------------------------------------------------------------
#include <ParametricInCorePointCloud.h>
#include <LocalDescriptor.h>

//--------------------------------------------------------------------------
namespace X4
{
    class SpinImageSignature;
    class SpinImageDescriptor;

    /**
    *  Code originally taken from project_rl (Ruxandra's project) and rewritten
    *  to use spin images as usual descriptors for any feature points.
    **/
    class X3FOUNDATION_PRIVATE_API SpinImageDescriptor : public LocalDescriptor
    {
        X4_CLASS( SpinImageDescriptor );
        private:
            card32 m_spinSize;
            card32 m_histSize;
            //! whether to check surface normals for ignoring incompatibly oriented points.
            bool m_useNormals;

			friend class SpinImageSignature;

        public:
	        /// construct from scratch
	        SpinImageDescriptor(bool useNormals = true);

            /// destruct 
	        virtual ~SpinImageDescriptor();

            /// compute local signature for point
            virtual LocalSignature * computeSignature(const Vector3f &point, const Vector3f &normal, float size, UnstructuredInCorePointCloud * upc);

            /// returns true if signatures can be recomputed (e.g. false for all detector-based signatures)
            bool isRecomputeable() const {return true;}

            /// copy descriptor to descriptor
            void assign(const Object* obj, X4CopyContext *context = NULL);

            /// get size of the spin image (size x size), default 16
            card32 getSpinSize() const { return m_spinSize; }

            /// set spin size, default size is 16
            void setSpinSize(card32 size) { m_spinSize = size; }

            /// set histogram size to use when comparing two spin images by nmi
            void setHistogramSize(card32 size) { m_histSize = size; }
            
            /// return the length of the descriptor.
            card32 getDescriptorSize() const { return m_histSize; }

            /// compare two spin images, by using joint histograms
            virtual float compareByNMI(SpinImageSignature* one, SpinImageSignature* two);

            /// compare two spin images, by simple linear comparison
            virtual float compareLinear(SpinImageSignature* one, SpinImageSignature* two);
    };

    /**
    * SpinImage class representing by a signature. Signatures used as descriptors,
    * or at least the SpinImageDescriptor handles with them.
    **/
    class X3FOUNDATION_PRIVATE_API SpinImageSignature : public LocalSignature
    {
        X4_CLASS( SpinImageSignature );

        private:
            //std::vector<float32> m_spinImage;
            DVectorF m_spinImage;

            /// extents of spinimage
            card32 m_binsize;
            card32 m_bins;

            Vector3f m_center;
            Vector3f m_normal;
            float32 m_radius;

            //IMPLEMENT_VAR_ARRAY_METHODS_STL(m_spinImage);
            void setNumm_spinImages(card32 size) {m_spinImage.setDim(size);m_spinImage.setZero();}
            void deletem_spinImage( card32 index );
            float* getm_spinImage( card32 index ) {return (float*)&m_spinImage[index];}
            card32 getNumm_spinImages() {return m_spinImage.getDim();}
            std::string getm_spinImageName(card32 index ) {return "-";}
            

        public:
            SpinImageSignature();
            virtual ~SpinImageSignature();

            /// create spin image based on given parameters
			virtual void create(Vector3f center, Vector3f normal, float32 radius, const std::vector<Vector3f> &points, card32 spinSize = 16) { create(center, normal, radius, points, std::vector<Vector3f>(), spinSize); }
            virtual void create(Vector3f center, Vector3f normal, float32 radius, const std::vector<Vector3f> &points, const std::vector<Vector3f> &normals, card32 spinSize = 16);

            /// copy one signature to another
            void assign(const Object* obj, X4CopyContext *context = NULL);

            /// compares two signatures (descriptor is required for context information)
            virtual float compare(LocalSignature * other, LocalDescriptor * myDescriptor );
			virtual float compare(LocalSignature * other ) { return compare(other, NULL); }
            virtual float compare(const SpinImageSignature& src);

            /// compare signature to the given one by computing NMI (using joint histogramms)
            virtual float compareByNMI(LocalSignature* other, LocalDescriptor* myDescriptor) ;

            /// used for signature based feature filtering
            bool badFeature(LocalDescriptor * myDescriptor) const {return false;}

            /// distance operator, based on center distance
            virtual float32 distance(const SpinImageSignature& src, const float& rad);

            /// distance operator, based on normal deviation
            virtual float32 deviation(const SpinImageSignature& src);

            /// get number of spin image elements
            inline size_t getNumElements() const { return m_spinImage.size(); }

            /// get spin image elements
            inline float32 getElement(card32 i) const { return m_spinImage[i]; }
			inline float32& operator[](card32 i) { return m_spinImage[i]; }

			/// resize spin image (this will set all elements to 0)
			inline void setNumElements(card32 num) { setNumm_spinImages(num); }

			//virtual void write(OutputObjectStream *s) const;
			//virtual void read(InputObjectStream *s);	
	};

}
#endif