//---------------------------------------------------------------------------
#include "stdafx.h" 
//---------------------------------------------------------------------------
#include "Transformation.h"
#include "ClassMethods.h"
#include "ObjectClassProperty.h"
#include "NumericalClassProperty.h"
#include "DebugOutput.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

std::string getTransformationDesc( const card32 i )
{
	if ( i == TRANSFORMATION_EULER_ANGLES ){
		return "Euler Angles";
	} else if ( i == TRANSFORMATION_QUATERNIONS ){
		return "Quaternions";
	} else if ( i == TRANSFORMATION_ROTATION_VECTOR ){
		return "Rotation Vector";
	} else if ( i == TRANSFORMATION_ROTATION_VECTOR_EXACT ){
		return "Rotation Vector (exact)";
	} else {
		return "Undefined";
	}
}
IMPLEMENT_X4_CLASS( Transformation, 2 )
{
	BEGIN_CLASS_INIT( Transformation )
	ADD_CARD32_PROP_UPDATE_DESCR( mode, 2, NULL, getTransformationDesc, 4, NULL )
}

Transformation::Transformation()
{
	mode = TRANSFORMATION_QUATERNIONS;
	reset();
}

Transformation::~Transformation()
{
}

void Transformation::add( const StaticVector<float32,7> step )
{
	state += step;
	/* normalization is done in cleanup() called explicitly
	if ( mode == TRANSFORMATION_QUATERNIONS ){
		// normalize quaternion
		float32 qnorm = sqrt( state[3]*state[3] + state[4]*state[4] + state[5]*state[5] + state[6]*state[6] );
		state[3] /= qnorm;
		state[4] /= qnorm;
		state[5] /= qnorm;
		state[6] /= qnorm;
	}
	*/
}

void Transformation::reset()
{
	if ( mode == TRANSFORMATION_EULER_ANGLES ){
		for ( unsigned int n = 0; n < 7; n++ ) state[n] = 0;
	} else if ( mode == TRANSFORMATION_QUATERNIONS ){
		state[0] = 0;	// Translation x
		state[1] = 0;	// Translation y
		state[2] = 0;	// Translation z
		state[3] = 1;	// q0
		state[4] = 0;	// q1
		state[5] = 0;	// q2
		state[6] = 0;	// q3
	} else if ( mode == TRANSFORMATION_ROTATION_VECTOR_EXACT
			 || mode == TRANSFORMATION_ROTATION_VECTOR ) {
		for ( unsigned int n = 0; n < 7; n++ ) state[n] = 0;
	}
}

Matrix3f Transformation::getRotation() const
{
	if ( mode == TRANSFORMATION_EULER_ANGLES ){
		return eulerAnglesR();
	} else if ( mode == TRANSFORMATION_QUATERNIONS ){
		return Quaternion( state[3], state[4], state[5], state[6] ).getRotMat();
	} else if ( mode == TRANSFORMATION_ROTATION_VECTOR_EXACT
			 || mode == TRANSFORMATION_ROTATION_VECTOR ){
		return RotationVector( state[3], state[4], state[5] ).getRotMat();
	}
	return IDENTITY3F;
}

void Transformation::jacobian( StaticMatrix<float32,7,3>& J, const Vector3f& pos ) const
{
	if ( mode == TRANSFORMATION_EULER_ANGLES )
	{
		jacobianEulerAngles( J, pos );	
	} 
	else if ( mode == TRANSFORMATION_QUATERNIONS )
	{		
		jacobianQuaternion( J, pos, Quaternion( state[3], state[4], state[5], state[6] ) );		
	}
	else if ( mode == TRANSFORMATION_ROTATION_VECTOR_EXACT)
	{		
		jacobianRotationVector( J, pos, false, makeVector3f( state[3], state[4], state[5] ) );
	}
	else if ( mode == TRANSFORMATION_ROTATION_VECTOR)
	{
		jacobianRotationVector( J, pos, true );		
	}
}

Matrix3f Transformation::eulerAnglesR() const
{
	float32 sinAlpha = sin( state[3] );
	float32 cosAlpha = cos( state[3] );
	Matrix3f rotX = makeMatrix3f( 1,0,0,0,cosAlpha,-sinAlpha,0,sinAlpha,cosAlpha );
	float32 sinBeta = sin( state[4] );
	float32 cosBeta = cos( state[4] );
	Matrix3f rotY = makeMatrix3f( cosBeta,0,sinBeta, 0,1,0, -sinBeta,0,cosBeta );
	float32 sinGamma = sin( state[5] );
	float32 cosGamma = cos( state[5] );
	Matrix3f rotZ = makeMatrix3f( cosGamma,-sinGamma,0, sinGamma,cosGamma,0, 0,0,1);
	return rotX * rotY * rotZ;
}

void Transformation::hessianRotationVector( Vector3f &H, const card32 i, const card32 j, const Vector3f &pos)
{
	if (i<3 || j<3)
	{
		// second derivative for translation is zero
		H = NULL_VECTOR3F;
		return;
	}
	
	Matrix4f dF1, dF2, dF3;
	Vector4f dq; // dq/dv_j
	Matrix4f F1, F2, F3;
	Vector4f g;  // dq/dv_i  p.19. col 1	
	Vector4f dg; // dg/dv_j
	Vector4f h1, h2, h3; // columns 

	// rotation std::vector (for convenience)
	Vector3f v = makeVector3f(state[3], state[4], state[5]);
	
	// convert to quaternion			
	float32 norm_v = float32(sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]));
	float32 s = float32(sin(norm_v/2.0))/norm_v;
	Quaternion q( cos(norm_v/2.0), v[0]*s, v[1]*s, v[2]*s );

	// helper var
	float32 a = 0.5*(q[0]-2*s)/(norm_v*norm_v); // a/(2v^3) p.19 col 1
	float32 b = v[0]*(-s-3*a)/(2*norm_v*norm_v);
	
	F1 = makeMatrix4f(	q[0],  q[1], -q[2], -q[3],
					   -q[3],  q[2],  q[1], -q[0],
					    q[2],  q[3],  q[0],  q[1],
						   0,     0,     0,  1.0f );
	F2 = makeMatrix4f(	q[3],  q[2],  q[1],  q[0],
					    q[0], -q[1],  q[2], -q[3],
					   -q[1], -q[0],  q[3],  q[2],
						   0,     0,     0,  1.0f );
	F3 = makeMatrix4f( -q[2],  q[3], -q[0],  q[1],
					    q[1],  q[0],  q[3],  q[2],
					    q[0], -q[1], -q[2],  q[3],
						   0,     0,     0,  1.0f );	
	switch(i)
	{
	case 3:	
		g = makeVector4f( -v[0]* 0.5f * s, 
						s + v[1]*v[1]*a, 
						v[1]*v[2]*a, 
						v[1]*v[3]*a);
		switch(j)
		{
		case 3:
			dg = makeVector4f( -2*s + 0.5f*a*(-v[0]*v[0]),
									  3.0f*a*v[0] + v[0]*b*v[0]*v[0],
									  a*v[1] + v[0]*b*v[0]*v[1],
									  a*v[2] + v[0]*b*v[0]*v[2] );
			break;
		case 4:
			dg = makeVector4f(        0.5f*a*(-v[0]*v[1]),
									  a*v[1] + v[1]*b*v[0]*v[0],
									  a*v[0] + v[1]*b*v[0]*v[1],
											   v[1]*b*v[0]*v[2] );
			break;
		case 5:
			dg = makeVector4f(        0.5f*a*(-v[0]*v[2]),
									  a*v[2] + v[2]*b*v[0]*v[0],
									           v[2]*b*v[0]*v[1],
									  a*v[0] + v[2]*b*v[0]*v[2] );
			break;
		}
	case 4:
		g = makeVector4f( -v[1]* 0.5f * s, 
						v[1]*v[2]*a, 
						s + v[2]*v[2]*a, 
						v[2]*v[3]*a);
		switch(j)
		{
		case 3:
			dg = makeVector4f(        0.5f*a*(-v[0]*v[1]),
									  a*v[1] + v[0]*b*v[0]*v[1],
									  a*v[0] + v[0]*b*v[1]*v[1],
									           v[0]*b*v[1]*v[2] );
			break;
		case 4:
			dg = makeVector4f( -2*s + 0.5f*a*(-v[1]*v[1]),
									  a*v[0] + v[1]*b*v[0]*v[1],
								 3.0f*a*v[1] + v[1]*b*v[1]*v[1],
									  a*v[2] + v[1]*b*v[1]*v[2] );
			break;
		case 5:
			dg = makeVector4f(        0.5f*a*(-v[1]*v[2]),
									           v[2]*b*v[0]*v[1],
									  a*v[2] + v[2]*b*v[1]*v[1],
									  a*v[1] + v[2]*b*v[1]*v[2] );
			break;
		}
		break;
	case 5:
		g = makeVector4f( -v[2]* 0.5f * s, 
						v[1]*v[3]*a, 
						v[2]*v[3]*a, 
						s + v[3]*v[3]*a);
		switch(j)
		{
		case 3:
			dg = makeVector4f(        0.5f*a*(-v[0]*v[2]),
									  a*v[2] + v[0]*b*v[0]*v[2],
									           v[0]*b*v[1]*v[2],
									  a*v[0] + v[0]*b*v[2]*v[2] );
			break;
		case 4:
			dg = makeVector4f(        0.5f*a*(-v[1]*v[2]),
									           v[1]*b*v[0]*v[2],
								      a*v[2] + v[1]*b*v[1]*v[2],
									  a*v[1] + v[1]*b*v[2]*v[2] );
			break;
		case 5:
			dg = makeVector4f( -2*s + 0.5f*a*(-v[2]*v[2]),
									  a*v[0] + v[2]*b*v[0]*v[2],
									  a*v[1] + v[2]*b*v[1]*v[2],
								 3.0f*a*v[2] + v[2]*b*v[2]*v[2] );
			break;
		}
		break;
	}
	switch (j)
	{
	case 3:
		dq = makeVector4f( -v[0]* 0.5f * s, 
						s + v[1]*v[1]*a, 
						v[1]*v[2]*a, 
						v[1]*v[3]*a);
		break;
	case 4:
		dq = makeVector4f( -v[1]* 0.5f * s, 
						v[1]*v[2]*a, 
						s + v[2]*v[2]*a, 
						v[2]*v[3]*a);
		break;
	case 5:
		dq = makeVector4f( -v[2]* 0.5f * s, 
						v[1]*v[3]*a, 
						v[2]*v[3]*a, 
						s + v[3]*v[3]*a);
		break;
	}

	dF1 = makeMatrix4f(	dq[0],  dq[1], -dq[2], -dq[3],
					   -dq[3],  dq[2],  dq[1], -dq[0],
					    dq[2],  dq[3],  dq[0],  dq[1],
						    0,      0,      0,   1.0f );
	dF2 = makeMatrix4f(	dq[3],  dq[2],  dq[1],  dq[0],
					    dq[0], -dq[1],  dq[2], -dq[3],
					   -dq[1], -dq[0],  dq[3],  dq[2],
						    0,      0,      0,   1.0f );
	dF3 = makeMatrix4f( -dq[2],  dq[3], -dq[0],  dq[1],
					     dq[1],  dq[0],  dq[3],  dq[2],
					     dq[0], -dq[1], -dq[2],  dq[3],
						     0,      0,      0,   1.0f );	

	h1 = dF1 * g + F1 * dg;
	h2 = dF2 * g + F2 * dg;
	h3 = dF3 * g + F3 * dg;

	H[0] = h1[0]*pos[0] + h2[0]*pos[1] + h3[0]*pos[2];
	H[1] = h1[1]*pos[0] + h2[1]*pos[1] + h3[1]*pos[2];
	H[2] = h1[2]*pos[0] + h2[2]*pos[1] + h3[2]*pos[2];
}

void Transformation::hessianQuaternion( Vector3f &H, const card32 i, const card32 j, const Vector3f &pos)
{
	switch(i)
	{	
	case 3:
		switch(j)
		{
		case 3:
			H[0] = 2.0f*pos[0]; 
			H[1] = 2.0f*pos[1]; 
			H[2] = 2.0f*pos[2];
			break;
		case 4:
			H[0] = 0.0f; 
			H[1] = 2.0f*pos[2]; 
			H[2] =-2.0f*pos[1];
			break;
		case 5:
			H[0] =-2.0f*pos[2]; 
			H[1] = 0.0f; 
			H[2] = 2.0f*pos[0];
			break;
		case 6:
			H[0] = 2.0f*pos[1]; 
			H[1] =-2.0f*pos[0]; 
			H[2] = 0.0f;
			break;
		}
	case 4:
		switch(j)
		{
		case 3: //q0
			H[0] = 0.0f; 
			H[1] = 2.0f*pos[2]; 
			H[2] =-2.0f*pos[1];
			break;
		case 4: //q1
			H[0] = 2.0f*pos[0]; 
			H[1] =-2.0f*pos[1]; 
			H[2] =-2.0f*pos[2];
			break;
		case 5: //q2
			H[0] = 2.0f*pos[1]; 
			H[1] = 2.0f*pos[0]; 
			H[2] = 0.0f;
			break;
		case 6: //q3
			H[0] = 2.0f*pos[2]; 
			H[1] = 0.0f; 
			H[2] = 2.0f*pos[0];
			break;
		}
	case 5:
		switch(j)
		{
		case 3: //q0
			H[0] =-2.0f*pos[2]; 
			H[1] = 0.0f; 
			H[2] = 2.0f*pos[0];
			break;
		case 4: //q1
			H[0] = 2.0f*pos[1]; 
			H[1] =-2.0f*pos[0]; 
			H[2] = 0.0f;
			break;
		case 5: //q2
			H[0] =-2.0f*pos[0]; 
			H[1] = 2.0f*pos[1]; 
			H[2] =-2.0f*pos[2];
			break;
		case 6: //q3
			H[0] = 0.0f; 
			H[1] = 2.0f*pos[2]; 
			H[2] = 2.0f*pos[1];
			break;
		}		
	case 6:
		switch(j)
		{
		case 3: //q0
			H[0] = 2.0f*pos[1]; 
			H[1] =-2.0f*pos[0]; 
			H[2] = 0.0f;
			break;
		case 4: //q1
			H[0] = 2.0f*pos[2]; 
			H[1] = 0.0f; 
			H[2] = 2.0f*pos[0];
			break;
		case 5: //q2
			H[0] = 0.0f; 
			H[1] = 2.0f*pos[2]; 
			H[2] = 2.0f*pos[1];
			break;
		case 6: //q3
			H[0] =-2.0f*pos[0]; 
			H[1] =-2.0f*pos[1]; 
			H[2] = 2.0f*pos[2];
			break;
		}
	}
}

void Transformation::jacobianEulerAngles( StaticMatrix<float32,7,3>& J, const Vector3f& pos )
{

	/*
	const float32 alpha = state[3];
	const float32 beta = state[4];
	const float32 gamma = state[5];
	*/

	/*
	J[0][0] = -1;
	J[1][0] = 0;
	J[2][0] = 0;
	J[3][0] = 0;
	J[4][0] = pos[2] - c[2];
	J[5][0] = -( pos[1] - c[1] );
	J[6][0] = 0;
	// y
	J[0][1] = 0;
	J[1][1] = -1;
	J[2][1] = 0;
	J[3][1] = -( pos[2] - c[2] );
	J[4][1] = 0;
	J[5][1] = pos[0] - c[0];
	J[6][1] = 0;
	// z
	J[0][2] = 0;
	J[1][2] = 0;
	J[2][2] = -1;
	J[3][2] = pos[1] - c[1];
	J[4][2] = -( pos[0] - c[0] );
	J[5][2] = 0;
	J[6][2] = 0;
	*/

	/*
	// x
	J[0][0] = -1;
	J[1][0] = 0;
	J[2][0] = 0;
	J[3][0] = 0;
	J[4][0] = sin(beta)*cos(gamma)*(pos[0]-c[0])+sin(beta)*sin(gamma)*(pos[1]-c[1])+cos(beta)*(pos[2]-c[2]);
	J[5][0] = cos(beta)*sin(gamma)*(pos[0]-c[0])-cos(beta)*cos(gamma)*(pos[1]-c[1]);
	J[6][0] = 0;
	// y
	J[0][1] = 0;
	J[1][1] = -1;
	J[2][1] = 0;
	J[3][1] = -(cos(alpha)*sin(beta)*cos(gamma)+sin(alpha)*sin(gamma))*(pos[0]-c[0])-(cos(alpha)*sin(beta)*sin(gamma)-sin(alpha)*cos(gamma))*(pos[1]-c[1])-cos(alpha)*cos(beta)*(pos[2]-c[2]);
	J[4][1] = -sin(alpha)*cos(beta)*cos(gamma)*(pos[0]-c[0])-sin(alpha)*cos(beta)*sin(gamma)*(pos[1]-c[1])+sin(alpha)*sin(beta)*(pos[2]-c[2]);
	J[5][1] = -(-sin(alpha)*sin(beta)*sin(gamma)-cos(alpha)*cos(gamma))*(pos[0]-c[0])-(sin(alpha)*sin(beta)*cos(gamma)-cos(alpha)*sin(gamma))*(pos[1]-c[1]);
	J[6][1] = 0;
	// z
	J[0][2] = 0;
	J[1][2] = 0;
	J[2][2] = -1;
	J[3][2] = -(-sin(alpha)*sin(beta)*cos(gamma)+cos(alpha)*sin(gamma))*(pos[0]-c[0])-(-sin(alpha)*sin(beta)*sin(gamma)-cos(alpha)*cos(gamma))*(pos[1]-c[1])+sin(alpha)*cos(beta)*(pos[2]-c[2]);
	J[4][2] = -cos(alpha)*cos(beta)*cos(gamma)*(pos[0]-c[0])-cos(alpha)*cos(beta)*sin(gamma)*(pos[1]-c[1])+cos(alpha)*sin(beta)*(pos[2]-c[2]);
	J[5][2] = -(-cos(alpha)*sin(beta)*sin(gamma)+sin(alpha)*cos(gamma))*(pos[0]-c[0])-(cos(alpha)*sin(beta)*cos(gamma)+sin(alpha)*sin(gamma))*(pos[2]-c[1]);
	J[6][2] = 0;
	*/

	// small angle approx
	// x
	J[0][0] = 1; 
	J[1][0] = 0; 
	J[2][0] = 0; 
	J[3][0] = 0;	      
	J[4][0] = pos[2];  
	J[5][0] = -pos[1];	
	J[6][0] = 0;
	// y
	J[0][1] = 0; 
	J[1][1] = 1; 
	J[2][1] = 0; 
	J[3][1] = -pos[2]; 
	J[4][1] = 0;       
	J[5][1] = pos[0];	
	J[6][1] = 0;
	// z
	J[0][2] = 0; 
	J[1][2] = 0; 
	J[2][2] = 1; 
	J[3][2] = pos[1];  
	J[4][2] = -pos[0]; 
	J[5][2] = 0;		
	J[6][2] = 0;
}



void Transformation::jacobianQuaternion( StaticMatrix<float32,7,3>& J, const Vector3f& pos, const Quaternion& q )
{
	J[0][0] = 1;
	J[1][0] = 0; 
	J[2][0] = 0; 
	J[3][0] = 2*(  q[0]*pos[0] + q[3]*pos[1] - q[2]*pos[2]);
	J[4][0] = 2*(  q[1]*pos[0] + q[2]*pos[1] + q[3]*pos[2]);
	J[5][0] = 2*( -q[2]*pos[0] + q[1]*pos[1] - q[0]*pos[2]);
	J[6][0] = 2*( -q[3]*pos[0] + q[0]*pos[1] + q[1]*pos[2]);

	J[0][1] = 0;
	J[1][1] = 1; 
	J[2][1] = 0; 
	J[3][1] = 2*( -q[3]*pos[0] + q[0]*pos[1] + q[1]*pos[2]);
	J[4][1] = 2*(  q[2]*pos[0] - q[1]*pos[1] + q[0]*pos[2]);
	J[5][1] = 2*(  q[1]*pos[0] + q[2]*pos[1] + q[3]*pos[2]);
	J[6][1] = 2*( -q[0]*pos[0] - q[3]*pos[1] + q[2]*pos[2]);

	J[0][2] = 0;
	J[1][2] = 0; 
	J[2][2] = 1; 
	J[3][2] = 2*( q[2]*pos[0] - q[1]*pos[1] + q[0]*pos[2]);
	J[4][2] = 2*( q[3]*pos[0] - q[0]*pos[1] - q[1]*pos[2]);
	J[5][2] = 2*( q[0]*pos[0] + q[3]*pos[1] - q[2]*pos[2]);
	J[6][2] = 2*( q[1]*pos[0] + q[2]*pos[1] + q[3]*pos[2]);
}

void Transformation::jacobianRotationVector( StaticMatrix<float32,7,3>& J, const Vector3f& pos, bool approx, Vector3f v )
{	
	float32 norm_v = float32(sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]));
	float32 eps = 1e-6f;
	if (approx || norm_v < eps)
	{
		/// small angle approximation!
		J[0][0] = 1;
		J[1][0] = 0; 
		J[2][0] = 0; 
		J[3][0] =  0.0f;
		J[4][0] = -1.0f * pos[2];
		J[5][0] =  1.0f * pos[1];
		J[6][0] = 0; // dummy

		J[0][1] = 0;
		J[1][1] = 1; 
		J[2][1] = 0; 
		J[3][1] =  1.0f * pos[2];
		J[4][1] =  0.0f;
		J[5][1] = -1.0f * pos[0];
		J[6][1] = 0; // dummy

		J[0][2] = 0;
		J[1][2] = 0; 
		J[2][2] = 1; 
		J[3][2] = -1.0f * pos[1];
		J[4][2] =  1.0f * pos[0];
		J[5][2] =  0.0f;
		J[6][2] = 0; // dummy
	}
	else
	{		
		// convert to quaternion		
		float32 s = float32(sin(norm_v/2.0))/norm_v;
		Quaternion q( cos(norm_v/2.0), v[0]*s, v[1]*s, v[2]*s );

		Matrix4f F1, F2, F3;
		Vector4f g1, g2, g3, r11, r12, r13, r21, r22, r23, r31, r32, r33;
		F1 = makeMatrix4f(	q[0],  q[1], -q[2], -q[3],
						   -q[3],  q[2],  q[1], -q[0],
						    q[2],  q[3],  q[0],  q[1],
							   0,     0,     0,  1.0f );
		F2 = makeMatrix4f(	q[3],  q[2],  q[1],  q[0],
						    q[0], -q[1],  q[2], -q[3],
						   -q[1], -q[0],  q[3],  q[2],
							   0,     0,     0,  1.0f );
		F3 = makeMatrix4f( -q[2],  q[3], -q[0],  q[1],
						    q[1],  q[0],  q[3],  q[2],
						    q[0], -q[1], -q[2],  q[3],
							   0,     0,     0,  1.0f );
		float32 a = 0.5*(q[0]-2*s)/(norm_v*norm_v);
		g1 = makeVector4f( -v[0]* 0.5f * s, 
							s + v[1]*v[1]*a, 
							v[1]*v[2]*a, 
							v[1]*v[3]*a);
		g2 = makeVector4f( -v[1]* 0.5f * s, 
							v[1]*v[2]*a, 
							s + v[2]*v[2]*a, 
							v[2]*v[3]*a);
		g3 = makeVector4f( -v[2]* 0.5f * s, 
							v[1]*v[3]*a, 
							v[2]*v[3]*a, 
							s + v[3]*v[3]*a);
		r11 = F1*g1;
		r12 = F1*g2;
		r13 = F1*g3;
		r21 = F2*g1;
		r22 = F2*g2;
		r23 = F2*g3;
		r31 = F3*g1;
		r32 = F3*g2;
		r33 = F3*g3;
		///x
		J[0][0] = 1;
		J[1][0] = 0; 
		J[2][0] = 0; 
		J[3][0] = r11[0]*pos[0] + r12[0]*pos[1] + r13[0]*pos[2]; 
		J[4][0] = r11[1]*pos[0] + r12[1]*pos[1] + r13[1]*pos[2];
		J[5][0] = r11[2]*pos[0] + r12[2]*pos[1] + r13[2]*pos[2];
		J[6][0] = 0; // dummy
		///y
		J[0][1] = 0;
		J[1][1] = 1; 
		J[2][1] = 0; 
		J[3][1] = r21[0]*pos[0] + r22[0]*pos[1] + r23[0]*pos[2];
		J[4][1] = r21[1]*pos[0] + r22[1]*pos[1] + r23[1]*pos[2];
		J[5][1] = r21[2]*pos[0] + r22[2]*pos[1] + r23[2]*pos[2];
		J[6][1] = 0; // dummy
		///z
		J[0][2] = 0;
		J[1][2] = 0; 
		J[2][2] = 1; 
		J[3][2] = r31[0]*pos[0] + r32[0]*pos[1] + r33[0]*pos[2];
		J[4][2] = r31[1]*pos[0] + r32[1]*pos[1] + r33[1]*pos[2];
		J[5][2] = r31[2]*pos[0] + r32[2]*pos[1] + r33[2]*pos[2];
		J[6][2] = 0; // dummy
	}
}
/*
void Transformation::hessian( Vector3f & Hij, const Vector3f& pos, card32 i, card32 j ) const
{
	if ( mode == TRANSFORMATION_ROTATION_VECTOR_EXACT)
	{		
		hessianRotationVector( Hij, pos, i, j, makeVector3f( state[3], state[4], state[5] ) );
	}
	else
	{
		debugOutput << "no hessian implemented for other types than rotation std::vector exact.\n";
		debugOutput.flush();
	}
}

void Transformation::hessianRotationVector( Vector3f & Hij, const Vector3f& pos, card32 i, card32 j, Vector3f v )
{	
	// todo
	debugOutput << "not implemented yet.\n";
	debugOutput.flush();
	
}

Matrix3d Transformation::rotationAxisR() const
{
	float32 t = 1.0f - cos( state[3] );
	float32 sinPhi = sin( state[3] );
	float32 cosPhi = cos( state[3] );
	float32 rx = state[4];
	float32 ry = state[5];
	float32 rz = state[6];
	return makeMatrix3d(
		t*rx*rx + cosPhi,    t*rx*ry + sinPhi*rz, t*rx*rz - sinPhi*ry,
		t*rx*ry - sinPhi*rz, t*ry*ry + cosPhi,    t*ry*rz + sinPhi*rx,
		t*rx*rz + sinPhi*ry, t*ry*rz - sinPhi*rx, t*rz*rz + cosPhi );
}

void Transformation::jacobianRotationAxis( StaticMatrix<float32,7,3>& J, const Vector3d& pos ) const
{
	/*
	float32 t = 1.0f - cos( state[3] );
	float32 sinPhi = sin( state[3] );
	float32 cosPhi = cos( state[3] );
	float32 rx = state[4];
	float32 ry = state[5];
	float32 rz = state[6];
	float32 A = pos[0]*rx*rx - pos[0] + pos[1]*rx*ry + pos[2]*rx*rz; 
	float32 B = pos[1]*rz - pos[2]*ry;
	float32 C = pos[0]*rx*ry + pos[1]*ry*ry - pos[1] + pos[2]*ry*rz;
	float32 D = -pos[0]*rz + pos[2]*rx;
	float32 E = pos[0]*rx*rz + pos[1]*ry*rz + pos[2]*rz*rz - pos[2];
	float32 F = pos[0]*ry - pos[1]*rx;
	// x
	J[0][0] = 1; 
	J[1][0] = 0; 
	J[2][0] = 0; 
	J[3][0] = t*(2*pos[0]*rx + pos[1]*ry + pos[2]*rz); 
	J[4][0] = pos[1]*t*rx - pos[2]*sinPhi;    
	J[5][0] = pos[1]*sinPhi + pos[2]*t*rx;	
	J[6][0] = sinPhi*A + cosPhi*B;
	// y	
	J[0][1] = 0; 
	J[1][1] = 1; 
	J[2][1] = 0; 
	J[3][1] = pos[0]*t*ry + pos[2]*sinPhi;  
	J[4][1] = t*(pos[0]*rx + 2*pos[1]*ry + pos[2]*rz); 
	J[5][1] = -pos[0]*sinPhi + pos[2]*t*ry;	
	J[6][1] = sinPhi*C + cosPhi*D;
	// z
	J[0][2] = 0; 
	J[1][2] = 0; 
	J[2][2] = 1; 
	J[3][2] = pos[0]*t*rz - pos[1]*sinPhi; 
	J[4][2] = pos[0]*sinPhi + pos[1]*t*rz;            
	J[5][2] = t*(pos[0]*rx + pos[1]*ry + 2*pos[2]*rz);		
	J[6][2] = sinPhi*E + cosPhi*F;

	float32 sinPhi = sin( state[3] );
	float32 cosPhi = cos( state[3] );
	float32 rx = state[4];
	float32 ry = state[5];
	float32 rz = state[6];

	// x
	J[0][0] = 1; 
	J[1][0] = 0; 
	J[2][0] = 0; 
	J[3][0] = (sinPhi*rx*rx-sinPhi)*pos[0]+(sinPhi*rx*ry+cosPhi*rz)*pos[1]+(sinPhi*rx*rz-cosPhi*ry)*pos[2]; 
	J[4][0] = 2*(1-cosPhi)*rx*pos[0]+(1-cosPhi)*ry*pos[1]+(1-cosPhi)*rz*pos[2];    
	J[5][0] = (1-cosPhi)*rx*pos[1]-sinPhi*pos[2];	
	J[6][0] = sinPhi*pos[1]+(1-cosPhi)*rx*pos[2];
	// y	
	J[0][1] = 0; 
	J[1][1] = 1; 
	J[2][1] = 0; 
	J[3][1] = (sinPhi*rx*ry-cosPhi*rz)*pos[0]+(sinPhi*ry*ry-sinPhi)*pos[1]+(sinPhi*ry*rz+cosPhi*rx)*pos[2];  
	J[4][1] = (1-cosPhi)*ry*pos[0]+sinPhi*pos[2]; 
	J[5][1] = (1-cosPhi)*rx*pos[0]+2*(1-cosPhi)*ry*pos[1]+(1-cosPhi)*rz*pos[2];	
	J[6][1] = -sinPhi*pos[0]+(1-cosPhi)*ry*pos[2];
	// z
	J[0][2] = 0; 
	J[1][2] = 0; 
	J[2][2] = 1; 
	J[3][2] = (sinPhi*rx*rz+cosPhi*ry)*pos[0]+(sinPhi*ry*rz-cosPhi*rx)*pos[1]+(sinPhi*rz*rz-sinPhi)*pos[2]; 
	J[4][2] = (1-cosPhi)*rz*pos[0]-sinPhi*pos[1];            
	J[5][2] = sinPhi*pos[0]+(1-cosPhi)*rz*pos[1];		
	J[6][2] = (1-cosPhi)*rx*pos[0]+(1-cosPhi)*ry*pos[1]+2*(1-cosPhi)*rz*pos[2];	
*/

void Transformation::setTranslation( Vector3f translation )
{
	state[0] = translation[0];
	state[1] = translation[1];
	state[2] = translation[2];
}

void Transformation::setQuaternionRotation( const Quaternion& quaternion )
{
	state[3] = quaternion[0];
	state[4] = quaternion[1];
	state[5] = quaternion[2];
	state[6] = quaternion[3];
}

Quaternion Transformation::getQuaternion() const
{
	return Quaternion( state[3], state[4], state[5], state[6] );
}

void Transformation::cleanup()
{
	if ( mode == TRANSFORMATION_QUATERNIONS ){
		Quaternion q = getQuaternion();
		q.normalize();
		setQuaternionRotation( q );
	}
}

Matrix4f Transformation::getTransformation() const
{
	Matrix3f R = getRotation();
	Vector3f t = getTranslation();
	Matrix4f T;
	for ( unsigned int i = 0; i < 3; i++ ){
		for ( unsigned int j = 0; j < 3; j++ ){
			T[i][j] = R[i][j];
		}
	}
	for ( unsigned int i = 0; i < 3; i++ ){
		T[3][i] = t[i];
	}
	return T;
}

void Transformation::println() 
{
	for ( unsigned int i = 0; i < 7; i++ )
	{
		debugOutput << state[i] << " ";
	}
	debugOutput << "\n";
	debugOutput.flush();
}