#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "TriangleMeshHashGrid.h"
//---------------------------------------------------------------------------
#include "SparseGrid.inline.h"
#include "VertexArray.h"
#include "TriangleBBIntersection.h"
#include "PointCloudTools.h"
#include "DebugRenderer.h"
//---------------------------------------------------------------------------
using namespace X4;


TriangleMeshHashGrid::TriangleMeshHashGrid(void)
{
}

TriangleMeshHashGrid::~TriangleMeshHashGrid(void)
{
	clear();
}


void X4::TriangleMeshHashGrid::setup( UnstructuredInCoreTriangleMesh * mesh, float cellSize )
{
	m_BoundingBox = getPCBBox(mesh);
	m_BoundingBox.addBorder(cellSize * 0.5f);
	m_CellSize = cellSize;

	VertexArray va_vert(mesh);
	VertexArray va_tri(mesh->getTriangles());

	float sampleSpacing = cellSize / 2.0f;
	for( unsigned triangle_index=0;triangle_index<va_tri.getNumElements();triangle_index++ )
	{
		Vector3i tri = va_tri.getIndex3i(triangle_index);
		Vector3f corner0 = va_vert.getPosition3f(tri[0]);
		Vector3f corner1 = va_vert.getPosition3f(tri[1]);
		Vector3f corner2 = va_vert.getPosition3f(tri[2]);
		Vector3f normal = ( (corner1 - corner0).crossProduct(corner2 - corner0));
		float length = norm(normal);
		if( sqrt(length) < cellSize*.1f )
			continue;

		normal /= length;

		// determine triangles bounding box in grid coordinates
		BoundingBox3f bbGrid;
		bbGrid.lowerCorner = bbGrid.upperCorner = getNearestCellCenter(corner0);
		bbGrid.addPoint(getNearestCellCenter(corner1));
		bbGrid.addPoint(getNearestCellCenter(corner2));
		bbGrid.addBorder(cellSize);

		// rasterize triangle (simplest method, if time becomes an issue -> hierarchical implementation)
		for( float x=bbGrid.lowerCorner[0];x<=bbGrid.upperCorner[0];x+=cellSize )
		{
			for( float y=bbGrid.lowerCorner[1];y<=bbGrid.upperCorner[1];y+=cellSize )
			{
				for( float z=bbGrid.lowerCorner[2];z<=bbGrid.upperCorner[2];z+=cellSize )
				{
					BoundingBox3f bb = getCellBoundingBox3f(makeVector3f(x,y,z));
					bb.addBorder(cellSize*0.001f);
					if( triangleIntersectsBB(corner0,corner1,corner2,bb))
					{
						GridCell * cell;
						Vector3i currentCellIndex = m_Grid.calcIndex(makeVector3f(x,y,z), m_BoundingBox.lowerCorner, cellSize);
						if(	m_Grid.getObject(currentCellIndex,cell) )
						{
							if( cell->m_Indices.back() != triangle_index )
							{
								cell->m_Indices.push_back(triangle_index);
							}
						}
						else
						{
							cell = new GridCell;
							cell->m_Indices.push_back(triangle_index);
							m_Grid.setObject(currentCellIndex,cell);
						}
					}

				}
			}
		}

		continue;
	}

}

void X4::TriangleMeshHashGrid::clear()
{
	for( unsigned i=0;i<m_Grid.getNumElements();i++ )
	{
		delete m_Grid.getEntry(i);
	}
	m_Grid.clear();
}

TriangleMeshHashGrid::GridCell * X4::TriangleMeshHashGrid::getCell( const Vector3f &p)
{
	GridCell * ret;
	Vector3i index = m_Grid.calcIndex(p, m_BoundingBox.lowerCorner, m_CellSize);
	if(	!m_Grid.getObject(index,ret) )
		return NULL;
	else
		return ret;
}

TriangleMeshHashGrid::GridCell * X4::TriangleMeshHashGrid::getCell( mpcard i ) 
{
	return m_Grid.getEntry(i);
}

X4::BoundingBox3f X4::TriangleMeshHashGrid::getCellBoundingBox3f( const Vector3f &p ) const
{
	Vector3f cellCenter = getNearestCellCenter(p);
	BoundingBox3f bb;
	bb.lowerCorner = bb.upperCorner = cellCenter;
	bb.addBorder(m_CellSize/2.0f);
	return bb;
}

X4::BoundingBox3f X4::TriangleMeshHashGrid::getCellBoundingBox3f( mpcard i ) const
{
	Vector3i index = m_Grid.getKey(i);
	Vector3f cellCenter = m_BoundingBox.lowerCorner + makeVector3f(0.5f + index[0],0.5f + index[1],0.5f + index[2])*m_CellSize;
	BoundingBox3f bb;
	bb.lowerCorner = bb.upperCorner = cellCenter;
	bb.addBorder(m_CellSize/2.0f);
	return bb;
}

X4::mpcard X4::TriangleMeshHashGrid::getNumCells() const
{
	return m_Grid.getNumElements();
}

X4::Vector3f X4::TriangleMeshHashGrid::getNearestCellCenter( const Vector3f &p ) const
{
	Vector3i index = m_Grid.calcIndex(p, m_BoundingBox.lowerCorner, m_CellSize);
	return m_BoundingBox.lowerCorner + makeVector3f(0.5f + index[0],0.5f + index[1],0.5f + index[2])*m_CellSize;
}

void X4::TriangleMeshHashGrid::drawWithDR( const Vector3f &color ) const
{
	for( mpcard i=0;i<m_Grid.getNumElements();i++)
	{
		const BoundingBox3f bb = getCellBoundingBox3f(i);
		debugRenderer->addBoundingBox(bb,color,1);
	}
}
