//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "UniformPointCloudResampling.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "PointSet.h"
#include "PointSetKNNQuery.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------



IMPLEMENT_X4_CLASS( UniformPointCloudResampling ,0)
{
	BEGIN_CLASS_INIT( UniformPointCloudResampling );
}


// --


UniformPointCloudResampling::UniformPointCloudResampling() {
	maxSpacing = 1;
}

void UniformPointCloudResampling::assign(const Object* obj, X4CopyContext *context) {
	copyObjectProperties(obj, this);
}

void UniformPointCloudResampling::setup(float32 maxSpacing) {
	this->maxSpacing = maxSpacing;
}

UniformPointCloudResampling::~UniformPointCloudResampling() {}


// -

struct PQEntry {
	card32 index;
	float32 dist;
	bool operator<(const PQEntry &other) const {
		return dist < other.dist;
	}
};


PointSet *UniformPointCloudResampling::resamplePointSet( const PointSet *ps ) 
{
	x4Assert(ps);
	const card32 numPoints = ps->getNumEntries();

	std::vector<card32> overlap;
	overlap.resize(numPoints, 0);
	std::vector< std::vector<card32> > overlapPoints;
	overlapPoints.resize(numPoints);
	
	std::vector<bool> deletePtn;
	deletePtn.resize(numPoints,false);
	AAT POS = ps->getAAT("position", 3, VAD::DATA_FORMAT_FLOAT32);
	AttachedIndexedOctree *octree = PointSetKNNQuery::createOctree(ps);
	PointSetKNNQuery knn(ps, octree, NULL_VECTOR3F, POS);
	
	priority_queue<PQEntry> removalCandidates;

	for (card32 i=0; i<numPoints; i++) {
		Vector3f pos = ps->get3f(i, POS);
		knn.setTargetPoint(pos);
		card32 index = 0;
		bool found;
		Vector3f currentPointPos;
		knn.getNextNearestPointIndex(index, found);
		x4Assert(found);
		currentPointPos = ps->get3f(index, POS);
		float32 minDistance = 1E25f;
		while (found && norm(currentPointPos - pos) <= maxSpacing) {
			overlap[i]++;
			overlapPoints[i].push_back(index);
			knn.getNextNearestPointIndex(index, found);
			if (found) {
				currentPointPos = ps->get3f(index, POS);
				float32 distance = norm(currentPointPos - pos);
				if (distance < minDistance) {
					minDistance = distance;
				}
			}
		}
		if (minDistance < 1E25) {
			PQEntry e;
			e.dist = minDistance;
			e.index = i;
			removalCandidates.push(e);
		}
	}
	
	card32 newNumPoints = numPoints;
	
	while (!removalCandidates.empty()) {
		PQEntry e = removalCandidates.top();
		removalCandidates.pop();
		if (overlap[e.index] > 1) {
			bool othersOk = true;
			const size_t otherCount = overlapPoints[e.index].size();
			for (card32 i=0; i<otherCount; i++) {
				if (overlap[overlapPoints[e.index][i]] <= 1) {
					othersOk = false;
				}
			}
			if (othersOk) {
				overlap[e.index]--;
				for (card32 i=0; i<otherCount; i++) {
					overlap[overlapPoints[e.index][i]]--;
				}
				deletePtn[e.index] = true;
				
				x4Assert(newNumPoints > 0);
				newNumPoints--;
			}
		}
	}
	
	PointSet *result = (PointSet*)(ps->copy());
	result->clearAndSetup(1, newNumPoints, ps->getDescr());
	card32 targetPos = 0;
	for (card32 i=0; i<numPoints; i++) {
		if (!deletePtn[i]) {
			x4Assert(targetPos < newNumPoints);
			result->setVertex(targetPos, ps->getDataPointer(i));
			targetPos++;
		}
	}
	
	delete octree;
	
	return result;
}














/*PatchGraph *UniformPointCloudResampling::resamplePatchGraph(const PatchGraph *pg) {
	x4Assert(pg);
	const TrajectorySet *ts = pg->getTrajectorySet();
	const card32 numPatches = ts->getNumPatches();
	const card32 numFrames = ts->getNumFrames();

	std::vector<card32> overlap;
	overlap.resize(numPatches, 0);
	std::vector< std::vector<card32> > overlapPoints;
	overlapPoints.resize(numPatches);
	std::vector<bool> deletePtn;
	deletePtn.resize(numPatches, false);

	AAT POS = ps->getAAT("position", 3, VAD::DATA_FORMAT_FLOAT32);
	PointSetKNNQuery knn[] = new PointSetKNNQuery[numFrames];
	PointSet *ps[] = new PointSet*[numFrames];
	
	for (card32 frameIndex=0; frameIndex<numFrames; frameIndex++) {
		AttachedIndexedOctree *octree = ts->fetchOctree(frameIndex);
		knn[frameIndex](ps, octree, NULL_VECTOR3F);
	}
	
	priority_queue<PQEntry> removalCandidates;

	for (card32 i=0; i<numPatches; i++) {
		Vector3f pos = ps->get3f(i, POS);
		knn.setTargetPoint(pos);
		card32 index = 0;
		bool found;
		Vector3f currentPointPos;
		knn.getNextNearestPointIndex(index, found);
		x4Assert(found);
		currentPointPos = ps->get3f(index, POS);
		float32 minDistance = 1E25f;
		while (found && norm(currentPointPos - pos) <= maxSpacing) {
			overlap[i]++;
			overlapPoints[i].push_back(index);
			knn.getNextNearestPointIndex(index, found);
			if (found) {
				currentPointPos = ps->get3f(index, POS);
				float32 distance = norm(currentPointPos - pos);
				if (distance < minDistance) {
					minDistance = distance;
				}
			}
		}
		if (minDistance < 1E25) {
			PQEntry e;
			e.dist = minDistance;
			e.index = i;
			removalCandidates.push(e);
		}
	}
	
	card32 newNumPatches = numPatches;
	
	while (!removalCandidates.empty()) {
		PQEntry e = removalCandidates.top();
		removalCandidates.pop();
		if (overlap[e.index] > 1) {
			bool othersOk = true;
			const card32 otherCount = overlapPoints[e.index].size();
			for (card32 i=0; i<otherCount; i++) {
				if (overlap[overlapPoints[e.index][i]] <= 1) {
					othersOk = false;
				}
			}
			if (othersOk) {
				overlap[e.index]--;
				for (card32 i=0; i<otherCount; i++) {
					overlap[overlapPoints[e.index][i]]--;
				}
				deletePtn[e.index] = true;
				x4Assert(newNumPatches > 0);
				newNumPatches--;
			}
		}
	}
	
	PointSet *result = (PointSet*)(ps->copy());
	result->clearAndSetup(1, newNumPatches, ps->getDescr());
	card32 targetPos = 0;
	for (card32 i=0; i<numPatches; i++) {
		if (!deletePtn[i]) {
			x4Assert(targetPos < newNumPatches);
			result->setVertex(targetPos, ps->getDataPointer(i));
			targetPos++;
		}
	}
	
	delete[] knn;
	for (card32 frameIndex=0; frameIndex<numFrames; frameIndex++) {
		delete ps[frameIndex];
	}
	delete[] ps;
	
	
	return result;
}			 */

