//---------------------------------------------------------------------------
#include "stdafx.h"
#include "UnstructuredInCorePointCloudManifold.h"
#include "UnstructuredInCoreTriangleMeshManifold.h"
#include "ObjectClassProperty.h"
#include "TopologyAttachment.h"
#include "InCorePCTopologyGraph.h"
#include <omp.h>
#include "PointCloudTools.h"
#include "SeparatorClassProperty.h"
#include <GeometricTools.h>
#include <PCCResampler.h>
#include "x3foundation_private.h"

//---------------------------------------------------------------------------
using namespace X4;

IMPLEMENT_X4_CLASS(UnstructuredInCorePointCloudManifold, 1)
{
	BEGIN_CLASS_INIT(UnstructuredInCorePointCloudManifold);

	ADD_SEPARATOR("Point cloud data");
	ADD_OBJECT_PROP(m_PointCloud, 0, UnstructuredInCorePointCloud::getClass(), false);
	ADD_OBJECT_PROP(m_Tpg, 0, InCorePCTopologyGraph::getClass(), false);
	ADD_OBJECT_PROP(m_Subsampled, 1, UnstructuredInCorePointCloud::getClass(), true);
}


//---------------------------------------------------------------------------
UnstructuredInCorePointCloudManifold::UnstructuredInCorePointCloudManifold()
{
	m_hIt = NULL;
	m_Tpg = NULL;
	m_PointCloud = NULL;
	m_Subsampled = NULL;
}

//---------------------------------------------------------------------------
UnstructuredInCorePointCloudManifold::~UnstructuredInCorePointCloudManifold()
{
	if (m_hIt) delete m_hIt;
	if (m_Subsampled) delete m_Subsampled;
}

//---------------------------------------------------------------------------
void UnstructuredInCorePointCloudManifold::setPointCloud(UnstructuredInCorePointCloud* pc)
{
	m_PointCloud = pc;
	if (pc)
	{
		TopologyAttachment* tatt = dynamic_cast<TopologyAttachment*>(m_PointCloud->getAttachments()->getData(TopologyAttachment::getDefaultName()));
		if (tatt && tatt->getTopology()) m_Tpg = tatt->getTopology();
		else m_Tpg = NULL;
	}
}

//---------------------------------------------------------------------------
void UnstructuredInCorePointCloudManifold::subsamplePointCloud(float epsilonFactor)
{
	if (m_Subsampled) delete m_Subsampled;
	m_Subsampled = copyToUnstructuredInCorePointCloud(getPointCloud());
	PCCResampler::resamplePC(m_Subsampled, X4::getMedianPointDistance(m_Subsampled) * epsilonFactor);
}

//---------------------------------------------------------------------------
float UnstructuredInCorePointCloudManifold::computeDistanceAndFillHash(mpcard startIndex, mpcard toIndex, std::vector<DVectorF>* path)
{
	if (m_Tpg == NULL) return 0;

	if (startIndex >= m_PointCloud->getNumPoints()) return -1;
	if (toIndex >= m_PointCloud->getNumPoints()) return -1;

	if (toIndex == startIndex) return 0;

	PointSet* ps = m_PointCloud->getPointSet();
	InCorePCTopologyGraph* tpg = m_Tpg;

	AAT posAAT = ps->getAAT("position");
	AAT distAAT = NULL_AAT;
	if (tpg->getEdges()->providesAttribute("dist"))
		distAAT = tpg->getDescr()->getToken("dist");

	struct TDistanceIndexPair_Pre
	{
		float dist;
		mpcard index;
		mpcard predecessor;
		bool operator < (const TDistanceIndexPair_Pre &o) const {return dist > o.dist;}
		TDistanceIndexPair_Pre(float d, mpcard i,mpcard pre) {dist = d; index = i; predecessor = pre;}
	};

	std::vector<float32> _tempDistanceField(ps->getHeight(), 1e30f);

	const int PRED_Unassigned = (int)_tempDistanceField.size();
	const int PRED_UnassignedAndTarget = PRED_Unassigned+1;
	std::vector<int> predecessor( _tempDistanceField.size(), PRED_Unassigned );
	priority_queue<TDistanceIndexPair_Pre> pq;

	// flag target points
	predecessor[toIndex] = PRED_UnassignedAndTarget;

	mpcard numTargetPoints = 1;
	pq.push(TDistanceIndexPair_Pre(0,startIndex,startIndex));

	while(!pq.empty())
	{
		mpcard index = pq.top().index;
		float currentDistance = pq.top().dist;
		mpcard pre = pq.top().predecessor;
		pq.pop();

		if( predecessor[index] < PRED_Unassigned)
			continue;

		_tempDistanceField[index] = currentDistance;
		if( predecessor[index] == PRED_UnassignedAndTarget )
		{
			numTargetPoints--;
			predecessor[index] = pre;
			if( numTargetPoints == 0)
				break;
		}
		else
			predecessor[index] = pre;

		card32 numNeighbors = tpg->getNumAdjacentVertices(index);
		for( card32 i=0;i<numNeighbors;i++ )
		{
			InCorePCTopologyGraph::EdgeIndex edge = 0;
			mpcard neighbour = tpg->getAdjacentVertex(index,i, &edge);
			if( _tempDistanceField[neighbour] < 1e30f )
				continue;

			float distance = currentDistance;
			if (distAAT == NULL_AAT)
				distance += norm(ps->get3f(index, posAAT) - ps->get3f(neighbour, posAAT));
			else
				distance += tpg->getEdges()->get1f(edge, distAAT);

			if( _tempDistanceField[neighbour] > distance )
			{
				pq.push(TDistanceIndexPair_Pre(distance,neighbour,index));
			}
		}
	}

	// go through all distances and put value into the hash table if such was computed
	for( card32 i=0; i < _tempDistanceField.size(); i++ )
	{
		if (i == startIndex)
		{
			setHashData(startIndex, i, HashData(0.0f));
			//setDistanceHashed(startIndex, i, 0);
		}else if (i == toIndex)
		{
			setHashData(startIndex, i, HashData(_tempDistanceField[i]));
			//setDistanceHashed(startIndex, i, _tempDistanceField[i]);
		} else if (predecessor[i] < PRED_Unassigned)// _tempDistanceField[i] < 1e30f)
		{
			//float data = getDistanceHashed(startIndex, i);
			float data = getHashData(startIndex, i)->distance;

			if (data < 0 || data > _tempDistanceField[i]) 
				setHashData(startIndex, i, HashData(_tempDistanceField[i]));
				//setDistanceHashed(startIndex, i, _tempDistanceField[i]);
		}
	}

	// collect path
	if (path != NULL && startIndex != toIndex && predecessor[toIndex] < PRED_Unassigned)
	{
		path->clear();
		card32 index = toIndex;
		path->push_back(makeDVectorL(3, ps->get3f(index, posAAT).data()));
		while (predecessor[index] < PRED_Unassigned)
		{
			//InCorePCTopologyGraph::EdgeIndex edge;
			if (tpg->areVerticesAdjacent(index, predecessor[index]) == false)
				error(std::string(__FUNCTION__) + " : predecessor found, however not in the topology. This can not be the case!!!");

			// get coordinates on both ends
			//path->push_back(edge);
			index = predecessor[index];
			path->push_back(makeDVectorL(3, ps->get3f(index, posAAT).data()));

			if (index == startIndex) break;
		}

	}


	return _tempDistanceField[toIndex];
}

//---------------------------------------------------------------------------
std::vector<float> UnstructuredInCorePointCloudManifold::computeDistancesAndFillHash(mpcard i, const std::vector<mpcard>& j, std::vector<std::vector<DVectorF>>* paths)
{
	if (paths) paths->resize(j.size());

	std::vector<float> dist;
	for (unsigned i=0; i < j.size(); i++)
	{
		dist[i] = computeDistanceAndFillHash(i, j[i], paths ? &(paths->at(i)) : NULL);
	}
	return dist;
}

//---------------------------------------------------------------------------
mpcard UnstructuredInCorePointCloudManifold::getNumPoints()
{
	if (m_PointCloud == NULL) return 0;
	return m_PointCloud->getNumPoints();
}

#if 0
//---------------------------------------------------------------------------
DiscreteManifoldPoint UnstructuredInCorePointCloudManifold::getPoint(mpcard index)
{
	if (m_PointCloud == NULL) return DiscreteManifoldPoint(NON_VALID_POINT, DVectorF());
	AAT posAAT = m_PointCloud->getAAT("position");
	Vector3f pos = m_PointCloud->getPointSet()->get3f(index, posAAT);
	return DiscreteManifoldPoint(index, makeDVector(3, pos.data()));
}
#endif 

//---------------------------------------------------------------------------
ManifoldPoint* UnstructuredInCorePointCloudManifold::createPoint(const DVectorF& coordinate)
{
	mpcard index = findNearestPoint(coordinate);
	return createPoint(index, coordinate);
}

//---------------------------------------------------------------------------
ManifoldPoint* UnstructuredInCorePointCloudManifold::createPoint(mpcard index, const DVectorF& coordinate)
{
	if (index == NON_VALID_POINT)
	{
		DVectorF nullCoord(3);
		nullCoord.setZero();
		return new DiscreteManifoldPoint(NON_VALID_POINT, nullCoord, nullCoord);
	}

	if (m_PointCloud == NULL) return new DiscreteManifoldPoint(NON_VALID_POINT, DVectorF(), DVectorF());
	AAT posAAT = m_PointCloud->getAAT("position");
	AAT normalAAT = m_PointCloud->getAAT("normal");
	Vector3f pos = m_PointCloud->getPointSet()->get3f(index, posAAT);
	Vector3f normal = m_PointCloud->getPointSet()->get3f(index, normalAAT);
	
	if (coordinate.getDim() == 3)
		return new DiscreteManifoldPoint(index, coordinate, makeDVectorL(3, normal.data()));
	else
		return new DiscreteManifoldPoint(index, makeDVectorL(3, pos.data()), makeDVectorL(3, normal.data()));
}

//---------------------------------------------------------------------------
mpcard UnstructuredInCorePointCloudManifold::findNearestPoint(const DVectorF& point)
{
	if (m_PointCloud == NULL) return NON_VALID_POINT;

	if (m_bDirtyGeodesics || m_hIt == NULL)
	{
		if (m_hIt) delete m_hIt;
		m_hIt = new HierarchicalKNNIterator(m_PointCloud, 32, NULL);
	}

	Vector3f pos;
	pos[0] = point[0];
	pos[1] = point[1];
	pos[2] = point[2];

	m_hIt->setSeekPointAndReset(pos);
	return m_hIt->getCurrentPointIndex();
}

//---------------------------------------------------------------------------
float UnstructuredInCorePointCloudManifold::getMedianPointDistance()
{
	if (m_PointCloud == NULL) return -1;
	return X4::getMedianPointDistance(m_PointCloud);
}

//---------------------------------------------------------------------------
DiscreteManifold* X4::createDiscreteManifold(UnstructuredInCorePointCloud* pc)
{
	UnstructuredInCorePointCloudManifold* pcm = new UnstructuredInCorePointCloudManifold;
	
	UnstructuredInCorePointCloud* pcCopy = (UnstructuredInCorePointCloud*)pc->copy();
	pcm->setPointCloud(pcCopy);

	return pcm;
}

//---------------------------------------------------------------------------
DiscreteManifold* X4::createDiscreteManifold(UnstructuredInCoreTriangleMesh* pc)
{
	UnstructuredInCoreTriangleMeshManifold* pcm = new UnstructuredInCoreTriangleMeshManifold;

	UnstructuredInCoreTriangleMesh* pcCopy = (UnstructuredInCoreTriangleMesh*)pc->copy();
	pcm->setPointCloud(pcCopy);

	return pcm;
}
