//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "VSTreeOperations.h"
#include "SeparatorClassProperty.h"
#include "MathTools.h"
#include "PCA.h"
#include "VSTreeIterators.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

IMPLEMENT_X4_CLASS( VSTreeOperations, 0 )
{
	BEGIN_CLASS_INIT( VSTreeOperations )
	ADD_SEPARATOR( "Build" )
	ADD_CARD32_PROP( splitSize, 0 )
	ADD_CARD32_PROP( maxDepth, 0 )
	ADD_FLOAT32_PROP( maxTAngle, 0 )
}

VSTreeOperations::VSTreeOperations()
{
	splitSize = 10;
	maxDepth = 6;
	maxTAngle = 0.5f;
}

VSTreeOperations::~VSTreeOperations()
{
}

bool VSTreeOperations::buildVSTree( VSTree* tree, PointSet* ps )
{
	if ( ps == NULL ) return false;
	if ( !ps->providesAttribute( "position" ) ) return false;
	AAT POSITION = ps->getAAT( "position" );

	BoundingBox3f bbox = ps->getBoundingBox();
	bbox.enlarge( 1.05f );
	const Vector3f& center = bbox.getCenter();
	float32 s = bbox.getMaxSideLength() / 2.0f;
	bbox.lowerCorner = center - makeVector3f( s,s,s );
	bbox.upperCorner = center + makeVector3f( s,s,s );
	
	// Build tree
	currentID = 0;
	tree->clear();
	std::vector<card32> indices; indices.clear();
	for ( unsigned int pIndex = 0; pIndex < ps->getNumEntries(); pIndex++ ){
		indices.push_back( pIndex );
	}
	tree->setup( buildVSTreeVNode( 0, indices, ps, bbox ) );
	return true;
}

VSTreeNode* VSTreeOperations::buildVSTreeVNode( const card32& depth, std::vector<card32> indices, PointSet* ps, BoundingBox3f& bbox )
{
	// Leaf reached
	if ( ( depth >= maxDepth ) || ( indices.size() < splitSize ) ) {
		VSTreeVNode* vNode = new VSTreeVNode();
		vNode->setID( currentID ); currentID++;
		vNode->setBBox( bbox );
		vNode->setIndices( indices );
		return vNode;
	}

	// Check if t-layer reached
	AAT NORMAL = ps->getAAT( "normal" );
	Vector3f avgNormal = makeVector3f( 0,0,0 );
	for ( unsigned int n = 0; n < indices.size(); n++ ){
		const card32 pIndex = indices[n];
		const Vector3f& normal = ps->get3f( pIndex, NORMAL );
		avgNormal = ( avgNormal * n + normal ) / float32( n + 1 );
	}
	avgNormal.normalize();
	float32 maxAngle = 1;
	for ( unsigned int n = 0; n < indices.size(); n++ ){
		const card32 pIndex = indices[n];
		const Vector3f& normal = ps->get3f( pIndex, NORMAL );
		float32 angle = normal * avgNormal;
		if ( angle < maxAngle )
			maxAngle = angle;
	}
	if ( maxAngle > maxTAngle ){
		return buildVSTreeTNode( depth+1, indices, ps );
	}

	// Subdivide
	AAT POSITION = ps->getAAT( "position" );
	VSTreeVNode* vNode = new VSTreeVNode();
	vNode->setID( currentID ); currentID++;
	vNode->setBBox( bbox );
	std::vector< std::vector<card32> > subIndices; subIndices.resize( 8 ); 
	for ( unsigned int n = 0; n < 8; n++ ) 
		subIndices[n].clear();
	for ( unsigned int n = 0; n < indices.size(); n++ ){
		const card32 pIndex = indices[n];
		const Vector3f& p = ps->get3f( pIndex, POSITION );
		if ( !bbox.inBoundingBox( p ) ){
			debugOutput << "Should not happen!\n"; debugOutput.flush();
			return NULL;
		}
		card32 octand = 0;
		for ( unsigned int i = 0; i < 3; i++ ){
			if ( p[i] >= ( bbox.lowerCorner[i] + bbox.upperCorner[i] ) / 2.0f )
				octand += power( 2,i );
		}
		subIndices[octand].push_back( pIndex );
	}
	for ( unsigned int n = 0; n < 8; n++ ){
		BoundingBox3f subBBox;
		float32 s = bbox.getMaxSideLength() / 2.0f;
		Vector3i I;
		I[2] = n/4;
		I[1] = ( n - I[2]*4 ) / 2;
		I[0] = ( n - I[2]*4 - I[1]*2);
		for ( unsigned int i = 0; i < 3; i++ ){
			float32 e = I[i] * s;
			subBBox.lowerCorner[i] = bbox.lowerCorner[i] + e;
			subBBox.upperCorner[i] = subBBox.lowerCorner[i] + s;
		}
		vNode->setChild( n, buildVSTreeVNode( depth+1, subIndices[n], ps, subBBox ) );
	}
	return vNode;
}

VSTreeNode* VSTreeOperations::buildVSTreeTNode( const card32& depth, std::vector<card32> indices, PointSet* ps )
{
	AAT POSITION = ps->getAAT( "position" );
	VSTreeTNode* tNode = new VSTreeTNode();
	tNode->setID( currentID ); currentID++;
	PCA3f pca; pca.clear();
	for( unsigned int n = 0; n < indices.size(); n++ ){
		const card32 pIndex = indices[n];
		const Vector3f& p = ps->get3f( pIndex, POSITION );
		pca.addPoint( p );
	}
	Matrix3f eigenVectors;
	Vector3f eigenValues, centroid;
	pca.analyze( eigenValues, eigenVectors, centroid );
	tNode->setCenter( centroid );
	tNode->setT( eigenVectors );
	Matrix3f invT = tNode->getT().transpose();
	BoundingBox2f bbox;
	for( unsigned int n = 0; n < indices.size(); n++ ){
		const card32 pIndex = indices[n];
		Vector3f p = ps->get3f( pIndex, POSITION );
		p = invT * ( p - tNode->getCenter() );
		const Vector2f& p2D = makeVector2f( p[0], p[1] );
		if ( n == 0 ) bbox = BoundingBox2f( p2D );
		else bbox.addPoint( p2D );
	}
	bbox.enlarge( 1.05f );
	Vector2f bboxCenter = bbox.getCenter();
	float32 bboxExtend = bbox.getMaxSideLength() / 2.0f;
	bbox.lowerCorner = bboxCenter - makeVector2f( bboxExtend, bboxExtend );
	bbox.upperCorner = bboxCenter + makeVector2f( bboxExtend, bboxExtend );
	tNode->setChild( buildVSTreeSNode( depth+1, indices, ps, tNode, bbox ) );
	return tNode;
}

VSTreeSNode* VSTreeOperations::buildVSTreeSNode( 
	const card32& depth, 
	std::vector<card32> indices, 
	PointSet* ps, 
	VSTreeTNode* parentTNode, 
	const BoundingBox2f bbox )
{
	// Leaf reached
	if ( ( depth >= maxDepth ) || ( indices.size() < splitSize ) ) {
		VSTreeSNode* sNode = new VSTreeSNode();
		sNode->setID( currentID ); currentID++;
		sNode->setBBox( bbox );
		sNode->setIndices( indices );
		sNode->setParentTNode( parentTNode );
		return sNode;
	}

	// Subdivide
	AAT POSITION = ps->getAAT( "position" );
	VSTreeSNode* sNode = new VSTreeSNode();
	sNode->setID( currentID ); currentID++;
	sNode->setParentTNode( parentTNode );
	sNode->setBBox( bbox );
	Matrix3f invT = parentTNode->getT().transpose();
	std::vector< std::vector<card32> > subIndices; subIndices.resize( 4 ); 
	for ( unsigned int n = 0; n < 4; n++ ) 
		subIndices[n].clear();
	for ( unsigned int n = 0; n < indices.size(); n++ ){
		const card32 pIndex = indices[n];
		Vector3f p = ps->get3f( pIndex, POSITION );
		p = invT * ( p - parentTNode->getCenter() );
		const Vector2f& p2D = makeVector2f( p[0], p[1] );
		if ( !bbox.inBoundingBox( p2D ) ){
			debugOutput << "Should not happen!\n"; debugOutput.flush();
			return NULL;
		}
		card32 quadrant = 0;
		for ( unsigned int i = 0; i < 2; i++ ){
			if ( p2D[i] >= ( bbox.lowerCorner[i] + bbox.upperCorner[i] ) / 2.0f )
				quadrant += power( 2,i );
		}
		subIndices[quadrant].push_back( pIndex );
	}
	for ( unsigned int n = 0; n < 4; n++ ){
		BoundingBox2f subBBox;
		float32 s = bbox.getMaxSideLength() / 2.0f;
		Vector2i I;
		I[1] = n / 2;
		I[0] = n - I[1]*2;
		for ( unsigned int i = 0; i < 2; i++ ){
			float32 e = I[i] * s;
			subBBox.lowerCorner[i] = bbox.lowerCorner[i] + e;
			subBBox.upperCorner[i] = subBBox.lowerCorner[i] + s;
		}
		sNode->setChild( n, buildVSTreeSNode( depth+1, subIndices[n], ps, parentTNode, subBBox ) );
	}
	return sNode;
}

// Count number of nodes
void VSTreeOperations::getNumNodes( VSTree* tree, card32& numNodes, card32& numVNodes, card32& numTNodes, card32& numSNodes )
{
	VSTreeIterator it;
	it.setup( tree );
	numNodes = numVNodes = numTNodes = numSNodes = 0;
	while( !it.atEnd() ){
		VSTreeNode* node = it.next();
		VSTreeVNode* vNode = dynamic_cast<VSTreeVNode*>( node );
		VSTreeTNode* tNode = dynamic_cast<VSTreeTNode*>( node );
		VSTreeSNode* sNode = dynamic_cast<VSTreeSNode*>( node );
		if ( vNode != NULL ) numVNodes++;
		else if ( tNode != NULL ) numTNodes++;
		else if ( sNode != NULL ) numSNodes++;
		numNodes++;
	}
}