//---------------------------------------------------------------------------
#include "StdAfx.h"
#include <windows.h>
#include <mmsystem.h>
//---------------------------------------------------------------------------
#include "VSTreeRenderer.h"
#include "VSTreeSNode.h"
#include "VSTreeTNode.h"
#include "VSTreeVNode.h"
#include "VSTreeIterators.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

static std::string getNodeColorizationDescr( const card32 index )
{
	if ( index == COLORIZE_CONSTANT ) return "Constant";
	else if ( index == COLORIZE_NODE_COLOR ) return "Node-Color";
	else return "Undefined";
}

IMPLEMENT_X4_CLASS( VSTreeRenderer, 0 )
{
	BEGIN_CLASS_INIT( VSTreeRenderer )
	ADD_BOOLEAN_PROP( showVNodes, 0 )
	ADD_BOOLEAN_PROP( showTNodes, 0 )
	ADD_BOOLEAN_PROP( showSNodes, 0 )
	ADD_BOOLEAN_PROP( showNodeDataPoints, 0 )
	ADD_CARD32_PROP_UPDATE_DESCR( nodeColorization, 0, NULL, getNodeColorizationDescr, 2, NULL )
	ADD_VECTOR4F_PROP( vNodeColor, 0 )
	ADD_VECTOR4F_PROP( tNodeColor, 0 )
	ADD_VECTOR4F_PROP( sNodeColor, 0 )
}

VSTreeRenderer::VSTreeRenderer()
{
	nodeColorization = COLORIZE_CONSTANT;
	showVNodes = false;
	showTNodes = true;
	showSNodes = true;
	showNodeDataPoints = false;
	vNodeColor = makeVector4f( 0.5f, 0.5f, 0.5f, 1.0f );
	tNodeColor = makeVector4f( 0.0f, 0.0f, 0.5f, 1.0f );
	sNodeColor = makeVector4f( 0.5f, 0.0f, 0.0f, 1.0f );
}

VSTreeRenderer::~VSTreeRenderer()
{
}

void VSTreeRenderer::renderTree( VSTree* tree, PointSet* ps )
{
	if ( tree == NULL ) return;

	VSTreeIterator it;
	it.setup( tree );
	while( !it.atEnd() ){
		renderNode( it.next(), ps );
	}
}

void VSTreeRenderer::renderNode( VSTreeNode* node, PointSet* ps )
{
	if ( node == NULL ) return;
	VSTreeVNode* vNode = dynamic_cast<VSTreeVNode*>( node );
	VSTreeTNode* tNode = dynamic_cast<VSTreeTNode*>( node );
	VSTreeSNode* sNode = dynamic_cast<VSTreeSNode*>( node );
	if ( vNode != NULL ) renderVNode( vNode, ps );		
	else if ( tNode != NULL ) renderTNode( tNode, ps );
	else if ( sNode != NULL ) renderSNode( sNode, ps );
}

void VSTreeRenderer::renderVNode( VSTreeVNode* vNode, PointSet* ps )
{
	if ( vNode == NULL ) return;
	if ( !showVNodes ) return;
	BoundingBox3f bbox = vNode->getBBox();
	rt.disableLighting();
	rt.setRenderColor( getColor( vNode ) );
	//bbox.enlarge( 0.95f );
	rt.renderBox( bbox );
	if ( showNodeDataPoints ){
		AAT POSITION = ps->getAAT( "position" );
		for ( unsigned int n = 0; n < vNode->getNumIndices(); n++ ){
			const card32 pIndex = vNode->getIndex( n );
			const Vector3f p = ps->get3f( pIndex, POSITION );
			rt.renderPoint( p );
		}
	}
}

void VSTreeRenderer::renderTNode( VSTreeTNode* tNode, PointSet* ps )
{
	if ( tNode == NULL ) return;
	if ( !showTNodes ) return;
	VSTreeSNode* sNode = dynamic_cast<VSTreeSNode*>( tNode->getChild() );
	if ( sNode == NULL ) return;
	Vector2f p00 = sNode->getBBox().lowerCorner;
	Vector2f p11 = sNode->getBBox().upperCorner;
	Vector2f p01 = makeVector2f( p00[0], p11[1] );
	Vector2f p10 = makeVector2f( p11[0], p00[1] );
	Vector3f p00_3D = tNode->getT() * makeVector3f( p00[0], p00[1], 0 ) + tNode->getCenter();
	Vector3f p01_3D = tNode->getT() * makeVector3f( p01[0], p01[1], 0 ) + tNode->getCenter();
	Vector3f p10_3D = tNode->getT() * makeVector3f( p10[0], p10[1], 0 ) + tNode->getCenter();
	Vector3f p11_3D = tNode->getT() * makeVector3f( p11[0], p11[1], 0 ) + tNode->getCenter();
	rt.disableLighting();
	rt.setRenderColor( getColor( tNode ) );
	rt.renderLine( p00_3D, p01_3D );
	rt.renderLine( p01_3D, p11_3D );
	rt.renderLine( p11_3D, p10_3D );
	rt.renderLine( p10_3D, p00_3D );
	
}

void VSTreeRenderer::renderSNode( 
	VSTreeSNode* sNode, 
	PointSet* ps )
{
	if ( sNode == NULL ) return;
	if ( !showSNodes ) return;
	BoundingBox2f bbox = sNode->getBBox();
	//bbox.enlarge( 0.95f );
	Vector2f p00 = bbox.lowerCorner;
	Vector2f p11 = bbox.upperCorner;
	Vector2f p01 = makeVector2f( p00[0], p11[1] );
	Vector2f p10 = makeVector2f( p11[0], p00[1] );
	Vector3f p00_3D = sNode->getT() * makeVector3f( p00[0], p00[1], 0 ) + sNode->getCenter();
	Vector3f p01_3D = sNode->getT() * makeVector3f( p01[0], p01[1], 0 ) + sNode->getCenter();
	Vector3f p10_3D = sNode->getT() * makeVector3f( p10[0], p10[1], 0 ) + sNode->getCenter();
	Vector3f p11_3D = sNode->getT() * makeVector3f( p11[0], p11[1], 0 ) + sNode->getCenter();
	rt.disableLighting();
	rt.setRenderColor( getColor( sNode ) );
	rt.renderLine( p00_3D, p01_3D );
	rt.renderLine( p01_3D, p11_3D );
	rt.renderLine( p11_3D, p10_3D );
	rt.renderLine( p10_3D, p00_3D );
	if ( showNodeDataPoints ){
		AAT POSITION = ps->getAAT( "position" );
		for ( unsigned int n = 0; n < sNode->getNumIndices(); n++ ){
			const card32 pIndex = sNode->getIndex( n );
			const Vector3f p = ps->get3f( pIndex, POSITION );
			rt.renderPoint( p );
		}
	}
}