//---------------------------------------------------------------------------
#ifndef AttachedIndexedOctreeH
#define AttachedIndexedOctreeH
//---------------------------------------------------------------------------
#include "x3geometry.h"
//---------------------------------------------------------------------------
#include "AttachedData.h"
#include "SimpleOctree.h"
#include "BoundingBox.h"
#include "PointSet.h"
#include "PointCloud.h"
#include "PointCloudIterators.h"
//---------------------------------------------------------------------------
#include <queue>
#include <vector>
//---------------------------------------------------------------------------

namespace X4 {

/// helper structure, used in AttachedIndexedOctree
struct IndexedNode {
   std::vector<card32> m_index;
   BoundingBox3f m_bb;
	Sphere3f m_bsphere; 
};

typedef FastSimpleOctree<IndexedNode> AIOctreeNode;

/// helper structure, used in KNNQuery
struct PointEntry {
   card32 index;
   float32 quDist;
   PointEntry(const card32& i, const float32& q) {quDist = q; index = i;}
   inline bool operator<(const PointEntry &other) const {
      return quDist > other.quDist;
   }
};

/// helper structure, used in KNNQuery
struct BoxEntry {
   AIOctreeNode *node;
   float32 quDist;
   BoxEntry(AIOctreeNode *n, const float32& q) {quDist = q; node = n;}
   inline bool operator<(const BoxEntry &other) const {
      return quDist > other.quDist;
   }
};

template <class Type>
class  PriorityQueue_CachedArray{
public:
	std::vector<Type> data;
	unsigned mySize;

	PriorityQueue_CachedArray() {mySize=0;}
	inline unsigned size()const{return mySize;};
	inline bool empty()const{return mySize==0;}
	inline Type& top(){
		x4Assert(!empty());
		return data[0];}
	inline void push(const Type& d){
		if( mySize>=data.size() )
			data.push_back(d);
		data[mySize] = d;
		mySize++;
		push_heap(data.begin(), data.begin() + mySize);}
	inline void pop(){
		pop_heap(data.begin(), data.begin() + mySize);
		mySize--;
	}
	inline void clear(){
		mySize = 0;
	}

};

class AttachedIndexedOctree;

/// KNNQuery allows fast k-nearest-Neighbor Queries on non hierarchical PCs and
/// hierarchy-nodes
/// It's used by HierarchicalKNNIterator and ModifyableKNNIterator
/// direct use is nonrecommended
class X3GEOMETRY_API KNNQuery
{
 private:
   AttachedIndexedOctree* m_octree;
	SecondaryDataStructureIterator* m_sds;
   PriorityQueue_CachedArray<PointEntry> pqPoints;
   PriorityQueue_CachedArray<BoxEntry> pqBoxes;
   Vector3f seekPoint;
   AIOctreeNode* rootNode;
   const PointSet* pntSet;
   BasicPointCloudIterator* m_basic;
   AAT POSITION;
   card32 atPoint;
   bool nextPointReady;
   PointSet* pntSet_rw;
   float32 maxDistanceToSeekPoint_2;	// squared maximum distance to seek point (reports true in atEnd() 
										// when no further points within this distance are available)
   float32 m_SquaredDistanceToCurrentPoint;	// 


   virtual void setup(SceneObjectIterator* , const card32&, MemoryManager *, const bool &modifyable);

   void putBoxOnQueue(AIOctreeNode* box);
   void putPointOnQueue(const card32 &indexOfPoint);

 public:
   /// Use setSeekPoint after construction
   KNNQuery(SceneObjectIterator* , const card32&, MemoryManager *, const bool &modifyable);
   ~KNNQuery();

   /// restrict region of interest to a sphere with radius maxDistanceToSeekPoint
   /// default is inf
   void setMaxDistanceToSeekPoint(float32 d){maxDistanceToSeekPoint_2 = d*d;}

   void   setSeekPoint(const Vector3f&);
   /// returns PointSet Index of next nearest Neighbor
   card32 getNextPoint();
   /// returns true, if there is one more point in Pointset, access point by getVertex or getDataPointer
   bool  seekNextPoint();
	/// get nearest Neighbor and Distance to given point
   card32 getNearestPointIndexAndSqrDistance(const Vector3f &targetPosition, float32 &returnedDistance);
   /// get squared distance of current point to seek point
   float getCurrentSquaredDistance() const {return m_SquaredDistanceToCurrentPoint;}

   inline Vector3f getNearestPointPosition(){
      x4Assert(nextPointReady);
      return pntSet->get<float32,3>(atPoint, POSITION);
   }

   ///get current Vertex
   inline void getVertex(char *buffer) const {
      x4Assert(nextPointReady);
      pntSet->getVertex(atPoint, buffer);
   }

   /// get the datapointer of current vertex
   inline const char *getDataPointer() {
      x4Assert(nextPointReady);
      return pntSet->getDataPointer(atPoint);
   }
   /// get the datapointer of current vertex, its allowed to change everything but position
   /// change of position results in errors
   inline char *getModifyableDataPointer() {
      x4Assert(nextPointReady);
      x4Assert(pntSet_rw); // assert kNN is modifyable
      return pntSet_rw->getDataPointer(atPoint);
   }

   inline card32 getCurrentPointNumber(){return atPoint;}

   /// For use in HierarchicalKNNIterator only!!
   KNNQuery(SceneObjectIterator* , const Vector3f& , const card32&, MemoryManager *attachedOctreeMemManager, const bool &modifyable);

} ;


/// AttachedIndexedOctree is an Attachment - Octree of Indexes of points for large nodes
/// It is built dynamically for e.g. fast kNN Queries
class X3GEOMETRY_API AttachedIndexedOctree : public AttachedData {
	X4_CLASS( AttachedIndexedOctree )
private:
   static card64 m_globalAge;

   card64 m_age;
   AIOctreeNode m_octree;
   const PointSet* m_ps;

 public:

   AttachedIndexedOctree();

   /// copies indices only with the old point set reference
   /// call setPointSet afterwards if the point set has changed (e.g. has been copied as well)
   virtual void assign(const Object* obj, X4CopyContext *context = NULL);

   /// Ask for delete, do not delete if position of points didn't change
   virtual bool queryDelete(ChangeType whatChanged);


   /// build octree
   void octreesetup(
	   const PointSet* ps,		// Point set
	   card32 max_points		// Max points per leaf 
	);

   /// Build octree, this method is required if the initial point set does not contain all points
   void octreesetupWithBBox(
	   const PointSet* ps,		// Point set
	   card32 max_points,		// Max points per leaf 
	   BoundingBox3f bbox
	   );
   // Insert an additional index to the structure; required if underlying point set changes
   void insertPointIndex( 
	   const card32 index,
	   card32 max_points		// Max points per leaf 
	); 
   inline AIOctreeNode* getOctree(void){ return(&m_octree); };
   inline AIOctreeNode* findNode(const Vector3f& pos){
      if(!m_octree.getContent().m_bb.inBoundingBox(pos)) return(NULL);

      AIOctreeNode* node=&m_octree;
      while(!node->isLeaf()){
         for(card32 c=0; c<8; c++){
            if(node->getChild(c)->getContent().m_bb.inBoundingBox(pos)){
               node=node->getChild(c);
               break;
            }
         }
      }
      return(node);
   }
   inline const PointSet* getPointSet(void){ return(m_ps); };
   static void goOlder(void){ m_globalAge++; };
   inline bool isOld(void){ return(m_globalAge>m_age); };
   inline void setPointSet( PointSet* ps ){ this->m_ps = ps; }
   inline void setAge( card64 age ){ this->m_age = age; }
   inline card64 getGlobalAge(){ return m_globalAge; }
	void computeBoundingSpheres();

};

/*

/// deprecated
class X3GEOMETRY_API AIHierarchicalOctreeIterator :
                                 public PointCloudIterator,
                                 public BasicPointCloudIterator,
                                 public HierarchicalIterator,
                                 public SecondaryDataStructureIterator {

 private:
   static MetaClass *AIHierarchicalOctreeIteratorMetaClass;
   static Object* newInstance();

   SceneObjectIterator *m_pi;
   HierarchicalIterator* m_hpi;
   SecondaryDataStructureIterator* m_sds;
   BasicPointCloudIterator* m_bpci;

   bool m_inOctree;
   AIOctreeNode* m_node;
   card32 m_currentIndex;
   AttachedIndexedOctree* m_octree;
   PointSet* m_lockedPS;
   bool m_locked;

   MemoryManager *m_aiMemManage;

   void checkForOctree(void);

 public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass, card32 uniqueIdentifier);
   static void shutdown();

   virtual void assign(const Object* obj, X4CopyContext *context = NULL);

   virtual void clearAndSetup(PointCloud* pc, SceneObjectIterator* mpi, MemoryManager *aiMemManage = NULL);
   virtual void invalidateOctrees(void);

   AIHierarchicalOctreeIterator();
   virtual ~AIHierarchicalOctreeIterator();

   virtual void reset();
   virtual void next();
   virtual bool atEnd();

   /// get attributes in different formats
   virtual card8     get1ub(const AAT &token) const;
   virtual Vector2ub get2ub(const AAT &token) const;
   virtual Vector3ub get3ub(const AAT &token) const;
   virtual Vector4ub get4ub(const AAT &token) const;

   virtual int32    get1i(const AAT &token) const;
   virtual Vector2i get2i(const AAT &token) const;
   virtual Vector3i get3i(const AAT &token) const;
   virtual Vector4i get4i(const AAT &token) const;

   virtual float32  get1f(const AAT &token) const;
   virtual Vector2f get2f(const AAT &token) const;
   virtual Vector3f get3f(const AAT &token) const;
   virtual Vector4f get4f(const AAT &token) const;

   virtual void     getVertex(char *buffer) const;
   virtual bool nextBlock();
   virtual const PointSet * lockBlockRead();
   virtual void unlock();

   virtual card32 getNumChildren();
   virtual bool goToChild(card32 index);
   virtual bool goToParent();
   virtual BoundingBox3f getBoundingBox();
   virtual bool isLeaf() const;
   virtual bool getNodeInfo(NodeInfoType whichInfo, const void *parameter, void *data);

	virtual AttachedDataContainer *lockAttachments();
	virtual void unlockAttachments();

};

*/




}

#endif
