//---------------------------------------------------------------------------
#include "StdAfx.h"
#include "DiscreteManifold.h"
#include "CopyObjectProperties.h"
#include <omp.h>

//---------------------------------------------------------------------------
using namespace X4;

IMPLEMENT_X4_CLASS(DiscreteManifoldPoint, 0)
{
	BEGIN_CLASS_INIT(DiscreteManifoldPoint);

	ADD_CARD64_PROP(m_Index, 0);
}

//---------------------------------------------------------------------------
DiscreteManifoldPoint::DiscreteManifoldPoint() : ManifoldPoint()
{
}

//---------------------------------------------------------------------------
DiscreteManifoldPoint::~DiscreteManifoldPoint()
{
}

//---------------------------------------------------------------------------
DiscreteManifoldPoint::DiscreteManifoldPoint(const DiscreteManifoldPoint& point) : ManifoldPoint(point)
{
	m_Index = point.m_Index;
}

//---------------------------------------------------------------------------
DiscreteManifoldPoint::DiscreteManifoldPoint(const DiscreteManifoldPoint* point) : ManifoldPoint(point)
{
	m_Index = point->m_Index;
}


//---------------------------------------------------------------------------
IMPLEMENT_X4_ABSTRACT_CLASS(DiscreteManifold, 1)
{
	BEGIN_CLASS_INIT(DiscreteManifold);

	ADD_BOOLEAN_PROP(m_bUseSparseHashMap, 0);
	ADD_CARD32_PROP(m_SubdivisionLevel, 0);
	ADD_BOOLEAN_PROP(m_useHash, 1);
	ADD_BOOLEAN_PROP(m_hashPathsToo,1);
}

//---------------------------------------------------------------------------
DiscreteManifold::HashData DiscreteManifold::HASH_DATA_EMPTY = DiscreteManifold::HashData(-1);
DiscreteManifold::HashData DiscreteManifold::HASH_DATA_NON_VALID = DiscreteManifold::HashData();

//---------------------------------------------------------------------------
DiscreteManifold::DiscreteManifold()
{
	m_hashPathsToo = true;
	m_useHash = true;
	m_SubdivisionLevel = 0;
	m_DenseDistanceHash = NULL;
	m_SparseDistanceHash = NULL;
	useSparseHash(false);
}

//---------------------------------------------------------------------------
DiscreteManifold::~DiscreteManifold()
{
	clearGeodesicsMemory();
}

//---------------------------------------------------------------------------
void  DiscreteManifold::useSparseHash(bool use)
{
	if ((use && m_SparseDistanceHash) || (!use && m_DenseDistanceHash)) return;
	clearGeodesicsMemory();
	m_bUseSparseHashMap = use;

	if (use)
	{
		m_SparseDistanceHash = new SparseHashTable;
	}else
	{
		m_DenseDistanceHash = new DenseHashTable;
		m_DenseDistanceHash->set_empty_key(HashKey());
	}
}

//---------------------------------------------------------------------------
void DiscreteManifold::setHashData(mpcard i, mpcard j, const HashData& _data) const
{
	if (i == NON_VALID_POINT || j == NON_VALID_POINT || m_useHash == false) return;

	HashData data = _data;
	data.distance = _data.distance;

	if (m_hashPathsToo == false) data.path.clear();
	
	#pragma omp critical
	{
		HashKey key(i,j);
		if (m_SparseDistanceHash)
		{
			SparseHashTable::iterator it = m_SparseDistanceHash->find(key);
			if (it == m_SparseDistanceHash->end()) 
				m_SparseDistanceHash->insert(std::pair<HashKey, HashData>(key,data));
			else
				it->second = data;
		}else if (m_DenseDistanceHash)
		{
			DenseHashTable::iterator it = m_DenseDistanceHash->find(key);
			if (it == m_DenseDistanceHash->end()) 
				m_DenseDistanceHash->insert(std::pair<HashKey, HashData>(key,data));
			else
				it->second = data;
		}
	}
}

//---------------------------------------------------------------------------
bool DiscreteManifold::isHashDataValid(mpcard i, mpcard j) const
{
	if (i == NON_VALID_POINT || j == NON_VALID_POINT || m_useHash == false) return false;

	HashKey key(i,j);

	if (m_SparseDistanceHash)
	{
		SparseHashTable::const_iterator it = m_SparseDistanceHash->find(key);
		if (it == m_SparseDistanceHash->end()) return false;
	}else if (m_DenseDistanceHash)
	{
		DenseHashTable::const_iterator it = m_DenseDistanceHash->find(key);
		if (it == m_DenseDistanceHash->end()) return false;
	}
	return true;
}

//---------------------------------------------------------------------------
const DiscreteManifold::HashData* DiscreteManifold::getHashData(mpcard i, mpcard j) const
{
	if (i == NON_VALID_POINT || j == NON_VALID_POINT) return &DiscreteManifold::HASH_DATA_NON_VALID;
	if (m_useHash == false) return  &DiscreteManifold::HASH_DATA_EMPTY;

	HashKey key(i,j);

	if (m_SparseDistanceHash)
	{
		SparseHashTable::const_iterator it = m_SparseDistanceHash->find(key);
		if (it == m_SparseDistanceHash->end()) return &DiscreteManifold::HASH_DATA_EMPTY;
		return &(it->second);
	}else if (m_DenseDistanceHash)
	{
		DenseHashTable::const_iterator it = m_DenseDistanceHash->find(key);
		if (it == m_DenseDistanceHash->end()) return &DiscreteManifold::HASH_DATA_EMPTY;
		return &(it->second);
	}
	return &DiscreteManifold::HASH_DATA_EMPTY;
}


//---------------------------------------------------------------------------
const DiscreteManifold::HashData* DiscreteManifold::getHashData(const ManifoldPoint* pointA, const ManifoldPoint* pointB, mpcard* returnedIndexA, mpcard* returnedIndexB)
{
	x4Assert(pointA->isDiscretePoint() == false);
	x4Assert(pointB->isDiscretePoint() == false);

	mpcard i = findNearestPoint(pointA->getCoordinate());
	mpcard j = findNearestPoint(pointB->getCoordinate());

	if (returnedIndexA) *returnedIndexA = i;
	if (returnedIndexB) *returnedIndexB = j;

	if (m_useHash == false) return  &DiscreteManifold::HASH_DATA_EMPTY;

	if (i == NON_VALID_POINT || j == NON_VALID_POINT) return &DiscreteManifold::HASH_DATA_NON_VALID;

	return getHashData(i,j);
}

//---------------------------------------------------------------------------
const DiscreteManifold::HashData* DiscreteManifold::getHashData(const DiscreteManifoldPoint* pointA, const DiscreteManifoldPoint* pointB, mpcard* returnedIndexA, mpcard* returnedIndexB)
{
	x4Assert(pointA->isDiscretePoint() == true);
	x4Assert(pointB->isDiscretePoint() == true);

	mpcard i = pointA->getIndex();
	mpcard j = pointB->getIndex();

	if (returnedIndexA) *returnedIndexA = i;
	if (returnedIndexB) *returnedIndexB = j;

	if (m_useHash == false) return  &DiscreteManifold::HASH_DATA_EMPTY;

	if (i == NON_VALID_POINT || j == NON_VALID_POINT) return &DiscreteManifold::HASH_DATA_NON_VALID;

	return getHashData(i,j);
}


//---------------------------------------------------------------------------

std::vector<const DiscreteManifold::HashData*> DiscreteManifold::getHashDatas(const ManifoldPoint* pointA, const std::vector<const ManifoldPoint*>& toPoints, mpcard* returnedIndexA, std::vector<mpcard>* returnedIndicesB)
{
	std::vector<const HashData*> datas;

	mpcard i = NON_VALID_POINT;
	
	// in order to get distance we need to find out the discrete coordinates of the given points
	if (!pointA->isDiscretePoint()) i = findNearestPoint(pointA->getCoordinate());
	else
	{
		DiscreteManifoldPoint* dA = (DiscreteManifoldPoint*)(pointA);
		i = dA->getIndex();
	}

	if (returnedIndexA) *returnedIndexA = i;
	if (returnedIndicesB) returnedIndicesB->resize(toPoints.size(), NON_VALID_POINT);

	// setup a list of points which are not hashed
	for (size_t j=0; j < toPoints.size(); j++)
	{
		mpcard pj = NON_VALID_POINT;

		if (!toPoints[j]->isDiscretePoint()) pj = findNearestPoint(toPoints[j]->getCoordinate());
		else
		{
			DiscreteManifoldPoint* dB = (DiscreteManifoldPoint*)(toPoints[j]);
			pj = dB->getIndex();
		}

		if (returnedIndicesB) returnedIndicesB->at(j) = pj;
		datas.push_back(getHashData(i, pj));
	}

	return datas;
}

//---------------------------------------------------------------------------
float32 DiscreteManifold::getGeodeticDistance(const ManifoldPoint* pointA, const ManifoldPoint* pointB)
{
	if (m_bDirtyGeodesics)
	{
		if (m_bUseSparseHashMap) m_SparseDistanceHash->clear();
		else m_DenseDistanceHash->clear();

		m_bDirtyGeodesics = false;
	}

	mpcard i = NON_VALID_POINT;
	mpcard j = NON_VALID_POINT;

	float distance = getHashData(pointA, pointB, &i, &j)->distance;

	if (distance < 0) 
		distance = computeDistanceAndFillHash(i,j, NULL);

	return distance;
}

//---------------------------------------------------------------------------
float32 DiscreteManifold::getGeodeticDistance(const DiscreteManifoldPoint* pointA, const DiscreteManifoldPoint* pointB)
{
	if (pointA->getIndex() == pointB->getIndex()) return 0;

	if (m_bDirtyGeodesics)
	{
		if (m_bUseSparseHashMap) m_SparseDistanceHash->clear();
		else m_DenseDistanceHash->clear();

		m_bDirtyGeodesics = false;
	}

	mpcard i = NON_VALID_POINT;
	mpcard j = NON_VALID_POINT;

	float distance = getHashData(pointA, pointB, &i, &j)->distance;

	if (distance < 0) 
		distance = computeDistanceAndFillHash(i,j, NULL);

	return distance;

}

//---------------------------------------------------------------------------
std::vector<float32> DiscreteManifold::getGeodeticDistances(const ManifoldPoint* pointA, const std::vector<const ManifoldPoint*>& toPoints)
{
	if (m_bDirtyGeodesics)
	{
		if (m_bUseSparseHashMap) m_SparseDistanceHash->clear();
		else m_DenseDistanceHash->clear();

		m_bDirtyGeodesics = false;
	}

	mpcard i = NON_VALID_POINT;
	std::vector<mpcard> j;

	// get data from hash
	const std::vector<const HashData*>& datas = getHashDatas(pointA, toPoints, &i, &j);
	std::vector<float> dists(toPoints.size(), 1e30f);

	// mark those which still not computed
	std::vector<mpcard> targets;
	std::vector<mpcard> targetMap;
	for (size_t k=0; k < toPoints.size(); k++)
	{
		if (datas[k]->distance < 0)
		{
			targets.push_back(j[k]);
			targetMap.push_back(k);
		}else
		{
			dists[k] = datas[k]->distance;
		}
	}

	// compute distances for those which were uncomputed and copy it into the result field
	std::vector<float> computedDists = computeDistancesAndFillHash(i, targets, NULL);

	// copy distances accordingly
	for (size_t j=0; j < computedDists.size(); j++)
		dists[targetMap[j]] = computedDists[j];

	return dists;
}

//---------------------------------------------------------------------------

std::vector<DVectorF> DiscreteManifold::getGeodeticPath(const ManifoldPoint* pointA, const ManifoldPoint* pointB)
{
	if (m_bDirtyGeodesics)
	{
		if (m_bUseSparseHashMap) m_SparseDistanceHash->clear();
		else m_DenseDistanceHash->clear();

		m_bDirtyGeodesics = false;
	}

	mpcard i = NON_VALID_POINT;
	mpcard j = NON_VALID_POINT;

	const HashData* data = getHashData(pointA, pointB, &i, &j);

	std::vector<DVectorF> path = data->path;

	if (path.size() == 0) 
		computeDistanceAndFillHash(i,j, &path);

	return path;
}

//---------------------------------------------------------------------------

std::vector<DVectorF> DiscreteManifold::getGeodeticPath(const DiscreteManifoldPoint* pointA, const DiscreteManifoldPoint* pointB)
{
	if (m_bDirtyGeodesics)
	{
		if (m_bUseSparseHashMap) m_SparseDistanceHash->clear();
		else m_DenseDistanceHash->clear();

		m_bDirtyGeodesics = false;
	}

	mpcard i = NON_VALID_POINT;
	mpcard j = NON_VALID_POINT;

	const HashData* data = getHashData(pointA, pointB, &i, &j);

	std::vector<DVectorF> path = data->path;

	if (path.size() == 0) 
		computeDistanceAndFillHash(i,j, &path);

	return path;
}

//---------------------------------------------------------------------------

std::vector<const std::vector<DVectorF>*> DiscreteManifold::getGeodeticPaths(const ManifoldPoint* pointA, const std::vector<const ManifoldPoint*>& toPoints)
{
	if (m_bDirtyGeodesics)
	{
		if (m_bUseSparseHashMap) m_SparseDistanceHash->clear();
		else m_DenseDistanceHash->clear();

		m_bDirtyGeodesics = false;
	}

	mpcard i = NON_VALID_POINT;
	std::vector<mpcard> j;

	// get data from hash
	const std::vector<const HashData*>& datas = getHashDatas(pointA, toPoints, &i, &j);
	std::vector<const std::vector<DVectorF>*> paths(toPoints.size());

	// mark those which still not computed
	std::vector<mpcard> targets;
	std::vector<mpcard> targetMap;
	for (size_t k=0; k < toPoints.size(); k++)
	{
		if (datas[k]->path.size() == 0)
		{
			targets.push_back(j[k]);
			targetMap.push_back(k);
		}else
		{
			paths[k] = &(datas[k]->path);
		}
	}

	// compute distances for those which were uncomputed and copy it into the result field
	computeDistancesAndFillHash(i, targets, NULL);

	// get paths
	for (size_t j=0; j < targets.size(); j++)
		paths[targetMap[j]] = &(getHashData(pointA, toPoints[targetMap[j]])->path);

	return paths;
}

//---------------------------------------------------------------------------
void  DiscreteManifold::setCachePaths(bool use)
{
	if (m_hashPathsToo == use) return;
	if (use == false)
	{
		// clear all paths from the hash
		/*size_t size = getNumHashEntries();
		for (size_t i=0; i < size; i++)
		{
			getHashEntry(i)->second.path.clear();
		}*/
		if (m_DenseDistanceHash)
		{
			for (DenseHashTable::iterator it = m_DenseDistanceHash->begin(); it != m_DenseDistanceHash->end(); it++)
				it->second.path.clear();
		}else
		{
			for (SparseHashTable::iterator it = m_SparseDistanceHash->begin(); it != m_SparseDistanceHash->end(); it++)
				it->second.path.clear();
		}
	}

	m_hashPathsToo = use;
}

//---------------------------------------------------------------------------
void DiscreteManifold::setUseCaching(bool set)
{
	if (set == false)
	{
		if (m_bUseSparseHashMap) m_SparseDistanceHash->clear();
		else m_DenseDistanceHash->clear();
	}
	m_useHash = set;
}

//---------------------------------------------------------------------------
void DiscreteManifold::clearGeodesicsMemory()
{
	if (m_SparseDistanceHash)
	{
		delete m_SparseDistanceHash;
		m_SparseDistanceHash = NULL;
	}
	if (m_DenseDistanceHash)
	{
		delete m_DenseDistanceHash;
		m_DenseDistanceHash = NULL;
	}
	m_bDirtyGeodesics = true;
}

//---------------------------------------------------------------------------
void DiscreteManifold::assign(const Object* obj, X4CopyContext* context)
{
	const DiscreteManifold * o = dynamic_cast<const DiscreteManifold*>(obj);
	x4Assert( o != NULL );

	copyObjectProperties(obj,this);

	// copy hash
	useSparseHash(m_bUseSparseHashMap);
	if (m_bUseSparseHashMap) *m_SparseDistanceHash = *o->m_SparseDistanceHash;
	else *m_DenseDistanceHash = *o->m_DenseDistanceHash;

	/*card64 size = o->getNumHashEntries();
	for (card64 i=0; i < size; i++)
	{
		insertHashEntry(o->getHashEntry(i));
	}*/
}

//---------------------------------------------------------------------------
void DiscreteManifold::write(OutputObjectStream *s) const
{
	SceneObject::write(s);

	card64 size = getNumHashEntries();
	s->writeBuffer(&size, sizeof(card64));
	//for (card64 i=0; i < size; i++)
	if (m_DenseDistanceHash)
	{
		for (DenseHashTable::const_iterator it = m_DenseDistanceHash->begin(); it != m_DenseDistanceHash->end(); it++)
		{
			s->writeBuffer(&(it->first), sizeof(HashKey));
			s->writeBuffer(&(it->second.distance), sizeof(float));

			card64 length = it->second.path.size();
			s->writeBuffer(&length, sizeof(card64));
			for (size_t j=0; j < length; j++)
				it->second.path[j].write(s);
		}
	}else
	{
		for (SparseHashTable::const_iterator it = m_SparseDistanceHash->begin(); it != m_SparseDistanceHash->end(); it++)
		{
			s->writeBuffer(&(it->first), sizeof(HashKey));
			s->writeBuffer(&(it->second.distance), sizeof(float));

			card64 length = it->second.path.size();
			s->writeBuffer(&length, sizeof(card64));
			for (size_t j=0; j < length; j++)
				it->second.path[j].write(s);
		}
	}
		
	//}
}

//---------------------------------------------------------------------------
void DiscreteManifold::read(InputObjectStream *s)
{
	// read properties
	SceneObject::read(s);

	// setup hash
	m_DenseDistanceHash = NULL;
	m_SparseDistanceHash = NULL;
	useSparseHash(m_bUseSparseHashMap);

	// read hash
	card64 size = 0;
	s->readBuffer(&size, sizeof(card64));
	if (m_bUseSparseHashMap) m_SparseDistanceHash->rehash(size);
	else m_DenseDistanceHash->rehash(size);
	for (card64 i=0; i < size; i++)
	{
		HashKey key;
		HashData data;
		s->readBuffer(&key, sizeof(HashKey));
		s->readBuffer(&data.distance, sizeof(float));

		card64 length = 0;
		s->readBuffer(&length, sizeof(card64));
		data.path.resize(length);
		for (size_t j=0; j < length; j++)
			data.path[j].read(s);

		insertHashEntry(std::pair<HashKey,HashData>(key,data));
	}
	m_bDirtyGeodesics = false;
}

//---------------------------------------------------------------------------
SceneObjectIterator *DiscreteManifold::createIterator(SceneObjectIteratorType type)
{
	if (!supportsIteratorType(type)) {
		return NULL;
	} else {
		return new SceneObjectIterator();
	}
}

//---------------------------------------------------------------------------
bool DiscreteManifold::supportsIteratorType(SceneObjectIteratorType cap)
{
	SOIT possibleCaps(  SOIT::CAP_MANIFOLD | 
						SOIT::CAP_DISCRETE_MANIFOLD) ;

	if( !possibleCaps.providesCapsFor(cap))
		return false;
	else
		return true;
}
