/***************************************************************************
*   Copyright (c) 2010   Art Tevs                                         *
*                                                                         *
*   This library is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU Lesser General Public License as        *
*   published by the Free Software Foundation; either version 3 of        *
*   the License, or (at your option) any later version.                   *
*                                                                         *
*   This library is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU Lesse General Public License for more details.                    *
*                                                                         *
*   The full license is in LICENSE file included with this distribution.  *
***************************************************************************/

//---------------------------------------------------------------------------
#ifndef DISCRETE_MANIFOLD__H
#define DISCRETE_MANIFOLD__H

//---------------------------------------------------------------------------
#include "Manifold.h"
#include "InCorePCTopologyGraph.h"

#include <google/sparse_hash_map> 
#include <google/dense_hash_map> 

//---------------------------------------------------------------------------
namespace X4
{

	/**
	* Manifold point living on discrete manifold. This points has also an unique index of the underlying discrete data set.
	**/
	class X3GEOMETRY_API DiscreteManifoldPoint : public ManifoldPoint
	{
		X4_CLASS(DiscreteManifoldPoint);
	protected:
		card64 m_Index;

	public:
		DiscreteManifoldPoint();
		DiscreteManifoldPoint(card64 index, const DVectorF& coord, const DVectorF& normal) : ManifoldPoint(coord, normal), m_Index(index){}
		DiscreteManifoldPoint(const DiscreteManifoldPoint& point);
		DiscreteManifoldPoint(const DiscreteManifoldPoint* point);
		virtual ~DiscreteManifoldPoint();

		//! Get index of the discrete manifold point
		card64 getIndex() const { return m_Index; }
		virtual bool isDiscretePoint() const { return true; }
	};

	/**
	* This class represents a manifold which can be used to represent different kind of geometry.
	* Manifolds can be used to be registered against each other or for other operations.
	* Manifolds can be continuous or discrete. All given default manifold operations has to be 
	* valid in discrete as also in continuous space.
	**/
	class X3GEOMETRY_API DiscreteManifold : public Manifold
	{
		X4_ABSTRACT_CLASS(DiscreteManifold)
	protected:
		// ----------------------- Hash of geodesics --------------------------------
		struct HashData
		{
			float distance;
			std::vector<DVectorF> path;
			HashData() : distance(1e30f) {}
			HashData(const float& d) : distance(d){}
			HashData(const std::vector<DVectorF>& p)
			{
				path = p;
				if (path.size() == 0) distance = 1e30f;
				if (path.size() == 1) distance = 0;
				if (path.size() > 1)
				{
					distance = 0;
					for (size_t i=1; i < path.size(); i++) distance += norm(path[i-1] - path[i]);
				}
			}
			HashData(const float& d, const std::vector<DVectorF>& p) : distance(d), path(p) {}
		};

		struct HashKey
		{
			card32 a;
			card32 b;
			HashKey(card32 _a, card32 _b) : a(_a), b(_b) {}
			HashKey()
			{
				a = 0xFFFFFFFF;
				b = 0xFFFFFFFF;
			}
			HashKey(const HashKey& k) : a(k.a), b(k.b) {}
			bool operator==(const HashKey& k)
			{
				return (a == k.a && b == k.b) || (a == k.b && b == k.a);
			}
		};

		struct HashEql
		{
			bool operator()(const HashKey& obj1, const HashKey& obj2) const
			{
				return (obj1.a == obj2.a && obj1.b == obj2.b) || (obj1.a == obj2.b && obj1.b == obj2.a);
			}
		};

		struct HashFunction
		{
			size_t operator()(const HashKey& obj1) const
			{
				register card32 a = obj1.a;
				register card32 b = obj1.b;
				/*if (a < b) { a = obj1.b; b = obj1.a; }
				return a * 32771 + 1031 * b;
				*/
				return a * b + (a ^ b);
			}
		};

		typedef google::sparse_hash_map<HashKey, HashData, HashFunction, HashEql> SparseHashTable;
		typedef google::dense_hash_map<HashKey, HashData, HashFunction, HashEql> DenseHashTable;

		SparseHashTable* m_SparseDistanceHash;
		DenseHashTable* m_DenseDistanceHash;

		bool m_bUseSparseHashMap;
		card32 m_SubdivisionLevel;

		inline size_t getNumHashEntries() const
		{
			if (m_DenseDistanceHash) return m_DenseDistanceHash->size();
			if (m_SparseDistanceHash) return m_SparseDistanceHash->size();
			return 0;
		}
		/*inline const std::pair<const HashKey,HashData>* getHashEntry(mpcard index) const
		{
			if (m_DenseDistanceHash)
			{
				DenseHashTable::const_iterator it = m_DenseDistanceHash->begin();
				for (mpcard i=0; i < index; i++){it++;}
				return it.operator->();
			}else
			{
				SparseHashTable::const_iterator it = m_SparseDistanceHash->begin();
				for (mpcard i=0; i < index; i++){it++;}
				return it.operator->();
			}
		}
		inline std::pair<const HashKey,HashData>* getHashEntry(mpcard index) 
		{
			if (m_DenseDistanceHash)
			{
				DenseHashTable::iterator it = m_DenseDistanceHash->begin();
				for (mpcard i=0; i < index; i++){it++;}
				return it.operator->();
			}else
			{
				SparseHashTable::iterator it = m_SparseDistanceHash->begin();
				for (mpcard i=0; i < index; i++){it++;}
				return it.operator->();
			}
		}*/
		inline void insertHashEntry(const std::pair<const HashKey,HashData>* data)
		{
			insertHashEntry(*data);
		}
		inline void insertHashEntry(const std::pair<const HashKey,HashData>& data)
		{
			if (m_DenseDistanceHash)
				m_DenseDistanceHash->insert(data);
			else
				m_SparseDistanceHash->insert(data);
		}

		static HashData HASH_DATA_EMPTY;
		static HashData HASH_DATA_NON_VALID;

		bool m_hashPathsToo;
		bool m_useHash;

	protected:
		//! If true sparse hash std::map is used for storing geodesics, if false dense hash std::map will be used
		void  useSparseHash(bool use);

		void setHashData(mpcard i, mpcard j, const HashData& data) const;
		bool isHashDataValid(mpcard i, mpcard j) const;
		const HashData* getHashData(mpcard i, mpcard j) const;
		const HashData* getHashData(const DiscreteManifoldPoint* pointA, const DiscreteManifoldPoint* pointB, mpcard* returnedIndexA = NULL, mpcard* returnedIndexB = NULL);
		const HashData* getHashData(const ManifoldPoint* pointA, const ManifoldPoint* pointB, mpcard* returnedIndexA = NULL, mpcard* returnedIndexB = NULL);
		std::vector<const HashData*> getHashDatas(const ManifoldPoint* pointA, const std::vector<const ManifoldPoint*>& toPoints, mpcard* returnedIndexA = NULL, std::vector<mpcard>* returnedIndicesB = NULL);

		static const mpcard NON_VALID_POINT = ~mpcard(0);

		//! Compute geodesic distance and put data into hash
		virtual float computeDistanceAndFillHash(mpcard i, mpcard j, std::vector<DVectorF>* path) = 0;

		//! Compute distances to several points at once
		virtual std::vector<float> computeDistancesAndFillHash(mpcard i, const std::vector<mpcard>& j, std::vector<std::vector<DVectorF> >* paths) = 0;

		//! Implementation to react on subdivision level changes
		virtual void setSubdivisionLevelImplementation(card32 level) {}

	public:
		DiscreteManifold();
		virtual ~DiscreteManifold();

		/**
		* Set whenever the paths between points should be precached.
		**/
		virtual void setCachePaths(bool);
		virtual bool getCachePaths() { return m_hashPathsToo; }

		/**
		* Enable/disable use of hashes for geodetic distances
		**/
		virtual void setUseCaching(bool);
		virtual bool getUseCaching() { return m_useHash; }

		/**
		* Create point of the given coordinate living on the discrete manifold. The returned point should contain the given coordinate as its
		* own coordinates and could mismatch with the position returned from the discrete point cloud using point's index.
		**/
		virtual ManifoldPoint* createPoint(const DVectorF& coordinate) = 0;
		virtual ManifoldPoint* createPoint(mpcard index, const DVectorF& coordinate = DVectorF()) = 0;

		/**
		* Get geodetic distance between two points on the manifold.
		* If distance was computed before, then value from the cache will be used, if not it will be first computed.
		**/
		float32 getGeodeticDistance(const ManifoldPoint* pointA, const ManifoldPoint* pointB);
		float32 getGeodeticDistance(const DiscreteManifoldPoint* pointA, const DiscreteManifoldPoint* pointB);
		std::vector<float32> getGeodeticDistances(const ManifoldPoint* pointA, const std::vector<const ManifoldPoint*>& toPoints);

		/**
		* Get geodetic path between two points. The returned path consist also of geodetic points.
		**/
		std::vector<DVectorF> getGeodeticPath(const ManifoldPoint* pointA, const ManifoldPoint* pointB);
		std::vector<DVectorF> getGeodeticPath(const DiscreteManifoldPoint* pointA, const DiscreteManifoldPoint* pointB);
		std::vector<const std::vector<DVectorF>*> getGeodeticPaths(const ManifoldPoint* pointA, const std::vector<const ManifoldPoint*>& toPoints);
		//std::vector<const std::vector<DVectorF>*> getGeodeticPaths(const DiscreteManifoldPoint* pointA, const std::vector<const DiscreteManifoldPoint*>& toPoints);

		/**
		* Free memory used to store geodesics. After clearing memory any access to get geodesics will force a recomputation of them.
		**/
		virtual void clearGeodesicsMemory();

		/**
		* Check if derived manifold objects do support level of subdivision.
		**/
		virtual bool supportSubdivisionLevels() { return false; }

		/**
		* Set level of subdivision used by this manifold. Derived classes should reupdate internal strucutres on change of the subdivision levels.
		**/
		virtual void setSubdivisionLevel(card32 level) { setSubdivisionLevelImplementation(level); m_SubdivisionLevel = level; }
		card32 getSubdivisionLevel() { return m_SubdivisionLevel; }

		/**
		* Get number of discrete points in this manifold
		**/
		virtual mpcard getNumPoints() = 0;

		//! Find nearest discrete point to the given coordinate
		virtual mpcard findNearestPoint(const DVectorF& coordinate) = 0;

		//virtual DiscreteManifoldPoint getPoint(mpcard index) = 0;

		// ----------------- Persistent ----------------------------
		virtual void write(OutputObjectStream *s) const;
		virtual void read(InputObjectStream *s);
		virtual void assign(const Object* obj, X4CopyContext *context);

		virtual SceneObjectIterator *createIterator(SceneObjectIteratorType type);
		virtual bool                 supportsIteratorType(SceneObjectIteratorType cap);

	};
};

#endif 
