#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "FastSphereQuerry.h"
//---------------------------------------------------------------------------
#include "AttachedIndexedOctree.h"
#include "UnstructuredInCorePointCloud.h"
//---------------------------------------------------------------------------
using namespace X4;

FastSphereQuerry::FastSphereQuerry(PointCloud * pc)
{
	m_SceneIter = pc->createIterator(SOIT::CAP_ATTACH_SECONDARY|SOIT::CAP_BASIC_PC); 
	m_SDS = dynamic_cast<SecondaryDataStructureIterator*>(m_SceneIter);
	m_BasicIt = dynamic_cast<BasicPointCloudIterator*>(m_SceneIter);

	AttachedDataContainer *attachments = m_SDS->lockAttachments();

	m_PS = (PointSet*)m_BasicIt->lockBlockRead();

	m_Octree = dynamic_cast<AttachedIndexedOctree*>(attachments->getData("SimpleOctree"));
	if(m_Octree==NULL){
		m_Octree=new AttachedIndexedOctree();
		m_Octree->setup("SimpleOctree", AttachedData::ADF_QUERY_DELETE_ON_UPDATE,  NULL);
		m_Octree->octreesetup(m_PS, 32);
		attachments->attachData(m_Octree);
	}

	m_IndexCache.resize(1000);
	m_SquaredDistanceCache.resize(m_IndexCache.size());
}

FastSphereQuerry::FastSphereQuerry(AttachedIndexedOctree * octree, PointSet * ps)
{
	m_IndexCache.resize(1000);
	m_SquaredDistanceCache.resize(m_IndexCache.size());
	m_PS = ps;

	m_Octree = octree;
	m_SceneIter = NULL;
	m_SDS = NULL;
	m_BasicIt = NULL;
}

FastSphereQuerry::~FastSphereQuerry(void)
{
	if( m_BasicIt )
		m_BasicIt->unlock();
	if( m_SDS )
		m_SDS->unlockAttachments();
	delete m_SceneIter;
}

bool sphereIntersectsBoundingBox( const Vector3f &center, float radius, const BoundingBox3f &bb) {
	float s, d = 0; 

	//find the square of the distance
	//from the sphere to the box
	for( long i=0 ; i<3 ; i++ )
	{

		if( center[i] < bb.lowerCorner[i] )
		{

			s = center[i] - bb.lowerCorner[i];
			d += s*s; 

		}

		else if( center[i] > bb.upperCorner[i] )
		{

			s = center[i] - bb.upperCorner[i];
			d += s*s; 

		}

	}
	return d <= radius*radius;
}

bool sphereSurroundsBoundingBox2(const Vector3f &center, float sqr_radius, const BoundingBox3f &bb){

	float s, d = 0; 

	for(unsigned int i=0 ; i<3 ; ++i) {
		if(center[i] < bb.lowerCorner[i]) {
			s = center[i] - bb.upperCorner[i];
			d += s*s; 
		} else if( center[i] > bb.upperCorner[i]) {
			s = center[i] - bb.lowerCorner[i];
			d += s*s; 
		} else if(center[i]-bb.lowerCorner[i] < bb.upperCorner[i]-center[i]) {
			s = center[i] - bb.upperCorner[i];
			d += s*s;
		} else {
			s = center[i] - bb.lowerCorner[i];
			d += s*s; 
		}
	}
	return d < sqr_radius;
}

bool sphereIntersectsBoundingBox2(const Vector3f &center, float sqr_radius, const BoundingBox3f &bb)
{
	float s, d = 0; 

	//find the square of the distance
	//from the sphere to the box
	for( unsigned int i=0 ; i<3 ; i++ ) {
		if( center[i] < bb.lowerCorner[i] ) {
			s = center[i] - bb.lowerCorner[i];
			d += s*s; 
		} else if( center[i] > bb.upperCorner[i] ) {
			s = center[i] - bb.upperCorner[i];
			d += s*s; 
		}
	}
	return d <= sqr_radius;
}

bool sphereIntersectsBoundingBox2(const Vector3f &center, float min_sqr_radius, float max_sqr_radius, const BoundingBox3f &bb)
{
	float s;
	float dmin = 0;
	float dmax = 0; 

	for(unsigned int i=0 ; i<3 ; ++i) {
		if(center[i] < bb.lowerCorner[i]) {
			s = center[i] - bb.lowerCorner[i];
			dmax += s*s; 
			s = center[i] - bb.upperCorner[i];
			dmin += s*s; 
		} else if( center[i] > bb.upperCorner[i]) {
			s = center[i] - bb.upperCorner[i];
			dmax += s*s; 
			s = center[i] - bb.lowerCorner[i];
			dmin += s*s; 
		} else if(center[i]-bb.lowerCorner[i] < bb.upperCorner[i]-center[i]) {
			s = center[i] - bb.upperCorner[i];
			dmin += s*s;
		} else {
			s = center[i] - bb.lowerCorner[i];
			dmin += s*s; 
		}
	}
	return (dmax <= max_sqr_radius) && (dmin >= min_sqr_radius);
}



void X4::FastSphereQuerry::querry( const Vector3f &pos, const float minRadius, const float maxRadius, mpcard ** indices, mpcard &numPoints )
{
	const float32 max_rad_2 = maxRadius * maxRadius;
	const float32 min_rad_2 = minRadius * minRadius;
	
	AIOctreeNode * rootNode = m_Octree->getOctree();
	list<AIOctreeNode*> node_queue;
	node_queue.push_back(rootNode);

	m_NumPoints = 0;
	const PointSet * ps = m_Octree->getPointSet();
	AAT posAAT = ps->getAAT("position");

	if( sphereIntersectsBoundingBox2(pos,min_rad_2,max_rad_2,rootNode->getContent().m_bb) )
	while( !node_queue.empty() )
	{
		AIOctreeNode * node = node_queue.front();
		node_queue.pop_front();

		IndexedNode & content = node->getContent();
		if( node->isLeaf() )
		{
			const unsigned numPointsInNode = (unsigned)content.m_index.size();
			if( numPointsInNode )
			{
				card32 * tmpIndicies = &content.m_index[0];	// TODO_64BIT ATTENTION! Has to be changed for 64 bit! See definition of IndexedNode.
				for( unsigned i=0;i<numPointsInNode;i++ )
				{
					const Vector3f &tmpPos = ps->get3f(*tmpIndicies,posAAT);
					Vector3f v; v[0]=tmpPos[0]-pos[0]; v[1]=tmpPos[1]-pos[1]; v[2]=tmpPos[2]-pos[2];
					const float32 sqrlv = v[0]*v[0] + v[1]*v[1] + v[2]*v[2];
					if( sqrlv <= max_rad_2 && sqrlv >= min_rad_2 )
					{
						if(  m_IndexCache.size() == m_NumPoints )
							m_IndexCache.resize( m_IndexCache.size() * 2);
						m_IndexCache[m_NumPoints++] = *tmpIndicies;

					}
					tmpIndicies++;
				}
			}
		}
		else
		{
			// does sphere shell intersect this quandrant at all?
			// if (!sphereIntersectsBoundingBox2(pos,min_rad_2,max_rad_2,content.m_bb)) {
			if (sphereSurroundsBoundingBox2(pos,min_rad_2,content.m_bb)  ) {
				continue;
			}

			const Vector3f diff = pos - content.m_bb.getCenter();
			const float diff0_sqr = diff[0]*diff[0];
			const float diff1_sqr = diff[1]*diff[1];
			const float diff2_sqr = diff[2]*diff[2];

			// do all octants intersect?
			float diff_sqr = diff0_sqr + diff1_sqr + diff2_sqr;
			if( diff_sqr < max_rad_2 )
			{
				for( unsigned i=0;i<8;i++ )
				{
					AIOctreeNode * childNode = node->getChild(i);
					if( childNode )
						node_queue.push_back(childNode);
				}
				continue;
			}
			
			// determine octant in which pos is
			unsigned octant = (diff[0] < 0 ? 0:1)
				+ (diff[1] < 0 ? 0:2)
				+ (diff[2] < 0 ? 0:4);

			// determine if a halfspace doesn't intersect the sphere
			unsigned no_intersection = 1 << (~octant & 7);	// bitmask, mark complementary octant as culled
			if( diff0_sqr > max_rad_2 )
				no_intersection |= (diff[0] < 0)? 2+8+32+128 : 1+4+16+64;
			if( diff1_sqr > max_rad_2 )
				no_intersection |= (diff[1] < 0)? 4+8+64+128 : 1+2+16+32;
			if( diff2_sqr > max_rad_2 )
				no_intersection |= (diff[2] < 0)? 16+32+64+128 : 1+2+4+8;

			// check 3 edge neighbor cells for intersection (cells that share an edge)
			unsigned nIndex1 = (~octant & 3) | (octant & 4);
			unsigned nIndex2 = (~octant & 5) | (octant & 2);
			unsigned nIndex3 = (~octant & 6) | (octant & 1);
			unsigned no_intersection1 = diff0_sqr + diff1_sqr > max_rad_2 ? 1:0;
			unsigned no_intersection2 = diff0_sqr + diff2_sqr > max_rad_2 ? 1:0;
			unsigned no_intersection3 = diff1_sqr + diff2_sqr > max_rad_2 ? 1:0;
			no_intersection |= (no_intersection1 << nIndex1);
			no_intersection |= (no_intersection2 << nIndex2);
			no_intersection |= (no_intersection3 << nIndex3);


			// add children that intersect the half space
			AIOctreeNode * childNode;
			for( unsigned i=0;i<8;i++ )
			{
				childNode = node->getChild(i);
				if( (!(no_intersection & 1)) )
				{
					if( i == octant || sphereIntersectsBoundingBox2(pos,min_rad_2,max_rad_2,childNode->getContent().m_bb) )
					{
						node_queue.push_back(childNode);
					}
				}
				no_intersection = no_intersection >> 1;
			}


		}
	}
	numPoints = m_NumPoints;
	if(numPoints > 0)
		*indices = &m_IndexCache[0];
	else
		*indices = NULL;
}

void X4::FastSphereQuerry::querry( const Vector3f &pos, const float radius, mpcard ** indices, mpcard &numPoints )
{
	const float rad_2 = radius * radius;
	
	AIOctreeNode * rootNode = m_Octree->getOctree();
	node_queue.clear();
	node_queue.push_back(rootNode);

	m_NumPoints = 0;
	const PointSet * ps = m_PS;
	AAT posAAT = ps->getAAT("position");

	if( sphereIntersectsBoundingBox2(pos,rad_2,rootNode->getContent().m_bb) )
	while( !node_queue.empty() )
	{
		AIOctreeNode * node = node_queue.pop_back();
		//node_queue.pop_back();

		IndexedNode & content = node->getContent();
		if( node->isLeaf() )
		{
			const size_t numPointsInNode = content.m_index.size();
			if( numPointsInNode )
			{
				card32 * tmpIndicies = &content.m_index[0];	// TODO_64BIT ATTENTION! Has to be changed for 64 bit! See definition of IndexedNode.
				for( unsigned i=0;i<numPointsInNode;i++ )
				{
					const Vector3f &tmpPos = ps->get3f(*tmpIndicies,posAAT);
					Vector3f tmpDiff = tmpPos - pos;
					float dist2 = //normQuad(tmpDiff);
						tmpDiff[0]*tmpDiff[0] + tmpDiff[1]*tmpDiff[1] + tmpDiff[2]*tmpDiff[2];
					if( dist2 <= rad_2 )
					{
						if(  m_IndexCache.size() == m_NumPoints )
						{
							m_IndexCache.resize( m_IndexCache.size() * 2);
							m_SquaredDistanceCache.resize( m_IndexCache.size() );
						}
						m_SquaredDistanceCache[m_NumPoints] = dist2;
						m_IndexCache[m_NumPoints++] = *tmpIndicies;

					}
					tmpIndicies++;
				}
			}
		}
		else
		{
			const Vector3f diff = pos - content.m_bb.getCenter();
			const float diff0_sqr = diff[0]*diff[0];
			const float diff1_sqr = diff[1]*diff[1];
			const float diff2_sqr = diff[2]*diff[2];
			
			// do all octants intersect?
			float diff_sqr = diff0_sqr + diff1_sqr + diff2_sqr;
			if( diff_sqr < rad_2 )
			{
				for( unsigned i=0;i<8;i++ )
				{
					AIOctreeNode * childNode = node->getChild(i);
					if( childNode )
						node_queue.push_back(childNode);
				}
				continue;
			}
			
			// determine octant in which pos is
			unsigned octant = (diff[0] < 0 ? 0:1)
				+ (diff[1] < 0 ? 0:2)
				+ (diff[2] < 0 ? 0:4);

			// determine if a halfspace doesn't intersect the sphere
			unsigned no_intersection = 1 << (~octant & 7);	// bitmask, mark complementary octant as culled
			if( diff0_sqr > rad_2 )
				no_intersection |= (diff[0] < 0)? 2+8+32+128 : 1+4+16+64;
			if( diff1_sqr > rad_2 )
				no_intersection |= (diff[1] < 0)? 4+8+64+128 : 1+2+16+32;
			if( diff2_sqr > rad_2 )
				no_intersection |= (diff[2] < 0)? 16+32+64+128 : 1+2+4+8;

			// check 3 edge neighbor cells for intersection (cells that share an edge)
			unsigned nIndex1 = (~octant & 3) | (octant & 4);
			unsigned nIndex2 = (~octant & 5) | (octant & 2);
			unsigned nIndex3 = (~octant & 6) | (octant & 1);
			unsigned no_intersection1 = diff0_sqr + diff1_sqr > rad_2 ? 1:0;
			unsigned no_intersection2 = diff0_sqr + diff2_sqr > rad_2 ? 1:0;
			unsigned no_intersection3 = diff1_sqr + diff2_sqr > rad_2 ? 1:0;
			no_intersection |= (no_intersection1 << nIndex1);
			no_intersection |= (no_intersection2 << nIndex2);
			no_intersection |= (no_intersection3 << nIndex3);


			// add children that intersect the half space
			AIOctreeNode * childNode;
			for( unsigned i=0;i<8;i++ )
			{
				childNode = node->getChild(i);
				if( (!(no_intersection & 1)) )
				{
					if( i == octant || sphereIntersectsBoundingBox2(pos,rad_2,childNode->getContent().m_bb) )
					{
						node_queue.push_back(childNode);
					}
				}
				no_intersection = no_intersection >> 1;
			}


		}
	}
	numPoints = m_NumPoints;
	if(numPoints > 0)
		*indices = &m_IndexCache[0];
	else
		*indices = NULL;
}

void X4::FastSphereQuerry::process( const Vector3f &pos, const float radius, Object * obj, ProcessPointMethod method )
{
	const float rad_2 = radius * radius;

	AIOctreeNode * rootNode = m_Octree->getOctree();
	list<AIOctreeNode*> node_queue;
	node_queue.push_back(rootNode);

	const PointSet * ps = m_Octree->getPointSet();
	AAT posAAT = ps->getAAT("position");

	if( sphereIntersectsBoundingBox2(pos,rad_2,rootNode->getContent().m_bb) )
		while( !node_queue.empty() )
		{
			AIOctreeNode * node = node_queue.front();
			node_queue.pop_front();

			IndexedNode & content = node->getContent();
			if( node->isLeaf() )
			{
				const size_t numPointsInNode = content.m_index.size();
				if( numPointsInNode )
				{
					card32 * tmpIndicies = &content.m_index[0];	// TODO_64BIT ATTENTION! Has to be changed for 64 bit! See definition of IndexedNode.
					for( unsigned i=0;i<numPointsInNode;i++ )
					{
						const Vector3f tmpPos = ps->get3f(*tmpIndicies,posAAT);
						float dist_2 = normQuad(tmpPos - pos);
						if( dist_2 <= rad_2 )
						{
							(obj->*method)(tmpPos, dist_2, *tmpIndicies);
						}
						tmpIndicies++;
					}
				}
			}
			else
			{
				const Vector3f diff = pos - content.m_bb.getCenter();
				const float diff0_sqr = diff[0]*diff[0];
				const float diff1_sqr = diff[1]*diff[1];
				const float diff2_sqr = diff[2]*diff[2];

				// do all octants intersect?
				float diff_sqr = diff0_sqr + diff1_sqr + diff2_sqr;
				if( diff_sqr < rad_2 )
				{
					for( unsigned i=0;i<8;i++ )
					{
						AIOctreeNode * childNode = node->getChild(i);
						if( childNode )
							node_queue.push_back(childNode);
					}
					continue;
				}

				// determine octant in which pos is
				unsigned octant = (diff[0] < 0 ? 0:1)
					+ (diff[1] < 0 ? 0:2)
					+ (diff[2] < 0 ? 0:4);

				// determine if a halfspace doesn't intersect the sphere
				unsigned no_intersection = 1 << (~octant & 7);	// bitmask, mark complementary octant as culled
				if( diff0_sqr > rad_2 )
					no_intersection |= (diff[0] < 0)? 2+8+32+128 : 1+4+16+64;
				if( diff1_sqr > rad_2 )
					no_intersection |= (diff[1] < 0)? 4+8+64+128 : 1+2+16+32;
				if( diff2_sqr > rad_2 )
					no_intersection |= (diff[2] < 0)? 16+32+64+128 : 1+2+4+8;

				// check 3 edge neighbor cells for intersection (cells that share an edge)
				unsigned nIndex1 = (~octant & 3) | (octant & 4);
				unsigned nIndex2 = (~octant & 5) | (octant & 2);
				unsigned nIndex3 = (~octant & 6) | (octant & 1);
				unsigned no_intersection1 = diff0_sqr + diff1_sqr > rad_2 ? 1:0;
				unsigned no_intersection2 = diff0_sqr + diff2_sqr > rad_2 ? 1:0;
				unsigned no_intersection3 = diff1_sqr + diff2_sqr > rad_2 ? 1:0;
				no_intersection |= (no_intersection1 << nIndex1);
				no_intersection |= (no_intersection2 << nIndex2);
				no_intersection |= (no_intersection3 << nIndex3);


				// add children that intersect the half space
				AIOctreeNode * childNode;
				for( unsigned i=0;i<8;i++ )
				{
					childNode = node->getChild(i);
					if( (!(no_intersection & 1)) )
					{
						if( i == octant || sphereIntersectsBoundingBox2(pos,rad_2,childNode->getContent().m_bb) )
						{
							node_queue.push_back(childNode);
						}
					}
					no_intersection = no_intersection >> 1;
				}


			}
		}
}

/* didn't brought any improvement ...

void X4::optimizePointOrder( UnstructuredInCorePointCloud * pc )
{
	// get/create octree
	AttachedIndexedOctree * octree = dynamic_cast<AttachedIndexedOctree*>(pc->getAttachments()->getData("SimpleOctree"));
	if(octree==NULL){
		octree=new AttachedIndexedOctree();
		octree->setup("SimpleOctree", AttachedData::ADF_QUERY_DELETE_ON_UPDATE,  NULL);
		octree->octreesetup(pc->getPointSet(), 32);
		pc->getAttachments()->attachData(octree);
	}

	// create a reordered point set
	PointSet * ps_new = new PointSet;
	PointSet * ps = pc->getPointSet();
	ps_new->clearAndSetup(1, ps->getNumEntries(), ps->getDescr());

	list<AIOctreeNode*> nodes;
	nodes.push_back( octree->getOctree() );
	char * dest_ptr = ps_new->getDataPointer(0);
	const unsigned vertex_size = ps->getDescr()->getSize();

	while( !nodes.empty() )
	{
		AIOctreeNode * node = nodes.front();
		nodes.pop_front();

		if( node->isLeaf() )
		{
			for( unsigned i=0;i<node->getContent().m_index.size();i++ )
			{
				ps->getVertex(node->getContent().m_index[i], dest_ptr);
				dest_ptr += vertex_size;
			}
		}
		else
		{
			for( unsigned i=0;i<8;i++ )
				nodes.push_back(node->getChild(i));
		}
	}
	pc->setPointSet(ps_new);

	// recreate octree
	octree = dynamic_cast<AttachedIndexedOctree*>(pc->getAttachments()->getData("SimpleOctree"));
	delete octree;
	octree=new AttachedIndexedOctree();
	octree->setup("SimpleOctree", AttachedData::ADF_QUERY_DELETE_ON_UPDATE,  NULL);
	octree->octreesetup(pc->getPointSet(), 32);
	pc->getAttachments()->attachData(octree);
}*/