//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "GaussianPointCloudSmoother.h"
#include "X4Exception.h"
#include "FastSphereQuerry.h"
#include "UnstructuredInCorePointCloud.h"
#include "HierarchicalKNNIterator.h"
#include "MathTools.inline.h"
//#include "PointCloudTools.h"
#include <float.h>
#include "DynamicLinearAlgebra.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------


static std::string descrType(card32 i) {
   if (i == GaussianPointCloudSmoother::TYPE_GAUSS) return "TYPE_GAUSS";
//   if (i == GaussianPointCloudSmoother::TYPE_COS) return "TYPE_COS";
   return "UNKNOWN";
}

IMPLEMENT_X4_CLASS(GaussianPointCloudSmoother, 0)
{
   BEGIN_CLASS_INIT(GaussianPointCloudSmoother)
   
   ADD_FLOAT32_PROP(stdDevPerGridCell, 0)
   ADD_CARD32_PROP_UPDATE_DESCR(type, 0, NULL, descrType, 1, NULL)
}

//---------------------------------------------------------------------------------------

X4::GaussianPointCloudSmoother::GaussianPointCloudSmoother()
{
   stdDevPerGridCell = 1/sqrt(2.0f);
   type = 0;
}

//---------------------------------------------------------------------------------------

float32 X4::GaussianPointCloudSmoother::computeGaussianWeight(float32 x, float32 sigma) const
{
   // variance = sigma^2
   return expf(-x*x / (2.0 * sigma * sigma)); // don't need to normalize!
}


//---------------------------------------------------------------------------------------

UnstructuredInCorePointCloud * X4::GaussianPointCloudSmoother::downsamplePointCloud(
   UnstructuredInCorePointCloud * input, 
   float32 currentGridSpacing, 
   float32 smoothingFactor) const
{
   if (input == NULL) return NULL;
   if (!input->providesAttribute( "position" )) {
      throw EInvalidState("GaussianPointCloudSmoother::downsamplePointCloud() - point cloud does not provide attribute 'position'.");
   }
   if (smoothingFactor < 1.0f) {
      //Todo: (robert) implement upsampling instead
      throw X4Exception("GaussianPointCloudSmoother::downsamplePointCloud() - Smooting factor must be >= 1.");
   }
   if (currentGridSpacing <= 0.0f) {
      // not estimated the grid spacing
      throw X4Exception("GaussianPointCloudSmoother::downsamplePointCloud() - currentGridSpacing is zero.");
   }
   // Create iterators
   SceneObjectIterator * iter = input->createIterator( SceneObjectIteratorType::CAP_BASIC_PC );
   BasicPointCloudIterator * basicIt = dynamic_cast<BasicPointCloudIterator*>(iter);
   if ( basicIt == NULL ) {
      throw  X4Exception("GaussianPointCloudSmoother::downsamplePointCloud() - Cannot create basic iterator.");
   }
   const VertexDescriptor * vd = input->getDescr();
   UnstructuredInCorePointCloud * target = new UnstructuredInCorePointCloud();
   target->clearAndSetup( vd, 0);
   PointSet * psTarget = target->getPointSet(); 


   card32 numPoints = input->getNumPoints();
   const PointSet *sourcePS = basicIt->lockBlockRead();
   /// michael: for attribute filtering, i need to assume that a single point set represents the leave node data, and that the iterator starts here.
   if (numPoints != sourcePS->getNumEntries()) {
      basicIt->unlock();
      delete target;
      throw ERangeCheck("GaussianPointCloudSmoother::downsamplePointCloud() - The point cloud is not a single block container (at the initial iterator resolution) - aborted. This is a limitation of the current implementation (mw-11/13/12).");
   }
   
   std::vector<char> tmp_vertex( vd->getSize());
   std::vector<card8> flags(numPoints, 0);

   AAT POSITION = input->getAAT( "position" );

   //FastSphereQuerry * sphereQuery = new FastSphereQuerry(input);
   HierarchicalKNNIterator* knnIt = new HierarchicalKNNIterator( input, 32, NULL );

   const float32 downsampleRatio = smoothingFactor;
   const float32 sigma = currentGridSpacing * stdDevPerGridCell * downsampleRatio;
   const float32 searchRadius = sigma * 3.0f; // hopefully 3 stdev. should be enough
   const float32 maxRadiusSqr = sqr(searchRadius);
   const float32 targetGridSpacing = currentGridSpacing * downsampleRatio;
   const float32 targetGridSpacingSqr = sqr(targetGridSpacing);

   // Reset all flags
   basicIt->reset();
   card32 targetPointIndex = 0;
   card32 sourcePointIndex = 0;

   while( !basicIt->atEnd() ){
      const Vector3f& p = basicIt->get3f( POSITION );
      if ( _isnan( p[0] ) || _isnan( p[1] ) || _isnan( p[2] ) ){
         continue;
      }
      if ( !_finite( p[0] ) || !_finite( p[1] ) || !_finite( p[2] ) ){
         continue;
      }
      DVectorF accumulator(vd->computeOverallDimension());
      DVectorF temp(vd->computeOverallDimension());
      const int32 flag = flags[sourcePointIndex];
      if ( flag == 0 ){
         Vector3f newPos;
         newPos.setZero();
         float32 sumWeights = 0.0f;
         accumulator.setZero();
         basicIt->getVertex( &tmp_vertex[0] );
         psTarget->addPoint( &tmp_vertex[0] );
         knnIt->setSeekPointAndReset( p );
         while( !knnIt->atEnd() ){
            const Vector3f& np = knnIt->get3f( POSITION );
            const mpcard knnIndex = (mpcard)knnIt->getCurrentPointIndex();
            vd->getAsDVector((const card8*)sourcePS->getDataPointer(knnIndex), temp);
            
            float sqrD = ( p - np ) * ( p - np );
            if (_isnan(sqrD) || !_finite(sqrD))
               continue;
            if( sqrD > maxRadiusSqr )
               break;
            if (sqrD < targetGridSpacingSqr) {
               const card32 index = knnIndex;
               flags[index] = 1;
            }
            const float32 w = computeGaussianWeight(sqrD, sigma);
            temp *= w;
            accumulator += temp;
            newPos += w * np;
            sumWeights += w;
            knnIt->next();
         }
         float32 rescale = 1.0f/max(1e-9f, sumWeights);
         newPos *= rescale;
         accumulator *= rescale;
         psTarget->set3f(targetPointIndex, POSITION, newPos);
         vd->setByDVector((card8*)psTarget->getDataPointer(targetPointIndex), accumulator);
         targetPointIndex++;
      }
      basicIt->next();
      sourcePointIndex++;
   }

   delete knnIt;
   basicIt->unlock();
   delete iter;

   return target;
}

//---------------------------------------------------------------------------------------

UnstructuredInCorePointCloud * X4::GaussianPointCloudSmoother::createSmoothed(
    UnstructuredInCorePointCloud *input, 
    float32 currentGridSpacing, 
    float32 smoothingFactor) const
{
	// 1. compute nearest neighbor radius R from standard dev. and currentGridSpacing
	// 2. downsample the pointcloud uniformly with given target grid spacing and store in pcTarget
   // 3. loop over all points in pcTarget and query epsilon neighborhood in input within radius R
   // 4. for each point compute the Gaussian weighted average over all found nearest neighbor points (based on Euclidean distance)
   // 5. set the weighted average for position (and other attributes) in the pcTarget
   UnstructuredInCorePointCloud * pcTarget = downsamplePointCloud(input, currentGridSpacing, smoothingFactor);
	return pcTarget;
}