//---------------------------------------------------------------------------
#ifndef GenericSwapAllOutOfCorePointCloudH
#define GenericSwapAllOutOfCorePointCloudH
//---------------------------------------------------------------------------
#include "x3geometry.h"
//---------------------------------------------------------------------------
#include "PointCloud.h"
#include <QtCore/QMutex>
//---------------------------------------------------------------------------

namespace X4 {

class GSAOOCPointCloudIterator;

///
///   This class encapsulates a point cloud that is completely swapped out of core if not needed
///
class X3GEOMETRY_API GenericSwapAllOutOfCorePointCloud : public PointCloud {
	X4_CLASS( GenericSwapAllOutOfCorePointCloud )
 private:
	class GSAPointCloudPool {
	 private:
		struct EntryType {
			PointCloud* pc;
			mpcard size;
		   card32 lockCounter;
		   bool dirty;
		   list<std::string>::iterator lruEntry;
		};
	 
		typedef std::map<std::string, EntryType> MapType;
		
		mpint maxMem;
		mpint memUsed;
		MapType inCore;

		list<std::string> lruQueue;
		
		void checkSwap();
		mpcard computeSize(PointCloud *pc);
		void updateSize(EntryType &e);

		QMutex mutex;
		
	 public:
		GSAPointCloudPool(mpcard maxMem = 1024*1024*1024);
		void sync();
		PointCloud *fetchPC(std::string filename);
		enum SetDirtyFlag {setClean, setDirty, noChange};
		/// put in front of lru queue, update memory usage, update lock status
		void update(std::string filename, card32 lockCounter, SetDirtyFlag dirty);
		void setMaxMem(mpcard maxMem);
		mpcard getMaxMem() {return maxMem;}
		~GSAPointCloudPool();
	};
	
	static GSAPointCloudPool *pcPool;
	
	std::string filename;
	bool relativePath;
	mutable card32 lockCounter;
	mutable VertexDescriptor *vertexDescriptorCopy;
	
	void incrementLockCounter() const;
	void decrementLockCounter() const;
	void updatePC();
	void setDirty();
	static void truncatePathRelativeTo(std::string basePath, std::string filename, std::string &result, bool &matches);
	
 public:
	GenericSwapAllOutOfCorePointCloud();

	void clearAndSetup(const PointCloud *pc, std::string filename, bool relativePath = true);

	virtual void assign(const Object* obj, X4CopyContext *context = NULL);
	virtual void read(InputObjectStream *s);
	virtual void write(OutputObjectStream *s) const;
   virtual SceneObject *createOOCAwareCopy(std::string idPrefix = "", bool insertOriginal = true, std::string idSuffix = "");
	
   virtual const VertexDescriptor *getDescr() const;
   virtual PointCloud             *appendFields(const VertexDescriptor *newAttribs) const;
   virtual card64                  getNumPoints() const;
   virtual void                    clear();
   virtual SceneObjectIterator    *createIterator(SceneObjectIteratorType type);
   virtual bool                    supportsIteratorType(SceneObjectIteratorType cap);

	virtual PointCloud *fetchPC();

	void lock() { incrementLockCounter(); }
	void unlock() { x4Assert(lockCounter > 0); decrementLockCounter(); }

	std::string getFileName() {return filename;}
	void setFileName(const std::string &fnew) {filename = fnew;}

	static mpcard getMaxMem();
	static void setMaxMem(mpcard mm);

	static void finalShutdownCache();
   ~GenericSwapAllOutOfCorePointCloud();
 friend GSAOOCPointCloudIterator;
};


class GSAOOCPointCloudIterator : public PointCloudIterator,
 public SecondaryDataStructureIterator, public HierarchicalIterator, 
 public BasicPointCloudIterator, public IndexedPointCloudIterator, 
 public RandomAccessPointCloudIterator, public ModifyablePointCloudIterator,
 public DynamicPointCloudIterator, public DynamicParametricPointCloudIterator, 
 public ParametricPointCloudIterator, public TopologyIterator, public VolumetricIterator {

	X4_CLASS( GSAOOCPointCloudIterator )
 private:
	GenericSwapAllOutOfCorePointCloud *myPC;
	SceneObjectIterator *it;
	SecondaryDataStructureIterator *interfaceSecondaryDataStructureIterator;
	HierarchicalIterator *interfaceHierarchicalIterator;
	BasicPointCloudIterator *interfaceBasicPointCloudIterator;
	IndexedPointCloudIterator *interfaceIndexedPointCloudIterator;
	RandomAccessPointCloudIterator *interfaceRandomAccessPointCloudIterator;
	ModifyablePointCloudIterator *interfaceModifyablePointCloudIterator;
	DynamicPointCloudIterator *interfaceDynamicPointCloudIterator;
	DynamicParametricPointCloudIterator *interfaceDynamicParametricPointCloudIterator;
	ParametricPointCloudIterator *interfaceParametricPointCloudIterator;
	TopologyIterator *interfaceTopologyIterator;
	VolumetricIterator *interfaceVolumetricIterator;

 public:
	GSAOOCPointCloudIterator();
	GSAOOCPointCloudIterator(GenericSwapAllOutOfCorePointCloud *myPC, SceneObjectIterator *it);
 
	virtual void assign(const Object* obj, X4CopyContext *context = NULL);

	virtual AttachedDataContainer *lockAttachments();
	virtual void unlockAttachments();

   virtual card32 getNumChildren();
   virtual bool goToChild(card32 index);
   virtual bool goToParent();
   virtual BoundingBox3f getBoundingBox();
   virtual bool isLeaf() const;
   virtual bool getNodeInfo(HierarchicalIterator::NodeInfoType whichInfo, const void *parameter, void *data);

   virtual void reset();
   virtual void next();
   virtual bool atEnd();
   virtual card8     get1ub(const AAT &token) const;
   virtual Vector2ub get2ub(const AAT &token) const;
   virtual Vector3ub get3ub(const AAT &token) const;
   virtual Vector4ub get4ub(const AAT &token) const;
   virtual int32    get1i(const AAT &token) const;
   virtual Vector2i get2i(const AAT &token) const;
   virtual Vector3i get3i(const AAT &token) const;
   virtual Vector4i get4i(const AAT &token) const;
   virtual card32   get1u(const AAT &token) const;
   virtual Vector2u get2u(const AAT &token) const;
   virtual Vector3u get3u(const AAT &token) const;
   virtual Vector4u get4u(const AAT &token) const;
   virtual float32  get1f(const AAT &token) const;
   virtual Vector2f get2f(const AAT &token) const;
   virtual Vector3f get3f(const AAT &token) const;
   virtual Vector4f get4f(const AAT &token) const;
   virtual void     getVertex(char *buffer) const;
   virtual bool nextBlock();
   virtual const PointSet * lockBlockRead();
   virtual void unlock();

   virtual card64 getNumPointsInBlock() const; 
   virtual card64 getCurrentPointIndex() const;

   virtual void goToIndex(card64 i);

   virtual void set1ub(const AAT &token, const card8 &x)     const;
   virtual void set2ub(const AAT &token, const Vector2ub &v) const;
   virtual void set3ub(const AAT &token, const Vector3ub &v) const;
   virtual void set4ub(const AAT &token, const Vector4ub &v) const;

   virtual void set1i(const AAT &token, const int32 &x)     const;
   virtual void set2i(const AAT &token, const Vector2i &v) const;
   virtual void set3i(const AAT &token, const Vector3i &v) const;
   virtual void set4i(const AAT &token, const Vector4i &v) const;

   virtual void set1u(const AAT &token, const card32 &x)     const;
   virtual void set2u(const AAT &token, const Vector2u &v) const;
   virtual void set3u(const AAT &token, const Vector3u &v) const;
   virtual void set4u(const AAT &token, const Vector4u &v) const;

   virtual void set1f(const AAT &token, const float32 &x)  const;
   virtual void set2f(const AAT &token, const Vector2f &v) const;
   virtual void set3f(const AAT &token, const Vector3f &v) const;
   virtual void set4f(const AAT &token, const Vector4f &v) const;

   virtual void setVertex(const char *data) const;
   virtual PointSet *lockBlockReadWrite(ModifyablePointCloudIterator::LockRWHint hint = ModifyablePointCloudIterator::CHANGE_EVERYTHING);

   virtual void add(const char *newPoint);
   virtual void add(const PointSet *newPoints);
   virtual void deleteCurrent();

   virtual void setSize(card32 width, card32 height);

   virtual void goTo(int32 x, int32 y);
   virtual card32 getWidth();
   virtual card32 getHeight();

	virtual card32 getNumNeighbours();
	virtual PointIndex getNeighbour( card32 nIndex );

	virtual void goTo(card32 x, card32 y, card32 z);
	virtual Vector3i getVolumeSize();
	virtual BoundingBox3f getVolumeRegion();
	virtual const RegularVolume *lockVolumeBlockRead();
	virtual RegularVolume *lockVolumeBlockReadWrite();
	virtual void unlockVolume();

	virtual ~GSAOOCPointCloudIterator();
};


}

#endif