//---------------------------------------------------------------------------
#ifndef GeometricRangeH
#define GeometricRangeH
//---------------------------------------------------------------------------
#include "x3geometry.h"
//---------------------------------------------------------------------------
#include "BoundingBox.h"
#include "LinearAlgebra.h"
#include "math_include.h"
//---------------------------------------------------------------------------

namespace X4 {


/// parent interface class for all geometric ranges. ranges are created in the worldcoordinate system independent of cloud- or SceneGraph-transformations, so you always have to call setTransformation() with the appropriate ctm-matrix before using a range.
class X3GEOMETRY_API GeometricRange { /* INTERFACE CLASS */
 public:
   /// transforms the geometric range, e.g. according to the cloud transformation in SceneGraph. Ranges are created in the worldcoordinate system independent of cloud- or SceneGraph-transformations, so you always have to call setTransformation() with the appropriate ctm-matrix before using a range.
   virtual void setTransformation(const Matrix4f&) = 0;
   /// returns true if the geometric range intersects the given BoundingBox
   virtual bool intersects(const BoundingBox3f&)const = 0;
   /// returns true if the given point is inside the geometric range
   virtual bool intersects(const Vector3f&)const = 0;
};

/// class for AABB geometric range, defined by two points.
class X3GEOMETRY_API AABBRange :public GeometricRange{
 private:
   Vector3f lowerCorner;
   Vector3f upperCorner;

 public:
   AABBRange(Vector3f lowerCorner, Vector3f upperCorner);

   virtual void setTransformation(const Matrix4f& m);
   virtual bool intersects(const BoundingBox3f& bb)const;
   virtual bool intersects(const Vector3f& p)const;

};

/// class for spherical geometric range, defined by center and (squared) radius.
class X3GEOMETRY_API SphereRange :public GeometricRange{
 private:
   /// defines the sphere center
   Vector3f Center;
   /// holds transformed center after setTransformation(ctm) is applied
   Vector3f TransCenter;
   /// this is the square radius of the sphere
   float32 SqrRadius;
   /// this is the transformed square radius of the sphere after setTransformation(ctm) is applied
   float32 TransSqrRadius;
   /// this is the radius of the sphere
   float32 Radius;
   /// this is the transformed radius of the sphere after setTransformation(ctm) is applied
   float32 TransRadius;
   /// ??? needed for what???
   float32 scaleFactor;

 public:
   SphereRange(Vector3f cent, float32 sqrrad);

   virtual void setTransformation(const Matrix4f&);
   virtual bool intersects(const BoundingBox3f&)const;
   virtual bool intersects(const Vector3f&)const;

   /// returns transformed center
   inline Vector3f getTransCenter() const { return TransCenter; };
   /// returns center
   inline Vector3f getCenter() const { return Center; };
   /// returns radius
   inline float32 getRadius() const { return Radius; };
   /// sets radius
   inline void setRadius(float32 r) { Radius = r; SqrRadius = r*r; };
};



/// class for a cone shaped geometric range, defined by a coordinate system with cone axis as direction std::vector, an origin and the (squared) cone radius in ray-z-distance 1.0f defining the opening angle.
class X3GEOMETRY_API RayConeRange :public GeometricRange{
 private:
   /// cone coordinate system right-std::vector
   Vector3f RayRight;
   /// cone coordinate system up-std::vector
   Vector3f RayUp;
   /// cone coordinate system direction-std::vector
   Vector3f RayDir;
   /// cone coordinate system origin
   Vector3f RayOrigin;
   /// squared cone radius at z-distance 1.0f in cone-coordinate-system defining the opening angle
   float32 SqrRadiusAt1;
   /// cone radius at z-distance 1.0f in cone-coordinate-system defining the opening angle
   float32 RadiusAt1;
   /// ray coordinate system
   Matrix4f RaySystem;
   /// transformed coordinate system, see setTransformation()
   Matrix4f TransRaySystem;
   /// ??? needed for what???
   float32 scaleFactor;

 public:
   RayConeRange(Vector3f right, Vector3f up, Vector3f dir, Vector3f origin, float32 sra1);
   virtual void setTransformation(const Matrix4f&);
   virtual bool intersects(const BoundingBox3f&)const;
   virtual bool intersects(const Vector3f&)const;

   /// returns the z-distance of the input point from ray origin in ray system coordinates
   inline float32 getPointZInRayCoords(const Vector3f& Point) const {
      // transform point into ray coordinate system
      Vector3f PointRayPos = transformVector3f(TransRaySystem, Point);
      // return z value (negative if in ray direction!)
      return PointRayPos[2];
   }

   /// returns the squared distance from footpoint on the ray axis to input point and the z-distance of the input point from ray origin in ray system coordinates
   inline void getSqrDistanceAndZ(const Vector3f& Point, float32& retSqrDist, float32& retZ) const {
      // transform point into ray coordinate system
      Vector3f PointRayPos = transformVector3f(TransRaySystem, Point);
      // return squared distance and z value (negative if in ray direction!)
      retSqrDist = PointRayPos[1]*PointRayPos[1]+PointRayPos[0]*PointRayPos[0];
      retZ = PointRayPos[2];
   }

   /// returns the "normalized" squared distance from footpoint on the ray axis to input point (SqrDist to the point divided by ConeRadius at this z-value results in a float between 0.0 and 1.0).
   float32 getDistNorm(const Vector3f&)const;

   /// returns cone coordinate system right-std::vector
   inline Vector3f getRayRight() const { return RayRight; };
   /// returns cone coordinate system up-std::vector
   inline Vector3f getRayUp() const { return RayUp; };
   /// returns cone coordinate system direction-std::vector
   inline Vector3f getRayDir() const { return RayDir; };
   /// returns cone coordinate system origin
   inline Vector3f getRayOrigin() const { return RayOrigin; };
   /// returns cone coordinate system
   inline Matrix4f getRaySystem() const { return RaySystem; };
   /// returns transformed cone coordinate system, see setTransformation()
   inline Matrix4f getTransRaySystem() const { return TransRaySystem; };
   /// returns squared radius at z-distance 1, see SqrRadiusAt1
   inline float32 getSqrRadiusAt1() const { return SqrRadiusAt1; };
};



/// class for frustum shaped geometric range, defined by a coordinate system with a direction std::vector, an origin and the (squared) half width and half height at ray-z-distance 1.0f defining the horizontal and vertical opening angle.
class X3GEOMETRY_API RayPyramidalRange :public GeometricRange{
 protected:
   /// frustum coordinate system right-std::vector
   Vector3f RayRight;
   /// frustum coordinate system up-std::vector
   Vector3f RayUp;
   /// frustum coordinate system direction-std::vector
   Vector3f RayDir;
   /// frustum coordinate system origin
   Vector3f RayOrigin;
   /// squared half frustum width at z-distance 1.0f in frustum-coordinate-system defining the horizontal opening angle
   float32 SqrHalfWidthAt1;
   /// squared half frustum height at z-distance 1.0f in frustum-coordinate-system defining the vertical opening angle
   float32 SqrHalfHeightAt1;
   /// half frustum width at z-distance 1.0f in frustum-coordinate-system defining the horizontal opening angle
   float32 HalfWidthAt1;
   /// half frustum height at z-distance 1.0f in frustum-coordinate-system defining the vertical opening angle
   float32 HalfHeightAt1;
   /// returns frustum coordinate system
   Matrix4f RaySystem;
   /// returns transformed frustum coordinate system, see setTransformation()
   Matrix4f TransRaySystem;
   /// ??? needed for what???
   float32 scaleFactor;

 public:
   RayPyramidalRange(Vector3f right, Vector3f up, Vector3f dir, Vector3f origin, float32 sqrhw, float32 sqrhh);
   virtual void setTransformation(const Matrix4f&);
   virtual bool intersects(const BoundingBox3f&)const;
   virtual bool intersects(const Vector3f&)const;

   /// returns the z-distance of the input point from ray origin in ray system coordinates
   inline float32 getPointZInRayCoords(const Vector3f& Point) const {
      // transform point into ray coordinate system
      Vector3f PointRayPos = transformVector3f(TransRaySystem, Point);
      // return z value (negative if in ray direction!)
      return PointRayPos[2];
   }

   /// returns the squared distance from footpoint on the ray axis to input point and the z-distance of the input point from ray origin in ray system coordinates
   inline void getSqrDistanceAndZ(const Vector3f& Point, float32& retSqrDist, float32& retZ) const {
      // transform point into ray coordinate system
      Vector3f PointRayPos = transformVector3f(TransRaySystem, Point);
      // return squared distance and z value (negative if in ray direction!)
      retSqrDist = PointRayPos[1]*PointRayPos[1]+PointRayPos[0]*PointRayPos[0];
      retZ = PointRayPos[2];
   }

   /// returns frustum coordinate system right-std::vector
   inline Vector3f getRayRight() const { return RayRight; };
   /// returns frustum coordinate system up-std::vector
   inline Vector3f getRayUp() const { return RayUp; };
   /// returns frustum coordinate system direction-std::vector
   inline Vector3f getRayDir() const { return RayDir; };
   /// returns frustum coordinate system origin
   inline Vector3f getRayOrigin() const { return RayOrigin; };
   /// returns frustum coordinate system
   inline Matrix4f getRaySystem() const { return RaySystem; };
   /// returns transformed frustum coordinate system, see setTransformation()
   inline Matrix4f getTransRaySystem() const { return TransRaySystem; };

   /// returns squared half frustum width at z-distance 1, see SqrHalfWidthAt1
   inline float32 getSqrHalfWidthAt1() const { return SqrHalfWidthAt1; };
   /// returns squared half frustum height at z-distance 1, see SqrHalfHeightAt1
   inline float32 getSqrHalfHeightAt1() const { return SqrHalfHeightAt1; };
};



/// class for clipped frustum geometric range, defined by a coordinate system with a direction std::vector, an origin, the (squared) half width and half height at ray-z-distance 1.0f defining the horizontal and vertical opening angle and a near and a far clipping plane.
class X3GEOMETRY_API RayClippedPyramidalRange :public RayPyramidalRange{
 private:
   /// z-value of frustum near-clipping plane in frustum coordinate system
   float32 nearClip;
   /// z-value of frustum far-clipping plane in frustum coordinate system
   float32 farClip;

 public:
   RayClippedPyramidalRange(Vector3f right, Vector3f up, Vector3f dir, Vector3f origin, float32 sqrhw, float32 sqrhh, float32 nclip, float32 fclip);
   virtual bool intersects(const BoundingBox3f&)const;
   virtual bool intersects(const Vector3f&)const;

   /// returns z-value of frustum near-clipping plane in frustum coordinate system
   inline float32 getNearClip() const { return nearClip; };
   /// returns z-value of frustum far-clipping plane in frustum coordinate system
   inline float32 getFarClip() const { return farClip; };
};

}
#endif
