//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "HalfEdgeGraph.h"
//---------------------------------------------------------------------------
#include <map>
#include "UnstructuredInCoreTriangleMesh.h"
#include "VertexArray.h"
#include "PointSet.h"
#include "PointSet.inline.h"
#include "PriorityQueue.h"
#include "LinearAlgebra.inline.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

#define HALF_EDGE_DEBUG

HalfEdgeGraph::HalfEdgeGraph(UnstructuredInCoreTriangleMesh * mesh)
{
   m_VertexDescriptor = NULL;
   m_VertexArray = NULL;
	buildGraph( mesh );
}

HalfEdgeGraph::~HalfEdgeGraph(void)
{
	delete m_VertexDescriptor;
   delete m_VertexArray;
}

void HalfEdgeGraph::optimize()
{
   // remove unconnected triangles
   card32 lastValidTriangle = getNumTriangles()-1;
   for( int32 i=lastValidTriangle;i>=0; i++ )
      if( isValidTriangle(i) )
      {
         lastValidTriangle = i;
         break;
      }

   card32 numTriangles = 0;
   for( card32 i=0;(i<getNumTriangles()) && (i<=lastValidTriangle);i++ )
   {
      THalfEdge he1 = m_HalfEdgeTable[i*3+0];

      if( he1.origin == -1 )
      {
         swapTriangles(i, lastValidTriangle );
         do
         {
            x4Assert( lastValidTriangle > 0 );
            lastValidTriangle--;
         }
         while( !isValidTriangle(lastValidTriangle) );
      }

      numTriangles++;
   }
   m_HalfEdgeTable.resize( numTriangles*3 );

   // search for unconnected vertices and remove them
   VertexArray points( &m_VertexTable[0], getNumVertices(), m_VertexDescriptor );
   std::vector<card32> vertexIndexTable( getNumVertices() );
   card32 newIndex = 0;
   for( card32 i=0;i<getNumVertices();i++ )
   {
      if( m_VertexToHalfedgeTable[i] != -1 ) // connected ?
      {
         points.copyVertex( i, newIndex );
         m_VertexToHalfedgeTable[newIndex] = m_VertexToHalfedgeTable[i];
         vertexIndexTable[i] = newIndex;
         newIndex++;
      }
   }

   m_VertexTable.resize( m_VertexDescriptor->getSize() * newIndex );
   m_VertexToHalfedgeTable.resize( newIndex );
   delete m_VertexArray;
   m_VertexArray = new VertexArray( &m_VertexTable[0], newIndex, m_VertexDescriptor );

   for( card32 i=0;i<m_HalfEdgeTable.size();i++ )
      m_HalfEdgeTable[i].origin = vertexIndexTable[ m_HalfEdgeTable[i].origin ];
}

void HalfEdgeGraph::exportToMesh(UnstructuredInCoreTriangleMesh * mesh)
{
   optimize();
   PointSet * points = mesh->getPointSet();

   if( *points->getDescr() == m_VertexDescriptor )
      points->changeHeight( getNumVertices() );
   else
      points->clearAndSetup( 1, getNumVertices(), m_VertexDescriptor );

   PointSet * triangles = mesh->getTriangles();
   x4Assert( triangles->providesAttribute( "index" ) );
   triangles->changeHeight( getNumVertices() );

   VertexArray srcPoints( &m_VertexTable[0], getNumVertices(), m_VertexDescriptor );
   VertexArray dstPoints( points );
   dstPoints.copyVertices( srcPoints );

   VertexArray dstTriangles( triangles );

   for( card32 i=0;i<m_HalfEdgeTable.size() / 3;i++ )
   {
      THalfEdge he1 = m_HalfEdgeTable[i*3+0];
      THalfEdge he2 = m_HalfEdgeTable[i*3+1];
      THalfEdge he3 = m_HalfEdgeTable[i*3+2];

      dstTriangles.setIndex3i( i, makeVector3i( he1.origin, he2.origin, he3.origin ) );
   }

}

void HalfEdgeGraph::swapTriangles( int32 i1, int32 i2 )
{
   for( int32 i=0;i<3;i++ )
   {
      THalfEdge tmp = m_HalfEdgeTable[i1*3+i];
      m_HalfEdgeTable[i1*3+i] = m_HalfEdgeTable[i2*3+i];
      m_HalfEdgeTable[i2*3+i] = tmp;

      // validate VertexToHalfEdgeTable
      if( m_HalfEdgeTable[i1*3+i].origin >= 0 )
         if( m_VertexToHalfedgeTable[ m_HalfEdgeTable[i1*3+i].origin ] == i2*3+i )
            m_VertexToHalfedgeTable[ m_HalfEdgeTable[i1*3+i].origin ] = i1*3+i;
      if( m_HalfEdgeTable[i2*3+i].origin >= 0 )
         if( m_VertexToHalfedgeTable[ m_HalfEdgeTable[i2*3+i].origin ] == i1*3+i )
            m_VertexToHalfedgeTable[ m_HalfEdgeTable[i2*3+i].origin ] = i2*3+i;

      // reconnect next pointer
      m_HalfEdgeTable[i2*3+i].next = m_HalfEdgeTable[i2*3+i].next - i1 + i2;
      m_HalfEdgeTable[i1*3+i].next = m_HalfEdgeTable[i1*3+i].next - i2 + i1;

      // validate inverse pointer
      if( m_HalfEdgeTable[i1*3+i].inverse >= i1*3 &&
         m_HalfEdgeTable[i1*3+i].inverse < i1*3+3 )
         m_HalfEdgeTable[i1*3+i].inverse += i2 - i1;
      if( m_HalfEdgeTable[i2*3+i].inverse >= i2*3 &&
         m_HalfEdgeTable[i2*3+i].inverse < i2*3+3 ) 
         m_HalfEdgeTable[i2*3+i].inverse += i1 - i2;

      // reconnect inverse'inverse pointer
      if( m_HalfEdgeTable[i1*3+i].inverse >= 0 )
         if( m_HalfEdgeTable[ m_HalfEdgeTable[i1*3+i].inverse ].inverse >= 0 )
            m_HalfEdgeTable[ m_HalfEdgeTable[i1*3+i].inverse ].inverse = i1*3+i;
      if( m_HalfEdgeTable[i2*3+i].inverse >= 0 )
         if( m_HalfEdgeTable[ m_HalfEdgeTable[i2*3+i].inverse ].inverse >= 0 )
            m_HalfEdgeTable[ m_HalfEdgeTable[i2*3+i].inverse ].inverse = i2*3+i;

   }
}


/*int32 HalfEdgeGraph::addTriangle( int32 vertex1, int32 vertex2, int32 vertex3 )
{
}

int32 HalfEdgeGraph::addVertex( card8 * vPtr )
{
}

void HalfEdgeGraph::deleteTriangle( int32 triangle )
{
	for( card32 i=0;i<3;i++ )
	{
		int32 he = m_HalfEdgeTable[ triangle * 3 + i].inverse;
		f( he != -1
	}
}

void HalfEdgeGraph::deleteVertex( int32 vertex )
{
}*/

HalfEdgeIterator HalfEdgeGraph::getHalfEdgeIterator( int32 tri_index )
{
	x4Assert( tri_index / 3 < (int32)m_HalfEdgeTable.size() );
	HalfEdgeIterator ret( this, tri_index / 3 );
	return ret;
}

void HalfEdgeGraph::buildGraph( UnstructuredInCoreTriangleMesh * mesh )
{
	delete m_VertexDescriptor;
	m_VertexDescriptor = (VertexDescriptor *) mesh->getDescr()->copy();
	const PointSet * ps = mesh->getPointSet();

	// init vertices
   m_VertexTable.resize( m_VertexDescriptor->getSize() * ps->getNumEntries() );
	m_VertexToHalfedgeTable.resize( ps->getNumEntries() , -1 );
	const card8 * ptr = (const card8*)ps->getDataPointer();
	for( card32 i=0;i<m_VertexTable.size();i++ )
	{
		m_VertexTable[i] = ptr[i];
	}

   delete m_VertexArray;
   m_VertexArray = new VertexArray( &m_VertexTable[0], ps->getNumEntries(), m_VertexDescriptor );

	VertexArray points( &m_VertexTable[0], ps->getNumEntries(), m_VertexDescriptor  );
	VertexArray triangles( mesh->getTriangles() );

	// put every triangle into HE table
	m_HalfEdgeTable.clear();
   std::map<int64,int32> edgeMap;
	for( card32 i=0;i<triangles.getNumElements();i++ )
	{
		Vector3i tri = triangles.getIndex3i(i);

      size_t tri_start = m_HalfEdgeTable.size();

      for( card32 j=0;j<3;j++ )
      {
         int32 v1 = tri[j];
         int32 v2 = tri[(j+1)%3];

         THalfEdge edge;
         edge.origin = v1;
         mpcard edgeIndex = tri_start + j;
         edge.next = tri_start + ((j+1) % 3);
         edge.inverse = -1;

         int64 edgeID = (v2 > v1) ? v1 * ps->getNumEntries() + v2 :
            v2 * ps->getNumEntries() + v1;
         if( edgeMap.find( edgeID ) != edgeMap.end() )
         {
            edge.inverse = edgeMap[ edgeID ];
            m_HalfEdgeTable[ edge.inverse ].inverse = edgeIndex;
         }
         else
         {
            edgeMap[ edgeID ] = edgeIndex;
         }

         if( m_VertexToHalfedgeTable[v1] < 0 )
         {
             m_VertexToHalfedgeTable[v1] = edgeIndex;
         }

         m_HalfEdgeTable.push_back( edge );

      }
	}
	
	checkIntegrity();
	m_IsOptimized = true;

}

bool HalfEdgeGraph::checkHalfEdgeCollapse( int32 he )
{
   x4Assert( he < (int32)m_HalfEdgeTable.size() );
   x4Assert( he >= 0 );
   x4Assert( isValidHalfEdge( he ) );

   if( m_HalfEdgeTable[ he ].inverse >= 0 )
   {
      // case: no border
      HalfEdgeIterator iter(this,he);
      if( !iter.hasNextTriangle() )
         return false;

      iter.gotoNext();
      if( !iter.hasPreviousTriangle() )
         return false;


      iter = HalfEdgeIterator( this,he );
      iter.gotoInverse();

      int32 tmpHE = iter.getHalfEdge();
      if( !iter.hasNextTriangle() )
         return false;

      bool border_vertex = true;
      do
      {
         iter.gotoNextTriangle();
         if( iter.getHalfEdge() == tmpHE )
         {
            border_vertex = false;
            break;
         }

      } while( iter.hasNextTriangle() );

      if( !border_vertex )
         return false;
   }
   else
   {
       // border case
       HalfEdgeIterator iter( this, he );
       if( !iter.hasNextTriangle() )
         return false;
       iter.gotoNext();
       if( !iter.hasPreviousTriangle() )
         return false;
   }

   return true;
}


void HalfEdgeGraph::collapseHalfEdge( int32 he )
{
   x4Assert( he < (int32)m_HalfEdgeTable.size() );
   x4Assert( he >= 0 );
   x4Assert( m_HalfEdgeTable[he].origin >= 0 );
   x4Assert( checkHalfEdgeCollapse( he)  );

   int32 he2 = m_HalfEdgeTable[he].next;
   int32 he3 = m_HalfEdgeTable[he2].next;

   int32 newOrigin = m_HalfEdgeTable[m_HalfEdgeTable[he].next ].origin;
   x4Assert( newOrigin >= 0 );

   std::vector<int32> oldOriginEdges = getVertexHalfEdges( m_HalfEdgeTable[he].origin );
   for( card32 i=0;i<oldOriginEdges.size();i++ )
      m_HalfEdgeTable[ oldOriginEdges[i] ].origin = newOrigin;

   // mark vertex as deleted
   m_VertexToHalfedgeTable[ m_HalfEdgeTable[he].origin ] = -1;


   // connect neighbour triangles
   m_HalfEdgeTable[ m_HalfEdgeTable[he2].inverse ].inverse = m_HalfEdgeTable[he3].inverse;
   m_HalfEdgeTable[ m_HalfEdgeTable[he3].inverse ].inverse = m_HalfEdgeTable[he2].inverse;

   // update newOrigin's vertextohalfedge-entry
   m_VertexToHalfedgeTable[ newOrigin ] = m_HalfEdgeTable[he3].inverse;
   x4Assert( newOrigin == m_HalfEdgeTable[m_HalfEdgeTable[he3].inverse].origin );

   // update he3's origin v-to-he-entry
   m_VertexToHalfedgeTable[ m_HalfEdgeTable[he3].origin ] = m_HalfEdgeTable[he2].inverse;
   x4Assert( m_HalfEdgeTable[ m_HalfEdgeTable[he2].inverse].origin == m_HalfEdgeTable[he3].origin );


   HalfEdgeIterator iter( this, he );
   if( iter.hasInverse() )
   {
      iter.gotoInverse();
      int32 inv_he = iter.getHalfEdge();
      iter.gotoNext();
      int32 inv_he2 = iter.getHalfEdge();
      iter.gotoNext();
      int32 inv_he3 = iter.getHalfEdge();

      // connect neighbour triangles
      m_HalfEdgeTable[ m_HalfEdgeTable[inv_he2].inverse ].inverse = m_HalfEdgeTable[inv_he3].inverse;
      m_HalfEdgeTable[ m_HalfEdgeTable[inv_he3].inverse ].inverse = m_HalfEdgeTable[inv_he2].inverse;

      // update he_inv3's origin v-to-he-entry
      m_VertexToHalfedgeTable[ m_HalfEdgeTable[inv_he3].origin ] = m_HalfEdgeTable[inv_he2].inverse;
      x4Assert( m_HalfEdgeTable[ m_HalfEdgeTable[inv_he2].inverse].origin == m_HalfEdgeTable[inv_he3].origin );

      x4Assert( m_VertexToHalfedgeTable[m_HalfEdgeTable[inv_he].origin] != inv_he );
      m_HalfEdgeTable[ inv_he ].erase();
      x4Assert( m_VertexToHalfedgeTable[m_HalfEdgeTable[inv_he2].origin] != inv_he2 );
      m_HalfEdgeTable[ inv_he2 ].erase();
      x4Assert( m_VertexToHalfedgeTable[m_HalfEdgeTable[inv_he3].origin] != inv_he3 );
      m_HalfEdgeTable[ inv_he3 ].erase();
   }
   else
/*   {
      // update edge-origins
      for( HalfEdgeIterator iter(this,he); iter.hasNextTriangle(); )
      {
         iter.gotoNextTriangle();
         iter.setOrigin( newOrigin );
      }
   }*/

   x4Assert( m_VertexToHalfedgeTable[m_HalfEdgeTable[he].origin] != he );
   m_HalfEdgeTable[ he ].erase();
   x4Assert( m_VertexToHalfedgeTable[m_HalfEdgeTable[he].origin] != he2 );
   m_HalfEdgeTable[ he2 ].erase();
   x4Assert( m_VertexToHalfedgeTable[m_HalfEdgeTable[he].origin] != he3 );
   m_HalfEdgeTable[ he3 ].erase();

}


std::vector<int32> HalfEdgeGraph::getVertexHalfEdges( int32 vertex )
{
   x4Assert( vertex < (int32)m_VertexToHalfedgeTable.size() );
   std::vector<int32> ret;
   int32 start_halfedge = m_VertexToHalfedgeTable[ vertex ];
   if( start_halfedge < 0 )
      return ret;

   bool border = true;
   for( HalfEdgeIterator iter = HalfEdgeIterator( this, start_halfedge );
      iter.hasNextTriangle();  )
   {
      ret.push_back( iter.getHalfEdge() );
      x4Assert( ret.size() < 100 );
      iter.gotoNextTriangle();
      if( iter.getHalfEdge() == start_halfedge )
      {
         border = false;
         break;
      }
   }

   if( border )
      for( HalfEdgeIterator iter = HalfEdgeIterator( this, start_halfedge );
         iter.hasPreviousTriangle();  )
      {
         iter.gotoPreviousTriangle();
         ret.push_back( iter.getHalfEdge() );
         x4Assert( ret.size() < 100 );
      }

   return ret;
}

HalfEdgeIterator HalfEdgeGraph::getVertexHalfEdgeIterator( int32 vertex )
{
   x4Assert( vertex < (int32)m_VertexToHalfedgeTable.size() );
   x4Assert( m_VertexToHalfedgeTable[ vertex ] >= 0 );
   return HalfEdgeIterator( this, m_VertexToHalfedgeTable[ vertex ] );
}

struct THE_Simpl
{
   bool operator < ( const THE_Simpl &s ) const
   {
      return he < s.he; // halfedge is unique
   }
	bool operator == ( const THE_Simpl &s) const
	{
		return he == s.he;
	}
   int32 he;
   float cost;
};


void HalfEdgeGraph::simplify()
{
   PriorityQueue<float,THE_Simpl> queue;
   std::vector<THE_Simpl> table;

   // calculate quadrics
   Quadric quad0;
   quad0.A = makeMatrix3f(0,0,0, 0,0,0, 0,0,0);
   quad0.b = makeVector3f(0,0,0);
   quad0.c = 0;
   std::vector<Quadric> quadrics( getNumVertices(), quad0 );
   for( card32 i=0;i<getNumTriangles();i++ )
   {
      if( !isValidTriangle(i) )
         continue;

      HalfEdgeIterator iter = getHalfEdgeIterator(i);
      Vector3f x1 = iter.getOriginPosition3f();
      int32 ind1 = iter.getOrigin();
      iter.gotoNext();
      Vector3f x2 = iter.getOriginPosition3f();
      int32 ind2 = iter.getOrigin();
      iter.gotoNext();
      Vector3f x3 = iter.getOriginPosition3f();
      int32 ind3 = iter.getOrigin();

      Quadric q;
      q.calculate( x1,x2,x3 );

      quadrics[ ind1 ] = quadrics[ ind1 ] + q;
      quadrics[ ind2 ] = quadrics[ ind2 ] + q;
      quadrics[ ind3 ] = quadrics[ ind3 ] + q;

   }

   for( card32 i=0;i<m_HalfEdgeTable.size();i++ )
   {
      THE_Simpl element;
      element.he = i;
      Quadric q = quadrics[ m_HalfEdgeTable[i].origin ];
      q = q + quadrics[ m_HalfEdgeTable[m_HalfEdgeTable[i].next].origin ];
      HalfEdgeIterator iter(this,i);
      iter.gotoNext();
      if( checkHalfEdgeCollapse(i) )
         element.cost = q.eval( iter.getOriginPosition3f() );
      else
         element.cost = 1e20f;
      queue.insert( element.cost, element );
      table.push_back( element );
   }

   for( card32 i=0;i<getNumVertices()/3;i++ )
   {
      THE_Simpl element = queue.pop_first();
      if( element.cost > 1e19 )
         break;

//      checkIntegrity();

      // collect edges that will be (geometrical) modified due edge collapse

      std::vector<int32> removedEdges;
      //removedEdges.push_back( element.he );
      removedEdges.push_back( getNext(element.he) );
      removedEdges.push_back( getPrevious(element.he) );
      if( getInverse( element.he ) >= 0 )
      {
         int32 tmpHE = getInverse( element.he );
         removedEdges.push_back( tmpHE );
         removedEdges.push_back( getNext(tmpHE) );
         removedEdges.push_back( getPrevious(tmpHE) );
      }

/*      std::vector<int32> modifiedEdges;
      modifiedEdges.push_back( element.he );
      HalfEdgeIterator iter1(this,element.he);
      iter1.gotoNext();
      modifiedEdges.push_back( iter1.getHalfEdge() );
      iter2.
      bool border = true;
      x4Assert( isValidHalfEdge(element.he) );
      for( HalfEdgeIterator iter( this, element.he ); iter.hasNextTriangle(); )
      {
         iter.gotoNextTriangle();
         x4Assert( iter.getOrigin() >= 0 );
         if( iter.getHalfEdge() == element.he )
         {
            border = false;   // no border at this point
            break;
         }
         modifiedEdges.push_back( iter.getHalfEdge() );
         HalfEdgeIterator iter1 = iter;
         iter1.gotoNext();
         modifiedEdges.push_back( iter1.getHalfEdge() );

         x4Assert( modifiedEdges.size() < 100 );

         if( iter1.hasInverse() )
         {
            iter1.gotoInverse();
            modifiedEdges.push_back( iter1.getHalfEdge() );
         }


      }
      if( border )
      for( HalfEdgeIterator iter( this, element.he ); iter.hasPreviousTriangle(); )
      {
         iter.gotoPreviousTriangle();
         modifiedEdges.push_back( iter.getHalfEdge() );
         HalfEdgeIterator iter1 = iter;
         iter1.gotoNext();
         modifiedEdges.push_back( iter1.getHalfEdge() );
         if( iter1.hasInverse() )
         {
            iter1.gotoInverse();
            modifiedEdges.push_back( iter1.getHalfEdge() );
         }

      }*/
//      std::vector<int32> modifiedEdges = getVertexHalfEdges( m_HalfEdgeTable[ element.he ].origin );

      // save new origin
      int32 newOrigin = m_HalfEdgeTable[ getNext( element.he ) ].origin;

      Quadric newQuadric = quadrics[ m_HalfEdgeTable[ element.he ].origin ]
         + quadrics[ m_HalfEdgeTable[ m_HalfEdgeTable[ element.he ].next ].origin ];

      quadrics[ m_HalfEdgeTable[ element.he ].origin ] = newQuadric;

      // collapse
      collapseHalfEdge( element.he );

      checkIntegrity();

      // remove invalid halfedges
      for( card32 i=0;i<removedEdges.size();i++ )
      {
         THE_Simpl element = table[ removedEdges[i] ];
         queue.remove( element.cost, element );
      }

      // update collapse costs
      std::vector<int32> modifiedEdges = getVertexHalfEdges(newOrigin);
      for( card32 i=0;i<modifiedEdges.size();i++ )
      {
         x4Assert( isValidHalfEdge( modifiedEdges[i] ) );

         THE_Simpl element = table[ modifiedEdges[i] ];
         queue.remove( element.cost, element );
//         if( isValidHalfEdge( modifiedEdges[i] ) )
         {
            Quadric q = quadrics[ m_HalfEdgeTable[element.he].origin ];
            q = q + quadrics[ m_HalfEdgeTable[m_HalfEdgeTable[element.he].next].origin ];
            HalfEdgeIterator iter(this,element.he);
            iter.gotoNext();
            if( checkHalfEdgeCollapse( modifiedEdges[i] ) )
               element.cost = q.eval( iter.getOriginPosition3f() );
            else
               element.cost = 1e20f;

            table[ modifiedEdges[i] ] = element;
            queue.insert( element.cost, element );
         }
      }

   }

}

Vector3f HalfEdgeGraph::calculateFaceNormal( int32 triangle )
{
   HalfEdgeIterator iter = getHalfEdgeIterator( triangle );
   Vector3f edge1 = -iter.getHalfEdge3f();
   iter.gotoNext();
   Vector3f edge2 = iter.getHalfEdge3f();
   Vector3f normal = edge1.crossProduct( edge2 );
   normal.normalize();
   return normal;
}

Vector3f HalfEdgeGraph::calculateBaryCenter( int32 triangle )
{
   HalfEdgeIterator iter = getHalfEdgeIterator( triangle );
   Vector3f ret = iter.getOriginPosition3f();
   iter.gotoNext();
   ret += iter.getOriginPosition3f();
   iter.gotoNext();
   ret += iter.getOriginPosition3f();
   ret /= 3.0f;
   return ret;
}

void HalfEdgeGraph::caculateTangents( int32 triangle, Vector3f &tangent_u, Vector3f &tangent_v )
{
   HalfEdgeIterator iter = getHalfEdgeIterator( triangle );
   tangent_u = iter.getHalfEdge3f();
   tangent_u.normalize();
   Vector3f normal = calculateFaceNormal( triangle );
   tangent_v = tangent_u.crossProduct( normal );
   tangent_u.normalize();
}



bool HalfEdgeGraph::isValidTriangle( int32 index )
{
   return isValidHalfEdge( index * 3 );
}

bool HalfEdgeGraph::isValidVertex( int32 vertex )
{
   x4Assert( vertex < (int32)m_VertexToHalfedgeTable.size() );
   return m_VertexToHalfedgeTable[vertex] >= 0;
}

bool HalfEdgeGraph::isValidHalfEdge( int32 he )
{
   x4Assert( he < (int32)m_HalfEdgeTable.size() );
   return m_HalfEdgeTable[he].origin >= 0;
}

void HalfEdgeGraph::checkIntegrity()
{
   // check halfedge-inverse and triangle
   for( card32 i=0;i<m_HalfEdgeTable.size();i++ )
   {
      if( !isValidHalfEdge(i) )
         continue;

      if( m_HalfEdgeTable[i].inverse >= 0 )
      {
         x4Assert( m_HalfEdgeTable[ m_HalfEdgeTable[i].inverse ].inverse == (int32)i );
         x4Assert( isValidHalfEdge( m_HalfEdgeTable[i].inverse ) );
      }

      int32 he2 = m_HalfEdgeTable[i].next;
      int32 he3 = m_HalfEdgeTable[he2].next;
      x4Assert( isValidHalfEdge(he2) );
      x4Assert( isValidHalfEdge(he3) );

      x4Assert( m_HalfEdgeTable[he3].next == (int32)i );
   }

   for( card32 i=0;i<m_VertexToHalfedgeTable.size();i++ )
   {
      if( m_VertexToHalfedgeTable[i] >= 0 )
      {
         x4Assert( m_VertexToHalfedgeTable[i] < (int32)m_HalfEdgeTable.size() );
         if( m_HalfEdgeTable[m_VertexToHalfedgeTable[i]].origin != (int32)i )
            throw "narf";
         x4Assert( m_HalfEdgeTable[m_VertexToHalfedgeTable[i]].origin == (int32)i );
         card32 numVertHE = 0;
         HalfEdgeIterator iter = HalfEdgeIterator(this, m_VertexToHalfedgeTable[i] );
         if( iter.hasNextTriangle() )
         {
            do
            {
               iter.gotoNextTriangle();
               numVertHE++;
               if( iter.getHalfEdge() == m_VertexToHalfedgeTable[i] )
                  break;
               x4Assert( numVertHE <= 100 );
            }
            while( iter.hasNextTriangle() );
         }

      }


   }
}


// ---------------------------------------------------------


HalfEdgeIterator::HalfEdgeIterator( HalfEdgeGraph * graph, int32 halfedge )
{
	m_Graph = graph;
	m_HalfEdge = halfedge;
}

Vector3f HalfEdgeIterator::getOriginPosition3f()
{
#ifdef HALF_EDGE_DEBUG
	x4Assert( m_HalfEdge < (int32)m_Graph->m_HalfEdgeTable.size() );
	x4Assert( m_Graph->m_HalfEdgeTable[ m_HalfEdge ].origin >= 0 );
	x4Assert( m_Graph->m_HalfEdgeTable[ m_HalfEdge ].origin < (int32)m_Graph->m_VertexArray->getNumElements() );
#endif
	return m_Graph->m_VertexArray->getPosition3f( m_Graph->m_HalfEdgeTable[ m_HalfEdge ].origin );
}

void HalfEdgeIterator::setOriginPosition3f( const Vector3f &vec )
{
#ifdef HALF_EDGE_DEBUG
	x4Assert( m_HalfEdge < (int32)m_Graph->m_HalfEdgeTable.size() );
	x4Assert( m_Graph->m_HalfEdgeTable[ m_HalfEdge ].origin >= 0 );
	x4Assert( m_Graph->m_HalfEdgeTable[ m_HalfEdge ].origin < (int32)m_Graph->m_VertexArray->getNumElements() );
#endif
	m_Graph->m_VertexArray->setPosition3f( m_Graph->m_HalfEdgeTable[ m_HalfEdge ].origin , vec);
}

Vector3f HalfEdgeIterator::getOriginColor3f()
{
#ifdef HALF_EDGE_DEBUG
	x4Assert( m_HalfEdge < (int32)m_Graph->m_HalfEdgeTable.size() );
	x4Assert( m_Graph->m_HalfEdgeTable[ m_HalfEdge ].origin >= 0 );
	x4Assert( m_Graph->m_HalfEdgeTable[ m_HalfEdge ].origin < (int32)m_Graph->m_VertexArray->getNumElements() );
#endif
	return m_Graph->m_VertexArray->getColor3f( m_Graph->m_HalfEdgeTable[ m_HalfEdge ].origin );
}

void HalfEdgeIterator::setOriginColor3f( const Vector3f &vec )
{
#ifdef HALF_EDGE_DEBUG
	x4Assert( m_HalfEdge < (int32)m_Graph->m_HalfEdgeTable.size() );
	x4Assert( m_Graph->m_HalfEdgeTable[ m_HalfEdge ].origin >= 0 );
	x4Assert( m_Graph->m_HalfEdgeTable[ m_HalfEdge ].origin < (int32)m_Graph->m_VertexArray->getNumElements() );
#endif
	m_Graph->m_VertexArray->setColor3f( m_Graph->m_HalfEdgeTable[ m_HalfEdge ].origin , vec);
}

Vector3f HalfEdgeIterator::getOriginNormal3f()
{
#ifdef HALF_EDGE_DEBUG
	x4Assert( m_HalfEdge < (int32)m_Graph->m_HalfEdgeTable.size() );
	x4Assert( m_Graph->m_HalfEdgeTable[ m_HalfEdge ].origin >= 0 );
	x4Assert( m_Graph->m_HalfEdgeTable[ m_HalfEdge ].origin < (int32)m_Graph->m_VertexArray->getNumElements() );
#endif
	return m_Graph->m_VertexArray->getNormal3f( m_Graph->m_HalfEdgeTable[ m_HalfEdge ].origin );
}

void HalfEdgeIterator::setOriginNormal3f( const Vector3f &vec )
{
#ifdef HALF_EDGE_DEBUG
	x4Assert( m_HalfEdge < (int32)m_Graph->m_HalfEdgeTable.size() );
	x4Assert( m_Graph->m_HalfEdgeTable[ m_HalfEdge ].origin >= 0 );
	x4Assert( m_Graph->m_HalfEdgeTable[ m_HalfEdge ].origin < (int32)m_Graph->m_VertexArray->getNumElements() );
#endif
	m_Graph->m_VertexArray->setNormal3f( m_Graph->m_HalfEdgeTable[ m_HalfEdge ].origin , vec);
}

Vector3f HalfEdgeIterator::getHalfEdge3f()
{                                
	int32 nextEdge = m_Graph->m_HalfEdgeTable[ m_HalfEdge ].next;
	return m_Graph->m_VertexArray->getPosition3f( m_Graph->m_HalfEdgeTable[ nextEdge ].origin)
		- m_Graph->m_VertexArray->getPosition3f( m_Graph->m_HalfEdgeTable[ m_HalfEdge ].origin);
}

void Quadric::calculate( const Vector3f &x1,const Vector3f &x2,const Vector3f &x3 )
{
   // compute tangents
   Vector3f tangent_u = x2-x1;
   tangent_u.normalize();
   Vector3f normal = (x2-x1).crossProduct(x3-x1);
   float area = norm( normal );
   Vector3f tangent_v = tangent_u.crossProduct( normal );
   tangent_u.normalize();

   // compute A
   Matrix3f U,V;
   for( card32 i=0;i<3;i++ )
      for( card32 j=0;j<3;j++ )
      {
         U[i][j] = tangent_u[i] * tangent_u[j];
         V[i][j] = tangent_v[i] * tangent_v[j];
      }

   A = IDENTITY3F - U - V;
   A *= area;

   // barycenter p
   Vector3f p = (x1+x2+x3)/3.0f;

   // compute b
   b = -(A*p);

   // compute tmp = p^T * A, then tmp * p
   Vector3f tmp;
   for( card32 i=0;i<3;i++ )
   {
      tmp[i] = p * A[i];
   }
   c = tmp * p;

}

float Quadric::eval( const Vector3f &v )
{
   Vector3f tmp;
   for( card32 i=0;i<3;i++ )
      tmp[i] = v * A[i];
   return tmp * v + b*v*2.0f + c;
}
