//---------------------------------------------------------------------------
#ifndef HalfEdgeGraphH
#define HalfEdgeGraphH
//---------------------------------------------------------------------------
#include "x3geometry.h"
//---------------------------------------------------------------------------
#include "LinearAlgebra.h"
//---------------------------------------------------------------------------

namespace X4
{

class HalfEdgeIterator;
class VertexArray;
class UnstructuredInCoreTriangleMesh;
class VertexDescriptor;

class X3GEOMETRY_API Quadric
{
public:
   Matrix3f A;
   Vector3f b;
   float    c;

   Quadric operator + ( const Quadric &q )
   {
      Quadric ret;
      ret.A = A + q.A;
      ret.b = b + q.b;
      ret.c = c + q.c;
      return ret;
   }

   Quadric operator * ( float f )
   {
      Quadric ret = *this;
      ret.A *= f;
      ret.b *= f;
      ret.c *= f;
      return ret;
   }

   void calculate( const Vector3f &x1,const Vector3f &x2,const Vector3f &x3 );

   float eval( const Vector3f &v );

};


/**
	halfedge graph for triangle meshes.
*/
class X3GEOMETRY_API HalfEdgeGraph
{
private:

	struct THalfEdge
	{
      void erase() {origin = next = inverse = -1;}
		mpint	next;
		int32	inverse;
		int32	origin;	// = vertex index
	};

	VertexDescriptor	*	m_VertexDescriptor;
	std::vector<card8>			m_VertexTable;
	std::vector<int32>			m_VertexToHalfedgeTable;
   std::vector<THalfEdge>	   m_HalfEdgeTable;
	VertexArray			*	m_VertexArray;
	bool					   m_IsOptimized;

public:
	HalfEdgeGraph( UnstructuredInCoreTriangleMesh * mesh );
	~HalfEdgeGraph(void);

	/// export to mesh
	void exportToMesh( UnstructuredInCoreTriangleMesh * mesh );

   mpcard getNumVertices() {return m_VertexToHalfedgeTable.size();}
   mpcard getNumTriangles() {return m_HalfEdgeTable.size() / 3;}

/*   ///
	void disconnectVertex( int32 vertex );

	/// deletes vertex and all neighbour triangles
	void deleteVertex( int32 vertex );

	/// deletes triangle and all leaving unconnected vertices
	void deleteTriangle( int32 triangle );

	/// add vertex, returns index
	int32 addVertex( card8 * vPtr );

	/// add triangle, returns triangle-index
	int32 addTriangle( int32 vertex1, int32 vertex2, int32 vertex3 );

*/

	/// returns halfedgeiterator from first halfedge of triangle[tri_index]
	HalfEdgeIterator getHalfEdgeIterator( int32 tri_index );

   /// returns halfedgeiterator from halfedge connected to vertex
   HalfEdgeIterator getVertexHalfEdgeIterator( int32 vertex );

// ---------------- geometric ----------------------------------
   /// calculates normal (normalized)
   Vector3f calculateFaceNormal( int32 triangle );

   /// calculates barycenter
   Vector3f calculateBaryCenter( int32 triangle );

   /// calculates quadric
   Quadric calculateQuadric( int32 triangle );

   /// calculates normalized tangent vectors from
   void caculateTangents( int32 triangle, Vector3f &tangent_u, Vector3f &tangent_v );

   void simplify();

private:

   int32 getNext( int32 he )
      {x4Assert( he < (int32)m_HalfEdgeTable.size() ); return m_HalfEdgeTable[he].next;}
   int32 getPrevious(int32 he)
      {return getNext( getNext( he ) );}
   int32 getInverse( int32 he )
      {x4Assert( he < (int32)m_HalfEdgeTable.size() ); return m_HalfEdgeTable[he].inverse;}

   /// check the whole structure
   void checkIntegrity();

   /// collect all halfedges with origin vertex
   std::vector<int32> getVertexHalfEdges( int32 vertex );

   /// return true if halfedge can be collapsed
   bool checkHalfEdgeCollapse( int32 he );

   /// collapses halfedge
   void collapseHalfEdge( int32 he);

	/// build halfedge graph
	void buildGraph( UnstructuredInCoreTriangleMesh * mesh );

	/// optimize memory usage, removes every unconnected vertex
	/// every HalfEdgeIterator or stored index-value is INVALID after optimize()!
	void optimize();

   bool isValidTriangle( int32 index );

	bool isValidVertex( int32 vertex );

   bool isValidHalfEdge( int32 he );

   /// swaps triangle i1 & i2
   void swapTriangles( int32 i1, int32 i2 );

friend HalfEdgeIterator;

};

class X3GEOMETRY_API HalfEdgeIterator
{
private:
	HalfEdgeGraph *   m_Graph;
	int32			      m_HalfEdge;
public:

	HalfEdgeIterator(HalfEdgeGraph * graph, int32 halfedge);
	HalfEdgeIterator( const HalfEdgeIterator &other )
   {
		m_Graph = other.m_Graph;
		m_HalfEdge = other.m_HalfEdge;
   }

	inline void operator = ( const HalfEdgeIterator &other )
	{
		m_Graph = other.m_Graph;
		m_HalfEdge = other.m_HalfEdge;
	}

	inline bool operator == ( const HalfEdgeIterator &other )
	{
		return (m_Graph == other.m_Graph) && (m_HalfEdge == other.m_HalfEdge);
	}

	inline bool operator != ( const HalfEdgeIterator &other )
	{
		return (m_Graph != other.m_Graph) || (m_HalfEdge != other.m_HalfEdge);
	}

// -------------------------------------------------------------------
// --------------- half edge traversal -------------------------------

	/// goto next halfedge
	inline void gotoNext()
		{m_HalfEdge = m_Graph->m_HalfEdgeTable[m_HalfEdge].next;}

	/// goto previous halfedge
	inline void gotoPrevious()
		{ gotoNext();gotoNext(); }

   /// goto halfedge in next triangle ( also connected to origin )
   /// is equal to: gotoPrevious(); gotoInverse();
   inline void gotoNextTriangle()
   {
      gotoPrevious();
      x4Assert( hasInverse() );
      gotoInverse();
   }

   /// return true if next triangle exists
   inline bool hasNextTriangle()
   {
      return m_Graph->m_HalfEdgeTable[
         m_Graph->m_HalfEdgeTable[m_Graph->m_HalfEdgeTable[m_HalfEdge].next].next ].inverse != -1;
   }

   /// goto halfedge in previous triangle ( also connected to origin )
   /// is equal to: gotoInverse(); gotoNext();
   inline void gotoPreviousTriangle()
   {
      x4Assert( hasInverse() );
      gotoInverse();
      gotoNext();
   }

   /// return true if previous triangle exists
   inline bool hasPreviousTriangle()
   {
      return hasInverse();
   }

	/// goto inverse halfedge ( throws exception if inverse doesn't exist )
	inline void gotoInverse()
		{x4Assert( hasInverse() ); m_HalfEdge = m_Graph->m_HalfEdgeTable[m_HalfEdge].inverse;}

	/// returns true if inverse halfedge exists
	inline bool hasInverse()
		{return m_Graph->m_HalfEdgeTable[m_HalfEdge].inverse != -1;}

	/// returns triangle of this halfedge
	inline int32 getTriangle() {return m_HalfEdge / 3;}


   inline void setOrigin( int32 origin )
   {
      m_Graph->m_HalfEdgeTable[m_HalfEdge].origin = origin;
   }

   inline int32 getOrigin()
   {
      return m_Graph->m_HalfEdgeTable[m_HalfEdge].origin;
   }

   inline void setInverse( int32 inverse )
   {
      m_Graph->m_HalfEdgeTable[m_HalfEdge].inverse = inverse;
   }

   inline int32 getInverse()
   {
      return m_Graph->m_HalfEdgeTable[m_HalfEdge].inverse;
   }

   inline int32 getHalfEdge()
   {
      return m_HalfEdge;
   }

// -------------------------------------------------------------------
// --------------- vertex interface -- -------------------------------

	// fast access for lazy programmers ;-)
	Vector3f getOriginPosition3f();
	Vector3f getOriginNormal3f();
	Vector3f getOriginColor3f();

	void setOriginPosition3f( const Vector3f &vec );
	void setOriginColor3f( const Vector3f &vec );
	void setOriginNormal3f( const Vector3f &vec );

	// returns std::vector from origin to next-origin
	Vector3f getHalfEdge3f();

};

}

#endif