//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "HierarchicalKNNIterator.h"
//---------------------------------------------------------------------------
#include "PointSet.h"
#include "PointSet.inline.h"
#include "DebugOutput.h"
//#include "EventCounterDisplay.h"
#include "LinearAlgebra.h"
#include "LinearAlgebra.inline.h"
#include "BoundingBox.inline.h"
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------


HierarchicalKNNIterator::HierarchicalKNNIterator(PointCloud* pc,  card32 subTreeSize , MemoryManager *attachedOctreeMemManager):
  PointCloudIterator(),
  BasicPointCloudIterator(),
  m_pc(pc),
  m_ps(NULL),
  m_buffer(NULL),
  m_attachedOctreeMemManager(attachedOctreeMemManager),
  m_subTreeSize(subTreeSize),
  modifyable(false)
{
   knn = NULL;
   m_seekPoint= makeVector3f(0,0,0);
   native = m_pc->createIterator(SOIT(SOIT::CAP_HIERARCHICAL));


   m_hr = dynamic_cast<HierarchicalIterator*>(native);
   nextPointReady=false;

   if ((m_hr->getNumChildren()>0) && !m_hr->isLeaf())  //m_hr->getNumChildren()==0 means no hr
   {
      hierachyMode=true; // put root box on Queue
      putBoxOnQueue(dynamic_cast<HierarchicalIterator*> (  dynamic_cast<SceneObjectIterator*>(m_hr)->copy()));
   }else{
      hierachyMode=false;
      knn = new KNNQuery(native, subTreeSize, m_attachedOctreeMemManager, modifyable);
   }

   cacheHit=0;
   cacheMiss=0;
   noChaching=true;

   cacheBox=BoundingBox3f(makeVector3f(0,0,0));
   getNewCacheBox=true;
	m_maxDistanceToSeekPoint_2 = 1e35f;
}

HierarchicalKNNIterator::~HierarchicalKNNIterator(){

   if(hierachyMode) clearQueues();
   if(!hierachyMode) delete  knn;
   delete (native);
}


void HierarchicalKNNIterator::setSeekPointAndReset(const Vector3f& sp){
   m_seekPoint=sp;
   if(hierachyMode){

      bool showPQs =false;
      //noChaching=true;

      bool newPointinOldCacheBox= cacheBox.inBoundingBox(m_seekPoint);
      if(noChaching|| !newPointinOldCacheBox){


         clearQueues();
         putBoxOnQueue(dynamic_cast<HierarchicalIterator*> (  dynamic_cast<SceneObjectIterator*>(m_hr)->copy()));
         noChaching=false;
         getNewCacheBox=true;
         cacheMiss++;

      }else{
         cacheRefresh();
         cacheHit++;
         showPQs=true;
      }

   #ifdef _DEBUG //display cache effency
/*         EventCounterDisplay::NewValueStepOnChange e;
         e.value = cacheMiss;
         debugOutput.displayEvent(EventCounterDisplay::getClass(), std::string("PCA: Chache MISS"), &e);

         e.value = cacheHit;
         if (cacheHit%1000==0)debugOutput.displayEvent(EventCounterDisplay::getClass(), std::string("PCA: Chache HIT"), &e);
         if (showPQs){
            e.value = pqBoxes.size();
            debugOutput.displayEvent(EventCounterDisplay::getClass(), std::string("PCA: Cached Boxes"), &e);
            e.value = pqKNN.size();
            debugOutput.displayEvent(EventCounterDisplay::getClass(), std::string("PCA: Cached kNNs"), &e);
         }*/
   #endif


      seekNextPoint();
   }else{
      knn->setSeekPoint(sp);
      nextPointReady = knn->seekNextPoint();
      if (nextPointReady)
         m_buffer = knn->getDataPointer();
   }

}

void HierarchicalKNNIterator::putAIkNNOnQueue(SceneObjectIterator* box){

   KNNQuery* boxQuery = new KNNQuery(box, m_seekPoint, m_subTreeSize, m_attachedOctreeMemManager, modifyable);

   if (boxQuery->seekNextPoint()){ //if at least one point in box

      Vector3f nearest= boxQuery->getNearestPointPosition();
      float32 qDistance = (nearest-m_seekPoint)*(nearest-m_seekPoint);
      KNNEntry pnt(boxQuery , qDistance, box);
      pqKNN.push(pnt);

   }else{
      delete boxQuery;
      delete box;
   }
}

void HierarchicalKNNIterator::putOldQueryOnQueue(KNNQuery* q, SceneObjectIterator* box){
      Vector3f nearest= q->getNearestPointPosition();
      float32 qDistance = (nearest-m_seekPoint)*(nearest-m_seekPoint);
      KNNEntry pnt(q , qDistance, box);
      pqKNN.push(pnt);
}


void HierarchicalKNNIterator::putBoxOnQueue(HierarchicalIterator* box){
   // calc box-point distance
   float32 qDistance;

   Sphere3f sp=box->getBoundingBox().getBoundingSphere();
   qDistance=(sp.center-m_seekPoint)*(sp.center-m_seekPoint);
   if(qDistance<sp.radius*sp.radius){
      qDistance= -1.0f;
   }else{
      qDistance=sqrt(qDistance);
      qDistance-=sp.radius;
      qDistance*=qDistance;
   }

   if( qDistance > m_maxDistanceToSeekPoint_2)
	   return;

   MHEntry bEntry(dynamic_cast<SceneObjectIterator*>(box), qDistance);
   pqBoxes.push(bEntry);

}


/// go to first point
void HierarchicalKNNIterator::reset(){
   if(hierachyMode){


      clearQueues();
      putBoxOnQueue(dynamic_cast<HierarchicalIterator*> (  dynamic_cast<SceneObjectIterator*>(m_hr)->copy()));
      noChaching=false;
      getNewCacheBox=true;

      seekNextPoint();
   }else{
      knn->setSeekPoint(m_seekPoint);
   }
}

/// go to next point
void HierarchicalKNNIterator::next(){
   if(hierachyMode){
      seekNextPoint();
   }else{
      nextPointReady = knn->seekNextPoint();
      if (nextPointReady)
         m_buffer = knn->getDataPointer();
   }
}

/// true if end-of-set
bool HierarchicalKNNIterator::atEnd(){
  return(!nextPointReady);
}

card64 HierarchicalKNNIterator::getNumPointsInBlock() const {
	return m_ps->getNumEntries();
}

card64 HierarchicalKNNIterator::getCurrentPointIndex() const {
   x4Assert(!hierachyMode);
   return knn->getCurrentPointNumber();
}


void HierarchicalKNNIterator::cacheRefresh(){
   card32 size= pqKNN.size();
   for(card32 i=0; i< size; i++){
      KNNQuery* boxQuery= pqKNN[i].query;
      boxQuery->setSeekPoint(m_seekPoint);
      if (boxQuery->seekNextPoint()){ //if at least one point in box
         Vector3f nearest= boxQuery->getNearestPointPosition();
         pqKNN[i].quDist= (nearest-m_seekPoint)*(nearest-m_seekPoint);
         }else{
            x4Assert(false);
         }
   }
   pqKNN.refresh();


   size= pqBoxes.size();
   for(card32 i=0; i< size; i++){
     HierarchicalIterator* box= dynamic_cast<HierarchicalIterator*> (pqBoxes[i].node);

     float32 qDistance;

     Sphere3f sp=box->getBoundingBox().getBoundingSphere();
     qDistance=(sp.center-m_seekPoint)*(sp.center-m_seekPoint);
     if(qDistance<sp.radius*sp.radius){
         qDistance= -1.0f;
     }else{
         qDistance=sqrt(qDistance);
         qDistance-=sp.radius;
         qDistance*=qDistance;
     }
     pqBoxes[i].quDist=qDistance;
   }
   pqBoxes.refresh();

}

void HierarchicalKNNIterator::clearQueues(){
   // clear PQs
   /*
   while(!pqKNN.empty()){
      delete (pqKNN.top().query);
      delete (pqKNN.top().node);
      pqKNN.pop();
   }
   while(!pqBoxes.empty()){
      delete (pqBoxes.top().node);
      pqBoxes.pop();
   }
   */

   card32 size= pqKNN.size();
   for(card32 i=0; i< size; i++){
      delete (pqKNN[i].query);
      delete (pqKNN[i].node);
   }
   pqKNN.clear();


   size= pqBoxes.size();
   for(card32 i=0; i< size; i++){
      delete (pqBoxes[i].node);
   }
   pqBoxes.clear();

}

bool HierarchicalKNNIterator::seekNextPoint(){

   while (true){


      //check for points
      if (pqBoxes.empty() && pqKNN.empty()){   // No more points in cloud
          nextPointReady=false;
          return false;
      }


      if (!pqKNN.empty()){
         if(pqBoxes.empty() || pqKNN.top().quDist <= pqBoxes.top().quDist) //point nearer than any box
         {
            knn = pqKNN.top().query;
            SceneObjectIterator* box = pqKNN.top().node;
            pqKNN.pop();

            m_buffer = knn->getDataPointer();

            nextPointReady=true;

            //if(cacheSphere.radius==0.0f)
               //cacheSphere=dynamic_cast<HierarchicalIterator*>(box)->getBoundingBox().getBoundingSphere();


            if(getNewCacheBox){
               cacheBox=dynamic_cast<HierarchicalIterator*>(box)->getBoundingBox();
               getNewCacheBox=false;
            }

            if(knn->seekNextPoint()){
               putOldQueryOnQueue(knn, box);
               return true;
            }else{
               delete knn;
               delete box;
               noChaching=true;
               knn = NULL;
               return true;
            }
         }
      }

      // go into Box

      x4Assert(!pqBoxes.empty());
      

      SceneObjectIterator *thisBox = pqBoxes.top().node;
      HierarchicalIterator* thisBoxHir = dynamic_cast<HierarchicalIterator*> (thisBox);
      pqBoxes.pop();

      if(thisBoxHir->isLeaf()){  //Put Box-Query on Queue
         putAIkNNOnQueue(thisBox);

      }else{ //Put Box Children on Queue
         const card32 children = thisBoxHir->getNumChildren();
         for(card32 i=0; i< children; i++){
            if(thisBoxHir->goToChild(i)){
               x4Assert(thisBox!=NULL);
               putBoxOnQueue(dynamic_cast<HierarchicalIterator*> (  dynamic_cast<SceneObjectIterator*>(thisBoxHir)->copy()) );
               bool ok= thisBoxHir->goToParent();
               x4Assert(ok);
            }
         }
         delete thisBox;
      }

   } //while
}


card8 HierarchicalKNNIterator::get1ub(const AAT &token) const {
   return genericGetS<card8>(token);
}
Vector2ub HierarchicalKNNIterator::get2ub(const AAT &token) const {
   return genericGet<card8, 2>(token);
}
Vector3ub HierarchicalKNNIterator::get3ub(const AAT &token) const {
   return genericGet<card8, 3>(token);
}
Vector4ub HierarchicalKNNIterator::get4ub(const AAT &token) const {
   return genericGet<card8, 4>(token);
}
int32 HierarchicalKNNIterator::get1i(const AAT &token) const {
   return genericGetS<int32>(token);
}
Vector2i HierarchicalKNNIterator::get2i(const AAT &token) const {
   return genericGet<int32, 2>(token);
}
Vector3i HierarchicalKNNIterator::get3i(const AAT &token) const {
   return genericGet<int32, 3>(token);
}
Vector4i HierarchicalKNNIterator::get4i(const AAT &token) const {
   return genericGet<int32, 4>(token);
}
card32 HierarchicalKNNIterator::get1u(const AAT &token) const {
   return genericGetS<card32>(token);
}
Vector2u HierarchicalKNNIterator::get2u(const AAT &token) const {
   return genericGet<card32, 2>(token);
}
Vector3u HierarchicalKNNIterator::get3u(const AAT &token) const {
   return genericGet<card32, 3>(token);
}
Vector4u HierarchicalKNNIterator::get4u(const AAT &token) const {
   return genericGet<card32, 4>(token);
}
float32 HierarchicalKNNIterator::get1f(const AAT &token) const {
   return genericGetS<float32>(token);
}
Vector2f HierarchicalKNNIterator::get2f(const AAT &token) const {
   return genericGet<float32, 2>(token);
}
Vector3f HierarchicalKNNIterator::get3f(const AAT &token) const {
   return genericGet<float32, 3>(token);
}
Vector4f HierarchicalKNNIterator::get4f(const AAT &token) const {
   return genericGet<float32, 4>(token);
}

void HierarchicalKNNIterator::getVertex(char *buffer) const {
     card32 elementSize= m_pc->getDescr()->getSize();
     for (card32 i=0; i<elementSize; i++) {
         buffer[i] = m_buffer[i];
     };
}


card32 HierarchicalKNNIterator::getPointSetPointNumber(){
   x4Assert(!hierachyMode);
   return knn->getCurrentPointNumber();
}


//==============================================================================







ModifyableKNNIterator::ModifyableKNNIterator(PointCloud* pc, card32 subTreeSize, MemoryManager *attachedOctreeMemManager)
:HierarchicalKNNIterator()
{
   modifyable=true;

   m_pc=pc;
   m_ps=NULL;
   m_buffer=NULL;
   m_buffer_rw=NULL;
   m_attachedOctreeMemManager=attachedOctreeMemManager;
   m_subTreeSize=subTreeSize;

   knn = NULL;
   m_seekPoint= makeVector3f(0,0,0);
   native = m_pc->createIterator(SOIT(SOIT::CAP_HIERARCHICAL | SOIT::CAP_MODIFYABLE_PC));

   m_hr = dynamic_cast<HierarchicalIterator*>(native);
   nextPointReady=false;

   if ((m_hr->getNumChildren()>0) && !m_hr->isLeaf())  //m_hr->getNumChildren()==0 means no hr
   {
      hierachyMode=true; // put root box on Queue
      putBoxOnQueue(dynamic_cast<HierarchicalIterator*> (  dynamic_cast<SceneObjectIterator*>(m_hr)->copy()));
   }else{
      hierachyMode=false;
      knn = new KNNQuery(native, subTreeSize, m_attachedOctreeMemManager, modifyable);
   }

   aatPosition= m_pc->getDescr()->getToken("position");


}



ModifyableKNNIterator::~ModifyableKNNIterator(){
}


void ModifyableKNNIterator::set1ub(const AAT &token, const card8 &x)     const {
   genericSetS<card8>(token, x);
}
void ModifyableKNNIterator::set2ub(const AAT &token, const Vector2ub &v) const {
   genericSet<card8, 2>(token, v);
}
void ModifyableKNNIterator::set3ub(const AAT &token, const Vector3ub &v) const {
   genericSet<card8, 3>(token, v);
}
void ModifyableKNNIterator::set4ub(const AAT &token, const Vector4ub &v) const {
   genericSet<card8, 4>(token, v);
}
void ModifyableKNNIterator::set1i(const AAT &token, const int32 &x)    const {
   genericSetS<int32>(token, x);
}
void ModifyableKNNIterator::set2i(const AAT &token, const Vector2i &v) const {
   genericSet<int32, 2>(token, v);
}
void ModifyableKNNIterator::set3i(const AAT &token, const Vector3i &v) const {
   genericSet<int32, 3>(token, v);
}
void ModifyableKNNIterator::set4i(const AAT &token, const Vector4i &v) const {
   genericSet<int32, 4>(token, v);
}
void ModifyableKNNIterator::set1u(const AAT &token, const card32 &x)    const {
   genericSetS<card32>(token, x);
}
void ModifyableKNNIterator::set2u(const AAT &token, const Vector2u &v) const {
   genericSet<card32, 2>(token, v);
}
void ModifyableKNNIterator::set3u(const AAT &token, const Vector3u &v) const {
   genericSet<card32, 3>(token, v);
}
void ModifyableKNNIterator::set4u(const AAT &token, const Vector4u &v) const {
   genericSet<card32, 4>(token, v);
}
void ModifyableKNNIterator::set1f(const AAT &token, const float32 &x)  const {
   genericSetS<float32>(token, x);
}
void ModifyableKNNIterator::set2f(const AAT &token, const Vector2f &v) const {
   genericSet<float32, 2>(token, v);
}
void ModifyableKNNIterator::set3f(const AAT &token, const Vector3f &v) const {
   genericSet<float32, 3>(token, v);
}
void ModifyableKNNIterator::set4f(const AAT &token, const Vector4f &v) const {
   genericSet<float32, 4>(token, v);
}

void ModifyableKNNIterator::setVertex(const char *buffer) const {
     
     card32 elementSize= m_pc->getDescr()->getSize();
     for (card32 i=0; i<elementSize; i++) {
         m_buffer_rw[i] = buffer[i];
     };
}


void ModifyableKNNIterator::setSeekPointAndReset(const Vector3f& sp){
   m_seekPoint=sp;
   if(hierachyMode){

      bool showPQs =false;
      //noChaching=true;

      bool newPointinOldCacheBox= cacheBox.inBoundingBox(m_seekPoint);
      if(noChaching|| !newPointinOldCacheBox){


         clearQueues();
         putBoxOnQueue(dynamic_cast<HierarchicalIterator*> (  dynamic_cast<SceneObjectIterator*>(m_hr)->copy()));
         noChaching=false;
         getNewCacheBox=true;
         cacheMiss++;

      }else{
         cacheRefresh();
         cacheHit++;
         showPQs=true;
      }

   #ifdef _DEBUG //display cache effency
         /*EventCounterDisplay::NewValueStepOnChange e;
         e.value = cacheMiss;
         debugOutput.displayEvent(EventCounterDisplay::getClass(), std::string("PCA: Chache MISS"), &e);

         e.value = cacheHit;
         if (cacheHit%1000==0)debugOutput.displayEvent(EventCounterDisplay::getClass(), std::string("PCA: Cache HIT"), &e);
         if (showPQs){
            e.value = pqBoxes.size();
            debugOutput.displayEvent(EventCounterDisplay::getClass(), std::string("PCA: Cached Boxes"), &e);
            e.value = pqKNN.size();
            debugOutput.displayEvent(EventCounterDisplay::getClass(), std::string("PCA: Cached kNNs"), &e);
         }*/
   #endif


      seekNextPoint();
   }else{
      knn->setSeekPoint(sp);
      nextPointReady = knn->seekNextPoint();
      if (nextPointReady)
      {
         m_buffer = knn->getDataPointer();
         m_buffer_rw = knn->getModifyableDataPointer();
      }
   }

}



void ModifyableKNNIterator::next(){
   if(hierachyMode){
      seekNextPoint();
   }else{
      nextPointReady = knn->seekNextPoint();
      if (nextPointReady)
      {
         m_buffer = knn->getDataPointer();
         m_buffer_rw = knn->getModifyableDataPointer();
      }
   }
}


bool ModifyableKNNIterator::seekNextPoint(){

   while (true){


      //check for points
      if (pqBoxes.empty() && pqKNN.empty()){   // No more Points in Cloud
          nextPointReady=false;
          return false;
      }


      if (!pqKNN.empty()){
         if(pqBoxes.empty() || pqKNN.top().quDist <= pqBoxes.top().quDist) //point nearer than any box
         {
            knn = pqKNN.top().query;
            SceneObjectIterator* box = pqKNN.top().node;
            pqKNN.pop();

            m_buffer = knn->getDataPointer();
            m_buffer_rw = knn->getModifyableDataPointer();

            nextPointReady=true;

            //if(cacheSphere.radius==0.0f)
               //cacheSphere=dynamic_cast<HierarchicalIterator*>(box)->getBoundingBox().getBoundingSphere();


            if(getNewCacheBox){
               cacheBox=dynamic_cast<HierarchicalIterator*>(box)->getBoundingBox();
               getNewCacheBox=false;
            }

            if(knn->seekNextPoint()){
               putOldQueryOnQueue(knn, box);
               return true;
            }else{
               delete knn;
               delete box;
               noChaching=true;
               knn = NULL;
               return true;
            }
         }
      }

      // go into Box


         x4Assert(!pqBoxes.empty());
      

      SceneObjectIterator *thisBox = pqBoxes.top().node;
      HierarchicalIterator* thisBoxHir = dynamic_cast<HierarchicalIterator*> (thisBox);
      pqBoxes.pop();

      if(thisBoxHir->isLeaf()){  //Put Box-Query on Queue
         putAIkNNOnQueue(thisBox);

      }else{ //Put Box Children on Queue
         const card32 children = thisBoxHir->getNumChildren();
         for(card32 i=0; i< children; i++){
            if(thisBoxHir->goToChild(i)){
               x4Assert(thisBox!=NULL);
               putBoxOnQueue(dynamic_cast<HierarchicalIterator*> (  dynamic_cast<SceneObjectIterator*>(thisBoxHir)->copy()) );
               bool ok= thisBoxHir->goToParent();
               x4Assert(ok);
            }
         }
         delete thisBox;
      }

   } //while
}




float X4::HierarchicalKNNIterator::getSquaredDistance() const
{
	if( !hierachyMode)
	{
		return knn->getCurrentSquaredDistance();
	}
	else
	{
		throw X4Exception("HierarchicalKNNIterator::getSquaredDistance() - not yet implemented.");
	}
}