//---------------------------------------------------------------------------
#ifndef HierarchicalKNNIteratorH
#define HierarchicalKNNIteratorH
//---------------------------------------------------------------------------
#include "x3geometry.h"
//---------------------------------------------------------------------------
#include "PointCloudIterators.h"
#include "PointCloud.h"
#include "PointSet.h"
#include "AttachedIndexedOctree.h"
//#include <queue.h>
//---------------------------------------------------------------------------


namespace X4 {


/// KNNEntry is only used in HierarchicalKNNIterator for the priorityQueue
struct KNNEntry {
   KNNQuery *query;
   float32 quDist;
   SceneObjectIterator *node;

   KNNEntry(KNNQuery *n, const float32& q, SceneObjectIterator *iter) {quDist = q; query = n; node = iter;}
   inline bool operator<(const KNNEntry &other) const {
      return quDist > other.quDist;
   }
};

/// MHEntry is only used in HierarchicalKNNIterator for the priorityQueu
struct MHEntry {
   SceneObjectIterator *node;
   float32 quDist;
   MHEntry(SceneObjectIterator* i, const float32& q) {quDist = q; node = i;}
   inline bool operator<(const MHEntry &other) const {
      return quDist > other.quDist;
   }
};

/// VectorPriorityQueue is a replacement for std::priority_queue
/// it adds methods like operator[] and clear,
/// that are needed for HierarchicalKNNIterator
template <class Type>
class  VectorPriorityQueue{
   public:
   std::vector<Type> data;

   inline card32 size()const{return (card32)data.size();};
   inline bool empty()const{return data.empty();}
   inline Type& top(){
      x4Assert(!data.empty());
      return data[0];}
   inline void push(const Type& d){
      data.push_back(d);
      push_heap(data.begin(), data.end());}
   inline void pop(){
      pop_heap(data.begin(), data.end());
      data.pop_back();}
   inline void clear(){
      data.clear();}

   /// push data WITHOUT priority update! you have to use refresh() afterwards!
   inline void pushLazy(const Type& d){ 
	   data.push_back(d);}

   /// random access erase, you have to use refresh() afterwards!!
   inline void erase(const unsigned &index){
		data.erase(data.begin() + index);}

   /// if elements of the Queue have changed (e.g. via operator[]) use refresh()
   inline void refresh(){
      make_heap(data.begin(), data.end());}
   /// if elements of the Queue have changed use refresh() !
   inline Type& operator[](const unsigned &index) {
      x4Assert(index<data.size());
      return data[index];}
   inline const Type& operator[](const unsigned &index) const {
      x4Assert(index<data.size());
      return data[index];}

};

template <class Type>
class  VectorPriorityQueue_Fast{
public:
	std::vector<Type> data;
	unsigned mySize;

	VectorPriorityQueue_Fast() {mySize=0;}
	inline unsigned size()const{return mySize;};
	inline bool empty()const{return mySize==0;}
	inline Type& top(){
		x4Assert(!empty());
		return data[0];}
	inline void push(const Type& d){
		if( mySize>=data.size() )
			data.push_back(d);
		data[mySize] = d;
		mySize++;
		push_heap(data.begin(), data.begin() + mySize);}
	inline void pop(){
		pop_heap(data.begin(), data.begin() + mySize);
		mySize--;
		}
	inline void clear(){
		mySize = 0;
		}

	/// push data WITHOUT priority update! you have to use refresh() afterwards!
	inline void pushLazy(const Type& d){ 
		data.push_back(d);}

	/// random access erase, you have to use refresh() afterwards!!
	inline void erase(const unsigned &index){
		mySize--;
		data.erase(data.begin() + index);}

	/// if elements of the Queue have changed (e.g. via operator[]) use refresh()
	inline void refresh(){
		make_heap(data.begin(), data.begin() + mySize);}
	/// if elements of the Queue have changed use refresh() !
	inline Type& operator[](const unsigned &index) {
		x4Assert(index<mySize);
		return data[index];}
	inline const Type& operator[](const unsigned &index) const {
		x4Assert(index<mySize);
		return data[index];}

};

/// Iterator for the k-nearest-Neighbours
/// Works on all kind of PointClouds
class X3GEOMETRY_API HierarchicalKNNIterator: public PointCloudIterator,
                                               public BasicPointCloudIterator,
                                               public IndexedPointCloudIterator {

 protected:
   bool hierachyMode;
   KNNQuery *knn;
   SceneObjectIterator *native;
   bool nextPointReady;

   bool noChaching, getNewCacheBox;
   card32 cacheHit, cacheMiss;
   BoundingBox3f cacheBox;

   PointCloud* m_pc;
   PointSet* m_ps;
   HierarchicalIterator* m_hr;
   Vector3f m_seekPoint;
   card32 m_subTreeSize;
   float32 m_maxDistanceToSeekPoint_2;		// squared maximum distance to seek point (reports true in atEnd() 
											// when no further points within this distance are available)

   VectorPriorityQueue_Fast<MHEntry> pqBoxes;
   VectorPriorityQueue_Fast<KNNEntry> pqKNN;

   AIOctreeNode* rootNode;
   bool modifyable;

   void putBoxOnQueue(HierarchicalIterator* box);
   void putAIkNNOnQueue(SceneObjectIterator* q);
   void putOldQueryOnQueue(KNNQuery* q, SceneObjectIterator* box);

   bool seekNextPoint();
   void clearQueues();
   void cacheRefresh();

   card32 m_lastIndex;
   const char* m_buffer;
   MemoryManager *m_attachedOctreeMemManager;


 public:
   /// Use setSeekPointAndReset after construction
   HierarchicalKNNIterator(PointCloud*, card32 subTreeSize, MemoryManager *attachedOctreeMemManager);
   ~HierarchicalKNNIterator();

   /// restrict region of interest to a sphere with radius maxDistanceToSeekPoint
   /// default is inf
   void setMaxDistanceToSeekPoint(float d) {m_maxDistanceToSeekPoint_2 = d*d;if( knn ) knn->setMaxDistanceToSeekPoint(d);}
   float32 getMaxDistanceToSeekPoint() const {return sqrt(m_maxDistanceToSeekPoint_2);}

   /// Initialize Iterator to kNN of seekPoint
   void setSeekPointAndReset(const Vector3f&);

   /// go to first nearest point
   virtual void reset();
   /// go to next nearest point
   virtual void next();
   /// true if end-of-set
   virtual bool atEnd();

   virtual card64 getNumPointsInBlock() const; 
   virtual card64 getCurrentPointIndex() const;

   template <class Type, int dim>
   inline StaticVector<Type, dim> genericGet(AAT token) const {
      const char *pos =  m_buffer + token.getOffset();
      return *((StaticVector<Type, dim>*)(pos));
   }

   template <class Type>
   inline Type genericGetS(const AAT& token) const {
      const char *pos =  m_buffer + token.getOffset();
      const StaticVector<Type, 1> v = *((StaticVector<Type, 1>*)(pos));
      return v[0];
   }


   /// get attributes in different formats
   virtual card8     get1ub(const AAT &token) const;
   virtual Vector2ub get2ub(const AAT &token) const;
   virtual Vector3ub get3ub(const AAT &token) const;
   virtual Vector4ub get4ub(const AAT &token) const;

   virtual int32    get1i(const AAT &token) const;
   virtual Vector2i get2i(const AAT &token) const;
   virtual Vector3i get3i(const AAT &token) const;
   virtual Vector4i get4i(const AAT &token) const;

   virtual card32   get1u(const AAT &token) const;
   virtual Vector2u get2u(const AAT &token) const;
   virtual Vector3u get3u(const AAT &token) const;
   virtual Vector4u get4u(const AAT &token) const;

   virtual float32  get1f(const AAT &token) const;
   virtual Vector2f get2f(const AAT &token) const;
   virtual Vector3f get3f(const AAT &token) const;
   virtual Vector4f get4f(const AAT &token) const;

   /// ask VertexDescriptor for buffer size
   virtual void     getVertex(char *buffer) const;

   /// get squared distance of current point to seek point
   float getSquaredDistance() const;



   // --- blockwise access NOT useable here!
   virtual bool nextBlock(){x4Assert(false); return false;}
   virtual const PointSet * lockBlockRead(){x4Assert(false); return NULL;}
   virtual void unlock(){x4Assert(false);}

   /// do not use
   card32 getPointSetPointNumber();

 protected:
   HierarchicalKNNIterator(){}; // for use in ModifyableKNNIterator ONLY

};



/// Iterator for the k-nearest-Neighbours that allows attribute modifications
/// position changes are currently not allowed!
/// Works on all kind of PointClouds, expect hierarchical pointclouds (comming soon)
class X3GEOMETRY_API ModifyableKNNIterator:
    public HierarchicalKNNIterator,
    public ModifyablePointCloudIterator{

 private:
   AAT aatPosition;
   char *m_buffer_rw;

 public:
   /// Use setSeekPointAndReset after construction
   ModifyableKNNIterator(PointCloud*, card32 subTreeSize, MemoryManager *attachedOctreeMemManager);
   ~ModifyableKNNIterator();

   /// Initialize Iterator to kNN of seekPoint
   void setSeekPointAndReset(const Vector3f&);
   virtual void next();
 protected:
   bool seekNextPoint();


   template <class Type, int dim>
   inline void genericSet(const AAT& token, const StaticVector<Type, dim> &v) const {
      x4Assert (token != aatPosition); //Position may not be changed yet !
      x4Assert(m_buffer_rw);
      char *pos =  m_buffer_rw + token.getOffset();
      *((StaticVector<Type, dim>*)(pos)) = v;
   }

   template <class Type>
   inline void genericSetS(const AAT& token, const Type &v) const {

      x4Assert(m_buffer_rw);
      StaticVector<Type, 1> v1; v1[0] = v;
      char *pos =  m_buffer_rw + token.getOffset();
      *((StaticVector<Type, 1>*)(pos)) = v1;
   }

/*
      boundingBoxValid = false;           //in lock block rw
      if (token == aatPosition) {
         attachedDataLargeChange = true;
      } else {
         attachedDataAttribChange = true;
      }
      m_ps->invalidateLastNumFlagged();
*/


 public:

   /// set attributes in different formats
   virtual void set1ub(const AAT &token, const card8 &x)     const ;
   virtual void set2ub(const AAT &token, const Vector2ub &v) const ;
   virtual void set3ub(const AAT &token, const Vector3ub &v) const ;
   virtual void set4ub(const AAT &token, const Vector4ub &v) const ;

   virtual void set1i(const AAT &token, const int32 &x)     const ;
   virtual void set2i(const AAT &token, const Vector2i &v) const ;
   virtual void set3i(const AAT &token, const Vector3i &v) const ;
   virtual void set4i(const AAT &token, const Vector4i &v) const ;

   virtual void set1u(const AAT &token, const card32 &x)     const ;
   virtual void set2u(const AAT &token, const Vector2u &v) const ;
   virtual void set3u(const AAT &token, const Vector3u &v) const ;
   virtual void set4u(const AAT &token, const Vector4u &v) const ;

   virtual void set1f(const AAT &token, const float32 &x)  const ;
   virtual void set2f(const AAT &token, const Vector2f &v) const ;
   virtual void set3f(const AAT &token, const Vector3f &v) const ;
   virtual void set4f(const AAT &token, const Vector4f &v) const ;

   /// ask VertexDescriptor for size
   virtual void setVertex(const char *data) const ;


   // --- blockwise access NOT used here!
   virtual PointSet *lockBlockReadWrite(LockRWHint hint = CHANGE_EVERYTHING){x4Assert(false); return NULL;}
};




}//namespace X4

#endif
