//---------------------------------------------------------------------------
#include "StdAfx.h"
#include "Configuration.h"
//---------------------------------------------------------------------------
#include "ImageDisplayerWidget.h"
//---------------------------------------------------------------------------
#include "PointSet.h"
#include "StringHelper.h"
#include "FileDialogs.h"
#include "RegularData.h"
#include "RegularData.inline.h"
#include "DebugOutput.h"
#include "ImageDisplayer.h"
#include "RGBAImage.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------


ImagePopup::ImagePopup(QWidget *parent) : PopupWindow(parent) {
	ui.setupUi(this);
}
void ImagePopup::on_doubleSpinBoxZoom_valueChanged(double) {
	emit changed();
}
void ImagePopup::on_spinBoxMaxHeight_valueChanged(int i) {
	emit newMaxHeight(i);
}

MappingPopup::MappingPopup(QWidget *parent) : PopupWindow(parent) {
	ui.setupUi(this);
	noSignals = false;
}
void MappingPopup::on_dsbScale_valueChanged(double) {
	if (noSignals) return;
	emit changed();
}
void MappingPopup::on_dsbOffset_valueChanged(double) {
	if (noSignals) return;
	emit changed();
}
void MappingPopup::on_dsbGamma_valueChanged(double) {
	if (noSignals) return;
	emit changed();
}
void MappingPopup::on_cbAuto_stateChanged(int){
	if (noSignals) return;
	if (ui.cbAuto->isChecked()) {
		ui.dsbOffset->setEnabled(false);
		ui.dsbScale->setEnabled(false);
	} else {
		ui.dsbOffset->setEnabled(true);
		ui.dsbScale->setEnabled(true);
	}
	emit changed();
}

FileIOPopup::FileIOPopup(QWidget *parent) : PopupWindow(parent) {
	ui.setupUi(this);
}

void FileIOPopup::on_pbSaveNow_clicked() {
	emit saveNow();
}

void FileIOPopup::on_pbCopyToConsole_clicked() {
	emit copyToConsole();
}


// --

ImageDisplayerWidget::ImageDisplayerWidget(QWidget *parent, Qt::WindowFlags f) : QWidget(parent, f)
{
	ui.setupUi(this);
	ui.scrollArea->setWidget(ui.labelImage);
	imgPopup = new ImagePopup(this);
	mapPopup = new MappingPopup(this);
	filePopup = new FileIOPopup(this);
	psImage = NULL;
	currentChannel = "color";
	maxHeight = 500;
	deactivateChannelChoice = false;
	QObject::connect(imgPopup, SIGNAL(changed()), this, SLOT(changed()));
	QObject::connect(mapPopup, SIGNAL(changed()), this, SLOT(changed()));
	QObject::connect(imgPopup, SIGNAL(newMaxHeight(int)), this, SLOT(newMaxHeight(int)));
	QObject::connect(filePopup, SIGNAL(saveNow()), this, SLOT(saveNow()));
	QObject::connect(filePopup, SIGNAL(copyToConsole()), this, SLOT(copyToConsole()));
}

void ImageDisplayerWidget::changed() {
	displayPointSet(psImage);
}


void ImageDisplayerWidget::on_pushButtonImage_clicked() {
	imgPopup->popup(ui.pushButtonImage);
}
void ImageDisplayerWidget::on_pushButtonMapping_clicked() {
	mapPopup->popup(ui.pushButtonMapping);
}
void ImageDisplayerWidget::on_pushButtonFileIO_clicked() {
	filePopup->popup(ui.pushButtonFileIO);
}
void ImageDisplayerWidget::newMaxHeight(int i) {
	maxHeight = min(i, 10000);
	rethinkLayout();
}

void ImageDisplayerWidget::on_cbChannel_currentIndexChanged(int) {
	if (deactivateChannelChoice) return;
	currentChannel = qString2STLString(ui.cbChannel->currentText());
	changed();
}
void ImageDisplayerWidget::on_cbChannel_editTextChanged(const QString &text) {
	if (deactivateChannelChoice) return;
	currentChannel = qString2STLString(text);
	changed();
}

void ImageDisplayerWidget::rethinkLayout() {
	int h = ui.labelImage->height();
	if (h > maxHeight) h = maxHeight;
	ui.scrollArea->setMinimumHeight(h);
	ui.scrollArea->setMaximumHeight(h);
	//setMaximumHeight(h + ui.bottomFrame->height());
}


Vector3i ImageDisplayerWidget::mapColor(Vector3f col, float32 scale, float32 offset, float32 gamma) {
	col *= scale;
	col += makeVector3f(offset,offset,offset);
	if (gamma != 1.0f && gamma > 0) {
		col[0] = pow(col[0], gamma);
		col[1] = pow(col[1], gamma);
		col[2] = pow(col[2], gamma);
	}
	
	Vector3i icol = makeVector3i((int32)(col[0] * 255.0f), (int32)(col[1] * 255.0f), (int32)(col[2] * 255.0f));
	
	if (icol[0] < 0) icol[0] = 0; if (icol[0] > 255) icol[0] = 255;
	if (icol[1] < 0) icol[1] = 0; if (icol[1] > 255) icol[1] = 255;
	if (icol[2] < 0) icol[2] = 0; if (icol[2] > 255) icol[2] = 255;
	
	return icol;
}

bool ImageDisplayerWidget::createImage(const PointSet *ps, QImage &img, double &minv, double &maxv) {
	if (!ps->providesAttribute(currentChannel)) {
		return false;
	}
	
	const float32 scaleParam = mapPopup->ui.dsbScale->value() / 100.0f;
	const float32 offsetParam = mapPopup->ui.dsbOffset->value();
	const float32 gammaParam = 1.0f/max(mapPopup->ui.dsbGamma->value(), 0.01);

	VAD vad = ps->getDescr()->getAttribute(currentChannel);
	AAT aat = ps->getAAT(currentChannel);
	int offset = aat.getOffset();
	const int numDispChan = min((card32)3, vad.numComponents);
	const int w = ps->getWidth();
	const int h = ps->getHeight();
	img = QImage(w,h, QImage::Format_ARGB32);
	minv = 1e40;
	maxv = -1e40;
	for (int y=0; y<h; y++) {
		for (int x=0; x<w; x++) {	
			Vector3f col = makeVector3f(0,0,0);
			switch (vad.dataFormat) {
				case VAD::DATA_FORMAT_CARD8: {
					card8 *pos = (card8*)(ps->getDataPointer(x,y) + offset);
					for (int j=0; j<numDispChan; j++) {
						col[j] = pos[j] / 255.0f;
						if ((double)col[j] < minv) minv = col[j];
						if ((double)col[j] > maxv) maxv = col[j];
					}
					break;
				}
				case VAD::DATA_FORMAT_INT32: {
					int32 *pos = (int32*)(ps->getDataPointer(x,y) + offset);
					for (int j=0; j<numDispChan; j++) {
						col[j] = pos[j] / 255.0f;
						if ((double)col[j] < minv) minv = col[j];
						if ((double)col[j] > maxv) maxv = col[j];
					}
					break;
				}
				case VAD::DATA_FORMAT_FLOAT32: {
					float32 *pos = (float32*)(ps->getDataPointer(x,y) + offset);
					for (int j=0; j<numDispChan; j++) {
						col[j] = pos[j];
						if ((double)col[j] < minv) minv = col[j];
						if ((double)col[j] > maxv) maxv = col[j];
					}
					break;
				}
			}
			Vector3i iCol = mapColor(col, scaleParam, offsetParam, gammaParam);
			img.setPixel(x,y, qRgb(iCol[0], iCol[1], iCol[2]));
		}
	}
	return true;
}
	
void ImageDisplayerWidget::displayPointSet(const PointSet *ps, std::string channel) {
	if (channel != "") {
		currentChannel = channel;
	}
	if (ps == NULL) return;

	deactivateChannelChoice = true;
	ui.cbChannel->clear();
	const VertexDescriptor *vd = ps->getDescr();
	for (card32 i=0; i<vd->getNumAttribs(); i++) {
		ui.cbChannel->addItem(QString(vd->getAttribute(i).name.c_str()));
	}
	int pos = ui.cbChannel->findText(currentChannel.c_str());
	if (pos != -1) {
		ui.cbChannel->setCurrentIndex(pos);
	} else {
		ui.cbChannel->addItem(currentChannel.c_str());
		ui.cbChannel->setCurrentIndex(ui.cbChannel->count()-1);
	}
	deactivateChannelChoice = false;
	
	psImage = (PointSet*)ps->copy();
	QImage img;
	double minv;
	double maxv;
	bool success = createImage(ps, img, minv, maxv);
	if (mapPopup->ui.cbAuto->isChecked()) {
		mapPopup->noSignals = true;
		double delta = maxv - minv;
		if (delta < 1e-37) delta = 1e-37;
		mapPopup->ui.dsbScale->setValue(100.0 / delta);
		mapPopup->ui.dsbOffset->setValue(-minv/delta);
		success = createImage(ps, img, minv, maxv);
		mapPopup->noSignals = false;
	}
	if (success) {
		QPixmap pix = QPixmap::fromImage(img);
		ui.labelImage->setPixmap(pix);
		int w = (int)(ps->getWidth() * imgPopup->ui.doubleSpinBoxZoom->value() / 100.0);
		int h = (int)(ps->getHeight() * imgPopup->ui.doubleSpinBoxZoom->value() / 100.0);
		if (w < 1) w = 1;
		if (h < 1) h = 1;
		if (w > 10000) w = 10000;
		if (h > 10000) h = 10000;
		ui.labelImage->setScaledContents(true);
		ui.labelImage->resize(w,h);
	} else {
		ui.labelImage->setText(" - channel not found - ");
		ui.labelImage->resize(181, 41);
	}
	rethinkLayout();
}

void ImageDisplayerWidget::displayQImage(const QImage *img) {
	if (img == NULL) return;
	if (img->width() <= 0 || img->height() <= 0) {
		return;
	}
	PointSet ps;
	VertexDescriptor vd;
	vd.pushAttrib(VAD("color", 3, VAD::DATA_FORMAT_CARD8));
	ps.clearAndSetup(img->width(), img->height(), &vd);
	AAT COLAAT = vd.getToken("color");
	const int w = img->width();
	const int h = img->height();
	for (int y=0; y<h; y++) {
		for (int x=0; x<w; x++) {
			QRgb pix = img->pixel(x,y);			
			Vector3ub c = makeVector3ub(qRed(pix), qGreen(pix), qBlue(pix));
			ps.set<card8, 3>(x,y,COLAAT, c);
		}
	}
	displayPointSet(&ps);
}

void ImageDisplayerWidget::display2DRegularData(const RegularDataF *img) {
	if (img == NULL) return;
	DVectorI extends = img->getExtends();
	if (extends.getDim() != 2) return;
	const int32 width = extends[0];
	const int32 height = extends[1];
	if (width <= 0 || height <= 0) {
		return;
	}
	PointSet ps;
	VertexDescriptor vd;
	vd.pushAttrib(VAD("color", 3, VAD::DATA_FORMAT_FLOAT32));
	ps.clearAndSetup(width, height, &vd);
	AAT COLAAT = vd.getToken("color");
	const float32 *sourcePos = img->dataPtr();
	for (int y=0; y<height; y++) {
		for (int x=0; x<width; x++) {
			float32 v = *sourcePos;
			sourcePos++;			
			Vector3f c = makeVector3f(v, v, v);
			ps.set<float32, 3>(x,y,COLAAT, c);
		}
	}
	displayPointSet(&ps);
}


void ImageDisplayerWidget::displayRGBAFloat32Image(const RGBAFloat32Image *img) {
	if (img == NULL) return;
	const int32 width = img->getSizeX();
	const int32 height = img->getSizeY();
	if (width <= 0 || height <= 0) {
		return;
	}
	PointSet ps;
	VertexDescriptor vd;
	vd.pushAttrib(VAD("color", 3, VAD::DATA_FORMAT_FLOAT32));
	ps.clearAndSetup(width, height, &vd);
	AAT COLAAT = vd.getToken("color");
	for (int y=0; y<height; y++) {
		for (int x=0; x<width; x++) {
			Vector3f c = shrink4To3(img->getPixel(x,y));
			ps.set<float32, 3>(x,y,COLAAT, c);
		}
	}
	displayPointSet(&ps);
}

card32 ImageDisplayerWidget::getLoggingFormat() {
	if (filePopup->ui.rbJPG->isChecked()) {
		return JPG;
	} else {
		return PNG;
	}
}

QPixmap ImageDisplayerWidget::getCurrentImage() {
	return *(ui.labelImage->pixmap());
}

void ImageDisplayerWidget::saveNow() {
	std::string selectedFilter;
	std::string filename = FileDialogs::getSaveFileName(this, "Save current image as...", 
		"*.png;;*.jpg", &selectedFilter, "", FileDialogs::EXPORT_DIR);
	if (filename != "") {
		QPixmap pix = getCurrentImage();
		pix.save(filename.c_str(), (selectedFilter=="*.png")?"PNG":"JPG");
	}
}

void ImageDisplayerWidget::copyToConsole() {
	static int numImg = 1;
	QPixmap pix = getCurrentImage();
	QImage img = pix.toImage();
	ImageDisplayer::DisplayPointSet ev(psImage); 
	debugOutput.displayEvent(ImageDisplayer::getClass(), std::string("imgcopy ")+intToStr(numImg), &ev);
	ImageDisplayer::ChangeZoom ez(imgPopup->ui.doubleSpinBoxZoom->value()/100.0); 
	debugOutput.displayEvent(ImageDisplayer::getClass(), std::string("imgcopy ")+intToStr(numImg), &ez);
	ImageDisplayer::ChangeMapping em(mapPopup->ui.dsbOffset->value(), mapPopup->ui.dsbScale->value(), mapPopup->ui.dsbGamma->value()); 
	debugOutput.displayEvent(ImageDisplayer::getClass(), std::string("imgcopy ")+intToStr(numImg), &em);
	numImg++;
}

void ImageDisplayerWidget::changeZoom(double factor) {
	imgPopup->ui.doubleSpinBoxZoom->setValue(factor*100);
	changed();
}

void ImageDisplayerWidget::changeMapping(double offset, double scale, double gamma) {
	mapPopup->ui.dsbOffset->setValue(offset);
	mapPopup->ui.dsbScale->setValue(scale);
	mapPopup->ui.dsbGamma->setValue(gamma);
	changed();
}

void ImageDisplayerWidget::on_pbZoomDouble_clicked() {
	imgPopup->ui.doubleSpinBoxZoom->setValue(imgPopup->ui.doubleSpinBoxZoom->value()*2.0);
	changed();
}

void ImageDisplayerWidget::on_pbZoomHalf_clicked() {
	imgPopup->ui.doubleSpinBoxZoom->setValue(max(imgPopup->ui.doubleSpinBoxZoom->value()/2.0,1.0));
	changed();
}


ImageDisplayerWidget::~ImageDisplayerWidget() {
	delete psImage;
}


#include "ImageDisplayerWidget_moc.h"
