//---------------------------------------------------------------------------
#ifndef InCorePCTopologyGraphH
#define InCorePCTopologyGraphH
//---------------------------------------------------------------------------
#include "x3geometry.h"
//---------------------------------------------------------------------------
#include "LinearAlgebra.h"
#include "PointCloudTopologyGraph.h"
#include "VertexDescriptor.h"
#include "PointSetIterator.h"
//---------------------------------------------------------------------------
#include <vector>
//---------------------------------------------------------------------------

namespace X4 {

class PointSet;
class VertexDescriptor;
//class InCorePCTopologyGraphIterator;
class PointSetIterator;


class X3GEOMETRY_API InCorePCTopologyGraph : public PointCloudTopologyGraph {
	X4_CLASS( InCorePCTopologyGraph )

	private:
	
		/// list of edges with attributes
		/// indices of neighbors are stored in channel "int32 vindex[2]"
		/// additional attributes may be added
		PointSet *edges;
		/// table of edges associated with each vertex (which is the index of the outer array)
		/// the current implementation is not optimized in any way
		std::vector< std::vector<EdgeIndex> > edgeTable;
		AAT aatVIndex;

      char *tmpBuffer;
      char *tmpBuffer2;
      card32 tmpBufferSize;
      
	public:
		InCorePCTopologyGraph();

		virtual void assign(const Object* obj, X4CopyContext *context = NULL);
		virtual void write(OutputObjectStream *s) const;
		virtual void read(InputObjectStream *s);

		virtual void clearAndSetup(const VertexDescriptor *vd);
		virtual void addEdge(VertexIndex v1, VertexIndex v2);      
		virtual void deleteEdge( VertexIndex v1, unsigned index );
		virtual void deleteEdge(EdgeIndex edge);
		virtual void deleteEdge(const std::vector<EdgeIndex>& edges);

		/// Remove all connection for a vertex in the graph (number of vertices does not change)
		virtual void clearVertex( VertexIndex v );
		
		/// remove all connections as also remove the vertex completely (number of vertices decreases by one). 
		/// WARNING: the indices changes in the way that the last vertex became v afterwards, so make sure that this is consistent with your data!!!
		virtual void deleteVertex( VertexIndex v );
		virtual void deleteVertex( const std::vector<VertexIndex>& v );

		virtual void addEdge(const char *data);
		virtual PointCloudIterator *createIterator(PointSet *ps);
		virtual void appendFields(const VertexDescriptor *newAttribs);
		/// Create copy of topology graph
		InCorePCTopologyGraph* copy() const;
		const VertexDescriptor* getDescr() { return edges->getDescr(); }
	
		/// get all adjacent edges (indices) of vertex with index v
		std::vector<EdgeIndex> getEdgeIndices(VertexIndex v) const;
		/// get index of adjacent edge for index v, number (index) i
		EdgeIndex getEdgeIndex(VertexIndex v, VertexIndex i) const;
		/// get number of edges, i.e. maximum edge index is this number minus one
		EdgeIndex getNumEdges() const;
		/// get (unsorted) table of all edges
		inline const PointSet *getEdges() const {return edges;}
		/// get (unsorted) table of all edges
		inline PointSet *getEdges() {return edges;}
		/// get number of vertices for which adjacency information might be available
		inline VertexIndex getNumVertices() const {return (VertexIndex)edgeTable.size();}

		/// get number of adjacent vertices at vertex v
		VertexIndex getNumAdjacentVertices(VertexIndex v) const;
		/// get adjacent vertex number i at vertex v
		VertexIndex getAdjacentVertex(VertexIndex v, VertexIndex i, EdgeIndex* edgeIndex = NULL) const;
		/// check whenever a given vertices are adjacent
		inline bool areVerticesAdjacent(VertexIndex v, VertexIndex w, EdgeIndex* edgeIndex = NULL) const
		{
			for (card32 i=0; i < getNumAdjacentVertices(v); i++)
				if (getAdjacentVertex(v,i) == w)
				{
					if (edgeIndex != NULL) *edgeIndex = getEdgeIndex(v,i);
					return true;
				}
			return false;
		}

		virtual ~InCorePCTopologyGraph(); 
};


class X3GEOMETRY_API InCorePCTopologyGraphPointSetIterator : public AttachedDataPointSetIterator, 
														public TopologyIterator
{

private:
	InCorePCTopologyGraph * m_Graph;


public:

	InCorePCTopologyGraphPointSetIterator(PointSet *ps, AttachedDataContainer *attData, 
		InCorePCTopologyGraph * graph);
	~InCorePCTopologyGraphPointSetIterator();

//	virtual void assign(const Object* obj, X4CopyContext *context = NULL);

	/// returns number of neighbours for current point
	virtual card32 getNumNeighbours();

	/// returns an index of a neighbour point
	virtual PointIndex getNeighbour( card32 nIndex );

};



}

#endif