//---------------------------------------------------------------------------
#ifndef IteratorAdaptorsH
#define IteratorAdaptorsH
//---------------------------------------------------------------------------
#include "x3geometry.h"
//---------------------------------------------------------------------------
#include "PointCloudIterators.h"
#include "GeometricRange.h"
#include "PointFlags.h"
//---------------------------------------------------------------------------
    
namespace X4 {


const card32 maxBlockNumPoints=100000;


///Abstract IteratorAdaptor Base Class
class X3GEOMETRY_API IteratorAdaptor : public BasicPointCloudIterator,  // INTERFACE
													 public IndexedPointCloudIterator, // INTERFACE
													 public ModifyablePointCloudIterator  // INTERFACE

{
 protected:

   BasicPointCloudIterator *m_base;
   const VertexDescriptor *m_vd;

   PointSet* tmp_ps;
   bool locked;
   bool next_block;
   PointCloudIterator* m_native;
   ModifyablePointCloudIterator *m_mod_base;
   IndexedPointCloudIterator *m_indexed;
   PointCloudIterator* m_native_remember_write;
   bool write;


   /// true if actual Point meets Condition of Adaptor
   virtual inline bool pointMeetsCond()=0;

   /// for internal use of child classes only
   IteratorAdaptor(){};


 public:

   IteratorAdaptor(PointCloudIterator* native, const VertexDescriptor *vd);
   virtual ~IteratorAdaptor();

   /// true if end-of-set
   virtual bool atEnd();
   /// go to first point in Range
   virtual void reset();
   /// go to next point in Range
   virtual void next();

	// retrieve indices
   virtual card64 getNumPointsInBlock() const; 
   virtual card64 getCurrentPointIndex() const;

   /// get attributes in different formats
   virtual card8     get1ub(const AAT &token) const;
   virtual Vector2ub get2ub(const AAT &token) const;
   virtual Vector3ub get3ub(const AAT &token) const;
   virtual Vector4ub get4ub(const AAT &token) const;

   virtual int32    get1i(const AAT &token) const;
   virtual Vector2i get2i(const AAT &token) const;
   virtual Vector3i get3i(const AAT &token) const;
   virtual Vector4i get4i(const AAT &token) const;

   virtual card32   get1u(const AAT &token) const;
   virtual Vector2u get2u(const AAT &token) const;
   virtual Vector3u get3u(const AAT &token) const;
   virtual Vector4u get4u(const AAT &token) const;

   virtual float32  get1f(const AAT &token) const;
   virtual Vector2f get2f(const AAT &token) const;
   virtual Vector3f get3f(const AAT &token) const;
   virtual Vector4f get4f(const AAT &token) const;

   /// ask VertexDescriptor for buffer size
   virtual void     getVertex(char *buffer) const;

   // --- blockwise access (faster)
   /// step foreward to next data block, false if end-of-set (affects the same position pointer as "next()")
   virtual bool nextBlock();
   /// access raw data of current data block; call unlock to free;
   /// one data block is guaranteed to easily fit into main memory
   virtual const PointSet *lockBlockRead();


   /// free locked PointSet again
   /// & write back chainged points
   void unlock();

   /// set attributes in different formats
   virtual void set1ub(const AAT &token, const card8 &x)     const;
   virtual void set2ub(const AAT &token, const Vector2ub &v) const;
   virtual void set3ub(const AAT &token, const Vector3ub &v) const;
   virtual void set4ub(const AAT &token, const Vector4ub &v) const;

   virtual void set1i(const AAT &token, const int32 &x)     const;
   virtual void set2i(const AAT &token, const Vector2i &v) const;
   virtual void set3i(const AAT &token, const Vector3i &v) const;
   virtual void set4i(const AAT &token, const Vector4i &v) const;

   virtual void set1u(const AAT &token, const card32 &x)     const;
   virtual void set2u(const AAT &token, const Vector2u &v) const;
   virtual void set3u(const AAT &token, const Vector3u &v) const;
   virtual void set4u(const AAT &token, const Vector4u &v) const;

   virtual void set1f(const AAT &token, const float32 &x)  const;
   virtual void set2f(const AAT &token, const Vector2f &v) const;
   virtual void set3f(const AAT &token, const Vector3f &v) const;
   virtual void set4f(const AAT &token, const Vector4f &v) const;

   /// ask VertexDescriptor for size
   virtual void setVertex(const char *data) const;

   /// access raw data of current data block for read-write access;
   /// call unlock to free locked PointSet
   virtual PointSet *lockBlockReadWrite(LockRWHint hint = CHANGE_EVERYTHING);

};


//====================================================


class X3GEOMETRY_API NotModGeometricRangeIteratorAdaptor : public IteratorAdaptor
{
 protected:

   protected:
   GeometricRange *m_range;
   AAT POSITION;
   AAT FLAGS;

   virtual inline bool pointMeetsCond(){
     return  (m_range->intersects(m_base->get3f(POSITION)) &&
             !(m_base->get1i(FLAGS) & PF_FLAG_MODIFIED));
             //in Range and not modified
   }

 public:

   NotModGeometricRangeIteratorAdaptor(PointCloudIterator* native, const VertexDescriptor *vd,GeometricRange *r);
   ~NotModGeometricRangeIteratorAdaptor();


};


//==============================================================================

class X3GEOMETRY_API NormalIteratorAdaptor : public IteratorAdaptor
{
 protected:

   protected:

   AAT NORMAL;
   Vector3f myNormal;

   virtual inline bool pointMeetsCond(){
     return  (m_base->get3f(NORMAL) == myNormal) ;
             //same Normal
     }

 public:
   NormalIteratorAdaptor(PointCloudIterator* native, const VertexDescriptor *vd,Vector3f n);
   ~NormalIteratorAdaptor(){}
};

//==============================================================================

class X3GEOMETRY_API AllPointsIteratorAdaptor : public IteratorAdaptor
{
 protected:

   protected:

   virtual inline bool pointMeetsCond(){
     return  true;
     }

 public:
   AllPointsIteratorAdaptor(PointCloudIterator* native, const VertexDescriptor *vd);
   ~AllPointsIteratorAdaptor(){}
};



} //ns




#endif
