/***************************************************************************
*   Copyright (c) 2010   Art Tevs                                         *
*                                                                         *
*   This library is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU Lesser General Public License as        *
*   published by the Free Software Foundation; either version 3 of        *
*   the License, or (at your option) any later version.                   *
*                                                                         *
*   This library is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU Lesse General Public License for more details.                    *
*                                                                         *
*   The full license is in LICENSE file included with this distribution.  *
***************************************************************************/

//---------------------------------------------------------------------------
#ifndef MANIFOLD__H
#define MANIFOLD__H

//---------------------------------------------------------------------------
#include "x3geometry.h"
#include "SceneObject.h"
#include "ObjectListProperty.h"

//---------------------------------------------------------------------------
namespace X4
{

	/**
	* Class representing one point on the manifold. Points are always discrete.
	**/
	class X3GEOMETRY_API ManifoldPoint : public Persistent
	{
		X4_CLASS(ManifoldPoint);
	protected:
		DVectorF m_Coordinate;
		DVectorF m_Normal;

	public:
		ManifoldPoint();
		ManifoldPoint(const DVectorF& coord, const DVectorF& normal) : m_Coordinate(coord), m_Normal(normal) {}
		ManifoldPoint(const ManifoldPoint& point);
		ManifoldPoint(const ManifoldPoint* point);

		virtual ~ManifoldPoint();

		//! Get point coordinate in embedded manifold space
		inline void setCoordinate(const DVectorF& coord) { m_Coordinate = coord; }
		inline DVectorF getCoordinate() const { return m_Coordinate; }
		inline void setNormal(const DVectorF& normal) { m_Normal = normal; }
		inline DVectorF getNormal() const { return m_Normal; }

		virtual void write(OutputObjectStream *s) const;
		virtual void read(InputObjectStream *s);
		virtual void assign(const Object* obj, X4CopyContext *context  = NULL);

		virtual bool isDiscretePoint() const { return false; }
	};

	/**
	* Class for collecting manifold points
	**/
	class X3GEOMETRY_API ManifoldPointList : public AttachedData
	{
		X4_CLASS(ManifoldPointList);
	protected:
		std::vector<ManifoldPoint*> m_ManifoldPoint;

	public:
		ManifoldPointList();
		virtual ~ManifoldPointList();

		IMPLEMENT_OBJECT_LIST_METHODS_STL(ManifoldPoint, ManifoldPoint);

		ManifoldPoint* getPoint(mpcard index) const
		{
			return static_cast<ManifoldPoint*>(m_ManifoldPoint[index]);
		}
		ManifoldPoint* operator[](mpcard index)
		{
			return getPoint(index);
		}

		static std::string getDefaultName() { return "ManifoldPointList"; }
		void addManifoldPointList(ManifoldPointList* list);
	};

	/**
	* This class represents a manifold which can be used to represent different kind of geometry.
	* Manifolds can be used to be registered against each other or for other operations.
	**/
	class X3GEOMETRY_API Manifold : public SceneObject
	{
		X4_ABSTRACT_CLASS(Manifold)
	public:
		Manifold();
		virtual ~Manifold();

		//! Type of the interpolation used when accessing the manifold		
		enum InterpolationAlgorithm
		{
			NEAREST, // nearest interpolation
			LINEAR,
			EXACT    // there is no real exact interpolation, but used to mark exact results
		};

		/**
		* Set interpolation algorithm used for following operations.
		**/
		virtual void setInterpolationAlgorithm(InterpolationAlgorithm alg) { m_InterpolationAlgorithm = alg; }
		InterpolationAlgorithm getInterpolationAlgorithm() { return m_InterpolationAlgorithm; }

		/**
		* Get geodetic distance between two points on the manifold.
		**/
		virtual float32 getGeodeticDistance(const ManifoldPoint* pointA, const ManifoldPoint* pointB) = 0;
		inline float32 getGeodeticDistance(const ManifoldPoint& pointA, const ManifoldPoint& pointB)
		{
			return getGeodeticDistance(&pointA, &pointB);
		}

		/**
		* Return list of geodetic distances to given points.
		**/
		virtual std::vector<float32> getGeodeticDistances(const ManifoldPoint* pointA, const std::vector<const ManifoldPoint*>& toPoints) = 0;
		inline std::vector<float32> getGeodeticDistances(const ManifoldPoint& pointA, const std::vector<const ManifoldPoint>& toPoints)
		{
			std::vector<const ManifoldPoint*> points;
			for (size_t i=0; i < toPoints.size(); i++) points.push_back(&(toPoints[i]));
			return getGeodeticDistances(&pointA, points);
		}

		/**
		* Get geodetic path between two points.
		**/
		virtual std::vector<DVectorF> getGeodeticPath(const ManifoldPoint* pointA, const ManifoldPoint* pointB) = 0;
		virtual std::vector<const std::vector<DVectorF>*> getGeodeticPaths(const ManifoldPoint* pointA, const std::vector<const ManifoldPoint*>& toPoints) = 0;
		/*{
			std::vector<std::vector<DVectorF> > paths(toPoints.size());
			for (size_t i=0; i < toPoints.size(); i++)
			{
				paths[i]= getGeodeticPath(pointA, toPoints[i]);
			}
			return paths;
		}*/

		/**
		* Create manifold point from the given coordinate.
		**/
		virtual ManifoldPoint* createPoint(const DVectorF& coordinate) = 0;

		/**
		* Dirty geodesics. This will mark geodesics as dirty and can be detected by derived classes, that this has to be recomputed for example.
		**/
		void dirtyGeodesics() { m_bDirtyGeodesics = true; }

		/** 
		* Project a point onto the manifold. Return false if projection fails (i.e. point to far away from manifold surface)
		**/
		virtual bool projectPoint( const DVectorF &point, float sigma, DVectorF &projPoint, DVectorF &projNormal ) { return false; }

		/** 
		* Get median point distance of the manifold. Since this distance is used in a lot of computations some meaningfull value 
		* should be given back here. Derived classes must implement this.
		**/
		virtual float getMedianPointDistance() = 0;


		virtual SceneObjectIterator *createIterator(SceneObjectIteratorType type);
		virtual bool                 supportsIteratorType(SceneObjectIteratorType cap);
		virtual void write(OutputObjectStream *s) const = 0;
		virtual void read(InputObjectStream *s) = 0;
		virtual void assign(const Object* obj, X4CopyContext *context) = 0;

	protected:
		InterpolationAlgorithm m_InterpolationAlgorithm;
		bool m_bDirtyGeodesics;

	};
};

#endif 
