//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "MultiResolutionInCoreVolume.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "RegularVolume.h"
#include "VolumeTools.h"
#include "BoundingBox.inline.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------



IMPLEMENT_X4_CLASS(MultiResolutionInCoreVolume, 0)
{
	BEGIN_CLASS_INIT(MultiResolutionInCoreVolume);
	ADD_VAR_ARRAY_PROP( voxels, 0, 
		new ObjectClassProperty( "element" , MetaClass::FIRST_VERSION, NULL, 
                               RegularVolume::getClass(), true, NULL )
	)
	ADD_VAR_ARRAY_PROP( hierarchyAttachments, 0, 
		new ObjectClassProperty( "hierarchyAttachments" , MetaClass::FIRST_VERSION, NULL, 
                               AttachedDataContainer::getClass(), true, NULL )
	)
}


IMPLEMENT_X4_CLASS(MultiResolutionInCoreVolumeIterator, 0)
{
	BEGIN_CLASS_INIT(MultiResolutionInCoreVolumeIterator);
}



// ----



MultiResolutionInCoreVolume::MultiResolutionInCoreVolume()
{
}

MultiResolutionInCoreVolume::~MultiResolutionInCoreVolume()
{
}

void MultiResolutionInCoreVolume::assign(const Object* obj, X4CopyContext *context)
{
	copyObjectProperties(obj, this);
}


SceneObjectIterator *MultiResolutionInCoreVolume::createIterator(SceneObjectIteratorType type) 
{
   if (voxels.empty()) return NULL;
   if (!supportsIteratorType(type)) {
      return NULL;
   } else {
		return new MultiResolutionInCoreVolumeIterator(&voxels, &hierarchyAttachments);
   }
}

bool MultiResolutionInCoreVolume::supportsIteratorType(SceneObjectIteratorType cap) 
{
   SOIT possibleCaps(  SOIT::CAP_BASIC_PC 
							| SOIT::CAP_INDEXED_PC 
							| SOIT::CAP_RANDOM_ACCESS_PC 
							| SOIT::CAP_MODIFYABLE_PC
							| SOIT::CAP_HIERARCHICAL  
							| SOIT::CAP_ATTACH_SECONDARY
							| SOIT::CAP_REGULAR_VOLUME );
   if( !possibleCaps.providesCapsFor(cap))
		return false;
	else
		return true;
}

const VertexDescriptor *MultiResolutionInCoreVolume::getDescr() const 
{
   if (voxels.size() == 0) return NULL;
   return voxels[0]->getDescr();
}

PointCloud *MultiResolutionInCoreVolume::appendFields(const VertexDescriptor *newAttribs) const 
{
   if (voxels.empty()) {
      x4Assert(false);
      return NULL;
   }
   MultiResolutionInCoreVolume *result = (MultiResolutionInCoreVolume*)this->copy();
   result->setVoxels(voxels[0]->appendFields(newAttribs));
   return result;
}

card64 MultiResolutionInCoreVolume::getNumPoints() const
{
   if (voxels.size() == 0) return 0;
   return voxels[0]->getWidth() * voxels[0]->getHeight() * voxels[0]->getDepth();
}

void MultiResolutionInCoreVolume::clear() 
{
   for (unsigned i=0; i<voxels.size(); i++) {
      delete voxels[i];
      delete hierarchyAttachments[i];
   }
   voxels.clear();
   hierarchyAttachments.clear();
}

void MultiResolutionInCoreVolume::clearAndSetup(card32 width, card32 height, card32 depth, const VertexDescriptor *descr) 
{
   clear();
   RegularVolume *v = new RegularVolume();
   v->clearAndSetup(width, height, depth, descr, BoundingBox3f(NULL_VECTOR3F, makeVector3f(1,1,1)));
   setVoxels(v);
}

void MultiResolutionInCoreVolume::clearAndSetup(card32 width, card32 height, card32 depth, const VertexDescriptor *descr, const BoundingBox3f &bb) 
{
   clear();
   RegularVolume *v = new RegularVolume();
   v->clearAndSetup(width, height, depth, descr, bb);
   setVoxels(v);
}

const RegularVolume *MultiResolutionInCoreVolume::getVoxels() const 
{
   if (voxels.size() > 0) {
      return voxels[0];
   } else {
      return NULL;
   }
}

void MultiResolutionInCoreVolume::setVoxels(RegularVolume *v) 
{
   clear();
   voxels.push_back(v);
   hierarchyAttachments.push_back(new AttachedDataContainer());
   recreateMRLevels();
}

RegularVolume *MultiResolutionInCoreVolume::getVoxels() 
{
   if (voxels.size() > 0) {
      return voxels[0];
   } else {
      return NULL;
   }
}

card32 MultiResolutionInCoreVolume::getNumMRLevels() 
{
   return (card32)voxels.size();
}

RegularVolume *MultiResolutionInCoreVolume::getDownsampled(card32 i) 
{
   x4Assert(voxels.size() > i);
   return voxels[i];
}

void MultiResolutionInCoreVolume::recreateMRLevels() 
{
   x4Assert(voxels.size() >= 1);
   if (voxels.empty()) return;
   for (unsigned i=1; i<voxels.size(); i++) {
      delete voxels[i];
      delete hierarchyAttachments[i];
   }
   voxels.resize(1);
   hierarchyAttachments.resize(1);
   
   card32 size = min(voxels.back()->getWidth(), min(voxels.back()->getHeight(), voxels.back()->getDepth()));
   while (size >= 2) {
      voxels.push_back(downsampleVolumeByTwo(voxels.back()));
      hierarchyAttachments.push_back(new AttachedDataContainer());
      size = min(voxels.back()->getWidth(), min(voxels.back()->getHeight(), voxels.back()->getDepth()));
   }
}


// --



MultiResolutionInCoreVolumeIterator::MultiResolutionInCoreVolumeIterator(
   std::vector<RegularVolume*> *voxels, std::vector<AttachedDataContainer*> *attData) 
{
   x4Assert(voxels->size() > 0);
   this->voxels = voxels;
   this->attData = attData;
   currentLevel = (card32)voxels->size()-1;
   lockedReadWrite = false;
   retarget((*voxels)[currentLevel]->getVoxelData(), (*attData)[currentLevel]);
}

AttachedDataContainer *MultiResolutionInCoreVolumeIterator::lockAttachments() {
	return (*attData)[currentLevel];
}

void MultiResolutionInCoreVolumeIterator::unlockAttachments() {}


card32 MultiResolutionInCoreVolumeIterator::getNumChildren() 
{
   if (currentLevel <= 0) {
      return 0;
   } else {
      return 1;
   }
}

bool MultiResolutionInCoreVolumeIterator::goToChild(card32 index) 
{
   if (getNumChildren() > 0 && index == 0) {
      currentLevel--;
      retarget((*voxels)[currentLevel]->getVoxelData(), (*attData)[currentLevel]);
      return true;
   } else {
      return false;
   }
}

card32 MultiResolutionInCoreVolumeIterator::getCurrentLevel(void)
{
	return currentLevel;
}

bool MultiResolutionInCoreVolumeIterator::goToParent() 
{
   if (currentLevel < voxels->size()-1) {
      currentLevel++;
      retarget((*voxels)[currentLevel]->getVoxelData(), (*attData)[currentLevel]);
      return true;
   }
   return false;
}

BoundingBox3f MultiResolutionInCoreVolumeIterator::getBoundingBox() 
{
   return (*voxels)[currentLevel]->getBoundingBox();
}

bool MultiResolutionInCoreVolumeIterator::isLeaf() const 
{
   return currentLevel == (*voxels).size()-1;
}

bool MultiResolutionInCoreVolumeIterator::getNodeInfo(NodeInfoType whichInfo, const void *parameter, void *data) 
{
   switch (whichInfo) {
    case niPointSpacing: {
      const RegularVolume *vol = (*voxels)[currentLevel];
      BoundingBox3f bb = vol->getBoundingBox();
      float32 xss = bb.getSideLength(0) / vol->getWidth();
      float32 yss = bb.getSideLength(1) / vol->getHeight();
      float32 zss = bb.getSideLength(2) / vol->getDepth();

      *((float32*)data) = max(xss, max(yss, zss));
      return true;
    }

    case niNumPointsInNode: {
      *((card64*)data) = (*voxels)[currentLevel]->getVoxelData()->getNumEntries();
      return true;
    }
    
    case niChildAvailable: {
      *((bool*)data) = true;
      return true;
    }

    default: {
      return false;
    }
   };
}

BoundingBox3f MultiResolutionInCoreVolumeIterator::getVolumeRegion() 
{
   return (*voxels)[currentLevel]->getBoundingBox();
}

void MultiResolutionInCoreVolumeIterator::goTo(card32 x, card32 y, card32 z)
{
	goToIndex( (z*(*voxels)[currentLevel]->getHeight() + y) * (*voxels)[currentLevel]->getWidth() + x );
}

Vector3i MultiResolutionInCoreVolumeIterator::getVolumeSize()
{
	return makeVector3i( (*voxels)[currentLevel]->getWidth(),
			(*voxels)[currentLevel]->getHeight(),
			(*voxels)[currentLevel]->getDepth() );
}

const RegularVolume *MultiResolutionInCoreVolumeIterator::lockVolumeBlockRead() {
	return (*voxels)[currentLevel];	
}

RegularVolume *MultiResolutionInCoreVolumeIterator::lockVolumeBlockReadWrite() {
	lockedReadWrite = true;
	return (*voxels)[currentLevel];
}

void MultiResolutionInCoreVolumeIterator::unlockVolume() {
	if (lockedReadWrite) {
		(*attData)[currentLevel]->deleteDelOnWriteData();
		lockedReadWrite = false;
	}
}


MultiResolutionInCoreVolumeIterator::~MultiResolutionInCoreVolumeIterator() {
	x4Assert(!lockedReadWrite);
   if (attachedDataAttribChange || attachedDataLargeChange || lockedReadWrite) {
		for (size_t i=0; i<attData->size(); i++) {
			(*attData)[i]->deleteDelOnWriteData();
		}
   }
}