//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "MultiScaleInCorePointCloud.h"
#include "X4Exception.h"
#include "UnstructuredInCorePointCloud.h"
#include "pointcloudsmoother.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------



IMPLEMENT_X4_CLASS(MultiScaleInCorePointCloud, 0)
{
   BEGIN_CLASS_INIT(MultiScaleInCorePointCloud)
   ADD_OBJECT_PROP(vertDescr, 0, VertexDescriptor::getClass(), true)
   ADD_OBJECT_PROP(params, 0, MultiScalePCParams::getClass(), true)
   ADD_OBJECT_LIST_PROP( levels, 0, PointSet::getClass() )
   ADD_OBJECT_LIST_PROP( hierarchyAttachments, 0, AttachedDataContainer::getClass() )
}

IMPLEMENT_X4_CLASS(MultiScaleInCorePointCloudIterator, 0)
{
	BEGIN_CLASS_INIT(MultiScaleInCorePointCloudIterator)
}



MultiScaleInCorePointCloud::MultiScaleInCorePointCloud()
{
   params = NULL;
   vertDescr = NULL;
}

PointCloud * MultiScaleInCorePointCloud::appendFields( const VertexDescriptor *newAttribs ) const
{
   MultiScaleInCorePointCloud *result = (MultiScaleInCorePointCloud*)this->copy();
   for (size_t i=0; i<levels.size(); i++) {
      result->levels[i] = result->levels[i]->appendFields(newAttribs);
   }
   result->vertDescr = (VertexDescriptor*)result->levels[0]->getDescr()->copy();
   return result;
}

card64 MultiScaleInCorePointCloud::getNumPoints() const
{
   if (levels.empty()) throw EInvalidState("MultiScaleInCorePointCloud::getNumPoints() - data structure not setup yet.");
   return levels[0]->getNumEntries();
}

void MultiScaleInCorePointCloud::clear()
{
   for (size_t i=0; i<levels.size(); i++) {
      delete levels[i];
   }
   levels.clear();
   for (size_t i=0; i<hierarchyAttachments.size(); i++) {
      delete hierarchyAttachments[i];
   }
   hierarchyAttachments.clear();
   delete vertDescr;
   vertDescr = NULL;
   delete params;
   params = NULL;
}

void MultiScaleInCorePointCloud::clearAndSetup( const VertexDescriptor *vertDescr )
{
   clear();
   delete this->vertDescr;
   this->vertDescr = (VertexDescriptor*)vertDescr->copy();
   PointSet *basePS = new PointSet();
   basePS->clearAndSetup(1,0,vertDescr);
   levels.push_back(basePS);
   hierarchyAttachments.push_back(new AttachedDataContainer());
}

void MultiScaleInCorePointCloud::addPoint( const card8 *pt )
{
   if (isBeingSetup())
   {
      levels[0]->addPoint((const char*) pt);
   } else {
      if (isBuilt()) throw EInvalidState("MultiScaleInCorePointCloud::addPoint() - Data structure has already been build.");
      throw EInvalidState("MultiScaleInCorePointCloud::addPoint() - Data structure has not yet been initialized (clearAndSetup() missing).");
   }
}

void MultiScaleInCorePointCloud::addPointSet( const PointSet *ps )
{
   if (isBeingSetup())
   {
      PointSet *ps0 = levels[0];
      if (ps->getDescr()->equal(ps0->getDescr())) {
         for (mpcard i=0; i<ps->getNumEntries(); i++) {
            ps0->addPoint(ps->getDataPointer(i));
         }
      } else {
         throw ERangeCheck("MultiScaleInCorePointCloud::addPointSet() - vertex descriptors are different.");
      }
   } else {
      if (isBuilt()) throw EInvalidState("MultiScaleInCorePointCloud::addPointSet() - Data structure has already been build.");
      throw EInvalidState("MultiScaleInCorePointCloud::addPointSet() - Data structure has not yet been initialized (clearAndSetup() missing).");
   }
}

void MultiScaleInCorePointCloud::assign( const Object* obj, X4CopyContext *context /*= NULL*/ )
{
   const MultiScaleInCorePointCloud *otherObj = dynamic_cast<const MultiScaleInCorePointCloud*>(obj);
   if (!otherObj) {
      throw ETypeMismatch("MultiScaleInCorePointCloud::assign() - wrong type.");
   }
   MultiScalePointCloud::assign(obj, context);
   clear();
   this->vertDescr = (VertexDescriptor*)otherObj->vertDescr->copy();
   levels.resize(otherObj->levels.size());
   for (size_t i=0; i<otherObj->levels.size(); i++) {
      levels[i] = (PointSet*)(otherObj->levels[i]->copy());
   }
   hierarchyAttachments.resize(otherObj->hierarchyAttachments.size());
   for (size_t i=0; i<otherObj->hierarchyAttachments.size(); i++) {
      hierarchyAttachments[i] = (AttachedDataContainer*)(otherObj->hierarchyAttachments[i]->copy());
   }
   params = (MultiScalePCParams*)(otherObj->params->copy());   
}

void MultiScaleInCorePointCloud::build( const MultiScalePCParams *params, const PointCloudSmoother *interp )
{
   if (!isBeingSetup()) throw EInvalidState("MultiScaleInCorePointCloud::build() - the data structure is not being build right now.");
   this->params = (MultiScalePCParams*)params->copy();
   levels.resize(params->numLevels+1,NULL);
   hierarchyAttachments.resize(params->numLevels+1,NULL);
   float32 currentGridSpacing = params->baseGridSpacing;
   for (size_t i=1; i<params->numLevels+1; i++) {
      UnstructuredInCorePointCloud *tempPC = new UnstructuredInCorePointCloud();
      tempPC->setPointSet((PointSet*)levels[i-1]->copy());
      UnstructuredInCorePointCloud *result;
      result = interp->createSmoothed(tempPC, currentGridSpacing, params->levelFactor);
      currentGridSpacing *= params->levelFactor;
      levels[i] = (PointSet*)(result->getPointSet()->copy());
      delete tempPC;
      delete result;
      hierarchyAttachments[i] = new AttachedDataContainer;
   }
}

void MultiScaleInCorePointCloud::build(const MultiScalePCParams *params, std::deque<PointSet*> psvec)
{
    this->params = (MultiScalePCParams*)params->copy();
    unsigned nlevels = params->numLevels+1;
    if (nlevels != psvec.size()) {
        throw X4Exception("MultiScaleInCorePointCloud::build - level number mismatch");
    }
    levels.resize(nlevels,NULL);
    hierarchyAttachments.resize(nlevels,NULL);
    for (size_t i=0; i<levels.size(); i++) {
        levels[i] = psvec[i];
        hierarchyAttachments[i] = new AttachedDataContainer;
    }
}

void MultiScaleInCorePointCloud::build(const MultiScalePCParams *params)
{
    this->params = (MultiScalePCParams*)params->copy();
    levels.resize(params->numLevels+1,NULL);
    hierarchyAttachments.resize(params->numLevels+1,NULL);
    for (size_t i=0; i<levels.size(); i++) {
        hierarchyAttachments[i] = new AttachedDataContainer;
    }
}

SceneObjectIterator * MultiScaleInCorePointCloud::createIterator( SceneObjectIteratorType type )
{
	if (isBuilt()) {
		return new MultiScaleInCorePointCloudIterator(&levels, &hierarchyAttachments, params);
	} else {
		throw EInvalidState("MultiScaleInCorePointCloud::build() - the data structure is not yet built.");
   }
}

bool MultiScaleInCorePointCloud::supportsIteratorType( SceneObjectIteratorType cap )
{
	SOIT possibleCaps(
	     SOIT::CAP_BASIC_PC 
		| SOIT::CAP_INDEXED_PC 
		| SOIT::CAP_RANDOM_ACCESS_PC 
		| SOIT::CAP_MODIFYABLE_PC
		| SOIT::CAP_HIERARCHICAL  
		| SOIT::CAP_ATTACH_SECONDARY );
	if( !possibleCaps.providesCapsFor(cap))
		return false;
	else
		return true;
}

UnstructuredInCorePointCloud * MultiScaleInCorePointCloud::createExcerptDiscrScale( const BoundingBox3f &region, card32 level )
{
	if (level >= levels.size()) {
		throw ERangeCheck("MultiScaleInCorePointCloud::createExcerptDiscrScale() - level out of range.");
	}
	PointSet *ps = levels[level];
	if (!ps->providesAttribute("position", 3, VAD::DATA_FORMAT_FLOAT32)) {
		throw ETypeMismatch("MultiScaleInCorePointCloud::createExcerptDiscrScale() - position channel required.");
	}
	AAT positionAAT = ps->getAAT("position", 3, VAD::DATA_FORMAT_FLOAT32);
	
	UnstructuredInCorePointCloud *result = new UnstructuredInCorePointCloud;
	result->clearAndSetup(ps->getDescr(), 0);
	PointSet *resultPS = result->getPointSet();
	
	mpcard n = ps->getNumEntries();
   for (mpcard i=0; i<n; i++) {
		if (region.inBoundingBox(ps->get3f(i,positionAAT))) {
			resultPS->addPoint(ps->getDataPointer(i));
		}
   }
   return result;
}

UnstructuredInCorePointCloud * MultiScaleInCorePointCloud::createExcerptContScale( const BoundingBox3f &region, float32 gridSpacing, const PointCloudSmoother *interp )
{
   notImplemented();
   return NULL;
}

const MultiScalePCParams *MultiScaleInCorePointCloud::getMultiScaleParameters() const
{
   return params;
}

PointSet *MultiScaleInCorePointCloud::getLevel(card32 level)
{
	if (level <levels.size()) {
		return levels[level];
	} else {
		x4Assert(false);
		return NULL;
	}
}

const PointSet *MultiScaleInCorePointCloud::getLevel(card32 level) const
{
	if (level <levels.size()) {
		return levels[level];
	} else {
		x4Assert(false);
		return NULL;
	}
}

void MultiScaleInCorePointCloud::attachData(AttachedData* att, unsigned level)
{
    if (level > hierarchyAttachments.size()) return;
    hierarchyAttachments[level]->attachData(att);
}

AttachedData* MultiScaleInCorePointCloud::getAttachedData(unsigned level, std::string name)
{
    if (level > hierarchyAttachments.size()) return NULL;
    return hierarchyAttachments[level]->getData(name);
}

// -----



MultiScaleInCorePointCloudIterator::MultiScaleInCorePointCloudIterator(
   std::vector<PointSet*> *levels, std::vector<AttachedDataContainer*> *attData, MultiScalePCParams *params) 
{
   x4Assert(levels->size() > 0);
   this->levels = levels;
   this->attData = attData;
   this->params = params;
   currentLevel = (card32)levels->size()-1;
   lockedReadWrite = false;
   retarget((*levels)[currentLevel], (*attData)[currentLevel]);
}

AttachedDataContainer *MultiScaleInCorePointCloudIterator::lockAttachments() {
   return (*attData)[currentLevel];
}

void MultiScaleInCorePointCloudIterator::unlockAttachments() {}


card32 MultiScaleInCorePointCloudIterator::getNumChildren() 
{
   if (currentLevel <= 0) {
      return 0;
   } else {
      return 1;
   }
}

bool MultiScaleInCorePointCloudIterator::goToChild(card32 index) 
{
   if (getNumChildren() > 0 && index == 0) {
      currentLevel--;
      retarget((*levels)[currentLevel], (*attData)[currentLevel]);
      return true;
   } else {
      return false;
   }
}

card32 MultiScaleInCorePointCloudIterator::getCurrentLevel(void)
{
   return currentLevel;
}

bool MultiScaleInCorePointCloudIterator::goToParent() 
{
   if (currentLevel < levels->size()-1) {
      currentLevel++;
      retarget((*levels)[currentLevel], (*attData)[currentLevel]);
      return true;
   }
   return false;
}

BoundingBox3f MultiScaleInCorePointCloudIterator::getBoundingBox() 
{
   return (*levels)[currentLevel]->getBoundingBox();
}

bool MultiScaleInCorePointCloudIterator::isLeaf() const 
{
   return currentLevel == (*levels).size()-1;
}

bool MultiScaleInCorePointCloudIterator::getNodeInfo(NodeInfoType whichInfo, const void *parameter, void *data) 
{
   switch (whichInfo) {
    case niPointSpacing: {
		float32 spacing = params->baseGridSpacing*powf(params->levelFactor, currentLevel);
       *((float32*)data) = spacing;
       return true;
    }

    case niNumPointsInNode: {
       *((card64*)data) = (*levels)[currentLevel]->getNumEntries();
       return true;
    }

    case niChildAvailable: {
       *((bool*)data) = true;
       return true;
    }

    default: {
       return false;
    }
   };
}

MultiScaleInCorePointCloudIterator::~MultiScaleInCorePointCloudIterator() {
   x4Assert(!lockedReadWrite);
   if (attachedDataAttribChange || attachedDataLargeChange || lockedReadWrite) {
      for (size_t i=0; i<attData->size(); i++) {
         (*attData)[i]->deleteDelOnWriteData();
      }
   }
}

