//---------------------------------------------------------------------------
#ifndef PointCloudIteratorsH
#define PointCloudIteratorsH
//---------------------------------------------------------------------------
#include "x3geometry.h"
//---------------------------------------------------------------------------
#include "LinearAlgebra.h"
#include "BoundingBox.h"
#include "VertexDescriptor.h"
#include "SceneObjectIterator.h"
//---------------------------------------------------------------------------

namespace X4 {

class AttachedData;


/** base class for all concrete point cloud iterator classes */

class X3GEOMETRY_API PointCloudIterator : public SceneObjectIterator {
 private:
   static MetaClass *PointCloudIteratorMetaClass;

 protected:
   #ifdef _DEBUG
   static inline card32 getAATSize(const AAT &aat) {return aat.size;}
   #endif

 public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

}; 




/** Basic point cloud iterator: step foreward through all points, access attributes, copy point sets in chunks
    If the iterator is also hierarchical, the currently selected node is traversed ONLY */
class X3GEOMETRY_API BasicPointCloudIterator {  /* INTERFACE CLASS */
 public:
   virtual ~BasicPointCloudIterator() { }

   /// go to first point
   virtual void reset() = 0;
   /// go to next point
   virtual void next() = 0;
   /// true if end-of-set
   virtual bool atEnd() = 0;
   
   /// get attributes in different formats
   virtual card8     get1ub(const AAT &token) const = 0;
   virtual Vector2ub get2ub(const AAT &token) const = 0;
   virtual Vector3ub get3ub(const AAT &token) const = 0;
   virtual Vector4ub get4ub(const AAT &token) const = 0;

   virtual int32    get1i(const AAT &token) const = 0;
   virtual Vector2i get2i(const AAT &token) const = 0;
   virtual Vector3i get3i(const AAT &token) const = 0;
   virtual Vector4i get4i(const AAT &token) const = 0;

   virtual float32  get1f(const AAT &token) const = 0;
   virtual Vector2f get2f(const AAT &token) const = 0;
   virtual Vector3f get3f(const AAT &token) const = 0;
   virtual Vector4f get4f(const AAT &token) const = 0;

   virtual card32    get1u(const AAT &token) const = 0;
   virtual Vector2u get2u(const AAT &token) const = 0;
   virtual Vector3u get3u(const AAT &token) const = 0;
   virtual Vector4u get4u(const AAT &token) const = 0;

   /// ask VertexDescriptor for buffer size
   virtual void     getVertex(char *buffer) const = 0;

   // --- blockwise access (faster)
   /// step foreward to next data block, false if end-of-set (affects the same position pointer as "next()")
   /// no effect for hierarchal iterators (guaranteed to yield only one block per node)
   virtual bool nextBlock() = 0;
   /// access raw data of current data block; call unlock to free;
   /// one data block is guaranteed to easily fit into main memory
   /// for hierarchical iterators: locks current node only
   virtual const PointSet * lockBlockRead() = 0;
   /// free locked PointSet again
   virtual void unlock() = 0;
};


/// Additional interface to access point clouds by index: adds ability to retrieve the index of the current point.
/// Typically used in conjunction with BasicPointCloudIterator.
class IndexedPointCloudIterator { /* INTERFACE CLASS */
 public:
   virtual ~IndexedPointCloudIterator() { }

	/// get number of points in the current block (as accessed by nextBlock() or in a node if beeing a hierarhical iterator)
   virtual card64 getNumPointsInBlock() const = 0; 
   /// get the index of the current point within the current block (current node in the hierarchical case)
   virtual card64 getCurrentPointIndex() const = 0;
};


/// Additional interface to access point clouds by index: adds ability to navigate to a previosly retrieved index
/// Typically used in conjunction with BasicPointCloudIterator and IndexedPointCloudIterator
class RandomAccessPointCloudIterator { /* INTERFACE CLASS */
 public:
   virtual ~RandomAccessPointCloudIterator() { }

   /// move iterator position to a certain position (within current block/node)
   virtual void goToIndex(card64 i) = 0;
};


/** Basic point cloud iterator with additional capability for modifying point attributes */
class X3GEOMETRY_API ModifyablePointCloudIterator { /* INTERFACE CLASS */
 public:
   virtual ~ModifyablePointCloudIterator() { }

   /// set attributes in different formats
   virtual void set1ub(const AAT &token, const card8 &x)     const = 0;
   virtual void set2ub(const AAT &token, const Vector2ub &v) const = 0;
   virtual void set3ub(const AAT &token, const Vector3ub &v) const = 0;
   virtual void set4ub(const AAT &token, const Vector4ub &v) const = 0;

   virtual void set1i(const AAT &token, const int32 &x)     const = 0;
   virtual void set2i(const AAT &token, const Vector2i &v) const = 0;
   virtual void set3i(const AAT &token, const Vector3i &v) const = 0;
   virtual void set4i(const AAT &token, const Vector4i &v) const = 0;

   virtual void set1u(const AAT &token, const card32 &x)     const = 0;
   virtual void set2u(const AAT &token, const Vector2u &v) const = 0;
   virtual void set3u(const AAT &token, const Vector3u &v) const = 0;
   virtual void set4u(const AAT &token, const Vector4u &v) const = 0;

   virtual void set1f(const AAT &token, const float32 &x)  const = 0;
   virtual void set2f(const AAT &token, const Vector2f &v) const = 0;
   virtual void set3f(const AAT &token, const Vector3f &v) const = 0;
   virtual void set4f(const AAT &token, const Vector4f &v) const = 0;

   /// ask VertexDescriptor for size
   virtual void setVertex(const char *data) const = 0;

	/// access raw data of current data block for read-write access;
	///CHANGE_NOFLAG_ATTRIBS_ONLY means, that you have to guarantee, that position and flags will no be changed
	enum LockRWHint {CHANGE_EVERYTHING, CHANGE_NOFLAG_ATTRIBS_ONLY};
	/// call unlock to free locked PointSet
   virtual PointSet *lockBlockReadWrite(LockRWHint hint = CHANGE_EVERYTHING) = 0;

};



/** Adds the capability of inserting and deleting points (esp.: modifying positions via insert/delete sequences) */
class X3GEOMETRY_API DynamicPointCloudIterator { /* INTERFACE CLASS */
 public:
   virtual ~DynamicPointCloudIterator() { }

   /** adds single point. data points to vertex data with size according to the
       vertex descriptor of the point cloud. */
   virtual void add(const char *newPoint) = 0;

   /** adds set of points */
   virtual void add(const PointSet *newPoints) = 0;


   /** deletes point at current position of the iterator */
   virtual void deleteCurrent() = 0;

   /** deletes a set of points according to the currently locked point set.
       the point set must be currently locked in read or read-write mode.
       the method causes the point set to be unlocked again */
   // virtual void deleteIndexedSetAndUnlock(const std::vector<card32> &pointSetIndices) = 0;
};



/** Adds the capability of resizing a parametric point cloud */
class X3GEOMETRY_API DynamicParametricPointCloudIterator { /* INTERFACE CLASS */
 public:
   virtual ~DynamicParametricPointCloudIterator() { }

   /** set width and height of parametric grid, invalidates all data */
   virtual void setSize(card32 width, card32 height) = 0;
};



/** Adds the capability of navigation on a parametric grid */
class X3GEOMETRY_API ParametricPointCloudIterator { /* INTERFACE CLASS */
 public:
   virtual ~ParametricPointCloudIterator() { }

   /** go to parametric position */
   virtual void goTo(int32 x, int32 y) = 0;

   /** get width of parametric grid */
   virtual card32 getWidth() = 0;

   /** get height of parametric grid */
   virtual card32 getHeight() = 0;
};

typedef card32 PointIndex;

/** Adds topological information */
class X3GEOMETRY_API TopologyIterator /* INTERFACE CLASS */
{
public:
	virtual ~TopologyIterator() { }

	/// returns number of neighbours for current point
	virtual card32 getNumNeighbours() = 0;

	/// returns an index of a neighbour point
	virtual PointIndex getNeighbour( card32 nIndex ) = 0;


};


class RegularVolume;

/** volumetric navigation */
class X3GEOMETRY_API VolumetricIterator /* INTERFACE CLASS */
{
public:
	virtual ~VolumetricIterator() { }

	/// goto voxel (x,y,z)
	virtual void goTo(card32 x, card32 y, card32 z) = 0;

	/// get extends
	virtual Vector3i getVolumeSize() = 0;

   /// get spatial bounding box
	virtual BoundingBox3f getVolumeRegion() = 0;
	
	/// direct access to the volume object
	virtual const RegularVolume *lockVolumeBlockRead() = 0;

	/// direct access to the volume object (modifyable)
	virtual RegularVolume *lockVolumeBlockReadWrite() = 0;

	/// write back modified volume
	virtual void unlockVolume() = 0;

};


}

#endif
