#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PolygonTriangulation.h"
#include "MathTools.h"
#include "Plane3f.h"
#include "BoundingBox.inline.h"
//---------------------------------------------------------------------------
using namespace X4;

BoundingBox3f expandBB(const BoundingBox2f &bb)
{
	BoundingBox3f ret(NULL_VECTOR3F);
	for( mpcard i=0;i<2;i++)
	{
		ret.lowerCorner[i] = bb.lowerCorner[i];
		ret.upperCorner[i] = bb.upperCorner[i];
	}
	return ret;
}

Vector3f PolygonTriangulation::expand2To3Custom(const Vector2f &v, float x2)
{
	return makeVector3f(v[0],v[1],x2);
}

/// quadtree node 
struct TtriangularizeComplexPolygon_Node
{
	void cutEdge(const Vector2f &p0, const Vector2f &p1, float &start, float &stop);
	BoundingBox2f  myBox;	// bounding box in 2d domain
	std::vector<mpcard> points;	// points within quad
	std::vector<mpcard> edges;	// edges within quad
	mpcard level;			// level of subdivision

	// for empty nodes: define a callback corner that a non-empty leaf node will
	// test for inside/outside and report the result to inside (0 outside, 1 inside, 2 undefined)
	mpcard inside;
	std::vector<TtriangularizeComplexPolygon_Node*> insideCallbacks;
	Vector2f callbackCorner;
};

void TtriangularizeComplexPolygon_Node::cutEdge(const Vector2f &p0, const Vector2f &p1, float &start, float &stop)
{
	Vector2f edgeDir = p1-p0;
	// cut line with half-spaces defined by bounding box
	for( mpcard i=0;i<4;i++ )
	{
		// setup half-space
		Vector2f c = (i&1)?myBox.lowerCorner:myBox.upperCorner;
		Vector2f dir = NULL_VECTOR2F;
		dir[i/2] = 1;
		if( dir*(myBox.getCenter()-c) < 0)
			dir*= -1;

		// intersect
		float t = edgeDir*dir;
		if( fabs(t)<1e-10f )
		{
			if( dir*(p0-c) <= 0 )
				start = stop+1;
			continue;
		}
		t = (dir * c - p0*dir) / t;
		if( dir * edgeDir < 0)
		{
			stop = min(stop,t);
		}
		else
		{
			start = max(start,t);
		}
	}
}


struct TtriangularizeComplexPolygon_point
{
	TtriangularizeComplexPolygon_point( const Vector2f &v, const Vector2f &center)
	{
		point = v;
		angle = atan2(v[1]-center[1],v[0]-center[0])+M_PI;
	}
	float angle;
	Vector2f point;
	bool operator < (const TtriangularizeComplexPolygon_point &p) const{return angle < p.angle;}
};

/// triangularization of a complex polygon. edges are specified
/// by successive vertices edge = (edges[i*2] to edges[i*2+1])
/// make sure that (edges[i*2+1] - edges[i*2]).crossProduct(polyNormal) faces inside
/// this is NOT a line strip!

std::vector<Vector3f> PolygonTriangulation::triangularizeComplexPolygon( const std::vector<Vector3f> &edges, const Vector3f polyNormal )
{
	std::vector<Vector3f> ret;
	if( edges.size() == 0)
		return ret;

	mpcard maxLevel = 15;

	// compute bounding box
	BoundingBox3f bb(edges.front());
	for( mpcard i=1;i<edges.size();i++ )
	{
		bb.addPoint(edges[i]);
	}
	Plane3f p;
	p.setNormal(polyNormal);
	p.setPoint(edges.front());
	Vector3f projBBCenter = p.projectPointOntoPlane(bb.getCenter());
	
	// define tangent frame
	Matrix4f localToWorld = makeTranslation4f(projBBCenter) * expand3To4(calcTangentSystem(-normalize(polyNormal))) * makeRotZ4f(0.1f);
	Matrix4f worldToLocal = invertFrame(localToWorld);

	// transform points into tangent frame
	std::vector<Vector2f> edges_2d(edges.size());
	BoundingBox2f bb_2d;
	for( mpcard i=0;i<edges.size();i++ )
	{
		edges_2d[i] = shrink3To2(transformVector3f(worldToLocal,edges[i])); 
		if( i == 0)
			bb_2d.lowerCorner = bb_2d.upperCorner = edges_2d[i];
		else
			bb_2d.addPoint(edges_2d[i]);
	}
// 	for( mpcard i=0;i<edges.size()/2;i++)
// 		debugRenderer->addLine(expand2To3(edges_2d[i*2],0),expand2To3(edges_2d[i*2+1],0),makeVector3f(1,1,1),makeVector3f(1,1,1)*0.25f,3);

	// define start node
	TtriangularizeComplexPolygon_Node root;
	for( mpcard i=0;i<edges.size();i++ )
	{
		if( i < edges.size() / 2)
			root.edges.push_back(i);
		root.points.push_back(i);
	}
	root.myBox = bb_2d;
	root.myBox.addBorder(bb_2d.getMaxSideLength()*0.1f);	// make sure all points are inside
	root.level = 0;

	// start quadtree subdivision
	list<TtriangularizeComplexPolygon_Node> nodes, emptyNodes; 
	nodes.push_back(root);
	while (!nodes.empty())
	{
		TtriangularizeComplexPolygon_Node node = nodes.back();
		nodes.pop_back();

		static int test = 0;
// 			if( test < 600 )
// 			{
// 				debugRenderer->beginRenderJob_OneFrame("sldjkfjlk",test++);
// 				for( mpcard i=0;i<node.edges.size();i++ )
// 				{
// 					mpcard ei = node.edges[i];
// 					Vector2f p0 = edges_2d[ei*2];
// 					Vector2f p1 = edges_2d[ei*2+1];
// 					debugRenderer->addLine(expand2To3(p0,0),expand2To3(p1,0));
// 				}
// 				for( mpcard i=0;i<node.points.size();i++ )
// 					debugRenderer->addPoint(expand2To3(edges_2d[node.points[i]],0),makeVector3f(1,0,0));
// 				debugRenderer->addBoundingBox(expandBB(node.myBox),makeVector3f(0.25f,1,0.25f)*0.5f,1);
// 				debugRenderer->endRenderJob();
// 			}
// 			else
// 				return ret;


		// ------------------------------------
		// corner case: quad contains a simple corner (points=2 means 1 point actually)
		if(node.points.size() == 2 && node.edges.size() == 2)
		{
			// compute line/box intersections
			std::vector<TtriangularizeComplexPolygon_point> tmpPoints;
			std::vector<Vector2f> edgeNormal;
			Vector2f center = edges_2d[node.points.front()];
			for( mpcard l=0;l<2;l++ )
			{
				mpcard ei = node.edges[l];
				Vector2f p0 = edges_2d[ei*2];
				Vector2f p1 = edges_2d[ei*2+1];
				float start=0,stop=1;
				node.cutEdge(p0,p1,start,stop);
				Vector2f en = p1-p0;
				swap(en[0],en[1]);
				en[1] *= -1;
				if( p1 != center && p1 != edges_2d[node.points.back()])
				{
					tmpPoints.push_back(TtriangularizeComplexPolygon_point(p0+(p1-p0)*stop,center));
					edgeNormal.push_back(en);
				}
				else
				{
					tmpPoints.insert(tmpPoints.begin(),TtriangularizeComplexPolygon_point(p0+(p1-p0)*start,center));
					edgeNormal.insert(edgeNormal.begin(),en);
				}
			}
			Vector2f startPoint = tmpPoints[1].point;
			float angleDiff = tmpPoints[1].angle - tmpPoints[0].angle;
			bool convex = (angleDiff <= (float)M_PI && angleDiff>=0) || (angleDiff <= -(float)M_PI);

			// gather included corners
			for(mpcard i=0;i<4;i++ )
			{
				Vector2f corner = node.myBox.getCorner(i);
				if( convex )
				{
					if( (corner-center)*edgeNormal[0] >= 0 && (corner-center)*edgeNormal[1] >= 0 )
						tmpPoints.push_back(TtriangularizeComplexPolygon_point(corner,center));
				}
				else
				{
					if( (corner-center)*edgeNormal[0] >= 0 || (corner-center)*edgeNormal[1] >= 0 )
						tmpPoints.push_back(TtriangularizeComplexPolygon_point(corner,center));
				}
			}

			sort(tmpPoints.begin(),tmpPoints.end());
			// search for start point
			mpcard startPointIndex=0;
			for( mpcard i=0;i<tmpPoints.size();i++ )
			{
				if( tmpPoints[i].point == startPoint)
				{
					startPointIndex = i;
					break;
				}
			}

			// output triangles
			for( mpcard i=1;i<tmpPoints.size();i++ )
			{
				mpcard i0 = (startPointIndex + i) % tmpPoints.size();
				mpcard i1 = (i0+1) % tmpPoints.size();
				ret.push_back(transformVector3f(localToWorld,expand2To3Custom(center,0)));
				ret.push_back(transformVector3f(localToWorld,expand2To3Custom(tmpPoints[i1].point,0)));
				ret.push_back(transformVector3f(localToWorld,expand2To3Custom(tmpPoints[i0].point,0)));
			}

			// evaluate callbacks
			for( mpcard i=0;i<node.insideCallbacks.size();i++ )
			{
				Vector2f callBackCorner = node.insideCallbacks[i]->callbackCorner;
				if( convex )
				{
					if( (callBackCorner-center)*edgeNormal[0] >= 0 && (callBackCorner-center)*edgeNormal[1] >= 0 )
					{
						node.insideCallbacks[i]->inside = 1;
					}
				}
				else
				{
					if( (callBackCorner-center)*edgeNormal[0] >= 0 || (callBackCorner-center)*edgeNormal[1] >= 0 )
					{
						node.insideCallbacks[i]->inside = 1;
					}
				}
			}

			continue;
		}

		// ------------------------------------
		// edge case: quad contains only one edge
		if(node.edges.size() == 1 )//&& node.points.empty())
		{
			Vector2f p0 = edges_2d[node.edges.front()*2];
			Vector2f p1 = edges_2d[node.edges.front()*2+1];
			float start = 0;
			float stop = 1;
			Vector2f edgeDir = p1-p0;
			Vector2f edgeNormal = edgeDir;
			swap(edgeNormal[0],edgeNormal[1]);
			edgeNormal[1] *= -1.0f;

			// cut edge with box
			node.cutEdge(p0,p1,start,stop);

			// construct center point on edge center
			Vector2f center = p0 + edgeDir*(start+stop)*0.5f;

			// gather intersections / corners
			std::vector<TtriangularizeComplexPolygon_point> tmpPoints;
			tmpPoints.push_back(TtriangularizeComplexPolygon_point(p0 + (p1-p0)*start,center));
			tmpPoints.push_back(TtriangularizeComplexPolygon_point(p0 + (p1-p0)*stop,center));
			for( mpcard i=0;i<4;i++ )
			{
				Vector2f corner = node.myBox.getCorner(i);
				if( edgeNormal*(corner-p0) >= 0 )
					tmpPoints.push_back(TtriangularizeComplexPolygon_point(corner,center));
			}
			sort(tmpPoints.begin(),tmpPoints.end());

			// output triangles
			for( mpcard i=2;i<tmpPoints.size();i++ )
			{
				ret.push_back(transformVector3f(localToWorld,expand2To3Custom(tmpPoints[0].point,0)));
				ret.push_back(transformVector3f(localToWorld,expand2To3Custom(tmpPoints[i].point,0)));
				ret.push_back(transformVector3f(localToWorld,expand2To3Custom(tmpPoints[i-1].point,0)));
			}

			// check if there are any callbacks waiting
			for( mpcard i=0;i<node.insideCallbacks.size();i++ )
			{
				TtriangularizeComplexPolygon_Node * qnode = node.insideCallbacks[i];
				if( edgeNormal * (qnode->callbackCorner-p0) > 1e-10f)
					qnode->inside = 1;
				else
					qnode->inside = 0;
			}

			continue;

		}

		if( node.level >= maxLevel)
			continue;

		// split box
		TtriangularizeComplexPolygon_Node child[4];
		Vector2f splitPoint = node.myBox.getCenter();
		for( mpcard i=0;i<4;i++ )
		{
			child[i].level = node.level+1;
			child[i].myBox = node.myBox;
			for( mpcard d=0;d<2;d++)
			{
				if( ((i >> d) & 1)==0 )
					child[i].myBox.upperCorner[d] = splitPoint[d];
				else
					child[i].myBox.lowerCorner[d] = splitPoint[d];
			}

		}

		// split points
		for( mpcard i=0;i<node.points.size();i++ )
		{
			Vector2f point = edges_2d[node.points[i]] - splitPoint;
			mpcard index = point[0] < 0 ? 0:1;
			index += point[1] < 0 ? 0:2;
			child[index].points.push_back(node.points[i]);
		}

		// split corner call backs
		for( mpcard i=0;i<node.insideCallbacks.size();i++ )
		{
			Vector2f corner = node.insideCallbacks[i]->callbackCorner - splitPoint;
			mpcard index = corner[0] < 0 ? 0:1;
			index += corner[1] < 0 ? 0:2;
			child[index].insideCallbacks.push_back(node.insideCallbacks[i]);
		}

		// distribute edges
		for( mpcard i=0;i<node.edges.size();i++ )
		{
			Vector2f p0 = edges_2d[node.edges[i]*2];
			Vector2f p1 = edges_2d[node.edges[i]*2+1];

			for( mpcard c=0;c<4;c++)
			{
				float start=0;
				float stop=1;
				child[c].cutEdge(p0,p1,start,stop);
				if( start < stop )
					child[c].edges.push_back(node.edges[i]);
			}
		}

		// here begins some fun with special bad poly cases: check for double lines
		for( mpcard i=0;i<4;i++ )
		{
			if( !child[i].points.empty() || child[i].edges.size()!=2)
				continue;

			Vector2f e0_p0 = edges_2d[ child[i].edges[0]*2 ];
			Vector2f e0_p1 = edges_2d[ child[i].edges[0]*2+1 ];
			Vector2f e1_p0 = edges_2d[ child[i].edges[1]*2 ];
			Vector2f e1_p1 = edges_2d[ child[i].edges[1]*2+1 ];
			Vector2f e0_dir = e0_p1-e0_p0;
			Vector2f e1_dir = e1_p1-e1_p0;

			// measure angle/spatial distance
			Vector2f e0_dir_n = normalize(e0_dir);
			Vector2f e1_dir_n = normalize(e1_dir);
			float t = (child[i].myBox.getCenter()-e0_p0)*e0_dir_n;
			Vector2f projCenter = e0_p0+e0_dir_n*t;
			float dist = norm(projCenter - e1_p0 - e1_dir_n*((projCenter-e1_p0)*e1_dir_n));
			float dot_dir = e0_dir_n*e1_dir_n;
			if( dot_dir > 0.99f && (dist < bb.getMaxSideLength()*1e-5f || child[i].level == maxLevel-1))
			{
				child[i].edges.pop_back();
			}
			if( dot_dir < -0.99f && (dist < bb.getMaxSideLength()*1e-5f || child[i].level == maxLevel-1) )
			{
				child[i].inside=3;	// special case...
				child[i].edges.clear();
			}
		}

		// add to queue
		std::vector<TtriangularizeComplexPolygon_Node*> callbacks;
		for( mpcard i=0;i<4;i++ )
		{
			if( child[i].edges.empty() && child[i].points.empty())
			{
				child[i].callbackCorner = splitPoint;
				if( child[i].inside != 3)
					child[i].inside=2;
				emptyNodes.push_back(child[i]);
				callbacks.push_back(&emptyNodes.back());
			}
			else
			{
				nodes.push_back(child[i]);
			}
		}
		if( callbacks.size())
		{
			nodes.back().insideCallbacks.insert(nodes.back().insideCallbacks.end(),callbacks.begin(),callbacks.end());
		}

		// prevent to block too much memory (in case sth went wrong)
		if( (nodes.size() + emptyNodes.size()) * sizeof(TtriangularizeComplexPolygon_Node) > 100*1024*1024 )
		{
			throw X4Exception("something went wrong in PolygonTriangulation::triangularizeComplexPolygon()."
				" Memory usage exceeded 100MB which indicates a serious bug or wrong usage.");
		}
	}

	// handle empty nodes
	while (!emptyNodes.empty())	// ironic isn't it?
	{
		TtriangularizeComplexPolygon_Node node = emptyNodes.back();
		emptyNodes.pop_back();

		if( node.inside==1)
		{
			// spread information to dependent nodes
			for( mpcard i=0;i<node.insideCallbacks.size();i++ )
				node.insideCallbacks[i]->inside = 1;

			std::vector<Vector2f> tmpPoints;
			tmpPoints.push_back(node.myBox.getCorner(0));
			tmpPoints.push_back(node.myBox.getCorner(1));
			tmpPoints.push_back(node.myBox.getCorner(3));
			tmpPoints.push_back(node.myBox.getCorner(2));

			// output triangles
			for( mpcard i=2;i<tmpPoints.size();i++ )
			{
				ret.push_back(transformVector3f(localToWorld,expand2To3Custom(tmpPoints[0],0)));
				ret.push_back(transformVector3f(localToWorld,expand2To3Custom(tmpPoints[i],0)));
				ret.push_back(transformVector3f(localToWorld,expand2To3Custom(tmpPoints[i-1],0)));
			}
		}

	}

	return ret;
}