#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PrecomputedDistanceField.h"
//---------------------------------------------------------------------------
#include "PointCloud.h"
#include "HierarchicalKNNIterator.h"
#include "BoundingBox.h"
#include "MathTools.h"
//---------------------------------------------------------------------------
using namespace X4;


PrecomputedDistanceField::~PrecomputedDistanceField(void)
{
	delete m_Grid;
	delete m_PCIter;
}

/// Return the bounding box of a point cloud (unfortunately this function was moved to scenegraph...  )
BoundingBox3f  getPCBBox_PDF( PointCloud* pc )
{
	BoundingBox3f bb;
	if ( !pc->providesAttribute( "position" ) ) return bb;
	SceneObjectIterator *newIt = pc->createIterator( SOIT( SOIT::CAP_BASIC_PC ) );
	BasicPointCloudIterator *basicIt = dynamic_cast<BasicPointCloudIterator*>(newIt);
	AAT POSITION = pc->getAAT( "position" );
	Matrix4f transformation = pc->getTransformation();
	basicIt->reset();
	bool bbIsEmpty = true;
	while( !basicIt->atEnd() ){
		if ( bbIsEmpty ){
			bb = BoundingBox3f( transformVector3f(transformation, basicIt->get3f( POSITION )) );
			bbIsEmpty = false;
		} else
			bb.addPoint( transformVector3f(transformation, basicIt->get3f( POSITION ) ));

		basicIt->next();
	}
	delete newIt;
	return bb;
}

X4::PrecomputedDistanceField::PrecomputedDistanceField( PointCloud * pc, float gridPrecision, float nearfield )
{
	m_PC = pc;
	m_PCIter = pc->createIterator(SOIT::CAP_BASIC_PC | SOIT::CAP_RANDOM_ACCESS_PC);
	m_PC_bIt = dynamic_cast<BasicPointCloudIterator*>(m_PCIter);
	m_PC_rIt = dynamic_cast<RandomAccessPointCloudIterator*>(m_PCIter);
	m_PosAAT = pc->getAAT("position");
	if( pc->providesAttribute("normal"))
		m_NormalAAT = pc->getAAT("normal");
	else
		m_NormalAAT = NULL_AAT;

	BoundingBox3f bb = getPCBBox_PDF(pc);
	bb.makeBoundingCube();
	bb.addBorder(max(nearfield,gridPrecision) * 2.0f );

	// compute number of levels (root node may be larger

	int totalGridSize = bb.getMaxSideLength() / gridPrecision;
	unsigned numLevels = 1;

	int currentGridSize = totalGridSize;	
	while( currentGridSize > TGrid::MaxRootNodeDimension || numLevels == 1)
	{
		currentGridSize /= TGrid::NodeDimension;
		numLevels++;
	}

	m_Grid = new TGrid(currentGridSize,false);
	m_Grid->bb = bb;

	AAT posAAT = pc->getAAT("position");
	AAT normalAAT = pc->getAAT("normal");
	
	SceneObjectIterator * iter = pc->createIterator(SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC);
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(iter);

	unsigned index = 0;
	while( !bIt->atEnd() )
	{
		Vector3f p = bIt->get3f(posAAT);
		m_Grid->insert(this, p, index, nearfield, numLevels-1);

		bIt->next();
		index++;
	}

	// top level distance field
	HierarchicalKNNIterator hIt(pc,32,NULL);
	unsigned dimension = m_Grid->dimension;
	unsigned numEntries = dimension * dimension * dimension;
	m_Grid->data = new unsigned[numEntries];
	float cellSideLength = m_Grid->bb.getSideLength(0) / dimension;
	for( unsigned i=0;i<numEntries;i++ )
	{
		int x = i % dimension;
		int y = (i / dimension) % dimension;
		int z = (i / dimension / dimension) % dimension;
		Vector3f cellCenter = bb.lowerCorner + makeVector3f((float)x + 0.5f,(float)y + 0.5f,(float)z + 0.5f) * cellSideLength;
		hIt.setSeekPointAndReset(cellCenter);

		m_Grid->data[i] = hIt.getCurrentPointIndex();
	}

}


std::vector<BoundingBox3f> X4::PrecomputedDistanceField::getBBs()
{
	std::vector<BoundingBox3f> ret;
	m_Grid->getBBs(ret);
	return ret;
}

Plane X4::PrecomputedDistanceField::getClosestPointPlane( const Vector3f &pos )
{
	Plane res;
	unsigned index = m_Grid->getClosestPoint(pos);
	m_PC_rIt->goToIndex(index);
	res.point = m_PC_bIt->get3f(m_PosAAT);
	res.normal = m_PC_bIt->get3f(m_NormalAAT);
	return res;
}

X4::Vector3f X4::PrecomputedDistanceField::getPointPosition( mpcard index ) const
{
	m_PC_rIt->goToIndex(index);
	return m_PC_bIt->get3f(m_PosAAT);
}

unsigned X4::PrecomputedDistanceField::getClosestPointIndex( const Vector3f & pos )
{
	return m_Grid->getClosestPoint(pos);
}
void X4::PrecomputedDistanceField::TGrid::insert( PrecomputedDistanceField * df, const Vector3f &p, mpcard pi, float nearfield, int level )
{
	
	if( subgrids )	// is node?
	{
		Vector3f lp = computeLocalVector(p);
		const float bbSideLength = bb.getSideLength(0);
		const float lnearfield = nearfield / bbSideLength;
		const float sub_bbSideLength = bbSideLength / dimension;


		Vector3i minVec = makeVector3i(lp[0] - lnearfield,lp[1] - lnearfield,lp[2] - lnearfield);
		Vector3i maxVec = makeVector3i(lp[0] + lnearfield,lp[1] + lnearfield,lp[2] + lnearfield);
		for( unsigned i=0;i<3;i++ )
		{
			minVec[i] = max(minVec[i],0);
			maxVec[i] = min(maxVec[i],(int)dimension-1);
		}

		for( int x=minVec[0];x<=maxVec[0];x++ )
		{
			for( int y=minVec[1];y<=maxVec[1];y++ )
			{
				for( int z=minVec[2];z<=maxVec[2];z++ )
				{
					unsigned index = x + dimension*(y + dimension*z);
					if( subgrids[index] == NULL )		
					{
						subgrids[index] = new TGrid(NodeDimension, level == 1);
						subgrids[index]->bb.lowerCorner = bb.lowerCorner + makeVector3f(x,y,z) * sub_bbSideLength;
						subgrids[index]->bb.upperCorner = bb.lowerCorner + makeVector3f(x+1,y+1,z+1) * sub_bbSideLength;
					}
					if( level > 0)
						subgrids[index]->insert(df,p,pi, nearfield,level-1);
				
				}
			}
		}

	}

	if( dimension == NodeDimension )
	{
		points.push_back(pi);
	}


//	if(0)
	if( dimension == NodeDimension )
	{
		const unsigned numEntries = dimension*dimension*dimension;

		bool newCell = (data == NULL);
		if( newCell )
			data = new unsigned[numEntries];

		float cellSideLength = bb.getSideLength(0) / dimension;

		for( unsigned i=0;i<numEntries;i++ )
		{
			if( newCell )
				data[i] = pi;
			else
			{
				int x = i % dimension;
				int y = (i / dimension) % dimension;
				int z = (i / dimension / dimension) % dimension;
				Vector3f cellCenter = bb.lowerCorner + makeVector3f((float)x + 0.5f,(float)y + 0.5f,(float)z + 0.5f) * cellSideLength;
				Vector3f storedPoint = df->getPointPosition(data[i]);
					
				float dist_2 = normQuad(cellCenter - storedPoint);
				float current_dist_2 = normQuad(cellCenter - p);
				if( current_dist_2 < dist_2 )
				{
					data[i] = pi;
				}

			}
		}
	}
}

Vector3f X4::PrecomputedDistanceField::TGrid::computeLocalVector( const Vector3f &p ) const
{
	const float mult = dimension / bb.getSideLength(0);
	return (p - bb.lowerCorner) * mult;
}

void X4::PrecomputedDistanceField::TGrid::getBBs( std::vector<BoundingBox3f> &bbs )
{
	bbs.push_back(bb);
	if( subgrids )
	{
		for( unsigned i=0;i<dimension*dimension*dimension;i++ )
		{
			if( subgrids[i] )
				subgrids[i]->getBBs(bbs);
				//bbs.push_back(subgrids[i]->bb);
		}
	}
	//else
	//	bbs.push_back(bb);
}

unsigned X4::PrecomputedDistanceField::TGrid::getClosestPoint( const Vector3f &p )
{
	TGrid * currentCell = this;

	while(true)
	{
		Vector3f lp = currentCell->computeLocalVector(p);
		unsigned index = 0;
		unsigned multiplier = 1;
		float maxEntry = (float)currentCell->dimension-1.0f;
		for( unsigned i=0;i<3;i++ )
		{
			unsigned tmp = clamp(lp[i],0.0f,maxEntry);
			index += tmp * multiplier;
			multiplier *= currentCell->dimension;		
		}

		if( !currentCell->subgrids )	// leaf?
		{
			return currentCell->data[index];
		}
		else
		{
			TGrid * g = currentCell->subgrids[index];
			if( g )
				currentCell = g;
			else
				return currentCell->data[index];
		}
	}

}