//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "RegularVolume.h"
#include "RegularVolume.inline.h"
//---------------------------------------------------------------------------
#include "ObjectClassProperty.h"
#include "FixedArrayClassProperty.h"
#include "NumericalClassProperty.h"
#include "PointSet.inline.h"
#include <fstream>
#include "Stringhelper.h"
#include "BoundingBox.inline.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------


IMPLEMENT_X4_CLASS( RegularVolume, 0 ) {
	BEGIN_CLASS_INIT( RegularVolume )
	ADD_OBJECT_PROP( voxels, 0, PointSet::getClass(), true )
	ADD_CARD32_PROP( width, 0 )
	ADD_CARD32_PROP( height, 0 )
	ADD_CARD32_PROP( depth, 0 )
	ADD_VECTOR3F_PROP( bb.lowerCorner, 0 )
	ADD_VECTOR3F_PROP( bb.upperCorner, 0 )
}

// --

RegularVolume::RegularVolume() {
	width = 0;
	height = 0;
	depth = 0;
	voxels = 0;
	bb = BoundingBox3f(NULL_VECTOR3F);
}

void RegularVolume::clearAndSetup(card32 width, card32 height, card32 depth, const VertexDescriptor *descr, const BoundingBox3f &bb) {
	this->width = width;
	this->height = height;
	this->depth = depth;
	this->bb = bb;
	delete voxels;
	voxels = new PointSet();
	voxels->clearAndSetup(1, width*height*depth, descr);
}

void RegularVolume::computeMinMax1f(std::string channel1f, float32 &min, float32 &max) {
	const card32 n = voxels->getNumEntries();
	AAT aat = voxels->getAAT(channel1f);
	min = voxels->get<float32, 1>(0, aat)[0];
	max = min;
	for (card32 i=1; i<n; i++) {
		float32 v = voxels->get<float32, 1>(i, aat)[0];
		if (v>max) max = v;
		if (v<min) min = v;
	}
}

void RegularVolume::setScalarValue(const AAT channel, card32 x, card32 y, card32 z, float32 value) {
	card32 pos = x + y*width + z*width*height;
	x4Assert(pos < voxels->getNumEntries());
	voxels->set<float32>(pos, channel, value);
}

void RegularVolume::createAttribute(VAD attribute) 
{
	if (!voxels->providesAttribute(attribute.name, attribute.numComponents, attribute.dataFormat)) {
		if (voxels->providesAttribute(attribute.name)) {
			throw X4Exception("RegularInCoreVolume::createAttribute() - channel exists already with different parameters (numComponents/dataFormat)");
		}
		VertexDescriptor newAttribs;
		newAttribs.pushAttrib(attribute);
		PointSet *newV = voxels->appendFields(&newAttribs);
		delete voxels;
		voxels = newV;
	}
}

void RegularVolume::setVoxelData(PointSet *ps, card32 width, card32 height, card32 depth) 
{
   if (voxels != ps) {
      delete voxels;
      voxels = ps;
      this->width = width;
      this->height = height;
      this->depth = depth;
      x4Assert(ps == NULL || width*height*depth == ps->getNumEntries());
   }
}

RegularVolume *RegularVolume::appendFields(const VertexDescriptor *newAttribs) const 
{
   RegularVolume *result = new RegularVolume();
   PointSet *newPS = voxels->appendFields(newAttribs);
   result->clearAndSetup(width, height, depth, newPS->getDescr(), bb);
   result->setVoxelData(newPS, width, height, depth);
   return result;
}

bool RegularVolume::loadFromMHDRaw( const char * filename, bool createPositionChannel , bool is16Bit) 
{
	std::ifstream input( filename );
	if( !input.is_open() )
	{
		return false;
	}

	std::string line;
	int dimX,dimY,dimZ;
	dimX = dimY = dimZ = 0;
	Vector3f elementSpacing = NULL_VECTOR3F;
	std::string rawFilename;
	bool isLittleEndian = false;
	while( getline( input, line) )
	{
		std::vector<std::string> words = splitString( line, " \t" );
		if( words.size() < 3 )
			continue;

		if( words[0] == "NDims")
			if( words[2] != "3" )
				return false;	// no volume data set!

		if( words[0] == "DimSize" )
		{
			if( words.size() < 5)
				return false;

			if( !strToInt( words[2], dimX ) )
				return false;
			if( !strToInt( words[3], dimY ) )
				return false;
			if( !strToInt( words[4], dimZ ) )
				return false;

		}

		if( words[0] == "ElementType")
		{
			if( (words[2] != "MET_USHORT") && (words[2] != "MET_SHORT"))
				return false;	
		}

		if( words[0] == "ElementSpacing" )
		{
			if( words.size() < 5)
				return false;

			if( !strToFloat( words[2], elementSpacing[0] ) )
				return false;
			if( !strToFloat( words[3], elementSpacing[1] ) )
				return false;
			if( !strToFloat( words[4], elementSpacing[2] ) )
				return false;

		}

		if( words[0] == "ElementByteOrderMSB")
		{
			if( words.size() >= 3 )
			{
				
				isLittleEndian = strLCase(words[2]) == "true";
			}
		}

		if( words[0] == "ElementDataFile")
			rawFilename = words[2];
	}
	input.close();
	width = dimX;
	height = dimY;
	depth = dimZ;
	bb.lowerCorner = elementSpacing * -0.5f;
	bb.upperCorner = elementSpacing *  0.5f;
	bb.upperCorner[0] *= dimX;
	bb.upperCorner[1] *= dimY;
	bb.upperCorner[2] *= dimZ;
	bb.lowerCorner[0] *= dimX;
	bb.lowerCorner[1] *= dimY;
	bb.lowerCorner[2] *= dimZ;

	if( rawFilename.length() > 0 && dimX*dimY*dimZ > 0)
	{
		std::string tmpFilename = removeFilenameFromPath(std::string(filename)) + rawFilename;
		std::ifstream rawFile( tmpFilename.c_str(), ios::binary);
		if( !rawFile.is_open())
			return false;

		VertexDescriptor * vd = new VertexDescriptor;
		if( createPositionChannel )
			vd->pushAttrib( VAD("position",3,VAD::DATA_FORMAT_FLOAT32));
		vd->pushAttrib( VAD("intensity",1,VAD::DATA_FORMAT_FLOAT32) );
		delete voxels;
		voxels = new PointSet;
		voxels->clearAndSetup( 1, dimX*dimY*dimZ, vd );

		AAT intAAT = voxels->getAAT( "intensity" );
		AAT posAAT = NULL_AAT;
		if( createPositionChannel )
			posAAT = voxels->getAAT( "position" );


		if(is16Bit)
		{
			std::vector<card16> tmpBuffer(dimX*dimY*dimZ);
			rawFile.read( (char*)&tmpBuffer[0], dimX*dimY*dimZ * sizeof(card16) );
			for(card32 x = 0; x < (card32)dimX; x++)
			{
				for(card32 y = 0; y < (card32)dimY; y++)
				{
					for(card32 z = 0; z < (card32)dimZ; z++)
					{
						card32 i = x+dimX*y+dimX*dimY*z;

						card16 test = tmpBuffer[i];
						//if( isLittleEndian )
						//	test = (test >> 16) + ((test&0xff) << 16);
						voxels->set1f(i, intAAT, (float)test / 655536.0f);

						if( createPositionChannel )
						{
							float32 xf = bb.getCorner(0)[0]+(((x+0.5)*bb.getSideLength(0))/(float32)dimX);
							float32 yf = bb.getCorner(0)[1]+(((y+0.5)*bb.getSideLength(1))/(float32)dimY);
							float32 zf = bb.getCorner(0)[2]+(((z+0.5)*bb.getSideLength(2))/(float32)dimZ);
							voxels->set3f(i, posAAT, makeVector3f(xf, yf, zf));
						}
					}
				}
			}
		}
		else
		{
			std::vector<card8> tmpBuffer(dimX*dimY*dimZ);
			rawFile.read( (char*)&tmpBuffer[0], dimX*dimY*dimZ * sizeof(card8) );
			for(card32 x = 0; x < (card32)dimX; x++)
			{
				for(card32 y = 0; y < (card32)dimY; y++)
				{
					for(card32 z = 0; z < (card32)dimZ; z++)
					{
						card32 i = x+dimX*y+dimX*dimY*z;

						card8 test = tmpBuffer[i];
						//if( isLittleEndian )
						//	test = (test >> 8) + ((test&0xff) << 8);
						voxels->set1f(i, intAAT, (float)test / 256.0f);

						if( createPositionChannel )
						{
							float32 xf = bb.getCorner(0)[0]+(((x+0.5)*bb.getSideLength(0))/(float32)dimX);
							float32 yf = bb.getCorner(0)[1]+(((y+0.5)*bb.getSideLength(1))/(float32)dimY);
							float32 zf = bb.getCorner(0)[2]+(((z+0.5)*bb.getSideLength(2))/(float32)dimZ);
							voxels->set3f(i, posAAT, makeVector3f(xf, yf, zf));
						}
					}
				}
			}
		}
	}
	return true;
}

RegularVolume *RegularVolume::resample( card32 dimX, card32 dimY, card32 dimZ ) 
{
	RegularVolume * ret = new RegularVolume;
	ret->clearAndSetup(dimX,dimY,dimZ, (VertexDescriptor*)voxels->getDescr()->copy(), bb);
	
	std::vector<AAT> tokens;
	std::vector<VAD> attributes;
	for( card32 i=0;i<voxels->getDescr()->getNumAttributes();i++ )
	{
		tokens.push_back( voxels->getDescr()->getToken(i));
		attributes.push_back( voxels->getDescr()->getAttribute(i) );
	}

	
	FOR_XYZ( ret )
	{
		Vector3f sourcePos = makeVector3f( (float)x/(float)dimX*(float)width, 
			(float)y/(float)dimY*(float)height,
			(float)z/(float)dimZ*(float)depth);
		card32 index = (z*dimY + y)*dimX + x;
		for( card32 i=0;i<attributes.size();i++)
		{
			if( attributes[i].dataFormat == VAD::DATA_FORMAT_FLOAT32 && attributes[i].numComponents == 1)
				ret->voxels->set1f(index,tokens[i], interpolateAttribute1f(sourcePos,tokens[i]));
		}

	}
	END_FOR_XYZ();

	return ret;
}

void RegularVolume::assign(const Object* obj, X4CopyContext *context) 
{
	forget(context);
	const RegularVolume *other = dynamic_cast<const RegularVolume*>(obj);
	if (!other) {
		x4Assert(false);
		return;
	}
	this->width = other->width;
	this->height = other->height;
	this->depth = other->depth;
	this->bb = other->bb;
	if (this->voxels) {
		this->voxels->assign(other->voxels);
	} else {
		this->voxels = dynamic_cast<PointSet*>(other->voxels->copy(context));
	}
}

X4::RegularVolume::~RegularVolume()
{
	delete voxels;
}


Vector3f RegularVolume::voxelToWorldPos(const float w, const float h, const float d)
{
//	if((w < 0) || (w >= width) || (h < 0) || (h >= height) || (d < 0) || (d >= depth))
//		throw X4Exception("voxelToWorldPos - at least one component out of bounds.");

	Vector3f ll = bb.getCorner(0);

	Vector3f result;

	result[0] = ll[0] + ((0.5f + w) * (bb.getSideLength(0) / width) );
	result[1] = ll[1] + ((0.5f + h) * (bb.getSideLength(1) / height));
	result[2] = ll[2] + ((0.5f + d) * (bb.getSideLength(2) / depth) );

	return result;
}

Vector3f RegularVolume::worldPosToVoxel(const Vector3f pos)
{
//	if(!bb.inBoundingBox(pos))
//		PException("worldPosToVoxel - at least one component out of bounds.");
	
	Vector3f ll = bb.getCorner(0);

	Vector3f result;

	result[0] = (((pos[0]-ll[0]) * width)  / bb.getSideLength(0)) - 0.5f;
	result[1] = (((pos[1]-ll[1]) * height) / bb.getSideLength(1)) - 0.5f;
	result[2] = (((pos[2]-ll[2]) * depth)  / bb.getSideLength(2)) - 0.5f;

	return result;
}

float RegularVolume::trilinearInterpolation1f(const float vPosX, const float vPosY, const float vPosZ, AAT channel)
{
// 	if((vPosX < 0) || (vPosX >= width) || (vPosY < 0) || (vPosY >= height) || (vPosZ < 0) || (vPosZ >= depth))
// 	{	
// 		debugOutput << "[" << vPosX << "," << vPosY << "," << vPosZ << "]\n"; 
// 		throw X4Exception("trilinearInterpolation1f - at least one component out of bounds.");
// 	}

	// discrete values
	int	xf			= floor(vPosX);
	int	yf			= floor(vPosY);
	int	zf			= floor(vPosZ);
	int	xc			= ceil(vPosX);
	int	yc			= ceil(vPosY);
	int	zc			= ceil(vPosZ);

	// randbedingung (voxelzentrum zum volumenrand = 0.5)
	// aber vorher klren ob es in der boundingbox liegt	
	if(xf < 0)
		xf = 0;
	if(yf < 0)
		yf = 0;
	if(zf < 0)
		zf = 0;
	if(xc > (int)width-1)
		xc = width-1;
	if(yc > (int)height-1)
		yc = height-1;
	if(zc > (int)depth-1)
		zc = depth-1;

	const float	v000		= getVoxel<float,1>(xf,yf,zf,channel)[0]; // left  lower front
	const float	v001		= getVoxel<float,1>(xf,yf,zc,channel)[0]; // left  lower back
	const float	v010		= getVoxel<float,1>(xf,yc,zf,channel)[0]; // left  upper front
	const float	v011		= getVoxel<float,1>(xf,yc,zc,channel)[0]; // left  upper back
	const float	v100		= getVoxel<float,1>(xc,yf,zf,channel)[0]; // right lower front
	const float	v101		= getVoxel<float,1>(xc,yf,zc,channel)[0]; // right lower back
	const float	v110		= getVoxel<float,1>(xc,yc,zf,channel)[0]; // right upper front
	const float	v111		= getVoxel<float,1>(xc,yc,zc,channel)[0]; // right upper back

	// weights
	const float	xd			= vPosX	- xf;
	const float xdInv		= 1.0f	- xd;
	const float	yd			= vPosY	- yf;
	const float ydInv		= 1.0f	- yd;
	const float	zd			= vPosZ	- zf;
	const float zdInv		= 1.0f	- zd;

	// interpolate z first ...
	const float	i1			= (v000 * zdInv) + (v001 * zd);
	const float	i2			= (v010 * zdInv) + (v011 * zd);
	const float	j1			= (v100 * zdInv) + (v101 * zd);
	const float	j2			= (v110 * zdInv) + (v111 * zd);
	// ... then y ...
	const float	w1			= (i1   * ydInv) + (i2   * yd);
	const float	w2			= (j1   * ydInv) + (j2   * yd);
	// then x
	const float	result	= (w1   * xdInv) + (w2   * xd);

	return result;
}

Vector3f RegularVolume::trilinearInterpolation3f(const float vPosX, const float vPosY, const float vPosZ, AAT channel)
{
	if((vPosX < 0) || (vPosX >= width) || (vPosY < 0) || (vPosY >= height) || (vPosZ < 0) || (vPosZ >= depth))
		throw X4Exception("trilinearInterpolation3f - at least one component out of bounds.");

	// discrete values
	int		xf			= floor(vPosX);
	int		yf			= floor(vPosY);
	int		zf			= floor(vPosZ);
	int		xc			= ceil(vPosX);
	int		yc			= ceil(vPosY);
	int		zc			= ceil(vPosZ);

	// randbedingung (voxelzentrum zum volumenrand = 0.5)
	// aber vorher klren ob es in der boundingbox liegt	
	if(xf < 0)
		xf = 0;
	if(yf < 0)
		yf = 0;
	if(zf < 0)
		zf = 0;
	if(xc > (int)width-1)
		xc = width-1;
	if(yc > (int)height-1)
		yc = height-1;
	if(zc > (int)depth-1)
		zc = depth-1;

	const Vector3f v000		= getVoxel<float,3>(xf,yf,zf,channel); // left  lower front
	const Vector3f v001		= getVoxel<float,3>(xf,yf,zc,channel); // left  lower back
	const Vector3f v010		= getVoxel<float,3>(xf,yc,zf,channel); // left  upper front
	const Vector3f v011		= getVoxel<float,3>(xf,yc,zc,channel); // left  upper back
	const Vector3f v100		= getVoxel<float,3>(xc,yf,zf,channel); // right lower front
	const Vector3f v101		= getVoxel<float,3>(xc,yf,zc,channel); // right lower back
	const Vector3f v110		= getVoxel<float,3>(xc,yc,zf,channel); // right upper front
	const Vector3f v111		= getVoxel<float,3>(xc,yc,zc,channel); // right upper back

	// weights
	const float		xd			= vPosX	- xf;
	const float		xdInv		= 1.0f	- xd;
	const float		yd			= vPosY	- yf;
	const float		ydInv		= 1.0f	- yd;
	const float		zd			= vPosZ	- zf;
	const float		zdInv		= 1.0f	- zd;

	// interpolate z first ...
	const Vector3f i1			= (v000 * zdInv) + (v001 * zd);
	const Vector3f i2			= (v010 * zdInv) + (v011 * zd);
	const Vector3f j1			= (v100 * zdInv) + (v101 * zd);
	const Vector3f j2			= (v110 * zdInv) + (v111 * zd);
	// ... then y ...
	const Vector3f w1			= (i1   * ydInv) + (i2   * yd);
	const Vector3f w2			= (j1   * ydInv) + (j2   * yd);
	// then x
	const Vector3f result	= (w1   * xdInv) + (w2   * xd);

	return result;
}

void RegularVolume::threshold1f(AAT channel, const float thresholdHigh, const float thresholdLow)
{
	const int nVoxels = voxels->getNumEntries();

	int index;
	float currentValue;
	for(index = 0; index < nVoxels; index++)
	{
		currentValue = voxels->get1f(index,channel);
		if((currentValue > thresholdHigh) || (currentValue < thresholdLow))
			voxels->set1f(index,channel,0.0f);
	}
}

void RegularVolume::normalizeChannel_1f_0_1(std::string channel)
{
	float min,max;
	computeMinMax1f(channel,min,max);
	const float range = max - min;
	if(range < 1e-7)
	{
		warning("--- RegularVolume::normalizeChannel_1f_0_1: " + channel + "entries do not vary. division by zero ---");
		return;
	}

	AAT aat = voxels->getAAT(channel);
	const int nVoxels = voxels->getNumEntries();
	for(int index = 0; index < nVoxels; index++)
	{
		float currentValue = voxels->get1f(index,aat);
		float newValue = (currentValue - min) / range;
		voxels->set1f(index,aat,newValue);
	}
}


int RegularVolume::transformWHDToIndex(const int w, const int h, const int d)
{
	return (w + (h*width) + (d*width*height));
}

Vector3i RegularVolume::transformIndexToWHD(const int index)
{
	Vector3i result;
	result[0] = index % width;
	result[1] = (index / width) % height;
	result[2] = index / (width * height);
	return result;
}
