//---------------------------------------------------------------------------
#ifndef RegularVolumeH
#define RegularVolumeH
//---------------------------------------------------------------------------
#include "x3geometry.h"
//---------------------------------------------------------------------------
#include "X4Types.h"
#include "BoundingBox.h"
#include "PointSet.h"
#include "VertexDescriptor.h"
//---------------------------------------------------------------------------


namespace X4 {


class X3GEOMETRY_API RegularVolume : public Persistent {
	X4_CLASS(RegularVolume)
 private:
	PointSet *voxels;
	card32 width, height, depth;
	BoundingBox3f bb;
	
 public:
	template<typename FloatTypeSource, unsigned dimSource, typename FloatTypeDest, unsigned dimDest>
	class VoxelMapper {
		typedef StaticVector<FloatTypeDest, dimDest> (*TransferFunction)(void*, const StaticVector<FloatTypeSource, dimSource> &);
	};
	
	RegularVolume();
	~RegularVolume();
	
   virtual void assign(const Object* obj, X4CopyContext *context = NULL);

	void clearAndSetup(card32 width, card32 height, card32 depth, const VertexDescriptor *descr, const BoundingBox3f &bb);
   void createAttribute(VAD attribute);

	const VertexDescriptor *getDescr() const {return (voxels==NULL)?(NULL):(voxels->getDescr());}
	const PointSet *getVoxelData() const {return voxels;}
	PointSet *getVoxelData() {return voxels;}
	void setVoxelData(PointSet *ps, card32 width, card32 height, card32 depth);
	
	card32 getWidth()  const {return width;}
	card32 getHeight() const {return height;}
	card32 getDepth()  const {return depth;}
	int32 getNumVoxels() const {return (width * height * depth);}
	card32 getMinExtends() const {return min(width, min(height, depth));}
	card32 getMaxExtends() const {return max(width, min(height, depth));}
	Vector3i getExtends() const {return makeVector3i(width, height, depth);}
	BoundingBox3f getBoundingBox() const {return bb;}
	void setBoundingBox( const BoundingBox3f &bb ) {this->bb = bb;}

   RegularVolume *appendFields(const VertexDescriptor *newAttribs) const;
	
   template<typename FloatType, int dim>
   inline void setVoxel(card32 x, card32 y, card32 z, const AAT channel, const StaticVector<FloatType, dim> &v);
	template<typename FloatType, int dim>
	inline void setVoxel(int32 index, const AAT channel, const StaticVector<FloatType, dim> &v);
   template<typename FloatType, int dim>
   inline void setVoxelChecked(int32 x, int32 y, int32 z, const AAT channel, const StaticVector<FloatType, dim> &v);
   template<typename FloatType>
   inline void setVoxel(card32 x, card32 y, card32 z, const AAT channel, const FloatType &v);
   template<typename FloatType>
	inline void setVoxel(int32 index, const AAT channel, const FloatType &v);
	template<typename FloatType>
   inline void setVoxelChecked(int32 x, int32 y, int32 z, const AAT channel, const FloatType &v);
   inline void setFullVoxel(card32 x, card32 y, card32 z, const char *buffer);
   inline void setFullVoxelChecked(int32 x, int32 y, int32 z, const char *buffer);

   template<typename FloatType, int dim>
   inline StaticVector<FloatType, dim> getVoxel(card32 x, card32 y, card32 z, const AAT channel) const;
	template<typename FloatType, int dim>
	inline StaticVector<FloatType, dim> getVoxel(int32 index, const AAT channel) const;
   template<typename FloatType, int dim>
   inline StaticVector<FloatType, dim> getVoxelChecked(int32 x, int32 y, int32 z, const AAT channel) const;
   template<typename FloatType>
   inline FloatType getVoxel(card32 x, card32 y, card32 z, const AAT channel) const;
	template<typename FloatType>
	inline FloatType getVoxel(int32 index, const AAT channel) const;
   template<typename FloatType>
   inline FloatType getVoxelChecked(int32 x, int32 y, int32 z, const AAT channel) const;
   inline const char *getFullVoxel(card32 x, card32 y, card32 z) const;
   inline const char *getFullVoxelChecked(int32 x, int32 y, int32 z) const;

	void setScalarValue(const AAT channel, card32 x, card32 y, card32 z, float32 value);
	
	/// write data to volume in width-first, height-second, depth-third order
	template<typename FloatType, unsigned dim>
	void setupChannel(const std::vector< StaticVector<FloatType, dim> > data, std::string channelName, 
						   card32 startWidth  = 0xFFFFFFFF,  card32 endWidth  = 0xFFFFFFFF,
						   card32 startHeight = 0xFFFFFFFF, card32 endHeight = 0xFFFFFFFF,
						   card32 startDepth  = 0xFFFFFFFF,  card32 endDepth  = 0xFFFFFFFF);
	/// get data from volume in width-first, height-second, depth-third order
	template<typename FloatType, unsigned dim>
	void getChannel(std::vector< StaticVector<FloatType, dim> > &data, std::string channelName, 
			 			 card32 startWidth  = 0xFFFFFFFF,  card32 endWidth  = 0xFFFFFFFF,
						 card32 startHeight = 0xFFFFFFFF, card32 endHeight = 0xFFFFFFFF,
						 card32 startDepth  = 0xFFFFFFFF,  card32 endDepth  = 0xFFFFFFFF);
	/// get data from volume and std::map it to a color
	template<typename FloatTypeSource, unsigned dimSource, typename FloatTypeDest, unsigned dimDest>
	void mapChannel(std::vector< StaticVector<FloatTypeDest, dimDest> > &data, std::string channelName, 
						 typename VoxelMapper<FloatTypeSource, dimSource, 
						 FloatTypeDest, dimDest>::TransferFunction mapping, void *mappingContext,
			 			 card32 startWidth  = 0xFFFFFFFF,  card32 endWidth  = 0xFFFFFFFF,
						 card32 startHeight = 0xFFFFFFFF, card32 endHeight = 0xFFFFFFFF,
						 card32 startDepth  = 0xFFFFFFFF,  card32 endDepth  = 0xFFFFFFFF);
	void computeMinMax1f(std::string channel1f, float32 &min, float32 &max);

	bool loadFromMHDRaw( const char * filename, bool createPositionChannel, bool is16Bit);

	RegularVolume * resample( card32 dimX, card32 dimY, card32 dimZ );

	Vector3f voxelToWorldPos(const float w, const float h, const float d);
	Vector3f worldPosToVoxel(const Vector3f pos);
	float trilinearInterpolation1f(const float vPosX, const float vPosY, const float vPosZ, AAT channel);
	Vector3f trilinearInterpolation3f(const float vPosX, const float vPosY, const float vPosZ, AAT channel);
	void threshold1f(AAT channel, const float thresholdHigh, const float thresholdLow);
	void normalizeChannel_1f_0_1(std::string channel);
	int transformWHDToIndex(const int w, const int h, const int d);
	Vector3i transformIndexToWHD(const int index);



	/// trilinear interpolation, pos in ( (0,0,0), (dimX,dimY,dimZ) )
	template<class DataType, unsigned int numDim>
	StaticVector<DataType,numDim> interpolateAttribute( Vector3f pos, AAT aat ) const;

	float interpolateAttribute1f(Vector3f pos, AAT aat) const
	{
		return interpolateAttribute<float,1>(pos,aat)[0];
	}
	Vector2f interpolateAttribute2f(Vector3f pos, AAT aat) const
	{
		return interpolateAttribute<float,2>(pos,aat);
	}
	Vector3f interpolateAttribute3f(Vector3f pos, AAT aat) const
	{
		return interpolateAttribute<float,3>(pos,aat);
	}
	Vector4f interpolateAttribute4f(Vector3f pos, AAT aat) const
	{
		return interpolateAttribute<float,4>(pos,aat);
	}
	
};

#define FOR_XYZ( regularVolume ) \
	{\
	const card32 maxDimX_RegularVolume = regularVolume->getWidth();\
	const card32 maxDimY_RegularVolume = regularVolume->getHeight();\
	const card32 maxDimZ_RegularVolume = regularVolume->getDepth();\
	for( card32 z=0;z<maxDimZ_RegularVolume;z++ ) \
		for( card32 y=0;y<maxDimY_RegularVolume;y++ ) \
			for( card32 x=0;x<maxDimX_RegularVolume;x++ ) {

#define END_FOR_XYZ() \
			}}


#define END_FOR_XYZ() \
	}}



}



#endif
