//---------------------------------------------------------------------------
#ifndef RegularVolumeTemplCodeH
#define RegularVolumeTemplCodeH
//---------------------------------------------------------------------------
#include "x3geometry.h"
//---------------------------------------------------------------------------
#include "RegularVolume.h"
//---------------------------------------------------------------------------


namespace X4 {


template<typename FloatType, unsigned dim>
void RegularVolume::setupChannel(const std::vector< StaticVector<FloatType, dim> > data, std::string channelName, 
	card32 startWidth,  card32 endWidth,
	card32 startHeight, card32 endHeight,
	card32 startDepth,  card32 endDepth) 
{
	if (startWidth == 0xFFFFFFFF) startWidth = 0;
	if (endWidth == 0xFFFFFFFF) endWidth = width-1;
	x4Assert(startWidth < width);
	x4Assert(endWidth < width);
	if (startHeight == 0xFFFFFFFF) startHeight = 0;
	if (endHeight == 0xFFFFFFFF) endHeight = height-1;
	x4Assert(startHeight < height);
	x4Assert(endHeight < height);
	if (startDepth == 0xFFFFFFFF) startDepth = 0;
	if (endDepth == 0xFFFFFFFF) endDepth = depth-1;
	x4Assert(startDepth < depth);
	x4Assert(endDepth < depth);
	
	if (!data.size() == width*height*depth) {
		throw ERangeCheck("RegularVolume::setupChannel() - data size does not match volume.");
	}
	AAT aat = voxels->getAAT(channelName);
	if (!voxels->getDescr()->getAttrib(channelName)->getSize() == sizeof(StaticVector<FloatType, dim>)) {
		throw ERangeCheck("RegularVolume::setupChannel() - data format does not match channel format.");
	}
	
	char *pos = voxels->getDataPtr();
	pos += aat.getOffset();
	card32 i=0;
	card32 step = voxels->getDescr()->getSize();
	for (card32 z=startDepth; z<=endDepth; z++) {
		for (card32 y=startHeight; y<=endHeight; y++) {
			for (card32 x=startWidth; x<=endWidth; x++) {
				StaticVector<FloatTypeSource, dimSource> *readPos = (StaticVector<FloatTypeSource, dimSource>*)(pos+step*(x+y*width+z*width*height));
				*writePos = data[i];
				i++;
			}
		}
	}
}


template<typename FloatTypeSource, unsigned dimSource, typename FloatTypeDest, unsigned dimDest>
void RegularVolume::mapChannel(
	std::vector< StaticVector<FloatTypeDest, dimDest> > &data, std::string channelName, 
	typename VoxelMapper<FloatTypeSource, dimSource, 
	FloatTypeDest, dimDest>::TransferFunction mapping, void *mappingContext,
	card32 startWidth,  card32 endWidth,
	card32 startHeight, card32 endHeight,
	card32 startDepth,  card32 endDepth)
{
	if (startWidth == 0xFFFFFFFF) startWidth = 0;
	if (endWidth == 0xFFFFFFFF) endWidth = width-1;
	x4Assert(startWidth < width);
	x4Assert(endWidth < width);
	if (startHeight == 0xFFFFFFFF) startHeight = 0;
	if (endHeight == 0xFFFFFFFF) endHeight = height-1;
	x4Assert(startHeight < height);
	x4Assert(endHeight < height);
	if (startDepth == 0xFFFFFFFF) startDepth = 0;
	if (endDepth == 0xFFFFFFFF) endDepth = depth-1;
	x4Assert(startDepth < depth);
	x4Assert(endDepth < depth);
	
	if (!data.size() == width*height*depth) {
		throw ERangeCheck("RegularVolume::mapChannel() - data size does not match volume.");
	}
	AAT aat = voxels->getAAT(channelName);
	if (!(voxels->getDescr()->getAttribute(channelName).getSize() == sizeof(StaticVector<FloatTypeSource, dimSource>))) {
		throw ERangeCheck("RegularVolume::mapChannel() - data format does not match channel format.");
	}
	
	char *pos = voxels->getDataPointer();
	pos += aat.getOffset();
	card32 i=0;
	card32 step = voxels->getDescr()->getSize();
	for (card32 z=startDepth; z<=endDepth; z++) {
		for (card32 y=startHeight; y<=endHeight; y++) {
			for (card32 x=startWidth; x<=endWidth; x++) {
				StaticVector<FloatTypeSource, dimSource> *readPos = (StaticVector<FloatTypeSource, dimSource>*)(pos+step*(x+y*width+z*width*height));
				data[i] = mapping(mappingContext, *readPos);
				i++;
			}
		}
	}
}
					 



template<typename FloatType, unsigned dim>
void RegularVolume::getChannel(std::vector< StaticVector<FloatType, dim> > &data, std::string channelName,
	card32 startWidth,  card32 endWidth,
	card32 startHeight, card32 endHeight,
	card32 startDepth,  card32 endDepth) 
{
	if (startWidth == 0xFFFFFFFF) startWidth = 0;
	if (endWidth == 0xFFFFFFFF) endWidth = width-1;
	x4Assert(startWidth < width);
	x4Assert(endWidth < width);
	if (startHeight == 0xFFFFFFFF) startHeight = 0;
	if (endHeight == 0xFFFFFFFF) endHeight = height-1;
	x4Assert(startHeight < height);
	x4Assert(endHeight < height);
	if (startDepth == 0xFFFFFFFF) startDepth = 0;
	if (endDepth == 0xFFFFFFFF) endDepth = depth-1;
	x4Assert(startDepth < depth);
	x4Assert(endDepth < depth);

	data.resize(width*height*depth);
	AAT aat = voxels->getAAT(channelName);
	if (!voxels->getDescr()->getAttrib(channelName)->getSize() == sizeof(StaticVector<FloatType, dim>)) {
		throw ERangeCheck("RegularVolume::getChannel() - data format does not match channel format.");
	}
	
	char *pos = voxels->getDataPtr();
	pos += aat.getOffset();
	card32 i=0;
	card32 step = voxels->getDescr()->getSize();
	for (card32 z=startDepth; z<=endDepth; z++) {
		for (card32 y=startHeight; y<=endHeight; y++) {
			for (card32 x=startWidth; x<=endWidth; x++) {
				StaticVector<FloatTypeSource, dimSource> *readPos = (StaticVector<FloatTypeSource, dimSource>*)(pos+step*(x+y*width+z*width*height));
				data[i] = *readPos;
				i++;
			}
		}
	}
}


template<typename FloatType, int dim>
inline void RegularVolume::setVoxel(card32 x, card32 y, card32 z, const AAT channel, const StaticVector<FloatType, dim> &v) 
{
   voxels->set(x + y*width + z*width*height, channel, v);
}

template<typename FloatType, int dim>
inline void RegularVolume::setVoxel(int32 index, const AAT channel, const StaticVector<FloatType, dim> &v) 
{
	voxels->set(index, channel, v);
}

template<typename FloatType, int dim>
inline void RegularVolume::setVoxelChecked(int32 x, int32 y, int32 z, const AAT channel, const StaticVector<FloatType, dim> &v) 
{
   if (x>=(int32)width) x=width-1;
   if (x<0) x=0;
   if (y>=(int32)height) y=height-1;
   if (y<0) x=0;
   if (z>=(int32)depth) z=depth-1;
   if (z<0) z=0;
   voxels->set(x + y*width + z*width*height, channel, v);
}

template<typename FloatType>
inline void RegularVolume::setVoxel(card32 x, card32 y, card32 z, const AAT channel, const FloatType &v) 
{
   voxels->set(x + y*width + z*width*height, channel, v);
}

template<typename FloatType>
inline void RegularVolume::setVoxel(int32 index, const AAT channel, const FloatType &v) 
{
	voxels->set(index, channel, v);
}

template<typename FloatType>
inline void RegularVolume::setVoxelChecked(int32 x, int32 y, int32 z, const AAT channel, const FloatType &v)
{
   if (x>=(int32)width) x=width-1;
   if (x<0) x=0;
   if (y>=(int32)height) y=height-1;
   if (y<0) x=0;
   if (z>=(int32)depth) z=depth-1;
   if (z<0) z=0;
   voxels->set(x + y*width + z*width*height, channel, v);
}

inline void RegularVolume::setFullVoxelChecked(int32 x, int32 y, int32 z, const char *buffer)
{
   if (x>=(int32)width) x=width-1;
   if (x<0) x=0;
   if (y>=(int32)height) y=height-1;
   if (y<0) x=0;
   if (z>=(int32)depth) z=depth-1;
   if (z<0) z=0;
   voxels->setVertex(x + y*width + z*width*height, buffer);
}

inline void RegularVolume::setFullVoxel(card32 x, card32 y, card32 z, const char *buffer)
{
   voxels->setVertex(x + y*width + z*width*height, buffer);
}




template<typename FloatType, int dim>
inline StaticVector<FloatType, dim> RegularVolume::getVoxel(card32 x, card32 y, card32 z, const AAT channel) const
{
   return voxels->get<FloatType, dim>(x + y*width + z*width*height, channel);
}

template<typename FloatType, int dim>
inline StaticVector<FloatType, dim> RegularVolume::getVoxel(int32 index, const AAT channel) const
{
	return voxels->get<FloatType, dim>(index, channel);
}

template<typename FloatType, int dim>
inline StaticVector<FloatType, dim> RegularVolume::getVoxelChecked(int32 x, int32 y, int32 z, const AAT channel) const
{
   if (x>=(int32)width) x=width-1;
   if (x<0) x=0;
   if (y>=(int32)height) y=height-1;
   if (y<0) x=0;
   if (z>=(int32)depth) z=depth-1;
   if (z<0) z=0;
   return voxels->get<FloatType, dim>(x + y*width + z*width*height, channel);
}

template<typename FloatType>
inline FloatType RegularVolume::getVoxel(card32 x, card32 y, card32 z, const AAT channel) const
{
   return voxels->get<FloatType>(x + y*width + z*width*height, channel);
}

template<typename FloatType>
inline FloatType RegularVolume::getVoxel(int32 index, const AAT channel) const
{
	return voxels->get<FloatType>(index, channel);
}

template<typename FloatType>
inline FloatType RegularVolume::getVoxelChecked(int32 x, int32 y, int32 z, const AAT channel) const
{
   if (x>=(int32)width) x=width-1;
   if (x<0) x=0;
   if (y>=(int32)height) y=height-1;
   if (y<0) x=0;
   if (z>=(int32)depth) z=depth-1;
   if (z<0) z=0;
   return voxels->get<FloatType>(x + y*width + z*width*height, channel);
}


inline const char *RegularVolume::getFullVoxel(card32 x, card32 y, card32 z) const
{
   return voxels->getDataPointer(x + y*width + z*width*height);
}

inline const char *RegularVolume::getFullVoxelChecked(int32 x, int32 y, int32 z) const
{
   if (x>=(int32)width) x=width-1;
   if (x<0) x=0;
   if (y>=(int32)height) y=height-1;
   if (y<0) x=0;
   if (z>=(int32)depth) z=depth-1;
   if (z<0) z=0;
   return voxels->getDataPointer(x + y*width + z*width*height);
}



template<class DataType, unsigned int dim>
StaticVector<DataType,dim> RegularVolume::interpolateAttribute( Vector3f pos, AAT aat ) const
{
   StaticVector<DataType,dim> ret;
   for( card32 i=0;i<dim;i++ )
      ret[i] = 0;

   pos[0] = max(0.0f, min((float)width-1.0f,  pos[0] ));
   pos[1] = max(0.0f, min((float)height-1.0f, pos[1] ));
   pos[2] = max(0.0f, min((float)depth-1.0f,  pos[2] ));

   const Vector3f floor_vec = makeVector3f( floor(pos[0]), floor(pos[1]), floor(pos[2]) );

   card32 index = (floor_vec[2]*height + floor_vec[1])*width + floor_vec[0];
   const card32 xInc = pos[0] < (width-2)  ? 1            : 0;
   const card32 yInc = pos[1] < (height-2) ? width        : 0;
   const card32 zInc = pos[2] < (depth-2)  ? width*height : 0;

   const Vector3f interp = pos - floor_vec;
   //return voxels->get<DataType,dim>( index, aat ); //debug

   ret += voxels->get<DataType,dim>( index,                aat ) * (1-interp[0]) * (1-interp[1]) * (1-interp[2]);
   ret += voxels->get<DataType,dim>( index+xInc,           aat ) * (interp[0])   * (1-interp[1]) * (1-interp[2]);
   ret += voxels->get<DataType,dim>( index+xInc+yInc,      aat ) * (interp[0])   * (interp[1])   * (1-interp[2]);
   ret += voxels->get<DataType,dim>( index     +yInc,      aat ) * (1-interp[0]) * (interp[1])   * (1-interp[2]);
   ret += voxels->get<DataType,dim>( index          +zInc, aat ) * (1-interp[0]) * (1-interp[1]) * (interp[2]);
   ret += voxels->get<DataType,dim>( index+xInc     +zInc, aat ) * (interp[0])   * (1-interp[1]) * (interp[2]);
   ret += voxels->get<DataType,dim>( index+xInc+yInc+zInc, aat ) * (interp[0])   * (interp[1])   * (interp[2]);
   ret += voxels->get<DataType,dim>( index     +yInc+zInc, aat ) * (1-interp[0]) * (interp[1])   * (interp[2]);

   return ret;
}


}

#endif