//---------------------------------------------------------------------------
#ifndef SceneObjectH
#define SceneObjectH
//---------------------------------------------------------------------------
#include "x3geometry.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
#include "LinearAlgebra.h"
#include "AttachedData.h"
#include "SceneObjectIterator.h"
//---------------------------------------------------------------------------

namespace X4 {

    /// Some single object, typically stored in the scene graph.
    /// Usually encapsulates some kind of geometry.
    /// The geometry is accessed via iterators; different interfaces might be provided,
    /// depending on the type of geometry that is encoded.
    /// Provides a material index for visualization, means to attach "Attachments" data objects,
    /// as well as its own homogeneous transformation (4x4 Matrix).
    /// The latter is provided for convinience, simplifying writing the editor.
    class X3GEOMETRY_API SceneObject : public Persistent {
        X4_ABSTRACT_CLASS(UnstructuredInCoreTriangleMesh);
    private:
        /// index to material stored in scene
        card32 materialIndex;
        /// each scene object has its own transformation matrix (for convinience, simplifies coding the editor)
        Matrix4f transformation;
        /// any kind of attached data
        AttachedDataContainer *attachments;

    public:
        SceneObject();

        // --- assignment
        virtual void assign(const Object* obj, X4CopyContext *context = NULL);
        /// assign settings (material, transform, persistent attachments) only, not content. called by assign as first step.
        virtual void assignSettings(const SceneObject *assignFrom);
        // add additional code for compatibility with legacy scenes (for current version it only calls Persistent::read())
        virtual void read(InputObjectStream *s);
        /// creates a copy of the scene object; for inCore objects, this is equivalent to ->copy(); this is also the implemented default behavior
        /// for out-of-core objects, the identifier on secondary storage (filename) is changed to
        /// create a real copy, not overwriting the original object
        virtual SceneObject *createOOCAwareCopy(std::string idPrefix = "", bool insertOriginal = true, std::string idSuffix = "");

        // --- material attribute
        void setMaterialIndex(card32 materialIndex) {this->materialIndex = materialIndex;}
        card32 getMaterialIndex() const {return materialIndex;}

        // --- transformation
        inline void setTransformation(const Matrix4f &t) {transformation = t;}
        inline Matrix4f getTransformation() const {return transformation;}

        // --- data access
        /// create an iterator for accessing the data. returns NULL if type is not supported
        virtual SceneObjectIterator *createIterator(SceneObjectIteratorType type) = 0;
        /// find out whether a certain iterator type is supported
        virtual bool                 supportsIteratorType(SceneObjectIteratorType cap) = 0;

        // --- attachments
        inline AttachedDataContainer *getAttachments() {
            return attachments;}
        inline const AttachedDataContainer *getAttachments() const {
            return attachments;}
        virtual void clearAttachments();
        virtual void deleteAttachment(const std::string &name);

        // --- visualization
        virtual void renderObjectGL() {}

        ~SceneObject();
    };

}


#endif
