#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "TopologicalKNNIterator.h"
//---------------------------------------------------------------------------
#include "PointCloud.h"
#include "PointCloudIterators.h"
#include "SceneObjectIterator.h"
#include "UnstructuredInCorePointCloud.h"
#include "HierarchicalKNNIterator.h"
//#include 
//---------------------------------------------------------------------------
using namespace X4;


TopologicalKNNIteratorBase::TopologicalKNNIteratorBase(PointCloud * pc, PointCloudIterator * topologyIter)
{
	m_Pc = pc;
	m_Buffer = NULL;
	m_posAAT = pc->getAAT( "position" );

	m_Iter = topologyIter;

	m_bIt = dynamic_cast<BasicPointCloudIterator*>(m_Iter);
	m_tIt = dynamic_cast<TopologyIterator*>(m_Iter);
	m_rIt = dynamic_cast<RandomAccessPointCloudIterator*>(m_Iter);

	x4Assert( m_bIt != NULL );
	x4Assert( m_tIt != NULL );
	x4Assert( m_rIt != NULL );

	m_PointSet = m_bIt->lockBlockRead();
	m_StartIndex = 0;
}

TopologicalKNNIteratorBase::~TopologicalKNNIteratorBase()
{
	m_bIt->unlock();
}

void TopologicalKNNIteratorBase::setStartPointAndReset(const PointIndex &index) 
{
	m_StartIndex = index;
	reset();
}

bool TopologicalKNNIteratorBase::atEnd() 
{
	return !m_NextPointExists;
}

card8 TopologicalKNNIteratorBase::get1ub(const AAT &token) const {
   return genericGetS<card8>(token);
}
Vector2ub TopologicalKNNIteratorBase::get2ub(const AAT &token) const {
   return genericGet<card8, 2>(token);
}
Vector3ub TopologicalKNNIteratorBase::get3ub(const AAT &token) const {
   return genericGet<card8, 3>(token);
}
Vector4ub TopologicalKNNIteratorBase::get4ub(const AAT &token) const {
   return genericGet<card8, 4>(token);
}
int32 TopologicalKNNIteratorBase::get1i(const AAT &token) const {
   return genericGetS<int32>(token);
}
Vector2i TopologicalKNNIteratorBase::get2i(const AAT &token) const {
   return genericGet<int32, 2>(token);
}
Vector3i TopologicalKNNIteratorBase::get3i(const AAT &token) const {
   return genericGet<int32, 3>(token);
}
Vector4i TopologicalKNNIteratorBase::get4i(const AAT &token) const {
   return genericGet<int32, 4>(token);
}
card32 TopologicalKNNIteratorBase::get1u(const AAT &token) const {
   return genericGetS<card32>(token);
}
Vector2u TopologicalKNNIteratorBase::get2u(const AAT &token) const {
   return genericGet<card32, 2>(token);
}
Vector3u TopologicalKNNIteratorBase::get3u(const AAT &token) const {
   return genericGet<card32, 3>(token);
}
Vector4u TopologicalKNNIteratorBase::get4u(const AAT &token) const {
   return genericGet<card32, 4>(token);
}
float32 TopologicalKNNIteratorBase::get1f(const AAT &token) const {
   return genericGetS<float32>(token);
}
Vector2f TopologicalKNNIteratorBase::get2f(const AAT &token) const {
   return genericGet<float32, 2>(token);
}
Vector3f TopologicalKNNIteratorBase::get3f(const AAT &token) const {
   return genericGet<float32, 3>(token);
}
Vector4f TopologicalKNNIteratorBase::get4f(const AAT &token) const {
   return genericGet<float32, 4>(token);
}

void TopologicalKNNIteratorBase::getVertex(char *buffer) const {
     card32 elementSize= m_Pc->getDescr()->getSize();
     for (card32 i=0; i<elementSize; i++) {
         buffer[i] = m_Buffer[i];
     };
}

PointIndex TopologicalKNNIteratorBase::getPointSetPointNumber() 
{
	return m_CurrentIndex;
}

//----------------------------------------------------------------------------------
TopologicalKNNIterator::TopologicalKNNIterator(PointCloud* pc, PointCloudIterator * topologyIter ) : 
   TopologicalKNNIteratorBase(pc, topologyIter)
{
}

TopologicalKNNIterator::~TopologicalKNNIterator() 
{
}

void TopologicalKNNIterator::reset() 
{
   m_Visited.clear();
   m_Queue.clear();
   m_NextPointExists = true;
   visit( m_StartIndex, 0.0f );
}

void TopologicalKNNIterator::next() 
{
   bool found = false;
   while( !found )
   {
      if( m_Queue.isEmpty() )
         break;

      float dist = m_Queue.getFirstKey();
      PointIndex nextIndex = m_Queue.pop_first();

      if( m_Visited.find(nextIndex) == m_Visited.end() )
      {
         visit( nextIndex, dist );
         found = true;
      }
   }
   m_NextPointExists = found;
}

void TopologicalKNNIterator::visit( PointIndex index,float distance) 
{
   m_CurrentIndex = index;
   m_CurrentTopologicalDistance = distance;
   m_rIt->goToIndex( index );
   m_Buffer = m_PointSet->getDataPointer( index );
   m_Visited[index] = distance;

   Vector3f pos = m_bIt->get3f( m_posAAT );

   // put neighbours on queue
   for( card32 i=0;i<m_tIt->getNumNeighbours();i++ )
   {
      PointIndex nIndex = m_tIt->getNeighbour(i);
      if( m_Visited.find( nIndex ) == m_Visited.end() )
      {
         Vector3f nPos = m_PointSet->get3f( nIndex, m_posAAT );
         float nDist = distance + norm(pos - nPos);

         m_Queue.insert( nDist, nIndex );
      }
   }
}

//----------------------------------------------------------------------------------
ModifyableTopologicalKNNIterator::ModifyableTopologicalKNNIterator(PointCloud* pc, PointCloudIterator * topologyIter) :
	TopologicalKNNIterator(pc, topologyIter)
{
	m_modIt = dynamic_cast<ModifyablePointCloudIterator*>(topologyIter);

	x4Assert(m_modIt != NULL);

	//if (m_modIt == NULL) 
	//	error("Given iterator must be of modifyable type"!);
}

ModifyableTopologicalKNNIterator::~ModifyableTopologicalKNNIterator()
{

}

void ModifyableTopologicalKNNIterator::set1ub(const AAT &token, const card8 &x)     const {
	m_modIt->set1ub(token,x);
}
void ModifyableTopologicalKNNIterator::set2ub(const AAT &token, const Vector2ub &v) const {
	m_modIt->set2ub(token,v);
}
void ModifyableTopologicalKNNIterator::set3ub(const AAT &token, const Vector3ub &v) const {
	m_modIt->set3ub(token,v);
}
void ModifyableTopologicalKNNIterator::set4ub(const AAT &token, const Vector4ub &v) const {
	m_modIt->set4ub(token,v);
}
void ModifyableTopologicalKNNIterator::set1i(const AAT &token, const int32 &x)    const {
	m_modIt->set1i(token,x);
}
void ModifyableTopologicalKNNIterator::set2i(const AAT &token, const Vector2i &v) const {
	m_modIt->set2i(token,v);
}
void ModifyableTopologicalKNNIterator::set3i(const AAT &token, const Vector3i &v) const {
	m_modIt->set3i(token,v);
}
void ModifyableTopologicalKNNIterator::set4i(const AAT &token, const Vector4i &v) const {
	m_modIt->set4i(token,v);
}
void ModifyableTopologicalKNNIterator::set1u(const AAT &token, const card32 &x)    const {
	m_modIt->set1u(token,x);
}
void ModifyableTopologicalKNNIterator::set2u(const AAT &token, const Vector2u &v) const {
	m_modIt->set2u(token,v);
}
void ModifyableTopologicalKNNIterator::set3u(const AAT &token, const Vector3u &v) const {
	m_modIt->set3u(token,v);
}
void ModifyableTopologicalKNNIterator::set4u(const AAT &token, const Vector4u &v) const {
	m_modIt->set4u(token,v);
}
void ModifyableTopologicalKNNIterator::set1f(const AAT &token, const float32 &x)  const {
	m_modIt->set1f(token,x);
}
void ModifyableTopologicalKNNIterator::set2f(const AAT &token, const Vector2f &v) const {
	m_modIt->set2f(token,v);
}
void ModifyableTopologicalKNNIterator::set3f(const AAT &token, const Vector3f &v) const {
	m_modIt->set3f(token,v);
}
void ModifyableTopologicalKNNIterator::set4f(const AAT &token, const Vector4f &v) const {
	m_modIt->set4f(token,v);
}

void ModifyableTopologicalKNNIterator::setVertex(const char *buffer) const {
	m_modIt->setVertex(buffer);
}




//----------------------------------------------------------------------------------
TopologicalKNNIterator_Fast::TopologicalKNNIterator_Fast(PointCloud* pc, PointCloudIterator * topologyIter ) : 
   TopologicalKNNIteratorBase(pc, topologyIter)
{
   m_Visited.resize(pc->getNumPoints(), -1.0);
}

TopologicalKNNIterator_Fast::~TopologicalKNNIterator_Fast() 
{
}

void TopologicalKNNIterator_Fast::reset() 
{
   m_Visited.clear();
   m_Visited.resize(m_Pc->getNumPoints(), -1.0);
   m_Queue.clear();
   m_NextPointExists = true;
   visit( m_StartIndex, 0.0f );
}

void TopologicalKNNIterator_Fast::next() 
{
   bool found = false;
   while( !found )
   {
      if( m_Queue.isEmpty() )
         break;

      float dist = m_Queue.getFirstKey();
      PointIndex nextIndex = m_Queue.pop_first();

      if( m_Visited[nextIndex] < 0 )
      {
         visit( nextIndex, dist );
         found = true;
      }
   }
   m_NextPointExists = found;
}

void TopologicalKNNIterator_Fast::stopFurther()
{
	if(m_Queue.isEmpty()) return;

	// get list of all children of the next node in the queue
	set<card32> neighbor;
	float dist = m_Queue.getFirstKey();
	PointIndex nextIndex = m_Queue.pop_first();
	neighbor.insert(nextIndex);
	m_rIt->goToIndex(nextIndex);
	card32 numNeighbours = m_tIt->getNumNeighbours();
	for( card32 i=0;i<numNeighbours;i++ )
		neighbor.insert(m_tIt->getNeighbour(i));

	// remove all found neighbors from the queue
	struct Elem { float d; PointIndex i; };
	std::vector<Elem> save;
	while(!m_Queue.isEmpty())
	{
		float dist = m_Queue.getFirstKey();
		PointIndex nextIndex = m_Queue.pop_first();
		if (!binary_search(neighbor.begin(), neighbor.end(), nextIndex))
		{
			Elem el;
			el.d = dist;
			el.i = nextIndex;
			save.push_back(el);
		}
	}

	for (unsigned i=0; i < save.size(); i++)
         m_Queue.insert( save[i].d , save[i].i );
}

void TopologicalKNNIterator_Fast::visit( PointIndex index,float distance) 
{
   m_CurrentIndex = index;
   m_CurrentTopologicalDistance = distance;
   m_rIt->goToIndex( index );
   m_Buffer = m_PointSet->getDataPointer( index );
   m_Visited[index] = distance;

   Vector3f pos = m_bIt->get3f( m_posAAT );

   // put neighbours on queue
   card32 numNeighbours = m_tIt->getNumNeighbours();
   for( card32 i=0;i<numNeighbours;i++ )
   {
      PointIndex nIndex = m_tIt->getNeighbour(i);
      if( m_Visited[nIndex] < 0 )
      {
         Vector3f nPos = m_PointSet->get3f( nIndex, m_posAAT );
         float nDist = distance + norm(pos - nPos);

         m_Queue.insert( nDist, nIndex );
      }
   }
}


//----------------------------------------------------------------------------------
X4::TopologicalRangeQuerry::TopologicalRangeQuerry( PointCloud * pc, PointCloudIterator* titer )
{
	UnstructuredInCorePointCloud * upc = dynamic_cast<UnstructuredInCorePointCloud*>(pc);
	if( !upc )
		throw "this querry is suitable for UnstructuredInCorePointCloud only!";

	if (titer == NULL) 
		titer = new FixedKNNTopologyIterator(upc->getPointSet(), upc->getAttachments(), upc, 10);

	m_PosAAT = pc->getAAT("position");
	m_hIt = new HierarchicalKNNIterator(pc,32,NULL);
	m_tIt = new TopologicalKNNIterator(pc, titer);
	m_PointSet = new PointSet;
	VertexDescriptor * vd = (VertexDescriptor*)pc->getDescr()->copy();
	vd->pushAttrib(VAD("psIndex", 1, VAD::DATA_FORMAT_INT32));
	m_PointSet->clearAndSetup(1,0,vd);
	m_IndexAAT = m_PointSet->getAAT("psIndex");
	m_NumPoints = 0;
}

void X4::TopologicalRangeQuerry::rangeQuerry( const Vector3f &pos, float topologicalRadius )
{
	m_hIt->setSeekPointAndReset(pos);
	card64 index = m_hIt->getCurrentPointIndex();
	rangeQuerry(index, topologicalRadius);


}

void X4::TopologicalRangeQuerry::rangeQuerry( const mpcard startIndex, float topologicalRadius )
{
	m_NumPoints = 0;
	std::vector<char> buffer(m_PointSet->getDescr()->getSize());

	m_tIt->setStartPointAndReset(startIndex);
	while( !m_tIt->atEnd())
	{
		float currentDistance = m_tIt->getTopologicalDistance();
		if( currentDistance > topologicalRadius )
			break;

		m_NumPoints++;

		if( m_NumPoints >= m_PointSet->getNumEntries() )
		{
			m_PointSet->changeHeight(m_NumPoints * 2);
			m_TopologicalDistances.resize(m_NumPoints * 2);
		}

		m_TopologicalDistances[m_NumPoints-1] = currentDistance;
		m_tIt->getVertex( &buffer[0] );
		m_PointSet->setVertex(m_NumPoints-1, &buffer[0] );
		m_PointSet->set1i(m_NumPoints-1, m_IndexAAT, m_tIt->getPointSetPointNumber());

		m_tIt->next(); 
	}
}
X4::Vector3f X4::TopologicalRangeQuerry::getPosition( unsigned i )
{
	return m_PointSet->get3f(i, m_PosAAT );
}

float X4::TopologicalRangeQuerry::getDistance( unsigned i )
{
	return m_TopologicalDistances[i];
}

X4::Vector3f X4::TopologicalRangeQuerry::getAttribute3f( unsigned i, AAT attribute )
{
	return m_PointSet->get3f(i, attribute );
}

X4::TopologicalRangeQuerry::~TopologicalRangeQuerry()
{
	delete m_hIt;
	delete m_tIt;
	delete m_PointSet;
}

X4::mpcard X4::TopologicalRangeQuerry::getPointIndexInPS( unsigned i )
{
	return m_PointSet->get1i(i, m_IndexAAT);
}