#ifndef TopologicalKNNIteratorH
#define TopologicalKNNIteratorH
//---------------------------------------------------------------------------
#include "x3geometry.h"
//---------------------------------------------------------------------------
#include "PointCLoudIterators.h"
#include <map>
#include "PointSet.h"
#include "VertexDescriptor.h"
#include "PriorityQueue.h"

//---------------------------------------------------------------------------

namespace X4
{
class PointCloud;
class HierarchicalKNNIterator;
class ModifyablePointCloudIterator;

/**
	topological k-nearest-neighbour iterator base class
*/
class X3GEOMETRY_API TopologicalKNNIteratorBase : public PointCloudIterator,
                                               public BasicPointCloudIterator
{

protected:

	virtual void visit( PointIndex index, float distance ) = 0;

	PointCloudIterator		*	m_Iter;
	BasicPointCloudIterator	*	m_bIt;
	RandomAccessPointCloudIterator	*	m_rIt;
	TopologyIterator			*	m_tIt;


	PointCloud				*	m_Pc;
	AAT							m_posAAT;

	PointIndex					m_StartIndex;
	PointIndex					m_CurrentIndex;
	PriorityQueue<float,PointIndex>	m_Queue;
	bool							m_NextPointExists;
	float							m_CurrentTopologicalDistance;

	const char				*	m_Buffer;
	const PointSet			*	m_PointSet;

public:


   TopologicalKNNIteratorBase(PointCloud*, PointCloudIterator * topologyIter );
	virtual ~TopologicalKNNIteratorBase();

   /// initialize iterator to start point
   void setStartPointAndReset(const PointIndex &index);

   /// go to starting point
   virtual void reset() = 0;
   /// go to next nearest point
   virtual void next() = 0;
   /// true if end-of-set
   virtual bool atEnd();

   template <class Type, int dim>
   inline StaticVector<Type, dim> genericGet(AAT token) const {
      const char *pos =  m_Buffer + token.getOffset();
      return *((StaticVector<Type, dim>*)(pos));
   }

   template <class Type>
   inline Type genericGetS(const AAT& token) const {
      const char *pos =  m_Buffer + token.getOffset();
      const StaticVector<Type, 1> v = *((StaticVector<Type, 1>*)(pos));
      return v[0];
   }


   /// get attributes in different formats
   virtual card8     get1ub(const AAT &token) const;
   virtual Vector2ub get2ub(const AAT &token) const;
   virtual Vector3ub get3ub(const AAT &token) const;
   virtual Vector4ub get4ub(const AAT &token) const;

   virtual int32    get1i(const AAT &token) const;
   virtual Vector2i get2i(const AAT &token) const;
   virtual Vector3i get3i(const AAT &token) const;
   virtual Vector4i get4i(const AAT &token) const;

   virtual card32   get1u(const AAT &token) const;
   virtual Vector2u get2u(const AAT &token) const;
   virtual Vector3u get3u(const AAT &token) const;
   virtual Vector4u get4u(const AAT &token) const;

   virtual float32  get1f(const AAT &token) const;
   virtual Vector2f get2f(const AAT &token) const;
   virtual Vector3f get3f(const AAT &token) const;
   virtual Vector4f get4f(const AAT &token) const;

   /// ask VertexDescriptor for buffer size
   virtual void     getVertex(char *buffer) const;

   PointIndex getPointSetPointNumber();

	/// returns approimate topological distance to start point
	float getTopologicalDistance() {return m_CurrentTopologicalDistance;}


   // --- blockwise access NOT useable here!
   virtual bool nextBlock(){x4Assert(false); return false;}
   virtual const PointSet * lockBlockRead(){x4Assert(false); return NULL;}
   virtual void unlock(){x4Assert(false);}
};


/**
topological k-nearest-neighbour iterator
*/
class X3GEOMETRY_API TopologicalKNNIterator : public TopologicalKNNIteratorBase
{

private:

   void visit( PointIndex index, float distance );

   std::map<PointIndex,float>	         m_Visited;		// holds distances from visited points

public:

   TopologicalKNNIterator(PointCloud*, PointCloudIterator * topologyIter );
   ~TopologicalKNNIterator();

   void reset();
   void next();
};


/**
topological k-nearest-neighbour iterator
*/
class X3GEOMETRY_API ModifyableTopologicalKNNIterator : public TopologicalKNNIterator, public ModifyablePointCloudIterator
{

protected:
	ModifyablePointCloudIterator* m_modIt;

public:

	ModifyableTopologicalKNNIterator(PointCloud*, PointCloudIterator * topologyIter );
	~ModifyableTopologicalKNNIterator();


	/// set attributes in different formats
	virtual void set1ub(const AAT &token, const card8 &x)     const ;
	virtual void set2ub(const AAT &token, const Vector2ub &v) const ;
	virtual void set3ub(const AAT &token, const Vector3ub &v) const ;
	virtual void set4ub(const AAT &token, const Vector4ub &v) const ;

	virtual void set1i(const AAT &token, const int32 &x)     const ;
	virtual void set2i(const AAT &token, const Vector2i &v) const ;
	virtual void set3i(const AAT &token, const Vector3i &v) const ;
	virtual void set4i(const AAT &token, const Vector4i &v) const ;

	virtual void set1u(const AAT &token, const card32 &x)     const ;
	virtual void set2u(const AAT &token, const Vector2u &v) const ;
	virtual void set3u(const AAT &token, const Vector3u &v) const ;
	virtual void set4u(const AAT &token, const Vector4u &v) const ;

	virtual void set1f(const AAT &token, const float32 &x)  const ;
	virtual void set2f(const AAT &token, const Vector2f &v) const ;
	virtual void set3f(const AAT &token, const Vector3f &v) const ;
	virtual void set4f(const AAT &token, const Vector4f &v) const ;

	/// ask VertexDescriptor for size
	virtual void setVertex(const char *data) const ;

	// --- blockwise access NOT used here!
	virtual PointSet *lockBlockReadWrite(LockRWHint hint = CHANGE_EVERYTHING){x4Assert(false); return NULL;}

};

/**
topological k-nearest-neighbour iterator
*/
class X3GEOMETRY_API TopologicalKNNIterator_Fast : public TopologicalKNNIteratorBase
{

private:

   void visit( PointIndex index, float distance );
   
   std::vector<float>	         m_Visited;		// holds distances from visited points

public:

   TopologicalKNNIterator_Fast(PointCloud*, PointCloudIterator * topologyIter );
   ~TopologicalKNNIterator_Fast();

   void reset();
   void next();

   //! stop propagation on the current point, this will remove all children of the current point from the internal queue
   void stopFurther();
};


class X3GEOMETRY_API TopologicalRangeQuerry
{
public:

	// use given iterator or check for topology attachment in the point cloud (if given iterator is NULL, then compute topology on the fly)
	TopologicalRangeQuerry( PointCloud * pc, PointCloudIterator* titer = NULL);	

	// compute fixed knn topology on the fly
	//TopologicalRangeQuerry( PointCloud * pc);
	~TopologicalRangeQuerry();

	void rangeQuerry( const Vector3f &pos, float topologicalRadius );
	void rangeQuerry( const mpcard startIndex, float topologicalRadius );


	unsigned getNumPoints() {return m_NumPoints;}
	Vector3f getPosition( unsigned i);
	float getDistance( unsigned i);
	Vector3f getAttribute3f( unsigned i, AAT attribute);
	mpcard getPointIndexInPS(unsigned i);

private:

	TopologicalKNNIterator *	m_tIt;
	HierarchicalKNNIterator *	m_hIt;
	PointSet *	m_PointSet;
	std::vector<float> m_TopologicalDistances;
	unsigned		m_NumPoints;	// != m_PointSet.getNumElements() (number of actually querried points)
	AAT			m_PosAAT;
	AAT			m_IndexAAT;
};

}

#endif