//---------------------------------------------------------------------------
#ifndef TopologyRangeSearchH
#define TopologyRangeSearchH
//---------------------------------------------------------------------------
#include "x3geometry.h"
//---------------------------------------------------------------------------
#include "LinearAlgebra.h"
#include "HashTables.h"
//---------------------------------------------------------------------------

#include <queue>
#include <set>

namespace X4 {

class InCorePCTopologyGraph;
class PointSet;   

/// search a range in a point cloud with topology that is topologically reachable 
class X3GEOMETRY_API TopNeighborhoodRangeSearch {
 private:

    struct Card32KeyComp {
       static bool equals(card32 pos1, card32 pos2) {
          return pos1 == pos2;
       }
    };

    class VisitedNodesHT : public HashTable<card32, card32, Card32KeyComp> {
    public:
       virtual max_platform_card hashFunction(card32 key) const {
          return key;
       }
       VisitedNodesHT(max_platform_card fillFactor = 1, max_platform_card startSizeIndex = 0) : HashTable<card32, card32, Card32KeyComp>(fillFactor, startSizeIndex) {}
    };

   const InCorePCTopologyGraph *tg;
   const PointSet *ps;
   std::vector<card32> indexStack;
   VisitedNodesHT visited;

 protected:
   virtual bool inRange(const PointSet *ps, card32 index) = 0;
   void setup(const InCorePCTopologyGraph *tg, const PointSet *ps, card32 startIndex);
   
 public:
   TopNeighborhoodRangeSearch();
   bool atEnd(); 
   void next();
   card32 getCurrentIndex();
};

/// search topologically reachable spherical region
class X3GEOMETRY_API TopNeighborhoodSphereSearch : public TopNeighborhoodRangeSearch {
 private:
	Vector3f center;
	float32 dist;
	AAT posChannel;

 protected:
   virtual bool inRange(const PointSet *ps, card32 index);
   
 public:
   void setup(const InCorePCTopologyGraph *tg, const PointSet *ps, card32 startIndex,
              const Vector3f &center, const float32 dist, const AAT &posChannel);
};


/// nearest-neighbors search along topology graph edges
class X3GEOMETRY_API TopNeighborhoodKNNSearch {
 private:
   const InCorePCTopologyGraph *tg;
   const PointSet *ps;
   Vector3f pos;
   AAT posChannel;
   
   struct QueueEntry {
		float32 dist;
		card32 index;

		inline QueueEntry() {}
		inline QueueEntry(float32 d, card32 i) {dist=d; index=i;}		
		inline bool operator<(const QueueEntry &e) const {return dist > e.dist;}
   };
   
   priority_queue<QueueEntry> neighbors;
   set<card32> visited;

 public:
   TopNeighborhoodKNNSearch(const PointSet *ps, const InCorePCTopologyGraph *tg, card32 startIndex, const AAT &posChannel);
   bool atEnd(); 
   void next();
   void reset(card32 startIndex);
   card32 getCurrentIndex();
};


/// nearest-neighbors search along topology graph edges
/// this version is faster (not using the std::map) if the iterator is _not_ recreated frequently (creation is O(#points))
class X3GEOMETRY_API TopNeighborhoodKNNSearch_Fast {
 private:
   const InCorePCTopologyGraph *tg;
   const PointSet *ps;
   Vector3f pos;
   AAT posChannel;
   
   struct QueueEntry {
		float32 dist;
		card32 index;

		inline QueueEntry() {}
		inline QueueEntry(float32 d, card32 i) {dist=d; index=i;}		
		inline bool operator<(const QueueEntry &e) const {return dist > e.dist;}
   };
   
   priority_queue<QueueEntry> neighbors;
   std::vector<card32> visited;

 public:
   TopNeighborhoodKNNSearch_Fast(const PointSet *ps, const InCorePCTopologyGraph *tg, card32 startIndex, const AAT &posChannel);
   bool atEnd(); 
   void next();
   void reset(card32 startIndex);
   card32 getCurrentIndex();
};

}

#endif