//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "TriangleMeshTools.h"
#include "AttachedIndexedOctree.h"
#include "PointSetKNNQuery.h"
#include "Mathtools.h"
#include "Mathtools.inline.h"
#include "PCA.h"
#include "VertexArray.h"
#include "InCorePCTopologyGraph.h"
#include "TopologyAttachment.h"
#include <float.h>
#include "PCA.inline.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

/** 
* Snap close vertices (attention naive implementation O(n^2))
*/
void X4::snapTriangleMeshMesh( UnstructuredInCoreTriangleMesh* mesh, const float32 maxDist )
{
	if ( ( mesh == NULL ) ){
		debugOutput << "smoothTriangleMeshNormals: invalid mesh.\n"; debugOutput.flush();
		return;
	}
	PointSet* vertices = mesh->getPointSet();
	if ( !vertices->providesAttribute( "position" ) || !vertices->providesAttribute( "normal" ) || !vertices->providesAttribute( "color"  ) || !vertices->providesAttribute( "flags") ) {
		debugOutput << "smoothTriangleMeshNormals: vertex attribute channel missing.\n"; debugOutput.flush();
		return;
	}
	AAT POSITION = vertices->getAAT( "position" );
	AAT NORMAL = vertices->getAAT( "normal" );
	AAT COLOR = vertices->getAAT( "color" );
	AAT FLAGS = vertices->getAAT( "flags" );
	card32 nIndex;
	bool found;
	std::vector<int32> map; map.resize( vertices->getNumEntries() );
	std::vector<card32> numMerged; numMerged.resize( vertices->getNumEntries() );
	for( unsigned int n = 0; n < map.size(); n++ ) { map[n] = -1; numMerged[n] = 1; }
	card32 numDiscarded = 0;

	AttachedIndexedOctree* octree = PointSetKNNQuery::createOctree( vertices );
	PointSetKNNQuery knn( vertices, octree, NULL_VECTOR3F, POSITION );	
	for ( unsigned int pointIndex = 0; pointIndex < vertices->getNumEntries(); pointIndex++ ){
		if ( map[pointIndex] < 0 ){
			Vector3f pos = vertices->get3f( pointIndex, POSITION );
			knn.setTargetPoint( pos );
			knn.getNextNearestPointIndex( nIndex, found );
			float32 d = norm( vertices->get3f( nIndex, POSITION ) - pos );;
			while( found && ( d < maxDist ) ){
				if( ( map[nIndex] < 0 ) && ( nIndex > pointIndex ) ){
					map[nIndex] = pointIndex;
					vertices->set3f( pointIndex, POSITION, vertices->get3f( pointIndex, POSITION ) + vertices->get3f( nIndex, POSITION ) );
					numMerged[pointIndex]++;
					numDiscarded++;
				}
				knn.getNextNearestPointIndex( nIndex, found );
				d = norm( vertices->get3f( nIndex, POSITION ) - pos );
			}
		}
	} 	
	delete octree;

	// Create new vertex point set
	PointSet* newVertices = new PointSet();
	card32 numNewVertices = vertices->getNumEntries() - numDiscarded;
	std::vector<int32> deletionMap; deletionMap.resize( vertices->getNumEntries() );
	newVertices->clearAndSetup( 1, numNewVertices, vertices->getDescr() );
	card32 newPointIndex = 0;
	for ( card32 pointIndex = 0; pointIndex < vertices->getNumEntries(); pointIndex++ ){
		if ( map[pointIndex] < 0 ){
			// Copy
			Vector3f newPos = vertices->get3f( pointIndex, POSITION ) / float32( numMerged[pointIndex] );
			newVertices->set3f( newPointIndex, POSITION, newPos );
			newVertices->set3f( newPointIndex, NORMAL, vertices->get3f( pointIndex, NORMAL ) );
			newVertices->set3f( newPointIndex, COLOR, vertices->get3f( pointIndex, COLOR ) );
			newVertices->set1i( newPointIndex, FLAGS, vertices->get1i( pointIndex, FLAGS ) );
			deletionMap[pointIndex] = newPointIndex;
			newPointIndex++;
		} else {
			deletionMap[pointIndex] = -1;
		}
	}

	// Update triangle point set
	std::vector<mpcard> flippedTriangles; flippedTriangles.clear();
	std::vector<Vector3i> newTriangleContainer; newTriangleContainer.clear();
	PointSet* triangles = mesh->getTriangles();
	AAT TRIANGLE = triangles->getAAT( "index" );
	for ( unsigned int triangleIndex = 0; triangleIndex < triangles->getNumEntries(); triangleIndex++ ){
		Vector3i t = triangles->get<int32,3>( triangleIndex, TRIANGLE );
		Vector3f oldNormal = ( vertices->get3f( t[2], POSITION ) - vertices->get3f( t[0], POSITION ) ). crossProduct( vertices->get3f( t[1], POSITION ) - vertices->get3f( t[0], POSITION ) );
		// Mapping of discarded points
		if ( map[t[0]] >= 0 ) t[0] = map[t[0]];
		if ( map[t[1]] >= 0 ) t[1] = map[t[1]];
		if ( map[t[2]] >= 0 ) t[2] = map[t[2]];
		// Mapping of points moved due to deletion of other points
		if ( deletionMap[t[0]] >= 0 ) t[0] = deletionMap[t[0]];
		if ( deletionMap[t[1]] >= 0 ) t[1] = deletionMap[t[1]];
		if ( deletionMap[t[2]] >= 0 ) t[2] = deletionMap[t[2]];

		// Check if still valid
		if ( ( t[0] != t[1] ) && ( t[0] != t[2] ) && ( t[1] != t[2] ) ){
			// Check if triangle is valid
			int32 newNumVertices = int32( newVertices->getNumEntries() );
			if( ( t[0] >= newNumVertices ) || ( t[1] >= newNumVertices ) || ( t[2] >= newNumVertices ) ){
				debugOutput << "index too large.\n"; debugOutput.flush();
			} else if ( (t[0] < 0) || (t[1] < 0) || (t[2] < 0) ){
				debugOutput << "negative index.\n"; debugOutput.flush();
			} else {
				Vector3f newNormal = ( newVertices->get3f( t[2], POSITION ) - newVertices->get3f( t[0], POSITION ) ). crossProduct( newVertices->get3f( t[1], POSITION ) - newVertices->get3f( t[0], POSITION ) );
				if ( newNormal * oldNormal < 0 ){
					flippedTriangles.push_back( newTriangleContainer.size() );
				}
				newTriangleContainer.push_back( t );
			}
		}
	}
	PointSet* newTriangles = new PointSet();
	newTriangles->clearAndSetup( 1, newTriangleContainer.size(), triangles->getDescr() );
	for ( unsigned int triangleIndex = 0; triangleIndex < newTriangleContainer.size(); triangleIndex++ ){
		newTriangles->set<int32,3>( triangleIndex, TRIANGLE, newTriangleContainer[triangleIndex] );
	}

	debugOutput << "#vertices: " << vertices->getNumEntries() << "->" << newVertices->getNumEntries() << "\n"; debugOutput.flush();
	debugOutput << "#triangles: " << triangles->getNumEntries() << "->" << newTriangles->getNumEntries() << "\n"; debugOutput.flush();
	//debugOutput << "flipped triangles: " << flippedTriangles.size() << "\n"; debugOutput.flush();

	mesh->setPointSet( newVertices );
	mesh->setTriangles( newTriangles );
}

/** 
* Smooth out normals
*/
void X4::smoothTriangleMeshNormals( UnstructuredInCoreTriangleMesh* mesh, float32 sigma )
{
	if ( ( mesh == NULL ) ){
		debugOutput << "smoothTriangleMeshNormals: invalid mesh.\n"; debugOutput.flush();
		return;
	}
	card32 numNormalSmoothings = 1;
	PointSet* vertices = mesh->getPointSet();
	if ( !vertices->providesAttribute( "position" ) || !vertices->providesAttribute( "normal" ) ){
		debugOutput << "smoothTriangleMeshNormals: vertex attribute channel missing.\n"; debugOutput.flush();
		return;
	}
	AAT POSITION = vertices->getAAT( "position" );
	AAT NORMAL = vertices->getAAT( "normal" );
	AttachedIndexedOctree* octree = PointSetKNNQuery::createOctree( vertices );
	PointSetKNNQuery knn( vertices, octree, NULL_VECTOR3F, POSITION );	
	card32 nIndex;
	bool found;
	for ( unsigned int iteration = 0; iteration < numNormalSmoothings; iteration++ ){
		for ( unsigned int pointIndex = 0; pointIndex < vertices->getNumEntries(); pointIndex++ ){
			Vector3f pos = vertices->get3f( pointIndex, POSITION );
			knn.setTargetPoint( pos );
			// Skip self
			knn.getNextNearestPointIndex( nIndex, found );
			knn.getNextNearestPointIndex( nIndex, found );
			Vector3f nPos = vertices->get3f( nIndex, POSITION );
			Vector3f avgNormal = makeVector3f( 0,0,0 );
			float32 d = norm( pos - nPos );
			card32 numNeighbors = 0;
			float32 summedWeight = 0;
			card32 numFound = 0;
			while( ( numFound < 1 ) || ( d < sigma ) ){
				float32 w = exp( - sqr( d ) / sqr( sigma ) );
				summedWeight += w;
				avgNormal += vertices->get3f( nIndex, NORMAL ) * w;
				numNeighbors++;

				knn.getNextNearestPointIndex( nIndex, found );
				nPos = vertices->get3f( nIndex, POSITION );
				d = norm( pos - nPos );
				numFound++;
			}
			avgNormal /= summedWeight;
			vertices->set3f( pointIndex, NORMAL, normalize( avgNormal ) );
		}
	}
	delete octree;
}

/** 
* Compute normals for the mesh
*/
void X4::computeTriangleMeshNormals( UnstructuredInCoreTriangleMesh* mesh )
{
	if ( ( mesh == NULL ) ){
		debugOutput << "computeTriangleMeshNormals: invalid mesh.\n"; debugOutput.flush();
		return;
	}
	PointSet* triangles = mesh->getTriangles();
	PointSet* vertices = mesh->getPointSet();
	if ( !vertices->providesAttribute( "position" ) || !vertices->providesAttribute( "normal" ) ){
		debugOutput << "computeTriangleMeshNormals: vertex attribute channel missing.\n"; debugOutput.flush();
		return;
	}
	if ( !triangles->providesAttribute( "index" ) ){
		debugOutput << "computeTriangleMeshNormals: triangle attribute channel missing.\n"; debugOutput.flush();
		return;
	}
	AAT NORMAL = vertices->getAAT( "normal" );
	AAT POSITION = vertices->getAAT( "position" );
	AAT TRIANGLE_INDEX = triangles->getAAT( "index" );
	std::vector<Vector3f> newNormals; newNormals.resize( vertices->getNumEntries() );
	std::vector<card32> numNewNormals; numNewNormals.resize( vertices->getNumEntries() );

	// Init container
	for ( unsigned int normalIndex = 0; normalIndex < vertices->getNumEntries(); normalIndex++ ){
		newNormals[normalIndex] = makeVector3f( 0,0,0 );
		numNewNormals[normalIndex] = 0;
	}

	// Compute normals based on triangles
	for ( unsigned int triangleIndex = 0; triangleIndex < mesh->getNumTriangles(); triangleIndex++ ){
		Vector3i t = triangles->get<int32,3>( triangleIndex, TRIANGLE_INDEX );
		Vector3f v0 = vertices->get3f( t[0], POSITION );
		Vector3f v1 = vertices->get3f( t[1], POSITION );
		Vector3f v2 = vertices->get3f( t[2], POSITION );
		
		/*StaticVector<double,3> v0d; 
		v0d[0] = (double)v0[0];
		v0d[1] = (double)v0[1];
		v0d[2] = (double)v0[2];


		StaticVector<double,3> v1d; 
		v1d[0] = (double)v1[0];
		v1d[1] = (double)v1[1];
		v1d[2] = (double)v1[2];

		StaticVector<double,3> v2d; 
		v2d[0] = (double)v2[0];
		v2d[1] = (double)v2[1];
		v2d[2] = (double)v2[2];

		StaticVector<double,3> tNormald = (v2d-v0d).crossProduct(v1d-v0d);
		double nrm = 0;
		for (int i=0; i<3; i++)
			nrm += tNormald[i] * tNormald[i];
		Vector3f tNormal = makeVector3f(tNormald[0] / nrm, tNormald[1] / nrm, tNormald[2] / nrm);
	*/

		Vector3f tNormal = ( v2 - v0 ).crossProduct( v1 - v0 );
		float32 n = norm( tNormal );
		if ( n > 1e-7 ) {
			tNormal /= n;
			for ( unsigned int vIndex = 0; vIndex < 3; vIndex++ ){
				newNormals[ t[vIndex] ] += tNormal;
				numNewNormals[ t[vIndex] ]++;
			}
		}
	}

	// Set newly calculated normals
	for ( unsigned int normalIndex = 0; normalIndex < vertices->getNumEntries(); normalIndex++ ){
		if ( numNewNormals[normalIndex] > 0 ){
			vertices->set3f( normalIndex, NORMAL, -newNormals[normalIndex] / float32( numNewNormals[normalIndex] ) );
			//debugOutput << numNewNormals[normalIndex] << "\n"; debugOutput.flush();
		} else {
			vertices->set3f( normalIndex, NORMAL, makeVector3f( 1,0,0 ) );
		}
	}
}

/** 
* Compute normals for the mesh
*/
void X4::adjustTriangleIndicesToNormals( UnstructuredInCoreTriangleMesh* mesh )
{
	if ( mesh == NULL ) 
		throw X4Exception("mesh == NULL");

	PointSet* triangles = mesh->getTriangles();
	PointSet* vertices = mesh->getPointSet();
	if ( !vertices->providesAttribute( "position" ) || !vertices->providesAttribute( "normal" ) ){
		debugOutput << "computeTriangleMeshNormals: vertex attribute channel missing.\n"; debugOutput.flush();
		return;
	}
	if ( !triangles->providesAttribute( "index" ) ){
		debugOutput << "computeTriangleMeshNormals: triangle attribute channel missing.\n"; debugOutput.flush();
		return;
	}
	AAT NORMAL = vertices->getAAT( "normal" );
	AAT POSITION = vertices->getAAT( "position" );
	AAT TRIANGLE_INDEX = triangles->getAAT( "index" );

	// Compute normals based on triangles
	for ( unsigned int triangleIndex = 0; triangleIndex < mesh->getNumTriangles(); triangleIndex++ ){
		Vector3i t = triangles->get<int32,3>( triangleIndex, TRIANGLE_INDEX );
		Vector3f v0 = vertices->get3f( t[0], POSITION );
		Vector3f v1 = vertices->get3f( t[1], POSITION );
		Vector3f v2 = vertices->get3f( t[2], POSITION );

		Vector3f newNormal = ( v2 - v0 ).crossProduct( v1 - v0 );
		float32 n = norm( newNormal );
		if ( n <= 1e-7 ) 
			continue;
		newNormal /= n;

		Vector3f oldNormal = NULL_VECTOR3F;
		for ( unsigned int vIndex = 0; vIndex < 3; vIndex++ ){
			oldNormal += -vertices->get3f( t[vIndex], NORMAL );
		}
		oldNormal /= 3;

		if( oldNormal * newNormal < 0 ) { // flip indices
			triangles->set3i( triangleIndex, TRIANGLE_INDEX, makeVector3i( t[0], t[2], t[1] ) );
		}
	}

	// recalculate normals
	computeTriangleMeshNormals(mesh);
}

Vector2f toEuler( Vector3f p )
{
	Vector2f coord;
	p = normalize( p );
	Vector3f x = makeVector3f( 1,0,0 );
	Vector3f y = makeVector3f( 0,1,0 );
	Vector3f z = makeVector3f( 0,0,1 );
	coord[0] = acos( y * p );
	// Project p into xz plane
	p[1] = 0; 
	float32 n = norm( p );
	if ( n < 1e-9 ){
		coord[1] = float32( M_PI );
	} else {
		p /= n;
		coord[1] = acos( x * p );
		if ( p * z < 0 ){
			//coord[1] = - coord[1];
		}
		coord[1] += float32( M_PI );
	}

	return coord;
}

/** 
 * Parameterize a triangle mesh
 */
void X4::parameterizeTriangleMesh( 
	UnstructuredInCoreTriangleMesh* mesh, 
	const card32 mode, 
	const float32 textureRepeat, 
	bool pcaROT, 
	card32 coordIndex0, 
	card32 coordIndex1 )
{
	if ( ( mesh == NULL ) ){
		debugOutput << "computeTriangleMeshNormals: invalid mesh.\n"; debugOutput.flush();
		return;
	}
	PointSet* vertices = mesh->getPointSet();
	if ( !vertices->providesAttribute( "position" ) || !vertices->providesAttribute( "texCoord" ) ){
		debugOutput << "parameterizeTriangleMesh: vertices attribute channel missing.\n"; debugOutput.flush();
		return;
	}
	AAT TEX_COORD = vertices->getAAT( "texCoord" );
	AAT POSITION = vertices->getAAT( "position" );

	// Compute center of object
	PCA<float32,3> pca; pca.clear();
	for ( unsigned int pointIndex = 0; pointIndex < vertices->getNumEntries(); pointIndex++ ){
		Vector3f pos = vertices->get3f( pointIndex, POSITION );
		pca.addPoint( pos );
	}
	Matrix3f eigenVectors;
	Vector3f eigenValues, centroid;
	pca.analyze( eigenValues, eigenVectors, centroid );

	Vector2f u = makeVector2f( 10000, -100000 );
	Vector2f v = makeVector2f( 10000, -100000 );
	Matrix3f T = eigenVectors.transpose();
	for ( unsigned int pointIndex = 0; pointIndex < vertices->getNumEntries(); pointIndex++ ){
		Vector3f pos = vertices->get3f( pointIndex, POSITION );
		// Projection to plane
		Vector3f proj = T * ( pos - centroid );
		// u parameters
		if ( proj[0] < u[0] ) u[0] = proj[0];	// min
		if ( proj[0] > u[1] ) u[1] = proj[0];	// max
		// v parameters
		if ( proj[1] < v[0] ) v[0] = proj[1];	// min
		if ( proj[1] > v[1] ) v[1] = proj[1];	// max
	}
	float32 scalePlane = max( u[1] - u[0], v[1] - v[0] );
	Vector2f tPlane = makeVector2f( u[0], v[0] );

	for ( unsigned int pointIndex = 0; pointIndex < vertices->getNumEntries(); pointIndex++ ){
		Vector3f pos = vertices->get3f( pointIndex, POSITION );
		Vector2f texCoord = makeVector2f( 0,0 );

		if ( mode == MESH_PARAMETERIZATION_MODE_SPHERE ){
			texCoord = toEuler( pos - centroid );
			texCoord[0] *= float32( textureRepeat / M_PI );
			texCoord[1] *= float32( textureRepeat / ( 2.0f * M_PI ) );
		} else if ( mode == MESH_PARAMETERIZATION_MODE_PLANE ){
			Vector3f p3D = (pcaROT?T:IDENTITY3F) * ( pos - centroid );
			Vector2f p2D = makeVector2f( p3D[coordIndex0], p3D[coordIndex1] );
			texCoord = ( p2D - tPlane ) * scalePlane * textureRepeat;
		} else if ( mode == MESH_PARAMETERIZATION_MODE_CYLINDER ){
			// Direction = eigenVectors[0]
			pos = normalize( pos - centroid );
			texCoord = toEuler( pos - centroid );
			texCoord[0] = ( pos * eigenVectors[0] ) * textureRepeat;
		}
		
		vertices->set2f( pointIndex, TEX_COORD, texCoord );
	}
}

void X3GEOMETRY_API X4::computeTriangleMeshNormalsByPCA( UnstructuredInCoreTriangleMesh* meshA )
{
	VertexArray* pcVertices = new VertexArray(meshA);

	AttachedDataContainer *attachments = meshA->getAttachments();
	InCorePCTopologyGraph *tg;
	TopologyAttachment *topAtt = dynamic_cast<TopologyAttachment*>(attachments->getData("TopologyAtt"));
	if (topAtt) {
		tg = topAtt->getTopology();
	} else {
		meshA->computeTopologyAttachment();
		topAtt = dynamic_cast<TopologyAttachment*>(attachments->getData("TopologyAtt"));
		tg = topAtt->getTopology();
	}


	AAT posAAT = meshA->getAAT( "position" );
	AAT normalAAT = meshA->getAAT( "normal" );

	PointSet* ps = meshA->getPointSet();
	std::vector<Vector3f> posBuffer;
	std::vector<int32> pts;

	for (mpcard i = 0; i < meshA->getNumPoints(); i++)
	{


		Vector3f pos = pcVertices->getPosition3f(i);
		Vector3f myNormal = pcVertices->getNormal3f(i);

		pts.clear();
		pts.push_back(i);


		mpcard numNeigh = tg->getNumAdjacentVertices(i);
		for (mpcard n = 0; n < numNeigh; n++)
		{
			mpcard nId = tg->getAdjacentVertex(i, n);
			pts.push_back(nId);
		}

		mpcard numPoints = pts.size();
		for (mpcard p = 0; p < numPoints; p++)
		{
			mpcard numNeigh = tg->getNumAdjacentVertices(pts[p]);
			for (mpcard n = 0; n < numNeigh; n++)
			{
				mpcard nId = tg->getAdjacentVertex(pts[p], n);
				pts.push_back(nId);
			}
		}

		sort(pts.begin(), pts.end());
		pts.erase(   unique(pts.begin(), pts.end())  , pts.end() );



		float32 baseSigma = 0;
		numPoints = pts.size();
		std::vector<float32> dists(numPoints);
		for (mpcard n = 0; n < numPoints; n++)
		{
			Vector3f nPos = pcVertices->getPosition3f(pts[n]);
			float32 distance = norm(pos-nPos);
			dists[n] = distance;
			if (distance > baseSigma)
				baseSigma = distance;
		}

		PCA3f pca;
		for (mpcard n = 0; n < numPoints; n++)
		{
			Vector3f p = pcVertices->getPosition3f(pts[n]);
			pca.addPoint( p, exp( - ( dists[n] ) / baseSigma ) );
		}
		Matrix3f eigenVectors;
		Vector3f eigenValues, centroid;
		pca.analyze( eigenValues, eigenVectors, centroid );
		eigenVectors[2].normalize();

		Vector3f calculatedNormal = eigenVectors[2];
		if ( _isnan( calculatedNormal[0] ) || _isnan( calculatedNormal[1] ) || _isnan( calculatedNormal[2] ) ){
			debugOutput << "ERROR! - estimateNormals: calculated normal is NAN point "<< i <<"\n"; debugOutput.flush();
			calculatedNormal = makeVector3f(1,0,0);
		}
		if ( !_finite( calculatedNormal[0] ) || !_finite( calculatedNormal[1] ) || !_finite( calculatedNormal[2] ) ){
			debugOutput << "ERROR! - estimateNormals: calculated normal is INF point "<<i << "\n"; debugOutput.flush();
			calculatedNormal = makeVector3f(1,0,0);
		}


		Vector3f oldNormal = pcVertices->getNormal3f(i);
		if (calculatedNormal*oldNormal>=0)
		{
			pcVertices->setNormal3f( i, calculatedNormal );
		}else
			pcVertices->setNormal3f( i, -calculatedNormal );


	}


}