//---------------------------------------------------------------------------
#include "StdAfx.h"

//---------------------------------------------------------------------------
#include "UnstructuredInCoreTriangleMesh.h"
//---------------------------------------------------------------------------
#include "HashTables.inline.h"
#include "LinearAlgebra.inline.h"
#include "PointSet.inline.h"
#include "HierarchicalKNNIterator.h"
#include <set>
#include <queue>
#include "InCorePCTopologyGraph.h"
#include "TopologyAttachment.h"
#include "Line3f.h"
#include "Plane3f.h"
#include "TopologicalKNNIterator.h"

//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------
#include "VertexArray.h"

IMPLEMENT_X4_CLASS(UnstructuredInCoreTriangleMesh, 0)
{
	BEGIN_CLASS_INIT(UnstructuredInCoreTriangleMesh);
	ADD_OBJECT_PROP( triangles, 0, PointSet::getClass(), true );
}

// ----


UnstructuredInCoreTriangleMesh::UnstructuredInCoreTriangleMesh()
{
   triangles = new PointSet();
   VertexDescriptor vd;
   vd.pushAttrib(mVAD("position", 3, VAD::DATA_FORMAT_FLOAT32));
   vd.pushAttrib(mVAD("normal", 3, VAD::DATA_FORMAT_FLOAT32));
   clearAndSetup( &vd, 0 );
   //setMaterialIndex(3);
}

void UnstructuredInCoreTriangleMesh::clearAndSetup(const VertexDescriptor *vertDescr, mpcard numPoints) {
   UnstructuredInCorePointCloud::clearAndSetup(vertDescr, numPoints);
   VertexDescriptor vd;
   vd.pushAttrib(mVAD("index", 3, VAD::DATA_FORMAT_INT32));
   triangles->clearAndSetup(1, 0, &vd);
}

void UnstructuredInCoreTriangleMesh::clearAndSetup(const VertexDescriptor *vertDescr, mpcard numPoints,
                                                   const VertexDescriptor *triangleDescr, mpcard numTriangles)
{
   UnstructuredInCorePointCloud::clearAndSetup(vertDescr, numPoints);
   triangles->clearAndSetup(1, numTriangles, triangleDescr);

}                                            

void UnstructuredInCoreTriangleMesh::assign(const Object* obj, X4CopyContext *context) {
   UnstructuredInCorePointCloud::assign(obj, context);
   const UnstructuredInCoreTriangleMesh *other = dynamic_cast<const UnstructuredInCoreTriangleMesh*>(obj);
   x4Assert(other);
   if (other != this) {
      triangles->assign(other->triangles);
   }
}

PointCloudIterator *UnstructuredInCoreTriangleMesh::createIterator(SceneObjectIteratorType type) {
   if (!supportsIteratorType(type)) {
      return NULL;
   }
   if (type.providesCapsFor(SOIT::CAP_TRIANGLE_MESH)) {
      return new ConnectivityPointSetIterator(triangles, getHierarchyAttachedDSs(), getPointSet()->getBoundingBox());
   } else {
      return UnstructuredInCorePointCloud::createIterator(type);
   }
}

bool UnstructuredInCoreTriangleMesh::supportsIteratorType(SceneObjectIteratorType cap) {
   if (!cap.providesCapsFor(SOIT::CAP_TRIANGLE_MESH)) {
      return UnstructuredInCorePointCloud::supportsIteratorType(cap);
   } else {
      SOIT possibleCaps(  SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC | SOIT::CAP_DYNAMIC_PC
                        | SOIT::CAP_HIERARCHICAL  | SOIT::CAP_ATTACH_SECONDARY | SOIT::CAP_TRIANGLE_MESH );
      return (possibleCaps.providesCapsFor(cap));
   }
}

void UnstructuredInCoreTriangleMesh::clear() {
   UnstructuredInCorePointCloud::clear();
   triangles->clear();
}

const PointSet *UnstructuredInCoreTriangleMesh::getTriangles() const {
   return triangles;
}

void UnstructuredInCoreTriangleMesh::setTriangles(PointSet *psTriangles) {
   if (triangles != psTriangles) {
      delete triangles;
      triangles = psTriangles;
   }
}

PointSet *UnstructuredInCoreTriangleMesh::getTriangles() {
   return triangles;
}

UnstructuredInCoreTriangleMesh *UnstructuredInCoreTriangleMesh::createCopyWithNewTriangles(PointSet *psTriangles) const {
   PointSet *tcopy = this->triangles;
   ((UnstructuredInCoreTriangleMesh*)this)->triangles = NULL;
   UnstructuredInCoreTriangleMesh *result = (UnstructuredInCoreTriangleMesh*)copy();
   ((UnstructuredInCoreTriangleMesh*)this)->triangles = tcopy;
   result->triangles = psTriangles;
   return result;
}

UnstructuredInCoreTriangleMesh * X4::UnstructuredInCoreTriangleMesh::createCopyOfGivenTriangles( const std::vector<int32> &indexes )
{

	UnstructuredInCoreTriangleMesh *result = new UnstructuredInCoreTriangleMesh();
	result->setPointSet((PointSet*)this->getPointSet()->copy());
	PointSet* givenTriangles=new PointSet();
	givenTriangles->clearAndSetup(1, indexes.size(), this->getTriangles()->getDescr());
	AAT INDEX=this->getTriangles()->getAAT("index");
	for (card32 i=0;i<indexes.size();i++)
	{
		Vector3i tri=this->getTriangles()->get3i(indexes[i],INDEX);
		givenTriangles->set3i(i,INDEX,tri);
	}

	result->setTriangles(givenTriangles);
	result->clearGarbage();
	return result;
}

static bool SortRemovePointPredicate (mpcard i, mpcard j) { return (i>j); }
static void __removePoints(PointSet* upc, const std::vector<mpcard>& _points)
{
	if (_points.size() == 0) return;

	std::vector<mpcard> points = _points;
	std::sort(points.begin(), points.end(), SortRemovePointPredicate);

	unsigned oldSize = upc->getHeight();

	// std::map of new indices
	std::vector<int> indices(upc->getHeight());
	std::vector<int> lastVertices(upc->getHeight(), -1);
	for (size_t v=0; v < indices.size(); v++)
		indices[v] = (int)v;

	int lastIndex = upc->getHeight() - 1;
	size_t newSize = upc->getHeight() - points.size();
	for (std::vector<mpcard>::iterator it = points.begin(); it != points.end(); it++)
	{
		if (*it < newSize)
		{
			for (; indices[lastIndex] == -1; lastIndex--)
				x4Assert(lastIndex >= (int)newSize);

			indices[*it] = lastIndex;
			lastVertices[lastIndex] = *it;
			indices[lastIndex] = -1;
		}else
		{
			x4Assert(indices[*it] >= 0);
			indices[*it] = -1;
		}
	}

	for (size_t v=0; v < indices.size(); v++)
		if (indices[v] == v) lastVertices[v] = (int)v;

	// do swapping
	for (unsigned i=0; i < newSize; i++)
	{
		// do nothing if index not changed
		x4Assert(indices[i] >= 0);
		if (i == indices[i]) continue;

		// copy from the stored index to correct position
		for (mpcard j=0; j < upc->getWidth(); j++)
			upc->setVertex(j, i, upc->getDataPointer(j, indices[i]));
	}

	upc->changeHeight(newSize);
}

void X4::UnstructuredInCoreTriangleMesh::cleanTopology()
{
	std::vector<mpcard> todelete;
	AAT indexAAT = triangles->getAAT("index");
	std::map<card64,card64> triMap;
	for( unsigned i=0;i<triangles->getNumEntries(); i++)
	{
		Vector3i tri = triangles->get3i(i,indexAAT);
		bool erasetriangle = ( tri[0] == tri[1] || tri[0] == tri[2] || tri[2] == tri[1]);

		Vector3i trisorted = tri;
		if (trisorted[0] > trisorted[1]) swap(trisorted[0], trisorted[1]);
		if (trisorted[1] > trisorted[2]) swap(trisorted[1], trisorted[2]);
		if (trisorted[0] > trisorted[1]) swap(trisorted[0], trisorted[1]);
		if (trisorted[1] > trisorted[2]) swap(trisorted[1], trisorted[2]);

		//unsigned minVI = 0;
		//if( tri[1] < tri[minVI] ) minVI = 1;
		//if( tri[2] < tri[minVI] ) minVI = 2;
		//card64 tri_id = tri[minVI]*ps->getNumEntries() + tri[(minVI+1)%3];

		card64 tri_id = (trisorted[0]*ps->getNumEntries() + trisorted[1]) * ps->getNumEntries() + trisorted[2];
		if( triMap.find(tri_id) != triMap.end() )
			erasetriangle = true;
		else
			triMap[tri_id] = i;

		if( erasetriangle && find(todelete.begin(), todelete.end(), i) == todelete.end())
			todelete.push_back(i);
	}
	__removePoints(triangles, todelete);
}

void X4::UnstructuredInCoreTriangleMesh::extractMeshBorder( std::vector<mpcard> &borderEdges )
{
	borderEdges.clear();
	AAT indexAAT = triangles->getAAT("index");
	set<card64> edgeSet;
	for( unsigned i=0;i<triangles->getNumEntries();i++ )	
	{
		Vector3i tri = triangles->get3i(i,indexAAT);
		for( unsigned j=0;j<3;j++ )
		{
			mpcard i0 = tri[j];
			mpcard i1 = tri[(j+1)%3];
			card64 id = min(i0,i1)*ps->getNumEntries() + max(i0,i1);
			if( edgeSet.find(id) != edgeSet.end())
				edgeSet.erase(id);
			else
				edgeSet.insert(id);
		}
	}

	for( unsigned i=0;i<triangles->getNumEntries();i++ )	
	{
		Vector3i tri = triangles->get3i(i,indexAAT);
		for( unsigned j=0;j<3;j++ )
		{
			mpcard i0 = tri[j];
			mpcard i1 = tri[(j+1)%3];
			card64 id = min(i0,i1)*ps->getNumEntries() + max(i0,i1);
			if( edgeSet.find(id) != edgeSet.end())
				borderEdges.push_back(i*3+j);
		}
	}
}


std::vector<UnstructuredInCoreTriangleMesh*> UnstructuredInCoreTriangleMesh::split(InCorePCTopologyGraph* tpg )
{
	bool tpgcomputed = false;
	if (tpg == NULL)
	{
		tpg = computeTopology(false);
		tpgcomputed = true;
	}

	std::vector<UnstructuredInCoreTriangleMesh*> results;

	std::vector<char> flags(getNumPoints(), 0);
	for (unsigned i=0; i < flags.size(); i++)
	{
		if (flags[i] == 1) continue;

		std::vector<mpcard> component;

		PointCloudIterator* iter = tpg->createIterator(getPointSet());
		TopologicalKNNIterator_Fast* tIt = new TopologicalKNNIterator_Fast(this, iter);

		tIt->setStartPointAndReset(i);
		while(!tIt->atEnd())
		{
			flags[tIt->getPointSetPointNumber()] = 1;
			component.push_back(tIt->getPointSetPointNumber());
			tIt->next();
		}

		delete tIt;
		delete iter;

		results.push_back(extractSubMesh(component));
	}

	if (tpgcomputed) delete tpg;


	return results;
}

void X4::UnstructuredInCoreTriangleMesh::clearAndSetupDefault( mpcard numPoints, mpcard numTriangles, bool hasNormal/*=true*/, bool hasColor /*= true*/, bool hasFlags /*= true*/ )
{
	UnstructuredInCorePointCloud::clearAndSetup(numPoints,true,hasNormal,hasColor,hasFlags);
	VertexDescriptor vdTri;
	vdTri.pushAttrib(VAD("index",3,VAD::DATA_FORMAT_INT32));
	triangles->clearAndSetup(1,numTriangles,&vdTri);
}

void X4::UnstructuredInCoreTriangleMesh::flipOrientation()
{
	VertexArray tri( getTriangles());
	for( unsigned i=0;i<tri.getNumElements();i++ )
	{
		Vector3i t = tri.getIndex3i(i);
		swap(t[0],t[2]);
		tri.setIndex3i(i,t);
	}
}
void X4::UnstructuredInCoreTriangleMesh::computeBoundaryTopologyAttachment(){
	InCorePCTopologyGraph * graph = new InCorePCTopologyGraph;
	VertexDescriptor vd;
	vd.pushAttrib(VAD("vindex", 2, VAD::DATA_FORMAT_INT32));
	vd.pushAttrib(VAD("edgeDistance", 1, VAD::DATA_FORMAT_INT32));
	graph->clearAndSetup(&vd);

	std::map<int64,int64> edgeMap;
	VertexArray tri( getTriangles() );
	VertexArray vert( getPointSet() );

	for( card32 i=0;i<tri.getNumElements();i++ )
	{
		for( card32 e=0;e<3;e++ )
		{
			int64 i1 = tri.getIndex3i(i)[e];
			int64 i2 = tri.getIndex3i(i)[(e+1)%3];
			int64 index = min(i1,i2) * getNumPoints() + max(i1,i2);
			if( edgeMap.find(index) == edgeMap.end() )
				edgeMap[index] = index;
			else
				edgeMap.erase( edgeMap.find(index) );
		}
	}


	for( std::map<int64,int64>::iterator iter = edgeMap.begin(); iter != edgeMap.end(); ++iter )
	{
		int64 i1 = iter->second / (int64)getNumPoints();
		int64 i2 = iter->second % (int64)getNumPoints();

		graph->addEdge(i1, i2);

	}

	TopologyAttachment * topAtt = new TopologyAttachment;
	topAtt->setup("BoundaryTopologyAtt",AttachedData::ADF_PERSISTENT);
	topAtt->setTopology(graph);
	this->getAttachments()->attachData(topAtt);


}


/// angle at b in a-b-c
float32 calcAngle( const Vector3f &pointA, const Vector3f &pointB, const Vector3f &pointC )
{
	
	// get VecA endA in direction of startA
	Vector3f vecA = pointB - pointA;

	//get VecB form startB in dir or EndB
	Vector3f vecB = pointB - pointC;

	//normalize a and b
	vecA = normalize(vecA);
	vecB = normalize(vecB);

	//calculate float angle
	float32 dotProd = vecA * vecB;
	if (dotProd > 1)
		dotProd = 1;
	if (dotProd < -1)
		dotProd = -1;

	float32 extrinsicAngle = acos( dotProd );
	return extrinsicAngle;
}


void X4::UnstructuredInCoreTriangleMesh::computeTopologyAttachment(bool  computeCotangentWeights)
{
	InCorePCTopologyGraph * graph = computeTopology(computeCotangentWeights);

	TopologyAttachment * topAtt = new TopologyAttachment;
	topAtt->setup(TopologyAttachment::getDefaultName(),AttachedData::ADF_PERSISTENT);
	topAtt->setTopology(graph);
	this->getAttachments()->attachData(topAtt);
}

X4::InCorePCTopologyGraph* X4::UnstructuredInCoreTriangleMesh::computeTopology(bool  computeCotangentWeights)
{
	InCorePCTopologyGraph * graph = new InCorePCTopologyGraph;
	VertexDescriptor vd;
	vd.pushAttrib(VAD("vindex", 2, VAD::DATA_FORMAT_INT32));
	vd.pushAttrib(VAD("edgeDistance", 1, VAD::DATA_FORMAT_FLOAT32));
	if (computeCotangentWeights)
		vd.pushAttrib(VAD("weight", 1, VAD::DATA_FORMAT_FLOAT32));
	
	graph->clearAndSetup(&vd);

	AAT IND = triangles->getAAT("index");
	for (mpcard i = 0; i <  triangles->getNumEntries(); i++)
	{
		Vector3i triInd = triangles->get3i(i, IND);
		graph->addEdge(triInd[0], triInd[1]);
		graph->addEdge(triInd[0], triInd[2]);
		graph->addEdge(triInd[1], triInd[2]);
	}
	

	if (!computeCotangentWeights)
		return graph;

	//calculate distances
	PointSet* edges = graph->getEdges();
	AAT POS = ps->getAAT("position");
	AAT VIND = edges->getAAT("vindex");
	AAT DIST = edges->getAAT("edgeDistance");

	for (mpcard i = 0; i<graph->getNumEdges(); ++i)
	{
		Vector2i edgeInd = edges->get2i(i, VIND);
		float32 dist = norm( ps->get3f(edgeInd[0], POS) - ps->get3f(edgeInd[1], POS) );
		edges->set1f(i, DIST, dist);
	}

	//compute cotangent weight for every edge
	AAT C_WEIGHT = edges->getAAT("weight");

	for (mpcard i = 0; i<graph->getNumEdges(); ++i)
	{
		//get triangles of edge
		Vector2i edgeInd = edges->get2i(i, VIND);
		int32 a = edgeInd[0];
		int32 b = edgeInd[1];
		std::vector<int32> thirdVertexesForTriangles;

		mpcard numNeiA = graph->getNumAdjacentVertices(a);
		std::vector<int32> neiA(numNeiA);
		for (mpcard j = 0; j < numNeiA; ++j)
			neiA[j] = graph->getAdjacentVertex(a, j);

		mpcard numNeiB = graph->getNumAdjacentVertices(b);
		for (mpcard jb = 0; jb < numNeiB; ++jb){
			int32 neiB = graph->getAdjacentVertex(b, jb);
			for (mpcard comp = 0; comp < numNeiA; ++comp)
			{
				if (neiB == neiA[comp])
				{
					thirdVertexesForTriangles.push_back(neiB);
				}
			}
		}
		//debugOutput << thirdVertexesForTriangles.size() << "\n";

		// compute angles
		std::vector<float32> angles;
		for (mpcard nTriangle = 0; nTriangle < thirdVertexesForTriangles.size(); ++nTriangle)
		{
			int32 thirdForTriangle = thirdVertexesForTriangles[nTriangle];		
			angles.push_back(calcAngle(ps->get3f(a, POS), ps->get3f(thirdForTriangle, POS), ps->get3f(b, POS)));
		}

		// calc cotan
		std::vector<float32> weights(thirdVertexesForTriangles.size());
		for (mpcard nTriangle = 0; nTriangle < thirdVertexesForTriangles.size(); ++nTriangle)
		{
			float32 angle = angles[nTriangle];
			x4Assert(angle>=0.0f);
			if (  (fabs(angle - M_PI_2) < 0.001f) )
			{
				weights[nTriangle] = 0.0f;
				continue;
			}
			if (fabs(angle - M_PI) < 0.001f  || angle < 0.001f)			
			{
				weights[nTriangle] = 10.0f;
				continue;
			}

			float32 tangensAngle = tan(angle);
			float32 coTanW = 1.0f/tangensAngle;
			weights[nTriangle] = coTanW;
		}

		// add weights
		float32 avWeight = 0;
		for (mpcard nTriangle = 0; nTriangle < thirdVertexesForTriangles.size(); ++nTriangle)
		{
			avWeight += weights[nTriangle];
		}
		avWeight/= (float)weights.size();
		
		edges->set1f(i, C_WEIGHT, avWeight);
	}

	return graph;
}

UnstructuredInCoreTriangleMesh* X4::UnstructuredInCoreTriangleMesh::extractSubMesh(const std::vector<mpcard>& vertices, ExtractFunction func )
{
	UnstructuredInCoreTriangleMesh* result = new UnstructuredInCoreTriangleMesh;
	result->clearAndSetup(this->getDescr(), 0, this->getTriangleDescr(), 0);
	PointSet* ps = result->getPointSet();
	PointSet* orgPS = this->getPointSet();
	std::vector<mpint> indexConverter(this->getNumPoints(), -1);
	for (mpcard i = 0; i < vertices.size(); i++)
	{
		ps->addPoint(orgPS->getDataPointer(vertices[i]));
		indexConverter[vertices[i]] = i;
	}

	std::vector< mpcard> sortedFirstInstance =  vertices;
	sort(sortedFirstInstance.begin(),  sortedFirstInstance.end());

	//tris
	PointSet* orgTris = this->getTriangles();
	VertexArray orgTriArr(orgTris);
	AAT INDAAT = orgTris->getAAT("index");
	for (mpcard i = 0; i < orgTris->getNumEntries(); i++)
	{
		Vector3i thisTri = orgTriArr.getIndex3i(i);
		//check if all three vertices belong to instance
		if (!binary_search(sortedFirstInstance.begin(), sortedFirstInstance.end(), thisTri[0]))
			continue;
		if (!binary_search(sortedFirstInstance.begin(), sortedFirstInstance.end(), thisTri[1]))
			continue;
		if (!binary_search(sortedFirstInstance.begin(), sortedFirstInstance.end(), thisTri[2]))
			continue;

		// tri is IN
		//convert indexes
		Vector3i newTri = makeVector3i(indexConverter[thisTri[0]], indexConverter[thisTri[1]], indexConverter[thisTri[2]]);
		result->getTriangles()->addPoint(orgTris->getDataPointer(i));
		result->getTriangles()->set3i(result->getTriangles()->getNumEntries()-1, INDAAT,newTri);
	}

	// mark all vertices and cut them to the hell
	if (func == CUT_ALONG_BORDER)
	{
		set<mpcard> border;

		// remove only those vertices, which are in the submesh and are not shared by any of the triangles within the main mesh
		for (mpcard i = 0; i < orgTris->getNumEntries(); i++)
		{
			Vector3i thisTri = orgTriArr.getIndex3i(i);

			char b1 = (binary_search(sortedFirstInstance.begin(), sortedFirstInstance.end(), thisTri[0]));
			char b2 = (binary_search(sortedFirstInstance.begin(), sortedFirstInstance.end(), thisTri[1]));
			char b3 = (binary_search(sortedFirstInstance.begin(), sortedFirstInstance.end(), thisTri[2]));

			// if at least one vertex of the triangle is not in the submesh, then the whole triangle must survive
			if (b1+b2+b3 == 1 || b1+b2+b3 == 2)
			{
				if (b1) border.insert(thisTri[0]);
				if (b2) border.insert(thisTri[1]);
				if (b3) border.insert(thisTri[2]);
			}
		}
		for (std::vector<mpcard>::iterator it=sortedFirstInstance.begin(); it != sortedFirstInstance.end(); )
		{
			if (binary_search(border.begin(), border.end(), *it)) it = sortedFirstInstance.erase(it);
			else it++;
		}
	}

	if (func == CUT_ALONG_BORDER || func == CUT)
	{
		// find all triangles which are not valid now
		std::vector<mpcard> todelete;
		for (mpcard i = 0; i < orgTris->getNumEntries(); i++)
		{
			Vector3i thisTri = orgTriArr.getIndex3i(i);

			bool p1 = (binary_search(sortedFirstInstance.begin(), sortedFirstInstance.end(), thisTri[0]));
			bool p2 = (binary_search(sortedFirstInstance.begin(), sortedFirstInstance.end(), thisTri[1]));
			bool p3 = (binary_search(sortedFirstInstance.begin(), sortedFirstInstance.end(), thisTri[2]));

			if (p1 || p2 || p3)  //delete this tri
				todelete.push_back(i);
		}
		deleteMarkedTriangles(todelete);
	}

	return result;


}

void X4::UnstructuredInCoreTriangleMesh::joinOtherMesh( UnstructuredInCoreTriangleMesh* other, const Matrix4f &T )
{
	AAT posAAT = getAAT("position");
	bool hasNormal = providesAttribute("normal");
	AAT normalAAT;
	if( hasNormal)
		normalAAT = getAAT("normal");

	Matrix3f TR = shrink4To3(T);
	TR = invertMatrix(TR).transpose();

	PointSet* otherPS = other->getPointSet();
	mpcard firstIndexOfNewPoints = ps->getNumEntries();
	for (mpcard i = 0; i < otherPS->getNumEntries(); i++)
	{
		ps->addPoint(otherPS->getDataPointer(i));
		Vector3f pos = otherPS->get3f(i,posAAT);
		ps->set3f(firstIndexOfNewPoints+i, posAAT, transformVector3f(T,pos));
		if( hasNormal)
		{
			Vector3f normal = otherPS->get3f(i,normalAAT);
			ps->set3f(firstIndexOfNewPoints+i,normalAAT,TR*normal);
		}
	}

	mpcard firstIndexOfNewTriangles = triangles->getNumEntries();
	Vector3i moveVec = makeVector3i(firstIndexOfNewPoints, firstIndexOfNewPoints, firstIndexOfNewPoints);

	PointSet* otherTriangles = other->getTriangles();
	for (mpcard i = 0; i < other->getNumTriangles(); i++)
	{
		triangles->addPoint(otherTriangles->getDataPointer(i));
	}

	VertexArray vert(triangles);
	for (mpcard i = firstIndexOfNewTriangles; i < getNumTriangles(); i++)
	{
		vert.setIndex3i(i, vert.getIndex3i(i) + moveVec);
	}

}
const VertexDescriptor *UnstructuredInCoreTriangleMesh::getTriangleDescr() const {
   return triangles->getDescr();
}

PointCloud *UnstructuredInCoreTriangleMesh::appendTriangleFields(const VertexDescriptor *newAttribs) const {
   return createCopyWithNewTriangles(triangles->appendFields(newAttribs));
}

void UnstructuredInCoreTriangleMesh::clearAndSetupTriangles(const VertexDescriptor *triangleDescr) {
   triangles->clearAndSetup(1, 0, triangleDescr);
}

card64 UnstructuredInCoreTriangleMesh::getNumTriangles() const {
   return triangles->getNumEntries();
}

UnstructuredInCoreTriangleMesh::~UnstructuredInCoreTriangleMesh() {
   delete triangles;
}




std::string UnstructuredInCoreTriangleMesh::calcHashString( const Vector3f &vec )
{
   std::string ret;
   card8 * tmp = (card8 *)&vec[0];
   for( card32 i=0;i<3*4;i++ )
      ret += tmp[i];
   return ret;
}

void UnstructuredInCoreTriangleMesh::closeMesh(bool checkNormals)
{
   MyHashTable hashTable;

   // new vertex index = indexTable[ old vertex index ]
   std::vector<card32> indexTable(getPointSet()->getNumEntries());
   card32 numVertices = 0;

   VertexArray array(getPointSet());

   for( card32 i=0;i<getPointSet()->getNumEntries();i++ )
   {
      std::string hashID = calcHashString( array.getPosition3f(i) );
	  if( checkNormals )
		  hashID += calcHashString( array.getNormal3f(i) );
      card32 index = 0;
      if( hashTable.getData( hashID, index ) )
      {
         // vertex exists, take index from existing vertex
         indexTable[i] = index;
      }
      else
      {
         // new vertex
         indexTable[i] = numVertices;
         array.copyVertex( i, numVertices );
         numVertices++;
         hashTable.addElement( hashID, indexTable[i] );
      }
   }

   // shrink ps
   if( numVertices != getPointSet()->getNumEntries() )
   {
      PointSet * newPS = new PointSet;
      newPS->clearAndSetup( 1, numVertices, getPointSet()->getDescr() );

      std::vector<char> tmpBuffer( getPointSet()->getDescr()->getSize() );

      for( card32 i=0;i<numVertices;i++ )
      {
         getPointSet()->getVertex(i, &tmpBuffer[0]);
         newPS->setVertex(i, &tmpBuffer[0] );
      }
      setPointSet(newPS);
   }

   // adopt triangles
   VertexArray triangleArray(triangles);
   mpcard numTriangles = 0;
   for( card32 i=0;i<triangles->getNumEntries();i++ )
   {
      Vector3i index = triangleArray.getIndex3i(i);
      for( card32 c=0;c<3;c++ )
         index[c] = indexTable[index[c]];
	  if( index[0] == index[1] || index[0] == index[2] || index[1] == index[2] )
		  continue;
      triangleArray.setIndex3i(numTriangles++,index);
   }
   if( numTriangles != triangles->getNumEntries())
   {
		debugOutput << "removed " << triangles->getNumEntries() - numTriangles << " triangles\n";
		triangles->changeHeight(numTriangles);
   }
}

void UnstructuredInCoreTriangleMesh::openMesh()
{
	card32 numVertices = triangles->getNumEntries() * 3;
	
	std::vector<card32> indexTable(numVertices);

	// iterate over all faces and add three points per face
	VertexArray triangleArray(triangles);
	for( card32 i=0; i < triangles->getNumEntries(); i++ )
	{
		Vector3i index = triangleArray.getIndex3i(i);
		indexTable[i*3 + 0] = index[0];
		indexTable[i*3 + 1] = index[1];
		indexTable[i*3 + 2] = index[2];

		triangleArray.setIndex3i(i, makeVector3i(i*3+0, i*3+1, i*3+2));
	}

	// create from the index array a point set which holds the points
	{
		PointSet * newPS = new PointSet;
		newPS->clearAndSetup( 1, numVertices, getPointSet()->getDescr() );

		std::vector<char> tmpBuffer( getPointSet()->getDescr()->getSize() );

		for( card32 i=0; i < numVertices; i++ )
		{
			getPointSet()->getVertex(indexTable[i], &tmpBuffer[0]);
			newPS->setVertex(i, &tmpBuffer[0] );
		}
		setPointSet(newPS);
	}

	// clear up buffers
	getHierarchyAttachedDSs()->clear();

}

void UnstructuredInCoreTriangleMesh::calculateVertexNormals()
{
   x4Assert( getPointSet()->providesAttribute( "normal" ) );

   VertexArray vertArray( getPointSet() );
   VertexArray triArray( triangles );

	// init normals to 0
	for( card32 i=0;i<vertArray.getNumElements();i++ )
	{
		vertArray.setNormal3f( i, makeVector3f(0,0,0));
	}

   // calculate triangle normal and add to vertex normals
	for( card32 i=0;i<triArray.getNumElements();i++ )
   {
      Vector3i tri = triArray.getIndex3i(i);
      Vector3f p1 = vertArray.getPosition3f( tri[0] );
      Vector3f p2 = vertArray.getPosition3f( tri[1] );
      Vector3f p3 = vertArray.getPosition3f( tri[2] );
      Vector3f normal = (p2-p1).crossProduct(p2-p3);
      //normal.normalize();
      vertArray.setNormal3f(tri[0], vertArray.getNormal3f(tri[0]) + normal );
      vertArray.setNormal3f(tri[1], vertArray.getNormal3f(tri[1]) + normal );
      vertArray.setNormal3f(tri[2], vertArray.getNormal3f(tri[2]) + normal );
   }

   // normalize normals
   for( card32 i=0;i<vertArray.getNumElements();i++ )
   {
      Vector3f normal = vertArray.getNormal3f(i);
      normal.normalize();
		normal= -normal;
      vertArray.setNormal3f(i,normal);
   }

}

void UnstructuredInCoreTriangleMesh::clearGarbage()
{
   VertexArray triArray( triangles );

   PointSet * tmpPS = new PointSet;
   tmpPS->clearAndSetup( 1, getPointSet()->getNumEntries(), getPointSet()->getDescr() );

   std::vector<int32> moveTo( getPointSet()->getNumEntries() );
   for( card32 i=0;i<moveTo.size();i++ )
      moveTo[i] = -1;

   int32 numVertices = 0;
   std::vector<char> tmpBuffer( getPointSet()->getDescr()->getSize() );
   for( card32 i=0;i<triArray.getNumElements();i++ )
   {
      Vector3i tri = triArray.getIndex3i(i);

//       if( tri[0] == tri[1]
//          || tri[1] == tri[2]
//          || tri[0] == tri[2] )
//       {
//          triArray.setIndex3i( i, makeVector3i(0,0,0) );
//          continue;
//       }

      for( card32 j=0;j<3;j++ )
      {
         if( tri[j] < 0 )
            continue;
         if( tri[j] >= (int32)getPointSet()->getNumEntries() )
            continue;
         if( moveTo[ tri[j] ] == -1 )
         {
            getPointSet()->getVertex(tri[j], &tmpBuffer[0]);
            //getPointSet()->getVertex(getPointSet()->getNumEntries()-1, &tmpBuffer[0] );
            tmpPS->setVertex(numVertices, &tmpBuffer[0] );
            moveTo[ tri[j] ] = numVertices;
            tri[j] = numVertices;
            numVertices++;
         }
         else
         {
            tri[j] = moveTo[ tri[j] ];
         }

      }

      triArray.setIndex3i( i, tri );

   }
   // shrink ps
   PointSet * newPS = new PointSet;
   newPS->clearAndSetup( 1, numVertices, getPointSet()->getDescr() );

   for( card32 i=0;i<newPS->getNumEntries();i++ )
   {
      tmpPS->getVertex(i, &tmpBuffer[0] );
      newPS->setVertex(i, &tmpBuffer[0] );
   }
   setPointSet(newPS);
   delete tmpPS;

}


void UnstructuredInCoreTriangleMesh::snapMesh( float max_tolerance, bool snapSelectedPointsOnly )
{
   std::vector<int32> moveTo( getPointSet()->getNumEntries() );
   std::vector<int32> done( getPointSet()->getNumEntries(),0);
   for( card32 i=0;i<moveTo.size();i++ )
      moveTo[i] = i;

   clearAttachments();
   HierarchicalKNNIterator * iter = new HierarchicalKNNIterator( this, 32, NULL );
   iter->setMaxDistanceToSeekPoint(max_tolerance);

   VertexArray vertices( getPointSet() );
   VertexArray triArray( triangles );
   unsigned numSnapped = 0;
   AAT flagsAAT;
   if(snapSelectedPointsOnly )
	   flagsAAT = getPointSet()->getAAT("flags");

   for( card32 i=0;i<moveTo.size();i++ )
   {
	   if( snapSelectedPointsOnly)
		   if( (vertices.getFlags(i)&PF_FLAG_SELECTED) == 0)
			   continue;
	   if( done[i] )
		   continue;
	   done[i] = 1;
      Vector3f pos = vertices.getPosition3f(i);
      iter->setSeekPointAndReset( pos );

      while(!iter->atEnd())
      {
         int32 tmpI = iter->getPointSetPointNumber();
         //float32 dist = norm( pos - vertices.getPosition3f( moveTo[ tmpI ] ) );
         if( done[tmpI]==0 && (!snapSelectedPointsOnly || (iter->get1i(flagsAAT)&PF_FLAG_SELECTED)))
         {
			numSnapped++;
            moveTo[tmpI] = moveTo[i];
            done[tmpI]=1;
			//break;
         }
         //if( dist > max_tolerance )
         //   break;

         iter->next();
      }

   }
   debugOutput << "num snapped: " << numSnapped << "\n";

	card32 numTriangles=0;

	for( card32 i=0;i<triArray.getNumElements();i++ )
	{
		Vector3i tri = triArray.getIndex3i( i );
		tri[0] = moveTo[ tri[0] ];
		tri[1] = moveTo[ tri[1] ];
		tri[2] = moveTo[ tri[2] ];

		if (tri[0]==tri[1] || tri[0]==tri[2] || tri[2]==tri[1])
		{
			continue;
		}

		if( i != numTriangles)
			triArray.copyVertex( i, numTriangles );
		triArray.setIndex3i(numTriangles,tri);
		numTriangles++;
	}

   if( numTriangles < triArray.getNumElements())
   {
	   debugOutput << "removed " << triArray.getNumElements() - numTriangles << " triangles\n";
	   triangles->changeHeight(numTriangles);
   }

   delete iter;

   clearGarbage();
}

void UnstructuredInCoreTriangleMesh::deletePoints(const std::vector<mpcard>& _todelete)
{
	VertexArray vertArray( getPointSet() );
	VertexArray triArray( triangles );
	std::vector<mpcard> pts = _todelete;
	sort(pts.begin(), pts.end());
	std::vector<mpcard> todelete;

	// calculate triangle normal and add to vertex normals
	for (unsigned i = 0; i < triArray.getNumElements(); i++)
	{
		Vector3i triIndexes = triArray.getIndex3i(i);
		bool p1 = binary_search(pts.begin(), pts.end(), triIndexes[0]);
		bool p2 = binary_search(pts.begin(), pts.end(), triIndexes[1]);
		bool p3 = binary_search(pts.begin(), pts.end(), triIndexes[2]);

		if (p1 || p2 || p3)  //delete this tri
			todelete.push_back(i);
	}
	deleteMarkedTriangles(todelete);
}

void UnstructuredInCoreTriangleMesh::deleteMarkedTriangles()
{
	VertexArray vertArray( getPointSet() );
	VertexArray triArray( triangles );
	std::vector<mpcard> todelete;

	// calculate triangle normal and add to vertex normals
	for (unsigned i = 0; i < triArray.getNumElements(); i++)
	{
		Vector3i triIndexes = triArray.getIndex3i(i);
		bool p1 = (vertArray.getFlags( triIndexes[0] )& PF_FLAG_SELECTED);
		bool p2 = (vertArray.getFlags( triIndexes[1] )& PF_FLAG_SELECTED);
		bool p3 = (vertArray.getFlags( triIndexes[2] )& PF_FLAG_SELECTED);

		if (p1 || p2 || p3)  //delete this tri
			todelete.push_back(i);
	}
	deleteMarkedTriangles(todelete);
}

void UnstructuredInCoreTriangleMesh::deleteMarkedTriangles(const std::vector<mpcard>& _remove)
{
	//VertexArray vertArray( getPointSet() );
	VertexArray triArray( triangles );

	int32 lastTriIndex=triArray.getNumElements()-1;
	std::vector<mpcard> remove = _remove;
	sort(remove.begin(), remove.end());
	unique(remove.begin(), remove.end());

	std::vector<mpcard>::iterator remit = remove.begin();
	int ii = lastTriIndex;
	for (ii=lastTriIndex; ii >= 0 && remit != remove.end(); ii--)
	{
		if (!binary_search(remove.begin(), remove.end(), ii))
		{
			triArray.copyVertex(ii, *remit);
			remove.erase(remove.begin());
		}else
			remove.erase(find(remove.begin(),remove.end(),ii));
	}
	lastTriIndex = ii;
	/*
	for (unsigned i=0; i < remove.size(); i++)
	{
		while (binary_search(remove.begin(), remove.end(), lastTriIndex))
			lastTriIndex--;
		triArray.copyVertex(lastTriIndex--, remove[i]);
		remove[i] = (mpcard)-1;
	}*/

	/*for (int32 i = 0; i <= lastTriIndex; i++)
	{
		Vector3i triIndexes = triArray.getIndex3i(i);
		bool p1 = (vertArray.getFlags( triIndexes[0])& PF_FLAG_SELECTED);
		bool p2 = (vertArray.getFlags( triIndexes[1] )& PF_FLAG_SELECTED);
		bool p3 = (vertArray.getFlags( triIndexes[2] )& PF_FLAG_SELECTED);

		if (p1 || p2 || p3)  //delete this tri
		{
			triArray.copyVertex(lastTriIndex, i);
			lastTriIndex--;
			i--;
		}
	}*/
	triangles->changeHeight(lastTriIndex+1);

	clearGarbage();
	//cleanTopology();
	clearAttachments();

	// in case we have polygon information we need to adapt this to the current state
	if( triangles->providesAttribute("polyIndex") )
	{
		AAT polyindexAAT = triangles->getAAT("polyIndex");
		mpcard numPolygons = 0;
		std::map<int32,mpcard> polygonMap;
		for( mpcard i=0;i<triangles->getNumEntries();i++ )
		{
			int32 index = triangles->get1i(i,polyindexAAT);
			if( polygonMap.find(index) == polygonMap.end() )
			{
				polygonMap[index] = numPolygons;
				numPolygons++;
			}
			triangles->set1i(i,polyindexAAT,polygonMap[index]);
		}
	}
	
}

//#include <public/Rendering/DebugRenderer.h>
void UnstructuredInCoreTriangleMesh::upsample2(float sampleSpacing/*, bool useDebugRenderer*/)
{
	PointSet * ps = getPointSet();
	PointSet * ps_tri = getTriangles();

	AAT posAAT = getAAT("position");
	AAT normalAAT = getAAT("normal");
	AAT colorAAT = getAAT("color");
	AAT indexAAT = ps_tri->getAAT("index");

	std::vector<UnstructuredInCoreTriangleMesh*> trimesh(ps_tri->getNumEntries(), NULL);

	UnstructuredInCoreTriangleMesh* newMesh = new UnstructuredInCoreTriangleMesh;
	newMesh->clearAndSetup(getDescr(), 0, getTriangleDescr(), 0);

	//debugRenderer->clearDebugData("triangle");
	//debugRenderer->beginRenderJob("triangle");

	// check each triangle how far it can be splitted until each of its side is almost sampleSpacing long
	for (unsigned i = 0; i < ps_tri->getNumEntries(); i++)
	{
		struct Triangle { Vector3f pos[3]; };
		std::vector<Triangle> newTriangles;

		// we split every triangle as long as its sides are larger then the sampleSpacing
		queue<Triangle> tris; 
		Triangle maintri;
		Vector3f normals[3];
		Vector3f colors[3];
		{
			Vector3i triangle = ps_tri->get3i(i, indexAAT);
			maintri.pos[0] = ps->get3f(triangle[0], posAAT);
			maintri.pos[1] = ps->get3f(triangle[1], posAAT);
			maintri.pos[2] = ps->get3f(triangle[2], posAAT);
			tris.push(maintri);

			normals[0] = ps->get3f(triangle[0], normalAAT);
			normals[1] = ps->get3f(triangle[1], normalAAT);
			normals[2] = ps->get3f(triangle[2], normalAAT);

			if (colorAAT != NULL_AAT)
			{
				colors[0] = ps->get3f(triangle[0], colorAAT);
				colors[1] = ps->get3f(triangle[1], colorAAT);
				colors[2] = ps->get3f(triangle[2], colorAAT);
			}
		}

		while(!tris.empty())
		{
			Triangle tri = tris.front(); tris.pop();
			
			float hdist[3];
			for (unsigned j=0; j < 3; j++)
			{
				float dist = norm(tri.pos[(j+1) % 3] - tri.pos[j]);
				hdist[j] = dist * 0.5;
			}

			Vector3f p[6];
			p[0] = tri.pos[0];
			p[1] = tri.pos[0] + (tri.pos[1] - tri.pos[0]) * 0.5f;
			p[2] = tri.pos[1];
			p[3] = tri.pos[1] + (tri.pos[2] - tri.pos[1]) * 0.5f;
			p[4] = tri.pos[2];
			p[5] = tri.pos[2] + (tri.pos[0] - tri.pos[2]) * 0.5f;

			// if all three sites can be splitted, then do so
			if (hdist[0] > sampleSpacing && hdist[1] > sampleSpacing && hdist[2] > sampleSpacing)
			{
				// split each side in half and put 4 new triangles into the queue
				Triangle t[4];

				t[0].pos[0] = p[0];
				t[0].pos[1] = p[1];
				t[0].pos[2] = p[5];

				t[1].pos[0] = p[1];
				t[1].pos[1] = p[2];
				t[1].pos[2] = p[3];

				t[2].pos[0] = p[3];
				t[2].pos[1] = p[4];
				t[2].pos[2] = p[5];

				t[3].pos[0] = p[1];
				t[3].pos[1] = p[3];
				t[3].pos[2] = p[5];

				tris.push(t[0]);
				tris.push(t[1]);
				tris.push(t[2]);
				tris.push(t[3]);

			// if two of the sides can be splitted
			}else if (hdist[0] > sampleSpacing && hdist[1] > sampleSpacing)
			{
				Triangle t[3];

				t[0].pos[0] = p[0];
				t[0].pos[1] = p[3];
				t[0].pos[2] = p[1];

				t[1].pos[0] = p[1];
				t[1].pos[1] = p[2];
				t[1].pos[2] = p[3];

				t[2].pos[0] = p[0];
				t[2].pos[1] = p[3];
				t[2].pos[2] = p[4];

				tris.push(t[0]);
				tris.push(t[1]);
				tris.push(t[2]);

			}else if (hdist[0] > sampleSpacing && hdist[2] > sampleSpacing)
			{

				Triangle t[3];

				t[0].pos[0] = p[0];
				t[0].pos[1] = p[1];
				t[0].pos[2] = p[5];

				t[1].pos[0] = p[1];
				t[1].pos[1] = p[2];
				t[1].pos[2] = p[5];

				t[2].pos[0] = p[2];
				t[2].pos[1] = p[4];
				t[2].pos[2] = p[5];

				tris.push(t[0]);
				tris.push(t[1]);
				tris.push(t[2]);

			}else if (hdist[1] > sampleSpacing && hdist[2] > sampleSpacing)
			{
				Triangle t[3];

				t[0].pos[0] = p[0];
				t[0].pos[1] = p[2];
				t[0].pos[2] = p[3];

				t[1].pos[0] = p[0];
				t[1].pos[1] = p[3];
				t[1].pos[2] = p[5];

				t[2].pos[0] = p[5];
				t[2].pos[1] = p[3];
				t[2].pos[2] = p[4];

				tris.push(t[0]);
				tris.push(t[1]);
				tris.push(t[2]);

			// if only one of the sides can be splitted
			}else if (hdist[0] > sampleSpacing)
			{
				Triangle t[2];

				t[0].pos[0] = p[0];
				t[0].pos[1] = p[1];
				t[0].pos[2] = p[4];

				t[1].pos[0] = p[1];
				t[1].pos[1] = p[2];
				t[1].pos[2] = p[4];

				tris.push(t[0]);
				tris.push(t[1]);

			}else if (hdist[1] > sampleSpacing)
			{
				Triangle t[2];

				t[0].pos[0] = p[0];
				t[0].pos[1] = p[2];
				t[0].pos[2] = p[3];

				t[1].pos[0] = p[0];
				t[1].pos[1] = p[3];
				t[1].pos[2] = p[4];

				tris.push(t[0]);
				tris.push(t[1]);

			}else if (hdist[2] > sampleSpacing)
			{
				Triangle t[2];

				t[0].pos[0] = p[0];
				t[0].pos[1] = p[2];
				t[0].pos[2] = p[5];

				t[1].pos[0] = p[2];
				t[1].pos[1] = p[4];
				t[1].pos[2] = p[5];

				tris.push(t[0]);
				tris.push(t[1]);

			// if non of the sides can be splitted, this means this is our final triangle which would not subdivided further
			}else
			{
				newTriangles.push_back(tri);
			}
		}


		// now we iterate over all new triangles and close its sub-mesh, i.e. combine all equal vertices
		trimesh[i] = new UnstructuredInCoreTriangleMesh;
		{
			trimesh[i]->clearAndSetup(getDescr(), newTriangles.size() * 3, getTriangleDescr(), newTriangles.size());

			Vector3f p0 = maintri.pos[0];
			Vector3f p1 = maintri.pos[1];
			Vector3f p2 = maintri.pos[2];

			float triangleArea = norm((p1 - p0).crossProduct(p2 - p0)) * 0.5f;
			AAT normalAAT = trimesh[i]->getAAT("normal");
			AAT indexAAT  = trimesh[i]->getTriangles()->getAAT("index");
			AAT posAAT    = trimesh[i]->getAAT("position");
			AAT colAAT    = trimesh[i]->getAAT("color");

			for (unsigned j=0; j < newTriangles.size(); j++)
			{
				// compute barycentric coordinates for each of the vertices of the new triangle
				// this gives us a way to interpolate properties of a triangle 
				for (unsigned k=0; k < 3; k++)
				{
					float lambda0 = norm((p1-newTriangles[j].pos[k]).crossProduct(p2-newTriangles[j].pos[k])) * 0.5f / triangleArea;
					float lambda1 = norm((p2-newTriangles[j].pos[k]).crossProduct(p0-newTriangles[j].pos[k])) * 0.5f / triangleArea;
					float lambda2 = norm((p0-newTriangles[j].pos[k]).crossProduct(p1-newTriangles[j].pos[k])) * 0.5f / triangleArea;

					Vector3f interpNormal = normalize( normals[0]*lambda0 + normals[1]*lambda1 + normals[2]*lambda2 );
					trimesh[i]->getPointSet()->set3f(j * 3 + k, normalAAT, interpNormal);
					trimesh[i]->getPointSet()->set3f(j*3+k, posAAT, newTriangles[j].pos[k]);

					if (colAAT != NULL_AAT)
					{
						Vector3f interpColor = ( colors[0]*lambda0 + colors[1]*lambda1 + colors[2]*lambda2 );
						trimesh[i]->getPointSet()->set3f(j * 3 + k, colAAT, interpColor);
					}
				}
				trimesh[i]->getTriangles()->set3i(j, indexAAT, makeVector3i(j*3+0, j*3+1, j*3+2));
			}
			trimesh[i]->closeMesh();
		} // end building small triangle mesh containing the subdivided triangle

		newMesh->joinOtherMesh(trimesh[i]);
		delete trimesh[i];
	} // end subdividing each triangle

	newMesh->closeMesh();

	// now we need to combine the subdivided triangles into a larger one
	getPointSet()->assign(newMesh->getPointSet());
	getTriangles()->assign(newMesh->getTriangles());
	delete newMesh;

	//debugRenderer->endRenderJob();
}

void UnstructuredInCoreTriangleMesh::upsample(float sampleSpacing/*, bool useDebugRenderer*/)
{
	UnstructuredInCoreTriangleMesh * newMesh = new UnstructuredInCoreTriangleMesh;
	newMesh->clearAndSetup(getDescr(),0,getTriangleDescr(),0);
	newMesh->setMaterialIndex(getMaterialIndex());
	
	PointSet * ps = getPointSet();
	PointSet * ps_tri = getTriangles();
	PointSet * ps_new = newMesh->getPointSet();
	PointSet * ps_newtri = newMesh->getTriangles();

	AAT posAAT = getAAT("position");
	AAT normalAAT = getAAT("normal");
	AAT colorAAT = getAAT("color");
	AAT indexAAT = ps_tri->getAAT("index");

	std::vector<char> vbuffer(getDescr()->getSize());
	char * vptr = &vbuffer[0];

	for( unsigned i=0;i<ps_tri->getNumEntries();i++ )
	{
		Vector3i triangle = ps_tri->get3i(i,indexAAT);

		// reorder triangle indices so that the longest edge is first
		unsigned shift = 0;
		float bestLength = norm(ps->get3f(triangle[0],posAAT) - ps->get3f(triangle[1],posAAT));
		for( unsigned j=1;j<3;j++ )
		{
			float dist = norm(ps->get3f(triangle[(j+1) % 3],posAAT) - ps->get3f(triangle[j],posAAT));
			if( bestLength < dist)
			{
				bestLength = dist;
				shift = j;
			}
		}
		
		Vector3i tmpTri = triangle;
		triangle[0] = tmpTri[shift];
		triangle[1] = tmpTri[(shift+1)%3];
		triangle[2] = tmpTri[(shift+2)%3];
		

		std::vector<Vector3f> vertices;
		// create border vertices
		for( unsigned j=0;j<3;j++ )
		{
			Vector3f p0 = ps->get3f(triangle[j],posAAT);
			Vector3f p1 = ps->get3f(triangle[(j+1)%3],posAAT);
			vertices.push_back(p0);
			Vector3f dir = (p1 - p0);
			float l = norm(dir);
			dir = normalize(dir);
			unsigned numAddSamples = l / sampleSpacing;
			if( numAddSamples > 0 )
			{
				float lineSampleSpacing = l / ((float)numAddSamples + 1);
				for( unsigned k=1;k<numAddSamples+1;k++ )
				{
					Vector3f newVec = p0 + dir*(lineSampleSpacing*k);
					vertices.push_back(newVec);
				}
			}
		}
		card32 numBorderVertices = (card32)vertices.size();
		// create hexagonal triangulation inside the triangle
		Vector3f p0 = ps->get3f(triangle[0],posAAT);
		Vector3f p1 = ps->get3f(triangle[1],posAAT);
		Vector3f p2 = ps->get3f(triangle[2],posAAT);
		Vector3f dir1 = (p1-p0);
		if( norm(dir1) < 1e-15f )
			continue;
		dir1 = normalize(dir1);
		Vector3f triangleNormal = dir1.crossProduct(p2-p0);
		Vector3f dir2 = (dir1.crossProduct(triangleNormal));
		if( norm(dir2) < 1e-15f )
			continue;
		dir2 = normalize(dir2);
		if( dir2 * (p2-p0)< 0 )
			dir2 *= -1;
		
	
		float l1 = max(norm(p1-p0),dir1 * ((p2-p0)));
		float l2 = fabs(dir2 * (p2-p0));

		unsigned numSamplesX = l1 / sampleSpacing + 4;
		float rowHeight = sampleSpacing * sqrt(0.75f);
		unsigned numSamplesY = l2 / rowHeight + 4;

		std::vector<Vector3f> hexaVertices;
		std::vector<bool> validHexaVertices;
		Plane3f edge0(normalize(triangleNormal.crossProduct(p1-p0)),p0);
		Plane3f edge1(normalize(triangleNormal.crossProduct(p2-p1)),p1);
		Plane3f edge2(normalize(triangleNormal.crossProduct(p0-p2)),p2);
		Vector3f basePoint = p0 - dir1 * sampleSpacing * 1.5f - dir2 * rowHeight * 1.5f;
		for( unsigned y=0;y<numSamplesY;y++ )
		{
			float offset = (y%2)==0?0:sampleSpacing/2.0f;
			Vector3f rowVec = dir2 * (rowHeight * y);
			for( unsigned x=0;x<numSamplesX;x++ )
			{
				Vector3f newVec = basePoint + dir1 * (sampleSpacing * x + offset)
					 + rowVec;
				hexaVertices.push_back(newVec);
				

				validHexaVertices.push_back( edge0.calculateSignedDistance(newVec) > -sampleSpacing*1.5f  &&
					edge1.calculateSignedDistance(newVec) > -sampleSpacing*1.5f  && 
					edge2.calculateSignedDistance(newVec) > -sampleSpacing*1.5f  );

				// minize critical points near triangle border
				if( fabs(edge0.calculateSignedDistance(newVec)) < sampleSpacing * 0.1f ) 
				{
					newVec += edge0.getNormal() * sampleSpacing * 0.1f;
				}
				if( fabs(edge1.calculateSignedDistance(newVec)) < sampleSpacing * 0.1f ) 
				{
					newVec += edge1.getNormal() * sampleSpacing * 0.1f;
				}
				if( fabs(edge2.calculateSignedDistance(newVec)) < sampleSpacing * 0.1f ) 
				{
					newVec += edge2.getNormal() * sampleSpacing * 0.1f;
				}


			}
		}

		// create triangluation
		//if (useDebugRenderer)
		//	debugRenderer->beginRenderJob_OneFrame("asdf",7);
		std::vector<unsigned> valence(hexaVertices.size(),0);
		std::vector< Vector3i > triangles;
		for( unsigned y=0;y<numSamplesY-1;y++ )
		{
			unsigned rowIndex0 = y * numSamplesX;
			unsigned rowIndex1 = (y+1) * numSamplesX;
			for( unsigned x=0;x<numSamplesX-1;x++ )
			{
				unsigned i0 = rowIndex0 + x;
				unsigned i1 = rowIndex0 + x + 1;
				unsigned i2 = rowIndex1 + x + 1;
				unsigned i3 = rowIndex1 + x;

				if( y % 2 == 1)
				{
					i0 = rowIndex1 + x;
					i1 = rowIndex0 + x;
					i2 = rowIndex0 + x + 1;
					i3 = rowIndex1 + x + 1;
				}

				Vector3i tri0 = makeVector3i(i0,i1,i3);
				Vector3i tri1 = makeVector3i(i1,i2,i3);
				//if( validHexaVertices[tri0[0]] || validHexaVertices[tri0[1]] || validHexaVertices[tri0[2]])
				{
					triangles.push_back(tri0);
					valence[tri0[0]] ++;
					valence[tri0[1]] ++;
					valence[tri0[2]] ++;
					/*if( useDebugRenderer )
					{
						debugRenderer->addLine(hexaVertices[tri0[0]],hexaVertices[tri0[1]]);
						debugRenderer->addLine(hexaVertices[tri0[1]],hexaVertices[tri0[2]]);
						debugRenderer->addLine(hexaVertices[tri0[2]],hexaVertices[tri0[0]]);
					}*/
				}
				//if( validHexaVertices[tri1[0]] || validHexaVertices[tri1[1]] || validHexaVertices[tri1[2]])
				{
					triangles.push_back(tri1);
					valence[tri1[0]] ++;
					valence[tri1[1]] ++;
					valence[tri1[2]] ++;
					/*if( useDebugRenderer )
					{
						debugRenderer->addLine(hexaVertices[tri1[0]],hexaVertices[tri1[1]]);
						debugRenderer->addLine(hexaVertices[tri1[1]],hexaVertices[tri1[2]]);
						debugRenderer->addLine(hexaVertices[tri1[2]],hexaVertices[tri1[0]]);
					}*/
				}

			}
		}

		// cut to triangle shape
		unsigned numInnerVertices = (unsigned)hexaVertices.size();
		for( unsigned j=0;j<3;j++ )
		{
			Vector3f p0 = ps->get3f(triangle[j],posAAT);
			Vector3f p1 = ps->get3f(triangle[(j+1)%3],posAAT);
			Plane3f cutPlane(normalize(triangleNormal.crossProduct(p1-p0)),p0);
			std::map< pair<unsigned,unsigned>, unsigned > newVertexMap;
			std::vector<Vector3i> newTriangles;

			for( int t=0;t<(int)triangles.size();t++ )
			{

				//if( t >= triangles.size() )
				//	throw X4Exception("asdf!!");
				Vector3i currentTri = triangles[t];
				bool cutRequired = cutPlane.calculateSignedDistance(hexaVertices[currentTri[0]]) <= 0
					|| cutPlane.calculateSignedDistance(hexaVertices[currentTri[1]]) <= 0
					|| cutPlane.calculateSignedDistance(hexaVertices[currentTri[2]]) <= 0;

				if( !cutRequired )
				{
					newTriangles.push_back(currentTri);
					continue;
				}

//				triangles.erase(triangles.begin() + t);
				

				if( cutRequired )
					if( cutPlane.calculateSignedDistance(hexaVertices[currentTri[0]]) <= 0
						&& cutPlane.calculateSignedDistance(hexaVertices[currentTri[1]]) <= 0
						&& cutPlane.calculateSignedDistance(hexaVertices[currentTri[2]]) <= 0)
					{
						// reject triangle
						continue;
					}

				std::vector<unsigned> polygon;
				
				for( unsigned k=0;k<3;k++ )
				{
					Vector3f cp0 = hexaVertices[currentTri[k]];										
					Vector3f cp1 = hexaVertices[currentTri[(k+1)%3]];
					Line3f l(cp1-cp0,cp0);
					EPlaneSide side = cutPlane.classify(l,0);
					if ( side == /*EPlaneSide::*/PS_SPANNING)
					{
						// add new vertex
						pair<unsigned,unsigned> searchEntry( currentTri[k], currentTri[(k+1)%3]);
						if( searchEntry.first > searchEntry.second )
							swap(searchEntry.first, searchEntry.second);
						if( newVertexMap.find(searchEntry) != newVertexMap.end())
						{
							polygon.push_back(newVertexMap[searchEntry]);
						}
						else
						{
							Vector3f newPos = cutPlane.calculateIntersection(l);
							unsigned newIndex = (unsigned)hexaVertices.size();
							hexaVertices.push_back(newPos);
							polygon.push_back(newIndex);
							newVertexMap[searchEntry] = newIndex;
							//debugRenderer->addFastSphere(newPos,sampleSpacing * 0.25f, makeVector3f(1,1,0)*.6,false);
						}

					}
					if( cutPlane.calculateSignedDistance(cp1) > 0)
					{
						polygon.push_back(currentTri[(k+1)%3]);
					}

				}

				for( unsigned k=2;k<polygon.size();k++ )
				{
					newTriangles.push_back( makeVector3i(polygon[0], polygon[k-1],polygon[k]));
				}

			}

			triangles = newTriangles;
		}

		vertices = hexaVertices;

		// optimize inner points
		for( unsigned step=0;step<3;step++ )
		{
			//debugRenderer->beginRenderJob_OneFrame("testTr",step + 2);
			
			std::vector<Vector3f> vcopy = vertices;
			vertices = std::vector<Vector3f>(vertices.size(),NULL_VECTOR3F);
			std::vector<float> weight(vertices.size(),0);
			for( unsigned t=0;t<triangles.size();t++ )
			{
				unsigned i0 = triangles[t][0];
				unsigned i1 = triangles[t][1];
				unsigned i2 = triangles[t][2];
				Vector3f p0 = vcopy[i0];
				Vector3f p1 = vcopy[i1];
				Vector3f p2 = vcopy[i2];
				/*debugRenderer->addLine(p0,p1);
				debugRenderer->addLine(p1,p2);
				debugRenderer->addLine(p2,p0);*/
				float area = norm((p1-p0).crossProduct(p2-p0)) * .5f;
				Vector3f middle = (p1+p2+p0) / 3;
				
				vertices[i0] += middle * area;
				vertices[i1] += middle * area;
				vertices[i2] += middle * area;
				weight[i0] += area;
				weight[i1] += area;
				weight[i2] += area;
			}

			for( unsigned j=0;j<vertices.size();j++ )
			{
				if( j >= numInnerVertices || weight[j] < 1e-10f)
					vertices[j] = vcopy[j];
				else
				{
					vertices[j] /= weight[j];
				}

			}

			//debugRenderer->endRenderJob();
		}

		// add new triangles/vertices to new mesh
		Vector3f normals[3];
		for( unsigned j=0;j<3;j++ )
			normals[j] = ps->get3f(triangle[j],normalAAT);

		unsigned numVerticesInNewMesh = newMesh->getNumPoints();
		std::vector<char> buffer( max(newMesh->getDescr()->getSize(),newMesh->getTriangleDescr()->getSize()),0 );
		Matrix3f paramToWorld;
		paramToWorld[0] = p1-p0;
		paramToWorld[1] = p2-p0;
		paramToWorld[2] = triangleNormal;
		float triangleArea = norm((p1-p0).crossProduct(p2-p0)) * 0.5f;

		int invalidVertex = -1;
		std::vector<int> newVertexIndexTable(vertices.size(),invalidVertex);
		unsigned currentNewVertexIndex = 0;
		for( unsigned j=0;j<triangles.size();j++ )
		{
			Vector3i tri = triangles[j];
			for(  unsigned k=0;k<3;k++ )
			{
				if( newVertexIndexTable[tri[k]] == invalidVertex )
				{
					unsigned index = newMesh->getPointSet()->getNumEntries();
					unsigned oldIndex = tri[k];
					newVertexIndexTable[tri[k]] = index;
					tri[k] = index;

					newMesh->getPointSet()->addPoint(&buffer[0]);
					newMesh->getPointSet()->set3f(index, posAAT, vertices[oldIndex]);

					// interpolate normal via barycentric coordinates
					float lambda0 = norm((p1-vertices[oldIndex]).crossProduct(p2-vertices[oldIndex])) * 0.5f / triangleArea;
					float lambda1 = norm((p2-vertices[oldIndex]).crossProduct(p0-vertices[oldIndex])) * 0.5f / triangleArea;
					float lambda2 = norm((p0-vertices[oldIndex]).crossProduct(p1-vertices[oldIndex])) * 0.5f / triangleArea;

					Vector3f interpNormal = normalize( normals[0]*lambda0 + 
						normals[1]*lambda1 +  normals[2]*lambda2 );
					newMesh->getPointSet()->set3f(index, normalAAT, interpNormal);
				}
				else
				{
					tri[k] = newVertexIndexTable[tri[k]];
				}
			}
			unsigned currentIndex = newMesh->getNumTriangles();
			newMesh->getTriangles()->addPoint(&buffer[0]);
			newMesh->getTriangles()->set3i(currentIndex, indexAAT, tri);
		}

		/*unsigned newVIndex[3];
		for( unsigned j=0;j<3;j++ )
		{
			unsigned i1 = triangle[j];
			unsigned i2 = triangle[(j+1)%3];
			unsigned newVertexID = min(i1,i2)*ps->getNumEntries()+ max(i1,i2);
			if( edgeMap.find(newVertexID) != edgeMap.end() )
			{
				newVIndex[j] = edgeMap[newVertexID];
			}
			else
			{
				newVIndex[j] = newMesh->getNumPoints();
				edgeMap[newVertexID] = newVIndex[j];
				const VertexDescriptor * vd = mesh->getDescr();
				char * ptr[2];
				ptr[0] = ps->getDataPointer(i1);
				ptr[1] = ps->getDataPointer(i2);
				float32 interp[] = {0.5f,0.5f};
				vd->interpolate( (const char**)ptr, interp,2,vptr);
				ps_new->addPoint(vptr);
				ps_new->set3f(newVIndex[j],normalAAT, normalize(ps_new->get3f(newVIndex[j],normalAAT)));
			}
		}
		Vector3i newTri1 = makeVector3i(triangle[0], newVIndex[0],newVIndex[2]);
		Vector3i newTri2 = makeVector3i(triangle[1], newVIndex[1],newVIndex[0]);
		Vector3i newTri3 = makeVector3i(triangle[2], newVIndex[2],newVIndex[1]);
		Vector3i newTri4 = makeVector3i(newVIndex[0], newVIndex[1],newVIndex[2]);
		ps_newtri->set3i(i,indexAAT,newTri1);
		ps_newtri->addPoint((char*)newTri2.data());
		ps_newtri->addPoint((char*)newTri3.data());
		ps_newtri->addPoint((char*)newTri4.data());*/
	}
	//if (useDebugRenderer)
	//	debugRenderer->endRenderJob();
	//newMesh->setTransformation(mesh->getTransformation());
	//addPointCloud(getScene(), newMesh, "upsampledMesh");
	getPointSet()->assign(newMesh->getPointSet());
	getTriangles()->assign(newMesh->getTriangles());
	delete newMesh;
}

// --


ConnectivityPointSetIterator::ConnectivityPointSetIterator(PointSet *meshPS, AttachedDataContainer *attData, BoundingBox3f bb)
: AttachedDataPointSetIterator(meshPS, attData) {
   this->bb = bb;
}

BoundingBox3f ConnectivityPointSetIterator::getBoundingBox() {
   return bb;
}


