//---------------------------------------------------------------------------
#ifndef UnstructuredInCoreTriangleMeshH
#define UnstructuredInCoreTriangleMeshH
//---------------------------------------------------------------------------
#include "x3geometry.h"
//---------------------------------------------------------------------------
#include "UnstructuredInCorePointCloud.h"
#include "TriangleMeshPointCloud.h"
#include "HashTables.h"
//---------------------------------------------------------------------------

namespace X4 {

    class InCorePCTopologyGraph;

    class X3GEOMETRY_API UnstructuredInCoreTriangleMesh :
        public UnstructuredInCorePointCloud,
        public TriangleMeshPointCloud {
            X4_CLASS(UnstructuredInCoreTriangleMesh);
    private:
        PointSet *triangles;

    public:
        UnstructuredInCoreTriangleMesh();

        virtual void assign(const Object* obj, X4CopyContext *context = NULL);


        // --- UnstructuredInCorePointCloud

        /// creates empty mesh with standard triangle face descriptor (index only)
        virtual void clearAndSetup(const VertexDescriptor *vertDescr, mpcard numPoints = 0);
        virtual void clearAndSetup(const VertexDescriptor *vertDescr, mpcard numPoints,
            const VertexDescriptor *triangleDescr, mpcard numTriangles);

        void clearAndSetupDefault(mpcard numPoints, mpcard numTriangles, bool hasNormal=true, bool hasColor = true, bool hasFlags = true);

        virtual PointCloudIterator *createIterator(SceneObjectIteratorType type);
        virtual bool supportsIteratorType(SceneObjectIteratorType cap);
        virtual void clear();


        /// --- additional methods, similar to UnstructuredInCorePointCloud point processing methods

        virtual const PointSet *getTriangles() const;
        virtual void setTriangles(PointSet *psTriangles);
        virtual PointSet *getTriangles();
        /// ownership of ps will be transfered to this
        virtual UnstructuredInCoreTriangleMesh *createCopyWithNewTriangles(PointSet *psTriangles) const;
        UnstructuredInCoreTriangleMesh *createCopyOfGivenTriangles(const std::vector<int32> &indexes);


        // ---  TriangleMeshPointCloud
        virtual const VertexDescriptor *getTriangleDescr() const;
        virtual PointCloud             *appendTriangleFields(const VertexDescriptor *newAttribs) const;
        virtual void                    clearAndSetupTriangles(const VertexDescriptor *triangleDescr);
        virtual card64                  getNumTriangles() const;





        /// transfer triangle topology to topology graph attachment
        void computeTopologyAttachment(bool computeCotangentWeights = false);
        void computeBoundaryTopologyAttachment();


        ~UnstructuredInCoreTriangleMesh();

        // --- standard mesh methods ---

        /// searches for double vertices with exactly equal positions and merges them
        /// option: merge only vertices with exactly equal positions _and_ normals
        void closeMesh(bool checkNormals=false);

        /// opens the mesh by duplicating vertices which share several faces. the resulting triangle mesh will consists of 3 points per face
        /// the functions is a reverse of closeMesh
        void openMesh();

        /// calculate vertex normals
        void calculateVertexNormals();

        ///remove all triangles, connected to selected points
        void deleteMarkedTriangles();
        void deleteMarkedTriangles(const std::vector<mpcard>& todelete);
        void deletePoints(const std::vector<mpcard>& todelete);

        /// removes any unconnected vertices from mesh
        void clearGarbage();

        /// remove double/invalid triangles from mesh
        void cleanTopology();

        /// snap vertices to closest neighbor 
        void snapMesh( float max_tolerance, bool snapSelectedPointsOnly=false );

        /// extract border from mesh (considers mesh topology only)
        /// returns edge ids from border (edgeID = triangleIndex*3+edgeIndex)
        void extractMeshBorder(std::vector<mpcard> &borderEdges);


        /// Gives copy of the submesh indicated by the given vertices. If cut = true, the submesh will be cut from the triangle mesh
        enum ExtractFunction
        {
            COPY,  // return just a copy 
            CUT,   // return a copy and remove all triangles connected to the submesh
            CUT_ALONG_BORDER // return copy and extract submesh in the way, that all border vertices survives (some kind of a gentle cut)
        };
        UnstructuredInCoreTriangleMesh* extractSubMesh(const std::vector<mpcard>& vertices, ExtractFunction func = COPY);

        /// split mesh into a set of connected components. If no tpg is given, then it will be computed from the triangulation
        std::vector<UnstructuredInCoreTriangleMesh*> split(InCorePCTopologyGraph* tpg = NULL);

        InCorePCTopologyGraph* computeTopology(bool  computeCotangentWeights);

        /// joins two meshes, assumes same vertex descriptors (also for triangles)
        /// other mesh will be transformed
        void joinOtherMesh(UnstructuredInCoreTriangleMesh* other, const Matrix4f &T=IDENTITY4F);

        /// flip index orientation from 0-1-2 to 2-1-0
        void flipOrientation();

        /// upsample triangle mesh by producing uniformly sized sub-triangles
        void upsample(float sampleSpacing);

        /// upsample triangle mesh by producing uniformly sized sub-triangles
        void upsample2(float sampleSpacing);


		  // --- geometric primitives ---


    private:
        class X3GEOMETRY_API StringComp {
        public:
            static bool equals(std::string obj1, std::string obj2) { return obj1 == obj2; }
        };

        class MyHashTable : public HashTable<std::string, card32, StringComp> {
        protected:
            virtual max_platform_card hashFunction(std::string key) const {
                card32 hk = 0;
                const card32 n = (const card32 )key.size();
                for (card32 i=0; i<n; i++) {
                    const card8 c = key[i];
                    hk <<= 1;
                    hk ^=c;
                }
                return (max_platform_card)hk;
            };
        public:
            MyHashTable(card32 fillFactor = 1, card32 startSizeIndex = 8) : HashTable<std::string, card32, StringComp>(fillFactor, startSizeIndex) {}

        };

        std::string calcHashString( const Vector3f &vec );

    };



    class ConnectivityPointSetIterator : public AttachedDataPointSetIterator {
    private:
        BoundingBox3f bb;

    public:
        ConnectivityPointSetIterator(PointSet *meshPS, AttachedDataContainer *attData, BoundingBox3f bb);
        /// must be redefined because triangles themselfs do not contain positions (but indices to positions)
        /// --> the bb must be provided externally
        virtual BoundingBox3f getBoundingBox();

    };


};


#endif
