//---------------------------------------------------------------------------
#ifndef VertexArrayH
#define VertexArrayH
//---------------------------------------------------------------------------
#include "x3geometry.h"
//---------------------------------------------------------------------------
#include "VertexDescriptor.h"
#include "LinearAlgebra.h"
#include "DynamicLinearAlgebra.h"
//---------------------------------------------------------------------------

namespace X4
{

class PointSet;
class PointCloud;
class SceneObjectIterator;

/**
	provides easier access common vertex-formats
	special methods:
		get / setPosition3f
		get / setNormal3f
		get / setColor3f
		get / setFlags
      get / setIndex
	fell free to add other common attributes :-)
	you may also use the templated get/set methods
	Example:	
		for( card32 i=0;i<... )
		{
			Vector3f pos = vertexarray.get<float,3>(i, "position" );	// slower !
			set( i, COLOR, pos );
		}

*/
class X3GEOMETRY_API VertexArray
{
private:

   void initVD( const VertexDescriptor * desc );
   bool initAAT( AAT &aat, std::string name, card32 format, card32 numComponents );

	VertexDescriptor	*	m_VD;
	card8				*	m_Buffer;
	card64					m_NumElements;

	AAT						POSITION, COLOR, NORMAL, FLAGS, INDEX, TYPEID, TYPEFLAG;
	bool					m_HasPosition3f,
							m_HasColor3f,
							m_HasNormal3f,
							m_HasFlags,
                     m_HasIndex3i,
                     m_HasTypeID,
                     m_HasTypeFlag;
	SceneObjectIterator * m_Iter;
	PointSet*			m_PointSet;	// just used to invalidate intern flag, no other purpose

public:
	VertexArray( card8 * buffer, card64 numElements, const VertexDescriptor * desc);
	VertexArray( PointSet * ps );
	VertexArray( PointCloud * pc);
	~VertexArray(void);

//	void setNumElements(card32 numElements ) {m_NumElements = numElements;}
	card64 getNumElements() {return m_NumElements;}

//	void setBuffer(card8 *buffer){ m_Buffer = buffer; }
	card8 * getBuffer(){ return m_Buffer; }

   card32 getBufferSize() const;

   VertexDescriptor * getVertexDescriptor() const {return m_VD;}

// --------------------------------------------------------
//                complete vertex access methods

   /// vertex[destIndex] = vertex[sourceIndex]
   void copyVertex( card32 sourceIndex, card32 destIndex );

   /// copies all vertices from source to this array
   /// vertexdescriptor and num elements must be equal !
   void copyVertices( const VertexArray &source);

   /// copy vertex from sourceArray[sourceIndex] to this[destIndex]
   /// vertexdescriptor must be identical
   void copyVertex( const VertexArray &sourceArray, card32 sourceIndex,
      card32 destIndex );

// --------------------------------------------------------
//                attribute access methods
	/// vertices[index].attr = vec
	template<class DataType, int dim> 
	inline void set( card32 index, const std::string &attr, const StaticVector<DataType,dim> &vec )
	{
		AAT aat = m_VD->getToken( attr );
		set( index, aat, vec );
	}

	template<class DataType, int dim>
	inline void set( card32 index, AAT token, const StaticVector<DataType,dim> &vec )
	{
		x4Assert( index < m_NumElements );
        StaticVector<DataType,dim> * ptr = (StaticVector<DataType,dim>*)&m_Buffer[index*m_VD->getSize() + token.getOffset()];
		*ptr = vec;
	}

	template<class DataType, int dim>
	inline StaticVector<DataType,dim> get( card32 index, const std::string &attr ) const
	{
		AAT aat = m_VD->getToken( attr );
		return get( index, aat );
	}

	template<class DataType, int dim>
	inline StaticVector<DataType,dim> get( card32 index, AAT token ) const
	{
		x4Assert( index < m_NumElements );
        StaticVector<DataType,dim> * ptr = (StaticVector<DataType,dim>*)&m_Buffer[index*m_VD->getSize() + token.getOffset()];
		return *ptr;
	}

	inline Matrix3f getMatrix3f(mpcard index, AAT token)
	{
		StaticVector<float,9> tmp = get<float,9>(index,token);
		return *((Matrix3f*)&tmp[0]);
	}

	inline void setMatrix3f(mpcard index, AAT token, const Matrix3f &m)
	{
		set<float,9>(index,token, *((StaticVector<float,9>*)&m[0][0]));
	}

	inline Vector3f getPosition3f( card32 index ) const
	{
		x4Assert( m_HasPosition3f );
		return get<float,3>( index, POSITION );
	}

	inline void setPosition3f( card32 index, const Vector3f &vec )
	{
		x4Assert( m_HasPosition3f );
		set<float32, 3>( index, POSITION, vec );
		if( m_PointSet)
			m_PointSet->invalidateBoundingBox();
	}

	inline Vector3f getColor3f( card32 index ) const
	{
		x4Assert( m_HasColor3f );
		return get<float,3>( index, COLOR );
	}

	inline void setColor3f( card32 index, const Vector3f &vec )
	{
		x4Assert( m_HasColor3f );
		set<float32, 3>( index, COLOR, vec );
	}

	inline Vector3f getNormal3f( card32 index ) const
	{
		x4Assert( m_HasNormal3f );
		return get<float,3>( index, NORMAL );
	}

	inline void setNormal3f( card32 index, const Vector3f &vec )
	{
		x4Assert( m_HasNormal3f );
		set<float32, 3>( index, NORMAL, vec );
	}

	inline int32 getFlags( card32 index ) const
	{
		x4Assert( m_HasFlags );
		return get<int32,1>( index, FLAGS )[0];
	}

	inline void setFlags( card32 index, const int32 &f )
	{
		x4Assert( m_HasFlags );
      StaticVector<int32,1> tmp;
      tmp[0] = f;
		set<int32, 1>( index, FLAGS, tmp );
	}

	inline Vector3i getIndex3i( card32 index ) const
	{
		x4Assert( m_HasIndex3i );
		return get<int32,3>( index, INDEX );
	}

	inline void setIndex3i( card32 index, const Vector3i &vec )
	{
		x4Assert( m_HasIndex3i );
		set<int32, 3>( index, INDEX, vec );
	}

   inline int32 getTypeID( card32 index) const
   {
      x4Assert( m_HasTypeID );
      return get<int32,1>( index, TYPEID )[0];
   }

   inline void setTypeID( card32 index, int32 typeID )
   {
      x4Assert( m_HasTypeID );
      StaticVector<int32,1> vec;
      vec[0] = typeID;
      set<int32,1>( index, TYPEID, vec );
   }

   inline int32 getTypeFlag( card32 index) const
   {
      x4Assert( m_HasTypeFlag );
      return get<int32,1>( index, TYPEFLAG )[0];
   }

   inline void setTypeFlag( card32 index, int32 typeFlag )
   {
      x4Assert( m_HasTypeFlag );
      StaticVector<int32,1> vec;
      vec[0] = typeFlag;
      set<int32,1>( index, TYPEFLAG, vec );
   }

   char * getAttributePointer( card32 index, std::string attr )
   {
	   return (char*)m_Buffer + m_VD->getAttributeOffset( m_VD->getIndex(attr) );
   }

   template< typename Iterator > 
   void getPositions( Iterator out ) {
	   for( int i1 = 0; i1 < getNumElements(); ++ i1 ) {
		   Vector3f p = getPosition3f(i1); 
		   *(out ++) = p[0];
		   *(out ++) = p[1];
		   *(out ++) = p[2];
	   }
   }

   template< typename Iterator > 
   void setPositions( Iterator in ) {
	   for( int i1 = 0; i1 < getNumElements(); ++ i1 ) {
		   Vector3f p = makeVector3f( *(in++), *(in++), *(in++) );
		   setPosition3f(i1, p);
	   }
   }


   template< typename Iterator > 
   void getPositions3f( Iterator out ) {
	   for( int i1 = 0; i1 < getNumElements(); ++ i1 ) {
		   *(out ++) = getPosition3f(i1);
	   }
   }

   template< typename Iterator > 
   void setPositions3f( Iterator in ) {
	   for( int i1 = 0; i1 < getNumElements(); ++ i1 ) {
		   setPosition3f(i1, *(in++));
	   }
   }
};


}

#endif