//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "VertexDescriptor.h"
//---------------------------------------------------------------------------
#include "Tools.h"
#include "X4Exception.h"
#include "NumericalClassProperty.h"
#include "StringClassProperty.h"
#include "PointSet.h"
#include "PointSet.inline.h"
#include "LinearAlgebra.inline.h"
#include "StringHelper.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

const bool DO_NOT_USE_EXPERIMENTAL_FAST_COPY = true;


const AttributeAccessToken X4::NULL_AAT = AAT::makeInvalidToken();


MetaClass* VertexDescriptor::VertexDescriptorMetaClass = NULL;

const MetaClass* VertexDescriptor::getInstanceClass() const {
   return getClass();
}

Object* VertexDescriptor::newInstance() {
   return new VertexDescriptor;
}

const MetaClass* VertexDescriptor::getClass() {
   if (VertexDescriptorMetaClass==NULL) {
      VertexDescriptorMetaClass = new MetaClass("VertexDescriptor", "no description.",
                                  newInstance, sizeof(VertexDescriptor));
   }
   return VertexDescriptorMetaClass;
}

void VertexDescriptor::init(const MetaClass *parentClass) {
   ((MetaClass*)getClass())->init(parentClass);
   VertexDescriptorMetaClass->copyParentProperties();
	// mh - 21.03.2006 - moved custom editor init to InitPointClouds
}

void VertexDescriptor::shutdown() {
   delete VertexDescriptorMetaClass;
}


// ----

AttributeAccessToken AttributeAccessToken::makeInvalidToken() {
   return AttributeAccessToken();
}


// --


VertexAttribDescriptor::VertexAttribDescriptor() {
   this->name = ""; this->numComponents = 0;
   this->dataFormat = 0;
}

VertexAttribDescriptor::VertexAttribDescriptor(std::string name, card32 numComponents, card32 dataFormat) {
   this->name = name;
   this->numComponents = numComponents;
   this->dataFormat = dataFormat;
   // TODO (daniel): Fails for DATA_FORMAT_CARD64. Make sure that 64 bit types are correctly supported, and then extend the assertion.
   x4Assert(dataFormat == DATA_FORMAT_CARD8 || dataFormat == DATA_FORMAT_INT32 || dataFormat == DATA_FORMAT_FLOAT32 || dataFormat == DATA_FORMAT_CARD32);
}

card32 VertexAttribDescriptor::getAttribElementSize() const {
   return dataFormat & 0xF;
}

card32 VertexAttribDescriptor::getSize() const {
   return getAttribElementSize()*numComponents;
}


std::string VertexAttribDescriptor::getDescription() const {
   char tmp[50];
   std::string c;
   if (dataFormat == DATA_FORMAT_FLOAT32) {
      c = "f";
   } else if (dataFormat == DATA_FORMAT_CARD8) {
      c = "ub";
   } else if (dataFormat == DATA_FORMAT_INT32) {
      c = "i";
   } else if (dataFormat == DATA_FORMAT_CARD32) {
      c = "u";
   } else {
      c = "?";
   }
   std::string suffix = std::string("(") + std::string(itoa(numComponents, tmp, 10)) + c + std::string(")");
   return name+suffix;
}

void VertexAttribDescriptor::write(OutputObjectStream *s) const {
   s->writeString(name);
   s->write(numComponents);
   s->write(dataFormat);
}

void VertexAttribDescriptor::read(InputObjectStream *s) {
   s->readString(name);
   s->read(numComponents);
   s->read(dataFormat);
}


// --


VertexDescriptor::VertexDescriptor() {
   offsets.push_back(0);
}

void VertexDescriptor::assign(const Object* obj, X4CopyContext *context) {
   const VertexDescriptor *other = dynamic_cast<const VertexDescriptor*>(obj);
   x4Assert(other != NULL);
   this->attributes = other->attributes;
   this->offsets = other->offsets;
   forget(context);
}

bool VertexDescriptor::operator==(const VertexDescriptor *vd) const {
   return equal(vd);
}

bool VertexDescriptor::operator!=(const VertexDescriptor *vd) const {
   return !equal(vd);
}


void VertexDescriptor::write(OutputObjectStream *s) const {
   const card32 n = (card32)attributes.size();
   s->write(n);
   for (card32 i=0; i<n; i++) {
      attributes[i].write(s);
   }
}

void VertexDescriptor::read(InputObjectStream *s) {
   card32 n;
   s->read(n);
   attributes.resize(n);
   offsets.resize(n+1);
   card32 offset = 0;
   for (card32 i=0; i<n; i++) {
      attributes[i].read(s);
      offsets[i] = offset;
      offset += attributes[i].getSize();
   }
   offsets[n] = offset;
   x4Assert(offsets.size() == attributes.size() + 1);
}

void VertexDescriptor::pushAttrib(VertexAttribDescriptor a) {
   attributes.push_back(a);
   card32 offset = a.getSize();
   offset += offsets[attributes.size()-1];
   offsets.push_back(offset);
   x4Assert(offsets.size() == attributes.size() + 1);
}

void VertexDescriptor::clear()
{
   attributes.clear();
   offsets.clear();
   offsets.push_back(0);
}

mpcard VertexDescriptor::getNumAttribs() const {
   return attributes.size();
}

mpcard VertexDescriptor::getNumAttributes() const {
   return attributes.size();
}

VertexAttribDescriptor VertexDescriptor::getAttribute(card32 i) const {
   return attributes[i];
}

card32 VertexDescriptor::getIndex(std::string name) const {
   const card32 n = (card32)attributes.size();
   for (card32 i=0; i<n; i++) {
      if (attributes[i].name == name) {
         return i;
      }
   }
   error("VertexDescriptor::getIndex() - name not found.");
   return n;
};

bool VertexDescriptor::providesAttribute(std::string name) const {
	// mb - added a way to access single components of a std::vector-attribute directly
	// example: "color:2" corresponds to the third color component
	std::string number = name;
	name = extractToChar(number,':');

   const size_t n = attributes.size();
   for (card32 i=0; i<n; i++) {
      if (attributes[i].name == name) {
         return true;
      }
   }
   return false;
}

bool VertexDescriptor::providesAttribute(std::string name, card32 numComponents) const {
   const size_t n = attributes.size();
   for (card32 i=0; i<n; i++) {
      if (attributes[i].name == name && attributes[i].numComponents == numComponents) {
         return true;
      }
   }
   return false;
}

bool VertexDescriptor::providesAttribute(std::string name, card32 numComponents, card32 dataFormat) const {
/*   const size_t n = attributes.size();
   for (card32 i=0; i<n; i++) {
      if (attributes[i].name == name && attributes[i].numComponents == numComponents && attributes[i].dataFormat == dataFormat) {
         return true;
      }
   }
   return false;*/

	// mb - added a way to access single components of a std::vector-attribute directly
	// example: "color:2" corresponds to the third color component
	std::string number = name;
	name = extractToChar(number,':');
	if( number.length() > 0)
	{
		const size_t n = attributes.size();
		for (card32 i=0; i<n; i++)
			if (attributes[i].name == name && attributes[i].dataFormat == dataFormat )
			{
				int num;
				if( strToInt(number, num))
				{
					unsigned numBytes = attributes[i].dataFormat & 15;
					return num + numComponents <= (int)attributes[i].numComponents;

				}
				return false;
			}
		return false;
	}
	else
	{
		const size_t n = attributes.size();
		for (card32 i=0; i<n; i++) {
			if (attributes[i].name == name && attributes[i].numComponents == numComponents && attributes[i].dataFormat == dataFormat) {
				return true;
			}
		}
		return false;
	}

}

bool VertexDescriptor::providesAttribute(VAD att) const {
   const size_t n = attributes.size();
   for (card32 i=0; i<n; i++) {
      if (attributes[i] == att) {
         return true;
      }
   }
   return false;
}



card32 VertexDescriptor::getSize() const {
   x4Assert(attributes.size() + 1 == offsets.size());
   return offsets[attributes.size()];
}


std::string VertexDescriptor::getDescription() const {
   const size_t n = attributes.size();
   if (n == 0) {
      return "(no attributes defined)";
   }
   std::string result;
   for (card32 i=0; i<n; i++) {
      result += attributes[i].getDescription();
      if (i != n-1) {
         result += std::string("; ");
      }
   }
   return result;
}


void VertexDescriptor::mergeAttributes(const VertexDescriptor *newAttribs) {
   const size_t nd = attributes.size();
   const size_t ns = newAttribs->attributes.size();
   for (card32 s=0; s<ns; s++) {
      const std::string currentName = newAttribs->attributes[s].name;
      bool attribFound = false;
      for (card32 d=0; d<nd; d++) {
         if (attributes[d].name == currentName) {
            if (!(   attributes[d].dataFormat    == newAttribs->attributes[s].dataFormat
                  && attributes[d].numComponents == newAttribs->attributes[s].numComponents) ) {
               error("VertexDescriptor::mergeAttributes - type mismatch in duplicate field.");
            } else {
               attribFound = true;
               break;
            }
         }
      }
      if (!attribFound) {
         pushAttrib(newAttribs->attributes[s]);
      }
   }
}


AttributeAccessToken VertexDescriptor::getToken(std::string name) const {
	if (!providesAttribute(name)) {
		error(std::string("VertexDescriptor::getToken() - Couldn't find attribute ") + name + ".");
        return AttributeAccessToken();
	}
	// mb - added a way to access single components of a std::vector-attribute directly
	// example: "color:2" corresponds to the third color component
	std::string number = name;
	name = extractToChar(number,':');
   const card32 i = getIndex(name);
	if( number.length() > 0)
	{
		int num;
		if( strToInt(number, num))
		{
			unsigned numBytes = attributes[i].dataFormat & 15;
			if( num < (int)attributes[i].numComponents )
				return AAT( attributes[i].getSize(), offsets[i] + numBytes*num );
		}
		error(std::string("VertexDescriptor::getToken() - invalid attribute access! Trying to access component #") + number + ".");
        return AttributeAccessToken();
	}
	else
		return getToken(i);
}

AttributeAccessToken VertexDescriptor::getToken(std::string name, card32 numComponents, card32 dataFormat) const {
	// get result and and the same time check for name
	AAT result = getToken(name);
	if (!providesAttribute(name, numComponents, dataFormat)) {
		error("VertexDescriptor::getToken() - data format does not match.");
	}
	return result;
}											


AttributeAccessToken VertexDescriptor::getToken(card32 i) const {
   x4Assert(i < offsets.size());
   x4Assert(offsets.size() == attributes.size() + 1);
   return AttributeAccessToken(attributes[i].getSize(), offsets[i]);
}

// --------- convertion routines -----------------------
bool VertexDescriptor::isCompatible( const VertexDescriptor &source) const
{
   for( card32 i=0;i<getNumAttributes();i++ )
   {
      VAD tmp = getAttribute(i);
      if( source.providesAttribute( tmp.name ) )
         return true;
   }
   return false;
}

bool VertexDescriptor::isFullCompatible( const VertexDescriptor &source) const 
{
   for( card32 i=0;i<getNumAttributes();i++ )
   {
      VAD tmp = getAttribute(i);
      if( !source.providesAttribute( tmp.name ) )
         return false;
   }
   return true;
}

card8 VertexDescriptor::convertInt32ToCard8( const card8 * ptr ) const
{
   int32 * tmp = (int32 *)ptr;
   card8 ret = (card8)*tmp;
   return ret;
}

card8 VertexDescriptor::convertFloat32ToCard8( const card8 * ptr ) const
{
   float32 * tmp = (float32 *)ptr;
   card8 ret = (card8)(*tmp * 255.0f);
   return ret;
}

int32 VertexDescriptor::convertCard8ToInt32( const card8 * ptr ) const
{
   int32 ret = (int32)(*ptr);
   return ret;
}

int32 VertexDescriptor::convertFloat32ToInt32( const card8 * ptr ) const
{
   float32 * tmp = (float32 *)ptr;
   int32 ret = (int32)(*tmp * 255.0f);
   return ret;
}

float32 VertexDescriptor::convertCard8ToFloat32( const card8 * ptr ) const
{
   card8 tmp = *ptr;
   float32 ret = tmp / 255.0f;
   return ret;
}

float32 VertexDescriptor::convertInt32ToFloat32( const card8 * ptr ) const
{
   int32 * tmp = (int32 *)ptr;
   float32 ret = *tmp / 255.0f;
   return ret;
}

void VertexDescriptor::convertVertices( const PointSet * ps, std::vector<card8>  &ret_vertices, const int32 includeFlags, const int32 excludeFlags ) const
{
   char * dataPtr = (char *)ps->getDataPtr();
   VertexDescriptor sourceVD = *ps->getDescr();
   convertVertices( dataPtr, ps->getNumEntries(), sourceVD, ret_vertices, includeFlags, excludeFlags );
}

void VertexDescriptor::convertVertices( const std::vector<card8> &vertices, const VertexDescriptor &sourceVD,
   std::vector<card8>  &ret_vertices, const int32 includeFlags, const int32 excludeFlags ) const
{
   size_t numVertices = vertices.size() / sourceVD.getSize();
   convertVertices( (char*)&vertices[0], numVertices, sourceVD, ret_vertices, includeFlags, excludeFlags );
}

void VertexDescriptor::convertVertices( const char * source_vertices, mpcard numVertices,const VertexDescriptor &sourceVD,
   std::vector<card8>  &ret_vertices, const int32 includeFlags, const int32 excludeFlags ) const
{
   card8 * vertices = (card8 *)source_vertices;
   card32 num_ret_vert = numVertices;
   if( sourceVD.providesAttribute( "flags" ) && (includeFlags || excludeFlags))
   {
      num_ret_vert = 0;
      card32 flags_offset = sourceVD.getAttributeOffset( sourceVD.getIndex( "flags" ) );
      for( card32 i=0;i<numVertices;i++ )
      {
         int32 flags = *((int32*)&source_vertices[flags_offset]);
         flags_offset += sourceVD.getSize();
         bool takePoint = ((flags & includeFlags) == includeFlags) && ((flags & excludeFlags) == 0);
         if( !takePoint )
            continue;
         num_ret_vert++;
      }
   }


   if( !isFullCompatible(sourceVD) )
		ret_vertices.resize(num_ret_vert*getSize(),0);
   else
	   ret_vertices.resize(num_ret_vert*getSize());

   bool has_flags = sourceVD.providesAttribute( "flags" );
   card32 flags_offset;

   for( card32 a=0;a<getNumAttributes();a++)
   {
      if( has_flags )
         flags_offset = sourceVD.getAttributeOffset( sourceVD.getIndex( "flags" ) );

      VAD dest_attrib = getAttribute(a);
      if( !sourceVD.providesAttribute( dest_attrib.name ) )
      {
         // ret_vertices already filled with zeros
      }
      else
      {
         card32 source_a = sourceVD.getIndex( dest_attrib.name );
         VAD source_attrib = sourceVD.getAttribute( source_a );
         card32 dest_offset = getAttributeOffset( a );
         card32 source_offset = sourceVD.getAttributeOffset( source_a );
         card32 numComp = std::min( dest_attrib.numComponents, source_attrib.numComponents );
         card32 numBytes = numComp;
         switch( dest_attrib.dataFormat )
         {
         case VAD::DATA_FORMAT_CARD8:
         break;
         case VAD::DATA_FORMAT_FLOAT32:
         case VAD::DATA_FORMAT_INT32:
            numBytes *= 4;
         break;
         }

         if( source_attrib.dataFormat == dest_attrib.dataFormat )
         {
            // direct copy
            card32 dest_off = dest_offset;
            card32 source_off = source_offset;
			if( num_ret_vert != numVertices || source_attrib.dataFormat == VAD::DATA_FORMAT_CARD8 || DO_NOT_USE_EXPERIMENTAL_FAST_COPY)
			{
				for( card32 i=0;i<numVertices;i++ )
				{
				   bool valid = true;
				   if( has_flags )
				   {
					  int32 flags = *((card32 *)&vertices[flags_offset]);
					  flags_offset += sourceVD.getSize();
					  valid = ((flags & includeFlags) == includeFlags) && ((flags & excludeFlags) == 0);
				   }

				   if( valid )
				   {
					  for( card32 j=0;j<numBytes;j++ )
						 ret_vertices[ dest_off + j  ] =
							vertices[ source_off + j ];

					  dest_off += getSize();
				   }
				   source_off += sourceVD.getSize();

				}
            }
			else
			{
				card32 * sourcePtr = (card32*)&vertices[source_off];
				card32 * destPtr = (card32*)&ret_vertices[dest_off];
				const card32 numQuads = numBytes / 4;
				const card32 rel_sourceVDSize = sourceVD.getSize()/4 - numQuads;
				const card32 rel_destVDSize = getSize()/4 - numQuads;
				for( card32 i=0;i<numVertices;i++ )
				{
					for( card32 j=0;j<numQuads;j++ )
					{
						*destPtr = *sourcePtr;
						destPtr++;
						sourcePtr++;
					}
					sourcePtr += rel_sourceVDSize;
					destPtr += rel_destVDSize;
				}
			}
         }
         else
         {
            // float <-> byte: f = b / 255.0f
            // int <-> byte i = b
            // float <-> int f = i / 255.0f

            // convert format (very slow)
            card32 dest_off = dest_offset;
            card32 source_off = source_offset;

            for( card32 i=0;i<numVertices;i++ )
            {
               bool valid = true;
               if( has_flags )
               {
                  int32 flags = *((int32 *)&vertices[flags_offset]);
                  flags_offset += sourceVD.getSize();
                  valid = ((flags & includeFlags) == includeFlags) && ((flags & excludeFlags) == 0);
               }

               if( valid )
               {
                  for( card32 j=0;j<numComp;j++ )
                  {
                     if( dest_attrib.dataFormat == VAD::DATA_FORMAT_CARD8 )
                     {
                        card8 tmp=0;
                        switch( source_attrib.dataFormat )
                        {
                        case VAD::DATA_FORMAT_INT32:
                           tmp = convertInt32ToCard8( &vertices[source_off + j*4] );
                        break;
                        case VAD::DATA_FORMAT_FLOAT32:
                           tmp = convertFloat32ToCard8( &vertices[source_off + j*4] );
                        break;
                        }
                        ret_vertices[ dest_off + j ] = tmp;
                     }

                     if( dest_attrib.dataFormat == VAD::DATA_FORMAT_INT32 )
                     {
                        int32 tmp=0;
                        switch( source_attrib.dataFormat )
                        {
                        case VAD::DATA_FORMAT_CARD8:
                           tmp = convertCard8ToInt32( &vertices[source_off + j] );
                        break;
                        case VAD::DATA_FORMAT_FLOAT32:
                           tmp = convertFloat32ToInt32( &vertices[source_off + j*4] );
                        break;
                        }

                        card8 * ptr = (card8 *)&tmp;
                        ret_vertices[ dest_off + j*4 + 0 ] = ptr[0];
                        ret_vertices[ dest_off + j*4 + 1 ] = ptr[1];
                        ret_vertices[ dest_off + j*4 + 2 ] = ptr[2];
                        ret_vertices[ dest_off + j*4 + 3 ] = ptr[3];
                     }

                     if( dest_attrib.dataFormat == VAD::DATA_FORMAT_FLOAT32 )
                     {
                        float32 tmp=0;
                        switch( source_attrib.dataFormat )
                        {
                        case VAD::DATA_FORMAT_CARD8:
                           tmp = convertCard8ToFloat32( &vertices[source_off + j] );
                        break;
                        case VAD::DATA_FORMAT_FLOAT32:
                           tmp = convertInt32ToFloat32( &vertices[source_off + j*4] );
                        break;
                        }

                        card8 * ptr = (card8 *)&tmp;
                        ret_vertices[ dest_off + j*4 + 0 ] = ptr[0];
                        ret_vertices[ dest_off + j*4 + 1 ] = ptr[1];
                        ret_vertices[ dest_off + j*4 + 2 ] = ptr[2];
                        ret_vertices[ dest_off + j*4 + 3 ] = ptr[3];
                     }

                  }
                  dest_off += getSize();
               }

               source_off += sourceVD.getSize();
            }
         }
      }
   }


}

void VertexDescriptor::buildTangentialQuads( PointSet * ps, std::vector<card8> & ret_vertices, float scale ) const
{
   buildTangentialQuads( ps->getDataPtr(), ps->getDataPtr(), ps->getNumEntries(), *ps->getDescr(), ret_vertices,scale );
}

void VertexDescriptor::buildTangentialQuads( const char * source_vertices, const char * source_tangents, card32 numVertices,
         const VertexDescriptor &sourceTangentVD, std::vector<card8> &ret_vertices, float scale   ) const
{
   x4Assert( providesAttribute("position") );
   x4Assert( sourceTangentVD.providesAttribute("tangent_u") );
   x4Assert( sourceTangentVD.providesAttribute("tangent_v") );

   VAD vad_pos = getAttribute( "position" );
   VAD vad_tangu = sourceTangentVD.getAttribute( "tangent_u" );
   VAD vad_tangv = sourceTangentVD.getAttribute( "tangent_v" );

   x4Assert( vad_pos.dataFormat == VAD::DATA_FORMAT_FLOAT32 );
   x4Assert( vad_tangu.dataFormat == VAD::DATA_FORMAT_FLOAT32 );
   x4Assert( vad_tangv.dataFormat == VAD::DATA_FORMAT_FLOAT32 );
   x4Assert( vad_tangu.numComponents == vad_tangv.numComponents );

   card32 offset_pos = getAttributeOffset( getIndex( "position" ) );
   card32 offset_tangu = sourceTangentVD.getAttributeOffset( sourceTangentVD.getIndex( "tangent_u") );
   card32 offset_tangv = sourceTangentVD.getAttributeOffset( sourceTangentVD.getIndex( "tangent_v") );

   // simple copy every vertex 4x
   ret_vertices = std::vector<card8>( numVertices * getSize() * 4 );
   for( card32 i=0;i<numVertices;i++ )
   {
      for( card32 j=0;j<getSize();j++ )
      {
         ret_vertices[i*4*getSize()+j] = source_vertices[i*getSize()+j];
         ret_vertices[(i*4+1)*getSize()+j] = source_vertices[i*getSize()+j];
         ret_vertices[(i*4+2)*getSize()+j] = source_vertices[i*getSize()+j];
         ret_vertices[(i*4+3)*getSize()+j] = source_vertices[i*getSize()+j];
      }
   }

   for( card32 i=0;i<numVertices;i++ )
   {
      float * dst_1 = (float*)&ret_vertices[offset_pos];
      float * dst_2 = (float*)&ret_vertices[offset_pos+getSize()];
      float * dst_3 = (float*)&ret_vertices[offset_pos+getSize()*2];
      float * dst_4 = (float*)&ret_vertices[offset_pos+getSize()*3];
      float * src_u = (float*)&source_tangents[offset_tangu];
      float * src_v = (float*)&source_tangents[offset_tangv];

      for( card32 j=0;(j<vad_tangu.numComponents) && (j<vad_pos.numComponents);j++ )
      {
         dst_1[j] += (src_u[j] - src_v[j])*scale;
         dst_2[j] += (src_u[j] + src_v[j])*scale;
         dst_3[j] += (-src_u[j] + src_v[j])*scale;
         dst_4[j] += (-src_u[j] - src_v[j])*scale;
      }

      offset_pos += getSize() * 4;
      offset_tangu += sourceTangentVD.getSize();
      offset_tangv += sourceTangentVD.getSize();
   }

}

void VertexDescriptor::buildSplattingQuads( const PointSet * ps, std::vector<card8> & ret_vertices,
   const int32 includeFlags, const int32 excludeFlags ) const
{
   buildSplattingQuads( ps->getDataPtr(), ps->getNumEntries(), ps->getDescr(), ret_vertices,
      includeFlags, excludeFlags );
}

void VertexDescriptor::buildSplattingQuads( const char * source_vertices, card32 numVertices,
      const VertexDescriptor * sourceVD, std::vector<card8> &ret_vertices,
      const int32 includeFlags, const int32 excludeFlags ) const
{
   x4Assert( providesAttribute("position") );
   x4Assert( providesAttribute("pcoord") );
   x4Assert( providesAttribute("tangent_u") );
   x4Assert( providesAttribute("tangent_v") );
   x4Assert( sourceVD->providesAttribute("position") );
   x4Assert( sourceVD->providesAttribute("tangent_u") );
   x4Assert( sourceVD->providesAttribute("tangent_v") );

   VAD vad_pos = getAttribute( "position" );
   VAD vad_tangu = getAttribute( "tangent_u" );
   VAD vad_tangv = getAttribute( "tangent_v" );
   VAD vad_pcoord = getAttribute( "pcoord" );
   VAD src_vad_pos = sourceVD->getAttribute( "position" );
   VAD src_vad_tangu = sourceVD->getAttribute( "tangent_u" );
   VAD src_vad_tangv = sourceVD->getAttribute( "tangent_v" );

   x4Assert( vad_pos.dataFormat == VAD::DATA_FORMAT_FLOAT32 );
   x4Assert( vad_tangu.dataFormat == VAD::DATA_FORMAT_FLOAT32 );
   x4Assert( vad_tangv.dataFormat == VAD::DATA_FORMAT_FLOAT32 );
   x4Assert( vad_tangu.numComponents == vad_tangv.numComponents );
   x4Assert( src_vad_pos.numComponents == vad_pos.numComponents );
   x4Assert( src_vad_pos.numComponents == 3 );
   x4Assert( src_vad_pos.dataFormat == VAD::DATA_FORMAT_FLOAT32 );
   x4Assert( src_vad_tangu.dataFormat == VAD::DATA_FORMAT_FLOAT32 );
   x4Assert( src_vad_tangv.dataFormat == VAD::DATA_FORMAT_FLOAT32 );
   x4Assert( src_vad_tangv.numComponents == src_vad_tangu.numComponents );
   x4Assert( src_vad_tangv.numComponents == vad_tangu.numComponents );
   x4Assert( vad_pcoord.dataFormat == VAD::DATA_FORMAT_FLOAT32 );
   x4Assert( vad_pcoord.numComponents >= 2 );


   card32 offset_pcoord = getAttributeOffset( getIndex( "pcoord") );

   std::vector<card8> tmpVertices;
   convertVertices( source_vertices, numVertices, *sourceVD, tmpVertices,includeFlags, excludeFlags );

   expandVertices( tmpVertices, 4, ret_vertices );
   for( card32 i=0;i<tmpVertices.size()/getSize();i++ )
   {
      card32 offset0 = i * 4 * getSize() + offset_pcoord;
      card32 offset1 = (i * 4 + 1) * getSize() + offset_pcoord;
      card32 offset2 = (i * 4 + 2) * getSize() + offset_pcoord;
      card32 offset3 = (i * 4 + 3) * getSize() + offset_pcoord;
      ((float*)&ret_vertices[offset0])[0] = -1.0f;
      ((float*)&ret_vertices[offset0])[1] = -1.0f;

      ((float*)&ret_vertices[offset1])[0] =  1.0f;
      ((float*)&ret_vertices[offset1])[1] = -1.0f;

      ((float*)&ret_vertices[offset2])[0] =  1.0f;
      ((float*)&ret_vertices[offset2])[1] =  1.0f;

      ((float*)&ret_vertices[offset3])[0] = -1.0f;
      ((float*)&ret_vertices[offset3])[1] =  1.0f;

   }


}


void VertexDescriptor::expandVertices( const std::vector<card8> &source_vertices, card32 n, std::vector<card8> &ret_vertices ) const
{
   expandVertices( &source_vertices[0], source_vertices.size() / getSize(), n, ret_vertices );
}

void VertexDescriptor::expandVertices( const card8 * source_vertices, mpcard numVertices, card32 n, std::vector<card8> &ret_vertices ) const
{
   ret_vertices = std::vector<card8>( numVertices * n * getSize() );

   card32 offset = 0;
   for( card32 i=0;i<numVertices;i++ )
      for( card32 j=0;j<n;j++ )
         for( card32 k=0;k<getSize();k++ )
         {
            ret_vertices[offset++] = source_vertices[i*getSize()+k];
         }
}

void VertexDescriptor::modifyVerticesf(  card8* dest_vertices, card32 numVertices, card32 period, card32 phase,
            const std::string &channel, card32 op, float param ) const
{
   float param_vec[4];
   param_vec[0] = param;
   param_vec[1] = param;
   param_vec[2] = param;
   param_vec[3] = param;
   modifyVertices4fv( dest_vertices, numVertices, period, phase, channel, op, param_vec );
}

void VertexDescriptor::modifyVertices4fv(  card8* dest_vertices, card32 numVertices, card32 period, card32 phase,
            const std::string &channel, card32 op, float * param ) const
{
   x4Assert( providesAttribute(channel) );
   x4Assert( period>0 );

   VAD vad = getAttribute( getIndex( channel ) );
   x4Assert( vad.dataFormat == VAD::DATA_FORMAT_FLOAT32 );

   card32 offset = getAttributeOffset( getIndex( channel ) );

   for( card32 i=phase;i<numVertices;i+=period )
   {
      float * attr = (float*)&dest_vertices[i*getSize() + offset];
      for( card32 j=0;j<vad.numComponents;j++ )
      {
         switch( op )
         {
         case MODIFIER_SET:
            attr[j] = param[j];
         break;
         case MODIFIER_ADD:
            attr[j] += param[j];
         break;
         case MODIFIER_MULT:
            attr[j] *= param[j];
         break;
         case MODIFIER_INVERT:
            attr[j] = -attr[j];
         break;
         }
      }
   }
}

void VertexDescriptor::modifyVerticesf(  card8* dest_vertices, mpcard numVertices, card32 period, card32 phase,
            const std::string &dest_channel, card32 op,const std::string &src_channel, float param ) const
{
   float param_vec[4];
   param_vec[0] = param;
   param_vec[1] = param;
   param_vec[2] = param;
   param_vec[3] = param;
   modifyVertices4fv( dest_vertices, numVertices, period, phase, dest_channel, op, src_channel, param_vec );
}

void VertexDescriptor::modifyVertices4fv(  card8* dest_vertices, card32 numVertices, card32 period, card32 phase,
            const std::string &dest_channel, card32 op, const std::string &src_channel, float * param ) const
{
   x4Assert( providesAttribute(dest_channel) );
   x4Assert( providesAttribute(src_channel) );
   x4Assert( period>0 );

   VAD vadDest = getAttribute( getIndex( dest_channel ) );
   VAD vadSrc = getAttribute( getIndex( src_channel ) );
   x4Assert( vadDest.dataFormat == VAD::DATA_FORMAT_FLOAT32 );
   x4Assert( vadSrc.dataFormat == VAD::DATA_FORMAT_FLOAT32 );
   x4Assert( vadDest.numComponents == vadSrc.numComponents );

   card32 destoffset = getAttributeOffset( getIndex( dest_channel ) );
   card32 srcoffset = getAttributeOffset( getIndex( src_channel ) );


   for( card32 i=phase;i<numVertices;i+=period )
   {
      float * attr = (float*)&dest_vertices[i*getSize() + destoffset];
      float * attr_src = (float*)&dest_vertices[i*getSize() + srcoffset];
      for( card32 j=0;j<vadDest.numComponents;j++ )
      {
         switch( op )
         {
         case MODIFIER_SET:
            attr[j] = attr_src[j] * param[j];
         break;
         case MODIFIER_ADD:
            attr[j] += attr_src[j] * param[j];
         break;
         case MODIFIER_MULT:
            attr[j] *= attr_src[j] * param[j];
         break;
         case MODIFIER_INVERT:
            attr[j] = -attr[j];
         break;
         }
      }
   }
}


bool VertexDescriptor::equal(const Object* obj) const {


   x4Assert(dynamic_cast<const VertexDescriptor*>(obj));
   VertexDescriptor* vd2 = (VertexDescriptor*)obj;

   if (getNumAttributes() != vd2->getNumAttributes()) return false;

   for( card32 i=0;i<getNumAttributes();i++ )
   {
      VAD tmp = getAttribute(i);
      VAD tmp2 = vd2->getAttribute(i);

      if (!((tmp.name==tmp2.name) &&
           (tmp.numComponents==tmp2.numComponents) &&
           (tmp.dataFormat==tmp2.dataFormat))) {
          return false;
       }
   }

   return true;

};


void VertexDescriptor::interpolate(const char **source, float32 *factors, card32 numVertices, char *dest) const {
   mpcard n = attributes.size();
   for (mpcard i=0; i<n; i++) {
      card32 offset = offsets[i];
      VAD vad = attributes[i];
      vad.zero(dest+offset);
      for (card32 v=0; v<numVertices; v++) {
         vad.add(dest+offset, source[v]+offset, factors[v]);
      }
   }
}


void VertexDescriptor::zero(char *data) const {
   mpcard n = attributes.size();
   for (mpcard i=0; i<n; i++) {
      card32 offset = offsets[i];
      VAD vad = attributes[i];
      vad.zero(data+offset);
   }
}

void VertexDescriptor::appendAttribs(const card8 *source, const VertexDescriptor *sourceVD, card8 *dest, const VertexDescriptor *destVD) {
   const int32 sourceSize = sourceVD->getSize();
   const int32 destSize = destVD->getSize();
   const int32 addDestOffset = destSize - sourceSize;
   x4Assert(addDestOffset >= 0);
   for (int32 b=0; b<sourceSize; b++) {
      *dest = *source;
      dest++;
      source++;
   }
   for (int32 b=sourceSize; b<destSize; b++) {
      *dest = 0;
      dest++;
   }
}
