//---------------------------------------------------------------------------
#include "stdafx.h"

//---------------------------------------------------------------------------
#include "GLBuffer.h"
//---------------------------------------------------------------------------
#include "GLContext.h"
#include "GLObject.h"
#include "X4Exception.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------



GLBuffer::GLBuffer(GLContext * context, GLenum target)
:GLObject(context)
{
   glGenBuffersARB(1, &m_BufferID );
   m_SizeInBytes = 0;
   m_BufferTarget = target;
}

GLBuffer::~GLBuffer()
{
   if( isValid() )
      //glDeleteBuffersARB(1,&m_BufferID);
      getContext()->deleteBuffer( m_BufferID );
}

bool GLBuffer::init(GLuint size,GLenum usage)
{
   if( !isValid() )
      throw X4Exception("GLBuffer::init() - invalid context");
   m_SizeInBytes = size;
   glBindBufferARB( m_BufferTarget, m_BufferID );
   glBufferDataARB( m_BufferTarget, size, 0, usage );
   GLenum error = glGetError();
   if( error != GL_NO_ERROR )
      return false;
   unbind();

   return true;
}

void GLBuffer::bind()
{
   if( !isValid() )
      throw X4Exception("GLBuffer::bind() - invalid context");
   glBindBufferARB( m_BufferTarget, m_BufferID );
}

void GLBuffer::unbind()
{
   if( !isValid() )
      throw X4Exception("GLBuffer::unbind() - invalid context");
   glBindBufferARB( m_BufferTarget, 0 );
}

void GLBuffer::writeToBuffer( void * data, GLint offset,GLuint numBytes )
{
   if( !isValid() )
      throw X4Exception("GLBuffer::writeToBuffer() - invalid context");
   if( offset + numBytes > m_SizeInBytes )
      throw X4Exception("GLBuffer::writeToBuffer | out of range!");
   bind();
   glBufferSubDataARB( m_BufferTarget, offset, numBytes, data );
   unbind();
}

void GLBuffer::readFromBuffer( std::vector<card8> &destBuffer, GLint offset, GLuint numBytes )
{
   if( !isValid() )
      throw X4Exception("GLBuffer::readFromBuffer() - invalid context");
   if( offset + numBytes > m_SizeInBytes )
      throw X4Exception("GLBuffer::readFromBuffer | out of range!");
   if( destBuffer.size() != numBytes )
      destBuffer.resize( numBytes );
   bind();
   glGetBufferSubDataARB( m_BufferTarget, offset, numBytes, &destBuffer[0] );
   unbind();
}

void GLBuffer::onDeleteContext()
{
   if( isValid() )
      //glDeleteBuffersARB(1,&m_BufferID);
      getContext()->deleteBuffer( m_BufferID );
   else
      throw X4Exception("GLBuffer::onDeleteContext() - invalid context");

}



GLVertexBuffer::GLVertexBuffer( GLContext * context, const VertexDescriptor & vdesc,int type )
:GLBuffer(context, GL_ARRAY_BUFFER)
{
   m_UseDefaultGLMapping = false;
   m_VertexDescriptor.assign( &vdesc );
   m_NumElements = 0;
   m_BufferType = type;
   if( m_BufferType == TYPE_RENDERED32 )
   {
      for( card32 i=0;i<m_VertexDescriptor.getNumAttributes();i++ )
      {
         x4Assert( m_VertexDescriptor[i].numComponents == 4 );
         x4Assert( m_VertexDescriptor[i].dataFormat == VAD::DATA_FORMAT_FLOAT32 );
      }
   }
}

bool GLVertexBuffer::init(GLuint numElements, GLenum usage )
{
   m_NumElements = numElements;
   if( m_BufferType == TYPE_DEFAULT )
      return GLBuffer::init( numElements * m_VertexDescriptor.getSize(), usage );
   else
   {
      glBindBufferARB(GL_PIXEL_PACK_BUFFER_EXT, m_BufferID);
      glBufferDataARB(GL_PIXEL_PACK_BUFFER_EXT,
         m_VertexDescriptor.getSize() * m_NumElements, 0, usage); // undefined data
      GLenum error = glGetError();
      if( error != GL_NO_ERROR )
         return false;
      glBindBufferARB(GL_PIXEL_PACK_BUFFER_EXT, 0);
      m_SizeInBytes = numElements * m_VertexDescriptor.getSize();

      return true;
   }
}

void GLVertexBuffer::bind()
{
   GLBuffer::bind();
   card32 numAttribs = m_VertexDescriptor.getNumAttribs();
   for( card32 i=0;i<numAttribs;i++ )
   {

      GLuint index=i;

      VertexAttribDescriptor attrib = m_VertexDescriptor.getAttribute(i);
      GLenum vType = GL_FLOAT;   // default: FLOAT
      switch( attrib.dataFormat )
      {
      case VertexAttribDescriptor::DATA_FORMAT_CARD8:
         vType = GL_UNSIGNED_BYTE;
      break;
      case VertexAttribDescriptor::DATA_FORMAT_FLOAT32:
         vType = GL_FLOAT;
      break;
      case VertexAttribDescriptor::DATA_FORMAT_INT32:
         vType = GL_INT;
      break;
      }

      if( !m_UseDefaultGLMapping )
      {
         glEnableVertexAttribArrayARB( index );
         glVertexAttribPointerARB( index, attrib.numComponents, vType,
                     false,   // do not normalize
                     m_VertexDescriptor.getSize(), //stride
                     (GLubyte *)(NULL + m_VertexDescriptor.getAttributeOffset(i)) );
      }
      else
      {

         if( attrib.name == "position" )
         {
               glEnableClientState( GL_VERTEX_ARRAY );
               glVertexPointer( attrib.numComponents, vType,
                     m_VertexDescriptor.getSize(), //stride
                     (GLubyte *)(NULL + m_VertexDescriptor.getAttributeOffset(i)) );
         }

         if( (attrib.name == "normal") && attrib.numComponents == 3 )
         {
               glEnableClientState( GL_NORMAL_ARRAY );
               glNormalPointer( vType,
                     m_VertexDescriptor.getSize(), //stride
                     (GLubyte *)(NULL + m_VertexDescriptor.getAttributeOffset(i)) );
         }

         if( attrib.name == "color" )
         {
               glEnableClientState( GL_COLOR_ARRAY );
               glColorPointer( attrib.numComponents, vType,
                     m_VertexDescriptor.getSize(), //stride
                     (GLubyte *)(NULL + m_VertexDescriptor.getAttributeOffset(i)) );
         }
         if( attrib.name == "texcoord" )
         {
               glEnableClientState( GL_TEXTURE_COORD_ARRAY );
               glTexCoordPointer( attrib.numComponents, vType,
                     m_VertexDescriptor.getSize(), //stride
                     (GLubyte *)(NULL + m_VertexDescriptor.getAttributeOffset(i)) );
          }
/*         switch( i )
         {
            case MAP_POSITION:
               glEnableClientState( GL_VERTEX_ARRAY );
               glVertexPointer( attrib.numComponents, vType,
                     m_VertexDescriptor.getSize(), //stride
                     (GLubyte *)(NULL + m_VertexDescriptor.getAttributeOffset(i)) );
            break;
            case MAP_NORMAL:
               glEnableClientState( GL_NORMAL_ARRAY );
               glNormalPointer( vType,
                     m_VertexDescriptor.getSize(), //stride
                     (GLubyte *)(NULL + m_VertexDescriptor.getAttributeOffset(i)) );
            break;
            case MAP_COLOR:
               glEnableClientState( GL_COLOR_ARRAY );
               glColorPointer( attrib.numComponents, vType,
                     m_VertexDescriptor.getSize(), //stride
                     (GLubyte *)(NULL + m_VertexDescriptor.getAttributeOffset(i)) );
            break;
            case MAP_TEXCOORD0:
               glEnableClientState( GL_TEXTURE_COORD_ARRAY );
               glTexCoordPointer( attrib.numComponents, vType,
                     m_VertexDescriptor.getSize(), //stride
                     (GLubyte *)(NULL + m_VertexDescriptor.getAttributeOffset(i)) );
            break;
         }*/
      }

   }
}

void GLVertexBuffer::unbind()
{
 card32 numAttribs = m_VertexDescriptor.getNumAttribs();
   for( card32 i=0;i<numAttribs;i++ )
   {
      GLuint index=i;
      VertexAttribDescriptor attrib = m_VertexDescriptor.getAttribute(i);
      if( !m_UseDefaultGLMapping )
         glDisableVertexAttribArrayARB( index );
      else
      {
         if( attrib.name == "position" )
         {
               glDisableClientState( GL_VERTEX_ARRAY );
         }

         if( (attrib.name == "normal") && attrib.numComponents == 3 )
         {
               glDisableClientState( GL_NORMAL_ARRAY );
         }

         if( attrib.name == "color" )
         {
               glDisableClientState( GL_COLOR_ARRAY );
         }
         if( attrib.name == "texcoord" )
         {
               glDisableClientState( GL_TEXTURE_COORD_ARRAY );
          }

/*
         switch( i )
         {
            case MAP_POSITION:
               glDisableClientState( GL_VERTEX_ARRAY );
            break;
            case MAP_NORMAL:
               glDisableClientState( GL_NORMAL_ARRAY );
            break;
            case MAP_COLOR:
               glDisableClientState( GL_COLOR_ARRAY );
            break;
            case MAP_TEXCOORD0:
               glDisableClientState( GL_TEXTURE_COORD_ARRAY );
            break;
         }*/
      }
   }

//   glDisableVertexAttribArrayARB( 0 );
   GLBuffer::unbind();
}

void GLVertexBuffer::writeToBuffer( void * data, GLint offset, GLuint numElements )
{
   GLBuffer::writeToBuffer( data, offset * m_VertexDescriptor.getSize(),
      numElements * m_VertexDescriptor.getSize() );
}

void GLVertexBuffer::readFromBuffer( std::vector<card8> &destBuffer, GLint offset, GLuint numElements )
{
   GLBuffer::readFromBuffer( destBuffer, offset * m_VertexDescriptor.getSize(),
      numElements * m_VertexDescriptor.getSize() );
}

void GLVertexBuffer::copyFromRenderTarget( card32 x, card32 y, card32 width, card32 height,
   GLenum rbuffer )
{
   x4Assert( m_BufferType == TYPE_RENDERED32 );
   card32 numElementsToCopy = width*height/ m_VertexDescriptor.getNumAttributes();
   x4Assert( numElementsToCopy  <= m_NumElements );

   glBindBufferARB(GL_PIXEL_PACK_BUFFER_EXT, m_BufferID);
   glReadBuffer(rbuffer);
   glReadPixels(x* m_VertexDescriptor.getNumAttributes(), y, width * m_VertexDescriptor.getNumAttributes(), height, GL_RGBA, GL_FLOAT, 0);

   glBindBufferARB(GL_PIXEL_PACK_BUFFER_EXT, 0);

}

void GLVertexBuffer::drawArrays( GLenum mode, GLuint first, GLuint count )
{
   if( count == 0 )
      count = m_NumElements;
   glDrawArrays( mode, first, count );
}
