//---------------------------------------------------------------------------
#ifndef GLBufferH
#define GLBufferH
//---------------------------------------------------------------------------
#include "x3openGL.h"
//---------------------------------------------------------------------------
#include "GLObject.h"
#include "VertexDescriptor.h"
#include "gl.h"
//---------------------------------------------------------------------------

namespace X4 {

class GLContext;

class X3OPENGL_API GLBuffer : public GLObject
{
public:
   GLBuffer( GLContext *, GLenum target );
   virtual ~GLBuffer();

   void bind();
   void unbind();

   virtual void onDeleteContext();

   inline GLuint getSizeInBytes() const {return m_SizeInBytes;}

protected:

   bool init( GLuint size, GLenum usage = GL_STATIC_DRAW_ARB );
   void writeToBuffer( void * data, GLint offset, GLuint numBytes );
   void readFromBuffer( std::vector<card8> &destBuffer, GLint offset, GLuint numBytes );

   GLuint   m_BufferID;
   GLuint   m_SizeInBytes;
   GLenum   m_BufferTarget;

   int *    m_Test;
};

class X3OPENGL_API GLVertexBuffer : public GLBuffer
{
public:

   static const int TYPE_DEFAULT = 0;      /// default vertex buffer object
   static const int TYPE_RENDERED32 = 1;  /// used to copy data from float pbuffers(32Bit), attributes must be 4 float32

   GLVertexBuffer( GLContext *, const VertexDescriptor &, int type=TYPE_DEFAULT);

   bool init( GLuint numElements, GLenum usage = GL_STATIC_DRAW );

   void bind();
   void unbind();

   void writeToBuffer( void * data, GLint offset, GLuint numElements );
   void readFromBuffer( std::vector<card8> &destBuffer, GLint offset, GLuint numElements );

   void copyFromRenderTarget( card32 x, card32 y, card32 width, card32 height,
      GLenum rBuffer = GL_FRONT );

   void drawArrays( GLenum mode, GLuint first=0, GLuint count=0 );

   GLuint getNumElements() {return m_NumElements;}

   bool getMappingFlag() const {return m_UseDefaultGLMapping;}
   void setMappingFlag( bool f ) {m_UseDefaultGLMapping=f;}

   VertexDescriptor * getVertexDescriptor() {return &m_VertexDescriptor;}

/*
   GL Default Mapping
   channel Nr  |  GL channel
   ---------------------------------
   0           |  Position
   1           |  Color
   2           |  Normal
   3           |  TexCoord0

*/

   static const int MAP_POSITION = 0;
   static const int MAP_COLOR = 1;
   static const int MAP_NORMAL = 2;
   static const int MAP_TEXCOORD0 = 3;
protected:

   VertexDescriptor  m_VertexDescriptor;
   GLuint            m_NumElements;
   bool              m_UseDefaultGLMapping;  // vertex, normal, color ,...
   int               m_BufferType;
};


template<class TElement> 
class GLTemplatedIndexBuffer : public GLBuffer
{
public:
   GLTemplatedIndexBuffer<TElement>( GLContext * context )
   :GLBuffer(context, GL_ELEMENT_ARRAY_BUFFER_ARB)
   {
      m_NumElements = 0;
   }

   bool init( GLuint numElements, GLenum usage=GL_STATIC_DRAW )
   {
      m_NumElements = numElements;
      return GLBuffer::init( numElements * sizeof( TElement ), usage );
   }

   void writeToBuffer( TElement * data, GLint offset,GLuint numElements )
   {
      GLBuffer::writeToBuffer( data, offset, numElements*sizeof( TElement ) );
   }

   void drawElements( GLenum mode, GLuint first=0, GLuint count=0 )
   {
   }

   GLuint getNumElements() {return m_NumElements;}

protected:

   GLuint   m_NumElements;
};

template<>
inline void GLTemplatedIndexBuffer<card32>::drawElements( GLenum mode, GLuint first, GLuint count )
{
   if( count == 0 )
      count = m_NumElements;
   glDrawElements( mode, count, GL_UNSIGNED_INT, (GLubyte *)NULL + 4*first );
}
template<>
inline void GLTemplatedIndexBuffer<card16>::drawElements( GLenum mode, GLuint first, GLuint count )
{
   if( count == 0 )
      count = m_NumElements;
   glDrawElements( mode, count, GL_UNSIGNED_SHORT,(GLubyte *)NULL + 2*first );
}

typedef GLTemplatedIndexBuffer<card32> GLIndexBuffer32;
typedef GLTemplatedIndexBuffer<card16> GLIndexBuffer16;


};

#endif
