#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "GLMultipleRenderTexture.h"
//---------------------------------------------------------------------------
#include "GLExtensions.h"
#include "GLTexture.h"
#include "GLContext.h"
#include "ImageDisplayer.h"
#include "RGBAImage.h"
//---------------------------------------------------------------------------
using namespace X4;

GLMultipleRenderTexture::GLMultipleRenderTexture(GLContext * context)
:GLObject(context)
{
	m_Status = None;
}

GLMultipleRenderTexture::~GLMultipleRenderTexture(void)
{
	for( unsigned i=0;i<m_RenderTextures.size();i++ )
	{
		delete m_RenderTextures[i];
	}
}

void GLMultipleRenderTexture::init(unsigned width, unsigned height, bool hasDepth, bool hasStencil) 
{
	m_Width = width;
	m_Height = height;
	m_HasDepth = hasDepth;
	m_HasStencil = hasStencil;
}

void GLMultipleRenderTexture::addRenderTexture(unsigned numComponents, bool isFloat, bool isFloat32) 
{
	GLTextureRect * texture = new GLTextureRect(getContext());
	if( isFloat )
		texture->createFloatTextureFromMemory(m_Width,m_Height,numComponents,NULL,isFloat32);
	else
		texture->createFromMemory(m_Width,m_Height,numComponents,NULL);
	m_RenderTextures.push_back(texture);
}

bool GLMultipleRenderTexture::create() 
{
	if( m_RenderTextures.empty() )
		return false;	// does not support #rt=0

	glGenFramebuffersEXT( 1, &m_FrameBufferID );
   glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, m_FrameBufferID );

   // attach textures to buffer
   for( unsigned i=0;i<m_RenderTextures.size();i++ )
   {
	   glFramebufferTexture2DEXT(    GL_FRAMEBUFFER_EXT,
		                             GL_COLOR_ATTACHMENT0_EXT + i,
			                         m_RenderTextures[i]->getTextureTarget(),
				                     m_RenderTextures[i]->getTextureID(), 0);
   }

   // attach depth buffer
   if( m_HasDepth )
   {
      glGenRenderbuffersEXT( 1, &m_DepthRenderBufferID );
      glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, m_DepthRenderBufferID);
      glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT,
                               GL_DEPTH_COMPONENT24, m_Width, m_Height);
      glFramebufferRenderbufferEXT( GL_FRAMEBUFFER_EXT,
                                    GL_DEPTH_ATTACHMENT_EXT,
                                    GL_RENDERBUFFER_EXT,
                                    m_DepthRenderBufferID);

   }

   // attach stencil buffer
   if( m_HasStencil )
   {
      glGenRenderbuffersEXT( 1, &m_StencilRenderBufferID );
      glBindRenderbufferEXT(GL_RENDERBUFFER, m_StencilRenderBufferID);
      glRenderbufferStorageEXT(GL_RENDERBUFFER,
                               GL_STENCIL_INDEX, m_Width, m_Height);
      glFramebufferRenderbufferEXT( GL_FRAMEBUFFER_EXT,
                                    GL_STENCIL_ATTACHMENT_EXT,
                                    GL_RENDERBUFFER_EXT,
                                    m_StencilRenderBufferID);

   }

   if( !checkStatus() )
      return false;

   glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, 0 );

   return true;
}

bool GLMultipleRenderTexture::checkStatus()
{
    GLenum status;
    status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
    if( status ==  GL_FRAMEBUFFER_COMPLETE_EXT || status == 0 /*ATI, dont ask!*/ )
         return true;

	message("GLMultipleRenderTexture::checkStatus() - ERROR: ");
	switch(status)
	{
	case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT:
		message("GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT"); break;
	case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT:
		message("GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT"); break;
	//case GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS:
	//	message("GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS"); break;
	//case GL_FRAMEBUFFER_INCOMPLETE_FORMATS:
	//	message("GL_FRAMEBUFFER_INCOMPLETE_FORMATS"); break;
	case  GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT:
		message(" GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER"); break;
	case GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT:
		message("GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER"); break;
	case GL_FRAMEBUFFER_UNSUPPORTED_EXT:
		message("GL_FRAMEBUFFER_UNSUPPORTED"); break;
	default:
		message("unkown error code"); break;
	}
	message("\n");

    return false;
}

void GLMultipleRenderTexture::activate() 
{
	if( m_Status != None )
		throw X4Exception("GLMultipleRenderTexture::activate() - unbind textures/use deactivate() method!");

	m_Status = Active;
	glGetIntegerv( GL_VIEWPORT, m_OldViewport );
	glPushAttrib(GL_VIEWPORT_BIT | GL_COLOR_BUFFER_BIT);
	glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, m_FrameBufferID );
	glViewport(0,0, m_Width, m_Height);

	std::vector<GLenum> buffers;
	for( unsigned i=0;i<m_RenderTextures.size();i++ )
	{
	   buffers.push_back(GL_COLOR_ATTACHMENT0_EXT + i);
	}
	glDrawBuffersARB((GLsizei)buffers.size(), &buffers[0] );
}

void GLMultipleRenderTexture::deactivate() 
{
	if( m_Status != Active )
		throw X4Exception("GLMultipleRenderTexture::deactivate() - targets not active!");
	m_Status = None;

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	glPopAttrib();
}

void GLMultipleRenderTexture::bindAll(unsigned firstStage) 
{
	if( m_Status != None )
		throw X4Exception("GLMultipleRenderTexture::bindAll() - unbind textures/use deactivate() method!");

	m_Status = Bound;
	m_LastStage = firstStage;
	for( unsigned i=0;i<m_RenderTextures.size();i++ )		
	{
		m_RenderTextures[i]->bind(firstStage + i);
	}
}

void GLMultipleRenderTexture::unbindAll() 
{
	if( m_Status != Bound && m_Status != PartiallyBound)
		throw X4Exception("GLMultipleRenderTexture::bindAll() - textures not bound!");

	if( m_Status == Bound )
	{
		for( unsigned i=0;i<m_RenderTextures.size();i++ )		
		{
			m_RenderTextures[i]->unbind(m_LastStage+i);
		}
	}
	else
	{
		for( unsigned i=0;i<m_BoundTextures.size();i++ )
		{
			m_RenderTextures[m_BoundTextures[i].id]->unbind(m_BoundTextures[i].stage);
		}
	}
	m_Status = None;
}

void GLMultipleRenderTexture::outputViaImageDisplayer() 
{
	for( unsigned i=0;i<m_RenderTextures.size();i++ )
	{
		RGBAFloat32Image * ret = new RGBAFloat32Image;
		ret->resize( getWidth(), getHeight() );
		m_RenderTextures[i]->bind();
		glGetTexImage( m_RenderTextures[i]->getTextureTarget(),0, GL_RGBA, GL_FLOAT,
			ret->getPixels() );
		m_RenderTextures[i]->unbind();
		
		ImageDisplayer::DisplayRGBAFloat32Image ev(ret);
		debugOutput.displayEvent(ImageDisplayer::getClass(), std::string("mrt") + intToStr(i), &ev);
		delete ret;
	}
}

void GLMultipleRenderTexture::bind(unsigned texid, unsigned stage) 
{
	if( m_Status != None && m_Status != PartiallyBound )
		throw X4Exception("GLMultipleRenderTexture::bind() - unbind textures/use deactivate() method!");

	m_Status = PartiallyBound;

	TBoundTexture e;
	e.id = texid;
	e.stage = stage;
	m_BoundTextures.push_back(e);

	m_RenderTextures[texid]->bind(stage);
}
