//---------------------------------------------------------------------------
#include "stdafx.h"

//---------------------------------------------------------------------------
#include "GLShader.h"
//---------------------------------------------------------------------------
#include <string>
#include <vector>
#include "cg_include.h"
#include "GLContext.h"
#include "GLObject.h"
#include "X4Exception.h"
#include "Tools.h"
//---------------------------------------------------------------------------
 
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------



ShaderConstant::ShaderConstant( const std::string &name, int index, int numComponents )
{
   m_Name = name;
   m_Index = index;
   m_NumComponents = numComponents;
   m_Data = std::vector<float>(numComponents);
}

GLShader::GLShader( GLContext * context )
:GLObject(context)
{
}

void GLShader::setConstants( const std::vector<ShaderConstant> &constants )
{
   for( unsigned int i=0; i<constants.size(); i++ )
   {
      setConstant( constants[i] );
   }
}


GLShaderARB::GLShaderARB( GLContext * context, GLenum target )
:GLShader(context)
{
   m_Target = target;
   glGenProgramsARB(1, &m_ProgramID );
   m_LastErrorPosition = -1;
}

GLShaderARB::~GLShaderARB()
{
   if( isValid() )
      //glDeleteProgramsARB(1, &m_ProgramID );
      getContext()->deleteShader( m_ProgramID );
}

void GLShaderARB::onDeleteContext()
{
   if( isValid() )
      //glDeleteProgramsARB(1, &m_ProgramID );
      getContext()->deleteShader( m_ProgramID );
   else
      throw X4Exception("GLShaderARB::onDeleteContext() - invalid context!");
}

bool GLShaderARB::compile( const std::string &source )
{
   if( !isValid() )
      throw X4Exception("GLShaderARB::compile() - invalid context!");
   glBindProgramARB( m_Target, m_ProgramID );
   glGetError();  // clear last error
   glProgramStringARB( m_Target, GL_PROGRAM_FORMAT_ASCII_ARB, (GLsizei)source.length(),
      source.c_str() );
   if( glGetError() == GL_INVALID_OPERATION )
   {
      GLubyte * tmp = (GLubyte *)glGetString( GL_PROGRAM_ERROR_STRING_ARB );
      if( tmp )
         m_LastError = (char *)tmp;
      glGetIntegerv( GL_PROGRAM_ERROR_POSITION_ARB, &m_LastErrorPosition );
      return false;
   }

   m_LastError = "";
   m_LastErrorPosition = -1;

   return true;
}

void GLShaderARB::bind()
{
   if( !isValid() )
      throw X4Exception("GLShaderARB::bind() - invalid context!");
   glEnable( m_Target );
   glBindProgramARB( m_Target, m_ProgramID );
}

void GLShaderARB::unbind()
{
   if( !isValid() )
      throw X4Exception("GLShaderARB::unbind() - invalid context!");
   glDisable( m_Target );
}

void GLShaderARB::setConstant( const ShaderConstant & sc )
{
   if( !isValid() )
      throw X4Exception("GLShaderARB::setConstant() - invalid context!");
   const float * dataptr = sc.getDataPtr();
   switch( sc.getNumComponents() )
   {
   case 1:
      glProgramLocalParameter4fARB( m_Target,sc.getIndex(), dataptr[0], 0, 0, 0);
   break;
   case 2:
      glProgramLocalParameter4fARB( m_Target,sc.getIndex(), dataptr[0],
      dataptr[1], 0, 0);
   break;
   case 3:
      glProgramLocalParameter4fARB( m_Target,sc.getIndex(), dataptr[0],
      dataptr[1],dataptr[2], 0);
   break;
   case 4:
      glProgramLocalParameter4fvARB( m_Target,sc.getIndex(), dataptr);
   break;
   }
}

void GLShaderARB::setMatrix4f( const std::string &name, card32 index, float * matrix )
{
   glMatrixMode( GL_MATRIX0_ARB + index );
   glLoadIdentity();
   glMultMatrixf( matrix );
   glMatrixMode( GL_MODELVIEW );
   forget(name);
}


void GLShaderARB::setLocalParameter4f( GLuint index,  float x, float y, float z, float w)
{
   if( !isValid() )
      throw X4Exception("GLShaderARB::setLocalParameter4f() - invalid context!");
   glProgramLocalParameter4fARB( m_Target, index, x,y,z,w );
}

void GLShaderARB::setLocalParameter4fv( GLuint index, float * f )
{
   if( !isValid() )
      throw X4Exception("GLShaderARB::setLocalParameter4fv() - invalid context!");
   glProgramLocalParameter4fvARB( m_Target, index, f );
}

void GLShaderARB::setGlobalParameter4f( GLuint index,  float x, float y, float z, float w)
{
   if( !isValid() )
      throw X4Exception("GLShaderARB::setGlobalParameter4f() - invalid context!");
   glProgramEnvParameter4fARB( m_Target, index, x,y,z,w );
}

void GLShaderARB::setGlobalParameter4fv( GLuint index, float * f )
{
   if( !isValid() )
      throw X4Exception("GLShaderARB::setGlobalParameter4fv() - invalid context!");
   glProgramEnvParameter4fvARB( m_Target, index, f );
}

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
//
//                   Shader CG
//
// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// --------------------------------------------------------------------------


GLShaderCG::GLShaderCG( GLContext * context, CGprofile profile )
:GLShader( context )
{
   m_Profile = profile;
   m_Program = 0;
   m_GLProfile = 0;
   switch( m_Profile )
   {
   case CG_PROFILE_GP5VP:
   case CG_PROFILE_GP4VP:
   case CG_PROFILE_VP40:
   case CG_PROFILE_ARBVP1:
      m_GLProfile = GL_VERTEX_PROGRAM_ARB;
   break;
   case CG_PROFILE_GP5FP:
   case CG_PROFILE_GP4FP:
   case CG_PROFILE_FP40:
   case CG_PROFILE_ARBFP1:
      m_GLProfile = GL_FRAGMENT_PROGRAM_ARB;
   break;
   case CG_PROFILE_FP30:
      m_GLProfile = GL_FRAGMENT_PROGRAM_NV;
   break;
   case CG_PROFILE_VP30:
   case CG_PROFILE_VP20:
      m_GLProfile = GL_VERTEX_PROGRAM_NV;
   break;
   }


}

GLShaderCG::~GLShaderCG()
{
   if( isValid() )
      if( cgIsProgram( m_Program ) )
         cgDestroyProgram( m_Program );
}

void GLShaderCG::onDeleteContext()
{
   if( isValid() )
   {
      if( cgIsProgram( m_Program ) )
      cgDestroyProgram( m_Program );
   }
   else
      throw X4Exception("GLShaderCG::onDeleteContext() - invalid context");
}

bool GLShaderCG::compile( const std::string &source )
{
   if( !isValid() )
      throw X4Exception("GLShaderCG::compile() - invalid context");

   if( cgIsProgram( m_Program ) )
   {
      cgDestroyProgram( m_Program );
      m_Program = 0;
   }

   m_Program = cgCreateProgram( getContext()->getCGContext(),
      CG_SOURCE, source.c_str(), m_Profile, "main",  0 );

   if( m_Program )
   {
      cgGLLoadProgram( m_Program );
      m_LastError = "";
      return true;
   }
   else
   {
      const char * tmp =  cgGetLastListing(getContext()->getCGContext());
      if( tmp )
         m_LastError = tmp;
      else
         m_LastError = "unkown error";
      return false;
   }
}

void GLShaderCG::bind()
{
   if( m_Program )
   {
      //cgGLEnableProfile( m_Profile );
      glEnable( m_GLProfile );
      cgGLBindProgram ( m_Program );
      updateDefaultMatrices();
   }
}

void GLShaderCG::unbind()
{
   glDisable( m_GLProfile );
//   cgGLDisableProfile( m_Profile );
}

void GLShaderCG::updateDefaultMatrices()
{
   cgGLSetStateMatrixParameter( cgGetNamedParameter( m_Program,"matMVP" ), CG_GL_MODELVIEW_PROJECTION_MATRIX ,
      CG_GL_MATRIX_IDENTITY );
   cgGLSetStateMatrixParameter( cgGetNamedParameter( m_Program,"matMV" ), CG_GL_MODELVIEW_MATRIX ,
      CG_GL_MATRIX_IDENTITY );
   cgGLSetStateMatrixParameter( cgGetNamedParameter( m_Program,"matMVinv" ), CG_GL_MODELVIEW_MATRIX ,
      CG_GL_MATRIX_INVERSE );
   cgGLSetStateMatrixParameter( cgGetNamedParameter( m_Program,"matMVinvtrans" ), CG_GL_MODELVIEW_MATRIX ,
      CG_GL_MATRIX_INVERSE_TRANSPOSE );
   cgGLSetStateMatrixParameter( cgGetNamedParameter( m_Program,"matMVtrans" ), CG_GL_MODELVIEW_MATRIX ,
      CG_GL_MATRIX_TRANSPOSE );
}

void GLShaderCG::setMatrix4f( const std::string &name, card32 index, float * matrix )
{
   setMatrixParameterc( name, matrix );
   forget(index);
}

void GLShaderCG::setConstant( const ShaderConstant &sc )
{
   setParameterxf( sc.getName(), sc.getNumComponents(), sc.getDataPtr() );
}

void GLShaderCG::setParameter1f( const std::string &name, float x )
{
   setParameterxf( name, 1, &x );
}
void GLShaderCG::setParameter2f( const std::string &name, float x, float y )
{
   float tmp[4];
   tmp[0] = x;
   tmp[1] = y;
   tmp[2] = 0;
   tmp[3] = 0;
   setParameterxf( name, 2, tmp );
}
void GLShaderCG::setParameter3f( const std::string &name, float x, float y, float z )
{
   float tmp[4];
   tmp[0] = x;
   tmp[1] = y;
   tmp[2] = z;
   tmp[3] = 0;
   setParameterxf( name, 3, tmp );
}
void GLShaderCG::setParameter4f( const std::string &name, float x, float y, float z, float w )
{
   float tmp[4];
   tmp[0] = x;
   tmp[1] = y;
   tmp[2] = z;
   tmp[3] = w;
   setParameterxf( name, 4, tmp );
}

void GLShaderCG::setParameterxf( const std::string &name, GLuint numParms, const float *f)
{
   CGparameter parameter = cgGetNamedParameter( m_Program, name.c_str() );
   switch( numParms )
   {
   case 1:
      cgGLSetParameter1fv( parameter, f );
   break;
   case 2:
      cgGLSetParameter2fv( parameter, f );
   break;
   case 3:
      cgGLSetParameter3fv( parameter, f );
   break;
   case 4:
      cgGLSetParameter4fv( parameter, f );
   break;
   }
}

void GLShaderCG::setMatrixParameterr( const std::string &name, float * f )
{
   CGparameter parameter = cgGetNamedParameter( m_Program, name.c_str() );
   cgGLSetMatrixParameterfr( parameter, f );
}

void GLShaderCG::setMatrixParameterc( const std::string &name, float * f )
{
   CGparameter parameter = cgGetNamedParameter( m_Program, name.c_str() );
   cgGLSetMatrixParameterfc( parameter, f );
}


CGprofile GLShaderCG::getLatestVSProfile(const std::string& hint) {
	if (hint.length() == 0)
	   return cgGLGetLatestProfile( CG_GL_VERTEX );
	else
		return cgGetProfile(hint.c_str());
}

CGprofile GLShaderCG::getLatestPSProfile(const std::string& hint) {
	if (hint.length() == 0)
		return cgGLGetLatestProfile( CG_GL_FRAGMENT );
	else
		return cgGetProfile(hint.c_str());
}



