//---------------------------------------------------------------------------
#ifndef GLShaderH
#define GLShaderH
//---------------------------------------------------------------------------
#include "x3openGL.h"
//---------------------------------------------------------------------------
#include "cg_include.h"
#include <string>
#include "GLObject.h"
#include "X4Types.h"
//---------------------------------------------------------------------------

namespace X4 {

class GLContext;
class GLObject;


class X3OPENGL_API ShaderConstant
{
public:
   ShaderConstant( const std::string &name, int index, int numComponents );

   int      getIndex() const {return m_Index;}
   std::string   getName() const {return m_Name;}

   int      getNumComponents() const {return m_NumComponents;}
   const float *  getDataPtr() const {return &m_Data[0];}
   void     setData( card32 i, float v ) {m_Data[i] = v;}
   void     setData( const float * v ) { for( card32 i=0;i<4;i++ ) m_Data[i]=v[i]; }

protected:

   std::vector<float>  m_Data;
   int         m_NumComponents;
   int         m_Index;
   std::string      m_Name;
};

/// abstact interface for shaders
class X3OPENGL_API GLShader : public GLObject
{
public:
   GLShader( GLContext * );

   /// trys to compile a shader with std::string source
   /// if compilation failes, you can query debug-msg with getLastError
   virtual bool compile( const std::string &source ) = 0;

   /// enables shaderprofile
   /// bind the shaderprogram
   virtual void bind() = 0;

   /// update default matrices (
   /// CG doesn't use modelview, mvp, ... from gl-state
   /// you have to use this method to set matrices in CG - shaders
   virtual void updateDefaultMatrices() {}

   /// disables shaderprofile
   virtual void unbind() = 0;

   /// sets shadercontants ( 4 floats )
   virtual void setConstant( const ShaderConstant & ) = 0;
   void setConstants( const std::vector<ShaderConstant> & );

   /// set matrix
   virtual void setMatrix4f( const std::string &name, card32 index, float * matrix ) = 0;

   /// returns result of the last compilation
   /// "" if last comp. succeded
   std::string   getLastError() const {return m_LastError;}

   virtual void onDeleteContext() {}

protected:

   std::string      m_LastError;

};

/// ARB Shader
class X3OPENGL_API GLShaderARB : public GLShader
{
public:
   GLShaderARB( GLContext * , GLenum target);
   virtual ~GLShaderARB();

   virtual bool compile( const std::string &source );
   virtual void bind();
   virtual void unbind();
   virtual void setConstant( const ShaderConstant & );
   virtual void setMatrix4f( const std::string &name, card32 index, float * matrix );

   virtual void onDeleteContext();

   // ARB spezifisch
   void  setLocalParameter4f( GLuint index, float x, float y, float z, float w);
   void  setLocalParameter4fv( GLuint index, float *f );
   void  setGlobalParameter4f( GLuint index, float x, float y, float z, float w);
   void  setGlobalParameter4fv( GLuint index, float *f );
   int   getLastErrorPosition() {return m_LastErrorPosition;}  // -1 == NO ERROR

protected:

   GLenum      m_Target;
   GLuint      m_ProgramID;
   int         m_LastErrorPosition;

};

/// CG shader
class X3OPENGL_API GLShaderCG : public GLShader
{
public:
   GLShaderCG( GLContext *, CGprofile profile );
   virtual ~GLShaderCG();

   void setProfile( CGprofile profile );

   virtual bool compile( const std::string &source );
   virtual void bind();
   virtual void unbind();
   virtual void setConstant( const ShaderConstant & );
   virtual void setMatrix4f( const std::string &name, card32 index, float * matrix );

   virtual void onDeleteContext();

   virtual void updateDefaultMatrices();

   // CG spezifisch
   void setParameter1f( const std::string &name, float x);
   void setParameter2f( const std::string &name, float x, float y);
   void setParameter3f( const std::string &name, float x, float y, float z);
   void setParameter4f( const std::string &name, float x, float y, float z, float w);
   void setParameterxf( const std::string &name, GLuint numParms, const float * f);

   void setMatrixParameterr( const std::string &name, float * f );
   void setMatrixParameterc( const std::string &name, float * f );

   static CGprofile getLatestVSProfile(const std::string& hint = std::string());
   static CGprofile getLatestPSProfile(const std::string& hint = std::string());

protected:

   CGprofile   m_Profile;
   CGprogram   m_Program;
   GLenum      m_GLProfile;
};

}; // namespace GL

#endif
