//---------------------------------------------------------------------------
#include "stdafx.h"
//---------------------------------------------------------------------------
#include "Canvas2d.h"
//---------------------------------------------------------------------------
#include "Tools.h"
#include "BoundingBox.inline.h"
//---------------------------------------------------------------------------
#include <QtGui/QPainter>
#include <QtGui/QMouseEvent>
#include <QtGui/QResizeEvent>
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------



Canvas2d::Canvas2d(QWidget *parent, Qt::WindowFlags f)
   : QWidget(parent, f)
{
   ctm = IDENTITY3F;
   viewPortTransform = IDENTITY3F;
   fullTransform = IDENTITY3F;
   fullTransformInv = IDENTITY3F;
   scale = 1.0f;
   translation = NULL_VECTOR2F;
   lastVP = IDENTITY3F;
   lastX = 0;
   lastY = 0;
   lastScale = 1.0f;
   lastTranslation = NULL_VECTOR2F;
}

Vector2i Canvas2d::transform(const Vector2f &p) {
   Vector3f ph = makeVector3f(p[0], p[1], 1);
   ph = fullTransform*ph;
   Vector2i result = makeVector2i(ph[0], ph[1]);
   return result;
}

Vector2f Canvas2d::untransform(const Vector2i &p) {
   Vector3f ph = makeVector3f(p[0], p[1], 1);
   ph = fullTransformInv*ph;
   Vector2f result = makeVector2f(ph[0], ph[1]);
   return result;
}

void Canvas2d::addElement(Canvas2dElement *e) {
   items.push_back(e);
}

void Canvas2d::autoSize() {
   const size_t n = items.size();
   if (n == 0) return;
   BoundingBox2f bb = items[0]->getBB();
   for (card32 i=1; i<n; i++) {
      bb.addBB(items[i]->getBB());
   }
   float32 max = bb.getMaxSideLength();
   if (max > 1E-5) {
      scale = 2.0f/max;
      translation = - bb.getCenter() * scale;
   }
   Matrix3f vp = computeVPTransform();
   setMatrices(ctm, vp);
}

void Canvas2d::repaintCanvas(QPainter &painter) {
	painter.fillRect(QRect(0,0,width(), height()), QBrush(QColor(255,255,255,255)));

   for (unsigned i=0; i<items.size(); i++) {
      items[i]->draw(this, painter);
   }

   //Vector2f upperLeft = untransform(makeVector2i(0,0));
   //Vector2f lowerRight = untransform(makeVector2i(PaintBox1->Width, PaintBox1->Height));
   //TVarRec txt1[1] = {scale*100.0f};
   //TVarRec txt2[4] = {upperLeft[0], lowerRight[1], lowerRight[0], upperLeft[1]};
   //Label1->Caption = Format("%.3g", txt1, 0) +AnsiString("%")+ Format(" / (%.3g,%.3g) --> (%.3g,%.3g)", txt2, 3);
}

void Canvas2d::clear() {
   for (unsigned i=0; i<items.size(); i++) {
      delete items[i];
   }
   items.clear();
}
                                     
void Canvas2d::paintEvent(QPaintEvent *event) {
	QPainter painter(this);
	repaintCanvas(painter);
}

Matrix3f Canvas2d::computeVPTransform() {
   float32 w = width();
   float32 h = height();
   Matrix3f pre = makeMatrix3f(
      scale,     0, translation[0],
          0, scale, translation[1],
          0,     0,              1
   );
   return makeMatrix3f(
      h/2,    0, w/2,
        0, -h/2, h/2,
        0,    0,   1
   ) * pre;
}

void Canvas2d::setMatrices(const Matrix3f &newCTM, const Matrix3f &newVP) {
   ctm = newCTM;
   viewPortTransform = newVP;
   fullTransform = newVP * newCTM;
   fullTransformInv = invertMatrix(fullTransform);
}

void Canvas2d::resizeEvent(QResizeEvent *event) {
   Matrix3f vp = computeVPTransform();
   setMatrices(ctm, vp);
}


Canvas2d::~Canvas2d() {
   clear();
}





void Canvas2dPoint::draw(Canvas2d *dest, QPainter &destPainter) {
   const Vector2i S = makeVector2i(size, size);
   Vector2i li = dest->transform(pos);
   Vector2i ui = li;
   Vector2i p = ui;
   li -= S;
	destPainter.setPen(QPen(QBrush(color), 1));
   destPainter.drawRect(li[0], li[1], 2*size, 2*size);
   destPainter.drawPoint(p[0], p[1]);
}


void Canvas2dLine::draw(Canvas2d *dest, QPainter &destPainter) {
   Vector2i si = dest->transform(start);
   Vector2i ei = dest->transform(end);

	destPainter.setPen(QPen(QBrush(color),1));
   destPainter.drawLine(si[0], si[1], ei[0], ei[1]);
}


Canvas2dImage::Canvas2dImage(const QImage &bmp, bool autosize) {
   this->bmp = bmp;
   if (autosize && bmp.height() > 0) {
      float a = (float)bmp.width() / (float)bmp.height();
      lowerLeft = makeVector2f(-a, -1);
      upperRight = makeVector2f(a, 1);
   } else {
      lowerLeft = makeVector2f(-1, -1);
      upperRight = makeVector2f(1, 1);
   }
}

void Canvas2dImage::draw(Canvas2d *dest, QPainter &destPainter) {
   Vector2i li = dest->transform(lowerLeft);
   Vector2i ui = dest->transform(upperRight);

	QPixmap pixmap;
	pixmap = QPixmap::fromImage(bmp);	
	destPainter.drawPixmap(QRectF(li[0],li[1],ui[0]-li[0], ui[1]-li[1]), pixmap, QRectF(0,0,pixmap.width(), pixmap.height()) );
}

void Canvas2d:: mouseMoveEvent(QMouseEvent *event) {
   int X = event->x(); 
   int Y = event->y();
      
   int action = 0;

   if (event->buttons() & Qt::LeftButton) {
      action = 1;
   } else if (event->buttons() & Qt::RightButton) {
      action = 2;
   }


   if (action == 1) {
      float32 yDiff = (float32)(Y - lastY) / (float32)height();
      scale = exp(yDiff) * lastScale;
      translation = lastTranslation * exp(yDiff);
      Matrix3f vp = computeVPTransform();
      setMatrices(ctm, vp);
      repaint();
   } else if (action == 2) {
      float32 h = height();
      float32 xDiff = (X - lastX) / h;
      float32 yDiff = (Y - lastY) / h;
      lastX = X;
      lastY = Y;
      translation += makeVector2f(xDiff, -yDiff)*2;
      Matrix3f vp = computeVPTransform();
      setMatrices(ctm, vp);
      repaint();
   }
}

void Canvas2d::MouseDownEvent(QMouseEvent *event) {
   int X = event->x(); 
   int Y = event->y();
   lastVP = viewPortTransform;
   lastX = X;
   lastY = Y;
   lastScale = scale;
   lastTranslation = translation;
}

void Canvas2d::resetView() {
   scale = 1.0f;
   translation = NULL_VECTOR2F;
   Matrix3f vp = computeVPTransform();
   setMatrices(ctm, vp);
   repaint();
}

#include "Canvas2d_moc.h"
