//---------------------------------------------------------------------------
#ifndef Canvas2dH
#define Canvas2dH
//---------------------------------------------------------------------------
#include "x3rendering.h"
//---------------------------------------------------------------------------
#include "LinearAlgebra.h"
#include "BoundingBox.h"
//---------------------------------------------------------------------------
#include <QWidget>
#include <QtGui/QImage>
#include <vector>
//---------------------------------------------------------------------------

namespace X4 {

class Canvas2dElement;

class X3RENDERING_API Canvas2d : public QWidget {
	Q_OBJECT
 private:
   /// identity --> (-1,-1), (1,1) should std::map to window
   Matrix3f ctm;
   /// transform to window coords
   Matrix3f viewPortTransform;
   /// ctm and vpt multiplied
   Matrix3f fullTransform;
   /// inverse of full transform
   Matrix3f fullTransformInv;
   /// view port scale
   float32 scale;
   /// view port translation
   Vector2f translation;

   std::vector<Canvas2dElement*> items;

   // --- interaction state
   Matrix3f lastVP;
   int lastX;
   int lastY;
   float32 lastScale;
   Vector2f lastTranslation;

   void setMatrices(const Matrix3f &newCTM, const Matrix3f &newVP);
   Matrix3f computeVPTransform();

 protected:
	virtual void paintEvent(QPaintEvent *event);
	virtual void mouseMoveEvent(QMouseEvent *event);
	virtual void MouseDownEvent(QMouseEvent *event);
	virtual void resizeEvent(QResizeEvent *event);
	virtual void resetView();
	
 public:
   Canvas2d(QWidget *parent = 0, Qt::WindowFlags f = 0);

   Vector2i transform(const Vector2f &p);
   Vector2f untransform(const Vector2i &p);

   void addElement(Canvas2dElement *e);
   /// call to redraw window. must always be called manually (except from windows events such as repaint, resize etc...)
   void repaintCanvas(QPainter &painter);
   void autoSize();
   void clear();

   ~Canvas2d();
};



class X3RENDERING_API Canvas2dElement {
 public:
   virtual void draw(Canvas2d *dest, QPainter &destPainter) = 0;
   virtual BoundingBox2f getBB() = 0;
   virtual ~Canvas2dElement() {}
};


class X3RENDERING_API Canvas2dPoint : public Canvas2dElement {
 public:
   Vector2f pos;
   float32 size;
   QColor color;

   Canvas2dPoint() {pos = NULL_VECTOR2F; size = 3; color = QColor(200,0,0);}
   virtual void draw(Canvas2d *dest, QPainter &destPainter);
   virtual BoundingBox2f getBB() {return BoundingBox2f(pos);}
};


class X3RENDERING_API Canvas2dLine : public Canvas2dElement {
 public:
   Vector2f start;
   Vector2f end;
   QColor color;

   Canvas2dLine() {start = NULL_VECTOR2F; end = NULL_VECTOR2F; color = QColor(0,0,0);}
   virtual void draw(Canvas2d *dest, QPainter &destPainter);
   virtual BoundingBox2f getBB() {BoundingBox2f bb(start); bb.addPoint(end); return bb;}
};


class X3RENDERING_API Canvas2dImage : public Canvas2dElement {
 public:
   Vector2f lowerLeft;
   Vector2f upperRight;
   QImage bmp;

   Canvas2dImage(const QImage &bmp, bool autosize = true);

   virtual void draw(Canvas2d *dest, QPainter &destPainter);
   virtual BoundingBox2f getBB() {BoundingBox2f bb(lowerLeft); bb.addPoint(upperRight); return bb;}
};

}


#endif
