#ifndef DebugRendererH
#define DebugRendererH
//---------------------------------------------------------------------------
#include "x3rendering.h"
//---------------------------------------------------------------------------
//#include "windows.h"
#include "gl.h"
#include "UnstructuredInCoreTriangleMesh.h"
#include "Plane3f.h"
#include "Polygon.h"
#include "SparseLinearAlgebra.inline.h"
#include <map>
//---------------------------------------------------------------------------

namespace X4
{

class GLContext;
class Scene;
class RenderingTool;
class GLMaterial;
class SceneObject;
class SGListNode;
class RGBAFloat32Image;

/**
	this class provides visualization tools for debugging
	(doesn't work for multiple windows, there's only one instance)
	Usage Example:
	
	std::string myid = "what the ... is happening in my code?";
	debugRenderer->clearDebugData( myid );	// clear data from last call
	debugRenderer->beginRenderJob( myid );	// specify job name 
	
	// add some lines
		debugRenderer->addLine( p1,p2 );
		debugRenderer->addLine( p3,p4 );

	debugRenderer->endRenderJob();
*/
class X3RENDERING_API DebugRenderer 
{
public:
    static unsigned DR_FRAME;

private:

	struct TRJLine
	{
		float32	linewidth;
		Vector3f p1,p2;
		Vector4f c1,c2;
	};
	struct TRJLine2D
	{
		float32	linewidth;
		Vector2f p1,p2;
		Vector3f c1,c2;
	};

	struct TRJSphere
	{
		Vector3f pos;
		float radius;
		Vector3f color;
	};

	struct TRJPolygon
	{
		std::vector<Vector3f> points;
		Vector3f color;
		Vector3f normal;
	};

	struct TRJEllipsoid
	{
		Vector3f pos;
		Matrix3f orientation;
		Vector3f scale;
		Vector3f color;
	};

	// display 2d normal distribution as textured rectangle
	struct TNormalDist2D
	{
		Vector3f color;
		Vector3f x[4];
	};

   struct TSceneObject
   {
      Scene* scene;
      SceneObject* object;
      GLMaterial* material;
   };
   
   struct TTextObject
   {
      std::string text;
      int32 size;
      std::string fontname;
	  Vector3f color;
      std::map<GLContext*, GLuint> gllist;
      std::map<GLContext*, bool> dirty;
      Vector3f pos;
	  bool is2D;

      TTextObject () : size(-24), is2D(false) {}
      void initFont(GLContext* context);
      void draw(GLContext* context);
   };

   struct TSimpleImage
   {
	   RGBAFloat32Image * img;
	   Vector3f pos;
	   Vector3f xDir,yDir;
   };


	struct TDebugRenderJob
	{
		~TDebugRenderJob() {delete mesh;delete surfelCloud;}
		std::string id;
		bool active;
		int time;
		int time_range;
		std::vector<TRJLine> lines;
		std::vector<TRJSphere> spheres;
		std::vector<Vector3f> points;
		std::vector<Vector3f> points_2D;
		std::vector<TRJLine2D> lines2D;
		std::vector<TRJPolygon> polygons;
		std::vector<TRJEllipsoid> ellipsoids;
		UnstructuredInCoreTriangleMesh * mesh;
		UnstructuredInCorePointCloud * surfelCloud;
		Matrix4f objectMatrix;
		bool smoothLines;
		bool enableBlending;
		bool enableDepthTest;
		float32 pointSize;
		std::vector<TNormalDist2D> glowingNormalDistributions;
		std::vector<TRJSphere> glowingSplats;
		std::vector<TSimpleImage> simpleImages;
		
		std::vector<TSceneObject> object;

      std::vector<TTextObject> texts;
	};

	std::vector<std::string>	m_RenderJobStack;
	std::string	m_CurrentRenderJob;
	TDebugRenderJob * m_CurrentRenderJobPtr;
	std::vector<TDebugRenderJob*>	m_RenderJobs;

	void drawJob(GLContext * glContext, Scene * scene, TDebugRenderJob* job, bool drawSceneObjects = true);
	std::map< GLContext*, RenderingTool *> m_RTools;

	void enableBrightLight();

public:
	DebugRenderer(void);
	~DebugRenderer(void);

	void disconnect( GLContext * context );


  
	/// deletes render job with id (if exists)
	/// when id is not specified, all render jobs are deleted
	void clearDebugData( const std::string& id = std::string(""));

	static const int DRAW_EVERYTIME = -10000;
	
	/// specify id of current render job, is active is true then the job is automatically rendered
	/// if time is specified, then the render job is only rendered for the given time
	/// all frames within [time,time+time_range) will show this job
	void beginRenderJob( std::string id, bool active = true, int time = DRAW_EVERYTIME, int time_range = 1 );
	void endRenderJob();

	/// convenience: creates a renderjob name with id + intToStr(time)
	void beginRenderJob_OneFrame( std::string id, int time, bool clear = true );

	/// convenience: creates a renderjob for a specific time range [time,time+time_range)
	/// name will be id + intToStr(time) + "_" + intToStr(time+timeRange)
	void beginRenderJob_MultiFrame( std::string id, int time, int time_range, bool clear = true );

	/// convenience: clears all renderjobs: "name" + all id s
	void clearRenderJob_AllFrames( std::string id );

	/// manual job rendering
	void renderJob( GLContext * glContext, Scene * scene, std::string id);

	/// manual rendering special case for high quality renderer
	void renderJobs( GLContext * glContext, Scene * scene, GLShaderMaterial * material );

	mpint getIndexForRenderJob( std::string id );

	/// add meshes of all renderjobs to scene
	void instanciateAllMeshes(Scene * scene, bool timeInterval=false, int startTime=-1000000, int endTime=1000000);

	/// instantiate mesh of a single render job
	void instanciateMesh(Scene * scene, std::string renderid, card32 materialID = 0, SGListNode * rootNode=NULL);
	
	/// instantiate all meshes of current frame (adds everything to rootNode if specified)
	void instantiateAllMeshesInFrame(Scene * scene, int frame, card32 materialID = 0, SGListNode * rootNode=NULL);

	void addTransformationMatrix(const Matrix4f& mat);
	void setTransformationMatrix(const Matrix4f& mat);

	void setSmoothLine(bool s);
	void setEnableBlending(bool b);
	void setEnableDepthTest(bool b);
	void setPointSize(float32 s);

	/// render scene object (this will automatically render through the material of the object)
	/// if no material specified then the attached to the given object will be used
	void addSceneObject(Scene* scene, SceneObject* obj, GLMaterial* material = NULL);
   
	void addPoint( const Vector3f &point, const Vector3f &color);
	void addSurfel( const Vector3f &point, const Vector3f &normal, const Vector3f &color);

	void addLine( const Vector3f &point1, const Vector3f &point2, bool cylinder3D=false, float32 cylinderDiameter = 0.01f);
	void addLine( const Vector3f &point1, const Vector3f &point2, const Vector3f &color1, const Vector3f &color2, float linewidth, bool cylinder3D=false, float32 cylinderDiameter = 0.01f);
	void addLine( const Vector3f &point1, const Vector3f &point2, const Vector4f &color1, const Vector4f &color2, float linewidth, bool cylinder3D=false, float32 cylinderDiameter = 0.01f);
	void addLineStrip(const std::vector<Vector3f> &points, const Vector3f &color, float linewidth=2.0f);
	void addLineArray(const std::vector<Vector3f> &points, const Vector3f &color, float linewidth=2.0f);
	void addPlanarEllipse( const Vector3f &center, const Vector3f &dir1, const Vector3f &dir2, const Vector3f &color, float linewidth);
	void addSphereWiremesh( const Vector3f &point, const float radius, const float linewidth, const Vector3f &color);
	/// add array of lines -> addLine(p0 + gen*i,p1 + gen*i,...) with i in [0,numLines)
	void addLinePattern( const Vector3f &line_p0, const Vector3f &line_p1,const Vector3f &generator, mpcard numLines, const Vector3f &color, float lineWidth);
	void addLineAAGrid2D( const Vector3f &origin, float gridSize, mpcard dimX, mpcard dimY, const Vector3f &color, float lineWidth);

	void addSphere( const Vector3f &point, const float radius, const Vector3f &color);
	void addCoordSystem( const Vector3f &pos, const Matrix3f &mat, float size );
	void addCoordSystem( const Matrix4f &mat, float size );
	void addBoundingBox( const BoundingBox3f &bb, const Vector3f &color, const float32 &linewidth=2);
	void addBoundingBox( const Matrix4f &T, const BoundingBox3f &bb, const Vector3f &color, const float32 &linewidth=2);
	void addPlane( const Plane3f &p, float size, const Vector3f &color);

	/// adds 2d point, viewport is (0,0) to (1,1)
	void add2DPoint( const Vector2f &point, const Vector3f &color);
	void add2DLine(const Vector2f &p0, const Vector2f &p1, const Vector3f &color, float lineWidth=2);
	void add2DLine(const Vector2f &p0, const Vector2f &p1, const Vector3f &color0, const Vector3f &color1, float lineWidth=2);
	void add2DCircle(const Vector2f &center, float radius, const Vector3f &color, float lineWidth=2);
	void add2DText( const Vector2f& pos, const std::string& str, card32 size=24, const std::string& font = "Verdana", const Vector3f &color = makeVector3f(1,1,1));

	/// add 2d visualization of a sparse matrix (non-zero entries are represented as points)
	/// columns in x-dim, rows in y-dim, coordinates are scaled by 'scale' and translated by 'origin'
	template<class T>
	void addSparseMatrix(const SparseMatrix<T> &A, mpcard numColumns, const Vector3f &origin, float scale, const Vector3f &color=makeVector3f(1,1,1), bool transpose=false);

	void addPatch( const Vector3f &point, const Matrix3f &paramToWorld, const StaticVector<float,6> &coeffs, int numRSegments, int numCSegments, const Vector3f &color, 
	   float parametricScale, float patchSize);

	void addPolygon( const Vector3f * points, card32 numPoints, const Vector3f &color, bool cylinder3D=false, float32 cylinderDiameter = 0.01f );
	void addPolygon( const std::vector<Vector3f> &points, const Vector3f &color, bool cylinder3D=false, float32 cylinderDiameter = 0.01f );
	void addPolygon( const SimplePolygon &poly, const Vector3f &color, bool cylinder3D=false, float32 cylinderDiameter = 0.01f);
	void addPolygonOutline( const SimplePolygon &poly, const Vector3f &color, float lineWidth);
	void addQuad( const Vector3f &p1, const Vector3f &p2, const Vector3f &p3, const Vector3f &p4, const Vector3f &color, bool cylinder3D=false, float32 cylinderDiameter = 0.01f);
	void addPolygonColored( const Vector3f * points, const Vector3f * colors, card32 numPoints, bool cylinder3D = false, float32 cylinderDiameter=0.01f );
	void addTriangle( const Vector3f &p1, const Vector3f &p2, const Vector3f &p3, const Vector3f &color, bool cylinder3D=false, float32 cylinderDiameter = 0.01f);
	void addTriangleColored(const Vector3f &p1, const Vector3f &p2, const Vector3f &p3, const Vector3f &color0, const Vector3f &color1, const Vector3f &color2);
	void addBox(const Vector3f &pos, const Matrix3f &orientation, const Vector3f &scale, const Vector3f &color, bool cylinder3D=false, float32 cylinderDiameter = 0.01f);
	void addCenteredBox(const Vector3f &center, const Matrix3f &orientation, const Vector3f &scale, const Vector3f &color, bool cylinder3D=false, float32 cylinderDiameter=0.01f);
	void addEllipsoid( const Vector3f &pos, const Matrix3f &orientation, const Vector3f &scale, const Vector3f &color );
	void addPositionConstraint( const Vector3f &refPoint, const Vector3f &center, const Matrix3f &covariance );

	/// add camera outlines, fov specified in degree (not in rad!)
	void addCamera( const Vector3f &camPos, const Vector3f &lookAt, const Vector3f &up, float fovX, float fovY, float scale, const Vector3f &color, float lineWidth=2);

	/// add text to be printed on screen
	void addText( const Vector3f& pos, const std::string& str, card32 size=24, const std::string& font = "Verdana", const Vector3f &color = makeVector3f(1,1,1));

	/// uses texture mapping to visualize a normal distribution (blend mode: GL_ONE,GL_ONE)
	void addGlowingNormalDitribution2D( const Vector3f &pos, const Vector3f &dir1, const Vector3f &dir2, const Vector3f &color);

	/// round splat, blended with (GL_ONE,GL_ONE)
	void addGlowingSplat( const Vector3f &point, const float radius, const Vector3f &color);

	/// add point sampled normal distribution (height field z direction)
	void add2DNormalDistributionAsHeightField(const Vector3f &center, const Matrix2f &cov, float supportRadius, card32 numSamplesPerDim, const Vector3f &color);
	void add2DNormalDistributionAsHeightField(const Matrix2f &cov, const Vector2f &b, float c, float supportRadius, card32 numSamplesPerDim, const Vector3f &color);

	/// add simple image
	void addImage( RGBAFloat32Image * img, const Vector3f &pos, const Vector3f &xDir, const Vector3f &yDir);

	/// creates triangles (much faster when rendering)
	void addFastSphere(const Vector3f &pos, float radius, const Vector3f &color, bool highRes = true);
	void addFastEllipse(const Vector3f &pos, const Matrix3f &orientation, const Vector3f &scale, const Vector3f &color, bool highRes = true);
	void addFastCylinder(const Vector3f &pos, const Vector3f &orientation, float length, float radius, const Vector3f &color, bool caps = true, bool highRes = true);
	void addFastCylinder(const Vector3f &pos, const Vector3f &pos2, float radius, const Vector3f &color, bool caps = true, bool highRes = true);
	void addFastCone(const Vector3f &head, const Vector3f &direction, float length, float radius, const Vector3f &color, bool highRes = true);
	void addFineBoundingBox( const Matrix4f &frame, const BoundingBox3f &bb, const Vector3f &color, bool highRes = true );
	void addFineCoordinateFrame( const Vector3f &pos, const Matrix3f &frame, float size, const Vector3f &color, bool highRes = true);
	void addFineCoordinateFrame( const Matrix4f &frame, float size, const Vector3f &color, bool highRes = true);
	void addFineArrow(const Vector3f &start, const Vector3f &end, const Vector3f &color, float radius, bool highRes = true);
	
	/// plane is visible only inside the bb
	//void addPlane( const Plane3f &plane, const BoundingBox3f &bb);

	void draw( GLContext * glContext, Scene * scene, bool drawSceneObjects = true );

	card32 getNumRenderJobs() const {return (card32)m_RenderJobs.size();}
	std::string getRenderJobName(card32 index) {return m_RenderJobs[index]->id;}
	bool getRenderJobActiveFlag(card32 index) {return m_RenderJobs[index]->active;}
	void setRenderJobActiveFlag(card32 index, bool b) {m_RenderJobs[index]->active = b;}

};

template<class T>
void X4::DebugRenderer::addSparseMatrix(const SparseMatrix<T> &M, mpcard numColumns, const Vector3f &start, float scale, const Vector3f &color, bool transpose)
{
	for( mpcard r=0;r<M.getNumRows();r++ )
	{
		const SparseVector<T> & row = M[r];
		for( mpcard i=0;i<row.getNumNonZeroEntries();i++ )
		{
			mpcard c = row.getListEntryIndex(i);
			Vector3f p = makeVector3f(c,r,0);
			if( transpose )
				swap(p[0],p[1]);
			p = start + p * scale;
			debugRenderer->addPoint(p,color);
		}
	}

	const Vector3f outlineColor = color * 0.5f;
	Vector3f p0 = start+makeVector3f(-0.5f,-0.5f,0) * scale;
	Vector3f p1 = makeVector3f(0.5f+numColumns,0.5f+M.getNumRows(),0) * scale;
	if( transpose )
		swap(p1[0],p1[1]);
	p1 += start;

	debugRenderer->addLine(p0,makeVector3f(p0[0],p1[1],0),outlineColor,outlineColor,1);
	debugRenderer->addLine(p0,makeVector3f(p1[0],p0[1],0),outlineColor,outlineColor,1);
	debugRenderer->addLine(p1,makeVector3f(p0[0],p1[1],0),outlineColor,outlineColor,1);
	debugRenderer->addLine(p1,makeVector3f(p1[0],p0[1],0),outlineColor,outlineColor,1);
}


extern X3RENDERING_API DebugRenderer * debugRenderer;

}

#endif