#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "ECPhotonTracer.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "ECRayCaster.h"
#include "UnstructuredInCorePointCloud.h"
#include "PointCloudTools.h"
#include "Random.h"
#include "DebugRenderer.h"
#include "VertexArray.h"
#include "StandardSceneGraphIterator.h"
#include "SceneGraphNode.h"
#include "SGObjectNode.h"
#include "SGTransformationNode.h"
#include "SceneObject.h"
//---------------------------------------------------------------------------
using namespace X4;

IMPLEMENT_X4_CLASS( ECPhotonTracer ,0)
{
	BEGIN_CLASS_INIT( ECPhotonTracer );
}

ECPhotonTracer::ECPhotonTracer(void)
{
	m_UPC = NULL;
}

ECPhotonTracer::~ECPhotonTracer(void)
{
}

void ECPhotonTracer::assign(const Object* obj, X4CopyContext *context)
{
	const ECPhotonTracer * o = dynamic_cast<const ECPhotonTracer*>(obj);
	x4Assert( o != NULL );

	copyObjectProperties(obj,this);
}

void X4::ECPhotonTracer::setup( UnstructuredInCorePointCloud * upc )
{
	m_UPC = upc;
}


std::vector<ECPhoton> X4::ECPhotonTracer::compute()
{
	if( m_UPC == NULL )
		throw X4Exception("ECPhotonTracer::compute() - setup before calling compute()!");

	const unsigned maxNumBounces = 4;
//	Sphere3f sceneSphere = getPCBBox(m_UPC).getBoundingSphere();
	ECRayCaster rc;
	rc.setup(m_UPC);
	float surfelSize = getMedianPointDistance(m_UPC)*2.5f;

	// trace samples
	std::vector<ECPhoton> tracedSamples;	tracedSamples.reserve(m_InitialPhotons.size());
	PointSet * ps = m_UPC->getPointSet();
	AAT colorAAT = m_UPC->getAAT("color");
	const float safetyBounceDistance = surfelSize * 0.25f;	// begin raycasting off surface
	for( unsigned i=0;i<m_InitialPhotons.size();i++ )
	{
		ECPhoton photon = m_InitialPhotons[i];
		Ray3f r;
		r.direction = diceBounceDirection(photon.m_SurfaceDir);
		r.origin = photon.m_Position + r.direction * safetyBounceDistance;
		//debugRenderer->addLine( r.origin,r.origin+r.direction*safetyBounceDistance );
		ECPhoton bouncedPhoton = photon;
		bouncedPhoton.m_Color *= 1.0f;
		for( unsigned b = 0;b<maxNumBounces;b++ )
		{
			if( !rc.castRay(r,surfelSize) )
				break;	// photon bounced out of scene

			mpcard pindex=rc.getLastHitPointIndex();
			bouncedPhoton.m_Color = bouncedPhoton.m_Color.componentProduct(ps->get3f(pindex,colorAAT))*max(0.0f,(-normalize(r.direction)*rc.getLastHitNormal()));
			float energy = norm(bouncedPhoton.m_Color);	// TODO: we need the correct probability here!!
			if( energy<1e-7f)
				break;
			if(rnd01() > energy * 0.5f && b+1 < maxNumBounces)
			{
				// bounce again
				r.direction = diceBounceDirection(rc.getLastHitNormal());
				r.origin = rc.getLastHitPosition() + r.direction * safetyBounceDistance;
			}
			else
			{
				bouncedPhoton.m_Position = rc.getLastHitPosition();
				bouncedPhoton.m_SurfaceDir = rc.getLastHitNormal();
				//bouncedPhoton.m_Color = makeVector3f(b==0?1:0,b==1?1:0,b>=2?1:0);
				tracedSamples.push_back(bouncedPhoton);
			}

		}
	}
	
	tracedSamples.insert(tracedSamples.end(),m_InitialPhotons.begin(), m_InitialPhotons.end());

	return tracedSamples;
}

X4::Vector3f X4::ECPhotonTracer::diceBounceDirection( const Vector3f &surfaceNormal ) const
{
	float phi = rnd01()*M_PI*2.0f;
	float x = rnd01() * 1.999999998f-0.999999999f;
	float z = sqrt(1 - x*x);
	Matrix3f tangentFrame = calcTangentSystem(surfaceNormal);
	return tangentFrame.transpose() * makeVector3f(x*sin(phi),x*cos(phi),z);
}

void X4::ECPhotonTracer::addWhiteHemisphere( const Vector3f &upVector, mpcard numSamples )
{
	if( m_UPC == NULL )
	{
		throw X4Exception("ECPhotonTracer::addWhiteHemisphere() - setup before calling addWhiteHemisphere()!");
	}

	Sphere3f sphere = getPCBBox(m_UPC).getBoundingSphere();
	for( unsigned i=0;i<numSamples;i++ )
	{
		ECPhoton p;
		p.m_Position = getUniformDistributedRandomDirection(upVector);
		p.m_SurfaceDir = -p.m_Position;
		p.m_Position = p.m_Position*sphere.radius*2.0f + sphere.center;
		p.m_Color = makeVector3f(1,1,1);
		m_InitialPhotons.push_back(p);
	}
}

X4::Vector3f X4::ECPhotonTracer::getUniformDistributedRandomDirection( const Vector3f &surfaceNormal ) const
{
	float phi = rnd01()*M_PI*2.0f;
	float theta = acos(rnd01());
	Matrix3f tangentFrame = calcTangentSystem(surfaceNormal);
	return tangentFrame.transpose() * makeVector3f(sin(theta)*sin(phi),sin(theta)*cos(phi),cos(theta));
}

struct ECPhotonTracer_StdVertex
{
	Vector3f pos;
	Vector3f normal;
	Vector3f color;
};

void gatherSelectedPoints(StandardSceneGraphIterator * sceneIter, const Matrix4f &modelMatrix, std::vector<ECPhotonTracer_StdVertex> &vertices,
						  float sampleSize)
{
	if (!sceneIter->getNode()->visible()) return;

	Matrix4f currentModelMatrix = modelMatrix;
	SGTransformationNode * tnode = dynamic_cast<SGTransformationNode*>(sceneIter->getNode());
	if( tnode )
	{
		currentModelMatrix = currentModelMatrix * tnode->getTransformation();
	}

	SGObjectNode *objNode = dynamic_cast<SGObjectNode*>(sceneIter->getNode());
	if (objNode != NULL) {
		SceneObject* obj = objNode->getSceneObject();

		if (obj != NULL && objNode->visible()) 
		{
			UnstructuredInCorePointCloud * upc = dynamic_cast<UnstructuredInCorePointCloud*>(obj);
			UnstructuredInCoreTriangleMesh * mesh = dynamic_cast<UnstructuredInCoreTriangleMesh*>(obj);
			Matrix4f tmpM = currentModelMatrix * obj->getTransformation();
			if( upc && !mesh)
			{
				AAT posAAT = upc->getAAT("position");
				AAT flagsAAT = upc->getAAT("flags");
				AAT normalAAT = upc->getAAT("normal");
				bool providesColor = upc->providesAttribute("color");
				AAT colorAAT;
				if( providesColor )
					colorAAT = upc->getAAT("color");
				Matrix3f tmpM3 = shrink4To3(tmpM);
				ECPhotonTracer_StdVertex v;
				v.color = makeVector3f(1,1,1);
				for( unsigned i=0;i<upc->getNumPoints();i++ )
				{
					if( (upc->getPointSet()->get1i(i,flagsAAT) & PF_FLAG_SELECTED) == 0 )
						continue;

					v.pos = transformVector3f(tmpM, upc->getPointSet()->get3f(i,posAAT));
					if( providesColor )
						v.color = upc->getPointSet()->get3f(i,colorAAT);
					v.normal = tmpM3 * upc->getPointSet()->get3f(i,normalAAT);
					vertices.push_back(v);
				}
			}
			if( mesh )
			{
				AAT posAAT = mesh->getAAT("position");
				AAT flagsAAT = mesh->getAAT("flags");
				//AAT normalAAT = mesh->getAAT("normal");
				bool providesColor = mesh->providesAttribute("color");
				AAT colorAAT;
				if( providesColor )
					colorAAT = mesh->getAAT("color");
				AAT triAAT = mesh->getTriangles()->getAAT("index");
				for( unsigned i=0;i<mesh->getNumTriangles();i++ )
				{
					Vector3i tri = mesh->getTriangles()->get3i(i,triAAT);
					if( !((mesh->getPointSet()->get1i(tri[0],flagsAAT) & PF_FLAG_SELECTED) && 
						(mesh->getPointSet()->get1i(tri[1],flagsAAT) & PF_FLAG_SELECTED) &&
						(mesh->getPointSet()->get1i(tri[2],flagsAAT) & PF_FLAG_SELECTED))	)
						continue;	// not all vertices seleceted

					Vector3f p0 = transformVector3f(tmpM,mesh->getPointSet()->get3f(tri[0],posAAT));
					Vector3f p1 = transformVector3f(tmpM,mesh->getPointSet()->get3f(tri[1],posAAT));
					Vector3f p2 = transformVector3f(tmpM,mesh->getPointSet()->get3f(tri[2],posAAT));
					Vector3f col = makeVector3f(1,1,1);
					if( providesColor )
					{
						col = (mesh->getPointSet()->get3f(tri[0],colorAAT)+
							mesh->getPointSet()->get3f(tri[1],colorAAT)+
							mesh->getPointSet()->get3f(tri[2],colorAAT)/3.0f);
					}

					float area = norm((p1-p0).crossProduct(p2-p0)) * 0.5f;
					unsigned numSamples = area / sqrt(sampleSize);
					ECPhotonTracer_StdVertex vertex;
					vertex.color = col;
					vertex.normal;
					for( unsigned j=0;j<numSamples;j++ )
					{
						 float u = rnd01();
						 float v = rnd01();
						 if (u+v > 1) {
							float nu = 1.0f - v;
							v = 1.0f - u;
							u = nu;
						 }

						 float w = 1.0f - u - v;
						 vertex.pos = p0*u + p1*v + p2*w;
						 vertices.push_back(vertex);
					}
						
				}
			}
		}
	}

	unsigned numChildren = sceneIter->getNumChildren();
	for( unsigned i=0;i<numChildren;i++ )
	{
		sceneIter->goToChild(i);
		gatherSelectedPoints(sceneIter,currentModelMatrix, vertices, sampleSize);
		sceneIter->goToParent();
	}

}

void ECPhotonTracer::addSelectedPointsAsLightSource(StandardSceneGraphIterator *sceneIt, mpcard numSamples, float sampleSize) 
{
	std::vector<ECPhotonTracer_StdVertex> vertices;
	gatherSelectedPoints(sceneIt, IDENTITY4F, vertices, sampleSize);

	for( unsigned i=0;i<numSamples;i++ )
	{
		ECPhoton p;
		unsigned ri = random32bit((card32)vertices.size());
		p.m_Position = vertices[ri].pos;
		p.m_SurfaceDir = vertices[ri].normal;
		p.m_Color = vertices[ri].color;
		m_InitialPhotons.push_back(p);
	}
}
