#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "ECRayCaster.h"
//---------------------------------------------------------------------------
#include "UnstructuredInCorePointCloud.h"
#include "AttachedIndexedOctree.h"
#include "RaytracingMath.inline.h"
#include "DebugRenderer.h"
#include "DebugOutput.h"
#include <float.h>
//---------------------------------------------------------------------------
using namespace X4;


ECRayCaster::ECRayCaster(void)
{
	m_UPC = NULL;
	m_RayHitSurfel = false;
}

ECRayCaster::~ECRayCaster(void)
{
}

void X4::ECRayCaster::setup( UnstructuredInCorePointCloud* pc )
{
	m_UPC = pc;
	m_PosAAT = pc->getAAT("position");
	m_NormalAAT = pc->getAAT("normal");
	PointSet * ps = m_UPC->getPointSet();
	m_Octree = dynamic_cast<AttachedIndexedOctree*>(pc->getAttachments()->getData("SimpleOctree"));
	if(m_Octree==NULL){
		m_Octree=new AttachedIndexedOctree();
		m_Octree->setup("SimpleOctree", AttachedData::ADF_QUERY_DELETE_ON_UPDATE,  NULL); // attachedOctreeMemManager !!
		m_Octree->octreesetup(ps, 32);
		pc->getAttachments()->attachData(m_Octree);
	}
}

struct TECRayCaster_nodeEntry
{
	AIOctreeNode * node;
	float bbIntersectionT;
};

struct ECRayCasterTChildBBOrderEntry
{
	float t;
	unsigned i;
	bool operator < (const ECRayCasterTChildBBOrderEntry &o ) const {return t > o.t;}
};

enum RTME_CLASSIFICATION
{ MMM, MMP, MPM, MPP, PMM, PMP, PPM, PPP, POO, MOO, OPO, OMO, OOP, OOM,
OMM,OMP,OPM,OPP,MOM,MOP,POM,POP,MMO,MPO,PMO,PPO};

struct RTME_ray
{	
	//common variables
	float x, y, z;		// ray origin	
	float i, j, k;		// ray direction	
	float ii, ij, ik;	// inverses of direction components

	// ray slope
	RTME_CLASSIFICATION classification;
	float ibyj, jbyi, kbyj, jbyk, ibyk, kbyi; //slope
	float c_xy, c_xz, c_yx, c_yz, c_zx, c_zy;	
};

void make_ray(float x, float y, float z, float i, float j, float k, RTME_ray *r)
{
	//common variables
	r->x = x;
	r->y = y;
	r->z = z;
	r->i = i;
	r->j = j;
	r->k = k;

	r->ii = 1.0f/i;
	r->ij = 1.0f/j;
	r->ik = 1.0f/k;

	//ray slope
	r->ibyj = r->i * r->ij;
	r->jbyi = r->j * r->ii;
	r->jbyk = r->j * r->ik;
	r->kbyj = r->k * r->ij;
	r->ibyk = r->i * r->ik;
	r->kbyi = r->k * r->ii;
	r->c_xy = r->y - r->jbyi * r->x;
	r->c_xz = r->z - r->kbyi * r->x;
	r->c_yx = r->x - r->ibyj * r->y;
	r->c_yz = r->z - r->kbyj * r->y;
	r->c_zx = r->x - r->ibyk * r->z;
	r->c_zy = r->y - r->jbyk * r->z;	

	//ray slope classification
	if(i < 0)
	{
		if(j < 0)
		{
			if(k < 0)
			{
				r->classification = MMM;
			}
			else if(k > 0){
				r->classification = MMP;
			}
			else//(k >= 0)
			{
				r->classification = MMO;
			}
		}
		else//(j >= 0)
		{
			if(k < 0)
			{
				r->classification = MPM;
				if(j==0)
					r->classification = MOM;
			}
			else//(k >= 0)
			{
				if((j==0) && (k==0))
					r->classification = MOO;	
				else if(k==0)
					r->classification = MPO;
				else if(j==0)
					r->classification = MOP;
				else
					r->classification = MPP;
			}
		}
	}
	else//(i >= 0)
	{
		if(j < 0)
		{
			if(k < 0)
			{
				r->classification = PMM;
				if(i==0)
					r->classification = OMM;
			}
			else//(k >= 0)
			{				
				if((i==0) && (k==0))
					r->classification = OMO;
				else if(k==0)
					r->classification = PMO;
				else if(i==0)
					r->classification = OMP;
				else
					r->classification = PMP;
			}
		}
		else//(j >= 0)
		{
			if(k < 0)
			{
				if((i==0) && (j==0))
					r->classification = OOM;
				else if(i==0)
					r->classification = OPM;
				else if(j==0)
					r->classification = POM;
				else
					r->classification = PPM;
			}
			else//(k > 0)
			{
				if(i==0)
				{
					if(j==0)
						r->classification = OOP;
					else if(k==0)
						r->classification = OPO;
					else
						r->classification = OPP;
				}
				else
				{
					if((j==0) && (k==0))
						r->classification = POO;
					else if(j==0)
						r->classification = POP;
					else if(k==0)
						r->classification = PPO;
					else
						r->classification = PPP;
				}
			}			
		}
	}
}

bool slopeint_mul(RTME_ray *r, const BoundingBox3f &bb, float *t){

	switch (r->classification)
	{
	case MMM:
		{
			if ((r->x < bb.lowerCorner[0]) || (r->y < bb.lowerCorner[1]) || (r->z < bb.lowerCorner[2])
				|| (r->jbyi * bb.lowerCorner[0] - bb.upperCorner[1] + r->c_xy > 0)
				|| (r->ibyj * bb.lowerCorner[1] - bb.upperCorner[0] + r->c_yx > 0)
				|| (r->jbyk * bb.lowerCorner[2] - bb.upperCorner[1] + r->c_zy > 0)
				|| (r->kbyj * bb.lowerCorner[1] - bb.upperCorner[2] + r->c_yz > 0)
				|| (r->kbyi * bb.lowerCorner[0] - bb.upperCorner[2] + r->c_xz > 0)
				|| (r->ibyk * bb.lowerCorner[2] - bb.upperCorner[0] + r->c_zx > 0)
				)
				return false;

			// compute the intersection distance

			*t = (bb.upperCorner[0] - r->x) * r->ii;
			float t1 = (bb.upperCorner[1] - r->y) * r->ij;
			if(t1 > *t)
				*t = t1;
			float t2 = (bb.upperCorner[2] - r->z) * r->ik;
			if(t2 > *t)
				*t = t2;

			return true;
		}


	case MMP:
		{		
			if ((r->x < bb.lowerCorner[0]) || (r->y < bb.lowerCorner[1]) || (r->z > bb.upperCorner[2])
				|| (r->jbyi * bb.lowerCorner[0] - bb.upperCorner[1] + r->c_xy > 0)
				|| (r->ibyj * bb.lowerCorner[1] - bb.upperCorner[0] + r->c_yx > 0)
				|| (r->jbyk * bb.upperCorner[2] - bb.upperCorner[1] + r->c_zy > 0)
				|| (r->kbyj * bb.lowerCorner[1] - bb.lowerCorner[2] + r->c_yz < 0)
				|| (r->kbyi * bb.lowerCorner[0] - bb.lowerCorner[2] + r->c_xz < 0)
				|| (r->ibyk * bb.upperCorner[2] - bb.upperCorner[0] + r->c_zx > 0)
				)
				return false;

			*t = (bb.upperCorner[0] - r->x) * r->ii;
			float t1 = (bb.upperCorner[1] - r->y) * r->ij;
			if(t1 > *t)
				*t = t1;
			float t2 = (bb.lowerCorner[2] - r->z) * r->ik;
			if(t2 > *t)
				*t = t2;

			return true;
		}

	case MPM:
		{		
			if ((r->x < bb.lowerCorner[0]) || (r->y > bb.upperCorner[1]) || (r->z < bb.lowerCorner[2])
				|| (r->jbyi * bb.lowerCorner[0] - bb.lowerCorner[1] + r->c_xy < 0) 
				|| (r->ibyj * bb.upperCorner[1] - bb.upperCorner[0] + r->c_yx > 0)
				|| (r->jbyk * bb.lowerCorner[2] - bb.lowerCorner[1] + r->c_zy < 0) 
				|| (r->kbyj * bb.upperCorner[1] - bb.upperCorner[2] + r->c_yz > 0)
				|| (r->kbyi * bb.lowerCorner[0] - bb.upperCorner[2] + r->c_xz > 0)
				|| (r->ibyk * bb.lowerCorner[2] - bb.upperCorner[0] + r->c_zx > 0)
				)
				return false;

			*t = (bb.upperCorner[0] - r->x) * r->ii;
			float t1 = (bb.lowerCorner[1] - r->y) * r->ij;
			if(t1 > *t)
				*t = t1;
			float t2 = (bb.upperCorner[2] - r->z) * r->ik;
			if(t2 > *t)
				*t = t2;

			return true;
		}

	case MPP:
		{
			if ((r->x < bb.lowerCorner[0]) || (r->y > bb.upperCorner[1]) || (r->z > bb.upperCorner[2])
				|| (r->jbyi * bb.lowerCorner[0] - bb.lowerCorner[1] + r->c_xy < 0) 
				|| (r->ibyj * bb.upperCorner[1] - bb.upperCorner[0] + r->c_yx > 0)
				|| (r->jbyk * bb.upperCorner[2] - bb.lowerCorner[1] + r->c_zy < 0)
				|| (r->kbyj * bb.upperCorner[1] - bb.lowerCorner[2] + r->c_yz < 0) 
				|| (r->kbyi * bb.lowerCorner[0] - bb.lowerCorner[2] + r->c_xz < 0)
				|| (r->ibyk * bb.upperCorner[2] - bb.upperCorner[0] + r->c_zx > 0)
				)
				return false;

			*t = (bb.upperCorner[0] - r->x) * r->ii;
			float t1 = (bb.lowerCorner[1] - r->y) * r->ij;
			if(t1 > *t)
				*t = t1;
			float t2 = (bb.lowerCorner[2] - r->z) * r->ik;
			if(t2 > *t)
				*t = t2;

			return true;
		}

	case PMM:
		{
			if ((r->x > bb.upperCorner[0]) || (r->y < bb.lowerCorner[1]) || (r->z < bb.lowerCorner[2])
				|| (r->jbyi * bb.upperCorner[0] - bb.upperCorner[1] + r->c_xy > 0)
				|| (r->ibyj * bb.lowerCorner[1] - bb.lowerCorner[0] + r->c_yx < 0)
				|| (r->jbyk * bb.lowerCorner[2] - bb.upperCorner[1] + r->c_zy > 0)
				|| (r->kbyj * bb.lowerCorner[1] - bb.upperCorner[2] + r->c_yz > 0)
				|| (r->kbyi * bb.upperCorner[0] - bb.upperCorner[2] + r->c_xz > 0)
				|| (r->ibyk * bb.lowerCorner[2] - bb.lowerCorner[0] + r->c_zx < 0)
				)
				return false;

			*t = (bb.lowerCorner[0] - r->x) * r->ii;
			float t1 = (bb.upperCorner[1] - r->y) * r->ij;
			if(t1 > *t)
				*t = t1;
			float t2 = (bb.upperCorner[2] - r->z) * r->ik;
			if(t2 > *t)
				*t = t2;

			return true;
		}


	case PMP:
		{
			if ((r->x > bb.upperCorner[0]) || (r->y < bb.lowerCorner[1]) || (r->z > bb.upperCorner[2])
				|| (r->jbyi * bb.upperCorner[0] - bb.upperCorner[1] + r->c_xy > 0)
				|| (r->ibyj * bb.lowerCorner[1] - bb.lowerCorner[0] + r->c_yx < 0)
				|| (r->jbyk * bb.upperCorner[2] - bb.upperCorner[1] + r->c_zy > 0)
				|| (r->kbyj * bb.lowerCorner[1] - bb.lowerCorner[2] + r->c_yz < 0)
				|| (r->kbyi * bb.upperCorner[0] - bb.lowerCorner[2] + r->c_xz < 0)
				|| (r->ibyk * bb.upperCorner[2] - bb.lowerCorner[0] + r->c_zx < 0)
				)
				return false;

			*t = (bb.lowerCorner[0] - r->x) * r->ii;
			float t1 = (bb.upperCorner[1] - r->y) * r->ij;
			if(t1 > *t)
				*t = t1;
			float t2 = (bb.lowerCorner[2] - r->z) * r->ik;
			if(t2 > *t)
				*t = t2;

			return true;
		}

	case PPM:
		{
			if ((r->x > bb.upperCorner[0]) || (r->y > bb.upperCorner[1]) || (r->z < bb.lowerCorner[2])
				|| (r->jbyi * bb.upperCorner[0] - bb.lowerCorner[1] + r->c_xy < 0)
				|| (r->ibyj * bb.upperCorner[1] - bb.lowerCorner[0] + r->c_yx < 0)
				|| (r->jbyk * bb.lowerCorner[2] - bb.lowerCorner[1] + r->c_zy < 0) 
				|| (r->kbyj * bb.upperCorner[1] - bb.upperCorner[2] + r->c_yz > 0)
				|| (r->kbyi * bb.upperCorner[0] - bb.upperCorner[2] + r->c_xz > 0)
				|| (r->ibyk * bb.lowerCorner[2] - bb.lowerCorner[0] + r->c_zx < 0)
				)
				return false;

			*t = (bb.lowerCorner[0] - r->x) * r->ii;
			float t1 = (bb.lowerCorner[1] - r->y) * r->ij;
			if(t1 > *t)
				*t = t1;
			float t2 = (bb.upperCorner[2] - r->z) * r->ik;
			if(t2 > *t)
				*t = t2;

			return true;
		}

	case PPP:
		{
			if ((r->x > bb.upperCorner[0]) || (r->y > bb.upperCorner[1]) || (r->z > bb.upperCorner[2])
				|| (r->jbyi * bb.upperCorner[0] - bb.lowerCorner[1] + r->c_xy < 0)
				|| (r->ibyj * bb.upperCorner[1] - bb.lowerCorner[0] + r->c_yx < 0)
				|| (r->jbyk * bb.upperCorner[2] - bb.lowerCorner[1] + r->c_zy < 0)
				|| (r->kbyj * bb.upperCorner[1] - bb.lowerCorner[2] + r->c_yz < 0)
				|| (r->kbyi * bb.upperCorner[0] - bb.lowerCorner[2] + r->c_xz < 0)
				|| (r->ibyk * bb.upperCorner[2] - bb.lowerCorner[0] + r->c_zx < 0)
				)
				return false;

			*t = (bb.lowerCorner[0] - r->x) * r->ii;
			float t1 = (bb.lowerCorner[1] - r->y) * r->ij;
			if(t1 > *t)
				*t = t1;
			float t2 = (bb.lowerCorner[2] - r->z) * r->ik;
			if(t2 > *t)
				*t = t2;

			return true;
		}

	case OMM:
		{
			if((r->x < bb.lowerCorner[0]) || (r->x > bb.upperCorner[0])
				|| (r->y < bb.lowerCorner[1]) || (r->z < bb.lowerCorner[2])
				|| (r->jbyk * bb.lowerCorner[2] - bb.upperCorner[1] + r->c_zy > 0)
				|| (r->kbyj * bb.lowerCorner[1] - bb.upperCorner[2] + r->c_yz > 0)
				)
				return false;

			*t = (bb.upperCorner[1] - r->y) * r->ij;
			float t2 = (bb.upperCorner[2] - r->z) * r->ik;
			if(t2 > *t)
				*t = t2;

			return true;
		}

	case OMP:
		{
			if((r->x < bb.lowerCorner[0]) || (r->x > bb.upperCorner[0])
				|| (r->y < bb.lowerCorner[1]) || (r->z > bb.upperCorner[2])
				|| (r->jbyk * bb.upperCorner[2] - bb.upperCorner[1] + r->c_zy > 0)
				|| (r->kbyj * bb.lowerCorner[1] - bb.lowerCorner[2] + r->c_yz < 0)
				)
				return false;

			*t = (bb.upperCorner[1] - r->y) * r->ij;
			float t2 = (bb.lowerCorner[2] - r->z) * r->ik;
			if(t2 > *t)
				*t = t2;

			return true;
		}

	case OPM:
		{
			if((r->x < bb.lowerCorner[0]) || (r->x > bb.upperCorner[0])
				|| (r->y > bb.upperCorner[1]) || (r->z < bb.lowerCorner[2])
				|| (r->jbyk * bb.lowerCorner[2] - bb.lowerCorner[1] + r->c_zy < 0) 
				|| (r->kbyj * bb.upperCorner[1] - bb.upperCorner[2] + r->c_yz > 0)
				)
				return false;

			*t = (bb.lowerCorner[1] - r->y) * r->ij;		
			float t2 = (bb.upperCorner[2] - r->z) * r->ik;
			if(t2 > *t)
				*t = t2;

			return true;
		}

	case OPP:
		{
			if((r->x < bb.lowerCorner[0]) || (r->x > bb.upperCorner[0])
				|| (r->y > bb.upperCorner[1]) || (r->z > bb.upperCorner[2])
				|| (r->jbyk * bb.upperCorner[2] - bb.lowerCorner[1] + r->c_zy < 0)
				|| (r->kbyj * bb.upperCorner[1] - bb.lowerCorner[2] + r->c_yz < 0)
				)
				return false;

			*t = (bb.lowerCorner[1] - r->y) * r->ij;		
			float t2 = (bb.lowerCorner[2] - r->z) * r->ik;
			if(t2 > *t)
				*t = t2;

			return true;
		}


	case MOM:
		{
			if((r->y < bb.lowerCorner[1]) || (r->y > bb.upperCorner[1])
				|| (r->x < bb.lowerCorner[0]) || (r->z < bb.lowerCorner[2]) 
				|| (r->kbyi * bb.lowerCorner[0] - bb.upperCorner[2] + r->c_xz > 0)
				|| (r->ibyk * bb.lowerCorner[2] - bb.upperCorner[0] + r->c_zx > 0)
				)
				return false;

			*t = (bb.upperCorner[0] - r->x) * r->ii;
			float t2 = (bb.upperCorner[2] - r->z) * r->ik;
			if(t2 > *t)
				*t = t2;

			return true;
		}


	case MOP:
		{
			if((r->y < bb.lowerCorner[1]) || (r->y > bb.upperCorner[1])
				|| (r->x < bb.lowerCorner[0]) || (r->z > bb.upperCorner[2]) 
				|| (r->kbyi * bb.lowerCorner[0] - bb.lowerCorner[2] + r->c_xz < 0)
				|| (r->ibyk * bb.upperCorner[2] - bb.upperCorner[0] + r->c_zx > 0)
				)
				return false;

			*t = (bb.upperCorner[0] - r->x) * r->ii;
			float t2 = (bb.lowerCorner[2] - r->z) * r->ik;
			if(t2 > *t)
				*t = t2;

			return true;
		}

	case POM:
		{
			if((r->y < bb.lowerCorner[1]) || (r->y > bb.upperCorner[1])
				|| (r->x > bb.upperCorner[0]) || (r->z < bb.lowerCorner[2])
				|| (r->kbyi * bb.upperCorner[0] - bb.upperCorner[2] + r->c_xz > 0)
				|| (r->ibyk * bb.lowerCorner[2] - bb.lowerCorner[0] + r->c_zx < 0)
				)
				return false;

			*t = (bb.lowerCorner[0] - r->x) * r->ii;
			float t2 = (bb.upperCorner[2] - r->z) * r->ik;
			if(t2 > *t)
				*t = t2;

			return true;
		}


	case POP:
		{
			if((r->y < bb.lowerCorner[1]) || (r->y > bb.upperCorner[1])
				|| (r->x > bb.upperCorner[0]) || (r->z > bb.upperCorner[2])
				|| (r->kbyi * bb.upperCorner[0] - bb.lowerCorner[2] + r->c_xz < 0)
				|| (r->ibyk * bb.upperCorner[2] - bb.lowerCorner[0] + r->c_zx < 0)
				)
				return false;

			*t = (bb.lowerCorner[0] - r->x) * r->ii;
			float t2 = (bb.lowerCorner[2] - r->z) * r->ik;
			if(t2 > *t)
				*t = t2;

			return true;
		}	

	case MMO:
		{
			if((r->z < bb.lowerCorner[2]) || (r->z > bb.upperCorner[2])
				|| (r->x < bb.lowerCorner[0]) || (r->y < bb.lowerCorner[1])  
				|| (r->jbyi * bb.lowerCorner[0] - bb.upperCorner[1] + r->c_xy > 0)
				|| (r->ibyj * bb.lowerCorner[1] - bb.upperCorner[0] + r->c_yx > 0)
				)
				return false;

			*t = (bb.upperCorner[0] - r->x) * r->ii;
			float t1 = (bb.upperCorner[1] - r->y) * r->ij;
			if(t1 > *t)
				*t = t1;

			return true;
		}	

	case MPO:
		{
			if((r->z < bb.lowerCorner[2]) || (r->z > bb.upperCorner[2])
				|| (r->x < bb.lowerCorner[0]) || (r->y > bb.upperCorner[1]) 
				|| (r->jbyi * bb.lowerCorner[0] - bb.lowerCorner[1] + r->c_xy < 0) 
				|| (r->ibyj * bb.upperCorner[1] - bb.upperCorner[0] + r->c_yx > 0)
				)
				return false;

			*t = (bb.upperCorner[0] - r->x) * r->ii;
			float t1 = (bb.lowerCorner[1] - r->y) * r->ij;
			if(t1 > *t)
				*t = t1;

			return true;
		}


	case PMO:
		{
			if((r->z < bb.lowerCorner[2]) || (r->z > bb.upperCorner[2])
				|| (r->x > bb.upperCorner[0]) || (r->y < bb.lowerCorner[1]) 
				|| (r->jbyi * bb.upperCorner[0] - bb.upperCorner[1] + r->c_xy > 0)
				|| (r->ibyj * bb.lowerCorner[1] - bb.lowerCorner[0] + r->c_yx < 0) 
				)
				return false;

			*t = (bb.lowerCorner[0] - r->x) * r->ii;
			float t1 = (bb.upperCorner[1] - r->y) * r->ij;
			if(t1 > *t)
				*t = t1;

			return true;
		}

	case PPO:
		{
			if((r->z < bb.lowerCorner[2]) || (r->z > bb.upperCorner[2])
				|| (r->x > bb.upperCorner[0]) || (r->y > bb.upperCorner[1]) 
				|| (r->jbyi * bb.upperCorner[0] - bb.lowerCorner[1] + r->c_xy < 0)
				|| (r->ibyj * bb.upperCorner[1] - bb.lowerCorner[0] + r->c_yx < 0)
				)
				return false;

			*t = (bb.lowerCorner[0] - r->x) * r->ii;
			float t1 = (bb.lowerCorner[1] - r->y) * r->ij;
			if(t1 > *t)
				*t = t1;

			return true;
		}


	case MOO:
		{
			if((r->x < bb.lowerCorner[0])
				|| (r->y < bb.lowerCorner[1]) || (r->y > bb.upperCorner[1])
				|| (r->z < bb.lowerCorner[2]) || (r->z > bb.upperCorner[2])
				)
				return false;

			*t = (bb.upperCorner[0] - r->x) * r->ii;
			return true;
		}

	case POO:
		{
			if((r->x > bb.upperCorner[0])
				|| (r->y < bb.lowerCorner[1]) || (r->y > bb.upperCorner[1])
				|| (r->z < bb.lowerCorner[2]) || (r->z > bb.upperCorner[2])
				)
				return false;

			*t = (bb.lowerCorner[0] - r->x) * r->ii;
			return true;
		}

	case OMO:
		{
			if((r->y < bb.lowerCorner[1])
				|| (r->x < bb.lowerCorner[0]) || (r->x > bb.upperCorner[0])
				|| (r->z < bb.lowerCorner[2]) || (r->z > bb.upperCorner[2])
				)
				return false;

			*t = (bb.upperCorner[1] - r->y) * r->ij;
			return true;
		}

	case OPO:
		{
			if((r->y > bb.upperCorner[1])
				|| (r->x < bb.lowerCorner[0]) || (r->x > bb.upperCorner[0])
				|| (r->z < bb.lowerCorner[2]) || (r->z > bb.upperCorner[2])
				)
				return false;

			*t = (bb.lowerCorner[1] - r->y) * r->ij;
			return true;
		}


	case OOM:
		{
			if((r->z < bb.lowerCorner[2])
				|| (r->x < bb.lowerCorner[0]) || (r->x > bb.upperCorner[0])
				|| (r->y < bb.lowerCorner[1]) || (r->y > bb.upperCorner[1])
				)
				return false;

			*t = (bb.upperCorner[2] - r->z) * r->ik;
			return true;
		}

	case OOP:
		{
			if((r->z > bb.upperCorner[2])
				|| (r->x < bb.lowerCorner[0]) || (r->x > bb.upperCorner[0])
				|| (r->y < bb.lowerCorner[1]) || (r->y > bb.upperCorner[1])
				)
				return false;

			*t = (bb.lowerCorner[2] - r->z) * r->ik;
			return true;
		}	
	}

	return false;
}

#define CheckNAN(x) \
	if( _isnan(x) ) debugOutput << "nan in line " << numToStr(__LINE__) << "\n";

bool X4::ECRayCaster::castRay( const Ray3f &r, float surfelSize, float * t_ret_ptr /*= NULL*/ )
{
	if( m_UPC == NULL )
		throw X4Exception("ECRayCaster::castRay() - no point cloud found! call setup before using this method!");
	PointSet * ps = m_UPC->getPointSet();
	m_RayHitSurfel = false;

	// check root bb
	AIOctreeNode * rootNode = m_Octree->getOctree();
	BoundingBox3f rootBB = rootNode->getContent().m_bb;
	rootBB.addBorder(surfelSize);
	
	float firstHit, secondHit;
	if( !intersectBoundingBox3f(r,rootBB,firstHit, secondHit))
		return false;
	if( secondHit < 0.0f )
		return false;

	// precompute child ordering the lazy way :-)
	unsigned childOrder[8];
	std::vector<ECRayCasterTChildBBOrderEntry> entries(8);
	Ray3f tmpRay = r;
	float tmpMult = 1;
	while (true)
	{
		if( !rootBB.inBoundingBox(tmpRay.origin) )
			break;
		tmpRay.origin -= tmpRay.direction * tmpMult; tmpMult *= 2.0f;
	}
	for( unsigned i=0;i<8;i++ )
	{
		BoundingBox3f bb = rootBB;
		Vector3f bbCenter = bb.getCenter();
		for( unsigned j=0;j<3;j++ )
		{
			if( (i&(1 << j)) == 0 )
				bb.upperCorner[j] = bbCenter[j];
			else
				bb.lowerCorner[j] = bbCenter[j];
		}
		ECRayCasterTChildBBOrderEntry e;
		float secondHit;
		intersectBoundingBox3f(tmpRay, bb, e.t, secondHit);
		e.i = i;
		entries[i] = e;
	}
	sort(entries.begin(),entries.end());
	for( unsigned i=0;i<8;i++ )
		childOrder[i] = entries[i].i;

	// prepare rayslope test (Eisemann et al.)
	RTME_ray rtme_ray;
	make_ray(r.origin[0],r.origin[1],r.origin[2],r.direction[0],r.direction[1],r.direction[2],&rtme_ray);

	// traverse tree
	std::vector<TECRayCaster_nodeEntry> nodes; nodes.reserve(200);
	TECRayCaster_nodeEntry rootEntry;
	rootEntry.node = rootNode;
	rootEntry.bbIntersectionT = firstHit;
	nodes.push_back(rootEntry);
	float surfelSize_2 = surfelSize * surfelSize;
	float bestHitT = 1e29f;
	unsigned frameID=0;
	while (!nodes.empty())
	{
		TECRayCaster_nodeEntry node = nodes.back();
		nodes.pop_back();

//  		debugRenderer->beginRenderJob_OneFrame("narf",frameID++);
//  		debugRenderer->addBoundingBox(node.node->getContent().m_bb, makeVector3f(1,1,0),1);
//  		debugRenderer->endRenderJob();

		if( m_RayHitSurfel && node.bbIntersectionT > bestHitT )
			continue;

		if( node.node->isLeaf() )
		{
			IndexedNode & cont = node.node->getContent();
			const unsigned numChilds = (unsigned)cont.m_index.size();
			for( unsigned i=0;i<numChilds;i++ )
			{
				unsigned pIndex = cont.m_index[i];
				Vector3f p = ps->get3f(pIndex,m_PosAAT);
				Vector3f n = ps->get3f(pIndex,m_NormalAAT);
				
				// check if ray is orthogonal to n
				float n_dot_dir = n * r.direction;
				if( fabs(n_dot_dir) < 1e-7f)
					continue;

				// compute intersection
				float t = (p-r.origin) * n / n_dot_dir;
				if( t < 0.0f || t > bestHitT)
					continue;	// reject if not in front of origin or behind best existing hit

				// check if intersection point is inside surfel
				Vector3f hitPoint = r.origin + r.direction * t;
				if( normQuad(hitPoint - p) > surfelSize_2 )
					continue;

//				debugRenderer->addPoint(p,makeVector3f(1,0,0));

				// found a closer hit
				m_LastHittedSurfelIndex = pIndex;
				m_LastHittedSurfelPosition = hitPoint;
				m_LastHittedSurfelNormal = n;
				m_RayHitSurfel = true;
				bestHitT = t;
			}
		}
		else
		{
			// node case
			for( unsigned i=0;i<8;i++ )			
			{
				unsigned childIndex = childOrder[i];
				if( node.node->getChild(childIndex) )
				{
					TECRayCaster_nodeEntry e;
					e.node = node.node->getChild(childIndex);
					BoundingBox3f bb = e.node->getContent().m_bb;
					bb.addBorder(surfelSize);
					//if( test_rayIntersectsBoundingBox(r.origin, r.direction, bb, e.bbIntersectionT,secondHit ) )
					if( slopeint_mul( &rtme_ray, bb, &e.bbIntersectionT) )
					{
						nodes.push_back(e);
					}
				}
			}
		}
	}
	if( m_RayHitSurfel && t_ret_ptr)
		*t_ret_ptr = bestHitT;

	return m_RayHitSurfel;
}