//---------------------------------------------------------------------------
#include "stdafx.h"
 
//---------------------------------------------------------------------------
#include "GLAttachments.h"
//---------------------------------------------------------------------------
#include "GLContext.h"
#include "GLBuffer.h"
#include "GLTexture.h"
#include "DebugOutput.h"
#include "PointSet.h"
#include "VertexArray.h"
#include "UnstructuredInCoreTriangleMesh.h"
//---------------------------------------------------------------------------
 
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------


SharedVertexBufferAttachment::SharedVertexBufferAttachment()
:MemoryManagedAttachment()
{
   m_VertexBuffer = NULL;
}

SharedVertexBufferAttachment::~SharedVertexBufferAttachment()
{
   delete m_VertexBuffer;
}

card32 SharedVertexBufferAttachment::getManagedSize()
{
   if( m_VertexBuffer )
      return m_VertexBuffer->getSizeInBytes();
   else
   {
      x4Assert(false);
      return 0;
   }
}

void SharedVertexBufferAttachment::setupMMA( const std::string &name, MemoryManager * mm, bool deleteOnUpdate )
{
   MemoryManagedAttachment::setupMMA( name, mm, deleteOnUpdate );
}

void SharedVertexBufferAttachment::setupVB(GLVertexBuffer * vbuffer, const std::string &addCustomID,
	MemoryManager * mm, bool deleteOnUpdate, const int32 includeFlags, const int32 excludeFlags, 
   Vector3f lastDirection, Matrix4f lastCTM)
{
   m_VertexBuffer = vbuffer;
   this->lastDirection = lastDirection;
   this->lastCTM = lastCTM;
   setupMMA( generateID( vbuffer->getVertexDescriptor(), vbuffer->getRootContext(), addCustomID, includeFlags, excludeFlags ), mm, deleteOnUpdate );
}


std::string SharedVertexBufferAttachment::generateID( const VertexDescriptor * vdesc, GLContext * context,
      const std::string &addCustomID, const int32 includeFlags, const int32 excludeFlags )
{
   char tmpbuf[64];
   _i64toa( context->getContextNumber(), tmpbuf, 10 );
   char incl[64];
   itoa(includeFlags, incl, 10);
   char excl[64];
   itoa(excludeFlags, excl, 10);
   std::string result = vdesc->getDescription() + std::string("_C") + std::string( tmpbuf ) + addCustomID + std::string("_I") + std::string(incl) + std::string("_E") + std::string(excl);
   return result;
}

GLTextureAttachment::GLTextureAttachment()
   :MemoryManagedAttachment()
{
   m_Texture = NULL;
}

GLTextureAttachment::~GLTextureAttachment()
{
   delete m_Texture;
}

void GLTextureAttachment::setupTA( GLTexture * texture, const std::string &addCustomID,
   MemoryManager * memManager, bool deleteOnUpdate )
{
   m_Texture = texture;
   setupMMA( generateID( texture->getRootContext(), addCustomID ), memManager, deleteOnUpdate );
}


std::string GLTextureAttachment::generateID( GLContext * context, const std::string &addCustomID )
{
   char tmpbuf[64];
   for( card32 i=0;i<64;i++ )
      tmpbuf[i] = 0;
   itoa( (int)context, tmpbuf, 10 );
   tmpbuf[63] = 0;

   return std::string( tmpbuf ) + addCustomID;
}

card32 GLTextureAttachment::getManagedSize()
{
   if( m_Texture )
      return m_Texture->getSizeInBytes();
   else
   {
      x4Assert(false);
      return 0;
   }
}



GLTextureMapAttachment::GLTextureMapAttachment() :MemoryManagedAttachment()
{
}

GLTextureMapAttachment::~GLTextureMapAttachment()
{
	for (std::map<int,GLTexture*>::iterator it=m_Texture.begin(); it != m_Texture.end(); it++)
		delete it->second;
}

void GLTextureMapAttachment::setupTA( GLTexture * texture, int index, MemoryManager * memManager, bool deleteOnUpdate )
{
	GLTexture* found = getTexture(index);
	if (found) delete found;
	m_Texture[index] = texture;
	setupMMA( std::string("texturemap_") + intToStr(index) + std::string("_") + intToStr(texture->getRootContext()->getContextNumber()), memManager, deleteOnUpdate );
}

GLTexture* GLTextureMapAttachment::getTexture(int index)
{
	std::map<int,GLTexture*>::iterator it = m_Texture.find(index);
	if (it != m_Texture.end()) return it->second;
	return NULL;
}

card32 GLTextureMapAttachment::getManagedSize()
{
	card32 size = 0;
	for (std::map<int,GLTexture*>::iterator it=m_Texture.begin(); it != m_Texture.end(); it++)
		size += it->second->getSizeInBytes();
	return size;
}

card32 GLMultiTextureAttachment::getManagedSize()
{
	card32 sizeCounter = 0;
	for(int i = 0; i < (int)m_Textures.size(); i++)
	{
		sizeCounter += (m_Textures[i])->getSizeInBytes();
	}
	return sizeCounter;
}

// void GLMultiTextureAttachment::setupTA(const std::string &addCustomID, MemoryManager * memManager, bool deleteOnUpdate )
// {
// 	setupMMA( generateID( (m_Textures[0])->getRootContext(), addCustomID ), memManager, deleteOnUpdate );
// }

X4::GLIndexAttachment::GLIndexAttachment()
{
	m_Buffer = NULL;
	lastDirection = NULL_VECTOR3F;
	lastCTM = IDENTITY4F;
}

X4::GLIndexAttachment::~GLIndexAttachment()
{
	delete m_Buffer;
}

void X4::GLIndexAttachment::setupIndexBuffer(
		const std::string &addCustomID, GLContext * context, MemoryManager * mm,
		bool deleteOnUpdate, const int32 includeFlags, const int32 excludeFlags, 
		UnstructuredInCoreTriangleMesh * mesh, 
		Vector3f lastDirection, Matrix4f lastCTM  )
{
	this->lastDirection = lastDirection;
	this->lastCTM = lastCTM;
	m_Buffer = new GLIndexBuffer32(context);
	m_Buffer->init(mesh->getNumTriangles()*3);

	// extract triangles
	std::vector<int32> buffer(mesh->getNumTriangles()*3);
	VertexArray triangles(mesh->getTriangles());
	for( mpcard i=0;i<triangles.getNumElements();i++ )
	{
		Vector3i tri = triangles.getIndex3i(i);
		for( mpcard d=0;d<3;d++)
			buffer[i*3+d] = tri[d];
	}
	m_Buffer->writeToBuffer((card32*)&buffer[0],0,(GLuint)buffer.size());

	std::string id = generateID(addCustomID,context,mm,includeFlags,excludeFlags);
	setupMMA(id,mm,deleteOnUpdate);
}

std::string X4::GLIndexAttachment::generateID( const std::string &addCustomID, GLContext * context, MemoryManager * mm, const int32 includeFlags, const int32 excludeFlags )
{
	std::string ret="IDX";
	ret += intToStr(context->getContextNumber())+"#";
	ret += intToStr(includeFlags) + ";";
	ret += intToStr(excludeFlags) + ";";
	return ret;
}

X4::card32 X4::GLIndexAttachment::getManagedSize()
{
	return m_Buffer->getSizeInBytes();
}
