//---------------------------------------------------------------------------
#include "stdafx.h"
//---------------------------------------------------------------------------
#include "GLShaderMaterial.h"
//---------------------------------------------------------------------------
#include "DebugOutput.h"
#include "FixedArrayClassProperty.h"
#include "StringClassProperty.h"
#include "BooleanClassProperty.h"
#include "NumericalClassProperty.h"

#include "X4Assert.h"
#include "PointCloudIterators.h"
#include "PointCloud.h"
#include "PointSet.h"
#include "PointSet.inline.h"
#include "GLBuffer.h"
#include "GLContext.h"
#include "GLShader.h"
#include "LinearAlgebra.h"
#include "LinearAlgebra.inline.h"
#include "VertexDescriptor.h"
#include "Shader.h"
#include "MPContainer.h"
#include "GLAttachments.h"
#include "MaterialParameter.h"
#include "StringHelper.h"
#include "UnstructuredInCoreTriangleMesh.h"
#include "GLHelper.h"
#include "Configuration.h"
#include "SceneGraphState.h"
#include "Camera.h"
#include "GenericSwapAllOutOfCorePointCloud.h"
//---------------------------------------------------------------------------
#include <fstream>
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------


MetaClass* GLShaderMaterial::GLShaderMaterialMetaClass = NULL;

const MetaClass* GLShaderMaterial::getInstanceClass() const {
   return getClass();
}

Object* GLShaderMaterial::newInstance() {
   return new GLShaderMaterial;
}

const MetaClass* GLShaderMaterial::getClass() {
   if (GLShaderMaterialMetaClass==NULL) {
      GLShaderMaterialMetaClass = new MetaClass("GLShaderMaterial", "no description.",
                                  newInstance, sizeof(GLShaderMaterial));
   }
   return GLShaderMaterialMetaClass;
}

void GLShaderMaterial::init(const MetaClass *parentClass) {
   ((MetaClass*)getClass())->init(parentClass, MetaClass::FIRST_VERSION + 10);
   GLShaderMaterialMetaClass->copyParentProperties();
}

void GLShaderMaterial::shutdown() {
   delete GLShaderMaterialMetaClass;
}


// ----


GLShaderMaterial::GLShaderMaterial()
{
   setName( "unnamed shadermaterial" );

   m_VertexDescriptor = new VertexDescriptor;
   m_VertexDescriptor->pushAttrib( VAD( "position", 3 , VAD::DATA_FORMAT_FLOAT32 ) );

   m_VertexShader = new Shader;
   m_PixelShader = new Shader;

   m_VertexShader->setShaderType( Shader::VERTEX );
   m_PixelShader->setShaderType( Shader::FRAGMENT );

   m_Version = 1;
   m_ScaleValue = 1.0f;
   m_pointSize = 1.0f;

   m_Parameters = new MPContainer;

   m_RenderMode = RENDERMODE_POINTS;

   m_depthTest = true;
   m_depthWrite = true;
   m_enableDepthWriteControl = false;
   m_blending = false;
   m_vsPointSizeControl = false;
   m_sourceBlendFunc = BM_SOURCE_ALPHA;
   m_destBlendFunc = BM_ONE_MINUS_SOURCE_ALPHA;

   m_ForceTriangleNormal = false;
   m_DrawWireMesh = false;
   m_WireMeshColor = 0;
	m_noStateChanges = false;

	m_pointSpriteMode = false;
	m_sortPoints = false;
	m_frontToBack = false;

	m_drawSphericalPoints = false;
	m_BackfaceCulling = false;
}

GLShaderMaterial::~GLShaderMaterial()
{
   delete m_VertexShader;
   delete m_PixelShader;
   delete m_VertexDescriptor;

   for( TShaderMap::iterator iter=m_glVertexShaders.begin();
      iter!=m_glVertexShaders.end(); ++iter )
         delete iter->second;
   for( TShaderMap::iterator iter=m_glPixelShaders.begin();
      iter!=m_glPixelShaders.end(); ++iter )
         delete iter->second;

   delete m_Parameters;
}

void GLShaderMaterial::assign(const Object* obj, X4CopyContext *context )
{
   const GLShaderMaterial * mat = dynamic_cast<const GLShaderMaterial*>( obj );
   x4Assert( mat != NULL );

   GLMaterial::assign( obj, context );
   m_Version = mat->m_Version;
   m_ScaleValue = mat->m_ScaleValue;
   m_pointSize = mat->m_pointSize;
   m_RenderMode = mat->m_RenderMode;

   m_depthTest = mat->m_depthTest;
   m_depthWrite = mat->m_depthWrite;
   m_enableDepthWriteControl = mat->m_enableDepthWriteControl;
   m_blending = mat->m_blending;
   m_vsPointSizeControl = mat->m_vsPointSizeControl;
   m_sourceBlendFunc = mat->m_sourceBlendFunc;
   m_destBlendFunc = mat->m_destBlendFunc;

   m_VertexShader->assign( mat->m_VertexShader );
   m_PixelShader->assign( mat->m_PixelShader );
   m_Parameters->assign( mat->m_Parameters );
   m_VertexDescriptor->assign( mat->m_VertexDescriptor );

   m_ForceTriangleNormal = mat->m_ForceTriangleNormal;
   m_DrawWireMesh = mat->m_DrawWireMesh;
   m_WireMeshColor = mat->m_WireMeshColor;

   m_glLastVertexShader = NULL;
   m_glLastPixelShader = NULL;
   m_glVertexShaders.clear();
   m_glPixelShaders.clear();

   m_noStateChanges = mat->m_noStateChanges;
   m_pointSpriteMode = mat->m_pointSpriteMode;
   m_sortPoints = mat->m_sortPoints;
   m_frontToBack = mat->m_frontToBack;
   m_drawSphericalPoints = mat->m_drawSphericalPoints;
}

void GLShaderMaterial::setRenderState() {
   glPointSize(m_pointSize);
   //glEnable(GL_POINT_SMOOTH);
   //glHint (GL_POINT_SMOOTH_HINT, GL_NICEST);
   if (m_blending) {
      glEnable(GL_BLEND);
      GLenum source = GL_ONE;
      GLenum dest = GL_ONE;
      if (m_sourceBlendFunc == BM_SOURCE_ALPHA) {
         source = GL_SRC_ALPHA;
      } else if (m_sourceBlendFunc == BM_ONE_MINUS_SOURCE_ALPHA) {
         source = GL_ONE_MINUS_SRC_ALPHA;
      } else if (m_sourceBlendFunc == BM_ZERO) {
         source = GL_ZERO;
      }
      if (m_destBlendFunc == BM_SOURCE_ALPHA) {
         dest = GL_SRC_ALPHA;
      } else if (m_destBlendFunc == BM_ONE_MINUS_SOURCE_ALPHA) {
         dest = GL_ONE_MINUS_SRC_ALPHA;
      } else if (m_destBlendFunc == BM_ZERO) {
         dest = GL_ZERO;
      }
      glBlendFunc(source, dest);
   } else {
      glDisable(GL_BLEND);
   }

	if (m_vsPointSizeControl) {
		glEnable(GL_VERTEX_PROGRAM_POINT_SIZE_ARB);
	} else {
		glDisable(GL_VERTEX_PROGRAM_POINT_SIZE_ARB);
	}

	if( m_pointSpriteMode )
	{
		glEnable( GL_POINT_SPRITE_ARB );
		glTexEnvf( GL_POINT_SPRITE_ARB, GL_COORD_REPLACE_ARB, GL_TRUE );
	}

   if (m_depthTest) {
      glEnable(GL_DEPTH_TEST);
   } else {
      glDisable(GL_DEPTH_TEST);
   }
   if (m_drawSphericalPoints)
   {
	   glEnable (GL_POINT_SMOOTH);
   }

/* mb - this options doesn't make any sense at all. Depth masking
   should be controlled by renderer! */

// mw - but i like the option so I added code to enable it if desired

   if (m_enableDepthWriteControl) {
      if (m_depthWrite) {
         glDepthMask(1);
      } else {
         glDepthMask(0);
      }
   }
}

void GLShaderMaterial::setShaderParameters(Scene * scene, SceneObject * obj,
                                           SceneObjectIterator *iter, GLContext * context, MemoryManager * mm)
{
   setParameters( scene, obj, iter, context, mm );
   m_glLastVertexShader->updateDefaultMatrices();
   m_glLastPixelShader->updateDefaultMatrices();
}

void GLShaderMaterial::draw( Scene * scene, SceneGraphState *state, SceneObject * obj,
      SceneObjectIterator *iter, GLContext * context, MemoryManager * mm,
      const int32 includeFlags, const int32 excludeFlags )
{
   GLVertexBuffer * glBuffer = NULL;
	PointSet * ps = NULL;
   BasicPointCloudIterator * bIter = dynamic_cast<BasicPointCloudIterator *>( iter );
   x4Assert( bIter );

	bool useSoftwareBuffers = !context->getUseVertexBufferObjects();

	UnstructuredInCoreTriangleMesh * mesh = dynamic_cast<UnstructuredInCoreTriangleMesh*>(obj);
	
	//// TODO: Points/single vertices of a pc, that is, or inherits UnstructuredInCoreTriangleMesh, are not drawn, only the triangles - Points/single vertices should be visible, too!
	if ((mesh && mesh->getNumTriangles() == 0) && (mesh->getNumPoints() != 0 ))
		mesh = NULL; //quick fix for no triangles case
	
	
	if( !useSoftwareBuffers && m_RenderMode != RENDERMODE_SLOWSPLATTING)
	{
		glBuffer = fetchVertexBuffer( scene, state, iter, context , mm, includeFlags, excludeFlags, (mesh != NULL), m_ForceTriangleNormal ? mesh : NULL);
	}

   glPushAttrib( GL_ENABLE_BIT | GL_DEPTH_BUFFER_BIT );

   if( !m_noStateChanges )
		setRenderState();

   if( bindShaders( context ) )
   {
      setShaderParameters(scene, obj, iter, context, mm);
	  setMatrix4fv(context,"modelmatrix", state->dynamicState->ctm.data());

		if( m_RenderMode == RENDERMODE_SLOWSPLATTING )
		{
			drawSplats_Slow(iter);			

			unbindShaders(context);
			
			// just in case, should be done already by pushAttrib...
			glDepthMask(1);

			glPopAttrib();
			unlockVertexBuffer( iter );
			return;
			
		}

		if( glBuffer )
			glBuffer->bind();
		else
		{
			ps = (PointSet*)bIter->lockBlockRead();
			GLHelper::bindSoftwareVertexBuffer( ps->getDataPointer(), 
				ps->getDescr(), m_VertexDescriptor );
		}
		context->debug( "draw: " );
      GLenum renderMode = GL_POINTS;
      if( m_RenderMode == RENDERMODE_QUADS ) renderMode = GL_QUADS;
      if( m_RenderMode == RENDERMODE_NORMALS ) renderMode = GL_LINES;
      if( m_RenderMode == RENDERMODE_SPLATTING )
      {
         renderMode = GL_QUADS;
         ((GLShaderCG*)m_glLastVertexShader)->setParameter1f( "scale", m_ScaleValue );
      }

	  if ( mesh == NULL ){
		  GenericSwapAllOutOfCorePointCloud* gen = dynamic_cast<GenericSwapAllOutOfCorePointCloud*>( obj );
		  if ( gen != NULL ){
			  mesh = dynamic_cast<UnstructuredInCoreTriangleMesh*>( gen->fetchPC() );
		  }
	  }

		if( mesh && includeFlags != PF_FLAG_SELECTED )
		{
			int32 * index_ptr = (int32*)mesh->getTriangles()->getDataPtr();
			if( !m_ForceTriangleNormal )
			{
			
				GLIndexBuffer32 * ibuffer = fetchIndexBuffer(scene,state,iter,context,mm,includeFlags,excludeFlags,mesh);
				glPushAttrib( GL_ALL_ATTRIB_BITS );
				if( m_BackfaceCulling )
					glEnable( GL_CULL_FACE );
				ibuffer->bind();
				ibuffer->drawElements(GL_TRIANGLES,0,0);
				ibuffer->unbind();
				glPopAttrib();

				
			
			 /*
			 
			 
			 if( mesh->getTriangles()->getDescr()->getNumAttributes() != 1 )
				return;

            glPushAttrib( GL_ALL_ATTRIB_BITS );
			if( m_BackfaceCulling )
				glEnable( GL_CULL_FACE );
            //glCullFace( GL_FRONT );
            glDrawElements( GL_TRIANGLES, mesh->getTriangles()->getNumEntries() * 3, GL_UNSIGNED_INT, index_ptr );
            glPopAttrib();*/
		 }
		 else
		 {
			 if( mesh )
			 {
				 glPushAttrib( GL_ALL_ATTRIB_BITS );
				 if( m_BackfaceCulling )
					 glEnable( GL_CULL_FACE );
				 //glCullFace( GL_FRONT );
				 glDrawArrays( GL_TRIANGLES, 0, mesh->getTriangles()->getNumEntries() * 3 );
				 glPopAttrib();
			 }
		 }

         unbindShaders(context);

         if( m_DrawWireMesh && glBuffer)	// don't draw in software buffer mode
         {
            glPushAttrib( GL_ALL_ATTRIB_BITS );
            //glEnable( GL_CULL_FACE );
            //glCullFace( GL_FRONT );
            glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
            //glColor3f(0,0,0);
            glColor3ubv( (card8*)&m_WireMeshColor );
				glBuffer->unbind();
				glBuffer->setMappingFlag( true );
				glBuffer->bind();
            glDrawElements( GL_TRIANGLES, mesh->getTriangles()->getNumEntries() * 3, GL_UNSIGNED_INT, index_ptr );
			glBuffer->setMappingFlag( false );
            glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
            glPopAttrib();
         }

      }
      else
      {
			if( glBuffer )
				glBuffer->drawArrays(renderMode);
			else
			{
				glDrawArrays( GL_POINTS, 0, ps->getNumEntries() );
			}

         unbindShaders(context);

      }

		if( glBuffer )
			glBuffer->unbind();
		else
		{
			GLHelper::unbindSoftwareVertexBuffer( ps->getDescr(), m_VertexDescriptor ); 
			bIter->unlock();
		}

   }

   // just in case, should be done already by pushAttrib...
   glDepthMask(1);

   glPopAttrib();
	unlockVertexBuffer( iter );
}

struct SortZEntry {
	card32 index;
	float32 z;
	bool operator<(const SortZEntry &other) const {
		return z < other.z;
	}
};

static PointSet *createSortedPointSet(const PointSet *ps, Vector3f direction, Matrix4f transform, bool frontToBack) {
	if (!ps) return NULL;
	PointSet *result = new PointSet();
	result->clearAndSetup(ps->getWidth(), ps->getHeight(), ps->getDescr()); 
	card32 n = ps->getNumEntries();
	AAT posAAT = ps->getAAT("position", 3, VAD::DATA_FORMAT_FLOAT32);
	if (posAAT == NULL_AAT) throw EInvalidState("createSortedPointSet() - cannot find position channel.");
	std::vector< SortZEntry > index(n);
	for (card32 i=0; i<n; i++) {
		Vector3f pos = transformVector3f(transform, ps->get3f(i, posAAT));
		float32 z = pos*direction;
		index[i].index = i;
		if (frontToBack) {
			index[i].z = z;
		} else {
			index[i].z = -z;
		}
	}
	sort(index.begin(), index.end());
	for (card32 i=0; i<n; i++) {
		result->setVertex(i, ps->getDataPointer(index[i].index));
	}
	return result;
}

static UnstructuredInCoreTriangleMesh *createSortedMesh(const UnstructuredInCoreTriangleMesh *mesh, Vector3f direction, Matrix4f transform, bool frontToBack) {
	if (!mesh) throw EInvalidState("createSortedMesh - mesh is NULL; this must not happen.");
	UnstructuredInCoreTriangleMesh *result = (UnstructuredInCoreTriangleMesh*)mesh->copy();
	//result->clearAndSetup(mesh->getDescr(), mesh->getNumPoints(), mesh->getTriangleDescr(), mesh->getNumTriangles());
	PointSet *psVertices = result->getPointSet();
	PointSet *psFaces = result->getTriangles();

	card32 nTriangles = psFaces->getNumEntries();
	AAT posAAT = psVertices->getAAT("position", 3, VAD::DATA_FORMAT_FLOAT32);
	if (posAAT == NULL_AAT) throw EInvalidState("createSortedMesh() - cannot find position channel.");
	AAT indexAAT = psFaces->getAAT("index", 3, VAD::DATA_FORMAT_INT32);
	if (indexAAT == NULL_AAT) throw EInvalidState("createSortedMesh() - cannot find index channel.");
	std::vector< SortZEntry > index(nTriangles);
	for (card32 i=0; i<nTriangles; i++) {
		Vector3i indices = psFaces->get3i(i, indexAAT);
		Vector3f pos = transformVector3f(transform, psVertices->get3f(indices[0], posAAT));
		pos += transformVector3f(transform, psVertices->get3f(indices[1], posAAT));
		pos += transformVector3f(transform, psVertices->get3f(indices[2], posAAT));
		float32 z = pos*direction;
		index[i].index = i;
		if (frontToBack) {
			index[i].z = z;
		} else {
			index[i].z = -z;
		}
	}
	sort(index.begin(), index.end());
	const PointSet *inputFaces = mesh->getTriangles();
	for (card32 i=0; i<nTriangles; i++) {
		psFaces->set3i(i, indexAAT, inputFaces->get3i(index[i].index, indexAAT));
	}
	return result;
}


GLVertexBuffer * GLShaderMaterial::fetchVertexBuffer( 
		Scene *scene, SceneGraphState *state, SceneObjectIterator *iter,
      GLContext * context, MemoryManager * mm, const int32 includeFlags, const int32 excludeFlags, bool partOfAMesh, UnstructuredInCoreTriangleMesh * mesh )
{
      BasicPointCloudIterator * bIter = dynamic_cast<BasicPointCloudIterator *>( iter );
      x4Assert( bIter );

      if (bIter->atEnd()) {
         return NULL;
      }


      SecondaryDataStructureIterator *dataIter = dynamic_cast<SecondaryDataStructureIterator *>
         (iter);
      x4Assert(dataIter);

      std::string addDataName = "";
	  if( mesh )
		  addDataName += "MESH";
      switch( m_RenderMode )
      {
         case RENDERMODE_QUADS:
            addDataName+="quads";
         break;
         case RENDERMODE_NORMALS:
            addDataName+="normals";
         break;
         case RENDERMODE_SPLATTING:
            addDataName+="splatting";
         break;
      }
      addDataName += "Ver";
      char verNum[64];
      _i64toa(m_Version, verNum, 10);
      addDataName += verNum;
      std::string dataName = SharedVertexBufferAttachment::generateID( m_VertexDescriptor, context->getRootContext(), addDataName, includeFlags, excludeFlags );

      // try to find a existing compatible vertexbuffer
		AttachedDataContainer *attachments = dataIter->lockAttachments();
      AttachedData * data = attachments->getData( dataName );
      SharedVertexBufferAttachment * vbuffer = NULL;
      if( data )
      {
         vbuffer = dynamic_cast<SharedVertexBufferAttachment*>(data);
      }

      // generate shared vertex buffer & attach

      if( vbuffer )
         vbuffer->touched();  // <- informs memory manager

		if ( m_sortPoints && !partOfAMesh ) {
			if (vbuffer) {
				if (vbuffer->lastDirection * state->staticState->camera.getOrthoNormViewDirection() < cos(60.0f*(float)M_PI/180.0f)
					 || state->dynamicState->ctm != vbuffer->lastCTM) {
					delete vbuffer;
					vbuffer = NULL;
					data = NULL;
				}
			}
		}

      if( vbuffer == NULL )
      {
         const PointSet * ps = bIter->lockBlockRead();

			if (m_sortPoints && !partOfAMesh) {
				ps = createSortedPointSet(ps, state->staticState->camera.getOrthoNormViewDirection(), state->dynamicState->ctm, m_frontToBack);
			}

         vbuffer = new SharedVertexBufferAttachment;

         std::vector<card8> expanded_data;

         switch( m_RenderMode )
         {
         case RENDERMODE_POINTS:
         break;
         case RENDERMODE_QUADS:
         {
            if( ps->providesAttribute( "tangent_u" ) &&
               ps->providesAttribute( "tangent_v" ) )
               {
                  // TODO: vertexdescriptor methods -> const !!!
                  ps->getDescr()->buildTangentialQuads((PointSet*)ps, expanded_data, m_ScaleValue );
               }
         }
         break;
         case RENDERMODE_NORMALS:
            if( ps->providesAttribute( "normal" ) )
            {
               ps->getDescr()->expandVertices( (card8 *)ps->getDataPtr(), ps->getNumEntries(), 2, expanded_data );
               ps->getDescr()->modifyVerticesf( &expanded_data[0], expanded_data.size() / ps->getDescr()->getSize(), 2, 1,
                     "position", VertexDescriptor::MODIFIER_ADD, "normal", m_ScaleValue );
            }
         break;
         }

         std::vector<card8>  renderData;
         if( m_RenderMode != RENDERMODE_SPLATTING ) {
            if( expanded_data.size() ) {
               m_VertexDescriptor->convertVertices( expanded_data, *ps->getDescr(), renderData, includeFlags, excludeFlags );
				} else {
					if( mesh != NULL && mesh->providesAttribute("normal")) {	// hack for meshes in case of m_forcetrianglenormal=true
						std::vector<card8> tmpRenderData;
						m_VertexDescriptor->convertVertices( ps, tmpRenderData, includeFlags, excludeFlags );
						PointSet * triangles = mesh->getTriangles();
						AAT indexAAT = triangles->getAAT("index");
						AAT posAAT = mesh->getAAT("position");
						AAT normalAAT = m_VertexDescriptor->getToken("normal");
						unsigned vertexSizeInBytes = m_VertexDescriptor->getSize();
						renderData = std::vector<card8>( triangles->getNumEntries() *  vertexSizeInBytes * 3);
						card8 * currentDestPtr = &renderData[0];
						for( unsigned t=0;t<triangles->getNumEntries();t++ )
						{
							Vector3i tri = triangles->get3i(t,indexAAT);
							Vector3f p0 = mesh->getPointSet()->get3f(tri[0],posAAT);
							Vector3f p1 = mesh->getPointSet()->get3f(tri[1],posAAT);
							Vector3f p2 = mesh->getPointSet()->get3f(tri[2],posAAT);
							Vector3f normal = normalize((p1-p0).crossProduct(p2-p0));
							card8 * normalPtr = (card8*)normal.data();
							for(  unsigned k=0;k<3;k++ )
							{
								card8 * srcPtr = &tmpRenderData[tri[k]*vertexSizeInBytes];
								memcpy(currentDestPtr, srcPtr, vertexSizeInBytes);
								memcpy(currentDestPtr+normalAAT.getOffset(),normalPtr,sizeof(Vector3f));
								currentDestPtr += vertexSizeInBytes;
							}
						}
					}
					else
					{
						m_VertexDescriptor->convertVertices( ps, renderData, includeFlags, excludeFlags );
					}
				}
         } else {
            bool allIsFine = true;
            allIsFine &= m_VertexDescriptor->providesAttribute( "position" );
            allIsFine &= m_VertexDescriptor->providesAttribute( "tangent_u" );
            allIsFine &= m_VertexDescriptor->providesAttribute( "tangent_v" );
            allIsFine &= m_VertexDescriptor->providesAttribute( "pcoord" );
            allIsFine &= ps->getDescr()->providesAttribute( "position" );
            allIsFine &= ps->getDescr()->providesAttribute( "tangent_u" );
            allIsFine &= ps->getDescr()->providesAttribute( "tangent_v" );
            if( !allIsFine ) {
					renderData = std::vector<card8>( m_VertexDescriptor->getSize(), 0 );
            } else {
                m_VertexDescriptor->buildSplattingQuads( ps, renderData, includeFlags, excludeFlags );
            }
         }

			if (m_sortPoints && !partOfAMesh) {
				// in this case, ps is a sorted copy
				delete ps;
			}
         bIter->unlock();

         // generate gl vertex buffer
         GLVertexBuffer * glVertexBuffer = new GLVertexBuffer( context, *m_VertexDescriptor );
         size_t numElements =  renderData.size() / m_VertexDescriptor->getSize();
         glVertexBuffer->init( (GLuint)numElements );
         glVertexBuffer->writeToBuffer( &renderData[0], 0, (GLuint)numElements );

         vbuffer->setupVB( glVertexBuffer, addDataName , mm, true, includeFlags, excludeFlags, 
									state->staticState->camera.getOrthoNormViewDirection(),
									state->dynamicState->ctm );
         attachments->attachData( vbuffer );
         return glVertexBuffer;
      } else {
         return vbuffer->getVertexBuffer();
      }
}

void GLShaderMaterial::unlockVertexBuffer( SceneObjectIterator *iter )
{
	SecondaryDataStructureIterator *dataIter = dynamic_cast<SecondaryDataStructureIterator *> (iter);
	if (dataIter)
		dataIter->unlockAttachments();
}


bool GLShaderMaterial::bindShaders(GLContext * context)
{
   GLShader * vshader = NULL;
   GLShader * pshader = NULL;
   if( m_glVertexShaders.find( context ) != m_glVertexShaders.end() )
   {
      vshader = m_glVertexShaders[ context ];
      if( !vshader->isValid() )
      {
         delete vshader;
         vshader = NULL;
      }
   }
   if( m_glPixelShaders.find( context ) != m_glPixelShaders.end() )
   {
      pshader = m_glPixelShaders[ context ];
      if( !pshader->isValid() )
      {
         delete pshader;
         pshader = NULL;
      }
   }

   if(  (m_VertexShader->isValid() && (m_PixelShader->isValid())) )
   {
      // create shaders if necessary
      if( m_VertexShader->wasChanged() || (vshader == NULL) )
      {
         if( vshader )
            delete vshader;
         vshader = m_VertexShader->createShader( context );
         m_glVertexShaders[ context ] = vshader;
      }

      if( m_PixelShader->wasChanged() || (pshader == NULL) )
      {
         if( pshader )
            delete pshader;
         pshader = m_PixelShader->createShader( context );
         m_glPixelShaders[ context ] = pshader;
      }

      // try to bind
      if( vshader )
          vshader->bind();
      if( pshader )
          pshader->bind();

      m_glLastVertexShader = vshader;
      m_glLastPixelShader = pshader;

      return vshader && pshader;
   }

   return false;
}

void GLShaderMaterial::setParameters( Scene * scene, SceneObject * obj,
      SceneObjectIterator *iter, GLContext * context, MemoryManager * mm )
{
	GLShaderCG * lastVShader = dynamic_cast<GLShaderCG*>(m_glLastVertexShader);
	GLShaderCG * lastPShader = dynamic_cast<GLShaderCG*>(m_glLastPixelShader);

	if( lastVShader )
		lastVShader->setParameter1f( "time",scene->getRootState()->dynamicState->time);
	if( lastPShader )
		lastPShader->setParameter1f( "time",scene->getRootState()->dynamicState->time);

   for( card32 i=0;i<m_Parameters->getNumParms();i++ )
   {
      MaterialParameter * param = (MaterialParameter *)m_Parameters->getParm(i);
      //param->setDefault( this );
      param->set( this, scene, obj, iter, context, mm );
   }
}

void GLShaderMaterial::unbindShaders(GLContext * context)
{
   if( m_glVertexShaders.find( context ) != m_glVertexShaders.end() )
      m_glVertexShaders[context]->unbind();
   if( m_glPixelShaders.find( context ) != m_glPixelShaders.end() )
      m_glPixelShaders[context]->unbind();
}

void GLShaderMaterial::write(OutputObjectStream *s) const
{
   GLMaterial::write(s);
   s->writeObject(m_VertexShader);
   s->writeObject(m_PixelShader);
   s->writeObject(m_VertexDescriptor);
   s->write(m_RenderMode);
   // v121
   s->write(m_ScaleValue);
   // v first + 1
   s->write(m_pointSize);
   // v first + 2
   s->writeObject((Persistent*)m_Parameters);
   // v first + 3;
   s->writeBool(m_depthTest);
   s->writeBool(m_depthWrite);
   s->writeBool(m_blending);
   s->write(m_sourceBlendFunc);
   s->write(m_destBlendFunc);

   // v first + 4;
   s->write(m_ForceTriangleNormal);
   s->write(m_DrawWireMesh);

   // v first + 5
   s->write(m_WireMeshColor);
   
   // v first + 6
   s->writeBool(m_vsPointSizeControl);

   // v first + 7
   s->writeBool(m_noStateChanges);

   // v first + 8
   s->writeBool(m_pointSpriteMode);

   // v first + 9
   s->writeBool(m_sortPoints);
   s->writeBool(m_frontToBack);
	
	// v first + 10
	s->writeBool(m_enableDepthWriteControl);
}

void GLShaderMaterial::read(InputObjectStream *s)
{
   GLMaterial::read(s);
   delete m_VertexShader;
   s->readObject((Persistent*&)m_VertexShader);
   delete m_PixelShader;
   s->readObject((Persistent*&)m_PixelShader);
   delete m_VertexDescriptor;
   s->readObject((Persistent*&)m_VertexDescriptor);
   s->read(m_RenderMode);
   s->read( m_ScaleValue );
   s->read(m_pointSize);
   delete m_Parameters;
   s->readObject((Persistent*&)m_Parameters);
   s->readBool(m_depthTest);
   s->readBool(m_depthWrite);
   s->readBool(m_blending);
   s->read(m_sourceBlendFunc);
   s->read(m_destBlendFunc);

   s->read(m_ForceTriangleNormal);
   s->read(m_DrawWireMesh);
   s->read(m_WireMeshColor);
   s->readBool(m_vsPointSizeControl);

   s->readBool(m_noStateChanges);
	s->readBool(m_pointSpriteMode);
	s->readBool(m_sortPoints);
	s->readBool(m_frontToBack);
	s->readBool(m_enableDepthWriteControl);
}

void GLShaderMaterial::setParameter1f( GLContext * context,  const std::string &name, float f)
{
   TShaderMap::iterator itV = m_glVertexShaders.find( context );
   if( itV != m_glVertexShaders.end() )
   {
      GLShaderCG * vshader = dynamic_cast< GLShaderCG* >( itV->second );
      if( vshader )
         vshader->setParameter1f( name, f );
   }

   TShaderMap::iterator itF = m_glPixelShaders.find( context );
   if( itF != m_glPixelShaders.end() )
   {
      GLShaderCG * pshader = dynamic_cast< GLShaderCG* >( itF->second );
      if( pshader )
         pshader->setParameter1f( name, f );
   }

}

void GLShaderMaterial::setParameter2fv( GLContext * context,  const std::string &name, float * parm )
{
   TShaderMap::iterator itV = m_glVertexShaders.find( context );
   if( itV != m_glVertexShaders.end() )
   {
      GLShaderCG * vshader = dynamic_cast< GLShaderCG* >( itV->second );
      if( vshader )
         vshader->setParameter2f( name, parm[0], parm[1] );
   }

   TShaderMap::iterator itF = m_glPixelShaders.find( context );
   if( itF != m_glPixelShaders.end() )
   {
      GLShaderCG * pshader = dynamic_cast< GLShaderCG* >( itF->second );
      if( pshader )
         pshader->setParameter2f( name, parm[0], parm[1] );
   }

}

void GLShaderMaterial::setParameter3fv( GLContext * context,  const std::string &name, float * parm )
{
   TShaderMap::iterator itV = m_glVertexShaders.find( context );
   if( itV != m_glVertexShaders.end() )
   {
      GLShaderCG * vshader = dynamic_cast< GLShaderCG* >( itV->second );
      if( vshader )
         vshader->setParameter3f( name, parm[0], parm[1], parm[2] );
   }

   TShaderMap::iterator itF = m_glPixelShaders.find( context );
   if( itF != m_glPixelShaders.end() )
   {
      GLShaderCG * pshader = dynamic_cast< GLShaderCG* >( itF->second );
      if( pshader )
         pshader->setParameter3f( name, parm[0], parm[1], parm[2] );
   }

}

void GLShaderMaterial::setParameter4fv( GLContext * context,  const std::string &name, float * parm )
{
   TShaderMap::iterator itV = m_glVertexShaders.find( context );
   if( itV != m_glVertexShaders.end() )
   {
      GLShaderCG * vshader = dynamic_cast< GLShaderCG* >( itV->second );
      if( vshader )
         vshader->setParameter4f( name, parm[0], parm[1], parm[2], parm[3] );
   }

   TShaderMap::iterator itF = m_glPixelShaders.find( context );
   if( itF != m_glPixelShaders.end() )
   {
      GLShaderCG * pshader = dynamic_cast< GLShaderCG* >( itF->second );
      if( pshader )
         pshader->setParameter4f( name, parm[0], parm[1], parm[2], parm[3] );
   }

}

void GLShaderMaterial::setMatrix4fv( GLContext * context,  const std::string &name, float * parm )
{
   TShaderMap::iterator itV = m_glVertexShaders.find( context );
   if( itV != m_glVertexShaders.end() )
   {
      itV->second->setMatrix4f( name, 1, parm );
   }

   TShaderMap::iterator itF = m_glPixelShaders.find( context );
   if( itF != m_glPixelShaders.end() )
   {
      itF->second->setMatrix4f( name, 1, parm );
   }

}


std::string extractSubString( std::string str, char cBegin, char cEnd )
{
   std::string ret;
   card32 i;
   for( i=0;i<str.length();i++ )
   {
      if( str[i] == cBegin )
      {
         i++;
         break;
      }
   }

   for( ;i<str.length();i++ )
   {
      if( str[i] == cEnd )
         return ret;
      else
         ret += str[i];
   }

   return ret;
}

bool GLShaderMaterial::importFromTextFile( std::string filename )
{
   std::ifstream input( filename.c_str() );
   if( !input.is_open() )
      return false;


   int status = 0;
   const int s_nothing = 0;
   const int s_vertex = 1;
   const int s_fragment = 2;

   std::string actLine;

   std::string vshader;
   std::string fshader;

   while( getline( input, actLine ) )
   {
      std::string testline = trimStr( removeDblSpace(actLine));
      if( testline == "<END>" )
         status = s_nothing;

      switch( status )
      {
         case s_nothing:
            if( testline.length() == 0 )
               break;
            if( testline[0] == '#' )
               continue;
            if( testline[0] == '<' )
            {
                std::string tmp = extractSubString( testline, '<','>' );
                if( tmp.length() )
                {
                  std::vector<std::string> words = splitString( tmp, " \t" );
                  if( words.size() >= 2 )
                  {
                     if( words[0] == "VERTEXSHADER" )
                     {
                        status = s_vertex;
                        if( words[1] == "CG" )
                           m_VertexShader->setShaderClass( Shader::CLASS_CG );
                        if( words[1] == "ARB" )
                           m_VertexShader->setShaderClass( Shader::CLASS_ARB );
						if (words.size() >= 3)
						   m_VertexShader->setShaderProfile(words[2]);

                     }
                     if( words[0] == "FRAGMENTSHADER" )
                     {
                        status = s_fragment;
                        if( words[1] == "CG" )
                           m_PixelShader->setShaderClass( Shader::CLASS_CG );
                        if( words[1] == "ARB" )
                           m_PixelShader->setShaderClass( Shader::CLASS_ARB );
						if (words.size() >= 3)
							m_PixelShader->setShaderProfile(words[2]);
                     }
                  }
                }
            }
         break;
         case s_vertex:
            vshader += actLine + "\n";
         break;
         case s_fragment:
            fshader += actLine + "\n";
         break;
      }
   }

   m_VertexShader->setSource( vshader );
   m_VertexShader->setChanged();
   m_PixelShader->setSource( fshader );
   m_PixelShader->setChanged();
   m_VertexShader->setShaderType( Shader::VERTEX );
   m_VertexShader->setShaderClass( Shader::CLASS_CG );
   m_PixelShader->setShaderClass( Shader::CLASS_CG );

   return (vshader.length()>0) && (fshader.length()>0);
}

void GLShaderMaterial::drawSceneObjectDefaultMode( SceneObjectIterator *iter )
{
      BasicPointCloudIterator * bIter = dynamic_cast<BasicPointCloudIterator *>( iter );
      x4Assert( bIter );

      if (bIter->atEnd()) {
         return;
      }

		const PointSet * ps = bIter->lockBlockRead();

		GLHelper::drawBuffer( ps->getDataPointer(), ps->getNumEntries() * ps->getDescr()->getSize(),
			ps->getDescr(), m_VertexDescriptor, GL_POINTS );

		bIter->unlock(); 
}

void GLShaderMaterial::addVectorParameter( std::string param_name ) 
{
	MPVector4 * vecParm = new MPVector4;
	vecParm->setName( param_name );
	getParameters()->addParm( vecParm );
}

void GLShaderMaterial::addMatrixParameter( std::string param_name ) 
{
	MPMatrix4x4 * matParm2 = new MPMatrix4x4;
	matParm2->setName( param_name );
	getParameters()->addParm( matParm2 );
}

struct GLShaderMaterial_TMapVertexElement
{
	unsigned from;
	unsigned to;
	unsigned size;
};

void X4::GLShaderMaterial::drawSplats_Slow( SceneObjectIterator *iter )
{
	BasicPointCloudIterator * bIter = dynamic_cast<BasicPointCloudIterator*>(iter);

	if( !m_VertexDescriptor->providesAttribute("pcoord"))
		return;	// splatting would make no sense here!
	
	if (bIter->atEnd()) {
		return;
	}
	// lock point set
	const PointSet * ps = bIter->lockBlockRead();

	// create a temporal vertex buffer
	// process in blocks
	card32 maxNumVerticesPerBlock = 2500;
	static std::vector<char> tmpBuffer(4*maxNumVerticesPerBlock * m_VertexDescriptor->getSize());

	
	// create mapping mask for copying vertex data
	// assumption: all elements are n times 32 bit
	std::vector<GLShaderMaterial_TMapVertexElement> mapping;
	for( unsigned i=0;i<m_VertexDescriptor->getNumAttributes();i++ )
	{
		VAD destVAD = m_VertexDescriptor->getAttribute(i);
		if( ps->providesAttribute(destVAD.name, destVAD.numComponents, destVAD.dataFormat))
		{
			GLShaderMaterial_TMapVertexElement m;
			AAT srcAAT = ps->getAAT(destVAD.name);
			m.from = srcAAT.getOffset()/4;
			m.to = m_VertexDescriptor->getAttributeOffset(i)/4;
			m.size = destVAD.numComponents;
			mapping.push_back(m);
		}
	}
	AAT pcoordAAT = m_VertexDescriptor->getToken("pcoord");
	unsigned pcoord_offset = pcoordAAT.getOffset() / 4;
	
	// create a buffer for a single vertex
	std::vector<card32> singleVertex(m_VertexDescriptor->getSize()/4,0);
	card32 * singleVertexPtr = &singleVertex[0];

	// process in blocks
	unsigned numBlocks = (ps->getNumEntries()+maxNumVerticesPerBlock-1) / maxNumVerticesPerBlock;
	AAT flagsAAT = ps->getAAT("flags");
	for( unsigned b=0;b<numBlocks;b++ )
	{
		unsigned firstVertex = b*maxNumVerticesPerBlock;
		unsigned numVertices = min((mpint)maxNumVerticesPerBlock,(mpint)ps->getNumEntries() - (mpint)firstVertex);
		unsigned vSize = m_VertexDescriptor->getSize();
		unsigned vSize_dword = vSize / 4;
		
		mpcard numVisibleVertices = 0;
		for( unsigned i=0;i<numVertices;i++ )
		{
			if( ps->get1i(firstVertex+i,flagsAAT)  )
				continue;

			// create a vertex copy
			const char * srcVertexPtr = ps->getDataPointer(firstVertex + i);
			card32 * srcVertexPtr32 = (card32*)srcVertexPtr;
			for( unsigned j=0;j<mapping.size();j++ )
			{
				GLShaderMaterial_TMapVertexElement * m = &mapping[j];
				for( unsigned k=0;k<m->size;k++ )
				{
					singleVertexPtr[m->to + k]	= srcVertexPtr32[m->from + k];
				}
			}


			// create 4 vertices with different pcoord channel
			for( unsigned j=0;j<4;j++ )
			{
				Vector2f pcoord = makeVector2f((j>0 && j<3) ? -1.0f:1.0f,(j>=2)? -1.0f:1.0f);

				card32 * pcoordPtr = (card32*)pcoord.data();
				singleVertexPtr[pcoord_offset] = pcoordPtr[0];
				singleVertexPtr[pcoord_offset+1] = pcoordPtr[1];

				char * destPtr = &tmpBuffer[(numVisibleVertices*4+j)*vSize];
				card32 * destPtr32 = (card32*)destPtr;
				card32 * srcPtr32 = singleVertexPtr;
				for( unsigned k=0;k<vSize_dword;k++ )
				{
					*destPtr32++ = *srcPtr32++;
				}
			}

			numVisibleVertices++;

		}

		// render array
		if( numVisibleVertices > 0)
		{
			GLHelper::bindSoftwareVertexBuffer(&tmpBuffer[0], m_VertexDescriptor, m_VertexDescriptor);
			glDrawArrays( GL_QUADS,0, numVisibleVertices*4);
			GLHelper::unbindSoftwareVertexBuffer(m_VertexDescriptor,m_VertexDescriptor);
		}


	}



	bIter->unlock(); 	



}

void X4::GLShaderMaterial::setDefaultParameter( const std::string &name, const Vector4f &v )
{
	MPVector4 * parm = NULL;
	// search
	for( unsigned i=0;i<m_Parameters->getNumParms();i++)
	{
		parm = dynamic_cast<MPVector4 *>(m_Parameters->getParm(i));
		if( parm && parm->getName() == name )
			break;
		parm = NULL;
	}
	if (!parm)
	{
		parm = new MPVector4;
		parm->setName(name);
		m_Parameters->addParm(parm);
	}
	parm->setDefaultValue(v);
}

void X4::GLShaderMaterial::setDefaultParameter( const std::string &name, const Matrix4f &v )
{
	MPMatrix4x4 * parm = NULL;
	// search
	for( unsigned i=0;i<m_Parameters->getNumParms();i++)
	{
		parm = dynamic_cast<MPMatrix4x4 *>(m_Parameters->getParm(i));
		if( parm && parm->getName() == name )
			break;
		parm = NULL;
	}
	if (!parm)
	{
		parm = new MPMatrix4x4;
		parm->setName(name);
		m_Parameters->addParm(parm);
	}
	parm->setMatrix(v);
}

GLIndexBuffer32 * X4::GLShaderMaterial::fetchIndexBuffer( Scene *scene, SceneGraphState *state, SceneObjectIterator *iter, GLContext * context, MemoryManager * mm, const int32 includeFlags, const int32 excludeFlags, UnstructuredInCoreTriangleMesh * mesh )
{
	std::string id = GLIndexAttachment::generateID("",context,mm,includeFlags,excludeFlags);
	SecondaryDataStructureIterator *dataIter = dynamic_cast<SecondaryDataStructureIterator *>
		(iter);

	AttachedDataContainer * att = dataIter->lockAttachments();
	GLIndexAttachment *indexAtt = dynamic_cast<GLIndexAttachment*>( att->getData(id) );
	if (m_sortPoints && indexAtt) {
		if (indexAtt->lastDirection * state->staticState->camera.getOrthoNormViewDirection() < cos(60.0f*(float)M_PI/180.0f)
			|| state->dynamicState->ctm != indexAtt->lastCTM) 
		{
			delete indexAtt;
			indexAtt = NULL;
		}
	}

	bool deleteMesh = false;
	if( indexAtt == NULL )
	{
		indexAtt = new GLIndexAttachment;
		if (m_sortPoints) {
			mesh = createSortedMesh(mesh, state->staticState->camera.getOrthoNormViewDirection(), state->dynamicState->ctm, m_frontToBack);
			deleteMesh = true;
		}
		indexAtt->setupIndexBuffer("",context,mm,true,includeFlags,excludeFlags,mesh, state->staticState->camera.getOrthoNormViewDirection(), state->dynamicState->ctm);
		att->attachData(indexAtt);
	}

	dataIter->unlockAttachments();
	if (deleteMesh) {
		// in this case, it's a copy
		delete mesh;
	}

	return indexAtt->getBuffer();
}

GLShaderMaterial *GLShaderMaterial::createFromShaderFile( const std::string &filename) 
{
	GLShaderMaterial * mat = new GLShaderMaterial;
	if( !mat->importFromTextFile(filename) )
	{
		message( std::string("GLShaderMaterial::createFromShaderFile() - Could not find shader file ") + filename + "\n");
	}
	return mat;
}
