//---------------------------------------------------------------------------
#ifndef GLShaderMaterialH
#define GLShaderMaterialH
//---------------------------------------------------------------------------
#include "x3rendering.h"
//---------------------------------------------------------------------------
#include "GLMaterial.h"
//---------------------------------------------------------------------------
#include <map>
#include "GLBuffer.h"
//---------------------------------------------------------------------------

namespace X4 {


class MPContainer;
class GLShaderMaterialEditorWidget;
class UnstructuredInCoreTriangleMesh;

class X3RENDERING_API GLShaderMaterial : public GLMaterial
{
	
 private:
   static MetaClass *GLShaderMaterialMetaClass;
   static Object* newInstance();

   Shader   * m_VertexShader;
   Shader   * m_PixelShader;
   VertexDescriptor * m_VertexDescriptor;
   GLShader * m_glLastVertexShader;
   GLShader * m_glLastPixelShader;

   typedef std::map<GLContext*,GLShader*> TShaderMap;

   TShaderMap   m_glVertexShaders;
   TShaderMap   m_glPixelShaders;

   MPContainer *  m_Parameters;

   card32   m_RenderMode;
   card64   m_Version;

   float32  m_ScaleValue;
   float32  m_pointSize;

   /// render settings
   bool m_depthTest;
   bool m_depthWrite;
   bool m_enableDepthWriteControl;
   bool m_blending;
   bool m_vsPointSizeControl;
   bool m_noStateChanges;
   card32 m_sourceBlendFunc;
   card32 m_destBlendFunc;
   bool m_pointSpriteMode;
   bool m_sortPoints;
   bool m_frontToBack;
   bool m_drawSphericalPoints;

   // triangle special options
   bool m_ForceTriangleNormal;   /// forces material to use triangle normals instead of vertex normals
   bool m_DrawWireMesh;          /// draws wiremesh after solid rendering ( warning: z fighting ! fix in vertex shader)
   bool m_BackfaceCulling;		 /// cull backfacing triangles
   card32 m_WireMeshColor;       /// color of wiremesh   

 public:
   static const card32 BM_ONE = 0;
   static const card32 BM_SOURCE_ALPHA = 1;
   static const card32 BM_ONE_MINUS_SOURCE_ALPHA = 2;
   static const card32 BM_ZERO = 3;

 private:
   friend GLShaderMaterialEditorWidget;

 protected:

 /// sets all GL state parameters (useful for use in descendant materials;
 virtual void setRenderState();
 /// prepares shader for rendering (useful for use in descendant materials;
 virtual void setShaderParameters(Scene * scene, SceneObject * obj,
	 SceneObjectIterator *iter, GLContext * context, MemoryManager * mm);

 /// draws a scene object with default gl arrays
 void drawSceneObjectDefaultMode( SceneObjectIterator *iter );

 void drawSplats_Slow(SceneObjectIterator *iter);

public:

	GLShaderMaterial();
	virtual ~GLShaderMaterial();

	static const card32 RENDERMODE_POINTS = 1;
	static const card32 RENDERMODE_QUADS = 2;
	static const card32 RENDERMODE_NORMALS = 3;
	static const card32 RENDERMODE_SPLATTING = 4;
	static const card32 RENDERMODE_SLOWSPLATTING = 5;

	card64 getVersion() const {return m_Version;}
	void increaseVersion() {m_Version++;}
	virtual void rebuildRenderObjects(){increaseVersion();}

	card32 getRenderMode() const {return m_RenderMode;}
   void setRenderMode( card32 rm ) {m_RenderMode = rm;}
   float32 getScaleValue() const {return m_ScaleValue;}
   void setScaleValue( float32 scale ) {m_ScaleValue = scale;}
   float32 getPointSize() const {return m_pointSize;}
   void setPointSize( float32 ps ) {m_pointSize = ps;}

   Shader * getVertexShader() {return m_VertexShader;}
   Shader * getFragmentShader() {return m_PixelShader;}
   VertexDescriptor * getVertexDescriptor() {return m_VertexDescriptor;}

   GLShader * getGLVertexShader(GLContext * context) {return m_glVertexShaders[context];}
   GLShader * getGLPixelShader(GLContext * context) {return m_glPixelShaders[context];}

   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

   virtual void draw( Scene * scene, SceneGraphState *state, SceneObject * obj,
      SceneObjectIterator *iter, GLContext * context, MemoryManager * mm = NULL,
      const int32 includeFlags = 0, const int32 excludeFlags = PF_FLAG_INVALID );

   virtual void write(OutputObjectStream *s) const;
   virtual void read(InputObjectStream *s);

   virtual void assign(const Object* obj, X4CopyContext *context=NULL );

   virtual bool bindShaders(GLContext *);
   virtual void setParameters( Scene * scene, SceneObject * obj,
      SceneObjectIterator *iter, GLContext * context, MemoryManager * mm );
   virtual void unbindShaders(GLContext *);

   /// works for cg shaders only
   void setParameter1f( GLContext * context, const std::string &name, float );
   void setParameter2fv( GLContext * context, const std::string &name, float * parm );
   void setParameter3fv( GLContext * context, const std::string &name, float * parm );
   void setParameter4fv( GLContext * context, const std::string &name, float * parm );
   void setMatrix4fv( GLContext * context, const std::string &name, float * parm );

   /// set default value of a parameter (and create parameter if necessary)
   void setDefaultParameter(const std::string &name, const Vector4f &v);
   void setDefaultParameter(const std::string &name, const Matrix4f &v);


   // render state
	bool getNoStateChanges() {return m_noStateChanges;}
	void setNoStateChanges(bool b) {m_noStateChanges = b;}
   bool getDepthTest() {return m_depthTest;}
   void setDepthTest(bool d) {m_depthTest = d;}
   bool getDepthWrite() {return m_depthWrite;}
   void setDepthWrite(bool d) {m_depthWrite = d;}
	bool getEnableDepthWriteControl() const {return m_enableDepthWriteControl;}
	void setEnableDepthWriteControl(bool v) {m_enableDepthWriteControl=v;}
   bool getBlending() {return m_blending;}
   void setBlending(bool b) {m_blending = b;}
   bool getVSPointSizeControl() {return m_vsPointSizeControl;}
   void setVSPointSizeControl(bool b) {m_vsPointSizeControl = b;}
   card32 getSourceBlendFunc() {return m_sourceBlendFunc;}
   void   setSourceBlendFunc(card32 b) {m_sourceBlendFunc = b;}
   card32 getDestBlendFunc() {return m_destBlendFunc;}
   void   setDestBlendFunc(card32 b) {m_destBlendFunc = b;}

   void setForceTriangleNormal( bool b ) {m_ForceTriangleNormal = b;}
   bool getForceTriangleNormal() {return m_ForceTriangleNormal;}
   void setDrawWireMesh( bool b ) {m_DrawWireMesh = b;}
   bool getDrawWireMesh() {return m_DrawWireMesh;}

   card32 getWireMeshColor() {return m_WireMeshColor;}
   void setWireMeshColor( card32 color ) {m_WireMeshColor = color;}

   bool importFromTextFile( std::string filename );
   static GLShaderMaterial * createFromShaderFile( const std::string &filename);

	// parameters
	MPContainer * getParameters() {return m_Parameters;}
	void addVectorParameter( std::string param_name );
	void addMatrixParameter( std::string param_name );

	bool getPointSpriteMode() const {return m_pointSpriteMode;}
	void setPointSpriteMode(bool v) {m_pointSpriteMode=v;}

	bool getSortPoints() const {return m_sortPoints;}
	void setSortPoints(bool v) {m_sortPoints=v;} 

	bool getFrontToBack() const {return m_frontToBack;} 
	void setFrontToBack(bool v) {m_frontToBack=v;}

	bool getDrawSphericalPoints() const {return m_drawSphericalPoints;} 
	void setDrawSphericalPoints(bool v){m_drawSphericalPoints = v;} 

	bool getBackfaceCulling() const {return m_BackfaceCulling;} 
	void setBackfaceCulling(bool v){m_BackfaceCulling = v;} 

protected:

   virtual GLVertexBuffer * fetchVertexBuffer( 
		Scene *scene, SceneGraphState *state, SceneObjectIterator *iter,
      GLContext * context, MemoryManager * mm, const int32 includeFlags, const int32 excludeFlags, bool partOfAMesh,
	  UnstructuredInCoreTriangleMesh * mesh = NULL);

   GLIndexBuffer32 * fetchIndexBuffer(Scene *scene, SceneGraphState *state, SceneObjectIterator *iter,
	   GLContext * context, MemoryManager * mm, const int32 includeFlags, const int32 excludeFlags,
	   UnstructuredInCoreTriangleMesh * mesh );

	void unlockVertexBuffer( SceneObjectIterator *iter );

};



}

#endif

