//---------------------------------------------------------------------------
#include "stdafx.h"
//---------------------------------------------------------------------------
#include "LCHierarchyRenderer.h"
#include "LCFilters.h"
#include "Scene.h"
#include "PointCloud.h"
#include "SceneGraphState.h"
#include "PointCloudIterators.h"
#include "ObjectClassProperty.h"
#include "MemoryManager.h"
#include "AttachedIndexedOctree.h"
#include "Tools.h"
#include "SGSceneObjectFilterAdaptor.h"
#include "StandardSceneGraphIterator.h"
#include "SceneGraphNode.h"
#include "SGObjectNode.h"
#include "RenderableObject.h"
#include "VolumeObject.h"
//---------------------------------------------------------------------------
 
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------


card32 X4::GLOBAL_VAR_LCHierarchyRenderer_numNodesRendered = 0;
card32 X4::GLOBAL_VAR_LCHierarchyRenderer_numPointsRendered = 0;


MetaClass* LCHierarchyRenderer::LCHierarchyRendererMetaClass = NULL;

const MetaClass* LCHierarchyRenderer::getInstanceClass() const {
   return getClass();
}

Object* LCHierarchyRenderer::newInstance() {
   return new LCHierarchyRenderer;
}

const MetaClass* LCHierarchyRenderer::getClass() {
   if (LCHierarchyRendererMetaClass==NULL) {
      LCHierarchyRendererMetaClass = new MetaClass("LCHierarchyRenderer", "no description.",
                                  newInstance, sizeof(LCHierarchyRenderer));
   }
   return LCHierarchyRendererMetaClass;
}

void LCHierarchyRenderer::init(const MetaClass *parentClass) {
   ((MetaClass*)getClass())->init(parentClass);
   LCHierarchyRendererMetaClass->copyParentProperties();
	BEGIN_CLASS_INIT(LCHierarchyRenderer)
   ADD_OBJECT_PROP(lcTrav, 0, LCTraversal::getClass(), true)
}

void LCHierarchyRenderer::shutdown() {
   delete LCHierarchyRendererMetaClass;
}


// ----


LCHierarchyRenderer::LCHierarchyRenderer() {
   memManager = NULL;
   lcTrav = new LCTraversal(new LCState());

   // inner nodes processing
   lcTrav->addSubdividedNodeProcessors(new LCDrawInnerBoxesNodeProcessor());

   // processing for inner nodes at which the traversal stops
   lcTrav->addNodeProcessors(new LCDrawBoxesNodeProcessor());
   lcTrav->addNodeProcessors(new LCDrawPointsMaterial());
   lcTrav->addNodeProcessors(new LCDrawCloudSelection());
   lcTrav->addNodeProcessors(new LCDrawPointSelection());

   // leaf node processing
   lcTrav->addUnSubdividableNodesProcessors(new LCDrawBoxesNodeProcessor());
   lcTrav->addUnSubdividableNodesProcessors(new LCDrawPointsMaterial());
   lcTrav->addUnSubdividableNodesProcessors(new LCDrawCloudSelection());
   lcTrav->addUnSubdividableNodesProcessors(new LCDrawPointSelection());

   // traversal
   lcTrav->addNodeFilters(new LCViewFrustumCulling());
   lcTrav->addNodeFilters(new LCProjectiveClassification());
   lcTrav->addNodeFilters(new LCNonPresentNodesCulling());
   
   lcTrav->addStatePostProcessors(new LCStatisticsOutput());
}

LCHierarchyRenderer::LCHierarchyRenderer(int createType) {
   memManager = NULL;
   lcTrav = new LCTraversal(new LCState());

   switch( createType )
   {
      default:
      case CREATE_NOGLSTATECHANGE:
      // inner nodes processing
      lcTrav->addSubdividedNodeProcessors(new LCDrawInnerBoxesNodeProcessor());

      // processing for inner nodes at which the traversal stops
      lcTrav->addNodeProcessors(new LCDrawBoxesNodeProcessor());
		lcTrav->addNodeProcessors(new LCDrawPointsMaterialNoStateChange());
      lcTrav->addNodeProcessors(new LCDrawCloudSelection());
      lcTrav->addNodeProcessors(new LCDrawPointSelection());

      // leaf node processing
      lcTrav->addUnSubdividableNodesProcessors(new LCDrawBoxesNodeProcessor());
      lcTrav->addUnSubdividableNodesProcessors(new LCDrawPointsMaterialNoStateChange());
      lcTrav->addUnSubdividableNodesProcessors(new LCDrawCloudSelection());
      lcTrav->addUnSubdividableNodesProcessors(new LCDrawPointSelection());

      // traversal
      lcTrav->addNodeFilters(new LCViewFrustumCulling());
      lcTrav->addNodeFilters(new LCProjectiveClassification());
	   lcTrav->addNodeFilters(new LCNonPresentNodesCulling());
   break;
   }
}

void LCHierarchyRenderer::connect(Scene* scene, GLContext* ctxt) {
   Renderer::connect(scene, ctxt);

   x4Assert( memManager == NULL );
   memManager = new MemoryManager( 256 * 1024 * 1024 ); // max. 256 MB for vertex buffers
}

void LCHierarchyRenderer::render(card32 viewportWidth, card32 viewportHeight) {

   Scene *scene = getScene();
   glPushAttrib(GL_ALL_ATTRIB_BITS);

   try {
      glEnable(GL_DEPTH_TEST);

      StandardSceneGraphIterator sceneIt;
      sceneIt.clearAndSetup(scene->getRootNode(), scene->getRootState(), SGDynamicState::tmRendering, 0, 0);

      renderRecursive(sceneIt, viewportWidth, viewportHeight);

   } catch(...) {
      glPopAttrib();
		throw;
   }

   glPopAttrib();

}

void LCHierarchyRenderer::renderRecursive(StandardSceneGraphIterator &sceneIt, card32 viewportWidth, card32 viewportHeight) {

   if (!sceneIt.getNode()->visible()) return;

   glMatrixMode(GL_MODELVIEW);
   glPushMatrix();
   glMultMatrixf(sceneIt.getState()->dynamicState->ctm.data());

	if (lcTrav->getTraversalState()->PARAM_drawNodeGLRenderings) {
		sceneIt.getNode()->renderNodeGL(sceneIt.getState());
	}

   SGObjectNode *objNode = dynamic_cast<SGObjectNode*>(sceneIt.getNode());
   if (objNode != NULL) {
		SceneObject* obj = objNode->getSceneObject();

		if (obj != NULL && objNode->visible()) {

			PointCloud *pc = dynamic_cast<PointCloud*>(obj);
			if (!pc) {
				glMatrixMode(GL_MODELVIEW);
				glPopMatrix();
            return;
			}
			if (!pc->getDescr()->providesAttribute("position", 3, VAD::DATA_FORMAT_FLOAT32)) 
				if( dynamic_cast<VolumeObject*>(pc) == NULL )			
				{
					glMatrixMode(GL_MODELVIEW);
					glPopMatrix();
					return;
				}

			if (lcTrav->getTraversalState()->PARAM_drawObjectGLRenderings) {
				obj->renderObjectGL();
			}

			// check for selection
			set<std::string> &sel = sceneIt.getState()->staticState->selectedNodes;
			std::string name = sceneIt.getNameWithPath();
			set<std::string>::iterator pos = sel.find(name);
			bool selected = (pos != sel.end());
			lcTrav->getTraversalState()->currentCloudIsSelected = selected;


			SceneGraphState *state = sceneIt.getState();
			SceneObjectIterator *it = obj->createIterator(SOIT(SOIT::CAP_HIERARCHICAL | SOIT::CAP_ATTACH_SECONDARY));
			HierarchicalIterator *hit = dynamic_cast<HierarchicalIterator*>(it);
			

			if (lcTrav->getTraversalState()->PARAM_drawAttachmentGLRenderings) {
				AttachedDataContainer *attachments = obj->getAttachments();
				std::vector<AttachedData*> allAtt;
				attachments->getAll(allAtt);
				size_t n = allAtt.size();
				for (size_t i=0; i<n; i++) {
					RenderableObject* renderableObj = dynamic_cast<RenderableObject*>(allAtt[i]);
					if (renderableObj != NULL) {
						renderableObj->renderGL(obj);
					}
				}
			}

			if (hit != NULL) {
				try {
					lcTrav->performTraversal(it, state, this, viewportWidth, viewportHeight,
													 getScene(), obj, getGLContext(), memManager);
					GLOBAL_VAR_LCHierarchyRenderer_numNodesRendered = lcTrav->getTraversalState()->numBoxes;
					GLOBAL_VAR_LCHierarchyRenderer_numPointsRendered = lcTrav->getTraversalState()->numPoints;
				} catch(...) {
					glMatrixMode(GL_MODELVIEW);
					glPopMatrix();
					delete it;
					throw;
				}
			}
			delete it;
		}
	}

   glMatrixMode(GL_MODELVIEW);
   glPopMatrix();


   card32 numChildren = sceneIt.getNumChildren();
   for (card32 i=0; i<numChildren; i++) {
      sceneIt.goToChild(i);
      renderRecursive(sceneIt, viewportWidth, viewportHeight);
      sceneIt.goToParent();
   }
}


void LCHierarchyRenderer::disconnect(GLContext* ctxt) {
   forget(ctxt);

   x4Assert( memManager != NULL );
   delete memManager;
   memManager = NULL;
}


LCHierarchyRenderer::~LCHierarchyRenderer() {
   delete memManager; // in case not yet disconnected
   delete lcTrav;
}



void LCHierarchyRenderer::setUseBBSelection( bool b )
{
	lcTrav->getTraversalState()->useBBSelection = b;
}