//---------------------------------------------------------------------------
#ifndef MaterialAttachmentH
#define MaterialAttachmentH
//---------------------------------------------------------------------------
#include "x3rendering.h"
//---------------------------------------------------------------------------
#include "AttachedData.h"
#include "LinearAlgebra.h"
//---------------------------------------------------------------------------

namespace X4 {

class GLTexture;
class GLContext;


class X3RENDERING_API Matrix4fAttachment : public AttachedData
{
private:
   static MetaClass *Matrix4fAttachmentMetaClass;
   static Object* newInstance();

public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

private:

   Matrix4f m_Matrix;

public:

   Matrix4fAttachment() : AttachedData()
   {
      m_Matrix = IDENTITY4F;
   }

   void setMatrix( const Matrix4f &mat ) {m_Matrix = mat; }
   Matrix4f getMatrix() const {return m_Matrix; }
   virtual void assign(const Object* obj, X4CopyContext *context = NULL);
};

template<class FloatType, unsigned dim>
class X3RENDERING_API VectorAttachmentBase : public AttachedData
{
protected:

	StaticVector<FloatType, dim> m_Vector;
	card32	m_NumComponentsUsed;	// out of [1,4]

public:

	VectorAttachmentBase() : AttachedData()
	{
		m_Vector.setZero();
		m_NumComponentsUsed = m_Vector.size();
	}

	virtual void setVector( const StaticVector<FloatType, dim> &vec ) {m_Vector = vec; }
	virtual StaticVector<FloatType, dim> getVector() const {return m_Vector;}
	virtual void setNumComponentsUsed( card32 n ) {m_NumComponentsUsed = n;}
	virtual card32 getNumComponentsUsed() {return m_NumComponentsUsed;}

	virtual void assign(const Object* obj, X4CopyContext *context = NULL);
};


class X3RENDERING_API Vector4fAttachment : public VectorAttachmentBase<float,4>
{
	X4_CLASS(Vector4fAttachment);
};

class X3RENDERING_API Vector4iAttachment : public VectorAttachmentBase<int,4>
{
	X4_CLASS(Vector4iAttachment);
};

class X3RENDERING_API TextureAttachment : public AttachedData
{
private:
   static MetaClass *TextureAttachmentMetaClass;

public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

   TextureAttachment() : AttachedData()
   {
      m_Container = new AttachedDataContainer;
   }

   ~TextureAttachment()
   {
      delete m_Container;
   }

   static const card32 TYPE_2D = 1;
   static const card32 TYPE_3D = 2;
   static const card32 TYPE_CUBE = 3;
   static const card32 TYPE_RECT_FLOAT = 4;

   virtual card32 getTextureType() const = 0;
   virtual GLTexture * createGLTexture(GLContext * context) = 0;

   virtual void assign( const Object * obj,  X4CopyContext *context = NULL );

   inline AttachedData * fetchData( const std::string &name)
   {
      return m_Container->getData( name );
   }

   inline void deleteData( const std::string &name )
   {
      m_Container->deleteData(name);
   }

   inline void attachData( AttachedData * data )
   {
      m_Container->attachData(data);
   }

   inline void clearAttachments()
   {
      m_Container->clear();
   }

	GLTexture * getCachedTexture(GLContext * glContext);


private:

   AttachedDataContainer   *  m_Container;

};



class X3RENDERING_API RandomTextureAttachment : public TextureAttachment
{
	X4_CLASS(RandomTextureAttachment)
public:
   virtual card32 getTextureType() const {return TYPE_2D; }
   virtual GLTexture * createGLTexture(GLContext * context);
   virtual bool needsUpdate() {return false;}

};



class X3RENDERING_API TestTextureAttachment : public TextureAttachment
{
public:
   virtual card32 getTextureType() const {return TYPE_2D; }
   virtual GLTexture * createGLTexture(GLContext * context);
   virtual bool needsUpdate() {return false;}

};

}

#endif
