//---------------------------------------------------------------------------
#ifndef MaterialParameterH
#define MaterialParameterH
//---------------------------------------------------------------------------
#include "x3rendering.h"
//---------------------------------------------------------------------------
#include "Persistent.h"
//---------------------------------------------------------------------------

namespace X4 {

class GLContext;
class AttachedData;
class AttachedDataContainer;
class GLMaterial;
class Scene;
class SceneObject;
class SceneObjectIterator;
class SecondaryDataStructureIterator;
class AttachedData;
class MemoryManager;
class TextureAttachment;


class X3RENDERING_API MaterialParameter : public Persistent
{
private:
   static MetaClass *MaterialParameterMetaClass;

public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();


private:

   // properties
   std::string   m_Name;  /// identifier
   bool     m_UseGlobalFirst;   /// if true, search in global datacontainers first
   bool     m_UseDefault;  /// if true & no container knows 'Name' -> use default value

public:

   MaterialParameter()
   :Persistent()
   {
      m_Name = "unnamed";
      m_UseGlobalFirst = false;
      m_UseDefault = true;
   }

   std::string getName() const {return m_Name;}
   void setName( const std::string &name ) {m_Name = name;}
   bool getUseGlobalFirst() const {return m_UseGlobalFirst;}
   void setUseGlobalFirst( bool b) {m_UseGlobalFirst = b;}
   bool getUseDefault() const {return m_UseDefault;}
   void setUseDefault( bool b) {m_UseDefault = b;}

   virtual void assign(const Object* obj, X4CopyContext *context = NULL);

   virtual void set( GLMaterial * mat, Scene * scene, SceneObject *obj,
      SceneObjectIterator * iter, GLContext * context, MemoryManager * mm )=0;

protected:

    virtual AttachedData * getData( GLMaterial * mat, Scene * scene, SceneObject * obj,
      SecondaryDataStructureIterator * iter );

};


class X3RENDERING_API MPMatrix4x4 : public MaterialParameter
{
private:
   static MetaClass *MPMatrix4x4MetaClass;
   static Object* newInstance();

public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

private:

   card32   m_Index; /// modelview[Index] for ARB shaders
   float32  m_DefaultMatrix[4*4];

public:


   MPMatrix4x4() : MaterialParameter()
   {
      // identity
      for( card32 i=0;i<4*4;i++ )
         m_DefaultMatrix[i] = (i % 5) == 0 ? 1.0f : 0.0f;
      m_Index = 0;
   }

   virtual void assign(const Object* obj, X4CopyContext *context=NULL );

   virtual void set( GLMaterial * mat, Scene * scene, SceneObject * obj,
      SceneObjectIterator * iter, GLContext * context, MemoryManager * mm );

   void setMatrix( const Matrix4f &mat) {for(int i=0;i<16;i++ ) m_DefaultMatrix[i] = mat.data()[i]; }

};

class X3RENDERING_API MPVector4 : public MaterialParameter
{
private:
   static MetaClass *MPVector4MetaClass;
   static Object* newInstance();

public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

private:

   card32   m_Index; /// parameter # for ARB shaders
   float32  m_DefaultVector[4];
   float32  m_MinVector[4];
   float32  m_MaxVector[4];
   bool     m_ClampMin;
   bool     m_ClampMax;

public:
   MPVector4() : MaterialParameter()
   {
      for( card32 i=0;i<4;i++ )
      {
         m_DefaultVector[i] = 0.0f;
         m_MinVector[i] = 0.0f;
         m_MaxVector[i] = 1.0f;
      }
      m_ClampMin = false;
      m_ClampMax = false;
      m_Index = 0;
   }
   virtual void assign(const Object* obj, X4CopyContext *context=NULL );

   virtual void set( GLMaterial * mat, Scene * scene, SceneObject * obj,
      SceneObjectIterator * iter, GLContext * context, MemoryManager * mm );

   void setDefaultValue(const Vector4f &v) {for( int i=0;i<4;i++ ) m_DefaultVector[i] = v[i];}

};

class X3RENDERING_API MPTexture : public MaterialParameter
{
private:
   static MetaClass *MPTextureMetaClass;
   static Object* newInstance();

public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();


private:

   //AttachedDataContainer * m_AttachmentContainer;
   card32   m_TextureStage;
   std::string   m_DefaultTextureName;
	TextureAttachment	*	m_DefaultTexture;

public:

   MPTexture() : MaterialParameter()
   {
      m_TextureStage = 0;
      m_DefaultTextureName = "texture white 2D";
		m_DefaultTexture = NULL;
   }
	~MPTexture();

	void setTextureStage(card32 stage) { m_TextureStage = stage; }

   virtual void assign(const Object* obj, X4CopyContext *context = NULL);

   virtual void set( GLMaterial * mat, Scene * scene, SceneObject * obj,
      SceneObjectIterator * iter, GLContext * context, MemoryManager * mm );
};

class X3RENDERING_API MPSingleFloat : public MaterialParameter
{
private:
   static MetaClass *MPSingleFloatMetaClass;
   static Object* newInstance();

public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();

private:
   float32  m_Minimum, m_Maximum;
   float32  m_Interpolator;

public:

   MPSingleFloat()
   {
      m_Minimum = 0.0f;
      m_Maximum = 1.0f;
      m_Interpolator = 0.0f;
   }

   virtual void set( GLMaterial * mat, Scene * scene, SceneObject * obj,
      SceneObjectIterator * iter, GLContext * context, MemoryManager * mm );

};


class X3RENDERING_API MPMieRayleighParameter : public MaterialParameter
{
	X4_CLASS(MPMieRayleighParameter)
private:

	// scattering parameters
	float32     m_HG;
	float32     m_Inscattering;
	float32     m_Rayleigh;
	float32     m_Mie;
	float32		m_SunPhi;
	float32		m_SunTheta;

	// see ATI SDK scattering demo
	struct TScatteringParameter
	{
		Vector3f m_vBetaRay;				// Rayleigh scattering coeff
		Vector3f m_vBetaDashRay;			// Rayleigh Angular scattering coeff without phase term.
		Vector3f m_vBetaMie;				// Mie scattering coeff
		Vector3f m_vBetaDashMie;			// Mie Angular scattering coeff without phase term.

		// sun
		float	m_SunIntensity;
		Vector3f	m_SunAttenuation;
		Vector3f	m_SunDirection;

		float	m_DepthScale;

	};

	TScatteringParameter m_ScatteringParameter;

	void calculateScatteringConstants();

public:

	MPMieRayleighParameter();

	virtual void set( GLMaterial * mat, Scene * scene, SceneObject * obj,
		SceneObjectIterator * iter, GLContext * context, MemoryManager * mm );

};

}

#endif
