//---------------------------------------------------------------------------
#include "stdafx.h"

//---------------------------------------------------------------------------
#include "Shader.h"
//---------------------------------------------------------------------------
#include "ShaderEditorWidget.h"
#include "NumericalClassProperty.h"

#include "StringClassProperty.h"
#include "GLShader.h"
#include "GLContext.h"
#include "X4Exception.h"
#include "X4Assert.h"
#include "Tools.h"
#include "DebugOutput.h"
#include <fstream>
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------



MetaClass* Shader::ShaderMetaClass = NULL;

const MetaClass* Shader::getInstanceClass() const {
    return getClass();
}

Object* Shader::newInstance() {
    return new Shader;
}

const MetaClass* Shader::getClass() {
    if (ShaderMetaClass==NULL)
    {
        ShaderMetaClass = new MetaClass("Shader", "no description.",
            newInstance, sizeof(Shader));
    }
    return ShaderMetaClass;
}


std::string ShaderTypeDesc(card32 i)
{
    switch(i)
    {
    case Shader::INVALID:
        return "invalid";
    case Shader::VERTEX:
        return "Vertexshader";
    case Shader::FRAGMENT:
        return "Fragmentshader";
    }
    return "unknown";
}


std::string ShaderClassDesc(card32 i)
{
    switch(i)
    {
    case Shader::INVALID:
        return "invalid";
    case Shader::CLASS_ARB:
        return "gl ARB";
    case Shader::CLASS_CG:
        return "CG";
    }

    return "unkown";
}

void Shader::init(const MetaClass *parentClass) {
    ((MetaClass*)getClass())->init(parentClass);
    ShaderMetaClass->copyParentProperties();
    BEGIN_CLASS_INIT(Shader)
        ADD_CARD32_PROP_UPDATE_DESCR(m_ShaderType ,0, NULL, &ShaderTypeDesc, 0, NULL)		// 0 --> display as text editor
        ADD_CARD32_PROP_UPDATE_DESCR(m_ShaderClass ,0, NULL, &ShaderClassDesc, 0, NULL)  // 0 --> display as text editor
        ADD_STRING_PROP_METHOD(m_ShaderSource ,0, &getSource, &setSource)

}

void Shader::shutdown() {
    delete ShaderMetaClass;
}


// ----

Shader::Shader()
{
    m_ShaderType = INVALID;
    m_ShaderClass = INVALID;
    m_Changed = false;
    m_FileName = "";
    m_ShaderSource = "";
}

void Shader::assign(const Object* obj, X4CopyContext *context )
{
    const Shader * sh = dynamic_cast<const Shader*>(obj);
    x4Assert( sh != NULL );

    m_ShaderType = sh->m_ShaderType;
    m_ShaderClass = sh->m_ShaderClass;
    m_Changed = true;
    m_ShaderSource = sh->m_ShaderSource;
    m_LastError = "";
    forget(context);
}


std::string Shader::getSource() const
{
    return m_ShaderSource;
}

void Shader::setSource(std::string source)
{
    m_ShaderSource = source;
}

void Shader::setShaderType( card32 stype )
{
    m_ShaderType = stype;               
}

void Shader::setShaderClass( card32 stype )
{
    m_ShaderClass = stype;
}


bool Shader::compile(GLContext * testcontext)
{
    m_LastError = "";
    testcontext->activate();
    GLShader * testshader = createShaderInstance(testcontext);
    bool ret = testshader->compile( m_ShaderSource );
    if( !ret )
        m_LastError = testshader->getLastError();
    else
        m_Changed = true;
    testcontext->deactivate();
    delete testshader;
    return ret;
}

GLShader * Shader::createShaderInstance( GLContext * context )
{
    if( (m_ShaderType == INVALID) || (m_ShaderClass == INVALID) )
        throw X4Exception("Shader::createShader - trying to create shader with INVALID ShaderType or ShaderClass");

    m_Changed = false;

    if( m_ShaderClass == CLASS_ARB)
    {
        if( m_ShaderType == VERTEX )
            return new GLShaderARB( context, GL_VERTEX_PROGRAM_ARB );
        if( m_ShaderType == FRAGMENT )
            return new GLShaderARB( context, GL_FRAGMENT_PROGRAM_ARB );
    }
    if( m_ShaderClass == CLASS_CG )
    {
        // Debugging, pj
        if ( GLShaderCG::getLatestVSProfile(m_ShaderProfile) == CG_PROFILE_UNKNOWN ){
            debugOutput << "Unknown vertex shader profile.\n"; 
        }
        if ( GLShaderCG::getLatestVSProfile(m_ShaderProfile) == CG_PROFILE_UNKNOWN ){
            debugOutput << "Unknown pixel shader profile.\n"; 
        }
        //debugOutput << GLShaderCG::getLatestVSProfile(); 
        //debugOutput.flush();

        if( m_ShaderType == VERTEX )

            return new GLShaderCG( context, GLShaderCG::getLatestVSProfile(m_ShaderProfile) );
        if( m_ShaderType == FRAGMENT )
            return new GLShaderCG( context, GLShaderCG::getLatestPSProfile(m_ShaderProfile) );
    }

    throw X4Exception("Shader::createShader - trying to create shader with unkown parameter (Type or Class)");
}

GLShader * Shader::createShader( GLContext * context )
{
    GLShader *ret=createShaderInstance( context );
    if( !ret->compile( m_ShaderSource ) )
    {
        debugOutput << "Shader::createShader() - couldn't compile shader: " << m_FileName
            //<< "\n----------------------\n" << m_ShaderSource
            << "\n";
        debugOutput << ret->getLastError() << "\n";
    }
    return ret;
}

bool Shader::isValid()
{
    return (m_ShaderType != INVALID) && (m_ShaderClass != INVALID) && (!m_ShaderSource.empty());
}

bool X4::Shader::loadFromFile( const std::string &filename )
{
    m_FileName = filename;
    std::ifstream input( filename.c_str() );
    if( !input.is_open() )
        return false;

    m_ShaderSource = "";
    std::string currentLine;
    while( getline(input,currentLine))	
    {
        m_ShaderSource += currentLine + "\n";
    }

    input.close();
    return true;
}