//---------------------------------------------------------------------------
#ifndef POINT_CLOUD_TOOLS_H
#define POINT_CLOUD_TOOLS_H
//---------------------------------------------------------------------------
#include "x3sceneGraph.h"
//---------------------------------------------------------------------------
#include "PointCloud.h"
#include "UnstructuredInCorePointCloud.h"
#include "UnstructuredInCoreTriangleMesh.h"
#include "RegressionPlane3f.h"
#include "HierarchicalKNNIterator.h"
#include "SGObjectNode.h"
#include "Scene.h"
#include "SGLegacyPCSceneAdaptor.h"
//---------------------------------------------------------------------------

#define NUM_NEIGHBORS_LOCAL_CONTROL_PLANE 12

namespace X4 {

   /// Find scene object node
   X3SCENEGRAPH_API SGObjectNode* getSceneGraphObject( Scene* scene, std::string pcID );

   /// Find point cloud in scene
   X3SCENEGRAPH_API PointCloud*  getPointCloud( Scene* scene, std::string pcID );

   template<typename Base> Base* getPointCloudAs( Scene* scene, std::string pcID ) { return dynamic_cast<Base*>( getPointCloud(scene, pcID) ); }

   /// Find any scene object
   X3SCENEGRAPH_API SceneObject*  getSceneObject( Scene* scene, std::string pcID );

	/// Get transformation of scene graph node by name, along with the node itself; parameter node is NULL if not found
   X3SCENEGRAPH_API void getSGNodeTransformation(Scene *scene, std::string nodeID, Matrix4f &transformation, SceneGraphNode *&node);

   /// Replace point cloud 'pcID' with 'newPC'
   X3SCENEGRAPH_API void  replacePointCloud( Scene* scene, std::string pcID, PointCloud* newPC );

   /// Replace point cloud 'pcID' with 'newPC' or add if it doesn't exist
   X3SCENEGRAPH_API void  addOrReplacePointCloud( Scene* scene, std::string pcID, PointCloud* newPC );

   /// Add point cloud 'pc' as a child of the root node; set node name to 'name'.
   X3SCENEGRAPH_API SceneGraphNode* addPointCloud( Scene* scene, PointCloud* pc, std::string name );

	/// Add scene object 'so' as a child of the root node; set node name to 'name'. Returns created ObjectNode.
	X3SCENEGRAPH_API SceneGraphNode* addSceneObject( Scene * scene, SceneObject *so, std::string name );


   /// Add point cloud 'pc' as a child of the root node; set node name to 'name'; returns really set name because this can differ from the requested name. always use this returned name afterwards!
   X3SCENEGRAPH_API std::string  addPointCloudAndReturnNodeName( Scene* scene, PointCloud* pc, std::string name );

   /// Check for attribute: Search attribute 'attrib' in point cloud 'id' in scene 'pcscn'. If it does not exist, create attribute as (numComponents, dataType)
   X3SCENEGRAPH_API void  checkAttribute( Scene* pcscn, std::string id, std::string attrib, card32 numComponents, card8 dataType ) ;

   /// Check for attribute: Search attribute 'attrib' in point cloud pc. If it does not exist, create attribute as (numComponents, dataType)
   X3SCENEGRAPH_API void  checkAttribute( PointCloud * pc, std::string attrib, card32 numComponents, card8 dataType ) ;

	/// CheckAttribute: Check for (position, normal, flags, color)
   X3SCENEGRAPH_API void  checkSceneAllAttributes( Scene* scene );

   /// Convert point cloud to unstructured point cloud, whatever the type was before
   X3SCENEGRAPH_API UnstructuredInCorePointCloud*  convertToUnstructuredInCorePointCloud( PointCloud* inPC );

   /// Return transformation that automatically aligns the point cloud
   X3SCENEGRAPH_API Matrix4f  getAutoAlignMatrix( PointCloud* pc );

   /// Returns resampled point cloud
   X3SCENEGRAPH_API PointCloud*  createResampled( PointCloud *input, card32 x, card32 y );


   /// returns NULL if the scene does not containe a node named "name".
   X3SCENEGRAPH_API SceneGraphNode*  getSceneGraphNodeByName(Scene *scene, std::string name);

   /// returns NULL if iven node does not contain a child (also grandchildren) with the given name
   X3SCENEGRAPH_API SceneGraphNode*  getSceneGraphNodeByName(SceneGraphNode *root, std::string name, const SceneGraphState* state);

   /// Return point cloud from node; returns NULL if the node does not contain a point cloud
   X3SCENEGRAPH_API PointCloud*  getPointCloud( SceneGraphNode *node );

   /// Creates a random permutation of points in a point set
   X3SCENEGRAPH_API void randomlyPermutePointSet(PointSet *ps);

   /// Wwaps two points from a PointSet
   X3SCENEGRAPH_API void swapPSPoints(PointSet *ps, mpcard i1, mpcard i2);


   /// Return the bounding box of all scene point clouds
   X3SCENEGRAPH_API BoundingBox3f  getScenePCBBox( Scene* scene );

   /// Return the bounding box of all visible scene point clouds
   X3SCENEGRAPH_API BoundingBox3f  getScenePCBBoxVisibleOnly( Scene* scene );

   /// Return the bounding box of a point cloud
   X3SCENEGRAPH_API BoundingBox3f  getPCBBox( PointCloud* pc );

	/// Return the average distance between two points of a point cloud
	X3SCENEGRAPH_API float getAveragePointDistance( PointCloud *pc, card32 numDirectNeighbours = 5 );

	/// returns median distance between two points (numSamples == 0 -> all points)
   X3SCENEGRAPH_API float getMedianPointDistance( PointCloud *pc, card32 numDirectNeighbours = 5, card32 numSamples = 1000 );
	/// returns median distance between two points (stores result in attachments, reuses if attachment exists)
	X3SCENEGRAPH_API float getCachedMedianPointDistance( PointCloud * pc, card32 numDirectNeighbours = 5, card32 numSamples = 1000 );

	/// Delete node with name and path 'id' from scene. Returns true on success.
   X3SCENEGRAPH_API bool  deleteSceneGraphNode( Scene* scene, std::string id);
   /// Delete point cloud 'pcID' from scene. Returns true on success.
   X3SCENEGRAPH_API bool  deletePointCloud( Scene* scene, std::string pcID);


   /// Compute the regression plane from closest points
   X3SCENEGRAPH_API void  getLocalPlane( RegressionPlane3f& plane, Vector3f pos, PointCloud* pc, HierarchicalKNNIterator* pcNNIt );

   /// Compute the regression plane from closest points
   X3SCENEGRAPH_API void  getLocalPlane( RegressionPlane3f& plane, Vector3f pos, float32 lfs, PointCloud* pc, HierarchicalKNNIterator* pcNNIt );
   
   /// retrieve transformation matrix of a scene graph node
   X3SCENEGRAPH_API Matrix4f getTransformation(Scene *scene, std::string node);

	/// smoothes attribute "inputChannel" and stores the (normalized?) result in "outputChannel"
	/// uses a gaussian filter kernel exp( -distance^2/ h^2), distance will be scaled by 1/averagePointDistance
	/// averagePointDistance will be estimated when not specified 
	/// this function doesn't perform smoothing in a two step way, therefore: inputChannel != outputChannel
	/// (otherwise there will be different results)
	X3SCENEGRAPH_API void smoothAttribute1f( PointCloud * pc, std::string inputChannel, std::string outputChannel, float h,
							  bool normalize=false, float averagePointDistance = 0.0f );
	X3SCENEGRAPH_API void smoothAttribute2f( PointCloud * pc, std::string inputChannel, std::string outputChannel, float h,
							  bool normalize=false, float averagePointDistance = 0.0f );
	X3SCENEGRAPH_API void smoothAttribute3f( PointCloud * pc, std::string inputChannel, std::string outputChannel, float h,
							  bool normalize=false, float averagePointDistance = 0.0f );
	X3SCENEGRAPH_API void smoothAttribute4f( PointCloud * pc, std::string inputChannel, std::string outputChannel, float h,
							  bool normalize=false, float averagePointDistance = 0.0f );


	/// create a vertexdescriptor
	X3SCENEGRAPH_API VertexDescriptor * createDefaultVertexDescriptor( bool position, bool color, bool normal, bool flags=true, bool tangents=false );

	/// merge two point clouds
	/// vertex descriptors should match!
	X3SCENEGRAPH_API void merge( PointCloud * pc1, const PointCloud * pc2);


	// get absolute point cloud positions
	X3SCENEGRAPH_API void getPCAbsolutPositions(PointCloud* pc, std::vector<Vector3f>& ret); 


	X3SCENEGRAPH_API void removePoints( UnstructuredInCorePointCloud* pc, const std::vector<bool>& keep );
	X3SCENEGRAPH_API void removePointsAdjustTriangles( UnstructuredInCoreTriangleMesh* mesh, const std::vector<bool>& keep );


	X3SCENEGRAPH_API Vector3f getPCCenterOfGravity( PointCloud* pc );

}

#endif