#ifndef ProcessPointsCmdObjH
#define ProcessPointsCmdObjH
//---------------------------------------------------------------------------
#include "x3sceneGraph.h"
//---------------------------------------------------------------------------
#include "CommandObject.h"
#include "VertexDescriptor.h"
//---------------------------------------------------------------------------

namespace X4
{

/**
	Special command object for processing points. Iterates through
	point cloud (and visualizes with progressbar). Offers basic/modify
	iterators and HierarchicalKNN on demand. Also creates AATs for
	basic channels (position, normal, color, flags) and also source/dest
	channls, if defined.

	Usage:
	 - derive from ProcessPointsCmdObj
	 - implement doAction(),canUndo(),undoAction(),getDescription()
	 - implement at least one prepareSomething and processSomething method
	 - create a specific context (not necessary)
	 - call PROCESS_ALL_POINTS( scene, &myContext, &MyClass::prepareSomthing,&MyClass::processSomthing ) in doAction()

	Example:

	------- header ---------------

	class MyClass : public ProcessPointsCmdObj
	{
	public:
   virtual void doAction(Object *context);
   virtual void undoAction(Object *context){}
   virtual bool canUndo() {return false;}
   virtual std::string getDescription() {return "todo";}

	void prepareSomething(PointInfo* info, void *context);
	void processSomething(PointInfo* info, void *context);
		
	};

	------ cpp --------------------

	struct TSomethingContext
	{
		int foo;
	};
	
	void MyClass::doAction(Object *context)
	{
	   Scene *scene = dynamic_cast<Scene*>(context); x4Assert(scene);

		// define required channels
		m_RequiredChannels.clear();
		m_RequiredChannels.push_back( VAD( m_DestChannel, 1, VAD::DATA_FORMAT_FLOAT32 ) );

		TSomethingContext myContext;
		myContext.foo = 1;
		PROCESS_ALL_POINTS( scene, &myContext, &MyClass::prepareSomthing,&MyClass::processSomthing );
	}

	void MyClass::prepareSomething(PointInfo* info, void *context)
	{
		TSomethingContext * myContext	= (TSomethingContext *)context;

		// prepare something, but don't change pc with checkAttributes!
	}

	void MyClass::processSomething(PointInfo* info, void *context)
	{
		TSomethingContext * myContext	= (TSomethingContext *)context;

		Vector3f pos = info->bIt->get3f( info->posAAT );
		float length = norm( pos );
		
		info->mIt->set1f( info->destAAT, length ); 
	}

*/

	
class BasicPointCloudIterator;
class ModifyablePointCloudIterator;
class	HierarchicalKNNIterator;
class Scene;
class PointCloud;

class X3SCENEGRAPH_API ProcessPointsCmdObj : public CommandObject
{
	X4_ABSTRACT_CLASS( ProcessPointsCmdObj )

private:

	bool		m_CreateHierarchicalIterator;	// default is true
	bool		m_InvalidateAttachments;		// default is true, calls clearAttachments for 
														// every point cloud after processing
	card32	m_SelectionMode;	// default is SM_SelectedOrAll;

public:
	ProcessPointsCmdObj(void);
	~ProcessPointsCmdObj(void);

	// settings interface
	void setCreateHierarchicalIterator(bool b) {m_CreateHierarchicalIterator = b;}
	bool getCreateHierarchicalIterator() const {return m_CreateHierarchicalIterator;}

	void setInvalidateAttachments(bool b) {m_InvalidateAttachments = b;}
	bool getInvalidateAttachments() const {return m_InvalidateAttachments;}

	std::string getSourceChannel() const {return m_SourceChannel;}
	void setSourceChannel(std::string v) {m_SourceChannel=v;}

	std::string getDestChannel() const {return m_DestChannel;}
	void setDestChannel(std::string v) {m_DestChannel=v;}

	enum ESelectionMode
	{
		SM_SelectedOrAll,		// use selected or, if nothing selected, use all
		SM_Selected,
		SM_All
	};
	void setSelectionMode( card32 mode ) {m_SelectionMode=mode;}
	card32 getSelectionMode() const {return m_SelectionMode;}

   virtual void assign(const Object* obj, X4CopyContext *context = NULL);

protected:

	std::string	m_SourceChannel;			// specifies source channel, default ""
	std::string	m_DestChannel;

	card32	m_SourceCompIndex;		// specifies source channels component index (float a = source[sourceCompIndex])
	card32	m_DestCompIndex;

	std::vector<VAD>	m_RequiredChannels;	// specifies required channels in point clouds. non-persistent, 
												// you must not set this anywere except in doAction()!
								
	
	/// passed to prepare and process method
	struct PointInfo
	{
		BasicPointCloudIterator * bIt;
		ModifyablePointCloudIterator *mIt;

		HierarchicalKNNIterator * hIt;
		unsigned int index;

		AAT	posAAT;
		AAT	normalAAT;
		AAT	colorAAT;
		AAT	flagsAAT;

		
		AAT	sourceAAT;	// specify m_SourceChannel before using this AAT!
		AAT	destAAT;		// specify m_DestChannel before using this AAT!
		VAD	sourceVAD;
		VAD	destVAD;

		unsigned int	sourceCompIndex;
		unsigned int	destCompIndex;

		PointCloud	* pc;
		Scene			* scene;
		std::string		pointCloudName;
	};
	
	typedef void (Object::* ProcessPointMethod)(PointInfo*,void *);
	typedef void (Object::* PrepareProcessMethod)(PointInfo*,void *);
	
	/// process all points - loops over all points in all pointclouds
	void processAllPoints( Scene * scene, void * context, PrepareProcessMethod prepare, ProcessPointMethod process );

	
	struct TPCInfo
	{
		std::string			pcName;
		PointCloud *	pc;
	};

	/// returns selected point clouds for processing ( see SelectionMode )
	std::vector<TPCInfo> getCurrentPointClouds(Scene * scene);

};

#define PROCESS_ALL_POINTS( scene, context, prepare, process ) \
	processAllPoints( scene, context, (PrepareProcessMethod)prepare, (ProcessPointMethod)process );

}

#endif