//---------------------------------------------------------------------------
#include "stdafx.h"

//---------------------------------------------------------------------------
#include "SGLegacyPCSceneAdaptor.h"
//---------------------------------------------------------------------------
#include "Scene.h"
#include "SGListNode.h"
#include "PointCloud.h"
#include "VisitSGNodesOnceIterator.h"
#include "SGSceneObjectFilterAdaptor.h"
#include "SGIndexedAndByNameSceneGraphIteratorAdaptor.h"
#include "SGObjectNode.h"
#include "GLMaterial.h"
#include "StringHelper.h"
//---------------------------------------------------------------------------
 
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------



// --- constructor

SGLegacyPCSceneAdaptor::SGLegacyPCSceneAdaptor(Scene *scene) {
   x4Assert(scene != NULL);
   this->scene = scene;

   if (scene->getRootNode() == NULL) {
      scene->setRootNode(new SGListNode());
   } else if (dynamic_cast<SGListNode*>(scene->getRootNode()) == 0) {
      SGListNode *newRoot = new SGListNode();
      newRoot->addNode(scene->getRootNode());
      scene->setRootNode(newRoot);
   }

   sceneIterator = NULL;
   pointCloudFilter = NULL;
   indexingAdaptor = NULL;

   rebuildIndex();
}

void SGLegacyPCSceneAdaptor::rebuildIndex() {
   delete sceneIterator;
   delete pointCloudFilter;
   delete indexingAdaptor;

   sceneIterator = new StandardSceneGraphIterator(); // xxx new VisitSGNodesOnceIterator();
   sceneIterator->clearAndSetup(scene->getRootNode(), scene->getRootState());
   pointCloudFilter = new SGSceneObjectFilterAdaptor();
   pointCloudFilter->setReportedSceneObjectType(PointCloud::getClass());
   pointCloudFilter->setupAndGoToFirstNode(sceneIterator);
   indexingAdaptor = new SGIndexedAndByNameSceneGraphIteratorAdaptor();
   indexingAdaptor->clearAndSetup(pointCloudFilter);
}


// --- access point clouds within scene graph

mpcard SGLegacyPCSceneAdaptor::getNumPointClouds() {
   return indexingAdaptor->getNumNodes();
}

PointCloud *SGLegacyPCSceneAdaptor::getPointCloud(card32 index) {
   indexingAdaptor->goTo(index);
   SceneGraphNode *n = indexingAdaptor->getNode();
   SGObjectNode *objn = dynamic_cast<SGObjectNode*>(n);
   x4Assert(objn != NULL);
   PointCloud *pc = dynamic_cast<PointCloud*>(objn->getSceneObject());
   x4Assert(pc != NULL);
   return pc;
}

PointCloud *SGLegacyPCSceneAdaptor::getPointCloud(std::string name) {
   bool found = indexingAdaptor->goToName(name);
   if (!found) return NULL;
   SceneGraphNode *n = indexingAdaptor->getNode();
   SGObjectNode *objn = dynamic_cast<SGObjectNode*>(n);
   x4Assert(objn != NULL);
   PointCloud *pc = dynamic_cast<PointCloud*>(objn->getSceneObject());
   x4Assert(pc != NULL);
   return pc;
}

SceneGraphNode *SGLegacyPCSceneAdaptor::getNode(card32 index) {
   indexingAdaptor->goTo(index);
   SceneGraphNode *n = indexingAdaptor->getNode();
   return n;
}

SceneGraphNode *SGLegacyPCSceneAdaptor::getNode(std::string name) {
   bool found = indexingAdaptor->goToName(name);
   if (!found) return NULL;
   SceneGraphNode *n = indexingAdaptor->getNode();
   return n;
}


std::string SGLegacyPCSceneAdaptor::getNameWithPath(card32 index) {
   indexingAdaptor->goTo(index);
   return indexingAdaptor->getNameWithPath();
}


mpint SGLegacyPCSceneAdaptor::getPointCloudIndex(std::string name) {
   bool found = indexingAdaptor->goToName(name);
   if (!found) {
      return -1;
   } else {
      return (mpint)indexingAdaptor->getCurrentIndex();
   }
}

bool SGLegacyPCSceneAdaptor::visible(card32 index) {
   return getNode(index)->visible();
}



// --- add, remove, replace point clouds

void SGLegacyPCSceneAdaptor::addPointCloud(PointCloud* newPC, std::string name) {
   SGListNode *rootNode = dynamic_cast<SGListNode*>(scene->getRootNode());
   x4Assert(rootNode != NULL);
   SGObjectNode *objNode = new SGObjectNode();
   objNode->clearAndSetup(newPC);
   if (name != "") {
      objNode->setName(name);
   }
   rootNode->addNode(objNode);
   rebuildIndex();
}

void SGLegacyPCSceneAdaptor::addNode(SceneGraphNode *node) {
   SGListNode *rootNode = dynamic_cast<SGListNode*>(scene->getRootNode());
   x4Assert(rootNode != NULL);
   rootNode->addNode(node);
   rebuildIndex();
}

void SGLegacyPCSceneAdaptor::deletePointClouds(const std::vector<card32> pcs) {
   std::vector<std::string> pcNames;
   for (card32 i=0; i<pcs.size(); i++) {
      x4Assert(pcs[i] < indexingAdaptor->getNumNodes());
      indexingAdaptor->goTo(pcs[i]);
      std::string name = indexingAdaptor->getNameWithPath();
      pcNames.push_back(name);
   }

   for (card32 i=0; i<pcNames.size(); i++) {
      bool found = indexingAdaptor->goToName(pcNames[i]);
      x4Assert(found);
      SceneGraphNode *parent = indexingAdaptor->getParent();
      parent->removeChildNode(removePathFromFilename(pcNames[i]));
      rebuildIndex();
   }
}

void SGLegacyPCSceneAdaptor::setSelectedPCs(const std::vector<card32> &sel) {
   SceneGraphState *state = scene->getRootState();
   state->staticState->selectedNodes.clear();
   for (card32 i=0; i<sel.size(); i++) {
      indexingAdaptor->goTo(sel[i]);
      state->staticState->selectedNodes.insert(indexingAdaptor->getNameWithPath());
   }
}

void SGLegacyPCSceneAdaptor::replacePointCloud(card32 i, PointCloud *pc) {
   x4Assert(i < indexingAdaptor->getNumNodes());
   indexingAdaptor->goTo(i);
   SceneGraphNode *n = indexingAdaptor->getNode();
   SGObjectNode *objn = dynamic_cast<SGObjectNode*>(n);
   x4Assert(objn != NULL);
   objn->clearAndSetup(pc);
}

void SGLegacyPCSceneAdaptor::clear() {
   SGListNode *root = new SGListNode();
   scene->setRootNode(root);
   rebuildIndex();
}


// --- state handling

SceneGraphState *SGLegacyPCSceneAdaptor::getRootState() {
   return scene->getRootState();
}

const SceneGraphState *SGLegacyPCSceneAdaptor::getState(card32 pcNum) {
   x4Assert(pcNum < indexingAdaptor->getNumNodes());
   indexingAdaptor->goTo(pcNum);
   return indexingAdaptor->getState();
}


// --- materials

void SGLegacyPCSceneAdaptor::addMaterial(GLMaterial *mat) {
   scene->addGLMaterial(mat);
}

GLMaterial *SGLegacyPCSceneAdaptor::getMaterial(card32 i) {
   return scene->getGLMaterial(i);
}

GLMaterial *SGLegacyPCSceneAdaptor::getMaterial(std::string name) {
   const max_platform_card n = scene->getNumGLMaterials();
   for (max_platform_card i=0; i<n; i++) {
      if (scene->getGLMaterial((card32)i)->getName() == name) {
         return scene->getGLMaterial((card32)i);
      }
   }
   return NULL;
}

int32 SGLegacyPCSceneAdaptor::getMaterialIndex(std::string name) {
   const max_platform_card n = scene->getNumGLMaterials();
   for (max_platform_card i=0; i<n; i++) {
      if (scene->getGLMaterial((card32)i)->getName() == name) {
         return (int32)i;
      }
   }
   return -1;
}

/// returns false if name already exists
bool SGLegacyPCSceneAdaptor::renameMat(card32 index, std::string newName) {
   if (getMaterialIndex(newName) != -1) return false;
   GLMaterial *mat = scene->getGLMaterial(index);
   mat->setName(newName);
   return true;
}


// --- attachments
inline AttachedDataContainer *SGLegacyPCSceneAdaptor::getAttachments() {
   return scene->getAttachments();
}

// --- selection

std::vector<card32> SGLegacyPCSceneAdaptor::getSelectedPCs() {
   std::vector<card32> result;
   SceneGraphState *state = scene->getRootState();
   set<std::string>::iterator pos = state->staticState->selectedNodes.begin();
   set<std::string>::iterator end = state->staticState->selectedNodes.end();
   for (; pos != end; ++pos) {
      std::string name = *pos;
      if (indexingAdaptor->goToName(name)) {
         result.push_back((card32)indexingAdaptor->getCurrentIndex());
      }
   }
   return result;
}

// --- command history
CommandHistory *SGLegacyPCSceneAdaptor::getCommandHistory() {
   return scene->getCommandHistory();
}

SGLegacyPCSceneAdaptor::~SGLegacyPCSceneAdaptor() {
   delete indexingAdaptor;
   delete pointCloudFilter;
   delete sceneIterator;
}
