//---------------------------------------------------------------------------
#include "stdafx.h"

//---------------------------------------------------------------------------
#include "Scene.h"
//---------------------------------------------------------------------------
#include "SceneGraphNode.h"
#include "SceneGraphState.h"
#include "AttachedData.h"
#include "ObjectClassProperty.h"
#include "GLMaterial.h"
#include "FixedArrayClassProperty.h"
#include "CommandHistory.h"
#include "AttachedDataTypes.h"
#include "CommandObjMacro.h"
#include "SceneGraphState.h"
#include "SGListNode.h"
#include "PointCloud.h"
#include "SGObjectNode.h"
#include "ObjectListProperty.h"
#include <limits.h>
#include "Statistics.h"
//---------------------------------------------------------------------------
 
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------


MetaClass* Scene::SceneMetaClass = NULL;

const MetaClass* Scene::getInstanceClass() const {
   return getClass();
}

Object* Scene::newInstance() {
   return new Scene;
}

const MetaClass* Scene::getClass() {
   if (SceneMetaClass==NULL) {
      SceneMetaClass = new MetaClass("Scene", "no description.",
                                  newInstance, sizeof(Scene));
   }
   return SceneMetaClass;
}

// --- property access methods - glMaterials ---
GLMaterial *Scene::getGLMat(mpcard index) const {
   if( glMaterials.size() > index )
      return glMaterials[index];
   else
      return NULL;
}
void Scene::setGLMat(mpcard index, GLMaterial *pcm) {
   if (pcm == NULL) return;
   if (glMaterials[index] != pcm && glMaterials[index] != NULL) {
      delete glMaterials[index];
      glMaterials[index] = NULL;
   }
   glMaterials[index] = pcm;
}
mpcard Scene::getNumGLMats() const {
   return glMaterials.size();
}
void Scene::setNumGLMats(mpcard size) {
   size_t oldSize = glMaterials.size();
   glMaterials.resize(size);
   for (size_t i=oldSize; i<size; i++) {
      glMaterials[i] = NULL;
   }
}
void Scene::pushGLMat() {
   glMaterials.push_back(NULL);
}
void Scene::popGLMat() {
   if (glMaterials.back() != NULL) {
      delete glMaterials.back();
   }
   glMaterials.pop_back();
}
// ---------------------------------------------

void Scene::addMaterial( Object * obj )
{
	glMaterials.push_back( (GLMaterial*)obj );
}
void Scene::deleteMaterials( mpcard begin, mpcard end )
{
	glMaterials.erase( glMaterials.begin() + begin, glMaterials.begin() + end );
}

Object * Scene::getMaterial( mpcard index )
{
	x4Assert( index < glMaterials.size() );
	return glMaterials[index];
}

void Scene::moveMaterial( mpcard oldPos, mpcard newPos )
{
	Object * mat = glMaterials[oldPos];
	glMaterials.erase( glMaterials.begin() + oldPos );
	glMaterials.insert( glMaterials.begin() + newPos, (GLMaterial*)mat );
}

mpcard Scene::getNumMaterials()
{
	return (mpcard)glMaterials.size();
}


std::string Scene::getMaterialName(mpcard index )
{
	x4Assert( index < glMaterials.size() );
	GLMaterial * mat = (GLMaterial*)glMaterials[index];
	return mat->getName();
}


void Scene::init(const MetaClass *parentClass) {
	((MetaClass*)getClass())->init(parentClass, MetaClass::FIRST_VERSION + 4);
	SceneMetaClass->copyParentProperties();
	SceneMetaClass->getGUIDescriptor()->setIcon(":/foundation/images/Classes/Scene.png");

	BEGIN_CLASS_INIT(Scene)
	ADD_OBJECT_PROP( rootState, 0, SceneGraphState::getClass(), true )
	ADD_OBJECT_PROP( rootNode, 0, SceneGraphNode::getClass(), true )   
	ADD_OBJECT_PROP( attachments, 0, AttachedDataContainer::getClass(), true )   
	ADD_OBJECT_LIST_PROP( Material, 3, GLMaterial::getClass() );
	ADD_OBJECT_PROP( commandHistory, 0, CommandHistory::getClass(), true );
	ADD_OBJECT_PROP( statistics, 4, Statistics::getClass(),true );
}

void Scene::shutdown() {
   delete SceneMetaClass;
}


// ----


Scene::Scene() {
   rootState = new SceneGraphState();
   rootNode = new SGListNode();
   rootNode->setName("root");
   rootNode->addReference(NULL);
   attachments = new AttachedDataContainer;
   commandHistory = new CommandHistory;
	m_HQRMaterial = NULL;
	statistics = new Statistics;
}

void Scene::read(InputObjectStream *s) {
	if (s->getClassVersion(getClass()) >= MetaClass::FIRST_VERSION + 2) {

		// standard behaviour

		Persistent::read(s);

	} else {

		// import legacy format ("PointCloudScene")

		SGListNode *root = new SGListNode;
		card32 n;
		s->read<card32>(n);
	
		for (card32 i=0; i<n; i++) {
			PointCloud *pc;
			s->readObject((Persistent*&)pc);
			SGObjectNode *obj = new SGObjectNode();
			obj->clearAndSetup(pc);
			root->addNode(obj);
		}
	
		SceneGraphState *state;
		s->readObject((Persistent*&)state);

		clearAndSetup(state, root);
		
		s->read<card32>(n);
		for (card32 i=0; i<n; i++) {
			GLMaterial *mat;
			s->readObject((Persistent*&)mat);
			addGLMaterial(mat);
		}
	
		delete commandHistory;
		commandHistory = NULL;
		s->readObject((Persistent*&)commandHistory);
	
		card32 currentObjectNameNumber;
		s->read(currentObjectNameNumber);
		forget(currentObjectNameNumber);
	
		delete attachments;
		attachments = NULL;
		s->readObject((Persistent*&)attachments);
	}
}

void Scene::clearAndSetup(SceneGraphState *rootState, SceneGraphNode *rootNode) {
   x4Assert(this->rootState != rootState);
   delete this->rootState;
   this->rootState = rootState;

   x4Assert(this->rootNode != rootNode);
   delete this->rootNode;
   this->rootNode = rootNode;
   rootNode->addReference(NULL);

   delete attachments;
   delete commandHistory;
   attachments = new AttachedDataContainer;
   commandHistory = new CommandHistory;
}

SceneGraphState *Scene::getRootState() {
   return rootState;

}

SceneGraphNode *Scene::getRootNode() {
   return rootNode;
}

void Scene::setRootState(SceneGraphState *r) {
   if (r != rootState) {
      delete rootState;
      rootState = r;
   }
}

void Scene::setRootNode(SceneGraphNode *r) {
   if (r != rootNode) {
      rootNode->releaseReference(NULL);
      rootNode = r;
      r->addReference(NULL);
   }
}


void Scene::addGLMaterial(GLMaterial *newMat) {
   glMaterials.push_back(newMat);
}

mpcard Scene::getNumGLMaterials() {
   return glMaterials.size();
}

mpcard Scene::getGLMaterialIndex(const std::string& name)
{
   for (unsigned i=0; i < glMaterials.size(); i++)
      if(glMaterials[i]->getName() == name) return i;
   return INT_MAX;
}

GLMaterial *Scene::getGLMaterial(mpcard i) {
   x4Assert(i < glMaterials.size());
   if (i >= glMaterials.size()) {
      if (glMaterials.size() == 0) return NULL;
      return glMaterials[0];
   }
   return glMaterials[i];
}

void Scene::deleteGLMaterial(mpcard i) {
   if (i > glMaterials.size()) {
      x4Assert(false);
      return;
   }
   delete glMaterials[i];
   glMaterials.erase(glMaterials.begin() + i);
}



void Scene::clear() {
   clearAndSetup(NULL, NULL);
}

CommandHistory *Scene::getCommandHistory() {
   return commandHistory;
}

void Scene::addMacro(CommandObjMacro* macro, std::string name) {
   if (attachments->getData("dictionary") == NULL) {
      AttachedDictionaryObject* attachedDic = new AttachedDictionaryObject();
      attachedDic->setup("dictionary", AttachedData::ADF_PERSISTENT);
      attachments->attachData(attachedDic);
   }

   AttachedDictionaryObject* attachedDic =
     (AttachedDictionaryObject*)attachments->getData("dictionary");

   attachedDic->getObjectDictionary()->addObject(macro, name);
}

Scene::~Scene() {
   if (rootNode != NULL) rootNode->releaseReference(NULL);
   delete attachments;
	// delete state last because nodes may still contain references (e.g. to allocators)
   delete rootState;
	delete statistics;
}


void X4::Scene::activateMyStatistics()
{
	stats = statistics;
}

void X4::Scene::rebuildRenderObjects()
{
	for (mpcard i = 0; i < glMaterials.size(); i++)
	{
		glMaterials[i]->rebuildRenderObjects() ;
	}
}
