//---------------------------------------------------------------------------
#ifndef SceneGraphIteratorsH
#define SceneGraphIteratorsH
//---------------------------------------------------------------------------
#include "x3sceneGraph.h"
//---------------------------------------------------------------------------


namespace X4 {

class SceneGraphNode;
class SceneGraphState;


typedef card32 SceneGraphIteratorType;
static const card32 SGITCAP_basic = 1;
static const card32 SGITCAP_indexed = 2;
static const card32 SGITCAP_byName = 4;


class X3SCENEGRAPH_API SceneGraphIterator : public Object {
 private:
   static MetaClass *SceneGraphIteratorMetaClass;
   static Object* newInstance();

 public:
   virtual const MetaClass *getInstanceClass() const;
   static const MetaClass *getClass();
   static void init(const MetaClass *parentClass);
   static void shutdown();
};


/// interface that allows to browse through scene graph nodes and
/// retrieve the nodes along with the scene graph state at the current node
class SGBasicNodesIterator { /* INTERFACE CLASS */
 public:
   /// go to next node in set
   virtual void next() = 0;
   /// true if end-of-set
   virtual bool atEnd() = 0;

   /// get current node
   virtual SceneGraphNode *getNode() = 0;
   /// get parent of current node (NULL if this is the root of the set)
   /// this method is useful, e.g. for removing objects from the parent object
   virtual SceneGraphNode *getParent() = 0;
   /// get current state
   virtual SceneGraphState *getState() = 0;
   /// get fully qualified name, i.e. name of node and path to the node through the graph
   virtual std::string getNameWithPath() = 0;
};


/// interface that allows accessing sets of scene graph nodes by index
class SGIndexedNodesIterator { /* INTERFACE CLASS */
 public:
   /// get number of nodes in indexed set
   virtual max_platform_card getNumNodes() = 0;
   /// go to a node by its index (no effect if index is out of range)
   virtual void goTo(max_platform_card nodeIndex) = 0;
   /// get current index (useful if modified by another operation of another interface)
   virtual max_platform_card getCurrentIndex() = 0;
};


/// interface that allows accessing sets of scene graph nodes by their names/paths
class SGByNameNodesIterator { /* INTERFACE CLASS */
 public:
   /// go to a node identified by its name
   /// (use path delimiters "/" to access nodes nested in the hierarchy)
   /// returns false if no node with the specified name is in the set
   virtual bool goToName(const std::string name) = 0;
};


/// interface that allows hierarchical access to the scene graph
/// (i.e. state tracking is provided in addition to the interface already present in SceneGraphNode)
class SGHierarchicalIterator { /* INTERFACE CLASS */
 public:
   /// get the number of children of the current node
   virtual max_platform_card getNumChildren() = 0;
   /// go to child number i; result is true if the child exists
   /// otherwise the operation has no effect
   virtual bool goToChild(max_platform_card i) = 0;
   /// go to parent node. result is false if the current node has no parent.
   /// in that case, the operation has no effect
   virtual bool goToParent() = 0;
};



}


#endif
 