//---------------------------------------------------------------------------
#ifndef SELECTION_TOOLS_H
#define SELECTION_TOOLS_H
//---------------------------------------------------------------------------
#include "x3sceneGraph.h"
//---------------------------------------------------------------------------
#include "PointCloud.h"
#include "Scene.h"
#include "Camera.h"
#include "ViewFrustum.h"
#include "GeometricRange.h"
#include "IteratorAdaptors.h"
//---------------------------------------------------------------------------

namespace X4 {

	class RayConeRange;

	// identifiers used in PCISelectionTool and CmdObjSelection
	const card32 SELMODE_SINGLECLOUD = 0;
	const card32 SELMODE_SINGLEPOINT = 1;
	const card32 SELMODE_BRUSHCIRCLE = 2;
	const card32 SELMODE_BRUSHQUAD = 3;
	const card32 SELMODE_AREACIRCLE = 4;
	const card32 SELMODE_AREARECTANGLE = 5;

	const card32 BUTTON_INVERTPOINTS = 0;
	const card32 BUTTON_ALLCLOUDS = 1;
	const card32 BUTTON_ALLPOINTS = 2;
	const card32 BUTTON_NOCLOUDS = 3;
	const card32 BUTTON_NOPOINTS = 4;
	const card32 BUTTON_EXTENDTOCOMPONENTS = 5;


	/// check if the given point cloud is selected
	X3SCENEGRAPH_API bool testPointCloudIsSelected(Scene *scene, std::string PCName);


	// in the following "closest point" will be used for the point with shortest distance to ray axis.
	// "nearest point" will be used for the point being nearest to camera measured in ray-z-direction coordinates.

	/// returns the name with path of the cloud the user has clicked on or "" if no cloud was clicked
	X3SCENEGRAPH_API std::string  getClickedCloud(int32 x, int32 y, card32 w, card32 h, Camera* cam, ViewFrustum* viewfrust, Scene* scene);



	/// returns true if the closest point relative to mouse x/y on screen
	/// (within a defined circle) is found. also returns point position 
	/// and name of corresponding cloud. runs only through selected clouds!
	X3SCENEGRAPH_API bool  getClosestPointPositionFromSelectedClouds(int32 x, int32 y, card32 w, card32 h, card32 tolerance, Camera* cam, ViewFrustum* viewfrust, Scene* scene, Vector3f& returnPointPosition, std::string& returnCloudName);

	/// returns true if the closest point relative to mouse x/y on screen
	/// (within a defined circle) is found. also returns point position 
	/// and name of corresponding cloud. runs only through selected points!
	X3SCENEGRAPH_API bool  getClosestPointPositionFromSelectedPoints(int32 x, int32 y, card32 w, card32 h, card32 tolerance, Camera* cam, ViewFrustum* viewfrust, Scene* scene, Vector3f& returnPointPosition, std::string& returnCloudName);

	/// returns true if the closest point relative to mouse x/y on screen 
	/// (within a defined circle) is found. also returns point position 
	/// and name of corresponding cloud. runs through all points.
	X3SCENEGRAPH_API bool  getClosestPointPositionFromAllPoints(int32 x, int32 y, card32 w, card32 h, card32 tolerance, Camera* cam, ViewFrustum* viewfrust, Scene* scene, Vector3f& returnPointPosition, std::string& returnCloudName);

	/// returns copied iterator pointing to and name of corresponding cloud 
	/// of the nearest point within a defined circle around mouse x/y on 
	/// screen. runs through all points.
	X3SCENEGRAPH_API PointCloudIterator*  getClosestPointIteratorFromAllPoints(int32 x, int32 y, card32 w, card32 h, card32 tolerance, Camera *cam, ViewFrustum *viewfrust, Scene *scene, std::string& returnCloudName, bool onlySelectedClouds = false, bool onlySelectedPoints = false, Vector3f * retPointPtr=NULL);

   /// returns copied iterator pointing to and name of corresponding cloud 
   /// of the closest point relative to mouse x/y on screen (within a 
   /// defined circle). runs through all points.
  // X3SCENEGRAPH_API PointCloudIterator*  getClosestPointIteratorFromAllPoints(int32 x, int32 y, card32 w, card32 h, card32 tolerance, Camera *cam, ViewFrustum *viewfrust, Scene *scene, std::string& returnCloudName,Vector3f * retPointPtr=NULL);
   
   /// returns the nearest found z-value (negative) in ray coordinates 
   /// inside a defined circle around mouse x/y on screen. runs through all points.
   X3SCENEGRAPH_API float32  getNearestZValueFromAllPoints(int32 x, int32 y, card32 w, card32 h, card32 tolerance, Camera *cam, ViewFrustum *viewfrust, Scene *scene);
   
   /// returns the nearest found z-value (negative) in ray coordinates 
   /// inside a defined box around mouse x/y on screen. runs through all points.
   X3SCENEGRAPH_API float32  getNearestZValueFromAllPoints(int32 x1, int32 y1, int32 x2, int32 y2, card32 w, card32 h, Camera *cam, ViewFrustum *viewfrust, Scene *scene);


	X3SCENEGRAPH_API bool isSceneGraphNodeVisible( SceneGraphNode* sgn );

	/// returns copied iterator pointing to the closest point relative 
	/// to mouse x/y on screen (within a defined circle/box). also 
	/// returns point position, name of corresponding cloud and the 
	/// nearest found z-value in ray coordinate system. you can filter 
	/// the search-range by different switches. returned iterator has 
	/// to be deleted separately in every case after use!
	X3SCENEGRAPH_API PointCloudIterator*  getClosestPointAndNearestZ(bool onlySelectedClouds, bool onlySelectedPoints, bool returnIterator, GeometricRange* range, Scene *scene, Vector3f& returnPointPosition, std::string& returnCloudName, float32& returnedNearestZ, bool returnNearest=false);

	/// returns a cone GeometricRange pointing from camera origin in the 
	/// direction of the mouse click on screen with a defined aperture angle. 
	/// the returned GeometricRange has to be deleted seperately after use!
	X3SCENEGRAPH_API RayConeRange*  getRayConeRangeFromMouseClick(int32 x, int32 y, card32 w, card32 h, card32 tolerance, Camera* cam, ViewFrustum* viewfrust);

	/// returns a sphere GeometricRange centered near to the nearest point 
	/// towards camera in a defined circle around mouse click with squared 
	/// radius adapted to the tolerance circle on screen. the returned 
	/// GeometricRange has to be deleted seperately after use!
	X3SCENEGRAPH_API GeometricRange*  getNearestSphereRangeFromMouseClick(int32 x, int32 y, card32 w, card32 h, card32 tolerance, Camera* cam, ViewFrustum* viewfrust, Scene* scene);

	/// returns a pyramidal GeometricRange pointing from camera origin 
	/// in the direction of the mouse click on screen with defined 
	/// horizontal and vertical aperture angles. the returned GeometricRange 
	/// has to be deleted seperately after use!
	X3SCENEGRAPH_API GeometricRange*  getRayPyramidalRangeFromMouseClick(int32 x1, int32 y1, int32 x2, int32 y2, card32 w, card32 h, Camera* cam, ViewFrustum* viewfrust);

	/// returns a clipped pyramidal GeometricRange pointing from camera origin 
	/// in the direction of the mouse click on screen with defined horizontal 
	/// and vertical aperture angles and a near and far clipping plane defined 
	/// by the nearest point towards camera (this at all makes a frustum). 
	/// the returned GeometricRange has to be deleted seperately after use!
	X3SCENEGRAPH_API GeometricRange*  getNearestClippedRayPyramidalRangeFromMouseClick(int32 x1, int32 y1, int32 x2, int32 y2, card32 w, card32 h, Camera* cam, ViewFrustum* viewfrust, Scene *scene);


	/// returns new MHGeoRangeIteratorAdaptor, which has to be deleted 
	/// separately after use. Use this function whenever possible, due 
	/// to code must only be modified here for implementation changes
	X3SCENEGRAPH_API IteratorAdaptor*  getGeoRangeIteratorAdaptor(PointCloud* ,  PointCloudIterator* , GeometricRange *r);

	/// returns new MHSelectionIteratorAdaptor, which has to be deleted 
	/// separately after use. Use this function whenever possible, due 
	/// to code must only be modified here for implementation changes
	X3SCENEGRAPH_API IteratorAdaptor*  getSelectionIteratorAdaptor(PointCloud* , PointCloudIterator* );

   
   
   /// Describe me!
   X3SCENEGRAPH_API Matrix3f  createCovariance(PointCloud* p, const Matrix4f ctm );

	/// Describe me!
	X3SCENEGRAPH_API Matrix3f  createCovarianceAndMean(PointCloud* pc, IteratorAdaptor* it, Vector3f &mean, card32 &numPoints, const Matrix4f ctm );

	/// Rotate u and v from local coordSystem to match cam up and right
	X3SCENEGRAPH_API void  considerCamPos( Scene* scene, Vector3f &u,Vector3f &v,Vector3f &n );

	/// return the selected nodes, but order them in the ordering of the scene graph
	X3SCENEGRAPH_API std::vector<std::string> getSelectedNodeNames( Scene* scene, bool ordered = true );

	X3SCENEGRAPH_API std::vector<PointCloud*> getSelectedPointClouds( Scene* scene, bool ordered = true );

	/// after changling the selection status of vertices using direct access with flags (specially with 
	/// VertArray) the selection is not updated in the view. This function updates it.
	/// call for each pointcloud and then
	///     scene->rebuildRenderObjects();
	///     ObjectViewsTable::update(scene);
	X3SCENEGRAPH_API void updateVertexSelection( PointCloud* pc );

	template<typename Base> 
	inline std::vector<Base*> getSelectedPointCloudsAs(Scene* scene, bool ordered = true) { 
		std::vector<PointCloud*> pcs = getSelectedPointClouds(scene, ordered); 
		std::vector<Base*> ret;
		ret.reserve(pcs.size());
		for( int i1 = 0; i1 < (int)pcs.size(); ++ i1 ) {
			Base* b = dynamic_cast<Base*>(pcs[i1]);
			if( b ) ret.push_back(b);
		}
		return ret; 
	}

	template<typename Base> 
	inline std::vector<std::pair< std::string, Base* > > getSelectedPointCloudsWithNameAs(Scene* scene, bool ordered = true) { 
		std::vector<std::string> sel = getSelectedNodeNames(scene, ordered);
		std::vector<std::pair< std::string, Base* > > ret;

		for( int i1 = 0; i1 < sel.size(); ++ i1 ) {
			Base* obj = dynamic_cast<Base*>( getPointCloud(scene, sel[i1]) ); 
			if( obj ) 
				ret.push_back(std::make_pair(sel[i1], obj) );
		}
		return ret; 
	}


	inline PointCloud* getSingleSelectedPointCloud( Scene* scene ) {
		std::vector<PointCloud*> ret(getSelectedPointClouds(scene, true));
		return ret.size() >= 1 ? ret.front() : NULL;
	}

	template<typename Base> 
	inline Base* getSingleSelectedPointCloudAs(Scene* scene) { 
		std::vector<Base*> ret(getSelectedPointCloudsAs<Base>(scene, true));
		return ret.size() >= 1 ? ret.front() : NULL;
	}

}

#endif
