//---------------------------------------------------------------------------
#include "stdafx.h"

//---------------------------------------------------------------------------
#include "StandardSceneGraphIterator.h"
//---------------------------------------------------------------------------
#include "SGStateChanger.h"
#include "SceneGraphNode.h"
//---------------------------------------------------------------------------
 
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------



MetaClass* StandardSceneGraphIterator::StandardSceneGraphIteratorMetaClass = NULL;

const MetaClass* StandardSceneGraphIterator::getInstanceClass() const {
   return getClass();
}

Object* StandardSceneGraphIterator::newInstance() {
   return new StandardSceneGraphIterator;
}

const MetaClass* StandardSceneGraphIterator::getClass() {
   if (StandardSceneGraphIteratorMetaClass==NULL) {
      StandardSceneGraphIteratorMetaClass = new MetaClass("StandardSceneGraphIterator", "no description.",
                                  newInstance, sizeof(StandardSceneGraphIterator));
   }
   return StandardSceneGraphIteratorMetaClass;
}

void StandardSceneGraphIterator::init(const MetaClass *parentClass) {
   ((MetaClass*)getClass())->init(parentClass);
   StandardSceneGraphIteratorMetaClass->copyParentProperties();
}

void StandardSceneGraphIterator::shutdown() {
   delete StandardSceneGraphIteratorMetaClass;
}


// ----


StandardSceneGraphIterator::StandardSceneGraphIterator() {
   currentState = NULL;
   originalState = NULL;
   endOfTraversal = false;
}

void StandardSceneGraphIterator::clear() {
   const size_t n = entriesStack.size();
   for (card32 i=0; i<n; i++) {
      delete entriesStack[i].childEntryStateChange;
      delete entriesStack[i].nodeEntryStateChange;
   }
   entriesStack.clear();
   delete currentState;
   currentState = NULL;
   originalState = NULL;
   includeNodeFlags = 0;
   excludeNodeFlags = 0;
   endOfTraversal = false;
}

void StandardSceneGraphIterator::applyStateChanges(SceneGraphState *state, Entry &e) {
   if (e.childEntryStateChange != NULL) {
      e.childEntryStateChange->doChange(state);
   }
   if (e.nodeEntryStateChange != NULL) {
      e.nodeEntryStateChange->doChange(state);
   }
}

void StandardSceneGraphIterator::undoStateChanges(SceneGraphState *state, Entry &e) {
   if (e.nodeEntryStateChange != NULL) {
      e.nodeEntryStateChange->undoChange(state);
   }
   if (e.childEntryStateChange != NULL) {
      e.childEntryStateChange->undoChange(state);
   }
}

bool StandardSceneGraphIterator::internProcessNode(SceneGraphNode *n) {
   card32 flags = n->getFlags();
   return    processNode(n)
          && ((flags & includeNodeFlags) == includeNodeFlags)
          && ((flags & excludeNodeFlags) == 0);
}

void StandardSceneGraphIterator::clearAndSetup(SceneGraphNode *rootNode, const SceneGraphState *rootState,
   SGDynamicState::TraversalModeType traversalMode, card32 includeNodeFlags, card32 excludeNodeFlags)
{
	if (traversalMode==SGDynamicState::tmDefault) {
		traversalMode = (SGDynamicState::TraversalModeType)rootState->staticState->defaultTraversalMode;
	}

   clear();

   originalState = rootState;
   currentState = (const_cast<SceneGraphState*>(rootState)->createTraversalState());
   currentState->dynamicState->traversalMode = traversalMode;
   this->includeNodeFlags = includeNodeFlags;
   this->excludeNodeFlags = excludeNodeFlags;

   if (internProcessNode(rootNode)) {
      Entry e;
      e.node = rootNode;
      e.childEntryStateChange = NULL;
      e.nodeEntryStateChange = rootNode->getNodeEntryStateChange(currentState);
      e.subnodeIndex = 0;
      entriesStack.push_back(e);
      endOfTraversal = false;
   } else {
      endOfTraversal = true;
   }
}

void StandardSceneGraphIterator::next() {
   if (endOfTraversal) return;

   SceneGraphNode *currentNode = entriesStack.back().node;

   if (currentNode->getNumChildNodes(currentState) > 0) {

      SceneGraphNode *nextNode = currentNode->getChildNode(currentState, 0);

      Entry e;
      e.node = nextNode;
      e.childEntryStateChange = currentNode->getChildNodeEntryStateChange(currentState, 0);
      if (e.childEntryStateChange != NULL) {
         e.childEntryStateChange->doChange(currentState);
      }
      e.nodeEntryStateChange = nextNode->getNodeEntryStateChange(currentState);
      if (e.nodeEntryStateChange != NULL) {
         e.nodeEntryStateChange->doChange(currentState);
      }
      e.subnodeIndex = 0;
      entriesStack.push_back(e);

   } else {

      bool found;
      do {
         Entry oldTopEntry = entriesStack.back();
         entriesStack.pop_back();

         if (oldTopEntry.nodeEntryStateChange != NULL) {
            oldTopEntry.nodeEntryStateChange->undoChange(currentState);
         }
         if (oldTopEntry.childEntryStateChange != NULL) {
            oldTopEntry.childEntryStateChange->undoChange(currentState);
         }

		   delete oldTopEntry.childEntryStateChange;
	      delete oldTopEntry.nodeEntryStateChange;

         if (entriesStack.empty()) {
            endOfTraversal = true;
            return;
         }

         Entry *currentTopEntry = &(entriesStack.back());

         currentTopEntry->subnodeIndex++;
         found = false;

         if (   internProcessNode(currentTopEntry->node)
             && currentTopEntry->subnodeIndex < currentTopEntry->node->getNumChildNodes(currentState))
         {
            Entry nextE;
            nextE.node = currentTopEntry->node->getChildNode(currentState, currentTopEntry->subnodeIndex);
            nextE.childEntryStateChange = currentTopEntry->node->getChildNodeEntryStateChange(currentState, currentTopEntry->subnodeIndex);
            if (nextE.childEntryStateChange != NULL) {
               nextE.childEntryStateChange->doChange(currentState);
            }
            nextE.nodeEntryStateChange = nextE.node->getNodeEntryStateChange(currentState);
            if (nextE.nodeEntryStateChange != NULL) {
               nextE.nodeEntryStateChange->doChange(currentState);
            }
            nextE.subnodeIndex = 0;
            entriesStack.push_back(nextE);

            found = true;
         }


      } while (!found);
   }
}

bool StandardSceneGraphIterator::atEnd() {
   return endOfTraversal;
}

SceneGraphNode *StandardSceneGraphIterator::getNode() {
   x4Assert(!entriesStack.empty());
   return entriesStack.back().node;
}

SceneGraphNode *StandardSceneGraphIterator::getParent() {
   const max_platform_card n = entriesStack.size();
   if (n >= 2) {
      return entriesStack[n-2].node;
   } else {
      return NULL;
   }
}

SceneGraphState *StandardSceneGraphIterator::getState() {
   return currentState;
}


std::string StandardSceneGraphIterator::getNameWithPath() {
   std::string result;
   size_t n = entriesStack.size();
   for (card32 i=0; i<n; i++) {
      if (i>0) result += "/";
      result += entriesStack[i].node->getName();
   }
   return result;
}

max_platform_card StandardSceneGraphIterator::getNumChildren() {
   return entriesStack.back().node->getNumChildNodes(currentState);
}

bool StandardSceneGraphIterator::goToChild(max_platform_card i) {
   if (getNumChildren() <= i) return false;

   SceneGraphNode *currentNode = entriesStack.back().node;
   SceneGraphNode *nextNode = currentNode->getChildNode(currentState, i);
   x4Assert(nextNode != NULL);

   Entry e;
   e.node = nextNode;
   e.childEntryStateChange = currentNode->getChildNodeEntryStateChange(currentState, i);
   if (e.childEntryStateChange != NULL) {
      e.childEntryStateChange->doChange(currentState);
   }
   e.nodeEntryStateChange = nextNode->getNodeEntryStateChange(currentState);
   if (e.nodeEntryStateChange != NULL) {
      e.nodeEntryStateChange->doChange(currentState);
   }
   e.subnodeIndex = 0;
   entriesStack.push_back(e);

   /// using the hierarchical interface voids the basic interface
   endOfTraversal = true;

   return true;
}

bool StandardSceneGraphIterator::goToParent() {
   if (entriesStack.size() <= 1) return false;

   Entry oldTopEntry = entriesStack.back();
   entriesStack.pop_back();

   if (oldTopEntry.nodeEntryStateChange != NULL) {
      oldTopEntry.nodeEntryStateChange->undoChange(currentState);
   }
   if (oldTopEntry.childEntryStateChange != NULL) {
      oldTopEntry.childEntryStateChange->undoChange(currentState);
   }

   delete oldTopEntry.childEntryStateChange;
   delete oldTopEntry.nodeEntryStateChange;

   /// using the hierarchical interface voids the basic interface
   endOfTraversal = true;

   return true;
}



StandardSceneGraphIterator::~StandardSceneGraphIterator() {
   clear();
}
