///////////////////////////////////////////////////////////////////////////////
//
//	AR4CachedLocalGeodesics.cpp
//
//	class for local geodesic computation
//
//	Alan Brunton, December 2012
//
///////////////////////////////////////////////////////////////////////////////


#include "stdafx.h"

#include "AR4CachedLocalGeodesics.h"

#include "TopologyAttachment.h"
#include "TopologicalKNNIterator.h"

#include "StringHelper.h"
#include "Timer.h"

#include "DynamicLinearAlgebra.h"

#include "Monomials2ndOrder.h"


void X4::AR4CachedLocalGeodesics::cacheGeodesics(X4::mpcard startIndex, const vector<mpcard>& destIndices)
{
	Timer timer;

	mpcard i;
	vector<unsigned> copyIndices;
	vector<float> dists;
	vector<Vector3f> gradients;

	PointSet* pslocal = NULL;
	PointSet* pssource = NULL;

	AAT possrc, posloc, idxloc;

	if (m_pSourcePC == NULL)
	{
		throw ERangeCheck("AR4CachedLocalGeodesics::cacheGeodesics(...): null source point cloud");
	}

	pssource = m_pSourcePC->getPointSet();

   // create topology iterator
	TopologyAttachment* tatt = dynamic_cast<TopologyAttachment*>(m_pSourcePC->getAttachments()->getData(TopologyAttachment::getDefaultName()));
	InCorePCTopologyGraph* tpg = tatt->getTopology(); 
	PointCloudIterator* top_it = tpg->createIterator(m_pSourcePC->getPointSet());


	if (m_cache.size() == 0)
	{
		m_cache.resize(m_pSourcePC->getNumPoints());
	}

	if (m_pLocalPC == NULL)
	{
		m_pLocalPC = new UnstructuredInCorePointCloud();
		m_pLocalPC->clearAndSetup(m_pSourcePC->getDescr(), 1);
		pslocal = m_pLocalPC->getPointSet();
		pslocal->setVertex(0, pssource->getDataPointer(startIndex));
	}
	else
	{
		pslocal = m_pLocalPC->getPointSet();
		pslocal->addPoint(pssource->getDataPointer(startIndex));
	}

	copyIndices.resize(destIndices.size());
	for (i = 0; i < destIndices.size(); i++)
		copyIndices[i] = (unsigned)destIndices[i];

	dists = computeGeodesicsEarlyStop(m_pSourcePC, top_it, m_sigma, startIndex, destIndices, gradients);

	possrc = m_pSourcePC->getAAT("position");
	posloc = m_pLocalPC->getAAT("position");
	idxloc = m_pLocalPC->getAAT(m_strIndex);

	for (i = 0; i < destIndices.size(); i++)
	{
		CacheEntry entry;
		entry.first = destIndices[i];
		entry.second.m_dist = dists[i];
		entry.second.m_grad = gradients[i];
		m_cache[startIndex].push_back(entry);
	}

	//sort indices for faster checking if a geodesic is cached
	std::sort<vector<CacheEntry>::iterator>(m_cache[startIndex].begin(), m_cache[startIndex].end(), compareIndices);

	delete m_pHknnIt;

	delete top_it;
}


float X4::AR4CachedLocalGeodesics::geodesic_alt(const Vector3f &from, const Vector3f &to, float sigma/* = 1.000000*/, Vector3f &gradient/* = Vector3f()*/, card32 numNeighbours/* = 4*/)
{


	card32 i, j, k;
	float d, dist = 0.f;
	Vector3f grad = makeVector3f(0.f, 0.f, 0.f);
	card64 currentIndex;
	vector<card64> startIndices, endIndices;
	vector<Vector3f> startPositions, endPositions;
	vector<float> endDistances;
	vector<Vector3f> endGradients;
	vector<float> weights;
	float w, weightSum;
	float rcpSigma2 = 1.f / (sigma*sigma);
	AAT idx = m_pLocalPC->getAAT(m_strIndex);

	startIndices.reserve(numNeighbours);
	endIndices.reserve(numNeighbours);
	startPositions.reserve(numNeighbours);
	endPositions.reserve(numNeighbours);
	endDistances.reserve(numNeighbours);
	endGradients.reserve(numNeighbours);
	weights.reserve(numNeighbours);

	AAT pos = m_pLocalPC->getAAT("position");
	if (m_pHknnIt == NULL)
		throw ERangeCheck("AR4CachedGeodesics::geodesic(...): no geodesics cached");

	m_pHknnIt->setSeekPointAndReset(from);
	for (i = 0; i < numNeighbours; i++)
	{
		while (!isCached(m_pHknnIt->get1u(idx)) && !m_pHknnIt->atEnd())
			m_pHknnIt->next();
		if (m_pHknnIt->atEnd())
			break;
		startIndices.push_back(m_pHknnIt->get1u(idx));
		startPositions.push_back(m_pHknnIt->get3f(pos));
		m_pHknnIt->next();
	}

	if (startIndices.size() == 0)
		throw ERangeCheck("X4::AR4CachedLocalGeodesics::geodesic(...): no cached start indices found");

	bool bcached;
	m_pHknnIt->setSeekPointAndReset(to);
	for (i = 0; i < numNeighbours; i++)
	{
		bcached = false;
		while (!bcached && !m_pHknnIt->atEnd())
		{
			bcached = true;
			for (j = 0; j < startIndices.size(); j++)
			{
				if (!isCached(startIndices[j], (mpcard)m_pHknnIt->get1u(idx)))
				{
					bcached = false;
					m_pHknnIt->next();
					break;
				}
			}
		}
		if (m_pHknnIt->atEnd())
			break;
		endIndices.push_back(m_pHknnIt->get1u(idx));
		endPositions.push_back(m_pHknnIt->get3f(pos));
		m_pHknnIt->next();
	}

	if (endIndices.size() == 0)
		throw ERangeCheck("X4::AR4CachedLocalGeodesics::geodesic(...): no cached end indices found");

	for (j = 0; j < endIndices.size(); j++)
	{
		weights.clear();
		weightSum = 0.f;
		for (i = 0; i < startIndices.size(); i++)
		{
			d = (startPositions[i] - from).getSqrNorm();
			w = exp(-d * rcpSigma2);
			weights.push_back(w);
			weightSum += w;
		}

		dist = 0.f;
		grad.setZero();
		for (i = 0; i < startIndices.size(); i++)
		{
			k = findCacheIndex(startIndices[i], endIndices[j]);
			dist += m_cache[startIndices[i]][k].second.m_dist * weights[i];
			grad += m_cache[startIndices[i]][k].second.m_grad * weights[i];
		}
		dist /= weightSum;
		grad /= weightSum;

		endDistances.push_back(dist);
		endGradients.push_back(grad);
	}

	dist = 0.f;
	grad.setZero();
	weights.clear();
	weightSum = 0.f;
	for (j = 0; j < endIndices.size(); j++)
	{
		d = (endPositions[j] - to).getSqrNorm();
		w = exp(-d * rcpSigma2);
		weights.push_back(w);
		weightSum += w;
	}
	for (j = 0; j < endIndices.size(); j++)
	{
		dist += endDistances[j] * weights[j];
		grad += endGradients[j] * weights[j];
	}
	dist /= weightSum;
	grad /= weightSum;

	gradient = grad;

	if (dist > 1e10 || dist < -1e10)
	{
		debugOutput << "exploded geodesic (" << dist << ") " << from << " --> " << to << "\n";
		debugOutput << "num start indices: " << startIndices.size() << "; num end indices: " << endIndices.size() << "\n";
		debugOutput << "weight sum: " << weightSum << "\n";
		for (j = 0; j < endIndices.size(); j++)
			debugOutput << endPositions[j] << "\n";
	}

	return dist;
}

float X4::AR4CachedLocalGeodesics::geodesic(const Vector3f &from, const Vector3f &to, float sigma/* = 1.000000*/, Vector3f &gradient/* = Vector3f()*/, card32 numNeighbours/* = 4*/)
{
	const mpcard NO_VALID_INDEX = (mpcard)-1;
	const float WEIGHT_EPSILON = 1e-8f;

	card32 i, j, k;
	float d, dist = 0.f;
	Vector3f grad = makeVector3f(0.f, 0.f, 0.f);
	card64 currentIndex;
	vector<card64> startIndices, endIndices;
	vector<Vector3f> startPositions, endPositions;
	vector<float> endDistances;
	vector<Vector3f> endGradients;
	vector<float> weights, weightsEnd;
	float w, weightSum, rcpWeightSum, weightStartSum;
	float rcpSigma2 = 1.f / (sigma*sigma);
	AAT idx = m_pLocalPC->getAAT(m_strIndex);

	startIndices.resize(numNeighbours, NO_VALID_INDEX);
	endIndices.resize(numNeighbours, NO_VALID_INDEX);
	startPositions.resize(numNeighbours);
	endPositions.resize(numNeighbours);
	endDistances.resize(numNeighbours);
	endGradients.resize(numNeighbours);
	weights.resize(numNeighbours, 0.f);
	weightsEnd.resize(numNeighbours, 0.f);


	DynamicVector<float> W(numNeighbours), D(numNeighbours), Gend(3);
	DynamicMatrix<float> G(numNeighbours, 3);

	W.setZero();
	D.setZero();
	G.setZero();

	AAT pos = m_pLocalPC->getAAT("position");

	if (m_pHknnIt == NULL)
		throw ERangeCheck("AR4CachedGeodesics::geodesic(...): no geodesics cached");

	m_pHknnIt->setSeekPointAndReset(from);
	weightSum = 0.f;
	for (i = 0; i < numNeighbours; i++)
	{
		while (!isCached((mpcard)m_pHknnIt->get1u(idx)) && !m_pHknnIt->atEnd())
			m_pHknnIt->next();
		if (m_pHknnIt->atEnd())
			break;
		startIndices[i] = m_pHknnIt->get1u(idx);
		startPositions[i] = m_pHknnIt->get3f(pos);
		d = (startPositions[i] - to).getSqrNorm();
		w = max(WEIGHT_EPSILON, exp(-d * rcpSigma2));
		weights[i] = w;
		weightSum += w;
		m_pHknnIt->next();
	}

	if (i == 0)
		throw ERangeCheck("X4::AR4CachedLocalGeodesics::geodesic(...): no cached start indices found");

	weightStartSum = weightSum;
	rcpWeightSum = 1.f / weightSum;
	for (i = 0; i < numNeighbours; i++)
		weights[i] *= rcpWeightSum;

	bool bcached;
	m_pHknnIt->setSeekPointAndReset(to);
	weightSum = 0.f;
	for (i = 0; i < numNeighbours; i++)
	{
		bcached = false;
		while (!bcached && !m_pHknnIt->atEnd())
		{
			d = (m_pHknnIt->get3f(pos) - to).getSqrNorm();
			w = max(WEIGHT_EPSILON, exp(-d * rcpSigma2));
			weightsEnd[i] = w;
			W[i] = w;
			weightSum += w;
			bcached = false;
			for (j = 0; j < startIndices.size(); j++)
			{
				if (startIndices[j] == NO_VALID_INDEX)
					break;
				if (isCached(startIndices[j], (mpcard)m_pHknnIt->get1u(idx)))
				{
					bcached = true;
					endIndices[i] = m_pHknnIt->get1u(idx);
					endPositions[i] = m_pHknnIt->get3f(pos);
					k = findCacheIndex(startIndices[j], endIndices[i]);
					D[j] = m_cache[startIndices[j]][k].second.m_dist * weights[j];
					G[j][0] = m_cache[startIndices[j]][k].second.m_grad[0] * weights[j];
					G[j][1] = m_cache[startIndices[j]][k].second.m_grad[1] * weights[j];
					G[j][2] = m_cache[startIndices[j]][k].second.m_grad[2] * weights[j];
				}
			}
			if (!bcached)
				m_pHknnIt->next();
		}
		if (m_pHknnIt->atEnd())
			break;

		m_pHknnIt->next();
	}

	if (i == 0)
		throw ERangeCheck("X4::AR4CachedLocalGeodesics::geodesic(...): no cached end indices found");

	rcpWeightSum = 1.f / weightSum;
	W *= rcpWeightSum;
	
	dist = D * W;
	Gend = G * W;
	gradient[0] = Gend[0];
	gradient[1] = Gend[1];
	gradient[2] = Gend[2];

	if (dist > 1e10 || dist < -1e10)
	{
		debugOutput << "exploded geodesic (" << dist << ") " << from << " --> " << to << "\n";
		debugOutput << "num start indices: " << startIndices.size() << "; num end indices: " << endIndices.size() << "\n";
		debugOutput << "weight sum: " << weightSum << "weight sum start: " << weightStartSum << "\n";
		for (j = 0; j < endIndices.size(); j++)
			debugOutput << endPositions[j] << "\n";
	}

	return dist;
}

bool X4::AR4CachedLocalGeodesics::isCached(mpcard sourceIndex)
{
	bool banswer = false;
	banswer = (sourceIndex < m_cache.size() && m_cache[sourceIndex].size() > 0);

	return banswer;
}

bool X4::AR4CachedLocalGeodesics::isCached(mpcard startIndex, mpcard destIndex)
{
	mpcard i;

	CacheEntry temp;
	temp.first = destIndex;
	if (startIndex > m_cache.size())
		throw ERangeCheck("AR4CachedLocalGeodesics::isCached(...): startIndex out of range");
	return std::binary_search<vector<CacheEntry>::iterator, CacheEntry>(m_cache[startIndex].begin(), m_cache[startIndex].end(), temp, compareIndices);
}

X4::mpcard X4::AR4CachedLocalGeodesics::findCacheIndex(mpcard startIndex, mpcard destIndex)
{
	CacheEntry temp;
	pair<vector<CacheEntry>::iterator, vector<CacheEntry>::iterator> range;
	temp.first = destIndex;
	range = std::equal_range<vector<CacheEntry>::iterator, CacheEntry>(m_cache[startIndex].begin(), m_cache[startIndex].end(), temp, compareIndices);
	return (mpcard)(range.first - m_cache[startIndex].begin());
}

class IndexDist
{
public:
	X4::PointIndex	m_index;
	float				m_dist;

	IndexDist(): m_index(0), m_dist(0.f)
	{
	}
	IndexDist(X4::PointIndex index, float dist): m_index(index), m_dist(dist) 
	{
	}

	IndexDist& operator=(const IndexDist& id)
	{
		m_index = id.m_index;
		m_dist = id.m_dist;
		return *this;
	}
	inline bool operator<(const IndexDist& id) const
	{
		return (m_dist > id.m_dist);
	}
};

//static function for geodesics with early exit once all destinations have been computed
vector<float> X4::AR4CachedLocalGeodesics::computeGeodesicsEarlyStop(PointCloud * pc, PointCloudIterator * topologyIter, float sigma, const mpcard startIndex, const vector<mpcard> &indices, vector<Vector3f> &derivates, vector<vector<mpcard>>* paths/* = NULL*/)
{
	TopologyIterator * tIter = dynamic_cast<TopologyIterator*>(topologyIter);
	x4Assert( tIter );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(topologyIter);
	RandomAccessPointCloudIterator * rIt = dynamic_cast<RandomAccessPointCloudIterator*>(topologyIter);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(topologyIter);
	AAT posAAT = pc->getAAT( "position" );

	SceneObjectIterator * iter = pc->createIterator( SceneObjectIteratorType::CAP_BASIC_PC | 
		SceneObjectIteratorType::CAP_RANDOM_ACCESS_PC );
	x4Assert( iter != NULL );
	BasicPointCloudIterator * neighbour_bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	RandomAccessPointCloudIterator * neighbour_rIt = dynamic_cast<RandomAccessPointCloudIterator*>(iter);

	float rad_2 = pow(sigma * 2.0f,2.0f);

	// -----------------------------------------
	// distance transformation for the complete point cloud
	vector<float> distanceField( pc->getNumPoints(), 1e30f );
	derivates.resize(indices.size());
	HierarchicalKNNIterator hIt(pc,32,NULL);

	priority_queue<IndexDist> indexQueue;
	indexQueue.push(IndexDist(startIndex, 0.f));
	distanceField[startIndex] = 0.0f;

	card32 numNeighbors;

	mpcard dest, nbr, numMustVisit = 0;
	vector<mpcard> destNeighbors;
	vector<bool> mustVisit(pc->getNumPoints(), false);
	for (dest = 0; dest < indices.size(); dest++)
	{
		if (mustVisit[indices[dest]])
			continue;

		mustVisit[indices[dest]] = true;
		numMustVisit++;

		rIt->goToIndex(indices[dest]);
		numNeighbors = tIter->getNumNeighbours();
		for (nbr = 0; nbr < numNeighbors; nbr++)
		{
			if (mustVisit[tIter->getNeighbour(nbr)])
				continue;
			mustVisit[tIter->getNeighbour(nbr)] = true;
			numMustVisit++;
		}
	}

	while( !indexQueue.empty() && numMustVisit > 0)
	{
		PointIndex index;
		float currentDistance;
		IndexDist inddist = indexQueue.top();
		indexQueue.pop();	// remove first element

		index = inddist.m_index;

		if (mustVisit[index])
		{
			mustVisit[index] = false;
			numMustVisit--;
		}

		rIt->goToIndex( index );
		Vector3f pos = bIt->get3f(posAAT);
		currentDistance = inddist.m_dist;

		numNeighbors = tIter->getNumNeighbours();
		for( card32 i=0;i<numNeighbors;i++ )
		{
			PointIndex neighbour = tIter->getNeighbour(i);

			if (currentDistance < distanceField[neighbour])
			{
				neighbour_rIt->goToIndex(neighbour);
				Vector3f neighbour_pos = neighbour_bIt->get3f(posAAT);

				float distance = max(1e-20f, norm( neighbour_pos - pos )) + currentDistance;
				if( distanceField[neighbour] > distance )
				{
					distanceField[neighbour] = distance;
					indexQueue.push(IndexDist(neighbour, distance));
				}
			}
		}
	}

	if( paths != NULL )
	{
		paths->clear();
	}

	// get distances from distance field
	vector<float> distances( indices.size() );
	float inv_sigma_2 = 1.0f / sigma / sigma;
	for( card32 i=0;i<indices.size();i++ )
	{
		PointIndex index = indices[i];
		rIt->goToIndex(index);
		Vector3f currentPoint = bIt->get3f(posAAT);

		distances[i] = distanceField[index];

		// no path leads to this point
		if( distanceField[index] >= 1e20f )
		{
			distances[i] = 1e20f;
			derivates[i] = NULL_VECTOR3F;
			continue;
		}
      
		// compute a tangent plane
		Vector3f center;
		Matrix3f frame;
		computeLocalFrame(currentPoint, sigma, pc, &hIt, frame, center);
		const Matrix3f worldToFrame = frame.transpose();

		// gather neighbor points and transform into tangent space
		// height is set to distance, ignores outliers with infinity distance
		hIt.setSeekPointAndReset(currentPoint);
		vector<Vector3f> tangentSpacePoints;
		while( !hIt.atEnd())
		{
			if( distanceField[hIt.getCurrentPointIndex()] >= 1e20f)
			{
			   hIt.next();
				continue;
			}
			Vector3f tmpPos = hIt.get3f(posAAT);
			float dist_2 = normQuad(tmpPos-currentPoint);
			if( dist_2 > rad_2)
				break;

			tmpPos = worldToFrame * (tmpPos-currentPoint);
			tmpPos[2] = distanceField[hIt.getCurrentPointIndex()];
			tangentSpacePoints.push_back(tmpPos);

			hIt.next();
		}

		// compute first order fit
		Vector3f coeffs;
		Monomials1stOrder::computeCoeffs(tangentSpacePoints, coeffs, sigma*sigma);
		
		// set slightly more precise distance
		distances[i] = distanceField[index];//coeffs[0];

		// compute derivates
		Vector3f tangentSpaceDerivate = makeVector3f(coeffs[1],coeffs[2],0);
		derivates[i] = frame * tangentSpaceDerivate;

      
		if( paths != NULL )
		{
			vector<mpcard> path;
			float lastDistance = 1e30f;
			while( index != startIndex )
			{
				rIt->goToIndex( index );
				path.push_back(index);
				if( path.size() > 50000 )
					break;

				if( path.size()>1 )
					size_t s = path.size();
					
				float minDist = 1e30f;
				PointIndex bestNeighbour = tIter->getNeighbour(0);
				for( card32 j=0;j<tIter->getNumNeighbours();j++ )
				{
					PointIndex neighbour = tIter->getNeighbour(j);
					if( distanceField[neighbour] < minDist )
					{
						minDist = distanceField[neighbour];
						bestNeighbour = neighbour;
					}
				}
				if( minDist > lastDistance )
				{
					break;
				}
				lastDistance = minDist;

				index = bestNeighbour;
			}
			paths->push_back( path );

		}
	}


	delete iter;
	return distances;	
}


