///////////////////////////////////////////////////////////////////////////////
//
//	AR4CachedLocalGeodesics.h
//
//	class for maintaining cache of local geodesic distance and derivatives
//	for fast look-up and incremental computation
//
//	Alan Brunton, December 2012
//
///////////////////////////////////////////////////////////////////////////////


#ifndef __AR4CACHEDLOCALGEODESICS_H__
#define __AR4CACHEDLOCALGEODESICS_H__


#include "threeParameterMatchingGMOD14.h"

#include "InCorePCTopologyGraph.h"
#include "UnstructuredInCorePointCloud.h"
#include "GeometricFeature.h"
#include "TopologicalKNNIterator.h"
#include "TopologyRangeSearch.h"
#include "GeometricTools.h"

#include "StringHelper.h"


namespace X4
{
	class AR4CachedGeodesic
	{
	public:
		float								m_dist;
		Vector3f							m_grad;

		AR4CachedGeodesic(): m_dist(-1.f)
		{
			m_grad.setZero();
		}

		AR4CachedGeodesic& operator=(const AR4CachedGeodesic& cg)
		{
			m_dist = cg.m_dist;
			m_grad = cg.m_grad;
			return *this;
		}
	};

	class AR4CachedLocalGeodesics
	{
	private:


		typedef pair<mpcard, AR4CachedGeodesic>			CacheEntry;

		static bool compareIndices(CacheEntry e1, CacheEntry e2)	{ return e1.first < e2.first; }
		static bool compareDistances(CacheEntry e1, CacheEntry e2)	{ return e1.second.m_dist < e2.second.m_dist; }

		vector<vector<CacheEntry>>						m_cache;

		string											m_strIndex;

		UnstructuredInCorePointCloud*		m_pSourcePC;
		UnstructuredInCorePointCloud*		m_pLocalPC;
		HierarchicalKNNIterator*			m_pHknnIt;
		bool								m_bDelHknnIt;

		float								m_sigma;




	public:

		const static int32					m_GeodesicAvailableMask = 2;

		AR4CachedLocalGeodesics(HierarchicalKNNIterator* pHknnIt = NULL)
		{
			m_pSourcePC = NULL;
			m_pLocalPC = NULL;
			m_pHknnIt = pHknnIt;
			m_bDelHknnIt = (m_pHknnIt == NULL);
			m_sigma = 0.1f;
		}

		~AR4CachedLocalGeodesics()
		{
			delete m_pLocalPC;
			if (m_bDelHknnIt)
				delete m_pHknnIt;
		}

		void setSourcePC(UnstructuredInCorePointCloud* upc)
		{
			m_pSourcePC = upc;
		}
		UnstructuredInCorePointCloud* getSourcePC()
		{
			return m_pSourcePC;
		}

		void setSourceIndexString(string strIndex)	{ m_strIndex = strIndex; }
		string getSourceIndexString()				{ return m_strIndex; }

		void setSigma(float sigma)			{ m_sigma = sigma; }
		float getSigma()					{ return m_sigma; }

		bool isCached(mpcard sourceIndex);
		bool isCached(mpcard startIndex, mpcard destIndex);
		mpcard findCacheIndex(mpcard startIndex, mpcard destIndex);

		mpcard getNumCachedStartPoints()	{ return m_cache.size(); }

		void getCachedIndices(mpcard startIndex, vector<mpcard>& destIndices)
		{
			int i;
			destIndices.clear();
			if (startIndex < m_cache.size())
			{
				destIndices.resize(m_cache[startIndex].size());
				for (i = 0; i < destIndices.size(); i++)
					destIndices[i] = m_cache[startIndex][i].first;
			}
		}

		void getCachedGeodesic(mpcard startIndex, mpcard endIndex, float& dist, Vector3f& grad)
		{
			mpcard k;
			if (!isCached(startIndex, endIndex))
				throw ERangeCheck("AR4CachedLocalGeodesics::getCachedGeodesic(...): geodesic information not cached for start " + intToStr(startIndex) + " and end " + intToStr(endIndex));

			k = findCacheIndex(startIndex, endIndex);
			dist = m_cache[startIndex][k].second.m_dist;
			grad = m_cache[startIndex][k].second.m_grad;

		}

		void cacheGeodesics(mpcard startIndex, const vector<mpcard>& destIndices);

		float geodesic(const Vector3f& from, const Vector3f& to, float sigma = 1.0, Vector3f& gradient = Vector3f(), card32 numNeighbours = 4);
		float geodesic_alt(const Vector3f &from, const Vector3f &to, float sigma = 1.0, Vector3f &gradient = Vector3f(), card32 numNeighbours = 4);

		static vector<float> computeGeodesicsEarlyStop(PointCloud * pc, PointCloudIterator * topologyIter, float sigma, const mpcard startIndex, const vector<mpcard> &indices, vector<Vector3f> &derivates, vector<vector<mpcard>>* paths = NULL);
	};
};


#endif //__AR4CACHEDLOCALGEODESICS_H__


