//---------------------------------------------------------------------------
#include "stdafx.h"
//---------------------------------------------------------------------------
#include "AR4DijkstraDistPC3f.h"
//---------------------------------------------------------------------------
#include "GeometricTools.h"
#include "AR4MLSManifold.h"
#include "InCorePCTopologyGraph.h"
#include <float.h>
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------


AR4DijkstraDistPC3f::AR4DijkstraDistPC3f(AR4MLSManifold3f *manifold)
	:AR4FuncOnManifold<AR4MLSManifold3f, float32>(manifold)
	,maxDist(0)
{}


void AR4DijkstraDistPC3f::compute(card32 startIndex)
{
    const mpcard n = getManifold()->getNumDiscrPoints();
    vector<Vector3f> points(n);
    for (mpcard i=0; i<n; i++)
    {
        points[i] = getManifold()->getDiscrPoint(i);
    }

    PointCloudIterator *top_it = getManifold()->getTopology()->createIterator(getManifold()->getPointSet());

    distances = computeGeodeticDistances( getManifold()->getPointCloud(), top_it, getManifold()->getDiscrPoint(startIndex), points, NULL);

    delete top_it;

	updateMinMax();
}


///Small helper for computing the distances on the mesh
struct DistanceQueuePoint
{
	DistanceQueuePoint(const mpcard i, const mpcard cameFrom, const float32 ld)
		:index(i)
		,cameFromIndex(cameFrom)
		,GeoDistance(ld)
	{}

	inline bool operator<(const DistanceQueuePoint& other) const
	{
		return (GeoDistance > other.GeoDistance);
	}

	mpcard index;
	mpcard cameFromIndex;
	float32 GeoDistance;
};


void AR4DijkstraDistPC3f::computeUpToDistance(const card32 startIndex, const float32 DesiredDistance)
{
	const mpcard NumPoints = getManifold()->getNumDiscrPoints();

	//The distance between starting point and any other point on the mesh.
	// - computed only for as much as we need it.
	distances.resize(NumPoints, FLT_MAX);
	fill(distances.begin(), distances.end(), FLT_MAX);
	distances[startIndex] = 0;

	//Queue to put our visited vertices into
	priority_queue< DistanceQueuePoint > GrowQueue;
	// - we start with the given point and the distance to itself
	GrowQueue.push(DistanceQueuePoint(startIndex, startIndex, 0));

	//Grow the patch using dijkstra until we reached the desired distance
	while (!GrowQueue.empty() && GrowQueue.top().GeoDistance <= DesiredDistance)
	{
		//Pop the top
		DistanceQueuePoint ThisPoint = GrowQueue.top();
		GrowQueue.pop();
		// - and get the actual position in space
		const Vector3f ThisPos = getManifold()->getDiscrPoint(ThisPoint.index);

		//Propagate distance to neighbors of the current point
		const mpcard NumNeighs = getManifold()->getNumNeighbors(ThisPoint.index);
		for (mpcard i=0;i<NumNeighs;i++)
		{
			//Get the neighbor
			const mpcard NeighIdx = getManifold()->getNeighborIndex(ThisPoint.index, i);

			//Compute distance between neighbor and ThisPoint; add to growing queue, if appropriate
			// - The loop is only entered in about 50% of cases on a triangle or KNN mesh. Nevertheless, no measurable performance gain (tested Max Planck trimesh).
			if (ThisPoint.GeoDistance < distances[NeighIdx]) //We should not run back to where we came from
			{
				//dist = accumulated GeoDistance at ThisPoint + edge(ThisPoint ---> neighbour)
				const Vector3f NeighPos = getManifold()->getDiscrPoint(NeighIdx);
				const float32 NeighGeoDistanceViaHere =
											ThisPoint.GeoDistance
											+ (NeighPos - ThisPos).getNorm();

				//If this is the shortest path to our neighbor found so far, then it follows the dijkstra algo
				if (NeighGeoDistanceViaHere < distances[NeighIdx])
				{
					//Set value and push into queue
					distances[NeighIdx] = NeighGeoDistanceViaHere;
					GrowQueue.push(DistanceQueuePoint(NeighIdx, ThisPoint.index, NeighGeoDistanceViaHere));
				}
			}
		}
	}

	updateMinMax();
}


void AR4DijkstraDistPC3f::updateMinMax()
{
	maxDist = 0;

    for (size_t i=0; i<distances.size(); i++)
    {
        if (maxDist < distances[i] && distances[i] < FLT_MAX) maxDist = distances[i];
    }
}


float32 AR4DijkstraDistPC3f::evaluate( mpcard discrPoint ) const
{
   if (distances.size() <= discrPoint) throw EInvalidState("insufficient precomputed distances");
   return distances[discrPoint];
}

