///////////////////////////////////////////////////////////////////////////////
//
//	AR4GDLCluster.cpp
//
//	Source file for the AR4GDLCluster class
//	Implements graph degree linkage agglomerative clustering, Zhang et al. ECCV 2012
//
//	Alan Brunton, MPII March 2013
//
///////////////////////////////////////////////////////////////////////////////


#include "stdafx.h"

#include "AR4GDLCluster.h"
#include "DebugOutput.h"
#include <queue>


void X4::AR4GDLCluster::init(const mpcard numPoints, const mpcard numNN, const vector<vector<mpcard>>& neighbors, const vector<vector<float32>>& distances, bool bPreMerge/* = true*/)
{
	mpcard i, j;

	const float32 leadSigmaFactor = m_weightScale / (float32)(numPoints * numNN);

	float32 sigma2 = 0.f;

	m_nbrsIn.resize(numPoints);
	m_nbrsOut.resize(numPoints);
	m_weightsIn.resize(numPoints);
	m_weightsOut.resize(numPoints);
	m_weightsTemp.resize(numPoints);

	for (i = 0; i < numPoints; i++)
	{
		m_nbrsIn[i].clear();
		m_weightsIn[i].clear();
	}

	if (m_externalSigma <= 0.f)
	{
		for (i = 0; i < numPoints; i++)
		{
			const mpcard k = neighbors[i].size();
			
			m_nbrsOut[i] = neighbors[i];

			m_weightsOut[i].resize(k);

			for (j = 0; j < k; j++)
			{
				sigma2 += distances[i][j]*distances[i][j];
				m_nbrsIn[neighbors[i][j]].push_back(i);
				m_weightsIn[neighbors[i][j]].push_back(0.f);
			}
		}

		sigma2 *= leadSigmaFactor;
	}
	else
	{
		sigma2 = m_externalSigma*m_externalSigma;

		for (i = 0; i < numPoints; i++)
		{
			const mpcard k = neighbors[i].size();
			
			m_nbrsOut[i] = neighbors[i];

			m_weightsOut[i].resize(k);

			for (j = 0; j < k; j++)
			{
				m_nbrsIn[neighbors[i][j]].push_back(i);
				m_weightsIn[neighbors[i][j]].push_back(0.f);
			}
		}
	}

	debugOutput << "AR4GDLCluster::init(...): sigma2 = " << sigma2 << "\n";

	const float32 rcpSigma2 = 1.f / sigma2;
	for (i = 0; i < numPoints; i++)
	{
		const mpcard kout = m_nbrsOut[i].size();
		const mpcard kin = m_nbrsIn[i].size();

		for (j = 0; j < kout; j++)
		{
			const float32 dist = distances[i][j];
			m_weightsOut[i][j] = exp(-dist*dist * rcpSigma2);
		}

		for (j = 0; j < kin; j++)
		{
			mpcard idxNbr = m_nbrsIn[i][j];
			mpcard m;
			for (m = 0; m < m_nbrsOut[idxNbr].size(); m++)
			{
				if (m_nbrsOut[idxNbr][m] == i)
					break;
			}
			if (m == m_nbrsOut[idxNbr].size())	continue;//something messed up happened.... indexed wrong somehow

			const float32 dist = distances[idxNbr][m];
			m_weightsIn[i][j] = exp(-dist*dist * rcpSigma2);
		}
	}

	const mpcard k0 = m_numNNInit;

	GDLClusterContext gdlcc(m_nbrsIn, m_nbrsOut, m_weightsIn, m_weightsOut, m_weightsTemp, m_incomp, m_incompTemp);

	//form initial clusters
	m_internalClusters.clear();

	if (bPreMerge)
	{
		//ensure a point is only in one cluster
		vector<bool> alreadyAssigned(numPoints, false);
		if (m_incomp.size() > 0)
		{
			for (i = 0; i < numPoints; i++)
			{
				if (alreadyAssigned[i])	continue;

				vector<mpcard> indices, incomp;
				vector<mpcard>::iterator oldEndIt;
				queue<mpcard> indicesToProcess;

				indicesToProcess.push(i);
				while (!indicesToProcess.empty())
				{
					mpcard curIdx = indicesToProcess.front();
					indicesToProcess.pop();
					if (alreadyAssigned[curIdx])	continue;

					//add index to cluster indices
					indices.push_back(curIdx);

					//append new point's incompatibility list to that of the cluster
					const mpcard numPrevIncomp = incomp.size();
					incomp.insert(incomp.end(), m_incomp[i].begin(), m_incomp[i].end());
					oldEndIt = incomp.begin() + numPrevIncomp;

					//merge so that it is sorted
					std::inplace_merge(incomp.begin(), oldEndIt, incomp.end());

					//mark as assigned so that it wont' get added to another cluster
					alreadyAssigned[curIdx] = true;

					const mpcard numNbrsOut = m_nbrsOut[curIdx].size();
					for (j = 0; j < k0 && j < numNbrsOut; j++)
					{
						if (!alreadyAssigned[m_nbrsOut[curIdx][j]] && !std::binary_search(incomp.begin(), incomp.end(), m_nbrsOut[curIdx][j]))
							indicesToProcess.push(m_nbrsOut[curIdx][j]);
					}
				}

				if (indices.size() > 0)
					m_internalClusters.push_back( InternalGDLCluster(indices, gdlcc) );//m_internalClusters.insert( InternalGDLCluster(indices, gdlcc) );
			}
		}
		else
		{
			for (i = 0; i < numPoints; i++)
			{
				if (alreadyAssigned[i])	continue;

				vector<mpcard> indices;
				queue<mpcard> indicesToProcess;

				indicesToProcess.push(i);
				while (!indicesToProcess.empty())
				{
					mpcard curIdx = indicesToProcess.front();
					indicesToProcess.pop();
					if (alreadyAssigned[curIdx])	continue;

					//add index to cluster indices
					indices.push_back(curIdx);

					//mark as assigned so that it wont' get added to another cluster
					alreadyAssigned[curIdx] = true;

					const mpcard numNbrsOut = m_nbrsOut[curIdx].size();
					for (j = 0; j < k0 && j < numNbrsOut; j++)
					{
						if (!alreadyAssigned[m_nbrsOut[curIdx][j]])
							indicesToProcess.push(m_nbrsOut[curIdx][j]);
					}
				}

				if (indices.size() > 0)
					m_internalClusters.push_back( InternalGDLCluster(indices, gdlcc) );//m_internalClusters.insert( InternalGDLCluster(indices, gdlcc) );
			}
		}
	}
	else
	{
		for (i = 0; i < numPoints; i++)
		{
			vector<mpcard> indices(1, i);
			m_internalClusters.push_back( InternalGDLCluster(indices, gdlcc) );//m_internalClusters.insert( InternalGDLCluster(indices, gdlcc) );
		}
	}
}

void X4::AR4GDLCluster::init(const mpcard numPoints, const mpcard numNN, const vector<vector<mpcard>>& neighbors, const vector<vector<float32>>& distances, const vector<float32>& externalWeights, bool bPreMerge/* = true*/)
{
	init(numPoints, numNN, neighbors, distances, bPreMerge);

	mpcard i, j;

	for (i = 0; i < numPoints; i++)
	{
		const mpcard numNbrsIn = m_nbrsIn[i].size();
		const mpcard numNbrsOut = m_nbrsOut[i].size();

		for (j = 0; j < numNbrsIn; j++)
		{
			const mpcard nbr = m_nbrsIn[i][j];
			m_weightsIn[i][j] *= externalWeights[i * numPoints + nbr];
		}

		for (j = 0; j < numNbrsOut; j++)
		{
			const mpcard nbr = m_nbrsOut[i][j];
			m_weightsOut[i][j] *= externalWeights[i * numPoints + nbr];
		}
	}
}

void X4::AR4GDLCluster::init(const mpcard numPoints, const mpcard numNN, const vector<vector<mpcard>>& neighbors, const vector<vector<float32>>& distances, const vector<vector<card32>>& incompatibility, bool bPreMerge/* = true*/)
{
	mpcard i, j;

	m_incomp.resize(numPoints);
	m_incompTemp.resize(numPoints);

	for (i = 0; i < numPoints; i++)
	{
		const mpcard numIncomp = incompatibility[i].size();
		m_incomp[i].resize(numIncomp);
		for (j = 0; j < numIncomp; j++)
			m_incomp[i][j] = (mpcard)incompatibility[i][j];

		sort(m_incomp[i].begin(), m_incomp[i].end());
	}

	init(numPoints, numNN, neighbors, distances, bPreMerge);
}

void X4::AR4GDLCluster::cluster(const mpcard numTargetClusters, vector<vector<mpcard>>& clusters, mpcard& numInlierClusters)
{
	mpcard i, j, k;
	vector<float32> clusterAffinities, clusterAffAsymm;
	vector<mpcard> clusterNearest;

	GDLClusterContext gdlcc(m_nbrsIn, m_nbrsOut, m_weightsIn, m_weightsOut, m_weightsTemp, m_incomp, m_incompTemp);

	clusters.clear();

	const mpcard numInitClusters = m_internalClusters.size();
	clusterAffinities.resize((numInitClusters * (numInitClusters + 1)) / 2, 0.f);
	clusterAffAsymm.resize(numInitClusters * numInitClusters, 0.f);
	clusterNearest.resize(numInitClusters);

	debugOutput << "AR4GDLCluster::cluster(...):\n";
	debugOutput << "\tnumber of initial clusters = " << numInitClusters << "\n";
	debugOutput << "\taffinity threshold: " << m_affinityThresh << "\n";

	k = 0;
	for (i = 0; i < numInitClusters; i++)
	{
		for (j = 0; j < numInitClusters; j++)
		{
			if (j != i)
				clusterAffAsymm[k] = m_internalClusters[i].computeAffinityTo(m_internalClusters[j], gdlcc);
			else
				clusterAffAsymm[k] = 1.f;
			k++;
		}
	}
	k = 0;
	for (i = 0; i < numInitClusters; i++)
	{
		clusterNearest[i] = i;
		clusterAffinities[k] = 0.f;
		k++;

		float32 affMax = 0.f;
		for (j = i + 1; j < numInitClusters; j++)
		{
			clusterAffinities[k] = clusterAffAsymm[i * numInitClusters + j] + clusterAffAsymm[j * numInitClusters + i];
			if (clusterAffinities[k] > affMax)
			{
				affMax = clusterAffinities[k];
				clusterNearest[i] = j;
			}
			k++;
		}
	}

	while (m_internalClusters.size() > numTargetClusters)
	{
		const mpcard numClusters = m_internalClusters.size();

		float32 aff, affMax = m_affinityThresh;
		mpcard imax = numClusters, jmax = numClusters;

		//find maximum affinity cluster pair

		for (i = 0; i < numClusters; i++)
		{
			k = sm_ut_lookup(i, clusterNearest[i], numClusters);
			aff = clusterAffinities[k];
			if (aff > affMax)
			{
				affMax = aff;
				imax = i;
				jmax = clusterNearest[i];
			}
		}


		if (imax == numClusters)
		{
			debugOutput << "\tno two clusters with affinity > " << m_affinityThresh << ", no further clustering will be performed\n";
			break;
		}
		if (jmax == imax)
		{
			debugOutput << "\tthis doesn't make sense\n";
		}

		debugOutput << "\tclusters to merge: " << imax << " and " << jmax << " with affinity " << affMax << "\n";

		//merge clusters with maximal affinity
		InternalGDLCluster cnew(m_internalClusters[imax], m_internalClusters[jmax], gdlcc);

		//affinity of new cluster (to be placed in imax) to each other clusters is just the sum of two combined cluster to each other cluster
		if (m_incomp.size() > 0)
		{
			for (i = 0; i < numClusters; i++)
			{
				if (i != imax && i != jmax)
				{
					clusterAffAsymm[imax * numInitClusters + i] = cnew.computeAffinityTo(m_internalClusters[i], gdlcc);
				}
			}
		}
		else
		{
			for (i = 0; i < numClusters; i++)
			{
				if (i != imax && i != jmax)
				{
					clusterAffAsymm[imax * numInitClusters + i] += clusterAffAsymm[jmax * numInitClusters + i];
				}
			}
		}

		m_internalClusters[imax] = cnew;
		m_internalClusters[jmax] = m_internalClusters.back();
		m_internalClusters.pop_back();

		const mpcard numClustersNew = m_internalClusters.size();
		vector<float32> clusterAffToNew(numClusters);

		//copy affinities moved from back
		//if jmax == numClustersNew, then jmax was the last cluster anyways, don't need to copy things around
		if (numClustersNew != jmax)
		{
			//copy affinities from jmax to others
			copy(clusterAffAsymm.begin() + numClustersNew * numInitClusters, clusterAffAsymm.begin() + (numClustersNew + 1) * numInitClusters, clusterAffAsymm.begin() + jmax * numInitClusters);
			
			//copy affinities from others to jmax
			for (i = 0; i < numClustersNew; i++)
			{
				clusterAffAsymm[i * numInitClusters + jmax] = clusterAffAsymm[i * numInitClusters + numClustersNew];
			}
		}

		//update affinities of other clusters to new cluster (this is the expensive part)
		for (i = 0; i < numClustersNew; i++)
		{
			if (i != imax)
			{
				clusterAffAsymm[i * numInitClusters + imax] = m_internalClusters[i].computeAffinityTo(m_internalClusters[imax], gdlcc);
			}
		}

		//update symmetric affinities
		k = 0;
		for (i = 0; i < numClustersNew; i++)
		{
			clusterNearest[i] = i;
			clusterAffinities[k] = 0.f;
			k++;

			affMax = 0.f;
			for (j = i + 1; j < numClustersNew; j++)
			{
				clusterAffinities[k] = clusterAffAsymm[i * numInitClusters + j] + clusterAffAsymm[j * numInitClusters + i];
				if (clusterAffinities[k] > affMax)
				{
					affMax = clusterAffinities[k];
					clusterNearest[i] = j;
				}
				k++;
			}
		}
	}

	debugOutput << "\toutputting clusters and distinguishing inliers from outliers...\n";

	//find inlier and outlier clusters
	//compute internal connectivity of each cluster
	const mpcard numFinalClusters = m_internalClusters.size();
	for (i = 0; i < numFinalClusters; i++)
		m_internalClusters[i].computeConnectivity(gdlcc);

	//sort by connectivity
	sort(m_internalClusters.begin(), m_internalClusters.end());

	//set output cluster indices and determine which are inlier clusters and which are outlier clusters
	clusters.resize(numFinalClusters);
	float32 connectivityDrop, biggestConnectivityDrop = 0.f;
	numInlierClusters = numFinalClusters;
	for (i = 0; i < numFinalClusters; i++)
	{
		clusters[i].resize(m_internalClusters[i].m_indices.size());
		copy(m_internalClusters[i].m_indices.begin(), m_internalClusters[i].m_indices.end(), clusters[i].begin());

		if (i > 0)
		{
			connectivityDrop = m_internalClusters[i - 1].m_connectivity - m_internalClusters[i].m_connectivity;
			if (connectivityDrop > biggestConnectivityDrop)
			{
				biggestConnectivityDrop = connectivityDrop;
				numInlierClusters = i;
			}
		}
	}
}


