///////////////////////////////////////////////////////////////////////////////
//
//	AR4GDLCluster.h
//
//	Header file for the AR4GDLCluster class
//	Implements "Graph Degree Linkage: Agglomerative Clustering on a Directed Graph",
//	Zhang, Wang, Zhao and Tang, ECCV 2012
//
//	Alan Brunton, MPII March 2013
//
///////////////////////////////////////////////////////////////////////////////


#ifndef __AR4GDLCLUSTER_H__
#define __AR4GDLCLUSTER_H__


#include "threeParameterMatchingGMOD14.h"
#include "X4Types.h"


namespace X4
{

class AR4GDLCluster
{
private:

	//this structre is used to provide clusters with access to global information without copying data around
	struct GDLClusterContext
	{
		vector<vector<mpcard>>&				m_nbrsIn;
		vector<vector<mpcard>>&				m_nbrsOut;
		vector<vector<float32>>&			m_weightsIn;
		vector<vector<float32>>&			m_weightsOut;
		vector<float32>&					m_weightsTemp;
		
		vector<vector<mpcard>>&				m_incomp;
		vector<card32>&						m_incompTemp;

		GDLClusterContext(	vector<vector<mpcard>>& nbrsIn, vector<vector<mpcard>>& nbrsOut, 
							vector<vector<float32>>& weightsIn, vector<vector<float32>>& weightsOut, vector<float32>& weightsTemp, 
							vector<vector<mpcard>>& incomp, vector<card32>& incompTemp):
							m_nbrsIn(nbrsIn), m_nbrsOut(nbrsOut), 
							m_weightsIn(weightsIn), m_weightsOut(weightsOut), m_weightsTemp(weightsTemp), 
							m_incomp(incomp), m_incompTemp(incompTemp) 
		{
		}
	};

	class InternalGDLCluster
	{
	public:
		vector<mpcard>						m_indices;
		float32								m_connectivity;
		vector<mpcard>						m_incompList;

		InternalGDLCluster()
		{
			m_connectivity = 0.f;
		}
		InternalGDLCluster(const vector<mpcard>& indices, const GDLClusterContext& gdlcc)
		{
			m_indices = indices;
			m_connectivity = 0.f;
			sort(m_indices.begin(), m_indices.end());
			if (gdlcc.m_incomp.size() > 0)
				compileIncompatibilityList(gdlcc);
		}
		InternalGDLCluster(const InternalGDLCluster& c1, const InternalGDLCluster& c2, const GDLClusterContext& gdlcc)
		{
			m_indices.clear();
			m_indices.resize(c1.m_indices.size() + c2.m_indices.size());
			m_connectivity = 0.f;
			std::merge(c1.m_indices.begin(), c1.m_indices.end(), c2.m_indices.begin(), c2.m_indices.end(), m_indices.begin());
			if (gdlcc.m_incomp.size() > 0)
				mergeIncompatibilityLists(c1.m_incompList, c2.m_incompList);
		}

		void compileIncompatibilityList(const GDLClusterContext& gdlcc)
		{
			const vector<vector<mpcard>>& incomp = gdlcc.m_incomp;

			mpcard i;
			const mpcard numMembers = m_indices.size();
			mpcard totalIncomp = 0;
			for (i = 0; i < numMembers; i++)
				totalIncomp += incomp[m_indices[i]].size();

			m_incompList.resize(totalIncomp);
			vector<mpcard>::iterator it = m_incompList.begin();
			for (i = 0; i < numMembers; i++)
				it = copy(incomp[m_indices[i]].begin(), incomp[m_indices[i]].end(), it);

			mpcard off1 = 0, off2 = incomp[m_indices[0]].size();
			for (i = 1; i < numMembers; i++)
			{
				const mpcard step = incomp[m_indices[i]].size();
				std::inplace_merge(m_incompList.begin(), m_incompList.begin() + off2, m_incompList.begin() + off2 + step);
				off2 += step;
			}

			m_incompList.erase(unique(m_incompList.begin(), m_incompList.end()), m_incompList.end());
		}

		void mergeIncompatibilityLists(const vector<mpcard>& list1, const vector<mpcard>& list2)
		{
			m_incompList.clear();
			m_incompList.resize(list1.size() + list2.size());
			std::merge(list1.begin(), list1.end(), list2.begin(), list2.end(), m_incompList.begin());
			m_incompList.erase(unique(m_incompList.begin(), m_incompList.end()), m_incompList.end());
		}

		inline bool operator<(const InternalGDLCluster& c) const
		{
			if (m_connectivity == c.m_connectivity)
				return m_indices[0] < c.m_indices[0];//on the off-chance that two cluster have the exact same internal connectivity, differentiate them by first member
			return (m_connectivity > c.m_connectivity);//sort by descending order of internal connectivity
		}

		inline InternalGDLCluster& operator=(const InternalGDLCluster& c)
		{
			m_indices = c.m_indices;
			m_connectivity = c.m_connectivity;
			m_incompList = c.m_incompList;
			return *this;
		}

		void computeConnectivity(const GDLClusterContext& gdlcc)
		{
			const mpcard numMembers = m_indices.size();
			const float32 rcpNumMembers = 1.f / (const float32)numMembers;

			mpcard i, j;
			pair<vector<mpcard>::iterator, vector<mpcard>::iterator> searchRes;

			const vector<vector<mpcard>>& nbrsIn = gdlcc.m_nbrsIn;
			const vector<vector<mpcard>>& nbrsOut = gdlcc.m_nbrsOut;
			const vector<vector<float32>>& weightsIn = gdlcc.m_weightsIn;
			const vector<vector<float32>>& weightsOut = gdlcc.m_weightsOut;
			const vector<float32>& weightsTemp = gdlcc.m_weightsTemp;

			m_connectivity = 0.f;

			for (i = 0; i < numMembers; i++)
			{
				const mpcard curIdx = m_indices[i];
				const mpcard numNNIn = nbrsIn[curIdx].size();
				const mpcard numNNOut = nbrsOut[curIdx].size();

				float32 curCon = 0.f;

				for (j = 0; j < numNNIn; j++)
				{
					searchRes = equal_range(m_indices.begin(), m_indices.end(), nbrsIn[curIdx][j]);
					if (searchRes.first != searchRes.second && searchRes.first != m_indices.end())
					{
						curCon += weightsIn[curIdx][j];
					}
				}

				for (j = 0; j < numNNOut; j++)
				{
					searchRes = equal_range(m_indices.begin(), m_indices.end(), nbrsOut[curIdx][j]);
					if (searchRes.first != searchRes.second && searchRes.first != m_indices.end())
					{
						curCon += weightsOut[curIdx][j];
					}
				}

				m_connectivity += curCon;
			}

			m_connectivity *= rcpNumMembers;
		}

		float32 computeAffinityTo(const InternalGDLCluster& c, const GDLClusterContext& gdlcc) const
		{
			const vector<vector<mpcard>>& incomp = gdlcc.m_incomp;

			const mpcard numMembers = m_indices.size();

			mpcard i;

			float32 affinity = 0.f;

			if (incomp.size() > 0)
			{
				vector<card32> intersection;
				set_intersection(m_incompList.begin(), m_incompList.end(), c.m_indices.begin(), c.m_indices.end(), back_inserter(intersection));
				if (intersection.size() > 0)
					return 0.f;
			}

			for (i = 0; i < numMembers; i++)
			{
				affinity += vertexToClusterAffinity(m_indices[i], c, gdlcc);
			}

			//modification to original algorithm: divide by size of cluster to reduce bias for large clusters
			return affinity / (float32)numMembers;
		}

		float32 computeAffinity(const InternalGDLCluster& c, const GDLClusterContext& gdlcc) const
		{
			return computeAffinityTo(c, gdlcc) + c.computeAffinityTo(*this, gdlcc);
		}

		static float32 vertexToClusterAffinity(const mpcard idx, const InternalGDLCluster& c, const GDLClusterContext& gdlcc)
		{
			const vector<vector<mpcard>>& nbrsIn = gdlcc.m_nbrsIn;
			const vector<vector<mpcard>>& nbrsOut = gdlcc.m_nbrsOut;
			const vector<vector<float32>>& weightsIn = gdlcc.m_weightsIn;
			const vector<vector<float32>>& weightsOut = gdlcc.m_weightsOut;
			vector<float32>& weightsTemp = gdlcc.m_weightsTemp;

			mpcard i;
			const mpcard numMembers = c.m_indices.size();
			const float32 rcpNumMembers = 1.f / (const float32)numMembers;

			const mpcard numNNIn = nbrsIn[idx].size();
			const mpcard numNNOut = nbrsOut[idx].size();

			float32 affinity = 0.f;
			float32 avgInDeg = 0.f, avgOutDeg = 0.f;

			//clear temp weights for cluster members
			for (i = 0; i < numMembers; i++)
				weightsTemp[c.m_indices[i]] = 0.f;

			//set temp weights for in-neighbors
			for (i = 0; i < numNNIn; i++)
				weightsTemp[nbrsIn[idx][i]] = weightsIn[idx][i];

			//sum temp weights for cluster members
			for (i = 0; i < numMembers; i++)
				avgInDeg += weightsTemp[c.m_indices[i]];

			//divide by cluster size
			avgInDeg *= rcpNumMembers;
			
			//clear temp weights for cluster members
			for (i = 0; i < numMembers; i++)
				weightsTemp[c.m_indices[i]] = 0.f;

			//set temp weights for out-neighbors
			for (i = 0; i < numNNOut; i++)
				weightsTemp[nbrsOut[idx][i]] = weightsOut[idx][i];

			//sum temp weights for cluster members
			for (i = 0; i < numMembers; i++)
				avgOutDeg += weightsTemp[c.m_indices[i]];

			//divide by cluster size
			avgOutDeg *= rcpNumMembers;

			//affinity is product of average indegree and average outdegree
			affinity = avgInDeg * avgOutDeg;

			return affinity;
		}
	};


	vector<vector<mpcard>>					m_nbrsIn;
	vector<vector<mpcard>>					m_nbrsOut;

	float32									m_weightScale;
	vector<vector<float32>>					m_weightsIn;
	vector<vector<float32>>					m_weightsOut;

	float32									m_externalSigma;

	vector<vector<mpcard>>					m_incomp;

	vector<float32>							m_weightsTemp;//to avoid searching

	vector<card32>							m_incompTemp;//to avoid searching

	mpcard									m_numNNInit;

	float32									m_affinityThresh;

	vector<InternalGDLCluster>				m_internalClusters;


	static inline mpcard sm_diagonal(const mpcard row, const mpcard n)
	{
		return row * n - ((row * (row + 1)) >> 1) + row;
	}
	static inline mpcard sm_lookup(const mpcard row, const mpcard col, const mpcard n)
	{
		const mpcard diff = row - col;
		if (col >= row)
			return sm_diagonal(row, n) + diff;
		return sm_diagonal(row, n) - diff * n + ((diff * (row + col - 1)) >> 1) + diff;
	}
	static inline mpcard sm_ut_lookup(const mpcard row, const mpcard col, const mpcard n)
	{
		const mpcard diff = col - row;
		return sm_diagonal(row, n) + diff;
	}
	static inline mpcard sm_lt_lookup(const mpcard row, const mpcard col, const mpcard n)
	{
		const mpcard diff = row - col;
		return sm_diagonal(row, n) - diff * n + ((diff * (row + col - 1)) >> 1) + diff;
	}


public:

	AR4GDLCluster()
	{
		m_weightScale = 10.f;
		m_numNNInit = 1;
		m_affinityThresh = 0.f;
		m_externalSigma = -1.f;
	}
	~AR4GDLCluster()
	{
	}

	void clear()
	{
		m_nbrsIn.clear();
		m_nbrsOut.clear();
		m_weightsIn.clear();
		m_weightsOut.clear();
		m_weightsTemp.clear();
		m_incomp.clear();
		m_incompTemp.clear();
	}

	void setAffinityThreshold(float32 thresh)	{ m_affinityThresh = thresh; }
	float32 getAffinityThreshold()				{ return m_affinityThresh; }

	void setExternalSigma(float32 extSigma)		{ m_externalSigma = extSigma; }
	float32 getExternalSigma()					{ return m_externalSigma; }

	void init(const mpcard numPoints, const mpcard numNN, const vector<vector<mpcard>>& neighbors, const vector<vector<float32>>& distances, bool bPreMerge = true);
	void init(const mpcard numPoints, const mpcard numNN, const vector<vector<mpcard>>& neighbors, const vector<vector<float32>>& distances, const vector<float32>& externalWeights, bool bPreMerge = true);
	void init(const mpcard numPoints, const mpcard numNN, const vector<vector<mpcard>>& neighbors, const vector<vector<float32>>& distances, const vector<vector<card32>>& incompatibility, bool bPreMerge = true);

	void cluster(const mpcard numTargetClusters, vector<vector<mpcard>>& clusters, mpcard& numInlierClusters);
};

}


#endif //__AR4GDLCLUSTER_H__


