//---------------------------------------------------------------------------
#include "stdafx.h"
//---------------------------------------------------------------------------
#include "AR4IncrementalMatcher.h"
#include "AR4IncrementalMatcher.inline.h"
#include "AR4PartialMatch.h"
#include "AR4PartialMatch.inline.h"
//---------------------------------------------------------------------------
#include "AR4MLSManifold.h"
//---------------------------------------------------------------------------
#include "DynamicLinearAlgebraTools.h"
#include "SVD.h"
#include "IterativeSolvers.h"
#include "IterativeSolvers.inline.h"
#include "ApproximatedGeodesicsGM14.h"
#include "DebugRenderer.h"
#include "ColorTools.h"
#include "StringHelper.h"
#include "Timer.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
//preprocessor definitions
//---------------------------------------------------------------------------
#define AR4_DEBUG_OUTPUT					0
#define AR4IM_INNER_DEBUG_RENDER			0
#define AR4IM_INNER_DEBUG_RENDER_JOBNAME	"AR4IncrementalMatcher_Inner"
#define AR4IM_PROFILE						1
#if AR4IM_PROFILE
#define AR4IM_START_PROFILE(profile) profile.start();
#define AR4IM_STOP_PROFILE(profile)	profile.stop();
#else
#define AR4IM_START_PROFILE(profile)
#define AR4IM_STOP_PROFILE(profile)
#endif //AR4IM_PROFILE

//---------------------------------------------------------------------------

//utility
static X4::Matrix3f multTwoVectors(X4::Vector3f v1, X4::Vector3f v2)
{
   X4::Matrix3f m;
   m[0][0] = v1[0] * v2[0];
   m[0][1] = v1[1] * v2[0];
   m[0][2] = v1[2] * v2[0];

   m[1][0] = v1[0] * v2[1];
   m[1][1] = v1[1] * v2[1];
   m[1][2] = v1[2] * v2[1];

   m[2][0] = v1[0] * v2[2];
   m[2][1] = v1[1] * v2[2];
   m[2][2] = v1[2] * v2[2];

   return m;   
}

///////////////////////////////////////////////////////////////////////////////
//bundle adjustment / GMDS
///////////////////////////////////////////////////////////////////////////////

template <class ManifoldTraits>
typename AR4IncrementalMatcher<ManifoldTraits>::FloatType AR4BundleAdjustment<ManifoldTraits>::computeGeodesicError(const std::vector< std::pair<VectorType, VectorType> >& corrs, atrec::ApproximatedGeodesicsGM14* pSourceGeod, atrec::ApproximatedGeodesicsGM14* pTargetGeod)
{
	mpcard i, j, numCorrs = (mpcard)corrs.size();
	FloatType error = 0;

	FloatType cur, ref, diff;

	for (i = 0; i < numCorrs; i++)
	{
		for (j = 0; j < numCorrs; j++)
		{
			if (j == i)
				continue;

			ref = pSourceGeod->interpolate(corrs[i].first, corrs[j].first, m_geodSigma);
			cur = pTargetGeod->interpolate(corrs[i].second, corrs[j].second, m_geodSigma);

			if (cur < 1e10 && ref < 1e10)
			{
				diff = cur - ref;
				error += diff*diff;
			}
		}
	}

	return error;
}


///////////////////////////////////////////////////////////////////////////////
//IndexedGeodesic and IndexedVectorCompare
//helper classes to keep geodesic information in the same order when applying bundle adjustment
///////////////////////////////////////////////////////////////////////////////

class IndexedGeodesic
{
public:
	mpcard				m_index;
	float				m_sourceDist;
	float				m_targetDist;
	Vector3f			m_targetGrad;

	IndexedGeodesic(): m_index(0), m_sourceDist(0), m_targetDist(0) 
	{
		m_targetGrad.setZero();
	}
	inline IndexedGeodesic& operator=(const IndexedGeodesic& ig)
	{
		m_index = ig.m_index;
		m_sourceDist = ig.m_sourceDist;
		m_targetDist = ig.m_targetDist;
		m_targetGrad = ig.m_targetGrad;
		return *this;
	}
	inline bool operator<(const IndexedGeodesic& ig) const
	{
		return (m_index < ig.m_index);
	}
};

typedef pair<mpcard, set<IndexedGeodesic>>		IndexedGeodesicSet;
typedef pair<mpcard, vector<IndexedGeodesic>>	IndexedGeodesicVector;

class IndexedGeodesicSetCompare
{
public:
	inline bool operator()(const IndexedGeodesicSet& igd1, const IndexedGeodesicSet& igd2) const
	{
		return (igd1.first < igd2.first);
	}
};

template<typename T1, typename T2>
class PairFirstComp
{
public:
	inline bool operator()(const pair<T1, T2>& p1, const pair<T1, T2>& p2) const
	{
		return (p1.first < p2.first);
	}
};

typedef PairFirstComp<mpcard, mpcard>								PFC_mpcard;//comparison operator of first of pair of indices
typedef set<pair<mpcard, mpcard>, PFC_mpcard>						AssocIndexSet;//associated index set, sorted by first index
typedef pair<mpcard, AssocIndexSet>									IAIS;//indexed associated index set (index attached to associated index set)
typedef PairFirstComp<mpcard, AssocIndexSet>						PFC_IAIS;//comparison for IAIS


class IndexDistCompare
{
public:
	inline bool operator()(const pair<mpcard, float>& p1, const pair<mpcard, float>& p2) const
	{
		return (p1.second < p2.second);
	}
};


///////////////////////////////////////////////////////////////////////////////
//AR4DenseBundle3f
//concrete template-free class for dense bundle adjustment
///////////////////////////////////////////////////////////////////////////////

void AR4DenseBundle3f::improveMatch(X4::AR4PartialMatch<ManifoldTraits3f> *match)
{
#define AR4BUNDLE_LIMIT_NBRHD	1
#define AR4BUNDLE_LEVEL1_NBRHD	1

	Timer timer;

	mpcard iter;

	mpcard i, k, j, m, n;
	mpcard numCorr = match->getNumMatchedIndices();
	float32 error, besterror = 1e30f, lasterror;
	float32 nonfixed = 1.f;

	float32 dcur, dref;

	vector<mpcard> validNbrs;

	DynamicMatrix<float> M, U, V;
	DynamicVector<float> B, D, result, intermed;

	Vector3f sol;
	float32 lengthmin, lengthmax, length, lengthrange, rcplengthrange;

	vector< pair<float, Vector3f> > solution(numCorr);
	vector< pair<Vector3f, Vector3f> > corrs(numCorr);
	vector<bool> fixed(numCorr);

	float stepSize = m_stepScale;
	Vector3f step, projPoint, projNorm;

	Vector3f sourcePti, targetPti, sourcePtk, targetPtk;
	Vector3f ikgrad, kigrad;
	Matrix3f kiikMat, kikiMat;

	mpcard numfixed = 0, ifixed = 0, usedcols = 0;
	vector<mpcard> colmap(numCorr);

	//retrieve non-fixed matches and valid neighbors
#if AR4BUNDLE_LIMIT_NBRHD
	vector<mpcard> originalIndices, singleNbr(1);
	vector<mpcard> actualIndices;
	set<mpcard> actualIndexSet;
	set<mpcard>::iterator aisIt;
	vector<mpcard> targetGeos, targetGeosIn;
	IAIS geoIndexSet;
	set<IAIS, PFC_IAIS> geoIndexCollect;
	set<IAIS, PFC_IAIS>::iterator gicIt;
	AssocIndexSet::iterator pisIt;
	vector<vector<float>> sourceDistLookup;
	vector<vector<mpcard>> allTargetGeos, allTargetGeosAsEnd;
	vector<vector<mpcard>> sourceDests, corrDests;

	mpcard startCorr = 0;
	if (match->getNumMatchCohorts() > 1)
		startCorr = match->getMatchCohortCount(match->getNumMatchCohorts() - 2);

	//collect geodesics _starting_ from non-fixed matches
	for (i = startCorr; i < numCorr; i++)
	{
		k = match->getMatchedSourceIndex(i);
		if (!match->isFixed(k))
		{
			actualIndexSet.insert(k);
#if AR4BUNDLE_LEVEL1_NBRHD
			match->getMatchedGeodesicsIndicesMostLocal(k, validNbrs, targetGeos, targetGeosIn);
#else
			match->getValidMatchedGeodesicsIndices(k, validNbrs, targetGeos);
#endif //AR4BUNDLE_LEVEL1_NBRHD

			geoIndexSet.second.clear();
			for (j = 0; j < validNbrs.size(); j++)
			{
				geoIndexSet.second.insert(pair<mpcard, mpcard>(validNbrs[j], targetGeos[j]));
				actualIndexSet.insert(validNbrs[j]);
				originalIndices.push_back(k);
			}
			
			geoIndexSet.first = k;
			geoIndexCollect.insert(geoIndexSet);
		}
	}

	//collect geodesics from matches that were added as destinations of geodesics from non-fixed matches
	//this will give the reverse geodesics (from fixed to non-fixed)
	//(in general it can also include fixed to fixed geodesics, which will not get used in the optimization)
	numCorr = actualIndexSet.size();
	for (aisIt = actualIndexSet.begin(); aisIt != actualIndexSet.end(); aisIt++)
	{
		k = *aisIt;
		actualIndices.push_back(k);
		if (match->isFixed(k))
		{
#if AR4BUNDLE_LEVEL1_NBRHD
			match->getMatchedGeodesicsIndicesMostLocal(k, validNbrs, targetGeos, targetGeosIn);
#else
			match->getValidMatchedGeodesicsIndices(k, validNbrs, targetGeos);
#endif //AR4BUNDLE_LEVEL1_NBRHD

			geoIndexSet.second.clear();
			for (j = 0; j < validNbrs.size(); j++)
			{
				if (actualIndexSet.find(validNbrs[j]) != actualIndexSet.end())
					geoIndexSet.second.insert(pair<mpcard, mpcard>(validNbrs[j], targetGeos[j]));
			}

			geoIndexSet.first = k;
			geoIndexCollect.insert(geoIndexSet);
		}
	}

	//now put everything into vectors in consistent order for easy access
	//note that everything will be in the same order as actualIndices
	sourceDistLookup.resize(numCorr);
	allTargetGeos.resize(numCorr);
	allTargetGeosAsEnd.resize(numCorr);
	sourceDests.resize(numCorr);
	corrDests.resize(numCorr);

	i = 0;
	for (gicIt = geoIndexCollect.begin(); gicIt != geoIndexCollect.end(); gicIt++)
	{
		sourceDistLookup[i].reserve(numCorr);
		allTargetGeos[i].reserve(numCorr);
		sourceDests[i].reserve(numCorr);
		corrDests[i].reserve(numCorr);

		for (pisIt = (*gicIt).second.begin(); pisIt != (*gicIt).second.end(); pisIt++)
		{
			sourceDests[i].push_back((*pisIt).first);
			sourceDistLookup[i].push_back(match->sourceGeodesic((*gicIt).first, (*pisIt).first));
			allTargetGeos[i].push_back((*pisIt).second);
			j = lower_bound<vector<mpcard>::iterator, mpcard>(actualIndices.begin(), actualIndices.end(), (*pisIt).first) - actualIndices.begin();
			corrDests[i].push_back(j);
		}

		i++;
	}

	corrs.resize(numCorr);
	fixed.resize(numCorr);
	solution.resize(numCorr);
	colmap.resize(numCorr);
#endif //AR4BUNDLE_LIMIT_NBRHD

	//retrieve existing matches
	for (i = 0; i < numCorr; i++)
	{
#if AR4BUNDLE_LIMIT_NBRHD
		k = actualIndices[i];
#else
		k = match->getMatchedSourceIndex(startCorr + i);
#endif //AR4BUNDLE_LIMIT_NBRHD
		corrs[i].first = match->getSourceManifold()->getDiscrPoint(k);
		corrs[i].second = match->getMatch(k);
		fixed[i] = match->isFixed(k);
		if (fixed[i])
			numfixed++;
		else
		{
			colmap[i] = usedcols;
			usedcols++;
		}
	}//for (i...

	if (m_bDebugOutput)
		debugOutput << "AR4DenseBundle3f::improveMatch(...):\n" << "\tnumber of fixed points: " << numfixed << "\n";

	//compute starting geodesic error
	m_lastStartError = besterror = lasterror = evalError(match, sourceDistLookup, allTargetGeos, fixed);
	if (m_bDebugOutput)
		debugOutput << "\tstarting error: " << besterror << "\n";

	match->backupNonFixedMatchesAndGeodesics();

	//retrieve source and target surfaces
	AR4MLSManifold3f* pSourceMLS = dynamic_cast<AR4MLSManifold3f*>(match->getSourceManifold());
	if (pSourceMLS == NULL)
		throw ERangeCheck("AR4DenseBundle3f::improveMatch(...) - source is not an MLS manifold.");

	mpcard dimopt = 3;
	mpcard cols, rows;
	cols = (numCorr - numfixed) * 3;
	rows = numCorr * 3;

	bool bNoConv = true;//not yet converged

	//iterate
	for (iter = 0; iter < m_nIterations && bNoConv; iter++)
	{
		//setup matrix and vector
		M.setDimension(cols, rows);
		B.setDim(rows);
		M.setZero();
		B.setZero();

		ifixed = 0;

		//fill matrix M and vector B
		for (k = 0; k < numCorr; k++)
		{
			//get correspondence k
			sourcePtk = corrs[k].first;
			targetPtk = corrs[k].second;

			for (m = 0; m < corrDests[k].size(); m++)
			{
				i = corrDests[k][m];
				if (i == k)	continue;

				//i is the column of the matrix, k is the row
				//get correspondence i
				sourcePti = corrs[i].first;
				targetPti = corrs[i].second;

				//get geodesics and gradients for target surface
				for (n = 0; n < corrDests[i].size(); n++)
				{
					if (corrDests[i][n] == k)
						break;
				}
				match->targetGeodesic(allTargetGeos[i][n], ikgrad);
				dcur = match->targetGeodesic(allTargetGeos[k][m], kigrad);

				//compute 3x3 matrices
				kiikMat = multTwoVectors(kigrad, ikgrad);
				kikiMat = multTwoVectors(kigrad, kigrad);

				if (!fixed[k])
				{
					// set the i==k elements    
					j = colmap[k];
					M[j*dimopt + 0][k*dimopt + 0] += kikiMat[0][0];
					M[j*dimopt + 1][k*dimopt + 0] += kikiMat[1][0];
					M[j*dimopt + 2][k*dimopt + 0] += kikiMat[2][0];

					M[j*dimopt + 0][k*dimopt + 1] += kikiMat[0][1];
					M[j*dimopt + 1][k*dimopt + 1] += kikiMat[1][1];
					M[j*dimopt + 2][k*dimopt + 1] += kikiMat[2][1];

					M[j*dimopt + 0][k*dimopt + 2] += kikiMat[0][2];
					M[j*dimopt + 1][k*dimopt + 2] += kikiMat[1][2];
					M[j*dimopt + 2][k*dimopt + 2] += kikiMat[2][2];
				}

				if (!fixed[i])
				{
					// set i != k elements
					j = colmap[i];
					M[j*dimopt + 0][k*dimopt + 0] = kiikMat[0][0];
					M[j*dimopt + 1][k*dimopt + 0] = kiikMat[1][0];
					M[j*dimopt + 2][k*dimopt + 0] = kiikMat[2][0];

					M[j*dimopt + 0][k*dimopt + 1] = kiikMat[0][1];
					M[j*dimopt + 1][k*dimopt + 1] = kiikMat[1][1];
					M[j*dimopt + 2][k*dimopt + 1] = kiikMat[2][1];

					M[j*dimopt + 0][k*dimopt + 2] = kiikMat[0][2];
					M[j*dimopt + 1][k*dimopt + 2] = kiikMat[1][2];
					M[j*dimopt + 2][k*dimopt + 2] = kiikMat[2][2];
				}

				//dcur was returned above when retrieve the gradient for allTargetGeos[k][i]
				dref = sourceDistLookup[k][m];

				B[k*dimopt + 0] += (dcur - dref) * kigrad[0];
				B[k*dimopt + 1] += (dcur - dref) * kigrad[1];
				B[k*dimopt + 2] += (dcur - dref) * kigrad[2];
			}//for (i..
		}//for (k...

		//use SVD to compute the solution
		SVD<float32>::computeSVD(M, U, D, V);
		U.transpose();
		V.transpose();
		for (i = 0; i < D.getDim(); i++)
		{
			if (fabs(D[i]) > 0.001f)//FUTURE PARAMETER
				D[i] = 1.f / D[i];
			else
				D[i] = 0.f;
		}//for (i...

		intermed = U * B;
		for (i = 0; i < D.getDim(); i++)
			D[i] *= intermed[i];
		result = V * D;

		//normalize the movement vector, need to compute min and max length
		lengthmin = 1e30f;
		lengthmax = 0.f;
		for (i = 0; i < numCorr; i++)
		{
			if (fixed[i])
			{
				length = 0;
				sol = NULL_VECTOR3F;
				solution[i] = pair<float, Vector3f>(length, sol);
				continue;
			}

			j = colmap[i];
			sol = makeVector3f(result[j * dimopt + 0], result[j * dimopt + 1], result[j * dimopt + 2]);
			length = norm(sol);
			sol.normalize();
			if (length > lengthmax) lengthmax = length;
			if (length < lengthmin) lengthmin = length;
			solution[i] = pair<float, Vector3f>(length, sol);
		}//for (i...
		lengthrange = lengthmax - lengthmin;
		if (usedcols > 1)
			rcplengthrange = 1.f / lengthrange;
		else
		{
			lengthmin = 0.f;
			rcplengthrange = 1.f / lengthmax;
		}

		//move the points by the solution we found and project them back to the surface
		for (i = 0; i < numCorr; i++)
		{
			// if length of the solution is too small, then don't move
			if (solution[i].second.getSqrNorm() < 1e-5)	continue;//FUTURE PARAMETER

			// scale the length of the solution vector accordingly
			solution[i].second *= (solution[i].first - lengthmin) * rcplengthrange;

			// move the feature just in small steps
			step = solution[i].second * stepSize;

			match->getTargetManifold()->project(corrs[i].second + step, projPoint, projNorm);

			corrs[i].second = projPoint;

			//doing this here is new, based on the assumption that geodesics on target surface are recomputed every step
			//within setMatch... some changes below relate to this... this only happens for non-fixed points because of above condition on solution[i].second.getSqrNorm()
			match->setMatch(actualIndices[i], corrs[i].second);
		}//for (i...

		//update geodesics on target surface
#if AR4BUNDLE_LEVEL1_NBRHD
		match->updateTargetGeodesicsLimited();
#else
		match->updateTargetGeodesics();
#endif //AR4BUNDLE_LEVEL1_NBRHD
		
		//compute current correspondence error
		error = evalError(match, sourceDistLookup, allTargetGeos, fixed);

		//if it is better than the previous error, store the result
		if (error < besterror)
		{
			besterror = error;
			match->backupNonFixedMatchesAndGeodesics();
		}
		else
		{
			//roll-back partial match and retrieve matches that gave lower error
			match->rollbackMatchesAndGeodesics();

			for (i = 0; i < numCorr; i++)
			{
				k = actualIndices[i];
				corrs[i].second = match->getMatch(k);
			}

			//try reducing step size to improve chances of minimizing error
		}

		//error threshold
		if (error < m_errorStopThresh || fabs(error - lasterror) < m_errorChangeStopThresh)//PARAMETERS
			bNoConv = false;

		lasterror = error;
	}//for (iter...

	m_lastEndError = besterror;

	if (m_bDebugOutput)
	{
		debugOutput << "\tnumber of iterations: " << iter << "\n";
		debugOutput << "\tfinal error: " << m_lastEndError << "\n";
		debugOutput << "\ttime: " << convertTimeToString(timer.getDeltaValue()) << "\n";
	}
}

//concrete template-free class for sparse bundle adjustment
void AR4SparseBundle3f::improveMatch(X4::AR4PartialMatch<ManifoldTraits3f> *match)
{
	Timer timer;

	CGSolver<float> solver;

	SparseMatrix<float> A;
	SparseMatrix<float> Id;
	DynamicVector<float> X, B;

	mpcard i, j, k, m;
	mpcard numCorr = match->getNumMatchedIndices();
	mpcard dimopt = 3;
	mpcard probSize = numCorr * dimopt;
	mpcard iter;

	float32 error, besterror = 1e30f, lasterror;

	vector<pair<Vector3f, Vector3f>> corrs(numCorr);
	vector< pair<float, Vector3f> > solution(numCorr);
	vector<vector<mpcard>> corrDepend(numCorr), corrDependIndex(numCorr);

	float32 curMatEntry;
	float32 dcur, dref;

	vector<mpcard> validNbrs;

	Vector3f sol;
	float32 lengthmin, lengthmax, length, lengthrange, rcplengthrange;

	vector<bool> fixed(numCorr, false);

	float stepSize = m_stepScale;
	Vector3f step, projPoint, projNorm;

	Vector3f sourcePti, targetPti, sourcePtk, targetPtk;
	Vector3f ikgrad, kigrad;
	Matrix3f kiikMat, kikiMat;

	mpcard numfixed = 0, ifixed = 0, usedcols = 0;
	vector<mpcard> colmap(numCorr);

	X.setDim(probSize);
	B.setDim(probSize);
	A.setRows(probSize);

	float lambda = 0.1f;
	Id.setRows(probSize);
	Id.setZero();
	for (i = 0; i< probSize; i++)
	{
		Id[i].setEntry(i, lambda);
	}

#if AR4_DEBUG_OUTPUT
		debugOutput << "AR4SparseBundle3f::improveMatch(...):\n";
#endif //AR4_DEBUG_OUTPUT

	vector<vector<float>> sourceDistLookup(numCorr);
	vector<vector<mpcard>> allTargetGeos(numCorr);

	//collect geodesics _starting_ from matches
	for (i = 0; i < numCorr; i++)
	{
		k = match->getMatchedSourceIndex(i);
		corrs[i].first = match->getSourceManifold()->getDiscrPoint(k);
		corrs[i].second = match->getMatch(k);

		//get other matches for which we have valid geodesic measurements between this point and there
		//corrDepend[i] contains index in [0,numCorr-1] of correspondence ("correspondence index" instead of index within source surface)
		match->getValidMatchedGeodesicsIndices(k, validNbrs, corrDepend[i], allTargetGeos[i]);

		sourceDistLookup[i].resize(validNbrs.size());
		for (j = 0; j < validNbrs.size(); j++)
			sourceDistLookup[i][j] = match->sourceGeodesic(k, validNbrs[j]);
	}

	//compute starting geodesic error
	m_lastStartError = besterror = lasterror = evalError(match, sourceDistLookup, allTargetGeos, fixed);
#if AR4_DEBUG_OUTPUT
		debugOutput << "\tstarting error: " << besterror << "\n";
#endif //AR4_DEBUG_OUTPUT

	match->backupMatchesAndGeodesics();

	bool bNoConv = true;//not yet converged

	//set solver parameters
	solver.maxNumIterations = m_nSolverIters;
	solver.stopRes = m_residualThreshold;

	X.setZero();

	//iterate
	for (iter = 0; iter < m_nIterations && bNoConv; iter++)
	{
		A.setZero();
		B.setZero();

		B = X * lambda;
		A = Id;

		//fill in the system
		for (k = 0; k < numCorr; k++)
		{
			sourcePtk = corrs[k].first;
			targetPtk = corrs[k].second;

			const mpcard numDepend = corrDepend[k].size();
			for (j = 0; j < numDepend; j++)
			{
				i = corrDepend[k][j];
				if (i == k)	continue;

				//i is the column of the matrix, k is the row
				//get correspondence k
				sourcePti = corrs[i].first;
				targetPti = corrs[i].second;

				//look up the row for the other correspondence
				const mpcard numRevDepend = corrDepend[i].size();
				for (m = 0; m < numRevDepend; m++)
				{
					if (corrDepend[i][m] == k)
						break;
				}

				//target geodesics and gradients
				match->targetGeodesic(allTargetGeos[k][j], ikgrad);
				dcur = match->targetGeodesic(allTargetGeos[i][m], kigrad);

				//compute 3x3 matrices
				kiikMat = multTwoVectors(kigrad, ikgrad);
				kikiMat = multTwoVectors(kigrad, kigrad);

				if (!fixed[k])
				{
					// set the i==k elements    

					curMatEntry = A[k*dimopt + 0].getEntry(k*dimopt + 0);
					A[k*dimopt + 0].setEntry(k*dimopt + 0, curMatEntry + kikiMat[0][0]);
					curMatEntry = A[k*dimopt + 0].getEntry(k*dimopt + 1);
					A[k*dimopt + 0].setEntry(k*dimopt + 1, curMatEntry + kikiMat[1][0]);
					curMatEntry = A[k*dimopt + 0].getEntry(k*dimopt + 2);
					A[k*dimopt + 0].setEntry(k*dimopt + 2, curMatEntry + kikiMat[2][0]);

					curMatEntry = A[k*dimopt + 1].getEntry(k*dimopt + 0);
					A[k*dimopt + 1].setEntry(k*dimopt + 0, curMatEntry + kikiMat[0][1]);
					curMatEntry = A[k*dimopt + 1].getEntry(k*dimopt + 1);
					A[k*dimopt + 1].setEntry(k*dimopt + 1, curMatEntry + kikiMat[1][1]);
					curMatEntry = A[k*dimopt + 1].getEntry(k*dimopt + 2);
					A[k*dimopt + 1].setEntry(k*dimopt + 2, curMatEntry + kikiMat[2][1]);

					curMatEntry = A[k*dimopt + 2].getEntry(k*dimopt + 0);
					A[k*dimopt + 2].setEntry(k*dimopt + 0, curMatEntry + kikiMat[0][2]);
					curMatEntry = A[k*dimopt + 2].getEntry(k*dimopt + 1);
					A[k*dimopt + 2].setEntry(k*dimopt + 1, curMatEntry + kikiMat[1][2]);
					curMatEntry = A[k*dimopt + 2].getEntry(k*dimopt + 2);
					A[k*dimopt + 2].setEntry(k*dimopt + 2, curMatEntry + kikiMat[2][2]);
				}

				if (!fixed[i])
				{
					// set i != k elements

					curMatEntry = A[k*dimopt + 0].getEntry(i*dimopt + 0);
					A[k*dimopt + 0].setEntry(i*dimopt + 0, curMatEntry + kiikMat[0][0]);
					curMatEntry = A[k*dimopt + 0].getEntry(i*dimopt + 1);
					A[k*dimopt + 0].setEntry(i*dimopt + 1, curMatEntry + kiikMat[1][0]);
					curMatEntry = A[k*dimopt + 0].getEntry(i*dimopt + 2);
					A[k*dimopt + 0].setEntry(i*dimopt + 2, curMatEntry + kiikMat[2][0]);

					curMatEntry = A[k*dimopt + 1].getEntry(i*dimopt + 0);
					A[k*dimopt + 1].setEntry(i*dimopt + 0, curMatEntry + kiikMat[0][1]);
					curMatEntry = A[k*dimopt + 1].getEntry(i*dimopt + 1);
					A[k*dimopt + 1].setEntry(i*dimopt + 1, curMatEntry + kiikMat[1][1]);
					curMatEntry = A[k*dimopt + 1].getEntry(i*dimopt + 2);
					A[k*dimopt + 1].setEntry(i*dimopt + 2, curMatEntry + kiikMat[2][1]);

					curMatEntry = A[k*dimopt + 2].getEntry(i*dimopt + 0);
					A[k*dimopt + 2].setEntry(i*dimopt + 0, curMatEntry + kiikMat[0][2]);
					curMatEntry = A[k*dimopt + 2].getEntry(i*dimopt + 1);
					A[k*dimopt + 2].setEntry(i*dimopt + 1, curMatEntry + kiikMat[1][2]);
					curMatEntry = A[k*dimopt + 2].getEntry(i*dimopt + 2);
					A[k*dimopt + 2].setEntry(i*dimopt + 2, curMatEntry + kiikMat[2][2]);
				}

				//difference in geodesic distance between target and source
				//dcur is returned above when retrieving the gradient
				dref = sourceDistLookup[k][j];

				B[k*dimopt + 0] += (dcur - dref) * kigrad[0];
				B[k*dimopt + 1] += (dcur - dref) * kigrad[1];
				B[k*dimopt + 2] += (dcur - dref) * kigrad[2];
			}//for (j...
		}//for (i...

		//solve the system
		solver.solve(A, X, B);

		//normalize the movement vector, need to compute min and max length
		lengthmin = 1e30f;
		lengthmax = 0.f;
		for (i = 0; i < numCorr; i++)
		{
			sol = makeVector3f(X[i * dimopt + 0], X[i * dimopt + 1], X[i * dimopt + 2]);
			length = norm(sol);
			sol.normalize();
			if (length > lengthmax) lengthmax = length;
			if (length < lengthmin) lengthmin = length;
			solution[i] = pair<float, Vector3f>(length, sol);
		}//for (i...
		lengthrange = lengthmax - lengthmin;
		rcplengthrange = 1.f / lengthrange;

		//move the points by the solution we found and project them back to the surface
		for (i = 0; i < numCorr; i++)
		{
			// if length of the solution is too small, then don't move
			//this is not actually a parameter because solution[i].second is normalized above, 
			//so the norm will only be small if it was the zero vector in the first place
			if (solution[i].second.getSqrNorm() < 1e-5)	continue;

			// scale the length of the solution vector accordingly
			solution[i].second *= (solution[i].first - lengthmin) * rcplengthrange;

			// move the feature just in small steps
			step = solution[i].second * stepSize;

			bool bprjt = match->getTargetManifold()->project(corrs[i].second + step, projPoint, projNorm);
			if(!bprjt)
				debugOutput << "improveMatch project failed!\n";

			corrs[i].second = projPoint;

			//doing this here is new, based on the assumption that geodesics on target surface are recomputed every step
			//within setMatch... some changes below relate to this... this only happens for non-fixed points because of above condition on solution[i].second.getSqrNorm()
			k = match->getMatchedSourceIndex(i);
			match->setMatch(k, corrs[i].second);
		}//for (i...

		//update geodesics on target surface
		match->updateTargetGeodesics();

		//compute current correspondence error
		error = evalError(match, sourceDistLookup, allTargetGeos, fixed);

		//if it is better than the previous error, store the result
		if (error < besterror)
		{
			//the following section of code is removed due to assumption that match->setMatch(...) causes recomputation of geodesic on target surface
			//which is necessary for accurate measurements
			besterror = error;
			match->backupMatchesAndGeodesics();
		}
		else
		{
			//here should retrieve previous matches that had lower error
			match->rollbackMatchesAndGeodesics();

			for (i = 0; i < numCorr; i++)
			{
				k = match->getMatchedSourceIndex(i);
				corrs[i].second = match->getMatch(k);
			}
		}

		//error threshold
		if (error < m_errorStopThresh || fabs(error - lasterror) < m_errorChangeStopThresh)//PARAMETERS
			bNoConv = false;

		lasterror = error;
	}//for (iter...

	m_nMatchesLastRun = match->getNumMatchedIndices();

#if AR4_DEBUG_OUTPUT
		debugOutput << "\tnumber of iterations: " << iter << "\n";
		debugOutput << "\tfinal error: " << besterror << "\n";
		debugOutput << "\ttime: " << convertTimeToString(timer.getDeltaValue()) << "\n";
#endif //AR4_DEBUG_OUTPUT
	
}

//concrete template-free class for coarse-to-fine sparse bundle adjustment
void AR4CoarseToFineSparseBundle3f::optimizeAtLevel(mpcard level, AR4PartialMatch<ManifoldTraits3f> *match)
{
	Timer timer;

	CGSolver<float> solver;

	SparseMatrix<float> A;
	SparseMatrix<float> Id;
	DynamicVector<float> X, B;

	mpcard i, j, k, m;
	mpcard numCorr = match->getNumMatchedIndices();
	mpcard iter;

	float32 error, besterror = 1e30f, lasterror;

	vector<pair<Vector3f, Vector3f>> corrs(numCorr);
	vector< pair<float, Vector3f> > solution(numCorr);
	vector<vector<mpcard>> corrDepend(numCorr), corrDependIndex(numCorr);

	float32 curMatEntry;
	float32 dcur, dref;

	vector<mpcard> validNbrs;

	Vector3f sol;
	float32 lengthmin, lengthmax, length, lengthrange, rcplengthrange;

	vector<bool> fixed(numCorr, false);

	float stepSize = m_stepScale;
	Vector3f step, projPoint, projNorm;

	Vector3f sourcePti, targetPti, sourcePtk, targetPtk;
	Vector3f ikgrad, kigrad;
	Matrix3f kiikMat, kikiMat;

	mpcard numfixed = 0, ifixed = 0, usedcols = 0;
	vector<mpcard> colmap(numCorr);

	if (m_bDebugOutput)
		debugOutput << "AR4CoarseToFineSparseBundle3f::optimizeAtLevel(...):\n";

	vector<vector<float>> sourceDistLookup(numCorr);
	vector<vector<mpcard>> allTargetGeos(numCorr);

	//collect geodesics _starting_ from matches
	for (i = 0; i < numCorr; i++)
	{
		k = match->getMatchedSourceIndex(i);
		corrs[i].first = match->getSourceManifold()->getDiscrPoint(k);
		corrs[i].second = match->getMatch(k);

		//get other matches for which we have valid geodesic measurements between this point and there
		//corrDepend[i] contains index in [0,numCorr-1] of correspondence ("correspondence index" instead of index within source surface)
		match->getValidMatchedGeodesicsIndicesLevel(k, level, validNbrs, corrDepend[i], allTargetGeos[i]);

		sourceDistLookup[i].resize(validNbrs.size());
		for (j = 0; j < validNbrs.size(); j++)
			sourceDistLookup[i][j] = match->sourceGeodesic(k, validNbrs[j]);
	}

	vector<pair<Vector3f, Vector3f>> corrsCopy;
	map<mpcard, mpcard> corrMap;
	mpcard iUseCorr = 0;
	for (i = 0; i < numCorr; i++)
	{
		if (sourceDistLookup[i].size() > 0)
		{
			corrMap[i] = iUseCorr;
			corrDepend[iUseCorr] = corrDepend[i];
			allTargetGeos[iUseCorr] = allTargetGeos[i];
			sourceDistLookup[iUseCorr] = sourceDistLookup[i];
			corrsCopy.push_back(corrs[i]);
			iUseCorr++;
		}
	}

	numCorr = iUseCorr;
	corrs.resize(numCorr);
	for (i = 0; i < numCorr; i++)
	{
		corrs[i] = corrsCopy[i];

		iUseCorr = 0;
		for (j = 0; j < corrDepend[i].size(); j++)
		{
			if (corrMap.find(corrDepend[i][j]) != corrMap.end())
			{
				corrDepend[i][iUseCorr] = corrMap[corrDepend[i][j]];
				iUseCorr++;
			}
		}
	}

	const mpcard dimopt = 3;
	const mpcard probSize = numCorr * dimopt;

	X.setDim(probSize);
	B.setDim(probSize);
	A.setRows(probSize);

	float lambda = 0.1f;
	Id.setRows(probSize);
	Id.setZero();
	for (i = 0; i< probSize; i++)
	{
		Id[i].setEntry(i, lambda);
	}

	//compute starting geodesic error
	m_lastStartError = besterror = lasterror = evalError(match, sourceDistLookup, allTargetGeos, fixed);
	if (m_bDebugOutput)
		debugOutput << "\tstarting error: " << besterror << "\n";

	match->backupMatchesAndGeodesics();

	bool bNoConv = true;//not yet converged

	//set solver parameters
	solver.maxNumIterations = m_nSolverIters;
	solver.stopRes = m_residualThreshold;

	X.setZero();

	//iterate
	for (iter = 0; iter < m_nIterations && bNoConv; iter++)
	{
		A.setZero();
		B.setZero();
		B = X * lambda;
		A = Id;

		//fill in the system
		for (i = 0; i < numCorr; i++)
		{
			sourcePti = corrs[i].first;
			targetPti = corrs[i].second;

			for (j = 0; j < corrDepend[i].size(); j++)
			{
				k = corrDepend[i][j];
				if (k == i)	continue;

				//i is the row of the matrix, k is the column
				//get correspondence k
				sourcePtk = corrs[k].first;
				targetPtk = corrs[k].second;

				//target geodesics and gradients
				match->targetGeodesic(allTargetGeos[i][j], ikgrad);
				for (m = 0; m < corrDepend[k].size(); m++)
				{
					if (corrDepend[k][m] == i)
						break;
				}
				dcur = match->targetGeodesic(allTargetGeos[k][m], kigrad);

				//compute 3x3 matrices
				kiikMat = multTwoVectors(kigrad, ikgrad);
				kikiMat = multTwoVectors(kigrad, kigrad);

				if (!fixed[k])
				{
					// set the i==k elements    

					curMatEntry = A[k*dimopt + 0].getEntry(k*dimopt + 0);
					A[k*dimopt + 0].setEntry(k*dimopt + 0, curMatEntry + kikiMat[0][0]);
					curMatEntry = A[k*dimopt + 0].getEntry(k*dimopt + 1);
					A[k*dimopt + 0].setEntry(k*dimopt + 1, curMatEntry + kikiMat[1][0]);
					curMatEntry = A[k*dimopt + 0].getEntry(k*dimopt + 2);
					A[k*dimopt + 0].setEntry(k*dimopt + 2, curMatEntry + kikiMat[2][0]);

					curMatEntry = A[k*dimopt + 1].getEntry(k*dimopt + 0);
					A[k*dimopt + 1].setEntry(k*dimopt + 0, curMatEntry + kikiMat[0][1]);
					curMatEntry = A[k*dimopt + 1].getEntry(k*dimopt + 1);
					A[k*dimopt + 1].setEntry(k*dimopt + 1, curMatEntry + kikiMat[1][1]);
					curMatEntry = A[k*dimopt + 1].getEntry(k*dimopt + 2);
					A[k*dimopt + 1].setEntry(k*dimopt + 2, curMatEntry + kikiMat[2][1]);

					curMatEntry = A[k*dimopt + 2].getEntry(k*dimopt + 0);
					A[k*dimopt + 2].setEntry(k*dimopt + 0, curMatEntry + kikiMat[0][2]);
					curMatEntry = A[k*dimopt + 2].getEntry(k*dimopt + 1);
					A[k*dimopt + 2].setEntry(k*dimopt + 1, curMatEntry + kikiMat[1][2]);
					curMatEntry = A[k*dimopt + 2].getEntry(k*dimopt + 2);
					A[k*dimopt + 2].setEntry(k*dimopt + 2, curMatEntry + kikiMat[2][2]);
				}

				if (!fixed[i])
				{
					// set i != k elements

					curMatEntry = A[i*dimopt + 0].getEntry(k*dimopt + 0);
					A[i*dimopt + 0].setEntry(k*dimopt + 0, curMatEntry + kiikMat[0][0]);
					curMatEntry = A[i*dimopt + 0].getEntry(k*dimopt + 1);
					A[i*dimopt + 0].setEntry(k*dimopt + 1, curMatEntry + kiikMat[1][0]);
					curMatEntry = A[i*dimopt + 0].getEntry(k*dimopt + 2);
					A[i*dimopt + 0].setEntry(k*dimopt + 2, curMatEntry + kiikMat[2][0]);

					curMatEntry = A[i*dimopt + 1].getEntry(k*dimopt + 0);
					A[i*dimopt + 1].setEntry(k*dimopt + 0, curMatEntry + kiikMat[0][1]);
					curMatEntry = A[i*dimopt + 1].getEntry(k*dimopt + 1);
					A[i*dimopt + 1].setEntry(k*dimopt + 1, curMatEntry + kiikMat[1][1]);
					curMatEntry = A[i*dimopt + 1].getEntry(k*dimopt + 2);
					A[i*dimopt + 1].setEntry(k*dimopt + 2, curMatEntry + kiikMat[2][1]);

					curMatEntry = A[i*dimopt + 2].getEntry(k*dimopt + 0);
					A[i*dimopt + 2].setEntry(k*dimopt + 0, curMatEntry + kiikMat[0][2]);
					curMatEntry = A[i*dimopt + 2].getEntry(k*dimopt + 1);
					A[i*dimopt + 2].setEntry(k*dimopt + 1, curMatEntry + kiikMat[1][2]);
					curMatEntry = A[i*dimopt + 2].getEntry(k*dimopt + 2);
					A[i*dimopt + 2].setEntry(k*dimopt + 2, curMatEntry + kiikMat[2][2]);
				}

				//difference in geodesic distance between target and source
				//dcur is returned above when retrieving the gradient
				dref = sourceDistLookup[i][j];

				B[k*dimopt + 0] += (dcur - dref) * kigrad[0];
				B[k*dimopt + 1] += (dcur - dref) * kigrad[1];
				B[k*dimopt + 2] += (dcur - dref) * kigrad[2];
			}//for (j...
		}//for (i...

		//solve the system
		solver.solve(A, X, B);

		//normalize the movement vector, need to compute min and max length
		lengthmin = 1e30f;
		lengthmax = 0.f;
		for (i = 0; i < numCorr; i++)
		{
			sol = makeVector3f(X[i * dimopt + 0], X[i * dimopt + 1], X[i * dimopt + 2]);
			length = norm(sol);
			sol.normalize();
			if (length > lengthmax) lengthmax = length;
			if (length < lengthmin) lengthmin = length;
			solution[i] = pair<float, Vector3f>(length, sol);
		}//for (i...
		lengthrange = lengthmax - lengthmin;
		rcplengthrange = 1.f / lengthrange;

		//move the points by the solution we found and project them back to the surface
		for (i = 0; i < numCorr; i++)
		{
			// if length of the solution is too small, then don't move
			//this is not actually a parameter because solution[i].second is normalized above, 
			//so the norm will only be small if it was the zero vector in the first place
			if (solution[i].second.getSqrNorm() < 1e-5)	continue;

			// scale the length of the solution vector accordingly
			solution[i].second *= (solution[i].first - lengthmin) * rcplengthrange;

			// move the feature just in small steps
			step = solution[i].second * stepSize;

			match->getTargetManifold()->project(corrs[i].second + step, projPoint, projNorm);

			corrs[i].second = projPoint;

			//doing this here is new, based on the assumption that geodesics on target surface are recomputed every step
			//within setMatch... some changes below relate to this... this only happens for non-fixed points because of above condition on solution[i].second.getSqrNorm()
			k = match->getMatchedSourceIndex(i);
			match->setMatch(k, corrs[i].second);
		}//for (i...

		//update geodesics on target surface
		match->updateTargetGeodesics();

		//compute current correspondence error
		error = evalError(match, sourceDistLookup, allTargetGeos, fixed);

		//if it is better than the previous error, store the result
		if (error < besterror)
		{
			//the following section of code is removed due to assumption that match->setMatch(...) causes recomputation of geodesic on target surface
			//which is necessary for accurate measurements
			besterror = error;
			match->backupMatchesAndGeodesics();
		}
		else
		{
			//here should retrieve previous matches that had lower error
			match->rollbackMatchesAndGeodesics();

			for (i = 0; i < numCorr; i++)
			{
				k = match->getMatchedSourceIndex(i);
				corrs[i].second = match->getMatch(k);
			}
		}

		//error threshold
		if (error < m_errorStopThresh || fabs(error - lasterror) < m_errorChangeStopThresh)//PARAMETERS
			bNoConv = false;

		lasterror = error;
	}//for (iter...

	m_nMatchesLastRun = match->getNumMatchedIndices();

	if (m_bDebugOutput)
	{
		debugOutput << "\tnumber of iterations: " << iter << "\n";
		debugOutput << "\tfinal error: " << besterror << "\n";
		debugOutput << "\ttime: " << convertTimeToString(timer.getDeltaValue()) << "\n";
	}
}

void AR4CoarseToFineSparseBundle3f::improveMatch(X4::AR4PartialMatch<ManifoldTraits3f> *match)
{
	const mpcard levelMax = match->getNumTopologyLevels() - 1;
	const mpcard levelMatch = match->getMatchingLevel();

	mpcard level;

	debugOutput << "AR4CoarseToFineSparseBundle3f::improveMatch(...):\n";
	for (level = levelMax; level >= levelMatch; level--)
	{
		debugOutput << "\tlevel " << level << "\n";
		optimizeAtLevel(level, match);
	}
}


///////////////////////////////////////////////////////////////////////////////
//growing matcher
///////////////////////////////////////////////////////////////////////////////

//concrete non-template class
void AR4GrowingMatcher3f::improveMatch(AR4PartialMatch<ManifoldTraits3f> *match)
{

	const mpcard MATCH_INVALID = (mpcard)-1;

	AR4IM_START_PROFILE(m_profStep)

	Timer timer;

	mpcard i, j, k, l, m, n;
	mpcard numCorr = match->getNumMatchedIndices();
	mpcard numNbrs;
	mpcard numRigid = 4;//PARAMETER

	vector< pair<Vector3f, Vector3f> > corrs(numCorr);

	vector<mpcard> corrsNewCandidateIndices, corrsNewIndices;
	vector< pair<Vector3f, Vector3f> > corrsNewCandidates, corrsNew;
	vector<Vector3f> sourcePoints(numCorr), targetPoints(numCorr);
	vector<Vector3f> rigidSource, rigidTarget;
	vector<mpcard> sourceNbrs;
	FloatType avgdist;

	FloatType geosigma = match->getGeodesicsSigma();
	FloatType error, errorba, errorThreshold, errorRatio, errorVar, errorStd;
	FloatType geodist, geomax;

	PointSet* pSourcePointSet;
	PointSet* pTargetPointSet;
	AAT normAATSource, normAATTarget;

	pair<Vector3f, Vector3f> corr;

	Vector3f srcpt, targpt;
	Vector3f srcnorm, targnorm;
	Vector3f candidatePoint, sourceNorm, sourceTang, sourceBinorm;
	TangentSpace sourceTS;

	Vector3f dbgcolor1, dbgcolor2;
	Vector3f step, projPoint, projNorm, targetTang, targetBinorm;
	TangentSpace targetTS;

	Matrix4f rigid;
	Vector4f rigidPoint;

	AR4PartialMatch3f* pTempMatch;

	AR4MLSManifold3f* pSourceMLS = dynamic_cast<AR4MLSManifold3f*>(match->getSourceManifold());
	if (pSourceMLS == NULL)
		throw ERangeCheck("AR4GrowingMatcher3f::improveMatch(...): source manifold is not an AR4MLSManifold");
	AR4MLSManifold3f* pTargetMLS = dynamic_cast<AR4MLSManifold3f*>(match->getTargetManifold());
	if (pTargetMLS == NULL)
		throw ERangeCheck("AR4GrowingMatcher3f::improveMatch(...): target manifold is not an Ar4MLSManifold");

#if AR4IM_INNER_DEBUG_RENDER
	debugRenderer->beginRenderJob_OneFrame(AR4IM_INNER_DEBUG_RENDER_JOBNAME, m_run);
#endif //AR4IM_INNER_DEBUG_RENDER

	//retrieve existing matches
	for (i = 0; i < numCorr; i++)
	{
		k = match->getMatchedSourceIndex(i);
		corrs[i].first = match->getSourceManifold()->getDiscrPoint(k);
		corrs[i].second = match->getMatch(k);
		sourcePoints[i] = corrs[i].first;
		targetPoints[i] = corrs[i].second;
	}//for (i...


	AR4IM_START_PROFILE(m_profAddPoints)


	//add points just outside the boundary of matched region
	findNewMatchCandidates(match, corrsNewCandidateIndices);

	
	//estimate matches by computing rigid transform and projecting onto surface
	//estimate rigid transform from existing matches
	//use normals; first compute average distance to other points for scaling
	if (m_projScaleSize < 0.f)
		m_projScaleSize = match->computeAvgSourceMatchLevelEdgeLength();
	avgdist = m_projScaleSize;

	pSourcePointSet = pSourceMLS->getPointSet();
	pTargetPointSet = pTargetMLS->getPointSet();
	normAATSource = pSourcePointSet->getAAT("normal");
	normAATTarget = pTargetPointSet->getAAT("normal");

	bool bprjsrc, bprjtrg;

	if (m_eGrowingMode == eGMParallelTransport)
	{

		const mpcard nParTranRelaxIters = 2;

		vector<mpcard> sourceMatchNbrs;
		vector<float> sourceGrowDists;
		vector<vector<Vector3f>> sourceGrowPaths;
		vector<Vector3f> sourceMatchPath, targetMatchPath;
		vector<Vector3f> targetGrowEstimates;

		Vector3f targptend;
		Vector3f dirSource, dirTarget, dirSourceNew, dirTargetNew;
		Vector3f crossSource, crossTarget;
		Vector3f grad, Dir;
		float dist, u, v, weight, weightSum;
		float growDistMax;

		for (j = 0; j < corrsNewCandidateIndices.size(); j++)
		{
			sourceMatchNbrs.clear();
			sourceGrowDists.clear();
			sourceGrowPaths.clear();
			targetGrowEstimates.clear();

			m = corrsNewCandidateIndices[j];

			if (m > match->getSourceManifold()->getNumDiscrPoints())
				debugOutput << "bad candidate index: " << m << "\n";

			//examine neighborhood to find previously matched points from which to triangulate new point
			//1-ring neighborhood is important in this case; 2-ring neighbors are too far away and parallel transport breaks down
			match->getSourceMatchingLevelNeighborhood(m, 1, sourceNbrs);
			numNbrs = sourceNbrs.size();
			sourceGrowPaths.resize(numNbrs);
			growDistMax = 0.f;

			for (l = 0; l < numNbrs; l++)
			{
				i = sourceNbrs[l];
				if (match->isMatched(i))
				{
					//retrieve geodesic information, esp. path from match
					dist = match->sourceGeodesic(i, m, sourceGrowPaths[sourceMatchNbrs.size()]);
					if (dist >= 0.f)
					{
						sourceGrowDists.push_back(dist);
						sourceMatchNbrs.push_back(i);
						if (dist > growDistMax)
							growDistMax = dist;
					}
				}
			}//for (l...
		
			//the following test checks whether or not we have tried to match this point before
			//if so, and we haven't since added any of its neighbors as matches, then the result will be the same
			//so we skip it
			if (m_matchedNbrCount[m] > 0 && (card32)sourceMatchNbrs.size() == m_matchedNbrCount[m])	continue;
			m_matchedNbrCount[m] = (card32)sourceMatchNbrs.size();

		
			//loop over already matched neighbors
			//for each, perform parallel transport to get estimate of corresponding point of new candidate (m) on target
			for (l = 0; l < sourceMatchNbrs.size(); l++)
			{

				//index of matched neighbor (of point m on source)
				i = sourceMatchNbrs[l];

				//get another neighbor so we can have a reference direction
				//in the process we will get the path between the two existing matches on the source
				//which we use below to get a common reference frame for the direction to the new point
				//match->sourceGeodesic(...) will return a distance < 0 if the two source indices do not have
				//a pre-computed geodesic between them
				if (l == 0)		n = sourceMatchNbrs.size() - 1;
				else			n = l - 1;
				k = sourceMatchNbrs[n];
				while (match->sourceGeodesic(i, k, sourceMatchPath) < 0.f && n != l)
				{
					if (n == 0)	n = sourceMatchNbrs.size() - 1;
					else		n--;
					k = sourceMatchNbrs[n];
				}

				//if we could not find any pre-computed geodesic between this matched neighbor and another matched neighbor
				//then we must skip this neighbor and not using it for triangulating the new correspondence via parallel transport
				//(I actually don't think this should happen)
				if (n == l)		continue;

				
				//relax the path to get more accurate direction
				AR4IM_START_PROFILE(m_profAddPoints_relax)
				match->getSparseSourceGeodesics()->setRelaxationSoftConstraint(0.01f);
				match->getSparseSourceGeodesics()->relaxPath(nParTranRelaxIters, sourceMatchPath, dist, grad,Dir);
				AR4IM_STOP_PROFILE(m_profAddPoints_relax)


				//existing correspondences on target
				srcpt = pSourceMLS->getDiscrPoint(i);
				targpt = match->getMatch(i);
				targptend = match->getMatch(k);



				//get the path between existing matches on target to get another known direction to fix reference for new direction
				//will be computed on-the-fly if not already computed
				match->targetGeodesic(i, k, targpt, targptend, targetMatchPath);

				
				
				//relax the path to get more accurate direction
				AR4IM_START_PROFILE(m_profAddPoints_relax)
				match->getSparseTargetGeodesics()->setRelaxationSoftConstraint(0.01f);
				match->getSparseTargetGeodesics()->relaxPath(nParTranRelaxIters, targetMatchPath, dist, grad, Dir);
				AR4IM_STOP_PROFILE(m_profAddPoints_relax)
	 

				//something went wrong with the path?compute the path on the source
				if(sourceMatchPath.size() < 2)
				{
					Vector3f derivative;
					match->getSparseSourceGeodesics()->computeGeodesicSingleDest(m_geodSigma, i, k, dist, derivative, sourceMatchPath);
		
				}


				//something went wrong with the path?compute the path on the target
				if(targetMatchPath.size() < 2)
				{
					Vector3f derivative;
					match->getSparseSourceGeodesics()->computeGeodesicSingleDest(m_geodSigma, targpt, targptend, dist, derivative, &targetMatchPath);
		
				}
					

				dirSource = sourceMatchPath[1] - sourceMatchPath[0];
				dirSource.normalize();
				
		
				dirTarget = targetMatchPath[1] - targetMatchPath[0];
				dirTarget.normalize();
				
					
				
				
			
				//now get local tangent space coordinate frame
				//on source
				bool bprojs = pSourceMLS->project(srcpt, projPoint, srcnorm);
				if(!bprojs) debugOutput << "AR4GrowingMatcher3f::improveMatch:: project on source failed\n";
				
				crossSource = srcnorm.crossProduct(dirSource);
				crossSource.normalize();
				dirSource = crossSource.crossProduct(srcnorm);
				dirSource.normalize();
				//on target
				bool bprojt = pTargetMLS->project(targpt, projPoint, targnorm);
				if(!bprojt) debugOutput << "AR4GrowingMatcher3f::improveMatch:: project on target failed\n";
				
				crossTarget = targnorm.crossProduct(dirTarget);
				crossTarget.normalize();
				dirTarget = crossTarget.crossProduct(targnorm);
				dirTarget.normalize();

			
				//now express new direction in tangent space coordinates on source and transfer to target
				dirSourceNew = sourceGrowPaths[l][1] - sourceGrowPaths[l][0];
				dirSourceNew.normalize();
				u = dirSourceNew * dirSource;
				v = dirSourceNew * crossSource;
				dirTargetNew = dirTarget * u + crossTarget * v;
				dirTargetNew.normalize();
				
		
				//now "walk the isometry" (parallel transport) to find corresponding point of new point m (re-using targptend)
				if (match->walkIsometry(i, m, sourceGrowPaths[l], targpt, dirTargetNew, targptend))
				{
					//save esimate
					sourceGrowDists[targetGrowEstimates.size()] = sourceGrowDists[l];
					targetGrowEstimates.push_back(targptend);
				}

		
			}//for (l...
	
			//estimates from parallel tranport may not agree due to deviations from isometry and numerical imprecision
			//average them while staying on the surface
			//the geodesic average/centroid is the point on the surface, which minimizes the sum of distances (on the surface) to the parallel transport estimates
			if (targetGrowEstimates.size() > 0)
			{
#if AR4IM_INNER_DEBUG_RENDER
				Vector3f ptcol = rainbowGradient3f((float)j / (float)corrsNewCandidateIndices.size());
				debugRenderer->addSphere(pSourceMLS->getDiscrPoint(m), 0.15f * avgdist, ptcol);
				debugOutput << "growDistMax = " << growDistMax << "\n";
#endif //AR4IM_INNER_DEBUG_RENDER

				match->getSparseTargetGeodesics()->clearDistanceField();
				for (l = 0; l < targetGrowEstimates.size(); l++)
				{
#if AR4IM_INNER_DEBUG_RENDER
					debugRenderer->addSphere(targetGrowEstimates[l], 0.1f * avgdist, NULL_VECTOR3F);
#endif //AR4IM_INNER_DEBUG_RENDER

					match->getSparseTargetGeodesics()->distanceTransform(vector<Vector3f>(1, targetGrowEstimates[l]), growDistMax * 2.f, l > 0);
				}

				//optionally, we could relax the distance field here (apply iterative Laplacian smoothing)

				//find the minimum distance to the estimates from parallel transport
				float distMin;
				match->getSparseTargetGeodesics()->getMinDistanceContinuousHint(match->getGeodesicsSigma() * 2, targetGrowEstimates, &targpt, &distMin);

#if AR4IM_INNER_DEBUG_RENDER
				debugOutput << "minimum distance (sum) to prediction points: " << distMin << "\n";
				debugRenderer->addSphere(targpt, 0.15f * avgdist, ptcol);
#endif //AR4IM_INNER_DEBUG_RENDER

				//project back to surface
				//first check that we were able to find a valid minimum -- if (one instance of) the parallel transport gives us a completely crazy estimate, we may not be able to
				//this check could even be more strict, like (distMin/targetGrowEstimates.size()) < growDistMax, or something
				if (distMin < AR4SparseGeodesics::m_HugeDistance && pTargetMLS->project(targpt, projPoint))
				{
					targpt = projPoint;
					corr.first = pSourceMLS->getDiscrPoint(m);
					corr.second = targpt;

					corrsNewCandidateIndices[corrsNewCandidates.size()] = m;
					corrsNewCandidates.push_back(corr);
				}//if projection successful
			}//if (targetGrowEstimates.size() > 0)
		}//for (j...
	
	}
	else if (m_eGrowingMode == eGMGeodesicTriangulation)
	{
		vector<mpcard> sourceMatchNbrs;
		vector<float> sourceGrowDists;
		vector<Vector3f> targetMatchNbrs;
		vector<Vector3f> targetGrowEstimates;
		vector<vector<Vector3f>> contours;
		vector<Vector3f> otherContourPoints;

		Vector3f targptend;
		Vector3f dirSource, dirTarget, dirSourceNew, dirTargetNew;
		Vector3f crossSource, crossTarget;
		float dist, u, v, weight, weightSum;
		float growDistMax;

		for (j = 0; j < corrsNewCandidateIndices.size(); j++)
		{
			sourceMatchNbrs.clear();
			sourceGrowDists.clear();
			targetMatchNbrs.clear();
			growDistMax = 0.f;

			m = corrsNewCandidateIndices[j];

			candidatePoint = pSourceMLS->getDiscrPoint(m);

			//examine neighborhood to find previously matched points from which to triangulate new point
			match->getSourceMatchingLevelNeighborhood(m, 2, sourceNbrs);
			numNbrs = sourceNbrs.size();
			for (l = 0; l < numNbrs; l++)
			{
				i = sourceNbrs[l];
				if (match->isMatched(i))
				{
					//retrieve geodesic distance to use for triangulation, and match index and corresponding point
					dist = match->sourceGeodesic(i, m);
					if (dist >= 0.f)
					{
						sourceGrowDists.push_back(dist);
						sourceMatchNbrs.push_back(i);
						targetMatchNbrs.push_back(match->getMatch(i));

						if (dist > growDistMax)
							growDistMax = dist;
					}
				}
			}//for (l...

			if (sourceMatchNbrs.size() < 3)
				continue;

			contours.resize(sourceMatchNbrs.size());

			card32 contourPointCount = 0;
			for (l = 0; l < sourceMatchNbrs.size(); l++)
			{
				//take geodesic distance on source, and use this to define a radius to grow to on the target
				//extract points on this iso-contour
				match->getSparseTargetGeodesics()->computeGeodesicIsoContour(targetMatchNbrs[l], sourceGrowDists[l], contours[l]);
				contourPointCount += (card32)contours[l].size();
			}

			match->getSparseTargetGeodesics()->clearDistanceField();

			//now we compute the sum of distance transforms using the contours as zero sets
			//the result is that the distance field on the target stores the sum of distances to the contours
			//the point that minimizes this sum is the best estimate of the corresponding point according to geodesic triangulation
			otherContourPoints.resize(contourPointCount);
			i = 0;
			for (l = 0; l < sourceMatchNbrs.size(); l++)
			{
				for (n = 0; n < contours[l].size(); n++)
				{
					otherContourPoints[i] = contours[l][n];
					i++;
				}
			}
			for (l = 0; l < sourceMatchNbrs.size(); l++)
			{
				//compute distance transform wrt contours[l]
				//stop once all of otherContourPoints have had their distance computed
				match->getSparseTargetGeodesics()->distanceTransform(contours[l], growDistMax, (l > 0));
			}

#if AR4IM_INNER_DEBUG_RENDER
			if (j == 0)
			{
				float distMax = 0.f;
				for (n = 0; n < otherContourPoints.size(); n++)
				{
					dist = match->getSparseTargetGeodesics()->getDistanceField(otherContourPoints[n]);
					if (dist > distMax)
						distMax = dist;
				}
				debugOutput << "max dist = " << distMax << "\n";
				for (n = 0; n < otherContourPoints.size(); n++)
				{
					dist = match->getSparseTargetGeodesics()->getDistanceField(otherContourPoints[n]);
					debugRenderer->addSphere(otherContourPoints[n], 0.1 * avgdist, rainbowGradient3f(dist / distMax));
				}
			}
#endif //AR4IM_INNER_DEBUG_RENDER

			//find the minimum of the distance field
			match->getSparseTargetGeodesics()->getMinDistanceContinuousHint(match->getGeodesicsSigma(), otherContourPoints, &targpt, &dist);
#if AR4IM_INNER_DEBUG_RENDER
			if (j == 0)
			{
				debugOutput << "min dist = " << dist << "\n";
				debugRenderer->addSphere(targpt, avgdist * 0.15f, makeVector3f(0.f, 0.f, 0.f));
			}
#endif //AR4IM_INNER_DEBUG_RENDER

			if (pTargetMLS->project(targpt, projPoint))
			{
				corr.first = candidatePoint;
				corr.second = projPoint;

				corrsNewCandidateIndices[corrsNewCandidates.size()] = m;
				corrsNewCandidates.push_back(corr);
			}
		}//for (j...
	}
	else
	{
		for (j = 0; j < corrsNewCandidateIndices.size(); j++)
		{
			m = corrsNewCandidateIndices[j];

			k = 0;
			rigidSource.clear();
			rigidTarget.clear();

			match->getSourceMatchingLevelNeighborhood(m, 2, sourceNbrs);
			numNbrs = sourceNbrs.size();
			for (l = 0; l < numNbrs; l++)
			{
				i = sourceNbrs[l];
				if (match->isMatched(i))
				{
					srcpt = pSourceMLS->getDiscrPoint(i);
					bprjsrc = pSourceMLS->project(srcpt, projPoint, projNorm, sourceTS);
					srcpt = projPoint;
					srcnorm = projNorm;

					targpt = match->getMatch(i);
					bprjtrg = pTargetMLS->project(targpt, projPoint, projNorm, targetTS);
					targpt = projPoint;
					targnorm = projNorm;

					if (bprjsrc && bprjtrg)
					{
						rigidSource.push_back(srcpt);
						rigidSource.push_back(srcpt + srcnorm * avgdist);
						rigidTarget.push_back(targpt);
						rigidTarget.push_back(targpt + targnorm * avgdist);
					}
					else
						continue;

					if (k == 1)
					{
						Vector3f dir;
						dir = rigidSource[0] - srcpt;
						dir.normalize();
						dir = srcnorm.crossProduct(dir);
						dir.normalize();
						rigidSource.push_back(srcpt + dir * avgdist);

						dir = rigidTarget[0] - targpt;
						dir.normalize();
						dir = targnorm.crossProduct(dir);
						dir.normalize();
						rigidTarget.push_back(targpt + dir * avgdist);
					}
					k++;
				}
			}//for (l...

			if (match->canBeTriangulated(m) && rigidSource.size() >= 4)
			{
				rigid = computeRigidTransformationFromCorrespondences(rigidTarget, rigidSource);

				candidatePoint = pSourceMLS->getDiscrPoint(m);
				rigidPoint = rigid * makeVector4f(candidatePoint[0], candidatePoint[1], candidatePoint[2], 1);
				bprjtrg = pTargetMLS->project(makeVector3f(rigidPoint[0], rigidPoint[1], rigidPoint[2]), projPoint, projNorm);

				if (bprjtrg)
				{
					corr.first = candidatePoint;
					corr.second = projPoint;

					corrsNewCandidateIndices[corrsNewCandidates.size()] = m;
					corrsNewCandidates.push_back(corr);
				}
			}
		}//for (j...

	}//if (m_eGrowingMode)

	AR4IM_STOP_PROFILE(m_profAddPoints)



	

	AR4IM_START_PROFILE(m_profLocalFit)

	//apply (local/dense) bundle-adjustment / GMDS to enforce isometry
	//applied separate for each new candidate point (below)
	AR4DenseBundle3f ba(source, target);

	ba.setDebugOutput(m_bDebugOutput);

	ba.setGeodesicSigma(geosigma);
	ba.setStepScale(avgdist * 0.05f);

	//create a local copy of the match, so that new points are only added to the original at the end
	pTempMatch = new AR4PartialMatch3f(source, target);
	pTempMatch->mimick(match);

	//fix all previous matches so that we are only optimizing one at a time
	pTempMatch->fixAllMatches();

	vector<mpcard> connectIndices, toTargGeos, fromTargGeos;
	vector<vector<mpcard>> allConnectIndices, allToTargGeos, allFromTargGeos;

#if AR4_DEBUG_OUTPUT	
	debugOutput << "found " << corrsNewCandidates.size() << " new potential matches\n";
#endif //AR4_DEBUG_OUTPUT

	for (i = 0; i < corrsNewCandidates.size(); i++)
	{
		//add new correspondence candidate to temporary match instance
		corr = corrsNewCandidates[i];
		k = corrsNewCandidateIndices[i];
		pTempMatch->setMatch(k, corr.second);

		//apply local metric fitting
		ba.improveMatch(pTempMatch);

		//retrieve new position on target surface
		corr.second = pTempMatch->getMatch(k);

		//measure error before and after bundle
		error = ba.getLastStartError();
		errorba = ba.getLastEndError();
		if (m_bDebugOutput)
		{
			debugOutput << "\tgeodesic error before bundle: " << error << "\n";
			debugOutput << "\tgeodesic error after bundle:  " << errorba << "\n";
		}

		if (errorba <= error)
		{
			//add new correspondence to (preliminarily) accepted ones
			//collect geodesic information computed in temp match for this point
			corrsNew.push_back(corr);
			corrsNewIndices.push_back(k);
			pTempMatch->collectAllMatchedGeodesics(k, connectIndices, toTargGeos, fromTargGeos);
			allConnectIndices.push_back(connectIndices);
			allToTargGeos.push_back(toTargGeos);
			allFromTargGeos.push_back(fromTargGeos);
		}

		//kick this correspondence back out of this match so that it does not influence the next one
		pTempMatch->removeLastMatch();


	}//for (i...


	//add new correspondences to original match
	for (i = 0; i < corrsNew.size(); i++)
	{		
		match->setMatchWithGeoData(corrsNewIndices[i], corrsNew[i].second, allConnectIndices[i], allToTargGeos[i], allFromTargGeos[i]);
		
	}	

	AR4IM_STOP_PROFILE(m_profLocalFit)

	


	/////////////////////////////////
	//end-of-growing-step management:
	/////////////////////////////////

	AR4IM_START_PROFILE(m_profEndStep)

	//find any additional geodesic connections (between matches) that were not computed during local bundle above
	match->checkForNewGeodesicConnections();

	//validate new paths
	match->validatePaths();

	//remove outliers from among the newly added points (based on metric distortion)
	match->removeOutliersLastCohort();
	match->finalizeMatchCohort();

	AR4IM_STOP_PROFILE(m_profEndStep)
	
		
	


	delete pTempMatch;

#if AR4IM_INNER_DEBUG_RENDER
	debugRenderer->endRenderJob();
#endif //AR4IM_INNER_DEBUG_RENDER

	if (m_bDebugOutput)
		debugOutput << "\trun " << m_run << " - time: " << convertTimeToString(timer.getDeltaValue()) << "\n";
	m_run++;

	AR4IM_STOP_PROFILE(m_profStep)


		debugOutput.flush();



}

void AR4MatchInterpolator3f::improveMatch(AR4PartialMatch<ManifoldTraits3f> *match)
{
	const mpcard MATCH_INVALID = (mpcard)-1;

	AR4IM_START_PROFILE(m_profStep)

	Timer timer;

	mpcard i, j, k, l, m, n;
	mpcard numCorr = match->getNumMatchedIndices();
	mpcard numNbrs;
	mpcard numRigid = 4;//PARAMETER

	vector< pair<Vector3f, Vector3f> > corrs(numCorr);

	vector<mpcard> corrsNewCandidateIndices, corrsNewIndices;
	vector< pair<Vector3f, Vector3f> > corrsNewCandidates, corrsNew;
	vector<Vector3f> sourcePoints(numCorr), targetPoints(numCorr);
	vector<Vector3f> rigidSource, rigidTarget;
	vector<mpcard> sourceNbrs;
	FloatType avgdist;

	FloatType geosigma = match->getGeodesicsSigma();
	FloatType error, errorba, errorThreshold, errorRatio, errorVar, errorStd;
	FloatType geodist, geomax;

	PointSet* pSourcePointSet;
	PointSet* pTargetPointSet;
	AAT normAATSource, normAATTarget;

	pair<Vector3f, Vector3f> corr;

	Vector3f srcpt, targpt;
	Vector3f srcnorm, targnorm;
	Vector3f candidatePoint, sourceNorm, sourceTang, sourceBinorm;
	TangentSpace sourceTS;

	Vector3f dbgcolor1, dbgcolor2;
	Vector3f step, projPoint, projNorm, targetTang, targetBinorm;
	TangentSpace targetTS;

	Matrix4f rigid;
	Vector4f rigidPoint;

	AR4PartialMatch3f* pTempMatch;

	AR4MLSManifold3f* pSourceMLS = dynamic_cast<AR4MLSManifold3f*>(match->getSourceManifold());
	if (pSourceMLS == NULL)
		throw ERangeCheck("AR4MatchInterpolator3f::improveMatch(...): source manifold is not an AR4MLSManifold");
	AR4MLSManifold3f* pTargetMLS = dynamic_cast<AR4MLSManifold3f*>(match->getTargetManifold());
	if (pTargetMLS == NULL)
		throw ERangeCheck("AR4MatchInterpolator3f::improveMatch(...): target manifold is not an Ar4MLSManifold");

#if AR4IM_INNER_DEBUG_RENDER
	debugRenderer->beginRenderJob_OneFrame(AR4IM_INNER_DEBUG_RENDER_JOBNAME, m_run);
#endif //AR4IM_INNER_DEBUG_RENDER

	//retrieve existing matches
	for (i = 0; i < numCorr; i++)
	{
		k = match->getMatchedSourceIndex(i);
		corrs[i].first = match->getSourceManifold()->getDiscrPoint(k);
		corrs[i].second = match->getMatch(k);
		sourcePoints[i] = corrs[i].first;
		targetPoints[i] = corrs[i].second;
	}//for (i...

	AR4IM_START_PROFILE(m_profAddPoints)

	//add points just outside the boundary of matched region
	findNewMatchCandidates(match, corrsNewCandidateIndices);

	//estimate matches by computing rigid transform and projecting onto surface
	//estimate rigid transform from existing matches
	//use normals; first compute average distance to other points for scaling
	if (m_projScaleSize < 0.f)
		m_projScaleSize = match->computeAvgSourceEdgeLength();
	avgdist = m_projScaleSize;

	pSourcePointSet = pSourceMLS->getPointSet();
	pTargetPointSet = pTargetMLS->getPointSet();
	normAATSource = pSourcePointSet->getAAT("normal");
	normAATTarget = pTargetPointSet->getAAT("normal");

	bool bprjsrc, bprjtrg;

	if (m_eGrowingMode == eGMParallelTransport)
	{
		const mpcard nParTranRelaxIters = 2;

		vector<mpcard> sourceMatchNbrs;
		vector<float> sourceGrowDists;
		vector<vector<Vector3f>> sourceGrowPaths;
		vector<Vector3f> sourceMatchPath, targetMatchPath, sourceGrowCopy;
		vector<Vector3f> targetGrowEstimates;

		Vector3f targptend;
		Vector3f dirSource, dirTarget, dirSourceNew, dirTargetNew;
		Vector3f crossSource, crossTarget;
		Vector3f grad, Dir;
		float dist, u, v, weight, weightSum;
		float growDistMax;

		for (j = 0; j < corrsNewCandidateIndices.size(); j++)
		{
			sourceMatchNbrs.clear();
			sourceGrowDists.clear();
			sourceGrowPaths.clear();
			targetGrowEstimates.clear();

			m = corrsNewCandidateIndices[j];

			//examine neighborhood to find previously matched points from which to triangulate new point
			//1-ring neighborhood is important in this case; 2-ring neighbors are too far away and parallel transport breaks down
			match->getSourceNeighborhood(m, 1, sourceNbrs);
			numNbrs = sourceNbrs.size();
			sourceGrowPaths.resize(numNbrs);
			growDistMax = 0.f;
			for (l = 0; l < numNbrs; l++)
			{
				i = sourceNbrs[l];
				if (match->isMatched(i))
				{
					//retrieve geodesic information, esp. path from match
					dist = match->sourceGeodesic(i, m, sourceGrowPaths[sourceMatchNbrs.size()]);
					if (dist >= 0.f)
					{
						sourceGrowDists.push_back(dist);
						sourceMatchNbrs.push_back(i);
						if (dist > growDistMax)
							growDistMax = dist;
					}
					else
					{
						match->getSparseSourceGeodesics()->computeGeodesicSingleDest(0.01f * match->getGeodesicsSigma(), i, m, dist, grad, sourceGrowPaths[sourceMatchNbrs.size()]);
						match->getSparseSourceGeodesics()->storeGeodesicData(i, m, dist, grad, &sourceGrowPaths[sourceMatchNbrs.size()]);
						sourceGrowCopy = sourceGrowPaths[sourceMatchNbrs.size()];
						reverse(sourceGrowCopy.begin(), sourceGrowCopy.end());
						match->getSparseSourceGeodesics()->storeGeodesicData(m, i, dist, -grad, &sourceGrowCopy);
						sourceGrowDists.push_back(dist);
						sourceMatchNbrs.push_back(i);
					}
				}
			}//for (l...

			//the following test checks whether or not we have tried to match this point before
			//if so, and we haven't since added any of its neighbors as matches, then the result will be the same
			//so we skip it
			if (m_matchedNbrCount[m] > 0 && (card32)sourceMatchNbrs.size() == m_matchedNbrCount[m])	continue;
			m_matchedNbrCount[m] = (card32)sourceMatchNbrs.size();

			const mpcard numSourceMatchNbrs = sourceMatchNbrs.size();
			if (numSourceMatchNbrs <= 1)
				continue;

			//loop over already matched neighbors
			//for each, perform parallel transport to get estimate of corresponding point of new candidate (m) on target
			for (l = 0; l < numSourceMatchNbrs; l++)
			{
				//index of matched neighbor (of point m on source)
				i = sourceMatchNbrs[l];

				//get another neighbor so we can have a reference direction
				//in the process we will get the path between the two existing matches on the source
				//which we use below to get a common reference frame for the direction to the new point
				//match->sourceGeodesic(...) will return a distance < 0 if the two source indices do not have
				//a pre-computed geodesic between them
				if (l == 0)		n = numSourceMatchNbrs - 1;
				else			n = l - 1;
				k = sourceMatchNbrs[n];
				while (match->sourceGeodesic(i, k, sourceMatchPath) < 0.f && n != l)
				{
					if (n == 0)	n = numSourceMatchNbrs - 1;
					else		n--;
					k = sourceMatchNbrs[n];
				}

				//if we could not find any pre-computed geodesic between this matched neighbor and another matched neighbor
				//then we must choose one other neighbor arbitrarily to obtain a direction
				if (n == l)
				{
					//take an adjacent point
					if (l == numSourceMatchNbrs - 1)	n = 0;
					else								n = l + 1;
					k = sourceMatchNbrs[n];
					match->getSparseSourceGeodesics()->computeGeodesicSingleDest(0.01f * match->getGeodesicsSigma(), i, k, dist, grad, sourceMatchPath);
				}

				//relax the path to get more accurate direction
				AR4IM_START_PROFILE(m_profAddPoints_relax)
				match->getSparseSourceGeodesics()->setRelaxationSoftConstraint(0.01f);
				match->getSparseSourceGeodesics()->relaxPath(nParTranRelaxIters, sourceMatchPath, dist, grad, Dir);
				AR4IM_STOP_PROFILE(m_profAddPoints_relax)

				//existing correspondences on target
				srcpt = pSourceMLS->getDiscrPoint(i);
				targpt = match->getMatch(i);
				targptend = match->getMatch(k);

				//get the path between existing matches on target to get another known direction to fix reference for new direction
				//will be computed on-the-fly if not already computed
				match->targetGeodesic(i, k, targpt, targptend, targetMatchPath);

				if (targetMatchPath.size() < 2)
				{
					warning("in run " + intToStr(m_run) + ": single node path (pre-relax) between points " + targpt.toString() + " and " + targptend.toString());
					continue;
				}

				//relax the path to get more accurate direction
				AR4IM_START_PROFILE(m_profAddPoints_relax)
				match->getSparseTargetGeodesics()->setRelaxationSoftConstraint(0.01f);
				match->getSparseTargetGeodesics()->relaxPath(nParTranRelaxIters, targetMatchPath, dist, grad, Dir);
				AR4IM_STOP_PROFILE(m_profAddPoints_relax)

				if (sourceMatchPath.size() < 2)
				{
					warning("single node path between points " + intToStr(i) + " and " + intToStr(k));
					continue;
				}
				if (targetMatchPath.size() < 2)
				{
					warning("single node path between points " + targpt.toString() + " and " + targptend.toString());
					continue;
				}

				//now get corresponding directions
				//first from the paths
				dirSource = sourceMatchPath[1] - sourceMatchPath[0];
				dirSource.normalize();
				dirTarget = targetMatchPath[1] - targetMatchPath[0];
				dirTarget.normalize();

				//now get local tangent space coordinate frame
				//on source
				pSourceMLS->project(srcpt, projPoint, srcnorm);
				crossSource = srcnorm.crossProduct(dirSource);
				crossSource.normalize();
				dirSource = crossSource.crossProduct(srcnorm);
				dirSource.normalize();
				//on target
				pTargetMLS->project(targpt, projPoint, targnorm);
				crossTarget = targnorm.crossProduct(dirTarget);
				crossTarget.normalize();
				dirTarget = crossTarget.crossProduct(targnorm);
				dirTarget.normalize();

				//now express new direction in tangent space coordinates on source and transfer to target
				dirSourceNew = sourceGrowPaths[l][1] - sourceGrowPaths[l][0];
				dirSourceNew.normalize();
				u = dirSourceNew * dirSource;
				v = dirSourceNew * crossSource;
				dirTargetNew = dirTarget * u + crossTarget * v;
				dirTargetNew.normalize();

				//now "walk the isometry" (parallel transport) to find corresponding point of new point m (re-using targptend)
				if (match->walkIsometry(i, m, sourceGrowPaths[l], targpt, dirTargetNew, targptend))
				{
					//save esimate
					sourceGrowDists[targetGrowEstimates.size()] = sourceGrowDists[l];
					targetGrowEstimates.push_back(targptend);
				}
			}//for (l...

			//estimates from parallel tranport may not agree due to deviations from isometry and numerical imprecision
			//average them while staying on the surface
			//the geodesic average/centroid is the point on the surface, which minimizes the sum of distances (on the surface) to the parallel transport estimates
			if (targetGrowEstimates.size() > 0)
			{
#if AR4IM_INNER_DEBUG_RENDER
				Vector3f ptcol = rainbowGradient3f((float)j / (float)corrsNewCandidateIndices.size());
				debugRenderer->addSphere(pSourceMLS->getDiscrPoint(m), 0.15f * avgdist, ptcol);
				debugOutput << "growDistMax = " << growDistMax << "\n";
#endif //AR4IM_INNER_DEBUG_RENDER

				match->getSparseTargetGeodesics()->clearDistanceField();
				for (l = 0; l < targetGrowEstimates.size(); l++)
				{
#if AR4IM_INNER_DEBUG_RENDER
					debugRenderer->addSphere(targetGrowEstimates[l], 0.1f * avgdist, NULL_VECTOR3F);
#endif //AR4IM_INNER_DEBUG_RENDER

					match->getSparseTargetGeodesics()->distanceTransform(vector<Vector3f>(1, targetGrowEstimates[l]), growDistMax * 2.f, l > 0);
				}

				//optionally, we could relax the distance field here (apply iterative Laplacian smoothing)

				//find the minimum distance to the estimates from parallel transport
				float distMin;
				match->getSparseTargetGeodesics()->getMinDistanceContinuousHint(match->getGeodesicsSigma() * 2, targetGrowEstimates, &targpt, &distMin);

#if AR4IM_INNER_DEBUG_RENDER
				debugOutput << "minimum distance (sum) to prediction points: " << distMin << "\n";
				debugRenderer->addSphere(targpt, 0.15f * avgdist, ptcol);
#endif //AR4IM_INNER_DEBUG_RENDER

				//project back to surface
				//first check that we were able to find a valid minimum -- if (one instance of) the parallel transport gives us a completely crazy estimate, we may not be able to
				//this check could even be more strict, like (distMin/targetGrowEstimates.size()) < growDistMax, or something
				if (distMin < AR4SparseGeodesics::m_HugeDistance && pTargetMLS->project(targpt, projPoint))
				{
					targpt = projPoint;
					corr.first = pSourceMLS->getDiscrPoint(m);
					corr.second = targpt;

					corrsNewCandidateIndices[corrsNewCandidates.size()] = m;
					corrsNewCandidates.push_back(corr);
				}//if projection successful
			}//if (targetGrowEstimates.size() > 0)
		}//for (j...
	}
	else if (m_eGrowingMode == eGMGeodesicTriangulation)
	{
		vector<mpcard> sourceMatchNbrs;
		vector<float> sourceGrowDists;
		vector<Vector3f> targetMatchNbrs;
		vector<Vector3f> targetGrowEstimates;
		vector<vector<Vector3f>> contours;
		vector<Vector3f> otherContourPoints;

		Vector3f grad;
		Vector3f targptend;
		Vector3f dirSource, dirTarget, dirSourceNew, dirTargetNew;
		Vector3f crossSource, crossTarget;
		float dist, u, v, weight, weightSum;
		float growDistMax;

		for (j = 0; j < corrsNewCandidateIndices.size(); j++)
		{
			sourceMatchNbrs.clear();
			sourceGrowDists.clear();
			targetMatchNbrs.clear();
			growDistMax = 0.f;

			m = corrsNewCandidateIndices[j];

			candidatePoint = pSourceMLS->getDiscrPoint(m);

			//examine neighborhood to find previously matched points from which to triangulate new point
			match->getSourceNeighborhood(m, 2, sourceNbrs);
			numNbrs = sourceNbrs.size();
			for (l = 0; l < numNbrs; l++)
			{
				i = sourceNbrs[l];
				if (match->isMatched(i))
				{
					//retrieve geodesic distance to use for triangulation, and match index and corresponding point
					dist = match->sourceGeodesic(i, m);
					if (dist >= 0.f)
					{
						sourceGrowDists.push_back(dist);
						sourceMatchNbrs.push_back(i);
						targetMatchNbrs.push_back(match->getMatch(i));

						if (dist > growDistMax)
							growDistMax = dist;
					}
					else
					{
						match->getSparseSourceGeodesics()->computeGeodesicSingleDest(0.01f * match->getGeodesicsSigma(), i, m, dist, grad);
						sourceGrowDists.push_back(dist);
						sourceMatchNbrs.push_back(i);
						targetMatchNbrs.push_back(match->getMatch(i));

						if (dist > growDistMax)
							growDistMax = dist;
					}
				}
			}//for (l...

			if (sourceMatchNbrs.size() < 3)
				continue;

			contours.resize(sourceMatchNbrs.size());

			card32 contourPointCount = 0;
			for (l = 0; l < sourceMatchNbrs.size(); l++)
			{
				//take geodesic distance on source, and use this to define a radius to grow to on the target
				//extract points on this iso-contour
				match->getSparseTargetGeodesics()->computeGeodesicIsoContour(targetMatchNbrs[l], sourceGrowDists[l], contours[l]);
				contourPointCount += (card32)contours[l].size();
			}

			match->getSparseTargetGeodesics()->clearDistanceField();

			//now we compute the sum of distance transforms using the contours as zero sets
			//the result is that the distance field on the target stores the sum of distances to the contours
			//the point that minimizes this sum is the best estimate of the corresponding point according to geodesic triangulation
			otherContourPoints.resize(contourPointCount);
			i = 0;
			for (l = 0; l < sourceMatchNbrs.size(); l++)
			{
				for (n = 0; n < contours[l].size(); n++)
				{
					otherContourPoints[i] = contours[l][n];
					i++;
				}
			}
			for (l = 0; l < sourceMatchNbrs.size(); l++)
			{
				//compute distance transform wrt contours[l]
				//stop once all of otherContourPoints have had their distance computed
				match->getSparseTargetGeodesics()->distanceTransform(contours[l], growDistMax, (l > 0));
			}

#if AR4IM_INNER_DEBUG_RENDER
			if (j == 0)
			{
				float distMax = 0.f;
				for (n = 0; n < otherContourPoints.size(); n++)
				{
					dist = match->getSparseTargetGeodesics()->getDistanceField(otherContourPoints[n]);
					if (dist > distMax)
						distMax = dist;
				}
				debugOutput << "max dist = " << distMax << "\n";
				for (n = 0; n < otherContourPoints.size(); n++)
				{
					dist = match->getSparseTargetGeodesics()->getDistanceField(otherContourPoints[n]);
					debugRenderer->addSphere(otherContourPoints[n], 0.1 * avgdist, rainbowGradient3f(dist / distMax));
				}
			}
#endif //AR4IM_INNER_DEBUG_RENDER

			//find the minimum of the distance field
			match->getSparseTargetGeodesics()->getMinDistanceContinuousHint(match->getGeodesicsSigma(), otherContourPoints, &targpt, &dist);
#if AR4IM_INNER_DEBUG_RENDER
			if (j == 0)
			{
				debugOutput << "min dist = " << dist << "\n";
				debugRenderer->addSphere(targpt, avgdist * 0.15f, makeVector3f(0.f, 0.f, 0.f));
			}
#endif //AR4IM_INNER_DEBUG_RENDER

			if (pTargetMLS->project(targpt, projPoint))
			{
				corr.first = candidatePoint;
				corr.second = projPoint;

				corrsNewCandidateIndices[corrsNewCandidates.size()] = m;
				corrsNewCandidates.push_back(corr);
			}
		}//for (j...
	}
	else
	{
		for (j = 0; j < corrsNewCandidateIndices.size(); j++)
		{
			m = corrsNewCandidateIndices[j];

			k = 0;
			rigidSource.clear();
			rigidTarget.clear();

			match->getSourceNeighborhood(m, 2, sourceNbrs);
			numNbrs = sourceNbrs.size();
			for (l = 0; l < numNbrs; l++)
			{
				i = sourceNbrs[l];
				if (match->isMatched(i))
				{
					srcpt = pSourceMLS->getDiscrPoint(i);
					bprjsrc = pSourceMLS->project(srcpt, projPoint, projNorm, sourceTS);
					srcpt = projPoint;
					srcnorm = projNorm;

					targpt = match->getMatch(i);
					bprjtrg = pTargetMLS->project(targpt, projPoint, projNorm, targetTS);
					targpt = projPoint;
					targnorm = projNorm;

					if (bprjsrc && bprjtrg)
					{
						rigidSource.push_back(srcpt);
						rigidSource.push_back(srcpt + srcnorm * avgdist);
						rigidTarget.push_back(targpt);
						rigidTarget.push_back(targpt + targnorm * avgdist);
					}
					else
						continue;

					if (k == 1)
					{
						Vector3f dir;
						dir = rigidSource[0] - srcpt;
						dir.normalize();
						dir = srcnorm.crossProduct(dir);
						dir.normalize();
						rigidSource.push_back(srcpt + dir * avgdist);

						dir = rigidTarget[0] - targpt;
						dir.normalize();
						dir = targnorm.crossProduct(dir);
						dir.normalize();
						rigidTarget.push_back(targpt + dir * avgdist);
					}
					k++;
				}
			}//for (l...

			if (match->canBeTriangulated(m) && rigidSource.size() >= 4)
			{
				rigid = computeRigidTransformationFromCorrespondences(rigidTarget, rigidSource);

				candidatePoint = pSourceMLS->getDiscrPoint(m);
				rigidPoint = rigid * makeVector4f(candidatePoint[0], candidatePoint[1], candidatePoint[2], 1);
				bprjtrg = pTargetMLS->project(makeVector3f(rigidPoint[0], rigidPoint[1], rigidPoint[2]), projPoint, projNorm);

				if (bprjtrg)
				{
					corr.first = candidatePoint;
					corr.second = projPoint;

					corrsNewCandidateIndices[corrsNewCandidates.size()] = m;
					corrsNewCandidates.push_back(corr);
				}
			}
		}//for (j...
	}//if (m_eGrowingMode)

	AR4IM_STOP_PROFILE(m_profAddPoints)

	AR4IM_START_PROFILE(m_profLocalFit)

	//apply (local/dense) bundle-adjustment / GMDS to enforce isometry
	//applied separate for each new candidate point (below)
	AR4DenseBundle3f ba(source, target);

	ba.setDebugOutput(m_bDebugOutput);

	ba.setGeodesicSigma(geosigma);
	ba.setStepScale(avgdist * 0.05f);

	//create a local copy of the match, so that new points are only added to the original at the end
	pTempMatch = new AR4PartialMatch3f(source, target);
	pTempMatch->mimick(match);

	//fix all previous matches so that we are only optimizing one at a time
	pTempMatch->fixAllMatches();

	vector<mpcard> connectIndices, toTargGeos, fromTargGeos;
	vector<vector<mpcard>> allConnectIndices, allToTargGeos, allFromTargGeos;
	

	debugOutput << "found " << corrsNewCandidates.size() << " new potential matches\n";
	for (i = 0; i < corrsNewCandidates.size(); i++)
	{
		//add new correspondence candidate to temporary match instance
		corr = corrsNewCandidates[i];
		k = corrsNewCandidateIndices[i];
		pTempMatch->setMatch(k, corr.second);

		//apply local metric fitting
		ba.improveMatch(pTempMatch);

		//retrieve new position on target surface
		corr.second = pTempMatch->getMatch(k);

		//measure error before and after bundle
		error = ba.getLastStartError();
		errorba = ba.getLastEndError();
		if (m_bDebugOutput)
		{
			debugOutput << "\tgeodesic error before bundle: " << error << "\n";
			debugOutput << "\tgeodesic error after bundle:  " << errorba << "\n";
		}

		if (errorba <= error)
		{
			//add new correspondence to (preliminarily) accepted ones
			//collect geodesic information computed in temp match for this point
			corrsNew.push_back(corr);
			corrsNewIndices.push_back(k);
			pTempMatch->collectAllMatchedGeodesics(k, connectIndices, toTargGeos, fromTargGeos);
			allConnectIndices.push_back(connectIndices);
			allToTargGeos.push_back(toTargGeos);
			allFromTargGeos.push_back(fromTargGeos);
		}

		//kick this correspondence back out of this match so that it does not influence the next one
		pTempMatch->removeLastMatch();
	}//for (i...

	//add new correspondences to original match
	for (i = 0; i < corrsNew.size(); i++)
	{
		match->setMatchWithGeoData(corrsNewIndices[i], corrsNew[i].second, allConnectIndices[i], allToTargGeos[i], allFromTargGeos[i]);
	}

	AR4IM_STOP_PROFILE(m_profLocalFit)

	/////////////////////////////////
	//end-of-growing-step management:
	/////////////////////////////////

	AR4IM_START_PROFILE(m_profEndStep)

	//find any additional geodesic connections (between matches) that were not computed during local bundle above
	match->checkForNewGeodesicConnections();


	//remove outliers from among the newly added points (based on metric distortion)
	match->finalizeMatchCohort();

	AR4IM_STOP_PROFILE(m_profEndStep)

	delete pTempMatch;

#if AR4IM_INNER_DEBUG_RENDER
	debugRenderer->endRenderJob();
#endif //AR4IM_INNER_DEBUG_RENDER

	if (m_bDebugOutput)
		debugOutput << "\trun " << m_run << " - time: " << convertTimeToString(timer.getDeltaValue()) << "\n";
	m_run++;

	AR4IM_STOP_PROFILE(m_profStep)
}



