#ifndef AR4IncrementalMatcherH
#define AR4IncrementalMatcherH

#include "threeParameterMatchingGMOD14.h"

#include "AR4DiscreteManifold.h"
#include "AR4ManifoldTraits.h"
#include "AR4SmoothManifold.h"
#include "AR4PartialMatch.h"



namespace X4 {


template <class ManifoldTraits>
class AR4IncrementalMatcher {
 public:
   typedef typename ManifoldTraits::FloatType  FloatType;
   typedef typename ManifoldTraits::VectorType VectorType;
   typedef typename ManifoldTraits AR4DiscreteManifoldTraits;
   typedef typename ManifoldTraits::TangentSpace TangentSpace;
   typedef typename ManifoldTraits::BoundingBoxType BoundingBoxType;

 protected:
   AR4DiscreteManifold<ManifoldTraits> *source;
   AR4SmoothManifold<ManifoldTraits>   *target;

   bool									m_bDebugOutput;
 
 public:
   AR4IncrementalMatcher(AR4DiscreteManifold<ManifoldTraits> *source, AR4SmoothManifold<ManifoldTraits> *target);

   virtual void improveMatch(AR4PartialMatch<ManifoldTraits> *match) = 0;

   void setDebugOutput(bool boutput)	{ m_bDebugOutput = boutput; }
   bool isDebugOutput()					{ return m_bDebugOutput; }
};


template <class ManifoldTraits>
class AR4GrowingMatcher : public AR4IncrementalMatcher<ManifoldTraits> {

public:

	enum EGrowingMode
	{
		eGMRigidProjection = 0,
		eGMGeodesicTriangulation,
		eGMParallelTransport,
		eNumGrowingModes
	};

protected:
	mpcard									m_nPointsToAdd;
	FloatType								m_geodSigma;
	mpcard									m_nPointsLastGlobalBundle;
	mpcard									m_run;
	FloatType								m_projScaleSize;
	EGrowingMode							m_eGrowingMode;

	vector<card32>							m_matchedNbrCount;

	AR4Profile								m_profAddPoints;
	AR4Profile								m_profAddPoints_relax;
	AR4Profile								m_profLocalFit;
	AR4Profile								m_profEndStep;
	AR4Profile								m_profStep;

 public:
   AR4GrowingMatcher(AR4DiscreteManifold<ManifoldTraits> *source, AR4SmoothManifold<ManifoldTraits> *target);

   virtual void improveMatch(AR4PartialMatch<ManifoldTraits> *match) = 0;

   void setProjScaleSize(FloatType scale)					{ m_projScaleSize = scale; }
   FloatType getProjScaleSize()								{ return m_projScaleSize; }

   void setGrowingMode(EGrowingMode eMode)					{ m_eGrowingMode = eMode; }
   EGrowingMode getGrowingMode()							{ return m_eGrowingMode; }

   virtual void findNewMatchCandidates(AR4PartialMatch<ManifoldTraits> *match, vector<mpcard>& candidateIndices);

   void reset()
   {
	   m_matchedNbrCount.resize(source->getNumDiscrPoints(), 0);
   }

   void reportProfiling();
};

class AR4GrowingMatcher3f : public AR4GrowingMatcher<ManifoldTraits3f>
{
public:
	AR4GrowingMatcher3f(AR4DiscreteManifold<ManifoldTraits3f> *source, AR4SmoothManifold<ManifoldTraits3f> *target): AR4GrowingMatcher<ManifoldTraits3f>(source, target)
	{
	}

	virtual void improveMatch(AR4PartialMatch<ManifoldTraits3f> *match);
};

template<class ManifoldTraits>
class AR4MatchInterpolator : public AR4GrowingMatcher<ManifoldTraits>
{
protected:

public:

	AR4MatchInterpolator(AR4DiscreteManifold<ManifoldTraits> *source, AR4SmoothManifold<ManifoldTraits> *target);

	void findNewMatchCandidates(AR4PartialMatch<ManifoldTraits> *match, vector<mpcard>& candidateIndices);
};

class AR4MatchInterpolator3f : public AR4MatchInterpolator<ManifoldTraits3f>
{
public:
	AR4MatchInterpolator3f(AR4DiscreteManifold<ManifoldTraits3f> *source, AR4SmoothManifold<ManifoldTraits3f> *target): AR4MatchInterpolator<ManifoldTraits3f>(source, target)
	{
	}

	virtual void improveMatch(AR4PartialMatch<ManifoldTraits3f> *match);
};

//bundle-adjustment / GMDS
template <class ManifoldTraits>
class AR4BundleAdjustment : public AR4IncrementalMatcher<ManifoldTraits> {

protected:
	mpcard									m_nIterations;
	FloatType								m_geodSigma;
	FloatType								m_errorStopThresh;
	FloatType								m_errorChangeStopThresh;
	FloatType								m_stepScale;

	FloatType								m_lastStartError;
	FloatType								m_lastEndError;

	FloatType evalError(AR4PartialMatch<ManifoldTraits> *match, const vector<pair<Vector3f, Vector3f>>& corrs, const vector<bool>& fixed);
	FloatType evalError(AR4PartialMatch<ManifoldTraits> *match, const vector<vector<float>>& sourceDists, const vector<vector<mpcard>>& targetGeodesics, const vector<bool>& fixed);

 public:
	AR4BundleAdjustment(AR4DiscreteManifold<ManifoldTraits> *source, AR4SmoothManifold<ManifoldTraits> *target);

	void setGeodesicSigma(FloatType sigma)	{ m_geodSigma = sigma; }
	FloatType getGeodesicSigma()			{ return m_geodSigma; }

	void setStepScale(FloatType stepScale)	{ m_stepScale = stepScale; }
	FloatType getStepScale()				{ return m_stepScale; }

	FloatType getLastStartError()			{ return m_lastStartError; }
	FloatType getLastEndError()				{ return m_lastEndError; }

	virtual void improveMatch(AR4PartialMatch<ManifoldTraits> *match) = 0;

	FloatType computeGeodesicError(const std::vector< std::pair<VectorType, VectorType> >& corrs, atrec::ApproximatedGeodesicsGM14* pSourceGeod, atrec::ApproximatedGeodesicsGM14* pTargetGeod);
};

template <class ManifoldTraits>
class AR4DenseBundle : public AR4BundleAdjustment<ManifoldTraits>
{
protected:

public:

	AR4DenseBundle(AR4DiscreteManifold<ManifoldTraits> *source, AR4SmoothManifold<ManifoldTraits> *target): AR4BundleAdjustment<ManifoldTraits>(source, target)
	{
	}

	virtual void improveMatch(AR4PartialMatch<ManifoldTraits> *match) = 0;
};

template <class ManifoldTraits>
class AR4SparseBundle : public AR4BundleAdjustment<ManifoldTraits>
{
protected:

	mpcard									m_nSolverIters;
	FloatType								m_residualThreshold;
	mpcard									m_nMatchesLastRun;

public:

	AR4SparseBundle(AR4DiscreteManifold<ManifoldTraits> *source, AR4SmoothManifold<ManifoldTraits> *target): AR4BundleAdjustment<ManifoldTraits>(source, target)
	{
		m_nSolverIters = 10000;
		m_residualThreshold = (FloatType)1e-7;
		m_nMatchesLastRun = 0;
	}

	void setSolverIters(mpcard nIters)		{ m_nSolverIters = nIters; }
	mpcard getSolverIters() const			{ return m_nSolverIters; }

	void setResidualThreshold(float thresh)	{ m_residualThreshold = thresh; }
	FloatType getResidualThreshold() const	{ return m_residualThreshold; }

	mpcard getNumMatchesLastRun() const		{ return m_nMatchesLastRun; }
	void resetNumMatchesLastRun()			{ m_nMatchesLastRun = 0; }

	virtual void improveMatch(AR4PartialMatch<ManifoldTraits> *match) = 0;
};

template <class ManifoldTraits>
class AR4CoarseToFineSparseBundle : public AR4SparseBundle<ManifoldTraits>
{
protected:

	virtual void optimizeAtLevel(mpcard level, AR4PartialMatch<ManifoldTraits> *match) = 0;

public:

	AR4CoarseToFineSparseBundle(AR4DiscreteManifold<ManifoldTraits> *source, AR4SmoothManifold<ManifoldTraits> *target): AR4SparseBundle<ManifoldTraits>(source, target)
	{
	}
};


///////////////////////////////////////////////////////////////////////////////
//concrete type-specific subclasses
//3-dimensional with 32-bit floats
///////////////////////////////////////////////////////////////////////////////

class AR4DenseBundle3f : public AR4DenseBundle<ManifoldTraits3f>
{
public:

	AR4DenseBundle3f(AR4DiscreteManifold<ManifoldTraits3f> *source, AR4SmoothManifold<ManifoldTraits3f> *target): AR4DenseBundle<ManifoldTraits3f>(source, target)
	{
	}

	void improveMatch(AR4PartialMatch<ManifoldTraits3f> *match);
};

class AR4SparseBundle3f : public AR4SparseBundle<ManifoldTraits3f>
{
public:

	AR4SparseBundle3f(AR4DiscreteManifold<ManifoldTraits3f> *source, AR4SmoothManifold<ManifoldTraits3f> *target): AR4SparseBundle<ManifoldTraits3f>(source, target)
	{
	}

	void improveMatch(AR4PartialMatch<ManifoldTraits3f> *match);
};

class AR4CoarseToFineSparseBundle3f : public AR4CoarseToFineSparseBundle<ManifoldTraits3f>
{
protected:

	void optimizeAtLevel(mpcard level, AR4PartialMatch<ManifoldTraits3f> *match);

public:

	AR4CoarseToFineSparseBundle3f(AR4DiscreteManifold<ManifoldTraits3f> *source, AR4SmoothManifold<ManifoldTraits3f> *target): AR4CoarseToFineSparseBundle<ManifoldTraits3f>(source, target)
	{
	}

	void improveMatch(AR4PartialMatch<ManifoldTraits3f> *match);
};


}

#endif
