#ifndef AR4IncrementalMatcherInlineH
#define AR4IncrementalMatcherInlineH

#include "threeParameterMatchingGMOD14.h"

#include "AR4IncrementalMatcher.h"



namespace X4 {


template <class ManifoldTraits>
AR4IncrementalMatcher<ManifoldTraits>::AR4IncrementalMatcher( AR4DiscreteManifold<ManifoldTraits> *source, AR4SmoothManifold<ManifoldTraits> *target )
{
   this->source = source;
   this->target = target;
   m_bDebugOutput = true;
}


template <class ManifoldTraits>
AR4BundleAdjustment<ManifoldTraits>::AR4BundleAdjustment(AR4DiscreteManifold<ManifoldTraits> *source, AR4SmoothManifold<ManifoldTraits> *target): AR4IncrementalMatcher<ManifoldTraits>(source, target)
{
	m_nIterations = 100;
	m_geodSigma = (FloatType)0.1;
	m_errorStopThresh = (FloatType)1e-4;
	m_errorChangeStopThresh = (FloatType)1e-7;
	m_stepScale = (FloatType)1.0;
	m_lastStartError = (FloatType)0;
	m_lastEndError = (FloatType)0;
}

template <class ManifoldTraits>
typename AR4IncrementalMatcher<ManifoldTraits>::FloatType AR4BundleAdjustment<ManifoldTraits>::evalError(AR4PartialMatch<ManifoldTraits> *match, const std::vector<pair<Vector3f,Vector3f> > &corrs, const std::vector<bool> &fixed)
{
	mpcard i, j, numCorr = corrs.size();
	FloatType dtarget, dsource, diff;
	FloatType error = 0;

	for (i = 0; i < numCorr; i++)
	{
		if (!fixed[i])
		{
			for (j = 0; j < numCorr; j++)
			{
				if (j == i)
					continue;

				dtarget = match->targetGeodesic(corrs[i].second, corrs[j].second);
				dsource = match->sourceGeodesic(corrs[i].first, corrs[j].first);

				diff = dtarget - dsource;
				error += diff*diff;
			}
		}
	}

	return error;
}

template <class ManifoldTraits>
typename AR4IncrementalMatcher<ManifoldTraits>::FloatType AR4BundleAdjustment<ManifoldTraits>::evalError(AR4PartialMatch<ManifoldTraits> *match, const vector<vector<float>>& sourceDists, const vector<vector<mpcard>>& targetGeodesics, const vector<bool>& fixed)
{
	mpcard i, j, numCorr = sourceDists.size();
	FloatType dtarget, dsource, diff;
	VectorType grad;
	FloatType error = 0;

	for (i = 0; i < numCorr; i++)
	{
		if (fixed[i])
			continue;
		for (j = 0; j < sourceDists[i].size(); j++)
		{
			dsource = sourceDists[i][j];
			dtarget = match->targetGeodesic(targetGeodesics[i][j], grad);//gradient not used
			diff = dtarget - dsource;
			error += diff*diff;
		}
	}

	return error;
}

template<class ManifoldTraits>
AR4GrowingMatcher<ManifoldTraits>::AR4GrowingMatcher(AR4DiscreteManifold<ManifoldTraits> *source, AR4SmoothManifold<ManifoldTraits> *target): AR4IncrementalMatcher<ManifoldTraits>(source, target)
{
	m_nPointsToAdd = 10;
	m_geodSigma = (FloatType)0.1;
	m_nPointsLastGlobalBundle = 0;
	m_run = 0;
	m_projScaleSize = (FloatType)(-1);
	m_eGrowingMode = eGMRigidProjection;
	m_matchedNbrCount.resize(source->getNumDiscrPoints(), 0);
	m_profAddPoints.setName("AR4GrowingMatcher - add points");
	m_profAddPoints_relax.setName("AR4GrowingMatcher - add points - path relaxation");
	m_profLocalFit.setName("AR4GrowingMatcher - local fit");
	m_profEndStep.setName("AR4GrowingMatcher - end of step");
	m_profStep.setName("AR4GrowingMatcher::improveMatch");
}

template<class ManifoldTraits>
void AR4GrowingMatcher<ManifoldTraits>::reportProfiling()
{
	m_profAddPoints.reportStats();
	m_profAddPoints_relax.reportStats();
	m_profLocalFit.reportStats();
	m_profEndStep.reportStats();
	m_profStep.reportStats();
}

template<class ManifoldTraits>
void AR4GrowingMatcher<ManifoldTraits>::findNewMatchCandidates(AR4PartialMatch<ManifoldTraits> *match, vector<mpcard>& candidateIndices)
{
	mpcard i, j, k, l;

	const mpcard numCorr = match->getNumMatchedIndices();
	const mpcard numSourcePoints = match->getSourceManifold()->getNumDiscrPoints();

	vector<bool> isCandidate(numSourcePoints, false);

	candidateIndices.clear();

	for (i = 0; i < numCorr; i++)
	{
		k = match->getMatchedSourceIndex(i);
		const mpcard numNbrs = match->getNumSourceNeighborsMatchingLevel(k);
		for (j = 0; j < numNbrs; j++)
		{
			l = match->getSourceNeighborMatchingLevel(k, j);
			if (!match->isMatched(l))
			{
				//check if it is already assigned for mapping
				if (isCandidate[l])		continue;

				//if not, add it to the list of potential new correspondences
				candidateIndices.push_back(l);
				isCandidate[l] = true;
			}
		}
	}
}


template<class ManifoldTraits>
AR4MatchInterpolator<ManifoldTraits>::AR4MatchInterpolator(AR4DiscreteManifold<ManifoldTraits> *source, AR4SmoothManifold<ManifoldTraits> *target): AR4GrowingMatcher<ManifoldTraits>(source, target)
{
}

template<class ManifoldTraits>
void AR4MatchInterpolator<ManifoldTraits>::findNewMatchCandidates(AR4PartialMatch<ManifoldTraits> *match, vector<mpcard>& candidateIndices)
{
	mpcard i, j, k, l;

	const mpcard numCorr = match->getNumMatchedIndices();
	const mpcard numSourcePoints = match->getSourceManifold()->getNumDiscrPoints();

	vector<bool> isCandidate(numSourcePoints, false);

	candidateIndices.clear();

	for (i = 0; i < numCorr; i++)
	{
		k = match->getMatchedSourceIndex(i);

		vector<mpcard> neighbors;
		match->getSourceNeighborhood(k, 1, neighbors);
		const mpcard numNbrs = neighbors.size();

		for (j = 0; j < numNbrs; j++)
		{
			l = neighbors[j];
			if (!match->isMatched(l))
			{
				//check if it is already assigned for mapping
				if (isCandidate[l])		continue;

				//if not, add it to the list of potential new correspondences
				candidateIndices.push_back(l);
				isCandidate[l] = true;
			}
		}
	}
}


}

#endif
