//---------------------------------------------------------------------------
#include "stdafx.h"
//---------------------------------------------------------------------------
#include "AR4MLSManifold.h"
#include "AR4MLSManifold.inline.h"
//---------------------------------------------------------------------------
#include "InCorePCTopologyGraph.h"
#include "PointSet.inline.h"
#include "PointSetKNNQuery.h"
#include "PointSetSphericalQuery.h"
#include "Monomials2ndOrder.h"
#include "PCA.h"
#include "StringHelper.h"
#include "PCCComputeTopology.h"
#include "PCCEstimateSampleSpacing.h"
#include "userInterface/PCCBoundaryDetector.h"
#include "basics/DeformationTools.h"
#include "TopologyRangeSearch.h"
#include "HalfSpace.inline.h"
#include "SparseGrid.inline.h"
#include <Timer.h>
#include "TopologyAttachment.h"
//---------------------------------------------------------------------------
#include "DebugRenderer.h"
//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------
#define AR4MLS_PROJECT_RADIUS_FACTOR		4.5
#define AR4MLS_PROJECT_LOOKUP				1
#define AR4MLS_PROJECT_LOOKUP_LRFACTOR		4
#define AR4MLS_HARMONIC_EPSILON_FACTOR		5.0
#define AR4_DEBUG_OUTPUT					0
//---------------------------------------------------------------------------

AR4MLSManifold3f::AR4MLSManifold3f(UnstructuredInCorePointCloud *pc) 
	:AR4MLSManifold< ManifoldTraits3f >(pc)
	,sQueryNG(NULL)
	,knnNG(NULL)
	,knnEucl(NULL)
	,octree(NULL)
{
	
	imlsGradFact = 0.25f;
	
	imlsConvThreshold = 1e-5f;
	imlsConvThresholdFact = 1e-5f;
	imlsMaxNumIterations = 100;

	m_sampleFactorHarmonics = AR4MLS_HARMONIC_EPSILON_FACTOR;

	AR4MLSManifold3f::resetup(pc);
}

void AR4MLSManifold3f::resetup(UnstructuredInCorePointCloud *pc)
{
   AR4MLSManifold<ManifoldTraits3f>::resetup(pc);
   if (ps) {
      octree = PointSetKNNQuery::createOctree(ps);
      AAT posAAT = ps->getAAT("position");
      knnNG = new TopNeighborhoodKNNSearch_Fast(ps, tg, 0, positionAAT);
      sQueryNG = new TopNeighborhoodSphereSearch();
      knnEucl = new PointSetKNNQuery(ps, octree, NULL_VECTOR3F, positionAAT);
      imlsConvThreshold = imlsConvThresholdFact*ps->getBoundingBox().getMaxSideLength();

	  initProjectionLookups();
	  preNormNormals();
   } else {
      throw ETypeMismatch("AR4MLSManifold3f::setup - NULL point set");
   }
}


AR4MLSManifold3f::~AR4MLSManifold3f() {
   delete knnNG;
   delete sQueryNG;
   delete octree;
}

void AR4MLSManifold3f::initProjectionLookups()
{
	const mpcard numPoints = pc->getNumPoints();
	
	mpcard i, j, k;

	BoundingBox3f bbox = getBoundingBox();
	const Vector3f origin = bbox.lowerCorner;

	m_avgSampleSpacing = 0.f;
	for (i = 0; i < numPoints; i++)
	{
		m_avgSampleSpacing += ps->get1f(i, sampleSpacingAAT);
	}
	m_avgSampleSpacing /= (float32)numPoints;

	m_projectRadius = m_avgSampleSpacing * AR4MLS_PROJECT_RADIUS_FACTOR;

	const float32 binWidth = m_avgSampleSpacing * AR4MLS_PROJECT_RADIUS_FACTOR;
	const float32 collectRadius = binWidth * 2.f;
	const float32 collectRadius2 = collectRadius*collectRadius;
	const mpcard gridResX = max((mpcard)1, (mpcard)ceil(bbox.getSideLength(0) / binWidth));
	const mpcard gridResY = max((mpcard)1, (mpcard)ceil(bbox.getSideLength(1) / binWidth));
	const mpcard gridResZ = max((mpcard)1, (mpcard)ceil(bbox.getSideLength(2) / binWidth));

	const float32 lrBinWidth = binWidth * AR4MLS_PROJECT_LOOKUP_LRFACTOR;
	const mpcard lrGridResX = 1 + (mpcard)ceil(bbox.getSideLength(0) / lrBinWidth);
	const mpcard lrGridResY = 1 + (mpcard)ceil(bbox.getSideLength(1) / lrBinWidth);
	const mpcard lrGridResZ = 1 + (mpcard)ceil(bbox.getSideLength(2) / lrBinWidth);
	const Vector3f lrOrigin = origin - makeVector3f(lrBinWidth * 0.5f, lrBinWidth * 0.5f, lrBinWidth * 0.5f);

	m_gridSize = binWidth;
	m_longRangeGridSize = lrBinWidth;
	m_longRangeOrigin = lrOrigin;

	vector<mpcard> pointBinCount(numPoints, 0);

	mpcard maxCellPoints = 0;

	Vector3i gridPos;
	gridPos[2] = 0;
	for (i = 0; i < gridResZ; i++)
	{
		gridPos[1] = 0;
		for (j = 0; j < gridResY; j++)
		{
			gridPos[0] = 0;
			for (k = 0; k < gridResX; k++)
			{
				const BoundingBox3f gridBox = m_sparseGrid.calcBoundingBox(gridPos, origin, binWidth);
				const Vector3f boxCenter = gridBox.getCenter();

				ProjCell cell;

				knnEucl->setTargetPoint(boxCenter);
				card32 nearest;
				bool bfound;
				knnEucl->getNextNearestPointIndex(nearest, bfound);
				if (bfound && normQuad(ps->get3f(nearest, positionAAT) - boxCenter) < collectRadius2)
				{
					while (bfound && normQuad(ps->get3f(nearest, positionAAT) - boxCenter) < collectRadius2)
					{
						cell.m_neighbors.push_back(nearest);
						if (gridBox.inBoundingBox(ps->get3f(nearest, positionAAT)))
							pointBinCount[nearest]++;
						if (gridBox.inBoundingBox(ps->get3f(nearest, positionAAT)))
							pointBinCount[nearest]++;
						cell.m_avgSampleSpacing += ps->get1f(nearest, sampleSpacingAAT);
						cell.m_centroid += ps->get3f(nearest, positionAAT);

						knnEucl->getNextNearestPointIndex(nearest, bfound);
					}

					cell.m_avgSampleSpacing /= (float)cell.m_neighbors.size();
					cell.m_centroid /= (float)cell.m_neighbors.size();
					if (cell.m_neighbors.size() > maxCellPoints)
						maxCellPoints = cell.m_neighbors.size();

					m_sparseGrid.setObject(gridPos, cell);
				}

				gridPos[0]++;
			}

			gridPos[1]++;
		}

		gridPos[2]++;
	}

	m_vertexCache.resize(maxCellPoints);
	m_normalCache.resize(maxCellPoints);


	for (i = 0; i < numPoints; i++)
	{
		if (pointBinCount[i] == 0)
		{
			debugOutput << "AR4MSLManifold3f::initProjectionLookups(): point " << i << " in no bin\n";
		}
	}

	gridPos[2] = 0;
	for (i = 0; i < lrGridResZ; i++)
	{
		gridPos[1] = 0;
		for (j = 0; j < lrGridResY; j++)
		{
			gridPos[0] = 0;
			for (k = 0; k < lrGridResX; k++)
			{
				const BoundingBox3f gridBox = m_sparseGrid.calcBoundingBox(gridPos, lrOrigin, lrBinWidth);
				Vector3f centerPos = NULL_VECTOR3F;
				mpcard ki, kj, kk;

				ProjCellLongRange lrCell;
				vector<Vector3i> residents;

				centerPos[2] = gridBox.lowerCorner[2] + binWidth * 0.5f;
				for (ki = 0; ki < AR4MLS_PROJECT_LOOKUP_LRFACTOR; ki++)
				{
					centerPos[1] = gridBox.lowerCorner[1] + binWidth * 0.5f;
					for (kj = 0; kj < AR4MLS_PROJECT_LOOKUP_LRFACTOR; kj++)
					{
						centerPos[0] = gridBox.lowerCorner[0] + binWidth * 0.5f;
						for (kk = 0; kk < AR4MLS_PROJECT_LOOKUP_LRFACTOR; kk++)
						{
							Vector3i gridIdx = m_sparseGrid.calcIndex(centerPos, origin, binWidth);
							ProjCell cell;

							if (m_sparseGrid.getObject(gridIdx, cell))
							{
								residents.push_back(gridIdx);
							}
							centerPos[0] += binWidth;
						}

						centerPos[1] += binWidth;
					}

					centerPos[2] += binWidth;
				}

				const mpcard numResidents = residents.size();
				if (numResidents > 0)
				{
					const Vector3i hugeGridIdx = makeVector3i(gridResX * gridResY * gridResZ * 1000, gridResX * gridResY * gridResZ * 1000, gridResX * gridResY * gridResZ * 1000);

					mpcard n, r;

					lrCell.m_fineGridSize = m_gridSize;
					lrCell.m_fineGridOrigin = origin;
					lrCell.m_gridSizeFactor = AR4MLS_PROJECT_LOOKUP_LRFACTOR;
					lrCell.m_bbox = gridBox;
					lrCell.m_nearest.resize(AR4MLS_PROJECT_LOOKUP_LRFACTOR*AR4MLS_PROJECT_LOOKUP_LRFACTOR*AR4MLS_PROJECT_LOOKUP_LRFACTOR * 6);

					n = 0;
					centerPos[2] = gridBox.lowerCorner[2] + m_gridSize * 0.5f;
					for (ki = 0; ki < AR4MLS_PROJECT_LOOKUP_LRFACTOR; ki++)
					{
						centerPos[1] = gridBox.lowerCorner[1] + m_gridSize * 0.5f;
						for (kj = 0; kj < AR4MLS_PROJECT_LOOKUP_LRFACTOR; kj++)
						{
							centerPos[0] = gridBox.lowerCorner[0] + m_gridSize * 0.5f;
							for (kk = 0; kk < AR4MLS_PROJECT_LOOKUP_LRFACTOR; kk++)
							{
								Vector3i gridIdx = m_sparseGrid.calcIndex(centerPos, origin, m_gridSize);
								int32 dsqr;
								vector<pair<int32, mpcard>> distsToResidents(numResidents);
								for (r = 0; r < numResidents; r++)
								{
									dsqr = (gridIdx - residents[r]).getSqrNorm();
									distsToResidents[r].first = dsqr;
									distsToResidents[r].second = r;
								}
								sort(distsToResidents.begin(), distsToResidents.end());

								for (r = 0; r < 6 && r < numResidents; r++)
								{
									lrCell.m_nearest[n * 6 + r] = residents[distsToResidents[r].second];
								}
								for (; r < 6; r++)
								{
									lrCell.m_nearest[n * 6 + r] = hugeGridIdx;
								}

								centerPos[0] += m_gridSize;
								n++;
							}
							centerPos[1] += m_gridSize;
						}
						centerPos[2] += m_gridSize;
					}
					m_longRangeGrid.setObject(gridPos, lrCell);
				}

				gridPos[0]++;
			}

			gridPos[1]++;
		}

		gridPos[2]++;
	}

	const float32 sigma2 = (float32)m_nFunctionBins / 9.f;
	m_weightFunc.setDim(m_nFunctionBins);
	m_weightGrad.setDim(m_nFunctionBins);
	for (i = 0; i < m_nFunctionBins; i++)
	{
		const float32 d2 =(float32)i / sigma2;
		const float32 d = sqrt(d2);
		m_weightFunc[i] = exp(-d2);
		m_weightGrad[i] = -2.f * d * m_weightFunc[i];
	}

	m_swfStepSize = 9.f / (float)m_nFunctionBins;
	m_swfStepSizeInv = 1.f / m_swfStepSize;
}

void AR4MLSManifold3f::preNormNormals()
{
	const mpcard n = ps->getNumEntries();
	mpcard i;

	for (i = 0; i < n; i++)
	{
		Vector3f normal = ps->get3f(i, normalAAT);
		if (normal.getSqrNorm() < 1e-5f)
			normal *= 1e5f;
		normal.normalize();
		ps->set3f(i, normalAAT, normal);
	}
}



//---------------------------------------------------------------------------



bool AR4MLSManifold3f::project(
   const Vector3f &startPos, 
   Vector3f &surfacePoint, 
   Vector3f &tangentU, 
   Vector3f &tangentV, 
   Vector3f &normal, 
   Vector3f &curvTensor_u_uv_v)
{
	bool trusted = true;
   if (!ps) throw EInvalidState("AR4MLSManifold::project - point set missing.");
   if (!tg) throw EInvalidState("AR4MLSManifold::project - topology missing.");
   if (!knnNG || !octree) throw EInvalidState("AR4MLSManifold::project - where is the knn?");
   if (sampleSpacingAAT == NULL_AAT) throw EInvalidState("AR4MLSManifold::project - no sample spacing channel found.");
   normal = NULL_VECTOR3F;
   surfacePoint = startPos;
   
	
	const BoundingBox3f bbox = ps->getBoundingBox();
	const Vector3f origin = bbox.lowerCorner;
	Vector3i gridIdx = m_sparseGrid.calcIndex(startPos, origin, m_gridSize);
	ProjCell cell;

	Vector3f proxyPos = startPos;

	if (!m_sparseGrid.getObject(gridIdx, cell))
	{
		trusted = false;

		Vector3i lrIdx = m_longRangeGrid.calcIndex(startPos, m_longRangeOrigin, m_longRangeGridSize);
		ProjCellLongRange lrCell;
		if (!m_longRangeGrid.getObject(lrIdx, lrCell))
		{
			return false;
		}
		gridIdx = lrCell.computeNearest(startPos);
		if (!m_sparseGrid.getObject(gridIdx, cell))
		{
			debugOutput << "why the fuck should this happen?\n";
			return false;
		}
	}

	const mpcard numNbrs = cell.m_neighbors.size();

	for (mpcard nbr = 0; nbr < numNbrs; nbr++)
	{
		const mpcard index = cell.m_neighbors[nbr];
		m_vertexCache[nbr] = ps->get3f(index, positionAAT);
		m_normalCache[nbr] = ps->get3f(index, normalAAT);
	}

	Vector3f pos = startPos;

	
	const float32 sigmaN = 0.5f;
	const float32 sigmaN2 = sigmaN*sigmaN;
	const float32 rcpSigmaN2 = 1.f / sigmaN2;

	Vector3f grad = NULL_VECTOR3F, weightGSum;
	float32 scalarValue = 0, weightSum = 0;

	for (mpcard i = 0; i < imlsMaxNumIterations; i++)
	{
		const float32 localSpacing = AR4MLS_PROJECT_RADIUS_FACTOR * cell.m_avgSampleSpacing;
		const float32 localSpacing2 = localSpacing*localSpacing;
		const float32 rcpLocalSpacing2 = 1.f / localSpacing2;

		const float32 radius = max(localSpacing, (pos - cell.m_centroid).getNorm());
		const float32 radius2 = radius*radius;
		const float32 rcpRadius2 = 1.f / radius2;

		const float32 sigmaR = localSpacing * 0.5f;
		const float32 rcpSigmaR2 = 4.f * rcpLocalSpacing2;

		const Vector3f prevGrad = grad;
		const float32 prevScalar = scalarValue;

		normal = NULL_VECTOR3F;
		grad = NULL_VECTOR3F;
		weightGSum = NULL_VECTOR3F;

		scalarValue = 0;
		weightSum = 0;

		if (i == 0)
		{
			for (mpcard nbr = 0; nbr < numNbrs; nbr++)
			{
				const Vector3f& posL = m_vertexCache[nbr];
				const Vector3f delta = pos - posL;
				const float32 distSqr = delta.getSqrNorm();
				if (distSqr > radius2)	continue;

				const Vector3f& normL = m_normalCache[nbr];
				const float32 localScalar = delta * normL;

				const float32 spWeight = weightFunc(distSqr * rcpLocalSpacing2);
				const float32 nmWeight = 1.f;

				const float32 weight = spWeight * nmWeight;
				const Vector3f weightG = delta * nmWeight * weightGrad(distSqr * rcpLocalSpacing2) * rcpLocalSpacing2;

				weightSum += weight;
				weightGSum += weightG;

				scalarValue += localScalar * weight;
				grad += weightG * localScalar;

				normal += weight * normL;
			}
		}
		else
		{
			for (mpcard nbr = 0; nbr < numNbrs; nbr++)
			{
				const Vector3f& posL = m_vertexCache[nbr];
				const Vector3f delta = pos - posL;
				const float32 distSqr = delta.getSqrNorm();
				if (distSqr > radius2)	continue;

				const Vector3f& normL = m_normalCache[nbr];
				const float32 localScalar = delta * normL;

				const float32 spWeight =	weightFunc(distSqr * rcpLocalSpacing2);
				const float32 nmWeight =	weightFunc(sqr(localScalar - prevScalar) * rcpSigmaR2) * 
											weightFunc((normL - prevGrad).getSqrNorm() * rcpSigmaN2);

				const float32 weight = spWeight * nmWeight;
				const Vector3f weightG = delta * nmWeight * weightGrad(distSqr * rcpLocalSpacing2) * rcpLocalSpacing2;

				weightSum += weight;
				weightGSum += weightG;

				scalarValue += localScalar * weight;
				grad += weightG * localScalar;

				normal += weight * normL;
			}
		}

		if (weightSum <= 1e-5) {return false;}

		const float32 rcpWeightSum = 1.f / weightSum;

		scalarValue *= rcpWeightSum;

		if (fabs(scalarValue) <= imlsConvThreshold || i== imlsMaxNumIterations)
		{
			normal *= rcpWeightSum;
			normal = normalize(normal);
			surfacePoint = pos;
			tangentU = makeOrthonormalVector(normal);
			tangentV = normal.crossProduct(tangentU);
			curvTensor_u_uv_v = NULL_VECTOR3F;
			return trusted;
		}

		grad -= weightGSum * scalarValue;
		grad += normal;
		grad *= rcpWeightSum;

		pos -= grad * scalarValue * imlsGradFact;
	}
	

	return trusted;
}

bool AR4MLSManifold3f::project(const vector<float32>& funcOnSurface, const Vector3f &startPos, Vector3f &surfacePoint, Vector3f &normal, float32& funcValue)
{
	bool trusted = true;
   if (!ps) throw EInvalidState("AR4MLSManifold::project - point set missing.");
   if (!tg) throw EInvalidState("AR4MLSManifold::project - topology missing.");
   if (!knnNG || !octree) throw EInvalidState("AR4MLSManifold::project - where is the knn?");
   if (sampleSpacingAAT == NULL_AAT) throw EInvalidState("AR4MLSManifold::project - no sample spacing channel found.");
   normal = NULL_VECTOR3F;
   surfacePoint = startPos;
   
	
	const BoundingBox3f bbox = ps->getBoundingBox();
	const Vector3f origin = bbox.lowerCorner;
	Vector3i gridIdx = m_sparseGrid.calcIndex(startPos, origin, m_gridSize);
	ProjCell cell;

	Vector3f proxyPos = startPos;

	if (!m_sparseGrid.getObject(gridIdx, cell))
	{
		trusted = false;

		Vector3i lrIdx = m_longRangeGrid.calcIndex(startPos, m_longRangeOrigin, m_longRangeGridSize);
		ProjCellLongRange lrCell;
		if (!m_longRangeGrid.getObject(lrIdx, lrCell))
		{
			return false;
		}
		gridIdx = lrCell.computeNearest(startPos);
		if (!m_sparseGrid.getObject(gridIdx, cell))
		{
			debugOutput << "why the fuck should this happen?\n";
			return false;
		}
	}

	const mpcard numNbrs = cell.m_neighbors.size();

	vector<float32> funcCache(numNbrs);

	for (mpcard nbr = 0; nbr < numNbrs; nbr++)
	{
		const mpcard index = cell.m_neighbors[nbr];
		m_vertexCache[nbr] = ps->get3f(index, positionAAT);
		m_normalCache[nbr] = ps->get3f(index, normalAAT);
		funcCache[nbr] = funcOnSurface[index];
	}

	Vector3f pos = startPos;

	const float32 sigmaN = 0.5f;
	const float32 sigmaN2 = sigmaN*sigmaN;
	const float32 rcpSigmaN2 = 1.f / sigmaN2;

	Vector3f grad = NULL_VECTOR3F, weightGSum;
	float32 scalarValue = 0, weightSum = 0;

	for (mpcard i = 0; i < imlsMaxNumIterations; i++)
	{
		const float32 localSpacing = AR4MLS_PROJECT_RADIUS_FACTOR * cell.m_avgSampleSpacing;
		const float32 localSpacing2 = localSpacing*localSpacing;
		const float32 rcpLocalSpacing2 = 1.f / localSpacing2;

		const float32 radius = max(localSpacing, (pos - cell.m_centroid).getNorm());
		const float32 radius2 = radius*radius;
		const float32 rcpRadius2 = 1.f / radius2;

		const float32 sigmaR = localSpacing * 0.5f;
		const float32 rcpSigmaR2 = 4.f * rcpLocalSpacing2;

		const Vector3f prevGrad = grad;
		const float32 prevScalar = scalarValue;

		normal = NULL_VECTOR3F;
		grad = NULL_VECTOR3F;
		weightGSum = NULL_VECTOR3F;

		funcValue = 0.f;

		scalarValue = 0;
		weightSum = 0;

		if (i == 0)
		{
			for (mpcard nbr = 0; nbr < numNbrs; nbr++)
			{
				const Vector3f& posL = m_vertexCache[nbr];
				const Vector3f delta = pos - posL;
				const float32 distSqr = delta.getSqrNorm();
				if (distSqr > radius2)	continue;

				const Vector3f& normL = m_normalCache[nbr];
				const float32 localScalar = delta * normL;

				const float32 spWeight = weightFunc(distSqr * rcpLocalSpacing2);
				const float32 nmWeight = 1.f;

				const float32 weight = spWeight * nmWeight;
				const Vector3f weightG = delta * nmWeight * weightGrad(distSqr * rcpLocalSpacing2) * rcpLocalSpacing2;

				weightSum += weight;
				weightGSum += weightG;

				scalarValue += localScalar * weight;
				grad += weightG * localScalar;

				normal += weight * normL;

				funcValue += weight * funcCache[nbr];
			}
		}
		else
		{
			for (mpcard nbr = 0; nbr < numNbrs; nbr++)
			{
				const Vector3f& posL = m_vertexCache[nbr];
				const Vector3f delta = pos - posL;
				const float32 distSqr = delta.getSqrNorm();
				if (distSqr > radius2)	continue;

				const Vector3f& normL = m_normalCache[nbr];
				const float32 localScalar = delta * normL;

				const float32 spWeight =	weightFunc(distSqr * rcpLocalSpacing2);
				const float32 nmWeight =	weightFunc(sqr(localScalar - prevScalar) * rcpSigmaR2) * 
											weightFunc((normL - prevGrad).getSqrNorm() * rcpSigmaN2);

				const float32 weight = spWeight * nmWeight;
				const Vector3f weightG = delta * nmWeight * weightGrad(distSqr * rcpLocalSpacing2) * rcpLocalSpacing2;

				weightSum += weight;
				weightGSum += weightG;

				scalarValue += localScalar * weight;
				grad += weightG * localScalar;

				normal += weight * normL;

				funcValue += weight * funcCache[nbr];
			}
		}

		if (weightSum <= 1e-5) {return false;}

		const float32 rcpWeightSum = 1.f / weightSum;

		scalarValue *= rcpWeightSum;

		if (fabs(scalarValue) <= imlsConvThreshold || i== imlsMaxNumIterations)
		{
			normal *= rcpWeightSum;
			normal = normalize(normal);
			surfacePoint = pos;
			funcValue *= rcpWeightSum;
			return trusted;
		}

		grad -= weightGSum * scalarValue;
		grad += normal;
		grad *= rcpWeightSum;

		pos -= grad * scalarValue * imlsGradFact;
	}

	
	return trusted;
}

bool AR4MLSManifold3f::project(const vector<Vector3f>& funcOnSurface, const Vector3f &startPos, Vector3f &surfacePoint, Vector3f &normal, Vector3f& funcValue)
{
	bool trusted = true;
   if (!ps) throw EInvalidState("AR4MLSManifold::project - point set missing.");
   if (!tg) throw EInvalidState("AR4MLSManifold::project - topology missing.");
   if (!knnNG || !octree) throw EInvalidState("AR4MLSManifold::project - where is the knn?");
   if (sampleSpacingAAT == NULL_AAT) throw EInvalidState("AR4MLSManifold::project - no sample spacing channel found.");
   normal = NULL_VECTOR3F;
   surfacePoint = startPos;
   
	const BoundingBox3f bbox = ps->getBoundingBox();
	const Vector3f origin = bbox.lowerCorner;
	Vector3i gridIdx = m_sparseGrid.calcIndex(startPos, origin, m_gridSize);
	ProjCell cell;

	Vector3f proxyPos = startPos;

	if (!m_sparseGrid.getObject(gridIdx, cell))
	{
		trusted = false;

		Vector3i lrIdx = m_longRangeGrid.calcIndex(startPos, m_longRangeOrigin, m_longRangeGridSize);
		ProjCellLongRange lrCell;
		if (!m_longRangeGrid.getObject(lrIdx, lrCell))
		{
			return false;
		}
		gridIdx = lrCell.computeNearest(startPos);
		if (!m_sparseGrid.getObject(gridIdx, cell))
		{
			debugOutput << "why the fuck should this happen?\n";
			return false;
		}
	}

	const mpcard numNbrs = cell.m_neighbors.size();

	vector<Vector3f> funcCache(numNbrs);

	for (mpcard nbr = 0; nbr < numNbrs; nbr++)
	{
		const mpcard index = cell.m_neighbors[nbr];
		m_vertexCache[nbr] = ps->get3f(index, positionAAT);
		m_normalCache[nbr] = ps->get3f(index, normalAAT);
		funcCache[nbr] = funcOnSurface[index];
	}

	Vector3f pos = startPos;

	const float32 sigmaN = 0.5f;
	const float32 sigmaN2 = sigmaN*sigmaN;
	const float32 rcpSigmaN2 = 1.f / sigmaN2;

	Vector3f grad = NULL_VECTOR3F, weightGSum;
	float32 scalarValue = 0, weightSum = 0;

	for (mpcard i = 0; i < imlsMaxNumIterations; i++)
	{
		const float32 localSpacing = AR4MLS_PROJECT_RADIUS_FACTOR * cell.m_avgSampleSpacing;
		const float32 localSpacing2 = localSpacing*localSpacing;
		const float32 rcpLocalSpacing2 = 1.f / localSpacing2;

		const float32 radius = max(localSpacing, (pos - cell.m_centroid).getNorm());
		const float32 radius2 = radius*radius;
		const float32 rcpRadius2 = 1.f / radius2;

		const float32 sigmaR = localSpacing * 0.5f;
		const float32 rcpSigmaR2 = 4.f * rcpLocalSpacing2;

		const Vector3f prevGrad = grad;
		const float32 prevScalar = scalarValue;

		normal = NULL_VECTOR3F;
		grad = NULL_VECTOR3F;
		weightGSum = NULL_VECTOR3F;

		funcValue = NULL_VECTOR3F;

		scalarValue = 0;
		weightSum = 0;

		if (i == 0)
		{
			for (mpcard nbr = 0; nbr < numNbrs; nbr++)
			{
				const Vector3f& posL = m_vertexCache[nbr];
				const Vector3f delta = pos - posL;
				const float32 distSqr = delta.getSqrNorm();
				if (distSqr > radius2)	continue;

				const Vector3f& normL = m_normalCache[nbr];
				const float32 localScalar = delta * normL;

				const float32 spWeight = weightFunc(distSqr * rcpLocalSpacing2);
				const float32 nmWeight = 1.f;

				const float32 weight = spWeight * nmWeight;
				const Vector3f weightG = delta * nmWeight * weightGrad(distSqr * rcpLocalSpacing2) * rcpLocalSpacing2;

				weightSum += weight;
				weightGSum += weightG;

				scalarValue += localScalar * weight;
				grad += weightG * localScalar;

				normal += weight * normL;

				funcValue += weight * funcCache[nbr];
			}
		}
		else
		{
			for (mpcard nbr = 0; nbr < numNbrs; nbr++)
			{
				const Vector3f& posL = m_vertexCache[nbr];
				const Vector3f delta = pos - posL;
				const float32 distSqr = delta.getSqrNorm();
				if (distSqr > radius2)	continue;

				const Vector3f& normL = m_normalCache[nbr];
				const float32 localScalar = delta * normL;

				const float32 spWeight =	weightFunc(distSqr * rcpLocalSpacing2);
				const float32 nmWeight =	weightFunc(sqr(localScalar - prevScalar) * rcpSigmaR2) * 
											weightFunc((normL - prevGrad).getSqrNorm() * rcpSigmaN2);

				const float32 weight = spWeight * nmWeight;
				const Vector3f weightG = delta * nmWeight * weightGrad(distSqr * rcpLocalSpacing2) * rcpLocalSpacing2;

				weightSum += weight;
				weightGSum += weightG;

				scalarValue += localScalar * weight;
				grad += weightG * localScalar;

				normal += weight * normL;

				funcValue += weight * funcCache[nbr];
			}
		}

		if (weightSum <= 1e-5) {return false;}

		const float32 rcpWeightSum = 1.f / weightSum;

		scalarValue *= rcpWeightSum;

		if (fabs(scalarValue) <= imlsConvThreshold || i== imlsMaxNumIterations)
		{
			normal *= rcpWeightSum;
			normal = normalize(normal);
			surfacePoint = pos;
			funcValue *= rcpWeightSum;
			return trusted;
		}

		grad -= weightGSum * scalarValue;
		grad += normal;
		grad *= rcpWeightSum;

		pos -= grad * scalarValue * imlsGradFact;
	}

	
	return trusted;
}

void AR4MLSManifold3f::estimateSampleSpacing3f(card32 numNN)
{
   PCCEstimateSampleSpacing::estimateSampleSpacing(pc, numNN);
   ps = pc->getPointSet();
}

void AR4MLSManifold3f::createTopology3f(card32 numNN, float32 maxSpacing)
{
   if (tg) {
      delete tg;
      tg = NULL;
   }
   PCCComputeTopology compTop;
   if (dynamic_cast<UnstructuredInCoreTriangleMesh*>(pc)) {
      compTop.setup(PCCComputeTopology::TOPTYPE_TRIMESH_CONNECTIVITY, maxSpacing, numNN);
   } else {
      compTop.setup(PCCComputeTopology::TOPTYPE_KNN_AND_EPS, maxSpacing, numNN);
   }
   
   tg = compTop.computeTopology(pc->getPointSet(), pc);
   attachTopology(pc, tg);
}

void AR4MLSManifold3f::getKNearestNeighborsEucl(Vector3f point, mpcard k, vector<mpcard> &neighbors)
{
   knnEucl->setTargetPoint(point);
   mpcard numRet = 0;
   neighbors.resize(k);
   do {
      bool ok = false;
      card32 nextIndex = (card32)-1;
      knnEucl->getNextNearestPointIndex(nextIndex, ok);
      if (ok) {
         neighbors[numRet] = nextIndex;
         numRet++;
      } else {
         break;
      }
   } while (numRet < k);
   neighbors.resize(numRet); // in case point cloud is smaller than k
}

void AR4MLSManifold3f::getSphericalNeighborsNG(mpcard index, FloatType radius, vector<mpcard> &neighbors)
{
   sQueryNG->setup(tg, ps, index, ps->get3f(index, positionAAT), radius, positionAAT);
   while (!sQueryNG->atEnd()) {
      neighbors.push_back(sQueryNG->getCurrentIndex());
      sQueryNG->next();
   }
}

void AR4MLSManifold3f::getSphericalNeighborsNG(Vector3f point, FloatType radius, vector<mpcard> &neighbors)
{
   knnEucl->setTargetPoint(point);
   card32 index;
   bool found;
   knnEucl->getNextNearestPointIndex(index, found);
   if (!found) throw EInvalidState("AR4MLSManifold3f::getSphericalNeighbors() - point set empty.");
   
   sQueryNG->setup(tg, ps, index, point, radius, positionAAT);
   while (!sQueryNG->atEnd()) {
      neighbors.push_back(sQueryNG->getCurrentIndex());
      sQueryNG->next();
   }
}

X4::BoundingBox3f X4::AR4MLSManifold3f::getBoundingBox()
{
   return ps->getBoundingBox();
}



void X4::AR4MLSManifold3f::setImlsConvThresholdFact( float32 val )
{
   imlsConvThresholdFact = val;
   if (ps) {
      imlsConvThreshold = imlsConvThresholdFact*ps->getBoundingBox().getMaxSideLength();   
   }
}

bool AR4MLSManifold3f::computeTangentRotation(const Vector3f& startPoint, const Vector3f& endPoint, Matrix3f& rotation)
{

	Vector3f proj1, proj2;
	Vector3f normal1, normal2;
	StaticMatrix<float32, 2, 3> tangents1, tangents2;

	Matrix3f frame1, frame2;

	if (!project(startPoint, proj1, normal1, tangents1))
		return false;
	if (!project(endPoint, proj2, normal2, tangents2))
		return false;

	frame1[0] = tangents1[0];
	frame1[1] = tangents1[1];
	frame1[2] = normal1;
	frame2[0] = tangents2[0];
	frame2[1] = tangents2[1];
	frame2[2] = normal2;

	rotation = frame2 * frame1.transpose();

	return true;
}

bool AR4MLSManifold3f::transportVector(const Vector3f& vStart, const vector<mpcard>& path, Vector3f& vEnd)
{
	mpcard i;
	const mpcard pathSteps = path.size();

	Vector3f dir;
	Vector3f p = getDiscrPoint(path[0]);
	Vector3f proj, norm;
	StaticMatrix<float32, 2, 3> tangents;
	Vector2f v;

	Matrix3f rotation;

	project(p, proj, norm, tangents);
	v[0] = tangents[0] * vStart;
	v[1] = tangents[1] * vStart;
	dir = tangents * v;

	for (i = 1; i < pathSteps; i++)
	{
		p = getDiscrPoint(path[i]);
		
		if (!computeTangentRotation(proj, p, rotation))
			return false;

		dir = rotation * dir;

		proj = p;
	}

	vEnd = dir;
	return true;
}

bool AR4MLSManifold3f::transportVector(const Vector3f& vStart, const vector<Vector3f>& path, Vector3f& vEnd)
{
	mpcard i;
	const mpcard pathSteps = path.size();

	Vector3f dir;
	Vector3f proj, norm;
	StaticMatrix<float32, 2, 3> tangents;
	Vector2f v;

	project(path[0], proj, norm, tangents);
	v[0] = tangents[0] * vStart;
	v[1] = tangents[1] * vStart;
	v.normalize();
	dir = tangents * v;
	dir.normalize();

	for (i = 1; i < pathSteps; i++)
	{
		project(path[i], proj, norm, tangents);
		v[0] = tangents[0] * dir;
		v[1] = tangents[1] * dir;
		v.normalize();
		dir = tangents * v;
		dir.normalize();
	}

	vEnd = dir;
	return true;
}

bool AR4MLSManifold3f::jumpVector(const Vector3f& startPoint, const Vector3f& endPoint, const Vector3f& vStart, Vector3f& vEnd)
{
	Matrix3f rotation;
	Vector3f dir;
	Vector3f proj, norm;
	StaticMatrix<float32, 2, 3> tangents;
	Vector2f v;

	project(startPoint, proj, norm, tangents);
	v[0] = tangents[0] * vStart;
	v[1] = tangents[1] * vStart;
	dir = tangents * v;

	if (!computeTangentRotation(startPoint, endPoint, rotation))
		return false;

	vEnd = rotation * dir;
	return true;
}

void AR4MLSManifold3f::computeSampleEpsilon()
{
	mpcard i;
	const mpcard n = pc->getNumPoints();

	float32 nbrDistMin = 1e30f;
	float32 nbrDistMax = 0.f;
	for (i = 0; i < n; i++)
	{
		const Vector3f vx = ps->get3f(i, positionAAT);
		const mpcard numNbrs = tg->getNumAdjacentVertices(i);
		for (mpcard j = 0; j < numNbrs; j++)
		{
			const mpcard nbr = tg->getAdjacentVertex(i, j);
			const Vector3f nx = ps->get3f(nbr, positionAAT);
			const float32 dist = (nx - vx).getNorm();
			if (dist < nbrDistMin)
				nbrDistMin = dist;
			if (dist > nbrDistMax)
				nbrDistMax = dist;
		}
	}

	m_sampleEpsilon = nbrDistMin;

	if (m_sampleFactorHarmonics * m_sampleEpsilon < nbrDistMax)
	{
		m_sampleFactorHarmonics = (nbrDistMax / m_sampleEpsilon) * 1.25f;//heuristic to ensure every point has harmonic neighbors
	}

}

void AR4MLSManifold3f::computeHarmonicNeighbors()
{
	const FloatType eps = m_sampleEpsilon * m_sampleFactorHarmonics;

	mpcard i;
	const mpcard n = pc->getNumPoints();

	m_harmonicNeighbors.resize(n);

	for (i = 0; i < n; i++)
	{
		const VectorType point = ps->get3f(i, positionAAT);

		FloatType dist = (FloatType)0;
		card32 idx;
		bool bfound;

		m_harmonicNeighbors[i].clear();

		knnEucl->setTargetPoint(point);

		knnEucl->getNextNearestPointIndex(idx, bfound);//skip the point itself
		knnEucl->getNextNearestPointIndex(idx, bfound);
		while (bfound && dist < eps)
		{
			const VectorType vnbr = ps->get3f(idx, positionAAT);

			dist = (vnbr - point).getNorm();
			if (dist > eps)
				break;

			m_harmonicNeighbors[i].push_back((mpcard)idx);

			knnEucl->getNextNearestPointIndex(idx, bfound);
		}
	}
}

void AR4MLSManifold3f::computePointAreas()
{
	mpcard i;
	const mpcard n = pc->getNumPoints();

	m_pointAreas.setDim(n);


	m_avgPointArea = 0;

	UnstructuredInCoreTriangleMesh* utm = dynamic_cast<UnstructuredInCoreTriangleMesh*>(pc);
	if (utm != NULL)
	{
		debugOutput << " computing point areas unstructured in core triangle mesh...\n"; 
		m_pointAreas.setZero();

		const mpcard ntris = utm->getNumTriangles();
		
		PointSet* ptris = utm->getTriangles();
		AAT triAAT = ptris->getAAT("index");
		for (i = 0; i < ntris; i++)
		{
			Vector3i tri = ptris->get3i(i, triAAT);
			card32 ip = tri[0], iq = tri[1], ir = tri[2];
			const Vector3f p = ps->get3f(ip, positionAAT);
			const Vector3f q = ps->get3f(iq, positionAAT);
			const Vector3f r = ps->get3f(ir, positionAAT);

			Vector3f pq = p - q, pr = p - r;
			Vector3f qr = q - r, qp = q - p;
			Vector3f rp = r - p, rq = r - q;
			const float32 lpq = pq.getNorm(), lpr = pr.getNorm();
			const float32 lqr = qr.getNorm(), lqp = qp.getNorm();
			const float32 lrp = rp.getNorm(), lrq = rq.getNorm();

			
			pq.normalize(); pr.normalize();
			qr.normalize(); qp.normalize();
			rp.normalize(); rq.normalize();

			float32 anglep = acos(pq * pr);
			float32 angleq = acos(qr * qp);
			float32 angler = acos(rp * rq);

			if (anglep > M_PI_2 || angleq > M_PI_2 || angler > M_PI_2)
			{
				const float32 halfPerim = (lpq + lqr + lrp) * 0.5f;
				const float32 areaT = sqrt(halfPerim * (halfPerim - lpr) * (halfPerim - lpq) * (halfPerim - lqr));

				if (anglep > M_PI_2)	m_pointAreas[ip] += areaT * 0.5f;
				else					m_pointAreas[ip] += areaT * 0.25f;
				if (angleq > M_PI_2)	m_pointAreas[iq] += areaT * 0.5f;
				else					m_pointAreas[iq] += areaT * 0.25f;
				if (angler > M_PI_2)	m_pointAreas[ir] += areaT * 0.5f;
				else					m_pointAreas[ir] += areaT * 0.25f;
			}
			else
			{
				m_pointAreas[ip] += 0.125f * (pow(lpr,2) / tan(angleq) + pow(lpq,2) / tan(angler));
				m_pointAreas[iq] += 0.125f * (pow(lqr,2) / tan(anglep) + pow(lpq,2) / tan(angler));
				m_pointAreas[ir] += 0.125f * (pow(lpr,2) / tan(angleq) + pow(lqr,2) / tan(anglep));
			}
		}

		float sumArea = 0.0;
		for (i = 0; i < n; i++)
			sumArea += m_pointAreas[i];
	
		debugOutput << "sum area " << sumArea << "\n";

		for (i = 0; i < n; i++)
			m_avgPointArea += m_pointAreas[i];
		m_avgPointArea /= (float32)n;

		m_bAreasDirty = false;
	}//if triangle mesh
	else
		warning("AR4MLSManifold3f::computePointAreas(): not implemented for general point clouds");

	if (m_harmonicNeighbors.size() != n)
		computeHarmonicNeighbors();


}

void AR4MLSManifold3f::computeLaplace()
{
	const FloatType xi = (FloatType)0.01;
	const FloatType t_param = (FloatType)pow((double)m_sampleEpsilon, (double)(1.0 / (2.0 + xi)));
	const FloatType exp_scale = (FloatType)(1.0 / (4.0 * t_param));
	const FloatType lead_scale = (FloatType)(1.0 / (4.0 * M_PI * t_param*t_param));
	mpcard i;
	const mpcard n = pc->getNumPoints();

	if (m_bAreasDirty)
		computePointAreas();

	m_laplace.setRows(n);
	m_laplace.setZero();

	UnstructuredInCoreTriangleMesh* utm = dynamic_cast<UnstructuredInCoreTriangleMesh*>(pc);
	if (utm != NULL)
	{
		const mpcard ntris = utm->getNumTriangles();
		PointSet* ptris = utm->getTriangles();
		AAT triAAT = ptris->getAAT("index");
		
		//iterate over all triangles
		for (i = 0; i < ntris; i++)
		{
			Vector3i tri = ptris->get3i(i, triAAT);
			card32 ip = tri[0], iq = tri[1], ir = tri[2];

			const Vector3f p = ps->get3f(ip, positionAAT);
			const Vector3f q = ps->get3f(iq, positionAAT);
			const Vector3f r = ps->get3f(ir, positionAAT);


			Vector3f pq = p - q, pr = p - r;
			Vector3f qr = q - r, qp = q - p;
			Vector3f rp = r - p, rq = r - q;
			const float32 lpq = pq.getNorm(), lpr = pr.getNorm();
			const float32 lqr = qr.getNorm(), lqp = qp.getNorm();
			const float32 lrp = rp.getNorm(), lrq = rq.getNorm();

			pq.normalize(); pr.normalize();
			qr.normalize(); qp.normalize();
			rp.normalize(); rq.normalize();

			float32 anglep = acos(pq * pr);
			float32 angleq = acos(qr * qp);
			float32 angler = acos(rp * rq);

			float32 wpq = 0.5 / tan(angler);
			float32 wpr = 0.5 / tan(angleq);
			float32 wqr = 0.5 / tan(anglep);
			float32 entry;


			//off diagonals
			//p q
			entry = m_laplace[ip].getEntry(iq);
			m_laplace[ip].setEntry(iq, entry - wpq);
			m_laplace[ip].setEntry(iq, entry - wpq);
			//symmetric-> pq same value with qp
			//q p
			m_laplace[iq].setEntry(ip, entry - wpq);
			//p r
			entry = m_laplace[ip].getEntry(ir);
			m_laplace[ip].setEntry(ir, entry - wpr);
			//r p
			m_laplace[ir].setEntry(ip, entry - wpr);
			//q r
			entry = m_laplace[iq].getEntry(ir);
			m_laplace[iq].setEntry(ir, entry - wqr);
			//r q
			m_laplace[ir].setEntry(iq, entry - wqr);
		}//for

		float32 entry;
		//iterate over all matrix entries to check for boundaries
		//when boundaries, double the contagent weight
		float sumRow = 0.0;
		for(i = 0; i < n; i++)
		{
			sumRow = 0.0;

			const mpcard numEntries = m_laplace[i].getNumSparseEntries();
			for (mpcard j = 0; j < numEntries; j++)
			{
				mpcard idx;
				m_laplace[i].getSparseEntry(j, idx, entry);
				sumRow -= entry;
			}
		
			//diagonal element
			m_laplace[i].setEntry(i,sumRow);

		}//for
	}
	else
		warning("AR4MLSManifold3f::computeLaplace(): not implemented for general point clouds");
}




void AR4MLSManifold3f::computePointNormals()
{

	mpcard i;
	const mpcard n = pc->getNumPoints();

	UnstructuredInCoreTriangleMesh* utm = dynamic_cast<UnstructuredInCoreTriangleMesh*>(pc);


	if ( ( utm == NULL ) ){
		debugOutput << "computePointNormals: invalid mesh.\n"; debugOutput.flush();
		return;
	}


	if (utm != NULL)
	{

		PointSet* vertices = utm->getPointSet();
		PointSet* ptris = utm->getTriangles();
		const mpcard ntris = utm->getNumTriangles();

		AAT triAAT = ptris->getAAT("index");
		AAT NORMAL = vertices->getAAT( "normal" );
		
		

		if ( !vertices->providesAttribute( "position" ) || !vertices->providesAttribute( "normal" ) ){
			debugOutput << "computePointNormals: vertex attribute channel missing.\n"; debugOutput.flush();
			return;
		}

		if ( !ptris->providesAttribute( "index" ) ){
			debugOutput << "computePointNormals: triangle attribute channel missing.\n"; debugOutput.flush();
			return;
		}

		
		vector<Vector3f> m_pointNormals; m_pointNormals.resize( vertices->getNumEntries() );
		vector<card32> num_m_pointNormals; num_m_pointNormals.resize( vertices->getNumEntries() );


		// Init container
		for ( unsigned int normalIndex = 0; normalIndex < vertices->getNumEntries(); normalIndex++ ){
			m_pointNormals[normalIndex] = makeVector3f( 0,0,0 );
			num_m_pointNormals[normalIndex] = 0;
		}
	
	
		
		//iterate over all triangles
		for (i = 0; i < ntris; i++)
		{
			
			
			Vector3i tri = ptris->get3i(i, triAAT);
			card32 ip = tri[0], iq = tri[1], ir = tri[2];

			const Vector3f p = ps->get3f(ip, positionAAT);
			const Vector3f q = ps->get3f(iq, positionAAT);
			const Vector3f r = ps->get3f(ir, positionAAT);


			Vector3f pq = p - q, pr = p - r;
			Vector3f qr = q - r, qp = q - p;
			Vector3f rp = r - p, rq = r - q;
			const float32 lpq = pq.getNorm(), lpr = pr.getNorm();
			const float32 lqr = qr.getNorm(), lqp = qp.getNorm();
			const float32 lrp = rp.getNorm(), lrq = rq.getNorm();

			pq.normalize(); pr.normalize();
			qr.normalize(); qp.normalize();
			rp.normalize(); rq.normalize();

			float32 anglep = acos(pq * pr);
			float32 angleq = acos(qr * qp);
			float32 angler = acos(rp * rq);

			float32 wpq = 0.5 / tan(angler);
			float32 wpr = 0.5 / tan(angleq);
			float32 wqr = 0.5 / tan(anglep);
		
			
			Vector3f triangleNormal =  pq.crossProduct(rq);
			float32 n = norm( triangleNormal );

			

			if ( n > 1e-7 ) 
			{
				triangleNormal /= n;

				m_pointNormals[ip] += (wpq + wpr) * triangleNormal;
				num_m_pointNormals[ ip ]++;

				m_pointNormals[iq] += (wpq + wqr) * triangleNormal;
				num_m_pointNormals[ iq ]++;

				m_pointNormals[ir] += (wpr + wqr) * triangleNormal;
				num_m_pointNormals[ ir ]++;
			}


		}

		




		//iterate over all vertices
		for (unsigned int normalIndex = 0; normalIndex < n; normalIndex++)
		{
			if ( num_m_pointNormals[normalIndex] > 0 ){

				vertices->set3f( normalIndex, NORMAL, -m_pointNormals[normalIndex] / float32( num_m_pointNormals[normalIndex] ) );

			}
			else {

				vertices->set3f( normalIndex, NORMAL, makeVector3f( 1,0,0 ) );
			
			}


		}


	//if it is a triangle mesh
	}



	




}








void AR4MLSManifold3f::computeGradientOperator()
{
	const mpcard numPts = getNumDiscrPoints();

	mpcard i, j;

	m_gradient.setRows(3 * numPts);
	m_divergence.setRows(numPts);


	for (i = 0; i < numPts; i++)
	{
		
		Vector3f pos, projPos, normal;
		StaticMatrix<float, 2, 3> ts;
		StaticMatrix<float, 3, 2> tst;
		Matrix3f tsn;

		pos = getDiscrPoint(i);

		project(pos, projPos, normal, ts);

		tst = ts.transpose();
		tsn[0] = ts[0];
		tsn[1] = ts[1];
		tsn[2] = normal;

		const mpcard nnbrs = m_harmonicNeighbors[i].size();
		const mpcard numOp = nnbrs + 1;
		const float rcpNumOp = 1.f / (float)numOp;

		float avgSqrDist = 0.f;
		
		for (j = 0; j < nnbrs; j++)
		{
			const mpcard nbr = m_harmonicNeighbors[i][j];
			const Vector3f nbrPt = getDiscrPoint(nbr);
			const Vector2f nbrPtTS = tst * (nbrPt - pos);
			const float u = nbrPtTS[0];
			const float v = nbrPtTS[1];
			avgSqrDist += u*u + v*v;
		}
		
		avgSqrDist /= (float)nnbrs;
		const float sigma2 = avgSqrDist;
		const float rcp2Sigma2 = 1.f / (2.f * sigma2);

		DynamicMatrix<float> weightMat;
		weightMat.setDimension(numOp, 3);

		weightMat[0][0] = 1.f;
		weightMat[0][1] = 0.f;
		weightMat[0][2] = 0.f;

		Matrix3f M, Minv, Mperm, Mfinal;
		M.setZero();
		M[0][0] = 1.f;

		for (j = 0; j < nnbrs; j++)
		{
			const mpcard k = j + 1;
			const mpcard nbr = m_harmonicNeighbors[i][j];
			const Vector3f nbrPt = getDiscrPoint(nbr);
			const Vector2f nbrPtTS = tst * (nbrPt - pos);
			const float u = nbrPtTS[0];
			const float v = nbrPtTS[1];
			const float w = exp(-(u*u + v*v) * rcp2Sigma2);

			// 1 u v uu uv vv
			M[0][0] += w;     M[0][1] += u*w;     M[0][2] += v*w;  
			M[1][0] += w*u;   M[1][1] += u*w*u;   M[1][2] += v*w*u;
			M[2][0] += w*v;   M[2][1] += u*w*v;   M[2][2] += v*w*v;

			weightMat[k][0] = w;
			weightMat[k][1] = w * u;
			weightMat[k][2] = w * v;
		}

	

		bool bSuccess;
		Minv = invertMatrix(M, &bSuccess, 1e-7f);
		if (!bSuccess)
		{
			debugOutput << "Could not invert matrix for point " << i << "\n";
		}
		
		Mperm.setZero();
		Mperm[1][0] = 1.f;
		Mperm[2][1] = 1.f;

		Mfinal = tsn * Mperm * Minv;

		m_gradient[i * 3 + 0].setEntry(i, Mfinal[0][0]);
		m_gradient[i * 3 + 1].setEntry(i, Mfinal[0][1]);
		m_gradient[i * 3 + 2].setEntry(i, Mfinal[0][2]);

		//cout << " iterate over the neighbors for the third time \n";
		for (j = 0; j < nnbrs; j++)
		{
			const mpcard k = j + 1;
			const mpcard nbr = m_harmonicNeighbors[i][j];
			const Vector3f temp = Mfinal[0] * weightMat[k][0] + Mfinal[1] * weightMat[k][1] + Mfinal[2] * weightMat[k][2];
			m_gradient[i * 3 + 0].setEntry(nbr, temp[0]);
			m_gradient[i * 3 + 1].setEntry(nbr, temp[1]);
			m_gradient[i * 3 + 2].setEntry(nbr, temp[2]);
			
		}

	}

	m_divergence = m_gradient.transpose();


}



void AR4MLSManifold3f::findScalarFieldMinimum(const vector<float>& scalarField, const vector<Vector3f>& hintSet, Vector3f& minimalPoint, float& minimalValue)
{
	const mpcard num_init_nearest = 6;
	const mpcard num_nbrs_fit = 10;

	const mpcard numFieldValues = scalarField.size();
	const mpcard numPoints = pc->getNumPoints();

	if (numFieldValues != numPoints)
		throw ERangeCheck("AR4MLSManifold3f::findScalarFieldMinimum(...): scalar field different size than point cloud");

	const mpcard numHints = hintSet.size();

	mpcard i, j;

	card32 idxnear;
	bool bfound;

	vector<card32> pointStatus(numPoints, 0);

	queue<card32> idxQueue;

	for (i = 0; i < numHints; i++)
	{
		knnEucl->setTargetPoint(hintSet[i]);
		knnEucl->getNextNearestPointIndex(idxnear, bfound);

		for (j = 0; j < num_init_nearest && bfound; j++)
		{
			idxQueue.push(idxnear);
			knnEucl->getNextNearestPointIndex(idxnear, bfound);
		}
	}

	float valMin = 1e30f;
	card32 idxMin = (card32)numPoints;

	while (!idxQueue.empty())
	{
		card32 idx = idxQueue.front();
		idxQueue.pop();

		if (pointStatus[idx] > 0)
			continue;

		pointStatus[idx]++;

		const float curval = scalarField[idx];
		if (curval < valMin)
		{
			valMin = curval;
			idxMin = idx;
		}

		const card32 numNbrs = tg->getNumAdjacentVertices(idx);
		for (i = 0; i < numNbrs; i++)
		{
			const card32 nbr = tg->getAdjacentVertex(idx, i);
			if (pointStatus[nbr] == 0 && scalarField[nbr] < scalarField[idx])
			{
				idxQueue.push(nbr);
			}
		}
	}

	if (idxMin == numPoints)
	{
		minimalValue = 1e30f;
		minimalPoint = NULL_VECTOR3F;
		return;
	}

	minimalPoint = ps->get3f(idxMin, positionAAT);
	minimalValue = valMin;

	vector<Vector3f> tangentSpacePoints, tangentSpacePointsSurface;
	Vector3f projPoint, projNorm;
	TangentSpace ts;
	Matrix3f matToTS, matFromTS;

	project(minimalPoint, projPoint, projNorm, ts);

	matFromTS[0] = ts[0];
	matFromTS[1] = ts[1];
	matFromTS[2] = projNorm;
	matToTS = matFromTS.transpose();

	knnEucl->setTargetPoint(minimalPoint);
	knnEucl->getNextNearestPointIndex(idxnear, bfound);
	for (i = 0; i < num_nbrs_fit && bfound; i++)
	{
		Vector3f tmpPos = ps->get3f(idxnear, positionAAT) - minimalPoint;

		tmpPos = matToTS * tmpPos;

		tangentSpacePointsSurface.push_back(tmpPos);

		tmpPos[2] = scalarField[idxnear];
		tangentSpacePoints.push_back(tmpPos);

		knnEucl->getNextNearestPointIndex(idxnear, bfound);
	}

	Vector3f critPoint;
	Vector6f coeffs;

	Monomials2ndOrder::computeCoeffs(tangentSpacePoints, coeffs);

	if (Monomials2ndOrder::evaluateCriticalPoint(coeffs, critPoint) && critPoint[2] >= 0 && critPoint[2] < minimalValue)
	{
		minimalValue = critPoint[2];

		Monomials2ndOrder::computeCoeffs(tangentSpacePointsSurface, coeffs);
		critPoint[2] = Monomials2ndOrder::evaluate(coeffs, makeVector2f(critPoint[0], critPoint[1]));
		critPoint = minimalPoint + matFromTS * critPoint;

		minimalPoint = critPoint;
	}
}

AR4MLSManifold3f::FloatType AR4MLSManifold3f::getSurfaceArea()
{
	FloatType area = 0;

	UnstructuredInCoreTriangleMesh* ptm = dynamic_cast<UnstructuredInCoreTriangleMesh*>(pc);
	if (ptm != NULL)
	{
		//we have a triangle mesh, compute the area by summing the area of the triangles
		mpcard t, numtri = ptm->getNumTriangles();
		PointSet* ptris = ptm->getTriangles();
		PointSet* ppts = ptm->getPointSet();
		const VertexDescriptor* ptrivd = ptm->getTriangleDescr();
		AAT IND = ptris->getAAT("index");
		AAT POS = ppts->getAAT("position");
		float areatri;

		for (t = 0; t < numtri; t++)
		{
			Vector3i indices = ptris->get3i(t, IND);
			Vector3f v1 = ppts->get3f(indices[0], POS);
			Vector3f v2 = ppts->get3f(indices[1], POS);
			Vector3f v3 = ppts->get3f(indices[2], POS);
			areatri = (v3 - v1).crossProduct(v2 - v1).getNorm() * 0.5f;
			area += areatri;
		}
	}
	else
	{
		if (m_bAreasDirty)
			computePointAreas();

		mpcard i;
		const mpcard nPts = pc->getNumPoints();
		for (i = 0; i < nPts; i++)
		{
			area += m_pointAreas[i];
		}
	}

	return area;
}

AR4MLSManifold3f* AR4MLSManifold3f::createFromPointCloud(PointCloud* pc,
														 const card32 NumNearestNeighbors,
														 const float32 epsilon /*= 0.1f*/)
{
	if (!pc)
	{
		throw ERangeCheck("AR4MLSManifold3f::createFromPointCloud() - no valid point cloud given.");
	}

	UnstructuredInCorePointCloud* upc = dynamic_cast<UnstructuredInCorePointCloud*>(pc);
	if (!upc)
	{
		throw ERangeCheck("AR4MLSManifold3f::createFromPointCloud() - input must currently be unstructured PC.");
	}

	UnstructuredInCoreTriangleMesh* utm = dynamic_cast<UnstructuredInCoreTriangleMesh*>(pc);
	if (utm != NULL)
	{

		utm->computeTopologyAttachment(true);
	}

	InCorePCTopologyGraph* tg = X4::getTopology(pc);
	if (!tg)
	{
		PCCComputeTopology ctop;
		ctop.setup(PCCComputeTopology::TOPTYPE_KNN, epsilon, NumNearestNeighbors);
		tg = ctop.computeKNNTopology(upc->getPointSet(), upc);
		attachTopology(upc, tg);
	}

	PCCEstimateSampleSpacing::estimateSampleSpacing(upc, NumNearestNeighbors);

	return new AR4MLSManifold3f(upc);
}

void AR4MLSManifold3f::detectBoundaries()
{
	checkAttribute(pc, "color", 3, VAD::DATA_FORMAT_FLOAT32);
	checkAttribute(pc, "boundary", 1, VAD::DATA_FORMAT_FLOAT32);
	checkAttribute(pc, "boundaryTangent", 3, VAD::DATA_FORMAT_FLOAT32);

	resetup(pc);

	AAT boundary_AAT = pc->getAAT("boundary"); 
	AAT boundary_tang_AAT = pc->getAAT("boundaryTangent");

	UnstructuredInCoreTriangleMesh* utm = dynamic_cast<UnstructuredInCoreTriangleMesh*>(pc);

	if (utm == NULL)
	{
		//not a triangle mesh
		PCCBoundaryDetector bd;

		bd.detectBoundaries(pc);
		return;
	}

	PointSet* triangles = utm->getTriangles();
	AAT TIND = triangles->getAAT("index");
	vector<Vector2i> borderEdges;
	
	utm->cleanTopology();
	

	//START DETECTING BOUNDARY EDGES
	X4_TIMER_START(my_boundary_detection);
	map<card64, card64> edgeMap;
	map<card64, card64>::iterator edge_it;

	for(mpcard i=0; i<ps->getNumEntries(); i++)
	{
		ps->set1f(i, boundary_AAT, 0.f);
		ps->set3f(i, boundary_tang_AAT, NULL_VECTOR3F);
	}

	for (mpcard i=0; i<triangles->getNumEntries(); ++i)
	{
		Vector3i tri = triangles->get3i(i,TIND);
		
		for( unsigned j=0;j<3;j++ )
		{
			mpcard i0 = tri[j];
			mpcard i1 = tri[(j+1)%3];
			card64 id = min(i0,i1)*ps->getNumEntries() + max(i0,i1);
			
			if (edgeMap.find(id) != edgeMap.end())
				edgeMap.erase(id);
			else
				edgeMap[id] = i;
		}
	}

	for(edge_it = edgeMap.begin(); edge_it != edgeMap.end(); ++edge_it)
	{
		const card64 i0 = edge_it->first / ps->getNumEntries();
		const card64 i1 = edge_it->first % ps->getNumEntries();
		const card64 t = edge_it->second;
		
		card64 i2;//other vertex of triangle

		borderEdges.push_back( makeVector2i(i0, i1) );

		Vector3i tri = triangles->get3i(t,TIND);

		if (tri[0] != i0 && tri[0] != i1)
			i2 = tri[0];
		else if (tri[1] != i0 && tri[1] != i1)
			i2 = tri[1];
		else
			i2 = tri[2];

		const float b0prev = ps->get1f(i0, boundary_AAT);
		const float b1prev = ps->get1f(i1, boundary_AAT);
		const Vector3f b0tanprev = ps->get3f(i0, boundary_tang_AAT);
		const Vector3f b1tanprev = ps->get3f(i1, boundary_tang_AAT);

		const Vector3f v0 = ps->get3f(i0, positionAAT);
		const Vector3f v1 = ps->get3f(i1, positionAAT);
		const Vector3f v2 = ps->get3f(i2, positionAAT);

		ps->set1f(i0, boundary_AAT, 1.f);
		ps->set1f(i1, boundary_AAT, 1.f);

		const Vector3f n0 = ps->get3f(i0, normalAAT);
		const Vector3f n1 = ps->get3f(i1, normalAAT);
		Vector3f edge = v1 - v0;
		edge.normalize();

		Vector3f e02 = v2 - v0;
		Vector3f e12 = v2 - v1;
		e02.normalize();
		e12.normalize();

		Vector3f en;//"edge-normal"

		//for a counter-clockwise ordered triangle, normal cross edge point into the triangle
		Vector3f t0 = n0.crossProduct(edge);
		en = edge.crossProduct(e02);
		en.normalize();
		if (en * n0 > 0.f)
			t0 = -t0;
		t0.normalize();

		Vector3f t1 = n1.crossProduct(-edge);
		en = -edge.crossProduct(e12);
		en.normalize();
		if (en * n1 > 0.f)
			t1 = -t1;
		t1.normalize();

		t0 += b0tanprev * b0prev;
		t1 += b1tanprev * b1prev;
		t0 /= (b0prev + 1.f);
		t1 /= (b1prev + 1.f);
		t0.normalize();
		t1.normalize();

		ps->set3f(i0, boundary_tang_AAT, t0);
		ps->set3f(i1, boundary_tang_AAT, t1);
	}

	X4_TIMER_STOP(my_boundary_detection);


}



void AR4MLSManifold3f::getRingNeighborhood(mpcard sourceIndex, mpcard rings, vector<mpcard>& neighbors)
{

	mpcard i, vert, numNbrs, nbr;
	pair<mpcard, mpcard> nbrdepth;
	queue<pair<mpcard, mpcard>> nbrsToVisit;
	const mpcard n = pc->getNumPoints();
	vector<bool> visited(n, false);

	InCorePCTopologyGraph* tg = X4::getTopology(pc);

	neighbors.clear();

	nbrdepth.first = sourceIndex;
	nbrdepth.second = 0;
	nbrsToVisit.push(nbrdepth);


	while (!nbrsToVisit.empty())
	{
		nbrdepth = nbrsToVisit.front();
		nbrsToVisit.pop();

		vert = nbrdepth.first;
		if (vert != sourceIndex)
			neighbors.push_back(vert);

		if (nbrdepth.second < rings)
		{
			nbrdepth.second++;
			numNbrs = tg->getNumAdjacentVertices(vert);
			for (i = 0; i < numNbrs; i++)
			{
				nbr = tg->getAdjacentVertex(vert,i);
				if (!visited[nbr])
				{
					nbrdepth.first = nbr;
					visited[nbr] = true;
					nbrsToVisit.push(nbrdepth);
				}//if
			}//for	
		}//if
	}//while



}
