#ifndef AR4MLSManifoldH
#define AR4MLSManifoldH

#include "threeParameterMatchingGMOD14.h"

#include "DynamicLinearAlgebra.h"

#include "AR4DiscreteManifold.h"
#include "AR4ManifoldTraits.h"
#include "AR4SmoothManifold.h"
#include "UnstructuredInCorePointCloud.h"
#include "LinearAlgebra.inline.h"
#include "DynamicLinearAlgebra.inline.h"
#include "SparseLinearAlgebra.inline.h"

#include "SparseGrid.h"


namespace X4 {

class InCorePCTopologyGraph;
class PointSet;
class PointSetKNNQuery;
class AttachedIndexedOctree;
class TopNeighborhoodKNNSearch_Fast;
class TopNeighborhoodSphereSearch;


/// this is a manifold represented implicitly by point samples. yes, once again.
template <class ManifoldTraits>
class THREE_PARAMETER_MATCHING_GMOD14_API AR4MLSManifold : public AR4DiscreteManifold<ManifoldTraits>, public AR4SmoothManifold<ManifoldTraits> {
   
 protected:
   UnstructuredInCorePointCloud *pc;
   PointSet *ps;
   InCorePCTopologyGraph *tg;
   AAT sampleSpacingAAT;
   AAT positionAAT;
   AAT normalAAT;

   //the following member for point-based manifold harmonics
   FloatType								m_sampleEpsilon;

   //neighborhood size for manifold harmonics (relative to sample spacing)
   FloatType								m_sampleFactorHarmonics;

   vector<vector<mpcard>>					m_harmonicNeighbors;

   DynamicVector<FloatType>					m_pointAreas;
   FloatType								m_avgPointArea;
   bool										m_bAreasDirty;

   SparseMatrix<FloatType>					m_laplace;
   SparseMatrix<FloatType>					m_gradient;
   SparseMatrix<FloatType>					m_divergence;


 
	vector<Vector3f>						m_pointNormals;

 
   
 public:
    /// does not take ownership, only references
    /// requires topology
	AR4MLSManifold(UnstructuredInCorePointCloud *pc);

   /// reinitialize, for example, if pointers change
   virtual void resetup(UnstructuredInCorePointCloud *pc);

   inline UnstructuredInCorePointCloud *getPointCloud() {return pc;}
   inline PointSet *getPointSet() {return ps;}
   inline InCorePCTopologyGraph *getTopology() {return tg;}

   inline float32 getSampleSpacing(const mpcard index) const
   {
		return ps->get1f(index, sampleSpacingAAT);
   }

   virtual mpcard getNumDiscrPoints();
   virtual VectorType getDiscrPoint(mpcard index);

   virtual mpcard getNumNeighbors(mpcard point);
   virtual mpcard getNeighborIndex(mpcard point, mpcard index);

   virtual bool transportVector(const VectorType& vStart, const vector<mpcard>& path, VectorType& vEnd) { warning("AR4MLSManifold<ManifoldTraits>::transportVector(...) not implemented for abstract types"); return false; }
   virtual bool transportVector(const VectorType& vStart, const vector<VectorType>& path, VectorType& vEnd) { warning("AR4MLSManifold<ManifoldTraits>::transportVector(...) not implemented for abstract types"); return false; }
   virtual bool jumpVector(const VectorType& startPoint, const VectorType& endPoint, const VectorType& vStart, VectorType& vEnd) { warning("AR4MLSManifold<ManifoldTraits>::jumpVector(...) not implemented for abstract types"); return false; }

   FloatType getAvgEdgeLength();

   FloatType getSampleFactorHarmonics()				{ return m_sampleFactorHarmonics; }
   void setSampleFactorHarmonics(FloatType factor)	{ m_sampleFactorHarmonics = factor; }
   void setPointSet(PointSet* psIn){ pc->setPointSet(psIn);}

	virtual ~AR4MLSManifold();
};


class THREE_PARAMETER_MATCHING_GMOD14_API AR4MLSManifold3f : public AR4MLSManifold<ManifoldTraits3f> {
 private:
   PointSetKNNQuery *knnEucl;
   TopNeighborhoodKNNSearch_Fast *knnNG;
   TopNeighborhoodSphereSearch *sQueryNG;
   AttachedIndexedOctree *octree;

   /// relative to max bb side length
   float32 imlsConvThresholdFact;
   /// actual convergence threshold (good value is ~1E-5*bb)
   float32 imlsConvThreshold;
   /// maximum number of iterations (default: 100)
   card32 imlsMaxNumIterations;
   /// step size of gradient descent. empirically: 0.5 works well
   float32 imlsGradFact;


   //for fast projection via look-ups
   class ProjCell
   {
   public:
	   vector<mpcard>								m_neighbors;
	   float32										m_avgSampleSpacing;
	   Vector3f										m_centroid;
	   ProjCell(): m_avgSampleSpacing(0.f), m_centroid(NULL_VECTOR3F)	{}
   };

   class ProjCellLongRange
   {
   public:
	   vector<Vector3i>								m_nearest;
	   float32										m_fineGridSize;
	   Vector3f										m_fineGridOrigin;
	   BoundingBox3f								m_bbox;
	   mpcard										m_gridSizeFactor;

	   ProjCellLongRange& operator=(const ProjCellLongRange& cell)
	   {
		   m_nearest = cell.m_nearest;
		   m_fineGridSize = cell.m_fineGridSize;
		   m_fineGridOrigin = cell.m_fineGridOrigin;
		   m_bbox = cell.m_bbox;
		   m_gridSizeFactor = cell.m_gridSizeFactor;
		   return *this;
	   }

	   inline Vector3i computeNearest(const Vector3f& p) const
	   {
		   const float32 rcpFGS = 1.f / m_fineGridSize;
		   const mpcard ystep = m_gridSizeFactor;
		   const mpcard zstep = m_gridSizeFactor*m_gridSizeFactor;

		   Vector3i nearestFineCell;

		   //compute integer coords
		   Vector3f plocal = p - m_bbox.lowerCorner;
		   Vector3i pint;
		   pint[0] = (int32)(max(plocal[0], 0.f) * rcpFGS);
		   pint[1] = (int32)(max(plocal[1], 0.f) * rcpFGS);
		   pint[2] = (int32)(max(plocal[2], 0.f) * rcpFGS);

		   //look-up possible nearest (max 6)
		   const mpcard idx = pint[2] * zstep + pint[1] * ystep + pint[0];
		   const mpcard offset = idx * 6;
		   const mpcard stop = offset + 6;

		   //find min dist from those
		   mpcard i;
		   float32 dist, distMin = 1e30f;
		   for (i = offset; i < stop; i++)
		   {
			   Vector3f fineGridCenter = m_fineGridOrigin;
			   fineGridCenter[0] += m_fineGridSize * (0.5f + (float32)m_nearest[i][0]);
			   fineGridCenter[1] += m_fineGridSize * (0.5f + (float32)m_nearest[i][1]);
			   fineGridCenter[2] += m_fineGridSize * (0.5f + (float32)m_nearest[i][2]);
			   dist = (fineGridCenter - p).getSqrNorm();
			   if (dist < distMin)
			   {
				   distMin = dist;
				   nearestFineCell = m_nearest[i];
			   }
		   }

		   return nearestFineCell;
	   }
   };


   SparseGrid<3, ProjCell, Object>					m_sparseGrid;
   float32											m_gridSize;
   float32											m_avgSampleSpacing;
   float32											m_projectRadius;

   SparseGrid<3, ProjCellLongRange, Object>			m_longRangeGrid;
   float32											m_longRangeGridSize;
   Vector3f											m_longRangeOrigin;

   const static mpcard								m_nFunctionBins = 1000;
   const static mpcard								m_lastFuncBin = m_nFunctionBins - 1;

   DynamicVector<float32>							m_weightFunc;
   DynamicVector<float32>							m_weightGrad;
   float32											m_swfStepSize;
   float32											m_swfStepSizeInv;

   vector<Vector3f>									m_vertexCache;
   vector<Vector3f>									m_normalCache;

   inline static float32 gaussApprox(const float32 dist2Scaled)
   {
	   //approximation of Gaussian: (1 - d)^4
	   //see Oeztireli et al. "Feature Preserving Point Set Surfaces based on Non-Linear Kernel Regression", Eurographics 2009
	   const float32 s = 1.f - dist2Scaled;
	   return s*s*s*s;
   }

   inline float32 weightFunc(const float32 dist2Scaled)
   {
	   const mpcard bin = min(m_lastFuncBin, (mpcard)(dist2Scaled * m_swfStepSizeInv));
	   return m_weightFunc[bin];
   }
   inline float32 weightFunc(const float32 dist2Scaled, float32& weightGrad)
   {
	   const mpcard bin = min(m_lastFuncBin, (mpcard)(dist2Scaled * m_swfStepSizeInv));
	   weightGrad = m_weightGrad[bin];
	   return m_weightFunc[bin];
   }
   inline float32 weightGrad(const float32 dist2Scaled)
   {
	   const mpcard bin = min(m_lastFuncBin, (mpcard)(dist2Scaled * m_swfStepSizeInv));
	   return m_weightGrad[bin];
   }

   void initProjectionLookups();
   void preNormNormals();

	void computeHarmonicNeighbors();

public:
   AR4MLSManifold3f(UnstructuredInCorePointCloud *pc);

   virtual void resetup(UnstructuredInCorePointCloud *pc);
	

   float32 getImlsGradFact() const { return imlsGradFact; }
   void setImlsGradFact(float32 val) { imlsGradFact = val; }
   float32 getImlsConvThresholdFact() const { return imlsConvThreshold; }
   void setImlsConvThresholdFact(float32 val);
   card32 getImlsMaxNumIterations() const { return imlsMaxNumIterations; }
   void setImlsMaxNumIterations(card32 val) { imlsMaxNumIterations = val; }

  
   virtual BoundingBox3f getBoundingBox();
   
   virtual void getKNearestNeighborsEucl(Vector3f point, mpcard k, vector<mpcard> &neighbors);
   virtual void getSphericalNeighborsNG(mpcard index, FloatType radius, vector<mpcard> &neighbors);
   virtual void getSphericalNeighborsNG(Vector3f point, FloatType radius, vector<mpcard> &neighbors);

   void estimateSampleSpacing3f(card32 numNN);
   void createTopology3f(card32 numNN, float32 maxSpacing);
   AR4MLSManifold3f *resample(float32 spacing);

   float getAvgEdgeLength()		{ return AR4MLSManifold<ManifoldTraits3f>::getAvgEdgeLength(); }

   virtual inline bool project(const Vector3f &startPos, Vector3f &surfacePoint);
   virtual inline bool project(const Vector3f &startPos, Vector3f &surfacePoint, Vector3f &normal);
   virtual bool project(const Vector3f &startPos, Vector3f &surfacePoint, Vector3f &normal, StaticMatrix<float32, 2, 3> &tangentSpace);
   
   bool project(const Vector3f &startPos, Vector3f &surfacePoint, Vector3f &tangentU, Vector3f &tangentV, Vector3f &normal, Vector3f &curvTensor_u_uv_v);
   bool project(const vector<float32>& funcOnSurface, const Vector3f &startPos, Vector3f &surfacePoint, Vector3f &normal, float32& funcValue);
   bool project(const vector<Vector3f>& funcOnSurface, const Vector3f &startPos, Vector3f &surfacePoint, Vector3f &normal, Vector3f& funcValue);

   bool computeTangentRotation(const Vector3f& startPoint, const Vector3f& endPoint, Matrix3f& rotation);

   bool transportVector(const Vector3f& vStart, const vector<mpcard>& path, Vector3f& vEnd);
   bool transportVector(const Vector3f& vStart, const vector<Vector3f>& path, Vector3f& vEnd);

   bool jumpVector(const Vector3f& startPoint, const Vector3f& endPoint, const Vector3f& vStart, Vector3f& vEnd);
   
   void computeSampleEpsilon();
   void computePointAreas();
   void computePointNormals();
   void computeLaplace();
   void computeGradientOperator();

   void computeGradient(const DynamicVector<float>& func, DynamicVector<float>& grad);
   void computeDivergence(const DynamicVector<float>& vecField, DynamicVector<float>& div);

   void detectBoundaries();

   void getRingNeighborhood(mpcard sourceIndex, mpcard rings, vector<mpcard>& neighbors);


   FloatType getPointArea(card32 iPoint)
   {
	   if (m_bAreasDirty)
		   return 0.f;
	   return m_pointAreas[iPoint];
   }
   FloatType getPointLaplace(card32 iPoint)
   {
	   if (m_bAreasDirty || m_laplace.getRows() < iPoint)
		   return 0.f;
	   return m_laplace[iPoint][iPoint];
   }
   void getPointLaplace(card32 iPoint, SparseVector<FloatType>& L)
   {
		if (m_bAreasDirty || m_laplace.getRows() < iPoint)
			L.clear();
		L = m_laplace[iPoint];
   }
   
   vector<mpcard> getHarmonicNeighbors(card32 iPoint)
   {
	   if (m_harmonicNeighbors.size() < iPoint)
		   return vector<mpcard>();
	   return m_harmonicNeighbors[iPoint];
   }

   DynamicVector<float> getPointAreas()				{ return m_pointAreas; }
   SparseMatrix<float> getLaplace()					{ return m_laplace; }
   vector<vector<mpcard>> getHarmonicNeighbors()	{ return m_harmonicNeighbors; }

   FloatType getSurfaceArea();

   void findScalarFieldMinimum(const vector<float>& scalarField, const vector<Vector3f>& hintSet, Vector3f& minimalPoint, float& minimalValue);

   virtual ~AR4MLSManifold3f();

   static AR4MLSManifold3f* createFromPointCloud(PointCloud* pc,
												const card32 NumNearestNeighbors,
												const float32 epsilon = 0.1f);
};


}

#endif
