///////////////////////////////////////////////////////////////////////////////
//
//	AR4PartialMatchClustering.cpp
//
//	Source file implementing clustering of partial isometric matches
//
//	Alan Brunton, MPII, March 2013
//
///////////////////////////////////////////////////////////////////////////////


#include "stdafx.h"

#include "AR4PartialMatchClustering.h"
#include "AR4PartialMatch.inline.h"

#include "AR4GDLCluster.h"

#include "AnnSearch.h"

#include <fstream>
#include "AR4Settings.h"

#define AR4PMC_USE_SYM_DIST					0


//file-scope helper functions
static inline float smootherstep(const float x)
{
	const float y = max(0.f, min(1.f, x));
    // Evaluate polynomial
    return y*y*y*(y*(y*6.f - 15.f) + 10.f);
}


//static member variable initialization


//member functions
void X4::AR4PartialMatchClusteringMDS3f::run(vector<AR4PartialMatch3f*>& partialMatches, vector<vector<mpcard>>& clusters)
{
	clusters.clear();

	AR4MLSManifold3f* pmlsSource = dynamic_cast<AR4MLSManifold3f*>(partialMatches.front()->getSourceManifold());
	AR4MLSManifold3f* pmlsTarget = dynamic_cast<AR4MLSManifold3f*>(partialMatches.front()->getTargetManifold());

	if (pmlsSource == NULL || pmlsTarget == NULL)
	{
		throw EInvalidState("AR4PartialMatchClusteringMDS3f::run(...) - source or target manifold not valid");
	}

	mpcard i, j, k, m;
	const mpcard numPointsSource = pmlsSource->getNumDiscrPoints();
	const float rcpNumPointsSource = 1.f / (float)numPointsSource;
	const float fracOverlapIgnore = 0.05f;
	const float fracOverlapFullWeight = 0.2f;
	const float overlapScale = 1.f / (fracOverlapFullWeight - fracOverlapIgnore);

	m_numPartialMatches = partialMatches.size();
	const mpcard numPartial = m_numPartialMatches;
	const mpcard symMatDim = numPartial * (numPartial + 1) / 2;
	const mpcard dim = m_embeddingDim;

	vector<OptType> vecStressGradient(numPartial * dim);
	vector<OptType> vecEmbedCentroid(dim, (OptType)0);

	OptType stress;

	m_symmatDissim.resize(symMatDim);
	m_symmatWeight.resize(symMatDim);
	m_symmatEmDist.resize(symMatDim);
	m_symmatV.resize(symMatDim);
	m_symmatUpdate.resize(symMatDim);

	m_vecEmbeddedPoints.resize(numPartial);

	OptType avgDist = 0.f;

	//compute distances, weight and fill in matrices
	k = 0;
	m = 0;
	for (i = 0; i < numPartial; i++)
	{
		//for grins, validate manifolds here
		if (dynamic_cast<AR4MLSManifold3f*>(partialMatches[i]->getSourceManifold()) != pmlsSource ||
			dynamic_cast<AR4MLSManifold3f*>(partialMatches[i]->getTargetManifold()) != pmlsTarget)
		{
			throw EInvalidState("AR4PartialMatchClusteringMDS3f::run(...) - source or target manifold not consistent");
		}

		//set diagonal elements
		m_symmatDissim[k] = 0;
		m_symmatWeight[k] = 1;
		k++;

		//we can start at i + 1 because the distances and weights are symmetric
		for (j = i + 1; j < numPartial; j++)
		{
#if AR4PMC_USE_SYM_DIST
			AR4SymmetricPartialMatchDistance3f distMeasure(partialMatches[i], partialMatches[j]);

			//get the distance/dissimilarity dij and the weight wij
			const float dij = distMeasure.evaluate();
			const float wij = 1.f;
#else
			AR4PartialMatchDistance3f distMeasure(partialMatches[i], partialMatches[j]);

			//get the distance/dissimilarity dij and the weight wij
			const float dij = distMeasure.evaluate();
			const mpcard cij = distMeasure.getNumSharedSourcePoints();
			const float fracij = (float)cij * rcpNumPointsSource;
			const float wij = smootherstep((fracij - fracOverlapIgnore) * overlapScale);
#endif //AR4PMC_USE_SYM_DIST

			m_symmatDissim[k] = dij;
			m_symmatWeight[k] = wij;

			avgDist += m_symmatDissim[k];

			k++;
		}

		for (j = 0; j < dim; j++)
		{
			m_vecEmbeddedPoints[m] = rnd01();
			m++;
		}
	}

	avgDist += (OptType)symMatDim;

	//compute centroid of initial (random) embedding
	m = 0;
	for (i = 0; i < numPartial; i++)
	{
		for (j = 0; j < dim; j++)
		{
			vecEmbedCentroid[j] += m_vecEmbeddedPoints[m];
			m++;
		}
	}
	for (j = 0; j < dim; j++)
		vecEmbedCentroid[j] /= (OptType)numPartial;

	//subtract centroid so that initial embedding is centred at origin
	m = 0;
	for (i = 0; i < numPartial; i++)
	{
		for (j = 0; j < dim; j++)
		{
			m_vecEmbeddedPoints[m] -= vecEmbedCentroid[j];
			m++;
		}
	}

	//scale embedding (hopefully helps numerically with solver)
	m = 0; 
	for (i = 0; i < numPartial; i++)
	{
		for (j = 0; j < dim; j++)
		{
			m_vecEmbeddedPoints[m] *= avgDist;
			m++;
		}
	}

	//initialize for optimization
	lbfgs_parameter_t optParams;
	lbfgs_parameter_init(&optParams);

	optParams.min_step = 1e-40f;

	//pre-compute V matrix
	computeVMatrix(numPartial, &m_symmatWeight[0], &m_symmatV[0]);

	//run Quasi-Newton optimization
	lbfgs(m_numPartialMatches, &m_vecEmbeddedPoints[0], &stress, lbfgsEvalObjectiveFuncGrad, NULL, (void*)this, &optParams);

	//next cluster the embedded points
	AnnSearch annSearch(0.f);

	//create ANN's stupid pointer-to-pointer array structure
	ANNpointArray pointArray = new ANNpoint[numPartial];
	for (i = 0; i < numPartial; i++)
	{
		pointArray[i] = new ANNcoord[dim];
		for (j = 0; j < dim; j++)
			pointArray[i][j] = m_vecEmbeddedPoints[i * dim + j];
	}

	annSearch.buildSearchStructure(numPartial, dim, pointArray);

	vector<vector<mpcard>> nearestNeighbors(numPartial);
	vector<vector<float>> neighborDistances(numPartial);

	const mpcard numNN = 10;
	float* pNNDists = new float[numNN];
	int32* pNNIdxs = new int32[numNN];
	ANNpoint queryPoint = new ANNcoord[dim];

	const OptType sigma = 10.0;
	const OptType rcp2Sigma2 = 1.0 / (2.0 * sigma*sigma);

	for (i = 0; i < numPartial; i++)
	{
		for (j = 0; j < dim; j++)
			queryPoint[j] = m_vecEmbeddedPoints[i * dim + j];

		annSearch.getkNearestNeighbors(queryPoint, numNN, pNNDists, pNNIdxs);

		nearestNeighbors[i].resize(numNN);
		neighborDistances[i].resize(numNN);
		for (k = 0; k < numNN; k++)
		{
			nearestNeighbors[i][k] = pNNIdxs[k];

			const mpcard idx = sm_lookup(i, pNNIdxs[k], numPartial);
			neighborDistances[i][k] = m_symmatEmDist[idx];
		}
	}

	//deallocate ANN crap
	delete [] pNNDists;
	delete [] pNNIdxs;
	delete [] queryPoint;
	for (i = 0; i < numPartial; i++)
		delete [] pointArray[i];
	delete [] pointArray;

	//now that we have weights we can use graph-based cluster
	//graph degree linkage (GDL) Zhang et al. ECCV 2012
	mpcard numInlierClusters;
	AR4GDLCluster gdl;
	gdl.init(numPartial, numNN, nearestNeighbors, neighborDistances, false);
	gdl.cluster((mpcard)sqrt((float)numPartial), clusters, numInlierClusters);
	
	//next create structures for visualizing embedded space
	//delete any existing point cloud
	delete m_pEmbeddedPC;
	m_pEmbeddedPC = NULL;

	//only visualize if we have embedded in 3-dimensional space
	if (dim == 3)
	{
		//create a point cloud
		UnstructuredInCorePointCloud* pEmbedPC = new UnstructuredInCorePointCloud();

		//vetex descriptor
		VertexDescriptor vd;

		//position, of course
		vd.pushAttrib(VAD("position", 3, VAD::DATA_FORMAT_FLOAT32));

		//associate each point with specific partial match via attribute
		vd.pushAttrib(VAD("partialMatchIndex", 1, VAD::DATA_FORMAT_CARD32));
		
		//add attribute for size of partial match, as well as maybe some measure of stretch (avg, max, whatever)
		vd.pushAttrib(VAD("matchSize", 1, VAD::DATA_FORMAT_FLOAT32));
		vd.pushAttrib(VAD("totalStretch", 1, VAD::DATA_FORMAT_FLOAT32));

		pEmbedPC->clearAndSetup(&vd, numPartial);

		AAT pos, pmi, ms, ts;
		pos = pEmbedPC->getAAT("position");
		pmi = pEmbedPC->getAAT("partialMatchIndex");
		ms = pEmbedPC->getAAT("matchSize");
		ts = pEmbedPC->getAAT("totalStretch");

		PointSet* ps = pEmbedPC->getPointSet();

		//fill in point attributes
		for (i = 0; i < numPartial; i++)
		{
			float32 matchSize = (float32)partialMatches[i]->getNumMatchedIndices();
			float32 totalstretch = (float32)partialMatches[i]->getTotalStretch() / matchSize;

			ps->set3f(i, pos, makeVector3f(m_vecEmbeddedPoints[i * 3 + 0], m_vecEmbeddedPoints[i * 3 + 1], m_vecEmbeddedPoints[i * 3 + 2]));
			ps->set1u(i, pmi, (card32)i);
			ps->set1f(i, ms, matchSize);
			ps->set1f(i, ts, totalstretch);
		}

		//add point cloud to the scene somehow... maybe have to do outside this class
		//so we have to save the point cloud as a member variable that can be accessed via a getter
		m_pEmbeddedPC = pEmbedPC;
	}
}

void X4::AR4PartialMatchClusteringMDS3f::run(AR4PartialMatch3f* pRefMatch, vector<GenericSwapAllOutOfCorePointCloud*>& partialMatches, vector<vector<mpcard>>& clusters)
{
	clusters.clear();
	
	AR4MLSManifold3f* pmlsSource = dynamic_cast<AR4MLSManifold3f*>(pRefMatch->getSourceManifold());
	AR4MLSManifold3f* pmlsTarget = dynamic_cast<AR4MLSManifold3f*>(pRefMatch->getTargetManifold());

	if (pmlsSource == NULL || pmlsTarget == NULL)
	{
		throw EInvalidState("AR4PartialMatchClusteringMDS3f::run(...) - source or target manifold not valid");
	}

	mpcard i, j, k, m;
	const mpcard numPointsSource = pRefMatch->getNumSourcePointsMatchingLevel();
	
	const float rcpNumPointsSource = 1.f / (float)numPointsSource;

	m_numPartialMatches = partialMatches.size();
	const mpcard numPartial = m_numPartialMatches;
	const mpcard symMatDim = numPartial * (numPartial + 1) / 2;
	const mpcard dim = m_embeddingDim;

	vector<float32> matchSizes(numPartial);
	vector<float32> totalStretches(numPartial);

	vector<OptType> vecStressGradient(numPartial * dim);
	vector<OptType> vecEmbedCentroid(dim, (OptType)0);

	OptType stress;

	m_symmatDissim.resize(symMatDim);
	m_symmatWeight.resize(symMatDim);
	m_symmatEmDist.resize(symMatDim);
	m_symmatV.resize(symMatDim);
	m_symmatUpdate.resize(symMatDim);

	m_vecEmbeddedPoints.resize(numPartial * dim);

	float avgMatchSize = 0.f;
	for (i = 0; i < numPartial; i++)
	{
		UnstructuredInCorePointCloud* pUPC = dynamic_cast<UnstructuredInCorePointCloud*>(partialMatches[i]->fetchPC());
		avgMatchSize += (float)pUPC->getNumPoints();
	}
	avgMatchSize /= (float)numPartial;

	const float fracOverlapFullWeight = avgMatchSize * rcpNumPointsSource;
	const float fracOverlapIgnore = 0.f;
	const float overlapScale = 1.f / (fracOverlapFullWeight - fracOverlapIgnore);

	const OptType largeValueIgnore = (OptType)1e30;

	OptType avgDist = (OptType)0;

	vector<vector<mpcard>> nearestNeighbors(numPartial);
	vector<vector<float>> neighborDistances(numPartial);

	const mpcard numNN = m_settings->clusteringNumNN;
	int qnret = -1;


	vector<float32> dissimMatrix(numPartial*numPartial);
	vector<float32> weightMatrix(numPartial*numPartial);

	vector<float32> OverlapMatrix(numPartial*numPartial);
	vector<float32> distancePartialMatrix(numPartial*numPartial);

	debugOutput << "\tcomputing distances\n";
	X4_TIMER_START(partialMatch_distances);
	//compute distances, weight and fill in matrices
	k = 0;
	m = 0;
	
	for (i = 0; i < numPartial; i++)
	{

		AR4PartialMatch3f* pPartial1 = new AR4PartialMatch3f(pmlsSource, pmlsTarget);
		pPartial1->copyGeodesicAndTopologyData(pRefMatch);
		pPartial1->setGeodesicsSigma(pRefMatch->getGeodesicsSigma());

		UnstructuredInCorePointCloud* pUPC1 = dynamic_cast<UnstructuredInCorePointCloud*>(partialMatches[i]->fetchPC());
		pPartial1->initThroughPointCloud(pUPC1);

		matchSizes[i] = (float32)pPartial1->getNumMatchedIndices();
		totalStretches[i] = (float32)pPartial1->getTotalStretch() / matchSizes[i];

		//set diagonal elements
		dissimMatrix[i * numPartial + i] = 0.f;
		weightMatrix[i * numPartial + i] = 1.f;


		OverlapMatrix[i * numPartial + i] = matchSizes[i];
		

		k++;

		//we can start at i + 1 because the distances and weights are symmetric
		for (j = i + 1; j < numPartial; j++)
		{
			AR4PartialMatch3f* pPartial2 = new AR4PartialMatch3f(pmlsSource, pmlsTarget);
			pPartial2->copyGeodesicAndTopologyData(pRefMatch);
			pPartial2->setGeodesicsSigma(pRefMatch->getGeodesicsSigma());

			UnstructuredInCorePointCloud* pUPC2 = dynamic_cast<UnstructuredInCorePointCloud*>(partialMatches[j]->fetchPC());
			pPartial2->initThroughPointCloud(pUPC2);

#if AR4PMC_USE_SYM_DIST
			AR4SymmetricPartialMatchDistance3f distMeasure(pPartial1, pPartial2);

			//get the distance/dissimilarity dij and the weight wij
			const float dij = distMeasure.evaluate();

			dissimMatrix[i * numPartial + j] = dij;
			dissimMatrix[j * numPartial + i] = dij;

			weightMatrix[i * numPartial + j] = 1.f;
			weightMatrix[j * numPartial + i] = 1.f;
#else
			AR4PartialMatchDistance3f distMeasure(pPartial1, pPartial2);

			//get the distance/dissimilarity dij and the weight wij
			const float dij = distMeasure.evaluate();
			const float cij = distMeasure.getOverlap();
			const float fracij = cij * rcpNumPointsSource;

		
			const float wij = smootherstep((fracij - fracOverlapIgnore) * overlapScale);
			
			OverlapMatrix[i * numPartial + j] = cij;
			OverlapMatrix[j * numPartial + i] = cij;
			distancePartialMatrix[i * numPartial + j] = dij;			
			distancePartialMatrix[j * numPartial + i] = dij;
			
			
			weightMatrix[i * numPartial + j] = wij;
			weightMatrix[j * numPartial + i] = wij;

			if (wij > 0.f)
			{
				dissimMatrix[i * numPartial + j] = dij;
				dissimMatrix[j * numPartial + i] = dij;
				
			}
			else
			{
				dissimMatrix[i * numPartial + j] = -1.f;
				dissimMatrix[j * numPartial + i] = -1.f;
			}
#endif //AR4PMC_USE_SYM_DIST

			k++;

			delete pPartial2;
		}

		delete pPartial1;
	}
	X4_TIMER_STOP(partialMatch_distances);

	debugOutput << "\tcomputing nearest neighbors\n";
	for (i = 0; i < numPartial; i++)
	{
		vector<pair<OptType, mpcard>> distIndices(numPartial);
		for (j = 0; j < numPartial; j++)
		{
			distIndices[j].second = j;
			if (dissimMatrix[i * numPartial + j] >= 0.f)
				distIndices[j].first = dissimMatrix[i * numPartial + j];
			else
				distIndices[j].first = largeValueIgnore * 1.5f;
		}

		sort(distIndices.begin(), distIndices.end());

		for (j = 0; j < numNN; j++)
		{
			if (distIndices[j + 1].first < largeValueIgnore)
			{
				nearestNeighbors[i].push_back(distIndices[j + 1].second);
				neighborDistances[i].push_back((float32)distIndices[j + 1].first);
			}
		}

	}

	const float32 pmSigma = pRefMatch->getStretchThreshold() * pRefMatch->getAvgSourceMatchLevelEdgeWeight();

	debugOutput << "\tclustering with GDL\n";
	//now that we have weights we can use graph-based cluster
	//graph degree linkage (GDL) Zhang et al. ECCV 2012
	mpcard numInlierClusters;
	AR4GDLCluster gdl;
	debugOutput << "\tpartial distance threshold: " << m_partialDistanceThreshold << "\n";
	X4_TIMER_START(gdl_cluster);
	gdl.setAffinityThreshold(m_partialDistanceThreshold);
	gdl.init(numPartial, numNN, nearestNeighbors, neighborDistances, weightMatrix, false);
	gdl.cluster(max(m_nMinClusters, (mpcard)sqrt((float)numPartial)), clusters, numInlierClusters);
	X4_TIMER_STOP(gdl_cluster);
	
	//next create structures for visualizing embedded space
	//delete any existing point cloud
	delete m_pEmbeddedPC;
	m_pEmbeddedPC = NULL;

	//only visualize if we have embedded in 3-dimensional space
	if (dim == 3 && qnret >= 0)
	{
		debugOutput << "\tcreating embedding point cloud\n";

		//create a point cloud
		UnstructuredInCorePointCloud* pEmbedPC = new UnstructuredInCorePointCloud();

		//vetex descriptor
		VertexDescriptor vd;

		//position, of course
		vd.pushAttrib(VAD("position", 3, VAD::DATA_FORMAT_FLOAT32));

		//associate each point with specific partial match via attribute
		vd.pushAttrib(VAD("partialMatchIndex", 1, VAD::DATA_FORMAT_CARD32));
		
		//add attribute for size of partial match, as well as maybe some measure of stretch (avg, max, whatever)
		vd.pushAttrib(VAD("matchSize", 1, VAD::DATA_FORMAT_FLOAT32));
		vd.pushAttrib(VAD("totalStretch", 1, VAD::DATA_FORMAT_FLOAT32));

		pEmbedPC->clearAndSetup(&vd, numPartial);

		AAT pos, pmi, ms, ts;
		pos = pEmbedPC->getAAT("position");
		pmi = pEmbedPC->getAAT("partialMatchIndex");
		ms = pEmbedPC->getAAT("matchSize");
		ts = pEmbedPC->getAAT("totalStretch");

		PointSet* ps = pEmbedPC->getPointSet();

		//fill in point attributes
		for (i = 0; i < numPartial; i++)
		{
			const float32 matchSize = matchSizes[i];
			const float32 totalstretch = totalStretches[i];

			ps->set3f(i, pos, makeVector3f(m_vecEmbeddedPoints[i * 3 + 0], m_vecEmbeddedPoints[i * 3 + 1], m_vecEmbeddedPoints[i * 3 + 2]));
			ps->set1u(i, pmi, (card32)i);
			ps->set1f(i, ms, matchSize);
			ps->set1f(i, ts, totalstretch);
		}

		//add point cloud to the scene somehow... maybe have to do outside this class
		//so we have to save the point cloud as a member variable that can be accessed via a getter
		m_pEmbeddedPC = pEmbedPC;
	}


}


void X4::AR4PartialMatchClusteringAP3f::run(AR4PartialMatch3f* pRefMatch, vector<GenericSwapAllOutOfCorePointCloud*>& partialMatches, vector<vector<mpcard>>& clusters)
{
	clusters.clear();

	AR4MLSManifold3f* pmlsSource = dynamic_cast<AR4MLSManifold3f*>(pRefMatch->getSourceManifold());
	AR4MLSManifold3f* pmlsTarget = dynamic_cast<AR4MLSManifold3f*>(pRefMatch->getTargetManifold());

	if (pmlsSource == NULL || pmlsTarget == NULL)
	{
		throw EInvalidState("AR4PartialMatchClusteringMDS3f::run(...) - source or target manifold not valid");
	}

	const mpcard numPartial = partialMatches.size();
	const mpcard numPointsSource = pRefMatch->getNumSourcePointsMatchingLevel();//pmlsSource->getNumDiscrPoints();
	const float rcpNumPointsSource = 1.f / (float)numPointsSource;

	mpcard i, j, k;

	float avgMatchSize = 0.f;
	for (i = 0; i < numPartial; i++)
	{
		UnstructuredInCorePointCloud* pUPC = dynamic_cast<UnstructuredInCorePointCloud*>(partialMatches[i]->fetchPC());
		avgMatchSize += (float)pUPC->getNumPoints();
	}
	avgMatchSize /= (float)numPartial;

	const float ignoreValue = -1e30f;
	const float simThresh = ignoreValue * 0.5f;

	const float fracOverlapFullWeight = avgMatchSize * rcpNumPointsSource;
	const float fracOverlapIgnore = 0.1f * fracOverlapFullWeight;
	const float overlapScale = 1.f / (fracOverlapFullWeight - fracOverlapIgnore);

	vector<float> similarities(numPartial * numPartial);
	vector<float> preferences(numPartial);

	X4_TIMER_START(partialMatch_distances);
	k = 0;
	for (i = 0; i < numPartial; i++)
	{
		AR4PartialMatch3f* pPartial1 = new AR4PartialMatch3f(pmlsSource, pmlsTarget);
		pPartial1->copyGeodesicAndTopologyData(pRefMatch);

		UnstructuredInCorePointCloud* pUPC1 = dynamic_cast<UnstructuredInCorePointCloud*>(partialMatches[i]->fetchPC());
		pPartial1->initThroughPointCloud(pUPC1);

		//we can start at i + 1 because the distances and weights are symmetric
		for (j = 0; j < numPartial; j++, k++)
		{
			if (j == i)
			{
				similarities[k] = 0.f;
				continue;
			}

			AR4PartialMatch3f* pPartial2 = new AR4PartialMatch3f(pmlsSource, pmlsTarget);
			pPartial2->copyGeodesicAndTopologyData(pRefMatch);

			UnstructuredInCorePointCloud* pUPC2 = dynamic_cast<UnstructuredInCorePointCloud*>(partialMatches[j]->fetchPC());
			pPartial2->initThroughPointCloud(pUPC2);

#if AR4PMC_USE_SYM_DIST
			AR4SymmetricPartialMatchDistance3f distMeasure(pPartial1, pPartial2);

			//get the distance/dissimilarity dij and the weight wij
			const float dij = distMeasure.evaluate();

			similarities[k] = -dij;
#else
			AR4PartialMatchDistance3f distMeasure(pPartial1, pPartial2);

			//get the distance/dissimilarity dij and the weight wij
			const float dij = distMeasure.evaluate();
			const float cij = distMeasure.getOverlap();
			const float fracij = cij * rcpNumPointsSource;
			const float wij = smootherstep((fracij - fracOverlapIgnore) * overlapScale);
			
			
			if (cij > 0.f)
				similarities[k] = -dij;
			else
				similarities[k] = ignoreValue;
#endif //AR4PMC_USE_SYM_DIST

			delete pPartial2;
		}

		vector<float> orderedSimilarities(numPartial);
		std::copy(similarities.begin() + i * numPartial, similarities.begin() + (i + 1) * numPartial, orderedSimilarities.begin());

		sort(orderedSimilarities.begin(), orderedSimilarities.end());

		preferences[i] = orderedSimilarities[numPartial / 2];

		delete pPartial1;
	}
	X4_TIMER_STOP(partialMatch_distances);

	ofstream simFile("similarities.txt");
	ofstream prefFile("preferences.txt");

	k = 0;
	for (i = 0; i < numPartial; i++)
	{
		prefFile << preferences[i] << "\n";

		for (j = 0; j < numPartial; j++)
		{
			if (j != i && similarities[k] > simThresh)
				simFile << i + 1 << "\t" << j + 1 << "\t" << similarities[k] << "\n";

			k++;
		}
	}

	simFile.close();
	prefFile.close();

	X4_TIMER_START(apcluster);
	system("apcluster.exe similarities.txt preferences.txt apout");
	X4_TIMER_STOP(apcluster);

	//read back idx file

	ifstream idxFile("apout/idx.txt");

	map<mpcard, mpcard> exemplarLU;

	for (i = 0; i < numPartial && idxFile.good(); i++)
	{
		mpcard exemplar;

		idxFile >> exemplar;

		if (exemplarLU.find(exemplar) == exemplarLU.end())
		{
			exemplarLU[exemplar] = clusters.size();
			vector<mpcard> clusterNew(2);
			clusterNew[0] = exemplar - 1;
			clusterNew[1] = i;
			clusters.push_back(clusterNew);
		}
		else
		{
			mpcard icluster = exemplarLU[exemplar];
			clusters[icluster].push_back(i);
		}
	}

	idxFile.close();
}


void X4::AR4PartialMatchDirectCorrClustering3f::run(AR4PartialMatch3f* pRefMatch, vector<GenericSwapAllOutOfCorePointCloud*>& partialMatches, vector<vector<mpcard>>& clusters)
{
	const mpcard numPartials = partialMatches.size();

	AR4MLSManifold3f* pSource = dynamic_cast<AR4MLSManifold3f*>(pRefMatch->getSourceManifold());
	AR4MLSManifold3f* pTarget = dynamic_cast<AR4MLSManifold3f*>(pRefMatch->getTargetManifold());

	AR4SparseGeodesics* pSourceGeos = pRefMatch->getSparseSourceGeodesics();
	AR4SparseGeodesics* pTargetGeos = pRefMatch->getSparseTargetGeodesics();

	map<mpcard, vector<mpcard>> sourceIndexToCorrs;
	map<mpcard, vector<mpcard>>::iterator sicIt;

	m_correspondences.clear();

	mpcard i, j, k;

	for (i = 0; i < numPartials; i++)
	{
		UnstructuredInCorePointCloud* pMatchPC = dynamic_cast<UnstructuredInCorePointCloud*>(partialMatches[i]->fetchPC());

		AAT idxAAT = pMatchPC->getAAT(AR4PartialMatch3f::getMatchAttNameIndex());
		AAT trgAAT = pMatchPC->getAAT(AR4PartialMatch3f::getMatchAttNameTargetPos());

		const mpcard numMatches = pMatchPC->getNumPoints();

		for (j = 0; j < numMatches; j++)
		{
			const card32 index = pMatchPC->getPointSet()->get1u(j, idxAAT);
			const Vector3f targPos = pMatchPC->getPointSet()->get3f(j, trgAAT);

			const mpcard corrIdx = m_correspondences.size();

			PointCorr corr(index, targPos);

			sicIt = sourceIndexToCorrs.find(index);
			if (sicIt != sourceIndexToCorrs.end())
				sicIt->second.push_back(corrIdx);
			else
				sourceIndexToCorrs[index] = vector<mpcard>(1, corrIdx);

			m_correspondences.push_back(corr);
		}
	}

	const mpcard numCorr = m_correspondences.size();

	vector<vector<mpcard>> nearestNeighbors(numCorr);
	vector<vector<float32>> neighborDistances(numCorr);
	vector<vector<card32>> incompatibility(numCorr);

	float avgNbrs = 0.f;
	for (i = 0; i < numCorr; i++)
	{
		const mpcard sourceIndex = m_correspondences[i].m_sourceIndex;

		//collect source geodesic information
		vector<mpcard> sourceDests;
		vector<float> sourceDists;
		vector<Vector3f> sourceGrads;
		pSourceGeos->getGeodesicData(sourceIndex, sourceDests, sourceDists, sourceGrads);

		const mpcard numDests = sourceDests.size();
		for (j = 0; j < numDests; j++)
		{
			const float32 refDist = sourceDists[j];
			const mpcard sdest = sourceDests[j];
			if (sourceIndexToCorrs.find(sdest) != sourceIndexToCorrs.end())
			{
				const mpcard numSharing = sourceIndexToCorrs[sdest].size();
				for (k = 0; k < numSharing; k++)
				{
					const mpcard icorr = sourceIndexToCorrs[sdest][k];
					PointCorr corr = m_correspondences[icorr];

					float targetDist;
					Vector3f targetGrad;
					pTargetGeos->computeGeodesicSingleDest(pRefMatch->getGeodesicsSigma() * 0.1f, m_correspondences[i].m_targetPoint, corr.m_targetPoint, targetDist, targetGrad);

					const float distDiff = targetDist - refDist;
					const float corrDist = fabs(distDiff);

					nearestNeighbors[i].push_back(icorr);
					neighborDistances[i].push_back(corrDist);
				}
			}
		}

		//exclude matches with the same source index
		const mpcard numSharing = sourceIndexToCorrs[sourceIndex].size();//number of correspondences sharing this source index
		for (j = 0; j < numSharing; j++)
		{
			const mpcard icorr = sourceIndexToCorrs[sourceIndex][j];
			if (icorr != i)
				incompatibility[i].push_back(icorr);
		}

		avgNbrs += (float)nearestNeighbors[i].size();
	}

	avgNbrs /= (float)numCorr;
	const mpcard numNN = (mpcard)avgNbrs;
	
	mpcard numInlierClusters;

	AR4GDLCluster gdl;
	X4_TIMER_START(GDL_INIT);
	gdl.init(numCorr, numNN, nearestNeighbors, neighborDistances, incompatibility);
	X4_TIMER_STOP(GDL_INIT);
	
	X4_TIMER_START(GDL_CLUSTER);
	gdl.cluster(max(m_nMinClusters, (mpcard)sqrt((float)numCorr)), clusters, numInlierClusters);
	X4_TIMER_STOP(GDL_CLUSTER);
}

void X4::AR4PartialMatchSpectralClustering3f::run(X4::AR4PartialMatch3f *pRefMatch, std::vector<GenericSwapAllOutOfCorePointCloud*> &partialMatches, std::vector<vector<mpcard> > &clusters)
{
	clusters.clear();

	AR4MLSManifold3f* pmlsSource = dynamic_cast<AR4MLSManifold3f*>(pRefMatch->getSourceManifold());
	AR4MLSManifold3f* pmlsTarget = dynamic_cast<AR4MLSManifold3f*>(pRefMatch->getTargetManifold());

	if (pmlsSource == NULL || pmlsTarget == NULL)
	{
		throw EInvalidState("AR4PartialMatchClusteringMDS3f::run(...) - source or target manifold not valid");
	}

	mpcard i, j, k, m;
	const mpcard numPointsSource = pRefMatch->getNumSourcePointsMatchingLevel();//pmlsSource->getNumDiscrPoints();
	const float rcpNumPointsSource = 1.f / (float)numPointsSource;

	const mpcard numPartials = partialMatches.size();
	const mpcard matSize = numPartials*numPartials;

	vector<float32> matchSizes(numPartials);
	vector<float32> totalStretches(numPartials);

	vector<float32> matDistances(matSize);
	SparseMatrix<float32> matAffinity;

	matAffinity.setRows(numPartials);

	float avgMatchSize = 0.f;
	for (i = 0; i < numPartials; i++)
	{
		UnstructuredInCorePointCloud* pUPC = dynamic_cast<UnstructuredInCorePointCloud*>(partialMatches[i]->fetchPC());
		avgMatchSize += (float)pUPC->getNumPoints();
	}
	avgMatchSize /= (float)numPartials;

	const float fracOverlapFullWeight = avgMatchSize * rcpNumPointsSource;
	const float fracOverlapIgnore = 0.1f * fracOverlapFullWeight;
	const float overlapScale = 1.f / (fracOverlapFullWeight - fracOverlapIgnore);

	float avgDist = 0.f;

	X4_TIMER_START(partialMatchDistances);
	//compute distances, weight and fill in matrices
	for (i = 0; i < numPartials; i++)
	{
		AR4PartialMatch3f* pPartial1 = new AR4PartialMatch3f(pmlsSource, pmlsTarget);
		pPartial1->copyGeodesicAndTopologyData(pRefMatch);
		pPartial1->setGeodesicsSigma(pRefMatch->getGeodesicsSigma());

		UnstructuredInCorePointCloud* pUPC1 = dynamic_cast<UnstructuredInCorePointCloud*>(partialMatches[i]->fetchPC());
		pPartial1->initThroughPointCloud(pUPC1);

		matchSizes[i] = (float32)pPartial1->getNumMatchedIndices();
		totalStretches[i] = (float32)pPartial1->getTotalStretch() / matchSizes[i];

		matDistances[i * numPartials + i] = 0.f;

		//we can start at i + 1 because the distances and weights are symmetric
		for (j = i + 1; j < numPartials; j++)
		{
			AR4PartialMatch3f* pPartial2 = new AR4PartialMatch3f(pmlsSource, pmlsTarget);
			pPartial2->copyGeodesicAndTopologyData(pRefMatch);
			pPartial2->setGeodesicsSigma(pRefMatch->getGeodesicsSigma());

			UnstructuredInCorePointCloud* pUPC2 = dynamic_cast<UnstructuredInCorePointCloud*>(partialMatches[j]->fetchPC());
			pPartial2->initThroughPointCloud(pUPC2);

			AR4SymmetricPartialMatchDistance3f distMeasure(pPartial1, pPartial2);

			//get the distance/dissimilarity dij and the weight wij
			const float dij = distMeasure.evaluate();

			matDistances[i * numPartials + j] = dij;
			matDistances[j * numPartials + i] = dij;

			avgDist += 2.f * dij;

			delete pPartial2;
		}

		delete pPartial1;
	}
	X4_TIMER_STOP(partialMatchDistances);

	avgDist /= (float)matSize;

	const float sigma2 = avgDist;
	const float rcpSigma2 = 1.f / sigma2;
	const float smallThresh = exp(-2.f);//2 standard deviations

	mpcard numAffinityEntries = 0;
	for (i = 0; i < numPartials; i++)
	{
		for (j = 0; j < numPartials; j++)
		{
			const float dist = matDistances[i * numPartials + j];
			const float affinity = exp(-dist * rcpSigma2);

			if (affinity > smallThresh)
			{
				matAffinity[i].setEntry(j, affinity);
				numAffinityEntries++;
			}
		}
	}

	debugOutput << "number of non-zero entries in affinity matrix: " << numAffinityEntries << "\n";
}
