///////////////////////////////////////////////////////////////////////////////
//
//	AR4PartialMatchClustering.h
//
//	Header file for implementation of clustering of partial isometric matches
//
//	Alan Brunton, MPII, March 2013
//
///////////////////////////////////////////////////////////////////////////////


#ifndef __AR4PARTIALMATCHCLUSTERING_H__
#define __AR4PARTIALMATCHCLUSTERING_H__


#include "threeParameterMatchingGMOD14.h"
#include "AR4PartialMatch.h"

#include "AR4Settings.h"

#include "GenericSwapAllOutOfCorePointCloud.h"

#include "lbfgs/lbfgs.h"


namespace X4
{

template<class ManifoldTraits>
class AR4PartialMatchClustering
{
public:
	typedef typename ManifoldTraits::FloatType  FloatType;
	typedef typename ManifoldTraits::VectorType VectorType;
	typedef typename ManifoldTraits AR4DiscreteManifoldTraits;
	typedef typename ManifoldTraits::TangentSpace TangentSpace;
	typedef typename ManifoldTraits::BoundingBoxType BoundingBoxType;

protected:

	mpcard								m_nMinClusters;
	FloatType							m_partialDistanceThreshold;

	const AR4Settings*					m_settings;

public:

	AR4PartialMatchClustering(const AR4Settings* settings): m_settings(settings)
	{
		m_nMinClusters = 0;
		m_partialDistanceThreshold = 0.f;
	}
	~AR4PartialMatchClustering()
	{
	}

	void setNumMinClusters(const mpcard minClusters)		{ m_nMinClusters = minClusters; }
	mpcard getNumMinClusters()								{ return m_nMinClusters; }

	void setPartialDistThresh(const FloatType distThresh)	{ m_partialDistanceThreshold = distThresh; }
	FloatType getPartialDistThresh()						{ return m_partialDistanceThreshold; }

	virtual void run(vector<AR4PartialMatch<ManifoldTraits>*>& partialMatches, vector<vector<mpcard>>& clusters) = 0;
	virtual void run(AR4PartialMatch3f* pRefMatch, vector<GenericSwapAllOutOfCorePointCloud*>& partialMatches, vector<vector<mpcard>>& clusters) = 0;
};


template<class ManifoldTraits>
class AR4PartialMatchClusteringMDS : public AR4PartialMatchClustering<ManifoldTraits>
{
protected:

	typedef lbfgsfloatval_t OptType;

	mpcard									m_numPartialMatches;
	mpcard									m_embeddingDim;

	vector<OptType>							m_symmatDissim;
	vector<OptType>							m_symmatWeight;
	vector<OptType>							m_symmatEmDist;
	vector<OptType>							m_symmatV;
	vector<OptType>							m_symmatUpdate;

	vector<OptType>							m_vecEmbeddedPoints;

	UnstructuredInCorePointCloud*			m_pEmbeddedPC;


	static inline mpcard sm_diagonal(const mpcard row, const mpcard n)
	{
		return row * n - ((row * (row + 1)) >> 1) + row;
	}
	static inline mpcard sm_lookup(const mpcard row, const mpcard col, const mpcard n)
	{
		if (col >= row)
			return sm_ut_lookup(row, col, n);
		return sm_lt_lookup(row, col, n);
	}
	static inline mpcard sm_ut_lookup(const mpcard row, const mpcard col, const mpcard n)
	{
		const mpcard diff = col - row;
		return sm_diagonal(row, n) + diff;
	}
	static inline mpcard sm_lt_lookup(const mpcard row, const mpcard col, const mpcard n)
	{
		const mpcard diff = row - col;
		return sm_diagonal(row, n) - diff * n + ((diff * (row + col - 1)) >> 1) + diff;
	}

	static inline void computeEmbedDistances(const mpcard numPoints, const mpcard dim, const OptType* pPoints, OptType* pDistances)
	{
		const mpcard numDistances = numPoints * (numPoints + 1) / 2;
		mpcard i, j, k, d;
		float dist, diff;
		k = 0;
		for (i = 0; i < numPoints; i++)
		{
			const mpcard offset_i = i * dim;

			pDistances[k] = 0;
			k++;

			for (j = i + 1; j < numPoints; j++)
			{
				const mpcard offset_j = j * dim;

				dist = 0.f;
				for (d = 0; d < dim; d++)
				{
					diff = pPoints[offset_i + d] - pPoints[offset_j + d];
					dist += diff*diff;
				}
				pDistances[k] = sqrt(dist);
				k++;
			}
		}
	}

	static inline OptType computeEmbedStress(const mpcard numDistances, const OptType* pEmbedDists, const OptType* pDissim, const OptType* pWeights)
	{
		mpcard i;
		OptType diff;
		OptType stress = (OptType)0;
		for (i = 0; i < numDistances; i++)
		{
			diff = pEmbedDists[i] - pDissim[i];
			stress += pWeights[i] * diff*diff;
		}
		return stress;
	}

	static inline void computeVMatrix(const mpcard numPoints, const OptType* pWeights, OptType* pVMatrix)
	{
		mpcard i, j, k, kd, kc;
		k = 0;
		for (i = 0; i < numPoints; i++)
		{
			kd = k;
			pVMatrix[kd] = (OptType)0;
			for (j = i + 1; j < numPoints; j++)
			{
				pVMatrix[k] = -pWeights[k];
				pVMatrix[kd] += pVMatrix[k];
				k++;
			}
			kc = kd;
			for (j = 0; j < i; j++)
			{
				kc -= numPoints - i + j;
				pVMatrix[kd] += pVMatrix[kc];
			}
		}
	}

	static inline void computeUpdateMatrix(const mpcard numPoints, const OptType* pDissim, const OptType* pDistances, const OptType* pWeights, OptType* pUpdateMatrix)
	{
		const OptType divide_non_zero = 1e-8;

		mpcard i, j, k, kc;
		k = 0;
		for (i = 0; i < numPoints; i++)
		{
			const mpcard kd = k;

			pUpdateMatrix[kd] = (OptType)0;
			k++;

			for (j = i + 1; j < numPoints; j++)
			{
				if (pDistances[k] > divide_non_zero)
					pUpdateMatrix[k] = -pWeights[k] * pDissim[k] / pDistances[k];
				else
					pUpdateMatrix[k] = (OptType)0;

				pUpdateMatrix[kd] -= pUpdateMatrix[k];

				k++;
			}

			kc = kd;
			for (j = 0; j < i; j++)
			{
				kc -= numPoints - i + j;
				pUpdateMatrix[kd] -= pUpdateMatrix[kc];
			}
		}
	}

	static inline void computeEmbedStressGradient(const mpcard numPoints, const mpcard dim, const OptType* pPoints, const OptType* pVMatrix, OptType* pUpdateMatrix, OptType* pGradient)
	{
		const mpcard numDist = numPoints * (numPoints + 1) / 2;

		mpcard i, j, k, m;

		for (i = 0; i < numDist; i++)
			pUpdateMatrix[i] = (OptType)2 * (pVMatrix[i] - pUpdateMatrix[i]);

		m = 0;
		for (i = 0; i < numPoints; i++)
		{
			const mpcard offset_i = i * dim;
			const mpcard md = m;

			for (j = 0; j < dim; j++)
				pGradient[offset_i + j] = (OptType)0;

			for (j = i + 1; j < numPoints; j++)
			{
				const mpcard offset_j = j * dim;
				for (k = 0; k < dim; k++)
					pGradient[offset_i + k] += pUpdateMatrix[m] * pPoints[offset_j + k];

				m++;
			}

			mpcard mc = md;
			for (j = 0; j < i; j++)
			{
				const mpcard offset_j = j * dim;
				mc -= numPoints - i + j;
				for (k = 0; k < dim; k++)
					pGradient[offset_i + k] += pUpdateMatrix[mc] * pPoints[offset_j + k];
			}
		}
	}


	static lbfgsfloatval_t lbfgsEvalObjectiveFuncGrad(void* instance, const lbfgsfloatval_t *x, lbfgsfloatval_t *g, const int n, const lbfgsfloatval_t step)
	{
		AR4PartialMatchClusteringMDS* pInst = (AR4PartialMatchClusteringMDS*)instance;

		lbfgsfloatval_t stress;

		const mpcard numPoints = pInst->m_numPartialMatches;
		const mpcard numDistances = numPoints * (numPoints + 1) / 2;
		const mpcard dim = pInst->m_embeddingDim;

		lbfgsfloatval_t* pDissim = &(pInst->m_symmatDissim[0]);
		lbfgsfloatval_t* pWeight = &(pInst->m_symmatWeight[0]);
		lbfgsfloatval_t* pEmDist = &(pInst->m_symmatEmDist[0]);
		lbfgsfloatval_t* pVMatrix = &(pInst->m_symmatV[0]);
		lbfgsfloatval_t* pUpdateMatrix = &(pInst->m_symmatUpdate[0]);

		//compute distances in embedding space and then stress
		computeEmbedDistances(numPoints, dim, x, pEmDist);
		stress = computeEmbedStress(numDistances, pEmDist, pDissim, pWeight);

		//compute the update matrix
		computeUpdateMatrix(numPoints, pDissim, pEmDist, pWeight, pUpdateMatrix);

		//compute the gradient
		computeEmbedStressGradient(numPoints, dim, x, pVMatrix, pUpdateMatrix, g);

		return stress;
	}


public:

	AR4PartialMatchClusteringMDS(const AR4Settings* settings, mpcard embeddingDim = 3): AR4PartialMatchClustering(settings)
	{
		m_numPartialMatches = 1;
		m_embeddingDim = embeddingDim;
		m_pEmbeddedPC = NULL;
	}

	~AR4PartialMatchClusteringMDS()
	{
	}

	UnstructuredInCorePointCloud* getEmbeddedPointCloud()		{ return m_pEmbeddedPC; }

	virtual void run(vector<AR4PartialMatch<ManifoldTraits>*>& partialMatches, vector<vector<mpcard>>& clusters) { }
	virtual void run(AR4PartialMatch3f* pRefMatch, vector<GenericSwapAllOutOfCorePointCloud*>& partialMatches, vector<vector<mpcard>>& clusters) { }
};


class AR4PartialMatchClusteringMDS3f : public AR4PartialMatchClusteringMDS<ManifoldTraits3f>
{
public:
	typedef AR4PartialMatch3f PartialMatch;

protected:

public:

	AR4PartialMatchClusteringMDS3f(const AR4Settings* settings, mpcard embeddingDim = 3): AR4PartialMatchClusteringMDS(settings, embeddingDim)
	{
	}

	void run(vector<AR4PartialMatch3f*>& partialMatches, vector<vector<mpcard>>& clusters);
	void run(AR4PartialMatch3f* pRefMatch, vector<GenericSwapAllOutOfCorePointCloud*>& partialMatches, vector<vector<mpcard>>& clusters);
};


template<class ManifoldTraits>
class AR4PartialMatchClusteringAP : public AR4PartialMatchClustering<ManifoldTraits>
{
public:

	AR4PartialMatchClusteringAP(const AR4Settings* settings): AR4PartialMatchClustering(settings)
	{
	}

	virtual void run(vector<AR4PartialMatch<ManifoldTraits>*>& partialMatches, vector<vector<mpcard>>& clusters) { }
	virtual void run(AR4PartialMatch3f* pRefMatch, vector<GenericSwapAllOutOfCorePointCloud*>& partialMatches, vector<vector<mpcard>>& clusters) { }
};


class AR4PartialMatchClusteringAP3f : public AR4PartialMatchClusteringAP<ManifoldTraits3f>
{
public:
	typedef AR4PartialMatch3f PartialMatch;

protected:

public:

	AR4PartialMatchClusteringAP3f(const AR4Settings* settings): AR4PartialMatchClusteringAP(settings)
	{
	}

	void run(vector<AR4PartialMatch3f*>& partialMatches, vector<vector<mpcard>>& clusters) {}
	void run(AR4PartialMatch3f* pRefMatch, vector<GenericSwapAllOutOfCorePointCloud*>& partialMatches, vector<vector<mpcard>>& clusters);
};


template<class ManifoldTraits>
class AR4PartialMatchDirectCorrClustering : public AR4PartialMatchClustering<ManifoldTraits>
{
public:

	AR4PartialMatchDirectCorrClustering(const AR4Settings* settings): AR4PartialMatchClustering(settings)
	{
	}

	~AR4PartialMatchDirectCorrClustering()
	{
	}

	virtual void run(vector<AR4PartialMatch<ManifoldTraits>*>& partialMatches, vector<vector<mpcard>>& clusters) { }
	virtual void run(AR4PartialMatch3f* pRefMatch, vector<GenericSwapAllOutOfCorePointCloud*>& partialMatches, vector<vector<mpcard>>& clusters) { }
};


class AR4PartialMatchDirectCorrClustering3f : public AR4PartialMatchDirectCorrClustering<ManifoldTraits3f>
{
public:

	typedef AR4PartialMatch3f PartialMatch;

protected:

	class PointCorr
	{
	public:
		mpcard			m_sourceIndex;
		Vector3f		m_targetPoint;

		PointCorr()
		{
			m_sourceIndex = 0;
			m_targetPoint = NULL_VECTOR3F;
		}
		PointCorr(mpcard index, const Vector3f& targPos): m_sourceIndex(index), m_targetPoint(targPos)
		{
		}
		
		PointCorr& operator=(const PointCorr& corr)
		{
			m_sourceIndex = corr.m_sourceIndex;
			m_targetPoint = corr.m_targetPoint;
			return *this;
		}

		inline bool operator<(const PointCorr& corr) const
		{
			return (m_sourceIndex < corr.m_sourceIndex);
		}
	};


	vector<PointCorr>						m_correspondences;


public:

	AR4PartialMatchDirectCorrClustering3f(const AR4Settings* settings): AR4PartialMatchDirectCorrClustering<ManifoldTraits3f>(settings)
	{
	}

	~AR4PartialMatchDirectCorrClustering3f()
	{
	}

	bool getPointMatch(const mpcard imatch, mpcard& sourceIndex, Vector3f& targetPoint)
	{
		if (imatch < m_correspondences.size())
		{
			sourceIndex = m_correspondences[imatch].m_sourceIndex;
			targetPoint = m_correspondences[imatch].m_targetPoint;
			return true;
		}

		sourceIndex = (mpcard)-1;
		targetPoint = NULL_VECTOR3F;
		return false;
	}

	virtual void run(AR4PartialMatch3f* pRefMatch, vector<GenericSwapAllOutOfCorePointCloud*>& partialMatches, vector<vector<mpcard>>& clusters);
};


template<class ManifoldTraits>
class AR4PartialMatchSpectralClustering : public AR4PartialMatchClustering<ManifoldTraits>
{
protected:

public:

	AR4PartialMatchSpectralClustering(const AR4Settings* settings): AR4PartialMatchClustering(settings)
	{
	}

	virtual void run(vector<AR4PartialMatch<ManifoldTraits>*>& partialMatches, vector<vector<mpcard>>& clusters) { }
	virtual void run(AR4PartialMatch3f* pRefMatch, vector<GenericSwapAllOutOfCorePointCloud*>& partialMatches, vector<vector<mpcard>>& clusters) { }
};


class AR4PartialMatchSpectralClustering3f : public AR4PartialMatchSpectralClustering<ManifoldTraits3f>
{
protected:

public:

	AR4PartialMatchSpectralClustering3f(const AR4Settings* settings): AR4PartialMatchSpectralClustering<ManifoldTraits3f>(settings)
	{
	}

	virtual void run(AR4PartialMatch3f* pRefMatch, vector<GenericSwapAllOutOfCorePointCloud*>& partialMatches, vector<vector<mpcard>>& clusters);
};


}


#endif //__AR4PARTIALMATCHCLUSTERING_H__

