///////////////////////////////////////////////////////////////////////////////
//
//	AR4PartialMatchMaterial.cpp
//
//	Source file for the AR4PartialMatchMaterial class
//	subclass of GLMaterial
//	for displaying partial surface matches (pointwise correspondences)
//
//	Alan Brunton, MPII, March 2013
//
///////////////////////////////////////////////////////////////////////////////


//---------------------------------------------------------------------------
#include "stdafx.h"
//---------------------------------------------------------------------------
#include "AR4PartialMatchMaterial.h"
//---------------------------------------------------------------------------
#include "CopyObjectProperties.h"
#include "PointCloudIterators.h"
#include "PointSet.h"
#include "SeparatorClassProperty.h"
#include "BooleanClassProperty.h"
#include "FixedArrayClassProperty.h"
#include "RenderingTool.h"
#include "PointSet.inline.h"
//---------------------------------------------------------------------------


namespace X4
{
	IMPLEMENT_X4_CLASS(AR4PartialMatchMaterial, 3)
	{
		BEGIN_CLASS_INIT(AR4PartialMatchMaterial);
		ADD_SEPARATOR("position")
		ADD_STRING_PROP(m_strPositionChannel, 0)
		ADD_STRING_PROP(m_strTargetPosChannel, 0)
		ADD_SEPARATOR("radius")
		ADD_BOOLEAN_PROP(m_bUseRadiusChannel, 0)
		ADD_STRING_PROP(m_strRadiusChannel, 0)
		ADD_FLOAT32_PROP(m_radiusScale, 1)
		ADD_SEPARATOR("color")
		ADD_BOOLEAN_PROP(m_bUseColorChannel, 0)
		ADD_STRING_PROP(m_strColorChannel, 0)
		ADD_VECTOR3F_PROP(m_defaultColorRepeatPeriod, 0)
		ADD_BOOLEAN_PROP(m_bOnlyHighFreqPattern, 3)
	}

	AR4PartialMatchMaterial::AR4PartialMatchMaterial()
	{
		m_strPositionChannel = "position";
		m_strTargetPosChannel = "targetPos";
		m_bUseRadiusChannel = true;
		m_strRadiusChannel = "sampleSpacing";
		m_radiusScale = 0.03f;
		m_bUseColorChannel = false;
		m_strColorChannel = "color";
		m_defaultColorRepeatPeriod = makeVector3f(1.f, 1.f, 1.f);
		m_bOnlyHighFreqPattern = false;
		setName("AR4PartialMatchMaterial");
	}

	AR4PartialMatchMaterial::~AR4PartialMatchMaterial()
	{
	}

	void AR4PartialMatchMaterial::assign(const Object* obj, X4CopyContext *context)
	{
		copyObjectProperties(obj, this);
	}

	void AR4PartialMatchMaterial::draw(	Scene *scene, SceneGraphState *state, SceneObject *obj,
										SceneObjectIterator *iter, GLContext *context, MemoryManager * mm,
										const int32 includeFlags, const int32 excludeFlags)
	{
		const float32 color_period_eps = 1e-5f;

		BasicPointCloudIterator *bit = dynamic_cast<BasicPointCloudIterator*>(iter);

		if (bit == NULL)
			return;

		const PointSet *ps = bit->lockBlockRead();

		AAT positionAAT = NULL_AAT;
		AAT targetPosAAT = NULL_AAT;
		AAT radiusAAT  = NULL_AAT;
		AAT colorAAT = NULL_AAT;

		Vector3f colorRepeatFreq;

		m_defaultColorRepeatPeriod[0] = max(fabs(m_defaultColorRepeatPeriod[0]), color_period_eps);
		m_defaultColorRepeatPeriod[1] = max(fabs(m_defaultColorRepeatPeriod[1]), color_period_eps);
		m_defaultColorRepeatPeriod[2] = max(fabs(m_defaultColorRepeatPeriod[2]), color_period_eps);
		colorRepeatFreq[0] = 1.f / m_defaultColorRepeatPeriod[0];
		colorRepeatFreq[1] = 1.f / m_defaultColorRepeatPeriod[1];
		colorRepeatFreq[2] = 1.f / m_defaultColorRepeatPeriod[2];

		if (ps->providesAttribute(m_strPositionChannel) && ps->providesAttribute(m_strTargetPosChannel))
		{
			positionAAT = ps->getAAT(m_strPositionChannel);
			targetPosAAT = ps->getAAT(m_strTargetPosChannel);

			if (m_bUseRadiusChannel && ps->providesAttribute(m_strRadiusChannel))
				radiusAAT = ps->getAAT(m_strRadiusChannel);

			if (m_bUseColorChannel && ps->providesAttribute(m_strColorChannel))
				colorAAT = ps->getAAT(m_strColorChannel);

			Vector3f color;

			glPushAttrib(GL_ALL_ATTRIB_BITS);
			glEnable(GL_DEPTH_TEST);
			glDisable(GL_BLEND);
			glEnable(GL_LIGHTING);
			glEnable(GL_COLOR_MATERIAL);
			glEnable(GL_LIGHT0);

			float ambient[] = { 0.1f, 0.1f, 0.1f, 1.0 };
			float specular[] = { 1.0, 1.0, 1.0, 1.0 };
			float shininess[] = { 20.0 };
			float light_pos[] = { 1.0, 1.0, 1.0, 0.0 };

			glMaterialfv( GL_FRONT, GL_AMBIENT, ambient );
			glMaterialfv( GL_FRONT, GL_SPECULAR, specular );
			glMaterialfv( GL_FRONT, GL_SHININESS, shininess );
			glLightfv( GL_LIGHT0, GL_POSITION, light_pos );

			glMatrixMode(GL_MODELVIEW);

			float radius = 1.f;
			GLUquadricObj *q = gluNewQuadric();

			const card32 n = ps->getNumEntries();

			if (radiusAAT != NULL_AAT)
			{
				if (colorAAT != NULL_AAT)
				{
					for (card32 i = 0; i < n; i++)
					{
						Vector3f position = ps->get3f(i, positionAAT);
						Vector3f targetPos = ps->get3f(i, targetPosAAT);
						radius = ps->get1f(i, radiusAAT);
						color = ps->get3f(i, colorAAT);

						glColor3f(color[0], color[1], color[2]);

						glPushMatrix();
						glTranslatef(position[0], position[1], position[2]);
						gluSphere(q, radius, 10, 10);
						glPopMatrix();

						glPushMatrix();
						glTranslatef(targetPos[0], targetPos[1], targetPos[2]);
						gluSphere(q, radius, 10, 10);
						glPopMatrix();
					}
				}
				else if (m_bOnlyHighFreqPattern)
				{
					for (card32 i = 0; i < n; i++)
					{
						Vector3f position = ps->get3f(i, positionAAT);
						Vector3f targetPos = ps->get3f(i, targetPosAAT);
						radius = ps->get1f(i, radiusAAT);

						color[0] = (fmod(fabs(position[0]), m_defaultColorRepeatPeriod[0]) * colorRepeatFreq[0]) >= 0.5f ? 1.f : 0.f;
						color[1] = (fmod(fabs(position[1]), m_defaultColorRepeatPeriod[1]) * colorRepeatFreq[1]) >= 0.5f ? 1.f : 0.f;
						color[2] = (fmod(fabs(position[2]), m_defaultColorRepeatPeriod[2]) * colorRepeatFreq[2]) >= 0.5f ? 1.f : 0.f;

						glColor3f(color[0], color[1], color[2]);

						glPushMatrix();
						glTranslatef(position[0], position[1], position[2]);
						gluSphere(q, radius, 10, 10);
						glPopMatrix();

						glPushMatrix();
						glTranslatef(targetPos[0], targetPos[1], targetPos[2]);
						gluSphere(q, radius, 10, 10);
						glPopMatrix();
					}
				}
				else
				{
					for (card32 i = 0; i < n; i++)
					{
						Vector3f position = ps->get3f(i, positionAAT);
						Vector3f targetPos = ps->get3f(i, targetPosAAT);
						radius = ps->get1f(i, radiusAAT);

						color[0] = fmod(fabs(position[0]), m_defaultColorRepeatPeriod[0]) * colorRepeatFreq[0];
						color[1] = fmod(fabs(position[1]), m_defaultColorRepeatPeriod[1]) * colorRepeatFreq[1];
						color[2] = fmod(fabs(position[2]), m_defaultColorRepeatPeriod[2]) * colorRepeatFreq[2];

						glColor3f(color[0], color[1], color[2]);

						glPushMatrix();
						glTranslatef(position[0], position[1], position[2]);
						gluSphere(q, radius, 10, 10);
						glPopMatrix();

						glPushMatrix();
						glTranslatef(targetPos[0], targetPos[1], targetPos[2]);
						gluSphere(q, radius, 10, 10);
						glPopMatrix();
					}
				}
			}
			else
			{
				radius = m_radiusScale;

				if (colorAAT != NULL_AAT)
				{
					for (card32 i = 0; i < n; i++)
					{
						Vector3f position = ps->get3f(i, positionAAT);
						Vector3f targetPos = ps->get3f(i, targetPosAAT);
						color = ps->get3f(i, colorAAT);

						glColor3f(color[0], color[1], color[2]);

						glPushMatrix();
						glTranslatef(position[0], position[1], position[2]);
						gluSphere(q, radius, 10, 10);
						glPopMatrix();

						glPushMatrix();
						glTranslatef(targetPos[0], targetPos[1], targetPos[2]);
						gluSphere(q, radius, 10, 10);
						glPopMatrix();
					}
				}
				else if (m_bOnlyHighFreqPattern)
				{
					for (card32 i = 0; i < n; i++)
					{
						Vector3f position = ps->get3f(i, positionAAT);
						Vector3f targetPos = ps->get3f(i, targetPosAAT);

						color[0] = (fmod(fabs(position[0]), m_defaultColorRepeatPeriod[0]) * colorRepeatFreq[0]) >= 0.5f ? 1.f : 0.f;
						color[1] = (fmod(fabs(position[1]), m_defaultColorRepeatPeriod[1]) * colorRepeatFreq[1]) >= 0.5f ? 1.f : 0.f;
						color[2] = (fmod(fabs(position[2]), m_defaultColorRepeatPeriod[2]) * colorRepeatFreq[2]) >= 0.5f ? 1.f : 0.f;

						glColor3f(color[0], color[1], color[2]);

						glPushMatrix();
						glTranslatef(position[0], position[1], position[2]);
						gluSphere(q, radius, 10, 10);
						glPopMatrix();

						glPushMatrix();
						glTranslatef(targetPos[0], targetPos[1], targetPos[2]);
						gluSphere(q, radius, 10, 10);
						glPopMatrix();
					}
				}
				else
				{
					for (card32 i = 0; i < n; i++)
					{
						Vector3f position = ps->get3f(i, positionAAT);
						Vector3f targetPos = ps->get3f(i, targetPosAAT);

						color[0] = fmod(fabs(position[0]), m_defaultColorRepeatPeriod[0]) * colorRepeatFreq[0];
						color[1] = fmod(fabs(position[1]), m_defaultColorRepeatPeriod[1]) * colorRepeatFreq[1];
						color[2] = fmod(fabs(position[2]), m_defaultColorRepeatPeriod[2]) * colorRepeatFreq[2];

						glColor3f(color[0], color[1], color[2]);

						glPushMatrix();
						glTranslatef(position[0], position[1], position[2]);
						gluSphere(q, radius, 10, 10);
						glPopMatrix();

						glPushMatrix();
						glTranslatef(targetPos[0], targetPos[1], targetPos[2]);
						gluSphere(q, radius, 10, 10);
						glPopMatrix();
					}
				}
			}

			gluDeleteQuadric(q);
			glPopAttrib();
		}

		bit->unlock();
	}
}


