///////////////////////////////////////////////////////////////////////////////
//
//	AR4PartialMatchMerger.cpp
//
//	Source file for the AR4PartialMatchMerger class
//	merges partial isometric matches
//
//	Alan Brunton, MPII, March 2013
//
///////////////////////////////////////////////////////////////////////////////


#include "stdafx.h"

#include "AR4PartialMatchMerger.h"
#include "AR4PartialMatch.inline.h"
#include "AR4IncrementalMatcher.h"
#include "AR4IncrementalMatcher.inline.h"

#include "DebugRenderer.h"



#define AR4PMM_ORDER_WEIGHTED_MERGER		1
#define AR4PMM_ORDER_WEIGHTED_MERGER_CTF	0


void X4::AR4PartialMatchWeightedAverageMerger3f::run(std::vector<GenericSwapAllOutOfCorePointCloud*> &partialMatches, const std::vector<mpcard> &matchIndices, X4::AR4PartialMatch3f *pMergeMatch)
{
	//number of partial matches in the cluster
	const mpcard numMatches = matchIndices.size();

	const float stretchMu = pMergeMatch->getStretchThreshold() * pMergeMatch->getAvgSourceMatchLevelEdgeWeight() * 0.5f;
	const float stretchLambda = 1.f / stretchMu;
	const float distStartMu = pMergeMatch->getAvgSourceMatchLevelEdgeWeight() * 5.f;
	const float distStartLambda = 1.f / distStartMu;

	mpcard i, j;

#if AR4PMM_ORDER_WEIGHTED_MERGER_CTF
	AR4CoarseToFineSparseBundle3f bundler(pMergeMatch->getSourceManifold(), pMergeMatch->getTargetManifold());
#else
	AR4SparseBundle3f bundler(pMergeMatch->getSourceManifold(), pMergeMatch->getTargetManifold());
#endif //AR4PMM_ORDER_WEIGHTED_MERGER_CTF

	AR4MLSManifold3f* pmlsTarget = dynamic_cast<AR4MLSManifold3f*>(pMergeMatch->getTargetManifold());
	if (pmlsTarget == NULL)
		throw EInvalidState("target is not an MLS manifold!");

	map<card32, vector<TWeightedPos>> sourceIndicesMatched;
	map<card32, vector<TWeightedPos>>::iterator simIt;

	//go over partial matches in the cluster
	for (i = 0; i < numMatches; i++)
	{
		const mpcard im = matchIndices[i];

		UnstructuredInCorePointCloud* pMatchPC = dynamic_cast<UnstructuredInCorePointCloud*>(partialMatches[im]->fetchPC());

		//number of points in the partial match
		const mpcard numPoints = pMatchPC->getNumPoints();

		AAT pos = pMatchPC->getAAT("position");
		AAT idx = pMatchPC->getAAT(AR4PartialMatch3f::getMatchAttNameIndex());
		AAT targ = pMatchPC->getAAT(AR4PartialMatch3f::getMatchAttNameTargetPos());
		AAT str = pMatchPC->getAAT(AR4PartialMatch3f::getMatchAttNameStretch());
		AAT dst = pMatchPC->getAAT(AR4PartialMatch3f::getMatchAttNameDistToStart());

		//go over number of points in the partial match
		for (j = 0; j < numPoints; j++)
		{
			const card32 index = pMatchPC->getPointSet()->get1u(j, idx);
			const float stretch = pMatchPC->getPointSet()->get1f(j, str);
			const float distStart = pMatchPC->getPointSet()->get1f(j, dst);
			const float weight = exp(-(distStart * distStartLambda));
			const Vector3f targetPos = pMatchPC->getPointSet()->get3f(j, targ);

			const Vector3f sourcePos = pMatchPC->getPointSet()->get3f(j,pos);

	
			simIt = sourceIndicesMatched.find(index);
			if (simIt == sourceIndicesMatched.end())
			{
				//add new match with weight
				TWeightedPos twp(weight, targetPos);
				sourceIndicesMatched[index] = vector<TWeightedPos>(1, twp);
			}
			else
			{
				//update weight and position
				TWeightedPos twp(weight, targetPos);
				simIt->second.push_back(twp);
			}
		}
	}

	const float dist_eps = pMergeMatch->getAvgSourceMatchLevelEdgeWeight();
	const float dist_trans_max = dist_eps * 5.f;

#if AR4PMM_ORDER_WEIGHTED_MERGER

	const mpcard numIndices = sourceIndicesMatched.size();

	vector<TIndexRating> indexPosVars(numIndices);

#if AR4PMM_ORDER_WEIGHTED_MERGER_CTF

	const mpcard levelMax = pMergeMatch->getNumTopologyLevels() - 1;

	mpcard k = 0;
	for (simIt = sourceIndicesMatched.begin(); simIt != sourceIndicesMatched.end(); simIt++)
	{
		const card32 index = simIt->first;
		const mpcard numTargPos = simIt->second.size();

		//first some house-keeping: normalize the weights
		float weightSum = 0.f;
		for (i = 0; i < numTargPos; i++)
			weightSum += simIt->second[i].first;
		const float rcpWeightSum = 1.f / weightSum;
		for (i = 0; i < numTargPos; i++)
			simIt->second[i].first *= rcpWeightSum;

		mpcard level;
		for (level = 0; level <= levelMax; level++)
		{
			if (!pMergeMatch->isAtLevel((mpcard)index, level))
				break;
		}

		//keep track of this
		indexPosVars[k].first = (float)(levelMax - level);
		indexPosVars[k].second = index;
		k++;
	}

#else

	mpcard k = 0;
	for (simIt = sourceIndicesMatched.begin(); simIt != sourceIndicesMatched.end(); simIt++)
	{
		const card32 index = simIt->first;
		const mpcard numTargPos = simIt->second.size();

		//first some house-keeping: normalize the weights
		float weightSum = 0.f;
		for (i = 0; i < numTargPos; i++)
			weightSum += simIt->second[i].first;
		const float rcpWeightSum = 1.f / weightSum;
		for (i = 0; i < numTargPos; i++)
			simIt->second[i].first *= rcpWeightSum;

		//now measure how much the predicted target points vary for this particular source point
		//according to how far each one is from the others
		float distSum = 0.f;

		if (numTargPos > 1)
		{
			const mpcard numTargPosM1 = numTargPos - 1;
			const float scaleFactor = 1.f / (float)(numTargPos * numTargPosM1);
			vector<float> dists(numTargPosM1);
			vector<Vector3f> grads(numTargPosM1);

			vector<Vector3f> destSet(numTargPosM1);
			for (i = 1; i < numTargPos; i++)
				destSet[i - 1] = simIt->second[i].second;
			for (i = 0; i < numTargPos; i++)
			{
				pMergeMatch->getSparseTargetGeodesics()->computeGeodesicsEarlyStop(0.01f * dist_eps, simIt->second[i].second, destSet, dists, grads);

				for (j = 0; j < numTargPosM1; j++)
					distSum += dists[j];

				if (i < numTargPosM1)
					destSet[i] = simIt->second[i].second;
			}
			distSum *= scaleFactor;
		}
		else
			distSum = -1.f;
			

		//keep track of this
		indexPosVars[k].first = distSum;
		indexPosVars[k].second = index;
		k++;
	}

	float avgDistSum = 0.f;
	k = 0;
	for (i = 0; i < numIndices; i++)
	{
		if (indexPosVars[i].first < 0.f)	continue;
		avgDistSum += indexPosVars[i].first;
		k++;
	}

	if(k > 0) avgDistSum /= (float)k;


	//assign average sum of distance to points where there is no variation
	for (i = 0; i < numIndices; i++)
	{
		if (indexPosVars[i].first < 0.f)
			indexPosVars[i].first = avgDistSum;
			
	}

#endif //AR4PMM_ORDER_WEIGHTED_MERGER_CTF

	//sort so that the points that vary least are at the start
	sort(indexPosVars.begin(), indexPosVars.end());

	for (i = 0; i < numIndices; i++)
	{
		const card32 index = indexPosVars[i].second;
		const mpcard numPredicted = sourceIndicesMatched[index].size();

		vector<Vector3f> hintSet(numPredicted);
		vector<Vector3f> zeroSet(1);

		float sumMax = 0.f;

		hintSet[0] = sourceIndicesMatched[index][0].second;
		zeroSet[0] = sourceIndicesMatched[index][0].second;
		pMergeMatch->getSparseTargetGeodesics()->distanceTransform(zeroSet, dist_trans_max, false, sourceIndicesMatched[index][0].first);
		sumMax += dist_trans_max;

	
		for (j = 1; j < numPredicted; j++)
		{
			hintSet[j] = sourceIndicesMatched[index][j].second;
			zeroSet[0] = sourceIndicesMatched[index][j].second;
			pMergeMatch->getSparseTargetGeodesics()->distanceTransform(zeroSet, dist_trans_max, true, sourceIndicesMatched[index][j].first);
			sumMax += dist_trans_max;
		}

		Vector3f geoCentroid;
		float geoCentroidDist;
		pMergeMatch->getSparseTargetGeodesics()->getMinDistanceContinuousHint(pMergeMatch->getGeodesicsSigma(), hintSet, &geoCentroid, &geoCentroidDist);

	
		if (geoCentroidDist < sumMax)
		{
			Vector3f surfPoint;
			if (pMergeMatch->getTargetManifold()->project(geoCentroid, surfPoint))
			{
				pMergeMatch->setMatch(index, surfPoint);


			}


		}


		if (i > 4 && pMergeMatch->getNumMatchedIndices() > bundler.getNumMatchesLastRun() * 2)
		{
			//find any additional geodesic connections (between matches) that were not computed during local bundle above
			pMergeMatch->checkForNewGeodesicConnections();

			//validate new paths
			pMergeMatch->validatePaths();

			//remove outliers from among the newly added points (based on metric distortion)
			pMergeMatch->finalizeMatchCohort();

			bundler.setStepScale(pMergeMatch->computeAvgSourceEdgeLength() * 0.5f);
			bundler.improveMatch(pMergeMatch);
		}


	}

#else

	for (simIt = sourceIndicesMatched.begin(); simIt != sourceIndicesMatched.end(); simIt++)
	{
		const card32 index = simIt->first;
		const mpcard numPos = simIt->second.size();

		vector<Vector3f> hintSet(numPos);

		pMergeMatch->getSparseTargetGeodesics()->distanceTransform(vector<Vector3f>(1, simIt->second[0].second), dist_trans_max, false, simIt->second[0].first);
		hintSet[0] = simIt->second[0].second;
		for (i = 1; i < numPos; i++)
		{
			pMergeMatch->getSparseTargetGeodesics()->distanceTransform(vector<Vector3f>(1, simIt->second[i].second), dist_trans_max, true, simIt->second[i].first);
			hintSet[i] = simIt->second[i].second;
		}

		Vector3f geoCentroid;
		float geoCentroidDist;
		pMergeMatch->getSparseTargetGeodesics()->getMinDistanceContinuousHint(pMergeMatch->getGeodesicsSigma(), hintSet, &geoCentroid, &geoCentroidDist);

		pMergeMatch->setMatch(index, geoCentroid);
	}

#endif //AR4PMM_ORDER_WEIGHTED_MERGER

	//find any additional geodesic connections (between matches) that were not computed during local bundle above
	pMergeMatch->checkForNewGeodesicConnections();

	//validate new paths
	pMergeMatch->validatePaths();

	//remove outliers from among the newly added points (based on metric distortion)
	pMergeMatch->finalizeMatchCohort();

	bundler.setStepScale(pMergeMatch->computeAvgSourceEdgeLength() * 0.5f);
	bundler.improveMatch(pMergeMatch);
}


void X4::AR4PartialMatchMetricPreservingMerger3f::run(vector<GenericSwapAllOutOfCorePointCloud*>& partialMatches, const vector<mpcard>& matchIndices, AR4PartialMatch3f* pMergeMatch)
{
	//number of partial matches in the cluster
	const mpcard numMatches = matchIndices.size();
	
	const float stretchMu = pMergeMatch->getStretchThreshold() * pMergeMatch->getAvgSourceMatchLevelEdgeWeight() * 0.5f;
	const float stretchLambda = 1.f / stretchMu;
	const float distStartMu = pMergeMatch->getAvgSourceMatchLevelEdgeWeight() * 5.f;
	const float distStartLambda = 1.f / distStartMu;
	const float distRefErrMu = pMergeMatch->getStretchThreshold() * pMergeMatch->getAvgSourceMatchLevelEdgeWeight() * 0.5f;
	const float distRefErrLambda = 1.f / distRefErrMu;

	mpcard i, j, k;

	map<card32, vector<TWeightedPos>> sourceIndicesMatched;
	map<card32, vector<TWeightedPos>>::iterator simIt;

	AR4SparseBundle3f bundler(pMergeMatch->getSourceManifold(), pMergeMatch->getTargetManifold());

	//go over partial matches in the cluster
	for (i = 0; i < numMatches; i++)
	{
		//im contains the index of the partial match
		const mpcard im = matchIndices[i];

		
		//pMatchPC contains the partial match point cloud
		UnstructuredInCorePointCloud* pMatchPC = dynamic_cast<UnstructuredInCorePointCloud*>(partialMatches[im]->fetchPC());

		const mpcard numPoints = pMatchPC->getNumPoints();
	
		AAT pos = pMatchPC->getAAT("position");
		AAT idx = pMatchPC->getAAT(AR4PartialMatch3f::getMatchAttNameIndex());
		AAT targ = pMatchPC->getAAT(AR4PartialMatch3f::getMatchAttNameTargetPos());
		AAT str = pMatchPC->getAAT(AR4PartialMatch3f::getMatchAttNameStretch());
		AAT dst = pMatchPC->getAAT(AR4PartialMatch3f::getMatchAttNameDistToStart());


	
	
		//go over points in the partial match
		for (j = 0; j < numPoints; j++)
		{

			card32 index = pMatchPC->getPointSet()->get1u(j, idx);			
			const float stretch = pMatchPC->getPointSet()->get1f(j, str);
			const float distStart = pMatchPC->getPointSet()->get1f(j, dst);
			const float weight = exp(-distStart * distStartLambda);

			
			const Vector3f sourcePos = pMergeMatch->getSourceManifold()->getDiscrPoint(index);
	
			const Vector3f targetPos = pMatchPC->getPointSet()->get3f(j, targ);


			simIt = sourceIndicesMatched.find(index);
			if (simIt == sourceIndicesMatched.end())
			{
				//add new match with weight
				TWeightedPos twp(weight, targetPos);
				sourceIndicesMatched[index] = vector<TWeightedPos>(1, twp);
			}
			else
			{
				//update weight and position
				TWeightedPos twp(weight, targetPos);
				simIt->second.push_back(twp);
			}





		}//go over points in the partial match
	
	
	}//go over partial matches in the cluster


	

	



	const float dist_eps = pMergeMatch->getAvgSourceMatchLevelEdgeWeight();
	const float dist_trans_max = dist_eps * 5.f;

	const mpcard numIndices = sourceIndicesMatched.size();
	vector<TIndexRating> indexPosVars(numIndices);

	k = 0;
	for (simIt = sourceIndicesMatched.begin(); simIt != sourceIndicesMatched.end(); simIt++)
	{

		const card32 index = simIt->first;
		const mpcard numTargPos = simIt->second.size();

	
		//first some house-keeping: normalize the weights
		float weightSum = 0.f;
		for (i = 0; i < numTargPos; i++)
			weightSum += simIt->second[i].first;
		const float rcpWeightSum = 1.f / weightSum;
		for (i = 0; i < numTargPos; i++)
			simIt->second[i].first *= rcpWeightSum;

		//now measure how much the predicted target points vary for this particular source point
		//according to how far each one is from the others
		float distSum = 0.f;
		if (numTargPos > 1)
		{
			const mpcard numTargPosM1 = numTargPos - 1;
			const float scaleFactor = 1.f / (float)(numTargPos * numTargPosM1);
			vector<float> dists(numTargPosM1);
			vector<Vector3f> grads(numTargPosM1);

			vector<Vector3f> destSet(numTargPosM1);
			for (i = 1; i < numTargPos; i++)
				destSet[i - 1] = simIt->second[i].second;
			for (i = 0; i < numTargPos; i++)
			{
				pMergeMatch->getSparseTargetGeodesics()->computeGeodesicsEarlyStop(0.01f * dist_eps, simIt->second[i].second, destSet, dists, grads);

				for (j = 0; j < numTargPosM1; j++)
					distSum += dists[j];

				if (i < numTargPosM1)
					destSet[i] = simIt->second[i].second;
			}
			
			distSum *= scaleFactor;
		}
		else
			distSum = -1.f;

		//keep track of this
		indexPosVars[k].first = distSum;
		indexPosVars[k].second = index;
		k++;
	}

	float avgDistSum = 0.f;
	k = 0;
	for (i = 0; i < numIndices; i++)
	{
		if (indexPosVars[i].first < 0.f)	continue;
		avgDistSum += indexPosVars[i].first;
		k++;
	}
	avgDistSum /= (float)k;

	for (i = 0; i < numIndices; i++)
	{
		if (indexPosVars[i].first < 0.f)
			indexPosVars[i].first = avgDistSum;
	}

	//sort so that the points that vary least are at the start
	sort(indexPosVars.begin(), indexPosVars.end());

	AR4MLSManifold3f* pTargetMLS = dynamic_cast<AR4MLSManifold3f*>(pMergeMatch->getTargetManifold());

	const mpcard numTargPts = pTargetMLS->getNumDiscrPoints();

	vector<float> currentDistTrans(numTargPts);
	vector<float> sumDistTrans(numTargPts, 0.f);

	for (i = 0; i < numIndices; i++)
	{
		const card32 index = indexPosVars[i].second;

		const mpcard numPredicted = sourceIndicesMatched[index].size();


		vector<mpcard> sourceDests;
		vector<float> sourceDists;
		vector<Vector3f> targetDests;
		vector<float> targetDists;
		vector<Vector3f> targetGrads;

		//for this index, minimize the sum of distances to target points predicted by partial matches and the sum the stretch to previous points
		float stretchTerm = 0.f;
		for (j = 0; j < i; j++)
		{
			const mpcard prevIndex = indexPosVars[i].second;

			float sourceDist;
			Vector3f sourceGrad;

			//if we have pre-computed distances on the source between these points, keep this point as a "reference"
			Vector3f dirAtSource;
			if (pMergeMatch->getSparseSourceGeodesics()->getGeodesicData(index, prevIndex, sourceDist, sourceGrad, dirAtSource) != AR4SparseGeodesics::m_InvalidIndex)
			{
				sourceDests.push_back(prevIndex);
				sourceDists.push_back(sourceDist);
				targetDests.push_back(pMergeMatch->getMatch(prevIndex));
			}
		}

		const mpcard numRefs = sourceDests.size();

		fill(sumDistTrans.begin(), sumDistTrans.end(), 0.f);
		pMergeMatch->getSparseTargetGeodesics()->clearDistanceField();

		float funcMax = 0.f;

		//find point that minimizes distance to predicted points and minimizes difference of distances with reference points
		//first compute distance-differences wrt reference points
		for (j = 0; j < numRefs; j++)
		{
			const mpcard prevIndex = sourceDests[j];
			const Vector3f targPos = pMergeMatch->getMatch(prevIndex);

			const float transMax = sourceDists[j] * 3.f;
			funcMax += transMax;
			
			vector<Vector3f> zeroSet(1, targPos);
			pMergeMatch->getSparseTargetGeodesics()->distanceTransform(zeroSet, transMax);

			pMergeMatch->getSparseTargetGeodesics()->copyDistanceField(currentDistTrans);

			const float distRef = sourceDists[j];
			for (k = 0; k < numTargPts; k++)
				sumDistTrans[k] += fabs(currentDistTrans[k] - distRef);
		}

		//now add distances to points predicted by partial matches
		vector<Vector3f> hintSet(numPredicted);
		for (j = 0; j < numPredicted; j++)
		{
			funcMax += dist_trans_max;

			vector<Vector3f> zeroSet(1, sourceIndicesMatched[index][j].second);
			pMergeMatch->getSparseTargetGeodesics()->distanceTransform(zeroSet, dist_trans_max);

			const float weight = sourceIndicesMatched[index][j].first;

			pMergeMatch->getSparseTargetGeodesics()->copyDistanceField(currentDistTrans);
			for (k = 0; k < numTargPts; k++)
				sumDistTrans[k] += currentDistTrans[k] * weight;

			hintSet[j] = sourceIndicesMatched[index][j].second;
		}

		//find minimum
		Vector3f distSumMinPoint;
		float distSumMin;
		pTargetMLS->findScalarFieldMinimum(sumDistTrans, hintSet, distSumMinPoint, distSumMin);
		if (distSumMin < funcMax)
		{
			pMergeMatch->setMatch(index, distSumMinPoint);

		}
	

		if (i > 4 && i > bundler.getNumMatchesLastRun() * 2)
		{
			//find any additional geodesic connections (between matches) that were not computed during local bundle above
			pMergeMatch->checkForNewGeodesicConnections();

			//validate new paths
			pMergeMatch->validatePaths();

			//remove outliers from among the newly added points (based on metric distortion)
			pMergeMatch->finalizeMatchCohort();

			bundler.setStepScale(pMergeMatch->computeAvgSourceEdgeLength() * 0.5f);
			bundler.improveMatch(pMergeMatch);
		}


	}


	//find any additional geodesic connections (between matches) that were not computed during local bundle above
	pMergeMatch->checkForNewGeodesicConnections();

	//validate new paths
	pMergeMatch->validatePaths();

	//remove outliers from among the newly added points (based on metric distortion)
	pMergeMatch->finalizeMatchCohort();

	bundler.setStepScale(pMergeMatch->computeAvgSourceEdgeLength() * 0.5f);
	bundler.improveMatch(pMergeMatch);



}


void X4::AR4PartialMatchBPDirectMerger3f::initDistanceFields(vector<GenericSwapAllOutOfCorePointCloud*>& partialMatches)
{
	//for each partial match
	//select subset of matches from which to compute geodesics on both surfaces
	//compute geodesics to all other points on both surfaces
	//store those corresponding to corresponded points

	const mpcard numPartials = partialMatches.size();

	mpcard i, j, k, m, n, p;

	map<mpcard, vector<PartialMapping>>::iterator pmIt;

	vector<mpcard> sourceZeros(1);
	vector<Vector3f> targetZeros(1);

	vector<vector<pair<mpcard, Vector3f>>> partialSamples(numPartials);

	for (i = 0; i < numPartials; i++)
	{
		UnstructuredInCorePointCloud* pMatchPC = dynamic_cast<UnstructuredInCorePointCloud*>(partialMatches[i]->fetchPC());

		const mpcard numPoints = pMatchPC->getNumPoints();

		AAT pos = pMatchPC->getAAT("position");
		AAT idx = pMatchPC->getAAT(AR4PartialMatch3f::getMatchAttNameIndex());
		AAT targ = pMatchPC->getAAT(AR4PartialMatch3f::getMatchAttNameTargetPos());
		AAT str = pMatchPC->getAAT(AR4PartialMatch3f::getMatchAttNameStretch());
		AAT dst = pMatchPC->getAAT(AR4PartialMatch3f::getMatchAttNameDistToStart());

		float32 distToStartMax = 0.f;

		partialSamples[i].resize(m_nFieldsPerPartial, pair<mpcard, Vector3f>(m_pSourceManifold->getNumDiscrPoints(), NULL_VECTOR3F));

		for (j = 0; j < numPoints; j++)
		{
			const card32 index = pMatchPC->getPointSet()->get1u(j, idx);
			const Vector3f targetPos = pMatchPC->getPointSet()->get3f(j, targ);
			const float32 distToStart = pMatchPC->getPointSet()->get1f(j, dst);

			if (distToStart > distToStartMax)
				distToStartMax = distToStart;

			if (j == 0)
			{
				partialSamples[i][0].first = (mpcard)index;
				partialSamples[i][0].second = targetPos;
			}

			PartialMapping pm(i, targetPos);

			pmIt = m_possibleMatches.find((mpcard)index);
			if (pmIt == m_possibleMatches.end())
			{
				m_possibleMatches[(mpcard)index] = vector<PartialMapping>(1, pm);
			}
			else
			{
				pmIt->second.push_back(pm);
			}
		}

		vector<bool> wasUsed(numPoints, false);
		wasUsed[0] = true;
		const float32 rScale = (float32)(numPoints - 1);
		for (j = 1; j < m_nFieldsPerPartial; j++)
		{
			float32 r = rnd01();
			mpcard idxm = (mpcard)(r * rScale) + 1;
			while (wasUsed[idxm])
			{
				r = rnd01();
				idxm = (mpcard)(r * rScale) + 1;
			}

			partialSamples[i][j].first = idxm;
			partialSamples[i][j].second = pMatchPC->getPointSet()->get3f(idxm, targ);
			wasUsed[idxm] = true;
		}
	}

	//compute distance transforms
	m_sourceDistFields.resize(numPartials * m_nFieldsPerPartial);
	m_targetDistFields.resize(numPartials * m_nFieldsPerPartial);
	const mpcard numSourcePoints = m_possibleMatches.size();
	m_nSourcePoints = numSourcePoints;

	m_sourceIndices.resize(numSourcePoints);
	m_sourceMatchOffsets.resize(numSourcePoints);

	m = k = 0;
	for (pmIt = m_possibleMatches.begin(); pmIt != m_possibleMatches.end(); pmIt++)
	{
		m_sourceIndices[k] = pmIt->first;
		m_sourceMatchOffsets[k] = m;

		const mpcard numPossible = pmIt->second.size();
		for (n = 0; n < numPossible; n++)
		{
			m++;
		}
		k++;
	}

	const mpcard numTargetPoints = m;
	m_nTargetPoints = numTargetPoints;

#if !AR4PMM_BP_PAIRWISE_POINT_DISTANCES
	k = 0;
	for (i = 0; i < numPartials; i++)
	{
		for (j = 0; j < m_nFieldsPerPartial; j++)
		{
			m_sourceDistFields[k].resize(numSourcePoints);
			m_targetDistFields[k].resize(numTargetPoints);

			sourceZeros[0] = partialSamples[i][j].first;
			targetZeros[0] = partialSamples[i][j].second;

			m_pSourceGeos->distanceTransform(sourceZeros);
			m_pTargetGeos->distanceTransform(targetZeros);

			m = 0;
			n = 0;
			for (pmIt = m_possibleMatches.begin(); pmIt != m_possibleMatches.end(); pmIt++)
			{
				m_sourceDistFields[k][m] = m_pSourceGeos->getDistanceField(pmIt->first);

				const mpcard numPossible = pmIt->second.size();
				for (p = 0; p < numPossible; p++)
				{
					m_targetDistFields[k][n] = m_pTargetGeos->getDistanceField(pmIt->second[p].m_targetPos);
					n++;
				}
				m++;
			}

			k++;
		}
	}
#endif //!AR4PMM_BP_PAIRWISE_POINT_DISTANCES
}

void X4::AR4PartialMatchBPDirectMerger3f::initPairwiseDistances(vector<GenericSwapAllOutOfCorePointCloud*>& partialMatches)
{
}

void X4::AR4PartialMatchBPDirectMerger3f::initFactorGraph()
{
	clearFactorGraph();

	const mpcard numSourcePoints = m_nSourcePoints;
	const mpcard numTargetPoints = m_nTargetPoints;

	mpcard i, j;
	map<mpcard, vector<PartialMapping>>::iterator pmIt, pmLU;

	map<pair<mpcard, mpcard>, PGCN*> varsLinked;

	i = 0;
	for (pmIt = m_possibleMatches.begin(); pmIt != m_possibleMatches.end(); pmIt++)
	{
		const mpcard numPossible = pmIt->second.size();
		const mpcard offset = m_sourceMatchOffsets[i];

		vector<mpcard> labelMap(numPossible);
		for (j = 0; j < numPossible; j++)
			labelMap[j] = pmIt->second[j].m_iPartial;

		BPVariableNode* pVar = new BPVariableNode(numPossible, i, labelMap);

		m_variables.push_back(pVar);

		vector<mpcard> sourceLinks;
		vector<float> dists;
		vector<Vector3f> grads;

		m_pSourceGeos->getGeodesicData(pmIt->first, sourceLinks, dists, grads);

		const mpcard numLinks = sourceLinks.size();
		for (j = 0; j < numLinks; j++)
		{
			pair<mpcard, mpcard> link(sourceLinks[j], pmIt->first);
			pair<mpcard, mpcard> linkrev(pmIt->first, sourceLinks[j]);

			pmLU = m_possibleMatches.find(sourceLinks[j]);
			if (pmLU != m_possibleMatches.end())
			{
				//create a factor between these two nodes
				if (varsLinked.find(link) == varsLinked.end())
				{
					PGCN* pgcn = new PGCN(&m_sourceDistFields[0], &m_targetDistFields[0], &m_sourceMatchOffsets[0], numSourcePoints, numTargetPoints, m_nFieldsPerPartial);
					pgcn->addNeighbor(pVar);
					m_factors.push_back(pgcn);
					varsLinked[linkrev] = pgcn;

#if AR4PMM_BP_PAIRWISE_POINT_DISTANCES
					pgcn->m_pairwiseErrors.resize(pVar->m_nLabels);
#endif //AR4PMM_BP_PAIRWISE_POINT_DISTANCES
				}
				else
				{
					PGCN* pgcn = varsLinked[link];
					pgcn->addNeighbor(pVar);

#if AR4PMM_BP_PAIRWISE_POINT_DISTANCES
					const mpcard nFirstLabels = pgcn->m_pairwiseErrors.size();
					for (mpcard m = 0; m < nFirstLabels; m++)
					{
						const Vector3f startPos = pmLU->second[m].m_targetPos;//m_possibleMatches[sourceLinks[j]][m].m_targetPos;

						const mpcard nSecondLabels = pVar->m_nLabels;
						pgcn->m_pairwiseErrors[m].resize(nSecondLabels);
						for (mpcard n = 0; n < nSecondLabels; n++)
						{
							float targetDist;
							Vector3f targetGrad;
							m_pTargetGeos->computeGeodesicSingleDest(0.01f, startPos, pmIt->second[n].m_targetPos, targetDist, targetGrad);
							const float distDiff = targetDist - dists[j];
							pgcn->m_pairwiseErrors[m][n] = distDiff*distDiff;
						}
					}
#endif //AR4PMM_BP_PAIRWISE_POINT_DISTANCES
				}
			}
		}

		i++;
	}

	const mpcard numVariables = m_variables.size();
	for (i = 0; i < numVariables; i++)
	{
		m_variables[i]->init();
	}

	const mpcard numFactors = m_factors.size();
	for (i = 0; i < numFactors; i++)
	{
		m_factors[i]->init();
	}
}

void X4::AR4PartialMatchBPDirectMerger3f::iterateFactorGraph()
{
	const mpcard numFactors = m_factors.size();
	const mpcard numVariables = m_variables.size();

	mpcard f, v;

	for (f = 0; f < numFactors; f++)
	{
		m_factors[f]->computeMessages();
		m_factors[f]->normalizeMessages();
	}

	for (v = 0; v < numVariables; v++)
	{
		m_variables[v]->computeMessages();
		m_variables[v]->normalizeMessages();
	}
}

void X4::AR4PartialMatchBPDirectMerger3f::computeMAPSolution(AR4PartialMatch3f* pMAPMatch)
{
	const mpcard numVariables = m_variables.size();

	mpcard v;

	for (v = 0; v < numVariables; v++)
	{
		m_variables[v]->computeBeliefs();

		mpcard sourceIdx = m_sourceIndices[m_variables[v]->m_appDataLU];
		
		if (m_variables[v]->m_MAPLabel < m_variables[v]->m_nLabels)
		{
			Vector3f MAPtargetPos = m_possibleMatches[sourceIdx][m_variables[v]->m_MAPLabel].m_targetPos;
			pMAPMatch->setMatch(sourceIdx, MAPtargetPos);
		}
	}

	pMAPMatch->finalizeMatchCohort();
}

void X4::AR4PartialMatchBPDirectMerger3f::run(vector<GenericSwapAllOutOfCorePointCloud*>& partialMatches, const vector<mpcard>& matchIndices, AR4PartialMatch3f* pMergeMatch)
{
	m_pSourceManifold = dynamic_cast<AR4MLSManifold3f*>(pMergeMatch->getSourceManifold());
	m_pTargetManifold = dynamic_cast<AR4MLSManifold3f*>(pMergeMatch->getTargetManifold());
	m_pSourceGeos = dynamic_cast<AR4SparseGeodesics*>(pMergeMatch->getSparseSourceGeodesics());
	m_pTargetGeos = dynamic_cast<AR4SparseGeodesics*>(pMergeMatch->getSparseTargetGeodesics());

	debugOutput << "initializing distance fields\n";
	initDistanceFields(partialMatches);

	debugOutput << "initializing factor graph\n";
	initFactorGraph();

	mpcard iter;

	debugOutput << "iterating factor graph\n";
	for (iter = 0; iter < 100; iter++)
		iterateFactorGraph();

	debugOutput << "computing MAP solution\n";
	computeMAPSolution(pMergeMatch);
}

