///////////////////////////////////////////////////////////////////////////////
//
//	AR4Profile.h
//
//	Header file for the AR4Profile class
//	collects statistics of multiple timings of (presumably) the same set of computations
//
//	Alan Brunton, January 2013
//
///////////////////////////////////////////////////////////////////////////////


#ifndef __AR4PROFILE_H__
#define __AR4PROFILE_H__


#include "threeParameterMatchingGMOD14.h"

#include "Timer.h"


namespace X4
{

class AR4Profile
{
private:

	Timer									m_timer;
	vector<card32>							m_times;

	float									m_total;
	float									m_mean;
	float									m_std;
	float									m_min;
	float									m_max;

	bool									m_bStatsDirty;

	string									m_strName;

public:

	AR4Profile()
	{
		m_total = m_mean = m_std = m_min = m_max = 0.f;
		m_bStatsDirty = false;
	}

	void setName(string strName)			{ m_strName = strName; }
	string getName()						{ return m_strName; }

	void start()							{ m_timer.getDeltaValue(); }
	void stop()
	{
		m_times.push_back(m_timer.getDeltaValue());
		m_bStatsDirty = true;
	}

	void reset()							{ m_times.clear(); }
	void collectStats()
	{
		card32 i;
		float ft;
		m_total = m_mean = m_std = 0.f;
		m_min = 1e30f;
		m_max = 0.f;

		for (i = 0; i < m_times.size(); i++)
		{
			ft = (float)m_times[i];
			m_total += ft;
			if (ft > m_max)
				m_max = ft;
			if (ft < m_min)
				m_min = ft;
		}

		m_mean = m_total / (float)m_times.size();

		for (i = 0; i < m_times.size(); i++)
		{
			ft = (float)m_times[i];
			ft -= m_mean;
			m_std += ft*ft;
		}

		if (m_times.size() > 1)
			m_std /= (float)(m_times.size() - 1);
		m_std = sqrt(m_std);

		m_bStatsDirty = false;
	}

	void reportStats()
	{
		if (m_bStatsDirty)
			collectStats();

		debugOutput << "Reporting profile statistics for " << m_strName << ":\n";
		debugOutput << "\tnumber runs: " << m_times.size() << "\n";
		debugOutput << "\ttotal run-time: " << convertTimeToStringFloat(m_total) << "\n";
		debugOutput << "\tmean run-time: " << convertTimeToStringFloat(m_mean) << "\n";
		debugOutput << "\tstd. dev.: " << convertTimeToStringFloat(m_std) << "\n";
		debugOutput << "\tmin run-time: " << convertTimeToStringFloat(m_min) << "\n";
		debugOutput << "\tmax run-time: " << convertTimeToStringFloat(m_max) << "\n";
	}
};

}


#endif //__AR4PROFILE_H__


