///////////////////////////////////////////////////////////////////////////////
//
//	AR4SparseGeodesics.cpp
//
//	Source file for the AR4SparseGeodesics class and subclasses
//
//	Alan Brunton, December 2012
//
///////////////////////////////////////////////////////////////////////////////


#include "stdafx.h"

#include "AR4SparseGeodesics.h"

#include "TopologyAttachment.h"
#include "TopologicalKNNIterator.h"

#include "StringHelper.h"
#include "Timer.h"

#include "DynamicLinearAlgebra.h"

#include "Monomials2ndOrder.h"



#define AR4SPARSEGEODESICS_PATH_ASTAR		0
#define AR4SPARSEGEODESICS_RELAX_SOLVER		1
#define AR4SPARSEGEODESICS_TINY_EDGE_FACTOR	(1e-3)

#define AR4SPARSEGEODESICS_PROFILE			0

#if AR4SPARSEGEODESICS_PROFILE
#define AR4SG_START_PROFILE(profile) profile.start();
#define AR4SG_STOP_PROFILE(profile)	profile.stop();
#else
#define AR4SG_START_PROFILE(profile)
#define AR4SG_STOP_PROFILE(profile)
#endif //AR4SPARSEGEODESICS_PROFILE


//utility
static X4::Matrix3f multTwoVectors(X4::Vector3f v1, X4::Vector3f v2)
{
   X4::Matrix3f m;
   m[0][0] = v1[0] * v2[0];
   m[0][1] = v1[1] * v2[0];
   m[0][2] = v1[2] * v2[0];

   m[1][0] = v1[0] * v2[1];
   m[1][1] = v1[1] * v2[1];
   m[1][2] = v1[2] * v2[1];

   m[2][0] = v1[0] * v2[2];
   m[2][1] = v1[1] * v2[2];
   m[2][2] = v1[2] * v2[2];

   return m;   
}


//initialize static member variables
const float X4::AR4SparseGeodesics::m_HugeDistance	= 1e30f;


void X4::AR4SparseGeodesics::computeGeodesicsEarlyStop(float sigma, const mpcard startIndex, const vector<mpcard> &indices, vector<float>& distances, vector<Vector3f> &derivates, vector<vector<mpcard>>* paths/* = NULL*/)
{
	AR4SG_START_PROFILE(m_profEarlyStop)

	card32 i, j;
	TopologyIterator * tIter = dynamic_cast<TopologyIterator*>(m_pPCIt);
	x4Assert( tIter );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(m_pPCIt);
	RandomAccessPointCloudIterator * rIt = dynamic_cast<RandomAccessPointCloudIterator*>(m_pPCIt);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(m_pPCIt);
	AAT posAAT = m_pPtCloud->getAAT( "position" );

	SceneObjectIterator * iter = m_pPtCloud->createIterator( SceneObjectIteratorType::CAP_BASIC_PC | 
		SceneObjectIteratorType::CAP_RANDOM_ACCESS_PC );
	x4Assert( iter != NULL );
	BasicPointCloudIterator * neighbour_bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	RandomAccessPointCloudIterator * neighbour_rIt = dynamic_cast<RandomAccessPointCloudIterator*>(iter);

	const float rad_2 = pow(sigma * 2.0f, 2);
	const mpcard numMaxPtsGrad = m_nMaxPtsGradEst;

	// -----------------------------------------
	// distance transformation for the complete point cloud
	const card32 pred_start_pred = m_pPtCloud->getNumPoints();
	const card32 pred_not_visited = pred_start_pred + 1;
	const card32 pred_not_visited_must_visit = pred_not_visited + 1;
	fill(m_distanceField.begin(), m_distanceField.end(), m_HugeDistance);
	fill(m_predecessorField.begin(), m_predecessorField.end(), pred_not_visited);

	derivates.resize(indices.size());
	distances.resize(indices.size());

	if (indices.size() == 0)
	{
		AR4SG_STOP_PROFILE(m_profEarlyStop)
		return;
	}

	AR4SG_START_PROFILE(m_profPropagate)

	priority_queue<IndexDist> indexQueue;

	card32 numNeighbors;
	mpcard dest, nbr, numMustVisit = 0;
	for (i = 0; i < indices.size(); i++)
	{
		dest = indices[i];
		if (m_predecessorField[dest] != pred_not_visited_must_visit)
		{
			m_predecessorField[dest] = pred_not_visited_must_visit;
			numMustVisit++;
		}

		rIt->goToIndex(dest);
		numNeighbors = tIter->getNumNeighbours();
		for (j = 0; j < numNeighbors; j++)
		{
			nbr = tIter->getNeighbour(j);
			if (m_predecessorField[nbr] == pred_not_visited_must_visit)
				continue;
			m_predecessorField[nbr] = pred_not_visited_must_visit;
			numMustVisit++;
		}
	}

#if AR4SPARSEGEODESICS_PATH_ASTAR
	vector<Vector3f> end_positions(indices.size());
	for (j = 0; j < end_positions.size(); j++)
	{
		rIt->goToIndex(indices[j]);
		end_positions[j] = bIt->get3f(posAAT);
	}

	Vector3f start_pos;
	float hstart;
	rIt->goToIndex(startIndex);
	start_pos = bIt->get3f(posAAT);
	hstart = (start_pos - end_positions[0]).getNorm();
	for (j = 1; j < end_positions.size(); j++)
		hstart = min(hstart, (start_pos - end_positions[j]).getNorm());
	indexQueue.push(IndexDist(startIndex, hstart));
#else
	indexQueue.push(IndexDist(startIndex, 0.f, pred_start_pred));
#endif //AR4SPARSEGEODESICS_PATH_ASTAR
	m_distanceField[startIndex] = 0.0f;

	while (!indexQueue.empty() && numMustVisit > 0)
	{
		PointIndex index;
		float currentDistance;
		IndexDist inddist = indexQueue.top();
		indexQueue.pop();	// remove first element

		index = inddist.m_index;

		if (m_predecessorField[index] < pred_not_visited)
			continue;

		if (m_predecessorField[index] == pred_not_visited_must_visit)
			numMustVisit--;

		m_predecessorField[index] = inddist.m_pred;

		rIt->goToIndex( index );
		Vector3f pos = bIt->get3f(posAAT);
		currentDistance = m_distanceField[index];

		numNeighbors = tIter->getNumNeighbours();
		for (i = 0; i < numNeighbors; i++)
		{
			PointIndex neighbour = tIter->getNeighbour(i);
			if (m_predecessorField[neighbour] < pred_not_visited)
				continue;

			if (currentDistance < m_distanceField[neighbour])
			{
				neighbour_rIt->goToIndex(neighbour);
				Vector3f neighbour_pos = neighbour_bIt->get3f(posAAT);

				float distance = max(1e-20f, norm( neighbour_pos - pos )) + currentDistance;
				if (m_distanceField[neighbour] > distance)
				{
					m_distanceField[neighbour] = distance;
#if AR4SPARSEGEODESICS_PATH_ASTAR
					float h = (neighbour_pos - end_positions[0]).getNorm();
					for (j = 1; j < end_positions.size(); j++)
						h = min(h, (neighbour_pos - end_positions[j]).getNorm());
					indexQueue.push(IndexDist(neighbour, distance + h));
#else
					indexQueue.push(IndexDist(neighbour, distance, index));
#endif //AR4SPARSEGEODESICS_PATH_ASTAR
				}
			}
		}
	}

	AR4SG_STOP_PROFILE(m_profPropagate)

	if( paths != NULL )
	{
		paths->clear();
	}

	// get distances from distance field
	float inv_sigma_2 = 1.0f / sigma / sigma;
	for(i = 0; i < indices.size(); i++ )
	{
		PointIndex index = indices[i];
		rIt->goToIndex(index);
		Vector3f currentPoint = bIt->get3f(posAAT);

		// no path leads to this point
		if (m_predecessorField[index] >= pred_not_visited)
		{
			distances[i] = m_HugeDistance;
			derivates[i] = NULL_VECTOR3F;
			continue;
		}
      
		AR4SG_START_PROFILE(m_profGradient)

		// compute a tangent plane
		Vector3f center;
		Matrix3f frame;
		computeLocalFrame(currentPoint, sigma, m_pPtCloud, m_pHknnIt, frame, center);
		const Matrix3f worldToFrame = frame.transpose();

		// gather neighbor points and transform into tangent space
		// height is set to distance, ignores outliers with infinity distance
		m_pKNN->setTargetPoint(currentPoint);
		vector<Vector3f> tangentSpacePoints;
		card32 nextNeareset;
		bool bfound;

		//counter to limit the number of nearby points we need to estimate the gradient (performance consideration)
		mpcard gradEstCount = 0;
		m_pKNN->getNextNearestPointIndex(nextNeareset, bfound);
		while (bfound && gradEstCount < numMaxPtsGrad)
		{
			if (m_predecessorField[nextNeareset] >= pred_not_visited)
			{
				m_pKNN->getNextNearestPointIndex(nextNeareset, bfound);
				continue;
			}
			Vector3f tmpPos = m_pPtCloud->getPointSet()->get3f(nextNeareset, m_posAAT);
			float dist_2 = normQuad(tmpPos-currentPoint);
			if( dist_2 > rad_2)
				break;

			tmpPos = worldToFrame * (tmpPos-currentPoint);
			tmpPos[2] = m_distanceField[nextNeareset];
			tangentSpacePoints.push_back(tmpPos);

			m_pKNN->getNextNearestPointIndex(nextNeareset, bfound);
			gradEstCount++;
		}

		// compute first order fit
		Vector3f coeffs;
		Monomials1stOrder::computeCoeffs(tangentSpacePoints, coeffs, sigma*sigma);
		
		// set slightly more precise distance
		distances[i] = m_distanceField[index];//coeffs[0];

		// compute derivates
		Vector3f tangentSpaceDerivate = makeVector3f(coeffs[1],coeffs[2],0);
		derivates[i] = frame * tangentSpaceDerivate;
		derivates[i].normalize();

		AR4SG_STOP_PROFILE(m_profGradient)

		if( paths != NULL )
		{
			AR4SG_START_PROFILE(m_profBackTrack)

			vector<mpcard> path;
			while( index != startIndex )
			{
				rIt->goToIndex( index );
				path.push_back(index);
				if( path.size() > 50000 )
					break;

				index = m_predecessorField[index];
				if (index >= pred_not_visited)
					break;
			}

			path.push_back(startIndex);
			reverse(path.begin(), path.end());
			paths->push_back( path );

			AR4SG_STOP_PROFILE(m_profBackTrack)
		}
	}

	delete iter;

	AR4SG_STOP_PROFILE(m_profEarlyStop);
}

void X4::AR4SparseGeodesics::computeGeodesicsEarlyStop(float sigma, const mpcard startIndex, const vector<mpcard> &indices, vector<float>& distances, vector<Vector3f> &derivates, vector<vector<Vector3f>>* paths/* = NULL*/)
{
	AR4SG_START_PROFILE(m_profEarlyStop)

	card32 i, j;

	TopologyIterator * tIter = dynamic_cast<TopologyIterator*>(m_pPCIt);
	x4Assert( tIter );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(m_pPCIt);
	RandomAccessPointCloudIterator * rIt = dynamic_cast<RandomAccessPointCloudIterator*>(m_pPCIt);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(m_pPCIt);
	AAT posAAT = m_pPtCloud->getAAT( "position" );

	SceneObjectIterator * iter = m_pPtCloud->createIterator( SceneObjectIteratorType::CAP_BASIC_PC | 
		SceneObjectIteratorType::CAP_RANDOM_ACCESS_PC );
	x4Assert( iter != NULL );
	BasicPointCloudIterator * neighbour_bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	RandomAccessPointCloudIterator * neighbour_rIt = dynamic_cast<RandomAccessPointCloudIterator*>(iter);

	PointSet* ps = m_pPtCloud->getPointSet();

	const float rad_2 = pow(sigma * 2.0f,2.0f);
	const mpcard numMaxPtsGrad = m_nMaxPtsGradEst;

	// -----------------------------------------
	// distance transformation for the complete point cloud
	const card32 pred_start_pred = m_pPtCloud->getNumPoints();
	const card32 pred_not_visited = pred_start_pred + 1;
	const card32 pred_not_visited_must_visit = pred_not_visited + 1;
	fill(m_distanceField.begin(), m_distanceField.end(), m_HugeDistance);
	fill(m_predecessorField.begin(), m_predecessorField.end(), pred_not_visited);

	derivates.resize(indices.size());
	distances.resize(indices.size());

	if (indices.size() == 0)
	{
		AR4SG_STOP_PROFILE(m_profEarlyStop)
		return;
	}

	AR4SG_START_PROFILE(m_profPropagate)

	priority_queue<IndexDist> indexQueue;

	card32 numNeighbors;
	mpcard dest, nbr, numMustVisit = 0;
	for (i = 0; i < indices.size(); i++)
	{
		dest = indices[i];
		if (m_predecessorField[dest] != pred_not_visited_must_visit)
		{
			m_predecessorField[dest] = pred_not_visited_must_visit;
			numMustVisit++;
		}

		rIt->goToIndex(dest);
		numNeighbors = tIter->getNumNeighbours();
		for (j = 0; j < numNeighbors; j++)
		{
			nbr = tIter->getNeighbour(j);
			if (m_predecessorField[nbr] == pred_not_visited_must_visit)
				continue;
			m_predecessorField[nbr] = pred_not_visited_must_visit;
			numMustVisit++;
		}
	}

#if AR4SPARSEGEODESICS_PATH_ASTAR
	vector<Vector3f> end_positions(indices.size());
	for (j = 0; j < end_positions.size(); j++)
	{
		rIt->goToIndex(indices[j]);
		end_positions[j] = bIt->get3f(posAAT);
	}

	Vector3f start_pos;
	float hstart;
	rIt->goToIndex(startIndex);
	start_pos = bIt->get3f(posAAT);
	hstart = (start_pos - end_positions[0]).getNorm();
	for (j = 1; j < end_positions.size(); j++)
		hstart = min(hstart, (start_pos - end_positions[j]).getNorm());
	indexQueue.push(IndexDist(startIndex, hstart));
#else
	indexQueue.push(IndexDist(startIndex, 0.f, pred_start_pred));
#endif //AR4SPARSEGEODESICS_PATH_ASTAR
	m_distanceField[startIndex] = 0.0f;

	while (!indexQueue.empty() && numMustVisit > 0)
	{
		PointIndex index;
		float currentDistance;
		IndexDist inddist = indexQueue.top();
		indexQueue.pop();	// remove first element

		index = inddist.m_index;

		if (m_predecessorField[index] < pred_not_visited)
			continue;

		if (m_predecessorField[index] == pred_not_visited_must_visit)
			numMustVisit--;

		m_predecessorField[index] = inddist.m_pred;

		rIt->goToIndex( index );
		Vector3f pos = bIt->get3f(posAAT);
		currentDistance = m_distanceField[index];

		numNeighbors = tIter->getNumNeighbours();
		for (i = 0; i < numNeighbors; i++)
		{
			PointIndex neighbour = tIter->getNeighbour(i);
			if (m_predecessorField[neighbour] < pred_not_visited)
				continue;

			if (currentDistance < m_distanceField[neighbour])
			{
				neighbour_rIt->goToIndex(neighbour);
				Vector3f neighbour_pos = neighbour_bIt->get3f(posAAT);

				float distance = max(1e-20f, norm( neighbour_pos - pos )) + currentDistance;
				if (m_distanceField[neighbour] > distance)
				{
					m_distanceField[neighbour] = distance;
#if AR4SPARSEGEODESICS_PATH_ASTAR
					float h = (neighbour_pos - end_positions[0]).getNorm();
					for (j = 1; j < end_positions.size(); j++)
						h = min(h, (neighbour_pos - end_positions[j]).getNorm());
					indexQueue.push(IndexDist(neighbour, distance + h));
#else
					indexQueue.push(IndexDist(neighbour, distance, index));
#endif //AR4SPARSEGEODESICS_PATH_ASTAR
				}
			}
		}
	}

	AR4SG_STOP_PROFILE(m_profPropagate)

	if( paths != NULL )
	{
		paths->clear();
	}

	// get distances from distance field
	float inv_sigma_2 = 1.0f / sigma / sigma;
	for(i = 0; i < indices.size(); i++ )
	{
		PointIndex index = indices[i];
		rIt->goToIndex(index);
		Vector3f currentPoint = bIt->get3f(posAAT);

		// no path leads to this point
		if (m_predecessorField[index] >= pred_not_visited)
		{
			distances[i] = m_HugeDistance;
			derivates[i] = NULL_VECTOR3F;
			continue;
		}

		if (sigma > 0.0)
		{
			AR4SG_START_PROFILE(m_profGradient)

			// compute a tangent plane
			Vector3f center;
			Matrix3f frame;
			computeLocalFrame(currentPoint, sigma, m_pPtCloud, m_pHknnIt, frame, center);
			const Matrix3f worldToFrame = frame.transpose();

			// gather neighbor points and transform into tangent space
			// height is set to distance, ignores outliers with infinity distance
			m_pKNN->setTargetPoint(currentPoint);
			vector<Vector3f> tangentSpacePoints;
			card32 nextNearest;
			bool bfound;

			mpcard gradEstCount = 0;
			m_pKNN->getNextNearestPointIndex(nextNearest, bfound);
			while (bfound && gradEstCount < numMaxPtsGrad)
			{
				if (m_predecessorField[nextNearest] >= pred_not_visited)
				{
					m_pKNN->getNextNearestPointIndex(nextNearest, bfound);
					continue;
				}
				Vector3f tmpPos = m_pPtCloud->getPointSet()->get3f(nextNearest, m_posAAT);
				float dist_2 = normQuad(tmpPos-currentPoint);
				if( dist_2 > rad_2)
					break;

				tmpPos = worldToFrame * (tmpPos-currentPoint);
				tmpPos[2] = m_distanceField[nextNearest];
				tangentSpacePoints.push_back(tmpPos);

				m_pKNN->getNextNearestPointIndex(nextNearest, bfound);
				gradEstCount++;
			}

			// compute first order fit
			Vector3f coeffs;
			Monomials1stOrder::computeCoeffs(tangentSpacePoints, coeffs, sigma*sigma);
			
			// set slightly more precise distance
			distances[i] = m_distanceField[index];//coeffs[0];

			// compute derivates
			Vector3f tangentSpaceDerivate = makeVector3f(coeffs[1],coeffs[2],0);
			derivates[i] = frame * tangentSpaceDerivate;
			derivates[i].normalize();

			AR4SG_STOP_PROFILE(m_profGradient)
		}
      
		if( paths != NULL )
		{
			AR4SG_START_PROFILE(m_profBackTrack)

			vector<Vector3f> path;

			while( index != startIndex )
			{
				rIt->goToIndex( index );
				currentPoint = bIt->get3f(posAAT);
				path.push_back(currentPoint);

				if( path.size() > 50000 )
					break;

				index = m_predecessorField[index];
				if (index >= pred_not_visited)
					break;
			}

			rIt->goToIndex(startIndex);
			path.push_back(bIt->get3f(posAAT));
			reverse(path.begin(), path.end());
			paths->push_back( path );

			AR4SG_STOP_PROFILE(m_profBackTrack)
		}
	}

	delete iter;

	AR4SG_STOP_PROFILE(m_profEarlyStop)
}






void X4::AR4SparseGeodesics::computeGeodesicsEarlyStop(float sigma, const Vector3f& startPoint, const vector<Vector3f> &points, vector<float>& dists, vector<Vector3f> &derivates, vector<vector<Vector3f>>* paths/* = NULL*/)
{
	mpcard i, j, n = points.size();
	mpcard startIndex;
	vector<mpcard> endIndices(n);
	vector<vector<mpcard>> pathIndices;
	AAT pos = m_pPtCloud->getAAT("position");
	PointSet* ptset = m_pPtCloud->getPointSet();

	if (n == 0)
		return;

	m_pKNN->setTargetPoint(startPoint);
	card32 nextNearest;
	bool bfound;
	m_pKNN->getNextNearestPointIndex(nextNearest, bfound);
	startIndex = nextNearest;

	for (i = 0; i < n; i++)
	{
		m_pKNN->setTargetPoint(points[i]);
		m_pKNN->getNextNearestPointIndex(nextNearest, bfound);
		endIndices[i] = nextNearest;
	}

	if (paths == NULL)
	{
		computeGeodesicsEarlyStop(sigma, startIndex, endIndices, dists, derivates);

		for (i = 0; i < n; i++)
			dists[i] += (startPoint - ptset->get3f(startIndex, pos)).getNorm() + (points[i] - ptset->get3f(endIndices[i], pos)).getNorm();
	}
	else
	{
		computeGeodesicsEarlyStop(sigma, startIndex, endIndices, dists, derivates, &pathIndices);

		const float linkThresh = m_avgEdgeLength * AR4SPARSEGEODESICS_TINY_EDGE_FACTOR;

		paths->resize(n);
		for (i = 0; i < n; i++)
		{
			const Vector3f startLink = startPoint - ptset->get3f(startIndex, pos);
			const Vector3f endLink = points[i] - ptset->get3f(endIndices[i], pos);
			const float startLinkNorm = startLink.getNorm();
			const float endLinkNorm = endLink.getNorm();

			mpcard pathOffset = 0, pathCount = pathIndices[i].size();

			if (startLinkNorm > linkThresh)
			{
				pathOffset = 1;
				pathCount++;
				dists[i] += startLinkNorm;
			}
			if (endLinkNorm > linkThresh)
			{
				pathCount++;
				dists[i] += endLinkNorm;
			}

			(*paths)[i].resize(pathCount);
			for (j = 0; j < pathIndices[i].size(); j++)
				(*paths)[i][j + pathOffset] = ptset->get3f(pathIndices[i][j], pos);
			(*paths)[i][0] = startPoint;
			(*paths)[i][pathCount - 1] = points[i];
		}
	}
}

void X4::AR4SparseGeodesics::computeGeodesicSingleDest(float sigma, const mpcard startIndex, const mpcard endIndex, float& dist, Vector3f& derivative, vector<mpcard>* path/* = NULL*/)
{
	AR4SG_START_PROFILE(m_profSingleDest)

	card32 i, j;

	TopologyIterator * tIter = dynamic_cast<TopologyIterator*>(m_pPCIt);
	x4Assert( tIter );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(m_pPCIt);
	RandomAccessPointCloudIterator * rIt = dynamic_cast<RandomAccessPointCloudIterator*>(m_pPCIt);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(m_pPCIt);
	AAT posAAT = m_pPtCloud->getAAT( "position" );

	SceneObjectIterator * iter = m_pPtCloud->createIterator( SceneObjectIteratorType::CAP_BASIC_PC | 
		SceneObjectIteratorType::CAP_RANDOM_ACCESS_PC );
	x4Assert( iter != NULL );
	BasicPointCloudIterator * neighbour_bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	RandomAccessPointCloudIterator * neighbour_rIt = dynamic_cast<RandomAccessPointCloudIterator*>(iter);

	PointSet* ps = m_pPtCloud->getPointSet();

	const float rad_2 = pow(sigma * 2.0f,2.0f);
	const mpcard numMaxPtsGrad = m_nMaxPtsGradEst;

	float heurDistRem;//heuristic (Euclidean distance) for distance remaining to end node

	// -----------------------------------------
	// distance transformation for the complete point cloud
	const card32 pred_start_pred = m_pPtCloud->getNumPoints();
	const card32 pred_not_visited = pred_start_pred + 1;
	const card32 pred_not_visited_must_visit = pred_not_visited + 1;
	fill(m_distanceField.begin(), m_distanceField.end(), m_HugeDistance);
	fill(m_predecessorField.begin(), m_predecessorField.end(), pred_not_visited);

	priority_queue<IndexDist> indexQueue;
	card32 numNeighbors;

	mpcard dest, nbr, numMustVisit = 0;

	m_predecessorField[endIndex] = pred_not_visited_must_visit;
	numMustVisit++;
	rIt->goToIndex(endIndex);
	numNeighbors = tIter->getNumNeighbours();
	for (j = 0; j < numNeighbors; j++)
	{
		nbr = tIter->getNeighbour(j);
		if (m_predecessorField[nbr] == pred_not_visited_must_visit)
			continue;
		m_predecessorField[nbr] = pred_not_visited_must_visit;
		numMustVisit++;
	}

	const Vector3f end_pos = m_pPtCloud->getPointSet()->get3f(endIndex, posAAT);
	heurDistRem = norm(m_pPtCloud->getPointSet()->get3f(startIndex, posAAT) - end_pos);

	AR4SG_START_PROFILE(m_profPropagate)

	indexQueue.push(IndexDist(startIndex, heurDistRem, pred_start_pred));
	m_distanceField[startIndex] = 0.f;

	while( !indexQueue.empty() && numMustVisit > 0)
	{
		PointIndex index;
		float currentDistance;
		IndexDist inddist = indexQueue.top();
		indexQueue.pop();	// remove first element

		index = inddist.m_index;

		if (m_predecessorField[index] < pred_not_visited)
			continue;

		if (m_predecessorField[index] == pred_not_visited_must_visit)
			numMustVisit--;

		m_predecessorField[index] = inddist.m_pred;

		rIt->goToIndex( index );
		Vector3f pos = bIt->get3f(posAAT);
		currentDistance = m_distanceField[index];

		numNeighbors = tIter->getNumNeighbours();
		for(i = 0; i < numNeighbors; i++)
		{
			const PointIndex neighbour = tIter->getNeighbour(i);
			if (m_predecessorField[neighbour] < pred_not_visited)
				continue;

			if (currentDistance < m_distanceField[neighbour])
			{
				neighbour_rIt->goToIndex(neighbour);
				const Vector3f neighbour_pos = neighbour_bIt->get3f(posAAT);

				heurDistRem = norm(neighbour_pos - end_pos);

				float distance = max(1e-20f, norm( neighbour_pos - pos )) + currentDistance;
				if (m_distanceField[neighbour] > distance )
				{
					m_distanceField[neighbour] = distance;
					indexQueue.push(IndexDist(neighbour, distance + heurDistRem, index));
				}
			}
		}
	}

	AR4SG_STOP_PROFILE(m_profPropagate)

	if( path != NULL )
	{
		path->clear();
	}

	// get distances from distance field
	float inv_sigma_2 = 1.0f / sigma / sigma;

	{
		PointIndex index = endIndex;
		rIt->goToIndex(index);
		const Vector3f currentPoint = bIt->get3f(posAAT);

		// no path leads to this point
		if (m_predecessorField[index] >= pred_not_visited)
		{
			dist = m_HugeDistance;
			derivative = NULL_VECTOR3F;
			return;
		}
      
		AR4SG_START_PROFILE(m_profGradient)

		// compute a tangent plane
		Vector3f center;
		Matrix3f frame;
		computeLocalFrame(currentPoint, sigma, m_pPtCloud, m_pHknnIt, frame, center);
		const Matrix3f worldToFrame = frame.transpose();

		// gather neighbor points and transform into tangent space
		// height is set to distance, ignores outliers with infinity distance
		m_pKNN->setTargetPoint(currentPoint);
		vector<Vector3f> tangentSpacePoints;
		card32 nextNearest;
		bool bfound;

		mpcard gradEstCount = 0;
		m_pKNN->getNextNearestPointIndex(nextNearest, bfound);
		while (bfound && gradEstCount < numMaxPtsGrad)
		{
			if (m_predecessorField[nextNearest] >= pred_not_visited)
			{
				m_pKNN->getNextNearestPointIndex(nextNearest, bfound);
				continue;
			}
			Vector3f tmpPos = m_pPtCloud->getPointSet()->get3f(nextNearest, m_posAAT);
			float dist_2 = normQuad(tmpPos-currentPoint);
			if( dist_2 > rad_2)
				break;

			tmpPos = worldToFrame * (tmpPos-currentPoint);
			tmpPos[2] = m_distanceField[nextNearest];
			tangentSpacePoints.push_back(tmpPos);

			m_pKNN->getNextNearestPointIndex(nextNearest, bfound);
			gradEstCount++;
		}

		// compute first order fit
		Vector3f coeffs;
		Monomials1stOrder::computeCoeffs(tangentSpacePoints, coeffs, sigma*sigma);
		
		// set slightly more precise distance
		dist = m_distanceField[index];//coeffs[0];

		// compute derivates
		Vector3f tangentSpaceDerivate = makeVector3f(coeffs[1],coeffs[2],0);
		derivative = frame * tangentSpaceDerivate;
		derivative.normalize();

		AR4SG_STOP_PROFILE(m_profGradient)
      
		if( path != NULL )
		{
			AR4SG_START_PROFILE(m_profBackTrack)

			while( index != startIndex )
			{
				rIt->goToIndex( index );
				path->push_back(index);

				if( path->size() > 50000 )
					break;

				index = m_predecessorField[index];
				if (index >= pred_not_visited)
					break;
			}

			path->push_back(startIndex);
			reverse(path->begin(), path->end());

			AR4SG_STOP_PROFILE(m_profBackTrack)
		}
	}

	delete iter;

	AR4SG_STOP_PROFILE(m_profSingleDest)
}

void X4::AR4SparseGeodesics::computeGeodesicSingleDest(float sigma, const mpcard startIndex, const mpcard endIndex, float& dist, Vector3f& derivative, vector<Vector3f>& path)
{
	AR4SG_START_PROFILE(m_profSingleDest)

	card32 i, j;

	TopologyIterator * tIter = dynamic_cast<TopologyIterator*>(m_pPCIt);
	x4Assert( tIter );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(m_pPCIt);
	RandomAccessPointCloudIterator * rIt = dynamic_cast<RandomAccessPointCloudIterator*>(m_pPCIt);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(m_pPCIt);
	AAT posAAT = m_pPtCloud->getAAT( "position" );

	SceneObjectIterator * iter = m_pPtCloud->createIterator( SceneObjectIteratorType::CAP_BASIC_PC | 
		SceneObjectIteratorType::CAP_RANDOM_ACCESS_PC );
	x4Assert( iter != NULL );
	BasicPointCloudIterator * neighbour_bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	RandomAccessPointCloudIterator * neighbour_rIt = dynamic_cast<RandomAccessPointCloudIterator*>(iter);

	PointSet* ps = m_pPtCloud->getPointSet();

	const float rad_2 = pow(sigma * 2.0f,2.0f);
	const mpcard numMaxPtsGrad = m_nMaxPtsGradEst;

	Vector3f end_pos;
	float heurDistRem;//heuristic (Euclidean distance) for distance remaining to end node

	// -----------------------------------------
	// distance transformation for the complete point cloud
	const card32 pred_start_pred = m_pPtCloud->getNumPoints();
	const card32 pred_not_visited = pred_start_pred + 1;
	const card32 pred_not_visited_must_visit = pred_not_visited + 1;
	fill(m_distanceField.begin(), m_distanceField.end(), m_HugeDistance);
	fill(m_predecessorField.begin(), m_predecessorField.end(), pred_not_visited);

	priority_queue<IndexDist> indexQueue;
	card32 numNeighbors;

	mpcard dest, nbr, numMustVisit = 0;

	m_predecessorField[endIndex] = pred_not_visited_must_visit;
	numMustVisit++;
	rIt->goToIndex(endIndex);
	numNeighbors = tIter->getNumNeighbours();
	for (j = 0; j < numNeighbors; j++)
	{
		nbr = tIter->getNeighbour(j);
		if (m_predecessorField[nbr] == pred_not_visited_must_visit)
			continue;
		m_predecessorField[nbr] = pred_not_visited_must_visit;
		numMustVisit++;
	}

	end_pos = m_pPtCloud->getPointSet()->get3f(endIndex, posAAT);
	heurDistRem = norm(m_pPtCloud->getPointSet()->get3f(startIndex, posAAT) - end_pos);

	AR4SG_START_PROFILE(m_profPropagate)

	indexQueue.push(IndexDist(startIndex, heurDistRem, pred_not_visited));
	m_distanceField[startIndex] = 0.f;

	while( !indexQueue.empty() && numMustVisit > 0)
	{
		PointIndex index;
		float currentDistance;
		IndexDist inddist = indexQueue.top();
		indexQueue.pop();	// remove first element

		index = inddist.m_index;

		if (m_predecessorField[index] < pred_not_visited)
			continue;

		if (m_predecessorField[index] == pred_not_visited_must_visit)
			numMustVisit--;

		m_predecessorField[index] = inddist.m_pred;

		rIt->goToIndex( index );
		Vector3f pos = bIt->get3f(posAAT);
		currentDistance = m_distanceField[index];

		numNeighbors = tIter->getNumNeighbours();
		for (i = 0; i < numNeighbors; i++)
		{
			PointIndex neighbour = tIter->getNeighbour(i);
			if (m_predecessorField[neighbour] < pred_not_visited)
				continue;

			if (currentDistance < m_distanceField[neighbour])
			{
				neighbour_rIt->goToIndex(neighbour);
				Vector3f neighbour_pos = neighbour_bIt->get3f(posAAT);

				heurDistRem = norm(neighbour_pos - end_pos);

				float distance = max(1e-20f, norm( neighbour_pos - pos )) + currentDistance;
				if (m_distanceField[neighbour] > distance )
				{
					m_distanceField[neighbour] = distance;
					indexQueue.push(IndexDist(neighbour, distance + heurDistRem, index));
				}
			}
		}
	}

	// get distances from distance field
	float inv_sigma_2 = 1.0f / sigma / sigma;

	AR4SG_STOP_PROFILE(m_profPropagate)

	path.clear();

	PointIndex index = endIndex;
	rIt->goToIndex(index);
	Vector3f currentPoint = bIt->get3f(posAAT);

	// no path leads to this point
	if (m_predecessorField[index] >= pred_not_visited)
	{
		dist = m_HugeDistance;
		derivative = NULL_VECTOR3F;
		return;
	}
  
	AR4SG_START_PROFILE(m_profGradient)

	// compute a tangent plane
	Vector3f center;
	Matrix3f frame;
	computeLocalFrame(currentPoint, sigma, m_pPtCloud, m_pHknnIt, frame, center);
	const Matrix3f worldToFrame = frame.transpose();

	// gather neighbor points and transform into tangent space
	// height is set to distance, ignores outliers with infinity distance
	m_pKNN->setTargetPoint(currentPoint);
	vector<Vector3f> tangentSpacePoints;
	card32 nextNearest;
	bool bfound;

	mpcard gradEstCount = 0;
	m_pKNN->getNextNearestPointIndex(nextNearest, bfound);
	while (bfound && gradEstCount < numMaxPtsGrad)
	{
		if (m_predecessorField[nextNearest] >= pred_not_visited)
		{
			m_pKNN->getNextNearestPointIndex(nextNearest, bfound);
			continue;
		}
		Vector3f tmpPos = m_pPtCloud->getPointSet()->get3f(nextNearest, m_posAAT);
		float dist_2 = normQuad(tmpPos-currentPoint);
		if( dist_2 > rad_2)
			break;

		tmpPos = worldToFrame * (tmpPos-currentPoint);
		tmpPos[2] = m_distanceField[nextNearest];
		tangentSpacePoints.push_back(tmpPos);

		m_pKNN->getNextNearestPointIndex(nextNearest, bfound);
		gradEstCount++;
	}

	// compute first order fit
	Vector3f coeffs;
	Monomials1stOrder::computeCoeffs(tangentSpacePoints, coeffs, sigma*sigma);
	
	// set slightly more precise distance
	dist = m_distanceField[index];//coeffs[0];

	// compute derivates
	Vector3f tangentSpaceDerivate = makeVector3f(coeffs[1],coeffs[2],0);
	derivative = frame * tangentSpaceDerivate;
	derivative.normalize();

	AR4SG_STOP_PROFILE(m_profGradient)
	AR4SG_START_PROFILE(m_profBackTrack)
  
	float distBack = 0.f;
	while( index != startIndex )
	{
		rIt->goToIndex( index );
		currentPoint = bIt->get3f(posAAT);
		path.push_back(currentPoint);
		if( path.size() > 50000 )
			break;

		index = m_predecessorField[index];
		if (index >= pred_not_visited)
			break;
	}

	rIt->goToIndex(startIndex);
	path.push_back(bIt->get3f(posAAT));
	reverse(path.begin(), path.end());

	AR4SG_STOP_PROFILE(m_profBackTrack)

	delete iter;

	AR4SG_STOP_PROFILE(m_profSingleDest)
}

void X4::AR4SparseGeodesics::computeGeodesicSingleDest(float sigma, const Vector3f& startPoint, const Vector3f& point, float& dist, Vector3f& derivative, vector<Vector3f>* path/* = NULL*/)
{
	mpcard i, j;
	mpcard startIndex, endIndex;
	vector<mpcard> pathIndices;
	AAT pos = m_pPtCloud->getAAT("position");
	PointSet* ptset = m_pPtCloud->getPointSet();

	card32 nextNearest;
	bool bfound;
	m_pKNN->setTargetPoint(startPoint);
	m_pKNN->getNextNearestPointIndex(nextNearest, bfound);
	startIndex = nextNearest;

	m_pKNN->setTargetPoint(point);
	m_pKNN->getNextNearestPointIndex(nextNearest, bfound);
	endIndex = nextNearest;

	if (path == NULL)
	{
		computeGeodesicSingleDest(sigma, startIndex, endIndex, dist, derivative);
		
		dist += (startPoint - ptset->get3f(startIndex, pos)).getNorm() + (point - ptset->get3f(endIndex, pos)).getNorm();
	}
	else
	{
		computeGeodesicSingleDest(sigma, startIndex, endIndex, dist, derivative, &pathIndices);

		const float linkThresh = m_avgEdgeLength * AR4SPARSEGEODESICS_TINY_EDGE_FACTOR;

		const Vector3f startLink = startPoint - ptset->get3f(startIndex, pos);
		const Vector3f endLink = point - ptset->get3f(endIndex, pos);
		const float startLinkNorm = startLink.getNorm();
		const float endLinkNorm = endLink.getNorm();

		mpcard pathOffset = 0, pathCount = pathIndices.size();

		if (startLinkNorm > linkThresh)
		{
			pathOffset = 1;
			pathCount++;
			dist += startLinkNorm;
		}
		if (endLinkNorm > linkThresh)
		{
			pathCount++;
			dist += endLinkNorm;
		}

		path->resize(pathCount);
		for (j = 0; j < pathIndices.size(); j++)
			(*path)[j + pathOffset] = ptset->get3f(pathIndices[j], pos);
		(*path)[0] = startPoint;
		(*path)[pathCount - 1] = point;
	}
}

void X4::AR4SparseGeodesics::computeMinGeodesicToSet(const mpcard startIndex, const vector<mpcard>& indices, float& distMin, mpcard& nearest)
{
	AR4SG_START_PROFILE(m_profEarlyStop)

	card32 i, j;

	TopologyIterator * tIter = dynamic_cast<TopologyIterator*>(m_pPCIt);
	x4Assert( tIter );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(m_pPCIt);
	RandomAccessPointCloudIterator * rIt = dynamic_cast<RandomAccessPointCloudIterator*>(m_pPCIt);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(m_pPCIt);
	AAT posAAT = m_pPtCloud->getAAT( "position" );

	SceneObjectIterator * iter = m_pPtCloud->createIterator( SceneObjectIteratorType::CAP_BASIC_PC | 
		SceneObjectIteratorType::CAP_RANDOM_ACCESS_PC );
	x4Assert( iter != NULL );
	BasicPointCloudIterator * neighbour_bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	RandomAccessPointCloudIterator * neighbour_rIt = dynamic_cast<RandomAccessPointCloudIterator*>(iter);

	PointSet* ps = m_pPtCloud->getPointSet();

	const mpcard numDests = indices.size();

	// -----------------------------------------
	// distance transformation for the complete point cloud
	const card32 pred_start_pred = m_pPtCloud->getNumPoints();
	const card32 pred_not_visited = pred_start_pred + 1;
	const card32 pred_not_visited_must_visit = pred_not_visited + 1;
	fill(m_distanceField.begin(), m_distanceField.end(), m_HugeDistance);
	fill(m_predecessorField.begin(), m_predecessorField.end(), pred_not_visited);

	if (indices.size() == 0)
	{
		AR4SG_STOP_PROFILE(m_profEarlyStop)
		return;
	}

	AR4SG_START_PROFILE(m_profPropagate)

	priority_queue<IndexDist> indexQueue;

	mpcard dest;
	for (i = 0; i < numDests; i++)
	{
		dest = indices[i];
		m_predecessorField[dest] = pred_not_visited_must_visit;
	}

	indexQueue.push(IndexDist(startIndex, 0.f, pred_start_pred));
	m_distanceField[startIndex] = 0.0f;

	while (!indexQueue.empty())
	{
		PointIndex index;
		float currentDistance;
		IndexDist inddist = indexQueue.top();
		indexQueue.pop();	// remove first element

		index = inddist.m_index;

		if (m_predecessorField[index] < pred_not_visited)
			continue;

		if (m_predecessorField[index] == pred_not_visited_must_visit)
		{
			distMin = m_distanceField[index];
			nearest = index;
			break;
		}

		m_predecessorField[index] = inddist.m_pred;

		rIt->goToIndex( index );
		Vector3f pos = bIt->get3f(posAAT);
		currentDistance = m_distanceField[index];

		const card32 numNeighbors = tIter->getNumNeighbours();
		for (i = 0; i < numNeighbors; i++)
		{
			const PointIndex neighbour = tIter->getNeighbour(i);
			if (m_predecessorField[neighbour] < pred_not_visited)
				continue;

			if (currentDistance < m_distanceField[neighbour])
			{
				neighbour_rIt->goToIndex(neighbour);
				Vector3f neighbour_pos = neighbour_bIt->get3f(posAAT);

				float distance = max(1e-20f, norm( neighbour_pos - pos )) + currentDistance;
				if (m_distanceField[neighbour] > distance)
				{
					m_distanceField[neighbour] = distance;
					indexQueue.push(IndexDist(neighbour, distance, index));
				}
			}
		}
	}

	AR4SG_STOP_PROFILE(m_profPropagate)
}

void X4::AR4SparseGeodesics::computeGeodesicsLimitRange(const mpcard startIndex, const float rangeMax, vector<mpcard>& destIndices, vector<float>& distances, vector<vector<Vector3f>>* pPaths)
{
	card32 i, j;

	TopologyIterator * tIter = dynamic_cast<TopologyIterator*>(m_pPCIt);
	x4Assert( tIter );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(m_pPCIt);
	RandomAccessPointCloudIterator * rIt = dynamic_cast<RandomAccessPointCloudIterator*>(m_pPCIt);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(m_pPCIt);
	AAT posAAT = m_pPtCloud->getAAT( "position" );

	SceneObjectIterator * iter = m_pPtCloud->createIterator( SceneObjectIteratorType::CAP_BASIC_PC | 
		SceneObjectIteratorType::CAP_RANDOM_ACCESS_PC );
	x4Assert( iter != NULL );
	BasicPointCloudIterator * neighbour_bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	RandomAccessPointCloudIterator * neighbour_rIt = dynamic_cast<RandomAccessPointCloudIterator*>(iter);

	PointSet* ps = m_pPtCloud->getPointSet();

	// -----------------------------------------
	// distance transformation for the complete point cloud
	const card32 pred_start_pred = m_pPtCloud->getNumPoints();
	const card32 pred_not_visited = pred_start_pred + 1;
	const card32 pred_not_visited_must_visit = pred_not_visited + 1;
	fill(m_distanceField.begin(), m_distanceField.end(), m_HugeDistance);
	fill(m_predecessorField.begin(), m_predecessorField.end(), pred_not_visited);

	destIndices.clear();
	distances.clear();

	AR4SG_START_PROFILE(m_profPropagate)

	priority_queue<IndexDist> indexQueue;

#if AR4SPARSEGEODESICS_PATH_ASTAR
	vector<Vector3f> end_positions(indices.size());
	for (j = 0; j < end_positions.size(); j++)
	{
		rIt->goToIndex(indices[j]);
		end_positions[j] = bIt->get3f(posAAT);
	}

	Vector3f start_pos;
	float hstart;
	rIt->goToIndex(startIndex);
	start_pos = bIt->get3f(posAAT);
	hstart = (start_pos - end_positions[0]).getNorm();
	for (j = 1; j < end_positions.size(); j++)
		hstart = min(hstart, (start_pos - end_positions[j]).getNorm());
	indexQueue.push(IndexDist(startIndex, hstart));
#else
	indexQueue.push(IndexDist(startIndex, 0.f, pred_start_pred));
#endif //AR4SPARSEGEODESICS_PATH_ASTAR
	m_distanceField[startIndex] = 0.0f;

	while (!indexQueue.empty())
	{
		PointIndex index;
		float currentDistance;
		IndexDist inddist = indexQueue.top();
		indexQueue.pop();	// remove first element

		index = inddist.m_index;
		if (m_predecessorField[index] < pred_not_visited)
			continue;
		m_predecessorField[index] = inddist.m_pred;

		////stopping criterion:
		////algorithm stops because no more neighbors are added to the queue (so, not immediately)
		////this way we get all neighbors (and triangles in the case of meshes) of points within the specified distance
		destIndices.push_back(index);
		if (inddist.m_dist > rangeMax)
			continue;

		
		rIt->goToIndex( index );
		Vector3f pos = bIt->get3f(posAAT);
		currentDistance = m_distanceField[index];

		const card32 numNeighbors = tIter->getNumNeighbours();
		for (i = 0; i < numNeighbors; i++)
		{
			const PointIndex neighbour = tIter->getNeighbour(i);
			if (m_predecessorField[neighbour] < pred_not_visited)
				continue;

			if (currentDistance < m_distanceField[neighbour])
			{
				neighbour_rIt->goToIndex(neighbour);
				Vector3f neighbour_pos = neighbour_bIt->get3f(posAAT);

				float distance = max(1e-20f, norm( neighbour_pos - pos )) + currentDistance;
				if (m_distanceField[neighbour] > distance)
				{
					m_distanceField[neighbour] = distance;
#if AR4SPARSEGEODESICS_PATH_ASTAR
					float h = (neighbour_pos - end_positions[0]).getNorm();
					for (j = 1; j < end_positions.size(); j++)
						h = min(h, (neighbour_pos - end_positions[j]).getNorm());
					indexQueue.push(IndexDist(neighbour, distance + h));
#else
					indexQueue.push(IndexDist(neighbour, distance, index));
#endif //AR4SPARSEGEODESICS_PATH_ASTAR
				}
			}
		}
	}

	AR4SG_STOP_PROFILE(m_profPropagate)

	// get distances from distance field
	const mpcard numDests = destIndices.size();
	distances.resize(numDests);
	if (pPaths != NULL)
		pPaths->resize(numDests);
	for(i = 0; i < numDests; i++ )
	{
		PointIndex index = destIndices[i];
		rIt->goToIndex(index);
		Vector3f currentPoint = bIt->get3f(posAAT);

		// no path leads to this point
		if (m_predecessorField[index] >= pred_not_visited)
		{
			distances[i] = m_HugeDistance;
			continue;
		}
      
		// set distance
		distances[i] = m_distanceField[index];
 
		if (pPaths != NULL)
		{
			AR4SG_START_PROFILE(m_profBackTrack)

			vector<Vector3f> path;

			while (index != startIndex)
			{
				rIt->goToIndex( index );
				currentPoint = bIt->get3f(posAAT);
				path.push_back(currentPoint);

				if( path.size() > 50000 )
					break;

				index = m_predecessorField[index];
				if (index >= pred_not_visited)
					break;
			}

			rIt->goToIndex(startIndex);
			path.push_back(bIt->get3f(posAAT));
			reverse(path.begin(), path.end());
			(*pPaths)[i] = path;

			AR4SG_STOP_PROFILE(m_profBackTrack)
		}
	}

	delete iter;
}

void X4::AR4SparseGeodesics::computeGeodesicsLimitRange(const float sigma, const mpcard startIndex, const float rangeMax, vector<mpcard>& destIndices, vector<float>& distances, vector<Vector3f>& gradients, vector<vector<Vector3f>>* pPaths)
{
	card32 i, j;

	TopologyIterator * tIter = dynamic_cast<TopologyIterator*>(m_pPCIt);
	x4Assert( tIter );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(m_pPCIt);
	RandomAccessPointCloudIterator * rIt = dynamic_cast<RandomAccessPointCloudIterator*>(m_pPCIt);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(m_pPCIt);
	AAT posAAT = m_pPtCloud->getAAT( "position" );

	SceneObjectIterator * iter = m_pPtCloud->createIterator( SceneObjectIteratorType::CAP_BASIC_PC | 
		SceneObjectIteratorType::CAP_RANDOM_ACCESS_PC );
	x4Assert( iter != NULL );
	BasicPointCloudIterator * neighbour_bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	RandomAccessPointCloudIterator * neighbour_rIt = dynamic_cast<RandomAccessPointCloudIterator*>(iter);

	PointSet* ps = m_pPtCloud->getPointSet();

	const float rad_2 = pow(sigma * 2.0f,2.0f);
	const mpcard numMaxPtsGrad = m_nMaxPtsGradEst;

	// -----------------------------------------
	// distance transformation for the complete point cloud
	const card32 pred_start_pred = m_pPtCloud->getNumPoints();
	const card32 pred_not_visited = pred_start_pred + 1;
	const card32 pred_not_visited_must_visit = pred_not_visited + 1;
	fill(m_distanceField.begin(), m_distanceField.end(), m_HugeDistance);
	fill(m_predecessorField.begin(), m_predecessorField.end(), pred_not_visited);

	destIndices.clear();
	distances.clear();

	AR4SG_START_PROFILE(m_profPropagate)

	priority_queue<IndexDist> indexQueue;

#if AR4SPARSEGEODESICS_PATH_ASTAR
	vector<Vector3f> end_positions(indices.size());
	for (j = 0; j < end_positions.size(); j++)
	{
		rIt->goToIndex(indices[j]);
		end_positions[j] = bIt->get3f(posAAT);
	}

	Vector3f start_pos;
	float hstart;
	rIt->goToIndex(startIndex);
	start_pos = bIt->get3f(posAAT);
	hstart = (start_pos - end_positions[0]).getNorm();
	for (j = 1; j < end_positions.size(); j++)
		hstart = min(hstart, (start_pos - end_positions[j]).getNorm());
	indexQueue.push(IndexDist(startIndex, hstart));
#else
	indexQueue.push(IndexDist(startIndex, 0.f, pred_start_pred));
#endif //AR4SPARSEGEODESICS_PATH_ASTAR
	m_distanceField[startIndex] = 0.0f;

	while (!indexQueue.empty())
	{
		PointIndex index;
		float currentDistance;
		IndexDist inddist = indexQueue.top();
		indexQueue.pop();	// remove first element

		index = inddist.m_index;

		if (m_predecessorField[index] < pred_not_visited)
			continue;
		m_predecessorField[index] = inddist.m_pred;

		//stopping criterion:
		//algorithm stops because no more neighbors are added to the queue (so, not immediately)
		//this way we get all neighbors (and triangles in the case of meshes) of points within the specified distance
		destIndices.push_back(index);
		if (inddist.m_dist > rangeMax)
			continue;

		rIt->goToIndex( index );
		Vector3f pos = bIt->get3f(posAAT);
		currentDistance = m_distanceField[index];

		const card32 numNeighbors = tIter->getNumNeighbours();
		for (i = 0; i < numNeighbors; i++)
		{
			const PointIndex neighbour = tIter->getNeighbour(i);
			if (m_predecessorField[neighbour] < pred_not_visited)
				continue;

			if (currentDistance < m_distanceField[neighbour])
			{
				neighbour_rIt->goToIndex(neighbour);
				Vector3f neighbour_pos = neighbour_bIt->get3f(posAAT);

				float distance = max(1e-20f, norm( neighbour_pos - pos )) + currentDistance;
				if (m_distanceField[neighbour] > distance)
				{
					m_distanceField[neighbour] = distance;
#if AR4SPARSEGEODESICS_PATH_ASTAR
					float h = (neighbour_pos - end_positions[0]).getNorm();
					for (j = 1; j < end_positions.size(); j++)
						h = min(h, (neighbour_pos - end_positions[j]).getNorm());
					indexQueue.push(IndexDist(neighbour, distance + h));
#else
					indexQueue.push(IndexDist(neighbour, distance, index));
#endif //AR4SPARSEGEODESICS_PATH_ASTAR
				}
			}
		}
	}

	AR4SG_STOP_PROFILE(m_profPropagate)

	// get distances from distance field
	const mpcard numDests = destIndices.size();
	distances.resize(numDests);
	gradients.resize(numDests);
	if (pPaths != NULL)
		pPaths->resize(numDests);
	for(i = 0; i < numDests; i++ )
	{
		PointIndex index = destIndices[i];
		rIt->goToIndex(index);
		Vector3f currentPoint = bIt->get3f(posAAT);

		// no path leads to this point
		if (m_predecessorField[index] >= pred_not_visited)
		{
			distances[i] = m_HugeDistance;
			continue;
		}
      
		AR4SG_START_PROFILE(m_profGradient)

		// compute a tangent plane
		Vector3f center;
		Matrix3f frame;
		computeLocalFrame(currentPoint, sigma, m_pPtCloud, m_pHknnIt, frame, center);
		const Matrix3f worldToFrame = frame.transpose();

		// gather neighbor points and transform into tangent space
		// height is set to distance, ignores outliers with infinity distance
		m_pKNN->setTargetPoint(currentPoint);
		vector<Vector3f> tangentSpacePoints;
		card32 nextNearest;
		bool bfound;

		mpcard gradEstCount = 0;
		m_pKNN->getNextNearestPointIndex(nextNearest, bfound);
		while (bfound && gradEstCount < numMaxPtsGrad)
		{
			if (m_predecessorField[nextNearest] >= pred_not_visited)
			{
				m_pKNN->getNextNearestPointIndex(nextNearest, bfound);
				continue;
			}
			Vector3f tmpPos = m_pPtCloud->getPointSet()->get3f(nextNearest, m_posAAT);
			float dist_2 = normQuad(tmpPos-currentPoint);
			if( dist_2 > rad_2)
				break;

			tmpPos = worldToFrame * (tmpPos-currentPoint);
			tmpPos[2] = m_distanceField[nextNearest];
			tangentSpacePoints.push_back(tmpPos);

			m_pKNN->getNextNearestPointIndex(nextNearest, bfound);
			gradEstCount++;
		}

		// compute first order fit
		Vector3f coeffs;
		Monomials1stOrder::computeCoeffs(tangentSpacePoints, coeffs, sigma*sigma);
		
		// compute derivates
		Vector3f tangentSpaceDerivate = makeVector3f(coeffs[1],coeffs[2],0);
		Vector3f derivative = frame * tangentSpaceDerivate;
		derivative.normalize();

		gradients[i] = derivative;

		AR4SG_STOP_PROFILE(m_profGradient)

		// set distance
		distances[i] = m_distanceField[index];
 
		if (pPaths != NULL)
		{
			AR4SG_START_PROFILE(m_profBackTrack)

			vector<Vector3f> path;

			while (index != startIndex)
			{
				rIt->goToIndex( index );
				currentPoint = bIt->get3f(posAAT);
				path.push_back(currentPoint);

				if( path.size() > 50000 )
					break;

				index = m_predecessorField[index];
				if (index >= pred_not_visited)
					break;
			}

			rIt->goToIndex(startIndex);
			path.push_back(bIt->get3f(posAAT));
			reverse(path.begin(), path.end());
			(*pPaths)[i] = path;

			AR4SG_STOP_PROFILE(m_profBackTrack)
		}
	}

	delete iter;
}

void X4::AR4SparseGeodesics::computeGeodesicsLimitRangeRelax(const mpcard startIndex, const float rangeMax, const mpcard numIters, vector<mpcard>& destIndices, vector<float>& distances, vector<Vector3f>& gradients, vector<Vector3f>& dirsAtSource, vector<vector<Vector3f>>& paths)
{
	computeGeodesicsLimitRange(startIndex, rangeMax, destIndices, distances, &paths);

	gradients.resize(distances.size(), NULL_VECTOR3F);

	relaxPaths(numIters, paths, distances, gradients, dirsAtSource);
}

void X4::AR4SparseGeodesics::computeMinWeightPath(InCorePCTopologyGraph* ptg, const mpcard startIndex, const mpcard endIndex, float& dist, vector<mpcard>& path)
{
	PointSet* pEdges = ptg->getEdges();
	if (!pEdges->providesAttribute("weight"))
		throw EInvalidState("AR4SparseGeodesics::computeMinWeightPath(...): topology must provide weight attribute for edges");
	AAT weightAAT = pEdges->getAAT("weight");
	InCorePCTopologyGraph::EdgeIndex ei;

	card32 i, j;

	PointSet* ps = m_pPtCloud->getPointSet();

	Vector3f end_pos;
	float heurDistRem;//heuristic (Euclidean distance) for distance remaining to end node

	AAT posAAT = m_pPtCloud->getAAT("position");

	// -----------------------------------------
	// distance transformation for the complete point cloud
	const card32 pred_start_pred = m_pPtCloud->getNumPoints();
	const card32 pred_not_visited = pred_start_pred + 1;
	const card32 pred_not_visited_must_visit = pred_not_visited + 1;
	fill(m_distanceField.begin(), m_distanceField.end(), m_HugeDistance);
	fill(m_predecessorField.begin(), m_predecessorField.end(), pred_not_visited);

	priority_queue<IndexDist> indexQueue;
	card32 numNeighbors;

	mpcard dest, nbr, numMustVisit = 0;

	m_predecessorField[endIndex] = pred_not_visited_must_visit;
	numMustVisit++;
	numNeighbors = ptg->getNumAdjacentVertices(endIndex);
	for (j = 0; j < numNeighbors; j++)
	{
		nbr = ptg->getAdjacentVertex(endIndex, j);
		if (m_predecessorField[nbr] == pred_not_visited_must_visit)
			continue;
		m_predecessorField[nbr] = pred_not_visited_must_visit;
		numMustVisit++;
	}

	end_pos = m_pPtCloud->getPointSet()->get3f(endIndex, posAAT);
	heurDistRem = norm(m_pPtCloud->getPointSet()->get3f(startIndex, posAAT) - end_pos);

	indexQueue.push(IndexDist(startIndex, heurDistRem, pred_not_visited));
	m_distanceField[startIndex] = 0.f;

	while( !indexQueue.empty() && numMustVisit > 0)
	{
		PointIndex index;
		float currentDistance;
		IndexDist inddist = indexQueue.top();
		indexQueue.pop();	// remove first element

		index = inddist.m_index;

		if (m_predecessorField[index] < pred_not_visited)
			continue;

		if (m_predecessorField[index] == pred_not_visited_must_visit)
			numMustVisit--;

		m_predecessorField[index] = inddist.m_pred;

		Vector3f pos = ps->get3f(index, posAAT);
		currentDistance = m_distanceField[index];

		numNeighbors = ptg->getNumAdjacentVertices(index);
		for (i = 0; i < numNeighbors; i++)
		{
			PointIndex neighbour = ptg->getAdjacentVertex(index, i, &ei);
			if (m_predecessorField[neighbour] < pred_not_visited)
				continue;

			if (currentDistance < m_distanceField[neighbour])
			{
				Vector3f neighbour_pos = ps->get3f(neighbour, posAAT);

				heurDistRem = norm(neighbour_pos - end_pos);

				float distance = max(1e-20f, pEdges->get1f(ei, weightAAT)) + currentDistance;
				if (m_distanceField[neighbour] > distance )
				{
					m_distanceField[neighbour] = distance;
					indexQueue.push(IndexDist(neighbour, distance + heurDistRem, index));
				}
			}
		}
	}

	PointIndex index = endIndex;

	path.clear();

	// no path leads to this point
	if (m_predecessorField[index] >= pred_not_visited)
	{
		dist = m_HugeDistance;
		return;
	}
  	
	// set distance
	dist = m_distanceField[index];

	while( index != startIndex )
	{
		path.push_back(index);
		if( path.size() > 50000 )
			break;

		index = m_predecessorField[index];
		if (index >= pred_not_visited)
			break;
	}

	path.push_back(startIndex);
	reverse(path.begin(), path.end());
}

void X4::AR4SparseGeodesics::computeGeodesicIsoContour(const mpcard startIndex, float distRadius, vector<Vector3f>& contour, bool bOrder/* = false*/)
{
	card32 i, j;

	TopologyIterator * tIter = dynamic_cast<TopologyIterator*>(m_pPCIt);
	x4Assert( tIter );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(m_pPCIt);
	RandomAccessPointCloudIterator * rIt = dynamic_cast<RandomAccessPointCloudIterator*>(m_pPCIt);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(m_pPCIt);
	AAT posAAT = m_pPtCloud->getAAT( "position" );

	SceneObjectIterator * iter = m_pPtCloud->createIterator( SceneObjectIteratorType::CAP_BASIC_PC | 
		SceneObjectIteratorType::CAP_RANDOM_ACCESS_PC );
	x4Assert( iter != NULL );
	BasicPointCloudIterator * neighbour_bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	RandomAccessPointCloudIterator * neighbour_rIt = dynamic_cast<RandomAccessPointCloudIterator*>(iter);

	PointSet* ps = m_pPtCloud->getPointSet();

	vector<IndexDist> contourTemp;

	// -----------------------------------------
	// distance transformation for the complete point cloud
	const card32 pred_start_pred = m_pPtCloud->getNumPoints();
	const card32 pred_not_visited = pred_start_pred + 1;
	const card32 pred_not_visited_must_visit = pred_not_visited + 1;
	fill(m_distanceField.begin(), m_distanceField.end(), m_HugeDistance);
	fill(m_predecessorField.begin(), m_predecessorField.end(), pred_not_visited);

	AR4SG_START_PROFILE(m_profPropagate)

	priority_queue<IndexDist> indexQueue;
	card32 numNeighbors;

	indexQueue.push(IndexDist(startIndex, 0.f, pred_start_pred));
	m_distanceField[startIndex] = 0.f;

	while ( !indexQueue.empty() )
	{
		PointIndex index;
		float currentDistance;
		IndexDist inddist = indexQueue.top();
		indexQueue.pop();	// remove first element

		index = inddist.m_index;

		//nearest point is outside the radius, we are done
		if (inddist.m_dist > distRadius)
			break;

		if (m_predecessorField[index] < pred_not_visited)
			continue;

		m_predecessorField[index] = inddist.m_pred;

		rIt->goToIndex( index );
		Vector3f pos = bIt->get3f(posAAT);
		currentDistance = m_distanceField[index];

		numNeighbors = tIter->getNumNeighbours();
		for (i = 0; i < numNeighbors; i++)
		{
			PointIndex neighbour = tIter->getNeighbour(i);
			if (m_predecessorField[neighbour] < pred_not_visited)
				continue;

			if (currentDistance < m_distanceField[neighbour])
			{
				neighbour_rIt->goToIndex(neighbour);
				Vector3f neighbour_pos = neighbour_bIt->get3f(posAAT);

				float edgeLen = norm( neighbour_pos - pos );
				float distance = max(1e-20f, edgeLen) + currentDistance;
				if (m_distanceField[neighbour] > distance )
				{
					m_distanceField[neighbour] = distance;
					IndexDist idnew(neighbour, distance, index);
					indexQueue.push(idnew);

					if (currentDistance <= distRadius && distance > distRadius)
						contourTemp.push_back(idnew);
				}
			}
		}
	}

	AR4SG_STOP_PROFILE(m_profPropagate)

	contour.clear();
	if (contourTemp.size() == 0)
		return;

	if (bOrder)
	{
		warning("AR4SparseGeodesics::computeGeodesicIsoContour(...): ordering of contour points is not yet implemented");
	}
	else
	{
		for (i = 0; i < contourTemp.size(); i++)
		{
			IndexDist id = contourTemp[i];
			if (m_distanceField[id.m_index] < distRadius)
				continue;

			float t, edgeLen;
			Vector3f pos_in, pos_out;
			pos_in = ps->get3f(id.m_pred, posAAT);
			pos_out = ps->get3f(id.m_index, posAAT);
			edgeLen = (pos_out - pos_in).getNorm();

			t = (m_distanceField[id.m_index] - distRadius) / (m_distanceField[id.m_index] - m_distanceField[id.m_pred]);

			contour.push_back( pos_in * t + pos_out * (1.f - t) );
		}
	}

	delete iter;
}

void X4::AR4SparseGeodesics::computeGeodesicIsoContour(const Vector3f& startPoint, float distRadius, vector<Vector3f>& contour, bool bOrder/* = false*/)
{
	const card32 num_init_neighbors = 6;

	card32 i, j;

	TopologyIterator * tIter = dynamic_cast<TopologyIterator*>(m_pPCIt);
	x4Assert( tIter );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(m_pPCIt);
	RandomAccessPointCloudIterator * rIt = dynamic_cast<RandomAccessPointCloudIterator*>(m_pPCIt);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(m_pPCIt);
	AAT posAAT = m_pPtCloud->getAAT( "position" );

	SceneObjectIterator * iter = m_pPtCloud->createIterator( SceneObjectIteratorType::CAP_BASIC_PC | 
		SceneObjectIteratorType::CAP_RANDOM_ACCESS_PC );
	x4Assert( iter != NULL );
	BasicPointCloudIterator * neighbour_bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	RandomAccessPointCloudIterator * neighbour_rIt = dynamic_cast<RandomAccessPointCloudIterator*>(iter);

	PointSet* ps = m_pPtCloud->getPointSet();

	vector<IndexDist> contourTemp;

	// -----------------------------------------
	// distance transformation for the complete point cloud
	const card32 pred_start_pred = m_pPtCloud->getNumPoints();
	const card32 pred_not_visited = pred_start_pred + 1;
	const card32 pred_not_visited_must_visit = pred_not_visited + 1;
	fill(m_distanceField.begin(), m_distanceField.end(), m_HugeDistance);
	fill(m_predecessorField.begin(), m_predecessorField.end(), pred_not_visited);

	AR4SG_START_PROFILE(m_profPropagate)

	priority_queue<IndexDist> indexQueue;
	card32 numNeighbors;

	//initialize nearest neighbors of start point and add them to the queue
	m_pHknnIt->setSeekPointAndReset(startPoint);
	for (i = 0; i < num_init_neighbors; i++)
	{
		j = m_pHknnIt->getPointSetPointNumber();
		float dist = (m_pHknnIt->get3f(posAAT) - startPoint).getNorm();
		if (dist < m_distanceField[j])
		{
			indexQueue.push(IndexDist(j, dist, pred_start_pred));
			m_distanceField[j] = dist;
		}
		m_pHknnIt->next();
	}

	while ( !indexQueue.empty() )
	{
		PointIndex index;
		float currentDistance;
		IndexDist inddist = indexQueue.top();
		indexQueue.pop();	// remove first element

		index = inddist.m_index;

		//nearest point is outside the radius, we are done
		if (inddist.m_dist > distRadius)
			break;

		if (m_predecessorField[index] < pred_not_visited)
			continue;

		m_predecessorField[index] = inddist.m_pred;

		rIt->goToIndex( index );
		Vector3f pos = bIt->get3f(posAAT);
		currentDistance = m_distanceField[index];

		numNeighbors = tIter->getNumNeighbours();
		for (i = 0; i < numNeighbors; i++)
		{
			PointIndex neighbour = tIter->getNeighbour(i);
			if (m_predecessorField[neighbour] < pred_not_visited)
				continue;

			if (currentDistance < m_distanceField[neighbour])
			{
				neighbour_rIt->goToIndex(neighbour);
				Vector3f neighbour_pos = neighbour_bIt->get3f(posAAT);

				float edgeLen = norm( neighbour_pos - pos );
				float distance = max(1e-20f, edgeLen) + currentDistance;
				if (m_distanceField[neighbour] > distance )
				{
					m_distanceField[neighbour] = distance;
					IndexDist idnew(neighbour, distance, index);
					indexQueue.push(idnew);

					if (currentDistance <= distRadius && distance > distRadius)
						contourTemp.push_back(idnew);
				}
			}
		}
	}

	AR4SG_STOP_PROFILE(m_profPropagate)

	contour.clear();
	if (contourTemp.size() == 0)
		return;

	if (bOrder)
	{
		warning("AR4SparseGeodesics::computeGeodesicIsoContour(...): ordering of contour points is not yet implemented");
	}
	else
	{
		for (i = 0; i < contourTemp.size(); i++)
		{
			IndexDist id = contourTemp[i];
			if (m_distanceField[id.m_index] < distRadius)
				continue;

			float t, edgeLen;
			Vector3f pos_in, pos_out;
			pos_in = ps->get3f(id.m_pred, posAAT);
			pos_out = ps->get3f(id.m_index, posAAT);
			edgeLen = (pos_out - pos_in).getNorm();

			t = (m_distanceField[id.m_index] - distRadius) / (m_distanceField[id.m_index] - m_distanceField[id.m_pred]);

			contour.push_back( pos_in * t + pos_out * (1.f - t) );
		}
	}

	delete iter;
}

void X4::AR4SparseGeodesics::distanceTransform(const vector<mpcard>& zeroSet, bool bAdditive/* = false*/)
{
	AR4SG_START_PROFILE(m_profDistanceTransform)

	card32 i, j;

	TopologyIterator * tIter = dynamic_cast<TopologyIterator*>(m_pPCIt);
	x4Assert( tIter );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(m_pPCIt);
	RandomAccessPointCloudIterator * rIt = dynamic_cast<RandomAccessPointCloudIterator*>(m_pPCIt);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(m_pPCIt);
	AAT posAAT = m_pPtCloud->getAAT( "position" );

	SceneObjectIterator * iter = m_pPtCloud->createIterator( SceneObjectIteratorType::CAP_BASIC_PC | 
		SceneObjectIteratorType::CAP_RANDOM_ACCESS_PC );
	x4Assert( iter != NULL );
	BasicPointCloudIterator * neighbour_bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	RandomAccessPointCloudIterator * neighbour_rIt = dynamic_cast<RandomAccessPointCloudIterator*>(iter);

	PointSet* ps = m_pPtCloud->getPointSet();

	// -----------------------------------------
	// distance transformation for the complete point cloud
	const card32 pred_start_pred = m_pPtCloud->getNumPoints();
	const card32 pred_not_visited = pred_start_pred + 1;
	const card32 pred_not_visited_must_visit = pred_not_visited + 1;

	//make a copy so we can add it to result at end
	vector<float> distFieldCopy;
	if (bAdditive)
		distFieldCopy = m_distanceField;

	fill(m_distanceField.begin(), m_distanceField.end(), m_HugeDistance);
	fill(m_predecessorField.begin(), m_predecessorField.end(), pred_not_visited);

	AR4SG_START_PROFILE(m_profPropagate)

	priority_queue<IndexDist> indexQueue;

	for (i = 0; i < zeroSet.size(); i++)
	{
		indexQueue.push(IndexDist(zeroSet[i], 0.f, pred_start_pred));
		m_distanceField[zeroSet[i]] = 0.f;
	}

	while ( !indexQueue.empty() )
	{
		PointIndex index;
		IndexDist inddist = indexQueue.top();
		indexQueue.pop();	// remove first element

		index = inddist.m_index;

		if (m_predecessorField[index] < pred_not_visited)
			continue;

		m_predecessorField[index] = inddist.m_pred;

		rIt->goToIndex( index );
		const Vector3f pos = bIt->get3f(posAAT);
		const float currentDistance = m_distanceField[index];

		const card32 numNeighbors = tIter->getNumNeighbours();
		for (i = 0; i < numNeighbors; i++)
		{
			PointIndex neighbour = tIter->getNeighbour(i);
			if (m_predecessorField[neighbour] < pred_not_visited)
				continue;

			if (currentDistance < m_distanceField[neighbour])
			{
				neighbour_rIt->goToIndex(neighbour);
				const Vector3f neighbour_pos = neighbour_bIt->get3f(posAAT);

				const float edgeLen = norm( neighbour_pos - pos );
				const float distance = max(1e-20f, edgeLen) + currentDistance;
				if (m_distanceField[neighbour] > distance )
				{
					m_distanceField[neighbour] = distance;
					indexQueue.push(IndexDist(neighbour, distance, index));
				}
			}
		}
	}

	AR4SG_STOP_PROFILE(m_profPropagate)

	if (bAdditive)
	{
		//should be open mp'd
		for (i = 0; i < m_distanceField.size(); i++)
		{
			if (distFieldCopy[i] < m_HugeDistance)
				m_distanceField[i] += distFieldCopy[i];
		}
	}

	delete iter;

	AR4SG_STOP_PROFILE(m_profDistanceTransform)
}

void X4::AR4SparseGeodesics::distanceTransform(const vector<Vector3f>& zeroSet, bool bAdditive/* = false*/)
{
	const card32 num_init_neighbors = 6;

	AR4SG_START_PROFILE(m_profDistanceTransform)

	card32 i, j;

	TopologyIterator * tIter = dynamic_cast<TopologyIterator*>(m_pPCIt);
	x4Assert( tIter );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(m_pPCIt);
	RandomAccessPointCloudIterator * rIt = dynamic_cast<RandomAccessPointCloudIterator*>(m_pPCIt);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(m_pPCIt);
	AAT posAAT = m_pPtCloud->getAAT( "position" );

	SceneObjectIterator * iter = m_pPtCloud->createIterator( SceneObjectIteratorType::CAP_BASIC_PC | 
		SceneObjectIteratorType::CAP_RANDOM_ACCESS_PC );
	x4Assert( iter != NULL );
	BasicPointCloudIterator * neighbour_bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	RandomAccessPointCloudIterator * neighbour_rIt = dynamic_cast<RandomAccessPointCloudIterator*>(iter);

	PointSet* ps = m_pPtCloud->getPointSet();

	card32 numNeighbors;

	// -----------------------------------------
	// distance transformation for the complete point cloud
	const card32 pred_start_pred = m_pPtCloud->getNumPoints();
	const card32 pred_not_visited = pred_start_pred + 1;
	const card32 pred_not_visited_must_visit = pred_not_visited + 1;

	//make a copy so we can add it to result at end
	vector<float> distFieldCopy;
	if (bAdditive)
		distFieldCopy = m_distanceField;

	fill(m_distanceField.begin(), m_distanceField.end(), m_HugeDistance);
	fill(m_predecessorField.begin(), m_predecessorField.end(), pred_not_visited);

	AR4SG_START_PROFILE(m_profPropagate)

	priority_queue<IndexDist> indexQueue;

	for (i = 0; i < zeroSet.size(); i++)
	{
		m_pKNN->setTargetPoint(zeroSet[i]);
		for (j = 0; j < num_init_neighbors; j++)
		{
			card32 idx;
			bool bfound;
			m_pKNN->getNextNearestPointIndex(idx, bfound);
			if (!bfound)
				break;
			float dist = (ps->get3f(idx, m_posAAT) - zeroSet[i]).getNorm();
			if (dist < m_distanceField[idx])
			{
				m_distanceField[idx] = dist;
				indexQueue.push(IndexDist(idx, dist, pred_start_pred));
			}
			else
				break;
		}
	}

	while ( !indexQueue.empty() )
	{
		PointIndex index;
		float currentDistance;
		IndexDist inddist = indexQueue.top();
		indexQueue.pop();	// remove first element

		index = inddist.m_index;

		if (m_predecessorField[index] < pred_not_visited)
			continue;

		m_predecessorField[index] = inddist.m_pred;

		rIt->goToIndex( index );
		Vector3f pos = bIt->get3f(posAAT);
		currentDistance = m_distanceField[index];

		numNeighbors = tIter->getNumNeighbours();
		for (i = 0; i < numNeighbors; i++)
		{
			PointIndex neighbour = tIter->getNeighbour(i);
			if (m_predecessorField[neighbour] < pred_not_visited)
				continue;

			if (currentDistance < m_distanceField[neighbour])
			{
				neighbour_rIt->goToIndex(neighbour);
				Vector3f neighbour_pos = neighbour_bIt->get3f(posAAT);

				float edgeLen = norm( neighbour_pos - pos );
				float distance = max(1e-20f, edgeLen) + currentDistance;
				if (m_distanceField[neighbour] > distance )
				{
					m_distanceField[neighbour] = distance;
					indexQueue.push(IndexDist(neighbour, distance, index));
				}
			}
		}
	}

	AR4SG_STOP_PROFILE(m_profPropagate)

	if (bAdditive)
	{
		//should be open mp'd
		for (i = 0; i < m_distanceField.size(); i++)
		{
			if (distFieldCopy[i] < m_HugeDistance)
				m_distanceField[i] += distFieldCopy[i];
		}
	}

	delete iter;

	AR4SG_STOP_PROFILE(m_profDistanceTransform)
}

void X4::AR4SparseGeodesics::distanceTransform(const vector<mpcard>& zeroSet, const vector<mpcard>& destSet, bool bAdditive/* = false*/)
{
	AR4SG_START_PROFILE(m_profDistanceTransform)

	card32 i, j;

	TopologyIterator * tIter = dynamic_cast<TopologyIterator*>(m_pPCIt);
	x4Assert( tIter );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(m_pPCIt);
	RandomAccessPointCloudIterator * rIt = dynamic_cast<RandomAccessPointCloudIterator*>(m_pPCIt);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(m_pPCIt);
	AAT posAAT = m_pPtCloud->getAAT( "position" );

	SceneObjectIterator * iter = m_pPtCloud->createIterator( SceneObjectIteratorType::CAP_BASIC_PC | 
		SceneObjectIteratorType::CAP_RANDOM_ACCESS_PC );
	x4Assert( iter != NULL );
	BasicPointCloudIterator * neighbour_bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	RandomAccessPointCloudIterator * neighbour_rIt = dynamic_cast<RandomAccessPointCloudIterator*>(iter);

	PointSet* ps = m_pPtCloud->getPointSet();

	card32 numNeighbors;

	// -----------------------------------------
	// distance transformation for the complete point cloud
	const card32 pred_start_pred = m_pPtCloud->getNumPoints();
	const card32 pred_not_visited = pred_start_pred + 1;
	const card32 pred_not_visited_must_visit = pred_not_visited + 1;

	//make a copy so we can add it to result at end
	vector<float> distFieldCopy;
	if (bAdditive)
		distFieldCopy = m_distanceField;

	fill(m_distanceField.begin(), m_distanceField.end(), m_HugeDistance);
	fill(m_predecessorField.begin(), m_predecessorField.end(), pred_not_visited);

	AR4SG_START_PROFILE(m_profPropagate)

	priority_queue<IndexDist> indexQueue;

	mpcard dest, numMustVisit = 0;
	for (i = 0; i < destSet.size(); i++)
	{
		dest = destSet[i];
		if (m_predecessorField[dest] != pred_not_visited_must_visit)
		{
			m_predecessorField[dest] = pred_not_visited_must_visit;
			numMustVisit++;
		}
	}

	for (i = 0; i < zeroSet.size(); i++)
	{
		indexQueue.push(IndexDist(zeroSet[i], 0.f, pred_start_pred));
		m_distanceField[zeroSet[i]] = 0.f;
	}

	while ( !indexQueue.empty() && numMustVisit > 0 )
	{
		PointIndex index;
		float currentDistance;
		IndexDist inddist = indexQueue.top();
		indexQueue.pop();	// remove first element

		index = inddist.m_index;

		if (m_predecessorField[index] < pred_not_visited)
			continue;

		if (m_predecessorField[index] == pred_not_visited_must_visit)
			numMustVisit--;

		m_predecessorField[index] = inddist.m_pred;

		rIt->goToIndex( index );
		Vector3f pos = bIt->get3f(posAAT);
		currentDistance = m_distanceField[index];

		numNeighbors = tIter->getNumNeighbours();
		for (i = 0; i < numNeighbors; i++)
		{
			PointIndex neighbour = tIter->getNeighbour(i);
			if (m_predecessorField[neighbour] < pred_not_visited)
				continue;

			if (currentDistance < m_distanceField[neighbour])
			{
				neighbour_rIt->goToIndex(neighbour);
				Vector3f neighbour_pos = neighbour_bIt->get3f(posAAT);

				float edgeLen = norm( neighbour_pos - pos );
				float distance = max(1e-20f, edgeLen) + currentDistance;
				if (m_distanceField[neighbour] > distance )
				{
					m_distanceField[neighbour] = distance;
					indexQueue.push(IndexDist(neighbour, distance, index));
				}
			}
		}
	}

	AR4SG_STOP_PROFILE(m_profPropagate)

	if (bAdditive)
	{
		//should be open mp'd
		for (i = 0; i < m_distanceField.size(); i++)
		{
			if (distFieldCopy[i] < m_HugeDistance)
				m_distanceField[i] += distFieldCopy[i];
		}
	}

	delete iter;

	AR4SG_STOP_PROFILE(m_profDistanceTransform)
}

void X4::AR4SparseGeodesics::distanceTransform(const vector<Vector3f>& zeroSet, const vector<Vector3f>& destSet, bool bAdditive/* = false*/)
{
	const card32 num_init_neighbors = 6;

	AR4SG_START_PROFILE(m_profDistanceTransform)

	card32 i, j;

	TopologyIterator * tIter = dynamic_cast<TopologyIterator*>(m_pPCIt);
	x4Assert( tIter );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(m_pPCIt);
	RandomAccessPointCloudIterator * rIt = dynamic_cast<RandomAccessPointCloudIterator*>(m_pPCIt);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(m_pPCIt);
	AAT posAAT = m_pPtCloud->getAAT( "position" );

	SceneObjectIterator * iter = m_pPtCloud->createIterator( SceneObjectIteratorType::CAP_BASIC_PC | 
		SceneObjectIteratorType::CAP_RANDOM_ACCESS_PC );
	x4Assert( iter != NULL );
	BasicPointCloudIterator * neighbour_bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	RandomAccessPointCloudIterator * neighbour_rIt = dynamic_cast<RandomAccessPointCloudIterator*>(iter);

	PointSet* ps = m_pPtCloud->getPointSet();

	card32 numNeighbors;

	// -----------------------------------------
	// distance transformation for the complete point cloud
	const card32 pred_start_pred = m_pPtCloud->getNumPoints();
	const card32 pred_not_visited = pred_start_pred + 1;
	const card32 pred_not_visited_must_visit = pred_not_visited + 1;

	//make a copy so we can add it to result at end
	vector<float> distFieldCopy;
	if (bAdditive)
		distFieldCopy = m_distanceField;

	fill(m_distanceField.begin(), m_distanceField.end(), m_HugeDistance);
	fill(m_predecessorField.begin(), m_predecessorField.end(), pred_not_visited);

	AR4SG_START_PROFILE(m_profPropagate)

	priority_queue<IndexDist> indexQueue;

	for (i = 0; i < zeroSet.size(); i++)
	{
		m_pHknnIt->setSeekPointAndReset(zeroSet[i]);
		for (j = 0; j < num_init_neighbors; j++)
		{
			card32 idx = m_pHknnIt->getPointSetPointNumber();
			float dist = (m_pHknnIt->get3f(posAAT) - zeroSet[i]).getNorm();
			if (dist < m_distanceField[idx])
			{
				m_distanceField[idx] = dist;
				indexQueue.push(IndexDist(idx, dist, pred_start_pred));
			}
			else
				break;
			m_pHknnIt->next();
		}
	}

	mpcard numMustVisit = 0;
	for (i = 0; i < destSet.size(); i++)
	{
		m_pHknnIt->setSeekPointAndReset(destSet[i]);
		for (j = 0; j < num_init_neighbors; j++)
		{
			card32 idx = m_pHknnIt->getPointSetPointNumber();
			if (m_predecessorField[idx] != pred_not_visited_must_visit)
			{
				numMustVisit++;
				m_predecessorField[idx] = pred_not_visited_must_visit;
			}
			m_pHknnIt->next();
		}
	}

	while ( !indexQueue.empty() && numMustVisit > 0 )
	{
		PointIndex index;
		float currentDistance;
		IndexDist inddist = indexQueue.top();
		indexQueue.pop();	// remove first element

		index = inddist.m_index;

		if (m_predecessorField[index] < pred_not_visited)
			continue;

		if (m_predecessorField[index] == pred_not_visited_must_visit)
			numMustVisit--;

		m_predecessorField[index] = inddist.m_pred;

		rIt->goToIndex( index );
		Vector3f pos = bIt->get3f(posAAT);
		currentDistance = m_distanceField[index];

		numNeighbors = tIter->getNumNeighbours();
		for (i = 0; i < numNeighbors; i++)
		{
			PointIndex neighbour = tIter->getNeighbour(i);
			if (m_predecessorField[neighbour] < pred_not_visited)
				continue;

			if (currentDistance < m_distanceField[neighbour])
			{
				neighbour_rIt->goToIndex(neighbour);
				Vector3f neighbour_pos = neighbour_bIt->get3f(posAAT);

				float edgeLen = norm( neighbour_pos - pos );
				float distance = max(1e-20f, edgeLen) + currentDistance;
				if (m_distanceField[neighbour] > distance )
				{
					m_distanceField[neighbour] = distance;
					indexQueue.push(IndexDist(neighbour, distance, index));
				}
			}
		}
	}

	AR4SG_STOP_PROFILE(m_profPropagate)

	if (bAdditive)
	{
		//should be open mp'd
		for (i = 0; i < m_distanceField.size(); i++)
		{
			if (distFieldCopy[i] < m_HugeDistance)
				m_distanceField[i] += distFieldCopy[i];
		}
	}

	delete iter;

	AR4SG_STOP_PROFILE(m_profDistanceTransform)
}

void X4::AR4SparseGeodesics::distanceTransform(const vector<mpcard>& zeroSet, float distMax, bool bAdditive/* = false*/)
{
	AR4SG_START_PROFILE(m_profDistanceTransform)

	card32 i, j;

	TopologyIterator * tIter = dynamic_cast<TopologyIterator*>(m_pPCIt);
	x4Assert( tIter );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(m_pPCIt);
	RandomAccessPointCloudIterator * rIt = dynamic_cast<RandomAccessPointCloudIterator*>(m_pPCIt);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(m_pPCIt);
	AAT posAAT = m_pPtCloud->getAAT( "position" );

	SceneObjectIterator * iter = m_pPtCloud->createIterator( SceneObjectIteratorType::CAP_BASIC_PC | 
		SceneObjectIteratorType::CAP_RANDOM_ACCESS_PC );
	x4Assert( iter != NULL );
	BasicPointCloudIterator * neighbour_bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	RandomAccessPointCloudIterator * neighbour_rIt = dynamic_cast<RandomAccessPointCloudIterator*>(iter);

	PointSet* ps = m_pPtCloud->getPointSet();

	card32 numNeighbors;

	// -----------------------------------------
	// distance transformation for the complete point cloud
	const card32 pred_start_pred = m_pPtCloud->getNumPoints();
	const card32 pred_not_visited = pred_start_pred + 1;
	const card32 pred_not_visited_must_visit = pred_not_visited + 1;

	//make a copy so we can add it to result at end
	vector<float> distFieldCopy;
	if (bAdditive)
		distFieldCopy = m_distanceField;

	fill(m_distanceField.begin(), m_distanceField.end(), m_HugeDistance);
	fill(m_predecessorField.begin(), m_predecessorField.end(), pred_not_visited);

	AR4SG_START_PROFILE(m_profPropagate)

	priority_queue<IndexDist> indexQueue;

	for (i = 0; i < zeroSet.size(); i++)
	{
		indexQueue.push(IndexDist(zeroSet[i], 0.f, pred_start_pred));
		m_distanceField[zeroSet[i]] = 0.f;
	}

	while ( !indexQueue.empty() )
	{
		PointIndex index;
		float currentDistance;
		IndexDist inddist = indexQueue.top();
		indexQueue.pop();	// remove first element

		index = inddist.m_index;

		if (m_predecessorField[index] < pred_not_visited)
			continue;

		m_predecessorField[index] = inddist.m_pred;

		//if the nearest point is outside the max distance, we can stop
		if (m_distanceField[index] > distMax)
			break;

		rIt->goToIndex( index );
		Vector3f pos = bIt->get3f(posAAT);
		currentDistance = m_distanceField[index];

		numNeighbors = tIter->getNumNeighbours();
		for (i = 0; i < numNeighbors; i++)
		{
			PointIndex neighbour = tIter->getNeighbour(i);
			if (m_predecessorField[neighbour] < pred_not_visited)
				continue;

			if (currentDistance < m_distanceField[neighbour])
			{
				neighbour_rIt->goToIndex(neighbour);
				Vector3f neighbour_pos = neighbour_bIt->get3f(posAAT);

				float edgeLen = norm( neighbour_pos - pos );
				float distance = max(1e-20f, edgeLen) + currentDistance;
				if (m_distanceField[neighbour] > distance )
				{
					m_distanceField[neighbour] = distance;
					indexQueue.push(IndexDist(neighbour, distance, index));
				}
			}
		}
	}

	AR4SG_STOP_PROFILE(m_profPropagate)

	if (bAdditive)
	{
		//should be open mp'd
		for (i = 0; i < m_distanceField.size(); i++)
		{
			m_distanceField[i] = min(m_distanceField[i] + distFieldCopy[i], m_HugeDistance);
		}
	}

	delete iter;

	AR4SG_STOP_PROFILE(m_profDistanceTransform)
}

void X4::AR4SparseGeodesics::distanceTransform(const vector<Vector3f>& zeroSet, const float distMax, bool bAdditive/* = false*/, const float weight/* = 1.f*/)
{
	const card32 num_init_neighbors = 6;

	AR4SG_START_PROFILE(m_profDistanceTransform)

	card32 i, j;

	TopologyIterator * tIter = dynamic_cast<TopologyIterator*>(m_pPCIt);
	x4Assert( tIter );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(m_pPCIt);
	RandomAccessPointCloudIterator * rIt = dynamic_cast<RandomAccessPointCloudIterator*>(m_pPCIt);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(m_pPCIt);
	AAT posAAT = m_pPtCloud->getAAT( "position" );

	SceneObjectIterator * iter = m_pPtCloud->createIterator( SceneObjectIteratorType::CAP_BASIC_PC | 
		SceneObjectIteratorType::CAP_RANDOM_ACCESS_PC );
	x4Assert( iter != NULL );
	BasicPointCloudIterator * neighbour_bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	RandomAccessPointCloudIterator * neighbour_rIt = dynamic_cast<RandomAccessPointCloudIterator*>(iter);

	PointSet* ps = m_pPtCloud->getPointSet();

	// -----------------------------------------
	// distance transformation for the complete point cloud
	const card32 num_points = m_pPtCloud->getNumPoints();
	const card32 pred_start_pred = num_points;
	const card32 pred_not_visited = pred_start_pred + 1;
	const card32 pred_not_visited_must_visit = pred_not_visited + 1;

	//make a copy so we can add it to result at end
	vector<float> distFieldCopy;
	if (bAdditive)
		distFieldCopy = m_distanceField;

	fill(m_distanceField.begin(), m_distanceField.end(), distMax);
	fill(m_predecessorField.begin(), m_predecessorField.end(), pred_not_visited);

	AR4SG_START_PROFILE(m_profPropagate)

	priority_queue<IndexDist> indexQueue;

	for (i = 0; i < zeroSet.size(); i++)
	{
		m_pKNN->setTargetPoint(zeroSet[i]);
		for (j = 0; j < num_init_neighbors; j++)
		{
			card32 idx;
			bool bfound;
			m_pKNN->getNextNearestPointIndex(idx, bfound);
			if (!bfound)
				break;
			float dist = (ps->get3f(idx, m_posAAT) - zeroSet[i]).getNorm();
			if (dist < m_distanceField[idx])
			{
				m_distanceField[idx] = dist;
				indexQueue.push(IndexDist(idx, dist, pred_start_pred));
			}
			else
				break;
		}
	}

	while ( !indexQueue.empty() )
	{
		PointIndex index;
		IndexDist inddist = indexQueue.top();
		indexQueue.pop();	// remove first element

		index = inddist.m_index;

		if (m_predecessorField[index] < pred_not_visited)
			continue;

		m_predecessorField[index] = inddist.m_pred;

		//if the nearest point in the queue is beyond the max distance, stop
		if (m_distanceField[index] > distMax)
			break;

		rIt->goToIndex( index );
//		Vector3f pos = bIt->get3f(posAAT);
		const Vector3f pos = ps->get3f(index, m_posAAT);
		const float currentDistance = m_distanceField[index];

		const card32 numNeighbors = tIter->getNumNeighbours();
		for (i = 0; i < numNeighbors; i++)
		{
			const PointIndex neighbour = tIter->getNeighbour(i);
			if (m_predecessorField[neighbour] < pred_not_visited)
				continue;

			if (currentDistance < m_distanceField[neighbour])
			{
				const Vector3f neighbour_pos = ps->get3f(neighbour, m_posAAT);

				const float edgeLen = norm( neighbour_pos - pos );
				const float distance = max(1e-20f, edgeLen) + currentDistance;
				if (m_distanceField[neighbour] > distance )
				{
					m_distanceField[neighbour] = distance;
					indexQueue.push(IndexDist(neighbour, distance, index));
				}
			}
		}
	}

	AR4SG_STOP_PROFILE(m_profPropagate)

	if (weight != 1.f)
	{
		//should be open mp'd
		for (i = 0; i < num_points; i++)
		{
			m_distanceField[i] *= weight;
		}
	}

	if (bAdditive)
	{
		//should be open mp'd
		for (i = 0; i < num_points; i++)
		{
			m_distanceField[i] = min(m_distanceField[i] + distFieldCopy[i], m_HugeDistance);
		}
	}

	delete iter;

	AR4SG_STOP_PROFILE(m_profDistanceTransform)
}

void X4::AR4SparseGeodesics::geodesicVoronoi(const vector<mpcard>& seeds, vector<mpcard>& tesselation)
{
	card32 i, j;

	TopologyIterator * tIter = dynamic_cast<TopologyIterator*>(m_pPCIt);
	x4Assert( tIter );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(m_pPCIt);
	RandomAccessPointCloudIterator * rIt = dynamic_cast<RandomAccessPointCloudIterator*>(m_pPCIt);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(m_pPCIt);
	AAT posAAT = m_pPtCloud->getAAT( "position" );

	SceneObjectIterator * iter = m_pPtCloud->createIterator( SceneObjectIteratorType::CAP_BASIC_PC | 
		SceneObjectIteratorType::CAP_RANDOM_ACCESS_PC );
	x4Assert( iter != NULL );
	BasicPointCloudIterator * neighbour_bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	RandomAccessPointCloudIterator * neighbour_rIt = dynamic_cast<RandomAccessPointCloudIterator*>(iter);

	PointSet* ps = m_pPtCloud->getPointSet();

	// -----------------------------------------
	// distance transformation for the complete point cloud
	const card32 pred_start_pred = m_pPtCloud->getNumPoints();
	const card32 pred_not_visited = pred_start_pred + 1;
	const card32 pred_not_visited_must_visit = pred_not_visited + 1;

	//make a copy so we can add it to result at end
	fill(m_distanceField.begin(), m_distanceField.end(), m_HugeDistance);
	fill(m_predecessorField.begin(), m_predecessorField.end(), pred_not_visited);

	tesselation.resize(m_pPtCloud->getNumPoints());

	AR4SG_START_PROFILE(m_profPropagate)

	priority_queue<IndexDist> indexQueue;

	const mpcard numSeeds = seeds.size();
	for (i = 0; i < numSeeds; i++)
	{
		indexQueue.push(IndexDist(seeds[i], 0.f, pred_start_pred));
		m_distanceField[seeds[i]] = 0.f;
		tesselation[seeds[i]] = seeds[i];
	}

	while ( !indexQueue.empty() )
	{
		PointIndex index;
		IndexDist inddist = indexQueue.top();
		indexQueue.pop();	// remove first element

		index = inddist.m_index;

		if (m_predecessorField[index] < pred_not_visited)
			continue;

		m_predecessorField[index] = inddist.m_pred;

		if (inddist.m_pred != pred_start_pred)
			tesselation[index] = tesselation[inddist.m_pred];

		rIt->goToIndex( index );
		const Vector3f pos = bIt->get3f(posAAT);
		const float currentDistance = m_distanceField[index];

		const card32 numNeighbors = tIter->getNumNeighbours();
		for (i = 0; i < numNeighbors; i++)
		{
			const PointIndex neighbour = tIter->getNeighbour(i);
			if (m_predecessorField[neighbour] < pred_not_visited)
				continue;

			if (currentDistance < m_distanceField[neighbour])
			{
				neighbour_rIt->goToIndex(neighbour);
				const Vector3f neighbour_pos = neighbour_bIt->get3f(posAAT);

				const float edgeLen = norm( neighbour_pos - pos );
				const float distance = max(1e-20f, edgeLen) + currentDistance;
				if (m_distanceField[neighbour] > distance )
				{
					m_distanceField[neighbour] = distance;
					indexQueue.push(IndexDist(neighbour, distance, index));
				}
			}
		}
	}

	AR4SG_STOP_PROFILE(m_profPropagate)

	delete iter;
}

void X4::AR4SparseGeodesics::geodesicVoronoi(const vector<mpcard>& seeds, const vector<mpcard>& interestPoints, vector<mpcard>& nearestNeighbors)
{
	card32 i, j;

	TopologyIterator * tIter = dynamic_cast<TopologyIterator*>(m_pPCIt);
	x4Assert( tIter );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(m_pPCIt);
	RandomAccessPointCloudIterator * rIt = dynamic_cast<RandomAccessPointCloudIterator*>(m_pPCIt);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(m_pPCIt);
	AAT posAAT = m_pPtCloud->getAAT( "position" );

	SceneObjectIterator * iter = m_pPtCloud->createIterator( SceneObjectIteratorType::CAP_BASIC_PC | 
		SceneObjectIteratorType::CAP_RANDOM_ACCESS_PC );
	x4Assert( iter != NULL );
	BasicPointCloudIterator * neighbour_bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	RandomAccessPointCloudIterator * neighbour_rIt = dynamic_cast<RandomAccessPointCloudIterator*>(iter);

	PointSet* ps = m_pPtCloud->getPointSet();

	// -----------------------------------------
	// distance transformation for the complete point cloud
	const card32 pred_start_pred = m_pPtCloud->getNumPoints();
	const card32 pred_not_visited = pred_start_pred + 1;
	const card32 pred_not_visited_must_visit = pred_not_visited + 1;

	//make a copy so we can add it to result at end
	fill(m_distanceField.begin(), m_distanceField.end(), m_HugeDistance);
	fill(m_predecessorField.begin(), m_predecessorField.end(), pred_not_visited);

	vector<mpcard> tesselation(m_pPtCloud->getNumPoints());

	nearestNeighbors.resize(interestPoints.size());

	AR4SG_START_PROFILE(m_profPropagate)

	priority_queue<IndexDist> indexQueue;

	const mpcard numSeeds = seeds.size();
	for (i = 0; i < numSeeds; i++)
	{
		indexQueue.push(IndexDist(seeds[i], 0.f, pred_start_pred));
		m_distanceField[seeds[i]] = 0.f;
		tesselation[seeds[i]] = seeds[i];
	}

	while ( !indexQueue.empty() )
	{
		PointIndex index;
		IndexDist inddist = indexQueue.top();
		indexQueue.pop();	// remove first element

		index = inddist.m_index;

		if (m_predecessorField[index] < pred_not_visited)
			continue;

		m_predecessorField[index] = inddist.m_pred;
		tesselation[index] = tesselation[inddist.m_pred];

		rIt->goToIndex( index );
		const Vector3f pos = bIt->get3f(posAAT);
		const float currentDistance = m_distanceField[index];

		const card32 numNeighbors = tIter->getNumNeighbours();
		for (i = 0; i < numNeighbors; i++)
		{
			const PointIndex neighbour = tIter->getNeighbour(i);
			if (m_predecessorField[neighbour] < pred_not_visited)
				continue;

			if (currentDistance < m_distanceField[neighbour])
			{
				neighbour_rIt->goToIndex(neighbour);
				const Vector3f neighbour_pos = neighbour_bIt->get3f(posAAT);

				const float edgeLen = norm( neighbour_pos - pos );
				const float distance = max(1e-20f, edgeLen) + currentDistance;
				if (m_distanceField[neighbour] > distance )
				{
					m_distanceField[neighbour] = distance;
					indexQueue.push(IndexDist(neighbour, distance, index));
				}
			}
		}
	}

	AR4SG_STOP_PROFILE(m_profPropagate)

	delete iter;
}

void X4::AR4SparseGeodesics::geodesicVoronoi(const vector<Vector3f>& seeds, vector<mpcard>& tesselation)
{
	card32 i, j;

	TopologyIterator * tIter = dynamic_cast<TopologyIterator*>(m_pPCIt);
	x4Assert( tIter );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(m_pPCIt);
	RandomAccessPointCloudIterator * rIt = dynamic_cast<RandomAccessPointCloudIterator*>(m_pPCIt);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(m_pPCIt);
	AAT posAAT = m_pPtCloud->getAAT( "position" );

	SceneObjectIterator * iter = m_pPtCloud->createIterator( SceneObjectIteratorType::CAP_BASIC_PC | 
		SceneObjectIteratorType::CAP_RANDOM_ACCESS_PC );
	x4Assert( iter != NULL );
	BasicPointCloudIterator * neighbour_bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	RandomAccessPointCloudIterator * neighbour_rIt = dynamic_cast<RandomAccessPointCloudIterator*>(iter);

	PointSet* ps = m_pPtCloud->getPointSet();

	// -----------------------------------------
	// distance transformation for the complete point cloud
	const card32 pred_start_pred = m_pPtCloud->getNumPoints();
	const card32 pred_not_visited = pred_start_pred + 1;
	const card32 pred_not_visited_must_visit = pred_not_visited + 1;

	//make a copy so we can add it to result at end
	fill(m_distanceField.begin(), m_distanceField.end(), m_HugeDistance);
	fill(m_predecessorField.begin(), m_predecessorField.end(), pred_not_visited);

	tesselation.resize(m_pPtCloud->getNumPoints());

	AR4SG_START_PROFILE(m_profPropagate)

	priority_queue<IndexDist> indexQueue;

	const mpcard numSeeds = seeds.size();
	for (i = 0; i < numSeeds; i++)
	{
		card32 nearest;
		bool bfound;
		m_pKNN->setTargetPoint(seeds[i]);
		m_pKNN->getNextNearestPointIndex(nearest, bfound);
		if (bfound)
		{
			const float startDist = norm(m_pPtCloud->getPointSet()->get3f((mpcard)nearest, m_posAAT) - seeds[i]);
			if (startDist < m_distanceField[nearest])
			{
				indexQueue.push(IndexDist(nearest, startDist, pred_start_pred));
				m_distanceField[nearest] = startDist;
				tesselation[nearest] = i;
			}
		}
	}

	while ( !indexQueue.empty() )
	{
		PointIndex index;
		IndexDist inddist = indexQueue.top();
		indexQueue.pop();	// remove first element

		index = inddist.m_index;

		if (m_predecessorField[index] < pred_not_visited)
			continue;

		m_predecessorField[index] = inddist.m_pred;

		if (inddist.m_pred != pred_start_pred)
			tesselation[index] = tesselation[inddist.m_pred];

		rIt->goToIndex( index );
		const Vector3f pos = bIt->get3f(posAAT);
		const float currentDistance = m_distanceField[index];

		const card32 numNeighbors = tIter->getNumNeighbours();
		for (i = 0; i < numNeighbors; i++)
		{
			const PointIndex neighbour = tIter->getNeighbour(i);
			if (m_predecessorField[neighbour] < pred_not_visited)
				continue;

			if (currentDistance < m_distanceField[neighbour])
			{
				neighbour_rIt->goToIndex(neighbour);
				const Vector3f neighbour_pos = neighbour_bIt->get3f(posAAT);

				const float edgeLen = norm( neighbour_pos - pos );
				const float distance = max(1e-20f, edgeLen) + currentDistance;
				if (m_distanceField[neighbour] > distance )
				{
					m_distanceField[neighbour] = distance;
					indexQueue.push(IndexDist(neighbour, distance, index));
				}
			}
		}
	}

	AR4SG_STOP_PROFILE(m_profPropagate)

	delete iter;
}

void X4::AR4SparseGeodesics::relaxDistanceField(mpcard numIters)
{
	mpcard i, j, iter, nbr, numNbrs, numNbrsCount;

	vector<float> distFieldCopy = m_distanceField;
	float distSum;
	float* pDistRead;
	float* pDistWrite;
	float* pDistSwap;

	TopologyIterator* tIt = dynamic_cast<TopologyIterator*>(m_pPCIt);
	RandomAccessPointCloudIterator* rIt = dynamic_cast<RandomAccessPointCloudIterator*>(m_pPCIt);

	pDistRead = &distFieldCopy[0];
	pDistWrite = &m_distanceField[0];

	//for an odd number of iterations so that we always have the result in m_distanceField at the end
	if ((numIters & 1) == 0)
		numIters++;

	for (iter = 0; iter < numIters; iter++)
	{
		for (i = 0; i < m_distanceField.size(); i++)
		{
			rIt->goToIndex(i);

			numNbrs = tIt->getNumNeighbours();
			distSum = 0.f;
			numNbrsCount = 0;
			for (j = 0; j < numNbrs; j++)
			{
				nbr = tIt->getNeighbour(j);
				if (pDistRead[nbr] < m_HugeDistance)
				{
					distSum += pDistRead[nbr];
					numNbrsCount++;
				}
			}

			pDistWrite[i] = distSum / (float)numNbrsCount;
		}

		pDistSwap = pDistRead;
		pDistRead = pDistWrite;
		pDistWrite = pDistSwap;
	}
}

void X4::AR4SparseGeodesics::getMinDistanceDiscreteExhaustive(mpcard* pidx, float* pdist)
{
	card32 i, imin;
	float dist, distMin = m_HugeDistance;

	const card32 numDists = (card32)m_distanceField.size();
	for (i = 0; i < numDists; i++)
	{
		dist = m_distanceField[i];
		if (dist < distMin)
		{
			distMin = dist;
			imin = i;
		}
	}

	if (pidx != NULL)
		*pidx = imin;
	if (pdist != NULL)
		*pdist = distMin;
}

void X4::AR4SparseGeodesics::getMinDistanceContinuousExhaustive(float sigma, Vector3f* ppoint, float* pdist)
{
	mpcard imin;
	float distdisc;
	AAT posAAT = m_pPtCloud->getAAT("position");

	getMinDistanceDiscreteExhaustive(&imin, &distdisc);

	Vector3f currentPoint = m_pPtCloud->getPointSet()->get3f(imin, posAAT);
	float radius = sigma * 2.f;

	Vector3f critPoint;
	Vector6f coeffs;

	// compute a tangent plane
	Vector3f center;
	Matrix3f frame;
	computeLocalFrame(currentPoint, sigma, m_pPtCloud, m_pHknnIt, frame, center);
	const Matrix3f worldToFrame = frame.transpose();

	// gather neighbor points and transform into tangent space
	// height is set to distance, ignores outliers with infinity distance
	m_pHknnIt->setSeekPointAndReset(currentPoint);
	vector<Vector3f> tangentSpacePoints, tangentSpacePointsSurface;
	while( !m_pHknnIt->atEnd())
	{
		if (m_distanceField[m_pHknnIt->getCurrentPointIndex()] >= m_HugeDistance)
		{
			m_pHknnIt->next();
			continue;
		}
		Vector3f tmpPos = m_pHknnIt->get3f(posAAT) - currentPoint;
		float dist = tmpPos.getNorm();
		if (dist > radius)
			break;

		tmpPos = worldToFrame * tmpPos;
		tangentSpacePointsSurface.push_back(tmpPos);
		tmpPos[2] = m_distanceField[m_pHknnIt->getCurrentPointIndex()];
		tangentSpacePoints.push_back(tmpPos);

		m_pHknnIt->next();
	}

	Monomials2ndOrder::computeCoeffs(tangentSpacePoints, coeffs);
	if (!Monomials2ndOrder::evaluateCriticalPoint(coeffs, critPoint))
		return;

	if (pdist != NULL)
		*pdist = critPoint[2];

	Monomials2ndOrder::computeCoeffs(tangentSpacePointsSurface, coeffs);
	critPoint[2] = Monomials2ndOrder::evaluate(coeffs, makeVector2f(critPoint[0], critPoint[1]));
	critPoint = currentPoint + frame * critPoint;

	if (ppoint != NULL)
		*ppoint = critPoint;
}

void X4::AR4SparseGeodesics::getMinDistanceDiscreteHint(const vector<mpcard>& hintSet, mpcard* pidx, float* pdist)
{
	mpcard i, j, nbr, imin = m_InvalidIndex;
	float dist, distMin = m_HugeDistance;

	const card32 pred_start_pred = m_pPtCloud->getNumPoints();
	const card32 pred_not_visited = pred_start_pred + 1;

	fill(m_predecessorField.begin(), m_predecessorField.end(), pred_not_visited);

	priority_queue<IndexDist> pq;

	TopologyIterator* tIt = dynamic_cast<TopologyIterator*>(m_pPCIt);
	BasicPointCloudIterator* bIt = dynamic_cast<BasicPointCloudIterator*>(m_pPCIt);
	RandomAccessPointCloudIterator* rIt = dynamic_cast<RandomAccessPointCloudIterator*>(m_pPCIt);

	const mpcard numHints = hintSet.size();
	for (i = 0; i < numHints; i++)
	{
		const mpcard h = hintSet[i];
		if (m_distanceField[h] < m_HugeDistance && m_predecessorField[h] == pred_not_visited)
		{
			pq.push(IndexDist(h, m_distanceField[h], pred_start_pred));
			m_predecessorField[h] = pred_start_pred;//so that we only set it once
		}
	}

	//the following explores the distance field looking for a minimum
	//the hintSet has been used to initialize the search
	//but it can expand if a node in the queue has neighbors with smaller distance
	while (!pq.empty())
	{
		IndexDist id = pq.top();
		pq.pop();

		card32 idx = id.m_index;

		if (m_predecessorField[idx] < pred_start_pred)
			continue;

		m_predecessorField[idx] = id.m_pred;

		rIt->goToIndex(idx);

		//compare distance to minimum found so far, update if necessary
		dist = m_distanceField[idx];
		if (dist < distMin)
		{
			distMin = dist;
			imin = idx;
		}

		//probe neighbors to see if one of them locally decreases the distance
		const mpcard numNbrs = (mpcard)tIt->getNumNeighbours();
		for (j = 0; j < numNbrs; j++)
		{
			nbr = tIt->getNeighbour((card32)j);

			//only add neighbors to the queue that have a lower distance than the current point, and have not yet been visited
			if (m_predecessorField[nbr] == pred_not_visited && m_distanceField[nbr] < dist)
				pq.push(IndexDist(nbr, m_distanceField[nbr], idx));
		}
	}//while (!pq.empty())

	if (pidx != NULL)
		*pidx = imin;
	if (pdist != NULL)
		*pdist = distMin;
}

void X4::AR4SparseGeodesics::getMinDistanceDiscreteHint(const vector<Vector3f>& hintSet, mpcard* pidx, float* pdist)
{
	const mpcard numNbrs = 6;
	const mpcard num3DHints = hintSet.size();

	vector<mpcard> discHints;

	mpcard i, j;

	for (i = 0; i < num3DHints; i++)
	{
		card32 nextNearest;
		bool bfound;

		m_pKNN->setTargetPoint(hintSet[i]);
		
		for (j = 0; j < numNbrs; j++)
		{
			m_pKNN->getNextNearestPointIndex(nextNearest, bfound);
			if (!bfound)	break;

			discHints.push_back((mpcard)nextNearest);
		}
	}

	getMinDistanceDiscreteHint(discHints, pidx, pdist);
}

void X4::AR4SparseGeodesics::getMinDistanceContinuousHint(float sigma, const vector<Vector3f>& hintSet, Vector3f* ppoint, float* pdist)
{
#define AR4SPARSEGEODESICS_MINDIST_CONTHINT_USEMONOMIALS	1

	const mpcard num_init_neighbors = 6;

	AR4SG_START_PROFILE(m_profContDistFieldMin)

	mpcard i, j, nbr;
	mpcard imin;
	float distdisc;
	AAT posAAT = m_pPtCloud->getAAT("position");

#if AR4SPARSEGEODESICS_MINDIST_CONTHINT_USEMONOMIALS
	const card32 pred_start_pred = m_pPtCloud->getNumPoints();
	const card32 pred_not_visited = pred_start_pred + 1;

	fill(m_predecessorField.begin(), m_predecessorField.end(), pred_not_visited);

	float dist, distMin = m_HugeDistance;
	priority_queue<IndexDist> pq;

	TopologyIterator* tIt = dynamic_cast<TopologyIterator*>(m_pPCIt);
	BasicPointCloudIterator* bIt = dynamic_cast<BasicPointCloudIterator*>(m_pPCIt);
	RandomAccessPointCloudIterator* rIt = dynamic_cast<RandomAccessPointCloudIterator*>(m_pPCIt);

	Vector3f currentPoint;
	const float radius = sigma * 2.f;

	Vector3f critPoint;
	Vector6f coeffs;

	//to compute a tangent plane
	Vector3f center;
	Matrix3f frame;

	//for KNN queries
	card32 nextNearest;
	bool bfound;

	PointSet* ps = m_pPtCloud->getPointSet();

	//initialize imin to invalid, so that we have an indicator if no minimum could be found (all distance are huge)
	imin = m_InvalidIndex;

	//use the hintSet to initialize the search of the distance field
	for (i = 0; i < hintSet.size(); i++)
	{
		m_pKNN->setTargetPoint(hintSet[i]);
		m_pKNN->getNextNearestPointIndex(nextNearest, bfound);
		for (j = 0; j < num_init_neighbors && bfound; j++)
		{
			nbr = nextNearest;
			if (m_distanceField[nbr] < m_HugeDistance && m_predecessorField[nbr] == pred_not_visited)
			{
				pq.push(IndexDist(nbr, m_distanceField[nbr], pred_start_pred));
				m_predecessorField[nbr] = pred_start_pred;//so that we only set it once
			}
			m_pKNN->getNextNearestPointIndex(nextNearest, bfound);
		}//for (j...
	}//for (i...

	//the following explores the distance field looking for a minimum
	//the hintSet has been used to initialize the search
	//but it can expand if a node in the queue has neighbors with smaller distance
	while (!pq.empty())
	{
		IndexDist id = pq.top();
		pq.pop();

		card32 idx = id.m_index;

		if (m_predecessorField[idx] < pred_start_pred)
			continue;

		m_predecessorField[idx] = id.m_pred;

		rIt->goToIndex(idx);

		//compare distance to minimum found so far, update if necessary
		dist = m_distanceField[idx];
		if (dist < distMin)
		{
			distMin = dist;
			imin = idx;
			currentPoint = bIt->get3f(posAAT);
		}

		//probe neighbors to see if one of them locally decreases the distance
		const card32 numNbrs = tIt->getNumNeighbours();
		for (j = 0; j < numNbrs; j++)
		{
			nbr = tIt->getNeighbour(j);

			//only add neighbors to the queue that have a lower distance than the current point, and have not yet been visited
			if (m_predecessorField[nbr] == pred_not_visited && m_distanceField[nbr] < dist)
				pq.push(IndexDist(nbr, m_distanceField[nbr], idx));
		}
	}//while (!pq.empty())

	if (imin == m_InvalidIndex)
	{
		if (pdist != NULL)
			*pdist = m_HugeDistance;
		if (ppoint != NULL)
			*ppoint = NULL_VECTOR3F;
		return;
	}

	currentPoint = m_pPtCloud->getPointSet()->get3f(imin, posAAT);

	computeLocalFrame(currentPoint, sigma, m_pPtCloud, m_pHknnIt, frame, center);
	const Matrix3f worldToFrame = frame.transpose();

	// gather neighbor points and transform into tangent space
	// height is set to distance, ignores outliers with infinity distance
	m_pKNN->setTargetPoint(currentPoint);
	m_pKNN->getNextNearestPointIndex(nextNearest, bfound);
	vector<Vector3f> tangentSpacePoints, tangentSpacePointsSurface;
	while (bfound)
	{
		Vector3f tmpPos = ps->get3f(nextNearest, m_posAAT) - currentPoint;
		dist = tmpPos.getNorm();
		if (dist > radius)
			break;

		if (m_distanceField[nextNearest] < m_HugeDistance)
		{
			tmpPos = worldToFrame * tmpPos;
			tangentSpacePointsSurface.push_back(tmpPos);//collect the actual position in tangent frame
			tmpPos[2] = sqr(m_distanceField[nextNearest]);
			tangentSpacePoints.push_back(tmpPos);//collect "distance surface" points in the tangent frame
		}

		m_pKNN->getNextNearestPointIndex(nextNearest, bfound);
	}//while (bfound)

	Monomials2ndOrder::computeCoeffs(tangentSpacePoints, coeffs, sigma*sigma);
	if (!Monomials2ndOrder::evaluateCriticalPoint(coeffs, critPoint))
	{
		if (pdist != NULL)
			*pdist = distMin;
		if (ppoint != NULL)
			*ppoint = currentPoint;
		return;
	}

	if (critPoint[2] >= 0.f && critPoint[2] < distMin)
	{
		if (pdist != NULL)
			*pdist = critPoint[2];

		Monomials2ndOrder::computeCoeffs(tangentSpacePointsSurface, coeffs, sigma*sigma);
		critPoint[2] = Monomials2ndOrder::evaluate(coeffs, makeVector2f(critPoint[0], critPoint[1]));
		critPoint = currentPoint + frame * critPoint;

		if (ppoint != NULL)
			*ppoint = critPoint;
	}
	else
	{
		if (pdist != NULL)
			*pdist = distMin;
		if (ppoint != NULL)
			*ppoint = currentPoint;
	}
#else
	float dist, distMin = m_HugeDistance;
	Vector3f pointMin;
	for (i = 0; i < hintSet.size(); i++)
	{
		dist = getDistanceField(hintSet[i]);
		if (dist < distMin)
		{
			distMin = dist;
			imin = i;
			pointMin = hintSet[i];
		}
	}

	if (pdist != NULL)
		*pdist = distMin;
	if (ppoint != NULL)
		*ppoint = pointMin;
#endif //AR4SPARSEGEODESICS_MINDIST_CONTHINT_USEMONOMIALS

	AR4SG_STOP_PROFILE(m_profContDistFieldMin)
}

float X4::AR4SparseGeodesics::computeEccentricity(mpcard startIndex)
{
	TopologyIterator * tIter = dynamic_cast<TopologyIterator*>(m_pPCIt);
	x4Assert( tIter );
	BasicPointCloudIterator * bIt = dynamic_cast<BasicPointCloudIterator*>(m_pPCIt);
	RandomAccessPointCloudIterator * rIt = dynamic_cast<RandomAccessPointCloudIterator*>(m_pPCIt);
	ModifyablePointCloudIterator * mIt = dynamic_cast<ModifyablePointCloudIterator*>(m_pPCIt);
	AAT posAAT = m_pPtCloud->getAAT( "position" );

	SceneObjectIterator * iter = m_pPtCloud->createIterator( SceneObjectIteratorType::CAP_BASIC_PC | 
		SceneObjectIteratorType::CAP_RANDOM_ACCESS_PC );
	x4Assert( iter != NULL );
	BasicPointCloudIterator * neighbour_bIt = dynamic_cast<BasicPointCloudIterator*>(iter);
	RandomAccessPointCloudIterator * neighbour_rIt = dynamic_cast<RandomAccessPointCloudIterator*>(iter);

	float eccentricity = 0.f;

	// -----------------------------------------
	// distance transformation for the complete point cloud
	const card32 pred_start_pred = m_pPtCloud->getNumPoints();
	const card32 pred_not_visited = pred_start_pred + 1;
	fill(m_distanceField.begin(), m_distanceField.end(), m_HugeDistance);
	fill(m_predecessorField.begin(), m_predecessorField.end(), pred_not_visited);

	priority_queue<IndexDist> indexQueue;
	indexQueue.push(IndexDist(startIndex, 0.f, pred_start_pred));
	m_distanceField[startIndex] = 0.0f;

	card32 numNeighbors;

	mpcard dest, nbr;
	while( !indexQueue.empty() )
	{
		PointIndex index;
		float currentDistance;
		IndexDist inddist = indexQueue.top();
		indexQueue.pop();	// remove first element

		index = inddist.m_index;

		if (m_predecessorField[index] < pred_not_visited)
			continue;

		m_predecessorField[index] = inddist.m_pred;

		rIt->goToIndex( index );
		Vector3f pos = bIt->get3f(posAAT);
		currentDistance = m_distanceField[index];

		numNeighbors = tIter->getNumNeighbours();
		for( card32 i=0;i<numNeighbors;i++ )
		{
			PointIndex neighbour = tIter->getNeighbour(i);
			if (m_predecessorField[neighbour] < pred_not_visited)
				continue;

			if (currentDistance < m_distanceField[neighbour])
			{
				neighbour_rIt->goToIndex(neighbour);
				Vector3f neighbour_pos = neighbour_bIt->get3f(posAAT);

				float distance = max(1e-20f, norm( neighbour_pos - pos )) + currentDistance;
				if (m_distanceField[neighbour] > distance )
				{
					m_distanceField[neighbour] = distance;
					indexQueue.push(IndexDist(neighbour, distance, index));
				}
			}
		}
	}

	for (dest = 0; dest < m_distanceField.size(); dest++)
	{
		if (m_distanceField[dest] > eccentricity && m_predecessorField[dest] != pred_not_visited)
			eccentricity = m_distanceField[dest];
	}

	return eccentricity;
}

static bool isIndeterminate(const float pV)
{
    return (pV != pV);
} 


void X4::AR4SparseGeodesics::relaxPath(mpcard numIters, vector<Vector3f>& path, float& dist, Vector3f& grad, Vector3f& dirAtSource)
{

	

#define AR4SG_HARDCORE_FINEGRAINED_RELAXATION_PROFILE		0

	AR4SG_START_PROFILE(m_profRelax)

	if (m_pManifold == NULL)
	{
		warning("AR4SparseGeodesics::relaxPath(...): must have a manifold to relax paths");
		return;
	}
	

#if AR4SG_HARDCORE_FINEGRAINED_RELAXATION_PROFILE
	AR4Profile profRelaxMatrix;
	AR4Profile profRelaxSolver;
	AR4Profile profRelaxProject;
	AR4Profile profRelaxLength;
	AR4Profile profRelaxCopy;

	profRelaxMatrix.setName("AR4SparseGeodesics::relaxPath - matrix");
	profRelaxSolver.setName("AR4SparseGeodesics::relaxPath - solver");
	profRelaxProject.setName("AR4SparseGeodesics::relaxPath - projection");
	profRelaxLength.setName("AR4SparseGeodesics::relaxPath - length");
	profRelaxCopy.setName("AR4SparseGeodesics::relaxPath - copy");
#endif //AR4SG_HARDCORE_FINEGRAINED_RELAXATION_PROFILE

	mpcard i, j, k, dimb, dimx, iter;
	const mpcard nPoints = path.size();

	Vector3f projPoint, projNorm, vecLrow;
	StaticMatrix<float, 2, 3> tang;
	vector<StaticMatrix<float, 2, 3>> tangents(path.size());

	vector<Vector3f> pathTemp = path;
	float pathLength, pathLengthMin;

	float stepSize;
	Vector3f pointNew, midpoint;
	vector<Vector3f> pathNext(path.size());
	vector<Vector2f> tsPath(path.size());

	if (nPoints < 2)
	{
		dist = 0.f;
		grad.setZero();
		dirAtSource.setZero();
		return;
	}
	if (nPoints < 3)
	{
		grad = (path[1] - path[0]);
		dist = grad.getNorm();
		grad.normalize();
		dirAtSource = (path[0] - path[1]);
		dirAtSource.normalize();
		return;
	}

	if (m_bRelaxInTangentSpace)
	{
		pathLength = 0.f;
		for (i = 0; i < nPoints; i++)
		{
#if AR4SG_HARDCORE_FINEGRAINED_RELAXATION_PROFILE
			profRelaxProject.start();
#endif //AR4SG_HARDCORE_FINEGRAINED_RELAXATION_PROFILE
			m_pManifold->project(path[i], projPoint, projNorm, tang);
			
			tangents[i] = tang;
#if AR4SG_HARDCORE_FINEGRAINED_RELAXATION_PROFILE
			profRelaxProject.stop();
#endif //AR4SG_HARDCORE_FINEGRAINED_RELAXATION_PROFILE
			if (i > 0)
				pathLength += (path[i] - path[i - 1]).getNorm();
		}
	}
	else
	{
		pathLength = 0.f;
		for (i = 1; i < nPoints; i++)
			pathLength += (path[i] - path[i - 1]).getNorm();
	}



	pathLengthMin = pathLength;

	CGSolver<float> solver;

	SparseMatrix<float> A;
	SparseMatrix<float> Id;
	DynamicVector<float> X, B;

	const mpcard nVars = nPoints - 2;
	const mpcard nVarsMinusOne = nVars - 1;
	mpcard dimopt;
	if (m_bRelaxInTangentSpace)
		dimopt = 2;
	else
		dimopt = 3;
	const mpcard probSize = nVars * dimopt;

	Vector3f sol, step;
	float32 lengthmin, lengthmax, length, lengthrange, rcplengthrange;
	float32 curMatEntry;
	vector< pair<float, Vector3f> > solution(nPoints);

	Vector3f ikgrad, kigrad;
	Matrix3f kiikMat, kikiMat;

	X.setDim(probSize);
	B.setDim(probSize);
	A.setRows(probSize);

	stepSize = 1.f;

	const float lambda = m_relaxLambda;

	if (!m_bRelaxInTangentSpace)
	{
		for (i = 0; i < nVars; i++)
		{
			X[i * 3 + 0] = path[i + 1][0];
			X[i * 3 + 1] = path[i + 1][1];
			X[i * 3 + 2] = path[i + 1][2];
		}
	}

	A.setZero();

	for (iter = 0; iter < numIters; iter++)
	{

#if AR4SG_HARDCORE_FINEGRAINED_RELAXATION_PROFILE
		profRelaxMatrix.start();
#endif //AR4SG_HARDCORE_FINEGRAINED_RELAXATION_PROFILE

		if (m_bRelaxInTangentSpace)
		{
			B.setZero();
			X.setZero();
			for (i = 0; i < nVars; i++)
			{
				Vector3f Lxstar;

				Lxstar = path[i + 1] * 2.f - path[i] - path[i + 2];

				B[i * 2 + 0] = -tangents[i + 1][0] * Lxstar;
				B[i * 2 + 1] = -tangents[i + 1][1] * Lxstar;

				if (i > 0)
				{
					k = i - 1;
					A[i * 2 + 0].setEntry(k * 2 + 0, -tangents[i + 1][0] * tangents[k + 1][0]);
					A[i * 2 + 0].setEntry(k * 2 + 1, -tangents[i + 1][0] * tangents[k + 1][1]);
					A[i * 2 + 1].setEntry(k * 2 + 0, -tangents[i + 1][1] * tangents[k + 1][0]);
					A[i * 2 + 1].setEntry(k * 2 + 1, -tangents[i + 1][1] * tangents[k + 1][1]);
				}

				A[i * 2 + 0].setEntry(i * 2 + 0, lambda + 2.f);
				A[i * 2 + 1].setEntry(i * 2 + 1, lambda + 2.f);

				if (i < nVarsMinusOne)
				{
					k = i + 1;
					A[i * 2 + 0].setEntry(k * 2 + 0, -tangents[i + 1][0] * tangents[k + 1][0]);
					A[i * 2 + 0].setEntry(k * 2 + 1, -tangents[i + 1][0] * tangents[k + 1][1]);
					A[i * 2 + 1].setEntry(k * 2 + 0, -tangents[i + 1][1] * tangents[k + 1][0]);
					A[i * 2 + 1].setEntry(k * 2 + 1, -tangents[i + 1][1] * tangents[k + 1][1]);
				}
			}
		}
		else
		{
			B = X * lambda;

			for (i = 0; i < nVars; i++)
			{
				if (i == 0)
				{
					B[0] += path[0][0];
					B[1] += path[0][1];
					B[2] += path[0][2];
				}
				else
				{
					k = i - 1;
					A[i * 3 + 0].setEntry(k * 3 + 0, -1.f);
					A[i * 3 + 1].setEntry(k * 3 + 1, -1.f);
					A[i * 3 + 2].setEntry(k * 3 + 2, -1.f);
				}//if (i > 0)

				A[i * 3 + 0].setEntry(i * 3 + 0, lambda + 2.f);
				A[i * 3 + 1].setEntry(i * 3 + 1, lambda + 2.f);
				A[i * 3 + 2].setEntry(i * 3 + 2, lambda + 2.f);

				if (i == nVarsMinusOne)
				{
					B[i * 3 + 0] += path[i + 1][0];
					B[i * 3 + 1] += path[i + 1][1];
					B[i * 3 + 2] += path[i + 1][2];
				}
				else
				{
					k = i + 1;
					A[i * 3 + 0].setEntry(k * 3 + 0, -1.f);
					A[i * 3 + 1].setEntry(k * 3 + 1, -1.f);
					A[i * 3 + 2].setEntry(k * 3 + 2, -1.f);
				}//if (i <...
			}//for (i...
		}

		

#if AR4SG_HARDCORE_FINEGRAINED_RELAXATION_PROFILE
		profRelaxMatrix.stop();
#endif //AR4SG_HARDCORE_FINEGRAINED_RELAXATION_PROFILE

#if AR4SG_HARDCORE_FINEGRAINED_RELAXATION_PROFILE
		profRelaxSolver.start();
#endif //AR4SG_HARDCORE_FINEGRAINED_RELAXATION_PROFILE

		//solve the system
		solver.solve(A, X, B);

#if AR4SG_HARDCORE_FINEGRAINED_RELAXATION_PROFILE
		profRelaxSolver.stop();
#endif //AR4SG_HARDCORE_FINEGRAINED_RELAXATION_PROFILE


		//move the points by the solution we found and project them back to the surface
		if (m_bRelaxInTangentSpace)
		{
			for (i = 0; i < nVars; i++)
			{
#if AR4SG_HARDCORE_FINEGRAINED_RELAXATION_PROFILE
				profRelaxProject.start();
#endif //AR4SG_HARDCORE_FINEGRAINED_RELAXATION_PROFILE

				Vector2f tangStep = makeVector2f(X[i * 2 + 0], X[i * 2 + 1]);
				step = tangents[i + 1] * tangStep;

				m_pManifold->project(path[i + 1] + step, projPoint, projNorm, tang);

				pathTemp[i + 1] = projPoint;
				tangents[i + 1] = tang;

#if AR4SG_HARDCORE_FINEGRAINED_RELAXATION_PROFILE
				profRelaxProject.stop();
#endif //AR4SG_HARDCORE_FINEGRAINED_RELAXATION_PROFILE
			}//for (i...
		}
		else
		{
			for (i = 0; i < nVars; i++)
			{
#if AR4SG_HARDCORE_FINEGRAINED_RELAXATION_PROFILE
				profRelaxProject.start();
#endif //AR4SG_HARDCORE_FINEGRAINED_RELAXATION_PROFILE

				step = makeVector3f(X[i * 3 + 0], X[i * 3 + 1], X[i * 3 + 2]);

				m_pManifold->project(step, projPoint, projNorm, tang);

				pathTemp[i + 1] = projPoint;
				tangents[i + 1] = tang;

#if AR4SG_HARDCORE_FINEGRAINED_RELAXATION_PROFILE
				profRelaxProject.stop();
#endif //AR4SG_HARDCORE_FINEGRAINED_RELAXATION_PROFILE
			}//for (i...
		}

#if AR4SG_HARDCORE_FINEGRAINED_RELAXATION_PROFILE
		profRelaxLength.start();
#endif //AR4SG_HARDCORE_FINEGRAINED_RELAXATION_PROFILE

		//evaluate length of new path
		pathLength = 0.f;
		for (i = 1; i < nPoints; i++)
			pathLength += (pathTemp[i] - pathTemp[i - 1]).getNorm();

#if AR4SG_HARDCORE_FINEGRAINED_RELAXATION_PROFILE
		profRelaxLength.stop();
#endif //AR4SG_HARDCORE_FINEGRAINED_RELAXATION_PROFILE

		//if less, keep it, otherwise stop
		if (pathLength < pathLengthMin)
		{
			pathLengthMin = pathLength;

#if AR4SG_HARDCORE_FINEGRAINED_RELAXATION_PROFILE
			profRelaxCopy.start();
#endif //AR4SG_HARDCORE_FINEGRAINED_RELAXATION_PROFILE

			path = pathTemp;

#if AR4SG_HARDCORE_FINEGRAINED_RELAXATION_PROFILE
			profRelaxCopy.stop();
#endif //AR4SG_HARDCORE_FINEGRAINED_RELAXATION_PROFILE
		}
		else
			break;


	}//for (iter...

	


	//return distance
	dist = pathLengthMin;

	//check if we have introduced tiny edges
	const float tinyEdgeThresh = m_avgEdgeLength * AR4SPARSEGEODESICS_TINY_EDGE_FACTOR;
	const float tinyEdgeThreshSqr = tinyEdgeThresh*tinyEdgeThresh;
	const mpcard endpt = nPoints - 1;
	card32 numRemove = 0;
	for (i = 1; i < nPoints; i++)
	{
		const Vector3f edge = path[i] - path[i - numRemove - 1];
		const float edgeLenSqr = edge.getSqrNorm();

		if (edgeLenSqr < tinyEdgeThreshSqr)
		{
			numRemove++;
			if (i == endpt)
			{
				if (numRemove == endpt)
					numRemove--;
				path[i - numRemove] = path[i];
			}
		}
		else if (numRemove > 0)
			path[i - numRemove] = path[i];
	}

	for (i = 0; i < numRemove; i++)
		path.pop_back();

	const mpcard nPointsEnd = path.size();

	//return gradient
	i = nPointsEnd - 1;
	grad = path[i] - path[i - 1];
	grad.normalize();

	Vector2f gts;
	gts[0] = tangents[i][0] * grad;
	gts[1] = tangents[i][1] * grad;
	grad = tangents[i] * gts;
	grad.normalize();


	
	
	//compute direction vector projected at source tangent space
	dirAtSource = path[0] - path[1];
	dirAtSource.normalize();

	
	gts[0] = tangents[0][0] * dirAtSource;
	gts[1] = tangents[0][1] * dirAtSource;
	dirAtSource = tangents[0] * gts;
	dirAtSource.normalize();
	


#if AR4SG_HARDCORE_FINEGRAINED_RELAXATION_PROFILE
	profRelaxMatrix.reportStats();
	profRelaxSolver.reportStats();
	profRelaxProject.reportStats();
	profRelaxLength.reportStats();
	profRelaxCopy.reportStats();
#endif //AR4SG_HARDCORE_FINEGRAINED_RELAXATION_PROFILE

	
	AR4SG_STOP_PROFILE(m_profRelax)
	

}

void X4::AR4SparseGeodesics::reportProfiling()
{
#if AR4SPARSEGEODESICS_PROFILE
	m_profSingleDest.reportStats();
	m_profEarlyStop.reportStats();
	m_profDistanceTransform.reportStats();
	m_profRelax.reportStats();
	m_profContDistFieldMin.reportStats();
	m_profPropagate.reportStats();
	m_profGradient.reportStats();
	m_profBackTrack.reportStats();
#endif //AR4SPARSEGEODESICS_PROFILE
}

///////////////////////////////////////////////////////////////////////////////
//subclass AR4StaticSparseGeodesics
///////////////////////////////////////////////////////////////////////////////

void X4::AR4StaticSparseGeodesics::storeGeodesicData(const mpcard startIndex, const vector<mpcard>& indices, const vector<float>& dists, const vector<Vector3f>& derivates, const vector<vector<mpcard>>* paths/* = NULL*/)
{
	mpcard i, j;
	EntryIndexCompare comp;

	if (m_store.size() == 0)
		m_store.resize(m_pPtCloud->getNumPoints());

	m_store[startIndex].clear();

	PointSet* ps = m_pPtCloud->getPointSet();

	for (i = 0; i < indices.size(); i++)
	{
		AR4SparseGeodesicEntry entry;
		entry.first = indices[i];
		entry.second.m_dist = dists[i];
		entry.second.m_grad = derivates[i];
		if (paths != NULL)
		{
			for (j = 0; j < (*paths)[i].size(); j++)
				entry.second.m_path.push_back(ps->get3f((*paths)[i][j], m_posAAT));
		}

		m_store[startIndex].push_back(entry);
	}

	sort<vector<AR4SparseGeodesicEntry>::iterator, EntryIndexCompare>(m_store[startIndex].begin(), m_store[startIndex].end(), comp);
}

void X4::AR4StaticSparseGeodesics::storeGeodesicData(const mpcard startIndex, const vector<mpcard>& indices, const vector<float>& dists, const vector<Vector3f>& derivates, const vector<vector<Vector3f>>* paths/* = NULL*/)
{
	mpcard i, j;
	EntryIndexCompare comp;

	if (m_store.size() == 0)
		m_store.resize(m_pPtCloud->getNumPoints());

	if (startIndex > m_store.size())
		throw ERangeCheck("AR4StaticSparseGeodesics::storeGeodesicData(...): startIndex out of range");

	m_store[startIndex].clear();

	for (i = 0; i < indices.size(); i++)
	{
		AR4SparseGeodesicEntry entry;
		entry.first = indices[i];
		entry.second.m_dist = dists[i];
		entry.second.m_grad = derivates[i];
		if (paths != NULL)
		{
			for (j = 0; j < (*paths)[i].size(); j++)
				entry.second.m_path.push_back((*paths)[i][j]);
		}

		m_store[startIndex].push_back(entry);
	}

	sort<vector<AR4SparseGeodesicEntry>::iterator, EntryIndexCompare>(m_store[startIndex].begin(), m_store[startIndex].end(), comp);
}

void X4::AR4StaticSparseGeodesics::storeGeodesicData(const mpcard startIndex, const mpcard endIndex, const float dist, const Vector3f& gradient, const vector<Vector3f>* pPath/* = NULL*/)
{
	if (m_store.size() == 0)
		m_store.resize(m_pPtCloud->getNumPoints());

	if (startIndex > m_store.size())
		throw ERangeCheck("AR4StaticSparseGeodesics::storeGeodesicData(...): startIndex out of range");

	const mpcard numPrev = m_store[startIndex].size();

	AR4SparseGeodesicEntry entry;
	entry.first = endIndex;
	entry.second.m_dist = dist;
	entry.second.m_grad = gradient;
	if (pPath != NULL)
		entry.second.m_path = *pPath;

	EntryIndexCompare comp;

	m_store[startIndex].push_back(entry);
	std::inplace_merge<vector<AR4SparseGeodesicEntry>::iterator, EntryIndexCompare>(m_store[startIndex].begin(), m_store[startIndex].begin() + numPrev, m_store[startIndex].end(), comp);
}

void X4::AR4StaticSparseGeodesics::appendGeodesicData(const mpcard startIndex, const vector<mpcard>& indices, const vector<float>& dists, const vector<Vector3f>& derivates, const vector<Vector3f>& dirsAtSource, const vector<vector<Vector3f>>* paths/* = NULL*/)
{
	mpcard i, j;
	EntryIndexCompare comp;

	if (m_store.size() == 0)
		m_store.resize(m_pPtCloud->getNumPoints());

	if (startIndex > m_store.size())
		throw ERangeCheck("AR4StaticSparseGeodesics::appendGeodesicData(...): startIndex out of range");

	
	for (i = 0; i < indices.size(); i++)
	{
		AR4SparseGeodesicEntry entry;
		entry.first = indices[i];
		entry.second.m_dist = dists[i];
		entry.second.m_grad = derivates[i];
		entry.second.m_dirAtSource = dirsAtSource[i];
		if (paths != NULL)
		{
			for (j = 0; j < (*paths)[i].size(); j++)
				entry.second.m_path.push_back((*paths)[i][j]);
		}
		
		m_store[startIndex].push_back(entry);
		
	}
	
	
	sort<vector<AR4SparseGeodesicEntry>::iterator, EntryIndexCompare>(m_store[startIndex].begin(), m_store[startIndex].end(), comp);
	
}


void X4::AR4StaticSparseGeodesics::appendGeodesicDataSymmetric(const mpcard startIndex, const vector<mpcard>& indices, const vector<float>& dists, const vector<Vector3f>& derivates, const vector<Vector3f>& dirsAtSource, const vector<vector<Vector3f>>* paths/* = NULL*/)
{
	mpcard i, j;
	EntryIndexCompare comp;

	if (m_store.size() == 0)
		m_store.resize(m_pPtCloud->getNumPoints());

	if (startIndex > m_store.size())
		throw ERangeCheck("AR4StaticSparseGeodesics::appendGeodesicData(...): startIndex out of range");

	
	for (i = 0; i < indices.size(); i++)
	{
		AR4SparseGeodesicEntry entry;
		entry.first = startIndex;
		entry.second.m_dist = dists[i];
		entry.second.m_grad = dirsAtSource[i];
		entry.second.m_dirAtSource = derivates[i];
		if (paths != NULL)
		{
		
			
			for (j = ((*paths)[i].size()-1); j >=1; j--)
				entry.second.m_path.push_back((*paths)[i][j]);

			entry.second.m_path.push_back((*paths)[i][0]);


		}
		
		m_store[indices[i]].push_back(entry);
		
	}
	
	
	

}


void X4::AR4StaticSparseGeodesics::appendGeodesicData(const mpcard startIndex, const mpcard& endIndex, const float dist, const Vector3f& grad, const Vector3f& dirAtSource, const vector<Vector3f>* path/* = NULL*/)	
{
	EntryIndexCompare comp;

	if (m_store.size() == 0)
		m_store.resize(m_pPtCloud->getNumPoints());

	if (startIndex > m_store.size())
		throw ERangeCheck("AR4StaticSparseGeodesics::appendGeodesicData(...): startIndex out of range");

	AR4SparseGeodesicEntry entry;
	entry.first = endIndex;
	entry.second.m_dist = dist;
	entry.second.m_grad = grad;
	entry.second.m_dirAtSource = dirAtSource;
	if (path != NULL)
	{
		for (mpcard j = 0; j < path->size(); j++)
			entry.second.m_path.push_back((*path)[j]);
	}

	m_store[startIndex].push_back(entry);

	sort<vector<AR4SparseGeodesicEntry>::iterator, EntryIndexCompare>(m_store[startIndex].begin(), m_store[startIndex].end(), comp);
}

void X4::AR4StaticSparseGeodesics::computeGeodesicsLimitRangeRelaxStore(const mpcard startIndex, const float rangeMax, const mpcard numIters)
{
	vector<mpcard> destIndices;
	vector<float> distances;
	vector<Vector3f> gradients, dirsAtSource;
	vector<vector<Vector3f>> paths;

	computeGeodesicsLimitRangeRelax(startIndex, rangeMax, numIters, destIndices, distances, gradients, dirsAtSource, paths);

	storeGeodesicData(startIndex, destIndices, distances, gradients, &paths);
}

void X4::AR4StaticSparseGeodesics::computeGeodesicsLimitRangeStore(const float sigma, const mpcard startIndex, const float rangeMax, bool bPaths)
{
	vector<mpcard> destIndices;
	vector<float> distances;
	vector<Vector3f> gradients;
	vector<vector<Vector3f>> paths;

	if (bPaths)
	{
		computeGeodesicsLimitRange(sigma, startIndex, rangeMax, destIndices, distances, gradients, &paths);
		storeGeodesicData(startIndex, destIndices, distances, gradients, &paths);
	}
	else
	{
		computeGeodesicsLimitRange(sigma, startIndex, rangeMax, destIndices, distances, gradients);
		storeGeodesicData(startIndex, destIndices, distances, gradients, (vector<vector<Vector3f>>*)NULL);
	}
}

X4::mpcard X4::AR4StaticSparseGeodesics::getGeodesicData(const mpcard startIndex, const mpcard endIndex, float& dist, Vector3f& grad, Vector3f& dirAtSource, vector<Vector3f>* path/* = NULL*/)
{

	vector<AR4SparseGeodesicEntry>::iterator storeIt;
	pair<vector<AR4SparseGeodesicEntry>::iterator, vector<AR4SparseGeodesicEntry>::iterator> sres;
	AR4SparseGeodesicEntry entry;
	EntryIndexCompare comp;

	if (m_store.size() == 0)
		return m_InvalidIndex;

	if (startIndex > m_store.size())
		throw ERangeCheck("X4::AR4StaticSparseGeodesics::getGeodesicData(...): invalid startIndex");

	entry.first = endIndex;
	sres = equal_range<vector<AR4SparseGeodesicEntry>::iterator, AR4SparseGeodesicEntry, EntryIndexCompare>(m_store[startIndex].begin(), m_store[startIndex].end(), entry, comp);

	if (sres.first == sres.second || sres.first == m_store[startIndex].end())
		return m_InvalidIndex;

	storeIt = sres.first;
	dist = storeIt->second.m_dist;
	grad = storeIt->second.m_grad;
	dirAtSource = storeIt->second.m_dirAtSource;
	if (path != NULL)
	{
		mpcard i;
		path->clear();
		for (i = 0; i < storeIt->second.m_path.size(); i++)
			path->push_back(storeIt->second.m_path[i]);
	}

	return (mpcard)(storeIt - m_store[startIndex].begin());
}

void X4::AR4StaticSparseGeodesics::getGeodesicData(const mpcard startIndex, vector<mpcard>& endIndices, vector<float>& dists, vector<Vector3f>& grads, vector<vector<Vector3f>>* paths/* = NULL*/)
{
	mpcard i, n;

	endIndices.clear();
	dists.clear();
	grads.clear();

	if (startIndex >= m_store.size())	return;

	n = m_store[startIndex].size();
	endIndices.resize(n);
	dists.resize(n);
	grads.resize(n);

	for (i = 0; i < n; i++)
	{
		endIndices[i] = m_store[startIndex][i].first;
		dists[i] = m_store[startIndex][i].second.m_dist;
		grads[i] = m_store[startIndex][i].second.m_grad;
	}

	if (paths != NULL)
	{
		(*paths).resize(n);
		for (i = 0; i < n; i++)
			(*paths)[i] = m_store[startIndex][i].second.m_path;
	}
}


///////////////////////////////////////////////////////////////////////////////
//subclass AR4DynamicSparseGeodesics
///////////////////////////////////////////////////////////////////////////////

X4::mpcard X4::AR4DynamicSparseGeodesics::storeGeodesicData(const Vector3f& startPoint, const Vector3f& endPoint, const float dist, const Vector3f& grad, const vector<Vector3f>* path/* = NULL*/)
{
	mpcard datumIndex = m_store.size();

	AR4SparseGeodesicEntry entry;
	vector<AR4SparseGeodesicEntry> singleEntry;
	vector<mpcard> singleLookup;

	entry.first = 0;
	entry.second.m_dist = dist;
	entry.second.m_grad = grad;
	if (path != NULL)
	{
		mpcard i;
		for (i = 0; i < path->size(); i++)
			entry.second.m_path.push_back((*path)[i]);
	}

	singleEntry.push_back(entry);
	singleLookup.push_back(m_InvalidIndex);
	m_store.push_back(singleEntry);
	m_backupLookup.push_back(singleLookup);

	return datumIndex;
}

void X4::AR4DynamicSparseGeodesics::updateGeodesicData(const mpcard entryIndex, const Vector3f& startPoint, const Vector3f& endPoint, const float dist, const Vector3f& grad, const vector<Vector3f>* path/* = NULL*/)
{
	if (entryIndex > m_store.size())
		throw ERangeCheck("AR4DynamicSparseGeodesics::updateGeodesicData(...): entry index out of range");

	m_store[entryIndex][0].second.m_dist = dist;
	m_store[entryIndex][0].second.m_grad = grad;
	if (path != NULL)
	{
		m_store[entryIndex][0].second.m_path.clear();
		mpcard i;
		for (i = 0; i < path->size(); i++)
			m_store[entryIndex][0].second.m_path.push_back((*path)[i]);
	}

	dirtyGeodesicData(entryIndex);
}

void X4::AR4DynamicSparseGeodesics::getGeodesicData(const mpcard entryIndex, float& dist, Vector3f& grad, vector<Vector3f>* path/* = NULL*/)
{
	if (entryIndex > m_store.size())
		throw ERangeCheck("AR4DynamicSparseGeodesics::getGeodesicData(...): entry index out of range");

	dist = m_store[entryIndex][0].second.m_dist;
	grad = m_store[entryIndex][0].second.m_grad;
	if (path != NULL)
	{
		path->clear();
		mpcard i;
		for (i = 0; i < m_store[entryIndex][0].second.m_path.size(); i++)
			path->push_back(m_store[entryIndex][0].second.m_path[i]);
	}
}

void X4::AR4DynamicSparseGeodesics::updateEndPointsAndRelax(mpcard numIters, const mpcard entry, const Vector3f& start, const Vector3f& end)
{
	if (entry > m_store.size())
		throw ERangeCheck("AR4DynamicSparseGeodesics::updateEndPointsAndRelax(...): entry index out of range");

	mpcard pathSize = m_store[entry][0].second.m_path.size();
	m_store[entry][0].second.m_path[0] = start;
	m_store[entry][0].second.m_path[pathSize - 1] = end;
	relaxPath(numIters, m_store[entry][0].second.m_path, m_store[entry][0].second.m_dist, m_store[entry][0].second.m_grad, m_store[entry][0].second.m_dirAtSource);
}


