///////////////////////////////////////////////////////////////////////////////
//
//	AR4SparseGeodesics.h
//
//	Header file for the AR4SparseGeodesics class and subclasses
//	Allows sparse storage of geodesic distances, gradients and pathes
//
//	Alan Brunton, December 2012
//
///////////////////////////////////////////////////////////////////////////////


#ifndef __AR4SPARSEGEODESICS_H__
#define __AR4SPARSEGEODESICS_H__


#include "threeParameterMatchingGMOD14.h"

#include "InCorePCTopologyGraph.h"
#include "UnstructuredInCorePointCloud.h"
#include "GeometricFeature.h"
#include "TopologyAttachment.h"
#include "TopologicalKNNIterator.h"
#include "TopologyRangeSearch.h"
#include "GeometricTools.h"

#include "PointSetKNNQuery.h"

#include "IterativeSolvers.h"
#include "IterativeSolvers.inline.h"

#include "AR4MLSManifold.h"

#include "AR4Profile.h"




namespace X4
{

class AR4GeodesicData
{
public:
	float				m_dist;
	Vector3f			m_grad;
	Vector3f			m_dirAtSource;
	vector<Vector3f>	m_path;
	bool				m_bValid;

	AR4GeodesicData(): m_dist(-1.f), m_bValid(false)
	{
		m_grad.setZero();
	}

	AR4GeodesicData& operator=(const AR4GeodesicData& gd)
	{
		m_dist = gd.m_dist;
		m_grad = gd.m_grad;
		m_path = gd.m_path;
		m_bValid = gd.m_bValid;
		return *this;
	}

	bool operator<(const AR4GeodesicData& gd)
	{
		return (m_dist < gd.m_dist);
	}
};


class AR4SparseGeodesics
{
protected:

	class IndexDist
	{
	public:
		X4::PointIndex	m_index;
		float				m_dist;
		X4::PointIndex	m_pred;

		IndexDist(): m_index(0), m_dist(0.f), m_pred(0)
		{
		}
		IndexDist(X4::PointIndex index, float dist, X4::PointIndex pred): m_index(index), m_dist(dist), m_pred(pred)
		{
		}

		inline IndexDist& operator=(const IndexDist& id)
		{
			m_index = id.m_index;
			m_dist = id.m_dist;
			m_pred = id.m_pred;
			return *this;
		}
		inline bool operator<(const IndexDist& id) const
		{
			return (m_dist > id.m_dist);
		}
	};

	typedef pair<mpcard, AR4GeodesicData> AR4SparseGeodesicEntry;
	class EntryIndexCompare
	{
	public:
		inline bool operator()(const AR4SparseGeodesicEntry& entry1, const AR4SparseGeodesicEntry& entry2) const
		{
			return (entry1.first < entry2.first);
		}
	};


	UnstructuredInCorePointCloud*			m_pPtCloud;
	InCorePCTopologyGraph*					m_pTopology;
	HierarchicalKNNIterator*				m_pHknnIt;
	PointCloudIterator*						m_pPCIt;
	AAT										m_posAAT;

	AttachedIndexedOctree*					m_pOctree;
	PointSetKNNQuery*						m_pKNN;

	AR4MLSManifold3f*						m_pManifold;

	vector<float>							m_distanceField;
	vector<card32>							m_predecessorField;

	mpcard									m_nMaxPtsGradEst;
	float									m_avgEdgeLength;

	vector<vector<AR4SparseGeodesicEntry>>	m_store;

	vector<vector<mpcard>>					m_backupLookup;
	vector<AR4GeodesicData>					m_backup;
	vector<bool>							m_geoDataDirty;

	bool									m_bRelaxInTangentSpace;
	float									m_relaxLambda;

	AR4Profile								m_profSingleDest;
	AR4Profile								m_profEarlyStop;
	AR4Profile								m_profDistanceTransform;
	AR4Profile								m_profPropagate;
	AR4Profile								m_profGradient;
	AR4Profile								m_profBackTrack;
	AR4Profile								m_profRelax;
	AR4Profile								m_profContDistFieldMin;


	virtual void dirtyGeodesicData(const mpcard entry)
	{
		//by default do nothing
		//provide real implementation in selected derived classes
	}

public:

	const static mpcard						m_InvalidIndex = 0xFFFFFFFFFFFFFFFF;
	const static float						m_HugeDistance;
	const static mpcard						m_defaultMaxPtsGradEst = 10;

	AR4SparseGeodesics()
	{
		m_pManifold = NULL;
		m_pPtCloud = NULL;
		m_pTopology = NULL;
		m_pHknnIt = NULL;
		m_pPCIt = NULL;
		m_pOctree = NULL;
		m_pKNN = NULL;
		m_nMaxPtsGradEst = m_defaultMaxPtsGradEst;
		m_avgEdgeLength = 0.f;
		m_bRelaxInTangentSpace = true;
		m_relaxLambda = 0.f;
		m_profSingleDest.setName("AR4SparseGeodesics::computeGeodesicsSingleDest");
		m_profEarlyStop.setName("AR4SparseGeodesics::computeGeodesicsEarlyStop");
		m_profDistanceTransform.setName("AR4SparseGeodesics::distanceTransform");
		m_profPropagate.setName("AR4SparseGeodesics - distance propagation");
		m_profGradient.setName("AR4SparseGeodesics - gradient estimation");
		m_profBackTrack.setName("AR4SparseGeodesics - path back-tracking");
		m_profRelax.setName("AR4SparseGeodesics::relaxPath");
		m_profContDistFieldMin.setName("AR4SparseGeodesics::getMinDistanceContinuousHint");
	}
	~AR4SparseGeodesics()
	{
		delete m_pPCIt;
		delete m_pKNN;
		delete m_pOctree;
	}

	void setPointCloud(UnstructuredInCorePointCloud* pc)
	{
		TopologyAttachment* tatt;

		if (pc != m_pPtCloud)
		{
			delete m_pHknnIt;
			m_pHknnIt = NULL;
			delete m_pPCIt;
			m_pPCIt = NULL;
			delete m_pKNN;
			m_pKNN = NULL;
			delete m_pOctree;
			m_pOctree = NULL;
		}

		m_pPtCloud = pc;
		tatt = dynamic_cast<TopologyAttachment*>(m_pPtCloud->getAttachments()->getData(TopologyAttachment::getDefaultName()));
		m_pTopology = tatt->getTopology();

		if (m_pHknnIt == NULL)
			m_pHknnIt = new HierarchicalKNNIterator(m_pPtCloud, 32, NULL);
		if (m_pPCIt == NULL)
			m_pPCIt = m_pTopology->createIterator(m_pPtCloud->getPointSet());

		m_posAAT = m_pPtCloud->getAAT("position");

		if (m_pOctree == NULL)
			m_pOctree = PointSetKNNQuery::createOctree(m_pPtCloud->getPointSet());
		if (m_pKNN == NULL)
			m_pKNN = new PointSetKNNQuery(m_pPtCloud->getPointSet(), m_pOctree, NULL_VECTOR3F, m_posAAT);

		m_distanceField.resize(m_pPtCloud->getNumPoints());
		m_predecessorField.resize(m_pPtCloud->getNumPoints());

		m_avgEdgeLength = 0.f;
		mpcard edgeCount = 0;
		mpcard i, j;
		const mpcard n = m_pPtCloud->getNumPoints();
		for (i = 0; i < n; i++)
		{
			const Vector3f p = m_pPtCloud->getPointSet()->get3f(i, m_posAAT);
			const mpcard numNbrs = m_pTopology->getNumAdjacentVertices(i);
			for (j = 0; j < numNbrs; j++)
			{
				const Vector3f q = m_pPtCloud->getPointSet()->get3f(m_pTopology->getAdjacentVertex(i, j), m_posAAT);
				m_avgEdgeLength += (p - q).getNorm();
				edgeCount++;
			}
		}
		m_avgEdgeLength /= (float)edgeCount;
	}
	UnstructuredInCorePointCloud* getPointCloud()	{ return m_pPtCloud; }

	void setManifold(AR4MLSManifold3f* pManifold)
	{
		m_pManifold = pManifold;
		setPointCloud(m_pManifold->getPointCloud());
	}
	AR4MLSManifold3f* getManifold()					{ return m_pManifold; }

	bool isRelaxationInTangentSpace()				{ return m_bRelaxInTangentSpace; }
	void setRelaxationInTangentSpace(bool btangent)	{ m_bRelaxInTangentSpace = btangent; }

	void setRelaxationSoftConstraint(float lambda)	{ m_relaxLambda = lambda; }
	float getRelaxationSoftConstraint()				{ return m_relaxLambda; }

	//set sigma <= 0 to avoid computing gradients with tangent spaces (slow)
	//set paths to NULL to avoid computing paths
	void computeGeodesicsEarlyStop(float sigma, const mpcard startIndex, const vector<mpcard> &indices, vector<float>& distances, vector<Vector3f> &derivates, vector<vector<mpcard>>* paths = NULL);
	void computeGeodesicsEarlyStop(float sigma, const mpcard startIndex, const vector<mpcard> &indices, vector<float>& distances, vector<Vector3f> &derivates, vector<vector<Vector3f>>* paths);
	void computeGeodesicsEarlyStop(float sigma, const Vector3f& startPoint, const vector<Vector3f> &points, vector<float>& distances, vector<Vector3f> &derivates, vector<vector<Vector3f>>* paths = NULL);
	void computeGeodesicSingleDest(float sigma, const mpcard startIndex, const mpcard endIndex, float& dist, Vector3f& derivative, vector<mpcard>* path = NULL);
	void computeGeodesicSingleDest(float sigma, const mpcard startIndex, const mpcard endIndex, float& dist, Vector3f& derivative, vector<Vector3f>& path);
	void computeGeodesicSingleDest(float sigma, const Vector3f& startPoint, const Vector3f& point, float& dist, Vector3f& derivative, vector<Vector3f>* path = NULL);


	void computeMinGeodesicToSet(const mpcard startIndex, const vector<mpcard>& indices, float& distMin, mpcard& nearest);

	void computeGeodesicsLimitRange(const mpcard startIndex, const float rangeMax, vector<mpcard>& destIndices, vector<float>& distances, vector<vector<Vector3f>>* pPaths = NULL);
	void computeGeodesicsLimitRange(const float sigma, const mpcard startIndex, const float rangeMax, vector<mpcard>& destIndices, vector<float>& distances, vector<Vector3f>& gradients, vector<vector<Vector3f>>* pPaths = NULL);
	void computeGeodesicsLimitRangeRelax(const mpcard startIndex, const float rangeMax, const mpcard numIters, vector<mpcard>& destIndices, vector<float>& distances, vector<Vector3f>& gradients, vector<Vector3f>& dirsAtSource, vector<vector<Vector3f>>& paths);

	void computeMinWeightPath(InCorePCTopologyGraph* ptg, const mpcard startIndex, const mpcard endIndex, float& dist, vector<mpcard>& path);

	void computeGeodesicIsoContour(const mpcard startIndex, float distRadius, vector<Vector3f>& contour, bool bOrder = false);
	void computeGeodesicIsoContour(const Vector3f& startPoint, float distRadius, vector<Vector3f>& contour, bool bOrder = false);

	void distanceTransform(const vector<mpcard>& zeroSet, bool bAdditive = false);
	void distanceTransform(const vector<Vector3f>& zeroSet, bool bAdditive = false);
	void distanceTransform(const vector<mpcard>& zeroSet, const vector<mpcard>& destSet, bool bAdditive = false);
	void distanceTransform(const vector<Vector3f>& zeroSet, const vector<Vector3f>& destSet, bool bAdditive = false);
	void distanceTransform(const vector<mpcard>& zeroSet, const float distMax, bool bAdditive = false);
	void distanceTransform(const vector<Vector3f>& zeroSet, const float distMax, bool bAdditive = false, const float weight = 1.f);

	void geodesicVoronoi(const vector<mpcard>& seeds, vector<mpcard>& tesselation);
	void geodesicVoronoi(const vector<mpcard>& seeds, const vector<mpcard>& interestPoints, vector<mpcard>& nearestNeighbors);
	void geodesicVoronoi(const vector<Vector3f>& seeds, vector<mpcard>& tesselation);

	float computeEccentricity(mpcard startIndex);

	void clearDistanceField()
	{
		fill(m_distanceField.begin(), m_distanceField.end(), m_HugeDistance);
	}

	float getDistanceField(const mpcard index)				{ return m_distanceField[index]; }
	float getDistanceField(const Vector3f& point)
	{
		const mpcard num_nearest_neighbors = 6;
		mpcard i;
		card32 idx;
		float dist = 0.f;
		float w, wsum = 0.f;
		AAT posAAT = m_pPtCloud->getAAT("position");
		m_pHknnIt->setSeekPointAndReset(point);
		i = 0;
		while (i < num_nearest_neighbors && !m_pHknnIt->atEnd())
		{
			idx = m_pHknnIt->getPointSetPointNumber();
			if (m_distanceField[idx] < m_HugeDistance)
			{
				w = exp(-(m_pHknnIt->get3f(posAAT) - point).getSqrNorm());
				wsum += w;
				dist += w * m_distanceField[idx];
				i++;
			}
			m_pHknnIt->next();
		}
		if (wsum == 0.f)
			dist = m_HugeDistance;
		else
			dist /= wsum;
		return dist;
	}

	void addToDistanceField(const float offset)
	{
		const mpcard numPoints = m_distanceField.size();
		for (mpcard i = 0; i < numPoints; i++)
		{
			m_distanceField[i] = min(m_HugeDistance, fabs(m_distanceField[i] + offset));
		}
	}

	void copyDistanceField(vector<float>& fieldCopy)
	{
		const mpcard numPts = m_distanceField.size();
		if (fieldCopy.size() < numPts)
			fieldCopy.resize(numPts);
		std::copy(m_distanceField.begin(), m_distanceField.end(), fieldCopy.begin());
	}
	void copyDistanceField(DynamicVector<float>& fieldCopy)
	{
		const mpcard numPts = m_distanceField.size();
		if (fieldCopy.getDim() < numPts)
			fieldCopy.setDim(numPts);
		for (mpcard i = 0; i < numPts; i++)
			fieldCopy[i] = m_distanceField[i];
	}

	void relaxDistanceField(mpcard numIters);

	void getMinDistanceDiscreteExhaustive(mpcard* pidx, float* pdist);
	void getMinDistanceContinuousExhaustive(float sigma, Vector3f* ppoint, float* pdist);
	void getMinDistanceDiscreteHint(const vector<mpcard>& hintSet, mpcard* pidx, float* pdist);
	void getMinDistanceDiscreteHint(const vector<Vector3f>& hintSet, mpcard* pidx, float* pdist);
	void getMinDistanceContinuousHint(float sigma, const vector<Vector3f>& hintSet, Vector3f* ppoint, float* pdist);

	virtual void reportProfiling();

	virtual void storeGeodesicData(const mpcard startIndex, const vector<mpcard>& indices, const vector<float>& dists, const vector<Vector3f>& derivates, const vector<vector<mpcard>>* paths = NULL)
	{
		//by default do nothing
		//provide real implementation in selected derived classes
	}
	virtual void storeGeodesicData(const mpcard startIndex, const vector<mpcard>& indices, const vector<float>& dists, const vector<Vector3f>& derivates, const vector<vector<Vector3f>>* paths = NULL)
	{
		//by default do nothing
		//provide real implementation in selected derived classes
	}
	virtual void storeGeodesicData(const mpcard startIndex, const mpcard endIndex, const float dist, const Vector3f& gradient, const vector<Vector3f>* pPath = NULL)
	{
		//by default do nothing
		//provide real implementation in selected derived classes
	}
	virtual void appendGeodesicData(const mpcard startIndex, const vector<mpcard>& indices, const vector<float>& dists, const vector<Vector3f>& derivates, const vector<Vector3f>& dirsAtSource, const vector<vector<Vector3f>>* paths = NULL)
	{
		//by default do nothing
		//provide real implementation in selected derived classes
	}

	virtual void appendGeodesicDataSymmetric(const mpcard startIndex, const vector<mpcard>& indices, const vector<float>& dists, const vector<Vector3f>& derivates, const vector<Vector3f>& dirsAtSource, const vector<vector<Vector3f>>* paths = NULL)
	{
		//by default do nothing
		//provide real implementation in selected derived classes
	}
	
	
	virtual void appendGeodesicData(const mpcard startIndex, const mpcard& endIndex, const float dist, const Vector3f& grad, const Vector3f& dirAtSource, const vector<Vector3f>* path = NULL)
	{
		//by default do nothing
		//provide real implementation in selected derived classes
	}
	virtual mpcard storeGeodesicData(const Vector3f& startPoint, const Vector3f& endPoint, const float dist, const Vector3f& grad, const vector<Vector3f>* path = NULL)
	{
		//by default do nothing
		//provide real implementation in selected derived classes
		return m_InvalidIndex;
	}
	virtual void updateGeodesicData(const mpcard entryIndex, const Vector3f& startPoint, const Vector3f& endPoint, const float dist, const Vector3f& grad, const vector<Vector3f>* path = NULL)
	{
		//by default do nothing
		//provide real implementation in selected derived classes
	}

	virtual mpcard getGeodesicData(const mpcard startIndex, const mpcard endIndex, float& dist, Vector3f& grad, Vector3f& dirAtSource, vector<Vector3f>* path = NULL)
	{
		//by default do nothing
		//provide real implementation in selected derived classes
		return m_InvalidIndex;
	}
	virtual void getGeodesicData(const mpcard startIndex, vector<mpcard>& endIndices, vector<float>& dists, vector<Vector3f>& grads, vector<vector<Vector3f>>* paths = NULL)
	{
		//by default do nothing
		//provide real implementation in selected derived classes
	}
	virtual void getGeodesicData(const mpcard entryIndex, float& dist, Vector3f& grad, vector<Vector3f>* path = NULL)
	{
		//by default do nothing
		//provide real implementation in selected derived classes
	}

	//for this version, both start and end are point indices in the point cloud
	virtual bool isPathValid(const mpcard start, const mpcard end)
	{
		//by default do nothing
		//provide real implementation in selected derived classes
		return true;
	}
	//for this version, start is a point index in the point cloud, and endidx is an index of the available end points from start
	virtual bool isPathValid_endidx(const mpcard start, const mpcard endidx)
	{
		//by default do nothing
		//provide real implementation in selected derived classes
		return true;
	}
	virtual bool isPathValid(const mpcard entry)
	{
		//by default do nothing
		//provide real implementation in selected derived classes
		return true;
	}

	//for this version, both start and end are point indices in the point cloud
	virtual void setPathValidity(const mpcard start, const mpcard end, bool bvalid)
	{
		//by default do nothing
		//provide real implementation in selected derived classes
	}
	//for this version, start is a point index in the point cloud, and endidx is an index of the available end points from start
	virtual void setPathValidity_endidx(const mpcard start, const mpcard endidx, bool bvalid)
	{
		//by default do nothing
		//provide real implementation in selected derived classes
	}
	virtual void setPathValidity(const mpcard entry, bool bvalid)
	{
		//by default do nothing
		//provide real implementation in selected derived classes
	}

	//backing up geodesic data; only makes sense for dynamic geodesics
	virtual void backupGeodesicData(const mpcard entry)
	{
		//by default do nothing
		//provide real implementation in selected derived classes
	}
	virtual void rollbackGeodesicData(const mpcard entry)
	{
		//by default do nothing
		//provide real implementation in selected derived classes
	}

	void clearAllBackups()
	{
		mpcard i, j;
		m_backup.clear();
		for (i = 0; i < m_backupLookup.size(); i++)
		{
			for (j = 0; j < m_backupLookup[i].size(); j++)
				m_backupLookup[i][j] = m_InvalidIndex;
		}
	}

	void relaxPath(mpcard numIters, vector<Vector3f>& path, float& dist, Vector3f& grad, Vector3f& Dir);

	void relaxPaths(mpcard numIters, vector<vector<Vector3f>>& paths, vector<float>& dists, vector<Vector3f>& grads, vector<Vector3f>& dirsAtSource)
	{
		
		mpcard i;
	
		//initialize direction vector at source
		mpcard neighborSize = dists.size();
		dirsAtSource.resize(neighborSize);
		grads.resize(neighborSize);
		for(i=0;i<neighborSize;i++)	
		{
			dirsAtSource[i] = NULL_VECTOR3F;	
			grads[i] = NULL_VECTOR3F;
		}
		
	
		for (i = 0; i < paths.size(); i++)
			relaxPath(numIters, paths[i], dists[i], grads[i], dirsAtSource[i]);
		
		
	}

	virtual void updateEndPointsAndRelax(mpcard numIters, const mpcard entry, const Vector3f& start, const Vector3f& end)
	{
		//by default do nothing
		//provide real implementation in selected derived classes
	}

	virtual void computeGeodesicsLimitRangeRelaxStore(const mpcard startIndex, const float rangeMax, const mpcard numIters)
	{
		//by default do nothing
		//provide real implementation in selected derived classes
	}
	virtual void computeGeodesicsLimitRangeStore(const float sigma, const mpcard startIndex, const float rangeMax, bool bPaths)
	{
		//by default do nothing
		//provide real implementation in selected derived classes
	}
};


class AR4StaticSparseGeodesics : public AR4SparseGeodesics
{
protected:

public:

	void storeGeodesicData(const mpcard startIndex, const vector<mpcard>& indices, const vector<float>& dists, const vector<Vector3f>& derivates, const vector<vector<mpcard>>* paths = NULL);
	void storeGeodesicData(const mpcard startIndex, const vector<mpcard>& indices, const vector<float>& dists, const vector<Vector3f>& derivates, const vector<vector<Vector3f>>* paths = NULL);
	void storeGeodesicData(const mpcard startIndex, const mpcard endIndex, const float dist, const Vector3f& gradient, const vector<Vector3f>* pPath = NULL);
	void appendGeodesicData(const mpcard startIndex, const vector<mpcard>& indices, const vector<float>& dists, const vector<Vector3f>& derivates, const vector<Vector3f>& dirsAtSource, const vector<vector<Vector3f>>* paths = NULL);
	void appendGeodesicDataSymmetric(const mpcard startIndex, const vector<mpcard>& indices, const vector<float>& dists, const vector<Vector3f>& derivates, const vector<Vector3f>& dirsAtSource, const vector<vector<Vector3f>>* paths = NULL);
	void appendGeodesicData(const mpcard startIndex, const mpcard& endIndex, const float dist, const Vector3f& grad, const Vector3f& dirAtSource, const vector<Vector3f>* path = NULL);
	mpcard getGeodesicData(const mpcard startIndex, const mpcard endIndex, float& dist, Vector3f& grad, Vector3f& dirsAtSource, vector<Vector3f>* path = NULL);
	void getGeodesicData(const mpcard startIndex, vector<mpcard>& endIndices, vector<float>& dists, vector<Vector3f>& grads, vector<vector<Vector3f>>* paths = NULL);

	void computeGeodesicsLimitRangeRelaxStore(const mpcard startIndex, const float rangeMax, const mpcard numIters);
	void computeGeodesicsLimitRangeStore(const float sigma, const mpcard startIndex, const float rangeMax, bool bPaths);

	//for this version, both start and end are point indices in the point cloud
	bool isPathValid(const mpcard start, const mpcard end)
	{
		vector<AR4SparseGeodesicEntry>::iterator storeIt;
		AR4SparseGeodesicEntry entry;
		EntryIndexCompare comp;

		if (start > m_store.size())
			throw ERangeCheck("AR4StaticSparseGeodesics::isPathValid(...): start index out of range");

		entry.first = end;
		storeIt = lower_bound<vector<AR4SparseGeodesicEntry>::iterator, AR4SparseGeodesicEntry, EntryIndexCompare>(m_store[start].begin(), m_store[start].end(), entry, comp);

		if (storeIt == m_store[start].end())
			return false;

		return (*storeIt).second.m_bValid;
	}
	//for this version, start is a point index in the point cloud, and endidx is an index of the available end points from start
	bool isPathValid_endidx(const mpcard start, const mpcard endidx)
	{
		if (start > m_store.size())
			throw ERangeCheck("AR4StaticSparseGeodesics::isPathValid_endidx(...): start index out of range");
		if (endidx > m_store[start].size())
			throw ERangeCheck("AR4StaticSparseGeodesics::isPathValid_endidx(...): endidx index out of range");

		return m_store[start][endidx].second.m_bValid;
	}

	//for this version, both start and end are point indices in the point cloud
	void setPathValidity(const mpcard start, const mpcard end, bool bvalid)
	{
		vector<AR4SparseGeodesicEntry>::iterator storeIt;
		AR4SparseGeodesicEntry entry;
		EntryIndexCompare comp;

		if (start > m_store.size())
			throw ERangeCheck("AR4StaticSparseGeodesics::setPathValidity(...): start index out of range");

		entry.first = end;
		storeIt = lower_bound<vector<AR4SparseGeodesicEntry>::iterator, AR4SparseGeodesicEntry, EntryIndexCompare>(m_store[start].begin(), m_store[start].end(), entry, comp);

		if (storeIt == m_store[start].end())
			return;

		(*storeIt).second.m_bValid = bvalid;
	}
	//for this version, start is a point index in the point cloud, and endidx is an index of the available end points from start
	void setPathValidity_endidx(const mpcard start, const mpcard endidx, bool bvalid)
	{
		if (start > m_store.size())
			throw ERangeCheck("AR4StaticSparseGeodesics::setPathValidity_endidx(...): start index out of range");
		if (endidx > m_store[start].size())
			throw ERangeCheck("AR4StaticSparseGeodesics::setPathValidity_endidx(...): endidx index out of range");

		m_store[start][endidx].second.m_bValid = bvalid;
	}
};


class AR4DynamicSparseGeodesics : public AR4SparseGeodesics
{
protected:

	void dirtyGeodesicData(const mpcard entry)
	{
		mpcard ibu;

		if (entry > m_store.size())
			throw ERangeCheck("AR4DynamicSparseGeodesics::dirtyGeodesicData(...): entry index out of range");

		if ((ibu = m_backupLookup[entry][0]) != m_InvalidIndex)
			m_geoDataDirty[ibu] = true;
	}

public:

	mpcard storeGeodesicData(const Vector3f& startPoint, const Vector3f& endPoint, const float dist, const Vector3f& grad, const vector<Vector3f>* path = NULL);
	void updateGeodesicData(const mpcard entryIndex, const Vector3f& startPoint, const Vector3f& endPoint, const float dist, const Vector3f& grad, const vector<Vector3f>* path = NULL);
	void getGeodesicData(const mpcard entryIndex, float& dist, Vector3f& grad, vector<Vector3f>* path = NULL);

	void updateEndPointsAndRelax(mpcard numIters, const mpcard entry, const Vector3f& start, const Vector3f& end);

	bool isPathValid(const mpcard entry)
	{
		if (entry > m_store.size())
			throw ERangeCheck("AR4DynamicSparseGeodesics::isPathValid(...): entry index out of range");

		return m_store[entry][0].second.m_bValid;
	}
	void setPathValidity(const mpcard entry, bool bvalid)
	{
		if (entry == m_InvalidIndex)
			return;
		if (entry > m_store.size())
			throw ERangeCheck("AR4DynamicSparseGeodesics::setPathValidity(...): entry index out of range");

		m_store[entry][0].second.m_bValid = bvalid;
	}

	void backupGeodesicData(const mpcard entry)
	{
		mpcard ibu;

		if (entry > m_store.size())
			throw ERangeCheck("AR4DynamicSparseGeodesics::backupGeodesicData(...): entry index out of range");

		if ((ibu = m_backupLookup[entry][0]) != m_InvalidIndex && m_geoDataDirty[ibu])
		{
			m_backup[ibu] = m_store[entry][0].second;
			m_geoDataDirty[ibu] = false;
		}
		else
		{
			ibu = m_backup.size();
			m_backup.push_back(m_store[entry][0].second);
			m_geoDataDirty.push_back(false);
			m_backupLookup[entry][0] = ibu;
		}
	}
	void rollbackGeodesicData(const mpcard entry)
	{
		mpcard ibu;

		if (entry > m_store.size())
			throw ERangeCheck("AR4DynamicSparseGeodesics::rollbackGeodesicData(...): entry index out of range");

		if ((ibu = m_backupLookup[entry][0]) != m_InvalidIndex && m_geoDataDirty[ibu])
		{
			m_store[entry][0].second = m_backup[ibu];
			m_geoDataDirty[ibu] = false;
		}
	}
};


}


#endif //__AR4SPARSEGEODESICS_H__

