/***************************************************************************
*   Copyright (c) 2008   Art Tevs                                         *
*                                                                         *
*   This library is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU Lesser General Public License as        *
*   published by the Free Software Foundation; either version 3 of        *
*   the License, or (at your option) any later version.                   *
*                                                                         *
*   This library is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU Lesse General Public License for more details.                    *
*                                                                         *
*   The full license is in LICENSE file included with this distribution.  *
***************************************************************************/

//---------------------------------------------------------------------------
#include "stdafx.h"

#include "ApproximatedGeodesicsGM14.h"

#include "GeometricTools.h"
#include "GeometricFeature.h"
#include "UnstructuredInCorePointCloud.h"
#include "UnstructuredInCoreTriangleMesh.h"
#include "TopologyAttachment.h"
#include "TopologicalKNNIterator.h"
#include "VertexArray.h"
#include "FeatureCorrespondence.h"
#include "PCCResampleByCurvature.h"
#include "UnstructuredInCoreTriangleMeshManifold.h"
#include "Monomials2ndOrder.h"

#include "VariableArrayClassProperty.h"
#include "ObjectListProperty.h"
#include "StringHelper.h"
#include "Timer.h"
#include "CopyObjectProperties.h"
#include "ProgressWindow.h"

#include "PCCComputeTopology.h"

//---------------------------------------------------------------------------
using namespace X4::atrec;

//---------------------------------------------------------------------------
IMPLEMENT_X4_CLASS( ApproximatedGeodesicsDataGM14, 0 )
{
   BEGIN_CLASS_INIT( ApproximatedGeodesicsDataGM14 );

   ADD_FLOAT32_PROP (dist, 0);
   ADD_VECTOR3F_PROP (gradient, 0);
   ADD_VECTOR3F_PROP (from, 0);
   ADD_VECTOR3F_PROP (to, 0);
}

//---------------------------------------------------------------------------
ApproximatedGeodesicsDataGM14::ApproximatedGeodesicsDataGM14() :
   dist(0), gradient(X4::NULL_VECTOR3F), from(X4::NULL_VECTOR3F), to(X4::NULL_VECTOR3F)
{
}

//---------------------------------------------------------------------------
ApproximatedGeodesicsDataGM14::ApproximatedGeodesicsDataGM14(const ApproximatedGeodesicsDataGM14& i)
{
   assign(&i);
}

//---------------------------------------------------------------------------
ApproximatedGeodesicsDataGM14::~ApproximatedGeodesicsDataGM14()
{

}

//---------------------------------------------------------------------------
void ApproximatedGeodesicsDataGM14::assign(const X4::Object* obj, X4::X4CopyContext *context)
{
   const ApproximatedGeodesicsDataGM14* o = dynamic_cast<const ApproximatedGeodesicsDataGM14*>(obj);
   x4Assert( o != NULL );

   copyObjectProperties(obj,this);
}


//---------------------------------------------------------------------------
IMPLEMENT_X4_CLASS( ApproximatedGeodesicsGM14, 2 )
{
   BEGIN_CLASS_INIT( ApproximatedGeodesicsGM14 );
   ADD_VAR_ARRAY_PROP(m_Database, 0, new ObjectClassProperty( "", 0, NULL, ApproximatedGeodesicsDataGM14::getClass(), true));
   ADD_VAR_ARRAY_PROP(m_FeatureDatabase, 1, new ObjectClassProperty( "", 0, NULL, ApproximatedGeodesicsDataGM14::getClass(), true));
   ADD_CARD32_PROP (m_DBWidth, 0);
   ADD_CARD32_PROP (m_FeatureDBWidth, 1);
   ADD_OBJECT_PROP(m_PC, 0, PointCloud::getClass(), true);
   ADD_BOOLEAN_PROP(m_useInterpolation, 2);
 }

//---------------------------------------------------------------------------
ApproximatedGeodesicsGM14::ApproximatedGeodesicsGM14() : X4::AttachedData()
{
	m_DBWidth = 0;
	m_FeatureDBWidth = 0;
	m_PC = NULL;
	m_hIt = NULL;
	m_useInterpolation = true;
	m_FeatureGeodeticCallback = NULL;
	m_DataGeodeticCallback = NULL;
	m_FeatureGeodeticCallbackUserData = NULL;
	m_DataGeodeticCallbackUserData = NULL;
	m_lastAddition = 0;//abrunton 11/12/2012
}

//---------------------------------------------------------------------------
ApproximatedGeodesicsGM14::~ApproximatedGeodesicsGM14()
{
   clear();
   clearFeatureToFeature();
}

//---------------------------------------------------------------------------
void ApproximatedGeodesicsGM14::assign(const X4::Object* obj, X4::X4CopyContext *context)
{
   const ApproximatedGeodesicsGM14* o = dynamic_cast<const ApproximatedGeodesicsGM14*>(obj);
   x4Assert( o != NULL );

   copyObjectProperties(obj,this);

   m_FeatureGeodeticCallback = o->m_FeatureGeodeticCallback;
   m_DataGeodeticCallback = o->m_DataGeodeticCallback;
   m_FeatureGeodeticCallbackUserData = o->m_FeatureGeodeticCallbackUserData;
   m_DataGeodeticCallbackUserData = o->m_DataGeodeticCallbackUserData;
}

//---------------------------------------------------------------------------
void ApproximatedGeodesicsGM14::clear()
{
   for (unsigned i=0; i < m_Database.size(); i++)
      delete m_Database[i];
   m_Database.erase( m_Database.begin(), m_Database.end());

   m_DBWidth = 0;
   if (m_PC) delete m_PC;
   if (m_hIt) delete m_hIt;
   m_PC = NULL;
   m_hIt = NULL;
}

//---------------------------------------------------------------------------
void ApproximatedGeodesicsGM14::clearFeatureToFeature()
{
   for (unsigned i=0; i < m_FeatureDatabase.size(); i++)
      delete m_FeatureDatabase[i];
   m_FeatureDatabase.erase( m_FeatureDatabase.begin(), m_FeatureDatabase.end());

   m_FeatureDBWidth = 0;
}

//---------------------------------------------------------------------------
float ApproximatedGeodesicsGM14::distanceFunction(int prev, int current, int next, X4::TGDFContext* context)
{
   if( current < 0 || next < 0 )
      return 0;

   X4::VertexArray va_vert(context->pc);
   Vector3f p0 = va_vert.getPosition3f(current);
   Vector3f p1 = va_vert.getPosition3f(next);
   Vector3f n0 = va_vert.getNormal3f(current);
   Vector3f n1 = va_vert.getNormal3f(next);

   return (acos(n0*n1) / M_PI);
}

//---------------------------------------------------------------------------
void ApproximatedGeodesicsGM14::createFeatureToFeature(X4::card32 numFeatures)
{
	clearFeatureToFeature();

	m_FeatureDBWidth = numFeatures;
	m_FeatureDatabase.resize(m_FeatureDBWidth * m_FeatureDBWidth, NULL);
	for (unsigned i=0; i < m_FeatureDatabase.size(); i++)
		m_FeatureDatabase[i] = new ApproximatedGeodesicsDataGM14;
}

//---------------------------------------------------------------------------
void ApproximatedGeodesicsGM14::createFeatureToFeature(X4::FeatureSet* fs, X4::UnstructuredInCorePointCloud* upc, float sigma)
{
   X4_TIMER_START(timer0);

   clearFeatureToFeature();
   
   // work on the point cloud, which has been previously resized
   if (m_PC == NULL)
   {
       error(__FUNCTION__ " you have first to create a approximated geodetics database, before do this for features");
   }

   // here we use the topology graph of the original point cloud, so of the given point cloud
   // this will ensure that the geodesics are more or less good when computing the initial feature to feature database
   // create topology iterator
   PointCloudIterator* titer = NULL;
   if (upc->getAttachments()->getData(TopologyAttachment::getDefaultName()) == NULL)
   {
      PCCComputeTopology* pccTpg = new PCCComputeTopology();
      pccTpg->setup(PCCComputeTopology::TOPTYPE_KNN);
      InCorePCTopologyGraph* tpg = pccTpg->computeKNNTopology(upc->getPointSet(), upc);
      delete pccTpg;
      titer = tpg->createIterator(upc->getPointSet());
      delete tpg;
   }else
   {
      TopologyAttachment* tatt = dynamic_cast<TopologyAttachment*>(upc->getAttachments()->getData(TopologyAttachment::getDefaultName()));
      InCorePCTopologyGraph* tpg = tatt->getTopology();
      titer = tpg->createIterator(upc->getPointSet());
   }

  
   // resize the feature database
   m_FeatureDBWidth = fs->getNumFeatures();
   m_FeatureDatabase.resize(m_FeatureDBWidth * m_FeatureDBWidth, NULL);
   for (unsigned i=0; i < m_FeatureDatabase.size(); i++)
      m_FeatureDatabase[i] = new ApproximatedGeodesicsDataGM14;

   // create a list of points
   vector<Vector3f> neighbour (fs->getNumFeatures());
   for (unsigned j=0; j < fs->m_Features.size(); j++)
       neighbour[j] = fs->m_Features[j]->m_Position;

   TGDFContext context;
   context.pc = upc;
   context.normalAAT = upc->getAAT("normal");
   context.posAAT = upc->getAAT("position");

   // for each feature compute the db entry
   for (unsigned i=0; i < fs->m_Features.size(); i++)
   {

      // compute results and stability
      vector<Vector3f> grad(fs->m_Features.size());
      vector<float> dist;

      dist = computeGeodeticDistances(upc, titer, sigma, fs->m_Features[i]->m_Position, neighbour, grad, NULL);
      // store results in the db
      for (unsigned j=0; j < fs->m_Features.size(); j++)
      {
         grad[j].normalize();
         getFeatureToFeatureEntry(j,i).gradient = -grad[j];

         // setup default values
         ApproximatedGeodesicsDataGM14& data = getFeatureToFeatureEntry(i,j);
         data.dist = dist[j];
         data.from = fs->m_Features[i]->m_Position;
         data.to = fs->m_Features[j]->m_Position;
      }

   }
   delete titer;

   debugOutput << "ApproximatedGeodetics: feature to feature database created - ";
   X4_TIMER_STOP(timer0);

}

//---------------------------------------------------------------------------
void ApproximatedGeodesicsGM14::createFeatureToFeature(X4::FeatureSet* fs, X4::UnstructuredInCoreTriangleMesh* upc, int subdivisionFactor, float sigma)
{
	X4_TIMER_START(timer0);

	debugOutput << "Begin computing geodesics database for features " << fs->m_Features.size() << "x" << fs->m_Features.size() << "\n"; debugOutput.flush();

	clearFeatureToFeature();

	// work on the point cloud, which has been previously resized
	if (m_PC == NULL)
	{
		error(__FUNCTION__ " you have first to create a approximated geodetics database, before do this for features");
	}

	// we convert temporarely the given triangle mesh to a maniforld, so that we can compute geodesics
	UnstructuredInCoreTriangleMeshManifold* mf = new UnstructuredInCoreTriangleMeshManifold;
	mf->setPointCloud(upc);
	mf->setSubdivisionLevel(subdivisionFactor);
	mf->setCachePaths(false);
	mf->setUseCaching(true);

	// resize the feature database
	m_FeatureDBWidth = fs->getNumFeatures();
	m_FeatureDatabase.resize(m_FeatureDBWidth * m_FeatureDBWidth, NULL);
	for (unsigned i=0; i < m_FeatureDatabase.size(); i++)
	{
		m_FeatureDatabase[i] = new ApproximatedGeodesicsDataGM14;
		m_FeatureDatabase[i]->dist = 1e20f;
	}

	// point-to-point geodesics
	vector<const ManifoldPoint*> points (m_FeatureDBWidth, NULL);
	for (unsigned i=0; i < m_FeatureDBWidth; i++)
	{
		DVectorF pos; pos.setDim(3);
		pos[0] = fs->m_Features[i]->m_Position[0];
		pos[1] = fs->m_Features[i]->m_Position[1];
		pos[2] = fs->m_Features[i]->m_Position[2];
		points[i] = mf->createPoint(pos);
	}

	// compute geodesics and gradients
	for (unsigned i=0; i < m_FeatureDBWidth; i++)
	{
		if ((i % (m_FeatureDBWidth / 10))==0)
		{
			debugOutput << float(i) / float(m_FeatureDBWidth) << "\n";
			debugOutput.flush();
		}

		const DiscreteManifoldPoint* key = dynamic_cast<const DiscreteManifoldPoint*>(points[i]);
		vector<float> dists = mf->getGeodeticDistances(key, points);

		for (card64 j=0; j < points.size(); j++)
		{
			ApproximatedGeodesicsDataGM14& data = getFeatureToFeatureEntry(i,j);
			data.dist = dists[j];
			data.from = fs->m_Features[i]->m_Position;
			data.to = fs->m_Features[j]->m_Position;
			// gradient is copied from the nearest usual point
			DVectorF from; from.setDim(3);
			from[0] = data.from[0];
			from[1] = data.from[1];
			from[2] = data.from[2];
			card64 fromIndex = mf->findNearestPoint(from);

			DVectorF to; to.setDim(3);
			to[0] = data.to[0];
			to[1] = data.to[1];
			to[2] = data.to[2];
			card64 toIndex = mf->findNearestPoint(to);

			ApproximatedGeodesicsDataGM14& pointData = getEntry(fromIndex, toIndex);
			data.gradient = pointData.gradient;
		}

	}


	for (unsigned i=0; i < m_FeatureDBWidth; i++)
		delete points[i];
	delete mf;

	debugOutput << "ApproximatedGeodetics for TriangleMeshes: feature to feature database created - ";
	debugOutput.flush();
	X4_TIMER_STOP(timer0);
}

//---------------------------------------------------------------------------
void ApproximatedGeodesicsGM14::create(X4::UnstructuredInCoreTriangleMesh* upc, int subdivisionFactor, float sigma)
{
	X4_TIMER_START(timer0);

	debugOutput << "Begin computing geodesics database " << upc->getNumPoints() << "x" << upc->getNumPoints() << "\n"; debugOutput.flush();

	clear();

	m_PC = (UnstructuredInCorePointCloud*)upc->copy();

	// we convert temporarely the given triangle mesh to a manifold, so that we can compute geodesics
	UnstructuredInCoreTriangleMeshManifold* mf = new UnstructuredInCoreTriangleMeshManifold;
	mf->setPointCloud(upc);
	mf->setSubdivisionLevel(subdivisionFactor);
	mf->setCachePaths(false);
	mf->setUseCaching(true);
	
	// prepare the database
	m_DBWidth = m_PC->getNumPoints();
	m_Database.resize(m_DBWidth * m_DBWidth, NULL);
	for (unsigned i=0; i < m_Database.size(); i++)
	{
		m_Database[i] = new ApproximatedGeodesicsDataGM14;
		m_Database[i]->dist = 1e20f;
	}

	// point-to-point geodesics
	vector<const ManifoldPoint*> points (m_DBWidth, NULL);
	for (unsigned i=0; i < m_DBWidth; i++)
	{
		points[i] = mf->createPoint(i);
	}

	AAT posAAT = upc->getAAT("position");
	float inv_sigma_2 = 1.0f / sigma / sigma;
	float rad_2 = pow(sigma * 2.0f,2.0f);

	HierarchicalKNNIterator hIt(upc,32,NULL);

	// compute geodesics and gradients
	for (unsigned i=0; i < m_DBWidth; i++)
	{
		if ((i % (m_DBWidth / 10))==0)
		{
			debugOutput << float(i) / float(m_DBWidth) << "\n";
			debugOutput.flush();
		}

		DiscreteManifoldPoint* key = dynamic_cast<DiscreteManifoldPoint*>(mf->createPoint(i));
		vector<float> dists = mf->getGeodeticDistances(key, points);

		for (card64 j=0; j < points.size(); j++)
		{
			ApproximatedGeodesicsDataGM14& data = getEntry(i,j);
			data.dist = dists[j];
			data.from = makeVector3f(points[i]->getCoordinate()[0], points[i]->getCoordinate()[1], points[i]->getCoordinate()[2]);
			data.to = makeVector3f(points[j]->getCoordinate()[0], points[j]->getCoordinate()[1], points[j]->getCoordinate()[2]);
		}

		// gradients
		for (card64 j=0; j < points.size(); j++)
		{
			ApproximatedGeodesicsDataGM14& data = getEntry(i,j);

			// this point is unreachable from our starting point, then continue
			if (data.dist >= 1e19f)
			{
				data.gradient = NULL_VECTOR3F;
				continue;
			}

			Vector3f currentPoint = data.to;
			hIt.setSeekPointAndReset(currentPoint);

			// compute a tangent plane
			Vector3f center;
			Matrix3f frame;
			computeLocalFrame(currentPoint, sigma, upc, &hIt, frame, center);
			const Matrix3f worldToFrame = frame.transpose();

			// gather neighbor points on the target and transform into tangent space
			hIt.setSeekPointAndReset(currentPoint);
			vector<Vector3f> tangentSpacePoints;
			while( !hIt.atEnd())
			{
				ApproximatedGeodesicsDataGM14& tda = getEntry(i,hIt.getCurrentPointIndex());
				if( tda.dist >= 1e20f)
				{
					hIt.next();
					continue;
				}
				Vector3f tmpPos = hIt.get3f(posAAT);
				float dist_2 = normQuad(tmpPos-currentPoint);
				if( dist_2 > rad_2)
					break;

				tmpPos = worldToFrame * (tmpPos-currentPoint);
				tmpPos[2] = tda.dist;
				tangentSpacePoints.push_back(tmpPos);

				hIt.next();
			}

			// compute local plane fit to the points which are reachable around the target
			Vector3f coeffs;
			X4::Monomials1stOrder::computeCoeffs(tangentSpacePoints, coeffs, sigma*sigma);

			// compute derivates
			Vector3f tangentSpaceDerivate = makeVector3f(coeffs[1],coeffs[2],0);
			Vector3f derivate = frame * tangentSpaceDerivate;
			derivate.normalize();

			data.gradient = -derivate;
		}

		delete key;
	}

	debugOutput << "ApproximatedGeodetics: database created - ";
	debugOutput.flush();
	X4_TIMER_STOP(timer0);

	for (unsigned i=0; i < m_DBWidth; i++)
		delete points[i];
	delete mf;
}

//---------------------------------------------------------------------------
void ApproximatedGeodesicsGM14::create(X4::UnstructuredInCorePointCloud* pc, float minSpacing, float maxSpacing, float sigma, bool onlySetupNoGeodesics)
{
   
   clear();

   // first resize the given point cloud accordingly
   m_PC = copyToUnstructuredInCorePointCloud(pc);
   
   // remove all attachments from the pc
   m_PC->getAttachments()->clear();
   
   // if there was a topology graph, then copy it too and attach to the new point cloud
   if (pc->getAttachments()->getData(TopologyAttachment::getDefaultName()))
   {
      TopologyAttachment* tatt = dynamic_cast<TopologyAttachment*>(pc->getAttachments()->getData(TopologyAttachment::getDefaultName()));
      InCorePCTopologyGraph* tpg = tatt->getTopology();
      InCorePCTopologyGraph* ttpg = tpg->copy();
      tatt = new TopologyAttachment;
      tatt->setup(TopologyAttachment::getDefaultName(), X4::AttachedData::ADF_PERSISTENT);
      tatt->setTopology(ttpg);
      m_PC->getAttachments()->attachData(tatt);
   }else
   {
      error("ApproximatedGeodesicsGM14::create - given PC doesn't have a valid topology graph attached!");
      return;
   }

   // do nothing more than that
   if (onlySetupNoGeodesics) return;

   X4_TIMER_START(timer0);
   
   // resample based on spacing
   if (minSpacing < maxSpacing && maxSpacing > 0)
   {
	   PCCResampleByCurvature cmd;
	   cmd.setMinSpacing(minSpacing);
	   cmd.setMaxSpacing(maxSpacing);
	   cmd.resamplePC(m_PC);
   }

  
      
   // create topology iterator
   TopologyAttachment* tatt = dynamic_cast<TopologyAttachment*>(m_PC->getAttachments()->getData(TopologyAttachment::getDefaultName()));
   InCorePCTopologyGraph* tpg = tatt->getTopology(); 
   PointCloudIterator* m_TopologicalIterator = tpg->createIterator(m_PC->getPointSet());
   AAT posATT = m_PC->getAAT( "position" );


   // prepare the database
   m_DBWidth = m_PC->getNumPoints();
   m_Database.resize(m_DBWidth * m_DBWidth, NULL);
   for (unsigned i=0; i < m_Database.size(); i++)
      m_Database[i] = new ApproximatedGeodesicsDataGM14;
   
   // create a list of points 
   vector<Vector3f> points(m_PC->getNumPoints());
   for (card64 i=0; i < m_PC->getNumPoints(); i++)
      points[i] = m_PC->getPointSet()->get3f(i, posATT);

   TGDFContext context;
   context.pc = m_PC;
   context.normalAAT = m_PC->getAAT("normal");
   context.posAAT = m_PC->getAAT("position");
      
   // for each point of the point cloud do compute the distances
   // and gradients to any other point of the point cloud
   for (card64 i=0; i < m_PC->getNumPoints(); i++)
   {

      vector<Vector3f> gradients(m_PC->getNumPoints());
      vector<float> dists = computeGeodeticDistances(m_PC, m_TopologicalIterator, sigma, points[i], points, gradients, NULL);
      for (card64 j=0; j < points.size(); j++)
      {
         gradients[j].normalize();
         getEntry(j,i).gradient = -gradients[j];
         
         // setup default values
         ApproximatedGeodesicsDataGM14& data = getEntry(i,j);
         data.dist = dists[j];
         data.from = points[i];
         data.to = points[j];
      }
   }

  
   debugOutput << "ApproximatedGeodetics: database created - ";
   X4_TIMER_STOP(timer0);

   // free used data
   delete m_TopologicalIterator;
}

void ApproximatedGeodesicsGM14::create(X4::UnstructuredInCorePointCloud* upc, vector<X4::mpcard>& localPatchIndices, float sigma /*= 0.1f*/)
{
	mpcard i, j, n;
	vector<unsigned> indicesCopy;
	vector<Vector3f> gradients, points;
	vector<float> dists;
	AAT pos;

	Timer timer;

	if (m_PC == NULL)
		m_PC = new UnstructuredInCorePointCloud();

	n = localPatchIndices.size();
	m_PC->clearAndSetup(upc->getDescr(), n);

	m_DBWidth = n;
	m_Database.resize(n*n);
	for (i = 0; i < m_Database.size(); i++)
		m_Database[i] = new ApproximatedGeodesicsDataGM14;

	gradients.resize(n);
	points.resize(n);
	dists.resize(n);

	pos = upc->getAAT("position");
	for (i = 0; i < localPatchIndices.size(); i++)
	{
		indicesCopy.push_back((unsigned)localPatchIndices[i]);
		m_PC->getPointSet()->setVertex(i, upc->getPointSet()->getDataPointer(localPatchIndices[i]));
		points[i] = upc->getPointSet()->get3f(localPatchIndices[i], pos);
	}

	progressWindow->pushStep(true, "local geodesics database");

   // create topology iterator
	TopologyAttachment* tatt = dynamic_cast<TopologyAttachment*>(upc->getAttachments()->getData(TopologyAttachment::getDefaultName()));
	InCorePCTopologyGraph* tpg = tatt->getTopology(); 
	PointCloudIterator* top_it = tpg->createIterator(upc->getPointSet());

	for (i = 0; i < n; i++)
	{
		dists = computeGeodeticDistances(upc, top_it, sigma, indicesCopy[i], indicesCopy, gradients);

		for (j = 0; j < n; j++)
		{
			gradients[j].normalize();
			getEntry(j,i).gradient = -gradients[j];

			ApproximatedGeodesicsDataGM14& data = getEntry(i,j);
			data.dist = dists[j];
			data.from = points[i];
			data.to = points[j];
		}

		progressWindow->progressf((float32)i / (float32)n);
	}

	bool babort;
	progressWindow->popStep(&babort);

	debugOutput << "ApproximatedGeodesicsGM14::create(...) - time: " << convertTimeToString(timer.getDeltaValue()) << "\n";

	delete top_it;
}

void ApproximatedGeodesicsGM14::update(X4::UnstructuredInCorePointCloud* upc, vector<X4::mpcard>& newIndices, const X4::int32 flagVisitMask, float minSpacing, float maxSpacing, float sigma/* = 0.1f*/)
{
	const int32 FLAG_CURV_MASK = 2;

	Timer timer;

	mpcard i, j, k, l, m, n;
	mpcard imaxcurv;
	vector<mpcard> curvIndices;
	vector<unsigned> indicesCopy;
	vector<Vector3f> gradients, points, pointsNew;
	vector<float> dists;
	AAT pos, curv, flags;
	PointSet* ps;
	Vector3f vpos, vnbr;
	Vector4f vcurv, vcurvnbr;
	int32 flagspt, flagsnbr;
	float32 curvMag, curvMagNbr, curvMin, curvMax, rcpCurvRange, curvMagNbrMax;
	float32 curvFrac, sqrSpacing, sqrSpacingNbr, sqrD;

	float32 dist1, dist2, distsum, distsummin;
	mpcard imin1;
	
	ApproximatedGeodesicsDataGM14* pentry;
	vector<ApproximatedGeodesicsDataGM14*> dbcopy;
	mpcard oldwidth;

	if (m_PC == NULL)
		error("ApproximatedGeodesicsGM14::update(...): database not initialized");

	pos = upc->getAAT("position");
	curv = upc->getAAT("curvature");
	flags = upc->getAAT("flags");
	ps = upc->getPointSet();

   // create topology iterator
	TopologyAttachment* tatt = dynamic_cast<TopologyAttachment*>(upc->getAttachments()->getData(TopologyAttachment::getDefaultName()));
	InCorePCTopologyGraph* tpg = tatt->getTopology(); 
	PointCloudIterator* pc_it = tpg->createIterator(upc->getPointSet());
	TopologicalKNNIterator* top_it = new TopologicalKNNIterator(upc, pc_it);

	if (minSpacing < 0.f && maxSpacing < 0.f)
	{
		//in this case, add all points
		for (i = 0; i < newIndices.size(); i++)
		{
			curvIndices.push_back(newIndices[i]);
			pointsNew.push_back(ps->get3f(newIndices[i], pos));
		}
	}
	else
	{
		curvMax = 0.f;
		curvMin = 1e30f;
		n = newIndices.size();
		for (i = 0; i < n; i++)
		{
			j = newIndices[i];
			vpos = ps->get3f(j, pos);
			vcurv = ps->get4f(j, curv);
			curvMag = fabs(vcurv[0]) + fabs(vcurv[1]);
			if (curvMag > curvMax)	curvMax = curvMag;
			if (curvMag < curvMin)	curvMin = curvMag;
		}

		rcpCurvRange = 1.f / (curvMax - curvMin);

		for (i = 0; i < n; i++)
		{
			j = newIndices[i];
			vpos = ps->get3f(j, pos);
			vcurv = ps->get4f(j, curv);
			flagspt = ps->get1i(j, flags);

			if ((flagspt & FLAG_CURV_MASK) != 0)
				continue;

			flagspt |= FLAG_CURV_MASK;
			ps->set1i(j, flags, flagspt);

			curvMag = fabs(vcurv[0]) + fabs(vcurv[1]);
			curvFrac = (curvMag - curvMin) * rcpCurvRange;
			sqrSpacing = sqr( minSpacing * curvFrac + maxSpacing * (1.f - curvFrac) );

			imaxcurv = j;
			curvMagNbrMax = curvMag;
			sqrD = 0.f;
			top_it->setStartPointAndReset((PointIndex)j);
			top_it->next();
			while (!top_it->atEnd() && sqrD < sqrSpacing)
			{
				flagsnbr = top_it->get1i(flags);
				if ((flagsnbr & flagVisitMask) == 1 && (flagsnbr & FLAG_CURV_MASK) == 0)
				{
					vnbr = top_it->get3f(pos);
					vcurvnbr = top_it->get4f(curv);
					sqrD = (vnbr - vpos).getSqrNorm();
					curvMagNbr = fabs(vcurvnbr[0]) + fabs(vcurvnbr[1]);
					curvFrac = (curvMagNbr - curvMin) * rcpCurvRange;
					sqrSpacingNbr = sqr( minSpacing * curvFrac + maxSpacing * (1.f - curvFrac) );
					if (sqrD < sqrSpacing && sqrD < sqrSpacingNbr)
					{
						if (curvMagNbr > curvMagNbrMax)
						{
							curvMagNbrMax = curvMagNbr;
							imaxcurv = top_it->getPointSetPointNumber();
						}
						flagsnbr |= FLAG_CURV_MASK;
						ps->set1i(top_it->getPointSetPointNumber(), flags, flagsnbr);
					}
				}
				top_it->next();
			}

			curvIndices.push_back(imaxcurv);
			pointsNew.push_back(ps->get3f(imaxcurv, pos));
		}
	}

	//add points indexed in curvIndices to database
	//make a copy of the database
	oldwidth = m_DBWidth;
	dbcopy.resize(m_Database.size());
	for (i = 0; i < m_Database.size(); i++)
		dbcopy[i] = m_Database[i];

	//resize the database
	m_DBWidth += (card32)curvIndices.size();
	m_Database.resize(m_DBWidth*m_DBWidth);

	//copy previous data
	for (i = 0; i < oldwidth; i++)
	{
		for (j = 0; j < oldwidth; j++)
			m_Database[i * m_DBWidth + j] = dbcopy[i * oldwidth + j];
	}

	for (i = 0; i < m_DBWidth; i++)
	{
		if (i < oldwidth)
		{
			for (j = oldwidth; j < m_DBWidth; j++)
				m_Database[i * m_DBWidth + j] = new ApproximatedGeodesicsDataGM14;
		}
		else
		{
			for (j = 0; j < m_DBWidth; j++)
				m_Database[i * m_DBWidth + j] = new ApproximatedGeodesicsDataGM14;
		}
	}

	//compute new geodesics
	for (i = oldwidth; i < m_DBWidth; i++)
	{
		//first new-to-old
		//compute new-to-(newest old), where (newest old) is everything added last during update
		//fill vector of points for direct geodesic computation
		points.clear();
		for (j = m_lastAddition; j < oldwidth; j++)
			points.push_back(dbcopy[j * oldwidth]->from);

		dists = X4::computeGeodeticDistances(upc, pc_it, sigma, pointsNew[i - oldwidth], points, gradients);

		for (j = m_lastAddition; j < oldwidth; j++)
		{
			pentry = m_Database[i * m_DBWidth + j];
			pentry->dist = dists[j - m_lastAddition];
			pentry->from = pointsNew[i - oldwidth];
			pentry->to = points[j - m_lastAddition];
			pentry->gradient = gradients[j - m_lastAddition];
		}

		for (k = 0; k < m_lastAddition; k++)
		{
			//compute geodesic between i and k by finding j s.t. dist(i,j) + dist(k,j) is minimized
			//this works because the set indexed by j separates the set indexed by k from the set indexed by i
			distsummin = 1e30f;
			for (j = m_lastAddition; j < oldwidth; j++)
			{
				//geodesic between points j and k
				dist1 = m_Database[k * m_DBWidth + j]->dist;

				//geodesic between points i and j
				dist2 = m_Database[i * m_DBWidth + j]->dist;
				
				distsum = dist1 + dist2;
				if (distsum < distsummin)
				{
					distsummin = distsum;
					imin1 = j;
				}
			}

			j = imin1;
			pentry = m_Database[i * m_DBWidth + k];
			pentry->dist = distsummin;
			pentry->from = pointsNew[i - oldwidth];
			pentry->to = m_Database[k]->to;
			//gradient
			//the following is based on the assumption that the gradient in each entry is the gradient of the distance wrt the the 'to' point
			pentry->gradient = m_Database[i * m_DBWidth + j]->gradient + m_Database[k * m_DBWidth + j]->gradient;
		}

		//now compute new-to-new geodesics
		dists = X4::computeGeodeticDistances(upc, pc_it, sigma, pointsNew[i - oldwidth], pointsNew, gradients);

		for (j = oldwidth; j < m_DBWidth; j++)
		{
			pentry = m_Database[i * m_DBWidth + j];
			pentry->dist = dists[j - oldwidth];
			pentry->from = pointsNew[i - oldwidth];
			pentry->to = pointsNew[j - oldwidth];
			pentry->gradient = gradients[j - oldwidth];

			//indices reversed
			pentry = m_Database[j * m_DBWidth + i];
			pentry->dist = dists[j - oldwidth];
			pentry->from = pointsNew[j - oldwidth];
			pentry->to = pointsNew[i - oldwidth];
			pentry->gradient = -gradients[j - oldwidth];
		}
	}

	//now copy (except for gradient, which is negated) old-to-new from new-to-old
	for (i = 0; i < oldwidth; i++)
	{
		for (j = oldwidth; j < m_DBWidth; j++)
		{
			pentry = m_Database[i * m_DBWidth + j];
			pentry->dist = m_Database[j * m_DBWidth + i]->dist;
			pentry->from = m_Database[j * m_DBWidth + i]->to;
			pentry->to = m_Database[j * m_DBWidth + i]->from;
			pentry->gradient = -m_Database[j * m_DBWidth + i]->gradient;
		}
	}
	
	m_lastAddition = oldwidth;

	debugOutput << "ApproximatedGeodesicsGM14::update(...) - time: " << convertTimeToString(timer.getDeltaValue()) << "\n";

	delete pc_it;
}

//---------------------------------------------------------------------------
float ApproximatedGeodesicsGM14::interpolate(const X4::Vector3f& from, const X4::Vector3f& to, float sigma, X4::Vector3f& gradient, X4::card32 numNeighbours)
{
   float distance = 0.0f;
   AAT POSITION = m_PC->getAAT( "position" );
   
   if (m_hIt == NULL) m_hIt = new HierarchicalKNNIterator(m_PC, 32, NULL);
         
   // find starting point
   m_hIt->setSeekPointAndReset(from);
   unsigned startPointIndex = m_hIt->getCurrentPointIndex();

   vector<unsigned> startIndices(numNeighbours);
   vector<Vector3f> startPositions (numNeighbours);
   for( unsigned i=0; i < numNeighbours; i++)
   {
      m_hIt->next();
      startIndices[i] = m_hIt->getCurrentPointIndex();
      startPositions[i] = m_hIt->get3f(POSITION);
   }

   // find neighbour points in the near of end point
   m_hIt->setSeekPointAndReset(to);
   unsigned endPointIndex = m_hIt->getCurrentPointIndex();
 
   vector<unsigned> endIndices(numNeighbours);
   vector<Vector3f> endPositions (numNeighbours);
   for( unsigned i=0; i < numNeighbours; i++)
   {
      m_hIt->next();
      endIndices[i] = m_hIt->getCurrentPointIndex();
      endPositions[i] = m_hIt->get3f(POSITION);
   }

   // first we compute the interpolated distances from the given point to any of the end points
   vector<float> endDistances(endIndices.size());
   vector<Vector3f> endGradients(endIndices.size());
   for (unsigned j=0; j < endIndices.size(); j++)
   {
      // for each start neighbour compute interpolate weights
      vector<float> w(startIndices.size());
      float wSum = 0.0;
      for (unsigned i=0; i < startIndices.size(); i++)
      {
         // we use the inverse distance weighting method
          float d = (startPositions[i] - from).getSqrNorm();
         w[i] = exp(-d / (sigma * sigma));
         wSum += w[i];
      }
      // compute the interpolated distance
      float dist = 0.0;
      Vector3f grad = makeVector3f(0,0,0);
      for (unsigned i=0; i < startIndices.size(); i++)
      {
         dist += getEntry(startIndices[i], endIndices[j]).dist * w[i];
         grad += getEntry(startIndices[i], endIndices[j]).gradient * w[i];
      }
      dist /= wSum;
      grad /= wSum;
      
      // the distance will be used in the consecutive step for backward interpolation
      endDistances[j] = dist;
      endGradients[j] = grad;
   }
   
   // now we do the same step but interpolate the end point based on the distances interpolated before
   distance = 0.0;
   gradient = makeVector3f(0,0,0);
   {   
      vector<float> w(endIndices.size());
      float wSum = 0.0;
      for (unsigned i=0; i < endIndices.size(); i++)
      {
         // we use the inverse distance weighting method
         float d = (endPositions[i] - to).getSqrNorm();
         w[i] = exp(-d / (sigma * sigma));
         wSum += w[i];
      }
      // compute the interpolated distance
      for (unsigned i=0; i < endIndices.size(); i++)
      {
         distance += endDistances[i] * w[i];
         gradient += endGradients[i] * w[i];
      }
      distance /= wSum;
      gradient /= wSum;
   }   
mpcard i;

   return distance;
}


