/***************************************************************************
*   Copyright (c) 2008   Art Tevs                                         *
*                                                                         *
*   This library is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU Lesser General Public License as        *
*   published by the Free Software Foundation; either version 3 of        *
*   the License, or (at your option) any later version.                   *
*                                                                         *
*   This library is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU Lesse General Public License for more details.                    *
*                                                                         *
*   The full license is in LICENSE file included with this distribution.  *
***************************************************************************/

//---------------------------------------------------------------------------
#ifndef PROJECTS_AT_APPROXIMATED_GEODESICS_GM14_H
#define PROJECTS_AT_APPROXIMATED_GEODESICS_GM14_H
//---------------------------------------------------------------------------
#include "threeParameterMatchingGMOD14.h"

#include "InCorePCTopologyGraph.h"
#include "UnstructuredInCorePointCloud.h"
#include "GeometricFeature.h"
#include "TopologicalKNNIterator.h"
#include "TopologyRangeSearch.h"
#include "GeometricTools.h"

namespace X4
{

   namespace atrec
   {

      class THREE_PARAMETER_MATCHING_GMOD14_API ApproximatedGeodesicsDataGM14 : public X4::Persistent
      {
         X4_CLASS (ApproximatedGeodesicsDataGM14);

         public:
            Vector3f from;
            Vector3f to;
            float dist;
            Vector3f gradient;

            ApproximatedGeodesicsDataGM14();
            ApproximatedGeodesicsDataGM14(const ApproximatedGeodesicsDataGM14&);
            virtual ~ApproximatedGeodesicsDataGM14();         
            void assign(const X4::Object* obj, X4::X4CopyContext *context = NULL);
         
      };

      /**
       * Class which is capable of computing approximated geodesics based on some set of 
       * correct geodesics. You define a point cloud for which point-to-point geodesics are
       * computed. Then you can get geodesic of any point to any point (even if it has not been
       * existing before).
      */
      class THREE_PARAMETER_MATCHING_GMOD14_API ApproximatedGeodesicsGM14 : public X4::AttachedData
      {
         X4_CLASS (ApproximatedGeodesicsGM14);

      private:
         vector<ApproximatedGeodesicsDataGM14*> m_Database;
         IMPLEMENT_VAR_ARRAY_METHODS_STL(m_Database)
         
         vector<ApproximatedGeodesicsDataGM14*> m_FeatureDatabase;
         IMPLEMENT_VAR_ARRAY_METHODS_STL(m_FeatureDatabase);
         
         card32 m_DBWidth;
         card32 m_FeatureDBWidth;
         UnstructuredInCorePointCloud* m_PC;
         HierarchicalKNNIterator* m_hIt;
         bool m_useInterpolation;

		 //abrunton 11/12/2012
		 card32 m_lastAddition;

      public:

         
         ApproximatedGeodesicsGM14();
         virtual ~ApproximatedGeodesicsGM14();

         static string getDefaultName() {return "ApproximatedGeodesicsGM14";}

         void assign(const Object* obj, X4CopyContext *context = NULL);

         /**
          * Create preliminary geodesics database. The database will then be used to approximate new samples.
          * The given point cloud will be resampled based on curvature 
          * according to the given minSpacing and maxSpacing parameters.
          *
		  * specify minSpacing = maxSpacing < 0 in order to prevent internal resampling of the point cloud
		  **/
         void create(UnstructuredInCorePointCloud* upc, float minSpacing, float maxSpacing, float sigma = 0.1f, bool onlySetupNoGeodesics = false);
		 void create(UnstructuredInCoreTriangleMesh* upc, int subdivisionFactor, float sigma = 0.1f);

		 //abrunton 07/12/2012
		 //create a database for a local patch; this is an complete graph on the local patch, therefore patch should not be too large
		 void create(UnstructuredInCorePointCloud* upc, vector<mpcard>& localPatchIndices, float sigma = 0.1f);

		 //abrunton 10/12/2012
		 //update a database that was started with a local patch by adding additional vertices to it
		 //does not continue complete graph connectivity, but only adds high curvature points or up to max spacing
		 void update(UnstructuredInCorePointCloud* upc, vector<mpcard>& newIndices, const int32 flagVisitMask, float minSpacing, float maxSpacing, float sigma = 0.1f);

         /**
         * Create database which will hold feature to feature entries. Call this method if you like to 
         * create the database. 
         **/
         void createFeatureToFeature(FeatureSet* fs, UnstructuredInCorePointCloud* upc, float sigma = 0.1f);
		 void createFeatureToFeature(FeatureSet* fs, UnstructuredInCoreTriangleMesh* upc, int subdivisionFactor, float sigma = 0.1f);
         
		 /**
		 * Create empty feature to feature database. This can then be used to filled with correct data
		 **/
		 void createFeatureToFeature(card32 numFeatures);

         /**
          * Clear database. This will remove all cached data and no sampling is possible any more.
          **/
         void clear();
         void clearFeatureToFeature();

         /**
          * Interpolate distance and gradient of a geodesics based on previously computed database.
          **/
         float interpolate(const Vector3f& from, const Vector3f& to, float sigma = 1.0, Vector3f& gradient = Vector3f(), card32 numNeighbours = 4);

         /**
          * Get feature to feature geodesic data.
          **/
         inline ApproximatedGeodesicsDataGM14& getFeatureToFeatureEntry(X4::card32 from, X4::card32 to)
         {
			if (m_FeatureGeodeticCallback) return (*m_FeatureGeodeticCallback)(from,to,m_FeatureGeodeticCallbackUserData);
            x4Assert(m_FeatureDBWidth > 0);
            return *m_FeatureDatabase[from * m_FeatureDBWidth + to];
         }

		 /**
		 * Get database entry based. Entry corresponding to point indices is returned back
		 **/
		 inline ApproximatedGeodesicsDataGM14& getEntry(X4::card32 from, X4::card32 to)
		 {
			 if (m_DataGeodeticCallback) return (*m_DataGeodeticCallback)(from,to,m_DataGeodeticCallbackUserData);
			 x4Assert(m_DBWidth > 0);
			 return *m_Database[from * m_DBWidth + to];
		 }
          
         //! Check whenever the database is valid         
         inline bool isValid() const { return m_PC != NULL; }
         inline bool isFeatureToFeatureValid() const { return m_FeatureDBWidth > 0; }
         
         //! Get point cloud which is used for samples
         inline UnstructuredInCorePointCloud* getPointCloud() { return m_PC; }
		 inline void setPointCloud(UnstructuredInCorePointCloud* pc) { m_PC = pc; }

		 inline bool getUseInterpolation() const { return m_useInterpolation; }
		 inline void setUseInterpolation(bool use) { m_useInterpolation = use;}

         static float distanceFunction(int prev, int current, int next, TGDFContext* context);

		 typedef ApproximatedGeodesicsDataGM14& (*ComputeGeodeticCallback)(mpcard i, mpcard j, void* userData);

		 void* m_FeatureGeodeticCallbackUserData;
		 void* m_DataGeodeticCallbackUserData;
		 ComputeGeodeticCallback m_FeatureGeodeticCallback;
		 ComputeGeodeticCallback m_DataGeodeticCallback;
      };



   };
};

#endif
