/*  Time-stamp: <@(#)IntrinsicWavesCompute.cpp   12/9/2012 6:41 PM   Tino Weinkauf>
 *
 *********************************************************************
 *  File    : IntrinsicWavesCompute.cpp
 *
 *  Project : X4 Animation Reconstruction IV
 *
 *  Package : animRec4
 *
 *  Company : MPII
 *
 *  Author  : Tino Weinkauf                            Date: 12/9/2012
 *
 *  Purpose : Implementation of methods for class IntrinsicWavesCompute
 *
 *********************************************************************
 * Version History:
 *
 * V 0.10  12/9/2012 18:41:29  TW : First Revision
 *
 *********************************************************************
 */

#include "stdafx.h"

#include "IntrinsicWavesCompute.h"

#include "AR4SparseGeodesics.h"

#include <Timer.h>
#include <DebugRenderer.h>
#include <float.h>
#include <assert.h>
#include <omp.h>

#include <limits>

using namespace X4;


IntrinsicWavesCompute::IntrinsicWavesCompute()
	:RadiusScale(1.0f)
	,NumCircles(160)
	,MinRadius(0.1f)
	,MaxRadius(1.0f)
	,ClusterRadius(0.5f)
	,pManifold(NULL)
	,NumLevels(1)
{
	NumThreads = omp_get_num_procs();
}


IntrinsicWavesCompute::~IntrinsicWavesCompute()
{

}


///Small helper for computing the distances on the mesh
struct DistanceQueuePoint
{
	DistanceQueuePoint(const mpcard i, const mpcard cameFrom, const float32 ld)
		:index(i)
		,cameFromIndex(cameFrom)
		,GeoDistance(ld)
	{}

	inline bool operator<(const DistanceQueuePoint& other) const
	{
		return (GeoDistance > other.GeoDistance);
	}

	mpcard index;
	mpcard cameFromIndex;
	float32 GeoDistance;
};


///Small helper for computing the distances on the mesh
struct BorderEdge
{
	BorderEdge(const mpcard argoutIndex, const mpcard argcameFromIndex)
		:outIndex(argoutIndex)
		,cameFromIndex(argcameFromIndex)
	{}

	bool operator<(const BorderEdge& other) const
	{
		return ( (cameFromIndex < other.cameFromIndex)
			  || (cameFromIndex == other.cameFromIndex && outIndex < other.outIndex));
	}

	bool operator==(const BorderEdge& other) const
	{
		return (cameFromIndex == other.cameFromIndex && outIndex == other.outIndex);
	}

	mpcard outIndex;
	mpcard cameFromIndex;
	Vector3f exactBorderPosition;
};


///Small helper for writing
void WriteIsopointsToAmira(const vector< BorderEdge >& CrossingEdges, const mpcard CircleIdx)
{
	//Write to Amira
	char FileName[1024];
	sprintf(FileName, "D:\\Projekte\\Mesh\\dataLocal\\AnimR4\\Cat0\\InterPoints%03d.sls", CircleIdx);

	FILE* file = fopen(FileName, "wt");

	fprintf(file, "SeedingLS\nVersion 1\n\n\nNumOfSeedingLines %d\nNumOfDataPerPoint 0\n", CrossingEdges.size());
	for(int i=0;i<CrossingEdges.size();i++)
	{
		fprintf(file, "\nLinePoints 1\n%g %g %g", CrossingEdges[i].exactBorderPosition[0], CrossingEdges[i].exactBorderPosition[1], CrossingEdges[i].exactBorderPosition[2]);
	}

	fclose(file);
}

#define X4_INTRINSICWAVESCOMPUTE_ONLYFORWARDEDGES

bool IntrinsicWavesCompute::ComputeWaves(const mpcard startIndex, vector< float32 >& RawCircleLengths) const
{
	#ifdef X4_INTRINSICWAVESCOMPUTE_VISUALIZE
		string idDebugRenderer = "IntrinsicWavesCompute";
		debugRenderer->clearDebugData(idDebugRenderer);
		debugRenderer->setPointSize(4);
	#endif

	//Get data channel for boundary of the mesh
	AAT aatBoundary;
	bool bHasBoundary(false);
	UnstructuredInCorePointCloud* pc = pManifold->getPointCloud();
	if (pc->providesAttribute("boundary"))
	{
		aatBoundary = pc->getAAT("boundary");
		bHasBoundary = true;
	}
	PointSet* pRawPointSet = pc->getPointSet();

	card32 NumTotalCircles = NumCircles * NumLevels;

	//Get memory for output
	RawCircleLengths.clear();
	RawCircleLengths.resize(NumTotalCircles, -1.f);

	//Edges of the mesh that cross the current isocurve of the geodesic distance field
	vector< BorderEdge > CrossingEdges;
	CrossingEdges.reserve(50);

	const mpcard NumPoints = pManifold->getNumDiscrPoints();
	if (startIndex >= NumPoints) return false;
	if (bHasBoundary && pRawPointSet->get1f(startIndex, aatBoundary) > 0.5f) return false;

	//The distance between starting point and any other point on the mesh.
	// - computed only for as much as we need it.
	vector< float32 > DistanceField(NumPoints, FLT_MAX);
	DistanceField[startIndex] = 0;

	//Queue to put our visited vertices into
	priority_queue< DistanceQueuePoint > GrowQueue;
	// - we start with the given point and the distance to itself
	GrowQueue.push(DistanceQueuePoint(startIndex, startIndex, 0));

	//Compute isocurves of the distance field between MinRadius and MaxRadius
	//scale max radius by number of levels
	const float32 OuterMaxRadius = MaxRadius * (float32) NumLevels;
	// - distance between two adjacent circles
	const float32 CircleDist = (MaxRadius - MinRadius) / (float32)NumCircles;
	
	for(mpcard CircleIdx=0; CircleIdx<NumTotalCircles; CircleIdx++)
	{
		#ifdef X4_INTRINSICWAVESCOMPUTE_VISUALIZE
			debugRenderer->beginRenderJob_OneFrame(idDebugRenderer, CircleIdx);
		#endif

		card32 level = CircleIdx / NumCircles;
		const float32 ThisRadius = (MinRadius + (float32)CircleIdx * CircleDist) * RadiusScale * (float32)(1 << level);

		//Grow the isocurve using dijkstra until we reach the current radius
		while (!GrowQueue.empty() && GrowQueue.top().GeoDistance <= ThisRadius)
		{
			//Pop the top
			DistanceQueuePoint ThisPoint = GrowQueue.top();
			GrowQueue.pop();
			// - and get the actual position in space
			const Vector3f ThisPos = pManifold->getDiscrPoint(ThisPoint.index);

			//If we hit the boundary, then our entire wave should be ignored.
			// We can skip all of this now and return to the caller.
			if (bHasBoundary && pRawPointSet->get1f(ThisPoint.index, aatBoundary) > 0.5f)	return false;

			//Propagate distance to neighbors of the current point
			const mpcard NumNeighs = pManifold->getNumNeighbors(ThisPoint.index);
			for (mpcard i=0;i<NumNeighs;i++)
			{
				//Get the neighbor
				const mpcard NeighIdx = pManifold->getNeighborIndex(ThisPoint.index, i);

				//Compute distance between neighbor and ThisPoint; add to growing queue, if appropriate
				if (ThisPoint.GeoDistance < DistanceField[NeighIdx]) //We should not run back to where we came from
				{
					//dist = accumulated GeoDistance at ThisPoint + edge(ThisPoint ---> neighbour)
					const Vector3f NeighPos = pManifold->getDiscrPoint(NeighIdx);
					const float32 NeighGeoDistanceViaHere =
												ThisPoint.GeoDistance
												+ (NeighPos - ThisPos).getNorm();

					//If this is the shortest path to our neighbor found so far, then it follows the dijkstra algo
					if (NeighGeoDistanceViaHere < DistanceField[NeighIdx])
					{
						//Set value and push into queue
						DistanceField[NeighIdx] = NeighGeoDistanceViaHere;
						GrowQueue.push(DistanceQueuePoint(NeighIdx, ThisPoint.index, NeighGeoDistanceViaHere));

						#ifdef X4_INTRINSICWAVESCOMPUTE_ONLYFORWARDEDGES
						if (DistanceField[NeighIdx] > ThisRadius && ThisPoint.GeoDistance <= ThisRadius)
						{
							CrossingEdges.push_back(BorderEdge(NeighIdx, ThisPoint.index));
						}
						#endif
					}
				}

				#ifndef X4_INTRINSICWAVESCOMPUTE_ONLYFORWARDEDGES
				//Did we cross the current isoline?
				// If so, record this edge. We will need it later.
				if (DistanceField[NeighIdx] > ThisRadius && ThisPoint.GeoDistance <= ThisRadius)
				{
					CrossingEdges.push_back(BorderEdge(NeighIdx, ThisPoint.index));
				}
				else if (DistanceField[NeighIdx] <= ThisRadius && ThisPoint.GeoDistance > ThisRadius)
				{
					debugOutput << "Did it\n";
					CrossingEdges.push_back(BorderEdge(ThisPoint.index, NeighIdx));
				}
				#endif
			}
		}

	
		//Take care of the edges crossing our isocurve: make unique
		// This is somewhat important, since we may have introduced duplicates up there
		sort(CrossingEdges.begin(), CrossingEdges.end());
		CrossingEdges.erase(unique(CrossingEdges.begin(), CrossingEdges.end()), CrossingEdges.end());

		//Compute intersections and remove edges that do not have one
		for(mpint i=CrossingEdges.size()-1;i>=0;i--)
		{
			if (DistanceField[CrossingEdges[i].outIndex] < ThisRadius)
			{
			
				CrossingEdges.erase(CrossingEdges.begin() + i);
			}
			else
			{
				const Vector3f endPos = pManifold->getDiscrPoint(CrossingEdges[i].outIndex);
				const Vector3f startPos = pManifold->getDiscrPoint(CrossingEdges[i].cameFromIndex);
				const Vector3f direction = normalize(endPos - startPos);
				const float32 wayTooFar = DistanceField[CrossingEdges[i].outIndex] - ThisRadius;
				CrossingEdges[i].exactBorderPosition = endPos - direction * wayTooFar;

				#ifdef X4_INTRINSICWAVESCOMPUTE_VISUALIZE
					//Visualize intersection points
					//debugRenderer->addGlowingSplat(CrossingEdges[i].exactBorderPosition, 0.25, makeVector3f(1,1,0));
					debugRenderer->addSphere(CrossingEdges[i].exactBorderPosition, 0.01f, makeVector3f(0.5f,0.5f,0));
					//debugRenderer->addPoint(CrossingEdges[i].exactBorderPosition, makeVector3f(0.75,0.75,0));
				#endif
			}
		}

	
		//Compute/Estimate the length of the isocurve
		// To do so, we do not reconstruct the entire curve in a proper manner as a single line.
		// We only need its length, which is approximated anyway since we do Dijkstra.
		// Furthermore, the proper isocurve may consist of several components anyway.
		// Hence, we take the two nearest neighbors of each vertex and add the two distances.
		// After adding this up for all neighbors, we have an estimate of 2 * IsoCurveLength.
		// We make sure to look at angles and connectivity.
		// This pic shows why we need to look at both neighbors, and cannot only take the closest one:
		//
		//     x---x-----------------x---x
		//

		//For computing distances to other points
		vector< pair< float32, mpcard > > DistanceAndIdx;
		if (CrossingEdges.size() > 0) DistanceAndIdx.reserve(CrossingEdges.size() - 1);

		//Visit each crossing edge and get its closest neighbors
		float32 IsoCurveLengthApprox(0);
		for(int i=0;i<CrossingEdges.size();i++)
		{
			const BorderEdge& ThisEdge = CrossingEdges[i];

			//What is our sample spacing around here? We do not want to connect to too-far-away isopoints
			const float32 SumSampleSpacingAtEdge2 = pow(3*(pManifold->getSampleSpacing(ThisEdge.cameFromIndex) + pManifold->getSampleSpacing(ThisEdge.outIndex)), 2);

			//Get the two closest intersection points
			DistanceAndIdx.clear();
			for(int j=0;j<CrossingEdges.size();j++)
			{
				if (i == j) continue;

				const float32 Dist2 = (ThisEdge.exactBorderPosition - CrossingEdges[j].exactBorderPosition).getSqrNorm();
				if (Dist2 < SumSampleSpacingAtEdge2)
				{
					DistanceAndIdx.push_back( pair< float32, mpcard >(sqrt(Dist2), j) );
				}
			}
			sort(DistanceAndIdx.begin(), DistanceAndIdx.end());

			//What shall we do with a drunken point?
			if (DistanceAndIdx.size() <= 0) continue;

			//Take the closest one
			IsoCurveLengthApprox += DistanceAndIdx[0].first;
			#ifdef X4_INTRINSICWAVESCOMPUTE_VISUALIZE
				debugRenderer->addLine(CrossingEdges[i].exactBorderPosition, CrossingEdges[DistanceAndIdx[0].second].exactBorderPosition,
						makeVector3f(0,1,0), makeVector3f(0,1,0), 4);
			#endif

			if (DistanceAndIdx.size() <= 1)
			{
				IsoCurveLengthApprox += DistanceAndIdx[0].first; //Such that our division by 2 makes sense later
				continue;
			}

			//Get the second closest wrt. a proper angle
			bool bFound(false);
			const Vector3f MiddleFirstDirection = normalize(CrossingEdges[DistanceAndIdx[0].second].exactBorderPosition - CrossingEdges[i].exactBorderPosition);
			for(int d=1;d<DistanceAndIdx.size();d++)
			{
				const Vector3f& SecondPoint = CrossingEdges[DistanceAndIdx[d].second].exactBorderPosition;

				//Direction [middle - second]
				const Vector3f MiddleSecondDirection = normalize(SecondPoint - CrossingEdges[i].exactBorderPosition);

				//Angle between [middle - first] and [middle - second]
				const float32 CosAngle = MiddleFirstDirection * MiddleSecondDirection;

				if (CosAngle < 0)
				{
					IsoCurveLengthApprox += DistanceAndIdx[d].first;

					#ifdef X4_INTRINSICWAVESCOMPUTE_VISUALIZE
						debugRenderer->addLine(CrossingEdges[i].exactBorderPosition, SecondPoint,
							makeVector3f(0,1,0), makeVector3f(0,1,0), 4);
					#endif

					bFound = true;
					break;
				}
			}

			if (!bFound)
			{
				//Just add the second in the list
				IsoCurveLengthApprox += DistanceAndIdx[1].first;

				//debugOutput << "It happend\n";

				#ifdef X4_INTRINSICWAVESCOMPUTE_VISUALIZE
					debugRenderer->addLine(CrossingEdges[i].exactBorderPosition, CrossingEdges[DistanceAndIdx[1].second].exactBorderPosition,
						makeVector3f(0,1,0), makeVector3f(1,0,0), 4);
					debugRenderer->addGlowingSplat(CrossingEdges[DistanceAndIdx[1].second].exactBorderPosition, 0.2f, makeVector3f(1,0,0));
					//debugRenderer->addSphere(CrossingEdges[DistanceAndIdx[1].second].exactBorderPosition, 0.35, makeVector3f(0.5,0,0));
				#endif
			}
		}
		
		if (IsoCurveLengthApprox == 0.f) 
			return false;
		
		IsoCurveLengthApprox *= 0.5f;
		RawCircleLengths[CircleIdx] = IsoCurveLengthApprox / /*LengthInPlane*/(2*M_PI*ThisRadius);

	
		#ifdef X4_INTRINSICWAVESCOMPUTE_VISUALIZE
			debugRenderer->endRenderJob();
		#endif

	}

	return true;
}

void IntrinsicWavesCompute::SmoothWaves(vector<float32>& RawCircleLengths) const
{
	const float32 sigma = 0.75f;
	const float32 rcp2Sigma2 = 1.f / (2.f * sigma*sigma);
	const float32 leadFactor = 1.f / (float32)sqrt(2.0 * M_PI * (double)sigma*sigma);
	const float32 kernel[5] = { leadFactor * exp(-4.f * rcp2Sigma2), leadFactor * exp(-1.f * rcp2Sigma2), leadFactor, leadFactor * exp(-1.f * rcp2Sigma2), leadFactor * exp(-4.f * rcp2Sigma2) };
	mpcard level, c, d;

	vector<float32> paddedCircleLengths(NumCircles + 4);

	for (level = 0; level < NumLevels; level++)
	{
		const mpcard offset = level * NumCircles;
		const mpcard offsetNext = (level + 1) * NumCircles;

		if (RawCircleLengths[offsetNext - 1] < 0.f)
			break;

		paddedCircleLengths[0] = RawCircleLengths[offset + 2];
		paddedCircleLengths[1] = RawCircleLengths[offset + 1];
		for (c = offset, d = 2; c < offsetNext; c++, d++)
		{
			paddedCircleLengths[d] = RawCircleLengths[c];
		}
		paddedCircleLengths[d] = RawCircleLengths[offsetNext - 1];
		paddedCircleLengths[d + 1] = RawCircleLengths[offsetNext - 2];

		for (c = offset, d = 2; c < offsetNext; c++, d++)
		{
			RawCircleLengths[c] =	paddedCircleLengths[d - 2] * kernel[0] + 
									paddedCircleLengths[d - 1] * kernel[1] + 
									paddedCircleLengths[d + 0] * kernel[2] + 
									paddedCircleLengths[d + 1] * kernel[3] + 
									paddedCircleLengths[d + 2] * kernel[4];
		}
	}
}

void IntrinsicWavesCompute::ComputeEntropy(const vector< vector< float32 > >& AllRawCircleLengths, vector< float32 >& Entropy) const
{
	const float32 EntropySigma = 10.0f;
	const float32 EntropyDistinctiveEpsilon = 0.1f;

	////////////////////////////////////

	const float32 EntropySigma2Sq = 2 * EntropySigma * EntropySigma;
	const float32 rcpEntropySigma2Sq = 1.f / EntropySigma2Sq;

	//Setup
	const mpcard NumPoints = AllRawCircleLengths.size();
	const int64 iNumPoints(NumPoints);
	Entropy.resize(NumPoints);
	fill(Entropy.begin(), Entropy.end(), 0.0f);

	int ActualNumThreads = NumThreads;
	if (ActualNumThreads < 1)
	{
		ActualNumThreads = omp_get_num_procs() + ActualNumThreads;
	}
	if (ActualNumThreads < 1)
	{
		ActualNumThreads = 1;
	}

	//To sum up the entropy of one point - we will fill this NumPoints-times
	vector< vector< float32 > > IntermediateScalarFields(ActualNumThreads);
	vector< float32 > SumsOfIntermediateScalarFields(ActualNumThreads);
	for(int i=0;i<ActualNumThreads;i++)
	{
		IntermediateScalarFields[i].resize(NumPoints);
		SumsOfIntermediateScalarFields[i] = 0;
	}

	const card32 TotalNumCircles = NumLevels * NumCircles;

	//Compute entropy for each point
	#pragma omp parallel for num_threads(ActualNumThreads) schedule(dynamic, 10)
	for(int64 i=0;i<iNumPoints;i++)
	{
		//Shortcuts
		vector< float32 >& Field = IntermediateScalarFields[omp_get_thread_num()];
		float32& Sum = SumsOfIntermediateScalarFields[omp_get_thread_num()];
		float32 fieldSum = 0, rcpFieldSum;

		for (card32 level = 0; level < NumLevels; level++)
		{
			if (AllRawCircleLengths[i][(level + 1) * NumCircles - 1] < 0.f)
			{
				if (level == 0)
					Entropy[i] = numeric_limits< float32 >::quiet_NaN();
				break;
			}

			//Init
			Sum = 0;

			//Compute the scalar field for point i: How different is the descriptor of point i compared to all other points j
			for(mpcard j=0;j<NumPoints;j++)
			{
				if (AllRawCircleLengths[j][(level + 1) * NumCircles - 1] < 0.f)
				{
					Field[j] = -1.f;
					continue;
				}

				//Compute distance between descriptors i and j
				float32 dist = 0.f;
				for(card32 k = level * NumCircles; k < (level + 1) * NumCircles; k++)
				{
					const float32 Diff = AllRawCircleLengths[i][k] - AllRawCircleLengths[j][k];
					dist += Diff*Diff;
				}
				dist *= rcpEntropySigma2Sq;
				Field[j] = exp(-dist);
				fieldSum += Field[j];
			}

			rcpFieldSum = 1.f / fieldSum;
			for (mpcard j = 0; j < NumPoints; j++)
			{
				if (Field[j] < 0.f)	continue;
				Field[j] *= rcpFieldSum;
				Sum -= Field[j] * log(Field[j]);
			}
			Entropy[i] = max(Entropy[i], 1.f / (EntropyDistinctiveEpsilon + Sum));//so that more distinctive points have larger values
		}
	}
}


void IntrinsicWavesCompute::ComputeEntropyAsDiff(const vector< vector< float32 > >& AllRawCircleLengths, vector< float32 >& Entropy) const
{
	//Setup
	const mpcard NumPoints = AllRawCircleLengths.size();
	const int64 iNumPoints(NumPoints);
	Entropy.resize(NumPoints);
	fill(Entropy.begin(), Entropy.end(), 0.0f);

	int ActualNumThreads = NumThreads;
	if (ActualNumThreads < 1)
	{
		ActualNumThreads = omp_get_num_procs() + ActualNumThreads;
	}
	if (ActualNumThreads < 1)
	{
		ActualNumThreads = 1;
	}

	const card32 TotalNumCircles = NumLevels * NumCircles;

	//Run over all vertices and compute the descriptor difference.
	//Serial Cat0:  1m 34s 903ms
	//Parallel Cat0 8 threads on quadcore: 24s 68ms (Factor 3.96)
	#pragma omp parallel for num_threads(ActualNumThreads) schedule(dynamic, 10)
	for(int64 k=0;k<iNumPoints;k++)
	{
		for (card32 level = 0; level < NumLevels; level++)
		{
			if (AllRawCircleLengths[k][(level + 1) * NumCircles - 1] < 0.f)
			{
				if (level == 0)
					Entropy[k] = numeric_limits< float32 >::quiet_NaN();
				break;
			}

			float32 sum = 0.f;
			mpcard measureCount = 0;
			for(mpcard i=0;i<NumPoints;i++)
			{
				if (AllRawCircleLengths[i][(level + 1) * NumCircles - 1] < 0.f)
					continue;

				//Sum up the component-wise difference
				for(card32 j = level * NumCircles; j < (level + 1) * NumCircles; j++)
				{
					sum += fabs(AllRawCircleLengths[i][j] - AllRawCircleLengths[k][j]);
				}
				measureCount++;
			}

			if (measureCount > 0)
				Entropy[k] = max(Entropy[k], sum / (float)measureCount);
		}
	}
}

void IntrinsicWavesCompute::ComputeEntropySubset(const vector<mpcard>& subset, const vector< vector< float32 > >& AllRawCircleLengths, vector< float32 >& Entropy) const
{
	const float32 EntropySigma = 10.0f;
	const float32 EntropyDistinctiveEpsilon = 0.1f;

	////////////////////////////////////

	const float32 EntropySigma2Sq = 2 * EntropySigma * EntropySigma;
	const float32 rcpEntropySigma2Sq = 1.f / EntropySigma2Sq;

	//Setup
	const mpcard NumPoints = subset.size();
	const int64 iNumPoints(NumPoints);
	Entropy.resize(pManifold->getNumDiscrPoints());
	fill(Entropy.begin(), Entropy.end(), 0.0f);

	int ActualNumThreads = NumThreads;
	if (ActualNumThreads < 1)
	{
		ActualNumThreads = omp_get_num_procs() + ActualNumThreads;
	}
	if (ActualNumThreads < 1)
	{
		ActualNumThreads = 1;
	}

	//To sum up the entropy of one point - we will fill this NumPoints-times
	vector< vector< float32 > > IntermediateScalarFields(ActualNumThreads);
	vector< float32 > SumsOfIntermediateScalarFields(ActualNumThreads);
	for(int i=0;i<ActualNumThreads;i++)
	{
		IntermediateScalarFields[i].resize(NumPoints);
		SumsOfIntermediateScalarFields[i] = 0;
	}

	const card32 TotalNumCircles = NumLevels * NumCircles;

	//Compute entropy for each point
	#pragma omp parallel for num_threads(ActualNumThreads) schedule(dynamic, 10)
	for(int64 i=0;i<iNumPoints;i++)
	{
		//Shortcuts
		vector< float32 >& Field = IntermediateScalarFields[omp_get_thread_num()];
		float32& Sum = SumsOfIntermediateScalarFields[omp_get_thread_num()];
		float32 fieldSum = 0, rcpFieldSum;

		//Init
		Sum = 0;

		int64 idxi = subset[i];
		int64 idxj;

		for (card32 level = 0; level < NumLevels; level++)
		{
			if (AllRawCircleLengths[idxi][(level + 1) * NumCircles - 1] < 0.f)
			{
				if (level == 0)
					Entropy[idxi] = numeric_limits< float32 >::quiet_NaN();
				break;
			}

			//Compute the scalar field for point i: How different is the descriptor of point i compared to all other points j
			for(mpcard j=0;j<NumPoints;j++)
			{
				idxj = subset[j];
				if (AllRawCircleLengths[idxj][(level + 1) * NumCircles - 1] < 0.f)
				{
					Field[j] = -1.f;
					continue;
				}

				//Compute distance between descriptors i and j
				float32 dist = 0.f;
				for(card32 k = level * NumCircles; k < (level + 1) * NumCircles; k++)
				{
					const float32 Diff = AllRawCircleLengths[idxi][k] - AllRawCircleLengths[idxj][k];
					dist += Diff*Diff;
				}
				dist *= rcpEntropySigma2Sq;
				Field[j] = exp(-dist);
				fieldSum += Field[j];
			}

			rcpFieldSum = 1.f / fieldSum;
			for (mpcard j = 0; j < NumPoints; j++)
			{
				if (Field[j] < 0.f)	continue;
				Field[j] *= rcpFieldSum;
				Sum -= Field[j] * log(Field[j]);
			}
			Entropy[idxi] = max(Entropy[idxi], 1.f / (EntropyDistinctiveEpsilon + Sum));//so that more distinctive points have larger values
		}
	}
}

void IntrinsicWavesCompute::ComputeEntropyAsDiffSubset(const vector<mpcard>& subset, const vector< vector< float32 > >& AllRawCircleLengths, vector< float32 >& Entropy) const
{
	//Setup
	const mpcard NumPoints = subset.size();
	const int64 iNumPoints(NumPoints);
	Entropy.resize(pManifold->getNumDiscrPoints());
	fill(Entropy.begin(), Entropy.end(), 0.0f);

	int ActualNumThreads = NumThreads;
	if (ActualNumThreads < 1)
	{
		ActualNumThreads = omp_get_num_procs() + ActualNumThreads;
	}
	if (ActualNumThreads < 1)
	{
		ActualNumThreads = 1;
	}

	int64 idxk, idxi;

	card32 TotalNumCircles = NumLevels * NumCircles;

	//run over all points in subset
//	#pragma omp parallel for num_threads(ActualNumThreads) schedule(dynamic, 10)
	for(int64 k=0;k<iNumPoints;k++)
	{
		idxk = subset[k];
		for (card32 level = 0; level < NumLevels; level++)
		{
			if (AllRawCircleLengths[idxk][(level + 1) * NumCircles - 1] < 0.f)
			{
				if (level == 0)
					Entropy[idxk] = numeric_limits< float32 >::quiet_NaN();
				break;
			}

			float32 sum = 0.f;
			mpcard measureCount = 0;
			for(mpcard i=0;i<NumPoints;i++)
			{
				idxi = subset[i];
				if (AllRawCircleLengths[idxi][(level + 1) * NumCircles - 1] < 0.f)
					continue;

				//Sum up the component-wise difference
				for(card32 j = level * NumCircles; j < (level + 1) * NumCircles; j++)
				{
					sum += fabs(AllRawCircleLengths[idxi][j] - AllRawCircleLengths[idxk][j]);
				}
				measureCount++;
			}

			if (measureCount > 0)
				Entropy[idxk] = max(Entropy[idxk], sum / (float)measureCount);
		}
	}
}


void IntrinsicWavesCompute::ComputeMostSignificantFeaturePoints(const vector< float32 >& Entropy, const int32 DesiredNum, vector< mpcard >& MostSignificantMaxima) const
{
	//We assume that the entropy is from this manifold
	assert(Entropy.size() == pManifold->getNumDiscrPoints());

	//Run over all vertices and extract local maxima
	vector< TEntropyAndIdx > LocalMaxima;
	LocalMaxima.reserve(100);
	const mpcard NumPoints = pManifold->getNumDiscrPoints();
	for(mpcard i=0;i<NumPoints;i++)
	{
		if (!_finite(Entropy[i])) continue;

		//Get its neighbors and check if they have a smaller entropy value
		const mpcard NumNeighs = pManifold->getNumNeighbors(i);
		bool bIsMaximum(true);
		for (mpcard j=0;j<NumNeighs;j++)
		{
			//Get the neighbor
			const mpcard NeighIdx = pManifold->getNeighborIndex(i, j);

			//Is the neighbor larger?
			if (_finite(Entropy[NeighIdx]) && Entropy[NeighIdx] > Entropy[i])
			{
				//Vertex i is not a maximum
				bIsMaximum = false;
				break;
			}
		}

		//Add to our maximum list
		if (bIsMaximum)
		{
			LocalMaxima.push_back(TEntropyAndIdx(Entropy[i], i));
		}
	}

	//Sort the local maxima according to their entropy
	sort(LocalMaxima.begin(), LocalMaxima.end(), greater< TEntropyAndIdx >());

	//Cluster them
	ClusterFeatures(LocalMaxima, (DesiredNum >= 0) ? DesiredNum : INT_MAX, MostSignificantMaxima);


	
}

void IntrinsicWavesCompute::ComputeMostSignificantFeaturePointsEntropy(const vector<float32>& Entropy, const int32 DesiredNum, vector<mpcard>& MostDistinctPoints) const
{
	//We assume that the entropy is from this manifold
	assert(Entropy.size() == pManifold->getNumDiscrPoints());

	//Run over all vertices and extract local minima
	vector< TEntropyAndIdx > LocalMinima;
	LocalMinima.reserve(DesiredNum);
	const mpcard NumPoints = pManifold->getNumDiscrPoints();

	for(mpcard i=0;i<NumPoints;i++)
	{
		if (!_finite(Entropy[i])) continue;

		//Get its neighbors and check if they have a smaller entropy value
		const mpcard NumNeighs = pManifold->getNumNeighbors(i);
		bool bIsMin(true);
		for (mpcard j=0;j<NumNeighs;j++)
		{
			//Get the neighbor
			const mpcard NeighIdx = pManifold->getNeighborIndex(i, j);

			//Is the neighbor larger?
			if (_finite(Entropy[NeighIdx]) && Entropy[NeighIdx] < Entropy[i])
			{
				//Vertex i is not a minimum
				bIsMin = false;
				break;
			}
		}

		//Add to our minimum list
		if (bIsMin)
		{
			LocalMinima.push_back(TEntropyAndIdx(Entropy[i], i));
		}
	}

	//Sort the local minima according to their entropy
	sort(LocalMinima.begin(), LocalMinima.end(), less< TEntropyAndIdx >());

	//Cluster them
	ClusterFeatures(LocalMinima, (DesiredNum >= 0) ? DesiredNum : INT_MAX, MostDistinctPoints);
}

void IntrinsicWavesCompute::ComputeSecondaryFeatures(const mpcard primaryFeature, const mpcard DesiredNum, const vector< vector< float32 > >& AllRawCircleLengths, vector<mpcard>& secondaryFeatures, vector<float>& localEntropy)
{
	mpcard i, j, k;

	secondaryFeatures.clear();
	localEntropy.clear();

	//Get some infos
	const mpcard NumPoints = pManifold->getNumDiscrPoints();
	// - actual radius
	const float32 innerRadius = MinRadius * RadiusScale * (float32)(2);
	const float32 outerRadius = MaxRadius * RadiusScale * (float32)(1 << (NumLevels - 1));

	vector<bool> withinBand(NumPoints, false);
	vector< float32 > DistanceField(NumPoints);

	//mark the points that are within the geodesic band we are interested
	MarkGeodesicBand(primaryFeature, innerRadius, outerRadius, DistanceField, withinBand);

	vector<mpcard> secondaryCandidates;
	vector<float> localEntropyCandidates;
	vector<TEntropyAndIdx> localEntropyIndices;

	for (i = 0; i < NumPoints; i++)
	{
		if (withinBand[i])
			secondaryCandidates.push_back(i);
	}

	ComputeEntropyAsDiffSubset(secondaryCandidates, AllRawCircleLengths, localEntropyCandidates);

	for (i = 0; i < secondaryCandidates.size(); i++)
	{
		if (!_finite(localEntropyCandidates[i])) continue;

		bool bLocalZenith = true;
		const mpcard numNbrs = pManifold->getNumNeighbors(secondaryCandidates[i]);
		for (j = 0; j < numNbrs; j++)
		{
			k = pManifold->getNeighborIndex(secondaryCandidates[i], j);
			if (withinBand[k])
			{
				if (localEntropyCandidates[k] > localEntropyCandidates[secondaryCandidates[i]])
				{
					bLocalZenith = false;
					break;
				}
			}
		}

		if (bLocalZenith)
			localEntropyIndices.push_back(TEntropyAndIdx(localEntropyCandidates[secondaryCandidates[i]], secondaryCandidates[i]));
	}

	sort(localEntropyIndices.begin(), localEntropyIndices.end(), greater< TEntropyAndIdx >());

	for (i = 0; i < localEntropyIndices.size() && (DesiredNum == 0 || i < DesiredNum); i++)
	{
		secondaryFeatures.push_back(localEntropyIndices[i].second);
		localEntropy.push_back(localEntropyIndices[i].first);
	}
}

void IntrinsicWavesCompute::EnsureSamplingDensity(const vector<float32>& Entropy, const float32 desiredSpacing, vector<mpcard>& MostSignificantPoints) const
{
	const mpcard numPoints = pManifold->getNumDiscrPoints();
	const mpcard numFeaturesMax = numPoints / 4;
	const mpcard numInitFeatures = MostSignificantPoints.size();

	const float32 halfDesiredSpacing = desiredSpacing * 0.5;
	const float32 distSigma = halfDesiredSpacing / 3.0;//want zero distance to nearest feature to be 3sigma probability
	const float32 distLambda = 1.0 / distSigma;
	const float32 distSpaceSigma = desiredSpacing * 0.1;
	const float32 rcp2DistSpaceSigma2 = 1.0 / (2.0 * distSpaceSigma*distSpaceSigma);

	vector<bool> alreadyFeature(numPoints, false);
	vector<card32> numFeaturesClose(numPoints, 0);

	mpcard i, j;

	AR4StaticSparseGeodesics* pGeos = new AR4StaticSparseGeodesics();

	pGeos->setManifold(pManifold);

	bool bLonelyPoint = false;

	pGeos->distanceTransform(MostSignificantPoints);
	for (i = 0; i < numPoints; i++)
	{
		if (!_finite(Entropy[i]))			continue;
		if (pGeos->getDistanceField(i) > desiredSpacing)
		{
			bLonelyPoint = true;
			break;
		}
	}

	while (bLonelyPoint)
	{
		float32 topEntropy = 0.f;
		mpcard mostSigPoint = numPoints;
		bLonelyPoint = false;

		for (i = 0; i < numPoints; i++)
		{
			if (!_finite(Entropy[i]))			continue;

			const float32 dist = pGeos->getDistanceField(i);

			if (dist > desiredSpacing)
			{
				bLonelyPoint = true;

				const float32 distCost = max(0.0, (double) halfDesiredSpacing - dist);
				const float32 distImport = distLambda * exp(-distCost * distLambda);
				const float32 distSpaceDiff = dist - desiredSpacing;
				const float32 distSpaceCost = distSpaceDiff*distSpaceDiff;
				const float32 distSpaceImport = exp(-distSpaceCost * rcp2DistSpaceSigma2);

				const float32 modifiedEntropy = Entropy[i] * distSpaceImport * distImport;

				if (modifiedEntropy > topEntropy)
				{
					topEntropy = modifiedEntropy;
					mostSigPoint = i;
				}
			}
		}

		if (mostSigPoint < numPoints)
		{
			MostSignificantPoints.push_back(mostSigPoint);
			pGeos->distanceTransform(MostSignificantPoints);
		}
	}

	delete pGeos;
}

void IntrinsicWavesCompute::EnsureSamplingDensity(const vector<float32>& Entropy, const float32 desiredSpacing, const mpcard desiredCount, vector<mpcard>& MostSignificantPoints) const
{
	const mpcard numPoints = pManifold->getNumDiscrPoints();
	const mpcard numFeaturesMax = numPoints / 4;
	const mpcard numInitFeatures = MostSignificantPoints.size();

	const float32 halfDesiredSpacing = desiredSpacing * 0.5;
	const float32 distSigma = halfDesiredSpacing / 3.0;//want zero distance to nearest feature to be 3sigma probability
	const float32 distLambda = 1.0 / distSigma;
	const float32 distSpaceSigma = desiredSpacing * 0.1;
	const float32 rcp2DistSpaceSigma2 = 1.0 / (2.0 * distSpaceSigma*distSpaceSigma);

	vector<bool> alreadyFeature(numPoints, false);
	vector<card32> numFeaturesClose(numPoints, 0);

	mpcard i, j;
	mpcard iLastLonely, lastLonelyNbrs;

	AR4StaticSparseGeodesics* pGeos = new AR4StaticSparseGeodesics();

	pGeos->setManifold(pManifold);

	bool bLonelyPoint = false;

	for (i = 0; i < numInitFeatures; i++)
	{
		vector<mpcard> zeroSet(1, MostSignificantPoints[i]);
		pGeos->distanceTransform(zeroSet);

		mpcard numNear = 0;
		for (j = 0; j < numInitFeatures; j++)
		{
			if (j == i)	continue;
			if (pGeos->getDistanceField(MostSignificantPoints[i]) < desiredSpacing)
				numNear++;
		}

		if (numNear < desiredCount)
		{
			bLonelyPoint = true;
			iLastLonely = i;
			lastLonelyNbrs = numNear;
			break;
		}
	}

	vector<pair<float32, mpcard>> pointSigOrder(numPoints);

	for (i = 0; i < numPoints; i++)
	{
		pointSigOrder[i].first = Entropy[i];
		pointSigOrder[i].second = i;
	}

	while (bLonelyPoint)
	{
		bLonelyPoint = false;

		for (i = 0; i < numPoints; i++)
		{
			if (!_finite(Entropy[i]))			continue;

			const float32 dist = pGeos->getDistanceField(i);

			const float32 distCost = max(0.0, (double) halfDesiredSpacing - dist);
			const float32 distImport = distLambda * exp(-distCost * distLambda);
			const float32 distSpaceDiff = dist - desiredSpacing;
			const float32 distSpaceCost = distSpaceDiff*distSpaceDiff;
			const float32 distSpaceImport = exp(-distSpaceCost * rcp2DistSpaceSigma2);

			const float32 modifiedEntropy = Entropy[i] * distSpaceImport * distImport;

			pointSigOrder[i].first = modifiedEntropy;
			pointSigOrder[i].second = i;
		}

		for (i = 0; i < numInitFeatures; i++)
			pointSigOrder[i].second = 0.f;

		sort(pointSigOrder.begin(), pointSigOrder.end(), less<pair<float32, mpcard>>());

		i = 0;
		while (lastLonelyNbrs < desiredCount)
		{
			MostSignificantPoints.push_back(pointSigOrder[i].second);
			lastLonelyNbrs++;
			i++;
		}

		const mpcard numFeatures = MostSignificantPoints.size();
		for (i = iLastLonely + 1; i < numFeatures; i++)
		{
			vector<mpcard> zeroSet(1, MostSignificantPoints[i]);
			pGeos->distanceTransform(zeroSet);

			mpcard numNear = 0;
			for (j = 0; j < numFeatures; j++)
			{
				if (j == i)	continue;
				if (pGeos->getDistanceField(MostSignificantPoints[i]) < desiredSpacing)
					numNear++;
			}

			if (numNear < desiredCount)
			{
				bLonelyPoint = true;
				iLastLonely = i;
				lastLonelyNbrs = numNear;
				break;
			}
		}
	}
}


///Computes union of A and B. Saves the result in B. @returns whether there was an overlap.
bool Union(const vector< bool >& A, vector< bool >& B)
{
	bool bOverlap(false);

	const mpcard Num = A.size();
	for(mpcard i=0;i<Num;i++)
	{
		const bool bA(A[i]);
		const bool bB(B[i]);

		if (bA && bB) bOverlap = true;

		B[i] = bA || bB;
	}

	return bOverlap;
}


void IntrinsicWavesCompute::ClusterFeatures(const vector< TEntropyAndIdx >& LocalMaxima, const int32 DesiredNum, vector< mpcard >& MostSignificantMaxima) const
{
	//Assumption: LocalMaxima is sorted by entropy; global max coming first

	//Prepare memory
	MostSignificantMaxima.clear();
	MostSignificantMaxima.reserve(min((int) DesiredNum, (int) LocalMaxima.size()));

	//Get some infos
	const mpcard NumPoints = pManifold->getNumDiscrPoints();
	// - actual radius
	const float32 ActualClusterRadius = ClusterRadius * RadiusScale;

	//Where we leave our marks
	vector< bool > GlobalMarkedOnMesh(NumPoints, false);
	vector< bool > LocalMarkedOnMesh(NumPoints, false);
	//The distance field for the dijkstra
	vector< float32 > DistanceField(NumPoints);

	//Run over all local maxima, mark their influence area on the mesh (first one wins!), and collect all winners.
	const mpcard NumLocalMaxima = LocalMaxima.size();
	for(mpcard i=0;i<NumLocalMaxima&&MostSignificantMaxima.size()<DesiredNum;i++)//Isthereanybetterprogramminglanguage?
	{
		//Where is the current maximum on the mesh?
		const mpcard idxCurrentMax = LocalMaxima[i].second;

		//Leave a mark (geodesic circle) on the mesh and check if it intersects an existing one
		// If it intersects an existing one, it is part of a different cluster (with a larger max)
		fill(LocalMarkedOnMesh.begin(), LocalMarkedOnMesh.end(), false);
		MarkGeodesicArea(idxCurrentMax, ActualClusterRadius, DistanceField, LocalMarkedOnMesh);
		const bool bOverlap = Union(LocalMarkedOnMesh, GlobalMarkedOnMesh);
		if (!bOverlap)
		{
			MostSignificantMaxima.push_back(idxCurrentMax);
		}
		//else: it intersected; ignore this maxima
	}
}


void IntrinsicWavesCompute::MarkGeodesicArea(const mpcard startIndex, const float32 ActualClusterRadius,
					  vector< float32 >& DistanceField, vector< bool >& MarkedOnMesh) const
{
	//The distance between starting point and any other point on the mesh.
	// - computed only for as much as we need it.
	fill(DistanceField.begin(), DistanceField.end(), FLT_MAX);
	DistanceField[startIndex] = 0;

	//Queue to put our visited vertices into
	priority_queue< DistanceQueuePoint > GrowQueue;
	// - we start with the given point and the distance to itself
	GrowQueue.push(DistanceQueuePoint(startIndex, startIndex, 0));

	//Grow the isocurve using dijkstra until we reached the given radius
	while (!GrowQueue.empty() && GrowQueue.top().GeoDistance <= ActualClusterRadius)
	{
		//Pop the top
		DistanceQueuePoint ThisPoint = GrowQueue.top();
		GrowQueue.pop();
		// - and get the actual position in space
		const Vector3f ThisPos = pManifold->getDiscrPoint(ThisPoint.index);
		// - mark that we have been here
		MarkedOnMesh[ThisPoint.index] = true;

		//Propagate distance to neighbors of the current point
		const mpcard NumNeighs = pManifold->getNumNeighbors(ThisPoint.index);
		for (mpcard i=0;i<NumNeighs;i++)
		{
			//Get the neighbor
			const mpcard NeighIdx = pManifold->getNeighborIndex(ThisPoint.index, i);

			//Compute distance between neighbor and ThisPoint; add to growing queue, if appropriate
			if (ThisPoint.GeoDistance < DistanceField[NeighIdx]) //We should not run back to where we came from
			{
				//dist = accumulated GeoDistance at ThisPoint + edge(ThisPoint ---> neighbour)
				const Vector3f NeighPos = pManifold->getDiscrPoint(NeighIdx);
				const float32 NeighGeoDistanceViaHere =
											ThisPoint.GeoDistance
											+ (NeighPos - ThisPos).getNorm();

				//If this is the shortest path to our neighbor found so far, then it follows the dijkstra algo
				if (NeighGeoDistanceViaHere < DistanceField[NeighIdx])
				{
					//Set value and push into queue
					DistanceField[NeighIdx] = NeighGeoDistanceViaHere;
					GrowQueue.push(DistanceQueuePoint(NeighIdx, ThisPoint.index, NeighGeoDistanceViaHere));
				}
			}
		}
	}
}

void IntrinsicWavesCompute::MarkGeodesicBand(const mpcard startIndex, const float32 innerRadius, const float32 outerRadius, vector< float32 >& DistanceField, vector< bool >& MarkedOnMesh) const
{
	//The distance between starting point and any other point on the mesh.
	// - computed only for as much as we need it.
	fill(DistanceField.begin(), DistanceField.end(), FLT_MAX);
	DistanceField[startIndex] = 0;

	//Queue to put our visited vertices into
	priority_queue< DistanceQueuePoint > GrowQueue;
	// - we start with the given point and the distance to itself
	GrowQueue.push(DistanceQueuePoint(startIndex, startIndex, 0));

	//Grow the isocurve using dijkstra until we reached the given radius
	while (!GrowQueue.empty())
	{
		//Pop the top
		DistanceQueuePoint ThisPoint = GrowQueue.top();
		GrowQueue.pop();

		// - if this point is in the band we want, mark it
		// - if it is outside the outer ring, we are done
		if (ThisPoint.GeoDistance > outerRadius)
			break;
		if (ThisPoint.GeoDistance >= innerRadius)
			MarkedOnMesh[ThisPoint.index] = true;

		// - and get the actual position in space
		const Vector3f ThisPos = pManifold->getDiscrPoint(ThisPoint.index);

		//Propagate distance to neighbors of the current point
		const mpcard NumNeighs = pManifold->getNumNeighbors(ThisPoint.index);
		for (mpcard i=0;i<NumNeighs;i++)
		{
			//Get the neighbor
			const mpcard NeighIdx = pManifold->getNeighborIndex(ThisPoint.index, i);

			//Compute distance between neighbor and ThisPoint; add to growing queue, if appropriate
			if (ThisPoint.GeoDistance < DistanceField[NeighIdx]) //We should not run back to where we came from
			{
				//dist = accumulated GeoDistance at ThisPoint + edge(ThisPoint ---> neighbour)
				const Vector3f NeighPos = pManifold->getDiscrPoint(NeighIdx);
				const float32 NeighGeoDistanceViaHere =
											ThisPoint.GeoDistance
											+ (NeighPos - ThisPos).getNorm();

				//If this is the shortest path to our neighbor found so far, then it follows the dijkstra algo
				if (NeighGeoDistanceViaHere < DistanceField[NeighIdx])
				{
					//Set value and push into queue
					DistanceField[NeighIdx] = NeighGeoDistanceViaHere;
					GrowQueue.push(DistanceQueuePoint(NeighIdx, ThisPoint.index, NeighGeoDistanceViaHere));
				}
			}
		}
	}
}


