/*  Time-stamp: <@(#)IntrinsicWavesCompute.h   12/9/2012 6:41 PM   Tino Weinkauf>
 *
 *********************************************************************
 *  File    : IntrinsicWavesCompute.h
 *
 *  Project : X4 Animation Reconstruction IV
 *
 *  Package : animRec4
 *
 *  Company : MPII
 *
 *  Author  : Tino Weinkauf                            Date: 12/9/2012
 *
 *  Purpose : Declaration of class IntrinsicWavesCompute
 *
 *********************************************************************
 * Version History:
 *
 * V 0.10  12/9/2012 18:41:16  TW : First Revision
 *
 *********************************************************************
 */

#ifndef ANIMREC4_INTRINSICWAVESCOMPUTE_H
#define ANIMREC4_INTRINSICWAVESCOMPUTE_H

#include "threeParameterMatchingGMOD14.h"
#include "X4Types.h"

#include "AR4MLSManifold.h"

namespace X4
{

/** Computes seeds based on intrinsic waves, or better to say: based on the characteristics
	of isocurves of the geodesic distance function. These isocurves change
	their length/circumference in a very descriptive manner. This is used here to find
	the most discriminative vertices/points on a surface.

@author Tino Weinkauf
*/
class THREE_PARAMETER_MATCHING_GMOD14_API IntrinsicWavesCompute
{
//Friends
//Types
private:
	typedef pair< float32, mpcard > TEntropyAndIdx;

//Construction / Deconstruction
public:
	IntrinsicWavesCompute();
	virtual ~IntrinsicWavesCompute();

//Methods
public:
	///Does the actual computation.
	bool ComputeWaves(const mpcard startIndex, vector< float32 >& RawCircleLengths) const;

	void SmoothWaves(vector<float32>& RawCircleLengths) const;

	void ComputeEntropy(const vector< vector< float32 > >& AllRawCircleLengths, vector< float32 >& Entropy) const;
	void ComputeEntropyAsDiff(const vector< vector< float32 > >& AllRawCircleLengths, vector< float32 >& Entropy) const;
	void ComputeEntropySubset(const vector<mpcard>& subset, const vector< vector< float32 > >& AllRawCircleLengths, vector< float32 >& Entropy) const;
	void ComputeEntropyAsDiffSubset(const vector<mpcard>& subset, const vector< vector< float32 > >& AllRawCircleLengths, vector< float32 >& Entropy) const;

	void ComputeMostSignificantFeaturePoints(const vector< float32 >& Entropy, const int32 DesiredNum, vector< mpcard >& MostSignificantMaxima) const;
	void ComputeMostSignificantFeaturePointsEntropy(const vector<float32>& Entropy, const int32 DesiredNum, vector<mpcard>& MostDistinctPoints) const;

	void ComputeSecondaryFeatures(const mpcard primaryFeature, const mpcard DesiredNum, const vector< vector< float32 > >& AllRawCircleLengths, vector<mpcard>& secondaryFeatures, vector<float>& localEntropy);

	void EnsureSamplingDensity(const vector<float32>& Entropy, const float32 desiredSpacing, vector<mpcard>& MostSignificantPoints) const;
	void EnsureSamplingDensity(const vector<float32>& Entropy, const float32 desiredSpacing, const mpcard desiredCount, vector<mpcard>& MostSignificantPoints) const;

private:
	void ClusterFeatures(const vector< TEntropyAndIdx >& LocalMaxima, const int32 DesiredNum, vector< mpcard >& MostSignificantMaxima) const;

	void MarkGeodesicArea(const mpcard startIndex, const float32 ActualClusterRadius, vector< float32 >& DistanceField, vector< bool >& MarkedOnMesh) const;
	void MarkGeodesicBand(const mpcard startIndex, const float32 innerRadius, const float32 outerRadius, vector< float32 >& DistanceField, vector< bool >& MarkedOnMesh) const;

//Attributes
public:
	float32 RadiusScale;
	card32 NumCircles;
	float32 MinRadius;
	float32 MaxRadius;
	float32 ClusterRadius;
	AR4MLSManifold3f* pManifold;

	card32 NumLevels;

	///Number of OpenMP Threads for parallel computation.
	int32 NumThreads;
};

}

#endif //ANIMREC4_INTRINSICWAVESCOMPUTE_H

