//---------------------------------------------------------------------------
//PCCTopologyAndFeaturePreservingResampling.cpp
//Source file for the PCCTopologyAndFeaturePreservingResampling class
//This class is basically a copy of PCCTopologyPreservingResampling, with
//an added check for a "feature flag" so that vertices marked as features
//do not get removed
//Alan Brunton, MPII, February 2013
//---------------------------------------------------------------------------
#include "StdAfx.h"
//---------------------------------------------------------------------------
#include "PCCTopologyAndFeaturePreservingResampling.h"
//---------------------------------------------------------------------------
#include "NumericalClassProperty.h"
#include "DebugOutput.h"
#include "PointCloudIterators.h"
#include "PointSet.h"
#include "PointSet.inline.h"
#include "UnstructuredInCorePointCloud.h"
#include "PointCloudTools.h"
#include "SceneGraphState.h"
#include "StandardSceneGraphIterator.h"
#include "SGObjectNode.h"
#include "SGAnimationNode.h"
#include "ProgressWindow.h"
#include "SeparatorClassProperty.h"
#include "InCorePCTopologyGraph.h"
#include "TopologyAttachment.h"
#include "TopologicalKNNIterator.h"
#include "GeometricTools.h"

//---------------------------------------------------------------------------
using namespace X4;
//---------------------------------------------------------------------------

IMPLEMENT_X4_CLASS( PCCTopologyAndFeaturePreservingResampling, 2)
{
   BEGIN_CLASS_INIT( PCCTopologyAndFeaturePreservingResampling )
   ADD_FLOAT32_PROP( m_Spacing, 1 )
   ADD_FLOAT32_PROP ( m_ReconnectFactor, 2 );
   ADD_CARD32_PROP( m_featureLevel, 0 )
}

string PCCTopologyAndFeaturePreservingResampling::m_strFeatureFlag = "FeatureFlag";

PCCTopologyAndFeaturePreservingResampling::PCCTopologyAndFeaturePreservingResampling()
{
   m_Spacing = 0.01f;
   m_ReconnectFactor = 3.1f;   
   m_featureLevel = 1;
}

PCCTopologyAndFeaturePreservingResampling::~PCCTopologyAndFeaturePreservingResampling()
{
}

void PCCTopologyAndFeaturePreservingResampling::resamplePC(UnstructuredInCorePointCloud* pc)
{
	// copy point cloud
	UnstructuredInCorePointCloud* pcCopy = copyToUnstructuredInCorePointCloud(pc);
	AAT POSITION = pcCopy->getAAT( "position" );
	AAT FLAGS = pcCopy->getAAT( "flags" );

	// get topology
	if (!pc->getAttachments()->getData(TopologyAttachment::getDefaultName()))
	{
		warning("Selected pointcloud does not contain topology information");
		return;
	}

	InCorePCTopologyGraph* tpg = NULL;
	TopologyAttachment* tatt = dynamic_cast<TopologyAttachment*>(pc->getAttachments()->getData(TopologyAttachment::getDefaultName()));
	tpg = tatt->getTopology();
	InCorePCTopologyGraph* tpgCopy = tpg->copy();
	tatt = new TopologyAttachment;
	tatt->setup(TopologyAttachment::getDefaultName(), AttachedData::ADF_PERSISTENT);
	tatt->setTopology(tpgCopy);            
	pcCopy->getAttachments()->attachData(tatt);         

	// Create iterators
	SceneObjectIterator* it = pcCopy->createIterator( SOIT( SOIT::CAP_BASIC_PC | SOIT::CAP_MODIFYABLE_PC | SOIT::CAP_DYNAMIC_PC | SOIT::CAP_INDEXED_PC) );
	BasicPointCloudIterator* basicIt = dynamic_cast<BasicPointCloudIterator*>( it );
	ModifyablePointCloudIterator* modIt = dynamic_cast<ModifyablePointCloudIterator*>( it );
	DynamicPointCloudIterator* dynIt = dynamic_cast<DynamicPointCloudIterator*>( it );
	IndexedPointCloudIterator* indIt = dynamic_cast<IndexedPointCloudIterator*>(it);

	ModifyableKNNIterator* knnIt = new ModifyableKNNIterator( pcCopy, 32, NULL );
	knnIt->reset();

	//abrunton: check for presence of feature flag
	AAT FEATS;
	bool bFeatures = pc->providesAttribute(m_strFeatureFlag);
	if (bFeatures)
	{
		FEATS = pc->getAAT(m_strFeatureFlag);
	}

	// Reset all flags
	basicIt->reset();
	while( !basicIt->atEnd() )
	{
		modIt->set1i( FLAGS, 0 );
		basicIt->next();
	}

	card32 numPointStart = pc->getNumPoints();
	card32 numNewPoints = numPointStart;
	float32 sqrSpacing = sqr( m_Spacing );

	// first, mark all features as keepers and non-features within spacing of them as ones to delete
	if (bFeatures)
	{
		basicIt->reset();
		while (!basicIt->atEnd())
		{
			const Vector3f& p = basicIt->get3f( POSITION );
			const int32 flag = basicIt->get1i( FLAGS );

			//if this point is a sufficiently strong feature, mark it as a keeper
			if (basicIt->get1u( FEATS ) >= m_featureLevel)
			{
				modIt->set1i(FLAGS, 2);

				//points within sample spacing have their flags set to 1, which indicates they should be removed (below)
				// find the point and get to the nearest neighbour
				knnIt->setSeekPointAndReset( p );
				knnIt->next();			// Skip self 
				if (!knnIt->atEnd())
				{
					float32 sqrD = 0;         
					do 
					{
						//the code within the if statement checks the distance of the point and marks it for removal (setting flag to 1)
						//if it is within m_Spacing of the current point
						//the condition of the if statement stipulates that the point has not yet been marked as a keeper or one to remove (first part)
						//AND that it is not a feature (second part) -- so that we don't mark features for removal
						if (knnIt->get1i(FLAGS) == 0 && knnIt->get1u(FEATS) < m_featureLevel) 
						{            
							const Vector3f& np = knnIt->get3f( POSITION );
							sqrD = ( p - np ) * ( p - np );
							if (sqrD < sqrSpacing)
							{
								knnIt->set1i( FLAGS, 1 );
								numNewPoints--;
							}
						}
						knnIt->next();
					}while( sqrD < sqrSpacing && !knnIt->atEnd());
				}
			}
			basicIt->next();
		}
	}

	// Mark all delete points
	basicIt->reset();
	while( !basicIt->atEnd() )
	{
		const Vector3f& p = basicIt->get3f( POSITION );
		const int32 flag = basicIt->get1i( FLAGS );

		// point was not visited before 
		if (flag == 0)
		{      
			// mark point as visited
			modIt->set1i(FLAGS, 2);

			//points within sample spacing have their flags set to 1, which indicates they should be removed (below)
			// find the point and get to the nearest neighbour
			knnIt->setSeekPointAndReset( p );
			knnIt->next();			// Skip self 
			if (!knnIt->atEnd())
			{
				float32 sqrD = 0;         
				do 
				{
					//the code within the if statement checks the distance of the point and marks it for removal (setting flag to 1)
					//if it is within m_Spacing of the current point
					//the condition of the if statement stipulates that the point has not yet been marked as a keeper or one to remove (first part)
					//since we have processed all features first, all features will have flags set to 2, so they will not be marked for deletion by this procedure
					if (knnIt->get1i(FLAGS) == 0) 
					{            
						const Vector3f& np = knnIt->get3f( POSITION );
						sqrD = ( p - np ) * ( p - np );
						if (sqrD < sqrSpacing)
						{
							knnIt->set1i( FLAGS, 1 );
							numNewPoints--;
						}
					}
					knnIt->next();
				}while( sqrD < sqrSpacing && !knnIt->atEnd());
			}
		}
		basicIt->next();
	}
	delete knnIt;

	x4Assert(numPointStart == pcCopy->getNumPoints());

	// -1 means point has been removed
	vector<int32> newToOldIndex(numNewPoints, -1);
	vector<int32> oldToNewIndex(numPointStart, -1);

	// setup indices correctly and copy points
	card32 numDeleted = 0;
	pc->clearAndSetup(pcCopy->getDescr(), numNewPoints);
	for (unsigned i=0, newindex=0; i < pcCopy->getNumPoints(); i++)
	{
		if (pcCopy->getPointSet()->get1i(i, FLAGS) != 1)
		{
			newToOldIndex[newindex] = i;
			oldToNewIndex[i] = newindex;
			pc->getPointSet()->setVertex(newindex, pcCopy->getPointSet()->getDataPointer(i));
			newindex++;
		}
		else
			numDeleted++;
	}

	
	// Reset all flags
	basicIt->reset();
	while( !basicIt->atEnd() )
	{
		modIt->set1i( FLAGS, 0 );
		basicIt->next();
	}

	delete it;
	it = NULL;

	x4Assert(numNewPoints == pc->getNumPoints());

	const float32 maxDistance = m_Spacing * min(m_ReconnectFactor, (float32)numPointStart / (float32)numNewPoints);
	// create new topology graph
	if (tpg)
	{
		// this would be our new topology graph
		InCorePCTopologyGraph *resultTg = new InCorePCTopologyGraph();
		resultTg->clearAndSetup(tpg->getEdges()->getDescr());

		// use this iterator to go through the topology graph
		PointCloudIterator* titer = tpg->createIterator(pcCopy->getPointSet());
		TopologicalKNNIterator knn(pcCopy, titer);

		// for each new we have to establish a new topological connection based on previous one
		for (unsigned i=0; i < pc->getNumPoints(); i++)
		{
			int32 searchCenter = newToOldIndex[i];
			Vector3f searchCenterPos = pcCopy->getPointSet()->get3f(searchCenter, POSITION);
			knn.setStartPointAndReset(searchCenter);
			while(!knn.atEnd())
			{
				card32 currentOldPoint = knn.getPointSetPointNumber();
				if (currentOldPoint != searchCenter)
				{
					Vector3f currentOldPointPos = pcCopy->getPointSet()->get3f(currentOldPoint, POSITION);
					float32 distance = norm(currentOldPointPos-searchCenterPos);
					if (distance > maxDistance)
						break;

					if (oldToNewIndex[currentOldPoint] != -1)
						resultTg->addEdge(i, oldToNewIndex[currentOldPoint]);
				}
				knn.next();
			}
		}

		TopologyAttachment* tatt = NULL;
		tatt = new TopologyAttachment;
		tatt->setup(TopologyAttachment::getDefaultName(), AttachedData::ADF_PERSISTENT);
		tatt->setTopology(resultTg);            
		pc->getAttachments()->attachData(tatt);	
	}

	delete pcCopy;
}

void PCCTopologyAndFeaturePreservingResampling::execute( Scene *scene )
{
   vector<string> selection = scene->getRootState()->staticState->getSelectedNodes();
   for ( unsigned int m = 0; m < selection.size(); m++ )
   {
      UnstructuredInCorePointCloud* pc = dynamic_cast<UnstructuredInCorePointCloud*>(getPointCloud( scene, selection[m] ));
      PCCTopologyAndFeaturePreservingResampling::resamplePC(pc);
   }
}