/*  Time-stamp: <@(#)PCIGMOD14Features.cpp   12/9/2012 4:48 PM   Tino Weinkauf>
 *
 *********************************************************************
 *  File    : PCIGMOD14Features.cpp
 *
 *  Project : X4 Animation Reconstruction IV
 *
 *  Package : animRec4
 *
 *  Company : MPII
 *
 *  Author  : Tino Weinkauf                            Date: 12/9/2012
 *
 *  Purpose : Implementation of class PCIGMOD14Features
 *
 *********************************************************************
 * Version History:
 *
 * V 0.10  12/9/2012 16:48:56  TW : First Revision
 *
 *********************************************************************
 */

#include "stdafx.h"

#include "PCIGMOD14Features.h"

#include "CopyObjectProperties.h"
#include "AR4Settings.h"
#include "AR4MLSManifold.h"
#include "AR4DijkstraDistPC3f.h"

#include "ColorTools.h"
#include "InCorePCTopologyGraph.h"
#include "ProgressWindow.h"
#include "Timer.h"
#include "FileDialogs.h"
#include "ObjectAttachment.h"
#include "AttachedFloatSignatures.h"
#include "SimpleAttachment.h"


#include <assert.h>
#include <float.h>

#include "IntrinsicWavesCompute.h"

#include "DebugRenderer.h"

#include <omp.h>


using namespace X4;


IMPLEMENT_X4_CLASS(PCIGMOD14Features, 1)
{
	BEGIN_CLASS_INIT(PCIGMOD14Features)
	ADD_OBJECT_PROP(settings, 0, AR4Settings::getClass(), true)

	//Params for finding seeds
	// - Scale of the intrinsic radius
	ADD_FLOAT32_PROP(IntrinsicRadiusScale, 0)
	ADD_FLOAT32_PROP(MinRadius, 0)
	ADD_FLOAT32_PROP(MaxRadius, 0)
	ADD_FLOAT32_PROP(ClusterRadius, 0)
	ADD_CARD32_PROP(NumCircles, 0)
	ADD_CARD32_PROP(NumLevels, 0)
	ADD_CARD32_PROP(SamplingLevel, 0)
	ADD_CARD32_PROP(NumSecondaryFeatures, 0)
	ADD_INT32_PROP(NumThreads, 0)
	ADD_INT32_PROP(NumSignificantPoints, 0)
	ADD_BOOLEAN_PROP(bEntropyAsDiff, 0)
	ADD_BOOLEAN_PROP(bVisualizeEntropy, 0)

	ADD_NOARGS_METHOD_DESC(PrecomputeSeedFindingDescriptors, "Precompute some data for finding starting points later.")
	
	ADD_NOARGS_METHOD_DESC(UnwrapFeaturePoints, "Unwraps the feature points of the selected mesh and puts it into the scene graph for rendering.")

	//Additional Params for computing geodesics and debugging
	ADD_INT32_PROP(StartIndex, 0)
	ADD_NOARGS_METHOD_DESC(ShowGeodesicArea, "Compute geodesic distances from start index.")
	ADD_NOARGS_METHOD_DESC(ShowBoundary, "Fills color channel with boundary info.")

	

	ADD_NOARGS_METHOD_DESC(FindSeeds, "Find starting points for the growing matcher.")
	
	ADD_NOARGS_METHOD_DESC(ExportCircleLengths, "Export circle length descriptors of the mesh that has been computed last.")

	//add the button here
	ADD_NOARGS_METHOD_DESC(DetectBoundaries, "Sets the edges with only one triangle assosiated with it as boundary edges.")
}


PCIGMOD14Features::PCIGMOD14Features()
{
	settings = new AR4Settings();

	IntrinsicRadiusScale = 0.05f;
	modelScale = 1.f;
	MinRadius = 0.9f; 
	MaxRadius = 1.7f;
	ClusterRadius = 0.9f;
	NumCircles = 10; 
	AveragingHalfWindow = 3; 
	NumSecondaryFeatures = 0;//by default take as many secondaries as we can find
	NumLevels = 1;//by default single-scale

	SamplingLevel = 4;

	NumThreads = omp_get_num_procs();
	NumSignificantPoints = 300;
	bEntropyAsDiff = true;
	bVisualizeEntropy = true;

	StartIndex = -1;
}


PCIGMOD14Features::~PCIGMOD14Features()
{
	delete settings;
	
}



void PCIGMOD14Features::SetupICompute(IntrinsicWavesCompute& iComp, AR4MLSManifold3f* pManifold)
{
	iComp.RadiusScale = IntrinsicRadiusScale;
	iComp.pManifold = pManifold;
	iComp.MinRadius = MinRadius;
	iComp.MaxRadius = MaxRadius;
	iComp.ClusterRadius = ClusterRadius;
	iComp.NumCircles = NumCircles;
	iComp.NumLevels = NumLevels;
	iComp.NumThreads = NumThreads;
}


bool FiniteSensitiveLesser(float32 elem1, float32 elem2)
{
	
	return elem1 < elem2;
}


void PCIGMOD14Features::PrecomputeSeedFindingDescriptors()
{
	//For all selected entries in the scene graph:
	// - (try to) Convert to our AR4 manifold
	// - run seed finder on it

	int ActualNumThreads = NumThreads;
	if (ActualNumThreads < 1)
	{
		ActualNumThreads = omp_get_num_procs() + ActualNumThreads;
	}
	if (ActualNumThreads < 1)
	{
		ActualNumThreads = 1;
	}

	vector<AR4MLSManifold3f*> manifoldsProcess;

	float32 comulativeSurfaceArea = 0, avgSurfaceArea;

	vector<AR4MLSManifold3f*> manifolds;
	vector<string> SelectedSGNodes = scene->getRootState()->staticState->getSelectedNodes();
	for (unsigned int node=0;node<SelectedSGNodes.size();node++)
	{
		//retrieve selected point cloud
		PointCloud* pc = getPointCloud(scene, SelectedSGNodes[node]);

		//create MLS manifold from point cloud
		AR4MLSManifold3f* pManifold = checkForManifold(getPointCloud(scene, SelectedSGNodes[node]));

		//TODO: Setting eps and so on
		if (pManifold == NULL)
		{
			pManifold = AR4MLSManifold3f::createFromPointCloud(getPointCloud(scene, SelectedSGNodes[node]), settings->numNN, settings->maxSampleSpacing);
			if (pManifold != NULL)
				manifolds.push_back(pManifold);
		}
		
		if (!pManifold) continue;

		manifoldsProcess.push_back(pManifold);

		
		//adds the average edge length of all the nodes selected.
		comulativeSurfaceArea += pManifold->getSurfaceArea();
		
	}
	//average the the average edge lengths of all the selected nodes.
	avgSurfaceArea = comulativeSurfaceArea / SelectedSGNodes.size();

	//use this as measure of the scale of the collection of models
	modelScale = sqrt( avgSurfaceArea / M_PI );

	//now re-scale radius parameters
	float32 irs = IntrinsicRadiusScale;
	IntrinsicRadiusScale *= modelScale;

	const card32 numCirclesTotal = NumLevels * NumCircles;

	//progressWindow->pushStep(true, "Find Seeds", SelectedSGNodes.size());
	for (mpcard node = 0; node < manifoldsProcess.size(); node++)
	{
		
		//retrieve selected manifold and point cloud
		AR4MLSManifold3f* pManifold = manifoldsProcess[node];
		PointCloud* pc = (PointCloud*)pManifold->getPointCloud();

		//Get Output
		AllRawCircleLengths.clear();
		AllRawCircleLengths.resize(pManifold->getNumDiscrPoints());

		IntrinsicWavesCompute comp;
		SetupICompute(comp, pManifold);

		const int64 iNumPoints = pManifold->getNumDiscrPoints();
		const mpcard NumPoints = pManifold->getNumDiscrPoints();

		X4_TIMER_START(IntrinsicWavesForAll);
		#pragma omp parallel for schedule(dynamic, 10) num_threads(ActualNumThreads)
		for(int64 i=0;i<iNumPoints;i++)
		{
			comp.ComputeWaves(i, AllRawCircleLengths[i]);
		}
		X4_TIMER_END(IntrinsicWavesForAll);
		
		//Compute scalar field highlighting most distinguishable points
		vector< float32 > Entropy;
		if (bEntropyAsDiff)
		{
			X4_TIMER_START(ComputeEntropyAsDiff);
			comp.ComputeEntropyAsDiff(AllRawCircleLengths, Entropy);
			X4_TIMER_END(ComputeEntropyAsDiff);
		}
		else
		{
			X4_TIMER_START(ComputeEntropy);
			comp.ComputeEntropy(AllRawCircleLengths, Entropy);
			X4_TIMER_END(ComputeEntropy);
		}

		//Get min/max
		float32 EntropyMin(FLT_MAX);
		float32 EntropyMax(-FLT_MAX);
		for(mpcard i=0;i<NumPoints;i++)
		{
			if (_finite(Entropy[i]) && EntropyMin > Entropy[i]) EntropyMin = Entropy[i];
			if (_finite(Entropy[i]) && EntropyMax < Entropy[i]) EntropyMax = Entropy[i];
		}

		//Visualize scalar field?
		if (bVisualizeEntropy)
		{
			UnstructuredInCorePointCloud* pc = pManifold->getPointCloud();
			checkAttribute(pc, "color", 3, VAD::DATA_FORMAT_FLOAT32);
			pManifold->resetup(pc); // reassign pointers
			AAT distColAAT = pc->getAAT("color");
			for (mpcard i=0;i<NumPoints;i++)
			{
				if (_finite(Entropy[i]))
				{
					const Vector3f col = rainbowGradient3f((Entropy[i] - EntropyMin) / (EntropyMax - EntropyMin));
					pc->getPointSet()->set3f(i, distColAAT, col);
				}
				else
				{
					pc->getPointSet()->set3f(i, distColAAT, makeVector3f(1,1,1));
				}
			}

			pc->deleteDelOnWriteAttachments();
		}

		//Identify the most significant points wrt. entropy
		vector< mpcard > MostSignificantMaxima;
		
		X4_TIMER_START(ComputeMostSignificantFeaturePoints);
		comp.ComputeMostSignificantFeaturePoints(Entropy, NumSignificantPoints, MostSignificantMaxima);
		X4_TIMER_END(ComputeMostSignificantFeaturePoints);
		

		X4_TIMER_START(EnsureSamplingDensity);
		comp.EnsureSamplingDensity(Entropy, pManifold->getAvgEdgeLength() * (float32)(1 << SamplingLevel), MostSignificantMaxima);
		X4_TIMER_STOP(EnsureSamplingDensity);

		vector<vector<mpcard>> secondaryFeatures(MostSignificantMaxima.size());
		vector<vector<float>> localEntropies(MostSignificantMaxima.size());
		if (NumSecondaryFeatures > 0)
		{
			//compute secondary features near each "primary" feature
			//these are used later to determine corresponding directions on the two surfaces
			X4_TIMER_START(ComputeSecondaryFeatures);
			for (mpcard i = 0; i < MostSignificantMaxima.size(); i++)
			{
				comp.ComputeSecondaryFeatures(MostSignificantMaxima[i], NumSecondaryFeatures, AllRawCircleLengths, secondaryFeatures[i], localEntropies[i]);
			}
			X4_TIMER_END(ComputeSecondaryFeatures);
		}

		//Add the descriptor and the entropy as attachments
		// for the n most significant points
		const mpcard NumFeaturePoints = MostSignificantMaxima.size();

		//Get attachment with the feature point set inside
		UnstructuredInCorePointCloud* pPointCloud = pManifold->getPointCloud();
		assert(pPointCloud);
		UnstructuredInCorePointCloud* pFeaturePoints = NULL;
		ObjectAttachment* pAttachment = dynamic_cast< ObjectAttachment* >(pPointCloud->getAttachments()->getData("FeaturePoints"));
		bool bpAttachmentWasAlreadyThere(true);
		if (!pAttachment)
		{
			//Create a new one and place a point set inside it
			pAttachment = new ObjectAttachment();
			pAttachment->setup("FeaturePoints", AttachedData::ADF_PERSISTENT);
			pFeaturePoints = new UnstructuredInCorePointCloud();
			pAttachment->setObject(pFeaturePoints);
			bpAttachmentWasAlreadyThere = false;
		}
		else
		{
			//Get the existing one, which we will overwrite in the following.
			pFeaturePoints = dynamic_cast< UnstructuredInCorePointCloud* >(pAttachment->getObject());
			if (!pFeaturePoints)
			{
				pFeaturePoints = new UnstructuredInCorePointCloud();
				pAttachment->replaceObject(pFeaturePoints);
			}
		}

		//Init the feature point set
		pFeaturePoints->clearAndSetup(pPointCloud->getDescr(), NumFeaturePoints);
	
		//Write into attachment - copy points
		for(mpcard i=0;i<NumFeaturePoints;i++)
		{
			pFeaturePoints->getPointSet()->setVertex(i, pPointCloud->getPointSet()->getDataPointer(MostSignificantMaxima[i]));
		}

		mpcard idx2, actualSecondaryCount = 0;
		UnstructuredInCorePointCloud* pSecondaryPoints = NULL;
		ObjectAttachment* pSecondObjAtt = NULL;
		bool bSecondaryPointsAlreadyThere = false;
		if (NumSecondaryFeatures > 0)
		{
			//create a point cloud for secondary features
			pSecondObjAtt = dynamic_cast< ObjectAttachment* >(pPointCloud->getAttachments()->getData("SecondaryFeaturePoints"));
			bSecondaryPointsAlreadyThere = true;
			if (pSecondObjAtt == NULL)
			{
				pSecondObjAtt = new ObjectAttachment();
				pSecondObjAtt->setup("SecondaryFeaturePoints", AttachedData::ADF_PERSISTENT);
				pSecondaryPoints = new UnstructuredInCorePointCloud();
				pSecondObjAtt->setObject(pSecondaryPoints);
				bSecondaryPointsAlreadyThere = false;
			}
			else
			{
				pSecondaryPoints = dynamic_cast< UnstructuredInCorePointCloud* >(pAttachment->getObject());
				if (pSecondaryPoints == NULL)
				{
					pSecondaryPoints = new UnstructuredInCorePointCloud();
					pSecondObjAtt->replaceObject(pSecondaryPoints);
				}
			}

			//keep attributes from original point cloud
			//compute number of points needed
			for (mpcard i = 0; i < NumFeaturePoints; i++)
			{
				actualSecondaryCount += secondaryFeatures[i].size();
			}
			pSecondaryPoints->clearAndSetup(pPointCloud->getDescr(), actualSecondaryCount);
			idx2 = 0;
			for (mpcard i = 0; i < NumFeaturePoints; i++)
			{
				for (mpcard j = 0; j < secondaryFeatures[i].size(); j++)
				{
					pSecondaryPoints->getPointSet()->setVertex(idx2, pPointCloud->getPointSet()->getDataPointer(secondaryFeatures[i][j]));
					idx2++;
				}
			}
		}

		//create attachment for feature parameters
		AttachedCard32Vector* pParamAttachment = dynamic_cast< AttachedCard32Vector* >(pFeaturePoints->getAttachments()->getData("FeatureParameters"));
		bool bParamAttachmentThere(true);
		if (pParamAttachment == NULL)
		{
			pParamAttachment = new AttachedCard32Vector();
			pParamAttachment->setup("FeatureParameters", AttachedData::ADF_PERSISTENT);
			bParamAttachmentThere = false;
		}

		pParamAttachment->data.resize(2);
		pParamAttachment->data[0] = NumCircles;
		pParamAttachment->data[1] = NumLevels;

		//Create storage for descriptor
		AttachedFloatVector* pDescrAttachment = dynamic_cast< AttachedFloatVector* >(pFeaturePoints->getAttachments()->getData("IntrinsicWavesDescriptor"));
		bool bpDescrAttachmentWasAlreadyThere(true);
		if (!pDescrAttachment)
		{
			pDescrAttachment = new AttachedFloatVector();
			pDescrAttachment->setup("IntrinsicWavesDescriptor", AttachedData::ADF_PERSISTENT);
			bpDescrAttachmentWasAlreadyThere = false;
		}

		AttachedCard32Vector* pNumSecondAtt = NULL;
		AttachedFloatVector* pDescrSecondAtt = NULL;
		bool bSecondCountAlreadyThere = false;
		bool bDescrSecondAlreadyThere = false;
		if (NumSecondaryFeatures > 0)
		{
			//create storage for secondary feature descriptors
			//first the number of secondary features for each primary
			pNumSecondAtt = dynamic_cast< AttachedCard32Vector* >(pFeaturePoints->getAttachments()->getData("SeconaryFeatureCount"));
			bSecondCountAlreadyThere = true;
			if (pNumSecondAtt == NULL)
			{
				pNumSecondAtt = new AttachedCard32Vector();
				pNumSecondAtt->setup("SecondaryFeatureCount", AttachedData::ADF_PERSISTENT);
				bSecondCountAlreadyThere = false;
			}

			//now descriptors of secondary features themselves
			pDescrSecondAtt = dynamic_cast< AttachedFloatVector* >(pFeaturePoints->getAttachments()->getData("IntrinsicWavesDescriptor_secondary"));
			bDescrSecondAlreadyThere = true;
			if (pDescrSecondAtt == NULL)
			{
				pDescrSecondAtt = new AttachedFloatVector();
				pDescrSecondAtt->setup("IntrinsicWavesDescriptor_secondary", AttachedData::ADF_PERSISTENT);
				bDescrSecondAlreadyThere = false;
			}
		}

		//Write into attachment - add descriptor to each point
		//Write into attachment - set entropy
		checkAttribute(pFeaturePoints, "entropy", 1, VAD::DATA_FORMAT_FLOAT32);
		AAT EntropyAAT = pFeaturePoints->getAAT("entropy");

		//abrunton: add point index field to feature points for backwards reference to original point set
		checkAttribute(pFeaturePoints, "index", 1, VAD::DATA_FORMAT_CARD32);
		AAT indexAAT = pFeaturePoints->getAAT("index");

		//get color AAT
		AAT colorAAT = pFeaturePoints->getAAT("color");

		AAT Entropy2AAT, index2AAT, color2AAT;
		if (NumSecondaryFeatures > 0)
		{
			//same attributes for secondary features
			checkAttribute(pSecondaryPoints, "entropy", 1, VAD::DATA_FORMAT_FLOAT32);
			Entropy2AAT = pSecondaryPoints->getAAT("entropy");
			checkAttribute(pSecondaryPoints, "index", 1, VAD::DATA_FORMAT_CARD32);
			index2AAT = pSecondaryPoints->getAAT("index");
			//get color AAT
			color2AAT = pSecondaryPoints->getAAT("color");
		}

		//compute size of attachments
		//primary descriptors
		pDescrAttachment->data.resize(MostSignificantMaxima.size() * numCirclesTotal);//assume always used the same number of radii

		if (NumSecondaryFeatures > 0)
		{
			//number of secondary features for each primary
			pNumSecondAtt->data.resize(MostSignificantMaxima.size());
			//secondary descriptors  
			pDescrSecondAtt->data.resize(actualSecondaryCount * numCirclesTotal);
		}

		//write in the values
		//iterator for primary descriptors
		vector<float32>::iterator attDataIt = pDescrAttachment->data.begin();
		//iterator for secondary descriptors
		vector<float32>::iterator attData2It;
		if (NumSecondaryFeatures > 0)
			attData2It = pDescrSecondAtt->data.begin();

		//counter for secondary descriptors
		idx2 = 0;
		for(mpcard i=0;i<NumFeaturePoints;i++)
		{
			// - add descriptor
			attDataIt = std::copy(AllRawCircleLengths[MostSignificantMaxima[i]].begin(), AllRawCircleLengths[MostSignificantMaxima[i]].end(), attDataIt);

			// - add entropy
			pFeaturePoints->getPointSet()->set1f(i, EntropyAAT, Entropy[MostSignificantMaxima[i]]);

			//abrunton
			// - add point index
			pFeaturePoints->getPointSet()->set1u(i, indexAAT, (card32)MostSignificantMaxima[i]);

			if (NumSecondaryFeatures > 0)
			{
				//secondary feature stuff
				//first the number for this primary
				pNumSecondAtt->data[i] = (card32)secondaryFeatures[i].size();

				//now descriptors themselves and point cloud attributes
				for (mpcard j = 0; j < secondaryFeatures[i].size(); j++)
				{
					//local entropy
					pSecondaryPoints->getPointSet()->set1f(idx2, Entropy2AAT, localEntropies[i][j]);
					//index to original point cloud
					pSecondaryPoints->getPointSet()->set1u(idx2, index2AAT, (card32)secondaryFeatures[i][j]);
					//set color to the same as the primary
					pSecondaryPoints->getPointSet()->set3f(idx2, color2AAT, pFeaturePoints->getPointSet()->get3f(i, colorAAT));
					//descriptor
					attData2It = std::copy(AllRawCircleLengths[secondaryFeatures[i][j]].begin(), AllRawCircleLengths[secondaryFeatures[i][j]].end(), attData2It);
					//increment counter
					idx2++;
				}
			}
		}

		//attach the feature parameter attachment
		if (!bParamAttachmentThere) pFeaturePoints->getAttachments()->attachData(pParamAttachment);

		//Set descriptor attachment into the attachment
		if (!bpDescrAttachmentWasAlreadyThere) pFeaturePoints->getAttachments()->attachData(pDescrAttachment);

		//Set the global feature point attachment
		if (!bpAttachmentWasAlreadyThere) pPointCloud->getAttachments()->attachData(pAttachment);

		//attach secondary feature points
		if (!bSecondaryPointsAlreadyThere) pPointCloud->getAttachments()->attachData(pSecondObjAtt);

		//attach numbers of secondary features for each primary to primary feature point cloud
		if (!bSecondCountAlreadyThere) pFeaturePoints->getAttachments()->attachData(pNumSecondAtt);

		//attach descriptors of secondary featrues to primary feature point cloud
		if (!bDescrSecondAlreadyThere) pFeaturePoints->getAttachments()->attachData(pDescrSecondAtt);

		//TODO: What is this good for and why am I setting this for all attachments?
		pPointCloud->deleteDelOnWriteAttachments();
	}

	//reset intrinsic radius scale for interface
	IntrinsicRadiusScale = irs;

	
}


void PCIGMOD14Features::UnwrapFeaturePoints()
{
	vector<string> SelectedSGNodes = scene->getRootState()->staticState->getSelectedNodes();
	for (unsigned int node=0;node<SelectedSGNodes.size();node++)
	{
		UnstructuredInCorePointCloud* pPointCloud = dynamic_cast< UnstructuredInCorePointCloud* >(getPointCloud(scene, SelectedSGNodes[node]));
		if (!pPointCloud) continue;

		ObjectAttachment* pAttachment = dynamic_cast< ObjectAttachment* >(pPointCloud->getAttachments()->getData("FeaturePoints"));
		if (pAttachment)
		{
			addPointCloud(scene, dynamic_cast< UnstructuredInCorePointCloud* >(pAttachment->getObject()->copy()), SelectedSGNodes[node] + "_FeaturePoints");
		}

		ObjectAttachment* pAttSecond = dynamic_cast<ObjectAttachment*>(pPointCloud->getAttachments()->getData("SecondaryFeaturePoints"));
		if (pAttSecond)
		{
			addPointCloud(scene, dynamic_cast<UnstructuredInCorePointCloud*>(pAttSecond->getObject()->copy()), SelectedSGNodes[node] + "_SecondaryFeaturePoints");
		}
	}
}


void PCIGMOD14Features::FindSeeds()
{
	return;
}

		


void PCIGMOD14Features::ShowGeodesicArea()
{
	const int32 idxStart = (StartIndex < 0) ? 0 : StartIndex;

	vector<string> SelectedSGNodes = scene->getRootState()->staticState->getSelectedNodes();
	for (unsigned int node=0;node<SelectedSGNodes.size();node++)
	{
		
		AR4MLSManifold3f* pManifold = checkForManifold(getPointCloud(scene, SelectedSGNodes[node]));

		if (pManifold == NULL)
		{
			pManifold = AR4MLSManifold3f::createFromPointCloud(getPointCloud(scene, SelectedSGNodes[node]), settings->numNN, settings->maxSampleSpacing);
			if (pManifold != NULL)
				manifolds.push_back(pManifold);
		}

		if (!pManifold) continue;

		const mpcard NumPoints = pManifold->getNumDiscrPoints();
		if (idxStart >= NumPoints) continue;

		const float32 actualMinRadius = MinRadius * IntrinsicRadiusScale * modelScale;
		const float32 actualMaxRadius = MaxRadius * IntrinsicRadiusScale * modelScale * (float32)(1 << (NumLevels - 1));
		AR4DijkstraDistPC3f Dijkstra(pManifold);
		Dijkstra.initialize();
		Dijkstra.computeUpToDistance(idxStart, actualMaxRadius);

		UnstructuredInCorePointCloud* pc = pManifold->getPointCloud();
		checkAttribute(pc, "color", 3, VAD::DATA_FORMAT_FLOAT32);
		pManifold->resetup(pc); // reassign pointers
		AAT distColAAT = pc->getAAT("color");
		for (mpcard i=0; i<NumPoints; i++)
		{
			Vector3f col = rainbowGradient3f((Dijkstra.evaluate(i) - actualMinRadius) / Dijkstra.getMaxValue());
			pc->getPointSet()->set3f(i, distColAAT, col);
		}

		pc->deleteDelOnWriteAttachments();
	}
}


void PCIGMOD14Features::ShowBoundary()
{
	vector<string> SelectedSGNodes = scene->getRootState()->staticState->getSelectedNodes();
	for (unsigned int node=0;node<SelectedSGNodes.size();node++)
	{
		AR4MLSManifold3f* pManifold = checkForManifold(getPointCloud(scene, SelectedSGNodes[node]));

		if (pManifold == NULL)
		{
			pManifold = AR4MLSManifold3f::createFromPointCloud(getPointCloud(scene, SelectedSGNodes[node]), settings->numNN, settings->maxSampleSpacing);
			if (pManifold != NULL)
				manifolds.push_back(pManifold);
		}

		if (!pManifold) continue;

		UnstructuredInCorePointCloud* pc = pManifold->getPointCloud();
		if (!pc->providesAttribute("boundary")) continue;

		//Get data channel for boundary of the mesh
		AAT aatBoundary = pc->getAAT("boundary");
		PointSet* pRawPointSet = pc->getPointSet();

		//Get or create color channel
		checkAttribute(pc, "color", 3, VAD::DATA_FORMAT_FLOAT32);
		pManifold->resetup(pc); // reassign pointers
		AAT distColAAT = pc->getAAT("color");

		const mpcard NumPoints = pManifold->getNumDiscrPoints();
		for (mpcard i=0;i<NumPoints;i++)
		{
			const float BoundaryValue = pRawPointSet->get1f(i, aatBoundary);

			if (BoundaryValue > 0.5f)
			{
				//At boundary
				pRawPointSet->set3f(i, distColAAT, makeVector3f(0.9f, 0.4f, 0.4f));
			}
			else
			{
				//Not at boundary
				pRawPointSet->set3f(i, distColAAT, makeVector3f(0.6f, 0.6f, 0.6f));
			}
		}

		pc->deleteDelOnWriteAttachments();
	}
}


void PCIGMOD14Features::ExportCircleLengths()
{
	string filename = FileDialogs::getSaveFileName(NULL, "Export Circle Length Descriptors", "*.sls", NULL, "", FileDialogs::EXPORT_DIR, "AllRawCircleLengths.sls");

	if (filename != "")
	{
		FILE* file = fopen(filename.c_str(), "wt");

		fprintf(file, "SeedingLS\nVersion 1\n\n\nNumOfSeedingLines %d\nNumOfDataPerPoint 0\n", AllRawCircleLengths.size());
		for(int i=0;i<AllRawCircleLengths.size();i++)
		{
			fprintf(file, "\nLinePoints %d\n", AllRawCircleLengths[i].size());
			for(int j=0;j<AllRawCircleLengths[i].size();j++)
			{
				fprintf(file, "%d %g 0\n", j, AllRawCircleLengths[i][j]);
			}
		}

		fclose(file);
	}
}



void PCIGMOD14Features::DetectBoundaries()
{	
	vector<string> SelectedSGNodes = scene->getRootState()->staticState->getSelectedNodes();

	for (unsigned int node=0;node<SelectedSGNodes.size();node++)
	{
		AR4MLSManifold3f* pManifold = checkForManifold(getPointCloud(scene, SelectedSGNodes[node]));

		if (pManifold == NULL)
		{
			pManifold = AR4MLSManifold3f::createFromPointCloud(getPointCloud(scene, SelectedSGNodes[node]), settings->numNN, settings->maxSampleSpacing);
			if (pManifold != NULL)
				manifolds.push_back(pManifold);
		}
	
		if (pManifold == NULL)
		{
			debugOutput << "no pManifold for the node: %ud" << node << "\n";
			continue;
		}
		
		UnstructuredInCorePointCloud* pc = pManifold->getPointCloud();
		UnstructuredInCoreTriangleMesh* utm = dynamic_cast<UnstructuredInCoreTriangleMesh*>(pc);

		pManifold->detectBoundaries();
	}
}
